/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file ANIMATE.C The main entry point for the animation module   */

/* This file contains the module's message loop and the code for  */
/* the main windows' window functions and the command dispatcher  */

#define MODULE_ANIMATE 1

#include "animate.h"

#include "callall.h"

#define NFILESLIST 5
#define LASTFILECMD 31000

typedef struct tagLASTFILESLIST{
  BOOL present;
  char file[256];
} LASTFILESLIST;

static VOID AnimatorMenuCmd(int,HWND,long);
static void ExecuteKeyboardKeyCommand(int key);
static BOOL SetTriViewCursor(BOOL type);
static void UpdateLastFilesList(char *newname);

extern HINSTANCE __cdecl ResourceX(void);

HBITMAP hBmBackground=NULL;
UINT AnimationTimer=0;

static HWND hWndHelp=NULL;
static BOOL bTrackMouse=FALSE,bTrackIcon=FALSE,bToolCaptured=FALSE,
            bSpaceDown=FALSE,bTrackPan=FALSE,bCtrlDown=FALSE,
            bShiftDown=FALSE,bAltDown=FALSE,bResizing=FALSE;
static int  ScreenXsize,ScreenYsize;
static UINT screen_ready=0;
static int  view_dragging=0,trap_scroll_cursor=0;
static int lmessage_table[2]={IDS_SWAPSVIEW,IDS_SWAP3VIEW};
static int tmessage_table[MAXTOOLS]={
    IDS_SELECTTOOL,IDS_NODESELECTOR1,IDS_PANTOOL,IDS_INZOOMTOOL,
    IDS_GRABBERTOOL,IDS_ROTATORTOOL,IDS_SIZERTOOL,
    IDS_TRACKER,IDS_PATHEDITORTOOL,IDS_SKELETORTOOL
                                    };
static char gszSTGsaved[256]={"unnamed.ofx"};
static LASTFILESLIST LastFilesList[NFILESLIST];
static int nLastFiles=0;

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 int i;
 MSG msg;
 HWND hWndPrev;
 HANDLE hAccel,hNewAccel;
#include "..\timeout.c"
 hWndPrev = FindWindow(szMainClass, NULL);
 if (hWndPrev != NULL){
   SetForegroundWindow(hWndPrev);
   BringWindowToTop(hWndPrev);
   ShowWindow (hWndPrev,SW_SHOW);
   return FALSE;
 }
 hInstance=ResourceX();
 InitCommonControls();
 debug=NULL;
 strcpy(ruler_name,"None");
 ghdc_triview_Bitmap[0]=ghdc_triview_Bitmap[1]=
 ghdc_triview_Bitmap[2]=ghdc_view_Bitmap=NULL;
 ghbm_triview[0]=ghbm_triview[1]=ghbm_triview[2]=ghbm_view=NULL;
 gCurTool=gCurCommand=gCurSelect= -1;
 if(!hPrevInstance){
   WNDCLASS wc;
   BOOL     rc;
   if((ghpaletteScreen=CreateAppPalette(&ScreenXsize,&ScreenYsize)) == NULL)
     return FALSE;
   wc.lpszClassName = szMainClass;
   wc.style = CS_HREDRAW | CS_VREDRAW;
   wc.lpfnWndProc = (WNDPROC)MainAnimatorWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = LoadIcon(hInstance,"ANIMATORICON");
   wc.hCursor = LoadCursor(NULL,IDC_ARROW);
   wc.hbrBackground = NULL;
   wc.lpszMenuName = MAKEINTRESOURCE(ANIMATOR_MENU);
   rc = RegisterClass( &wc );
   if(!rc)return FALSE;
   wc.lpszClassName = szTriViewClass;
   wc.style = CS_DBLCLKS;
   wc.lpfnWndProc = (WNDPROC)AnimatorTriViewWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL,IDC_ARROW);
   wc.hbrBackground = NULL;
   wc.lpszMenuName = NULL;
   rc = RegisterClass( &wc );
   if(!rc)return FALSE;
   wc.lpszClassName = szViewClass;
   wc.style = 0;
   wc.lpfnWndProc = (WNDPROC)AnimatorViewWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL,IDC_ARROW);
   wc.hbrBackground = NULL;
   wc.lpszMenuName = NULL;
   rc = RegisterClass( &wc );
   if(!rc)return FALSE;
   wc.lpszClassName = szToolboxClass;
   wc.style = 0;
   wc.lpfnWndProc = (WNDPROC)ToolboxWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = NULL; //GetStockObject(LTGRAY_BRUSH);
   wc.lpszMenuName = NULL;
   if(!RegisterClass(&wc))return FALSE;
   wc.lpszClassName = szToolBtnClass;
   wc.style = CS_DBLCLKS;
   wc.lpfnWndProc = (WNDPROC)ToolBtnWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = NULL; //(HBRUSH)(COLOR_WINDOW + 1);
   wc.lpszMenuName = NULL;
   if(!RegisterClass(&wc))return FALSE;
   wc.lpszClassName = szQuicktimeClass;
   wc.style = 0;
   wc.lpfnWndProc = (WNDPROC)QuicktimeWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL,IDC_ARROW);
   wc.hbrBackground = GetStockObject(DKGRAY_BRUSH);
   wc.lpszMenuName = NULL;
   rc = RegisterClass( &wc );
   if(!rc)return FALSE;
 }
 if((hLibSpin=LoadLibrary("spin.dll")) == NULL){
   SendPrgmQuery(IDQ_FAILEDSPIN,3);
 }
 if(!RegisterNoAlphaWndClass(hInstance)){
   SendPrgmQuery(IDQ_FAILEDTIDE,3);
 }
 ghinst_main=hInstance;
 LoadPreferences();
 LoadArrowIcons();
 /* WARNING no Message Boxes until Window LOOP */
 if(!SetupMainWindow(ghinst_main,nCmdShow))return FALSE;
 ActivatePreferences();
 if(!AnimatorStartup(&screen_ready))return FALSE;
 if(setjmp(j_buf) != 0){
   SendPrgmQuery(IDQ_FATALERROR,3);
   return 0;
 }
 hAccel=LoadAccelerators(ghinst_main,MAKEINTRESOURCE(ANIMATOR_ACCEL));
 if(hAccel == NULL){
   SendPrgmQuery(IDQ_FAILEDACCELS,3);
   return 0;
 }
 hNewAccel=hAccel;
 ghwndOpenGLview=NULL;
 if(EnableGLview() && opengl_visible)StartGLview(TRUE);
 if(StartedInAnimator){
   if(GetCommandLine() != NULL){
     int X__argc;
     long l;
     char *cp,*cp1,cmdl[256];
     strcpy(cmdl,GetCommandLine());
     AnsiLower(cmdl);
     if((cp=strstr(cmdl,".exe")) != NULL){
       if((cp=strchr(cp,' ')) != NULL){
         cp++;
         if((cp1=strchr(cp,'"')) != NULL)cp=cp1+1;
         if((cp1=strchr(cp,'"')) != NULL)*cp1=0;
         l=strlen(cp);
         if(l < 2)X__argc=0;
       }
       else X__argc=0;
     }
     else X__argc=0;
     if(X__argc > 1){
       WritePrivateProfileString("CALLBACK","FILE",cp,IniFilename);
       PostMessage(ghwnd_main,WM_COMMAND,IDM_FILE_LOAD_DIRECT,0);
     }
     else  LogOnMessage();
   }
   else  LogOnMessage();
   {
     char pref_id[32];
     GetPrivateProfileString(IniSection,"KEYVISIBLE","$$$",
                             pref_id,sizeof(pref_id),IniFilename);
     if(strncmp(pref_id,"YES",3) == 0)PostMessage(ghwnd_main,
        WM_COMMAND,(WPARAM)IDM_FILE_KEYFRAMER,(LPARAM)0);
   }
 }
 while (GetMessage(&msg,NULL,0,0)){
   if(!IsWindow(ghwndCoord1) || !IsDialogMessage(ghwndCoord1,&msg)){
     if(!IsWindow(ghwndDlgTools) || !IsDialogMessage(ghwndDlgTools,&msg)){
     if(!IsWindow(hWndHelp) || !IsDialogMessage(hWndHelp,&msg)){
       if(!TranslateAccelerator(ghwnd_main,hNewAccel,&msg)){
         TranslateMessage(&msg);
         DispatchMessage(&msg);
       }
     }
     }
   }
 }
 DisengageTool();
 if(debug != NULL)fclose(debug);
 if(IsWindow(ghwndTimeline) && IsWindowVisible(ghwndTimeline))
   WritePrivateProfileString(IniSection,"KEYVISIBLE","YES",IniFilename);
 else
   WritePrivateProfileString(IniSection,"KEYVISIBLE","NO",IniFilename);
 DestroyWindow(ghwnd_main);
 AnimatorShutdown();
 if(hLibSpin != NULL)FreeLibrary(hLibSpin);
 return (msg.wParam);
}

//
// The programs main window WINDOWPROC this windows contains the 3 views,
// the camera view window, the toolpannel window and the status display.
//

LRESULT CALLBACK MainAnimatorWndProc(HWND hwnd, UINT msg, WPARAM wparam,
                                    LPARAM lparam ){
 static BOOL slider_lock=FALSE,bCaptured=FALSE;
 static int  last_x,last_y,type=0;
 static RECT rc;
 int i,oldi,x,y,ltype;
 switch( msg ){
   case WM_CALLBACK_RENDER_GL_SWITCH:
     /* call to this point to switch from the camera view */
     if(View == TRIVIEW){
       if(!BigCameraView)PostMessage(ghwnd_main,
               WM_COMMAND,(WPARAM)IDM_WINDOW_BIGCAMERA,0);
       PostMessage(ghwnd_main,WM_COMMAND,
               (WPARAM)IDM_WINDOW_SWITCHWINDOWS,0);
     }
     else{
       PostMessage(ghwnd_main,WM_COMMAND,
               (WPARAM)IDM_WINDOW_SWITCHWINDOWS,0);
       if(BigCameraView)PostMessage(ghwnd_main,
               WM_COMMAND,(WPARAM)IDM_WINDOW_BIGCAMERA,0);
     }
     break;
   case WM_CALLBACK_RENDER_GL: /* OPEN_GL call to update script for DLL */
     UpdateGLview(TRUE);
     break;
   case WM_CREATE:{
       HMENU hmenu=GetMenu(hwnd);
#if __DEMO__
       EnableMenuItem(hmenu,IDM_FILE_SAVE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_SAVE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_EXP_ASCRIPT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_EXP_FSCRIPT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_PRINT,MF_GRAYED);
#endif
#if __LITE__
       EnableMenuItem(hmenu,IDM_FILE_EXP_VRML,MF_GRAYED);
#endif
     }
     if(bBackgroundImage){
       hBmBackground=LoadImage(ghinst_main,"animator.bmp",IMAGE_BITMAP,
                               0,0,LR_LOADFROMFILE);
       if(hBmBackground == NULL)
         hBmBackground=LoadBitmap(ghinst_main,"screenbitmap");
     }
     break;
   case WM_DESTROY:
     if(hBmBackground != NULL)DeleteObject(hBmBackground);
     hBmBackground=NULL;
     break;
   case WM_ACTIVATE:
     if(LOWORD(wparam) == WA_INACTIVE && (
       (HWND)lparam == ghwndCoord1 ||
       (HWND)lparam == ghwndDlgTools))
     SendMessage(hwnd,WM_NCACTIVATE,(WPARAM)TRUE,0);
     else return( DefWindowProc( hwnd, msg, wparam, lparam ) );
     break;
   case WM_ACTIVATEAPP:
     PostMessage(hwnd,WM_NCACTIVATE,wparam,0);
     break;
   case WM_SETCURSOR:
     if(((HWND)wparam == ghwnd_main ||
         (HWND)wparam == ghwnd_xscl) && current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
   case WM_MENUSELECT:{
       HMENU hmenu;
       UINT fuFlags,uItem,id,i;
       if(bSpaceDown){
         bSpaceDown=FALSE; bCtrlDown=FALSE; bShiftDown=FALSE; bAltDown=FALSE;
       }
       hmenu=(HMENU)lparam;
       fuFlags=(UINT)HIWORD(wparam);
       uItem=(UINT)LOWORD(wparam);
       id=0;
       if(fuFlags == 0xFFFF)id=0;
       else if(fuFlags & MF_POPUP){
       }
       else{
         id=CheckMenuHelpData(uItem);
       }
       DisplayMenuHelpData(ghwnd_info,id);
     }
     break;
   case WM_LBUTTONDOWN:
     x=(short)LOWORD(lparam); y=(short)HIWORD(lparam);
     type=0;
     if(View == TRIVIEW){
       if(x > WindowLeftOrigin+Window0sizeX - 20 &&
          x < WindowLeftOrigin+Window0sizeX + 20 &&
          y > WindowTopOrigin+Window0sizeY  - 20 &&
          y < WindowTopOrigin+Window0sizeY  + 20)type=1;
       else if(x > WindowLeftOrigin+Window0sizeX - 10 &&
               x < WindowLeftOrigin+Window0sizeX + 10 &&
               y > WindowTopOrigin+10 &&
               y < WindowTopOrigin+TriviewSpaceY)type=2;
       else if(y > WindowTopOrigin+Window0sizeY - 10 &&
               y < WindowTopOrigin+Window0sizeY + 10 &&
               x > WindowLeftOrigin+10 &&
               x < WindowLeftOrigin+TriviewSpaceX-10)type=3;
     }
     if     (type == 1)SetCursor(LoadCursor(NULL,IDC_SIZEALL));
     else if(type == 2)SetCursor(LoadCursor(NULL,IDC_SIZEWE));
     else if(type == 3)SetCursor(LoadCursor(NULL,IDC_SIZENS));
     if(type > 0){
       HDC hdc;
       hdc=GetDC(ghwnd_main);
       FocusActiveWindow(hdc,FALSE);
       ReleaseDC(ghwnd_main,hdc);
       SetCapture(hwnd);
       DeleteMemoryWindows();
       bResizing=TRUE;
       bCaptured=TRUE;
       GetClientRect(hwnd,&rc);
       last_x=x; last_y=y;
     }
     break;
   case WM_LBUTTONUP:
     if(bCaptured){
       HDC hdc;
       hdc=GetDC(ghwnd_main);
       FocusActiveWindow(hdc,TRUE);
       ReleaseDC(ghwnd_main,hdc);
       ReleaseCapture();
       bCaptured=FALSE;
       bResizing=FALSE;
       CreateMemoryWindows();
       InvalidateRect(ghwndQuicktime,NULL,TRUE);
       UpdateWindow(ghwndQuicktime);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview != NULL)UpdateGLview(TRUE);
       else PerspectiveView(0,0);
       type=0;
     }
     break;
   case WM_MOUSEMOVE:
     x=(short)LOWORD(lparam); y=(short)HIWORD(lparam);
     if(!bCaptured && View == TRIVIEW){
       ltype=0;
       if(x > WindowLeftOrigin+Window0sizeX - 20 &&
          x < WindowLeftOrigin+Window0sizeX + 20 &&
          y > WindowTopOrigin+Window0sizeY  - 20 &&
          y < WindowTopOrigin+Window0sizeY  + 20)ltype=1;
       else if(x > WindowLeftOrigin+Window0sizeX - 10 &&
               x < WindowLeftOrigin+Window0sizeX + 10 &&
               y > WindowTopOrigin+10 &&
               y < WindowTopOrigin+TriviewSpaceY)ltype=2;
       else if(y > WindowTopOrigin+Window0sizeY - 10 &&
               y < WindowTopOrigin+Window0sizeY + 10 &&
               x > WindowLeftOrigin+10 &&
               x < WindowLeftOrigin+TriviewSpaceX-10)ltype=3;
       if     (ltype == 1)SetCursor(LoadCursor(NULL,IDC_SIZEALL));
       else if(ltype == 2)SetCursor(LoadCursor(NULL,IDC_SIZEWE));
       else if(ltype == 3)SetCursor(LoadCursor(NULL,IDC_SIZENS));
     }
     if(bCaptured){
         int dx,dy;
         if     (type == 1)SetCursor(LoadCursor(NULL,IDC_SIZEALL));
         else if(type == 2)SetCursor(LoadCursor(NULL,IDC_SIZEWE));
         else if(type == 3)SetCursor(LoadCursor(NULL,IDC_SIZENS));
         dx=x-last_x; dy=y-last_y;
         if(dx != 0 || dy != 0){
           if     (type == 2)dy=0;
           else if(type == 3)dx=0;
           Window0sizeX += dx;
           Window0sizeX = max(20,Window0sizeX);
           Window0sizeX = min(TriviewSpaceX-30,Window0sizeX);
           Window0sizeY += dy;
           Window0sizeY = max(20,Window0sizeY);
           Window0sizeY = min(TriviewSpaceY-30,Window0sizeY);
           SetWindowSize(rc.right,rc.bottom,View,FALSE);
           GetTriview(FALSE);
           for(i=0;i<3;i++)
             MoveWindow(ghwnd_triview[i],
                      WindowsXmin[i],WindowsYmin[i],
                      WindowSizeX[i],WindowSizeY[i],TRUE);
           MoveWindow(ghwnd_view,
                      WindowViewXmin,WindowViewYmin,
                      WindowViewXsize,WindowViewYsize,TRUE);
           if(ghwndOpenGLview != NULL)MoveWindow(ghwndOpenGLview,
                      WindowViewXmin,WindowViewYmin,
                      WindowViewXsize,WindowViewYsize,TRUE);
           MoveWindow(ghwndQuicktime,
                      WindowViewSliderXpos,WindowViewSliderYpos,
                      WindowViewSliderXsize,WindowViewSliderYsize,TRUE);
           UpdateWindow(ghwnd_view);
           UpdateWindow(ghwnd_triview[1]);

         }
         last_x=x; last_y=y;
     }
     break;
   case WM_RBUTTONDOWN:
     SendMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
     break;
   case WM_ERASEBKGND:{
      HDC hDC;
      HPALETTE hpalT;
      HBRUSH hOldBrush;
      RECT rc1;
      hDC = (HDC)wparam;
      if(hBmBackground == NULL){
        hOldBrush = SelectObject (hDC,ghbrushScreen);
        GetClientRect(hwnd,(LPRECT)&rc1);
        PatBlt(hDC,0,0,rc1.right,rc1.bottom,PATCOPY);
        SelectObject (hDC, hOldBrush);
      }
      else{
        HDC hMemDC;
        HBITMAP hbmOld;
        BITMAP  bm;
        GetObject(hBmBackground,sizeof(BITMAP),&bm);
        hMemDC = CreateCompatibleDC(hDC);
        hbmOld = SelectObject(hMemDC,hBmBackground);
        for(x=0;x<(ScreenXsize/bm.bmWidth)+1;x++)
        for(y=0;y<(ScreenYsize/bm.bmHeight)+1;y++)
          BitBlt(hDC,
                  x*bm.bmWidth,y*bm.bmHeight,
                  bm.bmWidth,
                  bm.bmHeight,
                  hMemDC,
                  0,0,
                  SRCCOPY);
        SelectObject(hMemDC,hbmOld);
        DeleteDC(hMemDC);
      }
      return TRUE;
    }
  case WM_PAINT: {
       HDC hdc;
       PAINTSTRUCT ps;
       HBRUSH hOldBrush;
       HPALETTE hpalT;
       RECT rc1;
       hdc=BeginPaint(hwnd,&ps);
       if(hBmBackground == NULL){
         hOldBrush = SelectObject (hdc,ghbrushScreen);
         GetClientRect(hwnd,(LPRECT)&rc1);
         PatBlt(hdc,0,0,rc1.right,rc1.bottom,PATCOPY);
         SelectObject (hdc, hOldBrush);
       }
       if(!bResizing)FocusActiveWindow(hdc,TRUE);  // ????
       EndPaint(hwnd,&ps);
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         PostMessage(hwnd,WM_COMMAND,IDM_FILE_EXIT,0);
         break;
       default:
         return( DefWindowProc( hwnd, msg, wparam, lparam ) );
         break;
     }
     break;
   case WM_COMMAND:
     if(HIWORD(wparam) == 1)lparam=ActiveView; /* acelerator */
     if(HIWORD(wparam) == BN_CLICKED){
       AnimatorMenuCmd(LOWORD(wparam),hwnd,lparam);
     }
     else AnimatorMenuCmd(LOWORD(wparam),hwnd,lparam);
     break;
   case WM_HSCROLL:
     if((HWND)lparam != ghwnd_xscl)break;
     oldi = -1;
     if(!trap_scroll_cursor){
       GetWindowRect(ghwnd_xscl,&rc);
       ClipCursor(&rc);
       trap_scroll_cursor=1;
     }
     switch (LOWORD(wparam)) {
       case SB_ENDSCROLL:
        if(trap_scroll_cursor){
          ClipCursor(NULL);
          trap_scroll_cursor=0;
          SetFocus(ghwnd_main);
        }
        oldi=1; break;
       case SB_LINEDOWN:
        oldi=0;
        CurrentFrame++;
        if(GetKeyState(VK_CONTROL) & 0x8000)CurrentFrame=Nframes;
        break;
       case SB_LINEUP:
        oldi=0;
        CurrentFrame--;
        if(GetKeyState(VK_CONTROL) & 0x8000)CurrentFrame=1;
        break;
       case SB_PAGEDOWN:
         oldi=0;
         if(Nframes < 50)CurrentFrame += min(5,Nframes/5);
         else            CurrentFrame += min(10,Nframes/10);
         break;
       case SB_PAGEUP:
         oldi=0;
         if(Nframes < 50)CurrentFrame -= min(5,Nframes/5);
         else            CurrentFrame -= min(10,Nframes/10);
         break;
       case SB_THUMBTRACK:
         CurrentFrame=(long)HIWORD(wparam);  oldi=0; break;
       case SB_THUMBPOSITION:
         CurrentFrame=(long)HIWORD(wparam);  oldi=0; break;
       default: break;
     }
     if(CurrentFrame < 1)CurrentFrame=1;
     if(CurrentFrame > Nframes)CurrentFrame=Nframes;
     if(oldi >= 0){
//       SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
       SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
       if(View == TRIVIEW || View == ONEVIEW){
         if(oldi == 1)ghcurSave=SetCursor(ghcurWait);
         if(ghwndOpenGLview == NULL)PerspectiveView(0,oldi);
         if(oldi == 1){
           SetCursor(ghcurSave);
           ReDrawStageDisplay(TRUE);
           UpdateSelectedActor(FALSE);
           UpdateGLview(TRUE);
           if(IsWindow(ghwndRobotView))
             SendMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
         }
         else UpdateGLview(FALSE);
       }
     }
     break;
   case WM_SIZE:{
       int cx,cy;
       if(screen_ready){
         DeleteMemoryWindows();
         cx=LOWORD(lparam); cy=HIWORD(lparam);
         SetWindowSize(cx,cy,View,FALSE);
         GetTriview(FALSE);
         if(View == TRIVIEW){
           for(i=0;i<3;i++){
             MoveWindow(ghwnd_triview[i],WindowsXmin[i],WindowsYmin[i],
                        WindowSizeX[i],WindowSizeY[i],TRUE);
           }
         }
         else{
           MoveWindow(ghwnd_current,
                      WindowsXmin[ActiveView],WindowsYmin[ActiveView],
                      WindowSizeX[ActiveView],WindowSizeY[ActiveView],TRUE);
         }
         MoveWindow(ghwnd_view,WindowViewXmin,WindowViewYmin,
                    WindowViewXsize,WindowViewYsize,TRUE);
         if(ghwndOpenGLview != NULL)MoveWindow(ghwndOpenGLview,
                    WindowViewXmin,WindowViewYmin,
                    WindowViewXsize,WindowViewYsize,TRUE);
         MoveWindow(ghwndQuicktime,
                    WindowViewSliderXpos,WindowViewSliderYpos,
                    WindowViewSliderXsize,WindowViewSliderYsize,TRUE);
         if(keyframer_docked_and_visible && IsWindow(ghwndTimeline) &&
            IsWindowVisible(ghwndTimeline))
           MoveWindow(ghwndTimeline,0,KeyframerYpos,cx,KeyframerSize,TRUE);
         InvalidateRect(ghwndQuicktime,NULL,FALSE);
         MoveToolsWindow();
         SendMessage(ghwnd_info,msg,wparam,lparam);
         if(CreateMemoryWindows()){
           ReDrawStageDisplay(TRUE);
           if(ghwndOpenGLview == NULL)PerspectiveView(0,0);
         }
         else{
           DeleteMemoryWindows();
         }
       }
     }
     break;
   case WM_KEYDOWN:
     ExecuteKeyboardKeyCommand(wparam);
   case WM_KEYUP:
     if(!bTrackMouse && !bTrackIcon && !bToolCaptured && !bTrackPan){
       if(GetKeyState(VK_SPACE) & 0x8000)bSpaceDown=TRUE; else bSpaceDown=FALSE;
       if(GetKeyState(VK_CONTROL) & 0x8000)bCtrlDown=TRUE; else bCtrlDown=FALSE;
       if(GetKeyState(VK_SHIFT) & 0x8000)bShiftDown=TRUE; else bShiftDown=FALSE;
       if(GetKeyState(VK_MENU) & 0x8000)bAltDown=TRUE; else bAltDown=FALSE;
       if(CursorInTriview())SetTriViewCursor(FALSE);
     }
//     if(bAbortedDrawing){
//       bAbortedDrawing=FALSE;
//       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
//     }
//     break;
//   case WM_TIMER:
//     if(wparam == 1){
//       KillTimer(hwnd,1);
//       bAbortedDrawing=FALSE;
//       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
//     }
//     break;
   case WM_SETFOCUS:
     if(GetKeyState(VK_SPACE) & 0x8000)bSpaceDown=TRUE; else bSpaceDown=FALSE;
     if(GetKeyState(VK_SHIFT) & 0x8000)bShiftDown=TRUE; else bShiftDown=FALSE;
     if(GetKeyState(VK_CONTROL) & 0x8000)bCtrlDown=TRUE; else bCtrlDown=FALSE;
     if(GetKeyState(VK_MENU) & 0x8000)bAltDown=TRUE; else bAltDown=FALSE;
     if(CursorInTriview())SetTriViewCursor(FALSE);
     break;
   case WM_KILLFOCUS:
     if(bSpaceDown || bCtrlDown || bShiftDown || bAltDown){
       bSpaceDown=FALSE;
       bCtrlDown=FALSE;
       bShiftDown=FALSE;
       bAltDown=FALSE;
       if(CursorInTriview())SetTriViewCursor(FALSE);
     }
     break;
   case WM_SYSKEYDOWN:
     if(((lparam & 0x40000000) == 0) && wparam == VK_MENU){
       bAltDown=TRUE;
//       if(tool == INZOOM && CursorInTriview())SetTriViewCursor(FALSE);
       if((bSpaceDown || tool == INZOOM) && CursorInTriview())SetTriViewCursor(FALSE);
     }
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
   case WM_SYSKEYUP:
     bAltDown=FALSE;
//     if(tool == INZOOM && CursorInTriview())SetTriViewCursor(FALSE);
     if((bSpaceDown || tool == INZOOM) && CursorInTriview())SetTriViewCursor(FALSE);
     break;
   default:
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

// window functions for the main screen windows

LRESULT CALLBACK AnimatorTriViewWndProc(HWND hwnd, UINT msg, WPARAM wparam,
                                    LPARAM lparam ){
 BOOL flag1,flag2;
 HDC hdc;
 RECT rc;
 int id;
 static long evTime;
 static int iIconSave = -1,PickedX,PickedY,lastlocaltool;
 static BOOL bPickedNode,bPickedDbl=FALSE;
 switch( msg ){
   case WM_SETCURSOR:
     if(!SetTriViewCursor(TRUE))return(DefWindowProc(hwnd,msg,wparam,lparam));
     break;
   case WM_RBUTTONDOWN:
     if(bSpaceDown)break;
     if(bToolCaptured)break;
     if(!bTrackMouse){
       if(PickClosestNode((int)GetWindowLong(hwnd,GWL_ID),
                          (int)LOWORD(lparam),(int)HIWORD(lparam),TRUE) &&
          SelectedNode != NULL){
         {
           POINT pt;
           HMENU hMenu;
           hMenu=NULL;
           switch (SelectedNode->type){
             case CAMERA:
               hMenu=LoadMenu(ghinst_main,"CAMERACONTEXTMENU");
               break;
             case ROBOT:
             case NORMAL:
             case ANIMOBJ:
               hMenu=LoadMenu(ghinst_main,"ACTORCONTEXTMENU");
               break;
             case CSG:
               hMenu=LoadMenu(ghinst_main,"CSGCONTEXTMENU");
               break;
             case PATH:
               hMenu=LoadMenu(ghinst_main,"PATHCONTEXTMENU");
               break;
             case LIGHT:
               hMenu=LoadMenu(ghinst_main,"LIGHTCONTEXTMENU");
               break;
             case TARGET:
               hMenu=LoadMenu(ghinst_main,"TARGETCONTEXTMENU");
               break;
             case GROUND:
               hMenu=LoadMenu(ghinst_main,"GROUNDCONTEXTMENU");
               break;
             default: break;
           }
           if(hMenu != NULL){
             pt.x=LOWORD(lparam);
             pt.y=HIWORD(lparam);
             ClientToScreen(hwnd,&pt);
             TrackPopupMenu(GetSubMenu(hMenu,0),
                            TPM_LEFTALIGN | TPM_LEFTBUTTON,
                            pt.x,pt.y,0,ghwnd_main,NULL);
             DestroyMenu(hMenu);
           }
         }
       }
       else SendMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
     }
     else if(!bTrackIcon){
       if(tool == NOTOOL || tool == NODETOOL){;}
       else{
         if(tool == GRABBER)StageGrab(1,0);
       }
     }
     break;
   case WM_LBUTTONDBLCLK:
     if((tool == NOTOOL || tool == NODETOOL) && bPickedDbl){
       BOOL draw=FALSE;
       object *Op;
       bPickedDbl=FALSE;
       Op=GetObjectAtFrame(SelectedNode,CurrentFrame);
       if(Op != NULL){
         switch (Op->type){
           case CAMERA:
             LensSettings();
             draw=TRUE;
//             SelectedInfo();
             break;
           case NORMAL:
           case ANIMOBJ:
             Save_UndoA();
             DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_COSTUME),
                      ghwnd_main,(DLGPROC)EditObjectDlgProc,(LPARAM)Op);
             draw=TRUE; break;
           case CSG:
           case PATH:
           case TARGET:
           case ROBOT:
             Save_UndoA();
             EditObject(SelectedNode,(short)Op->lastframe,
                                     (short)Op->lastframe,NO,NO);
             draw=TRUE;
             break;
           case LIGHT:
             Save_UndoA();
             EditLightDialog(Op,NO);
             draw=TRUE;
             break;
           case GROUND:
             Save_UndoA();
             EditGroundDialog(Op,NO);
             draw=TRUE;
             break;
           default: break;
         }
         if(draw){
           CheckRecursiveFollow(SelectedNode,Op->firstframe,Op->lastframe);
           UpdateSelectedActor(FALSE);
           ReDrawStageDisplay(TRUE);
           if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
           else UpdateGLview(TRUE);
           UPDATETIMELINES(IDC_TIMEL_UPDATE)
         }
       }
     }
     break;
   case WM_LBUTTONDOWN:
     if(bToolCaptured){
       if(tool == TRACKER && subtool == SHIFTER)ShiftKeyFrameDeactivate();
       bToolCaptured=FALSE;
       bTrackMouse=FALSE;
       ClipCursor(NULL);
       SetCursor(ghcurSave);
       ReleaseCapture();
       UpdateWindow(hwnd);
       break;
     }
     if(hwnd != ghwnd_current){
       hdc=GetDC(ghwnd_main);
       FocusActiveWindow(hdc,FALSE);
       ghwnd_current=hwnd;  ActiveView=GetWindowLong(hwnd,GWL_ID);
       FocusActiveWindow(hdc,TRUE);
       ReleaseDC(ghwnd_main,hdc);
     }
     SetCapture(hwnd);
     if(bSpaceDown){
       PanToolDown((int)LOWORD(lparam),(int)HIWORD(lparam),
                    bCtrlDown,bAltDown,hwnd);
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ClipCursor(&rc);
       bTrackPan=TRUE;
     }
     else if((iIconSave=CheckWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam))) >= 0){
       InvertWindowIcon(iIconSave,hwnd);
       bTrackIcon=TRUE;
     }
     else if(tool == PAN/* || tool == INZOOM */){
       PanToolDown((int)LOWORD(lparam),(int)HIWORD(lparam),
                    FALSE,FALSE,hwnd);
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ClipCursor(&rc);
       bTrackPan=TRUE;
     }
     else if(tool == INZOOM){
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       ZoomView((int)LOWORD(lparam),(int)HIWORD(lparam));
     }
     else if(tool == ROTATOR ){
       StageRotate(GetWindowLong(hwnd,GWL_ID));
     }
     else if(tool == SIZER   ){
       Resize();
     }
     else if(tool == TRACKER && subtool == SHIFTER ){
       do_NOT_abort=YES;
       id=ShiftKeyFrameActivate((int)LOWORD(lparam),(int)HIWORD(lparam));
       do_NOT_abort=NO;
       if(id == 1)bToolCaptured=TRUE;
       else if(id < 0)break;
/* under test */
//       if(GetKeyState(VK_CONTROL) & 0x8000){
         do_NOT_abort=YES;
         if(!bToolCaptured)PickClosestNode((int)GetWindowLong(hwnd,GWL_ID),
          (int)LOWORD(lparam),(int)HIWORD(lparam),FALSE);
         do_NOT_abort=NO;
//       }
/*************/
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ghcurSave=SetCursor(ghcurCross);
       ClipCursor(&rc);
       bTrackMouse=TRUE;
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
     }
     else if(tool == TRACKER && subtool == ANGLER){
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       do_NOT_abort=YES;
       AngleKeyFrame((int)LOWORD(lparam),(int)HIWORD(lparam));
       do_NOT_abort=NO;
     }
     else if(tool == SKELETOR){
       if(Skeletor()){
/* under test */
         if(GetKeyState(VK_CONTROL) & 0x8000){
           do_NOT_abort=YES;
           PickClosestNode((int)GetWindowLong(hwnd,GWL_ID),
            (int)LOWORD(lparam),(int)HIWORD(lparam),FALSE);
           do_NOT_abort=NO;
         }
/*************/
         GetWindowRect(hwnd,&rc);
         rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
         ghcurSave=SetCursor(ghcurCross);
         ClipCursor(&rc);
         bTrackMouse=TRUE;
         Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       }
     }
     else if(tool == GRABBER){
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       do_NOT_abort=YES;
       PickClosestNode((int)GetWindowLong(hwnd,GWL_ID),
                       (int)LOWORD(lparam),(int)HIWORD(lparam),FALSE);
       do_NOT_abort=NO;
       if(StageGrabIdentify()){
         ReDrawStageDisplay(TRUE);
         PerspectiveView(0,0);
         GetWindowRect(hwnd,&rc);
         rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
         ghcurSave=SetCursor(ghcurCross);
         ClipCursor(&rc);
         bTrackMouse=TRUE;
         EnterGrabMode();
       }
     }
     else if(tool == PATHEDITOR){
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       do_NOT_abort=YES;
       if(!PickClosestPathPoint((int)GetWindowLong(hwnd,GWL_ID),
                       (int)LOWORD(lparam),(int)HIWORD(lparam)) ){
/* under test */
         if(GetKeyState(VK_CONTROL) & 0x8000){
           PickClosestNode((int)GetWindowLong(hwnd,GWL_ID),
                           (int)LOWORD(lparam),(int)HIWORD(lparam),FALSE);
         }
/*************/
         if(gToolFlags.flagP == 3)AdjustPathTension(GetWindowLong(hwnd,GWL_ID));
       }
       do_NOT_abort=NO;
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ghcurSave=SetCursor(ghcurCross);
       ClipCursor(&rc);
       bTrackMouse=TRUE;
     }
     else{
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ghcurSave=SetCursor(ghcurCross);
       ClipCursor(&rc);
       bTrackMouse=TRUE;
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       bPickedNode=FALSE;
       bPickedDbl=FALSE;
       if((GetKeyState(VK_CONTROL) & 0x8000) != 0x8000 &&
          (tool == NOTOOL || tool == NODETOOL)){
//       if(tool == NOTOOL || tool == NODETOOL){
         do_NOT_abort=YES;
         if(PickClosestNode((int)GetWindowLong(hwnd,GWL_ID),
            (int)LOWORD(lparam),(int)HIWORD(lparam),FALSE)){
           bPickedNode=TRUE;
           bPickedDbl=TRUE;
           PickedX=LOWORD(lparam); PickedY=HIWORD(lparam);
           evTime=GetTickCount()+5;
         }
         do_NOT_abort=NO;
       }
     }
     if(!bTrackPan && !bTrackIcon && !bTrackMouse && GetCapture() == hwnd
        && !(GetAsyncKeyState(VK_LBUTTON) & 0x8000)){
       ReleaseCapture();
     }
     break;
   case WM_LBUTTONUP:
     id=0;
     if(bToolCaptured)break;
     if(bTrackIcon){
       if(iIconSave >= 0){
         InvertWindowIcon(iIconSave,hwnd);
         if     (iIconSave == 0)iIconSave=IDM_PAN_LEFT;
         else if(iIconSave == 1)iIconSave=IDM_PAN_RIGHT;
         else if(iIconSave == 2)iIconSave=IDM_PAN_UP;
         else if(iIconSave == 3)iIconSave=IDM_PAN_DOWN;
         else if(iIconSave == 4)iIconSave=IDM_WINDOW_SWITCHWINDOWS;
         else if(iIconSave == 5)iIconSave=IDM_WINDOW_SWITCHWINDOW0;
         else if(iIconSave == 6)iIconSave=IDM_WINDOW_SWITCHWINDOW1;
         PostMessage(ghwnd_main,WM_COMMAND,(WPARAM)iIconSave,
           (LPARAM)GetWindowLong(hwnd,GWL_ID));
       }
       iIconSave = -1;
       bTrackIcon=FALSE;
     }
     else if(bTrackPan){
       if(bSpaceDown){
         flag1=bCtrlDown;
//         flag2=bShiftDown;
         flag2=bAltDown;
       }
       else{
         if(tool == INZOOM){
           if(bAltDown){flag1=FALSE; flag2=TRUE;}  /* zoom out */
           else        {flag1=TRUE; flag2=FALSE;}  /* zoom in */
         }
         else {flag1=FALSE; flag2=FALSE;}
       }
       PanToolUp((int)LOWORD(lparam),(int)HIWORD(lparam),
                 flag1,flag2,hwnd);
       bTrackPan=FALSE;
       ClipCursor(NULL);
       if(!(GetKeyState(VK_SPACE)&0x8000))bSpaceDown=FALSE;
       if(!(GetKeyState(VK_CONTROL)&0x8000))bCtrlDown=FALSE;
       if(!(GetKeyState(VK_SHIFT)&0x8000))bShiftDown=FALSE;
       if(!(GetKeyState(VK_MENU)&0x8000))bAltDown=FALSE;
       SetTriViewCursor(FALSE);
     }
     else if(bTrackMouse){
       bTrackMouse=FALSE;   //
       ClipCursor(NULL);    // h
       SetCursor(ghcurSave);//
       if(tool == GRABBER){
         if(bPickedNode){
           LeaveGrabMode();
           StageGrabOff(1);
           tool=lastlocaltool;
         }
         else LeaveGrabMode();
       }
       else if(tool == PATHEDITOR)id=1;
       // h - here?
       UpdateWindow(hwnd);
     }
     bPickedNode=FALSE;
     ReleaseCapture();
     if(tool == PATHEDITOR && id == 1){
       if(gToolFlags.flagP      == 1)CreatePathPoint();
       else if(gToolFlags.flagP == 2)DeletePathPoint();
     }
     break;
   case WM_MOUSEMOVE:
     if(bTrackMouse){
       Move3dCursor(0,(int)LOWORD(lparam),(int)HIWORD(lparam));
       do_NOT_abort=YES;
       if     (tool == GRABBER)StageGrab(0,0); /* draw to screen */
       else if(tool == PATHEDITOR)EditPath();
       else if(tool == TRACKER && subtool == SHIFTER && bToolCaptured){
         do_NOT_abort=YES;
         ShiftKeyFrame((int)LOWORD(lparam),(int)HIWORD(lparam));
         do_NOT_abort=NO;
       }
       else if((tool == NOTOOL || tool == NODETOOL) && bPickedNode){
         if(GetTickCount() > evTime){
           if(abs((int)LOWORD(lparam) - PickedX) >= 2 ||
              abs((int)HIWORD(lparam) - PickedY) >= 2){
             if(IsValidPosition() != NULL){
               lastlocaltool=tool;
               tool=GRABBER;
               Save_UndoA();
               StageGrabIdentify();
               ReDrawStageDisplay(TRUE);
               PerspectiveView(0,0);
               EnterGrabMode();
             }
             else{
               bPickedNode=FALSE;
               ReleaseCapture();
               bTrackMouse=FALSE;
               ClipCursor(NULL);
               Save_UndoA();
               StageGrabIdentify();
               ReDrawStageDisplay(TRUE);
               if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
               else UpdateGLview(TRUE);
             }
           }
         }
       }
       do_NOT_abort=NO;
     }
     else if(bTrackIcon){
       int new_icon;
       new_icon=CheckWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam));
       if(new_icon !=  iIconSave){
         if(iIconSave >= 0)InvertWindowIcon(iIconSave,hwnd);
         iIconSave=new_icon;
         if(iIconSave >= 0)InvertWindowIcon(iIconSave,hwnd);
       }
     }
     else if(bTrackPan){
       if(tool == PAN || tool == INZOOM || bSpaceDown){
         if(bSpaceDown){
           flag1=bCtrlDown;
           flag2=bAltDown;
         }
         else{
           if(tool == INZOOM && !bAltDown)flag1=TRUE; else flag1=FALSE;
           if(tool == INZOOM &&  bAltDown)flag2=TRUE; else flag2=FALSE;
         }
         PanToolMove((int)LOWORD(lparam),(int)HIWORD(lparam),
                     flag1,flag2,hwnd);
       }
     }
     else { /* get help on in screen icons */
       int icon,new_message;
       icon=CheckWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam));
       if(icon >= 0){
         if(icon == 4)new_message=lmessage_table[View];
         else         new_message=IDS_SWAPOTHER;
         if(new_message != current_message){
           current_message=new_message;
           SendPrgmMessage(current_message,1);
         }
       }
       else if(current_message != tmessage_table[tool]){
         current_message = tmessage_table[tool];
         SendPrgmMessage(current_message,1);
       }
     }
     break;
   case WM_PAINT:{
       PAINTSTRUCT ps;
       HPALETTE hpalT;
       hdc=BeginPaint(hwnd,&ps);
       if(!bResizing){
         GetClientRect(hwnd,&rc);
         id=(int)GetWindowLong(hwnd,GWL_ID);
         if(ghdc_triview_Bitmap[id] != NULL)BitBlt(hdc,0,0,
           rc.right,rc.bottom,ghdc_triview_Bitmap[id],0,0,SRCCOPY);
         DrawOne3dCursor(hdc,id);
       }
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

LRESULT CALLBACK AnimatorViewWndProc(HWND hwnd, UINT msg, WPARAM wparam,
                                    LPARAM lparam ){
 RECT rc;
 static int iIconSave = -1;
 switch( msg ){
   case WM_ERASEBKGND:{
      HDC hDC;
      HPALETTE hpalT;
      HBRUSH hOldBrush;
      if(bResizing)return TRUE;
      hDC       = (HDC)wparam;
      hOldBrush = SelectObject (hDC,ghbrushWindow);
      GetClientRect(hwnd,(LPRECT)&rc);
      PatBlt(hDC,0,0,rc.right,rc.bottom,PATCOPY);
      SelectObject (hDC, hOldBrush);
      return TRUE;
   }
   case WM_RBUTTONDOWN:{
       POINT pt;
       HMENU hMenu=LoadMenu(ghinst_main,"CameraViewPopupMenu");
       if(hMenu == NULL)break;
       pt.x=LOWORD(lparam);
       pt.y=HIWORD(lparam);
       ClientToScreen(hwnd,&pt);
       TrackPopupMenu(GetSubMenu(hMenu,0),TPM_LEFTALIGN | TPM_LEFTBUTTON,
                      pt.x,pt.y,0,ghwnd_main,NULL);
       DestroyMenu(hMenu);
     }
     break;
   case WM_LBUTTONDOWN:
     SetCapture(hwnd);
     if((iIconSave=CheckViewWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam))) >= 0){
       InvertViewWindowIcon(iIconSave,hwnd);
       bTrackIcon=TRUE;
       break;
     }
     GetClientRect(hwnd,&rc);
     //ghcurSave=SetCursor(ghcurWait);
     PerspectiveView(0,1);
     //SetCursor(ghcurSave);
     break;
   case WM_MOUSEMOVE:{
       int new_icon;
       new_icon=CheckViewWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam));
       if(bTrackIcon){
         if(new_icon !=  iIconSave){
           if(iIconSave >= 0)InvertViewWindowIcon(iIconSave,hwnd);
           iIconSave=new_icon;
           if(iIconSave >= 0)InvertViewWindowIcon(iIconSave,hwnd);
         }
       }
       if(new_icon > 0){
         current_message=lmessage_table[View];
         SendPrgmMessage(current_message,1);
       }
       else if(current_message != IDS_VIEWUPDATE){
         current_message = IDS_VIEWUPDATE;
         SendPrgmMessage(IDS_VIEWUPDATE,1);
       }
     }
     break;
   case WM_LBUTTONUP:
     ReleaseCapture();
     if(bTrackIcon){
       if(iIconSave >= 0){
         InvertViewWindowIcon(iIconSave,hwnd);
         if(iIconSave == 1){
           if(View == TRIVIEW){
             if(!BigCameraView)PostMessage(ghwnd_main,
               WM_COMMAND,(WPARAM)IDM_WINDOW_BIGCAMERA,
               (LPARAM)GetWindowLong(hwnd,GWL_ID));
             PostMessage(ghwnd_main,WM_COMMAND,
               (WPARAM)IDM_WINDOW_SWITCHWINDOWS,
               (LPARAM)GetWindowLong(hwnd,GWL_ID));
           }
           else{
             PostMessage(ghwnd_main,WM_COMMAND,
               (WPARAM)IDM_WINDOW_SWITCHWINDOWS,
               (LPARAM)GetWindowLong(hwnd,GWL_ID));
             if(BigCameraView)PostMessage(ghwnd_main,
               WM_COMMAND,(WPARAM)IDM_WINDOW_BIGCAMERA,
               (LPARAM)GetWindowLong(hwnd,GWL_ID));
           }
         }
       }
       iIconSave = -1;
       bTrackIcon=FALSE;
     }
     break;
   case WM_TIMER:
     if(AnimationTimer == 3){
       long qsave;
       qsave=global_quickdraw;
       global_quickdraw=1;
       PerspectiveView(1,1);
       global_quickdraw=qsave;
     }
     else{
       PerspectiveView(1,AnimationTimer-1);
     }
     CurrentFrame++;
     if(CurrentFrame > Nframes)CurrentFrame=1;
//     SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
     SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
     break;
   case WM_PAINT: {
       HDC hdc;
       PAINTSTRUCT ps;
       HPALETTE hpalT;
       hdc=BeginPaint(hwnd,&ps);
       if(!bResizing){
         GetClientRect(hwnd,&rc);
         if(ghdc_view_Bitmap != NULL){
           COLORREF oldtextcolour;
           int oldbkmode;
           char p[8];
//           sprintf(p,"%4.4ld",CurrentFrame);
           oldbkmode=SetBkMode(ghdc_view_Bitmap,TRANSPARENT);
           oldtextcolour=SetTextColor(ghdc_view_Bitmap,RGB(255,255,255));
//           TextOut(ghdc_view_Bitmap,1,rc.bottom-Textheight,p,strlen(p));
           if(AnimationTimer == 0){
             HDC hMemDC = CreateCompatibleDC(ghdc_view_Bitmap);
             HBITMAP hbmOld=SelectObject(hMemDC,hbmName[5]);
             BitBlt(ghdc_view_Bitmap,0,0,
                    WINDOW_ID_VSIZE,WINDOW_ID_YSIZE,
                    hMemDC,0,0,SRCCOPY);
             SelectObject(hMemDC, hbmOld);
             DeleteDC(hMemDC);
//
//             TextOut(ghdc_view_Bitmap,
//                     rc.right-8*Textwidth,rc.bottom-Textheight,
//                     "Camera",strlen("Camera"));
//
           }
           SetBkMode(ghdc_view_Bitmap,oldbkmode);
           SetTextColor(ghdc_view_Bitmap,oldtextcolour);
           BitBlt(hdc,0,0,rc.right,rc.bottom,ghdc_view_Bitmap,0,0,SRCCOPY);
         }
       }
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

// Set the mouse cursor to the correct cursor depending on
// the window the cursor is over and the state of the modifier keys
// e.g. the PAN and ZOOM key accelerators.

static BOOL SetTriViewCursor(BOOL type){
 if(bSpaceDown){
   if(bCtrlDown)SetCursor(ghcurZin);
   else if(bAltDown)SetCursor(ghcurZout);
   else SetCursor(ghcurPan);
 }
// else if(tool == INZOOM && bAltDown)SetCursor(ghcurZout);
 else if(tool == TRACKER && subtool == SHIFTER)SetCursor(ghcurTracker);
 else if(tool == TRACKER && subtool == ANGLER)SetCursor(ghcurDirector);
 else if(tool < NTOOLS)SetCursor(ghcurTools[tool]);
 else if(tool == PATHEDITOR)SetCursor(ghcurPathed);
 else if(tool == SKELETOR)SetCursor(ghcurSkeletor);
 else if(type)return FALSE;
 else SetCursor(ghcurArrow);
 return TRUE;
}

//
// The main windows COMMAND generating function 
//

#include "..\opengl\glmenu.h"

#if __HELP_WS__
#include "..\help_ws\include\htmlhelp.h"
#endif

static VOID AnimatorMenuCmd(int item, HWND hwnd, long lparam){
 BOOL play_now=FALSE;
 int i;
 node *local_node;
 char local_nodename[128],lname[255];
 static BOOL busy_flag=FALSE;
 static object *ObjectInModeler=NULL;
 switch (item){
   case IDM_FILE_LOAD_DIRECT:{
       char load_file[255];
       GetPrivateProfileString("CALLBACK","FILE","$$$",
           load_file,sizeof(load_file),IniFilename);
       if(strncmp(load_file,"$$$",3) == 0)break;
       strcpy(gszSTGfile,load_file);
       strcpy(gszSTGdir,load_file);
       *FileInPath(gszSTGdir) = '\0';
       UpdateLastFilesList(gszSTGfile);
       goto OPENJUMP;
     }
   case IDM_FILE_LOAD_STARTUP:
     strcpy(gszSTGfile,gszHomeDir); strcat(gszSTGfile,"startup.ofx");
     play_now=TRUE;
     goto OPENJUMP;
   case LASTFILECMD+1:
   case LASTFILECMD+2:
   case LASTFILECMD+3:
   case LASTFILECMD+4:
   case LASTFILECMD+5:
   case IDM_FILE_LOAD_ANIMATION:
     Save_UndoA();
     if(EDIT_ACTION){
       int fResult;
       fResult=SendPrgmQuery(IDQ_CHANGED,2);
       if(fResult == IDCANCEL)break;
#if !__DEMO__
       if(fResult == IDYES){
         if(SelectSfxFileName(1,gszSTGfile,gszSTGdir,IDX_MISC_SAVEANIMATION,
                           "(*.OFX)|*.ofx|",ghwnd_main) == TRUE){
           AppendFileExtension(gszSTGfile,".ofx");
           ghcurSave=SetCursor(ghcurWait);
           SaveStage(gszSTGfile,FALSE);
           SetCursor(ghcurSave);
         }
         else break;
       }
#endif
     }
     play_now=FALSE;
     if(item == LASTFILECMD+1 || item == LASTFILECMD+2 ||
        item == LASTFILECMD+3 || item == LASTFILECMD+4 ||
        item == LASTFILECMD+5){
       strcpy(gszSTGfile,LastFilesList[item-LASTFILECMD-1].file);
       goto OPENJUMP;
     }
     if(SelectSfxFileName(0,gszSTGfile,gszSTGdir,
      IDX_MISC_OPENANIMATION,"(*.OFX)|*.ofx|",ghwnd_main) == TRUE){
       OPENJUMP:
       ghcurSave=SetCursor(ghcurWait);
       if(LoadStage(gszSTGfile) == OK){
         UpdateLastFilesList(gszSTGfile);
         SetPrgmMessage(IDS_CURRENT,gszSTGfile);
         SendPrgmMessage(IDS_CURRENT,0);
#if !__DEMO__
         EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_SAVEOLD,MF_ENABLED);
#endif
         strcpy(gszSTGsaved,gszSTGfile);
       }
       else play_now=FALSE;
//       SetScrollRange(ghwnd_xscl,SB_CTL,1,(int)Nframes,FALSE);
//       SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
       SendMessage(ghwnd_xscl,TBM_SETRANGE,(WPARAM)TRUE,(LPARAM)MAKELONG(1,Nframes));
       SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
       else UpdateGLview(TRUE);
       if(coords_visible)UpdateRuler(0);
       UPDATETIMELINES(IDC_TIMEL_UPDATE)
       SetCursor(ghcurSave);
       if(play_now)PostMessage(hwnd,WM_COMMAND,IDM_FRAME_PLAY_QUICK,0);
     }
     break;
   case IDM_FILE_LOAD_TEMPLATE:
   case IDM_FILE_LOAD_NEW: {
       int iok=0;
       if(EDIT_ACTION){
         int fResult;
         Save_UndoA();
         fResult=SendPrgmQuery(IDQ_CHANGED,2);
         if(fResult == IDCANCEL)break;
#if !__DEMO__
         if(fResult == IDYES){
           if(SelectSfxFileName(1,gszSTGfile,gszSTGdir,
             IDX_MISC_SAVEANIMATION,"(*.OFX)|*.ofx|",ghwnd_main) == TRUE){
             AppendFileExtension(gszSTGfile,".ofx");
             ghcurSave=SetCursor(ghcurWait);
             SaveStage(gszSTGfile,FALSE);
             SetCursor(ghcurSave);
           }
           else break;
         }
#endif
         iok=1;
       }
       {
         object *Op;
         ScrapEverything();
         UpdateSelectedActor(FALSE);
         SendPrgmMessage(IDS_DEFAULT,0);
         EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_SAVEOLD,MF_GRAYED);
         gszSTGfile[0]='\0';
         strcpy(gszSTGsaved,"unnamed.ofx");
         GetTriview(TRUE);
         NpointerX=0; NpointerY=0; NpointerZ=0;
         if(CreateNode() != NULL){
           LoadString(ghinst_main,IDX_MISC_CAMERA,res_str,256);
           strcpy(MainNp->actorname,res_str);
           MainNp->type=CAMERA;
           SelectedCamera=MainNp;
           if((Op=CreateActor(MainNp,1,(short)Nframes)) != NULL){
             Op->type=MainNp->type;
             MainNp->fobj->morph=NO;
             Ncameras=1;
             if(item != IDM_FILE_LOAD_TEMPLATE){
               CreatePosition(MainNp,1,(short)Nframes);
               MainNp->fpos->finish[0]=0;
               MainNp->fpos->finish[1]=-TVsizeY/3; //2+TVsizeY/10;
               MainNp->fpos->finish[2]=0;
               MainNp->fpos->type=SPLINE;
               CreateAlign(FirstNp,1,(short)Nframes);
               CreateSize(FirstNp,1,(short)Nframes);
               CreateNode();
               LoadString(ghinst_main,IDX_MISC_TARGET,res_str,256);
               strcpy(MainNp->actorname,res_str);
               MainNp->type=TARGET;
               Op=CreateActor(MainNp,1,(short)Nframes);
               Op->type=MainNp->type;
               FirstNp->fali->type=TRACK;
               FirstNp->fali->topath=MainNp;
               CreatePosition(MainNp,1,Nframes);
               MainNp->fpos->type=SPLINE;
             }
           }
           else PostQuitMessage(0L);  /* this should be a 1 in a billion shot */
         }
         else PostQuitMessage(0L);
         EDIT_ACTION=NO;
         if(item == IDM_FILE_LOAD_TEMPLATE){
           StartDefault(0);
           Zoom(CENTREWORLD,0);
         }
//         SetScrollRange(ghwnd_xscl,SB_CTL,1,(int)Nframes,FALSE);
//         SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
         SendMessage(ghwnd_xscl,TBM_SETRANGE,(WPARAM)TRUE,(LPARAM)MAKELONG(1,Nframes));
         SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
         ReDrawStageDisplay(TRUE);
         if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
         else UpdateGLview(TRUE);
         if(coords_visible)UpdateRuler(0);
         UPDATETIMELINES(IDC_TIMEL_UPDATE)
       }
     }
     break;
#if !__DEMO__
   case IDM_FILE_SAVEOLD:
     ghcurSave=SetCursor(ghcurWait);
     if(SaveStage(gszSTGsaved,TRUE) == OK)EDIT_ACTION=NO; // changed FALSE so no question
     SetCursor(ghcurSave);
     break;
   case IDM_FILE_SAVE:
     if(SelectSfxFileName(1,gszSTGfile,gszSTGdir,IDX_MISC_SAVEANIMATION,
        "(*.OFX)|*.ofx|",ghwnd_main) == TRUE){
       AppendFileExtension(gszSTGfile,".ofx");
       UpdateLastFilesList(gszSTGfile);
       ghcurSave=SetCursor(ghcurWait);
       if(SaveStage(gszSTGfile,FALSE) == OK){
         EDIT_ACTION=NO;
         EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_SAVEOLD,MF_ENABLED);
         strcpy(gszSTGsaved,gszSTGfile);
         SetPrgmMessage(IDS_CURRENT,gszSTGfile);
         SendPrgmMessage(IDS_CURRENT,0);
       }
       SetCursor(ghcurSave);
     }
     break;
#endif
   case IDM_CALLBACK_FSCRIPT:{   /* only from renderer */
       HWND hwndR;
       BOOL isOK;
       long saved_current;
       saved_current=CurrentFrame;
       if(lparam != CurrentFrame){
         if(lparam > 0 && lparam <= Nframes)CurrentFrame=lparam;
       }
       isOK=MakeScript(3,1);
       CurrentFrame=saved_current;
       hwndR=FindWindow("OFX:RenderClass",NULL);
       if(IsWindow(hwndR)){
         if(isOK)SendMessage(hwndR,WM_COMMAND,IDM_CALLBACK_SETUP,0);
         else    SendMessage(hwndR,WM_COMMAND,IDM_CALLBACK_CANCEL,0);
       }
     }
     break;
   case IDM_CALLBACK_ASCRIPT:{  /* only from renderer */
       BOOL isOK;
       HWND hwndR;
       isOK=MakeScript(4,1);
       hwndR=FindWindow("OFX:RenderClass",NULL);
       if(IsWindow(hwndR)){
         if(isOK)SendMessage(hwndR,WM_COMMAND,IDM_CALLBACK_SETUP,0);
         else    SendMessage(hwndR,WM_COMMAND,IDM_CALLBACK_CANCEL,0);
       }
     }
     break;
   case IDM_FILE_EXP_FSCRIPT:
     if(!MakeScript(1,1))SendPrgmQuery(IDQ_SCRIPTOPENFAIL,0);
     break;
   case IDM_FILE_EXP_ASCRIPT:
     if(!MakeScript(0,1))SendPrgmQuery(IDQ_SCRIPTOPENFAIL,0);
     break;
   case IDM_FILE_EXP_FMSCRIPT:{
       char ttt[64];
       LoadString(ghinst_main,IDX_MISC_SCRIPTID,res_str,256);
       LoadString(ghinst_main,IDX_MISC_SCRIPTTITLE,ttt,64);
       i=RequestNumEntry(5,1,999,res_str,ttt);
       if(!MakeScript(1,i))SendPrgmQuery(IDQ_SCRIPTOPENFAIL,0);
     }
     break;
   case IDM_FILE_EXP_AMSCRIPT:{
       char ttt[64];
       LoadString(ghinst_main,IDX_MISC_SCRIPTID,res_str,256);
       LoadString(ghinst_main,IDX_MISC_SCRIPTTITLE,ttt,64);
       i=RequestNumEntry(5,1,999,res_str,ttt);
       if(!MakeScript(0,i))SendPrgmQuery(IDQ_SCRIPTOPENFAIL,0);
     }
     break;
   case IDM_FILE_EXP_OPENGL:
     StartExternGL();
     break;
   case IDM_FILE_EXP_VRML:
     if(!MakeVRMLScript(0))SendPrgmQuery(IDQ_SCRIPTOPENFAIL,0);
     break;
   case IDM_WINDOW_SHOWOPENGL:
     StartGLview(FALSE);
     break;
   case IDM_FILE_MODELER:
     ExecuteModeler();
     break;
   case IDM_FILE_KEYFRAMER:
     if(keyframer_docked){
       if(IsWindow(ghwndTimeline)){
         if(IsWindowVisible(ghwndTimeline)){
           TimelineEditor();
           keyframer_docked_and_visible=0;
           FormatScreen();
         }
         else{
           keyframer_docked_and_visible=1;
           FormatScreen();
           TimelineEditor();
         }
       }
       else{
         keyframer_docked_and_visible=1;
         FormatScreen();
         TimelineEditor();
       }
     }
     else TimelineEditor();
     break;
   case IDM_FILE_VIEWER:
     LoadUtility("mdiview.exe","OFX:Viewer",1004);
     break;
   case IDM_FILE_PLAYER:{
       int result;
       char modulename[255];
       strcpy(modulename,gszHomeDir);strcat(modulename,"winplay.exe ANIM");
       result=WinExec(modulename,SW_SHOW);
       if(result < 32){
         if(result == ERROR_FILE_NOT_FOUND ||
            result == ERROR_PATH_NOT_FOUND)SendPrgmQuery(IDQ_NOSTART,3);
       }
     }
     break;
   case IDM_FILE_PLAYAVI:
     LoadUtility("aviplay.exe","OFX:AVIPlayer",1004);
     break;
   case IDM_FILE_FLI2AVI:
     LoadUtility("fli2avi.exe","OFX:AVIConvertClass",-1);
     break;
   case IDM_FILE_FLI2GIF:
     LoadUtility("fli2gif.exe","RFS:GIFConvertClass",-1);
     break;
   case IDM_FILE_ANIM:
     LoadUtility("build.exe","OFX:BuildClass",-1);
     break;
   case IDM_FILE_AUDIO:
     LoadUtility("editsnd.exe","OFX:EditSoundClass",-1);
     break;
   case IDM_FILE_AVI2MPG:
     LoadUtility("avi2mpg.exe",NULL,-1);
     break;
   case IDM_FILE_GIF2AVI:
     LoadUtility("gif2avi.exe",NULL,-1);
     break;
   case IDM_FILE_PLAYMPG:
     LoadUtility("amovieplayer.exe",NULL,-1);
     break;
   case IDM_FILE_EXIT:
     if(EDIT_ACTION){
       int fResult;
       fResult=SendPrgmQuery(IDQ_CHANGED,2);
       if(fResult == IDNO)PostQuitMessage(0L);
#if __DEMO__
       PostQuitMessage(0L);
#else
       if(fResult == IDYES){
         if(SelectSfxFileName(1,gszSTGfile,gszSTGdir,IDX_MISC_SAVEANIMATION,
           "(*.OFX)|*.ofx|",ghwnd_main) == TRUE){
           AppendFileExtension(gszSTGfile,".ofx");
           ghcurSave=SetCursor(ghcurWait);
           if(SaveStage(gszSTGfile,FALSE) == OK)PostQuitMessage(0L);
         }
         SetCursor(ghcurSave);
       }
       break;
#endif
     }
     else PostQuitMessage(0L);
     break;
   case IDM_CALLBACK_SHUTDOWN:
     if(EDIT_ACTION)PostMessage(ghwnd_main,WM_COMMAND,IDM_FILE_EXIT,0);
     else PostQuitMessage(0L);
     break;
   case IDM_FILE_PRINT:
     PrintImageFile(ghwnd_main);
     break;
   case IDM_FILE_PRINTSETUP:
     break;
   case IDM_CONTEXT_TOOLSLEFT:
     bToolpannelLeft ^= 1;
     UpdateIconPositions();
     break;
   case IDM_CONTEXT_TOOLSSMALL:
     bToolpannelSmall ^= 1;
     ChangeToolboxBitmaps();
     UpdateIconPositions();
     break;
   case IDM_FILE_PREFERENCES:
     PreferencesPropertyDlgProc(ghwnd_main);
     SendMessage(ghwndTooltip,TTM_ACTIVATE,(WPARAM)Preferences.tooltips,0);
     if(Preferences.toolleft  != bToolpannelLeft ||
        Preferences.toolsmall != bToolpannelSmall){
       if(bToolpannelSmall != Preferences.toolsmall){
         bToolpannelSmall=Preferences.toolsmall;
         ChangeToolboxBitmaps();
       }
       bToolpannelLeft=Preferences.toolleft;
       UpdateIconPositions();
     }
     else ReDrawStageDisplay(TRUE);
     break;
   case IDM_ACTOR_ADD_A:
     AddQuickActor(-1);
     break;
   case IDM_ACTOR_ADD_C:
     Save_UndoA();
     local_node=SelectNode(local_nodename,YES,ghwnd_main);
     AddCostumeTimeLine(local_node);
     UPDATETIMELINES(IDC_TIMEL_UPDATE)
     break;
   case IDM_ACTOR_ADD_E:{
       object *Op1;
       if(SelectedNode != NULL){
         if(SelectedNode->type == NORMAL){
           Save_UndoA();
           if((Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
             if(Op1->effect != NULL){
               int fResult;
               fResult=SendPrgmQuery(IDQ_EFFECTEXISTS,2);
               if(fResult == IDYES){
                 DeleteEffect(Op1);
                 AddEffect(Op1,ghwnd_main);
               }
               else if(fResult == IDNO){
                 EditActorEffect(Op1,ghwnd_main);
               }
             }
             else AddEffect(Op1,ghwnd_main);
           }
           else{
             SendPrgmQuery(IDQ_OLD2,0);
           }
         }
         else{
           SendPrgmQuery(IDQ_OLD2,0);
         }
       }
       else{
         SendPrgmQuery(IDQ_OLD1,0);
       }
     }
     break;
   case IDM_ACTOR_DEL_A:
   case IDM_ACTOR_DEL_C:
     if(SelectedNode != NULL){
       Save_UndoA();
       if(SelectedNode->type == CAMERA){
         if(item == IDM_ACTOR_DEL_C){
           SendPrgmQuery(IDQ_CAMERANODELCOSTUME,0);
         }
         else if(SelectedNode == FirstNp){
           SendPrgmQuery(IDQ_CAMERANODELETE,0);
         }
         else if(SendPrgmQuery(IDQ_DELETECAMERA,1) == IDYES){
           if(SelectedNode == SelectedCamera)SelectedCamera=FirstNp;
           DeleteNode(SelectedNode);
           SelectedNode=NULL;
           UpdateSelectedActor(FALSE);
           Ncameras--;
           ReDrawStageDisplay(TRUE);
           if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
           else UpdateGLview(TRUE);
           UPDATETIMELINES(IDC_TIMEL_UPDATE)
         }
       }
       else{
         if(SendPrgmQuery(IDQ_DELETEACTOR,1) == IDYES){
           if(item == IDM_ACTOR_DEL_A){
             if(SelectedNode->type == GROUND)Nground=0;
             if(SelectedNode->type == SKY)Nskys=0;
             if(SelectedNode->type == DIRECTOR)Ndirectors=0;
             if(SelectedNode->type == ROBOT)Nrobots--;
             DeleteNode(SelectedNode);
             SelectedNode=NULL;
             UpdateSelectedActor(FALSE);
           }
           else if(item == IDM_ACTOR_DEL_C){
             if(SelectedNode->type == SKY)DeleteSky(SelectedNode,CurrentFrame);
             else DeleteActor(SelectedNode,CurrentFrame);
           }
           ReDrawStageDisplay(TRUE);
           if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
           else UpdateGLview(TRUE);
           UPDATETIMELINES(IDC_TIMEL_UPDATE)
         }
       }
     }
     else SendPrgmQuery(IDQ_OLD1,0);
     break;
   case IDM_ACTOR_DEL_E:{
       object *Op1;
       if(SelectedNode != NULL){
         Save_UndoA();
         if(SelectedNode->type == NORMAL){
           if((Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
             if(Op1->effect != NULL){
               if(SendPrgmQuery(IDQ_DELETEACTOR,1) == IDYES){
                  DeleteEffect(Op1);
               }
             }
             else {
               SendPrgmQuery(IDQ_NOEFFECT,0);
             }
           }
           else{
           }
         }
         else{
           SendPrgmQuery(IDQ_OLD2,0);
         }
       }
       else{
         SendPrgmQuery(IDQ_OLD1,0);
       }
     }
     break;
   case IDM_ACTOR_SELECT:
     if((local_node=SelectNode(local_nodename,YES,ghwnd_main)) != NULL){
       if(local_node->type == SKY || local_node->type == IMAGEP){
         object *lop;
         if((lop=GetObjectAtFrame(local_node,CurrentFrame)) != NULL){
           if(local_node->type == SKY)EditSkyDialog((sky *)lop,NO);
           else{
             if(lop->xip != NULL)
               lop->xip=EditObjectImageProcess(ghwnd_main,lop->xip);
             else MessageBeep(MB_OK);
           }
         }
         else AddCostumeTimeLine(local_node);
         UPDATETIMELINES(IDC_TIMEL_UPDATE)
       }
       else {
         if(tool != NOTOOL && tool != NODETOOL)DisengageTool();
         SelectedNode=local_node;
         UpdateSelectedActor(TRUE);
         ReDrawStageDisplay(TRUE);
         if(SelectedNode->type == GROUND){
           object *Op1;
           if((Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
             EditGroundDialog(Op1,NO);
           }
         }
       }
     };
     break;
   case IDM_ACTOR_INFO:
     SelectedInfo();
     break;
   case IDM_ACTOR_AXIS_SH:
     SnapTo(0);
     break;
   case IDM_ACTOR_AXIS_SNAP:
     Save_UndoA();
     DisengageTool();
     if(StageGrabIdentify()){
       SetNewPosition(NpointerX,NpointerY,NpointerZ,1);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
       else UpdateGLview(TRUE);
       if(coords_visible)UpdateRuler(0);
     }
     break;
   case IDM_ACTOR_SWAP_ALL:
   case IDM_ACTOR_SWAP:{
       object *Op1;
       if(SelectedNode != NULL){
         if(SelectedNode->type == NORMAL || SelectedNode->type == ROBOT){
           if((Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
             if(SelectSfxFileName(0,gszSCEfile,gszSCEdir,IDX_MISC_REPLACEMENT,
               "(*.MFX)|*.mfx|",ghwnd_main) == TRUE){
                Save_UndoA();
                if(item == IDM_ACTOR_SWAP_ALL)ReplaceObjectFile(Op1,gszSCEfile);
                else ReplaceOneObjectFile(SelectedNode,Op1,gszSCEfile);
                ReDrawStageDisplay(TRUE);
                if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
                else UpdateGLview(TRUE);
                UPDATETIMELINES(IDC_TIMEL_UPDATE)
              }
           }
           else SendPrgmQuery(IDQ_OLD2,0);
         }
         else SendPrgmQuery(IDQ_OLD2,0);
       }
       else SendPrgmQuery(IDQ_OLD1,0);
     }
     break;
   case IDM_ACTOR_RENAME:
     if(SelectedNode != NULL){
       Save_UndoA();
       if(SelectedNode->type == CAMERA &&  SelectedNode == FirstNp){
         SendPrgmQuery(IDQ_NOTFIRSTCAMERA,0);
         break;
       }
       LoadString(ghinst_main,IDX_MISC_RENAMEOBJECT,res_str,256);
       RequestCharString(64,SelectedNode->actorname,res_str);
       UPDATETIMELINES(IDC_TIMEL_UPDATE)
     }
     break;
   case IDM_ACTOR_QADD:
     ExecuteActorsMenu(hwnd);
     break;
   case IDM_ACTOR_QADD_MOD:
   case IDM_ACTOR_QADD_MORPH:
   case IDM_ACTOR_QADD_CAMERA:
   case IDM_ACTOR_QADD_TARGET:
   case IDM_ACTOR_QADD_LIGHT:
   case IDM_ACTOR_QADD_PATH:
   case IDM_ACTOR_QADD_SKY:
   case IDM_ACTOR_QADD_GROUND:
   case IDM_ACTOR_QADD_XIP:
   case IDM_ACTOR_QADD_ROBOT:
   case IDM_ACTOR_QADD_CSG:
   case IDM_ACTOR_QADD_DIRECTOR:
     AddQuickActor(item);
     break;
   case IDM_ACTOR_TOOL_GRABABS:
     if(tool != NOTOOL)DisengageTool();
     EDIT_ACTION=YES;
     Save_UndoA();
     InstantStageGrab();
     break;
   case IDM_ACTOR_TOOL_ROTATEABS:
     if(tool != NOTOOL)DisengageTool();
     EDIT_ACTION=YES;
     Save_UndoA();
     InstantStageRotate();
     break;
   case IDM_ACTOR_TOOL_SIZEABS:
     if(tool != NOTOOL)DisengageTool();
     EDIT_ACTION=YES;
     Save_UndoA();
     InstantStageXpnd();
     break;
   case IDM_ACTOR_TOOL_GRAB:
     if(tool != GRABBER)EngageTool(GRABBER);
     break;
   case IDM_ACTOR_TOOL_ROTATE:
     if(tool != ROTATOR)EngageTool(ROTATOR);
     break;
   case IDM_ACTOR_TOOL_SIZE:
     if(tool != SIZER)EngageTool(SIZER);
     break;
   case IDM_ACTOR_TOOL_NODE:
     if(tool != NODETOOL)EngageTool(NODETOOL);
     break;
   case IDM_EDIT_SPLINETWEEN:    // same as of tool button
   case IDM_ACTOR_TOOL_SHIFTER:
     if(tool != TRACKER)EngageTool(TRACKER);
     break;
   case IDM_ACTOR_TOOL_POSE:
     if(tool != SKELETOR)EngageTool(SKELETOR);
     break;

   case IDM_ACTOR_TOOL_UNDO:
     if(UNDO_TYPE == 1)AnimatorUndo();
     else              Restore_UndoA();
     break;
   case IDM_ACTOR_TOOL_PAN:
     if(tool != PAN)EngageTool(PAN);
     break;
   case IDM_ACTOR_TOOL_ZOOMIN:
     if(tool != INZOOM)EngageTool(INZOOM);
     break;
   case IDM_ACTOR_USEDIRECTOR:
     use_director ^= 1;
     if(use_director == 1)CheckMenuItem(GetMenu(ghwnd_main),
                          IDM_ACTOR_USEDIRECTOR,MF_CHECKED);
     else                 CheckMenuItem(GetMenu(ghwnd_main),
                          IDM_ACTOR_USEDIRECTOR,MF_UNCHECKED);
     if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
     else UpdateGLview(TRUE);
     break;
   case IDM_ACTOR_ACTCAMERA:
     if(SelectedNode != NULL){
       if(SelectedNode->type == CAMERA){
         SelectedCamera=SelectedNode;
         ReDrawStageDisplay(TRUE);
         if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
         else UpdateGLview(TRUE);
       }
     }
     else SendPrgmQuery(IDQ_OLD1,0);
     break;
   case IDM_EDIT_XIMAGE:
     DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_XIP),ghwnd_main,
        (DLGPROC)EditXipDlgProc);
     break;
   case IDM_PATH_EDIT:
     if(tool != PATHEDITOR)EngageTool(PATHEDITOR);
     break;
   case IDM_PATH_CURSORTO_START:
     SnapTo(1);
     break;
   case IDM_PATH_CURSORTO_END:
     SnapTo(2);
     break;
   case IDM_PATH_CURSORTO_POINT:
     SnapTo(3);
     break;
   case IDM_PATH_GRAPH_POSITION:
     EditPathMovement(1);
     break;
   case IDM_PATH_REVERSE:{
       object *Op1;
       if(SelectedNode != NULL){
         if(SelectedNode->type == PATH){
           if((Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
             EDIT_ACTION=YES;
             Save_UndoA();
             ReversePathDirection(Op1);
             ReDrawStageDisplay(TRUE);
             if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
             else UpdateGLview(TRUE);
             UpdateGLview(TRUE);
           }
           else SendPrgmQuery(IDQ_OLD2,0);
         }
         else SendPrgmQuery(IDQ_NOTAPATH,0);
       }
       else SendPrgmQuery(IDQ_OLD1,0);
     }
     break;

   case IDM_FRAME_GO_FIRST:
     CurrentFrame=1;
     ReDrawStageDisplay(TRUE);
     UpdateSelectedActor(FALSE);
     if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
     else UpdateGLview(TRUE);
//     SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
     SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
     break;
   case IDM_FRAME_GO_PREVIOUS:
     if(CurrentFrame > 1){
       if     (GetAsyncKeyState(VK_SHIFT  ) & 0x8000)CurrentFrame -=  5;
       else if(GetAsyncKeyState(VK_CONTROL) & 0x8000)CurrentFrame -= 10;
       else if(GetAsyncKeyState(VK_MENU   ) & 0x8000)CurrentFrame -= 20;
       else            CurrentFrame--;
       CurrentFrame=max(CurrentFrame,1);
       UpdateSelectedActor(FALSE);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL){
         PerspectiveView(0,1);
         UpdateWindow(ghwnd_view);
       }
       else UpdateGLview(TRUE);
//       SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
       SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
     }
     break;
   case IDM_FRAME_GO_NEXT:
     if(CurrentFrame < Nframes){
       if     (GetAsyncKeyState(VK_SHIFT  ) & 0x8000)CurrentFrame +=  5;
       else if(GetAsyncKeyState(VK_CONTROL) & 0x8000)CurrentFrame += 10;
       else if(GetAsyncKeyState(VK_MENU   ) & 0x8000)CurrentFrame += 20;
       else            CurrentFrame++;
       CurrentFrame=min(CurrentFrame,Nframes);
       UpdateSelectedActor(FALSE);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL){
         PerspectiveView(0,1);
         UpdateWindow(ghwnd_view);
       }
       else UpdateGLview(TRUE);
//       SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
       SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
     }
     break;
   case IDM_FRAME_GO_LAST:
     CurrentFrame=Nframes;
     UpdateSelectedActor(FALSE);
     ReDrawStageDisplay(TRUE);
     if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
     else UpdateGLview(TRUE);
//     SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
     SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
     break;
   case IDM_FRAME_GO_NEXTKEY:
     NextKey('N');
     if(IsWindow(ghwndRobotView))SendMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
     break;
   case IDM_FRAME_GO_LASTKEY:
     NextKey('L');
     if(IsWindow(ghwndRobotView))SendMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
     break;
   case IDM_FRAME_GO_FRAME:{
       int id;
       char ttt[64];
       LoadString(ghinst_main,IDX_MISC_GOTOFRAME,res_str,256);
       LoadString(ghinst_main,IDX_MISC_FRAMEID,ttt,64);
       id=RequestNumEntry(CurrentFrame,1,Nframes,res_str,ttt);
       if(id != CurrentFrame && id > 0 && id <= Nframes){
         CurrentFrame=(long)id;
         ReDrawStageDisplay(TRUE);
         UpdateSelectedActor(FALSE);
         if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
         else UpdateGLview(TRUE);
//         SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
         SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
         if(IsWindow(ghwndRobotView))SendMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
       }
     }
     break;
   case IDM_FRAME_HIDDENLINE:
    if(!MakeScript(3,1))SendPrgmQuery(IDQ_SCRIPTOPENFAIL,0);
    else LoadRenderer(0);
    break;
   case IDM_FRAME_QUICKRENDER:
   case IDM_FRAME_RENDERIN:
    if(!MakeScript(3,1))SendPrgmQuery(IDQ_SCRIPTOPENFAIL,0);
    else LoadRenderer(3);
    break;
   case IDM_FRAME_RENDERNOW:
     LoadRenderer(2);
     break;
   case IDM_FRAME_RENDER:
     LoadRenderer(1);
     break;
   case IDM_FRAME_SHOW:{
       int result;
       char modulename[255],cp[128],cn[128];
       GetPrivateProfileString("VIEWER","FILE","$$$",
          cn,sizeof(cn),IniFilename);
       GetPrivateProfileString("VIEWER","PATH","$$$",
          cp,sizeof(cp),IniFilename);
       if(strncmp(cn,"$$$",3) == 0 || strncmp(cp,"$$$",3) == 0){
         SendPrgmQuery(IDQ_NOLASTIMAGE,0);
       }
       else LoadUtility("winview.exe","OFX:ViewClass",107);
     }
     break;
   case IDM_FRAME_PLAY_WIRE:
     AnimationPlayer(0,hwnd);
     break;
   case IDM_FRAME_PLAY_QUICK:
     if((GetKeyState(VK_CONTROL) & 0x8000) != 0x8000){
       AnimationPlayer(2,hwnd);
       break;
     }
   case IDM_FRAME_PLAY_FULL:
     AnimationPlayer(1,hwnd);
     break;
   case IDM_FRAME_SETNUMBER:
     if(busy_flag)break;
     if(tool == NOTOOL){
       int id;
       id=Nframes;
       Save_UndoA();
       busy_flag=TRUE;
       SetNumberOfFrames(0);
       busy_flag=FALSE;
       if(id != Nframes){
         if(CurrentFrame > Nframes)CurrentFrame=Nframes;
//         SetScrollRange(ghwnd_xscl,SB_CTL,1,(int)Nframes,FALSE);
         SendMessage(ghwnd_xscl,TBM_SETRANGE,(WPARAM)TRUE,(LPARAM)MAKELONG(1,Nframes));
         UPDATETIMELINES(IDC_TIMEL_UPDATE)
         if(ghwndTimeline != NULL && IsZoomed(ghwndTimeline)){
//           SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,FALSE);
           SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
           break;
         }
         else{
//           SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
           SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
         }
         ReDrawStageDisplay(TRUE);
         if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
         else UpdateGLview(TRUE);
         if(coords_visible)UpdateRuler(0);
       }
     }
     else SendPrgmQuery(IDQ_NOT_IN_TOOL,0);
     break;
   case IDM_FRAME_PREVIEW_MAKE:
     if(MakePreview() == FAIL)break;
     break;
   case IDM_STOP_TOOL:
     if(tool != NOTOOL)DisengageTool();
     break;
   case IDM_QUIET_EDITCOSTUME:{
       BOOL draw=FALSE;
       object *Op;
       Op=GetObjectAtFrame(SelectedNode,CurrentFrame);
       if(Op != NULL){
         switch (Op->type){
           case NORMAL:
           case ANIMOBJ:
             Save_UndoA();
             DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_COSTUME),
                      ghwnd_main,(DLGPROC)EditObjectDlgProc,(LPARAM)Op);
             draw=TRUE; break;
           case CSG:
           case PATH:
           case TARGET:
           case ROBOT:
             Save_UndoA();
             EditObject(SelectedNode,(short)Op->lastframe,
                                     (short)Op->lastframe,NO,NO);
             draw=TRUE;
             break;
           case LIGHT:
             Save_UndoA();
             EditLightDialog(Op,NO);
             draw=TRUE;
             break;
           case GROUND:
             Save_UndoA();
             EditGroundDialog(Op,NO);
             draw=TRUE;
             break;
           default: break;
         }
         if(draw){
           CheckRecursiveFollow(SelectedNode,Op->firstframe,Op->lastframe);
           UpdateSelectedActor(FALSE);
           ReDrawStageDisplay(TRUE);
           if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
           else UpdateGLview(TRUE);
           UPDATETIMELINES(IDC_TIMEL_UPDATE)
         }
       }
     }
     break;
   case IDM_QUIET_EDITLIGHT:
   case IDM_QUIET_EDITLENS:
     LensSettings();
     ReDrawStageDisplay(TRUE);
     UPDATETIMELINES(IDC_TIMEL_UPDATE)
     break;
   case IDM_TIMELINE_LINE_ADDP:
     if(SelectedNode != NULL){
       int id;
       Save_UndoA();
       if(CheckForControl(SelectedNode,'P') == YES){
         id=SendPrgmQuery(IDQ_DELETECHANNEL,1);
         if(id == IDYES){
           DeletePosition(SelectedNode,CurrentFrame);
           UPDATETIMELINES(IDC_TIMEL_UPDATE)
         }
       }
       else MakeControl(SelectedNode,'P',ghwnd_main);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
       else UpdateGLview(TRUE);
     }
     else SendPrgmQuery(IDQ_NOACTORSELECTED,0);
     break;
   case IDM_TIMELINE_LINE_ADDR:
     if(SelectedNode != NULL){
       int id;
       Save_UndoA();
       if(CheckForControl(SelectedNode,'A') == YES){
         id=SendPrgmQuery(IDQ_DELETECHANNELX,1);
         if(id == IDYES){
           DeleteAlign(SelectedNode,CurrentFrame);
           UPDATETIMELINES(IDC_TIMEL_UPDATE)
         }
       }
       else MakeControl(SelectedNode,'A',ghwnd_main);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
       else UpdateGLview(TRUE);
     }
     else SendPrgmQuery(IDQ_NOACTORSELECTED,0);
     break;
   case IDM_TIMELINE_LINE_ADDX:
     if(SelectedNode != NULL){
       int id;
       Save_UndoA();
       if(CheckForControl(SelectedNode,'Z') == YES){
         id=SendPrgmQuery(IDQ_DELETECHANNELX,1);
         if(id == IDYES){
           DeleteSize(SelectedNode,CurrentFrame);
           UPDATETIMELINES(IDC_TIMEL_UPDATE)
         }
       }
       else MakeControl(SelectedNode,'Z',ghwnd_main);
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
       else UpdateGLview(TRUE);
     }
     else SendPrgmQuery(IDQ_NOACTORSELECTED,0);
     break;
   case IDM_TIMELINE_KEY_ADDP:
     Save_UndoA();
     MakeKey('P',CurrentFrame,SelectedNode);
     break;
   case IDM_TIMELINE_KEY_ADDR:
     Save_UndoA();
     MakeKey('A',CurrentFrame,SelectedNode);
     break;
   case IDM_TIMELINE_KEY_ADDX:
     Save_UndoA();
     MakeKey('X',CurrentFrame,SelectedNode);
     break;
   case IDM_TIMELINE_KEY_RESETR:
   case IDM_TIMELINE_KEY_RESETX:
   case IDM_TIMELINE_KEY_RESETP:{
       object   *Op=NULL;
       position *Pp=NULL;
       align    *Ap=NULL;
       size     *Xp=NULL;
       Save_UndoA();
       GetTimelinesAtFrame(SelectedNode,CurrentFrame,&Op,&Pp,&Ap,&Xp);
       if     (item == IDM_TIMELINE_KEY_RESETP && Pp != NULL){
         Pp->finish[0]=Pp->finish[1]=Pp->finish[2]=0;
       }
       else if(item == IDM_TIMELINE_KEY_RESETR && Ap != NULL){
         Ap->theta=0.0; Ap->phi=0.0; Ap->alpha=0.0;
       }
       else if(item == IDM_TIMELINE_KEY_RESETX && Xp != NULL){
         Xp->Sx=1.0; Xp->Sy=1.0; Xp->Sz=1.0;
       }
       ReDrawStageDisplay(TRUE);
       if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
       else UpdateGLview(TRUE);
     }
     break;
   case IDM_TIMELINE_KEY_SHIFT:
     ShiftAllKeys();
     break;
   case IDM_VIEW_DRAW_BOUNDED:
     if(DrawStatus == 0){
       char tt[128];
       DrawStatus=AllDrawStatus;
       EnableMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_SELECTED,MF_ENABLED);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_DESELECTED,MF_ENABLED);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_ALL,MF_ENABLED);
       LoadString(ghinst_main,IDX_MISC_DRAWBOUNDED,tt,128);
       ModifyMenu(GetMenu(ghwnd_main),IDM_VIEW_DRAW_BOUNDED,
           MF_BYCOMMAND|MF_STRING,IDM_VIEW_DRAW_BOUNDED,tt);
     }
     else{
       char tt[128];
       DrawStatus=0;
       EnableMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_SELECTED,MF_GRAYED);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_DESELECTED,MF_GRAYED);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_ALL,MF_GRAYED);
       LoadString(ghinst_main,IDX_MISC_DRAWFULL,tt,128);
       ModifyMenu(GetMenu(ghwnd_main),IDM_VIEW_DRAW_BOUNDED,
           MF_BYCOMMAND|MF_STRING,IDM_VIEW_DRAW_BOUNDED,tt);
     }
     ReDrawStageDisplay(TRUE);
     break;
   case IDM_VIEW_DRAW_SELECTED:
   case IDM_VIEW_DRAW_DESELECTED:
   case IDM_VIEW_DRAW_ALL:
     switch(item){
       case IDM_VIEW_DRAW_SELECTED:   AllDrawStatus=1; break;
       case IDM_VIEW_DRAW_DESELECTED: AllDrawStatus=2; break;
       case IDM_VIEW_DRAW_ALL:        AllDrawStatus=3; break;
     }
     CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_ALL       ,MF_UNCHECKED);
     CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_SELECTED  ,MF_UNCHECKED);
     CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_DESELECTED,MF_UNCHECKED);
     if(AllDrawStatus == 1)CheckMenuItem(GetMenu(ghwnd_main),
                           IDM_VIEW_DRAW_SELECTED  ,MF_CHECKED);
     if(AllDrawStatus == 2)CheckMenuItem(GetMenu(ghwnd_main),
                           IDM_VIEW_DRAW_DESELECTED,MF_CHECKED);
     if(AllDrawStatus == 3)CheckMenuItem(GetMenu(ghwnd_main),
                           IDM_VIEW_DRAW_ALL       ,MF_CHECKED);
     if(DrawStatus > 0)DrawStatus=AllDrawStatus;
     ReDrawStageDisplay(TRUE);
     break;
   case IDM_VIEW_LOCKBOUNDED:
     local_node=SelectedNode;
     if(local_node != NULL){
       if(local_node->type == NORMAL){
         int id;
         object *Op1;
         char objname[256];
         LoadString(ghinst_main,IDX_MISC_QUICKONLYFOR,res_str,256);
         sprintf(objname,res_str,SelectedNode->actorname);
         LoadString(ghinst_main,IDX_MISC_QUICKDRAW,res_str,256);
         id=MessageBox(NULL,objname,res_str,
            MB_YESNOCANCEL|MB_ICONQUESTION|MB_TASKMODAL|MB_DEFBUTTON3);
         if(id == IDCANCEL)break;
         Op1=local_node->fobj; while (Op1 != NULL){
           if(id == IDYES)Op1->quick_only=1;
           else           Op1->quick_only=0;
           Op1=Op1->next;
         }
         ReDrawStageDisplay(TRUE);
       }
       else SendPrgmQuery(IDQ_OLD2,0);
     }
     else SendPrgmQuery(IDQ_OLD1,0);
     break;
   case IDM_VIEW_GLOBALQUICKDRAW:
     global_quickdraw ^= 1;
     if(global_quickdraw)CheckMenuItem(GetMenu(ghwnd_main),
                         IDM_VIEW_GLOBALQUICKDRAW,MF_CHECKED);
     else                CheckMenuItem(GetMenu(ghwnd_main),
                         IDM_VIEW_GLOBALQUICKDRAW,MF_UNCHECKED);
     break;
   case IDM_VIEW_DRAW_GRID:
     draw_grid_on ^= 1;
     if(draw_grid_on)
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_CHECKED);
     else
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_UNCHECKED);
     ReDrawStageDisplay(TRUE);
     break;
   case IDM_VIEW_GRID:
     grid_on ^= 1;
     if(grid_on)CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_GRID,MF_CHECKED);
     else       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_GRID,MF_UNCHECKED);
//     if(draw_grid_on)ReDrawStageDisplay(TRUE);
     break;
   case IDM_VIEW_DEFINEGRID:
     SetUpGrid();
     if(draw_grid_on)
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_CHECKED);
     else
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_UNCHECKED);
     ReDrawStageDisplay(TRUE);
     break;
   case IDM_ZOOM_IN:
     Zoom(ZOOMIN,1);
     break;
   case IDM_ZOOM_OUT:
     Zoom(ZOOMOUT,1);
     break;
   case IDM_ZOOM_CENTRE:
     Zoom(RECENTRE,1);
     break;
   case IDM_VIEW_RESTORE:
     GetTriview(TRUE);
     NpointerX=NpointerY=NpointerZ=0;
     ReDrawStageDisplay(TRUE);
     if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
     else UpdateGLview(TRUE);
     UpdateRuler(0);
     break;
   case IDM_VIEW_FULL:
     Zoom(CENTREWORLD,1);
     break;
   case IDM_VIEW_FEEDBACK_CAMERA:
     show_camera_fov ^= 1;
     if(show_camera_fov)CheckMenuItem(GetMenu(ghwnd_main),
                        IDM_VIEW_FEEDBACK_CAMERA,MF_CHECKED);
     else               CheckMenuItem(GetMenu(ghwnd_main),
                        IDM_VIEW_FEEDBACK_CAMERA,MF_UNCHECKED);
     ReDrawStageDisplay(TRUE);
     break;
   case IDM_VIEW_FEEDBACK_LIGHT:
     show_light_cones ^= 1;
     if(show_light_cones)CheckMenuItem(GetMenu(ghwnd_main),
                         IDM_VIEW_FEEDBACK_LIGHT,MF_CHECKED);
     else                CheckMenuItem(GetMenu(ghwnd_main),
                         IDM_VIEW_FEEDBACK_LIGHT,MF_UNCHECKED);
     ReDrawStageDisplay(TRUE);
     break;
   case IDM_WINDOW_BIGCAMERA:
     BigCameraView ^= 1;
     if(BigCameraView)CheckMenuItem(GetMenu(ghwnd_main),
                      IDM_WINDOW_BIGCAMERA,MF_CHECKED);
     else             CheckMenuItem(GetMenu(ghwnd_main),
                      IDM_WINDOW_BIGCAMERA,MF_UNCHECKED);
     if(View == ONEVIEW){
       RECT rc;
       GetClientRect(ghwnd_main,&rc);
       SendMessage(ghwnd_main,WM_SIZE,0,MAKELPARAM(rc.right,rc.bottom));
       if(BigCameraView)ToolboxSetPosition(2);
       else             ToolboxSetPosition(1);
     }
     break;
   case IDM_WINDOW_SWITCHWINDOWS:
     SwitchWindow(-1);
     break;
   case IDM_VIEW_SWAPVIEWPOINT:
     WindowBox_view ^= 1;
     SetUpWindowBoxView(WindowBox_view);
     SwapViewBitmaps();
     ReDrawStageDisplay(TRUE);
     break;
   case IDM_WINDOW_SWITCHWINDOW0:
     SwitchWindow(0);
     break;
   case IDM_WINDOW_SWITCHWINDOW1:
     SwitchWindow(1);
     break;
   case IDM_WINDOW_CLOSE:
     CloseUtilities(0);
     break;
   case IDM_WINDOW_COORDS:{
       HMENU hmenu;
       int cv[2]={SW_HIDE,SW_SHOWNA};
       hmenu=GetMenu(ghwnd_main);
       coords_visible ^=1;
       if(IsWindow(ghwndCoord1))ShowWindow(ghwndCoord1,cv[coords_visible]);
     }
     break;
   case IDM_WINDOW_TOOLLEFT:
     break;
   case IDM_WINDOW_RESET:
     Window0sizeX=TriviewSpaceX/2;
     Window0sizeY=TriviewSpaceY/2;
     if(View == TRIVIEW)FormatScreen();
     break;
   case IDM_FILE_KEYFRAMERDOCK:
     i=0;
     if(IsWindow(ghwndTimeline)){
       if(IsWindowVisible(ghwndTimeline)){
         i=1; /* window was visible */
         TimelineEditor(); /* Hide It */
       }
       if(keyframer_docked)FormatScreen();
       DestroyTimelineEditor();
     }
     keyframer_docked ^=1;
     if(keyframer_docked)
       CheckMenuItem(GetMenu(ghwnd_main),IDM_FILE_KEYFRAMERDOCK,MF_CHECKED);
     else
       CheckMenuItem(GetMenu(ghwnd_main),IDM_FILE_KEYFRAMERDOCK,MF_UNCHECKED);
     if(i == 1){
       if(keyframer_docked)keyframer_docked_and_visible=1;
       FormatScreen();
       TimelineEditor();
     }
     break;
   case IDM_WINDOW_ACTIVATE: {
       RECT  r;
       POINT p;
       HDC hdc;
       int i;
       if(View == ONEVIEW){
         if(ActiveView == TRIFRONT)
           PostMessage(hwnd,WM_COMMAND,IDM_WINDOW_SWITCHWINDOW1,0);
         else
           PostMessage(hwnd,WM_COMMAND,IDM_WINDOW_SWITCHWINDOW0,0);
         break;
       }
       hdc=GetDC(ghwnd_main);
       FocusActiveWindow(hdc,FALSE);
       ActiveView++; if(ActiveView > 2)ActiveView=0;
       ghwnd_current=ghwnd_triview[ActiveView];
       FocusActiveWindow(hdc,TRUE);
       ReleaseDC(ghwnd_main,hdc);
     }
     break;
   case IDM_COORDS_DEFINE:
     UpdateRuler(3);
     break;
   case IDM_COORDS_ORIGIN:
     UpdateRuler(2);
     break;
   case IDM_COORDS_RELATIVE:
     UpdateRuler(1);
     break;
   case IDM_COORDS_MEASURE:
     UpdateRuler(0);
     break;
   case IDM_PAN_UP:
     Nudge(item);
     break;
   case IDM_PAN_DOWN:
     Nudge(item);
     break;
   case IDM_PAN_LEFT:
     Nudge(item);
     break;
   case IDM_PAN_RIGHT:
     Nudge(item);
     break;
   case IDM_VIEW_FEEDBACK_CURSOR:{
       char tt[128];
       Draw3dCursor();
       CursorToggle ^= 1;
       Draw3dCursor();
       if(CursorToggle){
         LoadString(ghinst_main,IDX_MISC_CROSSHAIR,tt,128);
         ModifyMenu(GetMenu(ghwnd_main),IDM_VIEW_FEEDBACK_CURSOR,
          MF_BYCOMMAND|MF_STRING,IDM_VIEW_FEEDBACK_CURSOR,tt);
       }
       else{
         LoadString(ghinst_main,IDX_MISC_CROSSWIRE,tt,128);
         ModifyMenu(GetMenu(ghwnd_main),IDM_VIEW_FEEDBACK_CURSOR,
          MF_BYCOMMAND|MF_STRING,IDM_VIEW_FEEDBACK_CURSOR,tt);
       }
     }
     break;
   case IDM_SELECT_ACTOR:
     PickNode(NORMAL,99);
     break;
   case IDM_SELECT_PATH:
     PickNode(PATH,99);
     break;
   case IDM_SELECT_LIGHT:{
       object *Op1;
       if((GetAsyncKeyState(VK_CONTROL) & 0x8000) &&
           SelectedNode != NULL && SelectedNode->type == LIGHT &&
             (Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
         EditLightDialog(Op1,NO);
       }
       else PickNode(LIGHT,99);
     }
     break;
   case IDM_SELECT_TARGET:
     PickNode(TARGET,99);
     break;
   case IDM_SELECT_CAMERA:
     PickNode(CAMERA,99);
     break;
   case IDM_SELECT_ROBOT:
     PickNode(ANIMOBJ,ROBOT);
     break;
   case IDM_SELECT_CSG:
     PickNode(CSG,99);
     break;
   case IDM_SELECT_GROUND:{
       node *oldnode;
       object *Op1;
       oldnode=SelectedNode;
       PickNode(GROUND,99);
       if(SelectedNode != NULL && SelectedNode == oldnode &&
          SelectedNode->type == GROUND){
         if((Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
           EditGroundDialog(Op1,NO);
         }
       }
     }
     break;
   case IDM_HELP_HOME:
#if __LITE__
     LoadString(ghinst_main,IDX_MISC_HOMEPAGEL,res_str,256);
#else
     LoadString(ghinst_main,IDX_MISC_HOMEPAGE,res_str,256);
#endif
     strcpy(lname,gszHomeDir);
     ShellExecute(ghwnd_main,NULL,res_str,NULL,lname,0);
     break;
   case IDM_HELP_SUPPORT:
#if __LITE__
     LoadString(ghinst_main,IDX_MISC_SUPPORTPAGEL,res_str,256);
#else
     LoadString(ghinst_main,IDX_MISC_SUPPORTPAGE,res_str,256);
#endif
     strcpy(lname,gszHomeDir);
     ShellExecute(ghwnd_main,NULL,res_str,NULL,lname,0);
     break;
   case IDM_HELP:{
       char res_str1[256],res_str2[256];
       LoadString(ghinst_main,IDX_MISC_HELPFILENAME,res_str,256);
       strcpy(lname,gszHomeDir); strcat(lname,"help\\");
       strcpy(res_str1,lname);
       strcat(res_str1,res_str);
#if !__HELP_WS__
       ShellExecute(ghwnd_main,NULL,res_str1,
                    NULL,lname,SW_SHOW);
#else
       {
         HH_WINTYPE hhwin;
         char localname[32],localname1[32]=">";
         sprintf(localname,"AHelp%ld",rand());
         strcat(localname1,localname);
         memset(&hhwin,0,sizeof(HH_WINTYPE));
         hhwin.cbStruct=sizeof(HH_WINTYPE);
         hhwin.fsValidMembers = HHWIN_PARAM_SHOWSTATE |
                                HHWIN_PARAM_PROPERTIES|
                                HHWIN_PARAM_CUR_TAB|
                                HHWIN_PARAM_EXPANSION|
                                HHWIN_PARAM_RECT;
         hhwin.rcWindowPos.top=0;
         hhwin.rcWindowPos.left=0;
         hhwin.rcWindowPos.right=700;
         hhwin.rcWindowPos.bottom=500;
         hhwin.nShowState=SW_SHOW;
         hhwin.pszType=localname;
         LoadString(ghinst_main,IDX_MISC_HELPTITLE1,res_str,256);
         hhwin.pszCaption=res_str;
         hhwin.fNotExpanded = FALSE;
         hhwin.fsWinProperties = HHWIN_PROP_TRI_PANE|HHWIN_PROP_TAB_SEARCH;
         hhwin.curNavType=HHWIN_NAVTYPE_TOC;
         strcpy(res_str2,res_str1);
         strcat(res_str2,"::toc.hhc");
         hhwin.pszToc=res_str2;
         HtmlHelp(ghwnd_main,NULL,HH_SET_WIN_TYPE,(DWORD)&hhwin);
         hWndHelp=HtmlHelp(ghwnd_main,localname1,HH_DISPLAY_TOPIC,(DWORD)res_str1);
       }
#endif
     }
     break;
   case IDM_HELP_GENERAL:{
       char res_str1[256],res_str2[256];
       LoadString(ghinst_main,IDX_MISC_HELPFILENAME1,res_str,256);
       strcpy(lname,gszHomeDir); strcat(lname,"help\\");
       strcpy(res_str1,lname);
       strcat(res_str1,res_str);
#if !__HELP_WS__
       ShellExecute(ghwnd_main,NULL,res_str1,
                    NULL,lname,SW_SHOW);
#else
       {
         HH_WINTYPE hhwin;
         char localname[32],localname1[32]=">";
         sprintf(localname,"GHelp%ld",rand());
         strcat(localname1,localname);
         memset(&hhwin,0,sizeof(HH_WINTYPE));
         hhwin.cbStruct=sizeof(HH_WINTYPE);
         hhwin.fsValidMembers = HHWIN_PARAM_SHOWSTATE |
                                HHWIN_PARAM_PROPERTIES|
                                HHWIN_PARAM_CUR_TAB|
                                HHWIN_PARAM_EXPANSION|
                                HHWIN_PARAM_RECT;
         hhwin.rcWindowPos.top=0;
         hhwin.rcWindowPos.left=0;
         hhwin.rcWindowPos.right=700;
         hhwin.rcWindowPos.bottom=500;
         hhwin.nShowState=SW_SHOW;
         hhwin.pszType=localname;
         LoadString(ghinst_main,IDX_MISC_HELPTITLE2,res_str,256);
         hhwin.pszCaption=res_str;
         hhwin.fNotExpanded = FALSE;
         hhwin.fsWinProperties = HHWIN_PROP_TRI_PANE|HHWIN_PROP_TAB_SEARCH;
         hhwin.curNavType=HHWIN_NAVTYPE_TOC;
         strcpy(res_str2,res_str1);
         strcat(res_str2,"::toc.hhc");
         hhwin.pszToc=res_str2;
         hhwin.hwndCaller=ghwnd_main;
         HtmlHelp(ghwnd_main,NULL,HH_SET_WIN_TYPE,(DWORD)&hhwin);
         hWndHelp=HtmlHelp(ghwnd_main,localname1,HH_DISPLAY_TOPIC,(DWORD)res_str1);
       }
#endif
     }
//     strcpy(lname,gszHomeDir); strcat(lname,"help\\readme.hlp");
//     WinHelp(hwnd,lname,HELP_INDEX,0L);
     break;
   case IDM_HELP_HELP:
//     WinHelp(hwnd,"WINHELP.HLP",HELP_INDEX,0L);
     break;
   case IDM_HELP_ABOUT:
       DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_ABOUT),ghwnd_main,
          (DLGPROC)AboutDlgProc);
     break;
   case IDM_CALLBACK_RELOAD:{
       if(!EditUpdateInUse)break;
       EditUpdateInUse=FALSE;
       EnableToolPannels(ALL_PANNELS,TRUE);
       if(ObjectInModeler != NULL){
         ReplaceObjectFile(ObjectInModeler,
                           ObjectInModeler->name);
         ReDrawStageDisplay(TRUE);
       }
       ObjectInModeler=NULL;
     }
     break;
   case IDM_CALLBACK_NORELOAD:
     EnableToolPannels(ALL_PANNELS,TRUE);
     EditUpdateInUse=FALSE;
     ObjectInModeler=NULL;
     break;
   case IDM_EDIT_MODEL:{
       HWND hWndModeler;
       object *Op1;
       if(EditUpdateInUse){
         SendPrgmQuery(IDQ_MODELERINUSE,0);
         break;
       }
       if(SelectedNode != NULL){
         if(SelectedNode->type == NORMAL || SelectedNode->type == ROBOT){
           if((Op1=GetObjectAtFrame(SelectedNode,CurrentFrame)) != NULL){
             WritePrivateProfileString("CALLBACK","FILE",Op1->name,
                                       IniFilename);
             ObjectInModeler=Op1;
             ExecuteModeler();
             if((hWndModeler=FindWindow("OFX:ModelerClass",NULL)) != NULL){
               PostMessage(hWndModeler,WM_COMMAND,IDM_CALLBACK_LOAD_MODELER,0);
               EditUpdateInUse=TRUE;
               EnableToolPannels(ALL_PANNELS,FALSE);
             }
             else SendPrgmQuery(IDQ_NOLOADMODELER,0);
           }
           else{
             SendPrgmQuery(IDQ_OLD2,0);
           }
         }
         else{
           SendPrgmQuery(IDQ_OLD2,0);
         }
       }
       else{
         SendPrgmQuery(IDQ_OLD1,0);
       }
     }
     break;
   case IDM_GLCAMERA_PREFERENCES:
   case IDM_GLCAMERA_HELP:
     if(ghwndOpenGLview != NULL){
       SendMessage(ghwndOpenGLview,WM_COMMAND,item,0);
     }
     break;

   default: break;
 }
}

static void ExecuteKeyboardKeyCommand(int key){
  switch(key){
    case VK_ESCAPE:{
        HWND hWnd;
        SendMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
        if((hWnd=FindWindow("OFX:RenderClass",NULL)) != NULL){
          PostMessage(hWnd,WM_CALLBACK_ABANDON,0,0);
        }
      }
      break;
    case VK_PRIOR:
//      Nudge(IDM_PAN_UP);
      break;
    case VK_NEXT:
//      Nudge(IDM_PAN_DOWN);
      break;
    case VK_HOME:
//      Nudge(IDM_PAN_LEFT);
      break;
    case VK_END:
//      Nudge(IDM_PAN_RIGHT);
      break;
    case VK_INSERT:      /* bank */
//      Nudge(IDM_ZOOM_IN);
      break;
    case VK_DELETE:      /* bank */
//      Nudge(IDM_ZOOM_OUT);
      break;
    default: break;
  }
}

// code to handle the last used files commands on the file menu

void InitializeLastFilesList(void){
 int i,id;
 HMENU hmenu;
 char iname[32],lname[256];
 hmenu=GetSubMenu(GetMenu(ghwnd_main),0);
 for(nLastFiles=0,i=0;i<NFILESLIST;i++){
   LastFilesList[i].present=FALSE;
   sprintf(iname,"STARTFILE%ld",i);
   GetPrivateProfileString(IniSection,iname,"$$$",
                            LastFilesList[i].file,256,IniFilename);
   if(strncmp(LastFilesList[i].file,"$$$",3) == 0)break;
   LastFilesList[i].present=TRUE;
   nLastFiles++;
   id=LASTFILECMD+1+i;
   sprintf(lname,"&%ld %s",i+1,FileInPath(LastFilesList[i].file));
   InsertMenu(hmenu,IDM_FILE_EXIT,MF_ENABLED | MF_STRING,id,lname);
 }
 if(nLastFiles > 0)InsertMenu(hmenu,IDM_FILE_EXIT,MF_SEPARATOR,LASTFILECMD,0);
}

static void UpdateLastFilesList(char *newname){
 int i,k,id;
 char iname[32],lname[256];
 HMENU hmenu;
 hmenu=GetSubMenu(GetMenu(ghwnd_main),0);
 if(nLastFiles > 0){
   for(i=0;i<nLastFiles;i++)
     if(strcmp(LastFilesList[i].file,newname) == 0)return;
   RemoveMenu(hmenu,LASTFILECMD,MF_BYCOMMAND);
   for(i=0;i<nLastFiles;i++)RemoveMenu(hmenu,LASTFILECMD+1+i,MF_BYCOMMAND);
   nLastFiles=min(nLastFiles+1,NFILESLIST);
   for(i=nLastFiles-1;i>0;i--){
     LastFilesList[i].present=LastFilesList[i-1].present;
     strcpy(LastFilesList[i].file,LastFilesList[i-1].file);
   }
 }
 else nLastFiles=1;
 LastFilesList[0].present=TRUE;
 strcpy(LastFilesList[0].file,newname);
 for(i=0;i<nLastFiles;i++){
   id=LASTFILECMD+1+i;
   sprintf(lname,"&%ld %s",i+1,FileInPath(LastFilesList[i].file));
   InsertMenu(hmenu,IDM_FILE_EXIT,MF_ENABLED | MF_STRING,id,lname);
//              FileInPath(LastFilesList[i].file));
 }
 InsertMenu(hmenu,IDM_FILE_EXIT,MF_SEPARATOR,LASTFILECMD,0);
 for(i=0;i<nLastFiles;i++){
   sprintf(iname,"STARTFILE%ld",i);
   WritePrivateProfileString(IniSection,iname,
                             LastFilesList[i].file,IniFilename);
 }
 return;
}

