/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file astruct.h */

#if __SC__
#pragma  ZTC align 1
#endif

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

#ifdef __X__NT__
#define _LINE_DRAW 1
#define X__SetPixel(a,b,c,d) SetPixelV(a,b,c,d)
#else
#define X__SetPixel(a,b,c,d) SetPixel(a,b,c,d)
#endif
#ifdef __X__NT__
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#else
#define X__Malloc(s)    GlobalAlloc(GMEM_FIXED,s)
#define X__Free(p)      GlobalFree((HGLOBAL)p)
#define X__Realloc(p,s) GlobalReAlloc((HGLOBAL)p,s,GMEM_MOVEABLE)
#endif

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                         }
#define UPDATETIMELINES(x)  { \
    if(ghwndTimeline != NULL)SendMessage(ghwndTimeline,WM_COMMAND,x,0); \
                            }

#define ALL_PANNELS     0x0FFF
#define MAIN_WINDOW     0x0001
#define SELECT_PANNEL   0x0002
#define COMMAND_PANNEL  0x0004
#define ZOOM_PANNEL     0x0008
#define TOOL_PANNEL     0x0010
#define TIME_PANNEL     0x0020
#define COORD_PANNEL    0x0040
#define KEYS_WINDOW     0x0080

#define NSWITCH         2

typedef long point[3];
typedef long Sedge[2];
typedef double vector[3];
typedef float Svertex[3];

typedef struct tagANI_STRUCTURE {
 char  *gszHomeDir;
 short *Nnodes,*Nground,*Nrobots,
       *Nskys,*Ncameras;
 struct NODE **MainNp,**FirstNp,**SelectedCamera,**SelectedNode;
 long   *NpointerX,*NpointerY,*NpointerZ,
        *TVpointX, *TVpointY, *TVpointZ,
        *TVsizeX,  *TVsizeY,  *TVsizeZ;
 long   *Nframes,*CurrentFrame;
 double *ruler;
 long   *rulerx,*rulery,*rulerz,*orulerx,*orulery,*orulerz;
 long   version;
 long   *Ndirectors;
 long   *dummyv[62];
 short (*fpGetTransform)(short,
             long, double,
             struct NODE *,  struct OBJECT *,
             point, point,
             double *, double *, double *, double *, double *,double *,
             short *, double *);
 void (*fp_scal)(double [4][4], double, double, double);
 void (*fp_rotz)(double [4][4], double);
 void (*fp_roty)(double [4][4], double);
 void (*fp_rotx)(double [4][4], double);
 void (*fp_tram)(double [4][4], double, double, double);
 void (*fp_m4by4)(double [4][4], double [4][4], double [4][4]);
 void (*fp_m4by1)(double [4][4], double, double, double,
            double *, double *, double *);
 void (*fp_null_transform)(double [4][4]);
 void (*fp_c4to4)(double [4][4], double [4][4]);
 void (*fp_arbitrary_rotate)(double, point, point, double [4][4]);
 void (*fp_L_cross)(point, point, point, point);
 void (*fpDrawPerspectiveView)(int, int, HDC, HWND, HPEN);
 void (*fpDeleteActor)(struct NODE *, long);
 struct OBJECT * (*fpCreateActor)(struct NODE *, long, long);
 void (*fpDeleteSize)(struct NODE *, long);
 struct SIZE * (*fpCreateSize)(struct NODE *, long, long);
 void (*fpDeleteAlign)(struct NODE *, long);
 struct ALIGN * (*fpCreateAlign)(struct NODE *, long, long);
 void (*fpDeletePosition)(struct NODE *, long);
 struct POSITION * (*fpCreatePosition)(struct NODE *, long, long);
 void (*fpDeleteSky)(struct NODE *, long);
 struct SSKY * (*fpCreateSky)(struct NODE *, long, long);
 void (*fpDeleteNode)(struct NODE *);
 struct NODE * (*fpCreateNode)(void);
 struct tagPATHEDITCONTROL * (*fpCreatePathEditControlPoint)(int *,
                 struct tagPATHEDITCONTROL **,
                 struct tagPATHEDITCONTROL *,
                 struct tagPATHEDITCONTROL *,
                 int, double);
 void (*fpDeletePathEditControlPoint)(int *,
                 struct tagPATHEDITCONTROL **,
                 struct tagPATHEDITCONTROL *);
 struct PATHPOINT * (*fpAppendPathPoint)(struct PATHPOINT *);
 void (*fpCreatePathPoint)(void);
 void (*fpDeletePathPoint)(void);
 struct PATHPOINT * (*fpCreateIPointPath)(short, struct NODE *, short,
                            short *, short *);
 void  (*fpSetNumberOfFrames)(short);
 int   (*fpChooseFile)(int, char *, char *, char *, char *, HWND);
 short (*fpStageLoadObject)(char *, struct OBJECT *, short, short, short);
 void  (*fpChooseColour)(unsigned char [], char *, HWND);
 short (*fpRequestNumEntry)(long, long, long, char *, char *);
 long  (*fpSelectNodeID)(HWND);
 void  (*fpAppendFileExtension)(char *, char *);
 void  (*fpCentreDlgOnC)(HWND);
 void  (*fpCentreDlgOnS)(HWND);
 void *dummyf[64];
} ANI_STRUCTURE;

typedef struct AWIREFRAME {
 long Np,Ne;
 Sedge *e;
 point *p;
} awire;

typedef struct PATHPOINT {
  struct PATHPOINT  *last;
  struct PATHPOINT  *next;
  short status;
  point p;
  double distance;
  double ptheta,pfi;
  double tension_p,tension_n;
} pathpoint;

typedef struct tagPATHEDITCONTROL {
  struct tagPATHEDITCONTROL  *last;
  struct tagPATHEDITCONTROL  *next;
  double distance;
  int frame;
  int x,y;
  double a,b,c,S,r;
} PATHEDITCONTROL;

typedef struct SSKY {
  char name[128];
  short type;
  unsigned char colour[3],zcolour[3];
  struct SSKY *last;
  struct SSKY *next;
  short firstframe,lastframe;
  double gamb,oamb,gamb_colour[3],oamb_colour[3];
  long  aff,alf,ast;
} sky;

typedef struct tagDIRECTOR {
  struct tagDIRECTOR *last;
  struct tagDIRECTOR *next;
  long  type;
  short firstframe,lastframe;
  struct NODE *ActiveCamera;
} director;


typedef struct SKELETON {
  long    id;               /* index of parent node in skeleton      */
  char    name[16];         /* name for skeleton element             */
  vector  u,v,w,            /* u,v,w are original axes               */
          uu,vv,ww;         /* transformed axes directions based pp  */
  double  R[4][4],          /* R is the rotational matrix (3x3)      */
          Q[4][4],          /* Q quaternion tweened R matrix         */
          T[4][4];          /* T is the transformation matrix        */
  point   p,                 /* p is original skeleton vertex        */
          pp,                /* transformed point                    */
          pt,                /* pp after Actor's transformation      */
          wk,                /* work point                           */
          bx[8];             /* bx is original bounding box          */
  long    IKlock,            /* this node is locked in IK chain      */
          uIKlock,           /* rotation round u is limited          */
          vIKlock,           /* rotation round v is limited          */
          wIKlock;           /* rotation round w is limited          */
  double  uIKmin,uIKmax,     /* limits to rotation (u) in IK chain   */
          vIKmin,vIKmax,     /* limits to rotation (v) in IK chain   */
          wIKmin,wIKmax;     /* limits to rotation (w) in IK chain   */
  LPARAM  lParam;
}skel;

typedef struct tagVECTOR4 {
 double x,y,z,w;
 BOOL   selected;
} vector4;

typedef struct tagNurbsProperties {
 char name[32];            /* name field             */
 long GranularityMax,      /* for visualising in the */
      GranularityMin;      /* view windows           */
 long     u_scale,
          v_scale;         /* texture coord scaling  */
 long     shader,          /* surface properties     */
          map,
          axis,
          param,
          smoothing,
          shiny,
          brilliance,
          reflectivity,
          transparency,
          glasstype,
          maptiled,
          mapshaded,
          mapdecal;
 long     pad[4];
 unsigned char colour[3],
          material_colour[3];
 char     hidden;
 char     padc[5];         /* struct 128 bytes       */
} NurbsProperties;

typedef struct tagNURBS {
 long id;
 long numU, numV;
 long orderU, orderV;
 double * kvU, * kvV;
 vector4 **points;
 NurbsProperties properties;
} nurbs;

typedef struct OBJECT {
  char name[255],effectname[255],*effect,*xip;
  struct OBJECT *last;
  struct OBJECT *next;
  short bWireframe;
  short firstframe,lastframe;
  short type;
  short morph,morphNo,quick_only;
  short fileID;
  short in_ram;
  awire w_frame;   /* a wire frame for use a bounding box */
  point *points;   /* array of vertices */
  Sedge *edges;    /* array of edges    */
  long  *skid;     /* array for each vertex to the skeleton node (if any) */
  long  npoints,nedges,nskeleton;
  point origin,offset;
  point outline[8];
  short groundtype;
  short lighttype; /* type of light */
  double light_intensity; /* light intensity [0 - 1] */
  short depth_cue;
  double depth_cue_c,depth_cue_l,depth_cue_q;
  short self_shadow,cast_shadow,show_shadow;
  unsigned char colour[3],acolour[3];
  skel *skeleton;  /* pointer to skeleton structure  */
  double *v;       /* pointer to velocity array */
  short pathtype,npathpoints;
  pathpoint *firstpathpoint;
  double pathlength;
  int npec;
  PATHEDITCONTROL *pec;
  double gspec,gtran,grefl,gbump;
  HMODULE hEffect;
  long  nNurbs;
  nurbs *Nurbs;
  long  bIKchain;
} object;

typedef struct POSITION {
  struct POSITION *last;
  struct POSITION *next;
  short firstframe,lastframe;
  short type;
  short fx,fy,fz;   // this is for new position
  point finish;
  struct NODE  *onpath;
  char pnodename[128];
  double tension_s,tension_e,easin,easout;
} position;

typedef struct ALIGN {
  struct ALIGN *last;
  struct ALIGN *next;
  short firstframe,lastframe;
  short type,im;
  double theta,phi,alpha,ima;
  struct NODE *topath;
  char anodename[128];
  double tension_s,tension_e,easin,easout;
} align;

typedef struct SIZE {
  double Sx,Sy,Sz;
  struct NODE *base;
  struct SIZE *last;
  struct SIZE *next;
  short firstframe,lastframe;
  short tweentype;
  double tension_s,tension_e,easin,easout;
} size;

typedef struct NODE {
  struct NODE        *last;
  struct NODE        *next;
  struct SSKY        *fsky;
  struct tagDIRECTOR *fdirector;
  struct OBJECT      *fobj;
  struct POSITION    *fpos;
  struct ALIGN       *fali;
  struct SIZE        *fsiz;
  short              type,ID;
  char               actorname[128];
} node;

typedef struct tagXIP {
  struct tagXIP *last;
  struct tagXIP *next;
  char *p;
} XIP;

typedef struct tagTOOLFLAGS {
  BOOL flagX,flagY,flagZ,flagDummy;
  int  flagA,flagT,flagR,flagP,flagK;
} TOOLFLAGS;

typedef struct tagPREFERENCES {
  BOOL show_select;
  BOOL buttons;
  BOOL quick_trick;
  BOOL default_start;
  BOOL use_3d_dialogs;
  BOOL use_OpenGL;
  BOOL undo;
  BOOL use_positions;
  BOOL ground;
  BOOL big_camera;
  BOOL tooltips;
  BOOL toolsmall;
  BOOL toolleft;
  BOOL autocentre;
  BOOL detail_auto;
  long detail_step;
  BOOL background;
  BOOL auto_update_keyframer;
  BOOL use_gl_robot;
  long splash;
} PREFERENCES;

#define PLANE            0              /* brush map types */
#define CYLINDER         1
#define PLANE_MOZIAC     2
#define CYLINDER_MOZIAC  3
#define MAP_BY_VERTEX    4
#define NOMORPH          0
#define MOPPH            1
#define QMORPH           2
#define OPAQ             0              /* brush color 0   */
#define CLEAR            1
#define GLOSS            0              /* brush shiny     */
#define MATT             1
#define TRIVIEW          0
#define ONEVIEW          1
#define TRITOP           0
#define TRIFRONT         1
#define TRIRIGHT         2

/* Macro Function Definitions */
#define POINT2VECTOR(p,v) v[0]=(double)p[0]; v[1]=(double)p[1]; \
                          v[2]=(double)p[2];
#define VECTOR2POINT(v,p) p[0]=(long)v[0]; p[1]=(long)v[1]; \
                          p[2]=(long)v[2];

/* definition of unit dimensions */
#define UNIT         32768L
#define UNIT2        65536L
#define MINUNIT       2048L   /* was 1024 until 21/12/92 */
//                2147483648L (2^31)
#define MAXUNIT   1073741824L
#ifndef PI
#define PI 3.1415926
#endif
#define PI2            PI/2
/* definition of flags and identifiers*/
#define NORMAL      0
#define PATH        1
#define TARGET      2
#define GROUND      3
#define LIGHT       4
#define ANIMOBJ     5
#define SKY         6
#define CAMERA      7
#define CSG         8
#define ROBOT       9
#define IMAGEP     10
#define DIRECTOR   11
#define OPEN        0
#define CLOSED      1
#define LOADED      2
#define COPIED      3
#define TWEEN       0
#define FOLLOW      1
#define SPLINE      2
#define FOLLOWOFF   3
#define FOLLOWON    4
#define FOLLOWAT    5
#define TOPATH      1
#define TRACK       2
#define COPY        3
#define SPLINEA     4
#define PLAIN       0
#define TEXTU       1
#define BRUM        2
#define TBRUM       3
#define MBRUM       4
#define PRESET      6
#define SPHERE      0
#define TUBE        1
#define CYLIND      2
#define DISK        3
#define NOSHADOW    1
#define SHADOWONLY  2
#define SPOTLIGHT   3
#define DUMMYL      4
#define SELECTED    1
#define DESELECTED  0
#define ERASED      2
#define EDITING     3
#define SKYCOLOURED 0
#define SKYGRADED   1
#define SKYMAPPED   2
#define BACKDROP    3
#define SKYANIMATED 4
/* general defines  */
#define NO               0
#define YES              1
#define FAIL            -1
#define OK               1
#define MAX_DIR        255
#define MAX_FILE       255
#define ZOOMOUT          0
#define ZOOMIN           1
#define PANUP            2
#define PANDOWN          3
#define PANLEFT          4
#define PANRIGHT         5
#define PANFORWARD       6
#define PANBACK          7
#define RECENTRE         8
#define CENTREWORLD      9
/* tools and actions */
#define NTOOLS           8
#define NOTOOL           0
#define NODETOOL         1
#define PAN              2
#define INZOOM           3
#define GRABBER          4
#define ROTATOR          5
#define SIZER            6
#define TRACKER          7

#define PATHEDITOR       8
#define SKELETOR         9

#define MAXTOOLS        10

#define SHIFTER          1
#define ANGLER           2
