/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file CAMERA.C  */

// This file contains the code to render the camera perspective
// view into the camera perspective window. It includes the
// code to render the animation preview.

#define MODULE ANI_PRSP 1

#include "animate.h"

#define GROUND_GRID_SIZE 10

static void DrawStagePerspective(HDC, HWND, int , point ,BOOL);
static void DrawQuickWire(HDC hdcView,
                          object *Op, double mr, double t1[4][4],
                          short pwincx, short pwincy);
static void DrawNurbsPersp(HDC hdcView,
                          object *Op, double mr, double t1[4][4],
                          short pwincx, short pwincy);
static BOOL CalculateEffectedVertices(double mr,
                                      double tall[4][4],
                                      double tc[4][4], double tp[4][4],
                                      BOOL  ground,
                                      point p, point gpos, point cpos,
                                      object *op, Svertex *Vp);
static void ClipToPlane(double *x1, double *y1, double *z1,
                        double x2, double y2, double z2, long Cp);
static BOOL ClipToWindow(long *x1, long *y1,
                         long *x2, long *y2,
                         long xmin, long ymin,
                         long xmax, long ymax);

static double trr[4][4];
static double scalex,scaley,dt;
static double CameraAlpha,CameraTheta,CameraHeight;
static short ep1[12]={0,1,2,3,4,5,6,7,0,1,2,3};
static short ep2[12]={1,2,3,0,5,6,7,4,4,5,6,7};

static void DrawStagePerspective(HDC hdcView, HWND hWnd,
                                 int DisplayType, point cP,
                                 BOOL check_abort){
 double  x1,y1,z1,x2,y2,z2,alpha,theta,phi,sx,sy,sz,ima,y_offset
         ,t1[4][4],t2[4][4],t3[4][4],t4[4][4],mr,rxyz[3];
 short   flag,im,pwinl,pwinr,pwint,pwinb,pwincx,pwincy;
 int     ix1,iy1,ix2,iy2;
 long    lx1,ly1,lx2,ly2;
 long    i,j,k,Nv,Nl,Nm,xyz[3],id[2],*sid,*sidl,qstep;
 node    *Np;
 object  *Op;
 Svertex *Vp;
 point   *Vm;
 point   Offset,TrackedOffset,grP;
 skel    *sp,*spl;
 RECT    rc;
 HPEN hGpen;
 BOOL gg=FALSE;
 mr=1.0;
 GetClientRect(hWnd,&rc);
 pwinl=pwint=0; pwinr=rc.right; pwinb=rc.bottom;
 pwincx=rc.right/2; pwincy=rc.bottom/2;
 if(hdcView == NULL)goto XXIT;
// if(bAbortedDrawing){
//   DisplayType=0;
//   KillTimer(ghwnd_main,1);
//   SetTimer(ghwnd_main,1,30,NULL);
// }
 //TRYAGAIN:
 if((Np=FirstNp) != NULL)while(Np != NULL){ /* ground and lights */
   if(Np->type == GROUND && ((Op=Np->fobj) != NULL))while(Op != NULL){
     if(CurrentFrame >= Op->firstframe && CurrentFrame <= Op->lastframe){
       y_offset = (double)(scaley*(dt*tan(-CameraAlpha)));
       ix1 = (int)( 1000.0*cos(-CameraTheta)
                             - y_offset*sin(-CameraTheta) + pwincx);
       iy1 = (int)( 1000.0*sin( CameraTheta)
                             - y_offset*cos( CameraTheta) + pwincy);
       ix2 = (int)(-1000.0*cos(-CameraTheta)
                             - y_offset*sin(-CameraTheta) + pwincx);
       iy2 = (int)(-1000.0*sin( CameraTheta)
                             - y_offset*cos( CameraTheta) + pwincy);
       hGpen=SelectObject(hdcView,GetStockObject(BLACK_PEN));
       MoveToEx(hdcView,ix1,iy1,NULL);
       LineTo(hdcView,ix2,iy2);
       GetTransform(0,CurrentFrame,1.0,
                    Np,Op,Offset,TrackedOffset,&phi,&theta,&alpha,
                    &sx,&sy,&sz,&im,&ima);
       gg=TRUE;
       VECCOPY(Offset,grP)
       if(!Preferences.ground){ /* ground grid as well as horizon */
         long xx,yy,dx,dy,dxx,dyy,grid_size,Cp=UNIT;
         theta *= PIo180;
         phi   *= PIo180;
         alpha *= PIo180;
         scal(t3,sx,sy,sz);
         if(im > 0){
           if     (im == 1)rotz(t1,ima*PIo180);
           else if(im == 2)rotx(t1,ima*PIo180);
           else if(im == 3)roty(t1,ima*PIo180);
           roty(t4,theta);
           m4by4(t4,t1,t2);
         }
         else roty(t2,theta);
         m4by4(t2,t3,t1);
         rotx(t2,alpha);
         m4by4(t2,t1,t3);
         rotz(t1,phi);
         m4by4(t1,t3,t2);
         tram(t1,(double)Offset[0],(double)Offset[1],(double)Offset[2]);
         m4by4(t1,t2,t3);
         m4by4(trr,t3,t1);
         grid_size=UNIT*4;
         for(i=0;i<2;i++){
           yy = (grid_size/2)*GROUND_GRID_SIZE; xx = -yy;
           if(i == 0){
             dx=0; dy = -grid_size*GROUND_GRID_SIZE;
             dxx=grid_size;    dyy=0;
           }
           else{
             dx=grid_size*GROUND_GRID_SIZE; dy = 0;
             dxx=0; dyy = -grid_size;
           }
           for(j=0;j<GROUND_GRID_SIZE+1;j++){
             flag=0;
             m4by1(t1,(double)xx,(double)yy,0.0,&x1,&y1,&z1);
             m4by1(t1,(double)(xx+dx),(double)(yy+dy),0.0,&x2,&y2,&z2);
             if(y1 > Cp || y2 > Cp){
               if(y1 < Cp)ClipToPlane(&x1,&y1,&z1,x2,y2,z2,Cp);
               lx1=(long)((double)pwincx+scalex*x1/y1);
               ly1=(long)((double)pwincy-scaley*z1/y1);
               if(y2 < Cp)ClipToPlane(&x2,&y2,&z2,x1,y1,z1,Cp);
               lx2=(long)((double)pwincx+scalex*x2/y2);
               ly2=(long)((double)pwincy-scaley*z2/y2);
               if(ClipToWindow(&lx1,&ly1,&lx2,&ly2,
                               pwinl,pwint,pwinr,pwinb)){
                 MoveToEx(hdcView,lx1,ly1,NULL);
                 LineTo(hdcView,lx2,ly2);
               }
             }
             xx=xx+dxx; yy=yy+dyy;
           }
         }
       }
       SelectObject(hdcView,hGpen);
       break;
     }
     Op=Op->next;
   }
   else if(Np->type == LIGHT && ((Op=Np->fobj) != NULL))while(Op != NULL){
     if(CurrentFrame >= Op->firstframe && CurrentFrame <= Op->lastframe){
       GetTransform(0,CurrentFrame,1.0,
                    Np,Op,Offset,TrackedOffset,&phi,&theta,&alpha,
                    &sx,&sy,&sz,&im,&ima);
       tram(t1,(double)Offset[0],(double)Offset[1],(double)Offset[2]);
       m4by4(trr,t1,t2);
       m4by1(t2,0.0,0.0,0.0,&x1,&y1,&z1);
       if(y1 > dt){
         ix1=(int)(pwincx+scalex*x1*dt/(y1));
         iy1=(int)(pwincy-scaley*z1*dt/(y1));
         if(ix1 < 0 || ix1 >= pwinr || iy1 < 0 || iy1 >= pwinb)break;
         hGpen=SelectObject(hdcView,GetStockObject(BLACK_PEN));
         MoveToEx(hdcView,ix1-2,iy1,NULL); LineTo(hdcView,ix1+3,iy1);
         MoveToEx(hdcView,ix1,iy1-2,NULL); LineTo(hdcView,ix1,iy1+3);
         SelectObject(hdcView,hGpen);
       }
       break;
     }
     Op=Op->next;
   }
   Np=Np->next;
 }

 if((Np=FirstNp) != NULL)while(Np != NULL){
   if(   (Np->type == NORMAL || Np->type == ANIMOBJ || Np->type == CSG)
      && ((Op=Np->fobj) != NULL))while(Op != NULL){
     if(CurrentFrame >= Op->firstframe && CurrentFrame <= Op->lastframe){
       GetTransform(0,CurrentFrame,1.0,
                    Np,Op,Offset,TrackedOffset,&phi,&theta,&alpha,
                    &sx,&sy,&sz,&im,&ima);
       theta *= PIo180;
       phi   *= PIo180;
       alpha *= PIo180;
       scal(t3,sx,sy,sz);
       if(im > 0){
         if     (im == 1)rotz(t1,ima*PIo180);
         else if(im == 2)rotx(t1,ima*PIo180);
         else if(im == 3)roty(t1,ima*PIo180);
         roty(t4,theta);
         m4by4(t4,t1,t2);
       }
       else roty(t2,theta);
       m4by4(t2,t3,t1);
       rotx(t2,alpha);
       m4by4(t2,t1,t3);
       rotz(t1,phi);
       m4by4(t1,t3,t2);
       tram(t1,(double)Offset[0],(double)Offset[1],(double)Offset[2]);
       m4by4(t1,t2,t3);  /* does't take object origin into account */
       m4by4(trr,t3,t1);
       if(DisplayType == 0 || Op->quick_only == 1 || Op->type == CSG){
        if(Op->morph != NO && Op->last != NULL){
          mr=(double)(Op->lastframe - Op->firstframe+1);
          mr=(double)(CurrentFrame - Op->firstframe+1)/mr;
        }
        if(Op->w_frame.p != NULL){
          DrawQuickWire(hdcView,Op,mr,t1,pwincx,pwincy);
          goto CONTINUE;
        }
        for(i=0;i<12;i++){
         flag=0;
        if(Op->morph != NO && Op->last != NULL){
           m4by1(t1,(double)(Op->outline[ep1[i]][0] -
                       Op->last->outline[ep1[i]][0])*mr +
                       (double)Op->last->outline[ep1[i]][0],
                    (double)(Op->outline[ep1[i]][1] -
                       Op->last->outline[ep1[i]][1])*mr +
                       (double)Op->last->outline[ep1[i]][1],
                    (double)(Op->outline[ep1[i]][2] -
                       Op->last->outline[ep1[i]][2])*mr +
                       (double)Op->last->outline[ep1[i]][2],
                    &x1,&y1,&z1);
           m4by1(t1,(double)(Op->outline[ep2[i]][0] -
                       Op->last->outline[ep2[i]][0])*mr +
                       (double)Op->last->outline[ep2[i]][0],
                    (double)(Op->outline[ep2[i]][1] -
                       Op->last->outline[ep2[i]][1])*mr +
                       (double)Op->last->outline[ep2[i]][1],
                    (double)(Op->outline[ep2[i]][2] -
                       Op->last->outline[ep2[i]][2])*mr +
                       (double)Op->last->outline[ep2[i]][2],
                    &x2,&y2,&z2);
         }
         else{
           m4by1(t1 ,(double)Op->outline[ep1[i]][0]
                    ,(double)Op->outline[ep1[i]][1]
                    ,(double)Op->outline[ep1[i]][2]
                    ,&x1,&y1,&z1);
           m4by1(t1 ,(double)Op->outline[ep2[i]][0]
                    ,(double)Op->outline[ep2[i]][1]
                    ,(double)Op->outline[ep2[i]][2]
                    ,&x2,&y2,&z2);
         }
         if(y1 > dt){ /* point infront of viewing plane */
           ix1=(int)(pwincx+scalex*x1*dt/(y1));
           iy1=(int)(pwincy-scaley*z1*dt/(y1));
           if(ix1 < -16000 || ix1 > 16000)flag=1;
           if(iy1 < -16000 || iy1 > 16000)flag=1;
         }
         else  flag=1;
         if(y2 > dt){
           ix2=(int)(pwincx+scalex*x2*dt/(y2));
           iy2=(int)(pwincy-scaley*z2*dt/(y2));
           if(ix2 < -16000 || ix2 > 16000)flag=1;
           if(iy2 < -16000 || iy2 > 16000)flag=1;
         }
         else flag=1;
         if(flag == 0){
           MoveToEx(hdcView,ix1,iy1,NULL);
           LineTo(hdcView,ix2,iy2);
         }
        }
        CONTINUE:;
       }
       else{ /* draw the object */
        if(Op->morph != NO && Op->last != NULL && Op->last->in_ram){
          mr=(double)(Op->lastframe - Op->firstframe+1);
          if(Op->lastframe == Nframes)mr += 1.0;
          mr=(double)(CurrentFrame - Op->firstframe+1)/mr;
          Nm=Op->last->npoints;
        }
        else Nm=0;
        if(Op->in_ram && Op->points != NULL){
         Nv=Op->npoints;
         if(Nv > 0){
          Vp=(Svertex *)X__Malloc(Nv*sizeof(Svertex));
          if(Vp != NULL){
           if(Nv != Nm){
             if(!CalculateEffectedVertices(-1.0,t1,trr,t3,
                                            gg,Offset,grP,cP,Op,Vp))
             for(i=0;i<Nv;i++){
               m4by1(t1,(double)(Op->points[i][0] - Op->origin[0]),
                        (double)(Op->points[i][1] - Op->origin[1]),
                        (double)(Op->points[i][2] - Op->origin[2]),
                        &x1,&y1,&z1);
               Vp[i][0]=(float)x1; Vp[i][1]=(float)y1; Vp[i][2]=(float)z1;
             }
           }
           else{
             long l,m;
             if(!CalculateEffectedVertices(mr,t1,trr,t3,
                                           gg,Offset,grP,cP,Op,Vp))
             for(i=0;i<Nv;i++){
               for(j=0;j<3;j++){
                 l = Op->points[i][j] - Op->origin[j];
                 m = Op->last->points[i][j] - Op->last->origin[j];
                 rxyz[j] = (double)(l-m)*mr + (double)m;
               }
               m4by1(t1,rxyz[0],rxyz[1],rxyz[2]
                       ,&x1,&y1,&z1);
               Vp[i][0]=(float)x1; Vp[i][1]=(float)y1; Vp[i][2]=(float)z1;
             }
           }
           Nl = Op->nedges;
           if(global_quickdraw){
             if(Nl < 100)qstep=1;
             else if(Nl <   1000)qstep=2;
             else if(Nl <   2500)qstep=3;
             else if(Nl <   5000)qstep=4;
             else if(Nl <  10000)qstep=6;
             else if(Nl <  20000)qstep=8;
             else if(Nl <  50000)qstep=12;
             else if(Nl < 100000)qstep=25;
             else qstep=40;
           }
           else qstep=1;
           if(Nl > 0){
#ifdef _LINE_DRAW
LPPOINT lppt;
LPBYTE  lpbTypes;
int     cCount;
long    pcount;
 if((lppt=(LPPOINT)LocalAlloc(LMEM_FIXED,2*Nl*sizeof(POINT))) == NULL)goto XXIT;
 if((lpbTypes=(LPBYTE)LocalAlloc(LMEM_FIXED,2*Nl*sizeof(BYTE))) == NULL){
  LocalFree((HLOCAL)lppt);
  goto XXIT;
 }
 cCount=0; pcount=0;
#endif
             for(i=0;i<Nl;i += qstep){
               id[0] = Op->edges[i][0];
               id[1] = Op->edges[i][1];
               flag=0;
               x1=Vp[id[0]][0];  x2=Vp[id[1]][0];
               y1=Vp[id[0]][1];  y2=Vp[id[1]][1];
               z1=Vp[id[0]][2];  z2=Vp[id[1]][2];
               if(y1 > dt){ /* point infront of viewing plane */
                 ix1=(int)(pwincx+scalex*x1*dt/(y1));
                 iy1=(int)(pwincy-scaley*z1*dt/(y1));
                 if(ix1 < -16000 || ix1 > 16000)flag=1;
                 if(iy1 < -16000 || iy1 > 16000)flag=1;
               }
               else  flag=1;
               if(y2 > dt){
                 ix2=(int)(pwincx+scalex*x2*dt/(y2));
                 iy2=(int)(pwincy-scaley*z2*dt/(y2));
                 if(ix2 < -16000 || ix2 > 16000)flag=1;
                 if(iy2 < -16000 || iy2 > 16000)flag=1;
               }
               else flag=1;
               if(flag == 0){
                 if( !Preferences.quick_trick || abs(ix1-ix2) > 2 || abs(iy1-iy2) > 2 ){
#ifdef _LINE_DRAW
    lppt[cCount].x=ix1; lppt[cCount].y=iy1;
    lpbTypes[cCount]=PT_MOVETO; cCount++;
    lppt[cCount].x=ix2; lppt[cCount].y=iy2;
    lpbTypes[cCount]=PT_LINETO; cCount++;
#else
                   MoveToEx(hdcView,ix1,iy1,NULL);
                   LineTo(hdcView,ix2,iy2);
                   if(bAbortedDrawing){
                     X__Free(Vp);
                     bAbortedDrawing=FALSE;
                     goto XXIT;
                   }
#if 0
                   if(check_abort && HIWORD(GetQueueStatus(QS_KEY|QS_MOUSEBUTTON))){
                     X__Free(Vp);
                     bAbortedDrawing=TRUE;
                     DisplayType=0;
                     SetTimer(ghwnd_main,1,50,NULL);
                     goto XXIT;
                     //goto TRYAGAIN;
                   }
#endif
#endif
                 }
               }
             }
#ifdef _LINE_DRAW
 if(PolyDraw(hdcView,lppt,lpbTypes,cCount) == FALSE){
   char message[128];
   sprintf(message,"ERROR: pcount=%ld ccount=%ld",pcount,cCount);
   MessageBox (NULL,message, NULL,
   MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
 }
 LocalFree((HLOCAL)lppt);
 LocalFree((HLOCAL)lpbTypes);
#endif
           }
           X__Free(Vp);
          }
          else{
           //  Don't do this because of multithreading
           //  SendPrgmQuery(IDQ_NOMEM3,0);
          }
         }
        }
        else{
          //  Don't do this because of multithreading
          //  SendPrgmQuery(IDQ_NOFINDMODEL,0);
        }
        // Draw Any NURBS
        DrawNurbsPersp(hdcView,Op,mr,t1,pwincx,pwincy);
       }
     }
     Op=Op->next;
   }
   else if(Np->type == ROBOT && ((Op=Np->fobj) != NULL))while(Op != NULL){
     if(CurrentFrame >= Op->firstframe && CurrentFrame <= Op->lastframe){
       if(Op->nskeleton == 0 || (sp=Op->skeleton) == NULL)break;
       GetTransform(0,CurrentFrame,1.0,
                    Np,Op,Offset,TrackedOffset,&phi,&theta,&alpha,
                    &sx,&sy,&sz,&im,&ima);
       RobotTransform(Offset,phi,theta,alpha,sx,sy,sz,im,ima,t1);
       m4by4(trr,t1,t3);
       mr=(double)(Op->lastframe - Op->firstframe+1);
       mr=(double)(CurrentFrame - Op->firstframe+1)/mr;
       InterpolateRobot(mr,Op,Op->last);
       if(Op->last != NULL)spl=Op->last->skeleton;
       else                spl=NULL;
       if(DisplayType == 0){
         for(k=0;k<Op->nskeleton;k++){
           flag=0;
           m4by4(t3,(Op->skeleton+k)->T,t4);
           for(i=0;i<12;i++){
             m4by1(t4 ,(double)(Op->skeleton+k)->bx[ep1[i]][0]
                      ,(double)(Op->skeleton+k)->bx[ep1[i]][1]
                      ,(double)(Op->skeleton+k)->bx[ep1[i]][2]
                      ,&x1,&y1,&z1);
             m4by1(t4 ,(double)(Op->skeleton+k)->bx[ep2[i]][0]
                      ,(double)(Op->skeleton+k)->bx[ep2[i]][1]
                      ,(double)(Op->skeleton+k)->bx[ep2[i]][2]
                      ,&x2,&y2,&z2);
             if(y1 > dt){ /* point infront of viewing plane */
               ix1=(int)(pwincx+scalex*x1*dt/(y1));
               iy1=(int)(pwincy-scaley*z1*dt/(y1));
               if(ix1 < -16000 || ix1 > 16000)flag=1;
               if(iy1 < -16000 || iy1 > 16000)flag=1;
             }
             else  flag=1;
             if(y2 > dt){
               ix2=(int)(pwincx+scalex*x2*dt/(y2));
               iy2=(int)(pwincy-scaley*z2*dt/(y2));
               if(ix2 < -16000 || ix2 > 16000)flag=1;
               if(iy2 < -16000 || iy2 > 16000)flag=1;
             }
             else flag=1;
             if(flag == 0){
               MoveToEx(hdcView,ix1,iy1,NULL);
               LineTo(hdcView,ix2,iy2);
             }
           }
         }
#if 0
         for(k=0;k<Op->nskeleton;k++){
           x1=(double)sp[k].pp[0];
           y1=(double)sp[k].pp[1];
           z1=(double)sp[k].pp[2];
           m4by1(t3,x1,y1,z1,&x2,&y2,&z2);
           sp[k].wk[0]=(long)x2;
           sp[k].wk[1]=(long)y2;
           sp[k].wk[2]=(long)z2;
         }
         for(k=1;k<Op->nskeleton;k++){
           flag=0;
           x1=(double)sp[k].wk[0];
           y1=(double)sp[k].wk[1];
           z1=(double)sp[k].wk[2];
           if((j=sp[k].id) < Op->nskeleton){
             x2=(double)sp[j].wk[0];
             y2=(double)sp[j].wk[1];
             z2=(double)sp[j].wk[2];
           }
           else{ x2=x1; y2=y1; z2=z1;}
           if(y1 > dt){ /* point infront of viewing plane */
             ix1=(int)(pwincx+scalex*x1*dt/(y1));
             iy1=(int)(pwincy-scaley*z1*dt/(y1));
             if(ix1 < -16000 || ix1 > 16000)flag=1;
             if(iy1 < -16000 || iy1 > 16000)flag=1;
           }
           else  flag=1;
           if(y2 > dt){
             ix2=(int)(pwincx+scalex*x2*dt/(y2));
             iy2=(int)(pwincy-scaley*z2*dt/(y2));
             if(ix2 < -16000 || ix2 > 16000)flag=1;
             if(iy2 < -16000 || iy2 > 16000)flag=1;
           }
           else flag=1;
           if(flag == 0){
             MoveToEx(hdcView,ix1,iy1,NULL);
             LineTo(hdcView,ix2,iy2);
           }
         }
#endif
       }
       else{ /* full robot */
         Vm=NULL;  Vp=NULL; sid=Op->skid;
         if(Op->morph != NO && Op->last != NULL && (sidl=Op->last->skid) != NULL){
           if(Op->last->in_ram){
             Nm=Op->last->npoints;
             if(Nm > 0){
               if((Vm=(point *)X__Malloc(Nm*sizeof(point))) != NULL){
                 for(i=0;i<Nm;i++){
                   CopyPoint(Op->last->points[i],xyz);
                   Vm[i][0] = xyz[0] - Op->last->origin[0];
                   Vm[i][1] = xyz[1] - Op->last->origin[1];
                   Vm[i][2] = xyz[2] - Op->last->origin[2];
                 }
               }
               else{
                 SendPrgmQuery(IDQ_NOMEM3,0);
                 Nm=0;
               }
             }
           }
         }
         else Nm=0;
         if(Op->in_ram){
           Nv=Op->npoints;
           if(Nv > 0){
             if((Vp=(Svertex *)X__Malloc(Nv*sizeof(Svertex))) != NULL){
               for(i=0;i<Nv;i++){
                 CopyPoint(Op->points[i],xyz);
                 m4by4(t3,sp[sid[i]].T,t4);
                 x1 = (double)(xyz[0]-Op->origin[0]);
                 y1 = (double)(xyz[1]-Op->origin[1]);
                 z1 = (double)(xyz[2]-Op->origin[2]);
                 if(Nv == Nm && Vm != NULL){ /* if object is morphing */
                   x1 = (x1 - (double)Vm[i][0])*mr + (double)Vm[i][0];
                   y1 = (y1 - (double)Vm[i][1])*mr + (double)Vm[i][1];
                   z1 = (z1 - (double)Vm[i][2])*mr + (double)Vm[i][2];
                 }
                 m4by1(t4,x1,y1,z1,&x2,&y2,&z2);
                 Vp[i][0]=(float)x2; Vp[i][1]=(float)y2; Vp[i][2]=(float)z2;
               }
             }
             else{
               SendPrgmQuery(IDQ_NOMEM3,0);
               Nv=0;
             }
           }
         }
         else Nv=0;
         Nl = Op->nedges;
         if(global_quickdraw){
           if(Nl < 100)qstep=1;
           else if(Nl <   1000)qstep=2;
           else if(Nl <   2500)qstep=3;
           else if(Nl <   5000)qstep=4;
           else if(Nl <  10000)qstep=6;
           else if(Nl <  20000)qstep=8;
           else if(Nl <  50000)qstep=12;
           else if(Nl < 100000)qstep=25;
           else qstep=40;
         }
         else qstep=1;
         if(Nl > 0 && Nv > 0 && Vp != NULL && sid != NULL)
         for(i=0;i<Nl;i+=qstep){
           id[0] = Op->edges[i][0];
           id[1] = Op->edges[i][1];
           flag=0;
           x1=Vp[id[0]][0];  x2=Vp[id[1]][0];
           y1=Vp[id[0]][1];  y2=Vp[id[1]][1];
           z1=Vp[id[0]][2];  z2=Vp[id[1]][2];
           if(y1 > dt){ /* point infront of viewing plane */
             ix1=(int)(pwincx+scalex*x1*dt/(y1));
             iy1=(int)(pwincy-scaley*z1*dt/(y1));
             if(ix1 < -16000 || ix1 > 16000)flag=1;
             if(iy1 < -16000 || iy1 > 16000)flag=1;
           }
           else  flag=1;
           if(y2 > dt){
             ix2=(int)(pwincx+scalex*x2*dt/(y2));
             iy2=(int)(pwincy-scaley*z2*dt/(y2));
             if(ix2 < -16000 || ix2 > 16000)flag=1;
             if(iy2 < -16000 || iy2 > 16000)flag=1;
           }
           else flag=1;
           if(flag == 0){
             MoveToEx(hdcView,ix1,iy1,NULL);
             LineTo(hdcView,ix2,iy2);
           }
           if(bAbortedDrawing){
             if(Vp != NULL){ X__Free(Vp); Vp=NULL;}
             if(Vm != NULL){ X__Free(Vm); Vm=NULL;}
             bAbortedDrawing=FALSE;
             goto XXIT;
           }
         }
         if(Vp != NULL){ X__Free(Vp); Vp=NULL;}
         if(Vm != NULL){ X__Free(Vm); Vm=NULL;}
       }
       break;
     }
     Op=Op->next;
   }
   Np=Np->next;
 }
 XXIT:
 return;
}

static void DrawQuickWire(HDC hdcView,
                          object *Op, double mr, double t1[4][4],
                          short pwincx, short pwincy){
 long i,j,flag,morphflag=0;
 double x1,y1,z1,x2,y2,z2;
 int ix1,iy1,ix2,iy2;
 point pl,pc;
 if(Op->w_frame.e != NULL && Op->w_frame.Np > 0){
   if(Op->morph != NO && Op->last != NULL && Op->last->w_frame.p != NULL &&
      Op->last->w_frame.e != NULL  && Op->last->w_frame.Np == Op->w_frame.Np)
      morphflag=1;
   for(i=0;i<Op->w_frame.Ne;i++){
     flag=0;
     if(morphflag){
       pl[0]=Op->last->w_frame.p[Op->last->w_frame.e[i][0]][0];
       pl[1]=Op->last->w_frame.p[Op->last->w_frame.e[i][0]][1];
       pl[2]=Op->last->w_frame.p[Op->last->w_frame.e[i][0]][2];
       SubPoints(pl,Op->last->origin,pl);
       AddPoints(pl,Op->origin,pl);
       pc[0]=Op->w_frame.p[Op->w_frame.e[i][0]][0];
       pc[1]=Op->w_frame.p[Op->w_frame.e[i][0]][1];
       pc[2]=Op->w_frame.p[Op->w_frame.e[i][0]][2];
       m4by1(t1,(double)(pc[0]-pl[0])*mr+(double)pl[0],
                (double)(pc[1]-pl[1])*mr+(double)pl[1],
                (double)(pc[2]-pl[2])*mr+(double)pl[2],
          &x1,&y1,&z1);
       pl[0]=Op->last->w_frame.p[Op->last->w_frame.e[i][1]][0];
       pl[1]=Op->last->w_frame.p[Op->last->w_frame.e[i][1]][1];
       pl[2]=Op->last->w_frame.p[Op->last->w_frame.e[i][1]][2];
       pc[0]=Op->w_frame.p[Op->w_frame.e[i][1]][0];
       pc[1]=Op->w_frame.p[Op->w_frame.e[i][1]][1];
       pc[2]=Op->w_frame.p[Op->w_frame.e[i][1]][2];
       m4by1(t1,(double)(pc[0]-pl[0])*mr+(double)pl[0],
                (double)(pc[1]-pl[1])*mr+(double)pl[1],
                (double)(pc[2]-pl[2])*mr+(double)pl[2],
          &x2,&y2,&z2);
     }
     else{
       m4by1(t1 ,(double)Op->w_frame.p[Op->w_frame.e[i][0]][0]
                ,(double)Op->w_frame.p[Op->w_frame.e[i][0]][1]
                ,(double)Op->w_frame.p[Op->w_frame.e[i][0]][2]
                ,&x1,&y1,&z1);
       m4by1(t1 ,(double)Op->w_frame.p[Op->w_frame.e[i][1]][0]
                ,(double)Op->w_frame.p[Op->w_frame.e[i][1]][1]
                ,(double)Op->w_frame.p[Op->w_frame.e[i][1]][2]
                ,&x2,&y2,&z2);
     }
     if(y1 > dt){ /* point infront of viewing plane */
       ix1=(int)(pwincx+scalex*x1*dt/(y1));
       iy1=(int)(pwincy-scaley*z1*dt/(y1));
       if(ix1 < -16000 || ix1 > 16000)flag=1;
       if(iy1 < -16000 || iy1 > 16000)flag=1;
     }
     else  flag=1;
     if(y2 > dt){
       ix2=(int)(pwincx+scalex*x2*dt/(y2));
       iy2=(int)(pwincy-scaley*z2*dt/(y2));
       if(ix2 < -16000 || ix2 > 16000)flag=1;
       if(iy2 < -16000 || iy2 > 16000)flag=1;
     }
     else flag=1;
     if(flag == 0){
        MoveToEx(hdcView,ix1,iy1,NULL);
        LineTo(hdcView,ix2,iy2);
     }
   }
 }
}

static void DrawNurbsPersp(HDC hdcView,
                          object *Op, double mr, double t1[4][4],
                          short pwincx, short pwincy){
 vector4 *p;
 nurbs *n;
 long k,ii,jj,flag,morphflag=0;
 double w,x1,y1,z1,x2,y2,z2;
 int ix1,iy1,ix2,iy2;
 point pl,pc;
 if(Op->nNurbs == 0 || Op->Nurbs == NULL)return;
 for(k=0;k<Op->nNurbs;k++){
   n=(Op->Nurbs+k);
   if(n->properties.hidden)continue;
   if(Op->morph != NO && Op->last != NULL && Op->last->Nurbs != NULL &&
      Op->nNurbs == Op->last->nNurbs)morphflag=1;
   // first pass
   for(ii=0;ii<n->numV;ii++)
   for(jj=0;jj<n->numU-1;jj++){
     flag=0;
     if(morphflag){
       p = &((Op->last->Nurbs+k)->points[ii][jj]);
       w=1.0/p->w; pl[0]=((p->x)*w); pl[1]=((p->y)*w); pl[2]=((p->z)*w);
       SubPoints(pl,Op->last->origin,pl);
       AddPoints(pl,Op->origin,pl);
       p = &(n->points[ii][jj]);
       w=1.0/p->w; pc[0]=((p->x)*w); pc[1]=((p->y)*w); pc[2]=((p->z)*w);
       m4by1(t1,(double)(pc[0]-pl[0])*mr+(double)pl[0],
                (double)(pc[0]-pl[1])*mr+(double)pl[1],
                (double)(pc[0]-pl[1])*mr+(double)pl[2],
          &x1,&y1,&z1);
       p = &((Op->last->Nurbs+k)->points[ii][jj+1]);
       w=1.0/p->w; pl[0]=((p->x)*w); pl[1]=((p->y)*w); pl[2]=((p->z)*w);
       p = &(n->points[ii][jj+1]);
       w=1.0/p->w; pc[0]=((p->x)*w); pc[1]=((p->y)*w); pc[2]=((p->z)*w);
       m4by1(t1,(double)(pc[0]-pl[0])*mr+(double)pl[0],
                (double)(pc[0]-pl[1])*mr+(double)pl[1],
                (double)(pc[0]-pl[1])*mr+(double)pl[2],
          &x2,&y2,&z2);
     }
     else{
       p = &(n->points[ii][jj]);
       w=1.0/p->w; x1=((p->x)*w); y1=((p->y)*w); z1=((p->z)*w);
       p = &(n->points[ii][jj+1]);
       w=1.0/p->w; x2=((p->x)*w); y2=((p->y)*w); z2=((p->z)*w);
       m4by1(t1 ,x1,y1,z1
                ,&x1,&y1,&z1);
       m4by1(t1 ,x2,y2,z2
                ,&x2,&y2,&z2);
     }
     if(y1 > dt){ /* point infront of viewing plane */
       ix1=(int)(pwincx+scalex*x1*dt/(y1));
       iy1=(int)(pwincy-scaley*z1*dt/(y1));
       if(ix1 < -16000 || ix1 > 16000)flag=1;
       if(iy1 < -16000 || iy1 > 16000)flag=1;
     }
     else  flag=1;
     if(y2 > dt){
       ix2=(int)(pwincx+scalex*x2*dt/(y2));
       iy2=(int)(pwincy-scaley*z2*dt/(y2));
       if(ix2 < -16000 || ix2 > 16000)flag=1;
       if(iy2 < -16000 || iy2 > 16000)flag=1;
     }
     else flag=1;
     if(flag == 0){
        MoveToEx(hdcView,ix1,iy1,NULL);
        LineTo(hdcView,ix2,iy2);
     }
   }
   // second pass
   for(jj=0;jj<n->numU;jj++)
   for(ii=0;ii<n->numV-1;ii++){
     flag=0;
     if(morphflag){
       p = &((Op->last->Nurbs+k)->points[ii][jj]);
       w=1.0/p->w; pl[0]=((p->x)*w); pl[1]=((p->y)*w); pl[2]=((p->z)*w);
       SubPoints(pl,Op->last->origin,pl);
       AddPoints(pl,Op->origin,pl);
       p = &(n->points[ii][jj]);
       w=1.0/p->w; pc[0]=((p->x)*w); pc[1]=((p->y)*w); pc[2]=((p->z)*w);
       m4by1(t1,(double)(pc[0]-pl[0])*mr+(double)pl[0],
                (double)(pc[0]-pl[1])*mr+(double)pl[1],
                (double)(pc[0]-pl[1])*mr+(double)pl[2],
          &x1,&y1,&z1);
       p = &((Op->last->Nurbs+k)->points[ii+1][jj]);
       w=1.0/p->w; pl[0]=((p->x)*w); pl[1]=((p->y)*w); pl[2]=((p->z)*w);
       p = &(n->points[ii+1][jj]);
       w=1.0/p->w; pc[0]=((p->x)*w); pc[1]=((p->y)*w); pc[2]=((p->z)*w);
       m4by1(t1,(double)(pc[0]-pl[0])*mr+(double)pl[0],
                (double)(pc[0]-pl[1])*mr+(double)pl[1],
                (double)(pc[0]-pl[1])*mr+(double)pl[2],
          &x2,&y2,&z2);
     }
     else{
       p = &(n->points[ii][jj]);
       w=1.0/p->w; x1=((p->x)*w); y1=((p->y)*w); z1=((p->z)*w);
       p = &(n->points[ii+1][jj]);
       w=1.0/p->w; x2=((p->x)*w); y2=((p->y)*w); z2=((p->z)*w);
       m4by1(t1 ,x1,y1,z1
                ,&x1,&y1,&z1);
       m4by1(t1 ,x2,y2,z2
                ,&x2,&y2,&z2);
     }
     if(y1 > dt){ /* point infront of viewing plane */
       ix1=(int)(pwincx+scalex*x1*dt/(y1));
       iy1=(int)(pwincy-scaley*z1*dt/(y1));
       if(ix1 < -16000 || ix1 > 16000)flag=1;
       if(iy1 < -16000 || iy1 > 16000)flag=1;
     }
     else  flag=1;
     if(y2 > dt){
       ix2=(int)(pwincx+scalex*x2*dt/(y2));
       iy2=(int)(pwincy-scaley*z2*dt/(y2));
       if(ix2 < -16000 || ix2 > 16000)flag=1;
       if(iy2 < -16000 || iy2 > 16000)flag=1;
     }
     else flag=1;
     if(flag == 0){
        MoveToEx(hdcView,ix1,iy1,NULL);
        LineTo(hdcView,ix2,iy2);
     }
   }
   // end of passes
 }
}

void PerspectiveView0(int smab, int DisplayType){
 RECT rc;
 if(ghwndOpenGLview != NULL){
   UpdateGLview(FALSE);
   return;
 }
 if(ghdc_view_Bitmap == NULL || ghwnd_view == NULL)return;
 UpdateFrameBox();
 GetClientRect(ghwnd_view,&rc);
 PatBlt(ghdc_view_Bitmap,0,0,rc.right,rc.bottom,PATCOPY);
 DrawPerspectiveView(smab,DisplayType,
                     ghdc_view_Bitmap,ghwnd_view,ghWireframePen);
 InvalidateRect(ghwnd_view,NULL,FALSE);
}

static BOOL bRendering=FALSE;

static void PerspectiveView1(void *xxx){
//static DWORD WINAPI PerspectiveView1(LPVOID xxx){
 int smab=0,DisplayType;
 RECT rc;
 bRendering=TRUE;
 DisplayType=(int)xxx;
 GetClientRect(ghwnd_view,&rc);
 PatBlt(ghdc_view_Bitmap,0,0,rc.right,rc.bottom,PATCOPY);
 DrawPerspectiveView(smab,DisplayType,
                     ghdc_view_Bitmap,ghwnd_view,ghWireframePen);
 InvalidateRect(ghwnd_view,NULL,FALSE);
 bRendering=FALSE;
 _endthread();
// ExitThread(0);
 return;
}

void PerspectiveView(int smab, int DisplayType){
 void (*fp)(void *);
// DWORD WINAPI (*fp)(LPVOID);
// DWORD threadID;
 if(ghwndOpenGLview != NULL){
   UpdateGLview(FALSE);
   return;
 }
 if(ghdc_view_Bitmap == NULL || ghwnd_view == NULL)return;
 if(DisplayType == 0){
   if(bRendering){
     bAbortedDrawing=TRUE;
     Sleep(10);
     if(bAbortedDrawing)return;
   }
   PerspectiveView0(smab,DisplayType);
   return;
 }
 UpdateFrameBox();
 if(bRendering){
   bAbortedDrawing=TRUE;
   Sleep(10);
   if(bAbortedDrawing)return;
 }
 fp=PerspectiveView1;
#if  __WATCOMC__ && __WATCOMC__ < 1100
  _beginthread(fp,0,65536,(void *)DisplayType);
#else
  _beginthread(fp,0,(void *)DisplayType);
#endif
// CreateThread(NULL,0,fp,(LPVOID)DisplayType,0,&threadID);
}

void DrawPerspectiveView(int smab, int DisplayType,
                         HDC hDC, HWND hWnd, HPEN hPen){
 node *selected_camera;
 RECT rc;
 double xc,yc,zc,alpha,theta,phi,t1[4][4],t2[4][4],t3[4][4],sx,sy,sz,ima;
 short im;
 point camera,TrackedOffset;
 double lenz=50.0;
 dt=1.0;
 GetClientRect(hWnd,&rc);
 scalex=scaley=(rc.right)/2/21.22*lenz;
 selected_camera=GetSelectedCamera(CurrentFrame,SelectedCamera);
 if(selected_camera == NULL)selected_camera=FirstNp;
 if(selected_camera->type != CAMERA)return;
 GetTransform(0,CurrentFrame,1.0,
              selected_camera,selected_camera->fobj,camera,TrackedOffset,
              &phi,&theta,&alpha,&sx,&sy,&sz,&im,&ima);
 /* IM is ignored no internal for camera */
 scalex *= sx;
 scaley *= sy;
 xc=(double)camera[0];
 yc=(double)camera[1];
 zc=(double)camera[2];
 theta *= PIo180;
 phi   *= PIo180;
 alpha *= PIo180;
 CameraAlpha=alpha;
 CameraTheta=theta;
 CameraHeight=zc;
 tram(t1,-xc,-yc,-zc);
 rotz(t2,-phi);
 m4by4(t2,t1,t3);
 rotx(t1,-alpha);
 m4by4(t1,t3,t2);
 roty(t3,-theta);   /* camera banking */
 m4by4(t3,t2,trr);
 SelectObject(hDC,hPen);
 DrawStagePerspective(hDC,hWnd,DisplayType,camera,TRUE);
 return;
}

// preview code

static void    DrawPreviewFrame(HWND hwnd, HDC hdc, HBITMAP hbm);
static PSTR    lpszAnimatorPreviewClass="OFX:PreviewClass";
static HBITMAP *PreviewBitmaps=NULL;
static int     nPreviewFrames=0,iPlayFrame,StartPreviewFrame,EndPreviewFrame;
static HWND    hwndPreview=NULL;
static UINT    PreviewTimer=0;
static int     preview_quickdraw=0;
static int     preview_speed=30;
static RECT    mRect;
static HBITMAP PreviewButtonBitmaps[6]={NULL,NULL,NULL,NULL,NULL,NULL};
static int     PreviewButtonIDs[6]={
                IDBM_PREV_FIRST,IDBM_PREV_PREV,IDBM_PREV_PLAY,
                IDBM_PREV_NEXT,IDBM_PREV_LAST,IDBM_PREV_PAUSE};

#define IDM_PREVIEW_FREE 0x10

static void DrawPreviewFrame(HWND hwnd, HDC hdc, HBITMAP hbm){
 node *selected_camera;
 RECT rc;
 double xc,yc,zc,alpha,theta,phi,t1[4][4],t2[4][4],t3[4][4],sx,sy,sz,ima;
 short im;
 point camera,TrackedOffset;
 double lenz=50.0;
 if(hdc == NULL || hbm == NULL)return;
 dt=1.0;
 GetClientRect(hwnd,&rc);
 PatBlt(hdc,0,0,rc.right,rc.bottom,PATCOPY);
 scalex=scaley=(rc.right)/2/21.22*lenz;
 selected_camera=GetSelectedCamera(CurrentFrame,SelectedCamera);
 if(selected_camera == NULL)selected_camera=FirstNp;
 if(selected_camera->type != CAMERA)return;
 GetTransform(0,CurrentFrame,1.0,
              selected_camera,selected_camera->fobj,camera,TrackedOffset,
              &phi,&theta,&alpha,&sx,&sy,&sz,&im,&ima);
/* IM is ignored no internal for camera */
 scalex *= sx;
 scaley *= sy;
 xc=(double)camera[0];
 yc=(double)camera[1];
 zc=(double)camera[2];
 theta *= PIo180;
 phi   *= PIo180;
 alpha *= PIo180;
 CameraAlpha=alpha;
 CameraTheta=theta;
 CameraHeight=zc;
 tram(t1,-xc,-yc,-zc);
 rotz(t2,-phi);
 m4by4(t2,t1,t3);
 rotx(t1,-alpha);
 m4by4(t1,t3,t2);
 roty(t3,-theta);   /* camera banking */
 m4by4(t3,t2,trr);
 SelectObject(hdc,GetStockObject(BLACK_PEN));
 DrawStagePerspective(hdc,hwnd,1,camera,FALSE);
 if(GetQueueStatus(QS_ALLINPUT)){
   MSG msg;
   while (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE)){
     TranslateMessage(&msg);
     DispatchMessage(&msg);
   }
 }
 return;
}

void FreePreview(void){
 int i;
 if(nPreviewFrames == 0 || PreviewBitmaps == NULL)return;
 for(i=0;i<nPreviewFrames;i++){
   if(PreviewBitmaps[i] != NULL)DeleteObject(PreviewBitmaps[i]);
 }
 X__Free(PreviewBitmaps);
 PreviewBitmaps=NULL;
 nPreviewFrames=0;
}

void MoveCtrlWindow(HWND hdlg,int id,
                   int tx, int ty, int cx, int cy,
                   int type){
 POINT pp;
 RECT wRect;
 GetWindowRect(GetDlgItem(hdlg,id),&wRect);
 pp.x=wRect.left; pp.y=wRect.top;
 ScreenToClient(hdlg,&pp);
 wRect.left=pp.x; wRect.top=pp.y;
 pp.x=wRect.right; pp.y=wRect.bottom;
 ScreenToClient(hdlg,&pp);
 wRect.right=pp.x; wRect.bottom=pp.y;
 if(type == 0){
   wRect.right  += (tx - cx);
   wRect.bottom += (ty - cy);
 }
 else{
   if(type == 1)OffsetRect(&wRect,(tx - cx),0);
   else         OffsetRect(&wRect,0,(ty - cy));
 }
 MoveWindow(GetDlgItem(hdlg,id),wRect.left,wRect.top,
            wRect.right-wRect.left,wRect.bottom-wRect.top,TRUE);
}

static BOOL CALLBACK PreviewDlgProc(HWND hdlg, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 static BOOL bMakingPreview=FALSE;
 static HWND hWndStatus;
 static int cx,cy,ox,oy;
 HWND hctl;
 int i,tx,ty;
 char str[32];
 RECT wRect;
 LPRECT lpRect;
 switch( msg ) {
   case WM_INITDIALOG:
     SetClassLong(hdlg,GCL_HICON,
                 (LONG)LoadIcon(ghinst_main,"ANIMATORICON"));
     hWndStatus=CreateWindowEx(0L,STATUSCLASSNAME,"",
                WS_CHILD | !WS_BORDER | WS_VISIBLE,
                0,0,0,0,
                hdlg,
                (HMENU)NULL,
                (HINSTANCE)NULL,
                NULL);
     SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)SBT_NOBORDERS,(LPARAM)NULL);
     if(preview_quickdraw == 1)
       SendDlgItemMessage(hdlg,DLG_PREVIEW_QUICK,BM_SETCHECK,TRUE,0);
     else
       SendDlgItemMessage(hdlg,DLG_PREVIEW_FULL,BM_SETCHECK,TRUE,0);
     if(preview_speed == 30)
       SendDlgItemMessage(hdlg,DLG_PREVIEW_30,BM_SETCHECK,TRUE,0);
     else if(preview_speed == 25)
       SendDlgItemMessage(hdlg,DLG_PREVIEW_25,BM_SETCHECK,TRUE,0);
     else if(preview_speed == 15)
       SendDlgItemMessage(hdlg,DLG_PREVIEW_15,BM_SETCHECK,TRUE,0);
     else if(preview_speed == 5)
       SendDlgItemMessage(hdlg,DLG_PREVIEW_5,BM_SETCHECK,TRUE,0);
     else if(preview_speed == 1)
       SendDlgItemMessage(hdlg,DLG_PREVIEW_1,BM_SETCHECK,TRUE,0);
     for(i=0;i<6;i++){
       PreviewButtonBitmaps[i]=LoadBitmap(ghinst_main,
         MAKEINTRESOURCE(PreviewButtonIDs[i]));
     }
     if(LOBYTE(LOWORD(GetVersion())) < 4){;}
     else{
       SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_FIRST),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)PreviewButtonBitmaps[0]);
       SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PREVIOUS),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)PreviewButtonBitmaps[1]);
       SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PLAY),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)PreviewButtonBitmaps[2]);
       SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_NEXT),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)PreviewButtonBitmaps[3]);
       SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_LAST),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)PreviewButtonBitmaps[4]);
     }
     SetDlgItemInt(hdlg,DLG_PREVIEW_RANGE1,1,FALSE);
     SetDlgItemInt(hdlg,DLG_PREVIEW_RANGEN,Nframes,FALSE);
//     CentreDialogOnScreen(hdlg);
     GetWindowRect(hdlg,&wRect);
     ox=wRect.right-wRect.left;
     oy=wRect.bottom-wRect.top;
     GetClientRect(hdlg,&wRect);
     cx=wRect.right;
     cy=wRect.bottom;
     PostMessage(hdlg,(WM_USER+100),0,0);
     return (TRUE);
   case (WM_USER+100):{
       int xp,yp,xl,yl;
       xp=GetPrivateProfileInt(IniSection,"PRPOSX",-1,IniFilename);
       yp=GetPrivateProfileInt(IniSection,"PRPOSY",-1,IniFilename);
       xl=GetPrivateProfileInt(IniSection,"PRSIZX",-1,IniFilename);
       yl=GetPrivateProfileInt(IniSection,"PRSIZY",-1,IniFilename);
       if(xp < 0 || yp < 0 || xl < 0 || yl < 0)CentreDialogOnScreen(hdlg);
       else     SetWindowPos(hdlg,NULL,xp,yp,xl,yl,SWP_NOZORDER);
       ShowWindow(hdlg,SW_SHOW);
     }
     break;
   case WM_DESTROY:{
       RECT rc; char str[32];
       for(i=0;i<6;i++)if(PreviewButtonBitmaps[i] != NULL)
         DeleteObject(PreviewButtonBitmaps[i]);
       GetWindowRect(hdlg,&rc);
       sprintf(str,"%ld",rc.left);
       WritePrivateProfileString(IniSection,"PRPOSX",str,IniFilename);
       sprintf(str,"%ld",rc.top);
       WritePrivateProfileString(IniSection,"PRPOSY",str,IniFilename);
       sprintf(str,"%ld",rc.right-rc.left);
       WritePrivateProfileString(IniSection,"PRSIZX",str,IniFilename);
       sprintf(str,"%ld",rc.bottom-rc.top);
       WritePrivateProfileString(IniSection,"PRSIZY",str,IniFilename);
     }
     break;
   case WM_HSCROLL:
     if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_PREVIEW_SCROLL))){
       i=iPlayFrame;
       switch (LOWORD(wparam)) {
         case TB_TOP:           i=0; break;
         case TB_BOTTOM:        i=nPreviewFrames-1; break;
         case TB_PAGEDOWN:      break;
         case TB_LINEDOWN:      i++; break;
         case TB_PAGEUP:        break;
         case TB_LINEUP:        i--; break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION: i=HIWORD(wparam);break;
         default: break;
       }
       i=min(nPreviewFrames-1,max(i,0));
       if(i != iPlayFrame){
         iPlayFrame=i;
         InvalidateRect(hwndPreview,NULL,FALSE);
       }
     }
     break;
   case WM_SIZING:
     lpRect=(LPRECT)lparam;
     if((lpRect->right - lpRect->left) < ox)
       lpRect->right = lpRect->left+ox;
     if((lpRect->bottom - lpRect->top) < oy)
       lpRect->bottom = lpRect->top+oy;
     return TRUE;
     break;
   case WM_SIZE:
     SendMessage(hWndStatus,WM_SIZE,wparam,lparam);
     tx=LOWORD(lparam); ty=HIWORD(lparam);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_WIN,tx,ty,cx,cy,0);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_TX1,tx,ty,cx,cy,0);
     MoveCtrlWindow(hdlg,IDCANCEL,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_MAKE,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_PROGRESS,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_RANGE1,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_RANGEN,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_TX2,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_TX3,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_TX4,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_TX5,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_TX6,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_QUICK,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_FULL,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_1,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_5,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_15,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_25,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_30,tx,ty,cx,cy,1);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_FIRST,tx,ty,cx,cy,2);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_PREVIOUS,tx,ty,cx,cy,2);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_PLAY,tx,ty,cx,cy,2);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_NEXT,tx,ty,cx,cy,2);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_LAST,tx,ty,cx,cy,2);
     MoveCtrlWindow(hdlg,DLG_PREVIEW_SCROLL,tx,ty,cx,cy,2);
     InvalidateRect(hdlg,NULL,TRUE);
     cx=tx; cy=ty;
     break;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case DLG_PREVIEW_FULL:  preview_quickdraw=0; break;
       case DLG_PREVIEW_QUICK: preview_quickdraw=1; break;
       case DLG_PREVIEW_30:    preview_speed=30;    break;
       case DLG_PREVIEW_25:    preview_speed=25;    break;
       case DLG_PREVIEW_15:    preview_speed=15;    break;
       case DLG_PREVIEW_5:     preview_speed=5;     break;
       case DLG_PREVIEW_1:     preview_speed=1;     break;
       case DLG_PREVIEW_FIRST:
         iPlayFrame=0;
         SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETPOS,
                            TRUE,iPlayFrame);
         InvalidateRect(hwndPreview,NULL,FALSE);
         break;
       case DLG_PREVIEW_PREVIOUS:
         iPlayFrame -= 1;
         iPlayFrame = max(0,iPlayFrame);
         SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETPOS,
                            TRUE,iPlayFrame);
         InvalidateRect(hwndPreview,NULL,FALSE);
         break;
       case DLG_PREVIEW_NEXT:
         iPlayFrame += 1;
         iPlayFrame = min(nPreviewFrames-1,iPlayFrame);
         SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETPOS,
                            TRUE,iPlayFrame);
         InvalidateRect(hwndPreview,NULL,FALSE);
         break;
       case DLG_PREVIEW_LAST:
         iPlayFrame=nPreviewFrames-1;
         SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETPOS,
                            TRUE,iPlayFrame);
         InvalidateRect(hwndPreview,NULL,FALSE);
         break;
       case DLG_PREVIEW_PLAY:
         if(preview_playing){
           preview_playing=0;
           KillTimer(hwndPreview,PreviewTimer); PreviewTimer=0;
           SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETPOS,
                              TRUE,iPlayFrame);
           LoadString(ghinst_main,IDX_MISC_PLAY,str,32);
           SendDlgItemMessage(hdlg,DLG_PREVIEW_PLAY,WM_SETTEXT,
                              0,(LPARAM)str);
           SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PLAY),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)PreviewButtonBitmaps[2]);
           EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_MAKE),TRUE);
           EnableWindow(GetDlgItem(hdlg,IDCANCEL),TRUE);
           EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_FIRST),TRUE);
           EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_LAST),TRUE);
           EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PREVIOUS),TRUE);
           EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_NEXT),TRUE);
         }
         else{
           if(preview_speed == 1)i=1000;
           if(preview_speed == 5)i=200;
           if(preview_speed == 15)i=66;
           if(preview_speed == 25)i=40;
           if(preview_speed == 30)i=33;
           if((PreviewTimer=SetTimer(hwndPreview,1,i,NULL)) != 0){
             preview_playing=1;
             LoadString(ghinst_main,IDX_MISC_STOP,str,32);
             SendDlgItemMessage(hdlg,DLG_PREVIEW_PLAY,WM_SETTEXT,
                                0,(LPARAM)str);
             SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PLAY),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)PreviewButtonBitmaps[5]);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_MAKE),FALSE);
             EnableWindow(GetDlgItem(hdlg,IDCANCEL),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_FIRST),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_LAST),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PREVIOUS),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_NEXT),FALSE);
           }
         }
         break;
       case DLG_PREVIEW_MAKE:
         if(bMakingPreview){
           FreePreview();
           break;
         }
         if(preview_playing)break;
         if(PreviewBitmaps != NULL)FreePreview();
         if(hwndPreview != NULL){
           HDC     hdc,hdcMem;
           RECT    rc;
           HBITMAP hbm,hbmold=NULL;
           char    str[32];
           long    i,cf_store,qd_store;
           BOOL    err;
           if((PreviewBitmaps=(HBITMAP *)X__Malloc(sizeof(HBITMAP)*Nframes))
              != NULL){
             StartPreviewFrame=GetDlgItemInt(hdlg,DLG_PREVIEW_RANGE1,&err,FALSE);
             EndPreviewFrame=GetDlgItemInt(hdlg,DLG_PREVIEW_RANGEN,&err,FALSE);
             StartPreviewFrame=max(1,StartPreviewFrame);
             StartPreviewFrame=min(Nframes,StartPreviewFrame);
             EndPreviewFrame=max(1,EndPreviewFrame);
             EndPreviewFrame=min(Nframes,EndPreviewFrame);
             EndPreviewFrame=max(StartPreviewFrame,EndPreviewFrame);
             nPreviewFrames=EndPreviewFrame-StartPreviewFrame+1;
//             nPreviewFrames=Nframes;  //
             bMakingPreview=TRUE;
             SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,
                                (LPARAM)MAKELONG(0,nPreviewFrames-1));
             SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETLINESIZE,0,1);
             SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETPAGESIZE,0,0);
             SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETTIC,0,
                                (LPARAM)(nPreviewFrames/2));
             cf_store=CurrentFrame;
             qd_store=global_quickdraw;
             ActivateAllMenus(ghwnd_main,MF_GRAYED);
             EnableToolPannels(ALL_PANNELS,FALSE);
             LoadString(ghinst_main,IDX_MISC_CANCELPREVIEW,res_str,256);
             SetDlgItemText(hdlg,DLG_PREVIEW_MAKE,res_str);
             EnableWindow(GetDlgItem(hdlg,IDCANCEL),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PLAY),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_FIRST),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_LAST),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PREVIOUS),FALSE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_NEXT),FALSE);
             for(i=0;i<nPreviewFrames;i++)PreviewBitmaps[i]=NULL;
             SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PROGRESS),PBM_SETRANGE,0,MAKELPARAM(0,nPreviewFrames));
             SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PROGRESS),PBM_SETSTEP,(WPARAM)1,0);
             SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PROGRESS),PBM_SETPOS,(WPARAM)0,0);
             hdc=GetDC(hwndPreview);
             GetClientRect(hwndPreview,&rc);
             CopyRect(&mRect,&rc);
             global_quickdraw=preview_quickdraw;
             for(i=0;i<nPreviewFrames;i++){
               CurrentFrame=StartPreviewFrame+i;
//               CurrentFrame=i+1;
               LoadString(ghinst_main,IDX_MISC_PREPARINGFRAME,res_str,256);
               sprintf(str,res_str,
                       CurrentFrame,Nframes);
               SendPrgmText(str);
               if((hbm=CreateBitmap(rc.right,rc.bottom,1,1,NULL)) == NULL){
                 ReleaseDC(hwndPreview,hdc);
                 CurrentFrame=cf_store;
                 global_quickdraw=qd_store;
                 ActivateAllMenus(ghwnd_main,MF_ENABLED);
                 EnableToolPannels(ALL_PANNELS,TRUE);
                 LoadString(ghinst_main,IDX_MISC_MAKEPREVIEW,res_str,256);
                 SetDlgItemText(hdlg,DLG_PREVIEW_MAKE,res_str);
                 EnableWindow(GetDlgItem(hdlg,IDCANCEL),TRUE);
                 EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PLAY),TRUE);
                 EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_FIRST),TRUE);
                 EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_LAST),TRUE);
                 EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PREVIOUS),TRUE);
                 EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_NEXT),TRUE);
                 FreePreview();
                 bMakingPreview=FALSE;
                 return FALSE;
               }
               PreviewBitmaps[i]=hbm;
               hdcMem=CreateCompatibleDC(hdc);
               hbmold=SelectObject(hdcMem,hbm);
               DrawPreviewFrame(hwndPreview,hdcMem,hbm);
               sprintf(str,"%4.4ld",CurrentFrame);
               TextOut(hdcMem,1,1,str,strlen(str));
               BitBlt(hdc,0,0,rc.right,rc.bottom,hdcMem,0,0,SRCCOPY);
               SelectObject(hdcMem,hbmold);
               DeleteDC(hdcMem);
               SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PROGRESS),PBM_SETPOS,
                          (WPARAM)i,0);
             }
             ReleaseDC(hwndPreview,hdc);
             SendMessage(GetDlgItem(hdlg,DLG_PREVIEW_PROGRESS),PBM_SETPOS,
                         (WPARAM)nPreviewFrames,0);
             CurrentFrame=cf_store;
             global_quickdraw=qd_store;
             ActivateAllMenus(ghwnd_main,MF_ENABLED);
             EnableToolPannels(ALL_PANNELS,TRUE);
             LoadString(ghinst_main,IDX_MISC_PREVIEWCOMPLETE,res_str,256);
             SendPrgmText(res_str);
             iPlayFrame=0;
             InvalidateRect(hwndPreview,NULL,FALSE);
             UpdateWindow(hwndPreview);
             LoadString(ghinst_main,IDX_MISC_MAKEPREVIEW,res_str,256);
             SetDlgItemText(hdlg,DLG_PREVIEW_MAKE,res_str);
             EnableWindow(GetDlgItem(hdlg,IDCANCEL),TRUE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PLAY),TRUE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_FIRST),TRUE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_LAST),TRUE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_PREVIOUS),TRUE);
             EnableWindow(GetDlgItem(hdlg,DLG_PREVIEW_NEXT),TRUE);
             SendDlgItemMessage(hdlg,DLG_PREVIEW_SCROLL,TBM_SETPOS,
                                TRUE,iPlayFrame);
             bMakingPreview=FALSE;
           }
         }
         break;
       case IDCANCEL:
         EndDialog(hdlg,TRUE);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static LRESULT CALLBACK PreviewWndProc(HWND hWnd,UINT uMsg,WPARAM wParam,
                                                           LPARAM lParam){
 HDC hDC;
 RECT rc;
 PAINTSTRUCT ps;
 switch (uMsg) {
   case WM_CREATE:
     hwndPreview=hWnd;
     break;
   case WM_TIMER:
     switch(wParam){
       case 1:
         iPlayFrame++;
         if(iPlayFrame >= nPreviewFrames)iPlayFrame=0;
         SendDlgItemMessage(GetParent(hWnd),DLG_PREVIEW_SCROLL,
                            TBM_SETPOS,TRUE,iPlayFrame);
         InvalidateRect(hWnd,NULL,FALSE);
     }
     break;
   case WM_PAINT:
     hDC=BeginPaint(hWnd,&ps);
     if(PreviewBitmaps != NULL && PreviewBitmaps[iPlayFrame] != NULL){
       HDC hdcMem;
       PAINTSTRUCT ps;
       RECT rc;
       HBITMAP hbmold;
       hdcMem=CreateCompatibleDC(hDC);
       hbmold=SelectObject(hdcMem,PreviewBitmaps[iPlayFrame]);
       GetClientRect(hWnd,&rc);
       if(rc.right == mRect.right && rc.bottom == mRect.bottom)
         BitBlt(hDC,0,0,rc.right,rc.bottom,hdcMem,0,0,SRCCOPY);
       else
         StretchBlt(hDC,   0,0,rc.right,rc.bottom,
                    hdcMem,0,0,mRect.right,mRect.bottom,SRCCOPY);
       SelectObject(hdcMem,hbmold);
       DeleteDC(hdcMem);
     }
     EndPaint(hWnd,&ps);
     break;
   default:
     return(DefWindowProc(hWnd,uMsg,wParam,lParam));
 }
 return( 0L );
}

int MakePreview(void){
 WNDCLASS   wc;
 static BOOL preview_busy=FALSE;
 if(preview_busy)return FAIL;
 if(Nframes < 2){
   SendPrgmQuery(IDQ_PREVIEW_NOMAKE,0);
   return FAIL;
 }
 wc.lpszClassName = lpszAnimatorPreviewClass;
 wc.style = CS_SAVEBITS;
 wc.lpfnWndProc = (WNDPROC)PreviewWndProc;
 wc.cbClsExtra = 0;
 wc.cbWndExtra = 0;
 wc.hInstance = ghinst_main;
 wc.hIcon = LoadIcon(ghinst_main,"ANIMATORICON");
 wc.hCursor = NULL;
 wc.hbrBackground = NULL;
 wc.lpszMenuName = NULL;
 if(!RegisterClass(&wc))return FAIL;
 preview_busy=TRUE;
 hwndPreview=NULL;
 iPlayFrame=0;
 DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_PREVIEW),ghwnd_main,
                (DLGPROC)PreviewDlgProc);
 FreePreview();
 hwndPreview=NULL;
 UnregisterClass(lpszAnimatorPreviewClass,ghinst_main);
 preview_busy=FALSE;
 return OK;
}

// end preview code

#include "..\effects\sfxinfo.h"

static BOOL CalculateEffectedVertices(double mr,
                                      double tall[4][4],
                                      double tcamera[4][4],
                                      double tpos[4][4],
                                      BOOL ground,
                                      point p, point gpos, point cpos,
                                      object *Op, Svertex *Vp){
 float t,vmax[3],vmin[3];
 sfxinfo SFX,*SFXinfo;
 sfxdata SFD,*SFXdata;
 FARPROC fpFun;
 void *(*fpFun1)(char *, sfxinfo *, Svertex *);
 void *(*fpFun2)(char *, sfxinfo *, sfxdata *, Svertex *);
 double x0,y0,z0,x1,y1,z1;
 long Nv,i,j;
 if(Op->hEffect == NULL || Op->effect == NULL)return FALSE;
 if((fpFun=GetProcAddress(Op->hEffect,"_PreviewExternalEffect")) == NULL){
   return FALSE;
 }
 Nv=Op->npoints;
 t=(float)(Op->lastframe - Op->firstframe+1);
 t=(float)(CurrentFrame - Op->firstframe+1)/t;
 SFXinfo=&SFX;
 SFX.nvert=Nv;
 SFX.time=t;
 SFX.ltime=(long)(t*1000);
 SFX.Nfaces=0;
 SFX.Faces=NULL;
#if __LITE__
 SFX.version=0;
#else
 SFX.version=1;
#endif
 for(j=0;j<3;j++){
   SFX.origin[j]=Op->origin[j];
   vmax[j]= -1.e30;
   vmin[j]=  1.e30;
 }
 fpFun1 = (void *)fpFun;
 if(mr < 0.0){
   for(i=0;i<Nv;i++){
     for(j=0;j<3;j++){
       Vp[i][j]=(float)(Op->points[i][j] - Op->origin[j]);
       vmin[j] = min(vmin[j],Vp[i][j]);
       vmax[j] = max(vmax[j],Vp[i][j]);
     }
   }
 }
 else if(Op->last != NULL){
   long l,m;
   for(i=0;i<Nv;i++){
     for(j=0;j<3;j++){
       l = Op->points[i][j] - Op->origin[j];
       m = Op->last->points[i][j] - Op->last->origin[j];
       Vp[i][j] = (double)(l-m)*mr + (double)m;
       vmin[j] = min(vmin[j],Vp[i][j]);
       vmax[j] = max(vmax[j],Vp[i][j]);
     }
   }
 }
 for(j=0;j<3;j++){
   SFX.vmin[j]=(long)vmin[j];
   SFX.vmax[j]=(long)vmax[j];
 }
 (*fpFun1)(Op->effect,SFXinfo,Vp);
 if((fpFun=GetProcAddress(Op->hEffect,"_PreviewGlobalEffect")) == NULL){
   for(i=0;i<Nv;i++){ /* just the local effect */
     x0=(double)Vp[i][0]; y0=(double)Vp[i][1]; z0=(double)Vp[i][2];
     m4by1(tall,x0,y0,z0,&x1,&y1,&z1);
     Vp[i][0]=(float)x1; Vp[i][1]=(float)y1; Vp[i][2]=(float)z1;
   }
 }
 else{
   for(i=0;i<Nv;i++){
     m4by1(tpos,
           (double)Vp[i][0],(double)Vp[i][1],(double)Vp[i][2],
           &x1,&y1,&z1);
     Vp[i][0]=(float)x1; Vp[i][1]=(float)y1; Vp[i][2]=(float)z1;
   }
   SFXdata=&SFD;
   memset(SFXdata,0,sizeof(sfxdata));
   SFD.size=sizeof(sfxdata);
   SFD.version=1;
   SFD.ground=ground;
   VECCOPY(p,SFD.pos)
   VECCOPY(gpos,SFD.ground_pos)
   VECCOPY(cpos,SFD.camera_pos)
   fpFun2 = (void *)fpFun;
   (*fpFun2)(Op->effect,SFXinfo,SFXdata,Vp);
   for(i=0;i<Nv;i++){
     m4by1(tcamera,
           (double)Vp[i][0],(double)Vp[i][1],(double)Vp[i][2],
           &x1,&y1,&z1);
     Vp[i][0]=(float)x1; Vp[i][1]=(float)y1; Vp[i][2]=(float)z1;
   }
 }
 return TRUE;
}

static void ClipToPlane(double *x1, double *y1, double *z1,
                        double x2, double y2, double z2, long Cp){
 double mu;
 mu=(y2 - *y1);
 if(fabs(mu) < 1.e-3)mu=1.e-3;
 mu=((double)Cp - *y1)/mu;
 *x1 = *x1 + mu * (x2 - *x1);
 *y1 = *y1 + mu * (y2 - *y1);
 *z1 = *z1 + mu * (z2 - *z1);
 return;
}

static void ProjectToX(long *x1, long *y1, long *x2, long *y2, long x){
 double dx,dy;
 dx=(double)(*x2 - *x1);
 dy=(double)(*y2 - *y1);
 *y1 = *y1 + (long)((double)(x - *x1)*dy/dx);
 *x1 = x;
}

static void ProjectToY(long *x1, long *y1, long *x2, long *y2, long y){
 double dx,dy;
 dx=(double)(*x2 - *x1);
 dy=(double)(*y2 - *y1);
 *x1 = *x1 + (long)((double)(y - *y1)*dx/dy);
 *y1 = y;
}

static BOOL ClipToWindow(long *x1, long *y1,
                         long *x2, long *y2,
                         long xmin, long ymin,
                         long xmax, long ymax){
 if(*x1 < xmin && *x2 < xmin)return FALSE;
 if(*x1 > xmax && *x2 > xmax)return FALSE;
 if(*y1 < ymin && *y2 < ymin)return FALSE;
 if(*y1 > ymax && *y2 > ymax)return FALSE;
 if(*x1 >= xmin && *x1 <= xmax &&
    *y1 >= ymin && *y1 <= ymax &&
    *x2 >= xmin && *x2 <= xmax &&
    *y2 >= ymin && *y2 <= ymax)return TRUE;
 if     (*x1 < xmin)ProjectToX(x1,y1,x2,y2,xmin);
 else if(*x1 > xmax)ProjectToX(x1,y1,x2,y2,xmax);
 if     (*x2 < xmin)ProjectToX(x2,y2,x1,y1,xmin);
 else if(*x2 > xmax)ProjectToX(x2,y2,x1,y1,xmax);
 if     (*y1 < ymin)ProjectToY(x1,y1,x2,y2,ymin);
 else if(*y1 > ymax)ProjectToY(x1,y1,x2,y2,ymax);
 if     (*y2 < ymin)ProjectToY(x2,y2,x1,y1,ymin);
 else if(*y2 > ymax)ProjectToY(x2,y2,x1,y1,ymax);
 return TRUE;
}

/* Warning on mirror??? should be  + scale*x1 etc WHY */

void DrawViewRobot(HDC hdcView, HBITMAP hbmView, HWND hWnd,
                   node *Np, int arm, int axis,
                   double ra, double ua,
                   BOOL zoom,BOOL zoom_state){
 RECT    rc;
 double  scale,size,x1,y1,z1,x2,y2,z2,t1[4][4],t2[4][4],t3[4][4];
 vector  cc;
 int     pwinr,pwinb,pwincx,pwincy;
 int     ix1,iy1,ix2,iy2;
 long    i,j,k,n;
 object  *Op;
 skel    *sp;
 if(hdcView == NULL || hbmView == NULL)return;
 GetClientRect(hWnd,&rc);
 PatBlt(hdcView,0,0,rc.right,rc.bottom,PATCOPY);
 GetClientRect(hWnd,&rc);
 pwinr=rc.right; pwinb=rc.bottom;
 pwincx=rc.right/2; pwincy=rc.bottom/2;
 if(Np->type == ROBOT && ((Op=Np->fobj) != NULL))while(Op != NULL){
   if(CurrentFrame >= Op->firstframe && CurrentFrame <= Op->lastframe){
     if((n=Op->nskeleton) == 0 || (sp=Op->skeleton) == NULL)break;
     /* InterpolateRobot - Not Needed other windows call it */
     size=(double)GetRobotVolume(n,NULL,cc);
     scale=0.75*(double)pwinr/size;
     if(zoom && arm >=0 && arm < n){
       if(zoom_state && (j=(sp+arm)->id) >= 0 && j < n){
         tram(t1,-((sp+j)->pp[0]),-((sp+j)->pp[1]),-((sp+j)->pp[2]));
         scale *= 5.0;
       }
       else{
         tram(t1,-((sp+arm)->pp[0]),-((sp+arm)->pp[1]),-((sp+arm)->pp[2]));
         scale *= 4.0;
       }
     }
     else tram(t1,-cc[0],-cc[1],-cc[2]);
     rotz(t2, ra*PIo180);
     m4by4(t2,t1,t3);
     rotx(t1,-ua*PIo180);
     m4by4(t1,t3,t2);
     /* Bounding Box */
     SelectObject(hdcView,GetStockObject(BLACK_PEN));
     for(k=0;k<n;k++){
       m4by4(t2,(sp+k)->T,t3);
       for(i=0;i<12;i++){
         m4by1(t3,(double)(sp+k)->bx[ep1[i]][0]
                 ,(double)(sp+k)->bx[ep1[i]][1]
                 ,(double)(sp+k)->bx[ep1[i]][2]
                 ,&x1,&y1,&z1);
         m4by1(t3,(double)(sp+k)->bx[ep2[i]][0]
                 ,(double)(sp+k)->bx[ep2[i]][1]
                 ,(double)(sp+k)->bx[ep2[i]][2]
                 ,&x2,&y2,&z2);
         ix1=(int)(pwincx-scale*x1);
         iy1=(int)(pwincy-scale*z1);
         ix2=(int)(pwincx-scale*x2);
         iy2=(int)(pwincy-scale*z2);
         MoveToEx(hdcView,ix1,iy1,NULL);
         LineTo(hdcView,ix2,iy2);
       }
     }
     /* skeleton */
     SelectObject(hdcView,ghWireframePen);
     for(k=0;k<n;k++){
       x1=(double)sp[k].pp[0]; y1=(double)sp[k].pp[1]; z1=(double)sp[k].pp[2];
       m4by1(t2,x1,y1,z1,&x2,&y2,&z2);
       sp[k].wk[0]=(long)x2; sp[k].wk[1]=(long)y2; sp[k].wk[2]=(long)z2;
     }
     for(k=0;k<n;k++){
       x1=(double)sp[k].wk[0]; y1=(double)sp[k].wk[1]; z1=(double)sp[k].wk[2];
       j=sp[k].id;  /* j is parent ID */
       if(j < 0)continue;
       x2=(double)sp[j].wk[0]; y2=(double)sp[j].wk[1]; z2=(double)sp[j].wk[2];
       ix1=(int)(pwincx-scale*x1); iy1=(int)(pwincy-scale*z1);
       ix2=(int)(pwincx-scale*x2); iy2=(int)(pwincy-scale*z2);
       MoveToEx(hdcView,ix1,iy1,NULL); LineTo(hdcView,ix2,iy2);
       if(k == arm && axis < 7){ /* draw node and axis */
         HPEN hPenT;             /* ix2,iy2 is parent  */
         vector pp,x={1.0,0.0,0.0},y={0.0,1.0,0.0},z={0.0,0.0,1.0};
         Ellipse(hdcView,ix1-3,iy1-3,ix1+4,iy1+4);
         if     (axis == 0)VECCOPY((double)sp[j].vv,pp)
         else if(axis == 1)VECCOPY((double)sp[j].uu,pp)
         else if(axis == 2)VECCOPY((double)sp[j].ww,pp)
         else if(axis == 3)VECCOPY(x,pp)
         else if(axis == 4)VECCOPY(y,pp)
         else if(axis == 5)VECCOPY(z,pp)
         if(axis == 0 && (i=sp[j].id) >= 0){ /* do axis 0 other way */
           x1=sp[i].pp[0]; y1=sp[i].pp[1]; z1=sp[i].pp[2];
         }
         else if(axis < 6){
           x1=(double)sp[j].pp[0]+(double)pp[0]*size*0.1;
           y1=(double)sp[j].pp[1]+(double)pp[1]*size*0.1;
           z1=(double)sp[j].pp[2]+(double)pp[2]*size*0.1;
         }
         else {x1=sp[k].pp[0]; y1=sp[k].pp[1]; z1=sp[k].pp[2];}
         m4by1(t2,x1,y1,z1,&x2,&y2,&z2);
         ix1=(int)(pwincx-scale*x2); iy1=(int)(pwincy-scale*z2);
         hPenT=SelectObject(hdcView,GetStockObject(WHITE_PEN));
         MoveToEx(hdcView,ix2,iy2,NULL); LineTo(hdcView,ix1,iy1);
         MoveToEx(hdcView,ix2-1,iy2-1,NULL); LineTo(hdcView,ix2+2,iy2-1);
         MoveToEx(hdcView,ix2-1,iy2  ,NULL); LineTo(hdcView,ix2+2,iy2  );
         MoveToEx(hdcView,ix2-1,iy2+1,NULL); LineTo(hdcView,ix2+2,iy2+1);
         SelectObject(hdcView,hPenT);
       }
     }
     break;
   }
   Op=Op->next;
 }
 InvalidateRect(hWnd,NULL,FALSE);
}
