/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  file DIALOGS.C  contains all the dialog handler functions */

// This file contains most of the code for the dialogs in the animation
// program. Most dialogs are modal and their "window" procedure contains
// all the action code.

#define MODULE_DIALOGS 1

#include "animate.h"
#include "spin_dll.h"

#define D_BORDER 4

typedef struct tag_dlghdr1 {
  HWND hwndDlg1;
  HWND hwndDlg2;
  HWND hwndDlg3;
  HWND hwndDlg4;
  HWND hwndDlg5;
  HWND hwndDlg6;
  RECT rcTab;
} DLGHDR1;

static void DisplayImageFile(char *);
static void EditImageFile(char *);

static DLGHDR1 DlgHdr,*lpDlgHdr;
static HWND    hwndTab;

#include "..\design\loadview.c"

static void UpdateGroundDialog(HWND hwnd, object *op){
 BOOL enable=FALSE,enable_shader=FALSE;
 int gt;
 gt=(int)op->groundtype;
 SendDlgItemMessage(hwnd,DLG_GROUND_PLAIN, BM_SETCHECK,0,0);
 SendDlgItemMessage(hwnd,DLG_GROUND_SHADER,BM_SETCHECK,0,0);
 SendDlgItemMessage(hwnd,DLG_GROUND_MAPPED,BM_SETCHECK,0,0);
 SetDlgItemText(hwnd,DLG_GROUND_SHADER_NAME,(LPCTSTR)"None");
 SendDlgItemMessage(hwnd,DLG_GROUND_MAP_SINGLE,BM_SETCHECK,0,0);
 SendDlgItemMessage(hwnd,DLG_GROUND_MAP_TILE,  BM_SETCHECK,0,0);
 SendDlgItemMessage(hwnd,DLG_GROUND_MAP_MOZAIC,BM_SETCHECK,0,0);
 SetDlgItemText(hwnd,DLG_GROUND_MAP_NAME,(LPCTSTR)"None");
 SetDlgItemText(hwnd,DLG_GROUND_MAP_PATH,(LPCTSTR)" ");
 if(gt >= 128){
   gt -= 128;
   SendDlgItemMessage(hwnd,DLG_GROUND_SHADER,BM_SETCHECK,1,0);
   SetDlgItemText(hwnd,DLG_GROUND_SHADER_NAME,(LPCTSTR)GroundTextList[gt]);
   enable_shader=TRUE;
 }
 else if(gt == BRUM || gt == TBRUM || gt == MBRUM){
   SendDlgItemMessage(hwnd,DLG_GROUND_MAPPED,BM_SETCHECK,1,0);
   if     (gt ==  BRUM)SendDlgItemMessage(hwnd,DLG_GROUND_MAP_SINGLE,BM_SETCHECK,1,0);
   else if(gt == TBRUM)SendDlgItemMessage(hwnd,DLG_GROUND_MAP_TILE,  BM_SETCHECK,1,0);
   else if(gt == MBRUM)SendDlgItemMessage(hwnd,DLG_GROUND_MAP_MOZAIC,BM_SETCHECK,1,0);
   SetDlgItemText(hwnd,DLG_GROUND_MAP_NAME,(LPCTSTR)short_form(op->name));
   SetDlgItemText(hwnd,DLG_GROUND_MAP_PATH,(LPCTSTR)truncate_path(op->name,40));
   enable=TRUE;
 }
 else {
   SendDlgItemMessage(hwnd,DLG_GROUND_PLAIN,BM_SETCHECK,1,0);
 }
 EnableWindow(GetDlgItem(hwnd,DLG_GROUND_MAP_SINGLE),enable);
 EnableWindow(GetDlgItem(hwnd,DLG_GROUND_MAP_TILE),enable);
 EnableWindow(GetDlgItem(hwnd,DLG_GROUND_MAP_MOZAIC),enable);
 EnableWindow(GetDlgItem(hwnd,DLG_GROUND_CHOOSE),enable);
 EnableWindow(GetDlgItem(hwnd,DLG_GROUND_CHOOSE_SHA),enable_shader);
 EnableWindow(GetDlgItem(hwnd,DLG_GROUND_VIEW),enable);
 EnableWindow(GetDlgItem(hwnd,DLG_GROUND_EDITAPP),enable);
}

static BOOL CALLBACK GroundDlgProc(HWND hwnd,UINT msg,
                                   WPARAM wparam,LPARAM lparam){
 static HBITMAP GndBtnBitmaps[3]={NULL,NULL,NULL};
 static object *op;
 static BOOL noaction;
 static short l_limit,u_limit;
 char str[32],tt[128];
 BOOL err;
 int i;
 LPDRAWITEMSTRUCT lpdis;
 switch( msg ) {
   case WM_INITDIALOG:{
       op=(object *)lparam;
       noaction=TRUE;
       GndBtnBitmaps[0]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_GROUND1));
       GndBtnBitmaps[1]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_GROUND2));
       GndBtnBitmaps[2]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_GROUND3));
       SendMessage(GetDlgItem(hwnd,DLG_GROUND_MAP_SINGLE),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)GndBtnBitmaps[0]);
       SendMessage(GetDlgItem(hwnd,DLG_GROUND_MAP_TILE),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)GndBtnBitmaps[1]);
       SendMessage(GetDlgItem(hwnd,DLG_GROUND_MAP_MOZAIC),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)GndBtnBitmaps[2]);
       UpdateGroundDialog(hwnd,op);
       noaction=FALSE;
       SetDlgItemInt(hwnd,DLG_GROUND_FRAME_FROM,op->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_GROUND_FRAME_TO  ,op->lastframe, FALSE);
       if(op->last != NULL)l_limit=op->last->lastframe+1;
       else l_limit=1;
       if(op->next != NULL)u_limit=op->next->firstframe-1;
       else u_limit=Nframes;
       SendDlgItemMessage(hwnd,DLG_GROUND_FRAME_SPINF,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_GROUND_FRAME_SPINF,SPNM_SETCRNTVALUE,
                          (WPARAM)op->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_GROUND_FRAME_SPINF,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_GROUND_FRAME_FROM));
       SendDlgItemMessage(hwnd,DLG_GROUND_FRAME_SPINT,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_GROUND_FRAME_SPINT,SPNM_SETCRNTVALUE,
                          (WPARAM)op->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_GROUND_FRAME_SPINT,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_GROUND_FRAME_TO));
       sprintf(str,"%.2lf",op->grefl);
       SetDlgItemText(hwnd,DLG_GROUND_REFL,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_GROUND_REFL,EM_LIMITTEXT,(WPARAM)12,0);
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_DESTROY:
     for(i=0;i<3;i++){
       if(GndBtnBitmaps[i] != NULL){
         if(!DeleteObject(GndBtnBitmaps[i]))MessageBox(NULL,"Bad Delete",NULL,MB_OK);
         GndBtnBitmaps[i]=NULL;
       }
     }
     break;
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_GROUND_GC || lpdis->CtlID == DLG_GROUND_SC){
         if(lpdis->CtlID == DLG_GROUND_SC){
           r=op->acolour[0]; g=op->acolour[1]; b=op->acolour[2];
         }
         else {
           r=op->colour[0]; g=op->colour[1]; b=op->colour[2];
         }
         if(lpdis->itemState & ODS_SELECTED)InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_GROUND_CHOOSE_SHA:
          EnableWindow(hwnd,FALSE);
          i=SelectScrolledItemList((int)NoGroundTexts,GroundTextList,
                                   "Texture",FALSE,hwnd);
          EnableWindow(hwnd,TRUE);
          if(i < 0)break;
          if(i > 0)op->groundtype=128+i;
          else     op->groundtype=PLAIN;
          noaction=TRUE;
          UpdateGroundDialog(hwnd,op);
          noaction=FALSE;
          break;
        case DLG_GROUND_CHOOSE:
          if(op->groundtype == BRUM  ||
             op->groundtype == TBRUM ||
             op->groundtype == MBRUM){
            EnableWindow(hwnd,FALSE);
            if(SelectSfxFileName(0,gszMAPfile,gszMAPdir,IDX_MISC_SELIMAGEMAP,
              "All formats|*.bmp;*.gif;*.tga;*.dib;*.jpg;*.tif;*.png|"
              "(*.GIF)|*.gif|(*.TGA)|*.tga|(*.TIF)|*.tif|(*.JPG)|*.jpg|(*.PNG)|*.png|",hwnd) == TRUE){
              strcpy(op->name,gszMAPfile);
            }
            SetActiveWindow(hwnd);
            EnableWindow(hwnd,TRUE);
            noaction=TRUE;
            UpdateGroundDialog(hwnd,op);
            noaction=FALSE;
          }
          break;
        case DLG_GROUND_VIEW:
          if(op->groundtype == BRUM  ||
             op->groundtype == TBRUM ||
             op->groundtype == MBRUM)DisplayImageFile(op->name);
          break;
        case DLG_GROUND_EDITAPP:
          if(op->groundtype == BRUM  ||
             op->groundtype == TBRUM ||
             op->groundtype == MBRUM)EditImageFile(op->name);
          break;
        case DLG_GROUND_MAP_SINGLE:
          if(noaction)break;
          if(op->groundtype == PLAIN || op->groundtype >= 128)break;
          op->groundtype=BRUM;
          noaction=TRUE;
          UpdateGroundDialog(hwnd,op);
          noaction=FALSE;
          break;
        case DLG_GROUND_MAP_TILE:
          if(noaction)break;
          if(op->groundtype == PLAIN || op->groundtype >= 128)break;
          op->groundtype=TBRUM;
          noaction=TRUE;
          UpdateGroundDialog(hwnd,op);
          noaction=FALSE;
          break;
        case DLG_GROUND_MAP_MOZAIC:
          if(noaction)break;
          if(op->groundtype == PLAIN || op->groundtype >= 128)break;
          op->groundtype=MBRUM;
          noaction=TRUE;
          UpdateGroundDialog(hwnd,op);
          noaction=FALSE;
          break;
        case DLG_GROUND_PLAIN:
          if(noaction)break;
          op->groundtype=PLAIN;
          noaction=TRUE;
          UpdateGroundDialog(hwnd,op);
          noaction=FALSE;
          break;
        case DLG_GROUND_SHADER:
          if(noaction)break;
          if(HIWORD(wparam) == BN_CLICKED){
            EnableWindow(hwnd,FALSE);
            i=SelectScrolledItemList((int)NoGroundTexts,GroundTextList,
                                     "Texture",FALSE,hwnd);
            EnableWindow(hwnd,TRUE);
            if(i < 0)break;
            if(i > 0)op->groundtype=128+i;
            else     op->groundtype=PLAIN;
            noaction=TRUE;
            UpdateGroundDialog(hwnd,op);
            noaction=FALSE;
          }
          break;
        case DLG_GROUND_MAPPED:
          if(noaction)break;
          EnableWindow(hwnd,FALSE);
          if(SelectSfxFileName(0,gszMAPfile,gszMAPdir,IDX_MISC_SELIMAGEMAP,
            "All formats|*.bmp;*.gif;*.tga;*.dib;*.jpg;*.tif;*.png|"
            "(*.GIF)|*.gif|(*.TGA)|*.tga|(*.TIF)|*.tif|(*.JPG)|*.jpg|(*.PNG)|*.png|",hwnd) == TRUE){
            strcpy(op->name,gszMAPfile);
            op->groundtype=BRUM;
          }
          SetActiveWindow(hwnd);
          EnableWindow(hwnd,TRUE);
          noaction=TRUE;
          UpdateGroundDialog(hwnd,op);
          noaction=FALSE;
          break;
        case DLG_GROUND_GC:
          SetSfxColour(op->colour,IDX_MISC_GROUND_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_GROUND_GC),NULL,FALSE);
          UpdateGroundDialog(hwnd,op);
          break;
        case DLG_GROUND_SC:
          SetSfxColour(op->acolour,IDX_MISC_GROUND_SHADER,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_GROUND_SC),NULL,FALSE);
          UpdateGroundDialog(hwnd,op);
          break;
        case IDCANCEL:
        case DLG_GROUND_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_GROUND_OK:
          if(GetDlgItemText(hwnd,DLG_GROUND_REFL,str,10) == 0){
            EndDialog(hwnd,FAIL);
            return TRUE;
          }
          op->grefl=atof(str);
          if(op->grefl < 0.0 || op->grefl > 100.0){
            EndDialog(hwnd,FAIL);
            return TRUE;
          }
          i=GetDlgItemInt(hwnd,DLG_GROUND_FRAME_FROM,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          op->firstframe = (short)i;
          i=GetDlgItemInt(hwnd,DLG_GROUND_FRAME_TO,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          op->lastframe = (short)i;
          if(op->firstframe > op->lastframe)goto CANCELX;
          EndDialog(hwnd,OK);
          return(TRUE);
          CANCELX:
          EndDialog(hwnd,FAIL);
          return TRUE;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short EditGroundDialog(object *op, short in_script){
 int i;
 object actor;
 HWND hwnd;
 memcpy(&actor,op,sizeof(object));
 EnableToolPannels(ALL_PANNELS,FALSE);
 if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_GROUND),hwnd,
            (DLGPROC)GroundDlgProc,(LPARAM)&actor)) == OK){
   EDIT_ACTION=YES;
   memcpy(op,&actor,sizeof(object));
 }
 EnableToolPannels(ALL_PANNELS,TRUE);
 return (short)i;
}

static void UpdateSkyDialog(HWND hwnd, sky *sp){
 BOOL enable=FALSE;
 int show=SW_HIDE;
 int st;
 st=(int)sp->type;
 if(st == SKYMAPPED   ||
    st == SKYANIMATED ||
    st == BACKDROP)enable=TRUE;
 if(st == SKYANIMATED)show=SW_SHOW;
 EnableWindow(GetDlgItem(hwnd,DLG_SKY_CHOOSE),enable);
 EnableWindow(GetDlgItem(hwnd,DLG_SKY_VIEW),enable);
 EnableWindow(GetDlgItem(hwnd,DLG_SKY_EDITAPP),enable);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_START),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_START_S),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_START_T),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_END),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_END_S),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_END_T),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_STEP),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_STEP_S),show);
 ShowWindow(GetDlgItem(hwnd,DLG_SKY_ANIM_STEP_T),show);
 return;
}

static BOOL CALLBACK SkyDlgProc(HWND hwnd,UINT msg,
                                WPARAM wparam,LPARAM lparam){
 char str[32],nomaptext[]="No image map in use";
 char tt[128];
 static sky *sp;
 BOOL err;
 int i;
 LPDRAWITEMSTRUCT lpdis;
 static short l_limit,u_limit;
 switch( msg ) {
   case WM_INITDIALOG:{
       int dlg_list[5]={DLG_SKY_PLAIN,DLG_SKY_GRADED,
                        DLG_SKY_ENVIRON,DLG_SKY_BACKDROP,DLG_SKY_ANIMATED};
       sp=(sky *)lparam;
       SendDlgItemMessage(hwnd,dlg_list[sp->type],BM_SETCHECK,1,0);
       if(sp->type == SKYCOLOURED || sp->type == SKYGRADED){
         SetDlgItemText(hwnd,DLG_SKY_MAP_NAME,(LPCTSTR)nomaptext);
         SetDlgItemText(hwnd,DLG_SKY_MAP_PATH,(LPCTSTR)" ");
       }
       else{
         SetDlgItemText(hwnd,DLG_SKY_MAP_NAME,(LPCTSTR)short_form(sp->name));
         SetDlgItemText(hwnd,DLG_SKY_MAP_PATH,(LPCTSTR)truncate_path(sp->name,40));
       }
       SetDlgItemInt(hwnd,DLG_SKY_FRAME_FROM,sp->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_SKY_FRAME_TO  ,sp->lastframe, FALSE);
       if(sp->last != NULL)l_limit=sp->last->lastframe+1;
       else l_limit=1;
       if(sp->next != NULL)u_limit=sp->next->firstframe-1;
       else u_limit=Nframes;
       SendDlgItemMessage(hwnd,DLG_SKY_FRAME_SPINF,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SKY_FRAME_SPINF,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_SKY_FRAME_SPINF,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SKY_FRAME_FROM));
       SendDlgItemMessage(hwnd,DLG_SKY_FRAME_SPINT,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SKY_FRAME_SPINT,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_SKY_FRAME_SPINT,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SKY_FRAME_TO));
       sprintf(str,"%.2lf",sp->gamb*100.0);
       SetDlgItemText(hwnd,DLG_SKY_AMB_GROUND,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SKY_AMB_GROUND,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(str,"%.2lf",sp->oamb*100.0);
       SetDlgItemText(hwnd,DLG_SKY_AMB_MODEL,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SKY_AMB_MODEL,EM_LIMITTEXT,(WPARAM)12,0);
       SetDlgItemInt(hwnd,DLG_SKY_ANIM_START,sp->aff,FALSE);
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_START_S,SPNM_SETRANGE,0,
                          MAKELPARAM(1,9999));
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_START_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->aff,0);
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_START_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SKY_ANIM_START));
       SetDlgItemInt(hwnd,DLG_SKY_ANIM_END,sp->alf,FALSE);
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_END_S,SPNM_SETRANGE,0,
                          MAKELPARAM(1,9999));
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_END_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->alf,0);
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_END_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SKY_ANIM_END));
       SetDlgItemInt(hwnd,DLG_SKY_ANIM_STEP,sp->ast,FALSE);
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_STEP_S,SPNM_SETRANGE,0,
                          MAKELPARAM(1,9999));
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_STEP_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->ast,0);
       SendDlgItemMessage(hwnd,DLG_SKY_ANIM_STEP_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SKY_ANIM_STEP));
       UpdateSkyDialog(hwnd,sp);
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       HPEN hOldPen;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_SKY_ZC || lpdis->CtlID == DLG_SKY_HC){
         if(lpdis->CtlID == DLG_SKY_ZC){
           r=sp->zcolour[0]; g=sp->zcolour[1]; b=sp->zcolour[2];
         }
         else {
           r=sp->colour[0]; g=sp->colour[1]; b=sp->colour[2];
         }
         if(lpdis->itemState & ODS_SELECTED)InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_SKY_CHOOSE:
          if(sp->type == SKYMAPPED   ||
             sp->type == SKYANIMATED ||
             sp->type == BACKDROP){
            EnableWindow(hwnd,FALSE);
            if(sp->type == SKYANIMATED)
              LoadString(ghinst_main,IDX_MISC_SELANIMAP,tt,128);
            else
              LoadString(ghinst_main,IDX_MISC_SELIMAGEMAP,tt,128);
            if(SelectFileName(0,gszMAPfile,gszMAPdir,tt,
              "All file formats|*.bmp;*.gif;*.tga;*.dib;*.jpg;*.tif;*.png|"
              "(*.GIF)|*.gif|(*.TGA)|*.tga|(*.TIF)|*.tif|(*.JPG)|*.jpg|(*.PNG)|*.png|",hwnd) == TRUE){
              strcpy(sp->name,gszMAPfile);
              SetDlgItemText(hwnd,DLG_SKY_MAP_NAME,(LPCTSTR)short_form(sp->name));
              SetDlgItemText(hwnd,DLG_SKY_MAP_PATH,(LPCTSTR)truncate_path(sp->name,40));
            }
            SetActiveWindow(hwnd);
            EnableWindow(hwnd,TRUE);
          }
          break;
        case DLG_SKY_VIEW:
          if(sp->type == SKYMAPPED   ||
             sp->type == SKYANIMATED ||
             sp->type == BACKDROP)DisplayImageFile(sp->name);
          break;
        case DLG_SKY_EDITAPP:
          if(sp->type == SKYMAPPED   ||
             sp->type == SKYANIMATED ||
             sp->type == BACKDROP)EditImageFile(sp->name);
          break;
        case DLG_SKY_PLAIN:
          sp->type=SKYCOLOURED;
          SetDlgItemText(hwnd,DLG_SKY_MAP_NAME,nomaptext);
          SetDlgItemText(hwnd,DLG_SKY_MAP_PATH,(LPCTSTR)" ");
          UpdateSkyDialog(hwnd,sp);
          break;
        case DLG_SKY_GRADED:
          sp->type=SKYGRADED;
          SetDlgItemText(hwnd,DLG_SKY_MAP_NAME,nomaptext);
          SetDlgItemText(hwnd,DLG_SKY_MAP_PATH,(LPCTSTR)" ");
          UpdateSkyDialog(hwnd,sp);
          break;
        case DLG_SKY_ENVIRON:
        case DLG_SKY_BACKDROP:
        case DLG_SKY_ANIMATED:
          EnableWindow(hwnd,FALSE);
          if(LOWORD(wparam) == DLG_SKY_ANIMATED)
            LoadString(ghinst_main,IDX_MISC_SELANIMAP,tt,128);
          else
            LoadString(ghinst_main,IDX_MISC_SELIMAGEMAP,tt,128);
          if(SelectFileName(0,gszMAPfile,gszMAPdir,tt,
            "All file formats|*.bmp;*.gif;*.tga;*.dib;*.jpg;*.tif;*.png|"
            "(*.GIF)|*.gif|(*.TGA)|*.tga|(*.TIF)|*.tif|(*.JPG)|*.jpg|(*.PNG)|*.png|",hwnd) == TRUE){
            strcpy(sp->name,gszMAPfile);
            if(LOWORD(wparam) == DLG_SKY_ENVIRON) sp->type=SKYMAPPED;
            if(LOWORD(wparam) == DLG_SKY_BACKDROP)sp->type=BACKDROP;
            if(LOWORD(wparam) == DLG_SKY_ANIMATED)sp->type=SKYANIMATED;
            SetDlgItemText(hwnd,DLG_SKY_MAP_NAME,(LPCTSTR)short_form(sp->name));
            SetDlgItemText(hwnd,DLG_SKY_MAP_PATH,(LPCTSTR)truncate_path(sp->name,40));
          }
          SetActiveWindow(hwnd);
          EnableWindow(hwnd,TRUE);
          UpdateSkyDialog(hwnd,sp);
          break;
        case DLG_SKY_HC:
          SetSfxColour(sp->colour,IDX_MISC_HORIZON_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_SKY_HC),NULL,FALSE);
          break;
        case DLG_SKY_ZC:
          SetSfxColour(sp->zcolour,IDX_MISC_ZENITH_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_SKY_ZC),NULL,FALSE);
          break;
        case IDCANCEL:
        case DLG_SKY_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_SKY_OK:
          if(GetDlgItemText(hwnd,DLG_SKY_AMB_GROUND,str,10) == 0){
            EndDialog(hwnd,FAIL);
            return TRUE;
          }
          sp->gamb=atof(str)/100.0;
          if(sp->gamb < 0.0 || sp->gamb > 1.0){
            EndDialog(hwnd,FAIL);
            return TRUE;
          }
          if(GetDlgItemText(hwnd,DLG_SKY_AMB_MODEL,str,10) == 0){
            EndDialog(hwnd,FAIL);
            return TRUE;
          }
          sp->oamb=atof(str)/100.0;
          if(sp->oamb < 0.0 || sp->oamb > 1.0){
            EndDialog(hwnd,FAIL);
            return TRUE;
          }
          i=GetDlgItemInt(hwnd,DLG_SKY_FRAME_FROM,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          sp->firstframe = (short)i;
          i=GetDlgItemInt(hwnd,DLG_SKY_FRAME_TO,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          sp->lastframe = (short)i;
          if(sp->firstframe > sp->lastframe)goto CANCELX;
          sp->aff=GetDlgItemInt(hwnd,DLG_SKY_ANIM_START,&err,FALSE);
          sp->alf=GetDlgItemInt(hwnd,DLG_SKY_ANIM_END,&err,FALSE);
          sp->ast=GetDlgItemInt(hwnd,DLG_SKY_ANIM_STEP,&err,FALSE);
          EndDialog(hwnd,OK);
          return(TRUE);
          CANCELX:
          EndDialog(hwnd,FAIL);
          return TRUE;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short EditSkyDialog(sky *sp, int in_script){
 int i;
 sky Sky;
 HWND hwnd;
 memcpy(&Sky,sp,sizeof(sky));
 EnableToolPannels(ALL_PANNELS,FALSE);
// if((hwnd=GetFocus()) != ghwnd_main)hwnd=ghwndTimeline;
 if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_SKY),hwnd,
            (DLGPROC)SkyDlgProc,(LPARAM)&Sky)) == OK){
   EDIT_ACTION=YES;
   memcpy(sp,&Sky,sizeof(sky));
 }
 EnableToolPannels(ALL_PANNELS,TRUE);
 return (short)i;
}

static BOOL CALLBACK DirectorDlgProc(HWND hwnd,UINT msg,
                                WPARAM wparam,LPARAM lparam){
// char tt[128];
 node *Np;
 BOOL err;
 int i;
 static director *dp;
 static short l_limit,u_limit;
 switch( msg ) {
   case WM_INITDIALOG:{
       dp=(director *)lparam;
       SetDlgItemInt(hwnd,DLG_DIRECTOR_FRAME_FROM,dp->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_DIRECTOR_FRAME_TO  ,dp->lastframe, FALSE);
       if(dp->last != NULL)l_limit=dp->last->lastframe+1;
       else l_limit=1;
       if(dp->next != NULL)u_limit=dp->next->firstframe-1;
       else u_limit=Nframes;
       SendDlgItemMessage(hwnd,DLG_DIRECTOR_FRAME_SPINF,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_DIRECTOR_FRAME_SPINF,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_DIRECTOR_FRAME_SPINF,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_DIRECTOR_FRAME_FROM));
       SendDlgItemMessage(hwnd,DLG_DIRECTOR_FRAME_SPINT,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_DIRECTOR_FRAME_SPINT,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_DIRECTOR_FRAME_SPINT,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_DIRECTOR_FRAME_TO));
       if((Np=dp->ActiveCamera) != NULL){
         SetDlgItemText(hwnd,DLG_DIRECTOR_TARGET,Np->actorname);
       }
       Np=FirstNp; while(Np != NULL){
         if(Np->type == CAMERA){
           i=SendDlgItemMessage(hwnd,DLG_DIRECTOR_SELECT,LB_ADDSTRING,0,
                              (LPARAM)Np->actorname);
           SendDlgItemMessage(hwnd,DLG_DIRECTOR_SELECT,LB_SETITEMDATA,
                              i,(LPARAM)Np);
         }
         Np=Np->next;
       }
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
          i=GetDlgItemInt(hwnd,DLG_DIRECTOR_FRAME_FROM,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          dp->firstframe = (short)i;
          i=GetDlgItemInt(hwnd,DLG_DIRECTOR_FRAME_TO,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          dp->lastframe = (short)i;
          if(dp->firstframe > dp->lastframe)goto CANCELX;
          i=SendDlgItemMessage(hwnd,DLG_DIRECTOR_SELECT,LB_GETCURSEL,0,0);
          if(i != LB_ERR){
            i=(long)SendDlgItemMessage(hwnd,DLG_DIRECTOR_SELECT,
                       LB_GETITEMDATA,i,(LPARAM)0);
            if(i != LB_ERR){
              Np=(node *)i;
              dp->ActiveCamera=Np;
            }
          }
          EndDialog(hwnd,OK);
          return(TRUE);
          CANCELX:
          EndDialog(hwnd,FAIL);
          return TRUE;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short EditDirectorDialog(director *dp, int in_script){
 int i;
 director Director;
 HWND hwnd;
 memcpy(&Director,dp,sizeof(director));
 EnableToolPannels(ALL_PANNELS,FALSE);
 if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_DIRECTOR),hwnd,
            (DLGPROC)DirectorDlgProc,(LPARAM)&Director)) == OK){
   EDIT_ACTION=YES;
   memcpy(dp,&Director,sizeof(director));
 }
 EnableToolPannels(ALL_PANNELS,TRUE);
 return (short)i;
}

static BOOL CALLBACK LightDlgProc(HWND hwnd,UINT msg,
                                WPARAM wparam,LPARAM lparam){
 object *op;
 BOOL err;
 int i;
 char str[32];
 LPDRAWITEMSTRUCT lpdis;
 static short l_limit,u_limit;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       int dlg_list[5]={DLG_LIGHT_NORMAL,DLG_LIGHT_NOSHADOW,
                        DLG_LIGHT_SHADOWONLY,DLG_LIGHT_SPOTLIGHT,
                        DLG_LIGHT_DUMMY};
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       op=(object *)lparam;
       if(op->last != NULL)l_limit=op->last->lastframe+1;
       else l_limit=1;
       if(op->next != NULL)u_limit=op->next->firstframe-1;
       else u_limit=Nframes;
       SendDlgItemMessage(hwnd,dlg_list[op->lighttype],BM_SETCHECK,1,0);
       SetDlgItemInt(hwnd,DLG_LIGHT_FRAME_FROM,op->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_LIGHT_FRAME_TO  ,op->lastframe, FALSE);
       SendDlgItemMessage(hwnd,DLG_LIGHT_FRAME_SPINF,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_LIGHT_FRAME_SPINF,SPNM_SETCRNTVALUE,
                          (WPARAM)op->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_LIGHT_FRAME_SPINF,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_LIGHT_FRAME_FROM));
       SendDlgItemMessage(hwnd,DLG_LIGHT_FRAME_SPINT,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_LIGHT_FRAME_SPINT,SPNM_SETCRNTVALUE,
                          (WPARAM)op->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_LIGHT_FRAME_SPINT,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_LIGHT_FRAME_TO));
       if(op->depth_cue)
         SendDlgItemMessage(hwnd,DLG_LIGHT_DEPTHCUE,BM_SETCHECK,TRUE,0);
       sprintf(str,"%.2lf",op->depth_cue_c);
       SetDlgItemText(hwnd,DLG_LIGHT_DEPTHCUE_C,str);
       sprintf(str,"%.2lf",op->depth_cue_l);
       SetDlgItemText(hwnd,DLG_LIGHT_DEPTHCUE_L,str);
       sprintf(str,"%.2lf",op->depth_cue_q);
       SetDlgItemText(hwnd,DLG_LIGHT_DEPTHCUE_Q,str);
#if __LITE__
       EnableWindow(GetDlgItem(hwnd,DLG_LIGHT_DEPTHCUE),FALSE);
       EnableWindow(GetDlgItem(hwnd,DLG_LIGHT_DEPTHCUE_C),FALSE);
       EnableWindow(GetDlgItem(hwnd,DLG_LIGHT_DEPTHCUE_L),FALSE);
       EnableWindow(GetDlgItem(hwnd,DLG_LIGHT_DEPTHCUE_Q),FALSE);
#endif
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_LIGHT_COLOUR){
         op=(object *)GetWindowLong(hwnd,GWL_USERDATA);
         r=op->colour[0]; g=op->colour[1]; b=op->colour[2];
         if(lpdis->itemState & ODS_SELECTED)InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_COMMAND:
      op=(object *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_LIGHT_NORMAL:
          op->lighttype=NORMAL;
          break;
        case DLG_LIGHT_NOSHADOW:
          op->lighttype=NOSHADOW;
          break;
        case DLG_LIGHT_SHADOWONLY:
          op->lighttype=SHADOWONLY;
          break;
        case DLG_LIGHT_SPOTLIGHT:
          op->lighttype=SPOTLIGHT;
          break;
        case DLG_LIGHT_DUMMY:
          op->lighttype=DUMMYL;
          break;
        case DLG_LIGHT_COLOUR:
          SetSfxColour(op->colour,IDX_MISC_LIGHT_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_LIGHT_COLOUR),NULL,FALSE);
          break;
        case IDCANCEL:
        case DLG_LIGHT_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_LIGHT_OK:
          i=GetDlgItemInt(hwnd,DLG_LIGHT_FRAME_FROM,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          op->firstframe = (short)i;
          i=GetDlgItemInt(hwnd,DLG_LIGHT_FRAME_TO,&err,FALSE);
          if(!err || i < (short)l_limit ||
                     i > (short)u_limit)goto CANCELX;
          op->lastframe = (short)i;
          if(op->firstframe > op->lastframe)goto CANCELX;
          if(SendDlgItemMessage(hwnd,DLG_LIGHT_DEPTHCUE,BM_GETCHECK,0,0))
               op->depth_cue=1;
          else op->depth_cue=0;
          if(GetDlgItemText(hwnd,DLG_LIGHT_DEPTHCUE_C,str,31) != 0)
            op->depth_cue_c=atof(str);
          if(GetDlgItemText(hwnd,DLG_LIGHT_DEPTHCUE_L,str,31) != 0)
            op->depth_cue_l=atof(str);
          if(GetDlgItemText(hwnd,DLG_LIGHT_DEPTHCUE_Q,str,31) != 0)
            op->depth_cue_q=atof(str);
          EndDialog(hwnd,OK);
          return(TRUE);
          CANCELX:
          EndDialog(hwnd,FAIL);
          return TRUE;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short EditLightDialog(object *op, int in_script){
 int i;
 object actor;
 HWND hwnd;
 memcpy(&actor,op,sizeof(object));
 EnableToolPannels(ALL_PANNELS,FALSE);
 if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_LIGHT),hwnd,
            (DLGPROC)LightDlgProc,(LPARAM)&actor)) == OK){
   EDIT_ACTION=YES;
   memcpy(op,&actor,sizeof(object));
 }
 EnableToolPannels(ALL_PANNELS,TRUE);
 return (short)i;
}

typedef struct tagSIZESTRUCT {
  size *sp;
  int  d;
} SIZESTRUCT;

static BOOL CALLBACK SizeDlgProc(HWND hwnd,UINT msg,
                                 WPARAM wparam,LPARAM lparam){
 char str[32];
 SIZESTRUCT *ss;
 size *sp;
 BOOL err;
 static double r,storeX=1.0,storeY=1.0,storeZ=1.0;
 static short l_limit,u_limit;
 int i;
 long range;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       ss=(SIZESTRUCT *)lparam;
       sp=ss->sp;
       SetWindowLong(hwnd,GWL_USERDATA,(LONG)sp);
       sprintf(str,"%.2lf",sp->Sx);
       SetDlgItemText(hwnd,DLG_SCALE_W,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALE_W,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(str,"%.2lf",sp->Sy);
       SetDlgItemText(hwnd,DLG_SCALE_L,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALE_L,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(str,"%.2lf",sp->Sz);
       SetDlgItemText(hwnd,DLG_SCALE_H,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALE_H,EM_LIMITTEXT,(WPARAM)12,0);
       if(sp->tweentype == 0)
         SendDlgItemMessage(hwnd,DLG_SCALE_BOUNCY_OFF,BM_SETCHECK,1,0);
       else
         SendDlgItemMessage(hwnd,DLG_SCALE_BOUNCY_ON,BM_SETCHECK,1,0);
       SetDlgItemInt(hwnd,DLG_SCALE_FROM,sp->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_SCALE_TO,  sp->lastframe,FALSE);
       if(sp->last != NULL)l_limit=sp->last->lastframe+1;
       else l_limit=1;
       if(sp->next != NULL)u_limit=sp->next->firstframe-1;
       else u_limit=Nframes;
       SendDlgItemMessage(hwnd,DLG_SCALE_FROM_S,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SCALE_FROM_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_SCALE_FROM_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SCALE_FROM));
       SendDlgItemMessage(hwnd,DLG_SCALE_TO_S,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SCALE_TO_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_SCALE_TO_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SCALE_TO));
       if(ss->d == 0){
         EnableWindow(GetDlgItem(hwnd,DLG_SCALE_DELETE),FALSE);
//         EnableWindow(GetDlgItem(hwnd,DLG_SCALE_CUT),FALSE);
       }
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      sp=(size *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_SCALE_RESET:
          sprintf(str,"%.2lf",1.0);
          SetDlgItemText(hwnd,DLG_SCALE_W,(LPCTSTR)str);
          SetDlgItemText(hwnd,DLG_SCALE_L,(LPCTSTR)str);
          SetDlgItemText(hwnd,DLG_SCALE_H,(LPCTSTR)str);
          break;
        case DLG_SCALE_SET:
          sprintf(str,"%.2lf",storeX);
          SetDlgItemText(hwnd,DLG_SCALE_W,(LPCTSTR)str);
          sprintf(str,"%.2lf",storeY);
          SetDlgItemText(hwnd,DLG_SCALE_L,(LPCTSTR)str);
          sprintf(str,"%.2lf",storeZ);
          SetDlgItemText(hwnd,DLG_SCALE_H,(LPCTSTR)str);
          break;
        case DLG_SCALE_STORE:
          if(GetDlgItemText(hwnd,DLG_SCALE_W,str,10) != 0)
            if((r=atof(str)) != 0.0)storeX=r;
          if(GetDlgItemText(hwnd,DLG_SCALE_L,str,10) != 0)
            if((r=atof(str)) != 0.0)storeY=r;
          if(GetDlgItemText(hwnd,DLG_SCALE_H,str,10) != 0)
            if((r=atof(str)) != 0.0)storeZ=r;
          break;
        case DLG_SCALE_COPY:
          if(sp->last != NULL){
            sprintf(str,"%.2lf",sp->Sx);
            SetDlgItemText(hwnd,DLG_SCALE_W,(LPCTSTR)str);
            sprintf(str,"%.2lf",sp->Sy);
            SetDlgItemText(hwnd,DLG_SCALE_L,(LPCTSTR)str);
            sprintf(str,"%.2lf",sp->Sz);
            SetDlgItemText(hwnd,DLG_SCALE_H,(LPCTSTR)str);
          }
          break;
        case DLG_SCALE_DELETE:
          EndDialog(hwnd,2);
          return(TRUE);
//        case DLG_SCALE_CUT:
//          EndDialog(hwnd,FAIL);
//          return(TRUE);
        case IDCANCEL:
        case DLG_SCALE_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_SCALE_OK:
          if(SendDlgItemMessage(hwnd,DLG_SCALE_BOUNCY_ON,BM_GETCHECK,0,0))
             sp->tweentype=1;
          else sp->tweentype=0;
          range=SendDlgItemMessage(hwnd,DLG_SCALE_FROM_S,SPNM_GETRANGE,0,0);
          i=GetDlgItemInt(hwnd,DLG_SCALE_FROM,&err,FALSE);
          if(!err || i < LOWORD(range) || i > HIWORD(range))goto CANCELX;
          if(err)sp->firstframe=(short)i;
          range=SendDlgItemMessage(hwnd,DLG_SCALE_TO_S,SPNM_GETRANGE,0,0);
          i=GetDlgItemInt(hwnd,DLG_SCALE_TO,&err,FALSE);
          if(!err || i < LOWORD(range) || i > HIWORD(range))goto CANCELX;
          if(err)sp->lastframe=(short)i;
          if(GetDlgItemText(hwnd,DLG_SCALE_W,str,10) == 0)goto CANCELX;
          if((sp->Sx=atof(str)) == 0.0)goto CANCELX;
          if(GetDlgItemText(hwnd,DLG_SCALE_L,str,10) == 0)goto CANCELX;
          if((sp->Sy=atof(str)) == 0.0)goto CANCELX;
          if(GetDlgItemText(hwnd,DLG_SCALE_H,str,10) == 0)goto CANCELX;
          if((sp->Sz=atof(str)) == 0.0)goto CANCELX;
          EndDialog(hwnd,OK);
          return(TRUE);
          CANCELX:
          EndDialog(hwnd,FAIL);
          return TRUE;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short RequestEditSize(double *sx, double *sy, double *sz, size *xp, int d){
 int i;
 size Size;
 SIZESTRUCT SS;
 HWND original;
 memcpy(&Size,xp,sizeof(size));
 original=GetActiveWindow();
 EnableToolPannels(ALL_PANNELS,FALSE);
 SS.sp=&Size; SS.d=d;
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_SCALE),ghwnd_main,
            (DLGPROC)SizeDlgProc,(LPARAM)&SS)) == OK){
   EDIT_ACTION=YES;
   if(Size.lastframe >= Size.firstframe){
     memcpy(xp,&Size,sizeof(size));
     *sx=Size.Sx;
     *sy=Size.Sy;
     *sz=Size.Sz;
   }
 }
 EnableToolPannels(ALL_PANNELS,TRUE);
 if(original != NULL)SetActiveWindow(original);
 return (short)i;
}

static BOOL CALLBACK SizeL_DlgProc(HWND hwnd,UINT msg,
                                 WPARAM wparam,LPARAM lparam){
 static short l_limit,u_limit;
 char str[32];
 size *sp;
 SIZESTRUCT *ss;
 BOOL err;
 static double r,storeX=1.0,storeY=1.0;
 int i;
 long range;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       ss=(SIZESTRUCT *)lparam;
       sp=ss->sp;
       SetWindowLong(hwnd,GWL_USERDATA,(LONG)sp);
       sprintf(str,"%.2lf",sp->Sx);
       SetDlgItemText(hwnd,DLG_SCALEL_CONE,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALEL_CONE,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(str,"%.2lf",sp->Sy);
       SetDlgItemText(hwnd,DLG_SCALEL_EDGE,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALEL_EDGE,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(str,"%.2lf",sp->Sz);
       SetDlgItemText(hwnd,DLG_SCALEL_RANGE,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALEL_RANGE,EM_LIMITTEXT,(WPARAM)12,0);
       SetDlgItemInt(hwnd,DLG_SCALEL_FROM,sp->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_SCALEL_TO,  sp->lastframe,FALSE);
       if(sp->last != NULL)l_limit=sp->last->lastframe+1;
       else l_limit=1;
       if(sp->next != NULL)u_limit=sp->next->firstframe-1;
       else u_limit=Nframes;
       SendDlgItemMessage(hwnd,DLG_SCALEL_FROM_S,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SCALEL_FROM_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_SCALEL_FROM_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SCALEL_FROM));
       SendDlgItemMessage(hwnd,DLG_SCALEL_TO_S,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SCALEL_TO_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_SCALEL_TO_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SCALEL_TO));
       if(ss->d == 0){
         EnableWindow(GetDlgItem(hwnd,DLG_SCALEL_DELETE),FALSE);
       }
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      sp=(size *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_SCALEL_DELETE:
          EndDialog(hwnd,2);
          return(TRUE);
        case IDCANCEL:
        case DLG_SCALEL_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_SCALEL_OK:
          range=SendDlgItemMessage(hwnd,DLG_SCALEL_FROM_S,SPNM_GETRANGE,0,0);
          i=GetDlgItemInt(hwnd,DLG_SCALEL_FROM,&err,FALSE);
          if(!err || i < LOWORD(range) || i > HIWORD(range))goto CANCELX;
          if(err)sp->firstframe=(short)i;
          range=SendDlgItemMessage(hwnd,DLG_SCALEL_TO_S,SPNM_GETRANGE,0,0);
          i=GetDlgItemInt(hwnd,DLG_SCALEL_TO,&err,FALSE);
          if(!err || i < LOWORD(range) || i > HIWORD(range))goto CANCELX;
          if(err)sp->lastframe=(short)i;
          if(GetDlgItemText(hwnd,DLG_SCALEL_CONE,str,10) == 0)goto CANCELX;
          if((sp->Sx=atof(str)) < 1.0)goto CANCELX;
          if(sp->Sx > 179.0)goto CANCELX;
          if(GetDlgItemText(hwnd,DLG_SCALEL_EDGE,str,10) == 0)goto CANCELX;
          if((sp->Sy=atof(str)) < 1.0)goto CANCELX;
          if(sp->Sy > 179.0)goto CANCELX;
          if(GetDlgItemText(hwnd,DLG_SCALEL_RANGE,str,10) == 0)goto CANCELX;
          sp->Sz=atof(str);
          EndDialog(hwnd,OK);
          return(TRUE);
          CANCELX:
          EndDialog(hwnd,FAIL);
          return TRUE;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short RequestEditLightSize(double *sx, double *sy, double *sz, size *xp, int d){
 int i;
 size Size;
 SIZESTRUCT SS;
 HWND original;
 memcpy(&Size,xp,sizeof(size));
 SS.sp=&Size; SS.d=d;
 original=GetActiveWindow();
 EnableToolPannels(ALL_PANNELS,FALSE);
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_SCALEL),ghwnd_main,
            (DLGPROC)SizeL_DlgProc,(LPARAM)&SS)) == OK){
   EDIT_ACTION=YES;
   if(Size.lastframe >= Size.firstframe){
     memcpy(xp,&Size,sizeof(size));
     *sx=Size.Sx;
     *sy=Size.Sy;
     *sz=Size.Sz;
   }
 }
 EnableToolPannels(ALL_PANNELS,TRUE);
 if(original != NULL)SetActiveWindow(original);
 return (short)i;
}

static BOOL CALLBACK SizeC_DlgProc(HWND hwnd,UINT msg,
                                 WPARAM wparam,LPARAM lparam){
 static short l_limit,u_limit;
 char str[32];
 static size *sp;
 SIZESTRUCT *ss;
 BOOL err;
 int i,ilens;
 double lens,fov,ar;
 long range;
 char *cameralenslist[]={" 15 mm"," 25 mm"," 50 mm","100 mm","200 mm"};
 int cameralens[]={15,25,50,100,200};
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       ss=(SIZESTRUCT *)lparam;
       sp=ss->sp;
//       SetWindowLong(hwnd,GWL_USERDATA,(LONG)sp);
       if( (lens=50.0 * sp->Sx) - floor(lens) > 0.5)lens += 1.0;
       ilens=(int)lens;       // (5 999)
       fov=180/PI*2*atan2(21.22,(double)ilens);  //(5.0 160.0)
       ar = sp->Sy / sp->Sx; // (0.1,5.0);
       for(i=0;i<5;i++)SendDlgItemMessage(hwnd,DLG_SCALEC_PRESET,
         CB_ADDSTRING,0,(LPARAM)cameralenslist[i]);
       SendDlgItemMessage(hwnd,DLG_SCALEC_PRESET,CB_SELECTSTRING,
         0,(LPARAM)cameralenslist[2]);
       SetDlgItemInt(hwnd,DLG_SCALEC_LENS,ilens,FALSE);
       SendDlgItemMessage(hwnd,DLG_SCALEC_LENS_S,SPNM_SETRANGE,0,
                          MAKELPARAM(5,999));
       SendDlgItemMessage(hwnd,DLG_SCALEC_LENS_S,SPNM_SETCRNTVALUE,
                          (WPARAM)ilens,0);
       SendDlgItemMessage(hwnd,DLG_SCALEC_LENS_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SCALEC_LENS));
       sprintf(str,"%.2lf",fov);
       SetDlgItemText(hwnd,DLG_SCALEC_FOV,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALEC_FOV,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(str,"%.2lf",ar);
       SetDlgItemText(hwnd,DLG_SCALEC_AR,(LPCTSTR)str);
       SendDlgItemMessage(hwnd,DLG_SCALEC_AR,EM_LIMITTEXT,(WPARAM)12,0);
       SetDlgItemInt(hwnd,DLG_SCALEC_FROM,sp->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_SCALEC_TO,  sp->lastframe,FALSE);
       if(sp->last != NULL)l_limit=sp->last->lastframe+1;
       else l_limit=1;
       if(sp->next != NULL)u_limit=sp->next->firstframe-1;
       else u_limit=Nframes;
       SendDlgItemMessage(hwnd,DLG_SCALEC_FROM_S,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SCALEC_FROM_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_SCALEC_FROM_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SCALEC_FROM));
       SendDlgItemMessage(hwnd,DLG_SCALEC_TO_S,SPNM_SETRANGE,0,
                          MAKELPARAM(l_limit,u_limit));
       SendDlgItemMessage(hwnd,DLG_SCALEC_TO_S,SPNM_SETCRNTVALUE,
                          (WPARAM)sp->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_SCALEC_TO_S,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SCALEC_TO));
       if(ss->d == 0){
         EnableWindow(GetDlgItem(hwnd,DLG_SCALEC_DELETE),FALSE);
       }
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
//      sp=(size *)GetWindowLong(hwnd,GWL_USERDATA);
     switch(HIWORD(wparam)){
       case CBN_SELENDOK:
         if(LOWORD(wparam) == DLG_SCALEC_PRESET){
           SendDlgItemMessage(hwnd,DLG_SCALEC_PRESET,WM_GETTEXT,
             (WPARAM)12,(LPARAM)str);
           for(i=0;i<5;i++)if(strcmp(str,cameralenslist[i]) == 0)break;
           ilens=cameralens[i];
           SetDlgItemInt(hwnd,DLG_SCALEC_LENS,ilens,FALSE);
           SendDlgItemMessage(hwnd,DLG_SCALEC_LENS_S,SPNM_SETCRNTVALUE,
                          (WPARAM)ilens,0);
           fov=180/PI*2*atan2(21.22,(double)ilens);
           sprintf(str,"%.2lf",fov);
           SetDlgItemText(hwnd,DLG_SCALEC_FOV,(LPCTSTR)str);
         }
         break;
       default: break;
      }
      switch(LOWORD(wparam)){
        case DLG_SCALEC_DELETE:
          EndDialog(hwnd,2);
          return(TRUE);
        case IDCANCEL:
        case DLG_SCALEC_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_SCALEC_OK:
          range=SendDlgItemMessage(hwnd,DLG_SCALEC_FROM_S,SPNM_GETRANGE,0,0);
          i=GetDlgItemInt(hwnd,DLG_SCALEC_FROM,&err,FALSE);
          if(!err || i < LOWORD(range) || i > HIWORD(range))goto CANCELX;
          if(err)sp->firstframe=(short)i;
          range=SendDlgItemMessage(hwnd,DLG_SCALEC_TO_S,SPNM_GETRANGE,0,0);
          i=GetDlgItemInt(hwnd,DLG_SCALEC_TO,&err,FALSE);
          if(!err || i < LOWORD(range) || i > HIWORD(range))goto CANCELX;
          if(err)sp->lastframe=(short)i;
          ilens=GetDlgItemInt(hwnd,DLG_SCALEC_LENS,&err,FALSE);
          if(!err)goto CANCELX;
          if(GetDlgItemText(hwnd,DLG_SCALEC_AR,str,10) == 0)goto CANCELX;
          if((ar=atof(str)) == 0.0)goto CANCELX;
          if(ilens == 50)sp->Sx=1.0;
          else           sp->Sx=(double)ilens/50.0;
          sp->Sy = sp->Sx * ar;
          sp->Sz = sp->Sx;
          EndDialog(hwnd,OK);
          return(TRUE);
          CANCELX:
          EndDialog(hwnd,FAIL);
          return TRUE;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short EditCameraSize(double *sx, double *sy, double *sz, size *xp, int d){
 int i;
 size Size;
 SIZESTRUCT SS;
 HWND original;
 memcpy(&Size,xp,sizeof(size));
 SS.sp=&Size; SS.d=d;
 original=GetActiveWindow();
 EnableToolPannels(ALL_PANNELS,FALSE);
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_SCALEC),ghwnd_main,
            (DLGPROC)SizeC_DlgProc,(LPARAM)&SS)) == OK){
   EDIT_ACTION=YES;
   if(Size.lastframe >= Size.firstframe){
     memcpy(xp,&Size,sizeof(size));
     *sx=Size.Sx;
     *sy=Size.Sy;
     *sz=Size.Sz;
   }
 }
 EnableToolPannels(ALL_PANNELS,TRUE);
 if(original != NULL)SetActiveWindow(original);
 return (short)i;
}

void LensSettings(void){
 size *tss=NULL;
 position *tpp=NULL;
 align *taa=NULL;
 object *too=NULL;
 GetTimelinesAtFrame(SelectedNode,CurrentFrame,&too,
                    &tpp,&taa,&tss);
 if(tss != NULL)EditSize(SelectedNode,(short)tss->lastframe,
                                      (short)tss->lastframe);
 return;
}

BOOL CALLBACK EditObjectDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 static BOOL actor_edit_flag;
 static object actor;
 static short l_limit,u_limit;
 static object *op;
 BOOL err;
 char effectname[256];
 int i,id;
 short l,f;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     op=(object *)lparam;
//     SetWindowLong(hdlg,GWL_USERDATA,lparam);
     memcpy(&actor,op,sizeof(object));
     l_limit=(short)(op->firstframe);
     if(op->last != NULL)l_limit=op->last->lastframe+1;
     else l_limit=1;
     u_limit=(short)(op->lastframe);
     if(op->next != NULL)u_limit=op->next->firstframe-1;
     else u_limit=Nframes;
     if(op->effect != NULL){
       actor.effect=(char *)X__Malloc(strlen(op->effect)+1);
       if(actor.effect != NULL){
         memcpy(actor.effect,op->effect,strlen(op->effect)+1);
       }
     }
     actor_edit_flag=FALSE;
     EnableToolPannels(ALL_PANNELS,FALSE);
     SetDlgItemInt(hdlg,DLG_COSTUME_FROM,op->firstframe,FALSE);
     SetDlgItemInt(hdlg,DLG_COSTUME_TO  ,op->lastframe, FALSE);
     SendDlgItemMessage(hdlg,DLG_COSTUME_FROMS,SPNM_SETRANGE,0,
                        MAKELPARAM(l_limit,u_limit));
     SendDlgItemMessage(hdlg,DLG_COSTUME_FROMS,SPNM_SETCRNTVALUE,
                         (WPARAM)op->firstframe,0);
     SendDlgItemMessage(hdlg,DLG_COSTUME_FROMS,SPNM_SETEDITCTRL,0,
                         (LPARAM)GetDlgItem(hdlg,DLG_COSTUME_FROM));
     SendDlgItemMessage(hdlg,DLG_COSTUME_TOS,SPNM_SETRANGE,0,
                        MAKELPARAM(l_limit,u_limit));
     SendDlgItemMessage(hdlg,DLG_COSTUME_TOS,SPNM_SETCRNTVALUE,
                        (WPARAM)op->lastframe,0);
     SendDlgItemMessage(hdlg,DLG_COSTUME_TOS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hdlg,DLG_COSTUME_TO));
     SetDlgItemText(hdlg,DLG_COSTUME_NAME,short_form(actor.name));
     SetDlgItemText(hdlg,DLG_COSTUME_PATH,truncate_path(actor.name,40));
     if(op->morph == QMORPH)
          SendDlgItemMessage(hdlg,DLG_COSTUME_MORPH_Q,BM_SETCHECK,TRUE,0);
     else if(op->morph == YES)
          SendDlgItemMessage(hdlg,DLG_COSTUME_MORPH_LINEAR,BM_SETCHECK,TRUE,0);
     else SendDlgItemMessage(hdlg,DLG_COSTUME_MORPH_NO,BM_SETCHECK,TRUE,0);
     if(op->bWireframe)
          SendDlgItemMessage(hdlg,DLG_COSTUME_WIREFRAME,BM_SETCHECK,TRUE,0);
     if(op->self_shadow)
          SendDlgItemMessage(hdlg,DLG_COSTUME_SELF_SHADOWS,BM_SETCHECK,TRUE,0);
     if(op->cast_shadow)
          SendDlgItemMessage(hdlg,DLG_COSTUME_CAST_SHADOWS,BM_SETCHECK,TRUE,0);
     if(op->show_shadow)
          SendDlgItemMessage(hdlg,DLG_COSTUME_SHOW_SHADOWS,BM_SETCHECK,TRUE,0);
     if(op->effect == NULL){
       EnableWindow(GetDlgItem(hdlg,DLG_COSTUME_EFFECT_EDIT),FALSE);
       SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_NAME,"No Effect in Use");
       SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_PATH," ");
     }
     else{
       SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_SET,"&Drop");
       sscanf(op->effect,"%s",effectname);
       RestoreWhiteSpaces(effectname);
       if(strlen(effectname) == 1){
         id=atoi(effectname);
         sprintf(effectname,"%seffects\\%s",gszHomeDir,
                 internal_effect_list[id-1]);
       }
       SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_NAME,short_form(effectname));
       SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_PATH,truncate_path(effectname,40));
     }
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
//      op=(object *)GetWindowLong(hdlg,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_COSTUME_CHANGE:
          if(SelectSfxFileName(0,gszSCEfile,gszSCEdir,IDX_MISC_REPLACEMENT,
             "(*.MFX)|*.mfx|",hdlg) == TRUE){
            strcpy(actor.name,gszSCEfile);
            actor_edit_flag=TRUE;
            SetDlgItemText(hdlg,DLG_COSTUME_NAME,short_form(actor.name));
            SetDlgItemText(hdlg,DLG_COSTUME_PATH,truncate_path(actor.name,40));
          }
          break;
        case DLG_COSTUME_EFFECT_SET:
          if(op->effect == NULL)AddEffect(op,hdlg);
          else{
            DeleteEffect(op);
          }
          if(op->effect == NULL){
            EnableWindow(GetDlgItem(hdlg,DLG_COSTUME_EFFECT_EDIT),FALSE);
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_NAME,"No Effect in Use");
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_PATH," ");
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_SET,"&Set");
          }
          else{
            sscanf(op->effect,"%s",effectname);
            RestoreWhiteSpaces(effectname);
            if(strlen(effectname) == 1){
              id=atoi(effectname);
              sprintf(effectname,"%seffects\\%s",gszHomeDir,
                            internal_effect_list[id-1]);
            }
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_SET,"&Drop");
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_NAME,short_form(effectname));
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_PATH,truncate_path(effectname,40));
            EnableWindow(GetDlgItem(hdlg,DLG_COSTUME_EFFECT_EDIT),TRUE);
          }
          break;
        case DLG_COSTUME_EFFECT_EDIT:
          if(op->effect == NULL)break;
          EditActorEffect(op,hdlg);
          if(op->effect == NULL){
            EnableWindow(GetDlgItem(hdlg,DLG_COSTUME_EFFECT_EDIT),FALSE);
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_NAME,"No Effect in Use");
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_PATH," ");
            SetDlgItemText(hdlg,DLG_COSTUME_EFFECT_SET,"&Set");
          }
          break;
        case DLG_COSTUME_MORPH_NO:
          actor.morph=NO;
          break;
        case DLG_COSTUME_MORPH_LINEAR:
          actor.morph=YES;
          break;
        case DLG_COSTUME_MORPH_Q:
          actor.morph=QMORPH;
          break;
        case IDCANCEL:
        case DLG_COSTUME_CANCEL:
          CANCELX:
          if(op->effect != NULL)DeleteEffect(op);
          if(actor.effect != NULL){
            LoadEffectLibrary(actor.effectname,op);
            op->effect=actor.effect;
          }
          EndDialog(hdlg,-1);
          return(TRUE);
        case IDOK:
        case DLG_COSTUME_OK:
          i=GetDlgItemInt(hdlg,DLG_COSTUME_FROM,&err,FALSE);
          if(!err || i < l_limit ||
                     i > u_limit)goto CANCELX;
          actor.firstframe = (short)i;
          i=GetDlgItemInt(hdlg,DLG_COSTUME_TO,&err,FALSE);
          if(!err || i < l_limit ||
                     i > u_limit)goto CANCELX;
          actor.lastframe = (short)i;
          if(actor.lastframe < actor.firstframe)goto CANCELX;
          if(actor_edit_flag){ /* change the models database etc */
            ReplaceObjectFile(op,actor.name);
          }
          f=actor.firstframe; l=actor.lastframe;
          if(actor.type == PATH && actor.v != NULL){
            op->v=ReTweenVelocity(&actor,actor.v,
                            (int)(op->lastframe-op->firstframe+1),
                            (int)(l-f+1),actor.pathlength,0);
          }
          op->firstframe=f; op->lastframe=l;
          op->morph=actor.morph;
          if(SendDlgItemMessage(hdlg,DLG_COSTUME_WIREFRAME,BM_GETCHECK,
                                0,0))op->bWireframe=1;
          else                       op->bWireframe=0;
          if(SendDlgItemMessage(hdlg,DLG_COSTUME_SELF_SHADOWS,BM_GETCHECK,
                                0,0))op->self_shadow=1;
          else                       op->self_shadow=0;
          if(SendDlgItemMessage(hdlg,DLG_COSTUME_SHOW_SHADOWS,BM_GETCHECK,
                                0,0))op->show_shadow=1;
          else                       op->show_shadow=0;
          if(SendDlgItemMessage(hdlg,DLG_COSTUME_CAST_SHADOWS,BM_GETCHECK,
                                0,0))op->cast_shadow=1;
          else                       op->cast_shadow=0;
          EndDialog(hdlg,-1);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK EditObjectXipDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 static char *local_xip;
 static short l_limit,u_limit;
 static object *op;
 char   xipfilename[256];
 BOOL err;
 int i,id;
 short l,f;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     op=(object *)lparam;
     l_limit=(short)(op->firstframe);
     if(op->last != NULL)l_limit=op->last->lastframe+1;
     else l_limit=1;
     u_limit=(short)(op->lastframe);
     if(op->next != NULL)u_limit=op->next->firstframe-1;
     else u_limit=Nframes;
     EnableToolPannels(ALL_PANNELS,FALSE);
     SetDlgItemInt(hdlg,DLG_COSTUMEXIP_FROM,op->firstframe,FALSE);
     SetDlgItemInt(hdlg,DLG_COSTUMEXIP_TO  ,op->lastframe, FALSE);
     SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_FROMS,SPNM_SETRANGE,0,
                        MAKELPARAM(l_limit,u_limit));
     SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_FROMS,SPNM_SETCRNTVALUE,
                         (WPARAM)op->firstframe,0);
     SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_FROMS,SPNM_SETEDITCTRL,0,
                         (LPARAM)GetDlgItem(hdlg,DLG_COSTUME_FROM));
     SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_TOS,SPNM_SETRANGE,0,
                        MAKELPARAM(l_limit,u_limit));
     SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_TOS,SPNM_SETCRNTVALUE,
                        (WPARAM)op->lastframe,0);
     SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_TOS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hdlg,DLG_COSTUME_TO));
     if(op->morph == YES)
          SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_MORPH_YES,BM_SETCHECK,TRUE,0);
     else SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_MORPH_NO,BM_SETCHECK,TRUE,0);
     if(op->xip == NULL){
       local_xip=NULL;
       SetDlgItemText(hdlg,DLG_COSTUMEXIP_NAME,"No Post Process Loaded");
       SetDlgItemText(hdlg,DLG_COSTUMEXIP_PATH," ");
     }
     else{
       local_xip=(char *)X__Malloc(strlen(op->xip)+1);
       if(local_xip != NULL)strcpy(local_xip,op->xip);
       sscanf(op->xip,"%s",xipfilename);
       RestoreWhiteSpaces(xipfilename);
       SetDlgItemText(hdlg,DLG_COSTUMEXIP_NAME,short_form(xipfilename));
       SetDlgItemText(hdlg,DLG_COSTUMEXIP_PATH,truncate_path(xipfilename,40));
     }
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_COSTUMEXIP_SET:{
            char *temp_xip;
            if((temp_xip=EditObjectImageProcess(hdlg,NULL)) != NULL){
              if(local_xip != NULL)X__Free(local_xip);
              local_xip=temp_xip;
              sscanf(local_xip,"%s",xipfilename);
              RestoreWhiteSpaces(xipfilename);
              SetDlgItemText(hdlg,DLG_COSTUMEXIP_NAME,short_form(xipfilename));
              SetDlgItemText(hdlg,DLG_COSTUMEXIP_PATH,truncate_path(xipfilename,40));
            }
          }
          break;
        case DLG_COSTUMEXIP_EDIT:
          if(local_xip != NULL){
            local_xip=EditObjectImageProcess(hdlg,local_xip);
          }
          break;
        case IDCANCEL:
          CANCELX:
          if(local_xip != NULL)X__Free(local_xip);
          EndDialog(hdlg,-1);
          return(TRUE);
        case DLG_COSTUMEXIP_DELETE:
          EndDialog(hdlg,0);
          return(TRUE);
        case DLG_COSTUMEXIP_CUTOUT:
          EndDialog(hdlg,3);
          return(TRUE);
        case DLG_COSTUMEXIP_COPY:
          EndDialog(hdlg,4);
          return(TRUE);
        case IDOK:
          i=GetDlgItemInt(hdlg,DLG_COSTUMEXIP_FROM,&err,FALSE);
          if(!err || i < l_limit ||
                     i > u_limit)goto CANCELX;
          f = (short)i;
          i=GetDlgItemInt(hdlg,DLG_COSTUMEXIP_TO,&err,FALSE);
          if(!err || i < l_limit ||
                     i > u_limit)goto CANCELX;
          l = (short)i;
          if(l < f)goto CANCELX;
          op->firstframe=f; op->lastframe=l;
          if(SendDlgItemMessage(hdlg,DLG_COSTUMEXIP_MORPH_YES,BM_GETCHECK,
                                0,0))op->morph=YES;
          else op->morph=NO;
          if(local_xip != NULL){
            X__Free(op->xip);
            op->xip=(char *)X__Malloc(strlen(local_xip)+1);
            if(op->xip != NULL)strcpy(op->xip,local_xip);
            X__Free(local_xip);
          }
          EndDialog(hdlg,-1);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void BuildImageListTable(HWND hctl, BOOL erase){
 XIP  *Ip;
 char name[256];
 if(erase)SendMessage(hctl,LB_RESETCONTENT,0,0);
 Ip=FirstIp;
 while(Ip != NULL){
   if(Ip->p != NULL){
     sscanf(Ip->p,"%s",name);
     RestoreWhiteSpaces(name);
     SendMessage(hctl,LB_ADDSTRING,0,(LPARAM)name);
   }
   Ip=Ip->next;
 }
 SendMessage(hctl,LB_SETCURSEL,0,0);
}

BOOL CALLBACK EditXipDlgProc(HWND hdlg,UINT msg,
                          WPARAM wparam,LPARAM lparam){
 HWND hctl;
 XIP  *Ip;
 char name[256];
 BOOL flag;
 int id,i;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     Ip=FirstIp;
     if(Ip == NULL){
       EnableWindow(GetDlgItem(hdlg,DLG_XIP_EDIT),FALSE);
       EnableWindow(GetDlgItem(hdlg,DLG_XIP_DELETE),FALSE);
       EnableWindow(GetDlgItem(hdlg,DLG_XIP_VECTOR),FALSE);
       EnableWindow(GetDlgItem(hdlg,DLG_XIP_MOVE_FIRST),FALSE);
       EnableWindow(GetDlgItem(hdlg,DLG_XIP_MOVE_LAST),FALSE);
     }
     else {
       hctl=GetDlgItem(hdlg,DLG_XIP_LIST);
       BuildImageListTable(hctl,FALSE);
     }
     CentreDialogOnScreen(hdlg);
     EnableToolPannels(ALL_PANNELS,FALSE);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_XIP_LIST:
          if(HIWORD(wparam) == LBN_SELCHANGE){
            ;//MessageBeep(MB_OK);
          }
          if(HIWORD(wparam) == LBN_DBLCLK){
            SendMessage(hdlg,WM_COMMAND,MAKEWPARAM(DLG_XIP_EDIT,0),0);
          }
          break;
        case DLG_XIP_MOVE_FIRST:
        case DLG_XIP_MOVE_LAST:
        case DLG_XIP_EDIT:
          if((id=SendDlgItemMessage(hdlg,DLG_XIP_LIST,LB_GETCURSEL,0,0))
             == LB_ERR || (Ip=FirstIp) == NULL)break;
          i=0; while(Ip != NULL){
            if(i == id){
              if(LOWORD(wparam) == DLG_XIP_MOVE_FIRST){
                if(Ip != FirstIp){
                  if(MainIp == Ip)MainIp=Ip->last;
                  FirstIp->last=Ip;
                  Ip->last->next=Ip->next;
                  if(Ip->next != NULL)Ip->next->last=Ip->last;
                  Ip->next=FirstIp;
                  Ip->last=NULL;
                  FirstIp=Ip;
                  hctl=GetDlgItem(hdlg,DLG_XIP_LIST);
                  BuildImageListTable(hctl,TRUE);
                }
              }
              else if(LOWORD(wparam) == DLG_XIP_MOVE_LAST){
                XIP *LastIp;
                LastIp=FirstIp;
                while(LastIp->next != NULL)LastIp=LastIp->next;
                if(Ip->next != NULL){
                  if(Ip == FirstIp)FirstIp=Ip->next;
                  Ip->next->last=Ip->last;
                  if(Ip->last != NULL)Ip->last->next=Ip->next;
                  Ip->next=NULL;
                  LastIp->next=Ip;
                  Ip->last=LastIp;
                  MainIp=Ip;
                  hctl=GetDlgItem(hdlg,DLG_XIP_LIST);
                  BuildImageListTable(hctl,TRUE);
                }
              }
              else EditImageProcess(hdlg,Ip);
              break;
            }
            i++; Ip=Ip->next;
          }
          break;
        case DLG_XIP_VECTOR:{
            char temp[MAX_FILE],new_dir[MAX_DIR];
            PSTR c,cn;
            int j;
            XIP *Jp;
            strcpy(new_dir,gszHomeDir);strcat(new_dir,"postprocess\\");
            if(!RequestCharString(64,new_dir,"Post Processor Path"))break;
            HideWhiteSpaces(new_dir);
            if((j=strlen(new_dir)) < 3)break;
            if(new_dir[j-1] != '\\')strcat(new_dir,"\\");
            Jp=FirstIp;  while(Jp != NULL){
              sscanf(Jp->p,"%s",name);
              if((c=FileInPath(name)) != NULL){
                strcpy(temp,c); strcpy(name,new_dir); strcat(name,temp);
                c=strchr(Jp->p,' ');
                j=strlen(Jp->p)+strlen(c)+2;
                if((cn=(PSTR)X__Malloc(j)) != NULL){
                  strcpy(cn,name); strcat(cn," "); strcat(cn,c+1);
                  X__Free(Jp->p);
                  Jp->p=cn;
                }
              }
              Jp=Jp->next;
            }
            hctl=GetDlgItem(hdlg,DLG_XIP_LIST);
            BuildImageListTable(hctl,TRUE);
          }
          break;
        case DLG_XIP_ADD:
          if(FirstIp == NULL)flag=TRUE; else flag=FALSE;
          if(CreateImageProcess(hdlg) != NULL){
            sscanf(MainIp->p,"%s",name);
            RestoreWhiteSpaces(name);
            SendDlgItemMessage(hdlg,DLG_XIP_LIST,LB_ADDSTRING,0,(LPARAM)name);
            if(flag){
              EnableWindow(GetDlgItem(hdlg,DLG_XIP_EDIT),TRUE);
              EnableWindow(GetDlgItem(hdlg,DLG_XIP_DELETE),TRUE);
              EnableWindow(GetDlgItem(hdlg,DLG_XIP_VECTOR),TRUE);
              EnableWindow(GetDlgItem(hdlg,DLG_XIP_MOVE_FIRST),TRUE);
              EnableWindow(GetDlgItem(hdlg,DLG_XIP_MOVE_LAST),TRUE);
              SendDlgItemMessage(hdlg,DLG_XIP_LIST,LB_SETCURSEL,0,0);
            }
          }
          break;
        case DLG_XIP_DELETE:
          if((id=SendDlgItemMessage(hdlg,DLG_XIP_LIST,LB_GETCURSEL,0,0))
             == LB_ERR || (Ip=FirstIp) == NULL)break;
          i=0; while(Ip != NULL){
            if(i == id){
              DeleteImageProcess(Ip);
              SendDlgItemMessage(hdlg,DLG_XIP_LIST,LB_DELETESTRING,(WPARAM)id,0);
              if(FirstIp == NULL){
                EnableWindow(GetDlgItem(hdlg,DLG_XIP_EDIT),FALSE);
                EnableWindow(GetDlgItem(hdlg,DLG_XIP_DELETE),FALSE);
                EnableWindow(GetDlgItem(hdlg,DLG_XIP_VECTOR),FALSE);
                EnableWindow(GetDlgItem(hdlg,DLG_XIP_MOVE_FIRST),FALSE);
                EnableWindow(GetDlgItem(hdlg,DLG_XIP_MOVE_LAST),FALSE);
              }
              else SendDlgItemMessage(hdlg,DLG_XIP_LIST,LB_SETCURSEL,0,0);
              break;
            }
            i++; Ip=Ip->next;
          }
          break;
        case IDCANCEL:
          EndDialog(hdlg,-1);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK ShiftKeysDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 int n;
 BOOL type,err;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     n=0;
     SetDlgItemInt(hdlg,DLG_SHIFTKEYS_NUMBER,n,FALSE);
     SendDlgItemMessage(hdlg,DLG_SHIFTKEYS_UP,BM_SETCHECK,TRUE,0);
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDOK:
          n=GetDlgItemInt(hdlg,DLG_SHIFTKEYS_NUMBER,&err,FALSE);
          if(SendDlgItemMessage(hdlg,DLG_SHIFTKEYS_DOWN,BM_GETCHECK,0,0))
            n = -n;
          EndDialog(hdlg,n);
          return(TRUE);
          break;
        case IDCANCEL:
          EndDialog(hdlg,0);
          return(TRUE);
          break;
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

#include "..\common\animate\about.h"

BOOL CALLBACK AboutDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 char name[255];
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     CentreDialogOnScreen(hdlg);
//     strcpy(name,gszHomeDir);
//     strcat(name,"OPENFX.AVI");
     SendDlgItemMessage(hdlg,DLG_ABOUT_VERSION,WM_SETTEXT,
                       0,(LPARAM)ABOUT_VERSION);
//     Animate_Open(GetDlgItem(hdlg,DLG_ABOUT_ANIMATE),name);
//     Animate_Play(GetDlgItem(hdlg, DLG_ABOUT_ANIMATE),0, -1, -1);
     return TRUE;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case IDOK:
          EndDialog(hdlg,TRUE);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK GeneralPreferencesDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL state,err;
 char *w, ss[32], yesyes[]="YES", nono[]="NO",appname[256];
 int i;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     if(Preferences.quick_trick)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_QUICKTRICK,BM_SETCHECK,state,0);
     if(Preferences.undo)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_UNDO,BM_SETCHECK,state,0);
     if(Preferences.ground)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_GROUND,BM_SETCHECK,state,0);
     if(Preferences.tooltips)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLTIPS,BM_SETCHECK,state,0);
     if(Preferences.autocentre)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_AUTOCENTRE,BM_SETCHECK,state,0);
     if(!Preferences.detail_auto)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_AUTODETAIL,BM_SETCHECK,state,0);
     if(Preferences.toolsmall)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLSMALL,BM_SETCHECK,state,0);
     if(Preferences.toolleft)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLLEFT,BM_SETCHECK,state,0);
     if(Preferences.use_gl_robot)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_ROBOTGL,BM_SETCHECK,state,0);
     SetDlgItemInt(hdlg,DLG_PREFERENCES_STEPDETAIL,
                   Preferences.detail_step,FALSE);
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_STEPDETAILS,SPNM_SETRANGE,0,
                        MAKELPARAM(1,1000));
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_STEPDETAILS,SPNM_SETCRNTVALUE,
                        (WPARAM)Preferences.detail_step,0);
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_STEPDETAILS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hdlg,DLG_PREFERENCES_STEPDETAIL));

     GetPrivateProfileString(IniSection,"EDITAPP","$$$",
                             appname,sizeof(appname),IniFilename);
     if(strncmp(appname,"$$$",3) != 0){
       SetDlgItemText(hdlg,DLG_PREFERENCES_EDITAPP,(LPCSTR)appname);
     }
     return (TRUE);
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       default: break;
       case DLG_PREFERENCES_EDITAPPSET:{
         char appdir[256];
         if(GetDlgItemText(hdlg,DLG_PREFERENCES_EDITAPP,appname,255) == 0){
           strcpy(appname,"*.exe");
         }
         strcpy(appdir,appname);
         if(SelectFileName(0,appname,appdir," ",
                           "(*.EXE)|*.exe|",ghwnd_main)
                           == TRUE){
           SetDlgItemText(hdlg,DLG_PREFERENCES_EDITAPP,(LPCSTR)appname);
         }
       }
       break;
     }
     break;
   case WM_NOTIFY:
     switch(((LPNMHDR)lparam)->code){
       case PSN_APPLY:
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_QUICKTRICK,BM_GETCHECK,0,0))
            Preferences.quick_trick=TRUE; else Preferences.quick_trick=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_UNDO,BM_GETCHECK,0,0))
            Preferences.undo=TRUE; else Preferences.undo=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_GROUND,BM_GETCHECK,0,0))
            Preferences.ground=TRUE; else Preferences.ground=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLTIPS,BM_GETCHECK,0,0))
            Preferences.tooltips=TRUE; else Preferences.tooltips=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_AUTOCENTRE,BM_GETCHECK,0,0))
            Preferences.autocentre=TRUE; else Preferences.autocentre=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLSMALL,BM_GETCHECK,0,0))
            Preferences.toolsmall=TRUE; else Preferences.toolsmall=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLLEFT,BM_GETCHECK,0,0))
            Preferences.toolleft=TRUE; else Preferences.toolleft=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_AUTODETAIL,BM_GETCHECK,0,0))
            Preferences.detail_auto=FALSE; else Preferences.detail_auto=TRUE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_ROBOTGL,BM_GETCHECK,0,0))
            Preferences.use_gl_robot=TRUE; else Preferences.use_gl_robot=FALSE;
         i=GetDlgItemInt(hdlg,DLG_PREFERENCES_STEPDETAIL,&err,FALSE);
         if(err && i > 0 && i < 1000)Preferences.detail_step=i;
         if(GetDlgItemText(hdlg,DLG_PREFERENCES_EDITAPP,appname,255) != 0)
           WritePrivateProfileString(IniSection,"EDITAPP",appname,IniFilename);
         else
           WritePrivateProfileString(IniSection,"EDITAPP",NULL,IniFilename);
         if(Preferences.quick_trick)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"QUICKTRICK",w,IniFilename);
         if(Preferences.undo)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"UNDO",w,IniFilename);
         if(Preferences.undo)UNDO_ACTIVE=YES; else UNDO_ACTIVE=NO;
         if(Preferences.ground)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"GROUND",w,IniFilename);
         if(Preferences.tooltips)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"TOOLTIPS",w,IniFilename);
         if(Preferences.autocentre)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"AUTOCENTRE",w,IniFilename);
         if(Preferences.toolsmall)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"TOOLSMALL",w,IniFilename);
         if(Preferences.toolleft)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"TOOLLEFT",w,IniFilename);
         if(Preferences.detail_auto)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"DETAILAUTO",w,IniFilename);
         sprintf(ss,"%ld",Preferences.detail_step);
         WritePrivateProfileString(IniSection,"DETAILSTEP",ss,
                                   IniFilename);
         if(Preferences.use_gl_robot)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"GLROBOT",w,IniFilename);
         break;
       default:  break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK StartupPreferencesDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL state,err;
 char str[32];
 char *w, yesyes[]="YES", nono[]="NO";
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     if(Preferences.buttons)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_BUTTONS,BM_SETCHECK,state,0);
//     if(Preferences.show_select)state=TRUE; else state=FALSE;
//     SendDlgItemMessage(hdlg,DLG_PREFERENCES_SELECT,BM_SETCHECK,state,0);
     if(Preferences.default_start)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_STARTUP,BM_SETCHECK,state,0);
     if(Preferences.use_positions)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_POSITIONS,BM_SETCHECK,state,0);
     if(Preferences.background)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_BACKGROUND,BM_SETCHECK,state,0);
     SetDlgItemInt(hdlg,DLG_PREFERENCES_SPLASH,Preferences.splash,FALSE);
     return (TRUE);
   case WM_NOTIFY:
     switch(((LPNMHDR)lparam)->code){
       case PSN_APPLY:
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_BUTTONS,BM_GETCHECK,0,0))
            Preferences.buttons=TRUE; else Preferences.buttons=FALSE;
//         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_SELECT,BM_GETCHECK,0,0))
//            Preferences.show_select=TRUE; else Preferences.show_select=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_STARTUP,BM_GETCHECK,0,0))
            Preferences.default_start=TRUE; else Preferences.default_start=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_POSITIONS,BM_GETCHECK,0,0))
            Preferences.use_positions=TRUE; else Preferences.use_positions=FALSE;
         Preferences.splash=GetDlgItemInt(hdlg,DLG_PREFERENCES_SPLASH,&err,FALSE);
         if(Preferences.buttons)w=yesyes; else w=nono;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_BACKGROUND,BM_GETCHECK,0,0))
            Preferences.background=TRUE; else Preferences.background=FALSE;
         WritePrivateProfileString(IniSection,"BUTTONS",w,IniFilename);
         if(Preferences.show_select)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"SELECTPANNEL",w,IniFilename);
         if(Preferences.default_start)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"DEFAULTSTART",w,IniFilename);
         if(Preferences.use_positions)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"POSITIONS",w,IniFilename);
         if(Preferences.background)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"BACKIMAGE",w,IniFilename);
         sprintf(str,"%ld",Preferences.splash);
         WritePrivateProfileString(IniSection,"SPLASH",str,IniFilename);
         break;
       default:  break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK ModelPathPreferencesDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 char txt[256],none[64];
 char tempstr[256],tempdir[256];
 static HBITMAP CrossButtonBitmap=NULL;
 static HBITMAP FolderButtonBitmap=NULL;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     LoadString(ghinst_main,IDX_MISC_NONE,none,63);
     GetPrivateProfileString(IniRenderSection,"MODEL-PATH1",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH1,txt);
     GetPrivateProfileString(IniRenderSection,"MODEL-PATH2",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH2,txt);
     GetPrivateProfileString(IniRenderSection,"MODEL-PATH3",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH3,txt);
     GetPrivateProfileString(IniRenderSection,"MODEL-PATH4",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH4,txt);
     CrossButtonBitmap=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_CROSS));
     FolderButtonBitmap=(HBITMAP)LoadImage(ghinst_main,"FOLDERS",IMAGE_ICON,
                               0,0,LR_DEFAULTCOLOR); // ICON

     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH1_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH2_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH3_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH4_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);

     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH1_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH2_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH3_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MODPATH4_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);

     return (TRUE);
   case WM_DESTROY:
     if(CrossButtonBitmap != NULL)DeleteObject(CrossButtonBitmap);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case DLG_PREFERENCES_MODPATH1_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH1,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH1,tempdir);
         }
         break;
       case DLG_PREFERENCES_MODPATH1_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH1,none);
         break;
       case DLG_PREFERENCES_MODPATH2_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH2,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH2,tempdir);
         }
         break;
       case DLG_PREFERENCES_MODPATH2_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH2,none);
         break;
       case DLG_PREFERENCES_MODPATH3_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH3,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH3,tempdir);
         }
         break;
       case DLG_PREFERENCES_MODPATH3_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH3,none);
         break;
       case DLG_PREFERENCES_MODPATH4_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH4,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH4,tempdir);
         }
         break;
       case DLG_PREFERENCES_MODPATH4_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH4,none);
         break;
       default: break;
     }
     break;
   case WM_NOTIFY:
     switch(((LPNMHDR)lparam)->code){
       case PSN_APPLY:
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH1,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MODEL-PATH1",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MODEL-PATH1",
                                        tempdir,IniFilename);
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH2,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MODEL-PATH2",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MODEL-PATH2",
                                        tempdir,IniFilename);
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH3,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MODEL-PATH3",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MODEL-PATH3",
                                        tempdir,IniFilename);
         GetDlgItemText(hdlg,DLG_PREFERENCES_MODPATH4,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MODEL-PATH4",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MODEL-PATH4",
                                        tempdir,IniFilename);
         break;
       default: break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK MapPathPreferencesDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 char txt[256],none[64];
 char tempstr[256],tempdir[256];
 static HBITMAP CrossButtonBitmap=NULL;
 static HBITMAP FolderButtonBitmap=NULL;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     LoadString(ghinst_main,IDX_MISC_NONE,none,63);
     GetPrivateProfileString(IniRenderSection,"MAP-PATH1",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH1,txt);
     GetPrivateProfileString(IniRenderSection,"MAP-PATH2",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH2,txt);
     GetPrivateProfileString(IniRenderSection,"MAP-PATH3",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH3,txt);
     GetPrivateProfileString(IniRenderSection,"MAP-PATH4",none,
                             txt,255,IniFilename);
     SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH4,txt);
     CrossButtonBitmap=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_CROSS));
     FolderButtonBitmap=(HBITMAP)LoadImage(ghinst_main,"FOLDERS",IMAGE_ICON,
                               0,0,LR_DEFAULTCOLOR); // ICON

     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH1_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH2_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH3_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH4_D),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)CrossButtonBitmap);

     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH1_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH2_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH3_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);
     SendMessage(GetDlgItem(hdlg,DLG_PREFERENCES_MAPPATH4_B),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)FolderButtonBitmap);

     return (TRUE);
   case WM_DESTROY:
     if(CrossButtonBitmap != NULL)DeleteObject(CrossButtonBitmap);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case DLG_PREFERENCES_MAPPATH1_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH1,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH1,tempdir);
         }
         break;
       case DLG_PREFERENCES_MAPPATH1_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH1,none);
         break;
       case DLG_PREFERENCES_MAPPATH2_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH2,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH2,tempdir);
         }
         break;
       case DLG_PREFERENCES_MAPPATH2_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH2,none);
         break;
       case DLG_PREFERENCES_MAPPATH3_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH3,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH3,tempdir);
         }
         break;
       case DLG_PREFERENCES_MAPPATH3_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH3,none);
         break;
       case DLG_PREFERENCES_MAPPATH4_B:
         strcpy(tempstr,"temp.scr");
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH4,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_CHOOSEPATH,txt,255);
         if(SelectDirectoryName(tempstr,tempdir,txt,
           "(*.SCR)|*.scr|",hdlg) == TRUE){
           if(strlen(tempdir) == 2)strcat(tempdir,"\\");
           SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH4,tempdir);
         }
         break;
       case DLG_PREFERENCES_MAPPATH4_D:
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         SetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH4,none);
         break;
       default: break;
     }
     break;
   case WM_NOTIFY:
     switch(((LPNMHDR)lparam)->code){
       case PSN_APPLY:
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH1,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MAP-PATH1",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MAP-PATH1",
                                        tempdir,IniFilename);
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH2,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MAP-PATH2",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MAP-PATH2",
                                        tempdir,IniFilename);
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH3,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MAP-PATH3",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MAP-PATH3",
                                        tempdir,IniFilename);
         GetDlgItemText(hdlg,DLG_PREFERENCES_MAPPATH4,tempdir,255);
         LoadString(ghinst_main,IDX_MISC_NONE,none,63);
         if(!strcmp(tempdir,none))
              WritePrivateProfileString(IniRenderSection,"MAP-PATH4",
                                        NULL,IniFilename);
         else WritePrivateProfileString(IniRenderSection,"MAP-PATH4",
                                        tempdir,IniFilename);
         break;
       default: break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

void PreferencesPropertyDlgProc(HWND hwndOwner){
 char title[256];
 int i;
 PROPSHEETPAGE psp[4];
 PROPSHEETHEADER psh;
 HWND hWndP;
 for(i=0;i<4;i++){
   psp[i].dwSize = sizeof(PROPSHEETPAGE);
   psp[i].dwFlags = PSP_DEFAULT;
   psp[i].hInstance = ghinst_main;
   psp[i].pszIcon = NULL;
   psp[i].lParam = 0;
 }
 psp[0].pszTemplate = MAKEINTRESOURCE(DLG_PREFERENCES_GENERAL);
 psp[0].pfnDlgProc = GeneralPreferencesDlgProc;
 psp[1].pszTemplate = MAKEINTRESOURCE(DLG_PREFERENCES_START);
 psp[1].pfnDlgProc = StartupPreferencesDlgProc;
 psp[2].pszTemplate = MAKEINTRESOURCE(DLG_PREFERENCES_MODPATH);
 psp[2].pfnDlgProc = ModelPathPreferencesDlgProc;
 psp[3].pszTemplate = MAKEINTRESOURCE(DLG_PREFERENCES_MAPPATH);
 psp[3].pfnDlgProc = MapPathPreferencesDlgProc;
 psh.dwSize = sizeof(PROPSHEETHEADER);
 psh.dwFlags = PSH_PROPSHEETPAGE | PSH_NOAPPLYNOW;
 psh.hwndParent = hwndOwner;
 psh.hInstance = ghinst_main;
 psh.nStartPage = 0;
 psh.pszIcon = NULL;
 LoadString(ghinst_main,IDX_MISC_PREF_TITLE,title,255);
 psh.pszCaption = (LPSTR)title;
 psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
 psh.ppsp = (LPCPROPSHEETPAGE) &psp;
 PropertySheet(&psh);
 return;
}

static BOOL CALLBACK TrackerToolDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     if(gToolFlags.flagT == 0)
       SendDlgItemMessage(hdlg,DLG_TRACK_MOVE,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagT == 1)
       SendDlgItemMessage(hdlg,DLG_TRACK_TIMESHIFT,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagT == 2)
       SendDlgItemMessage(hdlg,DLG_TRACK_DELETE,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagT == 3)
       SendDlgItemMessage(hdlg,DLG_TRACK_TENSION,BM_SETCHECK,TRUE,0);
     return (TRUE);
   case WM_DESTROY:
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_TRACK_MOVE:      gToolFlags.flagT=0; break;
        case DLG_TRACK_TIMESHIFT: gToolFlags.flagT=1; break;
        case DLG_TRACK_DELETE:    gToolFlags.flagT=2; break;
        case DLG_TRACK_TENSION:   gToolFlags.flagT=3; break;
        default: break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK DirectorToolDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL err;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     if(gToolFlags.flagR == 0)
       SendDlgItemMessage(hdlg,DLG_DIRECT_MOVE,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagR == 1)
       SendDlgItemMessage(hdlg,DLG_DIRECT_TIMESHIFT,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagR == 2)
       SendDlgItemMessage(hdlg,DLG_DIRECT_DELETE,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagR == 3)
       SendDlgItemMessage(hdlg,DLG_DIRECT_TENSION,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagA == 0)
       SendDlgItemMessage(hdlg,DLG_ROTATOR_X,BM_SETCHECK,TRUE,0);
     else if(gToolFlags.flagA == 1)
       SendDlgItemMessage(hdlg,DLG_ROTATOR_Y,BM_SETCHECK,TRUE,0);
     else if(gToolFlags.flagA == 2)
       SendDlgItemMessage(hdlg,DLG_ROTATOR_Z,BM_SETCHECK,TRUE,0);
     if(gToolFlags.flagR == 0)err=TRUE; else err=FALSE;
     EnableWindow(GetDlgItem(hdlg,DLG_ROTATOR_X),err);
     EnableWindow(GetDlgItem(hdlg,DLG_ROTATOR_Y),err);
     EnableWindow(GetDlgItem(hdlg,DLG_ROTATOR_Z),err);
     return (TRUE);
   case WM_DESTROY:
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_ROTATOR_X:        gToolFlags.flagA=0; break;
        case DLG_ROTATOR_Y:        gToolFlags.flagA=1; break;
        case DLG_ROTATOR_Z:        gToolFlags.flagA=2; break;
        case DLG_DIRECT_MOVE:      gToolFlags.flagR=0; goto SETFLAGS;
        case DLG_DIRECT_TIMESHIFT: gToolFlags.flagR=1; goto SETFLAGS;
        case DLG_DIRECT_DELETE:    gToolFlags.flagR=2; goto SETFLAGS;
        case DLG_DIRECT_TENSION:   gToolFlags.flagR=3;
        SETFLAGS:
          if(gToolFlags.flagR == 0)err=TRUE; else err=FALSE;
          EnableWindow(GetDlgItem(hdlg,DLG_ROTATOR_X),err);
          EnableWindow(GetDlgItem(hdlg,DLG_ROTATOR_Y),err);
          EnableWindow(GetDlgItem(hdlg,DLG_ROTATOR_Z),err);
          break;
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK SplineTweenDlgProc(HWND hdlg, UINT msg, WPARAM wparam,
                                     LPARAM lparam ){
 // resource must contain at least one control or DLG3D is bad //
 static int xpos= -1,ypos= -1;
 RECT rc;
 switch( msg ){
   case WM_INITDIALOG:{
       TC_ITEM tie;
       int h;
       DWORD dwDlgBase=GetDialogBaseUnits();
       if(xpos == -1 && ypos == -1){
         RECT rc1;
         GetWindowRect(hdlg,&rc1);
         GetWindowRect(ghwndTools,&rc);
         if(bToolpannelLeft)xpos=rc.right+5;
         else               xpos=rc.left-(rc1.right-rc1.left)-5;
         ypos=rc.top;
       }
       SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
       lpDlgHdr=&DlgHdr;
       GetClientRect(hdlg,&rc);
       h=rc.bottom-rc.top;
       hwndTab=CreateWindow(WC_TABCONTROL," ",
                            WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
                            (D_BORDER*LOWORD(dwDlgBase))/4,
                            (D_BORDER*HIWORD(dwDlgBase))/8,
                            rc.right-(2*D_BORDER*LOWORD(dwDlgBase))/4,
                            h-(2*D_BORDER*HIWORD(dwDlgBase))/8,
                            hdlg,NULL,ghinst_main,NULL);
       LoadString(ghinst_main,IDX_MISC_LANGUAGE,res_str,256);
       if(strcmp(res_str,"ENGLISH") == 0)
         SendMessage(hwndTab,WM_SETFONT, /* for text in tab */
                    (WPARAM)GetStockObject(ANSI_VAR_FONT),0);
       memset(&tie,0,sizeof(TC_ITEM));
       tie.mask = TCIF_TEXT | TCIF_IMAGE;
       tie.iImage = -1;
       tie.pszText="Track";
       TabCtrl_InsertItem(hwndTab,0,&tie);
       tie.pszText="Direction";
       TabCtrl_InsertItem(hwndTab,1,&tie);
       SetRectEmpty(&rc);
       TabCtrl_AdjustRect(hwndTab,FALSE,&rc);
       rc.left += (D_BORDER*LOWORD(dwDlgBase))/4;
       rc.top += (D_BORDER*HIWORD(dwDlgBase))/8;
       CopyRect(&(lpDlgHdr->rcTab),&rc);
       lpDlgHdr->hwndDlg1=NULL;
       lpDlgHdr->hwndDlg2=NULL;
       lpDlgHdr->hwndDlg3=NULL;
       lpDlgHdr->hwndDlg4=NULL;
       lpDlgHdr->hwndDlg5=NULL;
       lpDlgHdr->hwndDlg6=NULL;
       lpDlgHdr->hwndDlg1=CreateDialog(ghinst_main,
                                    MAKEINTRESOURCE(DLG_TRACK),
                                   hdlg,(DLGPROC)TrackerToolDlgProc);
       SetWindowPos(lpDlgHdr->hwndDlg1,NULL,
                    lpDlgHdr->rcTab.left,lpDlgHdr->rcTab.top,0,0,
                    SWP_NOSIZE | SWP_NOZORDER);
       ShowWindow(lpDlgHdr->hwndDlg1,SW_SHOW);
     }
     return TRUE;
   case WM_NOTIFY:
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGING){
       int id=TabCtrl_GetCurSel(hwndTab);
       if(id == 0 || id == 1)SwitchTrackerType();
    }
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGE){
       int id=TabCtrl_GetCurSel(hwndTab);
       if(lpDlgHdr->hwndDlg1 != NULL){
         DestroyWindow(lpDlgHdr->hwndDlg1);
         lpDlgHdr->hwndDlg1=NULL;
       }
       if(lpDlgHdr->hwndDlg2 != NULL){
         DestroyWindow(lpDlgHdr->hwndDlg2);
         lpDlgHdr->hwndDlg2=NULL;
       }
       if(id == 0){
         lpDlgHdr->hwndDlg1=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_TRACK),
                                     hdlg,(DLGPROC)TrackerToolDlgProc);
         SetWindowPos(lpDlgHdr->hwndDlg1,NULL,
                      lpDlgHdr->rcTab.left,lpDlgHdr->rcTab.top,0,0,
                      SWP_NOSIZE | SWP_NOZORDER);
         ShowWindow(lpDlgHdr->hwndDlg1,SW_SHOW);
       }
       else if(id == 1){
         lpDlgHdr->hwndDlg2=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_DIRECT),
                                     hdlg,(DLGPROC)DirectorToolDlgProc);
         SetWindowPos(lpDlgHdr->hwndDlg2,NULL,
                      lpDlgHdr->rcTab.left,lpDlgHdr->rcTab.top,0,0,
                      SWP_NOSIZE | SWP_NOZORDER);
         ShowWindow(lpDlgHdr->hwndDlg2,SW_SHOW);
       }
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       default:
         break;
       case SC_CLOSE:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
         break;
     }
     break;
   case WM_COMMAND:
     switch( LOWORD(wparam) ){
       default:
         break;
     }
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_DESTROY:
     if(lpDlgHdr->hwndDlg1 != NULL)DestroyWindow(lpDlgHdr->hwndDlg1);
     lpDlgHdr->hwndDlg1=NULL;
     if(lpDlgHdr->hwndDlg2 != NULL)DestroyWindow(lpDlgHdr->hwndDlg2);
     lpDlgHdr->hwndDlg2=NULL;
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     break;
   default: break;
 }
 return( FALSE );
}

static void DisplayImageFile(char *pathname){
 WritePrivateProfileString("VIEWER","MAPFILE",pathname,IniFilename);
 LoadViewer(2);
 return;
}

static void EditImageFile(char *pathname){
 char appname[256];
 GetPrivateProfileString(IniSection,"EDITAPP","$$$",
                         appname,sizeof(appname),IniFilename);
 if(strncmp(appname,"$$$",3) != 0){
   ShellExecute(ghwnd_main,NULL,appname,
                pathname,appname,SW_SHOW);
 }
 return;
}

BOOL CALLBACK PathToolDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 HWND hctl;
 BOOL err;
 RECT rc;
 static int xpos=-1,ypos=-1;
 switch( msg ) {
   case WM_INITDIALOG:
     if(gToolFlags.flagP == 0)
       SendDlgItemMessage(hdlg,DLG_PATHTOOL_MOVE,BM_SETCHECK,TRUE,0);
     else if(gToolFlags.flagP == 1)
       SendDlgItemMessage(hdlg,DLG_PATHTOOL_ADD,BM_SETCHECK,TRUE,0);
     else if(gToolFlags.flagP == 2)
       SendDlgItemMessage(hdlg,DLG_PATHTOOL_DELETE,BM_SETCHECK,TRUE,0);
     else if(gToolFlags.flagP == 3)
       SendDlgItemMessage(hdlg,DLG_PATHTOOL_TENSION,BM_SETCHECK,TRUE,0);
     if(xpos == -1 && ypos == -1){
       RECT rc1;
       GetWindowRect(hdlg,&rc1);
       GetWindowRect(ghwndTools,&rc);
       if(bToolpannelLeft)xpos=rc.right+5;
       else               xpos=rc.left-(rc1.right-rc1.left)-5;
       ypos=rc.top;
     }
     SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     return (TRUE);
   case WM_DESTROY:
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       default:
         break;
       case SC_CLOSE:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
         break;
     }
     break;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDCANCEL:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
         break;
       case DLG_PATHTOOL_MOVE:    gToolFlags.flagP=0; goto DRAWIT;
       case DLG_PATHTOOL_ADD:     gToolFlags.flagP=1; goto DRAWIT;
       case DLG_PATHTOOL_DELETE:  gToolFlags.flagP=2; goto DRAWIT;
       case DLG_PATHTOOL_TENSION: gToolFlags.flagP=3;
       DRAWIT:
       DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,EDITING);
       SelectedPathPoint=NULL;
       DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,EDITING);
       default:   break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

BOOL CALLBACK RotatorToolDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 HWND hctl;
 BOOL err;
 RECT rc;
 static int xpos=-1,ypos=-1;
 switch( msg ) {
   case WM_INITDIALOG:
     if(gToolFlags.flagA == 0)
       SendDlgItemMessage(hdlg,DLG_ROTATOR_X,BM_SETCHECK,TRUE,0);
     else if(gToolFlags.flagA == 1)
       SendDlgItemMessage(hdlg,DLG_ROTATOR_Y,BM_SETCHECK,TRUE,0);
     else if(gToolFlags.flagA == 2)
       SendDlgItemMessage(hdlg,DLG_ROTATOR_Z,BM_SETCHECK,TRUE,0);
     if(xpos == -1 && ypos == -1){
       RECT rc1;
       GetWindowRect(hdlg,&rc1);
       GetWindowRect(ghwndTools,&rc);
       if(bToolpannelLeft)xpos=rc.right+5;
       else               xpos=rc.left-(rc1.right-rc1.left)-5;
       ypos=rc.top;
     }
//     SetClassLong(hdlg,GCL_HICON,(LONG)LoadIcon(ghinst_main,"ANIMATORICON"));
     SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     return (TRUE);
   case WM_DESTROY:
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       default:
         break;
       case SC_CLOSE:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
         break;
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDCANCEL:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
       break;
       case DLG_ROTATOR_X: gToolFlags.flagA=0; break;
       case DLG_ROTATOR_Y: gToolFlags.flagA=1; break;
       case DLG_ROTATOR_Z: gToolFlags.flagA=2; break;
       default:             break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

BOOL CALLBACK SizerToolDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 HWND hctl;
 BOOL err;
 RECT rc;
 static int xpos= -1,ypos= -1;
 switch( msg ) {
   case WM_INITDIALOG:
     SendDlgItemMessage(hdlg,DLG_SIZER_X,BM_SETCHECK,gToolFlags.flagX,0);
     SendDlgItemMessage(hdlg,DLG_SIZER_Y,BM_SETCHECK,gToolFlags.flagY,0);
     SendDlgItemMessage(hdlg,DLG_SIZER_Z,BM_SETCHECK,gToolFlags.flagZ,0);
     if(xpos == -1 && ypos == -1){
       RECT rc1;
       GetWindowRect(hdlg,&rc1);
       GetWindowRect(ghwndTools,&rc);
       if(bToolpannelLeft)xpos=rc.right+5;
       else               xpos=rc.left-(rc1.right-rc1.left)-5;
       ypos=rc.top;
     }
     SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     return (TRUE);
   case WM_DESTROY:
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       default:
         break;
       case SC_CLOSE:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
         break;
     }
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
        break;
        case DLG_SIZER_X: gToolFlags.flagX = !gToolFlags.flagX; break;
        case DLG_SIZER_Y: gToolFlags.flagY = !gToolFlags.flagY; break;
        case DLG_SIZER_Z: gToolFlags.flagZ = !gToolFlags.flagZ; break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

#define BITMAPX 36
#define BITMAPY 20

BOOL CALLBACK SelectActorDlgProc(HWND hDlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 static node *np;
 static HBITMAP hbmpPicture,hbmpOld;
 static HBITMAP hbmpImages[N_OBJECTTYPELIST];
 HWND hListBox;
 TEXTMETRIC tm;
 int i,y,nItem;
 int tc;
 node *nnp;
 HDC hdcMem;
 LPMEASUREITEMSTRUCT lpmis;
 LPDRAWITEMSTRUCT lpdis;
 RECT rcBitmap;
 TCHAR tchBuf[MAX_PATH];
 switch( msg ) {
   case WM_INITDIALOG:
     np=(node *)lparam;
//     SetWindowLong(hDlg,GWL_USERDATA,lparam);
     for(i=0;i<N_OBJECTTYPELIST;i++){
       hbmpImages[i]=LoadBitmap(ghinst_main,
         MAKEINTRESOURCE(ObjectBitmapIdList[i]));
     }
     hListBox=GetDlgItem(hDlg,DLG_ADDACTOR_LISTBOX);
     for(i=0;i<N_OBJECTTYPELIST;i++){
       nItem=SendMessage(hListBox,LB_ADDSTRING,0,(LPARAM)ObjectTypeList[i]);
       SendMessage(hListBox,LB_SETITEMDATA,nItem,(LPARAM)hbmpImages[i]);
     }
     SetFocus(hListBox);
     SendMessage(hListBox,LB_SETCURSEL,0,0);
     tc=1;
     nnp=FirstNp; while(nnp != NULL){
       if(nnp->type == ObjectTypeListOrder[0])tc++;
       nnp=nnp->next;
     }
     sprintf(tchBuf,"%s%d",ObjectTypeList[0],tc);
     SetDlgItemText(hDlg,DLG_ADDACTOR_NAME,(LPCSTR)tchBuf);
     SendDlgItemMessage(hDlg,DLG_ADDACTOR_NAME,CB_LIMITTEXT,10,0);
     CentreDialogOnCursor(hDlg);
     return TRUE;
   case WM_MEASUREITEM:
     lpmis=(LPMEASUREITEMSTRUCT)lparam;
     lpmis->itemHeight=BITMAPY+2;  // use bigger number for extra spacing
     return TRUE;
   case WM_DRAWITEM:
     lpdis=(LPDRAWITEMSTRUCT)lparam;
     if(lpdis->itemID == -1)break;
     switch(lpdis->itemAction){
       case ODA_SELECT:
       case ODA_DRAWENTIRE:
         hbmpPicture=(HBITMAP)SendMessage(lpdis->hwndItem,LB_GETITEMDATA,
          lpdis->itemID,(LPARAM)0);
         hdcMem=CreateCompatibleDC(lpdis->hDC);
         hbmpOld=SelectObject(hdcMem,hbmpPicture);
         BitBlt(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                lpdis->rcItem.right-lpdis->rcItem.left,
                lpdis->rcItem.bottom-lpdis->rcItem.top,
                hdcMem,0,0,SRCCOPY);
         SendMessage(lpdis->hwndItem,LB_GETTEXT,lpdis->itemID,(LPARAM)tchBuf);
         GetTextMetrics(lpdis->hDC,&tm);
         y=(lpdis->rcItem.bottom+lpdis->rcItem.top-tm.tmHeight)/2;
         TextOut(lpdis->hDC,BITMAPX+6,y,tchBuf,strlen(tchBuf));
         SelectObject(hdcMem,hbmpOld);
         DeleteDC(hdcMem);
         if(lpdis->itemState & ODS_SELECTED){
           rcBitmap.left=lpdis->rcItem.left;
           rcBitmap.top=lpdis->rcItem.top;
           rcBitmap.right=lpdis->rcItem.left+BITMAPX;
           rcBitmap.bottom=lpdis->rcItem.top+BITMAPY;
           InvertRect(lpdis->hDC,&rcBitmap);
         }
         break;
       case ODA_FOCUS:
         break;
     }
     return TRUE;
   case WM_PAINT:
     PaintDialogBackground(hDlg,ghinst_main);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case DLG_ADDACTOR_LISTBOX:
         switch(HIWORD(wparam)){
           case LBN_DBLCLK:
             SendMessage(hDlg,WM_COMMAND,MAKEWPARAM(DLG_ADDACTOR_OK,0),0);
             break;
           case LBN_SELCHANGE:
             i=SendDlgItemMessage(hDlg,DLG_ADDACTOR_LISTBOX,LB_GETCURSEL,0,0);
             if(i == GROUND || i == 7 || i == 11) // Sky + director
               sprintf(tchBuf,"%s",ObjectTypeList[i]);
             else{
               tc=1;
               nnp=FirstNp; while(nnp != NULL){
                 if(nnp->type == ObjectTypeListOrder[i])tc++;
                 nnp=nnp->next;
               }
               sprintf(tchBuf,"%s%d",ObjectTypeList[i],tc);
             }
             SetDlgItemText(hDlg,DLG_ADDACTOR_NAME,(LPCSTR)tchBuf);
             break;
           default:
             break;
         }
         break;
       case IDOK:
       case DLG_ADDACTOR_OK:
         i=SendDlgItemMessage(hDlg,DLG_ADDACTOR_LISTBOX,LB_GETCURSEL,0,0);
//         np=(node *)GetWindowLong(hDlg,GWL_USERDATA);
#if __LITE__
         if(i == 5 || i == 8 || i == 10 ){
           MessageBox(NULL,"This object type is not available","Warning",MB_OK);
           break;
         }
#endif
         if(GetDlgItemText(hDlg,DLG_ADDACTOR_NAME,(LPSTR)tchBuf,10) == 0){
            EndDialog(hDlg,FAIL);
            return(TRUE);
         }
         if(ActorNameExists(hDlg,np,tchBuf))break;
         strcpy(np->actorname,tchBuf);
         EndDialog(hDlg,i);
         return(TRUE);
       case IDCANCEL:
       case DLG_ADDACTOR_CANCEL:
         EndDialog(hDlg,FAIL);
         return(TRUE);
       default:
         break;
     }
     break;
   case WM_DESTROY:
     for(i=0;i<N_OBJECTTYPELIST;i++)DeleteObject(hbmpImages[i]);
     return TRUE;
   default: break;
 }
 return(FALSE);
}

static void UpdateframesDlgProc(HWND hdlg, int lframe, int *action){
 BOOL status;
 if(*action < 0)*action=0;
 if(lframe == Nframes){
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_EXTEND),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_BLANKS),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_SCALE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_CHOP),FALSE);
   *action= -1;
 }
 else if(lframe > Nframes){
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_EXTEND),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_BLANKS),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_SCALE),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_CHOP),FALSE);
   if(*action == 3)*action=0;
 }
 else if(lframe < Nframes){
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_EXTEND),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_BLANKS),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_SCALE),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_SETFRAMES_CHOP),TRUE);
   if(*action == 0 || *action == 1)*action=3;
 }
 if(*action == 0)status=TRUE; else status=FALSE;
 SendDlgItemMessage(hdlg,DLG_SETFRAMES_EXTEND,BM_SETCHECK,(WPARAM)status,0);
 if(*action == 1)status=TRUE; else status=FALSE;
 SendDlgItemMessage(hdlg,DLG_SETFRAMES_BLANKS,BM_SETCHECK,(WPARAM)status,0);
 if(*action == 2)status=TRUE; else status=FALSE;
 SendDlgItemMessage(hdlg,DLG_SETFRAMES_SCALE,BM_SETCHECK,(WPARAM)status,0);
 if(*action == 3)status=TRUE; else status=FALSE;
 SendDlgItemMessage(hdlg,DLG_SETFRAMES_CHOP,BM_SETCHECK,(WPARAM)status,0);
 return;
}

BOOL CALLBACK SetFramesDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL err;
 RECT rc;
 char str[16];
 static int xpos= -1,ypos= -1,action,lframe;
 static BOOL inuse=FALSE;
 int *i,k;
 switch( msg ) {
   case WM_INITDIALOG:{
       char colist0[]={"              "};
       char colist1[]={"              "};
       char colist2[]={"              "};
       LoadString(ghinst_main,IDX_MISC_FFNTSC,colist0,256);
       LoadString(ghinst_main,IDX_MISC_FFPAL,colist1,256);
       LoadString(ghinst_main,IDX_MISC_FFFPS,colist2,256);
       SetWindowLong(hdlg,GWL_USERDATA,lparam);
       SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_RESETCONTENT,0,0);
       SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_ADDSTRING,
                            0,(LPARAM)colist0);
       SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_ADDSTRING,
                            0,(LPARAM)colist1);
       SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_ADDSTRING,
                            0,(LPARAM)colist2);
       if     (FramesPerSec == 30)
         SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_SELECTSTRING,
                            0,(LPARAM)colist0);
       else if(FramesPerSec == 25)
         SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_SELECTSTRING,
                            0,(LPARAM)colist1);
       else if(FramesPerSec == 15)
         SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_SELECTSTRING,
                            0,(LPARAM)colist2);
       SetDlgItemInt(hdlg,DLG_SETFRAMES_EDIT,Nframes,FALSE);
       SendDlgItemMessage(hdlg,DLG_SETFRAMES_SPIN,SPNM_SETRANGE,0,
                           MAKELPARAM(1,10000));
       SendDlgItemMessage(hdlg,DLG_SETFRAMES_SPIN,SPNM_SETCRNTVALUE,
                          (WPARAM)Nframes,0);
       SendDlgItemMessage(hdlg,DLG_SETFRAMES_SPIN,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hdlg,DLG_SETFRAMES_EDIT));
       sprintf(str,"%.2lf",(double)Nframes/(double)FramesPerSec);
       inuse=TRUE;
       SetDlgItemText(hdlg,DLG_SETFRAMES_TIME,str);
       inuse=FALSE;
       action= -1; lframe=Nframes;
       UpdateframesDlgProc(hdlg,lframe,&action);
       if(xpos == -1 && ypos == -1)CentreDialogOnCursor(hdlg);
       else SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     }
     return (TRUE);
   case WM_DESTROY:
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     break;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_COMMAND:
     switch( HIWORD(wparam) ){
       case CBN_SELENDOK:
         if(LOWORD(wparam) == DLG_SETFRAMES_TYPE){
           k=SendDlgItemMessage(hdlg,DLG_SETFRAMES_TYPE,CB_GETCURSEL,0,0);
           if(k == 0)FramesPerSec=30;
           if(k == 1)FramesPerSec=25;
           if(k == 2)FramesPerSec=15;
           sprintf(str,"%.2lf",(double)lframe/(double)FramesPerSec);
           inuse=TRUE;
           SetDlgItemText(hdlg,DLG_SETFRAMES_TIME,str);
           inuse=FALSE;
         }
         break;
       default:
         break;
     }
     switch(LOWORD(wparam)){
       case DLG_SETFRAMES_EDIT:
         if(HIWORD(wparam) == EN_UPDATE && !inuse){
           lframe=GetDlgItemInt(hdlg,DLG_SETFRAMES_EDIT,&err,FALSE);
           if(err && lframe > 0){
             UpdateframesDlgProc(hdlg,lframe,&action);
             sprintf(str,"%.2lf",(double)lframe/(double)FramesPerSec);
             inuse=TRUE;
             SetDlgItemText(hdlg,DLG_SETFRAMES_TIME,str);
             inuse=FALSE;
           }
           else MessageBeep(MB_OK);
           break;
         }
         break;
       case DLG_SETFRAMES_TIME:
         if(HIWORD(wparam) == EN_UPDATE && !inuse){
           GetDlgItemText(hdlg,DLG_SETFRAMES_TIME,str,15);
           lframe=(long)(atof(str)*(double)FramesPerSec);
           if(lframe > 0){
             UpdateframesDlgProc(hdlg,lframe,&action);
             inuse=TRUE;
             SetDlgItemInt(hdlg,DLG_SETFRAMES_EDIT,lframe,FALSE);
             SendDlgItemMessage(hdlg,DLG_SETFRAMES_SPIN,SPNM_SETCRNTVALUE,
                          (WPARAM)lframe,0);
             inuse=FALSE;
           }
           else MessageBeep(MB_OK);
           break;
         }
         break;
       case DLG_SETFRAMES_EXTEND: action=0; break;
       case DLG_SETFRAMES_BLANKS: action=1; break;
       case DLG_SETFRAMES_SCALE:  action=2; break;
       case DLG_SETFRAMES_CHOP:   action=3; break;
       case IDOK:
       case DLG_SETFRAMES_OK:
         i=(int *)GetWindowLong(hdlg,GWL_USERDATA);
         if(action < 0) *i=0;
         else if(action == 0 || action == 3) *i=0;
         else if(action == 1) *i=2;
         else if(action == 2) *i=1;
         EndDialog(hdlg,lframe);
         return(TRUE);
       case IDCANCEL:
       case DLG_SETFRAMES_CANCEL:
         EndDialog(hdlg,-1);
         return(TRUE);
       default:  break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK AddKeyframeDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDOK:
        case DLG_KEYMAKER_GOTO:
          EndDialog(hdlg,1);
          return(TRUE);
          break;
        case DLG_KEYMAKER_MAKE:
          EndDialog(hdlg,2);
          return(TRUE);
          break;
        case IDCANCEL:
        case DLG_KEYMAKER_CANCEL:
          EndDialog(hdlg,0);
          return(TRUE);
          break;
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

int SelectKeyMaker(void){
 int id;
 EnableToolPannels(ALL_PANNELS,FALSE);
 id=DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_KEYMAKER),
              ghwnd_main,(DLGPROC)AddKeyframeDlgProc);
 EnableToolPannels(ALL_PANNELS,TRUE);
 return id;
}

BOOL CALLBACK ObjectTimelineDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 char line1[256],line2[256],line3[256];
 struct NP {node *np; short ff,lf;} *NP;
 sky    *sp;
 object *op;
 char title[255];
 switch( msg ) {
   case WM_INITDIALOG:
     NP=(struct NP *)lparam;
     line1[0]=0; line2[0]=0; line3[0]=0;
     if(NP->np->type == SKY){
       LoadString(ghinst_main,IDX_MISC_DSKYINFO,line2,256);
       strcpy(line1,line2);
       SkyCostumeInfo(NP->np,(short)(NP->lf),line2);
     }
     else if(NP->np->type == DIRECTOR){
       LoadString(ghinst_main,IDX_MISC_DIRECTORINFO,line2,256);
       strcpy(line1,line2);
       DirectorCostumeInfo(NP->np,(short)(NP->lf),line2);
     }
     else{
       op=GetObjectAtFrame(NP->np,(long)(NP->lf));
       if(op == NULL){
         MessageBeep(MB_OK); /* should never happen */
         SendMessage(hdlg,WM_COMMAND,IDCANCEL,0);
       }
       CostumeInfo(op,line1,line2,line3);
     }
     LoadString(ghinst_main,IDX_MISC_DCOSTUME,res_str,256);
     sprintf(title,res_str,NP->np->actorname);
     SendMessage(hdlg,WM_SETTEXT,0,(LPARAM)title);
     EnableToolPannels(ALL_PANNELS,FALSE);
     SendDlgItemMessage(hdlg,DLG_LINEOPT_T1,WM_SETTEXT,0,(LPARAM)line1);
     SendDlgItemMessage(hdlg,DLG_LINEOPT_T2,WM_SETTEXT,0,(LPARAM)line2);
     SendDlgItemMessage(hdlg,DLG_LINEOPT_T3,WM_SETTEXT,0,(LPARAM)line3);
     if(NP->np->type == CSG ||
        NP->np->type == PATH ||
        NP->np->type == DIRECTOR ||
        NP->np->type == SKY){
       EnableWindow(GetDlgItem(hdlg,DLG_LINEOPT_CUT),FALSE);
       EnableWindow(GetDlgItem(hdlg,DLG_LINEOPT_COPY),FALSE);
     }
     if(NP->np->type == ROBOT ||
        NP->np->type == TARGET || NP->np->type == PATH){
       EnableWindow(GetDlgItem(hdlg,DLG_LINEOPT_EDIT),FALSE);
     }
     CentreDialogOnScreen(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_LINEOPT_CUT:    EndDialog(hdlg,3); return TRUE;
        case DLG_LINEOPT_COPY:   EndDialog(hdlg,4); return TRUE;
        case DLG_LINEOPT_EDIT:   EndDialog(hdlg,1); return TRUE;
        case DLG_LINEOPT_DELETE: EndDialog(hdlg,0); return TRUE;
        case IDCANCEL:
        case DLG_LINEOPT_CANCEL:
          EndDialog(hdlg,-1);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK PathSetDlgProc(HWND hdlg,UINT msg,
                             WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     CentreDialogOnScreen(hdlg);
     EnableToolPannels(ALL_PANNELS,FALSE);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PATHTYPE_OPEN:
          EndDialog(hdlg,0);
          return(TRUE);
          break;
        case DLG_PATHTYPE_CLOSED:
          EndDialog(hdlg,1);
          return(TRUE);
          break;
        case DLG_PATHTYPE_LOAD:
          EndDialog(hdlg,2);
          return(TRUE);
          break;
        case DLG_PATHTYPE_COPY:
          EndDialog(hdlg,3);
          return(TRUE);
          break;
        case IDCANCEL:
          EndDialog(hdlg,-1);
          return(TRUE);
          break;
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK CsgSetDlgProc(HWND hdlg,UINT msg,
                             WPARAM wparam,LPARAM lparam){
 static object *op,*tp;
 LPDRAWITEMSTRUCT lpdis;
 BOOL err;
 struct NC {node *np; object *op,*tp; int d;} *Nc;
 int i,j,lf,ff;
 char title[256];
 long range1,range2;
 switch( msg ) {
   case WM_INITDIALOG:
     Nc=(struct NC *)lparam;
     memcpy(Nc->tp,Nc->op,sizeof(object));
     op=Nc->tp;  tp=Nc->op;
     LoadString(ghinst_main,IDX_MISC_DCSGSETUP,res_str,256);
     sprintf(title,res_str,Nc->np->actorname);
     SendMessage(hdlg,WM_SETTEXT,0,(LPARAM)title);
     SetDlgItemInt(hdlg,DLG_CSG_FROM,op->firstframe,FALSE);
     SetDlgItemInt(hdlg,DLG_CSG_TO,op->lastframe,FALSE);
     if(op->last != NULL)ff=op->last->lastframe+1;
     else                ff=1;
     if(op->next != NULL)lf=op->next->firstframe-1;
     else                lf=Nframes;
     SendDlgItemMessage(hdlg,DLG_CSG_FROMS,SPNM_SETRANGE,0,
                        MAKELPARAM(ff,lf));
     SendDlgItemMessage(hdlg,DLG_CSG_FROMS,SPNM_SETCRNTVALUE,
                        (WPARAM)op->firstframe,0);
     SendDlgItemMessage(hdlg,DLG_CSG_FROMS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hdlg,DLG_CSG_FROM));
     SendDlgItemMessage(hdlg,DLG_CSG_TOS,SPNM_SETRANGE,0,
                        MAKELPARAM(ff,lf));
     SendDlgItemMessage(hdlg,DLG_CSG_TOS,SPNM_SETCRNTVALUE,
                        (WPARAM)op->lastframe,0);
     SendDlgItemMessage(hdlg,DLG_CSG_TOS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hdlg,DLG_CSG_TO));
     if(op->lighttype == SPHERE)
       SendDlgItemMessage(hdlg,DLG_CSG_SPHERE,BM_SETCHECK,TRUE,0);
     if(op->lighttype == TUBE)
       SendDlgItemMessage(hdlg,DLG_CSG_TUBE,BM_SETCHECK,TRUE,0);
     if(op->lighttype == CYLIND)
       SendDlgItemMessage(hdlg,DLG_CSG_CYLINDER,BM_SETCHECK,TRUE,0);
     if(op->lighttype == DISK)
       SendDlgItemMessage(hdlg,DLG_CSG_DISK,BM_SETCHECK,TRUE,0);
     if(op->groundtype == 0)
       SendDlgItemMessage(hdlg,DLG_CSG_ADD,BM_SETCHECK,TRUE,0);
     if(op->groundtype == 1)
       SendDlgItemMessage(hdlg,DLG_CSG_SUB,BM_SETCHECK,TRUE,0);
     if(Nc->d == 0)EnableWindow(GetDlgItem(hdlg,DLG_CSG_DELETE),FALSE);
     CentreDialogOnScreen(hdlg);
     EnableToolPannels(ALL_PANNELS,FALSE);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_CSG_COLOUR){
         r=op->colour[0]; g=op->colour[1]; b=op->colour[2];
         if(lpdis->itemState & ODS_SELECTED)InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_COMMAND:
       /* severe warning about using GWL_USERDATA */
      switch(LOWORD(wparam)){
        case DLG_CSG_COLOUR:
          SetSfxColour(op->colour,IDX_MISC_CSG_COLOUR,hdlg);
          InvalidateRect(GetDlgItem(hdlg,DLG_CSG_COLOUR),NULL,FALSE);
          break;
        case DLG_CSG_SPHERE:
          op->lighttype=SPHERE;
          break;
        case DLG_CSG_TUBE:
          op->lighttype=TUBE;
          break;
        case DLG_CSG_CYLINDER:
          op->lighttype=CYLIND;
          break;
        case DLG_CSG_DISK:
          op->lighttype=DISK;
          break;
        case DLG_CSG_ADD:
           op->groundtype=0;
           break;
        case DLG_CSG_SUB:
           op->groundtype=1;
           break;
        case DLG_CSG_DELETE:
          EndDialog(hdlg,1);
          return(TRUE);
        case IDOK:
          range1=SendDlgItemMessage(hdlg,DLG_CSG_FROMS,SPNM_GETRANGE,0,0);
          range2=SendDlgItemMessage(hdlg,DLG_CSG_TOS,SPNM_GETRANGE,0,0);
          i=GetDlgItemInt(hdlg,DLG_CSG_FROM,&err,FALSE);
          if(!err || i < LOWORD(range1) ||
                     i > HIWORD(range1))goto CANCELX;
          j=GetDlgItemInt(hdlg,DLG_CSG_TO,&err,FALSE);
          if(!err || j < LOWORD(range2) ||
                     j > HIWORD(range2))goto CANCELX;
          if(j < i)goto CANCELX;
          op->firstframe=i;
          op->lastframe=j;
          memcpy(tp,op,sizeof(object));
          EndDialog(hdlg,0);
          return(TRUE);
          CANCELX:
          MessageBeep(MB_OK);
        case IDCANCEL:
          EndDialog(hdlg,-1);
          return(TRUE);
          break;
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

// The popup dialog that appears when the Quick and Wireframe previews
// are being drawn

extern UINT AnimationTimer;

static BOOL CALLBACK QuickPlayDlgProc(HWND hwnd,UINT msg,
                                WPARAM wparam,LPARAM lparam){
 RECT rc;
 POINT p;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     if(View == ONEVIEW){
       GetWindowRect(hwnd,&rc);
       if(BigCameraView){
         p.x=WindowViewSliderXpos+WindowViewSliderXsize+15;
         p.y=WindowViewSliderYpos-15;
         p.y -= (rc.bottom - rc.top);
       }
       else{
         p.x=WindowViewSliderXpos+WindowViewSliderXsize;
         p.x -= (rc.right - rc.left);
         p.y=WindowViewSliderYpos+WindowViewSliderYsize;
       }
       ClientToScreen(ghwnd_main,&p);
       p.x -= 5; p.y += 10;
     }
     else{
       GetWindowRect(hwnd,&rc);
       p.x=WindowViewSliderXpos+WindowViewSliderXsize;
       p.x -= (rc.right - rc.left);
       p.y=WindowViewSliderYpos+WindowViewSliderYsize;
       ClientToScreen(ghwnd_main,&p);
       p.x -= 5; p.y += 10;
     }
     SetWindowPos(hwnd,NULL,p.x,p.y,0,0,SWP_NOSIZE | SWP_NOZORDER);
     return (TRUE);
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDCANCEL:
         EndDialog(hwnd,IDCANCEL);
         return(TRUE);
       case IDOK:
         EndDialog(hwnd,IDOK);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

void AnimationPlayer(int type, HWND hwnd){
 static int AnimationStoreFrame;
 int i,b;
 if((AnimationTimer=SetTimer(ghwnd_view,type+1,33,NULL)) == 0)return;
 AnimationStoreFrame=CurrentFrame;
 EnableToolPannels(ALL_PANNELS,FALSE);
 if(DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_QUICKPLAY),hwnd,
     (DLGPROC)QuickPlayDlgProc) == IDOK)b=0;
 else b=1;
 KillTimer(ghwnd_view,AnimationTimer); AnimationTimer=0;
 if(b)CurrentFrame=AnimationStoreFrame;
 else UpdateSelectedActor(FALSE);
// SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
 SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
 ReDrawStageDisplay(TRUE);
 if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
 else UpdateGLview(TRUE);
 EnableToolPannels(ALL_PANNELS,TRUE);
}
