/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file FILES.C */

// Utilitiy functions to provide file open/save dialogs. Loads strings
// from resources files to allow for internationalisation

#define MODULE_FILES 1

#include "animate.h"

int SelectFileName(int save, char *szfile, char *szdir,
                   char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[256],szdummy[MAX_FILE];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=MAX_FILE;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=MAX_FILE;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
   {MSG msg;  /* allow WM_COMMAND message to be processed */
    while(PeekMessage(&msg,NULL,0,0,PM_REMOVE)){
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
   }

 return i;
}

int SelectDirectoryName(char *szfile, char *szdir,
                   char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[80],szdummy[MAX_FILE];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=MAX_FILE;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=MAX_FILE;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 ofn.hInstance=ghinst_main;
 ofn.lpfnHook=NULL;
 ofn.lpTemplateName="DIRECTORYOPEN";
 ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|
            OFN_ENABLETEMPLATE|OFN_NONETWORKBUTTON|
            OFN_NOTESTFILECREATE;  /* for PAR problem */
 i=(int)(GetSaveFileName(&ofn));
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

int SelectSfxFileName(int save, char *szfile, char *szdir, long ID,
                      char *szfilter, HWND parent){
 char szTitle[256];
 LoadString(ghinst_main,ID,szTitle,255);
 return SelectFileName(save,szfile,szdir,szTitle,szfilter,parent);
}

static int MorphSetupValue;

static unsigned CALLBACK MorphFileOpenHookProc(HWND hwnd,UINT msg,
                              WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     MorphSetupValue=1;
     SendDlgItemMessage(hwnd,100,BM_SETCHECK,1,0);
     return FALSE;
   case WM_COMMAND:
    switch(LOWORD(wparam)){
      case 100: MorphSetupValue=1; break;
      case 101: MorphSetupValue=2; break;
      case 102: MorphSetupValue=3; break;
      default: break;
    }
    break;
   case WM_NOTIFY:{
       LPOFNOTIFY lp=(LPOFNOTIFY)lparam;
       if(lp->hdr.code == CDN_FILEOK){
//       LPNMHDR lp=(LPNMHDR)lparam;
//       if(lp->code == CDN_FILEOK){
         ;
       }
     }
     break;
   default: break;
 }
 return FALSE;
}

int SelectMorphFileName(char *szfile, char *szdir,
                   char *szTitle, char *szfilter, HWND parent){
 int i;
 DWORD version;
 version=GetVersion();
// if(((version & 0x80000000) == 0x80000000) &&
//     (LOBYTE(LOWORD(version)) >= 4)){ /* windows 95 */
//   i=SelectFileName(0,szfile,szdir,szTitle,szfilter,parent);
//   if(i > 0)return 4;
//   else     return i;
// }
 {
 OPENFILENAME ofn;
 char szFilter[80],szdummy[MAX_FILE];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=MAX_FILE;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=MAX_FILE;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 ofn.hInstance=ghinst_main;
 ofn.lpfnHook=MorphFileOpenHookProc;
 ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|
           OFN_ENABLEHOOK|OFN_ENABLETEMPLATE|OFN_NONETWORKBUTTON;
 if(LOBYTE(LOWORD(version)) >= 4){
   ofn.lpTemplateName="FILEOPEN95";
   ofn.Flags |= OFN_EXPLORER;
 }
 else{
   ofn.lpTemplateName="FILEOPEN";
 }
 i=(int)(GetOpenFileName(&ofn));
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
   i=MorphSetupValue;
 }
 return i;
 }
}

