/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file keyframe.c */

// The functions in this file handles all the "keyframer" interactions
// and quite a lot of its functions. Since the keyframer is used to
// create time lines and keyframes quite a few of the main program's
// commands call through to these functions.

#define MODULE_SCRIPT 1

#include "animate.h"
#include "spin_dll.h"

static short CreatePathObject(short action, node *Np, object *tempobj,
                       long firstframe, long lastframe);
static short SetupTimeLineEndPoints(short current, short *ff, short *lf,
                                    short *type,short line, HWND parent);
static void SetFollowAtOptions(position *, HWND);
static short EditEndPoints(short type, object *vp, position *pp, align *ap);
static void ChangeSizeEndPoints(node *Np, short frame, short code);
static void ChangeAlignEndPoints(node *, short, short, HWND, int);
static void ChangePositionEndPoints(node *, short, short, HWND ,int);
static short EditObjectTimeline(node *Np, short frame, short code,
                               short in_script);
static void CopyObject(node *Np, short firstframe, short lastframe);
static void CutOutObject(node *Np, short frame);
static void CutOutPosition(node *Np, short frame);
static void CutOutAlign(node *Np, short frame);
static node *LoadStandinModel(short *nmodels, char *title, char *command);
static LRESULT CALLBACK TimelineEditorWndProc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam );
static LRESULT CALLBACK TimelineChildWndProc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam );
static LRESULT CALLBACK pTimeLineEditDlgProc(HWND hwnd, UINT msg,
                                      WPARAM wparam, LPARAM lparam);
static LRESULT CALLBACK rTimeLineEditDlgProc(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam);
static LRESULT CALLBACK DlgCommandDlgProc(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam);
static LRESULT CALLBACK ReorderDlgProc(HWND hwnd, UINT msg,
                                       WPARAM wParam, LPARAM lParam);
static int GetPresets(int nPresets);
static BOOL BuildFromPreset(int action);
static void FreePresets(void);

static point Cube[8]={ UNIT, UNIT,-UNIT,
                      -UNIT, UNIT,-UNIT,
                      -UNIT,-UNIT,-UNIT,
                       UNIT,-UNIT,-UNIT,
                       UNIT, UNIT, UNIT,
                      -UNIT, UNIT, UNIT,
                      -UNIT,-UNIT, UNIT,
                       UNIT,-UNIT, UNIT};
static point Disk[8]={ UNIT, UNIT,-UNIT/10,
                      -UNIT, UNIT,-UNIT/10,
                      -UNIT,-UNIT,-UNIT/10,
                       UNIT,-UNIT,-UNIT/10,
                       UNIT, UNIT, UNIT/10,
                      -UNIT, UNIT, UNIT/10,
                      -UNIT,-UNIT, UNIT/10,
                       UNIT,-UNIT, UNIT/10};

// script part 1

node *AddOnlyNewNode(HWND hWnd, long action){
 node   *Np;
 object *Op;
#if __DEMO__
 if(Nnodes >= 10){
   MessageBox(NULL,"Only 10 actors allowed in demo","Demo",MB_OK);
   return NULL;
 }
#endif
 if((Np=CreateNode()) != NULL){
   if(action < 0){
     EnableToolPannels(ALL_PANNELS,FALSE);
     action=(long)DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_ADDACTOR),hWnd,
       (DLGPROC)SelectActorDlgProc,(LPARAM)Np);
     EnableToolPannels(ALL_PANNELS,TRUE);
   }
   if(action == FAIL)strcpy(Np->actorname,"FAIL"); /* dumy text */
   if((action == GROUND && Nground > 0) ||
      (action == 7 && Nskys > 0) ||
      (action == 11 && Ndirectors > 0)
     ){
      SendPrgmQuery(IDQ_ONLYONEALLOWED,0);
      action = FAIL;
   }
   if(action == GROUND)Nground=1;
   if(action == 8)Nrobots++;
   if(action == 7)Nskys=1;   /* sky */
   if(action == 11)Ndirectors=1;
   if(action < 0){
     DeleteNode(Np);
     Np=NULL;
   }
   else if(action == 6)Np->type = CSG;
   else if(action == 7)Np->type = SKY;
   else if(action == 8)Np->type = ROBOT;
   else if(action == 9){
     Np->type = CAMERA;
     if((Op=CreateActor(Np,1,(short)Nframes)) == NULL){
       DeleteNode(Np);
       Np=NULL;
     }
     else{
       Ncameras++;
       Op->type=Np->type;
       Np->fobj->morph=NO;
     }
   }
   else if(action == 10)Np->type = IMAGEP;
   else if(action == 11)Np->type = DIRECTOR;
   else  Np->type = action;
 }
 return Np;
}

void AddQuickActor(long command){
  long action;
  node *local_node;
  Save_UndoA();
  if(command < 0){
   local_node=AddOnlyNewNode(ghwnd_main,-1);
  }
  else{
    action=NORMAL;
    switch (command){
     case IDM_ACTOR_QADD_MOD:      action=NORMAL;  break;
     case IDM_ACTOR_QADD_MORPH:    action=ANIMOBJ; break;
     case IDM_ACTOR_QADD_CAMERA:   action=9;       break;
     case IDM_ACTOR_QADD_TARGET:   action=TARGET;  break;
     case IDM_ACTOR_QADD_LIGHT:    action=LIGHT;   break;
     case IDM_ACTOR_QADD_PATH:     action=PATH;    break;
     case IDM_ACTOR_QADD_SKY:      action=7;       break;
     case IDM_ACTOR_QADD_GROUND:   action=GROUND;  break;
     case IDM_ACTOR_QADD_XIP:      action=10;      break;
     case IDM_ACTOR_QADD_ROBOT:    action=8;       break;
     case IDM_ACTOR_QADD_CSG:      action=6;       break;
     case IDM_ACTOR_QADD_DIRECTOR: action=11;      break;
    }
    local_node=AddOnlyNewNode(ghwnd_main,action);
    if(local_node != NULL){
      char tchBuf[64];
      long i,tc;
      strcpy(local_node->actorname,"Actor");
      if     (local_node->type == GROUND)
        sprintf(tchBuf,"%s",ObjectTypeList[3]);
      else if(local_node->type == SKY)
        sprintf(tchBuf,"%s",ObjectTypeList[7]);
      else if(local_node->type == DIRECTOR)
        sprintf(tchBuf,"%s",ObjectTypeList[11]);
      else{
        node *nnp;
        tc=1;
        nnp=FirstNp; while(nnp != NULL){
           if(local_node != nnp && local_node->type == nnp->type)tc++;
           nnp=nnp->next;
        }
        if     (local_node->type == NORMAL)i=0;
        else if(local_node->type == PATH)i=1;
        else if(local_node->type == TARGET)i=2;
        else if(local_node->type == LIGHT)i=4;
        else if(local_node->type == ANIMOBJ)i=5;
        else if(local_node->type == CSG)i=6;
        else if(local_node->type == ROBOT)i=8;
        else if(local_node->type == CAMERA)i=9;
        else if(local_node->type == IMAGEP)i=10;
        sprintf(tchBuf,"%s%d",ObjectTypeList[i],tc);
      }
      strcpy(local_node->actorname,tchBuf);
    }
  }
  if(local_node != NULL){
    position *localpp;
    if(local_node->type != SKY &&
       local_node->type != IMAGEP &&
       local_node->type != DIRECTOR &&
      (localpp=CreatePosition(local_node,1,Nframes)) != NULL){
      localpp->type=SPLINE;
      localpp->finish[0]=NpointerX;
      localpp->finish[1]=NpointerY;
      localpp->finish[2]=NpointerZ;
      if(local_node->type != TARGET){
        CreateAlign(local_node,1,Nframes);
         CreateSize(local_node,1,Nframes);
      }
    }
    if(local_node->type != CAMERA){
      if(AddCostumeTimeLine(local_node) == FAIL){
        if(local_node->type == GROUND)Nground=0;
        if(local_node->type == SKY)Nskys=0;
        if(local_node->type == DIRECTOR)Ndirectors=0;
        if(local_node->type == ROBOT)Nrobots--;
        DeleteNode(local_node);
        local_node=NULL;
      }
    }
    if( local_node != NULL &&
       (local_node->type == CAMERA || local_node->type == LIGHT)){
      /* add target for lights and cameras */
      object *Op;
      CreateNode();
//      strcpy(MainNp->actorname,"Target");
      sprintf(MainNp->actorname,"Target(%s)",local_node->actorname);
      MainNp->type=TARGET;
      Op=CreateActor(MainNp,1,(short)Nframes);
      Op->type=MainNp->type;
      local_node->fali->type=TRACK;
      local_node->fali->topath=MainNp;
      localpp=CreatePosition(MainNp,1,Nframes);
      MainNp->fpos->finish[0]=local_node->fpos->finish[0];
      MainNp->fpos->finish[1]=local_node->fpos->finish[1]+TVsizeY/4;
      MainNp->fpos->finish[2]=local_node->fpos->finish[2]-TVsizeZ/4;
      MainNp->fpos->type=SPLINE;
    }
    ReDrawStageDisplay(TRUE);
    if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
    else UpdateGLview(TRUE);
    UPDATETIMELINES(IDC_TIMEL_UPDATE)
  }
}


short GetTimeGapAtFrame(node *Np, long frame,
                        short *start_gap, short *end_gap){
 object   *Op,*Op1=NULL;
 sky      *Sp,*Sp1=NULL;
 director *Dp,*Dp1=NULL;
 if(Np == NULL)return FAIL;
 if(Np->type == SKY){
   if((Sp=Np->fsky) != NULL)while(Sp != NULL){
     if(frame < Sp->firstframe){
       *end_gap=Sp->firstframe-1;
       if(Sp->last != NULL)*start_gap=Sp->last->lastframe+1;
       else                *start_gap=1;
       return OK;
     }
     Sp1=Sp;
     Sp=Sp->next;
   }
   *end_gap=Nframes;
   if(Sp1 != NULL)*start_gap=Sp1->lastframe+1;
   else           *start_gap=1;
   return OK;
 }
 else if(Np->type == DIRECTOR){
   if((Dp=Np->fdirector) != NULL)while(Dp != NULL){
     if(frame < Dp->firstframe){
       *end_gap=Dp->firstframe-1;
       if(Dp->last != NULL)*start_gap=Dp->last->lastframe+1;
       else                *start_gap=1;
       return OK;
     }
     Dp1=Dp;
     Dp=Dp->next;
   }
   *end_gap=Nframes;
   if(Dp1 != NULL)*start_gap=Dp1->lastframe+1;
   else           *start_gap=1;
   return OK;
 }
 else{
   if((Op=Np->fobj) != NULL)while(Op != NULL){
     if(frame < Op->firstframe){
       *end_gap=Op->firstframe-1;
       if(Op->last != NULL)*start_gap=Op->last->lastframe+1;
       else                *start_gap=1;
       return OK;
     }
     Op1=Op;
     Op=Op->next;
   }
   *end_gap=Nframes;
   if(Op1 != NULL)*start_gap=Op1->lastframe+1;
   else           *start_gap=1;
   return OK;
 }
 return FAIL;
}

object *GetObjectAtFrame(node *Np, long frame){
 object   *Op;
 sky      *Sp;
 director *Dp;
 if(Np != NULL){
   if(Np->type == SKY){
     if((Sp=Np->fsky) != NULL)while(Sp != NULL){
       if(frame >= Sp->firstframe && frame <= Sp->lastframe)
         return (object *)Sp;
       Sp=Sp->next;
     }
   }
   else if(Np->type == DIRECTOR){
     if((Dp=Np->fdirector) != NULL)while(Dp != NULL){
       if(frame >= Dp->firstframe && frame <= Dp->lastframe)
         return (object *)Dp;
       Dp=Dp->next;
     }
   }
   else{
     if((Op=Np->fobj) != NULL)while(Op != NULL){
       if(frame >= Op->firstframe && frame <= Op->lastframe)return Op;
       Op=Op->next;
     }
   }
 }
 return NULL;
}

void GetTimelinesAtFrame(node *Np, long frame, object **o,
                        position **p, align **a, size **s){
 object *Op;
 position *Pp;
 align *Ap;
 size *Sp;
 if(Np == NULL)return;
 if(Np->type == SKY)return;
 if((Op=Np->fobj) != NULL)while(Op != NULL){
   if(frame >= Op->firstframe && frame <= Op->lastframe){
     *o = Op; break;
   }
   Op=Op->next;
 }
 if((Pp=Np->fpos) != NULL)while(Pp != NULL){
   if(frame >= Pp->firstframe && frame <= Pp->lastframe){
     *p = Pp; break;
   }
   Pp=Pp->next;
 }
 if((Ap=Np->fali) != NULL)while(Ap != NULL){
   if(frame >= Ap->firstframe && frame <= Ap->lastframe){
     *a = Ap; break;
   }
   Ap=Ap->next;
 }
 if((Sp=Np->fsiz) != NULL)while(Sp != NULL){
   if(frame >= Sp->firstframe && frame <= Sp->lastframe){
     *s = Sp; break;
   }
   Sp=Sp->next;
 }
 return;
}

short AddCostumeTimeLine(node *Np){
  short ff,lf,retval=FAIL;
  point TVC,TVP;
  if(Np != NULL){
    if(GetObjectAtFrame(Np,CurrentFrame) == NULL){
      if(GetTimeGapAtFrame(Np,CurrentFrame,&ff,&lf) == OK){
        retval=(short)EditObject(Np,ff,lf,NO,NO);
        if(Preferences.autocentre  &&
           !(GetAsyncKeyState(VK_CONTROL) & 0x8000)){
          get_centre_stage(TVC,TVP);
          TVpointX=TVP[0]; TVpointY=TVP[1]; TVpointZ=TVP[2];
          TVsizeX=2*(TVC[0]-TVP[0]);
          TVsizeY=2*(TVC[1]-TVP[1]);
          TVsizeZ=2*(TVC[2]-TVP[2]);
          TVcentreX=NpointerX=TVC[0];
          TVcentreY=NpointerY=TVC[1];
          TVcentreZ=NpointerZ=TVC[2];
          GetTriview(FALSE);  /* this will adjust the TV size */
        }
        if(retval == OK){
          ReDrawStageDisplay(TRUE);
          if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
          else UpdateGLview(TRUE);
        }
        return retval;
      }
      else{
        SendPrgmQuery(IDQ_NOADDTIMELINE,0);
      }
    }
    else if(Np->type == SKY || Np->type == IMAGEP)
      retval=EditObject(Np,(short)CurrentFrame,(short)CurrentFrame,NO,NO);
    else {
      SendPrgmQuery(IDQ_COSTUMEEXISTS,0);
    }
  }
  return retval;
}

short EditObject(node *Np, short firstframe, short lastframe,
                 short in_script, short split){
  /* action > 0 then use this as preset if editing */
  sky      *tempsky;
  director *tempdirector;
  object   *tempobj;
  short action,type,i,n,retval=OK;
  HWND hwnd;
  struct NP {node *np; short ff,lf;} NP;
  struct NC {node *np; object *op,*tp; int d;} NC;
  if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;

  if(Np->type == SKY){ /* edit SKY */
    tempsky = CreateSky(Np,firstframe,lastframe);
    if((tempsky == NULL) && fail_op == NO){
      NP.np=Np; NP.ff=firstframe; NP.lf=lastframe;
      action=(short)DialogBoxParam(ghinst_main,
                            MAKEINTRESOURCE(DLG_LINEOPT),hwnd,
                            (DLGPROC)ObjectTimelineDlgProc,(LPARAM)&NP);
      if(action == 0)DeleteSky(Np,firstframe);
      else if(action == 1){
        if((tempsky=Np->fsky) != NULL)while(tempsky != NULL){
          if(firstframe >= tempsky->firstframe &&
             firstframe <= tempsky->lastframe){
            EditSkyDialog(tempsky,in_script);
            break;
          }
          tempsky=tempsky->next;
        }
      }
    }
    else if(tempsky != NULL){ /* add a new sky */
      if(EditSkyDialog(tempsky,in_script) < 0){
        DeleteSky(Np,firstframe);
        retval=FAIL;
      }
    }
    fail_op=NO;
  }

  else if(Np->type == DIRECTOR){ /* edit DIRECTOR */
    tempdirector = CreateDirector(Np,firstframe,lastframe);
    if((tempdirector == NULL) && fail_op == NO){
      NP.np=Np; NP.ff=firstframe; NP.lf=lastframe;
      action=(short)DialogBoxParam(ghinst_main,
                            MAKEINTRESOURCE(DLG_LINEOPT),hwnd,
                            (DLGPROC)ObjectTimelineDlgProc,(LPARAM)&NP);
      if(action == 0)DeleteDirector(Np,firstframe);
      else if(action == 1){
        if((tempdirector=Np->fdirector) != NULL)while(tempdirector != NULL){
          if(firstframe >= tempdirector->firstframe &&
             firstframe <= tempdirector->lastframe){
            EditDirectorDialog(tempdirector,in_script);
            break;
          }
          tempdirector=tempdirector->next;
        }
      }
    }
    else if(tempdirector != NULL){ /* add a new director */
      if(EditDirectorDialog(tempdirector,in_script) < 0){
        DeleteDirector(Np,firstframe);
        retval=FAIL;
      }
    }
    fail_op=NO;
  }

  else if(Np->type != CAMERA){
    tempobj=CreateActor(Np,firstframe,lastframe);
    if(fail_op == YES)fail_op=NO; /* creation failed no memory or overlap */
    else if((tempobj == NULL) && (firstframe == lastframe)){/* edit needed */
      NP.np=Np; NP.ff=firstframe; NP.lf=lastframe;
      if(Np->type == IMAGEP){ /* use a single dialog to get the action */
        action=EditObjectTimeline(Np,firstframe,0,in_script);
      }
      else /* use a general purpose action */
        action=(short)DialogBoxParam(ghinst_main,
                         MAKEINTRESOURCE(DLG_LINEOPT),hwnd,
                         (DLGPROC)ObjectTimelineDlgProc,(LPARAM)&NP);
      if     (action == 0)DeleteActor(Np,firstframe);
      else if(action == 1){
        EditObjectTimeline(Np,firstframe,0,in_script);
      }
      else if(action == 3) /* delete bar and shift left */
        CutOutObject(Np,firstframe);
      else if(action == 4)  /* copy to end of time bar */
        CopyObject(Np,firstframe,lastframe);
    }
    else if(tempobj != NULL){/* add a new object */
     tempobj->type=Np->type;
     CheckRecursiveFollow(Np,firstframe,lastframe);
     if(Np->type == NORMAL || (Np->type == ANIMOBJ && tempobj->next != NULL)){
       if((i=SelectMorphFileName(gszSCEfile,gszSCEdir,"Select Model",
         "(*.MFX)|*.mfx|",hwnd)) > 0){
         strcpy(tempobj->name,gszSCEfile);
         if(i > 3){ /* used for default if previous then make morph */
           i=1; /* don't automatically morph in 95  */
         }
         tempobj->morph=(short)(i-1);
         if(tempobj->morph < 0){
           DeleteActor(Np,firstframe);
           retval=FAIL;
         }
         else if(StageLoadObject(tempobj->name,tempobj,YES,YES,NO) == FAIL){
           DeleteActor(Np,firstframe);
           retval=FAIL;
         }
       }
       else{
         DeleteActor(Np,firstframe);
         retval=FAIL;
       }
     }
     else if(Np->type == ANIMOBJ){
       DeleteActor(Np,firstframe); /* this single object is of no use */
       LoadAnimObject(Np,firstframe,lastframe,hwnd);
     }
     else if(Np->type == GROUND){
       if(EditGroundDialog(tempobj,in_script) < 0){
         DeleteActor(Np,firstframe);
         retval=FAIL;
       }
     }
     else if(Np->type == LIGHT){
       tempobj->colour[0]=255;  tempobj->colour[1]=255;
       tempobj->colour[2]=255;  tempobj->lighttype=SPHERE;
       if(EditLightDialog(tempobj,(int)in_script) < 0){
         DeleteActor(Np,firstframe);
         retval=FAIL;
       }
     }
     else if(Np->type == PATH){
       action=(short)DialogBox(ghinst_main,
                          MAKEINTRESOURCE(DLG_PATHTYPE),
                          hwnd,(DLGPROC)PathSetDlgProc);
       strcpy(tempobj->name,"PATH1995"); /* used to recognise new paths */
       if(action < OPEN){
         DeleteActor(Np,firstframe);
         retval=FAIL;
       }
       else  CreatePathObject(action,Np,tempobj,firstframe,lastframe);
     }
     else if(Np->type == CSG){
       object temp_obj;
       tempobj->lighttype=0; tempobj->groundtype=0;
       tempobj->colour[0]=255; tempobj->colour[1]=255; tempobj->colour[2]=0;
       NC.op=tempobj;  NC.tp=&temp_obj; NC.np=Np;  NC.d=0;
       action=(short)DialogBoxParam(ghinst_main,
                          MAKEINTRESOURCE(DLG_CSG),
                          hwnd,(DLGPROC)CsgSetDlgProc,(LPARAM)&NC);
       if(action < 0 || action == 1){
         DeleteActor(Np,firstframe);  /* escape pressed */
         retval=FAIL;
       }
       else{
         for(i=0;i<8;i++){
           if(tempobj->lighttype==DISK)CopyPoint(Disk[i],tempobj->outline[i]);
           else                        CopyPoint(Cube[i],tempobj->outline[i]);
         }
       }
     }
     else if(Np->type == TARGET){ ; }/* no need to load anything */
     else if(Np->type == IMAGEP){
       if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;
       if((tempobj->xip=EditObjectImageProcess(hwnd,tempobj->xip)) == NULL){
         DeleteActor(Np,firstframe);
         retval=FAIL;
       }
       tempobj->morph=1;
     }
     else if(Np->type == ROBOT){
       if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;
       if(split && tempobj != NULL && tempobj->next != NULL){
         strcpy(gszSCEfile,tempobj->next->name);  /* make same name */
         strcpy(gszSCEdir,tempobj->next->name);
         *FileInPath(gszSCEdir) = '\0';
       }
       if(SelectSfxFileName(0,gszSCEfile,gszSCEdir,IDX_MISC_SELECTROBOT,
          "(*.MFX)|*.mfx|",hwnd) == TRUE){
         strcpy(tempobj->name,gszSCEfile);
         tempobj->morph=0;
         if(tempobj->morph < 0){
           DeleteActor(Np,firstframe);
           retval=FAIL;
         }
         else{
           if(StageLoadObject(tempobj->name,tempobj,YES,YES,YES) == FAIL){
             DeleteActor(Np,firstframe);
             retval=FAIL;
           }
           else if(split){
             object *Opl,*Opn;
             if(tempobj != NULL && (Opl=tempobj->last) != NULL
                                && (Opn=tempobj->next) != NULL){
               double mr;
               mr=(double)(Opn->lastframe - firstframe+1);
               mr=(double)(lastframe - firstframe+1)/mr;
               if((n=tempobj->nskeleton) == Opn->nskeleton &&
                  tempobj->skeleton != NULL && Opn->skeleton != NULL){
                 InterpolateRobot(mr,Opn,Opl);
                 for(i=0;i<n;i++){ /* Q is interpolated rotations */
                    c4to4((Opn->skeleton+i)->Q,(tempobj->skeleton+i)->R);
                 }
               }
             }
           }
         }
       }
       else{
         DeleteActor(Np,firstframe);
         retval=FAIL;
       }
     }
    }
  }
  UpdateSelectedActor(FALSE);
  return retval;
}

static short CreatePathObject(short action, node *Np, object *tempobj,
                       long firstframe, long lastframe){
  int i;
  tempobj->firstpathpoint=CreateIPointPath(action,Np,(short)firstframe,
                 &(tempobj->pathtype),&(tempobj->npathpoints));
  if(tempobj->firstpathpoint != NULL){
    tempobj->origin[0]=0;   /* leave this alone no offset required */
    tempobj->origin[1]=0;   /* for paths !!!!!!!                   */
    tempobj->origin[2]=0;
    tempobj->pathlength=PathLength(tempobj->firstpathpoint
                                  ,tempobj->pathtype
                                  ,tempobj->npathpoints);
    i=tempobj->lastframe - tempobj->firstframe + 1;
    tempobj->v=ReTweenVelocity(tempobj,NULL,i,i,tempobj->pathlength,0);
    return 0;
  }
  DeleteActor(Np,firstframe);
  return -1;
}

void MakeKey(char com, long frame, node *np){
  position *Pp,*tempPp;
  align    *Ap,*tempAp;
  size     *Xp,*tempXp;
  double ratio,sx,sy,sz,ddp,phi,theta,alpha,kx[4],ky[4],kz[4];
  point  Offset,dp,spl,spn;
  short  ff;
  if(np == NULL){
    SendPrgmQuery(IDQ_OLD1,0);
  }
  else if(com == 'P'){
    if((Pp=np->fpos) != NULL)while(Pp != NULL){
      if(frame >= Pp->firstframe && frame < Pp->lastframe){
        if((Pp->last == NULL) || (Pp->firstframe == Pp->lastframe))
          CopyPoint(Pp->finish,Offset);
        else{
          ratio=(double)(frame+1  - Pp->firstframe)
               /(double)(Pp->lastframe+1 - Pp->firstframe);
          if(Pp->type == TWEEN || Pp->type == FOLLOWAT){
            SubPoints(Pp->finish,Pp->last->finish,dp);
            ScalePoint(ratio,ratio,ratio,dp,dp);
            AddPoints(Pp->last->finish,dp,Offset);
          }
          else{
            if(Pp->last->last != NULL)CopyPoint(Pp->last->last->finish,spl);
            else{
              SubPoints(Pp->finish,Pp->last->finish,dp);
              SubPoints(Pp->last->finish,dp,spl);
            }
            if(Pp->next != NULL)CopyPoint(Pp->next->finish,spn);
            else{
              SubPoints(Pp->finish,Pp->last->finish,dp);
              AddPoints(Pp->finish,dp,spn);
            }
            SplinesK(kx,spl[0],Pp->last->finish[0],Pp->finish[0],spn[0]);
            SplinesK(ky,spl[1],Pp->last->finish[1],Pp->finish[1],spn[1]);
            SplinesK(kz,spl[2],Pp->last->finish[2],Pp->finish[2],spn[2]);
            Offset[0] = SplinesP(kx,ratio);
            Offset[1] = SplinesP(ky,ratio);
            Offset[2] = SplinesP(kz,ratio);
          }
        }
        ff=Pp->firstframe;
        Pp->firstframe=frame+1;
        tempPp=CreatePosition(np,ff,(short)frame);
        if(tempPp == NULL){  /* if not created put everything back */
          Pp->firstframe=ff;
          return;
        }
        tempPp->type = Pp->type;
        CopyPoint(Offset,tempPp->finish);
        tempPp->onpath=Pp->onpath;
        tempPp->tension_s=Pp->tension_s;
        tempPp->tension_e=Pp->tension_e;
        goto MADEKEY;
      }
      Pp=Pp->next;
    }
    SendPrgmQuery(IDQ_NOADDKEY,0);
  }
  else if(com == 'Z'){
    if((Xp=np->fsiz) != NULL)while(Xp != NULL){
      if(frame >= Xp->firstframe && frame < Xp->lastframe){
        if(Xp->last == NULL){
          sx=Xp->Sx; sy=Xp->Sy; sz=Xp->Sz;
        }
        else{
          ratio=(double)(frame+1  - Xp->firstframe)
               /(double)(Xp->lastframe+1 - Xp->firstframe);
          sx = (Xp->Sx - Xp->last->Sx)*ratio + Xp->last->Sx;
          sy = (Xp->Sy - Xp->last->Sy)*ratio + Xp->last->Sy;
          sz = (Xp->Sz - Xp->last->Sz)*ratio + Xp->last->Sz;
        }
        ff=Xp->firstframe;
        Xp->firstframe=frame+1;
        tempXp=CreateSize(np,ff,(short)frame);
        if(tempXp == NULL){  /* if not created put everything back */
          Xp->firstframe=ff;
          return;
        }
        tempXp->Sx=sx;
        tempXp->Sy=sy;
        tempXp->Sz=sz;
        goto MADEKEY;
      }
      Xp=Xp->next;
    }
    SendPrgmQuery(IDQ_NOADDKEY,0);
  }
  else if(com == 'A'){
    if((Ap=np->fali) != NULL)while(Ap != NULL){
      if(frame >= Ap->firstframe && frame < Ap->lastframe){
        if((Ap->last == NULL) || (Ap->firstframe == Ap->lastframe)){
          phi=Ap->phi;
          theta=Ap->theta;
          alpha=Ap->alpha;
        }
        else{
          ratio=(double)(frame+1  - Ap->firstframe)
               /(double)(Ap->lastframe+1 - Ap->firstframe);
          ddp = Ap->phi - Ap->last->phi;
          if(ddp < -180.0)ddp += 360.0;
          if(ddp >  180.0)ddp -= 360.0;
          ddp *= ratio;
          ddp += Ap->last->phi;
          if(ddp < -180.0)ddp += 360.0;
          if(ddp >  180.0)ddp -= 360.0;
          phi = ddp;
          ddp = Ap->theta - Ap->last->theta;
          if(ddp < -180.0)ddp += 360.0;
          if(ddp >  180.0)ddp -= 360.0;
          ddp *= ratio;
          ddp += Ap->last->theta;
          if(ddp < -180.0)ddp += 360.0;
          if(ddp >  180.0)ddp -= 360.0;
          theta = ddp;
          ddp = Ap->alpha - Ap->last->alpha;
          if(ddp < -180.0)ddp += 360.0;
          if(ddp >  180.0)ddp -= 360.0;
          ddp *= ratio;
          ddp += Ap->last->alpha;
          if(ddp < -180.0)ddp += 360.0;
          if(ddp >  180.0)ddp -= 360.0;
          alpha = ddp;
        }
        ff=Ap->firstframe;
        Ap->firstframe=frame+1;
        tempAp=CreateAlign(np,ff,(short)frame);
        if(tempAp == NULL){  /* if not created put everything back */
          Ap->firstframe=ff;
          return;
        }
        tempAp->type=Ap->type;
        tempAp->theta=theta;
        tempAp->phi=phi;
        tempAp->alpha=alpha;
        tempAp->im=Ap->im;     /* needs fixing if tweening   ??? */
        tempAp->ima=Ap->ima;
        tempAp->topath=Ap->topath;
        tempAp->tension_s=Ap->tension_s;
        tempAp->tension_e=Ap->tension_e;
        goto MADEKEY;
      }
      Ap=Ap->next;
    }
    SendPrgmQuery(IDQ_NOADDKEY,0);
  }
  return;
  MADEKEY:
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
  UpdateSelectedActor(FALSE);
  return;
}

void NextKey(char com){
 object   *Op;
 position *Pp;
 align    *Ap;
 size     *Xp;
 if(com == 'N'){
   if(SelectedNode == NULL)goto KEY_ERR;
   if(tool == NOTOOL || tool == NODETOOL || tool == SKELETOR){
    if((Op=SelectedNode->fobj) != NULL)while(Op != NULL){
     if(CurrentFrame < Op->lastframe){
       CurrentFrame=Op->lastframe;
       goto GOTIT;
     }
     Op=Op->next;
    }
    goto KEY_ERR;
   }
   else if(tool == GRABBER || (tool == TRACKER && subtool == SHIFTER)){
    if((Pp=SelectedNode->fpos) != NULL)while(Pp != NULL){
     if(CurrentFrame < Pp->lastframe){
       CurrentFrame=Pp->lastframe;
       goto GOTIT;
     }
     Pp=Pp->next;
    }
    goto KEY_ERR;
   }
   else if(tool == ROTATOR || (tool == TRACKER && subtool == ANGLER)){
    if((Ap=SelectedNode->fali) != NULL)while(Ap != NULL){
     if(CurrentFrame < Ap->lastframe){
       CurrentFrame=Ap->lastframe;
       goto GOTIT;
     }
     Ap=Ap->next;
    }
    goto KEY_ERR;
   }
   else if(tool == SIZER){
    if((Xp=SelectedNode->fsiz) != NULL)while(Xp != NULL){
     if(CurrentFrame < Xp->lastframe){
       CurrentFrame=Xp->lastframe;
       goto GOTIT;
     }
     Xp=Xp->next;
    }
    goto KEY_ERR;
   }
   else goto KEY_ERR;
 }
 else if(com == 'L'){
   if(SelectedNode == NULL)goto KEY_ERR;
   if(tool == NOTOOL || tool == NODETOOL || tool == SKELETOR){
    if((Op=SelectedNode->fobj) != NULL)while(Op != NULL){
     if(CurrentFrame <= Op->lastframe){
       if(Op->last == NULL)goto KEY_ERR;
       CurrentFrame=Op->last->lastframe;
       goto GOTIT;
     }
     Op=Op->next;
    }
    goto KEY_ERR;
   }
   else if(tool == GRABBER || (tool == TRACKER && subtool == SHIFTER)){
    if((Pp=SelectedNode->fpos) != NULL)while(Pp != NULL){
     if(CurrentFrame <= Pp->lastframe){
       if(Pp->last == NULL)goto KEY_ERR;
       CurrentFrame=Pp->last->lastframe;
       goto GOTIT;
     }
     Pp=Pp->next;
    }
    goto KEY_ERR;
   }
   else if(tool == ROTATOR || (tool == TRACKER && subtool == ANGLER)){
    if((Ap=SelectedNode->fali) != NULL)while(Ap != NULL){
     if(CurrentFrame <= Ap->lastframe){
       if(Ap->last == NULL)goto KEY_ERR;
       CurrentFrame=Ap->last->lastframe;
       goto GOTIT;
     }
     Ap=Ap->next;
    }
    goto KEY_ERR;
   }
   else if(tool == SIZER){
    if((Xp=SelectedNode->fsiz) != NULL)while(Xp != NULL){
     if(CurrentFrame <= Xp->lastframe){
       if(Xp->last == NULL)goto KEY_ERR;
       CurrentFrame=Xp->last->lastframe;
       goto GOTIT;
     }
     Xp=Xp->next;
    }
    goto KEY_ERR;
   }
   else goto KEY_ERR;
 }
 goto KEY_ERR;
 GOTIT:
 ReDrawStageDisplay(TRUE);
 UpdateSelectedActor(FALSE);
 if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
 else UpdateGLview(TRUE);
// SetScrollPos(ghwnd_xscl,SB_CTL,(int)CurrentFrame,TRUE);
 SendMessage(ghwnd_xscl,TBM_SETPOS,TRUE,(LPARAM)CurrentFrame);
 return;
 KEY_ERR:
 MessageBeep(MB_OK);
 //SendPrgmQuery(IDQ_NOKEYFOUND,0);
 return;
}

short MakeControl(node *Np, char typ, HWND parent){
/* This function assummes that a time line does not exist at that frame */
 short code,nextf,lastf,type;
 position *Pp;
 align    *Ap;
 size     *Xp;
 nextf=Nframes;
 lastf=1;
 code=FAIL;
 if(typ == 'P'){
   if((Pp=Np->fpos) != NULL){
    if(Np->type == SKY || Np->type == IMAGEP)return OK;
     while(Pp != NULL){
       if(Pp->firstframe > CurrentFrame){
         nextf=Pp->firstframe-1;
         break;
       }
       lastf=Pp->lastframe+1;
       Pp=Pp->next;
     }
   }
   EditPosition(Np,lastf,nextf,0,parent); code=OK;
 }
 if(typ == 'A'){
  if(Np->type == PATH || Np->type == SKY || Np->type == TARGET ||
     Np->type == IMAGEP)return OK;
  if((Ap=Np->fali) != NULL){
     while(Ap != NULL){
       if(Ap->firstframe > CurrentFrame){
         nextf=Ap->firstframe-1;
         break;
       }
       lastf=Ap->lastframe+1;
       Ap=Ap->next;
     }
   }
   EditAlign(Np,lastf,nextf,0,parent); code=OK;
 }
 if(typ == 'Z'){
   if(Np->type == PATH || Np->type == SKY || Np->type == TARGET ||
      Np->type == IMAGEP)return OK;
   if((Xp=Np->fsiz) != NULL){
     while(Xp != NULL){
       if(Xp->firstframe > CurrentFrame){
         nextf=Xp->firstframe-1;
         break;
       }
       lastf=Xp->lastframe+1;
       Xp=Xp->next;
     }
   }
   EditSize(Np,lastf,nextf); code=OK;
 }
 if(code == FAIL){;}
 else{
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
 }
 fail_op=NO;
 return code;
}

short CheckForControl(node *Np, char typ){
 position *Pp;
 align    *Ap;
 size     *Xp;
 if(typ == 'P'){
   if((Pp=Np->fpos) != NULL){
     while(Pp != NULL){
       if(Pp->firstframe <= CurrentFrame && Pp->lastframe >= CurrentFrame){
         return YES;
       }
       Pp=Pp->next;
     }
   }
 }
 if(typ == 'A'){
   if((Ap=Np->fali) != NULL){
     while(Ap != NULL){
       if(Ap->firstframe <= CurrentFrame && Ap->lastframe >= CurrentFrame){
         return YES;
       }
       Ap=Ap->next;
     }
   }
 }
 if(typ == 'Z'){
   if((Xp=Np->fsiz) != NULL){
     while(Xp != NULL){
       if(Xp->firstframe <= CurrentFrame && Xp->lastframe >= CurrentFrame){
         return YES;
       }
       Xp=Xp->next;
     }
   }
 }
 return NO;
}

void EditPosition(node *Np, short firstframe, short lastframe,
                  short type, HWND parent){
 position *tempposition;
 short action;
 double x,y,z;
 if(Np->type == SKY || Np->type == IMAGEP)return;
 tempposition = CreatePosition(Np,firstframe,lastframe);
 if(fail_op == YES)fail_op=NO; /* failed no memory or overlap do nothing*/
 else if(tempposition != NULL){
   tempposition->type=SPLINE;
   ChangePositionEndPoints(Np,firstframe,lastframe,parent,YES);
 }
 else if(firstframe == lastframe){
   ChangePositionEndPoints(Np,firstframe,firstframe,parent,NO);
 }
 UpdateSelectedActor(FALSE);
 return;
}

void EditAlign(node *Np, short firstframe, short lastframe,
               short type, HWND parent){
 align *tempalign;
 short action,dumyi1;
 double dr1,dr2,dr3,dr4;
 point dumypoint,dp1;
 if((Np->type == SKY) || (Np->type == TARGET) ||
    (Np->type == PATH)|| (Np->type == IMAGEP))return;
 tempalign = CreateAlign(Np,firstframe,lastframe);
 if(fail_op == YES)fail_op=NO;    /* failed no memory or overlap */
 else if(tempalign != NULL){
   ChangeAlignEndPoints(Np,firstframe,lastframe,parent,YES);
 }
 else if(firstframe == lastframe){
   ChangeAlignEndPoints(Np,firstframe,firstframe,parent,NO);
 }
 UpdateSelectedActor(FALSE);
 return;
}

void EditSize(node *Np, short firstframe, short lastframe){
 size *tempsize;
 char string[32];
 short action;
 double sx,sy,sz;
 if( (Np->type == SKY)  || (Np->type == TARGET)
  || (Np->type == PATH) || (Np->type == IMAGEP))return;
 tempsize = CreateSize(Np,firstframe,lastframe);
 if(fail_op == YES)fail_op=NO;
 else if((tempsize == NULL) && (firstframe == lastframe)){
   ChangeSizeEndPoints(Np,firstframe,0);/* edit etc */
 }
 else if(tempsize != NULL){
   sx=tempsize->Sx; sy=tempsize->Sy; sz=tempsize->Sz;
   if(Np->type == LIGHT){
     if(RequestEditLightSize(&sx,&sy,&sz,tempsize,0) < 0){
       DeleteSize(Np,firstframe);
       return;
     }
   }
   else if(Np->type == CAMERA){
     if(EditCameraSize(&sx,&sy,&sz,tempsize,0) < 0){
       DeleteSize(Np,firstframe);
       return;
     }
   }
   else{
     if(RequestEditSize(&sx,&sy,&sz,tempsize,0) < 0){
       DeleteSize(Np,firstframe);
       return;
     }
   }
   tempsize->Sx=sx; tempsize->Sy=sy; tempsize->Sz=sz;
 }
 UpdateSelectedActor(FALSE);
 return;
}

// script part 2

static node *LoadStandinModel(short *, char *, char *);

static void ChangeSizeEndPoints(node *Np, short frame, short code){
  size *tempsize;
  char string[24];
  short l,f,minf,maxf;
  double sx,sy,sz;
  int i;
  EDIT_ACTION=YES;
  if((tempsize = Np->fsiz) != NULL)while(tempsize != NULL){
    if(tempsize->firstframe <=frame && tempsize->lastframe >=frame){
      if(code == 0){
        sx=tempsize->Sx;  sy=tempsize->Sy;  sz=tempsize->Sz;
        if(Np->type == LIGHT){
          i=RequestEditLightSize(&sx,&sy,&sz,tempsize,1);
          if(i == OK){
            tempsize->Sx=sx;  tempsize->Sy=sy;  tempsize->Sz=sz;
          }
          else if(i == 2){
            DeleteSize(Np,frame);
          }
        }
        else if(Np->type == CAMERA){
          i=(int)EditCameraSize(&sx,&sy,&sz,tempsize,1);
          if(i == OK){
            tempsize->Sx=sx;  tempsize->Sy=sy;  tempsize->Sz=sz;
          }
          else if(i == 2){
            DeleteSize(Np,frame);
          }
        }
        else{
          i=(int)RequestEditSize(&sx,&sy,&sz,tempsize,1);
          if(i == OK){
            tempsize->Sx=sx;  tempsize->Sy=sy;  tempsize->Sz=sz;
          }
          else if(i == 2){
            DeleteSize(Np,frame);
          }
        }
      }
      break;
    }
    tempsize=tempsize->next;
  }
  else{
    SendPrgmQuery(IDQ_SYNTAXERROR,0);
  }
}

static void ChangeAlignEndPoints(node *Np, short ff, short lf,
                                 HWND parent,int new){
 int i;
 align *Op,Align;
 struct NP {node *np; align *ap; int new;} NP;
 if((Op=Np->fali) != NULL)while(Op != NULL){
  if(Op->firstframe <= ff && Op->lastframe >= lf){
    memcpy(&Align,Op,sizeof(align));
    NP.np=Np; NP.ap= &Align; NP.new=new;
    if((i=(int)DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_TLER),parent,
         (DLGPROC)rTimeLineEditDlgProc,(LPARAM)&NP)) >= 0){
      EDIT_ACTION=YES;
      if(i == 1)DeleteAlign(Np,lf);
      else if(i == 2)CutOutAlign(Np,lf);
      else memcpy(Op,&Align,sizeof(align));
    }
    else if(new == YES){
      DeleteAlign(Np,lf);
    }
    CheckRecursiveFollow(Np,Op->firstframe,Op->lastframe);
    return;
  }
  Op=Op->next;
 }
}

static void ChangePositionEndPoints(node *Np, short ff, short lf,
                                    HWND parent, int new){
 int i;
 position *Op,Position;
 struct NP {node *np; position *pp; int new;} NP;
 if((Op=Np->fpos) != NULL)while(Op != NULL){
  if(Op->firstframe <= ff && Op->lastframe >= lf){
    memcpy(&Position,Op,sizeof(position));
    NP.np=Np; NP.pp= &Position;  NP.new=new;
    if((i=(int)DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_TLEP),parent,
         (DLGPROC)pTimeLineEditDlgProc,(LPARAM)&NP)) >= 0){
      EDIT_ACTION=YES;
      if(i == 1)DeletePosition(Np,lf);
      else if(i == 2)CutOutPosition(Np,lf);
      else memcpy(Op,&Position,sizeof(position));
    }
    else if(new == YES){
      DeletePosition(Np,lf);
    }
    CheckRecursiveFollow(Np,Op->firstframe,Op->lastframe);
    return;
  }
  Op=Op->next;
 }
}

static short EditObjectTimeline(node *Np, short frame, short code,
                               short in_script){
 object *Op;
 char string[10];
 short f,l,minf,maxf,oldff,oldlf,action=-1,action1;
 double ref;
 HWND hwnd;
 EDIT_ACTION=YES;
 if(in_script)hwnd=ghwndTimeline; else hwnd=ghwnd_main;
 if((Op=Np->fobj) != NULL)while(Op != NULL){
  if(Op->firstframe <= frame && Op->lastframe >= frame){
   oldff=Op->firstframe; oldlf=Op->lastframe;
   if(Op->type == NORMAL || Op->type == ANIMOBJ){
     action=(short)DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_COSTUME),
                    hwnd,(DLGPROC)EditObjectDlgProc,(LPARAM)Op);
   }
   else if(Op->type == GROUND){
     if(EditGroundDialog(Op,YES) < 0){;}
   }
   else if(Op->type == LIGHT){
     if(EditLightDialog(Op,YES) < 0){;}
   }
   else if(Op->type == IMAGEP){
     action=(short)DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_COSTUMEXIP),
                    hwnd,(DLGPROC)EditObjectXipDlgProc,(LPARAM)Op);
   }
   else if(Op->type == CSG){
     struct NC {node *np; object *op,*tp; int d;} NC;
     object temp_obj;
     NC.op=Op;  NC.tp=&temp_obj; NC.np=Np;  NC.d=1;
     action=(short)DialogBoxParam(ghinst_main,
                          MAKEINTRESOURCE(DLG_CSG),
                          hwnd,(DLGPROC)CsgSetDlgProc,(LPARAM)&NC);
     if(action == 1){
       DeleteActor(Np,frame);
     }
   }
//   else if(Op->type == PATH){
//   }
   CheckRecursiveFollow(Np,Op->firstframe,Op->lastframe);
   return action;
  }
  Op=Op->next;
 }
 return action;
}

static void CopyObject(node *Np, short firstframe, short lastframe){
 object *Op,*tempobj,*lastobj;
 pathpoint *Ppp,*Qpp;
 short endframe,newstart,newend,i;
 long sks;
 lastobj=NULL;
 EDIT_ACTION=YES;
 if((Op=Np->fobj) != NULL)while(Op != NULL){lastobj=Op; Op=Op->next; }
 if(lastobj == NULL || lastobj->lastframe >= Nframes){
   SendPrgmQuery(IDQ_NOCOPYSPACE,0);
 }
 else if((Op=Np->fobj) != NULL){
   endframe=lastobj->lastframe;
   while(Op != NULL && Op->firstframe <= lastframe){
     if(Op->lastframe >= firstframe){
       newstart= endframe+1;
       newend  = newstart+(Op->lastframe-Op->firstframe);
       if(newend > Nframes)newend=Nframes;
       tempobj=CreateActor(Np,newstart,newend);
       if(tempobj == NULL){
         SendPrgmQuery(IDQ_COPYFAIL,0);
         break;
       }
       strcpy(tempobj->name,Op->name);
       tempobj->morph=Op->morph;
       tempobj->morphNo=Op->morphNo;
       if(Op->effect != NULL){
         tempobj->effect=(char *)X__Malloc(strlen(Op->effect)+1);
         strncpy(tempobj->effect,Op->effect,strlen(Op->effect)+1);
       }
       if(Op->xip != NULL){
         tempobj->xip=(char *)X__Malloc(strlen(Op->xip)+1);
         strncpy(tempobj->xip,Op->xip,strlen(Op->xip)+1);
       }
       tempobj->hEffect=Op->hEffect;
       strcpy(tempobj->effectname,Op->effectname);
       tempobj->type=Op->type;
       tempobj->quick_only=Op->quick_only;
       tempobj->fileID=Op->fileID;
       tempobj->in_ram=Op->in_ram;
       tempobj->bWireframe=Op->bWireframe;
       tempobj->points=Op->points;
       tempobj->edges=Op->edges;
       tempobj->nedges=Op->nedges;
       tempobj->npoints=Op->npoints;
       tempobj->w_frame.Np=Op->w_frame.Np;
       tempobj->w_frame.Ne=Op->w_frame.Ne;
       tempobj->w_frame.p=Op->w_frame.p;
       tempobj->w_frame.e=Op->w_frame.e;
       tempobj->nNurbs=Op->nNurbs;
       tempobj->Nurbs=Op->Nurbs;
       tempobj->bIKchain=Op->bIKchain;
       if(Op->nskeleton > 0){
         sks=(long)(Op->nskeleton)*(long)sizeof(skel);
         if((tempobj->skeleton=(skel *)X__Malloc(sks)) != NULL){
           tempobj->nskeleton = Op->nskeleton;
           memcpy(tempobj->skeleton,Op->skeleton,sks);
         }
         else{
           tempobj->nskeleton=0;
         }
       }
       tempobj->skid=Op->skid;
       CopyPoint(Op->origin,tempobj->origin);
       CopyPoint(Op->offset,tempobj->offset);
       for(i=0;i<8;i++){
         CopyPoint(Op->outline[i],tempobj->outline[i]);
       }
       tempobj->groundtype=Op->groundtype;
       tempobj->lighttype=Op->lighttype;
       tempobj->light_intensity=Op->light_intensity;
       tempobj->depth_cue=Op->depth_cue;
       tempobj->depth_cue_c=Op->depth_cue_c;
       tempobj->depth_cue_l=Op->depth_cue_l;
       tempobj->depth_cue_q=Op->depth_cue_l;
       tempobj->self_shadow=Op->self_shadow;
       tempobj->cast_shadow=Op->cast_shadow;
       tempobj->show_shadow=Op->show_shadow;
       tempobj->colour[0]=Op->colour[0];
       tempobj->acolour[0]=Op->acolour[0];
       tempobj->colour[1]=Op->colour[1];
       tempobj->acolour[1]=Op->acolour[1];
       tempobj->colour[2]=Op->colour[2];
       tempobj->acolour[2]=Op->acolour[2];
       tempobj->pathtype=Op->pathtype;
       tempobj->npathpoints=Op->npathpoints;
       tempobj->pathlength=Op->pathlength;
       tempobj->gspec=Op->gspec;
       tempobj->gtran=Op->gtran;
       tempobj->grefl=Op->grefl;
       tempobj->gbump=Op->gbump;
       if(Op->npathpoints > 0)
       for(i=0,Ppp=NULL,Qpp=Op->firstpathpoint;i<Op->npathpoints;i++){
          if(((Ppp=AppendPathPoint(Ppp)) != NULL) && Qpp != NULL){
            if(i == 0)tempobj->firstpathpoint=Ppp;
            Ppp->status=Qpp->status;
            CopyPoint(Qpp->p,Ppp->p);
            Ppp->distance=Qpp->distance;
            Ppp->ptheta=Qpp->ptheta;
            Ppp->pfi=Qpp->pfi;
            Ppp->tension_n=Qpp->tension_n;
            Ppp->tension_p=Qpp->tension_p;
            Qpp=Qpp->next;
          }
       }
       tempobj->pec=NULL;
       tempobj->npec=0;
       if(Op->v != NULL){            /* copy the velocity specification */
         tempobj->v=ReTweenVelocity(Op,Op->v,
           (int)(Op->lastframe-Op->firstframe+1),
           (int)(tempobj->lastframe-tempobj->firstframe+1),
           Op->pathlength,1);
       }
       endframe=newend;
       CheckRecursiveFollow(Np,newstart,newend);
       if(newend == Nframes){
         SendPrgmQuery(IDQ_NOCOPYSPACE,0);
         break;
       }
     }
     Op=Op->next;
   }
 }
}

static void CutOutObject(node *Np, short frame){
 object *Op,*Op1;
 short ff,lf,df;
 EDIT_ACTION=YES;
 if((Op=Np->fobj) != NULL)while(Op != NULL){
  if((ff=Op->firstframe) <= frame && (lf=Op->lastframe) >= frame){
    df=Op->lastframe - Op->firstframe+1;
    Op1=Op->next;
    DeleteActor(Np,frame);
    while(Op1 != NULL){
      Op1->firstframe -= df;
      Op1->lastframe  -= df;
      CheckRecursiveFollow(Np,Op1->firstframe,Op1->lastframe);
      Op1=Op1->next;
    }
    return;
  }
  Op=Op->next;
 }
}

static void CutOutPosition(node *Np, short frame){
 position *Op,*Op1;
 short ff,lf,df;
 EDIT_ACTION=YES;
 if((Op=Np->fpos) != NULL)while(Op != NULL){
  if((ff=Op->firstframe) <= frame && (lf=Op->lastframe) >= frame){
    df=Op->lastframe - Op->firstframe+1;
    Op1=Op->next;
    DeletePosition(Np,frame);
    while(Op1 != NULL){
      Op1->firstframe -= df;
      Op1->lastframe  -= df;
      CheckRecursiveFollow(Np,Op1->firstframe,Op1->lastframe);
      Op1=Op1->next;
    }
    return;
  }
  Op=Op->next;
 }
}

static void CutOutAlign(node *Np, short frame){
 align *Op,*Op1;
 short ff,lf,df;
 EDIT_ACTION=YES;
 if((Op=Np->fali) != NULL)while(Op != NULL){
  if((ff=Op->firstframe) <= frame && (lf=Op->lastframe) >= frame){
    df=Op->lastframe - Op->firstframe+1;
    Op1=Op->next;
    DeleteAlign(Np,frame);
    while(Op1 != NULL){
      Op1->firstframe -= df;
      Op1->lastframe  -= df;
      CheckRecursiveFollow(Np,Op1->firstframe,Op1->lastframe);
      Op1=Op1->next;
    }
    return;
  }
  Op=Op->next;
 }
}

static node *LoadStandinModel(short *nmodels, char *title, char *command){
 object *Op;
 if(CreateNode() == NULL)return NULL;
 sprintf(MainNp->actorname,"Model%d",(*nmodels+1));
 MainNp->type=NORMAL;
 if((Op=CreateActor(MainNp,1,(short)Nframes)) == NULL || fail_op)goto ERRORS;
 Op->type=MainNp->type;
 if(SelectFileName(0,gszSCEfile,gszSCEdir,title,
                   "(*.MFX)|*.mfx|",ghwnd_main) == TRUE){
   strcpy(Op->name,gszSCEfile);
   Op->morph=NO;
   if(StageLoadObject(gszSCEfile,Op,YES,YES,NO) == FAIL)
     DeleteActor(MainNp,1);
 }
 else DeleteActor(MainNp,1);
 (*nmodels)++;
 return MainNp;
 ERRORS:
 DeleteNode(MainNp);
 return NULL;
}

static BOOL CALLBACK FollowAtOptionsDlgProc(HWND hwnd,UINT msg,
                                 WPARAM wparam,LPARAM lparam){
 char str[32];
 position *pp;
 BOOL err;
 int i;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       pp=(position *)lparam;
       if(pp->fx)SendDlgItemMessage(hwnd,DLG_FOLLOWAT_LR,BM_SETCHECK,1,0);
       if(pp->fy)SendDlgItemMessage(hwnd,DLG_FOLLOWAT_FB,BM_SETCHECK,1,0);
       if(pp->fz)SendDlgItemMessage(hwnd,DLG_FOLLOWAT_UD,BM_SETCHECK,1,0);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      pp=(position *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_FOLLOWAT_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case DLG_FOLLOWAT_OK:
          if(SendDlgItemMessage(hwnd,DLG_FOLLOWAT_LR,BM_GETCHECK,0,0))
             pp->fx=1; else pp->fx=0;
          if(SendDlgItemMessage(hwnd,DLG_FOLLOWAT_FB,BM_GETCHECK,0,0))
             pp->fy=1; else pp->fy=0;
          if(SendDlgItemMessage(hwnd,DLG_FOLLOWAT_UD,BM_GETCHECK,0,0))
             pp->fz=1; else pp->fz=0;
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void SetFollowAtOptions(position *Pp, HWND parent){
 int i;
 position pp;
 HWND original;
 memcpy(&pp,Pp,sizeof(position));
 original=GetActiveWindow();
 if((i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_FOLLOWAT),parent,
            (DLGPROC)FollowAtOptionsDlgProc,(LPARAM)&pp)) == OK){
   EDIT_ACTION=YES;
   memcpy(Pp,&pp,sizeof(position));
 }
 if(original != NULL)SetActiveWindow(original);
 return;
}

static LRESULT CALLBACK pTimeLineEditDlgProc(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam){
 static position *dp;
 static node     *np;
 char title[128];
 struct NP {node *np; position *pp; int new;} *NP;
 node *OnPath;
 BOOL err,status;
 int i,TypeID[6]={DLG_TLEP_PT,DLG_TLEP_FOLLOW,DLG_TLEP_ST,
                 DLG_TLEP_FOFF,DLG_TLEP_FON,DLG_TLEP_FAT};
 int ff,lf;
 double x,y,z;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       EnableToolPannels(ALL_PANNELS,FALSE);
       NP=(struct NP *)lparam;
       dp=NP->pp; np=NP->np;
       sprintf(title,"Movement Timeline for %s Actor",NP->np->actorname);
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)title);
       if(NP->new == YES){
         EnableWindow(GetDlgItem(hwnd,DLG_TLEP_DELETE),FALSE);
         EnableWindow(GetDlgItem(hwnd,DLG_TLEP_CUT),FALSE);
       }
       SetDlgItemInt(hwnd,DLG_TLEP_F,dp->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_TLEP_L,dp->lastframe,FALSE);
       if(dp->last != NULL)ff=dp->last->lastframe+1;
       else                ff=1;
       if(dp->next != NULL)lf=dp->next->firstframe-1;
       else                lf=Nframes;
       SendDlgItemMessage(hwnd,DLG_TLEP_FS,SPNM_SETRANGE,0,
                          MAKELPARAM(ff,lf));
       SendDlgItemMessage(hwnd,DLG_TLEP_FS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_TLEP_FS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_TLEP_F));
       SendDlgItemMessage(hwnd,DLG_TLEP_LS,SPNM_SETRANGE,0,
                          MAKELPARAM(ff,lf));
       SendDlgItemMessage(hwnd,DLG_TLEP_LS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_TLEP_LS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_TLEP_L));
       SendDlgItemMessage(hwnd,TypeID[dp->type],BM_SETCHECK,TRUE,0);
       x = (double)(dp->finish[0]-lrulerx); x /= ruler;
       y = (double)(dp->finish[1]-lrulery); y /= ruler;
       z = (double)(dp->finish[2]-lrulerz); z /= ruler;
       sprintf(title,"%.2lf",x); SetDlgItemText(hwnd,DLG_TLEP_X,title);
       sprintf(title,"%.2lf",y); SetDlgItemText(hwnd,DLG_TLEP_Y,title);
       sprintf(title,"%.2lf",z); SetDlgItemText(hwnd,DLG_TLEP_Z,title);
       OnPath=dp->onpath;
       if(OnPath != NULL){
         SetDlgItemText(hwnd,DLG_TLEP_FOLLOWED,OnPath->actorname);
       }
       else{
         SetDlgItemText(hwnd,DLG_TLEP_FOLLOWED,"<None>");
       }
       if(OnPath == NULL)status=FALSE; else status=TRUE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOLLOW),status);
       if(OnPath != NULL && OnPath->type == PATH)status=FALSE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOFF),status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FON),status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FAT),status);
       if(dp->type == TWEEN || dp->type == SPLINE)
         status=FALSE; else status=TRUE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOLLOWED),status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_X),!status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_Y),!status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_Z),!status);
       if(OnPath == NULL)status=TRUE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLEP_SELECT),status);
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      /****************** NOTE: dont make radio buttons default *****************/
      if(dp == NULL)break;
      switch(LOWORD(wparam)){
        case DLG_TLEP_PT:      i=TWEEN;     goto PT1;
        case DLG_TLEP_ST:      i=SPLINE;    goto PT1;
        case DLG_TLEP_FOLLOW:  i=FOLLOW;    goto PT1;
        case DLG_TLEP_FOFF:    i=FOLLOWOFF; goto PT1;
        case DLG_TLEP_FON:     i=FOLLOWON;  goto PT1;
        case DLG_TLEP_FAT:     i=FOLLOWAT;
          PT1:
          dp->type=i;
          OnPath=dp->onpath;
          if(OnPath == NULL)status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOLLOW),status);
          if(OnPath != NULL && OnPath->type == PATH)status=FALSE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOFF),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FON),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FAT),status);
          if(OnPath != NULL)
            SetDlgItemText(hwnd,DLG_TLEP_FOLLOWED,OnPath->actorname);
          else
            SetDlgItemText(hwnd,DLG_TLEP_FOLLOWED,"<None>");
          if(dp->type == TWEEN || dp->type == SPLINE)
            status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOLLOWED),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_X),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_Y),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_Z),!status);
          if(OnPath == NULL)status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_SELECT),status);
          break;
        case DLG_TLEP_SELECT:
          if((OnPath=SelectNode(dp->pnodename,NO,hwnd)) != NULL){
            if(OnPath == np){
              MessageBox(hwnd,"Actor Cannot Follow Itself",NULL,MB_OK);
              break;
            }
            dp->onpath=OnPath;
            strcpy(dp->pnodename,OnPath->actorname);
          }
          OnPath=dp->onpath;
          if(OnPath == NULL)status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOLLOW),status);
          if(OnPath != NULL && OnPath->type == PATH)status=FALSE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOFF),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FON),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FAT),status);
          if(OnPath != NULL)
            SetDlgItemText(hwnd,DLG_TLEP_FOLLOWED,OnPath->actorname);
          else
            SetDlgItemText(hwnd,DLG_TLEP_FOLLOWED,"<None>");
          if(dp->type == TWEEN || dp->type == SPLINE)
            status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_FOLLOWED),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_X),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_Y),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_Z),!status);
          if(OnPath == NULL)status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLEP_SELECT),status);
          break;
        case DLG_TLEP_COPY:
          if(dp->last != NULL){
            x = (double)(dp->last->finish[0]-lrulerx); x /= ruler;
            y = (double)(dp->last->finish[1]-lrulery); y /= ruler;
            z = (double)(dp->last->finish[2]-lrulerz); z /= ruler;
            sprintf(title,"%.2lf",x); SetDlgItemText(hwnd,DLG_TLEP_X,title);
            sprintf(title,"%.2lf",y); SetDlgItemText(hwnd,DLG_TLEP_Y,title);
            sprintf(title,"%.2lf",z); SetDlgItemText(hwnd,DLG_TLEP_Z,title);
          }
          break;
        case IDCANCEL:
        case DLG_TLEP_CANCEL:
          EndDialog(hwnd,-1);
          return(TRUE);
        case DLG_TLEP_DELETE:
          EndDialog(hwnd,1);
          return(TRUE);
        case DLG_TLEP_CUT:
          EndDialog(hwnd,2);
          return(TRUE);
        case IDOK:
        case DLG_TLEP_OK:
          if(dp->type == TWEEN || dp->type == SPLINE){
             dp->onpath=NULL; dp->pnodename[0]='\0';
          }
          if(dp->type == FOLLOWAT)SetFollowAtOptions(dp,hwnd);
          i=GetDlgItemInt(hwnd,DLG_TLEP_F,&err,FALSE);
          if(err)dp->firstframe=(short)i;
          if(dp->last != NULL && dp->firstframe <= dp->last->lastframe)
             dp->firstframe=dp->last->lastframe+1;
          if(dp->firstframe < 1)dp->firstframe=1;
          i=GetDlgItemInt(hwnd,DLG_TLEP_L,&err,FALSE);
          if(err)dp->lastframe=(short)i;
          if(dp->next != NULL && dp->lastframe >= dp->next->firstframe)
             dp->lastframe=dp->next->firstframe-1;
          if(dp->lastframe > Nframes)dp->lastframe=Nframes;
          if(dp->lastframe < dp->firstframe)dp->lastframe=dp->firstframe;
          if(GetDlgItemText(hwnd,DLG_TLEP_X,title,10) != 0){
             x=atof(title); x *= ruler; dp->finish[0]=lrulerx+(long)x;
          }
          if(GetDlgItemText(hwnd,DLG_TLEP_Y,title,10) != 0){
             y=atof(title); y *= ruler; dp->finish[1]=lrulery+(long)y;
          }
          if(GetDlgItemText(hwnd,DLG_TLEP_Z,title,10) != 0){
             z=atof(title); z *= ruler; dp->finish[2]=lrulerz+(long)z;
          }
          EndDialog(hwnd,0);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static LRESULT CALLBACK rTimeLineEditDlgProc(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam){
 static align *dp;
 static node  *np;
 node *OnPath;
 BOOL err,status;
 char title[128];
 struct NP {node *np; align *ap; int new;} *NP;
 int i,TypeID[5]={DLG_TLER_PT,DLG_TLER_TOPATH,DLG_TLER_LOOKAT,
                  DLG_TLER_MIMIC,DLG_TLER_ST},
       InternalID[4]={DLG_TLER_NA,DLG_TLER_VA,DLG_TLER_SA,DLG_TLER_FA};
 int ff,lf;
 double a;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=NULL;
       EnableToolPannels(ALL_PANNELS,FALSE);
       NP=(struct NP *)lparam;
       dp=NP->ap;
       np=NP->np;
       sprintf(title,"Rotation Timeline for %s Actor",NP->np->actorname);
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)title);
       if(NP->new == YES){
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_DELETE),FALSE);
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_CUT),FALSE);
       }
       SetDlgItemInt(hwnd,DLG_TLER_F,dp->firstframe,FALSE);
       SetDlgItemInt(hwnd,DLG_TLER_L,dp->lastframe,FALSE);
       if(dp->last != NULL)ff=dp->last->lastframe+1;
       else                ff=1;
       if(dp->next != NULL)lf=dp->next->firstframe-1;
       else                lf=Nframes;
       SendDlgItemMessage(hwnd,DLG_TLER_FS,SPNM_SETRANGE,0,
                          MAKELPARAM(ff,lf));
       SendDlgItemMessage(hwnd,DLG_TLER_FS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->firstframe,0);
       SendDlgItemMessage(hwnd,DLG_TLER_FS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_TLER_F));
       SendDlgItemMessage(hwnd,DLG_TLER_LS,SPNM_SETRANGE,0,
                          MAKELPARAM(ff,lf));
       SendDlgItemMessage(hwnd,DLG_TLER_LS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->lastframe,0);
       SendDlgItemMessage(hwnd,DLG_TLER_LS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_TLER_L));
       SendDlgItemMessage(hwnd,TypeID[dp->type],BM_SETCHECK,TRUE,0);
       sprintf(title,"%.2lf",dp->theta);
       SetDlgItemText(hwnd,DLG_TLER_THETA,title);
       sprintf(title,"%.2lf",dp->alpha);
       SetDlgItemText(hwnd,DLG_TLER_ALPHA,title);
       sprintf(title,"%.2lf",dp->phi);
       SetDlgItemText(hwnd,DLG_TLER_PHI,title);
       OnPath=dp->topath;
       if(OnPath != NULL){
         SetDlgItemText(hwnd,DLG_TLER_FOLLOWED,OnPath->actorname);
       }
       else{
         SetDlgItemText(hwnd,DLG_TLER_FOLLOWED,"<None>");
       }
       if(OnPath == NULL)status=FALSE; else status=TRUE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_LOOKAT),status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_MIMIC),status);
       if(OnPath != NULL && OnPath->type != PATH)status=FALSE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_TOPATH),status);
       if(dp->type == TWEEN || dp->type == SPLINEA)
         status=FALSE; else status=TRUE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_FOLLOWED),status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_THETA),!status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_ALPHA),!status);
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_PHI),!status);
       if(OnPath == NULL)status=TRUE;
       EnableWindow(GetDlgItem(hwnd,DLG_TLER_SELECT),status);
       if(NP->np->type == CAMERA || NP->np->type == GROUND){
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_NA),FALSE);
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_VA),FALSE);
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_SA),FALSE);
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_FA),FALSE);
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_CLOCK),FALSE);
         EnableWindow(GetDlgItem(hwnd,DLG_TLER_ACLOCK),FALSE);
       }
       else{
         SendDlgItemMessage(hwnd,InternalID[dp->im],BM_SETCHECK,TRUE,0);
         if(dp->ima < 0.0)
              SendDlgItemMessage(hwnd,DLG_TLER_CLOCK,BM_SETCHECK,TRUE,0);
         else SendDlgItemMessage(hwnd,DLG_TLER_ACLOCK,BM_SETCHECK,TRUE,0);
       }
       CentreDialogOnScreen(hwnd);
     }
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      if(dp == NULL)break;
      switch(LOWORD(wparam)){
        case DLG_TLER_PT:      i=TWEEN;     goto PT1;
        case DLG_TLER_ST:      i=SPLINEA;   goto PT1;
        case DLG_TLER_TOPATH:  i=TOPATH;    goto PT1;
        case DLG_TLER_LOOKAT:  i=TRACK;     goto PT1;
        case DLG_TLER_MIMIC:   i=COPY;      goto PT1;
          PT1:
          dp->type=i;
          OnPath=dp->topath;
          if(OnPath == NULL)status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_LOOKAT),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_MIMIC),status);
          if(OnPath != NULL && OnPath->type != PATH)status=FALSE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_TOPATH),status);
          if(OnPath != NULL)
            SetDlgItemText(hwnd,DLG_TLER_FOLLOWED,OnPath->actorname);
          else
            SetDlgItemText(hwnd,DLG_TLER_FOLLOWED,"<None>");
          if(dp->type == TWEEN || dp->type == SPLINEA)
            status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_FOLLOWED),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_THETA),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_ALPHA),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_PHI),!status);
          if(OnPath == NULL)status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_SELECT),status);
          break;
        case DLG_TLER_SELECT:
          if((OnPath=SelectNode(dp->anodename,NO,hwnd)) != NULL){
            if(OnPath == np){
              MessageBox(hwnd,"Actor Cannot Look at Itself",NULL,MB_OK);
              break;
            }
            dp->topath=OnPath;
            strcpy(dp->anodename,OnPath->actorname);
          }
          OnPath=dp->topath;
          if(OnPath == NULL)status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_LOOKAT),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_MIMIC),status);
          if(OnPath != NULL && OnPath->type != PATH)status=FALSE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_TOPATH),status);
          if(OnPath != NULL)
            SetDlgItemText(hwnd,DLG_TLER_FOLLOWED,OnPath->actorname);
          else
            SetDlgItemText(hwnd,DLG_TLER_FOLLOWED,"<None>");
          if(dp->type == TWEEN || dp->type == SPLINEA)
            status=FALSE; else status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_FOLLOWED),status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_THETA),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_ALPHA),!status);
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_PHI),!status);
          if(OnPath == NULL)status=TRUE;
          EnableWindow(GetDlgItem(hwnd,DLG_TLER_SELECT),status);
          break;
        case DLG_TLER_NA:  dp->im=0;  break;
        case DLG_TLER_VA:  dp->im=1;  break;
        case DLG_TLER_SA:  dp->im=2;  break;
        case DLG_TLER_FA:  dp->im=3;  break;
        case DLG_TLER_CLOCK:
          dp->ima = -1.0;
          break;
        case DLG_TLER_ACLOCK:
          dp->ima = 1.0;
          break;
        case DLG_TLER_COPY:
          if(dp->last != NULL){
            sprintf(title,"%.2lf",dp->last->theta);
            SetDlgItemText(hwnd,DLG_TLER_THETA,title);
            sprintf(title,"%.2lf",dp->last->alpha);
            SetDlgItemText(hwnd,DLG_TLER_ALPHA,title);
            sprintf(title,"%.2lf",dp->last->phi);
            SetDlgItemText(hwnd,DLG_TLER_PHI,title);
          }
          break;
        case IDCANCEL:
        case DLG_TLER_CANCEL:
          EndDialog(hwnd,-1);
          return(TRUE);
        case DLG_TLER_DELETE:
          EndDialog(hwnd,1);
          return(TRUE);
        case DLG_TLER_CUT:
          EndDialog(hwnd,2);
          return(TRUE);
        case IDOK:
        case DLG_TLER_OK:
          if(dp->type == TWEEN || dp->type == SPLINEA){
             dp->topath=NULL; dp->anodename[0]='\0';
          }
          i=GetDlgItemInt(hwnd,DLG_TLER_F,&err,FALSE);
          if(err)dp->firstframe=(short)i;
          if(dp->last != NULL && dp->firstframe <= dp->last->lastframe)
             dp->firstframe=dp->last->lastframe+1;
          if(dp->firstframe < 1)dp->firstframe=1;
          i=GetDlgItemInt(hwnd,DLG_TLER_L,&err,FALSE);
          if(err)dp->lastframe=(short)i;
          if(dp->next != NULL && dp->lastframe >= dp->next->firstframe)
             dp->lastframe=dp->next->firstframe-1;
          if(dp->lastframe > Nframes)dp->lastframe=Nframes;
          if(dp->lastframe < dp->firstframe)dp->lastframe=dp->firstframe;
          if(GetDlgItemText(hwnd,DLG_TLER_ALPHA,title,10) != 0){
             a=atof(title);
             if(a >= -360.0 && a <= 360.0)dp->alpha=a;
          }
          if(GetDlgItemText(hwnd,DLG_TLER_THETA,title,10) != 0){
             a=atof(title);
             if(a >= -360.0 && a <= 360.0)dp->theta=a;
          }
          if(GetDlgItemText(hwnd,DLG_TLER_PHI,title,10) != 0){
             a=atof(title);
             if(a >= -360.0 && a <= 360.0)dp->phi=a;
          }
          EndDialog(hwnd,0);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static LRESULT CALLBACK DlgCommandDlgProc(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam){
 node *np;
 char title[128];
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     np=(node *)lparam;
     sprintf(title,"Edit: %s",np->actorname);
     SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)title);
     CentreDialogOnScreen(hwnd);
     return (TRUE);
   case WM_DESTROY:
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,-1);
          return(TRUE);
        case DLG_KEYF_COMMAND_DELETE:
          EndDialog(hwnd,IDM_TIMELINE_ACTOR_DELETE);
          return(TRUE);
        case DLG_KEYF_COMMAND_RENAME:
          EndDialog(hwnd,IDM_TIMELINE_ACTOR_RENAME);
          return(TRUE);
        case DLG_KEYF_COMMAND_INFO:
          EndDialog(hwnd,IDM_TIMELINE_ACTOR_INFO);
          return(TRUE);
        case DLG_KEYF_COMMAND_SAVEANI:
          EndDialog(hwnd,IDM_TIMELINE_ACTOR_SAVEANIOBJ);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

#define  CYORDERDLGLINE 18          // Height of a line.
static HWND hwndOrderDlg;           // Window handle of the Order/Group dlg.
static HWND hwndOrderList;          // Window handle of the list box.
static WNDPROC lpfnOldLBWndProc;    // Original list box window proc.
static BOOL cSelItems=FALSE;

static BOOL OrderDlgInsertHitTest(INT y,PINT piInsert){
 INT i;
 INT iPixel;
 INT iInsert;
 INT iLine;
 BOOL fInsertZone;
 if (!cSelItems) return FALSE;
 iPixel = y % CYORDERDLGLINE;
 if (iPixel < 3) {
   iLine = y / CYORDERDLGLINE;
   fInsertZone = TRUE;
 }
 else if (iPixel > CYORDERDLGLINE - 3) {
   iLine = (y / CYORDERDLGLINE) + 1;
   fInsertZone = TRUE;
 }
 else {
   fInsertZone = FALSE;
 }
 if (fInsertZone) {
   iInsert = iLine + (INT)SendMessage(hwndOrderList,
                          LB_GETTOPINDEX, 0, 0L);
   if (iInsert > Nnodes)fInsertZone = FALSE;
   else *piInsert = iInsert;
 }
 return fInsertZone;
}

static LRESULT CALLBACK OrderDlgLBWndProc(HWND hwnd,UINT msg,WPARAM wParam,
                                    LPARAM lParam){
 INT iInsert;
 int item;
 switch (msg) {
   case WM_SETCURSOR:
     if (LOWORD(lParam) == HTCLIENT)return TRUE;
     break;
    case WM_MOUSEMOVE:
     if (!(GetKeyState(VK_LBUTTON) & 0x8000) &&
           OrderDlgInsertHitTest(HIWORD(lParam), &iInsert))
       SetCursor(ghcurInsert);
     else
       SetCursor(ghcurArrow);
     break;
   case WM_LBUTTONDOWN:
     if(OrderDlgInsertHitTest(HIWORD(lParam), &iInsert)) {
       item=SendMessage(hwndOrderList,LB_GETCURSEL,0,0);
       if(item == 0 || iInsert == 0){
         SendPrgmQuery(IDQ_CAMERANOREORDER,0);
         SetFocus(hwnd);
       }
       else if(item != LB_ERR){
         char temp[64];
         cSelItems=FALSE;
         SendMessage(hwndOrderList,LB_GETTEXT,item,(LPARAM)temp);
         if(item > iInsert){
           SendMessage(hwndOrderList,LB_DELETESTRING,(WPARAM)item,0);
           SendMessage(hwndOrderList,LB_INSERTSTRING,(WPARAM)iInsert,(LPARAM)temp);
         }
         else if(item < iInsert){
           SendMessage(hwndOrderList,LB_INSERTSTRING,(WPARAM)iInsert,(LPARAM)temp);
           SendMessage(hwndOrderList,LB_DELETESTRING,(WPARAM)item,0);
         }
         if(iInsert != item){
           node *n_out=NULL,*n_in=NULL,*Np,*Lp;
           int i;
           Np=FirstNp; i=0; while(Np != NULL){
             if(i == item)  n_out=Np;
             if(i == iInsert)n_in=Np;
             Lp=Np;
             i++; Np=Np->next;
           }
           if(n_out != NULL){
             if(n_out->next != NULL)n_out->next->last = n_out->last;
             if(n_out->last != NULL)n_out->last->next = n_out->next;
             if(n_out == FirstNp)FirstNp=n_out->next;
             if(n_out == MainNp)MainNp=n_out->last;
             if(n_in != NULL){
               n_out->last=n_in->last;
               if(n_in->last != NULL)n_in->last->next=n_out;
               n_in->last=n_out;
               n_out->next=n_in;
               if(n_in == FirstNp)FirstNp=n_out;
             }
             else{ /* at end of list */
               if(Lp != MainNp)MessageBeep(MB_OK);
               n_out->next=NULL;
               n_out->last=Lp;
               Lp->next=n_out;
               MainNp=n_out;
             }
           }
         }
       }
       return FALSE;
     }
     break;
  }
  return CallWindowProc((WNDPROC)lpfnOldLBWndProc,hwnd,msg,wParam,lParam);
}

static void OrderDlgInit(HWND hwnd){
 int nItem;
 node *Np;
 hwndOrderDlg = hwnd;
 hwndOrderList = GetDlgItem(hwnd,DLG_REORDER_LIST);
 cSelItems=FALSE;
 Np=FirstNp; while(Np != NULL){
   nItem=SendMessage(hwndOrderList,LB_ADDSTRING,0,(LPARAM)Np->actorname);
   Np=Np->next;
 }
 lpfnOldLBWndProc = (WNDPROC)SetWindowLong(hwndOrderList,
                     GWL_WNDPROC, (DWORD)OrderDlgLBWndProc);
 CentreDialogOnScreen(hwnd);
}

static LRESULT CALLBACK ReorderDlgProc(HWND hwnd, UINT msg,
                                       WPARAM wParam, LPARAM lParam){
 switch (msg) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     OrderDlgInit(hwnd);
     return TRUE;
   case WM_MEASUREITEM:
     ((LPMEASUREITEMSTRUCT)lParam)->itemHeight = CYORDERDLGLINE;
     return TRUE;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis=(LPDRAWITEMSTRUCT)lParam;
       TCHAR szItem[MAX_PATH];
       TEXTMETRIC tm;
       int y;
       if(lpdis->itemState & ODS_SELECTED) {
         SetBkColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHT));
         SetTextColor(lpdis->hDC, GetSysColor(COLOR_HIGHLIGHTTEXT));
       }
       else {
         SetBkColor(lpdis->hDC, GetSysColor(COLOR_WINDOW));
         SetTextColor(lpdis->hDC, GetSysColor(COLOR_WINDOWTEXT));
       }
       switch(lpdis->itemAction){
         case ODA_SELECT:
         case ODA_DRAWENTIRE:
           SendMessage(lpdis->hwndItem,LB_GETTEXT,lpdis->itemID,(LPARAM)szItem);
           GetTextMetrics(lpdis->hDC,&tm);
           y=(lpdis->rcItem.bottom+lpdis->rcItem.top-tm.tmHeight)/2;
           ExtTextOut(lpdis->hDC,6,y,
                    ETO_OPAQUE | ETO_CLIPPED, &lpdis->rcItem,
                    szItem, lstrlen(szItem), NULL);
         default:
           break;
       }
       if(lpdis->itemState & ODS_FOCUS)
         DrawFocusRect(lpdis->hDC, &lpdis->rcItem);
     }
     return TRUE;
   case WM_COMMAND:
     switch (LOWORD(wParam)) {
       case DLG_REORDER_LIST:
         switch (HIWORD(wParam)) {
           case LBN_SELCHANGE:
             cSelItems=TRUE;
             break;
         }
         break;
       case IDCANCEL:
       case IDOK:
         EndDialog(hwnd, LOWORD(wParam));
         break;
     }
     return TRUE;
   default:
     return FALSE;
 }
 return FALSE;
}
// script part 3

#define UPDATETRIVIEW(x)           \
  if(ghwndTimeline != NULL){       \
    if(!IsZoomed(ghwndTimeline)){  \
      ReDrawStageDisplay(TRUE);    \
      if(ghwndOpenGLview == NULL)  \
        PerspectiveView(0,1);      \
      else UpdateGLview(TRUE);     \
    }                              \
  }

#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define A_H 60
#define F_W 10
#define L_H 10
#define WM_KEYFRAMEUPDATE (WM_USER+1)

static void DoKeyFramerCommand(WPARAM wparam, int local_actor);

static char  szTimelineChildClass[]="OFX:TimelineChildClass";
static char  szTimelineClass[]="OFX:TimelineClass";
static int TimelineStatusPartsList[3]={110,220,-1};
static int FirstTime=TRUE,IndicatorVisible = NO,IndicatorActive=NO,
           IndicatedActor= -1, IndicatedFrame= -1, IndicatedLine= -1,
           ScrollingTimeLine=0,TimeLineUpdateFlag=0,TimeLineVisibleFlag=0,
           IndicatorsChanged=NO;
static HBITMAP  hTimelineBitmap=NULL;
static HWND     hwndTimelineInfo=NULL;
static node     *IndicatedNode=NULL;
static sky      *IndicatedSky=NULL;
static director *IndicatedDirector=NULL;
static object   *IndicatedObject=NULL;
static position *IndicatedPosition=NULL;
static align    *IndicatedAlign=NULL;
static size     *IndicatedSize=NULL;
static DWORD fdwStyleP = WS_POPUP | WS_CLIPSIBLINGS | WS_CAPTION |
                         WS_SYSMENU | WS_THICKFRAME |
                         WS_MAXIMIZEBOX | WS_MINIMIZEBOX |
                         !WS_VSCROLL | !WS_HSCROLL;
static DWORD fdwStyleD = WS_CHILD | WS_CLIPSIBLINGS | !WS_CAPTION |
                         !WS_SYSMENU | !WS_THICKFRAME | !WS_MAXIMIZEBOX |
                         !WS_VSCROLL | !WS_HSCROLL | !WS_BORDER;
static DWORD fdwStyleC = WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_VISIBLE;
static int xScreen,yScreen,nXBorder,nXTitle,nYTitle;
static POINT ptSize;
static long NoldFrames= -1, NoldNodes = -1;
static char *linetype[5]={NULL,NULL,NULL,NULL,NULL};

static int ChangeIndicatedKeys(BOOL bEditKeyStart){
 int dn,fn,update=0;

 if(IndicatedObject != NULL){
   if(bEditKeyStart){ /* moving start of timeline */
     if(IndicatedObject->last == NULL)dn=1;
     else dn=IndicatedObject->last->firstframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedObject->next == NULL)dn=Nframes;
       else dn=IndicatedObject->next->firstframe-1;
       if(IndicatedFrame <= dn){
         IndicatedObject->firstframe=IndicatedFrame;
         if(IndicatedObject->firstframe > IndicatedObject->lastframe)
            IndicatedObject->firstframe=IndicatedObject->lastframe;
         if(IndicatedObject->last != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedObject->last->lastframe=min(IndicatedFrame-1,
                 IndicatedObject->last->lastframe);
           else IndicatedObject->last->lastframe=IndicatedFrame-1;
         }
       }
     }
   }
   else{  /* moving end of timeline */
     if(IndicatedObject->last == NULL)dn=1;
     else dn=IndicatedObject->last->lastframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedObject->next == NULL)dn=Nframes;
       else dn=IndicatedObject->next->lastframe-1;
       if(IndicatedFrame <= dn){
         IndicatedObject->lastframe=IndicatedFrame;
         if(IndicatedObject->firstframe > IndicatedObject->lastframe)
            IndicatedObject->firstframe=IndicatedObject->lastframe;
         if(IndicatedObject->next != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedObject->next->firstframe=max(IndicatedFrame+1,
                 IndicatedObject->next->firstframe);
           else IndicatedObject->next->firstframe=IndicatedFrame+1;
         }
       }
     }
   }
   update=1;
 }

 if(IndicatedSky != NULL){
   if(bEditKeyStart){ /* moving start of timeline */
     if(IndicatedSky->last == NULL)dn=1;
     else dn=IndicatedSky->last->firstframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedSky->next == NULL)dn=Nframes;
       else dn=IndicatedSky->next->firstframe-1;
       if(IndicatedFrame <= dn){
         IndicatedSky->firstframe=IndicatedFrame;
         if(IndicatedSky->firstframe > IndicatedSky->lastframe)
            IndicatedSky->firstframe=IndicatedSky->lastframe;
         if(IndicatedSky->last != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedSky->last->lastframe=min(IndicatedFrame-1,
                 IndicatedSky->last->lastframe);
           else IndicatedSky->last->lastframe=IndicatedFrame-1;
         }
       }
     }
   }
   else{
     if(IndicatedSky->last == NULL)dn=1;
     else dn=IndicatedSky->last->lastframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedSky->next == NULL)dn=Nframes;
       else dn=IndicatedSky->next->lastframe-1;
       if(IndicatedFrame <= dn){
         IndicatedSky->lastframe=IndicatedFrame;
         if(IndicatedSky->firstframe > IndicatedSky->lastframe)
            IndicatedSky->firstframe=IndicatedSky->lastframe;
         if(IndicatedSky->next != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedSky->next->firstframe=max(IndicatedFrame+1,
                 IndicatedSky->next->firstframe);
           else  IndicatedSky->next->firstframe=IndicatedFrame+1;
         }
       }
     }
   }
   update=1;
 }

 if(IndicatedDirector != NULL){
   if(bEditKeyStart){ /* moving start of timeline */
     if(IndicatedDirector->last == NULL)dn=1;
     else dn=IndicatedDirector->last->firstframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedDirector->next == NULL)dn=Nframes;
       else dn=IndicatedDirector->next->firstframe-1;
       if(IndicatedFrame <= dn){
         IndicatedDirector->firstframe=IndicatedFrame;
         if(IndicatedDirector->firstframe > IndicatedDirector->lastframe)
            IndicatedDirector->firstframe=IndicatedDirector->lastframe;
         if(IndicatedDirector->last != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedDirector->last->lastframe=min(IndicatedFrame-1,
                 IndicatedDirector->last->lastframe);
           else IndicatedDirector->last->lastframe=IndicatedFrame-1;
         }
       }
     }
   }
   else{
     if(IndicatedDirector->last == NULL)dn=1;
     else dn=IndicatedDirector->last->lastframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedDirector->next == NULL)dn=Nframes;
       else dn=IndicatedDirector->next->lastframe-1;
       if(IndicatedFrame <= dn){
         IndicatedDirector->lastframe=IndicatedFrame;
         if(IndicatedDirector->firstframe > IndicatedDirector->lastframe)
            IndicatedDirector->firstframe=IndicatedDirector->lastframe;
         if(IndicatedDirector->next != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedDirector->next->firstframe=max(IndicatedFrame+1,
                 IndicatedDirector->next->firstframe);
           else  IndicatedDirector->next->firstframe=IndicatedFrame+1;
         }
       }
     }
   }
   update=1;
 }

 if(IndicatedPosition != NULL){
   if(bEditKeyStart){ /* moving start of timeline */
     if(IndicatedPosition->last == NULL)dn=1;
     else dn=IndicatedPosition->last->firstframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedPosition->next == NULL)dn=Nframes;
       else dn=IndicatedPosition->next->firstframe-1;
       if(IndicatedFrame <= dn){
         IndicatedPosition->firstframe=IndicatedFrame;
         if(IndicatedPosition->firstframe > IndicatedPosition->lastframe)
            IndicatedPosition->firstframe=IndicatedPosition->lastframe;
         if(IndicatedPosition->last != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedPosition->last->lastframe=min(IndicatedFrame-1,
                 IndicatedPosition->last->lastframe);
           else IndicatedPosition->last->lastframe=IndicatedFrame-1;
         }
       }
     }
   }
   else{
     if(IndicatedPosition->last == NULL)dn=1;
     else dn=IndicatedPosition->last->lastframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedPosition->next == NULL)dn=Nframes;
       else  dn=IndicatedPosition->next->lastframe-1;
       if(IndicatedFrame <= dn){
         IndicatedPosition->lastframe=IndicatedFrame;
         if(IndicatedPosition->firstframe > IndicatedPosition->lastframe)
            IndicatedPosition->firstframe=IndicatedPosition->lastframe;
         if(IndicatedPosition->next != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedPosition->next->firstframe=max(IndicatedFrame+1,
                 IndicatedPosition->next->firstframe);
           else  IndicatedPosition->next->firstframe=IndicatedFrame+1;
         }
       }
     }
   }
   update=1;
 }
 if(IndicatedAlign != NULL){
   if(bEditKeyStart){ /* moving start of timeline */
     if(IndicatedAlign->last == NULL)dn=1;
     else dn=IndicatedAlign->last->firstframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedAlign->next == NULL)dn=Nframes;
       else dn=IndicatedAlign->next->firstframe-1;
       if(IndicatedFrame <= dn){
         IndicatedAlign->firstframe=IndicatedFrame;
         if(IndicatedAlign->firstframe > IndicatedAlign->lastframe)
            IndicatedAlign->firstframe=IndicatedAlign->lastframe;
         if(IndicatedAlign->last != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedAlign->last->lastframe=min(IndicatedFrame-1,
                 IndicatedAlign->last->lastframe);
           else IndicatedAlign->last->lastframe=IndicatedFrame-1;
         }
       }
     }
   }
   else{
     if(IndicatedAlign->last == NULL)dn=1;
     else dn=IndicatedAlign->last->lastframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedAlign->next == NULL)dn=Nframes;
       else dn=IndicatedAlign->next->lastframe-1;
       if(IndicatedFrame <= dn){
         IndicatedAlign->lastframe=IndicatedFrame;
         if(IndicatedAlign->firstframe > IndicatedAlign->lastframe)
            IndicatedAlign->firstframe=IndicatedAlign->lastframe;
         if(IndicatedAlign->next != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedAlign->next->firstframe=max(IndicatedFrame+1,
                 IndicatedAlign->next->firstframe);
           else  IndicatedAlign->next->firstframe=IndicatedFrame+1;
         }
       }
     }
   }
   update=1;
 }
 if(IndicatedSize != NULL){
   if(bEditKeyStart){ /* moving start of timeline */
     if(IndicatedSize->last == NULL)dn=1;
     else dn=IndicatedSize->last->firstframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedSize->next == NULL)dn=Nframes;
       else dn=IndicatedSize->next->firstframe-1;
       if(IndicatedFrame <= dn){
         IndicatedSize->firstframe=IndicatedFrame;
         if(IndicatedSize->firstframe > IndicatedSize->lastframe)
            IndicatedSize->firstframe=IndicatedSize->lastframe;
         if(IndicatedSize->last != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedSize->last->lastframe=min(IndicatedFrame-1,
                 IndicatedSize->last->lastframe);
           else IndicatedSize->last->lastframe=IndicatedFrame-1;
         }
       }
     }
   }
   else{
     if(IndicatedSize->last == NULL)dn=1;
     else dn=IndicatedSize->last->lastframe+1;
     if(IndicatedFrame >= dn){
       if(IndicatedSize->next == NULL)dn=Nframes;
       else dn=IndicatedSize->next->lastframe-1;
       if(IndicatedFrame <= dn){
         IndicatedSize->lastframe=IndicatedFrame;
         if(IndicatedSize->firstframe > IndicatedSize->lastframe)
          IndicatedSize->firstframe=IndicatedSize->lastframe;
         if(IndicatedSize->next != NULL){
           if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
                 IndicatedSize->next->firstframe = max(IndicatedFrame+1,
                 IndicatedSize->next->firstframe);
           else  IndicatedSize->next->firstframe=IndicatedFrame+1;
         }
       }
     }
   }
   update=1;
 }
 if(update)return 1;
 return 0;
}

void WaitForUp(void){
 MSG msg;
 POINT p;
 RECT  r;
 GetCursorPos(&p);
 r.top=p.y; r.left=p.x;
 r.bottom=p.y; r.right=p.x;
 ClipCursor(&r);
 while (1) {
   GetMessage(&msg,NULL,0,0);
   TranslateMessage(&msg);
     DispatchMessage(&msg);
   if(msg.message == WM_LBUTTONUP)break;
 }
}

static node *EditTimeLine(int actor, int frame, int line, BOOL *keystart){
 node     *Np;
 object   *Op,*Opp;
 sky      *Sp,*Spp;
 director *Dp,*Dpp;
 position *Pp,*Ppp;
 align    *Ap,*App;
 size     *Xp,*Xpp;
 int i,lf;
 short slf,sframe,type;
 EDIT_ACTION=YES;
 IndicatedSky=NULL;      IndicatedObject=NULL;
 IndicatedPosition=NULL; IndicatedAlign=NULL; IndicatedSize=NULL;
 IndicatedDirector=NULL;
 *keystart=FALSE;
 Np=FirstNp; if(actor > 1)for(i=1;i<actor;i++)Np=Np->next;
 if(Np == NULL)return NULL;

 if(line == 0 && Np->type == SKY){
  Sp=Np->fsky; Spp=Sp;
  while(Sp != NULL){
    Spp=Sp;
    if(Sp->firstframe > frame){
      if(Sp->last == NULL)lf=1;
      else                lf=Sp->last->lastframe+1;
      WaitForUp();
      EditObject(Np,(short)lf,(short)frame,YES,NO);
      UPDATETIMELINES(IDC_TIMEL_UPDATE)
      return NULL;
    }
    if(Sp->firstframe <= frame && Sp->lastframe >= frame){
      if(Sp->lastframe == frame){
        IndicatedSky=Sp;
        return Np;
      }
      else if(Sp->firstframe == frame){
        IndicatedSky=Sp;
        *keystart=TRUE;
        return Np;
      }
      else{
        WaitForUp();
        if(!(GetKeyState(VK_CONTROL) & 0x8000) &&
          SendPrgmQuery(IDQ_BREAKCHANNEL,1) == IDNO)return NULL;
        lf=Sp->firstframe;
        Sp->firstframe=frame+1;
        EditObject(Np,(short)lf,(short)frame,YES,YES);
        UPDATETIMELINES(IDC_TIMEL_UPDATE)
      }
      return NULL;
    }
    Sp=Sp->next;
  }
  if(Spp != NULL)lf=Spp->lastframe+1;
  else lf=1;
  WaitForUp();
  EditObject(Np,(short)lf,(short)frame,YES,NO);
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
  return NULL;
 }

 else if(line == 0 && Np->type == DIRECTOR){
  Dp=Np->fdirector; Dpp=Dp;
  while(Dp != NULL){
    Dpp=Dp;
    if(Dp->firstframe > frame){
      if(Dp->last == NULL)lf=1;
      else                lf=Dp->last->lastframe+1;
      WaitForUp();
      EditObject(Np,(short)lf,(short)frame,YES,NO);
      UPDATETIMELINES(IDC_TIMEL_UPDATE)
      return NULL;
    }
    if(Dp->firstframe <= frame && Dp->lastframe >= frame){
      if(Dp->lastframe == frame){
        IndicatedDirector=Dp;
        return Np;
      }
      else if(Dp->firstframe == frame){
        IndicatedDirector=Dp;
        *keystart=TRUE;
        return Np;
      }
      else{
        WaitForUp();
        if(!(GetKeyState(VK_CONTROL) & 0x8000) &&
          SendPrgmQuery(IDQ_BREAKCHANNEL,1) == IDNO)return NULL;
        lf=Dp->firstframe;
        Dp->firstframe=frame+1;
        EditObject(Np,(short)lf,(short)frame,YES,YES);
        UPDATETIMELINES(IDC_TIMEL_UPDATE)
      }
      return NULL;
    }
    Dp=Dp->next;
  }
  if(Dpp != NULL)lf=Dpp->lastframe+1;
  else lf=1;
  WaitForUp();
  EditObject(Np,(short)lf,(short)frame,YES,NO);
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
  return NULL;
 }

 else if (line == 0){
  if(Np->type == CAMERA)return NULL; //To allow Camera Timeline Editing MOD
  Op=Np->fobj; Opp=Op;
  while(Op != NULL){
    Opp=Op;
    if(Op->firstframe > frame){
      if(Op->last == NULL)lf=1;
      else                lf=Op->last->lastframe+1;
      WaitForUp();
      EditObject(Np,(short)lf,(short)frame,YES,NO);
      UPDATETIMELINES(IDC_TIMEL_UPDATE)
      UPDATETRIVIEW(0)
      return NULL;
    }
    if(Op->firstframe <= frame && Op->lastframe >= frame){
      if(Op->lastframe == frame){
        IndicatedObject=Op;
        return Np;
      }
      else if(Op->firstframe == frame){
        IndicatedObject=Op;
        *keystart=TRUE;
        return Np;
      }
      else{
        WaitForUp();
        if(!(GetKeyState(VK_CONTROL) & 0x8000) &&
          SendPrgmQuery(IDQ_BREAKCHANNEL,1) == IDNO)return NULL;
        lf=Op->firstframe;
        Op->firstframe=frame+1;
        if(Op->type == PATH){
           Op->v=ReTweenVelocity(Op,Op->v,
                                 Op->lastframe-lf+1,
                                 Op->lastframe-Op->firstframe+1,
                                 Op->pathlength,
                                 0);
        }
        EditObject(Np,(short)lf,(short)frame,YES,YES); // New bit
        UPDATETIMELINES(IDC_TIMEL_UPDATE)
        UPDATETRIVIEW(0)
      }
      return NULL;
    }
    Op=Op->next;
  }
  if(Opp != NULL)lf=Opp->lastframe+1;
  else lf=1;
  WaitForUp();
  EditObject(Np,(short)lf,(short)frame,YES,NO);
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
  UPDATETRIVIEW(0)
  return NULL;
 }
 if(line == 1){
  if(Np->type == SKY || Np->type == IMAGEP || Np->type == DIRECTOR)return NULL;
  Pp=Np->fpos; Ppp=Pp;
  while(Pp != NULL){
    Ppp=Pp;
    if(Pp->firstframe > frame){
      if(Pp->last == NULL)lf=1;
      else                lf=Pp->last->lastframe+1;
      WaitForUp();
      type=0; slf=(short)lf; sframe=(short)frame;
      EditPosition(Np,slf,sframe,type,ghwndTimeline);
      UPDATETIMELINES(IDC_TIMEL_UPDATE)
      UPDATETRIVIEW(0)
      return NULL;
    }
    if(Pp->firstframe <= frame && Pp->lastframe >= frame){
      if(Pp->lastframe == frame){
        IndicatedPosition=Pp;
        return Np;
      }
      else if(Pp->firstframe == frame){
        IndicatedPosition=Pp;
        *keystart=TRUE;
        return Np;
      }
      else{
        WaitForUp();
        if(!(GetKeyState(VK_CONTROL) & 0x8000) &&
          SendPrgmQuery(IDQ_BREAKCHANNEL,1) == IDNO)return NULL;
        MakeKey('P',frame,Np);
        UPDATETRIVIEW(0)
      }
      return NULL;
    }
    Pp=Pp->next;
  }
  if(Ppp != NULL)lf=Ppp->lastframe+1;
  else lf=1;
  WaitForUp();
  type=0; slf=(short)lf; sframe=(short)frame;
  EditPosition(Np,slf,sframe,type,ghwndTimeline);
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
  UPDATETRIVIEW(0)
  return NULL;
 }
 else if(line == 2){
  if(Np->type == SKY  || Np->type == TARGET || Np->type == DIRECTOR ||
     Np->type == PATH || Np->type == IMAGEP)return NULL;
  Ap=Np->fali; App=Ap;
  while(Ap != NULL){
    App=Ap;
    if(Ap->firstframe > frame){
      if(Ap->last == NULL)lf=1;
      else                lf=Ap->last->lastframe+1;
      WaitForUp();
      type=0; slf=(short)lf; sframe=(short)frame;
      EditAlign(Np,slf,sframe,type,ghwndTimeline);
      UPDATETIMELINES(IDC_TIMEL_UPDATE)
      UPDATETRIVIEW(0)
      return NULL;
    }
    if(Ap->firstframe <= frame && Ap->lastframe >= frame){
      if(Ap->lastframe == frame){
        IndicatedAlign=Ap;
        return Np;
      }
      else if(Ap->firstframe == frame){
        IndicatedAlign=Ap;
        *keystart=TRUE;
        return Np;
      }
      else{
        WaitForUp();
        if(!(GetKeyState(VK_CONTROL) & 0x8000) &&
          SendPrgmQuery(IDQ_BREAKCHANNEL,1) == IDNO)return NULL;
        MakeKey('A',frame,Np);
        UPDATETRIVIEW(0)
      }
      return NULL;
    }
    Ap=Ap->next;
  }
  if(App != NULL)lf=App->lastframe+1;
  else lf=1;
  WaitForUp();
  type=0; slf=(short)lf; sframe=(short)frame;
  EditAlign(Np,slf,sframe,type,ghwndTimeline);
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
  UPDATETRIVIEW(0)
  return NULL;
 }
 else if(line == 3){
  if(Np->type == SKY  || Np->type == TARGET || Np->type == DIRECTOR ||
     Np->type == PATH || Np->type == IMAGEP)return NULL;
  Xp=Np->fsiz; Xpp=Xp;
  while(Xp != NULL){
    Xpp=Xp;
    if(Xp->firstframe > frame){
      if(Xp->last == NULL)lf=1;
      else                lf=Xp->last->lastframe+1;
      WaitForUp();
      EditSize(Np,(short)lf,(short)frame);
      UPDATETIMELINES(IDC_TIMEL_UPDATE)
      UPDATETRIVIEW(0)
      return NULL;
    }
    if(Xp->firstframe <= frame && Xp->lastframe >= frame){
      if(Xp->lastframe == frame){
        IndicatedSize=Xp;
        return Np;
      }
      else if(Xp->firstframe == frame){
        IndicatedSize=Xp;
        *keystart=TRUE;
        return Np;
      }
      else{
        WaitForUp();
        if(!(GetKeyState(VK_CONTROL) & 0x8000) &&
          SendPrgmQuery(IDQ_BREAKCHANNEL,1) == IDNO)return NULL;
        MakeKey('Z',frame,Np);
        UPDATETRIVIEW(0)
      }
      return NULL;
    }
    Xp=Xp->next;
  }
  if(Xpp != NULL)lf=Xpp->lastframe+1;
  else lf=1;
  WaitForUp();
  EditSize(Np,(short)lf,(short)frame);
  UPDATETIMELINES(IDC_TIMEL_UPDATE)
  UPDATETRIVIEW(0)
  return NULL;
 }
 return NULL;
}

static void GetRealClientRect(HWND hwnd, PRECT lprc){
 DWORD dwStyle;
 dwStyle = GetWindowLong(hwnd, GWL_STYLE);
 GetClientRect(hwnd,lprc);
 if (dwStyle & WS_HSCROLL)lprc->bottom += GetSystemMetrics(SM_CYHSCROLL);
 if (dwStyle & WS_VSCROLL)lprc->right  += GetSystemMetrics(SM_CXVSCROLL);
}

static void SetScrollRanges(HWND hwnd,int update){
 RECT rc;
 int iRangeH,iRangeV,i;
 static int iSem = 0;
 if(!iSem){
   iSem++;
   GetRealClientRect(hwnd,&rc);
   for(i = 0; i < 2; i++){
     iRangeV = ptSize.y - rc.bottom;
     iRangeH = ptSize.x - rc.right;
     iRangeV = (iRangeV/A_H+1)*A_H;
     iRangeH = (iRangeH/F_W+1)*F_W;
     if(iRangeH < 0)iRangeH = 0;
     if(keyframer_docked)iRangeH=max(iRangeH,F_W); /* to keep hor scroll */
     if(iRangeV < 0)iRangeV = 0;
     SetScrollRange(hwnd, SB_VERT, 0, iRangeV, TRUE);
     SetScrollRange(hwnd, SB_HORZ, 0, iRangeH, TRUE);
     GetClientRect(hwnd, &rc);
   }
   iSem--;
   if(update)InvalidateRect(hwnd,NULL,TRUE);
 }
}

static void GetIndicators(int x, int y, RECT rc,
                          int *frame, int *actor, int *line,
                          int *firstframe, int *firstactor,
                          int *lastframe, int *lastactor){
 int f,a,ff,aa;
 ff=GetScrollPos(ghwndTimelineChild,SB_HORZ);
 aa=GetScrollPos(ghwndTimelineChild,SB_VERT);
 f=x+ff;
 a=y+aa;
 *frame =(f/F_W)+1;
 *actor =(a/A_H)+1;
 *line  =(a-(a/A_H)*A_H)/L_H - 2;
 *lastframe=(ff+rc.right)/F_W+1;
 *lastactor=(aa+rc.bottom)/A_H+1;
 *firstframe=ff/F_W+1;
 *firstactor=aa/A_H+1;
}

static void UpdateIndicators(int frame, int actor, int line,
                             int firstframe, int lastframe){
 RECT rc,rcW;
 char str[128];
 if(frame < 0)strcpy(str," ");
 else{
   sprintf(str,"%s %ld ",linetype[4],frame);
 }
 if(IsWindow(hwndTimelineInfo))
   SendMessage(hwndTimelineInfo,SB_SETTEXT,(WPARAM)0,(LPARAM)str);
 else SendPrgmText(str);
 strcpy(str," ");
 if(frame >= 0 && line >= 0 && line <= 3 && actor <= Nnodes)strcpy(str,linetype[line]);
 if(IsWindow(hwndTimelineInfo))
   SendMessage(hwndTimelineInfo,SB_SETTEXT,(WPARAM)1,(LPARAM)str);
}

static void DrawHairs(HDC ihdc, HWND hwnd, int frame, int actor, int line){
 int x,y,f,a,oldROP;
 HDC hdc;
 RECT rc;
 HPEN holdpen;
 HBRUSH holdbrush;
 if(ihdc == NULL)hdc=GetDC(hwnd);
 else hdc=ihdc;
 holdpen=SelectObject(hdc,GetStockObject(WHITE_PEN));
 oldROP=SetROP2(hdc,R2_XORPEN);
 GetClientRect(hwnd,&rc);
 f=GetScrollPos(ghwndTimelineChild,SB_HORZ)/F_W+1;
 a=GetScrollPos(ghwndTimelineChild,SB_VERT)/A_H+1;
 x=(frame-f)*F_W+F_W/3;
 y=(actor-a)*A_H+20+line*10;
 MoveToEx(hdc,x,rc.top,NULL); LineTo(hdc,x,rc.bottom);
#if 0
 if(line >= 0 && line <= 3 && frame <= Nframes){
   MoveToEx(hdc,x,rc.top,NULL); LineTo(hdc,x,y);
   holdbrush=SelectObject(hdc,GetStockObject(HOLLOW_BRUSH));
   Rectangle(hdc,x-5,y,x+5,y+9);
   MoveToEx(hdc,x,y+9,NULL); LineTo(hdc,x,rc.bottom);
   SelectObject(hdc,holdbrush);
 }
 else if(line == -1){
   MoveToEx(hdc,x-5,y+4,NULL); LineTo(hdc,x+6,y+4);
   MoveToEx(hdc,x-5,y-4,NULL); LineTo(hdc,x+6,y-4);
   MoveToEx(hdc,x,y-4,NULL); LineTo(hdc,x,y+5);
 }
 else {
   MoveToEx(hdc,x-5,y+4,NULL); LineTo(hdc,x+6,y+4);
   MoveToEx(hdc,x,rc.top,NULL); LineTo(hdc,x,rc.bottom);
 }
#endif
 SetROP2(hdc,oldROP);
 SelectObject(hdc,holdpen);
 if(ihdc == NULL)ReleaseDC(hwnd,hdc);
}

static void DrawTime(HDC hdc,int o,int y,int x0,int x1,HPEN hp, HPEN hps){
 HPEN holdpen;
// holdpen=SelectObject(hdc,hp);  // for single colour line
 holdpen=SelectObject(hdc,hps);
 MoveToEx(hdc,x0+1,y-3,NULL); LineTo(hdc,x0+1,y+3);
 MoveToEx(hdc,x0+1,y,NULL); LineTo(hdc,x1,y);
 y++;
 SelectObject(hdc,hps);
// Ellipse(hdc,x1,y-4,x1+7,y+3);
 Rectangle(hdc,x1,y-4,x1+7,y+3);
 SelectObject(hdc,holdpen);
}

static void DrawFrameIndicators(HDC hdc,int x, int y, int ff, int lf){
 int i;
 char str[128];
 SelectObject(hdc,GetStockObject(BLACK_PEN));
 for(i=ff;i<lf;i++,x += F_W){
   if(x >= 55){
     if(i%10 == 0){
       sprintf(str,"%ld",i);
       TextOut(hdc,x-4,y,str,strlen(str));
     }
     else{
       if(i < 100)Ellipse(hdc,x+1,y+6,x+5,y+10);
       else if(i < 999 && i%10 > 1)Ellipse(hdc,x+1,y+6,x+5,y+10);
       else if(i%10 > 2)Ellipse(hdc,x+1,y+6,x+5,y+10);
     }
   }
 }
}

static void DrawPopupButton(HDC hDC,int x,int y){
 HBITMAP hbm,hbmOld;
 BITMAP bmp;
 HDC hMemDC;
 int cx,cy;
 hbm=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_QT2));
 if(hbm != NULL){
   GetObject(hbm,sizeof(BITMAP),&bmp);
   cx = bmp.bmWidth;
   cy = bmp.bmHeight;
   hMemDC = CreateCompatibleDC(hDC);
   hbmOld = SelectObject(hMemDC,hbm);
   BitBlt(hDC,x,y+3,cx,cy,hMemDC,0,0,SRCAND);
   SelectObject(hMemDC,hbmOld);
   DeleteDC(hMemDC);
   DeleteObject(hbm);
 }
}

static void DrawActorNames(HWND hWnd, int fg, int ff, int lf){
 int a,na,ac=0;
 node *np;
 HDC hdc;
 RECT rc;
 hdc=GetDC(hWnd);
 GetClientRect(hWnd,&rc);
 SetBkMode(hdc,TRANSPARENT);
// SelectObject(hdc,GetStockObject(LTGRAY_BRUSH));
 SelectObject(hdc,GetStockObject(GRAY_BRUSH));
// SelectObject(hdc,GetStockObject(DKGRAY_BRUSH));
 SetTextColor(hdc,RGB(255,255,255));
 LoadString(ghinst_main,IDX_MISC_LANGUAGE,res_str,256);
 if(strcmp(res_str,"ENGLISH") == 0)
   SelectObject(hdc,GetStockObject(ANSI_VAR_FONT));
 a=GetScrollPos(ghwndTimelineChild,SB_VERT)/A_H+1;
 na=a+rc.bottom/A_H;
 if((np=FirstNp) != NULL)while(np != NULL){
   ac++;
   if(ac >= a && ac <= na){
     if(fg){
       DrawFrameIndicators(hdc,0,(ac-a)*A_H+7,ff,lf);
       TextOut(hdc,1,(ac-a)*A_H+7,np->actorname,min(strlen(np->actorname),8));
       SetTextColor(hdc,RGB(0,0,0));  // strange needed for SRCAND
       DrawPopupButton(hdc,45,(ac-a)*A_H+7);
       SetTextColor(hdc,RGB(255,255,255));
     }
//     else PatBlt(hdc,1,(ac-a)*A_H+1,100,20,PATCOPY);
     else PatBlt(hdc,0,(ac-a)*A_H+6,rc.right,14,PATCOPY);
   }
   np=np->next;
 }
 ReleaseDC(hWnd,hdc);
}

static void DrawActorTime(HDC hdc, HPEN hp, HPEN *hps, RECT rc, int f, int a,
                          int na, int fa, node *np, int ac, int f_f, int l_f){
 int ff,lf;
 sky *sp;
 director *dp;
 object *op;
 position *pp;
 align *ap;
 size  *xp;
 HBRUSH hOldBrush;
 HPEN   hPen,hOldPen;
 SelectObject(hdc,GetStockObject(BLACK_PEN));
 hOldBrush=SelectObject(hdc,GetStockObject(NULL_BRUSH));
 lf=(Nframes-f+1)*F_W;
 MoveToEx(hdc,0,(ac-a)*A_H+24,NULL); LineTo(hdc,lf,(ac-a)*A_H+24);
 MoveToEx(hdc,0,(ac-a)*A_H+34,NULL); LineTo(hdc,lf,(ac-a)*A_H+34);
 MoveToEx(hdc,0,(ac-a)*A_H+44,NULL); LineTo(hdc,lf,(ac-a)*A_H+44);
 MoveToEx(hdc,0,(ac-a)*A_H+54,NULL); LineTo(hdc,lf,(ac-a)*A_H+54);
 if(!ScrollingTimeLine){
   LoadString(ghinst_main,IDX_MISC_LANGUAGE,res_str,256);
   if(strcmp(res_str,"ENGLISH") == 0)
     SelectObject(hdc,GetStockObject(ANSI_VAR_FONT));
   SetTextColor(hdc,RGB(255,255,255));
   DrawFrameIndicators(hdc,0,(ac-a)*A_H+7,f_f,l_f);
   TextOut(hdc,1,(ac-a)*A_H+7,np->actorname,min(strlen(np->actorname),8));
   SetTextColor(hdc,RGB(0,0,0));
   DrawPopupButton(hdc,45,(ac-a)*A_H+7);
 }
 SelectObject(hdc,GetStockObject(BLACK_BRUSH));
 if(np->type == SKY){
   if((sp=np->fsky) != NULL)while(sp != NULL){
     ff=sp->firstframe; lf=sp->lastframe;
     if( (ff >= f && ff <= fa) ||
         (lf >= f && lf <= fa) ||
         (ff <  f && lf >  fa) ){
       DrawTime(hdc,0,(ac-a)*A_H+24,(ff-f)*F_W,(lf-f)*F_W,hp,hps[0]);
     }
     sp=sp->next;
   }
 }
 else if(np->type == DIRECTOR){
   if((dp=np->fdirector) != NULL)while(dp != NULL){
     ff=dp->firstframe; lf=dp->lastframe;
     if( (ff >= f && ff <= fa) ||
         (lf >= f && lf <= fa) ||
         (ff <  f && lf >  fa) ){
       DrawTime(hdc,0,(ac-a)*A_H+24,(ff-f)*F_W,(lf-f)*F_W,hp,hps[0]);
     }
     dp=dp->next;
   }
 }
 else{
   if((op=np->fobj) != NULL)while(op != NULL){
     ff=op->firstframe; lf=op->lastframe;
     if( (ff >= f && ff <= fa) ||
         (lf >= f && lf <= fa) ||
         (ff <  f && lf >  fa) ){
       DrawTime(hdc,0,(ac-a)*A_H+24,(ff-f)*F_W,(lf-f)*F_W,hp,hps[0]);
     }
     op=op->next;
   }
 }
 if((pp=np->fpos) != NULL)while(pp != NULL){
   ff=pp->firstframe; lf=pp->lastframe;
   if( (ff >= f && ff <= fa) ||
       (lf >= f && lf <= fa) ||
       (ff <  f && lf >  fa) ){
     DrawTime(hdc,1,(ac-a)*A_H+34,(ff-f)*F_W,(lf-f)*F_W,hp,hps[1]);
   }
   pp=pp->next;
 }
 if((ap=np->fali) != NULL)while(ap != NULL){
   ff=ap->firstframe; lf=ap->lastframe;
   if( (ff >= f && ff <= fa) ||
       (lf >= f && lf <= fa) ||
       (ff <  f && lf >  fa) ){
     DrawTime(hdc,2,(ac-a)*A_H+44,(ff-f)*F_W,(lf-f)*F_W,hp,hps[2]);
   }
   ap=ap->next;
 }
 if((xp=np->fsiz) != NULL)while(xp != NULL){
   ff=xp->firstframe; lf=xp->lastframe;
   if( (ff >= f && ff <= fa) ||
       (lf >= f && lf <= fa) ||
       (ff <  f && lf >  fa) ){
     DrawTime(hdc,3,(ac-a)*A_H+54,(ff-f)*F_W,(lf-f)*F_W,hp,hps[3]);
   }
   xp=xp->next;
 }
 hPen=CreatePen(PS_SOLID,2,RGB(192,192,192));
 hOldPen=SelectObject(hdc,hPen);
 MoveToEx(hdc,0,(ac-a)*A_H+59,NULL); LineTo(hdc,rc.right,(ac-a)*A_H+59);
 SelectObject(hdc,hOldPen);
 DeleteObject(hPen);
 SelectObject(hdc,hOldBrush);
 return;
}


// script part 4

static BOOL RegisterTimelineClass(HINSTANCE hInstance){
 WNDCLASS wc;
 /* main window */
 wc.lpszClassName = szTimelineClass;
 wc.style = 0;
 wc.lpfnWndProc = (WNDPROC)TimelineEditorWndProc;
 wc.cbClsExtra = 0;
 wc.cbWndExtra = 0;
 wc.hInstance = hInstance;
 wc.hIcon = LoadIcon(hInstance,"ANIMATORICON");
 wc.hCursor = LoadCursor(NULL,IDC_ARROW);
 wc.hbrBackground = (HBRUSH)GetStockObject(LTGRAY_BRUSH);
 if(keyframer_docked)wc.lpszMenuName = NULL;
 else                wc.lpszMenuName = MAKEINTRESOURCE(TIMELINE_MENU);
 if(!RegisterClass( &wc ))return FALSE;
 /* child window */
 wc.lpszClassName = szTimelineChildClass;
 wc.style = CS_DBLCLKS;
 wc.lpfnWndProc = (WNDPROC)TimelineChildWndProc;
 wc.cbClsExtra = 0;
 wc.cbWndExtra = 0;
 wc.hInstance = hInstance;
 wc.hIcon = NULL;
 wc.hCursor = LoadCursor(NULL,IDC_ARROW);
 wc.hbrBackground = (HBRUSH)GetStockObject(GRAY_BRUSH);
// wc.hbrBackground = (HBRUSH)GetStockObject(DKGRAY_BRUSH);
 wc.lpszMenuName = NULL;
 if(!RegisterClass( &wc ))return FALSE;
 FirstTime=FALSE;
 return TRUE;
}

static LRESULT CALLBACK TimelineEditorWndProc(HWND hWnd, UINT msg,
                            WPARAM wparam, LPARAM lparam ){
 int i;
 RECT  rc;
 static RECT ClientClip;
 switch( msg ){
   case WM_DESTROY:{
       char str[64];
       if(hTimelineBitmap != NULL)DeleteObject(hTimelineBitmap);
       hTimelineBitmap=NULL;
       DestroyWindow(ghwndTimelineChild);
       if(IsZoomed(hWnd) || IsIconic(hWnd) || keyframer_docked > 0)break;
       GetWindowRect(hWnd,&rc);
       sprintf(str,"%ld %ld %ld %ld",rc.left,rc.top,
               rc.right-rc.left,rc.bottom-rc.top);
       if(Preferences.use_positions)
         WritePrivateProfileString("KEYFRAMER","POSITION",str,IniFilename);
       for(i=0;i<5;i++){
         X__Free(linetype[i]); linetype[i]=NULL;
       }
     }
     break;
   case WM_CREATE: {
       HMENU hmenu = GetSystemMenu(hWnd, FALSE);
       RemoveMenu(hmenu, 7, MF_BYPOSITION);
//       RemoveMenu(hmenu, SC_MINIMIZE, MF_BYCOMMAND);
       RemoveMenu(hmenu, SC_TASKLIST, MF_BYCOMMAND);
       if(!keyframer_docked){
         hwndTimelineInfo=CreateWindowEx(0L,STATUSCLASSNAME,"",
           WS_CHILD | !WS_BORDER | WS_VISIBLE,
           0,0,0,0,
           hWnd,
           (HMENU)101,
           ghinst_main,
           NULL);
         SendMessage(hwndTimelineInfo,SB_SETPARTS,(WPARAM)3,
                     (LPARAM)TimelineStatusPartsList);
       }
       else hwndTimelineInfo=NULL;
       ghwndTimelineChild = CreateWindow(szTimelineChildClass," ",
                                         fdwStyleC,
                                         0,0,0,0,
                                         hWnd,
                                         (HMENU)102,
                                         ghinst_main,NULL);
       hTimelineBitmap=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TIMELINE));
     }
     break;
   case WM_PAINT: {
       HDC hdc;
       PAINTSTRUCT ps;
       hdc=BeginPaint(hWnd,&ps);
       if(keyframer_docked && hTimelineBitmap != NULL){
         HDC hMemDC;
         HBITMAP hbmOld;
         BITMAP bmp;
         hMemDC = CreateCompatibleDC(hdc);
         GetObject(hTimelineBitmap,sizeof(BITMAP),&bmp);
         hbmOld = SelectObject(hMemDC,hTimelineBitmap);
         BitBlt(hdc,0,0,bmp.bmWidth,bmp.bmHeight,hMemDC,0,0,SRCCOPY);
         SelectObject(hMemDC, hbmOld);
         DeleteDC(hMemDC);
       }
       EndPaint(hWnd,&ps);
     }
     break;
//#if 0
   case WM_ACTIVATE:
       if(IndicatorVisible)DrawHairs(NULL,ghwndTimelineChild,
         IndicatedFrame,IndicatedActor,IndicatedLine);
       if(wparam)IndicatorActive=YES;
       else      IndicatorActive=NO;
       IndicatorVisible=NO;
      break;
//#endif
   case WM_MOUSEMOVE:
   case WM_NCMOUSEMOVE:
     if(IndicatorVisible)DrawHairs(NULL,ghwndTimelineChild,
         IndicatedFrame,IndicatedActor,IndicatedLine);
     IndicatorVisible=NO;
     UpdateIndicators(-1,-1,-1,-1,-1);
     IndicatedFrame = -1; IndicatedActor = -1; IndicatedLine = -1;
     break;
   case WM_SIZE:{
       int yOff=0,xOff=0;
       if(keyframer_docked){
         yOff=1;
         xOff=56;
       }
       if(IsWindow(hwndTimelineInfo)){
         SendMessage(hwndTimelineInfo,msg,wparam,lparam);
         i=InfoWindowSize;
       }
       else i=0;
       MoveWindow(ghwndTimelineChild,xOff,yOff,
                  LOWORD(lparam)-xOff,
                  HIWORD(lparam)-i-yOff,TRUE);
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_MINIMIZE:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_FILE_KEYFRAMERDOCK,0);
         break;
       case SC_CLOSE:
         PostMessage(hWnd,WM_COMMAND,IDM_TIMELINE_HIDE,0);
         break;
       default:
         return( DefWindowProc( hWnd, msg, wparam, lparam ) );
         break;
     }
     break;
   case WM_COMMAND:{
     switch (LOWORD(wparam)){
        case IDM_TIMELINE_ACTOR_ADD:
          AddOnlyNewNode(hWnd,-1);
          UPDATETIMELINES(IDC_TIMEL_UPDATE)
          break;
        case IDM_TIMELINE_ACTOR_ORDER:
          DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_REORDER),
                         ghwndTimeline,(DLGPROC)ReorderDlgProc);
          UPDATETIMELINES(IDC_TIMEL_UPDATE)
          break;
        case IDM_TIMELINE_ACTOR_DELETE:
        case IDM_TIMELINE_ACTOR_RENAME:
        case IDM_TIMELINE_ACTOR_INFO:
        case IDM_TIMELINE_ACTOR_SAVEANIOBJ:
          DoKeyFramerCommand(wparam,0);
          break;
        case IDM_TIMELINE_FRAMES:
          SendMessage(ghwnd_main,WM_COMMAND,IDM_FRAME_SETNUMBER,0);
          break;
        case IDM_TIMELINE_HIDE:{
            char tt[128];
            ShowWindow(hWnd,SW_HIDE);
            TimeLineVisibleFlag=0;
            IndicatorVisible = NO;
            IndicatedActor= -1, IndicatedFrame= -1, IndicatedLine= -1;
            LoadString(ghinst_main,IDX_MISC_SKEYFRAMER,tt,128);
            ModifyMenu(GetMenu(ghwnd_main),IDM_FILE_KEYFRAMER,
                    MF_BYCOMMAND|MF_STRING,IDM_FILE_KEYFRAMER,tt);
          }
          break;
        case IDC_TIMEL_UPDATE:
          SendMessage(ghwndTimelineChild,WM_KEYFRAMEUPDATE,0,0);
          break;
        default: break;
     }
     break;
   }
   default:
     return( DefWindowProc( hWnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

static LRESULT CALLBACK TimelineChildWndProc(HWND hWnd, UINT msg,
                            WPARAM wparam, LPARAM lparam ){
 int i,iPos,iMax,iMin,dn,lastframe,lastactor,firstframe,firstactor;
 RECT  rc;
 POINT pt;
 static BOOL bTrackMouse=FALSE,bClientArea=FALSE,bEditKey=FALSE,
             bEditKeyStart=FALSE;
 static RECT ClientClip;
 switch( msg ){
   case WM_PAINT: {
       HDC hdc;
       PAINTSTRUCT ps;
       node *np;
       HPEN hp,hps[4];
       int f,a,ac=0,na,fa;
       hdc=BeginPaint(hWnd,&ps);
       SetBkMode(hdc,TRANSPARENT);
       f=GetScrollPos(hWnd,SB_HORZ)/F_W+1;
       a=GetScrollPos(hWnd,SB_VERT)/A_H+1;
       hp=CreatePen(PS_SOLID,2,RGB(255,0,0));
#if 0    // the original colours
       hps[0]=CreatePen(PS_SOLID,2,RGB(0,255,255));
       hps[1]=CreatePen(PS_SOLID,2,RGB(255,255,0));
       hps[2]=CreatePen(PS_SOLID,2,RGB(0,255,0));
       hps[3]=CreatePen(PS_SOLID,2,RGB(255,0,0));
#endif
       hps[0]=CreatePen(PS_SOLID,2,RGB(145,255,94));
       hps[1]=CreatePen(PS_SOLID,2,RGB(0,255,255));
       hps[2]=CreatePen(PS_SOLID,2,RGB(255,255,0));
       hps[3]=CreatePen(PS_SOLID,2,RGB(255,96,0));
       na=a+ps.rcPaint.bottom/A_H; fa=f+ps.rcPaint.right/F_W;
       if((np=FirstNp) != NULL)while(np != NULL){
         ac++;
         if(ac >= a && ac <= na){
           GetClientRect(hWnd,&rc);
           GetIndicators(0,0,rc,&i,&i,&i,&firstframe,&i,&lastframe,&i);
           DrawActorTime(hdc,hp,hps,ps.rcPaint,f,a,na,fa,np,ac,
                         firstframe,lastframe);
         }
         np=np->next;
       }
       if(IndicatorVisible){
         DrawHairs(hdc,hWnd,IndicatedFrame,IndicatedActor,IndicatedLine);
         UpdateWindow(hWnd);
       }
       DeleteObject(hp);
       DeleteObject(hps[0]); DeleteObject(hps[1]);
       DeleteObject(hps[2]); DeleteObject(hps[3]);
       EndPaint(hWnd,&ps);
     }
     break;
   case WM_LBUTTONDBLCLK:
     if(bEditKey){
       Save_UndoA();
       if(IndicatedLine == 0)EditObject(IndicatedNode,
                                        (short)IndicatedFrame,
                                        (short)IndicatedFrame,
                                        YES,NO);
       if(IndicatedLine == 1)EditPosition(IndicatedNode,
                             (short)IndicatedFrame,
                             (short)IndicatedFrame,-1,ghwndTimeline);
       if(IndicatedLine == 2)EditAlign(IndicatedNode,
                             (short)IndicatedFrame,
                             (short)IndicatedFrame,-1,ghwndTimeline);
       if(IndicatedLine == 3)EditSize(IndicatedNode,
                             (short)IndicatedFrame,(short)IndicatedFrame);
       IndicatedNode=NULL;
       UPDATETIMELINES(IDC_TIMEL_UPDATE)
       UPDATETRIVIEW(0)
       bEditKey=0;
     }
     /* could add keyframe command here */
     break;
   case WM_LBUTTONDOWN:
     if(IndicatorActive == YES){
//       SetCursor(ghcurCross); /* hide cursor */
       UpdateWindow(hWnd);
       if(!IndicatorVisible){
         GetClientRect(hWnd,&rc);
         GetIndicators((int)LOWORD(lparam),(int)HIWORD(lparam),rc,
                   &IndicatedFrame,&IndicatedActor,&IndicatedLine,
                   &firstframe,&firstactor,&lastframe,&lastactor);
         DrawHairs(NULL,hWnd,IndicatedFrame,IndicatedActor,IndicatedLine);
         UpdateWindow(hWnd);
         IndicatorVisible=YES;
       }
     }
     if(tool != NOTOOL){
       MessageBeep(MB_OK);
       break;
     }
     bEditKey=FALSE;
     if(IndicatedActor >  0 && IndicatedActor <= Nnodes  &&
        IndicatedFrame >  0 && IndicatedFrame <= Nframes &&
        IndicatedLine  >= 0 && IndicatedLine  <  4){
       int f,a;
       f=GetScrollPos(hWnd,SB_HORZ)/F_W;
       a=GetScrollPos(hWnd,SB_VERT)/A_H;
       SetCapture(hWnd);
       GetClientRect(hWnd,&rc);
       pt.x=0;
       pt.y=(IndicatedActor-1-a)*A_H+(IndicatedLine+2)*L_H;
       ClientClip.left=pt.x; ClientClip.top=pt.y;
       ClientToScreen(hWnd,&pt); rc.left=pt.x; rc.top=pt.y;
       pt.x=rc.right;  // pt.x=min(rc.right,(Nframes-f)*F_W);
       pt.y=min((IndicatedActor-1-a)*A_H+(IndicatedLine+3)*L_H,rc.bottom);
       ClientClip.right=pt.x; ClientClip.bottom=pt.y-1; //test
       ClientToScreen(hWnd,&pt); rc.right=pt.x; rc.bottom=pt.y;
       ClipCursor(&rc);
       bTrackMouse=TRUE;
       if((IndicatedNode=EditTimeLine(IndicatedActor,IndicatedFrame,
           IndicatedLine,&bEditKeyStart)) != NULL)bEditKey=TRUE;
     }
     else if((IndicatedLine == -1 || IndicatedLine == -2)
             && IndicatedActor <= Nnodes
//             && LOWORD(lparam)){
             && LOWORD(lparam) < 55 && LOWORD(lparam) > 42){
       Save_UndoA();
//       DoKeyFramerCommand(0,IndicatedActor);
       {
         POINT pt;
         HMENU hMenu=LoadMenu(ghinst_main,"KeyframerPopupMenu");
         node *local_node;
         if(hMenu == NULL)break;
         pt.x=LOWORD(lparam);
         pt.y=HIWORD(lparam);
         ClientToScreen(hWnd,&pt);
         local_node=FirstNp;
         if(IndicatedActor > 1)
           for(i=1;i<IndicatedActor;i++)local_node=local_node->next;
         if(local_node != NULL){
           if(local_node->type != NORMAL)
             EnableMenuItem(hMenu,IDM_TIMELINE_ACTOR_SAVEANIOBJ,MF_GRAYED);
           if(local_node->type == CAMERA ||
              local_node->type == DIRECTOR ||
              local_node->type == SKY)
             EnableMenuItem(hMenu,IDM_TIMELINE_ACTOR_INFO,MF_GRAYED);
         }
         TrackPopupMenu(GetSubMenu(hMenu,0),TPM_LEFTALIGN | TPM_LEFTBUTTON,
                        pt.x,pt.y,0,hWnd,NULL);
         DestroyMenu(hMenu);
       }
     }
     break;
   case WM_LBUTTONUP:
     if(IndicatorsChanged){
       IndicatorsChanged=NO;
       UPDATETRIVIEW(0)
       if(IndicatedNode != NULL){
         if(IndicatedObject != NULL && IndicatedObject->type == PATH){
           int dfn;
           dfn=IndicatedObject->lastframe - IndicatedObject->firstframe + 1;
           IndicatedObject->v=ReTweenVelocity(IndicatedObject,
                                              IndicatedObject->v,
                                              -1,dfn,
                                              IndicatedObject->pathlength,
                                              0);
         }
         CheckRecursiveFollow(IndicatedNode,1,(short)Nframes);
       }
     }
     if(bTrackMouse){
       bTrackMouse=FALSE;
       IndicatedSky=NULL;      IndicatedObject=NULL;
       IndicatedPosition=NULL; IndicatedAlign=NULL;
       IndicatedSize=NULL;     IndicatedDirector=NULL;
       ClipCursor(NULL);
       if(ScrollingTimeLine){
         if(IndicatorVisible)DrawHairs(NULL,hWnd,
                             IndicatedFrame,IndicatedActor,IndicatedLine);
         GetClientRect(hWnd,&rc);
         GetIndicators(0,0,rc,&i,&i,&i,&firstframe,&i,&lastframe,&i);
         DrawActorNames(hWnd,1,firstframe,lastframe);
         if(IndicatorVisible)DrawHairs(NULL,hWnd,
                             IndicatedFrame,IndicatedActor,IndicatedLine);
         ScrollingTimeLine=0;
       }
       ReleaseCapture();
     }
     break;
   case WM_MOUSEMOVE:
//     if(IndicatorActive == YES)SetCursor(ghcurCross); /* hidden cursor */
     bClientArea=TRUE;
     GetClientRect(hWnd,&rc);
     i=IndicatedFrame;
     iMin=IndicatedLine; iMax=IndicatedActor;
     GetIndicators((int)LOWORD(lparam),(int)HIWORD(lparam),rc,
                   &IndicatedFrame,&IndicatedActor,&IndicatedLine,
                   &firstframe,&firstactor,&lastframe,&lastactor);
     if(IndicatorActive && (i != IndicatedFrame ||
                         iMax != IndicatedActor || iMin != IndicatedLine)){
       UpdateWindow(hWnd);
       if(IndicatorVisible)DrawHairs(NULL,hWnd,i,iMax,iMin);
       DrawHairs(NULL,hWnd,IndicatedFrame,IndicatedActor,IndicatedLine);
       IndicatorVisible=YES;
     }
     UpdateIndicators(IndicatedFrame,IndicatedActor,IndicatedLine,
                      firstframe,lastframe);
     if(i != IndicatedFrame){
       if(ChangeIndicatedKeys(bEditKeyStart)){
         IndicatorsChanged=YES;
         InvalidateRect(hWnd,&ClientClip,TRUE);
         UpdateWindow(hWnd);
       }
     }
     if(bTrackMouse && (IndicatedFrame >= lastframe-1 ||
                        IndicatedFrame == firstframe) ){
       GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
       iPos = GetScrollPos(hWnd, SB_HORZ);
       if(IndicatedFrame >= lastframe-1)dn=  F_W;
       else                             dn= -F_W;
       if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
         if(IndicatorVisible)DrawHairs(NULL,hWnd,
           IndicatedFrame,IndicatedActor,IndicatedLine);
         DrawActorNames(hWnd,0,firstframe,lastframe);
         if(IndicatorVisible)DrawHairs(NULL,hWnd,
           IndicatedFrame,IndicatedActor,IndicatedLine);
         ScrollingTimeLine=1;
         ScrollWindow(hWnd, -dn, 0, NULL, NULL);
         SetScrollPos(hWnd, SB_HORZ, iPos + dn, TRUE);
         UpdateWindow(hWnd);
         GetCursorPos(&pt);
         if(IndicatedFrame >= lastframe-1)pt.x -= dn;
         else                             pt.x += (abs(dn));
         SetCursorPos(pt.x,pt.y);
       }
     }
     break;
   case WM_VSCROLL:
     GetScrollRange (hWnd, SB_VERT, &iMin, &iMax);
     iPos = GetScrollPos (hWnd, SB_VERT);
     GetClientRect (hWnd, &rc);
     switch (LOWORD(wparam)) {
       case SB_ENDSCROLL:
         dn=0;
         break;
       case SB_LINEDOWN:      dn =  A_H;                  break;
       case SB_LINEUP:        dn = -A_H;                  break;
       case SB_PAGEDOWN:      dn =  rc.bottom / 2  + A_H; break;
       case SB_PAGEUP:        dn = -rc.bottom / 2  - A_H; break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION:
         dn = HIWORD(wparam)-iPos;
         break;
       default:               dn = 0;
     }
     /* Limit scrolling to current scroll range */
     dn=(dn/A_H)*A_H;
     if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       ScrollWindow(hWnd, 0, -dn, NULL, NULL);
       SetScrollPos(hWnd, SB_VERT, iPos + dn, TRUE);
       IndicatedActor=(iPos+dn)/A_H+1;
     }
     break;
   case WM_HSCROLL:
     /* Calculate new horizontal scroll position */
     GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
     iPos = GetScrollPos(hWnd, SB_HORZ);
     GetClientRect (hWnd, &rc);
     switch (LOWORD(wparam)) {
       case SB_ENDSCROLL:
         if(ScrollingTimeLine){
           ScrollingTimeLine=0;
           GetIndicators(0,0,rc,&i,&i,&i,&firstframe,&i,&lastframe,&i);
           DrawActorNames(hWnd,1,firstframe,lastframe);
         }
         dn=0;
         break;
       case SB_LINEDOWN:      dn =  F_W;             break;
       case SB_LINEUP:        dn = -F_W;             break;
       case SB_PAGEDOWN:      dn =  rc.right/5+F_W;  break;
       case SB_PAGEUP:        dn = -rc.right/5+F_W;  break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION:
         dn = HIWORD(wparam) - iPos;
         break;
       default:               dn = 0;
     }
     /* Limit scrolling to current scroll range */
     dn=(dn/F_W)*F_W;
     if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       ScrollingTimeLine=1;
       GetIndicators(0,0,rc,&i,&i,&i,&firstframe,&i,&lastframe,&i);
       DrawActorNames(hWnd,0,firstframe,lastframe);
       ScrollWindow(hWnd, -dn, 0, NULL, NULL);
       SetScrollPos(hWnd, SB_HORZ, iPos + dn, TRUE);
       iMin=(iPos+dn)/F_W+1;
       iMax=(iPos+dn+rc.right)/F_W+1;
       UpdateIndicators(iMin,-1,-1,iMin,iMax);
     }
     break;
   case WM_SIZE:
     SetScrollRanges(hWnd,1);
     break;
   case WM_NCMOUSEMOVE:{
       if(!bClientArea)break;
       bClientArea=FALSE;
       if(IndicatorVisible)DrawHairs(NULL,hWnd,
         IndicatedFrame,IndicatedActor,IndicatedLine);
       IndicatorVisible=NO;
       UpdateIndicators(-1,-1,-1,-1,-1);
       IndicatedFrame = -1; IndicatedActor = -1; IndicatedLine = -1;
     }
     break;
   case WM_KEYFRAMEUPDATE:
     if(!TimeLineVisibleFlag)TimeLineUpdateFlag=1;
     else{
       TimeLineUpdateFlag=0;
       if(Nframes != NoldFrames || NoldNodes != Nnodes){
         ptSize.x=Nframes*F_W;   ptSize.y=Nnodes*A_H;
         SetScrollRanges(hWnd,0);
         if(Nframes < NoldFrames){
           GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
           SetScrollPos(hWnd,SB_HORZ,iMin,TRUE);
         }
         if(Nnodes  < NoldNodes){
           GetScrollRange(hWnd, SB_VERT, &iMin, &iMax);
           SetScrollPos(hWnd,SB_VERT,iMin,TRUE);
         }
         NoldFrames=Nframes; NoldNodes=Nnodes;
       }
       InvalidateRect(hWnd,NULL,TRUE);
       UpdateWindow(hWnd);
     }
     break;
   case WM_COMMAND:
     switch (LOWORD(wparam)){
        case IDM_TIMELINE_ACTOR_DELETE:
        case IDM_TIMELINE_ACTOR_RENAME:
        case IDM_TIMELINE_ACTOR_INFO:
        case IDM_TIMELINE_ACTOR_SAVEANIOBJ:
          DoKeyFramerCommand(wparam,IndicatedActor);
          break;
        default:
          break;
      }
   default:
     return( DefWindowProc( hWnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

static void DoKeyFramerCommand(WPARAM wparam, int local_actor){
 int i,id;
 node *local_node,*temp_node;
 char local_nodename[128];
 if(local_actor == 0){
   if((local_node=SelectNode(local_nodename,YES,ghwndTimeline))
                              == NULL)return;
 }
 else {
   local_node=FirstNp;
   if(local_actor > 1)for(i=1;i<local_actor;i++)local_node=local_node->next;
   if(local_node == NULL)return;
//   if((id=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_KEYF_COMMAND),
//       ghwndTimeline,(DLGPROC)DlgCommandDlgProc,(LPARAM)local_node)) < 0)
//     return;
//   wparam=id;
   strcpy(local_nodename,local_node->actorname);
 }
 switch (LOWORD(wparam)){
   case IDM_TIMELINE_ACTOR_DELETE:
     if(local_node->type == CAMERA){
       if(local_node == FirstNp){
         SendPrgmQuery(IDQ_CAMERANODELETE,0);
         break;
       }
       if(local_node == SelectedCamera)SelectedCamera=FirstNp;
       DeleteNode(local_node);
       SelectedNode=NULL;
       UpdateSelectedActor(FALSE);
       Ncameras--;
     }
     else {
       if(local_node->type == GROUND)Nground=0;
       if(local_node->type == SKY)Nskys=0;
       if(local_node->type == DIRECTOR)Ndirectors=0;
       if(local_node->type == ROBOT)Nrobots--;
       DeleteNode(local_node);
       SelectedNode=NULL;
       UpdateSelectedActor(FALSE);
     }
     UPDATETIMELINES(IDC_TIMEL_UPDATE)
     UPDATETRIVIEW(0)
     break;
   case IDM_TIMELINE_ACTOR_RENAME:
     LoadString(ghinst_main,IDX_MISC_RENAMEOBJECT,res_str,256);
     RequestCharString(64,local_node->actorname,res_str);
     UPDATETIMELINES(IDC_TIMEL_UPDATE)
     break;
   case IDM_TIMELINE_ACTOR_INFO:
     temp_node=SelectedNode;
     SelectedNode=local_node;
     SelectedInfo();
     SelectedNode=temp_node;
     break;
   case IDM_TIMELINE_ACTOR_SAVEANIOBJ:
     SaveAnimFile(local_node);
     break;
 }
}

void TimelineEditor(void){
 char str[64];
 char tt[128];
 int i,x,y,cx,cy,nf;
 if(FirstTime){
   if(!RegisterTimelineClass(ghinst_main))return;
   xScreen     = GetSystemMetrics(SM_CXSCREEN);
   yScreen     = GetSystemMetrics(SM_CYSCREEN);
   nXBorder    = (INT)GetSystemMetrics(SM_CXFRAME);
   nXTitle     = (INT)GetSystemMetrics(SM_CXSIZE);
   nYTitle     = (INT)GetSystemMetrics(SM_CYSIZE);
   x=0; y=nYTitle+GetSystemMetrics(SM_CYFRAME)+GetSystemMetrics(SM_CYMENU)
                 +GetSystemMetrics(SM_CYBORDER);
   if(Xres >= 800)nf=60; else nf=40;
   cx=(max(nf,Nframes))*F_W;   cy=(max(3,Nnodes))*A_H+25;
   GetPrivateProfileString("KEYFRAMER","POSITION","$$$",
                   str,sizeof(str),IniFilename);
   if(Preferences.use_positions && strncmp(str,"$$$",3) != 0){ /* if valid paramaters */
     int xx,yy,cxx,cyy;
     sscanf(str,"%ld %ld %ld %ld",&xx,&yy,&cxx,&cyy);
     if(xx < Xres-10 && yy < Yres-10){
       x=xx; y=yy;
       cx=cxx;
       cy=cyy;
     }
   }
   NoldFrames=Nframes; NoldNodes=Nnodes;
   for(i=0;i<5;i++)linetype[i]=X__Malloc(16);
   if(linetype[0] != NULL)LoadString(ghinst_main,IDX_MISC_LINE1,linetype[0],16);
   if(linetype[1] != NULL)LoadString(ghinst_main,IDX_MISC_LINE2,linetype[1],16);
   if(linetype[2] != NULL)LoadString(ghinst_main,IDX_MISC_LINE3,linetype[2],16);
   if(linetype[3] != NULL)LoadString(ghinst_main,IDX_MISC_LINE4,linetype[3],16);
   if(linetype[4] != NULL)LoadString(ghinst_main,IDX_MISC_LINE0,linetype[4],16);
   LoadString(ghinst_main,IDX_MISC_KEYFRAMER,str,64);
   if(keyframer_docked){
     RECT rc;
     GetClientRect(ghwnd_main,&rc);
     if((ghwndTimeline = CreateWindow(szTimelineClass,str,
        fdwStyleD,
        0,KeyframerYpos,
        rc.right,KeyframerSize,
        ghwnd_main,NULL,ghinst_main,NULL)) == NULL)return;
   }
   else{
     if((ghwndTimeline = CreateWindow(szTimelineClass,str,
        fdwStyleP,
        x,y,0,0,ghwnd_main,NULL,ghinst_main,NULL)) == NULL)return;
     SetWindowPos(ghwndTimeline,(HWND)NULL,0,0,
       min(Xres,cx),min(Yres-100,cy),SWP_NOMOVE | SWP_NOZORDER);
   }
   ShowWindow(ghwndTimeline,SW_SHOW);
   TimeLineVisibleFlag=1;
   LoadString(ghinst_main,IDX_MISC_HKEYFRAMER,tt,128);
   ModifyMenu(GetMenu(ghwnd_main),IDM_FILE_KEYFRAMER,
           MF_BYCOMMAND|MF_STRING,IDM_FILE_KEYFRAMER,tt);
   { int iMin,iMax;
     ptSize.x=Nframes*F_W;   ptSize.y=Nnodes*A_H;
     SetScrollRanges(ghwndTimelineChild,0);
     GetScrollRange(ghwndTimelineChild, SB_HORZ, &iMin, &iMax);
     SetScrollPos(ghwndTimelineChild,SB_HORZ,iMin,TRUE);
     GetScrollRange(ghwndTimelineChild, SB_VERT, &iMin, &iMax);
     SetScrollPos(ghwndTimelineChild,SB_VERT,iMin,TRUE);
     InvalidateRect(ghwndTimeline,NULL,TRUE);
     UpdateWindow(ghwndTimeline);
   }
   if(keyframer_docked)IndicatorActive=YES;
 }
 else{
   if(TimeLineVisibleFlag == 0){
     LoadString(ghinst_main,IDX_MISC_HKEYFRAMER,tt,128);
     ModifyMenu(GetMenu(ghwnd_main),IDM_FILE_KEYFRAMER,
             MF_BYCOMMAND|MF_STRING,IDM_FILE_KEYFRAMER,tt);
     ShowWindow(ghwndTimeline,SW_SHOW);
     TimeLineVisibleFlag=1;
     if(TimeLineUpdateFlag)UPDATETIMELINES(IDC_TIMEL_UPDATE)
   }
   else SendMessage(ghwndTimeline,WM_COMMAND,IDM_TIMELINE_HIDE,0);
 }
}

void DestroyTimelineEditor(void){
 if(FirstTime)return;
 DestroyWindow(ghwndTimeline);
 FirstTime=TRUE;
 ScrollingTimeLine=0;
 TimeLineUpdateFlag=0;
 TimeLineVisibleFlag=0;
 keyframer_docked_and_visible=0;
 IndicatorVisible = NO; IndicatorActive=NO; IndicatorsChanged=NO;
 IndicatedActor= -1; IndicatedFrame= -1; IndicatedLine= -1;
 UnregisterClass(szTimelineClass,ghinst_main);
 UnregisterClass(szTimelineChildClass,ghinst_main);
 ghwndTimeline=NULL;
 ghwndTimelineChild=NULL;
 hwndTimelineInfo=NULL;
 return;
}

void AutoUpdateKeyframerVpos(void){
 node *np;
 long c;
 if(!keyframer_docked_and_visible)return;
 if(SelectedNode == NULL)return;
 if(ghwndTimelineChild == NULL)return;
 if(!Preferences.auto_update_keyframer)return;
 c=0;
 if((np=FirstNp) != NULL)while(np != NULL){
   if(SelectedNode == np){
     SetScrollPos(ghwndTimelineChild,SB_VERT,c*A_H,TRUE);
     InvalidateRect(ghwndTimelineChild,NULL,TRUE);
//     UpdateWindow(ghwndTimelineChild);
     return;
   }
   c++;
   np=np->next;
 }
 return;
}
// end script part4

#define EXTEND 0
#define CHOP   0
#define SCALE  1
#define BLANKS 2

BOOL ActorNameExists(HWND parent,node *this_node,char *name){
 node *np;
 np=FirstNp; while(np != NULL){
   if(np != this_node){
     if(strcmpi(np->actorname,name) == 0){
       SendPrgmQuery(IDQ_ACTORNAMEEXISTS,0);
       return TRUE;
     }
   }
   np=np->next;
 }
 return FALSE;
}

void ShiftAllKeys(void){
 int n;
 node     *Np;
 object   *Op,*Op1;
 sky      *Sp,*Sp1;
 director *Dp,*Dp1;
 position *Pp,*Pp1;
 align    *Ap,*Ap1;
 size     *Xp,*Xp1;
 n=DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_SHIFTKEYS),
             ghwnd_main,(DLGPROC)ShiftKeysDlgProc);
 if(n == 0)return;
 if(n < -Nframes || n > Nframes){
   SendPrgmQuery(IDQ_TOOMANYFRAMES,0);
   return;
 }
 Save_UndoA();
 if((Np=FirstNp) != NULL)while(Np != NULL){
   if(n > 0){ /* shifting later   */
     if(Np->type != CAMERA && (Op=Np->fobj) != NULL)while(Op != NULL){
       Op1=Op; Op=Op->next;
       if(Op1->firstframe+n > Nframes)DeleteActor(Np,Op1->firstframe);
     }
     if((Sp=Np->fsky) != NULL)while(Sp != NULL){
       Sp1=Sp; Sp=Sp->next;
       if(Sp1->firstframe+n > Nframes)DeleteSky(Np,Sp1->firstframe);
     }
     if((Dp=Np->fdirector) != NULL)while(Dp != NULL){
       Dp1=Dp; Dp=Dp->next;
       if(Dp1->firstframe+n > Nframes)DeleteDirector(Np,Dp1->firstframe);
     }
     if((Pp=Np->fpos) != NULL)while(Pp != NULL){
       Pp1=Pp; Pp=Pp->next;
       if(Pp1->firstframe+n > Nframes)DeletePosition(Np,Pp1->firstframe);
     }
     if((Ap=Np->fali) != NULL)while(Ap != NULL){
       Ap1=Ap; Ap=Ap->next;
       if(Ap1->firstframe+n > Nframes)DeleteAlign(Np,Ap1->firstframe);
     }
     if((Xp=Np->fsiz) != NULL)while(Xp != NULL){
       Xp1=Xp; Xp=Xp->next;
       if(Xp1->firstframe+n > Nframes)DeleteSize(Np,Xp1->firstframe);
     }
   }
   else{      /* shifting earlier */
     if(Np->type != CAMERA && (Op=Np->fobj) != NULL)while(Op != NULL){
       Op1=Op; Op=Op->next;
       if(Op1->lastframe+n < 1)DeleteActor(Np,Op1->firstframe);
     }
     if((Sp=Np->fsky) != NULL)while(Sp != NULL){
       Sp1=Sp; Sp=Sp->next;
       if(Sp1->lastframe+n < 1)DeleteSky(Np,Sp1->firstframe);
     }
     if((Dp=Np->fdirector) != NULL)while(Dp != NULL){
       Dp1=Dp; Dp=Dp->next;
       if(Dp1->lastframe+n < 1)DeleteDirector(Np,Dp1->firstframe);
     }
     if((Pp=Np->fpos) != NULL)while(Pp != NULL){
       Pp1=Pp; Pp=Pp->next;
       if(Pp1->lastframe+n < 1)DeletePosition(Np,Pp1->firstframe);
     }
     if((Ap=Np->fali) != NULL)while(Ap != NULL){
       Ap1=Ap; Ap=Ap->next;
       if(Ap1->lastframe+n < 1)DeleteAlign(Np,Ap1->firstframe);
     }
     if((Xp=Np->fsiz) != NULL)while(Xp != NULL){
       Xp1=Xp; Xp=Xp->next;
       if(Xp1->lastframe+n < 1)DeleteSize(Np,Xp1->firstframe);
     }
   }
   if(Np->type != CAMERA && (Op=Np->fobj) != NULL)while(Op != NULL){
     Op->firstframe += n; Op->lastframe += n;
     Op->firstframe=max(1,Op->firstframe);
     Op->lastframe=min(Nframes,Op->lastframe);
     Op=Op->next;
   }
   if((Sp=Np->fsky) != NULL)while(Sp != NULL){
     Sp->firstframe += n; Sp->lastframe += n;
     Sp->firstframe=max(1,Sp->firstframe);
     Sp->lastframe=min(Nframes,Sp->lastframe);
     Sp=Sp->next;
   }
   if((Dp=Np->fdirector) != NULL)while(Dp != NULL){
     Dp->firstframe += n; Dp->lastframe += n;
     Dp->firstframe=max(1,Dp->firstframe);
     Dp->lastframe=min(Nframes,Dp->lastframe);
     Dp=Dp->next;
   }
   if((Pp=Np->fpos) != NULL)while(Pp != NULL){
     Pp->firstframe += n; Pp->lastframe += n;
     Pp->firstframe=max(1,Pp->firstframe);
     Pp->lastframe=min(Nframes,Pp->lastframe);
     Pp=Pp->next;
   }
   if((Ap=Np->fali) != NULL)while(Ap != NULL){
     Ap->firstframe += n; Ap->lastframe += n;
     Ap->firstframe=max(1,Ap->firstframe);
     Ap->lastframe=min(Nframes,Ap->lastframe);
     Ap=Ap->next;
   }
   if((Xp=Np->fsiz) != NULL)while(Xp != NULL){
     Xp->firstframe += n; Xp->lastframe += n;
     Xp->firstframe=max(1,Xp->firstframe);
     Xp->lastframe=min(Nframes,Xp->lastframe);
     Xp=Xp->next;
   }
   Np=Np->next;
 }
 EDIT_ACTION=YES;
 UpdateSelectedActor(FALSE);
 UPDATETIMELINES(IDC_TIMEL_UPDATE)
 ReDrawStageDisplay(TRUE);
 if(ghwndOpenGLview == NULL)PerspectiveView(0,1);
 else UpdateGLview(TRUE);
 return;
}

void SetNumberOfFrames(short query){
  char string[10];
  short number,appear,newappear,oldff,oldlf;
  int action;
  double   scale,dappear;
  node     *Np;
  object   *Op,*OpEnd;
  sky      *Sp,*SpEnd;
  director *Dp,*DpEnd;
  position *Pp,*PpEnd;
  align    *Ap,*ApEnd;
  size     *Xp,*XpEnd;
  HWND     hwnd;
  if(query < 0){
    number= -query;
    action=SCALE;
  }
  else if(query == 0){
    if((hwnd=GetActiveWindow()) != ghwndTimeline)hwnd=ghwnd_main;
    number=(short)DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_SETFRAMES),
           hwnd,(DLGPROC)SetFramesDlgProc,(LPARAM)&action);
    if(number < 1)return;
#if __DEMO__
    if(number > 70){
      LoadString(ghinst_main,IDX_MISC_DEMO1,res_str,256);
      MessageBox(NULL,res_str,NULL,MB_OK);
      number=70;
    }
#endif
    if(number < 0)return;
    if(number > 0  && number != Nframes){
      scale = (double)number/(double)Nframes;
    }
  }
  else{
    number=query;
    scale = (double)number/(double)Nframes;
    action=EXTEND;
  }
  EDIT_ACTION=YES;
  if((Np=FirstNp) != NULL)while(Np != NULL){
   OpEnd=NULL;
   if((Op=Np->fobj) != NULL)while(Op != NULL){OpEnd=Op;  Op=Op->next;}
   if(OpEnd != NULL && (action < BLANKS || Np->type == CAMERA)){
     if(action == SCALE){
       Op=Np->fobj;
       while(Op != NULL){
         oldff=Op->firstframe; oldlf=Op->lastframe;
         appear=Op->lastframe - Op->firstframe + 1 ;
         Op->firstframe = (Op->firstframe - 1) * scale + 1;
         newappear = (short)((double)appear * scale);
         Op->lastframe = Op->firstframe + newappear;
         if(Op->last != NULL){
           if(Op->last->lastframe >= Op->firstframe){
             Op->firstframe = Op->last->lastframe + 1;
             if(Op->lastframe < Op->firstframe)
               Op->lastframe = Op->firstframe;
           }
         }
         if(Op->type == PATH && Op->v != NULL){
           Op->v=ReTweenVelocity(Op,Op->v,
               (int)(oldlf-oldff+1),(int)(Op->lastframe-Op->firstframe+1),
               Op->pathlength,0);
         }
         Op=Op->next;
       }
     }
     if(number > OpEnd->lastframe && OpEnd->lastframe >= Nframes){
       oldff=OpEnd->firstframe; oldlf=OpEnd->lastframe;
       OpEnd->lastframe=number;
       if(OpEnd->type == PATH && OpEnd->v != NULL){
         OpEnd->v=ReTweenVelocity(OpEnd,OpEnd->v,
              (int)(oldlf-oldff+1),
              (int)(OpEnd->lastframe-OpEnd->firstframe+1),
              OpEnd->pathlength,0);
       }
     }
   }
   if(OpEnd != NULL)while(OpEnd != NULL){
     Op=OpEnd->last;
     if(OpEnd->firstframe > number)DeleteActor(Np,OpEnd->firstframe);
     else if(OpEnd->lastframe > number){
       oldff=OpEnd->firstframe; oldlf=OpEnd->lastframe;
       OpEnd->lastframe=number;
       if(OpEnd->type == PATH && OpEnd->v != NULL){
         OpEnd->v=ReTweenVelocity(OpEnd,OpEnd->v,
              (int)(oldlf-oldff+1),
              (int)(OpEnd->lastframe-OpEnd->firstframe+1),
              OpEnd->pathlength,0);
         }
     }
     OpEnd=Op;
   }
   SpEnd=NULL;
   if((Sp=Np->fsky) != NULL)while(Sp != NULL){SpEnd=Sp;  Sp=Sp->next;}
   if(SpEnd != NULL && action < BLANKS){
     if(action == SCALE){
       Sp=Np->fsky;
       while(Sp != NULL){
         appear=Sp->lastframe - Sp->firstframe + 1 ;
         Sp->firstframe = (Sp->firstframe - 1) * scale + 1;
         newappear = (short)((double)appear * scale);
         Sp->lastframe = Sp->firstframe + newappear;
         if(Sp->last != NULL){
           if(Sp->last->lastframe >= Sp->firstframe){
             Sp->firstframe = Sp->last->lastframe + 1;
             if(Sp->lastframe < Sp->firstframe)
               Sp->lastframe = Sp->firstframe;
           }
         }
       Sp=Sp->next;}
     }
     if(number > SpEnd->lastframe && SpEnd->lastframe >= Nframes)
       SpEnd->lastframe=number;
   }
   if(SpEnd != NULL)while(SpEnd != NULL){Sp=SpEnd->last;
     if(SpEnd->firstframe > number)DeleteSky(Np,SpEnd->firstframe);
     else if(SpEnd->lastframe > number)SpEnd->lastframe=number;
   SpEnd=Sp;}

   DpEnd=NULL;
   if((Dp=Np->fdirector) != NULL)while(Dp != NULL){DpEnd=Dp;  Dp=Dp->next;}
   if(DpEnd != NULL && action < BLANKS){
     if(action == SCALE){
       Dp=Np->fdirector;
       while(Dp != NULL){
         appear=Dp->lastframe - Dp->firstframe + 1 ;
         Dp->firstframe = (Dp->firstframe - 1) * scale + 1;
         newappear = (short)((double)appear * scale);
         Dp->lastframe = Dp->firstframe + newappear;
         if(Dp->last != NULL){
           if(Dp->last->lastframe >= Dp->firstframe){
             Dp->firstframe = Dp->last->lastframe + 1;
             if(Dp->lastframe < Dp->firstframe)
               Dp->lastframe = Dp->firstframe;
           }
         }
       Dp=Dp->next;}
     }
     if(number > DpEnd->lastframe && DpEnd->lastframe >= Nframes)
       DpEnd->lastframe=number;
   }
   if(DpEnd != NULL)while(DpEnd != NULL){Dp=DpEnd->last;
     if(DpEnd->firstframe > number)DeleteSky(Np,DpEnd->firstframe);
     else if(DpEnd->lastframe > number)DpEnd->lastframe=number;
   DpEnd=Dp;}

   PpEnd=NULL;
   if((Pp=Np->fpos) != NULL)while(Pp != NULL){PpEnd=Pp;  Pp=Pp->next;}
   if(PpEnd != NULL && action < BLANKS){
     if(action == SCALE){
       Pp=Np->fpos;
       while(Pp != NULL){
         appear=Pp->lastframe - Pp->firstframe + 1 ;
         Pp->firstframe = (Pp->firstframe - 1) * scale + 1;
         newappear = (short)((double)appear * scale);
         Pp->lastframe  =  Pp->firstframe + newappear;
         if(Pp->last != NULL){
           if(Pp->last->lastframe >= Pp->firstframe){
             Pp->firstframe = Pp->last->lastframe + 1;
             if(Pp->lastframe < Pp->firstframe)
               Pp->lastframe = Pp->firstframe;
           }
         }
       Pp=Pp->next;}
     }
     if(number > PpEnd->lastframe && PpEnd->lastframe >= Nframes)
       PpEnd->lastframe=number;
   }
   if(PpEnd != NULL)while(PpEnd != NULL){Pp=PpEnd->last;
     if(PpEnd->firstframe > number)DeletePosition(Np,PpEnd->firstframe);
     else if(PpEnd->lastframe > number)PpEnd->lastframe=number;
   PpEnd=Pp;}
   ApEnd=NULL;
   if((Ap=Np->fali) != NULL)while(Ap != NULL){ApEnd=Ap;  Ap=Ap->next;}
   if(ApEnd != NULL && action < BLANKS){
     if(action == SCALE){
       Ap=Np->fali;
       while(Ap != NULL){
         appear=Ap->lastframe - Ap->firstframe + 1 ;
         Ap->firstframe = (Ap->firstframe - 1) * scale + 1;
         newappear = (short)((double)appear * scale);
         Ap->lastframe  =  Ap->firstframe + newappear;
         if(Ap->last != NULL){
           if(Ap->last->lastframe >= Ap->firstframe){
             Ap->firstframe = Ap->last->lastframe + 1;
             if(Ap->lastframe < Ap->firstframe)
               Ap->lastframe = Ap->firstframe;
           }
         }
       Ap=Ap->next;}
     }
     if(number > ApEnd->lastframe && ApEnd->lastframe >= Nframes)
       ApEnd->lastframe=number;
   }
   if(ApEnd != NULL)while(ApEnd != NULL){Ap=ApEnd->last;
     if(ApEnd->firstframe > number)DeleteAlign(Np,ApEnd->firstframe);
     else if(ApEnd->lastframe > number)ApEnd->lastframe=number;
   ApEnd=Ap;}
   XpEnd=NULL;
   if((Xp=Np->fsiz) != NULL)while(Xp != NULL){XpEnd=Xp;  Xp=Xp->next;}
   if(XpEnd != NULL && action < BLANKS){
     if(action == SCALE){
       Xp=Np->fsiz;
       while(Xp != NULL){
         appear=Xp->lastframe - Xp->firstframe + 1 ;
         Xp->firstframe = (Xp->firstframe - 1) * scale + 1;
         newappear = (short)((double)appear * scale);
         Xp->lastframe  =  Xp->firstframe + newappear;
         if(Xp->last != NULL){
           if(Xp->last->lastframe >= Xp->firstframe){
             Xp->firstframe = Xp->last->lastframe + 1;
             if(Xp->lastframe < Xp->firstframe)
               Xp->lastframe = Xp->firstframe;
           }
         }
       Xp=Xp->next;}
     }
     if(number > XpEnd->lastframe && XpEnd->lastframe >= Nframes)
       XpEnd->lastframe=number;
   }
   if(XpEnd != NULL)while(XpEnd != NULL){Xp=XpEnd->last;
     if(XpEnd->firstframe > number)DeleteSize(Np,XpEnd->firstframe);
     else if(XpEnd->lastframe > number)XpEnd->lastframe=number;
   XpEnd=Xp;}

  Np=Np->next;}
  Nframes=number;
  UpdateSelectedActor(FALSE);
  EnableMenuItem(GetMenu(ghwnd_main),IDM_FRAME_RENDERNOW,MF_GRAYED);
}

static int  nActionsList=0;
static char *ActionsList[256],*ActionsText[256];

short StartDefault(short what_function){
 int nPresets=7;
 short action,nfr,nmodels=0;
 node     *Np,*Np1,*lastNode,*pathNode;
 object   *Op,*Op1;
 sky      *Sp;
 position *Pp;
 align    *Ap,*Ap1;
 size     *Xp;
 char response[16];
 nPresets=GetPresets(nPresets);
 LoadString(ghinst_main,IDX_MISC_TEMPLATE,res_str,256);
 action=SelectScrolledItemList(nPresets,ActionsText,res_str,
                               TRUE,ghwnd_main);
 if(action < 0)goto FINISHED;
 if(action > 6){
   if(!BuildFromPreset(action))goto ERRORS;
   goto FINISHED;
 }
 if(Nskys == 0){
   if(CreateNode() != NULL){
     LoadString(ghinst_main,IDX_MISC_NAMESKY,res_str,256);
     strcpy(MainNp->actorname,res_str);
     MainNp->type=SKY;
     MainNp->fsky=NULL;
     Nskys=1;
   }
   else goto FINISHED;
 }
 if((Sp=CreateSky(MainNp,1,(short)Nframes)) == NULL || fail_op)goto ERRORS;
 Sp->type=SKYGRADED;
 Sp->colour[0]=0; Sp->colour[1]=255; Sp->colour[2]=255;
 Sp->zcolour[0]=0; Sp->zcolour[1]=0; Sp->zcolour[2]=255;
 if(CreatePosition(FirstNp,1,(short)Nframes) == NULL || fail_op)goto ERRORS;
 FirstNp->fpos->finish[0]=TVpointX+TVsizeX*2;
 FirstNp->fpos->finish[1]=TVpointY+TVsizeY*2;
 FirstNp->fpos->finish[2]=TVpointZ+TVsizeZ*2;
 FirstNp->fpos->type=SPLINE;
 if((Ap=CreateAlign(FirstNp,1,(short)Nframes)) == NULL || fail_op)goto ERRORS;
 if(CreateNode() == NULL)goto ERRORS;
 LoadString(ghinst_main,IDX_MISC_NAMETARGET,res_str,256);
 strcpy(MainNp->actorname,res_str);
 MainNp->type=TARGET;
 if((Op=CreateActor(MainNp,1,(short)Nframes)) == NULL || fail_op)goto ERRORS;
 Op->type=MainNp->type;
 FirstNp->fali->type=TRACK; FirstNp->fali->topath=MainNp;
// strange bug !!!  if((Pp=CreatePosition(MainNp,1,Nframes)) == NULL || fail_op)goto ERRORS;
// anything will do
 if(action < 6){
   char ttt[64];
   LoadString(ghinst_main,IDX_MISC_SELECTMODEL,res_str,256);
   LoadString(ghinst_main,IDX_MISC_SELECT1,ttt,256);
   if((Np1=LoadStandinModel(&nmodels,res_str,ttt)) == NULL)
   goto ERRORS;
 }
 else if(action == 6){
   char ttt[64];
   LoadString(ghinst_main,IDX_MISC_MODELHEAD,res_str,256);
   LoadString(ghinst_main,IDX_MISC_MODELLOAD,ttt,256);
   if((Np1=LoadStandinModel(&nmodels,res_str,ttt))
          == NULL)goto ERRORS;
 }
 if(Np1->fobj != NULL){
   Op1=Np1->fobj;
   FirstNp->fpos->finish[0] += Op1->outline[2][0];
   FirstNp->fpos->finish[1] += Op1->outline[2][1];
   FirstNp->fpos->finish[2] += Op1->outline[2][2];
 }
 if(CreateNode() == NULL)goto ERRORS;
 LoadString(ghinst_main,IDX_MISC_NAMELIGHT,res_str,256);
 strcpy(MainNp->actorname,res_str);
 MainNp->type=LIGHT;
 if((Op=CreateActor(MainNp,1,Nframes)) == NULL)goto ERRORS;
 Op->type=MainNp->type;
 Op->lighttype=SPHERE; Op->colour[0]=Op->colour[1]=Op->colour[2]=255;
 if((Pp=CreatePosition(MainNp,1,(short)Nframes)) == NULL || fail_op)goto ERRORS;
 Pp->finish[0]=FirstNp->fpos->finish[0]+TVsizeX/2;
 Pp->finish[1]=FirstNp->fpos->finish[1]+TVsizeY;
 Pp->finish[2]=FirstNp->fpos->finish[2]+TVsizeZ*2;
 Pp->type=SPLINE;
 if(action < 1)goto FINISHED;
 if(action == 1){
   if(CreateNode() == NULL)goto ERRORS;
   LoadString(ghinst_main,IDX_MISC_NAMEGROUND,res_str,256);
   strcpy(MainNp->actorname,res_str);
   MainNp->type=GROUND; Nground=1;
   if((Op=CreateActor(MainNp,1,(short)Nframes)) == NULL)goto ERRORS;
   Op->type=MainNp->type;
   Op->groundtype=PLAIN;
   SetSfxColour(Op->colour,IDX_MISC_GROUND_COLOUR,ghwnd_main);
   if((Pp=CreatePosition(MainNp,1,Nframes)) == NULL || fail_op)goto ERRORS;
   Pp->finish[0]=0;
   Pp->finish[1]=0;
   Pp->finish[2]=TVpointZ;
   Pp->type=SPLINE;
   if((Xp=CreateSize(MainNp,1,Nframes)) == NULL || fail_op)goto ERRORS;
 }
 if(action > 1){
   char ttt[64];
   LoadString(ghinst_main,IDX_MISC_NUMFRAMES,res_str,256);
   LoadString(ghinst_main,IDX_MISC_FRAMES,ttt,64);
   nfr=RequestNumEntry(60,1,999,res_str,ttt);
#if __DEMO__
   if(nfr > 70){
     LoadString(ghinst_main,IDX_MISC_DEMO1,res_str,256);
     MessageBox(NULL,res_str,NULL,MB_OK);
     nfr=70;
   }
#endif
   if(nfr < 0)goto FINISHED;
   SetNumberOfFrames(nfr);
   if((Ap1=CreateAlign(Np1,1,Nframes)) == NULL || fail_op)goto ERRORS;
   if(action == 2){Ap1->im=1; Ap1->ima=1.0;}
 }
 if(action >= 3 && action <= 5){
   if(CreateNode() == NULL)goto ERRORS;
   LoadString(ghinst_main,IDX_MISC_NAMEPATH,res_str,256);
   strcpy(MainNp->actorname,res_str);
   MainNp->type=PATH;
   if((Op=CreateActor(MainNp,1,Nframes)) == NULL)goto ERRORS;
   Op->type=MainNp->type;
   strcpy(Op->name,"PATH1995");
   if(CreatePathObject((short)(action-3),MainNp,Op,1,Nframes) < 0)goto ERRORS;
   if((Pp=CreatePosition(Np1,1,Nframes)) == NULL || fail_op)goto ERRORS;
   Pp->type=FOLLOW; Pp->onpath=MainNp;
   Ap1->type=TOPATH; Ap1->topath=MainNp;
 }
 else if(action == 6){ /* train so load the other models */
   if(CreateNode() == NULL)goto ERRORS;
   LoadString(ghinst_main,IDX_MISC_NAMEPATH,res_str,256);
   strcpy(MainNp->actorname,res_str);
   pathNode=MainNp;
   MainNp->type=PATH;
   if((Op=CreateActor(MainNp,1,Nframes)) == NULL)goto ERRORS;
   Op->type=MainNp->type;
   strcpy(Op->name,"PATH1995");
   if(CreatePathObject(2,MainNp,Op,1,Nframes) < 0)goto ERRORS;
   if((Pp=CreatePosition(Np1,1,Nframes)) == NULL || fail_op)goto ERRORS;
   Pp->type=FOLLOW; Pp->onpath=pathNode;
   Ap1->type=TOPATH; Ap1->topath=pathNode;
   lastNode=Np1;
   while(1){
     char ttt[64];
     LoadString(ghinst_main,IDX_MISC_NEXTCANCEL,res_str,256);
     LoadString(ghinst_main,IDX_MISC_ADDNEXT,ttt,64);
     if((Np=LoadStandinModel(&nmodels,res_str,ttt))
           == NULL)goto ERRORS;
     if(Np->fobj == NULL){
       DeleteNode(Np);
       break;
     }
     if((Pp=CreatePosition(Np,1,Nframes)) == NULL || fail_op)goto ERRORS;
     Pp->type=FOLLOWON; Pp->onpath=lastNode;
     if((Ap=CreateAlign(Np,1,Nframes)) == NULL || fail_op)goto ERRORS;
     Ap->type=TOPATH; Ap->topath=pathNode;
     lastNode=Np;
   }
 }
 goto FINISHED;
 ERRORS:
 SendPrgmQuery(IDQ_OLD4,0);
 FINISHED:
 FreePresets();
 return 0;
}

static int GetPresets(int nPresets){
 int i;
 HANDLE h;
 FILE *ff;
 WIN32_FIND_DATA ffd,*lpffd;
 char str[255],title[64],tempstr[256];
 int PresetTypeListID[]={
   IDX_MISC_PRESET1,
   IDX_MISC_PRESET2,
   IDX_MISC_PRESET3,
   IDX_MISC_PRESET4,
   IDX_MISC_PRESET5,
   IDX_MISC_PRESET6,
   IDX_MISC_PRESET7
 };
 sprintf(str,"%s\\*.pfx",gszPSTdir);
 lpffd=&ffd;
 nActionsList=0;
 for(i=0;i<nPresets;i++){
   ActionsList[nActionsList]=(char *)X__Malloc(8*sizeof(char));
   if(ActionsList[nActionsList] != NULL)strcpy(ActionsList[nActionsList],"I");
   ActionsText[nActionsList]=(char *)X__Malloc(256*sizeof(char));
   if(ActionsText[nActionsList] != NULL){
     LoadString(ghinst_main,PresetTypeListID[nActionsList],tempstr,256);
     strcpy(ActionsText[nActionsList],tempstr);
   }
   nActionsList++;
 }
 if((h=FindFirstFile(str,lpffd)) != INVALID_HANDLE_VALUE){
   sprintf(str,"%s\\%s",gszPSTdir,lpffd->cFileName);
   ActionsList[nActionsList]=(char *)X__Malloc(256*sizeof(char));
   if(ActionsList[nActionsList] != NULL)strcpy(ActionsList[nActionsList],str);
   nActionsList++;
   while(FindNextFile(h,lpffd)){
     sprintf(str,"%s\\%s",gszPSTdir,lpffd->cFileName);
     ActionsList[nActionsList]=(char *)X__Malloc(256*sizeof(char));
     strcpy(ActionsList[nActionsList],str);
     nActionsList++;
   }
   if(GetLastError() != ERROR_NO_MORE_FILES){
     LoadString(ghinst_main,IDX_MISC_ERRORINNEXT,res_str,256);
     MessageBox(NULL,res_str,NULL,MB_OK);
   }
   if(!FindClose(h)){
     LoadString(ghinst_main,IDX_MISC_FILEERROR,res_str,256);
     MessageBox(NULL,res_str,NULL,MB_OK);
   }
   if(nActionsList > nPresets)for(i=nPresets;i<nActionsList;i++){
     if((ff=fopen(ActionsList[i],"r")) != NULL){
       fscanf(ff,"%[^\n]",title);
       ActionsText[i]=(char *)X__Malloc(256*sizeof(char));
       if(ActionsText[i] != NULL)
       strcpy(ActionsText[i],title);
       fclose(ff);
     }
   }
 }
 return nActionsList;
}

static void FreePresets(void){
 int i;
 if(nActionsList > 0){
   for(i=0;i<nActionsList;i++){
     if(ActionsList[i] != NULL)X__Free(ActionsList[i]);
     if(ActionsText[i] != NULL)X__Free(ActionsText[i]);
   }
   nActionsList=0;
 }
}

static BOOL BuildFromPreset(int action){
 BOOL status=FALSE;
 HMODULE hLib;
 FARPROC fpFun;
 ANI_STRUCTURE *Evs;
 BOOL (*fpFun1)(HWND, ANI_STRUCTURE *);
 int l;
 char fname[255];
 strcpy(fname,ActionsList[action]);
 l=strlen(fname);
 fname[l-1]='l'; fname[l-2]='l'; fname[l-3]='d';
 Evs=&AniStructure;
 if((hLib=LoadLibrary(fname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_ExternalPreset")) != NULL){
     fpFun1 = (void *)fpFun;
     status=(*fpFun1)(ghwnd_main,Evs);
   }
   else MessageBox(ghwnd_main,(LPCTSTR) "Failed function",(LPCTSTR)" ",
        MB_OK | MB_TASKMODAL );
   FreeLibrary(hLib);
 }
 else MessageBox(ghwnd_main,(LPCTSTR)fname,(LPCTSTR)"Failed to find DLL",
                MB_OK | MB_TASKMODAL );
 return status;
}
