/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file WROBOT.C    for Windows animator                        */
/* assume that of each robot actor the skeletons are consistent */

#define MODULE_ROBOT 1

#define TIME_DELAY 30

#include <io.h>
#include <fcntl.h>
#include "animate.h"

typedef double quaternion[4];

static void rotate_round_vector(double angle, vector v, double t[4][4]);
static void TransformAxis(skel *sid, vector v, vector vv);
static void SelectRobotArm(void);
static void NameRobotDirectory(char *script_file, char *robo_dir);
static void EditRobotTransform(int command);
static void RobotSkeletonTreeView(HWND hWndParent);
static void SelectTreeViewItem(void);
static void GetBasisTransform(vector x, vector y, vector z,
                              vector u, vector v, vector w,
                              double T[4][4]);

static node   *LastSelectedRobotNode=NULL;
static object *SelectedRobot=NULL;
static long    SelectedRobotArm = -1;

long GetRobotVolume(long n, skel *sp, vector c){
 long i,size=UNIT;
 static double xmax=UNIT2,ymax=UNIT2,zmax=UNIT2,
               xmin=0,ymin=0,zmin=0;
 if(sp != NULL){
   xmin=ymin=zmin =  MAXUNIT;
   xmax=ymax=zmax = -MAXUNIT;
   for(i=0;i<n;i++){
     if(sp->p[0]  > xmax)xmax=sp->p[0];
     if(sp->pp[0] > xmax)xmax=sp->pp[0];
     if(sp->p[0]  < xmin)xmin=sp->p[0];
     if(sp->pp[0] < xmin)xmin=sp->pp[0];
     if(sp->p[1]  > ymax)ymax=sp->p[1];
     if(sp->pp[1] > ymax)ymax=sp->pp[1];
     if(sp->p[1]  < ymin)ymin=sp->p[1];
     if(sp->pp[1] < ymin)ymin=sp->pp[1];
     if(sp->p[2]  > zmax)zmax=sp->p[2];
     if(sp->pp[2] > zmax)zmax=sp->pp[2];
     if(sp->p[2]  < zmin)zmin=sp->p[2];
     if(sp->pp[2] < zmin)zmin=sp->pp[2];
     sp++;
   }
 }
 size=max(xmax-xmin,ymax-ymin);
 size=max(size,zmax-zmin);
 c[0]=(double)(xmax+xmin)*0.5;
 c[1]=(double)(ymax+ymin)*0.5;
 c[2]=(double)(zmax+zmin)*0.5;
 return size;
}

void DrawRobot(HDC hdc[], object *Op, point Offset, double p, double t
             , double a, short im, double ima
             , double sx, double sy, double sz, short status){
 int sh1,sv1,sh2,sv2,i,j,k,ep;
 point  v1,v2;
 skel *sp;
 double mr,RoboT[4][4],x,y,z,xx,yy,zz;
 int ifd,ild;
 if(View == TRIVIEW){ifd=0;ild=3;}
 else               {ifd=ActiveView; ild=ifd+1;}
 if(status == DESELECTED)DrawActorCentre(hdc,Offset,p,t,a,im,ima,TVsizeX/24);
 if(status == SELECTED)DrawArrow(hdc,Offset,p,t,a,im,ima,TVsizeX/24);
 mr=(double)(Op->lastframe - Op->firstframe+1);   /* mr = 1 in key frame */
 mr=(double)(CurrentFrame - Op->firstframe+1)/mr;
 if(Op->nskeleton > 0){
   InterpolateRobot(mr,Op,Op->last);
   RobotTransform(Offset,p,t,a,sx,sy,sz,im,ima,RoboT);
   sp=Op->skeleton;
   for(k=0;k<Op->nskeleton;k++){ /* transform all the points */
     xx=(double)sp[k].pp[0];
     yy=(double)sp[k].pp[1];
     zz=(double)sp[k].pp[2];
     m4by1(RoboT,xx,yy,zz,&x,&y,&z);
     sp[k].pt[0]=(long)x;   /* keep transformed pt for selection */
     sp[k].pt[1]=(long)y;
     sp[k].pt[2]=(long)z;
   }
   for(k=0;k<Op->nskeleton;k++){ /* draw the points */
     CopyPoint(sp[k].pt,v1);
     if(k > 0 && (j=sp[k].id) < Op->nskeleton)CopyPoint(sp[j].pt,v2);
     else                                     CopyPoint(v1,v2);
     if(in_stage_triview(v1) || in_stage_triview(v2)){
       for(i=ifd;i<ild;i++){
         GetWindowCoords(i,v1[0],v1[1],v1[2],&sh1,&sv1);
         GetWindowCoords(i,v2[0],v2[1],v2[2],&sh2,&sv2);
         MoveToEx(hdc[i],sh1,sv1,NULL); LineTo(hdc[i],sh2,sv2);
         if(status == SELECTED){
           HPEN holdpen;
           if(k == SelectedRobotArm)holdpen=SelectObject(hdc[i],ghEditPen);
           Rectangle(hdc[i],(int)sh1-3,(int)sv1-3,(int)sh1+3,(int)sv1+3);
           if(k == SelectedRobotArm)SelectObject(hdc[i],holdpen);
         }
       }
     }
   }
 }
}

void DrawFullRobot(HDC hdc[], object *Op, point Offset, double p, double t
               , double a, short im, double ima
               , double sx, double sy, double sz, short status){
 short morphflag;
 int sh1,sv1,sh2,sv2,i,j,k,ep;
 point xyz,*Vp;
 skel *sp;
 double mr,RoboT[4][4],ViewT[4][4],x,y,z,xx,yy,zz;
 int ifd,ild;
 long *sid,Nv;
 morphflag=0;
 if(View == TRIVIEW){ifd=0;ild=3;}
 else               {ifd=ActiveView; ild=ifd+1;}
 if(status == DESELECTED)DrawActorCentre(hdc,Offset,p,t,a,im,ima,TVsizeX/24);
 if(status == SELECTED)DrawArrow(hdc,Offset,p,t,a,im,ima,TVsizeX/24);
 mr=(double)(Op->lastframe - Op->firstframe+1);   /* mr = 1 in key frame */
 mr=(double)(CurrentFrame - Op->firstframe+1)/mr;
 if(Op->nskeleton > 0 && (sp=Op->skeleton) != NULL && (sid=Op->skid) != NULL){
   InterpolateRobot(mr,Op,Op->last);
   RobotTransform(Offset,p,t,a,sx,sy,sz,im,ima,RoboT);
   for(k=0;k<Op->nskeleton;k++){ /* transform all the points */
     xx=(double)sp[k].pp[0];
     yy=(double)sp[k].pp[1];
     zz=(double)sp[k].pp[2];
     m4by1(RoboT,xx,yy,zz,&x,&y,&z);
     sp[k].pt[0]=(long)x;   /* keep transformed pt for selection */
     sp[k].pt[1]=(long)y;
     sp[k].pt[2]=(long)z;
   }
   if(Op->in_ram){
     Nv=Op->npoints;
     if(Nv > 0){
       if((Vp=(point *)X__Malloc(Nv*sizeof(point))) != NULL){
         for(i=0;i<Nv;i++){
           CopyPoint(Op->points[i],xyz);
           m4by4(RoboT,sp[sid[i]].T,ViewT);
           x = (double)(xyz[0]-Op->origin[0]);
           y = (double)(xyz[1]-Op->origin[1]);
           z = (double)(xyz[2]-Op->origin[2]);
           m4by1(ViewT,x,y,z,&xx,&yy,&zz);
           Vp[i][0]=(long)xx; Vp[i][1]=(long)yy; Vp[i][2]=(long)zz;
         }
       }
       else{
         SendPrgmQuery(IDQ_NOMEM3,0);
         Nv=0;
       }
       if(Nv > 0)for(k=0;k<Nv;k++){
         if(in_stage_triview(Vp[k])){
           for(i=ifd;i<ild;i++){
             GetWindowCoords(i,Vp[k][0],Vp[k][1],Vp[k][2],&sh1,&sv1);
             MoveToEx(hdc[i],sh1,sv1,NULL); LineTo(hdc[i],sh1+2,sv1);
           }
         }
       }
       if(Vp != NULL){ X__Free(Vp); Vp=NULL;}
     }
   }
 }
 DrawRobot(hdc,Op,Offset,p,t
          ,a,im,ima
          ,sx,sy,sz,status);
}

void SkeletorOn(void){
 object *Op;
 EDIT_ACTION=YES;
 if(SelectedNode == NULL){
   SendPrgmQuery(IDQ_NRSELECT,0);
   return;
 }
 if(SelectedNode->type != ROBOT){
   SendPrgmQuery(IDQ_NOTAROBOT,0);
   return;
 }
 if(SelectedNode->fobj == NULL){
   SendPrgmQuery(IDQ_NOROBOTA,0);
   return;
 }
 Op=SelectedNode->fobj;
 while(Op != NULL){
  if(CurrentFrame == Op->lastframe){
    SelectedRobot=Op;
    tool=SKELETOR;
//    if(!IsWindow(ghwndDlgTools)){
//      ghwndDlgTools=CreateDialog(ghinst_main,MAKEINTRESOURCE(DLG_ROBOT),
//                                  ghwnd_main,(DLGPROC)RobotToolDlgProc);
//      if(IsWindow(ghwndDlgTools))ShowWindow(ghwndDlgTools,SW_SHOW);
//    }
    if(!IsWindow(ghwndSkEdit))RobotSkeletonTreeView(ghwnd_main);
    ReDrawStageDisplay(TRUE);
    return;
  }
  Op=Op->next;
 }
 SendPrgmQuery(IDQ_NOREDIT,0);
 return;
}

void SkeletorOff(short redraw){
 tool=NOTOOL;
 if(IsWindow(ghwndSkEdit))RobotSkeletonTreeView(ghwnd_main);
// if(IsWindow(ghwndDlgTools)){
//   DestroyWindow(ghwndDlgTools);
//   ghwndDlgTools=NULL;
// }
 SelectedRobot=NULL;
 SelectedRobotArm = -1;
 if(redraw)ReDrawStageDisplay(TRUE);
}

static void SelectRobotArm(void){
 BOOL flag=FALSE;
 long i;
 skel *sp;
 double d,dmin,dth;
 int sho,svo,H,V;
 POINT p;
 if(SelectedRobot == NULL)return;
 DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,SELECTED);
// now keep selection   SelectedRobotArm=-1;
 sp=SelectedRobot->skeleton;
 dmin=(double)TVsizeX*(double)TVsizeX;
 dth=dmin/32/32;
 GetCursorPos(&p); ScreenToClient(ghwnd_current,&p);
 H=(int)p.x; V=(int)p.y;
 if(SelectedRobot->nskeleton > 0)for(i=0;i<SelectedRobot->nskeleton;i++){
   GetWindowCoords(ActiveView,sp->pt[0],sp->pt[1],sp->pt[2],&sho,&svo);
   if( (sho > H-4) && (sho < H+4) && (svo > V-4) && (svo < V+4)){
     d=(double)(sp->pt[0]-NpointerX)*(double)(sp->pt[0]-NpointerX)
      +(double)(sp->pt[1]-NpointerY)*(double)(sp->pt[1]-NpointerY)
      +(double)(sp->pt[2]-NpointerZ)*(double)(sp->pt[2]-NpointerZ);
     if(d < dmin){
       dmin=d;
       SelectedRobotArm=i;
       flag=TRUE;
     }
   }
   sp++;
 }
 DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,SELECTED);
 if(View == ONEVIEW)UpdateWindow(ghwnd_current);
 else for(i=0;i<3;i++)UpdateWindow(ghwnd_triview[i]);
 if(flag && IsWindow(ghwndRobotView))PostMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
}

static int A_normalize(vector v){
 double n,nn;
 n= (double)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return FAIL;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE(n,v,v)
 return OK;
}

static void TransformAxis(skel *sid, vector v, vector vv){
 vector t;
 VECSCALE((double)UNIT2*1000.0,v,t)
 VECSUM(t,(double)sid->p,t)
 m4by1(sid->T,t[0],t[1],t[2],&vv[0],&vv[1],&vv[2]);
 VECSUB(vv,(double)sid->pp,vv)
 A_normalize(vv);
}

static void GetBasisTransform(vector x, vector y, vector z,
                              vector u, vector v, vector w,
                              double T[4][4]){
 int i,j;                     /* return transformation matrix from the   */
 double a[3][3],t[4][4];      /* transformed basis to the original basis */
 double det;
 null_transform(T);
 null_transform(t);
 a[0][0]=x[0]; a[0][1]=x[1]; a[0][2]=x[2];
 a[1][0]=y[0]; a[1][1]=y[1]; a[1][2]=y[2];
 a[2][0]=z[0]; a[2][1]=z[1]; a[2][2]=z[2];
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
 m4by1(t,u[0],v[0],w[0],&T[0][0],&T[0][1],&T[0][2]);
 m4by1(t,u[1],v[1],w[1],&T[1][0],&T[1][1],&T[1][2]);
 m4by1(t,u[2],v[2],w[2],&T[2][0],&T[2][1],&T[2][2]);
}

static void InvertRotationMatrix(double T[4][4], double t[4][4]){
 int i,j;
 double a[3][3];
 double det;
 null_transform(t);
 for(i=0;i<3;i++)for(j=0;j<3;j++)a[i][j]=T[i][j];
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
}

static void rotate_round_vector(double angle, vector v, double t[4][4]){
  double dx,dy,dz,phi,theta,dxy;
  double t1[4][4],t2[4][4],t3[4][4];
  angle *= PI/180.0;
  dx=v[0]; dy=v[1]; dz=v[2];
  dxy=dx*dx+dy*dy;
  if(dxy < 0.00001){ /* vertical so just rotate about z  and return */
    if(dz > 0.0)rotz(t, angle);
    else        rotz(t,-angle);
    return;
  }
  dxy=sqrt(dxy);
  if     (dx == 0.0 && dy > 0.0)phi =  PI2;
  else if(dx == 0.0 && dy < 0.0)phi = -PI2;
  else phi = atan2(dy,dx);
  theta = atan2(dz,dxy);
  rotz(t3, -phi);
  roty(t2, -theta);
  m4by4(t2,t3,t1);
  rotx(t2, angle);
  m4by4(t2,t1,t3);
  roty(t2,theta);
  m4by4(t2,t3,t1);
  rotz(t2,phi);
  m4by4(t2,t1,t);
}

int Skeletor(void){
 POINT pt;
 int xi,xo,MickeyX,I;
 double ltr[4][4],ctr[4][4];
 vector Tv,Rv,Xv={1.0,0.0,0.0},Yv={0.0,1.0,0.0},Zv={0.0,0.0,1.0};
 long i,j,k,evTime;
 skel *s;
 object *op;
 if(SelectedNode == NULL)return 1;
 if(SelectedNode->type != ROBOT)return 1;
 op=GetObjectAtFrame(SelectedNode,CurrentFrame);
 UpdateRobot(op);
// if(op != NULL)SelectedRobot=op;  // now in UpdateRobot;
 if(SelectedRobot != NULL && SelectedRobot->lastframe != CurrentFrame){
   SendPrgmQuery(IDQ_ONLYKEYFRAME,0);
   return 0;
 }
 SelectRobotArm();
 if(SelectedRobot != NULL && SelectedRobotArm >= 0){
   SelectTreeViewItem();
   if(CurrentFrame == SelectedRobot->lastframe){
     s=SelectedRobot->skeleton;
     i=SelectedRobotArm;
     if(gToolFlags.flagK == 6){
       if(i > 0){
         VECSUB((double)(s+((s+i)->id))->p,(double)(s+i)->p,Tv)
         A_normalize(Tv);
         m4by1((s+i)->R,Tv[0],Tv[1],Tv[2],&Rv[0],&Rv[1],&Rv[2]);
       }
       else{
         VECCOPY(Zv,Rv)
       }
     }
     else if(gToolFlags.flagK > 2 && i > 0){
       if(     gToolFlags.flagK == 3)VECCOPY(Xv,Tv)
       else if(gToolFlags.flagK == 4)VECCOPY(Yv,Tv)
       else                          VECCOPY(Zv,Tv)
       j=s[i].id;
       TransformAxis((s+j),s[j].u,s[j].uu);
       TransformAxis((s+j),s[j].v,s[j].vv);
       TransformAxis((s+j),s[j].w,s[j].ww);
       GetBasisTransform(s[j].uu,s[j].vv,s[j].ww,s[j].u,s[j].v,s[j].w,ltr);
       m4by1(ltr,Tv[0],Tv[1],Tv[2],&Rv[0],&Rv[1],&Rv[2]);
     }
     GetCursorPos(&pt); xo=(int)pt.x;
     evTime=GetTickCount()+TIME_DELAY;
     do {
       if(evTime > GetTickCount())continue;
       GetCursorPos(&pt); xi=pt.x;
       MickeyX = xi-xo; xo=xi;
       if(MickeyX != 0){
         if(MickeyX > 0)I=max( 1,min( 45,MickeyX));
         else           I=min(-1,max(-45,MickeyX));
         DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,SELECTED);
         if(i > 0){
           j=s[i].id;
           if(gToolFlags.flagK == 0) /* rotate about v (along axis) */
             rotate_round_vector(I,s[j].v,ltr);
           else if(gToolFlags.flagK == 1)
             rotate_round_vector(I,s[j].u,ltr);
           else if(gToolFlags.flagK == 2)
             rotate_round_vector(I,s[j].w,ltr);
           else if(gToolFlags.flagK > 2 && gToolFlags.flagK <= 6)
             rotate_round_vector(I,Rv,ltr);
           else null_transform(ltr);
           m4by4(ltr,s[i].R,ctr); c4to4(ctr,s[i].R);
         }
         DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,SELECTED);
         if(ghwndOpenGLview == NULL)PerspectiveView(0,0);
         if(IsWindow(ghwndRobotView)){
           SendMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
           UpdateWindow(ghwndRobotView);
         }
         if(View == ONEVIEW)UpdateWindow(ghwnd_current);
         else{
           for(I=0;I<3;I++)UpdateWindow(ghwnd_triview[I]);
         }
         UpdateWindow(ghwnd_view);
       }
       evTime=GetTickCount()+TIME_DELAY;
     }
     while (GetAsyncKeyState(VK_LBUTTON) & 0x8000);

     if(ghwndOpenGLview != NULL)UpdateGLview(TRUE);
     else PerspectiveView(0,1);

     return 0;
   }
   else{ /* not key frame */
     ;
   }
 }
 return 1;
}

void RobotTransform(point Offset, double p, double t, double a,
                    double sx, double sy, double sz,
                    short im, double ima, double RoboT[4][4]){
 double tr0[4][4],tr1[4][4],tr2[4][4],tr3[4][4],x,y,z;
 scal(tr1,sx,sy,sz);
 if(im > 0){
   if     (im == 1)rotz(tr0,ima*PIo180);
   else if(im == 2)rotx(tr0,ima*PIo180);
   else if(im == 3)roty(tr0,ima*PIo180);
   roty(tr2,t*PI/180.0);
   m4by4(tr2,tr0,tr3);
 }
 else{
  roty(tr3,t*PIo180);
 }
 m4by4(tr3,tr1,tr2);
 rotx(tr3,a*PIo180);
 m4by4(tr3,tr2,tr1);
 rotz(tr2,p*PIo180);
 m4by4(tr2,tr1,tr3);
 tram(tr1,Offset[0],Offset[1],Offset[2]);
 m4by4(tr1,tr3,RoboT);
 return;
}

static void NameRobotDirectory(char *script_file, char *robo_dir){
  long l;
  l=strlen(script_file);
  strcpy(robo_dir,script_file);
  strcpy((robo_dir+l-4),".skd");
}

short OpenRobotDirectory(char * script_file){
  SECURITY_ATTRIBUTES sa;
  char str[255];
  sa.nLength=sizeof(sa);
  sa.lpSecurityDescriptor=NULL;
  sa.bInheritHandle=FALSE;
  NameRobotDirectory(script_file,str);
  if(CreateDirectory(str,&sa))return 1;
  /* try to remove an old  one before giving up */
  RemoveRobotDirectory(script_file);
  if(CreateDirectory(str,&sa))return 1;
  else               return 0;
}

void RemoveRobotDirectory(char *script_file){
 HANDLE h;
 WIN32_FIND_DATA ffd,*lpffd;
 char str[255],strr[255],title[64];
 NameRobotDirectory(script_file,str);
 sprintf(strr,"%s\\$*.*",str);
 lpffd=&ffd;
 if((h=FindFirstFile(strr,lpffd)) == INVALID_HANDLE_VALUE){
   if(!RemoveDirectory(str)){
     LoadString(ghinst_main,IDX_MISC_ERROR1,title,64);
     MessageBox(NULL,str,title,MB_OK);
   }
   return;
 }
 sprintf(strr,"%s\\%s",str,lpffd->cFileName);
 if(!DeleteFile(strr))MessageBeep(MB_OK);
 while(FindNextFile(h,lpffd)){
   sprintf(strr,"%s\\%s",str,lpffd->cFileName);
   if(!DeleteFile(strr))MessageBeep(MB_OK);
 }
 if(GetLastError() != ERROR_NO_MORE_FILES){
   SendPrgmQuery(IDQ_BADNEXTFILE,3);
 }
 if(!FindClose(h))SendPrgmQuery(IDQ_BADFILE,3);
 if(!RemoveDirectory(str))MessageBeep(MB_OK);
}

void GetRobotScriptName(char *script_file, long node_id,
                        short lastframe, char *robot_name){
  char str[255];
  NameRobotDirectory(script_file,str);
  sprintf(robot_name,"%s\\$%d.%ld",str,lastframe,node_id);
}

short RobotFileExists(char *robot_name){
  int fd;
  if((fd=open(robot_name,O_RDONLY)) < 0)return 0;
  close(fd);
  return 1;
}

static void EditRobotTransform(int command){
 object *Op;
 skel   *sp;
 long j,id;
 if((Op=SelectedRobot) == NULL)return;
 if(command == 0){ /* reset robot */
   sp=Op->skeleton;
   for(j=0;j<Op->nskeleton;j++){
     null_transform(sp->R);
     null_transform(sp->Q);
     null_transform(sp->T);
     sp++;
   }
 }
 else if(command == 2){ /* reset selected */
   if(SelectedRobotArm >= 0 && SelectedRobotArm < Op->nskeleton){
     sp=(Op->skeleton+SelectedRobotArm);
     null_transform(sp->R);
     null_transform(sp->Q);
     null_transform(sp->T);
     sp=Op->skeleton;
     if((GetAsyncKeyState(VK_SHIFT) & 0x8000) == 0x8000){
       for(j=0;j<Op->nskeleton;j++){
         id=sp->id;
         while(id != 0){
           if(id == SelectedRobotArm){
             null_transform(sp->R);
             null_transform(sp->Q);
             null_transform(sp->T);
             break;
           }
           id=(Op->skeleton+id)->id;
         }
         sp++;
       }
     }
   }
 }
 else if(command == 1){/* copy from last */
   if(Op->last != NULL && Op->last->nskeleton == Op->nskeleton){
     j=(long)(Op->last->nskeleton)*(long)sizeof(skel);
     memcpy(Op->skeleton,Op->last->skeleton,j);
   }
   else MessageBeep(MB_OK);
 }
 ReDrawStageDisplay(TRUE);
 PerspectiveView(0,0);
 return;
}

void WriteRobot(object *Op, char *robot_name){
  long i,j,n;
  FILE *frobot;
  skel *sk;
  if((frobot = fopen(robot_name,"w")) != NULL){
    if((n=Op->nskeleton) > 0 && (sk=Op->skeleton) != NULL){
      fprintf(frobot,"%ld\n",n);
      for(i=0;i<n;i++){
        for(j=0;j<4;j++)fprintf(frobot,"%lf %lf %lf %lf\n",
         sk[i].R[j][0],sk[i].R[j][1],sk[i].R[j][2],sk[i].R[j][3]);
      }
    }
    else{
      fprintf(frobot,"0\n");
    }
    fclose(frobot);
  }
}

static int nSkeleton;
skel   *Skeleton;

static void ApplyTransformToChildren(int id, double T[4][4]){
 static double x,y,z;
 static double sst[4][4],t1[4][4],t2[4][4];
 int i;
 for(i=1;i<nSkeleton;i++){ /* must start at one since root is never a child */
   if(Skeleton[i].id == id){
     tram(t1,-(Skeleton+id)->p[0],-(Skeleton+id)->p[1],-(Skeleton+id)->p[2]);
     m4by4((Skeleton+i)->Q,t1,t2);
     tram(t1, (Skeleton+id)->p[0], (Skeleton+id)->p[1], (Skeleton+id)->p[2]);
     m4by4(t1,t2,sst);
     m4by4(T,sst,Skeleton[i].T);
     m4by1(Skeleton[i].T,  /* transform point as well and store for drawing */
           (double)Skeleton[i].p[0],
           (double)Skeleton[i].p[1],
           (double)Skeleton[i].p[2],&x,&y,&z);
           Skeleton[i].pp[0]=(long)x;
           Skeleton[i].pp[1]=(long)y;
           Skeleton[i].pp[2]=(long)z;
     ApplyTransformToChildren(i,Skeleton[i].T);
   }
 }
 return;
}

#define EPSILON 1.0e-6

static void MatrixToQuaternion(double m[4][4], quaternion q){
 double w4,x2;
 q[0]=0.25*(1.0+m[0][0]+m[1][1]+m[2][2]);
 if(q[0] > EPSILON){
   q[0]=sqrt(q[0]);
   w4=1.0/(4.0*q[0]);
   q[1]=(m[1][2]-m[2][1])*w4;
   q[2]=(m[2][0]-m[0][2])*w4;
   q[3]=(m[0][1]-m[1][0])*w4;
 }
 else{
   q[0]=0.0;
   q[1]= -0.5*(m[1][1]+m[2][2]);
   if(q[1] > EPSILON){
     q[1]=sqrt(q[1]);
     x2=1.0/(2.0*q[1]);
     q[2]=m[0][1]*x2;
     q[3]=m[0][2]*x2;
   }
   else{
     q[1]=0.0;
     q[2]=0.5*(1.0-m[2][2]);
     if(q[2] > EPSILON){
       q[2]=sqrt(q[2]);
       q[3]=m[1][2]/(2.0*q[2]);
     }
     else{
       q[2]=0.0;
       q[3]=1.0;
     }
   }
 }
 return;
}

static void QuaternionToMatrix(quaternion q, double m[4][4]){
 double w,x,y,z,x2,y2,z2;
 w=q[0]; x=q[1]; y=q[2]; z=q[3]; x2=x*x; y2=y*y; z2=z*z;
 m[0][0]=1.0-2*y2-2*z2; m[0][1]=2*x*y+2*w*z;   m[0][2]=2*x*z-2*w*y;
 m[1][0]=2*x*y-2*w*z;   m[1][1]=1.0-2*x2-2*z2; m[1][2]=2*y*z+2*w*x;
 m[2][0]=2*x*z+2*w*y;   m[2][1]=2*y*z-2*w*x;   m[2][2]=1.0-2*x2-2*y2;
 m[3][3]=1.0; m[0][3]=m[1][3]=m[2][3]=m[3][0]=m[3][1]=m[3][2]=0.0;
 return;
}

static void Slerp(double mr, quaternion q1, quaternion q2, quaternion q){
 double sum,beta1,beta2,theta;
 sum=q1[0]*q2[0]+q1[1]*q2[1]+q1[2]*q2[2]+q1[3]*q2[3];
 if(sum > 1.0)sum=1.0;
 theta=acos(sum);
 if(theta < EPSILON){ beta1=1.0-mr; beta2=mr;}
 else{
  beta1=sin((1.0-mr)*theta)/sin(theta);
  beta2=sin(mr*theta)/sin(theta);
 }
 q[0]=beta1*q1[0]+beta2*q2[0];
 q[1]=beta1*q1[1]+beta2*q2[1];
 q[2]=beta1*q1[2]+beta2*q2[2];
 q[3]=beta1*q1[3]+beta2*q2[3];
}

void InterpolateRobot(double mr, object *Op, object *LastOp){
 int i;
 quaternion q,q1,q2;
 nSkeleton=Op->nskeleton;
 Skeleton=Op->skeleton;
 if(mr < 1.0 && LastOp != NULL && Op->nskeleton == LastOp->nskeleton){
   /* interpopate rotation matrices using Quaternion tweening */
#if 0
   for(i=0;i<nSkeleton;i++){ /* this interpolation always from original - bad? */
     MatrixToQuaternion((LastOp->skeleton+i)->R,q1);
     MatrixToQuaternion((Skeleton+i)->R,q2);
     Slerp(mr,q1,q2,q);
     QuaternionToMatrix(q,(Skeleton+i)->Q);
   }
#else
   /* this interpolation always from last frame rotation as origin */
   quaternion Qi={1.0,0.0,0.0,0.0};
   double Q[4][4],Ri[4][4];
   for(i=0;i<nSkeleton;i++){
     InvertRotationMatrix((LastOp->skeleton+i)->R,Ri);
     m4by4((Skeleton+i)->R,Ri,Q);
     MatrixToQuaternion(Q,q2);
     Slerp(mr,Qi,q2,q);
     QuaternionToMatrix(q,Q);
//     m4by4((LastOp->skeleton+i)->R,Q,(Skeleton+i)->Q); fault !!!
     m4by4(Q,(LastOp->skeleton+i)->R,(Skeleton+i)->Q);
   }
#endif
 }
 else{
   for(i=0;i<nSkeleton;i++)c4to4((Skeleton+i)->R,(Skeleton+i)->Q);
 }
 ApplyTransformToChildren(0,Skeleton[0].T);
 return;
}

// robot treeview dialog

#define ptrNMHDR       ((LPNMHDR)lparam)
#define ptrNM_TREEVIEW ((NM_TREEVIEW *)lparam)
#define ptrTV_DISPINFO ((TV_DISPINFO *)lparam)

#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))

#define IDM_SKTB_V            100
#define IDM_SKTB_U            101
#define IDM_SKTB_W            102
#define IDM_SKTB_X            103
#define IDM_SKTB_Y            104
#define IDM_SKTB_Z            105
#define IDM_SKTB_R            106
#define IDM_SKTB_RESET        107
#define IDM_SKTB_COPY         108
#define IDM_SKTB_SAVE         109
#define IDM_SKTB_LOAD         110
#define IDM_SKTB_ZOOM         111
#define IDM_SKTB_ZOOM_STATE   112
#define IDM_SKTB_RESET_SELECT 113
#define IDM_SKTB_OVERLAY      114
#define IDM_SKTB_TRANSPARENT  115

static LRESULT CALLBACK RobotOglViewWndProc(HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK RobotHdcViewWndProc(HWND, UINT, WPARAM, LPARAM);

static PSTR        lpszSkTreeViewClass="SkTreeViewClass";
static PSTR        lpszRobotViewClass="RobotViewClass";
static HWND        hWndTreeView=NULL;
static HWND        hTBWnd=NULL;
static HIMAGELIST  hCoasterImageList=NULL;
static HDC         hdcRobotView=NULL;
static HBITMAP     hbmRobotView=NULL,hbmOldRobotView=NULL;
static int         RobotViewXpos=-1,RobotViewYpos=-1,
                   RobotViewXsize=0,RobotViewYsize=0;
                   SkTreeViewXpos=-1,SkTreeViewYpos=-1,
                   SkTreeViewXsize=0,SkTreeViewYsize=0;
static BOOL        RobotZoomFlag=FALSE,RobotZoomState=FALSE,
                   RobotOverlayFlag=FALSE,RobotGlassFlag=FALSE;

#define NTOOLBARBUTTONS 21 // 18

static TBBUTTON tbb[NTOOLBARBUTTONS] = {
          {9,IDM_SKTB_LOAD,    TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {10,IDM_SKTB_SAVE,   TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {0,0,                TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          {7,IDM_SKTB_RESET,   TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {13,IDM_SKTB_RESET_SELECT, TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {8,IDM_SKTB_COPY,    TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {0,0,                TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          {11,IDM_SKTB_ZOOM,   TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                                TBSTYLE_CHECK,0,0 },
          {12,IDM_SKTB_ZOOM_STATE, TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                                TBSTYLE_CHECK,0,0 },

          {0,0,                TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          {14,IDM_SKTB_OVERLAY,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                                TBSTYLE_CHECK,0,0 },
          {15,IDM_SKTB_TRANSPARENT,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                                TBSTYLE_CHECK,0,0 },

          {0,0,                TBSTATE_ENABLED|TBSTATE_WRAP,TBSTYLE_SEP,0,0 },
          {0,0,                TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          {0,IDM_SKTB_V,TBSTATE_ENABLED|TBSTATE_CHECKED,TBSTYLE_BUTTON|
                                               TBSTYLE_CHECKGROUP, 0,0 },
          {1,IDM_SKTB_U,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                               TBSTYLE_CHECKGROUP, 0,0 },
          {2,IDM_SKTB_W,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                               TBSTYLE_CHECKGROUP, 0,0 },
          {6,IDM_SKTB_R,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                               TBSTYLE_CHECKGROUP, 0,0 },
          {5,IDM_SKTB_Y,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                               TBSTYLE_CHECKGROUP, 0,0 },
          {4,IDM_SKTB_X,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                               TBSTYLE_CHECKGROUP, 0,0 },
          {3,IDM_SKTB_Z,TBSTATE_ENABLED,TBSTYLE_BUTTON|
                                               TBSTYLE_CHECKGROUP, 0,0 }
                         };

static HTREEITEM AddTreeViewItem (HWND        hWndTV,
                                  HTREEITEM   hParent,
                                  HTREEITEM   hInsertAfter,
                                  int         iImage,
                                  int         iSelectedImage,
                                  LPSTR       szText,
                                 LPARAM      lParam){
 TV_INSERTSTRUCT       tvIns;
 tvIns.item.mask           = TVIF_TEXT  | TVIF_IMAGE |
                         TVIF_SELECTEDIMAGE | TVIF_PARAM;
 tvIns.item.pszText        = szText;
 tvIns.item.cchTextMax     = sizeof(szText);
 tvIns.item.iImage         = iImage;
 tvIns.item.iSelectedImage = iSelectedImage;
 tvIns.item.lParam         = lParam;
 tvIns.hParent             = hParent;
 tvIns.hInsertAfter        = hInsertAfter;
 return TreeView_InsertItem(hWndTV,&tvIns);
}

static void SelectTreeViewItem(void){
 skel *sp;
 int n;
 if(SelectedRobot == NULL)return;
 if((sp=SelectedRobot->skeleton) == NULL)return;
 n=SelectedRobot->nskeleton;
 if(SelectedRobotArm < 0 || SelectedRobotArm >= n)return;
 TreeView_EnsureVisible(hWndTreeView,(HTREEITEM)(sp+SelectedRobotArm)->lParam);
 TreeView_Select(hWndTreeView,(HTREEITEM)(sp+SelectedRobotArm)->lParam,TVGN_CARET);
}

static void AddTwigsTo(long id, HWND hWndTV){
 int i;
 skel *sp;
 HTREEITEM hParent;
 sp=(SelectedRobot->skeleton+1);
 for(i=1;i<SelectedRobot->nskeleton;i++){
   if(sp->id == id){
     hParent=(HTREEITEM)((SelectedRobot->skeleton+id)->lParam);
     sp->lParam=(LPARAM)AddTreeViewItem(hWndTV,hParent,(HTREEITEM)TVI_FIRST,
                                      0,1,
                                      (LPSTR)sp->name,
                                      (LPARAM)i);
     AddTwigsTo(i,hWndTV);
   }
   sp++;
 }
}

static void UpdateSkTreeView(void){
 int i;
 skel *sp;
 object *Op;
 if(!IsWindow(hWndTreeView))return;
 TreeView_DeleteItem(hWndTreeView,TVI_ROOT);
 SelectedRobot->skeleton->lParam=(LPARAM)AddTreeViewItem(hWndTreeView,
                                     TVI_ROOT,
                                     (HTREEITEM)TVI_FIRST,
                                     0,1,
                                     (LPSTR)SelectedRobot->skeleton->name,
                                     (LPARAM)0);
 AddTwigsTo(0,hWndTreeView);
 for(i=1;i<SelectedRobot->nskeleton;i++){
  sp=(SelectedRobot->skeleton+i);
  if(sp->id == 0){
    TreeView_EnsureVisible(hWndTreeView,(HTREEITEM)sp->lParam);
    TreeView_Expand(hWndTreeView,(HTREEITEM)sp->lParam,TVE_EXPAND);
   }
 }
 /* now copy all the TV Handles into all the channels skel LPARAMS's in case
    we change to different pose channel - this assumes that all poses have
    the same skeleton structure and that the TV is NOT changed. */
 if(SelectedNode != NULL){
   Op=SelectedNode->fobj;
   while(Op != NULL){
     if(Op != SelectedRobot){
       for(i=0;i<SelectedRobot->nskeleton;i++){
         if(i < Op->nskeleton)
           (Op->skeleton+i)->lParam=(SelectedRobot->skeleton+i)->lParam;
       }
     }
     Op=Op->next;
   }
 }
}

static BOOL TreeviewMenuCommand(HWND hWnd, WORD id){
 FILE *pz;
 long k;
 switch (id) {
   case IDM_SKTB_V:
     gToolFlags.flagK=0;
     break;
   case IDM_SKTB_U:
     gToolFlags.flagK=1;
     break;
   case IDM_SKTB_W:
     gToolFlags.flagK=2;
     break;
   case IDM_SKTB_X:
     gToolFlags.flagK=3;
     break;
   case IDM_SKTB_Y:
     gToolFlags.flagK=4;
     break;
   case IDM_SKTB_Z:
     gToolFlags.flagK=5;
     break;
   case IDM_SKTB_R:
     gToolFlags.flagK=6;
     break;
   case IDM_SKTB_RESET:
     EditRobotTransform(0);
     break;
   case IDM_SKTB_RESET_SELECT:
     EditRobotTransform(2);
     break;
   case IDM_SKTB_COPY:
     EditRobotTransform(1);
     break;
   case IDM_SKTB_ZOOM:
     RobotZoomFlag ^= 1;
     break;
   case IDM_SKTB_ZOOM_STATE:
     RobotZoomState ^= 1;
     break;
   case IDM_SKTB_SAVE:
     if(SelectSfxFileName(1,gszPZEfile,gszPZEdir,
        IDX_MISC_SAVEPOSE,"(*.PZE)|*.pze|",ghwnd_main) == TRUE){
       AppendFileExtension(gszPZEfile,".pze");
       if((pz=fopen(gszPZEfile,"wb")) != NULL){
         k=0xFEFE;
         fwrite(&k,sizeof(long),1,pz);
         k=0;
         fwrite(&k,sizeof(long),1,pz); /* use this for future */
         fwrite(SelectedRobot->name,255,1,pz);
         fwrite(&(SelectedRobot->nskeleton),sizeof(long),1,pz);
         if(SelectedRobot->nskeleton > 0)
           for(k=0;k<SelectedRobot->nskeleton;k++){
           fwrite(SelectedRobot->skeleton[k].R,sizeof(double),16,pz);
         }
         fclose(pz);
       }
       else MessageBeep(MB_OK);
     }
     break;
   case IDM_SKTB_LOAD:
     if(SelectSfxFileName(0,gszPZEfile,gszPZEdir,
        IDX_MISC_LOADPOSE,"(*.PZE)|*.pze|",ghwnd_main) == TRUE){
       if((pz=fopen(gszPZEfile,"rb")) != NULL){
         fread(&k,sizeof(long),1,pz);
         if(k == 0xFEFE){
           char robo_name[256];
           fread(&k,sizeof(long),1,pz); /* use this for future */
           fread(robo_name,255,1,pz);
           fread(&k,sizeof(long),1,pz);
           if(k == SelectedRobot->nskeleton){
             if(SelectedRobot->nskeleton > 0){
               for(k=0;k<SelectedRobot->nskeleton;k++){
                 fread(SelectedRobot->skeleton[k].R,sizeof(double),16,pz);
               }
               ReDrawStageDisplay(TRUE);
               PerspectiveView(0,0);
             }
           }
           else SendPrgmQuery(IDQ_BADPOSE,0);
         }
         else SendPrgmQuery(IDQ_NOTPOSE,0);
         fclose(pz);
       }
       else MessageBeep(MB_OK);
     }
     break;
   case IDM_SKTB_OVERLAY:
     RobotOverlayFlag ^= 1;
     break;
   case IDM_SKTB_TRANSPARENT:
     RobotGlassFlag ^= 1;
     break;
   default:
     break;
 }
 if(IsWindow(ghwndRobotView))PostMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
 return TRUE;
}

static LRESULT CALLBACK SkTreeViewWndProc(HWND hwnd, UINT msg, WPARAM wparam,
                                    LPARAM lparam ){
 static int ToolbarHeight;
 TV_ITEM  tvi;
 RECT rc,rc1;
 int dx,dy,y;
 DWORD dwStyle;
 skel *sp;
 switch( msg ){
   case WM_CREATE:
     dwStyle=WS_CHILD | WS_VISIBLE | WS_BORDER;
     if(Preferences.tooltips)dwStyle = dwStyle | TBSTYLE_TOOLTIPS;
     hTBWnd = CreateToolbarEx(hwnd,dwStyle,
        0,      // control identifier for the toolbar
        16,     // bitmaps images
        ghinst_main,
        IDBM_SKTOOLBAR,
        tbb,
        NTOOLBARBUTTONS,
        16,15,  // size of buttons
        16,15,  // size of bitmap
        sizeof(TBBUTTON)
       );
     SendMessage(hTBWnd,TB_AUTOSIZE,0,0);
     gToolFlags.flagK=0;
     if(0){dx=32; dy=32;} else {dx=16; dy=16;}
     hCoasterImageList=ImageList_Create(
                         dx,dy,
                         TRUE,
                         2,
                         5
                         );
     ImageList_AddIcon(hCoasterImageList,
                              LoadIcon(ghinst_main,"TREEVIEWICON0"));
     ImageList_AddIcon(hCoasterImageList,
                              LoadIcon(ghinst_main,"TREEVIEWICON1"));
     GetClientRect(hTBWnd,&rc1); ToolbarHeight=rc1.bottom+1;
     GetClientRect(hwnd,&rc);
     hWndTreeView = CreateWindowEx(0,WC_TREEVIEW,
                                 " ",
                                 WS_VISIBLE | WS_CHILD | WS_BORDER |
                                 TVS_HASLINES | TVS_HASBUTTONS |
                                 TVS_EDITLABELS | TVS_LINESATROOT,
                                 0,ToolbarHeight,
                                 rc.right, rc.bottom,
                                 hwnd,
                                 (HMENU)0,
                                 ghinst_main,
                                 NULL
                               );
     if(hWndTreeView != NULL){
       TreeView_SetImageList(hWndTreeView,hCoasterImageList,0);
       ImageList_SetBkColor(hCoasterImageList,GetSysColor(COLOR_WINDOW));
       UpdateSkTreeView();
     }
     LastSelectedRobotNode=SelectedNode;
     UpdateRobot(SelectedRobot);
     break;
   case WM_DESTROY:{
       char str[32];
       GetWindowRect(hwnd,&rc);
       sprintf(str,"%ld",rc.left);
       WritePrivateProfileString(IniSection,"RSPOSX",str,IniFilename);
       sprintf(str,"%ld",rc.top);
       WritePrivateProfileString(IniSection,"RSPOSY",str,IniFilename);
       sprintf(str,"%ld",rc.right-rc.left);
       WritePrivateProfileString(IniSection,"RSSIZX",str,IniFilename);
       sprintf(str,"%ld",rc.bottom-rc.top);
       WritePrivateProfileString(IniSection,"RSSIZY",str,IniFilename);
       SkTreeViewXpos=rc.left;           SkTreeViewYpos=rc.top;
       SkTreeViewXsize=rc.right-rc.left; SkTreeViewYsize=rc.bottom-rc.top;
       if(hCoasterImageList != NULL)ImageList_Destroy(hCoasterImageList);
       hCoasterImageList=NULL;
       DestroyWindow(hWndTreeView);
       hWndTreeView=NULL;
       DestroyWindow(hTBWnd);
       hTBWnd=NULL;
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
//         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
         PostQuitMessage(0);   // only for case when THREADED
         break;
       default:
         return(DefWindowProc(hwnd,msg,wparam,lparam));
      }
      break;
   case WM_NOTIFY:
     switch (ptrNMHDR->code){
       case TTN_NEEDTEXT:{
           int idButton;
           LPTOOLTIPTEXT lpttt;
           lpttt=(LPTOOLTIPTEXT)lparam;
           lpttt->hinst=ghinst_main;
           idButton=lpttt->hdr.idFrom;
           switch(idButton){
             case IDM_SKTB_V:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_1); break;
             case IDM_SKTB_U:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_2); break;
             case IDM_SKTB_W:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_3); break;
             case IDM_SKTB_X:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_4); break;
             case IDM_SKTB_Y:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_5); break;
             case IDM_SKTB_Z:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_6); break;
             case IDM_SKTB_R:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_7); break;
             case IDM_SKTB_RESET:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_8); break;
             case IDM_SKTB_COPY:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_9); break;
             case IDM_SKTB_LOAD:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_0); break;
             case IDM_SKTB_SAVE:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_A); break;
             case IDM_SKTB_ZOOM:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_B); break;
             case IDM_SKTB_ZOOM_STATE:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_C); break;
             case IDM_SKTB_RESET_SELECT:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_D); break;
              case IDM_SKTB_OVERLAY:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_E); break;
              case IDM_SKTB_TRANSPARENT:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_K_F); break;
           }
         }
         break;
       case TVN_SELCHANGED:
         if(ptrNM_TREEVIEW->action == TVC_BYKEYBOARD ||
            ptrNM_TREEVIEW->action == TVC_BYMOUSE){
           tvi.mask  = TVIF_PARAM;
           tvi.hItem = ptrNM_TREEVIEW->itemNew.hItem;
           TreeView_GetItem(hWndTreeView,&tvi);
           DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,SELECTED);
           SelectedRobotArm=(long)tvi.lParam;
           DrawInvertNode(ghdc_triview_Bitmap,SelectedNode,SELECTED);
           if(IsWindow(ghwndRobotView))PostMessage(ghwndRobotView,WM_DRAWROBOT,0,0);
         }
         break;
       case TVN_BEGINLABELEDIT:
         return 1;
         break;
       default: break;
     }
     return (DefWindowProc(hwnd,msg,wparam,lparam));
   case WM_COMMAND:
     return TreeviewMenuCommand(hwnd,LOWORD(wparam));
     break;
   case WM_SIZE:
     if(hTBWnd != NULL){
       SendMessage(hTBWnd,msg,wparam,lparam);
     }
     if(hWndTreeView != NULL){
       SetWindowPos(hWndTreeView,NULL,
                    0,ToolbarHeight,
                    LOWORD(lparam),HIWORD(lparam)-ToolbarHeight,
                    SWP_NOZORDER);
     }
     break;
   default:
     return(DefWindowProc(hwnd,msg,wparam,lparam));
 }
 return( 0L );
}

static void ThreadSkeletonTreeView(void *);
static void RobotSkeletonTreeView(HWND hWndParent){
 void (*fp)(void *);
 if(ghwndSkEdit != NULL){
   PostMessage(ghwndSkEdit,WM_SYSCOMMAND,SC_CLOSE,0);
   return;
 }
 fp=ThreadSkeletonTreeView;
#if  __WATCOMC__ && __WATCOMC__ < 1100
  _beginthread(fp,0,65536,(void *)hWndParent);
#else
  _beginthread(fp,0,(void *)hWndParent);
#endif
}

static void ThreadSkeletonTreeView(void *arg){
 WNDCLASS   wc;
 RECT       rc;
 int xp,yp,xl,yl;
 HWND hWndParent;
 hWndParent=(HWND)arg;
#if 0
static void RobotSkeletonTreeView(HWND hWndParent){
 WNDCLASS   wc;
 RECT       rc;
 int xp,yp,xl,yl;
 if(ghwndSkEdit != NULL){
   DestroyWindow(ghwndSkEdit);
   UnregisterClass(lpszSkTreeViewClass,ghinst_main);
   ghwndSkEdit=NULL;
   if(ghwndRobotView != NULL){
     DestroyWindow(ghwndRobotView);
     UnregisterClass(lpszRobotViewClass,ghinst_main);
     ghwndRobotView=NULL;
   }
   return;
 }
#endif
 RobotZoomFlag=FALSE;
 RobotZoomState=FALSE;
 RobotOverlayFlag=FALSE;
 RobotGlassFlag=FALSE;
 wc.lpszClassName = lpszSkTreeViewClass;
 wc.style = CS_HREDRAW | CS_VREDRAW;
 wc.lpfnWndProc = (WNDPROC)SkTreeViewWndProc;
 wc.cbClsExtra = 0;
 wc.cbWndExtra = 0;
 wc.hInstance = ghinst_main;
 wc.hIcon = LoadIcon(ghinst_main,"ANIMATORICON");
 wc.hCursor = LoadCursor(NULL,IDC_ARROW);
 wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
 wc.lpszMenuName = NULL;
 if(!RegisterClass(&wc))return;
 GetWindowRect(ghwndTools,&rc);
 if(SkTreeViewXpos < 0){
   xp=GetPrivateProfileInt(IniSection,"RSPOSX",-1,IniFilename);
   yp=GetPrivateProfileInt(IniSection,"RSPOSY",-1,IniFilename);
   xl=GetPrivateProfileInt(IniSection,"RSSIZX",-1,IniFilename);
   yl=GetPrivateProfileInt(IniSection,"RSSIZY",-1,IniFilename);
   if(xp < 0 || yp < 0 || xl < 0 || yl < 0){
     SkTreeViewXsize=185;
     SkTreeViewYsize=255;
     if(bToolpannelLeft)SkTreeViewXpos=rc.right+5;
     else               SkTreeViewXpos=rc.left-SkTreeViewXsize-5;
     SkTreeViewYpos=rc.top;
   }
   else{
     SkTreeViewXsize=xl;
     SkTreeViewYsize=yl;
     SkTreeViewXpos=xp;
     SkTreeViewYpos=yp;
   }
 }
 LoadString(ghinst_main,IDX_MISC_ROBOTHIERARCHY,res_str,256);
 ghwndSkEdit = CreateWindow(
     wc.lpszClassName,res_str,
     WS_POPUP | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME,
     SkTreeViewXpos,SkTreeViewYpos,
     SkTreeViewXsize,SkTreeViewYsize,
     hWndParent,NULL,ghinst_main,NULL);
 ShowWindow(ghwndSkEdit,SW_SHOW);
 UpdateWindow(ghwndSkEdit);
 if(ghwndSkEdit == NULL){
   UnregisterClass(lpszSkTreeViewClass,ghinst_main);
   return;
 }
 wc.lpszClassName = lpszRobotViewClass;
 wc.style = CS_HREDRAW | CS_VREDRAW;
 if(Preferences.use_gl_robot)wc.lpfnWndProc = (WNDPROC)RobotOglViewWndProc;
 else                        wc.lpfnWndProc = (WNDPROC)RobotHdcViewWndProc;
 wc.cbClsExtra = 0;
 wc.cbWndExtra = 0;
 wc.hInstance = ghinst_main;
 wc.hIcon = LoadIcon(ghinst_main,"ANIMATORICON");
 wc.hCursor = ghcurPan; //LoadCursor(NULL,IDC_ARROW);
 wc.hbrBackground = NULL;
 wc.lpszMenuName = NULL;
 if(!RegisterClass(&wc))return;
 LoadString(ghinst_main,IDX_MISC_ROBOTVIEW,res_str,256);
 if(RobotViewXpos < 0){
   xp=GetPrivateProfileInt(IniSection,"RVPOSX",-1,IniFilename);
   yp=GetPrivateProfileInt(IniSection,"RVPOSY",-1,IniFilename);
   xl=GetPrivateProfileInt(IniSection,"RVSIZX",-1,IniFilename);
   yl=GetPrivateProfileInt(IniSection,"RVSIZY",-1,IniFilename);
   if(xp < 0 || yp < 0 || xl < 0 || yl < 0){
     RobotViewXpos=150;  RobotViewYpos=100;
     RobotViewXsize=300; RobotViewYsize=300;
   }
   else{
     RobotViewXpos=xp;  RobotViewYpos=yp;
     RobotViewXsize=xl; RobotViewYsize=yl;
   }
 }
 ghwndRobotView = CreateWindow(
     wc.lpszClassName,res_str,
     WS_POPUP | WS_CAPTION | !WS_SYSMENU | WS_THICKFRAME |
     WS_VSCROLL | WS_HSCROLL |
     WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
     RobotViewXpos,RobotViewYpos,
     RobotViewXsize,RobotViewYsize,
     hWndParent,NULL,ghinst_main,NULL);
 ShowWindow(ghwndRobotView,SW_SHOW);
 UpdateWindow(ghwndRobotView);
 if(ghwndRobotView == NULL){
   UnregisterClass(lpszRobotViewClass,ghinst_main);
   return;
 }
//
 {
   MSG msg;
   while(GetMessage(&msg,NULL,0,0)){
     TranslateMessage(&msg);
     DispatchMessage(&msg);
   }
   DestroyWindow(ghwndSkEdit);
   UnregisterClass(lpszSkTreeViewClass,ghinst_main);
   ghwndSkEdit=NULL;
   if(ghwndRobotView != NULL){
     DestroyWindow(ghwndRobotView);
     UnregisterClass(lpszRobotViewClass,ghinst_main);
     ghwndRobotView=NULL;
   }
   PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
   _endthread();
 }
//
 return;
}

void UpdateRobot(object *Op){
 if(ghwndSkEdit == NULL)return;
 if(SelectedNode == NULL || SelectedNode->type != ROBOT){
   PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
   return;
 }
 if(SelectedNode != LastSelectedRobotNode){
// Just discard it for now
   PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP_TOOL,0);
   return;
 }
 if(Op != SelectedRobot){
   vector c;
   SelectedRobot=Op;
   if(Op == NULL)SelectedRobotArm= -1;
   else GetRobotVolume(Op->nskeleton,Op->skeleton,c);
 }
 return;
}

static LRESULT CALLBACK RobotHdcViewWndProc(HWND hwnd, UINT msg,
                                            WPARAM wparam,LPARAM lparam ){
 static BOOL flag=FALSE,mmoved=FALSE,bTrackPan=FALSE;
 static long xs,ys;
 static double up_angle=0.0,round_angle=0.0;
 char str[32];
 double old;
 HDC  hdc;
 RECT rc;
 int i,iPos,iMax,iMin,dn;
 switch( msg ){
   case WM_CREATE:{
       vector c;
       GetRobotVolume(SelectedRobot->nskeleton,SelectedRobot->skeleton,c);
       hdc=GetDC(hwnd);
       hdcRobotView=CreateCompatibleDC(hdc);
       ReleaseDC(hwnd,hdc);
       SetScrollRange(hwnd, SB_VERT, -89,89,TRUE);
       SetScrollRange(hwnd, SB_HORZ, -179,179,TRUE);
       SetScrollPos(hwnd,SB_HORZ,(int)round_angle,TRUE);
       SetScrollPos(hwnd,SB_VERT,(int)up_angle,TRUE);
     }
     break;
   case WM_DESTROY:
     if(hbmOldRobotView != NULL)SelectObject(hdcRobotView,hbmOldRobotView);
     if(hdcRobotView != NULL)DeleteDC(hdcRobotView);
     hdcRobotView=NULL;
     if(hbmRobotView != NULL)DeleteObject(hbmRobotView);
     hbmRobotView=NULL; hbmOldRobotView=NULL;
     GetWindowRect(hwnd,&rc);
     RobotViewXpos=rc.left; RobotViewYpos=rc.top;
     RobotViewXsize=rc.right-rc.left; RobotViewYsize=rc.bottom-rc.top;
     sprintf(str,"%ld",rc.left);
     WritePrivateProfileString(IniSection,"RVPOSX",str,IniFilename);
     sprintf(str,"%ld",rc.top);
     WritePrivateProfileString(IniSection,"RVPOSY",str,IniFilename);
     sprintf(str,"%ld",rc.right-rc.left);
     WritePrivateProfileString(IniSection,"RVSIZX",str,IniFilename);
     sprintf(str,"%ld",rc.bottom-rc.top);
     WritePrivateProfileString(IniSection,"RVSIZY",str,IniFilename);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         break;
       default:
         return(DefWindowProc(hwnd,msg,wparam,lparam));
      }
      break;
   case WM_SIZE:
     if(hbmOldRobotView != NULL)SelectObject(hdcRobotView,hbmOldRobotView);
     if(hbmRobotView != NULL)DeleteObject(hbmRobotView);
     hbmRobotView=NULL;
//     SelectPalette(hdcRobotView,ghpaletteScreen,FALSE);
//     RealizePalette(hdcRobotView);
     SelectObject(hdcRobotView,ghRobotBrush);
     GetClientRect(hwnd,&rc);
     hdc=GetDC(hwnd);
     hbmRobotView=CreateCompatibleBitmap(hdc,rc.right,rc.bottom);
     ReleaseDC(hwnd,hdc);
     hbmOldRobotView=SelectObject(hdcRobotView,hbmRobotView);
     PatBlt(hdcRobotView,0,0,rc.right,rc.bottom,PATCOPY);
     SendMessage(hwnd,WM_DRAWROBOT,0,0);
     break;
   case WM_LBUTTONDOWN:
     mmoved=FALSE;
     xs=LOWORD(lparam);
     ys=HIWORD(lparam);
     SetCapture(hwnd);
     bTrackPan=TRUE;
     break;
   case WM_MOUSEMOVE:
     if(bTrackPan){
       long dx,dy;
       dx=(short)LOWORD(lparam)-xs; xs=(short)LOWORD(lparam);
       dy=(short)HIWORD(lparam)-ys; ys=(short)HIWORD(lparam);
       if(labs(dx) > 0 || labs(dy) > 0){
         mmoved=TRUE;
         round_angle=round_angle+(double)dx;
         while(round_angle < -180.0)round_angle += 360.0;
         while(round_angle >  180.0)round_angle -= 360.0;
         SetScrollPos(hwnd,SB_HORZ,(int)round_angle,TRUE);
         up_angle=up_angle+(double)dy;
         if(up_angle > -90 && up_angle < 90)
           SetScrollPos(hwnd,SB_VERT,(int)up_angle,TRUE);
         SendMessage(hwnd,WM_DRAWROBOT,0,0);
         UpdateWindow(ghwndRobotView);
//         InvalidateRect(hwnd, NULL, FALSE);
       }
     }
     break;
   case WM_LBUTTONUP:
     if(bTrackPan){
       ReleaseCapture();
       bTrackPan=FALSE;
     }
     if(mmoved){
       mmoved=FALSE;
       break;
     }
     mmoved=FALSE;
   case WM_DRAWROBOT:
     if(SelectedRobot != NULL && (i=SelectedRobotArm) > 0){
       skel *s;
       s=SelectedRobot->skeleton;
       if((i=s[i].id) >= 0){
         if(gToolFlags.flagK == 0)TransformAxis((s+i),s[i].v,s[i].vv);
         if(gToolFlags.flagK == 1)TransformAxis((s+i),s[i].u,s[i].uu);
         if(gToolFlags.flagK == 2)TransformAxis((s+i),s[i].w,s[i].ww);
       }
     }
     DrawViewRobot(hdcRobotView,hbmRobotView,hwnd,
                  SelectedNode,SelectedRobotArm,(int)gToolFlags.flagK,
                  round_angle,up_angle,RobotZoomFlag,RobotZoomState);
     break;
   case WM_VSCROLL:
     GetScrollRange(hwnd,SB_VERT,&iMin,&iMax);
     iPos = GetScrollPos(hwnd, SB_VERT);
     switch (LOWORD(wparam)) {
       case SB_LINEDOWN:      dn =  1;    break;
       case SB_LINEUP:        dn = -1;    break;
       case SB_PAGEDOWN:      dn =  5;    break;
       case SB_PAGEUP:        dn = -5;    break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wparam)-iPos;    break;
       default:               dn = 0;
     }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       SetScrollPos(hwnd, SB_VERT, iPos + dn, TRUE);
       old=up_angle;
       up_angle=(double)(iPos+dn);
       if(old != up_angle){
         SendMessage(hwnd,WM_DRAWROBOT,0,0);
         UpdateWindow(ghwndRobotView);
       }
     }
     break;
   case WM_HSCROLL:
     GetScrollRange(hwnd, SB_HORZ, &iMin, &iMax);
     iPos = GetScrollPos(hwnd, SB_HORZ);
     switch (LOWORD(wparam)) {
       case SB_LINEDOWN:      dn =  1;     break;
       case SB_LINEUP:        dn = -1;     break;
       case SB_PAGEDOWN:      dn =  5;     break;
       case SB_PAGEUP:        dn = -5;     break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wparam) - iPos; break;
       default:               dn = 0;
     }
     if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       SetScrollPos(hwnd, SB_HORZ, iPos + dn, TRUE);
       old=round_angle;
       round_angle=(double)(iPos+dn);
       if(old != round_angle){
         SendMessage(hwnd,WM_DRAWROBOT,0,0);
         UpdateWindow(ghwndRobotView);
       }
     }
     break;
   case WM_PAINT: {
       HDC hdc;
       PAINTSTRUCT ps;
       HPALETTE hpalT;
       hdc=BeginPaint(hwnd,&ps);
//       hpalT = SelectPalette(hdc,ghpaletteScreen,FALSE);
//       RealizePalette (hdc);
       GetClientRect(hwnd,&rc);
       if(hdcRobotView != NULL)
       BitBlt(hdc,0,0,rc.right,rc.bottom,hdcRobotView,0,0,SRCCOPY);
//       SelectPalette(hdc,hpalT,FALSE);
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return(DefWindowProc(hwnd,msg,wparam,lparam));
 }
 return( 0L );
}

// end robot treeview dialog

// robot opengl viewer

#include <gl\gl.h>
#include <gl\glu.h>

static HPALETTE hRobotPalette=NULL,hRobotOldPalette=NULL;
static HGLRC  hRC=NULL;
static HDC    hRobotDC=NULL;
static RECT oldRobotRect;
static GLfloat near_plane,far_plane;
static GLfloat gbRed=0.0,gbGreen=0.1,gbBlue=0.3;

static unsigned char threeto8[8] = {
    0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

static unsigned char twoto8[4] = {
    0, 0x55, 0xaa, 0xff
};

static unsigned char oneto8[2] = {
    0, 255
};

static int defaultOverride[13] = {
    0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift){
 unsigned char val;
 val = (unsigned char) (i >> shift);
 switch (nbits) {
    case 1:
        val &= 0x1;
        return oneto8[val];
    case 2:
        val &= 0x3;
        return twoto8[val];
    case 3:
        val &= 0x7;
        return threeto8[val];
    default:
        return 0;
 }
}

static void CreateRGBPalette(HDC hDC){
 PIXELFORMATDESCRIPTOR pfd;
 LOGPALETTE *pPal;
 int n, i;
 n = GetPixelFormat(hDC);
 DescribePixelFormat(hDC, n, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
 if(pfd.dwFlags & PFD_NEED_PALETTE) {
   n = 1 << pfd.cColorBits;
   pPal = (PLOGPALETTE)LocalAlloc(LMEM_FIXED, sizeof(LOGPALETTE) +
                                   n * sizeof(PALETTEENTRY));
   pPal->palVersion = 0x300;
   pPal->palNumEntries = n;
   for (i=0; i<n; i++) {
     pPal->palPalEntry[i].peRed =
            ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
     pPal->palPalEntry[i].peGreen =
            ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
     pPal->palPalEntry[i].peBlue =
            ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
     pPal->palPalEntry[i].peFlags = 0;
   }
   if((pfd.cColorBits == 8)                           &&
       (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
       (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
       (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)) {
     for(i = 1 ; i <= 12 ; i++)
        pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
   }
   hRobotPalette = CreatePalette(pPal);
   LocalFree(pPal);
   hRobotOldPalette = SelectPalette(hDC,hRobotPalette,FALSE);
   n = RealizePalette(hDC);
 }
}

static BOOL bSetupPixelFormat(HDC hDC){
 static PIXELFORMATDESCRIPTOR pfd = {
   sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
   1,                              // version number
   PFD_DRAW_TO_WINDOW |            // support window
   PFD_SUPPORT_OPENGL |            // support OpenGL
   PFD_DOUBLEBUFFER,               // double buffered
   PFD_TYPE_RGBA,                  // RGBA type
   24,                             // 24-bit color depth
   0, 0, 0, 0, 0, 0,               // color bits ignored
   0,                              // no alpha buffer
   0,                              // shift bit ignored
   0,                              // no accumulation buffer
   0, 0, 0, 0,                     // accum bits ignored
   32,                             // 32-bit z-buffer
   0,                              // no stencil buffer
   0,                              // no auxiliary buffer
   PFD_MAIN_PLANE,                 // main layer
   0,                              // reserved
   0, 0, 0                         // layer masks ignored
 };
 int pixelformat;
 if ( (pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0 ){
   MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
   return FALSE;
 }
 if (SetPixelFormat(hDC, pixelformat, &pfd) == FALSE) {
   MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
   return FALSE;
 }
 CreateRGBPalette(hDC);
 return TRUE;
}

static GLvoid initializeGLrobot(HWND hWnd){
 GLfloat   aspect;
 GLfloat d[]={1.0,1.0,1.0,1.0};
 GLfloat a[]={0.1,0.1,0.1,1.0};
 GLfloat s[]={1.0,1.0,1.0,1.0};
 GLfloat p[]={0.5,0.5,0.5,0.0};
 if(1)p[3]=1.0;
 else p[3]=0.0;
 GetClientRect(hWnd, &oldRobotRect);
 glClearColor(gbRed,gbGreen,gbBlue,1.0);
 glMatrixMode(GL_PROJECTION);
 aspect = (GLfloat) oldRobotRect.right / oldRobotRect.bottom;;
 near_plane = 1.5; far_plane = 9.5;
 gluPerspective(45.0, aspect, near_plane, far_plane);
 glMatrixMode( GL_MODELVIEW );
 glLoadIdentity();
 glClearDepth(1.0);
 glEnable(GL_DEPTH_TEST);
 if(0)glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
 else glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
 glEnable(GL_LIGHT0);
 glLightfv(GL_LIGHT0,GL_DIFFUSE,d);
 glLightfv(GL_LIGHT0,GL_AMBIENT,a);
 glLightfv(GL_LIGHT0,GL_SPECULAR,s);
 glLightfv(GL_LIGHT0,GL_POSITION,p);
 glLightf(GL_LIGHT0,GL_LINEAR_ATTENUATION,0.125);
 glLightf(GL_LIGHT0,GL_QUADRATIC_ATTENUATION,0.02);
 glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
 glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
 glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

static void DrawViewRobotGL(HWND hWnd,
                   node *Np, int arm, int axis,
                   double ra, double ua,
                   BOOL zoom,BOOL zoom_state){
 GLfloat glos_color[]={1.0,1.0,1.0,1.0};
 GLfloat shiny[]={100.0};
 short   ep1[6]={0,4,0,1,2,0};
 short   ep2[6]={1,7,4,5,6,3};
 short   ep3[6]={2,6,5,6,7,7};
 short   ep4[6]={3,5,1,2,3,4};
 double  scale,size,x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,
         t1[4][4],t2[4][4],t3[4][4];
 vector  cc,p1,p2,p3,p4,p21,p41,nn;
 long    i,j,k,n;
 object  *Op;
 skel    *sp;
 if(Np->type == ROBOT && ((Op=Np->fobj) != NULL))while(Op != NULL){
   if(CurrentFrame >= Op->firstframe && CurrentFrame <= Op->lastframe){
     if((n=Op->nskeleton) == 0 || (sp=Op->skeleton) == NULL)break;
     /* InterpolateRobot - Not Needed other windows call it */
     size=(double)GetRobotVolume(n,NULL,cc);
     scale=3.25/size;
     if(zoom && arm >=0 && arm < n){
       if(zoom_state && (j=(sp+arm)->id) >= 0 && j < n){
         tram(t1,-((sp+j)->pp[0]),-((sp+j)->pp[1]),-((sp+j)->pp[2]));
         scale *= 5.0;
       }
       else{
         tram(t1,-((sp+arm)->pp[0]),-((sp+arm)->pp[1]),-((sp+arm)->pp[2]));
         scale *= 4.0;
       }
     }
     else tram(t1,-cc[0],-cc[1],-cc[2]);
     rotz(t2, ra*PIo180);
     m4by4(t2,t1,t3);
     rotx(t1, ua*PIo180);
     m4by4(t1,t3,t2);
     /* Bounding Box */
     glEnable(GL_LIGHTING);
     glEnable(GL_COLOR_MATERIAL);
     if(RobotGlassFlag){
       glEnable(GL_BLEND);
       glDepthMask(GL_FALSE);
       glBlendFunc(GL_SRC_ALPHA,GL_ONE);
     }
     glBegin(GL_QUADS);
     glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
     glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,glos_color);
     for(k=0;k<n;k++){
       m4by4(t2,(sp+k)->T,t3);
       if(k == arm)glColor4f(1.0,0.3,0.3,1.0);
       else        glColor4f(1.0,1.0,0.3,1.0);
       for(i=0;i<6;i++){
         m4by1(t3,(double)(sp+k)->bx[ep1[i]][0]
                 ,(double)(sp+k)->bx[ep1[i]][1]
                 ,(double)(sp+k)->bx[ep1[i]][2]
                 ,&x1,&y1,&z1);
         p1[0]=scale*(x1); p1[1]=scale*(y1); p1[2]=scale*(z1);
         m4by1(t3,(double)(sp+k)->bx[ep2[i]][0]
                 ,(double)(sp+k)->bx[ep2[i]][1]
                 ,(double)(sp+k)->bx[ep2[i]][2]
                 ,&x2,&y2,&z2);
         p2[0]=scale*(x2); p2[1]=scale*(y2); p2[2]=scale*(z2);
         m4by1(t3,(double)(sp+k)->bx[ep3[i]][0]
                 ,(double)(sp+k)->bx[ep3[i]][1]
                 ,(double)(sp+k)->bx[ep3[i]][2]
                 ,&x3,&y3,&z3);
         p3[0]=scale*(x3); p3[1]=scale*(y3); p3[2]=scale*(z3);
         m4by1(t3,(double)(sp+k)->bx[ep4[i]][0]
                 ,(double)(sp+k)->bx[ep4[i]][1]
                 ,(double)(sp+k)->bx[ep4[i]][2]
                 ,&x4,&y4,&z4);
         p4[0]=scale*(x4); p4[1]=scale*(y4); p4[2]=scale*(z4);
         VECSUB(p2,p1,p21)
         VECSUB(p4,p1,p41)
         CROSS(p21,p41,nn)
         A_normalize(nn);
         glNormal3f(nn[0],nn[2],-nn[1]);
         glVertex3f(p1[0],p1[2],-p1[1]);
         glVertex3f(p2[0],p2[2],-p2[1]);
         glVertex3f(p3[0],p3[2],-p3[1]);
         glVertex3f(p4[0],p4[2],-p4[1]);
       }
     }
     glEnd();
     if(RobotGlassFlag){
       glDepthMask(GL_TRUE);
       glDisable(GL_BLEND);
     }
     glDisable(GL_COLOR_MATERIAL);
     glDisable(GL_LIGHTING);
     /* skeleton */
     for(k=0;k<n;k++){
       x1=(double)sp[k].pp[0]; y1=(double)sp[k].pp[1]; z1=(double)sp[k].pp[2];
       m4by1(t2,x1,y1,z1,&x2,&y2,&z2);
       sp[k].wk[0]=(long)x2; sp[k].wk[1]=(long)y2; sp[k].wk[2]=(long)z2;
     }
     glLineWidth(2.0);
     if(RobotOverlayFlag)glDepthFunc(GL_ALWAYS);
     glBegin(GL_LINES);
     for(k=0;k<n;k++){
       x1=(double)sp[k].wk[0]; y1=(double)sp[k].wk[1]; z1=(double)sp[k].wk[2];
       j=sp[k].id;  /* j is parent ID */
       if(j < 0)continue;
       x2=(double)sp[j].wk[0]; y2=(double)sp[j].wk[1]; z2=(double)sp[j].wk[2];
       x1=scale*(x1); y1=scale*(y1); z1=scale*(z1);
       x2=scale*(x2); y2=scale*(y2); z2=scale*(z2);
       glColor3f(1.0,1.0,0.0);
       glVertex3f(x1,z1,-y1);
       glVertex3f(x2,z2,-y2);
       if(k == arm && axis < 7){ /* draw node and axis */
         vector pp,x={2.0,0.0,0.0},y={0.0,2.0,0.0},z={0.0,0.0,2.0};
//         Ellipse(hdcView,ix1-3,iy1-3,ix1+4,iy1+4);
         if     (axis == 0)VECCOPY((double)sp[j].vv,pp)
         else if(axis == 1)VECCOPY((double)sp[j].uu,pp)
         else if(axis == 2)VECCOPY((double)sp[j].ww,pp)
         else if(axis == 3)VECCOPY(x,pp)
         else if(axis == 4)VECCOPY(y,pp)
         else if(axis == 5)VECCOPY(z,pp)
         if(axis == 0 && (i=sp[j].id) >= 0){ /* do axis 0 other way */
           x1=sp[i].pp[0]; y1=sp[i].pp[1]; z1=sp[i].pp[2];
         }
         else if(axis < 6){
           x1=(double)sp[j].pp[0]+(double)pp[0]*size*0.1;
           y1=(double)sp[j].pp[1]+(double)pp[1]*size*0.1;
           z1=(double)sp[j].pp[2]+(double)pp[2]*size*0.1;
         }
         else {x1=sp[k].pp[0]; y1=sp[k].pp[1]; z1=sp[k].pp[2];}
         m4by1(t2,x1,y1,z1,&x3,&y3,&z3);
         x3=scale*(x3); y3=scale*(y3); z3=scale*(z3);
         glColor3f(1.0,1.0,1.0);
         glLineWidth(4.0);
         glVertex3f(x2,z2,-y2);
         glVertex3f(x3,z3,-y3);
         glLineWidth(2.0);
       }
     }
     glEnd();
     if(RobotOverlayFlag)glDepthFunc(GL_LESS);
     break;
   }
   Op=Op->next;
 }
}

static GLvoid drawRobotGLscene(HWND hWnd, double up_angle, double round_angle){
 HDC hDC;
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 if(1)glShadeModel(GL_FLAT);
 else glShadeModel(GL_SMOOTH);
 glPushMatrix();
 glLoadIdentity();
 glTranslatef(0.0, 0.0, -5.0);
// glRotatef(up_angle,1.0,0.0,0.0);
// glRotatef(round_angle,0.0,1.0,0.0);
 DrawViewRobotGL(hWnd,
                 SelectedNode,SelectedRobotArm,(int)gToolFlags.flagK,
                 round_angle,up_angle,RobotZoomFlag,RobotZoomState);
 glPopMatrix();
 glFinish();
 hDC = wglGetCurrentDC();
 SwapBuffers(hDC);
}

static GLvoid resizeRobotGL(HWND hWnd){
 RECT    rect;
 GLfloat aspect;
 GetClientRect(hWnd, &rect);
 glViewport(0,0,rect.right,rect.bottom);
 oldRobotRect.right = rect.right;
 oldRobotRect.bottom = rect.bottom;
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 aspect = (GLfloat)oldRobotRect.right/oldRobotRect.bottom;
 gluPerspective( 45.0, aspect, near_plane, far_plane);
 glMatrixMode( GL_MODELVIEW );
 glLoadIdentity();
}

static LRESULT CALLBACK RobotOglViewWndProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam ){
 static BOOL flag=FALSE,mmoved=FALSE,bTrackPan=FALSE;
 static long xs,ys;
 static double up_angle=0.0,round_angle=0.0;
 char str[32];
 double old;
 HDC  hDC;
 RECT rc;
 int i,iPos,iMax,iMin,dn;
 switch( msg ){
   case WM_CREATE:{
       vector c;
       GetRobotVolume(SelectedRobot->nskeleton,SelectedRobot->skeleton,c);
       SetScrollRange(hwnd, SB_VERT, -89,89,TRUE);
       SetScrollRange(hwnd, SB_HORZ, -179,179,TRUE);
       SetScrollPos(hwnd,SB_HORZ,(int)round_angle,TRUE);
       SetScrollPos(hwnd,SB_VERT,(int)up_angle,TRUE);
       hRobotDC = GetDC(hwnd);
       bSetupPixelFormat(hRobotDC);
       hRC = wglCreateContext(hRobotDC);
       wglMakeCurrent(hRobotDC,hRC);
       initializeGLrobot(hwnd);
     }
     break;
   case WM_DESTROY:
     wglMakeCurrent(NULL,NULL);
     if(hRC != NULL)wglDeleteContext(hRC);
     if(hRobotDC != NULL)ReleaseDC(hwnd,hRobotDC);
     GetWindowRect(hwnd,&rc);
     RobotViewXpos=rc.left; RobotViewYpos=rc.top;
     RobotViewXsize=rc.right-rc.left; RobotViewYsize=rc.bottom-rc.top;
     sprintf(str,"%ld",rc.left);
     WritePrivateProfileString(IniSection,"RVPOSX",str,IniFilename);
     sprintf(str,"%ld",rc.top);
     WritePrivateProfileString(IniSection,"RVPOSY",str,IniFilename);
     sprintf(str,"%ld",rc.right-rc.left);
     WritePrivateProfileString(IniSection,"RVSIZX",str,IniFilename);
     sprintf(str,"%ld",rc.bottom-rc.top);
     WritePrivateProfileString(IniSection,"RVSIZY",str,IniFilename);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         break;
       default:
         return(DefWindowProc(hwnd,msg,wparam,lparam));
      }
      break;
   case WM_SIZE:
     resizeRobotGL(hwnd);
     break;
   case WM_LBUTTONDOWN:
     mmoved=FALSE;
     xs=LOWORD(lparam);
     ys=HIWORD(lparam);
     SetCapture(hwnd);
     bTrackPan=TRUE;
     break;
   case WM_MOUSEMOVE:
     if(bTrackPan){
       long dx,dy;
       dx=(short)LOWORD(lparam)-xs; xs=(short)LOWORD(lparam);
       dy=(short)HIWORD(lparam)-ys; ys=(short)HIWORD(lparam);
       if(labs(dx) > 0 || labs(dy) > 0){
         mmoved=TRUE;
         round_angle=round_angle+(double)dx;
         while(round_angle < -180.0)round_angle += 360.0;
         while(round_angle >  180.0)round_angle -= 360.0;
         SetScrollPos(hwnd,SB_HORZ,(int)round_angle,TRUE);
         up_angle=up_angle+(double)dy;
         if(up_angle > -90 && up_angle < 90)
           SetScrollPos(hwnd,SB_VERT,(int)up_angle,TRUE);
         InvalidateRect(hwnd, NULL, FALSE);
       }
     }
     break;
   case WM_LBUTTONUP:
     if(bTrackPan){
       ReleaseCapture();
       bTrackPan=FALSE;
     }
     if(mmoved){
       mmoved=FALSE;
       break;
     }
     mmoved=FALSE;
   case WM_DRAWROBOT:
     if(SelectedRobot != NULL && (i=SelectedRobotArm) > 0){
       skel *s;
       s=SelectedRobot->skeleton;
       if((i=s[i].id) >= 0){
         if(gToolFlags.flagK == 0)TransformAxis((s+i),s[i].v,s[i].vv);
         if(gToolFlags.flagK == 1)TransformAxis((s+i),s[i].u,s[i].uu);
         if(gToolFlags.flagK == 2)TransformAxis((s+i),s[i].w,s[i].ww);
       }
     }
     InvalidateRect(hwnd,NULL,FALSE);
     break;
   case WM_VSCROLL:
     GetScrollRange(hwnd,SB_VERT,&iMin,&iMax);
     iPos = GetScrollPos(hwnd, SB_VERT);
     switch (LOWORD(wparam)) {
       case SB_LINEDOWN:      dn =  1;    break;
       case SB_LINEUP:        dn = -1;    break;
       case SB_PAGEDOWN:      dn =  5;    break;
       case SB_PAGEUP:        dn = -5;    break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wparam)-iPos;    break;
       default:               dn = 0;
     }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       SetScrollPos(hwnd, SB_VERT, iPos + dn, TRUE);
       old=up_angle;
       up_angle=(double)(iPos+dn);
       if(old != up_angle){
         SendMessage(hwnd,WM_DRAWROBOT,0,0);
         UpdateWindow(ghwndRobotView);
       }
     }
     break;
   case WM_HSCROLL:
     GetScrollRange(hwnd, SB_HORZ, &iMin, &iMax);
     iPos = GetScrollPos(hwnd, SB_HORZ);
     switch (LOWORD(wparam)) {
       case SB_LINEDOWN:      dn =  1;     break;
       case SB_LINEUP:        dn = -1;     break;
       case SB_PAGEDOWN:      dn =  5;     break;
       case SB_PAGEUP:        dn = -5;     break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wparam) - iPos; break;
       default:               dn = 0;
     }
     if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       SetScrollPos(hwnd, SB_HORZ, iPos + dn, TRUE);
       old=round_angle;
       round_angle=(double)(iPos+dn);
       if(old != round_angle){
         SendMessage(hwnd,WM_DRAWROBOT,0,0);
         UpdateWindow(ghwndRobotView);
       }
     }
     break;
   case WM_PAINT: {
       PAINTSTRUCT ps;
       BeginPaint(hwnd,&ps);
       drawRobotGLscene(hwnd,up_angle,round_angle);
       EndPaint(hwnd,&ps);
     }
     break;
   case WM_QUERYNEWPALETTE:{
       HPALETTE hpalOld;
       if(hRobotPalette != NULL){
         hDC = GetDC(hwnd);
         hpalOld = SelectPalette(hDC, hRobotPalette, FALSE);
         RealizePalette(hDC);
         InvalidateRect(hwnd,NULL,TRUE);
         UpdateWindow(hwnd);
         if(hpalOld != NULL)SelectPalette(hDC, hpalOld, FALSE);
         ReleaseDC(hwnd,hDC);
         return TRUE;
       }
       return FALSE;
     }
   case WM_PALETTECHANGED:{
       HPALETTE    hpalOld;
       if(hRobotPalette != NULL){
         if (wparam != (WPARAM)hwnd){
           hDC = GetDC(hwnd);
           hpalOld = SelectPalette(hDC,hRobotPalette,FALSE);
           RealizePalette(hDC);
           UpdateColors(hDC);
           if(hpalOld != NULL)SelectPalette(hDC,hpalOld,FALSE);
           ReleaseDC(hwnd,hDC);
         }
       }
     }
     break;
   default:
     return(DefWindowProc(hwnd,msg,wparam,lparam));
 }
 return( 0L );
}

// end robot opengl viewer

