/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file SUPERCLS.C          superclassing  */

#define MODULE_SUPERCLASS 1

#include <windows.h>
#include "supercls.h"

ATOM WINAPI RegisterSuperClass (LPWNDCLASS WndClass,
   WNDPROC lpfpSCWndProc, int cbClsAdditional,
   int cbWndAdditional);
WNDPROC WINAPI GetBCWndProc (HWND hWnd);
WORD  WINAPI SetSCClassWord (HWND hWnd, int nIndex, WORD wNewWord);
WORD  WINAPI GetSCClassWord (HWND hWnd, int nIndex);
DWORD WINAPI SetSCClassLong (HWND hWnd, int nIndex, DWORD dwNewLong);
DWORD WINAPI GetSCClassLong (HWND hWnd, int nIndex);
WORD  WINAPI SetSCWindowWord (HWND hWnd, int nIndex, WORD wNewWord);
WORD  WINAPI GetSCWindowWord (HWND hWnd, int nIndex);
DWORD WINAPI GetSCWindowLong (HWND hWnd, int nIndex);
DWORD WINAPI SetSCWindowLong (HWND hWnd, int nIndex,
   DWORD dwNewLong);
LRESULT CALLBACK NoAlphaWndProc (HWND hWnd, UINT uMsg,
   WPARAM wParam, LPARAM lParam);

#define BCWNDPROCINDEX     (sizeof(WNDPROC))
#define CBBCCLSEXTRAINDEX  (sizeof(WNDPROC) + sizeof(int))
#define CBBCWNDEXTRAINDEX  \
   (sizeof(WNDPROC) + sizeof(int) + sizeof(int))

#define MINCBCLSADDTIONAL  \
      (sizeof(WNDPROC) + sizeof(int) + sizeof(int))

typedef struct {
   WNDPROC lpfnSuperClsWndProc;
   WNDPROC lpfnBCWndProc;
   int cbBCClsExtra;
   int cbBCWndExtra;
} SUPERCLSINIT, FAR *LPSUPERCLSINIT;


static LRESULT CALLBACK SuperClsInitWndProc (HWND hWnd, UINT uMsg,
   WPARAM wParam, LPARAM lParam) {
   LPSUPERCLSINIT lpSuperClsInit;
   int cbClsTotal;
   switch (uMsg) {
      case WM_CREATE:
         cbClsTotal = GetClassLong(hWnd, GCL_CBCLSEXTRA);
         SetClassLong(hWnd, cbClsTotal - BCWNDPROCINDEX,
            (LONG) (((LPCREATESTRUCT) lParam)->lpCreateParams));
         break;
      case WM_NCDESTROY:
         cbClsTotal = GetClassLong(hWnd, GCL_CBCLSEXTRA);
         lpSuperClsInit = (LPSUPERCLSINIT)
            GetClassLong(hWnd, cbClsTotal - BCWNDPROCINDEX);
         SetClassLong(hWnd, cbClsTotal - BCWNDPROCINDEX,
            (LONG) lpSuperClsInit->lpfnBCWndProc);
         SetClassWord(hWnd, cbClsTotal - CBBCCLSEXTRAINDEX,
            (WORD)lpSuperClsInit->cbBCClsExtra);
         SetClassWord(hWnd, cbClsTotal - CBBCWNDEXTRAINDEX,
            (WORD)lpSuperClsInit->cbBCWndExtra);
         SetClassLong(hWnd, GCL_WNDPROC,
            (LONG) lpSuperClsInit->lpfnSuperClsWndProc);
      break;
   }
   return(DefWindowProc(hWnd, uMsg, wParam, lParam));
}

ATOM WINAPI RegisterSuperClass (LPWNDCLASS WndClass,
   WNDPROC lpfnSCWndProc, int cbClsAdditional,
   int cbWndAdditional) {
   HWND hWnd;
   SUPERCLSINIT SuperClsInit;
   SuperClsInit.lpfnSuperClsWndProc = lpfnSCWndProc;
   SuperClsInit.lpfnBCWndProc = WndClass->lpfnWndProc;
   SuperClsInit.cbBCClsExtra  = WndClass->cbClsExtra;
   SuperClsInit.cbBCWndExtra  = WndClass->cbWndExtra;
   WndClass->cbClsExtra += cbClsAdditional + MINCBCLSADDTIONAL;
   WndClass->cbWndExtra += cbWndAdditional;
   WndClass->lpfnWndProc = SuperClsInitWndProc;
   if (RegisterClass(WndClass) == 0) return 0;
   hWnd = CreateWindow(WndClass->lpszClassName, "", 0, 0, 0, 0, 0,
      NULL, NULL, WndClass->hInstance, (LPSTR) &SuperClsInit);
   if (hWnd == NULL) {
      UnregisterClass(WndClass->lpszClassName,
         WndClass->hInstance);
      return(FALSE);
   }
   DestroyWindow(hWnd);
   _ValidRangeMsg = RegisterWindowMessage("ValidRange");
   _SetRangeMsg   = RegisterWindowMessage("SetRange");
   return(TRUE);
}

WNDPROC WINAPI GetBCWndProc (HWND hWnd) {
   return((WNDPROC) GetClassLong(hWnd,
      GetClassLong(hWnd, GCL_CBCLSEXTRA) - BCWNDPROCINDEX));
}

static int NEAR CalcClassByteIndex (HWND hWnd, int nIndex) {
   int cbBCClsExtraIndex, cbBCClsExtra;
   if (nIndex < 0) return(nIndex);
   cbBCClsExtraIndex =
      GetClassLong(hWnd, GCL_CBCLSEXTRA) - CBBCCLSEXTRAINDEX;
   cbBCClsExtra = GetClassWord(hWnd, cbBCClsExtraIndex);
   return(nIndex + cbBCClsExtra);
}

static int NEAR CalcWindowByteIndex (HWND hWnd, int nIndex) {
   int cbBCWndExtraIndex, cbBCWndExtra;
   if (nIndex < 0) return(nIndex);
   cbBCWndExtraIndex =
      GetClassLong(hWnd, GCL_CBCLSEXTRA) - CBBCWNDEXTRAINDEX;
   cbBCWndExtra = GetClassWord(hWnd, cbBCWndExtraIndex);
   return(nIndex + cbBCWndExtra);
}

WORD WINAPI SetSCClassWord (HWND hWnd, int nIndex,
   WORD wNewWord) {
   nIndex = CalcClassByteIndex(hWnd, nIndex);
   return(SetClassWord(hWnd, nIndex, wNewWord));
}

WORD WINAPI GetSCClassWord (HWND hWnd, int nIndex) {
   nIndex = CalcClassByteIndex(hWnd, nIndex);
   return(GetClassWord(hWnd, nIndex));
}

DWORD WINAPI SetSCClassLong (HWND hWnd, int nIndex,
   DWORD dwNewLong) {
   nIndex = CalcClassByteIndex(hWnd, nIndex);
   return(SetClassLong(hWnd, nIndex, dwNewLong));
}

DWORD WINAPI GetSCClassLong (HWND hWnd, int nIndex) {
   nIndex = CalcClassByteIndex(hWnd, nIndex);
   return(GetClassLong(hWnd, nIndex));
}

WORD WINAPI SetSCWindowWord (HWND hWnd, int nIndex,
   WORD wNewWord) {
   nIndex = CalcWindowByteIndex(hWnd, nIndex);
   return(SetWindowWord(hWnd, nIndex, wNewWord));
}

WORD WINAPI GetSCWindowWord (HWND hWnd, int nIndex) {
   nIndex = CalcWindowByteIndex(hWnd, nIndex);
   return(GetWindowWord(hWnd, nIndex));
}

DWORD WINAPI SetSCWindowLong (HWND hWnd, int nIndex,
   DWORD dwNewLong) {
   nIndex = CalcWindowByteIndex(hWnd, nIndex);
   return(SetWindowLong(hWnd, nIndex, dwNewLong));
}

DWORD WINAPI GetSCWindowLong (HWND hWnd, int nIndex) {
   nIndex = CalcWindowByteIndex(hWnd, nIndex);
   return(GetWindowLong(hWnd, nIndex));
}

/*****************************************************************/
#define NOALPHA_CBCLSEXTRA    (0)

#define NOALPHA_CBWNDEXTRA    (4)
#define GSCWW_LOVALUE         (0)
#define GSCWW_HIVALUE         (2)

ATOM RegisterNoAlphaWndClass (HINSTANCE hInstance) {
   WNDCLASS WndClass;
   GetClassInfo(NULL, "EDIT", &WndClass);
   WndClass.lpszClassName = "TIDE";
   WndClass.hInstance     = hInstance;
   return(RegisterSuperClass(
      &WndClass,           // Address of WNDCLASS structure.
      NoAlphaWndProc,      // Addr of superclass window procedure.
      NOALPHA_CBCLSEXTRA,  // # of additional class extra bytes.
      NOALPHA_CBWNDEXTRA   // # of additional window extra bytes.
      ));
}

LRESULT CALLBACK NoAlphaWndProc (HWND hWnd, UINT uMsg,
   WPARAM wParam, LPARAM lParam) {
   LRESULT lResult = 0L;
   BOOL fCallBaseClassWndProc = TRUE;
   char szValue[10];
   int nValue, nLoValue, nHiValue;
   if (uMsg == _SetRangeMsg) {
      SetSCWindowWord(hWnd, GSCWW_LOVALUE, LOWORD(lParam));
      SetSCWindowWord(hWnd, GSCWW_HIVALUE, HIWORD(lParam));
      fCallBaseClassWndProc = FALSE;
   }
   if (uMsg == _ValidRangeMsg) {
      GetWindowText(hWnd, szValue, sizeof(szValue));
      nValue = atoi(szValue);
      nLoValue = GetSCWindowWord(hWnd, GSCWW_LOVALUE);
      nHiValue = GetSCWindowWord(hWnd, GSCWW_HIVALUE);
      lResult = (nLoValue <= nValue) && (nValue <= nHiValue);
      fCallBaseClassWndProc = FALSE;
   }
   switch (uMsg) {
      case WM_CREATE:
         SetSCWindowWord(hWnd, GSCWW_LOVALUE, 0);
         SetSCWindowWord(hWnd, GSCWW_HIVALUE, 32767);
         break;
      case WM_CHAR:
         if (wParam >= 'A' && wParam <= 'Z')
            fCallBaseClassWndProc = FALSE;
         if (wParam >= 'a' && wParam <= 'z')
            fCallBaseClassWndProc = FALSE;
         if (fCallBaseClassWndProc == FALSE) MessageBeep(0);
         break;
      default: // Pass other messages to base class procedure.
         break;
   }
   if (fCallBaseClassWndProc) {
      lResult =
         CallWindowProc((WNDPROC) GetBCWndProc(hWnd), hWnd,
            uMsg, wParam, lParam);
   }
   return(lResult);
}
