/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file TOOLBOX .C */

#define MODULE_TOOLBOX 1

#include "animate.h"

typedef struct tagQUICKTIME{
  HWND hwnd;
  WNDPROC fpSclass;
  HBITMAP hbm;
  int idBitmap;
  int idCommand,idHelp,iTip;
} QUICKTIME;

typedef struct tagTOOL_STRUCT{
 long  Message,Help,Tip;
} TOOL_STRUCT;

extern HBITMAP hBmBackground;

#define LARGE_TOOL_X 50
#define LARGE_TOOL_Y 40
#define SMALL_TOOL_X 32
#define SMALL_TOOL_Y 32
#define N_ICONS       9
static HBITMAP hBitmapB=NULL,
               hBitmapH=NULL,
               hBitmapD=NULL,
               hBitmapI=NULL,
               hBitmapM=NULL;
static BOOL    tools_enable=TRUE;
static long    cxToolBtn,cyToolBtn;
static char    szToolsClass[]="OFX:AnimatorToolsClass";
static LRESULT CALLBACK ToolsWndProc(HWND,UINT,WPARAM,LPARAM);
static TOOL_STRUCT aTools[N_ICONS]={
  IDM_STOP_TOOL,        IDI_TOOL1,           IDI_TIP_T_1,
  IDM_ACTOR_TOOL_NODE,  IDI_TOOL2,           IDI_TIP_T_2,
  IDM_ACTOR_TOOL_PAN,   IDI_TOOL3,           IDI_TIP_T_3,
  IDM_ACTOR_TOOL_ZOOMIN,IDI_TOOL4,           IDI_TIP_T_4,
  IDM_ACTOR_TOOL_GRAB,  IDI_TOOL5,           IDI_TIP_T_5,
  IDM_ACTOR_TOOL_ROTATE,IDI_TOOL6,           IDI_TIP_T_6,
  IDM_ACTOR_TOOL_SIZE,  IDI_TOOL7,           IDI_TIP_T_7,
  IDM_ACTOR_QADD,       IDI_TOOL8,           IDI_TIP_T_8,
  IDM_FRAME_RENDER,     IDI_COMMAND_RENDER,  IDI_TIP_A_1
};

static void ToolboxCreate(void);
static void PannelSelectTool(HWND hwnd, int *oldtool, int newtool);
static void DoAddTooltip(HWND hwndOwner, HWND tipon, int resid);
static HWND DoCreateTooltip(HWND hwndOwner);

#define NO_MESSAGE -1
#define NO_HELP    -1

static QUICKTIME QtWin[6]={
 NULL,NULL,NULL,0,       -1,                  NO_HELP,           IDI_TIP_Q_2,
 NULL,NULL,NULL,IDBM_QT1,IDM_FRAME_GO_LASTKEY,IDI_TIME_LASTKEY,  IDI_TIP_Q_1,
 NULL,NULL,NULL,IDBM_QT4,IDM_FRAME_GO_NEXTKEY,IDI_TIME_NEXTKEY,  IDI_TIP_Q_5,
 NULL,NULL,NULL,IDBM_QT2,IDM_FRAME_PLAY_QUICK,IDI_TIME_PLAYQUICK,IDI_TIP_Q_3,
 NULL,NULL,NULL,IDBM_QT3,IDM_FRAME_PLAY_WIRE, IDI_TIME_PLAYWIRE, IDI_TIP_Q_4,
// NULL,NULL,NULL,0,       -1,                  NO_HELP,           IDI_TIP_Q_2
 NULL,NULL,NULL,0,       IDM_FRAME_GO_FRAME,  IDI_GOFRAME,       IDI_TIP_Q_2
};

void ToolboxShow(BOOL fShow){
 if(!ghwndQuicktime)ToolboxCreate();
}

void ToolboxDestroy(void){
 if(ghwndQuicktime)DestroyWindow(ghwndQuicktime);
 if(ghwndCoord1)DestroyWindow(ghwndCoord1);
 if(ghwndTools)DestroyWindow(ghwndTools);
 ghwndQuicktime=NULL;
 ghwndCoord1=NULL;
 ghwndTools=NULL;
}

void AugmentToolPannel(BOOL state){
 if(state)UNDO_TYPE=1;
 else     UNDO_TYPE=0;
}

void EnableTools(BOOL status){
  tools_enable=status;
  InvalidateRect(ghwndTools,NULL,FALSE);
}

void MoveToolsWindow(void){
 int i,x,y;
 RECT rc1;
 POINT pt;
 if(ghwndTools == NULL)return;
 if(BigCameraView)GetWindowRect(ghwnd_triview[ActiveView],&rc1);
 else             GetWindowRect(ghwnd_view,&rc1);
 if(bToolpannelLeft)x=WindowGap;
 else x=rc1.right+3;
 y=rc1.top;
 pt.x=x; pt.y=y;
 ScreenToClient(ghwnd_main,&pt);
 x=pt.x; y=pt.y;
 for(i=0;i<N_ICONS;i++){
   MoveWindow(GetDlgItem(ghwndTools,i),
              0,cyToolBtn*i,cxToolBtn,cyToolBtn,FALSE);
 }
 MoveWindow(ghwndTools,x,y,cxToolBtn,cyToolBtn*N_ICONS,TRUE);
}

void ChangeToolboxBitmaps(void){
 if(hBitmapB != NULL)DeleteObject(hBitmapB);  hBitmapB=NULL;
 if(hBitmapH != NULL)DeleteObject(hBitmapH);  hBitmapH=NULL;
 if(hBitmapD != NULL)DeleteObject(hBitmapD);  hBitmapD=NULL;
 if(hBitmapI != NULL)DeleteObject(hBitmapI);  hBitmapI=NULL;
 if(hBitmapM != NULL)DeleteObject(hBitmapM);  hBitmapM=NULL;
 if(bToolpannelSmall){
   cxToolBtn=SMALL_TOOL_X;
   cyToolBtn=SMALL_TOOL_Y;
   hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_B));
   hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_H));
   hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_D));
   hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_I));
   hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_M));
 }
 else{
   cxToolBtn=LARGE_TOOL_X;
   cyToolBtn=LARGE_TOOL_Y;
   hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_B));
   hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_H));
   hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_D));
   hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_I));
   hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_M));
 }
}

static long GetWhichButton(void){
  long  i,button_index = -1;
  POINT pt;
  RECT  rc;
  GetCursorPos(&pt);
  for(i=0;i<N_ICONS;i++){
    GetWindowRect(GetDlgItem(ghwndTools,i),&rc);
    if(PtInRect(&rc,pt))button_index=i;
  }
  return button_index;
}

static void ToolboxCreate(void){
 char title[256];
 int x,y;
 RECT rc,rc1;
 POINT pt;
// if(Preferences.tooltips){
 if((ghwndTooltip=DoCreateTooltip(ghwnd_main)) == NULL)return;
 SetWindowPos(ghwndTooltip,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE | SWP_NOSIZE);
 SendMessage(ghwndTooltip,TTM_ACTIVATE,(WPARAM)Preferences.tooltips,0);
// }
 if((ghwndQuicktime = CreateWindow(szQuicktimeClass,NULL,
    WS_CHILD | WS_VISIBLE | !WS_BORDER,
    WindowViewSliderXpos,WindowViewSliderYpos,
    WindowViewSliderXsize,WindowViewSliderYsize,
    ghwnd_main,NULL,ghinst_main,NULL)) == NULL)return;
 GetWindowRect(ghwnd_view,&rc1);
 if(bToolpannelLeft)x=WindowGap;
 else x=rc1.right+3;
 y=rc1.top;
 pt.x=x; pt.y=y;
 ScreenToClient(ghwnd_main,&pt);
 x=pt.x; y=pt.y;
 if(bToolpannelSmall){
   cxToolBtn=SMALL_TOOL_X;
   cyToolBtn=SMALL_TOOL_Y;
 }
 else{
   cxToolBtn=LARGE_TOOL_X;
   cyToolBtn=LARGE_TOOL_Y;
 }
 ghwndTools=CreateWindowEx(
                         WS_EX_TRANSPARENT,
                         szToolboxClass,NULL,
                         WS_CHILD | !WS_BORDER | WS_VISIBLE,
                         x,y,
                         cxToolBtn,cyToolBtn*N_ICONS,
                         ghwnd_main,
                         (HMENU)101,
                         ghinst_main,NULL);
 ghwndCoord1=CreateDialog(ghinst_main,MAKEINTRESOURCE(DLG_COORDS),
                          ghwnd_main,(DLGPROC)CoordinatesDlgProc);
 return;
}

LRESULT CALLBACK ToolboxWndProc(HWND hwnd,UINT msg,
                                WPARAM wParam,LPARAM lParam){
 long i;
 HWND hWndChild;
 switch (msg) {
   case WM_CREATE:
     if(bToolpannelSmall){
       hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_B));
       hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_H));
       hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_D));
       hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_I));
       hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_M));
     }
     else{
       hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_B));
       hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_H));
       hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_D));
       hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_I));
       hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_M));
     }
     for(i=0;i<N_ICONS;i++){
       hWndChild=CreateWindowEx(
                              WS_EX_TRANSPARENT,
                              szToolBtnClass,NULL,
                              WS_CHILD | WS_VISIBLE,
                              0,cyToolBtn*i,
                              cxToolBtn,cyToolBtn,
                              hwnd,(HMENU)i,
                              ghinst_main,NULL);
       DoAddTooltip(ghwnd_main,hWndChild,aTools[i].Tip);
     }
     break;
   case WM_DESTROY:
     if(hBitmapB != NULL)DeleteObject(hBitmapB);  hBitmapB=NULL;
     if(hBitmapH != NULL)DeleteObject(hBitmapH);  hBitmapH=NULL;
     if(hBitmapD != NULL)DeleteObject(hBitmapD);  hBitmapD=NULL;
     if(hBitmapI != NULL)DeleteObject(hBitmapI);  hBitmapI=NULL;
     if(hBitmapM != NULL)DeleteObject(hBitmapM);  hBitmapM=NULL;
     break;
   default:
     return DefWindowProc(hwnd, msg, wParam, lParam);
 }
 return 0;
}

LRESULT CALLBACK ToolBtnWndProc(HWND hwnd,UINT msg,
                                WPARAM wParam,LPARAM lParam){
 static int  toolTimer = -1, down_tool;
 static BOOL bCaptured=FALSE,bCancel=FALSE,bLastCancel=FALSE,bBusy=FALSE;
 long w,wold,hid;
 MSG mmsg;
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 hid=(int)GetWindowLong(hwnd,GWL_ID);
 switch (msg) {
   case WM_TIMER:
     if(bBusy || toolTimer < 0)break;
     bBusy=TRUE;
     w=GetWhichButton();
     if(w < 0 && toolTimer >= 0){
       KillTimer(hwnd,toolTimer);
       toolTimer = -1;
     }
     if( w < 0 && gCurTool >= 0){
       wold=gCurTool;
       gCurTool= -1;
       InvalidateRect(GetDlgItem(ghwndTools,wold),NULL,FALSE);
     }
     bBusy=FALSE;
     break;
   case WM_SETCURSOR:{
       hid=aTools[hid].Help;
       if(current_message != hid){
         if(hid == NO_HELP)SendPrgmMessage(IDS_ZERO,1);
         else SendIconHelpMessage(hid);
       }
       current_message=hid;
     }
     return DefWindowProc(hwnd,msg,wParam,lParam);
   case WM_MOUSEMOVE:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     w=GetWhichButton();
     if(bCaptured){
       if(w == hid)bCancel=FALSE;
       else        bCancel=TRUE;
       if(bCancel != bLastCancel)InvalidateRect(hwnd,NULL,FALSE);
       bLastCancel=bCancel;
     }
     else{
       wold=gCurTool;
       gCurTool=w;
       if(wold != w){
         if(wold >= 0 && wold  < N_ICONS)
           InvalidateRect(GetDlgItem(ghwndTools,wold),NULL,FALSE);
         if(gCurTool >= 0 && gCurTool < N_ICONS)
           InvalidateRect(GetDlgItem(ghwndTools,gCurTool),NULL,FALSE);
       }
       if(toolTimer < 0)toolTimer=SetTimer(hwnd,100,500,NULL);
     }
     break;
   case WM_LBUTTONDBLCLK:
     w=GetWhichButton();
     if(w == 2)PostMessage(ghwnd_main,WM_COMMAND,IDM_ZOOM_CENTRE,0);
     if(w == 3)PostMessage(ghwnd_main,WM_COMMAND,IDM_VIEW_FULL,0);
     if(w == 4)PostMessage(ghwnd_main,WM_COMMAND,IDM_ACTOR_TOOL_GRABABS,0);
     if(w == 5)PostMessage(ghwnd_main,WM_COMMAND,IDM_ACTOR_TOOL_ROTATEABS,0);
     if(w == 6)PostMessage(ghwnd_main,WM_COMMAND,IDM_ACTOR_TOOL_SIZEABS,0);
     break;
   case WM_LBUTTONDOWN:
     if(toolTimer >= 0)KillTimer(hwnd,toolTimer); toolTimer = -1;
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     bCaptured=TRUE;
     bCancel=bLastCancel=FALSE;
     down_tool=hid;
     SetCapture(hwnd);
     InvalidateRect(hwnd,NULL,FALSE);
     break;
   case WM_LBUTTONUP:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     ReleaseCapture();
     w=GetWhichButton();
     bCaptured=FALSE;
     if(w < 0 && gCurTool >= 0){
       wold=gCurTool;
       gCurTool = -1;
       InvalidateRect(GetDlgItem(ghwndTools,wold),NULL,FALSE);
     }
     InvalidateRect(hwnd,NULL,FALSE);
     if(down_tool == w && aTools[down_tool].Message != NO_MESSAGE)
       PostMessage(ghwnd_main,WM_COMMAND,aTools[w].Message,0);
     gCurTool= -1;
     break;
   case WM_RBUTTONDOWN:{
       POINT pt;
       HMENU hMenu=LoadMenu(ghinst_main,"ToolbarPopupMenu");
       if(hMenu == NULL)break;
       pt.x=LOWORD(lParam);
       pt.y=HIWORD(lParam);
       ClientToScreen(hwnd,&pt);
       if(bToolpannelLeft)
         CheckMenuItem(hMenu,IDM_CONTEXT_TOOLSLEFT,MF_CHECKED);
       if(bToolpannelSmall)
         CheckMenuItem(hMenu,IDM_CONTEXT_TOOLSSMALL,MF_CHECKED);
       TrackPopupMenu(GetSubMenu(hMenu,0),TPM_LEFTALIGN | TPM_LEFTBUTTON,
                      pt.x,pt.y,0,ghwnd_main,NULL);
       DestroyMenu(hMenu);
     }
     break;
   case WM_PAINT:{
       HDC hDC;
       PAINTSTRUCT ps;
       hDC=BeginPaint(hwnd,&ps);
       if(hBitmapB != NULL && hBitmapD != NULL  && hBitmapH != NULL){
         HDC hMemDC;
         HBITMAP hbmOld;
         BOOL blit=FALSE;
         hMemDC = CreateCompatibleDC(hDC);
         if(!tools_enable){
           hbmOld = SelectObject(hMemDC,hBitmapI);
           blit=TRUE;
         }
         else if( (tool == NOTOOL     && hid == 0)      ||
                  (tool == NODETOOL   && hid == 1)      ||
                  (tool == TRACKER    && hid == 1)      ||
                  (tool == SKELETOR   && hid == 1)      ||
                  (tool == PATHEDITOR && hid == 1)      ||
                  (tool == PAN        && hid == 2)      ||
                  (tool == INZOOM     && hid == 3)      ||
                  (tool == GRABBER    && hid == 4)      ||
                  (tool == ROTATOR    && hid == 5)      ||
                  (tool == SIZER      && hid == 6)      ||
                  (hwnd == GetCapture() && !bCancel)  ||
                  FALSE)
           hbmOld = SelectObject(hMemDC,hBitmapD);
         else if(hid == gCurTool)
           hbmOld = SelectObject(hMemDC,hBitmapH);
         else{
           hbmOld = SelectObject(hMemDC,hBitmapB);
           blit=TRUE;
         }
         if(hBmBackground == NULL)blit=FALSE;
         if(blit){
           HDC hdc_view;
           RECT rc;
           POINT pp;
           BITMAP bm;
           HBITMAP hbm_view,hbmoldv;
           hbm_view=CreateCompatibleBitmap(hMemDC,cxToolBtn,cyToolBtn);
           hdc_view=CreateCompatibleDC(hMemDC);
           hbmoldv=SelectObject(hdc_view,hbm_view);
           GetWindowRect(hwnd,&rc);
           pp.x=rc.left; pp.y=rc.top;
           ScreenToClient(ghwnd_main,&pp);
           GetObject(hBmBackground,sizeof(BITMAP),&bm);
           pp.x = pp.x % bm.bmWidth;
           pp.y = pp.y % bm.bmHeight;
           SelectObject(hMemDC,hBmBackground);
           if(pp.y + cyToolBtn > bm.bmHeight){
             if(pp.x + cxToolBtn > bm.bmWidth){
               BitBlt(hDC,
                  0,0,
                  bm.bmWidth -pp.x,
                  bm.bmHeight-pp.y,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
               BitBlt(hDC,
                  bm.bmWidth-pp.x,bm.bmHeight-pp.y,
                  cxToolBtn - (bm.bmWidth -pp.x),
                  cyToolBtn - (bm.bmHeight-pp.y),
                  hMemDC,
                  (pp.x + (bm.bmWidth -pp.x)) % bm.bmWidth,
                  (pp.y + (bm.bmHeight-pp.y)) % bm.bmHeight,
                  SRCCOPY);
             }
             else{
               BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  bm.bmHeight-pp.y,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
               BitBlt(hDC,
                  0,bm.bmHeight-pp.y,
                  cxToolBtn,
                  cyToolBtn - (bm.bmHeight-pp.y),
                  hMemDC,
                  pp.x,(pp.y + (bm.bmHeight-pp.y)) % bm.bmHeight,
                  SRCCOPY);
             }
           }
           else if(pp.x + cxToolBtn > bm.bmWidth){
             BitBlt(hDC,
                  0,0,
                  bm.bmWidth -pp.x,
                  cyToolBtn,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
             BitBlt(hDC,
                  bm.bmWidth-pp.x,0,
                  cxToolBtn - (bm.bmWidth -pp.x),
                  cyToolBtn,
                  hMemDC,
                  (pp.x + (bm.bmWidth -pp.x)) % bm.bmWidth,pp.y,
                  SRCCOPY);
           }
           else BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
           SelectObject(hMemDC,hBitmapM);
           BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  0x00220326); // DSna
           if(tools_enable)SelectObject(hMemDC,hBitmapB);
           else            SelectObject(hMemDC,hBitmapI);
           BitBlt(hdc_view,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  SRCCOPY);
           SelectObject(hMemDC,hBitmapM);
           BitBlt(hdc_view,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  SRCAND);
           BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hdc_view,
                  0,0,
                  SRCPAINT);
           SelectObject(hdc_view,hbmoldv);
           DeleteObject(hbm_view);
           DeleteDC(hdc_view);
         }
         else BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  SRCCOPY);
         SelectObject(hMemDC,hbmOld);
         DeleteDC(hMemDC);
       }
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return DefWindowProc(hwnd,msg,wParam,lParam);
 }
 return 0;
}

void ToolboxSetPosition(int w){
}

static void DrawQtBtn(BOOL selected,HDC hDC, LPRECT lprc, HBITMAP hbm){
 HDC hMemDC;
 BITMAP  bmp;
 HBITMAP hbmOld;
 HBRUSH holdbrush,hbrushFace;
 HPEN   holdpen,hpenShadow,hpenHighlight;
 int x,y,cx,cy;
 hbrushFace=CreateSolidBrush(RGB(192,192,192));
 holdbrush=SelectObject(hDC,hbrushFace);
 hpenShadow=CreatePen(PS_SOLID,0,RGB(128,128,128));
 hpenHighlight=CreatePen(PS_SOLID,0,RGB(255,255,255));
 holdpen=SelectObject(hDC,GetStockObject(BLACK_PEN));
 Rectangle(hDC,0,0,lprc->right+1,lprc->bottom);
 SelectObject(hDC,hpenShadow);
 if(selected){
   MoveToEx(hDC,1,lprc->bottom-2,NULL);
   LineTo(hDC,1,1);
   LineTo(hDC,lprc->right-1,1);
 }
 else{
   MoveToEx(hDC,1,lprc->bottom-2,NULL);
   LineTo(hDC,lprc->right-1,lprc->bottom-2);
   LineTo(hDC,lprc->right-1,1);
   MoveToEx(hDC,2,lprc->bottom-3,NULL);
   LineTo(hDC,lprc->right-2,lprc->bottom-3);
   LineTo(hDC,lprc->right-2,3);
   SelectObject(hDC,hpenHighlight);
   MoveToEx(hDC,1,lprc->bottom-3,NULL);
   LineTo(hDC,1,1);
   LineTo(hDC,lprc->right-1,1);
   MoveToEx(hDC,2,lprc->bottom-4,NULL);
   LineTo(hDC,2,2);
   LineTo(hDC,lprc->right-2,2);
 }
 SelectObject(hDC,holdpen);
 DeleteObject(hpenShadow);
 DeleteObject(hpenHighlight);
 SelectObject(hDC,holdbrush);
 DeleteObject(hbrushFace);
 if(hbm == NULL)return;
 GetObject(hbm,sizeof(BITMAP),&bmp);
 cx = bmp.bmWidth;
 cy = bmp.bmHeight;
 x=max(0,(lprc->right-cx)/2); y=max(0,(lprc->bottom-cy)/2)+1;
 if(selected){x++; y++;}
 hMemDC = CreateCompatibleDC(hDC);
 hbmOld = SelectObject(hMemDC, hbm);
 BitBlt(hDC,x,y,cx,cy,hMemDC,0,0,SRCAND);
 SelectObject(hMemDC, hbmOld);
 DeleteDC(hMemDC);
}

static LRESULT CALLBACK QuicktimeSubclassProc(HWND hwnd,UINT msg,
                                       WPARAM wParam,LPARAM lParam){
 int hid;
 MSG mmsg;
 LRESULT lRet;
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 switch (msg) {
   case WM_KEYDOWN:
    SetFocus(ghwnd_main);
    return 0;
   case WM_LBUTTONDOWN:
   case WM_LBUTTONUP:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     break;
   case WM_MOUSEMOVE:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     hid=QtWin[GetWindowLong(hwnd,GWL_ID)].idHelp;
     if(hid == NO_HELP)SendPrgmMessage(IDS_ZERO,1);
         else {
           if(current_message != hid){
              SendIconHelpMessage(hid);
              current_message=hid;
           }
         }
     break;
   default: break;
 }
 lRet=CallWindowProc(QtWin[GetWindowLong(hwnd,GWL_ID)].fpSclass,
                     hwnd,msg,wParam,lParam);
 switch (msg) {
   case WM_LBUTTONUP:
     SetFocus(ghwnd_main);
     break;
   default: break;
 }
 return lRet;
}

#define STATICSIZE 40

LRESULT CALLBACK QuicktimeWndProc(HWND hwnd,UINT msg,WPARAM wParam,
                                  LPARAM lParam){
 static BOOL bCaptured=FALSE;
 static RECT wrc;
 HDC  hDC;
 HWND hdesk;
 BITMAP bmp;
 static int  xc,yc;
 int i,xp,sxsize,scsize,btnsize,btnxsize;
 switch (msg) {
   case WM_CREATE:
     for(i=0;i<6;i++){
       if(i == 0){
//
//         QtWin[i].hwnd=CreateWindow("SCROLLBAR",NULL,
//           WS_CHILD | WS_VISIBLE | SBS_HORZ | SBS_TOPALIGN,
//           0,0,0,0,hwnd,(HMENU)i,ghinst_main,NULL);
// for reference
         QtWin[i].hwnd=CreateWindowEx(0,TRACKBAR_CLASS,"",
           WS_CHILD | WS_BORDER | WS_VISIBLE | TBS_BOTTOM,// | TBS_AUTOTICKS,
           0,0,0,0,
           hwnd,(HMENU)i,ghinst_main,NULL);
//         DoAddTooltip(ghwnd_main,QtWin[i].hwnd,QtWin[i].iTip);
//
         QtWin[i].fpSclass=NULL;
         ghwnd_xscl=QtWin[i].hwnd;
       }
       else if(i == 5){ /* static text */
         QtWin[i].hwnd=CreateWindow("button",NULL,
           WS_CHILD | WS_VISIBLE | !WS_BORDER | BS_PUSHBUTTON,
           0,0,0,0,hwnd,(HMENU)i,ghinst_main,NULL);
         QtWin[i].fpSclass=(WNDPROC)SetWindowLong(QtWin[i].hwnd,GWL_WNDPROC,
           (LONG)QuicktimeSubclassProc);
         DoAddTooltip(ghwnd_main,QtWin[i].hwnd,QtWin[i].iTip);
//         QtWin[i].hwnd=CreateWindow("static",NULL,
//           WS_CHILD | WS_VISIBLE | WS_BORDER | SS_CENTER,
//           0,0,0,0,hwnd,(HMENU)i,ghinst_main,NULL);
         SendMessage(QtWin[i].hwnd,WM_SETFONT,
                    (WPARAM)GetStockObject(ANSI_VAR_FONT),0);
         SendMessage(QtWin[i].hwnd,WM_SETTEXT,0,(LPARAM)"0001");
       }
       else{
         QtWin[i].hwnd=CreateWindow("button",NULL,
           WS_CHILD | WS_VISIBLE | !WS_BORDER | BS_OWNERDRAW | BS_PUSHBUTTON,
           0,0,0,0,hwnd,(HMENU)i,ghinst_main,NULL);
         QtWin[i].fpSclass=(WNDPROC)SetWindowLong(QtWin[i].hwnd,GWL_WNDPROC,
           (LONG)QuicktimeSubclassProc);
         QtWin[i].hbm=LoadBitmap(ghinst_main,
           MAKEINTRESOURCE(QtWin[i].idBitmap));
         DoAddTooltip(ghwnd_main,QtWin[i].hwnd,QtWin[i].iTip);
       }
     }
     return 0;
   case WM_SIZE:
     xp=0;btnsize=HIWORD(lParam); sxsize=LOWORD(lParam);
//     if(Xres > 800)btnxsize=2*btnsize;
//     else          btnxsize=(3*btnsize/2);
     if(Xres <= 800)btnxsize=(2*btnsize/3); // New resolution for bigger buttons
     else           btnxsize=btnsize;       // RSF
     for(i=0;i<6;i++){
       if(i == 0)scsize=sxsize-4*btnxsize-STATICSIZE;
       else scsize=btnxsize;
       if(i == 0){ /* for Windows95 */
         BOOL v4;
         DWORD version;
         version=GetVersion();
         if((version & 0x80000000) == 0x80000000){
          if(LOBYTE(LOWORD(version)) < 4)v4=FALSE;
           else                           v4=TRUE;
         }
         else{
          if(LOBYTE(LOWORD(version)) < 4)v4=FALSE;
           else                           v4=TRUE;
         }
         if(v4)MoveWindow(QtWin[i].hwnd,xp,0,scsize+1,btnsize,TRUE);
         else  MoveWindow(QtWin[i].hwnd,xp,0,scsize+1,btnsize,TRUE);
       }
       else if(i == 5){
//         MoveWindow(QtWin[i].hwnd,xp,0,STATICSIZE,btnsize,TRUE);   // text
         MoveWindow(QtWin[i].hwnd,xp,1,STATICSIZE,btnsize-1,TRUE); // Button
       }
       else MoveWindow(QtWin[i].hwnd,xp,0,scsize,btnsize,TRUE);
       xp += scsize;
     }
     break;
   case WM_SETCURSOR:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     return( DefWindowProc(hwnd,msg,wParam,lParam));
   case WM_DRAWITEM:{
       UINT idCtl;
       LPDRAWITEMSTRUCT lpdis;
       idCtl=(UINT)wParam;
       lpdis=(LPDRAWITEMSTRUCT)lParam;
       if(idCtl < 5){
         if(lpdis->itemState & ODS_SELECTED){
           DrawQtBtn(TRUE,lpdis->hDC,&(lpdis->rcItem),QtWin[idCtl].hbm);
         }
         else{
           DrawQtBtn(FALSE,lpdis->hDC,&(lpdis->rcItem),QtWin[idCtl].hbm);
         }
       }
     }
     return TRUE;
   case WM_LBUTTONDOWN:
     if(0){ /* comment out drag facility */
       bCaptured=TRUE;
       SetCapture(hwnd);
       hdesk=GetDesktopWindow();
       hDC=GetWindowDC(hdesk);
       GetWindowRect(hwnd,&wrc);
       DrawFocusRect(hDC,&wrc);
       ReleaseDC(hdesk,hDC);
       xc=(int)((short)LOWORD(lParam)); yc=(int)((short)HIWORD(lParam));
     }
     break;
   case WM_MOUSEMOVE:
     if(bCaptured){
       hdesk=GetDesktopWindow();
       hDC=GetWindowDC(hdesk);
       DrawFocusRect(hDC,&wrc);
       OffsetRect(&wrc,(int)((short)LOWORD(lParam)-(short)xc),
                       (int)((short)HIWORD(lParam)-(short)yc));
       DrawFocusRect(hDC,&wrc);
       ReleaseDC(hdesk,hDC);
       xc=LOWORD(lParam); yc=HIWORD(lParam);
     }
     break;
   case WM_LBUTTONUP:
     if(bCaptured){
       POINT pt;
       ReleaseCapture();
       hdesk=GetDesktopWindow();
       hDC=GetWindowDC(hdesk);
       DrawFocusRect(hDC,&wrc);
       ReleaseDC(hdesk,hDC);
       bCaptured=FALSE;
       pt.x=wrc.left; pt.y=wrc.top;
       ScreenToClient(ghwnd_main,&pt); /* for Child Window */
       SetWindowPos(hwnd,NULL,pt.x,pt.y,0,0,SWP_NOSIZE);
       InvalidateRect(hwnd,NULL,TRUE); /* for Child Window */
     }
     break;
   case WM_COMMAND:
     if(HIWORD(wParam) == BN_CLICKED){
       i = -1;
       if((HWND)lParam == QtWin[1].hwnd)i=1;
       if((HWND)lParam == QtWin[2].hwnd)i=2;
       if((HWND)lParam == QtWin[3].hwnd)i=3;
       if((HWND)lParam == QtWin[4].hwnd)i=4;
       if((HWND)lParam == QtWin[5].hwnd)i=5;
       if(i >= 0)SendMessage(ghwnd_main,WM_COMMAND,QtWin[i].idCommand,0);
     }
     SetFocus(ghwnd_main);
     break;
   case WM_HSCROLL:
     PostMessage(ghwnd_main,WM_HSCROLL,wParam,lParam);
     break;
   case WM_DESTROY:
     for(i=0;i<6;i++){
       if(QtWin[i].fpSclass != NULL)SetWindowLong(QtWin[i].hwnd,GWL_WNDPROC,
            (LONG)QtWin[i].fpSclass);
       if(QtWin[i].hbm != NULL)DeleteObject(QtWin[i].hbm);
       QtWin[i].hbm = NULL;
     }
     break;
   default:
     return DefWindowProc(hwnd, msg, wParam, lParam);
 }
 return 0;
}

void UpdateFrameBox(void){
 char str[32];
 sprintf(str,"%4.4ld",CurrentFrame);
 SendMessage(QtWin[5].hwnd,WM_SETTEXT,0,(LPARAM)str);
 return;
}

static void GetFormatString(char *format, double x, double y, double z,
 double dx, double dy, double dz){
 if(fabs(x)  > 9999999.0 || fabs(y)  > 9999999.0 ||
    fabs(z)  > 9999999.0 || fabs(dx) > 9999999.0 ||
    fabs(dy) > 9999999.0 || fabs(dz) > 9999999.0)
   sprintf(format,"%s","%9.1le");
 else if(fabs(x)  > 99999.0 || fabs(y)  > 99999.0 ||
         fabs(z)  > 99999.0 || fabs(dx) > 99999.0 ||
         fabs(dy) > 99999.0 || fabs(dz) > 99999.0)
   sprintf(format,"%s","%9.0lf");
 else  sprintf(format,"%s","%9.2lf");
}

static void GetDoubleCoordinates(int ix, int iy, int iz,
                     double *x,  double *y,  double *z,
                     double *dx, double *dy, double *dz){
 *x = (double)(ix-lrulerx);
 *y = (double)(iy-lrulery);
 *z = (double)(iz-lrulerz);
 *dx = (double)(ix-rulerx);
 *dy = (double)(iy-rulery);
 *dz = (double)(iz-rulerz);
 *x  /= ruler; *y   /=ruler; *z   /=ruler;
 *dx /= ruler; *dy /= ruler; *dz /= ruler;
}

#define NC_BUTTONS 5

static struct QC {int idBitmap; int dlgID; int idHelp; int iTip;
       HWND hwnd;WNDPROC fpSclass; HBITMAP hbm;} QtCoord[NC_BUTTONS]={
  IDBM_CO1,DLG_COORDS_GOTO,     IDI_COORD_MEASURE, IDI_TIP_C_1,NULL,NULL,NULL,
  IDBM_CO2,DLG_COORDS_CANCEL,   IDI_COORD_RELATIVE,IDI_TIP_C_2,NULL,NULL,NULL,
  IDBM_CO3,DLG_COORDS_SETCENTRE,IDI_COORD_ORIGIN,  IDI_TIP_C_3,NULL,NULL,NULL,
  IDBM_CO4,DLG_COORDS_DEFINE,   IDI_COORD_DEFINE,  IDI_TIP_C_4,NULL,NULL,NULL,
  IDBM_CO5,DLG_COORDS_MEASURE,  IDI_COORD_NEAREST, IDI_TIP_C_5,NULL,NULL,NULL
};

static LRESULT CALLBACK CoordinatesSubclassProc(HWND hwnd,UINT msg,
                                       WPARAM wParam,LPARAM lParam){
 MSG mmsg;
 int i,hid;
 for(i=0;i<NC_BUTTONS;i++){
   if(QtCoord[i].hwnd == hwnd)goto XXIT;
 }
 MessageBeep(MB_OK);
 XXIT:
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 switch (msg) {
   case WM_LBUTTONDOWN:
   case WM_LBUTTONUP:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,
                                         0,(LPARAM)(LPMSG)&mmsg);
     break;
   case WM_MOUSEMOVE:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,
                                         0,(LPARAM)(LPMSG)&mmsg);
     hid=QtCoord[i].idHelp;
     if(hid == NO_HELP)SendPrgmMessage(IDS_ZERO,1);
         else SendIconHelpMessage(hid);
     break;
   default: break;
 }
 return CallWindowProc(QtCoord[i].fpSclass,hwnd,msg,wParam,lParam);
}

LRESULT CALLBACK CoordinatesDlgProc(HWND hdlg,UINT msg,
                            WPARAM wparam,LPARAM lparam){
 BOOL err;
 RECT rc;
 int i,ix,iy,iz;
 double x,y,z,dx,dy,dz;
 char string[32],format[10];
 static int xpos=-1,ypos=-1,infocus=0,lastid=-1,udflag=1;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     for(i=0;i<NC_BUTTONS;i++){
       QtCoord[i].hbm=LoadBitmap(ghinst_main,
                                 MAKEINTRESOURCE(QtCoord[i].idBitmap));
       QtCoord[i].hwnd=GetDlgItem(hdlg,QtCoord[i].dlgID);
       if(QtCoord[i].hwnd != NULL){
         QtCoord[i].fpSclass=(WNDPROC)SetWindowLong(QtCoord[i].hwnd,
                              GWL_WNDPROC,(LONG)CoordinatesSubclassProc);
         DoAddTooltip(ghwnd_main,QtCoord[i].hwnd,QtCoord[i].iTip);
       }
     }
     if(xpos == -1 && ypos == -1){
       xpos=GetPrivateProfileInt(IniSection,"XCOORDPOS",-1,
                                 IniFilename);
       if(xpos < 0)xpos=52;
       ypos=GetPrivateProfileInt(IniSection,"YCOORDPOS",-1,
                                 IniFilename);
       if(ypos < 0)ypos=50;
     }
     SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     udflag=1;
     PostMessage(hdlg,WM_COMMAND,IDC_COORD_UPDATE,0);
     return (TRUE);
   case WM_CTLCOLORSTATIC:
   case WM_CTLCOLORDLG:
     SetBkColor((HDC)wparam,RGB(192,192,192));
     return (LRESULT)GetStockObject(LTGRAY_BRUSH);
   case WM_DRAWITEM:{
       BOOL status;
       HBITMAP hbm;
       LPDRAWITEMSTRUCT lpdis;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->itemState & ODS_SELECTED)status=TRUE; else status=FALSE;
       if     (lpdis->CtlID == DLG_COORDS_GOTO)hbm=QtCoord[0].hbm;
       else if(lpdis->CtlID == DLG_COORDS_CANCEL)hbm=QtCoord[1].hbm;
       else if(lpdis->CtlID == DLG_COORDS_SETCENTRE)hbm=QtCoord[2].hbm;
       else if(lpdis->CtlID == DLG_COORDS_DEFINE)hbm=QtCoord[3].hbm;
       else if(lpdis->CtlID == DLG_COORDS_MEASURE)hbm=QtCoord[4].hbm;
       DrawQtBtn(status,lpdis->hDC,&(lpdis->rcItem),hbm);
     }
     return TRUE;
   case WM_DESTROY:
     for(i=0;i<NC_BUTTONS;i++){
       if(QtCoord[i].hbm != NULL)DeleteObject(QtCoord[i].hbm);
       if(QtCoord[i].fpSclass != NULL)SetWindowLong(QtCoord[i].hwnd,
         GWL_WNDPROC,(LONG)QtCoord[i].fpSclass);
     }
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     sprintf(string,"%ld",xpos);
     WritePrivateProfileString(IniSection,"XCOORDPOS",string,IniFilename);
     sprintf(string,"%ld",ypos);
     WritePrivateProfileString(IniSection,"YCOORDPOS",string,IniFilename);
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_COMMAND:
     switch(HIWORD(wparam)){
       case EN_SETFOCUS:
         infocus=1;
         break;
       case EN_KILLFOCUS:
         infocus=0;
         break;
       case EN_CHANGE:
         if(!infocus)break;
         udflag=1;
         i=(int)LOWORD(wparam);
         if(GetDlgItemText(hdlg,i,string,30) != 0){
           if(i == DLG_COORDS_X){
             if(lastid == DLG_COORDS_DX)break;
             ix=(int)(atof(string)*ruler)+lrulerx;
             GetDoubleCoordinates(ix,NpointerY,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,dx);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_DX,string);
             lastid=-1;
           }
           else if(i == DLG_COORDS_Y){
             if(lastid == DLG_COORDS_DY)break;
             iy=(int)(atof(string)*ruler)+lrulery;
             GetDoubleCoordinates(NpointerX,iy,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,dy);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_DY,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_Z){
             if(lastid == DLG_COORDS_DZ)break;
             iz=(int)(atof(string)*ruler)+lrulerz;
             GetDoubleCoordinates(NpointerX,NpointerY,iz,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,dz);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_DZ,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_DX){
             if(lastid == DLG_COORDS_X)break;
             ix=(int)(atof(string)*ruler)+rulerx;
             GetDoubleCoordinates(ix,NpointerY,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,x);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_X,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_DY){
             if(lastid == DLG_COORDS_Y)break;
             iy=(int)(atof(string)*ruler)+rulery;
             GetDoubleCoordinates(NpointerX,iy,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,y);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_Y,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_DZ){
             if(lastid == DLG_COORDS_Z)break;
             iz=(int)(atof(string)*ruler)+rulerz;
             GetDoubleCoordinates(NpointerX,NpointerY,iz,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,z);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_Z,string);
             lastid= -1;
           }
         }
         break;
       default:   break;
     }
     switch(LOWORD(wparam)){
       case IDCANCEL:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_WINDOW_COORDS,0);
         break;
       case IDOK:
       case DLG_COORDS_GOTO:{
           point TVC;
           if(GetDlgItemText(hdlg,DLG_COORDS_X,string,30) == 0)break;
           TVC[0]=(long)(atof(string)*ruler)+lrulerx;
           if(GetDlgItemText(hdlg,DLG_COORDS_Y,string,30) == 0)break;
           TVC[1]=(long)(atof(string)*ruler)+lrulery;
           if(GetDlgItemText(hdlg,DLG_COORDS_Z,string,30) == 0)break;
           TVC[2]=(long)(atof(string)*ruler)+lrulerz;
           if(in_stage_triview(TVC)){
             Draw3dCursor();
             NpointerX=TVC[0]; NpointerY=TVC[1]; NpointerZ=TVC[2];
             Draw3dCursor();
           }
           else{
             NpointerX=TVC[0]; NpointerY=TVC[1]; NpointerZ=TVC[2];
             TVpointX=NpointerX - TVsizeX/2;
             TVpointY=NpointerY - TVsizeY/2;
             TVpointZ=NpointerZ - TVsizeZ/2;
             ReDrawStageDisplay(TRUE);
             PerspectiveView(0,0);
           }
           PostMessage(hdlg,WM_COMMAND,IDC_COORD_UPDATE,0);
         }
         break;
       case DLG_COORDS_CANCEL:
         if(HIWORD(wparam) != BN_CLICKED)break;
         udflag=1;
         PostMessage(hdlg,WM_COMMAND,IDC_COORD_UPDATE,0);
         break;
       case DLG_COORDS_SETCENTRE:
         if(HIWORD(wparam) != BN_CLICKED)break;
         SendMessage(ghwnd_main,WM_COMMAND,IDM_COORDS_ORIGIN,0);
         break;
       case DLG_COORDS_MEASURE:
         if(HIWORD(wparam) != BN_CLICKED)break;
         if(SelectedNode != NULL){
           point p,tp;
           short im;
           double phi,theta,alpha,mratio,sx,sy,sz,ima;
           GetTransform(0,CurrentFrame,1.0,
                        SelectedNode,SelectedNode->fobj,p,tp,
                        &phi,&theta,&alpha,&sx,&sy,&sz,&im,&ima);
           dx=(double)(p[0]-NpointerX)/ruler;
           dy=(double)(p[1]-NpointerY)/ruler;
           dz=(double)(p[2]-NpointerZ)/ruler;
           ima=sqrt(dx*dx+dy*dy+dz*dz);
           sprintf(string,"D %17.3lf",ima);
         }
         else sprintf(string,"N/A");
         SetDlgItemText(hdlg,DLG_COORDS_DIST,string);
         break;
       case DLG_COORDS_DEFINE:
         if(HIWORD(wparam) != BN_CLICKED)break;
         SendMessage(ghwnd_main,WM_COMMAND,IDM_COORDS_DEFINE,0);
         break;
       case IDC_COORD_UPDATE:
         GetDoubleCoordinates(NpointerX,NpointerY,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
         GetFormatString(format,x,y,z,dx,dy,dz);
         sprintf(string,format,x);  SetDlgItemText(hdlg,DLG_COORDS_X,string);
         sprintf(string,format,y);  SetDlgItemText(hdlg,DLG_COORDS_Y,string);
         sprintf(string,format,z);  SetDlgItemText(hdlg,DLG_COORDS_Z,string);
         rulerx=NpointerX; rulery=NpointerY; rulerz=NpointerZ;
         if(udflag){
           sprintf(string,format,0.0);SetDlgItemText(hdlg,DLG_COORDS_DX,string);
           sprintf(string,format,0.0);SetDlgItemText(hdlg,DLG_COORDS_DY,string);
           sprintf(string,format,0.0);SetDlgItemText(hdlg,DLG_COORDS_DZ,string);
           udflag=0;
         }
         break;
       default:   break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static void DoAddTooltip(HWND hwndOwner, HWND tipon, int resid){
 TOOLINFO ti;
 if(ghwndTooltip == NULL || tipon == NULL)return;
 memset(&ti,0,sizeof(TOOLINFO));
 ti.cbSize = sizeof(TOOLINFO);
 ti.uFlags = TTF_IDISHWND;
 ti.hwnd = hwndOwner;
 ti.hinst = ghinst_main;
 ti.uId = (UINT)tipon;
 ti.lpszText = (LPSTR)resid;
 SendMessage(ghwndTooltip,TTM_ADDTOOL,0,(LPARAM) (LPTOOLINFO) &ti);
 return;
}

static HWND DoCreateTooltip(HWND hwndOwner){
 HWND hwndTT;
 hwndTT = CreateWindow(TOOLTIPS_CLASS, (LPSTR) NULL, TTS_ALWAYSTIP,
     CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
     hwndOwner, (HMENU) NULL, ghinst_main, NULL);
 return hwndTT;
}

