/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* view_cel.c */


static FILE *Bit_File;
static unsigned char *Screen=NULL;

static short RenderFlcBrun(FILE *f);
static short RenderFlcColor(FILE *f, short type);
static short LoadRenderAnimCel(FILE *f, Flc_head *flc, short info_only);

static long nx,ny,nx4;

static long LoadCEL(char *name, short info_only){
 long i,cel_type;
 struct pic_header aa_cel;
 Flc_head flc_cel;
 unsigned char cc,*pix;
 if((Bit_File=fopen(name,"rb")) == NULL)return 0;
 if(fread(&aa_cel,sizeof(struct pic_header),1,Bit_File) != 1)cel_type=0;
 else if(aa_cel.type == PIC_MAGIC){
   cel_type=1;
   nx = x_size = aa_cel.w;
   ny = y_size = aa_cel.h;
 }
 else{
   fseek(Bit_File,0,SEEK_SET);
   if(fread(&flc_cel,sizeof(Flc_head),1,Bit_File) != 1)cel_type=0;
   else if(flc_cel.type == FLCH_MAGIC){
     cel_type=2;
     nx = x_size = flc_cel.width;
     ny = y_size = flc_cel.height;
   }
 }
 nx4=ALIGNULONG(nx);
 if(cel_type == 0){fclose(Bit_File); return(0);}
 if(cel_type == 2){
   if(LoadRenderAnimCel(Bit_File,&flc_cel,info_only) < 0){
     fclose(Bit_File);
     return(0);
   }
 }
 else{
   for(i = 0; i < 256 ; i++) {
     fread(&cc,1,1,Bit_File); p_red[i]  = cc << 2;
     fread(&cc,1,1,Bit_File); p_green[i]  = cc << 2;
     fread(&cc,1,1,Bit_File); p_blue[i]  = cc << 2;
   }
   if(info_only){
     fclose(Bit_File);
     return 1;
   }
   for(i=0;i<ny;i++){
     pix=Screen+(ny-1-i)*nx4;
     if(fread(pix,1,nx,Bit_File) != nx){
       fclose(Bit_File);
       return(0);
     }
   }
 }
 fclose(Bit_File);
 return(1);
}

static short LoadRenderAnimCel(FILE *f,Flc_head *flc, short info_only){
 Flc_frame f_frame;
 Flc_chunk f_chunk;
 long i,j,k;
 char c;
 unsigned char *p;
 fseek(f,flc->oframe1,SEEK_SET);
 if(fread(&f_frame,sizeof(Flc_frame),1,f) != 1 ||
           f_frame.type != FLCF_MAGIC)return -1;
/*printf("point2  %ld chunks\n",f_frame.chunks);*/
 for(i=0;i<f_frame.chunks;i++){
   if(fread(&f_chunk,sizeof(Flc_chunk),1,f) != 1)return -1;
/*printf("chunk type %ld\n",(long)f_chunk.type);*/
   if(f_chunk.type == FLC_COLOR256 || f_chunk.type == FLC_COLOR){
     if(RenderFlcColor(f,f_chunk.type) < 0)return -1;
     if(info_only)return 0;
   }
   else if(!info_only && f_chunk.type == FLC_BRUN){
     if(RenderFlcBrun(f) < 0)return -1;
   }
   else if(!info_only && f_chunk.type == FLC_COPY){
     for (k=0;k<ny;k++){
       p=Screen+(ny-1-k)*nx4;
      if(fread(p,1,nx,f) != nx)return -1;
     }
   }
   else{
     for(j=0;j<f_chunk.size-6;j++)fread(&c,1,1,f);
   }
 }
 return 0;
}

static short RenderFlcColor(FILE *f, short type){
 short i,j,pk,packets,ci=0;
 unsigned char skip=0,change,colour;
 if(fread(&packets,2,1,f) != 1)return -1;
 for (pk=0; pk<packets; pk++ ) {
   if(fread(&skip,1,1,f) != 1)return -1;
   ci += skip;
   if(fread(&change,1,1,f) != 1)return -1;
   if (change == 0) change=255;
   for (i=0; i<=change; i++) {
     if(fread(&colour,1,1,f) != 1)return -1;
     if(type == FLC_COLOR)colour = colour << 2;
     p_red[ci] = colour;
     if(fread(&colour,1,1,f) != 1)return -1;
     if(type == FLC_COLOR)colour = colour << 2;
     p_green[ci] = colour;
     if(fread(&colour,1,1,f) != 1)return -1;
     if(type == FLC_COLOR)colour = colour << 2;
     p_blue[ci] = colour;
     ci++;
   }
 }
 return 0;
}

static short RenderFlcBrun(FILE *f){
 char size=0;
 unsigned char packet,*p,colour;
 short i,line,pk;
 for (line=0;line<ny;line++){
   p=Screen+(ny-1-line)*nx4;
   if(fread(&packet,1,1,f) != 1)return -1;
   for(pk=0; pk<nx; pk+=abs(size)){
     if(fread(&size,1,1,f) != 1)return -1;
     if(size > 0){
       if(fread(&colour,1,1,f) != 1)return -1;
       for (i=0; i<size; i++) *p++ = colour;
     }
     else{
       for (i = size; i<0; i++){
         if(fread(&colour,1,1,f) != 1)return -1;
         *p++ = colour;
       }
     }
   }
 }
 return 0;
}

static int celview(HWND hWnd){
 int i;
 char temp[128];
 long imagesize;
 LPBITMAPINFOHEADER lpbi;
 FreeDib();
 rowcount=0;
 strcpy(temp,CurrentFile);
 if(LoadCEL(temp,1) < 0){
   LoadErrMsg(IDX_ERR_NOCEL);
   return 0;
 }
 x4_size = ALIGNULONG(x_size);
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)
                                +256L*sizeof(RGBQUAD));
 if(lpBitmapHeader == NULL){
   LoadErrMsg(IDX_ERR_MEMORY);
   return 0;
 }
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x4_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=8;
 lpbi->biCompression=BI_RGB;
 imagesize = lpbi->biSizeImage = (DWORD)x4_size*(DWORD)y_size;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=256;
 lpbi->biClrImportant=256;
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<256;i++){
   pRgb[i].rgbRed = p_red[i];
   pRgb[i].rgbGreen = p_green[i];
   pRgb[i].rgbBlue = p_blue[i];
 }
 hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits=(LPSTR)X__Malloc(imagesize);
 if(lpBitmapBits == NULL){
   LoadErrMsg(IDX_ERR_MEMORY);
   FreeDib();
   return 0;
 }
 memset(lpBitmapBits,0,imagesize);
 Screen=(unsigned char *)lpBitmapBits;
 if(LoadCEL(temp,0) < 0){
   PutMsg("Error reading CEL file");
   FreeDib();
   return 0;
 }
 hbmCurrent=BitmapFromDib(lpBitmapHeader,lpBitmapBits,hpalCurrent);
 if(hbmCurrent == NULL)LoadErrMsg(IDX_ERR_NOBITMAP);
 SizeWindow(hWnd);
 return 1;
}
