/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* File VRMLX.C DLL   ||   */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <windows.h>
#if __HELP_WS__
#include "..\help_ws\include\htmlhelp.h"
#endif

#include "astruct.h"
#include "dstruct.h"

#include "vrmlX.h"

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HWND      hParent;
static HINSTANCE hThisInstance;

#define DOS386
#define _X__FAR
#define PI  3.1415926

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

/* definition of flags and identifiers*/

static long indent=0;
static long digits=3,interpolate=3;
static BOOL bIndent=TRUE,bPrefix=FALSE;
static BOOL bViewpoints=TRUE,bInterpolate=FALSE;
static char map_path[256]="../objects/";
static char globalfile[256]="";
static char back_image[256]="";
static char front_image[256]="";
static char left_image[256]="";
static char right_image[256]="";
static char top_image[256]="";
static char bottom_image[256]="";

static BOOL WriteDataToRegistry(char *keyname){
 HKEY hkey;
 DWORD result;
 DWORD dw;
 if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,keyname,0,NULL,
                   REG_OPTION_NON_VOLATILE,
                   KEY_ALL_ACCESS,NULL,&hkey,&result) != ERROR_SUCCESS)
   return FALSE;
 RegSetValueEx(hkey,"object_prefix",0,REG_SZ,(LPBYTE)map_path,strlen(map_path)+1);
 RegSetValueEx(hkey,"anim_file",0,REG_SZ,(LPBYTE)globalfile,strlen(globalfile)+1);
 dw=(DWORD)bPrefix;
 RegSetValueEx(hkey,"use_object_prefix",0,REG_DWORD,(LPBYTE)(&dw),sizeof(DWORD));
 dw=(DWORD)digits;
 RegSetValueEx(hkey,"digits",0,REG_DWORD,(LPBYTE)(&dw),sizeof(DWORD));
 dw=(DWORD)bIndent;
 RegSetValueEx(hkey,"indent",0,REG_DWORD,(LPBYTE)(&dw),sizeof(DWORD));

 dw=(DWORD)bViewpoints;
 RegSetValueEx(hkey,"viewpoints",0,REG_DWORD,(LPBYTE)(&dw),sizeof(DWORD));
 dw=(DWORD)bInterpolate;
 RegSetValueEx(hkey,"interpolate",0,REG_DWORD,(LPBYTE)(&dw),sizeof(DWORD));
 dw=(DWORD)interpolate;
 RegSetValueEx(hkey,"interpolate_value",0,REG_DWORD,(LPBYTE)(&dw),sizeof(DWORD));
 RegSetValueEx(hkey,"back_image",0,REG_SZ,(LPBYTE)back_image,strlen(back_image)+1);
 RegSetValueEx(hkey,"front_image",0,REG_SZ,(LPBYTE)front_image,strlen(front_image)+1);
 RegSetValueEx(hkey,"left_image",0,REG_SZ,(LPBYTE)left_image,strlen(left_image)+1);
 RegSetValueEx(hkey,"right_image",0,REG_SZ,(LPBYTE)right_image,strlen(right_image)+1);
 RegSetValueEx(hkey,"top_image",0,REG_SZ,(LPBYTE)top_image,strlen(top_image)+1);
 RegSetValueEx(hkey,"bottom_image",0,REG_SZ,(LPBYTE)bottom_image,strlen(bottom_image)+1);

 RegCloseKey(hkey);
 return TRUE;
}

static BOOL ReadDataFromRegistry(char *keyname){
 HKEY hkey;
 DWORD result;
 DWORD len,type,dw;
 char  str[256];
 if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,keyname,0,KEY_ALL_ACCESS,&hkey)
   != ERROR_SUCCESS)return FALSE;
 len=256;
 if(RegQueryValueEx(hkey,"object_prefix",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(map_path,str);
 len=256;
 if(RegQueryValueEx(hkey,"anim_file",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(globalfile,str);
 len=sizeof(DWORD);
 if(RegQueryValueEx(hkey,"use_object_prefix",NULL,&type,(LPBYTE)(&dw),&len)
    == ERROR_SUCCESS)bPrefix=(BOOL)dw;
 len=sizeof(DWORD);
 if(RegQueryValueEx(hkey,"indent",NULL,&type,(LPBYTE)(&dw),&len)
    == ERROR_SUCCESS)bIndent=(BOOL)dw;
 len=sizeof(DWORD);
 if(RegQueryValueEx(hkey,"digits",NULL,&type,(LPBYTE)(&dw),&len)
    == ERROR_SUCCESS)digits=(long)dw;

 len=sizeof(DWORD);
 if(RegQueryValueEx(hkey,"viewpoints",NULL,&type,(LPBYTE)(&dw),&len)
    == ERROR_SUCCESS)bViewpoints=(BOOL)dw;
 len=sizeof(DWORD);
 if(RegQueryValueEx(hkey,"interpolate",NULL,&type,(LPBYTE)(&dw),&len)
    == ERROR_SUCCESS)bInterpolate=(BOOL)dw;
 len=sizeof(DWORD);
 if(RegQueryValueEx(hkey,"interpolate_value",NULL,&type,(LPBYTE)(&dw),&len)
    == ERROR_SUCCESS)interpolate=(long)dw;
 len=256;
 if(RegQueryValueEx(hkey,"back_image",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(back_image,str);
 len=256;
 if(RegQueryValueEx(hkey,"front_image",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(front_image,str);
 len=256;
 if(RegQueryValueEx(hkey,"left_image",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(left_image,str);
 len=256;
 if(RegQueryValueEx(hkey,"right_image",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(right_image,str);
 len=256;
 if(RegQueryValueEx(hkey,"top_image",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(top_image,str);
 len=256;
 if(RegQueryValueEx(hkey,"bottom_image",NULL,&type,(LPBYTE)str,&len)
    == ERROR_SUCCESS)strcpy(bottom_image,str);
 RegCloseKey(hkey);
 return TRUE;
}
static char *FileOnly(char *file){
 char *t;
 t=strrchr(file,'\\');
 if(t == NULL)return file;
 return (t+1);
}

static void ReplaceExt(char *str){
 char *t;
 t=strrchr(str,'.');
 if(t == NULL)return;
 strcpy((t+1),"wrl");
 return;
}

static void DeleteExt(char *str){
 char *t;
 t=strchr(str,'.');
 if(t == NULL)return;
 *t='\0';
 t=strchr(str,' ');
 if(t == NULL)return;
 *t='\0';
 return;
}

static void Id(FILE *f){
 int i;
 if(!bIndent)return;
 if(indent > 0)for(i=0;i<indent;i++)fprintf(f,"  ");
}

typedef double quaternion[4];

void EulerToQuaternion(double p, double t, double a, quaternion q){
  a *= 0.5; t *= 0.5; p *= 0.5;
  q[0]=cos(a)*cos(t)*cos(p)+sin(a)*sin(t)*sin(p);
  q[1]=sin(a)*cos(t)*cos(p)-cos(a)*sin(t)*sin(p);
  q[2]=cos(a)*sin(t)*cos(p)+sin(a)*cos(t)*sin(p);
  q[3]=cos(a)*cos(t)*sin(p)-sin(a)*sin(t)*cos(p);
}

void TransformOrientation(double phi, double theta, double alpha,
                          double *ux, double *uy, double *uz,
                          double *ua){
 double w;
 quaternion q;
 *ux = 0.0; *uy = 0.0; *uz= 1.0; *ua = 0.0;
 EulerToQuaternion(phi*PI/180.0,theta*PI/180.0,alpha*PI/180.0,q);
 if(fabs(q[0] - 1.0) < 1.e-5)return;
 *ua = 2.0*acos(q[0]);
 w=1.0/sqrt(1.0-q[0]*q[0]);
 *ux = q[1]*w;
 *uy = q[2]*w;
 *uz = q[3]*w;
 return;
}

void CreateVRMLfile(FILE *dxfp){
 char format2[64],format3[64],format4[64],tempstr[256],tempstr1[256];
 position *Pp;
 node     *Np;
 object   *Op;
 sky      *Sp;
 long i,k,ccf;
 point p,tp;
 short im,refl;
 double phi,theta,alpha,mratio,sx,sy,sz,ima,gamb,oamb,m2ratio,frac;
 unsigned char sr,sg,sb,zr,zg,zb,gr,gg,gb;
 double ux,uy,uz,ua;
 char filename[255];
 BOOL bSky=FALSE,bGround=FALSE;
 sprintf(format3,"%%.%ldf %%.%ldf %%.%ldf",digits,digits,digits);
 sprintf(format4,"%%.%ldf %%.%ldf %%.%ldf %%.%ldf",digits,digits,digits,digits);
 sprintf(format2,"%%.%ldf %%.%ldf",digits,digits);
 // write-group
 Id(dxfp); fprintf(dxfp,"DEF SFX-ANIM Group {\n");  indent++;
 Id(dxfp); fprintf(dxfp,"children [\n"); indent++;
 // Viewpoints
 if(bViewpoints){
 Np=FirstNp; while(Np!=NULL){
   if(Np->type == CAMERA && Np == SelectedCamera){
     if((Op=Np->fobj) != NULL && (Pp=Np->fpos) != NULL)while(Pp != NULL){
       ccf=CurrentFrame;
       CurrentFrame=Pp->lastframe;
       GetTransform(0,Pp->lastframe,1.0,
                    Np,Op,p,tp,&phi,&theta,&alpha,
                    &sx,&sy,&sz,&im,&ima);
       CurrentFrame=ccf;
       Id(dxfp); fprintf(dxfp,"DEF %s Viewpoint {\n",
                         Np->actorname /*,Pp->lastframe */);
       indent++;
       Id(dxfp); fprintf(dxfp,"fieldOfView 0.78\n");
       Id(dxfp); fprintf(dxfp,"description \"%s Key-%ld\"\n",Np->actorname,Pp->lastframe);
       Id(dxfp); fprintf(dxfp,"position ");
       Id(dxfp); fprintf(dxfp,format3,
                              (double)p[0]/(double)UNIT,
                              (double)p[2]/(double)UNIT,
                             -(double)p[1]/(double)UNIT);
       Id(dxfp); fprintf(dxfp,"\n");
       Id(dxfp); fprintf(dxfp,"# phi=%lf theta=%lf alpha=%lf\n",phi,theta,alpha);
       Id(dxfp); fprintf(dxfp,"orientation ");
       // theta is bank alpha is pitch phi is turn ????
       TransformOrientation(phi,-theta,alpha,&ux,&uy,&uz,&ua);
       Id(dxfp); fprintf(dxfp,format4,ux,uz,-uy,ua);
       Id(dxfp); fprintf(dxfp,"\n");
       indent--;
       Id(dxfp); fprintf(dxfp,"}\n");
       Pp=Pp->next;
     }
     break;
   }
   Np=Np->next;
 }
 }
 // TimeSensors
 if(bInterpolate){
   Np=FirstNp; while(Np!=NULL){
     if(Np->type == CAMERA && Np == SelectedCamera)break;
     Np=Np->next;
   }
   frac=1.0/(double)interpolate;
   Id(dxfp);
     fprintf(dxfp,"DEF TIMER-%s TimeSensor { loop FALSE cycleInterval %.3lf}\n",
   Np->actorname,frac);

   Id(dxfp); fprintf(dxfp,
     "DEF POS-TWEEN-%s PositionInterpolator{\n",Np->actorname); indent++;
   Id(dxfp); fprintf(dxfp,"key ["); indent++;
   for(i=1,k=1;i<=Nframes;i+=interpolate,k++){
     fprintf(dxfp,"%.3lf",(double)(i)/(double)(Nframes));
     if(k%10 == 0){
       fprintf(dxfp,",\n");
       Id(dxfp);
     }
     else fprintf(dxfp,", ");
   }
   indent--;
   fprintf(dxfp,"\n");
   Id(dxfp); fprintf(dxfp,"]\n"); // end key
   Id(dxfp); fprintf(dxfp,"keyValue ["); indent++;
   ccf=CurrentFrame;
   for(CurrentFrame=1,k=1;CurrentFrame<=Nframes;CurrentFrame+=interpolate,k++){
     Np=FirstNp; while(Np!=NULL){
       if(Np->type == CAMERA && Np == SelectedCamera){
         GetTransform(0,CurrentFrame,1.0,
                      Np,Op,p,tp,&phi,&theta,&alpha,
                      &sx,&sy,&sz,&im,&ima);
         fprintf(dxfp,format3,
                              (double)p[0]/(double)UNIT,
                              (double)p[2]/(double)UNIT,
                             -(double)p[1]/(double)UNIT);
         if(k%4 == 0){
           fprintf(dxfp,",\n");
           Id(dxfp);
         }
         else fprintf(dxfp,", ");
         break;
       }
       Np=Np->next;
     }
   }
   indent--;
   fprintf(dxfp,"\n");
   Id(dxfp); fprintf(dxfp,"]\n"); // end keyValue
   indent--;
   Id(dxfp); fprintf(dxfp,"}\n");  // position interpolator
   CurrentFrame=ccf;

   Id(dxfp); fprintf(dxfp,
     "DEF ROT-TWEEN-%s OrientationInterpolator{\n",Np->actorname); indent++;
   Id(dxfp); fprintf(dxfp,"key ["); indent++;
   for(i=1,k=1;i<=Nframes;i+=interpolate,k++){
     fprintf(dxfp,"%.3lf",(double)(i)/(double)(Nframes));
     if(k%10 == 0){
       fprintf(dxfp,",\n");
       Id(dxfp);
     }
     else fprintf(dxfp,", ");
   }
   indent--;
   fprintf(dxfp,"\n");
   Id(dxfp); fprintf(dxfp,"]\n"); // end key
   Id(dxfp); fprintf(dxfp,"keyValue ["); indent++;
   ccf=CurrentFrame;
   for(CurrentFrame=1,k=1;CurrentFrame<=Nframes;CurrentFrame+=interpolate,k++){
     Np=FirstNp; while(Np!=NULL){
       if(Np->type == CAMERA && Np == SelectedCamera){
         GetTransform(0,CurrentFrame,1.0,
                      Np,Op,p,tp,&phi,&theta,&alpha,
                      &sx,&sy,&sz,&im,&ima);
         TransformOrientation(phi,-theta,alpha,&ux,&uy,&uz,&ua);
         fprintf(dxfp,format4,ux,uz,-uy,ua);
         if(k%3 == 0){
           fprintf(dxfp,",\n");
           Id(dxfp);
         }
         else fprintf(dxfp,", ");
         break;
       }
       Np=Np->next;
     }
   }
   indent--;
   fprintf(dxfp,"\n");
   Id(dxfp); fprintf(dxfp,"]\n"); // end keyValue
   indent--;
   Id(dxfp); fprintf(dxfp,"}\n");  // orientation interpolator
   CurrentFrame=ccf;

// Put route statement here

 }
 // Models
 Np=FirstNp; while(Np!=NULL){
   if(Np->type == SKY){
     if((Sp=Np->fsky) != NULL)while(Sp != NULL){
       if(CurrentFrame >= Sp->firstframe && CurrentFrame <= Sp->lastframe){
         sr=Sp->colour[0];
         sg=Sp->colour[1];
         sb=Sp->colour[2];
         zr=Sp->zcolour[0];
         zg=Sp->zcolour[1];
         zb=Sp->zcolour[2];
         bSky=TRUE;
         break;
       }
       Sp=Sp->next;
     }
   }
   else{
     if((Op=Np->fobj) != NULL)while(Op != NULL){
       if(CurrentFrame >= Op->firstframe && CurrentFrame <= Op->lastframe){
         if(Op->type == LIGHT){
           sr=Op->colour[0]; sg=Op->colour[1]; sb=Op->colour[2];
           GetTransform(0,CurrentFrame,1.0,
                        Np,Op,p,tp,&phi,&theta,&alpha,
                        &sx,&sy,&sz,&im,&ima);
           if(Op->lighttype >= SPOTLIGHT){
           ;}
         }
         else if(Op->type == GROUND){
           gr=Op->colour[0];  gg=Op->colour[1];  gb=Op->colour[2];
           refl = Op->grefl;
           bGround=TRUE;
         }
         else if(Op->type == NORMAL || Op->type == ANIMOBJ){
           strcpy(tempstr,FileOnly(Op->name));
           ReplaceExt(tempstr);
           strcpy(tempstr1,FileOnly(Op->name));
           DeleteExt(tempstr1);
           GetTransform(0,CurrentFrame,1.0,
                        Np,Op,p,tp,&phi,&theta,&alpha,
                        &sx,&sy,&sz,&im,&ima);
           Id(dxfp); fprintf(dxfp,"DEF SFX-%s Transform {\n",tempstr1);
           indent++;
           Id(dxfp); fprintf(dxfp,"translation ");
           Id(dxfp); fprintf(dxfp,format3,
                              (double)p[0]/(double)UNIT,
                              (double)p[2]/(double)UNIT,
                             -(double)p[1]/(double)UNIT);
           Id(dxfp); fprintf(dxfp,"\n");
           //Id(dxfp); fprintf(dxfp,"# phi=%lf theta=%lf alpha=%lf\n",phi,theta,alpha);
           Id(dxfp); fprintf(dxfp,"rotation ");
           TransformOrientation(phi,-theta,alpha,&ux,&uy,&uz,&ua);
           Id(dxfp); fprintf(dxfp,format4,ux,uz,-uy,ua);
           Id(dxfp); fprintf(dxfp,"\n");
           Id(dxfp); fprintf(dxfp,"scale ");
           Id(dxfp); fprintf(dxfp,format3,sx,sz,sy);
           Id(dxfp); fprintf(dxfp,"\n");
           Id(dxfp); fprintf(dxfp,"children [\n"); indent++;
           Id(dxfp); fprintf(dxfp,"Inline {");
           if(bPrefix)fprintf(dxfp,"url \"%s",map_path);
           else       fprintf(dxfp,"url \"");
           fprintf(dxfp,"%s\"}\n",tempstr);
           indent--;
           Id(dxfp); fprintf(dxfp,"]\n");
           indent--;
           Id(dxfp); fprintf(dxfp,"}\n");
         }
         break;
       }
       Op=Op->next;
     }
   }
   Np=Np->next;
 }
 if(1 || bGround || bSky){
   Id(dxfp); fprintf(dxfp,"DEF BACKGROUND Background {\n");
   indent++;
   if(bSky){
     Id(dxfp); fprintf(dxfp,"skyAngle 1.57\n");
     Id(dxfp); fprintf(dxfp,"skyColor [");
     fprintf(dxfp,format3,(double)sr/255.0,(double)sg/255.0,(double)sb/255.0);
     fprintf(dxfp,", ");
     fprintf(dxfp,format3,(double)zr/255.0,(double)zg/255.0,(double)zb/255.0);
     fprintf(dxfp,"]\n");
   }
   if(bGround){
     Id(dxfp); fprintf(dxfp,"groundAngle 1.57\n");
     Id(dxfp); fprintf(dxfp,"groundColor [");
     fprintf(dxfp,format3,(double)gr/255.0,(double)gg/255.0,(double)gb/255.0);
     fprintf(dxfp,", ");
     fprintf(dxfp,format3,(double)gr/255.0,(double)gg/255.0,(double)gb/255.0);
     fprintf(dxfp,"]\n");
   }
   if(back_image[0] != '\0')  {Id(dxfp); fprintf(dxfp,"backUrl   [\"%s\"] \n",back_image);  }
   if(front_image[0] != '\0') {Id(dxfp); fprintf(dxfp,"frontUrl  [\"%s\"] \n",front_image); }
   if(left_image[0] != '\0')  {Id(dxfp); fprintf(dxfp,"leftUrl   [\"%s\"] \n",left_image);  }
   if(right_image[0] != '\0') {Id(dxfp); fprintf(dxfp,"rightUrl  [\"%s\"] \n",right_image); }
   if(top_image[0] != '\0')   {Id(dxfp); fprintf(dxfp,"topUrl    [\"%s\"] \n",top_image);   }
   if(bottom_image[0] != '\0'){Id(dxfp); fprintf(dxfp,"bottomUrl [\"%s\"] \n",bottom_image);}
   indent--;
   Id(dxfp); fprintf(dxfp,"}\n"); // Background
 }
 indent--;
 Id(dxfp); fprintf(dxfp,"]\n"); // children
 indent--;
 Id(dxfp); fprintf(dxfp,"}\n"); // group
 return;
}

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hThisInstance=hDLL;
      break;
    }
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return TRUE;
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static unsigned CALLBACK NameHookProc(HWND hwnd,UINT msg,
                              WPARAM wparam,LPARAM lparam){
 BOOL err;
 switch( msg ) {
   case WM_INITDIALOG:
     if(bPrefix)SendDlgItemMessage(hwnd,DLG_VRML_USEPREFIX,BM_SETCHECK,1,0);
     if(bIndent)SendDlgItemMessage(hwnd,DLG_VRML_INDENTATION,BM_SETCHECK,1,0);
     if(bViewpoints)SendDlgItemMessage(hwnd,DLG_VRML_KEYFRAME,BM_SETCHECK,1,0);
     if(bInterpolate)SendDlgItemMessage(hwnd,DLG_VRML_KEYINTER,BM_SETCHECK,1,0);
     SetDlgItemInt(hwnd,DLG_VRML_DIGITS,digits,FALSE);
     SetDlgItemInt(hwnd,DLG_VRML_KEYINTERVAL,interpolate,FALSE);
     SetDlgItemText(hwnd,DLG_VRML_PREFIX,map_path);
     SetDlgItemText(hwnd,DLG_VRML_IMAGE_TOP,top_image);
     SetDlgItemText(hwnd,DLG_VRML_IMAGE_BOTTOM,bottom_image);
     SetDlgItemText(hwnd,DLG_VRML_IMAGE_LEFT,left_image);
     SetDlgItemText(hwnd,DLG_VRML_IMAGE_RIGHT,right_image);
     SetDlgItemText(hwnd,DLG_VRML_IMAGE_FRONT,front_image);
     SetDlgItemText(hwnd,DLG_VRML_IMAGE_BACK,back_image);
//     SetWindowPos(GetParent(hwnd),HWND_TOP,50,5,0,0,SWP_NOSIZE);
//     CentreDlgOnS(GetParent(hwnd));
     return FALSE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDOK:{
          goto JUMPHERE;
         }
         break;
       case  DLG_VRML_HELP:{
            char lname[256],res_str1[256],res_str2[256];
            GetModuleFileName(hThisInstance,lname,225);
            *FileInPath(lname) = '\0';
            strcat(lname,"help\\");
            strcpy(res_str1,lname);
#if !__HELP_WS__
            strcat(res_str1,"ani_help.chm");
            ShellExecute(hwnd,NULL,res_str1,
                         NULL,lname,SW_SHOW);
#else
            {
             HH_WINTYPE hhwin;
             memset(&hhwin,0,sizeof(HH_WINTYPE));
             hhwin.cbStruct=sizeof(HH_WINTYPE);
             hhwin.nShowState=SW_SHOW;
             hhwin.pszType="MyHelp";
             LoadString(hThisInstance,IDX_VRML_STRING2,lname,255);
             hhwin.pszCaption=lname;
             hhwin.fsValidMembers = HHWIN_PARAM_SHOWSTATE |
                                    HHWIN_PARAM_PROPERTIES|
                                    HHWIN_PARAM_CUR_TAB|
                                    HHWIN_PARAM_EXPANSION|
                                    HHWIN_PARAM_RECT;
             hhwin.rcWindowPos.top=0;
             hhwin.rcWindowPos.left=0;
             hhwin.rcWindowPos.right=500;
             hhwin.rcWindowPos.bottom=500;
             hhwin.nShowState=SW_SHOW;
             strcat(res_str1,"ani_help.chm");
             hhwin.fNotExpanded = TRUE;
             hhwin.fsWinProperties = HHWIN_PROP_TRI_PANE|HHWIN_PROP_TAB_SEARCH;
             hhwin.curNavType=HHWIN_NAVTYPE_TOC;
             strcpy(res_str2,res_str1);
             strcat(res_str2,"::toc.hhc");
             hhwin.pszToc=res_str2;
             strcat(res_str1,"::/vrml_exporter.htm");
             HtmlHelp(hwnd,NULL,HH_SET_WIN_TYPE,(DWORD)&hhwin);
             HtmlHelp(hwnd,">MyHelp",HH_DISPLAY_TOPIC,(DWORD)res_str1);
           }
#endif
         }
         break;
       default: break;
     }
     break;
   case WM_NOTIFY:{
       LPOFNOTIFY lp=(LPOFNOTIFY)lparam;
       if(lp->hdr.code == CDN_FILEOK){
         JUMPHERE:
         if(SendDlgItemMessage(hwnd,DLG_VRML_USEPREFIX,BM_GETCHECK,0,0))
              bPrefix=TRUE;
         else bPrefix=FALSE;
         if(SendDlgItemMessage(hwnd,DLG_VRML_INDENTATION,BM_GETCHECK,0,0))
              bIndent=TRUE;
         else bIndent=FALSE;
         if(SendDlgItemMessage(hwnd,DLG_VRML_KEYFRAME,BM_GETCHECK,0,0))
              bViewpoints=TRUE;
         else bViewpoints=FALSE;
         if(SendDlgItemMessage(hwnd,DLG_VRML_KEYINTER,BM_GETCHECK,0,0))
              bInterpolate=TRUE;
         else bInterpolate=FALSE;
         digits=GetDlgItemInt(hwnd,DLG_VRML_DIGITS,&err,FALSE);
         if(digits < 0)digits=0;
         interpolate=GetDlgItemInt(hwnd,DLG_VRML_KEYINTERVAL,&err,FALSE);
         if(digits < 1)digits=1;
         GetDlgItemText(hwnd,DLG_VRML_PREFIX,map_path,255);
         GetDlgItemText(hwnd,DLG_VRML_IMAGE_BACK,back_image,255);
         GetDlgItemText(hwnd,DLG_VRML_IMAGE_FRONT,front_image,255);
         GetDlgItemText(hwnd,DLG_VRML_IMAGE_LEFT,left_image,255);
         GetDlgItemText(hwnd,DLG_VRML_IMAGE_RIGHT,right_image,255);
         GetDlgItemText(hwnd,DLG_VRML_IMAGE_TOP,top_image,255);
         GetDlgItemText(hwnd,DLG_VRML_IMAGE_BOTTOM,bottom_image,255);
       }
     }
     break;
   default: break;
 }
 return FALSE;
}

static BOOL LocalSelectFileName(char *szfile, char *szfilter,
                           HWND parent, HINSTANCE hInst){
 int i;
 OPENFILENAME ofn;
 char szFilter[80],title[256];
 DWORD version;
 version=GetVersion();
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=0;
 ofn.lpstrFile=szfile;
 ofn.nMaxFile=255;
 ofn.lpstrFileTitle=NULL;
 ofn.nMaxFileTitle=0;
 ofn.lpstrInitialDir=NULL;
 LoadString(hThisInstance,IDX_VRML_STRING1,title,255);
 ofn.lpstrTitle=title;
 ofn.hInstance=hInst;
 ofn.lpfnHook=NameHookProc;
 ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|
           OFN_ENABLEHOOK|OFN_ENABLETEMPLATE|OFN_NONETWORKBUTTON;
//version=3;
 if(LOBYTE(LOWORD(version)) >= 4){
   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_VRML);
   ofn.Flags |= OFN_EXPLORER;
 }
 else{
   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_OLDVRML);
 }

 return GetOpenFileName(&ofn);
}

long _Export(HWND parent_window, char *filename, ANI_STRUCTURE *lpevi){
 char localfile[256];
 FILE *dxfp;
 lpEVI=lpevi;
 ReadDataFromRegistry("SOFTWARE\\OPENFX\\VRML");
 strcpy(localfile,globalfile);
 if(!LocalSelectFileName(localfile,
                         "VRML files (*.WRL)|*.wrl",
                          parent_window,hThisInstance))return 1;
 strcpy(globalfile,localfile);
 WriteDataToRegistry("SOFTWARE\\OPENFX\\VRML");
 AppendFileExtension(localfile,".wrl");
 if((dxfp=fopen(localfile,"w")) != NULL){
   fprintf(dxfp,"#VRML V2.0 utf8\n\n# Animation Script Produced by OpenFX Version 1.0\n\n");
   CreateVRMLfile(dxfp);
   fclose(dxfp);
 }
 else{
   MessageBox(parent_window,"File open Error",NULL,MB_OK);
 }
 return 1;
}

