/* file WRITEAVI.c                                    */

/* Code to create an AVI file (Open write each frame  */
/* etc). given a pointer to a bitmap structure        */

/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */


#ifndef STRICT
#define STRICT
#endif

#define  INC_OLE2
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <vfw.h>
#include <stdlib.h>

#define AVIIF_KEYFRAME  0x00000010L // this frame is a key frame.

void configureAVIfile(HWND, char *);
void openAVIfile(char *Name, LPBITMAPINFOHEADER alpbi, int compress);
void appendAVIfile(LPBITMAPINFOHEADER alpbi);
void closeAVIfile(void);

int AVI_frame_rate=15;

static int fci=0;
static PAVIFILE pfile = NULL;
static PAVISTREAM ps = NULL, psCompressed = NULL;
static BOOL compress_flag=TRUE, options=FALSE;
static AVICOMPRESSOPTIONS  aviopt;

static PSTR AVIFileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

void configureAVIfile(HWND hWnd, char *preview){
 char filename[256];
 PAVIFILE pfile = NULL;
 PAVISTREAM ps = NULL;
 LPAVICOMPRESSOPTIONS lpaviopt;
 HRESULT hr;
 if(preview != NULL)strcpy(filename,preview);
 else{
   GetModuleFileName(NULL,filename,256);
   strcpy(AVIFileInPath(filename),"blank.avi");
 }
 AVIFileInit();
 hr = AVIFileOpen(&pfile,filename,OF_READ,NULL);
 if(hr == AVIERR_OK){
   int i;
   AVISTREAMINFO avis;
   for(i=0;i<10;i++){
     if(AVIFileGetStream(pfile,&ps,0,i) == AVIERR_OK){
       AVIStreamInfo(ps,&avis,sizeof(avis));
       if(avis.fccType == streamtypeVIDEO){
         lpaviopt=&aviopt;
         if(AVISaveOptions(hWnd,
               ICMF_CHOOSE_KEYFRAME|ICMF_CHOOSE_DATARATE|ICMF_CHOOSE_PREVIEW,
               1,&ps,&lpaviopt) == TRUE)options=TRUE;
         AVISaveOptionsFree(1,&lpaviopt);
         break;
       }
     }
   }
   if(pfile != NULL)AVIFileRelease(pfile);
 }
 AVIFileExit();
 return;
}

void openAVIfile(char *Name, LPBITMAPINFOHEADER alpbi, int compress){
 int i;
 AVISTREAMINFO strhdr;
 AVICOMPRESSOPTIONS opts;
 HRESULT hr;
 WORD wVer;
 fci=0;
 ps = NULL; psCompressed = NULL; pfile=NULL;
 wVer = HIWORD(VideoForWindowsVersion());
 if (wVer < 0x010a){
   MessageBeep(MB_ICONHAND);
   MessageBox(NULL, "Version",NULL,MB_OK|MB_ICONSTOP);
   return;
 }
 AVIFileInit();
 hr = AVIFileOpen(&pfile,             // returned file pointer
          Name,
          OF_WRITE | OF_CREATE,       // mode to open file with
          NULL);                      // use handler determined
                                      // from file extension....
 if (hr != AVIERR_OK)goto error;
 _fmemset(&strhdr, 0, sizeof(strhdr));
 strhdr.fccType                = streamtypeVIDEO;// stream type
 strhdr.fccHandler             = 0;
 strhdr.dwScale                = 1;
 strhdr.dwRate                 = AVI_frame_rate;
 strhdr.dwSuggestedBufferSize  = alpbi->biSizeImage;
 SetRect(&strhdr.rcFrame, 0, 0,              // rectangle for stream
     (int)alpbi->biWidth,
     (int)alpbi->biHeight);
 hr = AVIFileCreateStream(pfile,             // file pointer
        &ps,                                 // returned stream pointer
        &strhdr);                            // stream header
 if (hr != AVIERR_OK)goto error;
 _fmemset(&opts, 0, sizeof(opts));
// opts.dwFlags = AVICOMPRESSF_VALID | AVICOMPRESSF_KEYFRAMES;
// opts.fccHandler = mmioFOURCC('C', 'V', 'I', 'D');
 opts.dwFlags = AVICOMPRESSF_VALID;
 opts.fccType = streamtypeVIDEO;
 opts.fccHandler = mmioFOURCC('M', 'S', 'V', 'C');
 opts.dwQuality = 6500;
 opts.dwKeyFrameEvery = 10;
 if(options){
   compress=TRUE;
   memcpy(&opts,&aviopt,sizeof(AVICOMPRESSOPTIONS));
 }
 if(compress){
   compress_flag=TRUE;
   hr = AVIMakeCompressedStream(&psCompressed, ps, &opts, NULL);
   if (hr != AVIERR_OK)goto error;
 }
 else{
   compress_flag=FALSE;
   psCompressed=ps;
 }
 hr = AVIStreamSetFormat(psCompressed, 0,
             alpbi,            // stream format
             alpbi->biSize +   // format size
             alpbi->biClrUsed * sizeof(RGBQUAD));
 if (hr != AVIERR_OK)goto error;
 return;
 error:
 closeAVIfile();
}

void appendAVIfile(LPBITMAPINFOHEADER alpbi){
 HRESULT hr;
 if(!ps || !psCompressed || !pfile)return;
 hr = AVIStreamWrite(psCompressed, // stream pointer
  fci,                             // time of this frame
  1,                               // number to write
  (LPBYTE) alpbi +                 // pointer to data
   alpbi->biSize +
   alpbi->biClrUsed * sizeof(RGBQUAD),
   alpbi->biSizeImage,            // size of this frame
   AVIIF_KEYFRAME,                // flags....
   NULL,
   NULL);
 if (hr != AVIERR_OK)closeAVIfile();
 fci++;
}

void closeAVIfile(void){
 if (ps)AVIStreamClose(ps); ps=NULL;
 if(compress_flag){
   if(psCompressed)AVIStreamClose(psCompressed); psCompressed=NULL;
 }
 if (pfile)AVIFileClose(pfile); pfile=NULL;
 AVIFileExit();
 return;
}

