/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file RSTRUCT.H   include file for Renderer external structure */
/*                  pointers to callback functions               */

/*****************************************************************\
   The pointers in this structure are pointers to functions
   in the Renderer that can be called from an external shader.
   The do not have to be used but a pointer to this structure
   is passed to the shader function so this header file must
   be included.
   Use the definitions in the file DEFINES.H to provide easy
   to use references to the functions. See the example shaders
   on how to use them.
   The functions provide standard Perlin type random variables
   that can be used in developing fractal type textures.
   normalize() will scale a vector to have unit length.
   Some functions are more useful than others the one that are
   useful are -:
     normalize(n);          normalize vector "n" to unit length
     sNoise(x,y,z,r);       returns (r) a Perlin random variable [0-1]
                            given an (x,y,z) co-ordinate
     vNoise(v,x,y,z);       returns a normalized Perlin random vector
                            given an (x,y,z) co-ordinate
     sTurbulence(x,y,z,r);  returns (r) a turbulence value given co-ordinate
     vTurbulence(v,x,y,z);  returns a turbulence vector given a co-ordinate
\*****************************************************************/

typedef struct tagX__SHADER {
 void   (*fp_sNoise)(double x, double y, double z, double *result);
 void   (*fp_sTurbulence)(double x, double y, double z, double *result);
 void   (*fp_Cycloidal)(double value, double *result);
 void   (*fp_vNoise)(vector result, double x, double y, double z);
 void   (*fp_vTurbulence)(vector result, double x, double y, double z);
 short  (*fp_normalize)(vector n);
 vector *fp_Wave_Sources;
 long   *dummyf[32];
 long   version;
 vector *fp_axis_o,*fp_axis_u,*fp_axis_v,*fp_axis_n;
 vector *fp_axis_unorm,*fp_axis_vnorm;
 double *u_length,*v_length;
 double *tr,*itr;     // pointers to 4 x 4 arrays
 long   *dummyv[32];
} X__SHADER;

