/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file 3dscon.c */

#include <windows.h>
#include <commctrl.h>

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <io.h>
#include <string.h>
#include <ctype.h>
#include <sys\types.h>
#include <sys\stat.h>

#include "struct.h"
#include "dstruct.h"
#include "3dscon.h"

int  (*X__pri)(const char *,...);
int  X__local_print(const char *,...);

HWND      dlg=NULL;
HWND      hWndParent;
HWND      hWndProgress=NULL;
HINSTANCE hInstance;
long      file_size=1;

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BCC__
int APIENTRY DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
      hInstance=hDLL;
      break;
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     hWndProgress=GetDlgItem(hwnd,DLG_PROGRESS);
     SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,99));
     SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
     SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
     CentreDlgOnS(hwnd);
     break;
   default: break;
 }
 return( FALSE );
}

BOOL CALLBACK DlgProcScale(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[32];
 static struct SC {double scale,xc,yc,zc,c;} *Sc;
 switch( msg ) {
   case WM_INITDIALOG:
     Sc = (struct SC *)lparam;
     sprintf(str,"%.3le",Sc->scale);
     SetDlgItemText(hwnd,DLG_SCALE_SCALE,(LPCTSTR)str);
     sprintf(str,"%.1le",Sc->xc);
     SetDlgItemText(hwnd,DLG_SCALE_XCEN,(LPCTSTR)str);
     sprintf(str,"%.1le",Sc->yc);
     SetDlgItemText(hwnd,DLG_SCALE_YCEN,(LPCTSTR)str);
     sprintf(str,"%.1le",Sc->zc);
     SetDlgItemText(hwnd,DLG_SCALE_ZCEN,(LPCTSTR)str);
     if(Sc->c == 1)SendDlgItemMessage(hwnd,DLG_SCALE_CEN,BM_SETCHECK,TRUE,0);
     CentreDlgOnS(hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(SendDlgItemMessage(hwnd,DLG_SCALE_CEN,BM_GETCHECK,0,0))Sc->c=1;
          else Sc->c=0;
          if(GetDlgItemText(hwnd,DLG_SCALE_SCALE,str,31) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          sscanf(str,"%lf",&(Sc->scale));
          if(Sc->scale < 1.e-10)Sc->scale=1.e-10;
          if(GetDlgItemText(hwnd,DLG_SCALE_XCEN,str,31) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          sscanf(str,"%lf",&(Sc->xc));
          if(GetDlgItemText(hwnd,DLG_SCALE_YCEN,str,31) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          sscanf(str,"%lf",&(Sc->yc));
          if(GetDlgItemText(hwnd,DLG_SCALE_ZCEN,str,31) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          sscanf(str,"%lf",&(Sc->zc));
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default: break;
      }
   default: break;
 }
 return( FALSE );
}

void UpdateProgress(FILE *dxfp){
 long p;
 if(hWndProgress == NULL)return;
 p=ftell(dxfp);
 p=(p*100)/file_size;
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)p,0);
 Sleep(0);
 return;
}

long _Convert(
     HWND parent,
     char *iname,
     char *oname,
     X__STRUCTURE *lpevi
                              ){
 char *s[32];
 long i,r;
 int fhandle;
 lpEVI=lpevi;
 hWndParent=parent;
 if((fhandle=open(iname,O_RDONLY)) >= 0){
   struct stat buffer;
   if(fstat(fhandle,&buffer) == 0){
     file_size=buffer.st_size;
   }
   else MessageBox(NULL,"File size",NULL,MB_OK);
   close(fhandle);
 }
 else MessageBox(NULL,"Open for size",NULL,MB_OK);

 X__pri=X__local_print;
 for(i=1;i<3;i++)s[i]=X__Malloc(256);
 strcpy(s[1],iname);
 strcpy(s[2],oname);
 dlg=CreateDialog(hInstance,MAKEINTRESOURCE(DLG),parent,(DLGPROC)DlgProc);
 r=(long)Do3dsConvert(3,s);
 if(dlg != NULL)DestroyWindow(dlg);
 for(i=1;i<3;i++)X__Free(s[i]);
 return r;
}

int X__local_print(const char *format,...){
 char line[80];
 va_list arg_ptr;
 va_start(arg_ptr,format);
 vsprintf(line,format,arg_ptr);
 va_end(arg_ptr);
// WindowedRenderString(line);
 return 1;
}
