/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* 3DSCON1.C */

#define MODULE_SAPC

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <windows.h>

#undef OPAQUE

#include "3dscon.h"

extern HWND      dlg;
extern HWND      hWndParent;
extern HINSTANCE hInstance;

extern BOOL CALLBACK DlgProcScale(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);
extern BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

extern void UpdateProgress(FILE *);
extern int (*X__pri)(const char *,...);

#ifdef __ZTC__
  #ifdef DOS386
    #define _X__FAR
  #else
    #define _X__FAR _far
  #endif
#endif

#define DOS386
#define _X__FAR
#define PI  3.1415926

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

typedef double vector[3];
typedef long   point[3];
typedef long   twoLongs[2];

typedef struct MAPGLUE {
 float x,y;
} mapglue;

typedef struct WIREFRAME {
 long Np,Ne;
 twoLongs *e;
 point *p;
} wireframe;

typedef struct SKELETON {
 struct SKELETON _X__FAR *at,*next,*last;
 short  id,offset;
 vector u,v,w;
 double t[4][4];
 point  xyz,tp,bx[8];
 float  p[3];
 char   name[16];
} skel;

typedef struct VERTEX {
 struct VERTEX    *last;
 struct VERTEX    *next;
 struct VERTEX    **adj;
 struct SKELETON  *sp;
 struct MAPGLUE   *gp;
 char   status,pad;
 unsigned short id,n;
 float  p[3];
 long   xyz[3];
} vertex;

typedef struct EDGE {
 struct EDGE _X__FAR *last;
 struct EDGE _X__FAR *next;
 struct VERTEX _X__FAR *V[2];
} edge;


typedef struct FACE{
 struct FACE _X__FAR *last;
 struct FACE _X__FAR *next;
 struct VERTEX _X__FAR *V[3];
 unsigned char color[3],matcol[3],texture,brush,axis;
} face;

typedef struct {
       vertex Origin;
       long Xdirection[3];
       long Ydirection[3];
       long Zdirection[3];
       vertex Offset;
} axis;

typedef struct {
  char  name[8];
  short Nc,Np,flag;
  vertex **Vc,**Vp;
} bezier_curve;

typedef struct tagIMAGEMAP {
short   in_use;
char    N[32],*S,*R,*B;
char    s,r,b;
short   sp,rp,bp;
char    Type,Map,Lock;
short   AnimFF,AnimDL,AnimLF,Angle,SkId;
point   P,X,Y;
skel    *Sk;
vertex  *Vlock[3];
unsigned char d_colour[3],s_colour[3],a_colour[3];
} IMAGEMAP;

typedef struct tagSHADERAXIS {
char    N[32];
point   P,X,Y;
} SHADERAXIS;

typedef struct tagXSHADER {
  long in_use;
  char name[128],*parameters;
} XSHADER;

short silent_3ds=0;

#define PLANE            0              /* brush map types */
#define CYLINDER         1
#define PLANE_MOZIAC     2
#define CYLINDER_MOZIAC  3
#define MAP_BY_VERTEX    4
#define OPAQUE           0              /* brush color 0   */
#define CLEAR            1
#define GLOSS            0              /* brush shiny     */
#define MATT             1
#define INEDITOR         6

#define TRIVIEW   0
#define FRONTVIEW 1
#define TOPVIEW   2
#define RIGHTVIEW 3
#define PERSPVIEW 4
#define BIGWIN    0
#define TRITOP    1
#define TRIFRONT  2
#define TRIRIGHT  3

/* definition of unit dimensions */
#define UNIT         32768L
#define UNIT2        65536L
#define MINUNIT       2048L     // was 1024 trying larger to prevent error
#define MAXUNIT    4194304L     //1073741824L
#define PI2            PI/2
/* definition of flags and identifiers*/
#define DESELECTED       0
#define SELECTED         1
#define HIDDEN           2
#define NO               0
#define YES              1
#define OK               1   // used to be 0
#define FAIL            -1   // used to be 1

#define STATICBRUSH      0
#define ANIMBRUSH        1
#define BUMPBRUSH        2
#define REFLBRUSH        3
#define PSETBRUSH        4

#define G_FILEOPENFAIL 1000

unsigned short _cdecl ERRORFLAG= 1,
               LOADOPERATION=0,
               Xres=640,Yres=480,Xres1=639,Yres1=479,
               StringToggle=0,CursorToggle=0,HelpBarToggle=0,Hires1=0,
               UNDO_ON=YES,EDIT_ACTION=NO,
               zoom_redraw=NO,zoom_abort=NO,
               WindowBox_view=0,icon_arrow_ids=1;
char   *WorkDrive;                      /* from TEMP environment */
char   ruler_name[8]={"None "};
double ruler=(double)UNIT;
long   rulerx=0,rulery=0,rulerz=0,    /* relative */
       lrulerx=0,lrulery=0,lrulerz=0, /* origin of ruler */
       draw_grid_on=0,grid_on=0,grid_size=UNIT;
long NpointerX=0,NpointerY=0,NpointerZ=0
    ,TVpointX=(-UNIT),TVpointY=(-UNIT),TVpointZ=(-UNIT)
    ,TVsizeX=UNIT2,TVsizeY=UNIT2,TVsizeZ=UNIT2;

wireframe w_frame={0,0,NULL,NULL};
bezier_curve *Bcurves=NULL;

axis _X__FAR ObjectAxis={NULL,NULL,0,0,
                      0L,0L,0L,
                      0,0L,0L,
                      UNIT,0L,0L,
                      0L,UNIT,0L,
                      0L,0L,0L};

vertex _X__FAR *MainVp=NULL,*LastToolVertex=NULL;
edge   _X__FAR *MainEp=NULL;
face   _X__FAR *MainFp=NULL;
skel   _X__FAR *MainSp=NULL,*LastSkelVertex=NULL,*FirstSp=NULL,
           *current_skeleton=NULL;
unsigned long   Nvert=0,NvertSelect=0,NvertDeselect=0,NvertGlue=0,
                Nface=0,Nedge=0,Nskel=0,N_Bcurves=0;
short        SelectedBrush = -1,FontLoaded=NO,ShadowFlag=NO,MirrorFlag = -1;
short    nImaps=0,nLockediMaps=0,nImap4=0,nSha=0;
SHADERAXIS iSha[64];
IMAGEMAP iMap[64];
XSHADER  xShaders[8];
unsigned char _X__FAR LocalColour[7]={0,0,0,0,128,128,128},
              _X__FAR FaceColour[3]={255,255,255},
              _X__FAR ShaderColour[3]={128,128,128};
long DefaultBrilliance=8,DefaultTransparency=8,DefaultReflectivity=0,
     DefaultTexture=0,DefaultShiny=0,DefaultSmoothing=0;

void SaveObject(char *Filename, short status);
void WriteTextures(short );
void WriteVertices(short );
void WriteEdges(short );
void WriteFaces(short );
void WriteSkeleton(void);


#define BRUSHBIT   0x40
#define SMOOTHBIT  0x80
#define TILEBIT    0x20;
static vertex **p=NULL;
static FILE *fp,*fo;
static long newface,newedge,newvert,newglue,minX,minY,minZ,maxX,maxY,maxZ;
static char str[10];
static vertex *vlast;
static axis _X__FAR TempAxis;
static face _X__FAR *FaceBase;
void outsht(short sht){
 fputc((char)(sht >>  8),fo);
 fputc((char)(sht      ),fo);
}
void outlng(long lng){
 fputc((char)(lng >> 24),fo);
 fputc((char)(lng >> 16),fo);
 fputc((char)(lng >>  8),fo);
 fputc((char)(lng      ),fo);
}
short intriview(vertex _X__FAR *vp){return 1;}
void Output_Message(short id){;}
void disinform(void){;}
void inform(char *aa){;}
void pause(short delay){;}
void DrawOldMarker(void){;}
short RequestItemSelect(short Nitems, char **ItemList, char *Message){
 return 1;
}


/* uncomment to allow saving of mapping co-ords */
#define __RSF__
#include "3dssave.c"

static void mes3ds(char *s){
 if(silent_3ds)return;
 X__pri("%s",s);
}

void CreateEdge1(vertex _X__FAR *v1, vertex _X__FAR *v2, edge *ep){
 edge _X__FAR *temp;
 if(( temp=(edge _X__FAR *)malloc(sizeof(edge)) ) == NULL){
   mes3ds("out of memory\n");
   return;
 }
 if(MainEp != NULL)MainEp->next=temp;
 temp->last=MainEp;
 temp->V[0]=v1;
 temp->V[1]=v2;
 temp->next=NULL;
 MainEp=temp;
 Nedge++;
}

void CreateFace(vertex _X__FAR *v1, vertex _X__FAR *v2,
                vertex _X__FAR *v3, short check){
 face _X__FAR *temp;
 if(check){
 temp=MainFp;  /* check to see if face exists */
 while(temp != NULL)   /* check to see if this face already exists */
 {if(   ((temp->V[0] == v1) && (temp->V[1] == v2) && (temp->V[2] == v3))
     || ((temp->V[0] == v1) && (temp->V[1] == v3) && (temp->V[2] == v2))
     || ((temp->V[0] == v2) && (temp->V[1] == v3) && (temp->V[2] == v1))
     || ((temp->V[0] == v2) && (temp->V[1] == v1) && (temp->V[2] == v3))
     || ((temp->V[0] == v3) && (temp->V[1] == v1) && (temp->V[2] == v2))
     || ((temp->V[0] == v3) && (temp->V[1] == v2) && (temp->V[2] == v1))
  )return;
 temp=temp->last;}}
 if(( temp=(face _X__FAR *)malloc(sizeof(face)) ) == NULL){
   mes3ds("out of memory\n");
   return;
 }
 if(MainFp != NULL)MainFp->next=temp;
 temp->last=MainFp;
 temp->V[0]=v1;
 temp->V[1]=v2;
 temp->V[2]=v3;
 temp->next=NULL;
 temp->color[0]=FaceColour[0];
 temp->color[1]=FaceColour[1];
 temp->color[2]=FaceColour[2];
 temp->matcol[0]=(((unsigned char)(DefaultBrilliance) << 4)
                + (ShaderColour[0] >> 4) );
 temp->matcol[1]=(((unsigned char)(DefaultTransparency) << 4)
                + (ShaderColour[1] >> 4) );
 temp->matcol[2]=(((unsigned char)(DefaultReflectivity) << 4)
                + (ShaderColour[2] >> 4) );
 temp->texture=DefaultTexture;
 if(DefaultShiny)temp->texture     |= 0x20;  /* set bit #5 */
 if(DefaultSmoothing)temp->texture |= 0x80;  /* set smoothing bit */
 temp->brush=0;
 temp->axis=0;
 MainFp=temp;
 Nface++;
}
void CreateVertex(void){
 vertex _X__FAR *temp;
 if(( temp=(vertex _X__FAR *)malloc(sizeof(vertex)) ) == NULL){
   mes3ds("out of memory\n");
   return;
 }
 if(MainVp != NULL)MainVp->next=temp;
 temp->last=MainVp;
 temp->xyz[0]=NpointerX;
 temp->xyz[1]=NpointerY;
 temp->xyz[2]=NpointerZ;
 temp->next=NULL;
 temp->status=SELECTED;
 temp->sp=NULL;
 temp->gp=NULL;
 temp->adj=NULL;
 temp->n=0;
 temp->p[0]=0.0;
 temp->p[1]=0.0;
 temp->p[2]=0.0;
 MainVp=temp;
 Nvert++;
 NvertSelect++;
}
void CreateSkeleton(skel *at){
 skel _X__FAR *temp;
 int i,j;
 if(( temp=(skel _X__FAR *)malloc(sizeof(skel)) ) == NULL){
   mes3ds("out of memory\n");
   return;
 }
 if(MainSp != NULL)MainSp->next=temp;
 temp->last=MainSp;
 temp->xyz[0]=NpointerX;
 temp->xyz[1]=NpointerY;
 temp->xyz[2]=NpointerZ;
 temp->p[0]=temp->p[1]=temp->p[2]=0.0;
 temp->next=NULL;
 temp->at=at;
 temp->id=0;
 temp->offset=0;
 for(i=0;i<4;i++)
  for(j=0;j<4;j++){
    temp->t[i][j]=0.0;
    if(i == j)temp->t[i][j]=1.0;
  }
 MainSp=temp;
 Nskel++;
 if(Nskel == 1)strcpy(temp->name,"Root");
 else sprintf(temp->name,"%ld",Nskel-1);
}
static short CreateModelerBrush(void){
  long i;
  EDIT_ACTION=YES;
  if(nImaps > 62)return FAIL;
  iMap[nImaps].s=0;    iMap[nImaps].r=0;    iMap[nImaps].b=0;
  iMap[nImaps].S=NULL; iMap[nImaps].R=NULL; iMap[nImaps].B=NULL;
  iMap[nImaps].sp=100; iMap[nImaps].rp=50; iMap[nImaps].bp=50;
  for(i=0;i<3;i++){
    iMap[nImaps].d_colour[i]=0;
    iMap[nImaps].s_colour[i]=0;
    iMap[nImaps].a_colour[i]=0;
  }
  sprintf(iMap[nImaps].N,"unnamed_%ld",nImaps+1);
  {
    iMap[nImaps].P[0] = TVpointX + TVsizeX/10;
    iMap[nImaps].P[1] = TVpointY + TVsizeY/2;
    iMap[nImaps].P[2] = TVpointZ + TVsizeZ - TVsizeZ/10;
    iMap[nImaps].X[0] = TVpointX + TVsizeX - TVsizeX/10;
    iMap[nImaps].X[1] = TVpointY + TVsizeY/2;
    iMap[nImaps].X[2] = TVpointZ + TVsizeZ - TVsizeZ/10;
    iMap[nImaps].Y[0] = TVpointX + TVsizeX/10;
    iMap[nImaps].Y[1] = TVpointY + TVsizeY/2;
    iMap[nImaps].Y[2] = TVpointZ + TVsizeZ/10;
  }
  iMap[nImaps].Type = STATICBRUSH;    /* Still Anim */
  iMap[nImaps].Angle  = 360;          /* full rotation if required */
  iMap[nImaps].AnimFF = 1;
  iMap[nImaps].AnimDL = 1;
  iMap[nImaps].AnimLF = 1;
  iMap[nImaps].SkId = -1;        /* Brush is not assigned to any skeletal */
  iMap[nImaps].Sk = NULL;        /* component */
  iMap[nImaps].Lock = 0;         /* not locked */
  iMap[nImaps].Vlock[0] = NULL;  /* lock P */
  iMap[nImaps].Vlock[1] = NULL;  /* lock X */
  iMap[nImaps].Vlock[2] = NULL;  /* lock Y */
  iMap[nImaps].Map=MAP_BY_VERTEX;  /* brush map */
//  iMap[nImaps].Map=PLANE;  /* brush map */
  nImap4++;
  nImaps++;
  return OK;
}

#define _DOS_GETDRIVE _dos_getdrive
#define _DOS_SETDRIVE _dos_setdrive

typedef struct tagMAP3DS {
  char    N[32];
  short   shiny,transp,evi;
  unsigned char d_colour[3];
} MAP3DS;

static float scale_3ds,gmax[3],gmin[3];
static long posn,nJmaps=0;
static FILE *fi;
static MAP3DS *jMap;
static unsigned drive;
static vertex *vertex_object_start=NULL;

void InsertInVertexList(vertex *vj, vertex *vi);
void BuildEdges(void);

static long GetLong(FILE *fp){
 long val;
 fread(&val,sizeof(long),1,fp);
 return val;
}

static short GetShort(FILE *fp){
 short val;
 fread(&val,sizeof(short),1,fp);
 return val;
}

static char GetByte(FILE *fp){
 char val;
 val  = ((char)getc(fp));
 return val;
}

void get_material(long count, FILE *fi){
  char c,mapname[128],imagename[32],str[128];
  unsigned char r,g,b;
  unsigned short id,sv,id1,first_time;
  long cc,p,i,l,j;
  while(1){
    p=ftell(fi);  id=GetShort(fi); cc=GetLong(fi);
    if(id < 0xa000 || id >= 0xafff)break;
    if(id == 0xa000){
      j=0; while((c=GetByte(fi)) != 0)mapname[j++]=c;
      mapname[j]=0;
      sprintf(str,"Material [%s]\n",mapname);
      mes3ds(str);
      first_time=1; nJmaps++;
      if(nJmaps == 1){
        if((jMap=(MAP3DS *)malloc(nJmaps*sizeof(MAP3DS))) == NULL){
          mes3ds("Out of memory\n"); return;
        }
      }
      else {
        if((jMap=(MAP3DS *)realloc(jMap,nJmaps*sizeof(MAP3DS))) == NULL){
          mes3ds("Out of memory\n"); return;
        }
      }
      jMap[nJmaps-1].evi = -1;
      strcpy(jMap[nJmaps-1].N,mapname);
    }
    else if(id == 0xa010){
      id=GetShort(fi); l=GetLong(fi);
      if(id != 0x11){mes3ds("Error in material\n"); break;}
      r=GetByte(fi); g=GetByte(fi); b=GetByte(fi);
    }
    else if(id == 0xa020){
      id=GetShort(fi); l=GetLong(fi);
      if(id != 0x11){mes3ds("Error in material\n"); break;}
      r=GetByte(fi); g=GetByte(fi); b=GetByte(fi);
      jMap[nJmaps-1].d_colour[0]=r;
      jMap[nJmaps-1].d_colour[1]=g;
      jMap[nJmaps-1].d_colour[2]=b;
    }
    else if(id == 0xa030){
      id=GetShort(fi); l=GetLong(fi);
      if(id != 0x11){mes3ds("Error in material\n"); break;}
      r=GetByte(fi); g=GetByte(fi); b=GetByte(fi);
    }
    else if(id == 0xa200 || id == 0xa210 || id == 0xa220 || id == 0xa230){
      if(first_time){
        CreateModelerBrush();
        strcpy(iMap[nImaps-1].N,jMap[nJmaps-1].N);
        jMap[nJmaps-1].evi=nImaps-1;
        first_time=0;
      }
      if(id == 0xa200){mes3ds("Texture map "); id1=1;}
      if(id == 0xa210){mes3ds("Opacity map "); id1=2;}
      if(id == 0xa220){mes3ds("Reflect map "); id1=3;}
      if(id == 0xa230){mes3ds("Bump    map "); id1=4;}
      id=GetShort(fi); l=GetLong(fi);
      if(id != 0x30){mes3ds("Error in material\n"); break;}
      sv=GetShort(fi);
      sprintf(str,"Map value %d ",sv);
      mes3ds(str);
      id=GetShort(fi); l=GetLong(fi);
      if(id != 0xa300){mes3ds("Error in material\n"); break;}
      j=0; while((c=GetByte(fi)) != 0)imagename[j++]=c;
      imagename[j]=0;
      sprintf(str,"Map [%s]\n",imagename);
      mes3ds(str);
      if(id1 == 1){
        iMap[nImaps-1].s=1;
        iMap[nImaps-1].sp=sv;
        iMap[nImaps-1].S=(char *)malloc(256);
        sprintf(iMap[nImaps-1].S,"%c:\\3ds2\\maps\\",drive+'@');
        strcat(iMap[nImaps-1].S,imagename);
      }
      if(id1 == 3){
        iMap[nImaps-1].r=1;
        iMap[nImaps-1].rp=sv;
        iMap[nImaps-1].R=(char *)malloc(256);
        sprintf(iMap[nImaps-1].R,"%c:\\3ds2\\maps\\",drive+'@');
        strcat(iMap[nImaps-1].R,imagename);
      }
      if(id1 == 4){
        iMap[nImaps-1].b=1;
        iMap[nImaps-1].bp=sv;
        iMap[nImaps-1].B=(char *)malloc(256);
        sprintf(iMap[nImaps-1].B,"%c:\\3ds2\\maps\\",drive+'@');
        strcat(iMap[nImaps-1].B,imagename);
      }
    }
    fseek(fi,p,SEEK_SET); for(i=0;i<cc;i++)GetByte(fi);
  }
}

void get_vertices(FILE *fi){
 float f1,f2,f3,fmin[3],fmax[3];
 unsigned short n,i;
 long l;
 n=GetShort(fi);
 if((p=(vertex **)malloc((long)n*sizeof(vertex *))) == NULL){
  mes3ds("out of memory\n");
  return;
 }
 for(i=0;i<3;i++){
   fmin[i] =   1.e30;
   fmax[i] = - 1.e30;
 }
 for(i=0;i<n;i++){
   fread(&f1,sizeof(float),1,fi);
   fread(&f2,sizeof(float),1,fi);
   fread(&f3,sizeof(float),1,fi);
   if(f1 < fmin[0])fmin[0]=f1; if(f1 > fmax[0])fmax[0]=f1;
   if(f2 < fmin[1])fmin[1]=f2; if(f2 > fmax[1])fmax[1]=f2;
   if(f3 < fmin[2])fmin[2]=f3; if(f3 > fmax[2])fmax[2]=f3;
   if(f1 < gmin[0])gmin[0]=f1; if(f1 > gmax[0])gmax[0]=f1;
   if(f2 < gmin[1])gmin[1]=f2; if(f2 > gmax[1])gmax[1]=f2;
   if(f3 < gmin[2])gmin[2]=f3; if(f3 > gmax[2])gmax[2]=f3;
   CreateVertex();
   MainVp->p[0]=f1;
   MainVp->p[1]=f2;
   MainVp->p[2]=f3;
   p[i]=MainVp;
   MainVp->sp=MainSp;
 }
 for(i=0;i<3;i++){
   MainSp->p[i]=(fmax[i]+fmin[i])/2.0;
//   MainSp->xyz[i]=(long)(scale_3ds*MainSp->p[i]);
 }
 n=GetShort(fi); l=GetShort(fi);
//mes3ds("Next packet=%x\n",n);
 UpdateProgress(fi);
}

void get_mapping_coords(FILE *fi){
 float f1,f2;
 unsigned short n,i;
 long l;
 n=GetShort(fi);
//mes3ds("%d Vertices have mapping\n",n);
 for(i=0;i<n;i++){
   fread(&f1,sizeof(float),1,fi);
   fread(&f2,sizeof(float),1,fi);
//printf("%ld\t\t %lf \t%lf\n",i,f1,f2);
   if((p[i]->gp=(mapglue _X__FAR *)malloc(sizeof(mapglue)) ) == NULL){
     mes3ds("out of memory\n");
     return;
   }
   NvertGlue++;
   p[i]->gp->x=f1;
   p[i]->gp->y=f2;
 }
 n=GetShort(fi); l=GetShort(fi);
//mes3ds("Next packet=%x\n",n);
}


long GetMap(char *mapname){
  long i;
  if(nJmaps > 0){
    for(i=0;i<nJmaps;i++){
      if(strcmp(jMap[i].N,mapname) == 0)return i;
    }
  }
  return -1; /* no map found */
}

void InsertInVertexList(vertex *vj, vertex *vi){
 long i,j;
 if(vi->n > 0)for(i=0;i<vi->n;i++)if(vi->adj[i] == vj)return;
 if(vj->n > 0)for(j=0;j<vj->n;j++)if(vj->adj[j] == vi)return;
 if(vi->n == 0){
   if((vi->adj = (vertex **)malloc((vi->n + 1)*sizeof(vertex *))) == NULL){
     mes3ds("out of memory\n");  return;
   }
 }
 else{
   if((vi->adj = (vertex **)realloc(vi->adj,(vi->n + 1)*sizeof(vertex *))) == NULL){
     mes3ds("out of memory\n");  return;
   }
 }
 vi->adj[vi->n] = vj;
 (vi->n)++;
}

void get_faces(FILE *fi){
 char cc,matname[128],str[128];
 unsigned short i,n,nf;
 unsigned short a,b,c,f;
 long j,l,mn;
 edge *last_edge=MainEp;
 face **ff=NULL;
 nf=n=GetShort(fi);
//mes3ds("%d Faces ",n);
 if((ff=(face **)malloc((long)nf*sizeof(face *))) == NULL){
  mes3ds("out of memory\n");
  return;
 }
 for(i=0;i<n;i++){
  a=GetShort(fi);  b=GetShort(fi);  c=GetShort(fi);  f=GetShort(fi);
  CreateFace(*(p+a),*(p+b),*(p+c),0); *(ff+i)=MainFp;
  InsertInVertexList(*(p+b),*(p+a));
  InsertInVertexList(*(p+c),*(p+b));
  InsertInVertexList(*(p+a),*(p+c));
 }
 UpdateProgress(fi);
 BuildEdges();
 loop:
 n=GetShort(fi); l=GetLong(fi);
//mes3ds("Next packet=%x  length=%ld\n",n,l);
 if(n == 0x4130){
   j=0; while((cc=GetByte(fi)) != 0)matname[j++]=cc;
   matname[j]=0;
   sprintf(str,"Material [%s] on surfaces\n",matname);
   mes3ds(str);
   mn=GetMap(matname);
   n=GetShort(fi);
   for(i=0;i<n;i++){
     f=GetShort(fi);
     if(mn >= 0){
       for(j=0;j<3;j++){
         (*(ff+f))->color[j]=jMap[mn].d_colour[j];
       }
       (*(ff+f))->texture |= 0x20;  /* brilliance bit */
       if(jMap[mn].evi >= 0){                   /* brush is on */
         (*(ff+f))->texture |= 0x40;            /* image map on       */
         (*(ff+f))->brush = jMap[mn].evi;       /* image map          */
         (*(ff+f))->brush |= 0x80;              /* tiled by default   */
       }
     }
   }
   goto loop;
 }
 else if(n == 0x4150){ /* smothing group */
   if((l-6)/4 != nf){
     mes3ds("Error in face data\n"); return;
   }
   for(i=0;i<nf;i++){
     l=GetLong(fi);
     (*(ff+i))->texture |= 0x80;  /* smoothing */
   }
   goto loop;
 }
 if(ff != NULL)free(ff);
 UpdateProgress(fi);
}


void BuildEdges(void){
 long i;
 vertex *vp;
 vp=MainVp; while(vp != vertex_object_start){
  if(vp->n > 0)for(i=0;i<vp->n;i++)CreateEdge1(vp,vp->adj[i],NULL);
  vp=vp->last;
 }
}

void get_object(long count, FILE *fi){
  char c,objectname[64],str[128];
  unsigned short id;
  long cc,j,i,l;
  p=NULL; vertex_object_start=MainVp;
//printf("loading Object\n");
  j=0; while((c=GetByte(fi)) != 0){
    objectname[j++]=c;
//printf("j=%ld c=%c\n",j,c);
  }
  objectname[j]=0;
//printf("object name loaded\n",j,c);
  sprintf(str,"\n");
  mes3ds(str);
  sprintf(str,"Object [%s]\n",objectname);
  mes3ds(str);
  id=GetShort(fi); l=GetLong(fi);
  if(id == 0x4100){  /* mesh object */
    CreateSkeleton(FirstSp);
    strcpy(MainSp->name,objectname);
    while(1){
      j=ftell(fi);  id=GetShort(fi); cc=GetLong(fi);
      if(id < 0x4001 || id > 0x5000)break;
//printf("Object Section %x\t length=%ld\n",id,cc);
      if(id == 0x4110)get_vertices(fi);
      if(id == 0x4140)get_mapping_coords(fi);
      if(id == 0x4120)get_faces(fi);
      fseek(fi,j,SEEK_SET); for(i=0;i<cc;i++)GetByte(fi);
   }
 }
 else if(id == 0x4600){
   mes3ds("Light\n");
 }
 else if(id == 0x4700){
   mes3ds("Camera\n");
 }
 if(p != NULL)free(p);
}

void rescale_model(short option){
  vertex *vp;
  skel   *sp;
  char str[128];
  long i;
  float c[3],maxd=1.e-30;
  double scan;
  for(i=0;i<3;i++){
    FirstSp->p[i]=c[i]=(gmax[i]+gmin[i])/2.0;
    if(gmax[i]-gmin[i] > maxd)maxd=gmax[i]-gmin[i];
  }
  if(option == 0){
    scale_3ds=(float)MAXUNIT/2.0/maxd;
  }
  else if(option == 1){ /* scale set by param  centre from file */
    struct SC {double scan,xc,yc,zc,c;} Sc;
    Sc.xc=c[0]; Sc.yc=c[1]; Sc.zc=c[2]; Sc.c=1;
    Sc.scan=(double)(UNIT2)/(double)maxd;
    scale_3ds=(float)Sc.scan;
    if(DialogBoxParam(hInstance,MAKEINTRESOURCE(DLG_SCALE),hWndParent,
                     (DLGPROC)DlgProcScale,(LPARAM)&Sc)){
      scale_3ds=(float)Sc.scan;
      if(Sc.c == 1){c[0]=c[1]=c[2]=0.0;}
      else{c[0]=Sc.xc; c[1]=Sc.yc; c[2]=Sc.zc;}
    }
  }
  vp=MainVp; while(vp != NULL){
    for(i=0;i<3;i++)vp->xyz[i] = (long)(scale_3ds*(vp->p[i]-c[i]));
    vp=vp->last;
  }
  sp=MainSp; while(sp != NULL){
    for(i=0;i<3;i++)sp->xyz[i] = (long)(scale_3ds*(sp->p[i]-c[i]));
    sp=sp->last;
  }
}

BOOL Do3dsConvert(int argc, char **argv){
 float size=500.0;
 long count,fze,sc=1;
 unsigned short i,id;
 char c,argv1[128],str[128];
 if(argc < 2)return FALSE;
 strcpy(argv1,argv[1]);
 if((i=strlen(argv1)) < 4 || argv1[i-4] != '.')strcat(argv1,".3DS");
 nImaps=0; nLockediMaps=0;
 strcpy(iSha[0].N,"INTERNAL"); nSha=1;
 iSha[0].P[0]=0;          iSha[0].P[1]=0;         iSha[0].P[2]=TVpointZ;
 iSha[0].X[0]=TVsizeX/2;  iSha[0].X[1]=0;         iSha[0].X[2]=TVpointZ;
 iSha[0].Y[0]=0;          iSha[0].Y[1]=TVsizeZ/2; iSha[0].Y[2]=TVpointZ;
 xShaders[0].in_use=0; xShaders[1].in_use=0;
 xShaders[2].in_use=0; xShaders[3].in_use=0;
 xShaders[4].in_use=0; xShaders[5].in_use=0;
 xShaders[6].in_use=0; xShaders[7].in_use=0;
 CreateSkeleton(NULL); FirstSp=MainSp;
 drive=3;
 posn=0;
 if(argc > 3){
   i=3;
   while(i < argc){
     if(strcmp(argv[i],"-sc") == 0){
       i++; size=atof(argv[i]);
       scale_3ds=(float)MAXUNIT/2.0/size;
       sc=1;
     }
     else if(strcmp(argv[i],"-q") == 0)silent_3ds=1;
     i++;
   }
 }
 for(i=0;i<3;i++){
   gmax[i] = -1.0e30;
   gmin[i] =  1.0e30;
 }
 if((fi=fopen(argv1,"rb")) == NULL){
  mes3ds("Failed to open file\n");
  return FALSE;
 }
 id=GetShort(fi); posn += 2;
 if(id != 0x4d4d){
  mes3ds("Not 3DS file\n"); goto EXIT;
 }
 fze=count=GetLong(fi); posn += 4;
 while(1){
   id=GetShort(fi); if(posn >= fze)goto EXIT;
   count=GetLong(fi);
   if(id == 0x3d3d){
     posn += 6;
     while(1){
       id=GetShort(fi); if(posn >= fze)goto EXIT;
       count=GetLong(fi);
       if(id == 0x4000){
         UpdateProgress(fi);
         get_object(count,fi);
         fseek(fi,posn,SEEK_SET);
       }
       if(id == 0xafff){
         UpdateProgress(fi);
         get_material(count,fi);
         fseek(fi,posn,SEEK_SET);
       }
       posn += count;
       fseek(fi,posn,SEEK_SET);
     }
   }
   else{
     posn += count;
     fseek(fi,posn,SEEK_SET);
     UpdateProgress(fi);
   }
 }
 EXIT:
 UpdateProgress(fi);
 fclose(fi);
#if 0
 sprintf(str,"%ld Vertices   %ld Edges  and  %ld Faces\n",Nvert,Nedge,Nface);
 mes3ds(str);
 sprintf(str,"\txmin=%8.4f\txmax=%8.4f\n",gmin[0],gmax[0]);
 mes3ds(str);
 sprintf(str,"\tymin=%8.4f\tymax=%8.4f\n",gmin[1],gmax[1]);
 mes3ds(str);
 sprintf(str,"\tzmin=%8.4f\tzmax=%8.4f\n",gmin[2],gmax[2]);
 mes3ds(str);
#endif
 if(dlg != NULL)DestroyWindow(dlg); dlg=NULL;
 if(sc)rescale_model(1);
 else  rescale_model(0);
 if(argc < 3)return FALSE;
 strcpy(argv1,argv[2]);
 if((i=strlen(argv1)) < 4 || argv1[i-4] != '.')strcat(argv1,".mfx");
 dlg=CreateDialog(hInstance,MAKEINTRESOURCE(DLGWRIT),hWndParent,(DLGPROC)DlgProc);
 SaveObject(argv1,1);
 return TRUE;
}

