/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* File 3dSOUT.C DLL to create a 3DS from EVI  */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <windows.h>

#include "struct.h"
#include "dstruct.h"

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HWND      hParent;
static HINSTANCE hThisInstance;

#define DOS386
#define _X__FAR
#define PI  3.1415926

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

#define MINUNIT       2048L     // was 1024 trying larger to prevent error
#define MAXUNIT    4194304L     //1073741824L
#define PI2            PI/2
/* definition of flags and identifiers*/
#define DESELECTED       0
#define SELECTED         1
#define HIDDEN           2
#define INEDITOR         6

typedef struct tagMap3DS {
 char name[15];
 unsigned char dr,dg,db;
} Map3DS;

static Map3DS *map3DSlist=NULL;
static long   map3DScount;
static face   **faces=NULL;
static long   fcc;
static unsigned short *faceid=NULL;
static long lrulerx,lrulery,lrulerz;
static double ruler;

static void mes3ds(char *s){
 return;
}

static long AdjacentTo(face *f, face *F){
 long F0,F1,F2,f0,f1,f2;
 F0=F->V[0]; F1=F->V[1]; F2=F->V[2];
 f0=f->V[0]; f1=f->V[1]; f2=f->V[2];
 if(f0 == F0 && f2 == F1)return 1;
 if(f2 == F0 && f0 == F1){f->V[2]=f0; f->V[0]=f2; return 1;}
 if(f2 == F0 && f1 == F1)return 1;
 if(f1 == F0 && f2 == F1){f->V[1]=f2; f->V[2]=f1; return 1;}
 if(f1 == F0 && f0 == F1)return 1;
 if(f0 == F0 && f1 == F1){f->V[0]=f1; f->V[1]=f0; return 1;}
 if(f0 == F1 && f2 == F2)return 1;
 if(f2 == F1 && f0 == F2){f->V[2]=f0; f->V[0]=f2; return 1;}
 if(f2 == F1 && f1 == F2)return 1;
 if(f1 == F1 && f2 == F2){f->V[1]=f2; f->V[2]=f1; return 1;}
 if(f1 == F1 && f0 == F2)return 1;
 if(f0 == F1 && f1 == F2){f->V[0]=f1; f->V[1]=f0; return 1;}
 if(f0 == F2 && f2 == F0)return 1;
 if(f2 == F2 && f0 == F0){f->V[2]=f0; f->V[0]=f2; return 1;}
 if(f2 == F2 && f1 == F0)return 1;
 if(f1 == F2 && f2 == F0){f->V[1]=f2; f->V[2]=f1; return 1;}
 if(f1 == F2 && f0 == F0)return 1;
 if(f0 == F2 && f1 == F0){f->V[0]=f1; f->V[1]=f0; return 1;}
 return 0;
}

static void NormaliseAdjacentFaces(face *F){
 long i;
 face *f;
 if(fcc > 0)for(i=0;i<fcc;i++){
   if((f=faces[i]) != NULL && AdjacentTo(f,F)){
     faces[i]=NULL;
     NormaliseAdjacentFaces(f);
   }
 }
}

static void MakeNormalsPointOut(unsigned short nf, char *name){
 long i;
 long fp;
 face *Fp;
 char str[80];
 if((faces = (face **)X__Malloc((long)nf*(long)sizeof(face *))) == NULL){
   mes3ds("out of memory");
   return;
 }
 fcc=0;
 fp=0; while(fp < Nface){
   Fp=(MainFp+fp);
   if((MainVp+Fp->V[0])->status == INEDITOR &&
      (MainVp+Fp->V[1])->status == INEDITOR &&
      (MainVp+Fp->V[2])->status == INEDITOR)faces[fcc++]=Fp;
   fp++;
 }
 sprintf(str,"%6ld polygons in object %s",fcc,name);
 mes3ds(str);
 if(fcc > 0)for(i=0;i<fcc;i++){
   if((Fp=faces[i]) != NULL){
     faces[i]=NULL;
     NormaliseAdjacentFaces(Fp);
   }
 }
 X__Free(faces);
}

static void Add3dsMaterial(face *f, unsigned char *color){
 long i;
 if(map3DScount > 0){
   for(i=0;i<map3DScount;i++){
     if(map3DSlist[i].dr == color[0] &&
        map3DSlist[i].dg == color[1] &&
        map3DSlist[i].db == color[2]){
       f->pad=(unsigned char)i;
       return;
     }
   }
 }
 if(map3DScount == 0)
   map3DSlist = (Map3DS *)X__Malloc((map3DScount+1)*sizeof(Map3DS));
 else
   map3DSlist = (Map3DS *)X__Realloc(map3DSlist,(map3DScount+1)*sizeof(Map3DS));
 if(map3DSlist == NULL){
   mes3ds("out of memory");
   return;
 }
 map3DSlist[map3DScount].dr=color[0];
 map3DSlist[map3DScount].dg=color[1];
 map3DSlist[map3DScount].db=color[2];
 f->pad=(unsigned char)map3DScount;
 map3DScount++;
}

void WriteMapChunk(FILE *fi){
 long i,pos_map,
        len_map,len_name,len14=14,len15=15,len8=8,len6=6,
        cpos;
 char name[16];
 /* sixbytes is 0x11 and size of 3 chars giving colours */
 /* sixbytes1 is 0x30 and size of 2 bytes for attr      */
 char onebyte=0,sixbytes[6]={17,0,9,0,0,0},zero=0,
               sixbytes1[6]={48,0,8,0,0,0};
 unsigned char white[3]={255,255,255},grey[3]={64,64,64};
 char     shinyness=50,transparency=0;
 unsigned short magic_afff=0xafff,
                magic_a000=0xa000,   /* name     */
                magic_a010=0xa010,   /* ambient  */
                magic_a020=0xa020,   /* diffuse  */
                magic_a030=0xa030,   /* specular */
                magic_a040=0xa040,
                magic_a050=0xa050,
                magic_a052=0xa052,
                magic_a053=0xa053,
                magic_a100=0xa100,
                magic_a081=0xa081,   /* 2 sided  */
       oneshort;
 for(i=0;i<map3DScount;i++){
   fwrite(&magic_afff,sizeof(short),1,fi);
   len_map=0;
   pos_map=ftell(fi); fwrite(&len_map,sizeof(long),1,fi);
   sprintf(name,"MATERIAL%ld",i);
   strcpy(map3DSlist[i].name,name);
   fwrite(&magic_a000,sizeof(short),1,fi);
   len_name=strlen(name)+1+6;
   fwrite(&len_name,sizeof(long),1,fi);
   len_name -= 6;
   fwrite(name,1,len_name,fi);
   fwrite(&magic_a010,sizeof(short),1,fi); /* ambient */
   fwrite(&len15,sizeof(long),1,fi);
   fwrite(sixbytes,1,6,fi);
   fwrite(grey,1,3,fi);
   fwrite(&magic_a020,sizeof(short),1,fi); /* diffuse */
   fwrite(&len15,sizeof(long),1,fi);
   fwrite(sixbytes,1,6,fi);
   fwrite(&(map3DSlist[i].dr),1,1,fi);
   fwrite(&(map3DSlist[i].dg),1,1,fi);
   fwrite(&(map3DSlist[i].db),1,1,fi);
   fwrite(&magic_a030,sizeof(short),1,fi); /* specular colour */
   fwrite(&len15,sizeof(long),1,fi);
   fwrite(sixbytes,1,6,fi);
   fwrite(white,1,3,fi);
   fwrite(&magic_a040,sizeof(short),1,fi); /* shinyness  0 - 100 */
   fwrite(&len14,sizeof(long),1,fi);
   fwrite(sixbytes1,1,6,fi);
   fwrite(&shinyness,1,1,fi);
   fwrite(&onebyte,1,1,fi);
   fwrite(&magic_a050,sizeof(short),1,fi);
   fwrite(&len14,sizeof(long),1,fi);
   fwrite(sixbytes1,1,6,fi);
   fwrite(&transparency,1,1,fi);
   fwrite(&onebyte,1,1,fi);
   fwrite(&magic_a052,sizeof(short),1,fi);
   fwrite(&len14,sizeof(long),1,fi);
   fwrite(sixbytes1,1,6,fi);
   fwrite(&zero,1,1,fi);
   fwrite(&onebyte,1,1,fi);
   fwrite(&magic_a053,sizeof(short),1,fi);
   fwrite(&len14,sizeof(long),1,fi);
   fwrite(sixbytes1,1,6,fi);
   fwrite(&zero,1,1,fi);
   fwrite(&onebyte,1,1,fi);
   fwrite(&magic_a100,sizeof(short),1,fi);  /* a_100 */
   fwrite(&len8,sizeof(long),1,fi);
   oneshort=3;
   fwrite(&oneshort,sizeof(short),1,fi);
 fwrite(&magic_a081,sizeof(short),1,fi);  /* 2 sided */
 fwrite(&len6,sizeof(long),1,fi);
   cpos=ftell(fi);
   fseek(fi,pos_map,SEEK_SET);
   pos_map = cpos-pos_map+2;
   fwrite(&pos_map,sizeof(long),1,fi);
   fseek(fi,0L,SEEK_END);
 }
}

void Create3DSfile(FILE *fi){
 char layer[]={"MODEL"},cl[64];
 long fp,vp;
 vertex *Vp;
 face   *Fp;
 skel   *sp;
 long i;
 unsigned short nv,nf,j,k;
 float x,y,z,scale3ds=0.0001;
 long xmax = -MAXUNIT, ymax = -MAXUNIT, zmax = -MAXUNIT,
      xmin =  MAXUNIT, ymin =  MAXUNIT, zmin =  MAXUNIT;
 unsigned short magic_4d4d=0x4d4d,
                magic_3d3d=0x3d3d,
                magic_4000=0x4000,
                magic_4100=0x4100,
                magic_4110=0x4110,
                magic_4120=0x4120,
                magic_4130=0x4130,
                magic_4150=0x4150,
       attribs=7;
 long pos_file,pos_sect,pos_obj,pos_mesh,pos_vert,pos_face,pos_mat,
      pos_num,pos_smo,
      len_file,len_sect,len_obj,len_mesh,len_vert,len_face,len_mat,
      len_smo,
      cpos,group;
 map3DSlist=NULL;
 map3DScount=0;
 len_file=len_sect=len_mesh=0;
 vp=0; while(vp <Nvert){
   Vp=(MainVp+vp);
   if(Vp->status == SELECTED){
     if(Vp->xyz[0] > xmax)xmax=Vp->xyz[0];
     if(Vp->xyz[1] > ymax)ymax=Vp->xyz[1];
     if(Vp->xyz[2] > zmax)zmax=Vp->xyz[2];
     if(Vp->xyz[0] < xmin)xmin=Vp->xyz[0];
     if(Vp->xyz[1] < ymin)ymin=Vp->xyz[1];
     if(Vp->xyz[2] < zmin)zmin=Vp->xyz[2];
   }
   vp++;
 }
 fwrite(&magic_4d4d,sizeof(short),1,fi);
 pos_file=ftell(fi); fwrite(&len_file,sizeof(long),1,fi);
 fwrite(&magic_3d3d,sizeof(short),1,fi);
 pos_sect=ftell(fi); fwrite(&len_sect,sizeof(long),1,fi);
 fp=0; while(fp < Nface){
   Fp=(MainFp+fp);
   if((MainVp+Fp->V[0])->status == SELECTED &&
      (MainVp+Fp->V[1])->status == SELECTED &&
      (MainVp+Fp->V[2])->status == SELECTED)
      Add3dsMaterial(Fp,Fp->color);
   fp++;
 }
 if(map3DScount > 0)WriteMapChunk(fi);
 sp=MainSp; for(;;){ /* sp will be null for vertices not on any node */
   if(sp != NULL)strcpy(cl,sp->name);
   else          strcpy(cl,layer);
   vp=0; nv=0; while(vp < Nvert){
     Vp=(MainVp+vp);
     if(Vp->status == SELECTED && Vp->sp == sp){
       Vp->status=INEDITOR;
       nv++;
     }
     vp++;
   }
   nf=0; fp=0; while(fp < Nface){
     Fp=(MainFp+fp);
     if((MainVp+Fp->V[0])->status == INEDITOR &&
        (MainVp+Fp->V[1])->status == INEDITOR &&
        (MainVp+Fp->V[2])->status == INEDITOR)nf++;
     fp++;
   }
   if(nf > 0){ /* only write if there are any faces */
     unsigned short id;
     MakeNormalsPointOut(nf,cl);
     len_obj=len_mesh=len_vert=len_face=0;
     fwrite(&magic_4000,sizeof(short),1,fi);
     pos_obj=ftell(fi); fwrite(&len_obj,sizeof(long),1,fi);
     fwrite(cl,1,strlen(cl)+1,fi);
     fwrite(&magic_4100,sizeof(short),1,fi);
     pos_mesh=ftell(fi); fwrite(&len_mesh,sizeof(long),1,fi);
     fwrite(&magic_4110,sizeof(short),1,fi);
     len_vert=6L+2L+3L*sizeof(float)*(long)nv;
     pos_vert=ftell(fi); fwrite(&len_vert,sizeof(long),1,fi);
     fwrite(&nv,sizeof(short),1,fi);
     i=0; vp=0; while(vp < Nvert){
       Vp=(MainVp+vp);
       if(Vp->status == INEDITOR){
          x=(float)(Vp->xyz[0]-lrulerx)/(float)ruler;
          y=(float)(Vp->xyz[1]-lrulery)/(float)ruler;
          z=(float)(Vp->xyz[2]-lrulerz)/(float)ruler;
          fwrite(&x,sizeof(float),1,fi);
          fwrite(&y,sizeof(float),1,fi);
          fwrite(&z,sizeof(float),1,fi);
          Vp->id = i++;
       }
       else Vp->id = 65535;
       vp++;
     }
     fwrite(&magic_4120,sizeof(short),1,fi);
     pos_face=ftell(fi); len_face=0; fwrite(&len_face,sizeof(long),1,fi);
     fwrite(&nf,sizeof(short),1,fi);
     fp=0; j=0; while(fp < Nface){
       Fp=(MainFp+fp);
       if((MainVp+Fp->V[0])->status == INEDITOR &&
          (MainVp+Fp->V[1])->status == INEDITOR &&
          (MainVp+Fp->V[2])->status == INEDITOR){
         id=(unsigned short)((MainVp+Fp->V[0])->id);
         fwrite(&id,sizeof(unsigned short),1,fi);
         id=(unsigned short)((MainVp+Fp->V[1])->id);
         fwrite(&id,sizeof(unsigned short),1,fi);
         id=(unsigned short)((MainVp+Fp->V[2])->id);
         fwrite(&id,sizeof(unsigned short),1,fi);
         fwrite(&attribs,sizeof(short),1,fi);
         faceid[fp] = (unsigned short)(j++);
       }
       fp++;
     }
     if(map3DScount > 0){
       for(i=0;i<map3DScount;i++){
         j=0;
         fp=0; while(fp < Nface){
           Fp=(MainFp+fp);
           if((MainVp+Fp->V[0])->status == INEDITOR &&
              (MainVp+Fp->V[1])->status == INEDITOR &&
              (MainVp+Fp->V[2])->status == INEDITOR &&
              Fp->pad == (unsigned char)i){
             if(j == 0){
               fwrite(&magic_4130,sizeof(short),1,fi); /* material assign */
               pos_mat=ftell(fi); len_mat=0;
               fwrite(&len_mat,sizeof(long),1,fi);
               j=strlen(map3DSlist[i].name)+1;
               fwrite(map3DSlist[i].name,1,j,fi);
               pos_num=ftell(fi); j=0;
               fwrite(&j,sizeof(short),1,fi);
             }
             k=faceid[fp];
             fwrite(&k,sizeof(short),1,fi);
             j++;
           }
           fp++;
         }
         if(j > 0){
           cpos=ftell(fi);
           fseek(fi,pos_mat,SEEK_SET);
           pos_mat = cpos-pos_mat+2;
           fwrite(&pos_mat,sizeof(long),1,fi);
           fseek(fi,pos_num,SEEK_SET);
           fwrite(&j,sizeof(short),1,fi);
           fseek(fi,0L,SEEK_END);
         }
       }
     }
     fwrite(&magic_4150,sizeof(short),1,fi); /* smoothing */
     pos_smo=ftell(fi); len_smo=6+nf*sizeof(long);
     fwrite(&len_smo,sizeof(long),1,fi);
     fp=0; while(fp < Nface){
       Fp=(MainFp+fp);
       if((MainVp+Fp->V[0])->status == INEDITOR &&
          (MainVp+Fp->V[1])->status == INEDITOR &&
          (MainVp+Fp->V[2])->status == INEDITOR){
         if((Fp->attrib & 0x80) == 0x80)group=1;
         else                           group=0;
         fwrite(&group,sizeof(long),1,fi);
       }
       fp++;
     }
     cpos=ftell(fi);
     fseek(fi,pos_face,SEEK_SET);
     len_face = cpos-pos_face+2;
     fwrite(&len_face,sizeof(long),1,fi);
     fseek(fi,pos_obj,SEEK_SET);
     pos_obj = cpos-pos_obj+2;
     fwrite(&pos_obj,sizeof(long),1,fi);
     fseek(fi,pos_mesh,SEEK_SET);
     pos_mesh = cpos-pos_mesh+2;
     fwrite(&pos_mesh,sizeof(long),1,fi);
     fseek(fi,0L,SEEK_END);
   }
   vp=0; while(vp < Nvert){ /* reset status */
     Vp=(MainVp+vp);
     if(Vp->status == INEDITOR)Vp->status=SELECTED;
     vp++;
   }
   if(sp == NULL)break;
   sp=sp->last;
 }  /* end of for(;;) */
 cpos=ftell(fi);
 fseek(fi,pos_sect,SEEK_SET);
 pos_sect = cpos-pos_sect+2;
 fwrite(&pos_sect,sizeof(long),1,fi);
 fseek(fi,pos_file,SEEK_SET);
 pos_file = cpos-pos_file+2;
 fwrite(&pos_file,sizeof(long),1,fi);
 fseek(fi,0L,SEEK_END);
}

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hThisInstance=hDLL;
      break;
    }
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return TRUE;
}

long _Export
 (HWND parent_window, char *filename,X__STRUCTURE *lpevi){
 char layer[8]={"MODEL"},cl[16];
 FILE *dxfp;
 vertex *Vp,*V0,*V1,*V2;
 face   *Fp;
 skel   *sp;
 long   vp,fp;
 long i,nf,colour=7;
 long xmax = -MAXUNIT, ymax = -MAXUNIT, zmax = -MAXUNIT,
      xmin =  MAXUNIT, ymin =  MAXUNIT, zmin =  MAXUNIT;
 lpEVI=lpevi;
 if(filename == NULL)return 0;
 if(NvertSelect == 0 || Nface == 0){
   MessageBox(parent_window,"Too few vertices or faces selected",NULL,MB_OK);
   return 0;
 }
 lrulerx= (*(lpEVI->orulerx));
 lrulery= (*(lpEVI->orulery));
 lrulerz= (*(lpEVI->orulerz));
 ruler  = (*(lpEVI->ruler));
 if((dxfp=fopen(filename,"wb")) != NULL){
   if((faceid=(unsigned short *)X__Malloc(Nface*sizeof(unsigned short))) != NULL){
     Create3DSfile(dxfp);
     X__Free(faceid);
   }
   else MessageBox(NULL,"No memory",NULL,MB_OK);
   fclose(dxfp);
 }
 else{
   MessageBox(parent_window,"File open Error",NULL,MB_OK);
 }
 if(map3DSlist != NULL)X__Free(map3DSlist);
 return 1;
}
