/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* File OBJSAVE.C      object file format saver  in LOADER.C */

static void WriteWireframe(void);
static void WriteBezierCurve(void);
static void WriteVertexMapping(short status);
static void WriteVerticesWithOffset(point p, long invert, short short_form,
                                    short status);
static void WriteOffsetTextures(point p, unsigned char maps, long invert);
static void WriteFacesWithMapCheck(short short_form);
static void WriteShaders(short status);
static void WriteShadersAxes(void);
static void WriteOffsetShadersAxes(point p, long invert);

#ifdef __X__WINDOWS__
#else
static char *file_check_list[]={"Cancel","Save Over"};
#endif

void SaveObject(char *Filename, short status){
  long pos1,pos2,pos3,pos4,pos5,pos6,pos7,pos8,pos9,posA,posB,
       posC,posD,posE,posF,pose;
  long FORMsize;
  short id,short_form=0;
  FORMsize=0; newvert=0; newface=0; newedge=0; newglue=0;
#ifdef __DEMO__
  inform("NOT IN DEMO VERSION"); pause(30); disinform();
  return;
#endif
  /* status = 3 -> saving with memory problem - no extra display  !! */
  if(status < 3){   /* no memory problem */
#ifdef __X__WINDOWS__
;
#else
    if( (fo=fopen(Filename,"r")) != NULL){ /* check for old file */
      fclose(fo);
      id=RequestItemSelect(2,file_check_list,"File exists Save Over ?");
      if(id < 1)return;
    }
#endif
  }
  if(status == 1)EDIT_ACTION=NO;
  if( (fo=fopen(Filename,"wb")) == NULL){   /* now open for reading */
#ifdef __X__WINDOWS__
#else
    Output_Message(G_FILEOPENFAIL);
#endif
    return;
  }
#ifdef __X__WINDOWS__
#else
  if(status < 3)DrawOldMarker();
  if(status < 2)inform("Saving ...");
#endif
  fprintf(fo,"FORM");
  pos1=ftell(fo); outlng(FORMsize);
  fprintf(fo,"OFXM");
  fprintf(fo,"AXIS"); outlng(48L); /* size of data chunk */
  if(status == 2){  /* dont use offset for rendering in triview */
    outlng(0); outlng(0); outlng(0);
  }
  else{
    outlng(ObjectAxis.Origin.xyz[0]);
    outlng(ObjectAxis.Origin.xyz[1]);
    outlng(ObjectAxis.Origin.xyz[2]);
  }
  outlng(iSha[0].P[0]);  /* store the algorithmic texture origin */
  outlng(iSha[0].P[1]);  /* in the object axis structure         */
  outlng(iSha[0].P[2]);
  outlng(iSha[0].X[0]);
  outlng(iSha[0].X[1]);
  outlng(iSha[0].X[2]);
  outlng(iSha[0].Y[0]);
  outlng(iSha[0].Y[1]);
  outlng(iSha[0].Y[2]);
  fprintf(fo,"OFFS"); outlng(16L);      /* follow point */
  outlng(3L);
  outlng(ObjectAxis.Offset.xyz[0]);
  outlng(ObjectAxis.Offset.xyz[1]);
  outlng(ObjectAxis.Offset.xyz[2]);
  if(Nvert > 0){
    fprintf(fo,"VERT"); pos2=ftell(fo); outlng(12*newvert);
    WriteVertices(status);
  }
  else pos2 = -1;
  if(Nedge > 0){
    fprintf(fo,"EDGE"); pos3=ftell(fo); outlng( 8*newedge); WriteEdges(0);
  }
  else pos3 = -1;
  if(Nface > 0){
    if(nSha <= 1){
      fprintf(fo,"FACE"); pos4=ftell(fo); outlng(20*newface); WriteFaces(0);
    }
    else{
      fprintf(fo,"SFAC"); pos4=ftell(fo); outlng(21*newface); WriteFaces(1);
    }
  }
  else pos4 = -1;
  if(NvertGlue > 0){
    fprintf(fo,"VTEX"); posB=ftell(fo); outlng(12*newglue);
    WriteVertexMapping(status);
  }
  else posB = -1;
  if(xShaders[0].in_use || xShaders[1].in_use ||
     xShaders[2].in_use || xShaders[3].in_use){
    fprintf(fo,"SHDR"); posC=ftell(fo); outlng(0L);
    WriteShaders(status);
    posD=ftell(fo);
  }
  else posC = -1;
  if(nSha > 1){
    fprintf(fo,"AXES"); posE=ftell(fo); outlng(68*(nSha-1));
    WriteShadersAxes();
    posF=ftell(fo);
  }
  else posE = -1;
  if(nImaps > 0){
    fprintf(fo,"TEXT"); pos5=ftell(fo); outlng(0L);
    WriteTextures(1);
    pos6=ftell(fo);
  }
  else pos5 = -1;         /*! root node only implies no skeleton */
  if((status == 1 || status == 3) &&
      Nskel > 1){/* if saving whole object then save skeleton */
    fprintf(fo,"SK02"); pos7=ftell(fo); outlng(0L);
    WriteSkeleton();
    pos8=ftell(fo);
  }
  else pos7 = -1;
  if(MirrorFlag >= 0 && (status == 1 || status == 3)){ /* only save if full */
    fprintf(fo,"MPID"); outlng((long)(1*sizeof(long)));
    outlng((long)MirrorFlag);
  }
  fprintf(fo,"MATL"); outlng((long)(5*sizeof(long)));
  outlng(4L);
  outlng((long)ShadowFlag);
  outlng(DefaultBrilliance);     /* these are not used any more       */
  outlng(DefaultTransparency);
  outlng(DefaultReflectivity);   /* but remain for file compatibility */
  fprintf(fo,"UNIT"); outlng((long)(7*sizeof(long)+8));
  outlng((long)ruler);
  fwrite(ruler_name,1,8,fo);
  outlng(lrulerx); outlng(lrulery); outlng(lrulerz);
  outlng(rulerx);  outlng(rulery);  outlng(rulerz);
  fprintf(fo,"GRID"); outlng((long)(2*sizeof(long)));
  outlng(grid_size); outlng(grid_size); /* second is unused */
  WriteWireframe();
  if((status == 1 || status == 3) && N_Bcurves > 0){
     fprintf(fo,"BEZ1"); pos9=ftell(fo); outlng(0L);
     WriteBezierCurve();
     posA=ftell(fo);
  }
  else pos9 = -1;
  pose=ftell(fo);  /* the end of the object */
  if(pos2 > 0){fseek(fo,pos2,0); outlng(12*newvert);}
  if(short_form){
    if(pos3 > 0){fseek(fo,pos3,0); outlng( 4*newedge);}
    if(pos4 > 0){fseek(fo,pos4,0); outlng(14*newface);}
  }
  else{
    if(pos3 > 0){fseek(fo,pos3,0); outlng( 8*newedge);}
    if(pos4 > 0){
      fseek(fo,pos4,0);
      if(nSha <= 1)outlng(20*newface);
      else         outlng(21*newface);
    }
  }
  if(posB > 0){fseek(fo,posB,0); outlng(12*newglue);}
  if(posE > 0){fseek(fo,posE,0); outlng(posF-posE-4);} /* shaders new axes  */
  if(posC > 0){fseek(fo,posC,0); outlng(posD-posC-4);} /* extern shader     */
  if(pos5 > 0){fseek(fo,pos5,0); outlng(pos6-pos5-4);} /* brushes !inc size */
  if(pos7 > 0){fseek(fo,pos7,0); outlng(pos8-pos7-4);} /* skeleton          */
  if(pos9 > 0){fseek(fo,pos9,0); outlng(posA-pos9-4);} /* bezier curves     */

  fseek(fo,pos1,0); outlng(pose-8);
#ifdef __X__WINDOWS__
#else
  if(status < 2)disinform();
  if(status < 3)DrawOldMarker();
#endif
  fclose(fo);
}

static void WriteShadersAxes(void){
 long i,j;
 for(i=1;i<nSha;i++){
   fwrite(iSha[i].N,1,32,fo);
   for(j=0;j<3;j++)outlng(iSha[i].P[j]);
   for(j=0;j<3;j++)outlng(iSha[i].X[j]);
   for(j=0;j<3;j++)outlng(iSha[i].Y[j]);
 }
}

void WriteTextures(short version){
 long i,temp;
 skel *sp;
 vertex *v;
 char tempstr[256]="VERSION1.TXU";
 i=0;
 sp=FirstSp; while(sp != NULL){ /* set up sk id */
   sp->id=i; i++;
   sp=sp->next;
 }
 fwrite(&nImaps,1,1,fo);
 for(i=0;i<nImaps;i++){
   outlng(256);                 /* size of Texture list string */
   fwrite(tempstr,1,256,fo);    /* to maintain compatibility   */
   outlng(iMap[i].P[0]); outlng(iMap[i].P[1]); outlng(iMap[i].P[2]);
   outlng(iMap[i].X[0]); outlng(iMap[i].X[1]); outlng(iMap[i].X[2]);
   outlng(iMap[i].Y[0]); outlng(iMap[i].Y[1]); outlng(iMap[i].Y[2]);
   temp=iMap[i].Type; outlng(temp);
   temp=iMap[i].AnimFF; outlng(temp);
   temp=iMap[i].AnimDL; outlng(temp);
   temp=iMap[i].AnimLF; outlng(temp);
   if(iMap[i].Sk != NULL){
     temp=((iMap[i].Sk)->id)+1; outlng(temp);/* 1 is first file brush*/
   }
   else{
     temp=0; outlng(temp);
   }
   temp=iMap[i].Map; outlng(temp);   /* 0 = plane 1 = cyl. map  2/3 = moz */
   if(version == 0)temp=0;
   else{
     temp=32+3+6+9+2;    /* 52 bytes + */
     if(iMap[i].s)temp += 256;
     if(iMap[i].r)temp += 256;
     if(iMap[i].b)temp += 256;
     temp /= 4;   /* make 4 byte long equivalent */
   }
   temp += 5;          outlng(temp); /* for expansion, count - from here */
   temp=iMap[i].Angle; outlng(temp); /* angle for cylinder brushes */
   if(iMap[i].Lock &&
      iMap[i].Vlock[0]->pad >= 0 &&        /* out a tex map if lock */
      iMap[i].Vlock[1]->pad >= 0 &&        /* and if vertices OK    */
      iMap[i].Vlock[2]->pad >= 0 ){
     temp=iMap[i].Lock;  outlng(temp);
     v=iMap[i].Vlock[0];
     temp=(((long)v->id) | ((long)v->pad << 16)); outlng(temp);
     v=iMap[i].Vlock[1];
     temp=(((long)v->id) | ((long)v->pad << 16)); outlng(temp);
     v=iMap[i].Vlock[2];
     temp=(((long)v->id) | ((long)v->pad << 16)); outlng(temp);
   }
   else{  /* map is not locked */
     temp=0; outlng(temp); outlng(temp); outlng(temp); outlng(temp);
   }
   if(version > 0){
     fwrite(iMap[i].N,1,32,fo);
     fwrite(&iMap[i].s,1,1,fo);
     fwrite(&iMap[i].r,1,1,fo);
     fwrite(&iMap[i].b,1,1,fo);
     outsht(iMap[i].sp);
     outsht(iMap[i].rp);
     outsht(iMap[i].bp);
     fwrite(iMap[i].d_colour,1,3,fo);
     fwrite(iMap[i].s_colour,1,3,fo);
     fwrite(iMap[i].a_colour,1,3,fo);
     if(iMap[i].s)fwrite(iMap[i].S,1,256,fo);
     if(iMap[i].r)fwrite(iMap[i].R,1,256,fo);
     if(iMap[i].b)fwrite(iMap[i].B,1,256,fo);
     outsht(0); /* to make up to multiple of 4 bytes and possible expansion */
   }
 }
}

void WriteVertices(short status){
 long i;
 vertex  *vp,*fv;
 fv=NULL; vp=MainVp; while(vp != NULL){fv=vp; vp=vp->last;}
 i=0;
 vp=fv; while(vp != NULL){
  if(status == 2){ /* for use in renderer */
    if(intriview(vp)){
      outlng(vp->xyz[0]); outlng(vp->xyz[1]); outlng(vp->xyz[2]);
      vp->pad = (char)(i >> 16);
      vp->id=(unsigned short)(i & 0x0000ffff);
      i++; newvert++;
    }
    else vp->pad = -1;
  }
  else if( (status==1) || (status == 3) || (vp->status==SELECTED) ){
    outlng(vp->xyz[0]); outlng(vp->xyz[1]); outlng(vp->xyz[2]);
    vp->pad = (char)(i >> 16);
    vp->id=(unsigned short)(i & 0x0000ffff);
    i++; newvert++;
  }
  else vp->pad = -1;
  vp=vp->next;
 }
}

static void WriteVerticesWithOffset(point p, long invert, short short_form,
                                    short status){
 long i;
 vertex  *vp,*fv;
 fv=NULL; vp=MainVp; while(vp != NULL){fv=vp; vp=vp->last;}
 i=0;
 vp=fv; while(vp != NULL){
  if(vp->status==status){
    outlng((vp->xyz[0]-p[0])*invert);
    outlng((vp->xyz[1]-p[1])*invert);
    outlng(vp->xyz[2]-p[2]);
    vp->pad = (char)(i >> 16);
    vp->id=(unsigned short)(i & 0x0000ffff);
    i++; newvert++;
  }
  else vp->pad = -1;
  vp=vp->next;
 }
}

void WriteVertexMapping(short status){
 long i,ix,iy;
 vertex *vp,*fv;
 fv=NULL; vp=MainVp; while(vp != NULL){fv=vp; vp=vp->last;}
 i=0;
 vp=fv; while(vp != NULL){
  if(status == 2){ /* for use in renderer */
    if(intriview(vp)){
#if o
      if(vp->gp){
        ix=(long)(vp->gpx * 1.0e6); iy=(long)(vp->gpy * 1.0e6);
        outlng(i); outlng(ix); outlng(iy);
        newglue++;
      }
#else
      if(vp->gp != NULL){
        ix=(long)(vp->gp->x * 1.0e6); iy=(long)(vp->gp->y * 1.0e6);
        outlng(i); outlng(ix); outlng(iy);
        newglue++;
      }
#endif
      i++;
    }
  }
  else if( (status==1) || (status == 3) || (vp->status==SELECTED) ){
    if(vp->gp != NULL){
      ix=(long)(vp->gp->x * 1.0e6); iy=(long)(vp->gp->y * 1.0e6);
      outlng(i); outlng(ix); outlng(iy);
      newglue++;
    }
    i++;
  }
  vp=vp->next;
 }
}

void WriteEdges(short short_form){
 long id;
 edge *ep;
 ep=MainEp;
 while(ep != NULL){
   if((ep->V[0]->pad >= 0) && (ep->V[1]->pad >= 0)){
     id=(((long)ep->V[0]->id) | ((long)ep->V[0]->pad << 16)); outlng(id);
     id=(((long)ep->V[1]->id) | ((long)ep->V[1]->pad << 16)); outlng(id);
     newedge++;
   }
   ep=ep->last;
 }
}

void WriteFaces(short shader){
 face *f,*ff;
 long id;
 ff=NULL; f=MainFp; while(f != NULL){ff=f; f=f->last;}
 if((f=ff) != NULL)while(f != NULL){
   if((f->V[0]->pad >= 0) &&
      (f->V[1]->pad >= 0) &&
      (f->V[2]->pad >= 0)){
     id=(((long)f->V[0]->id) | ((long)f->V[0]->pad << 16)); outlng(id);
     id=(((long)f->V[1]->id) | ((long)f->V[1]->pad << 16)); outlng(id);
     id=(((long)f->V[2]->id) | ((long)f->V[2]->pad << 16)); outlng(id);
     fwrite(f->color,1,3,fo);
     fwrite(&f->texture,1,1,fo);
     fwrite(f->matcol,1,3,fo);
     fwrite(&f->brush,1,1,fo);
     if(shader)fwrite(&f->axis,1,1,fo);
     newface++;
   }
   f=f->next;
 }
}

static void WriteFacesWithMapCheck(short short_form){
 unsigned char map;
 face *f,*ff;
 long id;
 ff=NULL; f=MainFp; while(f != NULL){ff=f; f=f->last;}
 if((f=ff) != NULL)while(f != NULL){
   if((f->V[0]->pad >= 0) &&
      (f->V[1]->pad >= 0) &&
      (f->V[2]->pad >= 0)){
     id=(((long)f->V[0]->id) | ((long)f->V[0]->pad << 16)); outlng(id);
     id=(((long)f->V[1]->id) | ((long)f->V[1]->pad << 16)); outlng(id);
     id=(((long)f->V[2]->id) | ((long)f->V[2]->pad << 16)); outlng(id);
     fwrite(f->color,1,3,fo);
     fwrite(&f->texture,1,1,fo);
     fwrite(f->matcol,1,3,fo);
     if((f->texture & 0x40) == 0x40){/* map in use */
       map=(f->brush & 0x1f);
       map=(unsigned char)(iMap[(long)map].in_use);  /* this must be >= 0*/
       map=((f->brush & 0xe0) | map);
       fwrite(&map,1,1,fo);
     }
     else fwrite(&f->brush,1,1,fo);
     newface++;
   }
   f=f->next;
 }
}

static void WriteBezierCurve(void){
 long i,j,id;
 outlng(N_Bcurves);
 for(i=0;i<N_Bcurves;i++){
   outlng((long)Bcurves[i].Nc); outlng((long)Bcurves[i].Np);
   fwrite((char *)Bcurves[i].name,1,8,fo);
   for(j=0;j<Bcurves[i].Nc;j++){
     id=((long)((Bcurves[i].Vc[j])->id) |
        ((long)((Bcurves[i].Vc[j])->pad) << 16));
     outlng(id);
   }
   for(j=0;j<Bcurves[i].Np;j++){
     id=((long)((Bcurves[i].Vp[j])->id) |
        ((long)((Bcurves[i].Vp[j])->pad) << 16));
     outlng(id);
   }
 }
}

void WriteSkeleton(void){
 short i,j;
 skel *sp;
 vertex *vp, *fv;
 long l;
 i=0;
 sp=FirstSp; while(sp != NULL){
   sp->id=i; i++;
   sp=sp->next;
 }
 outlng(Nskel);
 sp=FirstSp; while(sp != NULL){
   outlng(sp->xyz[0]); outlng(sp->xyz[1]); outlng(sp->xyz[2]);
   if(sp->at != NULL)outsht(sp->at->id);
   else              outsht(0);
   fwrite(sp->name,sizeof(char),16,fo);
   for(j=0;j<3;j++){l=(long)(sp->u[j]*1.0e8); outlng(l);}
   for(j=0;j<3;j++){l=(long)(sp->v[j]*1.0e8); outlng(l);}
   for(j=0;j<3;j++){l=(long)(sp->w[j]*1.0e8); outlng(l);}
   for(i=0;i<8;i++)for(j=0;j<3;j++)outlng(sp->bx[i][j]);
   sp=sp->next;
 }
 outlng(Nvert);
 fv=NULL; vp=MainVp; while(vp != NULL){fv=vp; vp=vp->last;}
 vp=fv; while(vp != NULL){
   if(vp->sp != NULL)outsht(vp->sp->id);
   else              outsht(0);
   vp=vp->next;
 }
}

static void WriteWireframe(void){
 long i;
 if(w_frame.Np == 0 || w_frame.Ne == 0)return;
 fprintf(fo,"WFRM");
 outlng((long)(((w_frame.Np*3)+(w_frame.Ne*2)+2)*sizeof(long)));
 outlng(w_frame.Np); outlng(w_frame.Ne);
 for(i=0;i<w_frame.Np;i++){
   outlng(w_frame.p[i][0]);
   outlng(w_frame.p[i][1]);
   outlng(w_frame.p[i][2]);
 }
 for(i=0;i<w_frame.Ne;i++){
   outlng(w_frame.e[i][0]); outlng(w_frame.e[i][1]);
 }
}

void SaveObjects(char *rootname, short do_all, short rotate180){
 char Filename[255],NodeName[16];
 vertex *vp;
 face   *fp;
 edge   *ep;
 skel   *sp,*sp1;   /* sp1 gives the offset for follow point */
 long i,nf,ne,map,invert;
 long pos1,pos2,pos3,pos4,pos5,pos6,posA,posB,posC,posD,pose;
 long FORMsize;
 short id;
#ifdef __DEMO__
 inform("NOT IN DEMO VERSION"); pause(30); disinform();
 return;
#endif
 if(rotate180)invert = -1;
 else         invert =  1;
 sp=MainSp; while(sp != NULL){
   if(sp->at == FirstSp || do_all){
     vp=MainVp; while(vp != NULL){
       if(vp->status == SELECTED && vp->sp == sp)vp->status=INEDITOR;
       vp=vp->last;
     }
     sp1=MainSp; while(sp1 != NULL){
       if(sp1->at == sp)break;
       sp1=sp1->last;
     }
     if(nImaps > 0)for(i=0;i<nImaps;i++)iMap[i].in_use = -1;
     nf=0; fp=MainFp; while(fp != NULL){
       if(fp->V[0]->status == INEDITOR && fp->V[1]->status == INEDITOR &&
          fp->V[2]->status == INEDITOR){
            if((fp->texture & 0x40) == 0x40){/* map in use */
              map=(long)(fp->brush & 0x1f);
              iMap[map].in_use=1;
            }
            nf++;
          }
       fp=fp->last;
     }
     map=0;
     if(nImaps > 0)for(i=0;i<nImaps;i++){
       if(iMap[i].in_use == 1){
         iMap[i].in_use = (short)map;
         map++;
       }
     }
     ne=0; if(nf == 0){
       ep=MainEp; while(ep != NULL){/* no faces check edges */
         if(ep->V[0]->status == INEDITOR && ep->V[1]->status == INEDITOR)ne++;
         ep=ep->last;
       }
     }
     if(nf > 0 || ne > 0){ /* only write if there are any faces or edges */
       FORMsize=0; newvert=0; newface=0; newedge=0;  /* edges => faces   */
       strcpy(NodeName,sp->name);
       id=min(strlen(sp->name),8); NodeName[id]=0;
       sprintf(Filename,"Saving: %s%s.mfx ",rootname,NodeName);
#ifdef __X__WINDOWS__
#else
       inform(Filename); pause(1); disinform();
#endif
       sprintf(Filename,"%s%s.mfx",rootname,NodeName);
       if( (fo=fopen(Filename,"r")) != NULL){ /* check for old file */
         fclose(fo);
#ifdef __X__WINDOWS__
#else
         id=RequestItemSelect(2,file_check_list,"File exists Save Over ?");
         if(id < 1)goto SKIPIT;
#endif
       }
       if( (fo=fopen(Filename,"wb")) == NULL){   /* now open for reading */
#ifdef __X__WINDOWS__
#else
         Output_Message(G_FILEOPENFAIL);
#endif
         goto SKIPIT;
       }
       fprintf(fo,"FORM");
       pos1=ftell(fo); outlng(FORMsize);
       fprintf(fo,"OFXM");
       fprintf(fo,"AXIS"); outlng(48L); /* size of data chunk */
       outlng(0);  outlng(0);  outlng(0);
       outlng((iSha[0].P[0]-sp->xyz[0])*invert);  /* store the algorithmic texture origin */
       outlng((iSha[0].P[1]-sp->xyz[1])*invert);  /* in the object axis structure         */
       outlng( iSha[0].P[2]-sp->xyz[2]);
       outlng((iSha[0].X[0]-sp->xyz[0])*invert);
       outlng((iSha[0].X[1]-sp->xyz[1])*invert);
       outlng( iSha[0].X[2]-sp->xyz[2]);
       outlng((iSha[0].Y[0]-sp->xyz[0])*invert);
       outlng((iSha[0].Y[1]-sp->xyz[1])*invert);
       outlng( iSha[0].Y[2]-sp->xyz[2]);
       fprintf(fo,"OFFS"); outlng(16L);      /* follow point */
       outlng(3L);
       outlng((sp1->xyz[0]-sp->xyz[0])*invert);
       outlng((sp1->xyz[1]-sp->xyz[1])*invert);
       outlng(sp1->xyz[2]-sp->xyz[2]);
       fprintf(fo,"VERT"); pos2=ftell(fo); outlng(12*newvert);
       WriteVerticesWithOffset(sp->xyz,invert,0,INEDITOR);
       fprintf(fo,"EDGE"); pos3=ftell(fo); outlng( 8*newedge);
       WriteEdges(0);
       if(nf > 0){
         fprintf(fo,"FACE"); pos4=ftell(fo); outlng(20*newface);
         WriteFacesWithMapCheck(0);
       }
       else pos4 = -1;
       if(map > 0){
         fprintf(fo,"TEXT"); pos5=ftell(fo); outlng(0L);
         WriteOffsetTextures(sp->xyz,(unsigned char)map,invert);
         pos6=ftell(fo);
       }
       else pos5 = -1;

       if(xShaders[0].in_use || xShaders[1].in_use ||
          xShaders[2].in_use || xShaders[3].in_use){
         fprintf(fo,"SHDR"); posA=ftell(fo); outlng(0L);
         WriteShaders(1);
         posB=ftell(fo);
       }
       else posA = -1;
       if(nSha > 1){
         fprintf(fo,"AXES"); posC=ftell(fo); outlng(68*(nSha-1));
         WriteOffsetShadersAxes(sp->xyz,invert);
         posD=ftell(fo);
       }
       else posC = -1;

       fprintf(fo,"MATL"); outlng((long)(5*sizeof(long)));
       outlng(4L);
       outlng((long)ShadowFlag);
       outlng(0); outlng(0); outlng(0); /* not used now */
       fprintf(fo,"UNIT"); outlng((long)(7*sizeof(long)+8));
       outlng((long)ruler);
       fwrite(ruler_name,1,8,fo);
       outlng(lrulerx); outlng(lrulery); outlng(lrulerz);
       outlng(rulerx);  outlng(rulery);  outlng(rulerz);
       fprintf(fo,"GRID"); outlng((long)(2*sizeof(long)));
       outlng(grid_size); outlng(grid_size); /* second is unused */
       pose=ftell(fo);  /* the end of the object */
       if(pos2 > 0){fseek(fo,pos2,0); outlng(12*newvert);}
       if(pos3 > 0){fseek(fo,pos3,0); outlng( 8*newedge);}
       if(pos4 > 0){fseek(fo,pos4,0); outlng(20*newface);}
       if(pos5 > 0){fseek(fo,pos5,0); outlng(pos6-pos5-4);} /* not include count */
       if(posA > 0){fseek(fo,posA,0); outlng(posB-posA-4);}
       if(posC > 0){fseek(fo,posC,0); outlng(posD-posC-4);}
       fseek(fo,pos1,0); outlng(pose-8);
       fclose(fo);
     }
     SKIPIT:
     vp=MainVp; while(vp != NULL){ /* reset status */
       if(vp->status == INEDITOR)vp->status=SELECTED;
       vp=vp->last;
     }
   }
   sp=sp->last;
 }
}

static void WriteOffsetTextures(point p, unsigned char nmap, long invert){
 long i,temp,k;
 skel *sp;
 vertex *v;
 char tempstr[256]="VERSION1.TXU";
 i=0;
 sp=FirstSp; while(sp != NULL){ /* set up sk id */
   sp->id=i; i++;
   sp=sp->next;
 }
 fwrite(&nmap,1,1,fo); k=0;
 for(i=0;i<nImaps;i++){
   if(iMap[i].in_use < 0)continue;  /* image map not in use        */
   k++;
   outlng(256);                     /* size of map list string     */
   fwrite(tempstr,1,256,fo);        /* to maintain compatibility   */
   outlng((iMap[i].P[0]-p[0])*invert);
   outlng((iMap[i].P[1]-p[1])*invert);
   outlng( iMap[i].P[2]-p[2]);
   outlng((iMap[i].X[0]-p[0])*invert);
   outlng((iMap[i].X[1]-p[1])*invert);
   outlng( iMap[i].X[2]-p[2]);
   outlng((iMap[i].Y[0]-p[0])*invert);
   outlng((iMap[i].Y[1]-p[1])*invert);
   outlng( iMap[i].Y[2]-p[2]);
   temp=iMap[i].Type; outlng(temp);
   temp=iMap[i].AnimFF; outlng(temp);
   temp=iMap[i].AnimDL; outlng(temp);
   temp=iMap[i].AnimLF; outlng(temp);
   if(iMap[i].Sk != NULL){
     temp=((iMap[i].Sk)->id)+1; outlng(temp);/* 1 is first file brush*/
   }
   else{
     temp=0; outlng(temp);
   }
   temp=iMap[i].Map; outlng(temp);   /* 0 = plane 1 = cyl. map  2/3 = moz */
   temp=32+3+6+9+2;    /* 52 bytes + */
   if(iMap[i].s)temp += 256;
   if(iMap[i].r)temp += 256;
   if(iMap[i].b)temp += 256;
   temp /= 4;   /* make 4 byte long equivalent */
   temp += 5;          outlng(temp); /* for expansion, count - from here */
   temp=iMap[i].Angle; outlng(temp); /* angle for cylinder brushes */
   if(iMap[i].Lock &&
      iMap[i].Vlock[0]->pad >= 0 &&        /* out a tex map if lock */
      iMap[i].Vlock[1]->pad >= 0 &&        /* and if vertices OK    */
      iMap[i].Vlock[2]->pad >= 0 ){
     temp=iMap[i].Lock;  outlng(temp);
     v=iMap[i].Vlock[0];
     temp=(((long)v->id) | ((long)v->pad << 16)); outlng(temp);
     v=iMap[i].Vlock[1];
     temp=(((long)v->id) | ((long)v->pad << 16)); outlng(temp);
     v=iMap[i].Vlock[2];
     temp=(((long)v->id) | ((long)v->pad << 16)); outlng(temp);
   }
   else{  /* map is not locked */
     temp=0; outlng(temp); outlng(temp); outlng(temp); outlng(temp);
   }
   fwrite(iMap[i].N,1,32,fo);
   fwrite(&iMap[i].s,1,1,fo);
   fwrite(&iMap[i].r,1,1,fo);
   fwrite(&iMap[i].b,1,1,fo);
   outsht(iMap[i].sp);
   outsht(iMap[i].rp);
   outsht(iMap[i].bp);
   fwrite(iMap[i].d_colour,1,3,fo);
   fwrite(iMap[i].s_colour,1,3,fo);
   fwrite(iMap[i].a_colour,1,3,fo);
   if(iMap[i].s)fwrite(iMap[i].S,1,256,fo);
   if(iMap[i].r)fwrite(iMap[i].R,1,256,fo);
   if(iMap[i].b)fwrite(iMap[i].B,1,256,fo);
   outsht(0); /* to make up to multiple of 4 bytes and possible expansion */
 }
 if(k != (long)nmap){
#ifdef __X__WINDOWS__
#else
   inform("Error in writing maps - file corrupt");
   pause(200); disinform();
#endif
 }
}

static void WriteOffsetShadersAxes(point p, long invert){
 long i;
 for(i=1;i<nSha;i++){
   fwrite(iSha[i].N,1,32,fo);
   outlng((iSha[i].P[0]-p[0])*invert);
   outlng((iSha[i].P[1]-p[1])*invert);
   outlng( iSha[i].P[2]-p[2]);
   outlng((iSha[i].X[0]-p[0])*invert);
   outlng((iSha[i].X[1]-p[1])*invert);
   outlng( iSha[i].X[2]-p[2]);
   outlng((iSha[i].Y[0]-p[0])*invert);
   outlng((iSha[i].Y[1]-p[1])*invert);
   outlng( iSha[i].Y[2]-p[2]);
 }
}

static void WriteShaders(short status){
 long i,lp;
 outlng(4);
 for(i=0;i<4;i++){
   outlng(xShaders[i].in_use);
   if(xShaders[i].in_use){
     lp=strlen(xShaders[i].parameters)+1; /* include the trailing zero */
     outlng(128); outlng(lp);
     fwrite(xShaders[i].name,1,128,fo);
     fwrite(xShaders[i].parameters,1,lp,fo);
   }
 }
}
