/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  BLOBBY.CPP  */

#include "vtk.hh"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#define UNIT         32768L
#define DESELECTED       0
#define SELECTED         1
#define INEDITOR         6

#ifdef __cplusplus
extern "C" {
#endif

#include "struct.h"
#include "dstruct.h"

#ifdef __cplusplus
}
#endif

#include "blobby.h"

#include "sfx2vtk.h"
#include "sfx2vtk.cpp"

static HWND      hParent;
static HINSTANCE hThisInstance;

static BOOL CALLBACK SfxDlgProc(HWND hwnd, UINT msg,
                                WPARAM wparam, LPARAM lparam);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=(HINSTANCE)hDLL;
      if(hDLL == NULL)MessageBeep(MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}


extern "C" BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 HCURSOR hSave;
 static struct bp {
  float f1,f2;
  int   type;
  int   i,j,k;
 } d;
 lpEVI=lpevi;
 hParent=parent_window;
 d.type=1;
 d.f1=0.15;
 d.f2=0.1;
 d.i=25;
 d.j=25;
 d.k=25;
 if(DialogBoxParam(hThisInstance,MAKEINTRESOURCE(DLG_VTK),hParent,
             (DLGPROC)SfxDlgProc,(LPARAM)&d) == FALSE)return FALSE;
 hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
 {
  SfxBound  sfxbound;
  SfxReader sfxinput;
  SfxWriter sfxoutput;
  SfxPointReader sfxpointinput;
  SfxEdgeReader sfxedgeinput;

  double scale=sfxbound.GetMaximumSize();
  d.f1 /= scale;
  d.f2 /= scale;
  d.f1 *= RulerScale/(float)UNIT;
  d.f2 *= RulerScale/(float)UNIT;
  sfxpointinput.SetScale(1.0/scale);
  sfxedgeinput.SetScale(1.0/scale);
  sfxinput.SetScale(1.0/scale);

  vtkImplicitModeller *imp = new vtkImplicitModeller;
  if(d.type == 0)imp->SetInput(sfxpointinput.GetOutput());
  if(d.type == 1)imp->SetInput(sfxedgeinput.GetOutput());
  if(d.type == 2)imp->SetInput(sfxinput.GetOutput());
  imp->SetSampleDimensions(d.i,d.j,d.k);
  imp->SetMaximumDistance(d.f1);

  vtkContourFilter *cf = new vtkContourFilter;
  cf->SetInput(imp->GetOutput());
  cf->SetValue(0,d.f2);

  vtkCleanPolyData *cp = new vtkCleanPolyData;
  cp->SetTolerance(0.0001);
  cp->SetInput(cf->GetOutput());

  vtkTriangleFilter *tf = new vtkTriangleFilter;
  tf->SetInput(cp->GetOutput());

  sfxoutput.SetScale(scale);
  sfxoutput.SetInput(tf->GetOutput());
  sfxoutput.Write();

  delete tf;
  delete cp;
  delete cf;
  delete imp;

 }
 SetCursor(hSave);
 return TRUE;
}

static BOOL CALLBACK SfxDlgProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam){
 BOOL err;
 double f;
 int i;
 static struct bp {
  float f1,f2;
  int   type;
  int   i,j,k;
 } *dp;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     dp=(struct bp *)lparam;
     SetDlgItemInt(hwnd,DLG_VTK_XRES,dp->i,FALSE);
     SetDlgItemInt(hwnd,DLG_VTK_YRES,dp->j,FALSE);
     SetDlgItemInt(hwnd,DLG_VTK_ZRES,dp->k,FALSE);
     sprintf(str,"%.2lf",dp->f1);
     SetDlgItemText(hwnd,DLG_VTK_MAX_SAMPLING_DISTANCE,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->f2);
     SetDlgItemText(hwnd,DLG_VTK_CONTOUR_DISTANCE,(LPCTSTR)str);
     if(dp->type == 0)
       SendDlgItemMessage(hwnd,DLG_VTK_TYPE0,BM_SETCHECK,TRUE,0);
     if(dp->type == 1)
       SendDlgItemMessage(hwnd,DLG_VTK_TYPE1,BM_SETCHECK,TRUE,0);
     if(dp->type == 2)
       SendDlgItemMessage(hwnd,DLG_VTK_TYPE2,BM_SETCHECK,TRUE,0);
     CentreDlgOnS(hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          i=GetDlgItemInt(hwnd,DLG_VTK_XRES,&err,FALSE);
          if(err)dp->i=i;
          i=GetDlgItemInt(hwnd,DLG_VTK_YRES,&err,FALSE);
          if(err)dp->j=i;
          i=GetDlgItemInt(hwnd,DLG_VTK_ZRES,&err,FALSE);
          if(err)dp->k=i;
          if(GetDlgItemText(hwnd,DLG_VTK_MAX_SAMPLING_DISTANCE,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->f1=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_CONTOUR_DISTANCE,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->f2=atof(str);
          if(SendDlgItemMessage(hwnd,DLG_VTK_TYPE0,BM_GETCHECK,0,0))
             dp->type=0;
          else if(SendDlgItemMessage(hwnd,DLG_VTK_TYPE1,BM_GETCHECK,0,0))
             dp->type=1;
          else if(SendDlgItemMessage(hwnd,DLG_VTK_TYPE2,BM_GETCHECK,0,0))
             dp->type=2;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

