/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file D3DVIEW.CPP */

#define INITGUID
#include <d3drmwin.h>
#include <windows.h>
#include <direct.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

#include "struct.h"
#include "dstruct.h"

#ifdef __cplusplus
}
#endif

#include "glmenu.h"
#include "d3dview.h"
#include "..\openfx\about.h"

#define PLANE            0
#define CYLINDER         1
#define PLANE_MOZIAC     2
#define CYLINDER_MOZIAC  3
#define MAP_BY_VERTEX    4
#define MAP_SPHERICAL    5
#define MAX_FILE       255
#define MAX_DIR        255
#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))

#define IDM_WINDOW_SHOWOPENGL   17130

#define StartWait() hcurSave = SetCursor(LoadCursor(NULL,IDC_WAIT))
#define EndWait()   SetCursor(hcurSave)

#define WM_INIT       (WM_USER+1)
#define WM_UPDATE     (WM_USER+2)
#define WM_UPDATE_MAP (WM_USER+3)
#define WM_LOAD_MAPS  (WM_USER+4)

typedef float normal[3];

static void RegisterD3D(HINSTANCE);
static void UnregisterD3D(HINSTANCE);
static LONG WINAPI D3dWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam);
static void LoadPrefs(void);
static PSTR FileInPath(PSTR pstrPath);
static BOOL CALLBACK PrefsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam);
static void LocalCommands(WORD cmd,HWND hWnd);
LPGUID FindDevice(D3DCOLORMODEL cm);
static BOOL Render(void);

static HMENU     D3dMenu=NULL;
static HWND      D3dWnd=NULL,D3dWndParent;
static HINSTANCE d3dInstance;
static CHAR      szAppName[]="OFX:D3Dview";
static HCURSOR   hcurSave=NULL;
static DWORD     fdwStyle = WS_CHILD | !WS_CAPTION | !WS_SYSMENU |
                            !WS_MAXIMIZEBOX | !WS_THICKFRAME | WS_BORDER |
                            WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
static double  *PerspTheta,*PerspFi;
static float up_angle=0.0,round_angle=0.0;
static float gbRed=0.0,gbGreen=0.1,gbBlue=0.3;
static float near_plane,far_plane;
static int flat=0,bounding_box=0;
static BOOL  Mapped=FALSE,Update=TRUE,Fastest=FALSE,Clamped=FALSE,
             Cmaterial=TRUE,AttenuateLight=TRUE,
             ScrollBars=FALSE;
static char gszBMPfile[MAX_FILE],gszBMPdir[MAX_DIR];
unsigned char *pixels = NULL;

typedef struct _AppInfo
{
    LPDIRECT3DRMFRAME scene, camera, object;
    LPDIRECT3DRMDEVICE dev;
    LPDIRECT3DRMVIEWPORT view;
    D3DRMCOLORMODEL model;
    BOOL bMinimized;
} AppInfo;

AppInfo *active_window = NULL;

LPDIRECT3DRM lpD3DRM = 0;
LPDIRECTDRAWCLIPPER lpDDClipper = 0;
BOOL bQuit = FALSE;

static void Idle();
char* MyErrorToString(HRESULT error);
void __cdecl Msg( LPSTR fmt, ... );
char* LSTRRCHR( const char* lpString, int bChar );
static BOOL CreateDevice(HWND, AppInfo *);
static void MakeD3dModel(AppInfo *);
static BOOL ResizeViewport(HWND win, AppInfo* info, int width, int height);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BCC__
int APIENTRY DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
      RegisterD3D(hDLL);
      break;
    case DLL_PROCESS_DETACH:
// Don't do this for d3d Violation;
//      UnregisterD3D(hDLL);
      break;
  }
  return TRUE;
}

static void UnregisterD3D(HINSTANCE hInstance){
 if(D3dWnd != NULL)DestroyWindow(D3dWnd);
 UnregisterClass(szAppName,hInstance);
}

static void RegisterD3D(HINSTANCE hInstance){
 WNDCLASS   wc;
 wc.style         = 0;
 wc.lpfnWndProc   = (WNDPROC)D3dWndProc;
 wc.cbClsExtra    = 0;
 wc.cbWndExtra    = 0;
 wc.hInstance     = hInstance;
 wc.hIcon         = LoadIcon(hInstance,"GLVIEW_ICON");
 wc.hCursor       = LoadCursor(hInstance,MAKEINTRESOURCE(IDC_HAND));
 wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
 wc.lpszMenuName  = NULL;
 wc.lpszClassName = szAppName;
 d3dInstance=hInstance;
 if (!RegisterClass (&wc) )return;
 return;
}

extern "C" HWND _AttachViewOpenGL(
     HWND parent,
     char *tt,
     LPRECT pr,
     double *lpPerspTheta,
     double *lpPerspFi,
     X__STRUCTURE *lpevi
                              ){
 RECT rc;
 HWND hWnd;
 lpEVI=lpevi;
 D3dWndParent=parent;
 PerspTheta=lpPerspTheta;
 PerspFi=lpPerspFi;
 GetModuleFileName(d3dInstance,gszBMPdir,MAX_DIR);
 *FileInPath(gszBMPdir) = '\0';
 gszBMPfile[0]='\0'; strcat(gszBMPdir,"gif");
 HRESULT  rval = Direct3DRMCreate(&lpD3DRM);
 if(rval != D3DRM_OK) {
   Msg("Failed to create Direct3DRM.\n%s", MyErrorToString(rval));
   return NULL;
 }
 hWnd = CreateWindow(
             szAppName,
             " ",
             fdwStyle,
             pr->left,
             pr->top,
             pr->right,
             pr->bottom,
             parent,
             NULL,
             d3dInstance,
             NULL);
 if (!hWnd)return NULL;
 D3dWnd=hWnd;
 if(FAILED(DirectDrawCreateClipper(0, &lpDDClipper, NULL))) {
    return NULL;
 }
 if(FAILED(lpDDClipper->SetHWnd(0,hWnd))){
   RELEASE(lpDDClipper);
   return NULL;
 }
 active_window = (AppInfo*)X__Malloc(sizeof(AppInfo));
 memset(active_window,0,sizeof(AppInfo));
// active_window->model = D3DCOLOR_MONO;
 active_window->model = D3DCOLOR_RGB;
 if(!CreateDevice(hWnd,active_window))return NULL;
 SendMessage(hWnd, WM_INIT, 0, 0L);
 ShowWindow(hWnd,SW_SHOW);
 UpdateWindow(hWnd);
 return hWnd;
}

extern "C" void _DetachViewOpenGL(void){
 if(D3dWnd != NULL)DestroyWindow(D3dWnd);
 return;
}

static LONG WINAPI D3dWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam){
 static BOOL bBusy=FALSE;
 float old;
 int i,iPos,iMax,iMin,dn;
 LONG    lRet = 1;
 static BOOL flag=FALSE,mmoved=FALSE,bTrackPan=FALSE;
 static long xs,ys;
 AppInfo *info=active_window;
 switch (uMsg){
   case WM_SIZE:
     if(info != NULL){
       int width = LOWORD(lParam);
       int height = HIWORD(lParam);
       if (width && height) {
         if (!ResizeViewport(hWnd,info,width,height)) {
            bQuit = TRUE;
            lRet=0;
            break;
         }
         active_window->bMinimized = FALSE;
       }
       else{
         active_window->bMinimized = TRUE;
       }
     }
     lRet=0;
     break;
   case WM_ACTIVATE:
     if(info != NULL){
       LPDIRECT3DRMWINDEVICE windev;
       if(SUCCEEDED(info->dev->QueryInterface(IID_IDirect3DRMWinDevice,
           (void **)&windev))){
         if (FAILED(windev->HandleActivate(wParam)))
                    Msg("Failed to handle WM_ACTIVATE.\n");
         windev->Release();
       }
       else{
         Msg("Failed to create Windows device to handle WM_ACTIVATE.\n");
       }
     }
     break;
   case WM_PAINT:
     if(info != NULL){
       RECT r;
       PAINTSTRUCT ps;
       LPDIRECT3DRMWINDEVICE windev;
       if(GetUpdateRect(hWnd,&r,FALSE)){
         BeginPaint(hWnd,&ps);
         if(SUCCEEDED(info->dev->QueryInterface(IID_IDirect3DRMWinDevice,
             (void **) &windev))){
           if(FAILED(windev->HandlePaint(ps.hdc)))
              Msg("Failed to handle WM_PAINT.\n");
           windev->Release();
         }
         else{
           Msg("Failed to create Windows device to handle WM_PAINT.\n");
         }
         EndPaint(hWnd,&ps);
       }
       if(!bBusy){
         bBusy=TRUE;
         if(!Render())Msg("Rendering failed.\n");
         bBusy=FALSE;
       }
     }
     else return DefWindowProc(hWnd,uMsg,wParam,lParam);
     break;
   case WM_CREATE:{
       INT menu_flags;
       HDC hDC;
       HMENU hmenu;
       char str[128];
       LoadPrefs();
       up_angle=*PerspTheta;
       round_angle=*PerspFi;
       if(ScrollBars){
         SetScrollRange(hWnd, SB_VERT, -89,89,TRUE);
         SetScrollRange(hWnd, SB_HORZ, -179,179,TRUE);
         SetScrollPos(hWnd,SB_HORZ,(int)round_angle,TRUE);
         SetScrollPos(hWnd,SB_VERT,(int)up_angle,TRUE);
       }
       else{
         SetScrollRange(hWnd, SB_VERT, 0,0,TRUE);
         SetScrollRange(hWnd, SB_HORZ, 0,0,TRUE);
         SetScrollPos(hWnd,SB_HORZ,0,TRUE);
         SetScrollPos(hWnd,SB_VERT,0,TRUE);
       }
       hmenu=GetMenu(D3dWndParent);
       D3dMenu=CreatePopupMenu();
       AppendMenu(D3dMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_TWOSIDE,"Update");
       AppendMenu(D3dMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_HELP,"Load...");
       AppendMenu(D3dMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_SCROLLBARS,"Save...");
       LoadString(d3dInstance,IDS_ABOUT,str,127);
       AppendMenu(D3dMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_ABOUT,str);
       LoadString(d3dInstance,IDS_MENU,str,127);
       InsertMenu(hmenu,GLMENU_POSITION,MF_POPUP|MF_BYPOSITION,
                 (UINT)D3dMenu,str);
       DrawMenuBar(D3dWndParent);
     }
     break;
   case WM_LBUTTONDOWN:
     mmoved=FALSE;
     xs=LOWORD(lParam);
     ys=HIWORD(lParam);
     SetCapture(hWnd);
     bTrackPan=TRUE;
     break;
   case WM_MOUSEMOVE:
     if(bTrackPan){
       long dx,dy;
       dx=(short)LOWORD(lParam)-xs; xs=(short)LOWORD(lParam);
       dy=(short)HIWORD(lParam)-ys; ys=(short)HIWORD(lParam);
       if(labs(dx) > 0 || labs(dy) > 0){
         mmoved=TRUE;
         round_angle=round_angle+(double)dx;
         while(round_angle < -180.0)round_angle += 360.0;
         while(round_angle >  180.0)round_angle -= 360.0;
         *PerspFi = (double)round_angle;
         SetScrollPos(hWnd,SB_HORZ,(int)round_angle,TRUE);
         up_angle=up_angle+(double)dy;
         *PerspTheta = (double)up_angle;
         if(up_angle > -90 && up_angle < 90)
           SetScrollPos(hWnd,SB_VERT,(int)up_angle,TRUE);
         if(info->object != NULL){
           info->object->AddRotation(D3DRMCOMBINE_REPLACE,
             D3DVAL(1.0f),D3DVAL(0.0f),D3DVAL(0.0f),
             D3DVAL(-3.1415926f*(*PerspTheta)/180.0f));
           info->object->AddRotation(D3DRMCOMBINE_BEFORE,
             D3DVAL(0.0f),D3DVAL(1.0f),D3DVAL(0.0f),
             D3DVAL(-3.1415926f*(*PerspFi)/180.0f));
           info->object->SetPosition(info->scene,
           D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0));
         }
         InvalidateRect(hWnd, NULL, FALSE);
       }
     }
     break;
   case WM_LOAD_MAPS:
     SendMessage(hWnd,WM_COMMAND,IDM_GLVIEW_MAP,0);
     break;
   case WM_UPDATE_MAP:
     if(!Mapped)break;
     if(!Update)break;
   case WM_LBUTTONUP:
       if(!Render())Msg("Rendering failed.\n");
     if(bTrackPan){
       ReleaseCapture();
       bTrackPan=FALSE;
     }
     if(mmoved){
       mmoved=FALSE;
//       InvalidateRect(hWnd, NULL, FALSE); // using wireframe
       break;
     }
     mmoved=FALSE;
     break;
   case WM_UPDATE:
//     InvalidateRect(hWnd, NULL, FALSE);
     break;
   case WM_INIT:
     break;
  case WM_VSCROLL:
    GetScrollRange(hWnd, SB_VERT, &iMin, &iMax);
    iPos = GetScrollPos(hWnd, SB_VERT);
    switch (LOWORD(wParam)) {
      case SB_ENDSCROLL:
         dn=0;
        if(flag)InvalidateRect(hWnd, NULL, FALSE);
        flag=FALSE;
        break;
      case SB_LINEDOWN:      dn =  1;    break;
      case SB_LINEUP:        dn = -1;    break;
      case SB_PAGEDOWN:      dn =  5;    break;
      case SB_PAGEUP:        dn = -5;    break;
      case SB_THUMBTRACK:
      case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wParam)-iPos;    break;
      default:               dn = 0;
    }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      SetScrollPos(hWnd, SB_VERT, iPos + dn, TRUE);
      old=up_angle;
      up_angle=(float)(iPos+dn);
      *PerspTheta = (double)up_angle;
      if(old != up_angle){
        if(!flag && (GetKeyState(VK_CONTROL) & 0x8000))flag=TRUE;
        if(!flag)InvalidateRect(hWnd, NULL, FALSE);
      }
    }
    break;
  case WM_HSCROLL:
    GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
    iPos = GetScrollPos(hWnd, SB_HORZ);
    switch (LOWORD(wParam)) {
      case SB_ENDSCROLL:
        if(flag)InvalidateRect(hWnd, NULL, FALSE);
        flag=FALSE;
        dn=0;
        break;
      case SB_LINEDOWN:      dn =  1;     break;
      case SB_LINEUP:        dn = -1;     break;
      case SB_PAGEDOWN:      dn =  5;     break;
      case SB_PAGEUP:        dn = -5;     break;
      case SB_THUMBTRACK:
      case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wParam) - iPos; break;
      default:               dn = 0;
    }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      SetScrollPos(hWnd, SB_HORZ, iPos + dn, TRUE);
      old=round_angle;
      round_angle=(float)(iPos+dn);
      *PerspFi = (double)round_angle;
      if(old != round_angle){
        if(!flag && (GetKeyState(VK_CONTROL) & 0x8000))flag=TRUE;
        if(!flag)InvalidateRect(hWnd, NULL, FALSE);
      }
    }
    break;
   case WM_COMMAND:
     LocalCommands(LOWORD(wParam),hWnd);
     break;
   case WM_CLOSE:
     PostMessage(D3dWndParent,WM_COMMAND,IDM_WINDOW_SHOWOPENGL,0);
     break;
   case WM_DESTROY:{
       char *w, yesyes[]="YES", nono[]="NO",str[64];
       D3dWnd=NULL;
       RemoveMenu(GetMenu(D3dWndParent),GLMENU_POSITION,MF_BYPOSITION);
       if(D3dMenu != NULL)DestroyMenu(D3dMenu);
       DrawMenuBar(D3dWndParent);
       D3dMenu=NULL;
       RELEASE(active_window->scene);
       RELEASE(active_window->camera);
       RELEASE(active_window->view);
       RELEASE(active_window->dev);
       RELEASE(lpD3DRM);
       RELEASE(lpDDClipper);
     }
     break;
   default:
     lRet = DefWindowProc(hWnd, uMsg, wParam, lParam);
   break;
 }
 return lRet;
}

static BOOL RebuildDevice(HWND win, AppInfo* info, int width, int height){
    HRESULT rval;
    int old_dither = info->dev->GetDither();
    D3DRMRENDERQUALITY old_quality = info->dev->GetQuality();
    int old_shades = info->dev->GetShades();

    RELEASE(info->view);
    RELEASE(info->dev);
    rval = lpD3DRM->CreateDeviceFromClipper(lpDDClipper, FindDevice(info->model),
                                width, height, &info->dev);
    if (rval != D3DRM_OK) {
        Msg("Creating a device from HWND failed while rebuilding device.\n%s",
            MyErrorToString(rval));
        return FALSE;
    }

    if (FAILED(info->dev->SetDither(old_dither)))
        goto generic_error;
    if (FAILED(info->dev->SetQuality(old_quality)))
        goto generic_error;
    if (FAILED(info->dev->SetShades(old_shades)))
        goto generic_error;
    width = info->dev->GetWidth();
    height = info->dev->GetHeight();
    if (FAILED(lpD3DRM->CreateViewport(info->dev, info->camera,
                          0, 0, width, height, &info->view)))
                          goto generic_error;
    if (FAILED(info->view->SetBack(D3DVAL(400.0))))
        goto generic_error;
    return TRUE;
generic_error:
    Msg("A failure occurred while rebuilding the device.\n");
    return FALSE;
}

static BOOL ResizeViewport(HWND win, AppInfo* info, int width, int height){
    int view_width = info->view->GetWidth();
    int view_height = info->view->GetHeight();
    int dev_width = info->dev->GetWidth();
    int dev_height = info->dev->GetHeight();

    if (view_width == width && view_height == height)
        return TRUE;
    
    if (width <= dev_width && height <= dev_height) {
        RELEASE(info->view);
        if(FAILED(lpD3DRM->CreateViewport(info->dev,
           info->camera, 0, 0, width, height, &info->view)))
            goto generic_error;
        if (FAILED(info->view->SetBack(D3DVAL(400.0))))
            goto generic_error;
    }

    if (!RebuildDevice(win, info, width, height))
        return FALSE;
    return TRUE;
generic_error:
    Msg("A failure occurred while resizing the viewport.\n");
    return FALSE;
}

static BOOL CreateScene(AppInfo* info)
{
    LPDIRECT3DRMFRAME frame = NULL;
    LPDIRECT3DRMFRAME light = NULL;
    LPDIRECT3DRMMESHBUILDER builder = NULL;
    LPDIRECT3DRMLIGHT light1 = NULL;
    LPDIRECT3DRMLIGHT light2 = NULL;
    LPDIRECT3DRMMATERIAL mat = NULL;
    LPDIRECT3DRMTEXTURE lpTx = NULL;
    HRESULT rval;

    if (FAILED(lpD3DRM->CreateFrame(NULL, &info->scene)))
        goto generic_error;
    if (FAILED(lpD3DRM->LoadTexture("c:\\book\\cloud.bmp",&lpTx)))
        goto generic_error;
    info->scene->SetSceneBackgroundRGB(D3DVAL(0.5),D3DVAL(0.0),D3DVAL(0.0));
    info->scene->SetSceneBackgroundImage(lpTx);
    RELEASE(lpTx);

    if (FAILED(lpD3DRM->CreateLightRGB(D3DRMLIGHT_DIRECTIONAL,
      D3DVAL(1.0), D3DVAL(1.0), D3DVAL(1.0), &light1)))
        goto generic_error;
    if (FAILED(lpD3DRM->CreateLightRGB(D3DRMLIGHT_AMBIENT,
      D3DVAL(0.1), D3DVAL(0.1), D3DVAL(0.1), &light2)))
        goto generic_error;
    if (FAILED(lpD3DRM->CreateFrame(info->scene, &light)))
        goto generic_error;
    if (FAILED(light->SetPosition(info->scene,
      D3DVAL(2.0), D3DVAL(2.0), D3DVAL(5.0))))
        goto generic_error;
    if (FAILED(light->SetOrientation(info->scene,
      D3DVAL(-1.0), D3DVAL(-1.0), D3DVAL(1.0),
      D3DVAL(0.0), D3DVAL(1.0), D3DVAL(0.0))))
        goto generic_error;
    if (FAILED(light->AddLight(light1)))
        goto generic_error;
    RELEASE(light1);
    if (FAILED(info->scene->AddLight(light2)))
        goto generic_error;
    RELEASE(light2);
    RELEASE(light);
    if (FAILED(lpD3DRM->CreateFrame(info->scene, &frame)))
        goto generic_error;
    frame->AddRotation(D3DRMCOMBINE_REPLACE,
         D3DVAL(1.0f),D3DVAL(0.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspTheta)/180.0f));
    frame->AddRotation(D3DRMCOMBINE_BEFORE,
         D3DVAL(0.0f),D3DVAL(1.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspFi)/180.0f));
    frame->SetPosition(info->scene,
         D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0));
#if 0
    if (FAILED(lpD3DRM->CreateMeshBuilder(&builder)))
        goto generic_error;
    rval = builder->Load("torus.x", NULL, D3DRMLOAD_FROMFILE,
        NULL, NULL);
    if (rval != D3DRM_OK) {
        Msg("Failed to load.\n%s", MyErrorToString(rval));
        goto ret_with_error;
    }
    if (FAILED(builder->SetColorRGB(D3DVAL(0.8), D3DVAL(0.8), D3DVAL(0.8))))
        goto generic_error;
    if (FAILED(lpD3DRM->CreateMaterial(D3DVAL(10.0), &mat)))
        goto generic_error;
    if (FAILED(builder->SetMaterial(mat)))
        goto generic_error;
    RELEASE(mat);
    if (FAILED(frame->AddVisual(builder)))
        goto generic_error;
    RELEASE(builder);
#endif
    // record the objexts frame but don't release it  // RELEASE(frame);
    info->object=frame;
    if (FAILED(lpD3DRM->CreateFrame(info->scene, &info->camera)))
        goto generic_error;
    if (FAILED(info->camera->SetPosition(info->scene,
       D3DVAL(0.0), D3DVAL(0.0), D3DVAL(0.0))))
        goto generic_error;
    return TRUE;
generic_error:
    Msg("A failure occurred while creating the scene.\n");
ret_with_error:
    RELEASE(frame);
    RELEASE(light);
    RELEASE(builder);
    RELEASE(light1);
    RELEASE(light2);
    RELEASE(mat);
    RELEASE(lpTx);
    return FALSE;
}

static BOOL Render() {
    if (FAILED(active_window->scene->Move(D3DVAL(1.0))))
        return FALSE;
    if (FAILED(active_window->view->Clear()))
        return FALSE;
    if (FAILED(active_window->view->Render(active_window->scene)))
        return FALSE;
    if (FAILED(active_window->dev->Update()))
        return FALSE;
    return TRUE;
}

void DeleteObject(AppInfo* info){
 if(info->object == NULL)return;
 info->scene->DeleteChild(info->object);
 info->object=NULL;
 return;
#if 0
 LPDIRECT3DRMFRAME child,frame;
 LPDIRECT3DRMFRAMEARRAY frames;
 LPDIRECT3DRMMESHBUILDER mesh;
 frame=info->scene;
 frame->GetChildren(&frames);
 if(frames){
   frames->GetElement(frames->GetSize() - 2,&child);
   if(child){
     frame->DeleteChild(child);
   }
   frames->Release();
 }
#endif
}

void OpenNewModel(char *file, AppInfo *info){
 LPDIRECT3DRMMATERIAL mat;
 LPDIRECT3DRMMESHBUILDER builder;
 LPDIRECT3DRMFRAME frame;
 HRESULT rval;
 if (FAILED(lpD3DRM->CreateMeshBuilder(&builder))) {
   Msg("Failed the create a builder for the new mesh.\n");
   return;
 }
 rval = builder->Load(file, NULL, D3DRMLOAD_FROMFILE, NULL, NULL);
 if (rval != D3DRM_OK) {
   Msg("Loading %s failed.\n%s", file, MyErrorToString(rval));
   builder->Release();
   return;
 }
 if (FAILED(lpD3DRM->CreateMaterial(D3DVAL(10.0), &mat))){;}
 if (FAILED(builder->SetMaterial(mat))){;}
 RELEASE(mat);
 if(FAILED(lpD3DRM->CreateFrame(info->scene, &frame))){
   Msg("Placing the mesh in the scene failed.\n");
   builder->Release();
   return;
 }
 if(FAILED(frame->AddVisual(builder))){
   Msg("Placing the mesh in the scene failed.\n");
   builder->Release();
   return;
 }
 frame->AddRotation(D3DRMCOMBINE_REPLACE,
         D3DVAL(1.0f),D3DVAL(0.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspTheta)/180.0f));
 frame->AddRotation(D3DRMCOMBINE_BEFORE,
         D3DVAL(0.0f),D3DVAL(1.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspFi)/180.0f));
 frame->SetPosition(info->scene,
         D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0));
 if (FAILED(frame->SetPosition(info->scene,
      D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0)))){
   Msg("Placing the mesh in the scene failed.\n");
   builder->Release();
   return;
 }
 info->object=frame;
// to remove the backgound image
// info->scene->SetSceneBackgroundImage(NULL);
 builder->Release();
// don't release the frame
// frame->Release();
}

void SaveModel(char *file, AppInfo *info){
 LPDIRECT3DRMVISUAL visual;
 LPDIRECT3DRMVISUALARRAY visuals;
 LPDIRECT3DRMMESHBUILDER builder;
 info->object->GetVisuals(&visuals);
 if(visuals){
   visuals->GetElement(visuals->GetSize()-1,&visual);
   if(visual){
     if(SUCCEEDED(visual->QueryInterface(IID_IDirect3DRMMeshBuilder,
                (void **) &builder))){
//       builder->Save(file,D3DRMXOF_TEXT,D3DRMXOFSAVE_NORMALS);
       builder->Save(file,D3DRMXOF_TEXT,D3DRMXOFSAVE_ALL);
       builder->Release();
     }
     visual->Release();
   }
   visuals->Release();
 }
}


#include "d3dview0.cpp"
#include "d3dview1.cpp"
#include "d3dview2.cpp"

