/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// d3dview1.cpp

#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define VECCOPY(a,c)    { c[0]=a[0]; c[1]=a[1]; c[2]=a[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }

#define HIDDEN 2

static BOOL normalise(float *v){
 float n,nn;
 n= (float)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return FALSE;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE(n,v,v)
 return TRUE;
}

static BOOL Normalize(point x,point y,point z,float *n){
 float dz[3],dy[3];
 VECSUB((float)y,(float)x,dy)
 VECSUB((float)z,(float)x,dz)
 CROSS(dy,dz,n)
 if(normalise(n))return TRUE;
 return FALSE;
}

static BOOL inview(vertex  *vp){
 if(vp->status == HIDDEN)return 0;
 if( ((vp->xyz[0] > TVpointX-TVsizeX) && (vp->xyz[0] < TVpointX+2*TVsizeX))
  && ((vp->xyz[1] > TVpointY-TVsizeX) && (vp->xyz[1] < TVpointY+2*TVsizeY))
  && ((vp->xyz[2] > TVpointZ-TVsizeX) && (vp->xyz[2] < TVpointZ+2*TVsizeZ))
   )return TRUE;
 return FALSE;
}

#if 0
static void MakeD3dModel(AppInfo *info){
// double sided by copying second side   (if comments removed)
 unsigned long *ff;
 D3DVECTOR *vv;
 float x,y,z,scale,n[3];
 face *fp;
 vertex *v,*v1,*v2,*v0;
 long i,j,kv,kn,kf;
 if(Nface == 0)return;
 scale=10.0/TVsizeX;
 fp=MainFp; kn=0; for(i=0;i<Nface;i++,fp++){
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   if(inview(v0) || inview(v1) || inview(v2))kn++;
 }
 if((vv = (D3DVECTOR *)X__Malloc(3*kn*sizeof(D3DVECTOR))) == NULL){
   return;
 }
 if((ff = (unsigned long *)X__Malloc((8*kn+1)*sizeof(long))) == NULL){
   X__Free(vv);
   return;
 }
 fp=MainFp; kv=kn=kf=0; for(i=0;i<Nface;i++,fp++){
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   if(inview(v0) || inview(v1) || inview(v2)){
     if((fp->matcol[1] & 0xf0) == 0){ /* Not glass */
       ff[kf++]=3;
       for(j=0;j<3;j++){
         v=(MainVp+fp->V[j]);
         x=((float)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
         y=((float)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
         z=((float)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
         vv[kv].x=D3DVAL(x);
         vv[kv].y=D3DVAL(z);
         vv[kv].z=D3DVAL(-y);
         ff[kf++]=kv;
         kv++;
       }
//       ff[kf++]=3;     // second side
//       ff[kf++]=kv-3;
//       ff[kf++]=kv-1;
//       ff[kf++]=kv-2;
       kn++;
     }
   }
 }
 ff[kf++]=0;
 LPDIRECT3DRMMESHBUILDER builder;
 LPDIRECT3DRMFRAME frame;
 lpD3DRM->CreateMeshBuilder(&builder);
 builder->AddFaces(kv,vv,0,NULL,ff,NULL);
 // generate normals
 builder->GenerateNormals();

 builder->SetColor(D3DRMCreateColorRGB(D3DVAL(1.0),D3DVAL(0.3),D3DVAL(0.2)));
 lpD3DRM->CreateFrame(info->scene,&frame);
 frame->AddVisual(builder);
 builder->Release();
 frame->AddRotation(D3DRMCOMBINE_REPLACE,
         D3DVAL(1.0f),D3DVAL(0.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspTheta)/180.0f));
 frame->AddRotation(D3DRMCOMBINE_BEFORE,
         D3DVAL(0.0f),D3DVAL(1.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspFi)/180.0f));
 frame->SetPosition(info->scene,
         D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0));
 info->object=frame;
 X__Free(vv);
 X__Free(ff);
}
#endif

#if 0
static void MakeD3dModel(AppInfo *info){
// single sided - with normals
 unsigned long *ff;
 D3DVECTOR *vv,*nn;
 float x,y,z,scale,n[3];
 face *fp;
 vertex *v,*v1,*v2,*v0;
 long i,j,kv,kn,kf;
 if(Nface == 0)return;
 scale=10.0/TVsizeX;
 fp=MainFp; kn=0; for(i=0;i<Nface;i++,fp++){
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   if(inview(v0) || inview(v1) || inview(v2))kn++;
 }
 if((nn = (D3DVECTOR *)X__Malloc(3*kn*sizeof(D3DVECTOR))) == NULL)
   return;
 if((vv = (D3DVECTOR *)X__Malloc(3*kn*sizeof(D3DVECTOR))) == NULL){
   X__Free(nn);
   return;
 }
 if((ff = (unsigned long *)X__Malloc((7*kn+1)*sizeof(long))) == NULL){
   X__Free(nn);
   X__Free(vv);
   return;
 }
 fp=MainFp; kv=kn=kf=0; for(i=0;i<Nface;i++,fp++){
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   if(inview(v0) || inview(v1) || inview(v2)){
     if((fp->matcol[1] & 0xf0) == 0){ /* Not glass */
       if(Normalize(v0->xyz,v1->xyz,v2->xyz,n)){
         nn[kn].x=D3DVAL(n[0]);
         nn[kn].y=D3DVAL(n[2]);
         nn[kn].z=D3DVAL(-n[1]);
         ff[kf++]=3;
         for(j=0;j<3;j++){
           v=(MainVp+fp->V[j]);
           x=((float)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
           y=((float)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
           z=((float)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
           vv[kv].x=D3DVAL(x);
           vv[kv].y=D3DVAL(z);
           vv[kv].z=D3DVAL(-y);
           ff[kf++]=kv;
           ff[kf++]=kn;
           kv++;
         }
         kn++;
       }
     }
   }
 }
 ff[kf++]=0;
 LPDIRECT3DRMMESHBUILDER builder;
 LPDIRECT3DRMFRAME frame;
 lpD3DRM->CreateMeshBuilder(&builder);
 builder->AddFaces(kv,vv,kn,nn,ff,NULL);
 builder->SetColor(D3DRMCreateColorRGB(D3DVAL(1.0), D3DVAL(1.0), D3DVAL(1.0)));
 lpD3DRM->CreateFrame(info->scene,&frame);
 frame->AddVisual(builder);
 builder->Release();
 frame->AddRotation(D3DRMCOMBINE_REPLACE,
         D3DVAL(1.0f),D3DVAL(0.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspTheta)/180.0f));
 frame->AddRotation(D3DRMCOMBINE_BEFORE,
         D3DVAL(0.0f),D3DVAL(1.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspFi)/180.0f));
 frame->SetPosition(info->scene,
         D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0));
 info->object=frame;
 X__Free(vv);
 X__Free(nn);
 X__Free(ff);
}
#endif

#undef CreateFace

#if 0
static void MakeD3dModel(AppInfo *info){
// separate faces  - no smoothing
 D3DVECTOR *vv, *nn;
 float x,y,z,scale,n[3];
 face *fp;
 unsigned long *ff;
 vertex *v,*v1,*v2,*v0;
 long i,j,kv,kn,kf;
 if(Nface == 0)return;
 scale=10.0/TVsizeX;
 if((vv = (D3DVECTOR *)X__Malloc(Nvert*sizeof(D3DVECTOR))) == NULL){
   return;
 }
 if((ff = (unsigned long *)X__Malloc(1*sizeof(unsigned long))) == NULL){
   X__Free(vv);
   return;
 }
 if((nn = (D3DVECTOR *)X__Malloc(Nface*sizeof(D3DVECTOR))) == NULL){
   X__Free(vv);
   X__Free(ff);
   return;
 }
 v=MainVp; kv=kf=0; for(i=0;i<Nvert;i++,v++){
   x=((float)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
   y=((float)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
   z=((float)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
   vv[kv].x=D3DVAL(x);
   vv[kv].y=D3DVAL(z);
   vv[kv].z=D3DVAL(-y);
   kv++;
 }
 ff[kf++]=0;
 kf=1;
 fp=MainFp; kn=0; for(i=0;i<Nface;i++,fp++){
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   Normalize(v0->xyz,v1->xyz,v2->xyz,n);
   nn[kn].x=D3DVAL(n[0]);
   nn[kn].y=D3DVAL(n[2]);
   nn[kn].z=D3DVAL(-n[1]);
   kn++;
 }
 LPDIRECT3DRMMESHBUILDER builder;
 LPDIRECT3DRMFRAME frame;
 LPDIRECT3DRMFACE df;
 lpD3DRM->CreateMeshBuilder(&builder);
 builder->AddFaces(kv,vv,kn,nn,ff,NULL);
 builder->SetColor(D3DRMCreateColorRGB(D3DVAL(1.0),D3DVAL(1.0),D3DVAL(0.0)));
 fp=MainFp; for(i=0;i<Nface;i++,fp++){
   builder->CreateFace(&df);
   for(j=0;j<3;j++){
     df->SetColorRGB(D3DVAL(fp->color[0]/255.0f),
                     D3DVAL(fp->color[1]/255.0f),
                     D3DVAL(fp->color[2]/255.0f));
     df->AddVertexAndNormalIndexed((DWORD)fp->V[j],(DWORD)i);
   }
   df->Release();
 }

 lpD3DRM->CreateFrame(info->scene,&frame);
 frame->AddVisual(builder);
 builder->Release();
 frame->AddRotation(D3DRMCOMBINE_REPLACE,
         D3DVAL(1.0f),D3DVAL(0.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspTheta)/180.0f));
 frame->AddRotation(D3DRMCOMBINE_BEFORE,
         D3DVAL(0.0f),D3DVAL(1.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspFi)/180.0f));
 frame->SetPosition(info->scene,
         D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0));
 info->object=frame;
 X__Free(vv);
 X__Free(ff);
 X__Free(nn);
}
#endif

//#if 0
static void MakeD3dModel(AppInfo *info){
// separate faces  - smoothing
 D3DVECTOR *vv, *nn;
 float x,y,z,scale,n[3];
 face *fp;
 unsigned long *ff;
 vertex *v,*v1,*v2,*v0;
 long i,j,kv,kn,kf;
 if(Nface == 0)return;
 scale=10.0/TVsizeX;
 if((vv = (D3DVECTOR *)X__Malloc(Nvert*sizeof(D3DVECTOR))) == NULL){
   return;
 }
 if((ff = (unsigned long *)X__Malloc(1*sizeof(unsigned long))) == NULL){
   X__Free(vv);
   return;
 }
 if((nn = (D3DVECTOR *)X__Malloc(Nvert*sizeof(D3DVECTOR))) == NULL){
   X__Free(vv);
   X__Free(ff);
   return;
 }
 v=MainVp;  for(i=0;i<Nvert;i++,v++){
   x=((float)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
   y=((float)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
   z=((float)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
   vv[i].x=D3DVAL(x);
   vv[i].y=D3DVAL(z);
   vv[i].z=D3DVAL(-y);
   nn[i].x = nn[i].y = nn[i].z = D3DVAL(0.0f);
 }
 ff[0]=0;
 kf=1;
 fp=MainFp; kn=0; for(i=0;i<Nface;i++,fp++){
   long i1,i2,i3;
   i1=fp->V[0]; i2=fp->V[1]; i3=fp->V[2];
   v0=(MainVp+i1); v1=(MainVp+i2); v2=(MainVp+i3);
   Normalize(v0->xyz,v1->xyz,v2->xyz,n);
   nn[i1].x += D3DVAL(n[0]);
   nn[i1].y += D3DVAL(n[2]);
   nn[i1].z += D3DVAL(-n[1]);
   nn[i2].x += D3DVAL(n[0]);
   nn[i2].y += D3DVAL(n[2]);
   nn[i2].z += D3DVAL(-n[1]);
   nn[i3].x += D3DVAL(n[0]);
   nn[i3].y += D3DVAL(n[2]);
   nn[i3].z += D3DVAL(-n[1]);
//   D3DRMVectorNormalize(&nn[i1]);
//   D3DRMVectorNormalize(&nn[i2]);
//   D3DRMVectorNormalize(&nn[i3]);
 }
 for(i=0;i<Nvert;i++){
   D3DRMVectorNormalize(&nn[i]);
 }
 LPDIRECT3DRMMESHBUILDER builder;
 LPDIRECT3DRMFRAME frame;
 LPDIRECT3DRMFACE df;
 lpD3DRM->CreateMeshBuilder(&builder);
 builder->AddFaces(Nvert,vv,Nvert,nn,ff,NULL);
 builder->SetColor(D3DRMCreateColorRGB(D3DVAL(1.0),D3DVAL(1.0),D3DVAL(0.0)));
 fp=MainFp; for(i=0;i<Nface;i++,fp++){
   builder->CreateFace(&df);
   for(j=0;j<3;j++){
     df->SetColorRGB(D3DVAL(fp->color[0]/255.0f),
                     D3DVAL(fp->color[1]/255.0f),
                     D3DVAL(fp->color[2]/255.0f));
     df->AddVertexAndNormalIndexed((DWORD)fp->V[j],(DWORD)fp->V[j]);
   }
   df->Release();
 }

 lpD3DRM->CreateFrame(info->scene,&frame);
 frame->AddVisual(builder);
 builder->Release();
 frame->AddRotation(D3DRMCOMBINE_REPLACE,
         D3DVAL(1.0f),D3DVAL(0.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspTheta)/180.0f));
 frame->AddRotation(D3DRMCOMBINE_BEFORE,
         D3DVAL(0.0f),D3DVAL(1.0f),D3DVAL(0.0f),
         D3DVAL(-3.1415926f*(*PerspFi)/180.0f));
 frame->SetPosition(info->scene,
         D3DVAL(0.0), D3DVAL(0.0), D3DVAL(10.0));
 info->object=frame;
 X__Free(vv);
 X__Free(ff);
 X__Free(nn);
}
//#endif
