/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file DESIGN.C   THE MAIN PROGRAM ENTRY POINT FOR THE DESIGNER MODULE  */

// This file also includes the program's main message loop plus code
// to create and handle the main viewing windows.

#define MODULE_DESIGN 1

#include "design.h"

#define SIZESTATUS1 195 // 130 //This must also be defined in SETUP.C

#define NFILESLIST 5
#define LASTFILECMD 31000

typedef struct tagLASTFILESLIST{
  BOOL present;
  char file[256];
} LASTFILESLIST;

static VOID DesignMenuCommand(int,HWND,long);
static void EngageSkeletonTool(int item);
static void EngageDesignerTool(int );
static void ExecuteDesignerAction(int );
static BOOL SetTriViewCursor(BOOL);
static void UpdateLastFilesList(char *);

extern HINSTANCE __cdecl ResourceX(void);

HBITMAP hBmBackground=NULL;

static HWND hWndHelp=NULL;
static BOOL bTrackMouse=FALSE,bTrackIcon=FALSE,bToolCaptured=FALSE,
            bSpaceDown=FALSE,bTrackPan=FALSE,bTrackZoom=FALSE,
            bCtrlDown=FALSE,bShiftDown=FALSE,bAltDown=FALSE;
static BOOL in_subroutine_mode=FALSE,bResizing=FALSE;
static int  ScreenXsize,ScreenYsize;
static UINT screen_ready=0;
static int  trap_scroll_cursor=0;
static int tmessage_table[MAXTOOLS]={
  IDS_T_NOTOOL, IDS_T_NODETOOL, IDS_T_PAN, IDS_T_INZOOM,
  IDS_T_GRABBER,IDS_T_ROTATOR,IDS_T_EXPANDER,IDS_T_PLOTTER,
  IDS_T_SHAPER, IDS_T_MAGNETTO, IDS_T_BUILDER, IDS_T_PLOT3D,
  IDS_T_PRIMS, IDS_T_SKANIMATE, IDS_T_SELECTOR, IDS_T_DESELECTOR,
  IDS_T_SKGRAB, IDS_T_SKADD, IDS_T_SKDELETE, IDS_T_SKSUBDIV,
  IDS_T_SKASSIGN, IDS_T_SKSELECT, IDS_T_SKATMAP,IDS_T_SKTWIST
                                    };
static int lmessage_table[2]={IDS_SWAPSVIEW,IDS_SWAP3VIEW};
static char gszSCEsaved[256]={"unnamed.mfx"};
static LASTFILESLIST LastFilesList[NFILESLIST];
static int nLastFiles=0;

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 BOOL bMsg;
 int i;
 MSG msg;
 HWND hWndPrev;
 HMODULE hLibSpin;
 HANDLE hAccel,hNewAccel;
#include "..\timeout.c"
 hWndPrev = FindWindow(szMainClass, NULL);
 if (hWndPrev != NULL){
   SetForegroundWindow(hWndPrev);
   BringWindowToTop(hWndPrev);
   ShowWindow (hWndPrev,SW_SHOW);
   return FALSE;
 }
 hInstance=ResourceX();
 InitCommonControls();
 debug=NULL;
// debug=fopen("d:debug.ofx","w");
 strcpy(ruler_name,"None");
 ghdc_triview_Bitmap[0]=ghdc_triview_Bitmap[1]=
 ghdc_triview_Bitmap[2]=ghdc_view_Bitmap=NULL;
 ghbm_triview[0]=ghbm_triview[1]=ghbm_triview[2]=ghbm_view=NULL;
 gCurTool=gCurCommand=gCurSelect=gCurZoom=gCurAction
         =gCurCoord=gCurLock=-1;
 if(!hPrevInstance){
   WNDCLASS wc;
   BOOL     rc;
   if((ghpaletteScreen=CreateAppPalette(&ScreenXsize,&ScreenYsize)) == NULL)
     return FALSE;
   wc.lpszClassName = szMainClass;
   wc.style = CS_HREDRAW | CS_VREDRAW;
   wc.lpfnWndProc = (WNDPROC)MainModelerWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = LoadIcon(hInstance,"MODELERICON");
   wc.hCursor = LoadCursor(NULL,IDC_ARROW);
   wc.hbrBackground = NULL;
   wc.lpszMenuName = MAKEINTRESOURCE(MODELER_MENU);
   rc = RegisterClass( &wc );
   if(!rc)return FALSE;
   wc.lpszClassName = szTriViewClass;
   wc.style = CS_DBLCLKS;
   wc.lpfnWndProc = (WNDPROC)ModelerTriViewWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL,IDC_ARROW); //(HCURSOR)NULL;   /* use so that cursor does not change auto */
   wc.hbrBackground = NULL;
   wc.lpszMenuName = NULL;
   rc = RegisterClass( &wc );
   if(!rc)return FALSE;
   wc.lpszClassName = szViewClass;
   wc.style = 0;
   wc.lpfnWndProc = (WNDPROC)ModelerViewWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL,IDC_ARROW);
   wc.hbrBackground = NULL;
   wc.lpszMenuName = NULL;
   rc = RegisterClass( &wc );
   if(!rc)return FALSE;
   wc.lpszClassName = szToolboxClass;
   wc.style = 0;
   wc.lpfnWndProc = (WNDPROC)ToolboxWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = NULL; //GetStockObject(LTGRAY_BRUSH);
   wc.lpszMenuName = NULL;
   if(!RegisterClass(&wc))return FALSE;
   wc.lpszClassName = szToolBtnClass;
   wc.style = CS_DBLCLKS;
   wc.lpfnWndProc = (WNDPROC)ToolBtnWndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
   wc.hCursor = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = NULL; //(HBRUSH)(COLOR_WINDOW + 1);
   wc.lpszMenuName = NULL;
   if(!RegisterClass(&wc))return FALSE;
 }
 if((hLibSpin=LoadLibrary("spin.dll")) == NULL){
   MessageBox (NULL,"SPIN.DLL",NULL,
               MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
   return FALSE;
 }
 ghinst_main=hInstance;
 LoadPreferences();
 LoadArrowIcons();
 /* WARNING no Message Boxes until Window LOOP */
 if(!SetupMainWindow(ghinst_main,nCmdShow))return FALSE;
 ActivatePreferences();
 if(!DesignStartup(&screen_ready))return FALSE;
 if(setjmp(j_buf) != 0){
   MessageBox (NULL,"Fatal ERROR", NULL, MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
   return 0;
 }
 hAccel=LoadAccelerators(ghinst_main,MAKEINTRESOURCE(MODELER_ACCEL));
 if(hAccel == NULL){
   MessageBox (NULL,"No accelerator table", NULL, MB_OK|MB_ICONEXCLAMATION|
               MB_TASKMODAL);
   return 0;
 }
 hNewAccel=hAccel;
 ghwndOpenGLview=NULL;
 StartOpenGL(TRUE);
 PostMessage(ghwnd_main,WM_COMMAND,IDM_WINDOW_SHOWALL,0);
 if(GetCommandLine() != NULL){
   long l,X__argc;
   char *cp,*cp1,cmdl[256];
   strcpy(cmdl,GetCommandLine());
   AnsiLower(cmdl);
   if((cp=strstr(cmdl,".exe")) != NULL){
     if((cp=strchr(cp,' ')) != NULL){
       cp++;
       if((cp1=strchr(cp,'"')) != NULL)cp=cp1+1;
       if((cp1=strchr(cp,'"')) != NULL)*cp1=0;
       l=strlen(cp);
       if(l < 2)X__argc=0;
     }
     else X__argc=0;
   }
   else X__argc=0;
   if(X__argc > 1){
     WritePrivateProfileString("CALLBACK","FILE",cp,IniFilename);
     PostMessage(ghwnd_main,WM_COMMAND,IDM_FILE_LOAD_DIRECT,0);
   }
 }
 while (GetMessage(&msg,NULL,0,0)){
   bMsg=TRUE;
   if(!IsWindow(ghwndCoord1) || !IsDialogMessage(ghwndCoord1,&msg)){
     if(!IsWindow(ghwndDlgAttribs) || !IsAttribDlgMessage(ghwndDlgAttribs,&msg)){
       if(!IsWindow(ghwndDlgExpand) || !IsDialogMessage(ghwndDlgExpand,&msg)){
         if(!IsWindow(ghwndDlgPrimitives) || !IsPrimitivesDialogMessage(&msg)){
           if(!IsWindow(ghwndSkEdit) || !IsDialogMessage(ghwndSkEdit,&msg)){
             if(!IsWindow(ghwndSkToolbar) || !IsDialogMessage(ghwndSkToolbar,&msg)){
               if(!IsWindow(hWndHelp) || !IsDialogMessage(hWndHelp,&msg)){
                 if(!TranslateAccelerator(ghwnd_main,hNewAccel,&msg)){
                   TranslateMessage(&msg);
                   DispatchMessage(&msg);
                 }
               }
             }
           }
         }
       }
     }
   }
 }
 if(debug != NULL)fclose(debug);
 DesignShutdown();
 FreeLibrary(hLibSpin);
 return (msg.wParam);
}

// Windows' message handler functions

LRESULT CALLBACK MainModelerWndProc(HWND hwnd, UINT msg, WPARAM wparam,
                                    LPARAM lparam ){
  static BOOL bCaptured=FALSE;
 static int  last_x,last_y,type=0;
 static RECT rc;
 BOOL flag;
 int i,x,y,ltype;
 switch( msg ){
   case WM_CREATE:{
       HMENU hmenu=GetMenu(hwnd);
#if __DEMO__
       EnableMenuItem(hmenu,IDM_FILE_SAVE_ALL,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_SAVE_OBJECT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_SAVE_NAMED,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_SAVE_LETTERS,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_SAVE_PATH,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_IMPORT_TDD,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_IMPORT_SCN,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_PRINT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_BEZIER,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_HEIGHT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_TRACEOUTLINE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_EXPORT_3DS,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_EXPORT_DXF,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_EXPORT_STL,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_EXPORT_VRML,MF_GRAYED);
#endif
#if __LITE__
       EnableMenuItem(hmenu,IDM_FILE_TRACEOUTLINE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_HEIGHT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_FILE_EXPORT_VRML,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_SELECT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_POSE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_STRETCH,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_BUILD,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_REMOVE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_INSERT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_ATTACH,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_MAP,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_HIERARCHY_TWIST,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ATTRIBUTES_MAPFROMMAP,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ATTRIBUTES_MAPONPATH,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ATTRIBUTES_SETVERTEXMAPPING,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ATTRIBUTES_MIRROR,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ATTRIBUTES_REFINDEX,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_EDIT_SELECTMAPPED,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_EDIT_DESELECTMAPPED,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ACTIONS_BOOLEAN,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ACTIONS_TWIST,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ACTIONS_ORIENTATE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ACTIONS_REVERSE,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_ACTIONS_JOINS,MF_GRAYED);
#endif
       if(bBackgroundImage){
         hBmBackground=LoadImage(ghinst_main,"modeler.bmp",IMAGE_BITMAP,
                                0,0,LR_LOADFROMFILE);
         if(hBmBackground == NULL)
           hBmBackground=LoadBitmap(ghinst_main,"screenbitmap");
       }
     }
     break;
   case WM_DESTROY:
     if(hBmBackground != NULL)DeleteObject(hBmBackground);
     hBmBackground=NULL;
     break;
   case WM_ACTIVATE:
     if(LOWORD(wparam) != WA_INACTIVE)SetFocus(hwnd); /* under test Aug 3/94   */
                                                      /* for external modelers */
     SendMessage(hwnd,WM_NCACTIVATE,(WPARAM)TRUE,0);
     break;
   case WM_ACTIVATEAPP:
     PostMessage(hwnd,WM_NCACTIVATE,wparam,0);
     break;
   case WM_SETCURSOR:  // must make sure NOT child window
     if((HWND)wparam == ghwnd_main && current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
   case WM_MENUSELECT:{
       HMENU hmenu;
       UINT fuFlags,uItem,id,i;
       bSpaceDown=FALSE; bCtrlDown=FALSE; bShiftDown=FALSE; bAltDown=FALSE;
       hmenu=(HMENU)lparam;
       fuFlags=(UINT)HIWORD(wparam);
       uItem=(UINT)LOWORD(wparam);
       id=0;
       if(fuFlags == 0xFFFF)id=0;
       else if(fuFlags & MF_POPUP){
       }
       else{
         id=CheckMenuHelpData(uItem);
       }
       DisplayMenuHelpData(ghwnd_info,id);
     }
     break;
   case WM_TIMER:
     if(update_timer == 1){
       KillTimer(ghwnd_main,update_timer); update_timer=0;
       if(bTrackPan || bTrackZoom)DrawQuickModel();
     }
     break;
   case WM_LBUTTONDOWN:
     x=(short)LOWORD(lparam); y=(short)HIWORD(lparam);
     type=0;
     if(View == TRIVIEW){
       if(x > WindowLeftOrigin+Window0sizeX - 20 &&
          x < WindowLeftOrigin+Window0sizeX + 20 &&
          y > WindowTopOrigin+Window0sizeY  - 20 &&
          y < WindowTopOrigin+Window0sizeY  + 20)type=1;
       else if(x > WindowLeftOrigin+Window0sizeX - 10 &&
               x < WindowLeftOrigin+Window0sizeX + 10 &&
               y > WindowTopOrigin+10 &&
               y < WindowTopOrigin+TriviewSpaceY)type=2;
       else if(y > WindowTopOrigin+Window0sizeY - 10 &&
               y < WindowTopOrigin+Window0sizeY + 10 &&
               x > WindowLeftOrigin+10 &&
               x < WindowLeftOrigin+TriviewSpaceX-10)type=3;
     }
     if     (type == 1)SetCursor(LoadCursor(NULL,IDC_SIZEALL));
     else if(type == 2)SetCursor(LoadCursor(NULL,IDC_SIZEWE));
     else if(type == 3)SetCursor(LoadCursor(NULL,IDC_SIZENS));
     if(type > 0){
       HDC hdc;
       hdc=GetDC(ghwnd_main);
       FocusActiveWindow(hdc,FALSE);
       ReleaseDC(ghwnd_main,hdc);
       SetCapture(hwnd);
       DeleteMemoryWindows();
       bResizing=TRUE;
       bCaptured=TRUE;
       GetClientRect(hwnd,&rc);
       last_x=x; last_y=y;
     }
     break;
   case WM_LBUTTONUP:
     if(bCaptured){
       HDC hdc;
       hdc=GetDC(ghwnd_main);
       FocusActiveWindow(hdc,TRUE);
       ReleaseDC(ghwnd_main,hdc);
       ReleaseCapture();
       bCaptured=FALSE;
       bResizing=FALSE;
       CreateMemoryWindows();
       DrawModel();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0);
       type=0;
     }
     break;
   case WM_MOUSEMOVE:
     x=(short)LOWORD(lparam); y=(short)HIWORD(lparam);
     if(!bCaptured && View == TRIVIEW){
       ltype=0;
       if(x > WindowLeftOrigin+Window0sizeX - 20 &&
          x < WindowLeftOrigin+Window0sizeX + 20 &&
          y > WindowTopOrigin+Window0sizeY  - 20 &&
          y < WindowTopOrigin+Window0sizeY  + 20)ltype=1;
       else if(x > WindowLeftOrigin+Window0sizeX - 10 &&
               x < WindowLeftOrigin+Window0sizeX + 10 &&
               y > WindowTopOrigin+10 &&
               y < WindowTopOrigin+TriviewSpaceY)ltype=2;
       else if(y > WindowTopOrigin+Window0sizeY - 10 &&
               y < WindowTopOrigin+Window0sizeY + 10 &&
               x > WindowLeftOrigin+10 &&
               x < WindowLeftOrigin+TriviewSpaceX-10)ltype=3;
       if     (ltype == 1)SetCursor(LoadCursor(NULL,IDC_SIZEALL));
       else if(ltype == 2)SetCursor(LoadCursor(NULL,IDC_SIZEWE));
       else if(ltype == 3)SetCursor(LoadCursor(NULL,IDC_SIZENS));
     }
     if(bCaptured){
         int dx,dy;
         if     (type == 1)SetCursor(LoadCursor(NULL,IDC_SIZEALL));
         else if(type == 2)SetCursor(LoadCursor(NULL,IDC_SIZEWE));
         else if(type == 3)SetCursor(LoadCursor(NULL,IDC_SIZENS));
         dx=x-last_x; dy=y-last_y;
         if(dx != 0 || dy != 0){
           if     (type == 2)dy=0;
           else if(type == 3)dx=0;
           Window0sizeX += dx;
           Window0sizeX = max(20,Window0sizeX);
           Window0sizeX = min(TriviewSpaceX-30,Window0sizeX);
           Window0sizeY += dy;
           Window0sizeY = max(20,Window0sizeY);
           Window0sizeY = min(TriviewSpaceY-30,Window0sizeY);
           SetWindowSize(rc.right,rc.bottom,View,FALSE);
           GetTriview(FALSE);
           for(i=0;i<3;i++)
             MoveWindow(ghwnd_triview[i],
                  WindowsXmin[i],WindowsYmin[i],
                  WindowSizeX[i],WindowSizeY[i],TRUE);
           MoveWindow(ghwnd_view,WindowViewXmin,WindowViewYmin,
                WindowViewXsize,WindowViewYsize,TRUE);
           if(ghwndOpenGLview != NULL)MoveWindow(ghwndOpenGLview,
                WindowViewXmin,WindowViewYmin,
                WindowViewXsize,WindowViewYsize,TRUE);
           else UpdateWindow(ghwnd_view);
           UpdateWindow(ghwnd_triview[1]);

         }
         last_x=x; last_y=y;
     }
     break;
   case WM_RBUTTONDOWN:
     SendMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
     break;
   case WM_ERASEBKGND:{
       HDC hDC;
       HPALETTE hpalT;
       HBRUSH hOldBrush;
       RECT rc1;
       hDC = (HDC)wparam;
       if(hBmBackground == NULL){
         hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
         RealizePalette (hDC);
         hOldBrush = SelectObject (hDC,ghbrushScreen);
         GetClientRect(hwnd,(LPRECT)&rc1);
         PatBlt(hDC,0,0,rc1.right,rc1.bottom,PATCOPY);
         SelectPalette(hDC,hpalT,FALSE);
         SelectObject (hDC, hOldBrush);
       }
       else{
         HDC hMemDC;
         HBITMAP hbmOld;
         BITMAP  bm;
         GetObject(hBmBackground,sizeof(BITMAP),&bm);
         hMemDC = CreateCompatibleDC(hDC);
         hbmOld = SelectObject(hMemDC,hBmBackground);
         for(x=0;x<(ScreenXsize/bm.bmWidth)+1;x++)
         for(y=0;y<(ScreenYsize/bm.bmHeight)+1;y++)
           BitBlt(hDC,
                   x*bm.bmWidth,y*bm.bmHeight,
                   bm.bmWidth,
                   bm.bmHeight,
                   hMemDC,
                   0,0,
                   SRCCOPY);
         SelectObject(hMemDC,hbmOld);
         DeleteDC(hMemDC);
       }
       return TRUE;
     }
   case WM_QUERYNEWPALETTE:{
      HDC hDC;
      HPALETTE hpalT;
      hDC=GetDC(hwnd);
      hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
      if(RealizePalette (hDC) > 0)InvalidateRect (hwnd, (LPRECT) (NULL), 1);
      SelectPalette(hDC,hpalT,FALSE);
      ReleaseDC(hwnd,hDC);
      return TRUE;
    }
   case WM_PALETTECHANGED:
     if ((HWND)wparam != hwnd ){
       HDC hDC;
       HPALETTE hpalT;
       hDC=GetDC(hwnd);
       hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
       if(RealizePalette (hDC) > 0)InvalidateRect (hwnd, (LPRECT) (NULL), 1);
       SelectPalette(hDC,hpalT,FALSE);
       ReleaseDC(hwnd,hDC);
     }
     break;
   case WM_PAINT: {
       HDC hdc;
       PAINTSTRUCT ps;
       HBRUSH hOldBrush;
       HPALETTE hpalT;
       RECT rc1;
       hdc=BeginPaint(hwnd,&ps);
       hpalT = SelectPalette(hdc,ghpaletteScreen,FALSE);
       RealizePalette (hdc);
       if(hBmBackground == NULL){
         hOldBrush = SelectObject (hdc,ghbrushScreen);
         GetClientRect(hwnd,(LPRECT)&rc1);
         PatBlt(hdc,0,0,rc1.right,rc1.bottom,PATCOPY);
         SelectObject (hdc, hOldBrush);
       }
       if(!bResizing)FocusActiveWindow(hdc,TRUE);
       SelectPalette(hdc,hpalT,FALSE);
       EndPaint(hwnd,&ps);
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         if(in_subroutine_mode)PostMessage(hwnd,WM_COMMAND,
           IDM_FILE_UPDATE_ANIMATOR,0);
         else PostMessage(hwnd,WM_COMMAND,IDM_FILE_EXIT,0);
         break;
       default:
         return( DefWindowProc( hwnd, msg, wparam, lparam ) );
         break;
     }
     break;
   case WM_COMMAND:
     if(HIWORD(wparam) == 1)lparam=ActiveView; /* acelerator */
     if(HIWORD(wparam) == BN_CLICKED){
       DesignMenuCommand(LOWORD(wparam),hwnd,lparam);
     }
     else DesignMenuCommand(LOWORD(wparam),hwnd,lparam);
     break;
   case WM_SIZE:{
       int cx,cy;
       int StatusPartsList[2]={0,-1};
       if(screen_ready){
         DeleteMemoryWindows();
         cx=LOWORD(lparam); cy=HIWORD(lparam);
         SetWindowSize(cx,cy,View,FALSE);
         GetTriview(FALSE);
         if(View == TRIVIEW){
           for(i=0;i<3;i++){
             MoveWindow(ghwnd_triview[i],WindowsXmin[i],WindowsYmin[i],
                        WindowSizeX[i],WindowSizeY[i],TRUE);
           }
         }
         else{
           MoveWindow(ghwnd_current,
                      WindowsXmin[ActiveView],WindowsYmin[ActiveView],
                      WindowSizeX[ActiveView],WindowSizeY[ActiveView],TRUE);
         }
         MoveWindow(ghwnd_view,WindowViewXmin,WindowViewYmin,
                    WindowViewXsize,WindowViewYsize,TRUE);
         if(ghwndOpenGLview != NULL)MoveWindow(ghwndOpenGLview,
                    WindowViewXmin,WindowViewYmin,
                    WindowViewXsize,WindowViewYsize,TRUE);
         StatusPartsList[0]=cx-SIZESTATUS1;
         SendMessage(ghwnd_info,SB_SETPARTS,(WPARAM)2,(LPARAM)StatusPartsList);
         MoveToolsWindow();
         SendMessage(ghwnd_info,msg,wparam,lparam);
         if(CreateMemoryWindows()){
           DrawModel();
           if(ghwndOpenGLview == NULL)Draw3dView(0,0);
           else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0);
         }
         else{
           DeleteMemoryWindows();
         }
       }
     }
     break;
   case WM_KEYDOWN:
     ExecuteKeyboardKeyCommand(wparam);
   case WM_KEYUP:
     if(!bTrackMouse && !bTrackIcon && !bToolCaptured && !bTrackPan){
       if(GetKeyState(VK_SPACE) & 0x8000)bSpaceDown=TRUE; else bSpaceDown=FALSE;
       if(GetKeyState(VK_CONTROL) & 0x8000)bCtrlDown=TRUE; else bCtrlDown=FALSE;
       if(GetKeyState(VK_SHIFT) & 0x8000)bShiftDown=TRUE; else bShiftDown=FALSE;
       if(GetKeyState(VK_MENU) & 0x8000)bAltDown=TRUE; else bAltDown=FALSE;
       if(CursorInTriview())SetTriViewCursor(FALSE);
     }
     break;
   case WM_SETFOCUS:
     if(GetKeyState(VK_SPACE) & 0x8000)bSpaceDown=TRUE; else bSpaceDown=FALSE;
     if(GetKeyState(VK_SHIFT) & 0x8000)bShiftDown=TRUE; else bShiftDown=FALSE;
     if(GetKeyState(VK_CONTROL) & 0x8000)bCtrlDown=TRUE; else bCtrlDown=FALSE;
     if(GetKeyState(VK_MENU) & 0x8000)bAltDown=TRUE; else bAltDown=FALSE;
     if(CursorInTriview())SetTriViewCursor(FALSE);
     break;
   case WM_KILLFOCUS:
     if(bSpaceDown || bCtrlDown || bShiftDown || bAltDown){
       bSpaceDown=FALSE;
       bCtrlDown=FALSE;
       bShiftDown=FALSE;
       bAltDown=FALSE;
       if(CursorInTriview())SetTriViewCursor(FALSE);
     }
     break;
   case WM_SYSKEYDOWN:
     if(((lparam & 0x40000000) == 0) && wparam == VK_MENU){
       bAltDown=TRUE;
       if((bSpaceDown || tool == INZOOM || tool == MAGNETTO) && CursorInTriview())SetTriViewCursor(FALSE);
     }
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
   case WM_SYSKEYUP:
     bAltDown=FALSE;
     if((bSpaceDown || tool == INZOOM || tool == MAGNETTO) && CursorInTriview())SetTriViewCursor(FALSE);
     break;
   default:
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

LRESULT CALLBACK ModelerTriViewWndProc(HWND hwnd, UINT msg, WPARAM wparam,
                                    LPARAM lparam ){
 HDC hdc;
 RECT rc;
 int id;
 static int iIconSave = -1;
 static BOOL bMouseDown=FALSE;
 BOOL flag1,flag2;
 switch( msg ){
   case WM_ERASEBKGND:{
       HDC hDC;
       HPALETTE hpalT;
       HBRUSH hOldBrush;
       if(bResizing)return TRUE;
       hDC       = (HDC)wparam;
       hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
       RealizePalette (hDC);
       hOldBrush = SelectObject (hDC,ghbrushWindow);
       GetClientRect(hwnd,(LPRECT)&rc);
       PatBlt(hDC,0,0,rc.right,rc.bottom,PATCOPY);
       SelectPalette(hDC,hpalT,FALSE);
       SelectObject (hDC, hOldBrush);
       return TRUE;
   }
   case WM_QUERYNEWPALETTE:{
       HDC hDC;
       HPALETTE hpalT;
       hDC=GetDC(hwnd);
       hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
       if(RealizePalette (hDC) > 0)InvalidateRect (hwnd, (LPRECT) (NULL), 1);
       SelectPalette(hDC,hpalT,FALSE);
       ReleaseDC(hwnd,hDC);
       return TRUE;
     }
   case WM_PALETTECHANGED:
     if ((HWND)wparam != hwnd ){
       HDC hDC;
       HPALETTE hpalT;
       hDC=GetDC(hwnd);
       hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
       if(RealizePalette (hDC) > 0)InvalidateRect (hwnd, (LPRECT) (NULL), 1);
       SelectPalette(hDC,hpalT,FALSE);
       ReleaseDC(hwnd,hDC);
     }
     break;
   case WM_SETCURSOR:
     if(!SetTriViewCursor(TRUE))return(DefWindowProc(hwnd,msg,wparam,lparam));
     break;
   case WM_RBUTTONDOWN:
     if(bToolCaptured){ /* release captured tool */
       if(bMouseDown || bTrackMouse)break;
       goto RELEASE_CAPTURE;
     }
     else if(!bTrackMouse)SendMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
     else if(!bTrackIcon){
       if(tool == NOTOOL){
//         SelectVertex(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
//         Empty_Undo();
       }
       else ToolAction((int)LOWORD(lparam),(int)HIWORD(lparam));
     }
     break;
   case WM_LBUTTONDBLCLK:
     break; if(tool != PLOTTER)break;
     if(bToolCaptured){
       RELEASE_CAPTURE:
       UpdateWindow(hwnd);
       ToolUp(hwnd,ActiveView,(int)LOWORD(lparam),(int)HIWORD(lparam));
       UpdateWindow(hwnd);
       bToolCaptured=FALSE;
       ClipCursor(NULL);
       SetCursor(ghcurSave);
       ReleaseCapture();
       Plot3dOff();
     }
     else{
       if(CheckWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam)) >= 0)break;
       if(bTrackMouse)MessageBeep(MB_OK); /* should not happen */
       SetCapture(hwnd);
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ghcurSave=SetCursor(ghcurCross);
       ClipCursor(&rc);
       ToolDown(hwnd,ActiveView,(int)LOWORD(lparam),(int)HIWORD(lparam));
       ToolAction((int)LOWORD(lparam),(int)HIWORD(lparam));
       bToolCaptured=TRUE;
     }
     break;
   case WM_LBUTTONDOWN:
     if(bToolCaptured){
       ToolAction((int)LOWORD(lparam),(int)HIWORD(lparam));
       bMouseDown=TRUE;
       break;
     }
     if(hwnd != ghwnd_current){
       hdc=GetDC(ghwnd_main);
       FocusActiveWindow(hdc,FALSE);
       ghwnd_current=hwnd;  ActiveView=GetWindowLong(hwnd,GWL_ID);
       FocusActiveWindow(hdc,TRUE);
       ReleaseDC(ghwnd_main,hdc);
     }
     SetCapture(hwnd);
     if(bSpaceDown){
       PanToolDown((int)LOWORD(lparam),(int)HIWORD(lparam),
                    bCtrlDown,bAltDown,hwnd);
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ClipCursor(&rc);
       bTrackPan=TRUE;
     }
     else if((iIconSave=CheckWindowIcon((int)LOWORD(lparam),
                                   (int)HIWORD(lparam))) >= 0){
       InvertWindowIcon(iIconSave,hwnd);
       bTrackIcon=TRUE;
     }
     else if(tool == PAN){
       PanToolDown((int)LOWORD(lparam),(int)HIWORD(lparam),
                    FALSE,FALSE,hwnd);
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ClipCursor(&rc);
       bTrackPan=TRUE;
     }
     else if(tool == INZOOM){
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       bTrackZoom=TRUE;
       ZoomView((int)LOWORD(lparam),(int)HIWORD(lparam));
       bTrackZoom=FALSE;
       ReleaseCapture();
     }
     else{
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ghcurSave=SetCursor(ghcurCross);
       ClipCursor(&rc);
       bTrackMouse=TRUE;
     }
     if(bTrackMouse){
       Move3dCursor(1,(int)LOWORD(lparam),(int)HIWORD(lparam));
       ToolDown(hwnd,ActiveView,(int)LOWORD(lparam),(int)HIWORD(lparam));
       if(!(GetAsyncKeyState(VK_LBUTTON) & 0x8000)){
         bTrackMouse=FALSE;
         ReleaseCapture();
         ClipCursor(NULL);
       }
     }
     break;
   case WM_LBUTTONUP:
     if(bToolCaptured){
       bMouseDown=FALSE;
       break;
     }
     if(bTrackIcon){
       if(iIconSave >= 0){
         InvertWindowIcon(iIconSave,hwnd);
         if     (iIconSave == 0)iIconSave=IDM_PAN_LEFT;
         else if(iIconSave == 1)iIconSave=IDM_PAN_RIGHT;
         else if(iIconSave == 2)iIconSave=IDM_PAN_UP;
         else if(iIconSave == 3)iIconSave=IDM_PAN_DOWN;
         else if(iIconSave == 4)iIconSave=IDM_WINDOW_SWITCHWINDOWS;
         else if(iIconSave == 5)iIconSave=IDM_WINDOW_SWITCHWINDOW0;
         else if(iIconSave == 6)iIconSave=IDM_WINDOW_SWITCHWINDOW1;
         PostMessage(ghwnd_main,WM_COMMAND,(WPARAM)iIconSave,
           (LPARAM)GetWindowLong(hwnd,GWL_ID));
       }
       iIconSave = -1;
       bTrackIcon=FALSE;
       ReleaseCapture();
     }
     else if(bTrackPan){
       ClipCursor(NULL);
       ReleaseCapture();
       if(bSpaceDown){
         flag1=bCtrlDown;
         flag2=bAltDown;
       }
       else{
         if(tool == INZOOM){
           if(bAltDown){flag1=FALSE; flag2=TRUE;}  /* zoom out */
           else        {flag1=TRUE; flag2=FALSE;}  /* zoom in */
         }
         else {flag1=FALSE; flag2=FALSE;}
       }
       PanToolUp((int)LOWORD(lparam),(int)HIWORD(lparam),
                 flag1,flag2,hwnd);
       bTrackPan=FALSE;
       if(!(GetKeyState(VK_SPACE)&0x8000))bSpaceDown=FALSE;
       if(!(GetKeyState(VK_CONTROL)&0x8000))bCtrlDown=FALSE;
       if(!(GetKeyState(VK_SHIFT)&0x8000))bShiftDown=FALSE;
       if(!(GetKeyState(VK_MENU)&0x8000))bAltDown=FALSE;
       SetTriViewCursor(FALSE);
     }
     else if(bTrackMouse){
       bTrackMouse=FALSE;
       ClipCursor(NULL);
       SetCursor(ghcurSave);
       UpdateWindow(hwnd);
       ToolUp(hwnd,ActiveView,(int)LOWORD(lparam),(int)HIWORD(lparam));
       UpdateWindow(hwnd);
       ReleaseCapture();
     }
     break;
   case WM_MOUSEMOVE:
     if(bToolCaptured){
       ToolMove(hwnd,ActiveView,(int)LOWORD(lparam),(int)HIWORD(lparam));
       break;
     }
     if(bTrackMouse){
       if(tool != EXPANDER && tool != ROTATOR && tool != SHAPER &&
         sktool == NO)Move3dCursor(0,(int)LOWORD(lparam),(int)HIWORD(lparam));
       ToolMove(hwnd,ActiveView,(int)LOWORD(lparam),(int)HIWORD(lparam));
     }
     else if(bTrackIcon){
       int new_icon;
       new_icon=CheckWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam));
       if(new_icon !=  iIconSave){
         if(iIconSave >= 0)InvertWindowIcon(iIconSave,hwnd);
         iIconSave=new_icon;
         if(iIconSave >= 0)InvertWindowIcon(iIconSave,hwnd);
       }
     }
     else if(bTrackPan){
       if(tool == PAN || tool == INZOOM || bSpaceDown){
         if(bSpaceDown){
           flag1=bCtrlDown;
           flag2=bAltDown;
         }
         else{
           if(tool == INZOOM && !bAltDown)flag1=TRUE; else flag1=FALSE;
           if(tool == INZOOM &&  bAltDown)flag2=TRUE; else flag2=FALSE;
         }
         PanToolMove((int)LOWORD(lparam),(int)HIWORD(lparam),
                     flag1,flag2,hwnd);
       }
     }
     else { /* get help on in screen icons */
       int icon,new_message;
       icon=CheckWindowIcon((int)LOWORD(lparam),(int)HIWORD(lparam));
       if(icon >= 0){
         if(icon == 4)new_message=lmessage_table[View];
         else         new_message=IDS_SWAPOTHER;
         if(new_message != current_message){
           current_message=new_message;
           SendPrgmMessage(current_message,1);
         }
       }
       else if(current_message != tmessage_table[tool]){
         current_message = tmessage_table[tool];
         SendPrgmMessage(current_message,1);
       }
     }
     break;
   case WM_PAINT:{
       PAINTSTRUCT ps;
       HPALETTE hpalT;
       hdc=BeginPaint(hwnd,&ps);
       if(!bResizing){
         hpalT = SelectPalette(hdc,ghpaletteScreen,FALSE);
         RealizePalette (hdc);
         GetClientRect(hwnd,&rc);
         id=(int)GetWindowLong(hwnd,GWL_ID);
         if(ghdc_triview_Bitmap[id] != NULL)BitBlt(hdc,0,0,
           rc.right,rc.bottom,ghdc_triview_Bitmap[id],0,0,SRCCOPY);
         DrawOne3dCursor(hdc,id);
         // removed to draw brush in Background windows
		 //if(SelectedBrush >= 0)DrawBrushInOne(hdc,id,SelectedBrush);
         //
		 if(SelectedShader >= 0)DrawShaderInOne(hdc,id,SelectedShader);
         if(sktool == YES && tool != SKANIMATE)DrawSkeletonInOne(hdc,id);
         SelectPalette(hdc,hpalT,FALSE);
       }
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

LRESULT CALLBACK ModelerViewWndProc(HWND hwnd, UINT msg, WPARAM wparam,
                                    LPARAM lparam ){
 int oldi;
 static BOOL bTrackMouseV=FALSE,bTrackPan=FALSE;
 static long xs,ys;
 RECT rc;
 switch( msg ){
   case WM_SETCURSOR:
     if     (tool == SELECTOR)SetCursor (ghcurSelect);
     else if(tool == DESELECTOR)SetCursor(ghcurDeselect);
     else{
       if(LOWORD(lparam) == HTCLIENT)SetCursor(ghcurPan);
       else SetCursor(ghcurArrow);
     }
     break;
   case WM_ERASEBKGND:{
     HDC hDC;
     HPALETTE hpalT;
     HBRUSH hOldBrush;
     if(bResizing)return TRUE;
     hDC       = (HDC)wparam;
     hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
     RealizePalette (hDC);
     hOldBrush = SelectObject (hDC,ghbrushWindow);
     GetClientRect(hwnd,(LPRECT)&rc);
     PatBlt(hDC,0,0,rc.right,rc.bottom,PATCOPY);
     SelectPalette(hDC,hpalT,FALSE);
     SelectObject (hDC, hOldBrush);
     return TRUE;
   }
   case WM_QUERYNEWPALETTE:{
       HDC hDC;
       HPALETTE hpalT;
       hDC=GetDC(hwnd);
       hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
       if(RealizePalette (hDC) > 0)InvalidateRect (hwnd, (LPRECT) (NULL), 1);
       SelectPalette(hDC,hpalT,FALSE);
       ReleaseDC(hwnd,hDC);
       return TRUE;
     }
   case WM_PALETTECHANGED:
     if ((HWND)wparam != hwnd ){
       HDC hDC;
       HPALETTE hpalT;
       hDC=GetDC(hwnd);
       hpalT = SelectPalette(hDC,ghpaletteScreen,FALSE);
       if(RealizePalette (hDC) > 0)InvalidateRect (hwnd, (LPRECT) (NULL), 1);
       SelectPalette(hDC,hpalT,FALSE);
       ReleaseDC(hwnd,hDC);
     }
     break;
   case WM_LBUTTONDOWN:
     if(tool == SELECTOR || tool == DESELECTOR){
       SetCapture(hwnd);
       GetWindowRect(hwnd,&rc);
       rc.left += 1; rc.top += 1; rc.right -= 1; rc.bottom -= 1;
       ClipCursor(&rc);
       bTrackMouseV=TRUE;
       SelectorDown(hwnd,-1,(int)LOWORD(lparam),(int)HIWORD(lparam));
     }
     else {
       GetClientRect(hwnd,&rc);
       if(LOWORD(lparam) > rc.right-5*Textwidth &&
          LOWORD(lparam) < rc.right &&
          HIWORD(lparam) > rc.bottom-Textheight &&
          HIWORD(lparam) < rc.bottom)
            PostMessage(ghwnd_main,WM_COMMAND,IDM_VIEW_HIDDENLINE,0);
       else{
         xs=LOWORD(lparam);
         ys=HIWORD(lparam);
         SetCapture(hwnd);
         bTrackPan=TRUE;
       }
     }
     break;
   case WM_MOUSEMOVE:
     if(bTrackMouseV)SelectorMove(hwnd,-1,(int)LOWORD(lparam),(int)HIWORD(lparam));
     else if(bTrackPan){
       long dx,dy;
       dx=(short)LOWORD(lparam)-xs; xs=(short)LOWORD(lparam);
       dy=(short)HIWORD(lparam)-ys; ys=(short)HIWORD(lparam);
       if(labs(dx) > 0 || labs(dy) > 0){
         PerspFi=PerspFi+(double)dx;
         while(PerspFi < -180.0)PerspFi += 360.0;
         while(PerspFi >  180.0)PerspFi -= 360.0;
         SetScrollPos(hwnd,SB_HORZ,(int)PerspFi+180,TRUE);
         PerspTheta=PerspTheta+(double)dy;
         if(PerspTheta > -90 && PerspTheta < 90)
           SetScrollPos(hwnd,SB_VERT,(int)PerspTheta+90,TRUE);
         DrawQuick3dView(0);
       }
     }
     else{
       GetClientRect(hwnd,&rc);
       if(LOWORD(lparam) > rc.right-5*Textwidth &&
          LOWORD(lparam) < rc.right &&
          HIWORD(lparam) > rc.bottom-Textheight &&
          HIWORD(lparam) < rc.bottom){
         if(current_message < 0){
           current_message=0;
           SendPrgmMessage(IDS_PERSPVIEW,1);
         }
       }
       else if(current_message != IDS_VIEWUPDATE){
         current_message = IDS_VIEWUPDATE;
         SendPrgmMessage(IDS_VIEWUPDATE,1);
       }
     }
     break;
   case WM_LBUTTONUP:
     if(bTrackMouseV){
       if(tool == SELECTOR || tool == DESELECTOR)
         SelectorUp(hwnd,-1,(int)LOWORD(lparam),(int)HIWORD(lparam),TRUE,FALSE);
       bTrackMouseV=FALSE;
       ClipCursor(NULL);
       ReleaseCapture();
     }
     else if(bTrackPan){
       ReleaseCapture();
       bTrackPan=FALSE;
       PostMessage(ghwnd_main,WM_COMMAND,IDM_VIEW_WIREFRAME,0);
     }
     break;
   case WM_RBUTTONDOWN:
     if(bTrackMouseV || bTrackPan)break;
     else{
       POINT pt;
       HMENU hMenu=LoadMenu(ghinst_main,"ViewWindowPopupMenu");
       if(hMenu == NULL)break;
       pt.x=LOWORD(lparam);
       pt.y=HIWORD(lparam);
       ClientToScreen(hwnd,&pt);
       TrackPopupMenu(GetSubMenu(hMenu,0),TPM_LEFTALIGN | TPM_LEFTBUTTON,
                      pt.x,pt.y,0,ghwnd_main,NULL);
       DestroyMenu(hMenu);
     }
     break;
   case WM_HSCROLL:
     oldi=(int)PerspFi;
     switch (LOWORD(wparam)) {
       case SB_ENDSCROLL:
        break;
       case SB_LINEDOWN:
        PerspFi +=  5; PerspFi = min( 180,PerspFi);  break;
       case SB_LINEUP:
        PerspFi -=  5; PerspFi = max(-180,PerspFi);  break;
       case SB_PAGEDOWN:
        PerspFi += 45; PerspFi = min( 180,PerspFi);  break;
       case SB_PAGEUP:
        PerspFi -= 45; PerspFi = max(-180,PerspFi);  break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION:
         PerspFi=(double)HIWORD(wparam)-180;  break;
       default: break;
     }
     if(oldi != (int)PerspFi){
       SetScrollPos(hwnd,SB_HORZ,(int)PerspFi+180,TRUE);
       Draw3dView(0,0);
       UpdateWindow(ghwnd_view);
     }
     break;
   case WM_VSCROLL:
     oldi=(int)PerspTheta;
     switch (LOWORD(wparam)) {
       case SB_ENDSCROLL:
        break;
       case SB_LINEDOWN:
        PerspTheta +=  5; PerspTheta = min( 90,PerspTheta);  break;
       case SB_LINEUP:
        PerspTheta -=  5; PerspTheta = max(-90,PerspTheta);  break;
       case SB_PAGEDOWN:
        PerspTheta += 20; PerspTheta = min( 90,PerspTheta);  break;
       case SB_PAGEUP:
        PerspTheta -= 20; PerspTheta = max(-90,PerspTheta);  break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION:
         PerspTheta=(double)HIWORD(wparam)-90;
         break;
       default: break;
     }
     if(oldi != (int)PerspTheta){
       SetScrollPos(hwnd,SB_VERT,(int)PerspTheta+90,TRUE);
       Draw3dView(0,0);
       UpdateWindow(ghwnd_view);
     }
     break;
   case WM_PAINT: {
       HDC hdc;
       PAINTSTRUCT ps;
       HPALETTE hpalT;
       hdc=BeginPaint(hwnd,&ps);
       if(!bResizing){
         hpalT = SelectPalette(hdc,ghpaletteScreen,FALSE);
         RealizePalette (hdc);
         GetClientRect(hwnd,&rc);
         if(ghdc_view_Bitmap != NULL){
           COLORREF oldtextcolour;
           int oldbkmode;
           oldbkmode=SetBkMode(ghdc_view_Bitmap,TRANSPARENT);
           oldtextcolour=SetTextColor(ghdc_view_Bitmap,RGB(255,255,255));
           TextOut(ghdc_view_Bitmap,
                   1,1,//rc.right-5*Textwidth,rc.bottom-Textheight,
                   "View",strlen("View"));
           SetBkMode(ghdc_view_Bitmap,oldbkmode);
           SetTextColor(ghdc_view_Bitmap,oldtextcolour);
           BitBlt(hdc,0,0,rc.right,rc.bottom,ghdc_view_Bitmap,0,0,SRCCOPY);
         }
         SelectPalette(hdc,hpalT,FALSE);
       }
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return( DefWindowProc( hwnd, msg, wparam, lparam ) );
 }
 return( 0L );
}

// End of Windows' event handler functions

static BOOL SetTriViewCursor(BOOL type){
//
// Set the mouse cursor to reflect the tool and state of modifier keys.
//
 if(bSpaceDown){
   if(bCtrlDown)SetCursor(ghcurZin);
   else if(bAltDown)SetCursor(ghcurZout);
   else SetCursor(ghcurPan);
 }
 else if(tool == MAGNETTO && bCtrlDown && bAltDown)SetCursor(ghcurMagnetCu);
 else if(tool == MAGNETTO && bAltDown)SetCursor(ghcurMagnetFr);
 else if(tool == MAGNETTO && bCtrlDown)SetCursor(ghcurMagnetTw);
 else if(tool < NTOOLS)SetCursor(ghcurTools[tool]);
 else if(tool == SELECTOR){
   if(bShiftDown)SetCursor(ghcurSel_l_a);
   else if(bCtrlDown)SetCursor(ghcurSel_c_a);
   else SetCursor(ghcurSel_a);//SetCursor(ghcurSelect);
 }
 else if(tool == DESELECTOR){
   if(bShiftDown)SetCursor(ghcurSel_l_s);
   else if(bCtrlDown)SetCursor(ghcurSel_c_s);
   else SetCursor(ghcurSel_s);//SetCursor(ghcurDeselect);
 }
 else if(tool == PRIMS)SetCursor(ghcurPrims);
 else if(sktool == YES)SetCursor(ghcurArrowD);
 else if(type)return FALSE;
 else SetCursor(ghcurArrow);
 return TRUE;
}

// code to execute the commands (from menus+buttons) 

#include "glmenu.h"
#include "callback.h"
#include "..\animate\callall.h"
#if __HELP_WS__
#include "..\help_ws\include\htmlhelp.h"
#endif

#define FI_SCE 0
#define FI_DXF 1
#define FI_3DS 2
#define FI_STL 3

static long identify_file_format(char *name){
 char cf[256];
 int ff=FI_SCE;
 strcpy(cf,name);
 if     (strstr(cf,".DXF") || strstr(cf,".dxf"))ff=FI_DXF;
 else if(strstr(cf,".3DS") || strstr(cf,".3ds"))ff=FI_3DS;
 else if(strstr(cf,".STL") || strstr(cf,".stl"))ff=FI_STL;
 return ff;
}

static VOID DesignMenuCommand(int item, HWND hwnd, long lparam){
 BOOL bUpdate=FALSE;
 int id;
 char temp_str[256];
 if(item >= 30000 && item < LASTFILECMD){
   if(ExecuteExternalAction(item-30000,ghwnd_main)){
     DrawModel();
     UpdateCounters();
     UpdateSkTreeView();
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
   }
   return;
 }
 switch (item){
   case IDC_CENTRE_DIALOG_CURSOR:
     CentreDialogOnCursor((HWND)lparam);
     break;
   case IDC_CENTRE_DIALOG_SCREEN:
     break;
   case IDC_UPDATE_VERTEXHEAP:
     UpdateVertexHeap((long)lparam);
     break;
   case IDC_UPDATE_EDGEHEAP:
     UpdateEdgeHeap((long)lparam);
     break;
   case IDC_UPDATE_FACEHEAP:
     UpdateFaceHeap((long)lparam);
     break;
   case LASTFILECMD+1:
   case LASTFILECMD+2:
   case LASTFILECMD+3:
   case LASTFILECMD+4:
   case LASTFILECMD+5:
   case IDM_FILE_NEW:
   case IDM_FILE_LOAD_MODEL:
     if(tool != NOTOOL)SendMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
     if(EDIT_ACTION){
       int fResult;
       fResult=SendPrgmQuery(IDQ_CHANGED,2);
       if(fResult == IDCANCEL)break;
       if(fResult == IDYES){
#if !__DEMO__
         if(SelectSfxFileName(1,gszSCEfile,gszSCEdir,IDX_MISC_SAVEMODEL,
                           "(*.MFX)|*.mfx|",hwnd) == TRUE){
           AppendFileExtension(gszSCEfile,".mfx");
           UpdateLastFilesList(gszSCEfile);
           ghcurSave=SetCursor(ghcurWait);
           SaveObject(gszSCEfile,1);
           SetCursor(ghcurSave);
         }
         else break;
#endif
       }
     }
     if(item == IDM_FILE_NEW)goto NEWMODELPOINT;
     if(item == LASTFILECMD+1 || item == LASTFILECMD+2 ||
        item == LASTFILECMD+3 || item == LASTFILECMD+4 ||
        item == LASTFILECMD+5){
       strcpy(gszSCEfile,LastFilesList[item-LASTFILECMD-1].file);
       goto LOADALL_POINT;
     }
     LoadString(ghinst_main,IDX_MISC_LOADFILTER,temp_str,256);
     if(SelectSfxFileName(0,gszSCEfile,gszSCEdir,IDX_MISC_OPENMODEL,
          temp_str,hwnd) == TRUE){
       int i,fid;
       UpdateLastFilesList(gszSCEfile);
       LOADALL_POINT:
       fid=identify_file_format(gszSCEfile);
       ghcurSave=SetCursor(ghcurWait);
       EraseAll(1);
       for(i=0;i<3;i++)UpdateWindow(ghwnd_triview[i]);
       UpdateWindow(ghwnd_view); UpdateWindow(ghwnd_main);
       if(fid == FI_SCE && LoadObject(gszSCEfile,0) == OK){
         int check_flag;
         SetPrgmMessage(IDS_CURRENT,gszSCEfile);
         SendPrgmMessage(IDS_CURRENT,0);
         if(ShadowFlag)check_flag=MF_CHECKED; else check_flag=MF_UNCHECKED;
         CheckMenuItem(GetMenu(ghwnd_main),IDM_ATTRIBUTES_SHADOW,check_flag);
         EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_SAVE,MF_ENABLED);
         strcpy(gszSCEsaved,gszSCEfile);
       }
       else if(fid == FI_DXF || fid == FI_3DS || fid == FI_STL){
         strcpy(gszSCEsaved,"unnamed.mfx");
         if(fid == FI_DXF)LoadXmodel(gszSCEfile,"dxfcon.dll");
         if(fid == FI_3DS)LoadXmodel(gszSCEfile,"3dscon.dll");
         if(fid == FI_STL)LoadXmodel(gszSCEfile,"stlcon.dll");
         SendPrgmMessage(IDS_DEFAULT,0);
         EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_SAVE,MF_GRAYED);
         gszSCEfile[0]='\0';
       }
       UpdateCounters();
       UpdateSkTreeView();
       DrawModel();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+4),0,0); // update maps
       SetCursor(ghcurSave);
     }
     break;
   case IDM_FILE_LOAD_RELATIVE: id=1; goto LOAD_POINT;
   case IDM_FILE_LOAD_ABSOLUTE: id=2; goto LOAD_POINT;
     LOAD_POINT:
     if(SelectSfxFileName(0,gszSCEfile,gszSCEdir,IDX_MISC_INCLUDEMODEL,
        "(*.MFX)|*.mfx|",hwnd) == TRUE){
       int i;
       Empty_Undo();
       ghcurSave=SetCursor(ghcurWait);
       if(LoadObject(gszSCEfile,(short)id) != OK){
         SendPrgmMessage(IDS_FAILINCLUDE,1);
       }
       SendPrgmMessage(IDS_CURRENT,0);
       UpdateCounters();
       UpdateSkTreeView();
       DrawModel();
       SetCursor(ghcurSave);
     }
     break;
   case IDM_FILE_LOAD_PATH:
     if(SelectSfxFileName(0,gszPTHfile,gszPTHdir,IDX_MISC_LOADPATH,
        "(*.PTH)|*.pth|",hwnd) == TRUE){
       Empty_Undo();
       EDIT_ACTION=YES;
       ghcurSave=SetCursor(ghcurWait);
       LoadAApoly(0,gszPTHfile);
       UpdateCounters();
       DrawModel();
       SetCursor(ghcurSave);
     }
     break;
#if !__DEMO__
   case IDM_FILE_SAVE:
     ghcurSave=SetCursor(ghcurWait);
     SaveObject(gszSCEsaved,4);
     SetCursor(ghcurSave);
     break;
   case IDM_FILE_SAVE_ALL:
     if(SelectSfxFileName(1,gszSCEfile,gszSCEdir,IDX_MISC_SAVEMODEL,
                     "(*.MFX)|*.mfx|",hwnd) == TRUE){
       AppendFileExtension(gszSCEfile,".mfx");
       UpdateLastFilesList(gszSCEfile);
       ghcurSave=SetCursor(ghcurWait);
       SaveObject(gszSCEfile,1);
       SetCursor(ghcurSave);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_SAVE,MF_ENABLED);
       strcpy(gszSCEsaved,gszSCEfile);
       SetPrgmMessage(IDS_CURRENT,gszSCEfile);
       SendPrgmMessage(IDS_CURRENT,0);
     }
     break;
   case IDM_FILE_SAVE_OBJECT:
     if(SelectSfxFileName(1,gszSCEfile,gszSCEdir,IDX_MISC_SAVESELECTED,
                     "(*.MFX)|*.mfx|",hwnd) == TRUE){
       AppendFileExtension(gszSCEfile,".mfx");
       ghcurSave=SetCursor(ghcurWait);
       SaveObject(gszSCEfile,0);
       SetCursor(ghcurSave);
     }
     break;
   case IDM_FILE_SAVE_NAMED:{
       char temp[255];
       Empty_Undo();
       strcpy(temp,gszHomeDir);strcat(temp,"objects\\");
       RequestCharString(64,temp,"Root directory",ghwnd_main);
       if(strlen(temp) < 3)break;
       SaveObjects(temp,0,0);
     }
     break;
   case IDM_FILE_SAVE_LETTERS:{
       char temp[255];
       Empty_Undo();
       strcpy(temp,gszHomeDir);strcat(temp,"objects\\");
       RequestCharString(64,temp,"Root directory",ghwnd_main);
       if(strlen(temp) < 3)break;
       SaveObjects(temp,0,1);
     }
     break;
   case IDM_FILE_SAVE_PATH:
     Empty_Undo();
     SavePath();
     break;
#endif
   case IDM_PRIMITIVE_QADD_8:   // TTF
     Save_Undo(0);
     if(BuildTTFfontString(ghwnd_main)){
       Zoom(CENTREWORLD,1);
       DrawModel();
       UpdateCounters();
       UpdateSkTreeView();
     }
     break;
     break;
   case IDM_PRIMITIVE_QADD_9:   // ATM
   case IDM_FILE_IMPORT_ATM:
     Save_Undo(0);
     if(BuildATMfontString(ghwnd_main)){
       Zoom(CENTREWORLD,1);
       DrawModel();
       UpdateCounters();
       UpdateSkTreeView();
     }
     break;
#if 0
   case IDM_FILE_IMPORT_DXF:
     ConvertDXFmodel(ghwnd_main);
     break;
   case IDM_FILE_IMPORT_3DS:
     Convert3DSmodel(ghwnd_main);
     break;
#endif
   case IDM_FILE_IMPORT_TDD:
     ConvertTDDmodel(ghwnd_main);
     break;
   case IDM_FILE_IMPORT_SCN:
     ConvertSCNmodel(ghwnd_main);
     break;
   case IDM_FILE_EXPORT_DXF:
     ExportDXFmodel(ghwnd_main);
     break;
   case IDM_FILE_EXPORT_3DS:
     Export3DSmodel(ghwnd_main);
     break;
   case IDM_FILE_EXPORT_STL:
     ExportModel(NULL,"stlout.dll");
     break;
   case IDM_FILE_EXPORT_VRML:
     ExportModel(NULL,"vrmlout.dll");
     break;
   case IDM_FILE_BOUNDINGBOX_CLEAR:
     ClearWireFrame();
     break;
   case IDM_FILE_BOUNDINGBOX_AUTO:
   case IDM_FILE_BOUNDINGBOX_MAKE:
   case IDM_FILE_BOUNDINGBOX_UNMAKE:
     Empty_Undo();
     if(tool != NOTOOL && tool != NODETOOL){
       SendPrgmQuery(IDQ_TOOLNOTFINISHED,0);
       break;
     }
     if     (item == IDM_FILE_BOUNDINGBOX_MAKE)MakeWireFrame();
     else if(item == IDM_FILE_BOUNDINGBOX_UNMAKE)UnMakeWireFrame();
     else if(item == IDM_FILE_BOUNDINGBOX_AUTO)MakeAutoWireFrame();
     break;
   case IDM_FILE_BEZIER:
     if(ExecuteBezierModeler(ghwnd_main)){
       DrawModel();
       UpdateCounters();
     }
     break;
   case IDM_FILE_TRACEOUTLINE:
     if(TraceImageOutline(ghwnd_main)){
       DrawModel();
       UpdateCounters();
     }
     break;
   case IDM_FILE_VIEWER:
     LoadMDIviewer();
     break;
   case IDM_FILE_HEIGHT:
     if(LoadHeightField(ghwnd_main)){
       DrawModel();
       UpdateCounters();
     }
     break;
   case IDM_FILE_EXTERNAL:
     if(ExecuteExternalModeler(ghwnd_main)){
       DrawModel();
       UpdateCounters();
       UpdateSkTreeView();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     }
     break;
   case IDM_FILE_ANIMATOR:
     ExecuteAnimator(FALSE);
     break;
   case IDM_FILE_UPDATE_ANIMATOR:{
       char load_file[256];
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_UPDATE_ANIMATOR,MF_GRAYED);
       in_subroutine_mode=FALSE;
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_ANIMATOR,MF_ENABLED);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_EXIT,MF_ENABLED);
       GetPrivateProfileString("CALLBACK","FILE","$$$",
           load_file,sizeof(load_file),IniFilename);
       if(strncmp(load_file,"$$$",3) == 0){
         SendPrgmQuery(IDQ_MODELERACTIVATEFAIL,0);
         break;
       }
       if(EDIT_ACTION && SendPrgmQuery(IDQ_SAVEBEFOREUPDATING,1) == IDYES){
         ghcurSave=SetCursor(ghcurWait);
#if !__DEMO__
         SaveObject(load_file,3);
#endif
       }
       else ghcurSave=SetCursor(ghcurWait);
       Empty_Undo();
       EraseAll(1);
       DrawModel();
       UpdateCounters();
       SetCursor(ghcurSave);
       ExecuteAnimator(TRUE);
     }
     break;
   case IDM_FILE_PRINT:
     PrintImageFile(ghwnd_main);
     break;
   case IDM_CONTEXT_TOOLSLEFT:
     bToolpannelLeft ^= 1;
     UpdateIconPositions();
     break;
   case IDM_CONTEXT_TOOLSSMALL:
     bToolpannelSmall ^= 1;
     ChangeToolboxBitmaps();
     UpdateIconPositions();
     break;
   case IDM_FILE_PREFERENCES:{
       PreferencesPropertyDlgProc(ghwnd_main);
       SendMessage(ghwndTooltip,TTM_ACTIVATE,(WPARAM)Preferences.tooltips,0);
       if(Preferences.toolleft  != bToolpannelLeft ||
          Preferences.toolsmall != bToolpannelSmall){
         if(bToolpannelSmall != Preferences.toolsmall){
           bToolpannelSmall=Preferences.toolsmall;
           ChangeToolboxBitmaps();
         }
         bToolpannelLeft=Preferences.toolleft;
         UpdateIconPositions();
       }
     }
     break;
   case IDM_FILE_EXIT:
     if(EDIT_ACTION){
       int fResult;
       fResult=SendPrgmQuery(IDQ_CHANGED,2);
       if(fResult == IDNO)PostQuitMessage(0L);
       if(fResult == IDYES){
#if !__DEMO__
         if(SelectSfxFileName(1,gszSCEfile,gszSCEdir,IDX_MISC_SAVEMODEL,
                           "(*.MFX)|*.mfx|",hwnd) == TRUE){
          AppendFileExtension(gszSCEfile,".mfx");
           ghcurSave=SetCursor(ghcurWait);
           SaveObject(gszSCEfile,1);
           SetCursor(ghcurSave);
           PostQuitMessage(0L);
         }
#endif
       }
       break;
     }
//     else if(SendPrgmQuery(IDQ_CONFIRM,1) == IDYES)PostQuitMessage(0L);
     else PostQuitMessage(0L);
     break;
   case IDM_CALLBACK_SHUTDOWN:
     if(EDIT_ACTION)PostMessage(ghwnd_main,WM_COMMAND,IDM_FILE_EXIT,0);
     else PostQuitMessage(0L);
     break;
   case IDM_CALLBACK_LOAD_MODELER: {
      char load_file[255];
      if(EDIT_ACTION){
        HWND hWndPrev;
        SendPrgmQuery(IDQ_MODELERBUSY,0);
        PostMessage(ghwnd_main,WM_COMMAND,IDM_FILE_ANIMATOR,0);
        hWndPrev = FindWindow("OFX:AnimatorClass",NULL);
        if(hWndPrev != NULL)PostMessage(hWndPrev,WM_COMMAND,
                            IDM_CALLBACK_NORELOAD,0);
        break;
      }
      GetPrivateProfileString("CALLBACK","FILE","$$$",
          load_file,sizeof(load_file),IniFilename);
      if(strncmp(load_file,"$$$",3) == 0){
         SendPrgmQuery(IDQ_MODELERACTIVATEFAIL,0);
         break;
       }
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_UPDATE_ANIMATOR,MF_ENABLED);
       in_subroutine_mode=TRUE;
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_ANIMATOR,MF_GRAYED);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_EXIT,MF_GRAYED);
       strcpy(gszSCEfile,load_file);
       strcpy(gszSCEdir,load_file);
       *FileInPath(gszSCEdir) = '\0';
       goto LOADALL_POINT;
     }
     break;
   case IDM_FILE_LOAD_DIRECT: {
      char load_file[255];
      GetPrivateProfileString("CALLBACK","FILE","$$$",
          load_file,sizeof(load_file),IniFilename);
      if(strncmp(load_file,"$$$",3) == 0)break;
       strcpy(gszSCEfile,load_file);
       strcpy(gszSCEdir,load_file);
       *FileInPath(gszSCEdir) = '\0';
       goto LOADALL_POINT;
     }
     break;
   case IDM_EDIT_UNDO:
     Restore_Undo();
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;
   case IDM_KEY_DELETE:
     Save_Undo(2);
     bUpdate=DeleteSelectedNurbsSurfaces();
   case IDM_EDIT_ERASE_VERTICES:
     if(!bUpdate)Save_Undo(2);
     if(EraseVertex(YES) || bUpdate){
       EDIT_ACTION=YES;
       DrawModel();
       UpdateCounters();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     }
     break;
   case IDM_EDIT_ERASE_EDGES:
     Save_Undo(2);
     if(EraseEdge(YES)){
       EDIT_ACTION=YES;
       DrawModel();
       UpdateCounters();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     }
     break;
   case IDM_EDIT_ERASE_ALL:
     if(SendPrgmQuery(IDQ_CONFIRM,1) == IDYES){
       NEWMODELPOINT:
       Empty_Undo();
       EraseAll(1);
       DrawModel();
       UpdateCounters();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+4),0,0); // update everything
       SendPrgmMessage(IDS_DEFAULT,0);
       EnableMenuItem(GetMenu(ghwnd_main),IDM_FILE_SAVE,MF_GRAYED);
       strcpy(gszSCEsaved,"unnamed.mfx");
       gszSCEfile[0]='\0';
     }
     break;
   case IDM_EDIT_ERASE_INDICATEDEDGE:
     Save_Undo(2);
     if(EraseIndicatedEdge()){
       EDIT_ACTION=YES;
       DrawModel();
       UpdateCounters();
     }
     break;
   case IDM_EDIT_ERASE_INDICATEDVERTEX:
     Save_Undo(2);
     if(EraseIndicatedVertex()){
       EDIT_ACTION=YES;
       DrawModel();
       UpdateCounters();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     }
     break;
   case IDM_EDIT_ERASE_HIERARCHY:
     EDIT_ACTION=YES;
     EraseAllSkeleton();
     break;
   case IDM_EDIT_ERASE_VERTEXMAPPING:
     EDIT_ACTION=YES;
     EraseVertexMapping();
     break;

   case IDM_EDIT_SELECTALL:
   case IDM_EDIT_DESELECTALL:
   case IDM_EDIT_SWAPSELECT:
   case IDM_EDIT_SELECTCONNECT:
   case IDM_EDIT_DESELECTCONNECT:
   case IDM_EDIT_SELECTBY_COLOUR:
   case IDM_EDIT_SELECTBY_MAP:
   case IDM_EDIT_SELECTBY_SHADER:
   case IDM_EDIT_SELECTBY_BEZIER:
   case IDM_EDIT_SELECTBY_INVIEW:
   case IDM_EDIT_SELECTBY_OUTVIEW:
   case IDM_EDIT_SELECTMAPPED:
   case IDM_EDIT_DESELECTMAPPED:
   case IDM_EDIT_DESELECT_EVERY:
     Empty_Undo();
     ExecuteSelect(item);
     break;

   case IDM_TOOLS_NODE:
     if(tool != NODETOOL)EngageDesignerTool(NODETOOL);
     break;
   case IDM_TOOLS_PAN:
     if(tool != PAN)EngageDesignerTool(PAN);
     break;
   case IDM_TOOLS_ZOOMIN:
     if(tool != INZOOM)EngageDesignerTool(INZOOM);
     break;
   case IDM_TOOLS_GRABBER:
     if(tool != GRABBER)EngageDesignerTool(GRABBER);
     break;
   case IDM_TOOLS_BUILDER:
     if(tool != BUILDER)EngageDesignerTool(BUILDER);
     break;
   case IDM_TOOLS_EXPANDER:
     if(tool != EXPANDER)EngageDesignerTool(EXPANDER);
     break;
   case IDM_TOOLS_ROTATOR:
     if(tool != ROTATOR)EngageDesignerTool(ROTATOR);
     break;
   case IDM_TOOLS_PLOTTER:
     if(tool != PLOTTER && tool != PLOT3D)EngageDesignerTool(PLOTTER);
     break;
   case IDM_TOOLS_SHAPER:
     if(tool != SHAPER)EngageDesignerTool(SHAPER);
     else EngageDesignerTool(MAGNETTO);
     break;
   case IDM_TOOLS_MAGNET_ATTRACT:
     if(tool != MAGNETTO && tool != SHAPER)EngageDesignerTool(MAGNETTO);
     break;
   case IDM_TOOLS_PLOT3D:
     if(tool != PLOT3D)EngageDesignerTool(PLOT3D);
     else EngageDesignerTool(PLOTTER);
     break;

  case IDM_TOOLS_SELECTOR:
     if(tool != SELECTOR && tool != DESELECTOR)EngageDesignerTool(SELECTOR);
     break;
   case IDM_TOOLS_DESELECTOR:
     if(tool != DESELECTOR)EngageDesignerTool(DESELECTOR);
     else EngageDesignerTool(SELECTOR);
     break;
   case IDM_TOOLS_PRIMITIVES:
     if(tool != PRIMS)EngageDesignerTool(PRIMS);
     break;
   case IDM_PRIMITIVE_QADD:
     ExecutePrimitivesMenu(hwnd);
     break;
   case IDM_PRIMITIVE_QADD_1:
   case IDM_PRIMITIVE_QADD_2:
   case IDM_PRIMITIVE_QADD_3:
   case IDM_PRIMITIVE_QADD_4:
   case IDM_PRIMITIVE_QADD_5:
   case IDM_PRIMITIVE_QADD_6:
   case IDM_PRIMITIVE_QADD_7:
     AddSinglePrimitive(item);
     break;
   case IDM_SNAP_CLOSEST:
   case IDM_SNAP_CLOSESELECT:
   case IDM_SNAP_MIDPOINT:
   case IDM_SNAP_MIDSELECT:
   case IDM_SNAP_CENTROID:
   case IDM_SNAP_HPLANE:
   case IDM_SNAP_VPLANERL:
   case IDM_SNAP_VPLANEFB:
   case IDM_SNAP_STORE:
   case IDM_SNAP_TOSTORED:
   case IDM_SNAP_TOCOORD:
   case IDM_SNAP_INCREMENT:
     Empty_Undo();
     SnapCursorTo(item);
     break;
   case IDM_ACTIONS_BOOLEAN:
   case IDM_ACTIONS_TWIST:
   case IDM_ACTIONS_ORIENTATE:
   case IDM_ACTIONS_REVERSE:
   case IDM_ACTIONS_JOINS:
   case IDM_ACTIONS_MOVE:
   case IDM_ACTIONS_BUILD:
   case IDM_ACTIONS_SMOOTH:
   case IDM_ACTIONS_EFFECTS_PLAN:
   case IDM_ACTIONS_FACET:
   case IDM_ACTIONS_EXTRUDE_BEVEL:
   case IDM_ACTIONS_EXTRUDE_ROUND:
   case IDM_ACTIONS_BEZIER:
   case IDM_ACTIONS_EXPAND:
   case IDM_ACTIONS_FLIP:
   case IDM_ACTIONS_DUPLICATE_CONNECT:
   case IDM_ACTIONS_SPIN:
   case IDM_ACTIONS_DUPLICATE_MULTI:
   case IDM_ACTIONS_JOIN:
   case IDM_ACTIONS_EFFECTS_SHATTER:
   case IDM_ACTIONS_EXTRUDE_ALONG:
   case IDM_ACTIONS_EFFECTS_EXPLODE:
   case IDM_ACTIONS_ROTATE_CLOCKWISE:
   case IDM_ACTIONS_ROTATE_ANTICLOCKWISE:
   case IDM_ACTIONS_WELD:
   case IDM_ACTIONS_FILL:
   case IDM_ACTIONS_SUBDIVIDE:
   case IDM_TOOLS_EXTRUDE:
   case IDM_ACTIONS_DUPLICATE_MAKE:
   case IDM_TOOLS_ADDVERTEX:
   case IDM_TOOLS_BUILDFACET:
     ExecuteDesignerAction(item);
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;
   case IDM_ACTIONS_SPECIAL_3DPLOT:
   case IDM_ACTIONS_SPECIAL_HELIX:
   case IDM_ACTIONS_SPECIAL_SPIRAL:
   case IDM_ACTIONS_SPECIAL_3DSPIRAL:
   case IDM_ACTIONS_SPECIAL_LINE:
   case IDM_ACTIONS_SPECIAL_BUTTERFLY:
     EDIT_ACTION=YES;
     SpecialFunction(item);
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;

   case IDM_ATTRIBUTES_FACE:
     Empty_Undo();
     if(!IsWindow(ghwndDlgAttribs)){
       ghwndDlgAttribs=CreateDialog(ghinst_main,
                       MAKEINTRESOURCE(DLG_ATTRIBS),
                       ghwnd_main,(DLGPROC)AttribsDlgProc);
     }
     else{
        DestroyWindow(ghwndDlgAttribs);
        ghwndDlgAttribs=NULL;
     }
     break;
   case IDM_ATTRIBUTES_DEFAULT:
     DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_DEFAULT),ghwnd_main,
               (DLGPROC)DefaultAttribsDlgProc);
     break;
   case IDM_ATTRIBUTES_NEWMAP:
     if(axis_move_flag >= 0)break;
     Empty_Undo();
     DisengageDesignerTool();
     EditImageMap(-2);
     break;
   case IDM_ATTRIBUTES_MAP:
     Empty_Undo();
     DisengageDesignerTool();
     if(axis_move_flag >= 0){
       int id;
       id=axis_move_flag;
       axis_move_flag = -1;
       EditImageMap(id);
     }
     else {
       EditImageMap(-1);
     }
     break;
   case IDM_ATTRIBUTES_MAPAXES:
     DisengageDesignerTool();
     if(SelectedBrush < 0)PositionBrushOnOff(1,-1);
     else                 PositionBrushOnOff(0,SelectedBrush);

     break;
   case IDM_ATTRIBUTES_PURGE:
     PurgeMaps();
     break;
   case IDM_ATTRIBUTES_SHADER_PURGE:
     PurgeShaders();
     break;
   case IDM_ATTRIBUTES_VECTOR:
     VectorMaps();
     break;
   case IDM_ATTRIBUTES_SHADER_MOVE:
     DisengageDesignerTool();
     if(SelectedShader < 0)PositionShaderOnOff(1,-1);
     else                  PositionShaderOnOff(0,SelectedShader);
     break;
   case IDM_ATTRIBUTES_SHADER_NEW:
     AddAdditionalShader();
     break;
   case IDM_ATTRIBUTES_SHADER_DELETE:
     DumpShader(-1);
     break;
   case IDM_ATTRIBUTES_SHADER_PREVIEW:
     ShaderPreview();
     break;
   case IDM_ATTRIBUTES_CENTREPOINT:
     ObjectAxis.Origin.xyz[0]=NpointerX;
     ObjectAxis.Origin.xyz[1]=NpointerY;
     ObjectAxis.Origin.xyz[2]=NpointerZ;
     EDIT_ACTION=YES;
     DrawModel();
     break;
   case IDM_ATTRIBUTES_FOLLOWPOINT:
     ObjectAxis.Offset.xyz[0]=NpointerX;
     ObjectAxis.Offset.xyz[1]=NpointerY;
     ObjectAxis.Offset.xyz[2]=NpointerZ;
     EDIT_ACTION=YES;
     DrawModel();
     break;
   case IDM_ATTRIBUTES_SHADOW:{
       int check_flag;
       EDIT_ACTION=YES;
       ShadowFlag ^= 1;
       if(ShadowFlag)check_flag=MF_CHECKED; else check_flag=MF_UNCHECKED;
       CheckMenuItem(GetMenu(ghwnd_main),IDM_ATTRIBUTES_SHADOW,check_flag);
     }
     break;
   case IDM_ATTRIBUTES_MIRROR:
     MirrorFlag=RequestMirrorID(ghwnd_main);
     break;
   case IDM_ATTRIBUTES_REFINDEX:{
      char temp[32],title[32];
      double ri;
      sprintf(temp,"%lf",refractive_index);
      LoadString(ghinst_main,IDX_MISC_REFINDEX,title,32);
      RequestCharString(32,temp,title,ghwnd_main);
      ri=atof(temp);
      if(ri > 0.0 && ri < 10.0){
        EDIT_ACTION=YES;
        refractive_index=ri;
      }
     }
     break;
   case IDM_ATTRIBUTES_MAPFROMMAP:
     Empty_Undo();
     MapVerticesByBrush();
     break;
   case IDM_ATTRIBUTES_MAPONPATH:
     Empty_Undo();
     MapOntoPath();
     break;
   case IDM_ATTRIBUTES_SETVERTEXMAPPING:
     Empty_Undo();
     SetMapCoords();
     break;
   case IDM_ATTRIBUTES_EXTERNALSHADER:
     EnableToolPannels(ALL_PANNELS,FALSE);
     DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_SHADERLOAD),ghwnd_main,
               (DLGPROC)LoadExternalShaderDlgProc);
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case IDM_HIERARCHY_NAME:{
       static BOOL in_hm=FALSE;
       if(Preferences.attribs_sheet){
         Empty_Undo();
         EDIT_ACTION=YES;
         SkeletonTreeView(ghwnd_main);
       }
       else{
         if(in_hm)break;
         in_hm=TRUE;
         Empty_Undo();
         EDIT_ACTION=YES;
         DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_SKEL),ghwnd_main,
                   (DLGPROC)HierarchyDlgProc);
         in_hm=FALSE;
       }
     }
     break;
   case IDM_HIERARCHY_SELECT:
   case IDM_HIERARCHY_POSE:
   case IDM_HIERARCHY_STRETCH:
   case IDM_HIERARCHY_BUILD:
   case IDM_HIERARCHY_REMOVE:
   case IDM_HIERARCHY_INSERT:
   case IDM_HIERARCHY_ATTACH:
   case IDM_HIERARCHY_MAP:
   case IDM_HIERARCHY_TWIST:
     EngageSkeletonTool(item);
     break;
   case IDM_VIEW_HIDE:
     ExecuteHide(TRUE);  DrawModel(); UpdateCounters();
     if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,(WM_USER+2),0,0);
     break;
   case IDM_VIEW_REVEAL:
     ExecuteHide(FALSE); DrawModel(); UpdateCounters();
     if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,(WM_USER+2),0,0);
     break;
   case IDM_ZOOM_IN:
     Zoom(ZOOMIN,1);
     break;
   case IDM_ZOOM_OUT:
     Zoom(ZOOMOUT,1);
     break;
   case IDM_ZOOM_CENTRE:
     Zoom(RECENTRE,1);
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;
   case IDM_VIEW_RESTORE:
     GetTriview(TRUE);
     NpointerX=NpointerY=NpointerZ=0;
     DrawModel();
     UpdateRuler(0);
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;
   case IDM_VIEW_FULL:
     Zoom(CENTREWORLD,1);
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;
   case IDM_VIEW_DRAWING_HIDEEDGES:
     global_quickdraw ^= 1;
     if(global_quickdraw)CheckMenuItem(GetMenu(ghwnd_main),
                         IDM_VIEW_DRAWING_HIDEEDGES,MF_CHECKED);
     else                CheckMenuItem(GetMenu(ghwnd_main),
                         IDM_VIEW_DRAWING_HIDEEDGES,MF_UNCHECKED);
     DrawModel();
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;
   case IDM_VIEW_DRAWING_HIDEJOINS:
     draw_joins ^= 1;
     if(!draw_joins)CheckMenuItem(GetMenu(ghwnd_main),
                    IDM_VIEW_DRAWING_HIDEJOINS,MF_CHECKED);
     else           CheckMenuItem(GetMenu(ghwnd_main),
                    IDM_VIEW_DRAWING_HIDEJOINS,MF_UNCHECKED);
     DrawModel();
     if(ghwndOpenGLview == NULL)Draw3dView(0,0);
     else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     break;
   case IDM_VIEW_RENDERNOW:
     Render_View(0,1);
     break;
   case IDM_WINDOW_QUICKRENDER:
   case IDM_VIEW_RENDERINWINDOW:
     Render_View(0,2);
     break;
   case IDM_VIEW_RENDER:
     Render_View(0,0);
     break;
   case IDM_VIEW_SHOW:{
       int result;
       char modulename[255],cp[128],cn[128];
       GetPrivateProfileString("VIEWER","FILE","$$$",
          cn,sizeof(cn),IniFilename);
       GetPrivateProfileString("VIEWER","PATH","$$$",
          cp,sizeof(cp),IniFilename);
       if(strncmp(cn,"$$$",3) == 0 || strncmp(cp,"$$$",3) == 0){
         SendPrgmQuery(IDQ_NOVIEWIMAGE,0);
       }
       else LoadViewer(1);
     }
     break;
   case IDM_VIEW_WIREFRAME:
     if(ghwndOpenGLview == NULL)Draw3dView(0,1);
     break;
   case IDM_VIEW_HIDDENLINE:
     if(ghwndOpenGLview == NULL)Draw3dView(1,0);
     break;
   case IDM_VIEW_DRAW_GRID:
     draw_grid_on ^= 1;
     if(draw_grid_on)
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_CHECKED);
     else
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_UNCHECKED);
     DrawModel();
     break;
   case IDM_VIEW_GRID:
     grid_on ^= 1;
     if(grid_on)CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_GRID,MF_CHECKED);
     else       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_GRID,MF_UNCHECKED);
//     if(draw_grid_on)DrawModel();
     break;
   case IDM_VIEW_DEFINEGRID:
     SetUpGrid();
     if(draw_grid_on)
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_CHECKED);
     else
       CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_UNCHECKED);
     DrawModel();
     break;
   case IDM_VIEW_FEEDBACK_CURSOR:{
       char tt[128];
       Draw3dCursor();
       CursorToggle ^= 1;
       Draw3dCursor();
       if(CursorToggle){
         LoadString(ghinst_main,IDX_MISC_CROSSHAIR,tt,128);
         ModifyMenu(GetMenu(ghwnd_main),IDM_VIEW_FEEDBACK_CURSOR,
          MF_BYCOMMAND|MF_STRING,IDM_VIEW_FEEDBACK_CURSOR,tt);
       }
       else{
         LoadString(ghinst_main,IDX_MISC_CROSSWIRE,tt,128);
         ModifyMenu(GetMenu(ghwnd_main),IDM_VIEW_FEEDBACK_CURSOR,
          MF_BYCOMMAND|MF_STRING,IDM_VIEW_FEEDBACK_CURSOR,tt);
       }
     }
     break;
   case IDM_WINDOW_SWITCHWINDOWS:
     SwitchWindow(-1);
     break;
   case IDM_VIEW_SWAPVIEWPOINT:
     WindowBox_view ^= 1;
     SetUpWindowBoxView(WindowBox_view);
     SwapViewBitmaps();
     if(WindowBox_view)CheckMenuItem(GetMenu(ghwnd_main),
                       IDM_VIEW_SWAPVIEWPOINT,MF_CHECKED);
     else              CheckMenuItem(GetMenu(ghwnd_main),
                       IDM_VIEW_SWAPVIEWPOINT,MF_UNCHECKED);
     DrawModel();
     break;
   case IDM_WINDOW_SWITCHWINDOW0:
     SwitchWindow(0);
     break;
   case IDM_WINDOW_SWITCHWINDOW1:
     SwitchWindow(1);
     break;
   case IDM_WINDOW_COORDS:{
       HMENU hmenu;
       int cv[2]={SW_HIDE,SW_SHOWNA};
       hmenu=GetMenu(ghwnd_main);
       coords_visible ^=1;
       if(IsWindow(ghwndCoord1))ShowWindow(ghwndCoord1,cv[coords_visible]);
       if(coords_visible)CheckMenuItem(hmenu,IDM_WINDOW_COORDS,MF_CHECKED);
       else              CheckMenuItem(hmenu,IDM_WINDOW_COORDS,MF_UNCHECKED);
     }
     break;
   case IDM_WINDOW_ACTIVATE:{
        RECT  r;
        POINT p;
        HDC hdc;
        int i;
        if(sktool == YES && tool == SKANIMATE){
          TabSkeletonAxis();
          break;
        }
        if(View == ONEVIEW){
          if(ActiveView == TRIFRONT)
            PostMessage(hwnd,WM_COMMAND,IDM_WINDOW_SWITCHWINDOW1,0);
          else
            PostMessage(hwnd,WM_COMMAND,IDM_WINDOW_SWITCHWINDOW0,0);
          break;
        }
        hdc=GetDC(ghwnd_main);
        FocusActiveWindow(hdc,FALSE);
        ActiveView++; if(ActiveView > 2)ActiveView=0;
        ghwnd_current=ghwnd_triview[ActiveView];
        FocusActiveWindow(hdc,FALSE);
        ReleaseDC(ghwnd_main,hdc);
      }
      break;
   case IDM_WINDOW_DATA:{
       HMENU hmenu;
       int cv[2]={SW_HIDE,SW_SHOW};
       hmenu=GetMenu(ghwnd_main);
       if(ghwndData)data_visible ^=1;
       if(ghwndData)ShowWindow(ghwndData,cv[data_visible]);
       if(data_visible)CheckMenuItem(hmenu,IDM_WINDOW_DATA,MF_CHECKED);
       else            CheckMenuItem(hmenu,IDM_WINDOW_DATA,MF_UNCHECKED);
     }
     break;
   case IDM_WINDOW_TOOLLEFT:
     break;
   case IDM_WINDOW_RESET:
     Window0sizeX=TriviewSpaceX/2;
     Window0sizeY=TriviewSpaceY/2;
     if(View == TRIVIEW)FormatScreen();
     break;
   case IDM_WINDOW_SHOWALL:
     ToolboxShow(FALSE);
     ToolboxShow(TRUE);
     break;
   case IDM_WINDOW_SHOWOPENGL:
     StartOpenGL(FALSE);        // load maps as well
     if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,(WM_USER+4),0,0);
     break;
   case IDM_STOP:
     if(sktool == YES)DisengageSkeletonTool();
     else if(SelectedBrush >= 0)PositionBrushOnOff(0,SelectedBrush);
     else if(SelectedShader >= 0)PositionShaderOnOff(0,SelectedShader);
     else if(tool != NOTOOL)DisengageDesignerTool();
     break;
   case IDM_HELP_HOME:{
       char lname[256],res_str[256];
#if __LITE__
       LoadString(ghinst_main,IDX_MISC_HOMEPAGEL,res_str,256);
#else
       LoadString(ghinst_main,IDX_MISC_HOMEPAGE,res_str,256);
#endif
       strcpy(lname,gszHomeDir);
       ShellExecute(ghwnd_main,NULL,res_str,NULL,lname,0);
     }
     break;
   case IDM_HELP_SUPPORT:{
       char lname[256],res_str[256];
#if __LITE__
       LoadString(ghinst_main,IDX_MISC_SUPPORTPAGEL,res_str,256);
#else
       LoadString(ghinst_main,IDX_MISC_SUPPORTPAGE,res_str,256);
#endif
       strcpy(lname,gszHomeDir);
       ShellExecute(ghwnd_main,NULL,res_str,NULL,lname,0);
     }
     break;
   case IDM_HELP:{
       char lname[256],res_str[256],res_str1[256],res_str2[256];
       LoadString(ghinst_main,IDX_MISC_HELPFILENAME,res_str,256);
       strcpy(lname,gszHomeDir); strcat(lname,"help\\");
       strcpy(res_str1,lname);
       strcat(res_str1,res_str);
#if !__HELP_WS__
       ShellExecute(ghwnd_main,NULL,res_str1,
                    NULL,lname,SW_SHOW);
#else
       {
         HH_WINTYPE hhwin;
         char localname[32],localname1[32]=">";
         sprintf(localname,"MHelp%ld",rand());
         strcat(localname1,localname);
         memset(&hhwin,0,sizeof(HH_WINTYPE));
         hhwin.cbStruct=sizeof(HH_WINTYPE);
         hhwin.fsValidMembers = HHWIN_PARAM_SHOWSTATE |
                                HHWIN_PARAM_PROPERTIES|
                                HHWIN_PARAM_CUR_TAB|
                                HHWIN_PARAM_EXPANSION|
                                HHWIN_PARAM_RECT;
         hhwin.rcWindowPos.top=0;
         hhwin.rcWindowPos.left=0;
         hhwin.rcWindowPos.right=700;
         hhwin.rcWindowPos.bottom=500;
         hhwin.nShowState=SW_SHOW;
         hhwin.pszType=localname;
         LoadString(ghinst_main,IDX_MISC_HELPTITLE1,res_str,256);
         hhwin.pszCaption=res_str;
         hhwin.fNotExpanded = FALSE;
         hhwin.fsWinProperties = HHWIN_PROP_TRI_PANE|HHWIN_PROP_TAB_SEARCH;
         hhwin.curNavType=HHWIN_NAVTYPE_TOC;
         strcpy(res_str2,res_str1);
         strcat(res_str2,"::toc.hhc");
         hhwin.pszToc=res_str2;
         HtmlHelp(ghwnd_main,NULL,HH_SET_WIN_TYPE,(DWORD)&hhwin);
         hWndHelp=HtmlHelp(ghwnd_main,localname1,HH_DISPLAY_TOPIC,(DWORD)res_str1);
       }
#endif
     }
     break;
   case IDM_HELP_GENERAL:{
       char lname[256],res_str[256],res_str1[256],res_str2[256];
       LoadString(ghinst_main,IDX_MISC_HELPFILENAME1,res_str,256);
       strcpy(lname,gszHomeDir); strcat(lname,"help\\");
       strcpy(res_str1,lname);
       strcat(res_str1,res_str);
#if !__HELP_WS__
       ShellExecute(ghwnd_main,NULL,res_str1,
                    NULL,lname,SW_SHOW);
#else
       {
         HH_WINTYPE hhwin;
         char localname[32],localname1[32]=">";
         sprintf(localname,"GHelp%ld",rand());
         strcat(localname1,localname);
         memset(&hhwin,0,sizeof(HH_WINTYPE));
         hhwin.cbStruct=sizeof(HH_WINTYPE);
         hhwin.fsValidMembers = HHWIN_PARAM_SHOWSTATE |
                                HHWIN_PARAM_PROPERTIES|
                                HHWIN_PARAM_CUR_TAB|
                                HHWIN_PARAM_EXPANSION|
                                HHWIN_PARAM_RECT;
         hhwin.rcWindowPos.top=0;
         hhwin.rcWindowPos.left=0;
         hhwin.rcWindowPos.right=700;
         hhwin.rcWindowPos.bottom=500;
         hhwin.nShowState=SW_SHOW;
         hhwin.pszType=localname;
         LoadString(ghinst_main,IDX_MISC_HELPTITLE2,res_str,256);
         hhwin.pszCaption=res_str;
         hhwin.fNotExpanded = FALSE;
         hhwin.fsWinProperties = HHWIN_PROP_TRI_PANE|HHWIN_PROP_TAB_SEARCH;
         hhwin.curNavType=HHWIN_NAVTYPE_TOC;
         strcpy(res_str2,res_str1);
         strcat(res_str2,"::toc.hhc");
         hhwin.pszToc=res_str2;
         HtmlHelp(ghwnd_main,NULL,HH_SET_WIN_TYPE,(DWORD)&hhwin);
         hWndHelp=HtmlHelp(ghwnd_main,localname1,HH_DISPLAY_TOPIC,(DWORD)res_str1);
       }
#endif
     }
     break;
   case IDM_HELP_HELP:
//     WinHelp(hwnd,"WINHELP.HLP",HELP_INDEX,0L);
     break;
   case IDM_HELP_ABOUT:
     DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_ABOUT),ghwnd_main,
          (DLGPROC)AboutDlgProc);
     break;
   case IDM_COORDS_DEFINE:
     UpdateRuler(3);
     break;
   case IDM_COORDS_ORIGIN:
     UpdateRuler(2);
     break;
   case IDM_COORDS_RELATIVE:
     UpdateRuler(1);
     break;
   case IDM_COORDS_MEASURE:
     UpdateRuler(0);
     break;
   case IDM_PAN_UP1:
     if(lparam == TRITOP  )Zoom(PANFORWARD,1);
     if(lparam == TRIFRONT)Zoom(PANUP,1);
     if(lparam == TRIRIGHT)Zoom(PANUP,1);
     break;
   case IDM_PAN_UP:
     Nudge(item);
     break;
   case IDM_PAN_DOWN1:
     if(lparam == TRITOP  )Zoom(PANBACK,1);
     if(lparam == TRIFRONT)Zoom(PANDOWN,1);
     if(lparam == TRIRIGHT)Zoom(PANDOWN,1);
     break;
   case IDM_PAN_DOWN:
     Nudge(item);
     break;
   case IDM_PAN_LEFT1:
     if(lparam == TRITOP  )Zoom(PANLEFT,1);
     if(lparam == TRIFRONT)Zoom(PANLEFT,1);
     if(lparam == TRIRIGHT)Zoom(PANBACK,1);
     break;
   case IDM_PAN_LEFT:
     Nudge(item);
     break;
   case IDM_PAN_RIGHT1:
     if(lparam == TRITOP  )Zoom(PANRIGHT,1);
     if(lparam == TRIFRONT)Zoom(PANRIGHT,1);
     if(lparam == TRIRIGHT)Zoom(PANFORWARD,1);
     break;
   case IDM_PAN_RIGHT:
     Nudge(item);
     break;
   case IDM_LOCK_SQUARE:
   case IDM_LOCK_CIRCLE:
   case IDM_LOCK_TRIANGLE:
   case IDM_LOCK_LOCK:
     ExecuteLockAction(item,SelectedBrush);
     break;
   case IDM_GLVIEW_ABOUT:
   case IDM_GLVIEW_SCROLLBARS:
   case IDM_GLVIEW_TWOSIDE:
   case IDM_GLVIEW_OVERLAY:
   case IDM_GLVIEW_PREFS:
   case IDM_GLVIEW_HELP:
   case IDM_GLVIEW_MAP:
     if(ghwndOpenGLview != NULL){
       SendMessage(ghwndOpenGLview,WM_COMMAND,item,0);
     }
     break;
   case IDM_NURBS_PRIM1:
     MakeNewNurbsSurface(1);
     break;
   case IDM_NURBS_PRIM2:
     MakeNewNurbsSurface(2);
     break;
   case IDM_NURBS_PRIM3:
     MakeNewNurbsSurface(3);
     break;
   case IDM_NURBS_PRIM4:
     MakeNewNurbsSurface(4);
     break;
   case IDM_NURBS_CREATE:
     MakeNewNurbsSurface(0);
     break;
   case IDM_NURBS_SELECT:
     SelectAllNurbs(TRUE);
     break;
   case IDM_NURBS_DESELECT:
     SelectAllNurbs(FALSE);
     break;
   case IDM_NURBS_SWAP:
     SwapNurbsSelection();
     break;
   case IDM_NURBS_DELETE:
     Save_Undo(2);
     if(DeleteSelectedNurbsSurfaces()){
       EDIT_ACTION=YES;
       DrawModel();
       UpdateCounters();
       if(ghwndOpenGLview == NULL)Draw3dView(0,0);
       else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
     }
     break;
   case IDM_NURBS_EDIT:
     EditSelectedNurbsSurface();
     break;
   case IDM_NURBS_HIDE:
     HideNurbs(TRUE);
     break;
   case IDM_NURBS_REVEAL:
     HideNurbs(FALSE);
     break;
   case IDM_NURBS_SETWEIGHTS:
     SetSelectedNurbsWeights();
     break;
   case IDM_NURBS_SPLIT:
     Save_Undo(2);
     SplitNurbs();
     break;
   case IDM_NURBS_TRIANGULATE2:
     Save_Undo(0);
     MeshNurbsSurface();
     break;
   case IDM_NURBS_TRIANGULATE1:
     Save_Undo(0);
     TriangulateNurbsSurface();
     break;
   default: break;
 }
 return;
}

void ExecuteKeyboardKeyCommand(int key){
  switch(key){
    case VK_ESCAPE:{
        HWND hWnd;
        if(tool != NOTOOL)SendMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
        if((hWnd=FindWindow("OFX:RenderClass",NULL)) != NULL){
          PostMessage(hWnd,WM_CALLBACK_ABANDON,0,0);
        }
      }
      break;
    case VK_PRIOR:
//      Nudge(IDM_PAN_UP);
      break;
    case VK_NEXT:
//      Nudge(IDM_PAN_DOWN);
      break;
    case VK_HOME:
//      Nudge(IDM_PAN_LEFT);
      break;
    case VK_END:
//      Nudge(IDM_PAN_RIGHT);
      break;
    default: break;
  }
}

static void EngageSkeletonTool(int item){
 EDIT_ACTION=YES;
 DisengageDesignerTool();
 Empty_Undo();
 EnableToolPannels(SELECT_PANNEL,FALSE);
 ActivateAllMenus(ghwnd_main,MF_GRAYED);
 EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_ENABLED);
 DrawMenuBar(ghwnd_main);
 sktool=YES;
 SetSkToolbarState(FALSE);
 SetCursor(ghcurArrowD);
 switch (item){
   case IDM_HIERARCHY_POSE:    tool=SKANIMATE; break;
   case IDM_HIERARCHY_BUILD:   tool=SKADD;     break;
   case IDM_HIERARCHY_STRETCH: tool=SKGRAB;    break;
   case IDM_HIERARCHY_TWIST:   tool=SKTWIST;   break;
   case IDM_HIERARCHY_REMOVE:  tool=SKDELETE;  break;
   case IDM_HIERARCHY_INSERT:  tool=SKSUBDIV;  break;
   case IDM_HIERARCHY_ATTACH:  tool=SKASSIGN;  break;
   case IDM_HIERARCHY_SELECT:  tool=SKSELECT;  break;
   case IDM_HIERARCHY_MAP:     tool=SKATMAP;   break;
   default: break;
 }
 InvalidateRect(ghwndTools,NULL,FALSE);
 SkToolOn();
 if(tool == SKANIMATE){
   DrawModel();
   if(ghwndOpenGLview == NULL)Draw3dView(0,0);
 }
 else{
   SkeletonToolBar(ghwnd_main);
   DrawSkeleton();
 }
}

void DisengageSkeletonTool(void){
 int oldtool;
 ActivateAllMenus(ghwnd_main,MF_ENABLED);
 EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_GRAYED);
 EnableToolPannels(SELECT_PANNEL,TRUE);
 DrawMenuBar(ghwnd_main);
 oldtool=(int)tool;
 SkToolOff();
 SetSkToolbarState(TRUE);
 sktool=NO;
 tool=NOTOOL;
 SetCursor(ghcurArrow);
 InvalidateRect(ghwndTools,NULL,FALSE);
 if(oldtool == SKANIMATE)DrawModel();
 else{
   SkeletonToolBar(ghwnd_main);
   DrawSkeleton();
 }
}

static void EngageDesignerTool(int newtool){
 int oldtool;
 oldtool=tool;
 if(tool > INZOOM)EDIT_ACTION=YES;
 if(tool != NOTOOL || SelectedBrush >= 0 || SelectedShader >= 0)DisengageDesignerTool();
 Empty_Undo();
 tool=newtool;
 if(newtool == SELECTOR || newtool == DESELECTOR){
   if(tool == SELECTOR){
     SendPrgmMessage(IDS_T_SELECTOR,1);
     SetCursor(ghcurSelect);
   }
   else{
     SendPrgmMessage(IDS_T_DESELECTOR,1);
     SetCursor(ghcurDeselect);
   }
   goto UPD;
 }
 if(tool > INZOOM){
   EnableToolPannels(COMMAND_PANNEL|ACTION_PANNEL,FALSE);
   ActivateAllMenus(ghwnd_main,MF_GRAYED);
   DrawMenuBar(ghwnd_main);
 }
 if(tool != NOTOOL)EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_ENABLED);
 if(tool == NODETOOL){
   SendPrgmMessage(IDS_T_NODETOOL,1);
 }
 else if(tool == PLOTTER){
   SendPrgmMessage(IDS_T_PLOTTER,1);
   Save_Undo(0);
   Plot3dOn();
 }
 else if(tool == GRABBER){
   SendPrgmMessage(IDS_T_GRABBER,1);
   Save_Undo(1);
 }
 else if(tool == PLOT3D || tool == BUILDER){
   if(tool == PLOT3D)SendPrgmMessage(IDS_T_PLOT3D,1);
   if(tool == BUILDER)SendPrgmMessage(IDS_T_BUILDER,1);
   Plot3dOn();
 }
 else if(tool == EXPANDER){
   SendPrgmMessage(IDS_T_EXPANDER,1);
   Save_Undo(1);
   IxpandOn();
 }
 else if(tool == ROTATOR){
   SendPrgmMessage(IDS_T_ROTATOR,1);
   Save_Undo(1);
   IrotateOn();
 }
 else if(tool == PRIMS){
   SendPrgmMessage(IDS_T_PRIMS,1);
   Save_Undo(0);
   if(!PrimitiveOn()){
     DisengageDesignerTool();
   }
 }
 else if(tool == SHAPER){
   SendPrgmMessage(IDS_T_SHAPER,1);
//   Save_Undo(1);  Not needed for double clicking
 }
 else if(tool == MAGNETTO){
   SendPrgmMessage(IDS_T_MAGNETTO,1);
   Save_Undo(1);
 }
 else if(newtool == PAN){
   SendPrgmMessage(IDS_T_PAN,1);
   tool=newtool;
 }
 else if(newtool == INZOOM){
   SendPrgmMessage(IDS_T_INZOOM,1);
   tool=newtool;
 }
 UPD:
 if(newtool >= NTOOLS)newtool=NODETOOL;
 else if(tool != NOTOOL)SetCursor(ghcurTools[newtool]);
 InvalidateRect(ghwndTools,NULL,FALSE);
 return;
}

void DisengageDesignerTool(void){
 int otool;
 otool=tool;
 if(SelectedBrush >= 0)PositionBrushOnOff(0,SelectedBrush);
 if(SelectedShader >= 0)PositionShaderOnOff(0,SelectedShader);
 if(sktool == YES)DisengageSkeletonTool();
 if(tool == NOTOOL)return;
 if     (tool == SELECTOR || tool == DESELECTOR){
   tool=NOTOOL;
   goto UPD;
 }
 if(tool > INZOOM){
   ActivateAllMenus(ghwnd_main,MF_ENABLED);
//   EnableToolPannels(SELECT_PANNEL|COMMAND_PANNEL|ACTION_PANNEL,TRUE);
   EnableToolPannels(COMMAND_PANNEL|ACTION_PANNEL,TRUE);
   DrawMenuBar(ghwnd_main);
 }
 if(tool != NOTOOL)EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_GRAYED);
 if(tool == GRABBER){
   tool=NOTOOL;
 }
 else if(tool == PLOTTER || tool == PLOT3D || tool == BUILDER){
   Plot3dOff();
   tool=NOTOOL;
 }
 else if(tool == EXPANDER || tool == ROTATOR || tool == SHAPER){
   tool=NOTOOL;
 }
 else if(tool == MAGNETTO){
   tool=NOTOOL;
 }
 else if(tool == PRIMS){
   PrimitiveOff();
   tool=NOTOOL;
 }
 else if(tool == PAN || tool == INZOOM){
   tool=NOTOOL;
 }
 else if(tool == NODETOOL){
   tool=NOTOOL;
 }
 UPD:
 if(otool >= NTOOLS)otool=NODETOOL;
 InvalidateRect(ghwndTools,NULL,FALSE);
 SendPrgmMessage(IDS_T_NOTOOL,1);
 SetCursor(ghcurArrow);
}

static void ExecuteDesignerAction(int item){
 if(tool > INZOOM)DisengageDesignerTool();
 if(sktool || SelectedBrush >= 0 || SelectedShader >= 0){
   SendPrgmQuery(IDQ_TOOLNOTFINISHED,0);
   return;
 }
 Empty_Undo();
 EDIT_ACTION=YES;
 switch (item){
   case IDM_TOOLS_ADDVERTEX:
     Save_Undo(0);
     if(!UpdateVertexHeap(Nvert+1))break;
     CreateVertex();
     DrawVerticesOnly((MainVp+Nvert-1));
     UpdateCounters();
     break;
   case IDM_TOOLS_BUILDFACET:
     Save_Undo(0);
     AddFace();
     UpdateCounters();
     break;
   case IDM_ACTIONS_DUPLICATE_MAKE:
     if(Nnurbs == 0 && NvertSelect == 0){
       SendPrgmQuery(IDQ_TOOFEWVERTICES,0);
       break;
     }
     Save_Undo(0);
     DuplicateSelectedNurbs();
     if(NvertSelect != 0 && AddCopy(1)){
       UpdateCounters();
       tool=GRABBER;
       InvalidateRect(ghwndTools,NULL,FALSE);
       SetCursor(ghcurTools[tool]);
     }
     break;
   case IDM_TOOLS_EXTRUDE:
     if(NvertSelect == 0){
       SendPrgmQuery(IDQ_TOOFEWVERTICES,0);
       break;
     }
     Save_Undo(0);
     if(Extrude(0)){
       UpdateCounters();
       tool=GRABBER;
       InvalidateRect(ghwndTools,NULL,FALSE);
       SetCursor(ghcurTools[tool]);
     }
     break;
   case IDM_ACTIONS_SUBDIVIDE:
     if(NvertSelect < 2 || Nedge < 1){
       SendPrgmQuery(IDQ_TOOFEWVERTICES,0);
       break;
     }
     Save_Undo(2);
     if(Subdivide(0,0,1) == OK){
       do_NOT_abort=TRUE;
       DrawModel();
       do_NOT_abort=FALSE;
       UpdateCounters();
     }
     break;
   case IDM_ACTIONS_FILL:
      if(NvertSelect == 0){
        SendPrgmQuery(IDQ_TOOFEWVERTICES,0);
        break;
     }
     if(NvertSelect == 0)break;
     Save_Undo(0);
     if(AutoFillFace(0)){
       do_NOT_abort=TRUE;
       DrawModel();
       do_NOT_abort=FALSE;
       UpdateCounters();
     }
     break;
   case IDM_ACTIONS_WELD:
     Replace_Faces(0,0.0);
     DrawModel();
     UpdateCounters();
     break;
   case IDM_ACTIONS_ROTATE_CLOCKWISE:
   case IDM_ACTIONS_ROTATE_ANTICLOCKWISE:
     Save_Undo(1);
     ActionRotate(ActiveView,90.0,1);
     break;
   case IDM_ACTIONS_EFFECTS_EXPLODE:
     Explode(1);
     break;
   case IDM_ACTIONS_EXTRUDE_ALONG:
     ExtrudeAlong();
     break;
   case IDM_ACTIONS_EFFECTS_SHATTER:
     ShatterEffect();
     break;
   case IDM_ACTIONS_JOIN:
     JoinTwoCurves();
     break;
   case IDM_ACTIONS_SPIN:
     Spin(1,0,FALSE,ActiveView);
     break;
   case IDM_ACTIONS_DUPLICATE_MULTI:
     Spin(0,0,FALSE,ActiveView);
     break;
   case IDM_ACTIONS_DUPLICATE_CONNECT:
     CopyConnected();
     break;
   case IDM_ACTIONS_FLIP:
     Flip();
     break;
   case IDM_ACTIONS_EXPAND:
     if(!IsWindow(ghwndDlgExpand)){
       ghwndDlgExpand=CreateDialog(ghinst_main,MAKEINTRESOURCE(DLG_EXPAND),
                                    ghwnd_main,(DLGPROC)ExpandDlgProc);
       if(IsWindow(ghwndDlgExpand))ShowWindow(ghwndDlgExpand,SW_SHOW);
     }
     break;
   case IDM_ACTIONS_MOVE:
     Save_Undo(1);
     DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_MOVE),ghwnd_main,
               (DLGPROC)MoveDlgProc);
     do_NOT_abort=TRUE;
     DrawModel();
     do_NOT_abort=FALSE;
     UpdateCounters();
     break;
   case IDM_ACTIONS_BEZIER:
     MakeBezierCurve();
     break;
   case IDM_ACTIONS_FACET:
     AutoFacetCurveWithHoles(0,1,0,0.0,0.0);
     break;
   case IDM_ACTIONS_EXTRUDE_BEVEL:
     AutoFacetCurveWithHoles(1,1,0,0.0,0.0);
     break;
   case IDM_ACTIONS_EXTRUDE_ROUND:
     AutoFacetCurveWithHoles(2,1,0,0.0,0.0);
     break;
   case IDM_ACTIONS_EFFECTS_PLAN:
     if(NvertSelect == 0)return;
     UnPlan();
     break;
   case IDM_ACTIONS_SMOOTH:
     SmoothCurve();
     break;
   case IDM_ACTIONS_BUILD:
     SendPrgmMessage(IDS_PLOTTERR,1);
     Save_Undo(0);
     EnableToolPannels(SELECT_PANNEL|COMMAND_PANNEL|ACTION_PANNEL|
                       TOOL_PANNEL,FALSE);
     ActivateAllMenus(ghwnd_main,MF_GRAYED);
     DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_BUILD),ghwnd_main,
            (DLGPROC)BuildAccurateCurveDlgProc);
     ActivateAllMenus(ghwnd_main,MF_ENABLED);
     EnableToolPannels(SELECT_PANNEL|COMMAND_PANNEL|ACTION_PANNEL|
                       TOOL_PANNEL,TRUE);
     break;
   case IDM_ACTIONS_BOOLEAN:
     LoadInternalAction("boolean.dll");   break;
   case IDM_ACTIONS_TWIST:
     LoadInternalAction("twist.dll");   break;
   case IDM_ACTIONS_ORIENTATE:
     LoadInternalAction("orient.dll");   break;
   case IDM_ACTIONS_REVERSE:
     LoadInternalAction("reverse.dll");   break;
   case IDM_ACTIONS_JOINS:
     LoadInternalAction("joins.dll");   break;
   default: break;
 }
}

// Code to handle the last used files on the File menu 

void InitializeLastFilesList(void){
 int i,id;
 HMENU hmenu;
 char iname[32],lname[256];
 hmenu=GetSubMenu(GetMenu(ghwnd_main),0);
 for(nLastFiles=0,i=0;i<NFILESLIST;i++){
   LastFilesList[i].present=FALSE;
   sprintf(iname,"STARTFILE%ld",i);
   GetPrivateProfileString(IniSection,iname,"$$$",
                            LastFilesList[i].file,256,IniFilename);
   if(strncmp(LastFilesList[i].file,"$$$",3) == 0)break;
   LastFilesList[i].present=TRUE;
   nLastFiles++;
   id=LASTFILECMD+1+i;
   sprintf(lname,"&%ld %s",i+1,FileInPath(LastFilesList[i].file));
   InsertMenu(hmenu,IDM_FILE_EXIT,MF_ENABLED | MF_STRING,id,lname);
 }
 if(nLastFiles > 0)InsertMenu(hmenu,IDM_FILE_EXIT,MF_SEPARATOR,LASTFILECMD,0);
}

static void UpdateLastFilesList(char *newname){
 int i,k,id;
 char iname[32],lname[256];
 HMENU hmenu;
 hmenu=GetSubMenu(GetMenu(ghwnd_main),0);
 if(nLastFiles > 0){
   for(i=0;i<nLastFiles;i++)
     if(strcmp(LastFilesList[i].file,newname) == 0)return;
   RemoveMenu(hmenu,LASTFILECMD,MF_BYCOMMAND);
   for(i=0;i<nLastFiles;i++)RemoveMenu(hmenu,LASTFILECMD+1+i,MF_BYCOMMAND);
   nLastFiles=min(nLastFiles+1,NFILESLIST);
   for(i=nLastFiles-1;i>0;i--){
     LastFilesList[i].present=LastFilesList[i-1].present;
     strcpy(LastFilesList[i].file,LastFilesList[i-1].file);
   }
 }
 else nLastFiles=1;
 LastFilesList[0].present=TRUE;
 strcpy(LastFilesList[0].file,newname);
 for(i=0;i<nLastFiles;i++){
   id=LASTFILECMD+1+i;
   sprintf(lname,"&%ld %s",i+1,FileInPath(LastFilesList[i].file));
   InsertMenu(hmenu,IDM_FILE_EXIT,MF_ENABLED | MF_STRING,id,lname);
 }
 InsertMenu(hmenu,IDM_FILE_EXIT,MF_SEPARATOR,LASTFILECMD,0);
 for(i=0;i<nLastFiles;i++){
   sprintf(iname,"STARTFILE%ld",i);
   WritePrivateProfileString(IniSection,iname,
                             LastFilesList[i].file,IniFilename);
 }
 return;
}

