/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file DIALOGS.C */

// Code for the Dialog box functions for most of the dialogs

#define MODULE_DIALOGS 1

#include "design.h"
#include "..\animate\spin_dll.h"

static int CountHierarchyList(HWND hdlg,skel *Sp);
static int UpdateHierarchyList(HWND hdlg,skel *Sp, skel *Ssp);
static skel *GetSkelFromList(HWND hdlg, skel *Sp);


static BOOL CALLBACK ThreeRealsDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 char tempstr[16];
 struct np {char *title,*name1,*name2,*name3; double cal1,cal2,cal3;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(struct np *)lparam;
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)dp->title);
       SendDlgItemMessage(hwnd,DLG_3REALS_NAME1,WM_SETTEXT,0,(LPARAM)dp->name1);
       SendDlgItemMessage(hwnd,DLG_3REALS_NAME2,WM_SETTEXT,0,(LPARAM)dp->name2);
       SendDlgItemMessage(hwnd,DLG_3REALS_NAME3,WM_SETTEXT,0,(LPARAM)dp->name3);
       sprintf(tempstr,"%.2lf",dp->cal1);
       SendDlgItemMessage(hwnd,DLG_3REALS_CAL1,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_3REALS_CAL1,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(tempstr,"%.2lf",dp->cal2);
       SendDlgItemMessage(hwnd,DLG_3REALS_CAL2,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_3REALS_CAL2,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(tempstr,"%.2lf",dp->cal3);
       SendDlgItemMessage(hwnd,DLG_3REALS_CAL3,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_3REALS_CAL3,EM_LIMITTEXT,(WPARAM)12,0);
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case DLG_3REALS_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case IDOK:
        case DLG_3REALS_OK:
          dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
          if(GetDlgItemText(hwnd,DLG_3REALS_CAL1,tempstr,10) == 0)EndDialog(hwnd, FAIL);
          dp->cal1=atof(tempstr);
          if(GetDlgItemText(hwnd,DLG_3REALS_CAL2,tempstr,10) == 0)EndDialog(hwnd, FAIL);
          dp->cal2=atof(tempstr);
          if(GetDlgItemText(hwnd,DLG_3REALS_CAL3,tempstr,10) == 0)EndDialog(hwnd, FAIL);
          dp->cal3=atof(tempstr);
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

int Read3Reals(char *title, double *r1, double *r2, double *r3, int list){
 int rv;
 char *lists[2][3]={ {"Left/Right" ,"Back/Front", "Up/Down"},
                     {"Direction", "Elevation", "Bank"}
                  };
 struct np {char *title,*name1,*name2,*name3; double cal1,cal2,cal3;} d,*dp;
 d.title=title;
 d.name1=lists[list][0];
 d.name2=lists[list][1];
 d.name3=lists[list][2];
 d.cal1 = *r1;
 d.cal2 = *r2;
 d.cal3 = *r3;
 EnableToolPannels(ALL_PANNELS,FALSE);
 if(DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_3REALS),ghwnd_main,
            (DLGPROC)ThreeRealsDlgProc,(LPARAM)&d) == OK){
   *r1 = d.cal1;
   *r2 = d.cal2;
   *r3 = d.cal3;
   rv=OK;
 }
 else rv=FAIL;
 EnableToolPannels(ALL_PANNELS,TRUE);
 return rv;
}

static BOOL CALLBACK ThreeIntsDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err;
 int i;
 struct np {char *title,*name1,*name2,*name3;
            int i1,i2,i3,i1h,i2h,i3h,i1l,i2l,i3l;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(struct np *)lparam;
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)dp->title);
       SendDlgItemMessage(hwnd,DLG_3INTS_TEXT1,WM_SETTEXT,0,(LPARAM)dp->name1);
       SendDlgItemMessage(hwnd,DLG_3INTS_TEXT2,WM_SETTEXT,0,(LPARAM)dp->name2);
       SendDlgItemMessage(hwnd,DLG_3INTS_TEXT3,WM_SETTEXT,0,(LPARAM)dp->name3);
       SetDlgItemInt(hwnd,DLG_3INTS_EDIT1,dp->i1,FALSE);
       SendDlgItemMessage(hwnd,DLG_3INTS_EDIT1,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN1,SPNM_SETRANGE,0,
                          MAKELPARAM(dp->i1l,dp->i1h));
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN1,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->i1,0);
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN1,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_3INTS_EDIT1));
       SetDlgItemInt(hwnd,DLG_3INTS_EDIT2,dp->i2,FALSE);
       SendDlgItemMessage(hwnd,DLG_3INTS_EDIT2,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN2,SPNM_SETRANGE,0,
                          MAKELPARAM(dp->i2l,dp->i2h));
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN2,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->i2,0);
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN2,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_3INTS_EDIT2));
       SetDlgItemInt(hwnd,DLG_3INTS_EDIT3,dp->i3,FALSE);
       SendDlgItemMessage(hwnd,DLG_3INTS_EDIT3,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN3,SPNM_SETRANGE,0,
                          MAKELPARAM(dp->i3l,dp->i3h));
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN3,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->i3,0);
       SendDlgItemMessage(hwnd,DLG_3INTS_SPIN3,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_3INTS_EDIT3));
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case DLG_3INTS_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case IDOK:
        case DLG_3INTS_OK:
          dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
          i=GetDlgItemInt(hwnd,DLG_3INTS_EDIT1,&err,FALSE);
          if(!err)EndDialog(hwnd,FAIL);
          if(i > dp->i1h || i < dp->i1l)EndDialog(hwnd,FAIL);
          dp->i1=i;
          i=GetDlgItemInt(hwnd,DLG_3INTS_EDIT2,&err,FALSE);
          if(!err)EndDialog(hwnd,FAIL);
          if(i > dp->i2h || i < dp->i2l)EndDialog(hwnd,FAIL);
          dp->i2=i;
          i=GetDlgItemInt(hwnd,DLG_3INTS_EDIT3,&err,FALSE);
          if(!err)EndDialog(hwnd,FAIL);
          if(i > dp->i3h || i < dp->i3l)EndDialog(hwnd,FAIL);
          dp->i3=i;
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

int Read3Ints(char *title, char *it1, char *it2, char *it3,
                    int *i1, int *i2, int *i3, HWND parent){
 struct np {char *title,*name1,*name2,*name3;
            int i1,i2,i3,i1h,i2h,i3h,i1l,i2l,i3l;} d,*dp;
 int rv;
 d.title=title;
 d.name1=it1;
 d.name2=it2;
 d.name3=it3;
 d.i1 = *i1;  d.i1l=0; d.i1h=9000;
 d.i2 = *i2;  d.i2l=0; d.i2h=9000;
 d.i3 = *i3;  d.i3l=0; d.i3h=9000;
 if(DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_3INTS),parent,
            (DLGPROC)ThreeIntsDlgProc,(LPARAM)&d) == OK){
   *i1 = d.i1;
   *i2 = d.i2;
   *i3 = d.i3;
   rv=OK;
 }
 else rv=FAIL;
 return rv;
}

int Read2Ints(char *title, char *it1, char *it2,
                    int *i1, int *i2, HWND parent){
 return FAIL;
}

BOOL CALLBACK SpiralsDlgProc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 int i;
 struct np {int type,n,r;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(struct np *)lparam;
       SetDlgItemInt(hwnd,DLG_SPIRAL_EDITN,dp->n,FALSE);
       SendDlgItemMessage(hwnd,DLG_SPIRAL_EDITN,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_SPIRAL_SPINN,SPNM_SETRANGE,0,
                          MAKELPARAM(5,90));
       SendDlgItemMessage(hwnd,DLG_SPIRAL_SPINN,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->n,0);
       SendDlgItemMessage(hwnd,DLG_SPIRAL_SPINN,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SPIRAL_EDITN));
       SetDlgItemInt(hwnd,DLG_SPIRAL_EDITR,dp->r,FALSE);
       SendDlgItemMessage(hwnd,DLG_SPIRAL_EDITR,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_SPIRAL_SPINR,SPNM_SETRANGE,0,
                          MAKELPARAM(10,9999));
       SendDlgItemMessage(hwnd,DLG_SPIRAL_SPINR,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->r,0);
       SendDlgItemMessage(hwnd,DLG_SPIRAL_SPINR,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_SPIRAL_EDITR));
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       if(dp->type == 0)active=TRUE; else active=FALSE;
       SendDlgItemMessage(hwnd,DLG_SPIRAL_ARC,BM_SETCHECK,active,0);
       SendDlgItemMessage(hwnd,DLG_SPIRAL_LOG,BM_SETCHECK,!active,0);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_SPIRAL_ARC:
          dp->type=0;
          break;
        case DLG_SPIRAL_LOG:
          dp->type=1;
          break;
        case IDCANCEL:
        case DLG_SPIRAL_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case IDOK:
        case DLG_SPIRAL_OK:
          i=GetDlgItemInt(hwnd,DLG_SPIRAL_EDITN,&err,FALSE);
          if(!err)EndDialog(hwnd, -1);
          if(i > 90)i=90; if(i < 5)i=5;
          dp->n=i;
          i=GetDlgItemInt(hwnd,DLG_SPIRAL_EDITR,&err,FALSE);
          if(!err)EndDialog(hwnd, -1);
          if(i > 9999)i=9999; if(i < 10)i=10;
          dp->r=i;
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK OneIntOneRealDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 int i;
 BOOL err;
 char tempstr[16];
 struct np {char *title,*name1,*name2;
            int i,il,ih; double r;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(struct np *)lparam;
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)dp->title);
       SendDlgItemMessage(hwnd,DLG_1I1R_TEXTI,WM_SETTEXT,0,(LPARAM)dp->name1);
       SendDlgItemMessage(hwnd,DLG_1I1R_TEXTR,WM_SETTEXT,0,(LPARAM)dp->name2);
       SetDlgItemInt(hwnd,DLG_1I1R_EDITI,dp->i,FALSE);
       SendDlgItemMessage(hwnd,DLG_1I1R_EDITI,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_1I1R_SPINI,SPNM_SETRANGE,0,
                          MAKELPARAM(dp->il,dp->ih));
       SendDlgItemMessage(hwnd,DLG_1I1R_SPINI,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->i,0);
       SendDlgItemMessage(hwnd,DLG_1I1R_SPINI,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_1I1R_EDITI));
       sprintf(tempstr,"%.4lf",dp->r);
       SendDlgItemMessage(hwnd,DLG_1I1R_EDITR,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_1I1R_EDITR,EM_LIMITTEXT,(WPARAM)12,0);
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case DLG_1I1R_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case IDOK:
        case DLG_1I1R_OK:
          dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
          i=GetDlgItemInt(hwnd,DLG_1I1R_EDITI,&err,FALSE);
          if(!err)EndDialog(hwnd, -1);
          if(i > dp->ih)i=dp->ih; if(i < dp->il)i=dp->il;
          dp->i=i;
          if(GetDlgItemText(hwnd,DLG_1I1R_EDITR,tempstr,10) == 0)EndDialog(hwnd, FAIL);
          dp->r=atof(tempstr);
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

int Read1Int1Real(char *title, char *it1, char *it2,
                    int *i, int il, int ih,
                    double *r, double rl, double rh, HWND parent){
 struct np {char *title,*name1,*name2;
            int i,il,ih; double r;} d;
 int rv;
 d.title=title;
 d.name1=it1;
 d.name2=it2;
 d.i = *i;  d.il=il; d.ih=ih;
 d.r = *r;
// EnableToolPannels(ALL_PANNELS,FALSE);
 if(DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_1I1R),parent,
            (DLGPROC)OneIntOneRealDlgProc,(LPARAM)&d) == OK){
   *i = d.i;
   *r = d.r;
   rv=OK;
 }
 else rv=FAIL;
// EnableToolPannels(ALL_PANNELS,TRUE);
 return rv;
}

BOOL CALLBACK FunctionPlotDlgProc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam){
 BOOL err;
 int i;
 struct np {int t,x,y;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
        char *function_list[]={
          "Plain Grid",
           "Sine Wave",
           "Tangent Function",
           "Bi-variate Normal Dist.",
           "Plot of Sine(x) * Sine(y)",
           "Plot of Sine(x * y)",
           "Plot of Sine(x) * Sine(3*y)",
           "Plot of Sine(x) + Cosine(y)"
         };
       dp=(struct np *)lparam;
       for(i=0;i<8;i++)SendDlgItemMessage(hwnd,DLG_FPLOT_LIST,LB_ADDSTRING,0,
                                          (LPARAM)function_list[i]);
       SendDlgItemMessage(hwnd,DLG_FPLOT_LIST,LB_SETCURSEL,0,(LPARAM)dp->t);
       SetDlgItemInt(hwnd,DLG_FPLOT_EDITX,dp->x,FALSE);
       SendDlgItemMessage(hwnd,DLG_FPLOT_EDITX,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_FPLOT_SPINX,SPNM_SETRANGE,0,
                          MAKELPARAM(2,90));
       SendDlgItemMessage(hwnd,DLG_FPLOT_SPINX,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->x,0);
       SendDlgItemMessage(hwnd,DLG_FPLOT_SPINX,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_FPLOT_EDITX));
       SetDlgItemInt(hwnd,DLG_FPLOT_EDITY,dp->y,FALSE);
       SendDlgItemMessage(hwnd,DLG_FPLOT_EDITY,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_FPLOT_SPINY,SPNM_SETRANGE,0,
                          MAKELPARAM(2,90));
       SendDlgItemMessage(hwnd,DLG_FPLOT_SPINY,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->y,0);
       SendDlgItemMessage(hwnd,DLG_FPLOT_SPINY,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_FPLOT_EDITY));
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_FPLOT_LIST:
          switch(HIWORD(wparam)){
            case LBN_DBLCLK:
              SendMessage(hwnd,WM_COMMAND,MAKEWPARAM(DLG_FPLOT_OK,0),0);
              break;
            default:
              break;
          }
          break;
        case IDCANCEL:
        case DLG_FPLOT_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case IDOK:
        case DLG_FPLOT_OK:
          i=GetDlgItemInt(hwnd,DLG_FPLOT_EDITX,&err,FALSE);
          if(!err)EndDialog(hwnd, -1);
          if(i > 190)i=190;
          if(i < 2)i=2;
          dp->x=i;
          i=GetDlgItemInt(hwnd,DLG_FPLOT_EDITY,&err,FALSE);
          if(!err)EndDialog(hwnd, -1);
          if(i > 190)i=190;
          if(i < 2)i=2;
          dp->y=i;
          dp->t=SendMessage(GetDlgItem(hwnd,DLG_FPLOT_LIST),LB_GETCURSEL,0,0);
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK ExtrudeAlongDlgProc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam){
 char tempstr[16];
 BOOL err;
 int i;
 struct np {double expan,repeat_map;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       dp=(struct np *)lparam;
       sprintf(tempstr,"%.4lf",dp->expan);
       SendDlgItemMessage(hwnd,DLG_EXTRUDEALONG_EXP,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_EXTRUDEALONG_EXP,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(tempstr,"%.4lf",dp->repeat_map);
       SendDlgItemMessage(hwnd,DLG_EXTRUDEALONG_MAP,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_EXTRUDEALONG_MAP,EM_LIMITTEXT,(WPARAM)12,0);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case DLG_EXTRUDEALONG_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_EXTRUDEALONG_OK:
          if(GetDlgItemText(hwnd,DLG_EXTRUDEALONG_EXP,tempstr,10) == 0)EndDialog(hwnd, FAIL);
          if((dp->expan=atof(tempstr)) == 0)EndDialog(hwnd, FAIL);
          if(GetDlgItemText(hwnd,DLG_EXTRUDEALONG_MAP,tempstr,10) == 0)EndDialog(hwnd, FAIL);
          if((dp->repeat_map=atof(tempstr)) == 0)EndDialog(hwnd, FAIL);
          if(SendDlgItemMessage(hwnd,DLG_EXTRUDEALONG_EXPAND,BM_GETCHECK,0,0))
                EndDialog(hwnd,2);
          else  EndDialog(hwnd,1);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK FlipDlgProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam){
 BOOL active;
 int i;
 struct np {int plane,weld;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       dp=(struct np *)lparam;
       if(dp->plane)active=TRUE; else active=FALSE;
       SendDlgItemMessage(hwnd,DLG_FLIP_PLANEV,BM_SETCHECK,active,0);
       SendDlgItemMessage(hwnd,DLG_FLIP_PLANEH,BM_SETCHECK,!active,0);
       if(dp->weld)active=TRUE; else active=FALSE;
       SendDlgItemMessage(hwnd,DLG_FLIP_WELDON,BM_SETCHECK,active,0);
       SendDlgItemMessage(hwnd,DLG_FLIP_WELDOFF,BM_SETCHECK,!active,0);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_FLIP_PLANEH:  dp->plane=0; break;
        case DLG_FLIP_PLANEV:  dp->plane=1; break;
        case DLG_FLIP_WELDOFF: dp->weld=0; break;
        case DLG_FLIP_WELDON:  dp->weld=1; break;
        case IDCANCEL:
        case DLG_FLIP_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_FLIP_OK:
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK ExpandDlgProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam){
#define RANGE 100
 char tempstr[16];
 HWND hctl;
 int iMin,iMax;
 RECT rc;
 POINT p;
 static int i,xpos=-1,ypos=-1,xpand_a=0;
 static double xpand_x=1.0,xpand_y=1.0,xpand_z=1.0;
 static BOOL active,xpand_X=TRUE,xpand_Y=TRUE,xpand_Z=TRUE,
             xpand_vf=TRUE,xpand_cu=TRUE;
 double a;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     EnableWindow(ghwnd_main,TRUE);
     if(xpand_vf)active=TRUE; else active=FALSE;
     SendDlgItemMessage(hwnd,DLG_EXPAND_VERT,BM_SETCHECK,active,0);
     SendDlgItemMessage(hwnd,DLG_EXPAND_FACE,BM_SETCHECK,!active,0);
     if(xpand_cu)active=TRUE; else active=FALSE;
     SendDlgItemMessage(hwnd,DLG_EXPAND_CURSOR,BM_SETCHECK,active,0);
     SendDlgItemMessage(hwnd,DLG_EXPAND_CENTRE,BM_SETCHECK,!active,0);
     if(xpand_X)active=TRUE; else active=FALSE;
     SendDlgItemMessage(hwnd,DLG_EXPAND_LR,BM_SETCHECK,active,0);
     if(xpand_Y)active=TRUE; else active=FALSE;
     SendDlgItemMessage(hwnd,DLG_EXPAND_FB,BM_SETCHECK,active,0);
     if(xpand_Z)active=TRUE; else active=FALSE;
     SendDlgItemMessage(hwnd,DLG_EXPAND_UD,BM_SETCHECK,active,0);
     sprintf(tempstr,"%.2lf",xpand_x);
     SendDlgItemMessage(hwnd,DLG_EXPAND_EDITLR,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hwnd,DLG_EXPAND_EDITLR,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(tempstr,"%.2lf",xpand_y);
     SendDlgItemMessage(hwnd,DLG_EXPAND_EDITFB,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hwnd,DLG_EXPAND_EDITFB,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(tempstr,"%.2lf",xpand_z);
     SendDlgItemMessage(hwnd,DLG_EXPAND_EDITUD,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hwnd,DLG_EXPAND_EDITUD,EM_LIMITTEXT,(WPARAM)12,0);
//     SetScrollRange(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),SB_CTL,-RANGE,RANGE,FALSE);
//     SetScrollPos(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),SB_CTL,0,TRUE);
     SendMessage(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),
       TBM_SETRANGE,(WPARAM)TRUE,(LPARAM)MAKELONG(-RANGE,RANGE));
     for(i = -RANGE+20; i < RANGE; i += 20)
       SendMessage(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),
         TBM_SETTIC,(WPARAM)0,(LPARAM)i);
     SendMessage(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),
       TBM_SETPOS,TRUE,(LPARAM)0);
     if(xpos == -1 && ypos == -1){
       CentreDialogOnCursor(hwnd);
       GetWindowRect(hwnd,&rc);
       p.x=rc.right; p.y=rc.top;
       ScreenToClient(ghwnd_main,&p);
       xpos=p.x; ypos=p.y;
     }
     else SetWindowPos(hwnd,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     return (TRUE);
   case WM_DESTROY:
     GetWindowRect(hwnd,&rc);
     xpos=rc.left; ypos=rc.top;
     break;
   case WM_HSCROLL:
     if((HWND)lparam == (hctl=GetDlgItem(hwnd,DLG_EXPAND_SLIDER))){
       iMin=SendMessage(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),
         TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),
         TBM_GETRANGEMAX,0,0);
//       GetScrollRange(hctl,SB_CTL,&iMin,&iMax);
       switch (LOWORD(wparam)) {
         case SB_PAGEDOWN: xpand_a = min(iMax,xpand_a+5); break;
         case SB_LINEDOWN: xpand_a = min(iMax,xpand_a+1); break;
         case SB_PAGEUP:   xpand_a = max(iMin,xpand_a-5); break;
         case SB_LINEUP:   xpand_a = max(iMin,xpand_a-1); break;
//         case SB_THUMBTRACK:
//         case SB_THUMBPOSITION:
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           xpand_a=(int)((short)HIWORD(wparam));  break;
         default: break;
       }
//       SetScrollPos(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),SB_CTL,xpand_a,TRUE);
       SendMessage(GetDlgItem(hwnd,DLG_EXPAND_SLIDER),
         TBM_SETPOS,TRUE,(LPARAM)xpand_a);
       a=3.0*(double)xpand_a/(double)RANGE;
       if(xpand_a < 0){
         a=1.0/(1.0-a);
       }
       else if(xpand_a > 0){
         a=1.0+a;
       }
       else a=1.0;
       sprintf(tempstr,"%.2lf",a); xpand_y=xpand_z=xpand_x=a;
       SendDlgItemMessage(hwnd,DLG_EXPAND_EDITLR,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_EXPAND_EDITFB,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_EXPAND_EDITUD,WM_SETTEXT,0,(LPARAM)tempstr);
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case DLG_EXPAND_LR: xpand_X = !xpand_X; break;
       case DLG_EXPAND_FB: xpand_Y = !xpand_Y; break;
       case DLG_EXPAND_UD: xpand_Z = !xpand_Z; break;
       case DLG_EXPAND_VERT: xpand_vf=TRUE;    break;
       case DLG_EXPAND_FACE: xpand_vf=FALSE;   break;
       case DLG_EXPAND_CURSOR: xpand_cu=TRUE;  break;
       case DLG_EXPAND_CENTRE: xpand_cu=FALSE; break;
       case DLG_EXPAND_14: strcpy(tempstr,"0.250"); goto DISPLAY;
       case DLG_EXPAND_13: strcpy(tempstr,"0.333"); goto DISPLAY;
       case DLG_EXPAND_12: strcpy(tempstr,"0.500"); goto DISPLAY;
       case DLG_EXPAND_2:  strcpy(tempstr,"2.000"); goto DISPLAY;
       case DLG_EXPAND_3:  strcpy(tempstr,"3.000"); goto DISPLAY;
       case DLG_EXPAND_4:  strcpy(tempstr,"4.000"); goto DISPLAY;
         DISPLAY:
         sscanf(tempstr,"%lf",&xpand_x); xpand_y=xpand_z=xpand_x;
         SendDlgItemMessage(hwnd,DLG_EXPAND_EDITLR,WM_SETTEXT,0,(LPARAM)tempstr);
         SendDlgItemMessage(hwnd,DLG_EXPAND_EDITFB,WM_SETTEXT,0,(LPARAM)tempstr);
         SendDlgItemMessage(hwnd,DLG_EXPAND_EDITUD,WM_SETTEXT,0,(LPARAM)tempstr);
         break;
       case DLG_EXPAND_EXPAND:
       case IDCANCEL:
       case DLG_EXPAND_DONE:
         if(GetDlgItemText(hwnd,DLG_EXPAND_EDITLR,tempstr,10) != 0 &&
            (a=atof(tempstr)) != 0)xpand_x=a;
         if(GetDlgItemText(hwnd,DLG_EXPAND_EDITFB,tempstr,10) != 0 &&
            (a=atof(tempstr)) != 0)xpand_y=a;
         if(GetDlgItemText(hwnd,DLG_EXPAND_EDITUD,tempstr,10) != 0 &&
            (a=atof(tempstr)) != 0)xpand_z=a;
         if(LOWORD(wparam) == DLG_EXPAND_EXPAND){
           long  cx,cy,cz;
           double sx,sy,sz,medianX,medianY,medianZ;
           int i;
           vertex *vp,*V0,*V1,*V2;
           face *fp;
           Save_Undo(1);
           if(xpand_X)sx=xpand_x; else sx=1.0;
           if(xpand_Y)sy=xpand_y; else sy=1.0;
           if(xpand_Z)sz=xpand_z; else sz=1.0;
           cx=NpointerX; cy=NpointerY; cz=NpointerZ;
           if(!xpand_cu){
             point p,temp;
             get_centre(1,p,temp,FALSE);
             cx=p[0]; cy=p[1]; cz=p[2];
           }
           if(xpand_vf){/* expand using verts */
             vp=MainVp; if(Nvert > 0)for(i=0;i<Nvert;i++){
               if(vp->status==SELECTED){
                 vp->xyz[0]=(long)((double)(vp->xyz[0]-cx)*sx)+cx;
                 vp->xyz[1]=(long)((double)(vp->xyz[1]-cy)*sy)+cy;
                 vp->xyz[2]=(long)((double)(vp->xyz[2]-cz)*sz)+cz;
               }
               vp++;
             }
           }
           else{ /* expand using faces */
             if((fp=MainFp) != NULL)for(i=0;i<Nface;i++){
               V0=(MainVp+fp->V[0]);
               V1=(MainVp+fp->V[1]);
               V2=(MainVp+fp->V[2]);
               if(V0->status == SELECTED ||
                  V1->status == SELECTED ||
                  V2->status == SELECTED){
                 medianX=(V0->xyz[0] + V1->xyz[0] + V2->xyz[0])/3.0;
                 medianY=(V0->xyz[1] + V1->xyz[1] + V2->xyz[1])/3.0;
                 medianZ=(V0->xyz[2] + V1->xyz[2] + V2->xyz[2])/3.0;
                 V0->xyz[0]=(long)(((double)V0->xyz[0]-medianX)*sx)+medianX;
                 V0->xyz[1]=(long)(((double)V0->xyz[1]-medianY)*sy)+medianY;
                 V0->xyz[2]=(long)(((double)V0->xyz[2]-medianZ)*sz)+medianZ;
                 V1->xyz[0]=(long)(((double)V1->xyz[0]-medianX)*sx)+medianX;
                 V1->xyz[1]=(long)(((double)V1->xyz[1]-medianY)*sy)+medianY;
                 V1->xyz[2]=(long)(((double)V1->xyz[2]-medianZ)*sz)+medianZ;
                 V2->xyz[0]=(long)(((double)V2->xyz[0]-medianX)*sx)+medianX;
                 V2->xyz[1]=(long)(((double)V2->xyz[1]-medianY)*sy)+medianY;
                 V2->xyz[2]=(long)(((double)V2->xyz[2]-medianZ)*sz)+medianZ;
               }
               fp++;
             }
           }
           ExpandNurbsAction(cx,cy,cz,sx,sy,sz);
           DrawModel();
           if(ghwndOpenGLview == NULL)Draw3dView(0,0);
           else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0); // update
           break;
         }
         DestroyWindow(ghwndDlgExpand);
         ghwndDlgExpand=NULL;
//         EndDialog(hwnd,OK);
//         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
#undef RANGE
}

static skel *GetSkelFromList(HWND hdlg, skel *Sp){
 int i,n=0;
 skel *sp;
 HWND hctl;
 hctl=GetDlgItem(hdlg,DLG_SKEL_DIR);
 if((i=SendMessage(hctl,LB_GETCURSEL,0,0)) == LB_ERR)return NULL;
 if((sp=FirstSp) != NULL)while(sp != NULL){
   if(sp->at == Sp){
     if(i == n)return sp;
     n++;
   }
   sp=sp->next;
 }
 return NULL;
}

static int UpdateHierarchyList(HWND hdlg,skel *Sp, skel *Ssp){
 BOOL active;
 int n=0,id=0;
 HWND hctl;
 skel *sp,*spp=NULL,*sp1=NULL;
 hctl=GetDlgItem(hdlg,DLG_SKEL_DIR);
 SendMessage(hctl,LB_RESETCONTENT,0,0);
 if((sp=FirstSp) != NULL)while(sp != NULL){
   if(sp->at == Sp){
     SendMessage(hctl,LB_ADDSTRING,0,(LPARAM)(sp->name));
     if(sp == Ssp){id=n; spp=sp;}
     sp1=sp;
     n++;
   }
   sp=sp->next;
 }
 if(n == 0)SendMessage(hctl,LB_ADDSTRING,0,(LPARAM)"<end>");
 else      SendMessage(hctl,LB_SETCURSEL,(WPARAM)id,0);
 if(Sp == NULL)active=FALSE; else active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_ADD),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_DELETE),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_PARENT),active);
 if(n == 0)active=FALSE; else active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_RENAME),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_CHILD),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_SELECT),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_DESELECT),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_HIDE),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_ATTACH),active);
 EnableWindow(GetDlgItem(hdlg,DLG_SKEL_DETACH),active);
 {
   char c1[64],c2[64];
   if(spp == NULL){
     if(Ssp == NULL){
       if(sp1 == NULL){
         if(Sp == NULL)sp=FirstSp;
         else sp=Sp;
       }
       else sp=sp1;
     }
     else sp=Ssp;
   }
   else sp=spp;
   if(sp == NULL)sp=FirstSp; /* last resort */
   strcpy(c1,"\\"); strcat(c1,sp->name);
   sp=sp->at;
   while(sp != NULL){
     strcpy(c2,"\\"); strcat(c2,sp->name); strcat(c2,c1);
     strcpy(c1,c2);
     if(strlen(c1) > 60)break;
     sp=sp->at;
   }
   strcat(c1,"\\");
   SendDlgItemMessage(hdlg,DLG_SKEL_FAMILY,WM_SETTEXT,0,(LPARAM)c1);
 }
 return n;
}

BOOL CALLBACK HierarchyDlgProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam){
 static int Nlist;
 char local_name[16];
 int i;
 skel *sp;
 vertex *vp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     current_skeleton=NULL;
     Nlist=UpdateHierarchyList(hwnd,current_skeleton,FirstSp);
     EnableWindow(ghwnd_main,TRUE);
     EnableWindow(GetDlgItem(hwnd,DLG_SKEL_FAMILY),FALSE);
     CentreDialogOnCursor(hwnd);
     return (TRUE);
   case WM_COMMAND:
     switch(HIWORD(wparam)){
       case LBN_DBLCLK:
         SendMessage(hwnd,WM_COMMAND,MAKEWPARAM(DLG_SKEL_CHILD,0),0);
         break;
       default:
         break;
     }
     switch(LOWORD(wparam)){
       case DLG_SKEL_ROOT:
         current_skeleton=NULL;
         Nlist=UpdateHierarchyList(hwnd,current_skeleton,FirstSp);
         break;
       case DLG_SKEL_PARENT:
         if(current_skeleton != NULL){
           Nlist=UpdateHierarchyList(hwnd,current_skeleton->at,current_skeleton);
           current_skeleton=current_skeleton->at;
         }
         break;
       case DLG_SKEL_CHILD:
         if(Nlist > 0){
           sp=GetSkelFromList(hwnd,current_skeleton);
           if(sp != NULL){
             Nlist=UpdateHierarchyList(hwnd,sp,NULL);
             current_skeleton=sp;
           }
         }
         break;
       case DLG_SKEL_DELETE:
         sp=GetSkelFromList(hwnd,current_skeleton);
         if(sp != NULL && sp->at != NULL){
           SkDelete(sp);
           Nlist=UpdateHierarchyList(hwnd,current_skeleton,NULL);
         }
         break;
       case DLG_SKEL_RENAME:
         sp=GetSkelFromList(hwnd,current_skeleton);
         strcpy(local_name,sp->name);
         RequestCharString(12,local_name,"New Name",hwnd);
         if(strlen(local_name) > 0){
           strcpy(sp->name,local_name);
           Nlist=UpdateHierarchyList(hwnd,current_skeleton,sp);
         }
         break;
       case DLG_SKEL_ADD:
         if(current_skeleton == NULL)break;
         CreateSkeleton(current_skeleton);
         MainSp->xyz[2] = -(long)MINUNIT*4L*Nskel;
         ZeroSkeletonBoundingBox(MainSp);
         strcpy(local_name,MainSp->name);
         RequestCharString(12,local_name,"New Name",hwnd);
         if(strlen(local_name) > 0)strcpy(MainSp->name,local_name);
         Nlist=UpdateHierarchyList(hwnd,current_skeleton,MainSp);
         break;
       case DLG_SKEL_ATTACH:
         if((sp=GetSkelFromList(hwnd,current_skeleton)) == NULL)break;
         GetSkeletonBoundBox(sp,1); /* assign as well */
         break;
       case DLG_SKEL_DETACH:
         if((sp=GetSkelFromList(hwnd,current_skeleton)) == NULL)break;
         if((vp=MainVp) != NULL)for(i=0;i<Nvert;i++,vp++){
           if(vp->sp == sp)vp->sp=NULL;
         }
         break;
       case DLG_SKEL_SELECT:
       case DLG_SKEL_DESELECT:
       case DLG_SKEL_HIDE:
         if((sp=GetSkelFromList(hwnd,current_skeleton)) == NULL)break;
         SelectByHierarchy(sp,(int)LOWORD(wparam));
         break;
       case IDCANCEL:
       case DLG_SKEL_DONE:
         current_skeleton=NULL;
         EndDialog(hwnd,OK);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

BOOL CALLBACK SqueezeDlgProc(HWND hdlg,UINT msg,WPARAM wparam,LPARAM lparam){
 double *d,d1;
 char tempstr[16];
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     d=(double *)lparam;
     SetWindowLong(hdlg,GWL_USERDATA,lparam);
     EnableToolPannels(ALL_PANNELS,FALSE);
     sprintf(tempstr,"%.2lf",*d);
     SendDlgItemMessage(hdlg,DLG_SQUEEZE_WELDVALUE,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hdlg,DLG_SQUEEZE_WELDVALUE,EM_LIMITTEXT,(WPARAM)12,0);
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      d=(double *)GetWindowLong(hdlg,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hdlg,0);
          return(TRUE);
        case DLG_SQUEEZE_CUR:
          EndDialog(hdlg,1);
          return(TRUE);
        case DLG_SQUEEZE_CENT:
          EndDialog(hdlg,2);
          return(TRUE);
        case DLG_SQUEEZE_WELD:
          if(GetDlgItemText(hdlg,DLG_SQUEEZE_WELDVALUE,tempstr,10) == 0){
            EndDialog(hdlg,0);
            return(TRUE);
          }
          if((d1=atof(tempstr)) == 0){
            EndDialog(hdlg,0);
            return(TRUE);
          }
          *d=d1;
          EndDialog(hdlg,3);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK ExplodeDlgProc(HWND hdlg,UINT msg,WPARAM wparam,LPARAM lparam){
 struct TEMP {int id,id1; double dxx,dyy,dzz;} *tc;
 char tempstr[16];
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     tc=(struct TEMP *)lparam;
     SetWindowLong(hdlg,GWL_USERDATA,lparam);
     EnableToolPannels(ALL_PANNELS,FALSE);
     if(tc->id == 5)SendDlgItemMessage(hdlg,DLG_EXPLODE_FACES,BM_SETCHECK,TRUE,0);
     if(tc->id1 == 1)SendDlgItemMessage(hdlg,DLG_EXPLODE_FO_RANDOM,BM_SETCHECK,TRUE,0);
     sprintf(tempstr,"%.2lf",tc->dxx);
     SendDlgItemMessage(hdlg,DLG_EXPLODE_FO_PROPX,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hdlg,DLG_EXPLODE_FO_PROPX,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(tempstr,"%.2lf",tc->dyy);
     SendDlgItemMessage(hdlg,DLG_EXPLODE_FO_PROPY,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hdlg,DLG_EXPLODE_FO_PROPY,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(tempstr,"%.2lf",tc->dzz);
     SendDlgItemMessage(hdlg,DLG_EXPLODE_FO_PROPZ,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hdlg,DLG_EXPLODE_FO_PROPZ,EM_LIMITTEXT,(WPARAM)12,0);
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      tc=(struct TEMP *)GetWindowLong(hdlg,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_EXPLODE_SPHERE:     tc->id=1;  break;
        case DLG_EXPLODE_H:          tc->id=2;  break;
        case DLG_EXPLODE_VFB:        tc->id=3;  break;
        case DLG_EXPLODE_VLR:        tc->id=4;  break;
        case DLG_EXPLODE_FACES:      tc->id=5;  break;
        case DLG_EXPLODE_GRID:       tc->id=6;  break;
        case DLG_EXPLODE_FO_UNIFORM: tc->id1=0; break;
        case DLG_EXPLODE_FO_RANDOM:  tc->id1=1; break;
        case DLG_EXPLODE_FO_PROP:    tc->id1=2; break;
        case IDCANCEL:
          tc->id=0;
          EndDialog(hdlg,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hdlg,DLG_EXPLODE_FO_PROPX,tempstr,10) == 0)
             tc->id=0;
          if((tc->dxx=atof(tempstr)) == 0)tc->id=0;
          if(GetDlgItemText(hdlg,DLG_EXPLODE_FO_PROPY,tempstr,10) == 0)
             tc->id=0;
          if((tc->dyy=atof(tempstr)) == 0)tc->id=0;
          if(GetDlgItemText(hdlg,DLG_EXPLODE_FO_PROPZ,tempstr,10) == 0)
             tc->id=0;
          if((tc->dzz=atof(tempstr)) == 0)tc->id=0;
          EndDialog(hdlg,TRUE);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

// Preferences code

static BOOL CALLBACK GeneralPreferencesDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL state;
 char *w, yesyes[]="YES", nono[]="NO";
 static unsigned char uColour[3];
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     CentreDialogOnScreen(GetParent(hdlg));
     if(Preferences.buttons)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_BUTTONS,BM_SETCHECK,state,0);
     if(Preferences.quick_trick)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_QUICKTRICK,BM_SETCHECK,state,0);
     if(Preferences.undo)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_UNDO,BM_SETCHECK,state,0);
     if(Preferences.tooltips)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLTIPS,BM_SETCHECK,state,0);
     if(Preferences.bigv)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_BIGV,BM_SETCHECK,state,0);
     if(Preferences.use_positions)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_POSITIONS,BM_SETCHECK,state,0);
     if(Preferences.toolsmall)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLSMALL,BM_SETCHECK,state,0);
     if(Preferences.toolleft)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLLEFT,BM_SETCHECK,state,0);
     uColour[0]=Preferences.Vcolour[0];
     uColour[1]=Preferences.Vcolour[1];
     uColour[2]=Preferences.Vcolour[2];
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PREFERENCES_VCOLOUR){
         r=uColour[0]; g=uColour[1]; b=uColour[2];
         if(lpdis->itemState & ODS_SELECTED)
           InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_NOTIFY:
     switch(((LPNMHDR)lparam)->code){
       case PSN_APPLY:
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_BUTTONS,BM_GETCHECK,0,0))
            Preferences.buttons=TRUE; else Preferences.buttons=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_QUICKTRICK,BM_GETCHECK,0,0))
            Preferences.quick_trick=TRUE; else Preferences.quick_trick=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_UNDO,BM_GETCHECK,0,0))
            Preferences.undo=TRUE; else Preferences.undo=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLTIPS,BM_GETCHECK,0,0))
            Preferences.tooltips=TRUE; else Preferences.tooltips=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_BIGV,BM_GETCHECK,0,0))
            Preferences.bigv=TRUE; else Preferences.bigv=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_POSITIONS,BM_GETCHECK,0,0))
            Preferences.use_positions=TRUE; else Preferences.use_positions=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLSMALL,BM_GETCHECK,0,0))
            Preferences.toolsmall=TRUE; else Preferences.toolsmall=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_TOOLLEFT,BM_GETCHECK,0,0))
            Preferences.toolleft=TRUE; else Preferences.toolleft=FALSE;
         if(Preferences.buttons)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"BUTTONS",w,IniFilename);
         if(Preferences.quick_trick)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"QUICKTRICK",w,IniFilename);
         if(Preferences.undo){
           w=yesyes;
           UNDO_ON=1;
         }
         else{
           UNDO_ON=0;
           Empty_Undo();
           w=nono;
         }
         WritePrivateProfileString(IniSection,"UNDO",w,IniFilename);
         { char pref_str[32];
           Preferences.Vcolour[0]=uColour[0];
           Preferences.Vcolour[1]=uColour[1];
           Preferences.Vcolour[2]=uColour[2];
           sprintf(pref_str,"%ld %ld %ld",(long)uColour[0],
                   (long)uColour[1],(long)uColour[2]);
           WritePrivateProfileString(IniSection,"VIEWCOLOUR",
                                      pref_str,IniFilename);
         }
         if(Preferences.tooltips)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"TOOLTIPS",w,IniFilename);
         if(Preferences.bigv)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"BIGV",w,IniFilename);
         if(Preferences.use_positions)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"POSITIONS",w,IniFilename);
         if(Preferences.toolsmall)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"TOOLSMALL",w,IniFilename);
         if(Preferences.toolleft)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"TOOLLEFT",w,IniFilename);
         break;
       default:  break;
     }
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PREFERENCES_VCOLOUR:
          SetSfxColour(uColour,IDX_MISC_BACK_COLOUR,hdlg);
          InvalidateRect(GetDlgItem(hdlg,DLG_PREFERENCES_VCOLOUR),NULL,FALSE);
          break;
        default: break;
      }
      break;
   default: break;
 }
 return(FALSE);
}

static BOOL CALLBACK StartupPreferencesDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL state;
 char *w, yesyes[]="YES", nono[]="NO";
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     if(Preferences.use_OpenGL)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_OPENGL,BM_SETCHECK,state,0);
     if(Preferences.background)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFERENCES_BACKGROUND,BM_SETCHECK,state,0);
     return (TRUE);
   case WM_NOTIFY:
     switch(((LPNMHDR)lparam)->code){
       case PSN_APPLY:
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_OPENGL,BM_GETCHECK,0,0))
            Preferences.use_OpenGL=TRUE; else Preferences.use_OpenGL=FALSE;
         if(SendDlgItemMessage(hdlg,DLG_PREFERENCES_BACKGROUND,BM_GETCHECK,0,0))
            Preferences.background=TRUE; else Preferences.background=FALSE;
         if(Preferences.use_OpenGL)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"USEOPENGL",w,IniFilename);
         if(Preferences.background)w=yesyes; else w=nono;
         WritePrivateProfileString(IniSection,"BACKIMAGE",w,IniFilename);
         break;
       default:  break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

void PreferencesPropertyDlgProc(HWND hwndOwner){
 char title[256];
 int i;
 PROPSHEETPAGE psp[2];
 PROPSHEETHEADER psh;
 for(i=0;i<2;i++){
   psp[i].dwSize = sizeof(PROPSHEETPAGE);
   psp[i].dwFlags = PSP_DEFAULT;
   psp[i].hInstance = ghinst_main;
   psp[i].pszIcon = NULL;
   psp[i].lParam = 0;
 }
 psp[0].pszTemplate = MAKEINTRESOURCE(DLG_PREFERENCES_GENERAL);
 psp[0].pfnDlgProc = GeneralPreferencesDlgProc;
 psp[1].pszTemplate = MAKEINTRESOURCE(DLG_PREFERENCES_START);
 psp[1].pfnDlgProc = StartupPreferencesDlgProc;
 psh.dwSize = sizeof(PROPSHEETHEADER);
 psh.dwFlags = PSH_PROPSHEETPAGE | PSH_NOAPPLYNOW;
 psh.hwndParent = hwndOwner;
 psh.hInstance = ghinst_main;
 psh.nStartPage = 0;
 psh.pszIcon = NULL;
 LoadString(ghinst_main,IDX_MISC_PREF_TITLE,title,255);
 psh.pszCaption = (LPSTR)title;
 psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
 psh.ppsp = (LPCPROPSHEETPAGE) &psp;
 PropertySheet(&psh);
 return;
}

//end Preferences

#include "..\common\animate\about.h"

BOOL CALLBACK AboutDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 char name[255];
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     CentreDialogOnScreen(hdlg);
//     strcpy(name,gszHomeDir);
//     strcat(name,"OPENFX.AVI");
     SendDlgItemMessage(hdlg,DLG_ABOUT_VERSION,WM_SETTEXT,
                       0,(LPARAM)ABOUT_VERSION);
//     Animate_Open(GetDlgItem(hdlg, DLG_ABOUT_ANIMATE),name);
     return TRUE;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case IDOK:
          EndDialog(hdlg,TRUE);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK RotateDlgProc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 int i;
 struct np {int a; BOOL clock;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(struct np *)lparam;
       SetDlgItemInt(hwnd,DLG_ROTATE_ANGLE,dp->a,FALSE);
       SendDlgItemMessage(hwnd,DLG_ROTATE_ANGLE,EM_LIMITTEXT,(WPARAM)6,0);
       SendDlgItemMessage(hwnd,DLG_ROTATE_ANGLESPIN,SPNM_SETRANGE,0,
                          MAKELPARAM(0,360));
       SendDlgItemMessage(hwnd,DLG_ROTATE_ANGLESPIN,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->a,0);
       SendDlgItemMessage(hwnd,DLG_ROTATE_ANGLESPIN,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_ROTATE_ANGLE));
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       if(dp->clock == 0)active=TRUE; else active=FALSE;
       SendDlgItemMessage(hwnd,DLG_ROTATE_CLOCK,BM_SETCHECK,active,0);
       SendDlgItemMessage(hwnd,DLG_ROTATE_ANTICLOCK,BM_SETCHECK,!active,0);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
          i=GetDlgItemInt(hwnd,DLG_ROTATE_ANGLE,&err,FALSE);
          if(!err)EndDialog(hwnd,FAIL);
          dp->a=i;
          if(SendDlgItemMessage(hwnd,DLG_ROTATE_CLOCK,BM_GETCHECK,0,0))
               dp->clock=TRUE;
          else dp->clock=FALSE;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK MoveDlgProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam){
 long i,x,y,z;
 char temp[64];
 vertex *vp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       strcpy(temp,"0.00");
       SetDlgItemText(hwnd,DLG_MOVE_X,temp);
       SetDlgItemText(hwnd,DLG_MOVE_Y,temp);
       SetDlgItemText(hwnd,DLG_MOVE_Z,temp);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_MOVE_X,temp,32) == 0)
            EndDialog(hwnd,FALSE);
          x=(long)(atof(temp)*ruler);
          if(GetDlgItemText(hwnd,DLG_MOVE_Y,temp,32) == 0)
            EndDialog(hwnd,FALSE);
          y=(long)(atof(temp)*ruler);
          if(GetDlgItemText(hwnd,DLG_MOVE_Z,temp,32) == 0)
            EndDialog(hwnd,FALSE);
          z=(long)(atof(temp)*ruler);
          vp=MainVp; if(Nvert > 0)for(i=0;i<Nvert;i++){
            if(vp->status==SELECTED){
              vp->xyz[0] += x;
              vp->xyz[1] += y;
              vp->xyz[2] += z;
            }
            vp++;
          }
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

// Primitives Dialog code

extern long PrimitiveType;

#define D_BORDER 4
#define N_PRIMS 10

typedef struct tag_dlghdr1 {
  HWND hwndDlg[N_PRIMS];
  RECT rcTab;
} DLGHDR1;

static DLGHDR1 DlgHdrPrimitives,*lpDlgHdrPrimitives;
static HWND    hwndTabPrimitives;

static void UpdateCubePrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->width);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_WIDTH,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_WIDTH,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->length);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_LENGTH,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_LENGTH,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->height);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_HEIGHT,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CUBE_HEIGHT,EM_LIMITTEXT,(WPARAM)12,0);
 return;
}

BOOL CALLBACK CubePrimitiveDlgProc(HWND hwnd, UINT msg,
                                   WPARAM wparam, LPARAM lparam){
 char tempstr[16];
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateCubePrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_CUBE_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdateCubePrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_CUBE_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_CUBE_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_CUBE_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CUBE_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CUBE_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CUBE_WIDTH,tempstr,10);
          dp->width=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CUBE_LENGTH,tempstr,10);
          dp->length=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CUBE_HEIGHT,tempstr,10);
          dp->height=atof(tempstr);
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateSpherePrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 BOOL active;
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 if(dp->smooth == 1)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hwnd,DLG_PR_SPH_ON,BM_SETCHECK,active,0);
 SendDlgItemMessage(hwnd,DLG_PR_SPH_OFF,BM_SETCHECK,!active,0);
 if(dp->faces == 20)
          SendDlgItemMessage(hwnd,DLG_PR_SPH_20,BM_SETCHECK,1,0);
 if(dp->faces == 80)
          SendDlgItemMessage(hwnd,DLG_PR_SPH_80,BM_SETCHECK,1,0);
 if(dp->faces == 320)
          SendDlgItemMessage(hwnd,DLG_PR_SPH_320,BM_SETCHECK,1,0);
 if(dp->faces == 1280)
          SendDlgItemMessage(hwnd,DLG_PR_SPH_1280,BM_SETCHECK,1,0);
 return;
}

BOOL CALLBACK SpherePrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL active;
 char tempstr[16];
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateSpherePrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_SPH_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdateSpherePrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_SPH_20:
          dp->faces=20;
          break;
        case DLG_PR_SPH_80:
          dp->faces=80;
          break;
        case DLG_PR_SPH_320:
          dp->faces=320;
          break;
        case DLG_PR_SPH_1280:
          dp->faces=1280;
          break;
        case DLG_PR_SPH_ON:
          dp->smooth=1;
          break;
        case DLG_PR_SPH_OFF:
          dp->smooth=0;
          break;
        case DLG_PR_SPH_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_SPH_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_SPH_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_SPH_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_SPH_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_SPH_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateSphere1PrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 BOOL active;
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 if(dp->smooth == 1)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_ON,BM_SETCHECK,active,0);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_OFF,BM_SETCHECK,!active,0);
 SetDlgItemInt(hwnd,DLG_PR_SPH1_VERT,dp->nvert,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_VERT,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_VERTS,SPNM_SETRANGE,0,
                          MAKELPARAM(3,1000));
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_VERTS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nvert,0);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_VERTS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_SPH1_VERT));
 SetDlgItemInt(hwnd,DLG_PR_SPH1_HORZ,dp->nhorz,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_HORZ,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_HORZS,SPNM_SETRANGE,0,
                          MAKELPARAM(4,1000));
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_HORZS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nhorz,0);
 SendDlgItemMessage(hwnd,DLG_PR_SPH1_HORZS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_SPH1_HORZ));
 return;
}

BOOL CALLBACK Sphere1PrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 char tempstr[16];
 int i;
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateSphere1PrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_SPH1_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdateSphere1PrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_SPH1_ON:
          dp->smooth=1;
          break;
        case DLG_PR_SPH1_OFF:
          dp->smooth=0;
          break;
        case DLG_PR_SPH1_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_SPH1_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_SPH1_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_SPH1_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_SPH1_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_SPH1_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          i=GetDlgItemInt(hwnd,DLG_PR_SPH1_VERT,&err,FALSE);
          if(i > 9999)i=9999;
          if(i < 3)i=3;
          dp->nvert=i;
          i=GetDlgItemInt(hwnd,DLG_PR_SPH1_HORZ,&err,FALSE);
          if(i > 9999)i=9999;
          if(i < 4)i=4;
          dp->nhorz=i;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdatePolygonPrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 SetDlgItemInt(hwnd,DLG_PR_POLY_SIDES,dp->nsides,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_SIDES,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_SIDESS,SPNM_SETRANGE,0,
                          MAKELPARAM(4,1000));
 SendDlgItemMessage(hwnd,DLG_PR_POLY_SIDESS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nsides,0);
 SendDlgItemMessage(hwnd,DLG_PR_POLY_SIDESS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_POLY_SIDES));
 return;
}

BOOL CALLBACK PolygonPrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 char tempstr[16];
 int i;
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdatePolygonPrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_UPDATEPDIALOG:
     UpdatePolygonPrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDOK:
          i=GetDlgItemInt(hwnd,DLG_PR_POLY_SIDES,&err,FALSE);
          if(i > 9999)i=9999;
          if(i < 4)i=4;
          dp->nsides=i;
          GetDlgItemText(hwnd,DLG_PR_POLY_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_POLY_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_POLY_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_POLY_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateDiskPrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 SetDlgItemInt(hwnd,DLG_PR_DISK_SIDES,dp->nsides,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_SIDES,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_SIDESS,SPNM_SETRANGE,0,
                          MAKELPARAM(4,1000));
 SendDlgItemMessage(hwnd,DLG_PR_DISK_SIDESS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nsides,0);
 SendDlgItemMessage(hwnd,DLG_PR_DISK_SIDESS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_DISK_SIDES));
 return;
}

BOOL CALLBACK DiskPrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 char tempstr[16];
 int i;
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateDiskPrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_DISK_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdateDiskPrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_DISK_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_DISK_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_DISK_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_DISK_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_DISK_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_DISK_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          i=GetDlgItemInt(hwnd,DLG_PR_DISK_SIDES,&err,FALSE);
          if(i > 9999)i=9999; if(i < 4)i=4;
          dp->nsides=i;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateTubePrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->length);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_LENGTH,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_LENGTH,EM_LIMITTEXT,(WPARAM)12,0);
 SetDlgItemInt(hwnd,DLG_PR_TUBE_SIDES,dp->nsides,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_SIDES,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_SIDESS,SPNM_SETRANGE,0,
                          MAKELPARAM(4,1000));
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_SIDESS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nsides,0);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_SIDESS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_TUBE_SIDES));
 SetDlgItemInt(hwnd,DLG_PR_TUBE_NSTEPS,dp->nsteps,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_NSTEPS,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_NSTEPSS,SPNM_SETRANGE,0,
                          MAKELPARAM(1,1000));
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_NSTEPSS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nsteps,0);
 SendDlgItemMessage(hwnd,DLG_PR_TUBE_NSTEPSS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_TUBE_NSTEPS));
 return;
}

BOOL CALLBACK TubePrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 char tempstr[16];
 int i;
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateTubePrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_UPDATEPDIALOG:
     UpdateTubePrimitiveDlgProc(hwnd,dp);
     break;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_TUBE_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_TUBE_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_TUBE_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_TUBE_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TUBE_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TUBE_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TUBE_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TUBE_LENGTH,tempstr,10);
          dp->length=atof(tempstr);
          i=GetDlgItemInt(hwnd,DLG_PR_TUBE_SIDES,&err,FALSE);
          if(i > 9999)i=9999; if(i < 4)i=4;
          dp->nsides=i;
          i=GetDlgItemInt(hwnd,DLG_PR_TUBE_NSTEPS,&err,FALSE);
          if(i > 9999)i=9999; if(i < 1)i=1;
          dp->nsteps=i;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateCylinderPrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->length);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_LENGTH,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_LENGTH,EM_LIMITTEXT,(WPARAM)12,0);
 SetDlgItemInt(hwnd,DLG_PR_CYL_SIDES,dp->nsides,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_SIDES,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_SIDESS,SPNM_SETRANGE,0,
                          MAKELPARAM(4,1000));
 SendDlgItemMessage(hwnd,DLG_PR_CYL_SIDESS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nsides,0);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_SIDESS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_CYL_SIDES));
 SetDlgItemInt(hwnd,DLG_PR_CYL_NSTEPS,dp->nsteps,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_NSTEPS,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_NSTEPSS,SPNM_SETRANGE,0,
                          MAKELPARAM(1,1000));
 SendDlgItemMessage(hwnd,DLG_PR_CYL_NSTEPSS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nsteps,0);
 SendDlgItemMessage(hwnd,DLG_PR_CYL_NSTEPSS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_CYL_NSTEPS));
 return;
}

BOOL CALLBACK CylinderPrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 char tempstr[16];
 int i;
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateCylinderPrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_CYL_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdateCylinderPrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_CYL_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_CYL_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_CYL_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CYL_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CYL_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CYL_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CYL_LENGTH,tempstr,10);
          dp->length=atof(tempstr);
          i=GetDlgItemInt(hwnd,DLG_PR_CYL_SIDES,&err,FALSE);
          if(i > 9999)i=9999; if(i < 4)i=4;
          dp->nsides=i;
          i=GetDlgItemInt(hwnd,DLG_PR_CYL_NSTEPS,&err,FALSE);
          if(i > 9999)i=9999; if(i < 1)i=1;
          dp->nsteps=i;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateConePrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->length);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_LENGTH,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_LENGTH,EM_LIMITTEXT,(WPARAM)12,0);
 SetDlgItemInt(hwnd,DLG_PR_CONE_SIDES,dp->nsides,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_SIDES,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_SIDESS,SPNM_SETRANGE,0,
                    MAKELPARAM(4,1000));
 SendDlgItemMessage(hwnd,DLG_PR_CONE_SIDESS,SPNM_SETCRNTVALUE,
                    (WPARAM)dp->nsides,0);
 SendDlgItemMessage(hwnd,DLG_PR_CONE_SIDESS,SPNM_SETEDITCTRL,0,
                    (LPARAM)GetDlgItem(hwnd,DLG_PR_CONE_SIDES));
 return;
}

BOOL CALLBACK ConePrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 char tempstr[16];
 int i;
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateConePrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_CONE_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdateConePrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_CONE_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_CONE_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_CONE_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CONE_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CONE_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CONE_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_CONE_LENGTH,tempstr,10);
          dp->length=atof(tempstr);
          i=GetDlgItemInt(hwnd,DLG_PR_CONE_SIDES,&err,FALSE);
          if(i > 9999)i=9999; if(i < 4)i=4;
          dp->nsides=i;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void  UpdatePlanePrimitiveDlgProc(HWND hwnd,PRIMITIVE *dp){
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->width);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_WIDTH,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_WIDTH,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->length);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_LENGTH,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_PLANE_LENGTH,EM_LIMITTEXT,(WPARAM)12,0);
 return;
}

BOOL CALLBACK PlanePrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 char tempstr[16];
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdatePlanePrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_PLANE_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdatePlanePrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_PLANE_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_PLANE_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_PLANE_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_PLANE_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_PLANE_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_PLANE_WIDTH,tempstr,10);
          dp->width=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_PLANE_LENGTH,tempstr,10);
          dp->length=atof(tempstr);
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateTorusPrimitiveDlgProc(HWND hwnd, PRIMITIVE *dp){
 BOOL active;
 char tempstr[16];
 sprintf(tempstr,"%.2lf",dp->lr);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_LR,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_LR,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->fb);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_FB,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_FB,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->ud);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_UD,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_UD,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->radius);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_RADIUS,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_RADIUS,EM_LIMITTEXT,(WPARAM)12,0);
 sprintf(tempstr,"%.2lf",dp->torus_ratio);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_RATIO,WM_SETTEXT,0,(LPARAM)tempstr);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_RATIO,EM_LIMITTEXT,(WPARAM)12,0);
 if(dp->smooth == 1)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_ON,BM_SETCHECK,active,0);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_OFF,BM_SETCHECK,!active,0);
 SetDlgItemInt(hwnd,DLG_PR_TORUS_VERT,dp->nvert,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_VERT,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_VERTS,SPNM_SETRANGE,0,
                          MAKELPARAM(3,1000));
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_VERTS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nvert,0);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_VERTS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_TORUS_VERT));
 SetDlgItemInt(hwnd,DLG_PR_TORUS_HORZ,dp->nhorz,FALSE);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_HORZ,EM_LIMITTEXT,(WPARAM)6,0);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_HORZS,SPNM_SETRANGE,0,
                          MAKELPARAM(4,1000));
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_HORZS,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->nhorz,0);
 SendDlgItemMessage(hwnd,DLG_PR_TORUS_HORZS,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_PR_TORUS_HORZ));
 return;
}

BOOL CALLBACK TorusPrimitiveDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 BOOL err,active;
 char tempstr[16];
 int i;
 PRIMITIVE *dp=&PrimStruct;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateTorusPrimitiveDlgProc(hwnd,dp);
     return (TRUE);
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_PR_TORUS_COLOUR){
         r=dp->fc[0]; g=dp->fc[1]; b=dp->fc[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_UPDATEPDIALOG:
     UpdateTorusPrimitiveDlgProc(hwnd,dp);
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PR_TORUS_ON:
          dp->smooth=1;
          break;
        case DLG_PR_TORUS_OFF:
          dp->smooth=0;
          break;
        case DLG_PR_TORUS_COLOUR:
          SetSfxColour(dp->fc,IDX_MISC_CHOOSE_COLOUR,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_PR_TORUS_COLOUR),NULL,FALSE);
          break;
        case IDOK:
          GetDlgItemText(hwnd,DLG_PR_TORUS_LR,tempstr,10);
          dp->lr=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TORUS_FB,tempstr,10);
          dp->fb=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TORUS_UD,tempstr,10);
          dp->ud=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TORUS_RADIUS,tempstr,10);
          dp->radius=atof(tempstr);
          GetDlgItemText(hwnd,DLG_PR_TORUS_RATIO,tempstr,10);
          dp->torus_ratio=atof(tempstr);
          i=GetDlgItemInt(hwnd,DLG_PR_TORUS_VERT,&err,FALSE);
          if(i > 9999)i=9999;
          if(i < 3)i=3;
          dp->nvert=i;
          i=GetDlgItemInt(hwnd,DLG_PR_TORUS_HORZ,&err,FALSE);
          if(i > 9999)i=9999;
          if(i < 4)i=4;
          dp->nhorz=i;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL CALLBACK PrimitivesDlgProc(HWND hdlg,UINT msg,
                                WPARAM wparam,LPARAM lparam){
 static int xpos= -1,ypos= -1;
 int i,xp,yp;
 RECT rc,rd;
 char str[32];
 switch( msg ){
   // put in PAINT statement here if required RSF TODO
   case WM_INITDIALOG:{
       TC_ITEM tie;
       int h;
       DWORD dwDlgBase=GetDialogBaseUnits();
       if(PrimStruct.auto_interactive)
         SendDlgItemMessage(hdlg,DLG_TAB_PRIMES_AUTO,BM_SETCHECK,TRUE,0);
       else
         SendDlgItemMessage(hdlg,DLG_TAB_PRIMES_AUTO,BM_SETCHECK,FALSE,0);
       if(xpos == -1 && ypos == -1){
         xp=GetPrivateProfileInt(IniSection,"PRPOSX",-1,IniFilename);
         yp=GetPrivateProfileInt(IniSection,"PRPOSY",-1,IniFilename);
         if(xp < 0 || yp < 0){
           RECT rc1;
           GetWindowRect(hdlg,&rc1);
           GetWindowRect(ghwndTools,&rc);
           if(bToolpannelLeft)xpos=rc.right+5;
           else               xpos=rc.left-(rc1.right-rc1.left)-5;
           ypos=rc.top;
         }
         else{
           xpos=xp; ypos=yp;
         }
       }
       SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
       lpDlgHdrPrimitives=&DlgHdrPrimitives;
       GetWindowRect(hdlg,&rc);
       GetWindowRect(GetDlgItem(hdlg,IDOK),&rd);
       h=rd.top-rc.top-GetSystemMetrics(SM_CYCAPTION);
       GetClientRect(hdlg,&rc);
       hwndTabPrimitives=CreateWindowEx(
                            0,
                            WC_TABCONTROL,"Hello ",
                            TCS_MULTILINE |//TCS_FOCUSNEVER |
                            WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
                            (D_BORDER*LOWORD(dwDlgBase))/4,
                            (D_BORDER*HIWORD(dwDlgBase))/8,
                            rc.right-(2*D_BORDER*LOWORD(dwDlgBase))/4,
                            h-(2*D_BORDER*HIWORD(dwDlgBase))/8,
                            hdlg,NULL,ghinst_main,NULL);
       {char text_string[255];
         LoadString(ghinst_main,IDX_MISC_LANGUAGE,
                    text_string,sizeof(text_string));
         if(strcmp(text_string,"ENGLISH") == 0)
           SendMessage(hwndTabPrimitives,WM_SETFONT, /* for text in tab */
                  (WPARAM)GetStockObject(ANSI_VAR_FONT),0);
       }
       memset(&tie,0,sizeof(TC_ITEM));
       tie.mask = TCIF_TEXT;// | TCIF_IMAGE;
//       tie.iImage = -1;
       tie.pszText=str;
       LoadString(ghinst_main,IDX_PRIMS_RECTANGLE,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,0,&tie);
       LoadString(ghinst_main,IDX_PRIMS_CUBE,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,1,&tie);
       LoadString(ghinst_main,IDX_PRIMS_POLYGON,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,2,&tie);
       LoadString(ghinst_main,IDX_PRIMS_TUBE,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,3,&tie);
       LoadString(ghinst_main,IDX_PRIMS_DISK,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,4,&tie);
       LoadString(ghinst_main,IDX_PRIMS_CYLINDER,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,5,&tie);
       LoadString(ghinst_main,IDX_PRIMS_CONE,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,6,&tie);
       LoadString(ghinst_main,IDX_PRIMS_SPHERE,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,7,&tie);
       LoadString(ghinst_main,IDX_PRIMS_SPHERE1,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,8,&tie);
       LoadString(ghinst_main,IDX_PRIMS_TORUS,str,32);
       TabCtrl_InsertItem(hwndTabPrimitives,9,&tie);
       SetRectEmpty(&rc);
       TabCtrl_AdjustRect(hwndTabPrimitives,FALSE,&rc);
       rc.left += (D_BORDER*LOWORD(dwDlgBase))/4;
       rc.top += (D_BORDER*HIWORD(dwDlgBase))/8;
       CopyRect(&(lpDlgHdrPrimitives->rcTab),&rc);
       for(i=0;i<N_PRIMS;i++)lpDlgHdrPrimitives->hwndDlg[i]=NULL;
       lpDlgHdrPrimitives->hwndDlg[0]=CreateDialog(ghinst_main,
                                   MAKEINTRESOURCE(DLG_PR_PLANE),
                                   hdlg,(DLGPROC)PlanePrimitiveDlgProc);
       SetWindowPos(lpDlgHdrPrimitives->hwndDlg[0],NULL,
                    lpDlgHdrPrimitives->rcTab.left,
                    lpDlgHdrPrimitives->rcTab.top,0,0,
                    SWP_NOSIZE | SWP_NOZORDER);
       ShowWindow(lpDlgHdrPrimitives->hwndDlg[0],SW_SHOW);
     }
     return TRUE;
   case WM_NOTIFY:
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGING){
       int id=TabCtrl_GetCurSel(hwndTabPrimitives);
     }
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGE){
       int id=TabCtrl_GetCurSel(hwndTabPrimitives);
       for(i=0;i<N_PRIMS;i++){
         if(lpDlgHdrPrimitives->hwndDlg[i] != NULL){
           SendMessage(lpDlgHdrPrimitives->hwndDlg[i],
                       WM_COMMAND,(WPARAM)IDOK,0);
           DestroyWindow(lpDlgHdrPrimitives->hwndDlg[i]);
           lpDlgHdrPrimitives->hwndDlg[i]=NULL;
         }
       }
       PrimitiveType=id;
       if(id == 0)lpDlgHdrPrimitives->hwndDlg[0]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_PLANE),
                                     hdlg,(DLGPROC)PlanePrimitiveDlgProc);
       if(id == 1)lpDlgHdrPrimitives->hwndDlg[1]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_CUBE),
                                     hdlg,(DLGPROC)CubePrimitiveDlgProc);
       if(id == 2)lpDlgHdrPrimitives->hwndDlg[2]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_POLY),
                                     hdlg,(DLGPROC)PolygonPrimitiveDlgProc);
       if(id == 3)lpDlgHdrPrimitives->hwndDlg[3]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_TUBE),
                                     hdlg,(DLGPROC)TubePrimitiveDlgProc);
       if(id == 4)lpDlgHdrPrimitives->hwndDlg[4]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_DISK),
                                     hdlg,(DLGPROC)DiskPrimitiveDlgProc);
       if(id == 5)lpDlgHdrPrimitives->hwndDlg[5]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_CYL),
                                     hdlg,(DLGPROC)CylinderPrimitiveDlgProc);
       if(id == 6)lpDlgHdrPrimitives->hwndDlg[6]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_CONE),
                                     hdlg,(DLGPROC)ConePrimitiveDlgProc);
       if(id == 7)lpDlgHdrPrimitives->hwndDlg[7]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_SPH),
                                     hdlg,(DLGPROC)SpherePrimitiveDlgProc);
       if(id == 8)lpDlgHdrPrimitives->hwndDlg[8]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_SPH1),
                                     hdlg,(DLGPROC)Sphere1PrimitiveDlgProc);
       if(id == 9)lpDlgHdrPrimitives->hwndDlg[9]=CreateDialog(ghinst_main,
                                     MAKEINTRESOURCE(DLG_PR_TORUS),
                                     hdlg,(DLGPROC)TorusPrimitiveDlgProc);
       if(lpDlgHdrPrimitives->hwndDlg[id] != NULL){
         SetWindowPos(lpDlgHdrPrimitives->hwndDlg[id],NULL,
                      lpDlgHdrPrimitives->rcTab.left,
                      lpDlgHdrPrimitives->rcTab.top,0,0,
                      SWP_NOSIZE | SWP_NOZORDER);
         ShowWindow(lpDlgHdrPrimitives->hwndDlg[id],SW_SHOW);
       }
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       default:
         break;
       case SC_CLOSE:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
         break;
     }
     break;
   case WM_COMMAND:
     switch( LOWORD(wparam) ){
       case DLG_TAB_PRIMES_AUTO:
         if(SendDlgItemMessage(hdlg,DLG_TAB_PRIMES_AUTO,BM_GETCHECK,0,0))
           PrimStruct.auto_interactive=TRUE;
         else
           PrimStruct.auto_interactive=FALSE;
         break;
       case IDOK:
         for(i=0;i<N_PRIMS;i++){
           if(lpDlgHdrPrimitives->hwndDlg[i] != NULL)
            SendMessage(lpDlgHdrPrimitives->hwndDlg[i],WM_COMMAND,
                        (WPARAM)IDOK,0);
         }
         BuildPrimitive(PrimitiveType,ActiveView);
         break;
       case IDCANCEL:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
       default:
         break;
     }
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_UPDATEPDIALOG:
     /* send message to active dialog to update it */
     for(i=0;i<N_PRIMS;i++){
       if(lpDlgHdrPrimitives->hwndDlg[i] != NULL)
         SendMessage(lpDlgHdrPrimitives->hwndDlg[i],WM_UPDATEPDIALOG,0,0);
     }
     break;
   case WM_INTERROGATE:
     /* send message to interrogate the dialog it */
     for(i=0;i<N_PRIMS;i++){
       if(lpDlgHdrPrimitives->hwndDlg[i] != NULL)
         SendMessage(lpDlgHdrPrimitives->hwndDlg[i],WM_COMMAND,(WPARAM)IDOK,0);
     }
     break;
   case WM_DESTROY:
     for(i=0;i<N_PRIMS;i++){
       if(lpDlgHdrPrimitives->hwndDlg[i] != NULL)
         DestroyWindow(lpDlgHdrPrimitives->hwndDlg[i]);
       lpDlgHdrPrimitives->hwndDlg[i]=NULL;
     }
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     sprintf(str,"%ld",xpos);
     WritePrivateProfileString(IniSection,"PRPOSX",str,IniFilename);
     sprintf(str,"%ld",ypos);
     WritePrivateProfileString(IniSection,"PRPOSY",str,IniFilename);
     break;
   default: break;
 }
 return( FALSE );
}

BOOL IsPrimitivesDialogMessage(MSG *msg){
 int i;
 DLGHDR1 *lpDlgHdr;
 lpDlgHdr=&DlgHdrPrimitives;
 for(i=0;i<N_PRIMS;i++){
   if(lpDlgHdr->hwndDlg[i] != NULL){
     if(IsDialogMessage(lpDlgHdr->hwndDlg[i],msg))return TRUE;
   }
 }
 return IsDialogMessage(ghwndDlgPrimitives,msg);
}
