/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file DRAFT1.C */

char szToolboxClass[]="SFX:DraftToolboxClass",
     szToolBtnClass[]="SFX:DraftToolBtnClass";

typedef struct tagTOOL_CONTROL{
 int     idToolBtnUp,idToolBtnDown;
 HBITMAP hbmToolBtnUp,hbmToolBtnDown;
 int     Message,Help,Tip;
} TOOL_CONTROL;

#define TOOLBOXSTYLE  (WS_POPUP | WS_CLIPSIBLINGS | !WS_CAPTION | !WS_SYSMENU)

static HWND CreateCommandPannel(TOOL_CONTROL aTC[],int n,char *title,
                                int nc,int nr,int x,int y,int gap,HWND parent);
static void DeleteCommandPannel(TOOL_CONTROL aTC[],int n);
static void PannelDrawBitmap(HDC hDC,int tp,TOOL_CONTROL aT[],int type,
                             int ctool);
static void PannelSelectTool(HWND hwnd, int *oldtool, int newtool);
static void DoAddTooltip(HWND hwndOwner, HWND tipon, int resid);
static HWND DoCreateTooltip(HWND hwndOwner);

#define NO_MESSAGE -1
#define NO_HELP    -1

#define NTOOLS         7

#define TOOLCOLUMNS    1
#define TOOLROWS       7

static HWND hwndDraftTools=NULL,hwnd_dtools,hwndTooltip=NULL;
static int gCurDraftTool=-1;
static int tool=0;
static HINSTANCE ghinst_draft;

static TOOL_CONTROL aTools[NTOOLS]={
  IDBM_DFT0U,IDBM_DFT0D,NULL,NULL,IDM_TP_0,NO_HELP,IDI_TIP_0,
  IDBM_DFT1U,IDBM_DFT1D,NULL,NULL,IDM_TP_1,NO_HELP,IDI_TIP_1,
  IDBM_DFT2U,IDBM_DFT2D,NULL,NULL,IDM_TP_2,NO_HELP,IDI_TIP_2,
  IDBM_DFT3U,IDBM_DFT3D,NULL,NULL,IDM_TP_3,NO_HELP,IDI_TIP_3,
  IDBM_DFT4U,IDBM_DFT4D,NULL,NULL,IDM_TP_4,NO_HELP,IDI_TIP_4,
  IDBM_DFT5U,IDBM_DFT5D,NULL,NULL,IDM_TP_5,NO_HELP,IDI_TIP_5,
  IDBM_DFT6U,IDBM_DFT6D,NULL,NULL,IDM_TP_6,NO_HELP,IDI_TIP_6
};

static int ToolbuttonSize=32,cxToolBtn,cyToolBtn;

static int FakeCaptionHeight (void){
 return (GetSystemMetrics (SM_CYCAPTION) / 3) + 2; // + 5;
}

static void DrawFakeCaption (HDC dc, HWND Wnd){
 HBRUSH hbr;
 RECT rClient;
 int width, height, brushKind;
 GetClientRect (Wnd, &rClient);
 rClient.bottom = rClient.top + FakeCaptionHeight();
// brushKind = (GetParent (Wnd) == GetFocus ()) ?
//   COLOR_ACTIVECAPTION : COLOR_INACTIVECAPTION;
 brushKind = COLOR_ACTIVECAPTION;
 hbr = CreateSolidBrush (GetSysColor (brushKind));
 FillRect (dc, &rClient, hbr);
 DeleteObject (hbr);
#if 0
 { HFONT hf,hfo;
   hf=CreateFont(-10, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE,
      ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
      DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS,
   "MS Sans Serif"); //      "Arial");
   hfo=SelectObject(dc,hf);
   SetBkColor(dc,GetSysColor(COLOR_ACTIVECAPTION));
   SetTextColor(dc,GetSysColor(COLOR_CAPTIONTEXT));
   TextOut(dc,2,1,"Tools",strlen("Tools"));
   SelectObject(dc,hfo);
   DeleteObject(hf);
 }
#endif
 MoveToEx(dc, rClient.left, rClient.bottom,NULL);
 LineTo(dc, rClient.right, rClient.bottom);
#if 0
 rClient.right = (FakeCaptionHeight() * 3) / 2;
 MoveToEx(dc, rClient.right, rClient.top,NULL);
 LineTo(dc, rClient.right, rClient.bottom);
 FillRect (dc, &rClient, GetStockObject (LTGRAY_BRUSH));
 width = rClient.right - rClient.left;
 rClient.left += (width / 4) + 1;
 rClient.right = rClient.left + (width / 2) + 1;
 height = rClient.bottom - rClient.top;
 rClient.top += height / 2;
 rClient.bottom = rClient.top + 3;
 FillRect (dc, &rClient, GetStockObject (GRAY_BRUSH));
 OffsetRect (&rClient, -1, -1);
 FillRect (dc, &rClient, GetStockObject (WHITE_BRUSH));
 FrameRect (dc, &rClient, GetStockObject (BLACK_BRUSH));
#endif
}

static BOOL CaptionHitTest(HWND Wnd, LONG lParam, BOOL CloseBoxOnly){
 POINT pt;
 RECT rClient;
 GetClientRect (Wnd, &rClient);
 rClient.bottom = rClient.top + FakeCaptionHeight();
 if (CloseBoxOnly) rClient.right = (FakeCaptionHeight() * 3) / 2;
 pt.x = LOWORD (lParam);
 pt.y = HIWORD (lParam);
 ScreenToClient (Wnd, &pt);
 return PtInRect (&rClient, pt);
}

static void ToolboxDestroy(void){
 if(hwndDraftTools)DestroyWindow(hwndDraftTools);
}

static HWND CreateCommandPannel(TOOL_CONTROL aTC[],int n,char *title,
                                int nc,int nr,int x,int y,int gap,
                                HWND parent){
 int i,j,cx,cy,captionH;
 RECT rc,rc1;
 BITMAP bmp;
 HWND hwnd,hwndB;
 captionH=FakeCaptionHeight();
 if(n > 0)for(i=0;i<n;i++){
   if((aTC[i].hbmToolBtnUp=LoadBitmap(ghinst_draft,
      MAKEINTRESOURCE(aTC[i].idToolBtnUp))) == NULL)return NULL;
   if((aTC[i].hbmToolBtnDown=LoadBitmap(ghinst_draft,
      MAKEINTRESOURCE(aTC[i].idToolBtnDown))) == NULL)return NULL;
 }
 cxToolBtn = ToolbuttonSize; /* bmp.bmWidth;  */
 cyToolBtn = ToolbuttonSize; /* bmp.bmHeight; */
 rc.left=0; rc.top=0;
 rc.right=nc*(ToolbuttonSize);
 rc.bottom=nr*(ToolbuttonSize);
 AdjustWindowRect(&rc,TOOLBOXSTYLE,FALSE);
 cx=rc.right-rc.left-1+2;
 cy=rc.bottom-rc.top-1+captionH+2;
 if((hwnd = CreateWindow(szToolboxClass,title,TOOLBOXSTYLE,
    x,y,cx,cy,hwnd_dtools,NULL,ghinst_draft,NULL)) == NULL)return NULL;
 hwndTooltip=DoCreateTooltip(hwnd);
 x=1; y=ToolbuttonSize*gap+captionH+1;
 if(n > 0)for(i=0,j=0;i<n;i++){
   hwndB=CreateWindow(szToolBtnClass,NULL,WS_CHILD | WS_VISIBLE,x,y,
                      ToolbuttonSize,ToolbuttonSize,
                      hwnd,(HMENU)i,ghinst_draft,NULL);
   DoAddTooltip(hwnd,hwndB,aTC[i].Tip);
   j++;
   if(j == nc){
     j=0; x=1; y += ToolbuttonSize;
   }
   else x += ToolbuttonSize;
 }
 return hwnd;
}

static void DeleteCommandPannel(TOOL_CONTROL aTC[],int n){
 int i;
 for(i=0;i<n;i++){
   DeleteObject(aTC[i].hbmToolBtnUp);
   aTC[i].hbmToolBtnUp = NULL;
   DeleteObject(aTC[i].hbmToolBtnDown);
   aTC[i].hbmToolBtnDown = NULL;
 }
}

static LRESULT CALLBACK ToolboxWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam){
 RECT r;
 switch (msg) {
   case WM_CREATE:
     return 0;
   case WM_NCHITTEST:
     if (CaptionHitTest(hwnd,lParam,FALSE))return (HTCAPTION);
     return DefWindowProc (hwnd,msg,wParam,lParam);
   case WM_NCLBUTTONDOWN:
     if(CaptionHitTest (hwnd,lParam,TRUE)){
//       PostMessage(hwnd_dtools,WM_COMMAND,IDM_CLOSETBX,0);
     }
     return DefWindowProc(hwnd,msg,wParam,lParam);
   case WM_KEYDOWN:
     SetActiveWindow(GetParent(hwnd));
     break;
   case WM_NCACTIVATE:
     return TRUE;
   case WM_ACTIVATE:
     GetClientRect(hwnd, &r);
     r.bottom = r.top + FakeCaptionHeight();
     InvalidateRect (hwnd,&r,FALSE);
     return (0);
   case WM_PAINT:{
       HDC hDC;
       COLORREF oldcolour;
       PAINTSTRUCT ps;
       HFONT holdFont=NULL;
       hDC=BeginPaint(hwnd,&ps);
       DrawFakeCaption(hDC,hwnd);
       GetClientRect(hwnd,&r);
       FrameRect(hDC,&r,GetStockObject(BLACK_BRUSH));
       EndPaint(hwnd,&ps);
     }
     break;
   case WM_DESTROY:{
       if(hwnd == hwndDraftTools){
         DeleteCommandPannel(aTools,NTOOLS);
         hwndDraftTools = NULL;
       }
     }
     break;
   default:
     return DefWindowProc(hwnd, msg, wParam, lParam);
 }
 return 0;
}

static LRESULT CALLBACK ToolBtnWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam){
 static int captured=0,*lpbuttonid=NULL;
 MSG mmsg;
 HWND hwnd_parent;
 hwnd_parent=GetParent(hwnd);
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 switch (msg) {
   case WM_MOUSEMOVE:
     if(hwndTooltip != NULL)SendMessage(hwndTooltip,TTM_RELAYEVENT,
                                        0,(LPARAM)(LPMSG)&mmsg);
     if(captured){
       RECT rc;
       POINT p;
       p.x=LOWORD(lParam),p.y=HIWORD(lParam);
       ClientToScreen(hwnd,&p);
       GetWindowRect(hwnd,&rc);
       if(p.x < rc.left || p.x > rc.right || p.y < rc.top || p.y > rc.bottom){
         if(lpbuttonid != NULL)PannelSelectTool(hwnd_parent,lpbuttonid,-1);
       }
       else{
         if(lpbuttonid != NULL)PannelSelectTool(hwnd_parent,lpbuttonid,
                               (int)GetWindowLong(hwnd,GWL_ID));
       }
     }
     break;
   case WM_LBUTTONDOWN:
     if(hwndTooltip != NULL)SendMessage(hwndTooltip,TTM_RELAYEVENT,
                                        0,(LPARAM)(LPMSG)&mmsg);
     if(hwnd_parent == hwndDraftTools )lpbuttonid = &gCurDraftTool;
     if(lpbuttonid != NULL)PannelSelectTool(hwnd_parent,lpbuttonid,
                           (int)GetWindowLong(hwnd,GWL_ID));
     captured=1;
     SetCapture(hwnd);
     break;
   case WM_LBUTTONUP:
     if(hwndTooltip != NULL)SendMessage(hwndTooltip,TTM_RELAYEVENT,
                                        0,(LPARAM)(LPMSG)&mmsg);
     if(lpbuttonid != NULL){
       if(hwnd_parent == hwndDraftTools && gCurDraftTool >= 0){
         if(aTools[gCurDraftTool].Message != NO_MESSAGE)PostMessage(hwnd_dtools,
           WM_COMMAND,aTools[gCurDraftTool].Message,0);
         PannelSelectTool(hwnd_parent,&gCurDraftTool,-1);
       }
     }
     captured=0; lpbuttonid=NULL;
     ReleaseCapture();
     break;
   case WM_PAINT:{
       HDC hDC;
       PAINTSTRUCT ps;
       hDC=BeginPaint(hwnd,&ps);
       if(hwnd_parent == hwndDraftTools)PannelDrawBitmap(hDC,1,
         aTools,(int)GetWindowLong(hwnd,GWL_ID),gCurDraftTool);
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return DefWindowProc(hwnd, msg, wParam, lParam);
 }
 return 0;
}

static void PannelDrawBitmap(HDC hDC, int tp, TOOL_CONTROL aT[],
                             int type, int ctool){
 HDC hMemDC;
 HBITMAP hbm;
 HBITMAP hbmOld;
 if(tp == 1){
   if(tool < NTOOLS){
     hbm = (type == ctool || type == tool) ? aT[type].hbmToolBtnDown : aT[type].hbmToolBtnUp;
   }
   else{
     hbm = (type == ctool || type == 1) ? aT[type].hbmToolBtnDown : aT[type].hbmToolBtnUp;
   }
 }
 else{
   hbm = (type == ctool) ? aT[type].hbmToolBtnDown : aT[type].hbmToolBtnUp;
 }
 if(hbm == NULL)return;
 hMemDC = CreateCompatibleDC(hDC);
 hbmOld = SelectObject(hMemDC, hbm);
 BitBlt(hDC,0,0,cxToolBtn,cyToolBtn,hMemDC,0,0,SRCCOPY);
 SelectObject(hMemDC, hbmOld);
 DeleteDC(hMemDC);
}

static void PannelSelectTool(HWND hwnd, int *oldtool, int newtool){
 if(hwnd != NULL){
   if(*oldtool >= 0)InvalidateRect(GetDlgItem(hwnd,*oldtool),NULL,FALSE);
   if( newtool >= 0)InvalidateRect(GetDlgItem(hwnd,newtool),NULL,FALSE);
 }
 *oldtool=newtool;
}

BOOL OpenDraftToolbox(HWND hwnd_parent, HINSTANCE hinst_parent){
 int x,y,sep=5;
 RECT rc,rc1;
 WNDCLASS wc;
 ghinst_draft=hinst_parent;
 hwnd_dtools=hwnd_parent;
 wc.lpszClassName = szToolboxClass;
 wc.style = 0;
 wc.lpfnWndProc = (WNDPROC)ToolboxWndProc;
 wc.cbClsExtra = 0;
 wc.cbWndExtra = 0;
 wc.hInstance = ghinst_draft;
 wc.hIcon = NULL;
 wc.hCursor = LoadCursor(NULL, IDC_ARROW);
 wc.hbrBackground = GetStockObject(LTGRAY_BRUSH);
 wc.lpszMenuName = NULL;
 if(!RegisterClass(&wc))return FALSE;
 wc.lpszClassName = szToolBtnClass;
 wc.style = 0;
 wc.lpfnWndProc = (WNDPROC)ToolBtnWndProc;
 wc.cbClsExtra = 0;
 wc.cbWndExtra = 0;
 wc.hInstance = ghinst_draft;
 wc.hIcon = NULL;
 wc.hCursor = LoadCursor(NULL, IDC_ARROW);
 wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
 wc.lpszMenuName = NULL;
 if(!RegisterClass(&wc))return FALSE;
 GetWindowRect(hwnd_dtools,&rc1);
 x=rc1.right-3*ToolbuttonSize;
 y=rc1.top+3*ToolbuttonSize;
 if((hwndDraftTools=CreateCommandPannel(aTools,NTOOLS," ",
                TOOLCOLUMNS,TOOLROWS,x,y,0,hwnd_parent)) == NULL)return FALSE;
 ShowWindow(hwndDraftTools,SW_SHOW);
 return TRUE;
}

void DestroyDraftToolbox(void){
 if(hwndDraftTools == NULL)return;
 DestroyWindow(hwndDraftTools);
 UnregisterClass(szToolboxClass,ghinst_draft);
 UnregisterClass(szToolBtnClass,ghinst_draft);
}

void SelectDraftTool(int nt){
 if(hwndDraftTools == NULL)return;
 InvalidateRect(GetDlgItem(hwndDraftTools,tool),NULL,FALSE);
 tool=nt;
 InvalidateRect(GetDlgItem(hwndDraftTools,tool),NULL,FALSE);
}

static void DoAddTooltip(HWND hwndOwner, HWND tipon, int resid){
 TOOLINFO ti;
 if(hwndTooltip == NULL || tipon == NULL)return;
 memset(&ti,0,sizeof(TOOLINFO));
 ti.cbSize = sizeof(TOOLINFO);
 ti.uFlags = TTF_IDISHWND;
 ti.hwnd = hwndOwner;
 ti.hinst = ghinst_draft;
 ti.uId = (UINT)tipon;
 ti.lpszText = (LPSTR)resid;
 SendMessage(hwndTooltip,TTM_ADDTOOL,0,(LPARAM) (LPTOOLINFO) &ti);
 return;
}

static HWND DoCreateTooltip(HWND hwndOwner){
 HWND hwndTT;
 hwndTT = CreateWindow(TOOLTIPS_CLASS, (LPSTR) NULL, TTS_ALWAYSTIP,
     CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
     hwndOwner, (HMENU) NULL, ghinst_draft, NULL);
 return hwndTT;
}
