/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file draw.c  */

// Drawing code


#define MODULE_DRAW 1

#include "design.h"

static short in_modeller_view2(vertex *, vertex *);
static short in_modeller_view1(vertex *, vertex *);
static void DrawAllEdgesInOne(HDC,int);
static void DrawAllVerticesInOne(HDC, int);
static void DrawOneVertexInOne(vertex *, HDC, int);
static void DrawWireFrame(HDC hdc,int skelview);
static void DrawQuickWireFrame(HDC hdc,int skelview);
static void DrawHiddenLine(HDC hdc);

static double trr[4][4], t1[4][4],
              t2[4][4],  t3[4][4];
static double scale,rad;
static vertex vobserver;
static int zoom_abort=NO,abort_enable=YES;
static int order_list[3][3]={ {0,1,2},{1,2,0},{2,0,1} };

int CheckInterrupt(void){
 int i;
 MSG msg;
 if(do_NOT_abort)return 0;
 if(PeekMessage(&msg,NULL,WM_MOUSEFIRST,WM_MOUSELAST,PM_REMOVE)){
   if(msg.message == WM_LBUTTONDOWN){
     if(GetKeyState(VK_SPACE) & 0x8000 || tool == PAN ||
        tool == INZOOM){
       for(i=0;i<3;i++)if(msg.hwnd == ghwnd_triview[i]){
         if(msg.hwnd != ghwnd_current){
           HDC hdc;
           hdc=GetDC(ghwnd_main);
           FocusActiveWindow(hdc,FALSE);
           ghwnd_current=msg.hwnd;
           ActiveView=GetWindowLong(msg.hwnd,GWL_ID);
           FocusActiveWindow(hdc,TRUE);
           ReleaseDC(ghwnd_main,hdc);
         }
       }
       PostMessage(msg.hwnd,msg.message,msg.wParam,msg.lParam);
       zoom_abort=0;
       return 1;
     }
   }
   else if(msg.message == WM_LBUTTONUP){
     for(i=0;i<3;i++){
       if(msg.hwnd == ghwnd_triview[i]){
         if(msg.hwnd != ghwnd_current){
           HDC hdc;
           hdc=GetDC(ghwnd_main);
           FocusActiveWindow(hdc,FALSE);
           ghwnd_current=msg.hwnd;
           ActiveView=GetWindowLong(msg.hwnd,GWL_ID);
           FocusActiveWindow(hdc,TRUE);
           ReleaseDC(ghwnd_main,hdc);
         }
         ScreenToClient(ghwnd_triview[i],&msg.pt);
         if((zoom_abort=CheckWindowIcon((int)msg.pt.x,(int)msg.pt.y)+1) > 0)
           return 1;
       }
     }
   }
 }
 return 0;
}

void SelectInPerspective(int tl,int boxl,int box,int boyl,int boy){
 vertex *vp;
 double x1,y1,z1;
 int i,ix1,iy1,pwincx,pwincy;
 RECT rc;
 GetClientRect(ghwnd_view,&rc);
 pwincx=rc.right/2; pwincy=rc.bottom/2;
 vp=MainVp; if(MainVp != NULL && Nvert > 0)for(i=0;i<Nvert;i++){
   if(intriview(vp)){
     m4by1(trr,(double)vp->xyz[0],(double)vp->xyz[1],(double)vp->xyz[2]
          ,&x1,&y1,&z1);
     if(y1 >= 0.0){
       ix1=(int)((double)pwincx+scale*x1);
       iy1=(int)((double)pwincy-scale*z1);
       if(ix1 > min(box,boxl) && ix1 < max(box,boxl) &&
          iy1 > min(boy,boyl) && iy1 < max(boy,boyl)){
         if(tl == SELECTOR && vp->status == DESELECTED){
           vp->status=SELECTED;
           NvertSelect++; NvertDeselect--;
         }
         else if(tl == DESELECTOR && vp->status == SELECTED){
           vp->status=DESELECTED;
           NvertSelect--; NvertDeselect++;
         }
       }
     }
    }
   vp++;
 }
}

void UpdateCounters(void){
  char temp[64];
  sprintf(temp,"Vertices=%ld(%ld) Faces=%ld",NvertSelect,Nvert,Nface);
  SendMessage(ghwnd_info,SB_SETTEXT,(WPARAM)1,(LPARAM)temp);
  if(data_visible && ghwndData != NULL)InvalidateRect(ghwndData,NULL,FALSE);
  if(ghwndDlgAttribs != NULL && IsWindow(ghwndDlgAttribs))
    SendMessage(ghwndDlgAttribs,WM_COMMAND,DLG_ATTRIBS_UPDATE,0);
}

void DrawVerticesOnly(vertex *vs){
 int i,j,n,k;
 RECT rc;
 HCURSOR lhcurSave;
 if(View == TRIVIEW)n=3;
 else               n=1;
 if(vs == NULL)lhcurSave=SetCursor(ghcurWait);
 k=ActiveView;
 for(j=0;j<n;j++){
   i=order_list[k][j];
   if(ghdc_triview_Bitmap[i] == NULL)return;
   if(ghbm_triview[i] == NULL)return;
   if(vs == NULL)DrawAllVerticesInOne(ghdc_triview_Bitmap[i],i);
   else          DrawOneVertexInOne(vs,ghdc_triview_Bitmap[i],i);
   DrawArrowIcons(ghdc_triview_Bitmap[i],i);
   InvalidateRect(ghwnd_triview[i],NULL,FALSE);
   UpdateWindow(ghwnd_triview[i]);
 }
 if(vs == NULL)SetCursor(lhcurSave);
}

void DrawOneEdgeOnly(vertex *v1, vertex *v2){
 int i,j,k,n;
 RECT rc;
 int sh1,sv1,sh2,sv2;
 short (*in_modeller_view)(vertex *vp1, vertex *vp2);
 HDC hdc;
 HPEN holdpen;
 if(View == TRIVIEW)n=3;
 else               n=1;
 k=ActiveView;
 if(!draw_joins)in_modeller_view = in_modeller_view2; /* scroll lock   */
 else           in_modeller_view = in_modeller_view1; /* test function */
 for(j=0;j<n;j++){
   i=order_list[k][j];
   if((hdc=ghdc_triview_Bitmap[i]) == NULL)return;
   if(ghbm_triview[i] == NULL)return;
   if(in_modeller_view(v1,v2)){
     GetWindowCoords(i,v1->xyz[0],v1->xyz[1],v1->xyz[2],&sh1,&sv1);
     GetWindowCoords(i,v2->xyz[0],v2->xyz[1],v2->xyz[2],&sh2,&sv2);
     holdpen=SelectObject(hdc,ghEdgePen);
     MoveToEx(hdc,sh1,sv1,NULL); LineTo(hdc,sh2,sv2);
     if(v1->status == SELECTED)SelectObject(hdc,ghSelectedPen);
     else                      SelectObject(hdc,ghDeselectedPen);
     MoveToEx(hdc,sh1,sv1,NULL); LineTo(hdc,sh1+1,sv1);
     if(v2->status == SELECTED)SelectObject(hdc,ghSelectedPen);
     else                      SelectObject(hdc,ghDeselectedPen);
     MoveToEx(hdc,sh2,sv2,NULL); LineTo(hdc,sh2+1,sv2);
     SelectObject(hdc,holdpen);
   }
   DrawArrowIcons(ghdc_triview_Bitmap[i],i);
   InvalidateRect(ghwnd_triview[i],NULL,FALSE);
   UpdateWindow(ghwnd_triview[i]);
 }
}

void DrawModel(void){
 int i,j,n,k;
 RECT rc;
 HCURSOR lhcurSave;
 if(View == TRIVIEW)n=3;
 else               n=1;
 if(sktool == NO || tool != SKANIMATE)lhcurSave=SetCursor(ghcurWait);
 k=ActiveView;
 for(j=0;j<n;j++){
   i=order_list[k][j];
   if(ghdc_triview_Bitmap[i] == NULL)return;
   if(ghbm_triview[i] == NULL)return;
   GetClientRect(ghwnd_triview[i],&rc);
   PatBlt(ghdc_triview_Bitmap[i],0,0,rc.right,rc.bottom,PATCOPY);
   if(sktool == YES && tool == SKANIMATE){
     DrawTrSkeletonInOne(ghdc_triview_Bitmap[i],i);
   }
   else {
     if(/* grid_on && */draw_grid_on)DrawGridInOne(ghdc_triview_Bitmap[i],i);
     SelectObject(ghdc_triview_Bitmap[i],ghEdgePen);
     if(!global_quickdraw){
       DrawAllEdgesInOne(ghdc_triview_Bitmap[i],i);
       DrawArrowIcons(ghdc_triview_Bitmap[i],i);
       InvalidateRect(ghwnd_triview[i],NULL,FALSE);
       UpdateWindow(ghwnd_triview[i]);
       if(zoom_abort)break;
     }
     DrawAllVerticesInOne(ghdc_triview_Bitmap[i],i);
     DrawAllNurbsInOne(ghdc_triview_Bitmap[i],i,FALSE);
   }
   DrawArrowIcons(ghdc_triview_Bitmap[i],i);
   // inserted to draw brush in background windows
   if(SelectedBrush >= 0)DrawBrushInOne(ghdc_triview_Bitmap[i],i,SelectedBrush);
   //
   InvalidateRect(ghwnd_triview[i],NULL,FALSE);
   UpdateWindow(ghwnd_triview[i]);
   if(zoom_abort)break;
 }
 if(sktool == NO || tool != SKANIMATE)SetCursor(lhcurSave);
 if(zoom_abort > 0){
   if     (zoom_abort ==  1)zoom_abort=IDM_PAN_LEFT;
   else if(zoom_abort ==  2)zoom_abort=IDM_PAN_RIGHT;
   else if(zoom_abort ==  3)zoom_abort=IDM_PAN_UP;
   else if(zoom_abort ==  4)zoom_abort=IDM_PAN_DOWN;
   else if(zoom_abort ==  5)zoom_abort=IDM_WINDOW_SWITCHWINDOWS;
   else if(zoom_abort ==  6)zoom_abort=IDM_WINDOW_SWITCHWINDOW0;
   else if(zoom_abort ==  7)zoom_abort=IDM_WINDOW_SWITCHWINDOW1;
   else if(zoom_abort ==  8)zoom_abort=IDM_ZOOM_IN;
   else if(zoom_abort ==  9)zoom_abort=IDM_ZOOM_OUT;
   else if(zoom_abort == 10)zoom_abort=IDM_ACTIONS_ROTATE_ANTICLOCKWISE;
   else if(zoom_abort == 11)zoom_abort=IDM_ACTIONS_ROTATE_CLOCKWISE;
   PostMessage(ghwnd_main,WM_COMMAND,(WPARAM)zoom_abort,(LPARAM)ActiveView);
   zoom_abort=NO;
 }
 return;
}

static void DrawAllVerticesInOne(HDC hdc, int view){
 int i,sho,svo;
 vertex *vp;
 HPEN holdpen;
 holdpen=SelectObject(hdc,ghDeselectedPen);
 if(intriview(&ObjectAxis.Offset)){
   GetWindowCoords(view,ObjectAxis.Offset.xyz[0],
      ObjectAxis.Offset.xyz[1],ObjectAxis.Offset.xyz[2],&sho,&svo);
   MoveToEx(hdc,sho-3,svo,NULL); LineTo(hdc,sho+4,svo);
   MoveToEx(hdc,sho,svo-3,NULL); LineTo(hdc,sho,svo+4);
 }
 SelectObject(hdc,ghSelectedPen);
 if(intriview(&ObjectAxis.Origin)){
   GetWindowCoords(view,ObjectAxis.Origin.xyz[0],
      ObjectAxis.Origin.xyz[1],ObjectAxis.Origin.xyz[2],&sho,&svo);
   MoveToEx(hdc,sho-5,svo,NULL); LineTo(hdc,sho+6,svo);
   MoveToEx(hdc,sho,svo-5,NULL); LineTo(hdc,sho,svo+6);
 }
 if(Preferences.bigv){
   SelectObject(hdc,ghDeselectedPen);
   vp=MainVp; if(MainVp != NULL && NvertDeselect > 0)for(i=0;i<Nvert;i++){
    if(vp->status == DESELECTED)if(intriview(vp)){
      GetWindowCoords(view,vp->xyz[0],vp->xyz[1],vp->xyz[2],&sho,&svo);
//      MoveToEx(hdc,sho-1,svo-1,NULL); LineTo(hdc,sho+2,svo-1);
      MoveToEx(hdc,sho-1,svo,NULL); LineTo(hdc,sho+2,svo);
      MoveToEx(hdc,sho,svo-1,NULL); LineTo(hdc,sho,svo+2);
//      MoveToEx(hdc,sho-1,svo+1,NULL); LineTo(hdc,sho+2,svo+1);
      if(HIWORD(GetQueueStatus(QS_MOUSEBUTTON)) == QS_MOUSEBUTTON
         &&  CheckInterrupt())goto ABORT;
    }
    vp++;
   }
   SelectObject(hdc,ghSelectedPen);
   vp=MainVp; if(MainVp != NULL && NvertSelect > 0)for(i=0;i<Nvert;i++){
    if(vp->status == SELECTED)if(intriview(vp)){
      GetWindowCoords(view,vp->xyz[0],vp->xyz[1],vp->xyz[2],&sho,&svo);
//      MoveToEx(hdc,sho-1,svo-1,NULL); LineTo(hdc,sho+2,svo-1);
      MoveToEx(hdc,sho-1,svo,NULL); LineTo(hdc,sho+2,svo);
      MoveToEx(hdc,sho,svo-1,NULL); LineTo(hdc,sho,svo+2);
//      MoveToEx(hdc,sho-1,svo+1,NULL); LineTo(hdc,sho+2,svo+1);
      if(HIWORD(GetQueueStatus(QS_MOUSEBUTTON)) == QS_MOUSEBUTTON
         &&  CheckInterrupt())goto ABORT;
    }
    vp++;
   }
 }
 else{
   SelectObject(hdc,ghDeselectedPen);
   vp=MainVp; if(MainVp != NULL && NvertDeselect > 0)for(i=0;i<Nvert;i++){
    if(vp->status == DESELECTED)if(intriview(vp)){
      GetWindowCoords(view,vp->xyz[0],vp->xyz[1],vp->xyz[2],&sho,&svo);
      MoveToEx(hdc,sho,svo,NULL); LineTo(hdc,sho+1,svo);
      if(HIWORD(GetQueueStatus(QS_MOUSEBUTTON)) == QS_MOUSEBUTTON
         &&  CheckInterrupt())goto ABORT;
    }
    vp++;
   }
   SelectObject(hdc,ghSelectedPen);
   vp=MainVp; if(MainVp != NULL && NvertSelect > 0)for(i=0;i<Nvert;i++){
    if(vp->status == SELECTED)if(intriview(vp)){
      GetWindowCoords(view,vp->xyz[0],vp->xyz[1],vp->xyz[2],&sho,&svo);
      MoveToEx(hdc,sho,svo,NULL); LineTo(hdc,sho+1,svo);
      if(HIWORD(GetQueueStatus(QS_MOUSEBUTTON)) == QS_MOUSEBUTTON
         &&  CheckInterrupt())goto ABORT;
    }
    vp++;
   }
 }
 ABORT:
 SelectObject(hdc,holdpen);
 return;
}

static void DrawOneVertexInOne(vertex *Vp, HDC hdc, int view){
 int sho,svo;
 HPEN holdpen;
 if(intriview(Vp)){
   if(Vp->status == DESELECTED){
     holdpen=SelectObject(hdc,ghDeselectedPen);
   }
   else{
     holdpen=SelectObject(hdc,ghSelectedPen);
   }
   GetWindowCoords(view,Vp->xyz[0],Vp->xyz[1],Vp->xyz[2],&sho,&svo);
   if(Preferences.bigv){
//     MoveToEx(hdc,sho-1,svo-1,NULL); LineTo(hdc,sho+2,svo-1);
     MoveToEx(hdc,sho-1,svo,NULL); LineTo(hdc,sho+2,svo);
     MoveToEx(hdc,sho,svo-1,NULL); LineTo(hdc,sho,svo+2);
//     MoveToEx(hdc,sho-1,svo+1,NULL); LineTo(hdc,sho+2,svo+1);
   }
   else MoveToEx(hdc,sho,svo,NULL); LineTo(hdc,sho+1,svo);
 }
 return;
}

static short in_modeller_view1(vertex *vp1, vertex *vp2){
 if(vp1->status == HIDDEN && vp2->status == HIDDEN)return 0;
 if( ((vp1->xyz[0] > TVpointX) && (vp1->xyz[0] < TVpointX+TVsizeX))
  && ((vp1->xyz[1] > TVpointY) && (vp1->xyz[1] < TVpointY+TVsizeY))
  && ((vp1->xyz[2] > TVpointZ) && (vp1->xyz[2] < TVpointZ+TVsizeZ))
   )return 1;
 if( ((vp2->xyz[0] > TVpointX) && (vp2->xyz[0] < TVpointX+TVsizeX))
  && ((vp2->xyz[1] > TVpointY) && (vp2->xyz[1] < TVpointY+TVsizeY))
  && ((vp2->xyz[2] > TVpointZ) && (vp2->xyz[2] < TVpointZ+TVsizeZ))
   )return 1;
 return 0;
}

static short in_modeller_view2(vertex *vp1, vertex *vp2){
 if(vp1->status == HIDDEN || vp2->status == HIDDEN)return 0;
 if( ((vp1->xyz[0] > TVpointX) && (vp1->xyz[0] < TVpointX+TVsizeX))
  && ((vp1->xyz[1] > TVpointY) && (vp1->xyz[1] < TVpointY+TVsizeY))
  && ((vp1->xyz[2] > TVpointZ) && (vp1->xyz[2] < TVpointZ+TVsizeZ))
   )return 1;
 if( ((vp2->xyz[0] > TVpointX) && (vp2->xyz[0] < TVpointX+TVsizeX))
  && ((vp2->xyz[1] > TVpointY) && (vp2->xyz[1] < TVpointY+TVsizeY))
  && ((vp2->xyz[2] > TVpointZ) && (vp2->xyz[2] < TVpointZ+TVsizeZ))
   )return 1;
 return 0;
}

static void DrawAllEdgesInOne(HDC hdc, int view){
 int i,sh1,sv1,sh2,sv2;
 edge   *ep;
 vertex *v1,*v2;
 short (*in_modeller_view)(vertex *vp1, vertex *vp2);
 LPPOINT lppt;
 LPBYTE  lpbTypes;
 int     cCount;
 long    pcount;
 if(Nedge == 0)return;
#ifdef _LINE_DRAW
 if((lppt=(LPPOINT)LocalAlloc(LMEM_FIXED,2*Nedge*sizeof(POINT))) == NULL)return;
 if((lpbTypes=(LPBYTE)LocalAlloc(LMEM_FIXED,2*Nedge*sizeof(BYTE))) == NULL){
  LocalFree((HLOCAL)lppt);
  return;
 }
 cCount=0; pcount=0;
#endif
 if(!draw_joins)in_modeller_view = in_modeller_view2; /* scroll lock   */
 else           in_modeller_view = in_modeller_view1; /* test function */
 ep=MainEp;
 for(i=0;i<Nedge;i++){
  v1=(MainVp+ep->V[0]);
  v2=(MainVp+ep->V[1]);
  if(in_modeller_view(v1,v2)){
    GetWindowCoords(view,v1->xyz[0],v1->xyz[1],v1->xyz[2],&sh1,&sv1);
    GetWindowCoords(view,v2->xyz[0],v2->xyz[1],v2->xyz[2],&sh2,&sv2);
    if((!Preferences.quick_trick || abs(sh1-sh2) > 3 || abs(sv1-sv2) > 3) ){
#ifdef _LINE_DRAW
      lppt[cCount].x=sh1; lppt[cCount].y=sv1;
      lpbTypes[cCount]=PT_MOVETO; cCount++;
      lppt[cCount].x=sh2; lppt[cCount].y=sv2;
      lpbTypes[cCount]=PT_LINETO; cCount++;
#else
      MoveToEx(hdc,sh1,sv1,NULL); LineTo(hdc,sh2,sv2);
#endif
    }
    if(HIWORD(GetQueueStatus(QS_MOUSEBUTTON)) == QS_MOUSEBUTTON
       &&  CheckInterrupt())goto ABORT;
  }
  ep++;
 }
 ABORT:;
#ifdef _LINE_DRAW
 if(PolyDraw(hdc,lppt,lpbTypes,cCount) == FALSE){
   char message[128];
   sprintf(message,"ERROR: pcount=%ld ccount=%ld",pcount,cCount);
   MessageBox (NULL,message, NULL,
   MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
 }
 LocalFree((HLOCAL)lppt);
 LocalFree((HLOCAL)lpbTypes);
#endif
}

void Draw3dView(int smab,int skelview){
 RECT rc;
 HCURSOR lhcurSave;
 double xt,yt,zt,dr;
 double tt1[4][4],tt2[4][4],tt3[4][4];
 short i,j;
 for(i=0;i<4;i++) for(j=0;j<4;j++){
   trr[i][j]=0.0; if(i == j)trr[i][j]=1;
 }
 GetClientRect(ghwnd_view,&rc);
 scale=266.0*(double)rc.right/320.0/(double)TVsizeX;
 rad=(double)TVsizeX*10;
 dr=PI/180.0;
 xt=(double)(TVpointX+TVsizeX/2);
 yt=(double)(TVpointY+TVsizeY/2);
 zt=(double)(TVpointZ+TVsizeZ/2);
 tram(tt1,-xt,-yt,-zt);
 if(type_of_view){
   rotz(tt2,-PerspFi*dr);  /* to rotate to the like a camera */
   m4by4(tt2,tt1,tt3);
//   rotx(tt1,PerspTheta*dr);
   rotx(tt1,-PerspTheta*dr);
 }
 else{
   rotz(tt2,PerspFi*dr);
   m4by4(tt2,tt1,tt3);
//   rotx(tt1,-PerspTheta*dr);  /this is the way DOS does it ????
   rotx(tt1,PerspTheta*dr);
 }
 m4by4(tt1,tt3,tt2);
 tram(tt1,0.0,rad,0.0);
 m4by4(tt1,tt2,trr);
 if(ghdc_view_Bitmap == NULL || ghbm_view == NULL)return;
 PatBlt(ghdc_view_Bitmap,0,0,rc.right,rc.bottom,PATCOPY);
 SelectObject(ghdc_view_Bitmap,ghWireframePen);
 SelectPalette(ghdc_view_Bitmap,ghpaletteScreen,FALSE);
 RealizePalette(ghdc_view_Bitmap);
 if(smab){
   lhcurSave=SetCursor(ghcurWait);
   DrawHiddenLine(ghdc_view_Bitmap);
   SetCursor(lhcurSave);
 }
 else{
   DrawWireFrame(ghdc_view_Bitmap,skelview);
   DrawNurbsWireFrame(ghdc_view_Bitmap,skelview,scale,trr,
                      in_modeller_view1,in_modeller_view2,FALSE);
 }
 InvalidateRect(ghwnd_view,NULL,FALSE);
 return;
}

static void DrawWireFrame(HDC hdc,int skelview){
 edge *ep;
 vertex *v1,*v2,*vop;
 double x1,y1,z1,x2,y2,z2;
 int i,ix1,ix2,iy1,iy2,flag,line_flag=0,
     pwinl,pwinr,pwint,pwinb,pwincx,pwincy;
 short (*in_modeller_view)(vertex *vp1, vertex *vp2);
 MSG msg;
 RECT rc;
 LPPOINT lppt;
 LPBYTE  lpbTypes;
 int     cCount;
 long    pcount;
 if(Nedge == 0)return;
#ifdef _LINE_DRAW
 if(tool != SKANIMATE || skelview == 1)line_flag=1;
 if(line_flag){
   if((lppt=(LPPOINT)LocalAlloc(LMEM_FIXED,2*Nedge*sizeof(POINT))) == NULL)return;
   if((lpbTypes=(LPBYTE)LocalAlloc(LMEM_FIXED,2*Nedge*sizeof(BYTE))) == NULL){
     LocalFree((HLOCAL)lppt);
     return;
   }
 }
 cCount=0; pcount=0;
#endif
 if(!draw_joins)in_modeller_view = in_modeller_view2; /* scroll lock   */
 else           in_modeller_view = in_modeller_view1; /* test function */
 GetClientRect(ghwnd_view,&rc);
 pwinl=pwint=0; pwinr=rc.right; pwinb=rc.bottom;
 pwincx=rc.right/2; pwincy=rc.bottom/2;
 if(tool != SKANIMATE){ /* draw wireframe as is */
   ep=MainEp; for(i=0;i<Nedge;i++){
    v1=(MainVp+ep->V[0]);
    v2=(MainVp+ep->V[1]);
    if(in_modeller_view(v1,v2)){
      flag=0;
      m4by1(trr,(double)v1->xyz[0],(double)v1->xyz[1],(double)v1->xyz[2]
               ,&x1,&y1,&z1);
      m4by1(trr,(double)v2->xyz[0],(double)v2->xyz[1],(double)v2->xyz[2]
               ,&x2,&y2,&z2);
      if(y1 >= 0.0){
        ix1=(int)(pwincx+scale*x1);
        iy1=(int)(pwincy-scale*z1);
      }
      else  flag=1;
      if(y2 >= 0.0){
        ix2=(int)(pwincx+scale*x2);
        iy2=(int)(pwincy-scale*z2);
      }
      else flag=1;
      if(flag == 0 && (!Preferences.quick_trick || abs(ix1-ix2) > 2 || abs(iy1-iy2) > 2) ){
#ifdef _LINE_DRAW
        lppt[cCount].x=ix1; lppt[cCount].y=iy1;
        lpbTypes[cCount]=PT_MOVETO; cCount++;
        lppt[cCount].x=ix2; lppt[cCount].y=iy2;
        lpbTypes[cCount]=PT_LINETO; cCount++;
#else
        MoveToEx(hdc,ix1,iy1,NULL);
        LineTo(hdc,ix2,iy2);
#endif
      }
    }
    ep++;
   }
   vop=&vobserver;           /* draw the current cursor point in 3d */
   vop->status=SELECTED;
   vop->xyz[0]=NpointerX; vop->xyz[1]=NpointerY; vop->xyz[2]=NpointerZ;
   if(intriview(vop)){
#if 0
     m4by1(trr,(double)vop->xyz[0],(double)vop->xyz[1],(double)vop->xyz[2]
              ,&x1,&y1,&z1);
     if(y1 >= 0.0){
       HPEN holdpen=SelectObject(hdc,ghEdgePen);
       ix1=(int)(pwincx+scale*x1);
       iy1=(int)(pwincy-scale*z1);
       MoveToEx(hdc,ix1-5,iy1,NULL); LineTo(hdc,ix1+5,iy1);
       MoveToEx(hdc,ix1,iy1-5,NULL); LineTo(hdc,ix1,iy1+5);
       SelectObject(hdc,holdpen);
     }
#endif
     {long d,dx,dy,dz;
      HPEN holdpen=SelectObject(hdc,ghEdgePen);
      d=TVsizeX/30;
      for(i=0;i<3;i++){
        if     (i == 0){dx=d; dy=dz=0;}
        else if(i == 1){dy=d; dx=dz=0;}
        else           {dz=d; dx=dy=0;}
        m4by1(trr,(double)(vop->xyz[0]-dx),
                  (double)(vop->xyz[1]-dy),
                  (double)(vop->xyz[2]-dz),
                  &x1,&y1,&z1);
        m4by1(trr,(double)(vop->xyz[0]+dx),
                  (double)(vop->xyz[1]+dy),
                  (double)(vop->xyz[2]+dz),
                  &x2,&y2,&z2);
        if(y1 >= 0.0 || y2 >= 0){
          ix1=(int)(pwincx+scale*x1);
          iy1=(int)(pwincy-scale*z1);
          ix2=(int)(pwincx+scale*x2);
          iy2=(int)(pwincy-scale*z2);
          MoveToEx(hdc,ix1,iy1,NULL); LineTo(hdc,ix2,iy2);
        }
      }
      SelectObject(hdc,holdpen);
     }
   } /* end of cursor draw */
 }
 else{ /* draw using skeletal transformation */
   if(skelview == 0){
     static short ep1[12]={0,1,2,3,4,5,6,7,0,1,2,3};
     static short ep2[12]={1,2,3,0,5,6,7,4,4,5,6,7};
     int i;
     skel *sp;
     if((sp=MainSp) != NULL)while(sp != NULL){
       if(skintriview(sp->tp) || (sp->at != NULL && skintriview(sp->at->tp))){
         flag=0;
         m4by4(trr,sp->t,t1);
         for(i=0;i<12;i++){
           m4by1(t1 ,(double)sp->bx[ep1[i]][0]
                    ,(double)sp->bx[ep1[i]][1]
                    ,(double)sp->bx[ep1[i]][2]
                    ,&x1,&y1,&z1);
           m4by1(t1 ,(double)sp->bx[ep2[i]][0]
                    ,(double)sp->bx[ep2[i]][1]
                    ,(double)sp->bx[ep2[i]][2]
                    ,&x2,&y2,&z2);
           if(y1 >= 0.0){
             ix1=(int)(pwincx+scale*x1);
             iy1=(int)(pwincy-scale*z1);
           }
           else  flag=1;
           if(y2 > 0.0){
             ix2=(int)(pwincx+scale*x2);
             iy2=(int)(pwincy-scale*z2);
           }
           else flag=1;
           if(flag == 0){
             MoveToEx(hdc,ix1,iy1,NULL);
             LineTo(hdc,ix2,iy2);
           }
         }
       }
       sp=sp->last;
     }
   }
   else {
     ep=MainEp; for(i=0;i<Nedge;i++){
       v1=(MainVp+ep->V[0]);
       v2=(MainVp+ep->V[1]);
       if(in_modeller_view(v1,v2)){
         flag=0;
         if(v1->sp != NULL)m4by4(trr,v1->sp->t,t1);
         else              c4to4(trr,t1);
         m4by1(t1 ,(double)v1->xyz[0],(double)v1->xyz[1],(double)v1->xyz[2]
                  ,&x1,&y1,&z1);
         if(v2->sp != NULL)m4by4(trr,v2->sp->t,t1);
         else              c4to4(trr,t1);
          m4by1(t1 ,(double)v2->xyz[0],(double)v2->xyz[1],(double)v2->xyz[2]
                  ,&x2,&y2,&z2);
         if(y1 >= 0.0){
           ix1=(int)(pwincx+scale*x1);
           iy1=(int)(pwincy-scale*z1);
         }
         else  flag=1;
         if(y2 > 0.0){
           ix2=(int)(pwincx+scale*x2);
           iy2=(int)(pwincy-scale*z2);
         }
         else flag=1;
         if(flag == 0 && (!Preferences.quick_trick || abs(ix1-ix2) > 2 || abs(iy1-iy2) > 2) ){
#ifdef _LINE_DRAW
           lppt[cCount].x=ix1; lppt[cCount].y=iy1;
           lpbTypes[cCount]=PT_MOVETO; cCount++;
           lppt[cCount].x=ix2; lppt[cCount].y=iy2;
           lpbTypes[cCount]=PT_LINETO; cCount++;
#else
           MoveToEx(hdc,ix1,iy1,NULL);
           LineTo(hdc,ix2,iy2);
#endif
         }
       }
       ep++;
     }
   }
 }
#ifdef _LINE_DRAW
 if(line_flag){
   if(PolyDraw(hdc,lppt,lpbTypes,cCount) == FALSE){
     char message[128];
     sprintf(message,"ERROR: pcount=%ld ccount=%ld",pcount,cCount);
     MessageBox (NULL,message, NULL,
     MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
   }
   LocalFree((HLOCAL)lppt);
   LocalFree((HLOCAL)lpbTypes);
 }
#endif
 return;
}

void DrawQuick3dView(int skelview){
 RECT rc;
 HCURSOR lhcurSave;
 HDC hdc;
 double xt,yt,zt,dr;
 double tt1[4][4],tt2[4][4],tt3[4][4];
 short i,j;
 for(i=0;i<4;i++) for(j=0;j<4;j++){
   trr[i][j]=0.0; if(i == j)trr[i][j]=1;
 }
 GetClientRect(ghwnd_view,&rc);
 scale=266.0*(double)rc.right/320.0/(double)TVsizeX;
 rad=(double)TVsizeX*10;
 dr=PI/180.0;
 xt=(double)(TVpointX+TVsizeX/2);
 yt=(double)(TVpointY+TVsizeY/2);
 zt=(double)(TVpointZ+TVsizeZ/2);
 tram(tt1,-xt,-yt,-zt);
 if(type_of_view){
   rotz(tt2,-PerspFi*dr);  /* to rotate to the like a camera */
   m4by4(tt2,tt1,tt3);
   rotx(tt1,-PerspTheta*dr);
 }
 else{
   rotz(tt2,PerspFi*dr);
   m4by4(tt2,tt1,tt3);
   rotx(tt1,PerspTheta*dr);
 }
 m4by4(tt1,tt3,tt2);
 tram(tt1,0.0,rad,0.0);
 m4by4(tt1,tt2,trr);
 hdc=GetDC(ghwnd_view);
 SelectPalette(hdc,ghpaletteScreen,FALSE);
 RealizePalette (hdc);
 SelectObject(hdc,ghbrushWindow);
 PatBlt(hdc,0,0,rc.right,rc.bottom,PATCOPY);
 SelectObject(hdc,ghWireframePen);
 DrawQuickWireFrame(hdc,skelview);
 DrawNurbsWireFrame(hdc,skelview,scale,trr,
                    in_modeller_view1,in_modeller_view2,TRUE);
 ReleaseDC(ghwnd_view,hdc);
 return;
}

static void DrawQuickWireFrame(HDC hdc,int skelview){
 edge *ep;
 vertex *v1,*v2,*vop;
 double x1,y1,z1,x2,y2,z2;
 int i,ix1,ix2,iy1,iy2,flag,line_flag=0,j,nstep,
     pwinl,pwinr,pwint,pwinb,pwincx,pwincy;
 short (*in_modeller_view)(vertex *vp1, vertex *vp2);
 MSG msg;
 RECT rc;
 LPPOINT lppt;
 LPBYTE  lpbTypes;
 int     cCount;
 long    pcount;
 if(Nedge == 0)return;
 if(!draw_joins)in_modeller_view = in_modeller_view2; /* scroll lock   */
 else           in_modeller_view = in_modeller_view1; /* test function */
 GetClientRect(ghwnd_view,&rc);
 pwinl=pwint=0; pwinr=rc.right; pwinb=rc.bottom;
 pwincx=rc.right/2; pwincy=rc.bottom/2;
 if(tool != SKANIMATE){ /* draw wireframe as is */
   flag=0;
   nstep=5;
   for(v1=MainVp,i=0;i<Nvert;i+=nstep,v1+=nstep){
     if(intriview(v1)){
       m4by1(trr,(double)v1->xyz[0],(double)v1->xyz[1],(double)v1->xyz[2]
                ,&x1,&y1,&z1);
       if(y1 >= 0.0){
         ix1=(int)(pwincx+scale*x1);
         iy1=(int)(pwincy-scale*z1);
         MoveToEx(hdc,ix1-1,iy1,NULL);  LineTo(hdc,ix1+2,iy1);
         MoveToEx(hdc,ix1,iy1-1,NULL);  LineTo(hdc,ix1,iy1+2);
       }
       if(HIWORD(GetQueueStatus(QS_MOUSEMOVE)) == QS_MOUSEMOVE)return;
     }
   }
//   if(flag)return;
   ep=MainEp; for(i=0;i<Nedge;i++){
    v1=(MainVp+ep->V[0]);
    v2=(MainVp+ep->V[1]);
    if(in_modeller_view(v1,v2)){
      flag=0;
      m4by1(trr,(double)v1->xyz[0],(double)v1->xyz[1],(double)v1->xyz[2]
               ,&x1,&y1,&z1);
      m4by1(trr,(double)v2->xyz[0],(double)v2->xyz[1],(double)v2->xyz[2]
               ,&x2,&y2,&z2);
      if(y1 >= 0.0){
        ix1=(int)(pwincx+scale*x1);
        iy1=(int)(pwincy-scale*z1);
      }
      else  flag=1;
      if(y2 >= 0.0){
        ix2=(int)(pwincx+scale*x2);
        iy2=(int)(pwincy-scale*z2);
      }
      else flag=1;
      if(flag == 0 && (!Preferences.quick_trick || abs(ix1-ix2) > 2 || abs(iy1-iy2) > 2) ){
        MoveToEx(hdc,ix1,iy1,NULL);
        LineTo(hdc,ix2,iy2);
      }
      if(HIWORD(GetQueueStatus(QS_MOUSEMOVE)) == QS_MOUSEMOVE)break;
    }
    ep++;
   }
 }
 else{ /* draw using skeletal transformation */
   if(skelview == 0){
     static short ep1[12]={0,1,2,3,4,5,6,7,0,1,2,3};
     static short ep2[12]={1,2,3,0,5,6,7,4,4,5,6,7};
     int i;
     skel *sp;
     if((sp=MainSp) != NULL)while(sp != NULL){
       if(skintriview(sp->tp) || (sp->at != NULL && skintriview(sp->at->tp))){
         flag=0;
         m4by4(trr,sp->t,t1);
         for(i=0;i<12;i++){
           m4by1(t1 ,(double)sp->bx[ep1[i]][0]
                    ,(double)sp->bx[ep1[i]][1]
                    ,(double)sp->bx[ep1[i]][2]
                    ,&x1,&y1,&z1);
           m4by1(t1 ,(double)sp->bx[ep2[i]][0]
                    ,(double)sp->bx[ep2[i]][1]
                    ,(double)sp->bx[ep2[i]][2]
                    ,&x2,&y2,&z2);
           if(y1 >= 0.0){
             ix1=(int)(pwincx+scale*x1);
             iy1=(int)(pwincy-scale*z1);
           }
           else  flag=1;
           if(y2 > 0.0){
             ix2=(int)(pwincx+scale*x2);
             iy2=(int)(pwincy-scale*z2);
           }
           else flag=1;
           if(flag == 0){
             MoveToEx(hdc,ix1,iy1,NULL);
             LineTo(hdc,ix2,iy2);
           }
         }
       }
       sp=sp->last;
     }
   }
   else {
     ep=MainEp; for(i=0;i<Nedge;i++){
       v1=(MainVp+ep->V[0]);
       v2=(MainVp+ep->V[1]);
       if(in_modeller_view(v1,v2)){
         flag=0;
         if(v1->sp != NULL)m4by4(trr,v1->sp->t,t1);
         else              c4to4(trr,t1);
         m4by1(t1 ,(double)v1->xyz[0],(double)v1->xyz[1],(double)v1->xyz[2]
                  ,&x1,&y1,&z1);
         if(v2->sp != NULL)m4by4(trr,v2->sp->t,t1);
         else              c4to4(trr,t1);
          m4by1(t1 ,(double)v2->xyz[0],(double)v2->xyz[1],(double)v2->xyz[2]
                  ,&x2,&y2,&z2);
         if(y1 >= 0.0){
           ix1=(int)(pwincx+scale*x1);
           iy1=(int)(pwincy-scale*z1);
         }
         else  flag=1;
         if(y2 > 0.0){
           ix2=(int)(pwincx+scale*x2);
           iy2=(int)(pwincy-scale*z2);
         }
         else flag=1;
         if(flag == 0 && (!Preferences.quick_trick || abs(ix1-ix2) > 2 || abs(iy1-iy2) > 2) ){
           MoveToEx(hdc,ix1,iy1,NULL);
           LineTo(hdc,ix2,iy2);
         }
         if(HIWORD(GetQueueStatus(QS_MOUSEMOVE)) == QS_MOUSEMOVE)break;
       }
       ep++;
     }
   }
 }
 return;
}

int O_normalize(vector v){
 double n,nn;
 n= (double)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return FAIL;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE(n,v,v)
 return OK;
}

// Hidden line drawing code

#include "hidden.h"

#define HRES 800

static long    Zbuffer[HRES];
static double  Zdepth [HRES];
static double  DL[HRES],DR[HRES],D1[HRES];

static object  Object;
static long    XMIN,
               XMAX,
               YMIN,
               YMAX,
               Xcentre,
               Ycentre;

static void ModNormUpdate(rvertex *v, rface *f, eface *e);
static void ModRenderHide(RECT rc, HDC hdc, int left_point, int top_point);
static void ModUpdateScanBuffers(double xl, double xr, long id,
                       rface  *f, eface  *e, double d2);
static long IntersectZ(long scanline, rvertex  *v1, rvertex  *v2,
                double *xi);
static long ActiveZ(long scanline, rvertex  *v, rface  *f,
                      double *leftedge, double *rightedge);

static void DrawHiddenLine(HDC hdc){
 face      *fp;
 vertex    *vp,*V0,*V1,*V2;
 double   x1,y1,z1,x2,y2,z2;
 long     sv,mm,sf,se,me,i,j,k,FirstNurbsVertex;
 int      ix1,ix2,iy1,iy2,flag,pwinl,pwinr,pwint,pwinb,pwincx,pwincy;
 rvertex  *v;
 rface    *f;
 eface    *e;
 RECT rc;
 if((Nvert == 0 || Nface == 0) && Nnurbs == 0)return;
 GetClientRect(ghwnd_view,&rc);
 pwinl=pwint=0; pwinr=rc.right; pwinb=rc.bottom;
 pwincx=rc.right/2; pwincy=rc.bottom/2;
 XMIN=0; XMAX=(pwinr-pwinl);
 YMIN=0; YMAX=(pwinb-pwint);
 Xcentre = XMAX/2; Ycentre=YMAX/2;
 Object.NoVertices=0;
 Object.NoFaces=0;
 vp=MainVp; if(MainVp != NULL && Nvert > 0)for(i=0;i<Nvert;i++){
   vp->id = 65535;
   vp++;
 }
 fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++){
   V0=(MainVp+fp->V[0]);
   V1=(MainVp+fp->V[1]);
   V2=(MainVp+fp->V[2]);
   if(intriview(V0))V0->id=1;
   if(intriview(V1))V1->id=1;
   if(intriview(V2))V2->id=1;
   fp++;
 }
 fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++){
   V0=(MainVp+fp->V[0]);
   V1=(MainVp+fp->V[1]);
   V2=(MainVp+fp->V[2]);
   if(V0->id == 1 || V1->id == 1 || V2->id == 1 ){
     if(V0->id != 1)V0->id = 2;
     if(V1->id != 1)V1->id = 2;
     if(V2->id != 1)V2->id = 2;
   }
   fp++;
 }
 Object.NoVertices=0;
 vp=MainVp; if(MainVp != NULL && Nvert > 0)for(i=0;i<Nvert;i++){
   if(vp->id == 1 || vp->id == 2){Object.NoVertices++; vp->id = 1;}
   vp++;
 }
 if(Object.NoVertices > 65535){
   SendPrgmQuery(IDQ_NOHIDDEN,0);
   goto NOTHING;
 }
 Object.NoFaces=0;
 fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++){
   V0=(MainVp+fp->V[0]);
   V1=(MainVp+fp->V[1]);
   V2=(MainVp+fp->V[2]);
   if(V0->id == 1 && V1->id == 1 && V2->id == 1 ){
     Object.NoFaces++;
   }
   fp++;
 }
 //update number of vertices and faces for NURBS
 HiddenGetNurbsNumbers(&Object.NoVertices,&Object.NoFaces);
 if(Object.NoVertices == 0 || Object.NoFaces == 0)goto NOTHING;
 sv=sizeof(rvertex);
 mm=Object.NoVertices*(long)sv;
 Object.Vbase=(rvertex  *)X__Malloc(mm);
 if(Object.Vbase == NULL){
   SendPrgmQuery(IDQ_NOHIDDEN,0);
   goto NOTHING;
 }
 sf=sizeof(rface);
 mm=Object.NoFaces*(long)sf;
 se=sizeof(eface);
 me=Object.NoFaces*(long)se;
 Object.Fbase=(rface  *)X__Malloc(mm);
 if(Object.Fbase == NULL){
   SendPrgmQuery(IDQ_NOHIDDEN,0);
   X__Free(Object.Vbase);  goto NOTHING;
 }
 Object.Ebase=(eface  *)X__Malloc(me);
 if(Object.Ebase == NULL){
  SendPrgmQuery(IDQ_NOHIDDEN,0);
  X__Free(Object.Vbase); X__Free(Object.Fbase); goto NOTHING;
 }
 v=Object.Vbase;
 f=Object.Fbase;
 e=Object.Ebase;
 i=0;
 if(1){
// if(tool != SKANIMATE){ /* draw wireframe as is */
   vp=MainVp; if(Nvert > 0)for(k=0;k<Nvert;k++){
    if(vp->id == 1){
      m4by1(trr,(double)vp->xyz[0],(double)vp->xyz[1],(double)vp->xyz[2]
               ,&x1,&y1,&z1);
      v[i].p[0]=x1; v[i].p[1]=y1; v[i].p[2]=z1;
      if(y1 >= 0.0){
        v[i].x=Xcentre+scale*x1;
        v[i].y=Ycentre-scale*z1;
      }
      else {v[i].x = 0; v[i].y= 0;}
      vp->id= i;
      i++;
      if(i > Object.NoVertices){
        SendPrgmQuery(IDQ_HIDDEN1,0);
        goto FREEALL;
      }
    }
    vp++;
   }
   //add vertices for NURBS
   FirstNurbsVertex=i;
   HiddenAddNurbsVertices(&i,Object.Vbase,scale,trr,XMAX,YMAX,Xcentre,Ycentre);
 }
 else{ /* draw using skeletal transformation */
   vp=MainVp; for(k=0;k<Nvert;k++){
    if(vp->id == 1){
      if(vp->sp != NULL)m4by4(trr,vp->sp->t,t1);
      else              c4to4(trr,t1);
      m4by1(t1 ,(double)vp->xyz[0],(double)vp->xyz[1],(double)vp->xyz[2]
               ,&x1,&y1,&z1);
      v[i].p[0]=x1; v[i].p[1]=y1; v[i].p[2]=z1;
      if(y1 >= 0.0){
        v[i].x=Xcentre+scale*x1;
        v[i].y=Ycentre-scale*z1;
      }
      else {v[i].x = 0; v[i].y= 0;}
      vp->id= i;
      i++;
    }
    vp++;
   }
   //add vertices for NURBS
   FirstNurbsVertex=i;
   HiddenAddNurbsVertices(&i,Object.Vbase,scale,t1,XMAX,YMAX,Xcentre,Ycentre);
 }
 if(i != Object.NoVertices){
   SendPrgmQuery(IDQ_HIDDEN2,0);
   goto FREEALL;
 }
 f=Object.Fbase;
 e=Object.Ebase;
 i=0;
 fp=MainFp; for(k=0;k<Nface;k++){
   V0=(MainVp+fp->V[0]);
   V1=(MainVp+fp->V[1]);
   V2=(MainVp+fp->V[2]);
   if(V0->id < 65535 && V1->id < 65535 && V2->id < 65535){
     if(i >= Object.NoFaces){
       SendPrgmQuery(IDQ_HIDDEN3,0);
       goto FREEALL;
     }
     f[i].V[0]=V0->id;
     f[i].V[1]=V1->id;
     f[i].V[2]=V2->id;
     i++;
   }
   fp++;
 }
 //add Nurbs Faces
 HiddenAddNurbsFaces(&i,Object.Fbase,FirstNurbsVertex);
 if(i != Object.NoFaces){
   SendPrgmQuery(IDQ_HIDDEN4,0);
   goto FREEALL;
 }
 ModNormUpdate(v,f,e);
 ModRenderHide(rc,hdc,pwinl,pwint);
 FREEALL:
 X__Free(Object.Vbase);
 X__Free(Object.Fbase);
 X__Free(Object.Ebase);
 NOTHING:
 return;
}

static void ModNormUpdate(rvertex *v, rface *f, eface *e){
   long i,bad;
   double c,ftop,fbottom,fleft,fright;
   vector p1,p2,p0;
   for(i=0;i<Object.NoFaces;i++)
   {
    p0[0]=(v+f->V[0])->p[0];
    p0[1]=(v+f->V[0])->p[1];
    p0[2]=(v+f->V[0])->p[2];
    p1[0]=(v+f->V[1])->p[0];
    p1[1]=(v+f->V[1])->p[1];
    p1[2]=(v+f->V[1])->p[2];
    p2[0]=(v+f->V[2])->p[0];
    p2[1]=(v+f->V[2])->p[1];
    p2[2]=(v+f->V[2])->p[2];
    VECSUB(p2,p0,p2)
    VECSUB(p1,p0,p1)
    CROSS(p1,p2,e->n)
    if(O_normalize(e->n) == FAIL)bad=1;
    else                         bad=0;
    c=DOT(p0,e->n);
    e->c=(double)c;
      if(c > 0.0){
        VECSCALE((double)(-1.0),e->n,e->n)
        e->c=(double)c*(-1.0);
    }
    ftop=min((v+(f->V[0]))->y,(v+(f->V[1]))->y);
    ftop=min((v+(f->V[2]))->y,ftop);
    fbottom=max((v+(f->V[0]))->y,(v+(f->V[1]))->y);
    fbottom=max((v+(f->V[2]))->y,fbottom)+1;
    fleft=min((v+(f->V[0]))->x,(v+(f->V[1]))->x);
    fleft=min((v+(f->V[2]))->x,fleft);
    fright=max((v+(f->V[0]))->x,(v+(f->V[1]))->x);
    fright=max((v+(f->V[2]))->x,fright)+1;
    if(ftop < 0.0)f->top= -1;
    else if(ftop > YMAX)f->top=YMAX+1;
    else f->top=(int)ftop;
    if(fbottom < 0.0)f->bottom= -1;
    else if(fbottom > YMAX)f->bottom=YMAX+1;
    else f->bottom=(int)fbottom;
    if(fleft < 0.0)f->left= -1;
    else if(fleft > XMAX)f->left=XMAX+1;
    else f->left=(int)fleft;
    if(fright < 0.0)f->right= -1;
    else if(fright > XMAX)f->right=XMAX+1;
    else f->right=(int)fright;
    if( ( v[ f->V[0] ].p[1] <= 0)  /* nodes in face are either behind */
     && ( v[ f->V[1] ].p[1] <= 0)  /* or on viewing plane */
     && ( v[ f->V[2] ].p[1] <= 0)){
      f->top=YMAX+1;
    }
    if(f->left >= XMAX){
      f->top=YMAX+1;
    }
    if(f->right < XMIN){
      f->top=YMAX+1;
    }
    if(fabs(e->c) < 1.e-5){
      f->top=YMAX+1;
    }
    if(bad)f->top=YMAX+1;
    f++; e++;
   }
}

static void ModRenderHide(RECT rc, HDC hdc, int left_point, int top_point){
 long i,j,jj,k;
 long Zold[HRES];
 double d2,xl,xr;
 rvertex  *v;
 rface    *f;
 eface   *e;
 HPEN holdpen;
// LOGPEN lopn;
// GetObject(ghEdgePen,sizeof(LOGPEN),&lopn);
 holdpen=SelectObject(hdc,ghEdgePen);
 rc.bottom=rc.top;
 for(i=XMIN;i<XMAX;i++) {
   DL[i]=(double)(i - Xcentre);        DL[i]/=scale;
   DR[i]=(double)(i+1-Xcentre);        DR[i]/=scale;
   D1[i]=(double)(i - Xcentre) + 0.5;  D1[i]/=scale;
   Zold[i]= -1;
 }
 for(k=YMIN;k<YMAX;k++){
   d2=(double)(Ycentre-k);
   d2/=scale;
   for(i=XMIN;i<XMAX;i++){
     Zbuffer[i] = -1;
     Zdepth [i] = FARAWAY;
   }
   v=Object.Vbase;
   f=Object.Fbase;
   e=Object.Ebase;
   for(j=0;j<Object.NoFaces;j++)
     if( ActiveZ(k,v,f+j,&xl,&xr) ){
       ModUpdateScanBuffers(xl,xr,j,f+j,e+j,d2);
     }
   for(j=XMIN+1;j<XMAX;j++){
     if((Zbuffer[j] != Zbuffer[j-1])
        || (Zold[j] != Zbuffer[j])) {
//        X__SetPixel(hdc,j+left_point,k+top_point,lopn.lopnColor);
        MoveToEx(hdc,j+left_point,k+top_point,NULL);
        LineTo(hdc,j+left_point+1,k+top_point);
      }
      Zold[j]=Zbuffer[j];
   }
   if(GetAsyncKeyState(VK_ESCAPE) & 0x8000)break;
   rc.bottom++;
   InvalidateRect(ghwnd_view,&rc,FALSE);
   rc.top++;
   UpdateWindow(ghwnd_view);
 }
 SelectObject(hdc,holdpen);
 return;
}

static void ModUpdateScanBuffers(double xl, double xr, long id,
                       rface  *f, eface  *e, double d2){
 long i,ixl,ixr;
 double en1,depth,bL,bR,aL,aR,dL,dR,dl,dr,dd;
 ixl = (long)xl;
 ixr = (long)xr;
 if(ixl < 0)ixl=0;
 if(ixr < 0)ixr=0;
 if(ixl >= XMAX)ixl=XMAX-1;
 if(ixr >= XMAX)ixr=XMAX-1;
 if(fabs(e->n[1]) < 1.e-30)en1=1.e-30;
 else                      en1=e->n[1];
 if(ixl == ixr){
   dL = xl - (double)Xcentre; dL/=scale;
   dR = xr - (double)Xcentre; dR/=scale;
   aL = (e->c - e->n[0]*dL - e->n[2]*d2)/en1;
   aR = (e->c - e->n[0]*dR - e->n[2]*d2)/en1;
   depth=min(aL,aR);
   if(depth > 0 && depth <= Zdepth[ixl]){
     Zdepth[ixl]=depth; Zbuffer[ixl]=id;
   }
   return;
 }
 else{
   dl = (e->c - e->n[0]*DR[ixl] - e->n[2]*d2)/en1;
   dr = (e->c - e->n[0]*DL[ixr] - e->n[2]*d2)/en1;
   dd = (dr-dl)/(xr-xl);
   for(i=0;i<ixr-ixl;i++){
     depth = dl + i*dd;
     if(depth > 0 && depth <= Zdepth[i+ixl]){
       Zdepth[i+ixl]=depth; Zbuffer[i+ixl]=id;
     }
   }
 }
}

static long IntersectZ(long scanline, rvertex  *v1, rvertex  *v2,
                double *xi){
 double ddx,ddy,doff,ddxy;
 if((v1->y < scanline)  && (v2->y < scanline))return 0;
 if((v1->y > scanline)  && (v2->y > scanline))return 0;
 ddy = (v2->y) - (v1->y);
 if(fabs(ddy) < 0.0001)return 0;
 ddx = (v2->x) - (v1->x);
 doff=(double)scanline - v1->y;
 ddxy=ddx*doff/ddy;
 *xi=(v1->x)+ddxy;
 return 1;
}

static long ActiveZ(long scanline, rvertex  *v, rface  *f,
                      double *leftedge, double *rightedge){
 long v1,v2,v3
     ,i1,i2,i3;
 double x1,x2,x3;
 if(f->top > scanline)       return 0;
 if(f->bottom < scanline)    return 0;
 v1=f->V[0];
 v2=f->V[1];
 v3=f->V[2];
 i1=IntersectZ(scanline,(v+v1),(v+v2),&x1);
 i2=IntersectZ(scanline,(v+v2),(v+v3),&x2);
 i3=IntersectZ(scanline,(v+v3),(v+v1),&x3);
 if((i1 == 1) && (i2 == 1) && (i3 == 1))
 {*leftedge = min(min(x1,x2),x3);
  *rightedge= max(max(x1,x2),x3);}
 else if((i2 == 1) && (i3 == 1))
 {*leftedge = min(x2,x3);
  *rightedge= max(x2,x3);}
 else if((i3 == 1) && (i1 == 1))
 {*leftedge = min(x3,x1);
  *rightedge= max(x3,x1);}
 else if((i1 == 1) && (i2 == 1))
 {*leftedge = min(x1,x2);
  *rightedge= max(x1,x2);}
 else return 0;
 if(*leftedge >= XMAX)return 0;
 else *leftedge  = max(0,*leftedge);
 if(*rightedge < XMIN)return 0;
 *rightedge = min(XMAX-0.01,*rightedge);
 return 1;
}

// End of hidden line drawing code

void DrawRubberBoundBox(long xmin, long ymin, long zmin,
                        long xmax, long ymax, long zmax){
 int i,j,n;
 int x1,y1,x2,y2;
 RECT rc;
 HDC hDC;
 if(View == TRIVIEW)n=3;
 else               n=1;
 for(j=0;j<n;j++){
   i=order_list[ActiveView][j];
   hDC=GetDC(ghwnd_triview[i]);
   SelectObject(hDC,ghInvertPen);
   SetROP2(hDC,R2_XORPEN);
   SelectPalette(hDC,ghpaletteScreen,FALSE);
   RealizePalette(hDC);
   GetWindowCoords(i,xmin,ymin,zmin,&x1,&y1);
   GetWindowCoords(i,xmax,ymax,zmax,&x2,&y2);
   if(x2 == x1)x2++; if(y2 == y1)y2++;
   DrawOne3dCursor(hDC,i);
   MoveToEx(hDC,x1,y1,NULL);
   LineTo(hDC,x2,y1); LineTo(hDC,x2,y2); LineTo(hDC,x1,y2); LineTo(hDC,x1,y1);
   DrawOne3dCursor(hDC,i);
   ReleaseDC(ghwnd_triview[i],hDC);
 }
}

void DrawRubber3dLine(point p1,point p2){
 int i,j,n;
 int x1,y1,x2,y2;
 RECT rc;
 HDC hDC;
 if(View == TRIVIEW)n=3;
 else               n=1;
 for(j=0;j<n;j++){
   i=order_list[ActiveView][j];
   hDC=GetDC(ghwnd_triview[i]);
   SelectObject(hDC,ghInvertPen);
   SetROP2(hDC,R2_XORPEN);
   SelectPalette(hDC,ghpaletteScreen,FALSE);
   RealizePalette(hDC);
   GetWindowCoords(i,p1[0],p1[1],p1[2],&x1,&y1);
   GetWindowCoords(i,p2[0],p2[1],p2[2],&x2,&y2);
   if(x2 == x1)x2++; if(y2 == y1)y2++;
   DrawOne3dCursor(hDC,i);
   MoveToEx(hDC,x2,y2,NULL); LineTo(hDC,x1,y1);
   DrawOne3dCursor(hDC,i);
   ReleaseDC(ghwnd_triview[i],hDC);
 }
}

void DrawRubberLines(HWND hwnd, int av, long vtx, int x, int y){
 static long last_V,last_x,last_y,last_z;
 long xx,yy,zz;
 edge *ep;
 vertex *this_V;
 int i,j,n,k;
 int x1,y1,x2,y2,xo,yo;
 RECT rc;
 HDC hDC;
 if(Nedge == 0)return;
 if(vtx >= 0)last_V=vtx;
 if(View == TRIVIEW)n=3;
 else               n=1;
 if(ActiveView == TRIFRONT){
   xx=NpointerX; yy=(MainVp+last_V)->xyz[1]; zz=NpointerZ;
 }
 else if(ActiveView == TRIRIGHT){
   xx=(MainVp+last_V)->xyz[0]; yy=NpointerY; zz=NpointerZ;
 }
 else{
   xx=NpointerX; yy=NpointerY; zz=(MainVp+last_V)->xyz[2];
 }
 for(j=0;j<n;j++){
   i=order_list[ActiveView][j];
   hDC=GetDC(ghwnd_triview[i]);
   SelectObject(hDC,ghInvertPen);
   SetROP2(hDC,R2_XORPEN);
   SelectPalette(hDC,ghpaletteScreen,FALSE);
   RealizePalette(hDC);
   DrawOne3dCursor(hDC,i);
   if(vtx < 0){
     GetWindowCoords(i,last_x,last_y,last_z,&xo,&yo);
   }
   GetWindowCoords(i,xx,yy,zz,&x1,&y1);
   for(k=0;k<Nedge;k++){
     ep=(MainEp+k);
     if     (ep->V[0] == last_V)this_V=(MainVp+ep->V[1]);
     else if(ep->V[1] == last_V)this_V=(MainVp+ep->V[0]);
     else continue;
     if(this_V->status == HIDDEN)continue;
     GetWindowCoords(i,this_V->xyz[0],this_V->xyz[1],this_V->xyz[2],&x2,&y2);
     if(vtx < 0){
       MoveToEx(hDC,xo,yo,NULL); LineTo(hDC,x2,y2);
     }
     if(vtx > -2){
       MoveToEx(hDC,x1,y1,NULL); LineTo(hDC,x2,y2);
     }
   }
   DrawOne3dCursor(hDC,i);
   ReleaseDC(ghwnd_triview[i],hDC);
 }
 last_x=xx; last_y=yy; last_z=zz;
 if(vtx == -2){
   (MainVp+last_V)->xyz[0]=xx;
   (MainVp+last_V)->xyz[1]=yy;
   (MainVp+last_V)->xyz[2]=zz;
 }
}

void DrawBrush(int brushID){
 HDC hdc;
 int i,j,n,k;
 if(View == TRIVIEW)n=3;
 else               n=1;
 k=ActiveView;
 for(j=0;j<n;j++){
   i=order_list[k][j];
//   to draw the brush in the background windows
//   hdc=GetDC(ghwnd_triview[i]);
//   DrawBrushInOne(hdc,i,SelectedBrush);
   DrawBrushInOne(ghdc_triview_Bitmap[i],i,SelectedBrush);
   InvalidateRect(ghwnd_triview[i],NULL,FALSE);
   //   ReleaseDC(ghwnd_triview[i],hdc);
 }
}

void DrawShader(int shaderID){
 HDC hdc;
 int i,j,n,k;
 if(View == TRIVIEW)n=3;
 else               n=1;
 k=ActiveView;
 for(j=0;j<n;j++){
   i=order_list[k][j];
   hdc=GetDC(ghwnd_triview[i]);
   DrawShaderInOne(hdc,i,shaderID);
   ReleaseDC(ghwnd_triview[i],hdc);
 }
}

void DrawShaperSphere(long x,long y,long z,
                      double shaper_angle_theta,
                      double shaper_angle_phi,
                      double shaper_size){
 vector shaper_direction={0.0,0.0,1.0};
 int   x1,y1,x2,y2,x3,y3,i,j,n,radius;
 long  Px,Py,Pz;
 double shaper_scale;
 RECT rc;
 HDC hDC;
 shaper_direction[2]=cos(shaper_angle_theta);
 shaper_direction[0]=cos(shaper_angle_phi)*sin(shaper_angle_theta);
 shaper_direction[1]=sin(shaper_angle_phi)*sin(shaper_angle_theta);
 shaper_scale=(double)TVsizeX/shaper_size*1.2;
 Px=x+(long)(shaper_direction[0]*shaper_scale);
 Py=y+(long)(shaper_direction[1]*shaper_scale);
 Pz=z+(long)(shaper_direction[2]*shaper_scale);
 if(View == TRIVIEW)n=3;
 else               n=1;
 for(j=0;j<n;j++){
   i=order_list[ActiveView][j];
   hDC=GetDC(ghwnd_triview[i]);
   SelectObject(hDC,ghInvertPen);
   SelectObject(hDC,GetStockObject(HOLLOW_BRUSH));
   SetROP2(hDC,R2_XORPEN);
   SelectPalette(hDC,ghpaletteScreen,FALSE);
   RealizePalette(hDC);
   GetWindowCoords(i,x,y,z,&x1,&y1);
   GetWindowCoords(i,Px,Py,Pz,&x3,&y3);
   GetWindowCoords(i,x+(long)(TVsizeX/shaper_size),
                     y+(long)(TVsizeY/shaper_size),
                     z+(long)(TVsizeZ/shaper_size),&x2,&y2);
   if(x2 == x1)x2++; if(y2 == y1)y2++;
   DrawOne3dCursor(hDC,i);
//   MoveToEx(hDC,x1,y1,NULL); LineTo(hDC,x3,y3);  // SHAPER
   radius=abs(x2-x1);
   if(radius > 0)Ellipse(hDC,x1-radius,y1-radius,
                             x1+radius,y1+radius);
   DrawOne3dCursor(hDC,i);
   ReleaseDC(ghwnd_triview[i],hDC);
 }
}

void DrawSkeleton(void){
 HDC hdc;
 int i,j,n,k;
 if(View == TRIVIEW)n=3;
 else               n=1;
 k=ActiveView;
 for(j=0;j<n;j++){
   i=order_list[k][j];
   hdc=GetDC(ghwnd_triview[i]);
   DrawSkeletonInOne(hdc,i);
   ReleaseDC(ghwnd_triview[i],hdc);
 }
}

#if 0
void DrawQuickModel(void){
 int v;
 RECT rc;
 v=ActiveView;
 if(ghdc_triview_Bitmap[v] == NULL)return;
 if(ghbm_triview[v] == NULL)return;
 GetClientRect(ghwnd_triview[v],&rc);
 PatBlt(ghdc_triview_Bitmap[v],0,0,rc.right,rc.bottom,PATCOPY);
 if(sktool == YES && tool == SKANIMATE){
   DrawTrSkeletonInOne(ghdc_triview_Bitmap[v],v);
 }
 else {
   int i,j,sho,svo,sh1,sv1,sh2,sv2,k=0;
   edge   *ep;
   vertex *vp,*v1,*v2;
   HDC hdc;
   HPEN holdpen;
   short (*in_modeller_view)(vertex *vp1, vertex *vp2);
   hdc=ghdc_triview_Bitmap[v];
   holdpen=SelectObject(hdc,ghEdgePen);
   if(Nedge > 1){
     in_modeller_view = in_modeller_view1; /* test function */
     for(j=0;j<2;j++){
     ep=MainEp; if(j == 1)ep++; k=0;
     for(i=j;i<Nedge;i+=2){
       v1=(MainVp+ep->V[0]);
       v2=(MainVp+ep->V[1]);
       if(in_modeller_view(v1,v2)){
         GetWindowCoords(v,v1->xyz[0],v1->xyz[1],v1->xyz[2],&sh1,&sv1);
         GetWindowCoords(v,v2->xyz[0],v2->xyz[1],v2->xyz[2],&sh2,&sv2);
         if((!Preferences.quick_trick || abs(sh1-sh2) > 3 || abs(sv1-sv2) > 3) ){
           MoveToEx(hdc,sh1,sv1,NULL); LineTo(hdc,sh2,sv2); k++;
         }
       }
       if(HIWORD(GetQueueStatus(QS_MOUSEMOVE)) == QS_MOUSEMOVE && k > 75){
         update_timer=SetTimer(ghwnd_main,1,33,NULL);
         goto ABORT;
       }
       ep+=2;
     }
     }
   }
   DrawArrowIcons(ghdc_triview_Bitmap[v],v);
   InvalidateRect(ghwnd_triview[v],NULL,FALSE);
   UpdateWindow(ghwnd_triview[v]);
   SelectObject(hdc,ghDeselectedPen);
   if(intriview(&ObjectAxis.Offset)){
     GetWindowCoords(v,ObjectAxis.Offset.xyz[0],
        ObjectAxis.Offset.xyz[1],ObjectAxis.Offset.xyz[2],&sho,&svo);
     MoveToEx(hdc,sho-3,svo,NULL); LineTo(hdc,sho+4,svo);
     MoveToEx(hdc,sho,svo-3,NULL); LineTo(hdc,sho,svo+4);
   }
   SelectObject(hdc,ghSelectedPen);
   if(intriview(&ObjectAxis.Origin)){
     GetWindowCoords(v,ObjectAxis.Origin.xyz[0],
      ObjectAxis.Origin.xyz[1],ObjectAxis.Origin.xyz[2],&sho,&svo);
     MoveToEx(hdc,sho-5,svo,NULL); LineTo(hdc,sho+6,svo);
     MoveToEx(hdc,sho,svo-5,NULL); LineTo(hdc,sho,svo+6);
   }
   SelectObject(hdc,ghSelectedPen);
   vp=MainVp; if(MainVp != NULL)for(i=0;i<Nvert;i++){
    if(vp->status != HIDDEN)if(intriview(vp)){
      GetWindowCoords(v,vp->xyz[0],vp->xyz[1],vp->xyz[2],&sho,&svo);
      MoveToEx(hdc,sho,svo,NULL); LineTo(hdc,sho+1,svo);
      if(HIWORD(GetQueueStatus(QS_MOUSEMOVE)) == QS_MOUSEMOVE){
        if(update_timer == 0){
          update_timer=SetTimer(ghwnd_main,1,33,NULL);
        }
        goto ABORT;
      }
    }
    vp++;
   }
   ABORT:
   SelectObject(hdc,holdpen);
 }
 DrawArrowIcons(ghdc_triview_Bitmap[v],v);
 InvalidateRect(ghwnd_triview[v],NULL,FALSE);
 UpdateWindow(ghwnd_triview[v]);
 return;
}
#endif

void DrawQuickModel(void){
 int v;
 RECT rc;
 HDC hdc;
 HPALETTE hpalT;
 HBRUSH hOldBrush;
 v=ActiveView;
 hdc=GetDC(ghwnd_triview[v]);
 hpalT = SelectPalette(hdc,ghpaletteScreen,FALSE);
 RealizePalette (hdc);
 hOldBrush = SelectObject(hdc,ghbrushWindow);
 GetClientRect(ghwnd_triview[v],&rc);
 PatBlt(hdc,0,0,rc.right,rc.bottom,PATCOPY);
 if(sktool == YES && tool == SKANIMATE){
   DrawTrSkeletonInOne(hdc,v);
 }
 else {
   int i,j,sho,svo,sh1,sv1,sh2,sv2,k=0,Nstep;
   edge   *ep;
   vertex *vp,*v1,*v2;
   HPEN holdpen;
   short (*in_modeller_view)(vertex *vp1, vertex *vp2);
   if(draw_grid_on)DrawGridInOne(hdc,v);
   DrawArrowIcons(hdc,v);
   holdpen=SelectObject(hdc,ghDeselectedPen);
   if(intriview(&ObjectAxis.Offset)){
     GetWindowCoords(v,ObjectAxis.Offset.xyz[0],
        ObjectAxis.Offset.xyz[1],ObjectAxis.Offset.xyz[2],&sho,&svo);
     MoveToEx(hdc,sho-3,svo,NULL); LineTo(hdc,sho+4,svo);
     MoveToEx(hdc,sho,svo-3,NULL); LineTo(hdc,sho,svo+4);
   }
   SelectObject(hdc,ghSelectedPen);
   if(intriview(&ObjectAxis.Origin)){
     GetWindowCoords(v,ObjectAxis.Origin.xyz[0],
      ObjectAxis.Origin.xyz[1],ObjectAxis.Origin.xyz[2],&sho,&svo);
     MoveToEx(hdc,sho-5,svo,NULL); LineTo(hdc,sho+6,svo);
     MoveToEx(hdc,sho,svo-5,NULL); LineTo(hdc,sho,svo+6);
   }
   SelectObject(hdc,ghSelectedPen);
   Nstep=4;
   vp=MainVp; if(MainVp != NULL)for(i=0;i<Nvert;i+=Nstep){
    if(vp->status != HIDDEN)if(intriview(vp)){
      GetWindowCoords(v,vp->xyz[0],vp->xyz[1],vp->xyz[2],&sho,&svo);
      MoveToEx(hdc,sho,svo,NULL); LineTo(hdc,sho+1,svo);
      if(HIWORD(GetQueueStatus(QS_MOUSEMOVE)) == QS_MOUSEMOVE){
        goto ABORT;
      }
      if(HIWORD(GetQueueStatus(QS_MOUSEBUTTON)) == QS_MOUSEBUTTON)goto ABORT;
    }
    vp+=Nstep;
   }
   SelectObject(hdc,ghEdgePen);
   if(Nedge > 1){
     in_modeller_view = in_modeller_view1; /* test function */
     for(j=0;j<2;j++){
     ep=MainEp; if(j == 1)ep++; k=0;
     for(i=j;i<Nedge;i+=2){
       v1=(MainVp+ep->V[0]);
       v2=(MainVp+ep->V[1]);
       if(in_modeller_view(v1,v2)){
         GetWindowCoords(v,v1->xyz[0],v1->xyz[1],v1->xyz[2],&sh1,&sv1);
         GetWindowCoords(v,v2->xyz[0],v2->xyz[1],v2->xyz[2],&sh2,&sv2);
         if((!Preferences.quick_trick || abs(sh1-sh2) > 3 || abs(sv1-sv2) > 3) ){
           MoveToEx(hdc,sh1,sv1,NULL); LineTo(hdc,sh2,sv2); k++;
         }
       }
       if(HIWORD(GetQueueStatus(QS_MOUSEMOVE)) == QS_MOUSEMOVE && k > 75){
         goto ABORT;
       }
       if(HIWORD(GetQueueStatus(QS_MOUSEBUTTON)) == QS_MOUSEBUTTON)goto ABORT;
       ep+=2;
     }
     }
   }
   DrawAllNurbsInOne(hdc,v,TRUE);
   ABORT:
   SelectObject(hdc,holdpen);
 }
 DrawArrowIcons(hdc,v);
 SelectObject (hdc, hOldBrush);
 SelectPalette(hdc,hpalT,FALSE);
 ReleaseDC(ghwnd_triview[v],hdc);
 return;
}
