/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file dxcon.c */

#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>

#include "struct.h"
#include "dstruct.h"
#include "dxfcon.h"

int  (*X__pri)(const char *,...);
int  X__local_print(const char *,...);

HWND      hWndParent;
HINSTANCE hInstance;
HWND dlg=NULL,hwndProgress=NULL;

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BCC__
int APIENTRY DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
      hInstance=hDLL;
      break;
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     hwndProgress=GetDlgItem(hwnd,DLG_PROGRESS);
     SendMessage(hwndProgress,PBM_SETRANGE,0,MAKELPARAM(0,99));
     SendMessage(hwndProgress,PBM_SETSTEP,(WPARAM)1,0);
     SendMessage(hwndProgress,PBM_SETPOS,(WPARAM)0,0);
     CentreDlgOnS(hwnd);
     break;
   default: break;
 }
 return( FALSE );
}

BOOL CALLBACK DlgProcSettings(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[32];
 static struct bp {
   short h_type,l_freeze,b_freeze,no_weld,d2,d3,cs;
   double wt;
 } *dp;
 switch( msg ) {
   case WM_INITDIALOG:
     dp=(struct bp *)lparam;
     if(dp->h_type == 0)SendDlgItemMessage(hwnd,DLG_SET_HX,BM_SETCHECK,TRUE,0);
     if(dp->h_type == 1)SendDlgItemMessage(hwnd,DLG_SET_HL,BM_SETCHECK,TRUE,0);
     if(dp->h_type == 2)SendDlgItemMessage(hwnd,DLG_SET_HB,BM_SETCHECK,TRUE,0);
     if(dp->l_freeze == 1)SendDlgItemMessage(hwnd,DLG_SET_FL,BM_SETCHECK,TRUE,0);
     if(dp->no_weld == 0)SendDlgItemMessage(hwnd,DLG_SET_NW,BM_SETCHECK,TRUE,0);
     if(dp->d2 == 1)SendDlgItemMessage(hwnd,DLG_SET_D2,BM_SETCHECK,TRUE,0);
     sprintf(str,"%.4lf",dp->wt);
     SetDlgItemText(hwnd,DLG_SET_WT,(LPCTSTR)str);
     sprintf(str,"%ld",dp->cs);
     SetDlgItemText(hwnd,DLG_SET_CS,(LPCTSTR)str);
     CentreDlgOnS(hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(SendDlgItemMessage(hwnd,DLG_SET_HX,BM_GETCHECK,0,0))
            dp->h_type=0;
          else if(SendDlgItemMessage(hwnd,DLG_SET_HL,BM_GETCHECK,0,0))
            dp->h_type=1;
          else if(SendDlgItemMessage(hwnd,DLG_SET_HB,BM_GETCHECK,0,0))
            dp->h_type=2;
          if(SendDlgItemMessage(hwnd,DLG_SET_FL,BM_GETCHECK,0,0))
               dp->l_freeze=1;
          else dp->l_freeze=0;
          if(SendDlgItemMessage(hwnd,DLG_SET_NW,BM_GETCHECK,0,0))
               dp->no_weld=0;
          else dp->no_weld=1;
          if(SendDlgItemMessage(hwnd,DLG_SET_D2,BM_GETCHECK,0,0))
               dp->d2=1;
          else dp->d2=0;
          if(GetDlgItemText(hwnd,DLG_SET_WT,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->wt=atof(str);
          if(GetDlgItemText(hwnd,DLG_SET_CS,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->cs=atoi(str);
          if(dp->cs < 3)dp->cs=3;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default: break;
      }
   default: break;
 }
 return( FALSE );
}

BOOL CALLBACK DlgProcScale(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[32];
 static double *scale;
 switch( msg ) {
   case WM_INITDIALOG:
     scale = (double *)lparam;
     sprintf(str,"%.3le",*scale);
     SetDlgItemText(hwnd,DLG_SCALE_SCALE,(LPCTSTR)str);
     CentreDlgOnS(hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_SCALE_SCALE,str,31) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          sscanf(str,"%lf",scale);
          if(*scale < 1.e-10)*scale=1.e-10;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default: break;
      }
   default: break;
 }
 return( FALSE );
}


long _Convert(
     HWND parent,
     char *iname,
     char *oname,
     X__STRUCTURE *lpevi
                              ){
 long i,r;
 double wt = 0.0001;
 short h_type=0,l_freeze=0,b_freeze=0,no_weld=0,d2=0,
       d3=1,cs=15;
 struct bp {
   short h_type,l_freeze,b_freeze,no_weld,d2,d3,cs;
   double wt;
 } d;
 lpEVI=lpevi;
 hWndParent=parent;
 X__pri=X__local_print;
 d.h_type=h_type;
 d.l_freeze=l_freeze;
 d.b_freeze=b_freeze;
 d.no_weld=no_weld;
 d.d2=d2;
 d.d3=d3;
 d.cs=cs;
 d.wt=wt;
 if(!DialogBoxParam(hInstance,MAKEINTRESOURCE(DLG_SET),hWndParent,
                    (DLGPROC)DlgProcSettings,(LPARAM)&d))return 0;
 dlg=CreateDialog(hInstance,MAKEINTRESOURCE(DLG),parent,(DLGPROC)DlgProc);
 r=(long)DoDxfConvert(iname,oname,
                      d.wt,d.h_type,d.l_freeze,d.b_freeze,d.no_weld,
                      d.d2,d.d3,d.cs);
 if(dlg != NULL)DestroyWindow(dlg); dlg=NULL;
 return r;
}

int X__local_print(const char *format,...){
 char line[80];
 va_list arg_ptr;
 va_start(arg_ptr,format);
 vsprintf(line,format,arg_ptr);
 va_end(arg_ptr);
// WindowedRenderString(line);
 return 1;
}
