/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* DXFCON1.C  */

#include <windows.h>
#include <commctrl.h>
#include <fcntl.h>
#include <io.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys\types.h>
#include <sys\stat.h>

#include "dxfcon.h"

extern int (*X__pri)(const char *,...);
extern HWND      hWndParent;
extern HINSTANCE hInstance;
extern HWND dlg,hwndProgress;
extern BOOL CALLBACK DlgProc(HWND,UINT,WPARAM,LPARAM);
extern BOOL CALLBACK DlgProcScale(HWND,UINT,WPARAM,LPARAM);

#ifndef PI
#define PI 3.1415926
#endif

#ifdef __ZTC__
#ifdef DOS386
#define _far
#endif
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#else
#endif

#define HE  2147483647L
#define _X__FAR
#define E_TOOMANYVERTICES  1
#define G_FILEOPENFAIL     2

typedef double vector[3];
typedef long   point[3];
typedef long   twoLongs[2];

typedef struct VERTEX {
 struct VERTEX  *last;
 struct VERTEX  *next;
 unsigned short  id;
 char   status,pad;
 long   xyz[3];
 struct SKELETON *sp;
} vertex;

typedef struct EDGE {
 struct EDGE  *last;
 struct EDGE  *next;
 struct VERTEX  *V[2];
} edge;

typedef struct FACE{
 struct FACE  *last;
 struct FACE  *next;
 struct VERTEX  *V[3];
 struct VERTEX  *extra,*xxtra;
 unsigned char color[3],matcol[3],texture,brush;
} face;

typedef struct WIREFRAME {
 long Np,Ne;
 twoLongs *e;
 point *p;
} wireframe;

typedef struct SKELETON {
 struct SKELETON *at,*next,*last;
 short  id,offset;
 double t[4][4];
 point  xyz,tp;
 char   name[16];
} skel;

typedef struct {
       vertex Origin;
       long Xdirection[3];
       long Ydirection[3];
       long Zdirection[3];
       vertex Offset;
} axis;

typedef struct {
  char  name[8];
  short Nc,Np,flag;
  vertex **Vc,**Vp;
} bezier_curve;

/* definition of unit dimensions */
#define UNIT         32768L
#define UNIT2        65536L
#define MINUNIT       1024L
#define MAXUNIT    4194304L
#define UNITSCA    (MAXUNIT/16)
/* definition of flags and identifiers*/
#define NO               0
#define YES              1
#define OK               0
#define FAIL             1
#define SELECTED         1
#define DESELECTED       0
#define PLANE            0

short quiet=0;
axis ObjectAxis={NULL,NULL,0,0,
                 0L,0L,0L,
                 0,0L,0L,
                 UNIT,0L,0L,
                 0L,UNIT,0L,
                 0L,0L,0L};
vertex  *MainVp=NULL;
edge    *MainEp=NULL;
face    *MainFp=NULL;
long Nvert=0,NvertSelect=0,NvertDeselect=0,Nface=0,Nedge=0
    ,TVpointX=(-UNIT),TVpointY=(-UNIT),TVpointZ=(-UNIT)
    ,TVsizeX=UNIT2,TVsizeY=UNIT2,TVsizeZ=UNIT2;
short  ShadowFlag=NO;
point  TextureP[64],TextureX[64],TextureY[64];
long   TextureType[64],TextureAnimFF[64],TextureAnimDL[64],TextureAnimLF[64],
       TextureMap[64],TextureAngle[64];
char   *TextureList[64];
long Brilliance=0, Reflectivity=0, Transparency=0;
long EDIT_ACTION,NoTextures=0,Nskel=0,DefaultBrilliance,DefaultTransparency,
     DefaultReflectivity,ruler=1,lrulerx=0,lrulery=0,lrulerz=0,grid_size=UNIT,
     rulerx=0,rulery=0,rulerz=0,N_Bcurves=0;
skel *MainSp=NULL,*FirstSp=NULL,*TextureSk[64];

void CreateEdge(vertex  *v1, vertex  *v2, short check);
void CreateFace(vertex  *v1, vertex  *v2, vertex  *v3, short check);
void CreateVertex(void);
void WriteTextures(void);
void WriteVertices(short );
void WriteEdges(short );
void WriteFaces(short );
void WriteSkeleton(void);
static void WriteWireframe(void);
static void WriteBezierCurve(void);

static FILE *fp,*fo;
static long newface,newedge,newvert,minX,minY,minZ,maxX,maxY,maxZ;
static char str[10],ruler_name[8];
static vertex  **p,*vlast;
static axis TempAxis;
static face *FaceBase;
static bezier_curve *Bcurves=NULL;
static wireframe w_frame={0,0,NULL,NULL};


void CreateEdge(vertex  *v1, vertex  *v2, short check)
{edge  *temp;
 if(check){
 temp=MainEp;
 while(temp != NULL)   /* check to see if this edge already exists */
 {if(   ((temp->V[0] == v1) && (temp->V[1] == v2))
     || ((temp->V[0] == v2) && (temp->V[1] == v1)) )return;
 temp=temp->last;}}
 if(( temp=(edge  *)malloc(sizeof(edge)) ) == NULL)return;
 if(MainEp != NULL)MainEp->next=temp;
 temp->last=MainEp;
 temp->V[0]=v1;
 temp->V[1]=v2;
 temp->next=NULL;
 MainEp=temp;
 Nedge++;
}

void CreateFace(vertex  *v1, vertex  *v2, vertex  *v3, short check)
{face  *temp;
 if(check){
 temp=MainFp;  /* check to see if face exists */
 while(temp != NULL)   /* check to see if this face already exists */
 {if(   ((temp->V[0] == v1) && (temp->V[1] == v2) && (temp->V[2] == v3))
     || ((temp->V[0] == v1) && (temp->V[1] == v3) && (temp->V[2] == v2))
     || ((temp->V[0] == v2) && (temp->V[1] == v3) && (temp->V[2] == v1))
     || ((temp->V[0] == v2) && (temp->V[1] == v1) && (temp->V[2] == v3))
     || ((temp->V[0] == v3) && (temp->V[1] == v1) && (temp->V[2] == v2))
     || ((temp->V[0] == v3) && (temp->V[1] == v2) && (temp->V[2] == v1))
  )return;
 temp=temp->last;}}
 if(( temp=(face  *)malloc(sizeof(face)) ) == NULL)return;
 if(MainFp != NULL)MainFp->next=temp;
 temp->last=MainFp;
 temp->V[0]=v1;
 temp->V[1]=v2;
 temp->V[2]=v3;
 temp->next=NULL;
 temp->color[0]=0;
 temp->color[1]=0;
 temp->color[2]=0;
 temp->matcol[0]=0;
 temp->matcol[1]=0;
 temp->matcol[2]=0;
 temp->texture=0;
 temp->brush=0;
 temp->extra=NULL;
 temp->xxtra=NULL;
 MainFp=temp;
 Nface++;
}

void CreateVertex(void){
 vertex  *temp;
 if(( temp=(vertex  *)malloc(sizeof(vertex)) ) == NULL)return;
 if(MainVp != NULL)MainVp->next=temp;
 temp->last=MainVp;
 temp->xyz[0]=0;
 temp->xyz[1]=0;
 temp->xyz[2]=0;
 temp->next=NULL;
 temp->sp=NULL;
 temp->status=SELECTED;
 temp->id=0;
 MainVp=temp;
 Nvert++;
 NvertSelect++;
}

void CreateSkeleton(skel *at){
 skel _X__FAR *temp;
 int i,j;
 if(( temp=(skel _X__FAR *)malloc(sizeof(skel)) ) == NULL)return;
 if(MainSp != NULL)MainSp->next=temp;
 temp->last=MainSp;
 temp->xyz[0]=0;
 temp->xyz[1]=0;
 temp->xyz[2]=0;
 temp->next=NULL;
 temp->at=at;
 temp->id=0;
 temp->offset=0;
 for(i=0;i<4;i++)
  for(j=0;j<4;j++){
    temp->t[i][j]=0.0;
    if(i == j)temp->t[i][j]=1.0;
  }
 MainSp=temp;
 Nskel++;
 if(Nskel == 1)strcpy(temp->name,"Root");
 else sprintf(temp->name,"%ld",Nskel-1);
}

void check_error(void){
  if(ferror(fo) == 0)return;
  if(quiet)return;
  X__pri("Error Writing model file");
  return;
}

void outsht(short sht){
 if(fputc((char)(sht >>  8),fo) == EOF)check_error();
 if(fputc((char)(sht      ),fo) == EOF)check_error();
}

void outlng(long lng){
 if(fputc((char)(lng >> 24),fo) == EOF)check_error();
 if(fputc((char)(lng >> 16),fo) == EOF)check_error();
 if(fputc((char)(lng >>  8),fo) == EOF)check_error();
 if(fputc((char)(lng      ),fo) == EOF)check_error();
}

short SaveObject(char *Filename, short status){
  long pos1,pos2,pos3,pos4,pos5,pos6,pos7,pos8,pos9,posA,pose;
  long FORMsize;
  short id,short_form=0;
  FORMsize=0; newvert=0; newface=0; newedge=0;
  if( (fo=fopen(Filename,"wb")) == NULL){   /* now open for reading */
    return 1;
  }
  fprintf(fo,"FORM");
  pos1=ftell(fo); outlng(FORMsize);
  fprintf(fo,"OFXM");
  fprintf(fo,"AXIS"); outlng(48L); /* size of data chunk */
  if(status == 2){  /* dont use offset for drawing in triview */
    outlng(0); outlng(0); outlng(0);
  }
  else{
    outlng(ObjectAxis.Origin.xyz[0]);
    outlng(ObjectAxis.Origin.xyz[1]);
    outlng(ObjectAxis.Origin.xyz[2]);
  }
  outlng(TextureP[63][0]);  /* store the algorithmic texture origin */
  outlng(TextureP[63][1]);  /* in the object axis structure         */
  outlng(TextureP[63][2]);
  outlng(TextureX[63][0]);
  outlng(TextureX[63][1]);
  outlng(TextureX[63][2]);
  outlng(TextureY[63][0]);
  outlng(TextureY[63][1]);
  outlng(TextureY[63][2]);
  fprintf(fo,"OFFS"); outlng(16L);      /* follow point */
  outlng(3L);
  outlng(ObjectAxis.Offset.xyz[0]);
  outlng(ObjectAxis.Offset.xyz[1]);
  outlng(ObjectAxis.Offset.xyz[2]);
  if(Nvert > 0){
    fprintf(fo,"VERT"); pos2=ftell(fo); outlng(12*newvert);
    WriteVertices(status);
  }
  else pos2 = -1;
  if(Nedge > 0){
    fprintf(fo,"EDGE"); pos3=ftell(fo); outlng( 8*newedge); WriteEdges(0);
  }
  else pos3 = -1;
  if(Nface > 0){
    fprintf(fo,"FACE"); pos4=ftell(fo); outlng(20*newface); WriteFaces(0);
  }
  else pos4 = -1;
  if(NoTextures > 0){
    fprintf(fo,"TEXT"); pos5=ftell(fo); outlng(0L);
    WriteTextures();
    pos6=ftell(fo);
  }
  else pos5 = -1;         /*! root node only implies no skeleton */
  if((status == 1 || status == 3) &&
      Nskel > 1){/* if saving whole object then save skeleton */
    fprintf(fo,"SK01"); pos7=ftell(fo); outlng(0L);
    WriteSkeleton();
    pos8=ftell(fo);
  }
  else pos7 = -1;
  fprintf(fo,"MATL"); outlng((long)(5*sizeof(long)));
  outlng(4L);
  outlng((long)ShadowFlag);
  outlng(DefaultBrilliance);     /* these are not used any more       */
  outlng(DefaultTransparency);
  outlng(DefaultReflectivity);   /* but remain for file compatibility */
  fprintf(fo,"UNIT"); outlng((long)(7*sizeof(long)+8));
  outlng((long)ruler);
  strcpy(ruler_name,"none");
  fwrite(ruler_name,1,8,fo);
  outlng(lrulerx); outlng(lrulery); outlng(lrulerz);
  outlng(rulerx);  outlng(rulery);  outlng(rulerz);
  fprintf(fo,"GRID"); outlng((long)(2*sizeof(long)));
  outlng(grid_size); outlng(grid_size); /* second is unused */
  WriteWireframe();
  if((status == 1 || status == 3) && N_Bcurves > 0){
     fprintf(fo,"BEZ1"); pos9=ftell(fo); outlng(0L);
     WriteBezierCurve();
     posA=ftell(fo);
  }
  else pos9 = -1;
  pose=ftell(fo);  /* the end of the object */
  if(pos2 > 0){fseek(fo,pos2,0); outlng(12*newvert);}
  if(short_form){
    if(pos3 > 0){fseek(fo,pos3,0); outlng( 4*newedge);}
    if(pos4 > 0){fseek(fo,pos4,0); outlng(14*newface);}
  }
  else{
    if(pos3 > 0){fseek(fo,pos3,0); outlng( 8*newedge);}
    if(pos4 > 0){fseek(fo,pos4,0); outlng(20*newface);}
  }
  if(pos5 > 0){fseek(fo,pos5,0); outlng(pos6-pos5-4);} /* not include count */
  if(pos7 > 0){fseek(fo,pos7,0); outlng(pos8-pos7-4);}
  if(pos9 > 0){fseek(fo,pos9,0); outlng(posA-pos9-4);}
  fseek(fo,pos1,0); outlng(pose-8);
  fclose(fo);
  return 0;
}

void WriteTextures(void){
 long i,temp;
 skel _X__FAR *sp;
 i=0;
 sp=FirstSp; while(sp != NULL){ /* set up sk id */
   sp->id=i; i++;
   sp=sp->next;
 }
 fwrite(&NoTextures,1,1,fo);
 for(i=0;i<NoTextures;i++){
   outlng(256);    /* size of Texture list string */
   fwrite(TextureList[i],1,256,fo);
   outlng(TextureP[i][0]); outlng(TextureP[i][1]); outlng(TextureP[i][2]);
   outlng(TextureX[i][0]); outlng(TextureX[i][1]); outlng(TextureX[i][2]);
   outlng(TextureY[i][0]); outlng(TextureY[i][1]); outlng(TextureY[i][2]);
   temp=TextureType[i]; outlng(temp);
   temp=TextureAnimFF[i]; outlng(temp);
   temp=TextureAnimDL[i]; outlng(temp);
   temp=TextureAnimLF[i]; outlng(temp);
   if(TextureSk[i] != NULL){
     temp=((TextureSk[i])->id)+1; outlng(temp);/* 1 is first file brush*/
   }
   else{
     temp=0; outlng(temp);
   }
   temp=TextureMap[i]; outlng(temp);   /* 0 = plane 1 = cyl. map  2/3 = moz */
   temp=1;    outlng(temp);   /* for expansion  use as count */
   temp=TextureAngle[i]; outlng(temp);  /* angle for cylinder brushes */
 }
}

void WriteVertices(short status){
 long i;
 vertex _X__FAR *vp,*fv;
 fv=NULL; vp=MainVp; while(vp != NULL){fv=vp; vp=vp->last;}
 i=0;
 vp=fv; while(vp != NULL){
  if(vp->status == SELECTED){ /* only selected vertices */
    outlng(vp->xyz[0]); outlng(vp->xyz[1]); outlng(vp->xyz[2]);
    vp->pad = (char)(i >> 16);
    vp->id=(unsigned short)(i & 0x0000ffff);
    i++; newvert++;
  }
  else  vp->pad = -1;
  vp=vp->next;
 }
}

void WriteEdges(short short_form){
 long id;
 edge _X__FAR *ep;
 ep=MainEp;
 while(ep != NULL){
   if((ep->V[0]->pad >= 0) && (ep->V[1]->pad >= 0)){
     id=(((long)ep->V[0]->id) | ((long)ep->V[0]->pad << 16)); outlng(id);
     id=(((long)ep->V[1]->id) | ((long)ep->V[1]->pad << 16)); outlng(id);
     newedge++;
   }
   ep=ep->last;
 }
}

void WriteFaces(short short_form){
 face _X__FAR *f,*ff;
 long id;
 ff=NULL; f=MainFp; while(f != NULL){ff=f; f=f->last;}
 if((f=ff) != NULL)while(f != NULL){
   if((f->V[0]->pad >= 0) &&
      (f->V[1]->pad >= 0) &&
      (f->V[2]->pad >= 0)){
     id=(((long)f->V[0]->id) | ((long)f->V[0]->pad << 16)); outlng(id);
     id=(((long)f->V[1]->id) | ((long)f->V[1]->pad << 16)); outlng(id);
     id=(((long)f->V[2]->id) | ((long)f->V[2]->pad << 16)); outlng(id);
     fwrite(f->color,1,3,fo);
     fwrite(&f->texture,1,1,fo);
     fwrite(f->matcol,1,3,fo);
     fwrite(&f->brush,1,1,fo);
     newface++;
   }
   f=f->next;
 }
}

static void WriteBezierCurve(void){
 long i,j,id;
 outlng(N_Bcurves);
 for(i=0;i<N_Bcurves;i++){
   outlng((long)Bcurves[i].Nc); outlng((long)Bcurves[i].Np);
   fwrite((char *)Bcurves[i].name,1,8,fo);
   for(j=0;j<Bcurves[i].Nc;j++){
     id=( (long)((Bcurves[i].Vc[j])->id) |
          ((long)((Bcurves[i].Vc[j])->pad) << 16));
     outlng(id);
   }
   for(j=0;j<Bcurves[i].Np;j++){
     id=((long)((Bcurves[i].Vp[j])->id) |
        ((long)((Bcurves[i].Vp[j])->pad) << 16));
     outlng(id);
   }
 }
}

void WriteSkeleton(void){
 short i;
 skel _X__FAR *sp;
 vertex _X__FAR *vp, *fv;
 i=0;
 sp=FirstSp; while(sp != NULL){
   sp->id=i; i++;
   sp=sp->next;
 }
 outlng(Nskel);
 sp=FirstSp; while(sp != NULL){
   outlng(sp->xyz[0]); outlng(sp->xyz[1]); outlng(sp->xyz[2]);
   if(sp->at != NULL)outsht(sp->at->id);
   else              outsht(0);
   fwrite(sp->name,sizeof(char),16,fo);
   sp=sp->next;
 }
 outlng(Nvert);
 fv=NULL; vp=MainVp; while(vp != NULL){fv=vp; vp=vp->last;}
 vp=fv; while(vp != NULL){
   if(vp->sp != NULL)outsht(vp->sp->id);
   else              outsht(0);
   vp=vp->next;
 }
}

static void WriteWireframe(void){
 long i;
 if(w_frame.Np == 0 || w_frame.Ne == 0)return;
 fprintf(fo,"WFRM");
 outlng((long)(((w_frame.Np*3)+(w_frame.Ne*2)+2)*sizeof(long)));
 outlng(w_frame.Np); outlng(w_frame.Ne);
 for(i=0;i<w_frame.Np;i++){
   outlng(w_frame.p[i][0]);
   outlng(w_frame.p[i][1]);
   outlng(w_frame.p[i][2]);
 }
 for(i=0;i<w_frame.Ne;i++){
   outlng(w_frame.e[i][0]); outlng(w_frame.e[i][1]);
 }
}

#include "dxfcon2.c"

BOOL DoDxfConvert(char *infile, char *outfile, double wt,
   short h_type,short l_freeze, short b_freeze, short no_weld,
   short d2, short d3, short cs){
 long i;
 short error,report=0;
 char argv1[256];
 CreateSkeleton(NULL);               /* make a root node for skeleton */
 FirstSp=MainSp;
 ruler=32768;
 TextureP[63][0]=0; TextureP[63][1]=0; TextureP[63][2]=TVpointZ;
 TextureX[63][0]=TVsizeX/2; TextureX[63][1]=0;
 TextureX[63][2]=TVpointZ;
 TextureY[63][0]=0; TextureY[63][1]=TVsizeZ/2; TextureY[63][2]=TVpointZ;
 TextureSk[63]=NULL;
 TextureMap[63]=PLANE;
 TextureType[63]=0;
 if(quiet)report=0;
 strcpy(argv1,infile);
 if((i=strlen(argv1)) < 4 || argv1[i-4] != '.')strcat(argv1,".DXF");
 error=DxfLoad(argv1,h_type,l_freeze,b_freeze,report,wt,no_weld,d2,d3,cs);
 if(error){
   if(quiet){
     if(error == 5)return FALSE;
     return FALSE;
   }
   if(error == 5)X__pri("Model is too BIG try increasing the weld threshold");
   if(error == 2)X__pri("DXF file truncated");
   if(error == 1)X__pri("Failed to open file");
   return FALSE;
 }
 strcpy(argv1,outfile);
 if((i=strlen(argv1)) < 4 || argv1[i-4] != '.')strcat(argv1,".mfx");
 error=SaveObject(argv1,1);
 if(quiet){
   if(error)return FALSE;
   return TRUE;
 }
 if(error)X__pri("Error Opening model output file");
 else X__pri("Build finished: [ %ld Vertices  %ld Faces ]",Nvert,Nface);
 return TRUE;
}
