/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file DXFLOAD in utilities */

typedef struct DXFCOLOUR {
  unsigned char r,g,b;
} DXFcolour;

static DXFcolour DXFc[256];

typedef struct DXFBLOCK {
  char   name[32];
  long   offset,x,y,z,frozen;
  vertex *vs;
  edge   *es;
} DXFblock;

typedef struct DXFLAYER {
  long colour,frozen,nought;
  char name[32];
  skel *sp;
} DXFlayer;

typedef struct AUTOVERTEX {
  vertex *vp;
  long n;
  struct AUTOVERTEX **adj;
} autovertex;

static DXFlayer *DXFlayers=NULL;
static DXFblock *DXFblocks=NULL;
static long nDXFblocks,nDXFlayers,skel_type,l_freeze,b_freeze,verbose,
            current_colour,current_layer,current_nought,no_weld,inc2d,
            current_freeze,inc3d,circle_step;
static double weld_threshold=0.0001,weld_mult;
static char CurrentLayer[32];
static skel *current_sp;
static long file_size=1;

static short InsertDXFblock(FILE *, skel *);

static void UpdateProgress(FILE *dxfp){
 long p;
 if(hwndProgress == NULL)return;
 p=ftell(dxfp);
 p=(p*100)/file_size;
 SendMessage(hwndProgress,PBM_SETPOS,(WPARAM)p,0);
 Sleep(0);
// MSG msg;
// while (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE)){
//   TranslateMessage(&msg);
//   DispatchMessage(&msg);
// }
 return;
}

static void AddDXFface(vertex *v1, vertex *v2, vertex *v3, short check){
 CreateFace(v1,v2,v3,check);
 MainFp->color[0]=DXFc[current_colour].r;
 MainFp->color[1]=DXFc[current_colour].g;
 MainFp->color[2]=DXFc[current_colour].b;
}

void veccross(double *o, double *a, double *b){
  double r[3];
  r[0] = a[1] * b[2] - a[2] * b[1];
  r[1] = a[2] * b[0] - a[0] * b[2];
  r[2] = a[0] * b[1] - a[1] * b[0];
  o[0]=r[0];
  o[1]=r[1];
  o[2]=r[2];
}

vecnorm(double *o, double *a){
  double mag;
  mag = sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]);
  mag = 1.0/mag;
  o[0]=a[0]*mag;
  o[1]=a[1]*mag;
  o[2]=a[2]*mag;
}

void arbaxis(double *newaxis, double *oldaxis){
 double yaxis[3] = {0.0, 1.0, 0.0};
 double zaxis[3] = {0.0, 0.0, 1.0};
 if (fabs(oldaxis[0]) < 0.015625 && fabs(oldaxis[1]) < 0.015625)
    veccross(newaxis, yaxis, oldaxis);
 else
    veccross(newaxis, zaxis, oldaxis);
 vecnorm(newaxis, newaxis);
}

void gettrans(double *zdir, double m[3][3]){
 double xdir[3], ydir[3];
 vecnorm(zdir,zdir);
 arbaxis(xdir, zdir);
 veccross(ydir, zdir, xdir);
 vecnorm(ydir, ydir);
 m[0][0]=xdir[0];
 m[1][0]=xdir[1];
 m[2][0]=xdir[2];
 m[0][1]=ydir[0];
 m[1][1]=ydir[1];
 m[2][1]=ydir[2];
 m[0][2]=zdir[0];
 m[1][2]=zdir[1];
 m[2][2]=zdir[2];
}

#if 0
// the old way this appears to be wrong
static void ReOrientate(double Nx,double Ny,double Nz,vertex *vl,vertex *vs){
 vertex *vp;                  /* apply the AUTOCAD arbitrary axis rotation */
 double x,y,z,dx,dy,dz;
 long axis=0;
 /* X__pri("Not normal direction %lf %lf %lf",Nx,Ny,Nz); */
 if(fabs(Nx) < 0.015625 && fabs(Ny) < 0.015625)axis=1;
 vp=vl; while(vp != vs){
   dx=(double)vp->xyz[0]; dy=(double)vp->xyz[1]; dz=(double)vp->xyz[2];
   if(axis){
     x =  Nz*dx - Nx*Ny*dy + Nx*dz;
     y =  (Nx*Nx+Nz*Nz)*dy + Ny*dz;
     z = -Nx*dx + Nz*Ny*dy + Nz*dz;
   }
   else{
     x = -Ny*dx - Nx*Nz*dy + Nx*dz;
     y =  Nx*dx + Ny*Nz*dy + Ny*dz;
     z =  (Nx*Nx+Ny*Ny)*dy + Nz*dz;
   }
   vp->xyz[0]=(long)x; vp->xyz[1]=(long)y; vp->xyz[2]=(long)z;
   vp=vp->last;
 }
}
#endif

static void ReOrientate(double Nx,double Ny,double Nz,vertex *vl,vertex *vs){
 vertex *vp;
 long i;
 double z[3],m[3][3],dx,dy,dz;
 z[0]=Nx; z[1]=Ny; z[2]=Nz;
 gettrans(z,m);
 vp=vl; while(vp != vs){
   dx=(double)(vp->xyz[0]);
   dy=(double)(vp->xyz[1]);
   dz=(double)(vp->xyz[2]);
   for (i=0; i<3; i++){
     vp->xyz[i] = (long)(m[i][0] * dx + m[i][1] * dy + m[i][2] * dz);
   }
   vp=vp->last;
 }
}

static void DXFhsv2rgb(double hh, double ss, double vv,
             unsigned char *r, unsigned char *g, unsigned char *b){
 double rr,gg,bb,maxcol,mincol;
 mincol=vv*(1.0-ss);
 if(hh <= 120.0){bb=mincol;
   if(hh <=  60){rr=vv; gg=mincol+hh*(vv-mincol)/(120.0-hh);}
   else         {gg=vv; rr=mincol+(120.0-hh)*(vv-mincol)/hh;}
 }
 else if( hh <= 240){rr=mincol;
   if(hh <= 180){gg=vv; bb=mincol+(hh-120.0)*(vv-mincol)/(240.0-hh);}
   else         {bb=vv; gg=mincol+(240.0-hh)*(vv-mincol)/(hh-120.0);}
 }
 else {gg=mincol;
   if(hh <= 300){bb=vv; rr=mincol+(hh-240.0)*(vv-mincol)/(360.0-hh);}
   else         {rr=vv; bb=mincol+(360.0-hh)*(vv-mincol)/(hh-240.0);}
 }
 rr = rr*255.0; *r=(unsigned char)rr;
 gg = gg*255.0; *g=(unsigned char)gg;
 bb = bb*255.0; *b=(unsigned char)bb;
}

static void SetupDXFcolours(void){
  double hue;
  long i,j,k,l;
  DXFc[  0].r=  0; DXFc[  0].g=  0; DXFc[  0].b=  0;
  DXFc[  1].r=255; DXFc[  1].g=  0; DXFc[  1].b=  0;
  DXFc[  2].r=255; DXFc[  2].g=255; DXFc[  2].b=  0;
  DXFc[  3].r=  0; DXFc[  3].g=255; DXFc[  3].b=  0;
  DXFc[  4].r=  0; DXFc[  4].g=255; DXFc[  4].b=255;
  DXFc[  5].r=  0; DXFc[  5].g=  0; DXFc[  5].b=255;
  DXFc[  6].r=255; DXFc[  6].g=  0; DXFc[  6].b=255;
  DXFc[  7].r=255; DXFc[  7].g=255; DXFc[  7].b=255;
  DXFc[  8].r= 64; DXFc[  8].g= 64; DXFc[  8].b= 64;
  DXFc[  9].r=128; DXFc[  9].g=128; DXFc[  9].b=128;
  for(i=10;i<=240;i+=10){
    for(j=0;j<=8;j+=2){
      for(k=0;k<2;k++){
        l=i+j+k; hue=((double)i/240.0)*360.0;
        DXFhsv2rgb(hue,(double)(k+1)/2.0,0.1+(double)(8-j)/10.0,
          &(DXFc[l].r),&(DXFc[l].g),&(DXFc[l].b));
      }
    }
  }
  DXFc[250].r= 42; DXFc[250].g= 42; DXFc[250].b= 42;
  DXFc[251].r= 84; DXFc[251].g= 84; DXFc[251].b= 84;
  DXFc[252].r=126; DXFc[252].g=126; DXFc[252].b=126;
  DXFc[252].r=169; DXFc[253].g=169; DXFc[253].b=169;
  DXFc[254].r=210; DXFc[254].g=210; DXFc[254].b=210;
  DXFc[255].r=252; DXFc[255].g=252; DXFc[255].b=252;
}

/* changed to accomodate MAC style as well  see also file open */

static long getDXFcode(FILE *dxfp, long *vi, double *di, char *dstr){
 long gc,i,j;
 strnset(dstr,0,256);
// j=0; while(((i=fgetc(dxfp)) != 13) && i >= 0 && j < 256)
 //     if(i != 10) *(dstr+(j++))=i;
 j=0;
 while(1){
   i=fgetc(dxfp);
   if(i < 0 || j > 255)return -1;
   else if(i == 13 || i == 10){
     if((i=fgetc(dxfp)) < 0)break; /* end of file */
     if( !(i == 13 || i == 10))ungetc(i,dxfp);
     break;
   }
   *(dstr+(j++))=i;
 }
 *(dstr+j)=0;
//X__pri("[%s] j=%ld lenght=%ld",dstr,j,(long)strlen(dstr));
 if(sscanf(dstr,"%ld",&gc) != 1)return -1;
 strnset(dstr,0,256);
// j=0; while(((i=fgetc(dxfp)) != 13) && i >= 0 && j < 256)
//      if(i != 10) *(dstr+(j++))=i;
 j=0;
 while(1){
   i=fgetc(dxfp);
   if(i < 0 || j > 255)return -1;
   else if(i == 13 || i == 10){
     if((i=fgetc(dxfp)) < 0)break;
     if( !(i == 13 || i == 10))ungetc(i,dxfp);
     break;
   }
   *(dstr+(j++))=i;
 }
 *(dstr+j)=0;
//X__pri("[%s] j=%ld lenght=%ld",dstr,j,(long)strlen(dstr));
 if(gc < 10 || gc == 999){
   return gc;
 }
 else if(gc < 60){
   if(sscanf(dstr,"%lf",di) != 1)return -1;
   return gc;
 }
 else if(gc > 60 && gc < 80){
   if(sscanf(dstr,"%ld",vi) != 1)return -1;
   return gc;
 }
 else if(gc > 80 && gc < 210){ /* not in autocad manual !!!!!!! */
   return gc;
 }
 else if(gc >= 210 && gc < 240){
   if(sscanf(dstr,"%lf",di) != 1)return -1;
   return gc;
 }
 else if(gc >= 1000 && gc < 1010){
   return gc;
 }
 else if(gc >= 1010 && gc < 1060){
   if(sscanf(dstr,"%lf",di) != 1)return -1;
   return gc;
 }
 else if(gc >= 1060 && gc < 1080){
   if(sscanf(dstr,"%ld",vi) != 1)return -1;
   return gc;
 }
 return -1;
}

static void CreateEdgeWithCheck(vertex *v1, vertex *v2, edge *le){
 edge *ep;
 ep=MainEp;
 while(ep != le){
   if(   ((ep->V[0] == v1) && (ep->V[1] == v2))
      || ((ep->V[0] == v2) && (ep->V[1] == v1)) )return;
   ep=ep->last;
 }
 CreateEdge(v1,v2,0);
}

static short ReadDXFheader(FILE *dxfp){
 char dstr[260]; double di; long vi,gc;
 if(verbose > 1)X__pri("Header");
 for(;;){
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0 && strncmp(dstr,"ENDSEC",6) == 0)break;
 }
 UpdateProgress(dxfp);
 return 0;
}

static short AddLayerToTable(char *name, long colour, long frozen){
 DXFlayer *db;
 long i,j;
 char confirm;
 if(nDXFlayers > 0){
   for(i=0;i<nDXFlayers;i++){
     if(strcmp(name,DXFlayers[i].name) == 0){
       return 0;
     }
   }
 }
 db=DXFlayers;
 if(db == NULL)db=(DXFlayer *)malloc(sizeof(DXFlayer));
 else db=(DXFlayer *)realloc(db,(nDXFlayers+1)*sizeof(DXFlayer));
 if(db == NULL){
   if(DXFlayers != NULL)free(DXFlayers); DXFlayers=NULL; nDXFlayers=0;
   return -1;
 }
 DXFlayers=db;
 if(l_freeze){
//   frozen ^= 1;
   frozen = 0;
 }
 DXFlayers[nDXFlayers].colour=colour;
 DXFlayers[nDXFlayers].frozen=frozen;
 if(strcmp("0",name) == 0)DXFlayers[nDXFlayers].nought=1;
 else                     DXFlayers[nDXFlayers].nought=0;
 strcpy(DXFlayers[nDXFlayers].name,name);
 if(skel_type == 1){
   CreateSkeleton(FirstSp);
   j=min(12,strlen(name)); strncpy(MainSp->name,name,j); *(MainSp->name+j)=0;
   DXFlayers[nDXFlayers].sp=MainSp;
   MainSp->xyz[2]=MINUNIT*2*Nskel;
 }
 else DXFlayers[nDXFlayers].sp=NULL;
 nDXFlayers++;
 return 0;
}

static void SetLayerCurrents(char *name){
/* If the layer is layer 0 then if the entity is in a block the Layer + */
/* colour used is the colour of the block that is being        inserted */
 long i;
 if(nDXFlayers > 0){
   for(i=0;i<nDXFlayers;i++){
     if(strcmp(name,DXFlayers[i].name) == 0)goto GOTIT;
   }
 }
 if(AddLayerToTable(name,7,0) < 0)return;   /* colour 7 not frozen */
 i=nDXFlayers-1;
 GOTIT:
 current_nought = DXFlayers[i].nought;
 if(!current_nought){
   current_freeze=DXFlayers[i].frozen;
   current_colour=DXFlayers[i].colour;
   current_layer  = i;
   strcpy(CurrentLayer,DXFlayers[i].name);
   if(skel_type == 1)current_sp=DXFlayers[i].sp;
 }
}

static short ReadDXFtables(FILE *dxfp){
 char name[32],dstr[260]; double di; long vi,gc;
 short layer_table=0;
 long freeze=0;
 if(verbose > 0)X__pri("Tables");
 for(;;){
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0 && strncmp(dstr,"ENDSEC",6) == 0)break;
   if(gc == 0 && strncmp(dstr,"TABLE",5) == 0){
     if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
     if(gc == 2){
       if(verbose > 1)X__pri("Table [%s]",dstr);
       if(strncmp(dstr,"LAYER",5) == 0){
          layer_table=1;
          if(l_freeze)X__pri("Confirm the Freeze/Thaw status of layers");
       }
       else layer_table=0;
       for(;;){
         if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
         if(gc == 0 && strncmp(dstr,"ENDTAB",6) == 0)break;
         if(layer_table){
           if(gc == 2)strcpy(name,dstr);
           else if(gc == 70){
             if((vi & 1) == 1)freeze=1;
             else             freeze=0;
           }
           else if(gc == 62){
             if(AddLayerToTable(name,vi,freeze) < 0)return -1;
             freeze=0;
           }
         }
       }
       if(verbose > 1)X__pri("End Table");
     }
   }
 }
 UpdateProgress(dxfp);
 return 0;
}

static short CreateSimplePolyline(FILE *dxfp, vertex *lastvertex,
             edge *lastedge, long closed, long height, long line3d,
             double dx, double dy, double dz){
 char dstr[260]; double di; long vi,gc;
 long x=0,y=0,z=0,flag=0;
 vertex *vp,*tv,*f1=NULL,*f2=NULL,*l1=NULL,*l2=NULL,*l0=NULL;
//X__pri("Standard polyline c=%ld h=%ld 3d=%ld",closed,height,line3d);
 tv=MainVp;
 for(;;){
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   else if(gc == 0 && strncmp(dstr,"SEQEND",6) == 0)flag=1;
   if     (gc == 10)x=(long)(di*weld_mult);
   else if(gc == 20)y=(long)(di*weld_mult);
   else if(gc == 30)z=(long)(di*weld_mult);
   else if(gc == 62)current_colour=vi;
   else if(gc ==  8)SetLayerCurrents(dstr);
   if(flag || (gc == 0 && strncmp(dstr,"VERTEX",5) == 0)){
   /* next vertex or end of sequence so create the vertex */
     CreateVertex();
     MainVp->xyz[0]=x; MainVp->xyz[1]=y; MainVp->xyz[2]=z;
     l0=l1;
     if(l1 != NULL)CreateEdge(MainVp,l1,0); else f1=MainVp;
     l1=MainVp;
     if(height > 0){
       CreateVertex();
       MainVp->xyz[0]=x; MainVp->xyz[1]=y; MainVp->xyz[2]=z+height;
       CreateEdge(MainVp,l1,0);
       if(l2 != NULL){ /* l0 will also be non null */
         CreateEdge(MainVp,l2,0); CreateEdge(MainVp,l0,0);
         AddDXFface(MainVp,l2,l0,0);
         AddDXFface(MainVp,l0,l1,0);
       }
       else f2=MainVp;
       l2=MainVp;
     }
   }
   if(flag){ /* end of sequence so quit */
     if(closed && tv != NULL && f1 != NULL){ /* cloes off the line */
       CreateEdge(l1,f1,0);
       if(height > 0){
         CreateEdge(l2,f2,0); CreateEdge(f1,l1,0);
         AddDXFface(f2,l2,l1,0);
         AddDXFface(f2,l1,f1,0);
       }
     }
     if(current_sp != NULL){ /* assign to current_skeleton       */
        vp=MainVp; while(vp != tv){ vp->sp=current_sp; vp=vp->last; }
     }
     if(!line3d && (dx != 0.0 || dy != 0.0 || dz != 1.0)){
       ReOrientate(dx,dy,dz,MainVp,tv);
     }
     return 0;
   }
 }
}

void BuildEdges(long n, autovertex *av){
 long i,j,nn;
 for(j=0;j<n;j++){
  nn=(av+j)->n;
  if(nn > 0)for(i=0;i<nn;i++)
    CreateEdge((av+j)->vp,((av+j)->adj[i])->vp,0);
 }
}

static void InsertInVertexList(autovertex *vj, autovertex *vi){
 long i,j;
 if(vi->n > 0)for(i=0;i<vi->n;i++)if(vi->adj[i] == vj)return;
 if(vj->n > 0)for(j=0;j<vj->n;j++)if(vj->adj[j] == vi)return;
 if((vi->adj = (autovertex **)realloc(vi->adj,(vi->n + 1)
       * sizeof(autovertex *))) == NULL){
   if(verbose > 0)X__pri("out of memory");  exit(0);
 }
 vi->adj[vi->n] = vj;
 (vi->n)++;
}

static autovertex *NewDXFvertex(vertex *lastvertex, long x, long y, long z,
                                autovertex *av, long *nav){
  vertex *vp,*vf;
  autovertex *tav;
  vf=NULL; tav=av;
  if(tav == NULL)tav=(autovertex *)malloc(sizeof(autovertex));
  else tav=(autovertex *)realloc(av,sizeof(autovertex)*( *nav + 1));
  if(tav == NULL){
    if(av != NULL){
      long i;
      for(i=0;i < *nav;i++)if((av+i)->adj != NULL)free((av+i)->adj);
      free(av);
      return NULL;
    }
  }
  av=tav;
  (av + *nav)->n=0; (av + *nav)->adj=NULL;
  if(!no_weld){
    vp=MainVp; while(vp != lastvertex){
      if(x == vp->xyz[0] && y == vp->xyz[1] && z == vp->xyz[2]){
        vf=vp; break;
      }
      vp=vp->last;
    }
  }
  if(vf != NULL){
    av[*nav].vp=vf;
  }
  else{
    CreateVertex();
    MainVp->xyz[0]=x;
    MainVp->xyz[1]=y;
    MainVp->xyz[2]=z;
    av[*nav].vp=MainVp;
  }
  (*nav)++;
  return av;
}

static short CreateDXFpolyline(FILE *dxfp,DXFblock *blk){
 char dstr[260]; double di; long vi,gc;
 long line3d=0,polygon_mesh=0,polyface_mesh=0,polyn=0,polym=0,
      x=0,y=0,z=0,fid[4],nfid=0,facespec=0,nav=0,i,j,k,
      closed=0,height=0;
 short in1=1,in2=1,in3=1,make_vertex;
 double lastx=0,lasty=0,lastz=0,dx=0,dy=0,dz=1;
 vertex *lastvertex,*vp,*tv;
 edge   *lastedge;
 autovertex *av=NULL; tv=MainVp;
 if(blk == NULL || no_weld){ lastvertex=MainVp; lastedge=MainEp;   }
 else                      { lastvertex=blk->vs; lastedge=blk->es; }
 for(;;){ /* do the POLYLINE stuff */
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0 && strncmp(dstr,"VERTEX",5) == 0)break;   /* vertices must follow */
   else if(gc == 0 && strncmp(dstr,"SEQEND",6) == 0)return -1;
   else if(gc == 62)current_colour=vi;
   else if(gc ==  8)SetLayerCurrents(dstr);
   else if(gc == 70){
     if ((vi &  8) ==  8)line3d=1;
     if ((vi & 16) == 16)polygon_mesh=1;
     if ((vi & 64) == 64)polyface_mesh=1;
     if ((vi &  1) ==  1)closed=1;
   }
   else if(gc == 71)polyn=vi;
   else if(gc == 72)polym=vi;
   else if(gc == 39)height=(long)(di*weld_mult);
   else if(gc == 210)dx=di;
   else if(gc == 220)dy=di;
   else if(gc == 230)dz=di;
 }
 if(current_freeze){  /* this POLYLINE entity is on a frozen layer */
   for(;;){
     if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
     else if(gc == 0 && strncmp(dstr,"SEQEND",6) == 0)return 0;
   }
 }
 if(!polyface_mesh && !polygon_mesh){ /* a simple polyline */
   if(height == 0 && !inc2d){ /* skip the poly line */
     for(;;){
       if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
       if(gc == 0 && strncmp(dstr,"SEQEND",6) == 0)return 0;
     }
   }
   if(CreateSimplePolyline(dxfp,lastvertex,lastedge,closed,height,
                           line3d,dx,dy,dz) < 0)return -1;
   return  0;
 }
 LOOP1:
 make_vertex=0;
 LOOP:  /* something else 3d etc */
 if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
 if(polyface_mesh || polygon_mesh){
   if     (gc == 10){lastx=di; x=(long)(lastx*weld_mult); in1=0;}
   else if(gc == 20){lasty=di; y=(long)(lasty*weld_mult); in2=0;}
   else if(gc == 30){lastz=di; z=(long)(lastz*weld_mult); in3=0;}
   else if(gc == 62)current_colour=vi;
   else if(gc ==  8)SetLayerCurrents(dstr);
   else if(gc == 70){
     if((polyface_mesh && vi == 192) || (polygon_mesh && vi == 64)){
       make_vertex=1;
     }
     else if(polyface_mesh && vi == 128){ /* 128 bit -> code */
       facespec=1,nfid=0;        /* face -> vertex entity   */
     }
   }
   else if(polyface_mesh && gc > 70 && gc < 75){ /* face spec    */
     if(nfid > 3)return -1;     /*  more than 4 vertices per face */
     fid[nfid++]=abs(vi);       /*  gather all the face indices   */
     /* NB: the first index is 1 */
   }
 }
 if(gc != 0)goto LOOP;   /* wait until next entity */
 if(make_vertex){
   if((in1 || in2 || in3) && verbose > 0){
//ToDo     X__pri("Warning x y or z not initialised"); getch();
   }
   if((av=NewDXFvertex(lastvertex,x,y,z,av,&nav)) == NULL)return -1;
 }
 if(facespec > 0){ /* a face has been found build it */
   /* X__pri("%ld vertices around face",nfid); */
//   CreateEdgeWithCheck(av[fid[0]-1].vp,av[fid[1]-1].vp,lastedge);
//   CreateEdgeWithCheck(av[fid[1]-1].vp,av[fid[2]-1].vp,lastedge);
//   CreateEdgeWithCheck(av[fid[2]-1].vp,av[fid[0]-1].vp,lastedge);
     InsertInVertexList(&av[fid[0]-1],&av[fid[1]-1]);
     InsertInVertexList(&av[fid[1]-1],&av[fid[2]-1]);
     InsertInVertexList(&av[fid[2]-1],&av[fid[0]-1]);
   AddDXFface(av[fid[0]-1].vp,av[fid[1]-1].vp,av[fid[2]-1].vp,0);
   if(nfid > 3){ /* create 2nd face since 4 edges */
//     CreateEdgeWithCheck(av[fid[2]-1].vp,av[fid[3]-1].vp,lastedge);
//     CreateEdgeWithCheck(av[fid[3]-1].vp,av[fid[0]-1].vp,lastedge);
     InsertInVertexList(&av[fid[2]-1],&av[fid[3]-1]);
     InsertInVertexList(&av[fid[3]-1],&av[fid[0]-1]);
     AddDXFface(av[fid[0]-1].vp,av[fid[2]-1].vp,av[fid[3]-1].vp,0);
   }
   nfid=0; facespec=0;
 }
 if(gc == 0 && strncmp(dstr,"SEQEND",6) == 0){
   /* build polygon meshes if any */
   if(polygon_mesh == 1 && polyn > 1 && polym > 1 &&
     nav == polyn*polym){ /* build mesh ! ensure correct No.V's*/
     for(j=0;j<polyn;j++){
       for(k=0;k<polym;k++){
         i=j*polym+k;
         if(k < polym-1)CreateEdge(av[i].vp,av[i+1].vp,0);
         if(j > 0){
           CreateEdge(av[i].vp,av[i-polym].vp,0);
           if(k > 0){
             CreateEdge(av[i].vp,av[i-polym-1].vp,0);
             AddDXFface(av[i-1].vp,av[i].vp,av[i-polym-1].vp,0);
             MainFp->texture |= 0x80;
             AddDXFface(av[i].vp,av[i-polym].vp,av[i-polym-1].vp,0);
             MainFp->texture |= 0x80;
           }
         }
       }
     }
   }
   if(av != NULL){
     if(polyface_mesh)BuildEdges(nav,av);
     for(i=0;i<nav;i++)if((av+i)->adj != NULL)free((av+i)->adj);
     free(av); /* free any autovertices */
   }
   if(current_sp != NULL){ /* assign to current_skeleton       */
      vp=MainVp; while(vp != tv){ vp->sp=current_sp; vp=vp->last; }
   }
   return 0;
 }
 goto LOOP1;
}

static short CreateDXF3dface(FILE *dxfp, DXFblock *blk){
 char dstr[260]; double di; long vi,gc;
 vertex *LastUsedVertex,*vp,*vpi[4],*tv;
 edge   *LastUsedEdge;
 long j,Nf,Face_X[4],Face_Y[4],Face_Z[4];
 tv=MainVp;
 if(blk == NULL){ LastUsedVertex=MainVp;  LastUsedEdge=MainEp;  }
 else           { LastUsedVertex=blk->vs; LastUsedEdge=blk->es; }
 for(;;){
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc==33){ /* last vertex so now build the face/faces */
     if(!current_freeze){
       Face_Z[3]=(long)(di*weld_mult);
       if(Face_X[3] == Face_X[2] &&
          Face_Y[3] == Face_Y[2] &&
          Face_Z[3] == Face_Z[2] )Nf=3;/* a single triangle */
       else Nf=4;
       for(j=0;j<Nf;j++){
         vpi[j]=NULL;
         vp=MainVp; while(vp != LastUsedVertex){
           if(Face_X[j] == vp->xyz[0] &&
              Face_Y[j] == vp->xyz[1] &&
              Face_Z[j] == vp->xyz[2]){vpi[j]=vp; break;}
           vp=vp->last;
         }
       }
       for(j=0;j<Nf;j++){
         if(vpi[j] == NULL){
           CreateVertex();
           vpi[j]=MainVp;
           MainVp->xyz[0]=Face_X[j];
           MainVp->xyz[1]=Face_Y[j];
           MainVp->xyz[2]=Face_Z[j];
         }
       }
       CreateEdgeWithCheck(vpi[0],vpi[1],LastUsedEdge);
       CreateEdgeWithCheck(vpi[1],vpi[2],LastUsedEdge);
       AddDXFface(vpi[0],vpi[1],vpi[2],0);
       if(Nf > 3){
         AddDXFface(vpi[0],vpi[2],vpi[3],0);
         CreateEdgeWithCheck(vpi[2],vpi[3],LastUsedEdge);
         CreateEdgeWithCheck(vpi[3],vpi[0],LastUsedEdge);
         CreateEdge(vpi[0],vpi[2],0);
       }
       else{
         CreateEdgeWithCheck(vpi[2],vpi[0],LastUsedEdge);
       }
       if(current_sp != NULL){
         vp=MainVp; while(vp != tv){ vp->sp=current_sp; vp=vp->last; }
       }
     }
     break;  /* got all this entity go to next */
   }
   else if(gc==10){Face_X[0]=(long)(di*weld_mult);}
   else if(gc==20){Face_Y[0]=(long)(di*weld_mult);}
   else if(gc==30){Face_Z[0]=(long)(di*weld_mult);}
   else if(gc==11){Face_X[1]=(long)(di*weld_mult);}
   else if(gc==21){Face_Y[1]=(long)(di*weld_mult);}
   else if(gc==31){Face_Z[1]=(long)(di*weld_mult);}
   else if(gc==12){Face_X[2]=(long)(di*weld_mult);}
   else if(gc==22){Face_Y[2]=(long)(di*weld_mult);}
   else if(gc==32){Face_Z[2]=(long)(di*weld_mult);}
   else if(gc==13){Face_X[3]=(long)(di*weld_mult);}
   else if(gc==23){Face_Y[3]=(long)(di*weld_mult);}
   else if(gc == 8){
     if(strcmp(dstr,CurrentLayer) != 0){ /* if the 3dFace is not on the  */
       if(blk != NULL){                  /* same layer, then change the  */
         blk->vs=MainVp;                 /* the Weld origin, this speeds */
         blk->es=MainEp;                 /* up welding a lot !!!         */
       }
       LastUsedVertex=MainVp;
       LastUsedEdge=MainEp;
     }
     SetLayerCurrents(dstr);
   }
   else if(gc == 62)current_colour=vi;
 }
 return 0;
}

short CreateDXFsolid(FILE *dxfp, DXFblock *blk){
 char dstr[260]; double di; long vi,gc;
 long i,offset,x[4],y[4],z[4],height=0;
 double dx=0,dy=0,dz=1;
 vertex *v1,*v2,*v3,*v4,*v5,*v6,*v7,*v8,*vl;
 for(i=0;i<4;i++)x[i]=y[i]=z[i]=0;
 for(;;){
   offset=ftell(dxfp);
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0)break;  /* next entity */
   else if(gc ==  8)SetLayerCurrents(dstr);
   else if(gc == 62)current_colour=vi;
   else if(gc > 9 && gc < 14)x[gc-10]=(long)(di*weld_mult);
   else if(gc >19 && gc < 24)y[gc-20]=(long)(di*weld_mult);
   else if(gc >29 && gc < 34)z[gc-30]=(long)(di*weld_mult);
   else if(gc == 39)height=(long)(di*weld_mult);
   else if(gc == 210)dx=di;
   else if(gc == 220)dy=di;
   else if(gc == 230)dz=di;
 }
 vl=MainVp;
 if(height != 0){ /* make a solid */
   if(!current_freeze){
     CreateVertex(); v1=MainVp; CreateVertex(); v2=MainVp;
     CreateVertex(); v3=MainVp; CreateVertex(); v4=MainVp;
     CreateVertex(); v5=MainVp; CreateVertex(); v6=MainVp;
     CreateVertex(); v7=MainVp; CreateVertex(); v8=MainVp;
     if(current_sp != NULL){
       v1->sp=v2->sp=v3->sp=v4->sp=current_sp;
       v5->sp=v6->sp=v7->sp=v8->sp=current_sp;
     }
     CreateEdge(v1,v2,0); CreateEdge(v2,v3,0);   /* top */
     CreateEdge(v3,v4,0); CreateEdge(v4,v1,0);
     CreateEdge(v5,v6,0); CreateEdge(v6,v7,0);   /* bottom */
     CreateEdge(v7,v8,0); CreateEdge(v8,v5,0);
     CreateEdge(v1,v5,0); CreateEdge(v2,v6,0);   /* sides */
     CreateEdge(v3,v7,0); CreateEdge(v4,v8,0);
     CreateEdge(v1,v6,0); CreateEdge(v2,v7,0);   /* side diagonal */
     CreateEdge(v3,v8,0); CreateEdge(v4,v5,0);
     AddDXFface(v1,v2,v6,0); AddDXFface(v1,v6,v5,0); /* sides */
     AddDXFface(v2,v3,v7,0); AddDXFface(v2,v7,v6,0);
     AddDXFface(v3,v4,v8,0); AddDXFface(v3,v8,v7,0);
     AddDXFface(v4,v1,v5,0); AddDXFface(v4,v5,v8,0);
     CreateEdge(v1,v3,0); /* top diagonal */
     AddDXFface(v1,v2,v3,0); AddDXFface(v1,v3,v4,0); /* top */
     AddDXFface(v5,v6,v7,0); AddDXFface(v5,v7,v8,0); /* bottom */
     CreateEdge(v5,v7,0); /* bottom diagonal */
     v1->xyz[0]=x[0]; v1->xyz[1]=y[0]; v1->xyz[2]=z[0]+height;
     v2->xyz[0]=x[1]; v2->xyz[1]=y[1]; v2->xyz[2]=z[1]+height;
     v3->xyz[0]=x[3]; v3->xyz[1]=y[3]; v3->xyz[2]=z[3]+height;
     v4->xyz[0]=x[2]; v4->xyz[1]=y[2]; v4->xyz[2]=z[2]+height;
     v5->xyz[0]=x[0]; v5->xyz[1]=y[0]; v5->xyz[2]=z[0];
     v6->xyz[0]=x[1]; v6->xyz[1]=y[1]; v6->xyz[2]=z[1];
     v7->xyz[0]=x[3]; v7->xyz[1]=y[3]; v7->xyz[2]=z[3];
     v8->xyz[0]=x[2]; v8->xyz[1]=y[2]; v8->xyz[2]=z[2];
   }
 }
 else {
   CreateVertex(); v1=MainVp; CreateVertex(); v2=MainVp;
   CreateVertex(); v3=MainVp; CreateVertex(); v4=MainVp;
   if(current_sp != NULL){
     v1->sp=v2->sp=v3->sp=v4->sp=current_sp;
   }
   CreateEdge(v1,v2,0); CreateEdge(v2,v3,0);   /* top */
   CreateEdge(v3,v4,0); CreateEdge(v4,v1,0); CreateEdge(v1,v3,0);
   v1->xyz[0]=x[0]; v1->xyz[1]=y[0]; v1->xyz[2]=z[0];
   v2->xyz[0]=x[1]; v2->xyz[1]=y[1]; v2->xyz[2]=z[1];
   v3->xyz[0]=x[3]; v3->xyz[1]=y[3]; v3->xyz[2]=z[3];
   v4->xyz[0]=x[2]; v4->xyz[1]=y[2]; v4->xyz[2]=z[2];
   AddDXFface(v1,v2,v3,0);
   AddDXFface(v1,v3,v4,0);
 }
 if(dx != 0.0 || dy != 0.0 || dz != 1.0){
   ReOrientate(dx,dy,dz,MainVp,vl);
 }
 fseek(dxfp,offset,SEEK_SET); /* go back to get the next entity */
 return 0;
}

short CreateDXFline(FILE *dxfp, DXFblock *blk){
 char dstr[260]; double di; long vi,gc;
 long offset,x[2],y[2],z[2],height=0;
 double dx=0,dy=0,dz=1;
 vertex *v1,*v2,*v3,*v4,*vp,*vl;
 x[0]=x[1]=y[0]=y[1]=z[0]=z[1]=0;
 for(;;){
   offset=ftell(dxfp);
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0)break;  /* next entity */
   else if(gc ==  8)SetLayerCurrents(dstr);
   else if(gc == 62)current_colour=vi;
   else if(gc > 9 && gc < 12)x[gc-10]=(long)(di*weld_mult);
   else if(gc >19 && gc < 22)y[gc-20]=(long)(di*weld_mult);
   else if(gc >29 && gc < 32)z[gc-30]=(long)(di*weld_mult);
   else if(gc == 39)height=(long)(di*weld_mult);
   else if(gc == 210)dx=di;
   else if(gc == 220)dy=di;
   else if(gc == 230)dz=di;
 }
 vl=MainVp;
 if(height != 0){ /* make an extruded line */
   if(!current_freeze){
     CreateVertex(); v1=MainVp; CreateVertex(); v2=MainVp;
     CreateVertex(); v3=MainVp; CreateVertex(); v4=MainVp;
     if(current_sp != NULL){
       v1->sp=v2->sp=v3->sp=v4->sp=current_sp;
     }
     CreateEdge(v1,v2,0); CreateEdge(v2,v3,0);
     CreateEdge(v3,v4,0); CreateEdge(v4,v1,0);
     CreateEdge(v1,v3,0); /* top diagonal */
     AddDXFface(v1,v2,v3,0); AddDXFface(v1,v3,v4,0);
     v1->xyz[0]=x[0]; v1->xyz[1]=y[0]; v1->xyz[2]=z[0];
     v2->xyz[0]=x[1]; v2->xyz[1]=y[1]; v2->xyz[2]=z[1];
     v3->xyz[0]=x[1]; v3->xyz[1]=y[1]; v3->xyz[2]=z[1]+height;
     v4->xyz[0]=x[0]; v4->xyz[1]=y[0]; v4->xyz[2]=z[0]+height;
   }
 }
 else if(inc2d && !current_freeze){ /* 2D line if required */
//X__pri("2D line on layer [%s] %ld",CurrentLayer,current_freeze);
   v1=NULL; v2=NULL;
   if(blk != NULL && !no_weld){
     vp=MainVp; while(vp != blk->vs){
       if(vp->xyz[0] == x[0] && vp->xyz[1] == y[0] && vp->xyz[2] == z[0])v1=vp;
       if(vp->xyz[0] == x[1] && vp->xyz[1] == y[1] && vp->xyz[2] == z[1])v2=vp;
       vp=vp->last;
     }
   }
   if(v1 == NULL){
     CreateVertex(); v1=MainVp; if(current_sp != NULL)v1->sp=current_sp;
     v1->xyz[0]=x[0]; v1->xyz[1]=y[0]; v1->xyz[2]=z[0];
   }
   if(v2 == NULL){
     CreateVertex(); v2=MainVp; if(current_sp != NULL)v2->sp=current_sp;
     v2->xyz[0]=x[1]; v2->xyz[1]=y[1]; v2->xyz[2]=z[1];
   }
   CreateEdge(v1,v2,0);
 }
// lines should not need orientation
// if(dx != 0.0 || dy != 0.0 || dz != 1.0){
//   ReOrientate(dx,dy,dz,MainVp,vl);
// }
 fseek(dxfp,offset,SEEK_SET); /* go back to get the next entity */
 return 0;
}

short CreateDXFcircle(FILE *dxfp, DXFblock *blk, long fill){/* ARC fill=0 */
 char dstr[260]; double di; long vi,gc;
 long offset,x=0,y=0,z=0,height=0,i,sa,ea;
 double dx=0,dy=0,dz=1,d=0,z2=0,x2=0,y2=0;
 vertex *v1,*v2,*v1s,*v2s,*v1l,*v2l,*vc1,*vc2,*vl,*vp;
 skel *sp=NULL;
 for(;;){
   offset=ftell(dxfp);
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0)break;  /* next entity */
   else if(gc ==  8)SetLayerCurrents(dstr);
   else if(gc == 62)current_colour=vi;
   else if(gc == 10)x=(long)(di*weld_mult);
   else if(gc == 20)y=(long)(di*weld_mult);
   else if(gc == 30)z=(long)(di*weld_mult);
   else if(gc == 40)d=(double)(di*weld_mult);
   else if(gc == 39)height=(long)(di*weld_mult);
   else if(gc == 50)sa=(long)di;
   else if(gc == 51)ea=(long)di;
   else if(gc == 210)dx=di;
   else if(gc == 220)dy=di;
   else if(gc == 230)dz=di;
 }
 vl=MainVp;
 if(height != 0){ /* make an extruded circle or arc */
   if(!current_freeze){
     if(fill){  /* circle */
       CreateVertex(); vc1=MainVp; CreateVertex(); vc2=MainVp; /* Bottom */
       vc1->xyz[0]=x; vc1->xyz[1]=y; vc1->xyz[2]=z;
       vc2->xyz[0]=x; vc2->xyz[1]=y; vc2->xyz[2]=z+height;     /* Top    */
       for(i=0;i<360;i+=circle_step){
         z2=PI*(double)i/180; x2=x+(long)(d*cos(z2)); y2=y+(long)(d*sin(z2));
         CreateVertex(); v1=MainVp;
         CreateVertex(); v2=MainVp;
         v1->xyz[0]=x2; v1->xyz[1]=y2; v1->xyz[2]=z;
         v2->xyz[0]=x2; v2->xyz[1]=y2; v2->xyz[2]=z+height;
         CreateEdge(v1,v2,0);
         CreateEdge(v1,vc1,0); CreateEdge(v2,vc2,0);   /* bottom top */
         if(i == 0){v1s=v1; v2s=v2;}
         else{
          CreateEdge(v1,v1l,0); CreateEdge(v2,v2l,0); CreateEdge(v1l,v2,0);
          AddDXFface(v1l,v1,v2,0); AddDXFface(v2,v2l,v1l,0);
          AddDXFface(v1,vc1,v1l,0); AddDXFface(v2,vc2,v2l,0); /* B/T */
         }
         v1l=v1; v2l=v2;
       }
       if(v1s != NULL && v2s != NULL){
         CreateEdge(v1,v1s,0); CreateEdge(v2,v2s,0); CreateEdge(v1,v2s,0);
         AddDXFface(v1,v1s,v2s,0); AddDXFface(v2s,v2,v1l,0);
         AddDXFface(v1s,vc1,v1,0); AddDXFface(v2s,vc2,v2,0);   /* B/T */
       }
     }
     else { /* arc */
       if(ea < sa)ea += 360;
       i=sa; for(;;){
         LOOPARC1:
         z2=PI*(double)i/180; x2=x+(long)(d*cos(z2)); y2=y+(long)(d*sin(z2));
         CreateVertex(); v1=MainVp;
         CreateVertex(); v2=MainVp;
         v1->xyz[0]=x2; v1->xyz[1]=y2; v1->xyz[2]=z;
         v2->xyz[0]=x2; v2->xyz[1]=y2; v2->xyz[2]=z+height;
         CreateEdge(v1,v2,0);
         if(i != sa){
          CreateEdge(v1,v1l,0); CreateEdge(v2,v2l,0); CreateEdge(v1l,v2,0);
          AddDXFface(v1l,v1,v2,0); AddDXFface(v2,v2l,v1l,0);
         }
         v1l=v1; v2l=v2;
         i += circle_step; if(i > ea)break;
       }
       if(i < ea+circle_step){i=ea; goto LOOPARC1;}
     }
     if(current_sp != NULL){
        vp=MainVp; while(vp != vl){
          vp->sp=current_sp;
          vp=vp->last;
        }
     }
   }
 }
 else if(inc2d && !current_freeze){ /* 2d circle or arc */
//X__pri("circle or arc on layer [%s] %ld",CurrentLayer,current_freeze);
   if(fill == 0){ /* arc */
     if(ea < sa)ea += 360;
     i=sa; for(;;){
       LOOPARC:
       z2=PI*(double)i/180; x2=x+(long)(d*cos(z2)); y2=y+(long)(d*sin(z2));
       CreateVertex();
       MainVp->xyz[0]=x2; MainVp->xyz[1]=y2; MainVp->xyz[2]=z;
       if(i != sa)CreateEdge(MainVp,v1,0);
       v1=MainVp;
       i += circle_step; if(i > ea)break;
     }
     if(i < ea+circle_step){i=ea; goto LOOPARC;} /* make sure arc ends EA */
   }
   else{  /* full circle */
     for(i=0;i<360;i+=circle_step){
       z2=PI*(double)i/180; x2=x+(long)(d*cos(z2)); y2=y+(long)(d*sin(z2));
       CreateVertex(); v1=MainVp;
       v1->xyz[0]=x2; v1->xyz[1]=y2; v1->xyz[2]=z;
       if(i == 0)v1s=v1;
       else CreateEdge(v1,v1l,0);
       v1l=v1;
     }
     if(v1s != NULL)CreateEdge(v1,v1s,0);
   }
   if(current_sp != NULL){
     vp=MainVp; while(vp != vl){
       vp->sp=current_sp;
       vp=vp->last;
     }
   }
 }
 if(dx != 0.0 || dy != 0.0 || dz != 1.0){
   ReOrientate(dx,dy,dz,MainVp,vl);
//   ReOrientate2(dx,dy,dz,x,y,z,MainVp,vl);
 }
 fseek(dxfp,offset,SEEK_SET); /* go back to get the next entity */
 return 0;
}

static DXFblock *WhatBlock(char *name){
 long i;
 if(nDXFblocks > 0 && DXFblocks != NULL){
   for(i=0;i<nDXFblocks;i++){
     if(strcmp(DXFblocks[i].name,name) == 0)return &DXFblocks[i];
   }
 }
 return NULL;
}

static short CreateEntity(FILE *dxfp,DXFblock *blk,char *entity,skel *sp){
 if     (strncmp(entity,"POLYLINE",8) == 0 &&
         CreateDXFpolyline(dxfp,blk) < 0)return -1;
 else if(strncmp(entity,"LINE",4) == 0 &&
         CreateDXFline(dxfp,blk) < 0)return -1;
 else if(strncmp(entity,"CIRCLE",6) == 0 &&
         CreateDXFcircle(dxfp,blk,1) < 0)return -1;
 else if(strncmp(entity,"ARC",3) == 0 &&       /*  arc is like circle    */
         CreateDXFcircle(dxfp,blk,0) < 0)return -1;
 else if(strncmp(entity,"TRACE",5) == 0 &&      /* trace is same as solid */
         CreateDXFsolid(dxfp,blk) < 0)return -1;
 else if(strncmp(entity,"SOLID",5) == 0 &&
         CreateDXFsolid(dxfp,blk) < 0)return -1;
 else if(strncmp(entity,"3DFACE",6) == 0 &&
         CreateDXF3dface(dxfp,blk) < 0)return -1;
 else if(strncmp(entity,"INSERT",6) == 0 &&
         InsertDXFblock(dxfp,sp) < 0)return -1;
 return 0;
}

static short InsertDXFblock(FILE *dxfp,skel *parent){
 char dstr[260]; double di; long vi,gc;
 char block[32],layer[32];
 long offset,x=0,y=0,z=0,j;
 double theta,ctheta,stheta,rot=0.0,sx=1.0,sy=1.0,sz=1.0,
        dx,dy,dz,tx,ty,edx=0.0,edy=0.0,edz=1.0;
 DXFblock *blk;
 vertex *vp,*ibs;
 skel *sp;
 for(;;){
   offset=ftell(dxfp);
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0)break;  /* next entity */
   else if(gc == 2)strcpy(block,dstr);
   else if(gc == 8)SetLayerCurrents(dstr);
   else if(gc == 10)x=(long)(di*weld_mult);
   else if(gc == 20)y=(long)(di*weld_mult);
   else if(gc == 30)z=(long)(di*weld_mult);
   else if(gc == 50)rot=di;
   else if(gc == 41)sx=di;
   else if(gc == 42)sy=di;
   else if(gc == 43)sz=di;
   else if(gc == 210)edx=di;
   else if(gc == 220)edy=di;
   else if(gc == 230)edz=di;
 }
 /* insert the BLOCK */
 if(!current_freeze && (blk=WhatBlock(block)) != NULL && !blk->frozen){
   if(skel_type == 2){ /* create a child */
     sp=parent;
     CreateSkeleton(parent);
     j=min(12,strlen(block));
     strncpy(MainSp->name,block,j); *(MainSp->name+j)=0;
     MainSp->xyz[2]=MINUNIT*2*Nskel;
     current_sp=MainSp;
   }
   if(verbose > 1){
     X__pri("Inserting [%s] at %lf %lf %lf  @ %lf",blk->name,
      (double)x*weld_threshold,(double)y*weld_threshold,(double)z*weld_threshold,rot);
     X__pri("Scaling %lf %lf %lf",sx,sy,sz);
     X__pri("Block origin %lf %lf %lf",(double)blk->x*weld_threshold,
        (double)blk->y*weld_threshold,(double)blk->z*weld_threshold);
   }
   fseek(dxfp,blk->offset,SEEK_SET); /* go to start of block */
   blk->vs=MainVp; blk->es=MainEp; ibs=MainVp;
   theta=rot*0.017453292; ctheta=cos(theta); stheta=sin(theta);
   for(;;){
     if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
     if(gc == 0 && strncmp(dstr,"ENDBLK",6) == 0)break;
     else if(gc == 0){
       if(CreateEntity(dxfp,blk,dstr,current_sp) < 0)return -1;
     }
   }
   /* move block into place */
   vp=MainVp; while(vp != ibs){
     dx = (double)(vp->xyz[0] - blk->x);  dx = dx*sx;
     dy = (double)(vp->xyz[1] - blk->y);  dy = dy*sy;
     dz = (double)(vp->xyz[2] - blk->z);  dz = dz*sz;
     if(rot != 0.0){
       tx=(ctheta*dx-stheta*dy);
       ty=(stheta*dx+ctheta*dy);
       dx=tx; dy=ty;
     }
     vp->xyz[0] = x + (long)dx;
     vp->xyz[1] = y + (long)dy;
     vp->xyz[2] = z + (long)dz;
     vp=vp->last;
   }
   if(edx != 0.0 || edy != 0.0 || edz != 1.0){
     ReOrientate(edx,edy,edz,MainVp,ibs);
   }
   /* block now in place */
   if(skel_type == 2)current_sp=sp;
 }
 /* XXXX */
 fseek(dxfp,offset,SEEK_SET); /* go back to get the next entity */
 return 0;
}

static short CreateBlockHeader(FILE *dxfp, char *name,
                            double *bx, double *by, double *bz){
 char dstr[260]; double di; long vi,gc;
 DXFblock *db;
 long j,offset;
 skel *sp=NULL;
 *bx = *by = *bz = 0.0;
 for(;;){
   offset=ftell(dxfp);
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0){ /* next entity so break out */
     fseek(dxfp,offset,SEEK_SET); /* go back so that next entity is ready */
     break;
   }
   else if(gc ==  2)strcpy(name,dstr);
   else if(gc == 10) *bx=di;
   else if(gc == 20) *by=di;
   else if(gc == 30) *bz=di;
 }
 db=DXFblocks;
 if(db == NULL)db=(DXFblock *)malloc(sizeof(DXFblock));
 else db=(DXFblock *)realloc(db,(nDXFblocks+1)*sizeof(DXFblock));
 if(db == NULL){
   if(DXFblocks != NULL)free(DXFblocks);
   DXFblocks=NULL; nDXFblocks=0;
   return -1;
 }
 DXFblocks=db;
 DXFblocks[nDXFblocks].x = (long)((*bx)*weld_mult);
 DXFblocks[nDXFblocks].y = (long)((*by)*weld_mult);
 DXFblocks[nDXFblocks].z = (long)((*bz)*weld_mult);
 DXFblocks[nDXFblocks].vs=NULL;
 DXFblocks[nDXFblocks].es=NULL;
 DXFblocks[nDXFblocks].frozen=0;
 strcpy(DXFblocks[nDXFblocks].name,name);
 DXFblocks[nDXFblocks].offset=ftell(dxfp);
 nDXFblocks++;
 return 0;
}

static short ReadDXFblocks(FILE *dxfp){
 char dstr[260]; double di; long vi,gc,confirm;
 char block_name[32];
 double bx=0,by=0,bz=0;
 for(;;){
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0 && strncmp(dstr,"ENDSEC",6) == 0)break;
   if(gc == 0 && strncmp(dstr,"BLOCK",5) == 0){
     UpdateProgress(dxfp);
     if(CreateBlockHeader(dxfp,block_name,&bx,&by,&bz) < 0)return -1;
     if(verbose > 0 || b_freeze)X__pri("Block [%s]",block_name);
     for(;;){
       if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
       if(gc == 0 && strncmp(dstr,"ENDBLK",6) == 0)break;
     }
     if(verbose > 0)X__pri(" at %lf %lf %lf",
              (double)DXFblocks[nDXFblocks-1].x * weld_threshold,
              (double)DXFblocks[nDXFblocks-1].y * weld_threshold,
              (double)DXFblocks[nDXFblocks-1].z * weld_threshold);
//     if(b_freeze){
//       if(confirm == 'Y' || confirm == 'y')DXFblocks[nDXFblocks-1].frozen=1;
//       if(DXFblocks[nDXFblocks-1].frozen)X__pri("  now FROZEN ");
//     }
   }
 }
 return 0;
}

static short ReadDXFentities(FILE *dxfp){
 char dstr[260]; double di; long vi,gc;
 vertex *vp,*EntityStart;
 long i,dmax,dmin;
 double scale,scan;
 DXFblock EntityBlock;
 if(verbose > 0)X__pri("Entities");
 EntityBlock.offset=ftell(dxfp);   /* this is needed for welding */
 EntityBlock.x = 0;                /* entities in the Entities   */
 EntityBlock.y = 0;                /* section only               */
 EntityBlock.z = 0;
 EntityBlock.vs=MainVp;
 EntityBlock.es=MainEp;
 strcpy(EntityBlock.name,"NULL");
 strcpy(CurrentLayer,"Dummy-RSF"); current_layer = -1; current_nought = 0;
 EntityStart=MainVp; /* vertices after this are DXF */
 current_freeze=0;
 if(skel_type == 2)current_sp=FirstSp; else current_sp=NULL;
 for(;;){
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)return -1;
   if(gc == 0 && strncmp(dstr,"ENDSEC",6) == 0)break;
   else if(gc == 0){
     UpdateProgress(dxfp);
     if(CreateEntity(dxfp,&EntityBlock,dstr,FirstSp) < 0)return -1;
   }
 }
 /* scale the model */
 dmin = MAXUNIT; dmax = -MAXUNIT;
 dmin *= 256; dmax *= 256;
 vp=MainVp; while(vp != EntityStart){
   if(vp->xyz[0] < dmin)dmin=vp->xyz[0];
   if(vp->xyz[1] < dmin)dmin=vp->xyz[1];
   if(vp->xyz[2] < dmin)dmin=vp->xyz[2];
   if(vp->xyz[0] > dmax)dmax=vp->xyz[0];
   if(vp->xyz[1] > dmax)dmax=vp->xyz[1];
   if(vp->xyz[2] > dmax)dmax=vp->xyz[2];
   vp=vp->last;
 }
 scale=(double)(dmax-dmin);
 scale=(double)(UNIT2*2)/(double)scale*100.0;

 if(dlg != NULL)DestroyWindow(dlg); dlg=NULL;
 scan=scale;
 if(DialogBoxParam(hInstance,MAKEINTRESOURCE(DLG_SCALE),hWndParent,
              (DLGPROC)DlgProcScale,(LPARAM)&scan)){
   scale=scan;
 }
 dlg=CreateDialog(hInstance,MAKEINTRESOURCE(DLGWRIT),hWndParent,(DLGPROC)DlgProc);
 scale /= 100.0;
 vp=MainVp; while(vp != EntityStart){
   for(i=0;i<3;i++)vp->xyz[i] = (long)((double)vp->xyz[i] * scale);
   vp=vp->last;
 }
 if(verbose > 0)X__pri("End Entities");
 return 0;
}

short DxfLoad(char *filename,short skl,short l_fr,short b_fr, short report,
              double w_t,short nw,short d2,short d3, short cs){
 short error=0;
 char dstr[260]; double di; long vi,gc;
 FILE *dxfp;
 int fhandle;
 skel_type=skl; l_freeze=l_fr; b_freeze=b_fr; verbose=report; circle_step=cs;
 if(w_t > 0.0)weld_threshold=w_t; no_weld=nw; inc2d=d2; inc3d=1;
 nDXFblocks=0; nDXFlayers=0;
 current_colour=7; current_layer = -1; current_nought = 0;
 current_freeze=0; current_sp=NULL;
 if((fhandle=open(filename,O_RDONLY)) >= 0){
   struct stat buffer;
   if(fstat(fhandle,&buffer) == 0){
     file_size=buffer.st_size;
   }
   else MessageBox(NULL,"File size",NULL,MB_OK);
   close(fhandle);
 }
 else MessageBox(NULL,"Open for size",NULL,MB_OK);
 if((dxfp=fopen(filename,"rb")) != NULL){
   weld_mult=1.0/weld_threshold;
   SetupDXFcolours();
   LOOP:
   UpdateProgress(dxfp);
   if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)goto NOTOK;
   if(gc == 0 && strncmp(dstr,"EOF",3) == 0)goto CLOSE;
   if(gc == 0 && strncmp(dstr,"SECTION",7) == 0){
     UpdateProgress(dxfp);
     if((gc=getDXFcode(dxfp,&vi,&di,dstr)) < 0)goto NOTOK;
     if     (gc == 2 && strncmp(dstr,"HEADER",6) == 0 &&
        ReadDXFheader(dxfp) < 0)goto NOTOK;
     else if(gc == 2 && strncmp(dstr,"TABLES",6) == 0  &&
        ReadDXFtables(dxfp) < 0)goto NOTOK;
     else if(gc == 2 && strncmp(dstr,"BLOCKS",6) == 0  &&
        ReadDXFblocks(dxfp) < 0)goto NOTOK;
     else if(gc == 2 && strncmp(dstr,"ENTITIES",8) == 0  &&
        ReadDXFentities(dxfp) < 0)goto NOTOK;
   }
   goto LOOP;
   NOTOK:
   error=2;
   CLOSE:
   UpdateProgress(dxfp);
   fclose(dxfp);
 }
 else{
   error=1;
 }
 return error;
}
