/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* File DXFOUT.C DLL to create a DXF from EVI structure */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <windows.h>

#include "struct.h"
#include "dstruct.h"

#define MAXUNIT    4194304L
#define DESELECTED 0
#define SELECTED   1
#define INEDITOR   3

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HWND      hParent;
static HINSTANCE hThisInstance;

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
//      MessageBox (GetFocus(),"Attaching","Debug",MB_OK);
      hThisInstance=hDLL;
      break;
    }
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
//      MessageBox (GetFocus(),"Detaching","Debug",MB_OK);
      break;
  }
  return TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

long _Export
 (HWND parent_window, char *filename,X__STRUCTURE *lpevi){
 char layer[8]={"MODEL"},cl[16];
 FILE *dxfp;
 vertex *Vp,*V0,*V1,*V2;
 face   *Fp;
 skel   *sp;
 long   vp,fp;
 long i,nf,colour=7;
 long xmax = -MAXUNIT, ymax = -MAXUNIT, zmax = -MAXUNIT,
      xmin =  MAXUNIT, ymin =  MAXUNIT, zmin =  MAXUNIT;
 long lrulerx,lrulery,lrulerz;
 double ruler;
 lpEVI=lpevi;
 if(filename == NULL)return 0;
 if(NvertSelect == 0 || Nface == 0){
   MessageBox(parent_window,"Too few vertices or faces selected","DXF Write",MB_OK);
   return 0;
 }
 lrulerx= (*(lpEVI->orulerx));
 lrulery= (*(lpEVI->orulery));
 lrulerz= (*(lpEVI->orulerz));
 ruler  = (*(lpEVI->ruler));
 if((dxfp=fopen(filename,"w")) != NULL){
   vp=0; while(vp < Nvert){
     Vp=(MainVp+vp);
     if(Vp->status == SELECTED){
       if(Vp->xyz[0] > xmax)xmax=Vp->xyz[0];
       if(Vp->xyz[1] > ymax)ymax=Vp->xyz[1];
       if(Vp->xyz[2] > zmax)zmax=Vp->xyz[2];
       if(Vp->xyz[0] < xmin)xmin=Vp->xyz[0];
       if(Vp->xyz[1] < ymin)ymin=Vp->xyz[1];
       if(Vp->xyz[2] < zmin)zmin=Vp->xyz[2];
     }
     vp++;
   }
   fprintf(dxfp,"0\nSECTION\n2\nHEADER\n9\n$ACADVER\n1\nAC1008\n9\n");
   fprintf(dxfp,"$UCSORG\n10\n0.0\n20\n0.0\n30\n0.0\n9\n$UCSXDIR\n");
   fprintf(dxfp,"10\n1.0\n20\n0.0\n30\n0.0\n9\n$TILEMODE\n70\n1\n9\n");
   fprintf(dxfp,"$UCSYDIR\n10\n0.0\n20\n1.0\n30\n0.0\n");
   fprintf(dxfp,"9\n$EXTMIN\n10\n%lf\n20\n%lf\n30\n%lf\n",
     (double)(xmin-lrulerx)/ruler,
     (double)(ymin-lrulery)/ruler,
     (double)(zmin-lrulerz)/ruler);
   fprintf(dxfp,"9\n$EXTMAX\n10\n%lf\n20\n%lf\n30\n%lf\n",
     (double)(xmax-lrulerx)/ruler,
     (double)(ymax-lrulery)/ruler,
     (double)(zmax-lrulerz)/ruler);
   fprintf(dxfp,"0\nENDSEC\n0\nSECTION\n2\nTABLES\n0\nTABLE\n2\nLTYPE\n");
   fprintf(dxfp,"70\n1\n0\nLTYPE\n2\nCONTINUOUS\n70\n64\n3\nSolid line\n");
   fprintf(dxfp,"72\n65\n73\n0\n40\n0.0\n0\nENDTAB\n0\nTABLE\n2\nLAYER\n");
   fprintf(dxfp,"70\n1\n0\nLAYER\n2\n%s\n70\n0\n62\n7\n6\nCONTINUOUS\n",layer);
   if((sp=MainSp) != NULL) while(sp != NULL){
     fprintf(dxfp,"0\nLAYER\n2\n%s\n70\n0\n62\n%ld\n6\nCONTINUOUS\n",
             sp->name,colour);
     sp=sp->last;
   }
   fprintf(dxfp,"0\nENDTAB\n0\nTABLE\n2\nSTYLE\n70\n1\n0\n");
   fprintf(dxfp,"STYLE\n2\nSTANDARD\n70\n0\n40\n0.0\n41\n1.0\n50\n0.0\n");
   fprintf(dxfp,"71\n0\n42\n0.2\n3\ntxt\n4\n  \n0\nENDTAB\n0\nTABLE\n");
   fprintf(dxfp,"2\nUCS\n70\n0\n0\nENDTAB\n0\nENDSEC\n");
   fprintf(dxfp,"0\nSECTION\n2\nENTITIES\n");
   sp=MainSp; for(;;){ /* sp will be null for vertices not on any node */
     if(sp != NULL)strcpy(cl,sp->name);
     else          strcpy(cl,layer);
     vp=0; while(vp < Nvert){
       Vp=(MainVp+vp);
       if(Vp->status == SELECTED && Vp->sp == sp)Vp->status=INEDITOR;
       vp++;
     }
     nf=0; fp=0; while(fp < Nface){
       Fp=(MainFp+fp);
       V0=(MainVp+Fp->V[0]); V1=(MainVp+Fp->V[1]); V2=(MainVp+Fp->V[2]);
       if(V0->status == INEDITOR && V1->status == INEDITOR &&
          V2->status == INEDITOR)nf++;
       fp++;
     }
     if(nf > 0){ /* only write if there are any faces */
       fprintf(dxfp,"0\nPOLYLINE\n8\n%s\n",cl);
       fprintf(dxfp,"66\n1\n70\n64\n71\n%ld\n72\n%ld\n",NvertSelect,nf);
       i=1; vp=0; while(vp < Nvert){
         Vp=(MainVp+vp);
         if(Vp->status == INEDITOR){
           fprintf(dxfp,"0\nVERTEX\n8\n%s\n10\n%lf\n20\n%lf\n30\n%lf\n",cl,
              (double)(Vp->xyz[0]-lrulerx)/ruler,
              (double)(Vp->xyz[1]-lrulery)/ruler,
              (double)(Vp->xyz[2]-lrulerz)/ruler);
           fprintf(dxfp,"70\n192\n");
           Vp->id = i++;
         }
         else Vp->id = -1;
         vp++;
       }
       fp=0; while(fp < Nface){
         Fp=(MainFp+fp);
         V0=(MainVp+Fp->V[0]); V1=(MainVp+Fp->V[1]); V2=(MainVp+Fp->V[2]);
         if(V0->status == INEDITOR && V1->status == INEDITOR &&
            V2->status == INEDITOR){
           fprintf(dxfp,"0\nVERTEX\n8\n%s\n10\n0\n20\n0\n30\n0\n70\n128\n",
                   cl);
           fprintf(dxfp,"71\n%d\n72\n%d\n73\n%d\n",(short)(V0->id),
                   (short)(V1->id),(short)(V2->id));
         }
         fp++;
       }
       fprintf(dxfp,"0\nSEQEND\n");
     }
     vp=0; while(vp < Nvert){ /* reset status */
       Vp=(MainVp+vp);
       if(Vp->status == INEDITOR)Vp->status=SELECTED;
       vp++;
     }
     if(sp == NULL)break;
     sp=sp->last;
   }  /* end of for(;;) */
   fprintf(dxfp,"8\n%s\n0\nENDSEC\n0\nEOF\n",layer);
   fclose(dxfp);
 }
 else{
   MessageBox(parent_window,"File open Error","DXF Write",MB_OK);
 }
 return 1;
}
