/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* File EXAMPLE.C DLL template to create an external modeler */

/* This trivial example can serve as a template for building external */
/* modelers. It shows the Entry point called by the                   */
/* Modeler and the include files that define the accessible           */
/* data structures of the modeler and defines "Function callbacks" to */
/* functions in the modeler. Other examples will show the use of the  */
/* structures                                                         */

/* include any other C library header files needed here               */
#include <windows.h>

/* include the definitions of the Data structures                     */
#include "struct.h"

/* include defines for function names and global variables            */
#include "dstruct.h"

/* include external modeler defines eg. Dialog Box ID's               */
#include "example.h"

/* Definition needed for compilation on MIPS machines                 */
#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static BOOL CALLBACK ExampleDlgProc(HWND hwnd, UINT msg,
                                 WPARAM wparam, LPARAM lparam);

static HWND      hParent;
static HINSTANCE hThisInstance;


/* Main Entry point, called as Process attaches and detached          */
/* remove debug MessageBox statement as required                      */
#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=hDLL;
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      MessageBox ( GetFocus(),
                  (LPCTSTR) "Process attaching",
                  (LPCTSTR) " ",
                  MB_OK | MB_SYSTEMMODAL);
      break;
    }
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      MessageBox ( GetFocus(),
                  (LPCTSTR) "Process detaching",
                  (LPCTSTR) "",
                  MB_OK | MB_SYSTEMMODAL );
      break;
  }
  return TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

/* Function Called From the modeler                                  */

BOOL _Xmodeler
 /* parameters                                                        */
 ( HWND parent_window,     /* parent window of caller                 */
   HWND info_window,       /* send WM_SETTEXT  see StatusBar class    */
   X__STRUCTURE *lpevi    /* pointer to database structure           */
 ){


 /* assign a static global variable that is used in DEFINE's for      */
 /* calls to functions and access to global variables                 */
 /* this name must NOT be changed lpEVI is typed in DSTRUCT.H         */

 lpEVI=lpevi;

 /* This is where the Modeler action takes place                      */
 /* can be almost anything required                                   */
 if(DialogBox(hThisInstance,MAKEINTRESOURCE(DLG_EXAMPLE),GetFocus(),
           (DLGPROC)ExampleDlgProc) != OK)return FALSE;
 /* end of action                                                     */

 /* return TRUE if functions are successful FALSE otherwise           */
 return TRUE;
}


/* this is a very trivial Dialog Box that just basically just has     */
/* a cancel button. It is NOT a required part of an external modeler  */

static BOOL CALLBACK ExampleDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 int i;
 BOOL err;
 switch( msg ) {
   case WM_INITDIALOG:
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_EXAMPLE_CANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case DLG_EXAMPLE_OK:
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

