/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file GLDESIGN.C  The DLL that provided OpenGL support for the designer */

#include <windows.h>

#include <gl\gl.h>
#include <gl\glu.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "struct.h"
#include "dstruct.h"
#include "glmenu.h"
#include "gldesign.h"
#include "..\common\animate\about.h"

#define PLANE            0
#define CYLINDER         1
#define PLANE_MOZIAC     2
#define CYLINDER_MOZIAC  3
#define MAP_BY_VERTEX    4
#define MAP_SPHERICAL    5

#define MAX_FILE       255
#define MAX_DIR        255

typedef struct tagGLMAP {
 unsigned char *pixels;
 long x,y;
}
GLMAP;

#define IDM_WINDOW_SHOWOPENGL            17130

#define StartWait() hcurSave = SetCursor(LoadCursor(NULL,IDC_WAIT))
#define EndWait()   SetCursor(hcurSave)

#define WM_INIT       (WM_USER+1)
#define WM_UPDATE     (WM_USER+2)
#define WM_UPDATE_MAP (WM_USER+3)
#define WM_LOAD_MAPS  (WM_USER+4)

#define WM_SHOWMAP    (WM_USER+6)

typedef GLfloat normal[3];

void RegisterOpenGL(HINSTANCE);
void UnregisterOpenGL(HINSTANCE);

static BOOL normalise(GLfloat *);
static BOOL Normalize(point, point, point, GLfloat *);
static void Make3dListF(void);
static void Make3dListS(void);
static void drawModelE(void);
static void drawModelF(void);
static void DrawGlNurbs(GLfloat);
static void drawModelMaps(void);
static GLvoid initialize(HWND);
static GLvoid drawScene(HWND,BOOL);
static GLvoid resize(HWND);
static BOOL bSetupPixelFormat(HDC);
static LONG WINAPI OpenGLWndProc(HWND, UINT, WPARAM, LPARAM);
static void LoadPrefs(void);
static PSTR FileInPath(PSTR pstrPath);
static BOOL CALLBACK PrefsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam);
static void GluCommands(WORD, HWND);
static SetMappingQuality(void);
static void makeImageMap(int id);
static void GetMappingCoordP(int id, vector n, vector x, vector y, vector p0,
                             long *pv,
                             GLfloat *a, GLfloat *b);
static void GetMappingCoordC(int id, vector n, vector x, vector y, vector p0,
                long *pv,
                GLfloat *a, GLfloat *b);
static void GetMapNormal(int k, vector mP, vector mX, vector mY,
                         vector mN);
static void BuildMapList(void);
static void FreeMapList(void);
static unsigned char *LoadDIB(char *temp, long *x, long *y);
static unsigned char *LoadGIF(char *temp, long *x, long *y);
static unsigned char *LoadTGA(char *temp, long *x, long *y);
static unsigned char *LoadXXX(char *temp, long *x, long *y, long type);

#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))

#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define VECCOPY(a,c)    { c[0]=a[0]; c[1]=a[1]; c[2]=a[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }

char IniFilename[] = "OPENFX_.INI",
     IniSection[]  = "GLPREFERENCES";
static GLMAP     *glMapList=NULL;
static long      NglMapList=0;
static HMENU     glMenu=NULL;
static HWND      glWnd=NULL,glWndParent;
static HINSTANCE glInstance;
static CHAR      szAppName[]="OFX:OpenGL";
static HGLRC     hRC;
static HCURSOR   hcurSave=NULL;
HPALETTE         ghpalOld, ghPalette = NULL;
RECT             oldrect;
//static DWORD     fdwStyle1 = WS_POPUP | WS_CAPTION | WS_SYSMENU |
//                            WS_MAXIMIZEBOX | WS_THICKFRAME |
//                            WS_VSCROLL | WS_HSCROLL |
//                            WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
static DWORD     fdwStyle = WS_CHILD | !WS_CAPTION | !WS_SYSMENU |
                            !WS_MAXIMIZEBOX | !WS_THICKFRAME | WS_BORDER |
//                            WS_VSCROLL | WS_HSCROLL |
                            WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
static double  *PerspTheta,*PerspFi;
static GLfloat up_angle=0.0,round_angle=0.0;
static GLfloat gbRed=0.0,gbGreen=0.1,gbBlue=0.3;
//static GLfloat gwRed=1.0,gwGreen=0.5,gwBlue=0.0;
static GLfloat gwRed=1.0,gwGreen=1.0,gwBlue=1.0;
static GLfloat near_plane,far_plane;
static GLint listID=0,mode=1,flat=0,bounding_box=0,current_map=0;
static BOOL  Mapped=FALSE,Update=TRUE,Fastest=FALSE,Clamped=FALSE,
             Cmaterial=TRUE,AttenuateLight=TRUE,TwoSided=TRUE,
             ScrollBars=FALSE,bDrawMapping=FALSE,bOverlayWire=FALSE,
             bThickWire=FALSE,bNurbsWireOnly=FALSE;
static char gszBMPfile[MAX_FILE],gszBMPdir[MAX_DIR];
unsigned char *pixels = NULL;

#define checkImageWidth 64
#define checkImageHeight 64
static GLubyte checkImage[checkImageWidth][checkImageHeight][3];


#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BCC__
int APIENTRY DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
      RegisterOpenGL(hDLL);
      break;
    case DLL_PROCESS_DETACH:
      UnregisterOpenGL(hDLL);
      break;
  }
  return TRUE;
}

void UnregisterOpenGL(HINSTANCE hInstance){
 if(glWnd != NULL)DestroyWindow(glWnd);
 UnregisterClass(szAppName,hInstance);
}

void RegisterOpenGL(HINSTANCE hInstance){
 WNDCLASS   wc;
 wc.style         = 0;
 wc.lpfnWndProc   = (WNDPROC)OpenGLWndProc;
 wc.cbClsExtra    = 0;
 wc.cbWndExtra    = 0;
 wc.hInstance     = hInstance;
 wc.hIcon         = LoadIcon(hInstance,"GLVIEW_ICON");
// wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
 wc.hCursor       = LoadCursor(hInstance,MAKEINTRESOURCE(IDC_HAND));
 wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
 wc.lpszMenuName  = NULL;
 wc.lpszClassName = szAppName;
 glInstance=hInstance;
 if (!RegisterClass (&wc) )return;
 return;
}

HWND _AttachViewOpenGL(
     HWND parent,
     char *tt,
     LPRECT pr,
     double *lpPerspTheta,
     double *lpPerspFi,
     X__STRUCTURE *lpevi
                              ){
 RECT rc;
 HWND hWnd;
 lpEVI=lpevi;
 glWndParent=parent;
 PerspTheta=lpPerspTheta;
 PerspFi=lpPerspFi;
 GetModuleFileName(glInstance,gszBMPdir,MAX_DIR);
 *FileInPath(gszBMPdir) = '\0';
 gszBMPfile[0]='\0'; strcat(gszBMPdir,"maps");
 hWnd = CreateWindow(szAppName,
             "View:OpenGL",
             fdwStyle,
             pr->left,
             pr->top,
             pr->right,
             pr->bottom,
             parent,
             NULL,
             glInstance,
             NULL);
 if (!hWnd)return NULL;
 SendMessage(hWnd, WM_INIT, 0, 0L);
 ShowWindow (hWnd, SW_SHOW);
 UpdateWindow(hWnd);
 glWnd=hWnd;
 return hWnd;
}

void _DetachViewOpenGL(void){
 if(glWnd != NULL)DestroyWindow(glWnd);
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static LONG WINAPI OpenGLWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam){
 GLfloat old;
 int i,iPos,iMax,iMin,dn;
 LONG    lRet = 1;
 static BOOL flag=FALSE,mmoved=FALSE,bTrackPan=FALSE;
 static long xs,ys;
 switch (uMsg){
//   case WM_NCACTIVATE:
//     return TRUE;
//   case WM_KEYDOWN:
//     SetActiveWindow(glWndParent);
//     break;
   case WM_CREATE:{
       INT menu_flags;
       HDC hDC;
       HMENU hmenu;
       char str[128];
       hDC = GetDC(hWnd);
       bSetupPixelFormat(hDC);
       hRC = wglCreateContext( hDC );
       wglMakeCurrent( hDC, hRC );
       LoadPrefs();
       up_angle=*PerspTheta;
       round_angle=*PerspFi;
       if(ScrollBars){
         SetScrollRange(hWnd, SB_VERT, -89,89,TRUE);
         SetScrollRange(hWnd, SB_HORZ, -179,179,TRUE);
         SetScrollPos(hWnd,SB_HORZ,(int)round_angle,TRUE);
         SetScrollPos(hWnd,SB_VERT,(int)up_angle,TRUE);
       }
       else{
         SetScrollRange(hWnd, SB_VERT, 0,0,TRUE);
         SetScrollRange(hWnd, SB_HORZ, 0,0,TRUE);
         SetScrollPos(hWnd,SB_HORZ,0,TRUE);
         SetScrollPos(hWnd,SB_VERT,0,TRUE);
       }
       hmenu=GetMenu(glWndParent);
       glMenu=CreatePopupMenu();
       LoadString(glInstance,IDS_HINTS,str,127);
       AppendMenu(glMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_MAP,str);
       menu_flags=MF_ENABLED | MF_STRING;
       if(TwoSided)menu_flags |= MF_CHECKED;
       LoadString(glInstance,IDS_SIDES,str,127);
       AppendMenu(glMenu,menu_flags,IDM_GLVIEW_TWOSIDE,str);
       menu_flags=MF_ENABLED | MF_STRING;
       if(ScrollBars)menu_flags |= MF_CHECKED;
       LoadString(glInstance,IDS_SCROLL,str,127);
       AppendMenu(glMenu,menu_flags,IDM_GLVIEW_SCROLLBARS,str);

       menu_flags=MF_ENABLED | MF_STRING;
       if(bOverlayWire)menu_flags |= MF_CHECKED;
       LoadString(glInstance,IDS_OVERLAY,str,127);
       AppendMenu(glMenu,menu_flags,IDM_GLVIEW_OVERLAY,str);

       LoadString(glInstance,IDS_PREF,str,127);
       AppendMenu(glMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_PREFS,str);
       AppendMenu(glMenu,MF_SEPARATOR,0,0);
       LoadString(glInstance,IDS_HELP,str,127);
       AppendMenu(glMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_HELP,str);
       LoadString(glInstance,IDS_ABOUT,str,127);
       AppendMenu(glMenu,MF_ENABLED | MF_STRING,IDM_GLVIEW_ABOUT,str);
       LoadString(glInstance,IDS_MENU,str,127);
       InsertMenu(hmenu,GLMENU_POSITION,MF_POPUP|MF_BYPOSITION,
                 (UINT)glMenu,str);
       DrawMenuBar(glWndParent);
     }
     break;
   case WM_PAINT:{
       HDC hDC;
       PAINTSTRUCT  ps;
       hDC = BeginPaint(hWnd, &ps);
       drawScene(hWnd,TRUE);
       EndPaint(hWnd, &ps);
     }
     break;
   case WM_LBUTTONDOWN:
     mmoved=FALSE;
     xs=LOWORD(lParam);
     ys=HIWORD(lParam);
     SetCapture(hWnd);
     bTrackPan=TRUE;
     break;
   case WM_MOUSEMOVE:
     if(bTrackPan){
       long dx,dy;
       dx=(short)LOWORD(lParam)-xs; xs=(short)LOWORD(lParam);
       dy=(short)HIWORD(lParam)-ys; ys=(short)HIWORD(lParam);
       if(labs(dx) > 0 || labs(dy) > 0){
         mmoved=TRUE;
         round_angle=round_angle+(double)dx;
         while(round_angle < -180.0)round_angle += 360.0;
         while(round_angle >  180.0)round_angle -= 360.0;
         *PerspFi = (double)round_angle;
         SetScrollPos(hWnd,SB_HORZ,(int)round_angle,TRUE);
         up_angle=up_angle+(double)dy;
         *PerspTheta = (double)up_angle;
         if(up_angle > -90 && up_angle < 90)
           SetScrollPos(hWnd,SB_VERT,(int)up_angle,TRUE);
         if(mode > 0)InvalidateRect(hWnd, NULL, FALSE);
         else        drawScene(hWnd,FALSE);
       }
     }
     break;
   case WM_RBUTTONDOWN:{
       POINT pt;
       HMENU hMenu=LoadMenu(glInstance,"GLViewPopupMenu");
       if(hMenu == NULL)break;
       pt.x=LOWORD(lParam);
       pt.y=HIWORD(lParam);
       ClientToScreen(hWnd,&pt);
       if(TwoSided)CheckMenuItem(hMenu,IDM_GLVIEW_TWOSIDE,MF_CHECKED);
       else        CheckMenuItem(hMenu,IDM_GLVIEW_TWOSIDE,MF_UNCHECKED);
       if(bOverlayWire)CheckMenuItem(hMenu,IDM_GLVIEW_OVERLAY,MF_CHECKED);
       else            CheckMenuItem(hMenu,IDM_GLVIEW_OVERLAY,MF_UNCHECKED);
       if(Mapped)CheckMenuItem(hMenu,IDM_QUICKMAPCHANGE,MF_CHECKED);
       else      CheckMenuItem(hMenu,IDM_QUICKMAPCHANGE,MF_UNCHECKED);
       TrackPopupMenu(GetSubMenu(hMenu,0),TPM_LEFTALIGN | TPM_LEFTBUTTON,
                      pt.x,pt.y,0,hWnd,NULL);
       DestroyMenu(hMenu);
     }
     break;
   case WM_LOAD_MAPS:
     SendMessage(hWnd,WM_COMMAND,IDM_GLVIEW_MAP,0);
     break;
   case WM_SHOWMAP:
     bDrawMapping=(BOOL)wParam;
     current_map=(GLint)lParam;
     InvalidateRect(hWnd, NULL, FALSE);
     break;
   case WM_UPDATE_MAP:
     if(!Mapped)break;
     if(!Update)break;
   case WM_LBUTTONUP:
     if(bTrackPan){
       ReleaseCapture();
       bTrackPan=FALSE;
     }
     if(mmoved){
       mmoved=FALSE;
       if(mode == 0)InvalidateRect(hWnd, NULL, FALSE); // using wireframe
       break;
     }
     mmoved=FALSE;
   case WM_UPDATE:
     if(flat)Make3dListF();
     else    Make3dListS();
     InvalidateRect(hWnd, NULL, FALSE);
     break;
   case WM_INIT:
     initialize(hWnd);
     break;
   case WM_SIZE:
     resize(hWnd);
     lRet=0;
     break;
  case WM_VSCROLL:
    GetScrollRange(hWnd, SB_VERT, &iMin, &iMax);
    iPos = GetScrollPos(hWnd, SB_VERT);
    switch (LOWORD(wParam)) {
      case SB_ENDSCROLL:
         dn=0;
        if(mode == 0 || flag)InvalidateRect(hWnd, NULL, FALSE);
        flag=FALSE;
        break;
      case SB_LINEDOWN:      dn =  1;    break;
      case SB_LINEUP:        dn = -1;    break;
      case SB_PAGEDOWN:      dn =  5;    break;
      case SB_PAGEUP:        dn = -5;    break;
      case SB_THUMBTRACK:
      case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wParam)-iPos;    break;
      default:               dn = 0;
    }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      SetScrollPos(hWnd, SB_VERT, iPos + dn, TRUE);
      old=up_angle;
      up_angle=(GLfloat)(iPos+dn);
      *PerspTheta = (double)up_angle;
      if(old != up_angle){
        if(mode > 0){
          if(!flag && (GetKeyState(VK_CONTROL) & 0x8000))flag=TRUE;
          if(!flag)InvalidateRect(hWnd, NULL, FALSE);
        }
        else drawScene(hWnd,FALSE);
      }
    }
    break;
  case WM_HSCROLL:
    GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
    iPos = GetScrollPos(hWnd, SB_HORZ);
    switch (LOWORD(wParam)) {
      case SB_ENDSCROLL:
        if(mode == 0 || flag)InvalidateRect(hWnd, NULL, FALSE);
        flag=FALSE;
        dn=0;
        break;
      case SB_LINEDOWN:      dn =  1;     break;
      case SB_LINEUP:        dn = -1;     break;
      case SB_PAGEDOWN:      dn =  5;     break;
      case SB_PAGEUP:        dn = -5;     break;
      case SB_THUMBTRACK:
      case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wParam) - iPos; break;
      default:               dn = 0;
    }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      SetScrollPos(hWnd, SB_HORZ, iPos + dn, TRUE);
      old=round_angle;
      round_angle=(GLfloat)(iPos+dn);
      *PerspFi = (double)round_angle;
      if(old != round_angle){
        if(mode > 0){
          if(!flag && (GetKeyState(VK_CONTROL) & 0x8000))flag=TRUE;
          if(!flag)InvalidateRect(hWnd, NULL, FALSE);
        }
        else drawScene(hWnd,FALSE);
      }
    }
    break;
   case WM_QUERYNEWPALETTE:{
       HDC     hDC;
       if(ghPalette != NULL){
         hDC = GetDC(hWnd);
         ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
         RealizePalette(hDC);
         InvalidateRect(hWnd, NULL, TRUE);
         UpdateWindow(hWnd);
         if(ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
         ReleaseDC(hWnd, hDC);
         return TRUE;
       }
       return FALSE;
     }
   case WM_PALETTECHANGED:{
       HDC         hDC;
       if(ghPalette != NULL){
         if (wParam != (WPARAM)hWnd){
           hDC = GetDC(hWnd);
           ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
           RealizePalette(hDC);
           UpdateColors(hDC);
           if(ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
           ReleaseDC(hWnd, hDC);
         }
       }
     }
     break;
   case WM_COMMAND:
     GluCommands(LOWORD(wParam),hWnd);
     break;
   case WM_CLOSE:
     PostMessage(glWndParent,WM_COMMAND,IDM_WINDOW_SHOWOPENGL,0);
     break;
   case WM_DESTROY:{
       RECT    rc;
       POINT   pt1,pt2;
       char *w, yesyes[]="YES", nono[]="NO",str[64];
       HGLRC hRC;
       HDC    hDC;
       FreeMapList();
       hRC = wglGetCurrentContext();
       hDC = wglGetCurrentDC();
       wglMakeCurrent(NULL, NULL);
       if(hRC != NULL)wglDeleteContext(hRC);
       if(hDC != NULL)ReleaseDC(hWnd, hDC);
       glWnd=NULL;
       if(TwoSided)w=yesyes; else w=nono;
       WritePrivateProfileString(IniSection,"TWOSIDED",w,IniFilename);
       if(ScrollBars)w=yesyes; else w=nono;
       WritePrivateProfileString(IniSection,"SCROLLBARS",w,IniFilename);
       if(bOverlayWire)w=yesyes; else w=nono;
       WritePrivateProfileString(IniSection,"OVERLAY",w,IniFilename);
       RemoveMenu(GetMenu(glWndParent),GLMENU_POSITION,MF_BYPOSITION);
       if(glMenu != NULL)DestroyMenu(glMenu);
       DrawMenuBar(glWndParent);
       glMenu=NULL;
     }
     break;
   default:
     lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
   break;
 }
 return lRet;
}

static void GluCommands(WORD cmd,HWND hWnd){
 char HelpText[]="Click in Window to Prepare Updated View\n"
                 "Drag Scroll Bars to Move Viewpoint\n\n"
                 "Hold Ctrl while dragging scrollbar\n"
                 "to prevent intermediate updates";
 switch (cmd){
   case IDM_GLVIEW_PREFS:
     DialogBox(glInstance,MAKEINTRESOURCE(DLG_PREFS),hWnd,
               (DLGPROC)PrefsDlgProc);
     if(Mapped && glMapList == NULL)BuildMapList();
     PostMessage(glWnd,WM_UPDATE,0,0);
     break;
   case IDM_QUICKMAPCHANGE:
     Mapped ^=1;
   case IDM_GLVIEW_MAP:
     BuildMapList();
     PostMessage(glWnd,WM_UPDATE,0,0);
     break;
   case IDM_GLVIEW_HELP:
     MessageBox(hWnd,HelpText,"Help",MB_OK);
     break;
   case IDM_GLVIEW_ABOUT:
     MessageBox(hWnd,ABOUT_VERSION,"About Designer OpenGL Viewer",MB_OK);
     break;
   case IDM_GLVIEW_SCROLLBARS:
     ScrollBars ^= 1;
     if(ScrollBars){
       SetScrollRange(hWnd, SB_VERT, -89,89,TRUE);
       SetScrollRange(hWnd, SB_HORZ, -179,179,TRUE);
       SetScrollPos(hWnd,SB_HORZ,(int)round_angle,TRUE);
       SetScrollPos(hWnd,SB_VERT,(int)up_angle,TRUE);
     }
     else{
       SetScrollRange(hWnd, SB_VERT, 0,0,TRUE);
       SetScrollRange(hWnd, SB_HORZ, 0,0,TRUE);
       SetScrollPos(hWnd,SB_HORZ,0,TRUE);
       SetScrollPos(hWnd,SB_VERT,0,TRUE);
     }
     if(ScrollBars)CheckMenuItem(glMenu,IDM_GLVIEW_SCROLLBARS,MF_CHECKED);
     else          CheckMenuItem(glMenu,IDM_GLVIEW_SCROLLBARS,MF_UNCHECKED);
     break;
   case IDM_GLVIEW_OVERLAY:
     bOverlayWire ^= 1;
     if(bOverlayWire)CheckMenuItem(glMenu,IDM_GLVIEW_OVERLAY,MF_CHECKED);
     else            CheckMenuItem(glMenu,IDM_GLVIEW_OVERLAY,MF_UNCHECKED);
     PostMessage(glWnd,WM_UPDATE,0,0);
     break;
   case IDM_GLVIEW_TWOSIDE:
      TwoSided ^= 1;
      if(TwoSided)glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
      else        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
      if(TwoSided)CheckMenuItem(glMenu,IDM_GLVIEW_TWOSIDE,MF_CHECKED);
      else        CheckMenuItem(glMenu,IDM_GLVIEW_TWOSIDE,MF_UNCHECKED);
      PostMessage(glWnd,WM_UPDATE,0,0);
      break;
   case IDM_GLVIEW_SHOWOPENGL:
      PostMessage(hWnd,WM_CLOSE,0,0);
      break;
   default:
     break;
 }
}

static unsigned char threeto8[8] = {
    0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

static unsigned char twoto8[4] = {
    0, 0x55, 0xaa, 0xff
};

static unsigned char oneto8[2] = {
    0, 255
};

static int defaultOverride[13] = {
    0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift){
 unsigned char val;
 val = (unsigned char) (i >> shift);
 switch (nbits) {
    case 1:
        val &= 0x1;
        return oneto8[val];
    case 2:
        val &= 0x3;
        return twoto8[val];
    case 3:
        val &= 0x7;
        return threeto8[val];
    default:
        return 0;
 }
}

static void CreateRGBPalette(HDC hDC){
 PIXELFORMATDESCRIPTOR pfd;
 LOGPALETTE *pPal;
 int n, i;
 n = GetPixelFormat(hDC);
 DescribePixelFormat(hDC, n, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
 if(pfd.dwFlags & PFD_NEED_PALETTE) {
   n = 1 << pfd.cColorBits;
   pPal = (PLOGPALETTE)LocalAlloc(LMEM_FIXED, sizeof(LOGPALETTE) +
                                   n * sizeof(PALETTEENTRY));
   pPal->palVersion = 0x300;
   pPal->palNumEntries = n;
   for (i=0; i<n; i++) {
     pPal->palPalEntry[i].peRed =
            ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
     pPal->palPalEntry[i].peGreen =
            ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
     pPal->palPalEntry[i].peBlue =
            ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
     pPal->palPalEntry[i].peFlags = 0;
   }
   if((pfd.cColorBits == 8)                           &&
       (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
       (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
       (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)) {
     for(i = 1 ; i <= 12 ; i++)
        pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
   }
   ghPalette = CreatePalette(pPal);
   LocalFree(pPal);
   ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
   n = RealizePalette(hDC);
 }
}

static BOOL bSetupPixelFormat(HDC hDC){
 static PIXELFORMATDESCRIPTOR pfd = {
   sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
   1,                              // version number
   PFD_DRAW_TO_WINDOW |            // support window
   PFD_SUPPORT_OPENGL |            // support OpenGL
   PFD_DOUBLEBUFFER,               // double buffered
   PFD_TYPE_RGBA,                  // RGBA type
   24,                             // 24-bit color depth
   0, 0, 0, 0, 0, 0,               // color bits ignored
   0,                              // no alpha buffer
   0,                              // shift bit ignored
   0,                              // no accumulation buffer
   0, 0, 0, 0,                     // accum bits ignored
   32,                             // 32-bit z-buffer
   0,                              // no stencil buffer
   0,                              // no auxiliary buffer
   PFD_MAIN_PLANE,                 // main layer
   0,                              // reserved
   0, 0, 0                         // layer masks ignored
 };
 int pixelformat;
 if ( (pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0 ){
   MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
   return FALSE;
 }
 if (SetPixelFormat(hDC, pixelformat, &pfd) == FALSE) {
   MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
   return FALSE;
 }
 CreateRGBPalette(hDC);
 return TRUE;
}

static BOOL same_color(unsigned char *color){
 static unsigned char lastcolor[3]={0,0,0};
 if(labs((long)(color[0]) - (long)(lastcolor[0])) > 2 ||
    labs((long)(color[1]) - (long)(lastcolor[1])) > 2 ||
    labs((long)(color[2]) - (long)(lastcolor[2])) > 2){
   VECCOPY(color,lastcolor)
   return FALSE;
 }
 return TRUE;
}

static GLvoid initialize(HWND hWnd){
 GLfloat   aspect;
 GLfloat d[]={1.0,1.0,1.0,1.0};
 GLfloat a[]={0.2,0.2,0.2,1.0};
 GLfloat s[]={1.0,1.0,1.0,1.0};
 GLfloat p[]={0.0,0.0,1.0,0.0};
 if(AttenuateLight)p[3]=1.0;
 else              p[3]=0.0;
 GetClientRect(hWnd, &oldrect);
 glClearColor(gbRed,gbGreen,gbBlue,1.0);
 glMatrixMode(GL_PROJECTION);
 aspect = (GLfloat) oldrect.right / oldrect.bottom;;
// near_plane = 2.0; far_plane = 8.0;
 near_plane = 1.5; far_plane = 9.5;
 gluPerspective( 45.0, aspect, near_plane, far_plane);
 glMatrixMode( GL_MODELVIEW );
 glLoadIdentity();
 glClearDepth(1.0);
 glEnable(GL_DEPTH_TEST);
 if(TwoSided)glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
 else        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
 glEnable(GL_LIGHT0);
 glLightfv(GL_LIGHT0,GL_DIFFUSE,d);
 glLightfv(GL_LIGHT0,GL_AMBIENT,a);
 glLightfv(GL_LIGHT0,GL_SPECULAR,s);
 glLightfv(GL_LIGHT0,GL_POSITION,p);
 glLightf(GL_LIGHT0,GL_LINEAR_ATTENUATION,0.15);
 glLightf(GL_LIGHT0,GL_QUADRATIC_ATTENUATION,0.05);
 glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
 makeImageMap(0);
}

static GLvoid drawScene(HWND hWnd,BOOL shaded){
 HDC hDC;
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 if(flat)glShadeModel(GL_FLAT);
 else    glShadeModel(GL_SMOOTH);
 glPushMatrix();
 glLoadIdentity();
 glTranslatef(0.0, 0.0, -5.0);
 glRotatef(up_angle,1.0,0.0,0.0);
 glRotatef(round_angle,0.0,1.0,0.0);
 if(shaded)drawModelF();
 else      drawModelE();
 if(bDrawMapping)drawModelMaps();
 glPopMatrix();
 glFinish();
 hDC = wglGetCurrentDC();
 SwapBuffers(hDC);
}

static GLvoid resize(HWND hWnd){
 RECT    rect;
 GLfloat aspect;
 GetClientRect(hWnd, &rect);
 glViewport(0, 0, rect.right, rect.bottom);
 /* don't paint here */
 oldrect.right = rect.right;
 oldrect.bottom = rect.bottom;
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 aspect = (GLfloat) oldrect.right / oldrect.bottom;
 gluPerspective( 45.0, aspect, near_plane, far_plane);
 glMatrixMode( GL_MODELVIEW );
 glLoadIdentity();
}

static void drawModelMaps(void){
 long i;
 GLfloat x1,y1,z1,x2,y2,z2,scale,dscale=0.05;
 point dx,dy,d;
 if(nImaps > 0 && current_map < nImaps){
   glPushMatrix();
   i=current_map;
   scale=4.0/TVsizeX;
   VECSUB(iMap[i].X,iMap[i].P,dx)
   VECSUB(iMap[i].Y,iMap[i].P,dy)
   VECSUM(iMap[i].X,dy,d)
   // Transparent
   glEnable(GL_BLEND);
   glDepthMask(GL_FALSE);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE);
   glColor4f(1.0, 1.0, 1.0,0.3);
   glBegin(GL_QUADS);
   x1=((GLfloat)(iMap[i].P[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].P[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].P[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].X[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].X[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].X[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(d[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(d[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(d[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].Y[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].Y[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].Y[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   glEnd();
   glDepthMask(GL_TRUE);
   glDisable(GL_BLEND);
  //end transparent
   glLineWidth(4.0);
   glBegin(GL_LINES);
   glColor3f(0.0, 1.0, 0.0);
   x1=((GLfloat)(iMap[i].Y[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].Y[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].Y[2]-TVpointZ-TVsizeZ/2))*scale;
   x2=((GLfloat)(iMap[i].P[0]-TVpointX-TVsizeX/2))*scale;
   y2=((GLfloat)(iMap[i].P[1]-TVpointY-TVsizeY/2))*scale;
   z2=((GLfloat)(iMap[i].P[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   glVertex3f(x2,z2,-y2);
   x1=((GLfloat)(iMap[i].X[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].X[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].X[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x2,z2,-y2);
   glVertex3f(x1,z1,-y1);
   x2=((GLfloat)(d[0]-TVpointX-TVsizeX/2))*scale;
   y2=((GLfloat)(d[1]-TVpointY-TVsizeY/2))*scale;
   z2=((GLfloat)(d[2]-TVpointZ-TVsizeZ/2))*scale;
   glEnd();
   glLineWidth(1.0);
   glBegin(GL_LINES);
   glVertex3f(x1,z1,-y1);
   glVertex3f(x2,z2,-y2);
   x1=((GLfloat)(iMap[i].Y[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].Y[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].Y[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x2,z2,-y2);
   glVertex3f(x1,z1,-y1);
   glEnd();
   VECSCALE(dscale,(GLfloat)dx,dx)
   VECSCALE(dscale,(GLfloat)dy,dy)
   glBegin(GL_LINE_LOOP);
   x1=((GLfloat)(iMap[i].P[0]-dx[0]-dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].P[1]-dx[1]-dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].P[2]-dx[2]-dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].P[0]+dx[0]-dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].P[1]+dx[1]-dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].P[2]+dx[2]-dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].P[0]+dx[0]+dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].P[1]+dx[1]+dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].P[2]+dx[2]+dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].P[0]-dx[0]+dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].P[1]-dx[1]+dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].P[2]-dx[2]+dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   glEnd();
   glBegin(GL_LINE_LOOP);
   x1=((GLfloat)(iMap[i].Y[0]-dx[0]-dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].Y[1]-dx[1]-dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].Y[2]-dx[2]-dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].Y[0]+dx[0]-dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].Y[1]+dx[1]-dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].Y[2]+dx[2]-dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].Y[0]+dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].Y[1]+dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].Y[2]+dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   glEnd();
   VECSCALE(0.5,(GLfloat)dx,dx)
   VECSCALE(0.5,(GLfloat)dy,dy)
   glBegin(GL_LINE_LOOP);
   x1=((GLfloat)(iMap[i].X[0]-dx[0]-dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].X[1]-dx[1]-dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].X[2]-dx[2]-dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].X[0]+dx[0]-dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].X[1]+dx[1]-dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].X[2]+dx[2]-dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].X[0]+dx[0]+dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].X[1]+dx[1]+dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].X[2]+dx[2]+dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   x1=((GLfloat)(iMap[i].X[0]-dx[0]+dy[0]-TVpointX-TVsizeX/2))*scale;
   y1=((GLfloat)(iMap[i].X[1]-dx[1]+dy[1]-TVpointY-TVsizeY/2))*scale;
   z1=((GLfloat)(iMap[i].X[2]-dx[2]+dy[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x1,z1,-y1);
   glEnd();
   glLineWidth(1.0);
   glPopMatrix();
 }
 return;
}

static void drawModelEx(void){
 GLfloat x,y,z,scale;
 edge *ep;
 vertex *v1,*v2;
 int i;
 if(Nedge == 0 && Nnurbs == 0)return;
 glPushMatrix();
 if(bThickWire)glLineWidth(2.0);
 else          glLineWidth(1.0);
 glBegin(GL_LINES);
 glColor3f(gwRed,gwGreen,gwBlue);
 scale=4.0/TVsizeX;
 ep=MainEp; if(Nedge > 0)for(i=0;i<Nedge;i++){
   v1=(MainVp+ep->V[0]);
   v2=(MainVp+ep->V[1]);
   x=((GLfloat)(v1->xyz[0]-TVpointX-TVsizeX/2))*scale;
   y=((GLfloat)(v1->xyz[1]-TVpointY-TVsizeY/2))*scale;
   z=((GLfloat)(v1->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x,z,-y);
   x=((GLfloat)(v2->xyz[0]-TVpointX-TVsizeX/2))*scale;
   y=((GLfloat)(v2->xyz[1]-TVpointY-TVsizeY/2))*scale;
   z=((GLfloat)(v2->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x,z,-y);
   ep++;
 }
 glEnd();
 glLineWidth(1.0);
 DrawGlNurbs(GLU_OUTLINE_PATCH);
 glPopMatrix();
}

static void drawModelEw(void){
 GLfloat x,y,z,scale;
 point    *p;
 twoLongs *e;
 long  i,ne,v1,v2;
 if((ne=WireFrame->Ne) == 0){
   drawModelEx();
   return;
 }
 if((e=WireFrame->e) == NULL)return;
 if((p=WireFrame->p) == NULL)return;
 glPushMatrix();
 glLineWidth(1.0);
 glBegin(GL_LINES);
 glColor3f(gwRed,gwGreen,gwBlue);
 scale=4.0/TVsizeX;
 for(i=0;i<ne;i++){
   v1=e[i][0];
   v2=e[i][1];
   x=((GLfloat)(p[v1][0]-TVpointX-TVsizeX/2))*scale;
   y=((GLfloat)(p[v1][1]-TVpointY-TVsizeY/2))*scale;
   z=((GLfloat)(p[v1][2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x,z,-y);
   x=((GLfloat)(p[v2][0]-TVpointX-TVsizeX/2))*scale;
   y=((GLfloat)(p[v2][1]-TVpointY-TVsizeY/2))*scale;
   z=((GLfloat)(p[v2][2]-TVpointZ-TVsizeZ/2))*scale;
   glVertex3f(x,z,-y);
 }
 glEnd();
 glPopMatrix();
}

static void drawModelE(void){
 if(bounding_box == 1)drawModelEw();
 else                 drawModelEx();
 return;
}

static void drawModelF(void){
 glCallList(1);
}

static BOOL normalise(GLfloat *v){
 GLfloat n,nn;
 n= (GLfloat)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return FALSE;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE(n,v,v)
 return TRUE;
}

static BOOL Normalize(point x,point y,point z,GLfloat *n){
 GLfloat dz[3],dy[3];
 VECSUB((GLfloat)y,(GLfloat)x,dy)
 VECSUB((GLfloat)z,(GLfloat)x,dz)
 CROSS(dy,dz,n)
 if(normalise(n))return TRUE;
 return FALSE;
}

#define HIDDEN 2

static BOOL inview(vertex  *vp){
 if(vp->status == HIDDEN)return 0;
// if( ((vp->xyz[0] > TVpointX) && (vp->xyz[0] < TVpointX+TVsizeX))
//  && ((vp->xyz[1] > TVpointY) && (vp->xyz[1] < TVpointY+TVsizeY))
//  && ((vp->xyz[2] > TVpointZ) && (vp->xyz[2] < TVpointZ+TVsizeZ))
//   )return TRUE;
 if( ((vp->xyz[0] > TVpointX-TVsizeX) && (vp->xyz[0] < TVpointX+2*TVsizeX))
  && ((vp->xyz[1] > TVpointY-TVsizeX) && (vp->xyz[1] < TVpointY+2*TVsizeY))
  && ((vp->xyz[2] > TVpointZ-TVsizeX) && (vp->xyz[2] < TVpointZ+2*TVsizeZ))
   )return TRUE;
 return FALSE;
}

static void Make3dListF(void){
 GLfloat x,y,z,scale,n[3],color[4],*spec_color,
         glos_color[]={1.0,1.0,1.0,1.0},
         matt_color[]={0.0,0.0,0.0,0.0},
         shiny[]={100.0};
 face *fp;
 vertex *v,*v1,*v2,*v0;
 int i,j;
 if(listID > 0){glDeleteLists(1,1); listID=0;}
 if(Nface == 0 && Nnurbs == 0)return;
 color[3]=1.0;
 glNewList(1,GL_COMPILE);
 glEnable(GL_LIGHTING);
 if(Cmaterial)glEnable(GL_COLOR_MATERIAL);
 glBegin(GL_TRIANGLES);
 scale=4.0/TVsizeX;
 glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
 glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,glos_color);
 fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++,fp++){
   if(Mapped && ((fp->attrib & 0x40) == 0x40))continue;
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   if(inview(v0) || inview(v1) || inview(v2)){
     if((fp->matcol[1] & 0xf0) == 0){ /* Not glass */
       if(!same_color(fp->color)){
         color[0]=(GLfloat)fp->color[0]/255.0;
         color[1]=(GLfloat)fp->color[1]/255.0;
         color[2]=(GLfloat)fp->color[2]/255.0;
         if(Cmaterial)glColor4fv(color);
         else{
           glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,color);
           if((fp->attrib & 0x20) == 0x20)spec_color=glos_color;
           else                           spec_color=matt_color;
           glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec_color);
         }
       }
       if(Normalize(v0->xyz,v1->xyz,v2->xyz,n)){
         glNormal3f(n[0],n[2],-n[1]);
         for(j=0;j<3;j++){
           v=(MainVp+fp->V[j]);
           x=((GLfloat)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
           y=((GLfloat)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
           z=((GLfloat)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
           glVertex3f(x,z,-y);
         }
       }
     }
   }
 }
 glEnd();
 if(Mapped && Nface > 0){
   int k,brushID;
   vector mN,mX,mY,mP;
   GLfloat alpha,beta;
   color[0]=1.2; color[1]=1.2; color[2]=1.2;
   if(Cmaterial)glColor4fv(color);
   else{
     glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,color);
     if((fp->attrib & 0x20) == 0x20)spec_color=glos_color;
     else                           spec_color=matt_color;
     glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec_color);
   }
   if(nImaps > 0)for(k=0;k<nImaps;k++){
     GetMapNormal(k,mP,mX,mY,mN);
     glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
     if(glMapList != NULL && k < NglMapList){
       if(glMapList[k].pixels != NULL)
             glTexImage2D(GL_TEXTURE_2D,0,3,
                          glMapList[k].x,glMapList[k].y,
                          0,GL_RGB,GL_UNSIGNED_BYTE,
                          (GLvoid *)glMapList[k].pixels);
        else glTexImage2D(GL_TEXTURE_2D, 0, 3, checkImageWidth,
                          checkImageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
                          &checkImage[0][0][0]);
     }
     else{
       glTexImage2D(GL_TEXTURE_2D, 0, 3, checkImageWidth,
                    checkImageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
                    &checkImage[0][0][0]);
     }
     glEnable(GL_TEXTURE_2D);
     glBegin(GL_TRIANGLES);
     fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++,fp++){
       if((fp->attrib & 0x40) != 0x40)continue; /* if not mapped */
       brushID=fp->map;
       if(brushID != k)continue;         /* skip if not this map */
       v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
       if(inview(v0) || inview(v1) || inview(v2)){
         if(Normalize(v0->xyz,v1->xyz,v2->xyz,n)){
           glNormal3f(n[0],n[2],-n[1]);
           for(j=0;j<3;j++){
             v=(MainVp+fp->V[j]);
             x=((GLfloat)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
             y=((GLfloat)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
             z=((GLfloat)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
             if(iMap[k].Map == MAP_BY_VERTEX){
               alpha=(GLfloat)v->x;
               beta=(GLfloat)v->y;
             }
             else if(iMap[k].Map == CYLINDER ||
                     iMap[k].Map == MAP_SPHERICAL ||
                     iMap[k].Map == CYLINDER_MOZIAC)
                  GetMappingCoordC(k,mN,mX,mY,mP,v->xyz,&alpha,&beta);
             else GetMappingCoordP(k,mN,mX,mY,mP,v->xyz,&alpha,&beta);
             glTexCoord2f(alpha,beta);
             glVertex3f(x,z,-y);
           }
         }
       }
     }
     glEnd();
     glDisable(GL_TEXTURE_2D);
   }
 }
 DrawGlNurbs(GLU_FILL);
 if(Cmaterial)glDisable(GL_COLOR_MATERIAL);
 glDisable(GL_LIGHTING);
 bThickWire=TRUE;
 if(bOverlayWire)drawModelEx();
 bThickWire=FALSE;
 glEndList();
 listID=1;
 return;
}

static void Make3dListS(void){
 GLfloat x,y,z,scale,n[3],color[4],*spec_color,
         glos_color[]={1.0,1.0,1.0,1.0},
         matt_color[]={0.0,0.0,0.0,0.0},
         shiny[]={100.0},dot;
 face *fp;
 vertex *v,*v1,*v2,*v0;
 normal *nv;
 int i,j,Vi,V[3];
 if(Nface == 0 && Nnurbs == 0){
   if(listID > 0){glDeleteLists(1,1);listID=0;}
   return;
 }
 if(Nvert > 0){
   if((nv=(normal *)X__Malloc(sizeof(normal)*Nvert)) == NULL){
     MessageBeep(MB_OK); Make3dListF(); return;
   }
   for(i=0;i<Nvert;i++)nv[i][0]=nv[i][1]=nv[i][2]=0.0;
 }
 fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++){
   V[0]=fp->V[0]; V[1]=fp->V[1]; V[2]=fp->V[2];
   v0=(MainVp+V[0]); v1=(MainVp+V[1]); v2=(MainVp+V[2]);
   if(inview(v0) || inview(v1) || inview(v2)){
     if(fp->attrib >> 7 != 0 && /* smoothed */
        Normalize(v0->xyz,v1->xyz,v2->xyz,n)){
       for(j=0;j<3;j++){
         dot=DOT(nv[V[j]],n);
         if(fabs(dot) < 1.e-10){
           VECCOPY(n,nv[V[j]])
         }
         else if(dot >  0.5){
           VECSUM(nv[V[j]],n,nv[V[j]])
           normalise(nv[V[j]]);
         }
         else if(dot < -0.5){
           VECSUB(nv[V[j]],n,nv[V[j]])
           normalise(nv[V[j]]);
         }
       }
     }
   }
   fp++;
 }
 if(listID > 0){glDeleteLists(1,1);listID=0;}
 color[3]=1.0;
 glNewList(1,GL_COMPILE);
 glEnable(GL_LIGHTING);
 if(Cmaterial)glEnable(GL_COLOR_MATERIAL);
 glBegin(GL_TRIANGLES);
 scale=4.0/TVsizeX;
 glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
 glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,glos_color);
 fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++,fp++){
   if(Mapped && ((fp->attrib & 0x40) == 0x40))continue;
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   if(inview(v0) || inview(v1) || inview(v2)){
     if((fp->matcol[1] & 0xf0) == 0){ /* Not glass */
       if(!same_color(fp->color)){
         color[0]=(GLfloat)fp->color[0]/255.0;
         color[1]=(GLfloat)fp->color[1]/255.0;
         color[2]=(GLfloat)fp->color[2]/255.0;
         if(Cmaterial)glColor4fv(color);
         else{
           glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,color);
           if((fp->attrib & 0x20) == 0x20)spec_color=glos_color;
           else                           spec_color=matt_color;
           glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec_color);
         }
       }
       if(Normalize(v0->xyz,v1->xyz,v2->xyz,n)){
         for(j=0;j<3;j++){
           Vi=fp->V[j]; v=(MainVp+Vi);
           x=((GLfloat)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
           y=((GLfloat)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
           z=((GLfloat)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
           if(fp->attrib >> 7 != 0){ /* smoothed */
             dot=DOT(nv[Vi],n);
             if(fabs(dot) < 0.5)glNormal3f(n[0],n[2],-n[1]);
             else if(dot  < 0.0)glNormal3f(-nv[Vi][0],-nv[Vi][2], nv[Vi][1]);
             else               glNormal3f( nv[Vi][0], nv[Vi][2],-nv[Vi][1]);
           }
           else{/* not smoothed */
             glNormal3f( n[0], n[2],-n[1]);
           }
           glVertex3f(x,z,-y);
         }
       }
     }
   }
 }
 glEnd();
 if(Mapped && Nface > 0){
   int k,brushID;
   vector mN,mX,mY,mP;
   GLfloat alpha,beta;
   color[0]=1.2; color[1]=1.2; color[2]=1.2;
   if(Cmaterial)glColor4fv(color);
   else{
     glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,color);
     if((fp->attrib & 0x20) == 0x20)spec_color=glos_color;
     else                           spec_color=matt_color;
     glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec_color);
   }
   if(Nface > 0 && nImaps > 0)for(k=0;k<nImaps;k++){
     GetMapNormal(k,mP,mX,mY,mN);
     glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
     if(glMapList != NULL && k < NglMapList){
       if(glMapList[k].pixels != NULL)
             glTexImage2D(GL_TEXTURE_2D,0,3,
                          glMapList[k].x,glMapList[k].y,
                          0,GL_RGB,GL_UNSIGNED_BYTE,
                          (GLvoid *)glMapList[k].pixels);
        else glTexImage2D(GL_TEXTURE_2D, 0, 3, checkImageWidth,
                          checkImageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
                          &checkImage[0][0][0]);
     }
     else{
       glTexImage2D(GL_TEXTURE_2D, 0, 3, checkImageWidth,
                    checkImageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
                    &checkImage[0][0][0]);
     }
     glEnable(GL_TEXTURE_2D);
     glBegin(GL_TRIANGLES);
     fp=MainFp; if(Nface > 0)for(i=0;i<Nface;i++,fp++){
       if((fp->attrib & 0x40) != 0x40)continue; /* if not mapped */
       brushID=fp->map;
       if(brushID != k)continue;         /* skip if not this map */
       v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
       if(inview(v0) || inview(v1) || inview(v2)){
         if(Normalize(v0->xyz,v1->xyz,v2->xyz,n)){
           for(j=0;j<3;j++){
             Vi=fp->V[j]; v=(MainVp+Vi);
             x=((GLfloat)(v->xyz[0]-TVpointX-TVsizeX/2))*scale;
             y=((GLfloat)(v->xyz[1]-TVpointY-TVsizeY/2))*scale;
             z=((GLfloat)(v->xyz[2]-TVpointZ-TVsizeZ/2))*scale;
             if(fp->attrib >> 7 != 0){ /* smoothed */
               dot=DOT(nv[Vi],n);
               if(fabs(dot) < 0.5)glNormal3f(n[0],n[2],-n[1]);
               else if(dot  < 0.0)glNormal3f(-nv[Vi][0],-nv[Vi][2], nv[Vi][1]);
               else               glNormal3f( nv[Vi][0], nv[Vi][2],-nv[Vi][1]);
             }
             else{/* not smoothed */
               glNormal3f( n[0], n[2],-n[1]);
             }
             if(iMap[k].Map == MAP_BY_VERTEX){
               alpha=(GLfloat)v->x;
               beta=(GLfloat)v->y;
             }
             else if(iMap[k].Map == CYLINDER ||
                     iMap[k].Map == MAP_SPHERICAL ||
                     iMap[k].Map == CYLINDER_MOZIAC)
                  GetMappingCoordC(k,mN,mX,mY,mP,v->xyz,&alpha,&beta);
             else GetMappingCoordP(k,mN,mX,mY,mP,v->xyz,&alpha,&beta);
             glTexCoord2f(alpha,beta);
             glVertex3f(x,z,-y);
           }
         }
       }
     }
     glEnd();
     glDisable(GL_TEXTURE_2D);
   }
 }
 DrawGlNurbs(GLU_FILL);
 if(Cmaterial)glDisable(GL_COLOR_MATERIAL);
 glDisable(GL_LIGHTING);
 bThickWire=TRUE;
 if(bOverlayWire)drawModelEx();
 bThickWire=FALSE;
 glEndList();
 listID=1;
 X__Free(nv);
 return;
}

static void DrawGlNurbs(GLfloat drawtype){
 GLUnurbsObj *theNurb;
 GLfloat *sknots,*tknots,*ctlarray;
 nurbs *np;
 GLfloat w,xp,yp,zp,scale;
 long i,j,k;
 vector4 *p1;
 if(Nnurbs == 0 || MainNp == NULL)return;
 if(drawtype == GLU_FILL && bNurbsWireOnly)drawtype=GLU_OUTLINE_POLYGON;
 scale=4.0/TVsizeX;
 np=MainNp; while(np != NULL){
   if(!(np->properties.hidden) && (theNurb = gluNewNurbsRenderer()) != 0){
     sknots=(GLfloat *)X__Malloc(sizeof(GLfloat)*(np->numU+np->orderU));
     tknots=(GLfloat *)X__Malloc(sizeof(GLfloat)*(np->numV+np->orderV));
     if(sknots != NULL)for(j=0;j<np->numU+np->orderU;j++){ //knotsU
       sknots[j]=(GLfloat)np->kvU[j];
     }
     if(tknots != NULL)for(i=0;i<np->numV+np->orderV;i++){ //knotsV
       tknots[i]=(GLfloat)np->kvV[i];
     }
     ctlarray=(GLfloat *)X__Malloc(sizeof(GLfloat)*(4*np->numV*np->numU));
     if(ctlarray != NULL){
       k=0;
       for(j=0;j<np->numU;j++)
       for(i=0;i<np->numV;i++){
         p1 = &(np->points[i][j]);
         w=(GLfloat)(1.0/p1->w);
         xp=(GLfloat)((p1->x)*w);
         yp=(GLfloat)((p1->y)*w);
         zp=(GLfloat)((p1->z)*w);
         xp = (xp-(GLfloat)(TVpointX+TVsizeX/2))*scale;
         yp = (yp-(GLfloat)(TVpointY+TVsizeY/2))*scale;
         zp = (zp-(GLfloat)(TVpointZ+TVsizeZ/2))*scale;
         w=(GLfloat)(p1->w);
         *(ctlarray+(k++)) =  xp*w;
         *(ctlarray+(k++)) =  zp*w;
         *(ctlarray+(k++)) = -yp*w;
         *(ctlarray+(k++)) =  w;
       }
       if(drawtype == GLU_FILL)glEnable(GL_AUTO_NORMAL);
       gluBeginSurface(theNurb);
       glColor3f((GLfloat)(np->properties.colour[0])/255.0,
                 (GLfloat)(np->properties.colour[1])/255.0,
                 (GLfloat)(np->properties.colour[2])/255.0);
       if(drawtype == GLU_OUTLINE_POLYGON){
         gluNurbsProperty(theNurb,GLU_SAMPLING_METHOD,GLU_DOMAIN_DISTANCE);
         gluNurbsProperty(theNurb,GLU_U_STEP,5);
         gluNurbsProperty(theNurb,GLU_V_STEP,5);
       }
       else{
         gluNurbsProperty(theNurb,GLU_SAMPLING_TOLERANCE,50.0);
         gluNurbsProperty(theNurb,GLU_SAMPLING_METHOD,GLU_PATH_LENGTH);
       }
//gluNurbsProperty(theNurb, GLU_DISPLAY_MODE, GLU_OUTLINE_POLYGON);
//gluNurbsProperty(theNurb, GLU_DISPLAY_MODE, GLU_OUTLINE_PATCH);
//gluNurbsProperty(theNurb, GLU_DISPLAY_MODE, GLU_FILL);
       gluNurbsProperty(theNurb, GLU_DISPLAY_MODE, drawtype);
       gluNurbsSurface(theNurb,
            np->numU+np->orderU, sknots,
            np->numV+np->orderV, tknots,
            4 * np->numV,
            4,
            ctlarray,
            np->orderU, np->orderV,
            GL_MAP2_VERTEX_4);
       gluEndSurface(theNurb);
       if(drawtype == GLU_FILL)glDisable(GL_AUTO_NORMAL);
     }
     if(sknots != NULL)X__Free(sknots);
     if(tknots != NULL)X__Free(tknots);
     if(ctlarray != NULL)X__Free(ctlarray);
     gluDeleteNurbsRenderer(theNurb);
   }
   np=np->last;
 }
 return;
}

#define ALIGNULONG(i)   ((i+3)/4*4)

static char *short_form(char *filename){
 char *sps,*spc;
 sps=spc=filename;
 while((spc=strchr(spc,'\\')) != NULL){spc++; sps=spc;}
 return sps;
}

static void CheckMapPath(char *fn){
 long l;
 char *fs;
 FILE *Bit_File=NULL;
 char pref[32],preff[32],temp_name[255];
 if((Bit_File=fopen((char *)fn,"rb")) != NULL){
   fclose(Bit_File);
   return;
 }
 fs=short_form(fn);
 strcpy(preff,"MAP-PATH");
 for(l=1;l<=128;l++){ /*  try .INI settings for models and maps*/
   sprintf(pref,"%s%ld",preff,l);
   GetPrivateProfileString("RPREFERENCES",pref,"$$$",
                           temp_name,255,IniFilename);
   if(strncmp(temp_name,"$$$",3) == 0)break;
   if(temp_name[1] != ':')continue;
   if(temp_name[strlen(temp_name)-1] != '\\')strcat(temp_name,"\\");
   strcat(temp_name,fs);
   if((Bit_File=fopen((char *)temp_name,"rb")) != NULL){
     fclose(Bit_File);
     strcpy(fn,temp_name);
     return;
   }
 }
}

unsigned char *LoadMAP(char *InFile, long *xx,long *yy){
 int ff;
 char CurrentFile[255];
 strcpy(CurrentFile,InFile);
 CheckMapPath(CurrentFile);
 if     (strstr(CurrentFile,".GIF") || strstr(CurrentFile,".gif"))
   return LoadGIF(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".TGA") || strstr(CurrentFile,".tga"))
   return LoadTGA(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".JPG") || strstr(CurrentFile,".jpg"))
   return LoadXXX(CurrentFile,xx,yy,0);
 else if(strstr(CurrentFile,".TIF") || strstr(CurrentFile,".tif"))
   return LoadXXX(CurrentFile,xx,yy,1);
 else if(strstr(CurrentFile,".PNG") || strstr(CurrentFile,".png"))
   return LoadXXX(CurrentFile,xx,yy,2);
 return NULL;
}

static void BuildMapList(void){
 long i,x,y;
 unsigned char *p;
 FreeMapList();
 if(!Mapped || nImaps == 0)return;
 if((glMapList=(GLMAP *)X__Malloc(nImaps*sizeof(GLMAP))) == NULL)return;
 NglMapList=nImaps;
 StartWait();
 for(i=0;i<nImaps;i++){
   glMapList[i].pixels=NULL;
   if(iMap[i].S != NULL){
     if((p=LoadMAP(iMap[i].S,&x,&y)) != NULL){
       glMapList[i].pixels=p;
       glMapList[i].x=x;
       glMapList[i].y=y;
     }
   }
 }
 SetMappingQuality();
// glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
 glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
 EndWait();
 return;
}

static void FreeMapList(void){
 long i;
 if(NglMapList == 0 || glMapList == NULL)return;
 for(i=0;i<NglMapList;i++){
   if(glMapList[i].pixels != NULL)X__Free(glMapList[i].pixels);
 }
 X__Free(glMapList);
 glMapList=NULL;
 NglMapList=0;
 return;
}

static SetMappingQuality(void){
 if(Clamped){
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
 }
 else{
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
 }
 if(Fastest)glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);
 else       glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

static void makeImageMap(int id){
 int i, j, c;
 if(id == 0){ /* simple checkered map */
   for (i = 0; i < checkImageWidth; i++) {
     for (j = 0; j < checkImageHeight; j++) {
       c = ((((i&0x8)==0)^((j&0x8))==0))*255;
       checkImage[i][j][0] = (GLubyte) c;
       checkImage[i][j][1] = (GLubyte) c;
       checkImage[i][j][2] = (GLubyte) c;
     }
   }
 }
 SetMappingQuality();
 glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

#define TOL 1000.0

static void GetMapNormal(int k, vector mP, vector mX, vector mY,
                         vector mN){
 GLfloat n[3];
 if(iMap[k].Map == CYLINDER || iMap[k].Map == CYLINDER_MOZIAC){
   VECCOPY((double)iMap[k].P,mP)
   VECSUB((double)iMap[k].X,mP,mX)
   VECSUB((double)iMap[k].Y,mP,mY)
   CROSS(mY,mX,mN)
 }
 else{                                              /* arranged this way to */
   VECCOPY((double)iMap[k].Y,mP)                    /* have map with origin */
   VECSUB((double)iMap[k].X,(double)iMap[k].P,mX)   /* at bottom left       */
   VECSUB((double)iMap[k].P,mP,mY)
   CROSS(mX,mY,mN)
 }
 VECCOPY((GLfloat)mN,n)
 if(!normalise(n))MessageBeep(MB_OK);
 VECCOPY((double)n,mN)
 return;
}

static void GetMappingCoordP(int id, vector n, vector x, vector y, vector p0,
                             long *pv,
                             GLfloat *a, GLfloat *b){
 vector v1,p;
 GLfloat mu,det,ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3;
 VECSUB(p0,(double)pv,v1)
 mu=DOT(n,v1);
 VECSCALE(mu,n,v1)
 VECSUM((double)pv,v1,p)
 ve1=x[0]; ve2=x[1]; ve3=x[2];
 vp1=y[0]; vp2=y[1]; vp3=y[2];
 p1=p[0]-p0[0];
 p2=p[1]-p0[1];
 p3=p[2]-p0[2];
 det=ve1*vp2-vp1*ve2;
 if(det > TOL || det < -TOL)  /* e-10 appears not to work */
   {  *a=( vp2*p1-vp1*p2)/det;
      *b=(-ve2*p1+ve1*p2)/det;
      goto HIT;
   }
 det=ve1*vp3-vp1*ve3;
 if(det > TOL || det < -TOL)
   {  *a=( vp3*p1-vp1*p3)/det;
      *b=(-ve3*p1+ve1*p3)/det;
      goto HIT;
   }
 det=ve2*vp3-vp2*ve3;
 if(det > TOL || det < -TOL)
   {  *a=( vp3*p2-vp2*p3)/det;
      *b=(-ve3*p2+ve2*p3)/det;
      goto HIT;
   }
 *a = *b = 0.0;
 HIT:
 return;
}

#define RTD 0.159154943

static void GetMappingCoordC(int id, vector n, vector x, vector y, vector p0,
                long *pv,
                GLfloat *a, GLfloat *b){
 GLfloat theta,nn;
 GLfloat p1[3],p2[3],q[3];
 VECSUB((GLfloat)p0,(GLfloat)pv,p1)
 *b=DOT(p1,y)/DOT(y,y);
 VECSCALE(*b,y,p2)
 VECSUB(p1,p2,q)
 normalise(q);
 nn=DOT(q,(GLfloat)n);
 if     (nn >  1.0)nn =  1.0;
 else if(nn < -1.0)nn = -1.0;
 theta=acos(nn)*RTD;
 CROSS(q,n,p1)
 if(DOT(p1,y) <= 0.0){
   if(theta <= 0.25) *a = (0.25 - theta);
   else  *a = (1.25 - theta);
 }
 else    *a = (0.25 + theta);
 *a *= 360.0/(GLfloat)iMap[id].Angle;
 return;
}

#if 0

#define DibNumColors(bmih) (bmih.biClrUsed == 0 && bmih.biBitCount <= 8 \
                            ? (int)(1 << (int)bmih.biBitCount)          \
                            : (int)bmih.biClrUsed)

static int LoadDIB(char *DIBfile long *xxi, long **yyi){
 unsigned char cr,cg,cb;
 HANDLE hfbm;
 BITMAPFILEHEADER bmfh;
 DWORD dwRead;
 BITMAPINFOHEADER bmih;
 LPBITMAPINFOHEADER lpbi;
 LPBITMAPINFO lpinfo;
 long i,j,k,imagesize,nc,x,y,xi,yi;
 unsigned char *px;
 RGBQUAD   *pRgb          = NULL;
 LPSTR     lpBitmapHeader = NULL;
 LPSTR     lpBitmapBits   = NULL,lpB;
 hfbm=CreateFile(DIBfile,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(hfbm == INVALID_HANDLE_VALUE){
   MessageBox(NULL,"Invalid file",NULL,MB_OK);
   return 0;
 }
 ReadFile(hfbm,&bmfh,sizeof(BITMAPFILEHEADER),&dwRead,(LPOVERLAPPED)NULL);
 ReadFile(hfbm,&bmih,sizeof(BITMAPINFOHEADER),&dwRead,(LPOVERLAPPED)NULL);
 nc=DibNumColors(bmih);
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+
                   nc*sizeof(RGBQUAD));
 if(lpBitmapHeader == NULL){
   CloseHandle(hfbm);
   return 0;
 }
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpinfo=(LPBITMAPINFO)lpBitmapHeader;
 lpbi->biSize=bmih.biSize;
 x=lpbi->biWidth=bmih.biWidth;
 y=lpbi->biHeight=bmih.biHeight;
 lpbi->biPlanes=bmih.biPlanes;
 lpbi->biBitCount=bmih.biBitCount;
 lpbi->biCompression=bmih.biCompression;
 lpbi->biSizeImage=bmih.biSizeImage;
 lpbi->biXPelsPerMeter=bmih.biXPelsPerMeter;
 lpbi->biYPelsPerMeter=bmih.biYPelsPerMeter;
 lpbi->biClrUsed=bmih.biClrUsed;
 lpbi->biClrImportant=bmih.biClrImportant;
 if(nc > 0)ReadFile(hfbm,lpinfo->bmiColors,nc*sizeof(RGBQUAD),
                    &dwRead,(LPOVERLAPPED)NULL);
 imagesize=bmfh.bfSize-bmfh.bfOffBits;
 lpBitmapBits=(LPSTR)X__Malloc(imagesize);
 if(lpBitmapBits == NULL){
   if(lpBitmapBits   != NULL)X__Free((char *)lpBitmapBits);
   if(lpBitmapHeader != NULL)X__Free((char *)lpBitmapHeader);
   CloseHandle(hfbm);
   return 0;
 }
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 memset(lpBitmapBits,0,imagesize);
 ReadFile(hfbm,lpBitmapBits,imagesize,&dwRead,(LPOVERLAPPED)NULL);
 CloseHandle(hfbm);
 /* now make the IMAGE MAP */
 imagesize=x*y;
 if((pixels=(unsigned char *)X__Malloc(imagesize*3)) == NULL){
   if(lpBitmapBits   != NULL)X__Free((char *)lpBitmapBits);
   if(lpBitmapHeader != NULL)X__Free((char *)lpBitmapHeader);
   return 0;
 }
 px=pixels;
 lpB=lpBitmapBits;
 if(lpbi->biBitCount == 24){  /* true colour */
   for(j=0;j<y;j++){
     for(i=0;i<x;i++){
       cr = *lpB++;
       cg = *lpB++;
       cb = *lpB++;
       *px++ = cb;
       *px++ = cg;
       *px++ = cr;
     }
   }
 }
 else{ /* mapped colour */
{char t[256];
sprintf(t,"Paletted %ld colours x=%ld y=%ld",nc,x,y);
MessageBox(NULL,t,"Debug",MB_OK);
}
   for(j=0;j<y;j++){
     for(i=0;i<x;i++){
       k=(long)(*lpB++);
       *px++ = (unsigned char)(pRgb[k].rgbRed  );
       *px++ = (unsigned char)(pRgb[k].rgbGreen);
       *px++ = (unsigned char)(pRgb[k].rgbBlue );
     }
   }
 }
 if(lpBitmapBits   != NULL)X__Free((char *)lpBitmapBits);
 if(lpBitmapHeader != NULL)X__Free((char *)lpBitmapHeader);
 if     (x <  96)xi=64;
 else if(x < 192)xi=128;
 else if(x < 384)xi=256;
 else xi=256;
 if     (y <  96)yi=64;
 else if(y < 192)yi=128;
 else if(y < 384)yi=256;
 else yi=256;
 if((px=(unsigned char *)X__Malloc(xi*yi*3)) == NULL){
   X__Free((char *)pixels);
   return 0;
 }
 gluScaleImage(GL_RGB,x,y,GL_UNSIGNED_BYTE,pixels,xi,yi,GL_UNSIGNED_BYTE,px);
 X__Free((char *)pixels);
 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glTexImage2D(GL_TEXTURE_2D,0,3,xi,yi,0,GL_RGB,GL_UNSIGNED_BYTE,
              (GLvoid *)px);
 X__Free((char *)px);
 return 1;
}

#endif

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size,gif_iLace;

static int rowcount,rowinc;

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(y_size - 1 - rowcount)*(long)(x_size);
 memcpy(pixels + l, linepixels, linelen);
 rowcount+=abs(rowinc);
 if(rowcount >= y_size){
  if     (rowinc == -8){rowinc=8; rowcount=4;}
  else if(rowinc ==  8){rowinc=4; rowcount=2;}
  else if(rowinc ==  4){rowinc=2; rowcount=1;}
 }
 return 0;
}

static unsigned char *LoadGIF(char *GIFfile, long *xii, long *yii){
 unsigned char *px,*lpB,*pixels24;
 long i,j,k,imagesize,x,y,xi,yi;
 rowcount=0;
 rowinc=1;
 if(read_gif_image(GIFfile,1) < 0)return 0;
 if(gif_iLace)rowinc = -8;
 x=(x_size+3)/4*4;
 y=y_size;
 imagesize=x_size*y_size;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return 0;
 if((pixels24=(unsigned char*)X__Malloc(x*y_size*3)) == NULL){
   X__Free(pixels);
   return NULL;
 }
 if(read_gif_image(GIFfile,0) < 0){
   X__Free(pixels);
   X__Free(pixels24);
   return NULL;
 }
 lpB=pixels;
 for(j=0;j<y;j++){
   px=(pixels24+j*x*3);
   for(i=0;i<x_size;i++){
     k=(long)(*lpB++);
     *px++ = (unsigned char)(p_red[k]   << 2);
     *px++ = (unsigned char)(p_green[k] << 2);
     *px++ = (unsigned char)(p_blue[k]  << 2);
   }
 }
 X__Free(pixels);
 if     (x <  96)xi=32;
 else if(x < 192)xi=64;
 else if(x < 384)xi=128;
 else xi=128;
 if     (y <  96)yi=32;
 else if(y < 192)yi=64;
 else if(y < 384)yi=128;
 else yi=128;
 if((px=(unsigned char *)X__Malloc(xi*yi*3)) == NULL){
   X__Free((char *)pixels24);
   return NULL;
 }
 gluScaleImage(GL_RGB,x,y,GL_UNSIGNED_BYTE,pixels24,xi,yi,GL_UNSIGNED_BYTE,px);
 X__Free((char *)pixels24);
 *xii = xi;
 *yii = yi;
 return px;
}


static void tga_display_line(unsigned char *prgb,
                             long wide, long line,
                             long upsidedown, long hi){
 register long l,x,k;
 if(upsidedown)l = 3L*(long)(line)*(long)(x_size);
 else          l = 3L*(long)(hi - 1 - line)*(long)(x_size);
 memcpy(pixels + l, prgb, (int)(wide*3));
}

static short R_getvariword(long n, FILE *f, long *vw){
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;          if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

static short read_tga_image(char *filename, short info_only){
 char c0,c1,c2;
 long i,j,k,l,m,n,cmo,cml,xor,yor,wi,hi,iok=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat,wi3;
 FILE *t24;
 unsigned char *prgb;
 prgb=NULL;
 if((t24=fopen(filename,"rb")) != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0 : 1 ;
   if(info_only){
     // for scale image
     x_size=(int)ALIGNULONG(wi);
     y_size=(int)hi;
     fclose(t24);
     return 1;
   }
   if((prgb=(unsigned char *)X__Malloc((size_t)wi*3)) == NULL)goto CLOSEOUT;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi;
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   wi3=wi*3;
   if(image_type == 2){ /* uncompressed */
     m=0; k=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       prgb[m+2] = (char)( vw        & pat) << s;
       prgb[m+1] = (char)((vw >> s1) & pat) << s;
       prgb[m  ] = (char)((vw >> s2) & pat) << s;
       k++; m+=3;
       if(m == wi3){
         tga_display_line(prgb,wi,i,upsidedown,hi); m=0;
       }
     }
   }
   else{  /* compressed */
     m=0; l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           prgb[m+2] = (char)( vw        & pat) << s;
           prgb[m+1] = (char)((vw >> s1) & pat) << s;
           prgb[m  ] = (char)((vw >> s2) & pat) << s;
           k++; m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           prgb[m+2] = c0;
           prgb[m+1] = c1;
           prgb[m  ] = c2;
           k++;  m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
   if(prgb != NULL)X__Free((char *)prgb);
 }
 return (short)iok;
}

static unsigned char *LoadTGA(char *TGAfile, long *xx, long *yy){
 unsigned char *px;
 long imagesize,xi,yi;
 rowcount=0;
 if(read_tga_image(TGAfile,1) < 0)return NULL;
 imagesize=x_size*y_size*3;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return NULL;
 if(read_tga_image(TGAfile,0) < 0){
   X__Free(pixels);
   return NULL;
 }
 if     (x_size <  96)xi=32;
 else if(x_size < 192)xi=64;
 else if(x_size < 384)xi=128;
 else xi=128;
 if     (y_size <  96)yi=32;
 else if(y_size < 192)yi=64;
 else if(y_size < 384)yi=128;
 else yi=128;
 if((px=(unsigned char *)X__Malloc(xi*yi*3)) == NULL){
   X__Free((char *)pixels);
   return NULL;
 }
 gluScaleImage(GL_RGB,x_size,y_size,GL_UNSIGNED_BYTE,
               pixels,xi,yi,GL_UNSIGNED_BYTE,px);
 X__Free((char *)pixels);
 *xx = xi;
 *yy = yi;
 return px;
}

static unsigned char *LoadXXX(char *JPGfile, long *xx, long *yy, long type){
 unsigned char *p,*px,*R,*G,*B,*r,*g,*b;
 long i,j,imagesize,xi,yi;
 long x,y;
 FILE *t24;
 char lname[256];
 HMODULE hLib;
 FARPROC fpFun;
 BOOL (*fpGetMapSize)(char *, long *, long *);
 BOOL (*fpGetMap)(char *, long, long,
                          unsigned char *, unsigned char *, unsigned char *);
 GetModuleFileName(glInstance,lname,255);
 *FileInPath(lname) = '\0';
 if     (type == 0)strcat(lname,"jpg_map.dll");
 else if(type == 1)strcat(lname,"tif_map.dll");
 else if(type == 2)strcat(lname,"png_map.dll");
 R=G=B=NULL;
 hLib=LoadLibrary(lname);
 if(hLib != NULL){
   if((fpFun=GetProcAddress(hLib,"_GetExternalImageSize")) != NULL){
     fpGetMapSize = (void *)fpFun;
     if((*fpGetMapSize)(JPGfile,&x,&y)){
       if((R = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((G = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((B = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((fpFun=GetProcAddress(hLib,"_GetExternalImage")) != NULL){
         fpGetMap = (void *)fpFun;
         if(!(*fpGetMap)(JPGfile,x,y,R,G,B))goto CLOSEOUT;
       }
       else goto CLOSEOUT;
     }
     else goto CLOSEOUT;
   }
   else goto CLOSEOUT;
   FreeLibrary(hLib);
 }
 else return NULL;
// x_size = x; bad when used with scale image and not `long' aligned
 x_size=(int)ALIGNULONG(x);
 y_size = y;
 imagesize=x_size*y_size*3;
 if((p=(unsigned char *)X__Malloc(imagesize+3)) == NULL){
   X__Free(R);
   X__Free(G);
   X__Free(B);
   return NULL;
 }
 for(j=0;j<y;j++){
  px=(p+(y-j-1)*x_size*3);
  r=(R+j*x); g=(G+j*x); b=(B+j*x);
  for(i=0;i<x;i++){
    *px++ = *r++;
    *px++ = *g++;
    *px++ = *b++;
  }
 }
 X__Free(R);
 X__Free(G);
 X__Free(B);
 if     (x_size <  96)xi=32;
 else if(x_size < 192)xi=64;
 else if(x_size < 384)xi=128;
 else xi=128;
 if     (y_size <  96)yi=32;
 else if(y_size < 192)yi=64;
 else if(y_size < 384)yi=128;
 else yi=128;
 if((px=(unsigned char *)X__Malloc(xi*yi*3+3)) == NULL){
   X__Free((char *)p);
   return NULL;
 }
 gluScaleImage(GL_RGB,x_size,y_size,GL_UNSIGNED_BYTE,
               p,xi,yi,GL_UNSIGNED_BYTE,px);
 X__Free((char *)p);
 *xx = xi;
 *yy = yi;
 return px;
 CLOSEOUT:
 if(R != NULL)X__Free(R);
 if(G != NULL)X__Free(G);
 if(B != NULL)X__Free(B);
 FreeLibrary(hLib);
 return NULL;
}

static void LoadPrefs(void){
 double R,G,B;
 char pref_id[64];
 GetPrivateProfileString(IniSection,"FLAT","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",2)  == 0)flat=0;
 if(strncmp(pref_id,"YES",3) == 0)flat=1;
 GetPrivateProfileString(IniSection,"WIRE","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)mode=0;
 if(strncmp(pref_id,"NO",2)  == 0)mode=1;
 GetPrivateProfileString(IniSection,"WIREB","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)bounding_box=1;
 if(strncmp(pref_id,"NO",2)  == 0)bounding_box=0;
 GetPrivateProfileString(IniSection,"COLOUR","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"$$$",3) != 0){
   sscanf(pref_id,"%lf %lf %lf",&R,&G,&B);
   gbRed=(GLfloat)R;  gbGreen=(GLfloat)G;  gbBlue=(GLfloat)B;
 }
 GetPrivateProfileString(IniSection,"MAPPED","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)Mapped=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)Mapped=FALSE;
 GetPrivateProfileString(IniSection,"UPDATE","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)Update=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)Update=FALSE;
 GetPrivateProfileString(IniSection,"FASTEST","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)Fastest=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)Fastest=FALSE;
 GetPrivateProfileString(IniSection,"CLAMPED","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)Clamped=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)Clamped=FALSE;
 GetPrivateProfileString(IniSection,"DEPTHCUE","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)AttenuateLight=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)AttenuateLight=FALSE;
 GetPrivateProfileString(IniSection,"TWOSIDED","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)TwoSided=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)TwoSided=FALSE;
 GetPrivateProfileString(IniSection,"SCROLLBARS","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)ScrollBars=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)ScrollBars=FALSE;
 GetPrivateProfileString(IniSection,"OVERLAY","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)bOverlayWire=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)bOverlayWire=FALSE;
 GetPrivateProfileString(IniSection,"NURBSWIREONLY","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)bNurbsWireOnly=TRUE;
 if(strncmp(pref_id,"NO",2)  == 0)bNurbsWireOnly=FALSE;
 return;
}

static BOOL CALLBACK PrefsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL state;
 static double R,G,B;
 char *w, yesyes[]="YES", nono[]="NO",str[64];
 unsigned char uColour[3];
 switch( msg ) {
   case WM_INITDIALOG:
     CentreDlgOnC(hdlg);
     if(flat)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_FLAT,BM_SETCHECK,state,0);
     if(mode == 0)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_WIRE,BM_SETCHECK,state,0);
     if(bounding_box == 1)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_WIREB,BM_SETCHECK,state,0);
     if(Mapped == TRUE)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_MAPS,BM_SETCHECK,state,0);
     if(Update == TRUE)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_UPDATE,BM_SETCHECK,state,0);
     if(Fastest == TRUE)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_FASTEST,BM_SETCHECK,state,0);
     if(Clamped == TRUE)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_CLAMPED,BM_SETCHECK,state,0);
     if(AttenuateLight == TRUE)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_DEPTHC,BM_SETCHECK,state,0);
     if(bNurbsWireOnly == TRUE)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_NURBSR,BM_SETCHECK,state,0);
     R=(double)gbRed; G=(double)gbGreen; B=(double)gbBlue;
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_PREFS_COLOUR:
          uColour[0]=(unsigned char)(R*255.0);
          uColour[1]=(unsigned char)(G*255.0);
          uColour[2]=(unsigned char)(B*255.0);
          ChooseColour(uColour,"Background Color",hdlg);
          R=(double)uColour[0]/255.0;
          G=(double)uColour[1]/255.0;
          B=(double)uColour[2]/255.0;
          break;
        case IDOK:
          if(SendDlgItemMessage(hdlg,DLG_PREFS_FLAT,BM_GETCHECK,0,0))
            flat=1; else flat=0;
          if(flat == 1)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"FLAT",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_WIRE,BM_GETCHECK,0,0))
            mode=0; else mode=1;
          if(mode == 0)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"WIRE",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_WIREB,BM_GETCHECK,0,0))
            bounding_box=1; else bounding_box=0;
          if(bounding_box == 1)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"WIREB",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_MAPS,BM_GETCHECK,0,0))
            Mapped=TRUE; else Mapped=FALSE;
          if(Mapped == TRUE)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"MAPPED",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_UPDATE,BM_GETCHECK,0,0))
            Update=TRUE; else Update=FALSE;
          if(Update == TRUE)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"UPDATE",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_FASTEST,BM_GETCHECK,0,0))
            Fastest=TRUE; else Fastest=FALSE;
          if(Fastest == TRUE)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"FASTEST",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_CLAMPED,BM_GETCHECK,0,0))
            Clamped=TRUE; else Clamped=FALSE;
          if(Clamped == TRUE)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"CLAMPED",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_DEPTHC,BM_GETCHECK,0,0))
            AttenuateLight=TRUE; else AttenuateLight=FALSE;
          if(AttenuateLight == TRUE)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"DEPTHCUE",w,IniFilename);
          if(SendDlgItemMessage(hdlg,DLG_PREFS_NURBSR,BM_GETCHECK,0,0))
            bNurbsWireOnly=TRUE; else bNurbsWireOnly=FALSE;
          if(bNurbsWireOnly == TRUE)w=yesyes; else w=nono;
          WritePrivateProfileString(IniSection,"NURBSWIREONLY",w,IniFilename);

          gbRed=(GLfloat)R; gbGreen=(GLfloat)G; gbBlue=(GLfloat)B;
          SetMappingQuality();
          glClearColor(gbRed,gbGreen,gbBlue,1.0);
          InvalidateRect(GetParent(hdlg),NULL,FALSE);
          sprintf(str,"%lf %lf %lf",R,G,B);
          WritePrivateProfileString(IniSection,"COLOUR",str,IniFilename);
          EndDialog(hdlg,TRUE);
          return(TRUE);
          break;
        case IDCANCEL:
          EndDialog(hdlg,FALSE);
          return(TRUE);
          break;
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}


