/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* File HEIGHT.C DLL to create a height field */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <windows.h>

#include "struct.h"
#include "callback.h"
#include "height.h"
#include "..\animate\spin_dll.h"

#define DESELECTED 0
#define SELECTED   1
#ifdef __X__NT__
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#else
#define X__Malloc(s)    GlobalAlloc(GMEM_FIXED,s)
#define X__Free(p)      GlobalFree((HGLOBAL)p)
#define X__Realloc(p,s) GlobalReAlloc((HGLOBAL)p,s,GMEM_MOVEABLE)
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size;

int out_line(unsigned char *linepixels, int linelen);

static BOOL CALLBACK HeightFieldDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam);

static HWND      hParent;
static HINSTANCE hThisInstance;
static unsigned char *pixels;
static long inter=0,gridx=10,gridy=10;
static vertex *MainVp;
static edge   *MainEp;
static face   *MainFp;
static long Nvert,Nedge,Nface,NvertSelect;
static void (*CreateVertex)(void);
static void (*CreateEdge)(long, long);
static void (*CreateFace)(long, long, long);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BC__
BOOL WINAPI DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hThisInstance=hDLL;
//      MessageBox (GetFocus(),"Height field attaching","Debug",MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
//      MessageBox (GetFocus(),"Height field detaching","Debug",MB_OK);
      break;
  }
  return TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

BOOL _HeightField
 (HWND parent_window, char *filename,X__STRUCTURE *lpEVI){
 char str[256];
 unsigned char *lpBitmapBits,val,r,g,b;
 long x,y,lx,ly,total,height,fvp,vp;
 long NpointerX,NpointerY,NpointerZ;
 long TVsizeX,TVsizeY,TVsizeZ;
 long TVpointX,TVpointY,TVpointZ;
 hParent=parent_window;
 gridx=10; gridy=10; inter=4;
 if(DialogBox(hThisInstance,MAKEINTRESOURCE(DLG_HEIGHT),GetFocus(),
           (DLGPROC)HeightFieldDlgProc) != OK)return FALSE;
 if(read_gif_image(filename,1) < 0){
   MessageBox ( GetFocus(), (LPCTSTR) "Height Field: Invalid GIF file",
                (LPCTSTR) "", MB_OK | MB_SYSTEMMODAL );
   return FALSE;
 }
// sprintf(str,"File %s size %ld %ld",filename,x_size,y_size);
// MessageBox ( GetFocus(),(LPCTSTR)str,(LPCTSTR) "Processing", MB_OK | MB_SYSTEMMODAL);
 if((lpBitmapBits=(unsigned char *)X__Malloc(x_size*y_size)) == NULL){
   MessageBox ( GetFocus(), (LPCTSTR) "Height Field: Out of memory",
                (LPCTSTR) "", MB_OK | MB_SYSTEMMODAL );
   return FALSE;
 }
 pixels=lpBitmapBits;
 if(read_gif_image(filename,0) < 0){
   MessageBox ( GetFocus(), (LPCTSTR) "Height Field: Error reading GIF file",
                (LPCTSTR) "", MB_OK | MB_SYSTEMMODAL );
   X__Free(lpBitmapBits);
   return FALSE;
 }
 NpointerX = *(lpEVI->NpointerX);
 NpointerY = *(lpEVI->NpointerY);
 NpointerZ = *(lpEVI->NpointerZ);
 TVsizeX = *(lpEVI->TVsizeX);
 TVsizeY = *(lpEVI->TVsizeY);
 TVsizeZ = *(lpEVI->TVsizeZ);
 TVpointX = *(lpEVI->TVpointX);
 TVpointY = *(lpEVI->TVpointY);
 TVpointZ = *(lpEVI->TVpointZ);
 Nvert =       *(lpEVI->Nvert);
 Nedge =       *(lpEVI->Nedge);
 Nface =       *(lpEVI->Nface);
 NvertSelect = *(lpEVI->NvertSelect);
 CreateVertex = lpEVI->fpCV;
 CreateEdge   = lpEVI->fpCE;
 CreateFace   = lpEVI->fpCF;
 SendMessage(hParent,WM_COMMAND,IDC_UPDATE_VERTEXHEAP,
            (LPARAM)(Nvert+gridx*gridy));
 SendMessage(hParent,WM_COMMAND,IDC_UPDATE_EDGEHEAP,
            (LPARAM)(Nedge+3*gridx*gridy-2*(gridx+gridy)+1));
 SendMessage(hParent,WM_COMMAND,IDC_UPDATE_FACEHEAP,
            (LPARAM)(Nface+2*(gridx-1)*(gridy-1)));
 MainVp = *(lpEVI->MainVp);
 MainEp = *(lpEVI->MainEp);
 MainFp = *(lpEVI->MainFp);
 for(y=0; y<(y_size/gridy)*gridy; y+=y_size/gridy){
   for(x=0; x<(x_size/gridx)*gridx; x+=x_size/gridx){
     total=0;
     for(ly=0; ly<y_size/gridy; ++ly){
       for(lx=0; lx<x_size/gridx; ++lx){
         val = *(lpBitmapBits+(long)((long)(y+ly)*x_size+(x+lx)));
         r=(unsigned char)p_red[val] << 2;
         g=(unsigned char)p_green[val] << 2;
         b=(unsigned char)p_blue[val] << 2;
         switch(inter){
           case 0: height=r;                              break;
           case 1: height=g;                              break;
           case 2: height=b;                              break;
           case 3: height=max(r,g); height=max(height,b); break;
           case 4: height=(r+g+b)/3;                      break;
         }
         total+=height;
       }
     }
     total /= (y_size/gridy) * (x_size/gridx);
     CreateVertex();
     Nvert =       *(lpEVI->Nvert);
     NvertSelect = *(lpEVI->NvertSelect);
     if(x == 0 && y == 0)fvp=Nvert-1;  /* Pick out first vertex in mesh */
     (MainVp+Nvert-1)->xyz[0] = TVpointX + (long)(x*TVsizeX)/(long)x_size;
     (MainVp+Nvert-1)->xyz[1] = TVpointY+TVsizeY-(long)(y*TVsizeY)/(long)y_size;
     (MainVp+Nvert-1)->xyz[2] = NpointerZ + (NpointerZ+TVsizeZ)/(4*256) * total;
   }
 }
 vp=fvp;
 for(y=0;y<gridy;y++){
   for(x=0;x<gridx;x++){
     if(x > 0){
       CreateEdge(vp,vp-1); Nedge = *(lpEVI->Nedge);
     }
     if(y > 0){
       CreateEdge(vp,vp-gridx); Nedge = *(lpEVI->Nedge);
       if(x > 0){
         CreateEdge(vp,vp-gridx-1);          Nedge = *(lpEVI->Nedge);
         CreateFace(vp,vp-gridx,vp-gridx-1); Nface = *(lpEVI->Nface);
         CreateFace(vp,vp-gridx-1,vp-1);     Nface = *(lpEVI->Nface);
       }
     }
     vp++;
   }
 }
 X__Free(lpBitmapBits);
 return TRUE;
}

static BOOL CALLBACK HeightFieldDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 int i;
 BOOL err;
 switch( msg ) {
   case WM_INITDIALOG:
     SetDlgItemInt(hwnd,DLG_HEIGHT_COLUMNS,gridx,FALSE);
     SendDlgItemMessage(hwnd,DLG_HEIGHT_COLUMNS,EM_LIMITTEXT,(WPARAM)6,0);
     SendDlgItemMessage(hwnd,DLG_HEIGHT_COLUMNSS,SPNM_SETRANGE,0,
                        MAKELPARAM(2,200));
     SendDlgItemMessage(hwnd,DLG_HEIGHT_COLUMNSS,SPNM_SETCRNTVALUE,
                        (WPARAM)gridx,0);
     SendDlgItemMessage(hwnd,DLG_HEIGHT_COLUMNSS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hwnd,DLG_HEIGHT_COLUMNS));
     SetDlgItemInt(hwnd,DLG_HEIGHT_ROWS,gridy,FALSE);
     SendDlgItemMessage(hwnd,DLG_HEIGHT_ROWS,EM_LIMITTEXT,(WPARAM)6,0);
     SendDlgItemMessage(hwnd,DLG_HEIGHT_ROWSS,SPNM_SETRANGE,0,
                        MAKELPARAM(2,200));
     SendDlgItemMessage(hwnd,DLG_HEIGHT_ROWSS,SPNM_SETCRNTVALUE,
                        (WPARAM)gridy,0);
     SendDlgItemMessage(hwnd,DLG_HEIGHT_ROWSS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hwnd,DLG_HEIGHT_ROWS));
     SendDlgItemMessage(hwnd,DLG_HEIGHT_LOW,BM_SETCHECK,TRUE,0);
     SendDlgItemMessage(hwnd,DLG_HEIGHT_AVERAGE,BM_SETCHECK,TRUE,0);
     SendMessage(hParent,WM_COMMAND,IDC_CENTRE_DIALOG_CURSOR,(LPARAM)hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_HEIGHT_RED:     inter=0; break;
        case DLG_HEIGHT_GREEN:   inter=1; break;
        case DLG_HEIGHT_BLUE:    inter=2; break;
        case DLG_HEIGHT_MAX:     inter=3; break;
        case DLG_HEIGHT_AVERAGE: inter=4; break;
        case DLG_HEIGHT_LOW:     i=10;  goto SETSIZE;
        case DLG_HEIGHT_MEDIUM:  i=40;  goto SETSIZE;
        case DLG_HEIGHT_HIGH:    i=80;
          SETSIZE:
          gridx=gridy=i;
          SetDlgItemInt(hwnd,DLG_HEIGHT_COLUMNS,gridx,FALSE);
          SendDlgItemMessage(hwnd,DLG_HEIGHT_COLUMNSS,SPNM_SETCRNTVALUE,
                             (WPARAM)gridx,0);
          SetDlgItemInt(hwnd,DLG_HEIGHT_ROWS,gridy,FALSE);
          SendDlgItemMessage(hwnd,DLG_HEIGHT_ROWSS,SPNM_SETCRNTVALUE,
                             (WPARAM)gridy,0);
          break;
        case DLG_HEIGHT_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case DLG_HEIGHT_OK:
          i=GetDlgItemInt(hwnd,DLG_HEIGHT_COLUMNS,&err,FALSE);
          if(!err)EndDialog(hwnd, FAIL);
          if(i > 200 || i < 2)EndDialog(hwnd, FAIL);
          gridx=i;
          i=GetDlgItemInt(hwnd,DLG_HEIGHT_ROWS,&err,FALSE);
          if(!err)EndDialog(hwnd, FAIL);
          if(i > 200 || i < 2)EndDialog(hwnd, FAIL);
          gridy=i;
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

int out_line(unsigned char *linepixels, int linelen){
 memcpy(pixels, linepixels, linelen);
 pixels += linelen;
 return 0;
}

