/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  ISOSURF.CPP  */

#include "vtk.hh"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#define UNIT         32768L
#define DESELECTED       0
#define SELECTED         1
#define INEDITOR         6

#ifdef __cplusplus
extern "C" {
#endif

#include "struct.h"
#include "dstruct.h"

#ifdef __cplusplus
}
#endif

#include "isosurf.h"

#include "sfx2vtk.h"
#include "sfx2vtk.cpp"

static HWND      hParent;
static HINSTANCE hThisInstance;
static BOOL  inc_c1=TRUE, inc_c2=FALSE, inc_c3=FALSE, inc_c4=FALSE;
static float c1_val=0.0, c2_val=0.0, c3_val=0.0, c4_val=0.0;

static BOOL LocalSelectFileName(char *szfile, char *szfilter,
                           HWND parent, HINSTANCE hInst);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=(HINSTANCE)hDLL;
      if(hDLL == NULL)MessageBeep(MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

extern "C" BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 HCURSOR hSave;
 char localfile[256];
 lpEVI=lpevi;
 localfile[0]='\0';
 hParent=parent_window;
 if(!LocalSelectFileName(localfile,
                        "VTK Structured Points Datasets (*.VTK)|*.vtk",
                         parent_window,hThisInstance))return FALSE;
 hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
 {
  int i,j,k,Offset,jOffset,kOffset;
  float x,y,z,s;
  SfxWriter sfxoutput;

#if 0
#define SAMPLE_SIZE 26
  vtkStructuredPoints *vol = new vtkStructuredPoints;
  vol->SetDimensions(SAMPLE_SIZE,SAMPLE_SIZE,SAMPLE_SIZE);
  vol->SetOrigin(-0.5,-0.5,-0.5);
  float ar=1.0/((float)SAMPLE_SIZE-1.0);
  vol->SetAspectRatio(ar,ar,ar);
  vtkFloatScalars *scalars = new vtkFloatScalars(SAMPLE_SIZE*SAMPLE_SIZE*SAMPLE_SIZE);
  for(k=0;k<SAMPLE_SIZE;k++){
    z= -0.5 + k*ar;
    kOffset = k*SAMPLE_SIZE*SAMPLE_SIZE;
    for(j=0;j<SAMPLE_SIZE;j++){
      y = -0.5 + j*ar;
      jOffset = j*SAMPLE_SIZE;
      for(i=0;i<SAMPLE_SIZE;i++){
        x = - 0.5 + i*ar;
        s = x*x + y*y + z*z - (0.4*0.4);
        Offset = i + jOffset + kOffset;
        scalars->InsertScalar(Offset,s);
      }
    }
  }
  vol->GetPointData()->SetScalars(scalars);
  scalars->Delete();
  vtkDataSetWriter dw;
  dw.SetFilename("F:\\temp1.vtk");
  dw.SetInput(vol);
  dw.Update();
#endif

  vtkStructuredPointsReader *volR = new vtkStructuredPointsReader;
  volR->SetFilename(localfile);

  vtkContourFilter *cf = new vtkContourFilter;
  cf->SetInput(volR->GetOutput());
//  cf->SetInput(vol);  for structured points
  if(inc_c1)cf->SetValue(0,c1_val);
  if(inc_c2)cf->SetValue(0,c2_val);
  if(inc_c3)cf->SetValue(0,c3_val);
  if(inc_c4)cf->SetValue(0,c4_val);

  vtkCleanPolyData *cp = new vtkCleanPolyData;
  cp->SetTolerance(0.0001);
  cp->SetInput(cf->GetOutput());

  vtkTriangleFilter *tf = new vtkTriangleFilter;
  tf->SetInput(cp->GetOutput());

  sfxoutput.SetInput(tf->GetOutput());
  sfxoutput.Write();

  delete tf;
  delete cp;
  delete cf;
  delete volR;
 }
 SetCursor(hSave);
 return TRUE;
}

static unsigned CALLBACK NameHookProc(HWND hwnd,UINT msg,
                              WPARAM wparam,LPARAM lparam){
 char str[32];
 BOOL err;
 LPOFNOTIFY lp;
 switch( msg ) {
   case WM_INITDIALOG:
     if(inc_c1)SendDlgItemMessage(hwnd,DLG_C1,BM_SETCHECK,1,0);
     if(inc_c2)SendDlgItemMessage(hwnd,DLG_C2,BM_SETCHECK,1,0);
     if(inc_c3)SendDlgItemMessage(hwnd,DLG_C3,BM_SETCHECK,1,0);
     if(inc_c4)SendDlgItemMessage(hwnd,DLG_C4,BM_SETCHECK,1,0);
     sprintf(str,"%.2lf",c1_val);
     SetDlgItemText(hwnd,DLG_C1_V,str);
     sprintf(str,"%.2lf",c2_val);
     SetDlgItemText(hwnd,DLG_C2_V,str);
     sprintf(str,"%.2lf",c3_val);
     SetDlgItemText(hwnd,DLG_C3_V,str);
     sprintf(str,"%.2lf",c4_val);
     SetDlgItemText(hwnd,DLG_C4_V,str);
     return FALSE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDOK:{
           lp=NULL;
           goto JUMPHERE;
         }
         break;
       default: break;
     }
     break;
   case WM_NOTIFY:{
       lp=(LPOFNOTIFY)lparam;
       if(lp->hdr.code == CDN_FILEOK){
         JUMPHERE:
         if(SendDlgItemMessage(hwnd,DLG_C1,BM_GETCHECK,0,0))
           inc_c1=TRUE; else inc_c1=FALSE;
         if(SendDlgItemMessage(hwnd,DLG_C2,BM_GETCHECK,0,0))
           inc_c2=TRUE; else inc_c2=FALSE;
         if(SendDlgItemMessage(hwnd,DLG_C3,BM_GETCHECK,0,0))
           inc_c3=TRUE; else inc_c3=FALSE;
         if(SendDlgItemMessage(hwnd,DLG_C4,BM_GETCHECK,0,0))
           inc_c4=TRUE; else inc_c4=FALSE;
         GetDlgItemText(hwnd,DLG_C1_V,str,255);
         if(inc_c1)c1_val=atof(str);
         GetDlgItemText(hwnd,DLG_C2_V,str,255);
         if(inc_c2)c2_val=atof(str);
         GetDlgItemText(hwnd,DLG_C3_V,str,255);
         if(inc_c3)c3_val=atof(str);
         GetDlgItemText(hwnd,DLG_C4_V,str,255);
         if(inc_c4)c4_val=atof(str);
       }
     }
     break;
   default: break;
 }
 return FALSE;
}

static BOOL LocalSelectFileName(char *szfile, char *szfilter,
                           HWND parent, HINSTANCE hInst){
 int i;
 OPENFILENAME ofn;
 char szFilter[80],title[256];
 DWORD version;
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=0;
 ofn.lpstrFile=szfile;
 ofn.nMaxFile=255;
 ofn.lpstrFileTitle=NULL;
 ofn.nMaxFileTitle=0;
 ofn.lpstrInitialDir=NULL;
 LoadString(hThisInstance,IDX_VTK_STRING1,title,255);
 ofn.lpstrTitle=title;
 ofn.hInstance=hInst;
 ofn.lpfnHook=NameHookProc;
 version=GetVersion();
 ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|
           OFN_ENABLEHOOK|OFN_ENABLETEMPLATE|OFN_NONETWORKBUTTON;
 if(LOBYTE(LOWORD(version)) >= 4){
   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_VTK);
   ofn.Flags |= OFN_EXPLORER;
 }
 else{
   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_VTKOLD);
 }
 return GetOpenFileName(&ofn);
}
