/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file LIBRARY.C     library of functions that may be helpful */

// include some simple dialogs and drawing functions

#define MODULE_LIBS 1

#include "design.h"
#include "..\animate\spin_dll.h"

extern long LastToolVertex;

static long local_grid_size=UNIT;

typedef struct tagNUMENTRY {
  short low,high,val;
  char *title;
  char *message;
} NUMENTRY;

static BOOL CALLBACK NumEntryDlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam ){
 NUMENTRY *dp;
 BOOL err;
 int i;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(NUMENTRY *)lparam;
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)dp->title);
       SendDlgItemMessage(hwnd,DLG_NUMENTRY_TEXT,WM_SETTEXT,0,
                          (LPARAM)dp->message);
       SetDlgItemInt(hwnd,DLG_NUMENTRY_EDIT,dp->val,FALSE);
       SendDlgItemMessage(hwnd,DLG_NUMENTRY_SPIN,SPNM_SETRANGE,0,
                          MAKELPARAM(dp->low,dp->high));
       SendDlgItemMessage(hwnd,DLG_NUMENTRY_SPIN,SPNM_SETCRNTVALUE,
                          (WPARAM)dp->val,0);
       SendDlgItemMessage(hwnd,DLG_NUMENTRY_SPIN,SPNM_SETEDITCTRL,0,
                          (LPARAM)GetDlgItem(hwnd,DLG_NUMENTRY_EDIT));
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      dp=(NUMENTRY *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_NUMENTRY_SPIN:
          switch(HIWORD(wparam)){
            case SPNN_VALUECHANGE:
            break;
            default: break;
          }
          break;
        case IDCANCEL:
        case DLG_NUMENTRY_CANCEL:
          EndDialog(hwnd, -1);
          return(TRUE);
        case IDOK:
        case DLG_NUMENTRY_OK:
          i=GetDlgItemInt(hwnd,DLG_NUMENTRY_EDIT,&err,FALSE);
          if(!err)EndDialog(hwnd, -1);
          if(i > dp->high)i=dp->high; if(i < dp->low)i=dp->low;
          EndDialog(hwnd,i);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

short RequestNumEntry(short val, short minval, short maxval,
                      char *title, char *message){
 int i;
 NUMENTRY Numentry;
 Numentry.low=minval;
 Numentry.high=maxval;
 Numentry.val=val;
 Numentry.title=title;
 Numentry.message=message;
 EnableToolPannels(ALL_PANNELS,FALSE);
 i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_NUMENTRY),ghwnd_main,
            (DLGPROC)NumEntryDlgProc,(LPARAM)&Numentry);
 EnableToolPannels(ALL_PANNELS,TRUE);
 return (short)i;
}

typedef struct tagSTRINGDLGLIST {
  int  len;
  char *title;
  char *string;
} STRINGDLGLIST;

static BOOL CALLBACK StringDlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam ){
 HWND hctl;
 STRINGDLGLIST *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(STRINGDLGLIST *)lparam;
       hctl=GetDlgItem(hwnd,DLG_STRING_EDIT);
       SendMessage(hctl,WM_SETTEXT,0,(LPARAM)dp->string);
       SendMessage(hctl,EM_LIMITTEXT,(WPARAM)dp->len,0);
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)dp->title);
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
        case DLG_STRING_OK:
          dp=(STRINGDLGLIST *)GetWindowLong(hwnd,GWL_USERDATA);
          hctl=GetDlgItem(hwnd,DLG_STRING_EDIT);
          SendMessage(hctl,WM_GETTEXT,(WPARAM)dp->len,(LPARAM)dp->string);
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

int RequestCharString(int len, char *string, char *title, HWND parent){
 char tempstr[255];
 STRINGDLGLIST StringDlgList;
 strcpy(tempstr,string);
 StringDlgList.len=len;
 StringDlgList.string=tempstr;
 StringDlgList.title=title;
 if(DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_STRING),parent,
            (DLGPROC)StringDlgProc,(LPARAM)&StringDlgList)){
    strcpy(string,tempstr);
    return 1;
 }
 return 0;
}

void EnableToolPannels(int what_one,BOOL status){
  if(what_one & MAIN_WINDOW   )EnableWindow(ghwnd_main,status);
//  if(what_one & TOOL_PANNEL   )EnableTools(status);
  if(what_one & COORD_PANNEL  )EnableWindow(ghwndCoord1,status);
}

typedef struct tagSELECTDLGLIST {
  int n;
  char *title;
  char **list;
} SELECTDLGLIST;

BOOL CALLBACK ItemSelectDlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam ){
 int i;
 SELECTDLGLIST *dp;
 HWND hctl;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       dp=(SELECTDLGLIST *)lparam;
       hctl=GetDlgItem(hwnd,DLG_CHOOSE_LIST);
       for(i=0;i<dp->n;i++)SendMessage(hctl,LB_ADDSTRING,0,(LPARAM)dp->list[i]);
       SendMessage(hctl,LB_SETCURSEL,0,0);
       SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)dp->title);
       CentreDialogOnCursor(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_CHOOSE_LIST:
          switch(HIWORD(wparam)){
            case LBN_DBLCLK:
            SendMessage(hwnd,WM_COMMAND,MAKEWPARAM(DLG_CHOOSE_OK,0),0);
              break;
            default:
              break;
          }
          break;
        case IDCANCEL:
        case DLG_CHOOSE_CANCEL:
          EndDialog(hwnd, -1);
          return (TRUE);
        case IDOK:
        case DLG_CHOOSE_OK:
          hctl=GetDlgItem(hwnd,DLG_CHOOSE_LIST);
          i=SendMessage(hctl,LB_GETCURSEL,0,0);
          EndDialog(hwnd,i);
          return(TRUE);
        default:
          break;
      }
      break;
 }
 return(FALSE);
}


int SelectScrolledItemList(int n, char **list, char *title, HWND parent){
 int i;
 SELECTDLGLIST SelectDlgList;
 SelectDlgList.n=n;
 SelectDlgList.list=list;
 SelectDlgList.title=title;
 EnableToolPannels(ALL_PANNELS,FALSE);
 i=DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_CHOOSE),parent,
            (DLGPROC)ItemSelectDlgProc,(LPARAM)&SelectDlgList);
 EnableToolPannels(ALL_PANNELS,TRUE);
 return i;
}

void  ActivateAllMenus(HWND hwnd, UINT fuFlags){
 static WORD enable_state[256];  /* must be >= number of menu commands */
 int id,i,j,k,n,m,o,c=0;
 HMENU hmenu,hsubmenu,hsubsubmenu;
 hmenu=GetMenu(hwnd);
 n=GetMenuItemCount(hmenu);
 if(n < 1)return;
 for(i=0;i<n;i++){
  if(i > 1)continue; /* don't fiddle with these menus */
  if((hsubmenu=GetSubMenu(hmenu,i)) == NULL){
    if((id=GetMenuItemID(hmenu,i)) < 0xffffffff){
      if(fuFlags == MF_GRAYED){
        enable_state[c] = GetMenuState(hmenu,id,MF_BYCOMMAND);
        EnableMenuItem(hmenu,id,fuFlags);
      }
      else EnableMenuItem(hmenu,id,enable_state[c]);
      c++;
    }
  }
  else{
    m=GetMenuItemCount(hsubmenu);
    for(j=0;j<m;j++){
      if(i == 0 && j >= 0)continue;
      if(i == 1 && (j ==  0 || j ==  4 || j == 5  ||
                    j ==  7 || j ==  9 || j == 10 ||
                    j == 11 || j == 12))continue;
      if((hsubsubmenu=GetSubMenu(hsubmenu,j)) == NULL){
        if((id=GetMenuItemID(hsubmenu,j)) < 0xffffffff){
          if(fuFlags == MF_GRAYED){
            enable_state[c] = GetMenuState(hsubmenu,id,MF_BYCOMMAND);
            EnableMenuItem(hsubmenu,id,fuFlags);
          }
          else EnableMenuItem(hsubmenu,id,enable_state[c]);
          c++;
        }
      }
      else{
        o=GetMenuItemCount(hsubsubmenu);
        for(k=0;k<o;k++){
          if((id=GetMenuItemID(hsubsubmenu,k)) < 0xffffffff){
            if(fuFlags == MF_GRAYED){
              enable_state[c] = GetMenuState(hsubsubmenu,id,MF_BYCOMMAND);
              EnableMenuItem(hsubsubmenu,id,fuFlags);
            }
            else EnableMenuItem(hsubsubmenu,id,enable_state[c]);
            c++;
          }
        }
      }
    }
  }
 }
}

PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

#define ATOL 0.5

void GetWorldCoords0(int W, long *x, long *y, long *z,
                     int hpos, int vpos);
void GetWindowCoords0(int W, long x, long y, long z,
                     int *hpos, int *vpos);
void GetWorldCoords1(int W, long *x, long *y, long *z,
                    int hpos, int vpos);
void GetWindowCoords1(int W, long x, long y, long z,
                     int *hpos, int *vpos);


void GetWorldCoords0(int W, long *x, long *y, long *z,
                    int hpos, int vpos){
  if(W == TRITOP){
    *x=(long)((double)(hpos-WindowCentreX[0])*WindowPixelSize)+TVcentreX;
    *y=(long)((double)(WindowCentreY[0]-vpos)*WindowPixelSize)+TVcentreY;
    *z=NpointerZ;
  }
  else if(W == TRIFRONT){
    *x=(long)((double)(hpos-WindowCentreX[1])*WindowPixelSize)+TVcentreX;
    *y=NpointerY;
    *z=(long)((double)(WindowCentreY[1]-vpos)*WindowPixelSize)+TVcentreZ;
  }
  else if(W == TRIRIGHT){
    *x=NpointerX;
    *y=(long)((double)(hpos-WindowCentreX[2])*WindowPixelSize)+TVcentreY;
    *z=(long)((double)(WindowCentreY[2]-vpos)*WindowPixelSize)+TVcentreZ;
  }
}

void GetWindowCoords0(int W, long x, long y, long z, int *hpos, int *vpos){
  if(W == TRITOP){
   *hpos=(int)((double)WindowCentreX[0]+(double)(x-TVcentreX)/WindowPixelSize+ATOL);
   *vpos=(int)((double)WindowCentreY[0]-(double)(y-TVcentreY)/WindowPixelSize+ATOL);
  }
  else if(W == TRIFRONT){
   *hpos=(int)((double)WindowCentreX[1]+(double)(x-TVcentreX)/WindowPixelSize+ATOL);
   *vpos=(int)((double)WindowCentreY[1]-(double)(z-TVcentreZ)/WindowPixelSize+ATOL);
  }
  else if(W == TRIRIGHT){
   *hpos=(int)((double)WindowCentreX[2]+(double)(y-TVcentreY)/WindowPixelSize+ATOL);
   *vpos=(int)((double)WindowCentreY[2]-(double)(z-TVcentreZ)/WindowPixelSize+ATOL);
  }
}

void GetWorldCoords1(int W, long *x, long *y, long *z,
                    int hpos, int vpos){
  if(W == TRITOP){
    *x=(long)((double)(WindowCentreX[0]-hpos)*WindowPixelSize)+TVcentreX;
    *y=(long)((double)(vpos-WindowCentreY[0])*WindowPixelSize)+TVcentreY;
    *z=NpointerZ;
  }
  else if(W == TRIFRONT){
    *x=(long)((double)(WindowCentreX[1]-hpos)*WindowPixelSize)+TVcentreX;
    *y=NpointerY;
    *z=(long)((double)(WindowCentreY[1]-vpos)*WindowPixelSize)+TVcentreZ;
  }
  else if(W == TRIRIGHT){
    *x=NpointerX;
    *y=(long)((double)(WindowCentreX[2]-hpos)*WindowPixelSize)+TVcentreY;
    *z=(long)((double)(WindowCentreY[2]-vpos)*WindowPixelSize)+TVcentreZ;
  }
}

void GetWindowCoords1(int W, long x, long y, long z,
                     int *hpos, int *vpos){
 double wdx,wdy;
  if(W == TRITOP){
   *hpos=(int)((double)WindowCentreX[0]-(double)(x-TVcentreX)/WindowPixelSize+ATOL);
   *vpos=(int)((double)WindowCentreY[0]+(double)(y-TVcentreY)/WindowPixelSize+ATOL);
  }
  else if(W == TRIFRONT){
   *hpos=(int)((double)WindowCentreX[1]-(double)(x-TVcentreX)/WindowPixelSize+ATOL);
   *vpos=(int)((double)WindowCentreY[1]-(double)(z-TVcentreZ)/WindowPixelSize+ATOL);
  }
  else if(W == TRIRIGHT){
   *hpos=(int)((double)WindowCentreX[2]-(double)(y-TVcentreY)/WindowPixelSize+ATOL);
   *vpos=(int)((double)WindowCentreY[2]-(double)(z-TVcentreZ)/WindowPixelSize+ATOL);
  }
}

void SetUpWindowBoxView(int view){
 if(view == 0){
   GetWorldCoords  = GetWorldCoords0;
   GetWindowCoords = GetWindowCoords0;
   WindowBox_view=0;
 }
 else if(view == 1){
   WindowBox_view=1;
   GetWorldCoords  = GetWorldCoords1;
   GetWindowCoords = GetWindowCoords1;
 }
}

void GetTriview(BOOL reset){
 double scale;
 if(reset){
   TVsizeX=TVsizeY=TVsizeZ=UNIT2;
   TVcentreX=TVcentreY=TVcentreZ=0;
 }
 if(View == TRIVIEW){
   scale=(double)WindowSizeX[1]/(double)WindowSizeY[1];
   TVsizeX=(long)((double)TVsizeZ*scale);
   scale=(double)WindowSizeX[2]/(double)WindowSizeY[1];
   TVsizeY=(long)((double)TVsizeZ*scale);
   scale=(double)WindowSizeY[0]/(double)WindowSizeX[1];
   TVsizeY=max(TVsizeY,(long)((double)TVsizeX*scale));
 }
 else{
   scale=(double)WindowSizeX[ActiveView]/(double)WindowSizeY[ActiveView];
   if     (ActiveView == TRITOP){
     TVsizeX=(long)((double)TVsizeZ*scale);
   }
   else if(ActiveView == TRIFRONT){
     TVsizeY=TVsizeX=(long)((double)TVsizeZ*scale);
   }
   else if(ActiveView == TRIRIGHT){
     TVsizeY=TVsizeX=(long)((double)TVsizeZ*scale);
   }
 }
 TVpointX=(TVcentreX-TVsizeX/2);
 TVpointY=(TVcentreY-TVsizeY/2);
 TVpointZ=(TVcentreZ-TVsizeZ/2);
 WindowPixelSize=(double)TVsizeZ/(double)WindowSizeY[1];
}

void reset_mod_maxview(short option){
 point TVC,TVP;
 if(Nvert == 0 && Nnurbs == 0)return;
 get_centre(option,TVC,TVP,TRUE);           /* redraw full zoom */
 TVsizeX=2*(TVC[0]-TVP[0]);  NpointerX=TVcentreX=TVC[0];
 TVsizeY=2*(TVC[1]-TVP[1]);  NpointerY=TVcentreY=TVC[1];
 TVsizeZ=2*(TVC[2]-TVP[2]);  NpointerZ=TVcentreZ=TVC[2];
 TVsizeX += TVsizeX/5;
 TVsizeY += TVsizeY/5;
 TVsizeZ += TVsizeZ/5;
 GetTriview(FALSE);  /* this will adjust the TV size */
}

void get_centre(short option, point p, point TVp, BOOL all){
 vertex *vp;
 double x,y,z,an;
 short flag;
 long i,n,xmin,xmax,ymin,ymax,zmin,zmax,d;
 flag=0;
 if     (option == 0){x=0; y=0; z=0; n=0;}
 else if(option >= 1){xmin=ymin=zmin=MAXUNIT; xmax=ymax=zmax = -MAXUNIT;}
 vp=MainVp; if(MainVp != NULL && Nvert > 0)for(i=0;i<Nvert;i++){
   if(vp->status == SELECTED || all == TRUE){
     flag=1;
     if(option == 0){ /* centroid */
       n++; an=(double)(n-1)/(double)n;
       x=x*an+(double)(vp->xyz[0])/(double)n;
       y=y*an+(double)(vp->xyz[1])/(double)n;
       z=z*an+(double)(vp->xyz[2])/(double)n;
     }
     else if(option >= 1){ /* average */
      if(vp->xyz[0] < xmin)xmin=vp->xyz[0];
      if(vp->xyz[1] < ymin)ymin=vp->xyz[1];
      if(vp->xyz[2] < zmin)zmin=vp->xyz[2];
      if(vp->xyz[0] > xmax)xmax=vp->xyz[0];
      if(vp->xyz[1] > ymax)ymax=vp->xyz[1];
      if(vp->xyz[2] > zmax)zmax=vp->xyz[2];
     }
   }
   vp++;
 }
 if(option >= 1)IncludeNurbsBoundary(&flag,&xmin,&xmax,&ymin,&ymax,&zmin,&zmax);
 if(flag == 0){ /* no points found to average */
   p[0]=NpointerX; p[1]=NpointerY; p[2]=NpointerZ;
   TVp[0]=TVpointX; TVp[1]=TVpointY; TVp[2]=TVpointZ;
   return;
 }
 else if(option == 0){ /* centroid */
   p[0]=(long)x; p[1]=(long)y; p[2]=(long)z;
 }
 else if(option == 1){  /* midpoint */
  p[0]=(xmax+xmin)/2; p[1]=(ymax+ymin)/2; p[2]=(zmax+zmin)/2;
 }
 else if(option == 2){ /* horizontal xy plane */
  p[0]=(xmax+xmin)/2; p[1]=(ymax+ymin)/2; p[2]=NpointerZ;
 }
 else if(option == 3){ /* vertical F/B */
  p[0]=NpointerX; p[1]=(ymax+ymin)/2; p[2]=(zmax+zmin)/2;
 }
 else if(option == 4){ /* vertical L/R */
  p[0]=(xmax+xmin)/2; p[1]=NpointerY; p[2]=(zmax+zmin)/2;
 }
 d=max(labs(xmax-xmin),labs(ymax-ymin)); d=max(d,labs(zmax-zmin));
 d=max(d,MINUNIT); d=min(d,MAXUNIT); d /= 2;
 TVp[0]=p[0]-d; TVp[1]=p[1]-d; TVp[2]=p[2]-d;
}

void Draw3dCursor(void){
 int i;
 HDC hdc;
 int ifd,ild;
 if(View == TRIVIEW){ifd=0;ild=3;}
 else               {ifd=ActiveView; ild=ifd+1;}
 for(i=ifd;i<ild;i++){
   hdc=GetDC(ghwnd_triview[i]);
   SelectPalette(hdc,ghpaletteScreen,FALSE);
   RealizePalette (hdc);
   DrawOne3dCursor(hdc,i);
   ReleaseDC(ghwnd_triview[i],hdc);
 }
}

void DrawOne3dCursor(HDC hdc, int i){
 int sho,svo,oldROP;
 HPEN holdpen;
 GetWindowCoords(i,NpointerX,NpointerY,NpointerZ,&sho,&svo);
 oldROP=SetROP2(hdc,R2_XORPEN);
 holdpen=SelectObject(hdc,ghCursorPen);
 if(CursorToggle == 0){
   MoveToEx(hdc,sho-10,svo,NULL); LineTo(hdc,sho-3,svo);
   MoveToEx(hdc,sho+3,svo,NULL);  LineTo(hdc,sho+10,svo);
   MoveToEx(hdc,sho,svo-10,NULL); LineTo(hdc,sho,svo-3);
   MoveToEx(hdc,sho,svo+3,NULL);  LineTo(hdc,sho,svo+10);
 }
 else{
   MoveToEx(hdc,0,svo,NULL); LineTo(hdc,WindowSizeX[i],svo);
   MoveToEx(hdc,sho,0,NULL); LineTo(hdc,sho,WindowSizeY[i]);
 }
 SelectObject(hdc,holdpen);
 SetROP2(hdc,oldROP);
}

void Move3dCursor(int stor, int xw, int yw){
 int i;
 HDC hdc;
 HWND hwnd;
 static int xws=0,yws=0;
 long x,y,z,g2;
 int ifd,ild,dx,dy,dxa,dya;
 if((GetKeyState(VK_SHIFT) & 0x8000) && (tool == PLOTTER || tool == PLOT3D)){
   if(LastToolVertex >= 0){
     x=(MainVp+LastToolVertex)->xyz[0];
     y=(MainVp+LastToolVertex)->xyz[1];
     z=(MainVp+LastToolVertex)->xyz[2];
     GetWindowCoords(ActiveView,x,y,z,&xws,&yws);
   }
   else{
     GetWindowCoords(ActiveView,NpointerX,NpointerY,NpointerZ,&xws,&yws);
   }
   dx=xw-xws; dy=yw-yws;  dxa=abs(dx); dya=abs(dy);
   if(dya == 0)dya=1;
   if(dxa == 0)dxa=1;
   if     (dya < dxa/2)yw=yws;
   else if(dxa < dya/2)xw=xws;
   else if(dya < dxa)yw=yws+(int)((double)dxa*(double)dy/(double)dya);
   else if(dxa < dya)xw=xws+(int)((double)dya*(double)dx/(double)dxa);
 }
 else if(stor){
   xws=xw; yws=yw;
 }
 else if((GetKeyState(VK_SHIFT) & 0x8000) &&
         (tool == GRABBER || tool == NODETOOL)){
   dx=xw-xws; dy=yw-yws;  dxa=abs(dx); dya=abs(dy);
   if(dya == 0)dya=1;
   if(dxa == 0)dxa=1;
   if     (dya < dxa/2)yw=yws;
   else if(dxa < dya/2)xw=xws;
   else if(dya < dxa)yw=yws+(int)((double)dxa*(double)dy/(double)dya);
   else if(dxa < dya)xw=xws+(int)((double)dya*(double)dx/(double)dxa);
 }
 if(View == TRIVIEW){ifd=0;ild=3;}
 else               {ifd=ActiveView; ild=ifd+1;}
 for(i=ifd;i<ild;i++){
   hwnd=ghwnd_triview[i];
   hdc=GetDC(hwnd);
   SelectPalette(hdc,ghpaletteScreen,FALSE);
   RealizePalette (hdc);
   DrawOne3dCursor(hdc,i);
   ReleaseDC(hwnd,hdc);
 }
 GetWorldCoords(ActiveView,&x,&y,&z,xw,yw);
 if(grid_on){
//   g2=grid_size/2;
//   x = lrulerx+((x+(x >= 0  ? g2 : -g2)-lrulerx)/grid_size)*grid_size;
//   y = lrulery+((y+(y >= 0  ? g2 : -g2)-lrulery)/grid_size)*grid_size;
//   z = lrulerz+((z+(z >= 0  ? g2 : -g2)-lrulerz)/grid_size)*grid_size;
   g2=local_grid_size/2;
   x = lrulerx+((x+(x >= 0  ? g2 : -g2)-lrulerx)/local_grid_size)*local_grid_size;
   y = lrulery+((y+(y >= 0  ? g2 : -g2)-lrulery)/local_grid_size)*local_grid_size;
   z = lrulerz+((z+(z >= 0  ? g2 : -g2)-lrulerz)/local_grid_size)*local_grid_size;
 }
 NpointerX=x; NpointerY=y; NpointerZ=z;
 for(i=ifd;i<ild;i++){
   hwnd=ghwnd_triview[i];
   hdc=GetDC(hwnd);
   SelectPalette(hdc,ghpaletteScreen,FALSE);
   RealizePalette (hdc);
   DrawOne3dCursor(hdc,i);
   ReleaseDC(hwnd,hdc);
 }
 if(coords_visible)UpdateRuler(0);
}

#define ARROW_ICON_SIZE 12
#define WINDOW_ID_XSIZE 48
#define WINDOW_ID_YSIZE 20

static HBITMAP hbmArrow[6]={NULL,NULL,NULL,NULL,NULL,NULL},
                hbmName[5]={NULL,NULL,NULL,NULL,NULL};
static int ArrowOK,ArrowID[6]={IDBM_ARROW_Zp,IDBM_ARROW_Zm,
                               IDBM_ARROW_Xp,IDBM_ARROW_Xm,
                               IDBM_ARROW_Yp,IDBM_ARROW_Ym};

void LoadArrowIcons(void){
 int i;
 ArrowOK=1;
 for(i=0;i<6;i++){
   if((hbmArrow[i]=LoadBitmap(ghinst_main,
      MAKEINTRESOURCE(ArrowID[i]))) == NULL)ArrowOK=0;
 }
 if(!Preferences.buttons){
   if((hbmName[0]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_Th))) == NULL)ArrowOK=0;
   if((hbmName[1]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_Rh))) == NULL)ArrowOK=0;
   if((hbmName[2]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_Sh))) == NULL)ArrowOK=0;
   if((hbmName[4]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_Fh))) == NULL)ArrowOK=0;
 }
 else{
   if((hbmName[0]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_T))) == NULL)ArrowOK=0;
   if((hbmName[1]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_R))) == NULL)ArrowOK=0;
   if((hbmName[2]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_S))) == NULL)ArrowOK=0;
   if((hbmName[4]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_WIN_F))) == NULL)ArrowOK=0;
 }
}

void DrawArrowIcons(HDC hDC, int id){
 int i,ic,x,y;
 HDC hMemDC;
 HBITMAP hbmOld;
 HBRUSH holdBrush;
 static int arrow_table[2][3][4]={
  {{3,2,4,5} ,     /* top   */
   {3,2,0,1} ,     /* rear  */
   {5,4,0,1} } ,   /* right */
  {{2,3,5,4} ,
   {2,3,0,1} ,
   {4,5,0,1} }
 };
 if(!ArrowOK)return;
 hMemDC = CreateCompatibleDC(hDC);
 for(i=0;i<4;i++){
   ic=arrow_table[WindowBox_view][id][i];
   if(i == 0)hbmOld = SelectObject(hMemDC,hbmArrow[ic]);
   else      SelectObject(hMemDC,hbmArrow[ic]);
   if     (i == 0){x=0; y=WindowSizeY[id]/2-ARROW_ICON_SIZE/2;}
   else if(i == 1){x=WindowSizeX[id]-ARROW_ICON_SIZE-1;
                   y=WindowSizeY[id]/2-ARROW_ICON_SIZE/2;}
   else if(i == 2){x=WindowSizeX[id]/2-ARROW_ICON_SIZE/2; y=0;}
   else if(i == 3){x=WindowSizeX[id]/2-ARROW_ICON_SIZE/2;
                   y=WindowSizeY[id]-ARROW_ICON_SIZE-1;}
   BitBlt(hDC,x,y,ARROW_ICON_SIZE,ARROW_ICON_SIZE,hMemDC,0,0,SRCPAINT);
 }
 SelectObject(hMemDC,hbmName[id]);
 x=y=0;
 if(!Preferences.buttons)
      BitBlt(hDC,x,y,WINDOW_ID_XSIZE,WINDOW_ID_YSIZE,hMemDC,0,0,SRCPAINT);
 else BitBlt(hDC,x,y,WINDOW_ID_XSIZE,WINDOW_ID_YSIZE,hMemDC,0,0,SRCCOPY);
 if(View == ONEVIEW){
   x=WindowSizeX[id]-2*WINDOW_ID_XSIZE-1;
   y=0;
   holdBrush=SelectObject(hDC,GetStockObject(LTGRAY_BRUSH));
   for(i=0;i<3;i++)if(i != id){
     SelectObject(hMemDC,hbmName[i]);
     if(!Preferences.buttons){
       BitBlt(hDC,x,y,WINDOW_ID_XSIZE,WINDOW_ID_YSIZE,hMemDC,0,0,0x00220326); /*  DSna */
       BitBlt(hDC,x,y,WINDOW_ID_XSIZE,WINDOW_ID_YSIZE,hMemDC,0,0,0x00ea02e9); /* DPSao */
     }
     else BitBlt(hDC,x,y,WINDOW_ID_XSIZE,WINDOW_ID_YSIZE,hMemDC,0,0,SRCCOPY);
     x+=WINDOW_ID_XSIZE;
   }
   SelectObject(hDC,holdBrush);
 }
 SelectObject(hMemDC, hbmOld);
 DeleteDC(hMemDC);
}

int CheckWindowIcon(int xi, int yi){
#if 0
 int i,x,y;
 if(xi > WindowSizeX[ActiveView]-WINDOW_ID_XSIZE &&
    xi < WindowSizeX[ActiveView] &&
    yi > WindowSizeY[ActiveView]-WINDOW_ID_YSIZE &&
    yi < WindowSizeY[ActiveView])return 4;
 if(View == ONEVIEW){
   if(xi > 1 && xi < WINDOW_ID_XSIZE &&
      yi > WindowSizeY[ActiveView]-WINDOW_ID_YSIZE &&
      yi < WindowSizeY[ActiveView])return 5;
   if(xi > 1+WINDOW_ID_XSIZE && xi < 2*WINDOW_ID_XSIZE &&
      yi > WindowSizeY[ActiveView]-WINDOW_ID_YSIZE &&
      yi < WindowSizeY[ActiveView])return 6;
 }
#endif
 if(xi >=0 &&
    xi < WINDOW_ID_XSIZE &&
    yi >=0 &&
    yi < WINDOW_ID_YSIZE)return 4;
 if(View == ONEVIEW){
   if(xi > WindowSizeX[ActiveView]-2*WINDOW_ID_XSIZE &&
      xi < WindowSizeX[ActiveView]-WINDOW_ID_XSIZE &&
      yi > 0 &&
      yi < WINDOW_ID_YSIZE)return 5;
   if(xi > WindowSizeX[ActiveView]-WINDOW_ID_XSIZE &&
      xi < WindowSizeX[ActiveView] &&
      yi > 0 &&
      yi < WINDOW_ID_YSIZE)return 6;
 }
 return -1;
}

void InvertWindowIcon(int i, HWND hwnd){
 HDC hdc;
 RECT rc;
 if     (i == 0){rc.left=0;
                 rc.top=WindowSizeY[ActiveView]/2-ARROW_ICON_SIZE/2;}
 else if(i == 1){rc.left=WindowSizeX[ActiveView]-ARROW_ICON_SIZE-1;
                 rc.top=WindowSizeY[ActiveView]/2-ARROW_ICON_SIZE/2;}
 else if(i == 2){rc.left=WindowSizeX[ActiveView]/2-ARROW_ICON_SIZE/2;
                 rc.top=0;}
 else if(i == 3){rc.left=WindowSizeX[ActiveView]/2-ARROW_ICON_SIZE/2;
                 rc.top=WindowSizeY[ActiveView]-ARROW_ICON_SIZE-1;}
 if(i > 3){
#if 0
   rc.top = WindowSizeY[ActiveView]-WINDOW_ID_YSIZE;
   if     (i == 4)rc.left=WindowSizeX[ActiveView]-WINDOW_ID_XSIZE;
   else if(i == 5)rc.left=1;
   else if(i == 6)rc.left=1+WINDOW_ID_XSIZE;
#endif
   rc.top = 0;
   if     (i == 4)rc.left=0;
   else if(i == 5)rc.left=WindowSizeX[ActiveView]-2*WINDOW_ID_XSIZE;
   else if(i == 6)rc.left=WindowSizeX[ActiveView]-WINDOW_ID_XSIZE;
   rc.right=rc.left+WINDOW_ID_XSIZE;
   rc.bottom=rc.top+WINDOW_ID_YSIZE;
 }
 else{  /* other corner for arrows */
   rc.right=rc.left+ARROW_ICON_SIZE;
   rc.bottom=rc.top+ARROW_ICON_SIZE;
 }
 hdc=GetDC(hwnd);
 InvertRect(hdc,&rc);
 ReleaseDC(hwnd,hdc);
 return;
}

void DeleteArrowIcons(void){
 int i;
 for(i=0;i<6;i++){
   if(hbmArrow != NULL)DeleteObject(hbmArrow[i]);
   hbmArrow[i]=NULL;
 }
 if(hbmName[0] != NULL)DeleteObject(hbmName[0]); hbmName[0]=NULL;
 if(hbmName[1] != NULL)DeleteObject(hbmName[1]); hbmName[1]=NULL;
 if(hbmName[2] != NULL)DeleteObject(hbmName[2]); hbmName[2]=NULL;
 if(hbmName[4] != NULL)DeleteObject(hbmName[4]); hbmName[4]=NULL;
}

void SwapViewBitmaps(void){
 HBITMAP hbmTemp;
 hbmTemp=hbmName[1];
 hbmName[1]=hbmName[4];
 hbmName[4]=hbmTemp;
}

static BOOL CALLBACK CoordinateDlgProc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam){
 char tempstr[16];
 struct np {char *name; double cal;} *dp;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       RECT rcDlg;
       POINT p;
       dp=(struct np *)lparam;
       SendDlgItemMessage(hwnd,DLG_COORD_NAME,WM_SETTEXT,0,(LPARAM)dp->name);
       SendDlgItemMessage(hwnd,DLG_COORD_NAME,EM_LIMITTEXT,(WPARAM)6,0);
       sprintf(tempstr,"%.4lf",dp->cal);
       SendDlgItemMessage(hwnd,DLG_COORD_CAL,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_COORD_CAL,EM_LIMITTEXT,(WPARAM)12,0);
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       GetCursorPos(&p);
       GetWindowRect(hwnd,&rcDlg);
       OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
       OffsetRect(&rcDlg,p.x-rcDlg.right/2,p.y-rcDlg.bottom/2);
       if(rcDlg.left < 0)OffsetRect(&rcDlg,-rcDlg.left,0);
       if(rcDlg.top  < 0)OffsetRect(&rcDlg,0,-rcDlg.top);
       if(rcDlg.right  > Xres)OffsetRect(&rcDlg,Xres-rcDlg.right,0);
       if(rcDlg.bottom > Yres)OffsetRect(&rcDlg,Yres-rcDlg.bottom,0);
       SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
     }
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case DLG_COORD_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case IDOK:
        case DLG_COORD_OK:
          dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
          if(GetDlgItemText(hwnd,DLG_COORD_NAME,dp->name,7) == 0)EndDialog(hwnd, FAIL);
          if(GetDlgItemText(hwnd,DLG_COORD_CAL,tempstr,10) == 0)EndDialog(hwnd, FAIL);
          if((dp->cal=atof(tempstr)) == 0)EndDialog(hwnd, FAIL);
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

void UpdateRuler(short m){
 double dx,dy,dz;
 char tempstr[32];
 if(m == 1){ /* cancel */
   rulerx=NpointerX; rulery=NpointerY; rulerz=NpointerZ;
 }
 if(m == 2){ /* set origin*/
   if(SendPrgmQuery(IDQ_SET_COORD_ORIGIN,1) != IDYES)return;
   EDIT_ACTION=YES;
   lrulerx=NpointerX; lrulery=NpointerY; lrulerz=NpointerZ;
   DrawModel();
 }
 if(m == 3){ /* set ruler */
   char name[10];
   struct np {char *name; double cal;} d;
   dx = (double)(NpointerX-lrulerx);
   dy = (double)(NpointerY-lrulery);
   dz = (double)(NpointerZ-lrulerz);
   if(dx < (double)MINUNIT || dy < (double)MINUNIT || dz < (double)MINUNIT){
     double dd;
     EDIT_ACTION=YES;
     if(ruler_name[0] == 0)strcpy(name,"<cms>");
     else                  strcpy(name,ruler_name);
     dd=sqrt(dx*dx+dy*dy+dz*dz);
     d.name=name; d.cal=dd/ruler;
     EnableToolPannels(ALL_PANNELS,FALSE);
     if(DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_COORD),ghwnd_main,
                (DLGPROC)CoordinateDlgProc,(LPARAM)&d) == OK){
       strcpy(ruler_name,d.name);
       LoadString(ghinst_main,IDX_MISC_COORDS,tempstr,32);
       strcat(tempstr,ruler_name);
       SendMessage(ghwndCoord1,WM_SETTEXT,0,(LPARAM)tempstr);
       ruler = dd;
       if(d.cal > 0.0)ruler /= d.cal;
   }
   EnableToolPannels(ALL_PANNELS,TRUE);
   }
   else SendPrgmQuery(IDQ_COORD_BAD_RANGE,0);
 }
 SendMessage(ghwndCoord1,WM_COMMAND,IDC_COORD_UPDATE,0);
}

static BOOL CALLBACK GridSizeDlgProc(HWND hwnd, UINT msg,
                                     WPARAM wparam, LPARAM lparam){
 double d;
 char size[16];
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hwnd,ghinst_main);
     break;
   case WM_INITDIALOG:{
       RECT rcDlg;
       POINT p;
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       GetCursorPos(&p);
       GetWindowRect(hwnd,&rcDlg);
       OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
       OffsetRect(&rcDlg,p.x-rcDlg.right/2,p.y-rcDlg.bottom/2);
       if(rcDlg.left < 0)OffsetRect(&rcDlg,-rcDlg.left,0);
       if(rcDlg.top  < 0)OffsetRect(&rcDlg,0,-rcDlg.top);
       if(rcDlg.right  > Xres)OffsetRect(&rcDlg,Xres-rcDlg.right,0);
       if(rcDlg.bottom > Yres)OffsetRect(&rcDlg,Yres-rcDlg.bottom,0);
       SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
       if(draw_grid_on)SendDlgItemMessage(hwnd,DLG_GRIDSIZE_ON ,BM_SETCHECK,1,0);
       else            SendDlgItemMessage(hwnd,DLG_GRIDSIZE_OFF,BM_SETCHECK,1,0);
       sprintf(size,"%.4lf",((double)grid_size)/ruler);
       SendDlgItemMessage(hwnd,DLG_GRIDSIZE_EDIT,WM_SETTEXT,0,
                          (LPARAM)(LPSTR)size);
       SendDlgItemMessage(hwnd,DLG_GRIDSIZE_EDIT,EM_LIMITTEXT,12,0);
     }
     return (TRUE);  /* must be used unless SetFocus */
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
        case DLG_GRIDSIZE_CANCEL:
          EndDialog(hwnd, FAIL);
          return(TRUE);
        case IDOK:
        case DLG_GRIDSIZE_OK:
          if(SendDlgItemMessage(hwnd,DLG_GRIDSIZE_ON,BM_GETCHECK,0,0))
               draw_grid_on=1;
          else draw_grid_on=0;
          if(GetDlgItemText(hwnd,DLG_GRIDSIZE_EDIT,size,12)){
            if((d=atof(size)) > 0.0)grid_size=(long)(d*ruler);
          }
          EndDialog(hwnd,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

void SetUpGrid(void){
 EnableToolPannels(ALL_PANNELS,FALSE);
 DialogBox(ghinst_main,MAKEINTRESOURCE(DLG_GRIDSIZE),ghwnd_main,
            (DLGPROC)GridSizeDlgProc);
 EnableToolPannels(ALL_PANNELS,TRUE);
 return;
}

void DrawGridInOne(HDC hdc, int i){
 COLORREF OldColour;
 HPEN hPen1,hPen2,hPen3,hOldPen,hTempPen;
 BOOL flagx[50],flagy[50];
 int ng,ngx,ngy,ngz,ng1,ng2,j,k,xx[50],yy[50],d;
 long x,y,z,x1,y1,z1,g2,nx,ny,nz,iter;
 local_grid_size=grid_size/10;
 iter=0;
 TRYAGAIN:
 if(local_grid_size == 0)return;
 g2=local_grid_size/2;
 x=TVpointX; y=TVpointY; z=TVpointZ;
 x1 = lrulerx+((x+(x >= 0  ? g2 : -g2)-lrulerx)/local_grid_size)*local_grid_size;
 y1 = lrulery+((y+(y >= 0  ? g2 : -g2)-lrulery)/local_grid_size)*local_grid_size;
 z1 = lrulerz+((z+(z >= 0  ? g2 : -g2)-lrulerz)/local_grid_size)*local_grid_size;
 if(x1 <= TVpointX)x1 += local_grid_size;
 if(y1 <= TVpointY)y1 += local_grid_size;
 if(z1 <= TVpointZ)z1 += local_grid_size;
 ngx=(TVsizeX/local_grid_size);
 ngy=(TVsizeY/local_grid_size);
 ngz=(TVsizeZ/local_grid_size);
 ng = max(ngx,ngy);   ng=max(ngz,ng); ng++;
 if(ng == 0)return;
 if(ng > 30){
   local_grid_size *= 10;
   if(iter++ > 100)return;
   goto TRYAGAIN;
 }
 ng1 = ng2 = -1; x=x1; y=y1; z=z1;
 for(j=0;j<ng;j++){
   nx = labs(x-lrulerx)/local_grid_size;
   ny = labs(y-lrulery)/local_grid_size;
   nz = labs(z-lrulerz)/local_grid_size;
   flagx[j]=flagy[j]=FALSE;
   if(i == TRIRIGHT){
     nx=ny;
     ny=nz;
   }
   else if(i == TRIFRONT){
     ny=nz;
   }
   if(nx%10 == 0)flagx[j]=TRUE;
   if(ny%10 == 0)flagy[j]=TRUE;
   GetWindowCoords(i,x,y,z,&xx[j],&yy[j]);
   x += local_grid_size; y += local_grid_size; z += local_grid_size;
   if(WindowBox_view == 0){
     if(xx[j] >= WindowSizeX[i] && ng1 < 0)ng1=j;
     if(yy[j] <= 0              && ng2 < 0)ng2=j;
   }
   else{
     if(xx[j] <= 0 && ng1 < 0)ng1=j;
     if(i==TRITOP){if(yy[j] >= WindowSizeY[i] && ng2 < 0)ng2=j;}
     else         {if(yy[j] <= 0              && ng2 < 0)ng2=j;}
   }
 }
 if(ng1 < 0)ng1=ng; if(ng2 < 0)ng2=ng;
// if(iter == 0)hPen1=CreatePen(PS_DOT,0,RGB(118,118,118));
// else         hPen1=CreatePen(PS_SOLID,0,RGB(118,118,118));
// hPen2=CreatePen(PS_SOLID,0,RGB(151,151,151));
// hPen3=CreatePen(PS_SOLID,0,RGB(  0,  0,255));
 if(iter == 0)hPen1=CreatePen(PS_DOT,0,gGridColourRef1);
 else         hPen1=CreatePen(PS_SOLID,0,gGridColourRef1);
 hPen2=CreatePen(PS_SOLID,0,gGridColourRef2);
 hPen3=CreatePen(PS_SOLID,0,RGB(255,255,255));
 OldColour=SetBkColor(hdc,gScreenColourRef);
 hOldPen=SelectObject(hdc,hPen1);
 x= -1; y = -1;
 if(lrulerx > TVpointX || lrulerx < TVpointX+TVsizeX ||
    lrulery > TVpointY || lrulery < TVpointY+TVsizeY ||
    lrulerz > TVpointZ || lrulerz < TVpointZ+TVsizeZ){
   GetWindowCoords(i,lrulerx,lrulery,lrulerz,(int *)(&x),(int *)&y);
 }
 if(ng1 > 0 && ng2 > 0){
   for(j=0;j<ng2;j++){
     if(yy[j] == y)continue;
     if(flagy[j])hTempPen=SelectObject(hdc,hPen2);
     MoveToEx(hdc,0,yy[j],NULL); LineTo(hdc,WindowSizeX[i]+1,yy[j]);
     if(flagy[j])SelectObject(hdc,hTempPen);
   }
   for(k=0;k<ng1;k++){
     if(xx[k] == x)continue;
     if(flagx[k])hTempPen=SelectObject(hdc,hPen2);
     MoveToEx(hdc,xx[k],0,NULL); LineTo(hdc,xx[k],WindowSizeY[i]+1);
     if(flagx[k])SelectObject(hdc,hTempPen);
   }
 }
 SetBkColor(hdc,OldColour);
 SelectObject(hdc,hPen3);
 if(y >= 0 && y <= WindowSizeY[i]){
   MoveToEx(hdc,0,y,NULL); LineTo(hdc,WindowSizeX[i]+1,y);
 }
 if(x >= 0 && x <= WindowSizeX[i]){
   MoveToEx(hdc,x,0,NULL); LineTo(hdc,x,WindowSizeY[i]+1);
 }
 SelectObject(hdc,hOldPen);
 DeleteObject(hPen1);
 DeleteObject(hPen2);
 DeleteObject(hPen3);
 return;
}

static char *ColorSelectDlgString;

static BOOL CALLBACK ColorSelectDlgProc(HWND hwnd, UINT msg, WPARAM wparam,
                                        LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)ColorSelectDlgString);
     return TRUE;
   default:
     break;
 }
 return FALSE;
}

BOOL SetColour(unsigned char colour[], char *Title, HWND parent){
 BOOL result=FALSE;
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB(colour[0],colour[1],colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags=CC_FULLOPEN | CC_RGBINIT | CC_ENABLEHOOK;
 cc.lCustData=(DWORD)0;
 ColorSelectDlgString=Title;
 cc.lpfnHook=(LPCCHOOKPROC)ColorSelectDlgProc;
 if((result=ChooseColor(&cc))){
   colour[0]=GetRValue(cc.rgbResult);
   colour[1]=GetGValue(cc.rgbResult);
   colour[2]=GetBValue(cc.rgbResult);
 }
 return result;
}

BOOL SetSfxColour(unsigned char colour[], long ID, HWND parent){
 char szTitle[256];
 LoadString(ghinst_main,ID,szTitle,255);
 return SetColour(colour,szTitle,parent);
}

void scal(double t[4][4], double sx, double sy, double sz){
 long i,j;
 for(i=0;i<4;i++) for(j=0;j<4;j++) t[i][j]=0.0;
 t[0][0]=sx; t[1][1]=sy; t[2][2]=sz; t[3][3]=1;
 return;
}

void rotz(double tr[4][4], double ang){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0]= cos(ang);
  tr[0][1]=-sin(ang);
  tr[1][0]= sin(ang);
  tr[1][1]= cos(ang);
  return;
}

void roty(double tr[4][4], double ang){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0]= cos(ang);
  tr[0][2]=-sin(ang);
  tr[2][0]= sin(ang);
  tr[2][2]= cos(ang);
  return;
}

void rotx(double tr[4][4], double ang){
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[1][1]= cos(ang);
  tr[1][2]=-sin(ang);
  tr[2][1]= sin(ang);
  tr[2][2]= cos(ang);
  return;
}

void tram(double t[4][4], double dx, double dy, double dz){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  t[i][j]=0.0;
  if(i == j)t[i][j]=1.0;
 }
 t[0][3]=dx;
 t[1][3]=dy;
 t[2][3]=dz;
 t[3][3]=1;
 return;
}

void mv4by1(double t4[4][4], vector v1, vector v2){
 double xx,yy,zz;
 xx=t4[0][0]*v1[0]+t4[0][1]*v1[1]+t4[0][2]*v1[2]+t4[0][3];
 yy=t4[1][0]*v1[0]+t4[1][1]*v1[1]+t4[1][2]*v1[2]+t4[1][3];
 zz=t4[2][0]*v1[0]+t4[2][1]*v1[1]+t4[2][2]*v1[2]+t4[2][3];
 v2[0]=xx; v2[1]=yy; v2[2]=zz;
 return;
}

void m4by4(double t1[4][4], double t2[4][4], double tr[4][4]){
 short n=4,i,j,k;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  for(k=0;k<4;k++)tr[i][j]=tr[i][j]+t1[i][k]*t2[k][j];
 }
 return;
}

void m4by1(double t4[4][4], double x, double y, double z,
           double *xx, double *yy, double *zz){
 *xx=t4[0][0]*x+t4[0][1]*y+t4[0][2]*z+t4[0][3];
 *yy=t4[1][0]*x+t4[1][1]*y+t4[1][2]*z+t4[1][3];
 *zz=t4[2][0]*x+t4[2][1]*y+t4[2][2]*z+t4[2][3];
 return;
}

void c4to4(double tin[4][4], double tout[4][4]){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 tout[i][j]=tin[i][j];
}

void null_transform(double t[4][4]){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 if(i == j)t[i][j]=1.0;
 else      t[i][j]=0.0;
}

void arbitrary_rotate(double angle, point p1, point p2, double t[4][4]){
  double dx,dy,dz,phi,theta,dxy;
  double t1[4][4],t2[4][4],t3[4][4];
  angle *= PI/180.0;
  dx=p2[0]-p1[0]; dy=p2[1]-p1[1]; dz=p2[2]-p1[2];
  dxy=dx*dx+dy*dy;
  tram(t1, -p1[0], -p1[1], -p1[2]);
  if(dxy < 1.0){ /* vertical so just rotate about z  and return */
    if(dz > 0.0)rotz(t2, angle);
    else        rotz(t2,-angle);
    m4by4(t2,t1,t3);
    tram(t1,p1[0],p1[1],p1[2]);
    m4by4(t1,t3,t);
    return;
  }
  dxy=sqrt(dxy);
  if     (dx == 0.0 && dy > 0.0)phi =  PI2;
  else if(dx == 0.0 && dy < 0.0)phi = -PI2;
  else phi = atan2(dy,dx);
  theta = atan2(dz,dxy);
  rotz(t2, -phi);
  m4by4(t2,t1,t3);
  roty(t2, -theta);
  m4by4(t2,t3,t1);
  rotx(t2, angle);
  m4by4(t2,t1,t3);
  roty(t2,theta);
  m4by4(t2,t3,t1);
  rotz(t2,phi);
  m4by4(t2,t1,t3);
  tram(t1,p1[0],p1[1],p1[2]);
  m4by4(t1,t3,t);
}

void rotate_round_vector(double angle, vector v, double t[4][4]){
  double dx,dy,dz,phi,theta,dxy;
  double t1[4][4],t2[4][4],t3[4][4];
  angle *= PI/180.0;
  dx=v[0]; dy=v[1]; dz=v[2];
  dxy=dx*dx+dy*dy;
  if(dxy < 0.00001){ /* vertical so just rotate about z  and return */
    if(dz > 0.0)rotz(t, angle);
    else        rotz(t,-angle);
    return;
  }
  dxy=sqrt(dxy);
  if     (dx == 0.0 && dy > 0.0)phi =  PI2;
  else if(dx == 0.0 && dy < 0.0)phi = -PI2;
  else phi = atan2(dy,dx);
  theta = atan2(dz,dxy);
  rotz(t3, -phi);
  roty(t2, -theta);
  m4by4(t2,t3,t1);
  rotx(t2, angle);
  m4by4(t2,t1,t3);
  roty(t2,theta);
  m4by4(t2,t3,t1);
  rotz(t2,phi);
  m4by4(t2,t1,t);
}

void L_cross(point p1, point p2, point p3, point p4){
 double v1[3],v2[3],r[3];
 v1[0]=(double)(p2[0]-p1[0]);
 v1[1]=(double)(p2[1]-p1[1]);
 v1[2]=(double)(p2[2]-p1[2]);
 v2[0]=(double)(p3[0]-p1[0]);
 v2[1]=(double)(p3[1]-p1[1]);
 v2[2]=(double)(p3[2]-p1[2]);
 r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);
 r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);
 r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);
 p4[0] = p1[0] + (long)(r[0]/UNIT);
 p4[1] = p1[1] + (long)(r[1]/UNIT);
 p4[2] = p1[2] + (long)(r[2]/UNIT);
}

void CentreDialogOnCursor(HWND hwnd){
 RECT rcDlg;
 POINT p;
 GetCursorPos(&p);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,p.x-rcDlg.right/2,p.y-rcDlg.bottom/2);
 if(rcDlg.left < 0)OffsetRect(&rcDlg,-rcDlg.left,0);
 if(rcDlg.top  < 0)OffsetRect(&rcDlg,0,-rcDlg.top);
 if(rcDlg.right  > Xres)OffsetRect(&rcDlg,Xres-rcDlg.right,0);
 if(rcDlg.bottom > Yres)OffsetRect(&rcDlg,0,Yres-rcDlg.bottom);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

void CentreDialogOnScreen(HWND hWnd){
 RECT rc;
 GetWindowRect(hWnd, &rc);
 SetWindowPos(hWnd, NULL,
    (GetSystemMetrics(SM_CXSCREEN) - (rc.right - rc.left)) / 2,
    (GetSystemMetrics(SM_CYSCREEN) - (rc.bottom - rc.top)) / 2,
    0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

void PaintDialogBackground(HWND hDlg, HINSTANCE hInst){
 int x;
 int y;
 HDC hDC;
 HPALETTE hpalT;
 HBRUSH hOldBrush;
 RECT rc1,rc2;
 HDC hMemDC;
 HBITMAP hBmBackground,hbmOld;
 BITMAP  bm;
 PAINTSTRUCT ps;
 hDC = BeginPaint(hDlg,&ps);
 hBmBackground=LoadBitmap(hInst,MAKEINTRESOURCE(IDBM_BACKGROUND));
 if(hBmBackground != NULL){
   GetObject(hBmBackground,sizeof(BITMAP),&bm);
   hMemDC = CreateCompatibleDC(hDC);
   hbmOld = SelectObject(hMemDC,hBmBackground);
   GetClientRect(hDlg,&rc2);
   for(x=0;x<(rc2.right/bm.bmWidth)+1;x++)
   for(y=0;y<(rc2.bottom/bm.bmHeight)+1;y++)
     BitBlt(hDC,
          x*bm.bmWidth,y*bm.bmHeight,
          bm.bmWidth,
          bm.bmHeight,
          hMemDC,
          0,0,
          SRCCOPY);
   SelectObject(hMemDC,hbmOld);
   DeleteDC(hMemDC);
   DeleteObject(hBmBackground);
 }
 EndPaint(hDlg,&ps);
}

void AppendFileExtension(char *filename, char *ext){
 if(strrchr(filename,'.') == NULL){
   strcat(filename,ext);
 }
}

static short PanTo(long *, long *, long);

static short PanTo(long *start, long *centre, long increment){
 long newpoint;
 newpoint = *start + increment;
 if(newpoint > MAXUNIT || newpoint < -MAXUNIT){
   SendPrgmQuery(IDQ_PANTOOFAR,0);
   return 0;
 }
 *start = newpoint;
 *centre = *centre + increment;
 return 1;
}

void Zoom(int command, int redraw){
 long what_shift,cx,cy,cz;
 double zf;
 short f;
 if(command == ZOOMIN || command == ZOOMOUT){
   zf=1.2;
   if(GetAsyncKeyState(VK_SHIFT  ) & 0x8000)zf=1.5;
   if(GetAsyncKeyState(VK_CONTROL) & 0x8000)zf=2.0;
   if(GetAsyncKeyState(VK_MENU   ) & 0x8000)zf=4.0;
 }
 else{
   what_shift=TVsizeX/16;
   if(GetAsyncKeyState(VK_SHIFT  ) & 0x8000)what_shift=TVsizeX/2;
   if(GetAsyncKeyState(VK_CONTROL) & 0x8000)what_shift=TVsizeX;
   if(GetAsyncKeyState(VK_MENU   ) & 0x8000)what_shift=TVsizeX*1.5;
 }
 if(command == ZOOMIN){
  if(TVsizeX <= MINUNIT || TVsizeY <= MINUNIT || TVsizeZ <= MINUNIT){
    SendPrgmQuery(IDQ_NOZOOMIN,0);
    goto EXIT;
  }
  cx=TVpointX+TVsizeX/2; cy=TVpointY+TVsizeY/2; cz=TVpointZ+TVsizeZ/2;
  TVsizeX=(long)((double)TVsizeX/zf);
  TVsizeY=(long)((double)TVsizeY/zf);
  TVsizeZ=(long)((double)TVsizeZ/zf);
  WindowPixelSize=(double)TVsizeZ/(double)WindowSizeY[1];
  TVpointX=cx-TVsizeX/2;
  TVpointY=cy-TVsizeY/2;
  TVpointZ=cz-TVsizeZ/2;
 }
 else if(command == ZOOMOUT){
  if(TVsizeX >= MAXUNIT || TVsizeY >= MAXUNIT || TVsizeZ >= MAXUNIT){
    SendPrgmQuery(IDQ_NOZOOMOUT,0);
    goto EXIT;
  }
  cx=TVpointX+TVsizeX/2; cy=TVpointY+TVsizeY/2; cz=TVpointZ+TVsizeZ/2;
  TVsizeX=(long)((double)TVsizeX*zf);
  TVsizeY=(long)((double)TVsizeY*zf);
  TVsizeZ=(long)((double)TVsizeZ*zf);
  WindowPixelSize=(double)TVsizeZ/(double)WindowSizeY[1];
  TVpointX=cx-TVsizeX/2;
  TVpointY=cy-TVsizeY/2;
  TVpointZ=cz-TVsizeZ/2;
 }
 else if(command == RECENTRE){
  if(TVsizeX >= MAXUNIT || TVsizeY >= MAXUNIT || TVsizeZ >= MAXUNIT){
    goto EXIT;
  }
  TVpointX=NpointerX-TVsizeX/2; TVcentreX=NpointerX;
  TVpointY=NpointerY-TVsizeY/2; TVcentreY=NpointerY;
  TVpointZ=NpointerZ-TVsizeZ/2; TVcentreZ=NpointerZ;
 }
 else if(command == CENTREWORLD){
   reset_mod_maxview(1);
 }
 else { /* pan commands */
   if(WindowBox_view == 1){
    if(command == PANFORWARD ||
       command == PANBACK    ||
       command == PANRIGHT   ||
       command == PANLEFT)what_shift *= -1.0;
   }
   if(command == PANFORWARD){if(!PanTo(&TVpointY,&TVcentreY,+what_shift))goto EXIT;} /* forward   */
   if(command == PANBACK   ){if(!PanTo(&TVpointY,&TVcentreY,-what_shift))goto EXIT;} /* backwards */
   if(command == PANRIGHT  ){if(!PanTo(&TVpointX,&TVcentreX,+what_shift))goto EXIT;} /* right     */
   if(command == PANLEFT   ){if(!PanTo(&TVpointX,&TVcentreX,-what_shift))goto EXIT;} /* left      */
   if(command == PANUP     ){if(!PanTo(&TVpointZ,&TVcentreZ,+what_shift))goto EXIT;} /* up        */
   if(command == PANDOWN   ){if(!PanTo(&TVpointZ,&TVcentreZ,-what_shift))goto EXIT;} /* down      */
 }
 if(NpointerX < TVpointX)        NpointerX=TVpointX;
 if(NpointerX > TVpointX+TVsizeX)NpointerX=TVpointX+TVsizeX;
 if(NpointerY < TVpointY)        NpointerY=TVpointY;
 if(NpointerY > TVpointY+TVsizeY)NpointerY=TVpointY+TVsizeY;
 if(NpointerZ < TVpointZ)        NpointerZ=TVpointZ;
 if(NpointerZ > TVpointZ+TVsizeZ)NpointerZ=TVpointZ+TVsizeZ;
 if(redraw)DrawModel();
 EXIT:;
}
