/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file loader.c  - loads/saves the models from their files on disk  */

#define MODULE_LOADER 1

#include "design.h"

extern long LastToolVertex;

static void LoadRenderer(int anim);
static unsigned char getlon(long *);
static unsigned char getsht(short *);
static unsigned char getusht(unsigned short *);
static void outsht(short);
static void outusht(unsigned short);
static void outlng(long);
static void getchunk(long clen);
static void WriteTextures(void);
static void WriteVertices(short status);
static void WriteVertexMapping(short status);
static void WriteEdges(BOOL );
static void WriteFaces(BOOL );
static void WriteSkeleton(void);

#define BRUSHBIT   0x40
#define SMOOTHBIT  0x80
#define TILEBIT    0x20;

static DWORD dwRead;
static HANDLE fq;
static FILE *fo;
static long newface,newedge,newvert,newglue,minX,minY,minZ,maxX,maxY,maxZ;
static char str[10];
static axis TempAxis;
static long FaceBase,EdgeBase,VertexBase;

static unsigned char getlon(long *val){
 long l,l0,l1,l2,l3;
 ReadFile(fq,&l,sizeof(long),&dwRead,NULL);
 l0 = (l >> 24) & 0x000000ff;
 l1 = (l >> 16) & 0x000000ff;
 l2 = (l >>  8) & 0x000000ff;
 l3 = (l      ) & 0x000000ff;
 *val = (l3 << 24) | (l2 << 16) | (l1 <<  8) | l0;
 return 1;
}

static unsigned char getsht(short * val){
 short s,s0,s1;
 ReadFile(fq,&s,sizeof(short),&dwRead,NULL);
 s0 = (s >>  8) & 0x00ff;
 s1 = (s      ) & 0x00ff;
 *val = (s1 <<  8) | s0;
 return 1;
}

static unsigned char getusht(unsigned short * val){
 unsigned short s,s0,s1;
 ReadFile(fq,&s,sizeof(short),&dwRead,NULL);
 s0 = (s >>  8) & 0x00ff;
 s1 = (s      ) & 0x00ff;
 *val = (s1 <<  8) | s0;
 return 1;
}

static void getchunk(long clen){
 SetFilePointer(fq,clen,NULL,FILE_CURRENT);
}

static unsigned short GetUshortFromBuffer(unsigned char *b){
 return ((unsigned short)(*b) <<  8) | (unsigned short)(*(b+1));
}

static short GetShortFromBuffer(unsigned char *b){
 return ((short)(*b) <<  8) | (short)(*(b+1));
}

static long GetLongFromBuffer(unsigned char *b){
 return ((long)(*b) << 24) | ((long)(*(b+1)) << 16) |
        ((long)(*(b+2)) <<  8) | (long)(*(b+3));
}


static void outsht(short sht){
 fputc((char)(sht >>  8),fo);
 fputc((char)(sht      ),fo);
}

static void outusht(unsigned short sht){
 fputc((char)(sht >>  8),fo);
 fputc((char)(sht      ),fo);
}

static void outlng(long lng){
 fputc((char)(lng >> 24),fo);
 fputc((char)(lng >> 16),fo);
 fputc((char)(lng >>  8),fo);
 fputc((char)(lng      ),fo);
}

char *short_form(char *filename){
 char *sps,*spc;
 sps=spc=filename;
 while((spc=strchr(spc,'\\')) != NULL){spc++; sps=spc;}
 return sps;
}

// Load the MFX file

static void ReadModWireframe(long clen);
static void ReadBezierCurves(long clen);
static void ReadOldFaces(long clen, short shader);
static void ReadFaces(long clen, BOOL short_form);
static void ReadEdges(long clen, BOOL short_form);
static void ReadVertices(long clen);
static void ReadVertexMapping(long clen);
//static void ReadBezierCurves(long clen);
static void ReadSkeleton(long clen, short version);
static void ReadTexture(long clen, short relation, short version);
static void ReadImageMap(long clen, short relation);
static void ReadModWireframe(long clen);
static void ReadShadersAxes(long clen, short relation);
static void ReadShader(long clen);

int LoadObject(char *FileName, short relation){
/* relation = 0 for full load
            = 1 for local load
            = 2 abs load
            = 3 DXF load
*/
 int fail=OK;
 long CHUNKsize,FORMsize,i,uxx,reset_ruler=1;
 vertex *v;
 if(relation > 0)EDIT_ACTION=YES;
 maxX = maxY = maxZ = -MAXUNIT;
 minX = minY = minZ =  MAXUNIT;
 FaceBase=Nface; EdgeBase=Nedge; VertexBase=Nvert;
 SendPrgmMessage(IDS_LOADING,0);
 fq=CreateFile(FileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
    FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(fq == INVALID_HANDLE_VALUE){
   SendPrgmMessage(IDS_FILEOPENFAIL,1);
   return FAIL;
 }
 LOADOPERATION=1;
 TempAxis.Origin.xyz[0]=0;         /* make sure all local variables */
 TempAxis.Origin.xyz[1]=0;         /* are  reset before using for   */
 TempAxis.Origin.xyz[2]=0;         /* the loading process           */
 newface=0; newedge=0; newvert=0; newglue=0;
 ReadFile(fq,str,4,&dwRead,NULL);
 if (dwRead != 4 || strcmp(str,"FORM") != 0){
   fail=FAIL;
   goto end;
 }
 getlon(&FORMsize);
 ReadFile(fq,str,4,&dwRead,NULL);
 if (dwRead != 4 || (strcmp(str,"OFXM") != 0 && strcmp(str,"AAPO") != 0)){
   fail=FAIL;
   goto end;
}
loop:
 if(!ReadFile(fq,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4)goto end;
 getlon(&CHUNKsize);
 if(strcmp(str,"VERT") == 0){ReadVertices(CHUNKsize);}
 else if(strcmp(str,"edge") == 0)ReadEdges(CHUNKsize,TRUE);
 else if(strcmp(str,"EDGE") == 0)ReadEdges(CHUNKsize,FALSE);
 else if(strcmp(str,"SFCE") == 0)ReadFaces(CHUNKsize,TRUE);
 else if(strcmp(str,"LFCE") == 0)ReadFaces(CHUNKsize,FALSE);
 else if(strcmp(str,"FACE") == 0)ReadOldFaces(CHUNKsize,0);
 else if(strcmp(str,"SFAC") == 0)ReadOldFaces(CHUNKsize,1);
 else if(strcmp(str,"VTEX") == 0)ReadVertexMapping(CHUNKsize);
// else if(strcmp(str,"BEZ1") == 0)ReadBezierCurves(CHUNKsize);
 else if(strcmp(str,"AXIS") == 0){
   getlon(&TempAxis.Origin.xyz[0]);
   getlon(&TempAxis.Origin.xyz[1]);
   getlon(&TempAxis.Origin.xyz[2]);
   if(relation == 0){ /* full load */
     getlon(&iSha[0].P[0]);  /* use shader 0 as base shader */
     getlon(&iSha[0].P[1]);
     getlon(&iSha[0].P[2]);
     getlon(&iSha[0].X[0]);
     getlon(&iSha[0].X[1]);
     getlon(&iSha[0].X[2]);
     getlon(&iSha[0].Y[0]);
     getlon(&iSha[0].Y[1]);
     getlon(&iSha[0].Y[2]);
     for(i=0;i<3;i++)ObjectAxis.Origin.xyz[i]=TempAxis.Origin.xyz[i];
   }
   else{ /* dump the unwanted texture info */
     for(i=0;i<9;i++)getlon(&uxx);
   }
 }
 else if(strcmp(str,"OFFS") == 0  && relation == 0){/* only for global load*/
   getlon(&i);   /* no of offsets 1 for now allows possible expansion */
   getlon(&ObjectAxis.Offset.xyz[0]);  /* offset for objects to follow */
   getlon(&ObjectAxis.Offset.xyz[1]);  /* this one */
   getlon(&ObjectAxis.Offset.xyz[2]);
 }
 else if(strcmp(str,"TEXT") == 0)ReadTexture(CHUNKsize,relation,0); /* old */
 else if(strcmp(str,"IMAP") == 0)ReadImageMap(CHUNKsize,relation);
 else if(strcmp(str,"SKEL") == 0 && relation == 0)ReadSkeleton(CHUNKsize,0);
 else if(strcmp(str,"SK01") == 0 && relation == 0)ReadSkeleton(CHUNKsize,1);
 else if(strcmp(str,"SK02") == 0 && relation == 0)ReadSkeleton(CHUNKsize,2);
 else if(strcmp(str,"MATL") == 0 && relation == 0){
   getlon(&i);   /* no of properties   min 3   max 4  for now */
   if(i == 4)getlon(&i);
   if(i != 0)ShadowFlag=1;
   getlon(&i); if(i > 1000)refractive_index=((double)i/10000.0);
   getlon(&i);
   getlon(&i);
 }
 else if(strcmp(str,"MPID") == 0 && relation == 0){ /* if plane mirror then axis ID */
   getlon(&i); MirrorFlag=i;
 }
 else if(strcmp(str,"UNIT") == 0 && relation == 0){ /* not for local */
   getlon(&i);
   ruler = (double)i;
   ReadFile(fq,ruler_name,8,&dwRead,NULL);
   if(dwRead != 8)goto end;
   getlon(&lrulerx);   getlon(&lrulery);   getlon(&lrulerz);
   rulerx = lrulerx; rulery = lrulery; rulerz = lrulerz;
   getlon(&i);   getlon(&i);   getlon(&i); /* don't use : is relative ruler */
   reset_ruler=0;
 }
 else if(strcmp(str,"GRID") == 0 && relation == 0){ /* not for local */
   getlon(&i);
   grid_size = (double)i;
   getlon(&i); /* unused */
 }
 else if(strcmp(str,"WFRM") == 0 && relation == 0)ReadModWireframe(CHUNKsize);
 else if(strcmp(str,"SHDR") == 0 && relation == 0)ReadShader(CHUNKsize);
 else if(strcmp(str,"AXES") == 0)ReadShadersAxes(CHUNKsize,relation);
 else if(strcmp(str,"NURB") == 0)ReadNurbs(CHUNKsize,relation,fq);
// put Nurbs here
 else getchunk(CHUNKsize);
 goto loop;
end:
 CloseHandle(fq);
 if(relation == 1){   /* then load apply local offset  */
   v=(MainVp+VertexBase);  /* object origin is placed at cursor */
   for(i=VertexBase;i<Nvert;i++){/* the existing origin is retained */
     v->xyz[0] += NpointerX-TempAxis.Origin.xyz[0];
     v->xyz[1] += NpointerY-TempAxis.Origin.xyz[1];
     v->xyz[2] += NpointerZ-TempAxis.Origin.xyz[2];
     v++;
   }
 }
 else  EDIT_ACTION=NO;
 if(Nvert > 0 && (relation == 0 || relation == 3)){
//
   reset_mod_maxview(1);
 }
 if(relation == 0 && reset_ruler){
   strcpy(ruler_name,"      ");
   ruler=(double)UNIT;
   rulerx=0;rulery=0;rulerz=0;
   lrulerx=0;lrulery=0;lrulerz=0;
 }
 LOADOPERATION=0;
 return fail;
}

static void ReadOldFaces(long clen, short shader){
  long i;
  long id;
  face *Fp;
  unsigned char texture,brush;
  unsigned char *buffer=NULL,*b;
  if(shader)newface=clen/21;
  else      newface=clen/20;
  if(newface == 0)return;
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)longjmp(j_buf,0);  b=buffer;
  ReadFile(fq,buffer,clen,&dwRead,NULL);
  UpdateFaceHeap(Nface+newface);
  Fp=(MainFp+FaceBase);
  for(i=0;i<newface;i++){
    Fp=(MainFp+FaceBase+i);
    Fp->axis=Fp->pad=0;
    Fp->V[0]=VertexBase+GetLongFromBuffer(b); b+= 4;
    Fp->V[1]=VertexBase+GetLongFromBuffer(b); b+= 4;
    Fp->V[2]=VertexBase+GetLongFromBuffer(b); b+= 4;
    memcpy(Fp->color,b,3); b+= 3;
    texture = *b++;
    memcpy(Fp->matcol,b,3); b+= 3;
    brush = *b++;
    if(shader)Fp->axis = *b++;
    /* make texture byte correnpond to new arrangement */
    Fp->attrib=(texture & 0xf0) | (brush >> 4); Fp->attrib &= 0xfe;
    Fp->map = (brush & 0x1f);
    Fp->shader=(texture & 0xf);
    if((texture & 0x1f) == 0x04){ /* glass */
      Fp->shader=0; Fp->attrib |= 0x1;
      Fp->param=0;
    }
    else{
      Fp->param=      (Fp->matcol[1] & 0xf0);  /* bumpy params etc */
      Fp->matcol[1] = (Fp->matcol[1] & 0x0f);
    }
    Nface++;
    Fp++;
  }
  X__Free(buffer);
  return;
}

static void ReadFaces(long clen, BOOL short_form){
  long i;
  long id;
  unsigned short ids;
  face *Fp;
  unsigned char texture,brush;
  unsigned char *buffer=NULL,*b;
  if(short_form)newface=clen/18;
  else          newface=clen/24;
  if(newface == 0)return;
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)longjmp(j_buf,0);  b=buffer;
  ReadFile(fq,buffer,clen,&dwRead,NULL);
  UpdateFaceHeap(Nface+newface);
  Fp=(MainFp+FaceBase);
  for(i=0;i<newface;i++){
    Fp=(MainFp+FaceBase+i);
    if(short_form){
      Fp->V[0]=VertexBase+(long)GetUshortFromBuffer(b); b+= 2;
      Fp->V[1]=VertexBase+(long)GetUshortFromBuffer(b); b+= 2;
      Fp->V[2]=VertexBase+(long)GetUshortFromBuffer(b); b+= 2;
    }
    else{
      Fp->V[0]=VertexBase+(long)GetLongFromBuffer(b); b+= 4;
      Fp->V[1]=VertexBase+(long)GetLongFromBuffer(b); b+= 4;
      Fp->V[2]=VertexBase+(long)GetLongFromBuffer(b); b+= 4;
    }
    memcpy(Fp->color,b,3);  b += 3;
    memcpy(Fp->matcol,b,3); b += 3;
    Fp->attrib = *b++;
    Fp->param  = *b++;
    Fp->map    = *b++;
    Fp->shader = *b++;
    Fp->axis   = *b++;
    Fp->pad    = *b++;
    Nface++;
    Fp++;
  }
  X__Free(buffer);
  return;
}

static void ReadEdges(long clen, BOOL short_form){
  long i;
  long id1,id2;
  unsigned short ids1,ids2;
  edge *Ep;
  unsigned char *buffer=NULL,*b;
  if(short_form)newedge=clen/4;
  else          newedge=clen/8;
  if(newedge == 0)return;
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)longjmp(j_buf,0);  b=buffer;
  ReadFile(fq,buffer,clen,&dwRead,NULL);
  UpdateEdgeHeap(Nedge+newedge);
  Ep=(MainEp+EdgeBase);
  for(i=0;i<newedge;i++){
    if(short_form){
      Ep->V[0]=VertexBase+(long)GetUshortFromBuffer(b); b+= 2;
      Ep->V[1]=VertexBase+(long)GetUshortFromBuffer(b); b+= 2;
    }
    else{
      Ep->V[0]=VertexBase+(long)GetLongFromBuffer(b); b+= 4;
      Ep->V[1]=VertexBase+(long)GetLongFromBuffer(b); b+= 4;
    }
    Ep++;
    Nedge++;
  }
  X__Free(buffer);
  return;
}

static void ReadVertices(long clen){
  vertex *Vp;
  char c;
  long i,j;
  unsigned char *buffer=NULL,*b;
  newvert=clen/12;
  if(newvert == 0)return;
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)longjmp(j_buf,0);  b=buffer;
  ReadFile(fq,buffer,clen,&dwRead,NULL);
  UpdateVertexHeap(Nvert+newvert);
  Vp=(MainVp+VertexBase);
  for(i=0;i<newvert;i++){
    Vp->status=SELECTED;
    Vp->sp =  NULL;
    Vp->gp = 0;
    Vp->xyz[0]=GetLongFromBuffer(b); b+= 4;
    Vp->xyz[1]=GetLongFromBuffer(b); b+= 4;
    Vp->xyz[2]=GetLongFromBuffer(b); b+= 4;
    Vp->sp=NULL;
    if(Vp->xyz[0] > maxX)maxX=Vp->xyz[0];
    if(Vp->xyz[0] < minX)minX=Vp->xyz[0];
    if(Vp->xyz[1] > maxY)maxY=Vp->xyz[1];
    if(Vp->xyz[1] < minY)minY=Vp->xyz[1];
    if(Vp->xyz[2] > maxZ)maxZ=Vp->xyz[2];
    if(Vp->xyz[2] < minZ)minZ=Vp->xyz[2];
    NvertSelect++;
    Nvert++;
    Vp++;
  }
  X__Free(buffer);
  return;
}

static void ReadVertexMapping(long clen){
  long i,id,ix,iy,szz;
  vertex *Vp;
  unsigned char *buffer=NULL,*b;
  newglue=clen/12;
  if(newglue == 0)return;
  if(newglue > newvert){ /* dump the mapping */
    for(i=0;i<newglue;i++){
      getlon(&id); getlon(&ix); getlon(&iy);
    }
    return;
  }
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)longjmp(j_buf,0);  b=buffer;
  ReadFile(fq,buffer,clen,&dwRead,NULL);
  for(i=0;i<newglue;i++){
    id=GetLongFromBuffer(b); b+= 4;
    ix=GetLongFromBuffer(b); b+= 4;
    iy=GetLongFromBuffer(b); b+= 4;
    Vp=(MainVp+VertexBase+id);
    Vp->gp=1;
    Vp->x=(float)ix*0.000001;
    Vp->y=(float)iy*0.000001;
    NvertGlue++;
  }
  X__Free(buffer);
  return;
}

static void ReadSkeleton(long clen, short version){
  vertex *v;
  char c;
  long i,j,k,l,szz,n,x,y,z;
  skel *sp,**s;
  short dummy;
  getlon(&n);
  if(n == 0)return;
  EraseAllSkeleton();  /* start with new skeleton */
  szz=sizeof(skel *);
  s=(skel  **)X__Malloc((long)(szz*n));
  if(s == NULL){
    SendPrgmQuery(IDQ_NOMEM2,0);
    CloseHandle(fq);
    longjmp(j_buf,0);
  }
  getlon(&MainSp->xyz[0]); /* first Node into root node */
  getlon(&MainSp->xyz[1]);
  getlon(&MainSp->xyz[2]);
  getsht(&MainSp->id);
  if(version > 0){
    ReadFile(fq,MainSp->name,sizeof(char)*16,&dwRead,NULL);
  }
  if(version > 1){
    for(j=0;j<3;j++){getlon(&l); MainSp->u[j] = ((double)l)*1.0e-8;}
    for(j=0;j<3;j++){getlon(&l); MainSp->v[j] = ((double)l)*1.0e-8;}
    for(j=0;j<3;j++){getlon(&l); MainSp->w[j] = ((double)l)*1.0e-8;}
    for(j=0;j<8;j++)for(k=0;k<3;k++)getlon(&(MainSp->bx[j][k]));
  }
  MainSp->at=NULL;
  *s = MainSp;
  if(n > 1)for(i=1;i<n;i++){
    CreateSkeleton(NULL);
    getlon(&MainSp->xyz[0]);
    getlon(&MainSp->xyz[1]);
    getlon(&MainSp->xyz[2]);
    getsht(&MainSp->id);
    if(version > 0){
      ReadFile(fq,MainSp->name,sizeof(char)*16,&dwRead,NULL);
    }
    if(version > 1){
      for(j=0;j<3;j++){getlon(&l); MainSp->u[j] = ((double)l)*1.0e-8;}
      for(j=0;j<3;j++){getlon(&l); MainSp->v[j] = ((double)l)*1.0e-8;}
      for(j=0;j<3;j++){getlon(&l); MainSp->w[j] = ((double)l)*1.0e-8;}
      for(j=0;j<8;j++)for(k=0;k<3;k++)getlon(&(MainSp->bx[j][k]));
    }
    *(s+i)=MainSp;
  }
  if((sp = FirstSp->next) != NULL)while(sp != NULL){ /* fix up pointers */
    sp->at = *(s + (long)sp->id);
    sp=sp->next;
  }
  getlon(&n);
  if(n != newvert){
    SendPrgmQuery(IDQ_HERROR,0);
    for(i=0;i<n;i++)getsht(&dummy);
  }
  else{
    v=(MainVp+VertexBase);
    for(i=0;i<n;i++){
      getsht(&dummy);   /* put into new vertex the pointer to the sk joint */
      if(dummy > 0)v->sp = *(s+(long)dummy);
      else         v->sp = NULL;
      v++;
    }
  }
  if(nImaps > 0){ /* point brush skeletal pointer to assigned bone ! */
    for(i=0;i<nImaps;i++){
      if(iMap[i].SkId > -1)iMap[i].Sk = *(s+(long)iMap[i].SkId);
      else iMap[i].Sk=NULL;
    }
  }
  if(version < 2){ /* work out bounding boxes for old versions */
    TidyUpOldSkeltonType();
  }
  X__Free(s);
}

static void ReadTexture(long clen, short relation, short version){
 face *f;
 long i,k,l,m,ts,tid,temp,namesize,temp1;
 short temps;
 char j,temp_name[256];
 ReadFile(fq,&j,sizeof(char),&dwRead,NULL);
 ts=nImaps;     /* the present no of brushes loaded */
 if(j > 0){
   for(i=0;i<j;i++){
     k=nImaps;
     getlon(&namesize);
     ReadFile(fq,temp_name,namesize,&dwRead,NULL);
     getlon(&iMap[k].P[0]); getlon(&iMap[k].P[1]); getlon(&iMap[k].P[2]);
     getlon(&iMap[k].X[0]); getlon(&iMap[k].X[1]); getlon(&iMap[k].X[2]);
     getlon(&iMap[k].Y[0]); getlon(&iMap[k].Y[1]); getlon(&iMap[k].Y[2]);
     if(relation == 1){ /* the brush must be repositioned to new offset */
       iMap[k].P[0] += NpointerX-TempAxis.Origin.xyz[0];
       iMap[k].P[1] += NpointerY-TempAxis.Origin.xyz[1];
       iMap[k].P[2] += NpointerZ-TempAxis.Origin.xyz[2];
       iMap[k].X[0] += NpointerX-TempAxis.Origin.xyz[0];
       iMap[k].X[1] += NpointerY-TempAxis.Origin.xyz[1];
       iMap[k].X[2] += NpointerZ-TempAxis.Origin.xyz[2];
       iMap[k].Y[0] += NpointerX-TempAxis.Origin.xyz[0];
       iMap[k].Y[1] += NpointerY-TempAxis.Origin.xyz[1];
       iMap[k].Y[2] += NpointerZ-TempAxis.Origin.xyz[2];
     }
     getlon(&temp); iMap[k].Type=(char)temp;
     getlon(&temp); iMap[k].AnimFF=(short)temp;
     getlon(&temp); iMap[k].AnimDL=(short)temp;
     getlon(&temp); iMap[k].AnimLF=(short)temp;
     iMap[k].TypeR=iMap[k].TypeB=iMap[k].TypeT=STATICBRUSH;
     iMap[k].AnimFFr=iMap[k].AnimFFb=iMap[k].AnimFFt=1;
     iMap[k].AnimDLr=iMap[k].AnimDLb=iMap[k].AnimDLt=1;
     iMap[k].AnimLFr=iMap[k].AnimLFb=iMap[k].AnimLFt=1;
     getlon(&temp); iMap[k].SkId=(short)temp-1;  /* 0 is first brush id    */
     getlon(&temp); iMap[k].Map=(char)temp;      /* 0 plane 1 cylinder     */
     iMap[k].Angle=360.0;  /* default */         /* 2 = pl-moz 3 = cyl-moz */
     iMap[k].Lock=0;
     iMap[k].Vlock[0] = -1; iMap[k].Vlock[1] = -1; iMap[k].Vlock[2] = -1;
     getlon(&temp); /* expansion count */
     if(temp > 0){
       getlon(&temp1); iMap[k].Angle=temp1;
       if(temp > 1){
         getlon(&temp1); iMap[k].Lock = (char)temp1;
         if(iMap[k].Lock){ /* if map lock then get the vertex id's */
           short Lerror=0;
           getlon(&temp1);
           if(temp1 < 0 || temp1 >= newvert)Lerror=1;
           else iMap[k].Vlock[0] = VertexBase+temp1;
           getlon(&temp1);
           if(temp1 < 0 || temp1 >= newvert)Lerror=1;
           else iMap[k].Vlock[1] = VertexBase+temp1;
           getlon(&temp1);
           if(temp1 < 0 || temp1 >= newvert)Lerror=1;
           else iMap[k].Vlock[2] = VertexBase+temp1;
           if(Lerror){
             iMap[k].Lock=0;
             iMap[k].Vlock[0] = -1; iMap[k].Vlock[1] = -1; iMap[k].Vlock[2] = -1;
           }
           else     nLockediMaps++;  /* inc the lock count */
         }
         else{
           getlon(&temp1); getlon(&temp1); getlon(&temp1);
         }
         if(temp > 5){
           version=1;
           ReadFile(fq,iMap[k].N,32,&dwRead,NULL);
           ReadFile(fq,&iMap[k].s,1,&dwRead,NULL);
           ReadFile(fq,&iMap[k].r,1,&dwRead,NULL);
           ReadFile(fq,&iMap[k].b,1,&dwRead,NULL);
           iMap[k].t=0; iMap[k].Key=0;
           getsht(&iMap[k].sp);
           getsht(&iMap[k].rp);
           getsht(&iMap[k].bp);      iMap[k].tp=0;
           ReadFile(fq,iMap[k].d_colour,3,&dwRead,NULL);
           ReadFile(fq,iMap[k].s_colour,3,&dwRead,NULL);
           ReadFile(fq,iMap[k].a_colour,3,&dwRead,NULL);
           if(iMap[k].s){
             if((iMap[k].S=(char *)X__Malloc(namesize)) == NULL)goto ERR;
             ReadFile(fq,iMap[k].S,namesize,&dwRead,NULL);
           }
           else iMap[k].S=NULL;
           if(iMap[k].r){
             if((iMap[k].R=(char *)X__Malloc(namesize)) == NULL)goto ERR;
             ReadFile(fq,iMap[k].R,namesize,&dwRead,NULL);
           }
           else iMap[k].R=NULL;
           if(iMap[k].b){
             if((iMap[k].B=(char *)X__Malloc(namesize)) == NULL)goto ERR;
             ReadFile(fq,iMap[k].B,namesize,&dwRead,NULL);
           }
           else iMap[k].B=NULL;
           iMap[k].T=NULL;
           getsht(&temps);
         }
       }
     }
     if(version == 0){
       for(l=0;l<3;l++){
         iMap[k].d_colour[l]=iMap[k].s_colour[l]=iMap[k].a_colour[l]=0;
       }
       iMap[k].s=iMap[k].r=iMap[k].b=iMap[k].t=0;
       iMap[k].S=iMap[k].R=iMap[k].B=iMap[k].T=NULL;
       iMap[k].sp=100;
       iMap[k].rp= 50;
       iMap[k].bp= 50;
       iMap[k].tp=  0;
       strncpy(iMap[k].N,temp_name,30);
       if(iMap[k].Type == REFLBRUSH){
         if((iMap[k].R=(char *)X__Malloc(namesize)) == NULL)goto ERR;
         strncpy(iMap[k].R,temp_name,namesize); iMap[k].r=1;
       }
       else if(iMap[k].Type == BUMPBRUSH){
         if((iMap[k].B=(char *)X__Malloc(namesize)) == NULL)goto ERR;
         strncpy(iMap[k].B,temp_name,namesize); iMap[k].b=1;
       }
       else {
         if((iMap[k].S=(char *)X__Malloc(namesize)) == NULL)goto ERR;
         strncpy(iMap[k].S,temp_name,namesize); iMap[k].s=1;
       }
       if(iMap[k].Type != ANIMBRUSH)iMap[k].Type=STATICBRUSH;
     }
     if(nImaps >= NMAPSMAX){;}
     else nImaps++;
   }
   if(FaceBase > 0){               /* set new faces for appropriate brush */
     f=(MainFp+FaceBase); for(m=0;m<newface;m++){
       if((f->attrib & 0x40) == 0x40){
         tid = f->map;               /* extract the brush identifier */
         tid += ts;                  /* get the new brush id         */
         f->map=(unsigned char)tid;  /* put it back                  */
       }
       f++;
     }
   }
 }
 return;
 ERR:
 CloseHandle(fq);
 longjmp(j_buf,0);
}

static void ReadImageMap(long clen, short relation){
 face *f;
 long j,i,k,l,m,ts,tid,temp,temp1;
 short temps;
 char temp_str[256];
 unsigned char *buffer=NULL,*b;
 buffer=(unsigned char *)X__Malloc(clen);
 if(buffer == NULL)longjmp(j_buf,0);  b=buffer;
 ReadFile(fq,buffer,clen,&dwRead,NULL);
 j=GetLongFromBuffer(b); b+= 4;
 ts=nImaps;     /* the present no of brushes loaded */
 if(j > 0){
   for(i=0;i<j;i++){
     k=nImaps;
     iMap[k].P[0]=GetLongFromBuffer(b); b+= 4;
     iMap[k].P[1]=GetLongFromBuffer(b); b+= 4;
     iMap[k].P[2]=GetLongFromBuffer(b); b+= 4;
     iMap[k].X[0]=GetLongFromBuffer(b); b+= 4;
     iMap[k].X[1]=GetLongFromBuffer(b); b+= 4;
     iMap[k].X[2]=GetLongFromBuffer(b); b+= 4;
     iMap[k].Y[0]=GetLongFromBuffer(b); b+= 4;
     iMap[k].Y[1]=GetLongFromBuffer(b); b+= 4;
     iMap[k].Y[2]=GetLongFromBuffer(b); b+= 4;
     if(relation == 1){ /* the brush must be repositioned to new offset */
       iMap[k].P[0] += NpointerX-TempAxis.Origin.xyz[0];
       iMap[k].P[1] += NpointerY-TempAxis.Origin.xyz[1];
       iMap[k].P[2] += NpointerZ-TempAxis.Origin.xyz[2];
       iMap[k].X[0] += NpointerX-TempAxis.Origin.xyz[0];
       iMap[k].X[1] += NpointerY-TempAxis.Origin.xyz[1];
       iMap[k].X[2] += NpointerZ-TempAxis.Origin.xyz[2];
       iMap[k].Y[0] += NpointerX-TempAxis.Origin.xyz[0];
       iMap[k].Y[1] += NpointerY-TempAxis.Origin.xyz[1];
       iMap[k].Y[2] += NpointerZ-TempAxis.Origin.xyz[2];
     }
     iMap[k].Type=(char)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimFF=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimDL=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimLF=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].TypeR=(char)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimFFr=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimDLr=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimLFr=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].TypeB=(char)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimFFb=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimDLb=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimLFb=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].TypeT=(char)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimFFt=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimDLt=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].AnimLFt=(short)GetLongFromBuffer(b); b+= 4;
     iMap[k].SkId=(short)GetLongFromBuffer(b)-1; b+= 4;  /* 0 is first brush id    */
     iMap[k].Map=(char)GetLongFromBuffer(b); b+=4;       /* 0 plane 1 cylinder     */
     iMap[k].Angle=360.0;  /* default */         /* 2 = pl-moz 3 = cyl-moz */
     iMap[k].Lock=0;
     iMap[k].Vlock[0] = -1; iMap[k].Vlock[1] = -1; iMap[k].Vlock[2] = -1;
     iMap[k].Angle=GetLongFromBuffer(b); b+=4;
     iMap[k].Lock = (char)GetLongFromBuffer(b); b+=4;
     if(iMap[k].Lock){ /* if map lock then get the vertex id's */
       short Lerror=0;
       temp1=GetLongFromBuffer(b); b+=4;
       if(temp1 < 0 || temp1 >= newvert)Lerror=1;
       else iMap[k].Vlock[0] = VertexBase+temp1;
       temp1=GetLongFromBuffer(b); b+=4;
       if(temp1 < 0 || temp1 >= newvert)Lerror=1;
       else iMap[k].Vlock[1] = VertexBase+temp1;
       temp1=GetLongFromBuffer(b); b+=4;
       if(temp1 < 0 || temp1 >= newvert)Lerror=1;
       else iMap[k].Vlock[2] = VertexBase+temp1;
       if(Lerror){
         iMap[k].Lock=0;
         iMap[k].Vlock[0] = -1; iMap[k].Vlock[1] = -1; iMap[k].Vlock[2] = -1;
       }
       else     nLockediMaps++;  /* inc the lock count */
     }
     else{
       b += (3*4);
     }
     memcpy(iMap[k].N,b,32);  b+= 32;
     iMap[k].s = *b++;
     iMap[k].r = *b++;
     iMap[k].b = *b++;
     iMap[k].t = *b++;
     iMap[k].Key = *b++;
     iMap[k].sp=GetShortFromBuffer(b); b+= 2;
     iMap[k].rp=GetShortFromBuffer(b); b+= 2;
     iMap[k].bp=GetShortFromBuffer(b); b+= 2;
     iMap[k].tp=GetShortFromBuffer(b); b+= 2;
     memcpy(iMap[k].d_colour,b,3); b += 3;
     memcpy(iMap[k].s_colour,b,3); b += 3;
     memcpy(iMap[k].a_colour,b,3); b += 3;
     memcpy(iMap[k].k_colour,b,3); b += 3;
     memcpy(temp_str,b,256); b += 256;
     if(iMap[k].s){
       if((iMap[k].S=(char *)X__Malloc(256)) == NULL)goto ERR;
       memcpy(iMap[k].S,temp_str,256);
     }
     else iMap[k].S=NULL;
     memcpy(temp_str,b,256); b += 256;
     if(iMap[k].r){
       if((iMap[k].R=(char *)X__Malloc(256)) == NULL)goto ERR;
       memcpy(iMap[k].R,temp_str,256);
     }
     else iMap[k].R=NULL;
     memcpy(temp_str,b,256); b += 256;
     if(iMap[k].b){
       if((iMap[k].B=(char *)X__Malloc(256)) == NULL)goto ERR;
       memcpy(iMap[k].B,temp_str,256);
     }
     else iMap[k].B=NULL;
     memcpy(temp_str,b,256); b += 256;
     if(iMap[k].t){
       if((iMap[k].T=(char *)X__Malloc(256)) == NULL)goto ERR;
       memcpy(iMap[k].T,temp_str,256);
     }
     else iMap[k].T=NULL;
     if(nImaps >= NMAPSMAX){;}
     else nImaps++;
   }
   if(FaceBase > 0){               /* set new faces for appropriate brush */
     f=(MainFp+FaceBase); for(m=0;m<newface;m++){
       if((f->attrib & 0x40) == 0x40){
         tid = f->map;              /* extract the brush identifier */
         tid += ts;                 /* get the new brush id         */
         f->map=(unsigned char)tid; /* put it back                  */
       }
       f++;
     }
   }
 }
 X__Free(buffer);
 return;
 ERR:
 if(buffer != NULL)X__Free(buffer);
 CloseHandle(fq);
 longjmp(j_buf,0);
 return;
}

static void ReadModWireframe(long clen){
 long i,ne,np;
 getlon(&np);  getlon(&ne);
 i=((np*3)+(ne*2)+2)*sizeof(long);
 if(clen != i){   /* if chunk is void */
   getchunk(clen-2);
   return;
 }
 w_frame.p = (point *)X__Malloc(np*sizeof(point));
 w_frame.e = (twoLongs *)X__Malloc(ne*sizeof(twoLongs));
 if(w_frame.p == NULL || w_frame.e == NULL){
   if(w_frame.p != NULL)X__Free(w_frame.p); w_frame.p=NULL;
   if(w_frame.e != NULL)X__Free(w_frame.e); w_frame.e=NULL;
   getchunk(clen-2);
   return;
 }
 w_frame.Np=np;  w_frame.Ne=ne;
 for(i=0;i<w_frame.Np;i++){
   getlon(&w_frame.p[i][0]);
   getlon(&w_frame.p[i][1]);
   getlon(&w_frame.p[i][2]);
 }
 for(i=0;i<w_frame.Ne;i++){
   getlon(&w_frame.e[i][0]);
   getlon(&w_frame.e[i][1]);
 }
}

static void ReadShadersAxes(long clen, short relation){
 face *f;
 long i,j,k,n;
 char temp_name[32];
 unsigned char tid,ts;
 n=clen/68;     /* size of name + 4 vectors  32+4*9 */
 ts=nSha;       /* the present no of shaders loaded */
 if(n > 0)for(i=0;i<n;i++){
   k=nSha;
   ReadFile(fq,iSha[k].N,32,&dwRead,NULL);
   for(j=0;j<3;j++)getlon(&iSha[k].P[j]);
   for(j=0;j<3;j++)getlon(&iSha[k].X[j]);
   for(j=0;j<3;j++)getlon(&iSha[k].Y[j]);
   if(relation == 1){ /* the brush must be repositioned to new offset */
     iSha[k].P[0] += NpointerX-TempAxis.Origin.xyz[0];
     iSha[k].P[1] += NpointerY-TempAxis.Origin.xyz[1];
     iSha[k].P[2] += NpointerZ-TempAxis.Origin.xyz[2];
     iSha[k].X[0] += NpointerX-TempAxis.Origin.xyz[0];
     iSha[k].X[1] += NpointerY-TempAxis.Origin.xyz[1];
     iSha[k].X[2] += NpointerZ-TempAxis.Origin.xyz[2];
     iSha[k].Y[0] += NpointerX-TempAxis.Origin.xyz[0];
     iSha[k].Y[1] += NpointerY-TempAxis.Origin.xyz[1];
     iSha[k].Y[2] += NpointerZ-TempAxis.Origin.xyz[2];
   }
   if(nSha > NAXISMAX){;}
   else nSha++;
 }
 if(FaceBase > 0){
   f=(MainFp+FaceBase); for(j=FaceBase;j<Nface;j++){
     tid = f->axis;              /* extract the shader axis id      */
     tid += ts;                  /* get the shaders new axis id     */
     if(tid >= nSha)tid=0;       /* put it back -if out of range    */
     f->axis=tid;                /* set to base axis (default)      */
   }
   f=(MainFp+FaceBase+j);
 }
 return;
}

static void ReadShader(long clen){
 long i,ns,fs,ps,in_use;
 getlon(&ns);
 if(ns > 0)for(i=0;i<ns;i++){
   getlon(&in_use);
   xShaders[i].in_use=in_use;
   if(in_use){
     getlon(&fs); getlon(&ps);
     ReadFile(fq,xShaders[i].name,fs,&dwRead,NULL);
     if((xShaders[i].parameters=(char *)X__Malloc(ps)) == NULL)getchunk(ps);
     else ReadFile(fq,xShaders[i].parameters,ps,&dwRead,NULL);
   }
 }
}

// Save the MFX file

static void WriteWireframe(void);
static void WriteBezierCurve(void);
static void WriteVertexMapping(short status);
static void WriteVerticesWithOffset(point p, long invert, short short_form,
                                    short status);
static void WriteOffsetTextures(point p, unsigned char maps, long invert);
static void WriteFacesWithMapCheck(BOOL short_form);
static void WriteShaders(short status);
static void WriteShadersAxes(void);
static void WriteOffsetShadersAxes(point p, long invert);

void SaveObject(char *Filename, short status){
  long pos1,pos2,pos3,pos4,pos5,pos6,pos7,pos8,pos9,posA,posB,
       posC,posD,posE,posF,posG,posH,pose;
  long FORMsize;
  HCURSOR hSave;
  short id;
  FORMsize=0; newvert=0; newface=0; newedge=0; newglue=0;
#ifdef __DEMO__
  if(status != 2)return;
#endif
  /* status = 3 -> saving with - no extra display  save everything!! */
  /* status = 4 -> save as "1" but without file check                */
  if(status < 3){   /* no memory problem */
    if( (fo=fopen(Filename,"r")) != NULL){ /* check for old file */
      fclose(fo);
      if(SendPrgmQuery(IDQ_FILEEXISTS,1) != IDYES)return;
    }
  }
  if(status == 4)status=1;
  if( (fo=fopen(Filename,"wb")) == NULL){   /* now open for reading */
    SendPrgmQuery(IDQ_FILEOPENFAIL,0);
    return;
  }
  if(status < 2)SendPrgmMessage(IDS_SAVING,0);
  if(status < 3)hSave=SetCursor(ghcurWait);
  if(status == 1 || status == 3)EDIT_ACTION=NO;
  fprintf(fo,"FORM");
  pos1=ftell(fo); outlng(FORMsize);
  fprintf(fo,"OFXM");
  fprintf(fo,"AXIS"); outlng(48L); /* size of data chunk */
  if(status == 2){  /* dont use offset for rendering in triview */
    outlng(0); outlng(0); outlng(0);
  }
  else{
    outlng(ObjectAxis.Origin.xyz[0]);
    outlng(ObjectAxis.Origin.xyz[1]);
    outlng(ObjectAxis.Origin.xyz[2]);
  }
  outlng(iSha[0].P[0]);  /* store the algorithmic texture origin */
  outlng(iSha[0].P[1]);  /* in the object axis structure         */
  outlng(iSha[0].P[2]);
  outlng(iSha[0].X[0]);
  outlng(iSha[0].X[1]);
  outlng(iSha[0].X[2]);
  outlng(iSha[0].Y[0]);
  outlng(iSha[0].Y[1]);
  outlng(iSha[0].Y[2]);
  fprintf(fo,"OFFS"); outlng(16L);      /* follow point */
  outlng(3L);
  outlng(ObjectAxis.Offset.xyz[0]);
  outlng(ObjectAxis.Offset.xyz[1]);
  outlng(ObjectAxis.Offset.xyz[2]);
  if(Nvert > 0){
    fprintf(fo,"VERT"); pos2=ftell(fo); outlng(0L); /* filled in later */
    WriteVertices(status);
  }
  else pos2 = -1;
  if(Nedge > 0){
    if(Nedge < 65536){
      fprintf(fo,"edge"); pos3=ftell(fo); outlng(0L); WriteEdges(TRUE);
    }
    else {
      fprintf(fo,"EDGE"); pos3=ftell(fo); outlng(0L); WriteEdges(FALSE);
    }
  }
  else pos3 = -1;
  if(Nface > 0){
    if(Nface < 65536){
      fprintf(fo,"SFCE"); pos4=ftell(fo); outlng(0L); WriteFaces(TRUE);
    }
    else{
      fprintf(fo,"LFCE"); pos4=ftell(fo); outlng(0L); WriteFaces(FALSE);
    }
  }
  else pos4 = -1;
  if(NvertGlue > 0){
    fprintf(fo,"VTEX"); posB=ftell(fo); outlng(12*newglue);
    WriteVertexMapping(status);
  }
  else posB = -1;
  if(xShaders[0].in_use || xShaders[1].in_use ||
     xShaders[2].in_use || xShaders[3].in_use ||
     xShaders[4].in_use || xShaders[5].in_use ||
     xShaders[6].in_use || xShaders[7].in_use ){
    fprintf(fo,"SHDR"); posC=ftell(fo); outlng(0L);
    WriteShaders(status);
    posD=ftell(fo);
  }
  else posC = -1;
  if(nSha > 1){
    fprintf(fo,"AXES"); posE=ftell(fo); outlng(68*(nSha-1));
    WriteShadersAxes();
    posF=ftell(fo);
  }
  else posE = -1;
  if(nImaps > 0){
    fprintf(fo,"IMAP"); pos5=ftell(fo); outlng(0L);
    WriteTextures();
    pos6=ftell(fo);
  }
  else pos5 = -1;         /*! root node only implies no skeleton */
  if((status == 1 || status == 3) &&
      Nskel > 1){/* if saving whole object then save skeleton */
    fprintf(fo,"SK02"); pos7=ftell(fo); outlng(0L);
    WriteSkeleton();
    pos8=ftell(fo);
  }
  else pos7 = -1;
  if(MirrorFlag >= 0 && (status == 1 || status == 3)){ /* only save mirror in full save */
    fprintf(fo,"MPID"); outlng((long)(1*sizeof(long)));
    outlng((long)MirrorFlag);
  }
  fprintf(fo,"MATL"); outlng((long)(5*sizeof(long)));
  outlng(4L);
  outlng((long)ShadowFlag);
  outlng((long)(refractive_index*10000));
  outlng(0);
  outlng(0);
  fprintf(fo,"UNIT"); outlng((long)(7*sizeof(long)+8));
  outlng((long)ruler);
  fwrite(ruler_name,1,8,fo);
  outlng(lrulerx); outlng(lrulery); outlng(lrulerz);
  outlng(rulerx);  outlng(rulery);  outlng(rulerz);
  fprintf(fo,"GRID"); outlng((long)(2*sizeof(long)));
  outlng(grid_size); outlng(grid_size); /* second is unused */
  WriteWireframe();
  if((status == 1 || status == 3) && N_Bcurves > 0){
     fprintf(fo,"BEZ1"); pos9=ftell(fo); outlng(0L);
     WriteBezierCurve();
     posA=ftell(fo);
  }
  else pos9 = -1;
  if(Nnurbs > 0){
    fprintf(fo,"NURB"); posH=ftell(fo); outlng(0);
    WriteNurbs(fo);
    posG=ftell(fo);
  }
  else posH = -1;
  pose=ftell(fo);  /* the end of the object */
  if(pos2 > 0){fseek(fo,pos2,0); outlng(12*newvert);}
  if(pos3 > 0){
    fseek(fo,pos3,0);
    if(Nedge < 65536)outlng(4*newedge);
    else             outlng(8*newedge);
  }
  if(pos4 > 0){
    fseek(fo,pos4,0);
    if(Nface < 65536)outlng(18*newface);
    else             outlng(24*newface);
  }
  if(posB > 0){fseek(fo,posB,0); outlng(12*newglue);}
  if(posE > 0){fseek(fo,posE,0); outlng(posF-posE-4);} /* shaders new axes  */
  if(posC > 0){fseek(fo,posC,0); outlng(posD-posC-4);} /* extern shader     */
  if(pos5 > 0){fseek(fo,pos5,0); outlng(pos6-pos5-4);} /* brushes !inc size */
  if(pos7 > 0){fseek(fo,pos7,0); outlng(pos8-pos7-4);} /* skeleton          */
  if(pos9 > 0){fseek(fo,pos9,0); outlng(posA-pos9-4);} /* bezier curves     */
  if(posH > 0){fseek(fo,posH,0); outlng(posG-posH-4);} /* Nurbs surfaces    */
  fseek(fo,pos1,0); outlng(pose-8);
  fclose(fo);
  if(status < 2)SendPrgmMessage(IDS_CURRENT,0);
  if(status < 3)SetCursor(hSave);
  return;
}

static void WriteShadersAxes(void){
 long i,j;
 for(i=1;i<nSha;i++){
   fwrite(iSha[i].N,1,32,fo);
   for(j=0;j<3;j++)outlng(iSha[i].P[j]);
   for(j=0;j<3;j++)outlng(iSha[i].X[j]);
   for(j=0;j<3;j++)outlng(iSha[i].Y[j]);
 }
}

static void WriteTextures(void){
 long i,temp;
 skel *sp;
 long v;
 char n256[256];
 i=0;
 memset(n256,0,256);
 sp=FirstSp; while(sp != NULL){ /* set up sk id */
   sp->id=i; i++;
   sp=sp->next;
 }
 outlng((long)nImaps);
 for(i=0;i<nImaps;i++){
   outlng(iMap[i].P[0]); outlng(iMap[i].P[1]); outlng(iMap[i].P[2]);
   outlng(iMap[i].X[0]); outlng(iMap[i].X[1]); outlng(iMap[i].X[2]);
   outlng(iMap[i].Y[0]); outlng(iMap[i].Y[1]); outlng(iMap[i].Y[2]);
   temp=iMap[i].Type;   outlng(temp);   temp=iMap[i].AnimFF; outlng(temp);
   temp=iMap[i].AnimDL; outlng(temp);   temp=iMap[i].AnimLF; outlng(temp);
   temp=iMap[i].TypeR;   outlng(temp);  temp=iMap[i].AnimFFr; outlng(temp);
   temp=iMap[i].AnimDLr; outlng(temp);  temp=iMap[i].AnimLFr; outlng(temp);
   temp=iMap[i].TypeB;   outlng(temp);  temp=iMap[i].AnimFFb; outlng(temp);
   temp=iMap[i].AnimDLb; outlng(temp);  temp=iMap[i].AnimLFb; outlng(temp);
   temp=iMap[i].TypeT;   outlng(temp);  temp=iMap[i].AnimFFt; outlng(temp);
   temp=iMap[i].AnimDLt; outlng(temp);  temp=iMap[i].AnimLFt; outlng(temp);
   if(iMap[i].Sk != NULL){
     temp=((iMap[i].Sk)->id)+1; outlng(temp);/* 1 is first file brush*/
   }
   else{
     temp=0; outlng(temp);
   }
   temp=iMap[i].Map; outlng(temp);   /* 0 = plane 1 = cyl. map  2/3 = moz */
   temp=iMap[i].Angle; outlng(temp); /* angle for cylinder brushes */
   if(iMap[i].Lock &&
      (MainVp+iMap[i].Vlock[0])->id >= 0 &&        /* out a tex map if lock */
      (MainVp+iMap[i].Vlock[1])->id >= 0 &&        /* and if vertices OK    */
      (MainVp+iMap[i].Vlock[2])->id >= 0 ){
     temp=iMap[i].Lock;   outlng(temp);
     v=iMap[i].Vlock[0];
     temp=(MainVp+v)->id; outlng(temp);
     v=iMap[i].Vlock[1];
     temp=(MainVp+v)->id; outlng(temp);
     v=iMap[i].Vlock[2];
     temp=(MainVp+v)->id; outlng(temp);
   }
   else{  /* map is not locked */
     temp=0; outlng(temp); outlng(temp); outlng(temp); outlng(temp);
   }
   fwrite(iMap[i].N,1,32,fo);
   fwrite(&iMap[i].s,1,1,fo);
   fwrite(&iMap[i].r,1,1,fo);
   fwrite(&iMap[i].b,1,1,fo);
   fwrite(&iMap[i].t,1,1,fo);
   fwrite(&iMap[i].Key,1,1,fo);
   outsht(iMap[i].sp);
   outsht(iMap[i].rp);
   outsht(iMap[i].bp);
   outsht(iMap[i].tp);
   fwrite(iMap[i].d_colour,1,3,fo);
   fwrite(iMap[i].s_colour,1,3,fo);
   fwrite(iMap[i].a_colour,1,3,fo);
   fwrite(iMap[i].k_colour,1,3,fo);
   if(iMap[i].s)fwrite(iMap[i].S,1,256,fo); else fwrite(n256,1,256,fo);
   if(iMap[i].r)fwrite(iMap[i].R,1,256,fo); else fwrite(n256,1,256,fo);
   if(iMap[i].b)fwrite(iMap[i].B,1,256,fo); else fwrite(n256,1,256,fo);
   if(iMap[i].t)fwrite(iMap[i].T,1,256,fo); else fwrite(n256,1,256,fo);
 }
}

static void WriteVertices(short status){
 long i,j;
 vertex  *vp;
 i=0;
 vp=MainVp; for(j=0;j<Nvert;j++){
  if(status == 2){ /* for use in renderer */
    if(intriview(vp)){
      outlng(vp->xyz[0]); outlng(vp->xyz[1]); outlng(vp->xyz[2]);
      vp->id=i;
      i++; newvert++;
    }
    else vp->id = -1;
  }
  else if( (status==1) || (status == 3) || (vp->status==SELECTED) ){
    outlng(vp->xyz[0]); outlng(vp->xyz[1]); outlng(vp->xyz[2]);
    vp->id=i;
    i++; newvert++;
  }
  else vp->id = -1;
  vp++;
 }
}

static void WriteVerticesWithOffset(point p, long invert, short short_form,
                                    short status){
 long i,j;
 vertex  *vp;
 i=0;
 vp=MainVp; for(j=0;j<Nvert;j++){
  if(vp->status==status){
    outlng((vp->xyz[0]-p[0])*invert);
    outlng((vp->xyz[1]-p[1])*invert);
    outlng(vp->xyz[2]-p[2]);
    vp->id=i;
    i++; newvert++;
  }
  else vp->id = -1;
  vp++;
 }
}

static void WriteVertexMapping(short status){
 long j,i,ix,iy;
 vertex *vp;
 i=0;
 vp=MainVp; for(j=0;j<Nvert;j++){
  if(status == 2){ /* for use in renderer */
    if(intriview(vp)){
      if(vp->gp == 1){
        ix=(long)(vp->x * 1.0e6); iy=(long)(vp->y * 1.0e6);
        outlng(i); outlng(ix); outlng(iy);
        newglue++;
      }
      i++;
    }
  }
  else if( (status==1) || (status == 3) || (vp->status==SELECTED) ){
    if(vp->gp == 1){
      ix=(long)(vp->x * 1.0e6); iy=(long)(vp->y * 1.0e6);
      outlng(i); outlng(ix); outlng(iy);
      newglue++;
    }
    i++;
  }
  vp++;
 }
}

static void WriteEdges(BOOL short_form){
 unsigned short ids1,ids2;
 long i,idl1,idl2;
 edge *ep;
 vertex *V0,*V1;
 ep=MainEp;
 for(i=0;i<Nedge;i++){
   V0=(MainVp+ep->V[0]); V1=(MainVp+ep->V[1]);
   if((V0->id >= 0) && (V1->id >= 0)){
     if(short_form){
       ids1=(unsigned short)V0->id; ids2=(unsigned short)V1->id;
       outusht(ids1); outusht(ids2);
     }
     else{
       idl1=V0->id; idl2=V1->id;
       outlng(idl1); outlng(idl2);
     }
     newedge++;
   }
   ep++;
 }
}

static void WriteFaces(BOOL short_form){
 face *f;
 unsigned short ids;
 long i,idl;
 vertex *V0,*V1,*V2;
 f=MainFp;
 for(i=0;i<Nface;i++){
   V0=(MainVp+f->V[0]); V1=(MainVp+f->V[1]); V2=(MainVp+f->V[2]);
   if((V0->id >= 0) &&
      (V1->id >= 0) &&
      (V2->id >= 0)){
     if(short_form){
       ids=(unsigned short)V0->id; outusht(ids);
       ids=(unsigned short)V1->id; outusht(ids);
       ids=(unsigned short)V2->id; outusht(ids);
     }
     else{
       idl=V0->id; outlng(idl);
       idl=V1->id; outlng(idl);
       idl=V2->id; outlng(idl);
     }
     fwrite(f->color,1,3,fo);
     fwrite(f->matcol,1,3,fo);
     fwrite(&f->attrib,1,1,fo);
     fwrite(&f->param,1,1,fo);
     fwrite(&f->map,1,1,fo);
     fwrite(&f->shader,1,1,fo);
     fwrite(&f->axis,1,1,fo);
     fwrite(&f->pad,1,1,fo);
     newface++;
   }
   f++;
 }
}

static void WriteFacesWithMapCheck(BOOL short_form){
 unsigned char map;
 face *f;
 long i,idl;
 unsigned short ids;
 vertex *V0,*V1,*V2;
 f=MainFp;
 for(i=0;i<Nface;i++){
   V0=(MainVp+f->V[0]); V1=(MainVp+f->V[1]); V2=(MainVp+f->V[2]);
   if((V0->id >= 0) &&
      (V1->id >= 0) &&
      (V2->id >= 0)){
     if(short_form){
       ids=(unsigned short)V0->id; outusht(ids);
       ids=(unsigned short)V1->id; outusht(ids);
       ids=(unsigned short)V2->id; outusht(ids);
     }
     else{
       idl=V0->id; outlng(idl);
       idl=V1->id; outlng(idl);
       idl=V2->id; outlng(idl);
     }
     fwrite(f->color,1,3,fo);
     fwrite(f->matcol,1,3,fo);
     fwrite(&f->attrib,1,1,fo);
     fwrite(&f->param,1,1,fo);
     if((f->attrib & 0x40) == 0x40){/* map in use */
       map=f->map;
       map=(unsigned char)(iMap[(long)map].in_use);  /* this must be >= 0*/
       fwrite(&map,1,1,fo);
     }
     else fwrite(&f->map,1,1,fo);
     fwrite(&f->shader,1,1,fo);
     fwrite(&f->axis,1,1,fo);
     fwrite(&f->pad,1,1,fo);
     newface++;
   }
   f++;
 }
}

static void WriteBezierCurve(void){
 long i,j,id;
 vertex *vp;
 outlng(N_Bcurves);
 for(i=0;i<N_Bcurves;i++){
   outlng((long)Bcurves[i].Nc); outlng((long)Bcurves[i].Np);
   fwrite((char *)Bcurves[i].name,1,8,fo);
   for(j=0;j<Bcurves[i].Nc;j++){
     id=(MainVp+Bcurves[i].Vc[j])->id; outlng(id);
   }
   for(j=0;j<Bcurves[i].Np;j++){
     id=(MainVp+Bcurves[i].Vp[j])->id; outlng(id);
   }
 }
}

static void WriteSkeleton(void){
 short i,j;
 skel *sp;
 vertex *vp;
 long l;
 long k;
 i=0;
 sp=FirstSp; while(sp != NULL){
   sp->id=i; i++;
   sp=sp->next;
 }
 outlng(Nskel);
 sp=FirstSp; while(sp != NULL){
   outlng(sp->xyz[0]); outlng(sp->xyz[1]); outlng(sp->xyz[2]);
   if(sp->at != NULL)outsht(sp->at->id);
   else              outsht(0);
   fwrite(sp->name,sizeof(char),16,fo);
   for(j=0;j<3;j++){l=(long)(sp->u[j]*1.0e8); outlng(l);}
   for(j=0;j<3;j++){l=(long)(sp->v[j]*1.0e8); outlng(l);}
   for(j=0;j<3;j++){l=(long)(sp->w[j]*1.0e8); outlng(l);}
   for(i=0;i<8;i++)for(j=0;j<3;j++)outlng(sp->bx[i][j]);
   sp=sp->next;
 }
 outlng(Nvert);
 vp=MainVp; for(k=0;k<Nvert;k++){
   if(vp->sp != NULL)outsht(vp->sp->id);
   else              outsht(0);
   vp++;
 }
}

static void WriteWireframe(void){
 long i;
 if(w_frame.Np == 0 || w_frame.Ne == 0)return;
 fprintf(fo,"WFRM");
 outlng((long)(((w_frame.Np*3)+(w_frame.Ne*2)+2)*sizeof(long)));
 outlng(w_frame.Np); outlng(w_frame.Ne);
 for(i=0;i<w_frame.Np;i++){
   outlng(w_frame.p[i][0]);
   outlng(w_frame.p[i][1]);
   outlng(w_frame.p[i][2]);
 }
 for(i=0;i<w_frame.Ne;i++){
   outlng(w_frame.e[i][0]); outlng(w_frame.e[i][1]);
 }
}

void SaveObjects(char *rootname, short do_all, short rotate180){
 char Filename[255],NodeName[16];
 vertex *vp,*V0,*V1,*V2;
 face   *fp;
 edge   *ep;
 skel   *sp,*sp1;   /* sp1 gives the offset for follow point */
 long j,i,nf,ne,map,invert;
 long pos1,pos2,pos3,pos4,pos5,pos6,posA,posB,posC,posD,pose;
 long FORMsize;
 short id;
 BOOL check_exi=TRUE;
#ifdef __DEMO__
 MessageBox(NULL,"Not in Demo","Demo",MB_OK);
 return;
#else
 if(rotate180)invert = -1;
 else         invert =  1;
 sp=MainSp; while(sp != NULL){
   if(sp->at == FirstSp || do_all){
     if((vp=MainVp) != NULL)for(j=0;j<Nvert;j++){
       if(vp->status == SELECTED && vp->sp == sp)vp->status=INEDITOR;
       vp++;
     }
     sp1=MainSp; while(sp1 != NULL){
       if(sp1->at == sp)break;
       sp1=sp1->last;
     }
     if(nImaps > 0)for(i=0;i<nImaps;i++)iMap[i].in_use = -1;
     nf=0; if((fp=MainFp) != NULL)for(j=0;j<Nface;j++){
       V0=(MainVp+fp->V[0]); V1=(MainVp+fp->V[1]); V2=(MainVp+fp->V[2]);
       if(V0->status == INEDITOR && V1->status == INEDITOR &&
                                    V2->status == INEDITOR){
         if((fp->attrib & 0x40) == 0x40){/* map in use */
           map=(long)(fp->map);
           iMap[map].in_use=1;
         }
         nf++;
       }
       fp++;
     }
     map=0;
     if(nImaps > 0)for(i=0;i<nImaps;i++){
       if(iMap[i].in_use == 1){
         iMap[i].in_use = (short)map;
         map++;
       }
     }
     ne=0; if(nf == 0){
       if((ep=MainEp) != NULL)for(j=0;j<Nedge;j++){/* no faces check edges */
         V0=(MainVp+ep->V[0]); V1=(MainVp+ep->V[1]);
         if(V0->status == INEDITOR && V1->status == INEDITOR)ne++;
         ep++;
       }
     }
     if(nf > 0 || ne > 0){ /* only write if there are any faces or edges */
       FORMsize=0; newvert=0; newface=0; newedge=0;  /* edges => faces   */
       strcpy(NodeName,sp->name);
       id=min(strlen(sp->name),8); NodeName[id]=0;
       sprintf(Filename,"Saving: %s%s.mfx",rootname,NodeName);
       SendPrgmText(Filename);
       sprintf(Filename,"%s%s.mfx",rootname,NodeName);
       if(check_exi){
         if((fo=fopen(Filename,"r")) != NULL){ /* check for old file */
           fclose(fo);
           if(SendPrgmQuery(IDQ_FILESEXIST,1) != IDYES)goto SKIPIT;
           else check_exi=FALSE;
           if(id < 1)goto SKIPIT;
         }
       }
       if( (fo=fopen(Filename,"wb")) == NULL){   /* now open for reading */
         SendPrgmQuery(IDQ_FILEOPENFAIL,0);
         goto SKIPIT;
       }
       fprintf(fo,"FORM");
       pos1=ftell(fo); outlng(FORMsize);
       fprintf(fo,"OFXM");
       fprintf(fo,"AXIS"); outlng(48L); /* size of data chunk */
       outlng(0);  outlng(0);  outlng(0);
       outlng((iSha[0].P[0]-sp->xyz[0])*invert);  /* store the algorithmic texture origin */
       outlng((iSha[0].P[1]-sp->xyz[1])*invert);  /* in the object axis structure         */
       outlng( iSha[0].P[2]-sp->xyz[2]);
       outlng((iSha[0].X[0]-sp->xyz[0])*invert);
       outlng((iSha[0].X[1]-sp->xyz[1])*invert);
       outlng( iSha[0].X[2]-sp->xyz[2]);
       outlng((iSha[0].Y[0]-sp->xyz[0])*invert);
       outlng((iSha[0].Y[1]-sp->xyz[1])*invert);
       outlng( iSha[0].Y[2]-sp->xyz[2]);
       fprintf(fo,"OFFS"); outlng(16L);      /* follow point */
       outlng(3L);
       outlng((sp1->xyz[0]-sp->xyz[0])*invert);
       outlng((sp1->xyz[1]-sp->xyz[1])*invert);
       outlng(sp1->xyz[2]-sp->xyz[2]);
       fprintf(fo,"VERT"); pos2=ftell(fo); outlng(0L);
       WriteVerticesWithOffset(sp->xyz,invert,0,INEDITOR);
       if(Nedge < 65536){
         fprintf(fo,"edge"); pos3=ftell(fo);
         outlng(0L); WriteEdges(TRUE);
       }
       else {
         fprintf(fo,"EDGE"); pos3=ftell(fo);
         outlng(0L); WriteEdges(FALSE);
       }
       if(nf > 0){
         if(Nface < 65536){
           fprintf(fo,"SFCE"); pos4=ftell(fo);
           outlng(0L); WriteFacesWithMapCheck(TRUE);
         }
         else{
           fprintf(fo,"LFCE"); pos4=ftell(fo);
           outlng(0L); WriteFacesWithMapCheck(FALSE);
         }
       }
       else pos4 = -1;
       if(map > 0){
         fprintf(fo,"IMAP"); pos5=ftell(fo); outlng(0L);
         WriteOffsetTextures(sp->xyz,(unsigned char)map,invert);
         pos6=ftell(fo);
       }
       else pos5 = -1;

       if(xShaders[0].in_use || xShaders[1].in_use ||
          xShaders[2].in_use || xShaders[3].in_use ||
          xShaders[4].in_use || xShaders[5].in_use ||
          xShaders[6].in_use || xShaders[7].in_use ){
         fprintf(fo,"SHDR"); posA=ftell(fo); outlng(0L);
         WriteShaders(1);
         posB=ftell(fo);
       }
       else posA = -1;
       if(nSha > 1){
         fprintf(fo,"AXES"); posC=ftell(fo); outlng(68*(nSha-1));
         WriteOffsetShadersAxes(sp->xyz,invert);
         posD=ftell(fo);
       }
       else posC = -1;

       fprintf(fo,"MATL"); outlng((long)(5*sizeof(long)));
       outlng(4L);
       outlng((long)ShadowFlag);
       outlng((long)(refractive_index*10000));
       outlng(0); outlng(0); /* not used now */
       fprintf(fo,"UNIT"); outlng((long)(7*sizeof(long)+8));
       outlng((long)ruler);
       fwrite(ruler_name,1,8,fo);
       outlng(lrulerx); outlng(lrulery); outlng(lrulerz);
       outlng(rulerx);  outlng(rulery);  outlng(rulerz);
       fprintf(fo,"GRID"); outlng((long)(2*sizeof(long)));
       outlng(grid_size); outlng(grid_size); /* second is unused */
       pose=ftell(fo);  /* the end of the object */
       if(pos2 > 0){fseek(fo,pos2,0); outlng(12*newvert);}
       if(pos3 > 0){
         fseek(fo,pos3,0);
         if(Nedge < 65536)outlng(4*newedge);
         else             outlng(8*newedge);
       }
       if(pos4 > 0){
         fseek(fo,pos4,0);
         if(Nface < 65536)outlng(18*newface);
         else             outlng(24*newface);
       }
       if(pos5 > 0){fseek(fo,pos5,0); outlng(pos6-pos5-4);} /* not include count */
       if(posA > 0){fseek(fo,posA,0); outlng(posB-posA-4);}
       if(posC > 0){fseek(fo,posC,0); outlng(posD-posC-4);}

       fseek(fo,pos1,0); outlng(pose-8);
       fclose(fo);
     }
     SKIPIT:
     vp=MainVp; if(vp != NULL)for(j=0;j<Nvert;j++){ /* reset status */
       if(vp->status == INEDITOR)vp->status=SELECTED;
       vp++;
     }
   }
   sp=sp->last;
 }
#endif
}

static void WriteOffsetTextures(point p, unsigned char nmap, long invert){
 long i,temp,k;
 skel *sp;
 long v;
 char n256[256];
 i=0;
 memset(n256,0,256);
 sp=FirstSp; while(sp != NULL){ /* set up sk id */
   sp->id=i; i++;
   sp=sp->next;
 }
 outlng((long)nmap);
 k=0;
 for(i=0;i<nImaps;i++){
   if(iMap[i].in_use < 0)continue;  /* texture not in use          */
   k++;
   outlng((iMap[i].P[0]-p[0])*invert);
   outlng((iMap[i].P[1]-p[1])*invert);
   outlng( iMap[i].P[2]-p[2]);
   outlng((iMap[i].X[0]-p[0])*invert);
   outlng((iMap[i].X[1]-p[1])*invert);
   outlng( iMap[i].X[2]-p[2]);
   outlng((iMap[i].Y[0]-p[0])*invert);
   outlng((iMap[i].Y[1]-p[1])*invert);
   outlng( iMap[i].Y[2]-p[2]);
   temp=iMap[i].Type;    outlng(temp); temp=iMap[i].AnimFF;  outlng(temp);
   temp=iMap[i].AnimDL;  outlng(temp); temp=iMap[i].AnimLF;  outlng(temp);
   temp=iMap[i].TypeR;   outlng(temp); temp=iMap[i].AnimFFr; outlng(temp);
   temp=iMap[i].AnimDLr; outlng(temp); temp=iMap[i].AnimLFr; outlng(temp);
   temp=iMap[i].TypeB;   outlng(temp); temp=iMap[i].AnimFFb; outlng(temp);
   temp=iMap[i].AnimDLb; outlng(temp); temp=iMap[i].AnimLFb; outlng(temp);
   temp=iMap[i].TypeT;   outlng(temp); temp=iMap[i].AnimFFt; outlng(temp);
   temp=iMap[i].AnimDLt; outlng(temp); temp=iMap[i].AnimLFt; outlng(temp);
   if(iMap[i].Sk != NULL){
     temp=((iMap[i].Sk)->id)+1; outlng(temp);/* 1 is first file brush*/
   }
   else{
     temp=0; outlng(temp);
   }
   temp=iMap[i].Map; outlng(temp);   /* 0 = plane 1 = cyl. map  2/3 = moz */
   temp=iMap[i].Angle; outlng(temp); /* angle for cylinder brushes */
   if(iMap[i].Lock &&
      (MainVp+iMap[i].Vlock[0])->id >= 0 &&        /* out a tex map if lock */
      (MainVp+iMap[i].Vlock[1])->id >= 0 &&        /* and if vertices OK    */
      (MainVp+iMap[i].Vlock[2])->id >= 0 ){
     temp=iMap[i].Lock;  outlng(temp);
     v=iMap[i].Vlock[0];
     temp=(MainVp+v)->id; outlng(temp);
     v=iMap[i].Vlock[1];
     temp=(MainVp+v)->id; outlng(temp);
     v=iMap[i].Vlock[2];
     temp=(MainVp+v)->id; outlng(temp);
   }
   else{  /* map is not locked */
     temp=0; outlng(temp); outlng(temp); outlng(temp); outlng(temp);
   }
   fwrite(iMap[i].N,1,32,fo);
   fwrite(&iMap[i].s,1,1,fo);
   fwrite(&iMap[i].r,1,1,fo);
   fwrite(&iMap[i].b,1,1,fo);
   fwrite(&iMap[i].t,1,1,fo);
   fwrite(&iMap[i].Key,1,1,fo);
   outsht(iMap[i].sp);
   outsht(iMap[i].rp);
   outsht(iMap[i].bp);
   outsht(iMap[i].tp);
   fwrite(iMap[i].d_colour,1,3,fo);
   fwrite(iMap[i].s_colour,1,3,fo);
   fwrite(iMap[i].a_colour,1,3,fo);
   fwrite(iMap[i].k_colour,1,3,fo);
   if(iMap[i].s)fwrite(iMap[i].S,1,256,fo); else fwrite(n256,1,256,fo);
   if(iMap[i].r)fwrite(iMap[i].R,1,256,fo); else fwrite(n256,1,256,fo);
   if(iMap[i].b)fwrite(iMap[i].B,1,256,fo); else fwrite(n256,1,256,fo);
   if(iMap[i].t)fwrite(iMap[i].T,1,256,fo); else fwrite(n256,1,256,fo);
 }
 if(k != (long)nmap){
//   inform("Error in writing maps - file corrupt");
 }
}

static void WriteOffsetShadersAxes(point p, long invert){
 long i;
 for(i=1;i<nSha;i++){
   fwrite(iSha[i].N,1,32,fo);
   outlng((iSha[i].P[0]-p[0])*invert);
   outlng((iSha[i].P[1]-p[1])*invert);
   outlng( iSha[i].P[2]-p[2]);
   outlng((iSha[i].X[0]-p[0])*invert);
   outlng((iSha[i].X[1]-p[1])*invert);
   outlng( iSha[i].X[2]-p[2]);
   outlng((iSha[i].Y[0]-p[0])*invert);
   outlng((iSha[i].Y[1]-p[1])*invert);
   outlng( iSha[i].Y[2]-p[2]);
 }
}

static void WriteShaders(short status){
 long i,lp;
 outlng(8);
 for(i=0;i<8;i++){
   outlng(xShaders[i].in_use);
   if(xShaders[i].in_use){
     lp=strlen(xShaders[i].parameters)+1; /* include the trailing zero */
     outlng(128); outlng(lp);
     fwrite(xShaders[i].name,1,128,fo);
     fwrite(xShaders[i].parameters,1,lp,fo);
   }
 }
}


// end of Load/Save code

typedef struct tagSAVESKELETON {
  point xyz,bx[8];
} SAVESKELETON;

static SAVESKELETON *Undo_Skeleton  = NULL;
static long Undo_Vertex = -1;
static long Undo_Edge   = -1;
static long Undo_Face   = -1;
static point  *Undo_Point  = NULL;
static short  Undo_Active  = 0;
static char   *Undo_Status = NULL;

void Save_Undo(short model){
  long sizz,i,j;
  vertex *vp;
  skel *sp;
  char workfile[128],*ud;
  EDIT_ACTION=YES;
  if(!UNDO_ON)return;
  Empty_Undo();
  if(Nvert > 0){
    if((Undo_Status=(char *)X__Malloc(Nvert*sizeof(char))) != NULL){
      vp=MainVp; ud=Undo_Status; for(j=0;j<Nvert;j++){
        *ud++=vp->status;
        vp++;
      }
    }
  }
  if(model == 0){/* need only current numbers */
    Undo_Vertex=Nvert;
    Undo_Edge=Nedge;
    Undo_Face=Nface;
    Undo_Active=1;
  }
  else if(model == 1){/* need to save position info as well */
    Undo_Vertex=Nvert;
    Undo_Edge=Nedge;
    Undo_Face=Nface;
    Undo_Active=2;
    sizz=NvertSelect*sizeof(point);
    if(sizz == 0)return;
    if((Undo_Point=(point *)X__Malloc(sizz)) == NULL){
      Undo_Active=0;
      if(Undo_Status != NULL)X__Free(Undo_Status);
      Undo_Status = NULL;
    }
    else{
      i=0; vp=MainVp; for(j=0;j<Nvert;j++){
        if(vp->status == SELECTED){
          Undo_Point[i][0]=vp->xyz[0];
          Undo_Point[i][1]=vp->xyz[1];
          Undo_Point[i][2]=vp->xyz[2];
          i++;
        }
        vp++;
      }
    }
    if(Nskel > 0){
      if((Undo_Skeleton=(SAVESKELETON *)X__Malloc(Nskel*sizeof(SAVESKELETON)))
          != NULL){
        i=0; sp=FirstSp; while(sp != NULL){
          VECCOPY(sp->xyz,Undo_Skeleton[i].xyz)
          for(j=0;j<8;j++)VECCOPY(sp->bx[j],Undo_Skeleton[i].bx[j])
          i++; sp=sp->next;
        }
      }
    }
  }
  else{  /* model = 2 use  temp device */
    sprintf(workfile,"%sundo_$.$$$",WorkDrive);
    if((fo=fopen(workfile,"wb")) == NULL || Nvert > 65500){
      if(Undo_Status != NULL)X__Free(Undo_Status);
      Undo_Status = NULL;
      SendPrgmQuery(IDQ_NOUNDOSPACE,0);
      return;
    }
    outlng(12*(long)Nvert);
    if(Nedge < 65536)outlng( 4*(long)Nedge);
    else             outlng( 8*(long)Nedge);
    if(Nface < 65536)outlng(18*(long)Nface);
    else             outlng(24*(long)Nface);
    outlng(   (long)Nskel);
    outlng(12*(long)NvertGlue);
    outlng(   (long)nSha);
    outlng(   (long)Nnurbs);
    if(Nvert > 0)WriteVertices(1);
    if(Nedge > 0 && Nedge < 65536)WriteEdges(TRUE);
    else if(Nedge > 0)WriteEdges(FALSE);
    if(Nface > 0){
      if(Nface < 65536)WriteFaces(TRUE);
      else             WriteFaces(FALSE);
    }
    if(Nskel > 1)WriteSkeleton();
    if(NvertGlue > 0)WriteVertexMapping(1);
    if(Nnurbs > 0)WriteNurbs(fo);
    fclose(fo);
    Undo_Active=3;
  }
}

void Restore_Undo(void){
 short id;
 char workfile[128];
 long i,j,nv,ne,nf,ns,ng,nsa,nn;
 vertex *vp;
 skel *sp;
 if(Undo_Active == 0){
   SendPrgmQuery(IDQ_NOUNDO,0);
   return;
 }
 else if(SelectedBrush >= 0 || SelectedShader >= 0){
   SendPrgmQuery(IDQ_UNDONOTREADY,0);
   return;
 }
 LastToolVertex= -1;    /* incase undo in middle of tool */
 if(Undo_Active <= 2){  /* undo for extras or positions  */
   /* get rid of any Bezier curves in the extra vertices */
   if(Nvert > Undo_Vertex){
     vp=(MainVp+Undo_Vertex); for(i=Undo_Vertex;i<Nvert;i++){
       vp->status=SELECTED;
       vp++;
     }
     Delete_Bcurve(); Delete_MapLock();
     /* now get rid of vertices edges and faces */
     vp=(MainVp+Undo_Vertex);for(i=Undo_Vertex;i<Nvert;i++){
       if(vp->gp == 1){ /* if any glue mapping has been added, remove */
         NvertGlue--;
       }
       vp++;
     }
   }
   UpdateVertexHeap(Undo_Vertex); Nvert=Undo_Vertex;
   UpdateEdgeHeap(Undo_Edge);     Nedge=Undo_Edge;
   UpdateFaceHeap(Undo_Face);     Nface=Undo_Face;
   NvertSelect=0;  NvertDeselect=0;
   if(Nvert > 0 && MainVp != NULL){
     vp=MainVp; for(i=0;i<Nvert;i++){
       if(vp->status == SELECTED){
         if(Undo_Point != NULL){
           vp->xyz[0]=Undo_Point[NvertSelect][0];
           vp->xyz[1]=Undo_Point[NvertSelect][1];
           vp->xyz[2]=Undo_Point[NvertSelect][2];
         }
         NvertSelect++;
       }
       else  NvertDeselect++;
       vp++;
     }
   }
   if(Undo_Skeleton != NULL){
     i=0; sp=FirstSp; while(sp != NULL){
       VECCOPY(Undo_Skeleton[i].xyz,sp->xyz)
       for(j=0;j<8;j++)VECCOPY(Undo_Skeleton[i].bx[j],sp->bx[j])
       i++; sp=sp->next;
     }
   }
 }
 else { /* off disk */
   sprintf(workfile,"%sundo_$.$$$",WorkDrive);
   fq=CreateFile(workfile,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
   if(fq == INVALID_HANDLE_VALUE){
     SendPrgmQuery(IDQ_NOUNDOSPACE,0);
     Empty_Undo();
     return;
   }
   getlon(&nv);
   getlon(&ne);
   getlon(&nf);
   getlon(&ns);
   getlon(&ng);
   getlon(&nsa);
   getlon(&nn);
//   if(nv > 0){
     EraseAll(0);  /* erase all vertices faces edges nurbs and skeleton no redraw */
     if(nv > 0)ReadVertices(nv);
     if(ne > 0 && ne < 65536*4)ReadEdges(ne,TRUE);
     else if(ne > 0)           ReadEdges(ne,FALSE);
     if(nf > 0){
       if(nf < 65536*18)ReadFaces(nf,TRUE);
       else             ReadFaces(nf,FALSE);
     }
     if(ns > 1)ReadSkeleton(0L,2);  /* no param necessary */
     if(ng > 0)ReadVertexMapping(ng);
     if(nn > 0)ReadNurbs(0,0,fq);
//   }
   CloseHandle(fq);
   remove(workfile);
 }
 if(Undo_Status != NULL){
   NvertSelect=0;
   vp=MainVp; for(i=0;i<Nvert;i++){
    if((vp->status=Undo_Status[i]) == SELECTED)NvertSelect++;
    vp++;
   }
   NvertDeselect=Nvert-NvertSelect;
 }
 Empty_Undo();
 DrawModel();
 UpdateCounters();
 if(tool > INZOOM)Save_Undo(1);
}

void Empty_Undo(void){
 char workfile[128];
 if(Undo_Status != NULL)X__Free(Undo_Status);
 Undo_Status=NULL;
 if(Undo_Active == 1 || Undo_Active == 2){
   Undo_Vertex = -1;
   Undo_Edge   = -1;
   Undo_Face   = -1;
   if(Undo_Point != NULL)X__Free(Undo_Point);  Undo_Point=NULL;
   if(Undo_Skeleton != NULL)X__Free(Undo_Skeleton); Undo_Skeleton=NULL;
 }
 else if(Undo_Active == 3){
   sprintf(workfile,"%sundo_$.$$$",WorkDrive);
   remove(workfile);
 }
 Undo_Active=0;
}

#include "..\animate\callall.h"

static void LoadRenderer(int anim){
 char modulename[256];
 HWND hWnd,hWndCall;
 int result;
 if((hWnd=FindWindow("OFX:RenderClass",NULL)) != NULL){
   if     (anim == 1)
     PostMessage(hWnd,WM_COMMAND,IDM_CALLBACK_SETUP_PREVIEW,0);
   else if(anim == 2)
     PostMessage(hWnd,WM_COMMAND,IDM_CALLBACK_RUN,(WORD)1);
   else if(anim == 3){
     if(ghwndOpenGLview == NULL)hWndCall=ghwnd_view;
     else                       hWndCall=ghwndOpenGLview;
     PostMessage(hWnd,WM_CALLBACK_MATERIAL,0,(LPARAM)hWndCall);
   }
   else
     SendMessage(hWnd,IDM_CALLBACK_SETUP_MODELLER,0,0);
   if(anim != 3){
     SetForegroundWindow(hWnd);
     BringWindowToTop(hWnd);
     ShowWindow(hWnd,SW_SHOW);
   }
   return;
 }
 else{
   strcpy(modulename,gszHomeDir);strcat(modulename,"render.exe $SAL$");
   result=WinExec(modulename,SW_SHOW);
   if(result < 32){
     if(result == ERROR_FILE_NOT_FOUND ||
        result == ERROR_PATH_NOT_FOUND)SendPrgmQuery(IDQ_NOSTART,3);
   }
   if((hWnd=FindWindow("OFX:RenderClass",NULL)) != NULL){
     if     (anim == 1)
       PostMessage(hWnd,WM_COMMAND,IDM_CALLBACK_SETUP_PREVIEW,0);
     else if(anim == 2)
       PostMessage(hWnd,WM_COMMAND,IDM_CALLBACK_RUN,(WORD)1);
     else if(anim == 3){
       if(ghwndOpenGLview == NULL)hWndCall=ghwnd_view;
       else                       hWndCall=ghwndOpenGLview;
       PostMessage(hWnd,WM_CALLBACK_MATERIAL,0,(LPARAM)hWndCall);
     }
     else
       PostMessage(hWnd,IDM_CALLBACK_SETUP_MODELLER,0,0);
     if(anim != 3){
       SetForegroundWindow(hWnd);
       BringWindowToTop(hWnd);
       ShowWindow(hWnd,SW_SHOW);
     }
   }
   else SendPrgmQuery(IDQ_NOSTART,3);
 }
 return;
}

void Render_View(int light_pos, long how){
 char config_file[64],SCRIPTfilename[256],OBJECTfilename[256];
 long i,narg;
 FILE *fs,*fc;
 HCURSOR lhcurSave;
 double lpx,lpy,lpz,rad,dr,llpx,llpy,llpz,Fi,Theta;
 if(type_of_view){
   Fi = PerspFi; Theta = -PerspTheta;
 }
 else{
   Fi = -PerspFi; Theta = PerspTheta;
 }
 dr=PI/180.0;
 sprintf(OBJECTfilename,"%sW_$.mfx",WorkDrive);
 remove(OBJECTfilename);
 lhcurSave=SetCursor(ghcurWait);
 SaveObject(OBJECTfilename,2);    // 2 for saving only in window-box
 SetCursor(lhcurSave);
 sprintf(SCRIPTfilename,"%sW_$.scr",WorkDrive);
 if((fs=fopen(SCRIPTfilename,"w")) == NULL){
   remove(OBJECTfilename);
   return;
 }
 fprintf(fs,"Renderer-Script\n");
 fprintf(fs,"1 1\n");
 lpx=(double)(TVpointX+TVsizeX/2);
 lpy=(double)(TVpointY+TVsizeY/2);
 lpz=(double)(TVpointZ+TVsizeZ/2);
 rad=TVsizeX*5;
 llpx = lpx-rad*10.0*cos(-(Theta+30)*dr)*sin(-(Fi+45)*dr);
 llpy = lpy-rad*10.0*cos(-(Theta+30)*dr)*cos(-(Fi+45)*dr);
 llpz = lpz-rad*10.0*sin(-(Theta+30)*dr);
 lpx -= rad*cos(-Theta*dr)*sin(-Fi*dr);
 lpy -= rad*cos(-Theta*dr)*cos(-Fi*dr);
 lpz -= rad*sin(-Theta*dr);
 fprintf(fs,"e %.1f %.1f %.1f\n",lpx,lpy,lpz);
 fprintf(fs,"  %.1f %.1f %.1f\n",Fi,0.0,-Theta);
 fprintf(fs,"  %.3f %.3f %.3f\n",3.5,3.5,3.5); /* 3.5 3.0 3.0 */
 fprintf(fs,"  %d %d %.1f\n",0,0,0.0);
 if(light_pos == 0){ /* light over shoulder */
   fprintf(fs,"l 0\n  255 255 255 1.0\n");
   fprintf(fs,"  %.1f %.1f %.1f\n",llpx,llpy,llpz);
   fprintf(fs,"  0.0 0.0 0.0\n 0 0 0.0\n");
   fprintf(fs,"  0 1.0 0.0 0.0\n");
 }
 fprintf(fs,"s %ld %ld %ld\n",
         (long)Preferences.Vcolour[0],
         (long)Preferences.Vcolour[1],
         (long)Preferences.Vcolour[2]);
 fprintf(fs,"  %ld %ld %ld\n",
         (long)Preferences.Vcolour[0],
         (long)Preferences.Vcolour[1],
         (long)Preferences.Vcolour[2]);
 fprintf(fs," 0.0 0.1 0.08\n");
 fprintf(fs," 0 0 0\n");
 fprintf(fs," 0 0 0\n");
 fprintf(fs," 0\n");
 fprintf(fs,"o 1\n");
 fprintf(fs,"%s\n",OBJECTfilename);
 fprintf(fs,"0\n");
 fprintf(fs,"  %.1f %.1f %.1f\n",0.0,0.0,0.0);
 fprintf(fs,"  %.1f %.1f %.1f\n",0.0,0.0,0.0);
 fprintf(fs,"  %.3f %.3f %.3f\n",1.0,1.0,1.0);
 fprintf(fs,"  %d %d %.1f %d %.3f\n",0,0,0.0,0,1.0);
 fprintf(fs,"  1 1 1\n");
 fprintf(fs,"f\n");
 fprintf(fs,"q\n");
 fclose(fs);
 sprintf(SCRIPTfilename,"%sW_$",WorkDrive);
 if     (how == 1)LoadRenderer(2);   // Quick render with prev params
 else if(how == 0){                  // Setup renderer
   LoadRenderer(0);
   EnableMenuItem(GetMenu(ghwnd_main),IDM_VIEW_RENDERNOW,MF_ENABLED);
 }
 else if(how == 2)LoadRenderer(3);   // render into window
 return;
}

BOOL LoadHeightField(HWND hwnd){
 BOOL status=FALSE;
 HMODULE hLib;
 FARPROC fpFun;
 char lname[256];
 X__STRUCTURE *Evs;
 BOOL (*fpFun1)(HWND, char *, X__STRUCTURE *);
 if(SelectSfxFileName(0,gszGIFfile,gszGIFdir,IDX_MISC_HEIGHTFIELD,
   "(*.GIF)|*.gif|",ghwnd_main) == FALSE)return FALSE;
 EDIT_ACTION=YES;
 Evs=&EviStructure;
 strcpy(lname,gszHomeDir); strcat(lname,"height.dll");
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_HeightField")) != NULL){
     fpFun1 = (void *)fpFun;
     status=(*fpFun1)(hwnd,gszGIFfile,Evs);
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 return status;
}

BOOL TraceImageOutline(HWND hwnd){
 BOOL status=FALSE;
 static HMODULE hLibTrace=NULL;
 FARPROC fpFun;
 char lname[256];
 X__STRUCTURE *Evs;
 BOOL (*fpFun1)(HWND, HWND, char *, X__STRUCTURE *);
 if(hLibTrace != NULL){
   return FALSE;
 }
 if(SelectSfxFileName(0,gszIMGfile,gszIMGdir,IDX_MISC_TRACE,
   "(*.GIF)|*.gif|(*.TGA|*.tga|(*.CEL)|*.cel|",hwnd) == FALSE)return FALSE;
 EDIT_ACTION=YES;
 Evs=&EviStructure;
 strcpy(lname,gszHomeDir); strcat(lname,"trace.dll");
 if((hLibTrace=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLibTrace,"_TraceOutline")) != NULL){
     fpFun1 = (void *)fpFun;
     status=(*fpFun1)(hwnd,ghwnd_info,gszIMGfile,Evs);
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLibTrace);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 hLibTrace=NULL;
 return status;
}

void LoadAApoly(int aa_or_not_aa, char *filename){
 FILE *fpp;
 long vp,vpf;
 short  fail,points,x,y,z,i;
 long xx,yy,zz;
 unsigned char closed,magic;
 long   packer;
 if((fpp=fopen(filename,"rb")) == NULL){
   SendPrgmQuery(IDQ_FILEOPENFAIL,0);
   return;
 }
 if(fread(&points,2,1,fpp) != 1)goto NOREAD;
 if(fread(&packer,4,1,fpp) != 1)goto NOREAD;
 if(fread(&closed,1,1,fpp) != 1)goto NOREAD;
 if(aa_or_not_aa){
   if(fread(&magic,1,1 ,fpp)  != 1 || magic != 0x99)goto NOREAD;
 }
 else{
   if(fread(&magic,1,1 ,fpp)  != 1 || magic != 0x88)goto NOREAD;
 }
 if(points == 0)goto NOREAD;
 if(!UpdateVertexHeap(Nvert+(long)points))goto NOREAD;
 if(!UpdateEdgeHeap(Nedge+(long)points+1))goto NOREAD;
 for(i=0; i< points; i++){
   if(aa_or_not_aa){
     if(fread(&x,2,1,fpp) != 1)goto NOREAD;
     if(fread(&y,2,1,fpp) != 1)goto NOREAD;
     if(fread(&z,2,1,fpp) != 1)goto NOREAD;
     vp=Nvert-1;
     CreateVertex();
     (MainVp+Nvert-1)->xyz[0] = ((long)x-160)*200;
     (MainVp+Nvert-1)->xyz[1] = 0;
     (MainVp+Nvert-1)->xyz[2] = ((long)y-100)*(-200);
   }
   else{
     if(fread(&xx,4,1,fpp) != 1)goto NOREAD;
     if(fread(&yy,4,1,fpp) != 1)goto NOREAD;
     if(fread(&zz,4,1,fpp) != 1)goto NOREAD;
     vp=Nvert-1;
     CreateVertex();
     (MainVp+Nvert-1)->xyz[0] = xx;
     (MainVp+Nvert-1)->xyz[1] = yy;
     (MainVp+Nvert-1)->xyz[2] = zz;
   }
   if(i > 0 && vp >= 0){
     CreateEdge(Nvert-1,vp);
   }
   else if(i == 0)vpf=Nvert-1;
 }
 if(closed == 1 && vpf >= 0)CreateEdge(vpf,Nvert-1);
 goto READOK;
 NOREAD:
 if(aa_or_not_aa)SendPrgmQuery(IDQ_NOTAAPROPATH,0);
 else            SendPrgmQuery(IDQ_NOTPATHFILE,0);
 READOK:
 fclose(fpp);
 UpdateVertexHeap(Nvert);
 UpdateEdgeHeap(Nedge);
 return;
}

void LoadMDIviewer(void){
 BOOL n=FALSE;
 int result;
 char modulename[255];
 HWND hWnd;
 if((hWnd=FindWindow("OFX:Viewer",NULL)) == NULL){
   strcpy(modulename,gszHomeDir);strcat(modulename,"mdiview.exe");
   n=TRUE;
   result=WinExec(modulename,SW_SHOW);
   if(result < 32){
     if(result == ERROR_FILE_NOT_FOUND ||
        result == ERROR_PATH_NOT_FOUND)SendPrgmQuery(IDQ_NOSTART,3);
     return;
   }
   hWnd=FindWindow("OFX:Viewer",NULL);
 }
 if(IsWindow(hWnd)){
   SetForegroundWindow(hWnd);
   BringWindowToTop(hWnd);
   ShowWindow(hWnd,SW_SHOW);
   if(n)PostMessage(hWnd,WM_COMMAND,1004,0);
 }
 return;
}

//void LoadExternalOpenGL(void){
// BOOL n=FALSE;
// int result;
// char modulename[255];
// HWND hWnd;
// if((hWnd=FindWindow("OFX:SfxView",NULL)) == NULL){
//   strcpy(modulename,gszHomeDir);strcat(modulename,"sfxview.exe");
//   n=TRUE;
//   result=WinExec(modulename,SW_SHOW);
//   if(result < 32){
//     if(result == ERROR_FILE_NOT_FOUND ||
//        result == ERROR_PATH_NOT_FOUND)SendPrgmQuery(IDQ_NOSTART,3);
//     return;
//   }
//   hWnd=FindWindow("OFX:SfxView",NULL);
// }
// if(IsWindow(hWnd)){
//   SetForegroundWindow(hWnd);
//   BringWindowToTop(hWnd);
//   ShowWindow(hWnd,SW_SHOW);
// }
// return;
//}

void SavePath(void){
 vertex *Va;
 long *va;
 int a,id,Np;
 char closed;
 unsigned char magic;
 FILE *fpp;
 long Npath,packer=0,i;
 if((va=GetPathList(&Npath)) != NULL){
   a=MessageBox (ghwnd_main,"Save as Closed Path","Save Path",
                 MB_YESNOCANCEL | MB_ICONQUESTION | MB_TASKMODAL);
   if    (a == IDYES)id=1;
   else if(a == IDNO)id=0;
   else id = -1;
   if(id == 0 || id == 1){
     if(SelectSfxFileName(1,gszPTHfile,gszPTHdir,IDX_MISC_WRITEPATH,
        "(*.PTH)|*.pth|",ghwnd_main) == TRUE){
       AppendFileExtension(gszPTHfile,".pth");
       closed = (char)id;
       if((fpp=fopen(gszPTHfile,"wb")) == NULL){
         SendPrgmQuery(IDQ_FILEOPENFAIL,0);
         goto XXIT;
       }
       Np=(short)Npath;
       if(fwrite(&Np,2,1,fpp) != 1)goto NOWRITE;
       if(fwrite(&packer,4,1,fpp) != 1)goto NOWRITE;
       if(fwrite(&closed,1,1,fpp) != 1)goto NOWRITE;
       magic = 0x88;
       if(fwrite(&magic,1,1 ,fpp)  != 1)goto NOWRITE;
       if(Npath == 0)goto NOWRITE;
       for(i=0; i< Npath; i++){
         Va=(MainVp+va[i]);
         if(fwrite(&(Va->xyz[0]),4,1,fpp) != 1)goto NOWRITE;
         if(fwrite(&(Va->xyz[1]),4,1,fpp) != 1)goto NOWRITE;
         if(fwrite(&(Va->xyz[2]),4,1,fpp) != 1)goto NOWRITE;
       }
       goto READOK;
       NOWRITE:
       SendPrgmQuery(IDQ_SAVINGERROR,0);
       READOK:
       fclose(fpp);
     }
   }
   XXIT:
   X__Free(va);
   DrawVerticesOnly(NULL);
   UpdateCounters();
 }
 else{
   SendPrgmQuery(IDQ_NOPATHSTART,0);
 }
}

void ShaderPreview(void){
  LoadRenderer(1);
}

BOOL ExecuteExternalAction(int id,HWND hwnd){
 static BOOL inuse=FALSE;
 BOOL status=FALSE;
 HMODULE hLib;
 FARPROC fpFun;
 X__STRUCTURE *Evs;
 BOOL (*fpFun1)(HWND, HWND, X__STRUCTURE *);
 int l;
 char fname[255];
 if(tool > INZOOM)DisengageDesignerTool();
 if(sktool || SelectedBrush >= 0 || SelectedShader >= 0){
   SendPrgmQuery(IDQ_TOOLNOTFINISHED,0);
   return FALSE;
 }
 if(id >= nActionsList)return FALSE;
 Empty_Undo();
 EDIT_ACTION=YES;
 Save_Undo(2);
 strcpy(fname,ActionsList[id]);
 l=strlen(fname);
 fname[l-1]='l';
 fname[l-2]='l';
 fname[l-3]='d';
 if(inuse)return FALSE;
 inuse=TRUE;
 Evs=&EviStructure;
 if((hLib=LoadLibrary(fname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_Xmodeler")) != NULL){
     fpFun1 = (void *)fpFun;
     status=(*fpFun1)(hwnd,ghwnd_info,Evs);
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 inuse=FALSE;
 return status;
 return TRUE;
}

BOOL ExecuteExternalModeler(HWND hwnd){
 long l;
 static BOOL inuse=FALSE;
 BOOL status=FALSE;
 HMODULE hLib;
 FARPROC fpFun;
 X__STRUCTURE *Evs;
 BOOL (*fpFun1)(HWND, HWND, X__STRUCTURE *);
 if(inuse)return FALSE;
 l=strlen(gszDLMfile);
 strcpy((gszDLMfile+l-3),"dfx");
 if(SelectSfxFileName(0,gszDLMfile,gszDLMdir,IDX_MISC_EXTMODELER,
   "(*.dfx)|*.dfx|",ghwnd_main) == FALSE)return FALSE;
 l=strlen(gszDLMfile);
 strcpy((gszDLMfile+l-3),"dll");
 inuse=TRUE;
 EDIT_ACTION=YES;
 Evs=&EviStructure;
 if((hLib=LoadLibrary(gszDLMfile)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_Xmodeler")) != NULL){
     fpFun1 = (void *)fpFun;
     status=(*fpFun1)(hwnd,ghwnd_info,Evs);
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 inuse=FALSE;
 return status;
}

BOOL ExecuteBezierModeler(HWND hwnd){
 char lname[256];
 static BOOL inuse=FALSE;
 BOOL status=FALSE;
 HMODULE hLib;
 FARPROC fpFun;
 X__STRUCTURE *Evs;
 BOOL (*fpFun1)(HWND, HWND, X__STRUCTURE *);
 if(inuse)return FALSE;
 inuse=TRUE;
 EDIT_ACTION=YES;
 Evs=&EviStructure;
 strcpy(lname,gszHomeDir); strcat(lname,"bezier.dll");
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_BezierModeler")) != NULL){
     fpFun1 = (void *)fpFun;
     status=(*fpFun1)(hwnd,ghwnd_info,Evs);
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 inuse=FALSE;
 return status;
}

BOOL PrintImageFile(HWND hwnd){
 BOOL status=FALSE;
 HMODULE hLib;
 FARPROC fpFun;
 char lname[256];
 BOOL (*fpFun1)(HWND, char *);
 if(SelectSfxFileName(0,gszIMGfile,gszIMGdir,IDX_MISC_IMAGEFILE,
   "(*.GIF)|*.gif|(*.TGA)|*.tga|(*.CEL)|*.cel|(*.BMP)|*.bmp|",ghwnd_main) == FALSE)
   return FALSE;
 strcpy(lname,gszHomeDir); strcat(lname,"eprint.dll");
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_PrintBitmap")) != NULL){
     fpFun1 = (void *)fpFun;
     EnableToolPannels(ALL_PANNELS,FALSE);
     status=(*fpFun1)(hwnd,gszIMGfile);
     EnableToolPannels(ALL_PANNELS,TRUE);
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 return status;
}


void ConvertTDDmodel(HWND hwnd){
 int result;
 char modulename[512];
 DWORD version;
 if(SelectSfxFileName(0,gszTDDfile,gszTDDdir,IDX_MISC_IMPORTTDD,
   "(*.TDD)|*.tdd|",ghwnd_main) == FALSE)return;
 if(SelectSfxFileName(1,gszSCFfile,gszSCFdir,IDX_MISC_OUTPUTSCE,
   "(*.MFX)|*.mfx|",ghwnd_main) == FALSE)return;
 AppendFileExtension(gszSCFfile,".mfx");
 version=GetVersion();
 strcpy(modulename,gszHomeDir);
#ifdef __X__NT__
  strcat(modulename,"TDDwNT.exe");
#else
 if((version & 0x80000000) == 0x80000000 && LOBYTE(LOWORD(version)) < 4) /* Win32s */
   strcat(modulename,"TDDw31.exe");
 else
   strcat(modulename,"TDDwNT.exe");
#endif
 strcat(modulename," ");
 strcat(modulename,gszTDDfile);
 strcat(modulename," ");
 strcat(modulename,gszSCFfile);
 result=WinExec(modulename,SW_SHOW);
 if(result < 32){
 if(result == ERROR_FILE_NOT_FOUND ||
    result == ERROR_PATH_NOT_FOUND)SendPrgmQuery(IDQ_NOSTART,3);
 }
 return;
}

void ConvertSCNmodel(HWND hwnd){
 int result;
 char modulename[512];
 DWORD version;
 if(SelectSfxFileName(0,gszSCNfile,gszSCNdir,IDX_MISC_IMPORTSCN,
   "(*.SCN)|*.scn|",ghwnd_main) == FALSE)return;
 if(SelectSfxFileName(1,gszSCFfile,gszSCFdir,IDX_MISC_OUTPUTSCE,
   "(*.MFX)|*.mfx|",ghwnd_main) == FALSE)return;
 AppendFileExtension(gszSCFfile,".mfx");
 version=GetVersion();
 strcpy(modulename,gszHomeDir);
#ifdef __X__NT__
  strcat(modulename,"SCNwNT.exe");
#else
 if((version & 0x80000000) == 0x80000000 && LOBYTE(LOWORD(version)) < 4) /* Win32s */
   strcat(modulename,"SCNw31.exe");
 else
   strcat(modulename,"SCNwNT.exe");
#endif
 strcat(modulename," ");
 strcat(modulename,gszSCNfile);
 strcat(modulename," ");
 strcat(modulename,gszSCFfile);
 result=WinExec(modulename,SW_SHOW);
 if(result < 32){
 if(result == ERROR_FILE_NOT_FOUND ||
    result == ERROR_PATH_NOT_FOUND)SendPrgmQuery(IDQ_NOSTART,3);
 }
 return;
}

// include the loadviewer code - this is used by the animator and so
// is in a separate file

#include "loadview.c"

// invoke the external plugins through their dialog boxes

long LoadXmodel(char *name, char *library){
 char lname[255],text[128];
 long rv=0;
 HMODULE hLib=NULL;
 FARPROC fpFun;
 X__STRUCTURE *Evs;
 long (*fpFun1)(HWND, char *, char *, X__STRUCTURE *);
 Evs=&EviStructure;
 strcpy(lname,gszHomeDir); strcat(lname,library);
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_Convert")) != NULL){
     fpFun1 = (void *)fpFun;
     sprintf(lname,"%sW_$.mfx",WorkDrive);
     rv=(*fpFun1)(ghwnd_main,name,lname,Evs);
   }
   else{
     SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   }
   FreeLibrary(hLib); hLib=NULL;
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 if(rv > 0)LoadObject(lname,0);
 remove(lname);
 return rv;
}

void Export3DSmodel(HWND hwnd){
 if(SelectSfxFileName(1,gsz3DSfile,gsz3DSdir,IDX_MISC_EXPORT3DS,
   "(*.3DS)|*.3ds|",ghwnd_main) == FALSE)return;
 AppendFileExtension(gsz3DSfile,".3ds");
 ExportModel(gsz3DSfile,"3dsout.dll");
 return;
}

void ExportDXFmodel(HWND hwnd){
 if(SelectSfxFileName(1,gszDXFfile,gszDXFdir,IDX_MISC_EXPORTDXF,
   "(*.DXF)|*.dxf|",ghwnd_main) == FALSE)return;
 AppendFileExtension(gszDXFfile,".dxf");
 ExportModel(gszDXFfile,"dxfout.dll");
 return;
}

long ExportModel(char *name, char *library){
 char lname[255];
 long rv=0;
 HMODULE hLib=NULL;
 FARPROC fpFun;
 X__STRUCTURE *Evs;
 long (*fpFun1)(HWND, char *, X__STRUCTURE *);
 Evs=&EviStructure;
 strcpy(lname,gszHomeDir); strcat(lname,library);
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_Export")) != NULL){
     fpFun1 = (void *)fpFun;
     rv=(*fpFun1)(ghwnd_main,name,Evs);
   }
   else{
     SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   }
   FreeLibrary(hLib); hLib=NULL;
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 return rv;
}

BOOL LoadInternalAction(char *library){
 char lname[128];
 static BOOL inuse=FALSE;
 BOOL status=FALSE;
 HMODULE hLib;
 FARPROC fpFun;
 X__STRUCTURE *Evs;
 BOOL (*fpFun1)(HWND, HWND, X__STRUCTURE *);
 if(tool > INZOOM)DisengageDesignerTool();
 if(sktool || SelectedBrush >= 0 || SelectedShader >= 0){
   SendPrgmQuery(IDQ_TOOLNOTFINISHED,0);
   return FALSE;
 }
 if(inuse)return FALSE;
 inuse=TRUE;
 Empty_Undo();
 EDIT_ACTION=YES;
 Save_Undo(2);
 Evs=&EviStructure;
 strcpy(lname,gszHomeDir); strcat(lname,library);
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_Xmodeler")) != NULL){
     fpFun1 = (void *)fpFun;
     status=(*fpFun1)(ghwnd_main,ghwnd_info,Evs);
     DrawModel();
     UpdateCounters();
     UpdateSkTreeView();
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 inuse=FALSE;
 return TRUE;
}

