/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file MAPPING.C */

// This file handles MAPPING 

#define MODULE_MAPPING 1

#include "design.h"
#include "..\animate\spin_dll.h"

static void DisplayImageFile(char *pathname);
static int E_MapToPlaneBrush(vector vi,
                            vector p0, vector p1, vector p2,
                            double *a, double *b);
static double Distance3D(point a, long *b);

#define BRUSHBIT   0x40
#define DIS2(a) ((double)(a[0]-NpointerX)*(double)(a[0]-NpointerX)+  \
                 (double)(a[1]-NpointerY)*(double)(a[1]-NpointerY)+  \
                 (double)(a[2]-NpointerZ)*(double)(a[2]-NpointerZ))

static char *GetFlicFromDisk(int id, char *p, HWND parent, int type);
static char *GetMapFromDisk(char *p, HWND parent);
static char *GetPresetMap(char *p, HWND parent);
static void EditImageFile(char *pathname);

static int IdentifiedBrushPoint=-1,IdentifiedShaderPoint=-1;

static int LastUsedMapID=0;

// Mapping Dialog

static HBITMAP MapBtnBitmaps[8]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

static void UpdateTypeMappingSection(HWND hdlg,int m){
 int active;
 char str[32];
 if(iMap[m].Map == MAP_BY_VERTEX)active=FALSE; else active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_MOVE),active);

 if(nImaps >= NMAPSMAX)EnableWindow(GetDlgItem(hdlg,DLG_MAPS_NEW),FALSE);
 if(iMap[m].Map == PLANE || iMap[m].Map == PLANE_MOZIAC)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_P,BM_SETCHECK,active,0);
 if(iMap[m].Map == CYLINDER || iMap[m].Map == CYLINDER_MOZIAC)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_C,BM_SETCHECK,active,0);
 if(iMap[m].Map == MAP_SPHERICAL)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_S,BM_SETCHECK,active,0);
 if(iMap[m].Map == MAP_BY_VERTEX)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_G,BM_SETCHECK,active,0);

 // PM now reps plain tile CM reps Mozaic tile
 if(iMap[m].Map == PLANE_MOZIAC || iMap[m].Map == CYLINDER_MOZIAC)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_CM,BM_SETCHECK,active,0);
 if(iMap[m].Map == PLANE || iMap[m].Map == CYLINDER)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_PM,BM_SETCHECK,active,0);
 if(iMap[m].Map == PLANE || iMap[m].Map == CYLINDER ||
    iMap[m].Map == PLANE_MOZIAC || iMap[m].Map == CYLINDER_MOZIAC)
    active=TRUE; else active=FALSE;
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_CM),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_PM),active);

 SendDlgItemMessage(hdlg,DLG_MAPS_C_SCROLL,TBM_SETPOS,TRUE,iMap[m].Angle);
 sprintf(str,"%ld",iMap[m].Angle);
 SendDlgItemMessage(hdlg,DLG_MAPS_C_EDIT,WM_SETTEXT,0,(LPARAM)str);
 if(iMap[m].Map == CYLINDER ||
    iMap[m].Map == CYLINDER_MOZIAC)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_MAPS_C_SCROLL),active);
 ShowWindow(GetDlgItem(hdlg,DLG_MAPS_C_EDIT),active);

 if(iMap[m].Key == 0)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_KEY_INDEX,BM_SETCHECK,active,0);
 if(iMap[m].Key == 1)active=TRUE; else active=FALSE;
 SendDlgItemMessage(hdlg,DLG_MAPS_KEY_CHOSEN,BM_SETCHECK,active,0);
}

static void UpdateSurfaceMappingSection(HWND hdlg,int m){
 BOOL active;
 char str[32];
 if(iMap[m].S == NULL){
   active=FALSE;
   SendDlgItemMessage(hdlg,DLG_MAPS_S_NAME,WM_SETTEXT,0,(LPARAM)"<No image>");
 }
 else{
   active=TRUE;
   SendDlgItemMessage(hdlg,DLG_MAPS_S_NAME,WM_SETTEXT,0,(LPARAM)iMap[m].S);
 }
 SendDlgItemMessage(hdlg,DLG_MAPS_S_ACTIVE,BM_SETCHECK,active,0);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_S_DROP),active);
// EnableWindow(GetDlgItem(hdlg,DLG_MAPS_S_ANIMATED),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_S_SCROLL),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_S_VIEW),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_S_MODIFY),active);
 if(iMap[m].Type == ANIMBRUSH)active=TRUE; else active=FALSE;
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_S_PRESET),!active);
// SendDlgItemMessage(hdlg,DLG_MAPS_S_STILL,BM_SETCHECK,!active,0);
 SendDlgItemMessage(hdlg,DLG_MAPS_S_ANIMATED,BM_SETCHECK,active,0);
 SendDlgItemMessage(hdlg,DLG_MAPS_S_SCROLL,TBM_SETPOS,TRUE,iMap[m].sp);
 sprintf(str,"%ld",iMap[m].sp);
 SendDlgItemMessage(hdlg,DLG_MAPS_S_EDIT,WM_SETTEXT,0,(LPARAM)str);
 return;
}

static void UpdateReflectionMappingSection(HWND hdlg,int m){
 BOOL active;
 char str[32];
 if(iMap[m].R == NULL){
   active=FALSE;
   SendDlgItemMessage(hdlg,DLG_MAPS_R_NAME,WM_SETTEXT,0,(LPARAM)"<No image>");
 }
 else{
   active=TRUE;
   SendDlgItemMessage(hdlg,DLG_MAPS_R_NAME,WM_SETTEXT,0,(LPARAM)iMap[m].R);
 }
 SendDlgItemMessage(hdlg,DLG_MAPS_R_ACTIVE,BM_SETCHECK,active,0);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_R_VIEW),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_R_MODIFY),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_R_DROP),active);
// EnableWindow(GetDlgItem(hdlg,DLG_MAPS_R_ANIMATED),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_R_SCROLL),active);
 if(iMap[m].TypeR == ANIMBRUSH)active=TRUE; else active=FALSE;
// SendDlgItemMessage(hdlg,DLG_MAPS_R_STILL,BM_SETCHECK,!active,0);
 SendDlgItemMessage(hdlg,DLG_MAPS_R_ANIMATED,BM_SETCHECK,active,0);
 SendDlgItemMessage(hdlg,DLG_MAPS_R_SCROLL,TBM_SETPOS,TRUE,iMap[m].rp);
 sprintf(str,"%ld",iMap[m].rp);
 SendDlgItemMessage(hdlg,DLG_MAPS_R_EDIT,WM_SETTEXT,0,(LPARAM)str);
 return;
}

static void UpdateBumpMappingSection(HWND hdlg,int m){
 BOOL active;
 char str[32];
 if(iMap[m].B == NULL){
   active=FALSE;
   SendDlgItemMessage(hdlg,DLG_MAPS_B_NAME,WM_SETTEXT,0,(LPARAM)"<No image>");
 }
 else{
   active=TRUE;
   SendDlgItemMessage(hdlg,DLG_MAPS_B_NAME,WM_SETTEXT,0,(LPARAM)iMap[m].B);
 }
 SendDlgItemMessage(hdlg,DLG_MAPS_B_ACTIVE,BM_SETCHECK,active,0);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_B_VIEW),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_B_MODIFY),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_B_DROP),active);
// EnableWindow(GetDlgItem(hdlg,DLG_MAPS_B_ANIMATED),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_B_SCROLL),active);
 if(iMap[m].TypeB == ANIMBRUSH)active=TRUE; else active=FALSE;
// SendDlgItemMessage(hdlg,DLG_MAPS_B_STILL,BM_SETCHECK,!active,0);
 SendDlgItemMessage(hdlg,DLG_MAPS_B_ANIMATED,BM_SETCHECK,active,0);
 SendDlgItemMessage(hdlg,DLG_MAPS_B_SCROLL,TBM_SETPOS,TRUE,iMap[m].bp);
 sprintf(str,"%ld",iMap[m].bp);
 SendDlgItemMessage(hdlg,DLG_MAPS_B_EDIT,WM_SETTEXT,0,(LPARAM)str);
 return;
}

static void UpdateTransMappingSection(HWND hdlg,int m){
 BOOL active;
 char str[32];
 if(iMap[m].T == NULL){
   active=FALSE;
   SendDlgItemMessage(hdlg,DLG_MAPS_T_NAME,WM_SETTEXT,0,(LPARAM)"<No image>");
 }
 else{
   active=TRUE;
   SendDlgItemMessage(hdlg,DLG_MAPS_T_NAME,WM_SETTEXT,0,(LPARAM)iMap[m].T);
 }
 SendDlgItemMessage(hdlg,DLG_MAPS_T_ACTIVE,BM_SETCHECK,active,0);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_T_VIEW),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_T_MODIFY),active);
 EnableWindow(GetDlgItem(hdlg,DLG_MAPS_T_DROP),active);
// EnableWindow(GetDlgItem(hdlg,DLG_MAPS_T_ANIMATED),active);
// EnableWindow(GetDlgItem(hdlg,DLG_MAPS_T_SCROLL),active);
 if(iMap[m].TypeT == ANIMBRUSH)active=TRUE; else active=FALSE;
// SendDlgItemMessage(hdlg,DLG_MAPS_T_STILL,BM_SETCHECK,!active,0);
 SendDlgItemMessage(hdlg,DLG_MAPS_T_ANIMATED,BM_SETCHECK,active,0);
// SetScrollPos(GetDlgItem(hdlg,DLG_MAPS_T_SCROLL),SB_CTL,iMap[m].tp,TRUE);
// sprintf(str,"%ld",iMap[m].tp);
// SendDlgItemMessage(hdlg,DLG_MAPS_T_EDIT,WM_SETTEXT,0,(LPARAM)str);
 return;
}

static UpdateMappingDialog(HWND hdlg,int mapid){
 UpdateTypeMappingSection(hdlg,mapid);
 UpdateSurfaceMappingSection(hdlg,mapid);
 UpdateReflectionMappingSection(hdlg,mapid);
 UpdateBumpMappingSection(hdlg,mapid);
 UpdateTransMappingSection(hdlg,mapid);
}

static BOOL CALLBACK MappingDlgProc(HWND hdlg,UINT msg,
                                    WPARAM wparam,LPARAM lparam){
 static MapID;
 HWND hctl;
 BOOL err;
 int i,iMin,iMax;
 LPDRAWITEMSTRUCT lpdis;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     EDIT_ACTION=YES;
     EnableWindow(ghwnd_main,TRUE);
     MapID=(int)lparam;
     MapBtnBitmaps[0]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAPS3));
     MapBtnBitmaps[1]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAPS1));
     MapBtnBitmaps[2]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAPS2));
     MapBtnBitmaps[3]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAPS4));
     MapBtnBitmaps[4]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAPS6));
     MapBtnBitmaps[5]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAPS5));
     MapBtnBitmaps[6]=(HBITMAP)LoadImage(ghinst_main,"DELETEX",IMAGE_ICON,
                               0,0,LR_DEFAULTCOLOR); // ICON
     MapBtnBitmaps[7]=(HBITMAP)LoadImage(ghinst_main,"FOLDERS",IMAGE_ICON,
                               0,0,LR_DEFAULTCOLOR); // ICON
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_S_DROP),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[6]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_R_DROP),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[6]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_B_DROP),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[6]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_T_DROP),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[6]);

     SendMessage(GetDlgItem(hdlg,DLG_MAPS_S_SET),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[7]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_R_SET),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[7]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_B_SET),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[7]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_T_SET),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapBtnBitmaps[7]);

     SendMessage(GetDlgItem(hdlg,DLG_MAPS_P),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)MapBtnBitmaps[0]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_C),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)MapBtnBitmaps[1]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_S),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)MapBtnBitmaps[2]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_G),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)MapBtnBitmaps[3]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_PM),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)MapBtnBitmaps[4]);
     SendMessage(GetDlgItem(hdlg,DLG_MAPS_CM),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)MapBtnBitmaps[5]);
//     SetWindowLong(hdlg,GWL_USERDATA,(long)MapID);
     SendDlgItemMessage(hdlg,DLG_MAPS_C_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,360));
     SendDlgItemMessage(hdlg,DLG_MAPS_S_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,100));
     SendDlgItemMessage(hdlg,DLG_MAPS_R_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,100));
     SendDlgItemMessage(hdlg,DLG_MAPS_B_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,100));
//     SendDlgItemMessage(hdlg,DLG_MAPS_T_SCROLL,TBM_SETRANGE,
//                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,100));
     for(i=0;i<100;i+=10){
       SendDlgItemMessage(hdlg,DLG_MAPS_S_SCROLL,TBM_SETTIC,0,(LPARAM)i);
       SendDlgItemMessage(hdlg,DLG_MAPS_R_SCROLL,TBM_SETTIC,0,(LPARAM)i);
       SendDlgItemMessage(hdlg,DLG_MAPS_B_SCROLL,TBM_SETTIC,0,(LPARAM)i);
//       SendDlgItemMessage(hdlg,DLG_MAPS_T_SCROLL,TBM_SETTIC,0,(LPARAM)i);
     }
     for(i=0;i<360;i+=90)
       SendDlgItemMessage(hdlg,DLG_MAPS_C_SCROLL,TBM_SETTIC,0,(LPARAM)i);
     for(i=0;i<nImaps;i++){
       SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_ADDSTRING,0,
                               (LPARAM)iMap[i].N);
     }
     SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_SELECTSTRING,0,
                             (LPARAM)iMap[MapID].N);
     SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_LIMITTEXT,30,0);
     UpdateMappingDialog(hdlg,MapID);
     CentreDialogOnScreen(hdlg);
     return (TRUE);
   case WM_DESTROY:
     for(i=0;i<6;i++){  // the 7th and 8th are ICONs
       if(MapBtnBitmaps[i] != NULL){
         if(!DeleteObject(MapBtnBitmaps[i]))MessageBox(NULL,"Bad Delete",NULL,MB_OK);
         MapBtnBitmaps[i]=NULL;
       }
     }
     break;
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_MAPS_KEY_COLOUR){
         r=iMap[MapID].k_colour[0];
         g=iMap[MapID].k_colour[1];
         b=iMap[MapID].k_colour[2];
         if(lpdis->itemState & ODS_SELECTED)
           InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_HSCROLL:
     if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_MAPS_C_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wparam)) {
         case TB_PAGEDOWN: iMap[MapID].Angle = min(iMax,iMap[MapID].Angle+45); break;
         case TB_LINEDOWN: iMap[MapID].Angle = min(iMax,iMap[MapID].Angle+1); break;
         case TB_PAGEUP:   iMap[MapID].Angle = max(iMin,iMap[MapID].Angle-45); break;
         case TB_LINEUP:   iMap[MapID].Angle = max(iMin,iMap[MapID].Angle-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           iMap[MapID].Angle=HIWORD(wparam);  break;
         default: break;
       }
       UpdateTypeMappingSection(hdlg,MapID);
     }
     else if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_MAPS_S_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wparam)) {
         case TB_PAGEDOWN: iMap[MapID].sp = min(iMax,iMap[MapID].sp+20); break;
         case TB_LINEDOWN: iMap[MapID].sp = min(iMax,iMap[MapID].sp+1); break;
         case TB_PAGEUP:   iMap[MapID].sp = max(iMin,iMap[MapID].sp-20); break;
         case TB_LINEUP:   iMap[MapID].sp = max(iMin,iMap[MapID].sp-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           iMap[MapID].sp=HIWORD(wparam);  break;
         default: break;
       }
       UpdateSurfaceMappingSection(hdlg,MapID);
     }
     else if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_MAPS_R_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wparam)) {
         case TB_PAGEDOWN: iMap[MapID].rp = min(iMax,iMap[MapID].rp+20); break;
         case TB_LINEDOWN: iMap[MapID].rp = min(iMax,iMap[MapID].rp+1); break;
         case TB_PAGEUP:   iMap[MapID].rp = max(iMin,iMap[MapID].rp-20); break;
         case TB_LINEUP:   iMap[MapID].rp = max(iMin,iMap[MapID].rp-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           iMap[MapID].rp=HIWORD(wparam);  break;
         default: break;
       }
       UpdateReflectionMappingSection(hdlg,MapID);
     }
     else if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_MAPS_B_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wparam)) {
         case TB_PAGEDOWN: iMap[MapID].bp = min(iMax,iMap[MapID].bp+20); break;
         case TB_LINEDOWN: iMap[MapID].bp = min(iMax,iMap[MapID].bp+1); break;
         case TB_PAGEUP:   iMap[MapID].bp = max(iMin,iMap[MapID].bp-20); break;
         case TB_LINEUP:   iMap[MapID].bp = max(iMin,iMap[MapID].bp-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           iMap[MapID].bp=HIWORD(wparam);  break;
         default: break;
       }
       UpdateBumpMappingSection(hdlg,MapID);
     }
//     renderer assummes this is 0 and if used must be changed there too
//     else if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_MAPS_T_SCROLL))){
//       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
//       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
//       switch (LOWORD(wparam)) {
//         case TB_PAGEDOWN: iMap[MapID].tp = min(iMax,iMap[MapID].tp+20); break;
//         case TB_LINEDOWN: iMap[MapID].tp = min(iMax,iMap[MapID].tp+1); break;
//         case TB_PAGEUP:   iMap[MapID].tp = max(iMin,iMap[MapID].tp-20); break;
//         case TB_LINEUP:   iMap[MapID].tp = max(iMin,iMap[MapID].tp-1); break;
//         case TB_THUMBTRACK:
//         case TB_THUMBPOSITION:
//           iMap[MapID].tp=HIWORD(wparam);  break;
//         default: break;
//       }
//       UpdateTransMappingSection(hdlg,MapID);
//     }
     break;
   case WM_COMMAND:
     switch(HIWORD(wparam)){
       case CBN_EDITCHANGE:
         SendDlgItemMessage(hdlg,DLG_MAPS_LIST,WM_GETTEXT,(WPARAM)30,
                            (LPARAM)iMap[MapID].N);
         if(strlen(iMap[MapID].N) == 0)strcpy(iMap[MapID].N,"$");
         SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_DELETESTRING,
                                (WPARAM)MapID,0);
         SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_INSERTSTRING,
                                (WPARAM)MapID,(LPARAM)iMap[MapID].N);
         break;
       case CBN_SELENDOK:
         i=(int)SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_GETCURSEL,0,0);
         if(i != CB_ERR){
           MapID=(long)i;
           LastUsedMapID=MapID; // save global
           UpdateMappingDialog(hdlg,MapID);
           SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_SELECTSTRING,0,
                           (LPARAM)iMap[MapID].N);
         }
         break;
       default:
         break;
     }
     switch(LOWORD(wparam)){
       case DLG_MAPS_MOVE:
         // if openGL then update the OpenGL maps
         if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,(WM_USER+4),0,0);
         EndDialog(hdlg,MapID);
         return(TRUE);
       case DLG_MAPS_NEW:
         MapID=nImaps;
         LastUsedMapID=MapID; // save global
         CreateModelerBrush();
         SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_ADDSTRING,0,
                               (LPARAM)iMap[MapID].N);
         SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_SELECTSTRING,0,
                           (LPARAM)iMap[MapID].N);
         UpdateMappingDialog(hdlg,MapID);
         break;
       case DLG_MAPS_DELETE:
         SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_DELETESTRING,
                                (WPARAM)MapID,0);
         DumpBrush(MapID);
         // if openGL then update the OpenGL maps
         if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,
                                                (WM_USER+4),0,0);
         if(nImaps == 0){
           EndDialog(hdlg,-1);
           return(TRUE);
         }
         if(MapID >= nImaps)MapID=nImaps-1;
         LastUsedMapID=MapID; // save global
         UpdateMappingDialog(hdlg,MapID);
         SendDlgItemMessage(hdlg,DLG_MAPS_LIST,CB_SELECTSTRING,0,
                           (LPARAM)iMap[MapID].N);
         break;
       case DLG_MAPS_G:
         iMap[MapID].Map=MAP_BY_VERTEX;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_P:
         if(iMap[MapID].Map == CYLINDER_MOZIAC)iMap[MapID].Map=PLANE_MOZIAC;
         else iMap[MapID].Map=PLANE;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_C:
         if(iMap[MapID].Map == PLANE_MOZIAC)iMap[MapID].Map=CYLINDER_MOZIAC;
         else iMap[MapID].Map=CYLINDER;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_S:
         iMap[MapID].Map=MAP_SPHERICAL;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_PM:
         if(iMap[MapID].Map == PLANE_MOZIAC)iMap[MapID].Map=PLANE;
         if(iMap[MapID].Map == CYLINDER_MOZIAC)iMap[MapID].Map=CYLINDER;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_CM:
         if(iMap[MapID].Map == PLANE)iMap[MapID].Map=PLANE_MOZIAC;
         if(iMap[MapID].Map == CYLINDER)iMap[MapID].Map=CYLINDER_MOZIAC;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_KEY_INDEX:
         iMap[MapID].Key=0;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_KEY_CHOSEN:
         iMap[MapID].Key=1;
         UpdateTypeMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_KEY_COLOUR:
         SetSfxColour(iMap[MapID].k_colour,IDX_MISC_KEYCOLOUR,hdlg);
         InvalidateRect(GetDlgItem(hdlg,DLG_MAPS_KEY_COLOUR),NULL,FALSE);
         break;
       case DLG_MAPS_S_ACTIVE:
         if(iMap[MapID].S == NULL)SendMessage(hdlg,WM_COMMAND,DLG_MAPS_S_SET,0);
         else                     SendMessage(hdlg,WM_COMMAND,DLG_MAPS_S_DROP,0);
         break;
       case DLG_MAPS_S_SET:
         if(iMap[MapID].Type == ANIMBRUSH)
              iMap[MapID].S=GetFlicFromDisk(MapID,iMap[MapID].S,hdlg,0);
         else iMap[MapID].S=GetMapFromDisk(iMap[MapID].S,hdlg);
         if(iMap[MapID].S == NULL)iMap[MapID].s=0;
         else                     iMap[MapID].s=1;
         UpdateSurfaceMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_S_PRESET:
         if((iMap[MapID].S=GetPresetMap(iMap[MapID].S,hdlg)) == NULL)
              iMap[MapID].s=0;
         else iMap[MapID].s=1;
         UpdateSurfaceMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_S_DROP:
         if(iMap[MapID].S != NULL)X__Free(iMap[MapID].S);
         iMap[MapID].S=NULL; iMap[MapID].s=0;
         UpdateSurfaceMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_S_VIEW:
         if(iMap[MapID].S != NULL)DisplayImageFile(iMap[MapID].S);
         break;
       case DLG_MAPS_S_MODIFY:
         if(iMap[MapID].S != NULL)EditImageFile(iMap[MapID].S);
         break;
       case DLG_MAPS_S_ANIMATED:
       case DLG_MAPS_S_STILL:
         if(iMap[MapID].Type == ANIMBRUSH){
           iMap[MapID].Type=STATICBRUSH;
           if(iMap[MapID].S != NULL){
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_S_DROP,0);
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_S_SET,0);
           }
           else UpdateSurfaceMappingSection(hdlg,MapID);
         }
         else if(iMap[MapID].Type != ANIMBRUSH){
           iMap[MapID].Type=ANIMBRUSH;
           if(iMap[MapID].S != NULL){
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_S_DROP,0);
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_S_SET,0);
           }
           else UpdateSurfaceMappingSection(hdlg,MapID);
         }
         break;
       case DLG_MAPS_R_ACTIVE:
         if(iMap[MapID].R == NULL)SendMessage(hdlg,WM_COMMAND,DLG_MAPS_R_SET,0);
         else                     SendMessage(hdlg,WM_COMMAND,DLG_MAPS_R_DROP,0);
         break;
       case DLG_MAPS_R_SET:
         if(iMap[MapID].TypeR == ANIMBRUSH)
              iMap[MapID].R=GetFlicFromDisk(MapID,iMap[MapID].R,hdlg,1);
         else iMap[MapID].R=GetMapFromDisk(iMap[MapID].R,hdlg);
         if(iMap[MapID].R == NULL)iMap[MapID].r=0;
         else                     iMap[MapID].r=1;
         UpdateReflectionMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_R_DROP:
         if(iMap[MapID].R != NULL)X__Free(iMap[MapID].R);
         iMap[MapID].R=NULL; iMap[MapID].r=0;
         UpdateReflectionMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_R_VIEW:
         if(iMap[MapID].R != NULL)DisplayImageFile(iMap[MapID].R);
         break;
       case DLG_MAPS_R_MODIFY:
         if(iMap[MapID].R != NULL)EditImageFile(iMap[MapID].R);
         break;
       case DLG_MAPS_R_STILL:
       case DLG_MAPS_R_ANIMATED:
         if(iMap[MapID].TypeR == ANIMBRUSH){
           iMap[MapID].TypeR=STATICBRUSH;
           if(iMap[MapID].R != NULL){
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_R_DROP,0);
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_R_SET,0);
           }
           else UpdateReflectionMappingSection(hdlg,MapID);
         }
         else if(iMap[MapID].TypeR != ANIMBRUSH){
           iMap[MapID].TypeR=ANIMBRUSH;
           if(iMap[MapID].R != NULL){
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_R_DROP,0);
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_R_SET,0);
           }
           else UpdateReflectionMappingSection(hdlg,MapID);
         }
         break;
       case DLG_MAPS_B_ACTIVE:
         if(iMap[MapID].B == NULL)SendMessage(hdlg,WM_COMMAND,DLG_MAPS_B_SET,0);
         else                     SendMessage(hdlg,WM_COMMAND,DLG_MAPS_B_DROP,0);
         break;
       case DLG_MAPS_B_SET:
         if(iMap[MapID].TypeB == ANIMBRUSH)
              iMap[MapID].B=GetFlicFromDisk(MapID,iMap[MapID].B,hdlg,2);
         else iMap[MapID].B=GetMapFromDisk(iMap[MapID].B,hdlg);
         if(iMap[MapID].B == NULL)iMap[MapID].b=0;
         else                     iMap[MapID].b=1;
         UpdateBumpMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_B_DROP:
         if(iMap[MapID].B != NULL)X__Free(iMap[MapID].B);
         iMap[MapID].B=NULL; iMap[MapID].b=0;
         UpdateBumpMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_B_VIEW:
         if(iMap[MapID].B != NULL)DisplayImageFile(iMap[MapID].B);
         break;
       case DLG_MAPS_B_MODIFY:
         if(iMap[MapID].B != NULL)EditImageFile(iMap[MapID].B);
         break;
       case DLG_MAPS_B_ANIMATED:
       case DLG_MAPS_B_STILL:
         if(iMap[MapID].TypeB == ANIMBRUSH){
           iMap[MapID].TypeB=STATICBRUSH;
           if(iMap[MapID].B != NULL){
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_B_DROP,0);
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_B_SET,0);
           }
           else UpdateBumpMappingSection(hdlg,MapID);
         }
         else if(iMap[MapID].TypeB != ANIMBRUSH){
           iMap[MapID].TypeB=ANIMBRUSH;
           if(iMap[MapID].B != NULL){
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_B_DROP,0);
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_B_SET,0);
           }
           else UpdateBumpMappingSection(hdlg,MapID);
         }
         break;
       case DLG_MAPS_T_ACTIVE:
         if(iMap[MapID].T == NULL)SendMessage(hdlg,WM_COMMAND,DLG_MAPS_T_SET,0);
         else                     SendMessage(hdlg,WM_COMMAND,DLG_MAPS_T_DROP,0);
         break;
       case DLG_MAPS_T_SET:
         if(iMap[MapID].TypeT == ANIMBRUSH)
              iMap[MapID].T=GetFlicFromDisk(MapID,iMap[MapID].T,hdlg,3);
         else iMap[MapID].T=GetMapFromDisk(iMap[MapID].T,hdlg);
         if(iMap[MapID].T == NULL)iMap[MapID].t=0;
         else                     iMap[MapID].t=1;
         UpdateTransMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_T_DROP:
         if(iMap[MapID].T != NULL)X__Free(iMap[MapID].T);
         iMap[MapID].T=NULL; iMap[MapID].t=0;
         UpdateTransMappingSection(hdlg,MapID);
         break;
       case DLG_MAPS_T_VIEW:
         if(iMap[MapID].T != NULL)DisplayImageFile(iMap[MapID].T);
         break;
       case DLG_MAPS_T_MODIFY:
         if(iMap[MapID].T != NULL)EditImageFile(iMap[MapID].T);
         break;
       case DLG_MAPS_T_ANIMATED:
       case DLG_MAPS_T_STILL:
         if(iMap[MapID].TypeT == ANIMBRUSH){
           iMap[MapID].TypeT=STATICBRUSH;
           if(iMap[MapID].T != NULL){
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_T_DROP,0);
              SendMessage(hdlg,WM_COMMAND,DLG_MAPS_T_SET,0);
           }
           else UpdateTransMappingSection(hdlg,MapID);
         }
         else if(iMap[MapID].TypeT != ANIMBRUSH){
           iMap[MapID].TypeT=ANIMBRUSH;
           if(iMap[MapID].T != NULL){
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_T_DROP,0);
             SendMessage(hdlg,WM_COMMAND,DLG_MAPS_T_SET,0);
           }
           else UpdateTransMappingSection(hdlg,MapID);
         }
         break;
       case DLG_MAPS_PREVIEW:
         SendMessage(ghwnd_main,WM_COMMAND,(WPARAM)IDM_VIEW_RENDERINWINDOW,0);
         break;
       case IDCANCEL:
       case DLG_MAPS_DONE:
         // if openGL then update the OpenGL maps
         if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,
                                                (WM_USER+4),0,0);
         EndDialog(hdlg,-1);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
}
// End Mapping Dialog

static void EditImageFile(char *pathname){
 char appname[256];
 GetPrivateProfileString("APREFERENCES","EDITAPP","$$$",
                         appname,sizeof(appname),IniFilename);
 if(strncmp(appname,"$$$",3) != 0){
   ShellExecute(ghwnd_main,NULL,appname,
                pathname,appname,SW_SHOW);
 }
 return;
}

void EditImageMap(int id){
 int id1;
 if(id < 0){
  if(id == -2){ // create a new map;
    id1=nImaps;
    if(CreateModelerBrush() == FAIL)return;

  }
  else{
//    id1=RequestMapID(ghwnd_main,"Choose Map",1);
//    if(id1 < 0)return;
    id1=LastUsedMapID;
    if(id1 >= nImaps)id1=0;
    if(id1 == nImaps){
      if(CreateModelerBrush() == FAIL)return;
    }
  }
  id=id1;
 }
 EnableToolPannels(SELECT_PANNEL|COMMAND_PANNEL|TOOL_PANNEL|
                   ACTION_PANNEL|COORD_PANNEL|LOCK_PANNEL,FALSE);
 ActivateAllMenus(ghwnd_main,MF_GRAYED);
 id1=(int)DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_MAPS),ghwnd_main,
            (DLGPROC)MappingDlgProc,(LPARAM)id);
 ActivateAllMenus(ghwnd_main,MF_ENABLED);
 EnableToolPannels(SELECT_PANNEL|COMMAND_PANNEL|TOOL_PANNEL|
                   ACTION_PANNEL|COORD_PANNEL|LOCK_PANNEL,TRUE);
 if(id1 >= 0){
   axis_move_flag=id1;
   PositionBrushOnOff(1,id1);
 }
 else if(image_edit_flag == 1) /* go back to the faces dialog */
   PostMessage(ghwnd_main,WM_COMMAND,IDM_ATTRIBUTES_FACE,0);
}

int RequestMapID(HWND parent,char *message,int new){
 int i,id;
 long N,N1;
 char **ItemList,newname[32]="< new >";
 N=nImaps;
 if(new)N1=N+1; else N1=N;
 if((ItemList = (char **)X__Malloc(N1*sizeof(char *))) == NULL){
   SendPrgmQuery(IDQ_NOMEM2,0);
   return -1;
 }
 for(i=0;i<N;i++)ItemList[i] = iMap[i].N;
 if(new)ItemList[N] = newname;
 id=SelectScrolledItemList(N1,ItemList,message,parent);
 X__Free(ItemList);
 return id;
}

int CreateModelerBrush(void){
  long i;
  EDIT_ACTION=YES;
  if(nImaps >= NMAPSMAX)return FAIL;
  iMap[nImaps].Key=0;
  iMap[nImaps].s=0;    iMap[nImaps].r=0;    iMap[nImaps].b=0;     iMap[nImaps].t=0;
  iMap[nImaps].S=NULL; iMap[nImaps].R=NULL; iMap[nImaps].B=NULL;  iMap[nImaps].T=NULL;
  iMap[nImaps].sp=100; iMap[nImaps].rp=50;  iMap[nImaps].bp=50;   iMap[nImaps].tp=0;
  for(i=0;i<3;i++){
    iMap[nImaps].d_colour[i]=0;
    iMap[nImaps].s_colour[i]=0;
    iMap[nImaps].a_colour[i]=0;
    iMap[nImaps].t_colour[i]=0;
    iMap[nImaps].k_colour[i]=0;
  }
  sprintf(iMap[nImaps].N,"unnamed_%ld",nImaps+1);
  if(ActiveView == TRITOP){
    iMap[nImaps].P[0] = TVpointX + TVsizeX/10;
    iMap[nImaps].P[1] = TVpointY + TVsizeY-TVsizeY/10;
    iMap[nImaps].P[2] = TVpointZ + TVsizeZ/2;
    iMap[nImaps].X[0] = TVpointX + TVsizeX-TVsizeX/10;
    iMap[nImaps].X[1] = TVpointY + TVsizeY-TVsizeY/10;
    iMap[nImaps].X[2] = TVpointZ + TVsizeZ/2;
    iMap[nImaps].Y[0] = TVpointX + TVsizeX/10;
    iMap[nImaps].Y[1] = TVpointY + TVsizeY/10;
    iMap[nImaps].Y[2] = TVpointZ + TVsizeZ/2;
  }
  else if(ActiveView == TRIRIGHT){
    iMap[nImaps].P[0] = TVpointX + TVsizeX/2;
    iMap[nImaps].P[1] = TVpointY + TVsizeY/10;
    iMap[nImaps].P[2] = TVpointZ + TVsizeZ - TVsizeZ/10;
    iMap[nImaps].X[0] = TVpointX + TVsizeX/2;
    iMap[nImaps].X[1] = TVpointY + TVsizeY - TVsizeY/10;
    iMap[nImaps].X[2] = TVpointZ + TVsizeZ - TVsizeZ/10;
    iMap[nImaps].Y[0] = TVpointX + TVsizeX/2;
    iMap[nImaps].Y[1] = TVpointY + TVsizeY/10;
    iMap[nImaps].Y[2] = TVpointZ + TVsizeZ/10;
  }
  else{
    iMap[nImaps].P[0] = TVpointX + TVsizeX/10;
    iMap[nImaps].P[1] = TVpointY + TVsizeY/2;
    iMap[nImaps].P[2] = TVpointZ + TVsizeZ - TVsizeZ/10;
    iMap[nImaps].X[0] = TVpointX + TVsizeX - TVsizeX/10;
    iMap[nImaps].X[1] = TVpointY + TVsizeY/2;
    iMap[nImaps].X[2] = TVpointZ + TVsizeZ - TVsizeZ/10;
    iMap[nImaps].Y[0] = TVpointX + TVsizeX/10;
    iMap[nImaps].Y[1] = TVpointY + TVsizeY/2;
    iMap[nImaps].Y[2] = TVpointZ + TVsizeZ/10;
  }
  iMap[nImaps].Type = STATICBRUSH;    /* Still Anim */
  iMap[nImaps].Angle  = 360;          /* full rotation if required */
  iMap[nImaps].AnimFF = 1;
  iMap[nImaps].AnimDL = 1;
  iMap[nImaps].AnimLF = 1;
  iMap[nImaps].TypeR = STATICBRUSH;
  iMap[nImaps].AnimFFr = 1;
  iMap[nImaps].AnimDLr = 1;
  iMap[nImaps].AnimLFr = 1;
  iMap[nImaps].TypeB = STATICBRUSH;
  iMap[nImaps].AnimFFb = 1;
  iMap[nImaps].AnimDLb = 1;
  iMap[nImaps].AnimLFb = 1;
  iMap[nImaps].TypeT = STATICBRUSH;
  iMap[nImaps].AnimFFt = 1;
  iMap[nImaps].AnimDLt = 1;
  iMap[nImaps].AnimLFt = 1;
  iMap[nImaps].SkId = -1;        /* Brush is not assigned to any skeletal */
  iMap[nImaps].Sk = NULL;        /* component */
  iMap[nImaps].Lock = 0;         /* not locked */
  iMap[nImaps].Vlock[0] = -1;    /* lock P */
  iMap[nImaps].Vlock[1] = -1;    /* lock X */
  iMap[nImaps].Vlock[2] = -1;    /* lock Y */
  iMap[nImaps].Map=PLANE;        /* planar brush map */
  nImaps++;
  return OK;
}

void DumpBrush(int idd){
  int id,i,j;
  char bid;
  face *f;
  if(nImaps == 0){
   SendPrgmQuery(IDQ_NOBRUSH,0);
    return;
  }
  if(idd < 0)id=RequestMapID(ghwnd_main,"Choose Map to delete",0);
  else id=idd;
  if(id < 0 )return;
  EDIT_ACTION=YES;
  if((f=MainFp) != NULL && Nface > 0)for(i=0;i<Nface;i++){
    if( (f->attrib & BRUSHBIT) == BRUSHBIT){/* it's a map   */
      bid=f->map;                      /* extract map id    */
      if(bid == id){
        f->map=0;              /* no more maps this face    */
        f->attrib &= 0xf1;     /* set map styles off        */
        f->attrib &= 0xbf;     /* set to underlying texture */
      }                        /* zero bit 6 leave others   */
      else if(bid > id){       /* slide all the map id's down one */
        bid--;
        f->map = bid;
      }
    }
    f++;
  }
  if(iMap[id].Lock)nLockediMaps--;
  if(iMap[id].s && iMap[id].S != NULL)X__Free(iMap[id].S);
  if(iMap[id].r && iMap[id].R != NULL)X__Free(iMap[id].R);
  if(iMap[id].b && iMap[id].B != NULL)X__Free(iMap[id].B);
  if(iMap[id].t && iMap[id].T != NULL)X__Free(iMap[id].T);
  if(id < nImaps-1)for(i=id;i<nImaps-1;i++){
    strcpy(iMap[i].N,iMap[i+1].N);
    for(j=0;j<3;j++){
      iMap[i].P[j] = iMap[i+1].P[j];
      iMap[i].X[j] = iMap[i+1].X[j];
      iMap[i].Y[j] = iMap[i+1].Y[j];
      iMap[i].d_colour[j]=iMap[i+1].d_colour[j];
      iMap[i].s_colour[j]=iMap[i+1].s_colour[j];
      iMap[i].a_colour[j]=iMap[i+1].a_colour[j];
      iMap[i].k_colour[j]=iMap[i+1].k_colour[j];
    }
    iMap[i].Type = iMap[i+1].Type;
    iMap[i].AnimFF = iMap[i+1].AnimFF;
    iMap[i].AnimDL = iMap[i+1].AnimDL;
    iMap[i].AnimLF = iMap[i+1].AnimLF;
    iMap[i].TypeR = iMap[i+1].TypeR;
    iMap[i].AnimFFr = iMap[i+1].AnimFFr;
    iMap[i].AnimDLr = iMap[i+1].AnimDLr;
    iMap[i].AnimLFr = iMap[i+1].AnimLFr;
    iMap[i].TypeB = iMap[i+1].TypeR;
    iMap[i].AnimFFb = iMap[i+1].AnimFFr;
    iMap[i].AnimDLb = iMap[i+1].AnimDLr;
    iMap[i].AnimLFb = iMap[i+1].AnimLFr;
    iMap[i].TypeT = iMap[i+1].TypeT;
    iMap[i].AnimFFt = iMap[i+1].AnimFFt;
    iMap[i].AnimDLt = iMap[i+1].AnimDLt;
    iMap[i].AnimLFt = iMap[i+1].AnimLFt;
    iMap[i].Angle  = iMap[i+1].Angle;
    iMap[i].SkId = iMap[i+1].SkId;
    iMap[i].Sk = iMap[i+1].Sk;
    iMap[i].Map=iMap[i+1].Map;
    iMap[i].Lock=iMap[i+1].Lock;
    iMap[i].Key=iMap[i+1].Key;
    iMap[i].Vlock[0]=iMap[i+1].Vlock[0];
    iMap[i].Vlock[1]=iMap[i+1].Vlock[1];
    iMap[i].Vlock[2]=iMap[i+1].Vlock[2];
    iMap[i].s=iMap[i+1].s;
    iMap[i].r=iMap[i+1].r;
    iMap[i].b=iMap[i+1].b;
    iMap[i].t=iMap[i+1].t;
    iMap[i].S=iMap[i+1].S;
    iMap[i].R=iMap[i+1].R;
    iMap[i].B=iMap[i+1].B;
    iMap[i].T=iMap[i+1].T;
    iMap[i].sp=iMap[i+1].sp;
    iMap[i].rp=iMap[i+1].rp;
    iMap[i].bp=iMap[i+1].bp;
    iMap[i].tp=iMap[i+1].tp;
  }
  iMap[nImaps-1].S=NULL; iMap[nImaps-1].s=0;
  iMap[nImaps-1].R=NULL; iMap[nImaps-1].r=0;
  iMap[nImaps-1].B=NULL; iMap[nImaps-1].b=0;
  iMap[nImaps-1].T=NULL; iMap[nImaps-1].t=0;
  /* Fixup the identity of the last used brush */
  nImaps--;
  return;
}

void PurgeMaps(void){ /* remove any maps not in use by any face */
 int i,j,bid;
 BOOL found;
 face *f;
 LOOP:
 if(nImaps < 1)return;
 for(j=0;j<nImaps;j++){
   found=FALSE;
   if((f=MainFp) != NULL && Nface > 0)for(i=0;i<Nface;i++){
     if( (f->attrib & BRUSHBIT) == BRUSHBIT){
       bid=(int)(f->map);
       if(bid == j)found=TRUE;
     }
     f++;
   }
   if(!found){     /* map not needed so scrap it      */
     DumpBrush(j);
     goto LOOP;    /* go back until all maps are done */
   }
 }
 // if openGL then update the OpenGL maps
 if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,(WM_USER+4),0,0);
}

void PurgeShaders(void){
 int i,k,id;
 BOOL found;
 face *f;
 if(nSha < 2){
   SendPrgmQuery(IDQ_NOSHADERAXES,0);
   return;
 }
 for(i=0;i<nSha;i++){
   found=FALSE;
   if((f=MainFp) != NULL && Nface > 0)for(k=0;k<Nface;k++){ /* drop shader */
     if((id=(unsigned char)f->axis) == i)found=TRUE;
     if(id >= nSha)f->axis=0; // bad axis - reset to base axis
     f++;
   }
   if(!found)DumpShader(i);
 }
 return;
}

void VectorMaps(void){ /* re vector maps on face */
 char temp[MAX_FILE],new_dir[MAX_DIR];
 PSTR c;
 int j;
 if(nImaps < 1)return;
 strcpy(new_dir,gszHomeDir);strcat(new_dir,"maps\\");
 if(!RequestCharString(64,new_dir,"Root directory",ghwnd_main))return;
 if((j=strlen(new_dir)) < 3){
   return;
 }
 if(new_dir[j-1] != '\\')strcat(new_dir,"\\");
 for(j=0;j<nImaps;j++){
  if(iMap[j].S != NULL && iMap[j].s == 1){
    if((c=FileInPath(iMap[j].S)) != NULL){
      strcpy(temp,c); strcpy(iMap[j].S,new_dir); strcat(iMap[j].S,temp);
    }
  }
  if(iMap[j].R != NULL && iMap[j].r == 1){
    if((c=FileInPath(iMap[j].R)) != NULL){
      strcpy(temp,c); strcpy(iMap[j].R,new_dir); strcat(iMap[j].R,temp);
    }
  }
  if(iMap[j].B != NULL && iMap[j].b == 1){
    if((c=FileInPath(iMap[j].B)) != NULL){
      strcpy(temp,c); strcpy(iMap[j].B,new_dir); strcat(iMap[j].B,temp);
    }
  }
 }
}

void DrawBrushInOne(HDC hDC,int view,int j){
 HRGN hRgn;
 POINT ptp[14];
 HPEN holdPen;
 HBRUSH holdBrush;
 int  oldROP;
 int i,k,N=12,map=0;
 int sho,svo,shx,svx,shy,svy,sht,svt;
 vector n,xp,py,p,q,pp;
 double d,r,nn,dtheta,theta,cs,ss;
 long x,y,z;
 if(iMap[j].Map == CYLINDER ||
    iMap[j].Map == MAP_SPHERICAL ||
    iMap[j].Map == CYLINDER_MOZIAC)map=1;
 if(map){
   if(View != TRIVIEW)N *= 2;
   VECSUB((double)iMap[j].X,(double)iMap[j].P,xp)
   VECSUB((double)iMap[j].P,(double)iMap[j].Y,py)
   nn=DOT(py,py);
   if(nn > 1.0){
     nn=DOT(xp,py)/nn;
     VECSUB((double)iMap[j].X,nn*py,p)
     VECSUB(p,(double)iMap[j].P,xp)
     CROSS(py,xp,n);
     nn=(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]);
     if(nn > 1.0){
       VECCOPY(iMap[j].P,p)
       VECCOPY(iMap[j].Y,pp)
       r=sqrt(xp[0]*xp[0]+xp[1]*xp[1]+xp[2]*xp[2]);
       nn=r/sqrt(nn);
       VECSCALE(nn,n,n)
       dtheta=iMap[j].Angle/180.0*PI/(double)N;
     }
     else map=0;
   }
   else map=0;
 }
 holdPen=SelectObject(hDC,ghInvertPen);
 holdBrush=SelectObject(hDC,GetStockObject(HOLLOW_BRUSH));
 oldROP=SetROP2(hDC,R2_XORPEN);
 SelectPalette(hDC,ghpaletteScreen,FALSE);
 RealizePalette(hDC);
 GetWindowCoords(view,iMap[j].P[0],iMap[j].P[1],iMap[j].P[2],&sho,&svo);
 GetWindowCoords(view,iMap[j].X[0],iMap[j].X[1],iMap[j].X[2],&shx,&svx);
 GetWindowCoords(view,iMap[j].Y[0],iMap[j].Y[1],iMap[j].Y[2],&shy,&svy);
 GetWindowCoords(view,
                      iMap[j].Y[0]+(iMap[j].X[0]-iMap[j].P[0]),
                      iMap[j].Y[1]+(iMap[j].X[1]-iMap[j].P[1]),
                      iMap[j].Y[2]+(iMap[j].X[2]-iMap[j].P[2]),&sht,&svt);
 Rectangle(hDC,sho-5,svo-5,sho+5,svo+5);
 Rectangle(hDC,shx-2,svx-2,shx+2,svx+2);
 MoveToEx(hDC,shy-6,svy-3,NULL); LineTo(hDC,shy+5,svy-3);
 MoveToEx(hDC,shy+6,svy-3,NULL); LineTo(hDC,shy,svy+5);
 MoveToEx(hDC,shy-1,svy+4,NULL); LineTo(hDC,shy-6,svy-2);
 if(map){  /* cylindrical mapping */
   MoveToEx(hDC,sho,svo,NULL); //_setcolor(COLOUR05);
   LineTo(hDC,shx,svx);       // _setcolor(COLOUR14); 14=invert
   MoveToEx(hDC,shy,svy,NULL); LineTo(hDC,sho,svo);
   theta=0;
   for(k=0;k<=N;k++){
     cs=cos(theta); ss=sin(theta);
     VECSUM(cs*xp,ss*n,q)
     x=(long)(p[0]+q[0]); y=(long)(p[1]+q[1]); z=(long)(p[2]+q[2]);
     if(k > 0)MoveToEx(hDC,shy,svy,NULL);
     GetWindowCoords(view,x,y,z,&shy,&svy);
     if(k > 0)LineTo(hDC,shy,svy);
     x=(long)(pp[0]+0.98*q[0]);
     y=(long)(pp[1]+0.98*q[1]);
     z=(long)(pp[2]+0.98*q[2]);
     if(k > 0)MoveToEx(hDC,sht,svt,NULL);
     GetWindowCoords(view,x,y,z,&sht,&svt);
     if(k > 0)LineTo(hDC,sht,svt);
     else{
       LineTo(hDC,shy,svy);
       LineTo(hDC,sht,svt);
     }
     theta += dtheta;
   }
 }
 else{
   MoveToEx(hDC,sho,svo,NULL);
   LineTo(hDC,shx,svx);  LineTo(hDC,sht,svt);
   LineTo(hDC,shy,svy);  LineTo(hDC,sho,svo);
   ptp[0].x=sho; ptp[0].y=svo;
   ptp[1].x=shx; ptp[1].y=svx;
   ptp[2].x=sht; ptp[2].y=svt;
   ptp[3].x=shy; ptp[3].y=svy;
   hRgn=CreatePolygonRgn(ptp,4,WINDING);
   InvertRgn(hDC,hRgn);
   DeleteObject(hRgn);
 }
 SelectObject(hDC,holdPen);
 SelectObject(hDC,holdBrush);
 SetROP2(hDC,oldROP);
 return;
}

void PositionBrushOnOff(int OnOff,int id){
 int i;
 static int LastModellerBrushUsed = -1;
 if(!OnOff){
   lock_visible=NO;
   if(ghwndLock)ShowWindow(ghwndLock,SW_HIDE);
   controls_visible=YES;
   EnableToolPannels(COMMAND_PANNEL|SELECT_PANNEL|TOOL_PANNEL|ACTION_PANNEL,TRUE);
   ActivateAllMenus(ghwnd_main,MF_ENABLED);
   EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_GRAYED);
   DrawMenuBar(ghwnd_main);
   DrawBrush(SelectedBrush);
   if(ghwndOpenGLview != NULL)
     PostMessage(ghwndOpenGLview,(WM_USER+6),(WPARAM)FALSE,0); // GL map axis
   SelectedBrush = -1;
   if(IsWindow(ghwndDlgAttribs))ShowWindow(ghwndDlgAttribs,SW_SHOW);
   if(axis_move_flag >= 0){
     PostMessage(ghwnd_main,WM_COMMAND,IDM_ATTRIBUTES_MAP,0);
   }
   return;
 }
 if(SelectedBrush >= 0)return;
 if(nImaps == 0){
   SendPrgmQuery(IDQ_NOBRUSH,0);
   return;
 }
 if(id < 0){
   if(nImaps == 1)id=0;
   else if((id=RequestMapID(ghwnd_main,"Choose Map",0)) < 0)return;
 }
 if(iMap[id].Map == MAP_BY_VERTEX){
   SendPrgmQuery(IDQ_BRUSHGLUED,0);
   return;
 }
 EDIT_ACTION=YES;
 SelectedBrush=id;
 LastModellerBrushUsed=SelectedBrush;
 lock_visible=YES;
 if(ghwndLock)ShowWindow(ghwndLock,SW_SHOW);
 if(IsWindow(ghwndDlgAttribs))ShowWindow(ghwndDlgAttribs,SW_HIDE);
 controls_visible=NO;
 EnableToolPannels(COMMAND_PANNEL|SELECT_PANNEL|TOOL_PANNEL|ACTION_PANNEL,FALSE);
 ActivateAllMenus(ghwnd_main,MF_GRAYED);
 EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_ENABLED);
 DrawMenuBar(ghwnd_main);
 DrawBrush(SelectedBrush);
 if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,
   (WM_USER+6),(WPARAM)TRUE,(LPARAM)SelectedBrush); // GL map axis
 return;
}

void GrabBrushPoint(int x,int y){
 int id,x0,y0,x1,y1,x2,y2;
 double d,dmin;
 if(SelectedBrush < 0)return;
 GetWindowCoords(ActiveView,iMap[SelectedBrush].P[0],
           iMap[SelectedBrush].P[1],iMap[SelectedBrush].P[2],&x0,&y0);
 GetWindowCoords(ActiveView,iMap[SelectedBrush].X[0],
           iMap[SelectedBrush].X[1],iMap[SelectedBrush].X[2],&x1,&y1);
 GetWindowCoords(ActiveView,iMap[SelectedBrush].Y[0],
           iMap[SelectedBrush].Y[1],iMap[SelectedBrush].Y[2],&x2,&y2);
 id = -1; dmin=(double)TVsizeX*(double)TVsizeY*2.0;
 if(abs(x-x0) < 4 && abs(y-y0) < 4){
   d=DIS2(iMap[SelectedBrush].P); if(d < dmin){id = 0; dmin=d;}
 }
 if(abs(x-x1) < 4 && abs(y-y1) < 4){
   d=DIS2(iMap[SelectedBrush].X); if(d < dmin){id = 1; dmin=d;}
 }
 if(abs(x-x2) < 4 && abs(y-y2) < 4){
   d=DIS2(iMap[SelectedBrush].Y); if(d < dmin){id = 2; dmin=d;}
 }
 if(id >= 0){
   if(iMap[SelectedBrush].Lock){
    MessageBeep(MB_OK);
    SendPrgmMessage(IDS_MAPLOCKED,1);
   }
   else IdentifiedBrushPoint=id;
 }
}

void UnGrabBrushPoint(void){
  IdentifiedBrushPoint=-1;
  if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,(WM_USER+3),0,0);
}

void PullBrushPoint(int x,int y){
 int i,p,sb,ix,iy,iz;
 long x1,y1,z1,dx,dy,dz;
 if((p=IdentifiedBrushPoint) >= 0){
   sb=SelectedBrush;
   DrawBrush(sb);
   GetWorldCoords(ActiveView,&x1,&y1,&z1,x,y);
   ix=iy=iz=1;
   if     (ActiveView == TRITOP  )iz=0;
   else if(ActiveView == TRIFRONT)iy=0;
   else if(ActiveView == TRIRIGHT)ix=0;
   if(p == 0){
     dx=iMap[sb].X[0]-iMap[sb].P[0];
     dy=iMap[sb].X[1]-iMap[sb].P[1];
     dz=iMap[sb].X[2]-iMap[sb].P[2];
     if(ix)iMap[sb].X[0]=x1+dx;
     if(iy)iMap[sb].X[1]=y1+dy;
     if(iz)iMap[sb].X[2]=z1+dz;
     dx=iMap[sb].Y[0]-iMap[sb].P[0];
     dy=iMap[sb].Y[1]-iMap[sb].P[1];
     dz=iMap[sb].Y[2]-iMap[sb].P[2];
     if(ix){iMap[sb].Y[0]=x1+dx; iMap[sb].P[0]=x1;}
     if(iy){iMap[sb].Y[1]=y1+dy; iMap[sb].P[1]=y1;}
     if(iz){iMap[sb].Y[2]=z1+dz; iMap[sb].P[2]=z1;}
   }
   else if(p == 1){
     if(ix)iMap[sb].X[0]=x1;
     if(iy)iMap[sb].X[1]=y1;
     if(iz)iMap[sb].X[2]=z1;
   }
   else if(p == 2){
     if(ix)iMap[sb].Y[0]=x1;
     if(iy)iMap[sb].Y[1]=y1;
     if(iz)iMap[sb].Y[2]=z1;
   }
   DrawBrush(SelectedBrush);
   // inserted to draw brush in background windows
   UpdateWindow(ghwnd_triview[ActiveView]);
 }
}

void AddAdditionalShader(void){
 long i;
 char tempstr[30];
 if((i=nSha) > 63)return;
 sprintf(tempstr,"New%ld",i);
 if(RequestCharString(10,tempstr,"New Name",ghwnd_main) == 0)return;
 if(strlen(tempstr) > 0){
   strcpy(iSha[i].N,tempstr);
   iSha[i].P[0]=0;          iSha[i].P[1]=0;         iSha[i].P[2]=TVpointZ;
   iSha[i].X[0]=TVsizeX/2;  iSha[i].X[1]=0;         iSha[i].X[2]=TVpointZ;
   iSha[i].Y[0]=0;          iSha[i].Y[1]=TVsizeZ/2; iSha[i].Y[2]=TVpointZ;
   nSha++;
 }
}

void DumpShader(int idd){
  int id,i,j,k;
  unsigned char bid;
  face *f;
  if(nSha < 2){
    SendPrgmQuery(IDQ_NOSHADERAXES,0);
    return;
  }
  if(idd < 0){
    id=RequestShaderID(ghwnd_main,"Choose axis to delete");
    if(id < 0)return;  // cancel pressed
  }
  else id=idd;
  if(id < 1){
    SendPrgmQuery(IDQ_NODELETEROOTSHADER,0);
    return;
  }
  EDIT_ACTION=YES;
  if((f=MainFp) != NULL && Nface > 0)for(k=0;k<Nface;k++){ /* drop shader */
    if((bid = (unsigned char)f->axis) == id){
      f->axis=0;            /* goto default shader         */
    }
    else if(bid > id){      /* slide all the axes down one */
      bid--;
      f->axis=bid;
    }
    f++;
  }
  if(id < nSha-1)for(i=id;i<nSha-1;i++){
    strcpy(iSha[i].N,iSha[i+1].N);
    for(j=0;j<3;j++){
      iSha[i].P[j] = iSha[i+1].P[j];
      iSha[i].X[j] = iSha[i+1].X[j];
      iSha[i].Y[j] = iSha[i+1].Y[j];
    }
  }
  if     (id <  MirrorFlag)MirrorFlag--;   /* keep track of MirrorFlag */
  else if(id == MirrorFlag)MirrorFlag= -1; /* this axis IS mirror flag */
  nSha--;
  return;
}

void DrawShaderInOne(HDC hDC,int view,int j){
 HPEN holdPen;
 HBRUSH holdBrush;
 int  oldROP;
 int i,k;
 int sho,svo,shx,svx,shy,svy;
 long x,y,z;
 holdPen=SelectObject(hDC,ghInvertPen);
 holdBrush=SelectObject(hDC,GetStockObject(HOLLOW_BRUSH));
 oldROP=SetROP2(hDC,R2_XORPEN);
 SelectPalette(hDC,ghpaletteScreen,FALSE);
 RealizePalette(hDC);
 GetWindowCoords(view,iSha[j].P[0],iSha[j].P[1],iSha[j].P[2],&sho,&svo);
 GetWindowCoords(view,iSha[j].X[0],iSha[j].X[1],iSha[j].X[2],&shx,&svx);
 GetWindowCoords(view,iSha[j].Y[0],iSha[j].Y[1],iSha[j].Y[2],&shy,&svy);
 Rectangle(hDC,sho-5,svo-5,sho+5,svo+5);
 Rectangle(hDC,shx-2,svx-2,shx+2,svx+2);
 MoveToEx(hDC,shy-6,svy-3,NULL); LineTo(hDC,shy+5,svy-3);
 MoveToEx(hDC,shy+6,svy-3,NULL); LineTo(hDC,shy,svy+5);
 MoveToEx(hDC,shy-1,svy+4,NULL); LineTo(hDC,shy-6,svy-2);
 MoveToEx(hDC,shx,svx,NULL);  LineTo(hDC,sho,svo); LineTo(hDC,shy,svy);
 SelectObject(hDC,holdPen);
 SelectObject(hDC,holdBrush);
 SetROP2(hDC,oldROP);
 return;
}

void PositionShaderOnOff(int OnOff,int id){
 int i;
 static int LastShaderUsed = -1;
 if(!OnOff){
   EnableToolPannels(COMMAND_PANNEL|SELECT_PANNEL|TOOL_PANNEL|ACTION_PANNEL,TRUE);
   ActivateAllMenus(ghwnd_main,MF_ENABLED);
   EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_GRAYED);
   DrawMenuBar(ghwnd_main);
   lock_visible=NO;
   if(ghwndLock)ShowWindow(ghwndLock,SW_HIDE);
   controls_visible=YES;
   DrawShader(SelectedShader);
   SelectedShader = -1;
   if(IsWindow(ghwndDlgAttribs))ShowWindow(ghwndDlgAttribs,SW_SHOW);
   return;
 }
 if(SelectedShader >= 0)return;
 if(id < 0){
   if(nSha == 1)id=0;
   else if((id=RequestShaderID(ghwnd_main,"Choose Axis")) < 0)return;
 }
 EDIT_ACTION=YES;
 SelectedShader=id;
 LastShaderUsed=SelectedShader;
 EnableToolPannels(COMMAND_PANNEL|SELECT_PANNEL|TOOL_PANNEL|ACTION_PANNEL,FALSE);
 ActivateAllMenus(ghwnd_main,MF_GRAYED);
 EnableMenuItem(GetMenu(ghwnd_main),IDM_STOP,MF_ENABLED);
 DrawMenuBar(ghwnd_main);
 controls_visible=NO;
 lock_visible=YES;
 if(ghwndLock)ShowWindow(ghwndLock,SW_SHOW);
 if(IsWindow(ghwndDlgAttribs))ShowWindow(ghwndDlgAttribs,SW_HIDE);
 DrawShader(SelectedShader);
 return;
}

void GrabShaderPoint(int x,int y){
 int id,x0,y0,x1,y1,x2,y2;
 double d,dmin;
 if(SelectedShader < 0)return;
 GetWindowCoords(ActiveView,iSha[SelectedShader].P[0],
           iSha[SelectedShader].P[1],iSha[SelectedShader].P[2],&x0,&y0);
 GetWindowCoords(ActiveView,iSha[SelectedShader].X[0],
           iSha[SelectedShader].X[1],iSha[SelectedShader].X[2],&x1,&y1);
 GetWindowCoords(ActiveView,iSha[SelectedShader].Y[0],
           iSha[SelectedShader].Y[1],iSha[SelectedShader].Y[2],&x2,&y2);
 id = -1; dmin=(double)TVsizeX*(double)TVsizeY*2.0;
 if(abs(x-x0) < 4 && abs(y-y0) < 4){
   d=DIS2(iSha[SelectedShader].P); if(d < dmin){id = 0; dmin=d;}
 }
 if(abs(x-x1) < 4 && abs(y-y1) < 4){
   d=DIS2(iSha[SelectedShader].X); if(d < dmin){id = 1; dmin=d;}
 }
 if(abs(x-x2) < 4 && abs(y-y2) < 4){
   d=DIS2(iSha[SelectedShader].Y); if(d < dmin){id = 2; dmin=d;}
 }
 if(id >= 0){
   IdentifiedShaderPoint=id;
 }
}

void UnGrabShaderPoint(void){
  IdentifiedShaderPoint=-1;
}

void PullShaderPoint(int x,int y){
 int p,sb,ix,iy,iz;
 long x1,y1,z1,dx,dy,dz;
 if((p=IdentifiedShaderPoint) >= 0){
   sb=SelectedShader;
   DrawShader(sb);
   GetWorldCoords(ActiveView,&x1,&y1,&z1,x,y);
   ix=iy=iz=1;
   if     (ActiveView == TRITOP  )iz=0;
   else if(ActiveView == TRIFRONT)iy=0;
   else if(ActiveView == TRIRIGHT)ix=0;
   if(p == 0){
     dx=iSha[sb].X[0]-iSha[sb].P[0];
     dy=iSha[sb].X[1]-iSha[sb].P[1];
     dz=iSha[sb].X[2]-iSha[sb].P[2];
     if(ix)iSha[sb].X[0]=x1+dx;
     if(iy)iSha[sb].X[1]=y1+dy;
     if(iz)iSha[sb].X[2]=z1+dz;
     dx=iSha[sb].Y[0]-iSha[sb].P[0];
     dy=iSha[sb].Y[1]-iSha[sb].P[1];
     dz=iSha[sb].Y[2]-iSha[sb].P[2];
     if(ix){iSha[sb].Y[0]=x1+dx; iSha[sb].P[0]=x1;}
     if(iy){iSha[sb].Y[1]=y1+dy; iSha[sb].P[1]=y1;}
     if(iz){iSha[sb].Y[2]=z1+dz; iSha[sb].P[2]=z1;}
   }
   else if(p == 1){
     if(ix)iSha[sb].X[0]=x1;
     if(iy)iSha[sb].X[1]=y1;
     if(iz)iSha[sb].X[2]=z1;
   }
   else if(p == 2){
     if(ix)iSha[sb].Y[0]=x1;
     if(iy)iSha[sb].Y[1]=y1;
     if(iz)iSha[sb].Y[2]=z1;
   }
   DrawShader(SelectedShader);
 }
}

int RequestShaderID(HWND parent,char *message){
 int N,i,id;
 char **ItemList;
 N=nSha;
 if((ItemList = (char **)X__Malloc(N*sizeof(char *))) == NULL){
   SendPrgmQuery(IDQ_NOMEM2,0);
   return -1;
 }
 for(i=0;i<N;i++)ItemList[i] = iSha[i].N;
 id=SelectScrolledItemList(N,ItemList,message,parent);
 X__Free(ItemList);
 return id;
}

int RequestMirrorID(HWND parent){
 int i,id;
 long N,N1;
 char **ItemList,newname[32]="< none >";
 N=nSha;
 N1=N+1;
 if((ItemList = (char **)X__Malloc(N1*sizeof(char *))) == NULL){
   SendPrgmQuery(IDQ_NOMEM2,0);
   return -1;
 }
 for(i=0;i<N;i++)ItemList[i] = iSha[i].N;
 ItemList[N] = newname;
 id=SelectScrolledItemList(N1,ItemList,"Mirror Plane Axis",parent);
 X__Free(ItemList);
 if(id == N)return -1;
 EDIT_ACTION=YES;
 return id;
}

static char *GetMapFromDisk(char *p, HWND parent){
 char *pp;
 if(SelectSfxFileName(0,gszIMGfile,gszIMGdir,IDX_MISC_IMAGEFORMAP,
     "All Image Formats|*.bmp;*.gif;*.tga;*.cel;*.tif;*.jpg;*.png|"
     "BMP (*.BMP) |*.bmp|"
     "Compuserve GIF (*.GIF) |*.gif|"
     "Targa TGA (*.TGA) |*.tga|"
     "TIFF (*.TIF) |*.tif|"
     "JPEG (*.JPG) |*.jpg|"
     "PNG (*.PNG) |*.png|"
     "Autodesk Animator Cels (*.CEL) |*.cel|"
                   ,parent) == TRUE){
   if(p == NULL){
     pp=(char *)X__Malloc(256);
     if(pp == NULL)return p; else p=pp;
   }
   strcpy(p,gszIMGfile);
 }
 return p;
}

static char *GetFlicFromDisk(int id, char *p, HWND parent, int type){
 int ff,lf,dl;
 char *pp;
 char tfs[32],tff[32],tlf[32],tsk[32];
 if(SelectSfxFileName(0,gszFLXfile,gszFLXdir,IDX_MISC_FIRSTIMAGE,
   "8 Bit Images|*.gif|24 bit Images|*.tga|",parent) == TRUE){
   if(type == 0){
     ff=(int)iMap[id].AnimFF;
     lf=(int)iMap[id].AnimLF;
     dl=(int)iMap[id].AnimDL;
   }
   else if(type == 1){
     ff=(int)iMap[id].AnimFFr;
     lf=(int)iMap[id].AnimLFr;
     dl=(int)iMap[id].AnimDLr;
   }
   else if(type == 2){
     ff=(int)iMap[id].AnimFFb;
     lf=(int)iMap[id].AnimLFb;
     dl=(int)iMap[id].AnimDLb;
   }
   else if(type == 3){
     ff=(int)iMap[id].AnimFFt;
     lf=(int)iMap[id].AnimLFt;
     dl=(int)iMap[id].AnimDLt;
   }
   LoadString(ghinst_main,IDX_FLIC_SET1,tfs,31);
   LoadString(ghinst_main,IDX_FLIC_SET2,tff,31);
   LoadString(ghinst_main,IDX_FLIC_SET3,tlf,31);
   LoadString(ghinst_main,IDX_FLIC_SET4,tsk,31);
   if(Read3Ints(tfs,tff,tlf,tsk,&ff,&lf,&dl,parent) == FAIL)return p;
   if(type == 0){
     iMap[id].AnimFF=(short)ff;
     iMap[id].AnimLF=(short)lf;
     iMap[id].AnimDL=(short)dl;
   }
   else if(type == 1){
     iMap[id].AnimFFr=(short)ff;
     iMap[id].AnimLFr=(short)lf;
     iMap[id].AnimDLr=(short)dl;
   }
   else if(type == 2){
     iMap[id].AnimFFb=(short)ff;
     iMap[id].AnimLFb=(short)lf;
     iMap[id].AnimDLb=(short)dl;
   }
   else if(type == 3){
     iMap[id].AnimFFt=(short)ff;
     iMap[id].AnimLFt=(short)lf;
     iMap[id].AnimDLt=(short)dl;
   }
   if(p == NULL){
     pp=(char *)X__Malloc(256);
     if(pp == NULL)return p; else p=pp;
   }
   strcpy(p,gszFLXfile);
 }
 return p;
}

static char *GetPresetMap(char *p, HWND parent){
 int id;
 char *pp,lname[255];
 HMODULE hLib;
 FARPROC fpFun;
 int (*fpFun1)(HWND, char *);
#if __DEMO__
 MessageBox(NULL,"Not allowed in Demo",NULL,MB_OK);
 return p;
#endif
 EDIT_ACTION=YES;
 strcpy(lname,gszHomeDir); strcat(lname,"premap.dll");
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_GetPresetMapImage")) != NULL){
     fpFun1 = (void *)fpFun;
     strcpy(lname,gszHomeDir); strcat(lname,"maps\\presets\\textures.bmp");
     id=(*fpFun1)(parent,lname);
     if(id > 0 && id <= 30){
       if(p == NULL){
         pp=(char *)X__Malloc(256);
         if(pp == NULL)return p; else p=pp;
       }
       sprintf(p,"%smaps\\presets\\tx%03ld.gif",gszHomeDir,id);
     }
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 return p;
}

static void DisplayImageFile(char *pathname){
 WritePrivateProfileString("VIEWER","MAPFILE",pathname,IniFilename);
 LoadViewer(2);
 return;
}

static double Distance3D(point a, long *b){
 return sqrt((double)(a[0]-b[0])*(double)(a[0]-b[0])+
             (double)(a[1]-b[1])*(double)(a[1]-b[1])+
             (double)(a[2]-b[2])*(double)(a[2]-b[2]));
}

static int E_MapToPlaneBrush(vector vi,
                            vector p0, vector p1, vector p2,
                            double *a, double *b){
 double ve1,ve2,ve3,vp1,vp2,vp3,q1,q2,q3,det,
        det1,det2,det3,detmax,dm;
 int k;
  q1=vi[0]-p0[0];  q2=vi[1]-p0[1];  q3=vi[2]-p0[2];
 ve1=p2[0]-p0[0]; ve2=p2[1]-p0[1]; ve3=p2[2]-p0[2];
 vp1=p1[0]-p0[0]; vp2=p1[1]-p0[1]; vp3=p1[2]-p0[2];
 det1=ve1*vp2-vp1*ve2;    /* new way to ensure optimum numerics */
 det2=ve2*vp3-vp2*ve3;
 det3=ve1*vp3-vp1*ve3;
 k=0; detmax=200;
 if((dm=fabs(det1)) > detmax){k=1; detmax=dm;}
 if((dm=fabs(det2)) > detmax){k=2; detmax=dm;}
 if((dm=fabs(det3)) > detmax){k=3; detmax=dm;}
 if(k == 0)return 0;
 else if(k == 1){
   *a=( vp2*q1-vp1*q2)/det1;
   *b=(-ve2*q1+ve1*q2)/det1;
 }
 else if(k == 2){
   *a=( vp3*q2-vp2*q3)/det2;
   *b=(-ve3*q2+ve2*q3)/det2;
 }
 else if(k == 3){
   *a=( vp3*q1-vp1*q3)/det3;
   *b=(-ve3*q1+ve1*q3)/det3;
 }
 else return 0;
/* a=1 at p2      b=1 at p1  */
 return 1;
}

static BOOL CALLBACK MapFromBrushDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 struct TEMP {int it; int id;} *tc;
 int i;
 HWND hctl;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     tc=(struct TEMP *)lparam;
     SetWindowLong(hdlg,GWL_USERDATA,lparam);
     EnableToolPannels(ALL_PANNELS,FALSE);
     if(tc->it == 0)
       SendDlgItemMessage(hdlg,DLG_MAPMAP_P,BM_SETCHECK,TRUE,0);
     else
       SendDlgItemMessage(hdlg,DLG_MAPMAP_C,BM_SETCHECK,TRUE,0);
     hctl=GetDlgItem(hdlg,DLG_MAPMAP_LIST);
     for(i=0;i<nImaps;i++){
       SendMessage(hctl,LB_ADDSTRING,0,(LPARAM)iMap[i].N);
     }
     SendDlgItemMessage(hdlg,DLG_MAPMAP_LIST,LB_SETCURSEL,0,0);
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      tc=(struct TEMP *)GetWindowLong(hdlg,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case DLG_MAPMAP_LIST:
          switch(HIWORD(wparam)){
            case LBN_DBLCLK:
              SendMessage(hdlg,WM_COMMAND,MAKEWPARAM(IDOK,0),0);
              break;
            default:
              break;
          }                                                                           break;
        case IDCANCEL:
          EndDialog(hdlg,FALSE);
          return(TRUE);
        case IDOK:
          if(SendDlgItemMessage(hdlg,DLG_MAPMAP_C,BM_GETCHECK,0,0))
               tc->it=1;
          else tc->it=0;
          hctl=GetDlgItem(hdlg,DLG_MAPMAP_LIST);
          tc->id=SendMessage(hctl,LB_GETCURSEL,0,0);
          EndDialog(hdlg,TRUE);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

void MapVerticesByBrush(void){
 struct TEMP {int it; int id;} tc;
 int id,it;
 vector n,vi,p0,p1,p2,p10,p20,dv;
 double nn,a,b;
 vertex *vp;
 long i;
 if(NvertSelect == 0){
   SendPrgmQuery(IDQ_OLD2,0);
   return;
 }
 if(nImaps == 0){
   SendPrgmQuery(IDQ_NOBRUSH,0);
   return;
 }
 tc.it=0; tc.id=0;
 if(!DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_MAPMAP),ghwnd_main,
    (DLGPROC)MapFromBrushDlgProc,(LPARAM)&tc))return;
 id=tc.id;
 it=tc.it;
 EDIT_ACTION=YES;
 if(it == 0){
   POINT2VECTOR(iMap[id].P,p1)   // this should ensure that the
   POINT2VECTOR(iMap[id].Y,p0)   // mapping coords is origin at bottom left
   POINT2VECTOR(iMap[id].X,p2)   //
   VECSUB(p1,p0,p10)             //
   VECSUB(p2,p1,p20)             //
   VECSUM(p0,p20,p2)             //
   CROSS(p20,p10,n)              //
   if(O_normalize(n) != OK)return;
   vp=MainVp; for(i=0;i<Nvert;i++){
     if(vp->status == SELECTED){
       VECSUB(p0,(double)vp->xyz,vi)
       nn=DOT(vi,n);
       VECSCALE(nn,n,dv)
       VECSUM((double)vp->xyz,dv,vi)
       if(E_MapToPlaneBrush(vi,p0,p1,p2,&a,&b)){
         if(vp->gp == 1){
           vp->x=(float)a;
           vp->y=(float)b;
         }
         else{
           vp->gp=1;
           vp->x=(float)a;
           vp->y=(float)b;
           NvertGlue++;
         }
       }
     }
     vp++;
   }
 }
 else{ /* cylindrical map */
   double theta,RTD=0.159154943;
   vector y,q;
   POINT2VECTOR(iMap[id].P,p0)
   POINT2VECTOR(iMap[id].Y,p1)
   POINT2VECTOR(iMap[id].X,p2)
   VECSUB(p1,p0,y)
   VECSUB(p2,p0,p20)
   CROSS(y,p20,n)
   if(O_normalize(n) != OK)return;
   vp=MainVp; for(i=0;i<Nvert;i++){
     if(vp->status == SELECTED){
       VECSUB(p0,(double)vp->xyz,p1)
       b=DOT(p1,y)/DOT(y,y);
       VECSCALE(b,y,p2)
       VECSUB(p1,p2,q)
       O_normalize(q);
       nn=DOT(q,n);
       if     (nn >  1.0)nn =  1.0;
       else if(nn < -1.0)nn = -1.0;
       theta=acos(nn)*RTD;
       CROSS(q,n,p1)
       if(DOT(p1,y) <= 0.0){
         if(theta <= 0.25) a = (0.25 - theta);
         else  a = (1.25 - theta);
       }
       else    a = (0.25 + theta);
       a *= 360.0/(double)iMap[id].Angle;
       if(vp->gp == 1){
         vp->x=(float)a;
         vp->y=(float)b;
       }
       else{
         vp->gp=1;
         vp->x=(float)a;
         vp->y=(float)b;
         NvertGlue++;
       }
     }
     vp++;
   }
 }
}

static BOOL CALLBACK MapOntoPathDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 char tempstr[16];
 struct TEMP {int type; long n;} *tc;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     tc=(struct TEMP *)lparam;
     SetWindowLong(hdlg,GWL_USERDATA,lparam);
     EnableToolPannels(ALL_PANNELS,FALSE);
     if(tc->type == 0)
       SendDlgItemMessage(hdlg,DLG_MAPCURVE_LIN,BM_SETCHECK,TRUE,0);
     else
       SendDlgItemMessage(hdlg,DLG_MAPCURVE_ONE,BM_SETCHECK,TRUE,0);
     sprintf(tempstr,"%d",tc->n);
     SendDlgItemMessage(hdlg,DLG_MAPCURVE_N,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hdlg,DLG_MAPCURVE_N,EM_LIMITTEXT,(WPARAM)4,0);
     SendDlgItemMessage(hdlg,DLG_MAPCURVE_NS,SPNM_SETRANGE,0,
                        MAKELPARAM(1,9999));
     SendDlgItemMessage(hdlg,DLG_MAPCURVE_NS,SPNM_SETCRNTVALUE,
                        (WPARAM)tc->n,0);
     SendDlgItemMessage(hdlg,DLG_MAPCURVE_NS,SPNM_SETEDITCTRL,0,
                        (LPARAM)GetDlgItem(hdlg,DLG_MAPCURVE_N));
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      tc=(struct TEMP *)GetWindowLong(hdlg,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hdlg,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hdlg,DLG_MAPCURVE_N,tempstr,10) == 0){
            EndDialog(hdlg,FALSE);
            return(TRUE);
          }
          tc->n=atoi(tempstr);
          if(SendDlgItemMessage(hdlg,DLG_MAPCURVE_ONE,BM_GETCHECK,0,0))
               tc->type=1;
          else tc->type=0;
          EndDialog(hdlg,TRUE);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

void MapOntoPath(void){
 struct TEMP {int type; long n;} tc;
 vertex *Vp,*Va,*Va1;
 long i,*va,vp,np,n;
 int id;
 double a,l,ll;
 tc.n=1;
 tc.type=0;
 if(!DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_MAPCURVE),ghwnd_main,
    (DLGPROC)MapOntoPathDlgProc,(LPARAM)&tc))return;
 id=tc.type;
 n=max(1,tc.n);
 EDIT_ACTION=YES;
 if((va=GetPathList(&np)) != NULL){
   if(id == 0){
     l=0; /* get length of path */
     for(i=1;i<np;i++){
       Va=(MainVp+va[i]); Va1=(MainVp+va[i-1]);
       l += sqrt(
       (double)(Va->xyz[0] - Va1->xyz[0])*
       (double)(Va->xyz[0] - Va1->xyz[0])+
       (double)(Va->xyz[1] - Va1->xyz[1])*
       (double)(Va->xyz[1] - Va1->xyz[1])+
       (double)(Va->xyz[2] - Va1->xyz[2])*
       (double)(Va->xyz[2] - Va1->xyz[2]));
     }
   }
   a=0.0; ll=0.0;
   for(i=0;i<np;i++){
     Vp=(MainVp+va[i]);
     if(Vp->gp == 1){
       Vp->x=(float)a;
       Vp->y=0.0;
     }
     else{
       Vp->gp=1;
       Vp->x=(float)a;
       Vp->y=0.0;
       NvertGlue++;
     }
     if(id == 0){
       if(i < np-1){
         Va=(MainVp+va[i]); Va1=(MainVp+va[i+1]);
         ll += sqrt(
         (double)(Va1->xyz[0] - Va->xyz[0])*
         (double)(Va1->xyz[0] - Va->xyz[0])+
         (double)(Va1->xyz[1] - Va->xyz[1])*
         (double)(Va1->xyz[1] - Va->xyz[1])+
         (double)(Va1->xyz[2] - Va->xyz[2])*
         (double)(Va1->xyz[2] - Va->xyz[2]));
       }
       a = ll/l*(double)n;
     }
     else a += 1.0;
     if(Vp->status == DESELECTED){ /* so that we don't need to redraw */
       Vp->status=SELECTED;
       NvertSelect++; NvertDeselect--;
     }
   }
   X__Free(va);
   DrawVerticesOnly(NULL);
   UpdateCounters();
 }
}

static BOOL CALLBACK SetMapCoordsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 char tempstr[16];
 struct TEMP {int type; double v;} *tc;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     tc=(struct TEMP *)lparam;
     SetWindowLong(hdlg,GWL_USERDATA,lparam);
     EnableToolPannels(ALL_PANNELS,FALSE);
     if(tc->type == 0)
       SendDlgItemMessage(hdlg,DLG_MAPCOORD_H,BM_SETCHECK,TRUE,0);
     else
       SendDlgItemMessage(hdlg,DLG_MAPCOORD_V,BM_SETCHECK,TRUE,0);
     sprintf(tempstr,"%.2lf",tc->v);
     SendDlgItemMessage(hdlg,DLG_MAPCOORD_VALUE,WM_SETTEXT,0,(LPARAM)tempstr);
     SendDlgItemMessage(hdlg,DLG_MAPCOORD_VALUE,EM_LIMITTEXT,(WPARAM)12,0);
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_DESTROY:
     EnableToolPannels(ALL_PANNELS,TRUE);
     break;
   case WM_COMMAND:
      tc=(struct TEMP *)GetWindowLong(hdlg,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hdlg,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hdlg,DLG_MAPCOORD_VALUE,tempstr,10) == 0){
            EndDialog(hdlg,FALSE);
            return(TRUE);
          }
          tc->v=atof(tempstr);
          if(SendDlgItemMessage(hdlg,DLG_MAPCOORD_V,BM_GETCHECK,0,0))
               tc->type=1;
          else tc->type=0;
          EndDialog(hdlg,TRUE);
          return(TRUE);
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

void SetMapCoords(void){
 int it;
 double v;
 vertex *vp;
 long i;
 struct TEMP {int type; double v;} tc;
 if(NvertSelect == 0){
   SendPrgmQuery(IDQ_OLD2,0);
   return;
 }
 tc.type=0; tc.v=1.0;
 if(!DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_MAPCOORD),ghwnd_main,
    (DLGPROC)SetMapCoordsDlgProc,(LPARAM)&tc))return;
 it=tc.type;
 v=tc.v;
 EDIT_ACTION=YES;
 if(fabs(v) < 0.01)v=1.0;
 vp=MainVp; for(i=0;i<Nvert;i++){
   if(vp->status == SELECTED && vp->gp == 1){
     if(it == 0)vp->x=v;
     else       vp->y=v;
   }
   vp++;
 }
}

void  ExecuteLockAction(int item, int sb){
 long k,dX[3],dY[3];
 if(sb >= 0){
   if(sb < 0 || sb >= nImaps)return;
   DrawBrush(sb);
   switch(item){
     case IDM_LOCK_SQUARE:
       for(k=0;k<3;k++){
         dX[k]=iMap[sb].X[k]-iMap[sb].P[k];
         dY[k]=iMap[sb].Y[k]-iMap[sb].P[k];
       }
       iMap[sb].P[0] = NpointerX;
       iMap[sb].P[1] = NpointerY;
       iMap[sb].P[2] = NpointerZ;
       for(k=0;k<3;k++){
         iMap[sb].X[k] = iMap[sb].P[k]+dX[k];
         iMap[sb].Y[k] = iMap[sb].P[k]+dY[k];
       }
       break;
     case IDM_LOCK_CIRCLE:
       iMap[sb].X[0] = NpointerX;
       iMap[sb].X[1] = NpointerY;
       iMap[sb].X[2] = NpointerZ;
       break;
     case IDM_LOCK_TRIANGLE:
       iMap[sb].Y[0] = NpointerX;
       iMap[sb].Y[1] = NpointerY;
       iMap[sb].Y[2] = NpointerZ;
       break;
     case IDM_LOCK_LOCK:{
         double d,dmin;
         long vp,vl[3];
         long n=0,i,id[3];
         vertex *Vp;
         if(iMap[sb].Lock){ /* if locked then unlock */
           iMap[sb].Lock = 0; nLockediMaps--;
           iMap[sb].Vlock[0]=iMap[sb].Vlock[1]=iMap[sb].Vlock[2]= -1;
         }
         else{
           if(NvertSelect == 3){
             vp=0; while(vp < Nvert){
               if((MainVp+vp)->status == SELECTED)vl[n++]=vp;
               if(n == 3)break;
               vp++;
             }
             if(n != 3)return; /* just in case ! */
             id[0]=0; dmin=1.0e35;
             for(i=0;i<3;i++){
               if((d=Distance3D(iMap[sb].P,(MainVp+vl[i])->xyz)) < dmin){
                 id[0]=i; dmin=d;
               }
             }
             if(id[0] == 0){
               id[1]=1; id[2]=2;
               dmin=Distance3D(iMap[sb].X,(MainVp+vl[1])->xyz);
               if(Distance3D(iMap[sb].X,(MainVp+vl[2])->xyz) < dmin){
                 id[1]=2; id[2]=1;
               }
             }
             else if(id[0] == 1){
               id[1]=0; id[2]=2;
               dmin=Distance3D(iMap[sb].X,(MainVp+vl[0])->xyz);
               if(Distance3D(iMap[sb].X,(MainVp+vl[2])->xyz) < dmin){
                 id[1]=2; id[2]=0;
               }
             }
             else if(id[0] == 2){
               id[1]=0; id[2]=1;
               dmin=Distance3D(iMap[sb].X,(MainVp+vl[0])->xyz);
               if(Distance3D(iMap[sb].X,(MainVp+vl[1])->xyz) < dmin){
                 id[1]=1; id[2]=0;
               }
             }
             for(i=0;i<3;i++)iMap[sb].Vlock[i]=vl[id[i]];
             iMap[sb].Lock=1; nLockediMaps++;
             for(i=0;i<3;i++){
               iMap[sb].P[i] = (MainVp+(iMap[sb].Vlock[0]))->xyz[i];
               iMap[sb].X[i] = (MainVp+(iMap[sb].Vlock[1]))->xyz[i];
               iMap[sb].Y[i] = (MainVp+(iMap[sb].Vlock[2]))->xyz[i];
             }
           }
           else SendPrgmQuery(IDQ_MAPNOLOCK,0);
         }
       }
       break;
     default: break;
   }
   DrawBrush(sb);
 }
 else { /* shader */
   sb=SelectedShader;
   if(sb < 0 || sb >= nSha)return;
   DrawShader(sb);
   switch(item){
     case IDM_LOCK_SQUARE:
       for(k=0;k<3;k++){
         dX[k]=iSha[sb].X[k]-iSha[sb].P[k];
         dY[k]=iSha[sb].Y[k]-iSha[sb].P[k];
       }
       iSha[sb].P[0] = NpointerX;
       iSha[sb].P[1] = NpointerY;
       iSha[sb].P[2] = NpointerZ;
       for(k=0;k<3;k++){
         iSha[sb].X[k] = iSha[sb].P[k]+dX[k];
         iSha[sb].Y[k] = iSha[sb].P[k]+dY[k];
       }
       break;
     case IDM_LOCK_CIRCLE:
       iSha[sb].X[0] = NpointerX;
       iSha[sb].X[1] = NpointerY;
       iSha[sb].X[2] = NpointerZ;
       break;
     case IDM_LOCK_TRIANGLE:
       iSha[sb].Y[0] = NpointerX;
       iSha[sb].Y[1] = NpointerY;
       iSha[sb].Y[2] = NpointerZ;
       break;
     case IDM_LOCK_LOCK:
       MessageBeep(MB_OK);
       break;
     default: break;
   }
   DrawShader(sb);
 }
}
