/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file MATERIAL.C */

// Handle surface attributes shaders and materials

#define MODULE_ATTRIBS 1

#include "design.h"

#include "faceattr.h"

static int GetFaceAttribs(FACEATTRIBS *fa, int defaults);
static void UpdateAttribsDialog(HWND, int, FACEATTRIBS *, FACEATTRIBS *);
static void PutFaceAttribute(int what, FACEATTRIBS *f);
static BOOL CALLBACK SetExternalShaderDlgProc(HWND hdlg,UINT msg,
                                    WPARAM wparam,LPARAM lparam);
static void UpdateShaderLoadDialog(int id,HWND hdlg);
static void UpdateShaderSetDialog(int sa, int id, HWND hdlg);
static void LoadExternalShader(int id, HWND hdlg);
static int  EditExternalShader(int id, HWND hdlg);
static void DropExternalShader(int id, HWND hdlg);
static void DrawShaderPreview(HDC hdc,RECT rc);

static FACEATTRIBS FaceAttribs,DefAttribs;
static int nf;

static int GetFaceAttribs(FACEATTRIBS *fa, int defaults){
 unsigned char lfc[3],lsc[3];
 int ff,lsm,lbr,lgl,ltr,lrf,lsh,lsa,lbu,lba,lbt,lbs,lbc,lpa,lgt,lst;
 face *fp;
 vertex *V0,*V1,*V2;
 int i,j,nff=0;
 fa->flag_fc = 1;
 fa->flag_sc = 1;                    /* colours      */
 fa->flag_sm = 1;
 fa->flag_br = 1;                    /* brill on/off */
 fa->flag_gl = 1;                    /* glossiness   */
 fa->flag_gt = 1;                    /* glass type   */
 fa->flag_tr = 1;                    /* !opacity     */
 fa->flag_rf = 1;                    /* reflectivity */
 fa->flag_pa = 1;                    /* parameter    */
 fa->flag_sh = 1;                    /* shader       */
 fa->flag_sa = 1;                    /* shader axis  */
 fa->flag_st = 1;                    /* shader type  */
 fa->flag_bu = 1;                    /* map ID       */
 fa->flag_ba = 1;                    /* map active   */
 fa->flag_bt = 1;                    /* map tiled    */
 fa->flag_bs = 1;                    /* map shaded   */
 fa->flag_bc = 1;                    /* map clear    */
 for(i=0;i<3;i++){
   fa->fc[i]=LastColour[i];
   fa->sc[i]=ShaderColour[i];
 }
 fa->sm=  DefaultSmoothing;
 fa->br=  DefaultShiny;
 fa->gl=  DefaultBrilliance;
 fa->tr=  DefaultTransp;
 fa->gt=  0;
 fa->rf=  DefaultReflectivity;
 fa->pa=  DefaultParameter;
 fa->sh=  DefaultTexture;
 fa->sa=  0;
 fa->st=  0;
 fa->bu=  0;   /* map ID       */
 fa->ba=  0;   /* map active   */
 fa->bt=  0;   /* map tiled    */
 fa->bs=  0;   /* map shaded   */
 fa->bc=  0;   /* clear zero   */
 if(defaults)return 0;
 ff=1;
 if((fp=MainFp) != NULL)for(j=0;j<Nface;j++){
   V0=(MainVp+fp->V[0]); V1=(MainVp+fp->V[1]); V2=(MainVp+fp->V[2]);
   if(V0->status == SELECTED &&
      V1->status == SELECTED &&
      V2->status == SELECTED){
     nff++; /* make sure that at least one face is selected */
     fa->gl = (fp->matcol[0] & 0xf0) >> 4;   /* gloss value      */
     fa->tr = (fp->matcol[1] & 0xf0) >> 4;   /* glass value      */
     fa->rf = (fp->matcol[2] & 0xf0) >> 4;   /* reflection value */
     if((fp->attrib & 0x80) == 0x80)fa->sm=1; else fa->sm=0; /* smoothing on  */
     if((fp->attrib & 0x40) == 0x40)fa->ba=1; else fa->ba=0; /* bit 6 = 1 if map active */
     if((fp->attrib & 0x20) == 0x20)fa->br=1; else fa->br=0; /* brilliance on */
     if((fp->attrib & 0x10) == 0x10)fa->st=1; else fa->st=0; /* extern shader */
     fa->sh=fp->shader;
     fa->sa=fp->axis;
     fa->bu=fp->map;
     fa->pa=fp->param;
     if((fp->attrib & 0x08) == 0x08)fa->bt=1; else fa->bt=0;
     if((fp->attrib & 0x04) == 0x04)fa->bs=1; else fa->bs=0;
     if((fp->attrib & 0x02) == 0x02)fa->bc=1; else fa->bc=0;
     if((fp->attrib & 0x01) == 0x01)fa->gt=1; else fa->gt=0;
     fa->fc[0]=fp->color[0];
     fa->fc[1]=fp->color[1];
     fa->fc[2]=fp->color[2];
     fa->sc[0]=(fp->matcol[0] & 0xf) << 4;
     fa->sc[1]=(fp->matcol[1] & 0xf) << 4;
     fa->sc[2]=(fp->matcol[2] & 0xf) << 4;
     if(ff){
       for(i=0;i<3;i++){lfc[i]=fa->fc[i]; lsc[i]=fa->sc[i];}
       lsm=fa->sm;
       lbr=fa->br; lgl=fa->gl; ltr=fa->tr; lgt=fa->gt; lrf=fa->rf; lpa=fa->pa;
       lsh=fa->sh; lsa=fa->sa; lst=fa->st;
       lbu=fa->bu; lba=fa->ba;
       lbt=fa->bt; lbs=fa->bs; lbc=fa->bc;
       ff=0;
     }
     else{
       for(i=0;i<3;i++){
         if(lfc[i] != fa->fc[i])fa->flag_fc = -1;
         if(lsc[i] != fa->sc[i])fa->flag_sc = -1;
       }
       if(lsm != fa->sm)fa->flag_sm = -1;
       if(lbr != fa->br)fa->flag_br = -1;
       if(lgl != fa->gl)fa->flag_gl = -1;
       if(ltr != fa->tr)fa->flag_tr = -1;
       if(lgt != fa->gt)fa->flag_gt = -1;
       if(lrf != fa->rf)fa->flag_rf = -1;
       if(lpa != fa->pa)fa->flag_pa = -1;
       if(lsh != fa->sh)fa->flag_sh = -1;
       if(lsa != fa->sa)fa->flag_sa = -1;
       if(lst != fa->st)fa->flag_st = -1;
       if(lbu != fa->bu)fa->flag_bu = -1;
       if(lba != fa->ba)fa->flag_ba = -1;
       if(lbt != fa->bt)fa->flag_bt = -1;
       if(lbs != fa->bs)fa->flag_bs = -1;
       if(lbc != fa->bc)fa->flag_bc = -1;
     }
   }
   fp++;
 }
 return nff;
}

static void UpdateMapAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 int active;
 char *map_type_list[]={"Still","Animate"};
 char *map_attr_list[]={"Plane","Cylinder","Moziac-P","Moziac-C","Fixed To"};
 char str[64];
 if(nff != 0 && nImaps > 0 && (fa->flag_ba < 0))active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_MAP),active);
 if(nff != 0 && nImaps > 0 && !(fa->flag_ba < 0 || fa->ba == 0) && (fa->flag_bt < 0))active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_MAP1),active);
 if(nff != 0 && nImaps > 0 && !(fa->flag_ba < 0 || fa->ba == 0) && (fa->flag_bs < 0))active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_MAP2),active);
 if(nff != 0 && nImaps > 0 && !(fa->flag_ba < 0 || fa->ba == 0) && (fa->flag_bc < 0))active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_MAP3),active);
 if(nff == 0){
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_SET),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DROP),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_MOVE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_EDIT),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DELETE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_TILE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_SINGLE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_SHADED),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_UNSHADED),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_OPAQUE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_CLEAR),FALSE);
   SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_NAME,(LPCTSTR)"<No faces>");
   SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_TYPE,(LPCTSTR)"  ");
   return;
 }
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_SET),TRUE);
 if(fa->flag_ba < 0 || fa->ba == 0 || fa->flag_bu < 0 || nImaps == 0)active=FALSE;
 else                                                                active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_TILE),active);
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_SINGLE),active);
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_SHADED),active);
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_UNSHADED),active);
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_OPAQUE),active);
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_CLEAR),active);
 if(fa->flag_ba < 0){ /* brush activity is not consistant */
   SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_NAME,(LPCTSTR)"<Some faces have map>");
   SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_TYPE,(LPCTSTR)"  ");
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DROP),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_MOVE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_EDIT),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DELETE),FALSE);
 }
 else{
   if(fa->ba == 0 || nImaps == 0){
     SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_NAME,(LPCTSTR)"<No map>");
     SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_TYPE,(LPCTSTR)"  ");
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DROP),FALSE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_MOVE),FALSE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_EDIT),FALSE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DELETE),FALSE);
   }
   else if(fa->flag_bu < 0){
     SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_NAME,(LPCTSTR)"<Not all same map>");
     SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_TYPE,(LPCTSTR)"  ");
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DROP),TRUE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_MOVE),FALSE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_EDIT),FALSE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DELETE),FALSE);
   }
   else{
     SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_NAME,(LPCTSTR)iMap[fa->bu].N);
     sprintf(str,"%s and %s",map_type_list[min((int)iMap[fa->bu].Type,1)],
                                  map_attr_list[(int)(iMap[fa->bu].Map)] );
     SetDlgItemText(hdlg,DLG_ATTRIBS_MAP_TYPE,(LPCTSTR)str);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DROP),TRUE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_MOVE),TRUE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_EDIT),TRUE);
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_MAP_DELETE),TRUE);
   }
 }
 if(fa->flag_ba < 0 || fa->ba == 0 || fa->flag_bu < 0 ||
    fa->flag_bt < 0 || nImaps == 0){
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_SINGLE,BM_SETCHECK,0,0);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_TILE,BM_SETCHECK,0,0);
 }
 else{
   if(fa->bt)active=TRUE; else active=FALSE;
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_SINGLE,BM_SETCHECK,(WPARAM)!active,0);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_TILE,BM_SETCHECK,(WPARAM)active,0);
 }
 if(fa->flag_ba < 0 || fa->ba == 0 || fa->flag_bu < 0 ||
    fa->flag_bs < 0 || nImaps == 0){
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_SHADED,BM_SETCHECK,0,0);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_UNSHADED,BM_SETCHECK,0,0);
 }
 else{
   if(fa->bs)active=TRUE; else active=FALSE;
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_SHADED,BM_SETCHECK,(WPARAM)!active,0);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_UNSHADED,BM_SETCHECK,(WPARAM)active,0);
 }
 if(fa->flag_ba < 0 || fa->ba == 0 || fa->flag_bu < 0 ||
    fa->flag_bc < 0 || nImaps == 0){
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_OPAQUE,BM_SETCHECK,0,0);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_CLEAR,BM_SETCHECK,0,0);
 }
 else{
   if(fa->bc)active=TRUE; else active=FALSE;
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_OPAQUE,BM_SETCHECK,(WPARAM)!active,0);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_MAP_CLEAR,BM_SETCHECK,(WPARAM)active,0);
 }
}

static void UpdateShaderAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
 if(nff != 0 && (fa->flag_sh < 0 || fa->flag_st < 0))active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_SHA),active);
 if(nff == 0){
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_SET),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_INT),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_EXT),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_XCHOOSE),FALSE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_XMOVE),FALSE);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_INT,BM_SETCHECK,BST_INDETERMINATE,0);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_EXT,BM_SETCHECK,BST_INDETERMINATE,0);
   SetDlgItemText(hdlg,DLG_ATTRIBS_SHA_NAME,(LPCTSTR)"<No faces>");
 }
 else{
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_SET),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_INT),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_EXT),TRUE);
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_XCHOOSE),TRUE);
   if(fa->flag_st < 0){
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_INT,BM_SETCHECK,BST_INDETERMINATE,0);
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_EXT,BM_SETCHECK,BST_INDETERMINATE,0);
     SetDlgItemText(hdlg,DLG_ATTRIBS_SHA_NAME,(LPCTSTR)"<Not same>");
   }
   else{
     if(fa->st == 1){ // external
       SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_INT,BM_SETCHECK,BST_UNCHECKED,0);
       if(fa->flag_sh < 0)
         SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_EXT,BM_SETCHECK,BST_INDETERMINATE,0);
       else
         SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_EXT,BM_SETCHECK,BST_CHECKED,0);
       SetDlgItemText(hdlg,DLG_ATTRIBS_SHA_NAME,(LPCTSTR)"External");
     }
     else{  // internal
       SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_EXT,BM_SETCHECK,BST_UNCHECKED,0);
       if(fa->flag_sh < 0){
         SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_INT,BM_SETCHECK,BST_INDETERMINATE,0);
         SetDlgItemText(hdlg,DLG_ATTRIBS_SHA_NAME,(LPCTSTR)"<Not same>");
       }
       else{
         SendDlgItemMessage(hdlg,DLG_ATTRIBS_SHA_INT,BM_SETCHECK,BST_CHECKED,0);
         SetDlgItemText(hdlg,DLG_ATTRIBS_SHA_NAME,(LPCTSTR)AlTextList[fa->sh]);
       }
     }
   }

   if(fa->flag_sa < 0){
     EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_XMOVE),FALSE);
     SetDlgItemText(hdlg,DLG_ATTRIBS_SHA_XNAME,(LPCTSTR)"<Not same>");
   }
   else{
     SetDlgItemText(hdlg,DLG_ATTRIBS_SHA_XNAME,(LPCTSTR)iSha[fa->sa].N);
     if(nSha > 1)EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_XCHOOSE),TRUE);
     else        EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SHA_XCHOOSE),FALSE);
   }
 }
}

static void UpdateSmoothAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
// if(nff != 0 && fa->flag_sm < 0)active=SW_SHOW; else active=SW_HIDE;
// ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_SMOOTH),active);
 if(nff == 0){
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SMOOTH_ON),FALSE);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_SMOOTH_ON, BM_SETCHECK,BST_INDETERMINATE,0);
 }
 else {
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SMOOTH_ON),TRUE);
   if( fa->flag_sm < 0){
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_SMOOTH_ON, BM_SETCHECK,BST_INDETERMINATE,0);
   }
   else{
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_SMOOTH_ON, BM_SETCHECK,fa->sm,0);
   }
 }
}

static void UpdateColourAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 int active;
 if(nff != 0 && fa->flag_fc < 0)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_FC),active);
 if(nff != 0)active=TRUE; else active=FALSE;
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_FC_EDIT),active);
 if(nff != 0){
   VECCOPY(FaceAttribs.fc,DefAttribs.fc)
   VECCOPY(DefAttribs.fc,LastColour)
 }
 InvalidateRect(GetDlgItem(hdlg,DLG_ATTRIBS_FC_EDIT),NULL,FALSE);
}

static void UpdateShaderColourAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 int active;
 if(nff != 0 && fa->flag_sc < 0)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_SC),active);
 if(nff != 0)active=TRUE; else active=FALSE;
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_SC_EDIT),active);
 if(nff != 0){
   VECCOPY(FaceAttribs.sc,DefAttribs.sc)
   VECCOPY(DefAttribs.sc,LastColour)
 }
 InvalidateRect(GetDlgItem(hdlg,DLG_ATTRIBS_SC_EDIT),NULL,FALSE);
}

static void UpdateBrillAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
// if(nff != 0 && fa->flag_br < 0)active=SW_SHOW; else active=SW_HIDE;
// ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_SPEC),active);
 if(nff == 0){
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_GLS_ON),FALSE);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_GLS_ON, BM_SETCHECK,BST_INDETERMINATE,0);
 }
 else {
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_GLS_ON),TRUE);
   if( fa->flag_br < 0){
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_GLS_ON, BM_SETCHECK,BST_INDETERMINATE,0);
   }
   else{
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_GLS_ON, BM_SETCHECK,fa->br,0);
   }
 }
}

static void UpdateGlassTypeAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
// if(nff != 0 && fa->flag_gt < 0)active=SW_SHOW; else active=SW_HIDE;
// ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_TRNE),active);
 if(nff == 0){
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_TRN_TYPE1),FALSE);
   SendDlgItemMessage(hdlg,DLG_ATTRIBS_TRN_TYPE1,BM_SETCHECK,BST_INDETERMINATE,0);
 }
 else {
   EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_TRN_TYPE1),TRUE);
   if( fa->flag_gt < 0){
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_TRN_TYPE1,BM_SETCHECK,BST_INDETERMINATE,0);
   }
   else{
     SendDlgItemMessage(hdlg,DLG_ATTRIBS_TRN_TYPE1,BM_SETCHECK,fa->gt,0);
   }
 }
}

static void UpdateGlossAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
 char str[32];
 if(nff != 0 && fa->flag_gl < 0)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_GLS),active);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_GLS_SCROLL,TBM_SETPOS,TRUE,da->gl);
 sprintf(str,"%ld",da->gl);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_GLS_EDIT,WM_SETTEXT,0,(LPARAM)str);
 if(nff == 0)active=FALSE; else active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_GLS_SCROLL),active);
}

static void UpdateTransAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
 char str[32];
 if(nff != 0 && fa->flag_tr < 0)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_TRN),active);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_TRN_SCROLL,TBM_SETPOS,TRUE,da->tr);
 sprintf(str,"%ld",da->tr);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_TRN_EDIT,WM_SETTEXT,0,(LPARAM)str);
 if(nff == 0)active=FALSE; else active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_TRN_SCROLL),active);
}

static void UpdateParamAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
 char str[32];
 if(nff != 0 && fa->flag_pa < 0)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_PRM),active);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_PRM_SCROLL,TBM_SETPOS,TRUE,da->pa);
 sprintf(str,"%ld",da->pa);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_PRM_EDIT,WM_SETTEXT,0,(LPARAM)str);
 if(nff == 0)active=FALSE; else active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_PRM_SCROLL),active);
}

static void UpdateReflAttribsSection(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 BOOL active;
 char str[32];
 if(nff != 0 && fa->flag_rf < 0)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,DLG_ATTRIBS_WARN_REF),active);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_REF_SCROLL,TBM_SETPOS,TRUE,da->rf);
 sprintf(str,"%ld",da->rf);
 SendDlgItemMessage(hdlg,DLG_ATTRIBS_REF_EDIT,WM_SETTEXT,0,(LPARAM)str);
 if(nff == 0)active=FALSE; else active=TRUE;
 EnableWindow(GetDlgItem(hdlg,DLG_ATTRIBS_REF_SCROLL),active);
}

static void UpdateTitleDialog(HWND hdlg,int nff){
 char str[128];
 char ltext[256];
 LoadString(ghinst_main,IDX_MISC_SETATTRIBS,ltext,256);
 sprintf(str,ltext,nff);
 SendMessage(hdlg,WM_SETTEXT,0,(LPARAM)str);
}

static void UpdateAttribsDialog(HWND hdlg, int nff,
                                FACEATTRIBS *fa, FACEATTRIBS *da){
 UpdateTitleDialog(hdlg,nff);
 UpdateColourAttribsSection(hdlg,nff,fa,da);
 UpdateShaderColourAttribsSection(hdlg,nff,fa,da);
 UpdateGlossAttribsSection(hdlg,nff,fa,da);
 UpdateBrillAttribsSection(hdlg,nff,fa,da);
 UpdateGlassTypeAttribsSection(hdlg,nff,fa,da);
 UpdateTransAttribsSection(hdlg,nff,fa,da);
 UpdateReflAttribsSection(hdlg,nff,fa,da);
 UpdateParamAttribsSection(hdlg,nff,fa,da);
 UpdateSmoothAttribsSection(hdlg,nff,fa,da);
 UpdateShaderAttribsSection(hdlg,nff,fa,da);
 UpdateMapAttribsSection(hdlg,nff,fa,da);
}

static void PutFaceAttribute(int what, FACEATTRIBS *f){
/* what   0 = face colour
          1 = shader colour
          2 = gloss
          3 = trans
          4 = reflectance
          5 = smoothing
          6 = brilliance switch
          7 = shader ID
          8 = dump map
          9 = shader axis
         10 = set brush ID;
         11 = set tiled
         12 = set shaded
         13 = set clear
         14 = set parameter
         15 = set glass type
*/
 face *fp;
 vertex *V0,*V1,*V2;
 int i;
 EDIT_ACTION=YES;
 if((fp=MainFp) != NULL)for(i=0;i<Nface;i++){
   V0=(MainVp+fp->V[0]); V1=(MainVp+fp->V[1]); V2=(MainVp+fp->V[2]);
   if(  (V0->status == SELECTED)
     && (V1->status == SELECTED)
     && (V2->status == SELECTED)){
     if(what == 0){
       VECCOPY(f->fc,fp->color)
     }
     else if(what == 1){
       fp->matcol[0] = (fp->matcol[0] & 0xf0) + f->sc[0]/16;
       fp->matcol[1] = (fp->matcol[1] & 0xf0) + f->sc[1]/16;
       fp->matcol[2] = (fp->matcol[2] & 0xf0) + f->sc[2]/16;
     }
     else if(what == 2){ /* gloss */
       fp->matcol[0] = fp->matcol[0] & 0x0f;  /* clear 4 bits at top */
       fp->matcol[0] |= ( (unsigned char)(f->gl) << 4);
     }
     else if(what == 3){ /* trans ( bump is now parameter ) */
       fp->matcol[1] = fp->matcol[1] & 0x0f;  /* clear 4 bits at top */
       fp->matcol[1] |= ( (unsigned char)(f->tr) << 4);
     }
     else if(what == 4){ /* refl  */
       fp->matcol[2] = fp->matcol[2] & 0x0f;  /* clear 4 bits at top */
       fp->matcol[2] |= ( (unsigned char)(f->rf) << 4);
     }
     else if(what == 5){ /* smoothing */
       if(f->sm == 1)fp->attrib |= 0x80; /* Set Smoothing bit #7            */
       else          fp->attrib &= 0x7f; /* Clear Smoothing bit in faces #7 */
     }
     else if(what == 6){ /* brilliant switch */
       if(f->br == 1)fp->attrib |= 0x20;  /* Set brilliant bit #5            */
       else          fp->attrib &= 0xdf;  /* Clear brilliant bit in faces #5 */
     }
     else if(what == 7){ /* set shader value */
       fp->shader = (unsigned char)f->sh;
       if(f->st == 1)fp->attrib |= 0x10; /* Set external shader bit */
       else          fp->attrib &= 0xef;
     }
     else if(what == 8){ /* dump map - clear bit 6 */
       fp->attrib  &= 0xBF;
       fp->map=0;
     }
     else if(what == 9){ /* set shader axis ID */
       fp->axis  = (unsigned char)f->sa;
     }
     else if(what == 10){ /* set map */
       fp->attrib = (fp->attrib | 0x40);/* set bit 6 to flag map in use  */
       fp->map    = (unsigned char)f->bu;
     }
     else if(what == 11){ /* set tiled */
       if(f->bt == 1)fp->attrib |= 0x08;    /* bit 3 to 1 for tiled map  */
       else          fp->attrib &= 0xf7;    /* bit 3 clear */
     }
     else if(what == 12){ /* map shading */
       if(f->bs == 1)fp->attrib |= 0x04;    /* bit 2 to 1 for unshaded map */
       else          fp->attrib &= 0xfb;    /* bit 2 clear */
     }
     else if(what == 13){ /* map use clear colour */
       if(f->bc == 1)fp->attrib |= 0x02;    /* bit 1 to 1 if clear colour  */
       else          fp->attrib &= 0xfd;    /* bit 1 clear */
     }
     else if(what == 14){ /* set parameter */
       fp->param = (unsigned char)f->pa;
     }
     else if(what == 15){ /* set glass type */
       if(f->gt == 1)fp->attrib |= 0x01;    /* bit 0 to 1 for rounded glass */
       else          fp->attrib &= 0xfe;    /* bit 0 clear */
     }
   }
   fp++;
 }
 PutNurbsAttribute(what,f);
}

BOOL CALLBACK DefaultAttribsDlgProc(HWND hdlg,UINT msg,
                             WPARAM wparam,LPARAM lparam){
 HWND hctl;
 BOOL err;
 int i,iMin,iMax;
 LPDRAWITEMSTRUCT lpdis;
 static FACEATTRIBS DefAttribs;
 static int nf;
 FACEATTRIBS *da;
 char str[32];
 da=&DefAttribs;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     GetFaceAttribs(da,1);
     for(i=0;i<3;i++){
       da->fc[i]=FaceColour[i];
       da->sc[i]=ShaderColour[i];
     }
     SetScrollRange(GetDlgItem(hdlg,DLG_DEFAULT_GLS_SCROLL),SB_CTL,0,15,FALSE);
     SetScrollRange(GetDlgItem(hdlg,DLG_DEFAULT_BMP_SCROLL),SB_CTL,0,15,FALSE);
     SetScrollRange(GetDlgItem(hdlg,DLG_DEFAULT_REF_SCROLL),SB_CTL,0,15,FALSE);
     SetScrollRange(GetDlgItem(hdlg,DLG_DEFAULT_TRN_SCROLL),SB_CTL,0,15,FALSE);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_SMOOTH_OFF,BM_SETCHECK,da->sm^1,0);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_SMOOTH_ON, BM_SETCHECK,da->sm,  0);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_GLS_OFF,BM_SETCHECK,da->br^1,0);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_GLS_ON, BM_SETCHECK,da->br,  0);
     SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_GLS_SCROLL),SB_CTL,da->gl,TRUE);
     sprintf(str,"%ld",da->gl);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_GLS_EDIT,WM_SETTEXT,0,(LPARAM)str);
     SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_BMP_SCROLL),SB_CTL,da->pa,TRUE);
     sprintf(str,"%ld",da->pa);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_BMP_EDIT,WM_SETTEXT,0,(LPARAM)str);
     SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_REF_SCROLL),SB_CTL,da->rf,TRUE);
     sprintf(str,"%ld",da->rf);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_REF_EDIT,WM_SETTEXT,0,(LPARAM)str);
     SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_TRN_SCROLL),SB_CTL,da->tr,TRUE);
     sprintf(str,"%ld",da->tr);
     SendDlgItemMessage(hdlg,DLG_DEFAULT_TRN_EDIT,WM_SETTEXT,0,(LPARAM)str);
     CentreDialogOnScreen(hdlg);
     return (TRUE);
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_DEFAULT_FC_EDIT ||
          lpdis->CtlID == DLG_DEFAULT_SC_EDIT){
         if(lpdis->CtlID == DLG_DEFAULT_FC_EDIT){
           r=da->fc[0]; g=da->fc[1]; b=da->fc[2];
         }
         else {
           r=da->sc[0]; g=da->sc[1]; b=da->sc[2];
         }
         if(lpdis->itemState & ODS_SELECTED)
           InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_HSCROLL:
     if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_DEFAULT_GLS_SCROLL))){
       GetScrollRange(hctl,SB_CTL,&iMin,&iMax);
       switch (LOWORD(wparam)) {
         case SB_PAGEDOWN:
         case SB_LINEDOWN: da->gl = min(iMax,da->gl+1); break;
         case SB_PAGEUP:
         case SB_LINEUP:   da->gl = max(iMin,da->gl-1); break;
         case SB_THUMBTRACK:
         case SB_THUMBPOSITION:
           da->gl=HIWORD(wparam);  break;
         default: break;
       }
       SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_GLS_SCROLL),SB_CTL,da->gl,TRUE);
       sprintf(str,"%ld",da->gl);
       SendDlgItemMessage(hdlg,DLG_DEFAULT_GLS_EDIT,WM_SETTEXT,0,(LPARAM)str);
     }
     else if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_DEFAULT_BMP_SCROLL))){
       GetScrollRange(hctl,SB_CTL,&iMin,&iMax);
       switch (LOWORD(wparam)) {
         case SB_PAGEDOWN:
         case SB_LINEDOWN: da->pa = min(iMax,da->pa+1); break;
         case SB_PAGEUP:
         case SB_LINEUP:   da->pa = max(iMin,da->pa-1); break;
         case SB_THUMBTRACK:
         case SB_THUMBPOSITION:
           da->pa=HIWORD(wparam);  break;
         default: break;
       }
       SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_BMP_SCROLL),SB_CTL,da->pa,TRUE);
       sprintf(str,"%ld",da->pa);
       SendDlgItemMessage(hdlg,DLG_DEFAULT_BMP_EDIT,WM_SETTEXT,0,(LPARAM)str);
     }
     else if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_DEFAULT_REF_SCROLL))){
       GetScrollRange(hctl,SB_CTL,&iMin,&iMax);
       switch (LOWORD(wparam)) {
         case SB_PAGEDOWN:
         case SB_LINEDOWN: da->rf = min(iMax,da->rf+1); break;
         case SB_PAGEUP:
         case SB_LINEUP:   da->rf = max(iMin,da->rf-1); break;
         case SB_THUMBTRACK:
         case SB_THUMBPOSITION:
           da->rf=HIWORD(wparam);  break;
         default: break;
       }
       SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_REF_SCROLL),SB_CTL,da->rf,TRUE);
       sprintf(str,"%ld",da->rf);
       SendDlgItemMessage(hdlg,DLG_DEFAULT_REF_EDIT,WM_SETTEXT,0,(LPARAM)str);
     }
     else if((HWND)lparam == (hctl=GetDlgItem(hdlg,DLG_DEFAULT_TRN_SCROLL))){
       GetScrollRange(hctl,SB_CTL,&iMin,&iMax);
       switch (LOWORD(wparam)) {
         case SB_PAGEDOWN:
         case SB_LINEDOWN: da->tr = min(iMax,da->tr+1); break;
         case SB_PAGEUP:
         case SB_LINEUP:   da->tr = max(iMin,da->tr-1); break;
         case SB_THUMBTRACK:
         case SB_THUMBPOSITION:
           da->tr=HIWORD(wparam);  break;
         default: break;
       }
       SetScrollPos(GetDlgItem(hdlg,DLG_DEFAULT_TRN_SCROLL),SB_CTL,da->tr,TRUE);
       sprintf(str,"%ld",da->tr);
       SendDlgItemMessage(hdlg,DLG_DEFAULT_TRN_EDIT,WM_SETTEXT,0,(LPARAM)str);
     }
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case DLG_DEFAULT_FC_EDIT:
          SetSfxColour(da->fc,IDX_MISC_FACE_COLOUR,hdlg);
          InvalidateRect(GetDlgItem(hdlg,DLG_DEFAULT_FC_EDIT),NULL,FALSE);
          break;
        case DLG_DEFAULT_SC_EDIT:
          SetSfxColour(da->sc,IDX_MISC_SHADER_COLOUR,hdlg);
          InvalidateRect(GetDlgItem(hdlg,DLG_DEFAULT_SC_EDIT),NULL,FALSE);
          break;
        case DLG_DEFAULT_GLS_OFF:
          da->br=0;
          break;
        case DLG_DEFAULT_GLS_ON:
          da->br=1;
          break;
        case DLG_DEFAULT_SMOOTH_OFF:
          da->sm=0;
          break;
        case DLG_DEFAULT_SMOOTH_ON:
          da->sm=1;
          break;
        case IDCANCEL:
        case DLG_DEFAULT_CANCEL:
          EndDialog(hdlg,FAIL);
          return(TRUE);
        case IDOK:
        case DLG_DEFAULT_OK:
          for(i=0;i<3;i++){
            FaceColour[i]=da->fc[i];
            ShaderColour[i]=da->sc[i];
          }
          DefaultSmoothing=da->sm;
          DefaultShiny=da->br;
          DefaultBrilliance=da->gl;
          DefaultParameter=da->pa;
          DefaultReflectivity=da->rf;
          DefaultTransp=da->tr;
          DefaultTexture=da->sh;

          EndDialog(hdlg,OK);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void UpdateGLview(void){
 if(ghwndOpenGLview != NULL)PostMessage(ghwndOpenGLview,(WM_USER+2),0,0);
}

#define WM_RSFINIT (WM_USER+1000)

static HBITMAP MapButtonBitmaps[6]={NULL,NULL,NULL,NULL,NULL,NULL};

static BOOL APIENTRY PropColourDlgProc(HWND hDlg,UINT msg,
                                    WPARAM wParam, LPARAM lParam){
 HWND hctl;
 int i,iMin,iMax;
 LPDRAWITEMSTRUCT lpdis;
 FACEATTRIBS *fa,*da;
 fa=&FaceAttribs; da=&DefAttribs;
 switch(msg){
   case WM_PAINT:
     PaintDialogBackground(hDlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     image_edit_flag= -1;
     nf=GetFaceAttribs(da,1);
     nf=GetFaceAttribs(fa,0);
     UpdateTitleDialog(GetParent(hDlg),nf);
     UpdateColourAttribsSection(hDlg,nf,fa,da);
     UpdateBrillAttribsSection(hDlg,nf,fa,da);
     UpdateSmoothAttribsSection(hDlg,nf,fa,da);
     UpdateGlassTypeAttribsSection(hDlg,nf,fa,da);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_GLS_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,15));
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_GLS_SCROLL,TBM_SETLINESIZE,0,1);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_GLS_SCROLL,TBM_SETPAGESIZE,0,5);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_GLS_SCROLL,TBM_SETTICFREQ,1,(LPARAM)1);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_TRN_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,15));
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_TRN_SCROLL,TBM_SETLINESIZE,0,1);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_TRN_SCROLL,TBM_SETPAGESIZE,0,5);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_GLS_SCROLL,TBM_SETTICFREQ,1,(LPARAM)1);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_REF_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,15));
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_REF_SCROLL,TBM_SETLINESIZE,0,1);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_REF_SCROLL,TBM_SETPAGESIZE,0,5);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_GLS_SCROLL,TBM_SETTICFREQ,1,(LPARAM)1);
     if(nf > 0 && fa->flag_tr > 0)da->tr=fa->tr;
     UpdateTransAttribsSection(hDlg,nf,fa,da);
     if(nf > 0 && fa->flag_gl > 0)da->gl=fa->gl;
     UpdateGlossAttribsSection(hDlg,nf,fa,da);
     if(nf > 0 && fa->flag_rf > 0)da->rf=fa->rf;
     UpdateReflAttribsSection(hDlg,nf,fa,da);
     return TRUE;
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lParam;
       if(lpdis->CtlID == DLG_ATTRIBS_FC_EDIT){
         r=da->fc[0]; g=da->fc[1]; b=da->fc[2];
         if(lpdis->itemState & ODS_DISABLED){
           r=128; g=128; b=128;
         }
         if(lpdis->itemState & ODS_SELECTED)
           InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           if(lpdis->itemState & ODS_DISABLED)
                hbr=CreateHatchBrush(HS_DIAGCROSS,RGB(128,128,128));
           else hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_HSCROLL:
     if((HWND)lParam == (hctl=GetDlgItem(hDlg,DLG_ATTRIBS_GLS_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wParam)) {
         case TB_PAGEDOWN:
         case TB_LINEDOWN: da->gl = min(iMax,da->gl+1); break;
         case TB_PAGEUP:
         case TB_LINEUP:   da->gl = max(iMin,da->gl-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           da->gl=HIWORD(wParam);  break;
         case TB_ENDTRACK:
           fa->gl=da->gl; fa->flag_gl=1;
           PutFaceAttribute(2,fa);
           break;
         default: break;
       }
       UpdateGlossAttribsSection(hDlg,nf,fa,da);
     }
     else if((HWND)lParam == (hctl=GetDlgItem(hDlg,DLG_ATTRIBS_TRN_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wParam)){
         case TB_PAGEDOWN:
         case TB_LINEDOWN: da->tr = min(iMax,da->tr+1); break;
         case TB_PAGEUP:
         case TB_LINEUP:   da->tr = max(iMin,da->tr-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           da->tr=HIWORD(wParam);  break;
         case TB_ENDTRACK:
           fa->tr=da->tr; fa->flag_tr=1;
           PutFaceAttribute(3,fa);
           break;
         default: break;
       }
       UpdateTransAttribsSection(hDlg,nf,fa,da);
     }
     else if((HWND)lParam == (hctl=GetDlgItem(hDlg,DLG_ATTRIBS_REF_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wParam)) {
         case TB_PAGEDOWN:
         case TB_LINEDOWN: da->rf = min(iMax,da->rf+1); break;
         case TB_PAGEUP:
         case TB_LINEUP:   da->rf = max(iMin,da->rf-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           da->rf=HIWORD(wParam);  break;
         case TB_ENDTRACK:
           fa->rf=da->rf; fa->flag_rf=1;
           PutFaceAttribute(4,fa);
           break;
         default: break;
       }
       UpdateReflAttribsSection(hDlg,nf,fa,da);
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wParam)){
       case DLG_ATTRIBS_UPDATE:
         nf=GetFaceAttribs(fa,0);
         UpdateTitleDialog(GetParent(hDlg),nf);
         UpdateColourAttribsSection(hDlg,nf,fa,da);
         UpdateBrillAttribsSection(hDlg,nf,fa,da);
         UpdateSmoothAttribsSection(hDlg,nf,fa,da);
         UpdateGlassTypeAttribsSection(hDlg,nf,fa,da);
         if(nf > 0 && fa->flag_tr > 0)da->tr=fa->tr;
         UpdateTransAttribsSection(hDlg,nf,fa,da);
         if(nf > 0 && fa->flag_gl > 0)da->gl=fa->gl;
         UpdateGlossAttribsSection(hDlg,nf,fa,da);
         if(nf > 0 && fa->flag_rf > 0)da->rf=fa->rf;
         UpdateReflAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_FC_EDIT:
         if(SetSfxColour(da->fc,IDX_MISC_FACE_COLOUR,hDlg)){
           VECCOPY(da->fc,fa->fc) fa->flag_fc=1;
           VECCOPY(da->fc,LastColour)
           UpdateColourAttribsSection(hDlg,nf,fa,da);
           PutFaceAttribute(0,fa);
           UpdateGLview();
         }
         InvalidateRect(GetDlgItem(hDlg,DLG_ATTRIBS_FC_EDIT),NULL,FALSE);
         break;
       case DLG_ATTRIBS_SMOOTH_ON:
         i=SendDlgItemMessage(hDlg,DLG_ATTRIBS_SMOOTH_ON,BM_GETCHECK,0,0);
         if(i == BST_CHECKED)fa->sm=da->sm=0;
         else                fa->sm=da->sm=1;
         fa->flag_sm=1;
         UpdateSmoothAttribsSection(hDlg,nf,fa,da);
         PutFaceAttribute(5,fa);
         UpdateGLview();
         break;
       case DLG_ATTRIBS_GLS_ON:
         i=SendDlgItemMessage(hDlg,DLG_ATTRIBS_GLS_ON,BM_GETCHECK,0,0);
         if(i == BST_CHECKED)fa->br=da->br=0;
         else                fa->br=da->br=1;
         fa->flag_br=1;
         UpdateBrillAttribsSection(hDlg,nf,fa,da);
         PutFaceAttribute(6,fa);
         break;
       case DLG_ATTRIBS_TRN_TYPE1:
         i=SendDlgItemMessage(hDlg,DLG_ATTRIBS_TRN_TYPE1,BM_GETCHECK,0,0);
         if(i == BST_CHECKED)fa->gt=da->gt=0;
         else                fa->gt=da->gt=1;
         fa->flag_gt=1;
         UpdateGlassTypeAttribsSection(hDlg,nf,fa,da);
         PutFaceAttribute(15,fa);
         break;
       default:
         break;
     }
   default:   break;
 }
 return FALSE;
}

static BOOL APIENTRY PropTextureDlgProc(HWND hDlg,UINT msg,
                                       WPARAM wParam, LPARAM lParam){
 HWND hctl;
 int i,iMin,iMax;
 FACEATTRIBS *fa,*da;
 fa=&FaceAttribs; da=&DefAttribs;
 switch(msg){
   case WM_PAINT:
     PaintDialogBackground(hDlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     return TRUE;
     break;
   case WM_COMMAND:
     switch(LOWORD(wParam)){
       case DLG_ATTRIBS_UPDATE:
         nf=GetFaceAttribs(fa,0);
         UpdateTitleDialog(GetParent(hDlg),nf);
         break;
       default:
         break;
     }
   default:   break;
 }
 return FALSE;
}

static BOOL APIENTRY PropShaderDlgProc(HWND hDlg,UINT msg,
                                       WPARAM wParam, LPARAM lParam){
 HWND hctl;
 int i,iMin,iMax;
 LPDRAWITEMSTRUCT lpdis;
 FACEATTRIBS *fa,*da;
 fa=&FaceAttribs; da=&DefAttribs;
 switch(msg){
   case WM_PAINT:
     PaintDialogBackground(hDlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateShaderAttribsSection(hDlg,nf,fa,da);
     UpdateShaderColourAttribsSection(hDlg,nf,fa,da);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_PRM_SCROLL,TBM_SETRANGE,
                                (WPARAM)TRUE,(LPARAM)MAKELONG(0,255));
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_PRM_SCROLL,TBM_SETLINESIZE,0,1);
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_PRM_SCROLL,TBM_SETPAGESIZE,0,5);
     for(i=0;i<256;i+=16)
     SendDlgItemMessage(hDlg,DLG_ATTRIBS_PRM_SCROLL,TBM_SETTIC,0,(LPARAM)i);
     if(nf > 0 && fa->flag_pa > 0)da->pa=fa->pa;
     UpdateParamAttribsSection(hDlg,nf,fa,da);
     return TRUE;
     break;
   case WM_DRAWITEM:{
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lParam;
       if(lpdis->CtlID == DLG_ATTRIBS_SC_EDIT){
         r=da->sc[0]; g=da->sc[1]; b=da->sc[2];
         if(lpdis->itemState & ODS_DISABLED){
           r=128; g=128; b=128;
         }
         if(lpdis->itemState & ODS_SELECTED)
           InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           if(lpdis->itemState & ODS_DISABLED)
                hbr=CreateHatchBrush(HS_DIAGCROSS,RGB(128,128,128));
           else hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     return TRUE;
   case WM_HSCROLL:
     if((HWND)lParam == (hctl=GetDlgItem(hDlg,DLG_ATTRIBS_PRM_SCROLL))){
       iMin=SendMessage(hctl,TBM_GETRANGEMIN,0,0);
       iMax=SendMessage(hctl,TBM_GETRANGEMAX,0,0);
       switch (LOWORD(wParam)) {
         case TB_PAGEDOWN:
         case TB_LINEDOWN: da->pa = min(iMax,da->pa+1); break;
         case TB_PAGEUP:
         case TB_LINEUP:   da->pa = max(iMin,da->pa-1); break;
         case TB_THUMBTRACK:
         case TB_THUMBPOSITION:
           da->pa=HIWORD(wParam);  break;
         case TB_ENDTRACK:
           fa->pa=da->pa; fa->flag_pa=1;
           PutFaceAttribute(14,fa);
           break;
         default: break;
       }
       UpdateParamAttribsSection(hDlg,nf,fa,da);
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wParam)){
       case DLG_ATTRIBS_UPDATE:
         nf=GetFaceAttribs(fa,0);
         UpdateTitleDialog(GetParent(hDlg),nf);
         UpdateShaderAttribsSection(hDlg,nf,fa,da);
         UpdateShaderColourAttribsSection(hDlg,nf,fa,da);
         if(nf > 0 && fa->flag_pa > 0)da->pa=fa->pa;
         UpdateParamAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_SHA_SET:
         if(fa->flag_sh < 0  || fa->flag_st < 0){
           int sh;
//           if(SendPrgmQuery(IDQ_MISC_DSHADER,1) != IDYES)break;
           if((sh=SelectScrolledItemList(NoAlTexts,AlTextList,
                 "Shader",hDlg)) >= 0){
             fa->flag_sh=1; fa->sh=sh;
             fa->flag_st=1; fa->st=0;
             PutFaceAttribute(7,fa);
           }
         }
         else if(fa->st == 1){ /* external */
           DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_SHADERSET),hDlg,
                     (DLGPROC)SetExternalShaderDlgProc,(LPARAM)fa);
           fa->flag_sh=1; fa->flag_st=1;
           PutFaceAttribute(7,fa);
         }
         else{
           int sh;
           if((sh=SelectScrolledItemList(NoAlTexts,AlTextList,
                 "Shader",hDlg)) >= 0){
             fa->flag_sh=1; fa->sh=sh;
             fa->flag_st=1; fa->st=0;
             PutFaceAttribute(7,fa);
           }
         }
         UpdateShaderAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_SHA_INT:{
           int sh;
           if((sh=SelectScrolledItemList(NoAlTexts,AlTextList,
                 "Shader",hDlg)) >= 0){
             fa->flag_sh=1; fa->sh=sh;
             fa->flag_st=1; fa->st=0;
             PutFaceAttribute(7,fa);
             UpdateShaderAttribsSection(hDlg,nf,fa,da);
           }
         }
         break;
       case DLG_ATTRIBS_SHA_EXT:
         fa->flag_sh=1; //fa->sh=0;
         fa->flag_st=1; fa->st=1;
         DialogBoxParam(ghinst_main,MAKEINTRESOURCE(DLG_SHADERSET),hDlg,
                   (DLGPROC)SetExternalShaderDlgProc,(LPARAM)fa);
         PutFaceAttribute(7,fa);
         UpdateShaderAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_SHA_XCHOOSE:
         if(nSha < 2){
           SendPrgmQuery(IDQ_NOSHADERAXES,0);
           break;
         }
         else {
           int sa;
           if((sa=RequestShaderID(hDlg,"Choose axes to use")) >= 0){
             fa->flag_sa=1; fa->sa=sa;
             PutFaceAttribute(9,fa);
             UpdateShaderAttribsSection(hDlg,nf,fa,da);
           }
         }
         break;
       case DLG_ATTRIBS_SHA_XMOVE:
         if(tool != NOTOOL)DisengageDesignerTool();
         if(sktool == YES)DisengageSkeletonTool();
         PositionShaderOnOff(1,fa->sa);
         break;
       case DLG_ATTRIBS_SC_EDIT:
         if(SetSfxColour(da->sc,IDX_MISC_SHADER_COLOUR,hDlg)){
           VECCOPY(da->sc,fa->sc) fa->flag_sc=1;
           UpdateShaderColourAttribsSection(hDlg,nf,fa,da);
           PutFaceAttribute(1,fa);
         }
         InvalidateRect(GetDlgItem(hDlg,DLG_ATTRIBS_SC_EDIT),NULL,FALSE);
         break;
      default:
        break;
     }
   default:   break;
 }
 return FALSE;
}

static BOOL APIENTRY PropMappingDlgProc(HWND hDlg,UINT msg,
                                       WPARAM wParam, LPARAM lParam){
 FACEATTRIBS *fa,*da;
 fa=&FaceAttribs; da=&DefAttribs;
 switch(msg){
   case WM_PAINT:
     PaintDialogBackground(hDlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     SendMessage(GetDlgItem(hDlg,DLG_ATTRIBS_MAP_SHADED),BM_SETIMAGE,
                 (WPARAM)IMAGE_BITMAP,(LPARAM)MapButtonBitmaps[0]);
     SendMessage(GetDlgItem(hDlg,DLG_ATTRIBS_MAP_UNSHADED),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)MapButtonBitmaps[1]);
     SendMessage(GetDlgItem(hDlg,DLG_ATTRIBS_MAP_SINGLE),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)MapButtonBitmaps[2]);
     SendMessage(GetDlgItem(hDlg,DLG_ATTRIBS_MAP_TILE),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)MapButtonBitmaps[3]);
     SendMessage(GetDlgItem(hDlg,DLG_ATTRIBS_MAP_OPAQUE),BM_SETIMAGE,
                 (WPARAM)0,(LPARAM)MapButtonBitmaps[4]);

     SendMessage(GetDlgItem(hDlg,DLG_ATTRIBS_MAP_CLEAR),BM_SETIMAGE,
                 (WPARAM)IMAGE_ICON,(LPARAM)MapButtonBitmaps[5]);

     UpdateMapAttribsSection(hDlg,nf,fa,da);
     return TRUE;
     break;
   case WM_COMMAND:
     switch(LOWORD(wParam)){
       case DLG_ATTRIBS_UPDATE:
         nf=GetFaceAttribs(fa,0);
         UpdateTitleDialog(GetParent(hDlg),nf);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_MAP_SET:{
         int id;
           if((id=RequestMapID(hDlg,"Choose map",1)) < 0)break;
           if(id == nImaps){
             if(CreateModelerBrush() == FAIL){
               SendPrgmQuery(IDQ_NOMOREMAPS,0);
               break;
             }
           }
           fa->flag_ba=1; fa->ba=1;  /* brush    */
           fa->flag_bu=1; fa->bu=id; /* brush ID */
           PutFaceAttribute(10,fa);
           UpdateMapAttribsSection(hDlg,nf,fa,da);
           UpdateGLview();
         }
         break;
       case DLG_ATTRIBS_MAP_DROP:
         fa->flag_ba=1; fa->ba=0;
         fa->flag_bu=1; fa->bu &=0xe0;
         PutFaceAttribute(8,fa);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         UpdateGLview();
         break;
       case DLG_ATTRIBS_MAP_MOVE:
         if(tool != NOTOOL)DisengageDesignerTool();
         if(sktool == YES)DisengageSkeletonTool();
         PositionBrushOnOff(1,fa->bu);
         break;
       case DLG_ATTRIBS_MAP_EDIT:
         image_edit_flag=1;
         axis_move_flag=fa->bu;
         DestroyWindow(ghwndDlgAttribs);
         ghwndDlgAttribs=NULL;
         if(sktool == YES)DisengageSkeletonTool();
         PostMessage(ghwnd_main,WM_COMMAND,IDM_ATTRIBUTES_MAP,0);
         break;
       case DLG_ATTRIBS_MAP_SINGLE:
         fa->flag_bt=1; fa->bt=0;
         PutFaceAttribute(11,fa);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_MAP_TILE:
         fa->flag_bt=1; fa->bt=1;
         PutFaceAttribute(11,fa);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_MAP_SHADED:
         fa->flag_bs=1; fa->bs=0;
         PutFaceAttribute(12,fa);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_MAP_UNSHADED:
         fa->flag_bs=1; fa->bs=1;
         PutFaceAttribute(12,fa);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_MAP_OPAQUE:
         fa->flag_bc=1; fa->bc=0;
         PutFaceAttribute(13,fa);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_MAP_CLEAR:
         fa->flag_bc=1; fa->bc=1;
         PutFaceAttribute(13,fa);
         UpdateMapAttribsSection(hDlg,nf,fa,da);
         break;
       case DLG_ATTRIBS_MAP_DELETE:   /* not used yet */
         break;
       default:
         break;
     }
   default:   break;
 }
 return FALSE;
}

#define D_BORDER 4
#define N_PAGES  3

typedef struct tag_dlghdr1 {
  HWND hwndDlg[N_PAGES];
  RECT rcTab;
} DLGHDR1;

static DLGHDR1 DlgHdrAttribs,*lpDlgHdrAttribs;
static HWND    hwndTabAttribs;

BOOL CALLBACK AttribsDlgProc(HWND hdlg,UINT msg,
                             WPARAM wparam,LPARAM lparam){
 static BOOL bRollUp=FALSE;
 static RECT RollUpRect;
 static int xpos= -1,ypos= -1;
 int i;
 RECT rc,rd;
 char str[32];
 switch( msg ){
//   case WM_PAINT: // Include this if the tabbed page bacground is to be textured
//     PaintDialogBackground(hdlg,ghinst_main);
//     break;
   case WM_INITDIALOG:{
       TC_ITEM tie;
       int h;
       DWORD dwDlgBase=GetDialogBaseUnits();
       MapButtonBitmaps[0]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAP_FX1));
       MapButtonBitmaps[1]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAP_FX2));
       MapButtonBitmaps[2]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAP_FX3));
       MapButtonBitmaps[3]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAP_FX4));
       MapButtonBitmaps[4]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAP_FX5));
// for bitmap
//       MapButtonBitmaps[5]=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_MAP_FX6));
       MapButtonBitmaps[5]=(HBITMAP)LoadImage(ghinst_main,"MAPFX6",IMAGE_ICON,
                                0,0,LR_DEFAULTCOLOR); // ICON
       if(xpos == -1 && ypos == -1)CentreDialogOnScreen(hdlg);
       else SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
       lpDlgHdrAttribs=&DlgHdrAttribs;
       GetWindowRect(hdlg,&rc);
       GetWindowRect(GetDlgItem(hdlg,DLG_ATTRIBS_PREVIEW),&rd);
       h=rd.top-rc.top-GetSystemMetrics(SM_CYCAPTION);
       GetClientRect(hdlg,&rc);
       hwndTabAttribs=CreateWindow(WC_TABCONTROL," ",
                            WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
                            (D_BORDER*LOWORD(dwDlgBase))/4,
                            (D_BORDER*HIWORD(dwDlgBase))/8,
                            rc.right-(2*D_BORDER*LOWORD(dwDlgBase))/4,
                            h-(2*D_BORDER*HIWORD(dwDlgBase))/8,
                            hdlg,NULL,ghinst_main,NULL);
       {char text_string[255];
         LoadString(ghinst_main,IDX_MISC_LANGUAGE,
                    text_string,sizeof(text_string));
         if(strcmp(text_string,"ENGLISH") == 0)
           SendMessage(hwndTabAttribs,WM_SETFONT, /* for text in tab */
                  (WPARAM)GetStockObject(ANSI_VAR_FONT),0);
       }
       memset(&tie,0,sizeof(TC_ITEM));
       tie.mask = TCIF_TEXT | TCIF_IMAGE;
       tie.iImage = -1;
       tie.pszText=str;
       LoadString(ghinst_main,IDX_MISC_ATT_COLOUR,str,32);
       TabCtrl_InsertItem(hwndTabAttribs,0,&tie);
//       LoadString(ghinst_main,IDX_MISC_ATT_SETTINGS,str,32);
//       TabCtrl_InsertItem(hwndTabAttribs,3,&tie);
       LoadString(ghinst_main,IDX_MISC_ATT_SHADER,str,32);
       TabCtrl_InsertItem(hwndTabAttribs,1,&tie);
       LoadString(ghinst_main,IDX_MISC_ATT_MAPS,str,32);
       TabCtrl_InsertItem(hwndTabAttribs,2,&tie);
       SetRectEmpty(&rc);
       TabCtrl_AdjustRect(hwndTabAttribs,FALSE,&rc);
       rc.left += (D_BORDER*LOWORD(dwDlgBase))/4;
       rc.top += (D_BORDER*HIWORD(dwDlgBase))/8;
       CopyRect(&(lpDlgHdrAttribs->rcTab),&rc);
       for(i=0;i<N_PAGES;i++)lpDlgHdrAttribs->hwndDlg[i]=NULL;
       lpDlgHdrAttribs->hwndDlg[0]=CreateDialog(ghinst_main,
                                   "ATTRIBSPAGE1",
                                   hdlg,(DLGPROC)PropColourDlgProc);
       SetWindowPos(lpDlgHdrAttribs->hwndDlg[0],NULL,
                    lpDlgHdrAttribs->rcTab.left,
                    lpDlgHdrAttribs->rcTab.top,0,0,
                    SWP_NOSIZE | SWP_NOZORDER);
       ShowWindow(lpDlgHdrAttribs->hwndDlg[0],SW_SHOW);
     }
     return TRUE;
   case WM_NOTIFY:
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGING){
       int id=TabCtrl_GetCurSel(hwndTabAttribs);
     }
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGE){
       int id=TabCtrl_GetCurSel(hwndTabAttribs);
       for(i=0;i<N_PAGES;i++){
         if(lpDlgHdrAttribs->hwndDlg[i] != NULL){
           DestroyWindow(lpDlgHdrAttribs->hwndDlg[i]);
           lpDlgHdrAttribs->hwndDlg[i]=NULL;
         }
       }
       if(id == 0)lpDlgHdrAttribs->hwndDlg[0]=CreateDialog(ghinst_main,
                                   "ATTRIBSPAGE1",
                                   hdlg,(DLGPROC)PropColourDlgProc);
//       if(id == 3)lpDlgHdrAttribs->hwndDlg[3]=CreateDialog(ghinst_main,
//                                   "ATTRIBSPAGE2",
//                                   hdlg,(DLGPROC)PropTextureDlgProc);
       if(id == 1)lpDlgHdrAttribs->hwndDlg[1]=CreateDialog(ghinst_main,
                                   "ATTRIBSPAGE3",
                                   hdlg,(DLGPROC)PropShaderDlgProc);
       if(id == 2)lpDlgHdrAttribs->hwndDlg[2]=CreateDialog(ghinst_main,
                                   "ATTRIBSPAGE4",
                                   hdlg,(DLGPROC)PropMappingDlgProc);
       if(lpDlgHdrAttribs->hwndDlg[id] != NULL){
         SetWindowPos(lpDlgHdrAttribs->hwndDlg[id],NULL,
                      lpDlgHdrAttribs->rcTab.left,
                      lpDlgHdrAttribs->rcTab.top,0,0,
                      SWP_NOSIZE | SWP_NOZORDER);
         ShowWindow(lpDlgHdrAttribs->hwndDlg[id],SW_SHOW);
       }
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       default:
         break;
       case SC_MINIMIZE:
         if(bRollUp){
           SetWindowPos(hdlg,NULL,0,0,RollUpRect.right-RollUpRect.left,
                                      RollUpRect.bottom-RollUpRect.top,
                                      SWP_NOMOVE | SWP_NOZORDER);
         }
         else{
           GetWindowRect(hdlg,&RollUpRect);
           SetWindowPos(hdlg,NULL,0,0,RollUpRect.right-RollUpRect.left,
                                      GetSystemMetrics(SM_CYCAPTION)+
                                      2*GetSystemMetrics(SM_CYFIXEDFRAME),
                                      SWP_NOMOVE | SWP_NOZORDER);
         }
         bRollUp ^= 1;
         return TRUE;
         break;
       case SC_RESTORE:
         return TRUE;
         break;
       case SC_CLOSE:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_ATTRIBUTES_FACE,0);
         break;
     }
     break;
   case WM_COMMAND:
     switch( LOWORD(wparam) ){
       case DLG_ATTRIBS_UPDATE:
         for(i=0;i<N_PAGES;i++){
           if(lpDlgHdrAttribs->hwndDlg[i] != NULL)
             PostMessage(lpDlgHdrAttribs->hwndDlg[i],
             WM_COMMAND,DLG_ATTRIBS_UPDATE,0);
         }
         break;
       case DLG_ATTRIBS_PREVIEW:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_VIEW_RENDERINWINDOW,0);
         break;
       case DLG_ATTRIBS_CLOSE:
       case IDOK:
       case IDCANCEL:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_ATTRIBUTES_FACE,0);
         break;
       default:
         break;
     }
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_DESTROY:
     for(i=0;i<N_PAGES;i++){
       if(lpDlgHdrAttribs->hwndDlg[i] != NULL)
         DestroyWindow(lpDlgHdrAttribs->hwndDlg[i]);
       lpDlgHdrAttribs->hwndDlg[i]=NULL;
     }
     for(i=0;i<5;i++){  // the 6th is an ICON
       if(MapButtonBitmaps[i] != NULL){
         if(!DeleteObject(MapButtonBitmaps[i]))MessageBox(NULL,"Bad Delete",NULL,MB_OK);
         MapButtonBitmaps[i]=NULL;
       }
     }
     GetWindowRect(hdlg,&rc);
     bRollUp=FALSE;
     xpos=rc.left; ypos=rc.top;
     break;
   default: break;
 }
 return( FALSE );
}

BOOL IsAttribDlgMessage(HWND hWnd, LPMSG lpMsg){
 int i;
 DLGHDR1 *lpDlgHdr;
 lpDlgHdr=&DlgHdrAttribs;
 for(i=0;i<N_PAGES;i++){
   if(lpDlgHdr->hwndDlg[i] != NULL){
     if(IsDialogMessage(lpDlgHdr->hwndDlg[i],lpMsg))return TRUE;
   }
 }
 return IsDialogMessage(hWnd,lpMsg);
}

// Handle textures

static void UpdateShaderSetDialog(int sa, int id, HWND hdlg){
 int active;
 int BitID[8]={0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};
 int OnResourceID[8]={DLG_SHADERSET_ON1,DLG_SHADERSET_ON2,
                      DLG_SHADERSET_ON3,DLG_SHADERSET_ON4,
                      DLG_SHADERSET_ON5,DLG_SHADERSET_ON6,
                      DLG_SHADERSET_ON7,DLG_SHADERSET_ON8
                     };
 int OffResourceID[8]={DLG_SHADERSET_OFF1,DLG_SHADERSET_OFF2,
                       DLG_SHADERSET_OFF3,DLG_SHADERSET_OFF4,
                       DLG_SHADERSET_OFF5,DLG_SHADERSET_OFF6,
                       DLG_SHADERSET_OFF7,DLG_SHADERSET_OFF8
                      };
 int NameResourceID[8]={DLG_SHADERSET_NAME1,DLG_SHADERSET_NAME2,
                        DLG_SHADERSET_NAME3,DLG_SHADERSET_NAME4,
                        DLG_SHADERSET_NAME5,DLG_SHADERSET_NAME6,
                        DLG_SHADERSET_NAME7,DLG_SHADERSET_NAME8
                       };
 int EditResourceID[8]={DLG_SHADERSET_EDIT1,DLG_SHADERSET_EDIT2,
                        DLG_SHADERSET_EDIT3,DLG_SHADERSET_EDIT4,
                        DLG_SHADERSET_EDIT5,DLG_SHADERSET_EDIT6,
                        DLG_SHADERSET_EDIT7,DLG_SHADERSET_EDIT8
                       };
 int LoadResourceID[8]={DLG_SHADERSET_LOAD1,DLG_SHADERSET_LOAD2,
                        DLG_SHADERSET_LOAD3,DLG_SHADERSET_LOAD4,
                        DLG_SHADERSET_LOAD5,DLG_SHADERSET_LOAD6,
                        DLG_SHADERSET_LOAD7,DLG_SHADERSET_LOAD8
                       };
 if(xShaders[id].in_use == 1)active=TRUE; else active=FALSE;
 EnableWindow(GetDlgItem(hdlg,OnResourceID[id]),active);
// EnableWindow(GetDlgItem(hdlg,OffResourceID[id]),active);
 EnableWindow(GetDlgItem(hdlg,EditResourceID[id]),active);
 EnableWindow(GetDlgItem(hdlg,LoadResourceID[id]),!active);
 if(xShaders[id].in_use == 1)active=SW_SHOW; else active=SW_HIDE;
 ShowWindow(GetDlgItem(hdlg,EditResourceID[id]),active);
 ShowWindow(GetDlgItem(hdlg,LoadResourceID[id]),!active);
 if(xShaders[id].in_use == 1){
   SetDlgItemText(hdlg,NameResourceID[id],
                 (LPCTSTR)FileInPath(xShaders[id].name));
   if((sa & BitID[id]) == BitID[id])active=TRUE; else active=FALSE;
   SendDlgItemMessage(hdlg,OnResourceID[id],BM_SETCHECK,active,0);
//   SendDlgItemMessage(hdlg,OffResourceID[id],BM_SETCHECK,!active,0);
 }
 return;
}

static BOOL CALLBACK SetExternalShaderDlgProc(HWND hdlg,UINT msg,
                                    WPARAM wparam,LPARAM lparam){
 BOOL active;
 FACEATTRIBS *fa;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     SetWindowLong(hdlg,GWL_USERDATA,lparam);
     fa=(FACEATTRIBS *)lparam;
     UpdateShaderSetDialog(fa->sh,0,hdlg);
     UpdateShaderSetDialog(fa->sh,1,hdlg);
     UpdateShaderSetDialog(fa->sh,2,hdlg);
     UpdateShaderSetDialog(fa->sh,3,hdlg);
     UpdateShaderSetDialog(fa->sh,4,hdlg);
     UpdateShaderSetDialog(fa->sh,5,hdlg);
     UpdateShaderSetDialog(fa->sh,6,hdlg);
     UpdateShaderSetDialog(fa->sh,7,hdlg);
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_COMMAND:
      fa=(FACEATTRIBS *)GetWindowLong(hdlg,GWL_USERDATA);
     switch(LOWORD(wparam)){
       case DLG_SHADERSET_ON1:
         if((fa->sh & 0x01) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x01;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,0,hdlg);
         }
         else{
           fa->sh &= 0xfe;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,0,hdlg);
         }
         break;
       case DLG_SHADERSET_ON2:
         if((fa->sh & 0x02) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x02;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,1,hdlg);
         }
         else{
           fa->sh &= 0xfd;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,1,hdlg);
         }
         break;
       case DLG_SHADERSET_ON3:
         if((fa->sh & 0x04) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x04;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,2,hdlg);
         }
         else{
           fa->sh &= 0xfb;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,2,hdlg);
         }
         break;
       case DLG_SHADERSET_ON4:
         if((fa->sh & 0x08) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x08;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,3,hdlg);
         }
         else{
           fa->sh &= 0xf7;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,3,hdlg);
         }
         break;
       case DLG_SHADERSET_ON5:
         if((fa->sh & 0x10) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x10;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,4,hdlg);
         }
         else{
           fa->sh &= 0xef;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,4,hdlg);
         }
         break;
       case DLG_SHADERSET_ON6:
         if((fa->sh & 0x20) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x20;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,5,hdlg);
         }
         else{
           fa->sh &= 0xdf;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,5,hdlg);
         }
         break;
       case DLG_SHADERSET_ON7:
         if((fa->sh & 0x40) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x40;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,6,hdlg);
         }
         else{
           fa->sh &= 0xbf;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,6,hdlg);
         }
         break;
       case DLG_SHADERSET_ON8:
         if((fa->sh & 0x80) == 0x00){
           fa->sh &= 0xff;
           fa->sh |= 0x80;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,7,hdlg);
         }
         else{
           fa->sh &= 0x7f;
           PutFaceAttribute(7,fa); UpdateShaderSetDialog(fa->sh,7,hdlg);
         }
         break;

       case DLG_SHADERSET_EDIT1:
         EditExternalShader(0,hdlg);
         break;
       case DLG_SHADERSET_EDIT2:
         EditExternalShader(1,hdlg);
         break;
       case DLG_SHADERSET_EDIT3:
         EditExternalShader(2,hdlg);
         break;
       case DLG_SHADERSET_EDIT4:
         EditExternalShader(3,hdlg);
         break;
       case DLG_SHADERSET_EDIT5:
         EditExternalShader(4,hdlg);
         break;
       case DLG_SHADERSET_EDIT6:
         EditExternalShader(5,hdlg);
         break;
       case DLG_SHADERSET_EDIT7:
         EditExternalShader(6,hdlg);
         break;
       case DLG_SHADERSET_EDIT8:
         EditExternalShader(7,hdlg);
         break;

       case DLG_SHADERSET_LOAD1:
         LoadExternalShader(0,hdlg);
         UpdateShaderSetDialog(fa->sh,0,hdlg);
         break;
       case DLG_SHADERSET_LOAD2:
         LoadExternalShader(1,hdlg);
         UpdateShaderSetDialog(fa->sh,1,hdlg);
         break;
       case DLG_SHADERSET_LOAD3:
         LoadExternalShader(2,hdlg);
         UpdateShaderSetDialog(fa->sh,2,hdlg);
         break;
       case DLG_SHADERSET_LOAD4:
         LoadExternalShader(3,hdlg);
         UpdateShaderSetDialog(fa->sh,3,hdlg);
         break;
       case DLG_SHADERSET_LOAD5:
         LoadExternalShader(4,hdlg);
         UpdateShaderSetDialog(fa->sh,4,hdlg);
         break;
       case DLG_SHADERSET_LOAD6:
         LoadExternalShader(5,hdlg);
         UpdateShaderSetDialog(fa->sh,5,hdlg);
         break;
       case DLG_SHADERSET_LOAD7:
         LoadExternalShader(6,hdlg);
         UpdateShaderSetDialog(fa->sh,6,hdlg);
         break;
       case DLG_SHADERSET_LOAD8:
         LoadExternalShader(7,hdlg);
         UpdateShaderSetDialog(fa->sh,7,hdlg);
         break;

       case IDCANCEL:
       case DLG_SHADERSET_DONE:
         EndDialog(hdlg,OK);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static void UpdateShaderLoadDialog(int id, HWND hdlg){
 int EditResourceID[8]={DLG_SHADERLOAD_EDIT1,DLG_SHADERLOAD_EDIT2,
                        DLG_SHADERLOAD_EDIT3,DLG_SHADERLOAD_EDIT4,
                        DLG_SHADERLOAD_EDIT5,DLG_SHADERLOAD_EDIT6,
                        DLG_SHADERLOAD_EDIT7,DLG_SHADERLOAD_EDIT8};
 int DropResourceID[8]={DLG_SHADERLOAD_DROP1,DLG_SHADERLOAD_DROP2,
                        DLG_SHADERLOAD_DROP3,DLG_SHADERLOAD_DROP4,
                        DLG_SHADERLOAD_DROP5,DLG_SHADERLOAD_DROP6,
                        DLG_SHADERLOAD_DROP7,DLG_SHADERLOAD_DROP8};
 int NameResourceID[8]={DLG_SHADERLOAD_NAME1,DLG_SHADERLOAD_NAME2,
                        DLG_SHADERLOAD_NAME3,DLG_SHADERLOAD_NAME4,
                        DLG_SHADERLOAD_NAME5,DLG_SHADERLOAD_NAME6,
                        DLG_SHADERLOAD_NAME7,DLG_SHADERLOAD_NAME8};
 if(xShaders[id].in_use == 1){
   EnableWindow(GetDlgItem(hdlg,EditResourceID[id]),TRUE);
   EnableWindow(GetDlgItem(hdlg,DropResourceID[id]),TRUE);
   SetDlgItemText(hdlg,NameResourceID[id],xShaders[id].name);
 }
 else{
   EnableWindow(GetDlgItem(hdlg,EditResourceID[id]),FALSE);
   EnableWindow(GetDlgItem(hdlg,DropResourceID[id]),FALSE);
   SetDlgItemText(hdlg,NameResourceID[id],"<none>");
 }
}

static int EditExternalShader(int id, HWND hdlg){
 HMODULE hLib;
 FARPROC fpFun;
 X__MEMORY_MANAGER *Evs;
 char *(*fpFun1)(char *, HWND, X__MEMORY_MANAGER *);
 Evs=&EviMemoryManager;
 EDIT_ACTION=YES;
 hLib=LoadLibrary(xShaders[id].name);
 if(hLib == NULL){ /* try the default directory */
   char *sfname,tfname[256];
   sfname=short_form(xShaders[id].name);
   strcpy(tfname,gszHomeDir);  strcat(tfname,"textures\\");
   strcat(tfname,sfname);
   hLib=LoadLibrary(tfname);
 }
 if(hLib != NULL){
   if((fpFun=GetProcAddress(hLib,"_SetExternalParameters")) != NULL){
     fpFun1 = (void *)fpFun;
     if(xShaders[id].parameters != NULL)xShaders[id].parameters[0]='#';
     xShaders[id].parameters=(*fpFun1)(xShaders[id].parameters,hdlg,Evs);
   }
   else SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
   FreeLibrary(hLib);
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 if(xShaders[id].parameters == NULL)return FAIL;
 return OK;
}

static void DropExternalShader(int id, HWND hdlg){
 unsigned char ft;
 face *fp;
 int i;
 EDIT_ACTION=YES;
 if(Nface > 0){
   if(id < 4){ /* ordinary preview */
     fp=MainFp; for(i=0;i<Nface;i++){ /* kill this shader from any faces */
       if((fp->attrib & 0x10) == 0x10){
         ft=fp->shader;
         ft=ft & (~(1 << id ));
         fp->shader = ft;
       }
       fp++;
     }
   }
 }
 xShaders[id].in_use=0;
 if(xShaders[id].parameters != NULL)X__Free(xShaders[id].parameters);
 xShaders[id].parameters=NULL;
 UpdateShaderLoadDialog(id,hdlg);
 return;
}

static void LoadExternalShader(int id, HWND hdlg){
 long l;
 l=strlen(gszDLLfile);
 strcpy((gszDLLfile+l-3),"sfx");
 if(SelectSfxFileName(0,gszDLLfile,gszDLLdir,IDX_MISC_LOADXSHADER,
   "(*.TFX)|*.tfx|",hdlg) == TRUE){
   l=strlen(gszDLLfile);
   strcpy((gszDLLfile+l-3),"dll");
   strcpy(xShaders[id].name,gszDLLfile);
   xShaders[id].in_use=1;
   if(xShaders[id].parameters != NULL)X__Free(xShaders[id].parameters);
   xShaders[id].parameters=NULL;
   if(EditExternalShader(id,hdlg) == FAIL ||
      xShaders[id].parameters == NULL)DropExternalShader(id,hdlg);
   else UpdateShaderLoadDialog(id,hdlg);
 }
 return;
}

BOOL CALLBACK LoadExternalShaderDlgProc(HWND hdlg,UINT msg,
                                 WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     UpdateShaderLoadDialog(0,hdlg);
     UpdateShaderLoadDialog(1,hdlg);
     UpdateShaderLoadDialog(2,hdlg);
     UpdateShaderLoadDialog(3,hdlg);
     UpdateShaderLoadDialog(4,hdlg);
     UpdateShaderLoadDialog(5,hdlg);
     UpdateShaderLoadDialog(6,hdlg);
     UpdateShaderLoadDialog(7,hdlg);
     CentreDialogOnCursor(hdlg);
     return (TRUE);
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case DLG_SHADERLOAD_LOAD1:
         LoadExternalShader(0,hdlg);
         break;
       case DLG_SHADERLOAD_LOAD2:
         LoadExternalShader(1,hdlg);
         break;
       case DLG_SHADERLOAD_LOAD3:
         LoadExternalShader(2,hdlg);
         break;
       case DLG_SHADERLOAD_LOAD4:
         LoadExternalShader(3,hdlg);
         break;
       case DLG_SHADERLOAD_LOAD5:
         LoadExternalShader(4,hdlg);
         break;
       case DLG_SHADERLOAD_LOAD6:
         LoadExternalShader(5,hdlg);
         break;
       case DLG_SHADERLOAD_LOAD7:
         LoadExternalShader(6,hdlg);
         break;
       case DLG_SHADERLOAD_LOAD8:
         LoadExternalShader(7,hdlg);
         break;
       case DLG_SHADERLOAD_EDIT1:
        if(xShaders[0].in_use){
           if(EditExternalShader(0,hdlg) == FAIL ||
                  xShaders[0].parameters == NULL)DropExternalShader(0,hdlg);
            else UpdateShaderLoadDialog(0,hdlg);
         }
         break;
       case DLG_SHADERLOAD_EDIT2:
        if(xShaders[1].in_use){
           if(EditExternalShader(1,hdlg) == FAIL ||
                  xShaders[1].parameters == NULL)DropExternalShader(1,hdlg);
            else UpdateShaderLoadDialog(1,hdlg);
         }
         break;
       case DLG_SHADERLOAD_EDIT3:
        if(xShaders[2].in_use){
           if(EditExternalShader(2,hdlg) == FAIL ||
                  xShaders[2].parameters == NULL)DropExternalShader(2,hdlg);
            else UpdateShaderLoadDialog(2,hdlg);
         }
         break;
       case DLG_SHADERLOAD_EDIT4:
        if(xShaders[3].in_use){
           if(EditExternalShader(3,hdlg) == FAIL ||
                  xShaders[3].parameters == NULL)DropExternalShader(3,hdlg);
            else UpdateShaderLoadDialog(3,hdlg);
         }
         break;
       case DLG_SHADERLOAD_EDIT5:
        if(xShaders[4].in_use){
           if(EditExternalShader(4,hdlg) == FAIL ||
                  xShaders[4].parameters == NULL)DropExternalShader(4,hdlg);
            else UpdateShaderLoadDialog(4,hdlg);
         }
         break;
       case DLG_SHADERLOAD_EDIT6:
        if(xShaders[5].in_use){
           if(EditExternalShader(5,hdlg) == FAIL ||
                  xShaders[5].parameters == NULL)DropExternalShader(5,hdlg);
            else UpdateShaderLoadDialog(5,hdlg);
         }
         break;
       case DLG_SHADERLOAD_EDIT7:
        if(xShaders[6].in_use){
           if(EditExternalShader(6,hdlg) == FAIL ||
                  xShaders[6].parameters == NULL)DropExternalShader(6,hdlg);
            else UpdateShaderLoadDialog(5,hdlg);
         }
         break;
       case DLG_SHADERLOAD_EDIT8:
        if(xShaders[7].in_use){
           if(EditExternalShader(7,hdlg) == FAIL ||
                  xShaders[7].parameters == NULL)DropExternalShader(7,hdlg);
            else UpdateShaderLoadDialog(7,hdlg);
         }
         break;
       case DLG_SHADERLOAD_DROP1:
         DropExternalShader(0,hdlg);
         break;
       case DLG_SHADERLOAD_DROP2:
         DropExternalShader(1,hdlg);
         break;
       case DLG_SHADERLOAD_DROP3:
         DropExternalShader(2,hdlg);
         break;
       case DLG_SHADERLOAD_DROP4:
         DropExternalShader(3,hdlg);
         break;
       case DLG_SHADERLOAD_DROP5:
         DropExternalShader(4,hdlg);
         break;
       case DLG_SHADERLOAD_DROP6:
         DropExternalShader(5,hdlg);
         break;
       case DLG_SHADERLOAD_DROP7:
         DropExternalShader(6,hdlg);
         break;
       case DLG_SHADERLOAD_DROP8:
         DropExternalShader(7,hdlg);
         break;
       case DLG_SHADERLOAD_PREVIEW:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_VIEW_RENDERINWINDOW,0);
         break;
       case IDCANCEL:
       case DLG_SHADERLOAD_DONE:
         EndDialog(hdlg,OK);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
}


