/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file mstruct.h */

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

#ifdef __X__NT__
#define _LINE_DRAW 1
#define X__SetPixel(a,b,c,d) SetPixelV(a,b,c,d)
#else
#define X__SetPixel(a,b,c,d) SetPixel(a,b,c,d)
#endif
#ifdef __X__NT__
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#else
#define X__Malloc(s)    GlobalAlloc(GMEM_FIXED,s)
#define X__Free(p)      GlobalFree((HGLOBAL)p)
#define X__Realloc(p,s) GlobalReAlloc((HGLOBAL)p,s,GMEM_MOVEABLE)
//#define X__Malloc(s)    malloc(s)
//#define X__Free(s)      free(s)
//#define X__Realloc(p,s) realloc(p,s)
#endif

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                         }

#include "struct.h"
#include "..\animate\memory.h"

typedef struct {
       vertex Origin;
       long Xdirection[3];
       long Ydirection[3];
       long Zdirection[3];
       vertex Offset;
} axis;

typedef struct {
  char  name[8];
  short Nc,Np,flag;
  long *Vc,*Vp;
} bezier_curve;

typedef struct tagPRIMITIVE {
  unsigned char fc[3];
  int nsides,smooth,faces,ml,mw,nsteps;
  double length,width,height,radius;
  double lr,fb,ud;
  BOOL auto_interactive;
  int nvert,nhorz;
  double torus_ratio;
} PRIMITIVE;

typedef struct tagPREFERENCES {
  BOOL show_select;
  BOOL buttons;
  BOOL quick_trick;
  BOOL undo;
  BOOL use_3d_dialogs;
  BOOL use_OpenGL;
  unsigned char Vcolour[3];
  BOOL tooltips;
  BOOL attribs_sheet;
  BOOL bigv;
  BOOL use_positions;
  BOOL toolsmall;
  BOOL toolleft;
  BOOL background;
} PREFERENCES;

#define ALL_PANNELS     0x0FFF
#define MAIN_WINDOW     0x0001
#define SELECT_PANNEL   0x0002
#define COMMAND_PANNEL  0x0004
#define ZOOM_PANNEL     0x0008
#define TOOL_PANNEL     0x0010
#define ACTION_PANNEL   0x0020
#define COORD_PANNEL    0x0040
#define LOCK_PANNEL     0x0080

#define PLANE            0              /* brush map types */
#define CYLINDER         1
#define PLANE_MOZIAC     2
#define CYLINDER_MOZIAC  3
#define MAP_BY_VERTEX    4
#define MAP_SPHERICAL    5
#define OPAQ             0              /* brush color 0   */
#define CLEAR            1
#define GLOSS            0              /* brush shiny     */
#define MATT             1

#define TRIVIEW          0
#define ONEVIEW          1
#define TRITOP           0
#define TRIFRONT         1
#define TRIRIGHT         2

/* Macro Function Definitions */
#define POINT2VECTOR(p,v) v[0]=(double)p[0]; v[1]=(double)p[1]; \
                          v[2]=(double)p[2];
#define VECTOR2POINT(v,p) p[0]=(long)v[0]; p[1]=(long)v[1]; \
                          p[2]=(long)v[2];

/* definition of unit dimensions      */
#define FARAWAY      1.0e30
#define UNIT         32768L
#define UNIT2        65536L
#define MINUNIT       2048L     // was 1024 trying larger to prevent error
              //1073741824L
//#define MAXUNIT    4194304L
#define MAXUNIT   33554432L
#ifndef PI
#define PI 3.1415926
#endif
#define PI2            PI/2

/* definitionof  sizes etc            */
#define NMAPSMAX       128
#define NAXISMAX       128

/* definition of flags and identifiers*/
#define DESELECTED       0
#define SELECTED         1
#define HIDDEN           2
#define INEDITOR         3
#define STATICBRUSH      0
#define ANIMBRUSH        1
#define BUMPBRUSH        2
#define REFLBRUSH        3
#define PSETBRUSH        4
#define ATTRACT          0
#define REPEL            1

/* general defines  */
#define NO               0
#define YES              1
#define FAIL            -1
#define OK               1
#define MAX_DIR        255
#define MAX_FILE       255
#define ZOOMOUT          0
#define ZOOMIN           1
#define PANUP            2
#define PANDOWN          3
#define PANLEFT          4
#define PANRIGHT         5
#define PANFORWARD       6
#define PANBACK          7
#define RECENTRE         8
#define CENTREWORLD      9

/* tools and actions */

#define NTOOLS          12
#define NOTOOL           0
#define NODETOOL         1
#define PAN              2
#define INZOOM           3
#define GRABBER          4
#define ROTATOR          5
#define EXPANDER         6
#define PLOTTER          7
#define SHAPER           8
#define MAGNETTO         9
#define BUILDER         10
#define PLOT3D          11

#define PRIMS           12
#define SKANIMATE       13
#define SELECTOR        14
#define DESELECTOR      15

#define SKGRAB          16
#define SKADD           17
#define SKDELETE        18
#define SKSUBDIV        19
#define SKASSIGN        20
#define SKSELECT        21
#define SKATMAP         22
#define SKTWIST         23

#define MAXTOOLS        24

