/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  ORIENT.C  */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#include "struct.h"
#include "dstruct.h"

#define DLG_CUT 600

#include "orient.h"

#define DESELECTED 0
#define SELECTED   1
#define INEDITOR   3

static HWND      hParent;
static HINSTANCE hThisInstance;

static void Orientate(HWND);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=hDLL;
      if(hDLL == NULL)MessageBox( GetFocus()," NULL instance",NULL, MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

static BOOL CALLBACK OrientateDlgProc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     CentreDlgOnS(hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 HCURSOR hSave;
 lpEVI=lpevi;
 hParent=parent_window;
 if(DialogBox(hThisInstance,MAKEINTRESOURCE(DLG_CUT),hParent,
             (DLGPROC)OrientateDlgProc) == FALSE)return FALSE;
 hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
 Orientate(parent_window);
 SetCursor(hSave);
 return TRUE;
}

static int Nchange,Npieces;
static workface *WF;

static void MakeSameOrientation(long last, long id){
 face *f;
 long i0,i1,i2,l0,l1,l2,a0,a1,a2;
 f=(MainFp+WF[last].f); l0=f->V[0]; l1=f->V[1]; l2=f->V[2];
 f=(MainFp+WF[id].f);   i0=f->V[0]; i1=f->V[1]; i2=f->V[2];
 a0=WF[id].fa[0]; a1=WF[id].fa[1]; a2=WF[id].fa[2];
 if((i2 == l2 && i1 == l1) ||
    (i2 == l1 && i1 == l0) ||
    (i2 == l0 && i1 == l2)){
   WF[id].fa[2]=a0; WF[id].fa[0]=a2;
   f->V[2]=i1; f->V[1]=i2; Nchange++;
   return;
 }
 if((i0 == l2 && i2 == l1) ||
    (i0 == l1 && i2 == l0) ||
    (i0 == l0 && i2 == l2)){
   WF[id].fa[0]=a1; WF[id].fa[1]=a0;
   f->V[0]=i2; f->V[2]=i0; Nchange++;
   return;
 }
 if((i1 == l1 && i0 == l0) ||
    (i1 == l2 && i0 == l1) ||
    (i1 == l0 && i0 == l2)){
   WF[id].fa[1]=a2; WF[id].fa[2]=a1;
   f->V[1]=i0; f->V[0]=i1; Nchange++;
   return;
 }
 return;
}

static void SetConnectingID(long last,long id){
 if(WF[id].id == 1)return;
 WF[id].id=1;
 if(last >= 0)MakeSameOrientation(last,id);
 last=id;
 if(WF[id].fa[0] >= 0)SetConnectingID(last,WF[id].fa[0]);
 if(WF[id].fa[1] >= 0)SetConnectingID(last,WF[id].fa[1]);
 if(WF[id].fa[2] >= 0)SetConnectingID(last,WF[id].fa[2]);
}

static void Reorientate(workface *wf, long nf){
 long i,id;
 WF=wf;
 Nchange=0; Npieces=0;
 for(i=0;i<nf;i++)wf[i].id = -1;
 while(1){ /* keep going until all the bits have been processed */
   for(id=0;id<nf;id++){
     if(wf[id].id == -1)goto PROCESSIT;
   }
   break;      /* no bits have been found - so breakout */
   PROCESSIT:
   SetConnectingID(-1,id);
   Npieces++;
 }
 return;
}

static int FindAdjFaces(workface *wf, long nf){
 face *f;
 vertex *vp;
 vtxadj *vl=NULL;
 long c,nvl=0;
 long *fl,i,j,k,id,id0,id1,id2,fjd,fkd,n,m,kk,nl[3],ml[3];
 vp=MainVp; for(c=0;c<Nvert;c++){
   if(vp->status == SELECTED){ /* work vertex */
     vp->id=nvl;
     nvl++;
   }
   else vp->id = -10; /* just for fun */
   vp++;
 }
 if((vl=(vtxadj *)X__Malloc(nvl*sizeof(vtxadj))) == NULL)return 0;
 for(i=0;i<nvl;i++){
   vl[i].nf=0;
   vl[i].flist=NULL;
 }
 for(i=0;i<nf;i++){ /* make list of work faces attached to each vertex */
   f=(MainFp+wf[i].f);
   for(j=0;j<3;j++){
     id=(MainVp+(f->V[j]))->id;
     fl=vl[id].flist;
     n=vl[id].nf;
     if(fl == NULL){
       if((fl=(long *)X__Malloc(sizeof(long))) == NULL)return 0;
     }
     else{
       if((fl=(long *)X__Realloc(fl,(n+1)*sizeof(long))) == NULL)return 0;
     }
     fl[n]=i;
     vl[id].flist=fl;
     vl[id].nf=n+1;
   }
 }
 for(i=0;i<nf;i++){ /* find faces that are adjacent to face i */
   f=(MainFp+wf[i].f);
   id0=(MainVp+(f->V[0]))->id;
   id1=(MainVp+(f->V[1]))->id;
   id2=(MainVp+(f->V[2]))->id;
   nl[0]=id0; ml[0]=id1;
   nl[1]=id1; ml[1]=id2;
   nl[2]=id2; ml[2]=id0;
   for(kk=0;kk<3;kk++){
     wf[i].fa[kk] = -1;
     if((n=vl[nl[kk]].nf) > 0 && (m=vl[ml[kk]].nf) > 0){
       for(j=0;j<n;j++){
         fjd=vl[nl[kk]].flist[j];
         for(k=0;k<m;k++){
           fkd=vl[ml[kk]].flist[k];
           if(fkd == fjd && fkd != i){
             wf[i].fa[kk]=fkd;
             goto ENDF;
           }
         }
       }
     }
     ENDF:;
   }
 }
 for(i=0;i<nvl;i++){
   if(vl[i].flist != NULL)X__Free(vl[i].flist);
 }
 X__Free(vl);
 return nvl;
}

static void Orientate(HWND hWnd){
 /* workface is the face being worked on */
 /* tool face is the face doing the cutting */
 char str[256];
 workface *w;
 face *ff1;
 vertex *v0,*v1,*v2;
 long c,i,j;
 long n_work_face=0;
 if(NvertSelect == 0 || NvertSelect > 64000)return;
 if((ff1=MainFp) == NULL)return;
 w=NULL;
 for(c=0;c<Nface;c++){ /* find number of work and tool faces */
   v0=(MainVp+ff1->V[0]); v1=(MainVp+ff1->V[1]); v2=(MainVp+ff1->V[2]);
   if(v0->status == SELECTED &&
      v1->status == SELECTED &&
      v2->status == SELECTED)n_work_face++;
   ff1++;
 }
 if((w=(workface *)X__Malloc(n_work_face*(long)sizeof(workface)))
       == NULL)return;
 n_work_face=0;
 for(ff1=MainFp,c=0;c<Nface;c++){
  v0=(MainVp+ff1->V[0]); v1=(MainVp+ff1->V[1]); v2=(MainVp+ff1->V[2]);
  if(v0->status == SELECTED &&
     v1->status == SELECTED &&
     v2->status == SELECTED){
   w[n_work_face].f=c;
    for(i=0;i<3;i++){
      w[n_work_face].fa[i] = -1;
    }
    n_work_face++;
  }
  ff1++;
 }
 if(n_work_face != 0){
   FindAdjFaces(w,n_work_face);
   Reorientate(w,n_work_face);
 }
 if(w != NULL)X__Free(w);
 sprintf(str,"%ld Selected Faces in   \n%ld Separate pieces.   \n%ld Faces re-orientated  ",
         n_work_face,Npieces,Nchange);
 MessageBox(hWnd,str,"Orientate Faces",MB_OK|MB_ICONINFORMATION);
 return;
}

