/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  PREMAP.C  DLL to choose from preexiting maps */

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>

#ifdef __X__NT__
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#else
#define X__Malloc(s)    GlobalAlloc(GMEM_FIXED,s)
#define X__Free(p)      GlobalFree((HGLOBAL)p)
#define X__Realloc(p,s) GlobalReAlloc((HGLOBAL)p,s,GMEM_MOVEABLE)
#endif

#define FI_DIB 0
#define PALVERSION  0x300
#define MAXPALETTE  256
#define MAXTEXT     256

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif
static void FreeDib(void);
static LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,
                             LPARAM lParam);
static void AppPaint(HWND hWnd,HDC hDC,int x,int y);
static void LoadSelectionBitmap(char *, HWND);

static int       Selection = -1;
static HWND      hParent;
static HINSTANCE hThisInstance;
static HPALETTE  hpalCurrent    = NULL;         /* Handle to current palette            */
static HBITMAP   hbmCurrent     = NULL;         /* Handle to current memory BITMAP      */
static LPSTR     lpBitmapHeader = NULL;
static LPSTR     lpBitmapBits   = NULL;
static char *szViewClass="OFX:PremapClass";
static DWORD fdwStyle = WS_POPUP | WS_CAPTION | WS_SYSMENU |
                        !WS_MAXIMIZEBOX | !WS_MINIMIZEBOX | WS_THICKFRAME;

static int Viewer(char *mapfile){
 MSG msg;
 WNDCLASS wndclass;
 HWND hWnd;
 int xScreen,yScreen;
 wndclass.style         = 0;
 wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
 wndclass.cbClsExtra    = 0 ;
 wndclass.cbWndExtra    = 0 ;
 wndclass.hInstance     = hThisInstance ;
 wndclass.hIcon         = NULL;
 wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
 wndclass.hbrBackground = GetStockObject (WHITE_BRUSH) ;
 wndclass.lpszMenuName  = NULL;
 wndclass.lpszClassName = szViewClass;
 if (!RegisterClass (&wndclass))  return -1;
 xScreen = GetSystemMetrics (SM_CXSCREEN) ;
 yScreen = GetSystemMetrics (SM_CYSCREEN) ;
 hWnd = CreateWindow(
        szViewClass,
        "Click on Desired Postage Stamp Image",
        fdwStyle,
        25,
        25,
        xScreen-50,
        yScreen-50,
        hParent,
        NULL,
        hThisInstance,
        NULL) ;
 if(hWnd == NULL)return -1;
 EnableWindow(hParent,FALSE);
 LoadSelectionBitmap(mapfile,hWnd);
 ShowWindow (hWnd,SW_SHOWNA);
 while(GetMessage (&msg,NULL,0,0)){
   TranslateMessage (&msg);
   DispatchMessage (&msg);
 }
 if(IsWindow(hWnd))DestroyWindow(hWnd);
 UnregisterClass(wndclass.lpszClassName,wndclass.hInstance);
 EnableWindow(hParent,TRUE);
 return -1;
}

#define NMATRIX 6
#define NROWS 200
#define HCOLS 320
#define LIMIT 30

static LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,
                             LPARAM lParam){
 static BOOL bToolCaptured=FALSE;
 static HCURSOR hCursorSave;
 static POINT LastPt;
 PAINTSTRUCT ps;
 HDC hDC;
 int i;
 RECT rc;
 POINT pt;
 switch (iMessage) {
   case WM_DESTROY:
    FreeDib();
    break;
  case WM_RBUTTONUP:
    break;
  case WM_LBUTTONDOWN:
    Selection=1+((HIWORD(lParam)*NMATRIX)/NROWS)*NMATRIX
               + (LOWORD(lParam)*NMATRIX)/HCOLS;
    if(Selection > LIMIT || Selection < 0){
      MessageBeep(MB_OK);
      Selection = -1;
    }
    break;
  case WM_LBUTTONUP:
//{char str[32];
//sprintf(str,"%ld",Selection);
//MessageBox(NULL,str,"UP",MB_OK);
//}
    if(Selection > 0 && Selection <= LIMIT)PostQuitMessage(0);
    break;
  case WM_PAINT:
    hDC = BeginPaint(hWnd, &ps);
    AppPaint(hWnd,hDC,0,0);
    EndPaint(hWnd,&ps);
    break ;
  case WM_QUERYNEWPALETTE:{
      HDC hDC;
      HPALETTE hpalT;
      if(hpalCurrent == NULL)break;
      hDC=GetDC(hWnd);
      hpalT = SelectPalette(hDC,hpalCurrent,FALSE);
      if(RealizePalette (hDC) > 0)InvalidateRect (hWnd, (LPRECT) (NULL),FALSE);
      SelectPalette(hDC,hpalT,FALSE);
      ReleaseDC(hWnd,hDC);
      return TRUE;
    }
  case WM_PALETTECHANGED:
    if ((HWND)wParam != hWnd ){
      HDC hDC;
      HPALETTE hpalT;
      if(hpalCurrent == NULL)break;
      hDC=GetDC(hWnd);
      hpalT = SelectPalette(hDC,hpalCurrent,FALSE);
      if(RealizePalette (hDC) > 0)InvalidateRect (hWnd, (LPRECT) (NULL),FALSE);
      SelectPalette(hDC,hpalT,FALSE);
      ReleaseDC(hWnd,hDC);
    }
    break;
  case WM_SYSCOMMAND:
    if ((wParam & 0xfff0) == SC_SIZE)break;
    if ((wParam & 0xfff0)== SC_MINIMIZE)break;
    if ((wParam & 0xfff0)== SC_MAXIMIZE)break;
    if ((wParam & 0xfff0)== SC_CLOSE)PostQuitMessage(0);
    return DefWindowProc (hWnd, iMessage, wParam, lParam) ;
  default:
    return DefWindowProc (hWnd, iMessage, wParam, lParam) ;
 }
 return 0L ;
}

static void FreeDib(void){
 if (hpalCurrent) DeleteObject(hpalCurrent);
 if (hbmCurrent)  DeleteObject(hbmCurrent);
 if(lpBitmapBits   != NULL)X__Free((char *)lpBitmapBits);
 if(lpBitmapHeader != NULL)X__Free((char *)lpBitmapHeader);
 lpBitmapHeader = NULL;
 lpBitmapBits   = NULL;
 hpalCurrent    = NULL;
 hbmCurrent     = NULL;
}

static void AppPaint(HWND hWnd,HDC hDC,int x,int y){
 RECT rc;
 HPALETTE hpalT;
 POINT Origin;
 HDC hDCbits;
 BITMAP bm;
 GetClientRect(hWnd,&rc);
 if(hpalCurrent != NULL){
   hpalT = SelectPalette(hDC,hpalCurrent,FALSE);
   RealizePalette(hDC);
   if(hbmCurrent != NULL){ /* compatible bit map exists; its quicker, use it */
     hDCbits = CreateCompatibleDC(hDC);
     GetObject(hbmCurrent,sizeof(BITMAP),(LPSTR)&bm);
     SelectObject(hDCbits,hbmCurrent);
     BitBlt(hDC,0,0,bm.bmWidth,bm.bmHeight,hDCbits,0,0,SRCCOPY);
     DeleteDC(hDCbits);
   }
   else{
     SetDIBitsToDevice(hDC,0,0,
         (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biWidth,
         (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biHeight,
         0,0,
         0,
         (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biHeight,
         lpBitmapBits,
         (LPBITMAPINFO)lpBitmapHeader,
         DIB_RGB_COLORS);
   }
   SelectPalette(hDC,hpalT,FALSE);
 }
}

static HBITMAP BitmapFromDib(LPSTR lpHeader,LPSTR lpBits,HPALETTE hpal){
 HDC       hdc;         /* make BITMAP from a memory DIB */
 HBITMAP   hbm;
 HPALETTE  hpalT;
 if (lpHeader == NULL || lpBits == NULL)return NULL;
 hdc = GetDC(NULL);
 if(hpal != NULL){
   hpalT = SelectPalette(hdc,hpal,FALSE);
   RealizePalette(hdc);
 }
 hbm = CreateDIBitmap(hdc,
                (LPBITMAPINFOHEADER)lpHeader,
                (LONG)CBM_INIT,
                lpBits,
                (LPBITMAPINFO)lpHeader,
                DIB_RGB_COLORS );
 if(hpalCurrent)SelectPalette(hdc,hpalT,FALSE);
 ReleaseDC(NULL,hdc);
 return hbm;
}

static void LoadSelectionBitmap(char *mapfile, HWND hWnd){
 LOGPALETTE *pPal;
 FILE *dibfile;
 LPSTR s;
 BITMAPFILEHEADER hdr,*lphdr;
 BITMAPINFOHEADER *lpbi;
 DWORD ncolours;
 RGBQUAD *pRgb;
 long i,k,imagesize;
 lphdr=&hdr;
 if((dibfile=fopen(mapfile,"rb")) == NULL)return;
 pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE)+256*sizeof(PALETTEENTRY));
 if(pPal == NULL)return;
 lpBitmapHeader=(LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+
                                   256L*sizeof(RGBQUAD));
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 if(lpBitmapHeader == NULL){
   LocalFree((HANDLE)pPal);
   return;
 }
 pPal->palNumEntries = 256;
 pPal->palVersion    = PALVERSION;
 fread(lphdr,1,sizeof(BITMAPFILEHEADER),dibfile);
 fread(lpbi,1,sizeof(BITMAPINFOHEADER),dibfile);
 lpbi->biCompression=BI_RGB;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrImportant=256;
//{
//char str[128];
//sprintf(str,"(%ldx%ld) %ld\n%ld %ld\n%ld %d\n",lpbi->biWidth,lpbi->biHeight,lpbi->biSizeImage,
// sizeof(BITMAPFILEHEADER),sizeof(BITMAPINFOHEADER),lpbi->biClrUsed,lpbi->biBitCount);
//MessageBox(NULL,str,"test",MB_OK);
//}
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<256;i++){
   fread(&pRgb[i],1,sizeof(RGBQUAD),dibfile);
   pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
   pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
   pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
   pPal->palPalEntry[i].peFlags = (BYTE)0;
 }
 hpalCurrent = CreatePalette(pPal);
 LocalFree((HANDLE)pPal);
 imagesize=(long)lpbi->biHeight * (long)lpbi->biWidth;
 if((lpBitmapBits=(LPSTR)X__Malloc(imagesize)) == NULL){
   FreeDib();
   return;
 }
 fread(lpBitmapBits,1,imagesize,dibfile);
 { RECT Rectangle;
   Rectangle.left   = 0;
   Rectangle.top    = 0;
   Rectangle.right  = lpbi->biWidth;
   Rectangle.bottom = lpbi->biHeight;
   AdjustWindowRect (&Rectangle,fdwStyle,TRUE);
   SetWindowPos (hWnd, (HWND)NULL, 0, 0,
        Rectangle.right  - Rectangle.left,
        Rectangle.bottom - Rectangle.top-GetSystemMetrics(SM_CYCAPTION),
        SWP_NOMOVE | SWP_NOZORDER);
 }
// BitmapFromDib(lpBitmapHeader,lpBitmapBits,hpalCurrent);
 fclose(dibfile);
}

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
//      MessageBox (GetFocus(),"Attaching","Debug",MB_OK);
#endif
      hThisInstance=hDLL;
      break;
    }
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
//      MessageBox (GetFocus(),"Detaching","Debug",MB_OK);
#endif
      break;
  }
  return TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

BOOL _GetPresetMapImage
 (HWND parent, char *mapfile){
 hParent=parent;
 Selection= -1;
 Viewer(mapfile);
 return Selection;
}
