/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  QUADRIC.CPP  */

#include "vtk.hh"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#define UNIT         32768L
#define DESELECTED       0
#define SELECTED         1
#define INEDITOR         6

#ifdef __cplusplus
extern "C" {
#endif

#include "struct.h"
#include "dstruct.h"

#ifdef __cplusplus
}
#endif

#include "quadric.h"

#include "sfx2vtk.h"
#include "sfx2vtk.cpp"

static HWND      hParent;
static HINSTANCE hThisInstance;

static BOOL CALLBACK SfxDlgProc(HWND hwnd, UINT msg,
                                WPARAM wparam, LPARAM lparam);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=(HINSTANCE)hDLL;
      if(hDLL == NULL)MessageBeep(MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}


extern "C" BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 HCURSOR hSave;
 static struct bp {
  BOOL cap;
  float xmin,xmax,ymin,ymax,zmin,zmax;
  int i,j,k;
  float cv;
  float a0,a1,a2,a3,a4,a5,a6,a7,a8,a9;
 } d;
 lpEVI=lpevi;
 hParent=parent_window;
 d.cap = TRUE;
 d.xmin = d.ymin = d.zmin = -1.0f;
 d.xmax = d.ymax = d.zmax =  1.0f;
 d.cv = 0.2;
 d.i = d.j = d.k = 20;
 d.a0=0.5;
 d.a1=1.0;
 d.a2=2,0;
 d.a3=0.0;
 d.a4=0.1;
 d.a5=0.0;
 d.a6=0.0;
 d.a7=0.2;
 d.a8=0.0;
 d.a9=0.0;
 if(DialogBoxParam(hThisInstance,MAKEINTRESOURCE(DLG_VTK),hParent,
             (DLGPROC)SfxDlgProc,(LPARAM)&d) == FALSE)return FALSE;
 hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
 {
  SfxWriter sfxoutput;
  vtkQuadric *quadric = new vtkQuadric;
  quadric->SetCoefficients(d.a0,d.a1,d.a2,d.a3,d.a4,d.a5,d.a6,d.a7,d.a8,d.a9);

  vtkSampleFunction *sample = new vtkSampleFunction;
  sample->SetSampleDimensions(d.i,d.j,d.k);
  sample->SetModelBounds(d.xmin,d.xmax,d.ymin,d.ymax,d.zmin,d.zmax);
  sample->SetCapping((int)d.cap);
  sample->SetImplicitFunction(quadric);

  vtkContourFilter *cf = new vtkContourFilter;
  cf->SetInput(sample->GetOutput());
  cf->SetValue(0,d.cv);

  vtkCleanPolyData *cp = new vtkCleanPolyData;
  cp->SetTolerance(0.001);
  cp->SetInput(cf->GetOutput());

  vtkTriangleFilter *tf = new vtkTriangleFilter;
  tf->SetInput(cp->GetOutput());

  sfxoutput.SetInput(tf->GetOutput());
  sfxoutput.Write();

  delete tf;
  delete cp;
  delete cf;
  delete sample;
  delete quadric;

 }
 SetCursor(hSave);
 return TRUE;
}

static BOOL CALLBACK SfxDlgProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam){
 BOOL err;
 double f;
 int i;
 static struct bp {
  BOOL cap;
  float xmin,xmax,ymin,ymax,zmin,zmax;
  int i,j,k;
  float cv;
  float a0,a1,a2,a3,a4,a5,a6,a7,a8,a9;
 } *dp;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     dp=(struct bp *)lparam;
     SetDlgItemInt(hwnd,DLG_VTK_XRES,dp->i,FALSE);
     SetDlgItemInt(hwnd,DLG_VTK_YRES,dp->j,FALSE);
     SetDlgItemInt(hwnd,DLG_VTK_ZRES,dp->k,FALSE);
     sprintf(str,"%.2lf",dp->a0);
     SetDlgItemText(hwnd,DLG_VTK_a0,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a1);
     SetDlgItemText(hwnd,DLG_VTK_a1,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a2);
     SetDlgItemText(hwnd,DLG_VTK_a2,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a3);
     SetDlgItemText(hwnd,DLG_VTK_a3,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a4);
     SetDlgItemText(hwnd,DLG_VTK_a4,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a5);
     SetDlgItemText(hwnd,DLG_VTK_a5,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a6);
     SetDlgItemText(hwnd,DLG_VTK_a6,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a7);
     SetDlgItemText(hwnd,DLG_VTK_a7,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a8);
     SetDlgItemText(hwnd,DLG_VTK_a8,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->a9);
     SetDlgItemText(hwnd,DLG_VTK_a9,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->cv);
     SetDlgItemText(hwnd,DLG_VTK_CONTOUR,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->xmin);
     SetDlgItemText(hwnd,DLG_VTK_XMIN,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->xmax);
     SetDlgItemText(hwnd,DLG_VTK_XMAX,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->ymin);
     SetDlgItemText(hwnd,DLG_VTK_YMIN,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->ymax);
     SetDlgItemText(hwnd,DLG_VTK_YMAX,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->zmin);
     SetDlgItemText(hwnd,DLG_VTK_ZMIN,(LPCTSTR)str);
     sprintf(str,"%.2lf",dp->zmax);
     SetDlgItemText(hwnd,DLG_VTK_ZMAX,(LPCTSTR)str);
     if(dp->cap)SendDlgItemMessage(hwnd,DLG_VTK_CAP,BM_SETCHECK,TRUE,0);
     CentreDlgOnS(hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          i=GetDlgItemInt(hwnd,DLG_VTK_XRES,&err,FALSE);
          if(err)dp->i=i;
          i=GetDlgItemInt(hwnd,DLG_VTK_YRES,&err,FALSE);
          if(err)dp->j=i;
          i=GetDlgItemInt(hwnd,DLG_VTK_ZRES,&err,FALSE);
          if(err)dp->k=i;
          if(GetDlgItemText(hwnd,DLG_VTK_a0,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a0=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a1,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a1=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a2,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a2=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a3,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a3=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a4,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a4=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a5,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a5=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a6,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a6=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a7,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a7=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a8,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a8=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_a9,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->a9=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_CONTOUR,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->cv=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_XMIN,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->xmin=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_XMAX,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->xmax=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_YMIN,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->ymin=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_YMAX,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->ymax=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_ZMIN,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->zmin=atof(str);
          if(GetDlgItemText(hwnd,DLG_VTK_YMAX,str,10) == 0){
            EndDialog(hwnd,FALSE); return TRUE;
          }
          dp->ymax=atof(str);
          if(SendDlgItemMessage(hwnd,DLG_VTK_CAP,BM_GETCHECK,0,0))
               dp->cap=TRUE;
          else dp->cap=FALSE;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

