/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  RAWDATA.CPP  read/write raw data in VTK text format */

#include "vtk.hh"
#include "vtkpolywriter.hh"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#define UNIT         32768L
#define DESELECTED       0
#define SELECTED         1
#define INEDITOR         6

#ifdef __cplusplus
extern "C" {
#endif

#include "struct.h"
#include "dstruct.h"

#ifdef __cplusplus
}
#endif

#include "rawdata.h"

#include "sfx2vtk.h"
#include "sfx2vtk.cpp"

static HWND      hParent;
static HINSTANCE hThisInstance;
static BOOL  i_read=TRUE;

static BOOL LocalSelectFileName(char *szfile, char *szfilter,
                           HWND parent, HINSTANCE hInst);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=(HINSTANCE)hDLL;
      if(hDLL == NULL)MessageBeep(MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

extern "C" BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 HCURSOR hSave;
 char localfile[256];
 lpEVI=lpevi;
 localfile[0]='\0';
 hParent=parent_window;
 if(!LocalSelectFileName(localfile,
                        "VTK Structured Points Datasets (*.VTK)|*.vtk",
                         parent_window,hThisInstance))return FALSE;
 hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
 {
  if(i_read){
    SfxWriter sfxoutput;
    vtkPolyReader *volR = new vtkPolyReader;
    volR->SetFilename(localfile);
    vtkCleanPolyData *cp = new vtkCleanPolyData;
    cp->SetTolerance(0.0001);
    cp->SetInput(volR->GetOutput());
    vtkTriangleFilter *tf = new vtkTriangleFilter;
    tf->SetInput(cp->GetOutput());
    sfxoutput.SetInput(volR->GetOutput());
    sfxoutput.Write();
    delete cp;
    delete tf;
    delete volR;
  }
  else{
    SfxReader sfxinput;
    vtkPolyWriter *volR = new vtkPolyWriter;
    volR->SetFilename(localfile);
    volR->SetInput(sfxinput.GetOutput());
    volR->Update();
    delete volR;
  }
 }
 SetCursor(hSave);
 return TRUE;
}

static unsigned CALLBACK NameHookProc(HWND hwnd,UINT msg,
                              WPARAM wparam,LPARAM lparam){
 char str[32];
 BOOL err;
 LPOFNOTIFY lp;
 switch( msg ) {
   case WM_INITDIALOG:
     if(i_read)SendDlgItemMessage(hwnd,DLG_IO,BM_SETCHECK,1,0);
     return FALSE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDOK:{
           lp=NULL;
           goto JUMPHERE;
         }
         break;
       default: break;
     }
     break;
   case WM_NOTIFY:{
       lp=(LPOFNOTIFY)lparam;
       if(lp->hdr.code == CDN_FILEOK){
         JUMPHERE:
         if(SendDlgItemMessage(hwnd,DLG_IO,BM_GETCHECK,0,0))
         i_read=TRUE; else i_read=FALSE;
       }
     }
     break;
   default: break;
 }
 return FALSE;
}

static BOOL LocalSelectFileName(char *szfile, char *szfilter,
                           HWND parent, HINSTANCE hInst){
 int i;
 OPENFILENAME ofn;
 char szFilter[80],title[256];
 DWORD version;
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=0;
 ofn.lpstrFile=szfile;
 ofn.nMaxFile=255;
 ofn.lpstrFileTitle=NULL;
 ofn.nMaxFileTitle=0;
 ofn.lpstrInitialDir=NULL;
 LoadString(hThisInstance,IDX_VTK_STRING1,title,255);
 ofn.lpstrTitle=title;
 ofn.hInstance=hInst;
 ofn.lpfnHook=NameHookProc;
 version=GetVersion();
 ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|
           OFN_ENABLEHOOK|OFN_ENABLETEMPLATE|OFN_NONETWORKBUTTON;
 if(LOBYTE(LOWORD(version)) >= 4){
   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_VTK);
   ofn.Flags |= OFN_EXPLORER;
 }
 else{
   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_VTKOLD);
 }
 return GetOpenFileName(&ofn);
}

#if 0
  vol->GetPointData()->SetScalars(scalars);
  scalars->Delete();
  vtkDataSetWriter dw;
  dw.SetFilename("F:\\temp1.vtk");
  dw.SetInput(vol);
  dw.Update();
#endif

