/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  SECTIONS.C  */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#include "struct.h"
#include "dstruct.h"

#include "sections.h"

#define DESELECTED 0
#define SELECTED   1

static HWND      hParent;
static HINSTANCE hThisInstance;
static BOOL      i_full=TRUE;
static double    scale_v=1.0;

static BOOL LocalSelectFileName(char *szfile, char *szfilter,
                           HWND parent, HINSTANCE hInst);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=hDLL;
      if(hDLL == NULL)MessageBox ( GetFocus()," NULL instance",NULL, MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 char localfile[256];
 lpEVI=lpevi;
 localfile[0]='\0';
 hParent=parent_window;
 if(!LocalSelectFileName(localfile,
                        "Data Files (*.DAT)|*.dat",
                         parent_window,hThisInstance))return FALSE;
 else{
   HCURSOR hSave;
   FILE *fp;
   vertex *Vp;
   edge   *Ep;
   long   i,j,nsections,npoints,id;
   double x,y,z,scale=32768.0/327.0*scale_v;
   hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
   if((fp = fopen(localfile,"r")) != NULL){
     fscanf(fp,"%ld %ld",&nsections,&npoints);
//npoints -= 4;
     if(nsections > 0 && npoints > 0){
       for(j=0;j<nsections;j++){
         if(!UpdateVertexHeap(Nvert+npoints))goto XXIT;
         if(!UpdateEdgeHeap(Nedge+npoints))goto XXIT;
         for(i=0;i<npoints;i++){
           fscanf(fp,"%lf %lf %lf",&x,&y,&z);
           CreateVertex();
           Vp=(MainVp+Nvert-1);
           Vp->xyz[0]=(long)(x*scale);
           Vp->xyz[1]=(long)(y*scale);
           Vp->xyz[2]=(long)(z*scale);
           if(i > 0)CreateEdge(Nvert-1,Nvert-2);
           else     id=Nvert-1;
           if(i == (npoints-1))CreateEdge(Nvert-1,id);
         }
//for(i=0;i<4;i++)fscanf(fp,"%lf %lf %lf",&x,&y,&z);
         if(i_full && j > 0){
           if(!UpdateEdgeHeap(Nedge+2*npoints))goto XXIT;
           if(!UpdateFaceHeap(Nface+2*npoints))goto XXIT;
           for(i=0;i<npoints;i++){
             CreateEdge(Nvert-npoints+i,Nvert-npoints-npoints+i);
             if(i > 0){
               CreateEdge(Nvert-npoints+i,Nvert-npoints-npoints+i-1);
               CreateFace(Nvert-npoints+i,
                          Nvert-npoints-npoints+i-1,
                          Nvert-npoints-npoints+i);
               CreateFace(Nvert-npoints+i,
                          Nvert-npoints+i-1,
                          Nvert-npoints-npoints+i-1);
             }
             else{
               CreateEdge(Nvert-npoints,Nvert-npoints-1);
               CreateFace(Nvert-npoints,
                          Nvert-npoints-1,
                          Nvert-npoints-npoints);
               CreateFace(Nvert-npoints,
                          Nvert-1,
                          Nvert-npoints-1);
             }
           }
         }
       }
       XXIT:;
     }
     fclose(fp);
   }
   else MessageBox(NULL,"No FILE",NULL,MB_OK);
   SetCursor(hSave);
 }
 return TRUE;
}

static unsigned CALLBACK NameHookProc(HWND hwnd,UINT msg,
                              WPARAM wparam,LPARAM lparam){
 char str[32];
 BOOL err;
 LPOFNOTIFY lp;
 switch( msg ) {
   case WM_INITDIALOG:
     if(i_full)SendDlgItemMessage(hwnd,DLG_SECTIONS_FACET,BM_SETCHECK,1,0);
     sprintf(str,"%.2lf",scale_v);
     SetDlgItemText(hwnd,DLG_SECTIONS_SCALE,str);
     return FALSE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDOK:{
           lp=NULL;
           goto JUMPHERE;
         }
         break;
       default: break;
     }
     break;
   case WM_NOTIFY:{
       lp=(LPOFNOTIFY)lparam;
       if(lp->hdr.code == CDN_FILEOK){
         JUMPHERE:
         if(SendDlgItemMessage(hwnd,DLG_SECTIONS_FACET,BM_GETCHECK,0,0))
         i_full=TRUE; else i_full=FALSE;
         GetDlgItemText(hwnd,DLG_SECTIONS_SCALE,str,31);
         scale_v=atof(str);
         if(scale_v < 0.00001)scale_v=0.00001;
       }
     }
     break;
   default: break;
 }
 return FALSE;
}

static BOOL LocalSelectFileName(char *szfile, char *szfilter,
                           HWND parent, HINSTANCE hInst){
 int i;
 OPENFILENAME ofn;
 char szFilter[128],title[256];
 DWORD version;
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=0;
 ofn.lpstrFile=szfile;
 ofn.nMaxFile=255;
 ofn.lpstrFileTitle=NULL;
 ofn.nMaxFileTitle=0;
 ofn.lpstrInitialDir=NULL;
 LoadString(hThisInstance,IDX_STRING2,title,255);
 ofn.lpstrTitle=title;
 ofn.hInstance=hInst;
 ofn.lpfnHook=NameHookProc;
 version=GetVersion();
 ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|
           OFN_ENABLEHOOK|OFN_ENABLETEMPLATE|OFN_NONETWORKBUTTON;
// if(LOBYTE(LOWORD(version)) >= 4){
   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_SECTIONS);
   ofn.Flags |= OFN_EXPLORER;
// }
// else{
//   MessageBeep(MB_OK);
//   return FALSE;
//   ofn.lpTemplateName=MAKEINTRESOURCE(DLG_SECTIONS);
// }
 return GetOpenFileName(&ofn);
}
