/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file SETUP.C */

// This file contains code called as the application starts to build
// pens brushes cursors etc. etc. etc. 

#define MODULE_SETUP 1

#define SIZESTATUS1 195 // 130  //This must also be defined in DESIGN.C

#include "design.h"
#include "..\animate\callall.h"

static void UnloadRenderer(void);

static long palette256[16]={
  0x00c0c0c0,  /* screen     brush */
  0x00808080,  /* window     brush */
  0x0000ffff,  /* selected   brush */
  0x000000ff,  /* deselected brush */
  0x00ffffff,  /* invert     brush */
  0x0000ffff,  /* selected   pen   */
  0x000000ff,  /* deselected pen   */
  0x0000ff00,  /* edit       pen   */
  0x00f0f0f0,  /* wireframe  pen   */
  0x00808080,  /* cursor     pen   */
  0x0000ff00,  /* invert     pen   */
  0x00000000,  /* edge       pen   */
  0x0000ffff,  /* selector   pen   */
  0x000000ff,  /* deselector pen   */
  0x00767676,  /* grid1 118  pen   */
  0x00979797}; /* grid2 151  pen   */

static long palette24b[16]={
  0x00c0c0c0,  /* screen     brush */
  0x00606060,  /* window     brush */
  0x00ff0000,  /* selected   brush */
  0x00ffff00,  /* deselected brush */
  0x00ffffff,  /* invert     brush */
  0x0000ffff,  /* selected   pen   */
  0x000000ff,  /* deselected pen   */
  0x0000ff00,  /* edit       pen   */
  0x00f0f0f0,  /* wireframe  pen   */
  0x00ffff00,  /* cursor     pen   */
  0x00ffff00,  /* invert     pen   */
  0x00000000,  /* edge       pen   */
  0x0000ffff,  /* selector   pen   */
  0x000000ff,  /* deselector pen   */
  0x00767676,  /* grid1 118  pen   */
  0x00979797}; /* grid2 151  pen   */

#define PALVERSION 0x300

#define X__C(p,i) RGB(p->palPalEntry[i].peRed,p->palPalEntry[i].peGreen,p->palPalEntry[i].peBlue)

HPALETTE CreateAppPalette(int *ScreenX, int *ScreenY){
 char *cstr[16]={
                "screen",
                "window",
                "selected",
                "deselected",
                "invert",
                "selected",
                "deselected",
                "edit",
                "wireframe",
                "cursor",
                "invert",
                "edge",
                "selector",
                "deselector",
                "grid1",
                "grid2"
                };
 char strng[64];
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE red,green,blue;
 HWND hDesktopWnd;
 HDC hDCcaps;
 int  i,Dbits;
 long c,*palette;
 nNumColors = 256;
 pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) +
                     nNumColors * sizeof(PALETTEENTRY));
 if(pPal == NULL)return NULL;
 pPal->palNumEntries = nNumColors;
 pPal->palVersion    = PALVERSION;
 red = green = blue = 0;
 for(i=0;i<256;i++){
   pPal->palPalEntry[i].peRed   = red;
   pPal->palPalEntry[i].peGreen = green;
   pPal->palPalEntry[i].peBlue  = blue;
   pPal->palPalEntry[i].peFlags = (BYTE)0;
   if (!(red += 32))if (!(green += 32)) blue += 64;
 }
 hDesktopWnd = GetDesktopWindow();
 hDCcaps     = GetDC(hDesktopWnd);
 Dbits = GetDeviceCaps(hDCcaps,BITSPIXEL);
 *ScreenX = GetDeviceCaps(hDCcaps,HORZRES);
 *ScreenY = GetDeviceCaps(hDCcaps,VERTRES);
 ReleaseDC(hDesktopWnd,hDCcaps);
 if(Dbits <= 8)palette=palette256;
 else          palette=palette24b;
 for(i = 0; i < 16; i++){
   GetPrivateProfileString("MCOLOURS",cstr[i],"$$$",
                   strng,sizeof(strng),IniFilename);
   if(strncmp(strng,"$$$",3) == 0){
     pPal->palPalEntry[i].peRed   = (unsigned char)( palette[i]        & 0xff);
     pPal->palPalEntry[i].peGreen = (unsigned char)((palette[i] >> 8)  & 0xff);
     pPal->palPalEntry[i].peBlue  = (unsigned char)((palette[i] >> 16) & 0xff);
     pPal->palPalEntry[i].peFlags = (BYTE)0;
   }
   else{
     int r,g,b;
     sscanf(strng,"%ld %ld %ld",&r,&g,&b);
     pPal->palPalEntry[i].peRed   = (unsigned char)(r&0xff);
     pPal->palPalEntry[i].peGreen = (unsigned char)(g&0xff);
     pPal->palPalEntry[i].peBlue  = (unsigned char)(b&0xff);
     pPal->palPalEntry[i].peFlags = (BYTE)0;
   }
 }
 hpal = CreatePalette(pPal);
 ghbrushScreen     = CreateSolidBrush(PALETTEINDEX(0));
 ghbrushWindow     = CreateSolidBrush(PALETTEINDEX(1));
 ghSelectedBrush   = CreateSolidBrush(PALETTEINDEX(2));
 ghDeselectedBrush = CreateSolidBrush(PALETTEINDEX(3));
 ghInvertBrush     = CreateSolidBrush(PALETTEINDEX(4));
 ghSelectedPen     = CreatePen(PS_SOLID,0,X__C(pPal, 5));
 ghDotSelectedPen  = CreatePen(PS_DOT,0,X__C(pPal, /*5*/ 11)); // Edge Pen
 ghDeselectedPen   = CreatePen(PS_SOLID,0,X__C(pPal, 6));
 ghEditPen         = CreatePen(PS_SOLID,0,X__C(pPal, 7));
 ghWireframePen    = CreatePen(PS_SOLID,0,X__C(pPal, 8));
 ghCursorPen       = CreatePen(PS_SOLID,0,X__C(pPal, 9));
 ghInvertPen       = CreatePen(PS_SOLID,0,X__C(pPal,10));
 ghEdgePen         = CreatePen(PS_SOLID,0,X__C(pPal,11));
 ghSelectorPen     = CreatePen(PS_SOLID,0,X__C(pPal,12));
 ghDeselectorPen   = CreatePen(PS_SOLID,0,X__C(pPal,13));
 gScreenColourRef=(COLORREF)palette[1];
 gGridColourRef1=X__C(pPal,14);
 gGridColourRef2=X__C(pPal,15);
 if(ghbrushScreen     == NULL ||
    ghbrushWindow     == NULL ||
    ghSelectedBrush   == NULL ||
    ghDeselectedBrush == NULL ||
    ghInvertBrush     == NULL ||
    ghSelectedPen     == NULL ||
    ghDotSelectedPen  == NULL ||
    ghDeselectedPen   == NULL ||
    ghEditPen         == NULL ||
    ghWireframePen    == NULL ||
    ghCursorPen       == NULL ||
    ghInvertPen       == NULL ||
    ghEdgePen         == NULL ||
    ghSelectorPen     == NULL ||
    ghDeselectorPen   == NULL)SendPrgmQuery(IDQ_BADPENS,3);
 LocalFree((HANDLE)pPal);
 return hpal;
}

void LoadPreferences(void){
 char pref_id[32];
 Preferences.buttons=TRUE;
 Preferences.quick_trick=TRUE;
 Preferences.show_select=FALSE;
 Preferences.undo=TRUE;
 Preferences.use_3d_dialogs=TRUE;
 Preferences.use_OpenGL=TRUE;
 Preferences.Vcolour[0]=0;
 Preferences.Vcolour[1]=0;
 Preferences.Vcolour[2]=0;
 Preferences.tooltips=TRUE;
 Preferences.attribs_sheet=TRUE;
 Preferences.bigv=FALSE;
 Preferences.use_positions=TRUE;
 Preferences.toolsmall=FALSE;
 Preferences.toolleft=FALSE;
 Preferences.background=TRUE;
 select_visible=0;
 ActiveView=TRITOP;
 GetPrivateProfileString(IniSection,"BUTTONS","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",3) == 0){
   Preferences.buttons=FALSE;
 }
 GetPrivateProfileString(IniSection,"SELECTPANNEL","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0){
   Preferences.show_select=TRUE;
   select_visible=1;
 }
 GetPrivateProfileString(IniSection,"QUICKTRICK","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",3) == 0){
   Preferences.quick_trick=FALSE;
 }
 GetPrivateProfileString(IniSection,"UNDO","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",3) == 0){
   Preferences.undo=FALSE;
   UNDO_ON=0;
 }
 GetPrivateProfileString(IniSection,"DIALOGS3D","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",3) == 0){
   Preferences.use_3d_dialogs=FALSE;
 }
 GetPrivateProfileString(IniSection,"USEOPENGL","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",3) == 0){
   Preferences.use_OpenGL=FALSE;
 }
 GetPrivateProfileString(IniSection,"VIEWCOLOUR","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"$$$",3) != 0){
   long i1,i2,i3;
   sscanf(pref_id,"%ld %ld %ld",&i1,&i2,&i3);
   Preferences.Vcolour[0]=(unsigned char)i1;
   Preferences.Vcolour[1]=(unsigned char)i2;
   Preferences.Vcolour[2]=(unsigned char)i3;
 }
 GetPrivateProfileString(IniSection,"TOOLTIPS","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",2) == 0){
   Preferences.tooltips=FALSE;
 }
 GetPrivateProfileString(IniSection,"DLG95","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",2) == 0){
   Preferences.attribs_sheet=FALSE;
 }
 GetPrivateProfileString(IniSection,"BIGV","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0){
   Preferences.bigv=TRUE;
 }
 GetPrivateProfileString(IniSection,"TOOLLEFT","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)Preferences.toolleft=TRUE;
 if(strncmp(pref_id,"NO",2) == 0)Preferences.toolleft=FALSE;
 GetPrivateProfileString(IniSection,"TOOLSMALL","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)Preferences.toolsmall=TRUE;
 if(strncmp(pref_id,"NO",2) == 0)Preferences.toolsmall=FALSE;
 GetPrivateProfileString(IniSection,"BACKIMAGE","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)Preferences.background=TRUE;
 if(strncmp(pref_id,"NO",2) == 0)Preferences.background=FALSE;
 ActiveView=GetPrivateProfileInt(IniSection,"ACTIVEVIEW",TRITOP,
                                 IniFilename);
 GetPrivateProfileString(IniSection,"POSITIONS","$$$",
                   pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"NO",2) == 0){
   Preferences.use_positions=FALSE;
 }
 bBackgroundImage=Preferences.background;
 bToolpannelLeft=Preferences.toolleft;
 bToolpannelSmall=Preferences.toolsmall;
 GetPrivateProfileString(IniSection,"QUICKDRAW","$$$",
                          pref_id,sizeof(pref_id),IniFilename);
 if(strncmp(pref_id,"YES",3) == 0)global_quickdraw=1;
 if(strncmp(pref_id,"NO",2) == 0)global_quickdraw=0;
 DefaultBrilliance=GetPrivateProfileInt(IniSection,
  "DefaultBrilliance",DefaultBrilliance,IniFilename);
 DefaultParameter=GetPrivateProfileInt(IniSection,
  "DefaultParameter",DefaultParameter,IniFilename);
 DefaultReflectivity=GetPrivateProfileInt(IniSection,
  "DefaultReflectivity",DefaultReflectivity,IniFilename);
 DefaultTransp=GetPrivateProfileInt(IniSection,
  "DefaultReflectivity",DefaultReflectivity,IniFilename);
 DefaultTexture=GetPrivateProfileInt(IniSection,
  "DefaultTexture",DefaultTexture,IniFilename);
 DefaultShiny=GetPrivateProfileInt(IniSection,
  "DefaultShiny",DefaultShiny,IniFilename);
 DefaultSmoothing=GetPrivateProfileInt(IniSection,
  "DefaultSmoothing",DefaultSmoothing,IniFilename);
 FaceColour[0]=GetPrivateProfileInt(IniSection,
  "FaceColour0",FaceColour[0],IniFilename);
 FaceColour[1]=GetPrivateProfileInt(IniSection,
  "FaceColour1",FaceColour[1],IniFilename);
 FaceColour[2]=GetPrivateProfileInt(IniSection,
  "FaceColour2",FaceColour[2],IniFilename);
 ShaderColour[0]=GetPrivateProfileInt(IniSection,
  "ShaderColour0",ShaderColour[0],IniFilename);
 ShaderColour[1]=GetPrivateProfileInt(IniSection,
  "ShaderColour1",ShaderColour[1],IniFilename);
 ShaderColour[2]=GetPrivateProfileInt(IniSection,
  "ShaderColour2",ShaderColour[2],IniFilename);
 LastColour[0]=FaceColour[0];
 LastColour[1]=FaceColour[1];
 LastColour[2]=FaceColour[2];
 coords_visible=GetPrivateProfileInt(IniSection,"COORDSVISIBLE",
                                     coords_visible,IniFilename);
 draw_grid_on=GetPrivateProfileInt(IniSection,"DRAWGRIDON",
                                     draw_grid_on,IniFilename);
 grid_on=GetPrivateProfileInt(IniSection,"GRIDON",
                                     grid_on,IniFilename);
 return;
}

void ActivatePreferences(void){
 HMENU hmenu;
 hmenu=GetMenu(ghwnd_main);
 if(global_quickdraw)CheckMenuItem(hmenu,
                      IDM_VIEW_DRAWING_HIDEEDGES,MF_CHECKED);
 else                CheckMenuItem(hmenu,
                      IDM_VIEW_DRAWING_HIDEEDGES,MF_UNCHECKED);
 if(coords_visible){
   if(IsWindow(ghwndCoord1))ShowWindow(ghwndCoord1,SW_SHOWNA);
   CheckMenuItem(hmenu,IDM_WINDOW_COORDS,MF_CHECKED);
 }
 if(draw_grid_on)CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_DRAW_GRID,MF_CHECKED);
 if(grid_on)CheckMenuItem(GetMenu(ghwnd_main),IDM_VIEW_GRID,MF_CHECKED);
 if(GetPrivateProfileInt(IniSection,"VIEW",View,IniFilename)
     == ONEVIEW)PostMessage(ghwnd_main,WM_COMMAND,
                           (WPARAM)IDM_WINDOW_SWITCHWINDOWS,(LPARAM)0);
}

static char WORKDRIVE[32];

BOOL DesignStartup(UINT *screen_r){
 int id,i;
 HMENU hmenu,hmenu1;
 HANDLE h;
 FILE *ff;
 WIN32_FIND_DATA ffd,*lpffd;
 char str[255],title[64];
 InitialiseStringVectors();
 ghcurWait    = LoadCursor(NULL,IDC_WAIT);
 ghcurArrow   = LoadCursor(NULL,IDC_ARROW);
 ghcurArrowD  = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_OPENARROW));
 ghcurPan     = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_PAN));
 ghcurZout    = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_ZOOMOUT));
 ghcurZin     = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_ZOOMIN));
 ghcurZoom    = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_ZOOM));
 ghcurCross   = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_NONE));
 ghcurSelect  = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SELECTOR));
 ghcurDeselect= LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_DESELECTOR));
 ghcurPlotter = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_PLOTTER));
 ghcurPlot3d  = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_PLOT3D));
 ghcurGrabber = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_GRABBER));
 ghcurBuilder = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_BUILDER));
 ghcurExpander= LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_EXPANDER));
 ghcurRotator = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_ROTATOR));
 ghcurPrims   = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_PRIMS));
 ghcurShaper  = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SHAPER));
 ghcurMagnetTo= LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_MAGNETTO));
 ghcurMagnetFr= LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_MAGNETFROM));
 ghcurMagnetTw= LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_MAGNETTWIST));
 ghcurMagnetCu= LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_MAGNETCURL));
 ghcurSel_a   = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SEL_A));
 ghcurSel_s   = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SEL_S));
 ghcurSel_l_a = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SEL_L_A));
 ghcurSel_l_s = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SEL_L_S));
 ghcurSel_c_a = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SEL_C_A));
 ghcurSel_c_s = LoadCursor(ghinst_main,MAKEINTRESOURCE(IDC_SEL_C_S));
 ghcurTools[ 0]= ghcurArrow;
 ghcurTools[ 1]= ghcurArrowD;
 ghcurTools[ 2]= ghcurPan;
 ghcurTools[ 3]= ghcurZoom;
 ghcurTools[ 4]= ghcurGrabber;
 ghcurTools[ 5]= ghcurRotator;
 ghcurTools[ 6]= ghcurExpander;
 ghcurTools[ 7]= ghcurPlotter;
 ghcurTools[ 8]= ghcurShaper;
 ghcurTools[ 9]= ghcurMagnetTo;
 ghcurTools[10]= ghcurBuilder;
 ghcurTools[11]= ghcurPlot3d;
 if(FindWindow("OFX:AnimatorClass",NULL) == NULL)StartedInModeler=TRUE;
 else StartedInModeler=FALSE;
 GetModuleFileName(ghinst_main,gszHomeDir,MAX_DIR);
 *FileInPath(gszHomeDir) = '\0';
 gszSCEfile[0]='\0';strcpy(gszSCEdir,gszHomeDir);strcat(gszSCEdir,"models");
 gszSCFfile[0]='\0';strcpy(gszSCFdir,gszHomeDir);strcat(gszSCFdir,"models");
 gszSCNfile[0]='\0';strcpy(gszSCNdir,gszHomeDir);strcat(gszSCNdir,"models");
 gszPTHfile[0]='\0';strcpy(gszPTHdir,gszHomeDir);strcat(gszPTHdir,"models");
 gszIMGfile[0]='\0';strcpy(gszIMGdir,gszHomeDir);strcat(gszIMGdir,"maps");
 gszGIFfile[0]='\0';strcpy(gszGIFdir,gszHomeDir);strcat(gszGIFdir,"maps");
 gszDLLfile[0]='\0';strcpy(gszDLLdir,gszHomeDir);strcat(gszDLLdir,"textures");
 gszDLMfile[0]='\0';strcpy(gszDLMdir,gszHomeDir);strcat(gszDLMdir,"plugins");
 gszDXFfile[0]='\0';strcpy(gszDXFdir,gszHomeDir);strcat(gszDXFdir,"models");
 gsz3DSfile[0]='\0';strcpy(gsz3DSdir,gszHomeDir);strcat(gsz3DSdir,"models");
 gszTDDfile[0]='\0';strcpy(gszTDDdir,gszHomeDir);strcat(gszTDDdir,"models");
 gszFLXfile[0]='\0';strcpy(gszFLXdir,gszHomeDir);strcat(gszFLXdir,"movies");
 WorkDrive=WORKDRIVE;
 if(GetEnvironmentVariable((LPTSTR)"EVT",(LPTSTR)WorkDrive,32) == 0){
   strcpy(WorkDrive,gszHomeDir);
 }
// fprintf(debug,"[%s]\n",WorkDrive);
 SetUpWindowBoxView(0);
 nImaps=0; nLockediMaps=0;
 strcpy(iSha[0].N,"Base Axis"); nSha=1;
 iSha[0].P[0]=0;          iSha[0].P[1]=0;         iSha[0].P[2]=TVpointZ;
 iSha[0].X[0]=TVsizeX/2;  iSha[0].X[1]=0;         iSha[0].X[2]=TVpointZ;
 iSha[0].Y[0]=0;          iSha[0].Y[1]=TVsizeZ/2; iSha[0].Y[2]=TVpointZ;
 for(id=0;id<8;id++){xShaders[id].in_use=0; xShaders[id].parameters=NULL; }
 tool=NOTOOL; sktool=NO;

 sprintf(str,"%s\\*.afx",gszDLMdir);
 lpffd=&ffd;
 nActionsList=0;
 if((h=FindFirstFile(str,lpffd)) != INVALID_HANDLE_VALUE){
   sprintf(str,"%s\\%s",gszDLMdir,lpffd->cFileName);
   ActionsList[nActionsList]=(char *)X__Malloc(256*sizeof(char));
   if(ActionsList[nActionsList] != NULL)strcpy(ActionsList[nActionsList],str);
   nActionsList++;
   while(FindNextFile(h,lpffd)){
     sprintf(str,"%s\\%s",gszDLMdir,lpffd->cFileName);
     ActionsList[nActionsList]=(char *)X__Malloc(256*sizeof(char));
     strcpy(ActionsList[nActionsList],str);
     nActionsList++;
   }
   if(GetLastError() != ERROR_NO_MORE_FILES){
     SendPrgmQuery(IDQ_BADNEXTFILE,3);
   }
   if(!FindClose(h))SendPrgmQuery(IDQ_FILEERROR,3);
   hmenu=GetSubMenu(GetMenu(ghwnd_main),4);
   AppendMenu(hmenu,MF_SEPARATOR,0,0);

   hmenu1=CreatePopupMenu();
   LoadString(ghinst_main,IDX_MISC_PLUGINACTIONS,title,64);
   AppendMenu(hmenu,MF_POPUP,(UINT)hmenu1,title);

   if(nActionsList > 0)for(i=0;i<nActionsList;i++){
     if((ff=fopen(ActionsList[i],"r")) != NULL){
       fscanf(ff,"%[^\n]",title);
       AppendMenu(hmenu1,MF_ENABLED | MF_STRING,i+30000,title);
       fclose(ff);
     }
   }
 }
 InitializeLastFilesList();
 SendPrgmMessage(IDS_DEFAULT,0);
 EDIT_ACTION=NO;
 DeleteMemoryWindows();
 if(!CreateMemoryWindows()){
   MessageBox (NULL,"Fatal ERROR", NULL, MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
   return FALSE;
 }
 EraseAll(1);
 SetupStructures();
 *screen_r=1;
 update_timer=0;
 DrawModel();
 if(ghwndOpenGLview == NULL)Draw3dView(0,0);
 else PostMessage(ghwndOpenGLview,(WM_USER+2),0,0);
 return TRUE;
}

BOOL SetupMainWindow(HANDLE inst, int cmdshow){
 long i;
 RECT rc;
 HDC hdc;
 TEXTMETRIC tm;
 int StatusPartsList[2]={0,-1};
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 View=TRIVIEW;
 ghwndTools=NULL;
 ghwndCoord1=NULL; ghwndData=NULL;
 ghwnd_current=NULL; ghwndDlgAttribs=NULL;
 ghwndLock=NULL; ghwndDlgTools=NULL;
 ghwndDlgExpand=NULL; ghwndTooltip=NULL;
 ghwndSkEdit=NULL; ghwndSkToolbar=NULL;
 ghwndDlgPrimitives=NULL;
 ghwnd_main = CreateWindow(
     szMainClass,     /* class */
     "  ",            /* caption */
     WS_OVERLAPPED |  WS_CAPTION | WS_SYSMENU     | WS_THICKFRAME |
     WS_MAXIMIZE   | !WS_VISIBLE | WS_MAXIMIZEBOX | WS_MINIMIZEBOX,
     0,
     0,
     Xres,                   /* init. x size */
     Yres,                   /* init. y size */
     NULL,                  /* parent window */
     NULL,                  /* menu handle */
     inst,                  /* program handle */
     NULL                   /* create parms */
     );
 if( !ghwnd_main ) return( FALSE );
 ghwnd_info=CreateWindowEx(0L,STATUSCLASSNAME,"",
    WS_CHILD | !WS_BORDER | WS_VISIBLE,
    0,0,0,0,
    ghwnd_main,
    (HMENU)NULL,
    inst,
    NULL);
 if(!ghwnd_info)return(FALSE);
 StatusPartsList[0]=Xres-SIZESTATUS1;
 SendMessage(ghwnd_info,SB_SETPARTS,(WPARAM)2,(LPARAM)StatusPartsList);
 GetClientRect(ghwnd_info,&rc);
 InfoWindowSize=rc.bottom-rc.top;
 LoadMenuHelpTable(ghwnd_main);
 GetClientRect(ghwnd_main,&rc);
 SetWindowSize(rc.right,rc.bottom,View,TRUE);
 GetTriview(TRUE);
 for(i=0;i<3;i++){
   ghwnd_triview[i] = CreateWindow(
     szTriViewClass,
     NULL,
     WS_CHILD | WS_BORDER | WS_VISIBLE,
     WindowsXmin[i],WindowsYmin[i],WindowSizeX[i],WindowSizeY[i],
     ghwnd_main,
     (HMENU)i,
     inst,
     NULL
   );
   if(!ghwnd_triview[i])return(FALSE);
 }
 ghwnd_view = CreateWindow(
     szViewClass,
     NULL,
     WS_CHILD | WS_BORDER | WS_VISIBLE |
     WS_HSCROLL | WS_VSCROLL,
     WindowViewXmin,WindowViewYmin,WindowViewXsize,WindowViewYsize,
     ghwnd_main,
     NULL,
     inst,
     NULL
 );
 if(!ghwnd_view)return(FALSE);
 cmdshow=SW_SHOWMAXIMIZED;
 SetScrollRange(ghwnd_view,SB_HORZ,0,359,FALSE);
 SetScrollRange(ghwnd_view,SB_VERT,0,179,FALSE);
 SetScrollPos(ghwnd_view,SB_HORZ,(int)PerspFi+180,FALSE);
 SetScrollPos(ghwnd_view,SB_VERT,(int)PerspTheta+90,FALSE);
 ghwnd_current=ghwnd_triview[ActiveView];
 ToolboxShow(TRUE);
 if(!ghwndTools){
   MessageBox (NULL,"Can't find icon bitmaps","Error",
               MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
   return (FALSE);
 }
 hdc=GetDC(ghwnd_main);
 GetTextMetrics(hdc,&tm);
 ReleaseDC(ghwnd_main,hdc);
 Textwidth  = tm.tmAveCharWidth;
 Textheight = tm.tmHeight;
 ghFont1=CreateFont(-12,0,0,0,FW_DONTCARE,FALSE,FALSE,FALSE,
   ANSI_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY,
   FIXED_PITCH,"MS Sans Serif");
 ghFont2=CreateFont(-12, 0, 0, 0, FW_LIGHT, FALSE, FALSE, FALSE,
      ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
      PROOF_QUALITY, VARIABLE_PITCH | FF_SWISS,
      "Arial");
 ShowWindow( ghwnd_main, cmdshow);
 return TRUE;
}

static void UnloadRenderer(void){
 HWND hWnd;
 if((hWnd=FindWindow("OFX:RenderClass",NULL)) != NULL){
   SendMessage(hWnd,WM_COMMAND,IDM_CALLBACK_SHUTDOWN,0);
   return;
 }
}

void DesignShutdown(void){
 int i;
 HWND hWndPrev;
 char ss[32];
 DeleteArrowIcons();
 Empty_Undo();
 EraseAll(1);
 DeleteMemoryWindows();
 ToolboxDestroy();
 UnloadRenderer();
 hWndPrev = FindWindow("OFX:AnimatorClass",NULL);
 if (hWndPrev != NULL){
   SetForegroundWindow(hWndPrev);
   BringWindowToTop(hWndPrev);
   ShowWindow (hWndPrev,SW_SHOW);
   if(StartedInModeler)PostMessage(hWndPrev,WM_COMMAND,IDM_CALLBACK_SHUTDOWN,0);
 }
 hWndPrev = FindWindow("OFX:BezierClass",NULL);
 if(hWndPrev != NULL){
   PostMessage(hWndPrev,WM_COMMAND,106,0); /* abort */
 }
 hWndPrev = FindWindow("OFX:ViewClass",NULL);
 if(StartedInModeler && hWndPrev != NULL)
   PostMessage(hWndPrev,WM_COMMAND,106,0);
 hWndPrev = FindWindow("MDI:Viewer",NULL);
 if(StartedInModeler &&hWndPrev != NULL)
   PostMessage(hWndPrev,WM_COMMAND,1001,0);
 if(nActionsList > 0){
   for(i=0;i<nActionsList;i++){
     if(ActionsList[i] != NULL)X__Free(ActionsList[i]);
   }
   nActionsList=0;
 }
 FreeStringVectors();
 FreeMenuHelpData();
// for(i=0;i<NMENUHELPDATA;i++)
//   if(MenuHelpData[i].str != NULL)X__Free(MenuHelpData[i].str);
 DeleteObject(ghbrushScreen);
 DeleteObject(ghbrushWindow);
 DeleteObject(ghSelectedBrush);
 DeleteObject(ghDeselectedBrush);
 DeleteObject(ghInvertBrush);
 DeleteObject(ghSelectedPen);
 DeleteObject(ghDotSelectedPen);
 DeleteObject(ghDeselectedPen);
 DeleteObject(ghEditPen);
 DeleteObject(ghWireframePen);
 DeleteObject(ghCursorPen);
 DeleteObject(ghInvertPen);
 DeleteObject(ghEdgePen);
 DeleteObject(ghSelectorPen);
 DeleteObject(ghDeselectorPen);
 DeleteObject(ghcurArrowD);
 DeleteObject(ghcurCross);
 DeleteObject(ghcurSelect);
 DeleteObject(ghcurDeselect);
 DeleteObject(ghcurPan);
 DeleteObject(ghcurZout);
 DeleteObject(ghcurZin);
 DeleteObject(ghcurZoom);
 DeleteObject(ghcurPlotter);
 DeleteObject(ghcurPlot3d);
 DeleteObject(ghcurGrabber);
 DeleteObject(ghcurBuilder);
 DeleteObject(ghcurExpander);
 DeleteObject(ghcurRotator);
 DeleteObject(ghcurPrims);
 DeleteObject(ghcurShaper);
 DeleteObject(ghcurMagnetTo);
 DeleteObject(ghcurMagnetFr);
 DeleteObject(ghcurMagnetTw);
 DeleteObject(ghcurMagnetCu);
 DeleteObject(ghcurSel_a);
 DeleteObject(ghcurSel_s);
 DeleteObject(ghcurSel_l_a);
 DeleteObject(ghcurSel_l_s);
 DeleteObject(ghcurSel_c_a);
 DeleteObject(ghcurSel_c_s);
 if(ghFont1 != NULL)DeleteObject(ghFont1);
 if(ghFont2 != NULL)DeleteObject(ghFont2);
 sprintf(ss,"%ld",View);
 WritePrivateProfileString(IniSection,"VIEW",ss,IniFilename);
 sprintf(ss,"%ld",ActiveView);
 WritePrivateProfileString(IniSection,"ACTIVEVIEW",ss,IniFilename);
 if(global_quickdraw)
   WritePrivateProfileString(IniSection,"QUICKDRAW","YES",IniFilename);
 else
   WritePrivateProfileString(IniSection,"QUICKDRAW","NO",IniFilename);
 sprintf(ss,"%ld",DefaultBrilliance);
 WritePrivateProfileString(IniSection,"DefaultBrilliance",ss,IniFilename);
 sprintf(ss,"%ld",DefaultParameter);
 WritePrivateProfileString(IniSection,"DefaultParameter",ss,IniFilename);
 sprintf(ss,"%ld",DefaultReflectivity);
 WritePrivateProfileString(IniSection,"DefaultReflectivity",ss,IniFilename);
 sprintf(ss,"%ld",DefaultReflectivity);
 WritePrivateProfileString(IniSection,"DefaultReflectivity",ss,IniFilename);
 sprintf(ss,"%ld",DefaultTexture);
 WritePrivateProfileString(IniSection,"DefaultTexture",ss,IniFilename);
 sprintf(ss,"%ld",DefaultShiny);
 WritePrivateProfileString(IniSection,"DefaultShiny",ss,IniFilename);
 sprintf(ss,"%ld",DefaultSmoothing);
 WritePrivateProfileString(IniSection,"DefaultSmoothing",ss,IniFilename);
 sprintf(ss,"%ld",FaceColour[0]);
 WritePrivateProfileString(IniSection,"FaceColour0",ss,IniFilename);
 sprintf(ss,"%ld",FaceColour[1]);
 WritePrivateProfileString(IniSection,"FaceColour1",ss,IniFilename);
 sprintf(ss,"%ld",FaceColour[2]);
 WritePrivateProfileString(IniSection,"FaceColour2",ss,IniFilename);
 sprintf(ss,"%ld",ShaderColour[0]);
 WritePrivateProfileString(IniSection,"ShaderColour0",ss,IniFilename);
 sprintf(ss,"%ld",ShaderColour[1]);
 WritePrivateProfileString(IniSection,"ShaderColour1",ss,IniFilename);
 sprintf(ss,"%ld",ShaderColour[2]);
 WritePrivateProfileString(IniSection,"ShaderColour2",ss,IniFilename);
 sprintf(ss,"%ld",coords_visible);
 WritePrivateProfileString(IniSection,"COORDSVISIBLE",ss,IniFilename);
 sprintf(ss,"%ld",draw_grid_on);
 WritePrivateProfileString(IniSection,"DRAWGRIDON",ss,IniFilename);
 sprintf(ss,"%ld",grid_on);
 WritePrivateProfileString(IniSection,"GRIDON",ss,IniFilename);
 return;
}

void StartOpenGL(BOOL first_time){
 char oldname[256],newname[256];
 char lname[255],text[128];
 RECT rc;
 HWND hwnd=NULL;
 static HMODULE hLib=NULL;
 FARPROC fpFun;
 X__STRUCTURE *Evs;
 HWND (*fpFun1)(HWND, char *, LPRECT, double *, double *, X__STRUCTURE *);
 GetCurrentDirectory(255,oldname);
 strcpy(newname,gszHomeDir);
 if(SetCurrentDirectory(newname) == FALSE)
   MessageBox(ghwnd_main,newname,NULL,MB_OK);
 {char sysdir[256];
  HANDLE hfbm;
  strcpy(sysdir,"opengl32.dll");
  hfbm=CreateFile(sysdir,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
  if(hfbm == INVALID_HANDLE_VALUE){
    GetSystemDirectory(sysdir,256); strcat(sysdir,"\\opengl32.dll");
    hfbm=CreateFile(sysdir,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
        FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
    if(hfbm == INVALID_HANDLE_VALUE){
      SetCurrentDirectory(oldname);
      return;
    }
    CloseHandle(hfbm);
  }
  else CloseHandle(hfbm);
 }
 EnableMenuItem(GetMenu(ghwnd_main),IDM_WINDOW_SHOWOPENGL,MF_ENABLED);
 if(first_time && !Preferences.use_OpenGL)return;
 if(hLib != NULL){
   if(ghwndOpenGLview != NULL)DestroyWindow(ghwndOpenGLview);
   SetScrollPos(ghwnd_view,SB_HORZ,(int)PerspFi+180,TRUE);
   SetScrollPos(ghwnd_view,SB_VERT,(int)PerspTheta+90,TRUE);
   ShowWindow(ghwnd_view,SW_SHOWNA);
   LoadString(ghinst_main,IDX_MISC_SHOWGL,text,128);
   ModifyMenu(GetMenu(ghwnd_main),IDM_WINDOW_SHOWOPENGL,
         MF_BYCOMMAND|MF_STRING,IDM_WINDOW_SHOWOPENGL,text);
   FreeLibrary(hLib); hLib=NULL;
   ghwndOpenGLview=NULL;
   SetCurrentDirectory(oldname);
   return;
 }
 Evs=&EviStructure;
 rc.left=WindowViewXmin;
 rc.top=WindowViewYmin;
 rc.right=WindowViewXsize;
 rc.bottom=WindowViewYsize;
 strcpy(lname,gszHomeDir); strcat(lname,"gldesign.dll");
 if((hLib=LoadLibrary(lname)) != NULL){
   if((fpFun=GetProcAddress(hLib,"_AttachViewOpenGL")) != NULL){
     ShowWindow(ghwnd_view,SW_HIDE);
     fpFun1 = (void *)fpFun;
     hwnd=(*fpFun1)(ghwnd_main,NULL,&rc,&PerspTheta,&PerspFi,Evs);
   }
   else{
     SendPrgmQuery(IDQ_FAILEDFUNCTION,3);
     FreeLibrary(hLib); hLib=NULL;
   }
 }
 else SendPrgmQuery(IDQ_FAILEDLIBRARY,3);
 ghwndOpenGLview=hwnd;
 if(hwnd != NULL){
   LoadString(ghinst_main,IDX_MISC_HIDEGL,text,128);
   ModifyMenu(GetMenu(ghwnd_main),IDM_WINDOW_SHOWOPENGL,
         MF_BYCOMMAND|MF_STRING,IDM_WINDOW_SHOWOPENGL,text);
 }
 SetCurrentDirectory(oldname);
 return;
}

void ExecuteAnimator(BOOL reload){
 int result;
 char modulename[255];
 HWND hWnd;
 if((hWnd=FindWindow("OFX:AnimatorClass",NULL)) != NULL){
   SetForegroundWindow(hWnd);
   BringWindowToTop(hWnd);
   if(reload)PostMessage(hWnd,WM_COMMAND,IDM_CALLBACK_RELOAD,0);
 }
 else{
   strcpy(modulename,gszHomeDir);strcat(modulename,"openfx.exe");
   result=WinExec(modulename,SW_SHOW);
   if(result < 32){
     if(result == ERROR_FILE_NOT_FOUND ||
        result == ERROR_PATH_NOT_FOUND)SendPrgmQuery(IDQ_NOSTART,3);
   }
   if((hWnd=FindWindow("OFX:AnimatorClass",NULL)) != NULL){
     SetForegroundWindow(hWnd);
     BringWindowToTop(hWnd);
     if(reload)PostMessage(hWnd,WM_COMMAND,IDM_CALLBACK_RELOAD,0);
   }
   else SendPrgmQuery(IDQ_NOSTART,3);
 }
}

#include "..\animate\memory.c"

void SetupStructures(void){
  X__STRUCTURE *lpEVI;
  X__MEMORY_MANAGER *lpEVIMM;
  lpEVI=&EviStructure;
  lpEVI->Nvert = &Nvert;
  lpEVI->NvertSelect = &NvertSelect;
  lpEVI->NvertDeselect = &NvertDeselect;
  lpEVI->NvertGlue= &NvertGlue;
  lpEVI->Nskel = &Nskel;
  lpEVI->Nedge = &Nedge;
  lpEVI->Nface = &Nface;
  lpEVI->Nnurbs = &Nnurbs;
  lpEVI->nImaps = &nImaps;
  lpEVI->nSha = &nSha;
  lpEVI->MainVp = &MainVp;
  lpEVI->MainEp = &MainEp;
  lpEVI->MainFp = &MainFp;
  lpEVI->MainSp = &MainSp;
  lpEVI->MainNp = &MainNp;
  lpEVI->FirstSp = &FirstSp;
  lpEVI->iMap=iMap;
  lpEVI->iSha=iSha;
  lpEVI->xShaders=xShaders;
  lpEVI->NpointerX=&NpointerX;
  lpEVI->NpointerY=&NpointerY;
  lpEVI->NpointerZ=&NpointerZ;
  lpEVI->TVpointX=&TVpointX;
  lpEVI->TVpointY=&TVpointY;
  lpEVI->TVpointZ=&TVpointZ;
  lpEVI->TVsizeX=&TVsizeX;
  lpEVI->TVsizeY=&TVsizeY;
  lpEVI->TVsizeZ=&TVsizeZ;
  lpEVI->ruler=&ruler;
  lpEVI->rulerx=&rulerx; lpEVI->rulery=&rulery; lpEVI->rulerz=&rulerz;
  lpEVI->orulerx=&lrulerx; lpEVI->orulery=&lrulery; lpEVI->orulerz=&lrulerz;
  lpEVI->ActiveView=&ActiveView;
  lpEVI->w_frame=&w_frame;
#if __LITE__
  lpEVI->version=0;
#else
  lpEVI->version=1;
#endif
// 60 dummies
  lpEVI->fpDA=EraseAll;
  lpEVI->fpCV=CreateVertex;
  lpEVI->fpCE=CreateEdge;
  lpEVI->fpCF=CreateFace;
  lpEVI->fpCS=CreateSkeleton;
  lpEVI->fpUVH=UpdateVertexHeap;
  lpEVI->fpUEH=UpdateEdgeHeap;
  lpEVI->fpUFH=UpdateFaceHeap;
  lpEVI->fpTool_UP=UpdateCounters;
  lpEVI->fpTool_RD=DrawModel;
  lpEVI->fpTool_CP=AddCopy;
  lpEVI->fpTool_FC=AutoFacetCurveWithHoles;
  lpEVI->fpTool_AC=CreateAttachedCopy;
  lpEVI->fpChooseColour=SetColour;
  lpEVI->fpChooseFile=SelectFileName;
  lpEVI->fpCentreDlgOnC=CentreDialogOnCursor;
  lpEVI->fpCentreDlgOnS=CentreDialogOnScreen;
  lpEVI->fpCopyFaceProp=CopyFaceProp;
  lpEVI->fpZeroSkeletonBoundingBox=ZeroSkeletonBoundingBox;
  lpEVI->fpGetSkeletonBoundBox=GetSkeletonBoundBox;
  lpEVI->fpEraseVertex=EraseVertex;
  lpEVI->fpEraseVertexMapping=EraseVertexMapping;
  lpEVI->fpEraseSkeleton=EraseSkeleton;
  lpEVI->fpGetClosestVertex=get_closest_vertex;
  lpEVI->fpConnected=connected;
  lpEVI->fpNormalize=O_normalize;
  lpEVI->fpArbitraryRotate=arbitrary_rotate;
  lpEVI->fpOrientateFaceNormals=OrientateFaceNormals;
  lpEVI->fpReverseFaceOrientation=ReverseFaceOrientation;
  lpEVI->fpIsFOSame=IsFOSame;
  lpEVI->fpOrientateSelectedFaces=OrientateSelectedFaces;
  lpEVI->fpAppendFileExtension=AppendFileExtension;
  lpEVI->fpEraseAllNurbs=EraseAllNurbs;
  lpEVI->fpEraseNurbs=EraseNurbs;
  lpEVI->fpCreateNurbs=CreateNurbs;
  lpEVI->fpClearNurbsSkeleton=ClearNurbsSkeleton;
  lpEVI->fpAllocNurbs=AllocNurbs;
  lpEVI->fpCloneNurbs=CloneNurbs;
  lpEVI->fpReleaseNurbs=ReleaseNurbs;
// 52 dummies

  lpEVIMM=&EviMemoryManager;
  lpEVIMM->fpMalloc=X__AllocateRam;
  lpEVIMM->fpFree=X__FreeRam;
  lpEVIMM->lpAni=NULL;
  lpEVIMM->lpMod=NULL;
}

static PSTR apstr_StringTable[CSTRINGS];
static PSTR apstr_QueryTable[QSTRINGS];
static PSTR apstr_IconHelpTable[ISTRINGS];

void SendIconHelpMessage(int idString){
 PSTR pstr,p;
 INT cch;
 idString -= FIRST_IDI_STRING;
 if(apstr_IconHelpTable[idString] != NULL)p=apstr_IconHelpTable[idString];
 else{
   if((pstr = (PSTR)X__Malloc(CCHTEXTMAX)) == NULL)return;
   if((cch = LoadString(ghinst_main, idString+FIRST_IDI_STRING,pstr, CCHTEXTMAX)) == 0){
     X__Free(pstr);
     return;
   }
   p = apstr_IconHelpTable[idString] = pstr = (PSTR)X__Realloc(pstr, cch + 1);
   if(p == NULL)return;
 }
 SendMessage(ghwnd_info,SB_SETTEXT,(WPARAM)0,(LPARAM)p);
}

void SendPrgmText(PSTR p){
 SendMessage(ghwnd_info,SB_SETTEXT,(WPARAM)0,(LPARAM)p);
}

void SendPrgmMessage(int idString, int where){
 PSTR pstr,p;
 INT cch,string_id;
 idString -= FIRST_IDS_STRING;
 if(apstr_StringTable[idString] != NULL)p=apstr_StringTable[idString];
 else{
   if((pstr = (PSTR)X__Malloc(CCHTEXTMAX)) == NULL)return;
   string_id=idString+FIRST_IDS_STRING;
#if __LITE__
   if(idString == IDS_DEFAULT)string_id=IDS_DEFAULT_LITE;
   if(idString == IDS_CURRENT)string_id=IDS_CURRENT_LITE;
#endif
  if((cch = LoadString(ghinst_main,string_id,pstr,CCHTEXTMAX)) == 0){
     X__Free(pstr);
     return;
   }
   p = apstr_StringTable[idString] = pstr = (PSTR)X__Realloc(pstr, cch + 1);
   if(p == NULL)return;
 }
 if(where == 0)SendMessage(ghwnd_main,WM_SETTEXT,0,(LPARAM)p);
 else SendPrgmText(p);
}

int SendPrgmQuery(int idString, int response){
 int a;
 PSTR pstr,p;
 INT cch;
 char title[128];
 idString -= FIRST_IDQ_STRING;
 if(apstr_QueryTable[idString] != NULL)p=apstr_QueryTable[idString];
 else{
   if((pstr = (PSTR)X__Malloc(CCHTEXTMAX)) == NULL)return 0;
   if((cch = LoadString(ghinst_main,idString+FIRST_IDQ_STRING,pstr,CCHTEXTMAX)) == 0){
     X__Free(pstr);
     return 0;
   }
   p = apstr_QueryTable[idString] = pstr = (PSTR)X__Realloc(pstr, cch + 1);
   if(p == NULL)return 0;
 }
#if __LITE__
 LoadString(ghinst_main,IDS_DEFAULT_LITE,title,127);
#else
 LoadString(ghinst_main,IDS_DEFAULT,title,127);
#endif
 if(response > 0)MessageBeep(MB_ICONQUESTION);
 else            MessageBeep(MB_OK);
 if     (response == 3)a=MessageBox (GetFocus(),p,NULL,
                          MB_OK | MB_ICONEXCLAMATION | MB_TASKMODAL);
 else if(response == 2)a=MessageBox (GetFocus(),p,title,
                          MB_YESNOCANCEL | MB_ICONQUESTION | MB_TASKMODAL);
 else if(response == 1)a=MessageBox (GetFocus(),p,title,
                          MB_YESNO | MB_ICONQUESTION | MB_TASKMODAL);
 else                  a=MessageBox (GetFocus(),p,title,
                          MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
 return a;
}

void SetPrgmMessage(int idString, PSTR pstr){
 int l;
 PSTR p;
 idString -= FIRST_IDS_STRING;
 l=strlen(pstr)+1;
 if(apstr_StringTable[idString] == NULL){
   if((p=(PSTR)X__Malloc(l)) == NULL)return;
 }
 else{
   if((p=(PSTR)X__Realloc(apstr_StringTable[idString],l)) == NULL)return;
 }
 strcpy(p,pstr);
 apstr_StringTable[idString]=p;
}

void InitialiseStringVectors(void){
 int i;
 for(i=0;i<CSTRINGS;i++)apstr_StringTable[i]=NULL;
 for(i=0;i<QSTRINGS;i++)apstr_QueryTable[i]=NULL;
 for(i=0;i<ISTRINGS;i++)apstr_IconHelpTable[i]=NULL;
 return;
}

void FreeStringVectors(void){
 int i;
 for(i=0;i<CSTRINGS;i++)
   if(apstr_StringTable[i] != NULL)X__Free(apstr_StringTable[i]);
 for(i=0;i<QSTRINGS;i++)
   if(apstr_QueryTable[i] != NULL)X__Free(apstr_QueryTable[i]);
 for(i=0;i<ISTRINGS;i++)
   if(apstr_IconHelpTable[i] != NULL)X__Free(apstr_IconHelpTable[i]);
 return;
}

