/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

//  sfx2vtk.h

class SfxBound
{
public:
  SfxBound();
  ~SfxBound();
  double GetMaximumSize(void);
};


class SfxReader : public vtkPolySource
{
public:
  SfxReader();
  ~SfxReader();
  char *GetClassName() {return "SfxReader";};
  int GetNumberOfVertices(void);
  int GetNumberOfEdge(void);
  int GetNumberOfFaces(void);
  void SetScale(double);
protected:
  void Execute();
  int  Nv,Nf,Ne;
  double scale;
};

class SfxPointReader : public vtkPolySource
{
public:
  SfxPointReader();
  ~SfxPointReader();
  char *GetClassName() {return "SfxPointReader";};
  void SetScale(double);
protected:
  void Execute();
  int  Nv;
  double scale;
};

class SfxEdgeReader : public vtkPolySource
{
public:
  SfxEdgeReader();
  ~SfxEdgeReader();
  char *GetClassName() {return "SfxEdgeReader";};
  void SetScale(double);
protected:
  void Execute();
  int  Nv;
  int  Ne;
  double scale;
};

class SfxBaseWriter : public vtkObject
{
public:
  SfxBaseWriter();
  virtual void Write();
  void Update();

protected:
  vtkDataSet *Input;
  virtual void WriteData() = 0;
};

class SfxWriter : public SfxBaseWriter
{
public:
  SfxWriter();
  ~SfxWriter();
  char *GetClassName() {return "SfxWriter";};
  void SetInput(vtkPolyData *input);
  void SetInput(vtkPolyData &input) {this->SetInput(&input);};
  vtkPolyData *GetInput() {return (vtkPolyData *)this->Input;};
  int GetNumberOfVertices(void);
  int GetNumberOfEdge(void);
  int GetNumberOfFaces(void);
  void EraseOldStructure(long );
  void OffsetStructure(long , long , long , long );
  void SetScale(double);

protected:
  double scale;
  int Nv,Ne,Nf;
  void WriteData();
  void InsertInVertexList(long , long , long );
  void BuildEdges(long );
};
