/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* SIMPLIFY.H     */


#define DESELECTED 0
#define SELECTED   1
#define INEDITOR   6

typedef struct SELECTEDEDGE {
  long e;
  long ea[2]; /* faces adjacent to this edge (assumes no bifurcation)  */
} selectededge;

typedef struct SELECTEDFACE {
  long f;
  long ea[3]; /* id of edges round face                                */
  long fa[3]; /* id if faces adjacent to "f"  -1 => no adjacent face   */
  long del;
  vector n;
} selectedface;

typedef struct VTYADJ {
  long v;
  long nf,ne;
  long *flist;
  long *elist;
} vtyadj;

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                         }
#define POINT2VECTOR(p,v) v[0]=(double)p[0]; v[1]=(double)p[1]; \
                          v[2]=(double)p[2];
#define VECTOR2POINT(v,p) p[0]=(long)v[0]; p[1]=(long)v[1]; \
                          p[2]=(long)v[2];

#define NO               0
#define YES              1
#define FAIL            -1
#define OK               1
#ifndef PI
#define PI 3.1415926
#endif
#define PI2            PI/2

static BOOL CALLBACK SimplifyDlgProc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam);
static void FreeAdjFaces(vtyadj *vl,long nvl);
static vtyadj *ListAdjFaces(selectedface *wf, long nf,
                            selectededge *we, long ne, long *nvll);
static int Simplify1(vtyadj *vl, long nvl,
                        selectedface *wf, long nf,
                        selectededge *we, long ne,
                        double fan, double ean);
static int DeTesselate(vtyadj *vl, long nvl,
                        selectedface *wf, long nf,
                        selectededge *we, long ne,
                        double angle);
static void SimplifyObject(long nloop, double fan, double ean);

static int O_normalize(vector v){
 double n,nn;
 n= (double)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return FAIL;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE(n,v,v)
 return OK;
}

#define DLG_SIM                     600
#define DLG_SIM_ITER                601
#define DLG_SIM_FACE                602
#define DLG_SIM_EDGE                603
