/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  SIMPLEPLT.C                                                        */
/* Draw straight line curves in Window and build 3D structure          */
/* this is an example external modeler                                 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#include "struct.h"
#include "dstruct.h"

#include "simplplt.h"

#define DESELECTED 0
#define SELECTED   1


static HWND      hParent;
static HINSTANCE hThisInstance;
static char *szViewClass="OFX:DraftClass";
static SYSTEM_INFO sys_info;
static int DisplayWidth,DisplayHeight,DisplayBits,DisplayPlanes,
           xScreen,yScreen;
static DWORD fdwStyle = WS_POPUP | WS_CAPTION | ! WS_SYSMENU |
                        WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_THICKFRAME;
static HCURSOR hcurSave,hcurPen1=NULL;

static int pen=0,stretched=1,tracetool=0;
static long Lv= -1,Fv= -1,Nv=0,Ne=0;
static struct TVLIST {BOOL inc; int x,y; long id;}  *TVlist=NULL;
static struct TELIST {BOOL inc; int v1,v2;}         *TElist=NULL;

static LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,LPARAM lParam);
static BOOL MenuCommand(HWND hWnd, WORD id);
static BOOL Viewer(void);
static void DrawRubberLine(HWND hwnd, int x, int y);
static void DrawRubberLines(HWND hwnd, int x, int y);
static void DrawList(HWND hwnd, HDC hDC);
static void AddTraceVertex(int x,int y);
static int GetVertexAtCursor(int x, int y);
static void MoveVertex(HWND hWnd, int x, int y);
static void TraceSubdivide(long v);
static void BuildFromTrace(HWND hWnd);

static BOOL Viewer(void){
 MSG msg;
 HWND hDesktopWnd,hWnd;
 HDC hDCcaps;
 WNDCLASS wndclass;
 TVlist=NULL; TElist=NULL; Nv=0; Ne=0;
 hDesktopWnd = GetDesktopWindow();
 hDCcaps     = GetDC(hDesktopWnd);
 DisplayWidth= GetDeviceCaps(hDCcaps,HORZRES);
 DisplayHeight= GetDeviceCaps(hDCcaps,VERTRES);
 DisplayBits= GetDeviceCaps(hDCcaps,BITSPIXEL);
 DisplayPlanes= GetDeviceCaps(hDCcaps,PLANES);
 ReleaseDC(hDesktopWnd,hDCcaps);
 GetSystemInfo(&sys_info);
 wndclass.style         = 0;
 wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
 wndclass.cbClsExtra    = 0 ;
 wndclass.cbWndExtra    = 0 ;
 wndclass.hInstance     = hThisInstance ;
 wndclass.hIcon         = NULL;
 wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
 wndclass.hbrBackground = GetStockObject (WHITE_BRUSH) ;
 wndclass.lpszMenuName  = "tracemenu" ;
 wndclass.lpszClassName = szViewClass;
 if (!RegisterClass (&wndclass))  return FALSE ;
 xScreen = GetSystemMetrics (SM_CXSCREEN) ;
 yScreen = GetSystemMetrics (SM_CYSCREEN) ;
 hWnd = CreateWindow(
        szViewClass,
        "2D Drafting Utility",
        fdwStyle,
        25,
        25,
        xScreen-50,
        yScreen-50,
        hParent,
        NULL,
        hThisInstance,
        NULL) ;
 if(hWnd == NULL)return FALSE;
 hcurPen1=LoadCursor(hThisInstance,MAKEINTRESOURCE(IDC_PEN1));
 ShowWindow (hWnd,SW_SHOWNA);
 PostMessage(hWnd,WM_COMMAND,(WPARAM)IDM_DRAWIT,0);
 while(GetMessage (&msg,NULL,0,0)){
   TranslateMessage (&msg);
   DispatchMessage (&msg);
 }
 DestroyWindow(hWnd);
 if(hcurPen1 != NULL)DeleteObject(hcurPen1);
 UnregisterClass(wndclass.lpszClassName,wndclass.hInstance);
 if(TVlist != NULL)X__Free(TVlist); TVlist=NULL; Nv=0;
 if(TElist != NULL)X__Free(TElist); TElist=NULL; Ne=0;
 return TRUE;
}

LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,
                             LPARAM lParam){
 static BOOL bToolCaptured=FALSE;
 static HCURSOR hCursorSave;
 static POINT LastPt;
 PAINTSTRUCT ps;
 HDC hDC;
 int i;
 RECT rc;
 POINT pt;
 switch (iMessage) {
   case WM_DESTROY:
    break;
  case WM_COMMAND:
    return MenuCommand(hWnd,LOWORD(wParam));
    break;
  case WM_RBUTTONUP:
    if(!bToolCaptured)break;
    if(tracetool == 1)DrawRubberLine(hWnd,LOWORD(lParam),HIWORD(lParam));
    if(tracetool == 2){
      DrawRubberLines(hWnd,LOWORD(lParam),HIWORD(lParam));
      MoveVertex(hWnd,LOWORD(lParam),HIWORD(lParam));
    }
    SetCursor(hCursorSave);
    ClipCursor(NULL);
    ReleaseCapture();
    bToolCaptured=FALSE;
    Lv= -1;
    break;
  case WM_LBUTTONDOWN:
    if(tracetool == 0 || bToolCaptured)break;
    else if(tracetool == 2){
      if((Fv=GetVertexAtCursor(LOWORD(lParam),HIWORD(lParam))) < 0)break;
    }
    else if(tracetool == 3){ /* delete */
      if((Fv=GetVertexAtCursor(LOWORD(lParam),HIWORD(lParam))) >= 0){
        if(Ne > 0 && TElist != NULL)for(i=0;i<Ne;i++){
          if(TElist[i].v1 == Fv || TElist[i].v2 == Fv)TElist[i].inc=FALSE;
        }
        TVlist[Fv].inc=FALSE;
        InvalidateRect(hWnd,NULL,FALSE);
        Fv= -1;
      }
      break;
    }
    else if(tracetool == 4){ /* subdivide */
      if((Fv=GetVertexAtCursor(LOWORD(lParam),HIWORD(lParam))) >= 0){
        TraceSubdivide(Fv);
        InvalidateRect(hWnd,NULL,FALSE);
        Fv= -1;
      }
      break;
    }
    bToolCaptured=TRUE;
    SetCapture(hWnd);
    hCursorSave=SetCursor(hcurPen1);
    GetClientRect(hWnd,&rc);
    pt.x=rc.left; pt.y=rc.top;
    ClientToScreen(hWnd,&pt);
    rc.left=pt.x; rc.top=pt.y;
    pt.x=rc.right; pt.y=rc.bottom;
    ClientToScreen(hWnd,&pt);
    rc.right=pt.x; rc.bottom=pt.y;
    ClipCursor(&rc);
    break;
  case WM_MOUSEMOVE:
    if(!bToolCaptured)break;
    if(tracetool == 1){
      DrawRubberLine(hWnd,(int)LastPt.x,(int)LastPt.y);
      LastPt.x=LOWORD(lParam); LastPt.y=HIWORD(lParam);
      DrawRubberLine(hWnd,(int)LastPt.x,(int)LastPt.y);
    }
    if(tracetool == 2){
      DrawRubberLines(hWnd,(int)LastPt.x,(int)LastPt.y);
      LastPt.x=LOWORD(lParam); LastPt.y=HIWORD(lParam);
      DrawRubberLines(hWnd,(int)LastPt.x,(int)LastPt.y);
    }
    break;
  case WM_LBUTTONUP:
    if(!bToolCaptured)break;
    if(tracetool == 1){ /* add */
      LastPt.x=LOWORD(lParam); LastPt.y=HIWORD(lParam);
      if(Lv >= 0)DrawRubberLine(hWnd,(int)LastPt.x,(int)LastPt.y);
      AddTraceVertex((int)LOWORD(lParam),(int)HIWORD(lParam));
      DrawList(hWnd,NULL);
      DrawRubberLine(hWnd,(int)LastPt.x,(int)LastPt.y);
    }
    else if(tracetool == 2){ /* move */
      if(Lv < 0){
        Lv=Fv; Fv=-1;
        LastPt.x=LOWORD(lParam); LastPt.y=HIWORD(lParam);
        DrawRubberLines(hWnd,(int)LastPt.x,(int)LastPt.y);
      }
      else{
        DrawRubberLines(hWnd,LOWORD(lParam),HIWORD(lParam));
        MoveVertex(hWnd,LOWORD(lParam),HIWORD(lParam));
        SetCursor(hCursorSave);
        ClipCursor(NULL);
        ReleaseCapture();
        bToolCaptured=FALSE;
        Lv= -1;
      }
    }
    break;
  case WM_PAINT:
    hDC = BeginPaint(hWnd, &ps);
    GetClientRect(hWnd,&rc);
    PatBlt(hDC,0,0,rc.right,rc.bottom,PATCOPY);
    DrawList(hWnd,hDC);
    EndPaint(hWnd,&ps);
    break ;
  case WM_SIZE:
    InvalidateRect(hWnd,NULL,FALSE);
    return 0;
  default:
      return DefWindowProc (hWnd, iMessage, wParam, lParam) ;
 }
 return 0L ;
}

BOOL MenuCommand (HWND hWnd, WORD id){
 int i;
 switch (id) {
   case IDM_PEN_RED  : pen=0; goto PENID;
   case IDM_PEN_WHITE: pen=1; goto PENID;
   case IDM_PEN_BLACK: pen=2; goto PENID;
   case IDM_PEN_CYAN : pen=3; goto PENID;
     PENID:
     CheckMenuItem(GetMenu(hWnd),IDM_PEN_RED,MF_UNCHECKED);
     CheckMenuItem(GetMenu(hWnd),IDM_PEN_WHITE,MF_UNCHECKED);
     CheckMenuItem(GetMenu(hWnd),IDM_PEN_BLACK,MF_UNCHECKED);
     CheckMenuItem(GetMenu(hWnd),IDM_PEN_CYAN,MF_UNCHECKED);
     if(pen == 0)CheckMenuItem(GetMenu(hWnd),IDM_PEN_RED,MF_CHECKED);
     if(pen == 1)CheckMenuItem(GetMenu(hWnd),IDM_PEN_WHITE,MF_CHECKED);
     if(pen == 2)CheckMenuItem(GetMenu(hWnd),IDM_PEN_BLACK,MF_CHECKED);
     if(pen == 3)CheckMenuItem(GetMenu(hWnd),IDM_PEN_CYAN,MF_CHECKED);
     DrawList(hWnd,NULL);
     break;
   case IDM_ADD:
     Lv= -1;
     EnableMenuItem(GetMenu(hWnd),IDM_ADD,MF_GRAYED);
     EnableMenuItem(GetMenu(hWnd),IDM_MOVE,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_DELETE,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_INSERT,MF_ENABLED);
     DrawMenuBar(hWnd);
     tracetool=1;
     break;
   case IDM_MOVE:
     Lv= -1;
     EnableMenuItem(GetMenu(hWnd),IDM_ADD,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_MOVE,MF_GRAYED);
     EnableMenuItem(GetMenu(hWnd),IDM_DELETE,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_INSERT,MF_ENABLED);
     DrawMenuBar(hWnd);
     tracetool=2;
     break;
   case IDM_DELETE:
     Lv= -1;
     EnableMenuItem(GetMenu(hWnd),IDM_ADD,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_MOVE,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_DELETE,MF_GRAYED);
     EnableMenuItem(GetMenu(hWnd),IDM_INSERT,MF_ENABLED);
     DrawMenuBar(hWnd);
     tracetool=3;
     break;
   case IDM_INSERT:
     Lv= -1;
     EnableMenuItem(GetMenu(hWnd),IDM_ADD,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_MOVE,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_DELETE,MF_ENABLED);
     EnableMenuItem(GetMenu(hWnd),IDM_INSERT,MF_GRAYED);
     DrawMenuBar(hWnd);
     tracetool=4;
     break;
   case IDM_EXIT:
     BuildFromTrace(hWnd);
   case IDM_ABORT:
     PostQuitMessage(0);
     break;
   default:
     break;
 }
 return TRUE;
}

static void DrawRubberLine(HWND hwnd, int x, int y){
 HPEN hOldPen,MyPen;
 int oldROP;
 HDC hDClocal;
 if(Lv < 0)return;
 hDClocal=GetDC(hwnd);
 if(pen == 0)MyPen=CreatePen(PS_SOLID,0,RGB(255,0,0));
 if(pen == 1)MyPen=CreatePen(PS_SOLID,0,RGB(0,0,255));
 if(pen == 2)MyPen=CreatePen(PS_SOLID,0,RGB(0,0,0));
 if(pen == 3)MyPen=CreatePen(PS_SOLID,0,RGB(0,255,0));
 hOldPen=SelectObject(hDClocal,MyPen);
 oldROP=SetROP2(hDClocal,R2_NOT);
 MoveToEx(hDClocal,x,y,NULL);
 LineTo(hDClocal,TVlist[Lv].x,TVlist[Lv].y);
 SetROP2(hDClocal,oldROP);
 SelectObject(hDClocal,hOldPen);
 DeleteObject(MyPen);
 ReleaseDC(hwnd,hDClocal);
}

static void DrawRubberLines(HWND hwnd, int x, int y){
 int i,v1,v2;
 HPEN hOldPen;
 int oldROP;
 HDC hDClocal;
 if(Lv < 0 || Ne == 0)return;
 hDClocal=GetDC(hwnd);
 hOldPen=SelectObject(hDClocal,GetStockObject(WHITE_PEN));
 oldROP=SetROP2(hDClocal,R2_NOT);
 for(i=0;i<Ne;i++){
   if(!TElist[i].inc)continue;
   v1=TElist[i].v1; v2=TElist[i].v2;
   if(TVlist[v1].inc && v1 == Lv){
     MoveToEx(hDClocal,x,y,NULL);
     LineTo(hDClocal,TVlist[v2].x,TVlist[v2].y);
   }
   else if(TVlist[v2].inc && v2 == Lv){
     MoveToEx(hDClocal,x,y,NULL);
     LineTo(hDClocal,TVlist[v1].x,TVlist[v1].y);
   }
 }
 SetROP2(hDClocal,oldROP);
 SelectObject(hDClocal,hOldPen);
 ReleaseDC(hwnd,hDClocal);
}

static void DrawList(HWND hwnd, HDC hDC){
 int i,x,y;
 HPEN hOldPen,MyPen;
 HBRUSH hOldBrush;
 HDC hDClocal;
 if(hDC == NULL){
   hDClocal=GetDC(hwnd);
 }
 else hDClocal=hDC;
 if(pen == 0)MyPen=CreatePen(PS_SOLID,0,RGB(255,0,0));
 if(pen == 1)MyPen=CreatePen(PS_SOLID,0,RGB(0,0,255));
 if(pen == 2)MyPen=CreatePen(PS_SOLID,0,RGB(0,0,0));
 if(pen == 3)MyPen=CreatePen(PS_SOLID,0,RGB(0,255,0));
 hOldPen=SelectObject(hDClocal,MyPen);
 hOldBrush=SelectObject(hDClocal,GetStockObject(BLACK_BRUSH));
 if(Ne > 0 && TElist != NULL){
   for(i=0;i<Ne;i++){
     if(!TElist[i].inc)continue;
     x=TVlist[TElist[i].v1].x; y=TVlist[TElist[i].v1].y;
     MoveToEx(hDClocal,x,y,NULL);
     x=TVlist[TElist[i].v2].x; y=TVlist[TElist[i].v2].y;
     LineTo(hDClocal,x,y);
   }
 }
 if(Nv > 0 && TVlist != NULL){
   for(i=0;i<Nv;i++)if(TVlist[i].inc)Rectangle(hDClocal,
      TVlist[i].x-2,TVlist[i].y-2,TVlist[i].x+2,TVlist[i].y+2);
 }
 SelectObject(hDClocal,hOldPen);
 SelectObject(hDClocal,hOldBrush);
 DeleteObject(MyPen);
 if(hDC == NULL){
   ReleaseDC(hwnd,hDClocal);
 }
}

static void AddTraceVertex(int x,int y){
 int i,id,found=0,flag=0;
 struct TVLIST *v;
 struct TELIST *e;
 if(Nv > 0){
   for(i=0;i<Nv;i++){
     if(TVlist[i].inc && abs(x-TVlist[i].x) < 4 && abs(y-TVlist[i].y) < 4){
       found=1;
       if(Lv >= 0)flag=1;
       id=i;
     }
   }
 }
 if(!found){
   if(TVlist == NULL){
     if((v=(struct TVLIST *)X__Malloc(sizeof(struct TVLIST))) == NULL)return;
   }
   else{
     if((v=(struct TVLIST *)X__Realloc(TVlist,(Nv+1)*sizeof(struct TVLIST))) == NULL)return;
   }
   TVlist=v;
   v[Nv].x=x; v[Nv].y=y; v[Nv].inc=TRUE;
   id=Nv;
   Nv++;
 }
 if(Lv < 0){
   Lv=id;
 }
 else{
   if(TElist == NULL){
     if((e=(struct TELIST *)X__Malloc(sizeof(struct TELIST))) == NULL)return;
   }
   else{
     if((e=(struct TELIST *)X__Realloc(TElist,(Ne+1)*sizeof(struct TELIST))) == NULL)return;
   }
   TElist=e;
   e[Ne].v1=id; e[Ne].v2=Lv; e[Ne].inc=TRUE;
   Lv=id;
   Ne++;
 }
 if(flag)Lv= -1;
}

static void MoveVertex(HWND hWnd,int x, int y){
 if(Lv < 0)return;
 TVlist[Lv].x=x; TVlist[Lv].y=y;
 InvalidateRect(hWnd,NULL,FALSE);
 return;
}

int GetVertexAtCursor(int x, int y){
 int i,id= -1;
 if(Nv > 0)for(i=0;i<Nv;i++){
   if(TVlist[i].inc && abs(x-TVlist[i].x) < 4 && abs(y-TVlist[i].y) < 4)id=i;
 }
 return id;
}

static void TraceSubdivide(long vv){
 struct TVLIST *v;
 struct TELIST *e;
 int ne1,i,x,y;
 if(Nv == 0 || Ne == 0 || TVlist == NULL || TElist == NULL)return;
 ne1=Ne; for(i=0;i<ne1;i++){
   if(!TElist[i].inc)continue;
   x=(TVlist[TElist[i].v1].x+TVlist[TElist[i].v2].x)/2;
   y=(TVlist[TElist[i].v1].y+TVlist[TElist[i].v2].y)/2;
   if(TElist[i].v1 == vv){
     if((v=(struct TVLIST *)X__Realloc(TVlist,(Nv+1)*sizeof(struct TVLIST))) == NULL)return;
     TVlist=v; v[Nv].x=x; v[Nv].y=y; v[Nv].inc=TRUE;
     TElist[i].v1=Nv;
     if((e=(struct TELIST *)X__Realloc(TElist,(Ne+1)*sizeof(struct TELIST))) == NULL)return;
     TElist=e; e[Ne].v1=Nv; e[Ne].v2=vv; e[Ne].inc=TRUE;
     Ne++;
     Nv++;
   }
   else if(TElist[i].v2 == vv){
     if((v=(struct TVLIST *)X__Realloc(TVlist,(Nv+1)*sizeof(struct TVLIST))) == NULL)return;
     TVlist=v; v[Nv].x=x; v[Nv].y=y; v[Nv].inc=TRUE;
     TElist[i].v2=Nv;
     if((e=(struct TELIST *)X__Realloc(TElist,(Ne+1)*sizeof(struct TELIST))) == NULL)return;
     TElist=e; e[Ne].v1=Nv; e[Ne].v2=vv; e[Ne].inc=TRUE;
     Ne++;
     Nv++;
   }
 }
}

static void BuildFromTrace(HWND hWnd){
 RECT rc;
 long Scale,v1,v2;
 int i,vc=0,ec=0;
 GetClientRect(hWnd,&rc);
 if(Nv == 0 || TVlist == NULL)return;
 Scale=100;
 for(i=0;i<Nv;i++){
   if(TVlist[i].inc)vc++;
   TVlist[i].id = -1;
 }
 if(vc == 0)return;
 UpdateVertexHeap(Nvert+vc);
 for(i=0;i<Nv;i++){
   if(!TVlist[i].inc)continue;
   CreateVertex();
   (MainVp+Nvert-1)->xyz[0]=TVlist[i].x*Scale;
   (MainVp+Nvert-1)->xyz[1]=NpointerY;
   (MainVp+Nvert-1)->xyz[2]=(rc.bottom-TVlist[i].y)*Scale;
   TVlist[i].id=Nvert-1; /* use to point to actual vertex */
 }
 if(Ne == 0)return;
 for(i=0;i<Ne;i++)if(TVlist[i].inc)ec++;
 if(ec == 0)return;
 UpdateEdgeHeap(Nedge+vc);
 for(i=0;i<Ne;i++){
   if(!TElist[i].inc)continue;
   v1=TVlist[TElist[i].v1].id; v2=TVlist[TElist[i].v2].id;
   if(v1 >= 0 && v2 >= 0)CreateEdge(v1,v2);
 }
}

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=hDLL;
      if(hDLL == NULL)MessageBox ( GetFocus()," NULL instance",NULL, MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 lpEVI=lpevi;
 tracetool=0;
 hParent=parent_window;
 return Viewer();
}
