/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* FILE struct.h pointers to main data structures and elements */

typedef double vector[3];
typedef long   point[3];
typedef long   twoLongs[2];

typedef struct tagVECTOR4 {
 double x,y,z,w;
 BOOL   selected;
} vector4;

typedef struct tagX__STRUCTURE {
  /* NOTE: Don't use functions that return "double" use pointer "double *" */
  /*       it is OK if same compiler as DLL otherwise bad                  */
  unsigned long *Nvert,*Nedge,*Nface,*Nskel,
                *NvertSelect,*NvertDeselect,
                *NvertGlue;
  int    *nImaps,*nSha;
  struct VERTEX   **MainVp;
  struct EDGE     **MainEp;
  struct FACE     **MainFp;
  struct SKELETON **MainSp,**FirstSp;
  struct tagIMAGEMAP     *iMap;
  struct tagSHADERAXIS   *iSha;
  struct tagXSHADER      *xShaders;
  long   *NpointerX,*NpointerY,*NpointerZ,
         *TVpointX, *TVpointY, *TVpointZ,
         *TVsizeX,  *TVsizeY,  *TVsizeZ;
  double *ruler;
  long   *rulerx,*rulery,*rulerz,*orulerx,*orulery,*orulerz;
  long   *ActiveView;
  struct WIREFRAME *w_frame;
  long   version;
  struct tagNURBS  **MainNp;
  unsigned long    *Nnurbs;
  long   *dummyv[60];
  void   (*fpDA)(short);
  void   (*fpCV)(void);
  void   (*fpCE)(long, long);
  void   (*fpCF)(long, long, long);
  void   (*fpCS)(struct SKELETON *);
  BOOL   (*fpUVH)(long);
  BOOL   (*fpUEH)(long);
  BOOL   (*fpUFH)(long);
  void   (*fpTool_UP)(void);
  void   (*fpTool_RD)(void);
  BOOL   (*fpTool_CP)(int);
  int    (*fpTool_FC)(int,int,int,double,double);
  BOOL   (*fpTool_AC)(int);
  void   (*fpChooseColour)(unsigned char [], char *, HWND);
  int    (*fpChooseFile)(int, char *, char *, char *, char *, HWND);
  void   (*fpCentreDlgOnC)(HWND);
  void   (*fpCentreDlgOnS)(HWND);
  void   (*fpCopyFaceProp)(struct FACE *, struct FACE *);
  void   (*fpZeroSkeletonBoundingBox)(struct SKELETON *);
  void   (*fpGetSkeletonBoundBox)(struct SKELETON *, int);
  int    (*fpEraseVertex)(int);
  void   (*fpEraseVertexMapping)(void);
  short  (*fpEraseSkeleton)(struct SKELETON *);
  long   (*fpGetClosestVertex)(void);
  int    (*fpConnected)(long, long);
  int    (*fpNormalize)(vector);
  void   (*fpArbitraryRotate)(double, point, point, double [4][4]);
  void   (*fpOrientateFaceNormals)(long, long, long *);
  void   (*fpReverseFaceOrientation)(long, long *);
  BOOL   (*fpIsFOSame)(long, long, long, long, long, long);
  void   (*fpOrientateSelectedFaces)(void);
  void   (*fpAppendFileExtension)(char *, char *);
  void   (*fpEraseAllNurbs)(void);
  void   (*fpEraseNurbs)(struct tagNURBS *);
  void   (*fpCreateNurbs)(void);
  void   (*fpClearNurbsSkeleton)(struct SKELETON *);
  BOOL   (*fpAllocNurbs)(struct tagNURBS *, double *, double *);
  BOOL   (*fpCloneNurbs)(struct tagNURBS *, struct tagNURBS *);
  void   (*fpReleaseNurbs)(struct tagNURBS *);
  void   *dummyf[52];
} X__STRUCTURE;

typedef struct VERTEX {
 struct SKELETON *sp;
 char   status,gp;
 long   id;
 float  x,y;
 long   xyz[3];
} vertex;

typedef struct EDGE {
 long V[2];
} edge;

typedef struct FACE{
 long V[3];
 unsigned char color[3],matcol[3];
 unsigned char attrib,shader,map,axis,param;
 unsigned char pad;
} face;

typedef struct SKELETON {
 struct SKELETON *at,*next,*last;
 short  id,offset;
 vector u,v,w;
 double t[4][4];
 point  xyz,tp,bx[8];
 char   name[16];
 LPARAM lParam;
} skel;

typedef struct tagNurbsSurfSample {
    vector point, normal;   /* Point on surface, normal at that point */
    double normLen;         /* Used for normalizing normals */
    double u, v;            /* Parameters, e.g., used for texture mapping. */
    /* Note the parameter's range is determined by the surface's knot vector,
     * i.e., u goes from kvU[orderU-1] to kvU[numU], and likewise for v */
} NurbsSurfSample;

typedef struct tagNurbsProperties {
 char name[32];            /* name field             */
 long GranularityMax,      /* for visualising in the */
      GranularityMin;      /* view windows           */
 long     u_scale,
          v_scale;         /* texture coord scaling  */
 long     shader,          /* surface properties     */
          map,
          axis,
          param,
          smoothing,
          shiny,
          brilliance,
          reflectivity,
          transparency,
          glasstype,
          maptiled,
          mapshaded,
          mapdecal;
 long     pad[4];
 unsigned char colour[3],
          material_colour[3];
 char     hidden;
 char     padc[5];         /* struct 128 bytes       */
} NurbsProperties;

typedef struct tagNURBS {
 struct tagNURBS *next,*last;
 struct SKELETON *sp;
 long numU, numV;          /* Number of Points in the U and V directions */
 long orderU, orderV;      /* Order of the surface in U and V */
                           /* (must be >= 2, < MAXORDER) */
 double * kvU, * kvV;      /* Knot vectors, indexed as [0..numU+orderU-1]
                                                    and [0..numV+orderV-1] */
 vector4 **points;         /* Control points, indexed as points
                              [0..numV-1][0..numU-1] */
                           /* Note the w values are *premultiplied*
                              with the x, y and z values */
 BOOL selected;            /* TRUE if selected       */
 NurbsProperties properties;
 /* These fields are added to support subdivision                     */
 BOOL strV0, strVn,        /* Edge straightness flags for subdivision */
      strU0, strUn;
 BOOL flatV, flatU;        /* Surface flatness flags for subdivision  */
 NurbsSurfSample c00, c0n,
                 cn0, cnn; /* Corner data structures for subdivision  */
 /* eg.
  Cubic NURBS mimimum requirement 4 control points
  => num = 4  order = 4   => 8 Index:  knots [0 - 3]  points [0 - 3][0 - 3]
 in the terms of Watt and Watt spline spec. is n=3 k=4 => knots = 3+4+1
 Sum i=0,n  (points = n+1)
 In knot vector first parameter repeated "k" times [0 0 0 0  1 1 1 1]
 */
} nurbs;

typedef struct tagIMAGEMAP {
short   in_use;  /* use this for partial saves */
char    N[32],*S,*R,*B,*T;
char    s,r,b,t;
short   sp,rp,bp,tp;
char    Type,TypeR,TypeB,TypeT,Map,Lock,Key;
short   AnimFF,AnimDL,AnimLF,Angle,SkId;
short   AnimFFr,AnimDLr,AnimLFr;
short   AnimFFb,AnimDLb,AnimLFb;
short   AnimFFt,AnimDLt,AnimLFt;
point   P,X,Y;
skel    *Sk;
long    Vlock[3];
unsigned char d_colour[3],s_colour[3],a_colour[3],t_colour[3],k_colour[3];
} IMAGEMAP;

typedef struct tagSHADERAXIS {
char    N[32];
point   P,X,Y;
} SHADERAXIS;

typedef struct tagXSHADER {
  long in_use;
  char name[128],*parameters;
} XSHADER;

typedef struct WIREFRAME {
 long Np,Ne;
 twoLongs *e;
 point *p;
} wireframe;

