/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file TOOLBOX .C */

// Handles the user interface code to draw the tool pannel
// and handle its window function. It also handles the coordinates
// floating pannel and the maps / lock etc. pannel

#define MODULE_TOOLBOX 1

#include "design.h"

#define NO_MESSAGE -1
#define NO_HELP    -1

typedef struct tagTOOL_STRUCT{
 long  Message,Help,Tip;
} TOOL_STRUCT;

extern HBITMAP hBmBackground;

#define LARGE_TOOL_X 50
#define LARGE_TOOL_Y 40
#define SMALL_TOOL_X 32
#define SMALL_TOOL_Y 32
#define N_ICONS      12

static HBITMAP hBitmapB=NULL,
               hBitmapH=NULL,
               hBitmapD=NULL,
               hBitmapA=NULL,
               hBitmapI=NULL,
               hBitmapM=NULL;
static BOOL    tools_enable=TRUE;
static long    cxToolBtn,cyToolBtn;
static char    szToolsClass[]="OFX:AnimatorToolsClass";
static LRESULT CALLBACK ToolsWndProc(HWND,UINT,WPARAM,LPARAM);
static TOOL_STRUCT aTools[N_ICONS]={
  IDM_STOP,                IDI_OBJECT_SELECTOR,           IDI_TIP_T_1,
  IDM_TOOLS_NODE,          IDI_VERTEX_SELECTOR,           IDI_TIP_T_2,
  IDM_TOOLS_SELECTOR,      IDI_SELECT_SELECTOR,           IDI_TIP_S_1,
  IDM_TOOLS_PAN,           IDI_PAN,                       IDI_TIP_T_3,
  IDM_TOOLS_ZOOMIN,        IDI_ZOOM,                      IDI_TIP_T_4,
  IDM_TOOLS_GRABBER,       IDI_TOOLS_GRABBER,             IDI_TIP_T_5,
  IDM_TOOLS_ROTATOR,       IDI_TOOLS_ROTATOR,             IDI_TIP_T_6,
  IDM_TOOLS_EXPANDER,      IDI_TOOLS_EXPANDER,            IDI_TIP_T_7,
  IDM_TOOLS_PLOTTER,       IDI_TOOLS_PLOTTER,             IDI_TIP_T_8,
  IDM_TOOLS_MAGNET_ATTRACT,IDI_TOOLS_MAGNET_ATTRACT,      IDI_TIP_T_A,
  IDM_PRIMITIVE_QADD,      IDI_TOOLS_QPRIMITIVES,         IDI_TIP_T_B,
  IDM_VIEW_RENDER,         IDI_CONTROL_RENDER,            IDI_TIP_A_1
};

static void ToolboxCreate(void);
static void PannelSelectTool(HWND hwnd, int *oldtool, int newtool);
static void DoAddTooltip(HWND hwndOwner, HWND tipon, int resid);
static HWND DoCreateTooltip(HWND hwndOwner);


void ToolboxShow(BOOL fShow){
 if(ghwndTools == NULL)ToolboxCreate();
}

void ToolboxDestroy(void){
 if(ghwndCoord1)DestroyWindow(ghwndCoord1);
 if(ghwndLock)DestroyWindow(ghwndLock);
}

void EnableTools(BOOL status){
  tools_enable=status;
  InvalidateRect(ghwndTools,NULL,FALSE);
}

void MoveToolsWindow(void){
 int i,x,y;
 RECT rc1;
 POINT pt;
 if(ghwndTools == NULL)return;
 GetWindowRect(ghwnd_view,&rc1);
 if(bToolpannelLeft)x=WindowGap;
 else x=rc1.right+3;
 y=rc1.top;
 pt.x=x; pt.y=y;
 ScreenToClient(ghwnd_main,&pt);
 x=pt.x; y=pt.y;
 for(i=0;i<N_ICONS;i++){
   MoveWindow(GetDlgItem(ghwndTools,i),
              0,cyToolBtn*i,cxToolBtn,cyToolBtn,FALSE);
 }
 MoveWindow(ghwndTools,x,y,cxToolBtn,cyToolBtn*N_ICONS,TRUE);
}

void ChangeToolboxBitmaps(void){
 if(hBitmapB != NULL)DeleteObject(hBitmapB);  hBitmapB=NULL;
 if(hBitmapH != NULL)DeleteObject(hBitmapH);  hBitmapH=NULL;
 if(hBitmapD != NULL)DeleteObject(hBitmapD);  hBitmapD=NULL;
 if(hBitmapA != NULL)DeleteObject(hBitmapA);  hBitmapA=NULL;
 if(hBitmapI != NULL)DeleteObject(hBitmapI);  hBitmapI=NULL;
 if(hBitmapM != NULL)DeleteObject(hBitmapM);  hBitmapM=NULL;
 if(bToolpannelSmall){
   cxToolBtn=SMALL_TOOL_X;
   cyToolBtn=SMALL_TOOL_Y;
   hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_B));
   hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_H));
   hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_D));
   hBitmapA=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_A));
   hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_I));
   hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_M));
 }
 else{
   cxToolBtn=LARGE_TOOL_X;
   cyToolBtn=LARGE_TOOL_Y;
   hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_B));
   hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_H));
   hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_D));
   hBitmapA=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_A));
   hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_I));
   hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_M));
 }
}

static long GetWhichButton(void){
  long  i,button_index = -1;
  POINT pt;
  RECT  rc;
  GetCursorPos(&pt);
  for(i=0;i<N_ICONS;i++){
    GetWindowRect(GetDlgItem(ghwndTools,i),&rc);
    if(PtInRect(&rc,pt))button_index=i;
  }
  return button_index;
}

static void ToolboxCreate(void){
 char title[256];
 int x,y;
 RECT rc,rc1;
 POINT pt;
// if(Preferences.tooltips){
 if((ghwndTooltip=DoCreateTooltip(ghwnd_main)) == NULL)return;
 SetWindowPos(ghwndTooltip,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE | SWP_NOSIZE);
 SendMessage(ghwndTooltip,TTM_ACTIVATE,(WPARAM)Preferences.tooltips,0);
// }
 GetWindowRect(ghwnd_view,&rc1);
 if(bToolpannelLeft)x=WindowGap;
 else x=rc1.right+3;
 y=rc1.top;
 pt.x=x; pt.y=y;
 ScreenToClient(ghwnd_main,&pt);
 x=pt.x; y=pt.y;
 if(bToolpannelSmall){
   cxToolBtn=SMALL_TOOL_X;
   cyToolBtn=SMALL_TOOL_Y;
 }
 else{
   cxToolBtn=LARGE_TOOL_X;
   cyToolBtn=LARGE_TOOL_Y;
 }
 ghwndTools=CreateWindowEx(
                         WS_EX_TRANSPARENT,
                         szToolboxClass,NULL,
                         WS_CHILD | !WS_BORDER | WS_VISIBLE,
                         x,y,
                         cxToolBtn,cyToolBtn*N_ICONS,
                         ghwnd_main,
                         (HMENU)101,
                         ghinst_main,NULL);
 ghwndCoord1=CreateDialog(ghinst_main,MAKEINTRESOURCE(DLG_COORDS),
                          ghwnd_main,(DLGPROC)CoordinatesDlgProc);
 ghwndLock=CreateDialog(ghinst_main,MAKEINTRESOURCE(DLG_LOCK),
                        ghwnd_main,(DLGPROC)LockDlgProc);
 return;
}

LRESULT CALLBACK ToolboxWndProc(HWND hwnd,UINT msg,
                                WPARAM wParam,LPARAM lParam){
 long i;
 HWND hWndChild;
 switch (msg) {
   case WM_CREATE:
     if(bToolpannelSmall){
       hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_B));
       hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_H));
       hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_D));
       hBitmapA=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_A));
       hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_I));
       hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_STOOLBAR_M));
     }
     else{
       hBitmapB=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_B));
       hBitmapH=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_H));
       hBitmapD=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_D));
       hBitmapA=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_A));
       hBitmapI=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_I));
       hBitmapM=LoadBitmap(ghinst_main,MAKEINTRESOURCE(IDBM_TOOLBAR_M));
     }
     for(i=0;i<N_ICONS;i++){
       hWndChild=CreateWindowEx(
                              WS_EX_TRANSPARENT,
                              szToolBtnClass,NULL,
                              WS_CHILD | WS_VISIBLE,
                              0,cyToolBtn*i,
                              cxToolBtn,cyToolBtn,
                              hwnd,(HMENU)i,
                              ghinst_main,NULL);
       DoAddTooltip(ghwnd_main,hWndChild,aTools[i].Tip);
     }
     break;
   case WM_DESTROY:
     if(hBitmapB != NULL)DeleteObject(hBitmapB);  hBitmapB=NULL;
     if(hBitmapH != NULL)DeleteObject(hBitmapH);  hBitmapH=NULL;
     if(hBitmapD != NULL)DeleteObject(hBitmapD);  hBitmapD=NULL;
     if(hBitmapA != NULL)DeleteObject(hBitmapA);  hBitmapA=NULL;
     if(hBitmapI != NULL)DeleteObject(hBitmapI);  hBitmapI=NULL;
     if(hBitmapM != NULL)DeleteObject(hBitmapM);  hBitmapM=NULL;
     break;
   default:
     return DefWindowProc(hwnd, msg, wParam, lParam);
 }
 return 0;
}

LRESULT CALLBACK ToolBtnWndProc(HWND hwnd,UINT msg,
                                WPARAM wParam,LPARAM lParam){
 static int  toolTimer = -1, down_tool;
 static BOOL bCaptured=FALSE,bCancel=FALSE,bLastCancel=FALSE,bBusy=FALSE;
 long w,wold,hid;
 MSG mmsg;
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 hid=(int)GetWindowLong(hwnd,GWL_ID);
 switch (msg) {
   case WM_TIMER:
     if(bBusy || toolTimer < 0)break;
     bBusy=TRUE;
     w=GetWhichButton();
     if(w < 0 && toolTimer >= 0){
       KillTimer(hwnd,toolTimer);
       toolTimer = -1;
     }
     if( w < 0 && gCurTool >= 0){
       wold=gCurTool;
       gCurTool= -1;
       InvalidateRect(GetDlgItem(ghwndTools,wold),NULL,FALSE);
     }
     bBusy=FALSE;
     break;
   case WM_SETCURSOR:{
       hid=aTools[hid].Help;
       if(current_message != hid){
         if(hid == NO_HELP)SendPrgmMessage(IDS_ZERO,1);
         else{
           SendIconHelpMessage(hid);
         }
       }
       current_message=hid;
     }
     return DefWindowProc(hwnd,msg,wParam,lParam);
   case WM_MOUSEMOVE:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     w=GetWhichButton();
     if(bCaptured){
       if(w == hid)bCancel=FALSE;
       else        bCancel=TRUE;
       if(bCancel != bLastCancel)InvalidateRect(hwnd,NULL,FALSE);
       bLastCancel=bCancel;
     }
     else{
       wold=gCurTool;
       gCurTool=w;
       if(wold != w){
         if(wold >= 0 && wold  < N_ICONS)
           InvalidateRect(GetDlgItem(ghwndTools,wold),NULL,FALSE);
         if(gCurTool >= 0 && gCurTool < N_ICONS)
           InvalidateRect(GetDlgItem(ghwndTools,gCurTool),NULL,FALSE);
       }
       if(toolTimer < 0)toolTimer=SetTimer(hwnd,100,500,NULL);
     }
     break;
   case WM_LBUTTONDBLCLK:
     w=GetWhichButton();
     if(w == 2)PostMessage(ghwnd_main,WM_COMMAND,IDM_TOOLS_DESELECTOR,0);
     if(w == 3)PostMessage(ghwnd_main,WM_COMMAND,IDM_ZOOM_CENTRE,0);
     if(w == 4)PostMessage(ghwnd_main,WM_COMMAND,IDM_VIEW_FULL,0);
     if(w == 5)PostMessage(ghwnd_main,WM_COMMAND,IDM_ACTIONS_MOVE,0);
     if(w == 6)PostMessage(ghwnd_main,WM_COMMAND,IDM_ACTIONS_ROTATE_CLOCKWISE,0);
     if(w == 7)PostMessage(ghwnd_main,WM_COMMAND,IDM_ACTIONS_EXPAND,0);
     if(w == 8)PostMessage(ghwnd_main,WM_COMMAND,IDM_TOOLS_PLOT3D,0);
     if(w == 9)PostMessage(ghwnd_main,WM_COMMAND,IDM_TOOLS_SHAPER,0);
     break;
   case WM_LBUTTONDOWN:
     if(toolTimer >= 0)KillTimer(hwnd,toolTimer); toolTimer = -1;
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     bCaptured=TRUE;
     bCancel=bLastCancel=FALSE;
     down_tool=hid;
     SetCapture(hwnd);
     InvalidateRect(hwnd,NULL,FALSE);
     break;
   case WM_LBUTTONUP:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,0,(LPARAM)(LPMSG)&mmsg);
     ReleaseCapture();
     w=GetWhichButton();
     bCaptured=FALSE;
     if(w < 0 && gCurTool >= 0){
       wold=gCurTool;
       gCurTool = -1;
       InvalidateRect(GetDlgItem(ghwndTools,wold),NULL,FALSE);
     }
     InvalidateRect(hwnd,NULL,FALSE);
     if(down_tool == w && aTools[down_tool].Message != NO_MESSAGE)
       PostMessage(ghwnd_main,WM_COMMAND,aTools[w].Message,0);
     gCurTool= -1;
     break;
   case WM_RBUTTONDOWN:{
       POINT pt;
       HMENU hMenu=LoadMenu(ghinst_main,"ToolbarPopupMenu");
       if(hMenu == NULL)break;
       pt.x=LOWORD(lParam);
       pt.y=HIWORD(lParam);
       ClientToScreen(hwnd,&pt);
       if(bToolpannelLeft)
         CheckMenuItem(hMenu,IDM_CONTEXT_TOOLSLEFT,MF_CHECKED);
       if(bToolpannelSmall)
         CheckMenuItem(hMenu,IDM_CONTEXT_TOOLSSMALL,MF_CHECKED);
       TrackPopupMenu(GetSubMenu(hMenu,0),TPM_LEFTALIGN | TPM_LEFTBUTTON,
                      pt.x,pt.y,0,ghwnd_main,NULL);
       DestroyMenu(hMenu);
     }
     break;
   case WM_PAINT:{
       HDC hDC;
       PAINTSTRUCT ps;
       hDC=BeginPaint(hwnd,&ps);
       if(hBitmapB != NULL && hBitmapD != NULL  && hBitmapH != NULL){
         HDC hMemDC;
         HBITMAP hbmOld;
         BOOL blit=FALSE;
         hMemDC = CreateCompatibleDC(hDC);
         if(!tools_enable){
           hbmOld = SelectObject(hMemDC,hBitmapI);
           blit=TRUE;
         }
         else if ((hwnd == GetCapture() && !bCancel)  ||
                  FALSE){
           if( (tool == PLOT3D     && hid == 8)      ||
               (tool == DESELECTOR && hid == 2)      ||
               (tool == SHAPER     && hid == 9))
                hbmOld = SelectObject(hMemDC,hBitmapA);
           else hbmOld = SelectObject(hMemDC,hBitmapD);
         }
         else if( (tool == NOTOOL     && hid == 0)      ||
                  (tool == NODETOOL   && hid == 1)      ||
                  (tool == SELECTOR   && hid == 2)      ||
                  (tool == PAN        && hid == 3)      ||
                  (tool == INZOOM     && hid == 4)      ||
                  (tool == GRABBER    && hid == 5)      ||
                  (tool == ROTATOR    && hid == 6)      ||
                  (tool == EXPANDER   && hid == 7)      ||
                  (tool == PLOTTER    && hid == 8)      ||
                  (tool == MAGNETTO   && hid == 9)      ||
                  FALSE)
           hbmOld = SelectObject(hMemDC,hBitmapD);
         else if( (tool == PLOT3D     && hid == 8)      ||
                  (tool == DESELECTOR && hid == 2)      ||
                  (tool == SHAPER     && hid == 9))
           hbmOld = SelectObject(hMemDC,hBitmapA);
         else if(hid == gCurTool)
           hbmOld = SelectObject(hMemDC,hBitmapH);
         else{
           hbmOld = SelectObject(hMemDC,hBitmapB);
           blit=TRUE;
         }
         if(hBmBackground == NULL)blit=FALSE;
         if(blit){
           HDC hdc_view;
           RECT rc;
           POINT pp;
           BITMAP bm;
           HBITMAP hbm_view,hbmoldv;
           hbm_view=CreateCompatibleBitmap(hMemDC,cxToolBtn,cyToolBtn);
           hdc_view=CreateCompatibleDC(hMemDC);
           hbmoldv=SelectObject(hdc_view,hbm_view);
           GetWindowRect(hwnd,&rc);
           pp.x=rc.left; pp.y=rc.top;
           ScreenToClient(ghwnd_main,&pp);
           GetObject(hBmBackground,sizeof(BITMAP),&bm);
           pp.x = pp.x % bm.bmWidth;
           pp.y = pp.y % bm.bmHeight;
           SelectObject(hMemDC,hBmBackground);
           if(pp.y + cyToolBtn > bm.bmHeight){
             if(pp.x + cxToolBtn > bm.bmWidth){
               BitBlt(hDC,
                  0,0,
                  bm.bmWidth -pp.x,
                  bm.bmHeight-pp.y,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
               BitBlt(hDC,
                  bm.bmWidth-pp.x,bm.bmHeight-pp.y,
                  cxToolBtn - (bm.bmWidth -pp.x),
                  cyToolBtn - (bm.bmHeight-pp.y),
                  hMemDC,
                  (pp.x + (bm.bmWidth -pp.x)) % bm.bmWidth,
                  (pp.y + (bm.bmHeight-pp.y)) % bm.bmHeight,
                  SRCCOPY);
             }
             else{
               BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  bm.bmHeight-pp.y,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
               BitBlt(hDC,
                  0,bm.bmHeight-pp.y,
                  cxToolBtn,
                  cyToolBtn - (bm.bmHeight-pp.y),
                  hMemDC,
                  pp.x,(pp.y + (bm.bmHeight-pp.y)) % bm.bmHeight,
                  SRCCOPY);
             }
           }
           else if(pp.x + cxToolBtn > bm.bmWidth){
             BitBlt(hDC,
                  0,0,
                  bm.bmWidth -pp.x,
                  cyToolBtn,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
             BitBlt(hDC,
                  bm.bmWidth-pp.x,0,
                  cxToolBtn - (bm.bmWidth -pp.x),
                  cyToolBtn,
                  hMemDC,
                  (pp.x + (bm.bmWidth -pp.x)) % bm.bmWidth,pp.y,
                  SRCCOPY);
           }
           else BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  pp.x,pp.y,
                  SRCCOPY);
           SelectObject(hMemDC,hBitmapM);
           BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  0x00220326); // DSna
           if(tools_enable)SelectObject(hMemDC,hBitmapB);
           else            SelectObject(hMemDC,hBitmapI);
           BitBlt(hdc_view,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  SRCCOPY);
           SelectObject(hMemDC,hBitmapM);
           BitBlt(hdc_view,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  SRCAND);
           BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hdc_view,
                  0,0,
                  SRCPAINT);
           SelectObject(hdc_view,hbmoldv);
           DeleteObject(hbm_view);
           DeleteDC(hdc_view);
         }
         else BitBlt(hDC,
                  0,0,
                  cxToolBtn,
                  cyToolBtn,
                  hMemDC,
                  0,hid*cyToolBtn,
                  SRCCOPY);
         SelectObject(hMemDC,hbmOld);
         DeleteDC(hMemDC);
       }
       EndPaint(hwnd,&ps);
     }
     break;
   default:
     return DefWindowProc(hwnd,msg,wParam,lParam);
 }
 return 0;
}

void ToolboxSetPosition(int w){
}

/////////////////////////////////////////////////////////////////////

static void DrawQtBtn(BOOL selected,HDC hDC, LPRECT lprc, HBITMAP hbm){
 HDC hMemDC;
 BITMAP  bmp;
 HBITMAP hbmOld;
 HBRUSH holdbrush,hbrushFace;
 HPEN   holdpen,hpenShadow,hpenHighlight;
 int x,y,cx,cy;
 hbrushFace=CreateSolidBrush(RGB(192,192,192));
 holdbrush=SelectObject(hDC,hbrushFace);
 hpenShadow=CreatePen(PS_SOLID,0,RGB(128,128,128));
 hpenHighlight=CreatePen(PS_SOLID,0,RGB(255,255,255));
 holdpen=SelectObject(hDC,GetStockObject(BLACK_PEN));
 Rectangle(hDC,0,0,lprc->right+1,lprc->bottom);
 SelectObject(hDC,hpenShadow);
 if(selected){
   MoveToEx(hDC,1,lprc->bottom-2,NULL);
   LineTo(hDC,1,1);
   LineTo(hDC,lprc->right-1,1);
 }
 else{
   MoveToEx(hDC,1,lprc->bottom-2,NULL);
   LineTo(hDC,lprc->right-1,lprc->bottom-2);
   LineTo(hDC,lprc->right-1,1);
   MoveToEx(hDC,2,lprc->bottom-3,NULL);
   LineTo(hDC,lprc->right-2,lprc->bottom-3);
   LineTo(hDC,lprc->right-2,3);
   SelectObject(hDC,hpenHighlight);
   MoveToEx(hDC,1,lprc->bottom-3,NULL);
   LineTo(hDC,1,1);
   LineTo(hDC,lprc->right-1,1);
   MoveToEx(hDC,2,lprc->bottom-4,NULL);
   LineTo(hDC,2,2);
   LineTo(hDC,lprc->right-2,2);
 }
 SelectObject(hDC,holdpen);
 DeleteObject(hpenShadow);
 DeleteObject(hpenHighlight);
 SelectObject(hDC,holdbrush);
 DeleteObject(hbrushFace);
 if(hbm == NULL)return;
 GetObject(hbm,sizeof(BITMAP),&bmp);
 cx = bmp.bmWidth;
 cy = bmp.bmHeight;
 x=max(0,(lprc->right-cx)/2); y=max(0,(lprc->bottom-cy)/2)+1;
 if(selected){x++; y++;}
 hMemDC = CreateCompatibleDC(hDC);
 hbmOld = SelectObject(hMemDC, hbm);
 BitBlt(hDC,x,y,cx,cy,hMemDC,0,0,SRCAND);
 SelectObject(hMemDC, hbmOld);
 DeleteDC(hMemDC);
}

static void GetFormatString(char *format, double x, double y, double z,
 double dx, double dy, double dz){
 if(fabs(x)  > 9999999.0 || fabs(y)  > 9999999.0 ||
    fabs(z)  > 9999999.0 || fabs(dx) > 9999999.0 ||
    fabs(dy) > 9999999.0 || fabs(dz) > 9999999.0)
   sprintf(format,"%s","%9.1le");
 else if(fabs(x)  > 99999.0 || fabs(y)  > 99999.0 ||
         fabs(z)  > 99999.0 || fabs(dx) > 99999.0 ||
         fabs(dy) > 99999.0 || fabs(dz) > 99999.0)
   sprintf(format,"%s","%9.0lf");
 else  sprintf(format,"%s","%9.2lf");
}

static void GetDoubleCoordinates(int ix, int iy, int iz,
                     double *x,  double *y,  double *z,
                     double *dx, double *dy, double *dz){
 *x = (double)(ix-lrulerx);
 *y = (double)(iy-lrulery);
 *z = (double)(iz-lrulerz);
 *dx = (double)(ix-rulerx);
 *dy = (double)(iy-rulery);
 *dz = (double)(iz-rulerz);
 *x  /= ruler; *y   /=ruler; *z   /=ruler;
 *dx /= ruler; *dy /= ruler; *dz /= ruler;
}

static struct QC {int idBitmap; int dlgID; int idHelp; int iTip;
                 HWND hwnd;WNDPROC fpSclass; HBITMAP hbm;} QtCoord[4]={
  IDBM_CO1,DLG_COORDS_GOTO,     IDI_COORD_MEASURE, IDI_TIP_C_1,NULL,NULL,NULL,
  IDBM_CO2,DLG_COORDS_CANCEL,   IDI_COORD_RELATIVE,IDI_TIP_C_2,NULL,NULL,NULL,
  IDBM_CO3,DLG_COORDS_SETCENTRE,IDI_COORD_ORIGIN,  IDI_TIP_C_3,NULL,NULL,NULL,
  IDBM_CO4,DLG_COORDS_DEFINE,   IDI_COORD_DEFINE,  IDI_TIP_C_4,NULL,NULL,NULL
};

static LRESULT CALLBACK CoordinatesSubclassProc(HWND hwnd,UINT msg,
                                       WPARAM wParam,LPARAM lParam){
 MSG mmsg;
 int i,hid;
 for(i=0;i<4;i++){
   if(QtCoord[i].hwnd == hwnd)goto XXIT;
 }
 MessageBeep(MB_OK);
 XXIT:
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 switch (msg) {
   case WM_LBUTTONDOWN:
   case WM_LBUTTONUP:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,
                                         0,(LPARAM)(LPMSG)&mmsg);
     break;
   case WM_MOUSEMOVE:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,
                                         0,(LPARAM)(LPMSG)&mmsg);
     hid=QtCoord[i].idHelp;
     if(hid == NO_HELP)SendPrgmMessage(IDS_ZERO,1);
         else SendIconHelpMessage(hid);
     break;
   default: break;
 }
 return CallWindowProc(QtCoord[i].fpSclass,hwnd,msg,wParam,lParam);
}

LRESULT CALLBACK CoordinatesDlgProc(HWND hdlg,UINT msg,
                            WPARAM wparam,LPARAM lparam){
 BOOL err;
 RECT rc;
 int i,ix,iy,iz;
 double x,y,z,dx,dy,dz;
 char string[32],format[10];
 static int xpos=-1,ypos=-1,infocus=0,lastid=-1,udflag=1;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     for(i=0;i<4;i++){
       QtCoord[i].hbm=LoadBitmap(ghinst_main,
                                 MAKEINTRESOURCE(QtCoord[i].idBitmap));
       QtCoord[i].hwnd=GetDlgItem(hdlg,QtCoord[i].dlgID);
       if(QtCoord[i].hwnd != NULL){
         QtCoord[i].fpSclass=(WNDPROC)SetWindowLong(QtCoord[i].hwnd,
                              GWL_WNDPROC,(LONG)CoordinatesSubclassProc);
         DoAddTooltip(ghwnd_main,QtCoord[i].hwnd,QtCoord[i].iTip);
       }
     }
     if(xpos == -1 && ypos == -1){
       xpos=GetPrivateProfileInt(IniSection,"XCOORDPOS",-1,
                                 IniFilename);
       if(xpos < 0)xpos=52;
       ypos=GetPrivateProfileInt(IniSection,"YCOORDPOS",-1,
                                 IniFilename);
       if(ypos < 0)ypos=50;
     }
     SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     udflag=1;
     PostMessage(hdlg,WM_COMMAND,IDC_COORD_UPDATE,0);
     return (TRUE);
   case WM_CTLCOLORSTATIC:
   case WM_CTLCOLORDLG:
     SetBkColor((HDC)wparam,RGB(192,192,192));
     return (LRESULT)GetStockObject(LTGRAY_BRUSH);
   case WM_DRAWITEM:{
       BOOL status;
       HBITMAP hbm;
       LPDRAWITEMSTRUCT lpdis;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->itemState & ODS_SELECTED)status=TRUE; else status=FALSE;
       if     (lpdis->CtlID == DLG_COORDS_GOTO)hbm=QtCoord[0].hbm;
       else if(lpdis->CtlID == DLG_COORDS_CANCEL)hbm=QtCoord[1].hbm;
       else if(lpdis->CtlID == DLG_COORDS_SETCENTRE)hbm=QtCoord[2].hbm;
       else if(lpdis->CtlID == DLG_COORDS_DEFINE)hbm=QtCoord[3].hbm;
       DrawQtBtn(status,lpdis->hDC,&(lpdis->rcItem),hbm);
     }
     return TRUE;
   case WM_DESTROY:
     for(i=0;i<4;i++){
       if(QtCoord[i].hbm != NULL)DeleteObject(QtCoord[i].hbm);
       if(QtCoord[i].fpSclass != NULL)SetWindowLong(QtCoord[i].hwnd,
         GWL_WNDPROC,(LONG)QtCoord[i].fpSclass);
     }
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     sprintf(string,"%ld",xpos);
     WritePrivateProfileString(IniSection,"XCOORDPOS",string,IniFilename);
     sprintf(string,"%ld",ypos);
     WritePrivateProfileString(IniSection,"YCOORDPOS",string,IniFilename);
     break;
   case WM_NCHITTEST:
     if(current_message >= 0){
       current_message = -1;
       SendPrgmMessage(IDS_ZERO,1);
     }
     break;
   case WM_COMMAND:
     switch(HIWORD(wparam)){
       case EN_SETFOCUS:
         infocus=1;
         break;
       case EN_KILLFOCUS:
         infocus=0;
         break;
       case EN_CHANGE:
         if(!infocus)break;
         udflag=1;
         i=(int)LOWORD(wparam);
         if(GetDlgItemText(hdlg,i,string,30) != 0){
           if(i == DLG_COORDS_X){
             if(lastid == DLG_COORDS_DX)break;
             ix=(int)(atof(string)*ruler)+lrulerx;
             GetDoubleCoordinates(ix,NpointerY,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,dx);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_DX,string);
             lastid=-1;
           }
           else if(i == DLG_COORDS_Y){
             if(lastid == DLG_COORDS_DY)break;
             iy=(int)(atof(string)*ruler)+lrulery;
             GetDoubleCoordinates(NpointerX,iy,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,dy);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_DY,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_Z){
             if(lastid == DLG_COORDS_DZ)break;
             iz=(int)(atof(string)*ruler)+lrulerz;
             GetDoubleCoordinates(NpointerX,NpointerY,iz,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,dz);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_DZ,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_DX){
             if(lastid == DLG_COORDS_X)break;
             ix=(int)(atof(string)*ruler)+rulerx;
             GetDoubleCoordinates(ix,NpointerY,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,x);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_X,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_DY){
             if(lastid == DLG_COORDS_Y)break;
             iy=(int)(atof(string)*ruler)+rulery;
             GetDoubleCoordinates(NpointerX,iy,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,y);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_Y,string);
             lastid= -1;
           }
           else if(i == DLG_COORDS_DZ){
             if(lastid == DLG_COORDS_Z)break;
             iz=(int)(atof(string)*ruler)+rulerz;
             GetDoubleCoordinates(NpointerX,NpointerY,iz,
                              &x,&y,&z,&dx,&dy,&dz);
             GetFormatString(format,x,y,z,dx,dy,dz);
             sprintf(string,format,z);
             lastid=i;
             SetDlgItemText(hdlg,DLG_COORDS_Z,string);
             lastid= -1;
           }
         }
         break;
       default:   break;
     }
     switch(LOWORD(wparam)){
       case IDCANCEL:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_WINDOW_COORDS,0);
         break;
       case IDOK:
       case DLG_COORDS_GOTO:{
           point p;
           if(GetDlgItemText(hdlg,DLG_COORDS_X,string,30) == 0)break;
           p[0]=(long)(atof(string)*ruler)+lrulerx;
           if(GetDlgItemText(hdlg,DLG_COORDS_Y,string,30) == 0)break;
           p[1]=(long)(atof(string)*ruler)+lrulery;
           if(GetDlgItemText(hdlg,DLG_COORDS_Z,string,30) == 0)break;
           p[2]=(long)(atof(string)*ruler)+lrulerz;
           if( (p[0] > TVpointX && p[0] < TVpointX+TVsizeX)
            && (p[1] > TVpointY && p[1] < TVpointY+TVsizeY)
            && (p[2] > TVpointZ && p[2] < TVpointZ+TVsizeZ)){ /* move 3D cursor */
             Draw3dCursor();
             NpointerX=p[0];  NpointerY=p[1];  NpointerZ=p[2];
             Draw3dCursor();
           }
           else{
             NpointerX=p[0];  NpointerY=p[1];  NpointerZ=p[2];
             TVpointX=NpointerX-TVsizeX/2;  TVcentreX=NpointerX;
             TVpointY=NpointerY-TVsizeY/2;  TVcentreY=NpointerY;
             TVpointZ=NpointerZ-TVsizeZ/2;  TVcentreZ=NpointerZ;
             DrawModel();
           }
           PostMessage(hdlg,WM_COMMAND,IDC_COORD_UPDATE,0);
         }
         break;
       case DLG_COORDS_CANCEL:
         if(HIWORD(wparam) != BN_CLICKED)break;
         udflag=1;
         PostMessage(hdlg,WM_COMMAND,IDC_COORD_UPDATE,0);
         break;
       case DLG_COORDS_SETCENTRE:
         if(HIWORD(wparam) != BN_CLICKED)break;
         SendMessage(ghwnd_main,WM_COMMAND,IDM_COORDS_ORIGIN,0);
         break;
       case DLG_COORDS_DEFINE:
         if(HIWORD(wparam) != BN_CLICKED)break;
         SendMessage(ghwnd_main,WM_COMMAND,IDM_COORDS_DEFINE,0);
         break;
       case IDC_COORD_UPDATE:
         GetDoubleCoordinates(NpointerX,NpointerY,NpointerZ,
                              &x,&y,&z,&dx,&dy,&dz);
         GetFormatString(format,x,y,z,dx,dy,dz);
         sprintf(string,format,x);  SetDlgItemText(hdlg,DLG_COORDS_X,string);
         sprintf(string,format,y);  SetDlgItemText(hdlg,DLG_COORDS_Y,string);
         sprintf(string,format,z);  SetDlgItemText(hdlg,DLG_COORDS_Z,string);
         rulerx=NpointerX; rulery=NpointerY; rulerz=NpointerZ;
         if(udflag){
           sprintf(string,format,0.0);SetDlgItemText(hdlg,DLG_COORDS_DX,string);
           sprintf(string,format,0.0);SetDlgItemText(hdlg,DLG_COORDS_DY,string);
           sprintf(string,format,0.0);SetDlgItemText(hdlg,DLG_COORDS_DZ,string);
           udflag=0;
         }
         break;
       default:   break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static struct QL {int idBitmap; int dlgID; int idHelp; int iTip;
                 HWND hwnd;WNDPROC fpSclass; HBITMAP hbm;} QtLock[5]={
  IDBM_LOCK1,DLG_LOCK_SQ, IDI_LOCK_SNAP1,  IDI_TIP_M_1,NULL,NULL,NULL,
  IDBM_LOCK2,DLG_LOCK_CI, IDI_LOCK_SNAP2,  IDI_TIP_M_2,NULL,NULL,NULL,
  IDBM_LOCK3,DLG_LOCK_TR, IDI_LOCK_SNAP3,  IDI_TIP_M_3,NULL,NULL,NULL,
  IDBM_LOCK4,DLG_LOCK_LK, IDI_LOCK_LOCK,   IDI_TIP_M_4,NULL,NULL,NULL,
  IDBM_LOCK5,DLG_LOCK_XX, NO_HELP,         IDI_TIP_M_5,NULL,NULL,NULL
};

static LRESULT CALLBACK LockSubclassProc(HWND hwnd,UINT msg,
                                         WPARAM wParam,LPARAM lParam){
 MSG mmsg;
 int i,hid;
 for(i=0;i<5;i++){
   if(QtLock[i].hwnd == hwnd)goto XXIT;
 }
 MessageBeep(MB_OK);
 XXIT:
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 switch (msg) {
   case WM_LBUTTONDOWN:
   case WM_LBUTTONUP:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,
                                         0,(LPARAM)(LPMSG)&mmsg);
     break;
   case WM_MOUSEMOVE:
     if(ghwndTooltip != NULL)SendMessage(ghwndTooltip,TTM_RELAYEVENT,
                                         0,(LPARAM)(LPMSG)&mmsg);
     hid=QtLock[i].idHelp;
     if(hid == NO_HELP)SendPrgmMessage(IDS_ZERO,1);
         else SendIconHelpMessage(hid);
     break;
   default: break;
 }
 return CallWindowProc(QtLock[i].fpSclass,hwnd,msg,wParam,lParam);
}

LRESULT CALLBACK LockDlgProc(HWND hdlg,UINT msg,
                             WPARAM wparam,LPARAM lparam){
 RECT rc;
 int i;
 static int xpos=-1,ypos=-1;
 switch( msg ) {
   case WM_PAINT:
     PaintDialogBackground(hdlg,ghinst_main);
     break;
   case WM_INITDIALOG:
     for(i=0;i<5;i++){
       QtLock[i].hbm=LoadBitmap(ghinst_main,
                                MAKEINTRESOURCE(QtLock[i].idBitmap));
       QtLock[i].hwnd=GetDlgItem(hdlg,QtLock[i].dlgID);
       if(QtLock[i].hwnd != NULL){
         QtLock[i].fpSclass=(WNDPROC)SetWindowLong(QtLock[i].hwnd,
                              GWL_WNDPROC,(LONG)LockSubclassProc);
         DoAddTooltip(ghwnd_main,QtLock[i].hwnd,QtLock[i].iTip);
       }
     }
     if(xpos == -1 && ypos == -1){
       GetWindowRect(ghwndTools,&rc);
       xpos=52; ypos=200;
     }
     SetWindowPos(hdlg,HWND_TOP,xpos,ypos,0,0,SWP_NOSIZE);
     return (TRUE);
   case WM_CTLCOLORSTATIC:
   case WM_CTLCOLORDLG:
     SetBkColor((HDC)wparam,RGB(192,192,192));
     return (LRESULT)GetStockObject(LTGRAY_BRUSH);
   case WM_DRAWITEM:{
       BOOL status;
       HBITMAP hbm;
       LPDRAWITEMSTRUCT lpdis;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->itemState & ODS_SELECTED)status=TRUE; else status=FALSE;
       if     (lpdis->CtlID == DLG_LOCK_SQ)hbm=QtLock[0].hbm;
       else if(lpdis->CtlID == DLG_LOCK_CI)hbm=QtLock[1].hbm;
       else if(lpdis->CtlID == DLG_LOCK_TR)hbm=QtLock[2].hbm;
       else if(lpdis->CtlID == DLG_LOCK_LK)hbm=QtLock[3].hbm;
       else if(lpdis->CtlID == DLG_LOCK_XX)hbm=QtLock[4].hbm;
       DrawQtBtn(status,lpdis->hDC,&(lpdis->rcItem),hbm);
     }
     return TRUE;
   case WM_DESTROY:
     for(i=0;i<5;i++){
       if(QtLock[i].hbm != NULL)DeleteObject(QtLock[i].hbm);
       if(QtLock[i].fpSclass != NULL)SetWindowLong(QtLock[i].hwnd,
         GWL_WNDPROC,(LONG)QtLock[i].fpSclass);
     }
     GetWindowRect(hdlg,&rc);
     xpos=rc.left; ypos=rc.top;
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDCANCEL:
       case IDOK:
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
         break;
       case DLG_LOCK_SQ:
         if(HIWORD(wparam) != BN_CLICKED)break;
         PostMessage(ghwnd_main,WM_COMMAND,IDM_LOCK_SQUARE,0);
         break;
       case DLG_LOCK_CI:
         if(HIWORD(wparam) != BN_CLICKED)break;
         PostMessage(ghwnd_main,WM_COMMAND,IDM_LOCK_CIRCLE,0);
         break;
       case DLG_LOCK_TR:
         if(HIWORD(wparam) != BN_CLICKED)break;
         PostMessage(ghwnd_main,WM_COMMAND,IDM_LOCK_TRIANGLE,0);
         break;
       case DLG_LOCK_LK:
         if(HIWORD(wparam) != BN_CLICKED)break;
         PostMessage(ghwnd_main,WM_COMMAND,IDM_LOCK_LOCK,0);
         break;
       case DLG_LOCK_XX:
         if(HIWORD(wparam) != BN_CLICKED)break;
         PostMessage(ghwnd_main,WM_COMMAND,IDM_STOP,0);
         break;
       default:   break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

static void DoAddTooltip(HWND hwndOwner, HWND tipon, int resid){
 TOOLINFO ti;
 if(ghwndTooltip == NULL || tipon == NULL)return;
 memset(&ti,0,sizeof(TOOLINFO));
 ti.cbSize = sizeof(TOOLINFO);
 ti.uFlags = TTF_IDISHWND;
 ti.hwnd = hwndOwner;
 ti.hinst = ghinst_main;
 ti.uId = (UINT)tipon;
 ti.lpszText = (LPSTR)resid;
 SendMessage(ghwndTooltip,TTM_ADDTOOL,0,(LPARAM) (LPTOOLINFO) &ti);
 return;
}

static HWND DoCreateTooltip(HWND hwndOwner){
 HWND hwndTT;
 hwndTT = CreateWindow(TOOLTIPS_CLASS, (LPSTR) NULL, TTS_ALWAYSTIP,
     CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
     hwndOwner, (HMENU) NULL, ghinst_main, NULL);
 return hwndTT;
}
