/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file TOOLS4.C  4th designer tools file */

#define MODULE_TOOLS4 1

#include "design.h"

typedef struct WORKFACE {
  long f;
  long fa[3];
  int  id;
} workface;

typedef struct VTXADJ {
  long nf;
  long *flist;
} vtxadj;

static int Nchange,Npieces;
static workface *WF;

static void xMakeSameOrientation(long last, long id){
 face *f;
 long i0,i1,i2,l0,l1,l2,a0,a1,a2;
 f=(MainFp+WF[last].f); l0=f->V[0]; l1=f->V[1]; l2=f->V[2];
 f=(MainFp+WF[id].f);   i0=f->V[0]; i1=f->V[1]; i2=f->V[2];
 a0=WF[id].fa[0]; a1=WF[id].fa[1]; a2=WF[id].fa[2];
 if((i2 == l2 && i1 == l1) ||
    (i2 == l1 && i1 == l0) ||
    (i2 == l0 && i1 == l2)){
   WF[id].fa[2]=a0; WF[id].fa[0]=a2;
   f->V[2]=i1; f->V[1]=i2; Nchange++;
   return;
 }
 if((i0 == l2 && i2 == l1) ||
    (i0 == l1 && i2 == l0) ||
    (i0 == l0 && i2 == l2)){
   WF[id].fa[0]=a1; WF[id].fa[1]=a0;
   f->V[0]=i2; f->V[2]=i0; Nchange++;
   return;
 }
 if((i1 == l1 && i0 == l0) ||
    (i1 == l2 && i0 == l1) ||
    (i1 == l0 && i0 == l2)){
   WF[id].fa[1]=a2; WF[id].fa[2]=a1;
   f->V[1]=i0; f->V[0]=i1; Nchange++;
   return;
 }
 return;
}

static void xSetConnectingID(long last,long id){
 if(WF[id].id == 1)return;
 WF[id].id=1;
 if(last >= 0)xMakeSameOrientation(last,id);
 last=id;
 if(WF[id].fa[0] >= 0)xSetConnectingID(last,WF[id].fa[0]);
 if(WF[id].fa[1] >= 0)xSetConnectingID(last,WF[id].fa[1]);
 if(WF[id].fa[2] >= 0)xSetConnectingID(last,WF[id].fa[2]);
}

static void xReorientate(workface *wf, long nf){
 long i,id;
 WF=wf;
 Nchange=0; Npieces=0;
 for(i=0;i<nf;i++)wf[i].id = -1;
 while(1){ /* keep going until all the bits have been processed */
   for(id=0;id<nf;id++){
     if(wf[id].id == -1)goto PROCESSIT;
   }
   break;      /* no bits have been found - so breakout */
   PROCESSIT:
   xSetConnectingID(-1,id);
   Npieces++;
 }
 return;
}

static void xFindAdjFaces(workface *wf, long nf){
 face *f;
 vertex *vp;
 vtxadj *vl=NULL;
 long c,nvl=0;
 long *fl,i,j,k,id,id0,id1,id2,fjd,fkd,n,m,kk,nl[3],ml[3];
 vp=MainVp; for(c=0;c<Nvert;c++){
   if(vp->id == -1){
     vp->id=nvl;
     nvl++;
   }
   else vp->id = -10;
   vp++;
 }
 if((vl=(vtxadj *)X__Malloc(nvl*sizeof(vtxadj))) == NULL)return;
 for(i=0;i<nvl;i++){
   vl[i].nf=0;
   vl[i].flist=NULL;
 }
 for(i=0;i<nf;i++){ /* make list of work faces attached to each vertex */
   f=(MainFp+wf[i].f);
   for(j=0;j<3;j++){
     id=(MainVp+(f->V[j]))->id;
     fl=vl[id].flist;
     n=vl[id].nf;
     if(fl == NULL){
       if((fl=(long *)X__Malloc(sizeof(long))) == NULL)return;
     }
     else{
       if((fl=(long *)X__Realloc(fl,(n+1)*sizeof(long))) == NULL)return;
     }
     fl[n]=i;
     vl[id].flist=fl;
     vl[id].nf=n+1;
   }
 }
 for(i=0;i<nf;i++){ /* find faces that are adjacent to face i */
   f=(MainFp+wf[i].f);
   id0=(MainVp+(f->V[0]))->id;
   id1=(MainVp+(f->V[1]))->id;
   id2=(MainVp+(f->V[2]))->id;
   nl[0]=id0; ml[0]=id1;
   nl[1]=id1; ml[1]=id2;
   nl[2]=id2; ml[2]=id0;
   for(kk=0;kk<3;kk++){
     wf[i].fa[kk] = -1;
     if((n=vl[nl[kk]].nf) > 0 && (m=vl[ml[kk]].nf) > 0){
       for(j=0;j<n;j++){
         fjd=vl[nl[kk]].flist[j];
         for(k=0;k<m;k++){
           fkd=vl[ml[kk]].flist[k];
           if(fkd == fjd && fkd != i){
             wf[i].fa[kk]=fkd;
             goto ENDF;
           }
         }
       }
     }
     ENDF:;
   }
 }
 for(i=0;i<nvl;i++){
   if(vl[i].flist != NULL)X__Free(vl[i].flist);
 }
 X__Free(vl);
 return;
}

void OrientateFaceNormals(long sas, long n_work_face, long *face_list){
 // sas = same orientation as face ID. If < 0 then make arbitrary.
 // then a lsit of Faces
 workface *w;
 long i,j,v0,v1,v2;
 w=NULL;
 if(sas >= 0){
  v0=(MainFp+sas)->V[0];
  v1=(MainFp+sas)->V[1];
  v2=(MainFp+sas)->V[2];
 }
 if((w=(workface *)X__Malloc(n_work_face*(long)sizeof(workface)))
       == NULL)return;
 for(i=0;i<Nvert;i++)(MainVp+i)->id = 0;
 for(j=0;j<n_work_face;j++){
   w[j].f=face_list[j];
   for(i=0;i<3;i++){
     w[j].fa[i] = -1;
     (MainVp+(MainFp+w[j].f)->V[i])->id = -1;
   }
 }
 xFindAdjFaces(w,n_work_face);
 xReorientate(w,n_work_face);
 if(sas >= 0){
   if(!IsFOSame(v0,v1,v2,
                (MainFp+face_list[0])->V[0],
                (MainFp+face_list[0])->V[1],
                (MainFp+face_list[0])->V[2]))
     ReverseFaceOrientation(n_work_face,face_list);
 }
 if(w != NULL)X__Free(w);
 return;
}

void ReverseFaceOrientation(long nf, long *face_list){
 long i,id;
 face *fp;
 for(i=0;i<nf;i++){
   fp=(MainFp+face_list[i]);
   id=fp->V[2];
   fp->V[2]=fp->V[0];
   fp->V[0]=id;
 }
 return;
}

BOOL IsFOSame(long s0, long s1, long s2,
                     long v0, long v1, long v2){
  vector u,v,n1,n2;
  VECSUB((double)(MainVp+s1)->xyz,
         (double)(MainVp+s0)->xyz,u)
  VECSUB((double)(MainVp+s2)->xyz,
         (double)(MainVp+s0)->xyz,v)
  CROSS(u,v,n1)
  VECSUB((double)(MainVp+v1)->xyz,
         (double)(MainVp+v0)->xyz,u)
  VECSUB((double)(MainVp+v2)->xyz,
         (double)(MainVp+v0)->xyz,v)
  CROSS(u,v,n2)
  if(DOT(n1,n2) < 0)return FALSE;
  return TRUE;
}

void OrientateSelectedFaces(void){
 face *fp;
 long i,face_count,*face_list;
 face_count=0; face_list=NULL;
 if(Nface == 0)return;
 for(i=0,fp=MainFp;i<Nface;i++,fp++){
   if((MainVp+fp->V[0])->status == SELECTED ||
      (MainVp+fp->V[1])->status == SELECTED ||
      (MainVp+fp->V[2])->status == SELECTED)face_count++;
 }
 face_list = (long *)X__Malloc((face_count)*sizeof(long));
 if(face_list == NULL)return;
 for(i=0,face_count=0,fp=MainFp;i<Nface;i++,fp++){
   if((MainVp+fp->V[0])->status == SELECTED ||
      (MainVp+fp->V[1])->status == SELECTED ||
      (MainVp+fp->V[2])->status == SELECTED){
     face_list[face_count]=i;
     face_count++;
   }
 }
 if(face_count > 0){
   OrientateFaceNormals(0,face_count,face_list);
 }
 X__Free(face_list);
}

