/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  TWIST.C  */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#include "struct.h"
#include "dstruct.h"

#include "twist.h"

#define DESELECTED 0
#define SELECTED   1
#define FAIL      -1
#define OK         1

#define POINT2VECTOR(p,v) v[0]=(double)p[0]; v[1]=(double)p[1]; \
                          v[2]=(double)p[2];
#define VECTOR2POINT(v,p) p[0]=(long)v[0]; p[1]=(long)v[1]; \
                          p[2]=(long)v[2];
#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                         }

static HWND      hParent;
static HINSTANCE hThisInstance;

static BOOL CALLBACK TwistAlongDlgProc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam);

void MultMV(double t4[4][4], long x, long y, long z,
           long *xx, long *yy, long *zz){
 *xx=(long)(t4[0][0]*(double)x+t4[0][1]*(double)y+t4[0][2]*(double)z+t4[0][3]);
 *yy=(long)(t4[1][0]*(double)x+t4[1][1]*(double)y+t4[1][2]*(double)z+t4[1][3]);
 *zz=(long)(t4[2][0]*(double)x+t4[2][1]*(double)y+t4[2][2]*(double)z+t4[2][3]);
 return;
}

void CopyMM(double tin[4][4], double tout[4][4]){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 tout[i][j]=tin[i][j];
}

void MultMM(double t1[4][4], double t2[4][4], double tr[4][4]){
 short n=4,i,j,k;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  for(k=0;k<4;k++)tr[i][j]=tr[i][j]+t1[i][k]*t2[k][j];
 }
 return;
}

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=hDLL;
      if(hDLL == NULL)MessageBox ( GetFocus()," NULL instance",NULL, MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 BOOL Twist,Xtrude=TRUE;
 struct np {int nstep; double repeat_map,angle; BOOL Xtrude;} data;
 long vp,vf,*va,LastV;
 double t[4][4],tr[4][4],ts[4][4];
 double l,ll,ls,dmap,ddmap,theta,alpha,dz,repeat_map=4.0,twist=360.0;
 long i,j,Npath,Dn,Nstep=16;
 vertex *Vp,*Va,*Va1,*Va2;
 vector dr,drr,n;
 point pc,pn;
 HCURSOR hSave;
 lpEVI=lpevi;
 hParent=parent_window;
 if(NvertSelect < 1 || NvertSelect > 32000)return FALSE;
 if((vf=get_closest_vertex()) < 0)return FALSE;
 data.repeat_map=repeat_map;
 data.angle=twist;
 data.nstep=Nstep;
 data.Xtrude=Xtrude;
 if(DialogBoxParam(hThisInstance,MAKEINTRESOURCE(DLG_TWIST),
     hParent,(DLGPROC)TwistAlongDlgProc,(LPARAM)&data) == FAIL)return FALSE;
 repeat_map=data.repeat_map;
 twist=data.angle;
 Xtrude=data.Xtrude;
 Nstep=data.nstep; if(Nstep < 1)return FALSE;
 if((va=(long *)X__Malloc(NvertSelect*sizeof(long))) == NULL){
   return FALSE;
 }
 hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
 Npath=0; va[0]=vf; (MainVp+va[0])->id=1;
 (MainVp+vf)->status=DESELECTED; NvertSelect--; NvertDeselect++;
 vp=0; while(vp < Nvert){
   Vp=(MainVp+vp);
   if(Vp->status == SELECTED && Vp->id == 0 && Connected(va[Npath],vp)){
     Npath++;
     va[Npath]=vp;
     Vp->id=1;
     Vp->status=DESELECTED;
     NvertSelect--; NvertDeselect++;
     vp=0;
   }
   else vp++;
 }
 if(Npath > 0){ /* at least 2 points needed */
   l=0.0; ll=0.0; /* get length of path */
   for(i=1;i<=Npath;i++){
     Va=(MainVp+va[i]); Va1=(MainVp+va[i-1]);
     l += sqrt(
         (double)(Va->xyz[0] - Va1->xyz[0])*
         (double)(Va->xyz[0] - Va1->xyz[0])+
         (double)(Va->xyz[1] - Va1->xyz[1])*
         (double)(Va->xyz[1] - Va1->xyz[1])+
         (double)(Va->xyz[2] - Va1->xyz[2])*
         (double)(Va->xyz[2] - Va1->xyz[2]));
   }
   for(i=1;i<=Npath;i++){ /* extrude along each path segment */
//   for(i=1;i<=1;i++){ /* extrude along each path segment */
     Va=(MainVp+va[i]); Va1=(MainVp+va[i-1]);
     VECSUB((double)Va->xyz,(double)Va1->xyz,dr)
     if(oNormalize(dr) == FAIL)goto EXIT;
     ls = sqrt(
         (double)(Va->xyz[0] - Va1->xyz[0])*
         (double)(Va->xyz[0] - Va1->xyz[0])+
         (double)(Va->xyz[1] - Va1->xyz[1])*
         (double)(Va->xyz[1] - Va1->xyz[1])+
         (double)(Va->xyz[2] - Va1->xyz[2])*
         (double)(Va->xyz[2] - Va1->xyz[2]));
     ll += ls;
     dmap=ll/l*repeat_map;
     theta=twist/Nstep; dz=ls/Nstep; ddmap=dmap/Nstep; dmap=ddmap;
     Twist=FALSE;
     if(i < Npath){
       Va2=(MainVp+va[i+1]);
       VECSUB((double)Va2->xyz,(double)Va->xyz,drr)
       if(oNormalize(drr) != FAIL){
         if((alpha=fabs(DOT(dr,drr))) < 9.8){ /* bend in path */
           CROSS(dr,drr,n)
           VECSCALE((double)TVsizeZ,n,n)
           alpha = acos(alpha);
           alpha *= 180.0/3.14159;
           if(alpha > 2.0)Twist=TRUE;
           alpha /= Nstep;
         }
       }
     }
     LastV=Nvert;
     for(j=0;j<Nstep;j++){
       if(Xtrude){
         if(!CreateAttachedCopy(1))goto EXIT;
       }
       else{
         if(!AddCopy(0))goto EXIT;
       }
     }
     Va=(MainVp+va[i]); Va1=(MainVp+va[i-1]);
     Dn=(Nvert-LastV)/Nstep;
     VECCOPY(Va1->xyz,pc)
     VECSUM(pc,(long)n,pn)
     for(j=0;j<Nstep;j++){ /* twist each step in segment  */
       ArbitraryRotate(theta*(j+1),Va1->xyz,Va->xyz,tr);
       if(Twist){
         ArbitraryRotate(alpha*(j+1),pc,pn,ts);
         MultMM(ts,tr,t);
       }
       else CopyMM(tr,t);
       for(vp=LastV;vp<LastV+Dn;vp++){ /* move new points into position */
         Vp=(MainVp+vp);
         if(Vp->gp > 0){
           Vp->y = dmap;
         }
         MultMV(t,Vp->xyz[0],Vp->xyz[1],Vp->xyz[2],
                &Vp->xyz[0],&Vp->xyz[1],&Vp->xyz[2]);
         Vp->xyz[0] += (dz*(j+1))*dr[0];  // shift in direction of edge
         Vp->xyz[1] += (dz*(j+1))*dr[1];
         Vp->xyz[2] += (dz*(j+1))*dr[2];
       }
       LastV += Dn;
       dmap += ddmap;
     }
   }
   /* set curve now as selected  everything else deselected*/
   vp=0; while(vp < Nvert){
     Vp=(MainVp+vp);
     if(Vp->status == SELECTED){
       Vp->status=DESELECTED; NvertSelect--; NvertDeselect++;
     }
     vp++;
   }
   for(i=0;i<=Npath;i++)if((MainVp+va[i])->status == DESELECTED){
     (MainVp+va[i])->status=SELECTED;
     NvertSelect++; NvertDeselect--;
   }
 }
 EXIT:
 X__Free(va);
 SetCursor(hSave);
 return TRUE;
}

static BOOL CALLBACK TwistAlongDlgProc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam){
 char tempstr[16];
 BOOL err;
 int i;
 struct np {int nstep; double repeat_map,angle; BOOL Xtrude;} *dp;
 switch( msg ) {
   case WM_INITDIALOG:{
       SetWindowLong(hwnd,GWL_USERDATA,lparam);
       dp=(struct np *)lparam;
       SetDlgItemInt(hwnd,DLG_TWIST_STEP,dp->nstep,FALSE);
       sprintf(tempstr,"%.0lf",dp->angle);
       SendDlgItemMessage(hwnd,DLG_TWIST_ANGLE,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_TWIST_ANGLE,EM_LIMITTEXT,(WPARAM)12,0);
       sprintf(tempstr,"%.4lf",dp->repeat_map);
       SendDlgItemMessage(hwnd,DLG_TWIST_MAP,WM_SETTEXT,0,(LPARAM)tempstr);
       SendDlgItemMessage(hwnd,DLG_TWIST_MAP,EM_LIMITTEXT,(WPARAM)12,0);
       if(dp->Xtrude)
         SendDlgItemMessage(hwnd,DLG_TWIST_EXTRUDE,BM_SETCHECK,TRUE,0);
       else
         SendDlgItemMessage(hwnd,DLG_TWIST_COPY,BM_SETCHECK,TRUE,0);
       CentreDlgOnS(hwnd);
     }
     return (TRUE);
   case WM_COMMAND:
      dp=(struct np *)GetWindowLong(hwnd,GWL_USERDATA);
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FAIL);
          return(TRUE);
        case IDOK:
          dp->nstep=GetDlgItemInt(hwnd,DLG_TWIST_STEP,&err,FALSE);
          if(GetDlgItemText(hwnd,DLG_TWIST_ANGLE,tempstr,10) == 0){
            EndDialog(hwnd, FAIL);
            return(TRUE);
          }
          if((dp->angle=atof(tempstr)) == 0){
            EndDialog(hwnd, FAIL);
            return(TRUE);
          }
          if(GetDlgItemText(hwnd,DLG_TWIST_MAP,tempstr,10) == 0){
            EndDialog(hwnd, FAIL);
            return(TRUE);
          }
          if((dp->repeat_map=atof(tempstr)) == 0){
            EndDialog(hwnd, FAIL);
            return(TRUE);
          }
          if(SendDlgItemMessage(hwnd,DLG_TWIST_EXTRUDE,BM_GETCHECK,0,0))
               dp->Xtrude=TRUE;
          else dp->Xtrude=FALSE;
          EndDialog(hwnd,1);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}
