/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  VTK.CPP  */

#include "vtk.hh"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>

#define UNIT         32768L
#define DESELECTED       0
#define SELECTED         1
#define INEDITOR         6

#ifdef __cplusplus
extern "C" {
#endif

#include "struct.h"
#include "dstruct.h"

#ifdef __cplusplus
}
#endif

#include "vtk.h"

#include "sfx2vtk.h"
#include "sfx2vtk.cpp"

static HWND      hParent;
static HINSTANCE hThisInstance;

static BOOL CALLBACK SfxDlgProc(HWND hwnd, UINT msg,
                                WPARAM wparam, LPARAM lparam);

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      hThisInstance=(HINSTANCE)hDLL;
      if(hDLL == NULL)MessageBeep(MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
      break;
  }
  return TRUE;
}


extern "C" BOOL _Xmodeler
 (HWND parent_window,HWND info_window,X__STRUCTURE *lpevi){
 HCURSOR hSave;
 static struct bp {
  float f;
 } d;
 lpEVI=lpevi;
 hParent=parent_window;
 d.f=0.8;
 if(DialogBoxParam(hThisInstance,MAKEINTRESOURCE(DLG_VTK),hParent,
             (DLGPROC)SfxDlgProc,(LPARAM)&d) == FALSE)return FALSE;
 hSave=SetCursor(LoadCursor(NULL,IDC_WAIT));
 {
  SfxReader sfxinput;
  SfxWriter sfxoutput;

  SfxPointReader sfxpointinput;
  SfxEdgeReader sfxedgeinput;
  vtkRenderMaster rm;
  vtkRenderWindow *renWin;
  vtkRenderer *aren;
  vtkActor    *actor1, *actor2;
  vtkCubeSource *cube1,*cube2;
  vtkAppendPolyData *cube;
  vtkConeSource *cone;
  vtkLineSource *line;
  vtkPointSource *pts;
  vtkImplicitModeller *imp;
  vtkContourFilter *cf;
  vtkPolyMapper *mapper, *omapper;
  vtkOutlineFilter *outline;
  vtkRenderWindowInteractor *iren;

//  renWin  = rm.MakeRenderWindow();
//  iren = renWin->MakeRenderWindowInteractor();
//  aren    = renWin->MakeRenderer();
//  pts = new vtkPointSource;
//  pts->SetNumberOfPoints(10);

//line = new vtkLineSource;
//line->SetResolution(1);

//  cone = new vtkConeSource;
//    cone->SetResolution(8);

  cube1 = new vtkCubeSource(1,2,3);
  cube2 = new vtkCubeSource(2,3,4);
  cube2->SetCenter(3,3,3);

  cube = new vtkAppendPolyData;
  cube->AddInput(cube1->GetOutput());
  cube->AddInput(cube2->GetOutput());

  imp = new vtkImplicitModeller;
//    imp->SetInput(sfxinput.GetOutput());
//    imp->SetInput(sfxpointinput.GetOutput());
    imp->SetInput(sfxedgeinput.GetOutput());
//    imp->SetInput(cube1->GetOutput());
    imp->SetSampleDimensions(25,25,25);
    imp->SetMaximumDistance(0.1);
//    imp->SetMaximumDistance(0.2);

  cf = new vtkContourFilter;
//    cf->SetInput(imp->GetOutput());
//    cf->SetValue(0, 0.1);
////    cf->SetValue(0, 0.2);

    vtkSphere *sph1 = new vtkSphere;
    sph1->SetRadius(1.0);
    vtkSphere *sph2 = new vtkSphere;
    sph2->SetRadius(0.5);
    sph2->SetCenter(0.0,0.0,0.5);
    vtkImplicitBoolean *ibool = new vtkImplicitBoolean;
    ibool->AddFunction(sph1);
    ibool->AddFunction(sph2);
    ibool->SetOperationType(VTK_DIFFERENCE);
    vtkSampleFunction *sam = new vtkSampleFunction;
    sam->SetSampleDimensions(25,25,25);
    sam->SetImplicitFunction(ibool);
//    sam->SetImplicitFunction(sph1);
    cf->SetInput(sam->GetOutput());
    cf->SetValue(0, 0.0);

//  mapper = new vtkPolyMapper;
//    mapper->SetInput(cf->GetOutput());

//  actor1 = new vtkActor;
//    actor1->SetMapper(mapper);
//    actor1->GetProperty()->SetColor(0.8,1.0,0.9);

  // draw an outline
//  outline = new vtkOutlineFilter;
//    outline->SetInput(imp->GetOutput());

//  omapper = new vtkPolyMapper;
//    omapper->SetInput(outline->GetOutput());

//  actor2 = new vtkActor;
//    actor2->SetMapper(omapper);
//    actor2->GetProperty()->SetColor(1,1,1);

//  aren->AddActors(actor2);
//  aren->AddActors(actor1);
//  aren->SetBackground(0.2,0.2,0.2);

//  renWin->Render();

  // interact with data
//  iren->Start();

//  vtkShrinkPolyData shrink;
//  shrink.SetInput(sfxinput.GetOutput());
//  shrink.SetShrinkFactor(d.f);

  vtkCleanPolyData *cp = new vtkCleanPolyData;
  cp->SetTolerance(0.0001);
  cp->SetInput(cf->GetOutput());

  vtkTriangleFilter *tf = new vtkTriangleFilter;
  tf->SetInput(cp->GetOutput());

  sfxoutput.SetInput(tf->GetOutput());
  sfxoutput.Write();

  delete tf;
  delete cp;
  delete cf;
  delete imp;

 }
 SetCursor(hSave);
 return TRUE;
}

static BOOL CALLBACK SfxDlgProc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam){
 BOOL err;
 double f;
 int i;
 static struct bp {
   float f;
 } *dp;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     dp=(struct bp *)lparam;
     CentreDlgOnS(hwnd);
     return (TRUE);
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

