/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* DISSOLVE.C  */

#include <math.h>
#include <windows.h>
#include <commctrl.h>

#include "dissolve.h"

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#ifndef PI
#define PI 3.1415926
#endif

int version=1;

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL; /* use to pick up resources from DLL   */

#include "pstruct.h"

#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(LibMain)
#endif

long _RenderGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, vertex *Vlist){
 return 1;
}

void _PreviewGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, Svertex *Vlist){
 return;
}

static long GenRan(long Nf,face *f){
 long j;
 do
  j=(long)((double)Nf*(double)rand()/32767.0);
 while ((f+j)->mattr > 0);
 return j;
}

long _RenderExternalEffect(char *PrmList, sfxinfo *SFXinfo,
                                 vertex *Vlist){
 long i,type,inout,iend,iedge,N,Nglass;
 face *f,*fg;
 char buffer[128];
 unsigned char tv;
 double t;
 sscanf(PrmList,"%s %d %ld %ld %ld %ld",buffer,&version,
        &type,&inout,&iend,&iedge);
 t=(double)SFXinfo->time;
 N=SFXinfo->Nfaces;
 if     (inout == 2)t=(1.0-t);
 else if(inout == 3)t=1.0;
 if(type == 1){
   if(iend)tv=192;
   else    tv=240;
   f=(face *)SFXinfo->Faces;
   for(i=0;i<N;i++){f->mattr=0; f++;}
 }
 else{
   if(iend)tv=(unsigned char)(t * 192.0);
   else    tv=(unsigned char)(t * 240.0);
 }
 if(N > 0 && SFXinfo->Faces != NULL){
   f=(face *)SFXinfo->Faces;
   if(type == 1){ // random
     Nglass=(long)((double)N*t);
     if(Nglass > 0){
       for(i=0;i<Nglass;i++){
         if(Nglass > N - 3)fg=(f+i);
         else fg=(f+GenRan(N,f));
         fg->mattr = tv;
         if(iedge)fg->fattrib |= 0x01;  // edge
          else{
            fg->fattrib &= 0xfe;         // no edge
            fg->fattrib &= 0xdf;         // no gloss
         }
       }
     }
   }
   else for(i=0;i<N;i++){
     f->mattr = tv;
     if(iedge)f->fattrib |= 0x01;  // edge
     else{
       f->fattrib &= 0xfe;         // no edge
       f->fattrib &= 0xdf;         // no gloss
     }
     f++;
   }
 }
 return 1;
}

void _PreviewExternalEffect(char *PrmList, sfxinfo *SFXinfo,
                                    Svertex *Vlist){
 return;
}

BOOL CALLBACK EffectDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static long type=0,inout=1,iend=0,iedge=0;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
 char buffer[128];
 if(Op != NULL){  /* parameters exist so read them off the list */
   sscanf(Op,"%s %ld %ld %ld %ld %ld",buffer,&version,
          &type,&inout,&iend,&iedge);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_DISSOLVE),hWnd,
              (DLGPROC)EffectDlgProc) == FALSE)return Op;
   if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(buffer,"%s %ld %ld %ld %ld %ld",name,version,type,inout,iend,iedge);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"External effect: Out of memory",NULL,
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK EffectDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 int err;
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     if     (inout == 1)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_OUT,BM_SETCHECK,TRUE,0);
     else if(inout == 2)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_IN,BM_SETCHECK,TRUE,0);
     else if(inout == 3)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_HOLD,BM_SETCHECK,TRUE,0);
     if     (type == 0)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_FADE,BM_SETCHECK,TRUE,0);
     else if(type == 1)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_RANDOM,BM_SETCHECK,TRUE,0);
     if     (iedge == 1)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_EDGE,BM_SETCHECK,TRUE,0);
     if     (iend == 0)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_100,BM_SETCHECK,TRUE,0);
     else if(iend == 1)
       SendDlgItemMessage(hwnd,DLG_DISSOLVE_90,BM_SETCHECK,TRUE,0);
     LoadAnimatedClip(hwnd);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hwnd);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if     (SendDlgItemMessage(hwnd,DLG_DISSOLVE_OUT,BM_GETCHECK,0,0))
            inout=1;
          else if(SendDlgItemMessage(hwnd,DLG_DISSOLVE_IN,BM_GETCHECK,0,0))
            inout=2;
          else if(SendDlgItemMessage(hwnd,DLG_DISSOLVE_HOLD,BM_GETCHECK,0,0))
            inout=3;
          if     (SendDlgItemMessage(hwnd,DLG_DISSOLVE_FADE,BM_GETCHECK,0,0))
            type=0;
          else if(SendDlgItemMessage(hwnd,DLG_DISSOLVE_RANDOM,BM_GETCHECK,0,0))
            type=1;
          if     (SendDlgItemMessage(hwnd,DLG_DISSOLVE_EDGE,BM_GETCHECK,0,0))
               iedge=1;
          else iedge=0;
          if     (SendDlgItemMessage(hwnd,DLG_DISSOLVE_90,BM_GETCHECK,0,0))
               iend=1;
          else iend=0;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

