/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* FALL.C external animation effect  (Windows NT version)                  */

#include <math.h>
#include <windows.h>
#include <commctrl.h>

#include "fall.h"

#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

#ifndef PI
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "pstruct.h"

#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

static double CalculatePosition(long n, double zm, double z0,
                                double t,
                                double mu, double a, double d){
 int i;
 double dfall,ta,dt,h,deltaT;
 dfall=z0-(zm-d);   /* must be > 0 */
 ta=sqrt(2.0*dfall/a);
 if(t < ta)return (z0-a*t*t*0.5);
 dt=mu*ta; h=dfall;
 if(n > 0)for(i=0;i<3*n;i++){
   h=mu*mu*h;
   if(t < ta+dt){
     deltaT = dt - (t - ta);
     return (z0-(dfall-h+a*deltaT*deltaT*0.5));
   }
   ta += dt;
   if(t < ta+dt){
     deltaT = (t - ta);
     return (z0-(dfall-h+a*deltaT*deltaT*0.5));
   }
   ta += dt;
   dt *= mu;
 }
 return (z0-dfall);
}

static void  GetTandA(int nb, double mu, double d, double *t, double *a){
 int i;
 double t0;
 if(nb > 0){
   t0=mu;
   for(i=0;i<nb-1;i++)t0 = mu*(t0 + 1.0);
   t0 *= 2.0;
   t0 += 1.0;
 }
 else t0=1.0;
 t0  = 1.0/t0;
 *a=2.0*d/(t0*t0);
 *t=t0;
 return;
}

long _RenderGlobalEffect
(char *PrmList, sfxinfo *SFXinfo,sfxdata *SFXdata, vertex *Vlist){
 long i,nb,v,local,gnd;
 double z,zmax=-1.e20,zmin=1.e20,mu,d,t,t0,a;  /* mu is elasticity */
 char buffer[128];
#include "pro_key.c"
 sscanf(PrmList,"%s %ld %ld %ld %ld %f %f",buffer,&nb,&v,&local,&gnd,&mu,&d);
 if(local == 1)return 1;
 t=(double)SFXinfo->time;
 for(i=0;i<SFXinfo->nvert;i++){
   if(Vlist[i].p[2] > zmax)zmax=Vlist[i].p[2];
   if(Vlist[i].p[2] < zmin)zmin=Vlist[i].p[2];
 }
 if(SFXdata->ground && gnd == 1){
   if(v == 1 || v == 3)d=zmin-(double)SFXdata->ground_pos[2];
   else                d=zmax-(double)SFXdata->ground_pos[2]-10.0;
 }
 GetTandA(nb,mu,d,&t0,&a);
 if(v > 1){ /* final */
  v -= 2;
  t=1.0;
 }
 if(v == 1){
   z=(double)CalculatePosition(nb,zmax,zmax,t,mu,a,d);
   z=zmax-z;
   for(i=0;i<SFXinfo->nvert;i++)Vlist[i].p[2] -= z;
 }
 else{
   for(i=0;i<SFXinfo->nvert;i++){
     Vlist[i].p[2]=(double)CalculatePosition(nb,zmax,(double)Vlist[i].p[2],
                                         t,mu,a,d);
   }
 }
 return 1;
}

long _RenderExternalEffect
(char *PrmList, sfxinfo *SFXinfo,vertex *Vlist){
 long i,nb,v,local,gnd;
 double z,zmax=-1.e20,mu,d,t,t0,a;  /* mu is elasticity */
 char buffer[128];
#include "pro_key.c"
 sscanf(PrmList,"%s %ld %ld %ld %ld %f %f",buffer,&nb,&v,&local,&gnd,&mu,&d);
 if(local == 0)return 1;
 t=(double)SFXinfo->time;
 GetTandA(nb,mu,d,&t0,&a);
 for(i=0;i<SFXinfo->nvert;i++)if(Vlist[i].p[2] > zmax)zmax=Vlist[i].p[2];
 if(v > 1){ /* final */
  v -= 2;
  t=1.0;
 }
 if(v == 1){
   z=(double)CalculatePosition(nb,zmax,zmax,t,mu,a,d);
   z=zmax-z;
   for(i=0;i<SFXinfo->nvert;i++)Vlist[i].p[2] -= z;
 }
 else{
   for(i=0;i<SFXinfo->nvert;i++){
     Vlist[i].p[2]=(double)CalculatePosition(nb,zmax,(double)Vlist[i].p[2],
                                         t,mu,a,d);
   }
 }
 return 1;
}

void _PreviewGlobalEffect
(char *PrmList, sfxinfo *SFXinfo,sfxdata *SFXdata, Svertex *Vlist){
 long i,nb,v,local,gnd;
 double z,zmax=-1.e20,zmin=1.e20,mu,d,t,t0,a;  /* mu is elasticity */
 char buffer[128];
#include "pro_key1.c"
 sscanf(PrmList,"%s %ld %ld %ld %ld %f %f",buffer,&nb,&v,&local,&gnd,&mu,&d);
 if(local == 1)return;
 t=(double)SFXinfo->time;
 for(i=0;i<SFXinfo->nvert;i++){
   if(Vlist[i][2] > zmax)zmax=Vlist[i][2];
   if(Vlist[i][2] < zmin)zmin=Vlist[i][2];
 }
 if(SFXdata->ground && gnd == 1){
   if(v == 1 || v == 3)d=zmin-(double)SFXdata->ground_pos[2];
   else                d=zmax-(double)SFXdata->ground_pos[2]-10.0;
 }
 GetTandA(nb,mu,d,&t0,&a);
 if(v > 1){ /* final */
  v -= 2;
  t=1.0;
 }
 if(v == 1){
   z=(double)CalculatePosition(nb,zmax,zmax,t,mu,a,d);
   z=zmax-z;
   for(i=0;i<SFXinfo->nvert;i++)Vlist[i][2] -= z;
 }
 else{
   for(i=0;i<SFXinfo->nvert;i++){
     Vlist[i][2]=(double)CalculatePosition(nb,zmax,(double)Vlist[i][2],
                                         t,mu,a,d);
   }
 }
 return;
}

void _PreviewExternalEffect
(char *PrmList, sfxinfo *SFXinfo,Svertex *Vlist){
 long i,nb,v,local,gnd;
 double z,zmax=-1.e20,mu,d,t,t0,a;  /* mu is elasticity */
 char buffer[128];
#include "pro_key1.c"
 sscanf(PrmList,"%s %ld %ld %ld %ld %f %f",buffer,&nb,&v,&local,&gnd,&mu,&d);
 if(local == 0)return;
 t=(double)SFXinfo->time;
 GetTandA(nb,mu,d,&t0,&a);
 for(i=0;i<SFXinfo->nvert;i++)if(Vlist[i][2] > zmax)zmax=Vlist[i][2];
 if(v > 1){ /* final */
  v -= 2;
  t=1.0;
 }
 if(v == 1){
   z=(double)CalculatePosition(nb,zmax,zmax,t,mu,a,d);
   z=zmax-z;
   for(i=0;i<SFXinfo->nvert;i++)Vlist[i][2] -= z;
 }
 else{
   for(i=0;i<SFXinfo->nvert;i++){
     Vlist[i][2]=(double)CalculatePosition(nb,zmax,(double)Vlist[i][2],
                                         t,mu,a,d);
   }
 }
 return;
}

BOOL CALLBACK EffectDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static long nbounce=2,vvv=0;
static double elasticity=0.4,distance=1.0;
static long local=0,gnd=1;

char * _SetExternalParameters
 (char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
 char buffer[128];
 double d1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   sscanf(Op,"%s %ld %ld %ld %ld %f %f",buffer,&nbounce,&vvv,&local,&gnd,&elasticity,&d1);
   distance=d1/(double)ruler;
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_EFFECT),hWnd,
              (DLGPROC)EffectDlgProc) == FALSE)return Op;
   if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 d1=distance*(double)ruler;
 sprintf(buffer,"%s %ld %ld %ld %ld %f %f",name,nbounce,vvv,local,gnd,elasticity,d1);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
   MessageBox(GetFocus(),"Memory",NULL,MB_OK|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK EffectDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[32];
 int err;
 switch( msg ) {
   case WM_INITDIALOG:
     if(vvv >  1){
       SendDlgItemMessage(hwnd,DLG_EFFECT_FFF,BM_SETCHECK,TRUE,0);
       vvv -= 2;
     }
     if(vvv == 0)
       SendDlgItemMessage(hwnd,DLG_EFFECT_FALL,BM_SETCHECK,TRUE,0);
     else
       SendDlgItemMessage(hwnd,DLG_EFFECT_DROP,BM_SETCHECK,TRUE,0);
     if(local == 1)
       SendDlgItemMessage(hwnd,DLG_EFFECT_LOCAL,BM_SETCHECK,TRUE,0);
     else
       SendDlgItemMessage(hwnd,DLG_EFFECT_GLOBAL,BM_SETCHECK,TRUE,0);
     if(gnd == 1)
       SendDlgItemMessage(hwnd,DLG_EFFECT_GROUND,BM_SETCHECK,TRUE,0);
     SetDlgItemInt(hwnd,DLG_EFFECT_BOUNCE,nbounce,FALSE);
     sprintf(str,"%.2f",elasticity);
     SetDlgItemText(hwnd,DLG_EFFECT_ELASTICITY,str);
     sprintf(str,"%.2f",distance);
     SetDlgItemText(hwnd,DLG_EFFECT_DISTANCE,str);
     LoadAnimatedClip(hwnd);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hwnd);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         EndDialog(hwnd,FALSE);
         return(TRUE);
       default:
         break;
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(SendDlgItemMessage(hwnd,DLG_EFFECT_DROP,BM_GETCHECK,0,0))
               vvv=1;
          else vvv=0;
          if(SendDlgItemMessage(hwnd,DLG_EFFECT_FFF,BM_GETCHECK,0,0))
               vvv += 2;
          if(SendDlgItemMessage(hwnd,DLG_EFFECT_LOCAL,BM_GETCHECK,0,0))
               local=1;
          else local=0;
          if(SendDlgItemMessage(hwnd,DLG_EFFECT_GROUND,BM_GETCHECK,0,0))
               gnd=1;
          else gnd=0;
          if(GetDlgItemText(hwnd,DLG_EFFECT_ELASTICITY,str,12) != 0)
            elasticity=atof(str);
          if(GetDlgItemText(hwnd,DLG_EFFECT_DISTANCE,str,12) != 0)
            distance=atof(str);
          nbounce=(int)GetDlgItemInt(hwnd,DLG_EFFECT_BOUNCE,&err,TRUE);
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

