/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* flap.c */

#include <math.h>
#include <windows.h>
#include <commctrl.h>

#include "flap.h"

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#ifndef PI
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "pstruct.h"

#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
 switch (dwReason) {
   case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
     hDLLinstance = hDLL;  /* handle to DLL file */
     break;
   case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
     break;
 }
 return (int)TRUE;
}

#if __SC__
#pragma startaddress(LibMain)
#endif

long _RenderGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, vertex *Vlist){
 return 1;
}

void _PreviewGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, Svertex *Vlist){
 return;
}

long _RenderExternalEffect(char *parameters, sfxinfo *SFXinfo,
                                 vertex *v){
 double amplitude,flaplength,displacement,PRMamp,PRMuamp,PRMlngth,PRMfreq,
        d,maxd,scale;
 char buffer[256];
 int type, PRMaxis, axis, fade, paxis, paxis2;
 long i;
 sscanf(parameters,"%s %d %d %f %f %f %f %d",buffer,&type,&PRMaxis,
        &PRMuamp,&PRMamp,
        &PRMlngth,&PRMfreq,&fade);
//MessageBox(NULL,parameters,NULL,MB_OK);
 if     (PRMaxis == 3) axis=1;    /* F / B */
 else if(PRMaxis == 1) axis=2;    /* U / D */
 else if(PRMaxis == 2) axis=0;    /* L / R */
 if     (axis == 0){ paxis = 1; paxis2 = 2; }
 else if(axis == 2){ paxis = 1; paxis2 = 0; }
 else if(axis == 1){ paxis = 0; paxis2 = 2; }
 flaplength=max((SFXinfo->vmax[paxis]-SFXinfo->vmin[paxis]) * PRMlngth,1.0);
 amplitude=(PRMamp/25.0) * (flaplength/8);
 if(type == 2){                        /* Radial */
  maxd=0.00;
  for(i=0;i<SFXinfo->nvert;i++){
   d=sqrt(v[i].p[paxis]*v[i].p[paxis] + v[i].p[paxis2]*v[i].p[paxis2]);
   maxd=max(maxd,d);
  }
 }
 for(i=0;i<SFXinfo->nvert;i++){
  if(type == 1) d=v[i].p[paxis];       /* Linear */
  else if(type == 2){                  /* Radial */
   d=sqrt(v[i].p[paxis]*v[i].p[paxis] + v[i].p[paxis2]*v[i].p[paxis2]);
  }
  displacement=amplitude*cos(PI*2*((SFXinfo->time*PRMfreq)-d/flaplength));
  if(type == 1){                       /* Linear */
   d = v[i].p[paxis]-SFXinfo->vmin[paxis];
   maxd = SFXinfo->vmax[paxis]-SFXinfo->vmin[paxis];
  }
  if(fade == 2) displacement *= d / maxd;             /* fade up */
  else if(fade == 3) displacement *= 1 - d / maxd;    /* fade down */
  v[i].p[axis] += displacement;

  scale=PRMuamp*cos(PI*2*((SFXinfo->time*PRMfreq)-d/flaplength));
  if(fade == 2) scale *= d / maxd;
  else if(fade == 3) scale *= 1 - d / maxd;
  scale = 1.0 + scale;
  v[i].p[paxis2] *= scale;
 }
 return 1;
}

void _PreviewExternalEffect(char *parameters, sfxinfo *SFXinfo,
                                    Svertex *v){
 double amplitude,flaplength,displacement,PRMamp,PRMuamp,PRMlngth,PRMfreq,
        d,maxd,scale;
 char buffer[256];
 int type, PRMaxis, axis, fade, paxis, paxis2;
 long i;
 sscanf(parameters,"%s %d %d %f %f %f %f %d",buffer,&type,&PRMaxis,
        &PRMuamp,&PRMamp,
        &PRMlngth,&PRMfreq,&fade);
 if(PRMaxis == 3) axis=1;         /* F / B */
 else if(PRMaxis == 1) axis=2;    /* U / D */
 else if(PRMaxis == 2) axis=0;    /* L / R */
 if     (axis == 0){ paxis = 1; paxis2 = 2; }
 else if(axis == 2){ paxis = 1; paxis2 = 0; }
 else if(axis == 1){ paxis = 0; paxis2 = 2; }
 flaplength=max((SFXinfo->vmax[paxis]-SFXinfo->vmin[paxis]) * PRMlngth,1.0);
 amplitude=(PRMamp/25.0) * (flaplength/8);
 if(type == 2){                        /* Radial */
  maxd=0.00;
  for(i=0;i<SFXinfo->nvert;i++){
   d=sqrt(v[i][paxis]*v[i][paxis] + v[i][paxis2]*v[i][paxis2]);
   maxd=max(maxd,d);
  }
 }
 for(i=0;i<SFXinfo->nvert;i++){
  if(type == 1) d=v[i][paxis];       /* Linear */
  else if(type == 2){                  /* Radial */
   d=sqrt(v[i][paxis]*v[i][paxis] + v[i][paxis2]*v[i][paxis2]);
  }
  displacement=amplitude*cos(PI*2*((SFXinfo->time*PRMfreq)-d/flaplength));
  if(type == 1){                       /* Linear */
   d = v[i][paxis]-SFXinfo->vmin[paxis];
   maxd = SFXinfo->vmax[paxis]-SFXinfo->vmin[paxis];
  }
  if(fade == 2) displacement *= d / maxd;             /* fade up */
  else if(fade == 3) displacement *= 1 - d / maxd;    /* fade down */
  v[i][axis] += displacement;

  scale=PRMuamp*cos(PI*2*((SFXinfo->time*PRMfreq)-d/flaplength));
  if(fade == 2) scale *= d / maxd;
  else if(fade == 3) scale *= 1 - d / maxd;
  scale = 1.0 + scale;
  v[i][paxis2] *= scale;
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static int type=1,motion=1,xfade=1;
static double uamp=0.5,amp=25.0,length=1.0,freq=1.0;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
  ){
 char buffer[256];
 if(Op != NULL){  /* parameters exist so read them off the list          */
                  /* the name of the DLL file is always the first string */
   sscanf(Op,"%s %d %d %f %f %f %f %d",buffer,&type,&motion,
          &uamp,&amp,&length,&freq,&xfade);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_FLAP),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(buffer,"%s %d %d %f %f %f %f %d",name,type,motion,uamp,amp,
         length,freq,xfade);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"External effect: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hDlg,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     if(type == 2)SendDlgItemMessage(hDlg,DLG_FLAP_RADIAL,BM_SETCHECK,1,0);
     else         SendDlgItemMessage(hDlg,DLG_FLAP_LINEAR,BM_SETCHECK,1,0);
     if     (motion == 3)SendDlgItemMessage(hDlg,DLG_FLAP_FB,BM_SETCHECK,1,0);
     else if(motion == 2)SendDlgItemMessage(hDlg,DLG_FLAP_LR,BM_SETCHECK,1,0);
     else                SendDlgItemMessage(hDlg,DLG_FLAP_UD,BM_SETCHECK,1,0);
     if     (xfade == 3)SendDlgItemMessage(hDlg,DLG_FLAP_FADE_DOWN,BM_SETCHECK,1,0);
     else if(xfade == 2)SendDlgItemMessage(hDlg,DLG_FLAP_FADE_UP,BM_SETCHECK,1,0);
     else               SendDlgItemMessage(hDlg,DLG_FLAP_FADE_NONE,BM_SETCHECK,1,0);
     sprintf(str,"%.2f",amp);
     SetDlgItemText(hDlg,DLG_FLAP_AMP,str);
     sprintf(str,"%.2f",uamp);
     SetDlgItemText(hDlg,DLG_FLAP_UAMP,str);
     sprintf(str,"%.2f",length);
     SetDlgItemText(hDlg,DLG_FLAP_LENGTH,str);
     sprintf(str,"%.2f",freq);
     SetDlgItemText(hDlg,DLG_FLAP_FREQ,str);
     LoadAnimatedClip(hDlg);
     CentreDialogOnScreen(hDlg);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hDlg);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         EndDialog(hDlg,FALSE);
         return(TRUE);
       default:
         break;
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_FLAP_LINEAR:    type=1; break;
        case DLG_FLAP_RADIAL:    type=2; break;
        case DLG_FLAP_UD:        motion=1; break;
        case DLG_FLAP_LR:        motion=2; break;
        case DLG_FLAP_FB:        motion=3; break;
        case DLG_FLAP_FADE_NONE: xfade=1; break;
        case DLG_FLAP_FADE_UP:   xfade=2; break;
        case DLG_FLAP_FADE_DOWN: xfade=3; break;
        case DLG_FLAP_OK:
          if(GetDlgItemText(hDlg,DLG_FLAP_AMP,str,12) != 0)
            sscanf(str,"%f",&amp);
          if(GetDlgItemText(hDlg,DLG_FLAP_UAMP,str,12) != 0)
            sscanf(str,"%f",&uamp);
          if(GetDlgItemText(hDlg,DLG_FLAP_LENGTH,str,12) != 0)
            sscanf(str,"%f",&length);
          if(GetDlgItemText(hDlg,DLG_FLAP_FREQ,str,12) != 0)
            sscanf(str,"%f",&freq);
          EndDialog(hDlg,TRUE);
          return(TRUE);
        case DLG_FLAP_CANCEL:
          EndDialog(hDlg,FALSE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
