/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* spring.c */

#include <math.h>
#include <windows.h>
#include <commctrl.h>

#include "spring.h"

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#ifndef PI
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "pstruct.h"

#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(LibMain)
#endif

/* internal effects have no need to use this */

long _RenderGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, vertex *Vlist){
 return 1;
}

void _PreviewGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, Svertex *Vlist){
 return;
}

long _RenderExternalEffect(char *PrmList, sfxinfo *SFXinfo,
                                 vertex *Vlist){
 return 1;
}

void _PreviewExternalEffect(char *parameters, sfxinfo *SFXinfo,
                                    Svertex *v){
 double scale,dpmax,dpmin,t,sp;
 char buffer[128];
 int PRMcomp;
 long i;
 sscanf(parameters,"%s %d",buffer,&PRMcomp);
 t=SFXinfo->time;
 scale=4.0 * ((double)PRMcomp / 75.0) * (t-0.75);
 if     (t <= 0.50)sp=1.0-2.0*t*t;
 else if(t <= 0.75)sp=0.5+1.5*(t-0.5);
 else              sp=2.0-scale;
 dpmax=SFXinfo->vmax[1]-SFXinfo->vmin[1];
 for(i=0;i<SFXinfo->nvert;i++){
  if(t <= 0.75)v[i][1] = (v[i][1] - SFXinfo->vmin[1])*sp +
                           SFXinfo->vmin[1];
  else v[i][1] = (v[i][1] - SFXinfo->vmin[1])*sp + SFXinfo->vmin[1] +
                   scale*dpmax;
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static int compression=75;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
  ){
 char buffer[128];
 if(Op != NULL){  /* parameters exist so read them off the list          */
                  /* the name of the DLL file is always the first string */
   sscanf(Op,"%s %ld",buffer,&compression);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_SPRING),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(buffer,"8 %ld",compression);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"External effect: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hDlg,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[12];
 switch( msg ) {
   case WM_INITDIALOG:
     SetDlgItemInt(hDlg,DLG_SPRING_COMPRESSION,compression,FALSE);
     LoadAnimatedClip(hDlg);
     CentreDialogOnScreen(hDlg);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hDlg);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         EndDialog(hDlg,FALSE);
         return(TRUE);
       default:
         break;
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_SPRING_OK:
          if(GetDlgItemText(hDlg,DLG_SPRING_COMPRESSION,str,3) != 0)
            compression=atoi(str);
          EndDialog(hDlg,TRUE);
          return(TRUE);
        case DLG_SPRING_CANCEL:
          EndDialog(hDlg,FALSE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
