/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* Template for building EXTERNAL animation effects       */

#include <math.h>
#include <windows.h>

#include "template.h"

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#ifndef PI
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL; /* use to pick up resources from DLL   */

/************** Structures and memory routines for external address ********/

#include "pstruct.h"

/************************** Local Utility Functions ************************/

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 if(rcDlg.left < 0)OffsetRect(&rcDlg,-rcDlg.left,0);
 if(rcDlg.top  < 0)OffsetRect(&rcDlg,0,-rcDlg.top);
 if(rcDlg.right  > Xres)OffsetRect(&rcDlg,Xres-rcDlg.right,0);
 if(rcDlg.bottom > Yres)OffsetRect(&rcDlg,Yres-rcDlg.bottom,0);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

/************************ Functions called *********************/

long CALLBACK RenderExternalEffect(char *PrmList, sfxinfo *SFXinfo,
                                   vertex *Vlist);
void CALLBACK PreviewExternalEffect(char *parameters, sfxinfo *SFXinfo,
                                    Svertex *v);
char *CALLBACK SetExternalParameters(char *Op, HWND Hwnd, char *name,
                                     long ruler, X__MEMORY_MANAGER *lpEVI);

/************************** DLL entry point ********************************/

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(LibMain)
#endif

/*******************************************************************\
|                Code that define the effects                       |
\*******************************************************************/

long _RenderExternalEffect
(char *PrmList, sfxinfo *SFXinfo,vertex *Vlist){
 return 1;
}

long _RenderGlobalEffect
(char *PrmList, sfxinfo *SFXinfo,
 sfxdata *SFXdata, vertex *Vlist){
 return 1;
}

/*******************************************************************\
|         Code to draw preview into Animators                       |
\*******************************************************************/
void _PreviewGlobalEffect
(char *PrmList, sfxinfo *SFXinfo,
 sfxdata *SFXdata, Svertex *Vlist){
 return;
}

void _PreviewExternalEffect
(char *PrmList, sfxinfo *SFXinfo,Svertex *Vlist){
}

/*************** Functions used to set up effect ***************/

/* Dialog box callback  */
BOOL CALLBACK EffectDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);


/* Set defaults for the parameter variables, declare them as static so that
   they are available to the dialog function. They are here only for the
   function that edits the parameter list and must not be used in the
   code that is called from the renderer to execute the effect            */


/************ Functions called by the animator to set up effect *************/

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
/* output name and buffer should be as long as necessary to hold full string */
 char buffer[128];
 if(Op != NULL){  /* parameters exist so read them off the list          */
                  /* the name of the DLL file is always the first string */
   sscanf(Op,"%s",buffer);
 }
/*   Do the user interface as required to set up the effect, may use a   */
/*   dialog box etc. Return old string if effect is cancelled.           */
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_TEMPLATE),hWnd,
              (DLGPROC)EffectDlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
/* print the parameters into the buffer */
 sprintf(buffer,"%s",name);
/* Prepare the output buffer to take copy of parameter list */
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"External effect: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 /* Copy the parameter string to the output buffer */
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK EffectDlgProc(HWND hDlg,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     CentreDialogOnScreen(hDlg);
     return TRUE;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         EndDialog(hDlg,FALSE);
         return(TRUE);
       default:
         break;
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
