/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* warp.c */

#include <math.h>
#include <windows.h>
#include <commctrl.h>

#include "warp.h"

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#ifndef PI
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "pstruct.h"

#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(LibMain)
#endif

long _RenderGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, vertex *Vlist){
 return 1;
}

void _PreviewGlobalEffect
(char *PrmList, sfxinfo *SFXinfo, sfxdata *SFXdata, Svertex *Vlist){
 return;
}

long _RenderExternalEffect(char *PrmList, sfxinfo *SFXinfo,
                                 vertex *Vlist){
 return 1;
}

void _PreviewExternalEffect(char *parameters, sfxinfo *SFXinfo,
                                    Svertex *v){
 double scale,dpmax,dpmin,t,sp;
 long i;
 int PRMwarp;
 char *buffer[128];
 sscanf(parameters,"%s %d",buffer,&PRMwarp);
 t=SFXinfo->time;
 dpmax=SFXinfo->vmax[1]-SFXinfo->vmin[1];
 if(PRMwarp == 2){                  /* Warp in */
  if(t <= 0.5)sp=1.0+2.0*t;
  else        sp=2.0-2.0*(t-0.5);
  scale=2.0*(t-0.5);
  for(i=0;i<SFXinfo->nvert;i++){
   if(t <= 0.5)v[i][1] = (v[i][1] - SFXinfo->vmin[1])*sp + SFXinfo->vmin[1];
   else v[i][1] = (v[i][1] - SFXinfo->vmin[1])*sp +
                    SFXinfo->vmin[1] + scale*dpmax;
  }
 }
 else if(PRMwarp == 1){             /* Warp Out */
  t=1.0-SFXinfo->time;
  for(i=0;i<SFXinfo->nvert;i++){
   v[i][1] = v[i][1] + t*dpmax;
  }
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static int warp=1;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
 ){
 char buffer[128];
 if(Op != NULL){  /* parameters exist so read them off the list          */
                  /* the name of the DLL file is always the first string */
   sscanf(Op,"%s %ld",buffer,&warp);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_WARP),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(buffer,"7 %ld",warp);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"External effect: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hDlg,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[12];
 switch( msg ) {
   case WM_INITDIALOG:
     if(warp == 1)SendDlgItemMessage(hDlg,DLG_WARP_WARPIN,BM_SETCHECK,1,0);
     else SendDlgItemMessage(hDlg,DLG_WARP_WARPOUT,BM_SETCHECK,1,0);
     LoadAnimatedClip(hDlg);
     CentreDialogOnScreen(hDlg);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hDlg);
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wparam & 0xfff0)){
       case SC_CLOSE:
         EndDialog(hDlg,FALSE);
         return(TRUE);
       default:
         break;
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_WARP_WARPIN:    warp=1; break;
        case DLG_WARP_WARPOUT:   warp=2; break;
        case DLG_WARP_OK:
          EndDialog(hDlg,TRUE);
          return(TRUE);
        case DLG_WARP_CANCEL:
          EndDialog(hDlg,FALSE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
