#define MODULE_EFFECTS

/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

#include "render.h"

static void ExternalFX(char parameters[]);
static short ExternalFXram(char parameters[]);
static void ExternalFXfile(char parameters[]);
static void InternalFX_Wave(char parameters[]);
static void InternalFX_Modulate(char parameters[]);
static void InternalFX_Squash(char parameters[]);
static void InternalFX_Warp(char parameters[]);
static void InternalFX_Spring(char parameters[]);
static void InternalFX_Shake(char parameters[]);

typedef struct {
 long nvert;
 long ltime;
 long vmin[3];
 long vmax[3];
 long origin[3];
 double time;
} sfxinfo;

static sfxinfo SFXinfo;
static long FXObj;


#include "..\animate\wspaces.c"


#ifdef _SUNSTYLE
void EffectObjectVertices(O) long O; {
#else
void EffectObjectVertices(long O){
#endif
  double amplitude,wavelength,displacement;
  long i,j,n;
  vertex *v;
  double t,vmax[3],vmin[3];
  if((n=(long)Object[O].effect) < 4)return;
  for(i=0;i<3;i++){vmax[i] = (double)(BIGN); vmin[i] = (double)BIGP;}
  t=Object[O].effectratio;
  v=Object[O].Vbase;
  for(i=0;i<Object[O].NoVertices;i++){ /* get object limits */
    for(j=0;j<3;j++){
       v[i].p[j] -= Object[O].Origin[j];
       vmin[j] = min(vmin[j],v[i].p[j]);
       vmax[j] = max(vmax[j],v[i].p[j]);
    }
  }
  FXObj=O;
  SFXinfo.nvert=Object[O].NoVertices;
  SFXinfo.time=t;
  SFXinfo.ltime=(long)(t*1000);
  SFXinfo.vmin[0]=(long)vmin[0];   SFXinfo.vmax[0]=(long)vmax[0];
  SFXinfo.vmin[1]=(long)vmin[1];   SFXinfo.vmax[1]=(long)vmax[1];
  SFXinfo.vmin[2]=(long)vmin[2];   SFXinfo.vmax[2]=(long)vmax[2];
  SFXinfo.origin[0]=Object[O].Origin[0];
  SFXinfo.origin[1]=Object[O].Origin[1];
  SFXinfo.origin[2]=Object[O].Origin[2];
//MessageBox(NULL,Object[O].effect_param_list,"Debug",MB_OK);
  switch((short)n){
   case 4:  InternalFX_Wave(Object[O].effect_param_list);     break;
   case 5:  InternalFX_Modulate(Object[O].effect_param_list); break;
   case 6:  InternalFX_Squash(Object[O].effect_param_list);   break;
   case 7:  InternalFX_Warp(Object[O].effect_param_list);     break;
   case 8:  InternalFX_Spring(Object[O].effect_param_list);   break;
   case 9:  InternalFX_Shake(Object[O].effect_param_list);    break;
   case 10:                                                   break;
   case 50: ExternalFX(Object[O].effect_param_list);          break;
   default: break;
  }
  for(i=0;i<Object[O].NoVertices;i++){
   for(j=0;j<3;j++){
    v[i].p[j] += Object[O].Origin[j];
   }
  }
}

#ifdef _SUNSTYLE
static void InternalFX_Wave(parameters) char parameters[]; {
#else
static void InternalFX_Wave(char parameters[]){
#endif
 double amplitude,wavelength,displacement,PRMamp,PRMlngth,PRMfreq,d,maxd;
 char buffer[128];
 int type, PRMaxis, axis, fade, paxis, paxis2;
 long i;
 vertex *v;
 sscanf(parameters,"%s %d %d %f %f %f %d",buffer,&type,&PRMaxis,&PRMamp,
        &PRMlngth,&PRMfreq,&fade);
 v=Object[FXObj].Vbase;
 if(PRMaxis == 3) axis=1;         /* F / B */
 else if(PRMaxis == 1) axis=2;    /* U / D */
 else if(PRMaxis == 2) axis=0;    /* L / R */
 if(axis == 0){      paxis = 1; paxis2 = 2; }
 else if(axis == 2){ paxis = 1; paxis2 = 0; }
 else if(axis == 1){ paxis = 0; paxis2 = 2; }
 wavelength=max((SFXinfo.vmax[paxis]-SFXinfo.vmin[paxis]) * PRMlngth,1.0);
 amplitude=(PRMamp/25.0) * (wavelength/8);
 if(type == 2){                        /* Radial */
  maxd=0.00;
  for(i=0;i<SFXinfo.nvert;i++){
   d=sqrt(v[i].p[paxis]*v[i].p[paxis] + v[i].p[paxis2]*v[i].p[paxis2]);
   maxd=max(maxd,d);
  }
 }
 for(i=0;i<SFXinfo.nvert;i++){
  if(type == 1) d=v[i].p[paxis];       /* Linear */
  else if(type == 2){                  /* Radial */
   d=sqrt(v[i].p[paxis]*v[i].p[paxis] + v[i].p[paxis2]*v[i].p[paxis2]);
  }
  displacement=amplitude*cos(PI*2*((SFXinfo.time*PRMfreq)-d/wavelength));
  if(type == 1){                       /* Linear */
   d = v[i].p[paxis]-SFXinfo.vmin[paxis];
   maxd = SFXinfo.vmax[paxis]-SFXinfo.vmin[paxis];
  }
  if(fade == 2) displacement *= d / maxd;             /* fade up */
  else if(fade == 3) displacement *= 1 - d / maxd;    /* fade down */
  v[i].p[axis] += displacement;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Modulate(parameters) char parameters[]; {
#else
static void InternalFX_Modulate(char parameters[]){
#endif
 double scale,dpmax,amplitude,PRMamp,PRMlngth,PRMfreq;
 int PRMaxis, axis, paxis, paxis2;
 char buffer[128];
 long i;
 vertex *v;
 sscanf(parameters,"%s %d %f %f %f",buffer,&PRMaxis,&PRMamp,&PRMlngth,&PRMfreq);
 if(PRMaxis == 3) axis=1;         /* F / B */
 else if(PRMaxis == 1) axis=2;    /* U / D */
 else if(PRMaxis == 2) axis=0;    /* L / R */
 if(axis == 0){      paxis = 1; paxis2 = 2; }
 else if(axis == 2){ paxis = 1; paxis2 = 0; }
 else if(axis == 1){ paxis = 0; paxis2 = 2; }
 v=Object[FXObj].Vbase;
 dpmax=max((SFXinfo.vmax[axis]-SFXinfo.vmin[axis]) * PRMlngth,1.0);
 amplitude = 0.1 * (PRMamp / 25.0);
 for(i=0;i<SFXinfo.nvert;i++){
  scale=1.0+amplitude*cos(PI*2*((SFXinfo.time*PRMfreq)-v[i].p[axis]/dpmax));
  v[i].p[paxis]  *= scale;
  v[i].p[paxis2] *= scale;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Squash(parameters) char parameters[]; {
#else
static void InternalFX_Squash(char parameters[]){
#endif
 double scale,dpmax,dpmin,t,sp;
 long i;
 char buffer[128];
 int PRMcomp, PRMeffect, PRMplane, axis, paxis, paxis2;
 vertex *v;
 sscanf(parameters,"%s %d %d %d",buffer,&PRMeffect,&PRMplane,&PRMcomp);
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 t = -1.00 * (t*t) + 2.00 * t;
 if(PRMeffect == 2) t = 1.00 - t;
 if(PRMplane == 1){           /* Horizontal Plane */
  axis=2; paxis=1; paxis2=0;
 }
 else if(PRMplane == 2){      /* Vertical Plane */
  axis=1; paxis=0; paxis2=2;
 }
 dpmax=max(fabs(SFXinfo.vmin[axis]),fabs(SFXinfo.vmax[axis]));
 dpmin=dpmax-t*dpmax*((double)PRMcomp/100.0);
 sp=dpmin/dpmax;
 for(i=0;i<SFXinfo.nvert;i++){
  scale=1.0+((double)PRMcomp/100.0)*t*cos(PI/2*v[i].p[axis]/dpmax);
  v[i].p[axis]   *= sp;
  v[i].p[paxis]  *= scale;
  v[i].p[paxis2] *= scale;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Warp(parameters) char parameters[]; {
#else
static void InternalFX_Warp(char parameters[]){
#endif
 double scale,dpmax,dpmin,t,sp;
 long i;
 int PRMwarp;
 char buffer[128];
 vertex *v;
 sscanf(parameters,"%s %d",buffer,&PRMwarp);
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 dpmax=SFXinfo.vmax[1]-SFXinfo.vmin[1];
 if(PRMwarp == 2){                  /* Warp in */
  if(t <= 0.5)sp=1.0+2.0*t;
  else        sp=2.0-2.0*(t-0.5);
  scale=2.0*(t-0.5);
  for(i=0;i<SFXinfo.nvert;i++){
   if(t <= 0.5)v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp + SFXinfo.vmin[1];
   else v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp +
                    SFXinfo.vmin[1] + scale*dpmax;
  }
 }
 else if(PRMwarp == 1){             /* Warp Out */
  t=1.0-SFXinfo.time;
  for(i=0;i<SFXinfo.nvert;i++){
   v[i].p[1] = v[i].p[1] + t*dpmax;
  }
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Spring(parameters) char parameters[]; {
#else
static void InternalFX_Spring(char parameters[]){
#endif
 double scale,dpmax,dpmin,t,sp;
 char buffer[128];
 int PRMcomp;
 long i;
 vertex *v;
 sscanf(parameters,"%s %d",buffer,&PRMcomp);
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 scale=4.0 * ((double)PRMcomp / 75.0) * (t-0.75);
 if     (t <= 0.50)sp=1.0-2.0*t*t;
 else if(t <= 0.75)sp=0.5+1.5*(t-0.5);
 else              sp=2.0-scale;
 dpmax=SFXinfo.vmax[1]-SFXinfo.vmin[1];
 for(i=0;i<SFXinfo.nvert;i++){
  if(t <= 0.75)v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp +
                           SFXinfo.vmin[1];
  else v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp + SFXinfo.vmin[1] +
                   scale*dpmax;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Shake(parameters) char parameters[]; {
#else
static void InternalFX_Shake(char parameters[]){
#endif
 double displacement,dpmax,dpmin,t,sp;
 long i;
 vertex *v;
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 dpmax=(double)(SFXinfo.vmax[2]-SFXinfo.vmin[2])*
       (double)(SFXinfo.vmax[2]-SFXinfo.vmin[2]);
 dpmin=(double)(SFXinfo.vmax[0]-SFXinfo.vmin[0])/4.0;
 sp=dpmin*sin(2*PI*SFXinfo.time);
 for(i=0;i<SFXinfo.nvert;i++){
  displacement=sp*((double)(v[i].p[2]-SFXinfo.vmin[2])/dpmax)*
                   (double)(v[i].p[2]-SFXinfo.vmin[2]);
  v[i].p[0] += displacement;
 }
}

static short ExternalFXnt(char *parameters);

static void ExternalFX(char parameters[]){
  ExternalFXnt(parameters);
}

static short ExternalFXnt(char *parameters){
 struct  tvtx {vector p; double x,y;} *v;
 char fname[256],tparam[256];
 HMODULE hLib;
 FARPROC fpFun;
 long l,result=0;
 long (*fpFun1)(char *, sfxinfo *, struct tvtx *);
 if(Object[FXObj].Vbase == NULL)return 0;
 if((v=(struct tvtx *)X__Malloc(sizeof(struct tvtx)*Object[FXObj].NoVertices))
      == NULL)return 0;
 for(l=0;l<Object[FXObj].NoVertices;l++){
   VECCOPY(Object[FXObj].Vbase[l].p,v[l].p)
 }
 sscanf(parameters,"%s",tparam);
 RestoreWhiteSpaces(tparam);
 sscanf(tparam,"%[^\0]",fname);
 l=strlen(fname);
 fname[l-1]='x';   /* use this to avoid Windows 3.1 file sharing problem */
 fname[l-2]='l';   /* effects directory must contain duplicte DLL's      */
 fname[l-3]='d';
 if((hLib=LoadLibrary(fname)) == NULL){ /* try to get 'dll' */
   fname[l-1]='l';
   fname[l-2]='l';
   fname[l-3]='d';
   hLib=LoadLibrary(fname);
 }
 if(hLib != NULL){
   if((fpFun=GetProcAddress(hLib,"_RenderExternalEffect")) != NULL){
     fpFun1 = (void *)fpFun;
     result=(*fpFun1)(parameters,&SFXinfo,v);
     for(l=0;l<Object[FXObj].NoVertices;l++){
       VECCOPY(v[l].p,Object[FXObj].Vbase[l].p)
     }
   }
   else MessageBox(GetFocus(),"Effect: Function Fail",NULL,MB_OK);
   FreeLibrary(hLib);
 }
 else MessageBox(GetFocus(),
                (LPCTSTR) "Effect: Load Fail",NULL,MB_OK);
 X__Free(v);
 return 0;
}
