#define MODULE_LOADER

/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

#include "render.h"

typedef struct SKELETON {
  long id;
  vector p;
  double T[4][4],R[4][4],Q[4][4];
}skel;

typedef double quaternion[4];

extern unsigned char *LoadMAP(char *name, int *x, int *y);
extern unsigned char *MakeMapImage(unsigned char *, int, int, int *, int *);
extern BOOL Normalize(vector x,vector y,vector z,float *n);

static void InterpolateRobot(double mr, long n, skel *sk1, skel *sk2);
static long LoadQmorph(char *QmorphFile,long N,ivector *Qv, ivector Mq);
static void QuadInterpolate(short mo, double mr, double qr,
                            vector v0, vector v1, vector vm,
                            vector v);
static void RenderTransformToRobot(long O,
                             long *Skid, skel *Skeleton,
                             double mr,
                             char *robo_f1, char *robo_f2);
static long *RenderReadSkeleton(long clen, short version, skel **sp, long *ns);
static short R_ReadAxis(long clen);
static short R_ReadOldFaces(long  clen, long shader);
static short R_ReadFaces(long  clen, BOOL short_form);
static short R_ReadVertices(long clen);
static short R_ReadVertText(long clen);
static short R_ReadOldText(long clen, long frame,
                    char *objectfilename, char *str);
static short R_ReadText(long clen, long frame,
                    char *objectfilename, char *str);
static short R_ReadShader(long clen, long frame,
                    char *objectfilename, char *str);
static short R_ReadMorphShader(long clen,
                    char *objectfilename, char *str);
static void R_MorphVector(vector b, ivector Mx, double mr);
static void R_getchunk(long clen);

#if !__DLL__
typedef struct tagMEMORYOBJECT {
  char name[256];
  char *effect;
  ivector Origin;
  long NoVertices,NoFaces,NoMaps,NoSkeleton;
  face    *Fbase;
  vertex  *Vbase;
  iMap    *Mbase;
  long    *Skid;
  skel    *Skeleton;
} MEMORYOBJECT;

static MEMORYOBJECT *MemObjList=NULL;
static long nMemObj=0;

static BOOL GetMemoryObject(char *name, object *O, long **Skid,
                            skel **Skeleton, long *nSkeleton){
 int i;
 MEMORYOBJECT *t;
 if(nMemObj > 0){
   for(i=0;i<nMemObj;i++){
     t=(MemObjList+i);
     if(strcmpi(name,t->name) == 0){
       VECCOPY(t->Origin,O->Origin);
       if(t->Fbase != NULL){
         if((O->Fbase=(face *)X__Malloc(t->NoFaces * sizeof(face)))
                     == NULL)return FALSE;
         memcpy(O->Fbase,t->Fbase,sizeof(face)*t->NoFaces);
         O->NoFaces=t->NoFaces;
       }
       if(t->Vbase != NULL){
         if((O->Vbase=(vertex *)X__Malloc(t->NoVertices * sizeof(vertex)))
                     == NULL)return FALSE;
         memcpy(O->Vbase,t->Vbase,sizeof(vertex)*t->NoVertices);
         O->NoVertices=t->NoVertices;
       }
       if(t->Mbase != NULL){
         if((O->Mbase=(iMap *)X__Malloc(t->NoMaps * sizeof(iMap)))
               == NULL)return FALSE;
         memcpy(O->Mbase,t->Mbase,sizeof(iMap)*t->NoMaps);
         O->NoMaps=t->NoMaps;
       }
       if(t->Skid != NULL){
         if((*Skid=(long *)X__Malloc(t->NoVertices * sizeof(long)))
                 == NULL)return FALSE;
         memcpy(*Skid,t->Skid,sizeof(long)*t->NoVertices);
       }
       if(t->Skeleton != NULL){
         if((*Skeleton=(skel *)X__Malloc(t->NoSkeleton * sizeof(skel)))
                      == NULL)return FALSE;
         memcpy(*Skeleton,t->Skeleton,sizeof(skel)*t->NoSkeleton);
         *nSkeleton=t->NoSkeleton;
       }
//MessageBox(NULL,t->name,"Found",MB_OK);
       return TRUE;
     }
   }
 }
 return FALSE;
}

static void PutMemoryObject(char *name ,object *O, long *Skid,
                     skel *Skeleton, long nSkeleton){
 int i;
 MEMORYOBJECT *t;
 if(MemObjList == NULL)
   t=(MEMORYOBJECT *)X__Malloc(sizeof(MEMORYOBJECT));
 else
   t=(MEMORYOBJECT *)X__Realloc(MemObjList,sizeof(MEMORYOBJECT)*(nMemObj+1));
 if(t == NULL){
MessageBox(NULL,"No memory for objects",NULL,MB_OK);
   return;
 }
 MemObjList=t;
 t=(MemObjList+nMemObj);
 strcpy(t->name,name);
 VECCOPY(O->Origin,t->Origin);
 t->effect=NULL;
 t->Fbase=NULL;
 t->Vbase=NULL;
 t->Mbase=NULL;
 t->Skid=NULL;
 t->Skeleton=NULL;
 t->NoVertices=0;
 t->NoFaces=0;
 t->NoMaps=0;
 t->NoSkeleton=0;
 if(O->Fbase != NULL){
   if((t->Fbase=(face *)X__Malloc(O->NoFaces * sizeof(face)))
               == NULL){
MessageBox(NULL,"No memory for faces",NULL,MB_OK);
     return;
   }
   memcpy(t->Fbase,O->Fbase,sizeof(face)*O->NoFaces);
   t->NoFaces=O->NoFaces;
 }
 if(O->Vbase != NULL){
   if((t->Vbase=(vertex *)X__Malloc(O->NoVertices * sizeof(vertex)))
               == NULL){
MessageBox(NULL,"No memory for vertices",NULL,MB_OK);
     return;
   }
   memcpy(t->Vbase,O->Vbase,sizeof(vertex)*O->NoVertices);
   t->NoVertices=O->NoVertices;
 }
 if(O->Mbase != NULL){
   if((t->Mbase=(iMap *)X__Malloc(O->NoMaps * sizeof(iMap)))
               == NULL){
MessageBox(NULL,"No memory for maps",NULL,MB_OK);
     return;
   }
   /* load and scale the maps now to save time while playing */
   for(i=0;i<O->NoMaps;i++){
     int xm,ym,ix,iy;
     unsigned char *m_pixels;
     iMap *tMap;
     tMap=(O->Mbase+i);
     if(tMap->pp < 0 && tMap->rp > 0) /*  try a reflection map  */
           m_pixels=LoadMAP(tMap->frefl,&xm,&ym);
     else  m_pixels=LoadMAP(tMap->filename,&xm,&ym);
     if(m_pixels != NULL){
       m_pixels=MakeMapImage(m_pixels,xm,ym,&ix,&iy);
       tMap->ix = ix;
       tMap->iy = iy;
       tMap->pixels = m_pixels;
     }
     else{
       tMap->ix = 0;
       tMap->iy = 0;
       tMap->pixels = NULL;
     }
   }
   memcpy(t->Mbase,O->Mbase,sizeof(iMap)*O->NoMaps);
   t->NoMaps=O->NoMaps;
 }
 if(Skid != NULL){
   if((t->Skid=(long *)X__Malloc(O->NoVertices * sizeof(long)))
               == NULL){
MessageBox(NULL,"No memory for Skid",NULL,MB_OK);
     return;
   }
   memcpy(t->Skid,Skid,sizeof(long)*O->NoVertices);
 }
 if(Skeleton != NULL){
   if((t->Skeleton=(skel *)X__Malloc(nSkeleton * sizeof(skel)))
               == NULL){
MessageBox(NULL,"No memory for Skeleton",NULL,MB_OK);
     return;
   }
   memcpy(t->Skeleton,Skeleton,sizeof(skel)*nSkeleton);
   t->NoSkeleton=nSkeleton;
 }
//MessageBox(NULL,name,"Putting",MB_OK);
 nMemObj++;
}

void FreeMemoryObjects(void){
 int i,j;
 MEMORYOBJECT *t;
//MessageBox(NULL,"Freeing List","Debug",MB_OK);
 if(nMemObj > 0)for(i=0;i<nMemObj;i++){
   t=(MemObjList+i);
   if(t->Fbase != NULL)X__Free(t->Fbase);
   if(t->Vbase != NULL)X__Free(t->Vbase);
   if(t->Mbase != NULL){
     for(j=0;j<t->NoMaps;j++){ /* any stored maps */
       if((t->Mbase+j)->pixels != NULL)X__Free((t->Mbase+j)->pixels);
     }
     X__Free(t->Mbase);
   }
   if(t->effect!= NULL)X__Free(t->effect);
   if(t->Skid != NULL)X__Free(t->Skid);
   if(t->Skeleton != NULL)X__Free(t->Skeleton);
 }
 X__Free(MemObjList);
 MemObjList=NULL;
 nMemObj=0;
}

vertex  *MainVp=NULL;
face    *MainFp=NULL;
long    Nvert=0,Nface=0;

#include "optimise.c"
#endif          /* end DLL */

static HANDLE fp;
static DWORD dwRead;
static char str[10];
static long newface,newvert;

char *short_form(char *filename){
 char *sps,*spc;
 sps=spc=filename;
 while((spc=strchr(spc,'\\')) != NULL){spc++; sps=spc;}
 return sps;
}

char R_getl(long *val){
 long l,l0,l1,l2,l3;
 ReadFile(fp,&l,sizeof(long),&dwRead,NULL);
 l0 = (l >> 24) & 0x000000ff;
 l1 = (l >> 16) & 0x000000ff;
 l2 = (l >>  8) & 0x000000ff;
 l3 = (l      ) & 0x000000ff;
 *val = (l3 << 24) | (l2 << 16) | (l1 <<  8) | l0;
 return(1);
}

char R_getsht(short * val){
 short s,s0,s1;
 ReadFile(fp,&s,sizeof(short),&dwRead,NULL);
 s0 = (s >>  8) & 0x00ff;
 s1 = (s      ) & 0x00ff;
 *val = (s1 <<  8) | s0;
 return(1);
}

char R_getusht(unsigned short * val){
 unsigned short s,s0,s1;
 ReadFile(fp,&s,sizeof(short),&dwRead,NULL);
 s0 = (s >>  8) & 0x00ff;
 s1 = (s      ) & 0x00ff;
 *val = (s1 <<  8) | s0;
 return(1);
}

static unsigned short GetUshortFromBuffer(unsigned char *b){
 return ((unsigned short)(*b) <<  8) | (unsigned short)(*(b+1));
}

static short GetShortFromBuffer(unsigned char *b){
 return ((short)(*b) <<  8) | (short)(*(b+1));
}

static long GetLongFromBuffer(unsigned char *b){
 return ((long)(*b) << 24) | ((long)(*(b+1)) << 16) |
        ((long)(*(b+2)) <<  8) | (long)(*(b+3));
}

static void R_MorphVector(vector b, ivector Mx, double mr){
 long i,j;
 for(i=0;i<3;i++){
   R_getl(&j); j=j+Object[ObjectCount].Origin[i]-Mx[i];
   b[i]=(b[i]-(double)j)*mr + (double)j;
 }
}

long LoadMorphObject(short mo, char *FileName,double mr,
                     char *QmorphFile,double qr){
 unsigned char *buffer=NULL,*b;
 long CHUNKsize,FORMsize,nm,i,j,nf,bid,l,m,m_angle,l_temp,version,mc,
      file_len,shader,na;
 unsigned short uj;
 short s,m_pp,m_rp,m_bp;
 unsigned char mcolor[3],mtexture,mmatsp,mmattr,mmatre,
               mtexco[3],mbrush,uc_temp,t_colour[3];
 char k,filename[256],ps,rs,bs;
 double a;
 ivector Mx,Mv,Mq,*Qv;
 vector V0,V1,VM;
 vertex *v;
 face   *f;
 Qv=NULL;
 if(mo > 1){
   nm=(long)(Object[ObjectCount].NoVertices)*(long)(sizeof(ivector));
   if(mo > 1){
     if((Qv=(ivector *)X__Malloc(nm)) == NULL)mo=1;
   }
   if(mo > 1){
     if(LoadQmorph(QmorphFile,Object[ObjectCount].NoVertices,Qv,Mq) < 0){
        X__Free(Qv); Qv=NULL;
        mo=1; /* Q load fail ignore Q morph */
      }
   }
 }
 Mx[0]=0; Mx[1]=0; Mx[2]=0;
 fp=CreateFile(FileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(fp == INVALID_HANDLE_VALUE){
   goto X__ERROR;
 }
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || strcmp(str,"FORM") != 0){
   CloseHandle(fp);
   goto X__ERROR;
 }
 R_getl(&FORMsize);
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || (strcmp(str,"OFXM") != 0 && strcmp(str,"AAPO") != 0)){
   CloseHandle(fp);
   goto X__ERROR;
 }
loop:
 if(!ReadFile(fp,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4)goto end;
 if (R_getl(&CHUNKsize) != 1)goto end;
 if     (strcmp(str,"VERT") == 0){ /* offset assumes Axis is first segment */
   nm=CHUNKsize/12;
   if(nm != Object[ObjectCount].NoVertices){
     CloseHandle(fp);
     goto X__ERROR;
   }
   v=Object[ObjectCount].Vbase;
   if(mo == 1){
     for(i=0;i<nm;i++){
       R_getl(&j); Mv[0] = j+Object[ObjectCount].Origin[0]-Mx[0];
       R_getl(&j); Mv[1] = j+Object[ObjectCount].Origin[1]-Mx[1];
       R_getl(&j); Mv[2] = j+Object[ObjectCount].Origin[2]-Mx[2];
       v[i].p[0] = (v[i].p[0]-(double)Mv[0])*mr + (double)Mv[0];
       v[i].p[1] = (v[i].p[1]-(double)Mv[1])*mr + (double)Mv[1];
       v[i].p[2] = (v[i].p[2]-(double)Mv[2])*mr + (double)Mv[2];
     }
   }
   else{
     for(i=0;i<nm;i++){
       V1[0] = v[i].p[0] - (double)Object[ObjectCount].Origin[0];
       V1[1] = v[i].p[1] - (double)Object[ObjectCount].Origin[1];
       V1[2] = v[i].p[2] - (double)Object[ObjectCount].Origin[2];
       VM[0] = (double)(Qv[i][0] - Mq[0]);
       VM[1] = (double)(Qv[i][1] - Mq[1]);
       VM[2] = (double)(Qv[i][2] - Mq[2]);
       R_getl(&j); V0[0] = (double)(j - Mx[0]);
       R_getl(&j); V0[1] = (double)(j - Mx[1]);
       R_getl(&j); V0[2] = (double)(j - Mx[2]);
       QuadInterpolate(mo,mr,qr,V0,V1,VM,v[i].p);
       v[i].p[0] += (double)Object[ObjectCount].Origin[0];
       v[i].p[1] += (double)Object[ObjectCount].Origin[1];
       v[i].p[2] += (double)Object[ObjectCount].Origin[2];
     }
   }
 }
 else if(strcmp(str,"AXIS") == 0){/* Axis is assumed to be first chunk */
   R_getl(&j); Mx[0] = j; /* this is the morphed object offset */
   R_getl(&j); Mx[1] = j; /* before using the morph it must be reposition */
   R_getl(&j); Mx[2] = j; /* to have its offset the same as the object */
   R_MorphVector(Object[ObjectCount].Axis[0].bo,Mx,mr); /* Tx origin */
   R_MorphVector(Object[ObjectCount].Axis[0].bu,Mx,mr); /* Tx X      */
   R_MorphVector(Object[ObjectCount].Axis[0].bv,Mx,mr); /* Tx X      */
   VECCOPY(Object[ObjectCount].Axis[0].bo,Object[ObjectCount].Axis[0].boin)
   VECCOPY(Object[ObjectCount].Axis[0].bu,Object[ObjectCount].Axis[0].buin)
   VECCOPY(Object[ObjectCount].Axis[0].bv,Object[ObjectCount].Axis[0].bvin)
 }
 else if(strcmp(str,"FACE") == 0 || strcmp(str,"SFAC") == 0){ /* morph the face colour */
  if(strcmp(str,"FACE") == 0){nf=CHUNKsize/20,shader=0;}
  else                       {nf=CHUNKsize/21;shader=1;}
  if(nf != Object[ObjectCount].NoFaces){
    CloseHandle(fp);
    goto X__ERROR;
  }
  buffer=(unsigned char *)X__Malloc(CHUNKsize);
  if(buffer == NULL){
    CloseHandle(fp);
    goto X__ERROR;
  }
  ReadFile(fp,buffer,CHUNKsize,&dwRead,NULL);
  b=buffer;
  f=Object[ObjectCount].Fbase;
  for(i=0;i<nf;i++){
    b += 3*sizeof(long);
    memcpy(mcolor,b,3); b += 3;
    mtexture = *b++;
    (f+i)->color[0]=( (f+i)->color[0] - mcolor[0] )*mr + mcolor[0];
    (f+i)->color[1]=( (f+i)->color[1] - mcolor[1] )*mr + mcolor[1];
    (f+i)->color[2]=( (f+i)->color[2] - mcolor[2] )*mr + mcolor[2];
    mmatsp = *b++;
    mmattr = *b++;
    mmatre = *b++;
    mbrush = *b++;
    mtexco[0] = (mmatsp & 0x0f) << 4;
    mtexco[1] = (mmattr & 0x0f) << 4;
    mtexco[2] = (mmatre & 0x0f) << 4;
    mmatsp  = (mmatsp & 0xf0) >> 4;
    mmattr &= 0xf0;
    mmatre &= 0xf0;
    (f+i)->texco[0]=( (f+i)->texco[0] - mtexco[0] )*mr + mtexco[0];
    (f+i)->texco[1]=( (f+i)->texco[1] - mtexco[1] )*mr + mtexco[1];
    (f+i)->texco[2]=( (f+i)->texco[2] - mtexco[2] )*mr + mtexco[2];
    (f+i)->matsp=( (f+i)->matsp - mmatsp )*mr + mmatsp;
    (f+i)->mattr=( (f+i)->mattr - mmattr )*mr + mmattr;
    (f+i)->matre=( (f+i)->matre - mmatre )*mr + mmatre;
    if(shader)b++; /* dump the shader axis ID */
  }
  X__Free(buffer);
 }
 else if(strcmp(str,"SFCE") == 0 || strcmp(str,"LFCE") == 0){
  if(strcmp(str,"SFCE") == 0){nf=CHUNKsize/18,shader=0;}
  else                       {nf=CHUNKsize/24;shader=1;}
  if(nf != Object[ObjectCount].NoFaces){
    CloseHandle(fp);
    goto X__ERROR;
  }
  buffer=(unsigned char *)X__Malloc(CHUNKsize);
  if(buffer == NULL){
    CloseHandle(fp);
    goto X__ERROR;
  }
  ReadFile(fp,buffer,CHUNKsize,&dwRead,NULL);
  b=buffer;
  f=Object[ObjectCount].Fbase;
  for(i=0;i<nf;i++){
    if(shader){ b += 3*sizeof(long);}
    else      { b += 3*sizeof(unsigned short);}
    memcpy(mcolor,b,3); b += 3;
    (f+i)->color[0]=( (f+i)->color[0] - mcolor[0] )*mr + mcolor[0];
    (f+i)->color[1]=( (f+i)->color[1] - mcolor[1] )*mr + mcolor[1];
    (f+i)->color[2]=( (f+i)->color[2] - mcolor[2] )*mr + mcolor[2];
    mmatsp = *b++;
    mmattr = *b++;
    mmatre = *b++;
    mtexco[0] = (mmatsp & 0x0f) << 4;
    mtexco[1] = (mmattr & 0x0f) << 4;
    mtexco[2] = (mmatre & 0x0f) << 4;
    (f+i)->texco[0]=( (f+i)->texco[0] - mtexco[0] )*mr + mtexco[0];
    (f+i)->texco[1]=( (f+i)->texco[1] - mtexco[1] )*mr + mtexco[1];
    (f+i)->texco[2]=( (f+i)->texco[2] - mtexco[2] )*mr + mtexco[2];
    mmatsp  = (mmatsp & 0xf0) >> 4;
    b++;  /*  attrib           */
    mtexture = *b++;/*  param  */
    b++;  /* dump the map      */
    b++;  /* shader            */
    b++;  /* axis              */
    b++;  /* pad               */
  }
  X__Free(buffer);
 }
 else R_getchunk(CHUNKsize);
 goto loop;
end:
 CloseHandle(fp);
 if(Qv != NULL)X__Free(Qv);
 return 1;
 X__ERROR:
 if(Qv != NULL)X__Free(Qv);
 return 0;
}

static void QuadInterpolate(short mo, double mr, double qr,
                            vector v0, vector v1, vector vm,
                            vector v){
 double a,b,c;
 long j;
 for(j=0;j<3;j++){
   if(mo == 2){ /* backwards interpolation */
     a=(    (v1[j]-v0[j]) - (v0[j] - vm[j])/qr )/(1.0+qr);
     b=( qr*(v1[j]-v0[j]) + (v0[j] - vm[j])/qr )/(1.0+qr);
     c=v0[j];
   }
   else{        /* =3  forwards interpolation */
     a=(       -(v1[j]-v0[j]) + (vm[j] - v0[j])/(1.0+qr) )/qr;
     b=( (qr+1)*(v1[j]-v0[j]) - (vm[j] - v0[j])/(1.0+qr) )/qr;
     c=v0[j];
   }
   v[j] = a*mr*mr + b*mr +c;
 }
}

static long LoadQmorph(char *QmorphFile,long N,ivector *Qv, ivector Mq){
long CHUNKsize,FORMsize,nm,i,j;
 fp=CreateFile(QmorphFile,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(fp == INVALID_HANDLE_VALUE){
   goto X__ERROR;
 }
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || strcmp(str,"FORM") != 0){
   CloseHandle(fp);
   goto X__ERROR;
 }
 R_getl(&FORMsize);
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || (strcmp(str,"OFXM") != 0 && strcmp(str,"AAPO") != 0)){
   CloseHandle(fp);
   goto X__ERROR;
 }
loop:
 if(!ReadFile(fp,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4 )goto end;
 if (R_getl(&CHUNKsize) != 1)goto end;
 if     (strcmp(str,"VERT") == 0){ /* offset assumes Axis is first segment */
   nm=CHUNKsize/12;
   if(nm != N){
     CloseHandle(fp);
     goto X__ERROR;
   }
   for(i=0;i<nm;i++){
     R_getl(&j); Qv[i][0]=j;
     R_getl(&j); Qv[i][1]=j;
     R_getl(&j); Qv[i][2]=j;
   }
 }
 else if(strcmp(str,"AXIS") == 0){/* Axis is assummed to be first chumk */
   R_getl(&j); Mq[0] = j; /* this is the morphed object offset */
   R_getl(&j); Mq[1] = j; /* before using the morph it must be reposition */
   R_getl(&j); Mq[2] = j; /* to have its offset the same as the object */
   for(i=0;i<9;i++)R_getl(&j);
 }
 else R_getchunk(CHUNKsize);
 goto loop;
end:
 CloseHandle(fp);
 return  2;
 X__ERROR:
 return -1;
}

static long *RenderReadSkeleton(long clen, short version, skel **sp, long *ns){
 short s;
 skel *Skeleton;
 long n,i,j,k,l,l0,l1,l2,*jj,*Skid;
 char name[32];
 Skeleton=NULL;
 Skid=NULL;
 R_getl(&n); *ns=n;
 if((*sp=(skel *)X__Malloc(n*(long)sizeof(skel))) != NULL)Skeleton=*sp;
 for(i=0;i<n;i++){ /* dump the original skeleton */
   R_getl(&l0);  R_getl(&l1);  R_getl(&l2);  R_getsht(&s);
   if(Skeleton != NULL){
     (Skeleton+i)->id=(long)s;
     (Skeleton+i)->p[0] = (double)(l0);
     (Skeleton+i)->p[1] = (double)(l1);
     (Skeleton+i)->p[2] = (double)(l2);
   }
   if(version > 0){
     ReadFile(fp,name,sizeof(char)*16,&dwRead,NULL);
   }
   if(version > 1){
     for(j=0;j<3;j++){R_getl(&l);}
     for(j=0;j<3;j++){R_getl(&l);}
     for(j=0;j<3;j++){R_getl(&l);}
     for(j=0;j<8;j++)for(k=0;k<3;k++)R_getl(&l);
   }
 }
 R_getl(&n);  /* get number of vertices in skeleton structure */
 if(n > 0){
   if(n == newvert){ /* if skeleton assignment is consitent with newvert */
     if((Skid=(long *)X__Malloc(n*(long)sizeof(long))) != NULL)jj=Skid;
   }
   for(i=0;i<n;i++){
     R_getsht(&s);
     if(Skid != NULL)*jj++ = (long)s;
   }
 }
 return Skid;
}

#ifdef _SUNSTYLE
long R_LoadObject(FileName,frame,robo,robo_ratio,robo_file1,robo_file2)
char *FileName; long frame; short robo; double robo_ratio;
char *robo_file1; char *robo_file2; {
#else
long R_LoadObject(char *FileName, long frame, short robo, double robo_ratio,
                  char *robo_file1, char *robo_file2){
#endif
 long CHUNKsize,FORMsize,l,nSkeleton=0;
 long *Skids=NULL;
 skel *Skeleton=NULL;
 Object[ObjectCount].Fbase=NULL;
 Object[ObjectCount].Vbase=NULL;
 Object[ObjectCount].Mbase=NULL;
 Object[ObjectCount].NoVertices=0;
 Object[ObjectCount].NoFaces=0;
 Object[ObjectCount].NoMaps=0;
 Object[ObjectCount].Origin[0]=0;   /* initialize in case no axis */
 Object[ObjectCount].Origin[1]=0;   /* present */
 Object[ObjectCount].Origin[2]=0;
 Object[ObjectCount].Axis[0].bo[0]=0;
 Object[ObjectCount].Axis[0].bo[1]=0;
 Object[ObjectCount].Axis[0].bo[2]=0;
 Object[ObjectCount].Axis[0].bu[0]=UNIT;
 Object[ObjectCount].Axis[0].bu[1]=0;
 Object[ObjectCount].Axis[0].bu[2]=0;
 Object[ObjectCount].Axis[0].bv[0]=0;
 Object[ObjectCount].Axis[0].bv[1]=UNIT;
 Object[ObjectCount].Axis[0].bv[2]=0;
 VECCOPY(Object[ObjectCount].Axis[0].bo,Object[ObjectCount].Axis[0].boin)
 VECCOPY(Object[ObjectCount].Axis[0].bu,Object[ObjectCount].Axis[0].buin)
 VECCOPY(Object[ObjectCount].Axis[0].bv,Object[ObjectCount].Axis[0].bvin)
#if !__DLL__
 if(GetMemoryObject(FileName,(Object+ObjectCount),
    &Skids,&Skeleton,&nSkeleton))goto XX1;
#endif
 fp=CreateFile(FileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(fp == INVALID_HANDLE_VALUE){
   return 0;
 }
 newface=0; newvert=0;
 ReadFile(fp,str,4,&dwRead,NULL);
 if(strcmp(str,"FORM") != 0){
   return 0;
 }
 ReadFile(fp,&FORMsize,sizeof(long),&dwRead,NULL);
 ReadFile(fp,str,4,&dwRead,NULL);
 if(dwRead != 4 || (strcmp(str,"OFXM") != 0 && strcmp(str,"AAPO") != 0)){
   return 0;
 }
loop:
 if(!ReadFile(fp,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4)goto end;
 if (R_getl(&CHUNKsize) != 1)goto end;
 if     (strcmp(str,"VERT") == 0){if(R_ReadVertices(CHUNKsize)   < 0)return -1;}
 else if(strcmp(str,"VTEX") == 0){if(R_ReadVertText(CHUNKsize)   < 0)return -1;}
 else if(strcmp(str,"SFCE") == 0){if(R_ReadFaces(CHUNKsize,TRUE) < 0)return -1;}
 else if(strcmp(str,"LFCE") == 0){if(R_ReadFaces(CHUNKsize,FALSE)< 0)return -1;}
 else if(strcmp(str,"FACE") == 0){if(R_ReadOldFaces(CHUNKsize,0) < 0)return -1;}
 else if(strcmp(str,"SFAC") == 0){if(R_ReadOldFaces(CHUNKsize,1) < 0)return -1;}
 else if(strcmp(str,"AXIS") == 0){if(R_ReadAxis(CHUNKsize)       < 0)return -1;}
 else if(strcmp(str,"TEXT") == 0){
   if(R_ReadOldText(CHUNKsize,frame,FileName,str) < 0)return -1;
 }
 else if(strcmp(str,"IMAP") == 0){
   R_ReadText(CHUNKsize,frame,FileName,str);
 }
 else if(strcmp(str,"SK02") == 0 && robo == YES){
   Skids=RenderReadSkeleton(CHUNKsize,2,&Skeleton,&nSkeleton);
 }
 else if(strcmp(str,"SK01") == 0 && robo == YES){
   Skids=RenderReadSkeleton(CHUNKsize,1,&Skeleton,&nSkeleton);
 }
 else if(strcmp(str,"SKEL") == 0 && robo == YES){
   Skids=RenderReadSkeleton(CHUNKsize,0,&Skeleton,&nSkeleton);
 }
 else R_getchunk(CHUNKsize);
 goto loop;
end:
 CloseHandle(fp);
#if !__DLL__
 MainVp=Object[ObjectCount].Vbase;
 MainFp=Object[ObjectCount].Fbase;
 Nvert=Object[ObjectCount].NoVertices;
 Nface=Object[ObjectCount].NoFaces;
 ListAdjFaces();
 FixUpOrientation();
 StripFaces();
 PutMemoryObject(FileName,(Object+ObjectCount),Skids,Skeleton,nSkeleton);
 XX1:
#endif
 if(robo == YES && Skids != NULL && Skeleton != NULL)
   RenderTransformToRobot(ObjectCount,
                           Skids,Skeleton,
                           robo_ratio,
                           robo_file1,robo_file2);
 return 1;
}

void  R_getchunk(long clen){
 SetFilePointer(fp,clen,NULL,FILE_CURRENT);
}

static short R_ReadAxis(long clen){
 long j;
 R_getl(&Object[ObjectCount].Origin[0]); /* origin of Object (in object coords */
 R_getl(&Object[ObjectCount].Origin[1]);
 R_getl(&Object[ObjectCount].Origin[2]);
 R_getl(&j); Object[ObjectCount].Axis[0].bo[0]=j;  /* texture axis point */
 R_getl(&j); Object[ObjectCount].Axis[0].bo[1]=j;
 R_getl(&j); Object[ObjectCount].Axis[0].bo[2]=j;
 R_getl(&j); Object[ObjectCount].Axis[0].bu[0]=j;  /* texture axis off x */
 R_getl(&j); Object[ObjectCount].Axis[0].bu[1]=j;
 R_getl(&j); Object[ObjectCount].Axis[0].bu[2]=j;
 R_getl(&j); Object[ObjectCount].Axis[0].bv[0]=j;  /* texture axis off y */
 R_getl(&j); Object[ObjectCount].Axis[0].bv[1]=j;
 R_getl(&j); Object[ObjectCount].Axis[0].bv[2]=j;
 VECCOPY(Object[ObjectCount].Axis[0].bo,Object[ObjectCount].Axis[0].boin)
 VECCOPY(Object[ObjectCount].Axis[0].bu,Object[ObjectCount].Axis[0].buin)
 VECCOPY(Object[ObjectCount].Axis[0].bv,Object[ObjectCount].Axis[0].bvin)
 return 0;
}

static short R_ReadOldFaces(long  clen, long shader){
  long i,sf,se,sp;
  long id,mm,me,mp;
  face  *f;
  unsigned char *buffer=NULL,*b;
  if(shader)newface=clen/21;
  else      newface=clen/20;
  if(newface == 0)return -1;
  sf=sizeof(face);
  mm=newface*(long)sf;
  Object[ObjectCount].Fbase=(face *)X__Malloc(mm);
  Object[ObjectCount].NoFaces=newface;
  f=Object[ObjectCount].Fbase;
  if(f == NULL){return -1;}
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  for(i=0;i<newface;i++){
    f->A[0] = f->A[1] = f->A[2] = -1; f->next = -1;
    f->vid = -1; f->start=0;
    f->V[0]=(long)GetLongFromBuffer(b); b+= 4;
    f->V[1]=(long)GetLongFromBuffer(b); b+= 4;
    f->V[2]=(long)GetLongFromBuffer(b); b+= 4;
    memcpy(f->color,b,3); b+=3;
    f->texture = *b++;
    f->matsp   = *b++;
    f->mattr   = *b++;
    f->matre   = *b++;
    f->brush   = *b++;
    if(shader){
      f->x = *b++;
    }
    else  f->x=0;
    f->texture &= 0xe0;  /* keep smoothing and Brill and Map id */
    f->texco[0] = (f->matsp & 0x0f) << 4;
    f->texco[1] = (f->mattr & 0x0f) << 4;
    f->texco[2] = (f->matre & 0x0f) << 4;
    f->matsp  = (f->matsp & 0xf0) >> 4;   /* 0 - 15 posssible specs */
    f->mattr &= 0xf0;
    f->matre &= 0xf0;
    if((f->texture & 0x40) == 0x40)f->MapID = (short)(f->brush & 0x1f);
    f++;
  }
  X__Free(buffer);
  return 0;
}

static short R_ReadFaces(long  clen, BOOL short_form){
  long i,sf,se,sp;
  long id,mm,me,mp;
  unsigned short ids;
  unsigned char dummy;
  unsigned char *buffer=NULL,*b;
  face  *f;
  if(short_form)newface=clen/18;
  else          newface=clen/24;
  if(newface == 0)return -1;
  sf=sizeof(face);
  mm=newface*(long)sf;
  Object[ObjectCount].Fbase=(face *)X__Malloc(mm);
  Object[ObjectCount].NoFaces=newface;
  f=Object[ObjectCount].Fbase;
  if(f == NULL){return -1;}
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  for(i=0;i<newface;i++){
    f->A[0] = f->A[1] = f->A[2] = -1; f->next = -1;
    f->vid = -1; f->start=0;
    if(short_form){
      f->V[0]=(long)GetUshortFromBuffer(b); b+= 2;
      f->V[1]=(long)GetUshortFromBuffer(b); b+= 2;
      f->V[2]=(long)GetUshortFromBuffer(b); b+= 2;
    }
    else{
      f->V[0]=(long)GetLongFromBuffer(b); b+= 4;
      f->V[1]=(long)GetLongFromBuffer(b); b+= 4;
      f->V[2]=(long)GetLongFromBuffer(b); b+= 4;
    }
    memcpy(f->color,b,3); b+=3;
    f->matsp   = *b++;
    f->mattr   = *b++;
    f->matre   = *b++;
    f->texture = *b++;
    dummy      = *b++;
    f->brush   = *b++;
    dummy      = *b++;
    f->x       = *b++;
    dummy      = *b++;
    f->texture &= 0xe0;  /* keep smoothing and Brill and Map id */
    f->texco[0] = (f->matsp & 0x0f) << 4;
    f->texco[1] = (f->mattr & 0x0f) << 4;
    f->texco[2] = (f->matre & 0x0f) << 4;
    f->matsp  = (f->matsp & 0xf0) >> 4;   /* 0 - 15 posssible specs */
    f->mattr &= 0xf0;
    if(f->mattr > 0){ // set old texture to be glass
      f->texture |= 0x04;
    }
    f->matre &= 0xf0;
    if((f->texture & 0x40) == 0x40)f->MapID = (short)(f->brush & 0x1f);
    f++;
  }
  X__Free(buffer);
  return 0;
}

static short R_ReadVertices(long clen){
  vertex  *v;
  long i,xx,mm,sv;
  unsigned char *buffer=NULL,*b;
  newvert=clen/12;
  if(newvert == 0)return -1;
  sv=sizeof(vertex);
  mm=newvert*(long)sv;
  if(newvert > 65535L && sizeof(long) < 4){return -1;}
  Object[ObjectCount].Vbase=(vertex *)X__Malloc(mm);
  Object[ObjectCount].NoVertices=newvert;
  v=Object[ObjectCount].Vbase;
  if(v == NULL){ return -1; }
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  for(i=0;i<newvert;i++){
    v->nf=0; v->flist=NULL;
    v->p[0]=(double)GetLongFromBuffer(b); b+= 4;
    v->p[1]=(double)GetLongFromBuffer(b); b+= 4;
    v->p[2]=(double)GetLongFromBuffer(b); b+= 4;
    v++;
  }
  X__Free(buffer);
  return 0;
}

#ifdef _SUNSTYLE
static short R_ReadVertText(clen) long clen; {
#else
static short R_ReadVertText(long clen){
#endif
  vertex *t;
  long i,n,xx,id;
  n=clen/12;
  i=Object[ObjectCount].NoVertices;
  if(n == 0 || n > i)return -1;
  t=Object[ObjectCount].Vbase;
  if(t == NULL){ return -1; }
  for(i=0;i<n;i++){
    R_getl(&id);
    R_getl(&xx); t[id].x=(double)xx*0.000001;
    R_getl(&xx); t[id].y=(double)xx*0.000001;
  }
  return 0;
}

static short R_ReadOldText(long clen, long frame, char *objectfilename,
                        char *str){
  iMap *temp_map,*Map;
  unsigned char t_colour[3];
  char i,j,*cp1,*cp2,brushmaptype,brushmethod,moziac,p,r,b;
  char fsurf[256],frefl[256],fbump[256],tempstr[256];
  long k,l,m,firstframe,framedelay,lastframe,thisframe,brushangle,file_len;
  short s,version;
  ReadFile(fp,&j,sizeof(char),&dwRead,NULL);
  if(j > 0){
    temp_map = (iMap *)X__Malloc(j*sizeof(iMap));
    if(temp_map == NULL){
      MessageBox(NULL,"Out of Memory",NULL,MB_OK);
      return -1;
    }
    Object[ObjectCount].Mbase = temp_map;
    for(i=0;i<j;i++){
      Map=(temp_map+i);
      version=0;
      R_getl(&file_len);
      ReadFile(fp,tempstr,file_len,&dwRead,NULL);
      R_getl(&l); Map->p[0]= (double)l;
      R_getl(&l); Map->p[1]= (double)l;
      R_getl(&l); Map->p[2]= (double)l;
      R_getl(&l); Map->x[0]= (double)l;
      R_getl(&l); Map->x[1]= (double)l;
      R_getl(&l); Map->x[2]= (double)l;
      R_getl(&l); Map->y[0]= (double)l;
      R_getl(&l); Map->y[1]= (double)l;
      R_getl(&l); Map->y[2]= (double)l;
      R_getl(&l); brushmethod=(char)l;   /* still bump etc from old version */
      R_getl(&firstframe); R_getl(&framedelay); R_getl(&lastframe);
      R_getl(&l); /* used by skeleton for br. assign. not used by renderer */
      R_getl(&l); brushmaptype=(char)l;
      R_getl(&l); /* this is the expansion count */
      if(l > 0){
        R_getl(&brushangle);
        if(l > 1){
          for(k=1;k<5;k++)R_getl(&m); /* dump the brush lock info */
          if(l > 5){             /* this must be a version 2 file */
            version=1;
            ReadFile(fp,tempstr,32,&dwRead,NULL);
            ReadFile(fp,&p,1,&dwRead,NULL);
            ReadFile(fp,&r,1,&dwRead,NULL);
            ReadFile(fp,&b,1,&dwRead,NULL);
            R_getsht(&s); Map->pp=s; /* these will be set to -1 */
            R_getsht(&s); Map->rp=s; /* if a map of this type   */
            R_getsht(&s); Map->bp=s; /* is not used in Material */
            ReadFile(fp,t_colour,3,&dwRead,NULL);  /* not used yet */
            ReadFile(fp,t_colour,3,&dwRead,NULL);
            ReadFile(fp,t_colour,3,&dwRead,NULL);  /* ------------ */
            if(p)ReadFile(fp,fsurf,file_len,&dwRead,NULL);
            if(r)ReadFile(fp,frefl,file_len,&dwRead,NULL);
            if(b)ReadFile(fp,fbump,file_len,&dwRead,NULL);
            R_getsht(&s); /* should be 0 use as expansion count */
            if(s > 0)for(k=0;k<s;k++)R_getl(&m);
          }
        }
      }
      else brushangle=360;
      moziac=NO;
      if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
      if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
      if(version == 0){
        p=r=b=0;
        Map->pp=100;
        Map->rp= 50;
        Map->bp= 50;
        if(brushmethod == REFLBRUSH){
          strcpy(frefl,tempstr); r=1;
        }
        else if(brushmethod == BUMPBRUSH){
          strcpy(fbump,tempstr); b=1;
        }
        else{
          strcpy(fsurf,tempstr); p=1;
        }
        if(brushmethod != ANIMBRUSH)brushmethod=STILBRUSH; /* just switch now */
      }
      if(p && brushmethod == ANIMBRUSH){ /* animated brush calculate frame */
        cp1=strchr(fsurf,'0');
        cp2=strchr(fsurf,'.');
        if(cp1 == NULL || cp2 == NULL){
          sprintf(tempstr,"dummy.evi");
        }
        else{
          int ln;
          char *cp,extn[16],fmat[16];
          strcpy(extn,cp2);
          ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
          *cp1='\0';
          if(framedelay == 0)thisframe=frame+firstframe-1;
          else               thisframe=frame/(framedelay+1)+firstframe;
          while(thisframe > lastframe)thisframe -= lastframe;
          sprintf(fmat,"%%s%%0%ldld%%s",ln);
          sprintf(tempstr,fmat,fsurf,thisframe,extn);
        }
        strcpy(fsurf,tempstr);
      }
      if(0){
        if(p) strcpy(Map->filename,short_form(fsurf));
        else         Map->pp = -1;
        if(r) strcpy(Map->frefl,short_form(frefl));
        else         Map->rp = -1;
        if(b) strcpy(Map->fbump,short_form(fbump));
        else         Map->bp = -1;
      }
      else{
        if(p) strcpy(Map->filename,fsurf);
        else         Map->pp = -1;
        if(r) strcpy(Map->frefl,frefl);
        else         Map->rp = -1;
        if(b) strcpy(Map->fbump,fbump);
        else         Map->bp = -1;
      }
      Object[ObjectCount].NoMaps++;
      Map->map=(short)brushmaptype;
      Map->method=(short)brushmethod;
      Map->angle=360.0/(double)brushangle;
      Map->moziac=moziac;
      Map->ix=0; Map->ix=0;  /* no image map is loaded if not      */
      Map->pixels=NULL;      /* DLL then this will be loaded later */
    }
  }
  return 0;
}

static short R_ReadText(long clen, long frame, char *objectfilename,
                        char *str){ //ignores transmissivity
  iMap *temp_map,*Map;
  unsigned char t_colour[3];
  char *cp1,*cp2,brushmaptype,brushmethod,moziac,p,r,b,t,kt;
  char fsurf[256],frefl[256],fbump[256],tempstr[256];
  long i,j,k,l,m,firstframe,framedelay,lastframe,thisframe,
       brushangle,file_len;
  short s,version;
  R_getl(&j);
  if(j > 0){
    temp_map = (iMap *)X__Malloc(j*sizeof(iMap));
    if(temp_map == NULL){
      MessageBox(NULL,"Out of Memory",NULL,MB_OK);
      return -1;
    }
    Object[ObjectCount].Mbase = temp_map;
    for(i=0;i<j;i++){
      Map=(temp_map+i);
      version=0;
      R_getl(&l); Map->p[0]= (double)l;
      R_getl(&l); Map->p[1]= (double)l;
      R_getl(&l); Map->p[2]= (double)l;
      R_getl(&l); Map->x[0]= (double)l;
      R_getl(&l); Map->x[1]= (double)l;
      R_getl(&l); Map->x[2]= (double)l;
      R_getl(&l); Map->y[0]= (double)l;
      R_getl(&l); Map->y[1]= (double)l;
      R_getl(&l); Map->y[2]= (double)l;
      R_getl(&l); brushmethod=(char)l;   /* still bump etc from old version */
      R_getl(&firstframe); R_getl(&framedelay); R_getl(&lastframe);
      R_getl(&l);  /* refl type */
      R_getl(&l);  R_getl(&l); R_getl(&l);
      R_getl(&l);  /* bump type */
      R_getl(&l);  R_getl(&l); R_getl(&l);
      R_getl(&l);  /* trans type */
      R_getl(&l);  R_getl(&l); R_getl(&l);
      R_getl(&l); /* used by skeleton for br. assign. not used by renderer */
      R_getl(&l); brushmaptype=(char)l;
      R_getl(&brushangle);
      for(k=0;k<4;k++)R_getl(&m); /* dump the brush lock info */
      ReadFile(fp,tempstr,32,&dwRead,NULL);
      ReadFile(fp,&p,1,&dwRead,NULL);
      ReadFile(fp,&r,1,&dwRead,NULL);
      ReadFile(fp,&b,1,&dwRead,NULL);
      ReadFile(fp,&t,1,&dwRead,NULL);
      ReadFile(fp,&kt,1,&dwRead,NULL);
      R_getsht(&s); Map->pp=s; /* these will be set to -1 */
      R_getsht(&s); Map->rp=s; /* if a map of this type   */
      R_getsht(&s); Map->bp=s; /* is not used in Material */
      R_getsht(&s); // trans map
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,fsurf,256,&dwRead,NULL);
      ReadFile(fp,frefl,256,&dwRead,NULL);
      ReadFile(fp,fbump,256,&dwRead,NULL);
      ReadFile(fp,tempstr,256,&dwRead,NULL);
      moziac=NO;
      if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
      if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
      if(p && brushmethod == ANIMBRUSH){ /* animated brush calculate frame */
        cp1=strchr(fsurf,'0');
        cp2=strchr(fsurf,'.');
        if(cp1 == NULL || cp2 == NULL){
          sprintf(tempstr,"dummy.evi");
        }
        else{
          int ln;
          char *cp,extn[16],fmat[16];
          strcpy(extn,cp2);
          ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
          *cp1='\0';
          if(framedelay == 0)thisframe=frame+firstframe-1;
          else               thisframe=frame/(framedelay+1)+firstframe;
          while(thisframe > lastframe)thisframe -= lastframe;
          sprintf(fmat,"%%s%%0%ldld%%s",ln);
          sprintf(tempstr,fmat,fsurf,thisframe,extn);
        }
        strcpy(fsurf,tempstr);
      }
      if(0){
        if(p) strcpy(Map->filename,short_form(fsurf));
        else         Map->pp = -1;
        if(r) strcpy(Map->frefl,short_form(frefl));
        else         Map->rp = -1;
        if(b) strcpy(Map->fbump,short_form(fbump));
        else         Map->bp = -1;
      }
      else{
        if(p) strcpy(Map->filename,fsurf);
        else         Map->pp = -1;
        if(r) strcpy(Map->frefl,frefl);
        else         Map->rp = -1;
        if(b) strcpy(Map->fbump,fbump);
        else         Map->bp = -1;
      }
      Object[ObjectCount].NoMaps++;
      Map->map=(short)brushmaptype;
      Map->method=(short)brushmethod;
      Map->angle=360.0/(double)brushangle;
      Map->moziac=moziac;
      Map->ix=0; Map->ix=0;  /* no image map is loaded if not      */
      Map->pixels=NULL;      /* DLL then this will be loaded later */
    }
  }
  return 0;
}

#define EPSILON 1.0e-6

#ifdef _SUNSTYLE
static void InvertRotationMatrix(T,t); double T[4][4]; double t[4][4]; {
#else
static void InvertRotationMatrix(double T[4][4], double t[4][4]){
#endif
 int i,j;
 double a[3][3];
 double det;
 R_null_transform(t);
 for(i=0;i<3;i++)for(j=0;j<3;j++)a[i][j]=T[i][j];
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
}

#ifdef _SUNSTYLE
#else
static void MatrixToQuaternion(double m[4][4], quaternion q){
#endif
 double w4,x2;
 q[0]=0.25*(1.0+m[0][0]+m[1][1]+m[2][2]);
 if(q[0] > EPSILON){
   q[0]=sqrt(q[0]);
   w4=1.0/(4.0*q[0]);
   q[1]=(m[1][2]-m[2][1])*w4;
   q[2]=(m[2][0]-m[0][2])*w4;
   q[3]=(m[0][1]-m[1][0])*w4;
 }
 else{
   q[0]=0.0;
   q[1]= -0.5*(m[1][1]+m[2][2]);
   if(q[1] > EPSILON){
     q[1]=sqrt(q[1]);
     x2=1.0/(2.0*q[1]);
     q[2]=m[0][1]*x2;
     q[3]=m[0][2]*x2;
   }
   else{
     q[1]=0.0;
     q[2]=0.5*(1.0-m[2][2]);
     if(q[2] > EPSILON){
       q[2]=sqrt(q[2]);
       q[3]=m[1][2]/(2.0*q[2]);
     }
     else{
       q[2]=0.0;
       q[3]=1.0;
     }
   }
 }
 return;
}

#ifdef _SUNSTYLE
#else
static void QuaternionToMatrix(quaternion q, double m[4][4]){
#endif
 double w,x,y,z,x2,y2,z2;
 w=q[0]; x=q[1]; y=q[2]; z=q[3]; x2=x*x; y2=y*y; z2=z*z;
 m[0][0]=1.0-2*y2-2*z2; m[0][1]=2*x*y+2*w*z;   m[0][2]=2*x*z-2*w*y;
 m[1][0]=2*x*y-2*w*z;   m[1][1]=1.0-2*x2-2*z2; m[1][2]=2*y*z+2*w*x;
 m[2][0]=2*x*z+2*w*y;   m[2][1]=2*y*z-2*w*x;   m[2][2]=1.0-2*x2-2*y2;
 m[3][3]=1.0; m[0][3]=m[1][3]=m[2][3]=m[3][0]=m[3][1]=m[3][2]=0.0;
 return;
}

#ifdef _SUNSTYLE
#else
static void Slerp(double mr, quaternion q1, quaternion q2, quaternion q){
#endif
 double sum,beta1,beta2,theta;
 sum=q1[0]*q2[0]+q1[1]*q2[1]+q1[2]*q2[2]+q1[3]*q2[3];
 if(sum < 0.0)sum=0.0;
 if(sum > 1.0)sum=1.0;
 theta=acos(sum);
 if(theta < EPSILON){ beta1=1.0-mr; beta2=mr;}
 else{
  beta1=sin((1.0-mr)*theta)/sin(theta);
  beta2=sin(mr*theta)/sin(theta);
 }
 q[0]=beta1*q1[0]+beta2*q2[0];
 q[1]=beta1*q1[1]+beta2*q2[1];
 q[2]=beta1*q1[2]+beta2*q2[2];
 q[3]=beta1*q1[3]+beta2*q2[3];
}

static int nSkeleton;
skel   *Skeleton;

#ifdef _SUNSTYLE
#else
static void ApplyTransformToChildren(int id, double T[4][4]){
#endif
 static double x,y,z;
 static double sst[4][4],t1[4][4],t2[4][4];
 int i;
 for(i=1;i<nSkeleton;i++){ /* must start at one since root is never a child */
   if(Skeleton[i].id == id){
     R_tram(t1,-(Skeleton+id)->p[0],-(Skeleton+id)->p[1],-(Skeleton+id)->p[2]);
     R_m4by4((Skeleton+i)->Q,t1,t2);
     R_tram(t1, (Skeleton+id)->p[0], (Skeleton+id)->p[1], (Skeleton+id)->p[2]);
     R_m4by4(t1,t2,sst);
     R_m4by4(T,sst,Skeleton[i].T);
     ApplyTransformToChildren(i,Skeleton[i].T);
   }
 }
 return;
}

#ifdef _SUNSTYLE
#else
static void InterpolateRobot(double mr, long n, skel *sk1, skel *sk2){
#endif
 int i;
 quaternion q,q1,q2;
 nSkeleton=n;
 Skeleton=sk1;
 if(mr < 1.0){
   quaternion Qi={1.0,0.0,0.0,0.0};
   double Q[4][4],Ri[4][4];
   for(i=0;i<nSkeleton;i++){
     InvertRotationMatrix((sk2+i)->R,Ri);
     R_m4by4((Skeleton+i)->R,Ri,Q);
     MatrixToQuaternion(Q,q2);
     Slerp(mr,Qi,q2,q);
     QuaternionToMatrix(q,Q);
     R_m4by4(Q,(sk2+i)->R,(Skeleton+i)->Q);
   }
 }
 else{
   for(i=0;i<nSkeleton;i++)R_c4to4((Skeleton+i)->R,(Skeleton+i)->Q);
 }
 ApplyTransformToChildren(0,Skeleton[0].T);
 return;
}

static void RenderTransformToRobot(long O,
                             long *Skid, skel *Skel, double mr,
                             char *robo_f1, char *robo_f2){
  long i,j,n1=0,n2=0;
  FILE *frobot;
  skel *sk1=NULL,*sk2=NULL;
  vertex *v;
  double x1,y1,z1;
  if((frobot = fopen(robo_f1,"r")) != NULL){
    fscanf(frobot,"%ld\n",&n1);
    if(n1 > 0){  /* sk1 is the Robot  sk2 is version is moving from */
      if((sk1=Skel) == NULL)goto FINISHED;
      for(i=0;i<n1;i++){
        R_null_transform(sk1->R);
        R_null_transform(sk1->Q);
        R_null_transform(sk1->T);
        for(j=0;j<4;j++)fscanf(frobot,GET_MATX,
         &sk1[i].R[j][0], &sk1[i].R[j][1],
         &sk1[i].R[j][2], &sk1[i].R[j][3]);
      }
    }
    fclose(frobot);
  }
  if((frobot = fopen(robo_f2,"r")) != NULL){
    fscanf(frobot,"%ld\n",&n2);
    if(n2 > 0){
      if((sk2 = (skel *)X__Malloc(n2*(long)sizeof(skel))) == NULL)
           goto FINISHED;
      for(i=0;i<n2;i++){
        R_null_transform(sk2->R);
        R_null_transform(sk2->Q);
        R_null_transform(sk2->T);
        for(j=0;j<4;j++)fscanf(frobot,GET_MATX,
         &sk2[i].R[j][0], &sk2[i].R[j][1],
         &sk2[i].R[j][2], &sk2[i].R[j][3]);
      }
    }
    fclose(frobot);
  }
  if(n1 != n2 || n1 == 0 || n2 == 0)goto FINISHED;
  InterpolateRobot(mr,n1,sk1,sk2);
  v=Object[O].Vbase;
  for(i=0;i<Object[O].NoVertices;i++){
    R_m4by1(sk1[Skid[i]].T,v[i].p[0],v[i].p[1],v[i].p[2],&x1,&y1,&z1);
    v[i].p[0]=x1;
    v[i].p[1]=y1;
    v[i].p[2]=z1;
  }
  FINISHED:
  if(sk1 != NULL)X__Free(sk1);
  if(sk2 != NULL)X__Free(sk2);
  X__Free(Skid);
  return;
}



