/* file MAPS.C  included in SOLID.C                     */

/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* this file contains routines to handle mapping        */
/* loading maps etc.                                    */
/* It also contains the code for the Preferences dialog */

static void CheckMapPath(char *fn){
 long l;
 char *fs;
 FILE *Bit_File=NULL;
 char pref[32],preff[32],temp_name[255];
 if((Bit_File=fopen((char *)fn,"rb")) != NULL){
   fclose(Bit_File);
   return;
 }
 fs=short_form(fn);
 strcpy(preff,"MAP-PATH");
 for(l=1;l<=128;l++){ /*  try .INI settings for models and maps*/
   sprintf(pref,"%s%ld",preff,l);
   GetPrivateProfileString("RPREFERENCES",pref,"$$$",
                           temp_name,255,iniFilename);
   if(strncmp(temp_name,"$$$",3) == 0)break;
   if(temp_name[1] != ':')continue;
   if(temp_name[strlen(temp_name)-1] != '\\')strcat(temp_name,"\\");
   strcat(temp_name,fs);
   if((Bit_File=fopen((char *)temp_name,"rb")) != NULL){
     fclose(Bit_File);
     strcpy(fn,temp_name);
     return;
   }
 }
}

unsigned char *LoadMAP(char *InFile, int *xx, int *yy){
 int ff;
 char CurrentFile[255];
 strcpy(CurrentFile,InFile);
 CheckMapPath(CurrentFile);
 if     (strstr(CurrentFile,".GIF") || strstr(CurrentFile,".gif"))
   return LoadGIF(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".TGA") || strstr(CurrentFile,".tga"))
   return LoadTGA(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".JPG") || strstr(CurrentFile,".jpg"))
   return LoadXXX(CurrentFile,xx,yy,0);
 else if(strstr(CurrentFile,".TIF") || strstr(CurrentFile,".tif"))
   return LoadXXX(CurrentFile,xx,yy,1);
 else if(strstr(CurrentFile,".PNG") || strstr(CurrentFile,".png"))
   return LoadXXX(CurrentFile,xx,yy,2);
 return NULL;
}

static void makeNextMap(unsigned char *pixels24, int x, int y){
 int xi,yi;
 unsigned char *px;
 if     (x <  96)xi=64;
 else if(x < 192)xi=128;
 else if(x < 384)xi=256;
 else xi=256;
 if     (y <  96)yi=64;
 else if(y < 192)yi=128;
 else if(y < 384)yi=256;
 else yi=256;
 if((px=(unsigned char *)X__Malloc(xi*yi*3)) == NULL){
   X__Free((char *)pixels24);
   return;
 }
 gluScaleImage(GL_RGB,x,y,GL_UNSIGNED_BYTE,pixels24,xi,yi,GL_UNSIGNED_BYTE,px);
 X__Free((char *)pixels24);
 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glTexImage2D(GL_TEXTURE_2D,0,3,xi,yi,0,GL_RGB,GL_UNSIGNED_BYTE,(GLvoid *)px);
 X__Free((char *)px);
}

unsigned char *MakeMapImage(unsigned char *pixels24, int x, int y,
                       int *xi, int *yi){
 unsigned char *px;
 if     (x < 128)*xi=64;
 else *xi=128;
 if     (y < 128)*yi=64;
 else *yi=128;
 if((px=(unsigned char *)X__Malloc(*xi * *yi * 3)) == NULL){
   X__Free((char *)pixels24);
   return NULL;
 }
 gluScaleImage(GL_RGB,x,y,GL_UNSIGNED_BYTE,pixels24,
               *xi,*yi,GL_UNSIGNED_BYTE,px);
 X__Free((char *)pixels24);
 return px;
}

void MakeMapFromImage(unsigned char *px, long xi, long yi){
 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glTexImage2D(GL_TEXTURE_2D,0,3,xi,yi,0,GL_RGB,GL_UNSIGNED_BYTE,(GLvoid *)px);
}

static void makeCheckeredImageMap(void){
 GLfloat parameter3[]={GL_SPHERE_MAP};
 int i, j, c;
 for (i = 0; i < checkImageWidth; i++) {
   for (j = 0; j < checkImageHeight; j++) {
     c = ((((i&0x8)==0)^((j&0x8))==0))*255;
     checkImage[i][j][0] = (GLubyte) c;
     checkImage[i][j][1] = (GLubyte) c;
     checkImage[i][j][2] = (GLubyte) c;
   }
 }
 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glPixelStorei(GL_PACK_ALIGNMENT,1);
 glTexImage2D(GL_TEXTURE_2D, 0, 3, checkImageWidth,
              checkImageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
              &checkImage[0][0][0]);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
 if(ffastmapping){
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);
 }
 else{
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
 }
// glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
 glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
// glTexGenfv(GL_S,GL_SPHERE_MAP,0); /* for reflection map */
// glTexGenfv(GL_T,GL_SPHERE_MAP,0); /*   ""      ""       */
 glTexGenfv(GL_S,GL_TEXTURE_GEN_MODE,parameter3);
 glTexGenfv(GL_T,GL_TEXTURE_GEN_MODE,parameter3);
}

#define TOL 1000.0

static void GetMapNormal(int type, vector P, vector X, vector Y,
                         vector mP, vector mX, vector mY,
                         vector mN){
 GLfloat n[3];
 if(type == CYLINDER || type == CYLINDER_MOZIAC){
   VECCOPY((double)P,mP)
   VECSUB((double)X,mP,mX)
   VECSUB((double)Y,mP,mY)
   CROSS(mY,mX,mN)
 }
 else{                                      /* arranged this way to */
   VECCOPY((double)Y,mP)                    /* have map with origin */
   VECSUB((double)X,(double)P,mX)           /* at bottom left       */
   VECSUB((double)P,mP,mY)
   CROSS(mX,mY,mN)
 }
 VECCOPY((GLfloat)mN,n)
 if(!normalise(n))MessageBeep(MB_OK);
 VECCOPY((double)n,mN)
 return;
}

static void GetMappingCoordP(vector n, vector x, vector y, vector p0,
                             vector pv,/* point of intersection */
                             GLfloat *a, GLfloat *b){
 vector v1,p;
 GLfloat mu,det,ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3;
 VECSUB(p0,pv,v1)
 mu=DOT(n,v1);
 VECSCALE(mu,n,v1)
 VECSUM(pv,v1,p)
 ve1=x[0]; ve2=x[1]; ve3=x[2];
 vp1=y[0]; vp2=y[1]; vp3=y[2];
 p1=p[0]-p0[0];
 p2=p[1]-p0[1];
 p3=p[2]-p0[2];
 det=ve1*vp2-vp1*ve2;
 if(det > TOL || det < -TOL)  /* e-10 appears not to work */
   {  *a=( vp2*p1-vp1*p2)/det;
      *b=(-ve2*p1+ve1*p2)/det;
      goto HIT;
   }
 det=ve1*vp3-vp1*ve3;
 if(det > TOL || det < -TOL)
   {  *a=( vp3*p1-vp1*p3)/det;
      *b=(-ve3*p1+ve1*p3)/det;
      goto HIT;
   }
 det=ve2*vp3-vp2*ve3;
 if(det > TOL || det < -TOL)
   {  *a=( vp3*p2-vp2*p3)/det;
      *b=(-ve3*p2+ve2*p3)/det;
      goto HIT;
   }
 *a = *b = 0.0;
 HIT:
 return;
}

#define RTD 0.159154943

static void GetMappingCoordC(vector n, vector x, vector y, vector p0,
                double angle,
                vector pv,
                GLfloat *a, GLfloat *b){
 GLfloat theta,nn;
 GLfloat p1[3],p2[3],q[3];
 VECSUB((GLfloat)p0,(GLfloat)pv,p1)
 *b=DOT(p1,y)/DOT(y,y);
 VECSCALE(*b,y,p2)
 VECSUB(p1,p2,q)
 normalise(q);
 nn=DOT(q,(GLfloat)n);
 if     (nn >  1.0)nn =  1.0;
 else if(nn < -1.0)nn = -1.0;
 theta=acos(nn)*RTD;
 CROSS(q,n,p1)
 if(DOT(p1,y) <= 0.0){
   if(theta <= 0.25) *a = (0.25 - theta);
   else  *a = (1.25 - theta);
 }
 else    *a = (0.25 + theta);
 *a *= (GLfloat)angle;
 return;
}

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size,gif_iLace;

static int rowcount,rowinc;
static unsigned char *pixels = NULL;

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(y_size - 1 - rowcount)*(long)(x_size);
 memcpy(pixels + l, linepixels, linelen);
 rowcount+=abs(rowinc);
 if(rowcount >= y_size){
  if     (rowinc == -8){rowinc=8; rowcount=4;}
  else if(rowinc ==  8){rowinc=4; rowcount=2;}
  else if(rowinc ==  4){rowinc=2; rowcount=1;}
 }
 return 0;
}

static unsigned char *LoadGIF(char *GIFfile, int *xx, int *yy){
 unsigned char *px,*lpB,*pixels24;
 long i,j,k,imagesize,x,y,xi,yi;
 rowcount=0;
 rowinc=1;
 if(read_gif_image(GIFfile,1) < 0)return 0;
 if(gif_iLace)rowinc = -8;
 x=x_size; y=y_size;
 imagesize=x_size*y_size;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return 0;
 if((pixels24=(unsigned char*)X__Malloc(imagesize*3)) == NULL){
   X__Free(pixels);
   return NULL;
 }
 if(read_gif_image(GIFfile,0) < 0){
   X__Free(pixels);
   return NULL;
 }
 px=pixels24;
 lpB=pixels;
 for(j=0;j<y;j++){
   for(i=0;i<x;i++){
     k=(long)(*lpB++);
     *px++ = (unsigned char)(p_red[k]   << 2);
     *px++ = (unsigned char)(p_green[k] << 2);
     *px++ = (unsigned char)(p_blue[k]  << 2);
   }
 }
 *xx=x; *yy=y;
 X__Free(pixels);
 return pixels24;
}

static void tga_display_line(unsigned char *prgb,
                             long wide, long line,
                             long upsidedown, long hi){
 register long l,x,k;
 if(upsidedown)l = 3L*(long)(line)*(long)(x_size);
 else          l = 3L*(long)(hi - 1 - line)*(long)(x_size);
 memcpy(pixels + l, prgb, (int)(wide*3));
}

static short R_getvariword(long n, FILE *f, long *vw){
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;          if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

static short read_tga_image(char *filename, short info_only){
 char c0,c1,c2;
 long i,j,k,l,m,n,cmo,cml,xor,yor,wi,hi,iok=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat,wi3;
 FILE *t24;
 unsigned char *prgb;
 prgb=NULL;
 if((t24=fopen(filename,"rb")) != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0 : 1 ;
   if(info_only){
     x_size=(int)wi;
     y_size=(int)hi;
     fclose(t24);
     return 1;
   }
   if((prgb=(unsigned char *)X__Malloc((size_t)wi*3)) == NULL)goto CLOSEOUT;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi;
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   wi3=wi*3;
   if(image_type == 2){ /* uncompressed */
     m=0; k=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       prgb[m+2] = (char)( vw        & pat) << s;
       prgb[m+1] = (char)((vw >> s1) & pat) << s;
       prgb[m  ] = (char)((vw >> s2) & pat) << s;
       k++; m+=3;
       if(m == wi3){
         tga_display_line(prgb,wi,i,upsidedown,hi); m=0;
       }
     }
   }
   else{  /* compressed */
     m=0; l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           prgb[m+2] = (char)( vw        & pat) << s;
           prgb[m+1] = (char)((vw >> s1) & pat) << s;
           prgb[m  ] = (char)((vw >> s2) & pat) << s;
           k++; m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           prgb[m+2] = c0;
           prgb[m+1] = c1;
           prgb[m  ] = c2;
           k++;  m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
   if(prgb != NULL)X__Free((char *)prgb);
 }
 return (short)iok;
}

static unsigned char *LoadTGA(char *TGAfile, int *xx, int *yy){
 long imagesize;
 rowcount=0;
 if(read_tga_image(TGAfile,1) < 0)return 0;
 imagesize=x_size*y_size*3;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return NULL;
 if(read_tga_image(TGAfile,0) < 0){
   X__Free(pixels);
   return NULL;
 }
 *xx=x_size; *yy=y_size;
 return pixels;
}

static unsigned char *LoadXXX(char *JPGfile, int *xx, int *yy, long type){
 unsigned char *p,*px,*R,*G,*B,*r,*g,*b;
 long i,j,imagesize;
 long x,y;
 char lname[256];
 HMODULE hLib;
 FARPROC fpFun;
 BOOL (*fpGetMapSize)(char *,
                      long *, long *);
 BOOL (*fpGetMap)(char *,
                  long, long,
                          unsigned char *, unsigned char *, unsigned char *);
 GetModuleFileName(glInstance,lname,255);
 *FileInPath(lname) = '\0';
 if     (type == 0)strcat(lname,"jpg_map.dll");
 else if(type == 1)strcat(lname,"tif_map.dll");
 else if(type == 2)strcat(lname,"png_map.dll");
 R=G=B=NULL;
 hLib=LoadLibrary(lname);
 if(hLib != NULL){
   if((fpFun=GetProcAddress(hLib,"_GetExternalImageSize")) != NULL){
     fpGetMapSize = (void *)fpFun;
     if((*fpGetMapSize)(JPGfile,&x,&y)){
       if((R = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((G = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((B = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((fpFun=GetProcAddress(hLib,"_GetExternalImage")) != NULL){
         fpGetMap = (void *)fpFun;
         if(!(*fpGetMap)(JPGfile,x,y,R,G,B))goto CLOSEOUT;
       }
       else goto CLOSEOUT;
     }
     else goto CLOSEOUT;
   }
   else goto CLOSEOUT;
   FreeLibrary(hLib);
 }
 else return NULL;
 x_size=(int)ALIGNULONG(x);
 y_size = y;
 imagesize=x_size*y_size*3;
 if((p=(unsigned char *)X__Malloc(imagesize)) == NULL){
   X__Free(R);
   X__Free(G);
   X__Free(B);
   return NULL;
 }
 for(j=0;j<y;j++){
  px=(p+(y-j-1)*x_size*3);
  r=(R+j*x); g=(G+j*x); b=(B+j*x);
  for(i=0;i<x;i++){
    *px++ = *r++;
    *px++ = *g++;
    *px++ = *b++;
  }
 }
 X__Free(R);
 X__Free(G);
 X__Free(B);
 *xx=x_size; *yy=y_size;
 return p;
 CLOSEOUT:
 if(R != NULL)X__Free(R);
 if(G != NULL)X__Free(G);
 if(B != NULL)X__Free(B);
 FreeLibrary(hLib);
 return NULL;
}

#define BFT_BITMAP 0x4d42   /* 'BM' */

static void writeScreen(void){
 HANDLE hfbm;
 DWORD dwRead;
 BITMAPFILEHEADER hdr,*lphdr;
 BITMAPINFOHEADER bfi,*lpbi;
 long i,j,imagesize,nx,mx,ny;
 unsigned char *Screen1,*p,*Screen2,*q,cr,cg,cb;
 if(SelectFileName(1,szOutFileName,szOutDirName,szAppName,
             "(*.BMP)|*.bmp|",glWnd) != TRUE)return;
 nx=oldrect.right; mx=((nx*24+31)/32) * 4;
 ny=oldrect.bottom;
 if((Screen1=(unsigned char *)X__Malloc(nx*ny*3)) == NULL){
   MessageBox(NULL,"Out of Memory","Warning",MB_OK);
   return;
 }
 if((Screen2=(unsigned char *)X__Malloc(mx*ny)) == NULL){
   MessageBox(NULL,"Out of Memory","Warning",MB_OK);
   X__Free(Screen1);
   return;
 }
 glReadPixels(0,0,(GLsizei)nx,(GLsizei)ny,
              GL_RGB,GL_UNSIGNED_BYTE,
              (GLvoid *)Screen1);
 p=Screen1;
 q=Screen2;
 for(i=0;i<ny;i++){
   p=Screen1+i*nx*3;
   q=Screen2+i*mx;
   for(j=0;j<nx;j++){
     cr = (unsigned char)*p++;
     cg = (unsigned char)*p++;
     cb = (unsigned char)*p++;
     *q++ = (unsigned char)cb;
     *q++ = (unsigned char)cg;
     *q++ = (unsigned char)cr;
   }
 }
 X__Free(Screen1);
 if((hfbm=CreateFile(szOutFileName,GENERIC_WRITE,FILE_SHARE_WRITE,
   NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,
   (HANDLE)NULL)) == INVALID_HANDLE_VALUE){
   X__Free(Screen1);
   X__Free(Screen2);
   MessageBox(NULL,szOutFileName,"File Open Fail",MB_OK);
   MessageBeep(MB_OK);
   return;
 }
 lpbi=&bfi;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)nx;
 lpbi->biHeight=(DWORD)ny;
 lpbi->biPlanes=1;
 lpbi->biBitCount=24;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage = imagesize = (DWORD)(mx*ny);
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=0;
 lpbi->biClrImportant=0;
 lphdr=&hdr;
 hdr.bfType          = BFT_BITMAP;
 hdr.bfSize          = (sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)
                        + imagesize);
 hdr.bfReserved1     = 0;
 hdr.bfReserved2     = 0;
 hdr.bfOffBits       = (DWORD)sizeof(BITMAPFILEHEADER) + lpbi->biSize;
 WriteFile(hfbm,lphdr,sizeof(BITMAPFILEHEADER),&dwRead,(LPOVERLAPPED)NULL);
 WriteFile(hfbm,lpbi,sizeof(BITMAPINFOHEADER),&dwRead,(LPOVERLAPPED)NULL);
 WriteFile(hfbm,Screen2,imagesize,&dwRead,(LPOVERLAPPED)NULL);
 if(dwRead != imagesize){
   MessageBeep(MB_OK);
   MessageBox(NULL,"Bad write",NULL,MB_OK);
 }
 if(!CloseHandle(hfbm)){
   MessageBeep(MB_OK);
   MessageBox(NULL,"Bad close",NULL,MB_OK);
 }
 X__Free(Screen2);
 return;
}

typedef LPBITMAPINFOHEADER PDIB;
extern HPALETTE CreateMyHalftonePalette(void);
extern PDIB DibHalftoneDIB(PDIB pSource, HPALETTE hpalApp);
extern void openAVIfile(char *Name, LPBITMAPINFOHEADER alpbi, int compress);
extern void appendAVIfile(LPBITMAPINFOHEADER alpbi);
extern void closeAVIfile(void);
extern void configureAVIfile(HWND,char *);

static void writeAVI(void){
#if !__DLL__
 DWORD dwRead;
 BITMAPINFOHEADER *lpbi;
 HPALETTE hpalHalftone=NULL;
 long i,j,k,imagesize,nx,mx,ny;
 unsigned char *Screen1,*p,*Screen2,*q,cr,cg,cb;
 if(SelectFileName(1,szAVIFileName,szAVIDirName,szAppName,
             "Video files|*.avi|",glWndFrame) != TRUE)return;
 AppendFileExtension(szAVIFileName,".avi");
 configureAVIfile(glWndFrame,NULL);
 nx=oldrect.right; mx=((nx*24+31)/32) * 4;
 ny=oldrect.bottom;
 if((Screen1=(unsigned char *)X__Malloc(nx*ny*3)) == NULL){
   MessageBox(NULL,"Out of Memory","Warning",MB_OK);
   return;
 }
 if((Screen2=(unsigned char *)X__Malloc(mx*ny+sizeof(BITMAPINFOHEADER))) == NULL){
   MessageBox(NULL,"Out of Memory","Warning",MB_OK);
   X__Free(Screen1);
   return;
 }
 open_script(FALSE,1,szFileName);
 UpdateEverything(TRUE);
 UpdateWindow(glWnd);
 if(fhalf){
   if((hpalHalftone=CreateMyHalftonePalette()) == NULL){
     MessageBox(NULL,"Cannot create Palette","Warning",MB_OK);
     X__Free(Screen1);
     X__Free(Screen2);
     return;
   }
 }
 for(k=0; k<nframes; k++){
   memset(Screen1,0,nx*ny*3);
   memset(Screen2,0,mx*ny+sizeof(BITMAPINFOHEADER));
   glReadPixels(0,0,(GLsizei)nx,(GLsizei)ny,
                GL_RGB,GL_UNSIGNED_BYTE,
                (GLvoid *)Screen1);
   p=Screen1;
   q=Screen2;
   for(i=0;i<ny;i++){
     p=Screen1+i*nx*3;
     q=Screen2+sizeof(BITMAPINFOHEADER)+i*mx;
     for(j=0;j<nx;j++){
       cr = (unsigned char)*p++;
       cg = (unsigned char)*p++;
       cb = (unsigned char)*p++;
       *q++ = (unsigned char)cb;
       *q++ = (unsigned char)cg;
       *q++ = (unsigned char)cr;
     }
   }
   lpbi=(LPBITMAPINFOHEADER)Screen2;
   lpbi->biSize=sizeof(BITMAPINFOHEADER);
   lpbi->biWidth=(DWORD)nx;
   lpbi->biHeight=(DWORD)ny-1;
   lpbi->biPlanes=1;
   lpbi->biBitCount=24;
   lpbi->biCompression=BI_RGB;
   lpbi->biSizeImage = imagesize = (DWORD)(mx*ny);
   lpbi->biXPelsPerMeter=0;
   lpbi->biYPelsPerMeter=0;
   lpbi->biClrUsed=0;
   lpbi->biClrImportant=0;
   if(fhalf)lpbi=DibHalftoneDIB(lpbi,hpalHalftone);
   if(k == 0)openAVIfile(szAVIFileName,lpbi,1);
   appendAVIfile(lpbi);
   if(k == nframes-1)closeAVIfile();
//   if(k == nframes-1)MessageBeep(MB_OK);
//{char temp[255]; sprintf(temp,"Appending %ld to %ld",k,nframes);
//MessageBox(NULL,temp,"Debug",MB_OK);}
   if(fhalf && lpbi != NULL)X__Free(lpbi);
   next_frame(szFileName);
   UpdateEverything(FALSE);
   UpdateWindow(glWnd);
   RenderYield();
 }
 if(fhalf && hpalHalftone != NULL)DeleteObject(hpalHalftone);
 X__Free(Screen1);
 X__Free(Screen2);
#endif
 return;
}


static BOOL CALLBACK PrefsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam){
 BOOL state;
 RECT rc;
 switch( msg ) {
   case WM_INITDIALOG:
     GetWindowRect(hdlg, &rc);
     SetWindowPos(hdlg, NULL,
       (GetSystemMetrics(SM_CXSCREEN) - (rc.right - rc.left)) / 2,
       (GetSystemMetrics(SM_CYSCREEN) - (rc.bottom - rc.top)) / 2,
       0, 0, SWP_NOSIZE | SWP_NOZORDER);
     if(fmapped)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_MAPS,BM_SETCHECK,state,0);
     if(ffastmapping)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_FAST,BM_SETCHECK,state,0);
     if(fantialiased)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_ANTI,BM_SETCHECK,state,0);
     if(fflat)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_FLAT,BM_SETCHECK,state,0);
     if(fhalf)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_HALF,BM_SETCHECK,state,0);
     if(Optimised)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_ACCL,BM_SETCHECK,state,0);
     if(fLightAttenuated)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_ATTN,BM_SETCHECK,state,0);
     if(fFog)state=TRUE; else state=FALSE;
     SendDlgItemMessage(hdlg,DLG_PREFS_FOG,BM_SETCHECK,state,0);
     SetScrollRange(GetDlgItem(hdlg,DLG_PREFS_ATTNV),SB_CTL,1,MAX_ATTEN,FALSE);
     SetScrollPos(GetDlgItem(hdlg,DLG_PREFS_ATTNV),SB_CTL,LightAttenuation,TRUE);
     SetScrollRange(GetDlgItem(hdlg,DLG_PREFS_FOGV),SB_CTL,1,MAX_FOG_ATTEN,FALSE);
     SetScrollPos(GetDlgItem(hdlg,DLG_PREFS_FOGV),SB_CTL,FogAttenuation,TRUE);
     return (TRUE);
   case WM_HSCROLL:
     if((HWND)lparam == GetDlgItem(hdlg,DLG_PREFS_ATTNV)){
       switch (LOWORD(wparam)) {
         case SB_PAGEDOWN: LightAttenuation += MAX_ATTEN/5;   break;
         case SB_LINEDOWN: LightAttenuation += MAX_ATTEN/20;  break;
         case SB_PAGEUP:   LightAttenuation -= MAX_ATTEN/5;   break;
         case SB_LINEUP:   LightAttenuation -= MAX_ATTEN/20;  break;
         case SB_THUMBTRACK:
         case SB_THUMBPOSITION:
           LightAttenuation=HIWORD(wparam);  break;
         default: break;
       }
       LightAttenuation=max(0,LightAttenuation);
       LightAttenuation=min(LightAttenuation,MAX_ATTEN);
       SetScrollPos(GetDlgItem(hdlg,DLG_PREFS_ATTNV),SB_CTL,LightAttenuation,TRUE);
     }
     else if((HWND)lparam == GetDlgItem(hdlg,DLG_PREFS_FOGV)){
       switch (LOWORD(wparam)) {
         case SB_ENDSCROLL:{
     if(fFog){
       GLfloat xs;
       GLfloat fc[4]={0.8,0.8,0.8,1.0};
       if(Nsky == 0){
         fc[0]=gbRed; fc[1]=gbGreen; fc[2]=gbBlue;
       }
       else{
         fc[0]=(GLfloat)Sky.Zenith[0]/255.0;
         fc[1]=(GLfloat)Sky.Zenith[1]/255.0;
         fc[2]=(GLfloat)Sky.Zenith[2]/255.0;
       }
       glFogfv(GL_FOG_COLOR,fc);
//       xs=(GLfloat)UNIT*FogAttenuation;
       xs=(GLfloat)UNIT*(1.0+0.05*(GLfloat)(FogAttenuation*FogAttenuation));
       glFogf(GL_FOG_START,xs);
       glFogf(GL_FOG_END,xs+xs*0.4);
       glFogi(GL_FOG_MODE,GL_LINEAR);
//       glFogi(GL_FOG_MODE,GL_EXP);
//       glFogf(GL_FOG_DENSITY,0.0001/(GLfloat)FogAttenuation);
       glEnable(GL_FOG);
     }
     else{
       glDisable(GL_FOG);
     }
     InvalidateRect(glWnd, NULL, FALSE);
           }
           break;
         case SB_PAGEDOWN: FogAttenuation += MAX_FOG_ATTEN/20;  break;
         case SB_LINEDOWN: FogAttenuation += MAX_FOG_ATTEN/100; break;
         case SB_PAGEUP:   FogAttenuation -= MAX_FOG_ATTEN/20;  break;
         case SB_LINEUP:   FogAttenuation -= MAX_FOG_ATTEN/100; break;
         case SB_THUMBTRACK:
         case SB_THUMBPOSITION:
           FogAttenuation=HIWORD(wparam);  break;
         default: break;
       }
       FogAttenuation=max(0,FogAttenuation);
       FogAttenuation=min(FogAttenuation,MAX_FOG_ATTEN);
       SetScrollPos(GetDlgItem(hdlg,DLG_PREFS_FOGV),SB_CTL,FogAttenuation,TRUE);
     }
     break;
   case WM_COMMAND:
      switch(LOWORD(wparam)){
        case IDOK:
          if(SendDlgItemMessage(hdlg,DLG_PREFS_MAPS,BM_GETCHECK,0,0))
            fmapped=TRUE; else fmapped=FALSE;
          if(SendDlgItemMessage(hdlg,DLG_PREFS_FAST,BM_GETCHECK,0,0))
            ffastmapping=TRUE; else ffastmapping=FALSE;
          if(SendDlgItemMessage(hdlg,DLG_PREFS_ANTI,BM_GETCHECK,0,0))
            fantialiased=TRUE; else fantialiased=FALSE;
          if(SendDlgItemMessage(hdlg,DLG_PREFS_FLAT,BM_GETCHECK,0,0))
            fflat=TRUE; else fflat=FALSE;
          if(SendDlgItemMessage(hdlg,DLG_PREFS_HALF,BM_GETCHECK,0,0))
            fhalf=TRUE; else fhalf=FALSE;
          if(SendDlgItemMessage(hdlg,DLG_PREFS_ACCL,BM_GETCHECK,0,0))
            Optimised=TRUE; else Optimised=FALSE;
          if(SendDlgItemMessage(hdlg,DLG_PREFS_ATTN,BM_GETCHECK,0,0))
            fLightAttenuated=TRUE; else fLightAttenuated=FALSE;
          if(SendDlgItemMessage(hdlg,DLG_PREFS_FOG,BM_GETCHECK,0,0))
            fFog=TRUE; else fFog=FALSE;
          SavePreferencesGLA();
          EndDialog(hdlg,TRUE);
          return(TRUE);
          break;
        case IDCANCEL:
          EndDialog(hdlg,FALSE);
          return(TRUE);
          break;
        default:  break;
      }
      break;
    default: break;
 }
 return(FALSE);
}

static void LoadPreferencesGLA(void){
 long i;
 i=GetPrivateProfileInt("GLAPREFERENCES","MAPPED",-1,iniFilename);
 if(i == 0)fmapped=FALSE; else if(i == 1)fmapped=TRUE;
 i=GetPrivateProfileInt("GLAPREFERENCES","FASTMAP",-1,iniFilename);
 if(i == 0)ffastmapping=FALSE; else if(i == 1)ffastmapping=TRUE;
 i=GetPrivateProfileInt("GLAPREFERENCES","ANTIALIASED",-1,iniFilename);
 if(i == 0)fantialiased=FALSE; else if(i == 1)fantialiased=TRUE;
 i=GetPrivateProfileInt("GLAPREFERENCES","FLATSHADE",-1,iniFilename);
 if(i == 0)fflat=FALSE; else if(i == 1)fflat=TRUE;
 i=GetPrivateProfileInt("GLAPREFERENCES","HALFTONE",-1,iniFilename);
 if(i == 0)fhalf=FALSE; else if(i == 1)fhalf=TRUE;
 i=GetPrivateProfileInt("GLAPREFERENCES","OPTIMISED",-1,iniFilename);
 if(i == 0)Optimised=FALSE; else if(i == 1)Optimised=TRUE;
 i=GetPrivateProfileInt("GLAPREFERENCES","ATTENUATE",-1,iniFilename);
 if(i == 0)fLightAttenuated=FALSE; else if(i == 1)fLightAttenuated=TRUE;
 i=GetPrivateProfileInt("GLAPREFERENCES","FOG",-1,iniFilename);
 if(i == 0)fFog=FALSE; else if(i == 1)fFog=TRUE;
 LightAttenuation=GetPrivateProfileInt("GLAPREFERENCES","LIGHTATTEN",
                LightAttenuation,iniFilename);
 FogAttenuation=GetPrivateProfileInt("GLAPREFERENCES","FOGATTEN",
                FogAttenuation,iniFilename);
 return;
}

static void SavePreferencesGLA(void){
 char str[16];
 sprintf(str,"%ld",(long)fmapped);
 WritePrivateProfileString("GLAPREFERENCES","MAPPED",str,iniFilename);
 sprintf(str,"%ld",(long)ffastmapping);
 WritePrivateProfileString("GLAPREFERENCES","FASTMAP",str,iniFilename);
 sprintf(str,"%ld",(long)fantialiased);
 WritePrivateProfileString("GLAPREFERENCES","ANTIALIASED",str,iniFilename);
 sprintf(str,"%ld",(long)fflat);
 WritePrivateProfileString("GLAPREFERENCES","FLATSHADE",str,iniFilename);
 sprintf(str,"%ld",(long)fhalf);
 WritePrivateProfileString("GLAPREFERENCES","HALFTONE",str,iniFilename);
 sprintf(str,"%ld",(long)Optimised);
 WritePrivateProfileString("GLAPREFERENCES","OPTIMISED",str,iniFilename);
 sprintf(str,"%ld",(long)fLightAttenuated);
 WritePrivateProfileString("GLAPREFERENCES","ATTENUATE",str,iniFilename);
 sprintf(str,"%ld",(long)fFog);
 WritePrivateProfileString("GLAPREFERENCES","FOG",str,iniFilename);
 sprintf(str,"%ld",(long)LightAttenuation);
 WritePrivateProfileString("GLAPREFERENCES","LIGHTATTEN",str,iniFilename);
 sprintf(str,"%ld",(long)FogAttenuation);
 WritePrivateProfileString("GLAPREFERENCES","FOGATTEN",str,iniFilename);
 return;
}
