/* file OPTIMISE.C */

/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

static void ListAdjFaces(void){
 face *f;
 long i,j,*fl,id,n,m,id0,id1,id2,nl[3],ml[3],k,kk,fjd,fkd;
 for(i=0;i<Nface;i++){ /* make list of work faces attached to each vertex */
   f=(MainFp+i);
   for(j=0;j<3;j++){
     id=f->V[j];
     fl=(MainVp+id)->flist;
     n=(MainVp+id)->nf;
     if(fl == NULL){
       if((fl=(long *)X__Malloc(sizeof(long))) == NULL)return;
     }
     else{
       if((fl=(long *)X__Realloc(fl,(n+1)*sizeof(long))) == NULL)return;
     }
     fl[n]=i;
     (MainVp+id)->flist=fl;
     (MainVp+id)->nf=n+1;
   }
 }
 for(i=0;i<Nface;i++){ /* find faces that are adjacent to face i */
   f=(MainFp+i);
   id0=f->V[0];   id1=f->V[1];   id2=f->V[2];
   nl[0]=id0; ml[0]=id1;
   nl[1]=id1; ml[1]=id2;
   nl[2]=id2; ml[2]=id0;
   for(kk=0;kk<3;kk++){
     f->A[kk] = -1;
     if((n=(MainVp+nl[kk])->nf) > 0 && (m=(MainVp+ml[kk])->nf) > 0){
       for(j=0;j<n;j++){
         fjd=MainVp[nl[kk]].flist[j];
         for(k=0;k<m;k++){
           fkd=MainVp[ml[kk]].flist[k];
           if(fkd == fjd && fkd != i){
             f->A[kk]=fkd;
             goto ENDF;
           }
         }
       }
     }
     ENDF:;
   }
 }
 return;
}

static void MakeSameOrientation(long last, long id){
 face *f;
 long i0,i1,i2,l0,l1,l2,a0,a1,a2;
 f=(MainFp+last); l0=f->V[0]; l1=f->V[1]; l2=f->V[2];
 f=(MainFp+id);   i0=f->V[0]; i1=f->V[1]; i2=f->V[2];
                  a0=f->A[0]; a1=f->A[1]; a2=f->A[2];
 if((i2 == l2 && i1 == l1) ||
    (i2 == l1 && i1 == l0) ||
    (i2 == l0 && i1 == l2)){
   f->V[2]=i1; f->V[1]=i2; f->A[2]=a0; f->A[0]=a2;
   return;
 }
 if((i0 == l2 && i2 == l1) ||
    (i0 == l1 && i2 == l0) ||
    (i0 == l0 && i2 == l2)){
   f->V[0]=i2; f->V[2]=i0; f->A[0]=a1; f->A[1]=a0;
   return;
 }
 if((i1 == l1 && i0 == l0) ||
    (i1 == l2 && i0 == l1) ||
    (i1 == l0 && i0 == l2)){
   f->V[1]=i0; f->V[0]=i1; f->A[1]=a2; f->A[2]=a1;
   return;
 }
 return;
}

static void SetConnectingID(long last, long id){
 if((MainFp+id)->start == 1)return;
 (MainFp+id)->start=1;
 if(last >= 0)MakeSameOrientation(last,id);
 last=id;
 if((MainFp+id)->A[0] >= 0)SetConnectingID(last,(MainFp+id)->A[0]);
 if((MainFp+id)->A[1] >= 0)SetConnectingID(last,(MainFp+id)->A[1]);
 if((MainFp+id)->A[2] >= 0)SetConnectingID(last,(MainFp+id)->A[2]);
 return;
}

static void FixUpOrientation(void){
 long i;
 while(1){
   for(i=0;i<Nface;i++){
     if((MainFp+i)->start == 0)goto PROCESSIT;
   }
   break;
   PROCESSIT:
   SetConnectingID(-1,i);
 }
 for(i=0;i<Nface;i++)(MainFp+i)->start=0;
}

static long CanExtendStrip(face *f, long next){
 if((MainFp+next)->next != -1)next = -1;
 else {
   face *fn;
   vector n,nn;
   fn=(MainFp+next);
   if(abs(f->color[0] - fn->color[0]) > 10 ||
      abs(f->color[1] - fn->color[1]) > 10 ||
      abs(f->color[2] - fn->color[2]) > 10 ){
     next = -1;
   }
   else{
     Normalize((MainVp+(f->V[0]))->p,         /* too great an angle     */
               (MainVp+(f->V[1]))->p,         /* assumes that faces     */
               (MainVp+(f->V[2]))->p,n);      /* have been consistently */
     Normalize((MainVp+(fn->V[0]))->p,        /* numbered.              */
               (MainVp+(fn->V[1]))->p,
               (MainVp+(fn->V[2]))->p,nn);
     if(DOT(n,nn) < 0.01)next = -1;
   }
 }
 return next;
}

static long NextFace(int id, int v1, int v2, int last, int *v3){
 face *f;
 int i0,i1,i2,next;
 next = -10;
 f=(MainFp+last); i0=f->V[0]; i1=f->V[1]; i2=f->V[2];
 if(     i0 == v1 && i1 == v2){
   *v3 = 2;
   if(id == 0)next=f->A[1];
   else       next=f->A[2];
 }
 else if(i1 == v1 && i0 == v2){
   *v3 = 2;
   if(id == 0)next=f->A[2];
   else       next=f->A[1];
 }
 else if(i1 == v1 && i2 == v2){
   *v3 = 0;
   if(id == 0)next=f->A[2];
   else       next=f->A[0];
 }
 else if(i2 == v1 && i1 == v2){
   *v3 = 0;
   if(id == 0)next=f->A[0];
   else       next=f->A[2];
 }
 else if(i2 == v1 && i0 == v2){
   *v3 = 1;
   if(id == 0)next=f->A[0];
   else       next=f->A[1];
 }
 else if(i0 == v1 && i2 == v2){
   *v3 = 1;
   if(id == 0)next=f->A[1];
   else       next=f->A[0];
 }
 if(next == -10){
   MessageBox(NULL,"None Found",NULL,MB_OK);
   next=-1;
 }
 if(next >= 0 && (MainFp+next)->next != -1)next = -1; /* already used */
 if(next >= 0)next=CanExtendStrip(f,next); /* already used or other attrib */
 return next;
}

static void StripFaces(void){
 vector n,nn;
 int i,j,k,last,next,v1,v2,v3,v4,i0,i1,i2;
 face *fp1,*fp2;
 for(i=0,fp1=MainFp;i<Nface;i++){
   last=i; k= -1;
   if(fp1->next == -1){ /* not processed */
     if((next=fp1->A[1]) >= 0){
       j=1; next=CanExtendStrip(fp1,next); //already in use/other attrib
       if(next >= 0){
         k=j; /* if this strip is only two long try the others too */
         if(NextFace(0,fp1->V[1],fp1->V[2],next,&v3) < 0)next = -1;
       }
     }
     if(next < 0 && (next=fp1->A[0]) >= 0){
       j=0; next=CanExtendStrip(fp1,next); //already in use/other attrib
       if(next >= 0){
         k=j; /* if this strip is only two long try the others too */
         if(NextFace(0,fp1->V[0],fp1->V[1],next,&v3) < 0)next = -1;
       }
     }
     if(next < 0 && (next=fp1->A[2]) >= 0){
       j=2; next=CanExtendStrip(fp1,next); //already in use/other attrib
       if(next >= 0){
         k=j; /* if this strip is only two long try the others too */
         if(NextFace(0,fp1->V[2],fp1->V[0],next,&v3) < 0)next = -1;
       }
     }
     if(next == -1 && k >= 0){ /* use last best hope */
       j=k; next=fp1->A[j];
     }
     if(next >= 0){
       (MainFp+last)->start=1;
       (MainFp+last)->next=next;  /* face 0 */
       (MainFp+next)->next = -2;  /* end of strip face 1 */
       if(j == 0){
         i0=(MainFp+last)->V[0];
         i1=(MainFp+last)->V[1];
         i2=(MainFp+last)->V[2];
         (MainFp+last)->V[1]=i0;
         (MainFp+last)->V[2]=i1;
         (MainFp+last)->V[0]=i2;
       }
       else if(j == 2){
         i0=(MainFp+last)->V[0];
         i1=(MainFp+last)->V[1];
         i2=(MainFp+last)->V[2];
         (MainFp+last)->V[1]=i2;
         (MainFp+last)->V[2]=i0;
         (MainFp+last)->V[0]=i1;
       }
       v1=(MainFp+last)->V[1];    /* of face 0 */
       v2=(MainFp+last)->V[2];    /* of face 0 */
       while(1){
         last=next;                    /* last now equals face 1      */
         next=NextFace(0,v1,v2,last,&v3);
         (MainFp+last)->vid=v3;        /* index vertex (3) current end*/
         v3=(MainFp+last)->V[v3];      /* next vertex (3)             */
         if(next < 0)break;            /* cannot extend               */
         (MainFp+last)->next=next;     /* set pointer                 */
         (MainFp+next)->next = -2;     /* set this to an end of strip */
         last=next;                    /* last now equal face 2       */
         next=NextFace(1,v3,v2,last,&v4);
         (MainFp+last)->vid=v4;        /* index of next vertex face 2 */
         v4=(MainFp+last)->V[v4];
         if(next < 0)break;            /* cannot extend               */
         (MainFp+last)->next=next;
         (MainFp+next)->next = -2;
         v1=v3;                        /* v1 moves on to v3           */
         v2=v4;                        /* v2 moves on to v4           */
       }
     }
   }
   fp1++;
 }
 for(i=0;i<Nface;i++){ /* any two face strips make quad if possible */
   fp1=(MainFp+i);
   if(fp1->start == 1){
     fp2=(MainFp+(fp1->next));
     if(fp2->next == -2){ /* two face strip */
       Normalize((MainVp+(fp1->V[0]))->p,
                 (MainVp+(fp1->V[1]))->p,
                 (MainVp+(fp1->V[2]))->p,n);
       Normalize((MainVp+(fp2->V[0]))->p,
                 (MainVp+(fp2->V[1]))->p,
                 (MainVp+(fp2->V[2]))->p,nn);
       if(DOT(n,nn) > 0.95){
         fp1->start=2; /* quadilateral */
       }
       else{ /* set remaining two triangles to TRIANGLES ? */
//         fp2->next = -1;
//         fp1->next= -1;
//         fp1->start=FALSE;
       }
     }
   }
 }
}
