
#define MODULE_RENDER

/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

#include "render.h"

extern BOOL uselists;

extern void EffectObjectVertices(long O);
extern void MakeFrameList(BOOL);
extern void FreeMemoryObjects(void);

static void free_frame(void);
static short GetCommand(void);
void R_rotz(double tr[4][4], double ang);
void R_rotx(double tr[4][4], double ang);
void R_roty(double tr[4][4], double ang);

FILE *CF=NULL;
long frame,firstframe,nframes;

static char outputfile[256],inputfile[256],tempfile[256];

void close_script(void){
 free_frame();
 if(CF != NULL)fclose(CF);
 CF=NULL;
}

int open_script(BOOL inew, long firstframe, char *scriptfile){
 char c,temp_str[80];
 long i,failure=NO;
 close_script();
#if !__DLL__
 if(inew)FreeMemoryObjects();
#endif
 frame=0;
 strcpy(inputfile,scriptfile);
 if((CF=fopen((char *)strlwr(inputfile),"r")) == NULL){
   MessageBox(NULL,"Failed to open inputfile",inputfile,MB_OK);
   goto EXITPOINT;
 }
 fscanf(CF,"%s",temp_str);
 if(temp_str[0] == 'R'){ /* get the number of frames in the script */
   fscanf(CF,"%s",temp_str);
   nframes=atoi(temp_str);
   fscanf(CF,"%s",temp_str); /* start frame in animation */
 }
 else{
   fscanf(CF,"%s",temp_str);
   nframes=-1;
 }
 NEXTFRAME:
 frame++;
 if(frame < firstframe){
   char tt[256];
   LOOP:
   if(fscanf(CF,"%s",tt) == EOF)goto EXITPOINT;
   if(tt[0] != 'f')goto LOOP;
   goto NEXTFRAME;
 }
 if(GetCommand() < 0){ /* last frame (q command)*/
   free_frame();
 }
 else if(uselists)MakeFrameList(TRUE);
 EXITPOINT:
 return (int)failure;
}

void next_frame(char *scriptfile){
 static BOOL busy=FALSE;
 if(busy)return;
 if(CF == NULL)return;
 busy=TRUE;
 free_frame();
 frame++;
 if(GetCommand() < 0){ /* last frame (q command)*/
   open_script(FALSE,1,scriptfile);
 }
 else if(uselists){
   MakeFrameList(TRUE);
 }
 busy=FALSE;
}

static void free_frame(void){
 int i;
 if(ObjectCount > 0){
   for(i=0;i<ObjectCount;i++){
     if(Object[i].Vbase != NULL)X__Free(Object[i].Vbase);
     if(Object[i].Fbase != NULL)X__Free(Object[i].Fbase);
     if(Object[i].Mbase != NULL)X__Free(Object[i].Mbase);
     if(Object[i].effect_param_list != NULL)
       X__Free(Object[i].effect_param_list);
     if(Object[i].Axis  != NULL)X__Free(Object[i].Axis);
   }
 }
 if(Object != NULL){X__Free(Object); Object = NULL;}
 ObjectCount=0;
 if(Lights != NULL){X__Free(Lights); Lights = NULL;}
 Nlights=0;
 Nground=0;
 Nsky=0;
 if(nImageProcesses > 0 && ImageProcesses != NULL){
   for(i=0;i<nImageProcesses;i++){
     if(ImageProcesses[i].parameters != NULL)
       X__Free(ImageProcesses[i].parameters);
   }
   X__Free(ImageProcesses);
   ImageProcesses=NULL;
   nImageProcesses=0;
 }
}

void read_vec(vector v){
 if(fscanf(CF,GET_VEC,&v[0],&v[1],&v[2]) != 3){ ; }
}

static short GetCommand(void){
 long a=0,b=0,c,d,e,f;
 double p,t,al,du,sx,sy,sz,gbtr[4][4],tr1[4][4],tr2[4][4],
        qmorph_ratio,robo_ratio,i_ratio;
 short imo,mo,robo=NO;
 char   Com[16],second_morph[256],robo_file1[256],robo_file2[256];
 object *temp_object;
 light  *temp_light;
 long   flag;
 XIP    *temp_xip;
 while (fscanf(CF,"%s",Com) != EOF){
//MessageBox(NULL,Com,"Command",MB_OK);
  switch(Com[0]) {
    case 'e' :
      read_vec(ViewPoint);
      fscanf(CF,GET_VEC,&CamPhi,&CamTheta,&CamAlpha);
      fscanf(CF,GET_VEC,&CamSx,&CamSy,&CamSz);
      fscanf(CF,"%ld %ld",&a,&b); CamIm=b;  /* camera internal */
      fscanf(CF,GET_DOB,&CamIma);
      break;
    case 'i' :
      fscanf(CF,"%ld %ld %ld %ld",&imo,&a,&c,&b);
      fscanf(CF,GET_DOB,&i_ratio);
      if(ImageProcesses == NULL)temp_xip = (XIP *)X__Malloc(sizeof(XIP));
      else                      temp_xip = (XIP *)X__Realloc(ImageProcesses,
                                    (nImageProcesses+1)*sizeof(XIP));
      if(temp_xip == NULL)return -1;
      temp_xip[nImageProcesses].imo=(long)imo;
      temp_xip[nImageProcesses].first_frame=a;
      temp_xip[nImageProcesses].last_frame=b;
      temp_xip[nImageProcesses].fmr=i_ratio;
      temp_xip[nImageProcesses].parameters=NULL;
      temp_xip[nImageProcesses].mparameters=NULL;
      fscanf(CF," %[^\n]",tempfile); c=strlen(tempfile)+1;
      if((temp_xip[nImageProcesses].parameters=(char *)X__Malloc(c)) != NULL)
        strcpy(temp_xip[nImageProcesses].parameters,tempfile);
      if(imo == 1){
        fscanf(CF," %[^\n]",tempfile); c=strlen(tempfile)+1;
        if((temp_xip[nImageProcesses].mparameters=(char *)X__Malloc(c)) != NULL)
          strcpy(temp_xip[nImageProcesses].mparameters,tempfile);
      }
      ImageProcesses=temp_xip;
      nImageProcesses++;
      break;
    case 't' : {  /* target */
        vector temp_vec;
        read_vec(temp_vec);
      }
      break;
    case 'l' :
      if(Lights == NULL)temp_light = (light *)X__Malloc(sizeof(light));
      else              temp_light = (light *)X__Realloc(Lights,
                                        (Nlights+1)*sizeof(light));
      if(temp_light == NULL){
        if(Lights != NULL)X__Free(Lights);
        Lights=NULL;
        return -1;
      }
      Lights=temp_light;
      fscanf(CF,"%ld",&a); Lights[Nlights].type=a;
      fscanf(CF,"%ld %ld %ld",&a,&b,&c);
      fscanf(CF,GET_DOB,&Lights[Nlights].Intensity);
      Lights[Nlights].color[0] = (unsigned char)a;
      Lights[Nlights].color[1] = (unsigned char)b;
      Lights[Nlights].color[2] = (unsigned char)c;
      read_vec(Lights[Nlights].p);
      VECCOPY(Lights[Nlights].p,Lights[Nlights].pin)
      fscanf(CF,GET_VEC,&Lights[Nlights].dPhi
                       ,&Lights[Nlights].dTheta
                       ,&Lights[Nlights].dAlpha);
       /* Theta axis bank  (Not used)                  */
       /* Phi is direction   Alpha is elevation angle  */
       fscanf(CF,"%ld %ld",&a,&b);  /* light internals */
       Lights[Nlights].internal_type=b; /* a=0 for expansion */
       fscanf(CF,GET_DOB,&Lights[Nlights].internal_size);
       fscanf(CF,"%ld",&a);  /* light depth cue */
       Lights[Nlights].dc=a;
       fscanf(CF,GET_DOB,&Lights[Nlights].dc_c);
       fscanf(CF,GET_DOB,&Lights[Nlights].dc_l);
       fscanf(CF,GET_DOB,&Lights[Nlights].dc_q);
       if(Lights[Nlights].type >= SPOTLIGHT){
         fscanf(CF,GET_DOB,&Lights[Nlights].cone1); /* angle  */
         fscanf(CF,GET_DOB,&Lights[Nlights].falloff);  /* degree */
       }
       Nlights++;
       break;
     case 'g' :
       Nground=1;
       fscanf(CF,"%ld",&a); Ground.type=a;
       fscanf(CF,"%ld %ld %ld",&a,&b,&c);
       Ground.color[0] = (unsigned char)a;
       Ground.color[1] = (unsigned char)b;
       Ground.color[2] = (unsigned char)c;
       Ground.ambient_light=0;
       Ground.brush = -1; Ground.texture = -1; Ground.shader = -1;
       if(Ground.type == PLAIN)Ground.texture=0;
       else if(Ground.type > 127){
         fscanf(CF,"%ld %ld %ld",&a,&b,&c);
         Ground.acolor[0] = (unsigned char)a;
         Ground.acolor[1] = (unsigned char)b;
         Ground.acolor[2] = (unsigned char)c;
         Ground.texture = Ground.type-128;
         Ground.type = TEXTU;
         if(Ground.texture == 0)Ground.type=PLAIN;
         else if(Ground.texture > 15){  /* external shader   */
           fscanf(CF,"  %[^\n]",inputfile); /* no parameters yet */
           Ground.texture=0; Ground.type=PLAIN;
         }
       }
       if(Ground.type == BRUM || Ground.type == TBRUM || Ground.type == MBRUM){
         fscanf(CF,"  %[^\n]",inputfile);
         strcpy(Ground.Map.filename,inputfile);
       }
       read_vec(Ground.p);
       fscanf(CF,GET_VEC,&p,&t,&al);
       fscanf(CF,GET_VEC,&sx,&sy,&sz);
       fscanf(CF,"%ld %ld",&a,&b);  /* ground internals */
       fscanf(CF,GET_DOB,&du);
       fscanf(CF,GET_DOB,&Ground.refl); Ground.refl *= 0.01;
       Ground.angle=p;
       Ground.height=sz;
       Ground.x[0]=Ground.p[0]+(double)UNIT*sx*4.0;  /* ground is flat */
       Ground.x[1]=Ground.p[1];
       Ground.x[2]=Ground.p[2];
       Ground.y[0]=Ground.p[0];
       Ground.y[1]=Ground.p[1]-(double)UNIT*sy*4.0;
       Ground.y[2]=Ground.p[2];       /* this may change */
       if(Ground.type == TEXTU){;}
       else if(Ground.type == BRUM || Ground.type == TBRUM ||
               Ground.type == MBRUM){
         Ground.Map.p[0]=0.0;
         Ground.Map.p[1]=0.0;
         Ground.Map.p[2]=0.0;
         Ground.Map.x[0]=Ground.Map.p[0]+UNIT*4*sx;
         Ground.Map.x[1]=Ground.Map.p[1];
         Ground.Map.x[2]=Ground.Map.p[2];
         Ground.Map.y[0]=Ground.Map.p[0];
         Ground.Map.y[1]=Ground.Map.p[1]-UNIT*4*sy;
         Ground.Map.y[2]=Ground.Map.p[2];
         if(b > 0){
           if(b == 1)R_rotz(tr1,du*PI/180.0);
           if(b == 2)R_rotx(tr1,du*PI/180.0);
           if(b == 3)R_roty(tr1,du*PI/180.0);
           R_rotz(tr2,p*PI/180.0);
           R_m4by4(tr2,tr1,gbtr);
         }
         else R_rotz(gbtr,p*PI/180.0);
         R_m4by1(gbtr, Ground.Map.x[0], Ground.Map.x[1],
                       Ground.Map.x[2],&Ground.Map.x[0],
                      &Ground.Map.x[1],&Ground.Map.x[2]);
         R_m4by1(gbtr, Ground.Map.y[0], Ground.Map.y[1],
                       Ground.Map.y[2],&Ground.Map.y[0],
                      &Ground.Map.y[1],&Ground.Map.y[2]);
         VECSUM(Ground.Map.p,Ground.p,Ground.Map.p)
         VECSUM(Ground.Map.x,Ground.p,Ground.Map.x)
         VECSUM(Ground.Map.y,Ground.p,Ground.Map.y)
       }
       VECCOPY(Ground.p,Ground.pin)
       VECCOPY(Ground.x,Ground.xin)
       VECCOPY(Ground.y,Ground.yin)
       break;
     case 's' :
       Nsky=1;
       fscanf(CF,"%ld %ld %ld %ld %ld %ld ",&a,&b,&c,&d,&e,&f);
       Sky.Horizon[0]=a; Sky.Horizon[1]=b; Sky.Horizon[2]=c;
       Sky.Zenith[0]=d; Sky.Zenith[1]=e; Sky.Zenith[2]=f;
       fscanf(CF,GET_DOB,&Sky.blendratio);
       fscanf(CF,GET_DOB,&object_ambient_light);
       fscanf(CF,GET_DOB,&ground_ambient_light);
       fscanf(CF,"%ld %ld %ld %ld %ld %ld ",&a,&b,&c,&d,&e,&f);
       Sky.Zblend[0]=a; Sky.Zblend[1]=b; Sky.Zblend[2]=c;
       Sky.Hblend[0]=d; Sky.Hblend[1]=e; Sky.Hblend[2]=f;
       fscanf(CF,"%ld",&a);
       Sky.type = (short)a;
       Sky.brush = -1;
       if(Sky.type > SKYGRADED){
         fscanf(CF,"  %[^\n]",inputfile);
         strcpy((char *)Sky.MapName,inputfile);
         if(Sky.type != BACKDROP)Sky.type=SKYCOLOUR;
       }
       break;
     case 'h' :
     case 'w' :
     case 'o' :
       if(Object == NULL)temp_object = (object *)X__Malloc(sizeof(object));
       else              temp_object = (object *)X__Realloc(Object,
                                       (ObjectCount+1)*sizeof(object));
       if(temp_object == NULL){
         if(Object != NULL)X__Free(Object);
         Object=NULL;
         return -1;
       }
       Object=temp_object;
       Object[ObjectCount].effect_param_list=NULL;
       fscanf(CF," %ld",&a);
       fscanf(CF," %[^\n]",tempfile);
       strcpy(inputfile,tempfile);
       Object[ObjectCount].NoMaps=0;
       Object[ObjectCount].Origin[0]=0;
       Object[ObjectCount].Origin[1]=0;
       Object[ObjectCount].Origin[2]=0;
       Object[ObjectCount].morphratio=1.0;
       Object[ObjectCount].Axis=(axis *)X__Malloc(sizeof(axis));
       fscanf(CF,"%ld",&a); mo=a;
       robo=NO;
       if(mo == 4){ /* robot */
         robo=YES;
         fscanf(CF," %[^\n]",tempfile);
         strcpy(robo_file1,tempfile);
         fscanf(CF,"  %[^\n]",tempfile);
         strcpy(robo_file2,tempfile);
         fscanf(CF,GET_DOB,&robo_ratio);
         mo=0;     /* robots can't morph */
       }
       if(R_LoadObject(inputfile,frame,robo,robo_ratio,
                       robo_file1,robo_file2) <= 0){
         return (-1);
       }
       if(mo > 0){
         fscanf(CF,GET_DOB,&Object[ObjectCount].morphratio);
         fscanf(CF,"  %[^\n]",tempfile);
         strcpy(inputfile,tempfile);
         if(mo > 1){/* 2 or 3 for - or + morph */
           fscanf(CF,GET_DOB,&qmorph_ratio);
           fscanf(CF,"  %[^\n]",tempfile);
           strcpy(second_morph,tempfile);
         }
         if(LoadMorphObject(mo,inputfile,Object[ObjectCount].morphratio,
            second_morph,qmorph_ratio) == 0){;}
       }
       read_vec(Object[ObjectCount].Offset);
       fscanf(CF,GET_VEC,&Object[ObjectCount].fi
                        ,&Object[ObjectCount].theta
                        ,&Object[ObjectCount].alpha);
       fscanf(CF,GET_VEC,&Object[ObjectCount].sx
                        ,&Object[ObjectCount].sy
                        ,&Object[ObjectCount].sz);
       fscanf(CF,"%ld %ld",&a,&b); /* a is zero for expansion */
       Object[ObjectCount].im=b;
       fscanf(CF,GET_DOB,&Object[ObjectCount].ima);
       fscanf(CF,"%ld",&a); Object[ObjectCount].effect=a;
       fscanf(CF,GET_DOB,&Object[ObjectCount].effectratio);
       fscanf(CF,"%ld %ld %ld",&Object[ObjectCount].self_shadow,
         &Object[ObjectCount].cast_shadow,&Object[ObjectCount].show_shadow);
       if(a > 20){
         c=0L;
         Object[ObjectCount].effect_param_list=NULL;
         fscanf(CF,"%s",tempfile);
         b=strlen(tempfile)+1L;
         if((Object[ObjectCount].effect_param_list=(char *)X__Malloc(b)) == NULL){
           Object[ObjectCount].effect=0;
           goto NOFX;
         }
         else strcpy(Object[ObjectCount].effect_param_list,tempfile);
         c += b;
         while(1){
           fscanf(CF,"%s",tempfile);
           if(strcmp(tempfile,"~") == 0) break;
           b=strlen(tempfile)+2L;
           c += b;
           if((Object[ObjectCount].effect_param_list=(char *)
               X__Realloc(Object[ObjectCount].effect_param_list,c)
              ) == NULL){ Object[ObjectCount].effect=0; goto NOFX; }
           else {
             strcat(Object[ObjectCount].effect_param_list," ");
             strcat(Object[ObjectCount].effect_param_list,tempfile);
           }
         }
         /* Determine if F/X is internal */
         sscanf(Object[ObjectCount].effect_param_list,"%s",tempfile);
         if(strlen(tempfile) <= 2 && tempfile[0] > 48 &&
            tempfile[0] < 58)Object[ObjectCount].effect=atoi(tempfile);
         else Object[ObjectCount].effect=50;
         if(Object[ObjectCount].effect >= 4)EffectObjectVertices(ObjectCount);
         NOFX:;
       }
       ObjectCount++;
       break;
     case 'f' : return (0); break;
     case 'q' :
       fclose(CF);
       CF=NULL;
       return (-1);
     default:   break;
  }
 }
 return (-1);
}

void R_rotz(double tr[4][4], double ang){
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][1] = -sin(ang);
  tr[1][0] =  sin(ang);
  tr[1][1] =  cos(ang);
  return;
}

void R_rotx(double tr[4][4], double ang){
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[1][1] =  cos(ang);
  tr[1][2] = -sin(ang);
  tr[2][1] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

void R_roty(double tr[4][4], double ang){
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][2] = -sin(ang);
  tr[2][0] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

void R_m4by4(double t1[4][4], double t2[4][4], double tr[4][4]){
 long i,j,k;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  for(k=0;k<4;k++)tr[i][j]=tr[i][j]+t1[i][k]*t2[k][j];
 }
 return;
}

void R_m4by1(double t4[4][4], double x, double y, double z,
           double *xx, double *yy, double *zz){
   *xx=t4[0][0]*x+t4[0][1]*y+t4[0][2]*z+t4[0][3];
   *yy=t4[1][0]*x+t4[1][1]*y+t4[1][2]*z+t4[1][3];
   *zz=t4[2][0]*x+t4[2][1]*y+t4[2][2]*z+t4[2][3];
 return;
}

void R_tram(double t[4][4], double dx, double dy, double dz){
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  t[i][j]=0.0;
  if(i == j)t[i][j]=1.0;
 }
 t[0][3]=dx;
 t[1][3]=dy;
 t[2][3]=dz;
 t[3][3]=1;
 return;
}

void R_c4to4(double tin[4][4], double tout[4][4]){
 int i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 tout[i][j]=tin[i][j];
 return;
}

void R_null_transform(double t[4][4]){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 if(i == j)t[i][j]=1.0;
 else      t[i][j]=0.0;
}
