/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*   RENDER.H     include file for the renderer       */

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <windows.h>
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)

#define YES       1
#define NO        0
#define FAIL     -1
#define OK        1

#define GET_MATX  "%f %f %f %f"
#define GET_VEC   "%f %f %f"
#define GET_DOB   "%f"
#define double float
#define PI 3.1415926

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

#define LIGHTMAX 16

#define UNIT  32768L
#define UNIT2 65526L
#define BIGP  2147483647L
#define BIGN -2147483647L
#define FARAWAY 1.0e30
#define max_spec 1000

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define INVERT(a)       { a[0] = -a[0]; a[1] = -a[1]; a[2] = -a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }

#define SINGLE    1
#define TILE      2
#define PLANE           0 /*  Brush map type must be the same as modeler */
#define CYLINDER        1 /*         "                              "    */
#define PLANE_MOZIAC    2
#define CYLINDER_MOZIAC 3
#define MAP_BY_VERTEX   4
#define EOPAQUE   0       /*  Brush transparent colour 0                 */
#define TRANSP    1       /*  must be same as modeler                    */
#define GLOSS     0       /*  Brush is shiny  must be same as modeler    */
#define MATT      1       /*  Brush is dull                              */
#define STILBRUSH 0
#define ANIMBRUSH 1
#define BUMPBRUSH 2
#define REFLBRUSH 3
#define PSETBRUSH 4
#define PLAIN     0
#define TEXTU     1
#define BRUM      2
#define TBRUM     3
#define MBRUM     4
#define SKYCOLOUR 0
#define SKYGRADED 1
#define SKYMAPPED 2
#define BACKDROP  3
#define NORMAL    0    /* light types */
#define NOSHADOW  1
#define SHADOWS   2
#define SPOTLIGHT 3

typedef double         vector[3];
typedef long           ivector[3];

typedef struct tagiMAP {
        vector n,p,x,y;
        char filename[128],frefl[128],fbump[128];
        short image,irefl,ibump;           /* map     PLANE / CYL         */
        short map          ;               /* image   YES if loaded       */
        short method,moziac;               /* method  anim/still          */
        short pp,rp,bp;                    /* map percent                 */
        double angle;                      /* angle   CLY type maps only  */
        double pd,rd;
        int  ix,iy;                        /* size of scaled image        */
        unsigned char *pixels;             /* scaled image                */
} iMap;

typedef struct tagXIP {
  long   imo,first_frame,last_frame;
  double fmr;
  char   *parameters;
  char   *mparameters;
} XIP;

typedef struct VERTEX { /* structure for each vertex   */
 vector p;              /* real position  3D space     */
 double x,y;            /* texture co-ordinates        */
 long nf,*flist;
} vertex;

typedef struct FACE {
 unsigned long V[3];
 long A[3],next;
 char vid,start;
 unsigned char color[3];
 unsigned char matsp,mattr,matre,texco[3];
 unsigned char texture,brush,x;
 short MapID;
} face;

typedef struct {
        vector pin;
        vector p,d;  /* p  is position  d is normalized dirn. for spots */
        double dot1,falloff,dPhi,dTheta,dAlpha,cone1,cone2,dot2,
               internal_size,scale;
        short type;
        float *shadow_buffer;
        double shadow_transform[4][4];
        unsigned char color[3],internal_type,*object_buffer;
        unsigned short *face_buffer;
        double Ire,Igr,Ibl,Intensity;
        BOOL   dc;
        double dc_c,dc_l,dc_q;
} light;

typedef struct {
       vector boin,bo;
       vector buin,bu;
       vector bvin,bv;
       vector bn;
       double tl;
} axis;

typedef struct { /* info for each object */
        ivector Origin;
        axis    *Axis;
        vertex  *Vbase;
        face    *Fbase;
        iMap    *Mbase;
        long  NoFaces,NoVertices,NoMaps;
        vector Offset;
        double trpos[4][4];
        double fi,theta,alpha,ima;
        double sx,sy,sz;
        double morphratio,effectratio,ambient_light;
        short im,effect;
        BOOL    wireframe;
        BOOL    self_shadow,cast_shadow,show_shadow;
        char  *effect_param_list;
} object;

typedef struct {
        vector        pin,xin,yin;
        vector        p,x,y,n,dx,dy,dyy;
        double        lengthx,lengthy,height,range,bump,angle;
        double        spec,trans,refl,ambient_light;
        unsigned char color[3],acolor[3];
        short         type,brush,brushtype,texture,shiny,ispec,shader;
        iMap          Map;
} ground;

typedef struct {
        unsigned char Horizon[3],
                      Zenith[3],
                      Hblend[3],
                      Zblend[3];
        double blendratio;
        short  type,brush;
        unsigned char MapName[256];
} sky;

typedef struct { /* camera */
 vector ViewPoint;
 double Phi,Theta,Alpha,Ima;
 int Im;
 double sx,sy,sz;
} camera;

#ifdef MODULE_RENDER
object        *Object=NULL;
vector        ViewPoint;
light         *Lights=NULL;
ground        Ground;
sky           Sky;
XIP           *ImageProcesses=NULL;
long          ObjectCount=0,Nlights=0,CamIm=0,Nground=0,Nsky=0,
              nImageProcesses=0;
double         scalex,scaley,
               BigL=FARAWAY,CamPhi,
               CamTheta,CamAlpha,CamIma,
               CamSx=1.0,CamSy=1.0,CamSz=1.0;
double ground_ambient_light,object_ambient_light;
#else
extern object        *Object;
extern vector        ViewPoint;
extern light         *Lights;
extern ground        Ground;
extern sky           Sky;
extern XIP           *ImageProcesses;
extern long          ObjectCount,Nlights,CamIm,Nground,Nsky,
                     nImageProcesses;
extern double        scalex,scaley,
                     BigL,CamPhi,
                     CamTheta,CamAlpha,CamIma,
                     CamSx,CamSy,CamSz;
extern double ground_ambient_light,object_ambient_light;
#endif

#ifdef MODULE_RENDER
#define WHERE
#else
#define WHERE extern
#endif
WHERE void R_tram(double t[4][4], double dx, double dy, double dz);
WHERE void R_c4to4(double tin[4][4], double tout[4][4]);
WHERE void R_null_transform(double t[4][4]);
WHERE void R_m4by1(double t4[4][4], double x, double y, double z,
           double *xx, double *yy, double *zz);
void R_m4by4(double t1[4][4], double t2[4][4], double tr[4][4]);
#undef WHERE

#ifdef MODULE_LOADER
#define WHERE
#else
#define WHERE extern
#endif
WHERE char *short_form(char *filename);
WHERE char R_getl(long *val);
WHERE char R_getsht(short * val);
WHERE long LoadMorphObject(short mo, char *FileName, double mr,
                     char *QmorphFile, double qr);
WHERE long R_LoadObject(char *FileName, long frame, short robo,
                  double robo_ratio, char *robo_file1, char *robo_file2);
WHERE void R_getchunk(long clen);

#undef WHERE
