/* file SOLID.C OpenGL animator - This module does the work */

/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

#define MODULE_MAIN

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include <gl\gl.h>
#include <gl\glu.h>

#include "render.h"
#include "main.h"
#include "glmenu.h"
#include "jitter.h"

typedef GLfloat normal[3];

#define IDM_WINDOW_SHOWOPENGL 1661


/* messages sent to animator  */
#define WM_WIRE      (WM_USER+2)
#define WM_CALLBACK  (WM_USER+3)
#define WM_SWITCH    (WM_USER+4)

/* input message from animator */
#define WM_INIT      (WM_USER+1)
#define WM_FULLWIRE  (WM_USER+4)

#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define ALIGNULONG(i)   ((i+3)/4*4)

#define MAX_ATTEN       100
#define MAX_FOG_ATTEN   100

extern void FreeMemoryObjects(void);
extern int open_script(BOOL, long, char *);
extern void close_script(void);
extern void next_frame(char *);
extern GLvoid drawWire(HWND, int);

extern long frame,nframes;

static void LoadPreferencesGLA(void);
static void SavePreferencesGLA(void);
static void UpdateEverything(BOOL reset);
static BOOL normalise(GLfloat *v);
static void draw3dF(void);
static void draw3dS(void);
static void drawGround(void);
static void drawSky(void);
static GLvoid initialize(HWND);
static GLvoid drawScene(HWND,BOOL);
static GLvoid resize(HWND,int x, int y);
static BOOL bSetupPixelFormat(HDC);
static LONG WINAPI MainOpenGLWndProc(HWND, UINT, WPARAM, LPARAM);
static LONG WINAPI OpenGLWndProc(HWND, UINT, WPARAM, LPARAM);
static BOOL CALLBACK PrefsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam);
static void GluCommands(WORD, HWND);
static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent);
static PSTR FileInPath(PSTR pstrPath);
static void GetMappingCoordP(vector n, vector x, vector y, vector p0,
                             vector pv,
                             GLfloat *a, GLfloat *b);
static void GetMappingCoordC(vector n, vector x, vector y, vector p0,
                double angle,
                vector pv,
                GLfloat *a, GLfloat *b);
static void GetMapNormal(int type, vector P, vector X, vector Y,
                         vector mP, vector mX, vector mY,
                         vector mN);
static void makeCheckeredImageMap(void);
static void makeNextMap(unsigned char *pixels24, int x, int y);
static unsigned char *LoadGIF(char *GIFfile, int *x, int *y);
static unsigned char *LoadTGA(char *TGAfile, int *x, int *y);
static unsigned char *LoadXXX(char *JPGfile, int *x, int *y, long type);
static void writeScreen(void);
static void writeAVI(void);
static void accFrustum(GLdouble left, GLdouble right,
    GLdouble bottom, GLdouble top,
    GLdouble znear, GLdouble zfar, GLdouble pixdx, GLdouble pixdy,
    GLdouble eyedx, GLdouble eyedy, GLdouble focus);
static void accPerspective(GLdouble fovy, GLdouble aspect,
    GLdouble znear, GLdouble zfar, GLdouble pixdx, GLdouble pixdy,
    GLdouble eyedx, GLdouble eyedy, GLdouble focus);
static BOOL CALLBACK PrefsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam);
static void GetDepthBufferRange(GLfloat *, GLfloat *);

void RegisterOpenGL(HINSTANCE);
void UnregisterOpenGL(HINSTANCE);
HWND RegisterWindow(HWND, LPRECT);
void MakeFrameList(BOOL);
unsigned char *LoadMAP(char *name, int *x, int *y);
void MakeMapFromImage(unsigned char *, long , long );
unsigned char *MakeMapImage(unsigned char *, int, int, int *, int *);
BOOL Normalize(vector x,vector y,vector z,GLfloat *n);

char      iniFilename[]="OPENFX_.INI";
HWND      glWnd=NULL,glWndFrame=NULL,glWndParent;
WNDPROC   fpSclass=NULL;
HINSTANCE glInstance;
BOOL      RendererStandAlone=FALSE,uselists=TRUE;
#if __DLL__
static CHAR      szAppName[]="OFX:PlayDllGL";
static CHAR      szCldName[]="OFX:PlayDLLGL:Child";
#else
static CHAR      szAppName[]="OFX:PlayGL";
static CHAR      szCldName[]="OFX:PlayGL:Child";
#endif
static HWND      ghWndButton=NULL;
static HMENU     glMenu=NULL;
static CHAR      tempstr[255];
static CHAR      szFileName[256],szDirName[256],gszHomeDir[256],gszWork[256];
static CHAR      szOutFileName[256],szOutDirName[256],*gszWorkDrive;
static CHAR      szAVIFileName[256],szAVIDirName[256];
static HGLRC     hRC;
static HPALETTE  ghpalOld, ghPalette = NULL;
static RECT      oldrect;
static DWORD     fdwStyle;
static DWORD gdwStyle = WS_CHILD | WS_VISIBLE | !WS_BORDER |
                        WS_HSCROLL |
                        WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
static UINT      AnimationTimer=0,ToolbarHeight=0;
static long      thisframe,listID=0,LightAttenuation=1,FogAttenuation=1;
static BOOL      fmapped=TRUE,ffastmapping=FALSE,fantialiased=FALSE,
                 fflat=FALSE,fhalf=FALSE,fwire=FALSE,busy=FALSE,fmenu=FALSE,
                 Cmaterial=TRUE,Optimised=TRUE,fLightAttenuated=TRUE,
                 fFog=FALSE;
static HWND      hTBWnd=NULL;

//static GLfloat gbRed=0.0,gbGreen=0.1,gbBlue=0.3;
static GLfloat gbRed=0.15,gbGreen=0.15,gbBlue=0.15;

#define checkImageWidth 64
#define checkImageHeight 64
static GLubyte checkImage[checkImageWidth][checkImageHeight][3];

#define NTOOLBARBUTTONS 11
#define NOTOOLBARIMAGES 9

static TBBUTTON tbb[NTOOLBARBUTTONS] = {
          {0,0,              TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          {0,IDM_PREVIOUS,   TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {1,IDM_PLAY,       TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {2,IDM_NEXT,       TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {3,IDM_STOP,       TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {0,0,              TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          {4,IDM_SCRIPT,     TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {5,IDM_MAKEAVI,    TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {6,IDM_CAPTURE,    TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {7,IDM_PREFERENCES,TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {8,IDM_HELP,       TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 }
                                        };

static void UpdateEverything(BOOL reset){
 char temp[256];
 if(RendererStandAlone && AnimationTimer > 0){
   InvalidateRect(glWnd, NULL, FALSE);
   return;
 }
 if(reset && !RendererStandAlone)SetScrollRange(glWnd, SB_HORZ, 1,nframes,TRUE);
#if __DLL__
 if(reset)InvalidateRect(glWnd, NULL, FALSE);
#else
 if(reset)InvalidateRect(glWnd, NULL, TRUE);
#endif
 else     InvalidateRect(glWnd, NULL, FALSE);
 if(RendererStandAlone){
   sprintf(temp,"Script %s : %ld",short_form(szFileName),frame);
   SendMessage(glWndFrame,WM_SETTEXT,0,(LPARAM)temp);
 }
#if !__DLL__
 else{
   sprintf(temp,"Frame %ld",frame);
   SendMessage(glWndFrame,WM_SETTEXT,0,(LPARAM)temp);
 }
#endif
 SetScrollPos(glWnd,SB_HORZ,frame,TRUE);
}

void UnregisterOpenGL(HINSTANCE hInstance){
 UnregisterClass(szAppName,hInstance);
 UnregisterClass(szCldName,hInstance);
}

void RenderYield(void){
 MSG msg;
 while (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE)){
   TranslateMessage(&msg);
   DispatchMessage(&msg);
 }
}

void RegisterOpenGL(HINSTANCE hInstance){
 WNDCLASS   wc;
 wc.style         = 0;
 wc.lpfnWndProc   = (WNDPROC)MainOpenGLWndProc;
 wc.cbClsExtra    = 0;
 wc.cbWndExtra    = 0;
 wc.hInstance     = hInstance;
 wc.hIcon         = LoadIcon(hInstance,"SHOWICON");
 wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
// wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
 wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
#if __DLL__
 wc.lpszMenuName  = NULL;
#else
// if(!RendererStandAlone) wc.lpszMenuName  = NULL;
// else {fmenu=TRUE;       wc.lpszMenuName  = "GL_Menu";}
 wc.lpszMenuName  = NULL;
#endif
 wc.lpszClassName = szAppName;
 glInstance=hInstance;
 if (!RegisterClass (&wc) )return;
 wc.style         = 0;
 wc.lpfnWndProc   = (WNDPROC)OpenGLWndProc;
 wc.cbClsExtra    = 0;
 wc.cbWndExtra    = 0;
 wc.hInstance     = hInstance;
 wc.hIcon         = NULL;
 wc.hCursor       = LoadCursor(NULL,IDC_ARROW);
// wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
 wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
 wc.lpszMenuName  = NULL;
 wc.lpszClassName = szCldName;
 RegisterClass(&wc);
 return;
}

HWND RegisterWindow(HWND parent, LPRECT pr){
 RECT rc;
 HWND hWnd;
 glWndParent=parent;
 GetModuleFileName(glInstance,gszHomeDir,256);
 *FileInPath(gszHomeDir) = '\0';
 strcpy(szFileName,"temp");
 strcpy(szDirName,gszHomeDir); strcat(szDirName,"scripts");
 strcpy(szOutFileName,"temp.bmp");
 strcpy(szOutDirName,gszHomeDir); strcat(szOutDirName,"scripts");
 strcpy(szAVIFileName,"temp.avi");
 strcpy(szAVIDirName,gszHomeDir); strcat(szAVIDirName,"movies");
 if(GetEnvironmentVariable((LPTSTR)"OFX",(LPTSTR)gszWork,255) == 0){
   strcpy(gszWork,gszHomeDir);
 }
 gszWorkDrive=&gszWork[0];
 LoadPreferencesGLA();
 if(parent == NULL)fdwStyle = WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
                              WS_SYSMENU | WS_MAXIMIZEBOX |
                              WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
 else              fdwStyle = WS_CHILD | !WS_CAPTION | !WS_SYSMENU |
                              !WS_MAXIMIZEBOX | WS_BORDER |
                              WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
 LoadString(glInstance,IDX_MISC_CAMERA,tempstr,256);
 hWnd = CreateWindow(szAppName,
             tempstr,
             fdwStyle,
             pr->left,
             pr->top,
             pr->right,
             pr->bottom,
             parent,
             NULL,
             glInstance,
             NULL);
 if (!hWnd)return NULL;
 initialize(hWnd);
 if(IsWindow(glWnd))GetClientRect(glWnd,&oldrect);
 if(parent == NULL)SetWindowPos(hWnd,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE | SWP_NOSIZE);
 ShowWindow (hWnd, SW_SHOW);
 UpdateWindow(hWnd);
 glWndFrame=hWnd;
 return hWnd;
}

static LRESULT CALLBACK SubclassButtonProc(HWND hwnd,UINT msg,
                                       WPARAM wParam,LPARAM lParam){
 // To allow the focus to be returned to the main WINANI window
 LRESULT lRet;
 int hid;
 MSG mmsg;
 mmsg.hwnd=hwnd;
 mmsg.message=msg;
 mmsg.wParam=wParam;
 mmsg.lParam=lParam;
 switch (msg) {
   case WM_KEYDOWN:
     SetFocus(glWndParent);
     return 0;
   default: break;
 }
 lRet=CallWindowProc(fpSclass,hwnd,msg,wParam,lParam);
 switch (msg) {
   case WM_LBUTTONUP:
     SetFocus(glWndParent);
     break;
   default: break;
 }
 return lRet;
}


static LONG WINAPI MainOpenGLWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam){
 LONG lRet = 1;
 switch (uMsg){
#if !__DLL__
   case WM_NCACTIVATE:
     return TRUE;
   case WM_KEYDOWN:
     if(glWndParent != NULL)SetActiveWindow(glWndParent);
     break;
#endif
   case WM_CREATE:{
       HMENU hmenu = GetSystemMenu(hWnd, FALSE);
       RECT rc;
#if !__DLL__
       hTBWnd = CreateToolbarEx(
         hWnd,
         WS_CHILD | WS_VISIBLE | !WS_BORDER | TBSTYLE_TOOLTIPS | CCS_BOTTOM,
         0,
         NOTOOLBARIMAGES,
         glInstance,
         TOOLBAR_IDBM,
         tbb,
         NTOOLBARBUTTONS,
         16,15, // button sizes
         16,15, // bitmap image sizes
         sizeof(TBBUTTON)
       );
       if(!RendererStandAlone)SendMessage(hTBWnd,
         TB_HIDEBUTTON,(WPARAM)IDM_SCRIPT,(LPARAM)MAKELONG(TRUE,0));
       GetClientRect(hTBWnd,&rc); ToolbarHeight=rc.bottom-rc.top;
#endif
       GetClientRect(hWnd,&rc);
       glWnd=CreateWindow(
         szCldName,
         NULL,
         gdwStyle,
         rc.left,
         rc.top,
         rc.right,
         rc.bottom-ToolbarHeight,
         hWnd,
         (HMENU)1,
         glInstance,
         NULL);
#if !__DLL__
       RemoveMenu(hmenu,7,MF_BYPOSITION);    // Second separator.
       RemoveMenu(hmenu,SC_MINIMIZE,MF_BYCOMMAND);
       RemoveMenu(hmenu,SC_TASKLIST,MF_BYCOMMAND);
       AppendMenu(hmenu,MF_SEPARATOR,0,0);
       LoadString(glInstance,IDX_MISC_NEXT,tempstr,256);
       AppendMenu(hmenu,MF_GRAYED | MF_STRING,IDM_NEXT,tempstr);
       LoadString(glInstance,IDX_MISC_PLAY,tempstr,256);
       AppendMenu(hmenu,MF_GRAYED | MF_STRING,IDM_PLAY,tempstr);
       LoadString(glInstance,IDX_MISC_STOP,tempstr,256);
       AppendMenu(hmenu,MF_GRAYED | MF_STRING,IDM_STOP,tempstr);
       AppendMenu(hmenu,MF_SEPARATOR,0,0);
       LoadString(glInstance,IDX_MISC_OPENSCRIPT,tempstr,256);
       AppendMenu(hmenu,MF_ENABLED | MF_STRING,IDM_SCRIPT,tempstr);
       LoadString(glInstance,IDX_MISC_SAVEIMAGE,tempstr,256);
       AppendMenu(hmenu,MF_GRAYED | MF_STRING,IDM_CAPTURE,tempstr);
       LoadString(glInstance,IDX_MISC_WRITEAVI,tempstr,256);
       AppendMenu(hmenu,MF_GRAYED | MF_STRING,IDM_MAKEAVI,tempstr);
       AppendMenu(hmenu,MF_SEPARATOR,0,0);
       LoadString(glInstance,IDX_MISC_SETTINGS,tempstr,256);
       AppendMenu(hmenu,MF_ENABLED | MF_STRING,IDM_PREFERENCES,tempstr);
       LoadString(glInstance,IDX_MISC_ABOUT,tempstr,256);
       AppendMenu(hmenu,MF_ENABLED | MF_STRING,IDM_HELP,tempstr);
#else
       ghWndButton=CreateWindow("BUTTON",
             "CAMERA",
             WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
             0,
             0,
             60,
             20,
             glWnd,
             (HMENU)1,
             glInstance,
             NULL);
       SendMessage(ghWndButton,WM_SETFONT,
                  (WPARAM)GetStockObject(ANSI_VAR_FONT),0);
       fpSclass=(WNDPROC)SetWindowLong(ghWndButton,GWL_WNDPROC,
         (LONG)SubclassButtonProc);
       hmenu=GetMenu(glWndParent);
       glMenu=CreatePopupMenu();
       LoadString(glInstance,IDX_MISC_SETTINGS,tempstr,256);
       AppendMenu(glMenu,MF_ENABLED | MF_STRING,IDM_GLCAMERA_PREFERENCES,
                  tempstr);
       LoadString(glInstance,IDX_MISC_ABOUT,tempstr,256);
       AppendMenu(glMenu,MF_ENABLED | MF_STRING,IDM_GLCAMERA_HELP,tempstr);
       LoadString(glInstance,IDX_MISC_CAMERAGL,tempstr,256);
       InsertMenu(hmenu,GLMENU_POSITION,MF_POPUP|MF_BYPOSITION,
                 (UINT)glMenu,tempstr);
       DrawMenuBar(glWndParent);
#endif
     }
     break;
   case WM_TIMER:
     if(AnimationTimer == 1){
       next_frame(szFileName);
       UpdateEverything(FALSE);
     }
     break;
   case WM_SIZE:
     if(AnimationTimer > 0)return lRet;
     if(hTBWnd != NULL)SendMessage(hTBWnd,uMsg,wParam,lParam);
     SetWindowPos(glWnd,NULL,
                  0,0,
                  LOWORD(lParam),HIWORD(lParam)-ToolbarHeight,
                  SWP_NOZORDER);
     break;
   case WM_QUERYNEWPALETTE:{
       HDC     hDC;
       if(ghPalette != NULL){
         hDC = GetDC(hWnd);
         ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
         RealizePalette(hDC);
         InvalidateRect(hWnd, NULL, TRUE);
         UpdateWindow(hWnd);
         if(ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
         ReleaseDC(hWnd, hDC);
         return TRUE;
       }
       return FALSE;
     }
   case WM_PALETTECHANGED:{
       HDC         hDC;
       if(ghPalette != NULL){
         if (wParam != (WPARAM)hWnd){
           hDC = GetDC(hWnd);
           ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
           RealizePalette(hDC);
           UpdateColors(hDC);
           if(ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
           ReleaseDC(hWnd, hDC);
         }
       }
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wParam & 0xfff0)){
       case IDM_SCRIPT:
         PostMessage(hWnd,WM_COMMAND,IDM_SCRIPT,0);
         break;
       case IDM_PREVIOUS:
         PostMessage(hWnd,WM_COMMAND,IDM_PREVIOUS,0);
         break;
       case IDM_NEXT:
         PostMessage(hWnd,WM_COMMAND,IDM_NEXT,0);
         break;
       case IDM_PLAY:
         PostMessage(hWnd,WM_COMMAND,IDM_PLAY,0);
         break;
       case IDM_STOP:
         PostMessage(hWnd,WM_COMMAND,IDM_STOP,0);
         break;
       case IDM_HELP:
         PostMessage(hWnd,WM_COMMAND,IDM_HELP,0);
         break;
       case IDM_CAPTURE:
         PostMessage(hWnd,WM_COMMAND,IDM_CAPTURE,0);
         break;
       case IDM_MAKEAVI:
         PostMessage(hWnd,WM_COMMAND,IDM_MAKEAVI,0);
         break;
       case IDM_PREFERENCES:
         PostMessage(hWnd,WM_COMMAND,IDM_PREFERENCES,0);
         break;
       default:
         return( DefWindowProc(hWnd,uMsg,wParam,lParam));
      }
      break;
   case WM_NOTIFY:
     switch (((LPNMHDR)lParam)->code){
       case TTN_NEEDTEXT:{
           int idButton;
           LPTOOLTIPTEXT lpttt;
           lpttt=(LPTOOLTIPTEXT)lParam;
           lpttt->hinst=glInstance;
           idButton=lpttt->hdr.idFrom;
           switch(idButton){
             case IDM_PREVIOUS:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_1); break;
             case IDM_PLAY:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_2); break;
             case IDM_NEXT:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_3); break;
             case IDM_STOP:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_4); break;
             case IDM_PREFERENCES:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_5); break;
             case IDM_SCRIPT:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_6); break;
             case IDM_MAKEAVI:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_7); break;
             case IDM_CAPTURE:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_8); break;
             case IDM_HELP:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_9); break;
           }
         }
         break;
       default: break;
     }
     return (DefWindowProc(hWnd,uMsg,wParam,lParam));
   case WM_INIT:
     if(busy)break;
     strcpy(szFileName,gszWorkDrive);
#if __DLL__
     strcat(szFileName,"x_$.scr");
#else
     strcat(szFileName,"y_$.scr");
#endif
     open_script(TRUE,1,szFileName);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_NEXT,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_PLAY,MF_ENABLED);
     if(nframes > 1)EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_ENABLED);
     else           EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_CAPTURE,MF_ENABLED);
     if(fmenu){
       HMENU hmenu=GetMenu(hWnd);
       EnableMenuItem(hmenu,IDM_NEXT,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_PLAY,MF_ENABLED);
       if(nframes > 1)EnableMenuItem(hmenu,IDM_MAKEAVI,MF_ENABLED);
       else           EnableMenuItem(hmenu,IDM_MAKEAVI,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_CAPTURE,MF_ENABLED);
       DrawMenuBar(hWnd);
     }
     UpdateEverything(TRUE);
     break;
   case WM_WIRE:
     if(busy)break;
#if __DLL__
     fwire=TRUE;
     drawWire(glWnd,0);
     fwire=FALSE;
#endif
     break;
   case WM_FULLWIRE:
     if(busy)break;
#if __DLL__
     fwire=TRUE;
     drawWire(glWnd,1);
     fwire=FALSE;
#endif
     break;
   case WM_COMMAND:
     GluCommands(LOWORD(wParam),hWnd);
     break;
   case WM_CLOSE:
     if(glWndParent == NULL)PostQuitMessage(0);
     else PostMessage(glWndParent,WM_COMMAND,IDM_WINDOW_SHOWOPENGL,0);
     break;
   case WM_DESTROY:{
#if __DLL__
       if(fpSclass != NULL)SetWindowLong(ghWndButton,GWL_WNDPROC,
                                         (LONG)fpSclass);
       fpSclass=NULL;
       RemoveMenu(GetMenu(glWndParent),GLMENU_POSITION,MF_BYPOSITION);
       if(glMenu != NULL)DestroyMenu(glMenu);
       DrawMenuBar(glWndParent);
       glMenu=NULL;
#else
       RECT    rc;
       POINT   pt1,pt2;
       char    str[64];
       if(!IsZoomed(hWnd)){
         GetWindowRect(hWnd,&rc);
         pt1.x=rc.left;  pt1.y=rc.top;
         pt2.x=rc.right; pt2.y=rc.bottom;
         sprintf(str,"%ld %ld %ld %ld",pt1.x,pt1.y,pt2.x,pt2.y);
         WritePrivateProfileString("GLAPREFERENCES","GLWINDOW",str,iniFilename);
       }
       FreeMemoryObjects();
#endif
       close_script();
     }
     break;
   default:
     lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
   break;
 }
 return lRet;
}

static LONG WINAPI OpenGLWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam){
 char temp[256];
 GLfloat old;
 int i,iPos,iMax,iMin,dn;
 LONG lRet = 1;
 switch (uMsg){
#if !__DLL__
   case WM_NCACTIVATE:
     return TRUE;
   case WM_KEYDOWN:
     if(glWndParent != NULL)SetActiveWindow(glWndParent);
     break;
#else
   case WM_COMMAND:
     if(HIWORD(wParam) == BN_CLICKED &&
        LOWORD(wParam) == 1 &&
        (HWND)lParam == ghWndButton){
       if(IsWindow(glWndParent))PostMessage(glWndParent,WM_SWITCH,0,0);
     }
     break;
#endif
   case WM_CREATE:{
       HDC hDC;
       HMENU hmenu = GetSystemMenu(hWnd, FALSE);
       hDC = GetDC(hWnd);
       bSetupPixelFormat(hDC);
       hRC = wglCreateContext( hDC );
       wglMakeCurrent( hDC, hRC );
       SetScrollRange(hWnd, SB_HORZ, 1,1,TRUE);
       SetScrollPos(hWnd,SB_HORZ,1,TRUE);
     }
     break;
   case WM_PAINT:{
       HDC hDC;
       PAINTSTRUCT  ps;
       hDC = BeginPaint(hWnd, &ps);
       if(!fwire)drawScene(hWnd,TRUE);
       EndPaint(hWnd, &ps);
     }
     break;
   case WM_LBUTTONDOWN:
#if __DLL__
     if(IsWindow(glWndParent))PostMessage(glWndParent,WM_CALLBACK,0,0);
#else
     InvalidateRect(hWnd, NULL, FALSE);
#endif
     break;
   case WM_RBUTTONDOWN:{
#if __DLL__
       POINT pt;
       HMENU hMenu=LoadMenu(glInstance,"GLCameraViewPopupMenu");
       if(hMenu == NULL)break;
       pt.x=LOWORD(lParam);
       pt.y=HIWORD(lParam);
       ClientToScreen(hWnd,&pt);
       if(fmapped && (!Optimised))CheckMenuItem(hMenu,IDM_OGL_MAPS,MF_CHECKED);
       else                       CheckMenuItem(hMenu,IDM_OGL_MAPS,MF_UNCHECKED);
       TrackPopupMenu(GetSubMenu(hMenu,0),TPM_LEFTALIGN | TPM_LEFTBUTTON,
                      pt.x,pt.y,0,glWndFrame,NULL);
       DestroyMenu(hMenu);
#endif
       break;
     }
   case WM_HSCROLL:{
       if(AnimationTimer > 0)break;
       GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
       iPos = GetScrollPos(hWnd, SB_HORZ);
       switch (LOWORD(wParam)) {
         case SB_ENDSCROLL:
           dn=0;
           if(thisframe > 0 && thisframe <= nframes){
             open_script(FALSE,thisframe,szFileName);
             UpdateEverything(FALSE);
           }
           break;
         case SB_LINEDOWN:      dn =  1;     break;
         case SB_LINEUP:        dn = -1;     break;
         case SB_PAGEDOWN:      dn =  5;     break;
         case SB_PAGEUP:        dn = -5;     break;
         case SB_THUMBTRACK:
         case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wParam) - iPos; break;
         default:               dn = 0;
       }
       if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
         SetScrollPos(hWnd, SB_HORZ, iPos + dn, TRUE);
         thisframe=(iPos+dn);
         if(RendererStandAlone){
           sprintf(temp,"Script %s : %ld",short_form(szFileName),thisframe);
           SendMessage(glWndFrame,WM_SETTEXT,0,(LPARAM)temp);
         }
#if !__DLL__
         else{
           sprintf(temp,"Frame %ld",thisframe);
           SendMessage(glWndFrame,WM_SETTEXT,0,(LPARAM)temp);
         }
#endif
       }
     }
     break;
   case WM_SIZE:
     if(AnimationTimer > 0)return lRet;
     resize(hWnd,(int)LOWORD(lParam),(int)HIWORD(lParam));
     lRet=0;
     break;
   case WM_QUERYNEWPALETTE:{
       HDC     hDC;
       if(ghPalette != NULL){
         hDC = GetDC(hWnd);
         ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
         RealizePalette(hDC);
         InvalidateRect(hWnd, NULL, TRUE);
         UpdateWindow(hWnd);
         if(ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
         ReleaseDC(hWnd, hDC);
         return TRUE;
       }
       return FALSE;
     }
   case WM_PALETTECHANGED:{
       HDC         hDC;
       if(ghPalette != NULL){
         if (wParam != (WPARAM)hWnd){
           hDC = GetDC(hWnd);
           ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
           RealizePalette(hDC);
           UpdateColors(hDC);
           if(ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
           ReleaseDC(hWnd, hDC);
         }
       }
     }
     break;
   case WM_DESTROY:{
       HGLRC hRC;
       HDC    hDC;
       RECT    rc;
       hRC = wglGetCurrentContext();
       hDC = wglGetCurrentDC();
       wglMakeCurrent(NULL, NULL);
       if(hRC != NULL)wglDeleteContext(hRC);
       if(hDC != NULL)ReleaseDC(hWnd, hDC);
       glWnd=NULL;
    }
     break;
   default:
     lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
   break;
 }
 return lRet;
}

static void AppendFileExtension(char *filename, char *ext){
 if(strrchr(filename,'.') == NULL){
   strcat(filename,ext);
 }
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[80],szdummy[256];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=256;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=256;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

BOOL CALLBACK AboutDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     return( TRUE );
   case WM_COMMAND:
      if( wparam == IDOK ) {
        EndDialog( hwnd, TRUE );
        return( TRUE );
      }
      break;
 }
 return( FALSE );
}

static void GluCommands(WORD cmd,HWND hWnd){
 int i;
 char temp[256];
 switch (cmd){
   case IDM_OGL_QUIT:  // from popup menu
     PostMessage(glWndFrame,WM_CLOSE,0,0);
     break;
   case IDM_EXIT:
     if(glWndParent == NULL)PostQuitMessage(0);
     break;
   case IDM_SCRIPT:
     if(SelectFileName(0,szFileName,szDirName,szAppName,
             "(*.SCR)|*.scr|",hWnd) == TRUE){
       open_script(TRUE,1,szFileName);
       EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_NEXT,MF_ENABLED);
       EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_PLAY,MF_ENABLED);
       if(nframes > 1)EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_ENABLED);
       else           EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_GRAYED);
       EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_CAPTURE,MF_ENABLED);
       if(fmenu){
         HMENU hmenu=GetMenu(hWnd);
         EnableMenuItem(hmenu,IDM_NEXT,MF_ENABLED);
         EnableMenuItem(hmenu,IDM_PLAY,MF_ENABLED);
         if(nframes > 1)EnableMenuItem(hmenu,IDM_MAKEAVI,MF_ENABLED);
         else           EnableMenuItem(hmenu,IDM_MAKEAVI,MF_GRAYED);
         EnableMenuItem(hmenu,IDM_CAPTURE,MF_ENABLED);
         DrawMenuBar(hWnd);
       }
       UpdateEverything(TRUE);
     }
     break;
   case IDM_PREVIOUS:
     thisframe = frame-1;
     if(thisframe > 0 && thisframe <= nframes){
       open_script(FALSE,thisframe,szFileName);
       UpdateEverything(FALSE);
     }
     break;
   case IDM_NEXT:
     next_frame(szFileName);
     UpdateEverything(FALSE);
     break;
   case IDM_PLAY:
     if(hTBWnd != NULL)for(i=0;i<NTOOLBARBUTTONS;i++){
       if(tbb[i].idCommand != IDM_STOP)SendMessage(hTBWnd,
       TB_ENABLEBUTTON,(WPARAM)tbb[i].idCommand,(LPARAM)MAKELONG(FALSE,0));
     }
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_NEXT,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_SCRIPT,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_PLAY,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_STOP,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_CAPTURE,MF_GRAYED);
     if(fmenu){
       HMENU hmenu=GetMenu(hWnd);
       EnableMenuItem(hmenu,IDM_NEXT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_SCRIPT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_PLAY,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_STOP,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_MAKEAVI,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_CAPTURE,MF_GRAYED);
       DrawMenuBar(hWnd);
     }
     AnimationTimer=SetTimer(hWnd,1,30,NULL);
//     AnimationTimer=SetTimer(hWnd,1,5,NULL);
     break;
   case IDM_STOP:
     if(hTBWnd != NULL)for(i=0;i<NTOOLBARBUTTONS;i++){
       if(tbb[i].idCommand != IDM_STOP)SendMessage(hTBWnd,
       TB_ENABLEBUTTON,(WPARAM)tbb[i].idCommand,(LPARAM)MAKELONG(TRUE,0));
     }
     if(AnimationTimer != 0)KillTimer(hWnd,AnimationTimer);
     AnimationTimer=0;
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_NEXT,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_SCRIPT,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_PLAY,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_STOP,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_CAPTURE,MF_ENABLED);
     if(fmenu){
       HMENU hmenu=GetMenu(hWnd);
       EnableMenuItem(hmenu,IDM_NEXT,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_SCRIPT,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_PLAY,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_STOP,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_MAKEAVI,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_CAPTURE,MF_ENABLED);
       DrawMenuBar(hWnd);
     }
     if(RendererStandAlone)UpdateEverything(FALSE);
     break;
   case IDM_GLCAMERA_HELP:
   case IDM_HELP:
     DialogBox(glInstance,MAKEINTRESOURCE(IDD_ABOUT),hWnd,AboutDlgProc);
     break;
   case IDM_CAPTURE:
     writeScreen();
     break;
#if !__DLL__
   case IDM_MAKEAVI:
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_NEXT,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_SCRIPT,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_PLAY,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_STOP,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_GRAYED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_CAPTURE,MF_GRAYED);
     if(fmenu){
       HMENU hmenu=GetMenu(hWnd);
       EnableMenuItem(hmenu,IDM_NEXT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_SCRIPT,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_PLAY,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_STOP,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_MAKEAVI,MF_GRAYED);
       EnableMenuItem(hmenu,IDM_CAPTURE,MF_GRAYED);
       DrawMenuBar(hWnd);
     }
     AnimationTimer=1; busy=TRUE;
     writeAVI();
     AnimationTimer=0; busy=FALSE;
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_NEXT,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_SCRIPT,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_PLAY,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_STOP,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_MAKEAVI,MF_ENABLED);
     EnableMenuItem(GetSystemMenu(hWnd, FALSE),IDM_CAPTURE,MF_ENABLED);
     if(fmenu){
       HMENU hmenu=GetMenu(hWnd);
       EnableMenuItem(hmenu,IDM_NEXT,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_SCRIPT,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_PLAY,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_STOP,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_MAKEAVI,MF_ENABLED);
       EnableMenuItem(hmenu,IDM_CAPTURE,MF_ENABLED);
       DrawMenuBar(hWnd);
     }
     break;
#endif
   case IDM_GLCAMERA_PREFERENCES:
   case IDM_PREFERENCES:
     DialogBox(glInstance,MAKEINTRESOURCE(DLG_PREFS),hWnd,
               (DLGPROC)PrefsDlgProc);
     MakeFrameList(uselists);
     UpdateEverything(FALSE);
     break;
   case IDM_OGL_MAPS:
     fmapped ^= 1;
     if(fmapped)Optimised=FALSE;
     else       Optimised=TRUE;
     MakeFrameList(uselists);
     UpdateEverything(FALSE);
     break;
   default:
     break;
 }
}

static unsigned char threeto8[8] = {
    0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

static unsigned char twoto8[4] = {
    0, 0x55, 0xaa, 0xff
};

static unsigned char oneto8[2] = {
    0, 255
};

static int defaultOverride[13] = {
    0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift){
 unsigned char val;
 val = (unsigned char) (i >> shift);
 switch (nbits) {
    case 1:
        val &= 0x1;
        return oneto8[val];
    case 2:
        val &= 0x3;
        return twoto8[val];
    case 3:
        val &= 0x7;
        return threeto8[val];
    default:
        return 0;
 }
}

static void CreateRGBPalette(HDC hDC){
 PIXELFORMATDESCRIPTOR pfd;
 LOGPALETTE *pPal;
 int n, i;
 n = GetPixelFormat(hDC);
 DescribePixelFormat(hDC, n, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
 if(pfd.dwFlags & PFD_NEED_PALETTE) {
   n = 1 << pfd.cColorBits;
   pPal = (PLOGPALETTE)LocalAlloc(LMEM_FIXED, sizeof(LOGPALETTE) +
                                   n * sizeof(PALETTEENTRY));
   pPal->palVersion = 0x300;
   pPal->palNumEntries = n;
   for (i=0; i<n; i++) {
     pPal->palPalEntry[i].peRed =
            ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
     pPal->palPalEntry[i].peGreen =
            ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
     pPal->palPalEntry[i].peBlue =
            ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
     pPal->palPalEntry[i].peFlags = 0;
   }
   if((pfd.cColorBits == 8)                           &&
       (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
       (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
       (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)) {
     for(i = 1 ; i <= 12 ; i++)
        pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
   }
   ghPalette = CreatePalette(pPal);
   LocalFree(pPal);
   ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
   n = RealizePalette(hDC);
 }
}

static BOOL bSetupPixelFormat(HDC hDC){
 static PIXELFORMATDESCRIPTOR pfd = {
   sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
   1,                              // version number
   PFD_DRAW_TO_WINDOW |            // support window
   PFD_SUPPORT_OPENGL |            // support OpenGL
   PFD_DOUBLEBUFFER,               // double buffered
   PFD_TYPE_RGBA,                  // RGBA type
   24,                             // 24-bit color depth
   0, 0, 0, 0, 0, 0,               // color bits ignored
   0,                              // no alpha buffer
   0,                              // shift bit ignored
   0,                              // no accumulation buffer are we shure ???
   0, 0, 0, 0,                     // accum bits ignored
   32,                             // 32-bit z-buffer
   0,                              // no stencil buffer
   0,                              // no auxiliary buffer
   PFD_MAIN_PLANE,                 // main layer
   0,                              // reserved
   0, 0, 0                         // layer masks ignored
 };
 int pixelformat;
 if ( (pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0 ){
   MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
   return FALSE;
 }
 if (SetPixelFormat(hDC, pixelformat, &pfd) == FALSE) {
   MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
   return FALSE;
 }
 CreateRGBPalette(hDC);
 return TRUE;
}

static GLvoid initialize(HWND hWnd){
 GLfloat d[]={1.0,1.0,1.0,1.0};
 GLfloat a[]={0.2,0.2,0.2,1.0};
 GLfloat s[]={1.0,1.0,1.0,1.0};
 glClearDepth(1.0);
 glEnable(GL_DEPTH_TEST);
 glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
 glEnable(GL_LIGHT0);
 glLightfv(GL_LIGHT0,GL_DIFFUSE,d);
 glLightfv(GL_LIGHT0,GL_AMBIENT,a);
 glLightfv(GL_LIGHT0,GL_SPECULAR,s);
 makeCheckeredImageMap();
}

static unsigned char black[3]={0,0,0};

static BOOL same_color(unsigned char *color){
 static unsigned char lastcolor[3]={0,0,0};
 if(abs(color[0] - lastcolor[0]) > 9 ||
    abs(color[1] - lastcolor[1]) > 9 ||
    abs(color[2] - lastcolor[2]) > 9){
   VECCOPY(color,lastcolor)
   return FALSE;
 }
 return TRUE;
}

static GLvoid resize(HWND hWnd,int x, int y){
 RECT  rect;
// GetClientRect(hWnd, &rect);
// if(rect.bottom-rect.top < 1)return;
// if(rect.right-rect.left < 1)return;
 rect.right=x; rect.bottom=y;
 glViewport(0, 0, rect.right, rect.bottom);
 oldrect.right = rect.right;
 oldrect.bottom = rect.bottom;
 MakeFrameList(uselists);
// PostMessage (hWnd, WM_PAINT, 0, 0L);
}

static BOOL normalise(GLfloat *v){
 GLfloat n,nn;
 n= (GLfloat)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return FALSE;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE(n,v,v)
 return TRUE;
}

BOOL Normalize(vector x,vector y,vector z,GLfloat *n){
 GLfloat dz[3],dy[3];
 VECSUB((GLfloat)y,(GLfloat)x,dy)
 VECSUB((GLfloat)z,(GLfloat)x,dz)
 CROSS(dy,dz,n)
 if(normalise(n))return TRUE;
 return FALSE;
}

static GLvoid drawScene(HWND hWnd,BOOL shaded){
 HDC hDC;
 if(uselists){
   if(listID > 0)glCallList(listID);
 }
 else MakeFrameList(FALSE);
 glFinish();
 hDC = wglGetCurrentDC();
 SwapBuffers(hDC);
}

#include "makelist.c"

#include "maps.c"

#include "accpersp.c"
