/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* hilite.c  image post-processor                                         */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <windows.h>
#include "struct.h"           /* general structures    */
#include "ximage.h"           /* XIMAGE structure      */
#include "local.h"

#include "hilite.h"

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL; /* use to pick up resources from DLL   */
static long version=1;

#include "utils.h"

#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BC__
BOOL WINAPI DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

static void SetColour(unsigned char *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(unsigned char)GetRValue(cc.rgbResult);
   colour[1]=(unsigned char)GetGValue(cc.rgbResult);
   colour[2]=(unsigned char)GetBValue(cc.rgbResult);
 }
}

static void DrawInBuffer1(long x, long y, fullscreenbuffer *Screen,
                          long X, long Y, long radius,
                          double r, double g, double b){
 /* mix in a white halo at the screen buffer location of the light source */
 long i,j;
 double d;
 fullscreenbuffer *S;
 for(i=x-radius;i<=x+radius;i++)for(j=y-radius;j<=y+radius;j++){
   if(i < 0 || i >= X || j < 0 || j >= Y)continue;
   S=(Screen + j*X + i);
   if(radius < 1)d=1.0;
   else{
     d=sqrt((double)(i-x)*(double)(i-x) + (double)(j-y)*(double)(j-y));
     d=255.0*((double)radius-d)/(double)radius;
     d=max(0.0,d);
   }
   /* mix to colour */
   S->R = (unsigned char)min(255.0,(double)S->R+d*r);
   S->G = (unsigned char)min(255.0,(double)S->G+d*g);
   S->B = (unsigned char)min(255.0,(double)S->B+d*b);
 }
 return;
}

void DrawInBuffer(fullscreenbuffer *Screen, int Xmax, int Ymax,
                  int x, int y, double v,
                  double rr, double gg, double bb){
 int r,g,b;
 fullscreenbuffer *S;
 if(x < 0 || y < 0 || x >= Xmax || y >= Ymax)return;
 S=(Screen+y*Xmax+x);
 r=(int)S->R+(int)((double)v*rr);
 g=(int)S->G+(int)((double)v*gg);
 b=(int)S->B+(int)((double)v*bb);
 S->R=(r>255) ? 255:r;
 S->G=(g>255) ? 255:g;
 S->B=(b>255) ? 255:b;
}

#define frac(z) fmod(z,1.0)

void DrawAApixel(double x, double y, double w, double h, double C,
                 fullscreenbuffer *S, int X, int Y,
                 double r, double g, double b){
  int xa,xb,ya,yb,i,j,brite;
  double x2,y2,lfrac,rfrac,tfrac,bfrac,xfrac,yfrac;
//  if(w > 4 || h > 4){
//    DrawInBuffer1((long)x,(long)y,S,X,Y,(long)w,
//                   r*C/255.0,g*C/255.0,b*C/255.0);
//    return;
//  }
  brite=(int)C;
  if(x < 0 || y < 0)return;
  x -= w*0.5; y -= h*0.5;
  x2 = x + w;
  y2 = y + w;
  xa = (int)x;
  xb = (int)x2;
  ya = (int)y;
  yb = (int)y2;
  lfrac = 1.0-frac(x);
  rfrac = frac(x2);
  tfrac = 1.0-frac(y);
  bfrac = frac(y2);
  if (xa==xb) {
    xfrac = lfrac+rfrac-1.0;
    if (ya==yb) {
      DrawInBuffer(S,X,Y,xa,ya,xfrac*(tfrac+bfrac-1.0)*brite,r,g,b);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,xfrac*tfrac*brite,r,g,b);
      for (j=ya+1; j<yb; j++) DrawInBuffer(S,X,Y,xa,j,xfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xa,yb,xfrac*bfrac*brite,r,g,b);
    }
  }
  else {
    if (ya==yb) {
      yfrac = tfrac+bfrac-1.0;
      DrawInBuffer(S,X,Y,xa,ya,yfrac*lfrac*brite,r,g,b);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,ya,yfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xb,ya,yfrac*rfrac*brite,r,g,b);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,tfrac*lfrac*brite,r,g,b);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,ya,tfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xb,ya,tfrac*rfrac*brite,r,g,b);
      for (j=ya+1; j<yb; j++) {
        DrawInBuffer(S,X,Y,xa,j,lfrac*brite,r,g,b);
        for (i=xa+1; i<xb; i++)DrawInBuffer(S,X,Y,i,j,brite,r,g,b);
        DrawInBuffer(S,X,Y,xb,j,rfrac*brite,r,g,b);
      }
      DrawInBuffer(S,X,Y,xa,yb,bfrac*lfrac*brite,r,g,b);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,yb,bfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xb,yb,bfrac*rfrac*brite,r,g,b);
    }
  }
}

#define DIST(i,j)  sqrt((double)i * (double)i + (double)j * (double)j)

static double GetW(double l, double dx, double dy,
                   double Wmin, double Wmax,
                   BOOL linear){
 double d,f;
 d=DIST(dx,dy);
 f=d/l;
 return (Wmax-(Wmax-Wmin)*f);
}

static double GetC(double l, double dx, double dy,
                   double Wmin, double Wmax,
                   BOOL linear){
 double d,f;
 d=DIST(dx,dy);
 f=d/l;
 if(f < 0.5)return Wmax;
 return (Wmin+(Wmax-Wmin)*2.0*(1.0-f));
}

static void DrawFlareInBuffer(long x, long y, long idx, long idy,
                              fullscreenbuffer *Screen,
                              long X, long Y,
                              double Cin,
                              double r, double g, double b,
                              double Wmin, double Wmax,
                              double z, float *Zb,
                              BOOL bDraw, long hide, BOOL linear){
 long i,j;
 double grad,error,W,l,C;
 C=Cin;
 l=DIST(idx,idy);
 if(abs(idx) >= abs(idy)){
   grad=(double)idy/(double)idx; j=y;
   if(idy >= 0)error = -0.5; else error = 0.5;
   if(idx >= 0){
     for(i=x;i<x+idx;i++){
       error += grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(i-x),(double)(j-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y,r,g,b);
     }
   }
   else{
     for(i=x;i>x+idx;i--){
       error -= grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(i-x),(double)(j-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y,r,g,b);
     }
   }
 }
 else{
   grad=(double)idx/(double)idy; j=x;
   if(idx >= 0)error = -0.5; else error = 0.5;
   if(idy >= 0){
     for(i=y;i<y+idy;i++){
       error += grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(j-x),(double)(i-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y,r,g,b);
     }
   }
   else{
     for(i=y;i>y+idy;i--){
       error -= grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(j-x),(double)(i-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y,r,g,b);
     }
   }
 }
 return;
}

long _RenderImageProcess(char *PrmList, XIMAGE *lpXimage){
 int i,j;
 char dummy[255];
 double x,y,z,mr=0.0,r,g,b,d2r,a0;
 long xx,yy,radius,iradius,hide,miradius,type,id,param,c1,c2,c3,rotate,
      mparam,mc1,mc2,mc3,mhide,mtype,mid;
 float *Zb;
 BOOL  bDraw,bMorph=FALSE;
#include "pro_key.c"
 if(lpXimage->Nlights == 0 || lpXimage->Lights == NULL)return 1;
 sscanf(PrmList,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",dummy,&version,
        &iradius,&hide,&type,&id,&param,&c1,&c2,&c3,&rotate);
 if(lpXimage->Morph && lpXimage->mParameters != NULL){
   bMorph=TRUE; mr=lpXimage->MorphRatio;
   sscanf(lpXimage->mParameters,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld",
          dummy,&version,&miradius,
          &mhide,&mtype,&mid,&mparam,&mc1,&mc2,&mc3);
   iradius=(long)((double)miradius + mr*((double)iradius-(double)miradius));
   c1=(long)((double)mc1+mr*((double)c1-(double)mc1));
   c2=(long)((double)mc2+mr*((double)c2-(double)mc2));
   c3=(long)((double)mc3+mr*((double)c3-(double)mc3));
 }
 if(!bMorph){
   mr=(double)(lpXimage->last_frame - lpXimage->first_frame);
   if(mr < 1.0)return 1;
   mr=(double)(lpXimage->this_frame - lpXimage->first_frame)/mr;
 }
 d2r=3.1415626/180.0;
 a0= -95.0;
 if(rotate == 1)a0 += 360.0*mr;
 Zb=lpXimage->Zbuffer;
 for(i=0;i<lpXimage->Nlights;i++){
   if(type == 2 && lpXimage->Lights[i].type != DUMMYL)continue;
   if(type == 3 && lpXimage->Lights[i].AnimatorID != id)continue;
   x=lpXimage->Lights[i].p[0];
   y=lpXimage->Lights[i].p[1];
   z=lpXimage->Lights[i].p[2];
   if(y >= 1.000){
     xx=lpXimage->Xmax/2 +
       (long)(lpXimage->Xscale*x/y);
     yy=lpXimage->Ymax/2 -
       (long)(lpXimage->Yscale*z/y);
     if(xx >= 0 && xx < lpXimage->Xmax &&
        yy >= 0 && yy < lpXimage->Ymax){
       if(Zb != NULL){
         if(y < *(Zb + (yy * lpXimage->Xmax) + xx))bDraw=TRUE;
         else bDraw=FALSE;
         if(hide == 1 && !bDraw)continue;
       }
       else bDraw=TRUE;
       radius = (long)(lpXimage->Xscale*(double)iradius/y);
       r=(double)c1/255.0;
       g=(double)c2/255.0;
       b=(double)c3/255.0;
//       DrawInBuffer1(xx,yy,lpXimage->Screen,
//                     lpXimage->Xmax,lpXimage->Ymax,radius/5,
//                     r,g,b);
       for(j=0;j<param;j++){  /* spines */
         double a,l,ddx,ddy;
         long   xs,ys,idx,idy;
         a= a0+360.0/(double)param*(double)j;
         a *= d2r;
         l=radius;
         ddx=l*cos(a); ddy=l*sin(a);
         idx=(long)ddx; idy=(long)ddy;
         DrawFlareInBuffer(xx,yy,idx,idy,
                           lpXimage->Screen,
                           lpXimage->Xmax,
                           lpXimage->Ymax,
                           255.0,
                           r,g,b,
                           0.0,(double)radius*0.02,
                           y,Zb,bDraw,hide,TRUE);
         DrawFlareInBuffer(xx,yy,idx,idy,
                           lpXimage->Screen,
                           lpXimage->Xmax,
                           lpXimage->Ymax,
                           4.0,
                           r,g,b,
                           (double)radius*0.1,(double)radius*0.08,
                           y,Zb,bDraw,hide,TRUE);
         l=(double)radius*2.0/3.0;
         ddx=l*cos(a); ddy=l*sin(a);
         idx=(long)ddx; idy=(long)ddy;
         DrawFlareInBuffer(xx,yy,idx,idy,
                           lpXimage->Screen,
                           lpXimage->Xmax,
                           lpXimage->Ymax,
                           5.0,
                           r,g,b,
                           0.0,(double)radius*0.1,
                           y,Zb,bDraw,hide,TRUE);
         l=(double)radius/2.0;
         ddx=l*cos(a); ddy=l*sin(a);
         idx=(long)ddx; idy=(long)ddy;
         DrawFlareInBuffer(xx,yy,idx,idy,
                           lpXimage->Screen,
                           lpXimage->Xmax,
                           lpXimage->Ymax,
                           8.0,
                           r,g,b,
                           0.0,(double)radius*0.15,
                           y,Zb,bDraw,hide,TRUE);
         l=(double)radius/3.0;
         ddx=l*cos(a); ddy=l*sin(a);
         idx=(long)ddx; idy=(long)ddy;
         DrawFlareInBuffer(xx,yy,idx,idy,
                           lpXimage->Screen,
                           lpXimage->Xmax,
                           lpXimage->Ymax,
                           8.0,
                           r,g,b,
                           0.0,(double)radius*0.2,
                           y,Zb,bDraw,hide,TRUE);
       }
     }
   }
 }
 return 1;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static double rad=1.0;
static long   hide=1,type=1,id = -1,c1=255,c2=255,c3=0,param=5,rotate=0;
static X__MEMORY_MANAGER *lpLocalEVI;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
 char buffer[256];
 long irad;       /* integer scaled radius value                */
 if(Op != NULL){  /* parameters exist so read them off the list */
   sscanf(Op,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",buffer,&version,
          &irad,&hide,&type,&id,&param,&c1,&c2,&c3,&rotate);
   /* transform from World units to User units                  */
   rad=(double)irad/(double)ruler;
 }
 lpLocalEVI=lpEVI;
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_HILITE),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 irad=(long)(rad*(double)ruler);
 sprintf(buffer,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",name,version,
         irad,hide,type,id,param,c1,c2,c3,rotate);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"External effect: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 BOOL err;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.3f",rad);
     SetDlgItemText(hwnd,DLG_HILITE_RADIUS,str);
     SetDlgItemInt(hwnd,DLG_HILITE_PARAM,param,FALSE);
     if(rotate == 1)SendDlgItemMessage(hwnd,DLG_HILITE_ROTATE,BM_SETCHECK,TRUE,0);
     if(hide == 1)SendDlgItemMessage(hwnd,DLG_HILITE_HIDE,BM_SETCHECK,TRUE,0);
     if(type == 1)SendDlgItemMessage(hwnd,DLG_HILITE_TYPE1,BM_SETCHECK,TRUE,0);
     if(type == 2)SendDlgItemMessage(hwnd,DLG_HILITE_TYPE2,BM_SETCHECK,TRUE,0);
     if(type == 3)SendDlgItemMessage(hwnd,DLG_HILITE_TYPE3,BM_SETCHECK,TRUE,0);
     if(id >= 0)SendDlgItemMessage(hwnd,DLG_HILITE_NAME,WM_SETTEXT,
                0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id));
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hwnd);
     break;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_HILITE_COLOUR){
         r=(BYTE)c1; g=(BYTE)c2; b=(BYTE)c3;
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_HILITE_COLOUR:{
            unsigned char colour[3];
            colour[0]=(unsigned char)c1;
            colour[1]=(unsigned char)c2;
            colour[2]=(unsigned char)c3;
            SetColour(colour,hwnd);
            c1=(long)colour[0]; c2=(long)colour[1]; c3=(long)colour[2];
            InvalidateRect(GetDlgItem(hwnd,DLG_HILITE_COLOUR),NULL,FALSE);
          }
          break;
        case DLG_HILITE_SETID:
          id=GetActorsID(lpLocalEVI->lpAni,hwnd);
          if(id >= 0)SendDlgItemMessage(hwnd,DLG_HILITE_NAME,WM_SETTEXT,
                     0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id));
          break;
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          param=GetDlgItemInt(hwnd,DLG_HILITE_PARAM,&err,FALSE);
          if(GetDlgItemText(hwnd,DLG_HILITE_RADIUS,str,12) != 0)
            sscanf(str,"%f",&rad);
          if(SendDlgItemMessage(hwnd,DLG_HILITE_HIDE,BM_GETCHECK,0,0))
               hide=1;
          else hide=0;
          if(SendDlgItemMessage(hwnd,DLG_HILITE_ROTATE,BM_GETCHECK,0,0))
               rotate=1;
          else rotate=0;
          if(SendDlgItemMessage(hwnd,DLG_HILITE_TYPE1,BM_GETCHECK,0,0))type=1;
          if(SendDlgItemMessage(hwnd,DLG_HILITE_TYPE2,BM_GETCHECK,0,0))type=2;
          if(SendDlgItemMessage(hwnd,DLG_HILITE_TYPE3,BM_GETCHECK,0,0))type=3;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

