/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* lasers.c  image post-processor                                         */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <windows.h>
#include "struct.h"           /* general structures    */
#include "ximage.h"           /* XIMAGE structure      */
#include "local.h"

#define y1 myy1

#include "lasers.h"

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL; /* use to pick up resources from DLL   */
static long version=1;

#include "utils.h"
#include "paint.c"


#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BC__
BOOL WINAPI DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

static void SetColour(unsigned char *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(unsigned char)GetRValue(cc.rgbResult);
   colour[1]=(unsigned char)GetGValue(cc.rgbResult);
   colour[2]=(unsigned char)GetBValue(cc.rgbResult);
 }
}

static void DrawInBuffer1(long x, long y, fullscreenbuffer *Screen,
                          long X, long Y, long radius,
                          double r, double g, double b,
                          double z, float *Zb, BOOL bDraw){
 long i,j;
 double d;
 fullscreenbuffer *S;
 for(i=x-radius;i<=x+radius;i++)for(j=y-radius;j<=y+radius;j++){
   if(i < 0 || i >= X || j < 0 || j >= Y)continue;
   S=(Screen + j*X + i);
   if(radius < 1)d=1.0;
   else{
     d=sqrt((double)(i-x)*(double)(i-x) + (double)(j-y)*(double)(j-y));
     d=255.0*((double)radius-d)/(double)radius;
     d=max(0.0,d);
   }
   if(Zb == NULL || !bDraw || z < *(Zb + j*X + i)){
     S->R = (unsigned char)min(255.0,(double)S->R+d*r);
     S->G = (unsigned char)min(255.0,(double)S->G+d*g);
     S->B = (unsigned char)min(255.0,(double)S->B+d*b);
   }
 }
 return;
}

void DrawInBuffer(fullscreenbuffer *Screen, int Xmax, int Ymax,
                  int x, int y, double v,
                  double rr, double gg, double bb,
                  double z, float *Zb,BOOL bDraw){
 int r,g,b;
 fullscreenbuffer *S;
 if(x < 0 || y < 0 || x >= Xmax || y >= Ymax)return;
 if(Zb == NULL || !bDraw || z < *(Zb + y*Xmax + x)){
   S=(Screen+y*Xmax+x);
   r=(int)S->R+(int)((double)v*rr);
   g=(int)S->G+(int)((double)v*gg);
   b=(int)S->B+(int)((double)v*bb);
   S->R=(r>255) ? 255:r;
   S->G=(g>255) ? 255:g;
   S->B=(b>255) ? 255:b;
 }
}

#define frac(z) fmod(z,1.0)
#define imod(z,a) ((z) - ((z)/(a))*(a))

void DrawAApixel(double x, double y, double w, double h, double C,
                 fullscreenbuffer *S, int X, int Y,
                 double r, double g, double b,
                 double z, float *Zb, BOOL bDraw){
  int xa,xb,ya,yb,i,j,brite;
  double x2,y2,lfrac,rfrac,tfrac,bfrac,xfrac,yfrac;
  if(w > 3 || h > 3){
    DrawInBuffer1((long)x,(long)y,S,X,Y,(long)w,r,g,b,z,Zb,bDraw);
    return;
  }
  brite=(int)C;
  if(x < 0 || y < 0)return;
  x -= w*0.5; y -= h*0.5;
  x2 = x + w;
  y2 = y + w;
  xa = (int)x;
  xb = (int)x2;
  ya = (int)y;
  yb = (int)y2;
  lfrac = 1.0-frac(x);
  rfrac = frac(x2);
  tfrac = 1.0-frac(y);
  bfrac = frac(y2);
  if (xa==xb) {
    xfrac = lfrac+rfrac-1.0;
    if (ya==yb){
      DrawInBuffer(S,X,Y,xa,ya,xfrac*(tfrac+bfrac-1.0)*brite,
                   r,g,b,z,Zb,bDraw);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,xfrac*tfrac*brite,r,g,b,z,Zb,bDraw);
      for (j=ya+1; j<yb; j++)DrawInBuffer(S,X,Y,xa,j,xfrac*brite,
                                           r,g,b,z,Zb,bDraw);
      DrawInBuffer(S,X,Y,xa,yb,xfrac*bfrac*brite,r,g,b,z,Zb,bDraw);
    }
  }
  else {
    if (ya==yb) {
      yfrac = tfrac+bfrac-1.0;
      DrawInBuffer(S,X,Y,xa,ya,yfrac*lfrac*brite,r,g,b,z,Zb,bDraw);
      for (i=xa+1; i<xb; i++)DrawInBuffer(S,X,Y,i,ya,yfrac*brite,
                                           r,g,b,z,Zb,bDraw);
      DrawInBuffer(S,X,Y,xb,ya,yfrac*rfrac*brite,r,g,b,z,Zb,bDraw);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,tfrac*lfrac*brite,r,g,b,z,Zb,bDraw);
      for (i=xa+1; i<xb; i++)DrawInBuffer(S,X,Y,i,ya,tfrac*brite,
                             r,g,b,z,Zb,bDraw);
      DrawInBuffer(S,X,Y,xb,ya,tfrac*rfrac*brite,r,g,b,z,Zb,bDraw);
      for (j=ya+1; j<yb; j++) {
        DrawInBuffer(S,X,Y,xa,j,lfrac*brite,r,g,b,z,Zb,bDraw);
        for (i=xa+1; i<xb; i++)DrawInBuffer(S,X,Y,i,j,brite,
                                            r,g,b,z,Zb,bDraw);
        DrawInBuffer(S,X,Y,xb,j,rfrac*brite,r,g,b,z,Zb,bDraw);
      }
      DrawInBuffer(S,X,Y,xa,yb,bfrac*lfrac*brite,r,g,b,z,Zb,bDraw);
      for (i=xa+1; i<xb; i++)DrawInBuffer(S,X,Y,i,yb,bfrac*brite,
                                          r,g,b,z,Zb,bDraw);
      DrawInBuffer(S,X,Y,xb,yb,bfrac*rfrac*brite,r,g,b,z,Zb,bDraw);
    }
  }
}

#define DIST(i,j)  sqrt((double)i * (double)i + (double)j * (double)j)

#define DIST3(i,j,k)  sqrt((double)(i) * (double)(i) + \
                           (double)(j) * (double)(j) + \
                           (double)(k) * (double)(k) )

static double GetW(double l, double dx, double dy,
                   double Wmin, double Wmax){
 double d,f;
 d=DIST(dx,dy);
 f=d/l;
 return (Wmin+(Wmax-Wmin)*f);
}

static double GetZ(long i, long j , long x, long y, long xt, long yt,
                   double y1, double y2){
 double d;
 d=DIST((xt-x),(yt-y));
 if(d < 1.e-10)return y1;
 d=DIST((i-x),(j-y))/d;
 return y1+d*(y2-y1);
}

static void DrawLineInBuffer(long x, long y, long xt, long yt,
                             double x1, double y1, double z1,
                             double x2, double y2, double z2,
                             fullscreenbuffer *Screen,
                             long X, long Y,
                             double Cin,
                             double r, double g, double b,
                             double Wmin, double Wmax,
                             float *Zb, BOOL bDraw){
 long i,j,idx,idy;
 double z,grad,error,W,l,C;
 idx=xt-x; idy=yt-y;
 C=Cin;
 l=DIST(idx,idy);
 if(abs(idx) >= abs(idy)){
   grad=(double)idy/(double)idx; j=y;
   if(idy >= 0)error = -0.5; else error = 0.5;
   if(idx >= 0){
     for(i=x;i<x+idx;i++){
       error += grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax);
       z=GetZ(i,j,x,y,xt,yt,y1,y2);
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y,r,g,b,z,Zb,bDraw);
     }
   }
   else{
     for(i=x;i>x+idx;i--){
       error -= grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax);
       z=GetZ(i,j,x,y,xt,yt,y1,y2);
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y,r,g,b,z,Zb,bDraw);
     }
   }
 }
 else{
   grad=(double)idx/(double)idy; j=x;
   if(idx >= 0)error = -0.5; else error = 0.5;
   if(idy >= 0){
     for(i=y;i<y+idy;i++){
       error += grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax);
       z=GetZ(j,i,x,y,xt,yt,y1,y2);
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y,r,g,b,z,Zb,bDraw);
     }
   }
   else{
     for(i=y;i>y+idy;i--){
       error -= grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax);
       z=GetZ(j,i,x,y,xt,yt,y1,y2);
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y,r,g,b,z,Zb,bDraw);
     }
   }
 }
 return;
}

static double GetYint(double x, double y, double dx, double dy, double xe){
 double grad;
 grad=dy/dx;
 return grad*(xe-x)+y;
}

static double GetXint(double x, double y, double dx, double dy, double ye){
 double grad;
 grad=dx/dy;
 return grad*(ye-y)+x;
}

static void GetBorderPosition(long xc, long yc,
                              long *x, long *y,
                              long Xmax, long Ymax){
 long dx,dy;
 double ddx,ddy,dxc,dyc,ye,xe,ds,de,fr;
 dx = *x - xc;
 dy = *y - yc;
 if(dx == 0 && dy == 0)dx=1;
 ddx=(double)dx; ddy=(double)dy;
 dxc=(double)xc; dyc=(double)yc;
 if(dx == 0){
   xe=xc;
   if(dy > 0)ye=(double)Ymax;
   else      ye=0.0;
 }
 else if(dy == 0){
   ye=yc;
   if(dx < 0)xe=0.0;
   else      xe=(double)Xmax;
 }
 else {
   if(dx > 0)xe=(double)Xmax;
   else      xe=0.0;
   ye=GetYint(dxc,dyc,ddx,ddy,xe);
   if(ye < 0){
     xe=GetXint(dxc,dyc,ddx,ddy,0);
     ye=0.0;
   }
   else if(ye > Ymax){
     xe=GetXint(dxc,dyc,ddx,ddy,(double)Ymax);
     ye=(double)Ymax;
   }
 }
 *x=(long)xe;
 *y=(long)ye;
 return;
}

static BOOL ClipView(double *x1, double *y1, double *z1,
                     double *x2, double *y2, double *z2){
 double mu;
 if(*y1 >= 0.0 && *y2 >= 0.0)return TRUE;
 if(*y1 <= 0.0 && *y2 <= 0.0)return FALSE;
 if(*y1 < 0.0){
   mu = (*y1 - *y2) / *y1;
   *y1 = 0.1;
   *x1 = *x1 + mu*(*x2 - *x1);
   *z1 = *z1 + mu*(*z2 - *z1);
 }
 else{
   mu = (*y2 - *y1) / *y2;
   *y2 = 0.1;
   *x2 = *x2 + mu*(*x1 - *x2);
   *z2 = *z2 + mu*(*z1 - *z2);
 }
 return TRUE;
}

static BOOL ClipScreen(long *x1, long *y1,
                       long *x2, long *y2,
                       long xmin, long xmax,
                       long ymin, long ymax){
 if((*x1 < 0 || *x1 >= xmax) &&
    (*y1 < 0 || *y1 >= ymax) &&
    (*x2 < 0 || *x2 >= xmax) &&
    (*y2 < 0 || *y2 >= ymax))return FALSE;  // both outside
 if((*x1 >= 0 && *x1 < xmax) &&
    (*y1 >= 0 && *y1 < ymax) &&
    (*x2 >= 0 && *x2 < xmax) &&
    (*y2 >= 0 && *y2 < ymax))return TRUE;  // both inside
 if((*x1 >= 0 && *x1 < xmax) &&
    (*y1 >= 0 && *y1 < ymax)){
   GetBorderPosition(*x1,*y1,x2,y2,xmax,ymax);
 }
 else{
   GetBorderPosition(*x2,*y2,x1,y1,xmax,ymax);
 }
 return TRUE;
}

long _RenderImageProcess(char *PrmList, XIMAGE *lpXimage){
 int i,j;
 char buffer[255];
 double x1,y1,z1,x2,y2,z2,mr=0.0,r,g,b,rad,dx,dy,dz;
 long xx1,yy1,xx2,yy2,dframe;
 long irad,hide,hold,type,c1,c2,c3,type1,id1,ix1,iy1,iz1,type2,id2,ix2,iy2,iz2,
      mirad,mhide,mhold,mtype,mc1,mc2,mc3,mtype1,mid1,mix1,miy1,miz1,
      mtype2,mid2,mix2,miy2,miz2;
 float *Zb;
 BOOL  bDraw,bMorph=FALSE;
#include "pro_key.c"
 sscanf(PrmList,
  "%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
  buffer,&version,&irad,
  &hide,&hold,&type,&c1,&c2,&c3,
  &type1,&id1,&ix1,&iy1,&iz1,
  &type2,&id2,&ix2,&iy2,&iz2);
 if(lpXimage->Morph && lpXimage->mParameters != NULL){
   bMorph=TRUE; mr=lpXimage->MorphRatio;
   sscanf(PrmList,
    "%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
    buffer,&version,&mirad,
    &mhide,&mhold,&mtype,&mc1,&mc2,&mc3,
    &mtype1,&mid1,&mix1,&miy1,&miz1,
    &mtype2,&mid2,&mix2,&miy2,&miz2);
   irad=(long)((double)mirad + mr*((double)irad-(double)mirad));
   c1=(long)((double)mc1+mr*((double)c1-(double)mc1));
   c2=(long)((double)mc2+mr*((double)c2-(double)mc2));
   c3=(long)((double)mc3+mr*((double)c3-(double)mc3));
   ix1=(long)((double)mix1+mr*((double)ix1-(double)mix1));
   iy1=(long)((double)miy1+mr*((double)iy1-(double)miy1));
   iz1=(long)((double)miz1+mr*((double)iz1-(double)miz1));
   ix2=(long)((double)mix2+mr*((double)ix2-(double)mix2));
   iy2=(long)((double)miy2+mr*((double)iy2-(double)miy2));
   iz2=(long)((double)miz2+mr*((double)iz2-(double)miz2));
 }
 if(!bMorph){
   mr=(double)(lpXimage->last_frame - lpXimage->first_frame);
   if(mr < 1.0)return 1;
   mr=(double)(lpXimage->this_frame - lpXimage->first_frame)/mr;
 }
 dframe=lpXimage->this_frame - lpXimage->first_frame;
 r=(double)c1/255.0;
 g=(double)c2/255.0;
 b=(double)c3/255.0;
 Zb=lpXimage->Zbuffer;
 if(hide && Zb != NULL)bDraw=TRUE;
 else                  bDraw=FALSE;
 rad = (double)irad*(double)lpXimage->Xmax/320.0;
 if(type1 == 1 || lpXimage->Nlights == 0)
   TransformIntoView(lpXimage->ViewTransform,
                     (double)ix1,(double)iy1,(double)iz1,&x1,&y1,&z1);
 else for(i=0;i<lpXimage->Nlights;i++){
   if(lpXimage->Lights[i].AnimatorID != id1)continue;
   x1=lpXimage->Lights[i].p[0];
   y1=lpXimage->Lights[i].p[1];
   z1=lpXimage->Lights[i].p[2];
   break;
 }
 if(type2 == 1 || lpXimage->Nlights == 0)
   TransformIntoView(lpXimage->ViewTransform,
                     (double)ix2,(double)iy2,(double)iz2,&x2,&y2,&z2);
 else for(i=0;i<lpXimage->Nlights;i++){
   if(lpXimage->Lights[i].AnimatorID != id2)continue;
   x2=lpXimage->Lights[i].p[0];
   y2=lpXimage->Lights[i].p[1];
   z2=lpXimage->Lights[i].p[2];
   break;
 }
 if(!ClipView(&x1,&y1,&z1,&x2,&y2,&z2))return 1;
 if(type == 2){
   double f=0.1,l,px,py,pz,px1,py1,pz1,px2,py2,pz2;
   dx=x2-x1; dy=y2-y1; dz=z2-z1;
   px=x1+mr*dx; py=y1+mr*dy; pz=z1+mr*dz;
   l=DIST3(dx,dy,dz);
   px1=px-f*dx; py1=py-f*dy; pz1=pz-f*dz;
   if(DIST3((x1-px1),(y1-py1),(z1-pz1)) < l*f){px1=x1; py1=y1; pz1=z1;}
   px2=px+f*dx; py2=py+f*dy; pz2=pz+f*dz;
   if(DIST3((x2-px2),(y2-py2),(z2-pz2)) < l*f){px2=x2; py2=y2; pz2=z2;}
   x1=px1; y1=py1; z1=pz1;
   x2=px2; y2=py2; z2=pz2;
 }
 else if(hold == 1){
   dx=x2-x1; dy=y2-y1; dz=z2-z1;
   x2=x1+mr*dx; y2=y1+mr*dy; z2=z1+mr*dz;
 }
 else if(hold == 3){
   dx=x2-x1; dy=y2-y1; dz=z2-z1;
   x1=x1+mr*dx; y1=y1+mr*dy; z1=z1+mr*dz;
 }
 if(type == 3){
   long i,istart,n=30,ninc=3;
   double px1,py1,pz1,px2,py2,pz2;
   dx=x2-x1; dy=y2-y1; dz=z2-z1;
   dx /= (double)n; dy /= (double)n; dz /= (double)n;
   px1=x1; py1=y1; pz1=z1;
   istart=imod(dframe,ninc);
   for(i=istart;i<n;i+=ninc){
     px1=x1+(double)i*dx; px2=px1+dx;
     py1=y1+(double)i*dy; py2=py1+dy;
     pz1=z1+(double)i*dz; pz2=pz1+dz;
     xx1=lpXimage->Xmax/2 + (long)(lpXimage->Xscale*px1/py1);
     yy1=lpXimage->Ymax/2 - (long)(lpXimage->Yscale*pz1/py1);
     xx2=lpXimage->Xmax/2 + (long)(lpXimage->Xscale*px2/py2);
     yy2=lpXimage->Ymax/2 - (long)(lpXimage->Yscale*pz2/py2);
     if(!ClipScreen(&xx1,&yy1,&xx2,&yy2,
                  0,lpXimage->Xmax,0,lpXimage->Ymax))return 1;
     DrawLineInBuffer(xx1,yy1,xx2,yy2,
                      px1,py1,pz1,
                      px2,py2,pz2,
                      lpXimage->Screen,
                      lpXimage->Xmax,lpXimage->Ymax,
                      255.0,r,g,b,
                      (double)rad,(double)rad,
                      Zb,bDraw);
   }
 }
 else{
   xx1=lpXimage->Xmax/2 + (long)(lpXimage->Xscale*x1/y1);
   yy1=lpXimage->Ymax/2 - (long)(lpXimage->Yscale*z1/y1);
   xx2=lpXimage->Xmax/2 + (long)(lpXimage->Xscale*x2/y2);
   yy2=lpXimage->Ymax/2 - (long)(lpXimage->Yscale*z2/y2);
   if(!ClipScreen(&xx1,&yy1,&xx2,&yy2,
                  0,lpXimage->Xmax,0,lpXimage->Ymax))return 1;
   DrawLineInBuffer(xx1,yy1,xx2,yy2,
                    x1,y1,z1,
                    x2,y2,z2,
                    lpXimage->Screen,
                    lpXimage->Xmax,lpXimage->Ymax,
                    255.0,r,g,b,
                    (double)rad,(double)rad,
                    Zb,bDraw);
 }
 return 1;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static double x1=0.0,y1=0.0,z1=0.0,x2=0.0,y2=1.0,z2=0.0;
static long   irad=2,hide=1,hold=2,type=1,id1 = -1, id2 = -1,type1=1,type2=1,
              c1=255,c2=0,c3=0;
static X__MEMORY_MANAGER *lpLocalEVI;


char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
 char buffer[512];
 long ix1,iy1,iz1,ix2,iy2,iz2;
 if(Op != NULL){
   sscanf(Op,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
          buffer,&version,&irad,
          &hide,&hold,&type,&c1,&c2,&c3,
          &type1,&id1,&ix1,&iy1,&iz1,
          &type2,&id2,&ix2,&iy2,&iz2);
   x1=(double)ix1/(double)ruler;
   y1=(double)iy1/(double)ruler;
   z1=(double)iz1/(double)ruler;
   x2=(double)ix2/(double)ruler;
   y2=(double)iy2/(double)ruler;
   z2=(double)iz2/(double)ruler;
 }
 lpLocalEVI=lpEVI;
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_LASER),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 ix1=(long)(x1*(double)ruler);
 iy1=(long)(y1*(double)ruler);
 iz1=(long)(z1*(double)ruler);
 ix2=(long)(x2*(double)ruler);
 iy2=(long)(y2*(double)ruler);
 iz2=(long)(z2*(double)ruler);
 sprintf(buffer,
         "%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
         name,version,irad,
         hide,hold,type,c1,c2,c3,
         type1,id1,ix1,iy1,iz1,
         type2,id2,ix2,iy2,iz2);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"External effect: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 BOOL err;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     SetDlgItemInt(hwnd,DLG_LASER_RADIUS,irad,TRUE);
     if(hide == 1)SendDlgItemMessage(hwnd,DLG_LASER_HIDE,BM_SETCHECK,TRUE,0);
     if(type == 1)SendDlgItemMessage(hwnd,DLG_LASER_TYPE1,BM_SETCHECK,TRUE,0);
     if(type == 2)SendDlgItemMessage(hwnd,DLG_LASER_TYPE2,BM_SETCHECK,TRUE,0);
     if(type == 3)SendDlgItemMessage(hwnd,DLG_LASER_TYPE3,BM_SETCHECK,TRUE,0);
     if(hold == 1)SendDlgItemMessage(hwnd,DLG_LASER_IN,BM_SETCHECK,TRUE,0);
     if(hold == 2)SendDlgItemMessage(hwnd,DLG_LASER_HOLD,BM_SETCHECK,TRUE,0);
     if(hold == 3)SendDlgItemMessage(hwnd,DLG_LASER_OUT,BM_SETCHECK,TRUE,0);
     if(type1 == 1)SendDlgItemMessage(hwnd,DLG_LASER_1_TYPE1,BM_SETCHECK,TRUE,0);
     if(type1 == 2)SendDlgItemMessage(hwnd,DLG_LASER_1_TYPE2,BM_SETCHECK,TRUE,0);
     if(id1 >= 0)SendDlgItemMessage(hwnd,DLG_LASER_1_NAME,WM_SETTEXT,
                0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id1));
     sprintf(str,"%.3f",x1); SetDlgItemText(hwnd,DLG_LASER_1_X,str);
     sprintf(str,"%.3f",y1); SetDlgItemText(hwnd,DLG_LASER_1_Y,str);
     sprintf(str,"%.3f",z1); SetDlgItemText(hwnd,DLG_LASER_1_Z,str);
     if(type2 == 1)SendDlgItemMessage(hwnd,DLG_LASER_2_TYPE1,BM_SETCHECK,TRUE,0);
     if(type2 == 2)SendDlgItemMessage(hwnd,DLG_LASER_2_TYPE2,BM_SETCHECK,TRUE,0);
     if(id2 >= 0)SendDlgItemMessage(hwnd,DLG_LASER_2_NAME,WM_SETTEXT,
                0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id2));
     sprintf(str,"%.3f",x2); SetDlgItemText(hwnd,DLG_LASER_2_X,str);
     sprintf(str,"%.3f",y2); SetDlgItemText(hwnd,DLG_LASER_2_Y,str);
     sprintf(str,"%.3f",z2); SetDlgItemText(hwnd,DLG_LASER_2_Z,str);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hwnd);
     break;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_LASER_COLOUR){
         r=(BYTE)c1; g=(BYTE)c2; b=(BYTE)c3;
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_LASER_COLOUR:{
            unsigned char colour[3];
            colour[0]=(unsigned char)c1;
            colour[1]=(unsigned char)c2;
            colour[2]=(unsigned char)c3;
            SetColour(colour,hwnd);
            c1=(long)colour[0]; c2=(long)colour[1]; c3=(long)colour[2];
            InvalidateRect(GetDlgItem(hwnd,DLG_LASER_COLOUR),NULL,FALSE);
          }
          break;
        case DLG_LASER_1_SETID:
          id1=GetActorsID(lpLocalEVI->lpAni,hwnd);
          if(id1 >= 0)SendDlgItemMessage(hwnd,DLG_LASER_1_NAME,WM_SETTEXT,
                       0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id1));
          break;
        case DLG_LASER_2_SETID:
          id2=GetActorsID(lpLocalEVI->lpAni,hwnd);
          if(id2 >= 0)SendDlgItemMessage(hwnd,DLG_LASER_2_NAME,WM_SETTEXT,
                       0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id2));
          break;
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          irad=GetDlgItemInt(hwnd,DLG_LASER_RADIUS,&err,FALSE);
          if(SendDlgItemMessage(hwnd,DLG_LASER_HIDE,BM_GETCHECK,0,0))
               hide=1;
          else hide=0;
          if(SendDlgItemMessage(hwnd,DLG_LASER_TYPE1,BM_GETCHECK,0,0))type=1;
          if(SendDlgItemMessage(hwnd,DLG_LASER_TYPE2,BM_GETCHECK,0,0))type=2;
          if(SendDlgItemMessage(hwnd,DLG_LASER_TYPE3,BM_GETCHECK,0,0))type=3;
          if(SendDlgItemMessage(hwnd,DLG_LASER_IN,BM_GETCHECK,0,0))hold=1;
          if(SendDlgItemMessage(hwnd,DLG_LASER_HOLD,BM_GETCHECK,0,0))hold=2;
          if(SendDlgItemMessage(hwnd,DLG_LASER_OUT,BM_GETCHECK,0,0))hold=3;
          if(SendDlgItemMessage(hwnd,DLG_LASER_1_TYPE1,BM_GETCHECK,0,0))
            type1=1;
          if(SendDlgItemMessage(hwnd,DLG_LASER_1_TYPE2,BM_GETCHECK,0,0))
            type1=2;
          if(GetDlgItemText(hwnd,DLG_LASER_1_X,str,12) != 0)
            sscanf(str,"%f",&x1);
          if(GetDlgItemText(hwnd,DLG_LASER_1_Y,str,12) != 0)
            sscanf(str,"%f",&y1);
          if(GetDlgItemText(hwnd,DLG_LASER_1_Z,str,12) != 0)
            sscanf(str,"%f",&z1);
          if(GetDlgItemText(hwnd,DLG_LASER_2_X,str,12) != 0)
            sscanf(str,"%f",&x2);
          if(GetDlgItemText(hwnd,DLG_LASER_2_Y,str,12) != 0)
            sscanf(str,"%f",&y2);
          if(GetDlgItemText(hwnd,DLG_LASER_2_Z,str,12) != 0)
            sscanf(str,"%f",&z2);
          if(SendDlgItemMessage(hwnd,DLG_LASER_2_TYPE1,BM_GETCHECK,0,0))
            type2=1;
          if(SendDlgItemMessage(hwnd,DLG_LASER_2_TYPE2,BM_GETCHECK,0,0))
            type2=2;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

