/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* lens.c  image post-processor                                         */
/*                                                                       */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include "struct.h"           /* general structures    */
#include "ximage.h"           /* XIMAGE structure      */
#include "local.h"

//FILE *debug;

#include "lens.h"

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL; /* use to pick up resources from DLL   */

#include "utils.h"

#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BC__
BOOL WINAPI DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

#include "lens1.c"

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static long   version=1,type=1,model=1,flicker=0,hide=1,id= -1,
              sx=0,sy=0,ipx=0,ipy=0,ipz=0;
static double bscale=1.0,rscale=1.0,angle=22.5;
static double px,py,pz;
static X__MEMORY_MANAGER *lpLocalEVI;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
 char buffer[256];
 if(Op != NULL){
   sscanf(Op,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %f %f %ld %f",
          buffer,&type,&flicker,&hide,&sx,&sy,&ipx,&ipy,&ipz,&model,&version,
          &bscale,&rscale,&id,&angle);
    px=(double)ipx/(double)ruler;
    py=(double)ipy/(double)ruler;
    pz=(double)ipz/(double)ruler;
 }
 lpLocalEVI=lpEVI;
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_LENS),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 ipx=(long)(px*(double)ruler);
 ipy=(long)(py*(double)ruler);
 ipz=(long)(pz*(double)ruler);
 sprintf(buffer,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %f %f %ld %f",
         name,type,flicker,hide,sx,sy,ipx,ipy,ipz,model,version,
         bscale,rscale,id,angle);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

#define D_BORDER 4
#define N_TABS 6

typedef struct tag_dlghdr1 {
  HWND hwndDlg[N_TABS];
  RECT rcTab;
} DLGHDR1;

static DLGHDR1 DlgHdrLens,*lpDlgHdrLens;
static HWND    hwndTabLens;
static HDRAWDIB hDrawDib = NULL;
static long BitmapSizeX,BitmapSizeY;
static LPSTR lpBitmapBits=NULL,lpBitmapHeader=NULL,lpBits;

static BOOL InitiatePreviewUpdateView(HWND);
static void TerminatePreviewUpdateView(void);
static void PaintPreviewUpdateView(HDC, RECT);
static void MakePreview(HWND);

BOOL CALLBACK Flare6DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          break;
        case IDOK:
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

BOOL CALLBACK Flare5DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          break;
        case IDOK:
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

BOOL CALLBACK Flare4DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          break;
        case IDOK:
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

BOOL CALLBACK Flare3DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          break;
        case IDOK:
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

BOOL CALLBACK Flare2DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     if     (model == 1)
       SendDlgItemMessage(hwnd,DLG_FLARE_MODEL1,BM_SETCHECK,TRUE,0);
     else if(model == 2)
       SendDlgItemMessage(hwnd,DLG_FLARE_MODEL2,BM_SETCHECK,TRUE,0);
     else if(model == 3)
       SendDlgItemMessage(hwnd,DLG_FLARE_MODEL3,BM_SETCHECK,TRUE,0);
     else if(model == 4)
       SendDlgItemMessage(hwnd,DLG_FLARE_MODEL4,BM_SETCHECK,TRUE,0);
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          break;
        case IDOK:
          if     (SendDlgItemMessage(hwnd,DLG_FLARE_MODEL4,BM_GETCHECK,0,0))
               model=4;
          else if(SendDlgItemMessage(hwnd,DLG_FLARE_MODEL3,BM_GETCHECK,0,0))
               model=3;
          else if(SendDlgItemMessage(hwnd,DLG_FLARE_MODEL2,BM_GETCHECK,0,0))
               model=2;
          else model=1;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

BOOL CALLBACK Flare1DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 BOOL err;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     if(hide)SendDlgItemMessage(hwnd,DLG_FLARE_HIDE,BM_SETCHECK,TRUE,0);
     if     (type == 1)
       SendDlgItemMessage(hwnd,DLG_FLARE_TYPE1,BM_SETCHECK,TRUE,0);
     else if(type == 2)
       SendDlgItemMessage(hwnd,DLG_FLARE_TYPE2,BM_SETCHECK,TRUE,0);
     else if(type == 3)
       SendDlgItemMessage(hwnd,DLG_FLARE_TYPE3,BM_SETCHECK,TRUE,0);
     else if(type == 4)
       SendDlgItemMessage(hwnd,DLG_FLARE_TYPE4,BM_SETCHECK,TRUE,0);
     else if(type == 5)
       SendDlgItemMessage(hwnd,DLG_FLARE_TYPE5,BM_SETCHECK,TRUE,0);
     sprintf(str,"%.3f",px);
     SetDlgItemText(hwnd,DLG_FLARE_PX,str);
     sprintf(str,"%.3f",py);
     SetDlgItemText(hwnd,DLG_FLARE_PY,str);
     sprintf(str,"%.3f",pz);
     SetDlgItemText(hwnd,DLG_FLARE_PZ,str);
     SetDlgItemInt(hwnd,DLG_FLARE_SX,sx,FALSE);
     SetDlgItemInt(hwnd,DLG_FLARE_SY,sy,FALSE);
     sprintf(str,"%.3f",bscale);
     SetDlgItemText(hwnd,DLG_FLARE_BSCALE,str);
     sprintf(str,"%.3f",rscale);
     SetDlgItemText(hwnd,DLG_FLARE_RSCALE,str);
     if(id >= 0)SendDlgItemMessage(hwnd,DLG_FLARE_NAME,WM_SETTEXT,
                0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id));
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_FLARE_SETID:
          id=GetActorsID(lpLocalEVI->lpAni,hwnd);
          if(id >= 0)SendDlgItemMessage(hwnd,DLG_FLARE_NAME,WM_SETTEXT,
                     0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id));
          break;
        case IDCANCEL:
          break;
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_FLARE_PX,str,12) != 0)
            sscanf(str,"%f",&px);
          if(GetDlgItemText(hwnd,DLG_FLARE_PY,str,12) != 0)
            sscanf(str,"%f",&py);
          if(GetDlgItemText(hwnd,DLG_FLARE_PZ,str,12) != 0)
            sscanf(str,"%f",&pz);
          if(GetDlgItemText(hwnd,DLG_FLARE_BSCALE,str,12) != 0)
            sscanf(str,"%f",&bscale);
          if(GetDlgItemText(hwnd,DLG_FLARE_RSCALE,str,12) != 0)
            sscanf(str,"%f",&rscale);
          sx=GetDlgItemInt(hwnd,DLG_FLARE_SX,&err,FALSE);
          sy=GetDlgItemInt(hwnd,DLG_FLARE_SY,&err,FALSE);
          if     (SendDlgItemMessage(hwnd,DLG_FLARE_TYPE5,BM_GETCHECK,0,0))
               type=5;
          else if(SendDlgItemMessage(hwnd,DLG_FLARE_TYPE4,BM_GETCHECK,0,0))
               type=4;
          else if(SendDlgItemMessage(hwnd,DLG_FLARE_TYPE3,BM_GETCHECK,0,0))
               type=3;
          else if(SendDlgItemMessage(hwnd,DLG_FLARE_TYPE2,BM_GETCHECK,0,0))
               type=2;
          else type=1;
          if(SendDlgItemMessage(hwnd,DLG_FLARE_HIDE,BM_GETCHECK,0,0))
               hide=1;
          else hide=0;
          break;
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

BOOL CALLBACK DlgProc(HWND hdlg,UINT msg,
                      WPARAM wparam,LPARAM lparam){
 int i,xp,yp;
 RECT rc,rd,re;
 char str[32];
 switch( msg ){
   case WM_INITDIALOG:{
       TC_ITEM tie;
       int h,w;
       DWORD dwDlgBase=GetDialogBaseUnits();
       lpDlgHdrLens=&DlgHdrLens;
       GetWindowRect(hdlg,&rc);
       GetWindowRect(GetDlgItem(hdlg,IDOK),&rd);
       GetWindowRect(GetDlgItem(hdlg,DLG_LENS_PREVIEW_W),&re);
       h=rc.bottom-rc.top-GetSystemMetrics(SM_CYCAPTION);
       w=re.left-rc.left;
       GetClientRect(hdlg,&rc);
       hwndTabLens=CreateWindowEx(
                            0,
                            WC_TABCONTROL," ",
                            TCS_MULTILINE |//TCS_FOCUSNEVER |
                            WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
                            (D_BORDER*LOWORD(dwDlgBase))/4,
                            (D_BORDER*HIWORD(dwDlgBase))/8,
                            w-(4*D_BORDER*LOWORD(dwDlgBase))/4,
                            h-(4*D_BORDER*HIWORD(dwDlgBase))/8,
                            hdlg,NULL,hDLLinstance,NULL);
       SendMessage(hwndTabLens,WM_SETFONT, /* for text in tab */
                  (WPARAM)GetStockObject(ANSI_VAR_FONT),0);
       memset(&tie,0,sizeof(TC_ITEM));
       tie.mask = TCIF_TEXT;// | TCIF_IMAGE;
       tie.pszText=str;
       LoadString(hDLLinstance,IDI_TEXT1,str,32);
       TabCtrl_InsertItem(hwndTabLens,0,&tie);
       LoadString(hDLLinstance,IDI_TEXT2,str,32);
       TabCtrl_InsertItem(hwndTabLens,1,&tie);
       LoadString(hDLLinstance,IDI_TEXT3,str,32);
       TabCtrl_InsertItem(hwndTabLens,2,&tie);
       LoadString(hDLLinstance,IDI_TEXT4,str,32);
       TabCtrl_InsertItem(hwndTabLens,3,&tie);
       LoadString(hDLLinstance,IDI_TEXT5,str,32);
       TabCtrl_InsertItem(hwndTabLens,4,&tie);
       LoadString(hDLLinstance,IDI_TEXT6,str,32);
       TabCtrl_InsertItem(hwndTabLens,6,&tie);
       SetRectEmpty(&rc);
       TabCtrl_AdjustRect(hwndTabLens,FALSE,&rc);
       rc.left += (D_BORDER*LOWORD(dwDlgBase))/4;
       rc.top += (D_BORDER*HIWORD(dwDlgBase))/8;
       CopyRect(&(lpDlgHdrLens->rcTab),&rc);
       for(i=0;i<N_TABS;i++)lpDlgHdrLens->hwndDlg[i]=NULL;
       lpDlgHdrLens->hwndDlg[0]=CreateDialog(hDLLinstance,
                                   MAKEINTRESOURCE(DLG_FLARE),
                                   hdlg,(DLGPROC)Flare1DlgProc);
       SetWindowPos(lpDlgHdrLens->hwndDlg[0],NULL,
                    lpDlgHdrLens->rcTab.left,
                    lpDlgHdrLens->rcTab.top,0,0,
                    SWP_NOSIZE | SWP_NOZORDER);
       ShowWindow(lpDlgHdrLens->hwndDlg[0],SW_SHOW);
     }
     InitiatePreviewUpdateView(GetDlgItem(hdlg,DLG_LENS_PREVIEW_W));
     CentreDialogOnScreen(hdlg);
     return TRUE;
   case WM_NOTIFY:
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGING){
       int id=TabCtrl_GetCurSel(hwndTabLens);
     }
     if( ((LPNMHDR)lparam)->code == TCN_SELCHANGE){
       int id=TabCtrl_GetCurSel(hwndTabLens);
       for(i=0;i<N_TABS;i++){
         if(lpDlgHdrLens->hwndDlg[i] != NULL){
           SendMessage(lpDlgHdrLens->hwndDlg[i],
                       WM_COMMAND,(WPARAM)IDOK,0);
           DestroyWindow(lpDlgHdrLens->hwndDlg[i]);
           lpDlgHdrLens->hwndDlg[i]=NULL;
         }
       }
       if(id == 0)lpDlgHdrLens->hwndDlg[0]=CreateDialog(hDLLinstance,
                                     MAKEINTRESOURCE(DLG_FLARE),
                                     hdlg,(DLGPROC)Flare1DlgProc);
       if(id == 1)lpDlgHdrLens->hwndDlg[1]=CreateDialog(hDLLinstance,
                                     MAKEINTRESOURCE(DLG_FLARE2),
                                     hdlg,(DLGPROC)Flare2DlgProc);
       if(id == 2)lpDlgHdrLens->hwndDlg[2]=CreateDialog(hDLLinstance,
                                     MAKEINTRESOURCE(DLG_FLARE3),
                                     hdlg,(DLGPROC)Flare3DlgProc);
       if(id == 3)lpDlgHdrLens->hwndDlg[3]=CreateDialog(hDLLinstance,
                                     MAKEINTRESOURCE(DLG_FLARE4),
                                     hdlg,(DLGPROC)Flare4DlgProc);
       if(id == 4)lpDlgHdrLens->hwndDlg[4]=CreateDialog(hDLLinstance,
                                     MAKEINTRESOURCE(DLG_FLARE5),
                                     hdlg,(DLGPROC)Flare5DlgProc);
       if(id == 4)lpDlgHdrLens->hwndDlg[5]=CreateDialog(hDLLinstance,
                                     MAKEINTRESOURCE(DLG_FLARE6),
                                     hdlg,(DLGPROC)Flare6DlgProc);
       if(lpDlgHdrLens->hwndDlg[id] != NULL){
         SetWindowPos(lpDlgHdrLens->hwndDlg[id],NULL,
                      lpDlgHdrLens->rcTab.left,
                      lpDlgHdrLens->rcTab.top,0,0,
                      SWP_NOSIZE | SWP_NOZORDER);
         ShowWindow(lpDlgHdrLens->hwndDlg[id],SW_SHOW);
       }
     }
     break;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       RECT rc;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_LENS_PREVIEW_W){
         if(lpdis->itemAction == ODA_DRAWENTIRE){
           PaintPreviewUpdateView(lpdis->hDC,lpdis->rcItem);
         }
       }
     }
     return TRUE;
   case WM_PAINT:
     PaintBackground(hdlg);
     break;
   case WM_COMMAND:
     switch( LOWORD(wparam) ){
       case IDOK:
         for(i=0;i<N_TABS;i++){
           if(lpDlgHdrLens->hwndDlg[i] != NULL)
            SendMessage(lpDlgHdrLens->hwndDlg[i],WM_COMMAND,
                        (WPARAM)IDOK,0);
         }
         EndDialog(hdlg,TRUE);
         return(TRUE);
       case IDCANCEL:
         EndDialog(hdlg,FALSE);
         return(TRUE);
       case DLG_LENS_PREVIEW:
         for(i=0;i<N_TABS;i++){
           if(lpDlgHdrLens->hwndDlg[i] != NULL)
            SendMessage(lpDlgHdrLens->hwndDlg[i],WM_COMMAND,
                        (WPARAM)IDOK,0);
         }
         MakePreview(GetDlgItem(hdlg,DLG_LENS_PREVIEW_W));
         break;
       default:
         break;
     }
     break;
   case WM_DESTROY:
     for(i=0;i<N_TABS;i++){
       if(lpDlgHdrLens->hwndDlg[i] != NULL)
         DestroyWindow(lpDlgHdrLens->hwndDlg[i]);
       lpDlgHdrLens->hwndDlg[i]=NULL;
     }
     TerminatePreviewUpdateView();
     break;
   default: break;
 }
 return( FALSE );
}

static BOOL InitiatePreviewUpdateView(HWND hwnd_control){
 long imagesize;
 int i;
 RECT rc;
 LPBITMAPINFOHEADER lpbi;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER));
 if(lpBitmapHeader == NULL){
   return FALSE;
 }
 GetClientRect(hwnd_control,&rc);
 BitmapSizeX=max(10,rc.right);
 BitmapSizeX=((BitmapSizeX+3)/4*4);
 BitmapSizeY=max(10,rc.bottom);
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)BitmapSizeX;   /* start with it as big as possible */
 lpbi->biHeight=(DWORD)BitmapSizeY;
 lpbi->biPlanes=1;
 lpbi->biBitCount=24;
 lpbi->biCompression=BI_RGB;
 imagesize = lpbi->biSizeImage = BitmapSizeX*BitmapSizeY*3L;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=0;
 lpbi->biClrImportant=0;
 lpBitmapBits=(LPSTR)X__Malloc(imagesize);
 if(lpBitmapBits == NULL){
   if(lpBitmapHeader != NULL)X__Free((char *)lpBitmapHeader);
   lpBitmapHeader = NULL;
   return FALSE;
 }
 memset(lpBitmapBits,0,imagesize);
 if((hDrawDib=DrawDibOpen()) == NULL)MessageBox(NULL,"No Dib",NULL,MB_OK);
 return TRUE;
}

static void TerminatePreviewUpdateView(void){
 if(hDrawDib != NULL)DrawDibClose(hDrawDib);
 if(lpBitmapBits   != NULL)X__Free((char *)lpBitmapBits);
 if(lpBitmapHeader != NULL)X__Free((char *)lpBitmapHeader);
 hDrawDib       = NULL;
 lpBitmapHeader = NULL;
 lpBitmapBits   = NULL;
 return;
}

static void PaintPreviewUpdateView(HDC hDC,RECT rc){
 if(hDrawDib != NULL && lpBitmapHeader != NULL && lpBitmapBits != NULL){
   DrawDibRealize(hDrawDib,hDC,TRUE);
   DrawDibDraw(hDrawDib,hDC,
            0,
            0,
            rc.right,
            rc.bottom,
           (LPBITMAPINFOHEADER)lpBitmapHeader,lpBitmapBits,
           0,0,
           (int)((LPBITMAPINFOHEADER)lpBitmapHeader)->biWidth,
           (int)((LPBITMAPINFOHEADER)lpBitmapHeader)->biHeight,
           0);
 }
 return;
}

static void MakePreview(HWND hwnd_control){
 LPSTR p;
 XIMAGE Ximage;
 fullscreenbuffer *fsb,*f;
 char buffer[256];
 long i,j,id,index;
 if(lpBitmapBits == NULL)return;
 if((fsb=(fullscreenbuffer *)
   X__Malloc(BitmapSizeX*BitmapSizeY*sizeof(fullscreenbuffer))) == NULL){
   MessageBeep(MB_OK);
   return;
 }
 memset(fsb,0,BitmapSizeX*BitmapSizeY*sizeof(fullscreenbuffer));
 sprintf(buffer,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %f %f %ld %f",
         "FLARE.DLL",3,flicker,0,(BitmapSizeX*2)/3,(BitmapSizeY*1)/3,
         ipx,ipy,ipz,model,version,
         bscale,rscale,id,angle);
 memset(&Ximage,0,sizeof(XIMAGE));
 Ximage.version=1;
 Ximage.Frame=1;
 Ximage.Xmax=BitmapSizeX;
 Ximage.Ymax=BitmapSizeY;
 Ximage.Screen=fsb;
 _RenderImageProcess(buffer,&Ximage);
 for(j=0;j<BitmapSizeY;j++){
   index = BitmapSizeY - 1 - j;
   p=(lpBitmapBits+index*BitmapSizeX*3);
   id=(j*Ximage.Ymax)/BitmapSizeY;
   id *= Ximage.Xmax;
   for(i=0;i<BitmapSizeX;i++){
     f = (fsb + id + (i*Ximage.Xmax)/BitmapSizeX);
     *p++ = f->B;
     *p++ = f->G;
     *p++ = f->R;
   }
   if((j%10) == 0){
     InvalidateRect(hwnd_control,NULL,FALSE);
     UpdateWindow(hwnd_control);
   }
 }
 InvalidateRect(hwnd_control,NULL,FALSE);
 UpdateWindow(hwnd_control);
 X__Free(fsb);
 return;
}
