/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* starfld.c  */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <windows.h>
#include <commctrl.h>
#include "struct.h"           /* general structures    */
#include "ximage.h"           /* XIMAGE structure      */
#include "local.h"

#include "starfld.h"

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL; /* use to pick up resources from DLL   */
static long version=1;

#include "utils.h"

#include "paint.c"

#if 0
static double GetYint(double x, double y, double dx, double dy, double xe){
 double grad;
 grad=dy/dx;
 return grad*(xe-x)+y;
}

static double GetXint(double x, double y, double dx, double dy, double ye){
 double grad;
 grad=dx/dy;
 return grad*(ye-y)+x;
}

static void GetStarPosition(double mr, long xc, long yc,
                            long Xmax, long Ymax,
                            long *x, long *y){
 long dx,dy;
 double ddx,ddy,dxc,dyc,ye,xe,ds,de,fr;
 dx = *x - xc;
 dy = *y - yc;
 if(dx == 0 && dy == 0)dx=1;
 ddx=(double)dx; ddy=(double)dy;
 dxc=(double)xc; dyc=(double)yc;
 if(dx == 0){
   xe=xc;
   if(dy > 0)ye=(double)Ymax;
   else      ye=0.0;
 }
 else if(dy == 0){
   ye=yc;
   if(dx < 0)xe=0.0;
   else      xe=(double)Xmax;
 }
 else {
   if(dx > 0)xe=(double)Xmax;
   else      xe=0.0;
   ye=GetYint(dxc,dyc,ddx,ddy,xe);
   if(ye < 0){
     xe=GetXint(dxc,dyc,ddx,ddy,0);
     ye=0.0;
   }
   else if(ye > Ymax){
     xe=GetXint(dxc,dyc,ddx,ddy,(double)Ymax);
     ye=(double)Ymax;
   }
 }
 ds=DIST(ddx,ddy);
 de=DIST(xe-dxc,ye-dyc);
 fr=frac(mr+ds/de);
 *x = xc+(long)((xe-dxc)*fr);
 *y = yc+(long)((ye-dyc)*fr);
 return;
}

 dmax=DIST(xc,yc);
 if((dc=DIST(Xmax-xc,yc)) > dmax)dmax=dc;
 if((dc=DIST(Xmax-xc,Ymax-yc)) > dmax)dmax=dc;
 if((dc=DIST(xc,Ymax-yc)) > dmax)dmax=dc;
 if(nstar > 0)for(i=0;i<nstar;i++){
   x=(long)((double)rand()*(double)Xmax/32767.0);
   y=(long)((double)rand()*(double)Ymax/32767.0);
   C=((double)rand()*(double)63/32767.0);
   GetStarPosition(mr,xc,yc,Xmax,Ymax,&x,&y);
   if((lpXimage->Screen+(long)y*Xmax+(long)x)->A == 0){
     dc=DIST(x-xc,y-yc);
     w=0.1+wmax*dc/dmax;
     DrawAApixel((double)x,(double)y,
     w,w,192+C,lpXimage->Screen,Xmax,Ymax,1.0,1.0,1.0);
   }
 }
#endif

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BC__
BOOL WINAPI DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

/* h is 0->360  s and v are 0-1 */
void Hsv2Rgb(double h, double s, double v,
             double *r, double *g, double *b){
 double rr,gg,bb,hh,ss,vv,maxcol,mincol;
 hh=(double)h; ss=(double)s; vv=(double)v;
 mincol=vv*(1.0-ss);
 if(hh <= 120.0){bb=mincol;
   if(hh <=  60){rr=vv; gg=mincol+hh*(vv-mincol)/(120.0-hh);}
   else         {gg=vv; rr=mincol+(120.0-hh)*(vv-mincol)/hh;}
 }
 else if( hh <= 240){rr=mincol;
   if(hh <= 180){gg=vv; bb=mincol+(hh-120.0)*(vv-mincol)/(240.0-hh);}
   else         {bb=vv; gg=mincol+(240.0-hh)*(vv-mincol)/(hh-120.0);}
 }
 else {gg=mincol;
   if(hh <= 300){bb=vv; rr=mincol+(hh-240.0)*(vv-mincol)/(360.0-hh);}
   else         {rr=vv; bb=mincol+(360.0-hh)*(vv-mincol)/(hh-240.0);}
 }
 *r = rr;
 *g = gg;
 *b = bb;
}

static void DrawInBuffer1(long x, long y, fullscreenbuffer *Screen,
                          long X, long Y, long radius,
                          double r, double g, double b){
 /* mix in a white halo at the screen buffer location of the light source */
 long i,j;
 double d;
 fullscreenbuffer *S;
 for(i=x-radius;i<=x+radius;i++)for(j=y-radius;j<=y+radius;j++){
   if(i < 0 || i >= X || j < 0 || j >= Y)continue;
   S=(Screen + j*X + i);
   if(radius < 1)d=1.0;
   else{
     d=sqrt((double)(i-x)*(double)(i-x) + (double)(j-y)*(double)(j-y));
     d=255.0*((double)radius-d)/(double)radius;
     d=max(0.0,d);
   }
   S->R = (unsigned char)min(255.0,(double)S->R+d*r);
   S->G = (unsigned char)min(255.0,(double)S->G+d*g);
   S->B = (unsigned char)min(255.0,(double)S->B+d*b);
 }
 return;
}

void DrawInBuffer(fullscreenbuffer *Screen, int Xmax, int Ymax,
                  int x, int y, double v,
                  double rr, double gg, double bb){
 int r,g,b;
 fullscreenbuffer *S;
 if(x < 0 || y < 0 || x >= Xmax || y >= Ymax)return;
 S=(Screen+y*Xmax+x);
 r=(int)S->R+(int)((double)v*rr);
 g=(int)S->G+(int)((double)v*gg);
 b=(int)S->B+(int)((double)v*bb);
 S->R=(r>255) ? 255:r;
 S->G=(g>255) ? 255:g;
 S->B=(b>255) ? 255:b;
}

#define DIST(i,j)  sqrt((double)(i) * (double)(i) + (double)(j) * (double)(j))

#define frac(z) fmod(z,1.0)

void DrawAApixel(double x, double y, double w, double h, double C,
                 fullscreenbuffer *S, int X, int Y,
                 double r, double g, double b){
  int xa,xb,ya,yb,i,j,brite;
  double x2,y2,lfrac,rfrac,tfrac,bfrac,xfrac,yfrac;

  if(w > 4 || h > 4){
    DrawInBuffer1((long)x,(long)y,S,X,Y,(long)w,r,g,b);
    return;
  }
  brite=(int)C;
  if(x < 0 || y < 0)return;
  x -= w*0.5; y -= h*0.5;
  x2 = x + w;
  y2 = y + w;
  xa = (int)x;
  xb = (int)x2;
  ya = (int)y;
  yb = (int)y2;
  lfrac = 1.0-frac(x);
  rfrac = frac(x2);
  tfrac = 1.0-frac(y);
  bfrac = frac(y2);
  if (xa==xb) {
    xfrac = lfrac+rfrac-1.0;
    if (ya==yb) {
      DrawInBuffer(S,X,Y,xa,ya,xfrac*(tfrac+bfrac-1.0)*brite,r,g,b);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,xfrac*tfrac*brite,r,g,b);
      for (j=ya+1; j<yb; j++) DrawInBuffer(S,X,Y,xa,j,xfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xa,yb,xfrac*bfrac*brite,r,g,b);
    }
  }
  else {
    if (ya==yb) {
      yfrac = tfrac+bfrac-1.0;
      DrawInBuffer(S,X,Y,xa,ya,yfrac*lfrac*brite,r,g,b);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,ya,yfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xb,ya,yfrac*rfrac*brite,r,g,b);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,tfrac*lfrac*brite,r,g,b);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,ya,tfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xb,ya,tfrac*rfrac*brite,r,g,b);
      for (j=ya+1; j<yb; j++) {
        DrawInBuffer(S,X,Y,xa,j,lfrac*brite,r,g,b);
        for (i=xa+1; i<xb; i++)DrawInBuffer(S,X,Y,i,j,brite,r,g,b);
        DrawInBuffer(S,X,Y,xb,j,rfrac*brite,r,g,b);
      }
      DrawInBuffer(S,X,Y,xa,yb,bfrac*lfrac*brite,r,g,b);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,yb,bfrac*brite,r,g,b);
      DrawInBuffer(S,X,Y,xb,yb,bfrac*rfrac*brite,r,g,b);
    }
  }
}

static double GetW(double l, double dx, double dy,
                   double Wmin, double Wmax,
                   BOOL linear){
 double d,f;
 d=DIST(dx,dy);
 f=d/l;
 return (Wmax-(Wmax-Wmin)*f);
}

static double GetC(double l, double dx, double dy,
                   double Wmin, double Wmax,
                   BOOL linear){
 double d,f;
 d=DIST(dx,dy);
 f=d/l;
 if(f < 0.5)return Wmax;
 return (Wmin+(Wmax-Wmin)*2.0*(1.0-f));
}

static void DrawFlareInBuffer(long x, long y, long idx, long idy,
                              fullscreenbuffer *Screen,
                              long X, long Y,
                              double Cin,
                              double r, double g, double b,
                              double Wmin, double Wmax,
                              BOOL linear){
 long i,j;
 double grad,error,W,l,C;
 C=Cin;
 l=DIST(idx,idy);
 if(abs(idx) >= abs(idy)){
   grad=(double)idy/(double)idx; j=y;
   if(idy >= 0)error = -0.5; else error = 0.5;
   if(idx >= 0){
     for(i=x;i<x+idx;i++){
       error += grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(i-x),(double)(j-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y,r,g,b);
     }
   }
   else{
     for(i=x;i>x+idx;i--){
       error -= grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(i-x),(double)(j-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y,r,g,b);
     }
   }
 }
 else{
   grad=(double)idx/(double)idy; j=x;
   if(idx >= 0)error = -0.5; else error = 0.5;
   if(idy >= 0){
     for(i=y;i<y+idy;i++){
       error += grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(j-x),(double)(i-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y,r,g,b);
     }
   }
   else{
     for(i=y;i>y+idy;i--){
       error -= grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax,linear);
       if(!linear)C=GetC(l,(double)(j-x),(double)(i-y),Cin*0.25,Cin,TRUE);
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y,r,g,b);
     }
   }
 }
 return;
}

long _RenderImageProcess(char *PrmList, XIMAGE *lpXimage){
 int i,j,nstar,nback,type,multic,xc,yc,Xmax,Ymax,accel,trails,cycles,scale;
 char dummy[255];
 double x,y,z,zt,xt,yt,Zmax;
 double mr,C,w,wmax=2.5;
 double r=1.0,g=1.0,b=1.0,h,s;
#include "pro_key.c"
 sscanf(PrmList,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",dummy,&version,
        &nstar,&nback,&type,&multic,&xc,&scale,&accel,&trails,&cycles);
 mr=(double)(lpXimage->last_frame - lpXimage->first_frame)+1.0;
 mr /= (double)cycles;
 if(mr < 1.0)return 0;
 mr=(double)(lpXimage->this_frame - lpXimage->first_frame)/mr;
 mr=frac(mr);
 if(type == 1 || type == 4)mr=1.0-mr;
 wmax=wmax*(double)scale*0.01;
 Xmax=lpXimage->Xmax;
 Ymax=lpXimage->Ymax;
 xc=Xmax/2; yc=Ymax/2;
 if(nback > 0)for(i=0;i<nback;i++){
   x=(long)((double)rand()*(double)Xmax/32767.0);
   y=(long)((double)rand()*(double)Ymax/32767.0);
   C=((double)rand()*(double)63/32767.0);
   if((lpXimage->Screen+(long)y*Xmax+(long)x)->A == 0)
       DrawAApixel((double)x,(double)y,
       1.5,1.5,128+C,lpXimage->Screen,Xmax,Ymax,1.0,1.0,1.0);
 }
 if((type == 1 || type == 2) && nstar > 0){
   Zmax=10.0*(double)Xmax;
   for(i=0;i<nstar;i++){
     if(multic == 1){
       h=360.0*(double)rand()/32767.0;
       s=0.5+0.5*(double)rand()/32767.0;
       Hsv2Rgb(h,s,1.0,&r,&g,&b);
     }
     x=((double)rand()*4.0*(double)Xmax/32767.0)-2*(double)Xmax;
     y=((double)rand()*4.0*(double)Ymax/32767.0)-2*(double)Ymax;
     if(fabs(x) < Xmax/2 && fabs(y) < Ymax/2){
       nstar++;
       continue;
     }
     z=((double)rand()*Zmax/32767.0);
     z=fmod(z+mr*Zmax,Zmax);
     z+=Xmax;
     if(trails == 1){
       if(type == 1)zt=z+Xmax;
       if(type == 2)zt=z-Xmax;
       xt=x*(double)Xmax/zt+(double)Xmax/2;
       yt=y*(double)Xmax/zt+(double)Ymax/2;
     }
     C=((double)rand()*(double)63/32767.0);
     x=x*(double)Xmax/z+(double)Xmax/2;
     y=y*(double)Xmax/z+(double)Ymax/2;
     if(x >= 0 && x < Xmax && y >= 0 && y < Ymax &&
       (lpXimage->Screen+(long)y*Xmax+(long)x)->A == 0){
       w=0.2+wmax*(Zmax-z+Xmax)/Zmax;
       if(w < 3.0)DrawAApixel((double)x,(double)y,
                              w,w,192+C,lpXimage->Screen,
                              Xmax,Ymax,
                              r,g,b);
       else DrawInBuffer1((long)x,(long)y,lpXimage->Screen,
                           Xmax,Ymax,(long)w,
                           r,g,b);
       if(trails == 1){
         DrawFlareInBuffer((long)x,(long)y,
                           (long)(xt-x),(long)(yt-y),
                           lpXimage->Screen,
                           Xmax,Ymax,
                           192+C,
                           r,g,b,
                           0.1,w,
                           TRUE);
       }
     }
   }
 }
 if((type == 3 || type == 4) && nstar > 0){
   Zmax=4.0*(double)Xmax;
   wmax /= 2.0;
   for(i=0;i<nstar;i++){
     if(multic == 1){
       h=360.0*(double)rand()/32767.0;
       s=0.5+0.5*(double)rand()/32767.0;
       Hsv2Rgb(h,s,1.0,&r,&g,&b);
     }
     x=((double)rand()*4.0*(double)Xmax/32767.0);
     x=fmod(x+mr*4.0*(double)Xmax,4.0*(double)Xmax);
     x -= 2.0*(double)Xmax;
     y=((double)rand()*4.0*(double)Ymax/32767.0);
     y -= 2.0*(double)Ymax;
     z=((double)rand()*Zmax/32767.0);
     C=((double)rand()*(double)63/32767.0);
     x=x*(double)Xmax/z+(double)Xmax/2;
     y=y*(double)Xmax/z+(double)Ymax/2;
     if(x >= 0 && x < Xmax && y >= 0 && y < Ymax &&
       (lpXimage->Screen+(long)y*Xmax+(long)x)->A == 0){
       w=1.0+wmax*(Zmax-z)/Zmax;
       if(w < 3.0)DrawAApixel((double)x,(double)y,w,w,192+C,
                              lpXimage->Screen,
                              Xmax,Ymax,
                              r,g,b);
       else DrawInBuffer1((long)x,(long)y,lpXimage->Screen,
                           Xmax,Ymax,(long)w,
                           r,g,b);
       if(trails == 1){
         if(type == 3)xt= -(double)Xmax*0.1;
         if(type == 4)xt=  (double)Xmax*0.1;
         DrawFlareInBuffer((long)x,(long)y,
                           (long)xt,0,
                           lpXimage->Screen,
                           Xmax,Ymax,
                           192+C,
                           r,g,b,
                           0.1,w,
                           TRUE);

       }
     }
   }
 }
 return 1;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static long nstar=100,nback=0,type=1,multic=0,xc=160,scale=100,accel=0,
            trails=0,cycles=1;

static X__MEMORY_MANAGER *lpLocalEVI;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
 char buffer[256];
 if(Op != NULL){  /* parameters exist so read them off the list */
   sscanf(Op,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",buffer,&version,
          &nstar,&nback,&type,&multic,&xc,&scale,&accel,&trails,&cycles);
 }
 lpLocalEVI=lpEVI;
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_STARFIELD),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(buffer,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",name,version,
         nstar,nback,type,multic,xc,scale,accel,trails,cycles);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
  MessageBox (GetFocus(),"Memory",NULL,MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

static void LoadAnimatedClip(HWND hDlg){
 char *c,modname[256];
 GetModuleFileName(hDLLinstance,modname,255);
 if((c=strrchr(modname,'.')) != NULL){
   strcpy(c,".avi");
   Animate_Open(GetDlgItem(hDlg,DLG_IMAGE),modname);
   Animate_Play(GetDlgItem(hDlg,DLG_IMAGE),0, -1, -1);
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 BOOL err;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     SetDlgItemInt(hwnd,DLG_STARFIELD_NSTAR,nstar,FALSE);
     SetDlgItemInt(hwnd,DLG_STARFIELD_NBACK,nback,FALSE);
     SetDlgItemInt(hwnd,DLG_STARFIELD_CYCLES,cycles,FALSE);
     if(type == 1)SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE1,BM_SETCHECK,TRUE,0);
     if(type == 2)SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE2,BM_SETCHECK,TRUE,0);
     if(type == 3)SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE3,BM_SETCHECK,TRUE,0);
     if(type == 4)SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE4,BM_SETCHECK,TRUE,0);
     if(multic == 1)SendDlgItemMessage(hwnd,DLG_STARFIELD_MULTICOLOUR,BM_SETCHECK,TRUE,0);
     if(trails == 1)SendDlgItemMessage(hwnd,DLG_STARFIELD_TRAILS,BM_SETCHECK,TRUE,0);
     SetDlgItemInt(hwnd,DLG_STARFIELD_SIZE,scale,FALSE);
     LoadAnimatedClip(hwnd);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hwnd);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          nstar=GetDlgItemInt(hwnd,DLG_STARFIELD_NSTAR,&err,FALSE);
          nback=GetDlgItemInt(hwnd,DLG_STARFIELD_NBACK,&err,FALSE);
          cycles=GetDlgItemInt(hwnd,DLG_STARFIELD_CYCLES,&err,FALSE);
          scale=GetDlgItemInt(hwnd,DLG_STARFIELD_SIZE,&err,FALSE);
          if(SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE1,BM_GETCHECK,0,0))
               type=1;
          if(SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE2,BM_GETCHECK,0,0))
               type=2;
          if(SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE3,BM_GETCHECK,0,0))
               type=3;
          if(SendDlgItemMessage(hwnd,DLG_STARFIELD_TYPE4,BM_GETCHECK,0,0))
               type=4;
          if(SendDlgItemMessage(hwnd,DLG_STARFIELD_TRAILS,BM_GETCHECK,0,0))
               trails=1;
          else trails=0;
          if(SendDlgItemMessage(hwnd,DLG_STARFIELD_MULTICOLOUR,BM_GETCHECK,0,0))
               multic=1;
          else multic=0;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

