/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* suprnova.c  image post-processor                                         */
/*                                                                       */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <windows.h>
#include <commctrl.h>
#include "struct.h"           /* general structures    */
#include "ximage.h"           /* XIMAGE structure      */
#include "local.h"

#include "suprnova.h"

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL; /* use to pick up resources from DLL   */

#include "utils.h"
#include "paint.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#elif __BC__
BOOL WINAPI DllEntryPoint(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif


static void GrayWholeScreen(double mr,
                            fullscreenbuffer *Screen,
                            float *Zbuffer, double y,
                            int Xmax, int Ymax){
 int i,j,r,g,b,v=128;
 fullscreenbuffer *S;
 if(mr < 0.12)v=(int)(160.0*mr/0.12);
 for(j=0;j<Ymax;j++)for(i=0;i<Xmax;i++){
   S=(Screen+j*Xmax+i);
//   if(y > *(Zbuffer+j*Xmax+i)){
   if(S->A < 64){
     r=(int)S->R+v;
     g=(int)S->G+v;
     b=(int)S->B+v;
     S->R=(r>255) ? 255:r;
     S->G=(g>255) ? 255:g;
     S->B=(b>255) ? 255:b;
   }
   else{
     S->R = (unsigned char)(0.5*(double)S->R+64);
     S->G = (unsigned char)(0.5*(double)S->G+64);
     S->B = (unsigned char)(0.5*(double)S->B+64);
   }
 }
}

double Grad(double x, double y1, double y2, int x1, int x2){
 double g;
 if(x <= x1)return y1;
 if(x >= x2)return y2;
 g=(y2-y1)/(double)(x2-x1);
 return y1+g*(x-(double)x1);
}

void DrawInBuffer(fullscreenbuffer *Screen, int Xmax, int Ymax,
                  int x, int y, unsigned char v ){
 int r,g,b;
 fullscreenbuffer *S;
 if(x < 0 || y < 0 || x >= Xmax || y >= Ymax)return;
 S=(Screen+y*Xmax+x);
 r=(int)S->R+(int)v;
 g=(int)S->G+(int)v;
 b=(int)S->B+(int)v;
 S->R=(r>255) ? 255:r;
 S->G=(g>255) ? 255:g;
 S->B=(b>255) ? 255:b;
}

#define frac(z) fmod(z,1.0)

void DrawAApixel(double x, double y, double w, double h, double C,
                 fullscreenbuffer *S, int X, int Y){
  int xa,xb,ya,yb,i,j,brite;
  double x2,y2,lfrac,rfrac,tfrac,bfrac,xfrac,yfrac;
  brite=(int)C;
  if(x < 0 || y < 0)return;
  x -= w*0.5; y -= h*0.5;
  x2 = x + w;
  y2 = y + w;
  xa = (int)x;
  xb = (int)x2;
  ya = (int)y;
  yb = (int)y2;
  lfrac = 1.0-frac(x);
  rfrac = frac(x2);
  tfrac = 1.0-frac(y);
  bfrac = frac(y2);
  if (xa==xb) {
    xfrac = lfrac+rfrac-1.0;
    if (ya==yb) {
      DrawInBuffer(S,X,Y,xa,ya,xfrac*(tfrac+bfrac-1.0)*brite);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,xfrac*tfrac*brite);
      for (j=ya+1; j<yb; j++) DrawInBuffer(S,X,Y,xa,j,xfrac*brite);
      DrawInBuffer(S,X,Y,xa,yb,xfrac*bfrac*brite);
    }
  }
  else {
    if (ya==yb) {
      yfrac = tfrac+bfrac-1.0;
      DrawInBuffer(S,X,Y,xa,ya,yfrac*lfrac*brite);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,ya,yfrac*brite);
      DrawInBuffer(S,X,Y,xb,ya,yfrac*rfrac*brite);
    }
    else {
      DrawInBuffer(S,X,Y,xa,ya,tfrac*lfrac*brite);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,ya,tfrac*brite);
      DrawInBuffer(S,X,Y,xb,ya,tfrac*rfrac*brite);
      for (j=ya+1; j<yb; j++) {
        DrawInBuffer(S,X,Y,xa,j,lfrac*brite);
        for (i=xa+1; i<xb; i++)DrawInBuffer(S,X,Y,i,j,brite);
        DrawInBuffer(S,X,Y,xb,j,rfrac*brite);
      }
      DrawInBuffer(S,X,Y,xa,yb,bfrac*lfrac*brite);
      for (i=xa+1; i<xb; i++) DrawInBuffer(S,X,Y,i,yb,bfrac*brite);
      DrawInBuffer(S,X,Y,xb,yb,bfrac*rfrac*brite);
    }
  }
}

#define DIST(i,j)  sqrt((double)i * (double)i + (double)j * (double)j)

static double GetW(double l, double dx, double dy,
                   double Wmin, double Wmax, double iScale,
                   BOOL linear){
 double d,f;
 d=DIST(dx,dy);
 f=d/l;
 if(linear){
   if(f < 0.5)return (Wmin+2.0*f*(Wmax-Wmin))*(double)iScale;
   return (Wmin+(Wmax-Wmin)*2.0*(1.0-f))*(double)iScale;
 }
 else{
   double y,dx,dw;
   dw=(Wmax-Wmin);
   if(f < 0.5)f=1.0-f;
   f = 2.0*(f-0.5);
   y=Wmax-dw*f*f*f;
   return y*(double)iScale;
 }
 return 0.0;
}

static double GetC(double l, double dx, double dy,
                   double Wmin, double Wmax, double iScale,
                   BOOL linear){
 double d,f;
 d=DIST(dx,dy);
 f=d/l;
 if(f < 0.5)return Wmax;
 return (Wmin+(Wmax-Wmin)*2.0*(1.0-f));
}

static void DrawFlareInBuffer(long x, long y, long idx, long idy,
                              fullscreenbuffer *Screen,
                              long X, long Y, double iScale,
                              double Cin,
                              double Wmin, double Wmax,
                              double z, float *Zb,
                              BOOL bDraw, BOOL linear){
 long i,j;
 double grad,error,W,l,C;
 if(iScale <1.1)Cin *= 2; Cin=min(255,Cin); C=Cin;
 l=DIST(idx,idy);
 if(abs(idx) >= abs(idy)){
   grad=(double)idy/(double)idx; j=y;
   if(idy >= 0)error = -0.5; else error = 0.5;
   if(idx >= 0){
     for(i=x;i<x+idx;i++){
       error += grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(double)(i-x),(double)(j-y),Cin*0.25,Cin,1.0,TRUE);
if(bDraw && (Screen+j*X+i)->A < 64)
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y);
     }
   }
   else{
     for(i=x;i>x+idx;i--){
       error -= grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(i-x),(double)(j-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(double)(i-x),(double)(j-y),Cin*0.25,Cin,1.0,TRUE);
if(bDraw && (Screen+j*X+i)->A < 64)
       DrawAApixel((double)i,(double)j,W,W,C,Screen,X,Y);
     }
   }
 }
 else{
   grad=(double)idx/(double)idy; j=x;
   if(idx >= 0)error = -0.5; else error = 0.5;
   if(idy >= 0){
     for(i=y;i<y+idy;i++){
       error += grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(double)(j-x),(double)(i-y),Cin*0.25,Cin,1.0,TRUE);
if(bDraw && (Screen+i*X+j)->A < 64)
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y);
     }
   }
   else{
     for(i=y;i>y+idy;i--){
       error -= grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(double)(j-x),(double)(i-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(double)(j-x),(double)(i-y),Cin*0.25,Cin,1.0,TRUE);
if(bDraw && (Screen+i*X+j)->A < 64)
       DrawAApixel((double)j,(double)i,W,W,C,Screen,X,Y);
     }
   }
 }
 return;
}

long _RenderImageProcess(char *PrmList, XIMAGE *lpXimage){
 int i,j,Nl,n=45;
 char dummy[255];
 double x,y,z,radius,atnB,atnL,atnV=1.0;
 long v,model,type,flicker,id,sx,sy,ipx,ipy,ipz;
 long xx,yy,Xmax,Ymax,Xc,Yc;
 long xBlue,yBlue,xBrown,yBrown,xGreen,yGreen,xT,yT;
 double iScale,ascale,lscale,dlscale,d2r,lc,dx,dy;
 double mr;
 float *Zb;
 BOOL  bDraw=TRUE;
#include "pro_key.c"
 sscanf(PrmList,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
        dummy,&type,&flicker,&id,&sx,&sy,&ipx,&ipy,&ipz,&model,&v);
 mr=(double)(lpXimage->last_frame - lpXimage->first_frame);
 if(mr < 1.0)return 1;
 mr=(double)(lpXimage->this_frame - lpXimage->first_frame)/mr;
 if(type == 1 && (lpXimage->Nlights == 0 || lpXimage->Lights == NULL))return 1;
 Xmax=lpXimage->Xmax;  Ymax=lpXimage->Ymax;
 Xc=Xmax/2; Yc=Ymax/2;
 iScale=(double)Xmax/320;
 ascale=5.0/32767.0;
 lscale=80.0*(double)iScale;
 dlscale=80.0/32767.0*(double)iScale;
 d2r=3.1415626/180.0;
 Zb=lpXimage->Zbuffer;
 if(flicker)srand(lpXimage->Frame);
 else       srand(1);
 if(type == 1)Nl=lpXimage->Nlights;
 else Nl=1;
 for(i=0;i<Nl;i++){
   if(type == 1){
     x=lpXimage->Lights[i].p[0];
     y=lpXimage->Lights[i].p[1];
     z=lpXimage->Lights[i].p[2];
   }
   else if(type == 2){
     TransformIntoView(lpXimage->ViewTransform,
                       (double)ipx,(double)ipy,(double)ipz,
                       &x,&y,&z);
   }
   else if(type == 3){
     y=1.1;
   }
   else if(type == 4){
     if(lpXimage->Lights[i].type != DUMMYL)continue;
     x=lpXimage->Lights[i].p[0];
     y=lpXimage->Lights[i].p[1];
     z=lpXimage->Lights[i].p[2];
   if(type == 3 && lpXimage->Lights[i].AnimatorID != id)continue;
   }
   else if(type == 5){
     if(lpXimage->Lights[i].AnimatorID != id)continue;
     x=lpXimage->Lights[i].p[0];
     y=lpXimage->Lights[i].p[1];
     z=lpXimage->Lights[i].p[2];
   }
   if(y >= 1.000){
     if(type == 3){
       xx=sx; yy=sy;
     }
     else{
       xx=Xc + (long)(lpXimage->Xscale*x/y);
       yy=Yc - (long)(lpXimage->Yscale*z/y);
     }
     if(1){
       dx=(double)(xx-Xc);  dy=(double)(yy-Yc);
       lc=DIST(dx,dy);  dx /= lc; dy /= lc;
       atnB=atnV; atnL=atnV;
       if(mr < 0.8){
         atnB = atnL = atnV*mr/0.8;
       }
       else if(mr > 0.9){
         atnB = atnL = atnV*(1.0-mr)/0.1;
       }
       if(model == 1){
         for(j=0;j<n;j++){  /* outlying spines */
           double a,l,br,mw;
           long   idx,idy;
           a=360.0/(double)n*(double)j+ascale*(double)rand();
           a *= d2r;
           l=lscale+dlscale*(double)rand();
           l *= atnL;
           idx=(long)(l*cos(a)); idy=(long)(l*sin(a));
           br=20.0*atnB;
           br=64.0*atnB;
           mw=0.8;
           mw=1.5;
           DrawFlareInBuffer(xx,yy,idx,idy,lpXimage->Screen,Xmax,Ymax,
                             iScale,
                             br, // brightness 0-255
                             0.2,// min width
                             mw, // max width
                             y,Zb,bDraw,TRUE);
         }
         for(i=0;i<3;i++)for(j=0;j<n;j++){ /* centre spines */
           double a,l,C;
           long   idx,idy;
           a=360.0/(double)n*(double)j+ascale*(double)rand();
           a *= d2r;
           l=lscale+dlscale*(double)rand();
           l *= atnL;
           if     (i == 0){l /= 5;            C=32.0 *atnB;}
           else if(i == 1){l /= 10;           C=64.0 *atnB;}
           else if(i == 2){l /= 20;           C=128.0*atnB;}
           idx=(long)(l*cos(a)); idy=(long)(l*sin(a));
           DrawFlareInBuffer(xx,yy,idx,idy,lpXimage->Screen,Xmax,Ymax,
                             iScale,
                             C,
                             0.2,0.8,
                             y,Zb,bDraw,TRUE);
         }
       }
       else{ /* model2 */
         for(j=0;j<5;j++){
           double a,l,ddx,ddy;
           long   xs,ys,idx,idy;
           a= -75.0 + 360.0/(double)5*(double)j;
           a *= d2r;
           l=lscale*1.5;
           l *= atnL;
           ddx=l*cos(a); ddy=l*sin(a);
           idx=(long)ddx; idy=(long)ddy;
           xs=xx+(long)(ddx*0.09); ys=yy+(long)(ddy*0.09);
           DrawFlareInBuffer(xs,ys,idx,idy,lpXimage->Screen,Xmax,Ymax,
                             iScale,
                             40.0*atnB,
                             0.0,1.4*atnB,
                             y,Zb,bDraw,FALSE);
           DrawFlareInBuffer(xs,ys,idx,idy,lpXimage->Screen,Xmax,Ymax,
                             iScale,
                             20.0*atnB,
                             0.0,2.8*atnB,
                             y,Zb,bDraw,FALSE);
           xs=xx+(long)(ddx*0.2); ys=yy+(long)(ddy*0.2);
           DrawFlareInBuffer(xs,ys,idx,idy,lpXimage->Screen,Xmax,Ymax,
                             iScale,
                             10.0*atnB,     // brightness 0-255
                             0.0,4.5*atnB,  // min width max width
                             y,Zb,bDraw,FALSE);
#if 0
#endif
         }
       }
       if(mr > 0.7 && mr < 0.9)GrayWholeScreen(mr-0.7,
                                               lpXimage->Screen,
                                               lpXimage->Zbuffer,y,
                                               Xmax,Ymax);
     }
   }
 }
 return 1;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

static long   version=1,type=1,model=1,flicker=0,id = -1 ,
              sx=0,sy=0,ipx=0,ipy=0,ipz=0;
static double bscale=1.0,rscale=1.0;
static double px,py,pz;
static X__MEMORY_MANAGER *lpLocalEVI;

char * _SetExternalParameters(
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  long ruler,               /* ruler scale value to facilitate scaling    */
  char *name,               /* name of DLL file with the effect           */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                                    ){
 char buffer[256];
 if(Op != NULL){
   sscanf(Op,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %f %f",
          buffer,&type,&flicker,&id,&sx,&sy,&ipx,&ipy,&ipz,&model,&version,
          &bscale,&rscale);
    px=(double)ipx/(double)ruler;
    py=(double)ipy/(double)ruler;
    pz=(double)ipz/(double)ruler;
 }
 lpLocalEVI=lpEVI;
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_SUPRNOVA),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 ipx=(long)(px*(double)ruler);
 ipy=(long)(py*(double)ruler);
 ipz=(long)(pz*(double)ruler);
 sprintf(buffer,"%s %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %f %f",
         name,type,flicker,id,sx,sy,ipx,ipy,ipz,model,version,bscale,rscale);
 if((Op=(char *)CALL_MALLOC(strlen(buffer)+1)) == NULL){
   return NULL;
 }
 strcpy(Op,buffer);
 return Op;
}

static void LoadAnimatedClip(HWND hDlg){
 char *c,modname[256];
 GetModuleFileName(hDLLinstance,modname,255);
 if((c=strrchr(modname,'.')) != NULL){
   strcpy(c,".avi");
   Animate_Open(GetDlgItem(hDlg,DLG_ANIM),modname);
   Animate_Play(GetDlgItem(hDlg,DLG_ANIM),0, -1, -1);
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 BOOL err;
 char str[32];
 switch( msg ) {
   case WM_INITDIALOG:
     if     (type == 1)
       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE1,BM_SETCHECK,TRUE,0);
     else if(type == 2)
       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE2,BM_SETCHECK,TRUE,0);
     else if(type == 3)
       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE3,BM_SETCHECK,TRUE,0);
     else if(type == 4)
       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE4,BM_SETCHECK,TRUE,0);
     else if(type == 5)
       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE5,BM_SETCHECK,TRUE,0);
//     if     (model == 2)
//       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_MODEL2,BM_SETCHECK,TRUE,0);
//     else if(model == 3)
//       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_MODEL3,BM_SETCHECK,TRUE,0);
//     else
//       SendDlgItemMessage(hwnd,DLG_SUPRNOVA_MODEL1,BM_SETCHECK,TRUE,0);
     sprintf(str,"%.3f",px);
     SetDlgItemText(hwnd,DLG_SUPRNOVA_PX,str);
     sprintf(str,"%.3f",py);
     SetDlgItemText(hwnd,DLG_SUPRNOVA_PY,str);
     sprintf(str,"%.3f",pz);
     SetDlgItemText(hwnd,DLG_SUPRNOVA_PZ,str);
     SetDlgItemInt(hwnd,DLG_SUPRNOVA_SX,sx,FALSE);
     SetDlgItemInt(hwnd,DLG_SUPRNOVA_SY,sy,FALSE);
//     sprintf(str,"%.3f",bscale);
//     SetDlgItemText(hwnd,DLG_SUPRNOVA_BSCALE,str);
//     sprintf(str,"%.3f",rscale);
//     SetDlgItemText(hwnd,DLG_SUPRNOVA_RSCALE,str);
     if(id >= 0)SendDlgItemMessage(hwnd,DLG_SUPRNOVA_NAME,WM_SETTEXT,
                0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id));
     LoadAnimatedClip(hwnd);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_PAINT:
     PaintBackground(hwnd);
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_SUPRNOVA_SETNAME:
          id=GetActorsID(lpLocalEVI->lpAni,hwnd);
          if(id >= 0)SendDlgItemMessage(hwnd,DLG_SUPRNOVA_NAME,WM_SETTEXT,
                     0,(LPARAM)GetActorsName(lpLocalEVI->lpAni,id));
          break;
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_SUPRNOVA_PX,str,12) != 0)
            sscanf(str,"%f",&px);
          if(GetDlgItemText(hwnd,DLG_SUPRNOVA_PY,str,12) != 0)
            sscanf(str,"%f",&py);
          if(GetDlgItemText(hwnd,DLG_SUPRNOVA_PZ,str,12) != 0)
            sscanf(str,"%f",&pz);
//          if(GetDlgItemText(hwnd,DLG_SUPRNOVA_BSCALE,str,12) != 0)
//            sscanf(str,"%f",&bscale);
//          if(GetDlgItemText(hwnd,DLG_SUPRNOVA_RSCALE,str,12) != 0)
//            sscanf(str,"%f",&rscale);
          sx=GetDlgItemInt(hwnd,DLG_SUPRNOVA_SX,&err,FALSE);
          sy=GetDlgItemInt(hwnd,DLG_SUPRNOVA_SY,&err,FALSE);
          if     (SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE5,BM_GETCHECK,0,0))
               type=5;
          else if(SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE4,BM_GETCHECK,0,0))
               type=4;
          else if(SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE3,BM_GETCHECK,0,0))
               type=3;
          else if(SendDlgItemMessage(hwnd,DLG_SUPRNOVA_TYPE2,BM_GETCHECK,0,0))
               type=2;
          else type=1;
//          if     (SendDlgItemMessage(hwnd,DLG_SUPRNOVA_MODEL2,BM_GETCHECK,0,0))
//               model=2;
//          else if(SendDlgItemMessage(hwnd,DLG_SUPRNOVA_MODEL3,BM_GETCHECK,0,0))
//               model=3;
//          else model=1;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

