/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file VIDEO2.C */

#include "..\animate\gif.c"

static int rowcount;
static unsigned char *pixels = NULL;

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(rowcount)*(long)(x_size);
 memcpy(pixels + l, linepixels, linelen);
 rowcount++;
 return 0;
}

static unsigned char *LoadGIF(char *GIFfile, int *xx, int *yy){
 unsigned char *px,*lpB,*pixels24;
 long i,j,k,imagesize,x,y,xi,yi;
 rowcount=0;
 if(read_gif_image(GIFfile,1) < 0)return 0;
 x=x_size; y=y_size;
 imagesize=x_size*y_size;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return 0;
 if((pixels24=(unsigned char*)X__Malloc(imagesize*3)) == NULL){
   X__Free(pixels);
   return NULL;
 }
 if(read_gif_image(GIFfile,0) < 0){
   X__Free(pixels);
   return NULL;
 }
 px=pixels24;
 lpB=pixels;
 for(j=0;j<y;j++){
   for(i=0;i<x;i++){
     k=(long)(*lpB++);
     *px++ = (unsigned char)(p_red[k]   << 2);
     *px++ = (unsigned char)(p_green[k] << 2);
     *px++ = (unsigned char)(p_blue[k]  << 2);
   }
 }
 *xx=x; *yy=y;
 X__Free(pixels);
 return pixels24;
}

static void tga_display_line(unsigned char *prgb,
                             long wide, long line,
                             long upsidedown, long hi){
 register long l,x,k;
 if(!upsidedown)l = 3L*(long)(line)*(long)(x_size);
 else           l = 3L*(long)(hi - 1 - line)*(long)(x_size);
 memcpy(pixels + l, prgb, (int)(wide*3));
}

static short R_getvariword(long n, FILE *f, long *vw){
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;          if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

static short read_tga_image(char *filename, short info_only){
 char c0,c1,c2;
 long i,j,k,l,m,n,cmo,cml,xor,yor,wi,hi,iok=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat,wi3;
 FILE *t24;
 unsigned char *prgb;
 prgb=NULL;
 if((t24=fopen(filename,"rb")) != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0 : 1 ;
   if(info_only){
     x_size=(int)wi;
     y_size=(int)hi;
     fclose(t24);
     return 1;
   }
   if((prgb=(unsigned char *)X__Malloc((size_t)wi*3)) == NULL)goto CLOSEOUT;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi;
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   wi3=wi*3;
   if(image_type == 2){ /* uncompressed */
     m=0; k=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       prgb[m+2] = (char)( vw        & pat) << s;
       prgb[m+1] = (char)((vw >> s1) & pat) << s;
       prgb[m  ] = (char)((vw >> s2) & pat) << s;
       k++; m+=3;
       if(m == wi3){
         tga_display_line(prgb,wi,i,upsidedown,hi); m=0;
       }
     }
   }
   else{  /* compressed */
     m=0; l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           prgb[m+2] = (char)( vw        & pat) << s;
           prgb[m+1] = (char)((vw >> s1) & pat) << s;
           prgb[m  ] = (char)((vw >> s2) & pat) << s;
           k++; m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           prgb[m+2] = c0;
           prgb[m+1] = c1;
           prgb[m  ] = c2;
           k++;  m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
   if(prgb != NULL)X__Free((char *)prgb);
 }
 return (short)iok;
}

static unsigned char *LoadTGA(char *TGAfile, int *xx, int *yy){
 long imagesize;
 rowcount=0;
 if(read_tga_image(TGAfile,1) < 0)return 0;
 imagesize=x_size*y_size*3;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return NULL;
 if(read_tga_image(TGAfile,0) < 0){
   X__Free(pixels);
   return NULL;
 }
 *xx=x_size; *yy=y_size;
 return pixels;
}

unsigned char *LoadMAP(char *CurrentFile, int *xx, int *yy){
 int ff;
 if     (strstr(CurrentFile,".GIF") || strstr(CurrentFile,".gif"))
   return LoadGIF(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".TGA") || strstr(CurrentFile,".tga"))
   return LoadTGA(CurrentFile,xx,yy);
 return NULL;
}
