/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// file allocmem.cpp
//                     Allocate memory for the renderer's multithreading
//                     algorithm. This file also include the dithering
//                     functions.


#define MODULE_DITHER

#include "render.h"

#if __NT__
extern HINSTANCE hThisInstance;
#endif


extern long           *gZbuffer, *gXbuffer,
                      *gZobject, *gXobject,
                      *gZglass[],
                      *gZglassO[],
                      *gZcsg;
extern double         *gZdepth,
                      *gZposn,
                      *gZglassd[],
                      *gZground;
extern unsigned char  *gRed,   *gGreen,   *gBlue,   *gAlfa,
                      *gL_Red, *gL_Green, *gL_Blue, *gL_Alfa,
                      *gLL_Red,*gLL_Green,*gLL_Blue,*gLL_Alfa;
extern char gszHomeDir[];

static THD MainTHD;
static short   *D_red=NULL,
               *D_green=NULL,
               *D_blue=NULL,
               *red_error=NULL,
               *green_error=NULL,
               *blue_error=NULL;
static unsigned long dither_seed1,dither_seed2,dither_seed3;

#ifdef _SUNSTYLE
static unsigned short Random_Dither1(val) unsigned short val; {
#else
static unsigned short Random_Dither1(unsigned short val){
#endif
dither_seed1 = (dither_seed1 * 125 + 7) % 1717;
return ((unsigned short)(dither_seed1 % (val+1)));
}

#ifdef _SUNSTYLE
static unsigned short Random_Dither2(val) unsigned short val; {
#else
static unsigned short Random_Dither2(unsigned short val){
#endif
dither_seed2  = (dither_seed2 * 256 + 3) % 32749;
return ((unsigned short)(dither_seed2 % (val+1)));
}

#ifdef _SUNSTYLE
static unsigned short Random_Dither3(val) unsigned short val; {
#else
static unsigned short Random_Dither3(unsigned short val){
#endif
dither_seed3  = (dither_seed3 * 1001 + 7) % 17417;
return ((unsigned short)dither_seed3 % (val+1));
}

/****************************************\
*  Returns a combined random cardinal *
*  such that Random_Dither(val) < 17416 *
*  and Random_Dither(val) <= val  *
\****************************************/

#ifdef _SUNSTYLE
static unsigned short Random_Dither(val) unsigned short val; {
#else
static unsigned short Random_Dither(unsigned short val){
#endif
if (Random_Dither1(1))
return (Random_Dither2(val));
else
return (Random_Dither3(val));
}

BOOL AllocateThreadStorage(THD *s){
 long i,size,sl,sd,su;
 UCHAR *p;
 memset(s,0,sizeof(THD));
 size=XMAX*((5+2*NGLASS)*sizeof(long)
           +(3+1*NGLASS)*sizeof(double)
           +(12)*sizeof(UCHAR)
           );
 sl=sizeof(long)*XMAX;
 sd=sizeof(double)*XMAX;
 su=sizeof(UCHAR)*XMAX;
 if((s->ram=(UCHAR *)X__Malloc(size)) == NULL)return FALSE;
 p=s->ram;
 s->Zbuffer=(long *)p; p+=sl;
 s->Zobject=(long *)p; p+=sl;
 s->Zcsg   =(long *)p; p+=sl;
 s->Xbuffer=(long *)p; p+=sl;
 s->Xobject=(long *)p; p+=sl;
 for(i=0;i<NGLASS;i++){
   s->Zglass[i] =(long *)p; p+=sl;
   s->ZglassO[i]=(long *)p; p+=sl;
 }
 s->Zdepth =(double *)p; p+=sd;
 s->Zposn  =(double *)p; p+=sd;
 s->Zground=(double *)p; p+=sd;
 for(i=0;i<NGLASS;i++){
   s->Zglassd[i]=(double *)p; p+=sd;
 }
 s->Red     =(UCHAR *)p; p+=su;
 s->Green   =(UCHAR *)p; p+=su;
 s->Blue    =(UCHAR *)p; p+=su;
 s->Alfa    =(UCHAR *)p; p+=su;
 s->L_Red   =(UCHAR *)p; p+=su;
 s->L_Green =(UCHAR *)p; p+=su;
 s->L_Blue  =(UCHAR *)p; p+=su;
 s->L_Alfa  =(UCHAR *)p; p+=su;
 s->LL_Red  =(UCHAR *)p; p+=su;
 s->LL_Green=(UCHAR *)p; p+=su;
 s->LL_Blue =(UCHAR *)p; p+=su;
 s->LL_Alfa =(UCHAR *)p; p+=su;
 return TRUE;
}

void FreeThreadStorage(THD *s){
 if(s->ram != NULL)X__Free(s->ram);
 memset(s,0,sizeof(THD));
 return;
}

#ifdef _SUNSTYLE
short SetUpDitherScanLines(){
#else
short SetUpDitherScanLines(void){
#endif
  long i;
  THD *s;
#if __DEMO__
  if(XMAX > 800){
    Render_Message(17,0,NULL);
    return FAIL;
  }
#endif
  s=&MainTHD;
  if(!AllocateThreadStorage(s))return FAIL;
  gZbuffer=s->Zbuffer;  gZobject=s->Zobject;
  gZcsg=s->Zcsg;
  gZdepth=s->Zdepth;    gZposn=s->Zposn;
  gXbuffer=s->Xbuffer;  gXobject=s->Xobject;
  gZground=s->Zground;
  for(i=0;i<NGLASS;i++){
    gZglass[i] =s->Zglass[i];
    gZglassO[i]=s->ZglassO[i];
    gZglassd[i]=s->Zglassd[i];
  }
  gRed    =s->Red;     gGreen   =s->Green;
  gBlue   =s->Blue;    gAlfa    =s->Alfa;
  gL_Red  =s->L_Red;   gL_Green =s->L_Green;
  gL_Blue =s->L_Blue;  gL_Alfa  =s->L_Alfa;
  gLL_Red =s->LL_Red;  gLL_Green=s->LL_Green;
  gLL_Blue=s->LL_Blue; gLL_Alfa =s->LL_Alfa;
  ZskyBuffer=NULL;
  DL=NULL; DR=NULL; D1=NULL;  DummyShadow=NULL;
  RendererUshortBuffer=NULL; RendererShortBuffer=NULL;
  if((ZskyBuffer=(short *)X__Malloc(XMAX*sizeof(short))) == NULL)return FAIL;
  if((DL=(double *)X__Malloc(XMAX*sizeof(double))) == NULL)return FAIL;
  if((DR=(double *)X__Malloc(XMAX*sizeof(double))) == NULL)return FAIL;
  if((D1=(double *)X__Malloc(XMAX*sizeof(double))) == NULL)return FAIL;
  if((DummyShadow=(UCHAR *)X__Malloc(XMAX*sizeof(UCHAR))) == NULL)return FAIL;
  if((RendererUshortBuffer=(unsigned short *)
     X__Malloc(XMAX*sizeof(unsigned short))) == NULL)return FAIL;
  if((RendererShortBuffer=(short *)
     X__Malloc(XMAX*sizeof(short))) == NULL)return FAIL;
  if(screenbufferON)FullScreenBuffer=(fullscreenbuffer *)
    X__Malloc(ResolutionX*ResolutionY*(long)sizeof(fullscreenbuffer));
  if(mblurON > 0){
    MotionBlurBuffer=(fullscreenbuffer *)
      X__Malloc(ResolutionX*ResolutionY*(long)sizeof(fullscreenbuffer));
    if(MotionBlurBuffer != NULL)memset(MotionBlurBuffer,
       0,ResolutionX*ResolutionY*(long)sizeof(fullscreenbuffer));
  }
  if(objectIDbufferON)ObjectIdBuffer=(unsigned char *)
    X__Malloc(ResolutionX*ResolutionY*(long)sizeof(unsigned char));
  if(fszBufferON)fszBuffer=(float *)
    X__Malloc(ResolutionX*ResolutionY*(long)sizeof(float));
  if(dither_on == NO)return OK;
  if(dither_on == FS_DITHER){
    if((D_red  =(short  *)X__Malloc(XMAX*sizeof(short))) == NULL)
      dither_on = NO;
    if((D_green=(short  *)X__Malloc(XMAX*sizeof(short))) == NULL){
      dither_on = NO; X__Free(D_red);}
    if((D_blue =(short  *)X__Malloc(XMAX*sizeof(short))) == NULL){
      dither_on = NO; X__Free(D_red); X__Free(D_green);}
    if((red_error  =(short  *)X__Malloc(XMAX*sizeof(short))) == NULL){
      dither_on = NO; X__Free(D_red); X__Free(D_green); X__Free(D_blue);}
    if((green_error=(short  *)X__Malloc(XMAX*sizeof(short))) == NULL){
      dither_on = NO; X__Free(D_red); X__Free(D_green); X__Free(D_blue);
      X__Free(red_error);}
    if((blue_error =(short  *)X__Malloc(XMAX*sizeof(short))) == NULL){
      dither_on = NO; X__Free(D_red); X__Free(D_green); X__Free(D_blue);
      X__Free(red_error); X__Free(green_error);}
    if(dither_on == FS_DITHER)for(i=0;i<XMAX;i++){
      D_red[i] = D_green[i] = D_blue[i] = 0;
    }
  }
  else if(dither_on == RANDOM_DITHER){
    dither_seed1=0;
    dither_seed2=1;
    dither_seed3=2;
  }
  else{
    dither_on= NO;
  }
  return OK;
}

#ifdef _SUNSTYLE
void ResetDither(){
#else
void ResetDither(void){
#endif
 long i;
 if(dither_on == FS_DITHER){
   for(i=0;i<XMAX;i++){
     D_red[i]=D_green[i]=D_blue[i]=0;
   }
 }
 else if(dither_on == RANDOM_DITHER){
   dither_seed1=0;
   dither_seed2=1;
   dither_seed3=2;
 }
}

#ifdef _SUNSTYLE
void FreeDitherScanLines(){
#else
void FreeDitherScanLines(void){
#endif
  long i;
  FreeThreadStorage(&MainTHD);
  gZbuffer=NULL; gZobject=NULL; gZcsg=NULL; gZdepth=NULL; gZposn=NULL;
  gXbuffer=NULL; gXobject=NULL; gZground=NULL;
  for(i=0;i<NGLASS;i++){
    gZglass[i]=NULL; gZglassO[i]=NULL;  gZglassd[i]=NULL;
  }
  gRed=NULL; gGreen=NULL; gBlue=NULL; gAlfa=NULL;
  gL_Red=NULL; gL_Green=NULL; gL_Blue=NULL; gL_Alfa=NULL;
  gLL_Red=NULL; gLL_Green=NULL; gLL_Blue=NULL; gLL_Alfa=NULL;
  if(FullScreenBuffer != NULL)X__Free(FullScreenBuffer); FullScreenBuffer=NULL;
  if(MotionBlurBuffer != NULL)X__Free(MotionBlurBuffer); MotionBlurBuffer=NULL;
  if(ObjectIdBuffer != NULL)X__Free(ObjectIdBuffer); ObjectIdBuffer=NULL;
  if(fszBuffer != NULL)X__Free(fszBuffer); fszBuffer=NULL;
  if(ZskyBuffer != NULL)X__Free(ZskyBuffer); ZskyBuffer=NULL;
  if(DL != NULL)X__Free(DL); DL=NULL;
  if(DR != NULL)X__Free(DR); DR=NULL;
  if(D1 != NULL)X__Free(D1); D1=NULL;
  if(DummyShadow != NULL)X__Free(DummyShadow); DummyShadow=NULL;
  if(RendererUshortBuffer != NULL)X__Free(RendererUshortBuffer);
     RendererUshortBuffer=NULL;
  if(RendererShortBuffer != NULL)X__Free(RendererShortBuffer);
     RendererShortBuffer=NULL;
  if(dither_on == NO)return;
  if(dither_on == FS_DITHER){
    if(D_red != NULL)X__Free(D_red);
    if(D_green != NULL)X__Free(D_green);
    if(D_blue != NULL)X__Free(D_blue);
    if(red_error != NULL)X__Free(red_error);
    if(green_error != NULL)X__Free(green_error);
    if(blue_error != NULL)X__Free(blue_error);
    D_red=D_green=D_blue=NULL;
    red_error=green_error=blue_error=NULL;
  }
  dither_on=NO;
}

#ifdef _SUNSTYLE
void DitherLine(scanline,xmax,R,G,B)
  long scanline, xmax; UCHAR *R,*G,*B;{
#else
void DitherLine(long scanline, long xmax, UCHAR *R, UCHAR *G,
                UCHAR *B){
#endif
 long i;
 short new_red, new_green, new_blue,Temp;
 unsigned short Ran;
 if(dither_on == FS_DITHER){
   for (i=0;i<xmax;i++){   /* add distributions from last line */
     new_red   = R[i] + D_red[i];
     new_green = G[i] + D_green[i];
     new_blue  = B[i] + D_blue[i];
     R[i] = (new_red>255) ? 255 : new_red;
     G[i] = (new_green>255) ? 255 : new_green;
     B[i] = (new_blue>255) ? 255 : new_blue;
     D_red[i]=D_green[i]=D_blue[i]=0;
   }
   for(i=0;i<xmax;i++){    /* distribute this line and calc dist.s for next */
     red_error[i] = R[i] %   8;
     green_error[i] = G[i] %  8;
     blue_error[i] = B[i] %  8;
     if (i < xmax-1) {
       new_red   = R[i+1] + (red_error[i] * 7) / 16;
       new_green = G[i+1] + (green_error[i] * 7) / 16;
       new_blue  = B[i+1] + (blue_error[i] * 7) / 16;
       R[i+1] = (new_red>255) ? 255 : new_red;
       G[i+1] = (new_green>255) ? 255 : new_green;
       B[i+1] = (new_blue>255) ? 255 : new_blue;
     }
     if (i != 0) {
       D_red[i-1]   += (red_error[i] * 3) / 16;
       D_green[i-1] += (green_error[i] * 3) / 16;
       D_blue[i-1]  += (blue_error[i] * 3) / 16;
     }
     D_red[i]   += (red_error[i] * 5) / 16;
     D_green[i] += (green_error[i] * 5) / 16;
     D_blue[i]  += (blue_error[i] * 5) / 16;
     if (i < xmax-1) {
       D_red[i+1]   += red_error[i] / 16;
       D_green[i+1] += green_error[i] / 16;
       D_blue[i+1]  += blue_error[i] / 16;
     }
   }
 }
 else if(dither_on == RANDOM_DITHER && dither_factor > 0){
   for(i=0;i<xmax;i++){
     Ran = Random_Dither(dither_factor) - (dither_factor >> 1);
     Temp = (short)R[i] + (short)Ran;
     if (Temp < 0)Temp = 0; else if (Temp > 255)Temp = 255;
     R[i]=(UCHAR)Temp;
     Temp = (short)G[i] + (short)Ran;
     if (Temp < 0)Temp = 0; else if (Temp > 255)Temp = 255;
     G[i]=(UCHAR)Temp;
     Temp = (short)B[i] + (short)Ran;
     if (Temp < 0)Temp = 0; else if (Temp > 255)Temp = 255;
     B[i]=(UCHAR)Temp;
   }
 }
}

#ifdef _SUNSTYLE
void Render_Message(number,query,str) short number,query; char *str;{
#else
void Render_Message(short number,short query, char *str){
#endif
#if __NT__
  char txt[256];
  if(number == 11){
    MessageBox(GetFocus(),(LPCTSTR)"Out of memory, no Phong smoothing",(LPCTSTR) "Phong", MB_OK | MB_SYSTEMMODAL );
  }
#endif
  if(renderer_silent)return;
#if !__NT__
#else
  switch(number){
    case 0:
     LoadString(hThisInstance,IDX_MESSAGE_0,txt,sizeof(txt));
     X_message(txt);
     break;
    case 1:
     LoadString(hThisInstance,IDX_MESSAGE_1,txt,sizeof(txt));
     X_message(txt);
     break;
    case 2:
     LoadString(hThisInstance,IDX_MESSAGE_2,txt,sizeof(txt));
     X_message(txt);
     break;
    case 3:
     LoadString(hThisInstance,IDX_MESSAGE_3,txt,sizeof(txt));
     X_message(txt);
     break;
    case 4:
     LoadString(hThisInstance,IDX_MESSAGE_4,txt,sizeof(txt));
     X_message(txt);
     break;
    case 5:
     LoadString(hThisInstance,IDX_MESSAGE_5,txt,sizeof(txt));
     X_message(txt);
     break;
    case 6:
     LoadString(hThisInstance,IDX_MESSAGE_6,txt,sizeof(txt));
     X_message(txt);
     break;
    case 7:
     LoadString(hThisInstance,IDX_MESSAGE_7,txt,sizeof(txt));
     X_message(txt);
     break;
    case 8:
     LoadString(hThisInstance,IDX_MESSAGE_8,txt,sizeof(txt));
     X_message(txt);
     break;
    case 9:
     LoadString(hThisInstance,IDX_MESSAGE_9,txt,sizeof(txt));
     X_message(txt);
     break;
    case 10:
     LoadString(hThisInstance,IDX_MESSAGE_10,txt,sizeof(txt));
     X_message(txt);
     break;
    case 11:
     LoadString(hThisInstance,IDX_MESSAGE_11,txt,sizeof(txt));
     X_message(txt);
     break;
    case 12:
     LoadString(hThisInstance,IDX_MESSAGE_12,txt,sizeof(txt));
     X_message(txt);
     break;
    case 13:
     LoadString(hThisInstance,IDX_MESSAGE_13,txt,sizeof(txt));
     X_message(txt);
     break;
    case 14:
     LoadString(hThisInstance,IDX_MESSAGE_14,txt,sizeof(txt));
     X_message(txt);
     break;
    case 15:
     LoadString(hThisInstance,IDX_MESSAGE_15,txt,sizeof(txt));
     X_message(txt);
     break;
    case 16:
     LoadString(hThisInstance,IDX_MESSAGE_16,txt,sizeof(txt));
     X_message(txt);
     break;
    case 17:
     LoadString(hThisInstance,IDX_MESSAGE_17,txt,sizeof(txt));
     X_message(txt);
     break;
    case 18:
     LoadString(hThisInstance,IDX_MESSAGE_18,txt,sizeof(txt));
     X_message(txt);
     break;
    case 19:
     LoadString(hThisInstance,IDX_MESSAGE_19,txt,sizeof(txt));
     MessageBox(GetFocus(),txt,NULL,MB_OK);
     break;
    case 20:
     LoadString(hThisInstance,IDX_MESSAGE_20,txt,sizeof(txt));
     X_message(txt);
     break;
    case 21:
     LoadString(hThisInstance,IDX_MESSAGE_21,txt,sizeof(txt));
     X_message(txt);
     break;
    case 22:
     LoadString(hThisInstance,IDX_MESSAGE_22,txt,sizeof(txt));
     X_message(txt);
     break;
    case 23:
     LoadString(hThisInstance,IDX_MESSAGE_23,txt,sizeof(txt));
     X_message(txt);
     break;
    case 24:
     LoadString(hThisInstance,IDX_MESSAGE_24,txt,sizeof(txt));
     X_message(txt);
     break;
    case 25:
     LoadString(hThisInstance,IDX_MESSAGE_25,txt,sizeof(txt));
     X_message(txt);
     break;
    case 26:
     LoadString(hThisInstance,IDX_MESSAGE_26,txt,sizeof(txt));
     X_message(txt);
     break;
    case 27:
     LoadString(hThisInstance,IDX_MESSAGE_27,txt,sizeof(txt));
     X_message(txt);
     break;
    case 28:
     LoadString(hThisInstance,IDX_MESSAGE_28,txt,sizeof(txt));
     X_message(txt);
     break;
    case 29:
     LoadString(hThisInstance,IDX_MESSAGE_29,txt,sizeof(txt));
     X_message(txt);
     break;
    case 30:
     LoadString(hThisInstance,IDX_MESSAGE_30,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 31:
     LoadString(hThisInstance,IDX_MESSAGE_31,txt,sizeof(txt));
     X_message(txt);
     break;
    case 32:
     LoadString(hThisInstance,IDX_MESSAGE_32,txt,sizeof(txt));
     X_message(txt);
     break;
    case 33:
     LoadString(hThisInstance,IDX_MESSAGE_33,txt,sizeof(txt));
     X_message(txt);
     break;
    case 34:
     LoadString(hThisInstance,IDX_MESSAGE_34,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 35:
     LoadString(hThisInstance,IDX_MESSAGE_35,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 36:
     LoadString(hThisInstance,IDX_MESSAGE_36,txt,sizeof(txt));
     X_message(txt);
     break;
    case 37:
     LoadString(hThisInstance,IDX_MESSAGE_37,txt,sizeof(txt));
     X_message(txt);
     break;
    case 38:
     LoadString(hThisInstance,IDX_MESSAGE_38,txt,sizeof(txt));
     X_message(txt);
     break;
    case 39:
     LoadString(hThisInstance,IDX_MESSAGE_39,txt,sizeof(txt));
     X_message(txt);
     break;
    case 40:
     LoadString(hThisInstance,IDX_MESSAGE_40,txt,sizeof(txt));
     X_message(txt);
     break;
    case 41:
     LoadString(hThisInstance,IDX_MESSAGE_41,txt,sizeof(txt));
     X_message(txt);
     break;
    case 42:
     LoadString(hThisInstance,IDX_MESSAGE_42,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 43:
     LoadString(hThisInstance,IDX_MESSAGE_43,txt,sizeof(txt));
     X_message(txt);
     break;
    case 44:
     LoadString(hThisInstance,IDX_MESSAGE_44,txt,sizeof(txt));
     X_message(txt);
     break;
    case 45:
     LoadString(hThisInstance,IDX_MESSAGE_45,txt,sizeof(txt));
     X_message(txt);
     break;
    case 46:
     LoadString(hThisInstance,IDX_MESSAGE_46,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 47:
     LoadString(hThisInstance,IDX_MESSAGE_47,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 48:
     LoadString(hThisInstance,IDX_MESSAGE_48,txt,sizeof(txt));
     X_message(txt);
     break;
    case 49:
     LoadString(hThisInstance,IDX_MESSAGE_49,txt,sizeof(txt));
     X_message(txt);
     break;
    case 50:
     LoadString(hThisInstance,IDX_MESSAGE_50,txt,sizeof(txt));
     X_message(txt);
     break;
    case 51:
     LoadString(hThisInstance,IDX_MESSAGE_51,txt,sizeof(txt));
     X_message(txt);
     break;
    case 52:
     LoadString(hThisInstance,IDX_MESSAGE_52,txt,sizeof(txt));
     X_message(txt);
     break;
    case 53:
     LoadString(hThisInstance,IDX_MESSAGE_53,txt,sizeof(txt));
     X_message(txt);
     break;
    case 54:
     LoadString(hThisInstance,IDX_MESSAGE_54,txt,sizeof(txt));
     X_message(txt);
     break;
    case 55:
     break;
    case 56:
     LoadString(hThisInstance,IDX_MESSAGE_56,txt,sizeof(txt));
     X_message(txt);
     break;
    case 57:
     LoadString(hThisInstance,IDX_MESSAGE_57,txt,sizeof(txt));
     X_message(txt);
     break;
    case 58:
     LoadString(hThisInstance,IDX_MESSAGE_58,txt,sizeof(txt));
     X_message(txt);
     break;
    case 59:
     LoadString(hThisInstance,IDX_MESSAGE_59,txt,sizeof(txt));
     X_message(txt);
     break;
    case 60:
     LoadString(hThisInstance,IDX_MESSAGE_60,txt,sizeof(txt));
     X_message(txt);
     break;
    case 61:
     LoadString(hThisInstance,IDX_MESSAGE_61,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 62:
     LoadString(hThisInstance,IDX_MESSAGE_62,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 63:
     LoadString(hThisInstance,IDX_MESSAGE_63,txt,sizeof(txt));
     X_message(txt);
     break;
    case 64:
     LoadString(hThisInstance,IDX_MESSAGE_64,txt,sizeof(txt));
     X_message(txt);
     break;
    case 65:
     LoadString(hThisInstance,IDX_MESSAGE_65,txt,sizeof(txt));
     X_message(txt);
     break;
    case 66:
     LoadString(hThisInstance,IDX_MESSAGE_66,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 67:
     LoadString(hThisInstance,IDX_MESSAGE_67,txt,sizeof(txt));
     X_message(txt);
     break;
    case 68:
     LoadString(hThisInstance,IDX_MESSAGE_68,txt,sizeof(txt));
     X_message(txt);
     break;
    case 69:
     LoadString(hThisInstance,IDX_MESSAGE_69,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 70:
     LoadString(hThisInstance,IDX_MESSAGE_70,txt,sizeof(txt));
     X_message(txt);
     break;
    case 71:
     LoadString(hThisInstance,IDX_MESSAGE_71,txt,sizeof(txt));
     X_message(txt);
     break;
    case 72:
     LoadString(hThisInstance,IDX_MESSAGE_72,txt,sizeof(txt));
     X_message(txt);
     break;
    case 73:
     LoadString(hThisInstance,IDX_MESSAGE_73,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 74:
     LoadString(hThisInstance,IDX_MESSAGE_74,txt,sizeof(txt));
     X_message(txt);
     break;
    case 75:
     LoadString(hThisInstance,IDX_MESSAGE_75,txt,sizeof(txt));
     X_message(txt);
     break;
    case 76:
     LoadString(hThisInstance,IDX_MESSAGE_76,txt,sizeof(txt));
     X_message(txt);
     break;
    case 77:
     LoadString(hThisInstance,IDX_MESSAGE_77,txt,sizeof(txt));
     X_message(txt);
     break;
    case 78:
     LoadString(hThisInstance,IDX_MESSAGE_78,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 79:
     LoadString(hThisInstance,IDX_MESSAGE_79,txt,sizeof(txt));
     X_message(txt);
     break;
    case 80:
     LoadString(hThisInstance,IDX_MESSAGE_80,txt,sizeof(txt));
     X_message(txt);
     break;
    case 81:
     LoadString(hThisInstance,IDX_MESSAGE_81,txt,sizeof(txt));
     X_message(txt);
     break;
    case 82: /* just a new line */
     break;
    case 83:
     LoadString(hThisInstance,IDX_MESSAGE_83,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 84:
     LoadString(hThisInstance,IDX_MESSAGE_84,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 85:
     LoadString(hThisInstance,IDX_MESSAGE_85,txt,sizeof(txt));
     X_message(txt);
     break;
    case 86:
     LoadString(hThisInstance,IDX_MESSAGE_86,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 87:
     LoadString(hThisInstance,IDX_MESSAGE_87,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 88:
     LoadString(hThisInstance,IDX_MESSAGE_88,txt,sizeof(txt));
     X_message(txt);
     break;
    case 89:
     LoadString(hThisInstance,IDX_MESSAGE_89,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 90:
     LoadString(hThisInstance,IDX_MESSAGE_90,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 91:
     LoadString(hThisInstance,IDX_MESSAGE_91,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 92:
     LoadString(hThisInstance,IDX_MESSAGE_92,txt,sizeof(txt));
     X_message(txt);
     break;
    case 93:
     LoadString(hThisInstance,IDX_MESSAGE_93,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 94:
     LoadString(hThisInstance,IDX_MESSAGE_94,txt,sizeof(txt));
     X_message(txt);
     break;
    case 95:
     LoadString(hThisInstance,IDX_MESSAGE_95,txt,sizeof(txt));
     X_message(txt);
     break;
    case 96:
     LoadString(hThisInstance,IDX_MESSAGE_96,txt,sizeof(txt));
     X_message(txt);
     break;
    case 97:
     LoadString(hThisInstance,IDX_MESSAGE_97,txt,sizeof(txt));
     X_message(txt);
     break;
    case 98:
     LoadString(hThisInstance,IDX_MESSAGE_98,txt,sizeof(txt));
     X_message(txt,str);
     break;
    case 99:
     LoadString(hThisInstance,IDX_MESSAGE_99,txt,sizeof(txt));
     X_message(txt);
     break;
    case 100:
     LoadString(hThisInstance,IDX_MESSAGE_100,txt,sizeof(txt));
     X_message(txt,str);
     break;
    default: break;
  }
#endif
#if !__NT__
  if(number == 91)X_message("\r");
  else            X_message("\n");
  fflush(stdout);
  if(number == 0 || number == 94 || number == 95 || number == 96)getch();
#endif
}

int  X__local_print(const char *format,...){
#if __NT__
 char line[512];
 va_list arg_ptr;
 if(renderer_silent)return 1;
 va_start(arg_ptr,format);
 vsprintf(line,format,arg_ptr);
 va_end(arg_ptr);
 WindowedRenderString(line);
#else
 va_list arg_ptr;
 if(renderer_silent)return 1;
 va_start(arg_ptr,format);
 vprintf(format,arg_ptr);
 va_end(arg_ptr);
#endif
 return 1;
}

typedef struct {
 char *old,*new;
} discarded_file;

static long n_render_discarded;

static discarded_file *discarded_render_file_list;

#ifdef _SUNSTYLE
void setup_render_discarded_list(){
#else
void setup_render_discarded_list(void){
#endif
 n_render_discarded=0; discarded_render_file_list=NULL;
}

#ifdef _SUNSTYLE
void free_render_discarded_list(){
#else
void free_render_discarded_list(void){
#endif
 long i;
 if(n_render_discarded > 0 && discarded_render_file_list != NULL){
   for(i=0;i<n_render_discarded;i++){
     if(discarded_render_file_list[i].old != NULL)
        X__Free(discarded_render_file_list[i].old);
     if(discarded_render_file_list[i].new != NULL)
        X__Free(discarded_render_file_list[i].new);
   }
 }
 n_render_discarded=0; discarded_render_file_list=NULL;
}

char *GetRenderReplacementFilename(char *name){
 //
 // If one of the model or map files has not be found we need to
 // prompt for a replacement. However we don't want to have to do
 // this every time a new frame is renderer, so when a replacement
 // file is selected it is stored in a list. When a new file is needed
 // but can't be found we first check to see if it is in the list before
 // bothering the user.
 //
 long i;
 char rep_name[258];
 discarded_file *temp;
 if(n_render_discarded > 0 && discarded_render_file_list != NULL){
   for(i=0;i<n_render_discarded;i++){
     if(strcmp(discarded_render_file_list[i].old,name) == 0){
        if(discarded_render_file_list[i].new == NULL){
          Render_Message(85,0,NULL);  /* dropping image */
          return NULL;
        }
        sprintf(rep_name,"%s",discarded_render_file_list[i].old);
        Render_Message(86,0,rep_name);  /* replacing */
        sprintf(rep_name,"%s",discarded_render_file_list[i].new);
        Render_Message(87,0,rep_name);  /* with      */
        return discarded_render_file_list[i].new;
     }
   }
 }
 if(n_render_discarded == 0)
   temp=(discarded_file *)X__Malloc(sizeof(discarded_file));
 else
   temp=(discarded_file *)X__Realloc(discarded_render_file_list,
        sizeof(discarded_file)*(n_render_discarded+1));
 if(temp == NULL)return NULL; /* just do nothing - soft landing */
 discarded_render_file_list=temp;
 if((discarded_render_file_list[n_render_discarded].old=(char *)X__Malloc(256))
   == NULL)return NULL;
 strcpy(discarded_render_file_list[n_render_discarded].old,name);
 /* new = NULL => no replacement just dump */
 discarded_render_file_list[n_render_discarded].new=NULL;
 Render_Message(88,0,NULL);
 if(GetRenderNTnewfile(name,rep_name) == 1){
   if(strlen(rep_name) > 4){
     if((discarded_render_file_list[n_render_discarded].new=
         (char *)X__Malloc(256)) == NULL)return NULL;
     strcpy(discarded_render_file_list[n_render_discarded].new,rep_name);
   }
 }
 n_render_discarded++;
 return discarded_render_file_list[n_render_discarded-1].new;
}

FILE *GetRenderRelativeFile(char *fn, char *objectfilename, char *newname){
 FILE *Bit_File=NULL;
 char *fs,*fp,temp_name[256],*evv;
 long l;
 if((Bit_File=fopen((char *)fn,"rb")) == NULL){
   /* try get the expected filename so try to see if it is elsewhere */
   char pref[32],preff[32];

   // Try same drive as renderer
   strcpy(temp_name,fn); temp_name[0]=gszHomeDir[0];
   if((Bit_File=fopen((char *)temp_name,"rb")) != NULL){
     if(newname != NULL)strcpy(newname,temp_name);
     return Bit_File;
   }

   // Try map and/or model paths
   fs=short_form(fn);
   if(objectfilename == NULL)strcpy(preff,"MODEL-PATH"); // Must be map on Model
   else{
     strcpy(preff,"MAP-PATH");
     if(strcmp(objectfilename,"dummy"))objectfilename=NULL; // NO OBJECT sky/ground
   }
   for(l=1;l<=128;l++){ /*  try .INI settings for models and maps*/
     sprintf(pref,"%s%ld",preff,l);
     GetPrivateProfileString("RPREFERENCES",pref,"$$$",
                             temp_name,255,"OPENFX_.INI");
     if(strncmp(temp_name,"$$$",3) == 0)break;
     if(temp_name[1] != ':')continue;
     if(temp_name[strlen(temp_name)-1] != '\\')strcat(temp_name,"\\");
     strcat(temp_name,fs);
     if((Bit_File=fopen((char *)temp_name,"rb")) != NULL){
       if(newname != NULL)strcpy(newname,temp_name);
       return Bit_File;
     }
   }
   fs=short_form(fn);
   if(objectfilename != NULL && strcmp(objectfilename,"dummy"))objectfilename=NULL;
   if(objectfilename != NULL){  /* else try to find in the same a model */
     strcpy(temp_name,objectfilename);
     fp=short_form(temp_name);
     strcpy(fp,fs); fs=temp_name;
   }
   if((Bit_File=fopen(fs,"rb")) == NULL){
     Render_Message(90,0,fs);
   }
   else{ //OK
     if(newname != NULL)strcpy(newname,fs);
   }
 }
 return Bit_File;
}

HANDLE GetRenderRelativeHandle(char *fn, char *objectfilename, char *newname){
 HANDLE Bit_File=INVALID_HANDLE_VALUE;
 char *fs,*fp,temp_name[256],*evv;
 long l;
 Bit_File=CreateFile(fn,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(Bit_File == INVALID_HANDLE_VALUE){ 
   /* try get the expected filename so try to see if it is elsewhere */
   char pref[32],preff[32];

   // Try same drive as renderer
   strcpy(temp_name,fn); temp_name[0]=gszHomeDir[0];
   Bit_File=CreateFile(temp_name,GENERIC_READ,
          FILE_SHARE_READ,NULL,OPEN_EXISTING,
          FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
   if(Bit_File != INVALID_HANDLE_VALUE){
     if(newname != NULL)strcpy(newname,temp_name);
     return Bit_File;
   }

   //Try MAP and/or MODEL path
   fs=short_form(fn);
   if(objectfilename == NULL)strcpy(preff,"MODEL-PATH"); // Must be map on Model
   else{
     strcpy(preff,"MAP-PATH");
     if(strcmp(objectfilename,"dummy"))objectfilename=NULL; // NO OBJECT sky/ground
   }
   for(l=1;l<=128;l++){ /*  try .INI settings for models and maps*/
     sprintf(pref,"%s%ld",preff,l);
     GetPrivateProfileString("RPREFERENCES",pref,"$$$",
                             temp_name,255,"OPENFX_.INI");
     if(strncmp(temp_name,"$$$",3) == 0)break;
     if(temp_name[1] != ':')continue;
     if(temp_name[strlen(temp_name)-1] != '\\')strcat(temp_name,"\\");
     strcat(temp_name,fs);
     Bit_File=CreateFile(temp_name,GENERIC_READ,
          FILE_SHARE_READ,NULL,OPEN_EXISTING,
          FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
     if(Bit_File != INVALID_HANDLE_VALUE){
       if(newname != NULL)strcpy(newname,temp_name);
       return Bit_File;
     }
   }

   fs=short_form(fn);
   if(objectfilename != NULL && strcmp(objectfilename,"dummy"))objectfilename=NULL;
   if(objectfilename != NULL){  /* else try to find in the same a model */
     strcpy(temp_name,objectfilename);
     fp=short_form(temp_name);
     strcpy(fp,fs); fs=temp_name;
   }
   Bit_File=CreateFile(fs,GENERIC_READ,
          FILE_SHARE_READ,NULL,OPEN_EXISTING,
          FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
   if(Bit_File == INVALID_HANDLE_VALUE){
     Render_Message(90,0,fs);
   }
   else{ // OK
     if(newname != NULL)strcpy(newname,fs);
   }
 }
 return Bit_File;
}
