/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file BUILDFC.C routines to encode FLC files  */

long Encode_FLC_colour(unsigned char *c){
 long i;
 *c++ = 1; *c++ = 0; *c++ = 0; *c++ = 0;
 for(i=0 ; i< AA_COLOURS; i++){
  *c++ = gif_p_red[i];
  *c++ = gif_p_green[i];
  *c++ = gif_p_blue[i];
 }
 return (AA_COLOURS * 3 + 4L);
}


static short Com_FLC_RunLine(unsigned char *input,
                             char *buffer,short x_res){
 long i;
 char *pack,*size,*data,run,length;
 short bufsize;
 pack=buffer; *pack =  1;     buffer++;
 size=buffer; *size =  1;     buffer++;
 data=buffer; *data = *input; buffer++; input++;
 bufsize = 3; run=1; length=0;
 for(i=1;i<x_res;i++){
   if( *input == *(input-1) ){ /* current byte is same as last byte */
     if(length == 1){   /* one byte of length change to run of 2 */
       *size=2; run=2; length=0;
     }
     else if(length > 1 || run == 127){ /* new packet */
       (*pack)++;
       size = buffer; *size = 1;      buffer++;
       data = buffer; *data = *input; buffer++;
       length = 0; bufsize += 2; run=1;
     }
     else{   /* increment count */
       run++; *size=run;
     }
   }
   else{
     if(run == 1){ /* one byte of run, change to length */
       *size = -2; length=2;
       data = buffer; *data = *input; buffer++;
       run=0; bufsize++;
     }
     else if(run > 1 || length == 127){  /* long run/length>27 new packet */
       (*pack)++;
       size = buffer; *size = -1;     buffer++;
       data = buffer; *data = *input; buffer++;
       length = 1; bufsize += 2; run=0;
     }
     else{ /* add to original packet */
       length++; *size = -length;
       data=buffer; *data = *input; buffer++; bufsize++;
     }
   }
   input++;
 }
 return bufsize;
}


static long Compress_SS2_Line(short *P, short *Q, unsigned char *B,
                              short xw, short *npacks)
{
 long i,j,ndif,bufsize,sames,lastbuf;
 short *p,*q;
 unsigned char *b,*szp,*ppk,skip,size,db[2048];
 ndif=0; p=P; q=Q;
 for(i=0;i<xw;i++){
   if(*p != *q ){db[i]=1; ndif++;}
   else          db[i]=0;
   p++; q++;
 }
 if(ndif == 0)return 0;
 *npacks=0;
 bufsize=0;
 b = B;
 i=0; p=P;skip=0;
 while(i < xw){
   if(db[i] == 1){
     if((j=i+1) < xw && db[j] == 1){
      (*npacks)++;
      *b++ = skip; skip=0;
      szp=b++;
      *((short *)b) = *p++; b += 2;
      bufsize +=4; lastbuf=bufsize;
      size=1;
      sames=1;
      while(db[j] == 1 && j < xw && size < 127){
       if(*p == *(p-1)){         /* count the number of repeats */
         sames++;
       }
       else{                     /* new pattern */
         if(sames > 4){          /* make new packet if worth while  */
           if(sames < size){     /* need extra pack for non repeats */
             *szp = size-sames;
             szp=(szp+(size-sames)*2+2);
             *(szp-1)=0;
             lastbuf=bufsize-sames*2+4;
            (*npacks)++;
           }
           *szp = (unsigned char)((char)(sames * -1)); /* fix up old pack */
           *(szp+3)=0;           /* no skips in this packet */
           szp=(szp+4);          /* new position for size count */
           bufsize = lastbuf+2;  /* inc to 3 immediately */
           lastbuf = bufsize+2;
           b=(szp+1);            /* incremented  immediately */
           (*npacks)++;
           size=0;               /* inc to 1 immediately */
         }
         sames=1;
       }
       size++;
       j++;
       *((short *)b) = *p++; b += 2;
       bufsize += 2;
      }
      if(sames == size){ /* this can be done with a copy */
        *szp = (unsigned char)( (char)(sames * -1) );
        bufsize=lastbuf; /* discard the multiple copies */
        b=(szp+3);       /* just modify the last packet */
      }
      else
        *szp=size;       /* update the size count */
      i=j;
     }                   /* end of while */
     else{
      (*npacks)++;
      *b++ = skip;     /* number of skips */
      *b++ = 1;
      *((short *)b) = *p++; b += 2;
      i++; skip=0;
      bufsize += 4;
     }
   }
   else{
     if(skip == 254){
       (*npacks)++;
       *b++ = skip;
       *b++ = 1;
       *((short *)b) = *p++; b += 2;
       i++; skip=0;
       bufsize += 4;
     }
     else{
       skip +=2; i++; p++;
     }
   }
 }
 return(bufsize);
}

static long Compress_SS2_Frame(unsigned char *S1,
                               unsigned char *S2,
                               unsigned char *c,
                               short x, short y)
{ short *count,skip=0,packs;
  long i,size,framesize;
  framesize=2;
  count = (short *)c; *count= 0; c += 2;
  for(i=0;i<y;i++){
    if((size=Compress_SS2_Line((short *)S1,
                               (short *)S2,
                               linebuffer,(short)(x/2),&packs)) == 0){
       skip++;
    }
    else{
      (*count)++;
      if(skip > 0){
        *((short *)c) = -skip;
        skip=0;
        framesize +=2; c+=2;
      }
      *((short *)c)=packs; c+=2; framesize += 2;
      memcpy(c,linebuffer,size);
      c += size;
      framesize += size;
    }
    S1 += x;
    S2 += x;
  }
  if(*count == 0)return 0;
  else return  framesize;
}

static long Encode_FLC_FirstFrame(unsigned char *S1,
                                  unsigned char *buffer,
                                  short x, short y,
                                  long max_size){
 unsigned char *c,*c1,*local_screen,*fcc;
 long i,nsize;
 short size;
 local_screen=S1;
 ffflc=(Flc_frame *)buffer;
 ffflc->size=16;
 ffflc->type=FLCF_MAGIC;
 ffflc->chunks=0;
 fcflc=(Flc_chunk *)(ffflc+1);
 fcflc->size=6;
 fcflc->type=FLC_COLOR;
 c=(unsigned char *)(fcflc)+6;
 fcflc->size += Encode_FLC_colour(c);
 ffflc->size += fcflc->size;
 ffflc->chunks++;
 c+=772;    /* size of uncompressed colour map */
 fcc=c; Clong(fcc,6);
 c+=6;
 nsize=0; c1=c;
 for(i=0;i<y;i++){
   size = Com_FLC_RunLine(S1,(char *)linebuffer,x);
   nsize += (long)size;
   if(nsize > max_size)break;
   memcpy(c1,linebuffer,size);
   c1 += size;
   S1 += x;
 }
 if(nsize > max_size){
   Along(fcc,max_size);
   Cshort(fcc+4,FLC_COPY);
   ffflc->size += Tlong(fcc);
   memcpy(c,local_screen,max_size);
   ffflc->chunks++;
 }
 else if(nsize > 0){  /* should be else */
   Along(fcc,nsize);
   Cshort(fcc+4,FLC_BRUN);
   ffflc->size += Tlong(fcc);
   ffflc->chunks++;
 }
 fhflc->frames++;
 fhflc->size += ffflc->size;
 return( (long) ffflc->size );
}

short  Write_first_FLC_frame(FILE *flc_file,
                             unsigned char *S1,
                             unsigned char *S2,
                             unsigned char *buffer,
                             short ispeed,
                             short x, short y)
{long i,k,size;
 size = (long)x*(long)y;
 fhflc = &FLC_Fhead;
 memset(fhflc,0,sizeof(Flc_head));
 fhflc->size    = 128;
 fhflc->type    = FLCH_MAGIC;
 fhflc->frames  = 0;
 fhflc->width   = x;
 fhflc->height  = y;
 fhflc->depth   = 8;
 fhflc->flags   = 0;
 fhflc->speed   = (ispeed*100)/7;
 fhflc->aapro_serial  = 0x464c4942;
 fhflc->aspectx = 1;
 fhflc->aspecty = 1;
 fhflc->oframe1 = 0;
 fhflc->oframe2 = 0;
 if(fwrite(fhflc,1,128,flc_file) != 128){
   return -1;
 }
 size=Encode_FLC_FirstFrame(S1,buffer,x,y,(long)x*(long)y);
 fhflc->oframe1=ftell(flc_file);
//{char tempstr[128]; sprintf(tempstr,"first frame size %ld",size);
//MessageBox(GetFocus(),tempstr,"debug",MB_OK);}
 if(fwrite(buffer,sizeof(unsigned char),size,flc_file) != size){
   return -1;
 }
 fhflc->oframe2=ftell(flc_file);
 memcpy(S2,S1,(long)x * (long)y);
 return 0;
}

short Write_FLC_frame(FILE *flc_file,
                      unsigned char *S1,
                      unsigned char *S2,
                      unsigned char *buffer,
                      short x, short y)
{unsigned char *c;
 long nsize,size,max_size;
 max_size=(long)x * (long)y;
 ffflc=(Flc_frame *)buffer;
 ffflc->size=sizeof(Flc_frame);
 ffflc->type=FLCF_MAGIC;
 ffflc->chunks=0;
 fcflc=(Flc_chunk *)(ffflc+1);
 fcflc->size=sizeof(Flc_chunk);
 fcflc->type=FLC_SS2;
 c=(unsigned char *)(fcflc)+6;
 if((nsize=(long)Compress_SS2_Frame(S1,S2,c,x,y)) > max_size){
   memcpy(c,S1,max_size);
   fcflc->size += max_size;
   fcflc->type = FLC_COPY;
   ffflc->size += fcflc->size;
   ffflc->chunks++;
 }
 else if(nsize > 0){
   fcflc->size += nsize;
   ffflc->size += fcflc->size;
   ffflc->chunks++;
 }
 fhflc->frames++;
 fhflc->size += ffflc->size;
 size = ffflc->size;
//{char tempstr[128]; sprintf(tempstr,"frame size %ld",size);
//MessageBox(GetFocus(),tempstr,"debug",MB_OK);}
 if(fwrite(buffer,sizeof(unsigned char),size,flc_file) != size)return -1;
 memcpy(S2,S1,max_size);
 return 0;
}

short Close_FLC_file(FILE *flc_file)
{long lastpos;
 fhflc->frames--;
 if (fseek (flc_file, 0L, SEEK_SET) < 0)return -1;
 fhflc->flags = (FLC_FINISHED | FLC_LOOPED);
 if (fwrite(fhflc,1,128,flc_file) != 128)return -2;
 return 0;
}

