/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  BUILDFI.C  routines to encode FLI files  */

static long EncodeColourMap(unsigned char *c){
 long i;
 *c++ = 1; *c++ = 0; *c++ = 0; *c++ = 0;
 for(i=0 ; i< AA_COLOURS ; i++){
   *c++ = gif_p_red[i];
   *c++ = gif_p_green[i];
   *c++ = gif_p_blue[i];
 }
//MessageBox(GetFocus(),"Colour Encoded","debug",MB_OK);
 return ((long)AA_COLOURS*3 + 4L);
}


static short CompressRunLine(unsigned char *input,char *buffer){
 long i;
 char *pack,*size,*data,run,length;
 short bufsize;
 pack=buffer; *pack =  1;     buffer++;
 size=buffer; *size =  1;     buffer++;
 data=buffer; *data = *input; buffer++; input++;
 bufsize = 3; run=1; length=0;
 for(i=1;i<320;i++){
   if( *input == *(input-1) ){ /* current byte is same as last byte */
     if(length == 1){   /* one byte of length change to run of 2 */
       *size=2; run=2; length=0;
     }
     else if(length > 1 || run == 127){ /* new packet */
       (*pack)++;
       size = buffer; *size = 1;      buffer++;
       data = buffer; *data = *input; buffer++;
       length = 0; bufsize += 2; run=1;
     }
     else{   /* increment count */
       run++; *size=run;
     }
   }
   else{
     if(run == 1){ /* one byte of run, change to length */
       *size = -2; length=2;
       data = buffer; *data = *input; buffer++;
       run=0; bufsize++;
     }
     else if(run > 1 || length == 127){  /* long run/length>27 new packet */
       (*pack)++;
       size = buffer; *size = -1;     buffer++;
       data = buffer; *data = *input; buffer++;
       length = 1; bufsize += 2; run=0;
     }
     else{ /* add to original packet */
       length++; *size = -length;
       data=buffer; *data = *input; buffer++; bufsize++;
     }
   }
   input++;
 }
 return bufsize;
}

static unsigned short RunEncodeFirst(unsigned char *S1,
                                     unsigned char *c){
  int j;
  unsigned short i,framesize;
  short size;
  framesize=size=0;
  for(i=0;i<200;i++){
    size = CompressRunLine(S1,linebuffer2);
    framesize += size;
    if(framesize > 60000)return (unsigned short)64000;
    if(size > 0)for(j=0;j<size;j++)*c++ = linebuffer2[j];
    S1 += 320;
  }
  return framesize;
}

static short CompressLine(unsigned char *P,
                          unsigned char *Q,
                          unsigned char *B){
 short i,j,ndif,bufsize,sames,lastbuf;
 unsigned char *p,*q,*b,*szp,*ppk,skip,size,npacks,db[320];
 ndif=0; p=P; q=Q;
 for(i=0;i<320;i++){
   if(*p != *q ){db[i]=1; ndif++;}
   else          db[i]=0;
   p++; q++;
 }
 if(ndif == 0)return 0;
 npacks=0;
 bufsize=1;
 ppk=B;
 b = (B+1);
 i=0; p=P;skip=0;
 while(i < 320){
   if(db[i] == 1){
     if((j=i+1) < 320 && db[j] == 1){
      npacks++;
      *b++ = skip; skip=0;
      szp=b++;
      *b++ = *p++;
      bufsize +=3; lastbuf=bufsize;
      size=1;
      sames=1;
      while(db[j] == 1 && j < 320 && size < 127){
       if(*p == *(p-1)){         /* count the number of repeats */
         sames++;
       }
       else{                     /* new pattern */
         if(sames > 4){          /* make new packet if worth while  */
           if(sames < size){     /* need extra pack for non repeats */
             *szp = size-sames;
             szp=(szp+size-sames+2);
             *(szp-1)=0;
             lastbuf=bufsize-sames+3;
             npacks++;
           }
           *szp = (unsigned char)((char)(sames * -1)); /* fix up old pack */
           *(szp+2)=0;           /* no skips in this packet */
           szp=(szp+3);          /* new position for size count */
           bufsize = lastbuf+2;  /* inc to 3 immediately */
           lastbuf = bufsize+1;
           b=(szp+1);            /* incremented  immediately */
           npacks++;
           size=0;               /* inc to 1 immediately */
         }
         sames=1;
       }
       size++;
       j++;
       *b++ = *p++;
       bufsize++;
      }
      if(sames == size){ /* this can be done with a copy */
        *szp = (unsigned char)( (char)(sames * -1) );
        bufsize=lastbuf; /* discard the multiple copies */
        b=(szp+2);       /* just modify the last packet */
      }
      else
        *szp=size;       /* update the size count */
      i=j;
     }
     else{
      npacks++;
      *b++ = skip;     /* number of skips */
      *b++ = 1;
      *b++ = *p++;
      i++; skip=0;
      bufsize += 3;
      if(npacks > 254)return(3201);
     }
   }
   else{
     if(skip == 255){
       npacks++;
       *b++ = skip;
       *b++ = 1;
       *b++ = *p++;
       i++; skip=0;
       bufsize += 3;
       if(npacks > 254)return(3201);
     }
     else{
       skip++; i++; p++;
     }
   }
 }
 *ppk = npacks;
 return(bufsize);
}

static unsigned short CompressFrame(unsigned char *S1,
                                    unsigned char *S2,
                                    unsigned char *c){
  long j;
  short *sames,*copies,same,copy,size,i;
  unsigned short framesize;
  framesize=4;
  sames  = (short *)c; *c++ = 0; *c++=0;
  copies = (short *)c; *c++ = 0; *c++=0;
//MessageBox(GetFocus(),"Compress Frame",NULL,MB_OK);
  for(i=0;i<200;i++){
    if((size=CompressLine(S1,S2,linebuffer)) == 0 && *copies == 0){
      (*sames)++;
    }
    else{
      if(size > 3200)return (unsigned short)64000;
      else if(size == 0){
        framesize++;
        *c++ = 0;
      }
      else{
        for(j=0;j<size;j++)*c++ = linebuffer[j];
        framesize += size;
        *copies = i+1;
      }
    }
    S1 += 320;
    S2 += 320;
    if(framesize > 60000)return (unsigned short)64000;
  }
  if(*sames == 200)return 0;
  framesize -= (200 - *copies);
  *copies -= *sames;
  return  framesize;
}

static unsigned short EncodeFirstFrame(unsigned char *S1,
                                       unsigned char *buffer)
{unsigned char *c,*fcc;
 unsigned short nsize;
 long i;
 ff=(Fli_frame *)buffer;
 ff->size=16;
 ff->type=FLIF_MAGIC;
 ff->chunks=0;
 fc=(Fli_chunk *)(ff+1);
 fc->size=6;
 fc->type=FLI_COLOR;
 c=(unsigned char *)(fc)+6;
 fc->size += EncodeColourMap(c);
 ff->size += fc->size;
 ff->chunks++;
 c+=772;    /* size of uncompressed colour map       */
 fcc=c; Clong(fcc,6);  /* 6 bytes for start of chunk */
 c+=6;
 if((nsize=RunEncodeFirst(S1,c)) > 60000){
   Along(fcc,64000L);
   Cshort(fcc+4,FLI_COPY);
   ff->size += Tlong(fcc);
   for(i=0;i<64000;i++)*c++ = *S1++;
   ff->chunks++;
 }
 else if(nsize > 0){
   Along(fcc,nsize);
   Cshort(fcc+4,FLI_BRUN);
   ff->size += Tlong(fcc);
   ff->chunks++;
 }
 fh->frame_count++;
 fh->size += ff->size;
 return( (unsigned short) ff->size );
}

static unsigned short EncodeFrame(unsigned char *S1,
                                  unsigned char *S2,
                                  unsigned char *buffer)
{unsigned char *c;
 unsigned short nsize;
 int j;
 ff=(Fli_frame *)buffer;
 ff->size=sizeof(Fli_frame);
 ff->type=FLIF_MAGIC;
 ff->chunks=0;
 fc=(Fli_chunk *)(ff+1);
 fc->size=sizeof(Fli_chunk);
 fc->type=FLI_LC;
 c=(unsigned char *)(fc)+6;
 if((nsize=CompressFrame(S1,S2,c)) > 60000){
   for(j=0;j<64000;j++)*c++ = *S1++;
   fc->size += 64000L;
   fc->type = FLI_COPY;
   ff->size += fc->size;
   ff->chunks++;
 }
 else if(nsize > 0){
   fc->size += nsize;
   ff->size += fc->size;
   ff->chunks++;
 }
 fh->frame_count++;
 fh->size += ff->size;
 return( (unsigned short) ff->size );
}

static void BackUpScreen(unsigned char *S1,
                         unsigned char *S2){
 memcpy(S2,S1,(size_t)64000);
//MessageBox(GetFocus(),"Copied frame","debug",MB_OK);
}

static short Write_first_FLI_frame(FILE *opf,short ispeed,
                                   unsigned char *S1,
                                   unsigned char *S2,
                                   unsigned char *buffer){
 long lsize;
 unsigned short size;
 int i,k;
 fh = &Fhead;
 memset(fh,0,sizeof(Fli_head));
 fh->type = FLIH_MAGIC;
 fh->size = 128;
 fh->width = 320;
 fh->height = 200;
 fh->bits_a_pixel = 8;
 fh->speed = ispeed;
 fh->frame_count = 0;
 if(fwrite(fh,1,128,opf) != 128)return -1;
 size=EncodeFirstFrame(S1,buffer);
 lsize=(long)size;
//{
//char xxx[128]; sprintf(xxx,"first frame size %d",lsize);
//MessageBox(GetFocus(),xxx,"debug",MB_OK);
//}
 if(fwrite(buffer,1,lsize,opf) != lsize)return -1;
 BackUpScreen(S1,S2);
 return 0;
}

static short Write_FLI_frame(FILE *opf,unsigned char *S1,
                                       unsigned char *S2,
                                       unsigned char *buffer)
{unsigned short size;
 size=EncodeFrame(S1,S2,buffer);
//{
//char xxx[128]; sprintf(xxx,"frame size %d",size);
//MessageBox(GetFocus(),xxx,"debug",MB_OK);
//}
 if(fwrite(buffer,1,size,opf) != size)return -1;
 BackUpScreen(S1,S2);
 return 0;
}

static short Close_FLI_file(FILE *opf){
 long lastpos;
 fh->frame_count--;
 lastpos = ftell(opf);
 if (fseek (opf, 0L, SEEK_SET) < 0){
   return -1;
 }
 fh->flags = (FLI_FINISHED | FLI_LOOPED);
 if (fwrite(fh,1,128,opf) != 128){
    return -1;
 }
 return 1;
}
