/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* FILE FLIC_DLL.C    DLL to build FLI/FLC files from TEMPORARY files */

/* NOTE all DLLS using the new Microsoft compiler must have CALLBACK  */


#include <stdlib.h>
#include <stdio.h>
#include <windows.h>

#ifdef __X__NT__
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#else
#define X__Malloc(s)    GlobalAlloc(GMEM_FIXED,s)
#define X__Free(p)      GlobalFree((HGLOBAL)p)
#define X__Realloc(p,s) GlobalReAlloc((HGLOBAL)p,s,GMEM_MOVEABLE)
#endif
#if !__WATCOMC__
#define fflush(a) 0
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

#include "flihead.h"

static Fli_head  Fhead;
static struct fli_head *fh;
static struct fli_frame *ff;
static struct fli_chunk *fc;

#include "flchead.h"

static Flc_head          FLC_Fhead;
static struct flc_head  *fhflc;
static struct flc_frame *ffflc;
static struct flc_chunk *fcflc;

static short compressbuffer[1024];
static char     linebuffer2[1024];
static unsigned char linebuffer[2048];
static unsigned char gif_p_red[256],gif_p_green[256],gif_p_blue[256];

#define GOOD 2
#define GIFCOLOURS 256

static unsigned long   *colorstats0;
static unsigned short  *colorstats1;

extern long getcolor(unsigned short crgb, long Npalette);
extern void SortColours(long Ngif, long N_colours);
extern void UnCompressWordLine(short bufsize, short *buffer,
                               unsigned short *output);

static void Cshort(unsigned char *c, short v){
  unsigned char ch,cl;
  ch = (unsigned char)(v >> 8);
  cl = (unsigned char)(v & 0x00ff);
  *c++ = cl;
  *c++ = ch;
}

static void Clong(unsigned char *c, long v){
  unsigned char ch,c2,c1,cl;
  ch = (unsigned char)((v >> 24) & 0x000000ff);
  c2 = (unsigned char)((v >> 16) & 0x000000ff);
  c1 = (unsigned char)((v >>  8) & 0x000000ff);
  cl = (unsigned char)(v & 0x000000ff);
  *c++ = cl;
  *c++ = c1;
  *c++ = c2;
  *c++ = ch;
}

static void Ashort(unsigned char *c,short v){
  unsigned char ch,cl;
  short ah,al,a;
  al = (short)(*c);
  ah = (short)(*(c+1));
  a  = (ah << 8) | al;
  a += v;
  ch = (unsigned char)(a >> 8);
  cl = (unsigned char)(a & 0x00ff);
  *c     = cl;
  *(c+1) = ch;
}

static void Along(unsigned char *c,long v){
  unsigned char ch,c2,c1,cl;
  long ah,a2,a1,al,a;
  al = (long)(*c);
  a1 = (long)(*(c+1));
  a2 = (long)(*(c+2));
  ah = (long)(*(c+3));
  a  = (ah << 24) | (a2 << 16) | (a1 << 8) | al;
  a += v;
  ch = (unsigned char)((a >> 24) & 0x000000ff);
  c2 = (unsigned char)((a >> 16) & 0x000000ff);
  c1 = (unsigned char)((a >>  8) & 0x000000ff);
  cl = (unsigned char)(a & 0x000000ff);
  *c     = cl;
  *(c+1) = c1;
  *(c+2) = c2;
  *(c+3) = ch;
}

static long Tlong(unsigned char *c){
  long ah,a2,a1,al,a;
  al = (long)(*c);
  a1 = (long)(*(c+1));
  a2 = (long)(*(c+2));
  ah = (long)(*(c+3));
  a  = (ah << 24) | (a2 << 16) | (a1 << 8) | al;
  return a;
}

static short Tshort(unsigned char *c){
  short ah,al,a;
  al = (short)(*c);
  ah = (short)(*(c+1));
  a  = (ah << 8) | al;
  return a;
}

#include "buildfi.c"
#include "buildfc.c"

// read and decompress code

#ifdef _SUNSTYLE
void  UnCompressWordLine(bufsize,buffer,output)
short bufsize;short *buffer; unsigned short *output;{
#else
void  UnCompressWordLine(short bufsize, short *buffer,unsigned short *output){
#endif
 short p,k;
 long i;
 k=0; while(k < bufsize){
   p = *buffer++; k++;
   if(p < 0)
     for(i=0; i<= -p; i++){*output++ = *buffer++; k++;}
   else{
     for(i=0; i<=  p; i++) *output++ = *buffer;
     buffer++; k++;
   }
 }
}

#ifdef _SUNSTYLE
static long distance(argb,brgb) unsigned short argb; unsigned short brgb; {
#else
static long distance(unsigned short argb, unsigned short brgb){
#endif
    long b,g,r;
    b  = (long)( argb        & 0x1f);
    g  = (long)((argb >>  5) & 0x1f);
    r  = (long)((argb >> 10) & 0x1f);
    b -= (long)( brgb        & 0x1f);
    g -= (long)((brgb >>  5) & 0x1f);
    r -= (long)((brgb >> 10) & 0x1f);
    return (r*r + g*g + b*b);
/*    return labs(r) + labs(g) + labs(b);
*/
}

#ifdef _SUNSTYLE
long getcolor(crgb, Npalette) unsigned short crgb; long Npalette; {
#else
long getcolor(unsigned short crgb, long Npalette){
#endif
  long i,j,best,dist;
/* look for an exact match in the color table (or minimal distance) */
    for (i=0; i < Npalette; i++) {
       if (colorstats1[i] == crgb) return (i);
       if (i == 0) {
         best = 0;
         dist = distance(colorstats1[i],crgb);
       }
       else if ((j=distance(colorstats1[i],crgb)) < dist) {
         best = i;
         dist = j;
       }
    }
    return(best);
}

#ifdef _SUNSTYLE
void SortColours(Ngif,N_colours) long Ngif; long N_colours; {
#else
void SortColours(long Ngif, long N_colours){
#endif
 long i,j,gap,t0,maxdis,used;
 unsigned short t1;
 for (i = 0; i < N_colours; i++) {
     colorstats1[i] = i;
 }
 for (gap = N_colours/2; gap > 0; gap /=2) {
  for (i = gap; i < N_colours; i++) {
    for (j = i-gap; j >= 0 &&
        colorstats0[j] < colorstats0[j+gap]; j -= gap) {
      t0 = colorstats0[j];
      t1 = colorstats1[j];
      colorstats0[j] = colorstats0[j+gap];
      colorstats1[j] = colorstats1[j+gap];
      colorstats0[j+gap] = t0;
      colorstats1[j+gap] = t1;
    }
  }
 }
 for (used = 0; used < N_colours && colorstats0[used] > 0; used++){;}
 if (used > Ngif) {
  for (maxdis = 2; maxdis < 1024; maxdis *= 2) {
   for (i=Ngif/2; i < Ngif; i++) {
    for (j = 0; j < i; j++) {
      if (distance(colorstats1[i],colorstats1[j])<maxdis) {
       for (j=i+1; j< used; j++) {
         colorstats0[j-1] = colorstats0[j];
         colorstats1[j-1] = colorstats1[j];
       }
       --used;
       --i;
       break;
      }
    }
    if (used <= Ngif) break;
   }
   if (used <= Ngif) break;
  }
 }
 return;
}

// end of read and uncompress code

#define FLI 1
#define FLC 2

/* range_l is last included frame+1 */

#if __SC__
//void CALLBACK BuildFli
#else
void _BuildFli
#endif
(  HWND parent_window, HWND status_window,
   int range_f, int range_l, char *flic_file,
   char *stem_file,
   int ResolutionX,int ResolutionY,
   int nCOLOUR, long speed, long tzeros,
   unsigned long *cstats0,unsigned short *cstats1){
 char filename[128],speed_string[128],*pp,format[64];
 long used,cpix,size,type,x_x,y_y;
 unsigned short val,Nt,RGB[1024];
 int i,j,frame,frames_found;
 unsigned char *Screen1,*Screen2,*screen,*fli_buffer;
 FILE *ipf, *opf;
 colorstats0=cstats0;
 colorstats1=cstats1;
// {
//  char str[128];
//  sprintf(str,"Stem=[%s]",stem_file);
//  MessageBox(GetFocus(),(LPCTSTR)str,(LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
//  sprintf(str,"flic=[%s] ff=%ld lf=%ld",flic_file,range_f,range_l);
//  MessageBox(GetFocus(),(LPCTSTR)str,(LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
//  sprintf(str,"x=%ld y=%ld speed=%ld",ResolutionX,ResolutionY,speed);
//  MessageBox(GetFocus(),(LPCTSTR)str,(LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
// }
 if((range_l - range_f) < 3)return;  /* don't animate < 3 frames */
 if(ResolutionX > 640 || ResolutionY > 480)return;   /* don't animate > 640 by 480 */
 if(ResolutionX > 320 || ResolutionY > 200){
   type=FLC;
   size=ResolutionX*ResolutionY;
 }
 else{
   type=FLI;
   size=64000;
 }
 if(speed < 0)speed=5;
 if((Screen1 = (unsigned char *)X__Malloc(size)) == NULL){
   MessageBox(GetFocus(),(LPCTSTR)"Out of memory",(LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
   return;
 }
 memset(Screen1,'0',size);
 if((Screen2 = (unsigned char *)X__Malloc(size)) == NULL){
   MessageBox(GetFocus(),(LPCTSTR)"Out of memory",(LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
   X__Free(Screen1);
   return;
 }
 memset(Screen2,'0',size);
 size += 10L*(long)ResolutionX;
 if((fli_buffer=(unsigned char *)X__Malloc(size)) == NULL){
   MessageBox(GetFocus(),(LPCTSTR)"Out of memory",(LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
   X__Free(Screen1); X__Free(Screen2);
   return;
 }
 SendMessage(status_window,WM_SETTEXT,0,(LPARAM)"Building color palette");
 SortColours(GIFCOLOURS,nCOLOUR);
 /* note << 1 is because the FLI/FLC is using 64**3 colours */
 for(i=0;i<GIFCOLOURS;i++){
   gif_p_red[i]  =(unsigned char)((colorstats1[i] >> 10) & 0x1F) << 1;
   gif_p_green[i]=(unsigned char)((colorstats1[i] >>  5) & 0x1F) << 1;
   gif_p_blue[i] =(unsigned char)((colorstats1[i]      ) & 0x1F) << 1;
 }
 for(i=0;i<nCOLOUR;i++)colorstats0[i] = GIFCOLOURS+1;
 if((opf=fopen(strlwr(flic_file),"wb")) == NULL)goto FATALOPEN2;
 for(frame=range_f;frame<=range_l;frame++){
   sprintf(speed_string,"Building frame %03ld",min(frame,range_l));
   SendMessage(status_window,WM_SETTEXT,0,(LPARAM)speed_string);
   sprintf(format,"%s%ld%s","%s%0",tzeros,"d.16b");
   if(frame < range_l)sprintf(filename,format,stem_file,frame);
   else               sprintf(filename,format,stem_file,range_f);
   if((ipf=fopen(strlwr(filename),"rb")) == NULL){
     goto FATALOPEN1;
   }
   if( fread((char *)&x_x,sizeof(long),1,ipf) != 1 ||
       fread((char *)&y_y,sizeof(long),1,ipf) != 1)goto FATALREAD;
   if(x_x != ResolutionX || y_y != ResolutionY)goto FATALREAD;
   screen=Screen1;
   for (i = 0; i < ResolutionY; i++) {
     if(fread((char *)&Nt,sizeof(short),1,ipf) != 1)goto FATALREAD;
     if(Nt == 0){
       if(fread((char *)RGB,sizeof(unsigned short),ResolutionX,ipf)
          != ResolutionX)goto FATALREAD;
     }
     else{
       if(fread((char *)compressbuffer,sizeof(short),Nt,ipf) != Nt)
         goto FATALREAD;
       UnCompressWordLine(Nt,compressbuffer,RGB);
     }
     for (j = 0; j < ResolutionX; j++) {
       val=RGB[j];
       if((cpix = (long)colorstats0[(long)val]) > GIFCOLOURS){
         cpix=getcolor(val,GIFCOLOURS);
         colorstats0[(long)val]=(unsigned long)cpix;
       }
       *screen++ = (unsigned char)cpix;
     }
     if(type == FLI)screen += (320-ResolutionX);
   }
   fclose(ipf);
   if(frame == range_f){
     if(type == FLI){
       if(Write_first_FLI_frame(opf,(short)speed,Screen1,Screen2,
          fli_buffer) < 0)goto FATALWRITE;
     }
     else if(type == FLC){
       if(Write_first_FLC_frame(opf,Screen1,Screen2,fli_buffer,(short)speed,
          (short)ResolutionX,(short)ResolutionY) <0)goto FATALWRITE;
     }
   }
   else if(frame == range_l){
     if(type == FLI){
       if(Write_FLI_frame(opf,Screen1,Screen2,fli_buffer) < 0)goto FATALWRITE;
       if(Close_FLI_file(opf) < 0)goto FATALWRITE;
     }
     else if(type == FLC){
       if(Write_FLC_frame(opf,Screen1,Screen2,fli_buffer,(short)ResolutionX,
          (short)ResolutionY) < 0)goto FATALWRITE;
       if(Close_FLC_file(opf) < 0)goto FATALWRITE;
     }
   }
   else{
     if(type == FLI){
       if(Write_FLI_frame(opf,Screen1,Screen2,fli_buffer) < 0)goto FATALWRITE;
     }
     else if(type == FLC){
       if(Write_FLC_frame(opf,Screen1,Screen2,fli_buffer,(short)ResolutionX,
          (short)ResolutionY) < 0)goto FATALWRITE;
     }
   }
 }
 fclose(opf);
 if(MessageBox(GetFocus(),(LPCTSTR)"Delete Temporary files",
          (LPCTSTR)"Flic",MB_YESNO|MB_TASKMODAL|MB_ICONQUESTION) == IDYES){
   FILE *FF;
   sprintf(format,"%s%ld%s","%s%0",tzeros,"d.16b");
   sprintf(filename,format,stem_file,range_f);
   if((FF=fopen(filename,"r")) != NULL){
     fclose(FF);
     for(i=range_f;i<range_l;i++){
       sprintf(format,"%s%ld%s","%s%0",tzeros,"d.16b");
       sprintf(filename,format,stem_file,i);
       if(DeleteFile(filename) == FALSE){
         MessageBox(GetFocus(),(LPCTSTR)strcat(filename," Failed during delete"),
         (LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
         break;
       }
     }
   }
   else MessageBox(GetFocus(),(LPCTSTR)"Failed to delete temporary files",
          (LPCTSTR)"Flic",MB_OK|MB_TASKMODAL);
 }
 goto XXIT;
 FATALWRITE:
   MessageBox(GetFocus(),(LPCTSTR)"Failed on flic write",(LPCTSTR)"Error",MB_OK|MB_TASKMODAL);
   fclose(ipf); fclose(opf); goto XXIT;
 FATALREAD:
   MessageBox(GetFocus(),(LPCTSTR)"Read failed on temporary file",(LPCTSTR)"Error",MB_OK|MB_TASKMODAL);
   fclose(ipf); fclose(opf); goto XXIT;
 FATALOPEN1:
   MessageBox(GetFocus(),(LPCTSTR)"Temporary file open failed",(LPCTSTR)"Error",MB_OK|MB_TASKMODAL);
   fclose(opf); goto XXIT;
 FATALOPEN2:
   MessageBox(GetFocus(),(LPCTSTR)"Flic file open failed",(LPCTSTR)"Error",MB_OK|MB_TASKMODAL);
 XXIT:
 X__Free(Screen1);
 X__Free(Screen2);
 X__Free(fli_buffer);
}

/************************** DLL entry point ********************************/

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
//      MessageBox (GetFocus(),"Flic builder attaching","Debug",MB_OK);
      break;
    }
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
//      MessageBox (GetFocus(),"Flic builder detaching","Debug",MB_OK);
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

