/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// file GEOMETRY.CPP

// This is the renderer's geometry file

#define MODULE_GEO

#include "render.h"

#ifdef _SUNSTYLE
static unsigned char *AllocateShadowBuffer();
static void ApplyVertexEffect();
#else
static unsigned char *AllocateShadowBuffer(void);
static void ApplyVertexEffect(long effect, vertex *v, long N, double time,
                              double vmin[], double vmax[], ivector Og);
#endif


#ifdef _SUNSTYLE
void SwapFaces(f1,f2,e1,e2) face  *f1; face  *f2;
eface  *e1; eface  *e2; {
#else
void SwapFaces(face  *f1, face  *f2, eface  *e1, eface  *e2){
#endif
 face  temp;
 eface etemp;
 memcpy(&temp,f1,sizeof(face));
 veccopy(e1->n,etemp.n);
 etemp.c=e1->c;
 memcpy(f1,f2,sizeof(face));
 veccopy(e2->n,e1->n);
 e1->c=e2->c;
 memcpy(f2,&temp,sizeof(face));
 veccopy(etemp.n,e2->n);
 e2->c=etemp.c;
}

#ifdef _SUNSTYLE
double FaceDepth(f,v) face  *f; vertex  *v; {
#else
double FaceDepth(face  *f, vertex  *v){
#endif
 double depth;
 depth = ( (v+f->V[0])->p[1] + (v+f->V[1])->p[1] + (v+f->V[2])->p[1])/3.0;
 return depth;
}

#ifdef _SUNSTYLE
void DepthSortFaces(O) long O; {
#else
void DepthSortFaces(long O){
#endif
 long i,j,gap;
 vertex  *v;
 face    *f;
 eface   *e;
 v=Object[O].Vbase;
 f=Object[O].Fbase;
 e=Object[O].Ebase;
 for (gap = Object[O].NoFaces; gap > 0; gap /=2) {
   for (i = gap; i < Object[O].NoFaces; i++) {
     for (j = i-gap; j >= 0 &&
          FaceDepth((f+j),v) < FaceDepth((f+j+gap),v); j -= gap){
       SwapFaces((f+j),(f+j+gap),(e+j),(e+j+gap));
     }
   }
 }
}

#ifdef _SUNSTYLE
void NewPoint(dp,v1,v2,p) double dp; vector v1; vector v2; vector p; {
#else
void NewPoint(double dp, vector v1, vector v2, vector p){
#endif
 double mu;
 mu=(v2[1]-v1[1]);
 if(fabs(mu) < 1.0)mu=0.0;
 else mu=(dp-v1[1])/mu;
 p[1]=dp;
 p[0]=v1[0]+mu*(v2[0]-v1[0]);
 p[2]=v1[2]+mu*(v2[2]-v1[2]);
}

#ifdef _SUNSTYLE
short ClipObject(O) long O; {
#else
short ClipObject(long O){
#endif
  vertex   *v;
  tvertex  *t;
  face     *f;
  eface    *e;
  pface    *p;
  double y1,y2,y3,dp;
  long i,i1,i2,i3,xv,xf,sz,vp,fp;
/* any faces that span the projection plane have to be turned into
   three faces with new vertices on the projection plane.
*/
  dp=1.000;
//dp=100.0; // rsf test
  xv=0; xf=0;
  v=Object[O].Vbase;
  f=Object[O].Fbase;
  p=Object[O].Pbase;
  for(i=0;i<Object[O].NoFaces;i++){
     i1= -1;
     y1=v[f[i].V[0]].p[1];
     y2=v[f[i].V[1]].p[1];
     y3=v[f[i].V[2]].p[1];
     if     (y1 <= dp && y2 <=dp && y3 >  dp)i1 = 2;
     else if(y1 <= dp && y2 > dp && y3 <= dp)i1 = 1;
     else if(y1 <= dp && y2 > dp && y3 >  dp)i1 = 0;
     else if(y1 >  dp && y2 <=dp && y3 <= dp)i1 = 0;
     else if(y1 >  dp && y2 <=dp && y3 >  dp)i1 = 1;
     else if(y1 >  dp && y2 > dp && y3 <= dp)i1 = 2;
     if(i1 >= 0){ /* calculate number of extra faces and vertices */
       xv += 2; xf += 2;
     }
  }
  if(xv == 0)return OK;  /* no clipping necessary */
  /* resize the arrays and do it */
  sz=(Object[O].NoVertices+xv)*(long)sizeof(vertex);
  if((v=(vertex  *)X__Realloc(Object[O].Vbase,sz)) == NULL)goto X__ERROR;
  Object[O].Vbase=v;
  if(Object[O].Tbase != NULL){
    sz=(Object[O].NoVertices+xv)*(long)sizeof(tvertex);
    if((t=(tvertex  *)X__Realloc(Object[O].Tbase,sz)) == NULL)goto X__ERROR;
    Object[O].Tbase=t;
  }
  sz=(Object[O].NoFaces+xf)*(long)sizeof(face);
  if((f=(face  *)X__Realloc(Object[O].Fbase,sz)) == NULL)goto X__ERROR;
  Object[O].Fbase=f;
  sz=(Object[O].NoFaces+xf)*(long)sizeof(pface);
  if((p=(pface  *)X__Realloc(Object[O].Pbase,sz)) == NULL)goto X__ERROR;
  Object[O].Pbase=p;
  sz=(Object[O].NoFaces+xf)*(long)sizeof(eface);
  if((e=(eface  *)X__Realloc(Object[O].Ebase,sz)) == NULL)goto X__ERROR;
  Object[O].Ebase=e;
  vp=Object[O].NoVertices;
  fp=Object[O].NoFaces;
  for(i=0;i<Object[O].NoFaces;i++){
     i1= -1; i2= -1; i3= -1;
     y1=v[f[i].V[0]].p[1];
     y2=v[f[i].V[1]].p[1];
     y3=v[f[i].V[2]].p[1];
     if     (y1 <= dp && y2 <=dp && y3 >  dp){i1 = 2; i2=0; i3=1;}
     else if(y1 <= dp && y2 > dp && y3 <= dp){i1 = 1; i2=0; i3=2;}
     else if(y1 <= dp && y2 > dp && y3 >  dp){i1 = 0; i2=1; i3=2;}
     else if(y1 >  dp && y2 <=dp && y3 <= dp){i1 = 0; i2=1; i3=2;}
     else if(y1 >  dp && y2 <=dp && y3 >  dp){i1 = 1; i2=0; i3=2;}
     else if(y1 >  dp && y2 > dp && y3 <= dp){i1 = 2; i2=0; i3=1;}
     if(i1 >= 0){ /* calculate the new vertex positions on viewing plane */
        /* calculate intersection point between points i1 - i2 and i1 - i3
           and place new vertices to interpolated position on viewing plane*/
       NewPoint(dp,v[f[i].V[i1]].p,v[f[i].V[i2]].p,v[vp].p);
       NewPoint(dp,v[f[i].V[i1]].p,v[f[i].V[i3]].p,v[vp+1].p);
       f[fp].V[0]=vp;
       f[fp].V[1]=f[i].V[i2];
       f[fp].V[2]=f[i].V[i3];
       f[fp+1].V[0]=vp;
       f[fp+1].V[1]=f[i].V[i3];
       f[fp+1].V[2]=vp+1;
       f[i].V[i2]=vp;
       f[i].V[i3]=vp+1;
       f[fp+1].color[0]=f[fp].color[0]=f[i].color[0];
       f[fp+1].color[1]=f[fp].color[1]=f[i].color[1];
       f[fp+1].color[2]=f[fp].color[2]=f[i].color[2];
       f[fp+1].fattrib=f[fp].fattrib=f[i].fattrib;
       f[fp+1].faxis=f[fp].faxis=f[i].faxis;
       f[fp+1].matsp=f[fp].matsp=f[i].matsp;
       f[fp+1].mattr=f[fp].mattr=f[i].mattr;
       f[fp+1].matre=f[fp].matre=f[i].matre;
       f[fp+1].texco[0]=f[fp].texco[0]=f[i].texco[0];
       f[fp+1].texco[1]=f[fp].texco[1]=f[i].texco[1];
       f[fp+1].texco[2]=f[fp].texco[2]=f[i].texco[2];
       f[fp+1].fmap=f[fp].fmap=f[i].fmap;
       f[fp+1].fshader=f[fp].fshader=f[i].fshader;
       f[fp+1].fparam=f[fp].fparam=f[i].fparam;
       p[fp+1].n[0][0] = p[fp].n[0][0] = p[i].n[0][0];
       p[fp+1].n[0][1] = p[fp].n[0][1] = p[i].n[0][1];
       p[fp+1].n[0][2] = p[fp].n[0][2] = p[i].n[0][2];
       p[fp+1].n[1][0] = p[fp].n[1][0] = p[i].n[1][0];
       p[fp+1].n[1][1] = p[fp].n[1][1] = p[i].n[1][1];
       p[fp+1].n[1][2] = p[fp].n[1][2] = p[i].n[1][2];
       p[fp+1].n[2][0] = p[fp].n[2][0] = p[i].n[2][0];
       p[fp+1].n[2][1] = p[fp].n[2][1] = p[i].n[2][1];
       p[fp+1].n[2][2] = p[fp].n[2][2] = p[i].n[2][2];
       v[vp].x=v[vp].y=v[vp+1].x=v[vp+1].y=0.0;
       if(Object[O].Tbase != NULL){
         t[vp].x=v[vp].y=v[vp+1].x=v[vp+1].y=0.0;
       }
       e[fp].c=e[fp+1].c=0.0;
       e[fp].n[0]=e[fp].n[1]=e[fp].n[2]=e[fp+1].n[0]
                 =e[fp+1].n[1]=e[fp+1].n[2]=0.0;
//not used now       f[fp].t=f[fp+1].t=0;
       vp += 2; fp += 2;
     }
  }
  Object[O].NoVertices = vp;
  Object[O].NoFaces = fp;
  return OK;
  X__ERROR:
  Render_Message(4,0,NULL);
  return FAIL;
}

#ifdef _SUNSTYLE
long IShadow(scanline,v1x,v1y,v2x,v2y,xi) long scanline; long v1x;
long v1y; long v2x; long v2y; long *xi; {
#else
long IShadow(long scanline, long v1x, long v1y, long v2x, long v2y,
             long *xi){
#endif
 double doff,ddxy;
 long dx,dy,off,dxy;
 if((v1y < scanline)  && (v2y < scanline))return 0;
 if((v1y > scanline)  && (v2y > scanline))return 0;
 dy = (v2y) - (v1y);
 if(dy == 0)return 0;
 dx = (v2x) - (v1x);
 off=scanline - v1y;
 doff=off;
 ddxy=dx*doff/dy;
 dxy=ddxy;
 *xi=(v1x)+dxy;
 return 1;
}

#ifdef _SUNSTYLE
void ShadowFillLine(scanline,vx,vy,leftedge,rightedge) long scanline;
long vx[3]; long vy[3]; long *leftedge; long *rightedge; {
#else
void ShadowFillLine(long scanline, long vx[3], long vy[3],
                      long *leftedge, long *rightedge){
#endif
 long x1,x2,x3,i1,i2,i3;
 i1=IShadow(scanline,vx[0],vy[0],vx[1],vy[1],&x1);
 i2=IShadow(scanline,vx[1],vy[1],vx[2],vy[2],&x2);
 i3=IShadow(scanline,vx[2],vy[2],vx[0],vy[0],&x3);
 if((i1 == 1) && (i2 == 1) && (i3 == 1))
 {*leftedge = min(min(x1,x2),x3);
  *rightedge= max(max(x1,x2),x3);}
 else if((i1 == 1) && (i2 == 1))
 {*leftedge = min(x1,x2);
  *rightedge= max(x1,x2);}
 else if((i2 == 1) && (i3 == 1))
 {*leftedge = min(x2,x3);
  *rightedge= max(x2,x3);}
 else if((i3 == 1) && (i1 == 1))
 {*leftedge = min(x3,x1);
  *rightedge= max(x3,x1);}
 else {
   Render_Message(5,0,NULL);
   return;
 }
 *leftedge  = max(0,min(XMAX-1,*leftedge));
 *rightedge = min(XMAX-1,max(0,*rightedge));
 return;
}

#ifdef _SUNSTYLE
static unsigned char *AllocateShadowBuffer(){
#else
static unsigned char *AllocateShadowBuffer(void){
#endif
 long i,screensize;
 unsigned char *s,*screen;
 screensize = (long)XMAX * (long)YMAX;
 if((s = (unsigned char *)X__Malloc(screensize)) == NULL){
   Render_Message(14,0,NULL);
   return NULL;
 }
 screen=s; for(i=0;i<screensize;i++)*screen++ = 0;
 return s;
}

#ifdef _SUNSTYLE
void ShadowGround(){
#else
void ShadowGround(void){
#endif
 long i,j,k,l,m;
 long vi[3];
 long xx[3],yy[3],xxmin,xxmax,yymin,yymax;
 face    *f;
 vertex  *v;
 vector p,t1;
 double mu,nu,de,temp4;
 UCHAR  *screen;
 if((VGA_screen=AllocateShadowBuffer()) == NULL)return;
 screen=VGA_screen;
 for(k=0;k<Nlights;k++){
   if(Lights[k].type == NORMAL || Lights[k].type == SHADOWS){
     for(i=0;i<ObjectCount;i++){
       if(!Object[i].cast_shadow)continue;
       v=Object[i].Vbase;
       f=Object[i].Fbase;
       for(j=0;j<Object[i].NoFaces;j++){
         if((f+j)->mattr > 0)continue; /* glass no shadow */
         vi[0]=(f+j)->V[0];
         vi[1]=(f+j)->V[1];
         vi[2]=(f+j)->V[2];
         for(l=0;l<3;l++){ /* project each vertex to shadow plane */
           p[0]=(v+vi[l])->p[0];
           p[1]=(v+vi[l])->p[1];
           p[2]=(v+vi[l])->p[2];
           vecsub(Ground.p,Lights[k].p,t1);
           nu=dot(Ground.n,t1);
           vecsub(p,Lights[k].p,t1);
           de=dot(Ground.n,t1);
           if(fabs(de) < 1.e-5)goto FACEBREAK;
           mu=nu/de;
           if(mu < 1.0)goto FACEBREAK;
           vecsub(p,Lights[k].p,t1);
           vecscale(mu,t1,t1);
           vecsum(Lights[k].p,t1,t1);
           if(t1[1] > 1.000){
             temp4=(long)(scalex*t1[0]/(t1[1]));
             xx[l]=Xcentre+temp4;
             temp4=(long)(scaley*t1[2]/(t1[1]));
             yy[l]=Ycentre-temp4;
           }
           else goto FACEBREAK;
         }
         xxmin=min(xx[0],min(xx[1],xx[2]));
         xxmax=max(xx[0],max(xx[1],xx[2]));
         yymin=min(yy[0],min(yy[1],yy[2]));
         yymax=max(yy[0],max(yy[1],yy[2]));
         if((yymax > yymin) && (xxmax > xxmin))
         for(l=max(0,yymin);l<=min(yymax,YMAX-1);l++){
           ShadowFillLine(l,xx,yy,&xxmin,&xxmax);
           for(m=xxmin;m<=xxmax;m++) *(screen+l*(long)XMAX+m) = 1;
         }
         FACEBREAK: continue;
       }
     }
   }
 }
 return;
}

#ifdef _SUNSTYLE
void ShadowObject(sn) long sn; {
#else
void ShadowObject(long sn){
#endif
 long i,j,k,l,m;
 long vi[3];
 long xx[3],yy[3],xxmin,xxmax,yymin,yymax;
 face    *f;
 vertex  *v;
 vector p,t1;
 double mu,nu,de,temp4;
 UCHAR *screen;
 if(!Object[sn].show_shadow)return;
 if(VGA_screen == NULL && R_Nground > 0)return;  /* alloc failed elsewhere */
 if(VGA_screen == NULL)VGA_screen=AllocateShadowBuffer();
 if(VGA_screen == NULL)return;
 screen=VGA_screen;
 for(k=0;k<Nlights;k++){
   if(Lights[k].type == NORMAL || Lights[k].type == SHADOWS){
     for(i=0;i<ObjectCount;i++){
       if(i == sn)continue; /* skip the object on which the shadow is cast */
       if(!Object[i].cast_shadow)continue;
       v=Object[i].Vbase;
       f=Object[i].Fbase;
       for(j=0;j<Object[i].NoFaces;j++){
         vi[0]=(f+j)->V[0];
         vi[1]=(f+j)->V[1];
         vi[2]=(f+j)->V[2];
         for(l=0;l<3;l++){ /* project each vertex to shadow plane */
           p[0]=(v+vi[l])->p[0];
           p[1]=(v+vi[l])->p[1];
           p[2]=(v+vi[l])->p[2];
           vecsub(Object[sn].Abase[0].bo,Lights[k].p,t1); /* use shader axis */
           nu=dot(Object[sn].Abase[0].bn,t1);             /* to define the    */
           vecsub(p,Lights[k].p,t1);                      /* object shadow    */
           de=dot(Object[sn].Abase[0].bn,t1);             /* plane            */

           mu=nu/de;  /* mu=0 at light 1 at face vertex */
           if(mu < 1.0)goto FACEBREAK; /* mu<1 => face beyond shadow plane */
           vecsub(p,Lights[k].p,t1);
           vecscale(mu,t1,t1);
           vecsum(Lights[k].p,t1,t1);
           if(t1[1] > 1.000){
             temp4=(long)(scalex*t1[0]/(t1[1]));
             xx[l]=Xcentre+temp4;
             temp4=(long)(scaley*t1[2]/(t1[1]));
             yy[l]=Ycentre-temp4;
           }
           else goto FACEBREAK;
         }
         xxmin=min(xx[0],min(xx[1],xx[2]));
         xxmax=max(xx[0],max(xx[1],xx[2]));
         yymin=min(yy[0],min(yy[1],yy[2]));
         yymax=max(yy[0],max(yy[1],yy[2]));
         if((yymax > yymin) && (xxmax > xxmin))
         for(l=max(0,yymin);l<=min(yymax,YMAX-1);l++){
           ShadowFillLine(l,xx,yy,&xxmin,&xxmax);
           for(m=xxmin;m<=xxmax;m++)
             *(screen+l*(long)XMAX+m) |= ( 1 << Object[sn].shadowed);
         }
         FACEBREAK: continue;
       }
     }
   }
 }
 return;
}

#if 0
********** the old way *************
#define TOL 1000.0
double hitpoint(vector n, vector x, vector y, vector pb, vector pf,
                double * a, double * b){
vector v1,p;
double mu,det,ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3;
VECSUB(pb,pf,v1)
mu=DOT(n,v1);
VECSCALE(mu,n,v1)
VECSUM(pf,v1,p)
ve1=x[0]; ve2=x[1]; ve3=x[2];
vp1=y[0]; vp2=y[1]; vp3=y[2];
p1=p[0]-pb[0];
p2=p[1]-pb[1];
p3=p[2]-pb[2];
det=ve1*vp2-vp1*ve2 ;
if(det > TOL || det < -TOL)  /* e-10 appears not to work */
  {  *a=( vp2*p1-vp1*p2)/det;
     *b=(-ve2*p1+ve1*p2)/det;
     goto HIT;
  }
det=ve1*vp3-vp1*ve3 ;
if(det > TOL || det < -TOL)
  {  *a=( vp3*p1-vp1*p3)/det;
     *b=(-ve3*p1+ve1*p3)/det;
     goto HIT;
  }
det=ve2*vp3-vp2*ve3 ;
if(det > TOL || det < -TOL)
  {  *a=( vp3*p2-vp2*p3)/det;
     *b=(-ve3*p2+ve2*p3)/det;
     goto HIT;
  }
*a = *b = 0.0;
HIT:
return mu;
}
**************************
#endif

#define TOL 1.e-4;

#ifdef _SUNSTYLE
double hitpoint(n,x,y,pb,pf,a,b) vector n; vector x; vector y; vector pb;
vector pf; double *a; double *b; {
#else
double hitpoint(vector n, vector x, vector y, vector pb, vector pf,
                double * a, double * b){
#endif
 /* n        is normal to plane of map
    x & y    are vectors along the edges of the map  x= left y = down
    pb       is position vector to base point on map
    fp       is position vector to vertex to be mapped
    a & b    are output of mapping coordinates [0 - 1]
 */
 int k;
 vector v1,p;
 double mu,dm,detmax,det1,det2,det3,ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3;
 VECSUB(pb,pf,v1)
 mu=DOT(n,v1);
 VECSCALE(mu,n,v1)
 VECSUM(pf,v1,p)
 ve1=x[0]; ve2=x[1]; ve3=x[2];
 vp1=y[0]; vp2=y[1]; vp3=y[2];
 p1=p[0]-pb[0];
 p2=p[1]-pb[1];
 p3=p[2]-pb[2];
 det1=ve1*vp2-vp1*ve2;
 det2=ve1*vp3-vp1*ve3;
 det3=ve2*vp3-vp2*ve3;
 k=0; detmax=TOL;
 if((dm=fabs(det1)) > detmax){k=1; detmax=dm;}
 if((dm=fabs(det2)) > detmax){k=2; detmax=dm;}
 if((dm=fabs(det3)) > detmax){k=3; detmax=dm;}
 if(k == 1){
   *a=( vp2*p1-vp1*p2)/det1;
   *b=(-ve2*p1+ve1*p2)/det1;
 }
 else if(k == 2){
   *a=( vp3*p1-vp1*p3)/det2;
   *b=(-ve3*p1+ve1*p3)/det2;
 }
 else if(k == 3){
   *a=( vp3*p2-vp2*p3)/det3;
   *b=(-ve3*p2+ve2*p3)/det3;
 }
 else {
   *a = *b = 0.0;
 }
 return mu;
}

#define RTD 0.159154943

#ifdef _SUNSTYLE
double hitcylinder(n,x,y,p0,p,a,b) vector n; vector x; vector y;
vector p0; vector p; double *a; double *b; {
#else
double hitcylinder(vector n, vector x, vector y, vector p0, vector p,
                double * a, double * b){
#endif
  double psi,theta,ddt;
  vector p1,q,p2;
  vecsub(p,p0,p1);
  psi =dot(p1,y)/dot(y,y);
  *b = psi;
  p2[0] = psi*y[0];     p2[1] = psi*y[1];      p2[2] = psi*y[2];
  q[0] = p1[0] - p2[0];  q[1] = p1[1] - p2[1];  q[2] = p1[2] - p2[2];
  normalize(q);
  ddt=DOT(q,n);
  if     (ddt >  1.0)ddt =  1.0;
  else if(ddt < -1.0)ddt = -1.0;
  theta=acos(ddt)*RTD;
  cross(q,n,p1);
  if(dot(p1,y) <= 0.0){
    if(theta <= 0.25) *a = (0.25 - theta);
    else  *a = (1.25 - theta);
  }
  else    *a = (0.25 + theta);
  return psi;
}

double hitsphere(vector y, vector dx, vector dy, vector p0, vector p,
                double * a, double * b){
  /* p0    is at centre of map
     dx    is vector in direction of seam
     dy    is directed to south pole
     p     is point of vertex to be assigned mapping coordinates
     a     is  phi  [0 - 1] (0 along x axis)
     b     is theta [0 - 1] (0  at north pole)
  */
  vector x,z,r;
  VECCOPY(dx,x)
  normalize(x);
  VECSCALE(-1.0,dy,z)
  normalize(z);
  VECSUB(p,p0,r)
  normalize(r);
  *b = acos(DOT(r,z))*0.3183;              /*  theta/pi                    */
  *a = atan2(DOT(r,y),DOT(r,x))*0.1592;    /*  phi/(2 pi)    0   < a < 0.5 */
  if(*a < 0.0) *a = 1.0 + *a;              /*                0.5 < a < 1.0 */
  return 0.0;
}

#define UpdateNoGlassBuffer(f,i,d,id,O) {                       \
  Zdepth[i]=d; Zbuffer[i]=id; Zobject[i]=O;                     \
}                                                               \

#if NGLASS == 1
#error BADGLASS1
#define UpdateGlassBuffer(f,i,d,id,O) {                         \
     if(f->mattr > 0){                                          \
       if(d < Zglassd[0][i]){                                   \
         Zglassd[0][i]=d; Zglass[0][i]=id; ZglassO[0][i]=O;     \
       }                                                        \
     }                                                          \
     else{                                                      \
       if(d <= Zglassd[0][i])Zglass[0][i] = -1;                 \
       Zdepth[i]=d; Zbuffer[i]=id; Zobject[i]=O;                \
     }                                                          \
}
#elif NGLASS == 2
#error BADGLASS2
#define UpdateGlassBuffer(f,i,d,id,O) {                         \
     if(f->mattr > 0){                                          \
       if(d < Zglassd[0][i]-600.0){                             \
         if(O != ZglassO[0][i] || id != Zglass[0][i]){          \
           Zglassd[1][i]=Zglassd[0][i];                         \
           ZglassO[1][i]=ZglassO[0][i];                         \
           Zglass[1][i]=Zglass[0][i];                           \
         }                                                      \
         Zglassd[0][i]=d; Zglass[0][i]=id; ZglassO[0][i]=O;     \
       }                                                        \
       else if(d > Zglassd[0][i]+600.0 &&                       \
               d < Zglassd[1][i]-600.0){                        \
         Zglassd[1][i]=d; Zglass[1][i]=id; ZglassO[1][i]=O;     \
       }                                                        \
     }                                                          \
     else{                                                      \
       if(d <= Zglassd[0][i]){                                  \
         Zglass[0][i] = -1;  Zglassd[0][i]=FARAWAY;             \
         Zglass[1][i] = -1;  Zglassd[1][i]=FARAWAY;             \
       }                                                        \
       else if(d < Zglassd[1][i]){                              \
         Zglass[1][i] = -1;                                     \
         Zglassd[1][i]=FARAWAY;                                 \
       }                                                        \
       Zdepth[i]=d; Zbuffer[i]=id; Zobject[i]=O;                \
     }                                                          \
}
#elif NGLASS == 4
#define UpdateGlassBuffer(f,i,d,id,O) {                         \
     if(f->mattr > 0){                                          \
       if(d < Zglassd[0][i]-600.0){                             \
         if(O != ZglassO[0][i] || id != Zglass[0][i]){          \
           Zglassd[3][i]=Zglassd[2][i];                         \
           ZglassO[3][i]=ZglassO[2][i];                         \
           Zglass[3][i]=Zglass[2][i];                           \
           Zglassd[2][i]=Zglassd[1][i];                         \
           ZglassO[2][i]=ZglassO[1][i];                         \
           Zglass[2][i]=Zglass[1][i];                           \
           Zglassd[1][i]=Zglassd[0][i];                         \
           ZglassO[1][i]=ZglassO[0][i];                         \
           Zglass[1][i]=Zglass[0][i];                           \
         }                                                      \
         Zglassd[0][i]=d; Zglass[0][i]=id; ZglassO[0][i]=O;     \
       }                                                        \
       else if(d > Zglassd[0][i]+600.0 &&                       \
               d < Zglassd[1][i]-600.0){                        \
         if(O != ZglassO[1][i] || id != Zglass[1][i]){          \
           Zglassd[3][i]=Zglassd[2][i];                         \
           ZglassO[3][i]=ZglassO[2][i];                         \
           Zglass[3][i]=Zglass[2][i];                           \
           Zglassd[2][i]=Zglassd[1][i];                         \
           ZglassO[2][i]=ZglassO[1][i];                         \
           Zglass[2][i]=Zglass[1][i];                           \
         }                                                      \
         Zglassd[1][i]=d; Zglass[1][i]=id; ZglassO[1][i]=O;     \
       }                                                        \
       else if(d > Zglassd[1][i]+600.0 &&                       \
               d < Zglassd[2][i]-600.0){                        \
         if(O != ZglassO[2][i] || id != Zglass[2][i]){          \
           Zglassd[3][i]=Zglassd[2][i];                         \
           ZglassO[3][i]=ZglassO[2][i];                         \
           Zglass[3][i]=Zglass[2][i];                           \
         }                                                      \
         Zglassd[2][i]=d; Zglass[2][i]=id; ZglassO[2][i]=O;     \
       }                                                        \
       else if(d > Zglassd[2][i]+600.0 &&                       \
               d < Zglassd[3][i]-600.0){                        \
         Zglassd[3][i]=d; Zglass[3][i]=id; ZglassO[3][i]=O;     \
       }                                                        \
     }                                                          \
     else{                                                      \
       if(d <= Zglassd[0][i]){                                  \
         Zglass[0][i] = -1;  Zglassd[0][i]=FARAWAY;             \
         Zglass[1][i] = -1;  Zglassd[1][i]=FARAWAY;             \
         Zglass[2][i] = -1;  Zglassd[2][i]=FARAWAY;             \
         Zglass[3][i] = -1;  Zglassd[3][i]=FARAWAY;             \
       }                                                        \
       else if(d < Zglassd[1][i]){                              \
         Zglass[1][i] = -1;  Zglassd[1][i]=FARAWAY;             \
         Zglass[2][i] = -1;  Zglassd[2][i]=FARAWAY;             \
         Zglass[3][i] = -1;  Zglassd[3][i]=FARAWAY;             \
       }                                                        \
       else if(d < Zglassd[2][i]){                              \
         Zglass[2][i] = -1;  Zglassd[2][i]=FARAWAY;             \
         Zglass[3][i] = -1;  Zglassd[3][i]=FARAWAY;             \
       }                                                        \
       else if(d < Zglassd[3][i]){                              \
         Zglass[3][i] = -1;  Zglassd[3][i]=FARAWAY;             \
       }                                                        \
       Zdepth[i]=d; Zbuffer[i]=id; Zobject[i]=O;                \
     }                                                          \
}
#else
#error BADGLASS0
#endif

#ifdef __AXP__
void UpdateGlassScanBuffers(double xl, double xr, long id, long O,
                       face  *f, eface  *e, double d2,
                       long *Zglass[], long *ZglassO[], double *Zglassd[],
                       long *Zbuffer, long *Zobject,
                       double *Zdepth, double *Zposn){
 long i,ixl,ixr;
 double depth,bL,bR,aL,aR,dL,dR;
/* xl += 0.0001; too small
   xr -= 0.0001;
   xl += 0.01;  too big
   xr -= 0.01;  */
 xl += 0.001;
 xr -= 0.001;
 ixl = (long)xl;
 ixr = (long)xr;
 if(ixl == ixr){
   dL = xl - (double)Xcentre; dL/=scalex;
   dR = xr - (double)Xcentre; dR/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL = FARAWAY; //MessageBox(
   else aL = (e->c)/bL;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY;   //MessageBox(
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=(xr+xl)/2.0;
   }
   return;
 }
 else{
   dL = xl - (double)Xcentre; dL/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL=FARAWAY; //MessageBox(NULL,"Below tollerance3",NULL,MB_OK);
   else aL = (e->c)/bL;
   bR = e->n[0]*DR[ixl] + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY; //MessageBox(NULL,"Below tollerance4",NULL,MB_OK);
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=xl;
   }
   bL = e->n[0]*DL[ixr] + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL=FARAWAY; //MessageBox(NULL,"Below tollerance5",NULL,MB_OK);
   else aL = (e->c)/bL;
   dR = xr - (double)Xcentre; dR/=scalex;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY; //MessageBox(NULL,"Below tollerance5",NULL,MB_OK);
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixr]){
     UpdateGlassBuffer(f,ixr,depth,id,O)
     Zposn[ixr]=xr;
   }
 }
 if(ixr-ixl  > 1)for(i=ixl+1;i<ixr;i++)
 {
   bL = e->n[0]*DL[i] + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL=FARAWAY; //MessageBox(NULL,"Below tollerance7",NULL,MB_OK);
   else aL = (e->c)/bL;
   bR = e->n[0]*DR[i] + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY; //MessageBox(NULL,"Below tollerance8",NULL,MB_OK);
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[i]){
     UpdateGlassBuffer(f,i,depth,id,O)
     Zposn[i]=(double)i+0.5;
   }
 }
}
#else
void UpdateGlassScanBuffers(double xl, double xr, long id, long O,
                       face  *f, eface  *e, double d2,
                       long *Zglass[], long *ZglassO[], double *Zglassd[],
                       long *Zbuffer, long *Zobject,
                       double *Zdepth, double *Zposn){
 long i,ixl,ixr;
 double depth,bL,bR,aL,aR,dL,dR;
/* xl += 0.0001; too small
   xr -= 0.0001;
   xl += 0.01;  too big
   xr -= 0.01;  */
 xl += 0.001;
 xr -= 0.001;
 ixl = (long)xl;
 ixr = (long)xr;
 if(ixl == ixr){
   dL = xl - (double)Xcentre; dL/=scalex;
   dR = xr - (double)Xcentre; dR/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=(xr+xl)/2.0;
   }
   return;
 }
 else{
   dL = xl - (double)Xcentre; dL/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   bR = e->n[0]*DR[ixl] + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=xl;
   }
   bL = e->n[0]*DL[ixr] + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   dR = xr - (double)Xcentre; dR/=scalex;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixr]){
     UpdateGlassBuffer(f,ixr,depth,id,O)
     Zposn[ixr]=xr;
   }
 }
 if(ixr-ixl  > 1)for(i=ixl+1;i<ixr;i++)
 {
   bL = e->n[0]*DL[i] + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   bR = e->n[0]*DR[i] + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[i]){
     UpdateGlassBuffer(f,i,depth,id,O)
     Zposn[i]=(double)i+0.5;
   }
 }
}
/* end of AXP */
#endif

#ifdef __AXP__
void UpdateNoGlassScanBuffers(double xl, double xr, long id, long O,
                       face  *f, eface  *e, double d2,
                       long *Zglass[], long *ZglassO[], double *Zglassd[],
                       long *Zbuffer, long *Zobject,
                       double *Zdepth, double *Zposn){
 long i,ixl,ixr;
 double depth,bL,bR,aL,aR,dL,dR;
/* xl += 0.0001; too small
   xr -= 0.0001;
   xl += 0.01;  too big
   xr -= 0.01;  */
 xl += 0.001;
 xr -= 0.001;
 ixl = (long)xl;
 ixr = (long)xr;
 if(ixl == ixr){
   dL = xl - (double)Xcentre; dL/=scalex;
   dR = xr - (double)Xcentre; dR/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL = FARAWAY; //MessageBox(
   else aL = (e->c)/bL;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY;   //MessageBox(
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateNoGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=(xr+xl)/2.0;
   }
   return;
 }
 else{
   dL = xl - (double)Xcentre; dL/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL=FARAWAY; //MessageBox(NULL,"Below tollerance3",NULL,MB_OK);
   else aL = (e->c)/bL;
   bR = e->n[0]*DR[ixl] + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY; //MessageBox(NULL,"Below tollerance4",NULL,MB_OK);
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateNoGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=xl;
   }
   bL = e->n[0]*DL[ixr] + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL=FARAWAY; //MessageBox(NULL,"Below tollerance5",NULL,MB_OK);
   else aL = (e->c)/bL;
   dR = xr - (double)Xcentre; dR/=scalex;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY; //MessageBox(NULL,"Below tollerance5",NULL,MB_OK);
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixr]){
     UpdateNoGlassBuffer(f,ixr,depth,id,O)
     Zposn[ixr]=xr;
   }
 }
 if(ixr-ixl  > 1)for(i=ixl+1;i<ixr;i++)
 {
   bL = e->n[0]*DL[i] + e->n[2]*d2 + e->n[1];
   if(fabs(bL) < 1.e-30)aL=FARAWAY; //MessageBox(NULL,"Below tollerance7",NULL,MB_OK);
   else aL = (e->c)/bL;
   bR = e->n[0]*DR[i] + e->n[2]*d2 + e->n[1];
   if(fabs(bR) < 1.e-30)aR=FARAWAY; //MessageBox(NULL,"Below tollerance8",NULL,MB_OK);
   else aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[i]){
     UpdateNoGlassBuffer(f,i,depth,id,O)
     Zposn[i]=(double)i+0.5;
   }
 }
}
#else
void UpdateNoGlassScanBuffers(double xl, double xr, long id, long O,
                       face  *f, eface  *e, double d2,
                       long *Zglass[], long *ZglassO[], double *Zglassd[],
                       long *Zbuffer, long *Zobject,
                       double *Zdepth, double *Zposn){
 long i,ixl,ixr;
 double depth,bL,bR,aL,aR,dL,dR;
/* xl += 0.0001; too small
   xr -= 0.0001;
   xl += 0.01;  too big
   xr -= 0.01;  */
 xl += 0.001;
 xr -= 0.001;
 ixl = (long)xl;
 ixr = (long)xr;
 if(ixl == ixr){
   dL = xl - (double)Xcentre; dL/=scalex;
   dR = xr - (double)Xcentre; dR/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateNoGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=(xr+xl)/2.0;
   }
   return;
 }
 else{
   dL = xl - (double)Xcentre; dL/=scalex;
   bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   bR = e->n[0]*DR[ixl] + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixl]){
     UpdateNoGlassBuffer(f,ixl,depth,id,O)
     Zposn[ixl]=xl;
   }
   bL = e->n[0]*DL[ixr] + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   dR = xr - (double)Xcentre; dR/=scalex;
   bR = e->n[0]*dR + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[ixr]){
     UpdateNoGlassBuffer(f,ixr,depth,id,O)
     Zposn[ixr]=xr;
   }
 }
 if(ixr-ixl  > 1)for(i=ixl+1;i<ixr;i++)
 {
   bL = e->n[0]*DL[i] + e->n[2]*d2 + e->n[1];
   aL = (e->c)/bL;
   bR = e->n[0]*DR[i] + e->n[2]*d2 + e->n[1];
   aR = (e->c)/bR;
   depth=min(aL,aR);
   if(depth > 1.000 && depth <= Zdepth[i]){
     UpdateNoGlassBuffer(f,i,depth,id,O)
     Zposn[i]=(double)i+0.5;
   }
 }
}
/* end of AXP */
#endif

#ifdef _SUNSTYLE
long Intersect(scanline,v1,v2,xi) long scanline; vertex  *v1;
vertex  *v2; double *xi; {
#else
long Intersect(long scanline, vertex  *v1, vertex  *v2,
                double *xi){
#endif
 double ddx,ddy,doff,ddxy;
 if((v1->y < (double)scanline)  && (v2->y < (double)scanline))return 0;
 if((v1->y > (double)scanline)  && (v2->y > (double)scanline))return 0;
 ddy = (v2->y) - (v1->y);
 if(fabs(ddy) < 0.0001)return 0;
 ddx = (v2->x) - (v1->x);
 doff=(double)scanline - v1->y;
 ddxy=ddx*doff/ddy;
 *xi=(v1->x)+ddxy;
 return 1;
}

#ifdef _SUNSTYLE
long Active(scanline,v,f,leftedge,rightedge) long scanline;
vertex  *v; face  *f; double *leftedge; double *rightedge; {
#else
long Active(long scanline, vertex  *v, face  *f,
                      double *leftedge, double *rightedge){
#endif
 long v1,v2,v3
     ,i1,i2,i3;
 double x1,x2,x3;
 if(f->top > scanline)       return 0;
 if(f->bottom < scanline)    return 0;
 v1=f->V[0];
 v2=f->V[1];
 v3=f->V[2];
 i1=Intersect(scanline,(v+v1),(v+v2),&x1);
 i2=Intersect(scanline,(v+v2),(v+v3),&x2);
 i3=Intersect(scanline,(v+v3),(v+v1),&x3);
 if((i1 == 1) && (i2 == 1) && (i3 == 1))
 {*leftedge = min(min(x1,x2),x3);
  *rightedge= max(max(x1,x2),x3);}
 else if((i2 == 1) && (i3 == 1))
 {*leftedge = min(x2,x3);
  *rightedge= max(x2,x3);}
 else if((i3 == 1) && (i1 == 1))
 {*leftedge = min(x3,x1);
  *rightedge= max(x3,x1);}
 else if((i1 == 1) && (i2 == 1))
 {*leftedge = min(x1,x2);
  *rightedge= max(x1,x2);}
 else return 0;
 if(*leftedge >= (double)XMAX-0.002)return 0;
 else *leftedge  = max(0,*leftedge);
 if(*rightedge < (double)XMIN+0.002)return 0;
 *rightedge = min((double)XMAX-0.01,*rightedge);
 return 1;
}

#ifdef _SUNSTYLE
void SetPerspectiveView(){
#else
void SetPerspectiveView(void){
#endif
 double t1[4][4],t2[4][4],t3[4][4];
 double lenz=50.0;
 if(ResolutionX > ResolutionY)
   scalex=scaley=(double)ResolutionX/2.0/21.22*lenz;
 else
   scalex=scaley=(double)ResolutionY/2.0/21.22*lenz;
 if(ResolutionX <= 320 && ResolutionY <= 200)
   scaley /= 1.2;     /* 3ds uses  *0.82 = /1.21 but this seems wrong */
 if(anti_alias == LOW)   {scalex *= 2;}
 if(anti_alias == MEDIUM){scalex *= 2; scaley *= 2;}
 if(anti_alias == HIGH)  {scalex *= 3; scaley *= 2;}
 if(anti_alias == ULTRA) {scalex *= 3; scaley *= 3;}
 CamTheta *= PI/180.0;
 CamPhi   *= PI/180.0;
 CamAlpha *= PI/180.0;
 scalex *= CamSx;  /* apply any more scaling or aspect ration */
 scaley *= CamSy;
 R_tram(t1,-ViewPoint[0],-ViewPoint[1],-ViewPoint[2]);
 R_rotz(t2,-CamPhi);
 R_m4by4(t2,t1,t3);
 R_rotx(t1,-CamAlpha);
 R_m4by4(t1,t3,t2);
 R_roty(t1,-CamTheta);   /* bank on axis */
 R_m4by4(t1,t2,trr);
 R_roty(t1,CamTheta);    /* for reflection maps get a transform back to */
 R_rotx(t2,CamAlpha);    /* co-ordinate system without camera, ie trr-1 */
 R_m4by4(t2,t1,t3);      /* but this is easier to compute               */
 R_rotz(t1,CamPhi);
 R_m4by4(t1,t3,AntiCamera);
 return;
}

#ifdef _SUNSTYLE
void SetTrPos(fi,theta,alpha,im,ima,sx,sy,sz,Offset,Position,trpos,trset)
double fi; double theta; double alpha; short im; double ima; double sx;
double sy; double sz; ivector Offset; vector Position;
double trpos[4][4]; double trset[4][4]; double trinv[4][4]; {
#else
void SetTrPos(double fi, double theta, double alpha
             ,short im, double ima
             ,double sx, double sy, double sz
             ,ivector Offset, vector Position,
             double trpos[4][4], double trset[4][4], double trinv[4][4]){
#endif
 double t1[4][4],t2[4][4],t3[4][4],t4[4][4];
 fi    *= PI/180.0;
 theta *= PI/180.0;
 alpha *= PI/180.0;
 R_tram(t2,-(double)Offset[0],-(double)Offset[1],-(double)Offset[2]);
 R_scal(t3,sx,sy,sz);
 R_m4by4(t3,t2,t1);
 if(im > 0){
   if(im == 1)R_rotz(t3,ima*PI/180.0);
   if(im == 2)R_rotx(t3,ima*PI/180.0);
   if(im == 3)R_roty(t3,ima*PI/180.0);
   R_m4by4(t3,t1,t2);
   R_roty(t1,theta);
   R_m4by4(t1,t2,t3);
 }
 else{
   R_roty(t2,theta);
   R_m4by4(t2,t1,t3);
 }
 R_rotx(t1,alpha);
 R_m4by4(t1,t3,t2);
 R_rotz(t1,fi);
 R_m4by4(t1,t2,t3);
 R_tram(t1,Position[0],Position[1],Position[2]);
 R_m4by4(t1,t3,trset);       // excluding the viewing transform
 R_m4by4(trr,trset,trpos);   // including the viewing transform

 /* Get the inverse object transformation */

 R_tram(t1,-Position[0],-Position[1],-Position[2]);
 R_rotz(t2,-fi);
 R_m4by4(t2,t1,t3);
 R_rotx(t1,-alpha);
 R_m4by4(t1,t3,t2);
 if(im > 0){
   R_roty(t1,-theta);
   if(im == 1)R_rotz(t3,-ima*PI/180.0);
   if(im == 2)R_rotx(t3,-ima*PI/180.0);
   if(im == 3)R_roty(t3,-ima*PI/180.0);
   R_m4by4(t3,t1,t4);
 }
 else{
   R_roty(t4,-theta);
 }
 R_m4by4(t4,t2,t1);
 R_scal(t2,1.0/sx,1.0/sy,1.0/sz);
 R_m4by4(t2,t1,t3);
 R_tram(t2,(double)Offset[0],(double)Offset[1],(double)Offset[2]);
 R_m4by4(t2,t3,t1);
 R_m4by4(t1,AntiCamera,trinv);
 return;
}

#ifdef _SUNSTYLE
void R_rotz(tr,ang) double tr[4][4]; double ang; {
#else
void R_rotz(double tr[4][4], double ang){
#endif
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][1] = -sin(ang);
  tr[1][0] =  sin(ang);
  tr[1][1] =  cos(ang);
  return;
}

#ifdef _SUNSTYLE
void R_rotx(tr,ang) double tr[4][4]; double ang; {
#else
void R_rotx(double tr[4][4], double ang){
#endif
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[1][1] =  cos(ang);
  tr[1][2] = -sin(ang);
  tr[2][1] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

#ifdef _SUNSTYLE
void R_roty(tr,ang) double tr[4][4]; double ang; {
#else
void R_roty(double tr[4][4], double ang){
#endif
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][2] = -sin(ang);
  tr[2][0] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

#ifdef _SUNSTYLE
void R_tram(t,dx,dy,dz) double t[4][4]; double dx; double dy; double dz; {
#else
void R_tram(double t[4][4], double dx, double dy, double dz){
#endif
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  t[i][j]=0.0;
  if(i == j)t[i][j]=1.0;
 }
 t[0][3]=dx;
 t[1][3]=dy;
 t[2][3]=dz;
 t[3][3]=1;
 return;
}

#ifdef _SUNSTYLE
void R_scal(t,sx,sy,sz) double t[4][4]; double sx; double sy; double sz; {
#else
void R_scal(double t[4][4], double sx, double sy, double sz){
#endif
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  t[i][j]=0.0;
  if(i == j)t[i][j]=1.0;
 }
 t[0][0]=sx;
 t[1][1]=sy;
 t[2][2]=sz;
 t[3][3]=1;
 return;
}

#ifdef _SUNSTYLE
void R_m4by4(t1,t2,tr) double t1[4][4]; double t2[4][4]; double tr[4][4]; {
#else
void R_m4by4(double t1[4][4], double t2[4][4], double tr[4][4]){
#endif
 long i,j,k;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  for(k=0;k<4;k++)tr[i][j]=tr[i][j]+t1[i][k]*t2[k][j];
 }
 return;
}

#ifdef _SUNSTYLE
void R_c4to4(t1,t2) double tin[4][4]; double tout[4][4]; {
#else
void R_c4to4(double tin[4][4], double tout[4][4]){
#endif
 int i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 tout[i][j]=tin[i][j];
 return;
}

#ifdef _SUNSTYLE
void R_null_transform(t) double t[4][4]; {
#else
void R_null_transform(double t[4][4]){
#endif
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 if(i == j)t[i][j]=1.0;
 else      t[i][j]=0.0;
}

#ifdef _SUNSTYLE
void R_m4by1(t4,x,y,z,xx,yy,zz) double t4[4][4]; double x; double y;
double z; double *xx; double *yy; double *zz; {
#else
void R_m4by1(double t4[4][4], double x, double y, double z,
           double *xx, double *yy, double *zz){
#endif
   *xx=t4[0][0]*x+t4[0][1]*y+t4[0][2]*z+t4[0][3];
   *yy=t4[1][0]*x+t4[1][1]*y+t4[1][2]*z+t4[1][3];
   *zz=t4[2][0]*x+t4[2][1]*y+t4[2][2]*z+t4[2][3];
 return;
}

#ifdef _SUNSTYLE
void R_reflect(t,p,n) double t[4][4]; vector p; vector n; {
#else
void R_reflect(double t[4][4], vector p, vector n){
#endif
 /* p is point in reflection plane  */
 /* n is normal to reflection plane */
 long i,j;
 double dxy,phi,theta;
 double t1[4][4],t2[4][4],t3[4][4];
//{
//char temp[256];
//sprintf(temp,"n= %lf %lf %lf p=%lf %lf %lf",n[0],n[1],n[2],p[0],p[1],p[2]);
//MessageBox(NULL,temp,"Debug",MB_OK);
//}
 dxy=n[0]*n[0]+n[1]*n[1];
 R_tram(t1,-p[0],-p[1],-p[2]);
 if(fabs(n[2]) > 0.7071){
//MessageBox(NULL,"transform to Z","Debug",MB_OK);
  if(fabs(n[2]) > 0.998){ /* vertical so just reflect in XY plane and return */
// MessageBox(NULL,"Vertical","Debug",MB_OK);
     for(i=0;i<4;i++)for(j=0;j<4;j++){
       if(i == j)t2[i][j]=1.0;
       else      t2[i][j]=0.0;
     }
     t2[2][2]=-1.0;
     R_m4by4(t2,t1,t3);
     R_tram(t1,p[0],p[1],p[2]);
     R_m4by4(t1,t3,t);
     return;
   }
   dxy=sqrt(dxy);
   if     (fabs(n[0]) < 1.e-6 && n[1] > 0.0)phi =  1.570796;
   else if(fabs(n[0]) < 1.e-6 && n[1] < 0.0)phi = -1.570796;
   else phi = atan2(n[1],n[0]);
   if(n[2] > 0.0)theta =  1.570796-atan2(n[2],dxy);
   else          theta = -1.570796-atan2(n[2],dxy);
   R_rotz(t2, -phi);
   R_m4by4(t2,t1,t3);
   R_roty(t2, theta);
   R_m4by4(t2,t3,t1);
   for(i=0;i<4;i++)for(j=0;j<4;j++){
     if(i == j)t2[i][j]=1.0;
     else      t2[i][j]=0.0;
   }
   t2[2][2]=-1.0;  /* reflect in XY plane - perpend`r to Z axis */
   R_m4by4(t2,t1,t3);
   R_roty(t2,-theta);
   R_m4by4(t2,t3,t1);
   R_rotz(t2,phi);
   R_m4by4(t2,t1,t3);
   R_tram(t1,p[0],p[1],p[2]);
   R_m4by4(t1,t3,t);
 }
 else{
   dxy=sqrt(dxy);
   if     (fabs(n[0]) < 1.e-6 && n[1] > 0.0)phi =  1.570796;
   else if(fabs(n[0]) < 1.e-6 && n[1] < 0.0)phi = -1.570796;
   else phi = atan2(n[1],n[0]);
   theta = atan2(n[2],dxy);
   R_rotz(t2, -phi);
   R_m4by4(t2,t1,t3);
   R_roty(t2, -theta);
   R_m4by4(t2,t3,t1);
   for(i=0;i<4;i++)for(j=0;j<4;j++){
     if(i == j)t2[i][j]=1.0;
     else      t2[i][j]=0.0;
   }
   t2[0][0]=-1.0;  /* reflect in YZ plane - perpend`r to X axis */
   R_m4by4(t2,t1,t3);
   R_roty(t2,theta);
   R_m4by4(t2,t3,t1);
   R_rotz(t2,phi);
   R_m4by4(t2,t1,t3);
   R_tram(t1,p[0],p[1],p[2]);
   R_m4by4(t1,t3,t);
 }
 return;
}

// Additional geometry routines

#ifdef _SUNSTYLE

void veccopy(v1,r)
vector v1;
vector r;
{
 r[0] = v1[0];
 r[1] = v1[1];
 r[2] = v1[2];
}

void vecsub(v1,v2,r)
vector v1;
vector v2;
vector r;
{
 r[0] = v1[0] - v2[0];
 r[1] = v1[1] - v2[1];
 r[2] = v1[2] - v2[2];
}

void vecsum(v1,v2,r)
vector v1;
vector v2;
vector r;
{
 r[0] = v1[0] + v2[0];
 r[1] = v1[1] + v2[1];
 r[2] = v1[2] + v2[2];
}

void vecscale(s,v,r)
double s;
vector v;
vector r;
{
 r[0] = s * v[0];
 r[1] = s * v[1];
 r[2] = s * v[2];
}

short normalize(v)
vector v;
{
 double n,nn;
 n= (double)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return 1;
 nn=sqrt(n);
 n = 1.0 / nn;
 vecscale(n,v,v);
 return 0;
}

double R_length(v)
vector v;
{
 double n,nn;
 n= (double)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 nn=sqrt(n);
 return (nn);
}

void cross(v1,v2,r)
vector v1;
vector v2;
vector r;
{
 r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);
 r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);
 r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);
}

double dot(v1,v2)
vector v1;
vector v2;
{
 double result;
 result  = v1[0] * v2[0];
 result += v1[1] * v2[1];
 result += v1[2] * v2[2];
 return result;
}

#else

void veccopy(vector v1, vector r){
 r[0] = v1[0];
 r[1] = v1[1];
 r[2] = v1[2];
}

void vecsub(vector v1, vector v2, vector r){
 r[0] = v1[0] - v2[0];
 r[1] = v1[1] - v2[1];
 r[2] = v1[2] - v2[2];
}

void vecsum(vector v1,vector v2, vector r){
 r[0] = v1[0] + v2[0];
 r[1] = v1[1] + v2[1];
 r[2] = v1[2] + v2[2];
}

void vecscale(double s, vector v, vector r){
 r[0] = s * v[0];
 r[1] = s * v[1];
 r[2] = s * v[2];
}

short normalize(vector v){
 double n,nn;
 n= (double)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return 1;
 nn=sqrt(n);
 n = 1.0 / nn;
 vecscale(n,v,v);
 return 0;
}

double R_length(vector v){
 double n,nn;
 n= (double)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 nn=sqrt(n);
 return (nn);
}

void cross(vector v1,vector v2, vector r){
 r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);
 r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);
 r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);
}

double dot(vector v1, vector v2){
 double result;
 result  = v1[0] * v2[0];
 result += v1[1] * v2[1];
 result += v1[2] * v2[2];
 return result;
}

#endif

// end of basic geometry routines
