/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file GIFLIB.H   */

#ifndef GIF_LIB_H
#define GIF_LIB_H

#define ERROR           0

#define OK              1

#ifndef TRUE
#define TRUE            1
#define FALSE           0
#endif

#define FILE_BUFFER_SIZE 16384       /* Files uses bigger buffers than usual */

typedef int             BooleanType;
typedef unsigned char   PixelType;
typedef unsigned char * RowType;
typedef unsigned char   ByteType;

#ifdef SYSV
#define VoidPtr char *
#else
#define VoidPtr void *
#endif /* SYSV */

typedef struct GifColorType {
    ByteType Red, Green, Blue;
} GifColorType;

/* Note entries prefixed with S are of Screen information, while entries     */
/* prefixed with I are of the current defined Image.                         */
typedef struct GifFileType {
    long SWidth, SHeight,                               /* Screen dimensions */
	SColorResolution, SBitsPerPixel, /* How many colors can we generate? */
	SBackGroundColor,               /* I hope you understand this one... */
	ILeft, ITop, IWidth, IHeight,            /* Current image dimensions */
	IInterlace,                           /* Sequential/Interlaced lines */
	IBitsPerPixel;                    /* How many colors this image has? */
    GifColorType *SColorMap, *IColorMap;               /* NULL if not exists */
    VoidPtr Private;      /* The regular user should not mesh with this one! */
} GifFileType;

typedef enum {
    UNDEFINED_RECORD_TYPE,
    SCREEN_DESC_RECORD_TYPE,
    IMAGE_DESC_RECORD_TYPE,                                /* Begin with ',' */
    EXTENSION_RECORD_TYPE,                                 /* Begin with '!' */
    TERMINATE_RECORD_TYPE                                  /* Begin with ';' */
} GifRecordType;


#ifdef _SUNSTYLE
GifFileType *EGifOpenFileName();
GifFileType *EGifOpenFileHandle();
long EGifPutScreenDesc();
long EGifPutImageDesc();
long EGifPutLine();
long EGifPutCode();
long EGifPutCodeNext();
long EGifCloseFile();
#else
GifFileType *EGifOpenFileName(char *GifFileName,
     long GifTestExistance, long vers);
GifFileType *EGifOpenFileHandle(char *filename, long vers);
long EGifPutScreenDesc(GifFileType *GifFile,
     long GifWidth, long GifHeight, long GifColorRes, long GifBackGround,
     long GifBitsPerPixel, GifColorType *GifColorMap);
long EGifPutImageDesc(GifFileType *GifFile,
     long GifLeft, long GifTop, long Width, long GifHeight, long GifInterlace,
     long GifBitsPerPixel, GifColorType *GifColorMap);
long EGifPutControlDesc(GifFileType *GifFile,
     long transp, ByteType transpid, long delay, long dummy);
long EGifPutBrowserControlDesc(GifFileType *GifFile,long repeat);
long EGifPutLine(GifFileType *GifFile, PixelType *GifLine, long GifLineLen);
long EGifPutCode(GifFileType *GifFile, long GifCodeSize, ByteType *GifCodeBlock);
long EGifPutCodeNext(GifFileType *GifFile, ByteType *GifCodeBlock);
long EGifCloseFile(GifFileType *GifFile);
#endif

#define E_GIF_ERR_OpenFailed    1                /* And EGif possible errors */
#define E_GIF_ERR_WriteFailed   2
#define E_GIF_ERR_HasScrnDscr   3
#define E_GIF_ERR_HasImagDscr   4
#define E_GIF_ERR_NoColorMap    5
#define E_GIF_ERR_DataTooBig    6
#define E_GIF_ERR_NotEnoughMem  7
#define E_GIF_ERR_DiskIsFull    8
#define E_GIF_ERR_CloseFailed   9
#define E_GIF_ERR_NotWriteable  10

#endif /* GIF_LIB_H */

