/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// file loadmaps.cpp

// load and prepare any image maps used by the renderer


#define MODULE_GIF

/* Output file writing + image map loading */

#include "render.h"
#undef OK                     /* OK is defined again in giflib.h */
#include "giflib.h"

typedef struct tagRenderIMAP {
 unsigned char *p,*pR,*pG,*pB,*p24R,*p24G,*p24B;
 short image,depth,xmax,ymax;
} RenderIMAP;

#ifdef _SUNSTYLE
/* for sun which does'nt appear to know  */
#define SEEK_SET 0
long out_line();
long get_byte();
extern short GIF_decoder();
static short R_getvariword();
static short R_putlong();
static short R_putshort();
static short R_putbyte();
static char *Put2B();
static short Load24bitBrush();
static long LoadCelBrush();
static long LoadBmpBrush();
static long Load8bitBrush();
static long distance();
static long GetRenderImage();
long getcolor();
void SortColours();
#else
long out_line(unsigned char *pixels, long linelen);
long get_byte(void);
extern short GIF_decoder(short linewidth);
static short R_getvariword(long n, FILE *f, long*vw);
static short R_putlong(long lng, FILE *f);
static short R_putshort(short sht, FILE *f);
static short R_putbyte(char c ,FILE *f);
static char *Put2B(char *buffer, long l, short key, long *bufsize);
static short Load24bitBrush(char *fn, char *ofname, RenderIMAP *map);
static short LoadExternalBrush(char *fn, char *ofname, RenderIMAP *map);
static long Load8bitBrush(char *fn, char *ofname, RenderIMAP *map);
static long LoadCelBrush(char *fn, char *ofname, RenderIMAP *map);
static long LoadBmpBrush(char *fn, char *ofname, RenderIMAP *map);
static long distance(unsigned short argb, unsigned short brgb);
static long GetRenderImage(imap *Map, char *mfname, char *ofname,
                           RenderIMAP *map);
long getcolor(unsigned short crgb, long Npalette);
void SortColours(long Ngif, long N_colours);
#endif

extern char gszHomeDir[];

static FILE *Bit_File;

static long colourmap_size,brush_size;
static long ScreenXMax,ScreenYMax,ScreenColorBits;
static PixelType *ScanLine;

#define GIFCOLOURS 256

GifColorType GifColourMap[GIFCOLOURS];
GifFileType *GifFile;
unsigned char *decoderline;

#ifdef _SUNSTYLE
short CompressWordLine(input,buffer,xmax) unsigned short *input;
short *buffer; long xmax;{
#else
short CompressWordLine(unsigned short *input, short *buffer, long xmax){
#endif
 long i;
 short *p,run,length,bufsize;
 p=buffer;  buffer++; *buffer++ = *input++;
 *p=0; run=0; length=0; bufsize = 2;
 for(i=1;i<xmax;i++){
   if( *input == *(input-1) ){
     if(length > 0){
       (*p)++; p = (buffer-1); *p = 1; *buffer++ = *input++; bufsize++;
       length=0; run=1;
     }
     else{
       (*p)++; input++; run++;
     }
   }
   else{
     if(run > 0){
       p=buffer++; *p=0; *buffer++ = *input++; bufsize +=2;
       run=0;
     }
     else{
       (*p)--; *buffer++ = *input++; bufsize++; length++;
     }
     if(bufsize > xmax-10)return 0;
   }
 }
 return bufsize;
}

// code to deal with anti-aliasing

extern unsigned char *gRed,   *gGreen,   *gBlue,
                     *gL_Red, *gL_Green, *gL_Blue;

static int n_per_line[5]={1,1,2,2,3};

long ScreenTo16bitFile(long frame, char *outputfile, char *out_root){
 char format[32];
 FILE *f16;
 fullscreenbuffer *f;
 unsigned short r,g,b,val,Nt;
 unsigned short *Ergb;
 short          *LineBuffer;
 long i,j,k,xmax;
 sprintf(format,"%s%ld%s","%s%0",tzeros,"d.16b");
 if(nScriptframes > 1)
   sprintf(outputfile,format,out_root,frame);
 else
   sprintf(outputfile,"%s.16b",out_root);
 if((f16=fopen(strlwr(outputfile),"wb")) == NULL)return FAIL;
 Render_Message(46,0,outputfile);
 Ergb=RendererUshortBuffer;
 LineBuffer=RendererShortBuffer;
 xmax=ResolutionX;
 if(fwrite((char *)&ResolutionX,sizeof(long),1,f16) != 1 ||
    fwrite((char *)&ResolutionY,sizeof(long),1,f16) != 1 ||
    fflush(f16) != 0){
   Render_Message(49,1,NULL);
   goto FPT;
 }
 for(i=0;i<ResolutionY;i++){
   f=FullScreenBuffer+i*xmax;
   for(j=0; j<xmax; j++){
     *(gRed+j)   = f->R;
     *(gGreen+j) = f->G;
     *(gBlue+j)  = f->B;
     f++;
   }
   if(dither_on != NO)DitherLine(i,xmax,gRed,gGreen,gBlue);
   for (k=0; k<xmax; k++) {
     val=((unsigned short)(gRed[k]   & 0xf8) << 7)
        |((unsigned short)(gGreen[k] & 0xf8) << 2)
        |((unsigned short)(gBlue[k]        ) >> 3);
     colorstats0[val]++;
     Ergb[k]=val;
   }
   Nt=CompressWordLine(Ergb,LineBuffer,xmax);
   if(fwrite((char *)&Nt,sizeof(short),1,f16) != 1)goto FPT;
   else if(fflush(f16) != 0)goto FPT;
   if(Nt == 0){
     if(fwrite((char *)Ergb,sizeof(unsigned short),xmax,f16) != xmax)goto FPT;
     else if(fflush(f16) != 0)goto FPT;
   }
   else{
     if(fwrite((char *)LineBuffer,sizeof(short),Nt,f16) != Nt)goto FPT;
     else if(fflush(f16) != 0)goto FPT;
   }
 }
 fclose(f16);
 return 0;
 FPT:
 fclose(f16);
 return FAIL;
}

long AntiAliasLine(long scanline,long aa,
         UCHAR *Red, UCHAR *Green, UCHAR *Blue, UCHAR *Alfa,
         UCHAR *L_Red, UCHAR *L_Green, UCHAR *L_Blue, UCHAR *L_Alfa,
         UCHAR *LL_Red, UCHAR *LL_Green, UCHAR *LL_Blue, UCHAR *LL_Alfa){
 fullscreenbuffer *f;
 unsigned short r,g,b,val,Nt;
 unsigned short *Ergb;
 short          *LineBuffer;
 long j,k,xmax;
 Ergb=RendererUshortBuffer;
 LineBuffer=RendererShortBuffer;
 xmax=ResolutionX;
 if(aa == NONE){
   OUTPUT:
   if(ntsc_colours == YES){ /* limit colours 0-240 */
     for(j=0;j<xmax;j++){
       Red[j]   = (unsigned char)( ((double)Red[j])   * 0.9412);
       Green[j] = (unsigned char)( ((double)Green[j]) * 0.9412);
       Blue[j]  = (unsigned char)( ((double)Blue[j])  * 0.9412);
     }
   }
   if(instant_callback == YES)
     AddToRenderImagePreview(scanline/n_per_line[aa],
                             ResolutionX,ResolutionY,
                             NULL,Red,Green,Blue);
   if(FullScreenBuffer != NULL){
     f=FullScreenBuffer+(scanline/n_per_line[aa])*xmax;
     for(j=0; j<xmax; j++){
       f->R = *(Red+j);  f->G = *(Green+j);
       f->B = *(Blue+j); f->A = *(Alfa+j);
       f++;
     }
     return 0;
   }
   if(dither_on != NO)DitherLine(scanline,xmax,Red,Green,Blue);
   if(ColourResolution != BITS24){
     if(OUTPUT_FILE_TYPE == FLC_FILE){
       if(AddToRenderFlicFile(scanline/n_per_line[aa],
                              Red,Green,Blue) != 1)return FAIL;
     }
     else{
       for (k=0; k<xmax; k++) {
         val=((unsigned short)(Red[k]   & 0xf8) << 7)
            |((unsigned short)(Green[k] & 0xf8) << 2)
            |((unsigned short)(Blue[k]        ) >> 3);
         colorstats0[val]++;
         Ergb[k]=val;
       }
       Nt=CompressWordLine(Ergb,LineBuffer,xmax);
       if(fwrite((char *)&Nt,sizeof(short),1,work) != 1) return FAIL;
       else if(fflush(work) != 0)return FAIL;
       if(Nt == 0){
         if(fwrite((char *)Ergb,sizeof(unsigned short),xmax,work) != xmax)
           return FAIL;
         else if(fflush(work) != 0)return FAIL;
       }
       else{
         if(fwrite((char *)LineBuffer,sizeof(short),Nt,work) != Nt)
           return FAIL;
         else if(fflush(work) != 0)return FAIL;
       }
     }
   }
   else{
     if(fwrite(Red  ,sizeof(unsigned char),xmax,work) != xmax)return FAIL;
     if(fwrite(Green,sizeof(unsigned char),xmax,work) != xmax)return FAIL;
     if(fwrite(Blue ,sizeof(unsigned char),xmax,work) != xmax)return FAIL;
     /* add write of alpha channel here */
     if(fflush(work) != 0)return FAIL;
   }
 }
 else if(aa == LOW){
   for (j = XMIN+1,k=0; j < XMAX ; j += 2,k++) {
     Red[k] = (unsigned char)(((long)Red[j]+(long)Red[j-1])/2);
     Green[k] = (unsigned char)(((long)Green[j]+(long)Green[j-1])/2);
     Blue[k] = (unsigned char)(((long)Blue[j]+(long)Blue[j-1])/2);
     Alfa[k] = (unsigned char)(((long)Alfa[j]+(long)Alfa[j-1])/2);
   }
   goto OUTPUT;
 }
 else if(aa == MEDIUM){
   if(scanline & 1){
     for (j = XMIN+1,k=0; j < XMAX ; j += 2,k++) {
       Red[k] = (unsigned char)(((long)Red[j]
              +(long)Red[j-1]+(long)L_Red[j]+(long)L_Red[j-1])/4);
       Green[k] = (unsigned char)(((long)Green[j]
              +(long)Green[j-1]+(long)L_Green[j]+(long)L_Green[j-1])/4);
       Blue[k] = (unsigned char)(((long)Blue[j]
              +(long)Blue[j-1]+(long)L_Blue[j]+(long)L_Blue[j-1])/4);
       Alfa[k] = (unsigned char)(((long)Alfa[j]
              +(long)Alfa[j-1]+(long)L_Alfa[j]+(long)L_Alfa[j-1])/4);
     }
     goto OUTPUT;
   }
   else{
     memcpy(L_Red,Red,XMAX);
     memcpy(L_Green,Green,XMAX);
     memcpy(L_Blue,Blue,XMAX);
     memcpy(L_Alfa,Alfa,XMAX);
   }
 }
 else if(aa == HIGH){
   if(scanline & 1){
     for (j = XMIN+2,k=0; j < XMAX ; j += 3,k++) {
       Red[k]   = (unsigned char)(0.5+
           ((double)Red[j]+(double)Red[j-1]+(double)Red[j-2]+
            (double)L_Red[j]+(double)L_Red[j-1]+(double)L_Red[j-2])/6.0);
       Green[k]   = (unsigned char)(0.5+
           ((double)Green[j]+(double)Green[j-1]+(double)Green[j-2]+
            (double)L_Green[j]+(double)L_Green[j-1]+(double)L_Green[j-2])/6.0);
       Blue[k]   = (unsigned char)(0.5+
           ((double)Blue[j]+(double)Blue[j-1]+(double)Blue[j-2]+
            (double)L_Blue[j]+(double)L_Blue[j-1]+(double)L_Blue[j-2])/6.0);
       Alfa[k]   = (unsigned char)(0.5+
           ((double)Alfa[j]+(double)Alfa[j-1]+(double)Alfa[j-2]+
            (double)L_Alfa[j]+(double)L_Alfa[j-1]+(double)L_Alfa[j-2])/6.0);
     }
     goto OUTPUT;
   }
   else{
     memcpy(L_Red,Red,XMAX);
     memcpy(L_Green,Green,XMAX);
     memcpy(L_Blue,Blue,XMAX);
     memcpy(L_Alfa,Alfa,XMAX);
   }
 }
 else if(aa == ULTRA){
   if(scanline%3 == 2){
     for (j = XMIN+2,k=0; j < XMAX ; j += 3,k++) {
       Red[k]   = (unsigned char)(0.5+
           ((double)Red[j]+(double)Red[j-1]+(double)Red[j-2]+
            (double)L_Red[j]+(double)L_Red[j-1]+(double)L_Red[j-2]+
            (double)LL_Red[j]+(double)LL_Red[j-1]+(double)LL_Red[j-2])/9.0);
       Green[k]   = (unsigned char)(0.5+
           ((double)Green[j]+(double)Green[j-1]+(double)Green[j-2]+
            (double)L_Green[j]+(double)L_Green[j-1]+(double)L_Green[j-2]+
            (double)LL_Green[j]+(double)LL_Green[j-1]+(double)LL_Green[j-2])/9.0);
       Blue[k]   = (unsigned char)(0.5+
           ((double)Blue[j]+(double)Blue[j-1]+(double)Blue[j-2]+
            (double)L_Blue[j]+(double)L_Blue[j-1]+(double)L_Blue[j-2]+
            (double)LL_Blue[j]+(double)LL_Blue[j-1]+(double)LL_Blue[j-2])/9.0);
       Alfa[k]   = (unsigned char)(0.5+
           ((double)Alfa[j]+(double)Alfa[j-1]+(double)Alfa[j-2]+
            (double)L_Alfa[j]+(double)L_Alfa[j-1]+(double)L_Alfa[j-2]+
            (double)LL_Alfa[j]+(double)LL_Alfa[j-1]+(double)LL_Alfa[j-2])/9.0);
     }
     goto OUTPUT;
   }
   else{
     if(scanline%3 == 1){
       memcpy(L_Red,Red,XMAX);
       memcpy(L_Green,Green,XMAX);
       memcpy(L_Blue,Blue,XMAX);
       memcpy(L_Alfa,Alfa,XMAX);
     }
     else if(scanline%3 == 0){
       memcpy(LL_Red,Red,XMAX);
       memcpy(LL_Green,Green,XMAX);
       memcpy(LL_Blue,Blue,XMAX);
       memcpy(LL_Alfa,Alfa,XMAX);
     }
   }
 }
 return 0;
}

void ScanlineToZbuffer(long k, double *Zdepth){
 float *buf;
 long i;
 if     (anti_alias == ULTRA){
   if(k%3 != 0)return;
   k=k/3;
 }
 else if(anti_alias == MEDIUM || anti_alias == HIGH){
   if(k%2 != 0)return;
   k=k/2;
 }
 buf=(fszBuffer+(long)ResolutionX*k);
 for(i=0;i<XMAX;i+=aaXstep) *buf++ = (float)Zdepth[i];
}

void ScanlineToZobject(long k, long *Zobject){
 unsigned char *buf;
 long i;
 if     (anti_alias == ULTRA){
   if(k%3 != 0)return;
   k=k/3;
 }
 else if(anti_alias == MEDIUM || anti_alias == HIGH){
   if(k%2 != 0)return;
   k=k/2;
 }
 buf=(ObjectIdBuffer+(long)ResolutionX*k);
 for(i=0;i<XMAX;i+=aaXstep) *buf++ = (unsigned char)Zobject[i];
}

void FillAlfaLine(UCHAR *Alfa, long *ZglassO[], long *Zobject){
 long i;
 for(i=XMIN;i<XMAX;i++){
   if(Zobject[i] == -2){
     Alfa[i]=0;  /* the sky */
     if(ZglassO[0][i] >= 0){ /* glass over sky */
       Alfa[i] = 128;
     }
   }
   else{
     if(alfagroundON || Zobject[i] != -1)Alfa[i]=255;
   }
 }
 return;
}

// end of anti-aliasing

// code to read and uncompress the temporary file format

#ifdef _SUNSTYLE
void  UnCompressWordLine(bufsize,buffer,output)
short bufsize;short *buffer; unsigned short *output;{
#else
void  UnCompressWordLine(short bufsize, short *buffer,unsigned short *output){
#endif
 short p,k;
 long i;
 k=0; while(k < bufsize){
   p = *buffer++; k++;
   if(p < 0)
     for(i=0; i<= -p; i++){*output++ = *buffer++; k++;}
   else{
     for(i=0; i<=  p; i++) *output++ = *buffer;
     buffer++; k++;
   }
 }
}

#ifdef _SUNSTYLE
static long distance(argb,brgb) unsigned short argb; unsigned short brgb; {
#else
static long distance(unsigned short argb, unsigned short brgb){
#endif
    long b,g,r;
    b  = (long)( argb        & 0x1f);
    g  = (long)((argb >>  5) & 0x1f);
    r  = (long)((argb >> 10) & 0x1f);
    b -= (long)( brgb        & 0x1f);
    g -= (long)((brgb >>  5) & 0x1f);
    r -= (long)((brgb >> 10) & 0x1f);
    return (r*r + g*g + b*b);
/*    return labs(r) + labs(g) + labs(b);
*/
}

#ifdef _SUNSTYLE
long getcolor(crgb, Npalette) unsigned short crgb; long Npalette; {
#else
long getcolor(unsigned short crgb, long Npalette){
#endif
  long i,j,best,dist;
/* look for an exact match in the color table (or minimal distance) */
    for (i=0; i < Npalette; i++) {
       if (colorstats1[i] == crgb) return (i);
       if (i == 0) {
         best = 0;
         dist = distance(colorstats1[i],crgb);
       }
       else if ((j=distance(colorstats1[i],crgb)) < dist) {
         best = i;
         dist = j;
       }
    }
    return(best);
}

#ifdef _SUNSTYLE
void SortColours(Ngif,N_colours) long Ngif; long N_colours; {
#else
void SortColours(long Ngif, long N_colours){
#endif
 long i,j,gap,t0,maxdis,used;
 unsigned short t1;
 for (i = 0; i < N_colours; i++) {
     colorstats1[i] = i;
 }
 for (gap = N_colours/2; gap > 0; gap /=2) {
  for (i = gap; i < N_colours; i++) {
    for (j = i-gap; j >= 0 &&
        colorstats0[j] < colorstats0[j+gap]; j -= gap) {
      t0 = colorstats0[j];
      t1 = colorstats1[j];
      colorstats0[j] = colorstats0[j+gap];
      colorstats1[j] = colorstats1[j+gap];
      colorstats0[j+gap] = t0;
      colorstats1[j+gap] = t1;
    }
  }
 }
 for (used = 0; used < N_colours && colorstats0[used] > 0; used++){;}
 if (used > Ngif) {
  for (maxdis = 2; maxdis < 1024; maxdis *= 2) {
   for (i=Ngif/2; i < Ngif; i++) {
    for (j = 0; j < i; j++) {
      if (distance(colorstats1[i],colorstats1[j])<maxdis) {
       for (j=i+1; j< used; j++) {
         colorstats0[j-1] = colorstats0[j];
         colorstats1[j-1] = colorstats1[j];
       }
       --used;
       --i;
       break;
      }
    }
    if (used <= Ngif) break;
   }
   if (used <= Ngif) break;
  }
 }
 return;
}

// end of read and uncompress code

#ifdef _SUNSTYLE
void OneGifPalette(firstframe,lastframe,file_root,iLace)
long firstframe; long lastframe; char *file_root; BOOL iLace; {
#else
void OneGifPalette(long firstframe, long lastframe, char *file_root, BOOL iLace){
#endif
 unsigned short *RGB;
 short Nt,*LineBuffer;
 char filename[128],FileName[128],messages[256],format[64];
 long  frame,x_x,y_y;
 unsigned short val;
 long  i,j,cpix,used;
 RGB=RendererUshortBuffer;
 LineBuffer=RendererShortBuffer;
 SortColours(GIFCOLOURS,nCOLOUR);
#if __NT__
 WindowedRenderMessage(WRM_COLOURMAP);
#else
 Render_Message(71,0,NULL);
#endif
 /* set up the colour map                */
 for(i=0;i<GIFCOLOURS;i++){
   GifColourMap[i].Red   = (((colorstats1[i] >> 10) & 0x1F) << 3);
   GifColourMap[i].Green = (((colorstats1[i] >>  5) & 0x1F) << 3);
   GifColourMap[i].Blue  = (((colorstats1[i]      ) & 0x1F) << 3);
 }
/* find the addresses of all colours and build LUT  for > 1 frame or 200 c */
 for(i=0;i<nCOLOUR;i++)  colorstats0[i] = GIFCOLOURS+1;
#if __NT__
 WindowedRenderMessage(WRM_WRITEGIF);
#else
 Render_Message(72,0,NULL);
#endif
 for(frame=firstframe; frame<lastframe; frame++){
   if(nScriptframes == 1){
     sprintf(filename,"%s.16b",file_root);
     sprintf(FileName,"%s.gif",file_root);
   }
   else{
     sprintf(format,"%s%ld%s","%s%0",tzeros,"d.16b");
     sprintf(filename,format,file_root,frame);
     sprintf(format,"%s%ld%s","%s%0",tzeros,"d.gif");
     sprintf(FileName,format,file_root,frame);
   }
   if((work=fopen((char *)strlwr(filename),"rb")) == NULL){
     Render_Message(73,1,filename);
     return;
   }
#if __NT__
   WindowedRenderMessage(WRM_WRITEGIF);
   sprintf(messages,"%s -> %s",filename,FileName);
#else
   sprintf(messages,"Procesing %s -> %s",filename,FileName);
#endif
   Render_Message(91,0,messages);
   if( fread((char *)&x_x,sizeof(long),1,work) != 1 ||
       fread((char *)&y_y,sizeof(long),1,work) != 1)goto FATALREAD;
   ScreenXMax = ResolutionX;
   ScreenYMax = ResolutionY;
   ScreenColorBits = 8;
   if((ScanLine=(PixelType *)
       X__Malloc(sizeof(PixelType)*ScreenXMax*ScreenYMax)) == NULL){
     Render_Message(9,0,NULL);
     return;
   }
   if((GifFile=EGifOpenFileName((char *)strlwr(FileName),FALSE,0)) == NULL){
     X__Free((char *) ScanLine);
     Render_Message(74,0,NULL);
     return;
   }
   if (EGifPutScreenDesc(GifFile, ScreenXMax, ScreenYMax, ScreenColorBits,
        0, ScreenColorBits, GifColourMap) == ERROR) {
     X__Free((char *) ScanLine);
     return;
   }
   if (EGifPutImageDesc(GifFile, 0, 0, ScreenXMax, ScreenYMax,
                        iLace,
                        1,NULL) == ERROR) {
     X__Free((char *) ScanLine);
     return;
   }
   for(i=YMIN;i<YMAX;i++) {
     if(fread((char *)&Nt,sizeof(short),1,work) != 1)goto FATALREAD;
     if(Nt == 0){
       if(fread((char *)RGB,sizeof(unsigned short),ResolutionX,work)
          != ResolutionX)goto FATALREAD;
     }
     else{
       if(fread((char *)LineBuffer,sizeof(short),Nt,work) != Nt)goto FATALREAD;
       UnCompressWordLine(Nt,LineBuffer,RGB);
     }
     for(j = 0;j<ResolutionX;j++){
       val=RGB[j];
       if((cpix = (long)colorstats0[(long)val]) > GIFCOLOURS){
         cpix=getcolor(val,GIFCOLOURS);
         colorstats0[(long)val]=(unsigned long)cpix;
       }
       *(ScanLine + i*ResolutionX +j) = (PixelType)cpix;
     }
   }
   if(iLace){
     for(i=0;i<ScreenYMax;i+=8)
       EGifPutLine(GifFile,(ScanLine+i*ScreenXMax),ScreenXMax);
     for(i=4;i<ScreenYMax;i+=8)
       EGifPutLine(GifFile,(ScanLine+i*ScreenXMax),ScreenXMax);
     for(i=2;i<ScreenYMax;i+=4)
       EGifPutLine(GifFile,(ScanLine+i*ScreenXMax),ScreenXMax);
     for(i=1;i<ScreenYMax;i+=2)
       EGifPutLine(GifFile,(ScanLine+i*ScreenXMax),ScreenXMax);
   }
   else{
     for(i=0;i<ScreenYMax;i++){
       if(EGifPutLine(GifFile,(ScanLine+i*ScreenXMax),ScreenXMax) == ERROR) {
         X__Free((char *) ScanLine);
         return;
       }
     }
   }
   if(EGifCloseFile(GifFile) == ERROR) {
     X__Free((char *) ScanLine);
     Render_Message(75,1,NULL);
     return;
   }
   X__Free((char *) ScanLine);
   fclose(work);
   remove(filename); /* delete 16 bit files as they are processed */
 }
 return;
 FATALREAD:
 Render_Message(76,1,NULL);
 return;
}

void OneAlfaSeparate(long frame, char *rawfile, char *out_root){
 long i,j;
 fullscreenbuffer *f;
 char outputfile[256],format[64];
 sprintf(format,"%s%ld%s","%s%0",tzeros,"d.gif");
 if(nScriptframes > 1)
   sprintf(outputfile,format,out_root,frame);
 else
   sprintf(outputfile,"%s.GIF",out_root);
 for(i=0;i<GIFCOLOURS;i++){
   GifColourMap[i].Red   = (unsigned char)i;
   GifColourMap[i].Green = (unsigned char)i;
   GifColourMap[i].Blue  = (unsigned char)i;
 }
 ScreenXMax = ResolutionX;
 ScreenYMax = ResolutionY;
 ScreenColorBits = 8;
 if ((ScanLine = (PixelType *) X__Malloc(sizeof(PixelType) * ScreenXMax))
    == NULL){
   Render_Message(9,0,NULL);
   return;
 }
 if((GifFile=EGifOpenFileName((char *)outputfile,FALSE,0)) == NULL){
   X__Free((char *) ScanLine);
   Render_Message(74,0,NULL);
   return;
 }
 if (EGifPutScreenDesc(GifFile, ScreenXMax, ScreenYMax, ScreenColorBits,
   0, ScreenColorBits, GifColourMap) == ERROR) {
   X__Free((char *) ScanLine);
   return;
 }
 if (EGifPutImageDesc(GifFile, 0, 0, ScreenXMax, ScreenYMax, FALSE, 1,
   NULL) == ERROR) {
   X__Free((char *) ScanLine);
   return;
 }
 f=FullScreenBuffer;
 for(i=0;i<ResolutionY;i++){
   for(j=0;j<ResolutionX;j++){
     ScanLine[j] = f->A; f++;
   }
   if(EGifPutLine(GifFile,ScanLine,ScreenXMax) == ERROR){
     X__Free((char *) ScanLine);
     return;
   }
 }
 if(EGifCloseFile(GifFile) == ERROR) {
   X__Free((char *) ScanLine);
   Render_Message(75,1,NULL);
   return;
 }
 X__Free((char *) ScanLine);
}

static UCHAR *screen;
static BOOL  gif_iLace;
static long  gif_row,gif_inc,gif_width,gif_height;

#ifdef _SUNSTYLE
long out_line(pixels,linelen) unsigned char *pixels; long linelen; {
#else
long out_line (unsigned char *pixels, long linelen){
#endif
 register long x;
 memcpy((screen + gif_row*gif_width),pixels,gif_width);
 gif_row+=abs(gif_inc);
 if(gif_row >= gif_height){
  if     (gif_inc == -8){gif_inc=8; gif_row=4;}
  else if(gif_inc ==  8){gif_inc=4; gif_row=2;}
  else if(gif_inc ==  4){gif_inc=2; gif_row=1;}
 }
 return 0;
}


long SetupImageMap(imap *Map, char *objectfilename){
 RenderIMAP temp_map,*map;
 long success;
 map=&temp_map;
 Map->image   = NO;  Map->irefl = Map->ibump = Map->itran = NO;
 Map->map     = PLANE;
 Map->angle   = 1.0;
 Map->moziac  = NO;
 Map->depth   = 8; Map->drefl = Map->dbump = Map->dtran = 8;
 Map->picture = NULL; Map->prefl = Map->pbump = Map->ptran = NULL;
 Map->palR     = Map->palG     = Map->palB     = NULL;
 Map->palRrefl = Map->palGrefl = Map->palBrefl = NULL;
 Map->palRbump = Map->palGbump = Map->palBbump = NULL;
 Map->palRtran = Map->palGtran = Map->palBtran = NULL;
 Map->p24R     = Map->p24G     = Map->p24B     = NULL;
 Map->p24Rrefl = Map->p24Grefl = Map->p24Brefl = NULL;
 Map->p24Rbump = Map->p24Gbump = Map->p24Bbump = NULL;
 Map->p24Rtran = Map->p24Gtran = Map->p24Btran = NULL;
 success=1;
 if(Map->pp >= 0){
  if(GetRenderImage(Map,Map->filename,objectfilename,map) == 1){
    Map->image=map->image;
    Map->depth=map->depth;
    Map->picture=map->p;
    Map->palR=map->pR;   Map->palG=map->pG;   Map->palB=map->pB;
    Map->p24R=map->p24R; Map->p24G=map->p24G; Map->p24B=map->p24B;
    Map->xmax=map->xmax;
    Map->ymax=map->ymax;
  }
  else success=0;
 }
 if(Map->rp >= 0){
  if(GetRenderImage(Map,Map->frefl,objectfilename,map) == 1){
    Map->irefl=map->image;
    Map->drefl=map->depth;
    Map->prefl=map->p;
    Map->palRrefl=map->pR;
    Map->palGrefl=map->pG;
    Map->palBrefl=map->pB;
    Map->p24Rrefl=map->p24R;
    Map->p24Grefl=map->p24G;
    Map->p24Brefl=map->p24B;
    Map->xmaxrefl=map->xmax;
    Map->ymaxrefl=map->ymax;
  }
  else success=0;
 }
 if(Map->bp >= 0){
  if(GetRenderImage(Map,Map->fbump,objectfilename,map) == 1){
    Map->ibump=map->image;
    Map->dbump=map->depth;
    Map->pbump=map->p;
    Map->palRbump=map->pR;
    Map->palGbump=map->pG;
    Map->palBbump=map->pB;
    Map->p24Rbump=map->p24R;
    Map->p24Gbump=map->p24G;
    Map->p24Bbump=map->p24B;
    Map->xmaxbump=map->xmax;
    Map->ymaxbump=map->ymax;
  }
  else success=0;
 }
 if(Map->tp >= 0){
  if(GetRenderImage(Map,Map->ftran,objectfilename,map) == 1){
    Map->itran=map->image;
    Map->dtran=map->depth;
    Map->ptran=map->p;
    Map->palRtran=map->pR;
    Map->palGtran=map->pG;
    Map->palBtran=map->pB;
    Map->p24Rtran=map->p24R;
    Map->p24Gtran=map->p24G;
    Map->p24Btran=map->p24B;
    Map->xmaxtran=map->xmax;
    Map->ymaxtran=map->ymax;
  }
  else success=0;
 }
 if(success == 0)FreeRenderBrush(Map);
 return success;
}

static long FoundExistingMap(imap *tMap, char *fn, int is24,
                             RenderIMAP *map){
 char map_no[10];
 if(tMap->pp >= 0 && strcmp(tMap->filename,fn) == 0){
   sprintf(map_no," ");  Render_Message(83,0,map_no);
   if(is24){
     map->p24R = tMap->p24R;
     map->p24G = tMap->p24G;
     map->p24B = tMap->p24B;
     map->depth = 24;
   }
   else{
     map->pR = tMap->palR;
     map->pG = tMap->palG;
     map->pB = tMap->palB;
     map->p  = tMap->picture;
   }
   map->xmax = tMap->xmax;
   map->ymax = tMap->ymax;
   return 1;
 }
 if(tMap->rp >= 0 && strcmp(tMap->frefl,fn) == 0){
   sprintf(map_no," ");  Render_Message(83,0,map_no);
   if(is24){
     map->p24R = tMap->p24Rrefl;
     map->p24G = tMap->p24Grefl;
     map->p24B = tMap->p24Brefl;
     map->depth = 24;
   }
   else{
     map->pR = tMap->palRrefl;
     map->pG = tMap->palGrefl;
     map->pB = tMap->palBrefl;
     map->p  = tMap->prefl;
   }
   map->xmax = tMap->xmaxrefl;
   map->ymax = tMap->ymaxrefl;
   return 1;
 }
 if(tMap->bp >= 0 && strcmp(tMap->fbump,fn) == 0){
   sprintf(map_no," ");  Render_Message(83,0,map_no);
   if(is24){
     map->p24R = tMap->p24Rbump;
     map->p24G = tMap->p24Gbump;
     map->p24B = tMap->p24Bbump;
     map->depth = 24;
   }
   else{
     map->pR = tMap->palRbump;
     map->pG = tMap->palGbump;
     map->pB = tMap->palBbump;
     map->p  = tMap->pbump;
   }
   map->xmax = tMap->xmaxbump;
   map->ymax = tMap->ymaxbump;
   return 1;
  }
 if(tMap->tp >= 0 && strcmp(tMap->ftran,fn) == 0){
   sprintf(map_no," ");  Render_Message(83,0,map_no);
   if(is24){
     map->p24R = tMap->p24Rtran;
     map->p24G = tMap->p24Gtran;
     map->p24B = tMap->p24Btran;
     map->depth = 24;
   }
   else{
     map->pR = tMap->palRtran;
     map->pG = tMap->palGtran;
     map->pB = tMap->palBtran;
     map->p  = tMap->ptran;
   }
   map->xmax = tMap->xmaxtran;
   map->ymax = tMap->ymaxtran;
   return 1;
  }
  return 0;
}

#ifdef _SUNSTYLE
static long GetRenderImage(Map,mfname,ofname,map) imap *Map;
char *mfname,*ofname; RenderIMAP *map; {
#else
static long GetRenderImage(imap *Map, char *mfname, char *ofname,
                           RenderIMAP *map){
#endif
   imap *tMap;
   long i,j,status,is24=0,iscel=0,isbmp=0,is8=0;
   char last4[8],*fn;
   map->p=NULL; map->image=NO; map->depth=8;
   map->pR=NULL; map->pG=NULL; map->pB=NULL;
   map->p24R=NULL; map->p24G=NULL; map->p24B=NULL;
   fn=(char *)strlwr(mfname);
   strncpy(last4,(fn+strlen(fn)-4),4);
   if(strncmp((char *)last4,".gif",4) == 0)is8=1;
   if(strncmp((char *)last4,".tga",4) == 0)is24=1;
   if(strncmp((char *)last4,".cel",4) == 0)iscel=1;
   if(strncmp((char *)last4,".bmp",4) == 0){is24=1; isbmp=1;}
   if(R_Nground > 0 && Map != &Ground.map &&
      FoundExistingMap(&Ground.map,fn,is24,map))return 1;
   if(Nsky > 0 && Map != &Sky.map &&
      FoundExistingMap(&Sky.map,fn,is24,map))return 1;
   if(ObjectCount > 0)for(i=0;i<ObjectCount;i++){
     if(Object[i].NoMaps > 0)for(j=0;j<Object[i].NoMaps;j++){
       tMap=(Object[i].Mbase+j);
       if(tMap != Map && FoundExistingMap(tMap,fn,is24,map))return 1;
     }
   }
   if(isbmp)return LoadBmpBrush(mfname,ofname,map);
   if(is24)return Load24bitBrush(mfname,ofname,map);
   else if(iscel)return LoadCelBrush(mfname,ofname,map);
   if(is8)return Load8bitBrush(mfname,ofname,map);
   return LoadExternalBrush(mfname,ofname,map);
}

static HANDLE Bit_Handle;
static unsigned char *Bit_Handle_Buffer,*Bit_Handle_Pointer;

long get_byte(void){
  return (long)(*Bit_Handle_Pointer++);
}

#ifdef _SUNSTYLE
static long Load8bitBrush(fn,ofname,map) long N;
char *fn,*ofname; RenderIMAP *map; {
#else
static long Load8bitBrush(char *fn, char *ofname,
                           RenderIMAP *map){
#endif
   DWORD file_size,dwRead;
   long i,j,status;
   unsigned finished, planes;
   unsigned char buffer[16];
   char local_message[256];
   if((map->pR=(unsigned char *)X__Malloc(256)) == NULL)
     {Render_Message(9,0,NULL); return(0);}
   if((map->pG=(unsigned char *)X__Malloc(256)) == NULL)
     {Render_Message(9,0,NULL); return(0);}
   if((map->pB=(unsigned char *)X__Malloc(256)) == NULL)
     {Render_Message(9,0,NULL); return(0);}
   status = 0;
   sprintf(local_message," %s",fn);
   Render_Message(84,0,local_message);
   if((Bit_Handle=GetRenderRelativeHandle(fn,ofname,NULL)) == NULL){
     if((fn=GetRenderReplacementFilename(fn)) == NULL)goto BAD1;
     Bit_Handle=CreateFile(fn,GENERIC_READ,
          FILE_SHARE_READ,NULL,OPEN_EXISTING,
          FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
     if(Bit_Handle == INVALID_HANDLE_VALUE)goto BAD1;
   }
   file_size=GetFileSize(Bit_Handle,NULL);
   Bit_Handle_Buffer=X__Malloc(file_size);
   if(Bit_Handle_Buffer == NULL)goto BAD;
   Bit_Handle_Pointer=Bit_Handle_Buffer;
   if(!ReadFile(Bit_Handle,Bit_Handle_Buffer,file_size,&dwRead,NULL))goto BAD;
   if(dwRead != file_size)goto BAD;
   if((decoderline = (unsigned char *) X__Malloc (2049)) == NULL)goto BAD;
   for (i=0;i<2049;i++)decoderline[i] = (unsigned char) 0;
   /* Get the screen description */
   gif_row=0; gif_inc=1;
   for (i = 0; i < 13; i++)buffer[i] = (unsigned char)get_byte();
   if (strncmp((char *)buffer,"GIF87a",3) ||
       buffer[3] < '0' || buffer[3] > '9' ||
       buffer[4] < '0' || buffer[4] > '9' ||
       buffer[5] < 'A' || buffer[5] > 'z' )goto BAD;
   planes = ((unsigned)buffer[10] & 0x0F) + 1;
   colourmap_size = (int)(1 << planes);
   if ((buffer[10] & 0x80) == 0)goto BAD;
   for (i = 0; i < colourmap_size ; i++) {
     map->pR[i]  = get_byte();
     map->pG[i]  = get_byte();
     map->pB[i]  = get_byte();
   }
   finished = FALSE;
   while (!finished) {
      switch (get_byte()) {
         case ';':                /* End of the GIF dataset */
            finished = TRUE;
            status = 0;
            break;
         case '!':                /* GIF Extension Block */
            get_byte();           /* read (and ignore) the ID */
            while ((i = get_byte()) > 0) /* get data len*/
            for (j = 0; j < i; j++)
                get_byte(); /* flush data */
            break;
         case ',': /* Start of image object. get description */
            for (i = 0; i < 9; i++) {
              buffer[i]=(unsigned char)get_byte();
            }
            if (status < 0) {
               finished = TRUE;
               break;
            }
            if((buffer[8] & 0x40) == 0x40)gif_iLace=TRUE;
            else                          gif_iLace=FALSE;
            if(gif_iLace)gif_inc = -8;
            map->xmax = (short)(buffer[4] | (buffer[5] << 8));
            map->ymax = (short)(buffer[6] | (buffer[7] << 8));
            brush_size=(long)(map->xmax)*(long)(map->ymax);
            gif_width=map->xmax;
            gif_height=map->ymax;
#ifdef __DEMO__
//            if(brush_size > 640L*480L)goto BAD;
#endif
            if((map->p=(unsigned char *)X__Malloc(brush_size)) == NULL)
              goto BAD;
            screen=map->p;
            if(map->xmax > 2048 || map->ymax > 1024)goto BAD;
            status = GIF_decoder((short)(map->xmax)); /*put bytes in Buf*/
            finished = TRUE;
            break;
         default:
            status = -1;
            finished = TRUE;
            break;
      }
   }
   X__Free(decoderline);
   X__Free(Bit_Handle_Buffer);
   CloseHandle(Bit_Handle);
   map->image=YES;
   return 1;
   BAD:
   if(Bit_Handle_Buffer != NULL)X__Free(Bit_Handle_Buffer);
   CloseHandle(Bit_Handle);
   BAD1:
   Render_Message(9,0,NULL);
   X__Free(map->pR); X__Free(map->pG); X__Free(map->pB);
   return 0;
}

void FreeRenderBrush(imap *Map){
 Map->filename[0]='\0'; Map->frefl[0]='\0';
 Map->fbump[0]='\0'; Map->ftran[0]='\0';
 if(Map->image == YES){
   if(Map->depth == 8){
     X__Free(Map->palR); X__Free(Map->palG); X__Free(Map->palB);
     X__Free(Map->picture);
   }
   else{
     X__Free(Map->p24R); X__Free(Map->p24G); X__Free(Map->p24B);
   }
   Map->image=NO;
 }
 if(Map->irefl == YES){
   if(Map->drefl == 8){
     X__Free(Map->palRrefl); X__Free(Map->palGrefl); X__Free(Map->palBrefl);
     X__Free(Map->prefl);
   }
   else{
     X__Free(Map->p24Rrefl); X__Free(Map->p24Grefl); X__Free(Map->p24Brefl);
   }
   Map->irefl=NO;
 }
 if(Map->ibump == YES){
   if(Map->dbump == 8){
     X__Free(Map->palRbump); X__Free(Map->palGbump); X__Free(Map->palBbump);
     X__Free(Map->pbump);
   }
   else{
     X__Free(Map->p24Rbump); X__Free(Map->p24Gbump); X__Free(Map->p24Bbump);
   }
   Map->ibump=NO;
 }
 if(Map->itran == YES){
   if(Map->dtran == 8){
     X__Free(Map->palRtran); X__Free(Map->palGtran); X__Free(Map->palBtran);
     X__Free(Map->ptran);
   }
   else{
     X__Free(Map->p24Rtran); X__Free(Map->p24Gtran); X__Free(Map->p24Btran);
   }
   Map->itran=NO;
 }
}

#ifdef _SUNSTYLE
void OneIffPalette(firstframe,lastframe,file_root,ham) long firstframe;
long lastframe; char *file_root; short ham; {;}
void Iff24Palette(firstframe,lastframe,file_root) long firstframe;
long lastframe; char *file_root; {;}
#else
void OneIffPalette(long firstframe, long lastframe, char *file_root,
                   short ham){;}
void Iff24Palette(long firstframe, long lastframe, char *file_root){;}
#endif

// code for reading 24 bit images

#ifdef _SUNSTYLE
static short R_getvariword(n,f,vw) long n; FILE *f; long *vw; {
#else
static short R_getvariword(long n, FILE *f, long *vw){
#endif
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;         if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

#ifdef _SUNSTYLE
static short R_putlong(lng,f) long lng; FILE *f; {
#else
static short R_putlong(long lng, FILE *f){
#endif
 if(fputc((char)(lng      ),f) == EOF)return FAIL; /* must be INTEL ordering */
 if(fputc((char)(lng >>  8),f) == EOF)return FAIL;
 if(fputc((char)(lng >> 16),f) == EOF)return FAIL;
 if(fputc((char)(lng >> 24),f) == EOF)return FAIL;
 return OK;
}

#ifdef _SUNSTYLE
static short R_putshort(sht,f) short sht; FILE *f;{
#else
static short R_putshort(short sht, FILE *f){
#endif
 if(fputc((char)(sht      ),f) == EOF)return FAIL; /* must be INTEL ordering */
 if(fputc((char)(sht >>  8),f) == EOF)return FAIL;
 return OK;
}

#ifdef _SUNSTYLE
static short R_putbyte(c,f) char c; FILE *f;{
#else
static short R_putbyte(char c ,FILE *f){
#endif
 if(fputc((char)(c),f) == EOF)return FAIL;
 return OK;
}

#ifdef _SUNSTYLE
static char *Put2B(buffer,l,key,bufsize) char *buffer; long l;
short key; long *bufsize; {
#else
static char *Put2B(char *buffer, long l, short key, long *bufsize){
#endif
            *bufsize += key;
            *buffer++ = (char)( l        & 0x000000ff);
            *buffer++ = (char)((l >>  8) & 0x000000ff);
 if(key > 2)*buffer++ = (char)((l >> 16) & 0x000000ff);
 if(key > 3)*buffer++ = (char)((l >> 24) & 0x000000ff);
 return buffer;
}

#if __DEMO__

void OneTargaFrame16(long frame, char *rawfile, char *out_root,
                    short truevision, short compress){
  Render_Message(17,0,NULL);
}
void OneTargaFrame32(long frame, char *rawfile, char *out_root,
                    short truevision, short compress){
  Render_Message(17,0,NULL);
}
void OneTiffFrame(long frame, char *rawfile, char *out_root){
 Render_Message(17,0,NULL);
}

#else

#ifdef _SUNSTYLE
void OneTargaFrame32(frame,rawfile,out_root,truevision,compress) long frame;
char *rawfile; char *out_root; short truevision; short compress; {
#else
void OneTargaFrame32(long frame, char *rawfile, char *out_root,
                  short truevision, short compress){
#endif
 fullscreenbuffer *f=NULL;
 FILE *f24=NULL,*t24;
 short iok=0;
 long x,y,i,j,*data_line;
 char outputfile[256],*line_buffer,format[64];
 if((data_line = (long *)X__Malloc(ResolutionX*sizeof(long))) == NULL){
   Render_Message(13,0,NULL);
   return;
 }
 if((line_buffer = (char *)X__Malloc(ResolutionX*2*sizeof(long))) == NULL){
   X__Free(data_line);
   Render_Message(13,0,NULL);
   return;
 }
 if(nScriptframes > 1){
   sprintf(format,"%s%ld%s","%s%0",tzeros,"d.TGA");
   sprintf(outputfile,format,out_root,frame);
 }
 else
   sprintf(outputfile,"%s.TGA",out_root);
 if(FullScreenBuffer == NULL){
   Render_Message(97,0,NULL);
   f24=fopen(rawfile,"rb");
 }
 if(FullScreenBuffer != NULL || f24 != NULL){
   if((t24=fopen(outputfile,"wb")) != NULL){
     if(FullScreenBuffer == NULL){
       if( fread((char *)&x,sizeof(long),1,f24) != 1 ||
           fread((char *)&y,sizeof(long),1,f24) != 1)goto FATALREAD;
     }
     else{
       x=ResolutionX; y=ResolutionY; f=FullScreenBuffer;
       if(truevision)f=f+(y-1)*x;
     }
     R_putbyte(0,t24); R_putbyte(0,t24);                      /* CI CMT */
     if(!compress)R_putbyte(2,t24); else R_putbyte(10,t24);   /* IT */
     R_putshort(0,t24); R_putshort(0,t24), R_putbyte(32,t24); /* CM */
     R_putshort(0,t24); R_putshort((short)(y-1),t24);         /* IS */
     R_putshort((short)x,t24); R_putshort((short)y,t24);      /* IS */
     R_putbyte(32,t24);  /* targa 32 for board speed             IS */
     if(truevision)R_putbyte(8,t24);  /* targa 32 has 8 attribute bits IS */
     else          R_putbyte(32+8,t24);
     for (i=0;i<y;i++){
       if(FullScreenBuffer == NULL){
         if(truevision)fseek(f24,8+(y-1-i)*x*3,SEEK_SET);
         if(fread((char *)gL_Red,sizeof(unsigned char),x,f24)
           != x)goto FATALREAD;
         if(fread((char *)gL_Green,sizeof(unsigned char),x,f24)
           != x)goto FATALREAD;
         if(fread((char *)gL_Blue,sizeof(unsigned char),x,f24)
           != x)goto FATALREAD;
 /* read Alfa line here */
         for (j=0;j<x;j++){
           data_line[j] = (((long)gL_Red[j]) << 16)
                        | (((long)gL_Green[j]) << 8)
                        | (((long)gL_Blue[j]));
         }
       }
       else{
         for (j=0;j<x;j++){
           data_line[j] = (((long)(f->R)) << 16)
                        | (((long)(f->G)) <<  8)
                        | (((long)(f->B))      )
                        | (((long)(f->A)) << 24); /* alfa channel */
           f++;
         }
         if(truevision)f=f-2*x;
       }
       if(!compress){
         for (j=0;j<x;j++){
           if(R_putlong(data_line[j],t24) == FAIL)goto FATALWRITE;
         }
       }
       else{
         long *l,run,raw,bufsize;
         char *p,*buffer,cc;
         run=0; raw=1; bufsize = 1;
         p=buffer=line_buffer; l=data_line; *p=0; buffer=p+1;
         buffer=Put2B(buffer,*l,4,&bufsize); l++;
         for(j=1;j<x;j++){
          if( *l == *(l-1) ){
            if(raw == 1){ /* convert into sames packet */
              *p = -127;
              raw=0; run=2;  l++;
            }
            else if(raw > 1){ /* last two are the same */
              (*p)--; p = (buffer-4); bufsize -= 4;
              *p = -127;
              buffer = p+1; bufsize++;
              buffer=Put2B(buffer,*l,4,&bufsize); l++;
              raw=0; run=2;
            }
            else if(run == 128){ /* over flow */
              p=buffer;
              *p = -128;
              buffer = p+1; bufsize++;
              buffer=Put2B(buffer,*l,4,&bufsize); l++;
              run=1;
            }
            else{
              (*p)++; l++; run++;
            }
          }
          else{               /* not same as last */
            if(run > 0 || raw == 128){ /* was it in run or is packet full */
              p=buffer;
              *p=0; buffer = p+1; bufsize++;
              buffer=Put2B(buffer,*l,4,&bufsize); l++;
              raw=1; run=0;
            }
            else{
              (*p)++; raw++; buffer=Put2B(buffer,*l,4,&bufsize); l++;
            }
          }
          if(bufsize > ResolutionX*8)goto FATALWRITE;
         }
         if(fwrite(line_buffer,1,bufsize,t24) != bufsize ||
            fflush(t24) != 0)goto FATALWRITE;
       }
     }
     goto IOK;
     FATALREAD:
     Render_Message(96,0,NULL);
     IOK:
     iok=1;
     FATALWRITE:
     fclose(t24);
     if(! iok)Render_Message(95,0,NULL);
   }
   else Render_Message(94,0,NULL);
   if(FullScreenBuffer == NULL){
     fclose(f24);
     remove(rawfile);
   }
   if(!iok)remove(outputfile);
 }
 else Render_Message(12,0,NULL);
 X__Free(line_buffer);
 X__Free(data_line);
}

#ifdef _SUNSTYLE
void OneTargaFrame16(frame,rawfile,out_root,truevision,compress) long frame;
char *rawfile; char *out_root; short truevision; short compress; {
#else
void OneTargaFrame16(long frame, char *rawfile, char *out_root,
                  short truevision, short compress){
#endif
 fullscreenbuffer *f=NULL;
 FILE *f24=NULL,*t24;
 short iok=0;
 long x,y,i,j;
 short *data_line;
 char outputfile[256],*line_buffer,format[64];
 if((data_line = (short *)X__Malloc(ResolutionX*sizeof(short))) == NULL){
   Render_Message(13,0,NULL);
   return;
 }
 if((line_buffer = (char *)X__Malloc(ResolutionX*2*sizeof(short))) == NULL){
   X__Free(data_line);
   Render_Message(13,0,NULL);
   return;
 }
 if(nScriptframes > 1){
   sprintf(format,"%s%ld%s","%s%0",tzeros,"d.TGA");
   sprintf(outputfile,format,out_root,frame);
 }
 else
   sprintf(outputfile,"%s.TGA",out_root);
 if(FullScreenBuffer == NULL)f24=fopen(rawfile,"rb");
 if(FullScreenBuffer != NULL || f24 != NULL){
   if((t24=fopen(outputfile,"wb")) != NULL){
     if(FullScreenBuffer == NULL){
       if( fread((char *)&x,sizeof(long),1,f24) != 1 ||
           fread((char *)&y,sizeof(long),1,f24) != 1)goto FATALWRITE;
     }
     else{
       x=ResolutionX; y=ResolutionY; f=FullScreenBuffer;
       if(truevision)f=f+(y-1)*x;
     }
     R_putbyte(0,t24); R_putbyte(0,t24);                      /* CI CMT */
     if(!compress)R_putbyte(2,t24); else R_putbyte(10,t24);   /* IT */
     R_putshort(0,t24); R_putshort(0,t24), R_putbyte(16,t24); /* CM */
     R_putshort(0,t24); R_putshort((short)(y-1),t24);         /* IS */
     R_putshort((short)x,t24); R_putshort((short)y,t24);      /* IS */
     R_putbyte(16,t24);  /* targa 32 for board speed             IS */
     if(truevision)R_putbyte(1,t24);  /* targa 16 has 1 attribute bit IS */
     else          R_putbyte(32+1,t24);
     for (i=0;i<y;i++){
       if(FullScreenBuffer == NULL){
         if(truevision)fseek(f24,8+(y-1-i)*x*3,SEEK_SET);
         if(fread((char *)gL_Red,sizeof(unsigned char),x,f24)
           != x)goto FATALWRITE;
         if(fread((char *)gL_Green,sizeof(unsigned char),x,f24)
           != x)goto FATALWRITE;
         if(fread((char *)gL_Blue,sizeof(unsigned char),x,f24)
           != x)goto FATALWRITE;
         /* read and dump alfa channel */
         for (j=0;j<x;j++){
           data_line[j] = (((short)(gL_Red[j])&0xF8)   << 7)
                        | (((short)(gL_Green[j])&0xF8) << 2)
                        | (((short)(gL_Blue[j])&0xF8)  >> 3);
         }
       }
       else{
         for (j=0;j<x;j++){
           data_line[j] = (((short)(f->R)&0xF8) << 7)
                        | (((short)(f->G)&0xF8) << 2)
                        | (((short)(f->B)&0xF8) >> 3);
           f++;
         }
         if(truevision)f=f-2*x;
       }
       if(!compress){
         for (j=0;j<x;j++){
           if(R_putshort(data_line[j],t24) == FAIL)goto FATALWRITE;
         }
       }
       else{
         long run,raw,bufsize;
         short *l;
         char *p,*buffer,cc;
         run=0; raw=1; bufsize = 1;
         p=buffer=line_buffer; l=data_line; *p=0; buffer=p+1;
         buffer=Put2B(buffer,(long)(*l),2,&bufsize); l++;
         for(j=1;j<x;j++){
          if( *l == *(l-1) ){
            if(raw == 1){ /* convert into sames packet */
              *p = -127;
              raw=0; run=2;  l++;
            }
            else if(raw > 1){ /* last two are the same */
              (*p)--; p = (buffer-2); bufsize -= 2;
              *p = -127;
              buffer = p+1; bufsize++;
              buffer=Put2B(buffer,(long)(*l),2,&bufsize); l++;
              raw=0; run=2;
            }
            else if(run == 128){ /* over flow */
              p=buffer;
              *p = -128;
              buffer = p+1; bufsize++;
              buffer=Put2B(buffer,(long)(*l),2,&bufsize); l++;
              run=1;
            }
            else{
              (*p)++; l++; run++;
            }
          }
          else{               /* not same as last */
            if(run > 0 || raw == 128){ /* was it in run or is packet full */
              p=buffer;
              *p=0; buffer = p+1; bufsize++;
              buffer=Put2B(buffer,(long)(*l),2,&bufsize); l++;
              raw=1; run=0;
            }
            else{
              (*p)++; raw++; buffer=Put2B(buffer,(long)(*l),2,&bufsize); l++;
            }
          }
          if(bufsize > ResolutionX*8)goto FATALWRITE;
         }
         if(fwrite(line_buffer,1,bufsize,t24) != bufsize ||
            fflush(t24) != 0)goto FATALWRITE;
       }
     }
     iok=1;
     FATALWRITE:
     fclose(t24);
     if(! iok)Render_Message(95,0,NULL);
   }
   else Render_Message(94,0,NULL);
   if(FullScreenBuffer == NULL){
     fclose(f24);
     remove(rawfile);
   }
   if(!iok)remove(outputfile);
 }
 else Render_Message(12,0,NULL);
 X__Free(line_buffer);
 X__Free(data_line);
}

#ifdef _SUNSTYLE
void OneTiffFrame(frame,rawfile,out_root) long frame;
char *rawfile; char *out_root; {
#else
void OneTiffFrame(long frame, char *rawfile, char *out_root){
#endif
 fullscreenbuffer *f=NULL;
 FILE *f24,*t24;
 short iok=0;
 long x,y,i,j,p,nh,adr0,adr1,adr2;
 char outputfile[256],format[64];
 if(nScriptframes > 1){
   sprintf(format,"%s%ld%s","%s%0",tzeros,"d.TIF");
   sprintf(outputfile,format,out_root,frame);
 }
 else
   sprintf(outputfile,"%s.TIF",out_root);
 if(FullScreenBuffer == NULL)f24=fopen(rawfile,"rb");
 if(FullScreenBuffer != NULL || f24 != NULL){
   if((t24=fopen(outputfile,"wb")) != NULL){
     if(FullScreenBuffer == NULL){
       if( fread((char *)&x,sizeof(long),1,f24) != 1 ||
           fread((char *)&y,sizeof(long),1,f24) != 1)goto FATALWRITE;
     }
     else{
       x=ResolutionX; y=ResolutionY; f=FullScreenBuffer;
     }
     R_putshort(0x4949,t24);
     R_putshort(0x002a,t24);
     R_putlong(8,t24); /* first IFD pointer imediate */
     nh=13; R_putshort((short)nh,t24); /* entry count */
     adr1=8+2+nh*12+4+6;  /* should = p */
     adr2=adr1+y*sizeof(long);
     adr0=adr1-6;
     R_putshort(0x00ff,t24); R_putshort(3,t24); R_putshort(1,t24);  /* subf */
                             R_putshort(0,t24); R_putshort(1,t24);
                             R_putshort(0,t24);
     R_putshort(0x0100,t24); R_putshort(4,t24); R_putlong(1,t24);  /* width */
                             R_putlong(x,t24);
     R_putshort(0x0101,t24); R_putshort(4,t24); R_putlong(1,t24);  /* lenght */
                             R_putlong(y,t24);
     R_putshort(0x0102,t24); R_putshort(3,t24); R_putshort(3,t24);  /* bps */
                             R_putshort(0,t24); R_putlong(adr0,t24);
     R_putshort(0x0103,t24); R_putshort(3,t24); R_putshort(1,t24);
                             R_putshort(0,t24); R_putshort(1,t24);
                             R_putshort(0,t24);/* comp */
     R_putshort(0x0106,t24); R_putshort(3,t24); R_putshort(1,t24);
                             R_putshort(0,t24); R_putshort(2,t24);
                             R_putshort(0,t24); /* photo */
     R_putshort(0x0111,t24); R_putshort(4,t24); R_putlong(y,t24);  /* S-off */
                             R_putlong(adr1,t24);
     R_putshort(0x0115,t24); R_putshort(3,t24); R_putshort(1,t24);
                             R_putshort(0,t24); R_putshort(3,t24);
                             R_putshort(0,t24); /* spp  */
     R_putshort(0x0116,t24); R_putshort(4,t24); R_putlong(1,t24);  /* Rpstr */
                             R_putlong(1,t24);
     R_putshort(0x0117,t24); R_putshort(4,t24); R_putlong(y,t24);  /* str-c*/
                             R_putlong(adr2,t24);
     R_putshort(0x0118,t24); R_putshort(3,t24); R_putshort(1,t24);
                             R_putshort(0,t24); R_putshort(0,t24);
                             R_putshort(0,t24); /* spp  */
     R_putshort(0x0119,t24); R_putshort(3,t24); R_putshort(1,t24);
                             R_putshort(0,t24); R_putshort(255,t24);
                             R_putshort(0,t24); /* spp  */
     R_putshort(0x011c,t24); R_putshort(3,t24); R_putshort(1,t24);
                             R_putshort(0,t24); R_putshort(1,t24);
                             R_putshort(0,t24); /* spp  */
     R_putlong(0,t24);
     R_putshort(8,t24); R_putshort(8,t24); R_putshort(8,t24); /*b/sample */
     p=ftell(t24);
     for(i=0;i<y;i++)R_putlong(p+2*sizeof(long)*y+i*x*3,t24);
     for(i=0;i<y;i++)R_putlong(x*3,t24);
     for (i=0;i<y;i++){
       if(FullScreenBuffer == NULL){
         if(fread((char *)gL_Red,sizeof(unsigned char),x,f24)
           != x)goto FATALWRITE;
         if(fread((char *)gL_Green,sizeof(unsigned char),x,f24)
           != x)goto FATALWRITE;
         if(fread((char *)gL_Blue,sizeof(unsigned char),x,f24)
           != x)goto FATALWRITE;
         for (j=0;j<x;j++){
           if(R_putbyte(gL_Red[j],t24) == FAIL)goto FATALWRITE;
           if(R_putbyte(gL_Green[j],t24) == FAIL)goto FATALWRITE;
           if(R_putbyte(gL_Blue[j],t24) == FAIL)goto FATALWRITE;
         }
       }
       else{
         for (j=0;j<x;j++){
           if(R_putbyte(f->R,t24) == FAIL)goto FATALWRITE;
           if(R_putbyte(f->G,t24) == FAIL)goto FATALWRITE;
           if(R_putbyte(f->B,t24) == FAIL)goto FATALWRITE;
           f++;
         }
       }
     }
     iok=1;
     FATALWRITE:
     fclose(t24);
     if(! iok)Render_Message(95,0,NULL);
   }
   else Render_Message(94,0,NULL);
   if(FullScreenBuffer == NULL){
     fclose(f24);
     remove(rawfile);
   }
   if(!iok)remove(outputfile);
 }
 else Render_Message(12,0,NULL);
}

#endif     // DEMO

#define UPSIDEDOWN  {l++; if(l == wi){l=0; pr -= wi*2; pg -= wi*2; pb -= wi*2; }}

#ifdef _SUNSTYLE
static long Load24bitBrush(fn,ofname,map)
char *fn,*ofname; RenderIMAP *map; {
#else
static short Load24bitBrush(char *fn, char *ofname, RenderIMAP *map){
#endif
 char c0,c1,c2;
 unsigned char *pr,*pg,*pb;
 long i,j,k,l,n,cmo,cml,xor,yor,wi,hi,iok=0,ialloc=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat;
 FILE *t24;
 Render_Message(61,0,fn);
 if((t24=GetRenderRelativeFile(fn,ofname,NULL)) == NULL){
   if((fn=GetRenderReplacementFilename(fn)) != NULL)t24=fopen(fn,"rb");
 }
 if(t24 != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0:1;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi; map->xmax = (short)wi; map->ymax = (short)hi;
   ialloc=1;
   if((map->p24R = (unsigned char *)X__Malloc(i)) == NULL)goto CLOSEOUT;
   if((map->p24G = (unsigned char *)X__Malloc(i)) == NULL)goto CLOSEOUT;
   if((map->p24B = (unsigned char *)X__Malloc(i)) == NULL)goto CLOSEOUT;
   if(upsidedown){
     pr=map->p24R+wi*(hi-1);
     pg=map->p24G+wi*(hi-1);
     pb=map->p24B+wi*(hi-1);
   }
   else{
     pr=map->p24R; pg=map->p24G; pb=map->p24B;
   }
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   if(image_type == 2){ /* uncompressed */
     l=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       *pb++ = (char)( vw        & pat) << s;
       *pg++ = (char)((vw >> s1) & pat) << s;
       *pr++ = (char)((vw >> s2) & pat) << s;
       if(upsidedown)UPSIDEDOWN
     }
   }
   else{  /* compressed */
     l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           *pb++ = (char)( vw        & pat) << s;
           *pg++ = (char)((vw >> s1) & pat) << s;
           *pr++ = (char)((vw >> s2) & pat) << s;
           k++;
           if(upsidedown)UPSIDEDOWN
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           *pb++ = c0;  *pg++ = c1;  *pr++ = c2; k++;
           if(upsidedown)UPSIDEDOWN
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
 }
 if(iok == 0){
   if(ialloc){
     if(map->p24R != NULL)X__Free(map->p24R); map->p24R = NULL;
     if(map->p24G != NULL)X__Free(map->p24G); map->p24G = NULL;
     if(map->p24B != NULL)X__Free(map->p24B); map->p24B = NULL;
   }
   Render_Message(62,0,fn);
   return 0;
 }
 map->image=YES;
 map->depth=24;
 return 1;
}
#undef UPSIDEDOWN

// end if 24 bit code

// code for mapping

static short LoadExternalBrush(char *fn, char *ofname, RenderIMAP *map){
 long x,y;
 FILE *t24;
 char lname[256],last3[8],dll[256],rep_name[256];
 HMODULE hLib;
 FARPROC fpFun;
 BOOL (*fpGetMapSize)(char *,
                      long *, long *);
 BOOL (*fpGetMap)(char *,
                  long, long,
                  unsigned char *, unsigned char *, unsigned char *);
 Render_Message(61,0,fn);
 strcpy(rep_name,"%%**%%"); //some dummy string that will never be a filename
 if((t24=GetRenderRelativeFile(fn,ofname,rep_name)) == NULL){
   if((fn=GetRenderReplacementFilename(fn)) != NULL){
     t24=fopen(fn,"rb");
   }
 }
 if(t24 == NULL)return 0;
 fclose(t24);
 if(strcmp(rep_name,"%%**%%") != 0){ // string was replaced
   fn=rep_name;
 }
 strcpy(lname,gszHomeDir);
 strncpy(last3,(fn+strlen(fn)-3),3);
 last3[3]=0;  // to ensure end of string null
 sprintf(dll,"%s_map.dll",last3);
 strcat(lname,dll);
 hLib=LoadLibrary(lname);
 if(hLib != NULL){
   if((fpFun=GetProcAddress(hLib,"_GetExternalImageSize")) != NULL){
     fpGetMapSize = (void *)fpFun;
     if((*fpGetMapSize)(fn,&x,&y)){
       if((map->p24R = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((map->p24G = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((map->p24B = (unsigned char *)X__Malloc(x*y)) == NULL)goto CLOSEOUT;
       if((fpFun=GetProcAddress(hLib,"_GetExternalImage")) != NULL){
         fpGetMap = (void *)fpFun;
         if(!(*fpGetMap)(fn,x,y,map->p24R,map->p24G,map->p24B))goto CLOSEOUT;
       }
       else goto CLOSEOUT;
     }
     else goto CLOSEOUT;
   }
   else goto CLOSEOUT;
   FreeLibrary(hLib);
 }
 else return 0;
 map->xmax = (short)x;
 map->ymax = (short)y;
 map->image=YES;
 map->depth=24;
 return 1;
 CLOSEOUT:
 if(map->p24R != NULL)X__Free(map->p24R); map->p24R = NULL;
 if(map->p24G != NULL)X__Free(map->p24G); map->p24G = NULL;
 if(map->p24B != NULL)X__Free(map->p24B); map->p24B = NULL;
 Render_Message(62,0,fn);
 FreeLibrary(hLib);
 return 0;
}

void OneExternalFrame(long frame, char *rawfile, char *out_root,
                      long type, long quality){
 long x,y;
 char lname[256],outputfile[256],dll[256],type_list[16],format[32];
 HMODULE hLib;
 FARPROC fpFun;
 BOOL (*fpPutMap)(char *, long, long, long, fullscreenbuffer *);
 if(FullScreenBuffer == NULL){
   Render_Message(97,0,NULL);
   return;
 }
 if     (type == J24_FILE)strcpy(type_list,"jpg");
 else if(type == B24_FILE)strcpy(type_list,"bmp");
 else if(type == PNG_FILE)strcpy(type_list,"png");
 x=ResolutionX;
 y=ResolutionY;
 strcpy(lname,gszHomeDir);
 sprintf(dll,"%s_map.dll",type_list);
 strcat(lname,dll);
 sprintf(format,"%s%ld%s","%s%0",tzeros,"d.%s");
 if(nScriptframes > 1)
   sprintf(outputfile,format,out_root,frame,type_list);
 else
   sprintf(outputfile,"%s.%s",out_root,type_list);
 hLib=LoadLibrary(lname);
 if(hLib != NULL){
   if((fpFun=GetProcAddress(hLib,"_PutExternalImage")) != NULL){
     fpPutMap = (void *)fpFun;
     if((*fpPutMap)(outputfile,x,y,quality,FullScreenBuffer)){
     }
   }
   FreeLibrary(hLib);
 }
 return;
}

// end of mapping code

// code for BMP files

#if !__NT__

#define DWORD unsigned long
#define WORD  unsigned short
#define BYTE  unsigned char

typedef struct tagRGBQUAD {
  BYTE rgbBlue;
  BYTE rgbGreen;
  BYTE rgbRed;
  BYTE rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPFILEHEADER {
  unsigned short bfType;
  unsigned long  bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned long  bfOffBits;
} BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER {
  DWORD biSize;
  DWORD biWidth;
  DWORD biHeight;
  WORD  biPlanes;
  WORD  biBitCount;
  DWORD biCompression;
  DWORD biSizeImage;
  DWORD biXPelsPerMeter;
  DWORD biYPelsPerMeter;
  DWORD biClrUsed;
  DWORD biClrImportant;
} BITMAPINFOHEADER;

typedef struct tagBITMAPINFO {
  BITMAPINFOHEADER bmiHeader;
  RGBQUAD          bmiColors[1];
} BITMAPINFO;

#define BFT_BITMAP 0x4d42   /* 'BM' */
#define BI_RGB     0L

#endif

#ifdef _SUNSTYLE
static long LoadBmpBrush(N,fn,ofname,map) long N;
char *fn,*ofname; RenderIMAP *map; {
#else
static long LoadBmpBrush(char *fn, char *ofname,
                           RenderIMAP *map){
#endif
 BITMAPFILEHEADER bmfh;
 DWORD dwRead;
 BITMAPINFOHEADER bmih;
 WORD bits;
 RGBQUAD *pRgb;
 FILE *t24;
 unsigned char *lpBitmapBits,*p,*pR,*pG,*pB;
 long imagesize,nc,wi,hi,fail=0,i,j,b;
 if((pRgb=(RGBQUAD *)X__Malloc(300*sizeof(RGBQUAD))) == NULL)return 0;
 Render_Message(61,0,fn);
 if((t24=GetRenderRelativeFile(fn,ofname,NULL)) == NULL){
   if((fn=GetRenderReplacementFilename(fn)) != NULL)t24=fopen(fn,"rb");
 }
 if(t24 == NULL)return 0;
 fread((char *)&bmfh,sizeof(BITMAPFILEHEADER),1,t24);
 fread((char *)&bmih,sizeof(BITMAPINFOHEADER),1,t24);
 if(bmih.biClrUsed != 0)nc=bmih.biClrUsed;
 else{
   bits = bmih.biBitCount;
   switch (bits){
     case 1:    nc=2;   break;
     case 4:    nc=16;  break;
     case 8:    nc=256; break;
     default:   nc=0;   break;
   }
 }
 if(nc > 0)fread((char *)pRgb,nc*sizeof(RGBQUAD),1,t24);
 imagesize=bmfh.bfSize-bmfh.bfOffBits;
 if((lpBitmapBits=(unsigned char *)X__Malloc(imagesize)) == NULL){
   fclose(t24);
   X__Free(pRgb);
   return 0;
 }
 memset(lpBitmapBits,0,imagesize);
 fread((char *)lpBitmapBits,imagesize,1,t24);
 fclose(t24);
 wi=bmih.biWidth; hi=bmih.biHeight;
 imagesize=wi*hi; map->xmax = (short)wi; map->ymax = (short)hi;
 if((map->p24R = (unsigned char *)X__Malloc(imagesize)) == NULL)fail=1;
 if((map->p24G = (unsigned char *)X__Malloc(imagesize)) == NULL)fail=1;
 if((map->p24B = (unsigned char *)X__Malloc(imagesize)) == NULL)fail=1;
 if(fail){
   if(map->p24R != NULL)X__Free(map->p24R);
   if(map->p24G != NULL)X__Free(map->p24G);
   if(map->p24B != NULL)X__Free(map->p24B);
   X__Free(pRgb);
   X__Free(lpBitmapBits);
   return 0;
 }
 for(i=0;i<hi;i++){
   if(nc > 0)p=(lpBitmapBits+wi*(hi-i-1));
   else      p=(lpBitmapBits+wi*(hi-i-1)*3);
   pR=(map->p24R+wi*i); pG=(map->p24G+wi*i); pB=(map->p24B+wi*i);
   for(j=0;j<wi;j++){
     if(nc > 0){
       b=(long)(*p++);
       *pR++ = pRgb[b].rgbRed;
       *pG++ = pRgb[b].rgbGreen;
       *pB++ = pRgb[b].rgbBlue;
     }
     else{
       *pR++ = *p++; *pG++ = *p++; *pB++ = *p++;
     }
   }
 }
 X__Free(pRgb);
 X__Free(lpBitmapBits);
 map->image=YES;
 map->depth=24;
 return 1;
}

#if !__NT__

#undef DWORD
#undef WORD
#undef BYTE

#endif

// end of BMP code

#include "flc.h"

#if _SUNSTYLE
static short RenderFlcBrun();
static short RenderFlcColor();
static short LoadRenderAnimCel();
#else
static short RenderFlcBrun(FILE *f, RenderIMAP *map);
static short RenderFlcColor(FILE *f, RenderIMAP *map, short type);
static short LoadRenderAnimCel(FILE *f, RenderIMAP *map, Flc_head *flc);
#endif

#ifdef _SUNSTYLE
static long LoadCelBrush(fn,ofname,map)
char *fn,*ofname; RenderIMAP *map; {
#else
static long LoadCelBrush(char *fn, char *ofname,
                           RenderIMAP *map){
#endif
 long i,cel_type;
 struct pic_header aa_cel;
 Flc_head flc_cel;
 unsigned char cc,*pix;
 char local_message[256];
 sprintf(local_message," %s",fn);
 Render_Message(84,0,local_message);
 if((Bit_File=GetRenderRelativeFile(fn,ofname,NULL)) == NULL){
   if((fn=GetRenderReplacementFilename(fn)) == NULL ||
      (Bit_File=fopen(fn,"rb")) == NULL){
        return(0);
   }
 }
 if(fread(&aa_cel,sizeof(struct pic_header),1,Bit_File) != 1)cel_type=0;
 else if(aa_cel.type == PIC_MAGIC){
   cel_type=1;
   map->xmax = aa_cel.w;
   map->ymax = aa_cel.h;
 }
 else{
   fseek(Bit_File,0,SEEK_SET);
   if(fread(&flc_cel,sizeof(Flc_head),1,Bit_File) != 1)cel_type=0;
   else if(flc_cel.type == FLCH_MAGIC){
     cel_type=2;
     map->xmax = flc_cel.width;
     map->ymax = flc_cel.height;
   }
 }
 if(cel_type == 0){Render_Message(6,0,NULL);  fclose(Bit_File);   return(0);}
 if((map->pR=(unsigned char *)X__Malloc(256)) == NULL){Render_Message(9,0,NULL); return(0);}
 if((map->pG=(unsigned char *)X__Malloc(256)) == NULL){Render_Message(9,0,NULL); return(0);}
 if((map->pB=(unsigned char *)X__Malloc(256)) == NULL){Render_Message(9,0,NULL); return(0);}
 brush_size=(long)(map->xmax)*(long)(map->ymax);
 if((map->p=(unsigned char *)X__Malloc(brush_size)) == NULL){
   Render_Message(9,0,NULL);
   X__Free(map->pR); X__Free(map->pG); X__Free(map->pB);
   fclose(Bit_File);
   return(0);
 }
 if(cel_type == 2){
   if(LoadRenderAnimCel(Bit_File,map,&flc_cel) < 0){
     Render_Message(80,1,NULL);
     fclose(Bit_File);
     X__Free(map->pR); X__Free(map->pG); X__Free(map->pB); X__Free(map->p);
     return(0);
   }
 }
 else{
   for(i = 0; i < 256 ; i++) {
     fread(&cc,1,1,Bit_File); map->pR[i]  = cc << 2;
     fread(&cc,1,1,Bit_File); map->pG[i]  = cc << 2;
     fread(&cc,1,1,Bit_File); map->pB[i]  = cc << 2;
   }
   pix=map->p;
   for(i=0;i<map->ymax;i++){
     if(fread(pix,1,map->xmax,Bit_File) != map->xmax){
       Render_Message(81,1,NULL);
       fclose(Bit_File);
       X__Free(map->pR); X__Free(map->pG); X__Free(map->pB); X__Free(map->p);
       return(0);
     }
     pix += (long)(map->xmax);
   }
 }
 fclose(Bit_File);
 map->image=YES;
 return(1);
}

#ifdef _SUNSTYLE
static short LoadRenderAnimCel(f,map,flc)
  FILE *f; RenderIMAP *map; Flc_head *flc; {
#else
static short LoadRenderAnimCel(FILE *f, RenderIMAP *map, Flc_head *flc){
#endif
 Flc_frame f_frame;
 Flc_chunk f_chunk;
 long i,j,k;
 char c;
 unsigned char *p;
/*printf("size of flc head %ld\n",(long)sizeof(Flc_head));*/
 fseek(f,flc->oframe1,SEEK_SET);
 if(fread(&f_frame,sizeof(Flc_frame),1,f) != 1 ||
           f_frame.type != FLCF_MAGIC)return -1;
/*printf("point2  %ld chunks\n",f_frame.chunks);*/
 for(i=0;i<f_frame.chunks;i++){
   if(fread(&f_chunk,sizeof(Flc_chunk),1,f) != 1)return -1;
/*printf("chunk type %ld\n",(long)f_chunk.type);*/
   if(f_chunk.type == FLC_COLOR256 || f_chunk.type == FLC_COLOR){
     if(RenderFlcColor(f,map,f_chunk.type) < 0)return -1;
   }
   else if(f_chunk.type == FLC_BRUN){
     if(RenderFlcBrun(f,map) < 0)return -1;
   }
   else if(f_chunk.type == FLC_COPY){
     p=map->p;
     for (k=0;k<map->ymax;k++){
      if(fread(p,1,map->xmax,f) != map->xmax)return -1;
       p += map->xmax;
     }
   }
   else{
     for(j=0;j<f_chunk.size-6;j++)fread(&c,1,1,f);
   }
 }
 return 0;
}

#ifdef _SUNSTYLE
static short RenderFlcColor(f,map,type)
  FILE *f; RenderIMAP *map; short type; {
#else
static short RenderFlcColor(FILE *f, RenderIMAP *map, short type){
#endif
 short i,j,pk,packets,ci=0;
 unsigned char skip=0,change,colour;
 if(fread(&packets,2,1,f) != 1)return -1;
 for (pk=0; pk<packets; pk++ ) {
   if(fread(&skip,1,1,f) != 1)return -1;
   ci += skip;
   if(fread(&change,1,1,f) != 1)return -1;
   if (change == 0) change=255;
   for (i=0; i<=change; i++) {
     if(fread(&colour,1,1,f) != 1)return -1;
     if(type == FLC_COLOR)colour = colour << 2;
     map->pR[ci] = colour;
     if(fread(&colour,1,1,f) != 1)return -1;
     if(type == FLC_COLOR)colour = colour << 2;
     map->pG[ci] = colour;
     if(fread(&colour,1,1,f) != 1)return -1;
     if(type == FLC_COLOR)colour = colour << 2;
     map->pB[ci] = colour;
     ci++;
   }
 }
 return 0;
}

#ifdef _SUNSTYLE
static short RenderFlcBrun(f,map)
  FILE *f; RenderIMAP *map; short type; {
#else
static short RenderFlcBrun(FILE *f, RenderIMAP *map){
#endif
 char size=0;
 unsigned char packet,*p,colour;
 short i,line,pk;
 p=map->p;
 for (line=0;line<map->ymax;line++){
   if(fread(&packet,1,1,f) != 1)return -1;
   for(pk=0; pk<map->xmax; pk+=abs(size)){
     if(fread(&size,1,1,f) != 1)return -1;
     if(size > 0){
       if(fread(&colour,1,1,f) != 1)return -1;
       for (i=0; i<size; i++) *p++ = colour;
     }
     else{
       for (i = size; i<0; i++){
         if(fread(&colour,1,1,f) != 1)return -1;
         *p++ = colour;
       }
     }
   }
 }
 return 0;
}
