/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// loadmesh.cpp  -- This file contains the code to load
//                  the mesh models
//

#define MODULE_LOADER

#include "render.h"

typedef struct SKELETON {
  long id;
  vector p;
  double T[4][4],R[4][4],Q[4][4];
}skel;

typedef double quaternion[4];

#ifdef _SUNSTYLE
static void InterpolateRobot();
static long LoadQmorph();
static void QuadInterpolate();
static void RenderTransformToRobot();
static long *RenderReadSkeleton();
static short R_ReadAxis();
static short R_ReadAxes();
static short R_ReadFaces();
static short R_ReadOldFaces();
static short R_ReadImaps();
static short R_ReadVertices();
static short R_ReadVertText();
static short R_ReadOldText();
static short R_ReadShader();
static short R_ReadMorphShader();
static void R_MorphVector();
#else
static void InterpolateRobot(double mr, long n, skel *sk1, skel *sk2);
static long LoadQmorph(char *QmorphFile,long N,ivector *Qv, ivector Mq);
static void QuadInterpolate(short mo, double mr, double qr,
                            vector v0, vector v1, vector vm,
                            vector v);
static void RenderTransformToRobot(long O,
                             long *Skid, skel *Skeleton,
                             double mr,
                             char *robo_f1, char *robo_f2);
static long RenderReadNurbs(long clen);
static long *RenderReadSkeleton(long clen, short version, skel **sp);
static short R_ReadAxis(long clen);
static short R_ReadAxes(long clen);
static short R_ReadFaces(long  clen, BOOL short_form);
static short R_ReadOldFaces(long  clen, long shader);
static short R_ReadVertices(long clen);
static short R_ReadVertText(long clen);
static short R_ReadOldText(long clen, long frame,
                    char *objectfilename, char *str);
static short R_ReadImaps(long clen, long frame,
                    char *objectfilename, char *str);
static short R_ReadShader(long clen, long frame,
                    char *objectfilename, char *str);
static short R_ReadMorphShader(long clen,
                    char *objectfilename, char *str);
static void R_MorphVector(vector b, ivector Mx, double mr);
#endif

HANDLE fp;
DWORD dwRead;
static char str[10];
static long newface,newvert;


#ifdef _SUNSTYLE
char R_getl(val) long *val; {
#else
char R_getl(long *val){
#endif
 long l,l0,l1,l2,l3;
 ReadFile(fp,&l,sizeof(long),&dwRead,NULL);
 l0 = (l >> 24) & 0x000000ff;
 l1 = (l >> 16) & 0x000000ff;
 l2 = (l >>  8) & 0x000000ff;
 l3 = (l      ) & 0x000000ff;
 *val = (l3 << 24) | (l2 << 16) | (l1 <<  8) | l0;
 return(1);
}

#ifdef _SUNSTYLE
char R_getsht(val) short *val; {
#else
char R_getsht(short * val){
#endif
 short s,s0,s1;
 ReadFile(fp,&s,sizeof(short),&dwRead,NULL);
 s0 = (s >>  8) & 0x00ff;
 s1 = (s      ) & 0x00ff;
 *val = (s1 <<  8) | s0;
 return(1);
}

#ifdef _SUNSTYLE
char R_getusht(val) unsigned short *val; {
#else
char R_getusht(unsigned short * val){
#endif
 unsigned short s,s0,s1;
 ReadFile(fp,&s,sizeof(short),&dwRead,NULL);
 s0 = (s >>  8) & 0x00ff;
 s1 = (s      ) & 0x00ff;
 *val = (s1 <<  8) | s0;
 return(1);
}

static unsigned short GetUshortFromBuffer(unsigned char *b){
 return ((unsigned short)(*b) <<  8) | (unsigned short)(*(b+1));
}

static short GetShortFromBuffer(unsigned char *b){
 return ((short)(*b) <<  8) | (short)(*(b+1));
}

static long GetLongFromBuffer(unsigned char *b){
 return ((long)(*b) << 24) | ((long)(*(b+1)) << 16) |
        ((long)(*(b+2)) <<  8) | (long)(*(b+3));
}

// This is the MORPHING code that reads other models and performs
// either linear, quadratic or DISSIMILAR morphing

#define DIST(a,b) sqrt(                                           \
                       (a[0]-(double)b[0]) * (a[0]-(double)b[0])  \
                      +(a[1]-(double)b[1]) * (a[1]-(double)b[1])  \
                      +(a[2]-(double)b[2]) * (a[2]-(double)b[2])  \
                      )

typedef struct tagMorphV{
 ivector v;
 long    id;
} MorphV;

static long NextId,NmorphV,NobjectV,*NobjectId;
static MorphV *Mvert;

static void AssignOvToMv(long id){
 long id1;
 static long i,n,n1,n2,n3,n4,n5,n6,n7,n8;
 static double x,y,z,an;
 static vertex *v;
 n=0; x=y=z=0.0;
 for(i=0;i<NmorphV;i++){
   if(Mvert[i].id == id){
     n++; an=(double)(n-1)/(double)n;
     x=x*an+(double)(Mvert[i].v[0])/(double)n;
     y=y*an+(double)(Mvert[i].v[1])/(double)n;
     z=z*an+(double)(Mvert[i].v[2])/(double)n;
   }
 }
 if(n < 2)return;   /* none or one vertex in group */
 n1=n2=n3=n4=n5=n6=n7=n8=0;
 for(i=0;i<NmorphV;i++){                // ToDo check this!!! Ray tracing ???
   if(Mvert[i].id == id){
     if(Mvert[i].v[0] > (long)x){
       if(Mvert[i].v[1] > (long)y){
         if(Mvert[i].v[2] > (long)z){
           Mvert[i].id=id; n1++;
         }
         else {
           Mvert[i].id=NextId; n2++;
         }
       }
       else{
         if(Mvert[i].v[2] > (long)z){
           Mvert[i].id=NextId+1; n3++;
         }
         else {
           Mvert[i].id=NextId+2; n4++;
         }
       }
     }
     else {
       if(Mvert[i].v[1] > (long)y){
         if(Mvert[i].v[2] > (long)z){
           Mvert[i].id=NextId+3; n5++;
         }
         else {
           Mvert[i].id=NextId+4; n6++;
         }
       }
       else{
         if(Mvert[i].v[2] > (long)z){
           Mvert[i].id=NextId+5; n7++;
         }
         else {
           Mvert[i].id=NextId+6; n8++;
         }
       }
     }
   }
 }
 if(n1 == n || n2 == n || n3 == n || n4 == n ||
    n5 == n || n6 == n || n7 == n || n8 == n)return; /* can't separate any */
 n=0; x=y=z=0.0;
 v=Object[ObjectCount].Vbase;
 for(i=0;i<NobjectV;i++){
   if(NobjectId[i] == id){
     n++; an=(double)(n-1)/(double)n;
     x=x*an+(double)(v[i].p[0])/(double)n;
     y=y*an+(double)(v[i].p[1])/(double)n;
     z=z*an+(double)(v[i].p[2])/(double)n;
   }
 }
 for(i=0;i<NobjectV;i++){
  if(NobjectId[i] == id){
     if(v[i].p[0] > x){
       if(v[i].p[1] > y){
         if(v[i].p[2] > z){
           NobjectId[i]=id;
         }
         else{
           NobjectId[i]=NextId;
         }
       }
       else{
         if(v[i].p[2] > z){
           NobjectId[i]=NextId+1;
         }
         else{
           NobjectId[i]=NextId+2;
         }
       }
     }
     else{
       if(v[i].p[1] > y){
         if(v[i].p[2] > z){
           NobjectId[i]=NextId+3;
         }
         else{
           NobjectId[i]=NextId+4;
         }
       }
       else{
         if(v[i].p[2] > z){
           NobjectId[i]=NextId+5;
         }
         else{
           NobjectId[i]=NextId+6;
         }
       }
     }
   }
 }
 id1=NextId; NextId += 7;
 AssignOvToMv(id);
 AssignOvToMv(id1);
 AssignOvToMv(id1+1);
 AssignOvToMv(id1+2);
 AssignOvToMv(id1+3);
 AssignOvToMv(id1+4);
 AssignOvToMv(id1+5);
 AssignOvToMv(id1+6);
 return;
}

#ifdef _SUNSTYLE
#else
static int InterpolateMorph1(long nm, long no, double mr, ivector Mv){
#endif
 long i,j,id,*Node;
 vertex *v;
 double d,dm; //,dl[3],dr[3];
// ivector cm,co;
 if(nm == 0 || no == 0)return 0;
 if((Mvert=(MorphV *)X__Malloc(nm*sizeof(MorphV))) == NULL)return 0;
 NmorphV=nm; NobjectV=no;
 if((NobjectId=(long *)X__Malloc(no*sizeof(long))) == NULL){
   X__Free(Mvert); return 0;
 }
// for(i=0;i<3;i++){dl[i] =  FARAWAY;  dr[i] = -FARAWAY;}
 for(i=0;i<nm;i++){
   R_getl(&j); Mvert[i].v[0]=j+Mv[0];
   R_getl(&j); Mvert[i].v[1]=j+Mv[1];
   R_getl(&j); Mvert[i].v[2]=j+Mv[2];
   Mvert[i].id=0;
//   for(j=0;j<3;j++){
//     dl[j]=min(dl[j],(double)Mvert[i].v[j]);
//     dr[j]=max(dr[j],(double)Mvert[i].v[j]);
//   }
 }
// for(i=0;i<3;i++)cm[i]=(long)((dl[i]+dr[i])*0.5);
// for(i=0;i<3;i++){dl[i] =  FARAWAY;  dr[i] = -FARAWAY;}
 v=Object[ObjectCount].Vbase;
 for(i=0;i<no;i++){
   NobjectId[i]=0;
//   for(j=0;j<3;j++){
//     dl[j]=min(dl[j],v[i].p[j]);
//     dr[j]=max(dr[j],v[i].p[j]);
//   }
 }
// for(i=0;i<3;i++)co[i]=(long)((dl[i]+dr[i])*0.5)-cm[i];
 NextId=1;
 AssignOvToMv(0);
 if((Node=(long *)X__Malloc(NextId*sizeof(long))) != NULL){
 /* fix up addresses */
   for(i=0;i<NextId;i++)Node[i] = -1;
   for(i=0;i<nm;i++){
     Node[Mvert[i].id]=i;
   }
   v=Object[ObjectCount].Vbase;
   for(i=0;i<no;i++){
     id=Node[NobjectId[i]];
     if(id < 0){/* no equivalent vertex - find closest */
       id=0; dm=FARAWAY;
//       for(j=0;j<nm;j++){
//         cm[0]=Mvert[j].v[0]+co[0];
//         cm[1]=Mvert[j].v[1]+co[1];
//         cm[2]=Mvert[j].v[2]+co[2];
//         if((d=DIST(v[i].p,cm)) < dm){dm=d; id=j;}
//       }
       for(j=0;j<no;j++){ /* closest in object that has been assigned */
         if(i == j || Node[NobjectId[j]] < 0)continue;
         if((d=DIST(v[i].p,v[j].p)) < dm){dm=d; id=Node[NobjectId[j]];}
       }
     }
     if(id < nm){
       v[i].p[0]=(v[i].p[0]-(double)Mvert[id].v[0])*mr+(double)Mvert[id].v[0];
       v[i].p[1]=(v[i].p[1]-(double)Mvert[id].v[1])*mr+(double)Mvert[id].v[1];
       v[i].p[2]=(v[i].p[2]-(double)Mvert[id].v[2])*mr+(double)Mvert[id].v[2];
     }
     else{;} // printf("\n\n\nError\n\n");
   }
 }
 X__Free(NobjectId); X__Free(Mvert);
 return 1;
}

#if 0

#ifdef _SUNSTYLE
#else
static int InterpolateMorph2(long nm, long no, double mr, ivector Mv){
#endif
 /* nm(morph) will be > no(object) */
 long i,j,id,*olist;
 vertex *v;
 ivector cm,co;
 ivector *MvertV;
 double d,dd,mr2,x,y,z;
 double dl[3],dr[3];
 if(nm == 0 || no == 0)return 0;
 if((MvertV=(ivector *)X__Malloc(nm*sizeof(ivector))) == NULL)return 0;
 if((olist=(long *)X__Malloc(no*sizeof(long))) == NULL){
  X__Free(olist);
  return 0;
 }
 for(i=0;i<3;i++){dl[i] =  FARAWAY;  dr[i] = -FARAWAY;}
 for(i=0;i<nm;i++){
   R_getl(&j); MvertV[i][0]=j+Mv[0];
   R_getl(&j); MvertV[i][1]=j+Mv[1];
   R_getl(&j); MvertV[i][2]=j+Mv[2];
   for(j=0;j<3;j++){
     dl[j]=min(dl[j],(double)MvertV[i][j]);
     dr[j]=max(dr[j],(double)MvertV[i][j]);
   }
 }
 v=Object[ObjectCount].Vbase;
 for(i=0;i<3;i++)cm[i]=(long)((dl[i]+dr[i])*0.5);
 for(i=0;i<3;i++){dl[i] =  FARAWAY;  dr[i] = -FARAWAY;}
 for(i=0;i<no;i++){
   olist[i] = -1;
   for(j=0;j<3;j++){
     dl[j]=min(dl[j],v[i].p[j]);
     dr[j]=max(dr[j],v[i].p[j]);
   }
 }
 for(i=0;i<3;i++)co[i]=(long)((dl[i]+dr[i])*0.5)-cm[i];
 for(i=0;i<nm;i++){
   id= -1; d=FARAWAY;
   for(j=0;j<no;j++){ /* find closest in object to each morph vertex */
     cm[0]=MvertV[i][0]+co[0];
     cm[1]=MvertV[i][1]+co[1];
     cm[2]=MvertV[i][2]+co[2];
     if((dd=DIST(v[j].p,cm)) < d){d=dd; id=j;}
   }
   if(id >= 0)olist[id]=i;
 }
 mr2=mr*mr*mr;
 for(i=0;i<no;i++){
   if((id=olist[i]) < 0){ /* for any unassigned vertices in object */
     d=FARAWAY;
     for(j=0;j<no;j++){
       if(olist[j] < 0 || i == j)continue;
       if((dd=DIST(v[i].p,v[j].p)) < d){d=dd; id=j;}
     }
     j=olist[id];
     x=(v[id].p[0]-(double)MvertV[j][0])*mr+(double)MvertV[j][0];
     y=(v[id].p[1]-(double)MvertV[j][1])*mr+(double)MvertV[j][1];
     z=(v[id].p[2]-(double)MvertV[j][2])*mr+(double)MvertV[j][2];
     v[i].p[0]=(v[i].p[0]-x)*mr2+x;
     v[i].p[1]=(v[i].p[1]-y)*mr2+y;
     v[i].p[2]=(v[i].p[2]-z)*mr2+z;
   }
 }
 for(i=0;i<no;i++){
   if((id=olist[i]) < 0)continue; /* for any assigned vertices in object */
   v[i].p[0]=(v[i].p[0]-(double)MvertV[id][0])*mr+(double)MvertV[id][0];
   v[i].p[1]=(v[i].p[1]-(double)MvertV[id][1])*mr+(double)MvertV[id][1];
   v[i].p[2]=(v[i].p[2]-(double)MvertV[id][2])*mr+(double)MvertV[id][2];
 }
 X__Free(olist);
 X__Free(MvertV);
 return 1;
}

#endif

#ifdef _SUNSTYLE
#else
static void R_MorphVector(vector b, ivector Mx, double mr){
#endif
 long i,j;
 for(i=0;i<3;i++){
   R_getl(&j); j=j+Object[ObjectCount].Origin[i]-Mx[i];
   b[i]=(b[i]-(double)j)*mr + (double)j;
 }
}

#ifdef _SUNSTYLE
long LoadMorphObject(mo,FileName,mr,QmorphFile,qr) short mo;
char *FileName; double mr; char *QmorphFile; double qr; {
#else
long LoadMorphObject(short mo, char *FileName,double mr,
                     char *QmorphFile,double qr){
#endif
 unsigned char *buffer=NULL,*b;
 long CHUNKsize,FORMsize,nm,i,j,nf,l,m,m_angle,l_temp,version,mc,
      file_len,shader,na;
 unsigned short uj;
 short s,m_pp,m_rp,m_bp,m_tp;
 unsigned char mcolor[3],mtexture,mmatsp,mmattr,mmatre,
               mtexco[3],mbrush,uc_temp,t_colour[3],
               mdummy;
 char k,filename[256],ps,rs,bs,ts;
 double a;
 ivector Mx,Mv,Mq,*Qv;
 vector V0,V1,VM;
 vertex *v;
 face   *f;
 int    same=1;
 Qv=NULL;
 if(mo > 1){
   nm=(long)(Object[ObjectCount].NoVertices)*(long)(sizeof(ivector));
   if(mo > 1){
     if((Qv=(ivector *)X__Malloc(nm)) == NULL)mo=1;
   }
   if(mo > 1){
     if(LoadQmorph(QmorphFile,Object[ObjectCount].NoVertices,Qv,Mq) < 0){
        X__Free(Qv); Qv=NULL;
        mo=1; /* Q load fail ignore Q morph */
      }
   }
 }
 Mx[0]=0; Mx[1]=0; Mx[2]=0;
 if((fp=GetRenderRelativeHandle((char *)strlwr(FileName),NULL,NULL))
       == INVALID_HANDLE_VALUE){
   Render_Message(20,1,NULL);
   goto X__ERROR;
 }
 ReadFile(fp,str,4,&dwRead,NULL);
 if(dwRead != 4 || strcmp(str,"FORM") != 0){
   Render_Message(21,1,NULL);
   CloseHandle(fp);
   goto X__ERROR;
 }
 ReadFile(fp,&FORMsize,4,&dwRead,NULL);
 ReadFile(fp,str,4,&dwRead,NULL);
 if(dwRead != 4 || (strcmp(str,"OFXM") != 0 && strcmp(str,"AAPO") != 0)){
   Render_Message(22,1,NULL);
   CloseHandle(fp);
   goto X__ERROR;
 }
loop:
 if(!ReadFile(fp,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4)goto end;
 if (R_getl(&CHUNKsize) != 1)goto end;
 if     (strcmp(str,"VERT") == 0){ /* offset assumes Axis is first segment */
   nm=CHUNKsize/12;
   if(nm != Object[ObjectCount].NoVertices){
     same=0;
     if(mo != 1){
       Render_Message(23,1,NULL);
       CloseHandle(fp);
       goto X__ERROR;
     }
   }
   if(same){
     v=Object[ObjectCount].Vbase;
     if(mo == 1){
       for(i=0;i<nm;i++){
         R_getl(&j); Mv[0] = j+Object[ObjectCount].Origin[0]-Mx[0];
         R_getl(&j); Mv[1] = j+Object[ObjectCount].Origin[1]-Mx[1];
         R_getl(&j); Mv[2] = j+Object[ObjectCount].Origin[2]-Mx[2];
         v[i].p[0] = (v[i].p[0]-(double)Mv[0])*mr + (double)Mv[0];
         v[i].p[1] = (v[i].p[1]-(double)Mv[1])*mr + (double)Mv[1];
         v[i].p[2] = (v[i].p[2]-(double)Mv[2])*mr + (double)Mv[2];
       }
     }
     else{
       for(i=0;i<nm;i++){
         V1[0] = v[i].p[0] - (double)Object[ObjectCount].Origin[0];
         V1[1] = v[i].p[1] - (double)Object[ObjectCount].Origin[1];
         V1[2] = v[i].p[2] - (double)Object[ObjectCount].Origin[2];
         VM[0] = (double)(Qv[i][0] - Mq[0]);
         VM[1] = (double)(Qv[i][1] - Mq[1]);
         VM[2] = (double)(Qv[i][2] - Mq[2]);
         R_getl(&j); V0[0] = (double)(j - Mx[0]);
         R_getl(&j); V0[1] = (double)(j - Mx[1]);
         R_getl(&j); V0[2] = (double)(j - Mx[2]);
         QuadInterpolate(mo,mr,qr,V0,V1,VM,v[i].p);
         v[i].p[0] += (double)Object[ObjectCount].Origin[0];
         v[i].p[1] += (double)Object[ObjectCount].Origin[1];
         v[i].p[2] += (double)Object[ObjectCount].Origin[2];
       }
     }
   }
   else{ /* try to interpolate morph - assume AXIS has been read */
     int result;
     VECSUB(Object[ObjectCount].Origin,Mx,Mv)
     if(nm > Object[ObjectCount].NoVertices)
       result=InterpolateMorph1(nm,Object[ObjectCount].NoVertices,mr,Mv);
     else
       result=InterpolateMorph1(nm,Object[ObjectCount].NoVertices,mr,Mv);
     if(!result){
       Render_Message(23,1,NULL);
       CloseHandle(fp);
       goto X__ERROR;
     }
   }
 }
 else if(strcmp(str,"AXIS") == 0){/* Axis is assumed to be first chunk */
   R_getl(&j); Mx[0] = j; /* this is the morphed object offset */
   R_getl(&j); Mx[1] = j; /* before using the morph it must be reposition */
   R_getl(&j); Mx[2] = j; /* to have its offset the same as the object */
   R_MorphVector(Object[ObjectCount].Abase[0].bo,Mx,mr); /* Tx origin */
   R_MorphVector(Object[ObjectCount].Abase[0].bu,Mx,mr); /* Tx X      */
   R_MorphVector(Object[ObjectCount].Abase[0].bv,Mx,mr); /* Tx X      */
   VECCOPY(Object[ObjectCount].Abase[0].bo,Object[ObjectCount].Abase[0].boin)
   VECCOPY(Object[ObjectCount].Abase[0].bu,Object[ObjectCount].Abase[0].buin)
   VECCOPY(Object[ObjectCount].Abase[0].bv,Object[ObjectCount].Abase[0].bvin)
 }
 else if(same && (strcmp(str,"AXES") == 0)){ /* extra shader axes */
   na=CHUNKsize/68; na++;
   if(na != Object[ObjectCount].NoAxes)R_getchunk(CHUNKsize-8);
   else{
     for(i=1;i<na;i++){
       ReadFile(fp,filename,32,&dwRead,NULL);
       R_MorphVector(Object[ObjectCount].Abase[i].bo,Mx,mr);
       R_MorphVector(Object[ObjectCount].Abase[i].bu,Mx,mr);
       R_MorphVector(Object[ObjectCount].Abase[i].bv,Mx,mr);
       VECCOPY(Object[ObjectCount].Abase[i].bo,Object[ObjectCount].Abase[i].boin)
       VECCOPY(Object[ObjectCount].Abase[i].bu,Object[ObjectCount].Abase[i].buin)
       VECCOPY(Object[ObjectCount].Abase[i].bv,Object[ObjectCount].Abase[i].bvin)
     }
   }
 }
 else if(same && (strcmp(str,"FACE") == 0 ||
                  strcmp(str,"SFAC") == 0)){ /* morph the face colour */
  if(strcmp(str,"FACE") == 0){nf=CHUNKsize/20,shader=0;}
  else                       {nf=CHUNKsize/21;shader=1;}
  if(nf != Object[ObjectCount].NoFaces){
    Render_Message(24,1,NULL);
    CloseHandle(fp);
    goto X__ERROR;
  }
  buffer=(unsigned char *)X__Malloc(CHUNKsize);
  if(buffer == NULL){
    CloseHandle(fp);
    goto X__ERROR;
  }
  ReadFile(fp,buffer,CHUNKsize,&dwRead,NULL);
  b=buffer;
  f=Object[ObjectCount].Fbase;
  for(i=0;i<nf;i++){
    b += 3*sizeof(long);
    memcpy(mcolor,b,3); b += 3;
    mtexture = *b++;
    (f+i)->color[0]=( (f+i)->color[0] - mcolor[0] )*mr + mcolor[0];
    (f+i)->color[1]=( (f+i)->color[1] - mcolor[1] )*mr + mcolor[1];
    (f+i)->color[2]=( (f+i)->color[2] - mcolor[2] )*mr + mcolor[2];
    mmatsp = *b++;
    mmattr = *b++;
    mmatre = *b++;
    mbrush = *b++;  /* dump the map ID */
    mtexco[0] = (mmatsp & 0x0f) << 4;
    mtexco[1] = (mmattr & 0x0f) << 4;
    mtexco[2] = (mmatre & 0x0f) << 4;
    mmatsp  = (mmatsp & 0xf0) >> 4;
    mmattr &= 0xf0;
    mmatre &= 0xf0;
    (f+i)->texco[0]=( (f+i)->texco[0] - mtexco[0] )*mr + mtexco[0];
    (f+i)->texco[1]=( (f+i)->texco[1] - mtexco[1] )*mr + mtexco[1];
    (f+i)->texco[2]=( (f+i)->texco[2] - mtexco[2] )*mr + mtexco[2];
    (f+i)->matsp=( (f+i)->matsp - mmatsp )*mr + mmatsp;
    (f+i)->fparam=( (f+i)->fparam - mmattr )*mr + mmattr;
    if((mtexture & 0x1f) != 0x04)mmattr=0; /* not old glass */
    (f+i)->mattr=( (f+i)->mattr - mmattr )*mr + mmattr;
    (f+i)->matre=( (f+i)->matre - mmatre )*mr + mmatre;
    if(shader)b++; /* dump the shader axis ID */
  }
  X__Free(buffer);
 }
 else if(same && (strcmp(str,"SFCE") == 0 ||
                  strcmp(str,"LFCE") == 0)){ /* new face colour */
  if(strcmp(str,"SFCE") == 0){nf=CHUNKsize/18,shader=0;}
  else                       {nf=CHUNKsize/24;shader=1;}
  if(nf != Object[ObjectCount].NoFaces){
    Render_Message(24,1,NULL);
    CloseHandle(fp);
    goto X__ERROR;
  }
  buffer=(unsigned char *)X__Malloc(CHUNKsize);
  if(buffer == NULL){
    CloseHandle(fp);
    goto X__ERROR;
  }
  ReadFile(fp,buffer,CHUNKsize,&dwRead,NULL);
  b=buffer;
  f=Object[ObjectCount].Fbase;
  for(i=0;i<nf;i++){
    if(shader){ b += 3*sizeof(long);}
    else      { b += 3*sizeof(unsigned short);}
    memcpy(mcolor,b,3); b += 3;
    (f+i)->color[0]=( (f+i)->color[0] - mcolor[0] )*mr + mcolor[0];
    (f+i)->color[1]=( (f+i)->color[1] - mcolor[1] )*mr + mcolor[1];
    (f+i)->color[2]=( (f+i)->color[2] - mcolor[2] )*mr + mcolor[2];
    mmatsp = *b++;
    mmattr = *b++;
    mmatre = *b++;
    mtexco[0] = (mmatsp & 0x0f) << 4;
    mtexco[1] = (mmattr & 0x0f) << 4;
    mtexco[2] = (mmatre & 0x0f) << 4;
    (f+i)->texco[0]=( (f+i)->texco[0] - mtexco[0] )*mr + mtexco[0];
    (f+i)->texco[1]=( (f+i)->texco[1] - mtexco[1] )*mr + mtexco[1];
    (f+i)->texco[2]=( (f+i)->texco[2] - mtexco[2] )*mr + mtexco[2];
    mmatsp  = (mmatsp & 0xf0) >> 4;
    mmattr &= 0xf0;
    mmatre &= 0xf0;
    (f+i)->matsp=( (f+i)->matsp - mmatsp )*mr + mmatsp;
    (f+i)->mattr=( (f+i)->mattr - mmattr )*mr + mmattr;
    (f+i)->matre=( (f+i)->matre - mmatre )*mr + mmatre;
    b++;  /* attrib */
    mdummy = *b++;  /* param  */
    (f+i)->fparam=( (f+i)->fparam - mdummy)*mr + mdummy; /* param */
    b++;  /* map    */
    b++;  /* shader */
    b++;  /* axis   */
    b++;  /* pad    */
  }
  X__Free(buffer);
 }
 else if(same && (strcmp(str,"MATL") == 0)){  /* morph the material properties */
   R_getl(&j);
   if(j == 4)R_getl(&j); /* cant morph the shadow flag just dump */
   R_getl(&j);
   if(j > 1000){
     a=(double)j/10000;
     Object[ObjectCount].refractive_index =
       (Object[ObjectCount].refractive_index + a)*0.5;
   }
   R_getl(&j);
   R_getl(&j);
 }
 else if(same && strcmp(str,"SHDR") == 0 && hidden_line == NO){
   R_ReadMorphShader(CHUNKsize,FileName,str);
 }
 else if(same && strcmp(str,"TEXT") == 0 && hidden_line == NO){ /* morph text pos */
   ReadFile(fp,&k,1,&dwRead,NULL);
   if(k > 0 && k <= Object[ObjectCount].NoMaps){
     imap *Map;
     for(i=0;i<k;i++){ /* assume brushes are compatible/Order + not > */
       Map=(Object[ObjectCount].Mbase+(long)i);
       version=0;
       R_getl(&j); file_len=j;
       ReadFile(fp,filename,file_len,&dwRead,NULL); /* version 2 is just a name */
       R_MorphVector(Map->p,Mx,mr);
       R_MorphVector(Map->x,Mx,mr);
       R_MorphVector(Map->y,Mx,mr);
       R_getl(&j);/* not needed for morphing brushes  STILL etc */
       R_getl(&j);    R_getl(&j);    R_getl(&j);  /*   FF DL FF */
       R_getl(&j);                                /* skeleton   */
       R_getl(&j);                                /* map        */
       R_getl(&j); /* expansion count */
       if(j > 0){
         R_getl(&m_angle);   /* temporary map wrap angle */
         Map->angle = 360.0/Map->angle; /* note odd transform  */
         Map->angle = (Map->angle - (double)m_angle)*mr
                          + (double)m_angle;        /*.angle is multiplier */
         Map->angle = 360.0/Map->angle;
         if(j > 1){
           for(mc=1;mc<5;mc++)R_getl(&l_temp); /* dump the map lock info   */
           if(j > 5){                     /* this must be a version 2 file */
             version=1;
             ReadFile(fp,filename,32,&dwRead,NULL);
             ReadFile(fp,&ps,1,&dwRead,NULL);
             ReadFile(fp,&rs,1,&dwRead,NULL);
             ReadFile(fp,&bs,1,&dwRead,NULL);
             R_getsht(&s); m_pp=s;
             R_getsht(&s); m_rp=s;
             R_getsht(&s); m_bp=s;
             if(ps)Map->pp = (short)((double)(Map->pp - m_pp)*mr
                                          +(double)m_pp);
             if(rs)Map->rp = (short)((double)(Map->rp - m_rp)*mr
                                          +(double)m_rp);
             if(bs)Map->bp = (short)((double)(Map->bp - m_bp)*mr
                                          +(double)m_bp);
             ReadFile(fp,t_colour,3,&dwRead,NULL);  /* not used yet */
             ReadFile(fp,t_colour,3,&dwRead,NULL);
             ReadFile(fp,t_colour,3,&dwRead,NULL);  /* ------------ */
             if(ps)ReadFile(fp,filename,file_len,&dwRead,NULL); /* skip the filenames */
             if(rs)ReadFile(fp,filename,file_len,&dwRead,NULL);
             if(bs)ReadFile(fp,filename,file_len,&dwRead,NULL);
             R_getsht(&s); /* should be 0 use as expansion count */
             if(s > 0)for(mc=0;mc<s;mc++)R_getl(&l_temp);
           }
         }
       }
     }
   }
   else{
     if(k > 0){
       Render_Message(25,0,NULL);
       for(i=0;i<k;i++){ /* get rid of the map information */
         R_getl(&j); ReadFile(fp,filename,j,&dwRead,NULL);
         for(l=0;l<15;l++)R_getl(&m);
         R_getl(&j); /* expansion */
         if(j > 0){ for(l=0;l<j;l++)R_getl(&m); }
       }
     }
   }
 }
 else if(same && strcmp(str,"IMAP") == 0 && hidden_line == NO){ /* morph text pos */
  R_getl(&j);  /* read number of maps */
  if(j > 0 && j <= Object[ObjectCount].NoMaps){
    imap *Map;
    for(i=0;i<j;i++){
      Map=(Object[ObjectCount].Mbase+(long)i);
      R_MorphVector(Map->p,Mx,mr);
      R_MorphVector(Map->x,Mx,mr);
      R_MorphVector(Map->y,Mx,mr);
      R_getl(&l);  /* anim / still */
      R_getl(&l);  R_getl(&l); R_getl(&l);
      R_getl(&l);  /* refl type */
      R_getl(&l);  R_getl(&l); R_getl(&l);
      R_getl(&l);  /* bump type */
      R_getl(&l);  R_getl(&l); R_getl(&l);
      R_getl(&l);  /* trans type */
      R_getl(&l);  R_getl(&l); R_getl(&l);
      R_getl(&l);
      R_getl(&l);
      R_getl(&m_angle);
      Map->angle = 360.0/Map->angle; /* note odd transform  */
      Map->angle = (Map->angle - (double)m_angle)*mr
           + (double)m_angle;        /*.angle is multiplier */
      Map->angle = 360.0/Map->angle;
      for(k=0;k<4;k++)R_getl(&l);
      ReadFile(fp,filename,32,&dwRead,NULL);
      ReadFile(fp,&ps,1,&dwRead,NULL);
      ReadFile(fp,&rs,1,&dwRead,NULL);
      ReadFile(fp,&bs,1,&dwRead,NULL);
      ReadFile(fp,&ts,1,&dwRead,NULL);
      ReadFile(fp,&mdummy,1,&dwRead,NULL);
      R_getsht(&s); m_pp=s; /* these will be set to -1 */
      R_getsht(&s); m_rp=s; /* if a map of this type   */
      R_getsht(&s); m_bp=s; /* is not used in Material */
      R_getsht(&s); m_tp=s;
      if(ps)Map->pp = (short)((double)(Map->pp - m_pp)*mr
                                          +(double)m_pp);
      if(rs)Map->rp = (short)((double)(Map->rp - m_rp)*mr
                                          +(double)m_rp);
      if(bs)Map->bp = (short)((double)(Map->bp - m_bp)*mr
                                          +(double)m_bp);
      if(ts)Map->tp = (short)((double)(Map->tp - m_tp)*mr
                                          +(double)m_tp);
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,t_colour,3,&dwRead,NULL);
      ReadFile(fp,filename,256,&dwRead,NULL);
      ReadFile(fp,filename,256,&dwRead,NULL);
      ReadFile(fp,filename,256,&dwRead,NULL);
      ReadFile(fp,filename,256,&dwRead,NULL);
    }
  }
  else{
    Render_Message(25,0,NULL);
    if(j > 0)for(i=0;i<CHUNKsize-4;i++)
      ReadFile(fp,&mdummy,1,&dwRead,NULL);
  }
 }
// ToDo Morphing Nurbs ???
 else R_getchunk(CHUNKsize); /* read any unused chunks */
 goto loop;
end:
 CloseHandle(fp);
 if(Qv != NULL)X__Free(Qv);
 return 1;
 X__ERROR:
 if(Qv != NULL)X__Free(Qv);
 return 0;
}

#ifdef _SUNSTYLE
static void QuadInterpolate(mo,mr,qr,v0,v1,vm,v) short mo;
double mr; double qr; vector v0; vector v1; vector vm; vector v; {
#else
static void QuadInterpolate(short mo, double mr, double qr,
                            vector v0, vector v1, vector vm,
                            vector v){
#endif
 double a,b,c;
 long j;
 for(j=0;j<3;j++){
   if(mo == 2){ /* backwards interpolation */
     a=(    (v1[j]-v0[j]) - (v0[j] - vm[j])/qr )/(1.0+qr);
     b=( qr*(v1[j]-v0[j]) + (v0[j] - vm[j])/qr )/(1.0+qr);
     c=v0[j];
   }
   else{        /* =3  forwards interpolation */
     a=(       -(v1[j]-v0[j]) + (vm[j] - v0[j])/(1.0+qr) )/qr;
     b=( (qr+1)*(v1[j]-v0[j]) - (vm[j] - v0[j])/(1.0+qr) )/qr;
     c=v0[j];
   }
   v[j] = a*mr*mr + b*mr +c;
 }
}

#ifdef _SUNSTYLE
static long LoadQmorph(QmorphFile,N,Qv,Mq) char *QmorphFile;
long N; ivector *Qv; ivector Mq; {
#else
static long LoadQmorph(char *QmorphFile,long N,ivector *Qv, ivector Mq){
#endif
 long CHUNKsize,FORMsize,nm,i,j;
 if((fp=GetRenderRelativeHandle((char *)strlwr(QmorphFile),NULL,NULL))
       == INVALID_HANDLE_VALUE){
   Render_Message(26,1,NULL);
   goto X__ERROR;
 }
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || strcmp(str,"FORM") != 0){
   Render_Message(21,1,NULL);
   CloseHandle(fp);
   goto X__ERROR;
 }
 ReadFile(fp,&FORMsize,4,&dwRead,NULL);
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || (strcmp(str,"OFXM") != 0 && strcmp(str,"AAPO") != 0)){
   Render_Message(22,1,NULL);
   CloseHandle(fp);
   goto X__ERROR;
 }
loop:
 if(!ReadFile(fp,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4 )goto end;
 if(R_getl(&CHUNKsize) != 1)goto end;
 if     (strcmp(str,"VERT") == 0){ /* offset assumes Axis is first segment */
   nm=CHUNKsize/12;
   if(nm != N){
     Render_Message(23,1,NULL);
     CloseHandle(fp);
     goto X__ERROR;
   }
   for(i=0;i<nm;i++){
     R_getl(&j); Qv[i][0]=j;
     R_getl(&j); Qv[i][1]=j;
     R_getl(&j); Qv[i][2]=j;
   }
 }
 else if(strcmp(str,"AXIS") == 0){/* Axis is assummed to be first chumk */
   R_getl(&j); Mq[0] = j; /* this is the morphed object offset */
   R_getl(&j); Mq[1] = j; /* before using the morph it must be reposition */
   R_getl(&j); Mq[2] = j; /* to have its offset the same as the object */
   for(i=0;i<9;i++)R_getl(&j);
 }
 else R_getchunk(CHUNKsize);
 goto loop;
end:
 CloseHandle(fp);
 return  2;
 X__ERROR:
 return -1;
}

// END of MORPHING code

#ifdef _SUNSTYLE
static long *RenderReadSkeleton(clen,version,sp) long clen;
 short version; skel **sp; {
#else
static long *RenderReadSkeleton(long clen, short version, skel **sp){
#endif
 short s;
 skel *Skeleton;
 long n,i,j,k,l,l0,l1,l2,*jj,*Skid;
 char name[32];
 Skeleton=NULL;
 Skid=NULL;
 R_getl(&n);
 if((*sp=(skel *)X__Malloc(n*(long)sizeof(skel))) != NULL)Skeleton=*sp;
 for(i=0;i<n;i++){ /* dump the original skeleton */
   R_getl(&l0);  R_getl(&l1);  R_getl(&l2);  R_getsht(&s);
   if(Skeleton != NULL){
     (Skeleton+i)->id=(long)s;
     (Skeleton+i)->p[0] = (double)(l0);
     (Skeleton+i)->p[1] = (double)(l1);
     (Skeleton+i)->p[2] = (double)(l2);
   }
   if(version > 0){
     ReadFile(fp,name,sizeof(char)*16,&dwRead,NULL);
   }
   if(version > 1){
     for(j=0;j<3;j++){R_getl(&l);}
     for(j=0;j<3;j++){R_getl(&l);}
     for(j=0;j<3;j++){R_getl(&l);}
     for(j=0;j<8;j++)for(k=0;k<3;k++)R_getl(&l);
   }
 }
 R_getl(&n);  /* get number of vertices in skeleton structure */
 if(n > 0){
   if(n == newvert){ /* if skeleton assignment is consitent with newvert */
     if((Skid=(long *)X__Malloc(n*(long)sizeof(long))) != NULL)jj=Skid;
   }
   for(i=0;i<n;i++){
     R_getsht(&s);
     if(Skid != NULL)*jj++ = (long)s;
   }
 }
 return Skid;
}

typedef struct tVECTOR4tag {
 tDOUBLE x,y,z,w;
 BOOL   selected;
 char   pad[4];
} tvector4;

static long RenderReadNurbs(long clen){
 DWORD dwRead;
 BOOL readit=TRUE;
 long i,j,k,N;
 nurbs *n,*Np;
 vector4 *p;
 tvector4 P;
 tDOUBLE tD;
//{ char ttt[256];
//sprintf(ttt,"size of prop %ld  vec4=%ld tvec4=%ld vec=%ld BOOL=%ld tD=%ld",
//sizeof(NurbsProperties),
//sizeof(vector4),sizeof(tvector4),
//sizeof(vector),
//sizeof(BOOL),sizeof(tDOUBLE));
//MessageBox(NULL,ttt,"Write",MB_OK);
//}
 ReadFile(fp,&N,sizeof(long),&dwRead,NULL);
 if(N == 0)return -1;
 if((Np=(nurbs *)X__Malloc(N*sizeof(nurbs))) == NULL)return -1;
 for(k=0;k<N;k++){
   n=(Np+k);
   ReadFile(fp,&(n->properties),sizeof(NurbsProperties),&dwRead,NULL);
   ReadFile(fp,&(n->numU),sizeof(long),&dwRead,NULL);
   ReadFile(fp,&(n->numV),sizeof(long),&dwRead,NULL);
   ReadFile(fp,&(n->orderU),sizeof(long),&dwRead,NULL);
   ReadFile(fp,&(n->orderV),sizeof(long),&dwRead,NULL);
   AllocNurbs(n,NULL,NULL);
   for(i=0;i<n->numU + n->orderU; i++){
     ReadFile(fp,&tD,sizeof(tDOUBLE),&dwRead,NULL);
     n->kvU[i]=(double)tD;
   }
   for(i=0;i<n->numV + n->orderV; i++){
     ReadFile(fp,&tD,sizeof(tDOUBLE),&dwRead,NULL);
     n->kvV[i]=(double)tD;
   }
   for(i=0;i<n->numV; i++)
   for(j=0;j<n->numU; j++){
     ReadFile(fp,&P,sizeof(tvector4),&dwRead,NULL);
     p=&(n->points[i][j]);
     p->x=(double)P.x;
     p->y=(double)P.y;
     p->z=(double)P.z;
     p->w=(double)P.w;
   }
 }
 if(readit){
   Object[ObjectCount].Nbase=Np;
   Object[ObjectCount].NoNurbs=N;
 }
 return 0;
}

static void R_GetAnimFileName(char *fsurf, char *tempstr, long frame,
                       long firstframe, long lastframe, long framestep){
 long thisframe;
 char *cp1,*cp2;
 cp1=strchr(fsurf,'0');
 cp2=strchr(fsurf,'.');
 if(cp1 == NULL || cp2 == NULL){
 Render_Message(27,0,NULL);
   sprintf(tempstr,"dummy.evi");
 }
 else{
   int ln;
   char *cp,extn[16],fmat[16];
   strcpy(extn,cp2);
   ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
   *cp1='\0';
   framestep=min(1,framestep);
   thisframe=firstframe-1+frame/framestep;
   while(thisframe > lastframe)thisframe -= lastframe;
   sprintf(fmat,"%%s%%0%ldld%%s",ln);
   sprintf(tempstr,fmat,fsurf,thisframe,extn);
 }
 strcpy(fsurf,tempstr);
}

#ifdef _SUNSTYLE
static short R_ReadOldText(clen,frame,objectfilename,str) long clen; long frame;
char *objectfilename,*str; {
#else
static short R_ReadOldText(long clen, long frame, char *objectfilename,
                        char *str){
#endif
  imap *temp_map,*Map;
  unsigned char t_colour[3];
  char i,j,brushmaptype,brushmethod,moziac,p,r,q;
  char fsurf[256],frefl[256],fbump[256],tempstr[256];
  long k,l,m,firstframe,framedelay,lastframe,brushangle,file_len;
  short s,version;
  ReadFile(fp,&j,sizeof(char),&dwRead,NULL);
  if(j > 0){
  temp_map = (imap *)X__Malloc(j*sizeof(imap));
  if(temp_map == NULL){
    Render_Message(13,0,NULL);
    for(l=0;l<clen-5;l++) ReadFile(fp,&i,sizeof(char),&dwRead,NULL);
    return -1;
  }
  Object[ObjectCount].Mbase=temp_map;
  Object[ObjectCount].NoMaps=j;
  for(i=0;i<j;i++){
    Map=(temp_map+(long)i);
    version=0;
    R_getl(&file_len);
    ReadFile(fp,tempstr,file_len,&dwRead,NULL);
    R_getl(&l); Map->p[0]= (double)l;
    R_getl(&l); Map->p[1]= (double)l;
    R_getl(&l); Map->p[2]= (double)l;
    VECCOPY(Map->p,Map->pin)
    R_getl(&l); Map->x[0]= (double)l;
    R_getl(&l); Map->x[1]= (double)l;
    R_getl(&l); Map->x[2]= (double)l;
    VECCOPY(Map->x,Map->xin)
    R_getl(&l); Map->y[0]= (double)l;
    R_getl(&l); Map->y[1]= (double)l;
    R_getl(&l); Map->y[2]= (double)l;
    VECCOPY(Map->y,Map->yin)
    R_getl(&l); brushmethod=(char)l;   /* still bump etc from old version */
    R_getl(&firstframe); R_getl(&framedelay); R_getl(&lastframe);
    R_getl(&l); /* used by skeleton for br. assign. not used by renderer */
    R_getl(&l); brushmaptype=(char)l;
    R_getl(&l); /* this is the expansion count */
    if(l > 0){
      R_getl(&brushangle);
      if(l > 1){
        for(k=1;k<5;k++)R_getl(&m); /* dump the brush lock info */
        if(l > 5){             /* this must be a version 2 file */
          version=1;
          ReadFile(fp,tempstr,32,&dwRead,NULL);
          ReadFile(fp,&p,1,&dwRead,NULL);
          ReadFile(fp,&r,1,&dwRead,NULL);
          ReadFile(fp,&q,1,&dwRead,NULL);
          R_getsht(&s); Map->pp=s; /* these will be set to -1 */
          R_getsht(&s); Map->rp=s; /* if a map of this type   */
          R_getsht(&s); Map->bp=s; /* is not used in Material */
          ReadFile(fp,t_colour,3,&dwRead,NULL);  /* not used yet */
          ReadFile(fp,t_colour,3,&dwRead,NULL);
          ReadFile(fp,t_colour,3,&dwRead,NULL);  /* ------------ */
          if(p)ReadFile(fp,fsurf,file_len,&dwRead,NULL);
          if(r)ReadFile(fp,frefl,file_len,&dwRead,NULL);
          if(q)ReadFile(fp,fbump,file_len,&dwRead,NULL);
          R_getsht(&s); /* should be 0 use as expansion count */
          if(s > 0)for(k=0;k<s;k++)R_getl(&m);
        }
      }
    }
    else brushangle=360;
    moziac=NO;
    if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
    if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
    if(version == 0){
      p=r=q=0;
      Map->pp=100;
      Map->rp= 50;
      Map->bp= 50;
      if(brushmethod == REFLBRUSH){
        strcpy(frefl,tempstr); r=1;
      }
      else if(brushmethod == BUMPBRUSH){
        strcpy(fbump,tempstr); q=1;
      }
      else{
        strcpy(fsurf,tempstr); p=1;
      }
      if(brushmethod != ANIMBRUSH)brushmethod=STILBRUSH; /* just switch now */
    }
    if(p && brushmethod == ANIMBRUSH){ /* animated brush calculate frame */
       R_GetAnimFileName(fsurf,tempstr,frame,firstframe,lastframe,framedelay);
    }
    Map->tp = -1;   /* no transmissivity old brushed */
    Map->kt =  0;
    if(DROP == YES){
      if(p) strcpy(Map->filename,short_form(fsurf));
      else         Map->pp = -1;
      if(r) strcpy(Map->frefl,short_form(frefl));
      else         Map->rp = -1;
      if(q) strcpy(Map->fbump,short_form(fbump));
      else         Map->bp = -1;
    }
    else{
      if(p) strcpy(Map->filename,fsurf);
      else         Map->pp = -1;
      if(r) strcpy(Map->frefl,frefl);
      else         Map->rp = -1;
      if(q) strcpy(Map->fbump,fbump);
      else         Map->bp = -1;
    }
    if(SetupImageMap(Map,objectfilename) == 1){
      Map->map=(short)brushmaptype;
      Map->angle=360.0/(double)brushangle;
      Map->moziac=moziac;
    }
    else{
      Render_Message(28,0,NULL);
      if((m=Object[ObjectCount].NoFaces) > 0)for(l=0;l<m;l++){
        if((Object[ObjectCount].Fbase[l].fattrib & 0x40) == 0x40)/* map    */
         if(Object[ObjectCount].Fbase[l].fmap == i) /* this one */
            Object[ObjectCount].Fbase[l].fattrib &= 0xbf;
      } /* clear to shader leave smoothing and gloss alone */
    }
  }
  }
  return 0;
}

static short R_ReadImaps(long clen, long frame, char *objectfilename,
                        char *str){
  unsigned char *buffer=NULL,*b;
  imap *temp_map,*Map;
  unsigned char t_colour[3];
  char brushmaptype,moziac,p,r,q,t;
  char fsurf[256],frefl[256],fbump[256],ftran[256],tempstr[256];
  long i,j,k,l,m,brushangle;
  long m_p,ff_p,fd_p,lf_p,
       m_r,ff_r,fd_r,lf_r,
       m_b,ff_b,fd_b,lf_b,
       m_t,ff_t,fd_t,lf_t;
  short s;
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  j=GetLongFromBuffer(b); b+= 4;  /* read number of maps */
  if(j > 0){
    temp_map = (imap *)X__Malloc(j*sizeof(imap));
    if(temp_map == NULL){
      Render_Message(13,0,NULL);
      X__Free(buffer);
      return -1;
    }
    Object[ObjectCount].Mbase=temp_map;
    Object[ObjectCount].NoMaps=j;
    for(i=0;i<j;i++){
      Map=(temp_map+(long)i);
      Map->p[0]=(double)GetLongFromBuffer(b); b+= 4;
      Map->p[1]=(double)GetLongFromBuffer(b); b+= 4;
      Map->p[2]=(double)GetLongFromBuffer(b); b+= 4;
      VECCOPY(Map->p,Map->pin)
      Map->x[0]=(double)GetLongFromBuffer(b); b+= 4;
      Map->x[1]=(double)GetLongFromBuffer(b); b+= 4;
      Map->x[2]=(double)GetLongFromBuffer(b); b+= 4;
      VECCOPY(Map->x,Map->xin)
      Map->y[0]=(double)GetLongFromBuffer(b); b+= 4;
      Map->y[1]=(double)GetLongFromBuffer(b); b+= 4;
      Map->y[2]=(double)GetLongFromBuffer(b); b+= 4;
      VECCOPY(Map->y,Map->yin)
      m_p =GetLongFromBuffer(b); b+= 4;
      ff_p=GetLongFromBuffer(b); b+= 4;
      fd_p=GetLongFromBuffer(b); b+= 4;
      lf_p=GetLongFromBuffer(b); b+= 4; /* surface */
      m_r =GetLongFromBuffer(b); b+= 4;
      ff_r=GetLongFromBuffer(b); b+= 4;
      fd_r=GetLongFromBuffer(b); b+= 4;
      lf_r=GetLongFromBuffer(b); b+= 4; /* refl */
      m_b =GetLongFromBuffer(b); b+= 4;
      ff_b=GetLongFromBuffer(b); b+= 4;
      fd_b=GetLongFromBuffer(b); b+= 4;
      lf_b=GetLongFromBuffer(b); b+= 4; /* bump */
      m_t =GetLongFromBuffer(b); b+= 4;
      ff_t=GetLongFromBuffer(b); b+= 4;
      fd_t=GetLongFromBuffer(b); b+= 4;
      lf_t=GetLongFromBuffer(b); b+= 4; /* trans */
      l=GetLongFromBuffer(b); b+= 4; /* used by skeleton for br. assign. not used by renderer */
      l=GetLongFromBuffer(b); b+= 4; brushmaptype=(char)l;
      brushangle=GetLongFromBuffer(b); b+= 4;
      for(k=0;k<4;k++){
        m=GetLongFromBuffer(b); b+= 4; /* dump the brush lock info */
      }
      memcpy(tempstr,b,32); b+= 32;
      p = *b++;
      r = *b++;
      q = *b++;
      t = *b++;
      Map->kt = *b++;
      Map->pp=GetShortFromBuffer(b); b+= 2; /* these will be set to -1 */
      Map->rp=GetShortFromBuffer(b); b+= 2; /* if a map of this type   */
      Map->bp=GetShortFromBuffer(b); b+= 2; /* is not used in Material */
      Map->tp=0; b+=2; /* no transparency slider  */
      memcpy(t_colour,b,3); b+=3;  /* not used yet */
      memcpy(t_colour,b,3); b+=3;
      memcpy(t_colour,b,3); b+=3;  /* ------------ */
      memcpy(Map->kc,b,3); b+=3;   /* key colour   */
      memcpy(fsurf,b,256); b+=256;
      memcpy(frefl,b,256); b+=256;
      memcpy(fbump,b,256); b+=256;
      memcpy(ftran,b,256); b+=256;
      moziac=NO;
      if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
      if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
      if(p && m_p == ANIMBRUSH){ /* animated brush calculate frame */
        R_GetAnimFileName(fsurf,tempstr,frame,ff_p,lf_p,fd_p);
      }
      if(r && m_r == ANIMBRUSH){ /* animated brush calculate frame */
        R_GetAnimFileName(frefl,tempstr,frame,ff_r,lf_r,fd_r);
      }
      if(q && m_b == ANIMBRUSH){ /* animated brush calculate frame */
        R_GetAnimFileName(fbump,tempstr,frame,ff_b,lf_b,fd_b);
      }
      if(t && m_t == ANIMBRUSH){ /* animated brush calculate frame */
        R_GetAnimFileName(ftran,tempstr,frame,ff_t,lf_t,fd_t);
      }
      if(p) strcpy(Map->filename,fsurf); else Map->pp = -1;
      if(r) strcpy(Map->frefl,frefl);    else Map->rp = -1;
      if(q) strcpy(Map->fbump,fbump);    else Map->bp = -1;
      if(t) strcpy(Map->ftran,ftran);    else Map->tp = -1;
      if(SetupImageMap(Map,objectfilename) == 1){
        Map->map=(short)brushmaptype;
        Map->angle=360.0/(double)brushangle;
        Map->moziac=moziac;
        if(Map->tp >= 0){ /* set face trans to facilitiate calculation */
          if((m=Object[ObjectCount].NoFaces) > 0)for(l=0;l<m;l++){
            if(Object[ObjectCount].Fbase[l].mattr == 0) /* not trans */
              if((Object[ObjectCount].Fbase[l].fattrib & 0x40) == 0x40)/* map */
                if(Object[ObjectCount].Fbase[l].fmap == i) /* this one */
                  Object[ObjectCount].Fbase[l].mattr=1;
          }
        }
      }
      else{
        Render_Message(28,0,NULL);
        if((m=Object[ObjectCount].NoFaces) > 0)for(l=0;l<m;l++){
          if((Object[ObjectCount].Fbase[l].fattrib & 0x40) == 0x40)/* map    */
           if(Object[ObjectCount].Fbase[l].fmap == i) /* this one */
              Object[ObjectCount].Fbase[l].fattrib &= 0xbf;
        } /*  clear to shader leave smoothing and gloss alone */
      }
    }
  }
  X__Free(buffer);
  return 0;
}

static short R_ReadMorphShader(long clen, char *objectfilename,
                        char *str){
 long i,m,l,ns,fs,ps,in_use;  /* ns is max of 4 */
  char tempname[128],tempparam[256];
  R_getl(&ns);
  if(ns > 0)for(i=0;i<ns;i++){
    R_getl(&in_use);
    if(in_use){
      R_getl(&fs); R_getl(&ps);
      ReadFile(fp,tempname,fs,&dwRead,NULL);
      if(ps > 0){
        if(ps > 256){R_getchunk(ps); return 0;}
        else ReadFile(fp,tempparam,ps,&dwRead,NULL);
      }
      MorphExternalTexture(tempname,tempparam,ObjectCount);
    }
  }
  return 0;
}

static short R_ReadShader(long clen, long frame, char *objectfilename,
                        char *str){
 long i,m,l,ns,fs,ps,in_use;  /* ns is max of 4 */
  char tempname[128],tempparam[256];
  R_getl(&ns);
  if(ns > 0)for(i=0;i<ns;i++){
    Object[ObjectCount].shader[i]=-1;
    R_getl(&in_use);
    if(in_use){
      R_getl(&fs); R_getl(&ps);
      ReadFile(fp,tempname,fs,&dwRead,NULL);
      if(ps > 0){
        if(ps > 256){R_getchunk(ps); goto DROPIT;}
        else ReadFile(fp,tempparam,ps,&dwRead,NULL);
      }
      if(LoadExternalTexture(tempname,tempparam) == OK){
        Object[ObjectCount].shader[i]=Nshader-1;
      }
      else{
        DROPIT:
        Render_Message(92,0,NULL);
        if((m=Object[ObjectCount].NoFaces) > 0)for(l=0;l<m;l++){
          if((Object[ObjectCount].Fbase[l].fattrib & 0x10) == 0x10)/* ext shad */
           if(Object[ObjectCount].Fbase[l].fshader == (1 << i)){  /* this one */
             Object[ObjectCount].Fbase[l].fattrib &= 0xef;
             Object[ObjectCount].Fbase[l].fshader=0;
           }
        } /* clear all face shaders to plain if any for this face missing */
      }
    }
  }
  return 0;
}

#ifdef _SUNSTYLE
long R_LoadObject(FileName,frame,robo,robo_ratio,robo_file1,robo_file2)
char *FileName; long frame; short robo; double robo_ratio;
char *robo_file1; char *robo_file2; {
#else
long R_LoadObject(char *FileName, long frame, short robo, double robo_ratio,
                  char *robo_file1, char *robo_file2){
#endif
 long CHUNKsize,FORMsize,l;
 long *Skids=NULL;
 skel *Skeleton=NULL;
 Object[ObjectCount].Fbase=NULL;
 Object[ObjectCount].Vbase=NULL;
 Object[ObjectCount].Vcopy=NULL;
 Object[ObjectCount].Ebase=NULL;
 Object[ObjectCount].Pbase=NULL;
 Object[ObjectCount].Tbase=NULL;
 Object[ObjectCount].Mbase=NULL;
 Object[ObjectCount].Nbase=NULL;
 Object[ObjectCount].Abase=NULL;
 Object[ObjectCount].mirror_buffer=NULL;
 Object[ObjectCount].NoVertices=0;
 Object[ObjectCount].NoFaces=0;
 Object[ObjectCount].NoMaps=0;
 Object[ObjectCount].NoAxes=0;
 Object[ObjectCount].NoNurbs=0;
 Object[ObjectCount].mirrorID=-1;
 if((fp=GetRenderRelativeHandle((char *)strlwr(FileName),NULL,NULL))
       == INVALID_HANDLE_VALUE){
   Render_Message(30,1,FileName);
   return 0;
 }
 newface=0; newvert=0;
 ReadFile(fp,str,4,&dwRead,NULL);
 if(strcmp(str,"FORM") != 0){
   Render_Message(21,1,NULL);
   return 0;
 }
 ReadFile(fp,&FORMsize,sizeof(long),&dwRead,NULL);
 ReadFile(fp,str,4,&dwRead,NULL);
 // allow old models
 if(dwRead != 4 || (strcmp(str,"OFXM") != 0  && strcmp(str,"AAPO") != 0)){
   Render_Message(22,1,NULL);
   return 0;
 }
 Object[ObjectCount].Origin[0]=0;   /* initialize in case no axis */
 Object[ObjectCount].Origin[1]=0;   /* present */
 Object[ObjectCount].Origin[2]=0;
 if((Object[ObjectCount].Abase=(axis *)X__Malloc(sizeof(axis)))
                            != NULL)Object[ObjectCount].NoAxes=1;
 else {
   CloseHandle(fp);
   return -1;
 }
 Object[ObjectCount].Abase[0].bo[0]=0;
 Object[ObjectCount].Abase[0].bo[1]=0;
 Object[ObjectCount].Abase[0].bo[2]=0;
 Object[ObjectCount].Abase[0].bu[0]=UNIT;
 Object[ObjectCount].Abase[0].bu[1]=0;
 Object[ObjectCount].Abase[0].bu[2]=0;
 Object[ObjectCount].Abase[0].bv[0]=0;
 Object[ObjectCount].Abase[0].bv[1]=UNIT;
 Object[ObjectCount].Abase[0].bv[2]=0;
loop:
 if(!ReadFile(fp,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4)goto end;
 if (R_getl(&CHUNKsize) != 1)goto end;
 if     (strcmp(str,"VERT") == 0){if(R_ReadVertices(CHUNKsize)   < 0)return -1;}
 else if(strcmp(str,"VTEX") == 0){if(R_ReadVertText(CHUNKsize)   < 0)return -1;}
 else if(strcmp(str,"SFCE") == 0){if(R_ReadFaces(CHUNKsize,TRUE) < 0)return -1;}
 else if(strcmp(str,"LFCE") == 0){if(R_ReadFaces(CHUNKsize,FALSE)< 0)return -1;}
 else if(strcmp(str,"FACE") == 0){if(R_ReadOldFaces(CHUNKsize,0) < 0)return -1;}
 else if(strcmp(str,"SFAC") == 0){if(R_ReadOldFaces(CHUNKsize,1) < 0)return -1;}
 else if(strcmp(str,"AXIS") == 0){if(R_ReadAxis(CHUNKsize)       < 0)return -1;}
 else if(strcmp(str,"AXES") == 0){if(R_ReadAxes(CHUNKsize)       < 0)return -1;}
 else if(strcmp(str,"TEXT") == 0 && hidden_line == NO){
   R_ReadOldText(CHUNKsize,frame,FileName,str);
 }
 else if(strcmp(str,"IMAP") == 0 && hidden_line == NO){
   R_ReadImaps(CHUNKsize,frame,FileName,str);
 }
 else if(strcmp(str,"SHDR") == 0 && hidden_line == NO){
   R_ReadShader(CHUNKsize,frame,FileName,str);
 }
 else if(strcmp(str,"MPID") == 0){
   R_getl(&l);
   if(trace_reflections == NO)Object[ObjectCount].mirrorID=(short)l;
 }
 else if(strcmp(str,"MATL") == 0){
   R_getl(&l);  /* used as count for future expansion */
   if(l == 4){
     R_getl(&l); /* shadow flag  1 = YES */
     if(l == 1 && number_of_shadows < 7 && trace_shadows == NO)
       Object[ObjectCount].shadowed = ++number_of_shadows;
   }
   R_getl(&l); if(l > 1000)Object[ObjectCount].refractive_index=(double)l/10000;
   R_getl(&l);
   R_getl(&l);
 }
 else if(strcmp(str,"SK02") == 0 && robo == YES){
   Skids=RenderReadSkeleton(CHUNKsize,2,&Skeleton);
 }
 else if(strcmp(str,"SK01") == 0 && robo == YES){
   Skids=RenderReadSkeleton(CHUNKsize,1,&Skeleton);
 }
 else if(strcmp(str,"SK01") == 0 && robo == YES){
   Skids=RenderReadSkeleton(CHUNKsize,1,&Skeleton);
 }
 else if(strcmp(str,"SKEL") == 0 && robo == YES){
   Skids=RenderReadSkeleton(CHUNKsize,0,&Skeleton);
 }
 else if(strcmp(str,"NURB") == 0){
   if(RenderReadNurbs(CHUNKsize) < 0)return -1;
   MeshNurbsSurfaces(ObjectCount);
 }
 else R_getchunk(CHUNKsize);
 goto loop;
end:
 CloseHandle(fp);
 if(Object[ObjectCount].NoFaces > 0){
  long i;  // test to make sure object counts are OK
  face *f=Object[ObjectCount].Fbase;
  for(i=0;i<Object[ObjectCount].NoFaces;i++){
   if(f[i].faxis >= Object[ObjectCount].NoAxes)f[i].faxis=0;
  }
 }
 if(robo == YES && Skids != NULL && Skeleton != NULL)
   RenderTransformToRobot(ObjectCount,
                           Skids,Skeleton,
                           robo_ratio,
                           robo_file1,robo_file2);
 return 1;
}

void  R_getchunk(long clen){
 SetFilePointer(fp,clen,NULL,FILE_CURRENT);
}

#ifdef _SUNSTYLE
static short R_ReadAxis(clen) long clen; {
#else
static short R_ReadAxis(long clen){
#endif
 long j;
 R_getl(&Object[ObjectCount].Origin[0]); /* origin of Object (in object coords */
 R_getl(&Object[ObjectCount].Origin[1]);
 R_getl(&Object[ObjectCount].Origin[2]);
 R_getl(&j); Object[ObjectCount].Abase[0].bo[0]=j;  /* shader axis point */
 R_getl(&j); Object[ObjectCount].Abase[0].bo[1]=j;
 R_getl(&j); Object[ObjectCount].Abase[0].bo[2]=j;
 R_getl(&j); Object[ObjectCount].Abase[0].bu[0]=j;  /* texture axis off x */
 R_getl(&j); Object[ObjectCount].Abase[0].bu[1]=j;
 R_getl(&j); Object[ObjectCount].Abase[0].bu[2]=j;
 R_getl(&j); Object[ObjectCount].Abase[0].bv[0]=j;  /* texture axis off y */
 R_getl(&j); Object[ObjectCount].Abase[0].bv[1]=j;
 R_getl(&j); Object[ObjectCount].Abase[0].bv[2]=j;
 VECCOPY(Object[ObjectCount].Abase[0].bo,Object[ObjectCount].Abase[0].boin)
 VECCOPY(Object[ObjectCount].Abase[0].bu,Object[ObjectCount].Abase[0].buin)
 VECCOPY(Object[ObjectCount].Abase[0].bv,Object[ObjectCount].Abase[0].bvin)
 return 0;
}

#ifdef _SUNSTYLE
static short R_ReadAxes(clen) long clen; {
#else
static short R_ReadAxes(long clen){
#endif
 long i,j,k,n;
 char str[32];
 axis *NewAxis;
 if(Object[ObjectCount].NoAxes != 1)return FAIL;
 n=clen/68;
 if((NewAxis=(axis *)X__Realloc(Object[ObjectCount].Abase,
     (n+1)*sizeof(axis))) == NULL)return FAIL;
 Object[ObjectCount].Abase=NewAxis;
 Object[ObjectCount].NoAxes=n+1;
 for(i=1;i<n+1;i++){
   ReadFile(fp,str,32,&dwRead,NULL);
   for(j=0;j<3;j++){R_getl(&k); Object[ObjectCount].Abase[i].bo[j]=(double)k;}
   for(j=0;j<3;j++){R_getl(&k); Object[ObjectCount].Abase[i].bu[j]=(double)k;}  /* texture axis off x */
   for(j=0;j<3;j++){R_getl(&k); Object[ObjectCount].Abase[i].bv[j]=(double)k;}  /* texture axis off y */
   VECCOPY(Object[ObjectCount].Abase[i].bo,Object[ObjectCount].Abase[i].boin)
   VECCOPY(Object[ObjectCount].Abase[i].bu,Object[ObjectCount].Abase[i].buin)
   VECCOPY(Object[ObjectCount].Abase[i].bv,Object[ObjectCount].Abase[i].bvin)
 }
 return 0;
}

static short R_ReadOldFaces(long  clen, long shader){
  long i,sf,se,sp;
  long id,mm,me,mp;
  unsigned char texture,brush;
  unsigned char *buffer=NULL,*b;
  face   *f;
  eface  *e;
  pface  *p;
  if(shader)newface=clen/21;
  else      newface=clen/20;
#if __DEMO__
  if(newface > 25000){
   Render_Message(18,0,NULL);
   return -1;
 }
#endif
  if(newface == 0)return -1;
  sf=sizeof(face);
  mm=newface*(long)sf;
  se=sizeof(eface);
  me=newface*(long)se;
  sp=sizeof(pface);
  mp=newface*(long)sp;
  Object[ObjectCount].Fbase=(face  *)X__Malloc(mm);
  Object[ObjectCount].NoFaces=newface;
  f=Object[ObjectCount].Fbase;
  if(f == NULL){Render_Message(9,0,NULL); return -1;}
  Object[ObjectCount].Ebase=(eface  *)X__Malloc(me);
  e=Object[ObjectCount].Ebase;
  if(e == NULL){Render_Message(9,0,NULL); return -1;}
  Object[ObjectCount].Pbase=(pface  *)X__Malloc(mp);
  p=Object[ObjectCount].Pbase;
  if(p == NULL){Render_Message(9,0,NULL); return -1;}
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  for(i=0;i<newface;i++){
    f->V[0]=(UNSIGNED)GetLongFromBuffer(b); b+= 4;
    f->V[1]=(UNSIGNED)GetLongFromBuffer(b); b+= 4;
    f->V[2]=(UNSIGNED)GetLongFromBuffer(b); b+= 4;
    memcpy(f->color,b,3); b+=3;
    texture  = *b++;
    f->matsp = *b++;
    f->mattr = *b++;
    f->matre = *b++;
    brush    = *b++;
    if(shader){
      f->faxis = *b++;
    }
    else  f->faxis=0;
    f->fattrib=(texture & 0xf0) | (brush >> 4); f->fattrib &= 0xfe;
    f->fmap = (brush & 0x1f);
    f->fshader=(texture & 0xf);
    f->texco[0] = (f->matsp & 0x0f) << 4;
    f->texco[1] = (f->mattr & 0x0f) << 4;
    f->texco[2] = (f->matre & 0x0f) << 4;
    f->matsp  = (f->matsp & 0xf0) >> 4;   /* 0 - 15 posssible specs */
    f->mattr &= 0xf0;                     /* scale 0->240           */
    f->matre &= 0xf0;                     /* scale 0->240           */
    if((texture & 0x1f) == 0x04){ /* old glass glass */
      f->fshader=0; f->fattrib |= 0x1; f->fparam=0;
    }
    else{
      f->fparam= f->mattr &= 0xf0; f->mattr = 0;
    }
    if(hidden_line == YES){ /* switch off shaders and maps              */
                            /* and make black white for hidden callback */
      if(hidden_callback){
        f->color[0]=255; f->color[1]=255; f->color[2]=255;
      }
      f->fattrib=0x80; f->fmap=0; f->fshader=0;
    }
    p->n[0][0]=0.0; p->n[0][1]=0.0; p->n[0][2]=0.0;
    p->n[1][0]=0.0; p->n[1][1]=0.0; p->n[1][2]=0.0;
    p->n[2][0]=0.0; p->n[2][1]=0.0; p->n[2][2]=0.0;
    e->n[0]=0.0; e->n[1]=0.0; e->n[2]=0.0;
    e->c=0.0;
    f++; e++;         /* used as flag for effects and visible faces before!*/
  }
  X__Free(buffer);
  return 0;
}

static short R_ReadFaces(long  clen, BOOL short_form){
  long i,sf,se,sp;
  long mm,me,mp;
  unsigned char texture,brush,pad;
  unsigned char *buffer=NULL,*b;
  face   *f;
  eface  *e;
  pface  *p;
  if(short_form)newface=clen/18;
  else          newface=clen/24;
#if __DEMO__
  if(newface > 25000){
   Render_Message(18,0,NULL);
   return -1;
 }
#endif
  if(newface == 0)return -1;
  sf=sizeof(face);
  mm=newface*(long)sf;
  se=sizeof(eface);
  me=newface*(long)se;
  sp=sizeof(pface);
  mp=newface*(long)sp;
  Object[ObjectCount].Fbase=(face  *)X__Malloc(mm);
  Object[ObjectCount].NoFaces=newface;
  f=Object[ObjectCount].Fbase;
  if(f == NULL){Render_Message(9,0,NULL); return -1;}
  Object[ObjectCount].Ebase=(eface  *)X__Malloc(me);
  e=Object[ObjectCount].Ebase;
  if(e == NULL){Render_Message(9,0,NULL); return -1;}
  Object[ObjectCount].Pbase=(pface  *)X__Malloc(mp);
  p=Object[ObjectCount].Pbase;
  if(p == NULL){Render_Message(9,0,NULL); return -1;}
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  for(i=0;i<newface;i++){
    if(short_form){
      f->V[0]=(UNSIGNED)GetUshortFromBuffer(b); b+= 2;
      f->V[1]=(UNSIGNED)GetUshortFromBuffer(b); b+= 2;
      f->V[2]=(UNSIGNED)GetUshortFromBuffer(b); b+= 2;
    }
    else{
      f->V[0]=(UNSIGNED)GetLongFromBuffer(b); b+= 4;
      f->V[1]=(UNSIGNED)GetLongFromBuffer(b); b+= 4;
      f->V[2]=(UNSIGNED)GetLongFromBuffer(b); b+= 4;
    }
    memcpy(f->color,b,3); b+=3;
    f->matsp   = *b++;
    f->mattr   = *b++;
    f->matre   = *b++;
    f->fattrib = *b++;
    f->fparam  = *b++;
    f->fmap    = *b++;
    f->fshader = *b++;
    f->faxis   = *b++;
    pad        = *b++;
    if(hidden_line == YES){ /* switch off shaders and maps              */
                            /* and make black white for hidden callback */
      if(hidden_callback){
        f->color[0]=255; f->color[1]=255; f->color[2]=255;
      }
      f->fattrib=0x80; f->fmap=0; f->fshader=0; f->faxis=0;
    }
    f->texco[0] = (f->matsp & 0x0f) << 4;
    f->texco[1] = (f->mattr & 0x0f) << 4;
    f->texco[2] = (f->matre & 0x0f) << 4;
    f->matsp  = (f->matsp & 0xf0) >> 4;   /* 0 - 15 posssible specs */
    f->mattr &= 0xf0;                     /* scale 0->240           */
    f->matre &= 0xf0;                     /* scale 0->240           */
    p->n[0][0]=0.0; p->n[0][1]=0.0; p->n[0][2]=0.0;
    p->n[1][0]=0.0; p->n[1][1]=0.0; p->n[1][2]=0.0;
    p->n[2][0]=0.0; p->n[2][1]=0.0; p->n[2][2]=0.0;
    e->n[0]=0.0; e->n[1]=0.0; e->n[2]=0.0;
    e->c=0.0;
    f++; e++;
  }
  X__Free(buffer);
  return 0;
}

#ifdef _SUNSTYLE
static short R_ReadVertices(clen) long clen; {
#else
static short R_ReadVertices(long clen){
#endif
  vertex   *v;
  long i,mm,sv;
  unsigned char *buffer=NULL,*b;
  newvert=clen/12;
  if(newvert == 0)return -1;
  sv=sizeof(vertex);
  mm=newvert*(long)sv;
#if __DEMO__
  if(newvert > 15000){
   Render_Message(18,0,NULL);
   return -1;
 }
#endif
  if(newvert > 65535L && sizeof(UNSIGNED) < 4){
    Render_Message(16,0,NULL);
    MessageBox(NULL,"The scene contains a model with > 65536 vertices, "
                    "you must split them up into smaller components.",
                     "OpenFX Renderer",MB_OK);
    return -1;
  }
  Object[ObjectCount].Vbase=(vertex *)X__Malloc(mm);
  Object[ObjectCount].NoVertices=newvert;
  v=Object[ObjectCount].Vbase;
  if(v == NULL){
    Render_Message(9,0,NULL);
    return -1;
  }
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  for(i=0;i<newvert;i++){
    v->p[0]=(double)GetLongFromBuffer(b); b+= 4;
    v->p[1]=(double)GetLongFromBuffer(b); b+= 4;
    v->p[2]=(double)GetLongFromBuffer(b); b+= 4;
    v->x=0; v->y=0;
    v++;
  }
  X__Free(buffer);
  return 0;
}

static short R_ReadVertText(long clen){
  tvertex *t;
  long i,n,xx,id;
  unsigned char *buffer=NULL,*b;
  n=clen/12;
  i=Object[ObjectCount].NoVertices;
  if(n == 0 || n > i)return -1;
  Object[ObjectCount].Tbase=(tvertex *)X__Malloc(i*(long)sizeof(tvertex));
  t=Object[ObjectCount].Tbase;
  if(t == NULL){ Render_Message(9,0,NULL); return -1; }
  memset(t,0,i*sizeof(tvertex));
  buffer=(unsigned char *)X__Malloc(clen);
  if(buffer == NULL)return -1;
  ReadFile(fp,buffer,clen,&dwRead,NULL);
  b=buffer;
  for(i=0;i<n;i++){
    id=GetLongFromBuffer(b); b+= 4;
    xx=GetLongFromBuffer(b); b+= 4; t[id].x=(double)xx*0.000001;
    xx=GetLongFromBuffer(b); b+= 4; t[id].y=(double)xx*0.000001;
  }
  X__Free(buffer);
  return 0;
}

long FindNumberVertices(char *ofile, long *nvs){
 long CHUNKsize,FORMsize;
 char FileName[256];
 strcpy(FileName,ofile);
 *nvs = -1;
 if((fp=GetRenderRelativeHandle((char *)strlwr(FileName),NULL,NULL))
       == INVALID_HANDLE_VALUE){
   return 0;
 }
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || strcmp(str,"FORM") != 0)goto end;
 ReadFile(fp,&FORMsize,4,&dwRead,NULL);
 ReadFile(fp,str,4,&dwRead,NULL);
 if (dwRead != 4 || (strcmp(str,"OFXM") != 0 && strcmp(str,"AAPO") != 0))goto end;
loop:
 if(!ReadFile(fp,str,4,&dwRead,NULL))goto end;
 if(dwRead != 4 )goto end;
 if(R_getl(&CHUNKsize) != 1)goto end;
 if(strcmp(str,"VERT") == 0){
   *nvs=CHUNKsize/12;
   goto end;
 }
 else R_getchunk(CHUNKsize);
 goto loop;
end:
 CloseHandle(fp);
 return 1;
}

#define EPSILON 1.0e-6

void InvertRotationMatrix(double T[4][4], double t[4][4]){
 int i,j;
 double a[3][3];
 double det;
 R_null_transform(t);
 for(i=0;i<3;i++)for(j=0;j<3;j++)a[i][j]=T[i][j];
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
}

#ifdef _SUNSTYLE
#else
static void MatrixToQuaternion(double m[4][4], quaternion q){
#endif
 double w4,x2;
 q[0]=0.25*(1.0+m[0][0]+m[1][1]+m[2][2]);
 if(q[0] > EPSILON){
   q[0]=sqrt(q[0]);
   w4=1.0/(4.0*q[0]);
   q[1]=(m[1][2]-m[2][1])*w4;
   q[2]=(m[2][0]-m[0][2])*w4;
   q[3]=(m[0][1]-m[1][0])*w4;
 }
 else{
   q[0]=0.0;
   q[1]= -0.5*(m[1][1]+m[2][2]);
   if(q[1] > EPSILON){
     q[1]=sqrt(q[1]);
     x2=1.0/(2.0*q[1]);
     q[2]=m[0][1]*x2;
     q[3]=m[0][2]*x2;
   }
   else{
     q[1]=0.0;
     q[2]=0.5*(1.0-m[2][2]);
     if(q[2] > EPSILON){
       q[2]=sqrt(q[2]);
       q[3]=m[1][2]/(2.0*q[2]);
     }
     else{
       q[2]=0.0;
       q[3]=1.0;
     }
   }
 }
 return;
}

#ifdef _SUNSTYLE
#else
static void QuaternionToMatrix(quaternion q, double m[4][4]){
#endif
 double w,x,y,z,x2,y2,z2;
 w=q[0]; x=q[1]; y=q[2]; z=q[3]; x2=x*x; y2=y*y; z2=z*z;
 m[0][0]=1.0-2*y2-2*z2; m[0][1]=2*x*y+2*w*z;   m[0][2]=2*x*z-2*w*y;
 m[1][0]=2*x*y-2*w*z;   m[1][1]=1.0-2*x2-2*z2; m[1][2]=2*y*z+2*w*x;
 m[2][0]=2*x*z+2*w*y;   m[2][1]=2*y*z-2*w*x;   m[2][2]=1.0-2*x2-2*y2;
 m[3][3]=1.0; m[0][3]=m[1][3]=m[2][3]=m[3][0]=m[3][1]=m[3][2]=0.0;
 return;
}

#ifdef _SUNSTYLE
#else
static void Slerp(double mr, quaternion q1, quaternion q2, quaternion q){
#endif
 double sum,beta1,beta2,theta;
 sum=q1[0]*q2[0]+q1[1]*q2[1]+q1[2]*q2[2]+q1[3]*q2[3];
 if(sum < 0.0)sum=0.0;
 if(sum > 1.0)sum=1.0;
 theta=acos(sum);
 if(theta < EPSILON){ beta1=1.0-mr; beta2=mr;}
 else{
  beta1=sin((1.0-mr)*theta)/sin(theta);
  beta2=sin(mr*theta)/sin(theta);
 }
 q[0]=beta1*q1[0]+beta2*q2[0];
 q[1]=beta1*q1[1]+beta2*q2[1];
 q[2]=beta1*q1[2]+beta2*q2[2];
 q[3]=beta1*q1[3]+beta2*q2[3];
}

static int nSkeleton;
skel   *Skeleton;

#ifdef _SUNSTYLE
#else
static void ApplyTransformToChildren(int id, double T[4][4]){
#endif
 static double x,y,z;
 static double sst[4][4],t1[4][4],t2[4][4];
 int i;
 for(i=1;i<nSkeleton;i++){ /* must start at one since root is never a child */
   if(Skeleton[i].id == id){
     R_tram(t1,-(Skeleton+id)->p[0],-(Skeleton+id)->p[1],-(Skeleton+id)->p[2]);
     R_m4by4((Skeleton+i)->Q,t1,t2);
     R_tram(t1, (Skeleton+id)->p[0], (Skeleton+id)->p[1], (Skeleton+id)->p[2]);
     R_m4by4(t1,t2,sst);
     R_m4by4(T,sst,Skeleton[i].T);
     ApplyTransformToChildren(i,Skeleton[i].T);
   }
 }
 return;
}

#ifdef _SUNSTYLE
#else
static void InterpolateRobot(double mr, long n, skel *sk1, skel *sk2){
#endif
 int i;
 quaternion q,q1,q2;
 nSkeleton=n;
 Skeleton=sk1;
 if(mr < 1.0){
   /* interpopate rotation matrices using Quaternion tweening */
#if 0
   for(i=0;i<nSkeleton;i++){ /* this interpolation always from original - bad? */
     MatrixToQuaternion((sk2+i)->R,q1);
     MatrixToQuaternion((Skeleton+i)->R,q2);
     Slerp(mr,q1,q2,q);
     QuaternionToMatrix(q,(Skeleton+i)->Q);
   }
#else
   /* this interpolation always from last frame rotation as origin */
   quaternion Qi={1.0,0.0,0.0,0.0};
   double Q[4][4],Ri[4][4];
   for(i=0;i<nSkeleton;i++){
     InvertRotationMatrix((sk2+i)->R,Ri);
     R_m4by4((Skeleton+i)->R,Ri,Q);
     MatrixToQuaternion(Q,q2);
     Slerp(mr,Qi,q2,q);
     QuaternionToMatrix(q,Q);
     R_m4by4(Q,(sk2+i)->R,(Skeleton+i)->Q);
   }
#endif
 }
 else{
   for(i=0;i<nSkeleton;i++)R_c4to4((Skeleton+i)->R,(Skeleton+i)->Q);
 }
 ApplyTransformToChildren(0,Skeleton[0].T);
 return;
}

#ifdef _SUNSTYLE
static void RenderTransformToRobot(O,Skid,Skelton,mr,robo_f1,robo_f2) long O;
long *Skid; skel *Skeleton; double mr; char *robo_f1; char *robo_f2; {
#else
static void RenderTransformToRobot(long O,
                             long *Skid, skel *Skel, double mr,
                             char *robo_f1, char *robo_f2){
#endif
  long i,j,n1=0,n2=0;
  FILE *frobot;
  skel *sk1=NULL,*sk2=NULL;
  vertex *v;
  double x1,y1,z1;
  if((frobot = fopen(robo_f1,"r")) != NULL){
    fscanf(frobot,"%ld\n",&n1);
    if(n1 > 0){  /* sk1 is the Robot  sk2 is version is moving from */
      if((sk1=Skel) == NULL)goto FINISHED;
      for(i=0;i<n1;i++){
        R_null_transform(sk1->R);
        R_null_transform(sk1->Q);
        R_null_transform(sk1->T);
        for(j=0;j<4;j++)fscanf(frobot,GET_MATX,
         &sk1[i].R[j][0], &sk1[i].R[j][1],
         &sk1[i].R[j][2], &sk1[i].R[j][3]);
      }
    }
    fclose(frobot);
  }
  if((frobot = fopen(robo_f2,"r")) != NULL){
    fscanf(frobot,"%ld\n",&n2);
    if(n2 > 0){
      if((sk2 = (skel *)X__Malloc(n2*(long)sizeof(skel))) == NULL)
           goto FINISHED;
      for(i=0;i<n2;i++){
        R_null_transform(sk2->R);
        R_null_transform(sk2->Q);
        R_null_transform(sk2->T);
        for(j=0;j<4;j++)fscanf(frobot,GET_MATX,
         &sk2[i].R[j][0], &sk2[i].R[j][1],
         &sk2[i].R[j][2], &sk2[i].R[j][3]);
      }
    }
    fclose(frobot);
  }
  if(n1 != n2 || n1 == 0 || n2 == 0)goto FINISHED;
  InterpolateRobot(mr,n1,sk1,sk2);
  v=Object[O].Vbase;
  for(i=0;i<Object[O].NoVertices;i++){
    R_m4by1(sk1[Skid[i]].T,v[i].p[0],v[i].p[1],v[i].p[2],&x1,&y1,&z1);
    v[i].p[0]=x1;
    v[i].p[1]=y1;
    v[i].p[2]=z1;
  }
  FINISHED:
  if(sk1 != NULL)X__Free(sk1);
  if(sk2 != NULL)X__Free(sk2);
  X__Free(Skid);
  return;
}

