/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// file material.c  

#define MODULE_TEXTURE

/* Calculate the surface colour for given Pixel now that we know which     */
/* face it is on.  Intepolate the PHONG vector first. The ground is an     */
/* independent entity so that its caclulations can be optimised.           */

/* most shaders have a function to caclulate their value                   */

#include "render.h"

#ifdef DOS386
extern short mirror_ground;
#endif
extern double Rip_Freq;

#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))

#ifdef _SUNSTYLE
static void Transform_to_Moziac();
static void GetPseudoMirror();
static void AttenuateGlassSurface();
static void AttenuateGlass();
static void PutBackPlanes();
#else
static void Transform_to_Moziac(double, double ,double *, double *);
static void GetPseudoMirror(long, long, long, double *);
static void AttenuateGlassSurface(long , long , long ,
                                  UCHAR *, UCHAR *, UCHAR *,
                                  long *[], long *[], double *[], double *);
static void AttenuateGlass(long, long,
                    UCHAR *, UCHAR *, UCHAR *Blue,
                    long *[], long *[], double *[], double *);
static void PutBackPlanes(long, long, double, long *,
                   UCHAR *, UCHAR *, UCHAR *,
                   double *,
                   long *, double *,
                   UCHAR *, long, double);
static void GetBackdropValue(long row, long col, vector p, vector d,
                             double *colour);
#endif

double spec_pow[16][max_spec+1],spec_gro[max_spec+1];

static double spec_val[16]={
       10.0, 15.0, 20.0, 25.0, 27.0, 30.0, 32.0, 35.0,
       40.0, 45.0, 50.0, 55.0, 60.0, 70.0, 80.0, 99.0};

#ifdef _SUNSTYLE
void init_spec(){
#else
void init_spec(void){
#endif
 long i,j; /* use LUT for specular highlight function to speed up rendering */
 double d;
 for(j=0;j<16;j++){
   spec_pow[j][max_spec] = 1.0;
   for(i=0;i<max_spec;i++){
     d = pow(((double)i/(double)max_spec),spec_val[j]);
     spec_pow[j][i]=d;
   }
 }
#ifdef __AXP__
 for(i=0;i<max_spec;i++)spec_gro[i] = pow((1.0/max_spec*(double)i),120.0);
#else
 for(i=0;i<max_spec;i++)spec_gro[i] = pow((1.0/max_spec*(double)i),200.0);
#endif
}


/* The functions for Bump and Reflection Mapping  */

/* Map->bumpX/Y is a scale factor to scale the change in normal            */
/* vector n is unity, max gradient is 255 -> scale factor is               */
/* 1/|x| * max_displ_of normal / 255 * height /100                         */
/* max_displ = tan(a) where a is the max angle the normal is bent          */
/* For Cylinder maps the X axis is defined by n X y which is normalized    */
/* so the appropriate scaling factor is different                          */

#ifdef _SUNSTYLE
void map_bumpy(type,obj,Map,p,n,bumpyness) short type;
long obj; imap *Map; vector p; vector n; {
#else
void map_bumpy(short type, long obj, imap *Map, vector p, vector n){
#endif
 long  i,j,ialpha,ibeta,k,kl,kr,kt,kb,pixel,ix,iy;
 double alpha0,beta0,alpha,beta,a,b,red[3],cc0,cl0,cr0,ct0,cb0;
 static double dalpha_bump[3]={0.0,0.01,0.0},
                dbeta_bump[3]={0.0,0.0,0.01};
 vector v1,v2;
 unsigned char *pP,*pR;
 j=Map->xmaxbump;
 if(Map->map == CYLINDER){
   hitcylinder(Map->n,Map->x,Map->y,Map->p
              ,p,&alpha0,&beta0);
   alpha0 *= Map->angle;
 }
 else if(Map->map == MAP_SPHERICAL){
   hitsphere(Map->n,Map->x,Map->y,Map->p
              ,p,&alpha0,&beta0);
 }
 else    /* assume planar map */
   hitpoint(Map->n,Map->x,Map->y,Map->p
           ,p,&alpha0,&beta0);

 for(i=0;i<3;i++){
 alpha = alpha0+dalpha_bump[i];
 beta  = beta0+dbeta_bump[i];
 if(type == TILE){
   if(Map->moziac)Transform_to_Moziac(alpha,beta,&a,&b);
   else{
     a=(alpha-floor(alpha));
     b=(beta-floor(beta));
   }
   alpha=a*Map->xmaxbump;
   beta=b*Map->ymaxbump;
 }
 else{
   alpha=Map->xmaxbump*alpha;
   beta =Map->ymaxbump*beta;
 }
 ialpha = (long)alpha; alpha -= (double)ialpha;
 ibeta  = (long)beta;  beta  -= (double)ibeta;
 if(ialpha < 0 || ialpha >= Map->xmaxbump)return;
 if(ibeta  < 0 || ibeta  >= Map->ymaxbump)return;
 k =ibeta*j+ialpha;                       /* do the bi-linear interpolation */
 if(type == TILE){
   if(Map->moziac){
     if(ialpha <   1)kl=1;                     else kl=k-1;
     if(ialpha > j-2)kr=k-1;                   else kr=k+1;
     if(ibeta  <   1)kb=k+j;                   else kb=k-j;
     if(ibeta  > Map->ymaxbump-2)kt=k-j; else kt=k+j;
   }
   else{
     if(ialpha <   1)kl=k+j-1;                            else kl=k-1;
     if(ialpha > j-2)kr=ibeta*j;                          else kr=k+1;
     if(ibeta  <   1)kb=(Map->ymaxbump-1)*j+ialpha; else kb=k-j;
     if(ibeta  > Map->ymaxbump-2)kt=ialpha;         else kt=k+j;
   }
 }
 else{
   if(ialpha <   1)kl=k; else kl=k-1;
   if(ialpha > j-2)kr=k; else kr=k+1;
   if(ibeta  <   1)kb=k; else kb=k-j;
   if(ibeta  > Map->ymaxbump-2)kt=k; else kt=k+j;
 }
 if(Map->dbump == 24){
   pR=Map->p24Gbump;
 }
 else{
   pP=Map->pbump;  k =pP[k];
   kl=pP[kl]; kr=pP[kr]; kt=pP[kt]; kb=pP[kb];
   pR=Map->palGbump;
 }
 cc0 = (double)pR[k];
 cl0 = (double)pR[kl];
 cr0 = (double)pR[kr];
 ct0 = (double)pR[kt];
 cb0 = (double)pR[kb];
 if(alpha > 0.5){
   alpha = alpha-0.5;
   if(beta > 0.5){
     beta = beta-0.5;
     red[i] = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
   }
   else{
     beta = 0.5-beta;
     red[i] = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
   }
 }
 else{
   alpha = 0.5-alpha;
   if(beta > 0.5){
     beta = beta-0.5;
     red[i] = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
   }
   else{
     beta = 0.5-beta;
     red[i] = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
   }
 }
 }
 if(Map->map == CYLINDER || Map->map == MAP_SPHERICAL){
   cross(n,Map->y,v1);
   normalize(v1);
   vecscale((red[1]-red[0])*Map->bumpx,v1,v1);
   vecscale((red[2]-red[0])*Map->bumpy,Map->y,v2);
 }
 else{
   vecscale((red[1]-red[0])*Map->bumpx,Map->x,v1);
   vecscale((red[2]-red[0])*Map->bumpy,Map->y,v2);
 }
 vecsum(n,v1,n);
 vecsum(n,v2,n);
 normalize(n);
 return;
}

#ifdef _SUNSTYLE
void BumpByCoordinates(Map,obj,f,p,n,bumpyness) imap *Map,
long obj; face *f; vector p; vector n; {
#else
void BumpByCoordinates(imap *Map, long obj, face *f, vector p, vector n){
#endif
 long  i,j,ialpha,ibeta,k,kl,kr,kt,kb,pixel,ix,iy;
 double alpha0,beta0,alpha,beta,a,b,red[3],cc0,cl0,cr0,ct0,cb0,
        ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3,
        detmax,dm,det1,det2,det3,
        t1x,t2x,t1y,t2y;
 tvertex  *T,*t0,*t1,*t2;
 vertex   *V,*v0,*v1,*v2;
 static double dalpha_bump[3]={0.0,0.01,0.0},
                dbeta_bump[3]={0.0,0.0,0.01};
 vector vv1,vv2;
 unsigned char *pP,*pR;
 j=Map->xmaxbump;
 if((T=Object[obj].Tbase) == NULL)return;
 if((V=Object[obj].Vbase) == NULL)return;
 v0=(V+(f->V[0])); t0=(T+(f->V[0]));
 v1=(V+(f->V[1])); t1=(T+(f->V[1]));
 v2=(V+(f->V[2])); t2=(T+(f->V[2]));
 ve1=v1->p[0] - v0->p[0]; ve2=v1->p[1] - v0->p[1]; ve3=v1->p[2] - v0->p[2];
 vp1=v2->p[0] - v0->p[0]; vp2=v2->p[1] - v0->p[1]; vp3=v2->p[2] - v0->p[2];
 p1=p[0]-v0->p[0];
 p2=p[1]-v0->p[1];
 p3=p[2]-v0->p[2];
 det1=ve1*vp2-vp1*ve2;
 det2=ve1*vp3-vp1*ve3;
 det3=ve2*vp3-vp2*ve3;
 k=0; detmax=1.e-3;  /* same as TOL in mirrors */
 if((dm=fabs(det1)) > detmax){k=1; detmax=dm;}
 if((dm=fabs(det2)) > detmax){k=2; detmax=dm;}
 if((dm=fabs(det3)) > detmax){k=3; detmax=dm;}
 if(k == 0)return;
 else if(k == 1){
   a=( vp2*p1-vp1*p2)/det1;
   b=(-ve2*p1+ve1*p2)/det1;
 }
 else if(k == 2){
   a=( vp3*p1-vp1*p3)/det2;
   b=(-ve3*p1+ve1*p3)/det2;
 }
 else if(k == 3){
   a=( vp3*p2-vp2*p3)/det3;
   b=(-ve3*p2+ve2*p3)/det3;
 }
 t0=T+(f->V[0]); t1=T+(f->V[1]); t2=T+(f->V[2]);
 t1x=t1->x - t0->x; t2x=t2->x - t0->x;
 t1y=t1->y - t0->y; t2y=t2->y - t0->y;
 if(fabs(t1x) > 4.0)t1x=0.5;
 if(fabs(t2x) > 4.0)t2x=0.5;
 if(fabs(t1y) > 4.0)t1y=0.5;
 if(fabs(t2y) > 4.0)t2y=0.5;
 alpha0 = t0->x + a*t1x + b*t2x;
 beta0  = t0->y + a*t1y + b*t2y;
/* alpha0 = t0->x + a*(t1->x - t0->x) + b*(t2->x - t0->x); */
/* beta0  = t0->y + a*(t1->y - t0->y) + b*(t2->y - t0->y); */
 for(i=0;i<3;i++){
   alpha = alpha0+dalpha_bump[i];
   beta  = beta0+dbeta_bump[i];
   a=(alpha-floor(alpha));
   b=1.0-(beta-floor(beta));
   alpha=a*Map->xmaxbump;
   beta=b*Map->ymaxbump;
   ialpha = (long)alpha; alpha -= (double)ialpha;
   ibeta  = (long)beta;  beta  -= (double)ibeta;
   if(ialpha < 0 || ialpha >= Map->xmaxbump)return;
   if(ibeta  < 0 || ibeta  >= Map->ymaxbump)return;
   k =ibeta*j+ialpha;                       /* do the bi-linear interpolation */
   if(ialpha <   1)kl=k+j-1;                            else kl=k-1;
   if(ialpha > j-2)kr=ibeta*j;                          else kr=k+1;
   if(ibeta  <   1)kb=(Map->ymaxbump-1)*j+ialpha; else kb=k-j;
   if(ibeta  > Map->ymaxbump-2)kt=ialpha;         else kt=k+j;
   if(Map->dbump == 24){
     pR=Map->p24Gbump;
   }
   else{
     pP=Map->pbump;  k =pP[k];
     kl=pP[kl]; kr=pP[kr]; kt=pP[kt]; kb=pP[kb];
     pR=Map->palGbump;
   }
   cc0 = (double)pR[k];
   cl0 = (double)pR[kl];
   cr0 = (double)pR[kr];
   ct0 = (double)pR[kt];
   cb0 = (double)pR[kb];
   if(alpha > 0.5){
     alpha = alpha-0.5;
     if(beta > 0.5){
       beta = beta-0.5;
       red[i] = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     }
     else{
       beta = 0.5-beta;
       red[i] = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     }
   }
   else{
     alpha = 0.5-alpha;
     if(beta > 0.5){
       beta = beta-0.5;
       red[i] = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     }
     else{
       beta = 0.5-beta;
       red[i] = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     }
   }
 }
 vecscale((red[1]-red[0])*Map->bumpx,Map->x,vv1);
 vecscale((red[2]-red[0])*Map->bumpy,Map->y,vv2);
 vecsum(n,vv1,n);
 vecsum(n,vv2,n);
 normalize(n);
 return;
}

#if 0
#ifdef _SUNSTYLE
void MapMirror(R,G,B,color,Map,p,n) double *R,*G,*B; imap *Map;
vector p; vector n; {
#else
void MapMirror(double *R,double *G, double *B, imap *Map, vector p, vector n){
#endif
 /* p is point on mirror n is normal  d is reflection direction */
 long pixel,k,ibeta,ialpha;
 double pdotn,alpha,beta,d12;
 vector d,dd;
 pdotn=dot(p,n);
 vecscale(2*pdotn,n,d);
/* vecsub(p,d,d);   use this to cut out the AntiCamera transform */
 vecsub(p,d,dd);   /* un camera the camera transform */
 R_m4by1(AntiCamera,dd[0],dd[1],dd[2],&d[0],&d[1],&d[2]);
 alpha=0.159152777*(1.5707963+atan2(d[1],d[0]))           /* get reflection */
      *(double)Map->xmaxrefl;                       /* vector angles  */
 ialpha=( Map->xmaxrefl/2 + (long)alpha ) % Map->xmaxrefl;
 d12=sqrt(d[0]*d[0] + d[1]*d[1]);
 beta=0.31830555*(1.5707963+atan2(d[2],d12))              /* + pixel in map */
     *(double)Map->ymaxrefl;
 ibeta=Map->ymaxrefl-((long)beta) % Map->ymaxrefl;
  k=ibeta*Map->xmaxrefl+ialpha;
 if(Map->drefl == 24){                              /* is it 24 bit ? */
   *R = (double)Map->p24Rrefl[k];
   *G = (double)Map->p24Grefl[k];
   *B = (double)Map->p24Brefl[k];
 }
 else{
   pixel = Map->prefl[k];                           /* 8 bit          */
   *R = (double)Map->palRrefl[ pixel ];
   *G = (double)Map->palGrefl[ pixel ];
   *B = (double)Map->palBrefl[ pixel ];
 }
 return;
}
#endif

void MapMirror(double *R,double *G, double *B, imap *Map, vector p, vector n){
 /* p is input direction n is normal  d is reflection direction */
 /* map 4 images round the sphere and 2 up down                 */
 /* atan2 returns in range -pi to pi                            */

//#define FOURDIV2PI 0.636619772
//#define TWODIVPI  0.318309886
#define FOURDIV2PI 0.318309886     /* 2 images in 360 rotation */
//#define TWODIVPI   0.636619772     /* 2 image in -90 to +90    */
#define TWODIVPI   0.318309886     /* 1 image in -90 to +90    */

 long j,pixel,k,ibeta,ialpha,kl,kr,kt,kb;
 double pdotn,alpha,beta,d12;
 double cc0,cc1,cc2,cl0,cl1,cl2,cr0,cr1,cr2,ct0,ct1,ct2,cb0,cb1,cb2;
 unsigned char *pP,*pR,*pG,*pB;
 vector d,dd;
 pdotn=dot(p,n);
 vecscale(2*pdotn,n,d);
 vecsub(p,d,dd);   /* un camera the camera transform */
 R_m4by1(AntiCamera,dd[0],dd[1],dd[2],&d[0],&d[1],&d[2]);
 d12=sqrt(d[0]*d[0] + d[1]*d[1]);

// alpha=fmod(FOURDIV2PI*(3.141592654+atan2(d[1],d[0])),1.0);
// beta=fmod(TWODIVPI*(1.570796327+atan2(d[2],d12)),1.0);
 alpha=FOURDIV2PI*(3.141592654+atan2(d[1],d[0]));
 beta=1.0-TWODIVPI*(1.570796327+atan2(d[2],d12));
 Transform_to_Moziac(alpha,beta,&alpha,&beta);

 alpha *= (double)Map->xmaxrefl;
 beta *= (double)Map->ymaxrefl;
 ialpha = (long)alpha; alpha -= (double)ialpha;
 ibeta=(long)beta; beta -= (double)ibeta;
 if(ialpha < 0 || ialpha >= Map->xmaxrefl)return;
 if(ibeta  < 0 || ibeta  >= Map->ymaxrefl)return;
 j=Map->xmaxrefl;
 k =ibeta*j+ialpha;

// if(ialpha <   1)kl=k+j-1;                      else kl=k-1;
// if(ialpha > j-2)kr=ibeta*j;                    else kr=k+1;
// if(ibeta  <   1)kb=(Map->ymaxrefl-1)*j+ialpha; else kb=k-j;
// if(ibeta  > Map->ymaxrefl-2)kt=ialpha;         else kt=k+j;
 if(ialpha <   1)kl=1;                 else kl=k-1;
 if(ialpha > j-2)kr=k-1;               else kr=k+1;
 if(ibeta  <   1)kb=k+j;               else kb=k-j;
 if(ibeta  > Map->ymaxrefl-2)kt=k-j;   else kt=k+j;

 if(Map->drefl == 24){
   pR=Map->p24Rrefl; pG=Map->p24Grefl; pB=Map->p24Brefl;
 }
 else{
   pP=Map->prefl;  k =pP[k];
   kl=pP[kl]; kr=pP[kr]; kt=pP[kt]; kb=pP[kb];
   pR=Map->palRrefl; pG=Map->palGrefl; pB=Map->palBrefl;
 }
 cc0 = (double)pR[k];  cc1 = (double)pG[k];  cc2 = (double)pB[k];
 cl0 = (double)pR[kl]; cl1 = (double)pG[kl]; cl2 = (double)pB[kl];
 cr0 = (double)pR[kr]; cr1 = (double)pG[kr]; cr2 = (double)pB[kr];
 ct0 = (double)pR[kt]; ct1 = (double)pG[kt]; ct2 = (double)pB[kt];
 cb0 = (double)pR[kb]; cb1 = (double)pG[kb]; cb2 = (double)pB[kb];
 if(alpha > 0.5){
   alpha = alpha-0.5;
   if(beta > 0.5){
     beta = beta-0.5;
     *R   = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *G   = (cr1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *B   = (cr2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *R   = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *G   = (cr1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *B   = (cr2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 else{
   alpha = 0.5-alpha;
   if(beta > 0.5){
     beta = beta-0.5;
     *R   = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *G   = (cl1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *B   = (cl2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *R   = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *G   = (cl1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *B   = (cl2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
}

/* end the Bump and reflection Mapping functions  */

/* The functions for Image mapping and fast       */
/* anti-aliasing line texturing                   */

/* functions for the image mapping   - some old code is still here (#if 0) */

#ifdef _SUNSTYLE
static void Transform_to_Moziac(alpha,beta,a,b) double alpha;
double beta; double *a; double *b;  {
#else
static void Transform_to_Moziac(double alpha, double beta,
                                double *a, double *b){
#endif
 long brkindx,brkindy;
 *a=FLOOR(alpha);  *b=FLOOR(beta);
 brkindx=((long)(*a) & 1L);
 brkindy=((long)(*b) & 1L);
 *a=(alpha - *a);  *b=(beta - *b);
 if(brkindx){
   if(brkindy){ *a = 1.0 - *a;  *b = 1.0 - *b; }
   else       { *a = 1.0 - *a; }
 }
 else if(brkindy){ *b = 1.0 - *b; }
 return;
}

/* bilinear interpolation  */
#ifdef _SUNSTYLE
long MapBrush(Map,type,clear,p,red,green,blue) imap *Map; short type;
short clear; vector p; double *red; double *green; double *blue; {
#else
long MapBrush(imap *Map, short type, short clear,
                        vector p , double *red
                                 , double *green
                                 , double *blue){
#endif
 /* return colour for map Map at point p */
 long  ialpha,ibeta,i,j,k,kl,kr,kt,kb,pixel;
 double alpha,beta,a,b,cc0,cc1,cc2,cl0,cl1,cl2,
           cr0,cr1,cr2,ct0,ct1,ct2,cb0,cb1,cb2;
 unsigned char *pP,*pR,*pG,*pB;
 /* checking for valid map must be done elsewhere */
 j=Map->xmax;
 if(Map->map == CYLINDER){
   hitcylinder(Map->n,Map->x,Map->y,Map->p
              ,p,&alpha,&beta);
   alpha *= Map->angle;
 }
 else if(Map->map == MAP_SPHERICAL){
   hitsphere(Map->n,Map->x,Map->y,Map->p
              ,p,&alpha,&beta);
 }
 else    /* assume planar map */
   hitpoint(Map->n,Map->x,Map->y,Map->p
           ,p,&alpha,&beta);
 if(type == TILE){
   if(Map->moziac)Transform_to_Moziac(alpha,beta,&a,&b);
   else{
     a=(alpha-floor(alpha));
     b=(beta-floor(beta));
   }
   alpha=a*Map->xmax;
   beta=b*Map->ymax;
 }
 else{
   alpha=Map->xmax*alpha;
   beta =Map->ymax*beta;
 }
 ialpha = (long)alpha; alpha -= (double)ialpha;
 ibeta  = (long)beta;  beta  -= (double)ibeta;
 if(ialpha < 0 || ialpha >= Map->xmax)return 0;
 if(ibeta  < 0 || ibeta  >= Map->ymax)return 0;
 k =ibeta*j+ialpha;
 if(type == TILE){
   if(Map->moziac){
     if(ialpha <   1)kl=1;                 else kl=k-1;
     if(ialpha > j-2)kr=k-1;               else kr=k+1;
     if(ibeta  <   1)kb=k+j;               else kb=k-j;
     if(ibeta  > Map->ymax-2)kt=k-j; else kt=k+j;
   }
   else{
     if(ialpha <   1)kl=k+j-1;                        else kl=k-1;
     if(ialpha > j-2)kr=ibeta*j;                      else kr=k+1;
     if(ibeta  <   1)kb=(Map->ymax-1)*j+ialpha; else kb=k-j;
     if(ibeta  > Map->ymax-2)kt=ialpha;         else kt=k+j;
   }
 }
 else{
   if(ialpha <   1)kl=k; else kl=k-1;
   if(ialpha > j-2)kr=k; else kr=k+1;
   if(ibeta  <   1)kb=k; else kb=k-j;
   if(ibeta  > Map->ymax-2)kt=k; else kt=k+j;
 }
 if(Map->depth == 24){
   pR=Map->p24R; pG=Map->p24G; pB=Map->p24B;
   if(clear == TRANSP){
//     if(pR[k] == colour0_red && pG[k] == colour0_green && // not used
//        pB[k] == colour0_blue)return 0;
     if(pR[k] == Map->kc[0] && pG[k] == Map->kc[1] &&  pB[k] == Map->kc[2])
       return 0;
   }
 }
 else{
   pP=Map->picture;  k =pP[k];
   kl=pP[kl]; kr=pP[kr]; kt=pP[kt]; kb=pP[kb];
   pR=Map->palR; pG=Map->palG; pB=Map->palB;
   if(clear == TRANSP){
//     if(colour0_flag == 0){
     if(Map->kt == 0){
       if(k == 0)return 0;
     }
//     else if(pR[k] == colour0_red && pG[k] == colour0_green &&
//             pB[k] == colour0_blue)return 0;
     else if(pR[k] == Map->kc[0] && pG[k] == Map->kc[1] &&
             pB[k] == Map->kc[2])return 0;
   }
 }
 cc0 = (double)pR[k];  cc1 = (double)pG[k];  cc2 = (double)pB[k];
 cl0 = (double)pR[kl]; cl1 = (double)pG[kl]; cl2 = (double)pB[kl];
 cr0 = (double)pR[kr]; cr1 = (double)pG[kr]; cr2 = (double)pB[kr];
 ct0 = (double)pR[kt]; ct1 = (double)pG[kt]; ct2 = (double)pB[kt];
 cb0 = (double)pR[kb]; cb1 = (double)pG[kb]; cb2 = (double)pB[kb];
 if(alpha > 0.5){
   alpha = alpha-0.5;
   if(beta > 0.5){
     beta = beta-0.5;
     *red   = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *red   = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 else{
   alpha = 0.5-alpha;
   if(beta > 0.5){
     beta = beta-0.5;
     *red   = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *red   = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 return 1;
}

#ifdef _SUNSTYLE
long MapByCoordinates(Map,obj,f,type,clear,p,red,green,blue) imap *Map; long obj;
face *f; short type; short clear; vector p; double *red; double *green; double *blue; {
#else

long MapByCoordinates(imap *Map,
                        long obj, face *f,
                        short type, short clear,
                        vector p,
                        double *red,
                        double *green,
                        double *blue){
#endif
 /* return colour for map pointed to by Map at point p on face f of object # obj */
 long  ialpha,ibeta,i,j,k,kl,kr,kt,kb,pixel;
 double alpha,beta,a,b,cc0,cc1,cc2,cl0,cl1,cl2,
           cr0,cr1,cr2,ct0,ct1,ct2,cb0,cb1,cb2,
           ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3,
           detmax,dm,det1,det2,det3,
           t1x,t2x,t1y,t2y;
 unsigned char *pP,*pR,*pG,*pB;
 tvertex  *T,*t0,*t1,*t2;
 vertex   *V,*v0,*v1,*v2;
 /* checking for valid map must be done elsewhere */
 j=Map->xmax;
 if((T=Object[obj].Tbase) == NULL)return 0;
 if((V=Object[obj].Vbase) == NULL)return 0;
 v0=(V+(f->V[0])); t0=(T+(f->V[0]));
 v1=(V+(f->V[1])); t1=(T+(f->V[1]));
 v2=(V+(f->V[2])); t2=(T+(f->V[2]));
 ve1=v1->p[0] - v0->p[0]; ve2=v1->p[1] - v0->p[1]; ve3=v1->p[2] - v0->p[2];
 vp1=v2->p[0] - v0->p[0]; vp2=v2->p[1] - v0->p[1]; vp3=v2->p[2] - v0->p[2];
 p1=p[0]-v0->p[0];
 p2=p[1]-v0->p[1];
 p3=p[2]-v0->p[2];
 det1=ve1*vp2-vp1*ve2;
 det2=ve1*vp3-vp1*ve3;
 det3=ve2*vp3-vp2*ve3;
 k=0; detmax=1.e-3;  /* same as TOL in mirrors */
 if((dm=fabs(det1)) > detmax){k=1; detmax=dm;}
 if((dm=fabs(det2)) > detmax){k=2; detmax=dm;}
 if((dm=fabs(det3)) > detmax){k=3; detmax=dm;}
 if(k == 0)return 0;
 else if(k == 1){
   a=( vp2*p1-vp1*p2)/det1;
   b=(-ve2*p1+ve1*p2)/det1;
 }
 else if(k == 2){
   a=( vp3*p1-vp1*p3)/det2;
   b=(-ve3*p1+ve1*p3)/det2;
 }
 else if(k == 3){
   a=( vp3*p2-vp2*p3)/det3;
   b=(-ve3*p2+ve2*p3)/det3;
 }
 t0=T+(f->V[0]); t1=T+(f->V[1]); t2=T+(f->V[2]);
 t1x=t1->x - t0->x; t2x=t2->x - t0->x;
 t1y=t1->y - t0->y; t2y=t2->y - t0->y;
 if(fabs(t1x) > 4.0)t1x=0.5;
 if(fabs(t2x) > 4.0)t2x=0.5;
 if(fabs(t1y) > 4.0)t1y=0.5;
 if(fabs(t2y) > 4.0)t2y=0.5;
 alpha = t0->x + a*t1x + b*t2x;
 beta  = t0->y + a*t1y + b*t2y;
 if(type != TILE){
   if(alpha < 0 || alpha >= 1.0)return 0;
   if(beta  < 0 || beta  >= 1.0)return 0;
 }
/* alpha = t0->x + a*(t1->x - t0->x) + b*(t2->x - t0->x); */
/* beta  = t0->y + a*(t1->y - t0->y) + b*(t2->y - t0->y); */
 a=(alpha-floor(alpha));
 b=1.0-(beta-floor(beta));  /* 1-  to have origin at bottom left */
/* b=(beta-floor(beta)); for origin at top left */
 alpha=a*Map->xmax;
 beta=b*Map->ymax;
 ialpha = (long)alpha; alpha -= (double)ialpha;
 ibeta  = (long)beta;  beta  -= (double)ibeta;
 if(ialpha < 0 || ialpha >= Map->xmax)return 0;
 if(ibeta  < 0 || ibeta  >= Map->ymax)return 0;
 k =ibeta*j+ialpha;
 if(ialpha <   1)kl=k+j-1;                        else kl=k-1;
 if(ialpha > j-2)kr=ibeta*j;                      else kr=k+1;
 if(ibeta  <   1)kb=(Map->ymax-1)*j+ialpha; else kb=k-j;
 if(ibeta  > Map->ymax-2)kt=ialpha;         else kt=k+j;
 if(Map->depth == 24){
   pR=Map->p24R; pG=Map->p24G; pB=Map->p24B;
   if(clear == TRANSP){
//     if(pR[k] == colour0_red && pG[k] == colour0_green &&
//        pB[k] == colour0_blue)return 0;
     if(pR[k] == Map->kc[0] && pG[k] == Map->kc[1] &&  pB[k] == Map->kc[2])
       return 0;
   }
 }
 else{
   pP=Map->picture;  k =pP[k];
   kl=pP[kl]; kr=pP[kr]; kt=pP[kt]; kb=pP[kb];
   pR=Map->palR; pG=Map->palG; pB=Map->palB;
   if(clear == TRANSP){
//     if(colour0_flag == 0){
     if(Map->kt == 0){
       if(k == 0)return 0;
     }
//     else if(pR[k] == colour0_red && pG[k] == colour0_green &&
//             pB[k] == colour0_blue)return 0;
     else if(pR[k] == Map->kc[0] && pG[k] == Map->kc[1] &&
             pB[k] == Map->kc[2])return 0;
   }
 }
 cc0 = (double)pR[k];  cc1 = (double)pG[k];  cc2 = (double)pB[k];
 cl0 = (double)pR[kl]; cl1 = (double)pG[kl]; cl2 = (double)pB[kl];
 cr0 = (double)pR[kr]; cr1 = (double)pG[kr]; cr2 = (double)pB[kr];
 ct0 = (double)pR[kt]; ct1 = (double)pG[kt]; ct2 = (double)pB[kt];
 cb0 = (double)pR[kb]; cb1 = (double)pG[kb]; cb2 = (double)pB[kb];
 if(alpha > 0.5){
   alpha = alpha-0.5;
   if(beta > 0.5){
     beta = beta-0.5;
     *red   = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *red   = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 else{
   alpha = 0.5-alpha;
   if(beta > 0.5){
     beta = beta-0.5;
     *red   = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *red   = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 return 1;
}

long MapTransByAxis(imap *Map, short type,
                    vector p , double *trans){
 long  ialpha,ibeta,i,j,k,kl,kr,kt,kb,pixel;
 double red,green,blue;
 double alpha,beta,a,b,cc0,cc1,cc2,cl0,cl1,cl2,
           cr0,cr1,cr2,ct0,ct1,ct2,cb0,cb1,cb2;
 unsigned char *pP,*pR,*pG,*pB;
 /* checking for valid map must be done elsewhere */
 j=Map->xmaxtran;
 if(Map->map == CYLINDER){
   hitcylinder(Map->n,Map->x,Map->y,Map->p
              ,p,&alpha,&beta);
   alpha *= Map->angle;
 }
 else    /* assume planar map */
   hitpoint(Map->n,Map->x,Map->y,Map->p
           ,p,&alpha,&beta);
 if(type == TILE){
   if(Map->moziac)Transform_to_Moziac(alpha,beta,&a,&b);
   else{
     a=(alpha-floor(alpha));
     b=(beta-floor(beta));
   }
   alpha=a*Map->xmaxtran;
   beta=b*Map->ymaxtran;
 }
 else{
   alpha=Map->xmaxtran*alpha;
   beta =Map->ymaxtran*beta;
 }
 ialpha = (long)alpha; alpha -= (double)ialpha;
 ibeta  = (long)beta;  beta  -= (double)ibeta;
 if(ialpha < 0 || ialpha >= Map->xmaxtran)return 0;
 if(ibeta  < 0 || ibeta  >= Map->ymaxtran)return 0;
 k =ibeta*j+ialpha;
 if(type == TILE){
   if(Map->moziac){
     if(ialpha <   1)kl=1;                 else kl=k-1;
     if(ialpha > j-2)kr=k-1;               else kr=k+1;
     if(ibeta  <   1)kb=k+j;               else kb=k-j;
     if(ibeta  > Map->ymaxtran-2)kt=k-j; else kt=k+j;
   }
   else{
     if(ialpha <   1)kl=k+j-1;                        else kl=k-1;
     if(ialpha > j-2)kr=ibeta*j;                      else kr=k+1;
     if(ibeta  <   1)kb=(Map->ymaxtran-1)*j+ialpha; else kb=k-j;
     if(ibeta  > Map->ymaxtran-2)kt=ialpha;         else kt=k+j;
   }
 }
 else{
   if(ialpha <   1)kl=k; else kl=k-1;
   if(ialpha > j-2)kr=k; else kr=k+1;
   if(ibeta  <   1)kb=k; else kb=k-j;
   if(ibeta  > Map->ymaxtran-2)kt=k; else kt=k+j;
 }
 if(Map->depth == 24){
   pR=Map->p24Rtran; pG=Map->p24Gtran; pB=Map->p24Btran;
 }
 else{
   pP=Map->ptran;  k =pP[k];
   kl=pP[kl]; kr=pP[kr]; kt=pP[kt]; kb=pP[kb];
   pR=Map->palRtran; pG=Map->palGtran; pB=Map->palBtran;
 }
 cc0 = (double)pR[k];  cc1 = (double)pG[k];  cc2 = (double)pB[k];
 cl0 = (double)pR[kl]; cl1 = (double)pG[kl]; cl2 = (double)pB[kl];
 cr0 = (double)pR[kr]; cr1 = (double)pG[kr]; cr2 = (double)pB[kr];
 ct0 = (double)pR[kt]; ct1 = (double)pG[kt]; ct2 = (double)pB[kt];
 cb0 = (double)pR[kb]; cb1 = (double)pG[kb]; cb2 = (double)pB[kb];
 if(alpha > 0.5){
   alpha = alpha-0.5;
   if(beta > 0.5){
     beta = beta-0.5;
     red   = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     green = (cr1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     blue  = (cr2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     red   = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     green = (cr1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     blue  = (cr2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 else{
   alpha = 0.5-alpha;
   if(beta > 0.5){
     beta = beta-0.5;
     red   = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     green = (cl1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     blue  = (cl2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     red   = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     green = (cl1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     blue  = (cl2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 *trans = 1.3071895e-3*(red+green+blue); /* scaled 0-1 black 0% transparent */
 return 1;
}

long MapTransByCoords(imap *Map, long obj, face *f,
                      vector p , double *trans){
 long  ialpha,ibeta,i,j,k,kl,kr,kt,kb,pixel;
 double red,green,blue;
 double alpha,beta,a,b,cc0,cc1,cc2,cl0,cl1,cl2,
           cr0,cr1,cr2,ct0,ct1,ct2,cb0,cb1,cb2,
           ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3,
           detmax,dm,det1,det2,det3,
           t1x,t2x,t1y,t2y;
 unsigned char *pP,*pR,*pG,*pB;
 tvertex  *T,*t0,*t1,*t2;
 vertex   *V,*v0,*v1,*v2;
 /* checking for valid map must be done elsewhere */
 j=Map->xmaxtran;
 if((T=Object[obj].Tbase) == NULL)return 0;
 if((V=Object[obj].Vbase) == NULL)return 0;
 v0=(V+(f->V[0])); t0=(T+(f->V[0]));
 v1=(V+(f->V[1])); t1=(T+(f->V[1]));
 v2=(V+(f->V[2])); t2=(T+(f->V[2]));
 ve1=v1->p[0] - v0->p[0]; ve2=v1->p[1] - v0->p[1]; ve3=v1->p[2] - v0->p[2];
 vp1=v2->p[0] - v0->p[0]; vp2=v2->p[1] - v0->p[1]; vp3=v2->p[2] - v0->p[2];
 p1=p[0]-v0->p[0];
 p2=p[1]-v0->p[1];
 p3=p[2]-v0->p[2];
 det1=ve1*vp2-vp1*ve2;
 det2=ve1*vp3-vp1*ve3;
 det3=ve2*vp3-vp2*ve3;
 k=0; detmax=1.e-3;  /* same as TOL in mirrors */
 if((dm=fabs(det1)) > detmax){k=1; detmax=dm;}
 if((dm=fabs(det2)) > detmax){k=2; detmax=dm;}
 if((dm=fabs(det3)) > detmax){k=3; detmax=dm;}
 if(k == 0)return 0;
 else if(k == 1){
   a=( vp2*p1-vp1*p2)/det1;
   b=(-ve2*p1+ve1*p2)/det1;
 }
 else if(k == 2){
   a=( vp3*p1-vp1*p3)/det2;
   b=(-ve3*p1+ve1*p3)/det2;
 }
 else if(k == 3){
   a=( vp3*p2-vp2*p3)/det3;
   b=(-ve3*p2+ve2*p3)/det3;
 }
 t0=T+(f->V[0]); t1=T+(f->V[1]); t2=T+(f->V[2]);
 t1x=t1->x - t0->x; t2x=t2->x - t0->x;
 t1y=t1->y - t0->y; t2y=t2->y - t0->y;
 if(fabs(t1x) > 4.0)t1x=0.5;
 if(fabs(t2x) > 4.0)t2x=0.5;
 if(fabs(t1y) > 4.0)t1y=0.5;
 if(fabs(t2y) > 4.0)t2y=0.5;
 alpha = t0->x + a*t1x + b*t2x;
 beta  = t0->y + a*t1y + b*t2y;
/* alpha = t0->x + a*(t1->x - t0->x) + b*(t2->x - t0->x); */
/* beta  = t0->y + a*(t1->y - t0->y) + b*(t2->y - t0->y); */
 a=(alpha-floor(alpha));
 b=1.0-(beta-floor(beta));  /* 1-  to have origin at bottom left */
/* b=(beta-floor(beta)); for origin at top left */
 alpha=a*Map->xmaxtran;
 beta=b*Map->ymaxtran;
 ialpha = (long)alpha; alpha -= (double)ialpha;
 ibeta  = (long)beta;  beta  -= (double)ibeta;
 if(ialpha < 0 || ialpha >= Map->xmaxtran)return 0;
 if(ibeta  < 0 || ibeta  >= Map->ymaxtran)return 0;
 k =ibeta*j+ialpha;
 if(ialpha <   1)kl=k+j-1;                        else kl=k-1;
 if(ialpha > j-2)kr=ibeta*j;                      else kr=k+1;
 if(ibeta  <   1)kb=(Map->ymaxtran-1)*j+ialpha; else kb=k-j;
 if(ibeta  > Map->ymaxtran-2)kt=ialpha;         else kt=k+j;
 if(Map->depth == 24){
   pR=Map->p24Rtran; pG=Map->p24Gtran; pB=Map->p24Btran;
 }
 else{
   pP=Map->ptran;  k =pP[k];
   kl=pP[kl]; kr=pP[kr]; kt=pP[kt]; kb=pP[kb];
   pR=Map->palRtran; pG=Map->palGtran; pB=Map->palBtran;
 }
 cc0 = (double)pR[k];  cc1 = (double)pG[k];  cc2 = (double)pB[k];
 cl0 = (double)pR[kl]; cl1 = (double)pG[kl]; cl2 = (double)pB[kl];
 cr0 = (double)pR[kr]; cr1 = (double)pG[kr]; cr2 = (double)pB[kr];
 ct0 = (double)pR[kt]; ct1 = (double)pG[kt]; ct2 = (double)pB[kt];
 cb0 = (double)pR[kb]; cb1 = (double)pG[kb]; cb2 = (double)pB[kb];
 if(alpha > 0.5){
   alpha = alpha-0.5;
   if(beta > 0.5){
     beta = beta-0.5;
     red   = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     green = (cr1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     blue  = (cr2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     red   = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     green = (cr1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     blue  = (cr2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 else{
   alpha = 0.5-alpha;
   if(beta > 0.5){
     beta = beta-0.5;
     red   = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     green = (cl1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     blue  = (cl2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     red   = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     green = (cl1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     blue  = (cl2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 *trans = 1.3071895e-3*(red+green+blue); /* scaled 0-1 black 0% transparent */
 return 1;
}

#ifdef _SUNSTYLE
long MapExternalByCoords(obj,f,p,alpha,beta,gamma) long obj,
face *f, vector p, double *alpha, double *beta, double *gamma; {
#else
long MapExternalByCoords(long obj,face *f, vector p,
     double *alpha, double *beta, double *gamma){
#endif
 /* return 2 D coords on face f of object # obj */
 long k;
 double a,b,
        ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3,
        detmax,dm,det1,det2,det3,
        t1x,t2x,t1y,t2y;
 unsigned char *pP,*pR,*pG,*pB;
 tvertex  *T,*t0,*t1,*t2;
 vertex   *V,*v0,*v1,*v2;
 /* checking for valid map must be done elsewhere */
 if((T=Object[obj].Tbase) == NULL)return 0;
 if((V=Object[obj].Vbase) == NULL)return 0;
 v0=(V+(f->V[0])); t0=(T+(f->V[0]));
 v1=(V+(f->V[1])); t1=(T+(f->V[1]));
 v2=(V+(f->V[2])); t2=(T+(f->V[2]));
 ve1=v1->p[0] - v0->p[0]; ve2=v1->p[1] - v0->p[1]; ve3=v1->p[2] - v0->p[2];
 vp1=v2->p[0] - v0->p[0]; vp2=v2->p[1] - v0->p[1]; vp3=v2->p[2] - v0->p[2];
 p1=p[0]-v0->p[0];
 p2=p[1]-v0->p[1];
 p3=p[2]-v0->p[2];
 det1=ve1*vp2-vp1*ve2;
 det2=ve1*vp3-vp1*ve3;
 det3=ve2*vp3-vp2*ve3;
 k=0; detmax=1.e-3;  /* same as TOL in mirrors */
 if((dm=fabs(det1)) > detmax){k=1; detmax=dm;}
 if((dm=fabs(det2)) > detmax){k=2; detmax=dm;}
 if((dm=fabs(det3)) > detmax){k=3; detmax=dm;}
 if(k == 0)return 0;
 else if(k == 1){
   a=( vp2*p1-vp1*p2)/det1;
   b=(-ve2*p1+ve1*p2)/det1;
 }
 else if(k == 2){
   a=( vp3*p1-vp1*p3)/det2;
   b=(-ve3*p1+ve1*p3)/det2;
 }
 else if(k == 3){
   a=( vp3*p2-vp2*p3)/det3;
   b=(-ve3*p2+ve2*p3)/det3;
 }
 t0=T+(f->V[0]); t1=T+(f->V[1]); t2=T+(f->V[2]);
 t1x=t1->x - t0->x; t2x=t2->x - t0->x;
 t1y=t1->y - t0->y; t2y=t2->y - t0->y;
 if(fabs(t1x) > 4.0)t1x=0.5;
 if(fabs(t2x) > 4.0)t2x=0.5;
 if(fabs(t1y) > 4.0)t1y=0.5;
 if(fabs(t2y) > 4.0)t2y=0.5;
 *alpha = t0->x + a*t1x + b*t2x;
 *beta  = t0->y + a*t1y + b*t2y;
 *gamma = 0.5;
 return 1;
}

long FastAntiAliasShadedScanline(
                      long scanline ,long aaline, double d2, long *Zcsg,
                      UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                      double *Zground, long Wsky, double Rsky,
                      long *Xbuffer, long *Xobject,
                      long *Zglass[], long *ZglassO[], double *Zglassd[],
                      long *Zbuffer, long *Zobject,
                      double *Zdepth, double *Zposn){
 face   *f;             /* super sample if faces are different  */
 eface  *e;             /* take it 1 pixel at a time            */
 pface  *g;
 vertex *v;
 imap   *Map;
 long i,j,k,l,f1,obj,AlreadyCalculated;
 double det,alpha,beta;
 vector n,v0,v1,v2;
 UCHAR  *ZobjChar;
 if(VGA_screen == NULL)ZobjChar=DummyShadow;
 else ZobjChar = &VGA_screen[scanline*XMAX];
 i=0; l=i-1;
 for(k=0;k<XMAX;k+=aaXstep){
   for(j=0;j<aaXstep;j++){
     f1=Zbuffer[i]; obj=Zobject[i];
     if(f1 == Xbuffer[i] && obj == Xobject[i])AlreadyCalculated=1;
     else if(j == 0)AlreadyCalculated=0;
     else if(f1 == Zbuffer[i-1] && obj == Zobject[i-1]){
       AlreadyCalculated=1;
       Red[i] = Red[l]; Green[i]=Green[l]; Blue[i]=Blue[l];
     }
     else AlreadyCalculated=0;
     if(!AlreadyCalculated){
       if(f1 >= 0){
         v=Object[obj].Vbase;
         f=Object[obj].Fbase + f1;
         e=Object[obj].Ebase + f1;
         g=Object[obj].Pbase + f1;
         if((f->fattrib & 0x40) == 0x40){ /* mapped */
           if((long)f->fmap >= Object[obj].NoMaps)Map=NULL;
           else Map=(Object[obj].Mbase+(long)f->fmap);
         }
         else Map=NULL;
         if( f->fattrib >> 7 == 0)VECCOPY(e->n,n)
         else{
           det=(
             (v[f->V[2]].y - v[f->V[0]].y) * (v[f->V[1]].x - v[f->V[0]].x)
              -(v[f->V[2]].x - v[f->V[0]].x) * (v[f->V[1]].y - v[f->V[0]].y)
               );
           if(det == 0){ VECCOPY(e->n,n) }
           else{
             VECCOPY(g->n[0],v0)
             VECCOPY(g->n[1],v1)
             VECCOPY(g->n[2],v2)
             det=1.0/det;
             alpha= (double)(
                     (v[f->V[2]].y - v[f->V[0]].y) *
                         (Zposn[i] - v[f->V[0]].x)
                    -(v[f->V[2]].x - v[f->V[0]].x) *
                 ((double)scanline - v[f->V[0]].y)
                            )*det;
             beta= (double)(
                  ((double)scanline - v[f->V[0]].y) *
                      (v[f->V[1]].x - v[f->V[0]].x)
                         -(Zposn[i] - v[f->V[0]].x) *
                      (v[f->V[1]].y - v[f->V[0]].y)
                           )*det;
             n[0]=v0[0] + alpha*(v1[0] - v0[0]) +  beta*(v2[0] - v0[0]);
             n[1]=v0[1] + alpha*(v1[1] - v0[1]) +  beta*(v2[1] - v0[1]);
             n[2]=v0[2] + alpha*(v1[2] - v0[2]) +  beta*(v2[2] - v0[2]);
             normalize(n);
           }
         }
         GetPixelValue(scanline,i,Map,e,f,obj,n,Red,Green,Blue,
                            Zdepth,ZobjChar);
         if(Zglass[0][i] > -1)AttenuateGlass(scanline,i,Red,Green,Blue,
                                             Zglass,ZglassO,Zglassd,Zposn);
       }
       else{
         PutBackPlanes(i,scanline,d2,Zcsg,Red,Green,Blue,
                       Zground,Zobject,Zdepth,ZobjChar,Wsky,Rsky);
         if(Zglass[0][i] > -1)AttenuateGlass(scanline,i,Red,Green,Blue,
                                             Zglass,ZglassO,Zglassd,Zposn);
       }
     }
     i++; l++;
   }
 }
 return 0;
}

long ShadedScanline(
                 long scanline ,double d2, long *Zcsg,
                 UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                 double *Zground, long Wsky, double Rsky,
                 long *Zglass[], long *ZglassO[], double *Zglassd[],
                 long *Zbuffer, long *Zobject,
                 double *Zdepth, double *Zposn){
 face    *f;
 eface   *e;
 pface   *g;
 vertex  *v;
 imap    *Map;
 long i,j,k,f1,obj;
 double det,alpha,beta;
 vector n,nn,dn,v0,v1,v2;
 UCHAR *ZobjChar;
 if(VGA_screen == NULL)ZobjChar=DummyShadow;
 else ZobjChar = &VGA_screen[scanline*XMAX]; /* shadow for this line */
 if(Ncsg > 0)GetCSGintersections(scanline,d2,Zcsg,Zground,
                                 Zbuffer,Zobject,Zdepth);
 i=XMIN;
LOOP:
 f1=Zbuffer[i];
 if(f1 >= 0){
   obj=Zobject[i];
   v=Object[obj].Vbase;
   f=Object[obj].Fbase + f1;
   e=Object[obj].Ebase + f1;
   g=Object[obj].Pbase + f1;
   if((f->fattrib & 0x40) == 0x40){                 /* mapped get map id */
     if((long)f->fmap >= Object[obj].NoMaps)Map=NULL;
     else Map=(Object[obj].Mbase+(long)f->fmap);
   }
   else Map=NULL;
 /* calculate the  alpha and beta for starting value */
   if( f->fattrib >> 7 == 0){   /* no smoothing required n is face normal */
     VECCOPY(e->n,n)
   }
   else{ /* interpolate normal over the face */
     det=(
         (v[f->V[2]].y - v[f->V[0]].y) * (v[f->V[1]].x - v[f->V[0]].x)
        -(v[f->V[2]].x - v[f->V[0]].x) * (v[f->V[1]].y - v[f->V[0]].y)
          );
     if(det == 0){
       VECCOPY(e->n,n)
     }
     else{
       VECCOPY(g->n[0],v0)
       VECCOPY(g->n[1],v1)
       VECCOPY(g->n[2],v2)
       det=1.0/det;
       alpha= (double)(
        (v[f->V[2]].y - v[f->V[0]].y) * (Zposn[i] - v[f->V[0]].x)
       -(v[f->V[2]].x - v[f->V[0]].x) * ((double)scanline - v[f->V[0]].y)
                      )*det;
       beta= (double)(
        ((double)scanline - v[f->V[0]].y) * (v[f->V[1]].x - v[f->V[0]].x)
       -(Zposn[i] - v[f->V[0]].x) * (v[f->V[1]].y - v[f->V[0]].y)
                    )*det;
       n[0]=v0[0] + alpha*(v1[0] - v0[0])
                  +  beta*(v2[0] - v0[0]);
       n[1]=v0[1] + alpha*(v1[1] - v0[1])
                  +  beta*(v2[1] - v0[1]);
       n[2]=v0[2] + alpha*(v1[2] - v0[2])
                  +  beta*(v2[2] - v0[2]);
       normalize(n);
     }
   }
   GetPixelValue(scanline,i,Map,e,f,obj,n,Red,Green,Blue,
                      Zdepth,ZobjChar); /* pixel 1*/
   if(Zglass[0][i] > -1)AttenuateGlass(scanline,i,Red,Green,Blue,
                                       Zglass,ZglassO,Zglassd,Zposn);
   j=i+1;  /* interpolate normal over pixels showing same face */
   if(j < XMAX){
    while((Zbuffer[j] == f1) && (Zobject[j] == Zobject[i]))
                                {j++; if(j == XMAX)break; }   /* how many */
    if(j > i+1){ /* more than one pixel */
      if(j > i+2){ /* more than 2 pixels */
        if((det==0) || (f->fattrib >> 7 == 0)){
          dn[0]=0; dn[1]=0; dn[2]=0;
        }
        else{
          dn[0]=( (v[f->V[1]].y - v[f->V[2]].y) * v0[0]
                + (v[f->V[2]].y - v[f->V[0]].y) * v1[0]
                + (v[f->V[0]].y - v[f->V[1]].y) * v2[0]
                )*det;
          dn[1]=( (v[f->V[1]].y - v[f->V[2]].y) * v0[1]
                + (v[f->V[2]].y - v[f->V[0]].y) * v1[1]
                + (v[f->V[0]].y - v[f->V[1]].y) * v2[1]
                )*det;
          dn[2]=( (v[f->V[1]].y - v[f->V[2]].y) * v0[2]
                + (v[f->V[2]].y - v[f->V[0]].y) * v1[2]
                + (v[f->V[0]].y - v[f->V[1]].y) * v2[2]
                )*det;
        }
        for(k=i+1; k<j-1; k++){
          VECSUM(n,dn,n)
          VECCOPY(n,nn)
          normalize(nn);
          GetPixelValue(scanline,k,Map,e,f,obj,nn,Red,Green,Blue,
                             Zdepth,ZobjChar);
          if(Zglass[0][k] > -1)AttenuateGlass(scanline,k,Red,Green,Blue,
                                              Zglass,ZglassO,Zglassd,Zposn);
        }
      }
      /* do last pixel this must be separate - may not be on integer boundary */
      if((det==0) || (f->fattrib >> 7 == 0))VECCOPY(e->n,n)
      else{
        alpha = (double)(
         (v[f->V[2]].y - v[f->V[0]].y) * (Zposn[j-1] - v[f->V[0]].x)
        -(v[f->V[2]].x - v[f->V[0]].x) * ((double)scanline - v[f->V[0]].y)
                        )*det;
        beta = (double)(
         ((double)scanline - v[f->V[0]].y) * (v[f->V[1]].x - v[f->V[0]].x)
        -(Zposn[j-1] - v[f->V[0]].x) * (v[f->V[1]].y - v[f->V[0]].y)
                       )*det;
        n[0]=v0[0] + alpha*(v1[0] - v0[0])
                   +  beta*(v2[0] - v0[0]);
        n[1]=v0[1] + alpha*(v1[1] - v0[1])
                   +  beta*(v2[1] - v0[1]);
        n[2]=v0[2] + alpha*(v1[2] - v0[2])
                   +  beta*(v2[2] - v0[2]);
        normalize(n);
      }
      GetPixelValue(scanline,j-1,Map,e,f,obj,n,Red,Green,Blue,
                         Zdepth,ZobjChar);
      if(Zglass[0][j-1] > -1)AttenuateGlass(scanline,j-1,Red,Green,Blue,
                                            Zglass,ZglassO,Zglassd,Zposn);
    }
   }
   i=j;
 }
 else{  /* if no object get back plane values */
  PutBackPlanes(i,scanline,d2,Zcsg,Red,Green,Blue,
                Zground,Zobject,Zdepth,
                ZobjChar,Wsky,Rsky);
  if(Zglass[0][i] > -1)AttenuateGlass(scanline,i,Red,Green,Blue,
                                      Zglass,ZglassO,Zglassd,Zposn);
  i++;
 }
 if(i < XMAX)goto LOOP;
 return 0;
}

static BOOL UpdateWireframeScanBuffer(long ixl,
                       long id, long O,
                       face *f, eface *e, double d2,
                       long *Zbuffer, long *Zobject,
                       double *Zdepth){
 BOOL infront;
 long i;
 double xl;
 double depth,bL,dL;
 infront=FALSE;
 xl = (double)ixl;
 dL = xl - (double)Xcentre; dL/=scalex;
 bL = e->n[0]*dL + e->n[2]*d2 + e->n[1];
 if(fabs(bL) < 1.e-30)depth = FARAWAY;
 else depth = (e->c)/bL;
 if(depth > 1.000 && depth <= Zdepth[ixl]){
// if(bHideWire && depth > 1.000 && depth <= Zdepth[ixl]+1000.0){
   Zbuffer[ixl]=id; Zobject[ixl]=O;
   infront=TRUE;
 }
 return infront;
}

static long IntersectWire(long scanline, vertex *v1, vertex *v2,
                long *xi, long *xj){
 double ddx,ddy,doff,ddxy;
 double dxi,dxj;
 if((v1->y < (double)(scanline))  &&
    (v2->y < (double)(scanline)))return 0;
 if((v1->y > (double)(scanline)+1.0)  &&
    (v2->y > (double)(scanline)+1.0))return 0;
 ddy = (v2->y) - (v1->y);
 if(fabs(ddy) < 2.0){
   *xi=(long)(v1->x);
   *xj=(long)(v2->x);
   return 1;
 }
 ddx = (v2->x) - (v1->x);
 doff=(double)(scanline)-0.5 - v1->y;
 ddxy=ddx*doff/ddy;
 dxi=((v1->x)+ddxy);
 doff=(double)(scanline)+1.5 - v1->y;
 ddxy=ddx*doff/ddy;
 dxj=((v1->x)+ddxy);
 if(dxi < v1->x && dxi < v2->x)dxi=min(v1->x,v2->x);
 if(dxi > v1->x && dxi > v2->x)dxi=max(v1->x,v2->x);
 *xi=(long)dxi;
 if(dxj < v1->x && dxj < v2->x)dxj=min(v1->x,v2->x);
 if(dxj > v1->x && dxj > v2->x)dxj=max(v1->x,v2->x);
 *xj=(long)dxj;
 return 1;
}

static long ActiveWire(long scanline, long ige, vertex *v, face *f,
                      long *leftedge, long *rightedge){
 long v1,v2,v3;
 long i;
 long x1,x2,x3;
 if(f->top > scanline+1)       return 0;
 if(f->bottom < scanline-1)    return 0;
 v1=f->V[0];
 v2=f->V[1];
 v3=f->V[2];
 if(ige == 0)i=IntersectWire(scanline,(v+v1),(v+v2),&x1,&x2);
 if(ige == 1)i=IntersectWire(scanline,(v+v2),(v+v3),&x1,&x2);
 if(ige == 2)i=IntersectWire(scanline,(v+v3),(v+v1),&x1,&x2);
 if(i == 0)return 0;
 if(x2 < x1){x3=x2; x2=x1; x1=x3;}
 x1=max(0,x1); x2=max(0,x2);
 *leftedge=min(XMAX-1,x1);
 *rightedge=min(XMAX-1,x2);
 return 1;
}

void RenderWireframeObjects(long k,
                 UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                 long *Zglass[], long *ZglassO[], double *Zglassd[],
                 long *Zbuffer, long *Zobject, double *Zdepth, double *Zposn){
 face    *f,*fj;
 eface   *e;
 vertex  *v;
 long kk,i,j,l,m,ixr,ixl;
 double d2,xr,xl;
 for(i=0;i<ObjectCount;i++)if((k > Object[i].ymin) &&
 (k < Object[i].ymax) && Object[i].wireframe &&(Object[i].visibility == YES)){
   v=Object[i].Vbase;
   f=Object[i].Fbase;
   e=Object[i].Ebase;
   for(kk=max(0,k+1-wireframe_thickness);kk<=k;kk++){
     d2=(Ycentre-kk)/scaley;
#if 0
     if(bHideWire && kk == max(0,k+1-wireframe_thickness){
       for(j=0;j<Object[i].NoFaces;j++){
         fj=(f+j);
         if( Active(kk,v,fj,&xl,&xr) ){
           UpdateScanBuffers(xl,xr,j,i,fj,e+j,d2,
                             Zglass,ZglassO,Zglassd,
                             Zbuffer,Zobject,
                             Zdepth,Zposn);
         }
       }
     }
#endif
     for(j=0;j<Object[i].NoFaces;j++){
       fj=(f+j);
       for(m=0;m<3;m++){
         if(ActiveWire(kk,m,v,fj,&ixl,&ixr) ){
           for(l=max(0,ixl-wireframe_thickness);
               l<=min(XMAX-1,ixr+wireframe_thickness);l++){
             if(UpdateWireframeScanBuffer(l,j,i,f+j,e+j,d2,
                                      Zbuffer,Zobject,Zdepth)){
               Red[l]=fj->color[0];
               Green[l]=fj->color[1];
               Blue[l]=fj->color[2];
             }
           }
         }
       }
#if 0
     if( Active(kk,v,fj,&xl,&xr) ){
       UpdateScanBuffers(xl,xr,j,i,fj,e+j,d2,
                         Zglass,ZglassO,Zglassd,
                         Zbuffer,Zobject,
                         Zdepth,Zposn);
     }
#endif
     }
   }
 }
 return;
}

/* end of Image mapping and fast AA line texturing     */

/* code for rendering specific elements such as ground */

#ifdef _SUNSTYLE
void GetSkyLocation(){
#else
void GetSkyLocation(void){
#endif
 double b;
 long i;
 for(i=XMIN;i<XMAX;i++){
   if(Sky.type == SKYMAPPED){
     b= (CamPhi+atan2(D1[i],1.000))/PI*180.0 * 4.0 ; /* 4 maps full rotation */
     while(b <  -180.0)b = b + 360.0;
     while(b >=  180.0)b = b - 360.0;
     ZskyBuffer[XMAX-1-i]=(short)((b+180.0)*skyXscale);
   }
   else{
     ZskyBuffer[XMAX-1-i]=0;
   }
 }
}

long PutBackIntersection(long scanline, double d2,
                         UCHAR *Alfa, double *Zground,
                         long *Zglass[], long *ZglassO[], double *Zglassd[],
                         long *Zbuffer, long *Zobject,
                         double *Zdepth, double *Zposn, double *Rsky){
 long i,j,Wsky;  /* background  setup */
 double b,c,mu;
 if(Sky.type == SKYGRADED){
   *Rsky=fabs((CamAlpha+atan2(d2,1.000))/PI*2.0);  /* 0 Hor  1 +/- Zenith */
 }
 else if(Sky.type == SKYMAPPED){ /* CamAlpha is 0 on horizontal +90 look up*/
  Wsky=0; /* Wsky scanline V pos. in env.bru 0 is brush top*/
  b=90.0 - (CamAlpha+atan2(d2,1.000))/PI*180.0 * 2;
  while(b <   0.0)b += 90.0;
  while(b > 180.0)b -= 90.0;
  if(b > 0){          /* b is angle in range 0-180 for view to scanline */
    if(b > 90)Wsky=(short)((180-b)*skyYscale);      /* 0 is up 180 down */
    else      Wsky=(short)(b*skyYscale);            /* b > 90 picture is */
    Wsky = min(Wsky,skyBmax);                       /* inverted */
  }
 }
 for(i=XMIN;i<XMAX;i++){
   Zbuffer[i] = -1;
   Zdepth [i] = FARAWAY;
   Zposn  [i] = (double)i+0.5;
   for(j=0;j<NGLASS;j++){
     Zglassd[j][i] = FARAWAY;
     ZglassO[j][i] = -1;
     Zglass [j][i] = -1;  /* this will be non zero if a glass face is on top */
   }
   Zobject[i] = -2;
   Alfa   [i] =  0;
   if(R_Nground > 0){
     b=Ground.n[0]*D1[i]+Ground.n[1]+Ground.n[2]*d2; // Z buffer
//{
// vector dr;     //  if Z buffer is to contain actual depths
// dr[0]=D1[i]; dr[1]=1.0; dr[2]=d2;
// b=DOT(Ground.n,dr);
//}
     if(fabs(b) > 1.e-4){
       c=dot(Ground.n,Ground.p);
       mu=c/b;
       if(mu > 0.0){
         Zdepth[i] = mu;
         Zground[i]= mu;
         Zobject[i]= -1;
       }
     }
   }
 }
 return Wsky;
}

#ifdef _SUNSTYLE
void GetGroundLight(p0,R,G,B) vector p0; double *R; double *G; double *B ; {
#else
void GetGroundLight(vector p0, double *R, double *G ,double *B){
#endif
/* the R G B values may exceed 255.0 they must be clipped elsewhere */
 long j,brkindx;
 double c,cc,alpha,beta,Iab,IsR,IsG,IsB,Bump_Scale,IdR,IdG,IdB,xRGB[3];
 vector p1,p2,ripple,gn,bv;
 double red,green,blue,map;
 IdR=IdG=IdB=0.0;
    if(Ground.texture > 0 && (Ground.texture < 4 || Ground.texture > 9)){
      switch(Ground.texture){
      case 1:  /* single wave  */
         calcripple(p0,Ground.p,wave_phase,Ground.lengthy,1.0,
         Ground.height,ripple); /* typical height 2.0 */
         Bump_Scale = - dot(p0,Ground.n)/R_length(p0);
         vecscale(Bump_Scale,ripple,ripple);
         vecsum(ripple,Ground.n,gn);
         normalize(gn);
         IsR=IsG=IsB=0.0;
         break;
      case 2: /* ripples using wave 2 */
        waves2(p0,Ground.n,Ground.dx,Ground.dyy,Ground.p,gn,
        Rip_Freq,wave_phase,Ground.height);  /* 0.2 is good for height */
        IsR=IsG=IsB=0.0;
        break;
      case 3: /* ocean multiple waves texture */
        Ocean(p0,Ground.n,gn);
        IsR=IsG=IsB=0.0;
        break;
      case 10: /* rough surface  ie bumpy */
        vecscale(Ground.bump,p0,p1);
        Dnoise2(gn,p1[0],p1[1],p1[2]);                  /* scale bumps into */
        Bump_Scale = dot(p0,Ground.n)/R_length(p0);     /* distance*/
        Bump_Scale = min(1.0, Bump_Scale*5)*Ground.height;
        vecscale(Bump_Scale,gn,gn);
        vecsum(Ground.n,gn,gn);
        normalize(gn);
        break;
     case 16: /* external textures */ {
// this is not used until -1's are fixed up
//          double alpha_channel=0.0,trans=0.0;
//          Bump_Scale = Ground.height;
//          VECCOPY(Ground.n,gn)
//          VECCOPY((double)Ground.color,xRGB)
//          hitpoint(Ground.n,Ground.dx,Ground.dy,Ground.p,p0,&alpha,&beta);
//          if(Ground.shader >= 0)GetExternalTexture(
//                             -1;
//                             (long)Ground.shader,
//                             0,
//                             p0,gn,
//                             alpha,beta,0.0,
//                             Bump_Scale,
//                             &alpha_channel,
//                             &trans,
//                             Ground.acolor,
//                             -1,NULL,
//                             xRGB);
        }
        break;
      default:
        veccopy(Ground.n,gn);
        break;
      }
    }
    else veccopy(Ground.n,gn);
    for(j=0;j<Nlights;j++)if(Lights[j].type != SHADOWS &&
                             Lights[j].type != DUMMYL){
      vecsub(Lights[j].p,p0,p1);
      cc=R_length(p1);
      vecscale(1.0/cc,p1,p2);
      Iab = Ground.ambient_light+(1.0-Ground.ambient_light)*fabs(DOT(p2,gn));
      if(Lights[j].dc){
        cc = DepthCueLight(j,cc);
        Iab *= cc;
      }
      if(Lights[j].type == SPOTLIGHT){
        if((alpha=DOT(p2,Lights[j].d)) < Lights[j].dot2)
          Iab=Ground.ambient_light;
        else if(alpha < Lights[j].dot1){
          Iab=Ground.ambient_light+(Iab-Ground.ambient_light)*
               (alpha-Lights[j].dot2)*Lights[j].falloff;
        }
        if(Iab > Ground.ambient_light && Lights[j].shadow_buffer != NULL)
            PointInShadow(p0,j,&Iab); /* for spotlights */
      }
      IdR += Iab*Lights[j].Ire;
      IdG += Iab*Lights[j].Igr;
      IdB += Iab*Lights[j].Ibl;
      if(Ground.shiny){
        vecsub(p1,p0,bv);
        normalize(bv);
        c=fabs(dot(bv,gn));
        if(c > 0.0){
          Iab = spec_gro[min(max_spec,(long)(c*max_spec))];
          if(Lights[j].dc)Iab *= cc;
          IsR += Iab*Lights[j].Ire;
          IsG += Iab*Lights[j].Igr;
          IsB += Iab*Lights[j].Ibl;
        }
      }
    }
    switch(Ground.texture){
    case 1:  /* waves have specular highlights */
    case 2:
    case 3:
      *R = min( ((double)Ground.color[0]*IdR + IsR*255), 255);
      *G = min( ((double)Ground.color[1]*IdG + IsG*255), 255);
      *B = min( ((double)Ground.color[2]*IdB + IsB*255), 255);
      break;
    case 4:  /* checkered ground */
      hitpoint(Ground.n,Ground.dx,Ground.dy,Ground.p,p0,&alpha,&beta);
      brkindx = (long)FLOOR(alpha) + (long)FLOOR(beta);
      if(brkindx & 1){
        *R = (double)Ground.acolor[0]*IdR;
        *G = (double)Ground.acolor[1]*IdG;
        *B = (double)Ground.acolor[2]*IdB;
      }
      else{
        *R = (double)Ground.color[0]*IdR;
        *G = (double)Ground.color[1]*IdG;
        *B = (double)Ground.color[2]*IdB;
      }
      break;
    case 5: /* grid texture */
      hitpoint(Ground.n,Ground.dx,Ground.dy,Ground.p,p0,&alpha,&beta);
      alpha=(alpha-floor(alpha));
      beta=(beta-floor(beta));
      if(beta < 0.1 || alpha < 0.1){
        *R = (double)Ground.acolor[0]*IdR;
        *G = (double)Ground.acolor[1]*IdG;
        *B = (double)Ground.acolor[2]*IdB;
      }
      else{
        *R = (double)Ground.color[0]*IdR;
        *G = (double)Ground.color[1]*IdG;
        *B = (double)Ground.color[2]*IdB;
      }
      break;
    case 6: /* stripes texure */
      hitpoint(Ground.n,Ground.dx,Ground.dy,Ground.p,p0,&alpha,&beta);
      alpha=(alpha-floor(alpha));
      beta=(beta-floor(beta));
      if(alpha < 0.1){
        *R = (double)Ground.acolor[0]*IdR;
        *G = (double)Ground.acolor[1]*IdG;
        *B = (double)Ground.acolor[2]*IdB;
      }
      else{
        *R = (double)Ground.color[0]*IdR;
        *G = (double)Ground.color[1]*IdG;
        *B = (double)Ground.color[2]*IdB;
      }
      break;
    case 7: /* dot texture */
      hitpoint(Ground.n,Ground.dx,Ground.dyy,Ground.p,p0,&alpha,&beta);
      alpha=(alpha-floor(alpha))-0.5;
      beta=(beta-floor(beta))-0.5;
      if(beta*beta+alpha*alpha < Ground.range){
        *R = (double)Ground.acolor[0]*IdR;
        *G = (double)Ground.acolor[1]*IdG;
        *B = (double)Ground.acolor[2]*IdB;
      }
      else{
        *R = (double)Ground.color[0]*IdR;
        *G = (double)Ground.color[1]*IdG;
        *B = (double)Ground.color[2]*IdB;
      }
      break;
    case 8: /* cross texture */
      hitpoint(Ground.n,Ground.dx,Ground.dyy,Ground.p,p0,&alpha,&beta);
      alpha=fabs((alpha-floor(alpha))-0.5);
      beta=fabs((beta-floor(beta))-0.5);
      if((beta < 0.3 && alpha < 0.1) || (alpha < 0.3 && beta < 0.1)){
        *R = (double)Ground.acolor[0]*IdR;
        *G = (double)Ground.acolor[1]*IdG;
        *B = (double)Ground.acolor[2]*IdB;
      }
      else{
        *R = (double)Ground.color[0]*IdR;
        *G = (double)Ground.color[1]*IdG;
        *B = (double)Ground.color[2]*IdB;
      }
      break;
    case 9: /* brick texture */
      hitpoint(Ground.n,Ground.dx,Ground.dy,Ground.p,p0,&alpha,&beta);
      alpha=(alpha-floor(alpha));
      beta=(beta-floor(beta));
      if((beta > 0.95 || (beta > 0.45 && beta < 0.5)) ||
         (beta < 0.5 && ( (alpha > 0.45 && alpha < 0.5) || alpha > 0.95) ) ||
         (beta > 0.5 && ((alpha > 0.25 && alpha < 0.3) ||
         (alpha > 0.7 && alpha < 0.75)))
         ){
        *R = (double)Ground.acolor[0]*IdR;
        *G = (double)Ground.acolor[1]*IdG;
        *B = (double)Ground.acolor[2]*IdB;
      }
      else{
        *R = (double)Ground.color[0]*IdR;
        *G = (double)Ground.color[1]*IdG;
        *B = (double)Ground.color[2]*IdB;
      }
      break;
    case 10:
      *R = (double)Ground.color[0]*IdR;
      *G = (double)Ground.color[1]*IdG;
      *B = (double)Ground.color[2]*IdB;
      break;  /* bumpy */
    case 16:  /* external texture */
      *R = xRGB[0]*IdR; *G = xRGB[1]*IdG; *B = xRGB[2]*IdB;
      break;
    default:  /* ground is mapped by a brush               */
      map=0;  /* ground brush has been successfully loaded */

      if(Ground.mapped){
         map=MapBrush(&Ground.map,Ground.maptype,0,
                      p0,&red,&green,&blue);
      }
      if(map == 1){
        *R = (red*IdR);
        *G = (green*IdG);
        *B = (blue*IdB);
      }
      else{
        *R = (double)Ground.color[0]*IdR;
        *G = (double)Ground.color[1]*IdG;
        *B = (double)Ground.color[2]*IdB;
      }
    break;
    }
}

double DepthCueLight(long j, double cc){
 double d;
 return 1.0/(Lights[j].dc_c + cc*(Lights[j].dc_l + cc*Lights[j].dc_q));
}

/* end of code for ground and other elements           */

static void PutBackPlanes(long i, long scanline, double d2, long *Zcsg,
                   UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                   double *Zground, long *Zobject, double *Zdepth,
                   UCHAR *ZobjChar, long Wsky, double Rsky){
 vector p0;
 double red,green,blue;
 long ix;
   if(Zobject[i] == -1){   /* ground at this pixel */
    p0[0]=(((double)i-(double)Xcentre)/scalex*Zground[i]);
    p0[1]=min(FARAWAY,Zdepth[i]);
    p0[2]=(((double)Ycentre-(double)scanline)/scaley*Zground[i]);
    GetGroundLight(p0,&red,&green,&blue);
    Red  [i] = (unsigned char)min(255.0,red);
    Green[i] = (unsigned char)min(255.0,green);
    Blue [i] = (unsigned char)min(255.0,blue);
    if((ZobjChar[i] & 1) == 1){
      Red[i]   *= shadow_density;     /* do the shadows before reflections*/
      Green[i] *= shadow_density;
      Blue[i]  *= shadow_density;
    }
#ifdef DOS386
    if(mirror_ground == 1){
      Add_Ground_Mirror(scanline,i,p0,&Red[i],&Green[i],&Blue[i]);
    }
#endif
   }
   else if(Zobject[i] == -3){
     GetCSGvalue(i,scanline,d2,Zcsg,Red,Green,Blue,Zground,Zdepth);
     return;  /* csg */
   }
   else if(Nsky > 0){
//RSF
//Red  [i]=Sky.Zenith[0];
//Green[i]=Sky.Zenith[1];
//Blue [i]=Sky.Zenith[2];
//return;
     /* if the map could not be loaded use this as well */
     if(Sky.type == SKYGRADED){
       Red  [i]=(unsigned char)(
                   ((double)Sky.Zenith[0]-(double)Sky.Horizon[0])*Rsky
                  + (double)Sky.Horizon[0]);
       Green[i]=(unsigned char)(
                   ((double)Sky.Zenith[1]-(double)Sky.Horizon[1])*Rsky
                  + (double)Sky.Horizon[1]);
       Blue [i]=(unsigned char)(
                   ((double)Sky.Zenith[2]-(double)Sky.Horizon[2])*Rsky
                  + (double)Sky.Horizon[2]);
     }
     else if(Sky.type <= SKYCOLOUR || !Sky.mapped){
       Red  [i]=Sky.Zenith[0];
       Green[i]=Sky.Zenith[1];
       Blue [i]=Sky.Zenith[2];
     }
     else if(Sky.type == SKYMAPPED){  /* Map the sky */
       ix=Wsky*Sky.map.xmax+ZskyBuffer[i];
       if(Sky.map.depth == 24){
         Red[i]   = Sky.map.p24R[ix];
         Green[i] = Sky.map.p24G[ix];
         Blue[i]  = Sky.map.p24B[ix];
       }
       else{
         Red[i]   = Sky.map.palR[ Sky.map.picture[ix] ];
         Green[i] = Sky.map.palG[ Sky.map.picture[ix] ];
         Blue[i]  = Sky.map.palB[ Sky.map.picture[ix] ];
       }
     }
     else if(Sky.type == BACKDROP){  /* Put in back drop */
       ix=((scanline*Sky.map.ymax)/YMAX)*Sky.map.xmax
         +(i*Sky.map.xmax)/XMAX;
       if(Sky.map.depth == 24){
         Red[i]   = Sky.map.p24R[ix];
         Green[i] = Sky.map.p24G[ix];
         Blue[i]  = Sky.map.p24B[ix];
       }
       else{
         Red[i]   = Sky.map.palR[ Sky.map.picture[ix] ];
         Green[i] = Sky.map.palG[ Sky.map.picture[ix] ];
         Blue[i]  = Sky.map.palB[ Sky.map.picture[ix] ];
       }
     }
   }
   else{ Red[i]=0;   Green[i]=0;   Blue[i]=0; }
}

static void GetBackdropValue(long row, long col, vector p, vector d,
                             double *colour){
 /* this is used by the Tracer to get the background */
 vector dp,gp;
 double pdotn,ddotn,mu,d2,Rsky;
 long ix;
 if(R_Nground > 0 && fabs(ddotn=dot(Ground.n,d)) > 1.e-3){
   VECSUB(Ground.p,p,dp)
   pdotn=DOT(Ground.n,dp);
   mu=pdotn/ddotn;
   if(mu > 0.0){
     vecscale(mu,d,gp);
     vecsum(p,gp,gp);
     GetGroundLight(gp,&colour[0],&colour[1],&colour[2]);
   }
   else goto NEXTBLOCK;
 }
 else{
   NEXTBLOCK:
   if(Sky.type == SKYGRADED){
     d2=(double)(Ycentre-row);
     d2/=scaley;
     Rsky=fabs((CamAlpha+atan2(d2,1.000))/PI*2.0);
     colour[0]=((double)Sky.Zenith[0]-(double)Sky.Horizon[0])*Rsky
               +(double)Sky.Horizon[0];
     colour[1]=((double)Sky.Zenith[1]-(double)Sky.Horizon[1])*Rsky
               +(double)Sky.Horizon[1];
     colour[2]=((double)Sky.Zenith[2]-(double)Sky.Horizon[2])*Rsky
               +(double)Sky.Horizon[2];
   }
   else if(Sky.type == SKYMAPPED && Sky.mapped){
     ddotn=d[2];
     if(fabs(ddotn) > 1.e-2){
       vecsub(Sky.map.p,p,dp);
       pdotn=dp[2];
       mu=pdotn/ddotn;
       vecscale(fabs(mu),d,gp);
       vecsum(p,gp,gp);
       MapBrush(&Sky.map,TILE,EOPAQUE,gp,&colour[0],&colour[1],&colour[2]);
     }
   }
   else if(Sky.type == BACKDROP && Sky.mapped){
     ix=((row*Sky.map.ymax)/YMAX)*Sky.map.xmax
       +(col*Sky.map.xmax)/XMAX;
     if(Sky.map.depth == 24){
       colour[0] = Sky.map.p24R[ix];
       colour[1] = Sky.map.p24G[ix];
       colour[2] = Sky.map.p24B[ix];
     }
     else{
       colour[0] = Sky.map.palR[ Sky.map.picture[ix] ];
       colour[1] = Sky.map.palG[ Sky.map.picture[ix] ];
       colour[2] = Sky.map.palB[ Sky.map.picture[ix] ];
     }
   }
   else{
     colour[0] = (double)Sky.Zenith[0];
     colour[1] = (double)Sky.Zenith[1];
     colour[2] = (double)Sky.Zenith[2];
   }
 }
 return;
}

#ifdef _SUNSTYLE
static void GetPseudoMirror(obj,row,col,color) long obj;
long row; long col; double *color; {
#else
static void GetPseudoMirror(long obj, long row, long col, double *color){
#endif
 int i,j;
 unsigned char *p;
 static xlist[]={1,2,2,3,3},ylist[]={1,1,2,2,3};
 if(in_mirror_phase || (p=Object[obj].mirror_buffer) == NULL){
   color[0]=color[1]=color[2]=0;
   return;
 }                     /* don't mirror the mirror */
 i=row/ylist[anti_alias];
 j=col/xlist[anti_alias];
 p+=((i*ResolutionX+j)*3);
 color[0]=(double)*p++;
 color[1]=(double)*p++;
 color[2]=(double)*p++;
 return;
}

void GetMirrorValues(double *color, vector p, vector din, vector n,
                     long trace_reflection_depth,
                     long trace_refraction_depth,
                     long trace_refractive_flag){
  /* p    is point on mirror surface where ray hits   */
  /* din  is incident direction                       */
  vector d,dp,gp,di;
  double pdotn,ddotn,mu,ld;
  ddotn = dot(din,n);
  vecscale(2*ddotn,n,d);  /* d is reflected direction (same length as din) */
  vecsub(din,d,d);        /* d=2(din.n)n-din                               */
  if(R_Nground > 0){
    ddotn=dot(Ground.n,d);
    if(fabs(ddotn) > 1.e-2){
      vecsub(Ground.p,p,dp);
      pdotn=dot(Ground.n,dp);
      mu=pdotn/ddotn;
      if(mu > 0.0){
        vecscale(mu,d,di);
        vecsum(p,di,gp);
        GetGroundLight(gp,&color[0],&color[1],&color[2]);
#ifdef TRACER
        if(trace_reflections == 1){
          ld = (double)((di[0]*di[0]) + (di[1]*di[1]) + (di[2]*di[2]));
          if(fabs(ld) > 1.e-10){
            ld=sqrt(ld);
            VECSCALE((1.0/ld),di,di)
            trace_reflection_ray(p,di,ld,color,
                                 trace_reflection_depth,
                                 trace_refraction_depth,
                                 trace_refractive_flag);
          }
        }
#endif
        return;
      }
    }
  }
  color[0] = (double)Sky.Zenith[0];
  color[1] = (double)Sky.Zenith[1];
  color[2] = (double)Sky.Zenith[2];
  if(Sky.type == SKYMAPPED){
    ddotn=d[2];
    if(fabs(ddotn) > 1.e-2){
      vecsub(Sky.map.p,p,dp);
      pdotn=dp[2];
      mu=pdotn/ddotn;
      vecscale(fabs(mu),d,gp);
      vecsum(p,gp,gp);
      /* check for existance of sky map elsewhere */
      MapBrush(&Sky.map,TILE,EOPAQUE,gp,&color[0],&color[1],&color[2]);
    }
  }
#ifdef TRACER
  if(trace_reflections == 1){
    ld= (double)((d[0]*d[0]) + (d[1]*d[1]) + (d[2]*d[2]));
    if(fabs(ld) > 1.e-10){
      ld=sqrt(ld);
      VECSCALE((1.0/ld),d,di)
      trace_reflection_ray(p,di,(double)BIGP,color,
                           trace_reflection_depth,
                           trace_refraction_depth,
                           trace_refractive_flag);
    }
  }
#endif
  return;
}

void GetPhongPixelValue(long row, long col, imap *Map, eface *e,
                        face *f, long obj, vector nn,  /* obj is object id */
                        UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                        double *Zdepth,UCHAR *ZobjChar){
 BOOL     xtext;
 vector   p,pr,bv,pl,n,d;
 long     j,flag_mirror=0,t;
 short    type,textID,clear;
 double   colors[3],colorr[3],
          colorb0,colorb1,colorb2;
 double   Iab,IdR,IdG,IdB,IsR,IsG,IsB,c,cc,
          pdotn,ldotn,amblight,falloff,alpha;
 double   bf=0.0,sf=1.0,srf=1.0;

//   VECCOPY((double)f->color,colors)   // for plain finish
//   Red[col]  = (unsigned char)min(255,colors[0]);
//   Green[col]= (unsigned char)min(255,colors[1]);
//   Blue[col] = (unsigned char)min(255,colors[2]);
//   return;

 /* p is the point on the face that is seen at this pixel */
 p[0]=(((double)col-(double)Xcentre)/scalex*Zdepth[col]);
 p[1]=Zdepth[col];
 p[2]=(((double)Ycentre-(double)row)/scaley*Zdepth[col]);
 IdR=IdG=IdB=0.0; IsR=IsG=IsB=0.0;
 textID = f->fshader; xtext=(BOOL)(f->fattrib & 0x10);
 /* n is to be normal to surface at that point p */
 /* if((pdotn=DOT(p,nn)) < 0)VECCOPY(nn,n) don't do this - bad pixels */
 if((pdotn=DOT(p,e->n)) < 0)VECCOPY(nn,n)
 else     {n[0] = -nn[0]; n[1] = -nn[1]; n[2] = -nn[2];}
 if(Map != NULL && Map->bp >= 0){
   if(Map->map == MAP_BY_VERTEX)
     BumpByCoordinates(Map,obj,f,p,n);
   else{
     if((f->fattrib & 0x08) == 0x08)map_bumpy(TILE,obj,Map,p,n);
     else                           map_bumpy(SINGLE,obj,Map,p,n);
   }
 }
 if(textID > 6 || xtext){  // else if  removed to allow for combo bump map
   VECCOPY((double)f->color,colors)
   if(xtext){ /* external shader */
     double alpha_channel=0.0,trans=0.0;
     int s_c,s_cid,t,ctype;
     double beta,gamma;
     t=f->faxis;
     if(t == 0 && Object[obj].Tbase != NULL){ /* if shader axis = default */
       ctype=1;                               /* + v's have map co-ords   */
       MapExternalByCoords(obj,f,p,&alpha,&beta,&gamma);
     }
     else{   /* use axes */
       ctype=0;
       gamma=hitpoint(Object[obj].Abase[t].bn,Object[obj].Abase[t].bu,
                 Object[obj].Abase[t].bv,Object[obj].Abase[t].bo,p,&alpha,&beta);
       gamma /= max(1.0,Object[obj].Abase[t].tl);
     }
     for(s_c=0; s_c < 8; s_c++){
       if((s_cid=Object[obj].shader[s_c]) >= 0 && (textID & (1 << s_c)))
       GetExternalTexture(
        t,
        s_cid,
        ctype,
        p,n,
        alpha,beta,gamma,
        (double)f->fparam*SHSCALE,  /* bumpyess */
        &alpha_channel,
        &trans,
        f->texco,
        obj,f,
        colors);
     }
   }
   else if(textID ==  7)new_bumpy  (obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
   else if(textID ==  8)aztec_bumpy(obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
   else if(textID == 13)dents      (obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
   else if(textID ==  9)wrinkles   (obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
 }
 for(j=0;j<Nlights;j++)if(Lights[j].type != SHADOWS &&
                          Lights[j].type != DUMMYL){
   VECSUB(Lights[j].p,p,pl)
   cc=R_length(pl);
   VECSCALE(1.0/cc,pl,pr)
/*   ldotn=DOT(pr,e->n); this was used to try to eliminate fringe pixels
     but it appears to cause some faces to de-phong incorrectly */
   ldotn=DOT(pr,n);
   if((pdotn > 0.0 && ldotn < 0.0)  || (pdotn <= 0.0 && ldotn >= 0.0)){
     amblight=Object[obj].ambient_light;
     c = max(0.0,DOT(pr,n));
     if(Lights[j].type == SPOTLIGHT){
/* note the < is because DOT is max when pr and n are aligned */
/* dot2 is the outer cone, dot1 is the inner cone             */
/* d of light is in OPPOSITE dirn. to the dirn of the light   */
/* pr is vector from point to light thus if pr.n = 1 light is */
/* pointing straight at the point                             */
       if((alpha=DOT(pr,Lights[j].d)) < Lights[j].dot2)falloff=0;
       else if(alpha < Lights[j].dot1){ /* in between cones */
         falloff=(alpha-Lights[j].dot2)*Lights[j].falloff;
       }
       else falloff=1.0; /* otherwise inside inner cone and no atten */
       c *= falloff;
       if(c > 0.0 && Lights[j].shadow_buffer != NULL &&
         Object[obj].show_shadow)PointInShadow(p,j,&c);
     }
     Iab = amblight+(1.0-amblight)*c;
   }
   else{
     Iab=Object[obj].ambient_light;
     falloff=1.0;
   }
   if(Lights[j].dc){
     cc = DepthCueLight(j,cc);
     Iab *= cc;
   }
   IdR += Iab*Lights[j].Ire;
   IdG += Iab*Lights[j].Igr;
   IdB += Iab*Lights[j].Ibl;
   if((f->fattrib & 0x20) == 0x20){/* shiny  switch is on  */
     VECSUB(pl,p,bv)                /* calculate specular highlight */
     normalize(bv);
     c=DOT(bv,n);
     if(c > 0.0)Iab = spec_pow[f->matsp][min(max_spec,(long)(c*max_spec))];
     else Iab = 0;
     if(Lights[j].type == SPOTLIGHT)Iab *= falloff;
     if(Lights[j].dc)Iab *= cc;
     IsR += Iab*Lights[j].Ire;
     IsG += Iab*Lights[j].Igr;
     IsB += Iab*Lights[j].Ibl;
   }
 }
 colorb0=colorb1=colorb2=0.0;
 if(Map != NULL){
   if(Map->rp >= 0){
     MapMirror(&colorr[0],&colorr[1],&colorr[2],Map,p,n);
     flag_mirror=1; srf=Map->rd;
   }
   if(Map->pp < 0)goto ALGOTEXT;
   if((f->fattrib & 0x02) == 0x02)clear = TRANSP;
   else                           clear = EOPAQUE;
   if(Map->map == MAP_BY_VERTEX){
     if((f->fattrib & 0x08) == 0x08)type = TILE;
     else                           type = SINGLE;
     if(MapByCoordinates(Map,obj,f,type,clear,p,&colorb0,&colorb1,&colorb2)
        == 0)goto ALGOTEXT;
   }
   else{
     if((f->fattrib & 0x08) == 0x08)type = TILE;
     else                           type = SINGLE;
     /* miss the map or in clear area */
    if(MapBrush(Map,type,clear,p,&colorb0,&colorb1,&colorb2) == 0){
       goto ALGOTEXT;
     }
   }
   if((f->fattrib & 0x04) == 0){       /* shaded map required */
     colorb0 = colorb0*IdR+IsR*255;
     colorb1 = colorb1*IdG+IsG*255;
     colorb2 = colorb2*IdB+IsB*255;
   }
   bf=Map->pd; sf=(1.0-bf);
   if(Map->pp == 100){
     colors[0]=colors[1]=colors[2]=0.0;
     goto SKIPALGO;
   }
 }
 /* algorithmic shaders */
 ALGOTEXT:
 if(xtext) goto SKIPALGO1;
 else if(textID == 0){
   VECCOPY((double)f->color,colors)
   goto SKIPALGO1; /* no shader applied */
 }
 else if(textID == 1){/* wood finish */
   VECCOPY((double)f->color,colors)
   wood_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 2){/* marble finish */
   VECCOPY((double)f->color,colors)
   marble_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 12){ /* sky  */
   VECCOPY((double)f->color,colors)
   bozo_grain(colors,obj,p,f->texco,
             (double)(f->fparam)*7.8125e-3,
             (double)(f->matsp)*0.375,(long)(f->faxis));
   IdR = IdG = IdB = 1.0;
   IsR = IsG = IsB = 0.0;
 }
 else if(textID == 3){ /* granite  */
   VECCOPY((double)f->color,colors)
   granite_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 if(textID == 4){   /* old glass */
   VECCOPY((double)f->color,colors)
   goto SKIPALGO1;
 }
 else if(textID == 5){ /* luminous */
   VECCOPY((double)f->color,colors)
   IdR = IdG = IdB = 1.0;
   IsR = IsG = IsB = 0.0;
 }
 else if(textID == 6){ /* flashing */
   VECCOPY((double)f->color,colors)
   if(flashing){
     IdR = IdG = IdB = 1.0;
     IsR = IsG = IsB = 0.0;
   }
   else{
     IdR = IdG = IdB = 0.1;
     IsR = IsG = IsB = 0.0;
   }
 }
 else if(textID == 10){ /* spotted  */
   VECCOPY((double)f->color,colors)
   spotted_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 11){ /* agate  */
   VECCOPY((double)f->color,colors)
   agate_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 14){/* blend finish */
   VECCOPY((double)f->color,colors)
   blend_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 15){/* more realistic wood finish */
   VECCOPY((double)f->color,colors)
/*   wood_grain2(colors,obj,p,f->texco,(double)f->fparam*);  */
   wood_grain2(colors,obj,p,f->texco,0.078125,(long)(f->faxis));
 }
 SKIPALGO1:
 colors[0] = colors[0]*IdR+IsR*255;
 colors[1] = colors[1]*IdG+IsG*255;
 colors[2] = colors[2]*IdB+IsB*255;
 SKIPALGO:
 colors[0] = colors[0]*sf + colorb0*bf;
 colors[1] = colors[1]*sf + colorb1*bf;
 colors[2] = colors[2]*sf + colorb2*bf;
 if((j=Object[obj].shadowed) != NO && ((ZobjChar[col] & (1 << j)) != 0)){
   colors[0] *= shadow_density;     /* do the shadows */
   colors[1] *= shadow_density;
   colors[2] *= shadow_density;
 }
 if(f->matre > 0 && ! flag_mirror){    /* mirror added to finish */
   if(f->faxis > 0 && Object[obj].mirrorID > 0)
        GetPseudoMirror(obj,row,col,colorr);
   else GetMirrorValues(colorr,p,p,n,0,0,0);
   srf = (double)(f->matre)/255.0;
   flag_mirror=1;
 }
 if(flag_mirror){
   bf=1.0-srf;
   Red[col]  = (unsigned char)min(255,colors[0]*bf+colorr[0]*srf);
   Green[col]= (unsigned char)min(255,colors[1]*bf+colorr[1]*srf);
   Blue[col] = (unsigned char)min(255,colors[2]*bf+colorr[2]*srf);
 }
 else{
   Red[col]  = (unsigned char)min(255,colors[0]);
   Green[col]= (unsigned char)min(255,colors[1]);
   Blue[col] = (unsigned char)min(255,colors[2]);
 }
}

static void AttenuateGlass(long row, long col,
                    UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                    long *Zglass[], long *ZglassO[], double *Zglassd[],
                    double *Zposn){
 long i;                          /* if the face on top is glass */
 for(i=NGLASS-1;i>=0;i--){        /* add in its surface attribs  */
   if(Zglass[i][col] >= 0)AttenuateGlassSurface(i,row,col,Red,Green,Blue,
                                                Zglass,ZglassO,Zglassd,
                                                Zposn);
 }
}

static void AttenuateGlassSurface(long id, long row, long col,
                         UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                         long *Zglass[], long *ZglassO[], double *Zglassd[],
                         double *Zposn){
 double tr,tr1,srf=0.0,srf1=1.0;
 double Iab,IdR,IdG,IdB,IsR,IsG,IsB,c,cc,det,alpha,beta,bend,bend1,falloff;
 double pdotn,ldotn;
 double colorb0=0,colorb1=0,colorb2=0,colorr0=0,colorr1=0,colorr2=0,
        colors[3],colorm[3];
 vector n,p,pr,pl,bv;
 long j,obj;
 short clear,type;
 face    *f;
 eface   *e;
 pface   *g;
 vertex  *v;
 imap    *Map;
 obj=ZglassO[id][col];
 f=Object[obj].Fbase+Zglass[id][col];
 v=Object[obj].Vbase;
 e=Object[obj].Ebase+Zglass[id][col];
 g=Object[obj].Pbase+Zglass[id][col];
 if((f->fattrib & 0x40) == 0x40){
   if((long)f->fmap >= Object[obj].NoMaps)Map=NULL;
   else Map=(Object[obj].Mbase+(long)f->fmap);
 }
 else Map=NULL;
 if(f->fattrib >> 7 == 0)VECCOPY(e->n,n)  /* phong = NO */
 else{
   det=(
       (v[f->V[2]].y - v[f->V[0]].y) * (v[f->V[1]].x - v[f->V[0]].x)
      -(v[f->V[2]].x - v[f->V[0]].x) * (v[f->V[1]].y - v[f->V[0]].y)
       );
   if(det == 0){
     VECCOPY(e->n,n)
   }
   else{
     det=1.0/det;
     alpha= (double)(
        (v[f->V[2]].y - v[f->V[0]].y) * (Zposn[col] - v[f->V[0]].x)
       -(v[f->V[2]].x - v[f->V[0]].x) * ((double)row - v[f->V[0]].y)
                    )*det;
     beta= (double)(
        ((double)row - v[f->V[0]].y) * (v[f->V[1]].x - v[f->V[0]].x)
       -(Zposn[col] - v[f->V[0]].x) * (v[f->V[1]].y - v[f->V[0]].y)
                   )*det;
     n[0]=g->n[0][0] + alpha*(g->n[1][0] - g->n[0][0])
                     +  beta*(g->n[2][0] - g->n[0][0]);
     n[1]=g->n[0][1] + alpha*(g->n[1][1] - g->n[0][1])
                     +  beta*(g->n[2][1] - g->n[0][1]);
     n[2]=g->n[0][2] + alpha*(g->n[1][2] - g->n[0][2])
                     +  beta*(g->n[2][2] - g->n[0][2]);
     normalize(n);
   }
 }
 /* p is the point on the glass face that is seen at this pixel */
 p[0]=(((double)col-(double)Xcentre)/scalex*Zglassd[id][col]);
 p[1]=Zglassd[id][col];
 p[2]=(((double)Ycentre-(double)row)/scaley*Zglassd[id][col]);
 if((pdotn=DOT(p,e->n)) > 0){n[0] = -n[0]; n[1] = -n[1]; n[2] = -n[2];}
 if(Map != NULL && Map->bp >= 0){
   if(Map->map == MAP_BY_VERTEX)
     BumpByCoordinates(Map,obj,f,p,n);
   else{
     if((f->fattrib & 0x08) == 0x08)map_bumpy(TILE,obj,Map,p,n);
     else                           map_bumpy(SINGLE,obj,Map,p,n);
   }
 }
 tr=((double)(f->mattr))*4.1666666e-3;
 VECCOPY((double)f->color,colors)  /* glass surface colour */
 if((f->fattrib & 0x10) == 0x10){ /* external shader */
   double alpha_channel=0.0;
   int s_c,s_cid,t,ctype;
   double beta,gamma;
   t=f->faxis;
   if(t == 0 && Object[obj].Tbase != NULL){ /* if shader axis = default */
     ctype=1;                               /* + v's have map co-ords   */
     MapExternalByCoords(obj,f,p,&alpha,&beta,&gamma);
   }
   else{   /* use axes */
     ctype=0;
     gamma=hitpoint(Object[obj].Abase[t].bn,Object[obj].Abase[t].bu,
               Object[obj].Abase[t].bv,Object[obj].Abase[t].bo,p,&alpha,&beta);
     gamma /= max(1.0,Object[obj].Abase[t].tl);
   }
   for(s_c=0; s_c < 8; s_c++){
     if((s_cid=Object[obj].shader[s_c]) >= 0 &&
        ((f->fshader) & (1 << s_c)))
     GetExternalTexture(
      t,
      s_cid,
      ctype,
      p,n,
      alpha,beta,gamma,
      (double)f->fparam*SHSCALE,  /* bumpyess */
      &alpha_channel,
      &tr,
      f->texco,
      obj,f,
      colors);
   }
 }
 IdR=IdG=IdB=0.0; IsR=IsG=IsB=0.0;
 for(j=0;j<Nlights;j++)if(Lights[j].type != SHADOWS &&
                          Lights[j].type != DUMMYL){
   vecsub(Lights[j].p,p,pl);
   cc=R_length(pl);
   vecscale(1.0/cc,pl,pr);
   ldotn=DOT(pr,n);
   if((pdotn > 0.0  && ldotn <  0.0) ||
      (pdotn <= 0.0 && ldotn >= 0.0)){
     Iab = max(0.1,dot(pr,n));
     if(Lights[j].type == SPOTLIGHT){
       if((alpha=DOT(pr,Lights[j].d)) < Lights[j].dot2)falloff=0.0;
       else if(alpha < Lights[j].dot1){ /* in between cones */
         falloff=(alpha-Lights[j].dot2)*Lights[j].falloff;
       }
       else falloff=1.0; /* inside inner */
       Iab = max(0.1,Iab*falloff);
     }
   }
   else{
     Iab=Object[obj].ambient_light;
     falloff=1.0;
   }
   if(Lights[j].dc){
     cc = DepthCueLight(j,cc);
     Iab *= cc;
   }
   IdR += Iab*Lights[j].Ire;
   IdG += Iab*Lights[j].Igr;
   IdB += Iab*Lights[j].Ibl;
   if((f->fattrib & 0x20) == 0x20){/* shiny glass */
     vecsub(pl,p,bv);
     normalize(bv);
     c=dot(bv,n);
     if(c > 0.0)Iab = spec_pow[f->matsp][min(max_spec,(long)(c*max_spec))];
     else Iab = 0;
     if(Lights[j].type == SPOTLIGHT)Iab *= falloff;
     if(Lights[j].dc)Iab *= cc;
     IsR += Iab*Lights[j].Ire;
     IsG += Iab*Lights[j].Igr;
     IsB += Iab*Lights[j].Ibl;
   }
 }

 if((f->fattrib & 0x1) == 0x1){       /* glass edge is on */
   if((bend=DOT(p,n)) < 0)bend = -bend;
   bend1=sqrt(bend/R_length(p));
   tr1  = 1.0-tr;
   tr *= bend1;                 /* to give glass some edge */
 }
 else{
   tr1  = 1.0-tr;
/* not used */
//     bend1=bend/R_length(p);   /* fade outside of glass */
//     bend1 *= bend1;
//     tr1  = bend1*(1.0-(f->mattr)/240.0);
//     tr = 1.0-tr1;
//     IdR=IdG=IdB=1.0;
/* under test */
 }

 if(Map != NULL){
   if(Map->rp >= 0){
     MapMirror(&colorr0,&colorr1,&colorr2,Map,p,n);
     srf=Map->rd; srf1=(1.0-srf);
   }
   if(Map->pp < 0)goto MISSBRUSH;
   if((f->fattrib & 0x02) == 0x02)clear = TRANSP;
   else                           clear = EOPAQUE;
   if(Map->map == MAP_BY_VERTEX){
     if((f->fattrib & 0x08) == 0x08)type = TILE;
     else                           type = SINGLE;
     if(MapByCoordinates(Map,obj,f,type,clear,p,&colorb0,&colorb1,&colorb2)
        == 0)goto MISSBRUSH1;
   }
   else{
     if((f->fattrib & 0x08) == 0x08)type  = TILE;
     else                           type  = SINGLE;
     /* miss the map or in clear area */
     if(MapBrush(Map,type,clear,p,&colorb0,&colorb1,&colorb2) == 0){
       goto MISSBRUSH1;
     }
   }
   if((f->fattrib & 0x04) == 0x04){       /* unshaded map required */
     IdR=IdG=IdB=1.0; IsR=IsG=IsB=0.0;
   }
   tr1=Map->pd; tr=(1.0-tr1);    /* tr is 0-1 % of transparency */
   *(colors  ) = colorb0;
   *(colors+1) = colorb1;
   *(colors+2) = colorb2;
   MISSBRUSH:
   if(Map->tp >= 0){ /* transparency map - white is fully transparent */
     if(Map->map == MAP_BY_VERTEX){
       if(MapTransByCoords(Map,obj,f,p,&tr) != 0)tr1=(1.0-tr);
     }
     else{
       if((f->fattrib & 0x08) == 0x08)type = TILE;
       else                           type = SINGLE;
       if(MapTransByAxis(Map,type,p,&tr) != 0)tr1=(1.0-tr);
     }
   }
   MISSBRUSH1:
   IdR=(double)Red[col]*tr   + (colors[0]*IdR+IsR*255)*tr1;
   IdG=(double)Green[col]*tr + (colors[1]*IdG+IsG*255)*tr1;
   IdB=(double)Blue[col]*tr  + (colors[2]*IdB+IsB*255)*tr1;
   IdR=IdR*srf1 + colorr0*srf;
   IdG=IdG*srf1 + colorr1*srf;
   IdB=IdB*srf1 + colorr2*srf;
 }
 else{
   IdR=(double)Red[col]*tr   + colors[0]*tr1*IdR+IsR*255;
   IdG=(double)Green[col]*tr + colors[1]*tr1*IdG+IsG*255;
   IdB=(double)Blue[col]*tr  + colors[2]*tr1*IdB+IsB*255;
   if(f->matre > 0){
     if(f->faxis > 0 && Object[obj].mirrorID > 0)
          GetPseudoMirror(obj,row,col,colorm);
     else GetMirrorValues(colorm,p,p,n,0,0,0);
     srf = (double)(f->matre)/255.0; srf1=(1.0-srf);
     IdR=IdR*srf1 + colorm[0]*srf;
     IdG=IdG*srf1 + colorm[1]*srf;
     IdB=IdB*srf1 + colorm[2]*srf;
   }
 }
 Red[col]  =(unsigned char)min(255.0,IdR);
 Green[col]=(unsigned char)min(255.0,IdG);
 Blue[col] =(unsigned char)min(255.0,IdB);
}

void GetTracePixelValue(long row, long col, imap *Map, eface *e,
                        face *f, long obj, vector nn,  /* obj is object id */
                        UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                        double *Zdepth,UCHAR *ZobjChar){
 BOOL     xtext;
 vector   p,pr,bv,pl,n,d;
 long     j,flag_mirror=0,t,flag_trans=0;
 short    type,textID,clear;
 double   colors[3],colorr[3],colorm[3],colort[3];
 double   Iab,IdR,IdG,IdB,IsR,IsG,IsB,c,cc,
          pdotn,ldotn,amblight,falloff,alpha;
 double   sf,sf1,tr,tr1,rf,rf1;
 /* p is the point on the face that is seen at this pixel */
 p[0]=(((double)col-(double)Xcentre)/scalex*Zdepth[col]);
 p[1]=Zdepth[col];
 p[2]=(((double)Ycentre-(double)row)/scaley*Zdepth[col]);
 IdR=IdG=IdB=0.0; IsR=IsG=IsB=0.0;
 textID = f->fshader; xtext=(BOOL)(f->fattrib & 0x10);
 tr=((double)(f->mattr))*4.1666666e-3;
 sf=1.0;
 /* n is to be normal to surface at that point p */
 VECCOPY((double)f->color,colors)
 if((pdotn=DOT(p,e->n)) < 0)VECCOPY(nn,n)
 else     {n[0] = -nn[0]; n[1] = -nn[1]; n[2] = -nn[2];}
 if(Map != NULL && Map->bp >= 0){
   if(Map->map == MAP_BY_VERTEX)
     BumpByCoordinates(Map,obj,f,p,n);
   else{
     if((f->fattrib & 0x08) == 0x08)map_bumpy(TILE,obj,Map,p,n);
     else                           map_bumpy(SINGLE,obj,Map,p,n);
   }
 }
 if(textID > 6 || xtext){  // else if  removed to allow for combo bump map
   if(xtext){ /* external shader */
     double alpha_channel=0.0;
     int s_c,s_cid,t,ctype;
     double beta,gamma;
     t=f->faxis;
     if(t == 0 && Object[obj].Tbase != NULL){ /* if shader axis = default */
       ctype=1;                               /* + v's have map co-ords   */
       MapExternalByCoords(obj,f,p,&alpha,&beta,&gamma);
     }
     else{   /* use axes */
       ctype=0;
       gamma=hitpoint(Object[obj].Abase[t].bn,Object[obj].Abase[t].bu,
                 Object[obj].Abase[t].bv,Object[obj].Abase[t].bo,p,&alpha,&beta);
       gamma /= max(1.0,Object[obj].Abase[t].tl);
     }
     for(s_c=0; s_c < 8; s_c++){
       if((s_cid=Object[obj].shader[s_c]) >= 0 && (textID & (1 << s_c)))
       GetExternalTexture(
        t,
        s_cid,
        ctype,
        p,n,
        alpha,beta,gamma,
        (double)f->fparam*SHSCALE,  /* bumpyess */
        &alpha_channel,
        &tr,
        f->texco,
        obj,f,
        colors);
     }
   }
   else if(textID ==  7)new_bumpy  (obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
   else if(textID ==  8)aztec_bumpy(obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
   else if(textID == 13)dents      (obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
   else if(textID ==  9)wrinkles   (obj,p,n,(double)f->fparam*SHSCALE,(long)(f->faxis));
#ifdef TRACER
   else if(textID == 17){
     t=f->faxis;
     waves2(p,n,Object[obj].Abase[t].bu,
                Object[obj].Abase[t].bv,
                Object[obj].Abase[t].bo,
                n,Rip_Freq,wave_phase,         /* 0.2 is max   */
                (double)f->fparam * 0.000390625);/* 0.02 is good */
   }
#endif
 }
 for(j=0;j<Nlights;j++)if(Lights[j].type != SHADOWS &&
                          Lights[j].type != DUMMYL){
   VECSUB(Lights[j].p,p,pl)
   cc=R_length(pl);
   VECSCALE(1.0/cc,pl,pr)
/*   ldotn=DOT(pr,e->n); this was used to try to eliminate fringe pixels
     but it appears to cause some faces to de-phong incorrectly */
   ldotn=DOT(pr,n);
   if((pdotn > 0.0 && ldotn < 0.0)  || (pdotn <= 0.0 && ldotn >= 0.0)){
     amblight=Object[obj].ambient_light;
     c = max(0.0,DOT(pr,n));
     if(Lights[j].type == SPOTLIGHT){
/* note the < is because DOT is max when pr and n are aligned */
/* dot2 is the outer cone, dot1 is the inner cone             */
/* d of light is in OPPOSITE dirn. to the dirn of the light   */
/* pr is vector from point to light thus if pr.n = 1 light is */
/* pointing straight at the point                             */
       if((alpha=DOT(pr,Lights[j].d)) < Lights[j].dot2)falloff=0;
       else if(alpha < Lights[j].dot1){ /* in between cones */
         falloff=(alpha-Lights[j].dot2)*Lights[j].falloff;
       }
       else falloff=1.0; /* otherwise inside inner cone and no atten */
       c *= falloff;
       if(c > 0.0 && Lights[j].shadow_buffer != NULL &&
         Object[obj].show_shadow)PointInShadow(p,j,&c);
     }
     Iab = amblight+(1.0-amblight)*c;
   }
   else{
     Iab=Object[obj].ambient_light;
     falloff=1.0;
   }
   if(Lights[j].dc){
     cc = DepthCueLight(j,cc);
     Iab *= cc;
   }
   IdR += Iab*Lights[j].Ire;
   IdG += Iab*Lights[j].Igr;
   IdB += Iab*Lights[j].Ibl;
   if((f->fattrib & 0x20) == 0x20){/* shiny  switch is on  */
     VECSUB(pl,p,bv)                /* calculate specular highlight */
     normalize(bv);
     c=DOT(bv,n);
     if(c > 0.0)Iab = spec_pow[f->matsp][min(max_spec,(long)(c*max_spec))];
     else Iab = 0;
     if(Lights[j].type == SPOTLIGHT)Iab *= falloff;
     if(Lights[j].dc)Iab *= cc;
     IsR += Iab*Lights[j].Ire;
     IsG += Iab*Lights[j].Igr;
     IsB += Iab*Lights[j].Ibl;
   }
 }
 colorm[0]=colorm[1]=colorm[2]=0.0;
 if(Map != NULL){
   if(Map->rp >= 0){
     MapMirror(&colorr[0],&colorr[1],&colorr[2],Map,p,n);
     flag_mirror=1; rf=Map->rd;
   }
   if(Map->tp >= 0){ /* transparency map - white is fully transparent */
#ifdef TRACER
     if(Map->map == MAP_BY_VERTEX){
       if(MapTransByCoords(Map,obj,f,p,&tr) != 0)tr1=(1.0-tr);
     }
     else{
       if((f->fattrib & 0x08) == 0x08)type = TILE;
       else                           type = SINGLE;
       if(MapTransByAxis(Map,type,p,&tr) != 0)tr1=(1.0-tr);
     }
     if(trace_refractions && tr > 0.0){
       c=(double)((p[0]*p[0]) + (p[1]*p[1]) + (p[2]*p[2]));
       c=sqrt(c);
       VECSCALE((1.0/c),p,d)
       GetBackdropValue(row,col,p,d,colort);
       trace_transmission_ray(p,d,n,obj,f,(double)BIGP,colort,0,0,0);
       flag_trans=1;
     }
#endif
   }
   if(Map->pp < 0)goto ALGOTEXT;
   if((f->fattrib & 0x02) == 0x02)clear = TRANSP;
   else                           clear = EOPAQUE;
   if(Map->map == MAP_BY_VERTEX){
     if((f->fattrib & 0x08) == 0x08)type = TILE;
     else                           type = SINGLE;
     if(MapByCoordinates(Map,obj,f,type,clear,p,
                         &colorm[0],&colorm[1],&colorm[2])
        == 0)goto ALGOTEXT;
   }
   else{
     if((f->fattrib & 0x08) == 0x08)type = TILE;
     else                           type = SINGLE;
     if(MapBrush(Map,type,clear,p,&colorm[0],&colorm[1],&colorm[2]) == 0){
       goto ALGOTEXT;  /* miss the map - or in clear area */
     }
     flag_trans=2;
   }
   if((f->fattrib & 0x04) == 0){       /* shaded map required */
     colorm[0] = colorm[0]*IdR+IsR*255;        /* put spec after mix  */
     colorm[1] = colorm[1]*IdG+IsG*255;
     colorm[2] = colorm[2]*IdB+IsB*255;
   }
   sf=(1.0-Map->pd);          /* map "pd" is 0.0-1.0  or pp */
   if(Map->pp == 100){        /* make sure only map is used */
     colors[0]=colors[1]=colors[2]=0.0;
     goto SKIPALGO;
   }
 }
 /* algorithmic shaders */
 ALGOTEXT:
 if(xtext)goto SKIPALGO;
 else if(textID == 0){
   goto SKIPALGO; /* no shader applied */
 }
 else if(textID == 1){/* wood finish */
   wood_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 2){/* marble finish */
   marble_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 12){ /* sky  */
   bozo_grain(colors,obj,p,f->texco,
             (double)(f->fparam)*7.8125e-3,
             (double)(f->matsp)*0.375,(long)(f->faxis));
   IdR = IdG = IdB = 1.0;
   IsR = IsG = IsB = 0.0;
 }
 else if(textID == 3){ /* granite  */
   granite_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 if(textID == 4){   /* old glass */
   goto SKIPALGO;
 }
 else if(textID == 5){ /* luminous */
   VECCOPY((double)f->color,colors)
   IdR = IdG = IdB = 1.0;
   IsR = IsG = IsB = 0.0;
 }
 else if(textID == 6){ /* flashing */
   if(flashing){
     IdR = IdG = IdB = 1.0;
     IsR = IsG = IsB = 0.0;
   }
   else{
     IdR = IdG = IdB = 0.1;
     IsR = IsG = IsB = 0.0;
   }
 }
 else if(textID == 10){ /* spotted  */
   spotted_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 11){ /* agate  */
   agate_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 14){/* blend finish */
   blend_grain(colors,obj,p,f->texco,(long)(f->faxis));
 }
 else if(textID == 15){/* more realistic wood finish */
   wood_grain2(colors,obj,p,f->texco,0.078125,(long)(f->faxis));
 }
#ifdef TRACER
 else if(trace_refractions && textID == 16){/* refractive glass */
   c=(double)((p[0]*p[0]) + (p[1]*p[1]) + (p[2]*p[2]));
   c=sqrt(c);
   VECSCALE((1.0/c),p,d)
   trace_refraction_ray(p,d,n,obj,f,(double)BIGP,colort,0,0,0);
   flag_trans=1;// tr=1.0;
   tr1=1.0-tr;
 }
 else if(trace_refractions && textID == 17){/* refractive liquid water */
   c=(double)((p[0]*p[0]) + (p[1]*p[1]) + (p[2]*p[2]));
   c=sqrt(c);
   VECSCALE((1.0/c),p,d)
   trace_refraction_ray(p,d,n,obj,f,(double)BIGP,colort,0,0,0);
   flag_trans=1;// tr=1.0;
   tr1=1.0-tr;
 }
#endif
 SKIPALGO:
 /* combine the shader and map with appropriate illimination */
 sf1=1.0-sf;
 colors[0] = colors[0]*IdR*sf + colorm[0]*sf1;
 colors[1] = colors[1]*IdG*sf + colorm[1]*sf1;
 colors[2] = colors[2]*IdB*sf + colorm[2]*sf1;
 /* colors now is the surface value including diffuse lighting */
#ifdef TRACER
 if(trace_shadows && Object[obj].show_shadow &&
    (j=check_for_shadow(obj,p,e->n)) != 0){
   IsR = IsG = IsB = 0.0;
   if(j < 0){
     colors[0] *= shadow_density;     /* do the shadows */
     colors[1] *= shadow_density;
     colors[2] *= shadow_density;
   }
   else{ /* j layers of glass */
     colors[0] *= glass_shadow_density;
     colors[1] *= glass_shadow_density;
     colors[2] *= glass_shadow_density;
   }
 }
 if(trace_refractions && f->mattr > 0 && flag_trans == 0){ /* double sided */
   double bend,bend1;
   if((f->fattrib & 0x1) == 0x1){       /* glass edge is on */
     if((bend=DOT(p,n)) < 0)bend = -bend;
     bend1=sqrt(bend/R_length(p));
     tr1  = 1.0-tr;
     tr *= bend1;                 /* to give glass some edge */
   }
   else{
     tr1  = 1.0-tr;
   }
   c=(double)((p[0]*p[0]) + (p[1]*p[1]) + (p[2]*p[2]));
   c=sqrt(c);
   VECSCALE((1.0/c),p,d)
   GetBackdropValue(row,col,p,d,colort);
   trace_transmission_ray(p,d,n,obj,f,(double)BIGP,colort,0,0,0);
   flag_trans=1;
 }
#endif
 if(flag_trans == 1){  /* mix any transmissivity */
   colors[0] = colort[0]*tr + colors[0]*tr1;
   colors[1] = colort[1]*tr + colors[1]*tr1;
   colors[2] = colort[2]*tr + colors[2]*tr1;
 }
 colors[0] += IsR*255; /* mix the specular highlights */
 colors[1] += IsG*255;
 colors[2] += IsB*255;
 if(f->matre > 0 && ! flag_mirror){    /* mirror added to finish */
   GetMirrorValues(colorr,p,p,n,0,0,0);
   rf = (double)(f->matre)/255.0;
   flag_mirror=1;
 }
 if(flag_mirror){
   rf1=1.0-rf;
   Red[col]  = (unsigned char)min(255,colorr[0]*rf+colors[0]*rf1);
   Green[col]= (unsigned char)min(255,colorr[1]*rf+colors[1]*rf1);
   Blue[col] = (unsigned char)min(255,colorr[2]*rf+colors[2]*rf1);
 }
 else{
   Red[col]  = (unsigned char)min(255,colors[0]);
   Green[col]= (unsigned char)min(255,colors[1]);
   Blue[col] = (unsigned char)min(255,colors[2]);
 }
}
