/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// file PLUGINS.C

// This is the file where the MESH effects POSTPROCESSOR
// and TEXTURE plugins are loaded and called.

#define MODULE_FX

#include "render.h"

#ifdef _SUNSTYLE
static void ExternalFX();
static short ExternalFXram();
static void ExternalFXfile();
static void InternalFX_Wave();
static void InternalFX_Modulate();
static void InternalFX_Squash();
static void InternalFX_Warp();
static void InternalFX_Spring();
static void InternalFX_Shake();
static short AllocateAnotherShader();
#else
static void ExternalFX(long global, char parameters[]);
static short ExternalFXram(char parameters[]);
static void ExternalFXfile(char parameters[]);
static void InternalFX_Wave(char parameters[]);
static void InternalFX_Modulate(char parameters[]);
static void InternalFX_Squash(char parameters[]);
static void InternalFX_Warp(char parameters[]);
static void InternalFX_Spring(char parameters[]);
static void InternalFX_Shake(char parameters[]);
static short AllocateAnotherShader(void);
#endif

#include "..\common\effects\sfxinfo.h"

static sfxinfo SFXinfo;
static sfxdata SFXdata;
static long FXObj;

#ifdef _SUNSTYLE
static short AllocateAnotherShader(){
#else
static short AllocateAnotherShader(void){
#endif
 shader *sp;
 if(Shader == NULL)sp=(shader *)X__Malloc(sizeof(shader));
 else              sp=(shader *)X__Realloc(Shader,sizeof(shader)*(Nshader+1));
 if(sp == NULL){
   Render_Message(13,0,NULL);
   return FAIL;
 }
 Shader=sp;
 return OK;
}

#include "..\common\animate\wspaces.c"

long MotionBlurFrame(long frame, long subframe, long command_id){
  long i,j;
  fullscreenbuffer *f,*b;
  double df1,df,dft[9]={0.9,0.85,0.8,0.75,0.7,0.65,0.6,0.55,0.5};
  if(mblurON < 2)return 0;  /* do it in a DLL */
//  if(mblurON > 10)mblurON=10;
  f=FullScreenBuffer;
  b=MotionBlurBuffer;
  if(mblurON > 10){
    df1=1.0/(double)subframe;
    df=(double)(subframe-1);
    for(i=0;i<ResolutionY;i++){
      for(j=0;j<ResolutionX;j++){
        b->R = (unsigned char)(((double)(f->R)+(double)(b->R)*df)*df1);
        b->G = (unsigned char)(((double)(f->G)+(double)(b->G)*df)*df1);
        b->B = (unsigned char)(((double)(f->B)+(double)(b->B)*df)*df1);
//        f->R = b->R; f->G = b->G; f->B = b->B;
        b++; f++;
      }
    }
    if(command_id == 0){
      memcpy(FullScreenBuffer,MotionBlurBuffer,ResolutionX*ResolutionY*(long)sizeof(fullscreenbuffer));
      memset(MotionBlurBuffer,0,ResolutionX*ResolutionY*(long)sizeof(fullscreenbuffer));
    }
    return 0;
  }
  if(frame == 1)df=1.0;
  else          df=dft[mblurON-2];
  df1=1.0-df;
  for(i=0;i<ResolutionY;i++){
    for(j=0;j<ResolutionX;j++){
      b->R = (unsigned char)((double)(f->R)*df+(double)(b->R)*df1);
      b->G = (unsigned char)((double)(f->G)*df+(double)(b->G)*df1);
      b->B = (unsigned char)((double)(f->B)*df+(double)(b->B)*df1);
      f->R = b->R; f->G = b->G; f->B = b->B;
      b++; f++;
    }
  }
  return 0;
}

#ifdef _SUNSTYLE
void EffectObjectVertices(global,O) long global, long O; {
#else
void EffectObjectVertices(long global, long O){
#endif
  double amplitude,wavelength,displacement;
  long i,j,n;
  vertex *v;
  double t,vmax[3],vmin[3];
  if((n=(long)Object[O].effect) < 4)return;
  if(global == 1 && n < 50)return; /* no globals for internal */
  for(i=0;i<3;i++){vmax[i] = (double)(BIGN); vmin[i] = (double)BIGP;}
  t=Object[O].effectratio;
  v=Object[O].Vbase;
  for(i=0;i<Object[O].NoVertices;i++){ /* get object limits */
    for(j=0;j<3;j++){
       if(global == 0)v[i].p[j] -= Object[O].Origin[j]; /* apply offset */
       vmin[j] = min(vmin[j],v[i].p[j]);
       vmax[j] = max(vmax[j],v[i].p[j]);
    }
  }
  FXObj=O;
#if __LITE__
  SFXinfo.version=0;
#else
  SFXinfo.version=1;
#endif
  SFXinfo.nvert=Object[O].NoVertices;
  SFXinfo.time=t;
  SFXinfo.ltime=(long)(t*1000);
  SFXinfo.vmin[0]=(long)vmin[0];   SFXinfo.vmax[0]=(long)vmax[0];
  SFXinfo.vmin[1]=(long)vmin[1];   SFXinfo.vmax[1]=(long)vmax[1];
  SFXinfo.vmin[2]=(long)vmin[2];   SFXinfo.vmax[2]=(long)vmax[2];
  SFXinfo.origin[0]=Object[O].Origin[0];
  SFXinfo.origin[1]=Object[O].Origin[1];
  SFXinfo.origin[2]=Object[O].Origin[2];
  SFXinfo.Nfaces=Object[O].NoFaces;
  SFXinfo.Faces=(void *)Object[O].Fbase;
  if(global == 1){ /* origin include in positon matrix already applied */
    n=50; /* only external on global */
    memset(&SFXdata,0,sizeof(sfxdata));
    SFXdata.size=sizeof(sfxdata);
    for(i=0;i<3;i++){
      SFXdata.pos[i]=(long)Object[O].Offset[i];
      SFXdata.camera_pos[i]=(long)ViewPoint[i];
    }
    if(R_Nground == 1){
      SFXdata.ground=TRUE;
      for(i=0;i<3;i++)SFXdata.ground_pos[i]=(long)Ground.pin[i];
    }
  }
  switch((short)n){
   case 4:  InternalFX_Wave(Object[O].effect_param_list);     break;
   case 5:  InternalFX_Modulate(Object[O].effect_param_list); break;
   case 6:  InternalFX_Squash(Object[O].effect_param_list);   break;
   case 7:  InternalFX_Warp(Object[O].effect_param_list);     break;
   case 8:  InternalFX_Spring(Object[O].effect_param_list);   break;
   case 9:  InternalFX_Shake(Object[O].effect_param_list);    break;
   case 10: X_message("Effect Flap not implemented\n");       break;
   case 50: ExternalFX(global,Object[O].effect_param_list);   break;
   default: break;
  }
  if(global == 1)return;
  for(i=0;i<Object[O].NoVertices;i++){
    for(j=0;j<3;j++){
      v[i].p[j] += Object[O].Origin[j];  /* undo offset */
    }
  }
}


#ifdef _SUNSTYLE
static void ExternalFXfile(parameters) char parameters[]; {
#else
static void ExternalFXfile(char parameters[]){
#endif
 FILE *fsfx;
 long size,lp,tl;
 short i,j;
 char fname[128], tfile[128], buffer[16], c;
 sscanf(parameters,"%s",fname);
 if(X__getenv("EVT") == NULL){
  Render_Message(63,1,NULL);
  return;
 }
 sprintf(tfile,"%ssfx_com.tmp",X__getenv("EVT"));
 fsfx=fopen(tfile,"wb");
 if(fsfx == NULL){
  Render_Message(64,1,NULL);
  return;
 }
 /* Info chunk */
 fprintf(fsfx,"INFO");
 size=(long)( sizeof(long)*4 + sizeof(double)*7 );
 fwrite(&size,sizeof(long),1,fsfx);
 fwrite(&SFXinfo.nvert,sizeof(long),1,fsfx);
 fwrite(&SFXinfo.ltime,sizeof(long),1,fsfx);
 fwrite(&SFXinfo.vmin,sizeof(long),3,fsfx);
 fwrite(&SFXinfo.vmax,sizeof(long),3,fsfx);
 fwrite(&SFXinfo.origin,sizeof(long),3,fsfx);
 /* Parameter chunk */
 fprintf(fsfx,"PRMS");
 size=(long)strlen(parameters);
 fwrite(&size,sizeof(long),1,fsfx);
 fwrite(parameters,sizeof(char),size,fsfx);
 /* Vert chunk */
 fprintf(fsfx,"VERT");
 size=(long)SFXinfo.nvert * (long)sizeof(double) * (long)3;
 fwrite(&size,sizeof(long),1,fsfx);
 for(lp=0L;lp<SFXinfo.nvert;lp++){
  for(j=0;j<3;j++){
   tl=(long)(Object[FXObj].Vbase[lp].p[j]);
   if(fwrite(&tl,sizeof(long),1,fsfx) != 1 || fflush(fsfx) != 0){
    Render_Message(65,1,NULL);
    fclose(fsfx); remove(tfile); return;
   }
  }
 }
 fclose(fsfx);
 i=spawnl(P_WAIT,fname,fname,tfile,NULL);
 if(i != 0){
  Render_Message(66,0,fname);
  remove(tfile); return;
 }
 fsfx=fopen(tfile,"rb");
 if(fsfx == NULL){
  Render_Message(0,1,NULL);
  return;
 }
 for(i=0; i<16; ++i)buffer[i]=0;
 while(1){
  if(fread(buffer,sizeof(char),4,fsfx) != 4) break;
  if(strcmp(buffer,"VERT") == 0){
   fread(&size,sizeof(long),1,fsfx);
   for(lp=0L; lp < SFXinfo.nvert; lp++){
    for(j=0; j < 3; j++){
     if(fread(&tl,sizeof(long),1,fsfx) != 1){
      Render_Message(67,0,NULL);
      fclose(fsfx); remove(tfile); return;
     }
     Object[FXObj].Vbase[lp].p[j]=(double)tl;
    }
   }
  }
  else {    /* Dump chunk */
   if(fread(&size,sizeof(long),1,fsfx) != 1) break;
   for(lp=0L; lp<size; ++lp) fread(&c,1,1,fsfx);
  }
 }
 Render_Message(68,0,NULL);
 fclose(fsfx);
 remove(tfile);
}

#ifdef _SUNSTYLE
static void InternalFX_Wave(parameters) char parameters[]; {
#else
static void InternalFX_Wave(char parameters[]){
#endif
 double amplitude,wavelength,displacement,PRMamp,PRMlngth,PRMfreq,d,maxd;
 char buffer[128];
 int type, PRMaxis, axis, fade, paxis, paxis2;
 long i;
 vertex *v;
 sscanf(parameters,"%s %d %d %f %f %f %d",buffer,&type,&PRMaxis,&PRMamp,
        &PRMlngth,&PRMfreq,&fade);
 v=Object[FXObj].Vbase;
 if(PRMaxis == 3) axis=1;         /* F / B */
 else if(PRMaxis == 1) axis=2;    /* U / D */
 else if(PRMaxis == 2) axis=0;    /* L / R */
 if(axis == 0){      paxis = 1; paxis2 = 2; }
 else if(axis == 2){ paxis = 1; paxis2 = 0; }
 else if(axis == 1){ paxis = 0; paxis2 = 2; }
 wavelength=max((SFXinfo.vmax[paxis]-SFXinfo.vmin[paxis]) * PRMlngth,1.0);
 amplitude=(PRMamp/25.0) * (wavelength/8);
 if(type == 2){                        /* Radial */
  maxd=0.00;
  for(i=0;i<SFXinfo.nvert;i++){
   d=sqrt(v[i].p[paxis]*v[i].p[paxis] + v[i].p[paxis2]*v[i].p[paxis2]);
   maxd=max(maxd,d);
  }
 }
 for(i=0;i<SFXinfo.nvert;i++){
  if(type == 1) d=v[i].p[paxis];       /* Linear */
  else if(type == 2){                  /* Radial */
   d=sqrt(v[i].p[paxis]*v[i].p[paxis] + v[i].p[paxis2]*v[i].p[paxis2]);
  }
  displacement=amplitude*cos(PI*2*((SFXinfo.time*PRMfreq)-d/wavelength));
  if(type == 1){                       /* Linear */
   d = v[i].p[paxis]-SFXinfo.vmin[paxis];
   maxd = SFXinfo.vmax[paxis]-SFXinfo.vmin[paxis];
  }
  if(fade == 2) displacement *= d / maxd;             /* fade up */
  else if(fade == 3) displacement *= 1 - d / maxd;    /* fade down */
  v[i].p[axis] += displacement;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Modulate(parameters) char parameters[]; {
#else
static void InternalFX_Modulate(char parameters[]){
#endif
 double scale,dpmax,amplitude,PRMamp,PRMlngth,PRMfreq;
 int PRMaxis, axis, paxis, paxis2;
 char buffer[128];
 long i;
 vertex *v;
 sscanf(parameters,"%s %d %f %f %f",buffer,&PRMaxis,&PRMamp,&PRMlngth,&PRMfreq);
 if(PRMaxis == 3) axis=1;         /* F / B */
 else if(PRMaxis == 1) axis=2;    /* U / D */
 else if(PRMaxis == 2) axis=0;    /* L / R */
 if(axis == 0){      paxis = 1; paxis2 = 2; }
 else if(axis == 2){ paxis = 1; paxis2 = 0; }
 else if(axis == 1){ paxis = 0; paxis2 = 2; }
 v=Object[FXObj].Vbase;
 dpmax=max((SFXinfo.vmax[axis]-SFXinfo.vmin[axis]) * PRMlngth,1.0);
 amplitude = 0.1 * (PRMamp / 25.0);
 for(i=0;i<SFXinfo.nvert;i++){
  scale=1.0+amplitude*cos(PI*2*((SFXinfo.time*PRMfreq)-v[i].p[axis]/dpmax));
  v[i].p[paxis]  *= scale;
  v[i].p[paxis2] *= scale;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Squash(parameters) char parameters[]; {
#else
static void InternalFX_Squash(char parameters[]){
#endif
 double scale,dpmax,dpmin,t,sp;
 long i;
 char buffer[128];
 int PRMcomp, PRMeffect, PRMplane, axis, paxis, paxis2;
 vertex *v;
 sscanf(parameters,"%s %d %d %d",buffer,&PRMeffect,&PRMplane,&PRMcomp);
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 t = -1.00 * (t*t) + 2.00 * t;
 if(PRMeffect == 2) t = 1.00 - t;
 if(PRMplane == 1){           /* Horizontal Plane */
  axis=2; paxis=1; paxis2=0;
 }
 else if(PRMplane == 2){      /* Vertical Plane */
  axis=1; paxis=0; paxis2=2;
 }
 dpmax=max(fabs(SFXinfo.vmin[axis]),fabs(SFXinfo.vmax[axis]));
 dpmin=dpmax-t*dpmax*((double)PRMcomp/100.0);
 sp=dpmin/dpmax;
 for(i=0;i<SFXinfo.nvert;i++){
  scale=1.0+((double)PRMcomp/100.0)*t*cos(PI/2*v[i].p[axis]/dpmax);
  v[i].p[axis]   *= sp;
  v[i].p[paxis]  *= scale;
  v[i].p[paxis2] *= scale;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Warp(parameters) char parameters[]; {
#else
static void InternalFX_Warp(char parameters[]){
#endif
 double scale,dpmax,dpmin,t,sp;
 long i;
 int PRMwarp;
 char buffer[128];
 vertex *v;
 sscanf(parameters,"%s %d",buffer,&PRMwarp);
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 dpmax=SFXinfo.vmax[1]-SFXinfo.vmin[1];
 if(PRMwarp == 2){                  /* Warp in */
  if(t <= 0.5)sp=1.0+2.0*t;
  else        sp=2.0-2.0*(t-0.5);
  scale=2.0*(t-0.5);
  for(i=0;i<SFXinfo.nvert;i++){
   if(t <= 0.5)v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp + SFXinfo.vmin[1];
   else v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp +
                    SFXinfo.vmin[1] + scale*dpmax;
  }
 }
 else if(PRMwarp == 1){             /* Warp Out */
  t=1.0-SFXinfo.time;
  for(i=0;i<SFXinfo.nvert;i++){
   v[i].p[1] = v[i].p[1] + t*dpmax;
  }
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Spring(parameters) char parameters[]; {
#else
static void InternalFX_Spring(char parameters[]){
#endif
 double scale,dpmax,dpmin,t,sp;
 char buffer[128];
 int PRMcomp;
 long i;
 vertex *v;
 sscanf(parameters,"%s %d",buffer,&PRMcomp);
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 scale=4.0 * ((double)PRMcomp / 75.0) * (t-0.75);
 if     (t <= 0.50)sp=1.0-2.0*t*t;
 else if(t <= 0.75)sp=0.5+1.5*(t-0.5);
 else              sp=2.0-scale;
 dpmax=SFXinfo.vmax[1]-SFXinfo.vmin[1];
 for(i=0;i<SFXinfo.nvert;i++){
  if(t <= 0.75)v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp +
                           SFXinfo.vmin[1];
  else v[i].p[1] = (v[i].p[1] - SFXinfo.vmin[1])*sp + SFXinfo.vmin[1] +
                   scale*dpmax;
 }
}

#ifdef _SUNSTYLE
static void InternalFX_Shake(parameters) char parameters[]; {
#else
static void InternalFX_Shake(char parameters[]){
#endif
 double displacement,dpmax,dpmin,t,sp;
 long i;
 vertex *v;
 v=Object[FXObj].Vbase;
 t=SFXinfo.time;
 dpmax=(double)(SFXinfo.vmax[2]-SFXinfo.vmin[2])*
       (double)(SFXinfo.vmax[2]-SFXinfo.vmin[2]);
 dpmin=(double)(SFXinfo.vmax[0]-SFXinfo.vmin[0])/4.0;
 sp=dpmin*sin(2*PI*SFXinfo.time);
 for(i=0;i<SFXinfo.nvert;i++){
  displacement=sp*((double)(v[i].p[2]-SFXinfo.vmin[2])/dpmax)*
                   (double)(v[i].p[2]-SFXinfo.vmin[2]);
  v[i].p[0] += displacement;
 }
}

/*       the rest is machine specific    */

#if !__NT__  && (__ZTC__ || __WATCOMC__)


#ifdef _SUNSTYLE
long ExternalImageProcess(long) long frame; {
#else
long ExternalImageProcess(long frame){
#endif
 return 0;
}

#ifdef _SUNSTYLE
static void ExternalFX(global,parameters) long global; char parameters[]; {
#else
static void ExternalFX(long global, char parameters[]){
#endif
#if __ZTC__ || __WATCOMC__
/*  if(ExternalFXram(parameters) < 0)ExternalFXfile(parameters); */
  ExternalFXfile(parameters);
#else
  ExternalFXfile(parameters);
#endif
}

#include "external.h"

short LoadExternalTexture(char *name, char *parameters){
 long l,i,result;
 char *n;
 long (*fpFun1)(long, long, char *);
 n=short_form(name);
 n=strlwr(n);
 l=strlen(n);
 *(n+l-4)=0;
 for(i=0;i<Render_NExternalTextures;i++){
   if(strcmp(n,Render_ExternalFunctionList[i].name) == 0){
//printf("Matches shader %ld [%s]\n",i,Render_ExternalFunctionList[i].name);
     if(AllocateAnotherShader() == FAIL)return FAIL;
     strcpy(Shader[Nshader].file,n);
     strncpy(Shader[Nshader].parameters,parameters,128);
     Shader[Nshader].module   = Render_ExternalFunctionList[i].ExternalTextureStartup;
     Shader[Nshader].function = Render_ExternalFunctionList[i].ExternalTextureProcedure;
     Shader[Nshader].close    = Render_ExternalFunctionList[i].ExternalTextureClose;
     Shader[Nshader].morph    = Render_ExternalFunctionList[i].ExternalTextureMorph;
     fpFun1 = (void *)Shader[Nshader].module;
     result=(*fpFun1)(RenderCurrentFrame,0,parameters);
     Nshader++;
     return OK;
   }
 }
 Render_Message(93,0,name);
 return FAIL;
}

void FreeExternalTextures(void){
 long i;
 void (*fpFun2)(void);
 if(Nshader > 0){
   for(i=0;i<Nshader;i++){
     if(Shader[i].close != NULL){
       fpFun2 = (void *)Shader[i].close;
       (*fpFun2)();
     }
//printf("\nExternal shader %s freed\n\n",Shader[i].file);
   }
 }
 if(Shader != NULL)X__Free(Shader);
 Shader=NULL;
 Nshader=0;
}

short MorphExternalTexture(char *name, char *parameters, long obj){
 long (*fpFun1)(char *, double);
 long n,result;
 if((n=Nshader-1) < 0)return FAIL; /* assumes called IMMEDIATELY after load */
 if(Shader[n].morph != NULL){
   fpFun1 = (void *)Shader[n].morph;
   result=(*fpFun1)(parameters,Object[obj].morphratio);
   return OK;
 }
 return FAIL;
}

void GetExternalTexture(long axisid,
                        long tid, long ctype,
                        vector p, vector n,
                        double alpha, double beta, double gamma,
                        double bump_scale,
                        double *ac,
                        double *transparency,
                        unsigned char sc[3],
                        long model_id, face *fp,
                        double colour[3]){
 long ( *fpTexFun1)(long,
                    vector, vector,
                    double,double,double,
                    double,
                    double,double,
                    double *,
                    unsigned char [3],
                    double [3],
                    double *, double *);
 long result;
 double reflectivity;
 /*  x and y are same a alppha beta */
 fpTexFun1=(void *)(Shader[tid].function);
 result=(*fpTexFun1)(
                     ctype,
                     p,n,
                     alpha,beta,gamma,
                     bump_scale,
                     alpha,beta,
                     ac,
                     sc,
                     colour,
                     &reflectivity,
                     transparency);
 return;
}
/* end of DOS  External shaders and effects and image processing*/
#endif

#if __NT__
static short ExternalFXnt(long global, char *parameters);
extern double noise2(double x, double y, double z, double *);
extern void Turbulence(double x, double y, double z, double *);
extern void cycloidal(double, double *);
extern void DTurbulence(vector result, double x, double y, double z);
extern double Triangle_Wave(double value);
extern vector Wave_Sources[];

extern char gszHomeDir[];

void init_structures(void);

#include "..\textures\rstruct.h"

static X__SHADER EviShader;

void init_structures(void){
 X__SHADER *lpEVI;
 lpEVI=&EviShader;
 lpEVI->fp_sNoise=noise2;
 lpEVI->fp_sTurbulence=Turbulence;
 lpEVI->fp_Cycloidal=cycloidal;
 lpEVI->fp_vNoise=Dnoise2;
 lpEVI->fp_vTurbulence=DTurbulence;
 lpEVI->fp_normalize=normalize;
 lpEVI->fp_Wave_Sources=Wave_Sources;
#if __LITE__
 lpEVI->version=0;
#else
 lpEVI->version=1;
#endif
 return;
}

static void ExternalFX(long global, char parameters[]){
  ExternalFXnt(global, parameters);
}

static short ExternalFXnt(long global, char *parameters){
 char fname[256],tparam[256];
 HMODULE hLib;
 FARPROC fpFun;
 long l,result=0;
 long (*fpFun1)(char *, sfxinfo *, vertex *);
 long (*fpFun2)(char *, sfxinfo *, sfxdata *, vertex *);
 sscanf(parameters,"%s",tparam);
 RestoreWhiteSpaces(tparam);
 sscanf(tparam,"%[^\0]",fname);
 l=strlen(fname);
 fname[l-1]='l';
 fname[l-2]='l';
 fname[l-3]='d';
 if((hLib=LoadLibrary(fname)) == NULL){ /* try to get 'dll' */
   fname[l-1]='l';
   fname[l-2]='l';
   fname[l-3]='d';
   hLib=LoadLibrary(fname);
 }
 if(hLib == NULL){ /* try the default effect directory */
   char *sfname,tfname[256];
   sfname=short_form(fname);
   strcpy(tfname,gszHomeDir);  strcat(tfname,"effects\\");
   strcat(tfname,sfname);
   hLib=LoadLibrary(tfname);
 }
 if(hLib != NULL){
   if(global == 0){
     if((fpFun=GetProcAddress(hLib,"_RenderExternalEffect")) != NULL){
       fpFun1 = (void *)fpFun;
       result=(*fpFun1)(parameters,&SFXinfo,Object[FXObj].Vbase);
     }
     else MessageBox(GetFocus(),(LPCTSTR)"Function Fail in DLL",NULL,MB_OK);
   }
   else{
     if((fpFun=GetProcAddress(hLib,"_RenderGlobalEffect")) != NULL){
       fpFun2 = (void *)fpFun;
       result=(*fpFun2)(parameters,&SFXinfo,&SFXdata,Object[FXObj].Vbase);
     }
   }
   FreeLibrary(hLib);
 }
 else MessageBox(GetFocus(),(LPCTSTR) "Load fail on DLL",NULL, MB_OK);
 return 0;
}

short LoadExternalTexture(char *fname, char *parameters){
 HMODULE hLib;
 FARPROC fpFun;
 long result=0;
 long (*fpFun1)(long, long, char *);
 void (*fpFun2)(void);
 long (*fpTexFun1)(long,
                   vector, vector,
                   double,double,double,
                   double,
                   double,double,
                   double *,
                   unsigned char [3],
                   double *, double *);
 fpTexFun1=NULL;
 hLib=LoadLibrary(fname);
 if(hLib == NULL){ /* try the default textures directory */
   char *sfname,tfname[256];
   sfname=short_form(fname);
   strcpy(tfname,gszHomeDir);  strcat(tfname,"textures\\");
   strcat(tfname,sfname);
   hLib=LoadLibrary(tfname);
 }
 if(hLib != NULL){
   if((fpFun=GetProcAddress(hLib,"_ExternalTextureStartup")) != NULL){
     fpFun1 = (void *)fpFun;
     result=(*fpFun1)(RenderCurrentFrame,0,parameters);
   }
   if((fpFun=GetProcAddress(hLib,"_ExternalTextureClose"))
     != NULL)fpFun2= (void *)fpFun;
   else      fpFun2=NULL;
   if((fpFun=GetProcAddress(hLib,"_ExternalTextureProcedure")) != NULL){
     fpTexFun1 = (void *)fpFun;
   }
   else{
     MessageBox(GetFocus(),"Failed Texture Function",NULL, MB_OK);
     FreeLibrary(hLib);
     return FAIL;
   }
 }
 else {
   MessageBox(GetFocus(),"Failed Texture Load",NULL,MB_OK);
    return FAIL;
 }
 if(AllocateAnotherShader() == FAIL){
   FreeLibrary(hLib);
   return FAIL;
 }
 strcpy(Shader[Nshader].file,fname);
 strncpy(Shader[Nshader].parameters,parameters,128);
 Shader[Nshader].module   = (void *)hLib;
 Shader[Nshader].function = (void *)fpTexFun1;
 Shader[Nshader].close    = (void *)fpFun2;
 Shader[Nshader].morph    = NULL;  /* to be used later */
 Nshader++;
 return OK;
}

short MorphExternalTexture(char *name, char *parameters, long obj){
 HMODULE hLibTex;
 FARPROC fpFun;
 long (*fpFun1)(char *, double);
 long n,result;
 if((n=Nshader-1) < 0)return FAIL; /* assumes called IMMEDIATELY after load */
 if((hLibTex=(HMODULE)Shader[n].module) != NULL){
   if((fpFun=GetProcAddress(hLibTex,"_ExternalTextureMorph")) != NULL){
     fpFun1 = (void *)fpFun;
     result=(*fpFun1)(parameters,Object[obj].morphratio);
   }
   else return FAIL;
 }
 else return FAIL;
 return OK;
}

void FreeExternalTextures(void){
 HMODULE hLibTex;
 void (*fpFun2)(void);
 long i;
 if(Nshader > 0){
   for(i=0;i<Nshader;i++){
     if(Shader[i].close != NULL){
       fpFun2 = (void *)Shader[i].close;
       (*fpFun2)();
     }
     hLibTex=(HMODULE)Shader[i].module;
     FreeLibrary(hLibTex);
   }
 }
 X__Free(Shader);
 Shader=NULL;
}

void GetExternalTexture(long axisid, long tid, long ctype,
                        vector p, vector n,
                        double alpha, double beta, double gamma,
                        double bump_scale,
                        double *ac,
                        double *transparency,
                        unsigned char sc[3],
                        long obj, face *fp,
                        double colour[3]){
 long (*fpTexFun1)(long,
                   vector, vector,
                   double,double,double,
                   double,
                   double,double,
                   double *,
                   unsigned char [3],
                   double [3],
                   double *, double *,
                   X__SHADER *);
 long result;
 double reflectivity;
 /*  x and y are same as alpha beta */
 fpTexFun1=(void *)(Shader[tid].function);
 EviShader.fp_axis_o=&Object[obj].Abase[axisid].bo;
 EviShader.fp_axis_u=&Object[obj].Abase[axisid].bu;
 EviShader.fp_axis_v=&Object[obj].Abase[axisid].bv;
 EviShader.fp_axis_n=&Object[obj].Abase[axisid].bn;
 EviShader.u_length=&Object[obj].Abase[axisid].tl;
 EviShader.v_length=&Object[obj].Abase[axisid].tv;
 EviShader.fp_axis_unorm=&Object[obj].Abase[axisid].bu_norm;
 EviShader.fp_axis_vnorm=&Object[obj].Abase[axisid].bv_norm;
 EviShader.tr=(double *)Object[obj].trpos;
 EviShader.itr=(double *)Object[obj].trinv;
 result=(*fpTexFun1)(
                     ctype,
                     p,n,
                     alpha,beta,gamma,
                     bump_scale,
                     alpha,beta,
                     ac,
                     sc,
                     colour,
                     &reflectivity,
                     transparency,
                     &EviShader);
}

#include "..\common\postprocess\ximage.h"

long ExternalImageProcess(long frame){
 int i,j,k;
 char fname[256],tparam[256];
 long AnimatorIdList[256];
 HMODULE hLib;
 FARPROC fpFun;
 XIMAGE Ximage,*lpXimage;
 long l,result=0;
 long (*fpFun1)(char *, XIMAGE *);
 if(nImageProcesses == 0 || ImageProcesses == NULL)return 0;
 if(FullScreenBuffer == NULL){
   MessageBox(GetFocus(),
              "Image Processing not possible without full screen buffer",
              NULL,MB_OK);
   return FAIL;
 }
#if __LITE__
 Ximage.version=0;
#else
 Ximage.version=1;
#endif
 Ximage.Frame=frame;
 Ximage.Xscale=scalex;
 Ximage.Yscale=scaley;
 if(anti_alias == LOW)   {Ximage.Xscale /= 2;}
 if(anti_alias == MEDIUM){Ximage.Xscale /= 2; Ximage.Yscale /= 2;}
 if(anti_alias == HIGH)  {Ximage.Xscale /= 3; Ximage.Yscale /= 2;}
 if(anti_alias == ULTRA) {Ximage.Xscale /= 3; Ximage.Yscale /= 3;}
 Ximage.Xmax=ResolutionX;
 Ximage.Ymax=ResolutionY;
 Ximage.Screen=FullScreenBuffer;
 Ximage.Blur=MotionBlurBuffer;
 Ximage.ObjectBuffer=ObjectIdBuffer;
 Ximage.Zbuffer=fszBuffer;
 Ximage.Nlights=Nlights;
 Ximage.Lights=Lights;
 Ximage.Ntargets=0;
 Ximage.Targets=NULL;
 Ximage.shadow_buffer_size=shadow_buffer_size;
 for(k=0;k<min(250,ObjectCount);k++)AnimatorIdList[k]=Object[k].AnimatorID;
 Ximage.AnimatorIdList=AnimatorIdList;
 Ximage.ViewPoint = (double *)ViewPoint;
 Ximage.CamPhi=CamPhi;
 Ximage.CamTheta=CamTheta;
 Ximage.CamAlpha=CamAlpha;
 Ximage.CamSx=CamSx;
 Ximage.CamSy=CamSy;
 Ximage.CamSz=CamSz;
 Ximage.CamIma=CamIma;
 Ximage.CamIm=CamIm;
 for(k=0;k<6;k++)Ximage.Pad[k]=0;
 Ximage.fp_Terminate=R_terminator;
 Ximage.fp_Yield=RenderYield;
 Ximage.fp_sNoise=noise2;
 Ximage.fp_sTurbulence=Turbulence;
 Ximage.fp_Cycloidal=cycloidal;
 Ximage.fp_vNoise=Dnoise2;
 Ximage.fp_vTurbulence=DTurbulence;
 Ximage.fp_normalize=normalize;
 lpXimage=&Ximage;
 for(k=0;k<nImageProcesses;k++){
   lpXimage->Morph=ImageProcesses[k].imo;
   lpXimage->MorphRatio=ImageProcesses[k].fmr;
   lpXimage->first_frame=ImageProcesses[k].first_frame;
   lpXimage->this_frame=ImageProcesses[k].this_frame;
   lpXimage->last_frame=ImageProcesses[k].last_frame;
   lpXimage->aParameters=ImageProcesses[k].parameters;
   lpXimage->mParameters=ImageProcesses[k].mparameters;
   sscanf(ImageProcesses[k].parameters,"%s",tparam);
   RestoreWhiteSpaces(tparam);
   sscanf(tparam,"%[^\0]",fname);
   Render_Message(98,0,fname);
   hLib=LoadLibrary(fname);
   lpXimage->ModuleName=fname;
   if(hLib == NULL){ /* try the default ximage directory */
     char *sfname,tfname[256];
     sfname=short_form(fname);
     strcpy(tfname,gszHomeDir);  strcat(tfname,"postprocess\\");
     strcat(tfname,sfname);
     hLib=LoadLibrary(tfname);
     lpXimage->ModuleName=tfname;
   }
   if(hLib != NULL){
     if((fpFun=GetProcAddress(hLib,"_RenderImageProcess")) != NULL){
       for(i=0;i<4;i++)for(j=0;j<4;j++)Ximage.ViewTransform[i][j]=trr[i][j];
       fpFun1 = (void *)fpFun;
       result=(*fpFun1)(ImageProcesses[k].parameters,lpXimage);
     }
     else MessageBox(GetFocus(),
        (LPCTSTR) "Failed Function", NULL, MB_OK | MB_SYSTEMMODAL );
     FreeLibrary(hLib);
   }
   else MessageBox(GetFocus(),(LPCTSTR) "Failed Load", NULL, MB_OK | MB_SYSTEMMODAL );
   if(!result)return FAIL;
 }
 return 0;
}

#endif
