/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

#define MODULE_TRACE

// This is the raytracing module - it uses an octree spatial
// decomposition to speed things up - see the book!

#define _DEBUG 0

#include <setjmp.h>
#include "render.h"

#define UC unsigned char
#define US unsigned short

typedef struct NODE {
  long n,mi[3],ma[3];
  UC *oid;
  US *fid;
  struct NODE *octree[8];
//  long id;
//  struct NODE *last;
} node;

#define THRESHOLDZ 2048
#define THRESHOLDF 10
#define THRESHOLDR 20
//#define VXT        512
//#define VXT        0

extern short R_terminator(void);

static jmp_buf x_buf;
static long threshold_f,threshold_z,threshold_dz,recursion,
            nfaces,kcount,maxfaces,minfaces,
            trace_depth_max=3,tree_depth,tree_depth_max;
static node *ROOT;
//static double refractive_index=1.2;
static long voxel_xmin,voxel_ymin,voxel_zmin,
            voxel_xmax,voxel_ymax,voxel_zmax;

#ifdef _SUNSTYLE
static node *which_voxel();
static node *track_ray();
static void destroyBST();
static void countBST();
#else
static node *which_voxel(vector p);
static node *track_ray(node *nc, vector p, vector d);
static void destroyBST(node *np);
static void countBST(node *np);
static int MakeVoxels(node *np);
#endif

#ifdef _SUNSTYLE
void freeBST(){
#else
void freeBST(void){
#endif
 if(ROOT != NULL)destroyBST(ROOT);
}

static int InsideBox(vector p, long mi[3], long ma[3]){
 if(p[0] >= (double)mi[0] && p[0] <= (double)ma[0] &&
    p[1] >= (double)mi[1] && p[1] <= (double)ma[1] &&
    p[2] >= (double)mi[2] && p[2] <= (double)ma[2])return 1;
 return 0;
}

static int IntersectXplane(vector p1, vector p2, long plane,
                           long amin, long amax, long bmin, long bmax){
 double mu,a,b;
// if(labs((long)p2[0]-(long)p1[0]) < 1)return 0;
 if((long)p1[0] <= plane/*-VXT*/ && (long)p2[0] <= plane/*-VXT*/)return 0;
 if((long)p1[0] >= plane/*+VXT*/ && (long)p2[0] >= plane/*+VXT*/)return 0;
 mu=((double)plane-p1[0])/(p2[0]-p1[0]);
 a=p1[1]+mu*(p2[1]-p1[1]);
 if((long)a <= amin/*-VXT*/ || (long)a >= amax/*+VXT*/)return 0;
 b=p1[2]+mu*(p2[2]-p1[2]);
 if((long)b <= bmin/*-VXT*/ || (long)b >= bmax/*+VXT*/)return 0;
 return 1;
}

static int IntersectYplane(vector p1, vector p2, long plane,
                           long amin, long amax, long bmin, long bmax){
 double mu,a,b;
// if(labs((long)p2[1]-(long)p1[1]) < 1)return 0;
 if((long)p1[1] <= plane/*-VXT*/ && (long)p2[1] <= plane/*-VXT*/)return 0;
 if((long)p1[1] >= plane/*+VXT*/ && (long)p2[1] >= plane/*+VXT*/)return 0;
 mu=((double)plane-p1[1])/(p2[1]-p1[1]);
 a=p1[0]+mu*(p2[0]-p1[0]);
 if((long)a <= amin/*-VXT*/ || (long)a >= amax/*+VXT*/)return 0;
 b=p1[2]+mu*(p2[2]-p1[2]);
 if((long)b <= bmin/*-VXT*/ || (long)b >= bmax/*+VXT*/)return 0;
 return 1;
}

static int IntersectZplane(vector p1, vector p2, long plane,
                           long amin, long amax, long bmin, long bmax){
 double mu,a,b;
// if(labs((long)p2[2]-(long)p1[2]) < 1)return 0;
 if((long)p1[2] <= plane/*-VXT*/ && (long)p2[2] <= plane/*-VXT*/)return 0;
 if((long)p1[2] >= plane/*+VXT*/ && (long)p2[2] >= plane/*+VXT*/)return 0;
 mu=((double)plane-p1[2])/(p2[2]-p1[2]);
 a=p1[0]+mu*(p2[0]-p1[0]);
 if((long)a <= amin/*-VXT*/ || (long)a >= amax/*+VXT*/)return 0;
 b=p1[1]+mu*(p2[1]-p1[1]);
 if((long)b <= bmin/*-VXT*/ || (long)b >= bmax/*+VXT*/)return 0;
 return 1;
}

static int IntersectTriangle(vector pf0, vector pf1, vector pf2, vector n,
                             long x1, long y1, long z1,
                             long x2, long y2, long z2){
 vector v1,pi;
 double mu,mu1,det1,det2,det3,ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3,a,b,dm,detmax;
 int k;
 double dd;
 vector p,q,d,da,db;
 p[0]=(double)x1; p[1]=(double)y1; p[2]=(double)z1;
 q[0]=(double)x2; q[1]=(double)y2; q[2]=(double)z2;
 VECSUB(p,pf0,da)
 VECSUB(q,pf0,db)
 if((DOT(da,n) * DOT(db,n)) > 0.0)return 0;
 VECSUB(q,p,d)
 if(normalize(d))return 0;

 mu1=DOT(d,n);
 VECSUB(pf0,p,v1)
 mu=DOT(n,v1)/mu1;
 VECSCALE(mu,d,v1)
 VECSUM(p,v1,pi)
 ve1=pf1[0]-pf0[0]; ve2=pf1[1]-pf0[1]; ve3=pf1[2]-pf0[2];
 vp1=pf2[0]-pf0[0]; vp2=pf2[1]-pf0[1]; vp3=pf2[2]-pf0[2];
 p1=pi[0]-pf0[0];
 p2=pi[1]-pf0[1];
 p3=pi[2]-pf0[2];
 det1=ve1*vp2-vp1*ve2;
 det2=ve1*vp3-vp1*ve3;
 det3=ve2*vp3-vp2*ve3;
 k=0; detmax=1.e-3;
 if((dm=fabs(det1)) > detmax){k=1; detmax=dm;}
 if((dm=fabs(det2)) > detmax){k=2; detmax=dm;}
 if((dm=fabs(det3)) > detmax){k=3; detmax=dm;}
 if(k == 0)return 0;
 else if(k == 1){
   a=( vp2*p1-vp1*p2)/det1;
   b=(-ve2*p1+ve1*p2)/det1;
 }
 else if(k == 2){
   a=( vp3*p1-vp1*p3)/det2;
   b=(-ve3*p1+ve1*p3)/det2;
 }
 else if(k == 3){
   a=( vp3*p2-vp2*p3)/det3;
   b=(-ve3*p2+ve2*p3)/det3;
 }
 if( a < 0.0 || a > 1.0 || b < 0.0 || b > 1.0 || (a+b) > 1.0)return 0;
 return 1;
}

static int FaceIntersectVoxel(vector p0, vector p1, vector p2, vector n,
                              long mi[3], long ma[3]){
 double xmax,xmin,ymax,ymin,zmin,zmax;
 xmax=max((max(p0[0],p1[0])),p2[0])/*+VXT*/;  /* VXT is for pinprick holes */
 ymax=max((max(p0[1],p1[1])),p2[1])/*+VXT*/;  /* in shadows                */
 zmax=max((max(p0[2],p1[2])),p2[2])/*+VXT*/;
 xmin=min((min(p0[0],p1[0])),p2[0])/*-VXT*/;
 ymin=min((min(p0[1],p1[1])),p2[1])/*-VXT*/;
 zmin=min((min(p0[2],p1[2])),p2[2])/*-VXT*/;
 if((long)xmax < mi[0])return 0;
 if((long)ymax < mi[1])return 0;
 if((long)zmax < mi[2])return 0;
 if((long)xmin > ma[0])return 0;
 if((long)ymin > ma[1])return 0;
 if((long)zmin > ma[2])return 0;
 if(trace_optimise == 0)return 1;
 if(InsideBox(p0,mi,ma))return 1;
 if(InsideBox(p1,mi,ma))return 1;
 if(InsideBox(p2,mi,ma))return 1;
 if(IntersectXplane(p0,p1,mi[0], mi[1],ma[1],mi[2],ma[2]))return 1;
 if(IntersectXplane(p0,p1,ma[0], mi[1],ma[1],mi[2],ma[2]))return 1;
 if(IntersectYplane(p0,p1,mi[1], mi[0],ma[0],mi[2],ma[2]))return 1;
 if(IntersectYplane(p0,p1,ma[1], mi[0],ma[0],mi[2],ma[2]))return 1;
 if(IntersectZplane(p0,p1,mi[2], mi[0],ma[0],mi[1],ma[1]))return 1;
 if(IntersectZplane(p0,p1,mi[2], mi[0],ma[0],mi[1],ma[1]))return 1;
 if(IntersectXplane(p1,p2,mi[0], mi[1],ma[1],mi[2],ma[2]))return 1;
 if(IntersectXplane(p1,p2,ma[0], mi[1],ma[1],mi[2],ma[2]))return 1;
 if(IntersectYplane(p1,p2,mi[1], mi[0],ma[0],mi[2],ma[2]))return 1;
 if(IntersectYplane(p1,p2,ma[1], mi[0],ma[0],mi[2],ma[2]))return 1;
 if(IntersectZplane(p1,p2,mi[2], mi[0],ma[0],mi[1],ma[1]))return 1;
 if(IntersectZplane(p1,p2,mi[2], mi[0],ma[0],mi[1],ma[1]))return 1;
 if(IntersectXplane(p2,p0,mi[0], mi[1],ma[1],mi[2],ma[2]))return 1;
 if(IntersectXplane(p2,p0,ma[0], mi[1],ma[1],mi[2],ma[2]))return 1;
 if(IntersectYplane(p2,p0,mi[1], mi[0],ma[0],mi[2],ma[2]))return 1;
 if(IntersectYplane(p2,p0,ma[1], mi[0],ma[0],mi[2],ma[2]))return 1;
 if(IntersectZplane(p2,p0,mi[2], mi[0],ma[0],mi[1],ma[1]))return 1;
 if(IntersectZplane(p2,p0,mi[2], mi[0],ma[0],mi[1],ma[1]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,mi[0],mi[1],mi[2],mi[0],mi[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,mi[0],ma[1],mi[2],mi[0],ma[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,ma[0],ma[1],mi[2],ma[0],ma[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,ma[0],mi[1],mi[2],ma[0],mi[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,mi[0],mi[1],ma[2],mi[0],ma[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,mi[0],ma[1],ma[2],ma[0],ma[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,ma[0],ma[1],ma[2],ma[0],mi[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,ma[0],mi[1],ma[2],mi[0],mi[1],ma[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,mi[0],mi[1],mi[2],mi[0],ma[1],mi[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,mi[0],ma[1],mi[2],ma[0],ma[1],mi[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,ma[0],ma[1],mi[2],ma[0],mi[1],mi[2]))return 1;
 if(IntersectTriangle(p0,p1,p2,n,ma[0],mi[1],mi[2],mi[0],mi[1],mi[2]))return 1;
 return 0;
}

static int MakeVoxels(node *np){
 static long i,j,o,f,n,n1,n2,n3,n4,n5,n6,n7,n8,c[8];
 static double x,y,z,an;
 static vertex *v,*v0,*v1,*v2;
 static face   *F;
 static eface  *E;
 static vector p;
 RenderYield();
 if(R_terminator() == FAIL)longjmp(x_buf,-1);
 if(tree_depth > recursion)return OK;
 if(np->n < threshold_f)return OK;   /* reached voxel minimum size */
 if((np->ma[0]-np->mi[0]) < threshold_z ||
    (np->ma[1]-np->mi[1]) < threshold_z ||
    (np->ma[2]-np->mi[2]) < threshold_z)return OK;
 n=0; x=y=z=0.0;
 for(i=0;i<np->n;i++){
   o = (long)(*(np->oid+i));   f = (long)(*(np->fid+i));
   F=(Object[o].Fbase+f);      v=Object[o].Vbase;
   v0=(v + F->V[0]);   v1=(v + F->V[1]); v2=(v + F->V[2]);
   n++; an=(double)(n-1)/(double)n;
   x=x*an+(double)(v0->p[0]+v1->p[0]+v2->p[0])/(double)n;
   y=y*an+(double)(v0->p[1]+v1->p[1]+v2->p[1])/(double)n;
   z=z*an+(double)(v0->p[2]+v1->p[2]+v2->p[2])/(double)n;
 }
 x=x*0.3333; y=y*0.3333; z=z*0.3333;
 x=max((double)(np->mi[0]+threshold_dz),x);
 x=min((double)(np->ma[0]-threshold_dz),x);
 y=max((double)(np->mi[1]+threshold_dz),y);
 y=min((double)(np->ma[1]-threshold_dz),y);
 z=max((double)(np->mi[2]+threshold_dz),z);
 z=min((double)(np->ma[2]-threshold_dz),z);
 n1=n2=n3=n4=n5=n6=n7=n8=0;
 for(i=0;i<np->n;i++){
   o = (long)(*(np->oid+i));   f = (long)(*(np->fid+i));
   F=(Object[o].Fbase+f);      v=Object[o].Vbase;
   v0=(v + F->V[0]);   v1=(v + F->V[1]); v2=(v + F->V[2]);
   p[0]=(v0->p[0]+v1->p[0]+v2->p[0])*0.33333;
   p[1]=(v0->p[1]+v1->p[1]+v2->p[1])*0.33333;
   p[2]=(v0->p[2]+v1->p[2]+v2->p[2])*0.33333;
   if((long)p[0] > (long)x){
     if((long)p[1] > (long)y){
       if((long)p[2] > (long)z)n1++;
       else                    n2++;
     }
     else{
       if((long)p[2] > (long)z)n3++;
       else                    n4++;
     }
   }
   else {
     if((long)p[1] > (long)y){
       if((long)p[2] > (long)z)n5++;
       else                    n6++;
     }
     else{
       if((long)p[2] > (long)z)n7++;
       else                    n8++;
     }
   }
 }
 if(n1 == n || n2 == n || n3 == n || n4 == n ||
    n5 == n || n6 == n || n7 == n || n8 == n){
   return OK; /* can't separate any */
 }
 if(tree_depth > tree_depth_max)tree_depth_max=tree_depth;
 tree_depth++;
 for(i=0;i<8;i++){
   if((np->octree[i]=(node *)X__Malloc(sizeof(node))) == NULL)return FAIL;
//   np->octree[i]->id=kcount;
//   np->octree[i]->last=np;
   kcount++;
   np->octree[i]->n=0;
   np->octree[i]->oid=NULL;
   np->octree[i]->fid=NULL;
   for(j=0;j<8;j++)np->octree[i]->octree[j]=NULL;
 }
 np->octree[0]->mi[0]=(long)x; np->octree[0]->ma[0]=np->ma[0];
 np->octree[0]->mi[1]=(long)y; np->octree[0]->ma[1]=np->ma[1];
 np->octree[0]->mi[2]=(long)z; np->octree[0]->ma[2]=np->ma[2];
 np->octree[1]->mi[0]=(long)x; np->octree[1]->ma[0]=np->ma[0];
 np->octree[1]->mi[1]=(long)y; np->octree[1]->ma[1]=np->ma[1];
 np->octree[1]->mi[2]=np->mi[2]; np->octree[1]->ma[2]=(long)z;
 np->octree[2]->mi[0]=(long)x; np->octree[2]->ma[0]=np->ma[0];
 np->octree[2]->mi[1]=np->mi[1]; np->octree[2]->ma[1]=(long)y;
 np->octree[2]->mi[2]=(long)z; np->octree[2]->ma[2]=np->ma[2];
 np->octree[3]->mi[0]=(long)x; np->octree[3]->ma[0]=np->ma[0];
 np->octree[3]->mi[1]=np->mi[1]; np->octree[3]->ma[1]=(long)y;
 np->octree[3]->mi[2]=np->mi[2]; np->octree[3]->ma[2]=(long)z;
 np->octree[4]->mi[0]=np->mi[0]; np->octree[4]->ma[0]=(long)x;
 np->octree[4]->mi[1]=(long)y; np->octree[4]->ma[1]=np->ma[1];
 np->octree[4]->mi[2]=(long)z; np->octree[4]->ma[2]=np->ma[2];
 np->octree[5]->mi[0]=np->mi[0]; np->octree[5]->ma[0]=(long)x;
 np->octree[5]->mi[1]=(long)y; np->octree[5]->ma[1]=np->ma[1];
 np->octree[5]->mi[2]=np->mi[2]; np->octree[5]->ma[2]=(long)z;
 np->octree[6]->mi[0]=np->mi[0]; np->octree[6]->ma[0]=(long)x;
 np->octree[6]->mi[1]=np->mi[1]; np->octree[6]->ma[1]=(long)y;
 np->octree[6]->mi[2]=(long)z; np->octree[6]->ma[2]=np->ma[2];
 np->octree[7]->mi[0]=np->mi[0]; np->octree[7]->ma[0]=(long)x;
 np->octree[7]->mi[1]=np->mi[1]; np->octree[7]->ma[1]=(long)y;
 np->octree[7]->mi[2]=np->mi[2]; np->octree[7]->ma[2]=(long)z;
 n=np->n;
 for(i=0;i<8;i++){
  c[i]=0;
  if((np->octree[i]->oid=(UC *)X__Malloc(sizeof(UC)*n)) == NULL)return FAIL;
  if((np->octree[i]->fid=(US *)X__Malloc(sizeof(US)*n)) == NULL)return FAIL;
 }
 for(i=0;i<n;i++){
   o = (long)(*(np->oid+i));
   f = (long)(*(np->fid+i));
   F=(Object[o].Fbase+f);
   E=(Object[o].Ebase+f);
   v=Object[o].Vbase;
   v0=(v + F->V[0]);
   v1=(v + F->V[1]);
   v2=(v + F->V[2]);
   for(j=0;j<8;j++){
     if(FaceIntersectVoxel(v0->p,v1->p,v2->p,E->n,
                           np->octree[j]->mi,np->octree[j]->ma)){
       *(np->octree[j]->oid+c[j])=(UC)o;
       *(np->octree[j]->fid+c[j])=(US)f;
       c[j]++;
     }
   }
 }
 for(i=0;i<8;i++){
   if(c[i] == 0){
     X__Free(np->octree[i]->oid); np->octree[i]->oid=NULL;
     X__Free(np->octree[i]->fid); np->octree[i]->fid=NULL;
   }
   else{
     np->octree[i]->oid=(UC *)X__Realloc(np->octree[i]->oid,sizeof(UC)*c[i]);
     np->octree[i]->fid=(US *)X__Realloc(np->octree[i]->fid,sizeof(US)*c[i]);
   }
   np->octree[i]->n=c[i];
 }
 X__Free(np->oid); np->oid=NULL;
 X__Free(np->fid); np->fid=NULL;
 np->n=0;

 /* and on to next level */
 MakeVoxels(np->octree[0]);
 MakeVoxels(np->octree[1]);
 MakeVoxels(np->octree[2]);
 MakeVoxels(np->octree[3]);
 MakeVoxels(np->octree[4]);
 MakeVoxels(np->octree[5]);
 MakeVoxels(np->octree[6]);
 MakeVoxels(np->octree[7]);
 tree_depth--;
 return OK;
}

#ifdef _SUNSTYLE
int buildBST(){
#else
int buildBST(void){
#endif
 char temp[256];
 long i,j,k,l,coord,nv=0;
 vertex *v;
 face   *f;
 nfaces=0; kcount=0; tree_depth=0; tree_depth_max=0;
 voxel_xmin=voxel_ymin=voxel_zmin=BIGP;
 voxel_xmax=voxel_ymax=voxel_zmax=BIGN;
 Render_Message(31,0,NULL);
 ROOT=NULL;
 for(i=0;i<ObjectCount;i++){
   nfaces += Object[i].NoFaces;
   nv += Object[i].NoVertices;
 }
 if(nfaces == 0)return 0;
 recursion=THRESHOLDR;
 threshold_f=THRESHOLDF;
 threshold_z=THRESHOLDZ;
 recursion=trace_optimise_depth;
 threshold_f=trace_optimise_threshold;
 threshold_z=trace_optimise_size*512;
 threshold_dz=threshold_z/4;
 threshold_f += nfaces/2000;
 if((ROOT=(node *)X__Malloc(sizeof(node))) == NULL)return FAIL;
 ROOT->n=0;
 if((ROOT->oid=(UC *)X__Malloc(sizeof(UC ) * nfaces)) == NULL){
   destroyBST(ROOT); return FAIL;
 }
 if((ROOT->fid=(US *)X__Malloc(sizeof(US ) * nfaces)) == NULL){
   destroyBST(ROOT); return FAIL;
 }
 ROOT->n=nfaces;
 for(i=0,k=0;i<ObjectCount;i++){
   f=Object[i].Fbase;
   for(j=0;j<Object[i].NoFaces;j++){
     *(ROOT->oid+k)=(UC)i;
     *(ROOT->fid+k)=(US)j;
     k++;
     f++;
   }
 }
 for(i=0;i<8;i++)ROOT->octree[i]=NULL;
 ROOT->mi[0]=ROOT->mi[1]=ROOT->mi[2]=BIGP;
 ROOT->ma[0]=ROOT->ma[1]=ROOT->ma[2]=BIGN;
 kcount++; //ROOT->id=kcount; ROOT->last=NULL;
 for(i=0,k=0;i<ObjectCount;i++){
   v = Object[i].Vbase;
   for(j=0;j<Object[i].NoVertices;j++){
     for(l=0;l<3;l++){
       if((coord=(long)(v->p[l])) < ROOT->mi[l])ROOT->mi[l]=coord;
       if( coord                  > ROOT->ma[l])ROOT->ma[l]=coord;
     }
     k++;
     v++;
   }
 }
 if(k == 0)return 0;
 for(i=0;i<3;i++){
   ROOT->mi[i] -= 2048;
   ROOT->ma[i] += 2048;
 }
 tree_depth=1;
 /* make the voxels */
 if(setjmp(x_buf) != 0 || MakeVoxels(ROOT) == FAIL){
   destroyBST(ROOT);
   return FAIL;
 }
 i=nfaces;
 nfaces=0;
 maxfaces=0;
 minfaces=65536;
 countBST(ROOT);
 sprintf(temp,
  "Scene: F=%ld V=%ld [Dp=%ld Th=%ld F/v=%ld] Size %ld(%ld) [Vx=%ld Fc=%ld]",
        i,nv,tree_depth_max,threshold_f,maxfaces,
        max(min(voxel_xmax,voxel_ymax),voxel_zmax)/512,
        min(min(voxel_xmin,voxel_ymin),voxel_zmin)/512,
        kcount,nfaces);
 WindowedRenderInfo(temp);
#if _DEBUG
{
// debug=fopen("d:\\debug.rnd","w");
 sprintf(temp,"%ld voxels %ld polygons %ld vertices depth %ld threshold %ld\n",
        kcount,i,nv,tree_depth_max,threshold_f);
 MessageBox(NULL,temp,"Debug",MB_OK);
 sprintf(temp,"voxels sizes\nRoot size %ld %ld %ld\nmin x=%ld y=%ld z=%ld\nmax x=%ld y=%ld z=%ld",
  ROOT->ma[0]-ROOT->mi[0],
  ROOT->ma[1]-ROOT->mi[1],
  ROOT->ma[2]-ROOT->mi[2],
  voxel_xmin,voxel_ymin,voxel_zmin,voxel_xmax,voxel_ymax,voxel_zmax);
 MessageBox(NULL,temp,"Debug",MB_OK);
 sprintf(temp,"%ld faces assigned\n max faces/voxel %ld\n min faces/voxel %ld",
 nfaces,maxfaces,minfaces);
 MessageBox(NULL,temp,"Debug",MB_OK);
// fclose(debug); debug=NULL;
}
#endif
 return 0;
}

#ifdef _SUNSTYLE
static void destroyBST(np) node *np; {
#else
static void destroyBST(node *np){
#endif
 if(np->octree[0] != NULL){destroyBST(np->octree[0]); np->octree[0]=NULL;}
 if(np->octree[1] != NULL){destroyBST(np->octree[1]); np->octree[1]=NULL;}
 if(np->octree[2] != NULL){destroyBST(np->octree[2]); np->octree[2]=NULL;}
 if(np->octree[3] != NULL){destroyBST(np->octree[3]); np->octree[3]=NULL;}
 if(np->octree[4] != NULL){destroyBST(np->octree[4]); np->octree[4]=NULL;}
 if(np->octree[5] != NULL){destroyBST(np->octree[5]); np->octree[5]=NULL;}
 if(np->octree[6] != NULL){destroyBST(np->octree[6]); np->octree[6]=NULL;}
 if(np->octree[7] != NULL){destroyBST(np->octree[7]); np->octree[7]=NULL;}
 if(np->fid != NULL){X__Free(np->fid); np->fid=NULL;}
 if(np->oid != NULL){X__Free(np->oid); np->oid=NULL;}
 X__Free(np); np=NULL;
 return;
}

#ifdef _SUNSTYLE
static void countBST(np) node *np; {
#else
static void countBST(node *np){
#endif
 char temp[256];
 int i;
 if(np->ma[0]-np->mi[0] < voxel_xmin)voxel_xmin=np->ma[0]-np->mi[0];
 if(np->ma[1]-np->mi[1] < voxel_ymin)voxel_ymin=np->ma[1]-np->mi[1];
 if(np->ma[2]-np->mi[2] < voxel_zmin)voxel_zmin=np->ma[2]-np->mi[2];
 if(np->ma[0]-np->mi[0] > voxel_xmax)voxel_xmax=np->ma[0]-np->mi[0];
 if(np->ma[1]-np->mi[1] > voxel_ymax)voxel_ymax=np->ma[1]-np->mi[1];
 if(np->ma[2]-np->mi[2] > voxel_zmax)voxel_zmax=np->ma[2]-np->mi[2];
 if(np->octree[0] != NULL)countBST(np->octree[0]);
 if(np->octree[1] != NULL)countBST(np->octree[1]);
 if(np->octree[2] != NULL)countBST(np->octree[2]);
 if(np->octree[3] != NULL)countBST(np->octree[3]);
 if(np->octree[4] != NULL)countBST(np->octree[4]);
 if(np->octree[5] != NULL)countBST(np->octree[5]);
 if(np->octree[6] != NULL)countBST(np->octree[6]);
 if(np->octree[7] != NULL)countBST(np->octree[7]);
 if(np->fid != NULL){
   if(np->n > maxfaces)maxfaces=np->n;
   if(np->n < minfaces)minfaces=np->n;
   nfaces += np->n;
#if 0
   fprintf(debug,"Voxel ID %ld  Parent ID %ld - %ld faces\n",np->id,
   ((np->last == NULL) ? (-1) : (np->last->id)),
   np->n);
   fprintf(debug,"min %ld %ld %ld\n",np->mi[0], np->mi[1], np->mi[2]);
   fprintf(debug,"max %ld %ld %ld\n",np->ma[0], np->ma[1], np->ma[2]);

   if(np->n > 0){
     for(i=0;i<np->n;i++)printf("Obj %ld face %ld\n",
        (int)(*(np->oid + i)),(int)(*(np->fid + i)));
   }
#endif
 }
 return;
}

#ifdef _SUNSTYLE
static node *which_voxel(p) vector p; {
#else
static node *which_voxel(vector p){
#endif
 int i;
 node *np,*np0;
 for(i=0;i<3;i++){
   if(p[i] < (double)ROOT->mi[i] || p[i] > (double)ROOT->ma[i])return NULL;
 }
 np=ROOT;
 NXT:
 for(i=0;i<8;i++){
   if((np0=np->octree[i]) == NULL)return np;
   if(p[0] >= (double)np0->mi[0] && p[0] <= (double)np0->ma[0] &&
      p[1] >= (double)np0->mi[1] && p[1] <= (double)np0->ma[1] &&
      p[2] >= (double)np0->mi[2] && p[2] <= (double)np0->ma[2]){
     np=np0;
     goto NXT;
   }
 }
#if _DEBUG
 MessageBox(NULL,"Bad Voxel",NULL,MB_OK);
#endif
 return NULL; /* should never be reached */
}

//#define TOL    1.e-3
//#define TOL1   2.0
//#define TOL3   25.0    /* 100 is OK  2 is too small   1024 is abs max */
//#define TOL    1.e-5   /* these figures are crucial to the operation  */
//#define TOL1   10.0    /* of the tracer. TOL1 is crucial to the non   */
//#define TOL3   100.0   /* appearance of pin-prick holes. TOL3 to mov- */
#define TOL    1.e-5     /* ing into the next voxel                     */
#define TOL1   0.0075
#define TOL3   5.0

#ifdef _SUNSTYLE
static node *track_ray(nc,p,d) node *nc, vector p, vector d; {
#else
static node *track_ray(node *nc, vector p, vector d){ /* d is normalized */
#endif
 double mu=BIGP,mu1,ddn,ppn,xmin,xmax,ymin,ymax,zmin,zmax;
 vector x;
 node *next;
 int ncc=0;
 xmin=(double)(nc->mi[0]); xmax=(double)(nc->ma[0]);
 ymin=(double)(nc->mi[1]); ymax=(double)(nc->ma[1]);
 zmin=(double)(nc->mi[2]); zmax=(double)(nc->ma[2]);
/*printf("track %lf %lf %lf\n",xmin,ymin,zmin); */
/*printf("      %lf %lf %lf\n",xmax,ymax,zmax); */
/*printf("p in  %lf %lf %lf\n",p[0],p[1],p[2]); */
/*printf("d     %lf %lf %lf\n",d[0],d[1],d[2]); */
 TRYAGAIN:
 ddn=d[0];
 if(fabs(ddn) > TOL){
   ppn=xmin-p[0]; mu1= ppn/ddn;
   if(mu1 > TOL1 && mu1 < mu)mu=mu1;
   ppn=xmax-p[0]; mu1= ppn/ddn;
   if(mu1 > TOL1 && mu1 < mu)mu=mu1;
 }
 ddn=d[1];
 if(fabs(ddn) > TOL){
   ppn=ymin-p[1]; mu1= ppn/ddn;
   if(mu1 > TOL1 && mu1 < mu)mu=mu1;
   ppn=ymax-p[1]; mu1= ppn/ddn;
   if(mu1 > TOL1 && mu1 < mu)mu=mu1;
 }
 ddn=d[2];
 if(fabs(ddn) > TOL){
   ppn=zmin-p[2]; mu1= ppn/ddn;
   if(mu1 > TOL1 && mu1 < mu)mu=mu1;
   ppn=zmax-p[2]; mu1= ppn/ddn;
   if(mu1 > TOL1 && mu1 < mu)mu=mu1;
 }
if(mu >= BIGP)Render_Message(33,0,NULL);
 mu += TOL3;  /* move into next cell */
 VECSCALE(mu,d,x)
 VECSUM(p,x,p)
 next = which_voxel(p);
 if(next == nc){
   ncc++; goto TRYAGAIN;
 }
#if _DEBUG
 if(ncc > 0)MessageBox(NULL,"Too many loops",NULL,MB_OK);
#endif
 return next;
}

#ifdef _SUNSTYLE
short check_for_shadow(obj,p,natp) long obj, vector p, vector natp; {
#else
short check_for_shadow(long obj, vector p, vector natp){
#endif
/* p is point for shadow   natp  is normal at point p */
 int i,j,o,glass_count=0;
 node *nps,*np,*nl,*npx;
 vector d,pc;
 face   *f;
 eface  *e;
 vertex *v;
 double depth,ld,alpha;
 if((nps=which_voxel(p)) == NULL)return 0;
 for(j=0;j<Nlights;j++)if(Lights[j].type != NOSHADOW &&
                          Lights[j].type != DUMMYL){
   if(Lights[j].type == SPOTLIGHT && Lights[j].shadow_buffer != NULL)continue;
   np=nps;
   nl=which_voxel(Lights[j].p);
   VECCOPY(p,pc)
   VECSUB(Lights[j].p,p,d)
    if(DOT(p,natp)*DOT(d,natp) > 0)continue; /* not illuminated by light */
    ld= (double)((d[0]*d[0]) + (d[1]*d[1]) + (d[2]*d[2]));
    if(ld > 1.e-10){
      ld=sqrt(ld);
      VECSCALE((1.0/ld),d,d)
      if(Lights[j].type == SPOTLIGHT){
        if((alpha=DOT(d,Lights[j].d)) < Lights[j].dot2){
             continue; /* outside light cone */
        }
      }
      for(;;){/* check for intersection with faces in this voxel */
         if(np->n > 0){ /* some faces */
           for(i=0;i<np->n;i++){
             o= (int)(*(np->oid + i));
             if(!Object[o].cast_shadow)continue;
             if(!Object[obj].self_shadow && obj == o)continue;
             f=Object[o].Fbase + (int)(*(np->fid + i));
             v=Object[o].Vbase;
             e=Object[o].Ebase + (int)(*(np->fid + i));
             if(R_intersect_face(p,d,
               (v + f->V[0])->p,(v + f->V[1])->p,(v + f->V[2])->p,
               e->n,&depth,2.0) && (depth < ld)){
               if(f->mattr > 16)glass_count++;
               else return -1;
            }
           }
         }
       if(np == nl)break;
       np=track_ray(np,pc,d); /* move on to next voxel */
       if(np == NULL)break;
     }
   }
 }
 return glass_count;
}

#if 1
short trace_reflection_ray(vector p, vector d, /* d is normalized */
                           double depth_max, double *colour,
                           long trace_reflection_depth,
                           long trace_refraction_depth,
                           long refractive_flag){
 int i,j,o,of,q,hit=0;
 node *np,*nl,*npx;
 vector pc;
 face   *f,*ff;
 eface  *e,*ef;
 pface  *h,*hf;
 vertex *v,*vf;
 double depth;//,r=0.0,g=0.0,b=0.0;
 if((np=which_voxel(p)) == NULL)return hit;
 VECCOPY(p,pc)
 for(;;){/* check for intersection with faces in this voxel */
     if(np->n > 0){ /* some faces */
       for(i=0;i<np->n;i++){
          o= (int)(*(np->oid + i));
          q= (int)(*(np->fid + i));
          v=Object[o].Vbase;
          f=(Object[o].Fbase + q);
          e=(Object[o].Ebase + q);
          h=(Object[o].Pbase + q);
          if(R_intersect_face(p,d,
            (v + f->V[0])->p,(v + f->V[1])->p,(v + f->V[2])->p,
            e->n,&depth,2.0) && (depth < depth_max)){
            hit=1;
            depth_max=depth;
            of=o; ff=f; hf=h; ef=e; vf=v;
          }
       }
     }
   np=track_ray(np,pc,d); /* move on to next voxel */
   if(np == NULL)break;
//   if(hit)break; // under test this should work but it doesn't
 }
 if(hit && trace_reflection_depth < trace_depth_max){
   trace_reflection_depth++;
   GetSurfaceValue(of,ff,hf,vf,ef,p,d,
   colour,(colour+1),(colour+2),
   trace_reflection_depth,
   trace_refraction_depth,
   refractive_flag);
 }
 return hit;
}
#else
short trace_reflection_ray(vector p, vector d, /* d is normalized */
                           double depth_max, double *colour,
                           long trace_reflection_depth,
                           long trace_refraction_depth,
                           long refractive_flag){
 int i,j,o,of,q,hit=0;
 node *np,*nl,*npx;
 vector pc;
 face   *f,*ff;
 eface  *e,*ef;
 pface  *h,*hf;
 vertex *v,*vf;
 double depth,threshold=2.0;
 if((np=which_voxel(p)) == NULL)return hit;
 VECCOPY(p,pc)
 for(;;){/* check for intersection with faces in this voxel */
     if(np->n > 0){ /* some faces */
       for(i=0;i<np->n;i++){
          o= (int)(*(np->oid + i));
          q= (int)(*(np->fid + i));
          v=Object[o].Vbase;
          f=(Object[o].Fbase + q);
          e=(Object[o].Ebase + q);
          h=(Object[o].Pbase + q);
          if(R_intersect_face(p,d,
            (v + f->V[0])->p,(v + f->V[1])->p,(v + f->V[2])->p,
            e->n,&depth,threshold) && (depth < depth_max)){
            hit=1;
            depth_max=depth;
            of=o; ff=f; hf=h; ef=e; vf=v;
          }
       }
     }
   np=track_ray(np,pc,d); /* move on to next voxel */
   if(np == NULL)break;
   if(hit)break; // under test this should work but it doesn't
   threshold=0.001;
 }
 if(hit && trace_reflection_depth < trace_depth_max){
   trace_reflection_depth++;
   GetSurfaceValue(of,ff,hf,vf,ef,p,d,
   colour,(colour+1),(colour+2),
   trace_reflection_depth,
   trace_refraction_depth,
   refractive_flag);
 }
 return hit;
}
#endif

void trace_refraction_ray(vector p, vector d, vector n,
                          int at_o, face *at_f,
                          double depth_max, double *colour,
                          long trace_reflection_depth,
                          long trace_refraction_depth,
                          long refractive_flag){
/*  p is point on glass face of ray hit
    d is incident direction of ray (normalized)
    n is normal of face on which ray is incident
    depth_max is maximum depth to trace to
*/
 int i,j,o,of,q,hit=0;
 node *np,*nl,*npx;
 vector pc,temp,qq;
 face   *f,*ff;
 eface  *e,*ef;
 pface  *h,*hf;
 vertex *v,*vf,*v0,*v1,*v2;
 double depth,ddn;
 vector dp,gp;
 double pdotn,ddotn,mu,qnorm;
 unsigned char red,green,blue;
 if((np=which_voxel(p)) == NULL)return;
 ddn=DOT(d,n);
 if(ddn > 0.0){
   ddn = -ddn; n[0] = -n[0]; n[1] = -n[1]; n[2] = -n[2];
 }
 VECSCALE(ddn,n,temp)
 VECSUB(d,temp,temp)
 if(refractive_flag == 0){
   VECSCALE(1.0/Object[at_o].refractive_index,temp,qq)
 }
 else{
   VECSCALE(Object[at_o].refractive_index,temp,qq)
 }
 refractive_flag ^= 1;  /* each intersection is assumed air/glass */
 qnorm=qq[0]*qq[0]+qq[1]*qq[1]+qq[2]*qq[2];
 if(qnorm > 1.0){
   VECSCALE(sqrt(qnorm-1.0),n,temp)
   VECSUM(qq,temp,d)
   ddn=d[0]*d[0] + d[1]*d[1] + d[2]*d[2];
   ddn=sqrt(ddn);
   VECSCALE((1.0/ddn),d,d)  /* d is normalised new ray direction */
 }
 else{
   VECSCALE(sqrt(1.0-qnorm),n,temp)
   VECSUB(qq,temp,d)
   ddn=d[0]*d[0] + d[1]*d[1] + d[2]*d[2];
   ddn=sqrt(ddn);
   VECSCALE((1.0/ddn),d,d)  /* d is normalised new ray direction */
 }
 /* get the ground or sky value for this ray in case no other Xn's */
 if(R_Nground > 0 && fabs(ddotn=dot(Ground.n,d)) > 1.e-3){
   VECSUB(Ground.p,p,dp)
   pdotn=DOT(Ground.n,dp);
   mu=pdotn/ddotn;
   if(mu > 0.0){
     vecscale(mu,d,gp);
     vecsum(p,gp,gp);
     GetGroundLight(gp,&colour[0],&colour[1],&colour[2]);
     depth_max = mu;
   }
   else goto NEXTBLOCK;
 }
 else{
   NEXTBLOCK:
   colour[0] = (double)Sky.Zenith[0];
   colour[1] = (double)Sky.Zenith[1];
   colour[2] = (double)Sky.Zenith[2];
   if(Sky.type == SKYMAPPED){
     ddotn=d[2];
     if(fabs(ddotn) > 1.e-2){
       vecsub(Sky.map.p,p,dp);
       pdotn=dp[2];
       mu=pdotn/ddotn;
       vecscale(fabs(mu),d,gp);
       vecsum(p,gp,gp);
       MapBrush(&Sky.map,TILE,EOPAQUE,gp,&colour[0],&colour[1],&colour[2]);
     }
   }
 }
 VECCOPY(p,pc)
 v0=Object[at_o].Vbase+at_f->V[0];
 v1=Object[at_o].Vbase+at_f->V[1];
 v2=Object[at_o].Vbase+at_f->V[2];
 for(;;){/* check for intersection with faces in this voxel */
     if(np->n > 0){ /* some faces */
       for(i=0;i<np->n;i++){
         o= (int)(*(np->oid + i));
         q= (int)(*(np->fid + i));
         v=Object[o].Vbase;
         f=(Object[o].Fbase + q);
         if(o == at_o && f == at_f)continue; /* make sure not same face */
         e=(Object[o].Ebase + q);
         h=(Object[o].Pbase + q);
         if(R_intersect_face(p,d,
           (v + f->V[0])->p,(v + f->V[1])->p,(v + f->V[2])->p,
/*   e->n,&depth,1000.0) && (depth < depth_max)){   OLD WAY */
           e->n,&depth,0.0) && (depth < depth_max)){
           if(depth < 1000 && ( /* same face - too close */
              (v+f->V[0]) == v0 || (v+f->V[0]) == v1 || (v+f->V[0]) == v2 ||
              (v+f->V[1]) == v0 || (v+f->V[1]) == v1 || (v+f->V[1]) == v2 ||
              (v+f->V[2]) == v0 || (v+f->V[2]) == v1 || (v+f->V[2]) == v2 ))
              continue;
            hit=1;
            depth_max=depth;
            of=o; ff=f; hf=h; ef=e; vf=v;
         }
       }
     }
//   if(hit)break; // under test
   np=track_ray(np,pc,d); /* move on to next voxel */
   if(np == NULL)break;
 }
 if(hit && trace_refraction_depth < trace_depth_max){
   trace_refraction_depth++;
   GetSurfaceValue(of,ff,hf,vf,ef,p,d,
                   colour,(colour+1),(colour+2),
                   trace_reflection_depth,
                   trace_refraction_depth,
                   refractive_flag);
 }
 return;
}

void trace_transmission_ray(vector p, vector d, vector n,
                          int at_o, face *at_f,
                          double depth_max, double *colour,
                          long trace_reflection_depth,
                          long trace_refraction_depth,
                          long refractive_flag){
/*  p is point on glass face of ray hit
    d is incident direction of ray (normalized)
    n is normal of face on which ray is incident
    depth_max is maximum depth to trace to
*/
 int i,j,o,of,q,hit=0;
 node *np,*nl,*npx;
 vector pc;
 face   *f,*ff;
 eface  *e,*ef;
 pface  *h,*hf;
 vertex *v,*vf,*v0,*v1,*v2;
 double depth;
 if((np=which_voxel(p)) == NULL)return;
 VECCOPY(p,pc)
 v0=Object[at_o].Vbase+at_f->V[0];
 v1=Object[at_o].Vbase+at_f->V[1];
 v2=Object[at_o].Vbase+at_f->V[2];
 for(;;){/* check for intersection with faces in this voxel */
     if(np->n > 0){ /* some faces */
       for(i=0;i<np->n;i++){
         o= (int)(*(np->oid + i));
         q= (int)(*(np->fid + i));
         v=Object[o].Vbase;
         f=(Object[o].Fbase + q);
         if(o == at_o && f == at_f)continue; /* make sure not same face */
         e=(Object[o].Ebase + q);
         h=(Object[o].Pbase + q);
         if(R_intersect_face(p,d,
           (v + f->V[0])->p,(v + f->V[1])->p,(v + f->V[2])->p,
           e->n,&depth,0.0) && (depth < depth_max)){
           if(depth < 1000 && (
              (v+f->V[0]) == v0 || (v+f->V[0]) == v1 || (v+f->V[0]) == v2 ||
              (v+f->V[1]) == v0 || (v+f->V[1]) == v1 || (v+f->V[1]) == v2 ||
              (v+f->V[2]) == v0 || (v+f->V[2]) == v1 || (v+f->V[2]) == v2 ))
              continue;
            hit=1;
            depth_max=depth;
            of=o; ff=f; hf=h; ef=e; vf=v;
         }
       }
     }
   np=track_ray(np,pc,d); /* move on to next voxel */
   if(np == NULL)break;
 }
 if(hit){
   GetSurfaceValue(of,ff,hf,vf,ef,p,d,
                   colour,(colour+1),(colour+2),
                   trace_reflection_depth,
                   trace_refraction_depth,
                   refractive_flag);
 }
 return;
}
