/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// File render.c

#define MODULE_RENDER

#include "render.h"

#ifdef DOS386
extern short mirror_ground;
#if __NT__
extern init_structures(void);
extern HINSTANCE hThisInstance;
#endif
extern void SetLightShadow(long i);
extern void Free_Mirror_On_Ground(void);
extern void Free_Lights_Shadows(void);
extern void Mirror_In_Ground(void);
#endif

static void ViewTr(vector, vector);
static void BrushTr(double [4][4], vector, vector, vector,
                    vector, vector, vector, vector);
static int TransformToView(long, long *, long, long);
static int RenderMirrorSurfaces(long *failure, long frame, long firstframe);
static short R_Transform(long O, long mid);
static void render_banner(void);
static void NormUpdate(vertex *v, face *f, eface *e, long O, long mid);
static void GetMirrorPlane(long i, vector p, vector n);
static int PsameSide(vector pt, vector p, vector n);
static long RenderHiddenLine(void);

extern char *preview_DLL_parameters,*preview_DLL;

FILE *CF=NULL;
long renderer_NT_abort_flag=0;
short End_of_animation;

static char ambient_flag=0,outputfile[256];

char *short_form(char *filename){
 char *sps,*spc;
 sps=spc=filename;
 while((spc=strchr(spc,'\\')) != NULL){spc++; sps=spc;}
 return sps;
}

short R_terminator(void){
 char c;
#if __NT__
 if(renderer_NT_abort_flag && ConfirmNTabort())return FAIL;
 renderer_NT_abort_flag=0;
 return NO;
#else
 if(kbhit()){
   c=getch();
   if(c == 27){
     if(renderer_silent)return FAIL;
     for(c=0;c<30;c++)X_message("\n");
     X_message("Do you want to abort Rendering Y/N ? >> "); fflush(stdout);
     c=getche();
     if(c == 'Y' || c == 'y'){
       putchar(13); for(c=0;c<60;c++)putchar(' '); putchar(13);
       X_message("Are you absolutely sure   !!!  Y/N ? >> "); fflush(stdout);
       c=getche();
       if(c == 'Y' || c == 'y'){
         putchar(13); for(c=0;c<60;c++)putchar(' '); putchar(13);
         return FAIL;
       }
     }
     putchar(13); for(c=0;c<60;c++)putchar(' '); putchar(13);
     X_message("Continuing ...."); fflush(stdout);
   }
 }
 return NO;
#endif
}

static void render_banner(void){
 X_message(VERSION);
 X_message(" for Extended DOS");
 X_message(" ( %s %s )",__DATE__,__TIME__);
 X_message("\n");
 return;
}

int render(int argc, char **argv){
 char  scriptfile[256];
 char  c,temp_str[80],format[64],temp_message[128];
 long  i,j,frame,subframe,command_id,firstframe,lastframe,
       failure=NO,x_xmax,y_ymax,nAnimStart;
 short flash_count,truevision=0,targa_compress=1;
 X_message=X__local_print;
 WindowedRenderInfo("Scene:");
 debug=NULL;
 renderer_silent=0;/* if printing output, vector printf() function */
 preview=0; firstframe=1; lastframe=65536;
 anti_alias=NONE; anti_alias_edges=NO;
 dither_on =NO; remove_hidden_faces=NO; hidden_line=NO; hidden_callback=NO;
 instant_callback=NO; shadow_buffer_size=300;
 FAST=NO; OUTPUT_FILE_TYPE=GIF_FILE; XMAX=320; YMAX=200;
 ColourResolution=BITS8; End_of_animation=0; colour0_flag=0;
 screenbufferON=NO; FullScreenBuffer=NULL; fszBufferON=NO; fszBuffer=NULL;
 objectIDbufferON=NO; ObjectIdBuffer=NULL;
 alfachannelON=NO; alfagroundON=NO; alfaseparate=NO;
 ntsc_colours=NO; full_mirror=FALSE;
 MotionBlurBuffer=NULL; mblurON=0; jpeg_quality=75;
 trace_shadows=NO; trace_reflections=NO; trace_refractions=NO;
 trace_optimise=YES; trace_optimise_depth=15;
 trace_optimise_threshold=20; trace_optimise_size=4;
 tzeros=3; wireframe_thickness=1;
 bWire=FALSE; bHideWire=FALSE; iLace=FALSE;
 if(argc < 3){
  render_banner();
  Render_Message(40,0,NULL);
/*
  printf("V=%ld F=%ld eF=%ld pF=%ld O=%ld M=%ld\n",sizeof(vertex),sizeof(face),
  sizeof(eface),sizeof(pface),sizeof(object),sizeof(brush));
*/
  goto EXITPOINT;
 }
 else if(argc > 3){
   i=3;
   while(i < argc){
     if     (strcmp(argv[i],"-16" ) == 0){
       ColourResolution=BITS16;
     }
     else if(strcmp(argv[i],"-ntc") == 0)targa_compress=0;
     else if(strcmp(argv[i],"-t16") == 0){
       ColourResolution=BITS24; OUTPUT_FILE_TYPE=T16_FILE;
       truevision=0;
     }
     else if(strcmp(argv[i],"-t32") == 0){
       ColourResolution=BITS24; OUTPUT_FILE_TYPE=T24_FILE;
       truevision=0;
     }
     else if(strcmp(argv[i],"-T32") == 0){
       ColourResolution=BITS24; OUTPUT_FILE_TYPE=T24_FILE;
       truevision=1;
     }
     else if(strcmp(argv[i],"-f24") == 0){
       ColourResolution=BITS24; OUTPUT_FILE_TYPE=F24_FILE;
     }
     else if(strcmp(argv[i],"-b24") == 0){
       ColourResolution=BITS24; OUTPUT_FILE_TYPE=B24_FILE;
     }
     else if(strcmp(argv[i],"-p24") == 0){
       ColourResolution=BITS24; OUTPUT_FILE_TYPE=PNG_FILE;
     }
     else if(strcmp(argv[i],"-j24") == 0){
       ColourResolution=BITS24; OUTPUT_FILE_TYPE=J24_FILE;
       i++; jpeg_quality = atol(argv[i]);
     }
     else if(strcmp(argv[i],"-flic") == 0){
       ColourResolution=BITS16; OUTPUT_FILE_TYPE=FLC_FILE;
     }
     else if(strcmp(argv[i],"-lce") == 0)iLace=TRUE;
     else if(strcmp(argv[i],"-cbk") == 0)instant_callback=YES;
     else if(strcmp(argv[i],"-h"   ) == 0){XMAX=640; YMAX=480;}
     else if(strcmp(argv[i],"-x"   ) == 0){i++; XMAX=atoi(argv[i]); }
     else if(strcmp(argv[i],"-y"   ) == 0){i++; YMAX=atoi(argv[i]); }
     else if(strcmp(argv[i],"-f"   ) == 0)FAST=YES;
     else if(strcmp(argv[i],"-ff"  ) == 0){i++; firstframe=atoi(argv[i]); }
     else if(strcmp(argv[i],"-lf"  ) == 0){i++; lastframe=atoi(argv[i]);  }
     else if(strcmp(argv[i],"-ae"  ) == 0)anti_alias_edges=YES;
     else if(strcmp(argv[i],"-a"   ) == 0)anti_alias=LOW;
     else if(strcmp(argv[i],"-aa"  ) == 0)anti_alias=MEDIUM;
     else if(strcmp(argv[i],"-aaa" ) == 0)anti_alias=HIGH;
     else if(strcmp(argv[i],"-aaaa") == 0)anti_alias=ULTRA;
     else if(strcmp(argv[i],"-z"   ) == 0)dither_on =FS_DITHER;
     else if(strcmp(argv[i],"-zz"  ) == 0){
       i++; dither_factor=atoi(argv[i]); dither_factor /= 2;
       if(dither_factor > 255)dither_factor=255;
      dither_on=RANDOM_DITHER;
     }
     else if(strcmp(argv[i],"-r"   ) == 0){
//       this is not used now and will require extra byte in face structure
//       remove_hidden_faces=YES;
     }
     else if(strcmp(argv[i],"-w"   ) == 0){
       hidden_line=YES;
       i++; hidden_line_angle = atol(argv[i]); /* angle */
     }
     else if(strcmp(argv[i],"-ac"  ) == 0){
       screenbufferON=YES;
       alfachannelON=YES;
     }
     else if(strcmp(argv[i],"-acg" ) == 0){
       alfagroundON=YES;
     }
     else if(strcmp(argv[i],"-acx" ) == 0){
       alfaseparate=YES;
     }
     else if(strcmp(argv[i],"-zb"  ) == 0)fszBufferON=YES;
     else if(strcmp(argv[i],"-al"  ) == 0){
       i++; ambient_light= ((double)atof(argv[i]))/100.0; /* % light */
       ambient_flag=1; /* override script ambients */
     }
     else if(strcmp(argv[i],"-sd"  ) == 0){
       i++; shadow_density = 1.0-((double)atof(argv[i])/100.0); /* %age */
       glass_shadow_density = 1.0-0.5*((double)atof(argv[i])/100.0);
     }
     else if(strcmp(argv[i],"-sc"  ) == 0){
       i++; smoothing_angle = cos((double)atof(argv[i])*PI/180.0); /* cos angle */
     }
     else if(strcmp(argv[i],"-c0"  ) == 0){ // not used now
       i++;             /* colour zero option */
       if     (strcmp(argv[i],"b") == 0){
//         colour0_flag=1; colour0_red=0; colour0_green=0; colour0_blue=0;
       }
       else if(strcmp(argv[i],"w") == 0){
//         colour0_flag=2; colour0_red=255; colour0_green=255; colour0_blue=255;
       }
       else  colour0_flag=0;
     }
     else if(strcmp(argv[i],"-sb"  ) == 0){
       i++; shadow_buffer_size=atoi(argv[i]);
       if(shadow_buffer_size < 0)shadow_buffer_size=0;
       if(shadow_buffer_size > 1024)shadow_buffer_size=1024;
     }
     else if(strcmp(argv[i],"-tz"  ) == 0){
       i++; tzeros = atol(argv[i]); /* trailing zeros */
     }
#if TRACER
     else if(strcmp(argv[i],"-trs"  ) == 0)trace_shadows=YES;
     else if(strcmp(argv[i],"-trr"  ) == 0)trace_reflections=YES;
     else if(strcmp(argv[i],"-trg"  ) == 0)trace_refractions=YES;
     else if(strcmp(argv[i],"-to"   ) == 0)trace_optimise=NO;
     else if(strcmp(argv[i],"-tod"  ) == 0){
       i++; trace_optimise_depth = atol(argv[i]); /* depth of tree */
     }
     else if(strcmp(argv[i],"-tot"  ) == 0){
       i++; trace_optimise_threshold = atol(argv[i]); /* face in voxel */
     }
     else if(strcmp(argv[i],"-tos"  ) == 0){
       i++; trace_optimise_size = atol(argv[i]); /* sizeo of voxel */
     }
#endif
     else if(strcmp(argv[i],"-q"   ) == 0){
       renderer_silent=1;
     }
     else if(strcmp(argv[i],"-prv" ) == 0){
       preview=1;
     }
     else if(strcmp(argv[i],"-wfm" ) == 0){
       hidden_callback=1; /* use wireframe to call back Ani */
       hidden_line=YES; hidden_line_angle = 1; /* 60; */
       ColourResolution=BITS16; OUTPUT_FILE_TYPE=FLC_FILE;
     }
     else if(strcmp(argv[i],"-ntsc") == 0){
       ntsc_colours=YES;
     }
     else if(strcmp(argv[i],"-blr") == 0){
       i++; mblurON = (int)atoi(argv[i]);
     }
     else if(strcmp(argv[i],"-fm") == 0){
       full_mirror=TRUE;
     }
     else if(strcmp(argv[i],"-ob") == 0){
       objectIDbufferON=YES;
     }
     i++;
   }
 }
#if !__NT__
 if(!renderer_silent)render_banner();
#endif
 ResolutionX=XMAX; ResolutionY=YMAX;
 aaXstep=1; aaYstep=1;
 if(hidden_line){
   anti_alias_edges=NO;
   anti_alias=NONE;
   dither_on=NO;
   trace_shadows=NO;
   trace_reflections=NO;
   trace_refractions=NO;
 }
 if(trace_refractions)UpdateScanBuffers=UpdateNoGlassScanBuffers;
 else                 UpdateScanBuffers=UpdateGlassScanBuffers;
 if(trace_reflections || trace_refractions || trace_shadows)
      GetPixelValue=GetTracePixelValue;
 else GetPixelValue=GetPhongPixelValue;
 if(anti_alias_edges)full_mirror=FALSE;
 if(anti_alias == LOW)   {XMAX *= 2; aaXstep=2; aaYstep=1;}
 if(anti_alias == MEDIUM){XMAX *= 2; YMAX *= 2; aaXstep=2; aaYstep=2;}
 if(anti_alias == HIGH)  {XMAX *= 3; YMAX *= 2; aaXstep=3; aaYstep=2;}
 if(anti_alias == ULTRA) {XMAX *= 3; YMAX *= 3; aaXstep=3; aaYstep=3;}
 if(ColourResolution == BITS24 && OUTPUT_FILE_TYPE != T16_FILE)dither_on=NO;
 if(OUTPUT_FILE_TYPE == FLC_FILE){
   screenbufferON=NO; fszBufferON=NO;
 }
#if __DEMO__
 if(ResolutionX > 320 || ResolutionY > 240 || anti_alias > MEDIUM ||
   ColourResolution == BITS24){
   if((hidden_line || (instant_callback && renderer_silent))
       && ResolutionX < 800)goto SKIPIT1;
   Render_Message(19,0,NULL);
   failure=FAIL;
   goto EXITPOINT;
   SKIPIT1:;
 }
#endif
 frame=0;
 command_id=0;
 if(firstframe < 1)firstframe=1;
 if(firstframe > lastframe)firstframe=lastframe;

 //debug=fopen("f:\debug.rnd","w");

 flash_count=0; wave_phase=0.0;
 if(Initialize_noise2() == FAIL){
   Render_Message(13,0,NULL); goto EXITPOINT;
 }
 init_noise();
 init_spec();
#if __NT__
 init_structures();
#endif
 setup_render_discarded_list();
 if(SetUpDitherScanLines() == FAIL){
   Render_Message(13,0,NULL); goto EXITPOINT;
 }
 if(OUTPUT_FILE_TYPE != FLC_FILE){
   for (i=0;i<nCOLOUR;i++)colorstats0[i] = 0; colorstats0[0]=65536*1024L;
 }
 for(i=0;i<XMAX;i++)DummyShadow[i]=0;
 sprintf(scriptfile,"%s.scr",argv[1]);
 if((CF=fopen((char *)strlwr(scriptfile),"r")) == NULL){
   Render_Message(41,1,NULL);
   goto EXITPOINT;
 }
 fscanf(CF,"%s",temp_str);
 if(temp_str[0] == 'R'){ /* get the number of frames in the script */
   fscanf(CF,"%s",temp_str);
   nScriptframes=atoi(temp_str);
   fscanf(CF,"%s",temp_str);
   nAnimStart=atoi(temp_str)-1;
 }
 else{
   fscanf(CF,"%s",temp_str);
   nScriptframes=9999;
   nAnimStart=0;
 }
 NEXTFRAME:
 frame++;
 subframe=0;
 NEXTSUBFRAME:
 subframe++;
 if(frame < firstframe){
   char tt[256];
   LOOP:
   if(fscanf(CF,"%s",tt) == EOF)goto EXITPOINT;
   if(tt[0] != 'f')goto LOOP;
   wave_phase += 0.1;          /* move onto next wave phase */
   if(wave_phase > 1.99999)wave_phase = 0.0;
   goto NEXTFRAME;
 }
#if __NT__
 if(!renderer_silent)UpdateNTFrame(frame,command_id);
#endif
 if(frame > lastframe){ End_of_animation=1; goto EXITPOINT;}
 sprintf(temp_message,"%ld",frame);
 Render_Message(42,0,temp_message);
 RenderCurrentFrame=frame;
 R_Nground=0; ObjectCount=0; Object=NULL; Ncsg=0; Csg=NULL;
 Nlights=0; Lights=NULL; nImageProcesses=0; ImageProcesses=NULL;
 VGA_screen=NULL; number_of_shadows=0; Nsky=0;
 Nshader=0; Shader=NULL;
 if((command_id=GetCommand(firstframe,frame,nAnimStart,argv[2])) < 0)
   goto EXITPOINT;
 Render_Message(44,0,NULL);
 Xcentre=(XMAX+XMIN)/2;
 Ycentre=(YMAX+YMIN)/2;
 SetPerspectiveView(); /* build view transformation TRR */
 in_mirror_phase=YES;
 if(RenderMirrorSurfaces(&failure,frame,firstframe) == FAIL)goto FREEPOINT;
 in_mirror_phase=NO;
 if(TransformToView(-1,&failure,frame,firstframe) == FAIL)goto FREEPOINT;
 if(FullScreenBuffer == NULL && OUTPUT_FILE_TYPE != FLC_FILE){
   if(ColourResolution == BITS24){
     char *evt;
     if((evt=X__getenv("EVT")) == NULL){
       if(nScriptframes > 1){
         sprintf(format,"%s%ld%s","%s%0",tzeros,"d.24b");
         sprintf(outputfile,format,argv[2],frame);
       }
       else
         sprintf(outputfile,"%s.24b",argv[2]);
     }
     else{
       if(nScriptframes > 1){
         sprintf(format,"%s%ld%s","%s%s%0",tzeros,"d.24b");
         sprintf(outputfile,format,evt,short_form(argv[2]),frame);
       }
       else
         sprintf(outputfile,"%s%s.24b",evt,short_form(argv[2]));
     }
   }
   else{
     if(nScriptframes > 1){
       sprintf(format,"%s%ld%s","%s%0",tzeros,"d.16b");
       sprintf(outputfile,format,argv[2],frame);
     }
     else
       sprintf(outputfile,"%s.16b",argv[2]);
   }
   if(ColourResolution == BITS8)
      Render_Message(46,0,outputfile);
   else if(ColourResolution == BITS16)
      Render_Message(47,0,outputfile);
   else if(ColourResolution == BITS24)
      Render_Message(46,0,outputfile);
   work=fopen((char *)strlwr(outputfile),"wb");
   if(work == NULL){
     Render_Message(48,1,NULL);
     failure = FAIL;   goto FREEPOINT;
   }
   else{
     if(fwrite((char *)&ResolutionX,sizeof(long),1,work) != 1 ||
        fwrite((char *)&ResolutionY,sizeof(long),1,work) != 1 ||
        fflush(work) != 0){
       Render_Message(49,1,NULL);
       failure = FAIL;   goto FREEPOINT;
     }
   }
 }
 if(hidden_line){ /* do the hidden line */
   if(RenderHiddenLine() == FAIL)failure=FAIL;
 }
 else if(anti_alias_edges == YES && anti_alias > LOW){
   if(FastAntiAliasRenderHide() == FAIL)failure=FAIL;
 }
 else{
   if(RenderHide() == FAIL)failure=FAIL;
 }
 /* if output is in temporary file then close it */
 if(FullScreenBuffer == NULL && OUTPUT_FILE_TYPE != FLC_FILE)fclose(work);
 if(failure != FAIL && nImageProcesses > 0 && hidden_callback == NO)
   failure=ExternalImageProcess(frame);
 if(failure != FAIL && FullScreenBuffer != NULL && MotionBlurBuffer != NULL)
   failure=MotionBlurFrame(frame,subframe,command_id);
 if(failure != FAIL && instant_callback == YES && FullScreenBuffer != NULL)
   AddToRenderImagePreview(0,
                           ResolutionX,ResolutionY,
                           FullScreenBuffer,NULL,NULL,NULL);
 /* if integrating for Motion blur or field rendering don't write yet   */
 if(command_id == 1)goto FREEPOINT;
 /* if doing a separate alpha channel then do it now to same name       */
 /* but as a GIF - so does not work if still gifs are being produced    */
 if(alfaseparate == YES && FullScreenBuffer != NULL && failure != FAIL &&
    ColourResolution != BITS8)
   OneAlfaSeparate(frame,outputfile,argv[2]);
 /* if output file type is 24 bit or 16 bit TARGA then do it now         */
 /* else if full screen buffer and 8 bit create the compressed truncated */
 /* temporary file. If full screen buffer buffer not used, this will     */
 /* already have been done                                               */
 if(OUTPUT_FILE_TYPE == T24_FILE && failure != FAIL)
   OneTargaFrame32(frame,outputfile,argv[2],truevision,targa_compress);
 else if(OUTPUT_FILE_TYPE == T16_FILE && failure != FAIL)
   OneTargaFrame16(frame,outputfile,argv[2],truevision,targa_compress);
 else if(OUTPUT_FILE_TYPE == F24_FILE && failure != FAIL)
   OneTiffFrame(frame,outputfile,argv[2]);
 else if(OUTPUT_FILE_TYPE == B24_FILE && failure != FAIL)
   OneExternalFrame(frame,outputfile,argv[2],B24_FILE,100);
 else if(OUTPUT_FILE_TYPE == J24_FILE && failure != FAIL)
   OneExternalFrame(frame,outputfile,argv[2],J24_FILE,jpeg_quality);
 else if(OUTPUT_FILE_TYPE == PNG_FILE && failure != FAIL)
   OneExternalFrame(frame,outputfile,argv[2],PNG_FILE,100);
 else if(failure != FAIL && FullScreenBuffer != NULL)/* still + animation */
   failure=ScreenTo16bitFile(frame,outputfile,argv[2]);
 FREEPOINT:
 if(ObjectCount > 0){
   for(i=0;i<ObjectCount;i++){
     if(Object[i].Vbase != NULL)X__Free(Object[i].Vbase);
     if(Object[i].Vcopy != NULL)X__Free(Object[i].Vcopy);
     if(Object[i].Fbase != NULL)X__Free(Object[i].Fbase);
     if(Object[i].Ebase != NULL)X__Free(Object[i].Ebase);
     if(Object[i].Pbase != NULL)X__Free(Object[i].Pbase);
     if(Object[i].Tbase != NULL)X__Free(Object[i].Tbase);
     if(Object[i].effect_param_list != NULL)
       X__Free(Object[i].effect_param_list);
     if(Object[i].Abase  != NULL)X__Free(Object[i].Abase);
     if(Object[i].mirror_buffer != NULL)X__Free(Object[i].mirror_buffer);
     if(Object[i].Mbase != NULL){
       for(j=0;j<Object[i].NoMaps;j++)
         FreeRenderBrush((Object[i].Mbase+j));
       X__Free(Object[i].Mbase);
     }
     if(Object[i].Nbase != NULL){
       FreeNurbs(Object[i].NoNurbs,Object[i].Nbase);
       Object[i].NoNurbs=0;
       Object[i].Nbase=NULL;
     }
   }
 }
 if(R_Nground > 0 && Ground.mapped)FreeRenderBrush(&Ground.map);
 if(Nsky > 0 && Sky.mapped)FreeRenderBrush(&Sky.map);
 if(Csg != NULL){ X__Free(Csg); Csg=NULL; }
#ifdef DOS386
 Free_Mirror_On_Ground();
 Free_Lights_Shadows();
#endif
 if(Shader != NULL)FreeExternalTextures();
 if(Object != NULL){X__Free(Object); Object = NULL;}
 if(Lights != NULL){X__Free(Lights); Lights = NULL;}
 if(nImageProcesses > 0 && ImageProcesses != NULL){
   for(i=0;i<nImageProcesses;i++){
     if(ImageProcesses[i].parameters != NULL)
       X__Free(ImageProcesses[i].parameters);
     if(ImageProcesses[i].mparameters != NULL)
       X__Free(ImageProcesses[i].mparameters);
   }
   X__Free(ImageProcesses);
   ImageProcesses=NULL;
   nImageProcesses=0;
 }
 if(VGA_screen != NULL){
   X__Free(VGA_screen);
   VGA_screen=NULL;
 }
#ifdef TRACER
 if(trace_reflections || trace_refractions || trace_shadows)freeBST();
#endif
 if(failure == FAIL){
   remove(outputfile);
   goto EXITPOINT;
 }
 if(command_id == 1){
   goto NEXTSUBFRAME;
 }
 wave_phase += 0.1;          /* move onto next wave phase */
 if(wave_phase > 1.99999)wave_phase = 0.0;
 flash_count++; if(flash_count > 5){flash_count=0; flashing ^= 1;}
#if !__NT__
 Render_Message(82,0,NULL);  /* new line before going on to next frame */
#endif
 goto NEXTFRAME;
 EXITPOINT:
 FreeTextureHashTable();
 if(End_of_animation)Make_up_output(firstframe,frame,argv[2],(BOOL)iLace);
 else{
#if __NT__
  WindowedRenderMessage(WRM_COMPLETE);
#endif
  Render_Message(43,0,NULL);
 }
 FreeDitherScanLines();
 free_render_discarded_list();
 if(CF != NULL)fclose(CF);
 if(debug != NULL)fclose(debug); debug=NULL;
#if !__NT__
 fflush(stdout);
#endif
 return (int)failure;
}

/*  perform geometry transformations   */

static void ViewTr(vector vin, vector vout){
 double x1,y1,z1;
 R_m4by1(trr,vin[0],vin[1],vin[2],&x1,&y1,&z1);
 vout[0]=x1;    vout[1]=y1;    vout[2]=z1;
}

static void BrushTr(double  trpos[4][4], vector  P, vector  X, vector Y,
                    vector  p, vector  x, vector y, vector n){
  R_m4by1(trpos,P[0],P[1],P[2],&p[0],&p[1],&p[2]);
  R_m4by1(trpos,X[0],X[1],X[2],&x[0],&x[1],&x[2]);
  R_m4by1(trpos,Y[0],Y[1],Y[2],&y[0],&y[1],&y[2]);
  vecsub(x,p,x);
  vecsub(y,p,y);
  cross (x,y,n);
  if(normalize(n)){ /* bad axis just set up unit */
    x[0]=1.0; x[1]=0.0; x[2]=0.0;
    y[0]=0.0; y[1]=1.0; y[2]=0.0;
    cross(x,y,n);
  }
/*
 if the map is perpendicular to the view plane then the
  intersection function will fail  x and y in same plane,
  so displace a little bit !! UNIT/10.
  if(fabs(n[1]) < 1.e-6){
    x[0] += UNIT/10;    x[1] += UNIT/10;    x[2] += UNIT/10;
    y[0] += UNIT/10;    y[1] += UNIT/10;    y[2] += UNIT/10;
    cross (x,y,n);
    normalize(n);
  }
 don't do this now the new hitpoint function should render this un-necessary
*/
}

static short R_Transform(long O, long mid){
/* now transform the object to the viewing plane and obtain bounds    */
/* the viewing plane is the XZ plane looking forward along the Y axis */
/* the transformed observer position is the origin (0,0,0)            */
/* the object PHI rotation is about the zaxis +ve anti clockwise      */
   long i,j;
   double temp4,xx,yy,theta_max,trset[4][4];
   vertex   *v;
   face     *f;
   eface    *e;
   Object[O].xmin = BIGP;   Object[O].xmax = BIGN;
   Object[O].ymin = BIGP;   Object[O].ymax = BIGN;
   SetTrPos(Object[O].fi,Object[O].theta,Object[O].alpha
           ,Object[O].im,Object[O].ima
           ,Object[O].sx,Object[O].sy,Object[O].sz
           ,Object[O].Origin,Object[O].Offset,
           Object[O].trpos,trset,
           Object[O].trinv);
   if((j=Object[O].NoMaps) > 0){
     imap *Map;
     for(i=0;i<j;i++){
       Map=(Object[O].Mbase+i);
       BrushTr(Object[O].trpos,
               Map->pin,Map->xin,Map->yin,
               Map->p,Map->x,Map->y,
               Map->n);
       if(Map->bp >= 0){
         theta_max=80 * 0.017453; theta_max=tan(theta_max); /* 60 */
         Map->bumpy = max(1.0,R_length(Map->y));
         Map->bumpy = -((double)Map->bp)*0.01*theta_max
                          /Map->bumpy/256;
         if(Map->map == CYLINDER ||
            Map->map == MAP_SPHERICAL ||
            Map->map == CYLINDER_MOZIAC)
               Map->bumpx = 1.0;
         else  Map->bumpx = max(1.0,R_length(Map->x));
         Map->bumpx = -((double)Map->bp)*0.01*theta_max
                          /Map->bumpx/256;
       }
       Map->pd=((double)Map->pp)*0.01;
       Map->rd=((double)Map->rp)*0.01;
     }
   }
   if(Object[O].NoAxes > 0)for(i=0;i<Object[O].NoAxes;i++){
     BrushTr(Object[O].trpos,
             Object[O].Abase[i].boin,
             Object[O].Abase[i].buin,
             Object[O].Abase[i].bvin,
             Object[O].Abase[i].bo, /* shader axes */
             Object[O].Abase[i].bu,
             Object[O].Abase[i].bv,
             Object[O].Abase[i].bn);
     Object[O].Abase[i].tl=max(1.0,R_length(Object[O].Abase[i].bu));
     Object[O].Abase[i].tv=max(1.0,R_length(Object[O].Abase[i].bv));
     vecscale(1.0/Object[O].Abase[i].tl,
              Object[O].Abase[i].bu,Object[O].Abase[i].bu_norm);
     vecscale(1.0/Object[O].Abase[i].tv,
              Object[O].Abase[i].bv,Object[O].Abase[i].bv_norm);
   }
   v=Object[O].Vbase;
   Object[O].visibility=NO;
   if(Object[O].effect >= 4){
     EffectObjectVertices(0,O); /* locally  */
     for(i=0;i<Object[O].NoVertices;i++){
       R_m4by1(trset, (v+i)->p[0], (v+i)->p[1], (v+i)->p[2]
                    ,&(v+i)->p[0],&(v+i)->p[1],&(v+i)->p[2]);
     }
     EffectObjectVertices(1,O); /* globally */
     for(i=0;i<Object[O].NoVertices;i++){
       R_m4by1(trr,(v+i)->p[0], (v+i)->p[1], (v+i)->p[2]
                  ,&(v+i)->p[0],&(v+i)->p[1],&(v+i)->p[2]);
       if((v+i)->p[1] > 1.000)Object[O].visibility=YES;
     }
   }
   else{
//fprintf(debug,"Object %ld Vertices %ld\n",O,Object[O].NoVertices);
     for(i=0;i<Object[O].NoVertices;i++){
//fprintf(debug,"B %ld   %lf %lf %lf\n",i,(v+i)->p[0],(v+i)->p[1],(v+i)->p[2]);
       R_m4by1(Object[O].trpos, (v+i)->p[0], (v+i)->p[1], (v+i)->p[2]
                            ,&(v+i)->p[0],&(v+i)->p[1],&(v+i)->p[2]);
       if((v+i)->p[1] > 1.000)Object[O].visibility=YES;
//fprintf(debug,"A %ld   %lf %lf %lf\n",i,(v+i)->p[0],(v+i)->p[1],(v+i)->p[2]);
     }
   }
   if(Object[O].visibility == YES){
//fprintf(debug,"Object Visible\n");
     if(mid < 0){
       if(ClipObject(O) == FAIL)return FAIL;
     }
     /* depth sort for effects only */
     if(FAST == NO && Object[O].effect > 0  /* furthest first */
                   && Object[O].effect < 4)DepthSortFaces(O);
   }
   v=Object[O].Vbase;
//fprintf(debug,"Object Screen Coordinates V=%ld\n",Object[O].NoVertices);
   for(i=0;i<Object[O].NoVertices;i++){
    if((v+i)->p[1] >= 1.000){  /* all points on or in front of view plane */
     temp4=(scalex*((v+i)->p[0])/((v+i)->p[1]));
     xx=(double)Xcentre + temp4;
     (Object[O].Vbase+i)->x=xx;
     temp4=(scaley*((v+i)->p[2])/((v+i)->p[1]));
     yy=(double)Ycentre - temp4;
     (Object[O].Vbase+i)->y=yy;
//fprintf(debug,"S %ld   %lf %lf\n",i,xx,yy);
    }
    else{ /* just dummy it any face with this node will not be used */
     (Object[O].Vbase+i)->x=xx=(double)Xcentre;  /* will be scrapped later */
     (Object[O].Vbase+i)->y=yy=(double)Ycentre;
    }
    if(xx < Object[O].xmin) Object[O].xmin = xx;
    if(xx > Object[O].xmax) Object[O].xmax = xx;
    if(yy < Object[O].ymin) Object[O].ymin = yy;
    if(yy > Object[O].ymax) Object[O].ymax = yy;
   }
   v=Object[O].Vbase;     /* NB the object may be invisible but its */
   f=Object[O].Fbase;     /* shadow/reflection may be !!            */
   e=Object[O].Ebase;
   NormUpdate(v,f,e,O,mid);
   if(mid >= 0){
     vector p,n;
     GetMirrorPlane(mid,p,n);
     for(i=0;i<Object[O].NoFaces;i++){
       if(PsameSide((v+f->V[0])->p,p,n) &&             /* move face right */
          PsameSide((v+f->V[1])->p,p,n) &&             /* out of view     */
          PsameSide((v+f->V[2])->p,p,n))f->top=YMAX+1;
       f++;
     }
   }
   return OK;
}

static int PsameSide(vector v, vector p0, vector n){
  vector pv;                            /* after reflection - remove faces  */
  VECSUB(p0,v,pv)                       /* if all vertices are on same side */
  if(DOT(pv,n)*DOT(p0,n) > 0)return 1;  /* as (0,0,0) - put off screen      */
  return 0;
}

static void NormUpdate(vertex *v, face *f, eface *e, long O, long mid){
 long i,buildNo,badflag;
 double c,ftop,fbottom,fleft,fright;
 vector p1,p2,p0;
 if(Object[O].effect == 1)
   buildNo=(long)((double)Object[O].NoFaces*Object[O].effectratio);
 else if(Object[O].effect == 2)
   buildNo=(long)((double)Object[O].NoFaces*(1.0-Object[O].effectratio));
 else if(Object[O].effect == 3) /* build from farthest */
   buildNo=(long)((double)Object[O].NoFaces*Object[O].effectratio);
 for(i=0;i<Object[O].NoFaces;i++){
    p0[0]=(v+f->V[0])->p[0];
    p0[1]=(v+f->V[0])->p[1];
    p0[2]=(v+f->V[0])->p[2];
    p1[0]=(v+f->V[1])->p[0];
    p1[1]=(v+f->V[1])->p[1];
    p1[2]=(v+f->V[1])->p[2];
    p2[0]=(v+f->V[2])->p[0];
    p2[1]=(v+f->V[2])->p[1];
    p2[2]=(v+f->V[2])->p[2];
    vecsub(p2,p0,p2);
    vecsub(p1,p0,p1);
    cross(p1,p2,e->n);
    if(normalize(e->n) == 1)badflag=1; else badflag=0;
    c=dot(p0,e->n);
    e->c=(double)c;
    if(c > 0.0){ /* all normals point to camera */
      vecscale((double)(-1.0),e->n,e->n);
      e->c=(double)c*(-1.0);
    }
    if(FAST == YES)f->fattrib &= 0x7f; /* all faceted */
/* get face screen limits */
    ftop=min((v+(f->V[0]))->y,(v+(f->V[1]))->y);
    ftop=min((v+(f->V[2]))->y,ftop);
    fbottom=max((v+(f->V[0]))->y,(v+(f->V[1]))->y);
    fbottom=max((v+(f->V[2]))->y,fbottom)+1;
    fleft=min((v+(f->V[0]))->x,(v+(f->V[1]))->x);
    fleft=min((v+(f->V[2]))->x,fleft);
    fright=max((v+(f->V[0]))->x,(v+(f->V[1]))->x);
    fright=max((v+(f->V[2]))->x,fright)+1;
    if(ftop < 0.0)f->top= -1;
    else if(ftop > YMAX)f->top=YMAX+1;
    else f->top=(short)ftop;
    if(fbottom < 0.0)f->bottom= -1;
    else if(fbottom > YMAX)f->bottom=YMAX+1;
    else f->bottom=(short)fbottom;
    if(fleft < 0.0)f->left= -1;
    else if(fleft > XMAX)f->left=XMAX+1;
    else f->left=(short)fleft;
    if(fright < 0.0)f->right= -1;
    else if(fright > XMAX)f->right=XMAX+1;
    else f->right=(short)fright;
/* eliminate faces with all vertices behind clipping plane,
   faces that span the clipping plane have been cliped in ClipObject,
   by setting the face top value to  > YMAX this face will then
   fail the first test in Active.  Also eliminate faces to left
   or right of screen by using the same trick */
    if( ( v[ f->V[0] ].p[1] <= 1.000)  /* nodes in face are either behind */
     && ( v[ f->V[1] ].p[1] <= 1.000)  /* or on viewing plane */
     && ( v[ f->V[2] ].p[1] <= 1.000)){
       f->top=YMAX+1;
    }
    if(f->left >= XMAX)      f->top=YMAX+1;
    if(f->right < XMIN)      f->top=YMAX+1;
/* eliminate faces that are perpendicular to the viewing plane */
    if(fabs(e->c) < 1.e-4)   f->top=YMAX+1;
/* and faces with bad normals this put in for DEC aXP and may be wrong */
    if(badflag == 1){
      f->top=YMAX+1; e->n[0]=e->n[1]=e->n[2]=1.0;
    }
    /* for build effects eliminate faces flagged for elimination */
    if((Object[O].effect == 1 || Object[O].effect == 2 ||
        Object[O].effect == 3) && i > buildNo)f->top=YMAX+1;
    f++; e++;
 }
}

static int TransformToView(long mid,long *failure,long frame, long firstframe){
 long i,tFaces=0,tVertices=0;
 double light_transformation[4][4],trset[4][4],trinv[4][4];
 ivector Zero_Vector;
 Zero_Vector[0]=Zero_Vector[1]=Zero_Vector[2]=0;
 if(Nlights > 0)for(i=0;i<Nlights;i++){
   Lights[i].shadow_buffer=NULL;
   Lights[i].object_buffer=NULL;
   Lights[i].face_buffer=NULL;
   Lights[i].atmospheric_shadow_buffer=NULL;
   if(Lights[i].type == SPOTLIGHT){
     SetTrPos(Lights[i].dPhi,0.0,Lights[i].dAlpha,
              Lights[i].internal_type,Lights[i].internal_size,
              1.0,1.0,1.0,
              Zero_Vector,Lights[i].pin,
              light_transformation,trset,trinv);
     ViewTr(Lights[i].pin,Lights[i].p);
     R_m4by1(light_transformation,
             0.0,(double)UNIT,0.0,
             &Lights[i].d[0],&Lights[i].d[1],&Lights[i].d[2]);
     vecsub(Lights[i].p,Lights[i].d,Lights[i].d);  //d is opposite to light dir
     normalize(Lights[i].d);
     Lights[i].cone2 = min(89.9,Lights[i].cone1/2+Lights[i].falloff);
     Lights[i].cone2 *= PI/180.0;   /* cone + edge */
     Lights[i].cone1 *= PI/360.0;   /* half viewing cone angle */
     Lights[i].dot1 = cos(Lights[i].cone1);  /* for dot product */
     Lights[i].dot2 = cos(Lights[i].cone2);
     Lights[i].falloff = 1.0/(Lights[i].dot1-Lights[i].dot2);
   }
   else ViewTr(Lights[i].pin,Lights[i].p);
   Lights[i].Ire = ((double)Lights[i].color[0])/255.0;
   Lights[i].Igr = ((double)Lights[i].color[1])/255.0;
   Lights[i].Ibl = ((double)Lights[i].color[2])/255.0;
 }
 else{
   if((Lights=(light *)X__Malloc(sizeof(light))) == NULL){
     *failure=FAIL; fclose(work); return FAIL;
   }
   Render_Message(50,0,NULL);
   Nlights=1;   /* camera is at origin after transform */
   Lights[0].type=NORMAL;
   Lights[0].shadow_buffer=NULL;
   Lights[0].atmospheric_shadow_buffer=NULL;
   Lights[0].AnimatorID=0;
   Lights[0].object_buffer=NULL;
   Lights[0].face_buffer=NULL;
   Lights[0].p[0] = 0.0; Lights[0].p[1] = -0.0; Lights[0].p[2] = 0.0;
   Lights[0].d[0] = 0.0; Lights[0].d[1] = -1.0; Lights[0].p[2] = 0.0;
   Lights[0].Ire = 1.0;
   Lights[0].Igr = 1.0;
   Lights[0].Ibl = 1.0;
   Lights[0].dc=FALSE;
   Lights[0].dc_c=1.0;
   Lights[0].dc_l=0.0;
   Lights[0].dc_q=0.0;
 }
 if(R_Nground > 0){
  ViewTr(Ground.pin,Ground.p);
  ViewTr(Ground.xin,Ground.x);
  ViewTr(Ground.yin,Ground.y);
  vecsub(Ground.x,Ground.p,Ground.dx);
  vecsub(Ground.y,Ground.p,Ground.dy);
  cross(Ground.dx,Ground.dy,Ground.n);
  normalize(Ground.n);
  Ground.lengthx=R_length(Ground.dx);
  Ground.lengthy=R_length(Ground.dy);
  Ground.bump=1.0/max(1.0,Ground.lengthx);
  Ground.range=Ground.lengthy/Ground.lengthx/2.0;
  Ground.range *= Ground.range;
  if(ambient_flag == 0)Ground.ambient_light=ground_ambient_light;
  veccopy(Ground.dy,Ground.dyy);
  normalize(Ground.dyy);
  vecscale(Ground.lengthx,Ground.dyy,Ground.dyy);
  if(Ground.texture > 0 && Ground.texture < 4)Ground.shiny = 1;
  else                                        Ground.shiny = 0;
  if(frame == firstframe)init_ripples();
  if(Ground.mapped){
    ViewTr(Ground.map.pin,Ground.map.p);
    ViewTr(Ground.map.xin,Ground.map.x);
    ViewTr(Ground.map.yin,Ground.map.y);
    vecsub(Ground.map.x,Ground.map.p,Ground.map.x);
    vecsub(Ground.map.y,Ground.map.p,Ground.map.y);
    cross (Ground.map.x,Ground.map.y,Ground.map.n);
    normalize(Ground.map.n);
  }
 }
 if(ambient_flag == 0)ambient_light=object_ambient_light;
 if(Sky.type == SKYMAPPED){  /* set up cylindrical map */
   if(Sky.mapped){
     ViewTr(Sky.map.pin,Sky.map.p);
     ViewTr(Sky.map.xin,Sky.map.x);
     ViewTr(Sky.map.yin,Sky.map.y);
     vecsub(Sky.map.x,Sky.map.p,Sky.map.x);
     vecsub(Sky.map.y,Sky.map.p,Sky.map.y);
     cross (Sky.map.x,Sky.map.y,Sky.map.n);
     normalize(Sky.map.n);
   }
 }
 if(Ncsg > 0)TransformCSG();
 if(ObjectCount > 0)for(i=0;i<ObjectCount;i++){
   if(Object[i].NoVertices == 0 || Object[i].NoFaces == 0)continue;
   tVertices += Object[i].NoVertices;
   tFaces    += Object[i].NoFaces;
   if(i == mid){ /* reflection of this object skip it */
     Object[i].visibility=NO;
     continue;
   }
   Object[i].ambient_light=ambient_light;
   if(preview){  /* use the axis scaling */
     int j;
     vector a,b;
     double xscale,yscale;
     if     (preview_xscale > 128)xscale = 1.0+
        ((double)preview_xscale-128.0)/128.0*10;
     else if(preview_xscale < 128)
       xscale = 0.1+0.00703*(double)preview_xscale;
     else xscale=1.0;
     if     (preview_yscale > 128)yscale = 1.0+
        ((double)preview_yscale-128.0)/128.0*10;
     else if(preview_yscale < 128)
       yscale = 0.1+0.00703*(double)preview_yscale;
     else yscale=1.0;
     vecsub(Object[i].Abase[0].buin,Object[i].Abase[0].boin,a);
     vecsub(Object[i].Abase[0].bvin,Object[i].Abase[0].boin,b);
     vecscale(xscale,a,a); vecscale(yscale,b,b);
     vecsum(Object[i].Abase[0].boin,a,Object[i].Abase[0].buin);
     vecsum(Object[i].Abase[0].boin,b,Object[i].Abase[0].bvin);
     if(!preview_background && Nsky){
       for(j=0;j<3;j++){
         Sky.Horizon[j]=Sky.Zenith[j]=Sky.Hblend[j]=Sky.Zblend[j]=0;
       }
     }
     if(preview_texture > 15 && preview_DLL != NULL &&
        LoadExternalTexture(preview_DLL,
        preview_DLL_parameters) == OK){
       Object[i].shader[0]=Nshader-1;
     }
   }
   if(R_Transform(i,mid) == FAIL){
     *failure=FAIL;
     return FAIL;
   }
 }
 for(i=0;i<Nlights;i++)if(Lights[i].type == SPOTLIGHT){
   SetLightShadow(i);
 }
#ifdef TRACER
 if(trace_reflections || trace_refractions || trace_shadows){
   if(buildBST() == FAIL){
     Render_Message(45,0,NULL);
     *failure=FAIL;
     return FAIL;
   }
 }
#endif
 if(!(trace_reflections || trace_refractions || trace_shadows)){
   char tstr[256],text_string[256];
   LoadString(hThisInstance,IDX_MISC_STATUS1,
              text_string,sizeof(text_string));
   sprintf(tstr,text_string,
           tFaces,tVertices,ObjectCount);
   WindowedRenderInfo(tstr);
 }
 return OK;
}

static void GetMirrorPlane(long i, vector p, vector n){
 double t[4][4],tt[4][4],ttt[4][4];
 long j;
 vector u,v;
 j=Object[i].mirrorID;
 SetTrPos(Object[i].fi,Object[i].theta,Object[i].alpha /* move the axis */
         ,Object[i].im,Object[i].ima                   /* of reflection */
         ,Object[i].sx,Object[i].sy,Object[i].sz       /* to reflection */
         ,Object[i].Origin,Object[i].Offset,t,tt,ttt); /* plane and get */
 BrushTr(t,                                            /* its transform */
         Object[i].Abase[j].boin,
         Object[i].Abase[j].buin,
         Object[i].Abase[j].bvin,
         p,u,v,n);
 return;
}

static int RenderMirrorSurfaces(long *failure, long frame, long firstframe){
 double trr_stor[4][4],t[4][4],tt[4][4],ttt[4][4];
 vector p,u,v,n;
 long i,j,k,s;
 long do_it=NO;
 unsigned char *pP;
 for(i=0;i<4;i++)for(j=0;j<4;j++)trr_stor[i][j]=trr[i][j];
 if(hidden_line == YES)return OK; /* not needed for line drawing  */
 if(preview) return OK;           /* not needed in preview mode   */
 if(ObjectCount == 0)return OK;   /* no objects - need I say more */
 for(i=0;i<ObjectCount;i++)if(Object[i].mirrorID >= 0)do_it=YES;
 if(do_it == NO)return OK;        /* no mirror surfaces           */
 for(i=0;i<ObjectCount;i++){      /* allocate RAM for verts+image */
   if(Object[i].NoVertices == 0)continue;
   if((Object[i].Vcopy=(vector *)X__Malloc(Object[i].NoVertices
       *sizeof(vector))) == NULL){
     *failure=FAIL;
     return FAIL;
   }
   for(j=0;j<Object[i].NoVertices;j++){
     VECCOPY(Object[i].Vbase[j].p,Object[i].Vcopy[j])
   }
   s=ResolutionX*ResolutionY*3*sizeof(unsigned char);
   if((Object[i].mirror_buffer=(unsigned char *)X__Malloc(s)) == NULL){
     *failure=FAIL;
     return FAIL;
   }
 }
 for(i=0;i<ObjectCount;i++){ /* now render each mirror in turn */
   if(Object[i].mirrorID >= 0){
     for(k=0;k<4;k++)for(j=0;j<4;j++)trr[k][j]=trr_stor[k][j];
     GetMirrorPlane(i,p,n);
     SetTrPos(Object[i].fi,Object[i].theta,Object[i].alpha /* move the axis */
             ,Object[i].im,Object[i].ima                   /* of reflection */
             ,Object[i].sx,Object[i].sy,Object[i].sz       /* to reflection */
             ,Object[i].Origin,Object[i].Offset,t,tt,ttt); /* plane and get */
     R_reflect(t,p,n); /* get reflection transformation */
     R_m4by4(t,trr_stor,trr); /* apply it               */
     /* transform everthing into mirror positions - ignore object i */
     if(TransformToView(i,failure,frame,firstframe) == FAIL)return FAIL;
     if(anti_alias_edges == YES && anti_alias > LOW){
       if(FastAntiAliasRenderHide() == FAIL){
         *failure=FAIL;
         return FAIL;
       }
     }
     else{
       if(RenderHide() == FAIL){
         *failure=FAIL;
         return FAIL;
       }
     }
     /* now copy screenbuffer to object mirror buffer */
     s=ResolutionX*ResolutionY;
     pP=Object[i].mirror_buffer;
     for(j=0;j<s;j++){
       *pP=FullScreenBuffer[j].R;  pP++;
       *pP=FullScreenBuffer[j].G;  pP++;
       *pP=FullScreenBuffer[j].B;  pP++;
     }
     for(k=0;k<ObjectCount;k++){    /* put back all the vertices */
       if(Object[k].NoVertices > 0)
       for(j=0;j<Object[k].NoVertices;j++){
         VECCOPY(Object[k].Vcopy[j],Object[k].Vbase[j].p)
       }
     }
   }
 }
 for(i=0;i<4;i++)for(j=0;j<4;j++)trr[i][j]=trr_stor[i][j];
 return OK;
}

/*  do the scan line rendering here */

/* functions to build the phong smoothing vectors and do the scanline */
/* algorithm. The geometry calculations are doen in file   GEO.C      */
/* and the surface properties are calculated in TEXTURE.C + TEXTUR1.C */
/* Also inclued the "CORE" of the scanline algorithm                  */

typedef struct tagPHONG1 {
  int fid,vid;
} PHONG1;

typedef struct PHONG {
  int n;
#if __NT__
  PHONG1 fv[8];   /* 10 seems to be too big */
#else
  PHONG1 *fv;
#endif
} phong;

void MakeVertexNormals(void){   /* this is a horrible thing that asembles */
                                /* the PHONG vectors 3 for each FACE      */
 vertex   *v;           /* umpteen different procedures have been */
 face     *f;           /* tried - this gives the best results    */
 eface    *e;           /* the diff. only occurs when faces are   */
 pface    *p;           /* perpend.r to the line of sight.        */
 long i,j,k,O,vi,ni,fj,vj,fk;   /* All other faces are turned to face the */
 double c;                      /* camera there is no problem there.       */
 vector nn;
 phong *ph;
 for(O=0; O<ObjectCount; O++){
   if(Object[O].NoVertices == 0 || Object[O].NoFaces == 0)continue;
   v=Object[O].Vbase;
   f=Object[O].Fbase;
   e=Object[O].Ebase;
   p=Object[O].Pbase;
   if(preview){  /* used for Shader preview only to allow shader to change */
     for(i=0;i<Object[O].NoFaces;i++){
       f[i].fattrib=0;
       if(preview_smoothing)f[i].fattrib |= 0x80;
       else                 f[i].fattrib &= 0x70;
       if(preview_brflag)f[i].fattrib |= 0x20;
       else              f[i].fattrib &= 0xdf;
       if(preview_texture == 17){
         f[i].fattrib |= 0x10;
         f[i].fshader=1;
       }
       else f[i].fshader = preview_texture;
       f[i].fmap  = 0;
       f[i].matsp = preview_gloss;
       f[i].mattr = preview_tr << 4;
       f[i].matre = preview_refl << 4;
       f[i].fparam= preview_pr;
       for(j=0;j<3;j++){
         f[i].color[j]=preview_fcolour[j];
         f[i].texco[j]=preview_tcolour[j];
       }
     }
   }
   for(i=0;i<Object[O].NoFaces;i++){
     for(j=0;j<3;j++)VECCOPY(e[i].n,p[i].n[j]);
   }
   if((ph = (phong *)X__Malloc(Object[O].NoVertices*sizeof(phong))) == NULL){
     Render_Message(11,0,NULL);
     return;
   }
   for(i=0;i<Object[O].NoVertices;i++){
     ph[i].n=0;
#if !__NT__
     ph[i].fv=NULL;
#endif
   }
   for(i=0;i<Object[O].NoFaces;i++){
     if(f[i].fattrib >> 7 != 0){ /* only smooth 'smoothed' faces */
       for(j=0;j<3;j++){  /* make list of faces adjacent to vertex */
         vi=f[i].V[j];
         ni=ph[vi].n;
#if __NT__
         if(ni < 8){ /* for windows just use ? (10 too big) so no reallocation needed */
           ph[vi].fv[ni].fid=i;
           ph[vi].fv[ni].vid=j;
           ph[vi].n++;
         }
#else
         if(ph[vi].fv == NULL){
           if((ph[vi].fv=(PHONG1 *)X__Malloc(sizeof(PHONG1))) == NULL){
             Render_Message(11,0,NULL); return;
           }
         }
         else if((ph[vi].fv=(PHONG1 *)X__Realloc(ph[vi].fv,sizeof(PHONG1)*(ni+1)))==NULL){
           Render_Message(11,0,NULL); return;
         }
         ph[vi].fv[ni].fid=i;
         ph[vi].fv[ni].vid=j;
         ph[vi].n++;
#endif
       }
     }
   }  /* now we have a list of faces adjacent to each vertex */

   if(remove_hidden_faces){                /* the bad faces will be removed */
#if 0
     for(i=0;i<Object[O].NoVertices;i++){  /* later in a full scanline pass */
       if((ni=ph[i].n) > 0){
         for(j=0;j<ni;j++){
           fj=ph[i].fv[j].fid; vj=ph[i].fv[j].vid;
           for(k=0;k<ni;k++)if(k != j){
             fk=ph[i].fv[k].fid;
             if(f[fk].t == 1){ /* visible */
               vecsum(p[fj].n[vj],e[fk].n,p[fj].n[vj]);
             }
           }
         }
       }
     }
#endif
   }
   else{                                  /* This is the most common case */
     for(i=0;i<Object[O].NoVertices;i++){
       if((ni=ph[i].n) > 0){
         for(j=0;j<ni;j++){
           fj=ph[i].fv[j].fid; vj=ph[i].fv[j].vid;
           for(k=0;k<ni;k++)if(k != j){
             fk=ph[i].fv[k].fid;
             if((c=dot(e[fj].n,e[fk].n)) > smoothing_angle /*0.01*/){ /* keep same */
               vecsum(p[fj].n[vj],e[fk].n,p[fj].n[vj]);
             }
             else if(c < -smoothing_angle /*-0.01*/){  /* invert before adding */
               nn[0] = -e[fk].n[0]; nn[1] = -e[fk].n[1]; nn[2] = -e[fk].n[2];
               vecsum(p[fj].n[vj],nn,p[fj].n[vj]);
             }
           }
         }
       }
     }
   }
   for(i=0;i<Object[O].NoFaces;i++){ /* normalize vertex vectors */
     for(j=0;j<3;j++)normalize(p[i].n[j]);
// removed with hidden faces     f[i].t=(f[i].brush & 0x1f);
   }
#if !__NT__
   for(i=0;i<Object[O].NoVertices;i++){
     if(ph[i].fv != NULL)X__Free(ph[i].fv);
   }
#endif
   X__Free(ph);
 }
}

#if 0
static void RemoveHiddenFaces(void){ // Not used now
 long i,j,jj,k;
 double d2,xl,xr;
 vertex *v;
 face   *f;
 eface  *e;
 for(i=XMIN;i<XMAX;i++) {
   DL[i]=(double)(i - Xcentre);        DL[i]/=scalex;
   DR[i]=(double)(i+1-Xcentre);        DR[i]/=scalex;
   D1[i]=(double)(i - Xcentre) + 0.5;  D1[i]/=scalex;
 }
 Render_Message(56,0,NULL);  /* that mess up Phong !                     */
 for(k=YMIN;k<YMAX;k++){     /* calculate the faces that are visible     */
   d2=(double)(Ycentre-k);   /* this is like renderering the whole thing */
   d2/=scaley;               /* again without phong                      */
   for(i=XMIN;i<XMAX;i++){
     gZbuffer[i] = -1;
     gZdepth [i] = FARAWAY;
     gZposn  [i] = (double)i+0.5;
     for(j=0;j<NGLASS;j++){
       gZglassd[j][i] = FARAWAY;
       gZglassO[j][i] = -1;
       gZglass [j][i] = -1;
     }
     gZobject[i] = -2;
   }
   for(i=0;i<ObjectCount;i++)
   if((k > Object[i].ymin) && (k < Object[i].ymax)
           &&(Object[i].visibility == YES)){
    v=Object[i].Vbase;
    f=Object[i].Fbase;
    e=Object[i].Ebase;
    for(j=0;j<Object[i].NoFaces;j++)
     if( Active(k,v,f+j,&xl,&xr) ){
       UpdateScanBuffers(xl,xr,j,i,f+j,e+j,d2,
                         gZglass,gZglassO,gZglassd,
                         gZbuffer,gZobject,gZdepth,gZposn);
     }
   }
   for(i=XMIN;i<XMAX;i++){  /* flag the visible faces using  f->t    */
     j=gZobject[i];         /* as flag to indicate a face is visible */
     if(j >= 0){            /* f->t is later used for texture info   */
       jj=gZbuffer[i];
       if(jj >= 0)Object[j].Fbase[jj].t=1;
     }
     j=gZglassO[0][i];       /* glass faces will also be visible      */
     if(j >= 0){
       jj=gZglass[0][i];
       if(jj >= 0)Object[j].Fbase[jj].t=1;
     }
   }
 }
 return;
}
#endif

static long RenderHiddenLine(void){
 unsigned char skyR,skyG,skyB;
 long i,j,jj,k,*Zhold,*Zhobj;
 double d2,xl,xr,h_dot;
 vertex *v;
 face   *f,*fc,*ft,*fl;
 eface  *e,*ec,*et,*el;
 if(R_terminator() == FAIL)return FAIL;
 h_dot=cos((double)hidden_line_angle*PI/180.0);
 if(Nsky == 0 || hidden_callback){skyR=skyG=skyB=0;}
 else{skyR=Sky.Zenith[0]; skyG=Sky.Zenith[1]; skyB=Sky.Zenith[2];}
 Render_Message(54,0,NULL);
 if((Zhold=(long *)X__Malloc(XMAX*sizeof(long))) == NULL)return FAIL;
 if((Zhobj=(long *)X__Malloc(XMAX*sizeof(long))) == NULL){
   Render_Message(9,0,NULL); X__Free(Zhobj); return FAIL;
 }
 for(i=XMIN;i<XMAX;i++){      /* nothing on the old line ! */
   DL[i]=(double)(i - Xcentre);        DL[i]/=scalex;
   DR[i]=(double)(i+1-Xcentre);        DR[i]/=scalex;
   D1[i]=(double)(i - Xcentre) + 0.5;  D1[i]/=scalex;
   Zhobj[i] = -2; Zhold[i] = -1;
 }
 for(k=YMIN;k<YMAX;k++){      /* calculate the faces that are visible */
   d2=(double)(Ycentre-k);    /* for scanline k                       */
   d2/=scaley;
   for(i=XMIN;i<XMAX;i++){
     gZbuffer[i] = -1;
     gZdepth [i] = FARAWAY;
     gZposn  [i] = (double)i+0.5;
     for(j=0;j<NGLASS;j++){
       gZglassd[j][i] = FARAWAY;
       gZglassO[j][i] = -1;
       gZglass [j][i] = -1;
     }
     gZobject[i] = -2;
     gRed[i]=skyR; gGreen[i]=skyG; gBlue[i]=skyB; gAlfa[i]=0;
   }
   for(i=0;i<ObjectCount;i++)                /* each object */
   if((k > Object[i].ymin) && (k < Object[i].ymax)
           &&(Object[i].visibility == YES)){ /* object visible + spans line */
    v=Object[i].Vbase;
    f=Object[i].Fbase;
    e=Object[i].Ebase;
    for(j=0;j<Object[i].NoFaces;j++)         /* each face of object */
     if( Active(k,v,f+j,&xl,&xr) ){
       UpdateScanBuffers(xl,xr,j,i,f+j,e+j,d2,
                         gZglass,gZglassO,gZglassd,
                         gZbuffer,gZobject,gZdepth,gZposn);
     }
   }
   if(fszBuffer != NULL)ScanlineToZbuffer(k,gZdepth);
   if(ObjectIdBuffer != NULL)ScanlineToZobject(k,gZobject);
   if(Ncsg > 0){
     GetCSGintersections(k,d2,gZcsg,gZground,gZbuffer,gZobject,gZdepth);
     for(i=XMIN+1;i<XMAX;i++){
       if(gZobject[i] == -3 && (gZobject[i-1] == -3 || Zhobj[i]  == -3)){
         gZbuffer[i] = gZcsg[i];   /* make Z buffer Csg id */
       }  /* CSG's all round */
     }
   }
   if(alfachannelON == YES)FillAlfaLine(gAlfa,gZglassO,gZobject);
   if(hidden_line_angle == 0){ /* show all faces */
     for(i=XMIN+1;i<XMAX;i++){
       if(gZobject[i] != gZobject[i-1] || gZobject[i] != Zhobj[i] ||
          gZbuffer[i] != gZbuffer[i-1] || gZbuffer[i] != Zhold[i]){
         f=NULL;
         if(gZobject[i] >= 0 && gZbuffer[i] >= 0)
           f=(Object[gZobject[i]].Fbase + gZbuffer[i]);
         else if(gZobject[i-1] >= 0 && gZbuffer[i-1] >= 0)
           f=(Object[gZobject[i-1]].Fbase + gZbuffer[i-1]);
         else if(Zhobj[i] >= 0 && Zhold[i] >= 0)
           f=(Object[Zhobj[i]].Fbase + Zhold[i]);
         if(f == NULL){
           gRed[i]=gGreen[i]=gBlue[i]=255;
         }
         else{
           gRed[i]=f->color[0]; gGreen[i]=f->color[1]; gBlue[i]=f->color[2];
         }
       }
       Zhobj[i]=gZobject[i]; Zhold[i]=gZbuffer[i];  /* update last line */
     }
   }
   else{  /* only faces colour change or angle > threshold */
     for(i=XMIN+1;i<XMAX;i++){
       if(gZobject[i] != gZobject[i-1] || gZobject[i] != Zhobj[i] ||
          gZbuffer[i] != gZbuffer[i-1] || gZbuffer[i] != Zhold[i]){
         ec=et=el=NULL; fc=ft=fl=NULL;
         if(gZobject[i] >= 0 && gZbuffer[i] >= 0){
           ec=(Object[gZobject[i]].Ebase + gZbuffer[i]);
           fc=(Object[gZobject[i]].Fbase + gZbuffer[i]);
         }
         if(gZobject[i-1] >= 0 && gZbuffer[i-1] >= 0){
           el=(Object[gZobject[i-1]].Ebase + gZbuffer[i-1]);
           fl=(Object[gZobject[i-1]].Fbase + gZbuffer[i-1]);
         }
         if(Zhobj[i] >= 0 && Zhold[i] >= 0){
           et=(Object[Zhobj[i]].Ebase + Zhold[i]);
           ft=(Object[Zhobj[i]].Fbase + Zhold[i]);
         }
         if(ec == NULL){
           if(fl != NULL){
             gRed[i]=fl->color[0];
             gGreen[i]=fl->color[1];
             gBlue[i]=fl->color[2];
           }
           else if(ft != NULL){
             gRed[i]=ft->color[0];
             gGreen[i]=ft->color[1];
             gBlue[i]=ft->color[2];
           }
           else { gRed[i]=gGreen[i]=gBlue[i]=255;}
         }
         else if(el == NULL || et == NULL){
           gRed[i]=fc->color[0];
           gGreen[i]=fc->color[1];
           gBlue[i]=fc->color[2];
         }
         else if(fc != NULL){
           if(fc->color[0] != fl->color[0] || fc->color[0] != ft->color[0] ||
              fc->color[1] != fl->color[1] || fc->color[1] != ft->color[1] ||
              fc->color[2] != fl->color[2] || fc->color[2] != ft->color[2] ||
            fabs(DOT(ec->n,el->n)) < h_dot ||
            fabs(DOT(ec->n,et->n)) < h_dot){
             gRed[i]=fc->color[0];
             gGreen[i]=fc->color[1];
             gBlue[i]=fc->color[2];
           }
         }
       }
       Zhobj[i]=gZobject[i]; Zhold[i]=gZbuffer[i];  /* to last line */
     }
   }
   if(hidden_callback){
     AnimatorHiddenLineCallback(k,XMAX,gRed,gGreen,gBlue);
     if(R_terminator() == FAIL){
       X__Free(Zhobj); X__Free(Zhold);
       return FAIL;
     }
   }
   else if(AntiAliasLine(k,anti_alias,
                         gRed,gGreen,gBlue,gAlfa,
                         gL_Red,gL_Green,gL_Blue,gL_Alfa,
                         gLL_Red,gLL_Green,gLL_Blue,gLL_Alfa) == FAIL){
     X__Free(Zhobj); X__Free(Zhold);
     Render_Message(0,0,NULL);
     return FAIL;
   }
 }
 X__Free(Zhobj); X__Free(Zhold);
 return 0;
}

#ifndef _MT

long RenderHide(void){                 /* The core of the hidden line */
 long i,j,k,Wsky;
 double d2,xl,xr,Rsky=0.0;
 vertex *v;
 face   *f;
 eface  *e;
#if __DEMO__
 if(XMAX > 800){
   Render_Message(17,0,NULL);
   return FAIL;
 }
#endif
 if(R_terminator() == FAIL)return FAIL;
 Render_Message(51,0,NULL);
 if(R_Nground > 0 && hidden_line == NO){  /* Make up the ground shadow mask */
   Render_Message(52,0,NULL);
   ShadowGround();
 }
 if(R_terminator() == FAIL)return FAIL;
 for(i=0;i<ObjectCount;i++)               /* Make up the Model shadow mask */
   if(Object[i].visibility == YES && Object[i].shadowed != NO){
     Render_Message(53,0,NULL);
     ShadowObject(i);
     if(R_terminator() == FAIL)return FAIL;
   }
// removed  if(remove_hidden_faces)RemoveHiddenFaces();
 Render_Message(57,0,NULL);
 MakeVertexNormals(); /* calculate Phong smoothing vertex normals */
 if(R_terminator() == FAIL)return FAIL;  /* Check for user cancel */
 if(R_Nground > 0 && Ground.refl > 0.0){
   Render_Message(58,0,NULL);
   Mirror_In_Ground();
 }
 if(in_mirror_phase)Render_Message(99,0,NULL);
 else Render_Message(59,0,NULL); /* HERE IS THE CORE !!!!! */
 for(i=XMIN;i<XMAX;i++) {     /* set scan line variables needed many times */
   DL[i]=(double)(i - Xcentre);        DL[i]/=scalex;
   DR[i]=(double)(i+1-Xcentre);        DR[i]/=scalex;
   D1[i]=(double)(i - Xcentre) + 0.5;  D1[i]/=scalex;
 }
 GetSkyLocation();       /* for environment sky(same X pos for each line */
 for(k=YMIN;k<YMAX;k++){        /* do the scan lines */
   d2=(double)(Ycentre-k);
   d2/=scaley;
   Wsky=PutBackIntersection(k,d2,gAlfa,gZground,  /* get background depth  */
                       gZglass,gZglassO,gZglassd, /* and iniitialise Z's   */
                       gZbuffer,gZobject,gZdepth,gZposn,
                       &Rsky);
   for(i=0;i<ObjectCount;i++)
   if((k > Object[i].ymin) && (k < Object[i].ymax) && !Object[i].wireframe
           &&(Object[i].visibility == YES)){     /* for each object         */
    v=Object[i].Vbase;
    f=Object[i].Fbase;
    e=Object[i].Ebase;
    for(j=0;j<Object[i].NoFaces;j++)             /* for each face           */
     if( Active(k,v,f+j,&xl,&xr) ){              /* anything on this line   */
       UpdateScanBuffers(xl,xr,j,i,f+j,e+j,d2,
                         gZglass,gZglassO,gZglassd,
                         gZbuffer,gZobject,
                         gZdepth,gZposn);        /* record face if in front */
     }
   }
   if(ShadedScanline(
                  k,d2,gZcsg,gRed,gGreen,gBlue,
                  gZground,Wsky,Rsky,
                  gZglass,gZglassO,gZglassd,
                  gZbuffer,gZobject,gZdepth,gZposn) == FAIL)return FAIL;
   if(bWire)RenderWireframeObjects(k,gRed,gGreen,gBlue,
                  gZglass,gZglassO,gZglassd,
                  gZbuffer,gZobject,gZdepth,gZposn);
   if(alfachannelON == YES)FillAlfaLine(gAlfa,gZglassO,gZobject);
   if(AntiAliasLine(k,anti_alias,gRed,gGreen,gBlue,gAlfa,
                         gL_Red,gL_Green,gL_Blue,gL_Alfa,
                         gLL_Red,gLL_Green,gLL_Blue,gLL_Alfa) == FAIL){
     Render_Message(0,0,NULL);                   /* if anti-alias store     */
     return FAIL;                                /* or combine as necessary */
   }
   if(fszBuffer != NULL)ScanlineToZbuffer(k,gZdepth);
   if(ObjectIdBuffer != NULL)ScanlineToZobject(k,gZobject);
#if __NT__
  if(!renderer_silent){
    if     (anti_alias == ULTRA)UpdateNTTrace((k+1)/3);
    else if(anti_alias == MEDIUM ||
            anti_alias == HIGH )UpdateNTTrace((k+1)/2);
    else                        UpdateNTTrace(k+1);
  }
#else
   if((k%60) == 0){
      if(k != 0)X_message("\n");
      if     (anti_alias == ULTRA)X_message("Line %03ld ",(k+1)/3);
      else if(anti_alias == MEDIUM ||
              anti_alias == HIGH )X_message("Line %03ld ",(k+1)/2);
      else                        X_message("Line %03ld ",k+1);
   }
   X_message(".");
#ifndef __ZTC__
   fflush(stdout);
#endif
#endif
   if(R_terminator() == FAIL)return FAIL;
 }                                                    /* END OF CORE */
 if(dither_on != NO)ResetDither(); /* set the dithering error=0 for next frame */
 return 0;
}

long FastAntiAliasRenderHide(void){  /* only get phong shading if each + sample */
                                     /* is different at each point              */
 long i,j,k,l,m,Wsky;
 double d2,xl,xr,Rsky=0.0;
 vertex  *v;
 face    *f;
 eface   *e;
 if(R_terminator() == FAIL)return FAIL;
 Render_Message(60,0,NULL);
 if(R_Nground > 0 && hidden_line == NO){
   Render_Message(52,0,NULL);
   ShadowGround();
 }
 if(R_terminator() == FAIL)return FAIL;
 for(i=0;i<ObjectCount;i++)
   if(Object[i].visibility == YES && Object[i].shadowed != NO){
     Render_Message(53,0,NULL);
     ShadowObject(i);
     if(R_terminator() == FAIL)return FAIL;
   }
 Render_Message(57,0,NULL);
 MakeVertexNormals(); /* calculate Phong smoothing vertex normals */
 if(R_terminator() == FAIL)return FAIL;  /* Check for user cancel */
 if(R_Nground > 0 && Ground.refl > 0.0){
   Render_Message(58,0,NULL);
   Mirror_In_Ground();  /* see file GEO.C */
 }
 if(in_mirror_phase)Render_Message(99,0,NULL);
 else Render_Message(59,0,NULL);
 for(i=XMIN;i<XMAX;i++) {
   DL[i]=(double)(i - Xcentre);        DL[i]/=scalex;
   DR[i]=(double)(i+1-Xcentre);        DR[i]/=scalex;
   D1[i]=(double)(i - Xcentre) + 0.5;  D1[i]/=scalex;
 }
 GetSkyLocation();       /* for environment sky(same X pos for each line */
 for(l=YMIN;l<YMAX;l+=aaYstep){
 for(i=0;i<XMAX;i++){gXbuffer[i] = -1; gXobject[i] = -4;}
 for(m=0;m<aaYstep;m++){
   k=l+m;
   d2=(double)(Ycentre-k);
   d2/=scaley;
   Wsky=PutBackIntersection(k,d2,gAlfa,gZground,gZglass,gZglassO,gZglassd,
                            gZbuffer,gZobject,gZdepth,gZposn,
                            &Rsky);
   for(i=0;i<ObjectCount;i++)
   if((k > Object[i].ymin) && (k < Object[i].ymax) && !Object[i].wireframe
           &&(Object[i].visibility == YES)){
    v=Object[i].Vbase;
    f=Object[i].Fbase;
    e=Object[i].Ebase;
    for(j=0;j<Object[i].NoFaces;j++)
     if( Active(k,v,f+j,&xl,&xr) ){
       UpdateScanBuffers(xl,xr,j,i,f+j,e+j,d2,
                         gZglass,gZglassO,gZglassd,
                         gZbuffer,gZobject,gZdepth,gZposn);
     }
   }
   if(Ncsg > 0)GetCSGintersections(k,d2,gZcsg,gZground,
                                   gZbuffer,gZobject,gZdepth);
   if(alfachannelON == YES)FillAlfaLine(gAlfa,gZglassO,gZobject);
   if(FastAntiAliasShadedScanline(
                       k,m,d2,gZcsg,gRed,gGreen,gBlue,
                       gZground,Wsky,Rsky,gXbuffer,gXobject,
                       gZglass,gZglassO,gZglassd,
                       gZbuffer,gZobject,gZdepth,gZposn) == FAIL)return FAIL;
   if(bWire)RenderWireframeObjects(k,gRed,gGreen,gBlue,
                  gZglass,gZglassO,gZglassd,
                  gZbuffer,gZobject,gZdepth,gZposn);
   if(AntiAliasLine(k,anti_alias,
                    gRed,gGreen,gBlue,gAlfa,
                    gL_Red,gL_Green,gL_Blue,gL_Alfa,
                    gLL_Red,gLL_Green,gLL_Blue,gLL_Alfa) == FAIL){
     Render_Message(0,0,NULL);
     return FAIL;
   }
   if(fszBuffer != NULL)ScanlineToZbuffer(k,gZdepth);
   if(ObjectIdBuffer != NULL)ScanlineToZobject(k,gZobject);
#if __NT__
  if(!renderer_silent){
    if     (anti_alias == ULTRA)UpdateNTTrace((k+1)/3);
    else if(anti_alias == MEDIUM ||
            anti_alias == HIGH )UpdateNTTrace((k+1)/2);
    else                        UpdateNTTrace(k+1);
  }
#else
   if((k%60) == 0){
      if(k != 0)X_message("\n");
      if     (anti_alias == ULTRA)X_message("Line %03ld ",(k+1)/3);
      else if(anti_alias == MEDIUM ||
              anti_alias == HIGH )X_message("Line %03ld ",(k+1)/2);
      else                        X_message("Line %03ld ",k+1);
   }
   X_message(".");
#ifndef __ZTC__
   fflush(stdout);
#endif
#endif
   if(R_terminator() == FAIL)return FAIL;
   for(i=0;i<XMAX;i++){gXbuffer[i] = gZbuffer[i]; gXobject[i] = gZobject[i];}
 }
 }
 if(dither_on != NO)ResetDither(); /* set the dithering error=0 for next frame */
 return 0;
}

#else
// MULTITHREADING

typedef struct {
 int start,skip;
 HANDLE h;
} THSTART;

static BOOL all_terminate_flag=FALSE;
static long last_update=0;

void RenderMultiThreaded(void *arg){
 long i,j,k,l,m,Wsky;
 double d2,xl,xr,Rsky=0.0;
 vertex *v;
 face   *f;
 eface  *e;
 THD    thd,*p;
 THSTART *s=(THSTART *)arg;
//{char temp[255]; sprintf(temp,"Normal Thread %ld -> %ld",s->start,s->skip);
//MessageBox(NULL,temp,"Thread",MB_OK);}
 SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_BELOW_NORMAL);
 p=&thd; if(!AllocateThreadStorage(p))goto THDERROR;
//MessageBeep(MB_OK);
 for(l=s->start*aaYstep;l<YMAX;l+=(aaYstep*(s->skip))){
   for(m=0;m<aaYstep;m++){
     k=l+m;
     d2=(double)(Ycentre-k);
     d2/=scaley;
     Wsky=PutBackIntersection(k,d2,p->Alfa,p->Zground,
                              p->Zglass,p->ZglassO,p->Zglassd,
                              p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn,
                              &Rsky);
     for(i=0;i<ObjectCount;i++)
     if((k > Object[i].ymin) && (k < Object[i].ymax) && !Object[i].wireframe
             &&(Object[i].visibility == YES)){
      v=Object[i].Vbase;
      f=Object[i].Fbase;
      e=Object[i].Ebase;
      for(j=0;j<Object[i].NoFaces;j++)
       if( Active(k,v,f+j,&xl,&xr) ){
         UpdateScanBuffers(xl,xr,j,i,f+j,e+j,d2,
                           p->Zglass,p->ZglassO,p->Zglassd,
                           p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn);
       }
     }
     if(ShadedScanline(
                    k,d2,p->Zcsg,p->Red,p->Green,p->Blue,
                    p->Zground,Wsky,Rsky,
                    p->Zglass,p->ZglassO,p->Zglassd,
                    p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn) == FAIL){
       goto THDERROR;
     }
     if(bWire)RenderWireframeObjects(k,p->Red,p->Green,p->Blue,
                    p->Zglass,p->ZglassO,p->Zglassd,
                    p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn);
     if(alfachannelON == YES)FillAlfaLine(p->Alfa,p->ZglassO,p->Zobject);
     if(AntiAliasLine(k,anti_alias,
                      p->Red,p->Green,p->Blue,p->Alfa,
                      p->L_Red,p->L_Green,p->L_Blue,p->L_Alfa,
                      p->LL_Red,p->LL_Green,p->LL_Blue,p->LL_Alfa) == FAIL){
       Render_Message(0,0,NULL);
       goto THDERROR;
     }
     if(fszBuffer != NULL)ScanlineToZbuffer(k,p->Zdepth);
     if(ObjectIdBuffer != NULL)ScanlineToZobject(k,p->Zobject);
     if(!renderer_silent){
       long ud;
       if     (anti_alias == ULTRA)                        ud=((k+1)/3);
       else if(anti_alias == MEDIUM || anti_alias == HIGH )ud=((k+1)/2);
       else                                                ud=(k+1);
       if(ud > last_update){
         last_update=ud;
         UpdateNTTrace(ud);
       }
     }
     if(all_terminate_flag)goto THDEND;
   }
 }
 goto THDEND;
 THDERROR:
 all_terminate_flag=TRUE;
 THDEND:
 FreeThreadStorage(p);
 SetEvent(s->h);
 _endthread();
 return;
}

void FastAntiAliasRenderMultiThreaded(void *arg){
 long i,j,k,l,m,Wsky;
 double d2,xl,xr,Rsky=0.0;
 vertex *v;
 face   *f;
 eface  *e;
 THD    thd,*p;
 THSTART *s=(THSTART *)arg;
 SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_BELOW_NORMAL);
 p=&thd; if(!AllocateThreadStorage(p))goto THDERROR;
//{char temp[255]; sprintf(temp,"Thread %ld -> %ld",s->start,s->skip);
//MessageBox(NULL,temp,"Thread",MB_OK);}
//MessageBeep(MB_OK);
 for(l=s->start*aaYstep;l<YMAX;l+=(aaYstep*(s->skip))){
   for(i=0;i<XMAX;i++){p->Xbuffer[i] = -1; p->Xobject[i] = -4;}
   for(m=0;m<aaYstep;m++){
//Sleep(0);
     k=l+m;
     d2=(double)(Ycentre-k);
     d2/=scaley;
     Wsky=PutBackIntersection(k,d2,p->Alfa,p->Zground,
                              p->Zglass,p->ZglassO,p->Zglassd,
                              p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn,
                              &Rsky);
     for(i=0;i<ObjectCount;i++)
     if((k > Object[i].ymin) && (k < Object[i].ymax) && !Object[i].wireframe
             &&(Object[i].visibility == YES)){
      v=Object[i].Vbase;
      f=Object[i].Fbase;
      e=Object[i].Ebase;
      for(j=0;j<Object[i].NoFaces;j++)
       if( Active(k,v,f+j,&xl,&xr) ){
         UpdateScanBuffers(xl,xr,j,i,f+j,e+j,d2,
                           p->Zglass,p->ZglassO,p->Zglassd,
                           p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn);
       }
     }
     if(Ncsg > 0)GetCSGintersections(k,d2,p->Zcsg,p->Zground,
                                     p->Zbuffer,p->Zobject,p->Zdepth);
     if(alfachannelON == YES)FillAlfaLine(p->Alfa,p->ZglassO,p->Zobject);
     if(FastAntiAliasShadedScanline(
                         k,m,d2,p->Zcsg,p->Red,p->Green,p->Blue,
                         p->Zground,Wsky,Rsky,p->Xbuffer,p->Xobject,
                         p->Zglass,p->ZglassO,p->Zglassd,
                         p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn) == FAIL){
       goto THDERROR;
     }
     if(bWire)RenderWireframeObjects(k,p->Red,p->Green,p->Blue,
                    p->Zglass,p->ZglassO,p->Zglassd,
                    p->Zbuffer,p->Zobject,p->Zdepth,p->Zposn);
     if(AntiAliasLine(k,anti_alias,
                      p->Red,p->Green,p->Blue,p->Alfa,
                      p->L_Red,p->L_Green,p->L_Blue,p->L_Alfa,
                      p->LL_Red,p->LL_Green,p->LL_Blue,p->LL_Alfa) == FAIL){
       Render_Message(0,0,NULL);
       goto THDERROR;
     }
     if(fszBuffer != NULL)ScanlineToZbuffer(k,p->Zdepth);
     if(ObjectIdBuffer != NULL)ScanlineToZobject(k,p->Zobject);
     if(!renderer_silent){
       long ud;
       if     (anti_alias == ULTRA)                        ud=((k+1)/3);
       else if(anti_alias == MEDIUM || anti_alias == HIGH )ud=((k+1)/2);
       else                                                ud=(k+1);
       if(ud > last_update){
         last_update=ud;
         UpdateNTTrace(ud);
       }
     }
     if(all_terminate_flag)goto THDEND;
     for(i=0;i<XMAX;i++){p->Xbuffer[i] = p->Zbuffer[i]; p->Xobject[i] = p->Zobject[i];}
   }
 }
 goto THDEND;
 THDERROR:
 all_terminate_flag=TRUE;
 THDEND:
 FreeThreadStorage(p);
 SetEvent(s->h);
 _endthread();
 return;
}

long FastAntiAliasRenderHide(void){  /* only get phong shading if each sample diff */
 void (*fp)(void *);
 long i,iThrd;
 HANDLE hThrd[16],hEvent[16];
 THSTART Sup[16],*p;
 DWORD tID,result,Np;
 char str[16];
 if(R_terminator() == FAIL)return FAIL;
 Render_Message(60,0,NULL);
 if(R_Nground > 0 && hidden_line == NO){
   Render_Message(52,0,NULL);
   ShadowGround();
 }
 if(R_terminator() == FAIL)return FAIL;
 for(i=0;i<ObjectCount;i++)
   if(Object[i].visibility == YES && Object[i].shadowed != NO){
     Render_Message(53,0,NULL);
     ShadowObject(i);
     if(R_terminator() == FAIL)return FAIL;
 }
 Render_Message(57,0,NULL);
 MakeVertexNormals();
 if(R_terminator() == FAIL)return FAIL;
 if(R_Nground > 0 && Ground.refl > 0.0){
   Render_Message(58,0,NULL);
   Mirror_In_Ground();
 }
 if(Nthreads > 0)Np=Nthreads;
 else            Np=min(16,Nprocessors);
 sprintf(str,"%ld processors %ld",Nprocessors,Np);
 if(in_mirror_phase)Render_Message(99,0,NULL);
 else Render_Message(100,0,str);
 for(i=XMIN;i<XMAX;i++){
   DL[i]=(double)(i - Xcentre);        DL[i]/=scalex;
   DR[i]=(double)(i+1-Xcentre);        DR[i]/=scalex;
   D1[i]=(double)(i - Xcentre) + 0.5;  D1[i]/=scalex;
 }
 GetSkyLocation();
 all_terminate_flag=FALSE; last_update=0;
//{char temp[255]; sprintf(temp,"Nps=%ld XMAX=%ld YMAX=%ld aaYstep=%ld,aaXstep=%ld",
// Nprocessors,XMAX,YMAX,aaYstep,aaXstep); MessageBox(NULL,temp,"Debug",MB_OK);}
 for(i=0;i<Np;i++){
   Sup[i].start=i; Sup[i].skip=Np;
   hEvent[i]=CreateEvent(NULL,TRUE,FALSE,NULL);
   p=&Sup[i]; p->h=hEvent[i];
   if(anti_alias_edges == YES && anti_alias > LOW)
        fp=FastAntiAliasRenderMultiThreaded;
   else fp=RenderMultiThreaded;
#if  __WATCOMC__ && __WATCOMC__ < 1100
  iThrd=_beginthread(fp,0,65536,(void *)p);
#else
  iThrd=_beginthread(fp,0,(void *)p);
#endif
   if(iThrd == -1){
     MessageBox(NULL,"Error Creating Thread",NULL,MB_OK);
     return FAIL;
   }
   hThrd[i]=(HANDLE)iThrd;
 }
 while(1){
   if(!all_terminate_flag && renderer_NT_abort_flag){
     for(i=0;i<Np;i++)SuspendThread(hThrd[i]);
     if(ConfirmNTabort()){
       for(i=0;i<Np;i++)ResumeThread(hThrd[i]);
       all_terminate_flag=TRUE;
       WaitForMultipleObjects(Np,hEvent,TRUE,INFINITE);
       break;
     }
     else renderer_NT_abort_flag=0;
     for(i=0;i<Np;i++)ResumeThread(hThrd[i]);
   }
   result=WaitForMultipleObjects(Np,hEvent,TRUE,75);
   if(result == WAIT_TIMEOUT)RenderYield();
   else break;
 }
 for(i=0;i<Np;i++){
   CloseHandle(hEvent[i]);
//   CloseHandle(hThrd[i]);  // Bad for MIPS !
 }
 if(dither_on != NO)ResetDither(); /* set the dithering error=0 for next frame */
 if(all_terminate_flag)return FAIL;
 return 0;
}

long RenderHide(void){
 return FastAntiAliasRenderHide();
}

// END MULTITHREADING
#endif
