/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*   RENDER.H     include file for the renderer                    */

/*  Important  RENDER uses FLOATS   Design/Animation use double    */
/*  DO NOT COMMUNICATE FLOATS / DOUBLES                            */
/*  Don't use C++ style comments in renderer SUN/VAX dont like it  */

/* define this as long for  > 65536 vertices in an object */
#if __BIG__
#define UNSIGNED long
#else
#define UNSIGNED unsigned short
#endif

/* a true double precison and float type */
typedef double tDOUBLE;
typedef float  tFLOAT;

#if __ZTC__ || __WATCOMC__
#include <conio.h>
#endif

#include <process.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#if __NT__
#include <windows.h>
#include "rmessage.h"
extern void RenderYield(void);
extern void WindowedRenderMessage(int message);
extern void WindowedRenderString(char *str);
extern void WindowedRenderInfo(char *str);
extern void UpdateNTTrace(long l);
extern void UpdateNTFrame(long f, long s);
extern int ConfirmNTabort(void);
extern int GetRenderNTnewfile(char *name, char *rep_name);
#define X__getenv(c)    R_GetEnv(c)
#ifdef __X__NT__
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#else
#define X__Malloc(s)    GlobalAlloc(GMEM_FIXED,s)
#define X__Free(p)      GlobalFree((HGLOBAL)p)
#define X__Realloc(p,s) GlobalReAlloc((HGLOBAL)p,s,GMEM_MOVEABLE)
#endif
#else
#define X__getenv(c)    getenv(c)
#define X__Malloc(s)    malloc(s)
#define X__Free(s)      free(s)
#define X__Realloc(p,s) realloc(p,s)
#define BOOL unsigned char
#define TRUE  1
#define FALSE 0
#endif

#if __ZTC__ || __WATCOMC__
#include <dos.h>
#endif

#define YES       1
#define NO        0
#define FAIL     -1
#define OK        1

#if __ZTC__
  #if __NT__
  #define DOS386 1
  #endif
  #define GET_MATX  "%f %f %f %f"
  #define GET_VEC   "%f %f %f"
  #define GET_DOB   "%f"
  #define double float
  #define DROP NO
  #ifdef DOS386
    #define VERSION "Render Module "
  #else
    #define VERSION "Render Module "
  #endif
/* the following lines are needed only for ansi compatibility */
/*  #define PI 3.1415926    */
/*  #define strlwr(a)  (a)  */
  #define flushall()
#endif

#if __WATCOMC__ || __XVC__
#define GET_MATX  "%f %f %f %f"
#define GET_VEC   "%f %f %f"
#define GET_DOB   "%f"
#define double float
#define DROP NO
#define VERSION "Render Module "
#define DOS386 1
#define PI 3.1415926
#endif


#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

#ifdef TRACER
#undef  VERSION
#define VERSION "Tracer Module "
#endif

#define UNIT  32768L
#define UNIT2 65526L
#define BIGP  2147483647L
#define BIGN -2147483647L
#define FARAWAY 1.0e30
#define max_spec 1000
#define SHSCALE  3.921571e-3
//define NGLASS 2
#define NGLASS 4

#define FS_DITHER 1
#define RANDOM_DITHER 2

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define INVERT(a)       { a[0] = -a[0]; a[1] = -a[1]; a[2] = -a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }

#define SINGLE    1
#define TILE      2
#define PLANE           0 /*  Brush map type must be the same as modeler */
#define CYLINDER        1 /*         "                              "    */
#define PLANE_MOZIAC    2
#define CYLINDER_MOZIAC 3
#define MAP_BY_VERTEX   4
#define MAP_SPHERICAL   5
#define EOPAQUE   0       /*  Brush transparent colour 0                 */
#define TRANSP    1       /*  must be same as modeler                    */
#define GLOSS     0       /*  Brush is shiny  must be same as modeler    */
#define MATT      1       /*  Brush is dull                              */
#define STILBRUSH 0
#define ANIMBRUSH 1
#define BUMPBRUSH 2
#define REFLBRUSH 3
#define PSETBRUSH 4
#define PLAIN     0
#define TEXTU     1
#define BRUM      2
#define TBRUM     3
#define MBRUM     4
#define SKYCOLOUR 0
#define SKYGRADED 1
#define SKYMAPPED 2
#define BACKDROP  3
#define NORMAL    0    /* light types */
#define NOSHADOW  1
#define SHADOWS   2
#define SPOTLIGHT 3
#define DUMMYL    4
#define NOCOLOUR  32768
#define BITS8     0   /* output colour resolutions */
#define BITS16    1
#define BITS24    2
#define GIF_FILE  0   /* output file types */
#define IFF_FILE  1
#define LBM_FILE  2
#define I24_FILE  3
#define T24_FILE  4
#define TIF_FILE  5
#define T16_FILE  6
#define F24_FILE  7
#define B24_FILE  8
#define J24_FILE  9
#define PNG_FILE 10
#define FLC_FILE 20
#define NONE      0   /* anti-aliasing  */
#define LOW       1
#define MEDIUM    2
#define HIGH      3
#define ULTRA     4

#define UCHAR  unsigned char
typedef double vector[3];
typedef long   ivector[3];

typedef struct SCREENBUFFER {
 unsigned char A,R,G,B;
} fullscreenbuffer;

typedef struct tagXIP {
  long   imo,first_frame,this_frame,last_frame;
  double fmr;
  char   *parameters;
  char   *mparameters;
} XIP;

typedef struct tagTHD {
 UCHAR  *ram; /* pointer to base all other variable are offset */
 long   *Zbuffer, *Xbuffer,
        *Zobject, *Xobject,
        *Zglass[NGLASS],
        *ZglassO[NGLASS],
        *Zcsg;
 double *Zdepth,
        *Zposn,
        *Zglassd[NGLASS],
        *Zground;
 UCHAR  *Red,   *Green,   *Blue,   *Alfa,
        *L_Red, *L_Green, *L_Blue, *L_Alfa,
        *LL_Red,*LL_Green,*LL_Blue,*LL_Alfa;
} THD;

typedef struct VERTEX { /* structure for each vertex   */
 vector p;              /* real position  3D space     */
 double x,y;            /* screen location pixel space */
} vertex;

typedef struct TVERTEX {
 double x,y;            /* texture coordinates         */
} tvertex;

typedef struct FACE {   /* basic face data, ideally all face structures should */
 UNSIGNED V[3];         /* combined but 16 bit segments would => too few faces */
 unsigned char color[3];
 unsigned char fparam,matsp,mattr,matre,texco[3];
 unsigned char fattrib,fmap,fshader,faxis;
 /* BOOL t is not used now equates with hidden faces (option never used) */
 short    top,bottom,left,right;  /* screen bounds  0 -> XMAX 0 -> YMAX  */
} face;

typedef struct EFACE {  /* normal to face and   p.n to speed calc */
 vector n;
 double c;
} eface;

typedef struct PFACE {  /* The phong normals (Ideally this should not be    */
 vector   n[3];         /* necessary one normal per vertex should be enough */
} pface;                /* alas the inconsistent face normals can give BIG  */
                        /* problems when the face is _|_'r to line of sight */
                        /* vector. Assembling this way seem to give better  */
                        /* images. I don't know why yet ! ... Thinking      */

typedef struct LIGHTtag {
        vector pin;
        vector p,d;  /* p  is position  d is normalized dirn. for spots */
        double dot1,falloff,dPhi,dTheta,dAlpha,cone1,cone2,dot2,
               internal_size,scale;
        short type,AnimatorID;
        float *shadow_buffer;
        double shadow_transform[4][4];
        unsigned char color[3],internal_type,*object_buffer;
        unsigned short *face_buffer;
        double Ire,Igr,Ibl,Intensity;
        BOOL   dc;             /* depth cue - default FALSE */
        double dc_c,dc_l,dc_q; /* depth cue - parameters    */
        float *atmospheric_shadow_buffer;
} light;

typedef struct AXIStag {
       vector boin,bo;
       vector buin,bu;
       vector bvin,bv;
       vector bn;
       double tl;  // the lenght of the "bu" axis
       double tv;  // the lenght of the "bv" axis
       vector bu_norm,bv_norm;
} axis;

typedef struct IMAPtag {
        vector pin,xin,yin;
        vector n,p,x,y;
        char filename[128],frefl[128],fbump[128],ftran[128];
        short image,irefl,ibump,itran;     /* map     PLANE / CYL         */
        short map;                         /* image   YES if loaded       */
        short moziac;                      /* method  anim/still          */
        short pp,rp,bp,tp;                 /* map percent                 */
        double bumpx,bumpy,angle;          /* angle   CLY type maps only  */
        double pd,rd;
        unsigned char *picture,*prefl,*pbump,*ptran;
        unsigned char *palR,*palG,*palB;   /* moziac  0 normal 1 moziac   */
        unsigned char *palRrefl,*palGrefl,*palBrefl;
        unsigned char *palRbump,*palGbump,*palBbump;
        unsigned char *palRtran,*palGtran,*palBtran;
        short depth,drefl,dbump,dtran;      /* 8 bits is mapped 16/24       */
        unsigned char *p24R,*p24G,*p24B;
        unsigned char *p24Rrefl,*p24Grefl,*p24Brefl;
        unsigned char *p24Rbump,*p24Gbump,*p24Bbump;
        unsigned char *p24Rtran,*p24Gtran,*p24Btran;
        short xmax,ymax,
              xmaxrefl,ymaxrefl,
              xmaxbump,ymaxbump,
              xmaxtran,ymaxtran;
        unsigned char kt,kc[3];
} imap;

typedef struct VECTOR4tag {
 double x,y,z,w;
 BOOL   selected;
} vector4;

typedef struct tagNurbsProperties {
 char name[32];            /* name field             */
 long GranularityMax,      /* for visualising in the */
      GranularityMin;      /* view windows           */
 long     u_scale,         /* texture coord scaling  */
          v_scale;
 long     shader,          /* surface properties     */
          map,
          axis,
          param,
          smoothing,
          shiny,
          brilliance,
          reflectivity,
          transparency,
          glasstype,
          maptiled,
          mapshaded,
          mapdecal;
 long     pad[4];
 unsigned char colour[3],
          material_colour[3];
 char     hidden;
 char     padc[5];
} NurbsProperties;

typedef struct tagNurbsSurfSample {
    vector point,normal;
    double normLen;
    double u,v;
} NurbsSurfSample;

typedef struct NURBStag {
 long id;
 long numU,numV;
 long orderU,orderV;
 double *kvU,*kvV;
 vector4 **points;
 NurbsProperties properties;
 BOOL strV0, strVn, strU0, strUn;
 BOOL flatV, flatU;
 NurbsSurfSample c00, c0n, cn0, cnn;
} nurbs;

typedef struct OBJECTtag { /* info for each object */
        ivector Origin;
        axis    *Abase;
        vertex  *Vbase;
        vector  *Vcopy;  /* copy of vertex positions used by mirrors */
        face    *Fbase;
        eface   *Ebase;
        pface   *Pbase;
        tvertex *Tbase;
        imap    *Mbase;
        nurbs   *Nbase;
        long    NoFaces,NoVertices,NoAxes,NoMaps,NoNurbs;
        long    xmin,xmax,ymin,ymax;
        vector  Offset;
        double  trpos[4][4],trinv[4][4];
        double  fi,theta,alpha,ima;
        double  sx,sy,sz;
        double  morphratio,effectratio,ambient_light,refractive_index;
        short   fshader,shader[8],
                visibility,shadowed,mirrorID,
                im,effect;
        BOOL    wireframe;
        BOOL    self_shadow,cast_shadow,show_shadow;
        long    AnimatorID;
        char    *effect_param_list;
        unsigned char *mirror_buffer;
} object;

typedef struct GROUNDtag {
        vector        pin,xin,yin;
        vector        p,x,y,n,dx,dy,dyy;
        double        lengthx,lengthy,height,range,bump;
        double        spec,trans,refl,ambient_light;
        unsigned char color[3],acolor[3];
        short         type,maptype,texture,shiny,ispec,shader;
        imap          map;
        BOOL          mapped;
} ground;

typedef struct SKYtag {
        unsigned char Horizon[3],
                      Zenith[3],
                      Hblend[3],
                      Zblend[3];
        double blendratio;
        short  type;
        imap   map;
        BOOL   mapped;
} sky;

typedef struct SHADERtag {
  char file[128],parameters[256];
  void *module;
  void *function;
  void *morph;
  void *close;
} shader;

typedef struct CSG {
  int type,bool,im;
  long xmin,xmax,ymin,ymax;
  vector pin,p0,l,u,v,w,nn;
  double r1,r2,ll,dl;
  double trpos[4][4];
  double fi,theta,alpha,ima;
  double sx,sy,sz;
  face   f;
  eface  e;
} csg;

extern FILE *debug;

#ifdef MODULE_RENDER

BOOL          full_mirror=FALSE;
long          ResolutionX,ResolutionY
             ,renderer_silent=0
             ,XMIN=0,XMAX=320,YMIN=0,YMAX=200,FAST=NO
             ,OUTPUT_FILE_TYPE=GIF_FILE
             ,anti_alias=NO,anti_alias_edges=NO,ntsc_colours=NO
             ,screenbufferON=NO,fszBufferON=NO,objectIDbufferON=NO
             ,alfachannelON=NO,alfagroundON=NO,alfaseparate=NO
             ,remove_hidden_faces=NO,in_mirror_phase=NO,mblurON=0
             ,hidden_line=NO,hidden_line_angle=1,hidden_callback=NO
             ,instant_callback=NO,jpeg_quality=75,iLace=FALSE
             ,nCOLOUR=NOCOLOUR,RenderCurrentFrame,nScriptframes;
FILE         *work;
fullscreenbuffer *FullScreenBuffer,*MotionBlurBuffer;
float            *fszBuffer;
XIP              *ImageProcesses;

#if __NT__
extern unsigned long  *colorstats0;
extern unsigned short *colorstats1;
extern char *R_GetEnv(char const *aa);
#else
unsigned long   colorstats0[NOCOLOUR];
unsigned short  colorstats1[NOCOLOUR];
#endif
object         *Object=NULL;
ground         Ground;
sky            Sky;
csg            *Csg=NULL;
shader         *Shader;
vector         ViewPoint;
light          *Lights;
char          dither_on,colour0_flag=0;
unsigned char dither_factor,number_of_shadows,
              colour0_red=0,colour0_green=0,colour0_blue=0;
long          ObjectCount=0
             ,Xcentre,Ycentre
             ,Nlights=0,R_Nground=0,Nsky=0,Ncsg=0
             ,Nshader=0,nImageProcesses=0
             ,shadow_buffer_size=320
             ,NewPalette=YES
             ,skyBmax
             ,flashing=0,CamIm=0
             ,trace_shadows,trace_reflections,trace_refractions
             ,trace_optimise,trace_optimise_depth
             ,trace_optimise_threshold,trace_optimise_size
             ,ColourResolution=BITS8
             ,tzeros=3
             ,wireframe_thickness=1
             ,aaXstep,aaYstep;

unsigned short     *RendererUshortBuffer;
short              *ZskyBuffer,*RendererShortBuffer;
long               *gZbuffer,*gXbuffer,
                   *gZobject,*gXobject,
                   *gZglass[NGLASS],
                   *gZglassO[NGLASS],
                   *gZcsg;
double             *gZdepth,
                   *gZposn,
                   *gZglassd[NGLASS],
                   *gZground,
                   *DL,*DR,*D1,
                   AntiCamera[4][4],
                   trr[4][4]={1.0,0.0,0.0,0.0
                                ,0.0,1.0,0.0,0.0
                                ,0.0,0.0,1.0,0.0
                                ,0.0,0.0,0.0,1.0};
unsigned char *gRed,   *gGreen,   *gBlue,   *gAlfa,
              *gL_Red, *gL_Green, *gL_Blue, *gL_Alfa,
              *gLL_Red,*gLL_Green,*gLL_Blue,*gLL_Alfa,
              *DummyShadow;
double         scalex,scaley,
               BigL=FARAWAY,CamPhi,
               CamTheta,CamAlpha,CamIma,
               CamSx,CamSy,CamSz,
               skyYscale,skyXscale,
               ambient_light=0.1,smoothing_angle=0.01,
               shadow_density=0.5,glass_shadow_density=0.8,
               ground_ambient_light,
               object_ambient_light,
               wave_phase;
unsigned char *VGA_screen=NULL,*ObjectIdBuffer=NULL;
int            Nprocessors=1,Nthreads=0;
unsigned char  preview,preview_texture,preview_smoothing,
               preview_brflag,preview_tr,preview_pr,
               preview_refl,preview_background,
               preview_fcolour[3],preview_tcolour[3],
               preview_gloss;
long           preview_xscale,preview_yscale;
BOOL           bWire,bHideWire;

#else

extern BOOL  full_mirror;
extern long  ResolutionX,ResolutionY
            ,renderer_silent
            ,XMIN,XMAX,YMIN,YMAX,FAST,OUTPUT_FILE_TYPE
            ,anti_alias,anti_alias_edges,ntsc_colours
            ,screenbufferON,fszBufferON,objectIDbufferON
            ,alfachannelON,alfagroundON,alfaseparate
            ,remove_hidden_faces,in_mirror_phase,mblurON
            ,hidden_line,hidden_line_angle,hidden_callback
            ,instant_callback,jpeg_quality,iLace
            ,nCOLOUR,RenderCurrentFrame,nScriptframes;
extern FILE *work;
extern fullscreenbuffer *FullScreenBuffer,*MotionBlurBuffer;
extern float            *fszBuffer;
extern XIP              *ImageProcesses;
#if __NT__
extern unsigned long  *colorstats0;
extern unsigned short *colorstats1;
extern char *R_GetEnv(char const *aa);
#else
extern unsigned long   colorstats0[];
extern unsigned short  colorstats1[];
#endif
extern object         *Object;
extern ground         Ground;
extern sky            Sky;
extern csg            *Csg;
extern shader         *Shader;
extern vector         ViewPoint;
extern light          *Lights;
extern char          dither_on,colour0_flag;
extern unsigned char dither_factor,number_of_shadows,
                     colour0_red,colour0_green,colour0_blue;
extern long          ObjectCount
                    ,Xcentre,Ycentre
                    ,Nlights,nImageProcesses,R_Nground,Nsky,Ncsg
                    ,Nshader,shadow_buffer_size
                    ,NewPalette
                    ,skyBmax
                    ,flashing,CamIm
                    ,trace_shadows,trace_reflections,trace_refractions
                    ,trace_optimise,trace_optimise_depth
                    ,trace_optimise_threshold,trace_optimise_size
                    ,trace_shadows,trace_reflections,trace_refractions
                    ,trace_optimise
                    ,ColourResolution
                    ,tzeros
                    ,wireframe_thickness
                    ,aaXstep,aaYstep;
extern unsigned short   *RendererUshortBuffer;
extern short            *ZskyBuffer,*RendererShortBuffer;

extern double           *DL,*DR,*D1,
                        AntiCamera[4][4],
                        trr[4][4];
extern unsigned char  *DummyShadow;
extern double         scalex,scaley,
                      BigL,CamPhi,CamTheta,CamAlpha,CamIma,
                      CamSx,CamSy,CamSz,
                      skyYscale,skyXscale,
                      ambient_light,smoothing_angle,
                      shadow_density,glass_shadow_density,
                      ground_ambient_light,
                      object_ambient_light,
                      wave_phase;
extern unsigned char  *VGA_screen,*ObjectIdBuffer;
extern int            Nprocessors,Nthreads;
extern BOOL           bWire,bHideWire;

#endif

#include "protos.h"
