/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// file shapes.c


#define MODULE_CSG

#include "render.h"

extern double spec_pow[16][max_spec+1];

static vector box[8] = { UNIT, UNIT,-UNIT,
                        -UNIT, UNIT,-UNIT,
                        -UNIT,-UNIT,-UNIT,
                         UNIT,-UNIT,-UNIT,
                         UNIT, UNIT, UNIT,
                        -UNIT, UNIT, UNIT,
                        -UNIT,-UNIT, UNIT,
                         UNIT,-UNIT, UNIT};

#ifdef _SUNSTYLE
void TransformCSG() {
#else
void TransformCSG(void){
#endif
 long i,j;
 double temp4,xx,yy,zmax,tpos[4][4],tinv[4][4];
 ivector zero;
 vector ux,vy,wz,b;
 zero[0]=zero[1]=zero[2]=0;
 ux[0]=1.0; ux[1]=0.0; ux[2]=0.0;
 vy[0]=0.0; vy[1]=1.0; vy[2]=0.0;
 wz[0]=0.0; wz[1]=0.0; wz[2]=1.0;
 for(i=0;i<Ncsg;i++){
   Csg[i].xmin = BIGP;   Csg[i].xmax = BIGN;
   Csg[i].ymin = BIGP;   Csg[i].ymax = BIGN;
   zmax = (double)BIGN;
   SetTrPos(Csg[i].fi,Csg[i].theta,Csg[i].alpha
           ,(short)Csg[i].im,Csg[i].ima
           ,Csg[i].sx,Csg[i].sy,Csg[i].sz
           ,zero,Csg[i].pin,Csg[i].trpos,tpos,tinv);
   R_m4by1(trr, Csg[i].pin[0], Csg[i].pin[1], Csg[i].pin[2]
           ,&b[0],&b[1],&b[2]); /* the origin point is moved */
   veccopy(b,Csg[i].p0);
   for(j=0;j<8;j++){ /* b is bounding box for specifing scanlines */
     R_m4by1(Csg[i].trpos, box[j][0], box[j][1], box[j][2]
           ,&b[0],&b[1],&b[2]);
     if(b[1] >= 1.000){  /* all points on or in front of view plane */
       temp4=(scalex*(b[0])/(b[1]));
       xx=(double)Xcentre + temp4;
       temp4=(scaley*(b[2])/(b[1]));
       yy=(double)Ycentre - temp4;
       if(b[1] > zmax)zmax=b[1];
     }
     if(xx < (double)Csg[i].xmin) Csg[i].xmin = (long)xx-1;
     if(xx > (double)Csg[i].xmax) Csg[i].xmax = (long)xx+1;
     if(yy < (double)Csg[i].ymin) Csg[i].ymin = (long)yy-1;
     if(yy > (double)Csg[i].ymax) Csg[i].ymax = (long)yy+1;
   }
   /* if zmax < 0 then the csg is invisible */
   if(zmax < 0.0)Csg[i].ymin = YMAX+1;
   R_m4by1(Csg[i].trpos, ux[0], ux[1], ux[2]
          ,&Csg[i].u[0],&Csg[i].u[1],&Csg[i].u[2]);
   vecsub(Csg[i].u,Csg[i].p0,Csg[i].u);
   normalize(Csg[i].u);
   R_m4by1(Csg[i].trpos, vy[0], vy[1], vy[2]
          ,&Csg[i].v[0],&Csg[i].v[1],&Csg[i].v[2]);
   vecsub(Csg[i].v,Csg[i].p0,Csg[i].v);
   normalize(Csg[i].v);
   R_m4by1(Csg[i].trpos, wz[0], wz[1], wz[2]
         ,&Csg[i].w[0],&Csg[i].w[1],&Csg[i].w[2]);
   vecsub(Csg[i].w,Csg[i].p0,Csg[i].w);
   normalize(Csg[i].w);
   Csg[i].r1 = Csg[i].sx*Csg[i].sx*(double)UNIT*(double)UNIT;
   if(Csg[i].type == 1 || Csg[i].type == 2){ /* tube  cylinder */
     vecscale(Csg[i].sz*(double)UNIT,Csg[i].w,Csg[i].l);
     Csg[i].ll = 1.0/dot(Csg[i].l,Csg[i].l);
     Csg[i].dl = dot(Csg[i].p0,Csg[i].l)*Csg[i].ll;
     vecscale(Csg[i].dl,Csg[i].l,b);
     vecsub(b,Csg[i].p0,Csg[i].nn);
     Csg[i].r2 = dot(Csg[i].nn,Csg[i].nn) -
                 Csg[i].sx*Csg[i].sx*(double)UNIT*(double)UNIT;
   }
   /* use sx for radius  sz for height of cylinder */
 }
 return;
 /* all CSG's should now be at the right place and direction */
 /* the xmin etc. are used to indicate scanline activity     */
}

void GetCSGintersections(long scanline, double d2, long *Zcsg,
                         double *Zground,
                         long *Zbuffer, long *Zobject, double *Zdepth){
 long i,j,hit,csgid,csgidi;
 vector v1,v2,dv; /* view direction */
 double z1,z2,zt,zd,zf,zb,z,zz,s,pdv,p0p,dl,arg,a,b,c,psi,psi1;
 for(j=XMIN;j<XMAX;j++){
   csgid = -1; csgidi= -1; zf = FARAWAY; zb = 0;
   z=Zdepth[j]; zz=0.0;
   dv[0]=D1[j]; dv[1]=1.000; dv[2]=d2;
   s=R_length(dv); if(s > 1.e-10)s = 1.0/s; VECSCALE(s,dv,dv)
   for(i=0;i<Ncsg;i++){
     hit=0;
     if(scanline >= Csg[i].ymin && scanline <= Csg[i].ymax &&
               j >= Csg[i].xmin &&        j <= Csg[i].xmax){
       switch(Csg[i].type){
         case 0:  /* sphere */
           pdv=DOT(Csg[i].p0,dv);
           p0p=DOT(Csg[i].p0,Csg[i].p0);
           if((arg=(4*pdv*pdv - 4*(p0p-Csg[i].r1))) > 0){
             hit=1;
             arg=sqrt(arg);
             z1=pdv - arg/2;
             z2=pdv + arg/2;
             if(z2 < z1){arg=z2; z2=z1; z1=arg;}
           }
           break;
         case 1:  /* tube */
         case 2:  /* cylinder */
           dl=DOT(dv,Csg[i].l)*Csg[i].ll;
           VECSCALE(dl,Csg[i].l,v1)
           VECSUB(dv,v1,v2)
           a=DOT(v2,v2);
           b=2.0*DOT(v2,Csg[i].nn);
           c=Csg[i].r2;
           if((arg=(b*b - 4*a*c)) > 0){
             arg=sqrt(arg);
             z1= (-b + arg)/(2*a);
             z2= (-b - arg)/(2*a);
             if(z2 < z1){arg=z2; z2=z1; z1=arg;}
             a=z1*dl - Csg[i].dl;
             b=z2*dl - Csg[i].dl;
             if(Csg[i].type == 2){ /* solid cylinder, may need ends !*/
               if(-1.0 < a && a < 1.0 && -1.0 < b && b < 1.0){
                 hit=1; psi1=a;
               }
               else{
                 if(fabs((pdv = DOT(dv,Csg[i].w))) > 0.0001){
                   zt=zd=FARAWAY;
                   VECSUM(Csg[i].p0,Csg[i].l,v2);
                   c=DOT(v2,Csg[i].w)/pdv;
                   VECSCALE(c,dv,v1)
                   VECSUB(v1,v2,v2);
                   if(DOT(v2,v2) < Csg[i].r1)zt=c;
                   VECSUB(Csg[i].p0,Csg[i].l,v2);
                   c=DOT(v2,Csg[i].w)/pdv;
                   VECSCALE(c,dv,v1)
                   VECSUB(v1,v2,v2);
                   if(DOT(v2,v2) < Csg[i].r1)zd=c;
                   if(zd < zt){arg=zd; zd=zt; zt=arg;}
                   if(-1.0 < a && a < 1.0){hit=1; psi1=a;}
                   else if(zt < z2)       {hit=1; psi1=2; z1=zt;}
                   if(-1.0 > b || b > 1.0)z2=zd;
                 }
               }
             }
             else{ /* tube */
               if     (-1.0 < a && a < 1.0){hit=1; psi1=a;}
               else if(-1.0 < b && b < 1.0){hit=1; psi1=b; z1=z2;}
             }
           }
           break;
         case 3:  /* disk */
           if(fabs((pdv = DOT(dv,Csg[i].w))) > 0.0001){
             a=DOT(Csg[i].p0,Csg[i].w)/pdv;
             VECSCALE(a,dv,v1)
             VECSUB(v1,Csg[i].p0,v2);
             if(DOT(v2,v2) < Csg[i].r1){
               hit=1; z1=a;
             }
           }
           break;
         default:
           break;
       }
     }
     if(hit){
       if(Csg[i].bool){
         if(z1 < zf)zf=z1;
         if(z2 > zb){zb=z2; csgidi=i;}
       }
       else{
         if(z1 < z){
           z  = z1;
           csgid=i;
           psi=psi1;
         }
         if(z2 > zz)zz=z2;
       }
     }
   }
   if(z > zf && z < zb){      /* z is in negative region */
     z=zb;
     csgid = csgidi;
     if(zz < zb)csgid = -1;   /* completey see thro' */
   }
   if(csgid >= 0){
     Zbuffer[j] = -1;         /* invalidate any face data */
     Zobject[j] = -3;         /* flag as csg */
     Zdepth[j]  =  z*s;       /* scaling for dv */
     Zcsg[j]    =  csgid;     /* CSG identity   */
     Zground[j] =  psi;       /* use since ground obscured */
   }
 } /* x scan loop */
}

void GetCSGvalue(long i, long scanline, double d2, long *Zcsg,
                 UCHAR *Red, UCHAR *Green, UCHAR *Blue,
                 double *Zground, double *Zdepth){
 vector p,n,pl,pr,bv;
 register long j,k;
 double c,cc,Iab,IdR,IdG,IdB,IsR,IsG,IsB,falloff,alpha;
 p[0]=D1[i]; p[1]=1.000; p[2]=d2;
 VECSCALE(Zdepth[i],p,p)
 k=Zcsg[i];
 switch(Csg[k].type){
   case 0:  /* sphere */
     VECSUB(p,Csg[k].p0,n)
     break;
   case 1:  /* tube */
     VECSCALE(Zground[i],Csg[k].l,n)
     VECSUM(Csg[k].p0,n,n)
     VECSUB(p,n,n)
     break;
   case 2:  /* cylinder */
     if(Zground[i] > 1.1){ /* hits an end */
       VECCOPY(Csg[k].w,n)
     }
     else{
       VECSCALE(Zground[i],Csg[k].l,n)
       VECSUM(Csg[k].p0,n,n)
       VECSUB(p,n,n)
     }
/* if(DOT(p,n) > 0.0){n[0] = -n[0]; n[1] = -n[1]; n[2] = -n[2];} */
     break;
   case 3:  /* disk */
     VECCOPY(Csg[k].w,n)
/* if(DOT(p,n) > 0.0){n[0] = -n[0]; n[1] = -n[1]; n[2] = -n[2];} */
     break;
   default:
     break;
 }
 if(DOT(p,n) > 0.0){n[0] = -n[0]; n[1] = -n[1]; n[2] = -n[2];}
/* if(Csg[k].bool)VECSCALE(-1.0,n,n)    inside of CSG */
 normalize(n);
 IdR=IdG=IdB=0.0; IsR=IsG=IsB=0.0;
 for(j=0;j<Nlights;j++)if(Lights[j].type != SHADOWS &&
                          Lights[j].type != DUMMYL){
   VECSUB(Lights[j].p,p,pl)
   cc=R_length(pl);
   VECSCALE(1.0/cc,pl,pr)
   Iab = (1.0-ambient_light)*DOT(pr,n);
   if(Lights[j].type == SPOTLIGHT){
     if((alpha=DOT(pr,Lights[j].d)) < Lights[j].dot2)falloff=0.0;
     else if(alpha < Lights[j].dot1){ /* in between cones */
       falloff=(alpha-Lights[j].dot2)*Lights[j].falloff;
     }
     else falloff=1.0; /* otherwise inside inner cone and no atten */
     Iab *= falloff;
   }
   Iab += ambient_light;
   IdR += Iab*Lights[j].Ire;
   IdG += Iab*Lights[j].Igr;
   IdB += Iab*Lights[j].Ibl;
   if(1){   /*(f->fattrib & 0x20) == 0x20){ shiny  switch is on  */
     VECSUB(pl,p,bv)
     normalize(bv);
     c=fabs(DOT(bv,n));
     Iab = spec_pow[8][min(max_spec,(long)(c*max_spec))];
     if(Lights[j].type == SPOTLIGHT)Iab *= falloff;
     IsR += Iab*Lights[j].Ire;
     IsG += Iab*Lights[j].Igr;
     IsB += Iab*Lights[j].Ibl;
   }
 }
 Red[i]  = (unsigned char)min(255,((double)Csg[k].f.color[0]*IdR + IsR*255));
 Green[i]= (unsigned char)min(255,((double)Csg[k].f.color[1]*IdG + IsG*255));
 Blue[i] = (unsigned char)min(255,((double)Csg[k].f.color[2]*IdB + IsB*255));
}

