/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file update.c */

/* this is the source of the update control taken from book */

#include <windows.h>
#include <windowsx.h>

#include "update.h"

#if __SC__
#pragma ZTC align 1
#endif

char _szControlName[] = "Update";

#define CBWNDEXTRA         (12)
#define GWL_PARTSINJOB     (0)
#define GWL_PARTSCOMPLETE  (4)
#define GWL_FONT           (8)

ATOM NEAR RegisterControlClass (HINSTANCE hInstance);
LRESULT CALLBACK MeterWndFn
         (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);


#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  static BOOL unregister=TRUE;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
//      MessageBox (GetFocus(),"Update","Process attaching",MB_OK|MB_SYSTEMMODAL);
      if(!RegisterControlClass(hDLL)){
        unregister=FALSE;
      }
      break;
    }
    case DLL_PROCESS_DETACH:
      if(unregister)UnregisterClass(_szControlName, hDLL);
//      MessageBox (GetFocus(),"Update","Process detaching",MB_OK|MB_SYSTEMMODAL);
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

ATOM NEAR RegisterControlClass (HINSTANCE hInstance) {
   WNDCLASS wc;
   wc.style         = CS_GLOBALCLASS | CS_HREDRAW | CS_VREDRAW;
   wc.lpfnWndProc   = MeterWndFn;
   wc.cbClsExtra    = 0;
   wc.cbWndExtra    = CBWNDEXTRA;
   wc.hInstance     = hInstance;
   wc.hIcon         = (HICON) NULL;
   wc.hCursor       = LoadCursor((HINSTANCE)NULL,IDC_ARROW);
   wc.hbrBackground = (HBRUSH) NULL;
   wc.lpszMenuName  = NULL;
   wc.lpszClassName = _szControlName;
   return(RegisterClass(&wc));
}

LRESULT CALLBACK MeterWndFn
       (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
   LRESULT lResult = 0;
   char szPercentage[10];
   RECT rcClient, rcPrcnt;
   PAINTSTRUCT ps;
   WORD wPartsInJob, wPartsComplete;
   HBRUSH hBrush,hOldBrush;
   DWORD dwColor;
   HFONT hFont;
   SIZE Size;
   switch (uMsg) {
      case WM_GETDLGCODE:
         lResult = DLGC_STATIC;
         break;
      case WM_SETFONT:
         SetWindowLong(hWnd, GWL_FONT, (LONG)wParam);
         if (lParam) // Redraw the control immediately
            InvalidateRect(hWnd, NULL, FALSE);
         break;
      case WM_GETFONT:
         lResult = GetWindowLong(hWnd, GWL_FONT);
         break;
      case WM_ENABLE:
         InvalidateRect(hWnd, NULL, FALSE);
         break;
      case WM_CREATE:   // lParam == &CreateStruct
         SendMessage(hWnd, MM_SETPARTSINJOB, 100, 0l);
         SendMessage(hWnd, MM_SETPARTSCOMPLETE, 50, 0);
         break;
      case WM_PAINT:
         wPartsInJob = (WORD)
            SendMessage(hWnd, MM_GETPARTSINJOB,  0, 0l);
         wPartsComplete = (WORD)
            SendMessage(hWnd, MM_GETPARTSCOMPLETE, 0, 0l);
         if (wPartsInJob == 0) {
            wPartsInJob = 1;
            wPartsComplete = 0;
         }
         wsprintf(szPercentage,"%d%%",(100 * wPartsComplete) / wPartsInJob);
         BeginPaint(hWnd, &ps);
         hFont = GetWindowFont(hWnd);
         if (hFont != NULL)SelectObject(ps.hdc, hFont);
         SetBkColor(ps.hdc, GetSysColor(COLOR_WINDOW));
         SetTextColor(ps.hdc, RGB(0,0,255));
         hBrush=CreateSolidBrush(RGB(0,0,255));
         hOldBrush=         SelectObject(ps.hdc, hBrush);
         dwColor = GetBkColor(ps.hdc);
         SetBkColor(ps.hdc, SetTextColor(ps.hdc, dwColor));
         GetClientRect(hWnd, &rcClient);
         SetRect(&rcPrcnt, 0, 0,
            (rcClient.right * wPartsComplete) / wPartsInJob,
            rcClient.bottom);
         SetTextAlign(ps.hdc, TA_CENTER | TA_TOP);
         GetTextExtentPoint(ps.hdc, "X", 1,&Size);
         ExtTextOut(ps.hdc, rcClient.right / 2,
            (rcClient.bottom - Size.cy) / 2,
            ETO_OPAQUE | ETO_CLIPPED, &rcPrcnt,
            szPercentage, lstrlen(szPercentage), NULL);
         rcPrcnt.left = rcPrcnt.right;
         rcPrcnt.right = rcClient.right;
         dwColor = GetBkColor(ps.hdc);
         SetBkColor(ps.hdc, SetTextColor(ps.hdc, dwColor));
         GetTextExtentPoint(ps.hdc, "X", 1,&Size);
         ExtTextOut(ps.hdc, rcClient.right / 2,
            (rcClient.bottom - Size.cy) / 2,
            ETO_OPAQUE | ETO_CLIPPED, &rcPrcnt,
            szPercentage, lstrlen(szPercentage), NULL);
         SelectObject(ps.hdc,hOldBrush);
         DeleteObject(hBrush);
         EndPaint(hWnd, &ps);
         break;
      case MM_SETPARTSINJOB:
         SetWindowLong(hWnd, GWL_PARTSINJOB, wParam);
         InvalidateRect(hWnd, NULL, FALSE);
         UpdateWindow(hWnd);
         break;
      case MM_GETPARTSINJOB:
         lResult = (LONG) GetWindowLong(hWnd, GWL_PARTSINJOB);
         break;
      case MM_SETPARTSCOMPLETE:
         SetWindowLong(hWnd, GWL_PARTSCOMPLETE, wParam);
         InvalidateRect(hWnd, NULL, FALSE);
         UpdateWindow(hWnd);
         break;
      case MM_GETPARTSCOMPLETE:
         lResult = (LONG) GetWindowLong(hWnd, GWL_PARTSCOMPLETE);
         break;
      default:
         lResult = DefWindowProc(hWnd, uMsg, wParam, lParam);
         break;
   }
   return(lResult);
}
