/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

//  brick2d.c

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <windows.h>

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#if __WATCOMC__
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "..\animate\memory.h" /* for memory definition */
#include "..\animate\memdef.h" /* local names           */
#include "defines.h"
#include "rstruct.h"
#include "brick2d.h"

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }

#define IMOD(z,a) ((z) - ((z)/(a))*(a))

#define PATTERN(x,y,z)   ( ((z)+16384L)*64536L +  \
                           ((y)+16384L)*32768L +  \
                           ((x)+16384L))
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))
#define ROOT32 0.866025403
#define ROOT3   1.73205080

static double ffmod(double x){
 if(x < 0){
   double a;
   a=fmod(-x,1.0);
   return 1.0-a;
 }
 return fmod(x,1.0);
}

static double step(double a, double x){
 if(x < a)return 0.0;
 return 1.0;
}

static void mix(double *r, double *g, double *b, double mx,
                double r1, double g1, double b1,
                double r2, double g2, double b2){
 *r = mx*r1 + (1.0-mx)*r2;
 *g = mx*g1 + (1.0-mx)*g2;
 *b = mx*b1 + (1.0-mx)*b2;
}

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  HANDLE ghMod;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

static double thickness=0.1;
static double c1[3]={255.0,255.0,255.0};
static double x=1.0,y=1.0,z=1.0;

#if 0
long CALLBACK ExternalTextureStartup(
#else
long _ExternalTextureStartup(
#endif
  long frame,long nframes,char *parameter_list){
 if(parameter_list[0] != '#'){
   MessageBox ( GetFocus(),
                (LPCTSTR) "External texture: Startup",
                (LPCTSTR) "Parameter list missing",
                MB_OK | MB_SYSTEMMODAL );
 }
 else {
  parameter_list++;
  sscanf(parameter_list,"%f %f %f  %f  %f %f %f",
         &x,&y,&z,&thickness,
         &c1[0],&c1[1],&c1[2]);
 }
 return 1;
}

#if 0
long CALLBACK ExternalTextureMorph(
#else
long _ExternalTextureMorph(
#endif
 char *parameter_list, double mr){
 int i;
 double thickness_m;
 double c1_m[3];
 double x_m,y_m,z_m;
 parameter_list++;
 sscanf(parameter_list,"%f %f %f  %f  %f %f %f",
        &x_m,&y_m,&z,&thickness_m,
        &c1_m[0],&c1_m[1],&c1_m[2]);
 for(i=0;i<3;i++)c1[i]=c1_m[i]+(c1[i]-c1_m[i])*mr;
 thickness=thickness_m+(thickness-thickness_m)*mr;
 x=x_m+(x-x_m)*mr;
 y=y_m+(y-y_m)*mr;
 z=z_m+(z-z_m)*mr;
 return 1;
}

#if 0
long CALLBACK ExternalTextureProcedure(
#else
long _ExternalTextureProcedure(
#endif
  long coord_type,  vector p, vector n,
  double alpha, double beta, double gamma,
  double bump,  double map_x,  double map_y,
  double *alpha_channel, unsigned char sc[3], double colour[3],
  double *reflectivity, double *transparency,
  X__SHADER *lpEVI
){
 long i;
 double a,cc,xx,yy,zz,noiseH,r,g,b;
 vector u,v;
 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;
{
  double ss,tt,tbrick,sbrick,w,h,mf=0.03;
  ss = xx / 1.0; tt= yy/ 1.0;
  if(ffmod(tt*0.5) > 0.5)ss += 0.5;
  tbrick=floor(tt);
  nNoise(0.0,tbrick*3,0.0,&noiseH);
  ss += 0.3*(noiseH - 0.5);
  sbrick=floor(ss);
  ss -= sbrick;
  tt -= tbrick;
  vNoise(v,10*xx,10*yy,0.5);
  ss += (v[0]-0.5)*0.02;
  tt += (v[1]-0.5)*0.02;
  w=step(mf,ss) - step(1-mf,ss);
  h=step(mf,tt) - step(1-mf,tt);
  mix(&r,&g,&b,w*h,
      (double)colour[0]/255.0,(double)colour[1]/255.0,(double)colour[2]/255.0,
      (double)c1[0]/255.0,(double)c1[1]/255.0,(double)c1[2]/255.0);
  if(w*h == 0){
    if(w == 0){
      if(ss < mf)  VECCOPY(-(1.0-ss/mf)*tAxis_unorm,v)
      if(ss > 1-mf)VECCOPY(((ss-(1.0-mf))/mf)*tAxis_unorm,v)
    }
    else{
      if(tt < mf)  VECCOPY((1.0-tt/mf)*tAxis_vnorm,v)
      if(tt > 1-mf)VECCOPY(-((tt-(1.0-mf))/mf)*tAxis_vnorm,v)
//      VECCOPY(tAxis_vnorm,v)
    }
    VECSUM(n,0.5*v,n)
    vNoise(u,40*xx,40*yy,0.5);
    VECSUM(0.2*u,n,n)
    normalize(n);
  }
  else{
    vNoise(v,2*xx,2*yy,0.5);
    VECSCALE(0.2,v,v)
    VECSUM(n,v,n)
    vNoise(v,20*xx,20*yy,0.5);
    VECSCALE(0.1,v,v)
    VECSUM(n,v,n)
    if(ss > 1 - 2*mf)VECSUM(0.3*tAxis_unorm,n,n)
    else if(ss < 2*mf)VECSUM(-0.3*tAxis_unorm,n,n)
    else if(tt < 2*mf)VECSUM(0.3*tAxis_vnorm,n,n)
    else if(tt > 1 - 2*mf)VECSUM(-0.3*tAxis_vnorm,n,n)
    normalize(n);
    nNoise(xx*5,yy*5,0.5,&noiseH);
    r -= (noiseH*0.05);
    g -= (noiseH*0.05);
    b -= (noiseH*0.05);
  }
  r=max(r,0.0); g=max(g,0.0); b=max(b,0.0);
}
 colour[0]  = r * 255.0;
 colour[1]  = g * 255.0;
 colour[2]  = b * 255.0;
 return 1;
}

#if 0
void CALLBACK ExternalTextureClose(void){
#else
void _ExternalTextureClose(void){
#endif
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static void SetColour(double *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(double)GetRValue(cc.rgbResult);
   colour[1]=(double)GetGValue(cc.rgbResult);
   colour[2]=(double)GetBValue(cc.rgbResult);
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

#if 0
char * CALLBACK SetExternalParameters(
#else
char * _SetExternalParameters(
#endif
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                             ){
 char szbuf[255],*Op1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   Op1=Op;
   Op1++;
   sscanf(Op1,"%f %f %f  %f  %f %f %f",
         &x,&y,&z,&thickness,
         &c1[0],&c1[1],&c1[2]);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(szbuf,"# %.2f %.2f %.2f  %.3f "
                 "%.2f %.2f %.2f ",
         x,y,z,thickness,
         c1[0],c1[1],c1[2]);
 if((Op=(char *)CALL_MALLOC(strlen(szbuf)+1)) == NULL){
  MessageBox (GetFocus(),"External shader: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,szbuf);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.2f",x);
     SendDlgItemMessage(hwnd,DLG_XSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_XSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",y);
     SendDlgItemMessage(hwnd,DLG_YSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_YSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",z);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_COLOUR1){
         if(lpdis->CtlID == DLG_COLOUR1){
           r=(BYTE)c1[0]; g=(BYTE)c1[1]; b=(BYTE)c1[2];
         }
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_COLOUR1:
          SetColour(c1,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_COLOUR1),NULL,FALSE);
          break;
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_XSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((x=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_YSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((y=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_ZSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((z=atof(str)) == 0)EndDialog(hwnd,FALSE);
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
