/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// crspots.c

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <windows.h>

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#if __WATCOMC__
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "..\animate\memory.h" /* for memory definition */
#include "..\animate\memdef.h" /* local names           */
#include "defines.h"
#include "rstruct.h"
#include "crspots.h"

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }

#define IMOD(z,a) ((z) - ((z)/(a))*(a))

#define PATTERN(x,y,z)   ( ((z)+16384L)*64536L +  \
                           ((y)+16384L)*32768L +  \
                           ((x)+16384L))
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))
#define ROOT32 0.866025403
#define ROOT3   1.73205080

static void mix(double *r, double *g, double *b, double mx,
                double r1, double g1, double b1,
                double r2, double g2, double b2){
 *r = mx*r1 + (1.0-mx)*r2;
 *g = mx*g1 + (1.0-mx)*g2;
 *b = mx*b1 + (1.0-mx)*b2;
}

static long ip[27][3]={
   { 0, 0, 0},
   { 1, 0, 0},
   { 1, 1, 0},
   { 0, 1, 0},
   {-1, 1, 0},
   {-1, 0, 0},
   {-1,-1, 0},
   { 0,-1, 0},
   { 1,-1, 0},
   { 0, 0,-1},
   { 1, 0,-1},
   { 1, 1,-1},
   { 0, 1,-1},
   {-1, 1,-1},
   {-1, 0,-1},
   {-1,-1,-1},
   { 0,-1,-1},
   { 1,-1,-1},
   { 0, 0, 1},
   { 1, 0, 1},
   { 1, 1, 1},
   { 0, 1, 1},
   {-1, 1, 1},
   {-1, 0, 1},
   {-1,-1, 1},
   { 0,-1, 1},
   { 1,-1, 1}
 };

#include "randno.c"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  HANDLE ghMod;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

static double x=1.0,y=1.0,z=1.0,bradius=0.5,lradius=0.005,rradius=0.75;
static int    rtype=0;

#if 0
long CALLBACK ExternalTextureStartup(
#else
long _ExternalTextureStartup(
#endif
  long frame,long nframes,char *parameter_list){
 if(parameter_list[0] != '#'){
   MessageBox ( GetFocus(),
                (LPCTSTR) "External texture: Startup",
                (LPCTSTR) "Parameter list missing",
                MB_OK | MB_SYSTEMMODAL );
 }
 else {
  parameter_list++;
  sscanf(parameter_list,"%f %f %f  %ld  %f %f %f",
         &x,&y,&z,&rtype,&bradius,&lradius,&rradius);
 }
 return 1;
}

#if 0
long CALLBACK ExternalTextureMorph(
#else
long _ExternalTextureMorph(
#endif
 char *parameter_list, double mr){
 double x_m,y_m,z_m;
 double bradius_m,lradius_m,rradius_m;
 parameter_list++;
 sscanf(parameter_list,"%f %f %f  %ld  %f %f %f",
        &x_m,&y_m,&z_m,&rtype,&bradius_m,&lradius_m,&rradius_m);
 x=x_m+(x-x_m)*mr;
 y=y_m+(y-y_m)*mr;
 z=z_m+(z-z_m)*mr;
 bradius=bradius_m+(bradius-bradius_m)*mr;
 lradius=lradius_m+(lradius-lradius_m)*mr;
 rradius=rradius_m+(rradius-rradius_m)*mr;
 return 1;
}


#if 0
long CALLBACK ExternalTextureProcedure(
#else
long _ExternalTextureProcedure(
#endif
  long coord_type,  vector p, vector n,
  double alpha, double beta, double gamma,
  double bump,  double map_x,  double map_y,
  double *alpha_channel, unsigned char sc[3], double colour[3],
  double *reflectivity, double *transparency,
  X__SHADER *lpEVI
){
 double xx,yy,zz,noiseH,r,g,b;
 vector  vclr[27];   // N=1
 double  dpoint[27]; //
 double dradius[27];
 vector u,v,pp[10];
 double a,dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long ix,iy,iz;
 BOOL ins;
 long i,j,jid,id,pattern,patterni;
 long N=1;

 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;
 xx *= 4.0;
 yy *= 4.0;
 zz *= 4.0;
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(double)ix;
 cy=yy-(double)iy;
 cz=zz-(double)iz;
 zb=-10.0;
 ins=TRUE;
 id = -1;
 for(j=0;j<27;j++){
 pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
 pattern=IMOD(N*pattern,1900);
 for(i=0;i<N;i++){
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((double)ip[j],rand_v[patterni],pp[i])
   VECCOPY(rand_v[patterni],vclr[jid])
   radius=bradius+lradius*(rand_v[patterni][1]-0.5);
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(cx-pp[i][0]);
   dy=(cy-pp[i][1]);
   dz=(cz-pp[i][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius){
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb){
       id=jid;
       zb=dc;
     }
   }
 }
 }
 zbb= -10.0;
 for(i=0;i<N*27;i++){
   if(i == id)continue;
   if(dpoint[i] >= 0.0){
     dc=(dradius[i]-dpoint[i]);
     if(dc > zbb){
       zbb=dc;
     }
   }
 }
 if((ccc=fabs(zbb - zb)) < 0.0)ins=FALSE;   // flat edge
 if(id >= 0 && ins
   && (rtype == 0 || ( zb < dradius[id]-rradius*dradius[id] ) ) // rings
   ){
   r=vclr[id][0];
   g=vclr[id][1];
   b=vclr[id][2];
   *transparency=0.0;
   if(zb < 0.1){
     mix(&r,&g,&b,
         zb/(0.1),
         r,g,b,
         (double)colour[0]/255.0,
         (double)colour[1]/255.0,
         (double)colour[2]/255.0
         );
   }
 }
 else{
   r=colour[0]/255.0;
   g=colour[1]/255.0;
   b=colour[2]/255.0;
 }
 colour[0]  = r*255.0;
 colour[1]  = g*255.0;
 colour[2]  = b*255.0;
 return 1;
}

#if 0
void CALLBACK ExternalTextureClose(void){
#else
void _ExternalTextureClose(void){
#endif
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

#if 0
char * CALLBACK SetExternalParameters(
#else
char * _SetExternalParameters(
#endif
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                             ){
 char szbuf[255],*Op1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   Op1=Op;
   Op1++;
   sscanf(Op1,"%f %f %f  %ld  %f %f %f",
          &x,&y,&z,&rtype,&bradius,&lradius,&rradius);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(szbuf,"# %.2f %.2f %.2f  %ld  %.4f %.4f %.4f",
         x,y,z,rtype,bradius,lradius,rradius);
 if((Op=(char *)CALL_MALLOC(strlen(szbuf)+1)) == NULL){
  MessageBox (GetFocus(),"External shader: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,szbuf);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 double scale=120.0*3.0;
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.2f",x);
     SendDlgItemMessage(hwnd,DLG_XSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_XSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",y);
     SendDlgItemMessage(hwnd,DLG_YSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_YSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",z);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.3f",bradius);
     SendDlgItemMessage(hwnd,DLG_BRADIUS,WM_SETTEXT,0,(LPARAM)str);
     sprintf(str,"%.3f",lradius);
     SendDlgItemMessage(hwnd,DLG_LRADIUS,WM_SETTEXT,0,(LPARAM)str);
     sprintf(str,"%.3f",rradius);
     SendDlgItemMessage(hwnd,DLG_RRADIUS,WM_SETTEXT,0,(LPARAM)str);
     if(rtype == 0)
        SendDlgItemMessage(hwnd,DLG_SPOTS,BM_SETCHECK,TRUE,0);
     if(rtype == 1)
        SendDlgItemMessage(hwnd,DLG_RINGS,BM_SETCHECK,TRUE,0);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_XSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((x=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_YSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((y=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_ZSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((z=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BRADIUS,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((bradius=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_LRADIUS,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((lradius=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_RRADIUS,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((rradius=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(SendDlgItemMessage(hwnd,DLG_SPOTS,BM_GETCHECK,0,0))
            rtype=0; else rtype=1;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

