/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file DEFINES.H     include with defines for external functions */

/*****************************************************************\
  This file contains defines that enable meaningful local names to
  be used when calling functions in the Renderers external shader
  structure. Note the use of Windows Local memory allocation.
  Wave_Sources is an array of 10 vectors for ripple like effects.

  It is also very important that all "double" are equated to "float"
  the Renderer uses "float" type real numbers in all calcuations.
  The Modeler and Animator use true "double" real numbers.
\*****************************************************************/

#define double float         /* The Renderer uses FLOAT real numbers */
typedef double vector[3];

#ifdef __X__NT__
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#else
#define X__Malloc(s)    GlobalAlloc(GMEM_FIXED,s)
#define X__Free(p)      GlobalFree((HGLOBAL)p)
#define X__Realloc(p,s) GlobalReAlloc((HGLOBAL)p,s,GMEM_MOVEABLE)
#endif
#define  nNoise(x,y,z,r)      (lpEVI->fp_sNoise(x,y,z,r))
#define  sTurbulence(x,y,z,r) (lpEVI->fp_sTurbulence(x,y,z,r))
#define  Cycloidal(v,r)       (lpEVI->fp_Cycloidal(v,r))
#define  vNoise(v,x,y,z)      (lpEVI->fp_vNoise(v,x,y,z))
#define  vTurbulence(r,x,y,z) (lpEVI->fp_vTurbulence(r,x,y,z))
#define  normalize(v)         (lpEVI->fp_normalize(v))
#define  Wave_Sources         (lpEVI->fp_Wave_Sources)

#define  tAxis_n              (*(lpEVI->fp_axis_n))
#define  tAxis_o              (*(lpEVI->fp_axis_o))
#define  tAxis_u              (*(lpEVI->fp_axis_u))
#define  tAxis_v              (*(lpEVI->fp_axis_v))
#define  tAxis_unorm          (*(lpEVI->fp_axis_unorm))
#define  tAxis_vnorm          (*(lpEVI->fp_axis_vnorm))
#define  tAxis_u_length       (*(lpEVI->u_length))
#define  tAxis_v_length       (*(lpEVI->v_length))
#define  tAxis_tr             (lpEVI->tr)    // pointer to 4 x 4 arrays
#define  tAxis_itr            (lpEVI->itr)
