/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// faceted.c

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <windows.h>

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#if __WATCOMC__
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "..\animate\memory.h" /* for memory definition */
#include "..\animate\memdef.h" /* local names           */
#include "defines.h"
#include "rstruct.h"
#include "faceted.h"

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )

#define IMOD(z,a) ((z) - ((z)/(a))*(a))
#define PATTERN(x,y,z)   ( ((z)+16384L)*64536L +  \
                           ((y)+16384L)*32768L +  \
                           ((x)+16384L))
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))

#include "randno.c"

static long ip[27][3]={
   { 0, 0, 0},
   { 1, 0, 0},
   { 1, 1, 0},
   { 0, 1, 0},
   {-1, 1, 0},
   {-1, 0, 0},
   {-1,-1, 0},
   { 0,-1, 0},
   { 1,-1, 0},
   { 0, 0,-1},
   { 1, 0,-1},
   { 1, 1,-1},
   { 0, 1,-1},
   {-1, 1,-1},
   {-1, 0,-1},
   {-1,-1,-1},
   { 0,-1,-1},
   { 1,-1,-1},
   { 0, 0, 1},
   { 1, 0, 1},
   { 1, 1, 1},
   { 0, 1, 1},
   {-1, 1, 1},
   {-1, 0, 1},
   {-1,-1, 1},
   { 0,-1, 1},
   { 1,-1, 1}
 };

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  HANDLE ghMod;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

static double x=1.0,y=1.0,z=1.0;

#if 0
long CALLBACK ExternalTextureStartup(
#else
long _ExternalTextureStartup(
#endif
  long frame,long nframes,char *parameter_list){
 if(parameter_list[0] != '#'){
   MessageBox ( GetFocus(),
                (LPCTSTR) "External texture: Startup",
                (LPCTSTR) "Parameter list missing",
                MB_OK | MB_SYSTEMMODAL );
 }
 else {
  parameter_list++;
  sscanf(parameter_list,"%f %f %f",
         &x,&y,&z);
 }
 return 1;
}

#if 0
long CALLBACK ExternalTextureMorph(
#else
long _ExternalTextureMorph(
#endif
 char *parameter_list, double mr){
 double x_m,y_m,z_m;
 parameter_list++;
 sscanf(parameter_list,"%f %f %f",
        &x_m,&y_m,&z);
 x=x_m+(x-x_m)*mr;
 y=y_m+(y-y_m)*mr;
 z=z_m+(z-z_m)*mr;
 return 1;
}

#if 0
long CALLBACK ExternalTextureProcedure(
#else
long _ExternalTextureProcedure(
#endif
  long coord_type,  vector p, vector n,
  double alpha, double beta, double gamma,
  double bump,  double map_x,  double map_y,
  double *alpha_channel, unsigned char sc[3], double colour[3],
  double *reflectivity, double *transparency,
  X__SHADER *lpEVI
){
 static long N=1;
 static double bradius=1.0,lradius=0.0;
 double a;
 vector u,v;
 double xx,yy,zz;
 double  dpoint[27];
 double dradius[27];
 vector pp[27];
 double dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long i,ix,iy,iz;
 BOOL ins;
 long k,j,jid,id,pattern,patterni;
 double len,rad;
 vector ppc;

 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;

 xx *= 2.0;
 yy *= 2.0;
 zz *= 2.0;
 for(k=0;k<2;k++){
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(double)ix;
 cy=yy-(double)iy;
 cz=zz-(double)iz;
 zb=-10.0;
 ins=TRUE;
 id = -1;
 len=sqrt(tAxis_u[0]*tAxis_u[0]+
          tAxis_u[1]*tAxis_u[1]+
          tAxis_u[2]*tAxis_u[2]);
 VECCOPY(-tAxis_n,u)
 VECSCALE(len,u,u)         // third axis
 VECCOPY(tAxis_o,v)
 VECSUM(v,cx*tAxis_u,v)
 VECSUM(v,cy*tAxis_v,v)
 VECSUM(v,cz*u,v)         // coord of point on surface
 for(j=0;j<27;j++){
 pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
 pattern=IMOD(N*pattern,1900);
 for(i=0;i<N;i++){
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((double)ip[j],rand_v[patterni],ppc)
   VECSUM(tAxis_o,ppc[0]*tAxis_u,pp[jid])
   VECSUM(pp[jid],ppc[1]*tAxis_v,pp[jid])
   VECSUM(pp[jid],ppc[2]*u,      pp[jid])

   radius=(bradius+lradius*(rand_v[patterni][1]-0.5))*len;
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius){
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb){
       id=jid;
       zb=dc;
     }
   }
 }
 }
 if(id >= 0 && ins){
   {
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)          //(v-pp) vector from centre of pattern
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad){
       double mu,moddn,rmu;
       vector c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu){
         a=1.0-moddn/rmu;
         normalize(dn);
         dn[0]=0.5*(double)((int)(dn[0]*2.0)); //(4)
         dn[1]=0.5*(double)((int)(dn[1]*2.0));
         dn[2]=0.5*(double)((int)(dn[2]*2.0));
         normalize(dn);                        //(4)
         VECSCALE(-1.0,dn,dn);  // dimples
         a=0.7; // (4)
         VECSUM((1.0-a)*dn,(a)*n,n)
         normalize(n);
       }
     }
   }
 }
 xx *= 2; yy *= 2; zz *=2;  //(4)
 }
 return 1;
}

#if 0
void CALLBACK ExternalTextureClose(void){
#else
void _ExternalTextureClose(void){
#endif
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static void SetColour(double *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(double)GetRValue(cc.rgbResult);
   colour[1]=(double)GetGValue(cc.rgbResult);
   colour[2]=(double)GetBValue(cc.rgbResult);
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

#if 0
char * CALLBACK SetExternalParameters(
#else
char * _SetExternalParameters(
#endif
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                             ){
 char szbuf[255],*Op1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   Op1=Op;
   Op1++;
   sscanf(Op1,"%f %f %f",
         &x,&y,&z);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(szbuf,"# %.3f %.3f %.3f",
         x,y,z);
 if((Op=(char *)CALL_MALLOC(strlen(szbuf)+1)) == NULL){
  MessageBox (GetFocus(),"External shader: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,szbuf);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.2f",x);
     SendDlgItemMessage(hwnd,DLG_XSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_XSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",y);
     SendDlgItemMessage(hwnd,DLG_YSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_YSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",z);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_XSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((x=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_YSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((y=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_ZSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((z=atof(str)) == 0)EndDialog(hwnd,FALSE);
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
