/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

//  paving.c

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <windows.h>

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#if __WATCOMC__
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "..\animate\memory.h" /* for memory definition */
#include "..\animate\memdef.h" /* local names           */
#include "defines.h"
#include "rstruct.h"
#include "paving.h"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  HANDLE ghMod;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

static double flat_edge=0.025,bevel_edge=0.06,blend_edge=0.06;
static double c1[3]={200.0,130.0, 90.0};
static double x=1.0,y=1.0,z=1.0;

#if 0
long CALLBACK ExternalTextureStartup(
#else
long _ExternalTextureStartup(
#endif
  long frame,long nframes,char *parameter_list){
 if(parameter_list[0] != '#'){
   MessageBox ( GetFocus(),
                (LPCTSTR) "External texture: Startup",
                (LPCTSTR) "Parameter list missing",
                MB_OK | MB_SYSTEMMODAL );
 }
 else {
  parameter_list++;
  sscanf(parameter_list,"%f %f %f  %f %f %f  %f %f %f",
         &x,&y,&z,&flat_edge,&bevel_edge,&blend_edge,
         &c1[0],&c1[1],&c1[2]);
 }
 return 1;
}

#if 0
long CALLBACK ExternalTextureMorph(
#else
long _ExternalTextureMorph(
#endif
 char *parameter_list, double mr){
 int i;
 double flat_edge_m,bevel_edge_m,blend_edge_m;
 double c1_m[3],c2_m[3],c3_m[3];
 double x_m,y_m,z_m;
 parameter_list++;
 sscanf(parameter_list,"%f %f %f  %f %f %f  %f %f %f",
        &x_m,&y_m,&z,&flat_edge_m,&bevel_edge_m,&blend_edge_m,
        &c1_m[0],&c1_m[1],&c1_m[2]);
 for(i=0;i<3;i++)c1[i]=c1_m[i]+(c1[i]-c1_m[i])*mr;
 flat_edge=flat_edge_m+(flat_edge-flat_edge_m)*mr;
 bevel_edge=bevel_edge_m+(bevel_edge-bevel_edge_m)*mr;
 blend_edge=blend_edge_m+(blend_edge-blend_edge_m)*mr;
 x=x_m+(x-x_m)*mr;
 y=y_m+(y-y_m)*mr;
 z=z_m+(z-z_m)*mr;
 return 1;
}

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }
#define IMOD(z,a) ((z) - ((z)/(a))*(a))
#define PATTERN(x,y,z)   ( ((z)+16384L)*64536L +  \
                           ((y)+16384L)*32768L +  \
                           ((x)+16384L))
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))
#define ROOT32 0.866025403
#define ROOT3   1.73205080

#include "randno.c"

static long ip[27][3]={
   { 0, 0, 0},
   { 1, 0, 0},
   { 1, 1, 0},
   { 0, 1, 0},
   {-1, 1, 0},
   {-1, 0, 0},
   {-1,-1, 0},
   { 0,-1, 0},
   { 1,-1, 0},
   { 0, 0,-1},
   { 1, 0,-1},
   { 1, 1,-1},
   { 0, 1,-1},
   {-1, 1,-1},
   {-1, 0,-1},
   {-1,-1,-1},
   { 0,-1,-1},
   { 1,-1,-1},
   { 0, 0, 1},
   { 1, 0, 1},
   { 1, 1, 1},
   { 0, 1, 1},
   {-1, 1, 1},
   {-1, 0, 1},
   {-1,-1, 1},
   { 0,-1, 1},
   { 1,-1, 1}
 };

static void mix(double *r, double *g, double *b, double mx,
                double r1, double g1, double b1,
                double r2, double g2, double b2){
 *r = mx*r1 + (1.0-mx)*r2;
 *g = mx*g1 + (1.0-mx)*g2;
 *b = mx*b1 + (1.0-mx)*b2;
}

#if 0
long CALLBACK ExternalTextureProcedure(
#else
long _ExternalTextureProcedure(
#endif
  long coord_type,  vector p, vector n,
  double alpha, double beta, double gamma,
  double bump,  double map_x,  double map_y,
  double *alpha_channel, unsigned char sc[3], double colour[3],
  double *reflectivity, double *transparency,
  X__SHADER *lpEVI
){
 double xx,yy,zz,r,g,b;
 double a;
 vector u,v;
 vector vclr[27];
 double dpoint[27];
 vector pp[10];
 double dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,noiseH;
 long ix,iy,iz;
 BOOL ins;
 long i,j,jid,id,pattern,patterni;

 static double radius=1.0;
 static double vc=0.1;   // discolouration

 xx=alpha*4.0 / x;
 yy=beta*4.0  / y;
 zz=gamma*4.0 / z;
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(double)ix;
 cy=yy-(double)iy;
 cz=zz-(double)iz;
 zb=-10.0;
 ins=TRUE;
 for(i=0,j=0;j<27;j++){
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(pattern,1900);   // should be N*pattern for n > 1
   jid=j+i;
   patterni=pattern+i;
   VECSUM((double)ip[j],rand_v[patterni],pp[i])
   VECCOPY(rand_v[patterni],vclr[jid])
   dpoint[jid] = -1.0;
   dx=(cx-pp[i][0]);
   dy=(cy-pp[i][1]);
   dz=(cz-pp[i][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius){
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb){
       id=jid;
       zb=dc;
     }
   }
 }
 zbb= -10.0;
 for(i=0;i<27;i++){
   if(i == id)continue;
   if(dpoint[i] >= 0.0){
     dc=(radius-dpoint[i]);
     if(dc > zbb){
       zbb=dc;
     }
   }
 }
 if((ccc=fabs(zbb - zb)) < flat_edge)ins=FALSE;
 if(ins){
   if(0){
     r=vclr[id][0];
     g=vclr[id][1];
     b=vclr[id][2];
   }
   else{
     r=colour[0]/255.0 + vc*(vclr[id][0]-0.5);
     g=colour[1]/255.0 + vc*(vclr[id][1]-0.5);
     b=colour[2]/255.0 + vc*(vclr[id][2]-0.5);
     r=min(r,1.0); g=min(g,1.0); b=min(b,1.0);
     nNoise(xx*2.0,yy*2.0,zz*2.0,&noiseH);
     r -= (noiseH*0.3);
     g -= (noiseH*0.3);
     b -= (noiseH*0.3);
     r=max(r,0.0); g=max(g,0.0); b=max(b,0.0);
   }
   if(ccc < bevel_edge){
     v[0]=(cx-pp[id][0]);
     v[1]=(cy-pp[id][1]);
     v[2]=(cz-pp[id][2]);
     normalize(v);
     CROSS(v,n,u)
     CROSS(n,u,v)
     normalize(v);
     a=0.4;
     VECSCALE(a,v,v)
     VECSCALE((1.0-a),n,n)
     VECSUM(v,n,n)
     normalize(n);
   }
   if(ccc < blend_edge)mix(&r,&g,&b,(ccc-flat_edge)/(blend_edge-flat_edge),
         r,g,b,
         (double)c1[0]/255.0,
         (double)c1[1]/255.0,
         (double)c1[2]/255.0
         );
 }
 else{
   r=c1[0]/255.0;
   g=c1[1]/255.0;
   b=c1[2]/255.0;
 }
 colour[0]=r*255;
 colour[1]=g*255;
 colour[2]=b*255;

 return 1;
}

#if 0
void CALLBACK ExternalTextureClose(void){
#else
void _ExternalTextureClose(void){
#endif
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static void SetColour(double *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(double)GetRValue(cc.rgbResult);
   colour[1]=(double)GetGValue(cc.rgbResult);
   colour[2]=(double)GetBValue(cc.rgbResult);
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

#if 0
char * CALLBACK SetExternalParameters(
#else
char * _SetExternalParameters(
#endif
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                             ){
 char szbuf[255],*Op1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   Op1=Op;
   Op1++;
   sscanf(Op1,"%f %f %f  %f %f %f  %f %f %f",
         &x,&y,&z,&flat_edge,&bevel_edge,&blend_edge,
         &c1[0],&c1[1],&c1[2]);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(szbuf,"# %.2f %.2f %.2f  %.3f %.3f %.3f "
                 "%.2f %.2f %.2f",
         x,y,z,flat_edge,bevel_edge,blend_edge,
         c1[0],c1[1],c1[2]);
 if((Op=(char *)CALL_MALLOC(strlen(szbuf)+1)) == NULL){
  MessageBox (GetFocus(),"External shader: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,szbuf);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.2f",x);
     SendDlgItemMessage(hwnd,DLG_XSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_XSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",y);
     SendDlgItemMessage(hwnd,DLG_YSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_YSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",z);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.3f",flat_edge);
     SendDlgItemMessage(hwnd,DLG_SIZE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_SIZE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.3f",bevel_edge);
     SendDlgItemMessage(hwnd,DLG_DEPTH,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_DEPTH,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.3f",blend_edge);
     SendDlgItemMessage(hwnd,DLG_BLEND,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BLEND,EM_LIMITTEXT,(WPARAM)12,0);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_COLOUR){
         if(lpdis->CtlID == DLG_COLOUR){
           r=(BYTE)c1[0]; g=(BYTE)c1[1]; b=(BYTE)c1[2];
         }
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_COLOUR:
          SetColour(c1,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_COLOUR),NULL,FALSE);
          break;
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_SIZE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((flat_edge=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_DEPTH,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((bevel_edge=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BLEND,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((blend_edge=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_XSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((x=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_YSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((y=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_ZSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((z=atof(str)) == 0)EndDialog(hwnd,FALSE);
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
