/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

//  plates.c

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <windows.h>

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#if __WATCOMC__
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "..\animate\memory.h" /* for memory definition */
#include "..\animate\memdef.h" /* local names           */
#include "defines.h"
#include "rstruct.h"
#include "plates.h"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  HANDLE ghMod;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

static int type=1, dirty=1;
static double dhue=0.05,dval=0.3;
static double x=1.0,y=1.0,z=1.0;

#if 0
long CALLBACK ExternalTextureStartup(
#else
long _ExternalTextureStartup(
#endif
  long frame,long nframes,char *parameter_list){
 if(parameter_list[0] != '#'){
   MessageBox ( GetFocus(),
                (LPCTSTR) "External texture: Startup",
                (LPCTSTR) "Parameter list missing",
                MB_OK | MB_SYSTEMMODAL );
 }
 else {
  parameter_list++;
  sscanf(parameter_list,"%f %f %f  %f %f  %ld %ld",
         &x,&y,&z,&dhue,&dval,&type,&dirty);
 }
 return 1;
}

#if 0
long CALLBACK ExternalTextureMorph(
#else
long _ExternalTextureMorph(
#endif
 char *parameter_list, double mr){
 double dhue_m,dval_m;
 double x_m,y_m,z_m;
 parameter_list++;
 sscanf(parameter_list,"%f %f %f  %f %f",
        &x_m,&y_m,&z_m,&dhue_m,&dval_m);
 dhue=dhue_m+(dhue-dhue_m)*mr;
 dval=dval_m+(dval-dval_m)*mr;
 x=x_m+(x-x_m)*mr;
 y=y_m+(y-y_m)*mr;
 z=z_m+(z-z_m)*mr;
 return 1;
}

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }
#define IMOD(z,a) ((z) - ((z)/(a))*(a))
#define PATTERN(x,y,z)   ( ((z)+16384L)*64536L +  \
                           ((y)+16384L)*32768L +  \
                           ((x)+16384L))
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))
#define ROOT32 0.866025403
#define ROOT3   1.73205080

#include "randno.c"

/* h is 0->360  s and v are 0-1 */
void Hsv2Rgb(double h, double s, double v,
             double *r, double *g, double *b){
 double rr,gg,bb,hh,ss,vv,maxcol,mincol;
 hh=(double)h; ss=(double)s; vv=(double)v;
 mincol=vv*(1.0-ss);
 if(hh <= 120.0){bb=mincol;
   if(hh <=  60){rr=vv; gg=mincol+hh*(vv-mincol)/(120.0-hh);}
   else         {gg=vv; rr=mincol+(120.0-hh)*(vv-mincol)/hh;}
 }
 else if( hh <= 240){rr=mincol;
   if(hh <= 180){gg=vv; bb=mincol+(hh-120.0)*(vv-mincol)/(240.0-hh);}
   else         {bb=vv; gg=mincol+(240.0-hh)*(vv-mincol)/(hh-120.0);}
 }
 else {gg=mincol;
   if(hh <= 300){bb=vv; rr=mincol+(hh-240.0)*(vv-mincol)/(360.0-hh);}
   else         {rr=vv; bb=mincol+(360.0-hh)*(vv-mincol)/(hh-240.0);}
 }
 *r = rr;
 *g = gg;
 *b = bb;
}

void Rgb2Hsv(double rr, double gg, double bb,  // rr etc. are (0-1)
             double *h, double *s, double *v){ // h (0-360)  s v (0-1)
 double maxcol,mincol;
 maxcol=max(rr,max(gg,bb));
 mincol=min(rr,min(gg,bb));
 *v=maxcol;
 if(maxcol == mincol){*s=0; *h=0;}
 else{
  *s=(maxcol-mincol)/maxcol;
  if(mincol == bb)
    *h = 120.0*(gg-mincol)/(rr+gg-2.0*mincol);
  else if(mincol == rr)
    *h = 120.0*(1.0+(bb-mincol)/(bb+gg-2.0*mincol));
  else
    *h = 120.0*(2.0+(rr-mincol)/(rr+bb-2.0*mincol));
 }
}

#if 0
long CALLBACK ExternalTextureProcedure(
#else
long _ExternalTextureProcedure(
#endif
  long coord_type,  vector p, vector n,
  double alpha, double beta, double gamma,
  double bump,  double map_x,  double map_y,
  double *alpha_channel, unsigned char sc[3], double colour[3],
  double *reflectivity, double *transparency,
  X__SHADER *lpEVI
){
 long id;
 double xx,yy,zz,noise;
 double xr,xl,yr,yl,zr,zl,d=0.05;
 vector v;
 xx=5.0*alpha / x;
 yy=5.0*beta  / y;
 zz=5.0*gamma / z;

 zz *=2.0;

 id=PATTERN((long)FLOOR(xx),(long)FLOOR(yy),(long)FLOOR(zz));
 id=IMOD(id,1900);
 if(type == 0){
   colour[0] = rand_v[id][0]*255.0;
   colour[1] = rand_v[id][1]*255.0;
   colour[2] = rand_v[id][2]*255.0;
 }
 else{
   double hue,sat,val;
   Rgb2Hsv(colour[0]/255.0,colour[1]/255.0,colour[2]/255.0,&hue,&sat,&val);
   hue += dhue*(rand_v[id][0]-0.5)*360.0;
   if(hue <   0.0)hue += 360.0;
   if(hue > 360.0)hue -= 360.0;
   val += val*dval*(rand_v[id][1]-0.5);
   val=min(val,1.0);
   val=max(val,0.0);
   Hsv2Rgb(hue,sat,val,&colour[0],&colour[1],&colour[2]);
   colour[0] *= 255.0;
   colour[1] *= 255.0;
   colour[2] *= 255.0;
 }
 if(dirty){
   nNoise(xx*2.0,yy*2.0,zz*1.0,&noise);
   colour[0] -= (noise*0.3)*255.0;
   colour[1] -= (noise*0.3)*255.0;
   colour[2] -= (noise*0.3)*255.0;
   colour[0]=max(colour[0],0.0);
   colour[1]=max(colour[1],0.0);
   colour[2]=max(colour[2],0.0);
 }

 xx -= FLOOR(xx);
 yy -= FLOOR(yy);
 zz -= FLOOR(zz);
 v[0]=0.0; v[1]=0.0; v[2]=0.0;
 if(xx < d){
   VECCOPY(-tAxis_u,v);
   normalize(v);
 }
 else if(xx > 1.0-d){
   VECCOPY(tAxis_u,v);
   normalize(v);
 }
 else if(yy < d){
   VECCOPY(-tAxis_v,v);
   normalize(v);
 }
 else if(yy > 1.0-d){
   VECCOPY(tAxis_v,v);
   normalize(v);
 }
 else if(zz < d){
   VECCOPY(-tAxis_v,v);
   normalize(v);
 }
 else if(zz > 1.0-d){
   VECCOPY(tAxis_v,v);
   normalize(v);
 }
 VECSUM(0.25*v,n,n);

 return 1;
}

#if 0
void CALLBACK ExternalTextureClose(void){
#else
void _ExternalTextureClose(void){
#endif
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static void SetColour(double *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(double)GetRValue(cc.rgbResult);
   colour[1]=(double)GetGValue(cc.rgbResult);
   colour[2]=(double)GetBValue(cc.rgbResult);
 }
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

#if 0
char * CALLBACK SetExternalParameters(
#else
char * _SetExternalParameters(
#endif
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                             ){
 char szbuf[255],*Op1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   Op1=Op;
   Op1++;
   sscanf(Op1,"%f %f %f  %f %f  %ld %ld",
         &x,&y,&z,&dhue,&dval,&type,&dirty);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG),hWnd,
              (DLGPROC)DlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(szbuf,"# %.2f %.2f %.2f  %.3f %.3f  %ld %ld",
        x,y,z,dhue,dval,type,dirty);
 if((Op=(char *)CALL_MALLOC(strlen(szbuf)+1)) == NULL){
  MessageBox (GetFocus(),"External shader: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,szbuf);
 return Op;
}

BOOL CALLBACK DlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.2f",x);
     SendDlgItemMessage(hwnd,DLG_XSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_XSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",y);
     SendDlgItemMessage(hwnd,DLG_YSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_YSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",z);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_ZSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.3f",dval);
     SendDlgItemMessage(hwnd,DLG_DVAL,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_DVAL,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.3f",dhue);
     SendDlgItemMessage(hwnd,DLG_DHUE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_DHUE,EM_LIMITTEXT,(WPARAM)12,0);
     if(dirty)SendDlgItemMessage(hwnd,DLG_DIRTY,BM_SETCHECK,TRUE,0);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case IDCANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
          if(GetDlgItemText(hwnd,DLG_DHUE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((dhue=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_DVAL,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((dval=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_XSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((x=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_YSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((y=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_ZSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((z=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(SendDlgItemMessage(hwnd,DLG_DIRTY,BM_GETCHECK,0,0))dirty=1;
          else dirty=0;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}
