/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// test.c

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <windows.h>

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#if __WATCOMC__
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "..\animate\memory.h" /* for memory definition */
#include "..\animate\memdef.h" /* local names           */
#include "defines.h"
#include "rstruct.h"
#include "test.h"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  HANDLE ghMod;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }

#define IMOD(z,a) ((z) - ((z)/(a))*(a))

#define PATTERN(x,y,z)   ( ((z)+16384L)*64536L +  \
                           ((y)+16384L)*32768L +  \
                           ((x)+16384L))
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))
#define ROOT32 0.866025403
#define ROOT3   1.73205080

#define nnNoise(x,y,z) (2.0*(nNoise(x,y,z)-0.5))

#define Noise3(nr,xx,yy,zz)        \
{                                  \
  nr = 0.5*(1+sin(6*3.1415926*nNoise(xx,yy,zz)));          \
}

//  nr += nNoise(xx*4,yy*4,zz*4);

static double turb=0.1,t1=0.4,t2=0.7,c[3]={255.0,255.0,0.0};
static double x=1.0,y=1.0,z=1.0;
static long   tr_type=0,tr_type1=0,tr_type2=0,tr_type3=0;

#if 0
long CALLBACK ExternalTextureStartup(
#else
long _ExternalTextureStartup(
#endif
  long frame,long nframes,char *parameter_list){
 if(parameter_list[0] != '#'){
   MessageBox ( GetFocus(),
                (LPCTSTR) "External texture: Startup",
                (LPCTSTR) "Parameter list missing",
                MB_OK | MB_SYSTEMMODAL );
 }
 else {
  parameter_list++;
  sscanf(parameter_list,"%f %f %f  %f  %f %f  %f %f %f %ld %ld %ld %ld",
         &c[0],&c[1],&c[2],&turb,&t1,&t2,&x,&y,&z,
         &tr_type,&tr_type1,&tr_type2,&tr_type3);
 }
 return 1;
}

#if 0
long CALLBACK ExternalTextureMorph(
#else
long _ExternalTextureMorph(
#endif
 char *parameter_list, double mr){
 int i;
 double mx,my,mz;
 double mturb,mt1,mt2,mc[3];
 parameter_list++;
 sscanf(parameter_list,"%f %f %f  %f  %f %f  %f %f %f",
         &mc[0],&mc[1],&mc[2],&mturb,&mt1,&mt2,&mx,&my,&mz);
 t1=mt1+(t1-mt1)*mr;
 t2=mt2+(t2-mt2)*mr;
 turb=(turb-mturb)*mr;
 for(i=0;i<3;i++)c[i]=mc[i]+(c[i]-mc[i])*mr;
 x=mx+(x-mx)*mr;
 y=my+(y-my)*mr;
 z=mz+(z-mz)*mr;
 return 1;
}

/* h is 0->360  s and v are 0-1 */
void Hsv2Rgb(double h, double s, double v,
             double *r, double *g, double *b){
 double rr,gg,bb,hh,ss,vv,maxcol,mincol;
 hh=(double)h; ss=(double)s; vv=(double)v;
 mincol=vv*(1.0-ss);
 if(hh <= 120.0){bb=mincol;
   if(hh <=  60){rr=vv; gg=mincol+hh*(vv-mincol)/(120.0-hh);}
   else         {gg=vv; rr=mincol+(120.0-hh)*(vv-mincol)/hh;}
 }
 else if( hh <= 240){rr=mincol;
   if(hh <= 180){gg=vv; bb=mincol+(hh-120.0)*(vv-mincol)/(240.0-hh);}
   else         {bb=vv; gg=mincol+(240.0-hh)*(vv-mincol)/(hh-120.0);}
 }
 else {gg=mincol;
   if(hh <= 300){bb=vv; rr=mincol+(hh-240.0)*(vv-mincol)/(360.0-hh);}
   else         {rr=vv; bb=mincol+(360.0-hh)*(vv-mincol)/(hh-240.0);}
 }
 *r = rr;
 *g = gg;
 *b = bb;
}

void Rgb2Hsv(double rr, double gg, double bb,  // rr etc. are (0-1)
             double *h, double *s, double *v){ // h (0-360)  s v (0-1)
 double maxcol,mincol;
 maxcol=max(rr,max(gg,bb));
 mincol=min(rr,min(gg,bb));
 *v=maxcol;
 if(maxcol == mincol){*s=0; *h=0;}
 else{
  *s=(maxcol-mincol)/maxcol;
  if(mincol == bb)
    *h = 120.0*(gg-mincol)/(rr+gg-2.0*mincol);
  else if(mincol == rr)
    *h = 120.0*(1.0+(bb-mincol)/(bb+gg-2.0*mincol));
  else
    *h = 120.0*(2.0+(rr-mincol)/(rr+bb-2.0*mincol));
 }
}

double hermite(double t, double tmin, double tmax){ // not a real hermite
 t=(t-tmin)/(tmax-tmin);
 return (t - t*t);
}

double sigmoid(double t, double tmin, double tmax){
 t=3.1415926*(2.0*(t-tmin)/(tmax-tmin));
 return 0.5*(1.0-cos(t));
}

double triangle(double value){
  register double offset;
  if (value >= 0.0) offset = value - floor(value);
  else offset = value - (-1.0 - floor(fabs(value)));
  if (offset >= 0.5) return (2.0 * (1.0 - offset));
  else return (2.0 * offset);
}

double costab(double value){
  if (value < 0.0)return 0.0;
  return 0.5*(1.0+cos((value-floor(value))*3.141592657*2));
}

double sintab(double value){
 if (value >= 0.0)
   return 0.5*(1.0+sin((value-floor(value))*3.141592657*2));
 return 0.0;
}

static double ffmod(double x){
 if(x < 0){
   double a;
   a=fmod(-x,1.0);
   return 1.0-a;
 }
 return fmod(x,1.0);
}

static double step(double a, double x){
 if(x < a)return 0.0;
 return 1.0;
}

static void mix(double *r, double *g, double *b, double mx,
                double r1, double g1, double b1,
                double r2, double g2, double b2){
 *r = mx*r1 + (1.0-mx)*r2;
 *g = mx*g1 + (1.0-mx)*g2;
 *b = mx*b1 + (1.0-mx)*b2;
}


#include "randno.c"

//long IX=9123,IY=8844,IZ=20846;
//void dSeedRandom(long i){
// srand(i);
// IX=rand(); IY=rand(); IZ=rand();
//}
//#define dSeedRandom(i)  srand(i)
//#define dRandom() ((double)rand() * 3.05185e-5)
//double dRandom(void){
//return (double)rand() * 3.05185e-5;
// IX=IMOD(171*IX,30269);
// IY=IMOD(172*IY,20207);
// IZ=IMOD(170*IZ,30323);
// return (double)fmod((double)IX/30269.0+
//                     (double)IY/30307.0+
//                     (double)IZ/30323.0,1.0);
//}

static long ip[27][3]={
   { 0, 0, 0},
   { 1, 0, 0},
   { 1, 1, 0},
   { 0, 1, 0},
   {-1, 1, 0},
   {-1, 0, 0},
   {-1,-1, 0},
   { 0,-1, 0},
   { 1,-1, 0},
   { 0, 0,-1},
   { 1, 0,-1},
   { 1, 1,-1},
   { 0, 1,-1},
   {-1, 1,-1},
   {-1, 0,-1},
   {-1,-1,-1},
   { 0,-1,-1},
   { 1,-1,-1},
   { 0, 0, 1},
   { 1, 0, 1},
   { 1, 1, 1},
   { 0, 1, 1},
   {-1, 1, 1},
   {-1, 0, 1},
   {-1,-1, 1},
   { 0,-1, 1},
   { 1,-1, 1}
 };

#if 0
long CALLBACK ExternalTextureProcedure(
#else
long _ExternalTextureProcedure(
#endif
  long coord_type,  vector p, vector n,
  double alpha, double beta, double gamma,
  double bump,  double map_x,  double map_y,
  double *alpha_channel, unsigned char sc[3], double colour[3],
  double *reflectivity, double *transparency,
  X__SHADER *lpEVI
){

#if 0
// 2D bricks
 long i;
 double a,cc,xx,yy,zz,noiseH,r,g,b;
 vector u,v;
 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;
{
  double ss,tt,tbrick,sbrick,w,h,mf=0.03;
  ss = xx / 1.0; tt= yy/ 1.0;
  if(ffmod(tt*0.5) > 0.5)ss += 0.5;
  tbrick=floor(tt);
  ss += 0.3*(nNoise(0.0,tbrick*3,0.0) - 0.5);
  sbrick=floor(ss);
  ss -= sbrick;
  tt -= tbrick;
  vNoise(v,10*xx,10*yy,0.5);
  ss += (v[0]-0.5)*0.02;
  tt += (v[1]-0.5)*0.02;
  w=step(mf,ss) - step(1-mf,ss);
  h=step(mf,tt) - step(1-mf,tt);
  mix(&r,&g,&b,w*h,
      (double)colour[0]/255.0,(double)colour[1]/255.0,(double)colour[2]/255.0,
      (double)c[0]/255.0,(double)c[1]/255.0,(double)c[2]/255.0);
  if(w*h == 0){
    if(w == 0){
      if(ss < mf)  VECCOPY(-(1.0-ss/mf)*tAxis_unorm,v)
      if(ss > 1-mf)VECCOPY(((ss-(1.0-mf))/mf)*tAxis_unorm,v)
    }
    else{
      if(tt < mf)  VECCOPY((1.0-tt/mf)*tAxis_vnorm,v)
      if(tt > 1-mf)VECCOPY(-((tt-(1.0-mf))/mf)*tAxis_vnorm,v)
//      VECCOPY(tAxis_vnorm,v)
    }
    VECSUM(n,0.5*v,n)
    vNoise(u,40*xx,40*yy,0.5);
    VECSUM(0.2*u,n,n)
    normalize(n);
  }
  else{
    vNoise(v,2*xx,2*yy,0.5);
    VECSCALE(0.2,v,v)
    VECSUM(n,v,n)
    vNoise(v,20*xx,20*yy,0.5);
    VECSCALE(0.1,v,v)
    VECSUM(n,v,n)
    if(ss > 1 - 2*mf)VECSUM(0.3*tAxis_unorm,n,n)
    else if(ss < 2*mf)VECSUM(-0.3*tAxis_unorm,n,n)
    else if(tt < 2*mf)VECSUM(0.3*tAxis_vnorm,n,n)
    else if(tt > 1 - 2*mf)VECSUM(-0.3*tAxis_vnorm,n,n)
    normalize(n);
    noiseH=nNoise(xx*5,yy*5,0.5);
    r -= (noiseH*0.05);
    g -= (noiseH*0.05);
    b -= (noiseH*0.05);
  }
  r=max(r,0.0); g=max(g,0.0); b=max(b,0.0);
}
 colour[0]  = r * 255.0;
 colour[1]  = g * 255.0;
 colour[2]  = b * 255.0;
#endif


#if 0
 long i;
 double a,cc,xx,yy,zz,noiseH,r,g,b;
 vector u,v,BozoTurbulence;
 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;
// interesting pattern
 noiseH = nNoise(xx, yy, zz);
 noiseH = sintab(noiseH*10.0);
 if(noiseH > 0.1){
   noiseH=(noiseH-0.1)/(1.0-0.1);
   r=(noiseH*(double)c[0]+(1.0-noiseH)*(double)colour[0]);
   g=(noiseH*(double)c[1]+(1.0-noiseH)*(double)colour[1]);
   b=(noiseH*(double)c[2]+(1.0-noiseH)*(double)colour[2]);
   r=r/255.0;
   g=g/255.0;
   b=b/255.0;
 }
 else{
  r=colour[0]/255.0;
  g=colour[1]/255.0;
  b=colour[2]/255.0;
 }
 colour[0]  = r * 255.0;
 colour[1]  = g * 255.0;
 colour[2]  = b * 255.0;
#endif

#if 0
// stars  (some odd behaviour - don't change parameters)
{double cellsize;
 int ncells=1,
     npoints=8;
 double rmax=4.0,   // proportion of size of corners
        rmin=0.5;   // rmin is radius of star;
 double rr,ss,tt,scell,tcell,sctr,tctr,sa,angle;
 vector p0,p1,d0,d1,dr;
 cellsize=1.0/ncells;
 ss=yy*ncells; tt=xx*ncells;
 sctr=0.5; tctr=0.5;
 scell=FLOOR(ss); tcell=FLOOR(tt);
 sctr=scell+0.5; tctr=tcell+0.5;
// sctr=cellsize*(scell + 0.5 +
//  0.6 * nnNoise(scell+0.5,tcell+0.5,0.5));
// tctr=cellsize*(tcell + 0.5 +
//  0.6 * nnNoise(scell+3.5,tcell+8.5,0.5));
 ss=ss*cellsize-sctr;
 tt=tt*cellsize-tctr;

 sa=2*3.1415926/npoints;
 p0[0]=rmax; p0[1]=0.0; p0[2]=0.0;
 p1[0]=rmin*cos(sa/2); p1[1]=rmin*sin(sa/2); p1[2]=0.0;
// ss=yy-sctr; tt=xx-tctr;
 angle=atan2(ss,tt)+3.1415926;
 rr=sqrt(ss*ss+tt*tt);
 angle /= sa;
 angle -= floor(angle);
 if(angle < 0.5)angle = 1-angle;
 VECSUB(p1,p0,d0)
 d1[0]=rr*cos(angle); d1[1]=rr*sin(angle); d1[2]=0.0;
 VECSUB(d1,p0,d1)
 CROSS(d0,d1,dr)
 mix(&r,&g,&b,step(0,dr[2]),
     (double)c[0]/255.0,
     (double)c[1]/255.0,
     (double)c[2]/255.0,
     (double)colour[0]/255.0,
     (double)colour[1]/255.0,
     (double)colour[2]/255.0
         );
//#endif
}
 colour[0]  = r * 255.0;
 colour[1]  = g * 255.0;
 colour[2]  = b * 255.0;
#endif

#if 0
 // randon pimples, blisters, moon, peas, colloieflower  crusted muscle
 long i;
 double a,cc,xx,yy,zz,noiseH,r,g,b;
 vector u,v,BozoTurbulence;
 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;
 {
 static vector    vclr[81]; // N=3
 static double  dpoint[81]; //
 static double dradius[81];
 static vector pp[81];
 double dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long ix,iy,iz;
 BOOL ins;
 long j,jid,id,pattern,patterni;

 long N=1;
 double bradius=0.5,lradius=0.01;  // for blisters - randon pimples N=1
// double bradius=0.85,lradius=0.1;  // for peas with N=3
// double bradius=1.2,lradius=1.0;  // alternate peas with N=1
// double bradius=0.4,lradius=0.3;  // moon N=1
 double len,rad;
 vector ppc;

 xx *= 4.0;
 yy *= 4.0;
 zz *= 4.0;

// vNoise(v,xx,yy,zz);   // crusted
// xx += 0.7*v[0];
// yy += 0.7*v[1];
// zz += 0.7*v[2];

 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(double)ix;
 cy=yy-(double)iy;
 cz=zz-(double)iz;
 zb=-10.0;
 ins=TRUE;
 id = -1;
 len=sqrt(tAxis_u[0]*tAxis_u[0]+
          tAxis_u[1]*tAxis_u[1]+
          tAxis_u[2]*tAxis_u[2]);
 VECCOPY(-tAxis_n,u)
 VECSCALE(len,u,u)         // third axis
 VECCOPY(tAxis_o,v)
 VECSUM(v,cx*tAxis_u,v)
 VECSUM(v,cy*tAxis_v,v)
 VECSUM(v,cz*u,v)         // coord of point on surface
 for(j=0;j<27;j++){
 pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
 pattern=IMOD(N*pattern,1900);
 for(i=0;i<N;i++){
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((double)ip[j],rand_v[patterni],ppc)
   VECSUM(tAxis_o,ppc[0]*tAxis_u,pp[jid])
   VECSUM(pp[jid],ppc[1]*tAxis_v,pp[jid])
   VECSUM(pp[jid],ppc[2]*u,      pp[jid])

   VECCOPY(rand_v[patterni],vclr[jid])
   radius=(bradius+lradius*(rand_v[patterni][1]-0.5))*len;
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius){
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb){
       id=jid;
       zb=dc;
     }
   }
 }
 }
 if(id >= 0 && ins){
   r=colour[0]/255.0;
   g=colour[1]/255.0;
   b=colour[2]/255.0;
   if(1){//zb < 0.2){
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)          //(v-pp) vector from centre of pattern
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad){
       double mu,moddn,rmu;
       vector c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu){
         a=1.0-moddn/rmu;
         normalize(dn);
//         VECSCALE(-1.0,dn,dn);  // dimples

         a=sin(a*1.57079);      // rounded
//         if(a > 0.5)a=1;    // flat centre  ( a= 1 at centre  0 at outside

//a=a*a;  // crusted // with small xx yy (peas)
//           a=sqrt(a);    // orange peel with small xx yy zz  (peas)
#if 0
         if(a < 0.06){ // moon
           a=0.1;  // moon
         }
         else{
           a=1;    // moon
           noiseH=nNoise(xx*5,yy*5,zz*5)-0.5; // moon
           r -= (0.18+(noiseH*0.11));
           g -= (0.18+(noiseH*0.11));
           b -= (0.18+(noiseH*0.11));
           r=max(r,0.0); g=max(g,0.0); b=max(b,0.0);
         } // moon
#endif
// a=1.0-(1.0-a)*0.5;  // blisters (reduced bump size)
         VECSUM((1.0-a)*dn,(a)*n,n)
//vNoise(v,xx*5,yy*5,zz*5);   // crusted
//VECSUM(0.8*v,n,n);          // crusted

         normalize(n);

//         c[0]/255.0;   // coloured
//         c[1]/255.0;
//         c[2]/255.0;
       }
     }
   }
 }
 else{
   r=colour[0]/255.0;
   g=colour[1]/255.0;
   b=colour[2]/255.0;

// noiseH=nNoise(xx*2.3,yy*2.3,zz*2.3); // moon
// r -= (noiseH*0.15);
// g -= (noiseH*0.15);
// b -= (noiseH*0.15);
// r=max(r,0.0); g=max(g,0.0); b=max(b,0.0); //moon
 }
 }
 colour[0]  = r * 255.0;
 colour[1]  = g * 255.0;
 colour[2]  = b * 255.0;
#endif

#if 0
 // leopard skin  with black spots  - under test
 vector pp[22]={{0.25,0.0,      0.0},                     //0
                {0.75,0.0,      0.0},
                {0.0, ROOT32/2, 0.0},
                {0.5, ROOT32/2, 0.0},
                {1.0, ROOT32/2, 0.0},
                {0.25,ROOT32,   0.0},
                {0.75,ROOT32,   0.0},
                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},  //7
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0}, //12
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};                 //21
 double dx,dy,dz,zb;
 double xr,xl,yr,yl,zr,zl;
 long i;
 double a,cc,xx,yy,zz,r,g,b;
 vector u,v;
 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;
 xx += 0.11*sin(FMOD(3*zz,2*3.1415926));
 xx += 0.2*sin(FMOD(0.1*yy,2*3.1415926));
 yy += 0.15*cos(FMOD(4*xx,2*3.1415926));
 yy += 0.2*sin(FMOD(0.1*zz,2*3.1415926));
 zz += 0.18*cos(FMOD(5*yy,2*3.1415926));
 zz += 0.2*sin(FMOD(0.1*xx,2*3.1415926));
 cc=0.25; // just spots
 yy /= ROOT32;
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=0.0;
 r=colour[0];
 g=colour[1];
 b=colour[2];
 for(i=0;i<22;i++){
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc){
     if((cc-a) > zb){
       zb=(cc-a);
       r=c[0];
       g=c[1];
       b=c[2];
       if(a > 0.20){
         mix(&r,&g,&b,(a-0.20)/(cc-0.20),
          (double)colour[0],
          (double)colour[1],
          (double)colour[2],
          r,g,b);
       }
     }
   }
 }
 colour[0]  = r;
 colour[1]  = g;
 colour[2]  = b;

 {
 long k;
 double xm[]={-1, 1,-1, 1,-1, 1,-1, 1,-1, 1, 0, 0, 0, 0},
        ym[]={-1,-1, 1, 1,-1,-1, 1, 1, 0, 0,-1, 1, 0, 0},
        zm[]={-1,-1,-1,-1, 1, 1, 1, 1, 0, 0, 0, 0,-1, 1};
 double delta,xxx,yyy,zzz;
 cc=0.14;
 cc=0.09; delta=0.12;
 xxx=xx; yyy=yy; zzz=zz;
 for(k=0;k<14;k++){
 zb=0.0;
 r=colour[0];
 g=colour[1];
 b=colour[2];
 xx = xxx+delta*xm[k]; yy = yyy+delta*ym[k]; zz = zzz+delta*zm[k];
 for(i=0;i<22;i++){
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if(fabs(dx) > cc)continue;
   if(fabs(dy) > cc)continue;
   if(fabs(dz) > cc)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc){
     if((cc-a) > zb){
       zb=(cc-a);
       r=96;
       g=96;
       b=96;
       if(a > 0.01){
         mix(&r,&g,&b,(a-0.01)/(cc-0.01),
          (double)colour[0],
          (double)colour[1],
          (double)colour[2],
          r,g,b);
       }
     }
   }
 }
 colour[0]  = r;
 colour[1]  = g;
 colour[2]  = b;
 }
 }
#endif


 return 1;
}

#if 0
void CALLBACK ExternalTextureClose(void){
#else
void _ExternalTextureClose(void){
#endif
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static void SetColour(double *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(double)GetRValue(cc.rgbResult);
   colour[1]=(double)GetGValue(cc.rgbResult);
   colour[2]=(double)GetBValue(cc.rgbResult);
 }
}

BOOL CALLBACK BozoDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

#if 0
char * CALLBACK SetExternalParameters(
#else
char * _SetExternalParameters(
#endif
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                             ){
 char szbuf[255],*Op1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   Op1=Op;
   Op1++;
   sscanf(Op1,"%f %f %f  %f  %f %f  %f %f %f %ld %ld %ld %ld",
          &c[0],&c[1],&c[2],&turb,&t1,&t2,&x,&y,&z,
          &tr_type,&tr_type1,&tr_type2,&tr_type3);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_BOZO),hWnd,
              (DLGPROC)BozoDlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(szbuf,"# %.2f %.2f %.2f  %.2f  %.2f %.2f  %.2f %.2f %.2f %ld %ld %ld %ld",
          c[0],c[1],c[2],turb,t1,t2,x,y,z,tr_type,tr_type1,tr_type2,tr_type3);
 if((Op=(char *)CALL_MALLOC(strlen(szbuf)+1)) == NULL){
  MessageBox (GetFocus(),"External shader: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,szbuf);
 return Op;
}

BOOL CALLBACK BozoDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.2f",turb);
     SendDlgItemMessage(hwnd,DLG_BOZO_TURB,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_TURB,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",t1);
     SendDlgItemMessage(hwnd,DLG_BOZO_T1,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_T1,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",t2);
     SendDlgItemMessage(hwnd,DLG_BOZO_T2,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_T2,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",x);
     SendDlgItemMessage(hwnd,DLG_BOZO_XSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_XSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",y);
     SendDlgItemMessage(hwnd,DLG_BOZO_YSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_YSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",z);
     SendDlgItemMessage(hwnd,DLG_BOZO_ZSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_ZSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     if(tr_type == 0)
        SendDlgItemMessage(hwnd,DLG_BOZO_SEMI,BM_SETCHECK,TRUE,0);
     else
        SendDlgItemMessage(hwnd,DLG_BOZO_SETT,BM_SETCHECK,TRUE,0);
     if(tr_type1 == 1)
        SendDlgItemMessage(hwnd,DLG_BOZO_CLEARC,BM_SETCHECK,TRUE,0);
     if(tr_type2 == 1)
        SendDlgItemMessage(hwnd,DLG_BOZO_CLEARS,BM_SETCHECK,TRUE,0);
     if(tr_type3 == 1)
        SendDlgItemMessage(hwnd,DLG_BOZO_CLEAR3,BM_SETCHECK,TRUE,0);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_BOZO_COLOUR){
         r=(BYTE)c[0]; g=(BYTE)c[1]; b=(BYTE)c[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_BOZO_COLOUR:
          SetColour(c,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_BOZO_COLOUR),NULL,FALSE);
          break;
        case IDCANCEL:
        case DLG_BOZO_CANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
        case DLG_BOZO_OK:
          if(GetDlgItemText(hwnd,DLG_BOZO_TURB,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((turb=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_T1,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((t1=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_T2,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((t2=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_XSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((x=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_YSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((y=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_ZSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((z=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(SendDlgItemMessage(hwnd,DLG_BOZO_SETT,BM_GETCHECK,0,0))
            tr_type=1; else tr_type=0;
          if(SendDlgItemMessage(hwnd,DLG_BOZO_CLEARC,BM_GETCHECK,0,0))
            tr_type1=1; else tr_type1=0;
          if(SendDlgItemMessage(hwnd,DLG_BOZO_CLEARS,BM_GETCHECK,0,0))
            tr_type2=1; else tr_type2=0;
          if(SendDlgItemMessage(hwnd,DLG_BOZO_CLEAR3,BM_GETCHECK,0,0))
            tr_type3=1; else tr_type3=0;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}



