/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

// test1.c

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <windows.h>

#if __ZTC__ || __SC__
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#endif

#if __WATCOMC__
#define PI 3.1415926
#endif

#if __X__MIPS__
BOOL WINAPI _CRT_INIT(HINSTANCE ,DWORD , LPVOID );
#endif

static HINSTANCE hDLLinstance=NULL;

#include "..\animate\memory.h" /* for memory definition */
#include "..\animate\memdef.h" /* local names           */
#include "defines.h"
#include "rstruct.h"
#include "test.h"

#if __WATCOMC__
int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#else
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
#endif
  HANDLE ghMod;
  switch (dwReason) {
    case DLL_PROCESS_ATTACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      hDLLinstance = hDLL;  /* handle to DLL file */
      break;
    case DLL_PROCESS_DETACH:
#if __X__MIPS__
      if(!_CRT_INIT(hDLL,dwReason,lpReserved))return(int)FALSE;
#endif
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(DllMain)
#endif

#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))
#define ROOT32 0.866025403
#define ROOT3   1.73205080

static double turb=0.1,t1=0.4,t2=0.7,c[3]={255.0,255.0,0.0};
static double x=1.0,y=1.0,z=1.0;
static long   tr_type=0,tr_type1=0,tr_type2=0,tr_type3=0;

#if 0
long CALLBACK ExternalTextureStartup(
#else
long _ExternalTextureStartup(
#endif
  long frame,long nframes,char *parameter_list){
 if(parameter_list[0] != '#'){
   MessageBox ( GetFocus(),
                (LPCTSTR) "External texture: Startup",
                (LPCTSTR) "Parameter list missing",
                MB_OK | MB_SYSTEMMODAL );
 }
 else {
  parameter_list++;
  sscanf(parameter_list,"%f %f %f  %f  %f %f  %f %f %f %ld %ld %ld %ld",
         &c[0],&c[1],&c[2],&turb,&t1,&t2,&x,&y,&z,
         &tr_type,&tr_type1,&tr_type2,&tr_type3);
 }
 return 1;
}

#if 0
long CALLBACK ExternalTextureMorph(
#else
long _ExternalTextureMorph(
#endif
 char *parameter_list, double mr){
 int i;
 double mx,my,mz;
 double mturb,mt1,mt2,mc[3];
 parameter_list++;
 sscanf(parameter_list,"%f %f %f  %f  %f %f  %f %f %f",
         &mc[0],&mc[1],&mc[2],&mturb,&mt1,&mt2,&mx,&my,&mz);
 t1=mt1+(t1-mt1)*mr;
 t2=mt2+(t2-mt2)*mr;
 turb=(turb-mturb)*mr;
 for(i=0;i<3;i++)c[i]=mc[i]+(c[i]-mc[i])*mr;
 x=mx+(x-mx)*mr;
 y=my+(y-my)*mr;
 z=mz+(z-mz)*mr;
 return 1;
}

/* h is 0->360  s and v are 0-1 */
void Hsv2Rgb(double h, double s, double v,
             double *r, double *g, double *b){
 double rr,gg,bb,hh,ss,vv,maxcol,mincol;
 hh=(double)h; ss=(double)s; vv=(double)v;
 mincol=vv*(1.0-ss);
 if(hh <= 120.0){bb=mincol;
   if(hh <=  60){rr=vv; gg=mincol+hh*(vv-mincol)/(120.0-hh);}
   else         {gg=vv; rr=mincol+(120.0-hh)*(vv-mincol)/hh;}
 }
 else if( hh <= 240){rr=mincol;
   if(hh <= 180){gg=vv; bb=mincol+(hh-120.0)*(vv-mincol)/(240.0-hh);}
   else         {bb=vv; gg=mincol+(240.0-hh)*(vv-mincol)/(hh-120.0);}
 }
 else {gg=mincol;
   if(hh <= 300){bb=vv; rr=mincol+(hh-240.0)*(vv-mincol)/(360.0-hh);}
   else         {rr=vv; bb=mincol+(360.0-hh)*(vv-mincol)/(hh-240.0);}
 }
 *r = rr;
 *g = gg;
 *b = bb;
}

double hermite(double t, double tmin, double tmax){ // not a real hermite
 t=(t-tmin)/(tmax-tmin);
 return (t - t*t);
}

double sigmoid(double t, double tmin, double tmax){
 t=3.1415926*(2.0*(t-tmin)/(tmax-tmin));
 return 0.5*(1.0-cos(t));
}

double triangle(double value){
  register double offset;
  if (value >= 0.0) offset = value - floor(value);
  else offset = value - (-1.0 - floor(fabs(value)));
  if (offset >= 0.5) return (2.0 * (1.0 - offset));
  else return (2.0 * offset);
}

double costab(double value){
  if (value < 0.0)return 0.0;
  return 0.5*(1.0+cos((value-floor(value))*3.141592657*2));
}

double sintab(double value){
 if (value >= 0.0)
   return 0.5*(1.0+sin((value-floor(value))*3.141592657*2));
 return 0.0;
}

static double ffmod(double x){
 if(x < 0){
   double a;
   a=fmod(-x,1.0);
   return 1.0-a;
 }
 return fmod(x,1.0);
}

static double step(double a, double x){
 if(x < a)return 0.0;
 return 1.0;
}

static void mix(double *r, double *g, double *b, double mx,
                double r1, double g1, double b1,
                double r2, double g2, double b2){
 *r = mx*r1 + (1.0-mx)*r2;
 *g = mx*g1 + (1.0-mx)*g2;
 *b = mx*b1 + (1.0-mx)*b2;
}

#if 0
long CALLBACK ExternalTextureProcedure(
#else
long _ExternalTextureProcedure(
#endif
  long coord_type,  vector p, vector n,
  double alpha, double beta, double gamma,
  double bump,  double map_x,  double map_y,
  double *alpha_channel, unsigned char sc[3], double colour[3],
  double *reflectivity, double *transparency,
  X__SHADER *lpEVI
){
 long i;
 double a,cc,xx,yy,zz,noiseH,r,g,b;
 vector u,v,BozoTurbulence;
 xx=alpha / x;
 yy=beta  / y;
 zz=gamma / z;


#if 0
 // soft spheres  CCP
 {
 vector pp[22]={{0.25,0.0,      0.0},                     //0
                {0.75,0.0,      0.0},
                {0.0, ROOT32/2, 0.0},
                {0.5, ROOT32/2, 0.0},
                {1.0, ROOT32/2, 0.0},
                {0.25,ROOT32,   0.0},
                {0.75,ROOT32,   0.0},
                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},  //7
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0}, //12
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};                 //21
 vector dn;
 double dx,dy,dz,cr,cg,cb,zb,noise;
// cc=0.28867;  // exact hexagon
// cc=0.3;  // tight hexagon
 cc=0.25;      // sphere
 nNoise(xx,yy,zz,&noise)
 Hsv2Rgb(noise*360.0,1.0,1.0,&cr,&cg,&cb);  // modulate color
                                                       // frequency
 yy /= ROOT32;
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=0.0;
 VECCOPY(n,dn)
 r=colour[0]/255.0;
 g=colour[1]/255.0;
 b=colour[2]/255.0;
 for(i=0;i<22;i++){
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc){
     if((cc-a) > zb){
       zb=(cc-a);
       r=cr;
       g=cg;
       b=cb;
       if(a > 0.20){
         mix(&r,&g,&b,(a-0.20)/(cc-0.20),
          (double)colour[0]/255.0,
          (double)colour[1]/255.0,
          (double)colour[2]/255.0,
          r,g,b);
       }
     }
   }
 }
 }
#endif

#if 0
 // hexagonal pattern (curvy edges flagstones)
 {
 vector pp[35]={{0.25,0.0,      0.0},                     //0
                {0.75,0.0,      0.0},                     //1
                {0.0, ROOT32/2, 0.0},                     //2
                {0.5, ROOT32/2, 0.0},                     //3
                {1.0, ROOT32/2, 0.0},                     //4
                {0.25,ROOT32,   0.0},                     //5
                {0.75,ROOT32,   0.0},                     //6

                {-0.25,0.0,     0.0},                     //7
                { 1.25,0.0,     0.0},                     //8
                {-0.25,ROOT32,  0.0},                     //9
                { 1.25,ROOT32,  0.0},                     //10
                {-0.25,0.0,    ROOT32},                   //11
                { 1.25,0.0,    ROOT32},                   //12
                {-0.25,ROOT32, ROOT32},                   //13
                { 1.25,ROOT32, ROOT32},                   //14
                {-0.25,1.0/2.0/ROOT3,ROOT32/2.0},         //15
                { 1.75,1.0/2.0/ROOT3,ROOT32/2.0},         //16
                {0.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},    //17
                {0.5,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},    //18
                {1.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},    //19

                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0}, //20
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0}, //21
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0}, //22
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0}, //23
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0}, //24
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0}, //25
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0}, //26
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0}, //27
                {0.25,0.0,      ROOT32},                  //28
                {0.75,0.0,      ROOT32},                  //29
                {0.0, ROOT32/2, ROOT32},                  //30
                {0.5, ROOT32/2, ROOT32},                  //31
                {1.0, ROOT32/2, ROOT32},                  //32
                {0.25,ROOT32,   ROOT32},                  //33
                {0.75,ROOT32,   ROOT32}};                 //34
 double dx,dy,dz,zb,ccc,dc,zbb;
 double flat_edge=0.025;
 double bevel_edge=0.06;
 double blend_edge=0.06;
// double flat_edge=0.075; // cracked
// double bevel_edge=0.145;
// double blend_edge=0.13;
 BOOL ins;
 int id;
 cc=0.8;            // flagstone
 nNoise(xx*15.0,yy*15.0,zz*15.0,noiseH);
 vNoise(v,xx,yy,zz);
 xx += 0.8*v[0];
 yy += 0.8*v[1];
 zz += 0.8*v[2];
 yy /= ROOT32;
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=-10.0;
 ins=TRUE;
 for(i=0;i<35;i++){
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc){
     dc=(cc-a);
     if(dc > zb){
       id=i;
       zb=dc;
     }
   }
 }
 zbb= -10.0;
 for(i=0;i<35;i++){
   if(i == id)continue;
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc){
     dc=(cc-a);
     if(dc > zbb){
       zbb=dc;
     }
   }
 }
 if((ccc=fabs(zbb - zb)) < flat_edge)ins=FALSE;   // flat edge
 if(ins){
   if(id == 3 || id == 31 || id == 23 || id == 24){
     r=sc[0]/255.0;
     g=sc[1]/255.0;
     b=sc[2]/255.0;
   }
   else{
     r=colour[0]/255.0;
     g=colour[1]/255.0;
     b=colour[2]/255.0;
   }
   r -= (noiseH*0.25);
   g -= (noiseH*0.25);
   b -= (noiseH*0.25);
   r=max(r,0.0); g=max(g,0.0); b=max(b,0.0);
   if(ccc < bevel_edge){
     v[0]=(xx-pp[id][0]);
     v[1]=(yy-pp[id][1]);
     v[2]=(zz-pp[id][2]);
     normalize(v);
     CROSS(v,n,u)
     CROSS(n,u,v)
     normalize(v);
     a=0.4;
     VECSCALE(a,v,v)
     VECSCALE((1.0-a),n,n)
     VECSUM(v,n,n)
     normalize(n);
   }
   if(ccc < blend_edge)mix(&r,&g,&b,(ccc-flat_edge)/(blend_edge-flat_edge),
         r,g,b,
         (double)c[0]/255.0,
         (double)c[1]/255.0,
         (double)c[2]/255.0
         );
 }
 else{
   r=c[0]/255.0;
   g=c[1]/255.0;
   b=c[2]/255.0;
 }

 }
#endif

#if 0
// interesting pattern
 nNoise(xx, yy, zz, &noiseH);
 noiseH = sintab(noiseH*10.0);
 if(noiseH > 0.1){
   noiseH=(noiseH-0.1)/(1.0-0.1);
   r=(noiseH*(double)c[0]+(1.0-noiseH)*(double)colour[0]);
   g=(noiseH*(double)c[1]+(1.0-noiseH)*(double)colour[1]);
   b=(noiseH*(double)c[2]+(1.0-noiseH)*(double)colour[2]);
   r=r/255.0;
   g=g/255.0;
   b=b/255.0;
 }
 else{
  r=colour[0]/255.0;
  g=colour[1]/255.0;
  b=colour[2]/255.0;
 }
#endif

 colour[0]  = r * 255.0;
 colour[1]  = g * 255.0;
 colour[2]  = b * 255.0;
 return 1;
}

#if 0
void CALLBACK ExternalTextureClose(void){
#else
void _ExternalTextureClose(void){
#endif
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static void SetColour(double *colour, HWND parent){
 CHOOSECOLOR cc;
 static COLORREF CustColours[16]={
   RGB(255,255,255), RGB(239,239,239), RGB(223,223,223), RGB(207,207,207),
   RGB(191,191,191), RGB(175,175,175), RGB(159,159,159), RGB(143,143,143),
   RGB(127,127,127), RGB(111,111,111), RGB( 95, 95, 95), RGB( 79, 79, 79),
   RGB( 63, 63, 63), RGB( 47, 47, 47), RGB( 31, 31, 31), RGB( 15, 15, 15) };
 cc.lStructSize=sizeof(CHOOSECOLOR);
 cc.hwndOwner=parent;
 cc.rgbResult=RGB((BYTE)colour[0],(BYTE)colour[1],(BYTE)colour[2]);
 cc.lpCustColors=(LPDWORD)CustColours;
// cc.Flags=CC_FULLOPEN | CC_RGBINIT | CC_ENABLEHOOK;
 cc.Flags= CC_RGBINIT;
 cc.lCustData=(DWORD)0;
 if(ChooseColor(&cc)){
   colour[0]=(double)GetRValue(cc.rgbResult);
   colour[1]=(double)GetGValue(cc.rgbResult);
   colour[2]=(double)GetBValue(cc.rgbResult);
 }
}

BOOL CALLBACK BozoDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);

#if 0
char * CALLBACK SetExternalParameters(
#else
char * _SetExternalParameters(
#endif
  char *Op,                 /* string for the parameters                  */
  HWND hWnd,                /* parent window                              */
  X__MEMORY_MANAGER *lpEVI /* pointer to structure with memory functions */
                             ){
 char szbuf[255],*Op1;
 if(Op != NULL){  /* parameters exist so read them off the list */
   Op1=Op;
   Op1++;
   sscanf(Op1,"%f %f %f  %f  %f %f  %f %f %f %ld %ld %ld %ld",
          &c[0],&c[1],&c[2],&turb,&t1,&t2,&x,&y,&z,
          &tr_type,&tr_type1,&tr_type2,&tr_type3);
 }
 if(DialogBox(hDLLinstance,MAKEINTRESOURCE(DLG_BOZO),hWnd,
              (DLGPROC)BozoDlgProc) == FALSE)return Op;
 if(Op != NULL)CALL_FREE(Op);  /* free the old string */
 sprintf(szbuf,"# %.2f %.2f %.2f  %.2f  %.2f %.2f  %.2f %.2f %.2f %ld %ld %ld %ld",
          c[0],c[1],c[2],turb,t1,t2,x,y,z,tr_type,tr_type1,tr_type2,tr_type3);
 if((Op=(char *)CALL_MALLOC(strlen(szbuf)+1)) == NULL){
  MessageBox (GetFocus(),"External shader: Out of memory","Error",
                MB_OK|MB_TASKMODAL|MB_ICONSTOP);
   return NULL;
 }
 strcpy(Op,szbuf);
 return Op;
}

BOOL CALLBACK BozoDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 char str[16];
 switch( msg ) {
   case WM_INITDIALOG:
     sprintf(str,"%.2f",turb);
     SendDlgItemMessage(hwnd,DLG_BOZO_TURB,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_TURB,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",t1);
     SendDlgItemMessage(hwnd,DLG_BOZO_T1,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_T1,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",t2);
     SendDlgItemMessage(hwnd,DLG_BOZO_T2,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_T2,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",x);
     SendDlgItemMessage(hwnd,DLG_BOZO_XSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_XSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",y);
     SendDlgItemMessage(hwnd,DLG_BOZO_YSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_YSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     sprintf(str,"%.2f",z);
     SendDlgItemMessage(hwnd,DLG_BOZO_ZSCALE,WM_SETTEXT,0,(LPARAM)str);
     SendDlgItemMessage(hwnd,DLG_BOZO_ZSCALE,EM_LIMITTEXT,(WPARAM)12,0);
     if(tr_type == 0)
        SendDlgItemMessage(hwnd,DLG_BOZO_SEMI,BM_SETCHECK,TRUE,0);
     else
        SendDlgItemMessage(hwnd,DLG_BOZO_SETT,BM_SETCHECK,TRUE,0);
     if(tr_type1 == 1)
        SendDlgItemMessage(hwnd,DLG_BOZO_CLEARC,BM_SETCHECK,TRUE,0);
     if(tr_type2 == 1)
        SendDlgItemMessage(hwnd,DLG_BOZO_CLEARS,BM_SETCHECK,TRUE,0);
     if(tr_type3 == 1)
        SendDlgItemMessage(hwnd,DLG_BOZO_CLEAR3,BM_SETCHECK,TRUE,0);
     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_DRAWITEM:{
       LPDRAWITEMSTRUCT lpdis;
       HBRUSH   hbr,hbrold;
       BYTE r,g,b;
       lpdis=(LPDRAWITEMSTRUCT)lparam;
       if(lpdis->CtlID == DLG_BOZO_COLOUR){
         r=(BYTE)c[0]; g=(BYTE)c[1]; b=(BYTE)c[2];
         if(lpdis->itemState & ODS_SELECTED)
            InvertRect(lpdis->hDC,&(lpdis->rcItem));
         else{
           hbr=CreateSolidBrush(RGB(r,g,b));
           hbrold=SelectObject(lpdis->hDC,hbr);
           Rectangle(lpdis->hDC,lpdis->rcItem.left,lpdis->rcItem.top,
                         lpdis->rcItem.right,lpdis->rcItem.bottom);
           SelectObject(lpdis->hDC,hbrold);
           DeleteObject(hbr);
         }
       }
     }
     break;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
        case DLG_BOZO_COLOUR:
          SetColour(c,hwnd);
          InvalidateRect(GetDlgItem(hwnd,DLG_BOZO_COLOUR),NULL,FALSE);
          break;
        case IDCANCEL:
        case DLG_BOZO_CANCEL:
          EndDialog(hwnd,FALSE);
          return(TRUE);
        case IDOK:
        case DLG_BOZO_OK:
          if(GetDlgItemText(hwnd,DLG_BOZO_TURB,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((turb=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_T1,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((t1=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_T2,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((t2=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_XSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((x=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_YSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((y=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(GetDlgItemText(hwnd,DLG_BOZO_ZSCALE,str,10) == 0)
             EndDialog(hwnd,FALSE);
          if((z=atof(str)) == 0)EndDialog(hwnd,FALSE);
          if(SendDlgItemMessage(hwnd,DLG_BOZO_SETT,BM_GETCHECK,0,0))
            tr_type=1; else tr_type=0;
          if(SendDlgItemMessage(hwnd,DLG_BOZO_CLEARC,BM_GETCHECK,0,0))
            tr_type1=1; else tr_type1=0;
          if(SendDlgItemMessage(hwnd,DLG_BOZO_CLEARS,BM_GETCHECK,0,0))
            tr_type2=1; else tr_type2=0;
          if(SendDlgItemMessage(hwnd,DLG_BOZO_CLEAR3,BM_GETCHECK,0,0))
            tr_type3=1; else tr_type3=0;
          EndDialog(hwnd,TRUE);
          return(TRUE);
        default:
          break;
      }
      break;
    default: break;
 }
 return FALSE;
}

