/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* text_dos  a TSR to return an external texture.c  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <int.h>
#include <dos.h>
#include <math.h>

#define double float
#define INTNO 0x60
#define _FAR _far

typedef struct tfxinfo_tag{
 long tid;       /* the id of the texture starts at 16        */
 long frame;     /* the frame of the animation                */
 double alpha,   /* position of surface point relative to the */
        beta,    /* shader axes alpha/beta in plane of axes   */
        gamma;   /* gamma at right angles to plane            */
 double x,y;     /* the surface mapping coords - if available */
 double   R,G,B; /* returned RGB colours on scale if 0-255.0  */
 unsigned char fc[3], /* original colour of face              */
               sc[3]; /* secondary (shader) colour for face   */
 unsigned char pad1,pad2;
} tfxinfo;

ExternalTextureProcedure(unsigned short tid, tfxinfo _FAR *TFXinfo);

int cdecl tsr_install (void);
int cdecl tsr_uninstall ( void );

main(int argc, char **argv){
int i;
 if((strncmp(argv[1],"/R",2) ==0)  || (strncmp(argv[1],"/r",2)==0)){
   i=tsr_uninstall();
   if(i==0)printf("Program removed\n");
   if(i==2)printf("Can not remove, Program not loaded!\n");
   if(i==3)printf("Can not remove, Another program loaded above\n");
   exit(0);
 }
 i=tsr_install();
 /* if it returns, error has occured       */
 if(i==1)printf("Can not load, program already loaded!\n");
 else    printf("Failed to install, error %i\n",i);
}

extern int cdecl _datapar;
extern int cdecl _progpar;

int cdecl tsr_test_vect ( unsigned vno , unsigned dataseg );
int cdecl tsr_already_in ( void );
int cdecl tsr_new_interrupt ( struct INT_DATA* );

char tsr_fprint[20] = "tsrclock.v1";  /* unique string*/
volatile int cdecl tsr_running;
volatile char cdecl oldinterrupt[4];
volatile int cdecl oldpsp;
int cdecl tsr_vect_changed;
unsigned cdecl tsr_codeseg;
struct SREGS segregs;
union REGS regs;

tsr_install(void){
 unsigned code,data,stack,extra;
 if(tsr_already_in() !=0)return 1;
 oldpsp=_psp;
 peek(0x0,INTNO*4,&oldinterrupt,4);
 int_intercept(INTNO,tsr_new_interrupt,1500);
 regs.x.dx=_datapar;
 regs.x.dx=regs.x.dx+_progpar;
 regs.h.ah=0x31;
 intdos(&regs,&regs);
}

tsr_uninstall(void){
 char near *np, far *fp;
 int near *inp,far *ifp;
 unsigned cseg,pid,mlen;
 char mz;
 unsigned tsr_ds,tsr_off;
 if(tsr_running==0){
    tsr_ds=tsr_already_in();
    if(tsr_ds == 0)return 2;
 }
 else tsr_ds=getDS();
 tsr_vect_changed=0;
 tsr_test_vect(INTNO,tsr_ds);
 if(tsr_vect_changed!=0)return 3;
 if(tsr_running==0){
   np=&tsr_fprint[0];
   *np=0;
   fp=MK_FP(getDS(),(unsigned) np);
   tsr_off=FP_OFF(fp);
   poke(tsr_ds,tsr_off,&tsr_fprint,1);
   np=&oldinterrupt[0];
   fp=MK_FP(getDS(),(unsigned) np);
   tsr_off=FP_OFF(fp);
   peek(tsr_ds,tsr_off,&oldinterrupt,4);
   inp=&oldpsp;
   ifp=MK_FP(getDS(),(unsigned) inp);
   tsr_off=FP_OFF(ifp);
   peek(tsr_ds,tsr_off,&oldpsp,2);
 }
 int_off();
 poke(0x0,INTNO*4,&oldinterrupt,4);
 int_on();
 peek(0,0x00ba,&cseg,2);
 cseg-=1;
un1:
 peek(cseg,0,&mz,1);
 if(mz != 0x4d)goto un9;
 peek(cseg,1,&pid,2);
 peek(cseg,3,&mlen,2);
 cseg+=1;
 if(pid == oldpsp){
   segregs.es=cseg;
   regs.h.ah=0x49;
   intdosx(&regs,&regs,&segregs);
 }
 cseg+=mlen;
 goto un1;
un9:
 return 0;
}

tsr_test_vect(vno,dataseg) unsigned vno,dataseg; {
unsigned lseg,loff;
 segread(&segregs);
 tsr_codeseg=dataseg-(segregs.ds-segregs.cs);
 int_getvector(vno,&loff,&lseg);
 if((lseg == tsr_codeseg) || (lseg == dataseg)) return;
 else  tsr_vect_changed+=1;
}

tsr_already_in(void){
 int tsr_hit;
 char near *np, far *fp;
 char tsr_match[20];
 unsigned tsr_seg,tsr_off,tsr_ds;
 tsr_hit=0;
 np=&tsr_fprint[0];
 *np=1;
 fp=MK_FP(getDS(),(unsigned) np);
 tsr_seg=FP_SEG(fp);
 tsr_off=FP_OFF(fp);
 tsr_ds=0x600;
 while((tsr_ds < tsr_seg) && tsr_hit==0) {
   peek(tsr_ds,tsr_off,&tsr_match,19);
   if(strcmp(tsr_fprint,tsr_match) == 0)tsr_hit=1;
   else                                 tsr_ds+=1;
 }
 if(tsr_hit==1)return(tsr_ds);
 else          return 0;
}

tsr_new_interrupt(struct INT_DATA *p){
 tfxinfo _FAR *TFXinfo;
 unsigned short off,seg,tid;
 if(tsr_running)return 0;
 tsr_running=1;
 off=(unsigned short)p->regs.x.ax;
 seg=(unsigned short)p->regs.x.bx;
 tid=(unsigned short)p->regs.x.cx;
 TFXinfo=MK_FP(seg,off);
 ExternalTextureProcedure(tid,TFXinfo);
 tsr_running=0;
 return 1;
}


ExternalTextureProcedure(unsigned short tid, tfxinfo _FAR *TFXinfo){
 TFXinfo->R=255.0;
 TFXinfo->G=255.0;
 TFXinfo->B=0.0;
}
