/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* 3DVIEW.C */

#include <windows.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <vfw.h>

#include "resource.h"

#define MAX_FILE       255
#define MAX_DIR        255
#define PALVERSION     0x300
#define MAXPALETTE     256

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define ALIGNULONG(i)   ((i+3)/4*4)
#define EVI_Malloc(s)   LocalAlloc(LMEM_FIXED,s)
#define EVI_Free(p)     LocalFree((HLOCAL)p)

static BOOL Register3dView(HINSTANCE);
static void Unregister3dView(HINSTANCE);
static HBITMAP BitmapFromDib(LPSTR ,LPSTR , HPALETTE );
static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER );
static HWND AttachView3dView(void);
static LONG WINAPI WndProc(HWND, UINT, WPARAM, LPARAM);
static void Commands(WORD, HWND);
static int LoadGIF(int, char *);
static void AppPaint(HWND, HDC, int, int );
static void FreeBitmaps(void);
static BOOL MakeImageName(int, char *, HWND);
static int  GetCurrentImage();
static PSTR FileInPath(PSTR );
static BOOL CALLBACK AboutDlgProc(HWND, UINT, WPARAM, LPARAM );

static BOOL      stretch=TRUE,dibdirect=TRUE;
static HWND      glWnd=NULL;
static HINSTANCE glInstance;
static CHAR      szAppName[]="3D:VIEW";
static HPALETTE  ghpalOld, ghPalette = NULL;
static HPALETTE  hpalCurrent    = NULL;
static HDRAWDIB  hDrawDib       = NULL;
static LPSTR     hbmCurrent[1024];
static int       nImages=0,resID,nR,nU,sizeX,sizeY;
static DWORD
fdwStyle = WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
           WS_HSCROLL | WS_VSCROLL |
           WS_SYSMENU | WS_MAXIMIZEBOX |
           WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
static float up_angle=0.0,round_angle=0.0;
static char CurrentDirFile[256],CurrentDirectory[256],CurrentFile[256];
static unsigned char *pixels = NULL;

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 int i;
 if(Register3dView(hInstance)) {
   AttachView3dView();
   if (glWnd != NULL)DestroyWindow(glWnd);
   Unregister3dView(hInstance);
 }
 else MessageBox(NULL,"Register Failed",NULL,MB_OK);
 for(i=0;i<1024;i++)hbmCurrent[i]=NULL;
 return 0;
}

static void Unregister3dView(HINSTANCE hInstance){
  UnregisterClass(szAppName,hInstance);
}

static BOOL Register3dView(HINSTANCE hInstance){
  WNDCLASS   wc;
  wc.style         = 0;
  wc.lpfnWndProc   = (WNDPROC)WndProc;
  wc.cbClsExtra    = 0;
  wc.cbWndExtra    = 0;
  wc.hInstance     = hInstance;
  wc.hIcon         = LoadIcon(hInstance,"VIEW_ICON");
  wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
  wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
  wc.lpszMenuName  = "VIEW_MENU";
  wc.lpszClassName = szAppName;
  glInstance=hInstance;
  if (!RegisterClass (&wc) )return FALSE;
  return TRUE;
}

static HWND AttachView3dView(void){
  RECT rc;
  LPRECT pr;
  HWND hWnd;
  MSG msg;
  pr=&rc;
  rc.left=100; rc.top=50; rc.right=400; rc.bottom=420;
  hWnd = CreateWindow(szAppName,
            "3D Viewer",
            fdwStyle,
            pr->left,
            pr->top,
            pr->right,
            pr->bottom,
            NULL,
            NULL,
            glInstance,
            NULL);
  if (!hWnd)return NULL;
  glWnd=hWnd;
  ShowWindow (hWnd, SW_SHOW);
  UpdateWindow(hWnd);
  while(GetMessage(&msg,NULL,0,0)) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  return hWnd;
}

static LONG WINAPI WndProc(HWND hWnd,UINT uMsg,
                           WPARAM wParam,LPARAM lParam){
  float old;
  int i,iPos,iMax,iMin,dn;
  LONG    lRet = 1;
  static BOOL flag=FALSE;
  switch (uMsg) {
  case WM_CREATE:
    if((hDrawDib=DrawDibOpen()) == NULL)MessageBox(NULL,"No Dib",NULL,MB_OK);
    SetScrollRange(hWnd, SB_VERT, -89,89,TRUE);
    SetScrollRange(hWnd, SB_HORZ, -179,179,TRUE);
    SetScrollPos(hWnd,SB_HORZ,0,TRUE);
    SetScrollPos(hWnd,SB_VERT,0,TRUE);
    up_angle=0.0; round_angle=0.0;
    break;
  case WM_DESTROY:
    if(hDrawDib != NULL)DrawDibClose(hDrawDib);
    hDrawDib=NULL;
    break;
  case WM_PAINT:{
      HDC hDC;
      PAINTSTRUCT  ps;
      hDC = BeginPaint(hWnd, &ps);
      AppPaint(hWnd,hDC,0,0);
      EndPaint(hWnd, &ps);
    }
    break;
  case WM_VSCROLL:
    GetScrollRange(hWnd, SB_VERT, &iMin, &iMax);
    iPos = GetScrollPos(hWnd, SB_VERT);
    switch (LOWORD(wParam)) {
    case SB_ENDSCROLL:
      dn=0;
      InvalidateRect(hWnd, NULL, FALSE);
      flag=FALSE;
      break;
    case SB_LINEDOWN:      dn =  1;    break;
    case SB_LINEUP:        dn = -1;    break;
    case SB_PAGEDOWN:      dn =  5;    break;
    case SB_PAGEUP:        dn = -5;    break;
    case SB_THUMBTRACK:
    case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wParam)-iPos;    break;
    default:               dn = 0;
    }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      SetScrollPos(hWnd, SB_VERT, iPos + dn, TRUE);
      old=up_angle;
      up_angle=(float)(iPos+dn);
      if (old != up_angle) {
        if (!flag)InvalidateRect(hWnd, NULL, FALSE);
      }
    }
    break;
  case WM_HSCROLL:
    GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
    iPos = GetScrollPos(hWnd, SB_HORZ);
    switch (LOWORD(wParam)) {
    case SB_ENDSCROLL:
      dn=0;
      InvalidateRect(hWnd, NULL, FALSE);
      flag=FALSE;
      break;
    case SB_LINEDOWN:      dn =  1;     break;
    case SB_LINEUP:        dn = -1;     break;
    case SB_PAGEDOWN:      dn =  5;     break;
    case SB_PAGEUP:        dn = -5;     break;
    case SB_THUMBTRACK:
    case SB_THUMBPOSITION: dn = (int)(short)HIWORD(wParam) - iPos; break;
    default:               dn = 0;
    }
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      SetScrollPos(hWnd, SB_HORZ, iPos + dn, TRUE);
      old=round_angle;
      round_angle=(float)(iPos+dn);
      if (old != round_angle) {
        if (!flag)InvalidateRect(hWnd, NULL, FALSE);
      }
    }
    break;
  case WM_QUERYNEWPALETTE:{
      HDC     hDC;
      if (ghPalette != NULL) {
        hDC = GetDC(hWnd);
        ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
        RealizePalette(hDC);
        InvalidateRect(hWnd, NULL, TRUE);
        UpdateWindow(hWnd);
        if (ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
        ReleaseDC(hWnd, hDC);
        return TRUE;
      }
      return FALSE;
    }
  case WM_PALETTECHANGED:{
      HDC         hDC;
      if (ghPalette != NULL) {
        if (wParam != (WPARAM)hWnd) {
          hDC = GetDC(hWnd);
          ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
          RealizePalette(hDC);
          UpdateColors(hDC);
          if (ghpalOld != NULL)SelectPalette(hDC, ghpalOld, FALSE);
          ReleaseDC(hWnd, hDC);
        }
      }
    }
    break;
  case WM_SIZE:{
      RECT rc;
      GetClientRect(hWnd,&rc);
      sizeX=rc.right;
      sizeY=rc.bottom;
      InvalidateRect(hWnd, NULL, FALSE);
    }
    break;
  case WM_SYSCOMMAND:
    switch(LOWORD(wParam & 0xfff0)) {
    case SC_CLOSE:
      PostMessage(hWnd,WM_COMMAND,IDM_EXIT,0);
      break;
    default:
      return( DefWindowProc(hWnd,uMsg,wParam,lParam));
      break;
    }
    break;
  case WM_COMMAND:
    Commands(LOWORD(wParam),hWnd);
    break;
  default:
    lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
    break;
  }
  return lRet;
}


static void Commands(WORD cmd,HWND hWnd){
  DLGPROC dlgprc;
  int i;
  OPENFILENAME ofn;
  char szTitle[60]="Choose First Image in Set",szFilter[80];
  switch (cmd) {
  case IDM_OPEN:
    CurrentDirFile[0]='\0'; CurrentFile[0]='\0';
    GetCurrentDirectory(sizeof(CurrentDirectory),CurrentDirectory);
    strcpy(szFilter,"GIF file (*.gif)|*.gif|All Files (*.*)|*.*|");
    i=0; while(szFilter[i] != '\0') {
      if (szFilter[i] == '|')szFilter[i]='\0';
      i++;
    }
    memset(&ofn,0,sizeof(OPENFILENAME));
    ofn.lStructSize=sizeof(OPENFILENAME);
    ofn.hwndOwner=hWnd;
    ofn.lpstrFilter=szFilter;
    ofn.nFilterIndex=0;
    ofn.lpstrFile=CurrentDirFile;   /* will contain file + path */
    ofn.nMaxFile=sizeof(CurrentDirFile);
    ofn.lpstrFileTitle=CurrentFile;  /* file name only */
    ofn.nMaxFileTitle=sizeof(CurrentFile);
    ofn.lpstrInitialDir=CurrentDirectory;
    ofn.lpstrTitle=szTitle;
    ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
    if (GetOpenFileName(&ofn)) {
      char ImageName[256];
      FreeBitmaps();
      nImages=0; resID=0; nR=0; nU=0;
      while(1){
        strcpy(ImageName,CurrentDirFile);
        if(!MakeImageName(nImages+1,ImageName,hWnd))break;
        SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)ImageName);
        LoadGIF(nImages,ImageName);
        nImages++;
      }
      SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)"3D View");
      if     (nImages == 162){resID=0; nR=18; nU=9;}
      else if(nImages == 198){resID=1; nR=18; nU=11;}
      else if(nImages == 216){resID=2; nR=24; nU=9;}
      else if(nImages == 330){resID=3; nR=30; nU=11;}
      else if(nImages == 396){resID=4; nR=36; nU=11;}
      else {
        MessageBox(hWnd,"Bad Image Sequence",NULL,MB_OK);
        resID= -1;
      }
      SetScrollPos(hWnd,SB_HORZ,0,TRUE);
      SetScrollPos(hWnd,SB_VERT,0,TRUE);
      up_angle=0.0; round_angle=0.0;
      if(hbmCurrent[0] != NULL){
        RECT Rectangle;
        Rectangle.left = 0; Rectangle.top    = 0;
        Rectangle.right  = (WORD)((LPBITMAPINFOHEADER)hbmCurrent[0])->biWidth;
        Rectangle.bottom = (WORD)((LPBITMAPINFOHEADER)hbmCurrent[0])->biHeight;
        Rectangle.right  += 2*GetSystemMetrics(SM_CXFRAME);
        Rectangle.bottom += 2*GetSystemMetrics(SM_CYFRAME);
        Rectangle.right  += GetSystemMetrics(SM_CXHSCROLL);
        Rectangle.bottom += GetSystemMetrics(SM_CYHSCROLL);
        Rectangle.bottom += GetSystemMetrics(SM_CYMENU);
        Rectangle.bottom += GetSystemMetrics(SM_CYCAPTION);
        SetWindowPos(hWnd,(HWND)NULL,0,0,
                     Rectangle.right,
                     Rectangle.bottom,
                     SWP_NOMOVE | SWP_NOZORDER);
      }
      InvalidateRect(hWnd, NULL, FALSE);
    }
    break;
  case IDM_EXIT:
    PostQuitMessage(0);
    break;
  case IDM_STRETCH:
    stretch ^= 1;
    if(stretch)CheckMenuItem(GetMenu(hWnd),IDM_STRETCH,MF_CHECKED);
    else       CheckMenuItem(GetMenu(hWnd),IDM_STRETCH,MF_UNCHECKED);
    InvalidateRect(hWnd,NULL,TRUE);
    break;
  case IDM_DRAWDIB:
    dibdirect ^= 1;
    if(dibdirect)CheckMenuItem(GetMenu(hWnd),IDM_DRAWDIB,MF_CHECKED);
    else         CheckMenuItem(GetMenu(hWnd),IDM_DRAWDIB,MF_UNCHECKED);
    InvalidateRect(hWnd,NULL,TRUE);
    InvalidateRect(hWnd,NULL,TRUE);
    break;
  case IDM_ABOUT:
    DialogBox(glInstance,MAKEINTRESOURCE(IDD_ABOUT),hWnd,AboutDlgProc);
    break;
  default:
    break;
  }
}

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size;

static int x4_size;

static int rowcount;

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(y_size - 1 - rowcount)*(long)(x_size);
 memcpy(pixels + l, linepixels, linelen);
 rowcount++;
 return 0;
}

static int LoadGIF(int id, char *GIFfile){
 int i;
 RGBQUAD *pRgb          = NULL;
 LPSTR   lpBitmapHeader = NULL;
 LPSTR   lpBitmapBits   = NULL;
 char temp[128];
 long imagesize;
 LPBITMAPINFOHEADER lpbi;
 rowcount=0;
 strcpy(temp,GIFfile);
 if(read_gif_image(temp,1) < 0){
   return 0;
 }
 x4_size = ALIGNULONG(x_size);
 imagesize = (DWORD)x4_size*(DWORD)y_size;
 lpBitmapHeader = (LPSTR)EVI_Malloc((long)sizeof(BITMAPINFOHEADER)
                                +256L*sizeof(RGBQUAD)+imagesize);
 if(lpBitmapHeader == NULL)return 0;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x4_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=8;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage=(DWORD)imagesize;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=256;
 lpbi->biClrImportant=256;
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<256;i++){
   pRgb[i].rgbRed = p_red[i] << 2;
   pRgb[i].rgbGreen = p_green[i] << 2;
   pRgb[i].rgbBlue = p_blue[i] << 2;
 }
 hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits = ((LPSTR)lpBitmapHeader)
              + (long)sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD);
 memset(lpBitmapBits,0,imagesize);
 pixels=(unsigned char *)lpBitmapBits;
 if(read_gif_image(temp,0) < 0)return 0;
 hbmCurrent[id]=lpBitmapHeader;
 if(hbmCurrent[id] == NULL)return 0;
 return 1;
}

static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi){
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE       red;
 BYTE       green;
 BYTE       blue;
 int        i;
 RGBQUAD    *pRgb;
 if(lpbi == NULL)return NULL;
 if(lpbi->biSize != sizeof(BITMAPINFOHEADER))return NULL;
 /* Get a pointer to the color table and the number of colors in it */
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 nNumColors = 256;
 if(nNumColors){
   pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   /* Fill in the palette entries from the DIB color table and
    * create a logical color palette. */
   for (i = 0; i < nNumColors; i++){
       pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
       pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
       pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 else if (lpbi->biBitCount == 24){
   /* A 24 bitcount DIB has no color table entries so, set the number of
    * to the maximum value (256). */
   nNumColors = MAXPALETTE;
   pPal=(LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   red = green = blue = 0;
   /* Generate 256 (= 8*8*4) RGB combinations to fill the palette entries. */
   for (i = 0; i < pPal->palNumEntries; i++){
       pPal->palPalEntry[i].peRed   = red;
       pPal->palPalEntry[i].peGreen = green;
       pPal->palPalEntry[i].peBlue  = blue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
       if (!(red += 32))if (!(green += 32)) blue += 64;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 return hpal;
}

static void AppPaint(HWND hWnd,HDC hDC,int x,int y){
 HPALETTE hpalT;
 POINT Origin;
 HDC hDCbits;
 BITMAP bm;
 int id;
 id=GetCurrentImage();
 SetWindowOrgEx(hDC,x,y,&Origin);
 SetBkMode (hDC,TRANSPARENT);
 if(hpalCurrent != NULL){
   if(hbmCurrent[id] != NULL){
     LPSTR lpBitmapBits;
     lpBitmapBits = ((LPSTR)hbmCurrent[id])
              + (long)sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD);
     if(!dibdirect){
       hpalT = SelectPalette(hDC,hpalCurrent,FALSE);
       RealizePalette (hDC);
       SetDIBitsToDevice(hDC,
                         0,0,
                         (WORD)((LPBITMAPINFOHEADER)hbmCurrent[id])->biWidth,
                         (WORD)((LPBITMAPINFOHEADER)hbmCurrent[id])->biHeight,
                         0,0,
                         0,
                         (WORD)((LPBITMAPINFOHEADER)hbmCurrent[id])->biHeight,
                         lpBitmapBits,
                         (LPBITMAPINFO)hbmCurrent[id],
                         DIB_RGB_COLORS);
       SelectPalette(hDC,hpalT,FALSE);
     }
     else{
       int sx,sy;
       if(!stretch){
         sx=(int)((LPBITMAPINFOHEADER)hbmCurrent[id])->biWidth;
         sy=(int)((LPBITMAPINFOHEADER)hbmCurrent[id])->biHeight;
       }
       else{ sx=sizeX; sy=sizeY;}
       DrawDibDraw(hDrawDib,hDC,0,0,
           sx,
           sy,
           (LPBITMAPINFOHEADER)hbmCurrent[id],lpBitmapBits,
           0,0,
           (int)((LPBITMAPINFOHEADER)hbmCurrent[id])->biWidth,
           (int)((LPBITMAPINFOHEADER)hbmCurrent[id])->biHeight,
           0);
     }
   }
 }
}

static void FreeBitmaps(void){
 int i;
 for(i=0;i<nImages;i++){
   if(hbmCurrent[i])EVI_Free(hbmCurrent[i]);
   hbmCurrent[i]=NULL;
 }
 if(hpalCurrent)DeleteObject(hpalCurrent);
 hpalCurrent=NULL;
}

static int GetCurrentImage(void){
 int iR,iU;
 double dR,dU;
 if(resID >= 0){
   if     (resID == 0){ dU=20.0;iU=(int)((-up_angle+80)/dU); }
   else if(resID == 1){ dU=15.0;iU=(int)((-up_angle+75)/dU); }
   else if(resID == 2){ dU=20.0;iU=(int)((-up_angle+80)/dU); }
   else if(resID == 3){ dU=15.0;iU=(int)((-up_angle+75)/dU); }
   else if(resID == 4){ dU=15.0;iU=(int)((-up_angle+75)/dU); }
   iU = max(0,iU); iU=min(iU,nU-1);
   dR = 360.0/(double)nR;
   iR = (int)((round_angle + 180.0)/dR);
   iR = max(0,iR); iR=min(iR,nR-1);
   return nR*iU+iR;
   return 0;
 }
 return 0;
}

static BOOL MakeImageName(int id, char *name, HWND hWnd){
 char fsurf[256],tempstr[256],*cp1,*cp2;
 FILE *ff;
 cp1=FileInPath(name);
 if(cp1 == NULL){
   MessageBox(hWnd,"Bad Path",NULL,MB_OK);
   return FALSE;
 }
 strcpy(fsurf,cp1);
 cp1=strchr(fsurf,'0');
 cp2=strchr(fsurf,'.');
 if(cp1 == NULL || cp2 == NULL){
   MessageBox(hWnd,"Bad Name In Image Sequence",NULL,MB_OK);
   return FALSE;
 }
 else{
   int ln;
   char *cp,extn[16],fmat[16];
   strcpy(extn,cp2);
   ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
   *cp1='\0';
   sprintf(fmat,"%%s%%0%ldld%%s",ln);
   sprintf(tempstr,fmat,fsurf,id,extn);
 }
 cp1=FileInPath(name);
 strcpy(cp1,tempstr);
 if((ff=fopen(name,"r")) == NULL)return FALSE;
 fclose(ff);
 return TRUE;
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static BOOL CALLBACK AboutDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     return( TRUE );
   case WM_COMMAND:
      if( wparam == IDOK ) {
        EndDialog( hwnd, TRUE );
        return( TRUE );
      }
      break;
 }
 return( FALSE );
}
