#include <windows.h>
#include <vfw.h>
#include <commctrl.h>

#include <stdio.h>
#include <string.h>
#include "global.h"
#include "video.h"

extern HWND hWndProgressF,hWndProgressM,hWndTextF,hWndTextM;

int sec_remaining = 0;
int time_valid = 0;

void putseq()
{
  /* this routine assumes (N % M) == 0 */
  int i, j, k, f, f0, n, np, nb, sxf, syf, sxb, syb;
  int ipflag;

  //time_t   current_time, el_time;

  char name[256];
  unsigned char *neworg[3], *newref[3];
  static char ipb[5] = {' ','I','P','B','D'};

  vbv_ovflow = vbv_unflow = 0;
  start_time = 0;

  rc_init_seq(); /* initialize rate control */

  /* sequence header, sequence extension and sequence display extension */
  putseqhdr();

  /* optionally output some text data (description, copyright or whatever) */
  if (strlen(id_string) > 1)
    putuserdata(id_string);

  /* loop through all frames in encoding/decoding order */

  SendMessage(hWndProgressM,PBM_SETRANGE,0,MAKELPARAM(0,nframes-1));
  SendMessage(hWndProgressM,PBM_SETSTEP,(WPARAM)1,0);
  SendMessage(hWndProgressM,PBM_SETPOS,(WPARAM)0,0);

  for (i=0; i<nframes; i++)
  {
    if (!quiet){ // ToDo
      char tempc[256];
      sprintf(tempc,"%ld(%ld)",i+1,nframes+1);
      SendMessage(hWndTextM,WM_SETTEXT,0,(LPARAM)tempc);
      SendMessage(hWndProgressM,PBM_SETPOS,(WPARAM)i,0);
      MyYield();
//      printf("frame %u",i);
    }

    /* f0: lowest frame number in current GOP
     *
     * first GOP contains N-(M-1) frames,
     * all other GOPs contain N frames
     */
    f0 = N*((i+(M-1))/N) - (M-1);

    if (f0<0)
      f0=0;

    if (i==0 || (i-1)%M==0)
    {
      /* I or P frame */
      for (j=0; j<3; j++)
      {
        /* shuffle reference frames */
        neworg[j] = oldorgframe[j];
        newref[j] = oldrefframe[j];
        oldorgframe[j] = neworgframe[j];
        oldrefframe[j] = newrefframe[j];
        neworgframe[j] = neworg[j];
        newrefframe[j] = newref[j];
      }

      /* f: frame number in display order */
      f = (i==0) ? 0 : i+M-1;
      if (f>=nframes)
        f = nframes - 1;

      if (i==f0) /* first displayed frame in GOP is I */
      {
        /* I frame */
        pict_type = I_TYPE;
        forw_hor_f_code = forw_vert_f_code = 15;
        back_hor_f_code = back_vert_f_code = 15;

        /* n: number of frames in current GOP
         *
         * first GOP contains (M-1) less (B) frames
         */
        n = (i==0) ? N-(M-1) : N;

        /* last GOP may contain less frames */
        if (n > nframes-f0)
          n = nframes-f0;

        /* number of P frames */
        if (i==0)
          np = (n + 2*(M-1))/M - 1; /* first GOP */
        else
          np = (n + (M-1))/M - 1;

        /* number of B frames */
        nb = n - np - 1;

        rc_init_GOP(np,nb);

        putgophdr(f0,i==0); /* set closed_GOP in first GOP only */
      }
      else
      {
        /* P frame */
        pict_type = P_TYPE;
        forw_hor_f_code = motion_data[0].forw_hor_f_code;
        forw_vert_f_code = motion_data[0].forw_vert_f_code;
        back_hor_f_code = back_vert_f_code = 15;
        sxf = motion_data[0].sxf;
        syf = motion_data[0].syf;
      }
    }
    else
    {
      /* B frame */
      for (j=0; j<3; j++)
      {
        neworg[j] = auxorgframe[j];
        newref[j] = auxframe[j];
      }

      /* f: frame number in display order */
      f = i - 1;
      pict_type = B_TYPE;
      n = (i-2)%M + 1; /* first B: n=1, second B: n=2, ... */
      forw_hor_f_code = motion_data[n].forw_hor_f_code;
      forw_vert_f_code = motion_data[n].forw_vert_f_code;
      back_hor_f_code = motion_data[n].back_hor_f_code;
      back_vert_f_code = motion_data[n].back_vert_f_code;
      sxf = motion_data[n].sxf;
      syf = motion_data[n].syf;
      sxb = motion_data[n].sxb;
      syb = motion_data[n].syb;
    }

    temp_ref = f - f0;
    frame_pred_dct = frame_pred_dct_tab[pict_type-1];
    q_scale_type = qscale_tab[pict_type-1];
    intravlc = intravlc_tab[pict_type-1];
    altscan = altscan_tab[pict_type-1];

#ifdef DEBUG
    RSFfprf(statfile,"Frame %d (#%d in display order): ",i,f);
    RSFfprf(statfile," picture_type=%c ",ipb[pict_type]);
    RSFfprf(statfile," temporal_reference=%d ",temp_ref);
    RSFfprf(statfile," frame_pred_frame_dct=%d ",frame_pred_dct);
    RSFfprf(statfile," q_scale_type=%d ",q_scale_type);
    RSFfprf(statfile," intra_vlc_format=%d ",intravlc);
    RSFfprf(statfile," alternate_scan=%d ",altscan);

    if (pict_type!=I_TYPE)
    {
      RSFfprf(statfile," forward search window: %d...%d / %d...%d ",
        -sxf,sxf,-syf,syf);
      RSFfprf(statfile," forward vector range: %d...%d.5 / %d...%d.5 ",
        -(4<<forw_hor_f_code),(4<<forw_hor_f_code)-1,
        -(4<<forw_vert_f_code),(4<<forw_vert_f_code)-1);
    }

    if (pict_type==B_TYPE)
    {
      RSFfprf(statfile," backward search window: %d...%d / %d...%d ",
        -sxb,sxb,-syb,syb);
      RSFfprf(statfile," backward vector range: %d...%d.5 / %d...%d.5 ",
        -(4<<back_hor_f_code),(4<<back_hor_f_code)-1,
        -(4<<back_vert_f_code),(4<<back_vert_f_code)-1);
    }
#else
  if(!quiet)
  {
//    RSFfprf(stderr," Type=%c",ipb[pict_type]);
    if(start_time == 0) time(&start_time);
  //  else if((ipb[pict_type] == 'I')||((time_valid)&&(sec_remaining < 120)))
  //  {
  //    time(&current_time);
  //    el_time = current_time - start_time;
  //    sec_remaining = (el_time / i+1) * (pfi.dwLength*frame_repl - 1 - i - frame0);
  //    time_valid = 1;
  //  }
  //  if(sec_remaining != 0)
  //  {
  //    if(sec_remaining <= 60)
  //      RSFfprf(stderr, " Est. time = %i seconds ", sec_remaining);
  //    else if(sec_remaining < 3600)
  //      RSFfprf(stderr, " Est. time < %i minutes ", (sec_remaining/60)+1);
  //    else
  //    {
  //      RSFfprf(stderr, " Est. time = %i hour(s) %i minute(s) ", sec_remaining/3600, (sec_remaining%3600)/60);
  //    }
  //  }
    RSFfprf(stderr,"Type=%c %u%% complete ",ipb[pict_type],
            i*100/(pfi.dwLength*frame_repl));
  }
#endif

    readframe(f+frame0, neworg);

    if (fieldpic)
    {
      if (!quiet)
      {
        RSFfprf(stderr," first field  (%s) ",topfirst ? "top" : "bot");
      }

      pict_struct = topfirst ? TOP_FIELD : BOTTOM_FIELD;

      motion_estimation(oldorgframe[0],neworgframe[0],
                        oldrefframe[0],newrefframe[0],
                        neworg[0],newref[0],
                        sxf,syf,sxb,syb,mbinfo,0,0);

      predict(oldrefframe,newrefframe,predframe,0,mbinfo);
      dct_type_estimation(predframe[0],neworg[0],mbinfo);
      transform(predframe,neworg,mbinfo,blocks);

      putpict(neworg[0]);

      for (k=0; k<mb_height2*mb_width; k++)
      {
        if (mbinfo[k].mb_type & MB_INTRA)
          for (j=0; j<block_count; j++)
            iquant_intra(blocks[k*block_count+j],blocks[k*block_count+j],
                         dc_prec,intra_q,mbinfo[k].mquant);
        else
          for (j=0;j<block_count;j++)
            iquant_non_intra(blocks[k*block_count+j],blocks[k*block_count+j],
                             inter_q,mbinfo[k].mquant);
      }

      itransform(predframe,newref,mbinfo,blocks);
#ifdef DEBUG
      calcSNR(neworg,newref);
      stats();
#endif

      if (!quiet)
      {
        RSFfprf(stderr,"second field (%s) ",topfirst ? "bot" : "top");
      }

      pict_struct = topfirst ? BOTTOM_FIELD : TOP_FIELD;

      ipflag = (pict_type==I_TYPE);
      if (ipflag)
      {
        /* first field = I, second field = P */
        pict_type = P_TYPE;
        forw_hor_f_code = motion_data[0].forw_hor_f_code;
        forw_vert_f_code = motion_data[0].forw_vert_f_code;
        back_hor_f_code = back_vert_f_code = 15;
        sxf = motion_data[0].sxf;
        syf = motion_data[0].syf;
      }

      motion_estimation(oldorgframe[0],neworgframe[0],
                        oldrefframe[0],newrefframe[0],
                        neworg[0],newref[0],
                        sxf,syf,sxb,syb,mbinfo,1,ipflag);

      predict(oldrefframe,newrefframe,predframe,1,mbinfo);
      dct_type_estimation(predframe[0],neworg[0],mbinfo);
      transform(predframe,neworg,mbinfo,blocks);

      putpict(neworg[0]);

      for (k=0; k<mb_height2*mb_width; k++)
      {
        if (mbinfo[k].mb_type & MB_INTRA)
          for (j=0; j<block_count; j++)
            iquant_intra(blocks[k*block_count+j],blocks[k*block_count+j],
                         dc_prec,intra_q,mbinfo[k].mquant);
        else
          for (j=0;j<block_count;j++)
            iquant_non_intra(blocks[k*block_count+j],blocks[k*block_count+j],
                             inter_q,mbinfo[k].mquant);
      }

      itransform(predframe,newref,mbinfo,blocks);
#ifdef DEBUG
      calcSNR(neworg,newref);
      stats();
#endif
    }
    else
    {
      pict_struct = FRAME_PICTURE;

      /* do motion_estimation
       *
       * uses source frames (...orgframe) for full pel search
       * and reconstructed frames (...refframe) for half pel search
       */

      motion_estimation(oldorgframe[0],neworgframe[0],
                        oldrefframe[0],newrefframe[0],
                        neworg[0],newref[0],
                        sxf,syf,sxb,syb,mbinfo,0,0);

      predict(oldrefframe,newrefframe,predframe,0,mbinfo);
      dct_type_estimation(predframe[0],neworg[0],mbinfo);
      transform(predframe,neworg,mbinfo,blocks);

      putpict(neworg[0]);

      for (k=0; k<mb_height*mb_width; k++)
      {
        if (mbinfo[k].mb_type & MB_INTRA)
          for (j=0; j<block_count; j++)
            iquant_intra(blocks[k*block_count+j],blocks[k*block_count+j],
                         dc_prec,intra_q,mbinfo[k].mquant);
        else
          for (j=0;j<block_count;j++)
            iquant_non_intra(blocks[k*block_count+j],blocks[k*block_count+j],
                             inter_q,mbinfo[k].mquant);
      }

      itransform(predframe,newref,mbinfo,blocks);
#ifdef DEBUG
      calcSNR(neworg,newref);
      stats();
#endif
    }

    sprintf(name,tplref,f+frame0);
    writeframe(name,newref);

  }  // End of frame loop

  putseqend();
  SendMessage(hWndProgressM,PBM_SETPOS,(WPARAM)nframes,0);
  SendMessage(hWndTextM,WM_SETTEXT,0,(LPARAM)"Complete");
}
