/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file AVIUTILS.c */

#define  STRICT
#define  INC_OLE2
#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <vfw.h>

#include "build.h"
#include "..\fli2gif\giflib.h"

#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)

#define BUFSIZE 256
#define AVIIF_KEYFRAME  0x00000010L // this frame is a key frame.
#define BFT_BITMAP      0x4d42      /* 'BM' */
#define GIFCOLOURS      256

extern void CertreDialogOnScreen(HWND);
extern void YieldTo(void);
extern int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent);
extern void AppendFileExtension(char *filename, char *ext);
extern void YieldTo(void);
extern void Query(int id);
extern void SetToolbarState(BOOL state);


extern HWND hWndProgress;
extern CHAR szOutDirName[];

static void writeGif(char *szOutFileName, LPBITMAPINFOHEADER lpbi){
 LPSTR lpBitmapBits;
 RGBQUAD *pRgb;
 GifColorType GifColourMap[GIFCOLOURS];
 GifFileType *GifFile;
 long i,j,ScreenXMax,ScreenYMax,ScreenColorBits;
 lpBitmapBits=(LPSTR)(lpbi);
 lpBitmapBits += (lpbi->biSize+GIFCOLOURS*sizeof(RGBQUAD));
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<GIFCOLOURS;i++){
   GifColourMap[i].Red   = pRgb[i].rgbRed;
   GifColourMap[i].Green = pRgb[i].rgbGreen;
   GifColourMap[i].Blue  = pRgb[i].rgbBlue;
 }
 ScreenXMax=(long)lpbi->biWidth;
 ScreenYMax=(long)lpbi->biHeight;
 ScreenColorBits = 8;
 if((GifFile=EGifOpenFileName(szOutFileName,FALSE)) == NULL)return;
 if(EGifPutScreenDesc(GifFile, ScreenXMax, ScreenYMax, ScreenColorBits,
     0, ScreenColorBits, GifColourMap) == ERROR)goto QQIT;
 if(EGifPutImageDesc(GifFile, 0, 0, ScreenXMax, ScreenYMax, FALSE, 1,
     NULL) == ERROR)goto QQIT;
 for(j=ScreenYMax-1;j>=0;j--){
   if(EGifPutLine(GifFile,(PixelType *)(lpBitmapBits+j*ScreenXMax),
       ScreenXMax) == ERROR)goto QQIT;
 }
 QQIT:
 EGifCloseFile(GifFile);
 return;
}

static void writeBitmap(char *szOutFileName, LPBITMAPINFOHEADER lpbi){
 HANDLE hfbm;
 DWORD dwRead;
 long imagesize;
 BITMAPFILEHEADER hdr,*lphdr;
 if((hfbm=CreateFile(szOutFileName,GENERIC_WRITE,FILE_SHARE_WRITE,
   NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,
   (HANDLE)NULL)) == INVALID_HANDLE_VALUE){
   MessageBox(NULL,szOutFileName,"File Open Fail",MB_OK);
   MessageBeep(MB_OK);
   return;
 }
 imagesize=lpbi->biSizeImage;
 lphdr=&hdr;
 hdr.bfType          = BFT_BITMAP;
 hdr.bfSize          = (sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)
                        + lpbi->biClrUsed*sizeof(RGBQUAD)
                        + imagesize);
 hdr.bfReserved1     = 0;
 hdr.bfReserved2     = 0;
 hdr.bfOffBits       = (DWORD)sizeof(BITMAPFILEHEADER) + lpbi->biSize+
                       lpbi->biClrUsed*sizeof(RGBQUAD);
 WriteFile(hfbm,(LPSTR)lphdr,sizeof(BITMAPFILEHEADER),&dwRead,(LPOVERLAPPED)NULL);
 WriteFile(hfbm,(LPSTR)lpbi,
                sizeof(BITMAPINFOHEADER)+
                lpbi->biClrUsed*sizeof(RGBQUAD)+
                imagesize,
                &dwRead,(LPOVERLAPPED)NULL);
 if(dwRead != sizeof(BITMAPINFOHEADER)+
              lpbi->biClrUsed*sizeof(RGBQUAD)+
              imagesize){
   MessageBeep(MB_OK);
   MessageBox(NULL,"Bad write",NULL,MB_OK);
 }
 if(!CloseHandle(hfbm)){
   MessageBeep(MB_OK);
   MessageBox(NULL,"Bad close",NULL,MB_OK);
 }
 return;
}

void MakeFileName(OPENFILENAME *ofn,  char *szFileName, char *szFilter, HWND hWnd){
 int i;
 strcpy(szFilter,"Video files|*.avi|");
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 ofn->lStructSize = sizeof(OPENFILENAME);
 ofn->hwndOwner = hWnd;
 ofn->hInstance = NULL;
 ofn->lpstrFilter = szFilter;
 ofn->lpstrCustomFilter = NULL;
 ofn->nMaxCustFilter = 0;
 ofn->nFilterIndex = 0;
 *szFileName = '\0';
 ofn->lpstrFile = szFileName;
 ofn->nMaxFile = BUFSIZE;
 ofn->lpstrFileTitle = NULL;
 ofn->nMaxFileTitle = 0;
 ofn->lpstrInitialDir = NULL;
 ofn->lpstrTitle = "Open";
 ofn->Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;
 ofn->nFileOffset = 0;
 ofn->nFileExtension = 0;
 ofn->lpstrDefExt = NULL;
 ofn->lCustData = 0;
 ofn->lpfnHook = NULL;
 ofn->lpTemplateName = NULL;
}

void AviUtilities(HWND hWnd, HWND hWndList){
 OPENFILENAME ofn;
 char szFileName[BUFSIZE];
 char szFilter[BUFSIZE];
 MakeFileName(&ofn,szFileName,szFilter,hWnd);
 if(GetOpenFileNamePreview(&ofn)){
   AVICOMPRESSOPTIONS   aviopt[2];
   LPAVICOMPRESSOPTIONS lpaviopt[2];
   LPBITMAPINFOHEADER lpBmi;
   LPWAVEFORMATEX  lpWavx;
   LPSTR pformat;
   char buffer[256];
   PAVIFILE pfile = NULL;
   PAVISTREAM ps = NULL, pss[2];
   PGETFRAME  pget;
   AVISTREAMINFO avis;
   HRESULT hr;
   long size,nsamples,key;
   int i;
   lpaviopt[0]=&aviopt[0];
   lpaviopt[1]=&aviopt[1];
   memset(lpaviopt[0],0,sizeof(AVICOMPRESSOPTIONS));
   memset(lpaviopt[1],0,sizeof(AVICOMPRESSOPTIONS));
   SendMessage(hWndList,LB_RESETCONTENT,0,0);
   SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)szFileName);
   AVIFileInit();
   hr = AVIFileOpen(&pfile,szFileName,OF_READ,NULL);
   if(hr == AVIERR_OK){
     for(i=0;i<10;i++){
       if(AVIFileGetStream(pfile,&ps,0,i) == AVIERR_OK){
         AVIStreamInfo(ps,&avis,sizeof(avis));
         if(avis.fccType == streamtypeVIDEO){
           char c0,c1,c2,c3;
           int i0,i1,i2,i3;
#if 0
           pss[0]=ps; pss[1]=ps;
           AVISaveOptions(hWnd,
                 ICMF_CHOOSE_KEYFRAME|ICMF_CHOOSE_DATARATE|ICMF_CHOOSE_PREVIEW,
                 2,pss,lpaviopt);
           AVISaveOptionsFree(2,lpaviopt);
#endif
           SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)"Video Stream");
           i0=c0=(char)((avis.fccHandler & 0x000000ff) >>  0);
           if(c0 < 32)c0=' ';
           i1=c1=(char)((avis.fccHandler & 0x0000ff00) >>  8);
           if(c1 < 32)c1=' ';
           i2=c2=(char)((avis.fccHandler & 0x00ff0000) >> 16);
           if(c2 < 32)c2=' ';
           i3=c3=(char)((avis.fccHandler & 0xff000000) >> 24);
           if(c3 < 32)c3=' ';
           sprintf(buffer,"Codec [%02x-%02x-%02x-%02x] (%c%c%c%c)",
                   i0,i1,i2,i3,c0,c1,c2,c3);
           SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
           AVIStreamFormatSize(ps,0,&size);
           if((pformat = (LPSTR)X__Malloc(size)) != NULL){
             if(AVIStreamReadFormat(ps,0,pformat,&size) == AVIERR_OK){
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)"Stream Format");
               lpBmi=(LPBITMAPINFOHEADER)pformat;
               sprintf(buffer,"Size (%ld %ld) Bits %ld",lpBmi->biWidth,lpBmi->biHeight,
               (long)(lpBmi->biBitCount));
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
               if(lpBmi->biCompression == BI_RGB)strcpy(buffer,"BI_RGB uncomressed");
               else if(lpBmi->biCompression == BI_RLE8)strcpy(buffer,"BI_RLE8");
               else if(lpBmi->biCompression == BI_RLE4)strcpy(buffer,"BI_RLE4");
               else if(lpBmi->biCompression == BI_BITFIELDS)strcpy(buffer,"BI_BITFIELDS");
               else strcpy(buffer,"Unknown Commpression");
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
               nsamples=AVIStreamEnd(ps);
               sprintf(buffer,"%ld Samples (%0.3lfs) %0.1lf fps",nsamples,
                 (double)AVIStreamSampleToTime(ps,nsamples)*0.001,
                 (double)nsamples*1000.0/
                 (double)AVIStreamSampleToTime(ps,nsamples));
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
               key=0; while (key >= 0){
                 sprintf(buffer,"Key frame %ld",key);
                 SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
                 key=AVIStreamNextKeyFrame(ps,key);
               }
               lpBmi->biCompression=BI_RGB;
               if(lpBmi->biBitCount > 8)lpBmi->biBitCount=24;
               lpBmi->biSizeImage=((((UINT)lpBmi->biBitCount*lpBmi->biWidth
                                  +31)&~31)/8)*lpBmi->biHeight;
               if((pget=AVIStreamGetFrameOpen(ps,lpBmi)) != NULL){
                 if(0)for(i=0;i<nsamples;i++){
                   lpBmi=(LPBITMAPINFOHEADER)AVIStreamGetFrame(pget,i);
                   if(lpBmi == NULL){
                     MessageBox(NULL,"Frame Fail read",NULL,MB_OK);
                     break;
                   }
                   sprintf(buffer,"Frame %ld Size (%ld %ld) Bits %ld Clr=%ld",i,lpBmi->biWidth,lpBmi->biHeight,
                   (long)(lpBmi->biBitCount),(long)(lpBmi->biClrUsed));
                   SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
                 }
                 AVIStreamGetFrameClose(pget);
               }
               else MessageBox(NULL,"No Frame Open",NULL,MB_OK);
             }
             else MessageBeep(MB_OK);
             X__Free(pformat);
           }
           else{
              MessageBox(NULL,"Memory",NULL,MB_OK);
           }
         }
         else if(avis.fccType == streamtypeAUDIO){
#if 0
           pss[0]=ps; pss[1]=ps;
           AVISaveOptions(hWnd,
                 0,
                 2,pss,lpaviopt);
           AVISaveOptionsFree(2,lpaviopt);
#endif
           SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)"Audio Stream");
           AVIStreamFormatSize(ps,0,&size);
           if((pformat = (LPSTR)X__Malloc(size)) != NULL){
             if(AVIStreamReadFormat(ps,0,pformat,&size) == AVIERR_OK){
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)"Stream Format");
               lpWavx=(LPWAVEFORMATEX)pformat;
               sprintf(buffer,"Channels=%ld Samples/sec=%ld Bits/sample=%ld",
                    (WORD)lpWavx->nChannels,(DWORD)lpWavx->nSamplesPerSec,
                    (WORD)lpWavx->wBitsPerSample);
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
               sprintf(buffer,"Av.Bytes=%ld Min.Data=%ld Formatsize=%ld",
                    (DWORD)lpWavx->nAvgBytesPerSec,
                    (DWORD)lpWavx->nBlockAlign,size);
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
               nsamples=AVIStreamEnd(ps);
               sprintf(buffer,"%ld Samples (%0.3lfs)",nsamples,
               (double)AVIStreamSampleToTime(ps,nsamples)*0.001);
               SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
             }
             else MessageBeep(MB_OK);
             X__Free(pformat);
           }
           else{
              MessageBox(NULL,"Memory",NULL,MB_OK);
           }
         }
         else {
           SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)"Other Stream");
         }
       }
     }
     if(pfile != NULL)AVIFileRelease(pfile);
   }
   AVIFileExit();
 }
}

static int istart,iend;
static char root_name[BUFSIZE];

static BOOL CALLBACK RangeDlgProc(HWND hwnd,UINT msg,
                                  WPARAM wparam,LPARAM lparam){
 BOOL err;
 switch( msg ) {
   case WM_INITDIALOG:
     SetDlgItemText(hwnd,IDD_RANGE_NAME,root_name);
     SetDlgItemInt(hwnd,IDD_RANGE_START,istart,FALSE);
     SetDlgItemInt(hwnd,IDD_RANGE_END,iend-1,FALSE);
     CentreDialogOnScreen(hwnd);
     return( TRUE );
   case WM_COMMAND:
      if(wparam == IDOK){
        istart=GetDlgItemInt(hwnd,IDD_RANGE_START,&err,FALSE);
        iend=GetDlgItemInt(hwnd,IDD_RANGE_END,&err,FALSE)+1;
        GetDlgItemText(hwnd,IDD_RANGE_NAME,root_name,255);
        EndDialog(hwnd,TRUE);
        return( TRUE );
      }
      else if(wparam == IDCANCEL){
        istart=iend= -1;
        EndDialog(hwnd,FALSE);
        return( TRUE );
      }
      break;
 }
 return( FALSE );
}

void AviWriteBmp(HWND hWnd, HWND hWndList, HINSTANCE hInst){
 char szFileName[BUFSIZE],szFileRoot[BUFSIZE],*fc;
 char szFilter[BUFSIZE];
 OPENFILENAME ofn;
 MakeFileName(&ofn,szFileName,szFilter,hWnd);
 if(GetOpenFileNamePreview(&ofn)){
   LPBITMAPINFOHEADER lpBmi;
   LPSTR pformat;
   char buffer[256];
   PAVIFILE pfile = NULL;
   PAVISTREAM ps = NULL;
   PGETFRAME  pget;
   AVISTREAMINFO avis;
   HRESULT hr;
   int i;
   long size,nsamples,key;
   strcpy(szFileRoot,szFileName);
   if((fc=strrchr(szFileRoot,'.')) == NULL)return;
   *fc='\0';
   SendMessage(hWndList,LB_RESETCONTENT,0,0);
   sprintf(buffer,"Reading %s",szFileName);
   SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
   AVIFileInit();
   hr = AVIFileOpen(&pfile,szFileName,OF_READ,NULL);
   if(hr == AVIERR_OK){
     for(i=0;i<10;i++){
       if(AVIFileGetStream(pfile,&ps,0,i) == AVIERR_OK){
         AVIStreamInfo(ps,&avis,sizeof(avis));
         if(avis.fccType == streamtypeVIDEO){
           AVIStreamFormatSize(ps,0,&size);
           if((pformat = (LPSTR)X__Malloc(size)) != NULL){
             if(AVIStreamReadFormat(ps,0,pformat,&size) == AVIERR_OK){
               lpBmi=(LPBITMAPINFOHEADER)pformat;
               nsamples=AVIStreamEnd(ps);
               lpBmi->biCompression=BI_RGB;
               if(lpBmi->biBitCount > 8){
                 lpBmi->biBitCount=24;
                 lpBmi->biClrUsed=0;
               }
               lpBmi->biSizeImage=((((UINT)lpBmi->biBitCount*lpBmi->biWidth
                                  +31)&~31)/8)*lpBmi->biHeight;
               if((pget=AVIStreamGetFrameOpen(ps,lpBmi)) != NULL){
                 istart=0; iend=nsamples; strcpy(root_name,szFileRoot);
                 DialogBox(hInst,MAKEINTRESOURCE(IDD_RANGE),hWnd,RangeDlgProc);
                 strcpy(szFileRoot,root_name);
                 if(istart < 0 || strlen(szFileRoot) < 1){istart = -1; iend = -1;}
                 else{
                   if(iend >= nsamples)iend=nsamples;
                   if(istart > iend)istart=iend;
                   if(iend < istart)iend=istart;
                 }
                 for(i=0;i<nsamples;i++){
                   YieldTo();
                   lpBmi=(LPBITMAPINFOHEADER)AVIStreamGetFrame(pget,i);
                   if(lpBmi == NULL){
                     MessageBox(NULL,"Frame Fail read",NULL,MB_OK);
                     break;
                   }
                   if(i >= istart && i < iend){
                     sprintf(buffer,"%s%03ld.bmp",szFileRoot,i);
                     SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
                     writeBitmap(buffer,lpBmi);
                   }
                 }
                 AVIStreamGetFrameClose(pget);
               }
               else MessageBox(NULL,"No Frame Open",NULL,MB_OK);
             }
             else MessageBeep(MB_OK);
             X__Free(pformat);
           }
           else{
              MessageBox(NULL,"Memory",NULL,MB_OK);
           }
         }
       }
     }
     if(pfile != NULL)AVIFileRelease(pfile);
   }
   AVIFileExit();
 }
}

void AviWriteGif(HWND hWnd, HWND hWndList, HINSTANCE hInst){
 char szFileName[BUFSIZE],szFileRoot[BUFSIZE],*fc;
 char szFilter[BUFSIZE];
 OPENFILENAME ofn;
 MakeFileName(&ofn,szFileName,szFilter,hWnd);
 if(GetOpenFileNamePreview(&ofn)){
   LPBITMAPINFOHEADER lpBmi;
   LPSTR pformat;
   char buffer[256];
   PAVIFILE pfile = NULL;
   PAVISTREAM ps = NULL;
   PGETFRAME  pget;
   AVISTREAMINFO avis;
   HRESULT hr;
   int i;
   long size,nsamples,key;
   strcpy(szFileRoot,szFileName);
   if((fc=strrchr(szFileRoot,'.')) == NULL)return;
   *fc='\0';
   SendMessage(hWndList,LB_RESETCONTENT,0,0);
   sprintf(buffer,"Reading %s",szFileName);
   SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
   AVIFileInit();
   hr = AVIFileOpen(&pfile,szFileName,OF_READ,NULL);
   if(hr == AVIERR_OK){
     for(i=0;i<10;i++){
       if(AVIFileGetStream(pfile,&ps,0,i) == AVIERR_OK){
         AVIStreamInfo(ps,&avis,sizeof(avis));
         if(avis.fccType == streamtypeVIDEO){
           AVIStreamFormatSize(ps,0,&size);
           if((pformat = (LPSTR)X__Malloc(size)) != NULL){
             if(AVIStreamReadFormat(ps,0,pformat,&size) == AVIERR_OK){
               lpBmi=(LPBITMAPINFOHEADER)pformat;
               nsamples=AVIStreamEnd(ps);
               lpBmi->biCompression=BI_RGB;  // read it back as 8 bit
               lpBmi->biBitCount=8;          // paletted image
               lpBmi->biClrUsed=256;         //
               lpBmi->biSizeImage=((((UINT)lpBmi->biBitCount*lpBmi->biWidth
                                  +31)&~31)/8)*lpBmi->biHeight;
               if((pget=AVIStreamGetFrameOpen(ps,lpBmi)) != NULL){
                 istart=0; iend=nsamples; strcpy(root_name,szFileRoot);
                 DialogBox(hInst,MAKEINTRESOURCE(IDD_RANGE),hWnd,RangeDlgProc);
                 strcpy(szFileRoot,root_name);
                 if(istart < 0 || strlen(szFileRoot) < 1){istart = -1; iend = -1;}
                 else{
                   if(iend >= nsamples)iend=nsamples;
                   if(istart > iend)istart=iend;
                   if(iend < istart)iend=istart;
                 }
                 for(i=0;i<nsamples;i++){
                   YieldTo();
                   lpBmi=(LPBITMAPINFOHEADER)AVIStreamGetFrame(pget,i);
                   if(lpBmi == NULL){
                     MessageBox(NULL,"Frame Fail read",NULL,MB_OK);
                     break;
                   }
                   if(i >= istart && i < iend){
                     sprintf(buffer,"%s%03ld.gif",szFileRoot,i);
                     SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buffer);
                     writeGif(buffer,lpBmi);
                   }
                 }
                 AVIStreamGetFrameClose(pget);
               }
               else MessageBox(NULL,"No Frame Open",NULL,MB_OK);
             }
             else MessageBeep(MB_OK);
             X__Free(pformat);
           }
           else{
              MessageBox(NULL,"Memory",NULL,MB_OK);
           }
         }
       }
     }
     if(pfile != NULL)AVIFileRelease(pfile);
   }
   AVIFileExit();
 }
}

static LONG CALLBACK my_save_callback(int n_percent){
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)n_percent,0);
 YieldTo();
 return AVIERR_OK;
}

#if 0
static BOOL PullFile(long id, char *fileslist, char *szFileName){
 int i=0,k=0,n=0;
 char *p0,*p,dir[256];
 p=p0=fileslist;
 while(1){
   if(*p == 0){
     n++;
     p++;
     if(*p == 0)break;
   }
   if(i++ > 1024)return FALSE;
   p++;
 }
 i=0;p=p0=fileslist;
 while(1){
   if(*p == 0){
//MessageBox(NULL,p0,"Trying",MB_OK);
     if(k == 0)strcpy(dir,p0);
     else if(k == id+1){
       strcpy(szFileName,dir);
       strcat(szFileName,"\\");
       strcat(szFileName,p0);
MessageBox(NULL,szFileName,"Found",MB_OK);
       return TRUE;
     }
     k++;
     p++;
     if(*p == 0)break;
     p0=p;
   }
   p++;
   if(i++ > 1000)break;
 }
 return FALSE;
}
#endif

void AddJoin(HWND hWnd, HWND hWndList, HINSTANCE hInst){
 char szFileName[BUFSIZE];
 char szFilter[BUFSIZE];
 OPENFILENAME ofn;
 MakeFileName(&ofn,szFileName,szFilter,hWnd);
 if(!GetOpenFileNamePreview(&ofn))return;
 SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)szFileName);
 return;
}

void AviJoin(HWND hWnd, HWND hWndList, HINSTANCE hInst){
 char szFileName[BUFSIZE],szOutFileName[BUFSIZE];
 char szFilter[BUFSIZE];
 AVICOMPRESSOPTIONS   aviopt[2];
 LPAVICOMPRESSOPTIONS lpaviopt[2];
 PAVIFILE pfile[100];
 PAVISTREAM ps,psv[100],psa[100],psout[2]={NULL,NULL};
 PGETFRAME  pget;
 AVISTREAMINFO avis;
 LPBITMAPINFOHEADER lpBmi;
 LPWAVEFORMATEX  lpWavx;
 LPSTR pformat;
 HRESULT hr;
 long i,j,fc,k,l,n;
 BOOL bCompatible;
 long size,CurrentX,CurrentY,CurrentB;
 szOutFileName[0]='\0';
 AVIFileInit();
 if((n=SendMessage(hWndList,LB_GETCOUNT,(WPARAM)0,(LPARAM)0)) < 2){
   MessageBox(NULL,"At least 2 AVI files required",NULL,MB_OK);
   return;
 }
 bCompatible=TRUE;
 for(fc=0,j=0;j<n;j++){
   if(SendMessage(hWndList,LB_GETTEXT,(WPARAM)j,(LPARAM)szFileName)
     == LB_ERR){
     MessageBox(NULL,"Bad Name",NULL,MB_OK);
   }
   hr = AVIFileOpen(&pfile[fc],szFileName,OF_READ,NULL);
   psv[fc] = psa[fc] = NULL;
   if(hr == AVIERR_OK){
     for(i=0;i<10;i++){
       if(AVIFileGetStream(pfile[fc],&ps,0,i) == AVIERR_OK){
         AVIStreamInfo(ps,&avis,sizeof(avis));
         if     (avis.fccType == streamtypeVIDEO){
           psv[fc]=ps;
           AVIStreamFormatSize(ps,0,&size);
           if((pformat = (LPSTR)X__Malloc(size)) != NULL){
             if(AVIStreamReadFormat(ps,0,pformat,&size) == AVIERR_OK){
               lpBmi=(LPBITMAPINFOHEADER)pformat;
               if(fc == 0){
                 CurrentX=(long)(lpBmi->biWidth);
                 CurrentY=(long)(lpBmi->biHeight);
                 CurrentB=(long)(lpBmi->biBitCount);
               }
               else{
                 if(CurrentX != (long)(lpBmi->biWidth))bCompatible=FALSE;
                 if(CurrentY != (long)(lpBmi->biHeight))bCompatible=FALSE;
                 if(CurrentB != (long)(lpBmi->biBitCount))bCompatible=FALSE;
               }
             }
             X__Free(pformat);
           }
         }
//         else if(avis.fccType == streamtypeAUDIO)psa[fc]=ps;
         else AVIStreamRelease(ps);
       }
       else break;
     }
     fc++;
   }
   else break;
 }
 if(!bCompatible){
   MessageBox(NULL,"The AVI files are incompatible in size or colour depth",
              NULL,MB_OK);
 }
 else if(fc > 1){
   if(CreateEditableStream(&psout[0],psv[0]) != AVIERR_OK){
     MessageBox(NULL,"Bad Edit Stream",NULL,MB_OK);
   }
   else{
     for(i=1;i<fc;i++){
       k=AVIStreamEnd(psout[0]);
       l=AVIStreamEnd(psv[i]);
       if(EditStreamPaste(psout[0],&k,&l,psv[i],0,l) != 0){
         MessageBox(NULL,"Bad Append",NULL,MB_OK);
         break;
       }
     }
     lpaviopt[0]=&aviopt[0];
     lpaviopt[1]=&aviopt[1];
     memset(lpaviopt[0],0,sizeof(AVICOMPRESSOPTIONS));
     memset(lpaviopt[1],0,sizeof(AVICOMPRESSOPTIONS));
     if(SelectFileName(1,szOutFileName,szOutDirName,
             "Joined Output AVI",
             "Video files|*.avi|",hWnd) == TRUE){
       AppendFileExtension(szOutFileName,".avi");
       if(AVISaveOptions(hWnd,
            ICMF_CHOOSE_KEYFRAME|ICMF_CHOOSE_DATARATE|ICMF_CHOOSE_PREVIEW,
            1,psout,lpaviopt)){
         SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,100));
         SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
         SetToolbarState(FALSE);
         if(AVISaveV((LPCTSTR)szOutFileName,NULL,(AVISAVECALLBACK)my_save_callback,
            1, psout,lpaviopt) != AVIERR_OK){
           MessageBox(NULL,"File Save Failed",NULL,MB_OK);
         }
         SetToolbarState(TRUE);
         SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
       }
       AVISaveOptionsFree(2,lpaviopt);
     }
     if(psout[0] != NULL)AVIStreamRelease(psout[0]);
     if(psout[1] != NULL)AVIStreamRelease(psout[1]);
   }
 }
 for(i=0;i<fc;i++){
   if(psv[i] != NULL)AVIStreamRelease(psv[i]);
   if(psa[i] != NULL)AVIStreamRelease(psa[i]);
   if(pfile[i] != NULL)AVIFileRelease(pfile[i]);
 }
 AVIFileExit();
}

#define NSTREAM 10

void Recompress(HWND hWnd, HWND hWndList, HINSTANCE hInst){
 char szFileName[BUFSIZE],szOutFileName[BUFSIZE];
 char szFilter[BUFSIZE];
 OPENFILENAME ofn;
 AVICOMPRESSOPTIONS   aviopt[NSTREAM];
 LPAVICOMPRESSOPTIONS lpaviopt[NSTREAM];
 PAVIFILE pfile=NULL;
 PAVISTREAM ps[NSTREAM];
 HRESULT hr;
 long i,k;
 szOutFileName[0]='\0';
 MakeFileName(&ofn,szFileName,szFilter,hWnd);
 if(!GetOpenFileNamePreview(&ofn))return;
 AVIFileInit();
 k=0;
 hr = AVIFileOpen(&pfile,szFileName,OF_READ,NULL);
 if(hr == AVIERR_OK){
   for(i=0;i<NSTREAM;i++){
     if(AVIFileGetStream(pfile,&ps[i],0,i) != AVIERR_OK)break;
     k++;
   }
 }
 else pfile=NULL;
 if(k > 0){
   for(i=0;i<k;i++){
     lpaviopt[i]=&aviopt[i];
     memset(lpaviopt[i],0,sizeof(AVICOMPRESSOPTIONS));
   }
   if(SelectFileName(1,szOutFileName,szOutDirName,
             "Recompressed Output",
             "Video files|*.avi|",hWnd) == TRUE){
     AppendFileExtension(szOutFileName,".avi");
     if(AVISaveOptions(hWnd,
         ICMF_CHOOSE_KEYFRAME|ICMF_CHOOSE_DATARATE|ICMF_CHOOSE_PREVIEW,
         k,ps,lpaviopt)){
       SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,100));
       SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
       SetToolbarState(FALSE);
       if(AVISaveV((LPCTSTR)szOutFileName,NULL,
            (AVISAVECALLBACK)my_save_callback,
             k, ps,lpaviopt) != AVIERR_OK)
         MessageBox(NULL,"File Save Failed",NULL,MB_OK);
       SetToolbarState(TRUE);
       SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
     }
     AVISaveOptionsFree(k,lpaviopt);
   }
   for(i=0;i<k;i++)if(ps[i] != NULL)AVIStreamRelease(ps[i]);
 }
 if(pfile != NULL)AVIFileRelease(pfile);
 AVIFileExit();
}
