/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file build.C  */

#define MODULE_MAIN

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include "..\..\animate\about.h"
#include "build.h"

#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)
#define X__Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#define StartWait() hcurSave = SetCursor(LoadCursor(NULL,IDC_WAIT))
#define EndWait()   SetCursor(hcurSave)

#define IDM_SYS_AVI  0x10
#define IDM_SYS_DIB  0x20
#define IDM_SYS_GIF  0x30
#define IDM_SYS_REC  0x40
#define IDM_SYS_ABT  0x50

extern int  AVI_frame_rate;
extern void AviUtilities(HWND,HWND);
extern void AviWriteBmp(HWND,HWND,HINSTANCE);
extern void AviWriteGif(HWND,HWND,HINSTANCE);
extern void AviJoin(HWND,HWND,HINSTANCE);
extern void AddJoin(HWND,HWND,HINSTANCE);
extern void configureAVIfile(HWND, char *);
extern void Recompress(HWND,HWND,HINSTANCE);

long getcolor(unsigned short crgb, long Npalette);
void SortColours(long Ngif, long N_colours);
void UnCompressWordLine(short bufsize, short *buffer,
                        unsigned short *output);
int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent);
void AppendFileExtension(char *filename, char *ext);

static BOOL CALLBACK AboutDlgProc(HWND,UINT,WPARAM,LPARAM);
static LONG WINAPI MainWndProc(HWND, UINT, WPARAM, LPARAM);
static BOOL CALLBACK PrefsDlgProc(HWND hdlg,UINT msg,
                           WPARAM wparam,LPARAM lparam);
static PSTR FileInPath(PSTR pstrPath);
static void GetFilesInSequence(void);
static void WriteAVI(void);
static void WriteFLI(void);
unsigned char *LoadMAP(char *CurrentFile, int *xx, int *yy);
static BOOL TestMAP(char *CurrentFile, int *xx, int *yy);

HWND             glWndFrame=NULL,hWndList=NULL,
                 hWndStatus=NULL,hWndProgress=NULL;
CHAR             szOutDirName[256];

static HINSTANCE glInstance;
static CHAR      szAppName[]="OFX:BuildClass";
static CHAR      szFileName[256],szDirName[256];
static CHAR      gszHomeDir[256],gszWork[256],*gszWorkDrive;
static CHAR      szOutFileName[256],stat_m[256];
static RECT      oldrect;
static DWORD     fdwStyle;
static UINT      ToolbarHeight=0,StatusbarHeight=0,ProgressbarHeight=20;
static HWND      hTBWnd=NULL;
static BOOL      bits24=TRUE,AVItype=FALSE,bConfigured=FALSE;
static HCURSOR   hcurSave;
static int       frame_rate=15;

#define NTOOLBARBUTTONS 19
#define NOTOOLBARIMAGES 13

static TBBUTTON tbb[NTOOLBARBUTTONS] = {
          { 0,0,         TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          { 0,IDM_OPEN,  TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          { 1,IDM_BUILD, TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          { 0,0,         TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          { 2,IDM_AVI,   TBSTATE_ENABLED,
                         TBSTYLE_BUTTON | TBSTYLE_CHECKGROUP,
                         0,0 },
          { 3,IDM_FLC,   TBSTATE_ENABLED | TBSTATE_CHECKED,
                         TBSTYLE_BUTTON | TBSTYLE_CHECKGROUP,
                         0,0 },
          { 0,0,         TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          { 4,IDM_8,     TBSTATE_ENABLED,
                         TBSTYLE_BUTTON | TBSTYLE_CHECKGROUP,
                         0,0 },
          { 5,IDM_24,    TBSTATE_ENABLED | TBSTATE_CHECKED,
                         TBSTYLE_BUTTON | TBSTYLE_CHECKGROUP,
                         0,0 },
          { 0,0,         TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          { 6,IDM_CLEAR, TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          { 7,IDM_DELETE,TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {10,IDM_RATE,  TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          { 8,IDM_CODEC, TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          { 0,0,         TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          {11,IDM_JLIST, TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          {12,IDM_JOIN,  TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 },
          { 0,0,         TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          { 9,IDM_HELP,  TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 }
                                        };

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

void YieldTo(void){
 MSG msg;
 while (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE)){
   TranslateMessage(&msg);
   DispatchMessage(&msg);
 }
}

void Query(int id){
  char mess[256];
  LoadString(glInstance,id,mess,255);
  MessageBox(glWndFrame,mess,NULL,MB_OK);
}

void SetToolbarState(BOOL state){
 int i;
 for(i=0;i<NTOOLBARBUTTONS;i++)SendMessage(hTBWnd,
   TB_ENABLEBUTTON,(WPARAM)tbb[i].idCommand,(LPARAM)MAKELONG(state,0));
}

void RegisterMainClass(HINSTANCE hInstance){
 WNDCLASS   wc;
 wc.style         = 0;
 wc.lpfnWndProc   = (WNDPROC)MainWndProc;
 wc.cbClsExtra    = 0;
 wc.cbWndExtra    = 0;
 wc.hInstance     = hInstance;
 wc.hIcon         = LoadIcon(hInstance,"SHOWICON");
 wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
 wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
 wc.lpszMenuName  = NULL;
 wc.lpszClassName = szAppName;
 glInstance=hInstance;
 if (!RegisterClass (&wc) )return;
 return;
}

void UnregisterMainClass(HINSTANCE hInstance){
 UnregisterClass(szAppName,hInstance);
}

static HWND RegisterWindow(HWND parent, LPRECT pr){
 RECT rc;
 HWND hWnd;
 GetModuleFileName(glInstance,gszHomeDir,256);
 *FileInPath(gszHomeDir) = '\0';
 strcpy(szFileName,"temp");
 strcpy(szDirName,gszHomeDir); strcat(szDirName,"images");
 strcpy(szOutFileName,"temp");
 strcpy(szOutDirName,gszHomeDir); strcat(szOutDirName,"movies");
 if(GetEnvironmentVariable((LPTSTR)"EVT",(LPTSTR)gszWork,255) == 0){
   strcpy(gszWork,gszHomeDir);
 }
 gszWorkDrive=&gszWork[0];
 fdwStyle = WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
            WS_SYSMENU | WS_MAXIMIZEBOX | WS_MINIMIZEBOX |
            WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
 LoadString(glInstance,IDI_TITLE,stat_m,255);
 hWnd = CreateWindow(szAppName,
             stat_m,
             fdwStyle,
             pr->left,
             pr->top,
             pr->right,
             pr->bottom,
             parent,
             NULL,
             glInstance,
             NULL);
 if(!hWnd)return NULL;
 glWndFrame=hWnd;
 CentreDialogOnScreen(hWnd);
 ShowWindow (hWnd, SW_SHOW);
 UpdateWindow(hWnd);
 return hWnd;
}

static BOOL CALLBACK FpsDlgProc(HWND hwnd,UINT msg,
                                WPARAM wparam,LPARAM lparam){
 BOOL err;
 int i;
 switch(msg){
   case WM_INITDIALOG:
     SetDlgItemInt(hwnd,DLG_FPS_RATE,frame_rate,FALSE);
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDCANCEL:
         EndDialog(hwnd,0);
         return(TRUE);
       case IDOK:
         frame_rate=GetDlgItemInt(hwnd,DLG_FPS_RATE,&err,FALSE);
         EndDialog(hwnd,1);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return FALSE;
}

static void Commands(WORD cmd,HWND hWnd){
 int i;
 char temp[256];
 switch (cmd){
   case IDM_OPEN:
     LoadString(glInstance,IDI_SELECTIN,stat_m,255);
     if(SelectFileName(0,szFileName,szDirName,
       stat_m,
       "All file formats|*.gif;*.tga;*.16b|"
       "Compuserve GIFs|*.gif|"
       "Targa TGAs|*.tga|"
       "Intermediate Files|*.16b|"
       ,hWnd) == TRUE){
       GetFilesInSequence();
     }
     break;
   case IDM_BUILD:
     LoadString(glInstance,IDI_SELECTOUT,stat_m,255);
     if(AVItype){
       if(SelectFileName(1,szOutFileName,szOutDirName,
             stat_m,
             "Video files|*.avi|",hWnd) != TRUE)break;
       AppendFileExtension(szOutFileName,".avi");
       SetToolbarState(FALSE);
       StartWait();
       if(!bConfigured)configureAVIfile(hWnd,NULL);
       WriteAVI();
       EndWait();
       SetToolbarState(TRUE);
       LoadString(glInstance,IDI_DONEAVI,stat_m,255);
       SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)stat_m);
       break;
     }
     else {
       if(SelectFileName(1,szOutFileName,szOutDirName,
             stat_m,
             "Flic Files|*.fl?|",hWnd) != TRUE)break;
       AppendFileExtension(szOutFileName,".flc");
       StartWait();
       SetToolbarState(FALSE);
       WriteFLI();
       SetToolbarState(TRUE);
       EndWait();
       LoadString(glInstance,IDI_DONEFLC,stat_m,255);
       SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)stat_m);
       break;
     }
     break;
   case IDM_AVI:
     AVItype=TRUE;
     break;
   case IDM_FLC:
     AVItype=FALSE;
     break;
   case IDM_8:
     bits24=FALSE;
     break;
   case IDM_24:
     bits24=TRUE;
     break;
   case IDM_CODEC:
     bConfigured=TRUE;
     configureAVIfile(hWnd,NULL);
     break;
   case IDM_CLEAR:
     SendMessage(hWndList,LB_RESETCONTENT,0,0);
     LoadString(glInstance,IDI_FILESLIST,stat_m,255);
     SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)stat_m);
     SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
     break;
   case IDM_DELETE:{
       int nc;
       nc=SendMessage(hWndList,LB_GETCOUNT,(WPARAM)0,(LPARAM)0);
       if(nc > 0)for(i=0;i<nc;i++){
         if(SendMessage(hWndList,LB_GETSEL,(WPARAM)i,(LPARAM)0) > 0){
           SendMessage(hWndList,LB_DELETESTRING,(WPARAM)i,(LPARAM)0);
           i--; nc--;
         }
       }
     }
     SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
     break;
   case IDM_JLIST:
     AddJoin(hWnd,hWndList,glInstance);
     break;
   case IDM_JOIN:
     AviJoin(hWnd,hWndList,glInstance);
     break;
   case IDM_HELP:
     DialogBox(glInstance,MAKEINTRESOURCE(IDD_ABOUT),hWnd,AboutDlgProc);
     break;
   case IDM_RATE:
     DialogBox(glInstance,MAKEINTRESOURCE(DLG_FPS),hWnd,(DLGPROC)FpsDlgProc);
     if(frame_rate < 1)frame_rate=1;
     AVI_frame_rate=frame_rate;
     break;
   case IDM_EXIT:
     PostQuitMessage(0);
     break;
   default:
     break;
 }
}

static LONG WINAPI MainWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam){
 RECT rc;
 LONG lRet = 1;
 switch (uMsg){
   case WM_CREATE:{
       HMENU hmenu = GetSystemMenu(hWnd, FALSE);
       hTBWnd = CreateToolbarEx(
         hWnd,
         WS_CHILD | WS_VISIBLE | !WS_BORDER | TBSTYLE_TOOLTIPS,
         0,
         NOTOOLBARIMAGES,
         glInstance,
         TOOLBAR_IDBM,
         tbb,
         NTOOLBARBUTTONS,
         16,15, // button sizes
         16,15, // bitmap image sizes
         sizeof(TBBUTTON)
       );
       GetClientRect(hTBWnd,&rc); ToolbarHeight=rc.bottom-rc.top;
       hWndStatus=CreateWindowEx(0L,STATUSCLASSNAME,"",
         WS_CHILD | !WS_BORDER | WS_VISIBLE,
         0,0,0,0,
         hWnd,
         (HMENU)102,
         glInstance,
         NULL);
       GetClientRect(hWndStatus,&rc); StatusbarHeight=rc.bottom-rc.top;
       GetClientRect(hWnd,&rc);
       rc.top = rc.bottom-StatusbarHeight-ProgressbarHeight+2;
       hWndProgress=CreateWindowEx(0L,PROGRESS_CLASS,"",
         WS_CHILD | !WS_BORDER | WS_VISIBLE,
         2,rc.top+4,
         rc.right-4,ProgressbarHeight-4,
         hWnd,
         (HMENU)103,
         glInstance,
         NULL);
       GetClientRect(hWnd,&rc);
       hWndList=CreateWindow("listbox",
         " ",
         WS_VISIBLE | WS_CHILD | WS_VSCROLL | WS_BORDER
         | LBS_NOINTEGRALHEIGHT | LBS_EXTENDEDSEL,
         rc.left+2,
         rc.top+ToolbarHeight+2,
         rc.right-4,
         rc.bottom-ToolbarHeight-StatusbarHeight-ProgressbarHeight-2,
         hWnd,
         (HMENU)101,
         glInstance,
         NULL);
       LoadString(glInstance,IDI_FILESLIST,stat_m,255);
       SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)stat_m);
       {
         AppendMenu(hmenu,MF_SEPARATOR,0,0);
         AppendMenu(hmenu,MF_STRING,IDM_SYS_AVI, "AVI Info");
         AppendMenu(hmenu,MF_STRING,IDM_SYS_DIB, "AVI Write Dib");
         AppendMenu(hmenu,MF_STRING,IDM_SYS_GIF, "AVI Write Gif");
         AppendMenu(hmenu,MF_STRING,IDM_SYS_REC, "Recompress");
         AppendMenu(hmenu,MF_SEPARATOR,0,0);
         AppendMenu(hmenu,MF_STRING,IDM_SYS_ABT, "About...");
       }
     }
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wParam & 0xfff0)){
       case IDM_SYS_AVI:
         AviUtilities(hWnd,hWndList);
         break;
       case IDM_SYS_DIB:
         AviWriteBmp(hWnd,hWndList,glInstance);
         break;
       case IDM_SYS_GIF:
         AviWriteGif(hWnd,hWndList,glInstance);
         break;
       case IDM_SYS_REC:
         Recompress(hWnd,hWndList,glInstance);
         break;
       case IDM_SYS_ABT:
         MessageBox(hWnd,"     "
                     ABOUT_VERSION
                     "        ",
                     "About OpenFX Building Utility...",MB_OK);
     break;
         break;
       default:
         return( DefWindowProc(hWnd,uMsg,wParam,lParam));
      }
      break;
   case WM_SIZE:
     if(hTBWnd != NULL)SendMessage(hTBWnd,uMsg,wParam,lParam);
     if(hWndStatus != NULL)SendMessage(hWndStatus,uMsg,wParam,lParam);
     SetWindowPos(hWndList,NULL,
                  0+2,ToolbarHeight+2,
                  LOWORD(lParam)-4,
                  HIWORD(lParam)-ToolbarHeight-StatusbarHeight
                                -ProgressbarHeight-2,
                  SWP_NOZORDER);
     GetClientRect(hWndList,&rc);
     if(hWndProgress != NULL)SetWindowPos(hWndProgress,NULL,
                  2,ToolbarHeight+6+rc.bottom,
                  LOWORD(lParam)-4,
                  ProgressbarHeight-4,
                  SWP_NOZORDER);
     break;
   case WM_NOTIFY:
     switch (((LPNMHDR)lParam)->code){
       case TTN_NEEDTEXT:{
           int idButton;
           LPTOOLTIPTEXT lpttt;
           lpttt=(LPTOOLTIPTEXT)lParam;
           lpttt->hinst=glInstance;
           idButton=lpttt->hdr.idFrom;
           switch(idButton){
             case IDM_OPEN:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_1); break;
             case IDM_BUILD:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_2); break;
             case IDM_AVI:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_3); break;
             case IDM_FLC:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_4); break;
             case IDM_8:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_5); break;
             case IDM_24:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_6); break;
             case IDM_CLEAR:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_7); break;
             case IDM_DELETE:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_8); break;
             case IDM_CODEC:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_9); break;
             case IDM_HELP:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_A); break;
             case IDM_JLIST:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_B); break;
             case IDM_JOIN:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_C); break;
             case IDM_RATE:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_D); break;
           }
         }
         break;
       default: break;
     }
     return (DefWindowProc(hWnd,uMsg,wParam,lParam));
   case WM_COMMAND:
     if((HWND)lParam == hWndList)break;
     Commands(LOWORD(wParam),hWnd);
     break;
   case WM_CLOSE:
     PostQuitMessage(0);
     break;
   default:
     lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
   break;
 }
 return lRet;
}

void AppendFileExtension(char *filename, char *ext){
 if(strrchr(filename,'.') == NULL){
   strcat(filename,ext);
 }
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

int SelectFileName(int save, char *szfile, char *szdir,
                        char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[256],szdummy[256];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=256;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=256;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

static void PrepareSequenceFilename(char *tname, long id, char *filename){
 int ln;
 char *cp,*cp1,*cp2,extn[16],fmat[16],fsurf[256];
 strcpy(tname,filename); // just use first if anything goes wrong
 strcpy(fsurf,filename);
 cp1=strchr(fsurf,'0');
 if(cp1 == NULL)cp1=strchr(fsurf,'1');
 if(cp1 == NULL)cp1=strchr(fsurf,'2');
 if(cp1 == NULL)cp1=strchr(fsurf,'3');
 if(cp1 == NULL)cp1=strchr(fsurf,'4');
 if(cp1 == NULL)cp1=strchr(fsurf,'5');
 if(cp1 == NULL)cp1=strchr(fsurf,'6');
 if(cp1 == NULL)cp1=strchr(fsurf,'7');
 if(cp1 == NULL)cp1=strchr(fsurf,'8');
 if(cp1 == NULL)cp1=strchr(fsurf,'9');
 cp2=strchr(fsurf,'.');
 if(cp1 == NULL || cp2 == NULL)return;
 strcpy(extn,cp2);
 ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
 *cp1='\0';
 sprintf(fmat,"%%s%%0%ldld%%s",ln);
 sprintf(tname,fmat,fsurf,id,extn);
}

static int GetFileNumber(char *filename){
 int ln;
 char *cp1,*cp2,*cp;
 char temp[256];
 strcpy(temp,filename);
 cp1=strchr(temp,'0');
 if(cp1 == NULL)cp1=strchr(temp,'1');
 if(cp1 == NULL)cp1=strchr(temp,'2');
 if(cp1 == NULL)cp1=strchr(temp,'3');
 if(cp1 == NULL)cp1=strchr(temp,'4');
 if(cp1 == NULL)cp1=strchr(temp,'5');
 if(cp1 == NULL)cp1=strchr(temp,'6');
 if(cp1 == NULL)cp1=strchr(temp,'7');
 if(cp1 == NULL)cp1=strchr(temp,'8');
 if(cp1 == NULL)cp1=strchr(temp,'9');
 cp2=strchr(temp,'.');
 if(cp1 == NULL || cp2 == NULL)return -1;
 cp=cp1; while (cp != cp2){
   if(!isdigit((int)*cp))return -1;
   cp++;
 }
 *cp2='\0';
 return atoi(cp1);
}

static void GetFilesInSequence(void){
 FILE *ff;
 char name[256];
 int i;
 if((i=GetFileNumber(szFileName)) < 0){
   MessageBeep(MB_OK);
   SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)szFileName);
   return;
 }
 for(;;){
   PrepareSequenceFilename(name,i++,szFileName);
   if((ff=fopen(name,"r")) ==NULL)break;
   SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)name);
   fclose(ff);
 }
 return;
}

static BOOL CALLBACK AboutDlgProc(HWND hwnd,UINT msg,
                                  WPARAM wparam,LPARAM lparam){
 switch( msg ) {
   case WM_INITDIALOG:
     CentreDialogOnScreen(hwnd);
     return( TRUE );
   case WM_COMMAND:
      if( wparam == IDOK ) {
        EndDialog( hwnd, TRUE );
        return( TRUE );
      }
      break;
 }
 return( FALSE );
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 MSG msg;
 RECT rc;
 InitCommonControls();
// Ctl3dRegister(hInstance);
// Ctl3dAutoSubclass(hInstance);
 RegisterMainClass(hInstance);
 rc.left=100; rc.top=100;
 rc.right=rc.left+185+22+22+22+10; rc.bottom=rc.top+250;
 rc.bottom +=  GetSystemMetrics(SM_CYMENU);
 RegisterWindow(NULL,&rc);
 while (GetMessage(&msg,NULL,0,0)){
   TranslateMessage(&msg);
   DispatchMessage(&msg);
 }
 DestroyWindow(glWndFrame);
 UnregisterMainClass(hInstance);
// Ctl3dUnAutoSubclass();
// Ctl3dUnregister(hInstance);
 return TRUE;
}

typedef LPBITMAPINFOHEADER PDIB;
extern HPALETTE CreateMyHalftonePalette(void);
extern PDIB DibHalftoneDIB(PDIB pSource, HPALETTE hpalApp);
extern void openAVIfile(char *Name, LPBITMAPINFOHEADER alpbi, int compress);
extern void appendAVIfile(LPBITMAPINFOHEADER alpbi);
extern void closeAVIfile(void);

static void WriteAVI(void){
 char filename[256],message[256];
 long nframes;
 BOOL fhalf=FALSE;
 DWORD dwRead;
 BITMAPINFOHEADER *lpbi;
 HPALETTE hpalHalftone=NULL;
 long i,j,k,imagesize,nx,mx,ny;
 int xx,yy;
 unsigned char *Screen1=NULL,*p,*Screen2=NULL,*q,cr,cg,cb;
 if((nframes=SendMessage(hWndList,LB_GETCOUNT,(WPARAM)0,(LPARAM)0)) < 2){
   Query(IDI_FEWPICS);
   return;
 }
 if(!bits24)fhalf=TRUE;
 else       fhalf=FALSE;
 nx=ny=0;
 SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,nframes-1));
 SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
 for(i=0;i<nframes;i++){
  if(SendMessage(hWndList,LB_GETTEXT,(WPARAM)i,(LPARAM)filename) == LB_ERR){
    Query(IDI_BADLIST);
    return;
  }
  SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(i),0);
  LoadString(glInstance,IDI_TESTING,stat_m,255);
  sprintf(message,stat_m,filename);
  SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)message);
  YieldTo();
  if(!TestMAP(filename,&xx,&yy)){
    Query(IDI_BADFILE);
    return;
  }
  if(xx > nx)nx=xx;
  if(yy > ny)ny=yy;
 }
 mx=((nx*24+31)/32) * 4;
 if(fhalf){
   if((hpalHalftone=CreateMyHalftonePalette()) == NULL){
     Query(IDI_BADPALETTE);
     return;
   }
 }
 if((Screen2=(unsigned char *)X__Malloc(mx*ny+sizeof(BITMAPINFOHEADER)))
            == NULL){
   Query(IDI_MEMORY);
   return;
 }
 SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,nframes-1));
 SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
 for(k=0; k<nframes; k++){
   if(SendMessage(hWndList,LB_GETTEXT,(WPARAM)k,(LPARAM)filename) == LB_ERR){
     Query(IDI_BADLIST);
     break;
   }
   SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(k),0);
   LoadString(glInstance,IDI_BUILDING,stat_m,255);
   sprintf(message,stat_m,filename);
   SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)message);
   YieldTo();
   if((Screen1=LoadMAP(filename,&xx,&yy)) == NULL){
     Query(IDI_BADFILE);
     break;
   }
   memset(Screen2,0,mx*ny+sizeof(BITMAPINFOHEADER));
   for(i=0;i<yy;i++){
     p=Screen1+i*xx*3;
     q=Screen2+sizeof(BITMAPINFOHEADER)+i*mx;
     for(j=0;j<xx;j++){
       cr = (unsigned char)*p++;
       cg = (unsigned char)*p++;
       cb = (unsigned char)*p++;
       *q++ = (unsigned char)cb;
       *q++ = (unsigned char)cg;
       *q++ = (unsigned char)cr;
     }
   }
   X__Free(Screen1);
   lpbi=(LPBITMAPINFOHEADER)Screen2;
   lpbi->biSize=sizeof(BITMAPINFOHEADER);
   lpbi->biWidth=(DWORD)nx;
   lpbi->biHeight=(DWORD)ny;
   lpbi->biPlanes=1;
   lpbi->biBitCount=24;
   lpbi->biCompression=BI_RGB;
   lpbi->biSizeImage = imagesize = (DWORD)(mx*ny);
   lpbi->biXPelsPerMeter=0;
   lpbi->biYPelsPerMeter=0;
   lpbi->biClrUsed=0;
   lpbi->biClrImportant=0;
   YieldTo();
   if(fhalf)lpbi=DibHalftoneDIB(lpbi,hpalHalftone);
   if(k == 0)openAVIfile(szOutFileName,lpbi,1);
   appendAVIfile(lpbi);
   if(k == nframes-1)closeAVIfile();
   if(fhalf && lpbi != NULL)X__Free(lpbi);
 }
 if(fhalf && hpalHalftone != NULL)DeleteObject(hpalHalftone);
 X__Free(Screen2);
 return;
}

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size;

static int rowcount;
static unsigned char *pixels = NULL;

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(y_size - 1 - rowcount)*(long)(x_size);
 memcpy(pixels + l, linepixels, linelen);
 rowcount++;
 return 0;
}

static unsigned char *LoadGIF(char *GIFfile, int *xx, int *yy){
 unsigned char *px,*lpB,*pixels24;
 long i,j,k,imagesize,x,y,xi,yi;
 rowcount=0;
 if(read_gif_image(GIFfile,1) < 0)return NULL;
 x=x_size; y=y_size;
 imagesize=x_size*y_size;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return NULL;
 if((pixels24=(unsigned char*)X__Malloc(imagesize*3)) == NULL){
   X__Free(pixels);
   return NULL;
 }
 if(read_gif_image(GIFfile,0) < 0){
   X__Free(pixels);
   return NULL;
 }
 px=pixels24;
 lpB=pixels;
 for(j=0;j<y;j++){
   for(i=0;i<x;i++){
     k=(long)(*lpB++);
     *px++ = (unsigned char)(p_red[k]   << 2);
     *px++ = (unsigned char)(p_green[k] << 2);
     *px++ = (unsigned char)(p_blue[k]  << 2);
   }
 }
 *xx=x; *yy=y;
 X__Free(pixels);
 return pixels24;
}

static void tga_display_line(unsigned char *prgb,
                             long wide, long line,
                             long upsidedown, long hi){
 register long l,x,k;
 if(upsidedown)l = 3L*(long)(line)*(long)(x_size);
 else          l = 3L*(long)(hi - 1 - line)*(long)(x_size);
 memcpy(pixels + l, prgb, (int)(wide*3));
}

static short R_getvariword(long n, FILE *f, long *vw){
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;          if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

static short read_tga_image(char *filename, short info_only){
 char c0,c1,c2;
 long i,j,k,l,m,n,cmo,cml,xor,yor,wi,hi,iok=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat,wi3;
 FILE *t24;
 unsigned char *prgb;
 prgb=NULL;
 if((t24=fopen(filename,"rb")) != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0 : 1 ;
   if(info_only){
     x_size=(int)wi;
     y_size=(int)hi;
     fclose(t24);
     return 1;
   }
   if((prgb=(unsigned char *)X__Malloc((size_t)wi*3)) == NULL)goto CLOSEOUT;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi;
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   wi3=wi*3;
   if(image_type == 2){ /* uncompressed */
     m=0; k=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       prgb[m+2] = (char)( vw        & pat) << s;
       prgb[m+1] = (char)((vw >> s1) & pat) << s;
       prgb[m  ] = (char)((vw >> s2) & pat) << s;
       k++; m+=3;
       if(m == wi3){
         tga_display_line(prgb,wi,i,upsidedown,hi); m=0;
       }
     }
   }
   else{  /* compressed */
     m=0; l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           prgb[m+2] = (char)( vw        & pat) << s;
           prgb[m+1] = (char)((vw >> s1) & pat) << s;
           prgb[m  ] = (char)((vw >> s2) & pat) << s;
           k++; m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           prgb[m+2] = c0;
           prgb[m+1] = c1;
           prgb[m  ] = c2;
           k++;  m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
   if(prgb != NULL)X__Free((char *)prgb);
 }
 return (short)iok;
}

static unsigned char *LoadTGA(char *TGAfile, int *xx, int *yy){
 long imagesize;
 rowcount=0;
 if(read_tga_image(TGAfile,1) <= 0)return NULL;
 imagesize=x_size*y_size*3;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return NULL;
 if(read_tga_image(TGAfile,0) <= 0){
   X__Free(pixels);
   return NULL;
 }
 *xx=x_size; *yy=y_size;
 return pixels;
}

static short compressbuffer[1024];

#if 0
void  UnCompressWordLine(short bufsize, short *buffer,unsigned short *output){
 short p,k;
 long i;
 k=0; while(k < bufsize){
   p = *buffer++; k++;
   if(p < 0)
     for(i=0; i<= -p; i++){*output++ = *buffer++; k++;}
   else{
     for(i=0; i<=  p; i++) *output++ = *buffer;
     buffer++; k++;
   }
 }
}
#endif

static short read_16b_image(char *filename, short info_only){
 FILE *ipf;
 long x_x,y_y;
 long i,j;
 unsigned short Nt,RGB[1024];
 unsigned char *px;
 if((ipf=fopen(strlwr(filename),"rb")) == NULL)return 0;
 if( fread((char *)&x_x,sizeof(long),1,ipf) != 1 ||
     fread((char *)&y_y,sizeof(long),1,ipf) != 1)goto FATALREAD;
 if(info_only){
   x_size=(int)x_x;
   y_size=(int)y_y;
   fclose(ipf);
   return 1;
 }
 px=pixels;
 for(i=0;i<y_y;i++) {
   if(fread((char *)&Nt,sizeof(short),1,ipf) != 1)goto FATALREAD;
   if(Nt == 0){
     if(fread((char *)RGB,sizeof(unsigned short),x_x,ipf)
        != x_x)goto FATALREAD;
   }
   else{
     if(fread((char *)compressbuffer,sizeof(short),Nt,ipf) != Nt)
         goto FATALREAD;
     UnCompressWordLine(Nt,compressbuffer,RGB);
   }
   px = (pixels + (long)(y_size - 1 - i)*3*(long)(x_size));
   for(j=0;j<x_x;j++){
     *px++ = (unsigned char)(((unsigned short)(RGB[j] & 0x7c00) <<  1) >> 8);
     *px++ = (unsigned char)(((unsigned short)(RGB[j] & 0x03e0) <<  6) >> 8);
     *px++ = (unsigned char)(((unsigned short)(RGB[j] & 0x001f) << 11) >> 8);
   }
 }
 fclose(ipf);
 return 1;
 FATALREAD:
 fclose(ipf);
 return 0;
}

static unsigned char *Load16B(char *B16file, int *xx, int *yy){
 long imagesize;
 rowcount=0;
 if(read_16b_image(B16file,1) <= 0)return NULL;
 imagesize=x_size*y_size*3;
 if((pixels=(unsigned char*)X__Malloc(imagesize)) == NULL)return NULL;
 if(read_16b_image(B16file,0) <= 0){
   X__Free(pixels);
   return NULL;
 }
 *xx=x_size; *yy=y_size;
 return pixels;
}

static unsigned char *LoadMAP(char *CurrentFile, int *xx, int *yy){
 if     (strstr(CurrentFile,".GIF") || strstr(CurrentFile,".gif"))
   return LoadGIF(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".TGA") || strstr(CurrentFile,".tga"))
   return LoadTGA(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".16b") || strstr(CurrentFile,".16b"))
   return Load16B(CurrentFile,xx,yy);
 return NULL;
}

static BOOL TestMAP(char *CurrentFile, int *xx, int *yy){
 if     (strstr(CurrentFile,".GIF") || strstr(CurrentFile,".gif")){
   if(read_gif_image(CurrentFile,1) <= 0)return FALSE;
   *xx=x_size;
   *yy=y_size;
   return TRUE;
 }
 else if(strstr(CurrentFile,".TGA") || strstr(CurrentFile,".tga")){
   if(read_tga_image(CurrentFile,1) <= 0)return FALSE;
   *xx=x_size;
   *yy=y_size;
   return TRUE;
 }
 else if(strstr(CurrentFile,".16b") || strstr(CurrentFile,".16b")){
   if(read_16b_image(CurrentFile,1) <= 0)return FALSE;
   *xx=x_size;
   *yy=y_size;
   return TRUE;
 }
 return FALSE;
}

#include "..\..\render\flihead.h"

static Fli_head  Fhead;
static struct fli_head *fh;
static struct fli_frame *ff;
static struct fli_chunk *fc;

#include "..\..\render\flchead.h"

static Flc_head          FLC_Fhead;
static struct flc_head  *fhflc;
static struct flc_frame *ffflc;
static struct flc_chunk *fcflc;

static char     linebuffer2[1024];
static unsigned char linebuffer[2048];
static unsigned char gif_p_red[256],gif_p_green[256],gif_p_blue[256];

#define GOOD 2
#define GIFCOLOURS 256
#define NOCOLOUR 32768

static unsigned long   *colorstats0;
static unsigned short  *colorstats1;


static void Cshort(unsigned char *c, short v){
  unsigned char ch,cl;
  ch = (unsigned char)(v >> 8);
  cl = (unsigned char)(v & 0x00ff);
  *c++ = cl;
  *c++ = ch;
}

static void Clong(unsigned char *c, long v){
  unsigned char ch,c2,c1,cl;
  ch = (unsigned char)((v >> 24) & 0x000000ff);
  c2 = (unsigned char)((v >> 16) & 0x000000ff);
  c1 = (unsigned char)((v >>  8) & 0x000000ff);
  cl = (unsigned char)(v & 0x000000ff);
  *c++ = cl;
  *c++ = c1;
  *c++ = c2;
  *c++ = ch;
}

static void Ashort(unsigned char *c,short v){
  unsigned char ch,cl;
  short ah,al,a;
  al = (short)(*c);
  ah = (short)(*(c+1));
  a  = (ah << 8) | al;
  a += v;
  ch = (unsigned char)(a >> 8);
  cl = (unsigned char)(a & 0x00ff);
  *c     = cl;
  *(c+1) = ch;
}

static void Along(unsigned char *c,long v){
  unsigned char ch,c2,c1,cl;
  long ah,a2,a1,al,a;
  al = (long)(*c);
  a1 = (long)(*(c+1));
  a2 = (long)(*(c+2));
  ah = (long)(*(c+3));
  a  = (ah << 24) | (a2 << 16) | (a1 << 8) | al;
  a += v;
  ch = (unsigned char)((a >> 24) & 0x000000ff);
  c2 = (unsigned char)((a >> 16) & 0x000000ff);
  c1 = (unsigned char)((a >>  8) & 0x000000ff);
  cl = (unsigned char)(a & 0x000000ff);
  *c     = cl;
  *(c+1) = c1;
  *(c+2) = c2;
  *(c+3) = ch;
}

static long Tlong(unsigned char *c){
  long ah,a2,a1,al,a;
  al = (long)(*c);
  a1 = (long)(*(c+1));
  a2 = (long)(*(c+2));
  ah = (long)(*(c+3));
  a  = (ah << 24) | (a2 << 16) | (a1 << 8) | al;
  return a;
}

static short Tshort(unsigned char *c){
  short ah,al,a;
  al = (short)(*c);
  ah = (short)(*(c+1));
  a  = (ah << 8) | al;
  return a;
}

#include "..\..\render\buildfi.c"
#include "..\..\render\buildfc.c"

// code to read and uncompress the temporary file format

#ifdef _SUNSTYLE
void  UnCompressWordLine(bufsize,buffer,output)
short bufsize;short *buffer; unsigned short *output;{
#else
void  UnCompressWordLine(short bufsize, short *buffer,unsigned short *output){
#endif
 short p,k;
 long i;
 k=0; while(k < bufsize){
   p = *buffer++; k++;
   if(p < 0)
     for(i=0; i<= -p; i++){*output++ = *buffer++; k++;}
   else{
     for(i=0; i<=  p; i++) *output++ = *buffer;
     buffer++; k++;
   }
 }
}

#ifdef _SUNSTYLE
static long distance(argb,brgb) unsigned short argb; unsigned short brgb; {
#else
static long distance(unsigned short argb, unsigned short brgb){
#endif
    long b,g,r;
    b  = (long)( argb        & 0x1f);
    g  = (long)((argb >>  5) & 0x1f);
    r  = (long)((argb >> 10) & 0x1f);
    b -= (long)( brgb        & 0x1f);
    g -= (long)((brgb >>  5) & 0x1f);
    r -= (long)((brgb >> 10) & 0x1f);
    return (r*r + g*g + b*b);
/*    return labs(r) + labs(g) + labs(b);
*/
}

#ifdef _SUNSTYLE
long getcolor(crgb, Npalette) unsigned short crgb; long Npalette; {
#else
long getcolor(unsigned short crgb, long Npalette){
#endif
  long i,j,best,dist;
/* look for an exact match in the color table (or minimal distance) */
    for (i=0; i < Npalette; i++) {
       if (colorstats1[i] == crgb) return (i);
       if (i == 0) {
         best = 0;
         dist = distance(colorstats1[i],crgb);
       }
       else if ((j=distance(colorstats1[i],crgb)) < dist) {
         best = i;
         dist = j;
       }
    }
    return(best);
}

#ifdef _SUNSTYLE
void SortColours(Ngif,N_colours) long Ngif; long N_colours; {
#else
void SortColours(long Ngif, long N_colours){
#endif
 long i,j,gap,t0,maxdis,used;
 unsigned short t1;
 for (i = 0; i < N_colours; i++) {
     colorstats1[i] = i;
 }
 for (gap = N_colours/2; gap > 0; gap /=2) {
  for (i = gap; i < N_colours; i++) {
    for (j = i-gap; j >= 0 &&
        colorstats0[j] < colorstats0[j+gap]; j -= gap) {
      t0 = colorstats0[j];
      t1 = colorstats1[j];
      colorstats0[j] = colorstats0[j+gap];
      colorstats1[j] = colorstats1[j+gap];
      colorstats0[j+gap] = t0;
      colorstats1[j+gap] = t1;
    }
  }
 }
 for (used = 0; used < N_colours && colorstats0[used] > 0; used++){;}
 if (used > Ngif) {
  for (maxdis = 2; maxdis < 1024; maxdis *= 2) {
   for (i=Ngif/2; i < Ngif; i++) {
    for (j = 0; j < i; j++) {
      if (distance(colorstats1[i],colorstats1[j])<maxdis) {
       for (j=i+1; j< used; j++) {
         colorstats0[j-1] = colorstats0[j];
         colorstats1[j-1] = colorstats1[j];
       }
       --used;
       --i;
       break;
      }
    }
    if (used <= Ngif) break;
   }
   if (used <= Ngif) break;
  }
 }
 return;
}

// end of read and uncompress code

#define FLI 1
#define FLC 2

static void WriteFLI(void){
 FILE *opf,*tt;
 char filename[256],message[256];
 long nframes;
 BOOL fhalf=FALSE;
 DWORD dwRead;
 BITMAPINFOHEADER *lpbi;
 HPALETTE hpalHalftone=NULL;
 long i,j,k,imagesize,nx,ny,speed,size,type,cpix;
 unsigned short val;
 int xx,yy;
 unsigned char *p,*s,*Picture=NULL,r,g,b,
               *Screen1=NULL,*Screen2=NULL,*fli_buffer=NULL;
 if((nframes=SendMessage(hWndList,LB_GETCOUNT,(WPARAM)0,(LPARAM)0)) < 2){
   Query(IDI_FEWPICS);
   return;
 }
 SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,nframes));
 SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
 if(!bits24)fhalf=TRUE;
 else       fhalf=FALSE;
 nx=ny=0;
 for(i=0;i<nframes;i++){
  if(SendMessage(hWndList,LB_GETTEXT,(WPARAM)i,(LPARAM)filename) == LB_ERR){
    Query(IDI_BADLIST);
    return;
  }
  LoadString(glInstance,IDI_TESTING,stat_m,255);
  sprintf(message,stat_m,filename);
  SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)message);
  SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(i),0);
  YieldTo();
  if(!TestMAP(filename,&xx,&yy)){
    Query(IDI_BADFILE);
    return;
  }
  if(xx > nx)nx=xx;
  if(yy > ny)ny=yy;
 }
 if(nx > 640 || ny > 480){
   MessageBeep(MB_OK);
   return;
 }
 colorstats0=NULL; colorstats1=NULL;
 if((colorstats0 = (unsigned long *)
   X__Malloc(NOCOLOUR*(long)sizeof(unsigned long))) == NULL){
   MessageBeep(MB_OK);
   return;
 }
 if((colorstats1 = (unsigned short *)
   X__Malloc(NOCOLOUR*(long)sizeof(unsigned short))) == NULL){
   X__Free(colorstats0);
   MessageBeep(MB_OK);
   return;
 }
 for(i=0;i<NOCOLOUR;i++){
   colorstats0[i]=0; colorstats1[i]=0;
 }
 colorstats0[0]=65536*1024L;
 for(k=0; k<nframes; k++){ // build palette
   if(SendMessage(hWndList,LB_GETTEXT,(WPARAM)k,(LPARAM)filename) == LB_ERR){
     Query(IDI_BADLIST);
     break;
   }
   LoadString(glInstance,IDI_BUILDPAL,stat_m,255);
   sprintf(message,stat_m,filename);
   SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)message);
   SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(k),0);
   if((Picture=LoadMAP(filename,&xx,&yy)) == NULL){
     Query(IDI_BADFILE);
     break;
   }
   p=Picture;
   for(j=0;j<yy;j++){
     for(i=0;i<xx;i++){
       r = *p++; g = *p++; b = *p++;
       val=((unsigned short)(r & 0xf8) << 7)
          |((unsigned short)(g & 0xf8) << 2)
          |((unsigned short)(b       ) >> 3);
       colorstats0[(long)val]++;
     }
   }
   YieldTo();
   X__Free(Picture);
 }
 SortColours(GIFCOLOURS,NOCOLOUR);
 for(i=0;i<GIFCOLOURS;i++){
   gif_p_red[i]  =(unsigned char)((colorstats1[i] >> 10) & 0x1F) << 1;
   gif_p_green[i]=(unsigned char)((colorstats1[i] >>  5) & 0x1F) << 1;
   gif_p_blue[i] =(unsigned char)((colorstats1[i]      ) & 0x1F) << 1;
 }
 for(i=0;i<NOCOLOUR;i++)colorstats0[i] = GIFCOLOURS+1;
 if(nx > 320 || ny > 200){
   type=FLC; size=nx*ny;
 }
 else{
   type=FLI; size=64000;
 }
 speed=5;
 if((Screen1 = (unsigned char *)X__Malloc(size)) == NULL){
   Query(IDI_MEMORY);
   X__Free(colorstats0); X__Free(colorstats1);
   return;
 }
 memset(Screen1,'0',size);
 if((Screen2 = (unsigned char *)X__Malloc(size)) == NULL){
   Query(IDI_MEMORY);
   X__Free(Screen1);
   X__Free(colorstats0); X__Free(colorstats1);
   return;
 }
 memset(Screen2,'0',size);
 size += 10L*(long)nx;
 if((fli_buffer=(unsigned char *)X__Malloc(size)) == NULL){
   X__Free(Screen1); X__Free(Screen2);
   X__Free(colorstats0); X__Free(colorstats1);
   Query(IDI_MEMORY);
   return;
 }
 if(type == FLI){
   if((p=(unsigned char *)strrchr(szOutFileName,'.')) != NULL)*p='\0';
   AppendFileExtension(szOutFileName,".fli");
 }
 if((opf=fopen(strlwr(szOutFileName),"wb")) != NULL){
   for(k=0; k<=nframes; k++){ // build
     if(k == nframes)i=0;
     else            i=k;
     if(SendMessage(hWndList,LB_GETTEXT,(WPARAM)i,(LPARAM)filename) == LB_ERR){
       Query(IDI_BADLIST);
       break;
     }
     LoadString(glInstance,IDI_BUILDING,stat_m,255);
     sprintf(message,stat_m,filename);
     SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)message);
     SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(k),0);
     if((Picture=LoadMAP(filename,&xx,&yy)) == NULL){
       Query(IDI_BADFILE);
       break;
     }
     YieldTo();
     p=Picture;
     s=Screen1; memset(Screen1,0,nx*ny);
     for(j=0;j<yy;j++){
       s=(Screen1+(yy-j-1)*nx);
       for(i=0;i<xx;i++){
         r = *p++; g = *p++; b = *p++;
         val=((unsigned short)(r & 0xf8) << 7)
            |((unsigned short)(g & 0xf8) << 2)
            |((unsigned short)(b       ) >> 3);
         if((cpix = (long)colorstats0[(long)val]) > GIFCOLOURS){
           cpix=getcolor(val,GIFCOLOURS);
           colorstats0[(long)val]=(unsigned long)cpix;
         }
         *s++ = (unsigned char)cpix;
       }
     }
     X__Free(Picture);
     YieldTo();
     if(k == 0){
       if(type == FLI){
         if(Write_first_FLI_frame(opf,(short)speed,Screen1,Screen2,
            fli_buffer) < 0)goto FATALWRITE;
       }
       else if(type == FLC){
         if(Write_first_FLC_frame(opf,Screen1,Screen2,fli_buffer,(short)speed,
            (short)nx,(short)ny) < 0)goto FATALWRITE;
       }
     }
     else if(k == nframes){
       if(type == FLI){
         if(Write_FLI_frame(opf,Screen1,Screen2,fli_buffer) < 0)goto FATALWRITE;
         if(Close_FLI_file(opf) < 0)goto FATALWRITE;
       }
       else if(type == FLC){
         if(Write_FLC_frame(opf,Screen1,Screen2,fli_buffer,(short)nx,
            (short)ny) < 0)goto FATALWRITE;
         if(Close_FLC_file(opf) < 0)goto FATALWRITE;
       }
     }
     else{
       if(type == FLI){
         if(Write_FLI_frame(opf,Screen1,Screen2,fli_buffer) < 0)goto FATALWRITE;
       }
       else if(type == FLC){
         if(Write_FLC_frame(opf,Screen1,Screen2,fli_buffer,(short)nx,
            (short)ny) < 0)goto FATALWRITE;
       }
     }
     YieldTo();
   }
   goto XXIT;
   FATALWRITE:
   ;
   XXIT:
   fclose(opf);
 }
 X__Free(colorstats0);
 X__Free(colorstats1);
 X__Free(Screen1);
 X__Free(Screen1);
 X__Free(fli_buffer);
 return;
}

