/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file editavi.C || */

#define MODULE_MAIN

#define  STRICT

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <vfw.h>
#include "resource.h"

typedef struct tagFILELIST {
  long type;
  char name;
} FILELIST;

typedef struct tagFRAME {
  long nameID;
  long frame;
  unsigned char *image;
  BOOL selected;
} FRAME;

typedef struct tagSCRIPT {
  long first_frame,last_frame;
  long type;
  long mixID;
} SCRIPT;

#define AVIIF_KEYFRAME  0x00000010L // this frame is a key frame.
#define BUFSIZE         256
#define SIZESTATUS1     130
#define IMAGE_X_SIZE     60
#define IMAGE_Y_SIZE     45
#define CELL_X_SIZE      65
#define CELL_Y_SIZE      50
#define X_OFFSET          2
#define Y_OFFSET          2

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

#define RSF_Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define RSF_Free(p)      LocalFree((HLOCAL)p)
#define RSF_Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#define StartWait() hcurSave = SetCursor(LoadCursor(NULL,IDC_WAIT))
#define EndWait()   SetCursor(hcurSave)

static LONG WINAPI MainWndProc(HWND, UINT, WPARAM, LPARAM);
static LONG WINAPI ChildWndProc(HWND, UINT, WPARAM, LPARAM);
static BOOL CALLBACK AboutDlgProc(HWND,UINT,WPARAM,LPARAM);
static int SelectFileName(int, char *, char *, char *, char *, HWND);
static void AppendFileExtension(char *, char *);
static PSTR FileInPath(PSTR);
static void InitNew(void);
static void CleanUp(void);
static void DrawChildWindow(HDC);
static BOOL GetCurrentFrame(long, long, long *, long *);
static void UpdateFrameStatus(long, long);

FILELIST *FileList=NULL;
FRAME    *Channel1=NULL,*Channel2=NULL;
SCRIPT   *Script=NULL;
long     nFrames1,nFrames2,FirstFrame,nDisplayFrames;

static BOOL       debug=FALSE;
static HWND       ghWndFrame=NULL,hWndChild=NULL,
                  hWndStatus=NULL,hWndProgress=NULL;
static HINSTANCE  ghInstance;
static CHAR       szAppName[]="OFX:EditVideoClass";
static CHAR       szWindowClass[]="OFX:EditVideoChildClass";
static CHAR       szVideoDirName[256],szAudioDirName[256];
static CHAR       szVideoFile[256],szAudioFile[256];
static CHAR       szAudioTrackFile[256],szWorkFile[256];
static CHAR       gszHomeDir[256],gszWork[256],*gszWorkDrive;
static CHAR       szOutFile[256],szOutDirName[256];
static CHAR       message[256],stat_m[256];
static DWORD      fdwStyle;
static UINT       ToolbarHeight=0,StatusbarHeight=0,ProgressbarHeight=20;
static HWND       hTBWnd=NULL;
static HCURSOR    hcurSave;
static int        TotalFrames=0,CurrentFrame=0;
static long       sAVIc=1,sAVIf=11025,sAVIb=1,sWAVc=1,sWAVf=11025,sWAVb=1;

#define NTOOLBARBUTTONS 2
#define NOTOOLBARIMAGES 1

static TBBUTTON tbb[NTOOLBARBUTTONS] = {
          { 0,0,           TBSTATE_ENABLED,TBSTYLE_SEP,   0,0 },
          { 0,IDM_OPEN,    TBSTATE_ENABLED,TBSTYLE_BUTTON,0,0 }
                                        };

static void AppendFileExtension(char *filename, char *ext){
 if(strrchr(filename,'.') == NULL){
   strcat(filename,ext);
 }
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[256],szdummy[256];
 szdummy[0]='\0';
 if(szfilter == NULL)AVIBuildFilter(szFilter,255,FALSE);
 else{
   strcpy(szFilter,szfilter);
   i=0; while(szFilter[i] != '\0'){
     if(szFilter[i] == '|')szFilter[i]='\0'; i++;
   }
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=256;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=256;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileNamePreview(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

static void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static BOOL CALLBACK AboutDlgProc(HWND hwnd,UINT msg,
                                  WPARAM wparam,LPARAM lparam){
 switch(msg){
   case WM_INITDIALOG:
     CentreDialogOnScreen(hwnd);
     return( TRUE );
   case WM_COMMAND:
      if(LOWORD(wparam) == IDOK){
        EndDialog(hwnd,TRUE);
        return(TRUE);
      }
      else if(LOWORD(wparam) == IDCANCEL){
        EndDialog(hwnd,FALSE);
        return(TRUE);
      }
      break;
 }
 return( FALSE );
}

void YieldTo(void){
 MSG msg;
 while (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE)){
   TranslateMessage(&msg);
   DispatchMessage(&msg);
 }
}

void Query(int id){
  char mess[256];
  LoadString(ghInstance,id,mess,255);
  MessageBox(ghWndFrame,mess,NULL,MB_OK);
}

void SetToolbarState(BOOL state){
 int i;
 for(i=0;i<NTOOLBARBUTTONS;i++)SendMessage(hTBWnd,
   TB_ENABLEBUTTON,(WPARAM)tbb[i].idCommand,(LPARAM)MAKELONG(state,0));
 YieldTo();
}

static LONG CALLBACK my_save_callback(int n_percent){
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)n_percent,0);
 YieldTo();
 return AVIERR_OK;
}

void RegisterMainClass(HINSTANCE hInstance){
 WNDCLASS   wc;
 wc.style         = 0;
 wc.lpfnWndProc   = (WNDPROC)MainWndProc;
 wc.cbClsExtra    = 0;
 wc.cbWndExtra    = 0;
 wc.hInstance     = hInstance;
 wc.hIcon         = LoadIcon(hInstance,"SHOWICON");
 wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
 wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
 wc.lpszMenuName  = "EditAviMenu";
 wc.lpszClassName = szAppName;
 ghInstance=hInstance;
 if(!RegisterClass(&wc))return;
 wc.style         = 0;
 wc.lpfnWndProc   = (WNDPROC)ChildWndProc;
 wc.cbClsExtra    = 0;
 wc.cbWndExtra    = 0;
 wc.hInstance     = hInstance;
 wc.hIcon         = NULL;
 wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
 wc.hbrBackground = (HBRUSH)(COLOR_BTNSHADOW+1);
 wc.lpszMenuName  = NULL;
 wc.lpszClassName = szWindowClass;
 RegisterClass(&wc);
 return;
}

void UnregisterMainClass(HINSTANCE hInstance){
 UnregisterClass(szWindowClass,hInstance);
 UnregisterClass(szAppName,hInstance);
}

static HWND SetupMainWindow(HWND parent, LPRECT pr){
 RECT rc;
 HWND hWnd;
 GetModuleFileName(ghInstance,gszHomeDir,256);
 *FileInPath(gszHomeDir) = '\0';

 strcpy(szVideoDirName,gszHomeDir); strcat(szVideoDirName,"flics");
 strcpy(szAudioDirName,gszHomeDir); strcat(szAudioDirName,"sounds");

 strcpy(szOutFile,"temp");

 strcpy(szOutDirName,gszHomeDir); strcat(szOutDirName,"flics");

 szVideoFile[0]='\0'; szAudioFile[0]='\0'; szAudioTrackFile[0]='\0';

 if(GetEnvironmentVariable((LPTSTR)"TEMP",(LPTSTR)gszWork,255) == 0){
   strcpy(gszWork,gszHomeDir);
 }
 strcpy(szWorkFile,gszWork);
 if(*(szWorkFile+strlen(szWorkFile)-1) != '\\')strcat(szWorkFile,"\\");
 strcat(szWorkFile,"$_RSF_$.AVI");
 gszWorkDrive=&gszWork[0];
 InitNew();
// DeleteFile(szWorkFile);

 fdwStyle = WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
            WS_SYSMENU | !WS_MAXIMIZEBOX | WS_MINIMIZEBOX |
            WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
 LoadString(ghInstance,IDI_TITLE,stat_m,255);
 hWnd = CreateWindow(szAppName,
             stat_m,
             fdwStyle,
             pr->left,
             pr->top,
             pr->right,
             pr->bottom,
             parent,
             NULL,
             ghInstance,
             NULL);
 if(!hWnd)return NULL;
 ghWndFrame=hWnd;
 ShowWindow(hWnd,SW_SHOW);
 UpdateWindow(hWnd);
 return hWnd;
}

static void InitNew(void){
 CleanUp();
 Channel1=(FRAME *)RSF_Malloc(sizeof(FRAME));
 if(Channel1 == NULL){ Query(IDI_MEMORY); return; }
 Channel1->nameID   = -1;
 Channel1->frame    = -1;
 Channel1->image    = NULL;
 Channel1->selected = FALSE;
 Channel2=(FRAME *)RSF_Malloc(sizeof(FRAME));
 if(Channel2 == NULL){ Query(IDI_MEMORY); return; }
 Channel2->nameID   = -1;
 Channel2->frame    = -1;
 Channel2->image    = NULL;
 Channel2->selected = FALSE;
 nFrames1=1;
 nFrames2=1;
 return;
}

static void CleanUp(void){
 int i;
 if(FileList != NULL)RSF_Free(FileList); FileList=NULL;
 if(Channel1 != NULL){
   for(i=0;i<nFrames1;i++)
     if((Channel1+i)->image != NULL)RSF_Free((Channel1+i)->image);
   RSF_Free(Channel1); Channel1=NULL;
 }
 if(Channel2 != NULL){
   for(i=0;i<nFrames2;i++)
     if((Channel2+i)->image != NULL)RSF_Free((Channel2+i)->image);
   RSF_Free(Channel2); Channel2=NULL;
 }
 if(Script   != NULL)RSF_Free(Script);   Script=NULL;
 nFrames1=0; nFrames2=0; FirstFrame=0;
}

static void UpdateFrameStatus(long f, long c){
 char str[64];
 LoadString(ghInstance,IDI_FRAMEFORMAT,str,64);
 sprintf(stat_m,str,f);
 SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)1,(LPARAM)stat_m);
 return;
}

static void Commands(WORD cmd,HWND hWnd){
 int i;
 RECT rc;
 char temp[256],temp1[256];
 switch (cmd){
   case IDM_OPEN:
     break;
   case IDM_HELP:
     DialogBox(ghInstance,MAKEINTRESOURCE(IDD_ABOUT),hWnd,AboutDlgProc);
     break;
   case IDM_EXIT:
     PostQuitMessage(0);
     break;
   default:
     break;
 }
}

static LONG WINAPI MainWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam){
 RECT rc;
 LONG lRet = 1;
 int StatusPartsList[2];
 switch (uMsg){
   case WM_CREATE:{
       HMENU hmenu=GetSystemMenu(hWnd,FALSE);
       AVIFileInit();
       hTBWnd = CreateToolbarEx(
         hWnd,
         WS_CHILD | WS_VISIBLE | !WS_BORDER | TBSTYLE_TOOLTIPS,
         0,
         NOTOOLBARIMAGES,
         ghInstance,
         TOOLBAR_IDBM,
         tbb,
         NTOOLBARBUTTONS,
         16,15,
         16,15,
         sizeof(TBBUTTON)
       );
       GetClientRect(hTBWnd,&rc); ToolbarHeight=rc.bottom-rc.top;
       hWndStatus=CreateWindowEx(0L,STATUSCLASSNAME,"",
         WS_CHILD | !WS_BORDER | WS_VISIBLE,
         0,0,0,0,
         hWnd,
         (HMENU)102,
         ghInstance,
         NULL);
       GetClientRect(hWndStatus,&rc); StatusbarHeight=rc.bottom-rc.top;
       StatusPartsList[0]=rc.right-SIZESTATUS1;
       StatusPartsList[1]= -1;
       SendMessage(hWndStatus,SB_SETPARTS,(WPARAM)2,(LPARAM)StatusPartsList);
       GetClientRect(hWnd,&rc);
       rc.top = rc.bottom-StatusbarHeight-ProgressbarHeight;
       hWndProgress=CreateWindowEx(0L,PROGRESS_CLASS,"",
         WS_CHILD | !WS_BORDER | WS_VISIBLE,
         0,rc.top+2,
         rc.right,ProgressbarHeight-2,
         hWnd,
         (HMENU)103,
         ghInstance,
         NULL);
       GetClientRect(hWnd,&rc);
       hWndChild=CreateWindow(szWindowClass," ",
         WS_CHILD | WS_BORDER | WS_VISIBLE | WS_HSCROLL,
         0,rc.top+ToolbarHeight,
         rc.right,rc.bottom-ProgressbarHeight-StatusbarHeight-ToolbarHeight,
         hWnd,
         (HMENU)101,
         ghInstance,
         NULL);
       LoadString(ghInstance,IDI_FILESLIST,stat_m,255);
       SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)stat_m);
     }
     break;
   case WM_DESTROY:
     AVIFileExit();
     break;
   case MCIWNDM_NOTIFYMODE:
     break;
   case WM_SYSCOMMAND:
     switch(LOWORD(wParam & 0xfff0)){
       default:
         return( DefWindowProc(hWnd,uMsg,wParam,lParam));
      }
      break;
   case WM_SIZING:{
       RECT *lprect;
       lprect=(RECT *)lParam;
       GetWindowRect(hWnd,&rc);
       lprect->top=rc.top;
       lprect->bottom=rc.bottom;
     }
     return TRUE;
   case WM_SIZE:
     nDisplayFrames=LOWORD(lParam)/CELL_X_SIZE + 1;
     if(hTBWnd != NULL)SendMessage(hTBWnd,uMsg,wParam,lParam);
     if(hWndStatus != NULL){
       SendMessage(hWndStatus,uMsg,wParam,lParam);
       StatusPartsList[0]=LOWORD(lParam)-SIZESTATUS1;
       StatusPartsList[1]= -1;
       SendMessage(hWndStatus,SB_SETPARTS,(WPARAM)2,(LPARAM)StatusPartsList);
     }
     if(hWndChild != NULL){
       rc.top=0; rc.left=0;
       rc.right=LOWORD(lParam); rc.bottom=HIWORD(lParam);
       SetWindowPos(hWndChild,NULL,
         0,rc.top+ToolbarHeight,
         rc.right,rc.bottom-ProgressbarHeight-StatusbarHeight-ToolbarHeight,
         SWP_NOZORDER);
     }
     rc.top=0; rc.left=0;
     rc.right=LOWORD(lParam); rc.bottom=HIWORD(lParam);
     rc.top = rc.bottom-StatusbarHeight-ProgressbarHeight;
     if(hWndProgress != NULL)SetWindowPos(hWndProgress,NULL,
          0,rc.top+2,
          rc.right,ProgressbarHeight-2,
          SWP_NOZORDER);
      break;
   case WM_NOTIFY:
     switch (((LPNMHDR)lParam)->code){
       case TTN_NEEDTEXT:{
           int idButton;
           LPTOOLTIPTEXT lpttt;
           lpttt=(LPTOOLTIPTEXT)lParam;
           lpttt->hinst=ghInstance;
           idButton=lpttt->hdr.idFrom;
           switch(idButton){
             case IDM_OPEN:
               lpttt->lpszText=MAKEINTRESOURCE(IDI_TIP_1); break;
           }
         }
         break;
       default: break;
     }
     return (DefWindowProc(hWnd,uMsg,wParam,lParam));
   case WM_COMMAND:
     Commands(LOWORD(wParam),hWnd);
     break;
   case WM_CLOSE:
     PostQuitMessage(0);
     break;
   default:
     lRet=DefWindowProc(hWnd,uMsg,wParam,lParam);
   break;
 }
 return lRet;
}

static void DrawChildWindow(HDC hdc){
 long i;
 long x0=X_OFFSET,x,y=Y_OFFSET;
 char str[32];
 FRAME *f;
 SetBkMode(hdc,TRANSPARENT);
 SelectObject(hdc,GetStockObject(NULL_BRUSH));
 SelectObject(hdc,GetStockObject(BLACK_PEN));
 x=x0;
 for(i=FirstFrame;i<FirstFrame+nDisplayFrames;i++){
   sprintf(str,"%ld",i);
   TextOut(hdc,x,y,str,strlen(str));
   x+=CELL_X_SIZE;
 }
 x=x0; y=y+20;
 for(i=FirstFrame;i<FirstFrame+nDisplayFrames;i++){
   if(i < nFrames1){
     f=(Channel1+i);
     if(f->selected)SelectObject(hdc,GetStockObject(WHITE_PEN));
     else           SelectObject(hdc,GetStockObject(BLACK_PEN));
     Rectangle(hdc,x,y,x+IMAGE_X_SIZE,y+IMAGE_Y_SIZE);
// if image
   }
   x+=CELL_X_SIZE;
 }
 x=x0; y+=CELL_Y_SIZE;
 for(i=FirstFrame;i<FirstFrame+nDisplayFrames;i++){
   if(i < nFrames2){
     f=(Channel2+i);
     if(f->selected)SelectObject(hdc,GetStockObject(WHITE_PEN));
     else           SelectObject(hdc,GetStockObject(BLACK_PEN));
     Rectangle(hdc,x,y,x+IMAGE_X_SIZE,y+IMAGE_Y_SIZE);
// if image
   }
   x+=CELL_X_SIZE;
 }
}

static BOOL GetCurrentFrame(long x, long y, long *f, long *c){
 if     (y < Y_OFFSET+20)return FALSE;
 else if(y < Y_OFFSET+20+IMAGE_Y_SIZE){
   *c=1;
   *f=(x-X_OFFSET)/CELL_X_SIZE;
   *f=max(*f,0);
   *f += FirstFrame;
   return TRUE;
 }
 else if(y < Y_OFFSET+20+CELL_Y_SIZE)return FALSE;
 else if(y < Y_OFFSET+20+CELL_Y_SIZE+IMAGE_Y_SIZE){
   *c=2;
   *f=(x-X_OFFSET)/CELL_X_SIZE;
   *f=max(*f,0);
   *f += FirstFrame;
   return TRUE;
 }
 return FALSE;
}

static LONG WINAPI ChildWndProc(HWND hWnd,UINT uMsg,
                                 WPARAM wParam,LPARAM lParam){
 static BOOL bCaptured=FALSE;
 long lRet=1,f,c;
 switch (uMsg){
   case WM_CREATE:
 SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,100));
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)50,0);
     break;
   case WM_PAINT:{
       HDC hdc;
       PAINTSTRUCT ps;
       hdc=BeginPaint(hWnd,&ps);
       DrawChildWindow(hdc);
       EndPaint(hWnd,&ps);
     }
     break;
   case WM_LBUTTONDOWN:
     SetCapture(hWnd);
     bCaptured=TRUE;
     break;
   case WM_LBUTTONUP:
     ReleaseCapture();
     bCaptured=FALSE;
     break;
   case WM_MOUSEMOVE:
     if(GetCurrentFrame((long)LOWORD(lParam),(long)HIWORD(lParam),&f,&c)){
       UpdateFrameStatus(f,c);
     }
     break;
   case WM_DESTROY:
     break;
   default:
     lRet=DefWindowProc(hWnd,uMsg,wParam,lParam);
   break;
 }
 return lRet;
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 MSG msg;
 RECT rc;
 InitCommonControls();
 if(!MCIWndRegisterClass())return FALSE;
 RegisterMainClass(hInstance);
 rc.left=20; rc.top=40;
 rc.right=GetSystemMetrics(SM_CXSCREEN)-2*rc.left;
 rc.bottom=280;
 nDisplayFrames=GetSystemMetrics(SM_CXSCREEN)/CELL_X_SIZE + 1;
 SetupMainWindow(NULL,&rc);
 while (GetMessage(&msg,NULL,0,0)){
   TranslateMessage(&msg);
   DispatchMessage(&msg);
 }
 DestroyWindow(ghWndFrame);
 UnregisterMainClass(hInstance);
 CleanUp();
 return TRUE;
}


