/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* utils.c */

static void AppendFileExtension(char *filename, char *ext){
 if(strrchr(filename,'.') == NULL){
   strcat(filename,ext);
 }
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static void UpdateTextInfo(void){
 if(ps_video[0] != NULL)
 sprintf(message,"Frame %ld of %ld (%.1lfs)",CurrentFrame,TotalFrames,
 (double)AVIStreamSampleToTime(ps_video[0],CurrentFrame)*0.001);
 else
 sprintf(message,"Frame %ld of %ld",CurrentFrame,TotalFrames);
 SendMessage(hWndText,WM_SETTEXT,0,(LPARAM)message);
}

static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[256],szdummy[256];
 szdummy[0]='\0';
 if(szfilter == NULL)AVIBuildFilter(szFilter,255,FALSE);
 else{
   strcpy(szFilter,szfilter);
   i=0; while(szFilter[i] != '\0'){
     if(szFilter[i] == '|')szFilter[i]='\0'; i++;
   }
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=256;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=256;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileNamePreview(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

static BOOL CALLBACK AboutDlgProc(HWND hwnd,UINT msg,
                                  WPARAM wparam,LPARAM lparam){
 switch(msg){
   case WM_INITDIALOG:
     CentreDialogOnScreen(hwnd);
     return( TRUE );
   case WM_COMMAND:
      if(LOWORD(wparam) == IDOK){
        EndDialog(hwnd,TRUE);
        return(TRUE);
      }
      else if(LOWORD(wparam) == IDCANCEL){
        EndDialog(hwnd,FALSE);
        return(TRUE);
      }
      break;
 }
 return( FALSE );
}

static BOOL CALLBACK InfoDlgProc(HWND hwnd,UINT msg,
                                 WPARAM wparam,LPARAM lparam){
 char temp[32];
 switch(msg){
   case WM_INITDIALOG:
     if(sAVIc == 1)SetDlgItemText(hwnd,DLG_AUDIO_INFO_AVI_CHANNELS,"Mono");
     if(sAVIc == 2)SetDlgItemText(hwnd,DLG_AUDIO_INFO_AVI_CHANNELS,"Stereo");
     if(sAVIb == 1)SetDlgItemText(hwnd,DLG_AUDIO_INFO_AVI_BPS,"8 Bits");
     if(sAVIb == 2)SetDlgItemText(hwnd,DLG_AUDIO_INFO_AVI_BPS,"16 Bits");
     sprintf(temp,"%.3lf kHz",((double)sAVIf)*0.001);
     SetDlgItemText(hwnd,DLG_AUDIO_INFO_AVI_FREQ,temp);
     if(sWAVc == 1)SetDlgItemText(hwnd,DLG_AUDIO_INFO_WAV_CHANNELS,"Mono");
     if(sWAVc == 2)SetDlgItemText(hwnd,DLG_AUDIO_INFO_WAV_CHANNELS,"Stereo");
     if(sWAVb == 1)SetDlgItemText(hwnd,DLG_AUDIO_INFO_WAV_BPS,"8 Bits");
     if(sWAVb == 2)SetDlgItemText(hwnd,DLG_AUDIO_INFO_WAV_BPS,"16 Bits");
     sprintf(temp,"%.3lf kHz",((double)sWAVf)*0.001);
     SetDlgItemText(hwnd,DLG_AUDIO_INFO_WAV_FREQ,temp);
     CentreDialogOnScreen(hwnd);
     return( TRUE );
   case WM_COMMAND:
      if(LOWORD(wparam) == IDOK){
        EndDialog(hwnd,TRUE);
        return(TRUE);
      }
      else if(LOWORD(wparam) == IDCANCEL){
        EndDialog(hwnd,FALSE);
        return(TRUE);
      }
      break;
 }
 return( FALSE );
}

static BOOL CALLBACK StreamDlgProc(HWND hwnd,UINT msg,
                                  WPARAM wparam,LPARAM lparam){
 switch(msg){
   case WM_INITDIALOG:
     if     (sAVIc == 1)
       SendDlgItemMessage(hwnd,DLG_AUDIO_SET_CHAN1,BM_SETCHECK,TRUE,0);
     else if(sAVIc == 2)
       SendDlgItemMessage(hwnd,DLG_AUDIO_SET_CHAN2,BM_SETCHECK,TRUE,0);
     if     (sAVIb == 1)
       SendDlgItemMessage(hwnd,DLG_AUDIO_SET_BPS8,BM_SETCHECK,TRUE,0);
     else if(sAVIb == 2)
       SendDlgItemMessage(hwnd,DLG_AUDIO_SET_BPS16,BM_SETCHECK,TRUE,0);
     if     (sAVIf == 11025)
       SendDlgItemMessage(hwnd,DLG_AUDIO_SET_FREQ1,BM_SETCHECK,TRUE,0);
     else if(sAVIf == 22050)
       SendDlgItemMessage(hwnd,DLG_AUDIO_SET_FREQ2,BM_SETCHECK,TRUE,0);
     else if(sAVIf == 44100)
       SendDlgItemMessage(hwnd,DLG_AUDIO_SET_FREQ3,BM_SETCHECK,TRUE,0);
     CentreDialogOnScreen(hwnd);
     return( TRUE );
   case WM_COMMAND:
      if(LOWORD(wparam) == IDOK){
        if     (SendDlgItemMessage(hwnd,DLG_AUDIO_SET_CHAN2,BM_GETCHECK,0,0))
               sAVIc=2;
        else   sAVIc=1;
        if     (SendDlgItemMessage(hwnd,DLG_AUDIO_SET_BPS16,BM_GETCHECK,0,0))
               sAVIb=2;
        else   sAVIb=1;
        if     (SendDlgItemMessage(hwnd,DLG_AUDIO_SET_FREQ3,BM_GETCHECK,0,0))
               sAVIf=44100;
        else if(SendDlgItemMessage(hwnd,DLG_AUDIO_SET_FREQ2,BM_GETCHECK,0,0))
               sAVIf=22050;
        else   sAVIf=11025;
        EndDialog(hwnd,TRUE);
        return(TRUE);
      }
      else if(LOWORD(wparam) == IDCANCEL){
        EndDialog(hwnd,FALSE);
        return(TRUE);
      }
      break;
 }
 return( FALSE );
}

static void UpdateInfo(void){
 strcpy(message," ");
 if(szVideoInfo[0] != '\0')strcat(message,szVideoInfo);
 if(szAudioInfo[0] != '\0')strcat(message,szAudioInfo);
 strcat(message,"   ");
 if(szSoundInfo[0] != '\0')strcat(message,szSoundInfo);
 SendMessage(hWndStatus,SB_SETTEXT,(WPARAM)0,(LPARAM)message);
}

static void GetInfo(BOOL video, char *name){
 LPBITMAPINFOHEADER lpBmi;
 LPWAVEFORMATEX  lpWavx;
 LPSTR pformat;
 HRESULT hr;
 PAVIFILE pfile;
 PAVISTREAM psin=NULL;
 PGETFRAME  pget;
 AVISTREAMINFO avis;
 BOOL bDone=FALSE;
 long i,size,nsamples;
 if(video){szAudioInfo[0]='\0'; szVideoInfo[0]='\0';}
 else      szSoundInfo[0]='\0';
 hr = AVIFileOpen(&pfile,name,OF_READ,NULL);
 if(hr != AVIERR_OK)pfile=NULL;
 if(hr == AVIERR_OK){
   for(i=0;i<10;i++){
     if(AVIFileGetStream(pfile,&psin,0,i) == AVIERR_OK){
       AVIStreamInfo(psin,&avis,sizeof(avis));
       if(video && avis.fccType == streamtypeVIDEO){
         AVIStreamFormatSize(psin,0,&size);
         if((pformat = (LPSTR)X__Malloc(size)) != NULL){
           if(AVIStreamReadFormat(psin,0,pformat,&size) == AVIERR_OK){
             lpBmi=(LPBITMAPINFOHEADER)pformat;
             nsamples=AVIStreamEnd(psin);
             sprintf(szVideoInfo,
                   "Video: (%ldx%ldx%ld %ld Frames %.1lfs)",
                   lpBmi->biWidth,lpBmi->biHeight,
                   (long)(lpBmi->biBitCount),nsamples,
                   (double)AVIStreamSampleToTime(psin,nsamples)*0.001);
           }
           X__Free(pformat);
         }
       }
       else if(avis.fccType == streamtypeAUDIO){
         AVIStreamFormatSize(psin,0,&size);
         if((pformat = (LPSTR)X__Malloc(size)) != NULL){
           if(AVIStreamReadFormat(psin,0,pformat,&size) == AVIERR_OK){
             lpWavx=(LPWAVEFORMATEX)pformat;
             nsamples=AVIStreamEnd(psin);
             if(video){
               sprintf(szAudioInfo,
                  "[%ld Ch %.2lf kHz %.1lfs] ",
                   (WORD)lpWavx->nChannels,
                   (double)lpWavx->nSamplesPerSec * 0.001,
                   (double)AVIStreamSampleToTime(psin,nsamples)*0.001);
               sAVIc=lpWavx->nChannels;
               sAVIf=lpWavx->nSamplesPerSec;
               sAVIb=lpWavx->wBitsPerSample/8;
             }
             else{
               sprintf(szSoundInfo,
                  "Sound: %ld Ch %.2lf kHz %.1lfs %ld",
                   (WORD)lpWavx->nChannels,
                   (double)lpWavx->nSamplesPerSec * 0.001,
                   (double)AVIStreamSampleToTime(psin,nsamples)*0.001,
                   (long)lpWavx->wBitsPerSample);
               sWAVc=lpWavx->nChannels;
               sWAVf=lpWavx->nSamplesPerSec;
               sWAVb=lpWavx->wBitsPerSample/8;
               if(lpWavx->wFormatTag == WAVE_FORMAT_ADPCM)
               MessageBox(NULL,"ADPCM format - cannot use",NULL,MB_OK);
               else if(lpWavx->wFormatTag != WAVE_FORMAT_PCM)
               MessageBox(NULL,"Not PCM wave format - cannot use",NULL,MB_OK);
             }
             bDone=TRUE;
           }
           X__Free(pformat);
         }
       }
       AVIStreamRelease(psin);
     }
     if(bDone)break;
   }
 }
 if(pfile != NULL)AVIFileRelease(pfile);
 UpdateInfo();
}

static void WriteAudioTrack(void){
 unsigned char *buffer;
 WAVEFORMATEX af;
 AVISTREAMINFO strhdr;
 PAVIFILE pfile=NULL;
 PAVISTREAM ps;
 HRESULT hr;
 long i,size,samps,bufsize;
 hr=AVIFileOpen(&pfile,
                szAudioTrackFile,
                OF_WRITE | OF_CREATE,
                NULL);
 if(hr != AVIERR_OK){ MessageBox(NULL,"BAD Audio File",NULL,MB_OK); return; }
 AVIStreamInfo(ps_video[1],&strhdr,sizeof(strhdr));
 hr=AVIFileCreateStream(pfile,&ps,&strhdr);
 if(hr != AVIERR_OK){ MessageBox(NULL,"BAD Stream",NULL,MB_OK); return; }
 memset(&af,0,sizeof(af));
 size=sizeof(PCMWAVEFORMAT);
 if(AVIStreamReadFormat(ps_video[1],0,&af,&size) != AVIERR_OK){
   MessageBox(NULL,"BAD Read Format",NULL,MB_OK); return;
 }
 hr=AVIStreamSetFormat(ps,0,&af,sizeof(af)-2); // size PCMWAVEFO..
 if (hr != AVIERR_OK){ MessageBox(NULL,"BAD SetFormat",NULL,MB_OK); return; }
 samps=AVIStreamEnd(ps_video[1]);
 bufsize=samps*af.nBlockAlign;
 buffer=(unsigned char*)X__Malloc(bufsize*sizeof(unsigned char *));
 if(buffer != NULL){
   if(AVIStreamRead(ps_video[1],0,samps,buffer,bufsize,NULL,NULL) != 0)
     MessageBox(NULL,"BAD Stream Read",NULL,MB_OK);
   if(AVIStreamWrite(ps,0,samps,buffer,bufsize,AVIIF_KEYFRAME,NULL,NULL) != 0)
     MessageBox(NULL,"BAD Write",NULL,MB_OK);
   X__Free(buffer);
 }
 else MessageBox(NULL,"Memory Error",NULL,MB_OK);
#if 0
 SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,samps));
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
 for(i=0;i<samps;i++){
   if(AVIStreamRead(ps_video[1],i,1,buffer,af.nBlockAlign,NULL,NULL) != 0){
     MessageBox(NULL,"BAD Stream Read",NULL,MB_OK); break;
   }
   if(AVIStreamWrite(ps,i,1,buffer,af.nBlockAlign,AVIIF_KEYFRAME,NULL,NULL) != 0){
     MessageBox(NULL,"BAD Write",NULL,MB_OK); break;
   }
   SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)i,0);
 }
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
#endif
 AVIStreamRelease(ps);
 AVIFileRelease(pfile);
}

static void MakeBlankSoundFile(long time, long chans, long freq, long bps){
 int i;
 unsigned char *buffer;
 WAVEFORMATEX af;
 AVISTREAMINFO strhdr;
 PAVIFILE pfile=NULL;
 PAVISTREAM ps;
 HRESULT hr;
 long samps,bytes;
 long bufsize;
 hr=AVIFileOpen(&pfile,
          szAudioTrackFile,
          OF_WRITE | OF_CREATE,
          NULL);
 if(hr != AVIERR_OK){
   MessageBox(NULL,"BAD Wave File",NULL,MB_OK);
   return;
 }
 memset(&strhdr,0,sizeof(strhdr));
 strhdr.fccType       = streamtypeAUDIO;
 strhdr.dwScale       = 1;
 strhdr.dwRate        = freq;
 strhdr.dwSampleSize  = 1;
 strhdr.dwQuality    = -1;
 strcpy(strhdr.szName,"Audio Track");
 hr=AVIFileCreateStream(pfile,&ps,&strhdr);
 if(hr != AVIERR_OK){
   MessageBox(NULL,"BAD Stream",NULL,MB_OK);
   return;
 }
 memset(&af,0,sizeof(af));
 af.wFormatTag=WAVE_FORMAT_PCM;
 af.nChannels=chans;
 af.nSamplesPerSec=freq;
 af.wBitsPerSample=bps;
 af.nBlockAlign=af.nChannels*af.wBitsPerSample/8;
 af.nAvgBytesPerSec=af.nSamplesPerSec*af.nBlockAlign;
 hr=AVIStreamSetFormat(ps,0,&af,sizeof(af)-2); // size PCMWAVEFO..
 if (hr != AVIERR_OK){
   MessageBox(NULL,"BAD SetFormat",NULL,MB_OK);
   return;
 }
 bufsize=af.nSamplesPerSec*time*af.nBlockAlign;
 buffer=(unsigned char *)X__Malloc(bufsize*sizeof(unsigned char));
 if(buffer != NULL){
   if(af.wBitsPerSample == 8){
     //  buffer[0]=127; // left  buffer[1]=127; // right
     memset(buffer,127,bufsize);
   }
   else{  // 16 bits
     //  buffer[0]=0; // left   buffer[1]=0; // left
     //  buffer[2]=0; // right  buffer[3]=0; // right
     memset(buffer,0,bufsize);
   }
   if(AVIStreamWrite(ps,0,
                     af.nSamplesPerSec*time,
                     buffer,
                     af.nSamplesPerSec*time*af.nBlockAlign,
                     AVIIF_KEYFRAME,&samps,&bytes) != 0)
     MessageBox(NULL,"BAD Write",NULL,MB_OK);
#if 0
   SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,af.nSamplesPerSec*time));
   SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
   for(i=0;i<af.nSamplesPerSec*time;i++){
     if(AVIStreamWrite(ps,i,1,buffer,af.nBlockAlign,
        AVIIF_KEYFRAME,&samps,&bytes) != 0){
       MessageBox(NULL,"BAD Write",NULL,MB_OK);
       break;
     }
     SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)i,0);
   }
#endif
   X__Free(buffer);
 }
 else MessageBox(NULL,"Memory Error",NULL,MB_OK);
 AVIStreamRelease(ps);
 AVIFileRelease(pfile);
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
}

static BOOL CopyAVIfile(char *szV, char *szW){
 LPBITMAPINFOHEADER lpBmi;
 LPWAVEFORMATEX  lpWavx;
 LPSTR pformat,pbuffer,sbuffer;
 HRESULT hr,hr1,hr2;
 PAVIFILE pinfile,poutfile;
 PAVISTREAM psin_v=NULL,psin_a=NULL,psout_v=NULL,psout_a=NULL;
 AVISTREAMINFO avis;
 long size,nsamples,nsa,key,vstream,astream,lBytes,lSamples;
 long TRST,TRET,TRSS,TRES;
 int i,j,k;
 vstream = -1,astream = -1;
 poutfile=NULL;
 hr1 = AVIFileOpen(&poutfile,szW,OF_WRITE | OF_CREATE,NULL);
 if(hr1 != AVIERR_OK)poutfile=NULL;
 hr2 = AVIFileOpen(&pinfile,szV,OF_READ,NULL);
 if(hr2 != AVIERR_OK)pinfile=NULL;
 if(hr1 == AVIERR_OK && hr2 == AVIERR_OK){
   for(i=0;i<10;i++){ /* get Video Stream */
     if(AVIFileGetStream(pinfile,&psin_v,0,i) == AVIERR_OK){
       AVIStreamInfo(psin_v,&avis,sizeof(avis));
       if(avis.fccType == streamtypeVIDEO){
         vstream=i;
         AVIStreamFormatSize(psin_v,0,&size);
         if((pformat = (LPSTR)X__Malloc(size)) != NULL){
           if(AVIStreamReadFormat(psin_v,0,pformat,&size) == AVIERR_OK){
             lpBmi=(LPBITMAPINFOHEADER)pformat;
             nsamples=AVIStreamEnd(psin_v);
             hr = AVIFileCreateStream(poutfile,&psout_v,&avis);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Error in output stream",NULL,MB_OK);
             }
             hr = AVIStreamSetFormat(psout_v,0,pformat,size);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Bad output format",NULL,MB_OK);
             }
           }
           X__Free(pformat);
         }
         break;
       }
     }
   }
   for(i=0;i<10;i++){ /* get Audio Stream */
     if(AVIFileGetStream(pinfile,&psin_a,0,i) == AVIERR_OK){
       AVIStreamInfo(psin_a,&avis,sizeof(avis));
       if(avis.fccType == streamtypeAUDIO){
         astream=i;
         AVIStreamFormatSize(psin_a,0,&size);
         if((pformat = (LPSTR)X__Malloc(size)) != NULL){
           if(AVIStreamReadFormat(psin_a,0,pformat,&size) == AVIERR_OK){
             lpWavx=(LPWAVEFORMATEX)pformat;
             nsa=AVIStreamEnd(psin_a);
             hr = AVIFileCreateStream(poutfile,&psout_a,&avis);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Error in output stream",NULL,MB_OK);
             }
             hr = AVIStreamSetFormat(psout_a,0,pformat,size);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Bad output format",NULL,MB_OK);
             }
           }
           X__Free(pformat);
         }
         break;
       }
     }
   }
   SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,nsamples));
   SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
   if(vstream >= 0){
     for(j=0;j<nsamples;j++){
       SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(j+1),0);
       YieldTo();
       hr = AVIStreamRead(psin_v,j,1,NULL,0,&lBytes,&lSamples);
       if((pbuffer = (LPSTR)X__Malloc(lBytes)) != NULL){
         hr = AVIStreamRead(psin_v,j,1,pbuffer,lBytes,NULL,NULL);
         hr = AVIStreamWrite(psout_v,j,1,pbuffer,lBytes,
                             j == 0 ? AVIIF_KEYFRAME : 0, NULL,NULL);
         X__Free(pbuffer);
         TRST=AVIStreamSampleToTime(psin_v,j);
         TRET=AVIStreamSampleToTime(psin_v,j+1);
         if(astream >= 0){
           long lBy;
           TRSS=min(AVIStreamTimeToSample(psin_a,TRST),
                    AVIStreamEnd(psin_a));
           TRES=min(AVIStreamTimeToSample(psin_a,TRET),
                    AVIStreamEnd(psin_a));
           hr = AVIStreamRead(psin_a,TRSS,TRES-TRSS,NULL,0,&lBy,NULL);
           if((sbuffer = (LPSTR)X__Malloc(lBy)) != NULL){
             hr = AVIStreamRead(psin_a,TRSS,TRES-TRSS,sbuffer,lBy,NULL,NULL);
             hr = AVIStreamWrite(psout_a,TRSS,TRES-TRSS,sbuffer,lBy,0,NULL,NULL);
             X__Free(sbuffer);
           }
         }
       }
     }
   }
   if(vstream >= 0){
     AVIStreamRelease(psin_v); AVIStreamRelease(psout_v);
   }
   if(astream >= 0){
     AVIStreamRelease(psin_a); AVIStreamRelease(psout_a);
   }
 }
 if(pinfile != NULL)AVIFileRelease(pinfile);
 if(poutfile != NULL)AVIFileRelease(poutfile);
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
 return TRUE;
}

#if 0
//   if(AVIPutFileOnClipboard(pfile) != 0){
//     MessageBox(NULL,"Bad ClipBoard Put",NULL,MB_OK);
//   }
//   AVIFileRelease(pfile);
//   if(AVIGetFromClipboard(&pfile) != 0){
//     MessageBox(NULL,"Bad ClipBoard Get",NULL,MB_OK);
//   }
//     if(AVIMakeFileFromStreams(&ptempfile,2,ps_video) != 0){
//       MessageBox(NULL,"Failed to make file",NULL,MB_OK);
//     }
//     if(AVIPutFileOnClipboard(ptempfile) != 0){
//       MessageBox(NULL,"Bad ClipBoard,",NULL,MB_OK);
//     }
//     AVIFileRelease(ptempfile); ptempfile=NULL;
static void InsertBlankSoundTrack(void){
 HRESULT hr;
 PAVIFILE pfile;
 PAVISTREAM psin=NULL,ptemp;
 AVISTREAMINFO avis;
 long i,j,n,tVideo,tAudio,s,f,ss;
 hr = AVIFileOpen(&pfile,"\\openfx\\win\\editsnd\\$$$.wav",OF_READ,NULL);
 if(hr != AVIERR_OK)pfile=NULL;
 if(hr == AVIERR_OK){
   for(i=0;i<10;i++){
     if(AVIFileGetStream(pfile,&psin,0,i) == AVIERR_OK){
       AVIStreamInfo(psin,&avis,sizeof(avis));
       if(avis.fccType == streamtypeAUDIO){
MessageBox(NULL,"Blank Soundtrack Audio Stream",NULL,MB_OK);
         tVideo=AVIStreamSampleToTime(ps_video[0],AVIStreamEnd(ps_video[0]));
         tAudio=AVIStreamSampleToTime(psin,AVIStreamEnd(psin));
         ss=s=AVIStreamTimeToSample(psin,tVideo);
         f=AVIStreamEnd(psin);
         if(CreateEditableStream(&ps_video[1],psin) != 0)MessageBeep(MB_OK);
         if(tAudio > tVideo){
           EditStreamCut(ps_video[1],&s,&f,&ptemp);
           AVIStreamRelease(ptemp);
         }
         else if(tAudio < tVideo){
//           n=(long)(tVideo/tAudio);
//           for(j=0;j<n;j++){
//             EditStreamPaste(ps_video[1],&ss,&s,psin,0,s);
//           }
         }
         AVIStreamRelease(psin);
         break;
       }
     }
   }
 }
 if(pfile != NULL)AVIFileRelease(pfile);
}
#endif

static void StripAudioAVI(char *szVideo1, char *szOutFileName){
 LPBITMAPINFOHEADER lpBmi;
 LPWAVEFORMATEX  lpWavx;
 LPSTR pformat,pbuffer,sbuffer;
 AVICOMPRESSOPTIONS opts;
 HRESULT hr,hr1,hr2;
 PAVIFILE pvideofile,poutfile;
 PAVISTREAM psin[10],psout[10];
 PGETFRAME  pget;
 AVISTREAMINFO avis;
 long size,nsamples,key,nstreams,vstream,astream,lBytes,lSamples;
 long TRST,TRET,TRSS,TRES;
 int i,j,k;
 if(szVideo1[0] == '\0')return;
 nstreams=0; vstream = -1,astream = -1;
 poutfile=NULL;
 hr1 = AVIFileOpen(&poutfile,szOutFileName,OF_WRITE | OF_CREATE,NULL);
 if(hr1 != AVIERR_OK)poutfile=NULL;
 hr2 = AVIFileOpen(&pvideofile,szVideo1,OF_READ,NULL);
 if(hr2 != AVIERR_OK)pvideofile=NULL;
 if(hr1 == AVIERR_OK && hr2 == AVIERR_OK){
   for(i=0;i<10;i++){
     psin[i]=psout[i]=NULL;
     if(AVIFileGetStream(pvideofile,&psin[i],0,i) == AVIERR_OK){
       AVIStreamInfo(psin[i],&avis,sizeof(avis));
       if(avis.fccType == streamtypeVIDEO){
         vstream=i;
         AVIStreamFormatSize(psin[i],0,&size);
         if((pformat = (LPSTR)X__Malloc(size)) != NULL){
           if(AVIStreamReadFormat(psin[i],0,pformat,&size) == AVIERR_OK){
             lpBmi=(LPBITMAPINFOHEADER)pformat;
             nsamples=AVIStreamEnd(psin[i]);
             hr = AVIFileCreateStream(poutfile,&psout[i],&avis);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Error in output stream",NULL,MB_OK);
             }
             hr = AVIStreamSetFormat(psout[i],0,pformat,size);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Bad output format",NULL,MB_OK);
             }
           }
           X__Free(pformat);
         }
       }
       nstreams++;
     }
     else {
        break;
     }
   }
   if(nstreams > 0 && vstream >= 0){
     SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,nsamples));
     SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
     for(j=0;j<nsamples;j++){
       SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(j+1),0);
       YieldTo();
       hr = AVIStreamRead(psin[vstream],j,1,NULL,0,&lBytes,&lSamples);
       if((pbuffer = (LPSTR)X__Malloc(lBytes)) != NULL){
         hr = AVIStreamRead(psin[vstream],j,1,pbuffer,lBytes,NULL,NULL);
         hr = AVIStreamWrite(psout[vstream],j,1,pbuffer,lBytes,
                             j == 0 ? AVIIF_KEYFRAME : 0, NULL,NULL);
         X__Free(pbuffer);
         TRST=AVIStreamSampleToTime(psin[vstream],j);
         TRET=AVIStreamSampleToTime(psin[vstream],j+1);
         if(astream >= 0){
           long lBy;
           TRSS=min(AVIStreamTimeToSample(psin[astream],TRST),
                    AVIStreamEnd(psin[astream]));
           TRES=min(AVIStreamTimeToSample(psin[astream],TRET),
                    AVIStreamEnd(psin[astream]));
           hr = AVIStreamRead(psin[astream],TRSS,TRES-TRSS,NULL,0,&lBy,NULL);
           if((sbuffer = (LPSTR)X__Malloc(lBy)) != NULL){
             hr = AVIStreamRead(psin[astream],TRSS,TRES-TRSS,sbuffer,lBy,NULL,NULL);
             hr = AVIStreamWrite(psout[astream],TRSS,TRES-TRSS,sbuffer,lBy,0,NULL,NULL);
             X__Free(sbuffer);
           }
         }
       }
     }
   }
   if(nstreams > 0)for(i=0;i<nstreams;i++){
     if(i == vstream || i == astream){
       AVIStreamRelease(psin[i]); AVIStreamRelease(psout[i]);
     }
   }
 }
 if(pvideofile != NULL)AVIFileRelease(pvideofile);
 if(poutfile != NULL)AVIFileRelease(poutfile);
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
}

static void DubSoundtrack(char *szVideo1, char *szAudio1,
                          char *szOutFileName){
 LPBITMAPINFOHEADER lpBmi;
 LPWAVEFORMATEX  lpWavx;
 LPSTR pformat,pbuffer,sbuffer;
 AVICOMPRESSOPTIONS opts;
 HRESULT hr,hr1,hr2,hr3;
 PAVIFILE pvideofile,paudiofile,poutfile;
 PAVISTREAM psin_v=NULL,psin_a=NULL,psout_v=NULL,psout_a=NULL;
 PGETFRAME  pget;
 AVISTREAMINFO avis;
 long size,nsamples,nsa,key,vstream,astream,lBytes,lSamples;
 long TRST,TRET,TRSS,TRES;
 int i,j,k;
 if(szVideo1[0] == '\0'|| szAudio1[0] == '\0')return;
 vstream = -1,astream = -1;
 poutfile=NULL;
 hr1 = AVIFileOpen(&poutfile,szOutFileName,OF_WRITE | OF_CREATE,NULL);
 if(hr1 != AVIERR_OK)poutfile=NULL;
 hr2 = AVIFileOpen(&pvideofile,szVideo1,OF_READ,NULL);
 if(hr2 != AVIERR_OK)pvideofile=NULL;
 hr3 = AVIFileOpen(&paudiofile,szAudio1,OF_READ,NULL);
 if(hr3 != AVIERR_OK)paudiofile=NULL;
 if(hr1 == AVIERR_OK && hr2 == AVIERR_OK && hr3 == AVIERR_OK){
   for(i=0;i<10;i++){
     if(AVIFileGetStream(pvideofile,&psin_v,0,i) == AVIERR_OK){
       AVIStreamInfo(psin_v,&avis,sizeof(avis));
       if(avis.fccType == streamtypeVIDEO){
         vstream=i;
         AVIStreamFormatSize(psin_v,0,&size);
         if((pformat = (LPSTR)X__Malloc(size)) != NULL){
           if(AVIStreamReadFormat(psin_v,0,pformat,&size) == AVIERR_OK){
             lpBmi=(LPBITMAPINFOHEADER)pformat;
             nsamples=AVIStreamEnd(psin_v);
             hr = AVIFileCreateStream(poutfile,&psout_v,&avis);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Error in output stream",NULL,MB_OK);
             }
             hr = AVIStreamSetFormat(psout_v,0,pformat,size);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Bad output format",NULL,MB_OK);
             }
           }
           X__Free(pformat);
         }
         break;
       }
     }
   }
   for(i=0;i<10;i++){
     if(AVIFileGetStream(paudiofile,&psin_a,0,i) == AVIERR_OK){
       AVIStreamInfo(psin_a,&avis,sizeof(avis));
       if(avis.fccType == streamtypeAUDIO){
         astream=i;
         AVIStreamFormatSize(psin_a,0,&size);
         if((pformat = (LPSTR)X__Malloc(size)) != NULL){
           if(AVIStreamReadFormat(psin_a,0,pformat,&size) == AVIERR_OK){
             lpWavx=(LPWAVEFORMATEX)pformat;
             nsa=AVIStreamEnd(psin_a);
             hr = AVIFileCreateStream(poutfile,&psout_a,&avis);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Error in output stream",NULL,MB_OK);
             }
             hr = AVIStreamSetFormat(psout_a,0,pformat,size);
             if(hr != AVIERR_OK){
               MessageBox(NULL,"Bad output format",NULL,MB_OK);
             }
           }
           X__Free(pformat);
         }
         break;
       }
     }
   }
   SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,nsamples));
   SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
   if(vstream >= 0){
     for(j=0;j<nsamples;j++){
       SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)(j+1),0);
       YieldTo();
       hr = AVIStreamRead(psin_v,j,1,NULL,0,&lBytes,&lSamples);
       if((pbuffer = (LPSTR)X__Malloc(lBytes)) != NULL){
         hr = AVIStreamRead(psin_v,j,1,pbuffer,lBytes,NULL,NULL);
         hr = AVIStreamWrite(psout_v,j,1,pbuffer,lBytes,
                             j == 0 ? AVIIF_KEYFRAME : 0, NULL,NULL);
         X__Free(pbuffer);
         TRST=AVIStreamSampleToTime(psin_v,j);
         TRET=AVIStreamSampleToTime(psin_v,j+1);
         if(astream >= 0){
           long lBy;
           TRSS=min(AVIStreamTimeToSample(psin_a,TRST),
                    AVIStreamEnd(psin_a));
           TRES=min(AVIStreamTimeToSample(psin_a,TRET),
                    AVIStreamEnd(psin_a));
           hr = AVIStreamRead(psin_a,TRSS,TRES-TRSS,NULL,0,&lBy,NULL);
           if((sbuffer = (LPSTR)X__Malloc(lBy)) != NULL){
             hr = AVIStreamRead(psin_a,TRSS,TRES-TRSS,sbuffer,lBy,NULL,NULL);
             hr = AVIStreamWrite(psout_a,TRSS,TRES-TRSS,sbuffer,lBy,0,NULL,NULL);
             X__Free(sbuffer);
           }
         }
       }
     }
   }
   if(vstream >= 0){
     AVIStreamRelease(psin_v); AVIStreamRelease(psout_v);
   }
   if(astream >= 0){
     AVIStreamRelease(psin_a); AVIStreamRelease(psout_a);
   }
 }
 if(pvideofile != NULL)AVIFileRelease(pvideofile);
 if(paudiofile != NULL)AVIFileRelease(paudiofile);
 if(poutfile != NULL)AVIFileRelease(poutfile);
 SendMessage(hWndProgress,PBM_SETPOS,(WPARAM)0,0);
}
