/*  FLI2AVI.c    Convert FLI/FLC to AVI files                               */

/* Note - At present using PAL_COLOURS not RGB  to change to RGB find       */
/* 2 commented lines for RGB     FLI uses OFF screen DDB  FLC uses DIB2DEV  */


#include <windows.h>

#include <stdio.h>
#include <string.h>
#include "fli2avi.h"
#include "..\..\render\flc.h"
#include "..\..\render\fli.h"
#include "..\..\animate\about.h"

#define EVI_Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define EVI_Free(p)      LocalFree((HLOCAL)p)
#define EVI_Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)

#define PALVERSION  0x300
#define MAXPALETTE  256
#define MAXTEXT     256

typedef struct tag_ram_frame{
 char *frame_header;
 char *frame;
}ram_frame;

extern int AVI_frame_rate;
extern void configureAVIfile(HWND, char *);
extern void openAVIfile(char *, LPBITMAPINFOHEADER, int);
extern void appendAVIfile(LPBITMAPINFOHEADER);
extern void closeAVIfile(void);

#define ALIGNULONG(i)   ((i+3)/4*4)

static ram_frame *RamFrames     = NULL;
static HWND      hDlgStatus     = NULL;
static HWND      hWndApp        = NULL;
static HINSTANCE hInst          = NULL;
static HPALETTE  hpalCurrent    = NULL;
static HBITMAP   hbmCurrent     = NULL;
static RGBQUAD   *pRgb          = NULL;
static LPSTR     lpBitmapHeader = NULL;
static LPSTR     lpBitmapBits   = NULL;
static Flc_head  flc_cel;
static Fli_head  fli_cel;
static FILE *FpF=NULL;
static char *szViewClass="OFX:AVIConvertClass";
static unsigned char *Screen,p_red[255],p_green[255],p_blue[255];
static int x_size,y_size,nframes=0,in_ram=0,Nframes=0;
static long nx,ny,nx4,frame2start,frame_rate=15;
static int DisplayWidth,DisplayHeight,DisplayBits,DisplayPlanes,
           xScreen,yScreen,x4_size,current_frame;
static char CurrentDirFile[256],CurrentDirectory[256],CurrentFile[256],
            TitleString[256],AVIfile[256],AVIdir[256];
static UINT  AnimationTimer=0;
static DWORD fdwStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
                        WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_THICKFRAME;
static BOOL AnimationOpen=FALSE,AnimationFLC=FALSE,StartedInAnimator=FALSE;
static BOOL compress=FALSE;

static int ErrMsg(HWND hWnd, char *sz,...);
static void PutMsg(char *s);
LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,LPARAM lParam);
BOOL MenuCommand(HWND hWnd, WORD id);
static void FreeAnimation(void);
static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi);
static WORD NumColours(LPBITMAPINFOHEADER lpbi);
static HBITMAP BitmapFromDib(LPSTR lpHeader,LPSTR lpBits,HPALETTE hpal);
static BOOL ShowFirstFrame(HWND hWnd);
static BOOL ShowNextFlcFrame(HWND hWnd);
static BOOL ShowNextFliFrame(HWND hWnd);
static BOOL FlcBrun(char *);
static BOOL FliBrun(char *);
static BOOL FlcColor(char *, short);
static BOOL FliColor(char *);
static BOOL LoadFirstFlcCel(FILE *f, Flc_head *flc, int info_only);
static BOOL flc_unlccomp(char *);
static BOOL fli_unlccomp(char *);
static BOOL OpenFLC(char *name, int info_only);
static BOOL OpenFLI(char *name, int info_only);

void YieldTo(void){
 MSG msg;
 while (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE)){
   TranslateMessage(&msg);
   DispatchMessage(&msg);
 }
}

void CentreDialogOnScreen(HWND hwnd){
 RECT rcDlg;
 long Xres,Yres;
 Yres=GetSystemMetrics(SM_CYSCREEN);
 Xres=GetSystemMetrics(SM_CXSCREEN);
 GetWindowRect(hwnd,&rcDlg);
 OffsetRect(&rcDlg,-rcDlg.left,-rcDlg.top);
 OffsetRect(&rcDlg,(Xres-rcDlg.right)/2,(Yres-rcDlg.bottom)/2);
 SetWindowPos(hwnd,HWND_TOP,rcDlg.left,rcDlg.top,0,0,SWP_NOSIZE);
 return;
}

static void AppendFileExtension(char *filename, char *ext){
 if(strrchr(filename,'.') == NULL){
   strcat(filename,ext);
 }
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[80],szdummy[256];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=255;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=255;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

int Player(HINSTANCE hInstance){
 int i;
 MSG msg;
 HWND hDesktopWnd,hWnd;
 HDC hDCcaps;
 WNDCLASS wndclass;
 for(i=0;i<255;i++){ p_red[i]=p_green[i]=p_blue[i]=0; }
 hInst       = hInstance;
 hDesktopWnd = GetDesktopWindow();
 hDCcaps     = GetDC(hDesktopWnd);
 DisplayWidth= GetDeviceCaps(hDCcaps,HORZRES);
 DisplayHeight= GetDeviceCaps(hDCcaps,VERTRES);
 DisplayBits= GetDeviceCaps(hDCcaps,BITSPIXEL);
 DisplayPlanes= GetDeviceCaps(hDCcaps,PLANES);
 ReleaseDC(hDesktopWnd,hDCcaps);
 wndclass.style         = 0;
 wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
 wndclass.cbClsExtra    = 0 ;
 wndclass.cbWndExtra    = 0 ;
 wndclass.hInstance     = hInstance ;
 wndclass.hIcon         = LoadIcon(hInst, "PLAYERICON");
 wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
 wndclass.hbrBackground = GetStockObject (WHITE_BRUSH) ;
 wndclass.lpszMenuName  = "ShowMenu" ;
 wndclass.lpszClassName = szViewClass;
 if (!RegisterClass (&wndclass))  return FALSE ;
 xScreen = GetSystemMetrics (SM_CXSCREEN) ;
 yScreen = GetSystemMetrics (SM_CYSCREEN) ;
 hWnd = CreateWindow(
        szViewClass,
        "FLI/FLC to AVI Movie Converter",
        fdwStyle,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        320,
        250,
        NULL,
        NULL,
        hInstance,
        NULL) ;
 if(hWnd == NULL)return FALSE;
 ShowWindow (hWndApp = hWnd, SW_SHOWNA);
 StartedInAnimator=FALSE;
 if(0){
   StartedInAnimator=TRUE;
 }
 GetModuleFileName(hInstance,CurrentDirectory,255);
 *FileInPath(CurrentDirectory) = '\0';
 strcpy(AVIdir,CurrentDirectory);
 strcat(CurrentDirectory,"flics");
 strcat(AVIdir,"flics");
 PostMessage(hWnd,WM_COMMAND,(WPARAM)IDM_OPEN,0);
 while(GetMessage (&msg, NULL, 0, 0)) {
   TranslateMessage (&msg);
   DispatchMessage (&msg);
 }
 FreeAnimation();
 UnregisterClass(wndclass.lpszClassName,wndclass.hInstance);
 return msg.wParam;
}

LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,
                             LPARAM lParam){
 PAINTSTRUCT ps;
 HDC hDC;
 int i,iPos,iMax,iMin,dn;
 RECT rc;
 HPALETTE hpalT;
 POINT Origin;
 HDC hDCbits;
 BITMAP bm;
 switch (iMessage) {
   case WM_SYSCOMMAND:
     switch(LOWORD(wParam & 0xfff0)){
       case SC_CLOSE:
         PostMessage(hWnd,WM_COMMAND,IDM_EXIT,0);
         break;
       default:
         return( DefWindowProc( hWnd, iMessage, wParam, lParam ) );
      }
      break;
  case WM_COMMAND:
    return MenuCommand(hWnd,LOWORD(wParam));
    break;
  case WM_PAINT:
    hDC = BeginPaint(hWnd, &ps);
    if(hpalCurrent != NULL){
      hpalT = SelectPalette(hDC,hpalCurrent,FALSE);
      RealizePalette (hDC);
      if(hbmCurrent != NULL){
        hDCbits = CreateCompatibleDC(hDC);
        GetObject(hbmCurrent,sizeof(BITMAP),(LPSTR)&bm);
        SelectObject(hDCbits,hbmCurrent);
        BitBlt(hDC,0,0,bm.bmWidth,bm.bmHeight,hDCbits,0,0,SRCCOPY);
        DeleteDC(hDCbits);
      }
      else{
        if(lpBitmapHeader != NULL && lpBitmapBits != NULL){
          SetDIBitsToDevice(hDC,0,0,
           (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biWidth,
           (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biHeight,
           0,0,0,
           (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biHeight,
           lpBitmapBits,(LPBITMAPINFO)lpBitmapHeader,DIB_RGB_COLORS);
        }
      }
      SelectPalette(hDC,hpalT,FALSE);
    }
    EndPaint(hWnd, &ps);
    break ;
  default:
      return DefWindowProc (hWnd, iMessage, wParam, lParam) ;
 }
 return 0L ;
}

static BOOL CALLBACK FpsDlgProc(HWND hwnd,UINT msg,
                                WPARAM wparam,LPARAM lparam){
 BOOL err;
 int i;
 switch(msg){
   case WM_INITDIALOG:
     SetDlgItemInt(hwnd,DLG_FPS_RATE,frame_rate,FALSE);
//     CentreDialogOnScreen(hwnd);
     return TRUE;
   case WM_COMMAND:
     switch(LOWORD(wparam)){
       case IDCANCEL:
         EndDialog(hwnd,0);
         return(TRUE);
       case IDOK:
         frame_rate=GetDlgItemInt(hwnd,DLG_FPS_RATE,&err,FALSE);
         EndDialog(hwnd,1);
         return(TRUE);
       default:
         break;
     }
     break;
   default: break;
 }
 return(FALSE);
}

BOOL MenuCommand(HWND hWnd,WORD id){
 int i;
 char lpMessage[MAXTEXT],lpCaption[64];
 switch (id) {
   case IDM_ABOUT:
     MessageBox(hWnd,"     "
                     ABOUT_VERSION
                     "        ",
                     "About the OpenFX FLIc to AVI Converter...",MB_OK);
     break;
   case IDM_COMPRESS:
//     compress ^= 1;
//     if(compress)CheckMenuItem(GetMenu(hWnd),IDM_COMPRESS,MF_CHECKED);
//     else        CheckMenuItem(GetMenu(hWnd),IDM_COMPRESS,MF_UNCHECKED);
     configureAVIfile(hWnd,NULL);
     compress=TRUE;
     break;
   case IDM_FPS:
     DialogBox(hInst,MAKEINTRESOURCE(DLG_FPS),hWnd,(DLGPROC)FpsDlgProc);
     if(frame_rate < 1)frame_rate=1;
     AVI_frame_rate=frame_rate;
     break;
   case IDM_OPEN:
     if(SelectFileName(0,CurrentFile,CurrentDirectory,"File to Convert",
                      "Flic files (*.FLI *.FLC)|*.FL?|",hWnd)){
       SendMessage(hWnd,WM_COMMAND,(WPARAM)IDM_DRAWIT,0);
     }
     break;
   case IDM_DRAWIT:
     ShowFirstFrame(hWnd);
     break;
   case IDM_PLAY:
     if(AnimationOpen && AnimationTimer != 0){
       AnimationTimer=0;
       EnableMenuItem(GetMenu(hWnd),IDM_OPEN,MF_ENABLED);
       ModifyMenu(GetMenu(hWnd),IDM_PLAY,MF_BYCOMMAND|MF_STRING,
                  IDM_PLAY,"&Write");
       SendMessage(hWnd,WM_COMMAND,(WPARAM)IDM_DRAWIT,0);
     }
     else if(AnimationOpen){
       MSG msg;
       static BOOL bBusy=FALSE;
       AnimationTimer=1;
       if(bBusy)break;
       if(!SelectFileName(1,AVIfile,AVIdir,"Output File",
                        "Video Files|*.avi|",hWnd))break;
       AppendFileExtension(AVIfile,".avi");
       EnableMenuItem(GetMenu(hWnd),IDM_OPEN,MF_GRAYED);
       ModifyMenu(GetMenu(hWnd),IDM_PLAY,MF_BYCOMMAND|MF_STRING,
                  IDM_PLAY,"&Stop");
//       if(compress)
//         openAVIfile(AVIfile,(LPBITMAPINFOHEADER)lpBitmapHeader,1);
//       else
//         openAVIfile(AVIfile,(LPBITMAPINFOHEADER)lpBitmapHeader,0);
       bBusy=TRUE;
       if(!compress)configureAVIfile(hWnd,NULL);
       openAVIfile(AVIfile,(LPBITMAPINFOHEADER)lpBitmapHeader,0);
       appendAVIfile((LPBITMAPINFOHEADER)lpBitmapHeader);
       for (i=1;i<Nframes;i++){
         YieldTo();
         sprintf(TitleString,"%s (%ld)",FileInPath(AVIfile),i);
         SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)TitleString);
         if(AnimationFLC)ShowNextFlcFrame(hWnd);
         else            ShowNextFliFrame(hWnd);
         UpdateWindow(hWnd);
         appendAVIfile((LPBITMAPINFOHEADER)lpBitmapHeader);
       }
       closeAVIfile();
       bBusy=FALSE;
       SendMessage(hWnd,WM_COMMAND,(WPARAM)IDM_PLAY,0);
     }
     break;
   case IDM_EXTERNAL:
   case IDM_EXIT:
     if(StartedInAnimator){
       HWND hWndPrev;
       hWndPrev = FindWindow("OFX:AnimatorClass",NULL);
       if(hWndPrev != NULL){
         SetForegroundWindow(hWndPrev);
         BringWindowToTop(hWndPrev);
         ShowWindow (hWndPrev,SW_SHOW);
       }
     }
     PostQuitMessage(0);
     break;
   default:
     break;
 }
 return TRUE;
}

static void FreeAnimation(void){
 if (hpalCurrent) DeleteObject(hpalCurrent);
 if (hbmCurrent)  DeleteObject(hbmCurrent);
 if(lpBitmapHeader != NULL)EVI_Free((char *)lpBitmapHeader);
 if(FpF != NULL)fclose(FpF);
 FpF            = NULL;
 lpBitmapHeader = NULL;
 lpBitmapBits   = NULL;
 hpalCurrent    = NULL;
 hbmCurrent     = NULL;
 pRgb           = NULL;
 AnimationOpen=FALSE;
 AnimationTimer=0;
 if(RamFrames != NULL && nframes > 0){
   int i;
   for(i=0;i<nframes;i++){
     if(RamFrames[i].frame_header != NULL)EVI_Free(RamFrames[i].frame_header);
     if(RamFrames[i].frame != NULL)EVI_Free(RamFrames[i].frame);
   }
   EVI_Free(RamFrames);
   RamFrames=NULL; nframes=0;
 }
 Nframes=0;
}

static int ErrMsg(HWND hWnd,char *sz,...){
 char ach[128];
 va_list arg_ptr;
 va_start(arg_ptr,sz);
 vsprintf(ach,sz,arg_ptr);   /* Format the string */
 va_end(arg_ptr);
 if(AnimationTimer > 0)KillTimer(hWnd,AnimationTimer);
 AnimationTimer=0;
 MessageBox (hWnd, ach, NULL, MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
 return FALSE;
}

static void PutMsg(char *s){
 MessageBox (NULL,s, "AVI Write", MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
}

static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi){
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE       red;
 BYTE       green;
 BYTE       blue;
 int        i;
 RGBQUAD    *pRgb;
 if(lpbi == NULL)return NULL;
 if(lpbi->biSize != sizeof(BITMAPINFOHEADER))return NULL;
 /* Get a pointer to the color table and the number of colors in it */
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 nNumColors = NumColours(lpbi);
 pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) +
         nNumColors * sizeof(PALETTEENTRY));
 if(pPal == NULL)return NULL;
 pPal->palNumEntries = nNumColors;
 pPal->palVersion    = PALVERSION;
 /* Fill in the palette entries from the DIB color table and
  * create a logical color palette. */
 for (i = 0; i < nNumColors; i++){
     pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
     pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
     pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
     pPal->palPalEntry[i].peFlags = (BYTE)0;
 }
 hpal = CreatePalette(pPal);
 LocalFree((HANDLE)pPal);
 return hpal;
}

static WORD NumColours(LPBITMAPINFOHEADER lpbi){
 WORD bits;
 if (lpbi->biClrUsed != 0)return (WORD)lpbi->biClrUsed;
 bits = lpbi->biBitCount;
 switch (bits){
  case 1:    return 2;
  case 4:    return 16;
  case 8:    return 256;
  default:   return 0; /* A 24 bitcount DIB has no color table */
 }
}

static BOOL OpenFLC(char *name, int info_only){
 if((FpF=fopen(name,"rb")) == NULL)return FALSE;
 if(fread(&flc_cel,sizeof(Flc_head),1,FpF) != 1)return FALSE;
 if(flc_cel.type != FLCH_MAGIC)return FALSE;
 nx = x_size = flc_cel.width;
 ny = y_size = flc_cel.height;
 nx4=ALIGNULONG(nx);
 return LoadFirstFlcCel(FpF,&flc_cel,info_only);
}

static BOOL LoadFirstFlcCel(FILE *f,Flc_head *flc, int info_only){
 Flc_frame f_frame;
 Flc_chunk *f_chunk;
 long i,j,s;
 char c,*frame,*buff;
 unsigned char *p;
 fseek(f,flc->oframe1,SEEK_SET);
 if(fread(&f_frame,sizeof(Flc_frame),1,f) != 1 ||
           f_frame.type != FLCF_MAGIC)return FALSE;
 s=f_frame.size-16;
 if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
 if(fread(frame,1,s,FpF) != s)return FALSE;
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Flc_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLC_COLOR256 || f_chunk->type == FLC_COLOR){
     if(!FlcColor(buff,f_chunk->type)){
       EVI_Free(frame);
       return FALSE;
     }
     if(info_only)goto XXIT;
   }
   else if(!info_only && f_chunk->type == FLC_BRUN){
     if(!FlcBrun(buff)){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(!info_only && f_chunk->type == FLC_COPY){
     for (j=0;j<ny;j++){
       p=Screen+(ny-1-j)*nx4;
       memcpy(p,buff+j*nx,nx);
     }
   }
   buff += (f_chunk->size - 6);
 }
 XXIT:
 EVI_Free(frame);
 frame2start=ftell(f);
 return TRUE;
}

static BOOL OpenFLI(char *name, int info_only){
 if((FpF=fopen(name,"rb")) == NULL)return FALSE;
 if(fread(&fli_cel,sizeof(Fli_head),1,FpF) != 1)return FALSE;
 if(fli_cel.type != FLIH_MAGIC){
   return FALSE;
 }
 nx = x_size = fli_cel.width;
 ny = y_size = fli_cel.height;
 nx4=ALIGNULONG(nx);
//ErrMsg(NULL,"%d %d   nx4=%d",nx,ny,nx4);
 return LoadFirstFliCel(FpF,&fli_cel,info_only);
}

static BOOL LoadFirstFliCel(FILE *f,Fli_head *fli, int info_only){
 Fli_frame f_frame;
 Fli_chunk *f_chunk;
 long i,j,s;
 char c,*frame,*buff;
 unsigned char *p;
 if(fread(&f_frame,sizeof(Fli_frame),1,f) != 1){
   return FALSE;
 }
 if(f_frame.type != FLIF_MAGIC){
   return FALSE;
 }
 s=f_frame.size-16;
 if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
 if(fread(frame,1,s,FpF) != s)return FALSE;
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Fli_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLI_COLOR){
     if(!FliColor(buff)){
       EVI_Free(frame);
       return FALSE;
     }
     if(info_only)goto XXIT;
   }
   else if(!info_only && f_chunk->type == FLI_BRUN){
     if(!FliBrun(buff)){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(!info_only && f_chunk->type == FLI_COPY){
     for (j=0;j<ny;j++){
       p=Screen+(ny-1-j)*nx4;
       memcpy(p,buff+j*nx,nx);
     }
   }
   buff += (f_chunk->size - 6);
 }
 XXIT:
 EVI_Free(frame);
 frame2start=ftell(f);
 return TRUE;
}

BOOL FliColor(char *cbuf1){
 short i,j,pk,packets,change;
 unsigned char skip=0,colour;
 packets = *cbuf1++;
 packets += *cbuf1++ *256;
 if (packets<=0) return TRUE;
 for (pk=0; pk<packets; pk++ ){
   skip += *cbuf1++;
   change = *cbuf1++;
   if(change == 0)change=256;
   for(i=0; i<change; i++){
     colour = (unsigned char)*cbuf1++;
     p_red[skip] = (colour << 2);
     colour = (unsigned char)*cbuf1++;
     p_green[skip] = (colour << 2);
     colour = (unsigned char)*cbuf1++;
     p_blue[skip++] = (colour << 2);
   }
 }
 return TRUE;
}

static BOOL FliBrun(char *cbuf1){
 char size;
 unsigned char pk,packet;
 short i,line;
 char colour,*address;
 for (line=0;line<200;line++){
   address=(char *)(Screen+(ny-1-line)*nx4);
   packet=*cbuf1++;
   for (pk=0; pk< packet; pk++){
     size=*cbuf1++;
     if(size > 0){
        colour = *cbuf1++;
        for(i = 0;i<size;i++)*address++ = colour;
     }
     else if(size < 0){
       for ( i = size; i<0; i++ )  *address++ = *cbuf1++;
     }
   }
 }
 return TRUE;
}

static BOOL FlcColor(char *buf, short type){
 short i,j,pk,packets,ci=0;
 unsigned char skip=0,change,colour;
 packets = *((short *)buf); buf += 2;
 for (pk=0; pk<packets; pk++ ) {
   skip = *((unsigned char *)buf); buf++;
   ci += skip;
   change = *((unsigned char *)buf); buf++;
   if (change == 0) change=255;
   for (i=0; i<=change; i++) {
     colour = *((unsigned char *)buf); buf++;
     if(type == FLC_COLOR)colour = colour << 2;
     p_red[ci] = colour;
     colour = *((unsigned char *)buf); buf++;
     if(type == FLC_COLOR)colour = colour << 2;
     p_green[ci] = colour;
     colour = (*(unsigned char *)buf); buf++;
     if(type == FLC_COLOR)colour = colour << 2;
     p_blue[ci] = colour;
     ci++;
   }
 }
 return TRUE;
}

static BOOL FlcBrun(char *buf){
 char size=0;
 unsigned char packet,*p,colour;
 short i,line,pk;
 for (line=0;line<ny;line++){
   p=Screen+(ny-1-line)*nx4;
   packet = *((unsigned char *)buf); buf++;
   for(pk=0; pk<nx; pk+=abs(size)){
     size = *((char *)buf);  buf++;
     if(size > 0){
       colour = *((unsigned char *)buf); buf++;
       for (i=0; i<size; i++) *p++ = colour;
     }
     else{
       for (i = size; i<0; i++){
         colour = *((unsigned char *)buf); buf++;
         *p++ = colour;
       }
     }
   }
 }
 return TRUE;
}

static BOOL flc_unlccomp(char *buff){
 char size;
 unsigned char skip,eol;
 short pk,packet;
 short line,lines,word,words[256],size2;
 unsigned char *p;
 long j,i;
 lines = *((short *)buff); buff+=2;
 line=0;
 for(j = 0; j < lines; j++){
   LOOP:
   packet = *((short *)buff); buff += 2;
   if(packet < 0){
     if((packet & 0x4000) == 0x0400){
       eol = (unsigned char)(packet & 0x00ff);
     }
     line += abs(packet);
     goto LOOP;
   }
   p=Screen+(long)(y_size-1-line)*(long)nx4; //x_size;
   for(pk=0; pk<packet; pk++){
     skip = *((unsigned char *)buff); buff++;
     p += (long)skip;
     size = *((char *)buff); buff++;
     size2=(short)size;
     if(size < 0){
       word = *((short *)buff); buff+=2;
       size = -size;
       for(i=0; i<size; i++){
         *((short *)p)=word; p += 2;
       }
     }
     else{
       size2=(short)size+(short)size;
       memcpy(p,buff,size2); buff += size2;  p += size2;
     }
   }
   line++;
 }
 return TRUE;
}

static BOOL fli_unlccomp(char *cbuf1){
 Fli_lc *flc;
 char *address;
 unsigned char skip,packet;
 short i,line;
 long s1,s2,size,pk;
 flc = (Fli_lc *)cbuf1;
 cbuf1 += 4;
 for( line = flc->skip; line < (short)(flc->skip)+(short)(flc->line); line++){
   packet = (unsigned char)(*cbuf1++);
   address=(char *)(Screen+((long)(199-line))*320L);
   for(pk=0; pk<packet; pk++){
     skip = (unsigned) *cbuf1++;
     address += (unsigned)skip;
     size = (long)(*cbuf1++);
     if(size < 0){
       size = -size;
       memset(address,*cbuf1,size);
       address += size;
       cbuf1++;
     }
     else{
       memcpy(address,cbuf1,size);
       cbuf1 += size; address += size;
     }
   }
 }
 return TRUE;
}

static BOOL ShowFirstFrame(HWND hWnd){
 int i,xc,yc,xs,ys,nf;
 char temp[128];
 long imagesize;
 LPBITMAPINFOHEADER lpbi;
 FreeAnimation();
 strcpy(temp,CurrentFile);
 if(strstr(strupr(temp),".FLI") != NULL)AnimationFLC=FALSE;
 else AnimationFLC=TRUE;
 if(AnimationFLC){
   if(!OpenFLC(temp,1)){
     ErrMsg(hWnd,"'%s' is NOT a valid FLC file", (LPSTR)temp);
     if(FpF != NULL)fclose(FpF); FpF=NULL;
     return FALSE;
   }
 }
 else{
   if(!OpenFLI(temp,1)){
     ErrMsg(hWnd,"'%ls' is NOT a valid FLI file", (LPSTR)temp);
     if(FpF != NULL)fclose(FpF); FpF=NULL;
     return FALSE;
   }
 }
 if(FpF != NULL)fclose(FpF); FpF=NULL;
 x4_size = ALIGNULONG(x_size);
 imagesize = (DWORD)x4_size*(DWORD)y_size;
 lpBitmapHeader = (LPSTR)EVI_Malloc((long)sizeof(BITMAPINFOHEADER)
                                +256L*sizeof(RGBQUAD)+imagesize);
 if(lpBitmapHeader == NULL){
   ErrMsg(hWnd,"Out of memory");
   return FALSE;
 }
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x4_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=8;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage=(DWORD)imagesize;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=256;
 lpbi->biClrImportant=256;
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<256;i++){
   pRgb[i].rgbRed = p_red[i];
   pRgb[i].rgbGreen = p_green[i];
   pRgb[i].rgbBlue = p_blue[i];
 }
 hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits=(LPSTR)(lpbi);
 lpBitmapBits += (lpbi->biSize+256*sizeof(RGBQUAD));
 if(lpBitmapBits == NULL){
   ErrMsg(hWnd,"Out of memory");
   FreeAnimation();
   return FALSE;
 }
 memset(lpBitmapBits,0,imagesize);
 Screen=(unsigned char *)lpBitmapBits;
 if(AnimationFLC){
   if(!OpenFLC(temp,0)){
     PutMsg("Error reading FLC file");
     FreeAnimation();
     return FALSE;
   }
 }
 else{
   if(!OpenFLI(temp,0)){
     PutMsg("Error reading FLI file");
     FreeAnimation();
     return FALSE;
   }
 }
 hbmCurrent=NULL;
 current_frame=0;
 if(AnimationFLC)nf=flc_cel.frames;
 else            nf=fli_cel.frame_count;
 Nframes=nf;
 sprintf(TitleString,"%s (%u Frames)",FileInPath(CurrentFile),nf);
 SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)TitleString);
 xs=x_size+2*GetSystemMetrics(SM_CXFRAME);
 ys=y_size+GetSystemMetrics(SM_CYFRAME)+
           GetSystemMetrics(SM_CYCAPTION)+
           GetSystemMetrics(SM_CYMENU);
 xc=(xScreen-xs)/2;
 yc=(yScreen-ys)/2;
 if(xs > xScreen){xs=xScreen; xc=0;}
 if(ys > yScreen){ys=yScreen; yc=0;}
 SetWindowPos(hWnd,NULL,xc,yc,xs,ys,SWP_NOMOVE);
 InvalidateRect(hWnd,NULL,TRUE);
 AnimationOpen=TRUE;
 EnableMenuItem(GetMenu(hWnd),IDM_PLAY,MF_ENABLED);
 return TRUE;
}

static BOOL ShowNextFlcFrame(HWND hWnd){
 Flc_frame f_frame;
 Flc_chunk *f_chunk;
 char c,*frame,*buff;
 int i,j;
 long s;
 current_frame++;
 if(current_frame > flc_cel.frames){
   if(FpF != NULL){
//     fseek(FpF,flc_cel.oframe2,SEEK_SET); /* this should be OK      */
     fseek(FpF,frame2start,SEEK_SET);   /* but just in case user ours */
   }
   current_frame=1;
 }
 if(FpF != NULL){
   if(fread(&f_frame,sizeof(Flc_frame),1,FpF) != 1){
     ErrMsg(hWnd,"Not an FLC frame 1");
     return FALSE;
   }
   if(f_frame.type != FLCF_MAGIC){
     ErrMsg(hWnd,"Not an FLC frame 2");
     return FALSE;
   }
   s=f_frame.size-16;
   if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
   if(fread(frame,1,s,FpF) != s)return FALSE;
 }
 else if(RamFrames != NULL){
   memcpy(&f_frame,RamFrames[current_frame-1].frame_header,sizeof(Flc_frame));
   frame=RamFrames[current_frame-1].frame;
 }
 else{
   MessageBox(NULL,"DISPLAY ERROR",NULL,MB_OK);
   PostQuitMessage(0);
   return FALSE;
 }
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Flc_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLC_SS2){   /* found cell */
     if(flc_unlccomp(buff) < 0){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(f_chunk->type == FLC_LC){
     if(fli_unlccomp(buff) < 0){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(f_chunk->type == FLC_COPY){
     unsigned char *p;
     for (j=0;j<ny;j++){
       p=Screen+(ny-1-j)*nx4;
       memcpy(p,buff+j*nx,nx);
     }
   }
   buff += (f_chunk->size - 6);
 }
 if(FpF != NULL)EVI_Free(frame);
 if (hbmCurrent)DeleteObject(hbmCurrent); hbmCurrent=NULL;
 InvalidateRect(hWnd,NULL,FALSE);
 return TRUE;
}

static BOOL ShowNextFliFrame(HWND hWnd){
 Fli_frame f_frame;
 Fli_chunk *f_chunk;
 char c,*frame,*buff;
 int i,j;
 long s;
 current_frame++;
 if(current_frame > fli_cel.frame_count){
   if(FpF != NULL){
     fseek(FpF,frame2start,SEEK_SET);   /* but just in case user ours */
   }
   current_frame=1;
 }
 if(FpF != NULL){
   if(fread(&f_frame,sizeof(Fli_frame),1,FpF) != 1){
     ErrMsg(hWnd,"Not an FLI frame 1");
     return FALSE;
   }
   if(f_frame.type != FLIF_MAGIC){
     ErrMsg(hWnd,"Not an FLI frame 2");
     return FALSE;
   }
   s=f_frame.size-16;
   if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
   if(fread(frame,1,s,FpF) != s)return FALSE;
 }
 else if(RamFrames != NULL){
   memcpy(&f_frame,RamFrames[current_frame-1].frame_header,sizeof(Fli_frame));
   frame=RamFrames[current_frame-1].frame;
 }
 else{
   MessageBox(NULL,"DISPLAY ERROR",NULL,MB_OK);
   PostQuitMessage(0);
   return FALSE;
 }
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Fli_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLI_LC){
     if(fli_unlccomp(buff) < 0){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(f_chunk->type == FLI_COPY){
     unsigned char *p;
     for (j=0;j<ny;j++){
       p=Screen+(ny-1-j)*nx4;
       memcpy(p,buff+j*nx,nx);
     }
   }
   buff += (f_chunk->size - 6);
 }
 if(FpF != NULL)EVI_Free(frame);
 if (hbmCurrent)DeleteObject(hbmCurrent); hbmCurrent=NULL;
 InvalidateRect(hWnd,NULL,FALSE);
 return TRUE;
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 HWND hWndPrev;
 int id;
 hWndPrev = FindWindow(szViewClass,NULL);
 if (hWndPrev != NULL){
   SetForegroundWindow(hWndPrev);
   BringWindowToTop(hWndPrev);
   ShowWindow (hWndPrev,SW_SHOW);
   return 32;
 }
 if(hPrevInstance != NULL)return FALSE;
 id=Player(hInstance);
 return id;
}
