/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file GIFHASH.C   */

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <windows.h>

#define EVI_Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define EVI_Free(p)      LocalFree((HLOCAL)p)
#define EVI_Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)

#undef OK

#include "giflib.h"
#include "gifhash.h"

#ifdef _SUNSTYLE
static long KeyItem();
#else
static long KeyItem(unsigned long Item);
#endif

/******************************************************************************
* Initialize HashTable - allocate the memory needed and clear it.             *
******************************************************************************/
#ifdef _SUNSTYLE
GifHashTableType *_InitHashTable(){
#else
GifHashTableType *_InitHashTable(void){
#endif
    GifHashTableType  *HashTable;
    if ((HashTable = (GifHashTableType *) EVI_Malloc(sizeof(GifHashTableType)))
                  == NULL) return NULL;
    _ClearHashTable(HashTable);
    return HashTable;
}

/******************************************************************************
* Routine to clear the HashTable to an empty state.                           *
* This part is a little machine depended. Use the commented part otherwise.   *
******************************************************************************/
#ifdef _SUNSTYLE
void _ClearHashTable(HashTable) GifHashTableType *HashTable; {
#else
void _ClearHashTable(GifHashTableType *HashTable){
#endif    
    memset(HashTable -> HTable, 0xFF, HT_SIZE * sizeof(long));
}

/******************************************************************************
* Routine to insert a new Item into the HashTable. The data is assumed to be  *
* new one.                                                                    *
******************************************************************************/
#ifdef _SUNSTYLE
void _InsertHashTable(HashTable,Key,Code) GifHashTableType *HashTable;
unsigned long Key; long Code; {
#else
void _InsertHashTable(GifHashTableType *HashTable,
        unsigned long Key, long Code){
#endif
    long HKey = KeyItem(Key);
    unsigned long *HTable = HashTable -> HTable;


    while (HT_GET_KEY(HTable[HKey]) != 0xFFFFFL) {
      HKey = (HKey + 1) & HT_KEY_MASK;
    }
    HTable[HKey] = HT_PUT_KEY(Key) | HT_PUT_CODE(Code);
}

/******************************************************************************
* Routine to test if given Key exists in HashTable and if so returns its code *
* Returns the Code if key was found, -1 if not.                               *
******************************************************************************/
#ifdef _SUNSTYLE
long _ExistsHashTable(HashTable,Key) GifHashTableType *HashTable;
unsigned long Key; {
#else
long _ExistsHashTable(GifHashTableType *HashTable, unsigned long Key){
#endif
    long HKey = KeyItem(Key);
    unsigned long *HTable = HashTable -> HTable, HTKey;


    while ((HTKey = HT_GET_KEY(HTable[HKey])) != 0xFFFFFL) {
if (Key == HTKey) return (long)(HT_GET_CODE(HTable[HKey]));
HKey = (HKey + 1) & HT_KEY_MASK;
    }

    return -1;
}

/******************************************************************************
* Routine to generate an HKey for the hashtable out of the given unique key.  *
* The given Key is assumed to be 20 bits as follows: lower 8 bits are the     *
* new postfix character, while the upper 12 bits are the prefix code.         *
* Because the average hit ratio is only 2 (2 hash references per entry),      *
* evaluating more complex keys (such as twin prime keys) does not worth it!   *
******************************************************************************/
#ifdef _SUNSTYLE
static long KeyItem(Item) unsigned long Item; {
#else
static long KeyItem(unsigned long Item){
#endif
  return (long)(((Item >> 12) ^ Item) & HT_KEY_MASK);
}

