/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>

#include "..\..\jpeg\jpeglib.h"

extern int x_size,y_size;

extern void jpg_display_line(unsigned char *pbgr,long line);

struct my_error_mgr {
  struct jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
};

typedef struct my_error_mgr * my_error_ptr;

METHODDEF(void)my_error_exit (j_common_ptr cinfo){
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  longjmp(myerr->setjmp_buffer, 1);
}

long ReadJPG(char *filename, short info_only){
  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr jerr;
  FILE * infile;
  JSAMPARRAY buffer;
  int row_stride;
  if ((infile = fopen(filename, "rb")) == NULL) {
    return 0;
  }
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  if (setjmp(jerr.setjmp_buffer)) {
    jpeg_destroy_decompress(&cinfo);
    fclose(infile);
    return 0;
  }
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, infile);
  (void) jpeg_read_header(&cinfo, TRUE);
  if(info_only){
    x_size=cinfo.image_width;
    y_size=cinfo.image_height;
    jpeg_destroy_decompress(&cinfo);
    fclose(infile);
    return 1;
  }
  (void) jpeg_start_decompress(&cinfo);
  row_stride = cinfo.output_width * cinfo.output_components;
  buffer = (*cinfo.mem->alloc_sarray)
           ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
  while (cinfo.output_scanline < cinfo.output_height) {
    (void) jpeg_read_scanlines(&cinfo, buffer, 1);
    jpg_display_line((unsigned char *)buffer[0],cinfo.output_scanline-1);
  }
  (void)jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  fclose(infile);
  return 1;
}


typedef struct SCREENBUFFER {
 unsigned char A,  /* Alpha channel (not available yet) */
               R,  /* red value  range - 255            */
               G,  /* green value  range - 255          */
               B;  /* blue value  range - 255           */
} fullscreenbuffer;


long WriteJPG(char *filename, long x, long y, long quality,
             fullscreenbuffer *b){
  int i,j; //quality=75;
  struct jpeg_compress_struct cinfo;
  struct my_error_mgr jerr;
  unsigned char *row,*p;
  FILE * outfile;
  JSAMPROW row_pointer[1];
  if((row = (unsigned char *)malloc(x*3*sizeof(unsigned char))) == NULL)
    return 0;
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  if (setjmp(jerr.setjmp_buffer)) {
    jpeg_destroy_compress(&cinfo);
    fclose(outfile);
    free(row);
    return 0;
  }
  if ((outfile = fopen(filename, "wb")) == NULL)return 0;
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo, outfile);
  cinfo.image_width = x;
  cinfo.image_height = y;
  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, quality, TRUE);
  jpeg_start_compress(&cinfo, TRUE);
  row_pointer[0] = (JSAMPROW)(&row[0]);
  i=0;
  while (cinfo.next_scanline < cinfo.image_height) {
    for(j=0,p=row;j<x;j++){
      *p++ = b->R;
      *p++ = b->G;
      *p++ = b->B;
      b++;
    }
    (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
    i++;
  }
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);
  fclose(outfile);
  free(row);
  return 1;
}
