/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

#include <stdio.h>
#include <windows.h>
#include <windowsx.h>

#include "mdiview.h"
#include "..\..\animate\about.h"

typedef LPBITMAPINFOHEADER PDIB;

#define MDIVIEW_APP_CLASS "OFX:Viewer"
#define MDIVIEW_DOC_CLASS "OFX:Viewer:Window"

#define CBWNDEXTRA            (12)
#define GWL_BITMAP            (0)
#define GWL_BITMAP_ICON       (4)
#define GWL_PALETTE           (8)

#define FI_GIF 0
#define FI_TGA 1
#define FI_CEL 2
#define FI_BMP 3
#define FI_DIB 4
#define FI_JPG 5
#define FI_TIF 6
#define FI_PNG 7

#define ISCHECKED(hwnd,id) (BOOL)(GetMenuState(GetMenu(hwnd), id, 0) & MF_CHECKED)
#define ALIGNULONG(i)   (((i+3)/4)*4)
#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)

HINSTANCE ghInstApp;    /* Instance handle */
HACCEL ghAccelApp;
HWND ghwndApp;          /* Handle to parent window */
HWND ghwndMdi;          /* Handle to MCI client window */
HWND ghwndAct=NULL;
char szFileName[256],szDirName[256];
POINT ptSize;
BOOL  dither;


extern HPALETTE CreateMyHalftonePalette(void);
extern PDIB DibHalftoneDIB(PDIB pSource, HPALETTE hpalApp);

LRESULT CALLBACK AppWndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK DocWndProc(HWND, UINT, WPARAM, LPARAM);

static BOOL AppInit(HINSTANCE, HINSTANCE, LPSTR, int);
static HWND mdiCreateDoc(LPSTR, LPSTR, LPARAM);
static HWND mdiCreateClient(HWND, HMENU);
static LRESULT mdiSendMessage(HWND, HWND, UINT, WPARAM, LPARAM);
static void mdiFilterSendMessage(HWND, UINT, WPARAM, LPARAM);
static void MCI_OnCommand(HWND, int, HWND, UINT);
static void MCI_OnInitMenuPopup(HWND, HMENU, int, BOOL);
static void MCI_OnSize(HWND, UINT, int, int);
static int SelectFileName(int, char *, char *, char *, char *, HWND);
static PSTR FileInPath(PSTR);
static int identify_file_format(char *);
static char *short_form(char *);
static HBITMAP LoadBmp(char *, HPALETTE *, HBITMAP *);
static HBITMAP LoadDib(char *, HPALETTE *, HBITMAP *);
static HBITMAP LoadGif(char *, HPALETTE *, HBITMAP *);
static HBITMAP LoadTga(char *, HPALETTE *, HBITMAP *);
static HBITMAP LoadJpg(char *, HPALETTE *, HBITMAP *);
static HBITMAP LoadTif(char *, HPALETTE *, HBITMAP *);
static HBITMAP LoadPng(char *, HPALETTE *, HBITMAP *);
static void AppPaint(HWND, HDC ,int, int);
static void SizeWindow(HWND);
static void SetScrollRanges(HWND);
static void GetRealClientRect(HWND, PRECT);

static BOOL AppInit(HINSTANCE hInst,HINSTANCE hPrev,LPSTR szCmd,int sw){
 char gszHomeDir[256];
 WNDCLASS cls;
 WORD wVer;
 HWND hDesktopWnd;
 HDC hDCcaps;
 int DisplayBits;
 ghInstApp = hInst;
 ghAccelApp = LoadAccelerators(ghInstApp, "AppAccel");
 if (!hPrev) {
     cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
     cls.hIcon          = LoadIcon(hInst,"AppIcon");
     cls.lpszMenuName   = "AppMenu";
     cls.lpszClassName  = MDIVIEW_APP_CLASS;
     cls.hbrBackground  = (HBRUSH)(COLOR_APPWORKSPACE+1);
     cls.hInstance      = hInst;
     cls.style          = 0;
     cls.lpfnWndProc    = (WNDPROC)AppWndProc;
     cls.cbClsExtra     = 0;
     cls.cbWndExtra     = 0;
     if (!RegisterClass(&cls))
         return FALSE;
     cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
     cls.hIcon          = NULL; //LoadIcon(hInst,"AppIcon");
     cls.lpszMenuName   = NULL;
     cls.lpszClassName  = MDIVIEW_DOC_CLASS;
     cls.hbrBackground  = (HBRUSH)(COLOR_APPWORKSPACE+1);
     cls.hInstance      = hInst;
     cls.style          = 0;
     cls.lpfnWndProc    = (WNDPROC)DocWndProc;
     cls.cbClsExtra     = 0;
     cls.cbWndExtra     = CBWNDEXTRA;
     if (!RegisterClass(&cls))
         return FALSE;
 }
 ghwndApp = CreateWindow(
     MDIVIEW_APP_CLASS,
     "MDI Image File Viewer",
     WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
     0,
     0,
     CW_USEDEFAULT,
     CW_USEDEFAULT,
     (HWND)NULL,        /* no parent */
     (HMENU)NULL,       /* use class menu */
     (HANDLE)hInst,     /* handle to window instance */
     (LPSTR)NULL);      /* no params to pass on */
 ShowWindow(ghwndApp,sw);
 GetModuleFileName(hInst,gszHomeDir,256);
 *FileInPath(gszHomeDir) = '\0'; szFileName[0]='\0';
 strcpy(szDirName,gszHomeDir);// strcat(szDirName,"render");
 hDesktopWnd = GetDesktopWindow();
 hDCcaps     = GetDC(hDesktopWnd);
 DisplayBits=GetDeviceCaps(hDCcaps,BITSPIXEL);
 ReleaseDC(hDesktopWnd,hDCcaps);
 if(DisplayBits <= 8)dither=TRUE; else dither=FALSE;
 if(dither)CheckMenuItem(GetMenu(ghwndApp),
                 MENU_DITHER,MF_CHECKED);
 else      CheckMenuItem(GetMenu(ghwndApp),
                 MENU_DITHER,MF_UNCHECKED);
 return TRUE;
}

int PASCAL WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,
                   LPSTR szCmdLine,int sw){
 MSG msg;
 if (!AppInit(hInstance,hPrevInstance,szCmdLine,sw))return FALSE;
 for (;;) {
    if (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE)) {
      if (msg.message == WM_QUIT)break;
      if (ghAccelApp && ghwndApp && TranslateAccelerator(ghwndApp, ghAccelApp, &msg))
                continue;
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
    else {
      WaitMessage();
    }
 }
 return msg.wParam;
}

static HWND mdiCreateDoc(LPSTR szClass,LPSTR szTitle,LPARAM l){
 static int x0=0,y0=0;
 MDICREATESTRUCT mdics;
 mdics.szClass   = szClass;
 mdics.szTitle   = szTitle;
 mdics.hOwner    = ghInstApp;
 mdics.x         = x0; x0 += 8;
 mdics.y         = y0; y0 += 8;
 mdics.cx        = 320;
 mdics.cy        = 240;
 mdics.style     = WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_HSCROLL | WS_VSCROLL;
 mdics.lParam    = l;
 return FORWARD_WM_MDICREATE(ghwndMdi,(LPMDICREATESTRUCT)&mdics,SendMessage);
}

static HWND mdiCreateClient(HWND hwndP,HMENU hmenuWindow){
 CLIENTCREATESTRUCT ccs;
 ccs.hWindowMenu = hmenuWindow;
 ccs.idFirstChild = 100;
 return CreateWindow("MDICLIENT",
            NULL,
            WS_CHILD | WS_CLIPCHILDREN | WS_VISIBLE,
            0,
            0,
            0,
            0,
            hwndP,
            0,
            ghInstApp,
            (LPVOID)&ccs);
}

static LRESULT mdiSendMessage(HWND hwndMdi,HWND hwnd,UINT msg,
                              WPARAM wParam,LPARAM lParam){
 if (hwnd == (HWND)-1) {
   for(hwnd = GetWindow(hwndMdi,GW_CHILD);
       hwnd;
       hwnd=GetWindow(hwnd,GW_HWNDNEXT))
     SendMessage(hwnd, msg, wParam, lParam);
   return 0L;
 }
 else {
   if (hwnd == NULL)hwnd = FORWARD_WM_MDIGETACTIVE(hwndMdi, SendMessage);
   if (hwnd)return SendMessage(hwnd, msg, wParam, lParam);
 }
 return 0L;
}

static void MCI_OnCommand(HWND hwnd,int id,HWND hwndCtl,UINT codeNotify){
 HWND hwndNew;
 HWND hwndMovie = FORWARD_WM_MDIGETACTIVE(ghwndMdi,SendMessage);
 switch(id) {
   case MENU_EXIT:
     PostMessage(hwnd,WM_CLOSE,0,0L);
     break;
   case MENU_CLOSE:
     PostMessage(hwndMovie, WM_CLOSE, 0, 0L);
     break;
   case MENU_ABOUT:
     MessageBox(hwnd,"     "
                     ABOUT_VERSION
                     "        ",
                     "About OpenFX Image Viewer...",MB_OK);
     break;
   case MENU_DITHER:
     dither ^= 1;
     if(dither)CheckMenuItem(GetMenu(hwnd),MENU_DITHER,MF_CHECKED);
     else      CheckMenuItem(GetMenu(hwnd),MENU_DITHER,MF_UNCHECKED);
     break;
   case MENU_OPEN:{
       HBITMAP  hBitmap=NULL,hBitmapIcon=NULL;
       HPALETTE hPalette=NULL;
       if(SelectFileName(0,szFileName,szDirName,"Viewer",
               "All file formats|*.bmp;*.gif;*.tga;*.dib;*.jpg;*.tif;*.png|"
               "Windows Bitmaps|*.bmp|"
               "Device Independent Bitmaps|*.dib|"
               "Compuserve GIFs|*.gif|"
               "Targa TGAs|*.tga|"
               "JPEG|*.jpg|"
               "TIFF|*.tif|"
               "PNG|*.png|"
                        ,hwnd) == TRUE){
         if (hwndNew = mdiCreateDoc(MDIVIEW_DOC_CLASS,
                       (LPSTR)short_form(szFileName),(LPARAM)0)){
           int fid;
           fid=identify_file_format(szFileName);
           ghwndAct=hwndNew;
           if     (fid == FI_BMP)hBitmap=LoadBmp(szFileName,&hPalette,&hBitmapIcon);
           else if(fid == FI_GIF)hBitmap=LoadGif(szFileName,&hPalette,&hBitmapIcon);
           else if(fid == FI_TGA)hBitmap=LoadTga(szFileName,&hPalette,&hBitmapIcon);
           else if(fid == FI_DIB)hBitmap=LoadDib(szFileName,&hPalette,&hBitmapIcon);
           else if(fid == FI_JPG)hBitmap=LoadJpg(szFileName,&hPalette,&hBitmapIcon);
           else if(fid == FI_TIF)hBitmap=LoadTif(szFileName,&hPalette,&hBitmapIcon);
           else if(fid == FI_PNG)hBitmap=LoadPng(szFileName,&hPalette,&hBitmapIcon);
           SetWindowLong(hwndNew,GWL_BITMAP,(LONG)hBitmap);
           SetWindowLong(hwndNew,GWL_BITMAP_ICON,(LONG)hBitmapIcon);
           SetWindowLong(hwndNew,GWL_PALETTE,(LONG)hPalette);
           SetFocus(hwndNew);
           SizeWindow(hwndNew);
         }
       }
     }
     break;
   case WM_MDITILE:
   case WM_MDICASCADE:
   case WM_MDIICONARRANGE:
     SendMessage(ghwndMdi, id, 0, 0);
     break;
   default:
     FORWARD_WM_COMMAND(NULL,id,hwndCtl,codeNotify,mdiFilterSendMessage);
     break;
 }
 return;
}

static void mdiFilterSendMessage(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam){
 mdiSendMessage(ghwndMdi, hwnd, msg, wParam, lParam);
}

static void MCI_OnInitMenuPopup(HWND hwnd,HMENU hMenu,
                                int item,BOOL fSystemMenu){
    HMENU hStylesMenu = GetSubMenu(GetSubMenu(GetMenu(hwnd), 1), 10);
    HMENU hFileMenu = GetSubMenu(GetMenu(hwnd), 0);
    HMENU hMovieMenu = GetSubMenu(GetMenu(hwnd), 1);
    HMENU hWindowMenu = GetSubMenu(GetMenu(hwnd), 2);
    HWND  hwndMovie = FORWARD_WM_MDIGETACTIVE(ghwndMdi, SendMessage);
    UINT f;
    if (fSystemMenu)return;
    else if (hMenu == hMovieMenu) {
    }
    else if (hMenu == hWindowMenu) {
    }
    else if (hMenu == hFileMenu) {
    }
    FORWARD_WM_INITMENUPOPUP(NULL, hMenu, item,
                             fSystemMenu, mdiFilterSendMessage);
    return;
}

static void MCI_OnSize(HWND hwnd,UINT state,int cx,int cy){
  MoveWindow(ghwndMdi,0,0,cx,cy,TRUE);
  return;
}

LRESULT CALLBACK AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam){
 static BOOL busy1=FALSE;
 switch (msg) {
   case WM_CREATE:
     ghwndMdi = mdiCreateClient(hwnd, GetSubMenu(GetMenu(hwnd), GetMenuItemCount(GetMenu(hwnd))-1));
     break;
   case WM_COMMAND:
     HANDLE_WM_COMMAND(hwnd, wParam, lParam, MCI_OnCommand);
     break;
   case WM_PALETTECHANGED:
     mdiSendMessage(ghwndMdi, (HWND)-1, msg, wParam, lParam);
     break;
   case WM_QUERYNEWPALETTE:
     return mdiSendMessage(ghwndMdi, NULL, msg, wParam, lParam);
     break;
   case WM_INITMENUPOPUP:
     HANDLE_WM_INITMENUPOPUP(hwnd, wParam, lParam, MCI_OnInitMenuPopup);
     break;
   case WM_SIZE:
     HANDLE_WM_SIZE(hwnd, wParam, lParam, MCI_OnSize);
     break;
   case WM_DESTROY:
     PostQuitMessage(0);
     return (0);
 }
 return DefFrameProc(hwnd,ghwndMdi,msg,wParam,lParam);
}

LRESULT CALLBACK DocWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam){
 int i,iPos,iMax,iMin,dn;
 RECT rc;
 HDC hDC;
 switch (msg) {
   case  WM_PAINT:{
       PAINTSTRUCT ps;
       hDC=BeginPaint(hwnd, &ps);
       AppPaint(hwnd,hDC,GetScrollPos(hwnd,SB_HORZ),GetScrollPos(hwnd,SB_VERT));
       EndPaint(hwnd,&ps);
     }
     break;
   case WM_VSCROLL:
     GetScrollRange (hwnd,SB_VERT, &iMin, &iMax);
     iPos = GetScrollPos (hwnd, SB_VERT);
     GetClientRect (hwnd, &rc);
     switch (LOWORD(wParam)) {
       case SB_LINEDOWN:      dn =  rc.bottom / 16 + 1;    break;
       case SB_LINEUP:        dn = -rc.bottom / 16 + 1;    break;
       case SB_PAGEDOWN:      dn =  rc.bottom / 2  + 1;    break;
       case SB_PAGEUP:        dn = -rc.bottom / 2  + 1;    break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION: dn = HIWORD(wParam)-iPos;    break;
        default:               dn = 0;
     }
     if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       ScrollWindow(hwnd, 0, -dn, NULL, NULL);
       SetScrollPos(hwnd, SB_VERT, iPos + dn, TRUE);
     }
     break;
   case WM_HSCROLL:
     GetScrollRange(hwnd, SB_HORZ, &iMin, &iMax);
     iPos = GetScrollPos(hwnd, SB_HORZ);
     GetClientRect (hwnd, &rc);
      switch (LOWORD(wParam)) {
       case SB_LINEDOWN:      dn =  rc.right / 16 + 1;     break;
       case SB_LINEUP:        dn = -rc.right / 16 + 1;     break;
       case SB_PAGEDOWN:      dn =  rc.right / 2  + 1;     break;
       case SB_PAGEUP:        dn = -rc.right / 2  + 1;     break;
       case SB_THUMBTRACK:
       case SB_THUMBPOSITION: dn = HIWORD(wParam) - iPos; break;
      default:               dn = 0;
     }
     if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
       ScrollWindow (hwnd, -dn, 0, NULL, NULL);
       SetScrollPos (hwnd, SB_HORZ, iPos + dn, TRUE);
     }
     break;
   case WM_SIZE:
     SetScrollRanges(hwnd);
     return DefMDIChildProc(hwnd,msg,wParam,lParam);
   case WM_DESTROY:{
       HBITMAP  hBitmap,hBitmapIcon;
       HPALETTE hPalette;
       hBitmap=(HBITMAP)GetWindowLong(hwnd,GWL_BITMAP);
       hBitmapIcon=(HBITMAP)GetWindowLong(hwnd,GWL_BITMAP_ICON);
       hPalette=(HPALETTE)GetWindowLong(hwnd,GWL_PALETTE);
       if(hBitmap != NULL)DeleteObject(hBitmap); hBitmap=NULL;
       if(hBitmapIcon != NULL)DeleteObject(hBitmapIcon); hBitmapIcon=NULL;
       if(hPalette != NULL)DeleteObject(hPalette); hPalette=NULL;
       SetWindowLong(hwnd,GWL_BITMAP,(LONG)hBitmap);
       SetWindowLong(hwnd,GWL_BITMAP_ICON,(LONG)hBitmapIcon);
       SetWindowLong(hwnd,GWL_PALETTE,(LONG)hPalette);
     }
     break;
   case WM_MDIACTIVATE:
     ghwndAct=(HWND)lParam;
     if(ghwndAct == hwnd)InvalidateRect(hwnd,(LPRECT)(NULL),1);
     break;
#if 0
   case WM_PALETTECHANGED:{
      if((HWND)wParam != hwnd){
        HDC hDC;
        HPALETTE hpalT,hpal;
        hpal=(HPALETTE)GetWindowLong(hwnd,GWL_PALETTE);
        if(hpal == NULL)break;
        hDC=GetDC(hwnd);
        hpalT = SelectPalette(hDC,hpal,FALSE);
        if(RealizePalette(hDC) > 0)InvalidateRect(hwnd,(LPRECT)(NULL),1);
        SelectPalette(hDC,hpalT,FALSE);
        ReleaseDC(hwnd,hDC);
      }
     }
     break;
#endif
   default:
     return DefMDIChildProc(hwnd,msg,wParam,lParam);
 }
 return 0;
}

static void GetRealClientRect(HWND hwnd, PRECT lprc){
 DWORD dwStyle;
 dwStyle = GetWindowLong (hwnd, GWL_STYLE);
 GetClientRect (hwnd,lprc);
 if (dwStyle & WS_HSCROLL)lprc->bottom += GetSystemMetrics(SM_CYHSCROLL);
 if (dwStyle & WS_VSCROLL)lprc->right  += GetSystemMetrics(SM_CXVSCROLL);
}

static void SizeWindow(HWND hWnd){
 DWORD dwStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
                 WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_THICKFRAME;
 RECT  Rectangle;
 int   dx=0,dy=0;
 POINT pt;
 LPBITMAPINFOHEADER bi;
 if (IsZoomed(hWnd))SetScrollRanges(hWnd);
 else {
   BITMAP bm;
   HBITMAP hbm;
   hbm=(HBITMAP)GetWindowLong(hWnd,GWL_BITMAP);
   if(hbm != NULL){
     GetObject(hbm,sizeof(BITMAP),(LPSTR)&bm);
     dx=bm.bmWidth; dy=bm.bmHeight;
     ptSize.x = (WORD)dx;
     ptSize.y = (WORD)dy;
   }
   Rectangle.left   = 0;
   Rectangle.top    = 0;
   Rectangle.right  = (WORD)dx;
   Rectangle.bottom = (WORD)dy;
   AdjustWindowRect(&Rectangle,dwStyle,TRUE);
   Rectangle.bottom -= GetSystemMetrics(SM_CYMENU);
   SetWindowPos (hWnd,(HWND)NULL,0,0,
                 Rectangle.right  - Rectangle.left,
                 Rectangle.bottom - Rectangle.top,
                 SWP_NOMOVE | SWP_NOZORDER);
 }
 InvalidateRect(hWnd,NULL,TRUE);
}

static void SetScrollRanges(HWND hwnd){
 RECT rc;
 int iRangeH,iRangeV,i;
 static int iSem = 0;
 if(!iSem){
   iSem++;
   GetRealClientRect(hwnd,&rc);
   for(i = 0; i < 2; i++){
     iRangeV = ptSize.y - rc.bottom;
     iRangeH = ptSize.x - rc.right;
     if(iRangeH < 0)iRangeH = 0;
     if(iRangeV < 0)iRangeV = 0;
     if (GetScrollPos(hwnd,SB_VERT)  > iRangeV ||
         GetScrollPos(hwnd, SB_HORZ) > iRangeH)
       InvalidateRect (hwnd, NULL, TRUE);
     SetScrollRange(hwnd, SB_VERT, 0, iRangeV, TRUE);
     SetScrollRange(hwnd, SB_HORZ, 0, iRangeH, TRUE);
     GetClientRect(hwnd, &rc);
   }
   iSem--;
 }
}

static void AppPaint(HWND hWnd,HDC hDC,int x,int y){
 HPALETTE hpalT,hpal;
 POINT Origin;
 HDC hDCbits;
 HBITMAP hbm;
 BITMAP bm;
 if(!IsIconic(hWnd)){
   SetWindowOrgEx(hDC,x,y,&Origin);
   hbm=(HBITMAP)GetWindowLong(hWnd,GWL_BITMAP);
 }
 else{
   SetWindowOrgEx(hDC,0,0,&Origin);
   hbm=(HBITMAP)GetWindowLong(hWnd,GWL_BITMAP_ICON);
 }
 hpal=(HPALETTE)GetWindowLong(hWnd,GWL_PALETTE);
 SetBkMode(hDC,TRANSPARENT);
 if(hpal != NULL){
   if(hWnd == ghwndAct){
     hpalT = SelectPalette(hDC,hpal,FALSE);
     RealizePalette(hDC);
   }
   if(hbm != NULL){
     hDCbits = CreateCompatibleDC(hDC);
     GetObject(hbm,sizeof(BITMAP),(LPSTR)&bm);
     SelectObject(hDCbits,hbm);
     BitBlt(hDC,0,0,bm.bmWidth,bm.bmHeight,hDCbits,0,0,SRCCOPY);
     DeleteDC(hDCbits);
   }
   if(hWnd == ghwndAct)SelectPalette(hDC,hpalT,FALSE);
 }
}

static int identify_file_format(char *filename){
 char CurrentFile[256];
 int ff=FI_BMP;
 strcpy(CurrentFile,filename);
 if     (strstr(CurrentFile,".GIF") || strstr(CurrentFile,".gif"))ff=FI_GIF;
 else if(strstr(CurrentFile,".CEL") || strstr(CurrentFile,".cel"))ff=FI_CEL;
 else if(strstr(CurrentFile,".TGA") || strstr(CurrentFile,".tga"))ff=FI_TGA;
 else if(strstr(CurrentFile,".DIB") || strstr(CurrentFile,".dib"))ff=FI_DIB;
 else if(strstr(CurrentFile,".BMP") || strstr(CurrentFile,".bmp") ||
         strstr(CurrentFile,".RLE") || strstr(CurrentFile,".rle"))ff=FI_BMP;
 else if(strstr(CurrentFile,".JPG") || strstr(CurrentFile,".jpg"))ff=FI_JPG;
 else if(strstr(CurrentFile,".TIF") || strstr(CurrentFile,".tif"))ff=FI_TIF;
 else if(strstr(CurrentFile,".PNG") || strstr(CurrentFile,".png"))ff=FI_PNG;
 return ff;
}

static char *short_form(char *filename){
 char *sps,*spc;
 sps=spc=filename;
 while((spc=strchr(spc,'\\')) != NULL){spc++; sps=spc;}
 return sps;
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[256],szdummy[256];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=256;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=256;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

#define PALVERSION  0x300
#define MAXPALETTE  256
#define DibNumColors(bmih) (bmih.biClrUsed == 0 && bmih.biBitCount <= 8 \
                            ? (int)(1 << (int)bmih.biBitCount)          \
                            : (int)bmih.biClrUsed)

static WORD NumColours(LPBITMAPINFOHEADER lpbi){
 WORD bits;
 if (lpbi->biClrUsed != 0)return (WORD)lpbi->biClrUsed;
 bits = lpbi->biBitCount;
 switch (bits){
  case 1:    return 2;
  case 4:    return 16;
  case 8:    return 256;
  default:   return 0; /* A 24 bitcount DIB has no color table */
 }
}

static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi){
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE       red;
 BYTE       green;
 BYTE       blue;
 int        i;
 RGBQUAD    *pRgb;
 if(lpbi == NULL)return NULL;
 if(lpbi->biSize != sizeof(BITMAPINFOHEADER))return NULL;
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 nNumColors = NumColours(lpbi);
 if(nNumColors){
   pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   for (i = 0; i < nNumColors; i++){
       pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
       pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
       pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 else if (lpbi->biBitCount == 24){
   nNumColors = MAXPALETTE;
   pPal=(LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   red = green = blue = 0;
   for (i = 0; i < pPal->palNumEntries; i++){
       pPal->palPalEntry[i].peRed   = red;
       pPal->palPalEntry[i].peGreen = green;
       pPal->palPalEntry[i].peBlue  = blue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
       if (!(red += 32))if (!(green += 32)) blue += 64;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 return hpal;
}

static HBITMAP BitmapFromDib(LPSTR lpHeader,
                             LPSTR lpBits,
                             HPALETTE hpal,
                             HBITMAP *hBitmapIcon){
 HDC       hdc;
 HBITMAP   hbm;
 HPALETTE  hpalT;
 if (lpHeader == NULL || lpBits == NULL)return NULL;
 hdc = GetDC(NULL);
 if(hpal != NULL){
   hpalT = SelectPalette(hdc,hpal,FALSE);
   RealizePalette(hdc);
 }
 hbm = CreateDIBitmap(hdc,
                (LPBITMAPINFOHEADER)lpHeader,
                (LONG)CBM_INIT,
                lpBits,
                (LPBITMAPINFO)lpHeader,
                DIB_RGB_COLORS );
 {
  LPBITMAPINFOHEADER lpbi;
  unsigned char Bits[4096];
  LPSTR iBits,bp;
  int i,j,x,dx,y,dy,nc,bpp;
  lpbi=(LPBITMAPINFOHEADER)lpHeader;
  iBits=(LPSTR)(&Bits[0]);
  nc = NumColours(lpbi); bpp=1;
  if(nc == 0)bpp=3;
  else       bpp=1;
  x=lpbi->biWidth; y=lpbi->biHeight;
  if(x < 32 || y < 32){
    if(x > 32)x=32; if(y > 32)y=32;
    lpbi->biWidth=x; lpbi->biHeight=y;
    bp=lpBits;
    for(i=0;i<y;i++)for(j=0;j<x;j++){
      *iBits++ = *bp++;
      if(bpp > 1){
        *iBits++ = *bp++;
        *iBits++ = *bp++;
      }
    }
  }
  else{
    x = ALIGNULONG(x);
    dx=x/32*bpp;
    dy=x*(y/32)*bpp;
    for(j=0;j<32;j++){
      bp=(lpBits + j*dy);
      for(i=0;i<32;i++){
        if(bpp == 3){
          *iBits++ = *(bp);
          *iBits++ = *(bp+1);
          *iBits++ = *(bp+2);
          bp += dx;
        }
        else{
          *iBits++ = *bp;
          bp += dx;
        }
      }
    }
    lpbi->biWidth=32; lpbi->biHeight=32;
  }
  iBits=(LPSTR)(&Bits[0]);
  *hBitmapIcon = CreateDIBitmap(hdc,
                lpbi,
                (LONG)CBM_INIT,
                iBits,
                (LPBITMAPINFO)lpHeader,
                DIB_RGB_COLORS );
 }
 if(hpal != NULL)SelectPalette(hdc,hpalT,FALSE);
 ReleaseDC(NULL,hdc);
 return hbm;
}

static HBITMAP RequiredBitmap(LPBITMAPINFOHEADER lpbi,
                              LPSTR lpBitmapHeader,
                              LPSTR lpBitmapBits,
                              HPALETTE *hpalCurrent,
                              HBITMAP *hBitmapIcon){

 HBITMAP hbmCurrent;
 if(!dither || lpbi->biBitCount < 8){
   hbmCurrent=BitmapFromDib(lpBitmapHeader,lpBitmapBits,*hpalCurrent,hBitmapIcon);
 }
 else{
   DeleteObject(*hpalCurrent); *hpalCurrent=NULL;
   *hpalCurrent=CreateMyHalftonePalette();
   lpbi=DibHalftoneDIB(lpbi,*hpalCurrent);
   lpBitmapBits = ((LPSTR)lpbi)
                + (long)sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD);
   hbmCurrent=BitmapFromDib((LPSTR)lpbi,lpBitmapBits,*hpalCurrent,hBitmapIcon);
   X__Free(lpbi);
 }
 X__Free(lpBitmapHeader);
 return hbmCurrent;
}

static HBITMAP LoadBmp(char *temp, HPALETTE *hpalCurrent,HBITMAP *hbmI){
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 HDC hdc;
 HANDLE hfbm;
 BITMAPFILEHEADER bmfh;
 DWORD dwRead;
 BITMAPINFOHEADER bmih;
 LPBITMAPINFOHEADER lpbi;
 LPBITMAPINFO lpinfo;
 long imagesize,nc;
 *hpalCurrent=NULL;
 hfbm=CreateFile(temp,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(hfbm == INVALID_HANDLE_VALUE)return NULL;
 ReadFile(hfbm,&bmfh,sizeof(BITMAPFILEHEADER),&dwRead,(LPOVERLAPPED)NULL);
 ReadFile(hfbm,&bmih,sizeof(BITMAPINFOHEADER),&dwRead,(LPOVERLAPPED)NULL);
 nc=DibNumColors(bmih);
 imagesize=bmfh.bfSize-bmfh.bfOffBits;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+
                   nc*sizeof(RGBQUAD)+imagesize);
 if(lpBitmapHeader == NULL)return 0;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpinfo=(LPBITMAPINFO)lpBitmapHeader;
 lpbi->biSize=bmih.biSize;
 lpbi->biWidth=bmih.biWidth;
 lpbi->biHeight=bmih.biHeight;
 lpbi->biPlanes=bmih.biPlanes;
 lpbi->biBitCount=bmih.biBitCount;
 lpbi->biCompression=bmih.biCompression;
 lpbi->biSizeImage=bmih.biSizeImage;
 lpbi->biXPelsPerMeter=bmih.biXPelsPerMeter;
 lpbi->biYPelsPerMeter=bmih.biYPelsPerMeter;
 lpbi->biClrUsed=bmih.biClrUsed;
 lpbi->biClrImportant=bmih.biClrImportant;
 if(nc > 0)ReadFile(hfbm,lpinfo->bmiColors,nc*sizeof(RGBQUAD),
                    &dwRead,(LPOVERLAPPED)NULL);
 *hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits = ((LPSTR)lpBitmapHeader)
              + (long)sizeof(BITMAPINFOHEADER)+nc*sizeof(RGBQUAD);
 memset(lpBitmapBits,0,imagesize);
 ReadFile(hfbm,lpBitmapBits,imagesize,&dwRead,(LPOVERLAPPED)NULL);
 CloseHandle(hfbm);
 return RequiredBitmap(lpbi,lpBitmapHeader,lpBitmapBits,hpalCurrent,hbmI);
}

static HBITMAP LoadDib(char *temp, HPALETTE *hpalCurrent,HBITMAP *hbmI){
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 HDC hdc;
 HANDLE hfbm;
 BITMAPFILEHEADER bmfh;
 DWORD dwRead;
 BITMAPINFOHEADER bmih;
 LPBITMAPINFOHEADER lpbi;
 LPBITMAPINFO lpinfo;
 long imagesize,nc;
 *hpalCurrent=NULL;
 hfbm=CreateFile(temp,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(hfbm == INVALID_HANDLE_VALUE)return NULL;
 ReadFile(hfbm,&bmfh,sizeof(BITMAPFILEHEADER),&dwRead,(LPOVERLAPPED)NULL);
 ReadFile(hfbm,&bmih,sizeof(BITMAPINFOHEADER),&dwRead,(LPOVERLAPPED)NULL);
//{char temp[256];
//sprintf(temp,"Width=%ld Height=%ld",bmih.biWidth,bmih.biHeight);
//MessageBox(NULL,temp,"DIB",MB_OK);}
 nc=DibNumColors(bmih);
 imagesize=bmfh.bfSize-bmfh.bfOffBits;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+
                   nc*sizeof(RGBQUAD)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpinfo=(LPBITMAPINFO)lpBitmapHeader;
 lpbi->biSize=bmih.biSize;
 lpbi->biWidth=bmih.biWidth;
 lpbi->biHeight=bmih.biHeight;
 lpbi->biPlanes=bmih.biPlanes;
 lpbi->biBitCount=bmih.biBitCount;
 lpbi->biCompression=bmih.biCompression;
 lpbi->biSizeImage=bmih.biSizeImage;
 lpbi->biXPelsPerMeter=bmih.biXPelsPerMeter;
 lpbi->biYPelsPerMeter=bmih.biYPelsPerMeter;
 lpbi->biClrUsed=bmih.biClrUsed;
 lpbi->biClrImportant=bmih.biClrImportant;
 if(nc > 0)ReadFile(hfbm,lpinfo->bmiColors,nc*sizeof(RGBQUAD),
                    &dwRead,(LPOVERLAPPED)NULL);
 *hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits = ((LPSTR)lpBitmapHeader)
              + (long)sizeof(BITMAPINFOHEADER)+nc*sizeof(RGBQUAD);
 memset(lpBitmapBits,0,imagesize);
 ReadFile(hfbm,lpBitmapBits,imagesize,&dwRead,(LPOVERLAPPED)NULL);
 CloseHandle(hfbm);
//X__Free(lpBitmapHeader);
 return NULL;
 return RequiredBitmap(lpbi,lpBitmapHeader,lpBitmapBits,hpalCurrent,hbmI);
}

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size,gif_iLace;

static int x4_size,rowcount,rowinc;
static LPSTR image_pixels;

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(y_size - 1 - rowcount)*(long)(x4_size);
 memcpy(image_pixels + l, linepixels, linelen);
 rowcount+=abs(rowinc);
 if(rowcount >= y_size){
  if     (rowinc == -8){rowinc=8; rowcount=4;}
  else if(rowinc ==  8){rowinc=4; rowcount=2;}
  else if(rowinc ==  4){rowinc=2; rowcount=1;}
 }
 return 0;
}

static HBITMAP LoadGif(char *temp, HPALETTE *hpalCurrent,HBITMAP *hbmI){
 long i,imagesize;
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 LPBITMAPINFOHEADER lpbi;
 RGBQUAD *pRgb;
 rowcount=0;
 rowinc=1;
 if(read_gif_image(temp,1) < 0)return NULL;
 if(gif_iLace)rowinc = -8;
 x4_size = ALIGNULONG(x_size);
 imagesize = (DWORD)x4_size*(DWORD)y_size;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)
                       +256L*sizeof(RGBQUAD)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=8;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage = (DWORD)imagesize;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=256;
 lpbi->biClrImportant=256;
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<256;i++){
   pRgb[i].rgbRed = p_red[i] << 2;
   pRgb[i].rgbGreen = p_green[i] << 2;
   pRgb[i].rgbBlue = p_blue[i] << 2;
 }
 *hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits = ((LPSTR)lpBitmapHeader)
              + (long)sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD);
 memset(lpBitmapBits,0,imagesize);
 image_pixels=lpBitmapBits;
 if(read_gif_image(temp,0) < 0){
   X__Free(lpBitmapHeader);
   return NULL;
 }
 return RequiredBitmap(lpbi,lpBitmapHeader,lpBitmapBits,hpalCurrent,hbmI);
}

static void tga_display_line(unsigned char *prgb,
                             long wide, long line,
                             long upsidedown, long hi){
 register long l,x,k;
// if(upsidedown)l = 3L*(long)(line)*(long)(x4_size);
// else          l = 3L*(long)(hi - 1 - line)*(long)(x4_size);
 if(upsidedown)l = (long)(line)*(long)(x4_size);
 else          l = (long)(hi - 1 - line)*(long)(x4_size);
 memcpy(image_pixels + l, prgb, (int)(wide*3));
}

static short R_getvariword(long n, FILE *f, long *vw){
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;          if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

static short ReadTGA(char *filename, short info_only){
 char c0,c1,c2;
 long i,j,k,l,m,n,cmo,cml,xor,yor,wi,hi,iok=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat,wi3;
 FILE *t24;
 unsigned char *prgb;
 prgb=NULL;
 if((t24=fopen(filename,"rb")) != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0 : 1 ;
   if(info_only){
     x_size=(int)wi;
     y_size=(int)hi;
     fclose(t24);
     return 1;
   }
   if((prgb=(unsigned char *)X__Malloc((size_t)wi*3)) == NULL)goto CLOSEOUT;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi;
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   wi3=wi*3;
   if(image_type == 2){ /* uncompressed */
     m=0; k=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       prgb[m  ] = (char)( vw        & pat) << s;
       prgb[m+1] = (char)((vw >> s1) & pat) << s;
       prgb[m+2] = (char)((vw >> s2) & pat) << s;
       k++; m+=3;
       if(m == wi3){
         tga_display_line(prgb,wi,i,upsidedown,hi); m=0;
       }
     }
   }
   else{  /* compressed */
     m=0; l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           prgb[m  ] = (char)( vw        & pat) << s;
           prgb[m+1] = (char)((vw >> s1) & pat) << s;
           prgb[m+2] = (char)((vw >> s2) & pat) << s;
           k++; m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           prgb[m  ] = c0;
           prgb[m+1] = c1;
           prgb[m+2] = c2;
           k++;  m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
   if(prgb != NULL)X__Free((char *)prgb);
 }
 return (short)iok;
}

static HBITMAP LoadTga(char *temp, HPALETTE *hpalCurrent,HBITMAP *hbmI){
 long i,imagesize;
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 LPBITMAPINFOHEADER lpbi;
 if(ReadTGA(temp,1) == 0)return NULL;
// x4_size=ALIGNULONG(x_size);
// imagesize = (DWORD)x4_size*(DWORD)y_size*3L;
 x4_size=ALIGNULONG(x_size*3);
 imagesize = (DWORD)x4_size*(DWORD)y_size;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=24;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage = imagesize;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=0;
 lpbi->biClrImportant=0;
 *hpalCurrent=CreateImagePalette(lpbi);  /* will set up a dummy palette */
 lpBitmapBits = ((LPSTR)lpBitmapHeader)+(long)sizeof(BITMAPINFOHEADER);
 memset(lpBitmapBits,0,imagesize);
 image_pixels=lpBitmapBits;
 if(ReadTGA(temp,0) == 0){
   X__Free(lpBitmapHeader);
   return NULL;
 }
 return RequiredBitmap(lpbi,lpBitmapHeader,lpBitmapBits,hpalCurrent,hbmI);
}

void jpg_display_line(unsigned char *pbgr, long line){
 unsigned char *p;
 register long i,l;
// l = 3L*(long)(y_size - 1 - line)*(long)(x4_size);
 l = (long)(y_size - 1 - line)*(long)(x4_size);
 p = (unsigned char *)(image_pixels+l);
 for(i=0;i<x_size;i++){
  *p++ = *(pbgr+2);
  *p++ = *(pbgr+1);
  *p++ = *(pbgr);
  pbgr+=3;
 }
}

extern long ReadJPG(char *filename, short info_only);

static HBITMAP LoadJpg(char *temp, HPALETTE *hpalCurrent,HBITMAP *hbmI){
 long i,imagesize;
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 LPBITMAPINFOHEADER lpbi;
 if(ReadJPG(temp,1) == 0)return NULL;
// x4_size=ALIGNULONG(x_size);
// imagesize = (DWORD)x4_size*(DWORD)y_size*3L;
 x4_size=ALIGNULONG(x_size*3);
 imagesize = (DWORD)x4_size*(DWORD)y_size;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=24;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage = imagesize;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=0;
 lpbi->biClrImportant=0;
 *hpalCurrent=CreateImagePalette(lpbi);  /* will set up a dummy palette */
 lpBitmapBits = ((LPSTR)lpBitmapHeader)+(long)sizeof(BITMAPINFOHEADER);
 memset(lpBitmapBits,0,imagesize);
 image_pixels=lpBitmapBits;
 if(ReadJPG(temp,0) == 0){
   X__Free(lpBitmapHeader);
   return NULL;
 }
 return RequiredBitmap(lpbi,lpBitmapHeader,lpBitmapBits,hpalCurrent,hbmI);
}

void tif_display_line(unsigned char *pbgr, long line){
 unsigned char *p;
 register long i,l;
 l = (long)(y_size - 1 - line)*(long)(x4_size);
 p = (unsigned char *)(image_pixels+l);
 for(i=0;i<x_size;i++){
  *p++ = *(pbgr+2);
  *p++ = *(pbgr+1);
  *p++ = *(pbgr);
  pbgr+=4;
 }
}

extern long ReadTIF(char *filename, short info_only);

static HBITMAP LoadTif(char *temp, HPALETTE *hpalCurrent,HBITMAP *hbmI){
 long i,imagesize;
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 LPBITMAPINFOHEADER lpbi;
 if(ReadTIF(temp,1) == 0)return NULL;
 x4_size=ALIGNULONG(x_size*3);
 imagesize = (DWORD)x4_size*(DWORD)y_size;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=24;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage = imagesize;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=0;
 lpbi->biClrImportant=0;
 *hpalCurrent=CreateImagePalette(lpbi);  /* will set up a dummy palette */
 lpBitmapBits = ((LPSTR)lpBitmapHeader)+(long)sizeof(BITMAPINFOHEADER);
 memset(lpBitmapBits,0,imagesize);
 image_pixels=lpBitmapBits;
 if(ReadTIF(temp,0) == 0){
   X__Free(lpBitmapHeader);
   return NULL;
 }
 return RequiredBitmap(lpbi,lpBitmapHeader,lpBitmapBits,hpalCurrent,hbmI);
}

void png_display_line(unsigned char *pbgr, long line){
 unsigned char *p;
 register long i,l;
 l = (long)(y_size - 1 - line)*(long)(x4_size);
 p = (unsigned char*)(image_pixels+l);
 for(i=0;i<x_size;i++){
  *p++ = *(pbgr+2);
  *p++ = *(pbgr+1);
  *p++ = *(pbgr);
  pbgr+=3;
 }
}

extern long ReadPNG(char *filename, short info_only);

static HBITMAP LoadPng(char *temp, HPALETTE *hpalCurrent,HBITMAP *hbmI){
 long i,imagesize;
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 LPBITMAPINFOHEADER lpbi;
 if(ReadPng(temp,1) == 0)return NULL;
 x4_size=ALIGNULONG(x_size*3);
 imagesize = (DWORD)x4_size*(DWORD)y_size;
 lpBitmapHeader = (LPSTR)X__Malloc((long)sizeof(BITMAPINFOHEADER)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=24;
 lpbi->biCompression=BI_RGB;
 lpbi->biSizeImage = imagesize;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=0;
 lpbi->biClrImportant=0;
 *hpalCurrent=CreateImagePalette(lpbi);  /* will set up a dummy palette */
 lpBitmapBits = ((LPSTR)lpBitmapHeader)+(long)sizeof(BITMAPINFOHEADER);
 memset(lpBitmapBits,0,imagesize);
 image_pixels=lpBitmapBits;
 if(ReadPng(temp,0) == 0){
   X__Free(lpBitmapHeader);
   return NULL;
 }
 return RequiredBitmap(lpbi,lpBitmapHeader,lpBitmapBits,hpalCurrent,hbmI);
}

