/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file PNG.C */

#include <stdlib.h>
#include <stdio.h>
//#include <windows.h>

#include "png.h"

extern int x_size,y_size;

extern void png_display_line(unsigned char *, long );

#define PNG_BYTES_TO_CHECK 4
static int check_if_png(FILE *fp){
 char buf[PNG_BYTES_TO_CHECK];
 if (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK)
    return 1;
 return(png_sig_cmp((png_bytep)buf, (png_size_t)0, PNG_BYTES_TO_CHECK));
}

long ReadPng(char *filename, short info_only){
 png_structp png_ptr;
 png_infop info_ptr;
 unsigned int sig_read = PNG_BYTES_TO_CHECK;
 png_uint_32 width, height;
 int bit_depth, color_type, interlace_type;
 long i,j;
 FILE *fp;
 png_bytep *row_pointers;
 if ((fp = fopen(filename, "rb")) == NULL)return 0;
 if(check_if_png(fp) != 0){
   fclose(fp);
   return 0;
 }
 png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
 if (png_ptr == NULL){
   fclose(fp);
   return 0;
 }
 info_ptr = png_create_info_struct(png_ptr);
 if (info_ptr == NULL){
   fclose(fp);
   png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
   return 0;
 }
 if (setjmp(png_ptr->jmpbuf))   {
   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
//MessageBox(NULL,"Bad reader",NULL,MB_OK);
   fclose(fp);
   return 0;
 }
 png_init_io(png_ptr, fp);
 png_set_sig_bytes(png_ptr, sig_read);
 png_read_info(png_ptr, info_ptr);
 png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
       &interlace_type, NULL, NULL);
 if(info_only){
   x_size=width;
   y_size=height;
 }
 else{
//MessageBox(NULL,"Image Read",NULL,MB_OK);
   /* tell libpng to strip 16 bit/color files down to 8 bits/color */
   png_set_strip_16(png_ptr);
   /* Strip alpha bytes from the input data without combining with the
    * background (not recommended).
    */
   png_set_strip_alpha(png_ptr);
   /* Extract multiple pixels with bit depths of 1, 2, and 4 from a single
    * byte into separate bytes (useful for paletted and grayscale images).
    */
   png_set_packing(png_ptr);
   /* Expand paletted colors into true RGB triplets */
   if (color_type == PNG_COLOR_TYPE_PALETTE)
     png_set_expand(png_ptr);
   /* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
   if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
     png_set_expand(png_ptr);
   /* Expand paletted or RGB images with transparency to full alpha channels
    * so the data will be available as RGBA quartets.
    */
   if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
     png_set_expand(png_ptr);
   if((row_pointers = (png_bytep *)malloc(height*sizeof(png_bytep))) != NULL){
     for (i = 0; i < height; i++){
       row_pointers[i] = malloc(png_get_rowbytes(png_ptr, info_ptr));
     }
     png_read_image(png_ptr, row_pointers);
     for(i=0;i<height;i++){
       if(row_pointers[i] != NULL){
         png_display_line((unsigned char*)row_pointers[i],i);
         free(row_pointers[i]);
       }
     }
     free(row_pointers);
   }
 }
 if(!info_only)png_read_end(png_ptr, info_ptr);
 png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
 fclose(fp);
//{char xxx[256];
//sprintf(xxx,"Info=%ld Width=%ld Height=%ld",info_only,width,height);
//MessageBox(NULL,xxx,NULL,MB_OK);
//}
 return 1;
}


typedef struct SCREENBUFFER {
 unsigned char A,  /* Alpha channel (not available yet) */
               R,  /* red value  range - 255            */
               G,  /* green value  range - 255          */
               B;  /* blue value  range - 255           */
} fullscreenbuffer;


long WritePng(char *filename, long x, long y, long quality,
             fullscreenbuffer *b){
 FILE *fp;
 png_structp png_ptr;
 png_infop info_ptr;
 unsigned char *row,*p;
 long i,j,pass;
 if((row = (unsigned char *)malloc(x*4*sizeof(unsigned char))) == NULL)
    return 0;
 fp = fopen(filename, "wb");
 if (fp == NULL){
  free(row);
  return 0;
 }
 png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
 if (png_ptr == NULL){
   fclose(fp);
   free(row);
   return 0;
 }
 info_ptr = png_create_info_struct(png_ptr);
 if (info_ptr == NULL){
   fclose(fp);
   free(row);
   png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
   return 0;
 }
 if (setjmp(png_ptr->jmpbuf)){
   fclose(fp);
   free(row);
   png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
   return 0;
 }
 png_init_io(png_ptr, fp);
 /* Set the image information here.  Width and height are up to 2^31,
  * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
  * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
  * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
  * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
  * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
  * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
  */
 png_set_IHDR(png_ptr,info_ptr,
              x,y,8,
              PNG_COLOR_TYPE_RGB_ALPHA, // ALPHA
              PNG_INTERLACE_NONE,
              PNG_COMPRESSION_TYPE_BASE,
              PNG_FILTER_TYPE_BASE);
 png_write_info(png_ptr, info_ptr);

 for (pass = 0; pass < 1; pass++){
   /* If you are only writing one row at a time, this works */
   for (i=0;i<y;i++){
     for(j=0,p=row;j<x;j++){
       *p++ = b->R;
       *p++ = b->G;
       *p++ = b->B;
       *p++ = b->A; // ALPHA
       b++;
     }
     png_write_rows(png_ptr, &row, 1);
   }
 }

 png_write_end(png_ptr, info_ptr);
 png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
 fclose(fp);
 free(row);
 return 1;
}
