/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/*  PLAYER.C    Player for  FLC / FLI files uses WinG */

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include "player.h"
#include "\EVI\render\flc.h"
#include "\EVI\render\fli.h"
#include "\wing\include\wing.h"

#if __SC__ || __WATCOMC__
extern INT _argc;
extern CHAR **_argv;
#else
extern INT __argc;
extern CHAR **__argv;
#endif

#if __SC__
#pragma ZTC align 1
#endif

#define EVI_Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define EVI_Free(p)      LocalFree((HLOCAL)p)
#define EVI_Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)

#define PALVERSION  0x300
#define MAXPALETTE  256
#define MAXTEXT     256

#define ALIGNULONG(i)   ((i+3)/4*4)

//extern void PASCAL fii_unlccomp(char *, char *);


typedef struct tag_ram_frame{
 char *frame_header;
 char *frame;
}ram_frame;

typedef struct header{
        BITMAPINFOHEADER Header;
        RGBQUAD aColors[256];
} header;

header BufferHeader;
long Orientation = -1;
void *pBuffer = NULL;
HDC Buffer = NULL;

typedef struct pal
{
        WORD Version;
        WORD NumberOfEntries;
        PALETTEENTRY aEntries[256];
} pal;

pal LogicalPalette =
{
        0x300,
        256
};

static ram_frame *RamFrames=NULL;
static HBITMAP   gbmOldMonoBitmap = NULL;
static HWND      hWndApp        = NULL;
static HINSTANCE hInst          = NULL;
static HPALETTE  hpalCurrent    = NULL;
static Flc_head  flc_cel;
static Fli_head  fli_cel;
static FILE *FpF=NULL;
static char *szViewClass="EVI:PlayClass";
static unsigned char *Screen,p_red[255],p_green[255],p_blue[255];
static int x_size,y_size,nframes=0;
static long nx,ny,nx4,frame2start;
static int DisplayWidth,DisplayHeight,DisplayBits,DisplayPlanes,
           xScreen,yScreen,x4_size,current_frame;
static char CurrentDirFile[256],CurrentDirectory[256],CurrentFile[256],
            TitleString[256];
static UINT  AnimationTimer=0;
static DWORD fdwStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
                        WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_THICKFRAME;
static BOOL AnimationOpen=FALSE,AnimationFLC=FALSE,StartedInAnimator=FALSE;

static BOOL LoadAnimationIntoRam(void);
static int ErrMsg(HWND hWnd, char *sz,...);
static void PutMsg(char *s);
LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,LPARAM lParam);
BOOL MenuCommand(HWND hWnd, WORD id);
static void FreeAnimation(void);
static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi);
static WORD NumColours(LPBITMAPINFOHEADER lpbi);
static BOOL ShowFirstFrame(HWND hWnd);
static BOOL ShowNextFlcFrame(HWND hWnd);
static BOOL ShowNextFliFrame(HWND hWnd);
static BOOL FlcBrun(char *);
static BOOL FliBrun(char *);
static BOOL FlcColor(char *, short);
static BOOL FliColor(char *);
static BOOL LoadFirstFlcCel(FILE *f, Flc_head *flc, int info_only);
static BOOL flc_unlccomp(char *);
static BOOL fli_unlccomp(char *);
static BOOL OpenFLC(char *name, int info_only);
static BOOL OpenFLI(char *name, int info_only);

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[80],szdummy[256];
 szdummy[0]='\0';
 strcpy(szFilter,szfilter);
 i=0; while(szFilter[i] != '\0'){
   if(szFilter[i] == '|')szFilter[i]='\0'; i++;
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=255;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=255;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

int Viewer(HINSTANCE hInstance){
 MSG msg;
 HWND hDesktopWnd,hWnd;
 HDC hDCcaps;
 WNDCLASS wndclass;
 hInst       = hInstance;
 hDesktopWnd = GetDesktopWindow();
 hDCcaps     = GetDC(hDesktopWnd);
 DisplayWidth= GetDeviceCaps(hDCcaps,HORZRES);
 DisplayHeight= GetDeviceCaps(hDCcaps,VERTRES);
 DisplayBits= GetDeviceCaps(hDCcaps,BITSPIXEL);
 DisplayPlanes= GetDeviceCaps(hDCcaps,PLANES);
 ReleaseDC(hDesktopWnd,hDCcaps);
 wndclass.style         = 0;
 wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
 wndclass.cbClsExtra    = 0 ;
 wndclass.cbWndExtra    = 0 ;
 wndclass.hInstance     = hInstance ;
 wndclass.hIcon         = LoadIcon(hInst, "PLAYERICON");
 wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
 wndclass.hbrBackground = GetStockObject (WHITE_BRUSH) ;
 wndclass.lpszMenuName  = "ShowMenu" ;
 wndclass.lpszClassName = szViewClass;
 if (!RegisterClass (&wndclass))  return FALSE ;
 xScreen = GetSystemMetrics (SM_CXSCREEN) ;
 yScreen = GetSystemMetrics (SM_CYSCREEN) ;
 hWnd = CreateWindow(
        szViewClass,
        "Envisage WinG Player",
        fdwStyle,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        320,
        250,
        NULL,
        NULL,
        hInstance,
        NULL) ;
 if(hWnd == NULL)return FALSE;
 ShowWindow (hWndApp = hWnd, SW_SHOWNA);
 StartedInAnimator=FALSE;
#if __SC__ || __WATCOMC__
 if(_argc > 1){
#else
 if(__argc > 1){
#endif
   StartedInAnimator=TRUE;
   GetModuleFileName(hInstance,CurrentDirectory,255);
   *FileInPath(CurrentDirectory) = '\0';
   strcat(CurrentDirectory,"flics");
 }
 PostMessage(hWnd,WM_COMMAND,(WPARAM)IDM_OPEN,0);
 while(GetMessage (&msg, NULL, 0, 0)) {
   TranslateMessage (&msg);
   DispatchMessage (&msg);
 }
 FreeAnimation();
 UnregisterClass(wndclass.lpszClassName,wndclass.hInstance);
 return msg.wParam;
}

LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,
                             LPARAM lParam){
 PAINTSTRUCT ps;
 HDC hDC;
 int i,iPos,iMax,iMin,dn;
 RECT rc;
 HPALETTE hpalT;
 POINT Origin;
 HDC hDCbits;
 BITMAP bm;
 switch (iMessage) {
   case WM_SYSCOMMAND:
     switch(LOWORD(wParam & 0xfff0)){
       case SC_CLOSE:
         PostMessage(hWnd,WM_COMMAND,IDM_EXIT,0);
         break;
       default:
         return( DefWindowProc( hWnd, iMessage, wParam, lParam ) );
      }
      break;
  case WM_COMMAND:
    return MenuCommand(hWnd,LOWORD(wParam));
    break;
  case WM_PAINT:
    hDC = BeginPaint(hWnd, &ps);
    if(hpalCurrent != NULL){
      hpalT = SelectPalette(hDC,hpalCurrent,FALSE);
      RealizePalette (hDC);
      if(Buffer != NULL)WinGBitBlt(hDC,0,0,x4_size,y_size,Buffer,0,0);
      SelectPalette(hDC,hpalT,FALSE);
    }
    EndPaint(hWnd, &ps);
    break ;
   case WM_TIMER:
     if(AnimationTimer == 1){
       if(AnimationFLC)ShowNextFlcFrame(hWnd);
       else            ShowNextFliFrame(hWnd);
     }
     break;
  default:
      return DefWindowProc (hWnd, iMessage, wParam, lParam) ;
 }
 return 0L ;
}

BOOL MenuCommand (HWND hWnd, WORD id){
 int i;
 char lpMessage[MAXTEXT],lpCaption[64];
 switch (id) {
   case IDM_OPEN:
     if(SelectFileName(0,CurrentFile,CurrentDirectory,"File to Play",
                      "Flic files (*.FLI *.FLC)|*.FL?|",hWnd)){
       SendMessage(hWnd,WM_COMMAND,(WPARAM)IDM_DRAWIT,0);
     }
     break;
   case IDM_DRAWIT:
     ShowFirstFrame(hWnd);
     break;
   case IDM_PLAY:
     if(AnimationOpen && AnimationTimer != 0){
//       KillTimer(hWnd,AnimationTimer);
       AnimationTimer=0;
       EnableMenuItem(GetMenu(hWnd),IDM_OPEN,MF_ENABLED);
       ModifyMenu(GetMenu(hWnd),IDM_PLAY,MF_BYCOMMAND|MF_STRING,
                  IDM_PLAY,"&Play");
       sprintf(TitleString,"Stopped %s (Frame %d)",
            FileInPath(CurrentFile),current_frame);
       SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)TitleString);
     }
     else if(AnimationOpen){
//       AnimationTimer=SetTimer(hWnd,1,33,NULL);
//       AnimationTimer=SetTimer(hWnd,1,5,NULL);
       EnableMenuItem(GetMenu(hWnd),IDM_OPEN,MF_GRAYED);
       ModifyMenu(GetMenu(hWnd),IDM_PLAY,MF_BYCOMMAND|MF_STRING,
                  IDM_PLAY,"&Stop");
       sprintf(TitleString,"Playing %s",FileInPath(CurrentFile));
       SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)TitleString);
{
       MSG msg;
       AnimationTimer=1;
       for (;;){
         if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)){
           TranslateMessage(&msg);
           DispatchMessage(&msg);
           if (msg.message == WM_COMMAND)break;
         }
         else {
           if(AnimationFLC)ShowNextFlcFrame(hWnd);
           else            ShowNextFliFrame(hWnd);
         }
       }
}
     }
     break;
   case IDM_EXTERNAL:
     if(!StartedInAnimator || AnimationTimer > 0)break;
   case IDM_EXIT:
     if(StartedInAnimator){
       HWND hWndPrev;
       hWndPrev = FindWindow("EVI:AnimatorClass",NULL);
       if(hWndPrev != NULL){
         SetForegroundWindow(hWndPrev);
         BringWindowToTop(hWndPrev);
         ShowWindow (hWndPrev,SW_SHOW);
       }
     }
     PostQuitMessage(0);
     break;
   default:
     break;
 }
 return TRUE;
}

static void FreeAnimation(void){
 if (hpalCurrent) DeleteObject(hpalCurrent);
 if(Buffer != NULL){
   HBITMAP hbm;
   hbm=(HBITMAP)SelectObject(Buffer,gbmOldMonoBitmap);
   DeleteObject(hbm);
   DeleteDC(Buffer);
   Buffer=NULL;
   pBuffer=NULL;
 }
 if(FpF != NULL)fclose(FpF);
 FpF            = NULL;
 hpalCurrent    = NULL;
 AnimationOpen=FALSE;
 AnimationTimer=0;
 if(RamFrames != NULL && nframes > 0){
   int i;
   for(i=0;i<nframes;i++){
     EVI_Free(RamFrames[i].frame_header);
     EVI_Free(RamFrames[i].frame);
   }
   EVI_Free(RamFrames);
   RamFrames=NULL; nframes=0;
 }
}

static int ErrMsg(HWND hWnd,char *sz,...){
 char ach[128];
 va_list arg_ptr;
 va_start(arg_ptr,sz);
 vsprintf(ach,sz,arg_ptr);   /* Format the string */
 va_end(arg_ptr);
 if(AnimationTimer > 0)KillTimer(hWnd,AnimationTimer);
 AnimationTimer=0;
 MessageBox (hWnd, ach, NULL, MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
 return FALSE;
}

static void PutMsg(char *s){
 MessageBox (NULL,s, "Player", MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
}

static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi){
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE       red;
 BYTE       green;
 BYTE       blue;
 int        i;
 RGBQUAD    *pRgb;
 if(lpbi == NULL)return NULL;
 if(lpbi->biSize != sizeof(BITMAPINFOHEADER))return NULL;
 /* Get a pointer to the color table and the number of colors in it */
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 nNumColors = NumColours(lpbi);
 pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) +
         nNumColors * sizeof(PALETTEENTRY));
 if(pPal == NULL)return NULL;
 pPal->palNumEntries = nNumColors;
 pPal->palVersion    = PALVERSION;
 /* Fill in the palette entries from the DIB color table and
  * create a logical color palette. */
 for (i = 0; i < nNumColors; i++){
     pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
     pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
     pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
     pPal->palPalEntry[i].peFlags = (BYTE)0;
 }
 hpal = CreatePalette(pPal);
 LocalFree((HANDLE)pPal);
 return hpal;
}

static WORD NumColours(LPBITMAPINFOHEADER lpbi){
 WORD bits;
 if (lpbi->biClrUsed != 0)return (WORD)lpbi->biClrUsed;
 bits = lpbi->biBitCount;
 switch (bits){
  case 1:    return 2;
  case 4:    return 16;
  case 8:    return 256;
  default:   return 0; /* A 24 bitcount DIB has no color table */
 }
}

static BOOL OpenFLC(char *name, int info_only){
 if((FpF=fopen(name,"rb")) == NULL)return FALSE;
 if(fread(&flc_cel,sizeof(Flc_head),1,FpF) != 1)return FALSE;
 if(flc_cel.type != FLCH_MAGIC)return FALSE;
 nx = x_size = flc_cel.width;
 ny = y_size = flc_cel.height;
 nx4=ALIGNULONG(nx);
 return LoadFirstFlcCel(FpF,&flc_cel,info_only);
}

static BOOL LoadFirstFlcCel(FILE *f,Flc_head *flc, int info_only){
 Flc_frame f_frame;
 Flc_chunk *f_chunk;
 long i,j,s;
 char c,*frame,*buff;
 unsigned char *p;
 fseek(f,flc->oframe1,SEEK_SET);
 if(fread(&f_frame,sizeof(Flc_frame),1,f) != 1 ||
           f_frame.type != FLCF_MAGIC)return FALSE;
 s=f_frame.size-16;
 if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
 if(fread(frame,1,s,FpF) != s)return FALSE;
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Flc_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLC_COLOR256 || f_chunk->type == FLC_COLOR){
     if(!FlcColor(buff,f_chunk->type)){
       EVI_Free(frame);
       return FALSE;
     }
     if(info_only)goto XXIT;
   }
   else if(!info_only && f_chunk->type == FLC_BRUN){
     if(!FlcBrun(buff)){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(!info_only && f_chunk->type == FLC_COPY){
     memcpy(Screen,buff,(long)nx*(long)ny);
   }
   buff += (f_chunk->size - 6);
 }
 XXIT:
 EVI_Free(frame);
 frame2start=ftell(f);
 return TRUE;
}

static BOOL OpenFLI(char *name, int info_only){
 if((FpF=fopen(name,"rb")) == NULL)return FALSE;
 if(fread(&fli_cel,sizeof(Fli_head),1,FpF) != 1)return FALSE;
 if(fli_cel.type != FLIH_MAGIC){
   return FALSE;
 }
 nx = x_size = fli_cel.width;
 ny = y_size = fli_cel.height;
 nx4=ALIGNULONG(nx);
//ErrMsg(NULL,"%d %d   nx4=%d",nx,ny,nx4);
 return LoadFirstFliCel(FpF,&fli_cel,info_only);
}

static BOOL LoadFirstFliCel(FILE *f,Fli_head *fli, int info_only){
 Fli_frame f_frame;
 Fli_chunk *f_chunk;
 long i,j,s;
 char c,*frame,*buff;
 unsigned char *p;
 if(fread(&f_frame,sizeof(Fli_frame),1,f) != 1){
   return FALSE;
 }
 if(f_frame.type != FLIF_MAGIC){
   return FALSE;
 }
 s=f_frame.size-16;
 if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
 if(fread(frame,1,s,FpF) != s)return FALSE;
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Fli_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLI_COLOR){
     if(!FliColor(buff)){
       EVI_Free(frame);
       return FALSE;
     }
     if(info_only)goto XXIT;
   }
   else if(!info_only && f_chunk->type == FLI_BRUN){
     if(!FliBrun(buff)){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(!info_only && f_chunk->type == FLI_COPY){
     for (j=0;j<ny;j++){
       p=Screen+(ny-1-j)*nx4;
       memcpy(p,buff+j*nx,nx);
     }
   }
   buff += (f_chunk->size - 6);
 }
 XXIT:
 EVI_Free(frame);
 frame2start=ftell(f);
 return TRUE;
}

BOOL FliColor(char *cbuf1){
 short i,j,pk,packets,change,colour;
 unsigned char skip=0;
 packets = *cbuf1++;
 packets += *cbuf1++ *256;
 if (packets<=0) return TRUE;
 for (pk=0; pk<packets; pk++ ){
   skip += *cbuf1++;
   change = *cbuf1++;
   if(change == 0)change=256;
   for(i=0; i<change; i++){
     colour = *cbuf1++;
     p_red[skip] = colour << 2;
     colour = *cbuf1++;
     p_green[skip] = colour << 2;
     colour = *cbuf1++;
     p_blue[skip++] = colour << 2;
   }
 }
 return TRUE;
}

static BOOL FliBrun(char *cbuf1){
 char size;
 unsigned char pk,packet;
 short i,line;
 char colour,*address;
 for (line=0;line<200;line++){
   address=(char *)(Screen+(long)line*(long)nx4);
   packet=*cbuf1++;
   for (pk=0; pk< packet; pk++){
     size=*cbuf1++;
     if(size > 0){
        colour = *cbuf1++;
        for(i = 0;i<size;i++)*address++ = colour;
     }
     else if(size < 0){
       for ( i = size; i<0; i++ )  *address++ = *cbuf1++;
     }
   }
 }
 return TRUE;
}

static BOOL FlcColor(char *buf, short type){
 short i,j,pk,packets,ci=0;
 unsigned char skip=0,change,colour;
 packets = *((short *)buf); buf += 2;
 for (pk=0; pk<packets; pk++ ) {
   skip = *((unsigned char *)buf); buf++;
   ci += skip;
   change = *((unsigned char *)buf); buf++;
   if (change == 0) change=255;
   for (i=0; i<=change; i++) {
     colour = *((unsigned char *)buf); buf++;
     if(type == FLC_COLOR)colour = colour << 2;
     p_red[ci] = colour;
     colour = *((unsigned char *)buf); buf++;
     if(type == FLC_COLOR)colour = colour << 2;
     p_green[ci] = colour;
     colour = (*(unsigned char *)buf); buf++;
     if(type == FLC_COLOR)colour = colour << 2;
     p_blue[ci] = colour;
     ci++;
   }
 }
 return TRUE;
}

static BOOL FlcBrun(char *buf){
 char size=0;
 unsigned char packet,*p,colour;
 short i,line,pk;
 for (line=0;line<ny;line++){
   p=Screen+(long)line*(long)nx4;
   packet = *((unsigned char *)buf); buf++;
   for(pk=0; pk<nx; pk+=abs(size)){
     size = *((char *)buf);  buf++;
     if(size > 0){
       colour = *((unsigned char *)buf); buf++;
       for (i=0; i<size; i++) *p++ = colour;
     }
     else{
       for (i = size; i<0; i++){
         colour = *((unsigned char *)buf); buf++;
         *p++ = colour;
       }
     }
   }
 }
 return TRUE;
}

static BOOL flc_unlccomp(char *buff){
 char size;
 unsigned char skip;
 short pk,packet;
 short lines,word,words[256],size2;
 unsigned char *p,*pl;
 long j,i;
 lines = *((short *)buff); buff+=2;
 p=Screen;
 for(j = 0; j < lines; j++){
   LOOP:
   packet = *((short *)buff); buff += 2;
   if(packet < 0){
     p += (long)abs(packet)*(long)x_size;
     goto LOOP;
   }
   pl=p;
   for(pk=0; pk<packet; pk++){
     skip = *((unsigned char *)buff); buff++;
     p += (long)skip;
     size = *((char *)buff); buff++;
     size2=(short)size;
     if(size < 0){
       word = *((short *)buff); buff+=2;
       size = -size;
       for(i=0; i<size; i++){
         *((short *)p)=word; p += 2;
       }
     }
     else{
       size2=(short)size+(short)size;
       memcpy(p,buff,size2); buff += size2;  p += size2;
     }
   }
   p=pl+(long)x_size;
 }
 return TRUE;
}

static BOOL fli_unlccomp(char *cbuf1){
 Fli_lc *flc;
 char *address;
 unsigned char skip,packet;
 short i,line;
 long s1,s2,size,pk;
 flc = (Fli_lc *)cbuf1;
 cbuf1 += 4;
 for( line = flc->skip; line < (short)(flc->skip)+(short)(flc->line); line++){
   packet = (unsigned char)(*cbuf1++);
   address=(char *)(Screen+(long)line*320L);
   for(pk=0; pk<packet; pk++){
     skip = (unsigned) *cbuf1++;
     address += (unsigned)skip;
     size = (long)(*cbuf1++);
     if(size < 0){
       size = -size;
       memset(address,*cbuf1,size);
       address += size;
       cbuf1++;
     }
     else{
       memcpy(address,cbuf1,size);
       cbuf1 += size; address += size;
     }
   }
 }
 return TRUE;
}

static BOOL ShowFirstFrame(HWND hWnd){
 int i,xc,yc,xs,ys;
 char temp[128];
 long imagesize;
 LPBITMAPINFOHEADER lpbi;
 HBITMAP hbm;
 RGBQUAD *pRgb;
 FreeAnimation();
 strcpy(temp,CurrentFile);
 if(strstr(strupr(temp),".FLI") != NULL)AnimationFLC=FALSE;
 else AnimationFLC=TRUE;
 if(AnimationFLC){
   if(!OpenFLC(temp,1)){
     ErrMsg(hWnd,"'%s' is NOT a valid FLC file", (LPSTR)temp);
     if(FpF != NULL)fclose(FpF); FpF=NULL;
     return FALSE;
   }
 }
 else{
   if(!OpenFLI(temp,1)){
     ErrMsg(hWnd,"'%ls' is NOT a valid FLI file", (LPSTR)temp);
     if(FpF != NULL)fclose(FpF); FpF=NULL;
     return FALSE;
   }
 }
 if(FpF != NULL)fclose(FpF); FpF=NULL;
 x4_size = ALIGNULONG(x_size);
 lpbi=(LPBITMAPINFOHEADER)&BufferHeader.Header;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x4_size;
 lpbi->biHeight=(DWORD)y_size*Orientation;
 lpbi->biPlanes=1;
 lpbi->biBitCount=8;
 lpbi->biCompression=BI_RGB;
 imagesize = lpbi->biSizeImage = (DWORD)x4_size*(DWORD)y_size;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=256;
 lpbi->biClrImportant=256;
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<256;i++){
   pRgb[i].rgbRed = p_red[i];
   pRgb[i].rgbGreen = p_green[i];
   pRgb[i].rgbBlue = p_blue[i];
 }
 hpalCurrent=CreateImagePalette(lpbi);
 Buffer=WinGCreateDC();
 hbm=WinGCreateBitmap(Buffer,(BITMAPINFO *)&BufferHeader,&pBuffer);
 gbmOldMonoBitmap=(HBITMAP)SelectObject(Buffer,hbm);
 if(pBuffer == NULL){
   ErrMsg(hWnd,"Out of memory");
   FreeAnimation();
   return FALSE;
 }
 Screen=(unsigned char *)pBuffer;
 memset(Screen,0,imagesize);
 if(AnimationFLC){
   if(!OpenFLC(temp,0)){
     PutMsg("Error reading FLC file");
     FreeAnimation();
     return FALSE;
   }
 }
 else{
   if(!OpenFLI(temp,0)){
     PutMsg("Error reading FLI file");
     FreeAnimation();
     return FALSE;
   }
 }
 current_frame=0;
 if(AnimationFLC)sprintf(TitleString,"%s (%u Frames)",
    FileInPath(CurrentFile),flc_cel.frames);
 else sprintf(TitleString,"%s (%u Frames)",
    FileInPath(CurrentFile),fli_cel.frame_count);
 SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)TitleString);
 xs=x_size+2*GetSystemMetrics(SM_CXFRAME);
 ys=y_size+GetSystemMetrics(SM_CYFRAME)+
           GetSystemMetrics(SM_CYCAPTION)+
           GetSystemMetrics(SM_CYMENU);
 xc=(xScreen-xs)/2;
 yc=(yScreen-ys)/2;
 if(xs > xScreen){xs=xScreen; xc=0;}
 if(ys > yScreen){ys=yScreen; yc=0;}
 SetWindowPos(hWnd,HWND_TOPMOST,xc,yc,xs,ys,SWP_NOMOVE);
 InvalidateRect(hWnd,NULL,TRUE);
 UpdateWindow(hWnd);
 AnimationOpen=TRUE;
 EnableMenuItem(GetMenu(hWnd),IDM_PLAY,MF_ENABLED);
 LoadAnimationIntoRam();
 return TRUE;
}

#define C current_frame

static BOOL LoadAnimationIntoRam(void){
 Fli_frame i_frame;
 Flc_frame f_frame;
 ram_frame *temp;
 char *frame;
 long s;
 SetCursor(LoadCursor(NULL,IDC_WAIT));
 MessageBeep(MB_OK);
 if(AnimationFLC){
   while(C < flc_cel.frames){
     if(RamFrames == NULL){
       RamFrames=(ram_frame *)EVI_Malloc(sizeof(ram_frame));
       if(RamFrames == NULL)goto ERRX;
     }
     else{
       temp=(ram_frame *)EVI_Realloc(RamFrames,sizeof(ram_frame)*(C+1));
       if(temp == NULL)goto ERRX;
       RamFrames=temp;
     }
     RamFrames[C].frame_header=NULL;
     RamFrames[C].frame=NULL;
     if(fread(&f_frame,sizeof(Flc_frame),1,FpF) != 1)goto ERRX;
     if(f_frame.type != FLCF_MAGIC)goto ERRX;
     s=f_frame.size-16;
     if((RamFrames[C].frame_header=(char *)
        EVI_Malloc(sizeof(Flc_frame))) == NULL)goto ERRX;
     memcpy(RamFrames[C].frame_header,&f_frame,sizeof(Flc_frame));
     s=f_frame.size-16;
     if((frame = (char *)EVI_Malloc(s)) == NULL)goto ERRX;
     RamFrames[C].frame=frame;
     if(fread(frame,1,s,FpF) != s)goto ERRX;
     C++;
   }
 }
 else{
   while(C < fli_cel.frame_count){
     if(RamFrames == NULL){
       RamFrames=(ram_frame *)EVI_Malloc(sizeof(ram_frame));
       if(RamFrames == NULL)goto ERRX;
     }
     else{
       temp=(ram_frame *)EVI_Realloc(RamFrames,sizeof(ram_frame)*(C+1));
       if(temp == NULL)goto ERRX;
       RamFrames=temp;
     }
     RamFrames[C].frame_header=NULL;
     RamFrames[C].frame=NULL;
     if(fread(&i_frame,sizeof(Fli_frame),1,FpF) != 1)goto ERRX;
     if(i_frame.type != FLIF_MAGIC)goto ERRX;
     s=i_frame.size-16;
          if((RamFrames[C].frame_header=(char *)
        EVI_Malloc(sizeof(Fli_frame))) == NULL)goto ERRX;
     memcpy(RamFrames[C].frame_header,&i_frame,sizeof(Fli_frame));
     s=i_frame.size-16;
     if((frame = (char *)EVI_Malloc(s)) == NULL)goto ERRX;
     RamFrames[C].frame=frame;
     if(fread(frame,1,s,FpF) != s)goto ERRX;
     C++;
   }
 }
 nframes=C;
 fclose(FpF); FpF=NULL;
 current_frame=0;
 SetCursor(LoadCursor(NULL,IDC_ARROW));
 MessageBeep(MB_OK);
 return TRUE;
 ERRX:
 MessageBox(NULL,"MEMORY ERROR",NULL,MB_OK);
 PostQuitMessage(0);
 SetCursor(LoadCursor(NULL,IDC_ARROW));
 return FALSE;
}

#undef C

static BOOL ShowNextFlcFrame(HWND hWnd){
 Flc_frame f_frame;
 Flc_chunk *f_chunk;
 char c,*frame,*buff;
 int i,j;
 long s;
 current_frame++;
 if(current_frame > flc_cel.frames){
   if(FpF != NULL){
     fseek(FpF,frame2start,SEEK_SET);   /* but just in case user ours */
   }
   current_frame=1;
 }
 if(FpF != NULL){
   if(fread(&f_frame,sizeof(Flc_frame),1,FpF) != 1){
     ErrMsg(hWnd,"Not an FLC frame 1");
     return FALSE;
   }
   if(f_frame.type != FLCF_MAGIC){
     ErrMsg(hWnd,"Not an FLC frame 2");
     return FALSE;
   }
   s=f_frame.size-16;
   if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
   if(fread(frame,1,s,FpF) != s)return FALSE;
 }
 else if(RamFrames != NULL){
   memcpy(&f_frame,RamFrames[current_frame-1].frame_header,sizeof(Flc_frame));
   frame=RamFrames[current_frame-1].frame;
 }
 else{
   MessageBox(NULL,"DISPLAY ERROR",NULL,MB_OK);
   PostQuitMessage(0);
   return FALSE;
 }
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Flc_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLC_SS2){   /* found cell */
     if(flc_unlccomp(buff) < 0){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(f_chunk->type == FLC_LC){
//     fii_unlccomp(buff,Screen);
     if(fli_unlccomp(buff) < 0){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(f_chunk->type == FLC_COPY){
     memcpy(Screen,buff,(long)nx*(long)ny);
   }
   buff += (f_chunk->size - 6);
 }
 if(FpF != NULL)EVI_Free(frame);
 InvalidateRect(hWnd,NULL,FALSE);
 return TRUE;
}

static BOOL ShowNextFliFrame(HWND hWnd){
 Fli_frame f_frame;
 Fli_chunk *f_chunk;
 char c,*frame,*buff;
 int i,j;
 long s;
 current_frame++;
 if(current_frame > fli_cel.frame_count){
   if(FpF != NULL){
     fseek(FpF,frame2start,SEEK_SET);   /* but just in case user ours */
   }
   current_frame=1;
 }
 if(FpF != NULL){
   if(fread(&f_frame,sizeof(Fli_frame),1,FpF) != 1){
     ErrMsg(hWnd,"Not an FLI frame 1");
     return FALSE;
   }
   if(f_frame.type != FLIF_MAGIC){
     ErrMsg(hWnd,"Not an FLI frame 2");
     return FALSE;
   }
   s=f_frame.size-16;
   if((frame = (char *)EVI_Malloc(s)) == NULL)return FALSE;
   if(fread(frame,1,s,FpF) != s)return FALSE;
 }
 else if(RamFrames != NULL){
   memcpy(&f_frame,RamFrames[current_frame-1].frame_header,sizeof(Fli_frame));
   frame=RamFrames[current_frame-1].frame;
 }
 else{
   MessageBox(NULL,"DISPLAY ERROR",NULL,MB_OK);
   PostQuitMessage(0);
   return FALSE;
 }
 buff=frame;
 for(i=0;i<f_frame.chunks;i++){
   f_chunk=(Fli_chunk *)buff;  buff += 6;
   if(f_chunk->type == FLI_LC){
     if(fli_unlccomp(buff) < 0){
       EVI_Free(frame);
       return FALSE;
     }
   }
   else if(f_chunk->type == FLI_COPY){
     unsigned char *p;
     for (j=0;j<ny;j++){
       p=Screen+j*nx4;
       memcpy(p,buff+j*nx,nx);
     }
   }
   buff += (f_chunk->size - 6);
 }
 if(FpF != NULL)EVI_Free(frame);
 InvalidateRect(hWnd,NULL,FALSE);
 return TRUE;
}


#if 0

int APIENTRY LibMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved){
  switch (dwReason) {
    case DLL_PROCESS_ATTACH: {
      MessageBox ( GetFocus(),
                  (LPCTSTR) "External texture: Process attaching",
                  (LPCTSTR) "ATTTACH",
                  MB_OK | MB_SYSTEMMODAL);
      break;
    }
    case DLL_PROCESS_DETACH:
      MessageBox ( GetFocus(),
                  (LPCTSTR) "External texture: Process detaching",
                  (LPCTSTR) "DETATCH",
                  MB_OK | MB_SYSTEMMODAL );
      break;
  }
  return (int)TRUE;
}

#if __SC__
#pragma startaddress(LibMain)
#endif

#endif

/* command line assumes a FULL path + filename */

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 HWND hWndPrev;
 hWndPrev = FindWindow(szViewClass,NULL);
 if (hWndPrev != NULL){
   SetForegroundWindow(hWndPrev);
   BringWindowToTop(hWndPrev);
   ShowWindow (hWndPrev,SW_SHOW);
   return 32;
 }
 if(hPrevInstance != NULL)return FALSE;
 return Viewer(hInstance);
}


