/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* Screen Saver Template */

#include <windows.h>
#include <commdlg.h>
#include <scrnsave.h>
#include <stdio.h>
#include <stdlib.h>
#include <vfw.h>

#include "saver1.h"

#define EVI_Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define EVI_Free(p)      LocalFree((HLOCAL)p)
#define EVI_Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#define PALVERSION  0x300
#define MAXPALETTE  256
#define DibNumColors(bmih) (bmih.biClrUsed == 0 && bmih.biBitCount <= 8 \
                            ? (int)(1 << (int)bmih.biBitCount)          \
                            : (int)bmih.biClrUsed)
#define GEN_STRING_SIZE 64

int      nbitmaps=2,pcount=0,disid=0;
LPSTR    lpgBitmap[2],lpgBitmapBits[2];
HPALETTE hPalette[2];
CHAR     wszScreenSaverTitle[GEN_STRING_SIZE];
CHAR     wszClassName[] = "GenericScreenSaverClass";
UINT     idTimer=0,xSize,ySize,wxSize=10,wySize=10;
float    xTransInc=2.0f,yTransInc=2.5f;
HWND     hwndFloat=NULL;
BOOL     random=FALSE,images=FALSE;

LRESULT WINAPI WndProc(HWND, UINT, WPARAM, LPARAM);

static WORD NumColours(LPBITMAPINFOHEADER lpbi){
 WORD bits;
 if (lpbi->biClrUsed != 0)return (WORD)lpbi->biClrUsed;
 bits = lpbi->biBitCount;
 switch (bits){
  case 1:    return 2;
  case 4:    return 16;
  case 8:    return 256;
  default:   return 0; /* A 24 bitcount DIB has no color table */
 }
}

static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi){
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE       red;
 BYTE       green;
 BYTE       blue;
 int        i;
 RGBQUAD    *pRgb;
 if(lpbi == NULL)return NULL;
 if(lpbi->biSize != sizeof(BITMAPINFOHEADER))return NULL;
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 nNumColors = NumColours(lpbi);
 if(nNumColors){
   pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   for (i = 0; i < nNumColors; i++){
       pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
       pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
       pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 else if (lpbi->biBitCount == 24){
   nNumColors = MAXPALETTE;
   pPal=(LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   red = green = blue = 0;
   for (i = 0; i < pPal->palNumEntries; i++){
       pPal->palPalEntry[i].peRed   = red;
       pPal->palPalEntry[i].peGreen = green;
       pPal->palPalEntry[i].peBlue  = blue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
       if (!(red += 32))if (!(green += 32)) blue += 64;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 return hpal;
}

static LPSTR LoadDib(char *temp, HPALETTE *hpalCurrent, LPSTR *lpBits){
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 HDC hdc;
 HANDLE hfbm;
 BITMAPFILEHEADER bmfh;
 DWORD dwRead;
 BITMAPINFOHEADER bmih;
 LPBITMAPINFOHEADER lpbi;
 LPBITMAPINFO lpinfo;
 long imagesize,nc;
 *hpalCurrent=NULL;
 hfbm=CreateFile(temp,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(hfbm == INVALID_HANDLE_VALUE)return NULL;
 ReadFile(hfbm,&bmfh,sizeof(BITMAPFILEHEADER),&dwRead,(LPOVERLAPPED)NULL);
 ReadFile(hfbm,&bmih,sizeof(BITMAPINFOHEADER),&dwRead,(LPOVERLAPPED)NULL);
 nc=DibNumColors(bmih);
 imagesize=bmfh.bfSize-bmfh.bfOffBits;
 lpBitmapHeader = (LPSTR)EVI_Malloc((long)sizeof(BITMAPINFOHEADER)+
                   nc*sizeof(RGBQUAD)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpinfo=(LPBITMAPINFO)lpBitmapHeader;
 lpbi->biSize=bmih.biSize;
 lpbi->biWidth=bmih.biWidth;
 lpbi->biHeight=bmih.biHeight;
 lpbi->biPlanes=bmih.biPlanes;
 lpbi->biBitCount=bmih.biBitCount;
 lpbi->biCompression=bmih.biCompression;
 lpbi->biSizeImage=bmih.biSizeImage;
 lpbi->biXPelsPerMeter=bmih.biXPelsPerMeter;
 lpbi->biYPelsPerMeter=bmih.biYPelsPerMeter;
 lpbi->biClrUsed=bmih.biClrUsed;
 lpbi->biClrImportant=bmih.biClrImportant;
 if(nc > 0)ReadFile(hfbm,lpinfo->bmiColors,nc*sizeof(RGBQUAD),
                    &dwRead,(LPOVERLAPPED)NULL);
 *hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits = ((LPSTR)lpBitmapHeader)
              + (long)sizeof(BITMAPINFOHEADER)+nc*sizeof(RGBQUAD);
 memset(lpBitmapBits,0,imagesize);
 ReadFile(hfbm,lpBitmapBits,imagesize,&dwRead,(LPOVERLAPPED)NULL);
 CloseHandle(hfbm);
 *lpBits=lpBitmapBits;
 return lpBitmapHeader;
}

BOOL WINAPI RegisterDialogClasses(HANDLE hinst){
  return TRUE;
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static int SelectFileName(int save, int avi, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[256],szdummy[256];
 szdummy[0]='\0';
 if(szfilter == NULL)AVIBuildFilter(szFilter,255,FALSE);
 else{
   strcpy(szFilter,szfilter);
   i=0; while(szFilter[i] != '\0'){
     if(szFilter[i] == '|')szFilter[i]='\0'; i++;
   }
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=256;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=256;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   if(avi)i=(int)(GetOpenFileNamePreview(&ofn));
   else   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

static BOOL WriteRegistryItem(char *keyname, char *keyvaluename,
                              char *keyvalue){
 HKEY hkey;
 DWORD result;
 if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,keyname,0,NULL,REG_OPTION_NON_VOLATILE,
                   KEY_ALL_ACCESS,NULL,&hkey,&result) != ERROR_SUCCESS)
   return FALSE;
 if     (result == REG_CREATED_NEW_KEY    ){;}
 else if(result == REG_OPENED_EXISTING_KEY){;}
 else                                      {;}
 if(keyvalue != NULL){
   if(RegSetValueEx(hkey,keyvaluename,0,REG_SZ,keyvalue,strlen(keyvalue)+1)
     != ERROR_SUCCESS)return FALSE;
 }
 RegCloseKey(hkey);
 return TRUE;
}

static BOOL ReadRegistryItem(char *keyname, char *keyvaluename,
                             LPBYTE value, DWORD size){
 HKEY hkey;
 DWORD result,type;
 if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,keyname,0,NULL,REG_OPTION_NON_VOLATILE,
                   KEY_ALL_ACCESS,NULL,&hkey,&result) != ERROR_SUCCESS)
   return FALSE;
 if     (result == REG_CREATED_NEW_KEY    ){;}
 else if(result == REG_OPENED_EXISTING_KEY){
   if(RegQueryValueEx(hkey,keyvaluename,NULL,&type,value,&size)
                      != ERROR_SUCCESS){
     RegCloseKey(hkey);
     return FALSE;
   }
   RegCloseKey(hkey);
   return TRUE;
 }
 else {;}
 RegCloseKey(hkey);
 return FALSE;
}

char keyname[]="System\\CurrentControlSet\\Control\\ProductOptions\\Saver1";
char moviekey[]="movie_name";
char image1key[]="image1_name";
char image2key[]="image2_name";
char typename[]="type";
char randomname[]="random";

BOOL WINAPI ScreenSaverConfigureDialog(HWND hDlg, UINT message,
                           WPARAM wParam, LPARAM lParam){
    static char moviename[256],image1name[256],image2name[256],temp[16];
    static char szFileName[256]=" ",szDirName[256]=" ";
    switch (message) {
        case WM_INITDIALOG:
            if(!ReadRegistryItem(keyname,moviekey,moviename,256))
              strcpy(moviename,"\\*.avi");
            SetDlgItemText(hDlg,DLG_SAVER_MOVIE,moviename);
            if(!ReadRegistryItem(keyname,image1key,image1name,256))
              strcpy(image1name,"\\*.bmp");
            SetDlgItemText(hDlg,DLG_SAVER_IMAGE1,image1name);
            if(!ReadRegistryItem(keyname,image2key,image2name,256))
              strcpy(image2name,"\\*.bmp");
            SetDlgItemText(hDlg,DLG_SAVER_IMAGE2,image2name);
            if(ReadRegistryItem(keyname,typename,temp,16)){
              if(temp[0] == 'Y')images=TRUE;
              else              images=FALSE;
            }
            if(ReadRegistryItem(keyname,randomname,temp,16)){
              if(temp[0] == 'Y')random=TRUE;
              else              random=FALSE;
            }
            if(random)
              SendDlgItemMessage(hDlg,DLG_SAVER_RANDOM,BM_SETCHECK,TRUE,0);
            else
              SendDlgItemMessage(hDlg,DLG_SAVER_DRIFT,BM_SETCHECK,TRUE,0);
            if(images)
              SendDlgItemMessage(hDlg,DLG_SAVER_DIB,BM_SETCHECK,TRUE,0);
            else
              SendDlgItemMessage(hDlg,DLG_SAVER_AVI,BM_SETCHECK,TRUE,0);
            return TRUE;
            break;
        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                case DLG_SAVER_MOVIE_C:
                  strcpy(szFileName,moviename);
                  strcpy(szDirName,moviename); *FileInPath(szDirName)='\0';
                  if(SelectFileName(0,1,szFileName,szDirName,
                     "Select Movie","Video Files|*.avi|",hDlg) == TRUE){
                    strcpy(moviename,szFileName);
                    SetDlgItemText(hDlg,DLG_SAVER_MOVIE,moviename);
                  }
                  break;
                case DLG_SAVER_IMAGE1_C:
                  strcpy(szFileName,image1name);
                  strcpy(szDirName,image1name); *FileInPath(szDirName)='\0';
                  if(SelectFileName(0,0,szFileName,szDirName,
                     "Select Bitmap","Image Files|*.bmp|",hDlg) == TRUE){
                    strcpy(image1name,szFileName);
                    SetDlgItemText(hDlg,DLG_SAVER_IMAGE1,image1name);
                  }
                  break;
                case DLG_SAVER_IMAGE2_C:
                  strcpy(szFileName,image2name);
                  strcpy(szDirName,image2name); *FileInPath(szDirName)='\0';
                  if(SelectFileName(0,0,szFileName,szDirName,
                     "Select Bitmap","Image Files|*.bmp|",hDlg) == TRUE){
                    strcpy(image2name,szFileName);
                    SetDlgItemText(hDlg,DLG_SAVER_IMAGE2,image2name);
                  }
                  break;
                case IDOK:
                  if(SendDlgItemMessage(hDlg,DLG_SAVER_RANDOM,BM_GETCHECK,0,0))
                    random=TRUE;
                  if(SendDlgItemMessage(hDlg,DLG_SAVER_DRIFT,BM_GETCHECK,0,0))
                    random=FALSE;
                  if(SendDlgItemMessage(hDlg,DLG_SAVER_DIB,BM_GETCHECK,0,0))
                    images=TRUE;
                  if(SendDlgItemMessage(hDlg,DLG_SAVER_AVI,BM_GETCHECK,0,0))
                    images=FALSE;
                  WriteRegistryItem(keyname,moviekey,moviename);
                  WriteRegistryItem(keyname,image1key,image1name);
                  WriteRegistryItem(keyname,image2key,image2name);
                  if(random)WriteRegistryItem(keyname,randomname,"Y");
                  else      WriteRegistryItem(keyname,randomname,"N");
                  if(images)WriteRegistryItem(keyname,typename,"Y");
                  else      WriteRegistryItem(keyname,typename,"N");
                  EndDialog(hDlg, TRUE);
                  break;
                case IDCANCEL:
                  EndDialog(hDlg, FALSE);
                  break;
                default:
                  break;
            }
            return TRUE;
            break;

        default:
            return 0;
    }
    return 0;
}

BOOL initClass(HANDLE hMainInstance){
 WNDCLASS wclass;
 wclass.style = CS_VREDRAW | CS_HREDRAW;
 wclass.lpfnWndProc = (WNDPROC)WndProc;
 wclass.cbClsExtra = 0;
 wclass.cbWndExtra = 0;
 wclass.hInstance = hMainInstance;
 wclass.hIcon = NULL;
 wclass.hCursor = NULL;
 wclass.hbrBackground = NULL;
 wclass.lpszMenuName = (LPSTR)NULL;
 wclass.lpszClassName = (LPSTR)wszClassName;
 return RegisterClass(&wclass);
}

BOOL initInstance(HANDLE hMainInstance,HWND parent){
 HDC hdc;
 DWORD style;
 ySize=GetSystemMetrics(SM_CYSCREEN);
 xSize=GetSystemMetrics(SM_CXSCREEN);
 style = WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
 hwndFloat = CreateWindowEx(WS_EX_TOPMOST,
                            wszClassName,
                            wszScreenSaverTitle,
                            style,
                            0,
                            0,
                            wxSize,
                            wySize,
                            parent,
                            NULL,
                            hMainInstance,
                            NULL
                            );
 if(hwndFloat == NULL)return FALSE;
 return TRUE;
}

void vMoveBuffer(int curX, int curY, int lastX, int lastY,
                int pxsize, int pysize){
 if(IsWindow(hwndFloat)){
   if(images)SetWindowPos(hwndFloat, HWND_TOPMOST, curX, curY, 0, 0, SWP_NOSIZE |
                 SWP_NOCOPYBITS);
   else SetWindowPos(hwndFloat, HWND_TOPMOST, curX, curY, 0, 0, SWP_NOSIZE);
 }
}

BOOL vShowBuffer(HWND hwnd){
 HDC hdc;
 static float xtrans = 0.0f;
 static float ytrans = 0.0f;
 static int lastXt = 0;
 static int lastYt = 0;
 static int bInited = FALSE;
 static int interBounce = 0;
 int bounce = FALSE;
 int xt, yt;
 if(!bInited){
   bInited = TRUE;
   srand(clock() & 0xffff);
   rand();
   xtrans = ((float)rand() / (float)RAND_MAX) * (float)(xSize - wxSize);
   ytrans = ((float)rand() / (float)RAND_MAX) * (float)(ySize - wySize);
   lastXt = (int) xtrans;
   lastYt = (int) ytrans;
 }
 xt = (int)xtrans;
 yt = (int)ytrans;
 vMoveBuffer(xt, yt, lastXt, lastYt, wxSize, wySize);
 lastXt = xt;
 lastYt = yt;
 xtrans += xTransInc;
 ytrans += yTransInc;
 if ((xtrans + wxSize) > xSize) {
   xtrans = (float) (xSize - wxSize);
   xTransInc = -xTransInc;
   bounce = TRUE;
 }
 else if (xtrans < 0.0f) {
   xtrans = 0.0f;
   xTransInc = -xTransInc;
   bounce = TRUE;
 }
 if ((ytrans + wySize) > ySize) {
   ytrans = (float) (ySize - wySize);
   yTransInc = -yTransInc;
   bounce = TRUE;
 }
 else if (ytrans < 0.0f) {
   ytrans = 0.0f;
   yTransInc = -yTransInc;
   bounce = TRUE;
 }
 interBounce++;
 if (bounce) {
   if (interBounce < 10)bounce = FALSE;
   else            interBounce = 0;
 }
 return bounce;
}

static void AppPaint(HWND hWnd,HDC hDC,int id){
 HPALETTE hpalT;
 if(IsIconic(hWnd))return;
 if(hPalette[id] != NULL){
   hpalT = SelectPalette(hDC,hPalette[id],FALSE);
   RealizePalette(hDC);
   if(lpgBitmap[id] != NULL)SetDIBitsToDevice(hDC,
                        0,0,
                        (WORD)((LPBITMAPINFOHEADER)lpgBitmap[id])->biWidth,
                        (WORD)((LPBITMAPINFOHEADER)lpgBitmap[id])->biHeight,
                        0,0,
                        0,
                        (WORD)((LPBITMAPINFOHEADER)lpgBitmap[id])->biHeight,
                        lpgBitmapBits[id],
                        (LPBITMAPINFO)lpgBitmap[id],
                        DIB_RGB_COLORS);
   SelectPalette(hDC,hpalT,FALSE);
 }
}

LRESULT WINAPI WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam){
  HDC hdc;
  PAINTSTRUCT ps;
  int retVal;
  switch(message){
    case WM_PAINT:
      hdc=BeginPaint(hwnd,&ps);
      AppPaint(hwnd,hdc,disid);
      EndPaint(hwnd, &ps);
      break;
    case WM_SYSCOMMAND:
    case WM_SETCURSOR:
    case WM_NCACTIVATE:
    case WM_ACTIVATE:
    case WM_ACTIVATEAPP:
    case WM_MOUSEMOVE:
    case WM_LBUTTONDOWN:
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_KEYDOWN:
    case WM_SYSKEYDOWN:
      retVal = DefScreenSaverProc(hwnd, message, wParam, lParam);
      DefScreenSaverProc(hMainWindow, message, wParam, lParam);
      return retVal;
    default:
      return DefWindowProc(hwnd, message, wParam, lParam);
  }
  return 0;
}

static void TimerProc(HWND hwnd){
  static BOOL busy=FALSE;
  if(busy)return;
  busy=TRUE;
  if(random || pcount > 100){
    disid++;
    if(disid >= nbitmaps)disid=0;
    pcount=0;
  }
  else pcount++;
  if(random){
    long curX,curY;
    curX=((long)rand()*((long)xSize-(long)wxSize))/32767L;
    curY=((long)rand()*((long)ySize-(long)wySize))/32767L;
     if(images){
       SetWindowPos(hwndFloat, HWND_TOPMOST, curX, curY, 0, 0, SWP_NOSIZE |
                   SWP_NOCOPYBITS);
       InvalidateRect(hwndFloat,NULL,FALSE);
     }
     else SetWindowPos(hwndFloat, HWND_TOPMOST, curX, curY, 0, 0, SWP_NOSIZE);
  }
  else vShowBuffer(hwnd);
  busy=FALSE;
}

LRESULT WINAPI ScreenSaverProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam){
 int i;
 HDC hdc;
 PAINTSTRUCT ps;
 LPBITMAPINFOHEADER hb;
 DWORD style;
 static char moviename[256],image1name[256],image2name[256],temp[16];
 switch (message){
   case WM_CREATE:
     for(i=0;i<nbitmaps;i++){lpgBitmap[i]=NULL; hPalette[i]=NULL;}
     LoadString(hMainInstance,IDS_DESCRIPTION,wszScreenSaverTitle,
                GEN_STRING_SIZE);
     if(!ReadRegistryItem(keyname,moviekey,moviename,256)){;}
     if(!ReadRegistryItem(keyname,image1key,image1name,256)){;}
     if(!ReadRegistryItem(keyname,image2key,image2name,256)){;}
     if(ReadRegistryItem(keyname,typename,temp,16)){
       if(temp[0] == 'Y')images=TRUE;
       else              images=FALSE;
     }
     if(ReadRegistryItem(keyname,randomname,temp,16)){
       if(temp[0] == 'Y')random=TRUE;
       else              random=FALSE;
     }
     if(images){
       lpgBitmap[0]=LoadDib(image1name,
                            &hPalette[0],&lpgBitmapBits[0]);
       lpgBitmap[1]=LoadDib(image2name,
                            &hPalette[1],&lpgBitmapBits[1]);
       hb=(LPBITMAPINFOHEADER)lpgBitmap[0];
       if(hb == NULL)hb=(LPBITMAPINFOHEADER)lpgBitmap[1];
       if(hb != NULL){
         wxSize=hb->biWidth;
         wySize=hb->biHeight;
         initClass(hMainInstance);
         initInstance(hMainInstance,hwnd);
         UpdateWindow(hwndFloat);
       }
     }
     else{
       if(MCIWndRegisterClass()){
         RECT rc;
         ySize=GetSystemMetrics(SM_CYSCREEN);
         xSize=GetSystemMetrics(SM_CXSCREEN);
         style = WS_POPUP | MCIWNDF_NOPLAYBAR | MCIWNDF_NOTIFYMODE;
         hwndFloat = CreateWindowEx(WS_EX_TOPMOST,
                              MCIWND_WINDOW_CLASS,
                              wszScreenSaverTitle,
                              style,
                              xSize/2-150,
                              ySize/2-120,
                              0,
                              0,
                              hwnd,
                              NULL,
                              hMainInstance,
                              NULL
                              );
         if(hwndFloat != NULL){
           MCIWndOpen(hwndFloat,moviename,0);
           MCIWndPlay(hwndFloat);
           GetWindowRect(hwndFloat,&rc);
           wxSize=rc.right-rc.left;
           wySize=rc.bottom-rc.top;
         }
       }
     }
     SetWindowPos(hwndFloat,HWND_TOPMOST,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE);
     idTimer=1;
     if(random)SetTimer(hwnd,idTimer,10000,0);
     else      SetTimer(hwnd,idTimer,15,0);
     ShowWindow(hwndFloat,SW_SHOW);
     break;
   case MCIWNDM_NOTIFYMODE:
     if((HWND)wParam == hwndFloat && lParam == MCI_MODE_STOP){
       if(random && IsWindow(hwndFloat)){
         long curX,curY;
         curX=((long)rand()*((long)xSize-(long)wxSize))/32767L;
         curY=((long)rand()*((long)ySize-(long)wySize))/32767L;
         SetWindowPos(hwndFloat, HWND_TOPMOST, curX, curY, 0, 0, SWP_NOSIZE);
       }
       else TimerProc(hwnd);
       MCIWndPlay(hwndFloat);
     }
     break;
   case WM_DESTROY:
     if(idTimer){
       KillTimer(hwnd, idTimer);
       idTimer=0;
     }
     if(IsWindow(hwndFloat))DestroyWindow(hwndFloat);
     for(i=0;i<nbitmaps;i++){
       if(lpgBitmap[i] != NULL)EVI_Free(lpgBitmap[i]);   lpgBitmap[i]=NULL;
       if(hPalette[i] != NULL)DeleteObject(hPalette[i]); hPalette[i]=NULL;
     }
     PostQuitMessage(0);
     break;
     case WM_PAINT:
       hdc=BeginPaint(hwnd,&ps);
       EndPaint(hwnd,&ps);
       break;
     case WM_TIMER:
       if(!random || images)TimerProc(hwnd);
       break;
     default:
       return DefScreenSaverProc(hwnd, message, wParam, lParam);
    }
    return 0;
}
