/* Screen Saver Template */

#include <windows.h>
#include <commdlg.h>
#include <scrnsave.h>
#include <stdio.h>
#include <stdlib.h>

#define EVI_Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define EVI_Free(p)      LocalFree((HLOCAL)p)
#define EVI_Realloc(p,s) LocalReAlloc((HLOCAL)p,s,LMEM_MOVEABLE)
#define PALVERSION  0x300
#define MAXPALETTE  256
#define DibNumColors(bmih) (bmih.biClrUsed == 0 && bmih.biBitCount <= 8 \
                            ? (int)(1 << (int)bmih.biBitCount)          \
                            : (int)bmih.biClrUsed)
#define GEN_STRING_SIZE 64

int      nbitmaps=2,pcount=0,disid=0;
LPSTR    lpgBitmap[2],lpgBitmapBits[2];
HPALETTE hPalette[2];
CHAR     wszScreenSaverTitle[GEN_STRING_SIZE];
CHAR     wszClassName[] = "GenericScreenSaverClass";
UINT     idTimer=0,xSize,ySize,wxSize,wySize;
float    xTransInc=2.0f,yTransInc=2.5f;
HWND     hwndFloat=NULL;

LRESULT WINAPI WndProc(HWND, UINT, WPARAM, LPARAM);

static WORD NumColours(LPBITMAPINFOHEADER lpbi){
 WORD bits;
 if (lpbi->biClrUsed != 0)return (WORD)lpbi->biClrUsed;
 bits = lpbi->biBitCount;
 switch (bits){
  case 1:    return 2;
  case 4:    return 16;
  case 8:    return 256;
  default:   return 0; /* A 24 bitcount DIB has no color table */
 }
}

static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi){
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE       red;
 BYTE       green;
 BYTE       blue;
 int        i;
 RGBQUAD    *pRgb;
 if(lpbi == NULL)return NULL;
 if(lpbi->biSize != sizeof(BITMAPINFOHEADER))return NULL;
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 nNumColors = NumColours(lpbi);
 if(nNumColors){
   pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   for (i = 0; i < nNumColors; i++){
       pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
       pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
       pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 else if (lpbi->biBitCount == 24){
   nNumColors = MAXPALETTE;
   pPal=(LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   red = green = blue = 0;
   for (i = 0; i < pPal->palNumEntries; i++){
       pPal->palPalEntry[i].peRed   = red;
       pPal->palPalEntry[i].peGreen = green;
       pPal->palPalEntry[i].peBlue  = blue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
       if (!(red += 32))if (!(green += 32)) blue += 64;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 return hpal;
}

static LPSTR LoadDib(char *temp, HPALETTE *hpalCurrent, LPSTR *lpBits){
 LPSTR lpBitmapHeader;
 LPSTR lpBitmapBits;
 HDC hdc;
 HANDLE hfbm;
 BITMAPFILEHEADER bmfh;
 DWORD dwRead;
 BITMAPINFOHEADER bmih;
 LPBITMAPINFOHEADER lpbi;
 LPBITMAPINFO lpinfo;
 long imagesize,nc;
 *hpalCurrent=NULL;
 hfbm=CreateFile(temp,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_READONLY,(HANDLE)NULL);
 if(hfbm == INVALID_HANDLE_VALUE)return NULL;
 ReadFile(hfbm,&bmfh,sizeof(BITMAPFILEHEADER),&dwRead,(LPOVERLAPPED)NULL);
 ReadFile(hfbm,&bmih,sizeof(BITMAPINFOHEADER),&dwRead,(LPOVERLAPPED)NULL);
 nc=DibNumColors(bmih);
 imagesize=bmfh.bfSize-bmfh.bfOffBits;
 lpBitmapHeader = (LPSTR)EVI_Malloc((long)sizeof(BITMAPINFOHEADER)+
                   nc*sizeof(RGBQUAD)+imagesize);
 if(lpBitmapHeader == NULL)return NULL;
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpinfo=(LPBITMAPINFO)lpBitmapHeader;
 lpbi->biSize=bmih.biSize;
 lpbi->biWidth=bmih.biWidth;
 lpbi->biHeight=bmih.biHeight;
 lpbi->biPlanes=bmih.biPlanes;
 lpbi->biBitCount=bmih.biBitCount;
 lpbi->biCompression=bmih.biCompression;
 lpbi->biSizeImage=bmih.biSizeImage;
 lpbi->biXPelsPerMeter=bmih.biXPelsPerMeter;
 lpbi->biYPelsPerMeter=bmih.biYPelsPerMeter;
 lpbi->biClrUsed=bmih.biClrUsed;
 lpbi->biClrImportant=bmih.biClrImportant;
 if(nc > 0)ReadFile(hfbm,lpinfo->bmiColors,nc*sizeof(RGBQUAD),
                    &dwRead,(LPOVERLAPPED)NULL);
 *hpalCurrent=CreateImagePalette(lpbi);
 lpBitmapBits = ((LPSTR)lpBitmapHeader)
              + (long)sizeof(BITMAPINFOHEADER)+nc*sizeof(RGBQUAD);
 memset(lpBitmapBits,0,imagesize);
 ReadFile(hfbm,lpBitmapBits,imagesize,&dwRead,(LPOVERLAPPED)NULL);
 CloseHandle(hfbm);
 *lpBits=lpBitmapBits;
 return lpBitmapHeader;
}

BOOL WINAPI RegisterDialogClasses(HANDLE hinst){
  return TRUE;
}

BOOL WINAPI ScreenSaverConfigureDialog(HWND hDlg, UINT message,
                           WPARAM wParam, LPARAM lParam){
    switch (message) {
        case WM_INITDIALOG:
            return TRUE;
            break;
        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                case IDOK:
                    EndDialog(hDlg, TRUE);
                    break;
                case IDCANCEL:
                    EndDialog(hDlg, FALSE);
                    break;
                default:
                    break;
            }
            return TRUE;
            break;

        default:
            return 0;
    }
    return 0;
}

BOOL initClass(HANDLE hMainInstance){
 WNDCLASS wclass;
 wclass.style = CS_VREDRAW | CS_HREDRAW;
 wclass.lpfnWndProc = (WNDPROC)WndProc;
 wclass.cbClsExtra = 0;
 wclass.cbWndExtra = 0;
 wclass.hInstance = hMainInstance;
 wclass.hIcon = NULL;
 wclass.hCursor = NULL;
 wclass.hbrBackground = NULL;
 wclass.lpszMenuName = (LPSTR)NULL;
 wclass.lpszClassName = (LPSTR)wszClassName;
 return RegisterClass(&wclass);
}

BOOL initInstance(HANDLE hMainInstance,HWND parent){
 HDC hdc;
 DWORD style;
 ySize=GetSystemMetrics(SM_CYSCREEN);
 xSize=GetSystemMetrics(SM_CXSCREEN);
 style = WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
 hwndFloat = CreateWindowEx(WS_EX_TOPMOST,
                            wszClassName,
                            wszScreenSaverTitle,
                            style,
                            0,
                            0,
                            wxSize,
                            wySize,
                            parent,
                            (HMENU)101,
                            hMainInstance,
                            NULL
                            );
 if(hwndFloat == NULL)return FALSE;
 return TRUE;
}

void vMoveBuffer(int curX, int curY, int lastX, int lastY,
                int pxsize, int pysize){
 SetWindowPos(hwndFloat, HWND_TOPMOST, curX, curY, 0, 0, SWP_NOSIZE |
                SWP_NOCOPYBITS);
}

BOOL vShowBuffer(HWND hwnd){
 HDC hdc;
 static float xtrans = 0.0f;
 static float ytrans = 0.0f;
 static int lastXt = 0;
 static int lastYt = 0;
 static int bInited = FALSE;
 static int interBounce = 0;
 int bounce = FALSE;
 int xt, yt;
 if(!bInited){
   bInited = TRUE;
   srand(clock() & 0xffff);
   rand();
   xtrans = ((float)rand() / (float)RAND_MAX) * (float)(xSize - wxSize);
   ytrans = ((float)rand() / (float)RAND_MAX) * (float)(ySize - wySize);
   lastXt = (int) xtrans;
   lastYt = (int) ytrans;
 }
 xt = (int)xtrans;
 yt = (int)ytrans;
 vMoveBuffer(xt, yt, lastXt, lastYt, wxSize, wySize);
 lastXt = xt;
 lastYt = yt;
 xtrans += xTransInc;
 ytrans += yTransInc;
 if ((xtrans + wxSize) > xSize) {
   xtrans = (float) (xSize - wxSize);
   xTransInc = -xTransInc;
   bounce = TRUE;
 }
 else if (xtrans < 0.0f) {
   xtrans = 0.0f;
   xTransInc = -xTransInc;
   bounce = TRUE;
 }
 if ((ytrans + wySize) > ySize) {
   ytrans = (float) (ySize - wySize);
   yTransInc = -yTransInc;
   bounce = TRUE;
 }
 else if (ytrans < 0.0f) {
   ytrans = 0.0f;
   yTransInc = -yTransInc;
   bounce = TRUE;
 }
 interBounce++;
 if (bounce) {
   if (interBounce < 10)bounce = FALSE;
   else            interBounce = 0;
 }
 return bounce;
}

static void AppPaint(HWND hWnd,HDC hDC,int id){
 HPALETTE hpalT;
 if(hPalette[id] != NULL){
   hpalT = SelectPalette(hDC,hPalette[id],FALSE);
   RealizePalette(hDC);
   if(lpgBitmap[id] != NULL)SetDIBitsToDevice(hDC,
                        0,0,
                        (WORD)((LPBITMAPINFOHEADER)lpgBitmap[id])->biWidth,
                        (WORD)((LPBITMAPINFOHEADER)lpgBitmap[id])->biHeight,
                        0,0,
                        0,
                        (WORD)((LPBITMAPINFOHEADER)lpgBitmap[id])->biHeight,
                        lpgBitmapBits[id],
                        (LPBITMAPINFO)lpgBitmap[id],
                        DIB_RGB_COLORS);
   SelectPalette(hDC,hpalT,FALSE);
 }
}

LRESULT WINAPI WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam){
  HDC hdc;
  PAINTSTRUCT ps;
  int retVal;
  switch(message){
    case WM_PAINT:
      hdc=BeginPaint(hwnd,&ps);
      AppPaint(hwnd,hdc,disid);
      EndPaint(hwnd, &ps);
      break;
    case WM_SYSCOMMAND:
    case WM_SETCURSOR:
    case WM_NCACTIVATE:
    case WM_ACTIVATE:
    case WM_ACTIVATEAPP:
    case WM_MOUSEMOVE:
    case WM_LBUTTONDOWN:
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_KEYDOWN:
    case WM_SYSKEYDOWN:
      retVal = DefScreenSaverProc(hwnd, message, wParam, lParam);
      DefScreenSaverProc(hMainWindow, message, wParam, lParam);
      return retVal;
    default:
      return DefWindowProc(hwnd, message, wParam, lParam);
  }
  return 0;
}

static void TimerProc(HWND hwnd){
  static BOOL busy=FALSE;
  if(busy)return;
  busy=TRUE;
  if(pcount > 100){
    disid++;
    if(disid >= nbitmaps)disid=0;
    pcount=0;
  }
  else pcount++;
  vShowBuffer(hwnd);
  busy=FALSE;
}

LRESULT WINAPI ScreenSaverProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam){
 int i;
 HDC hdc;
 PAINTSTRUCT ps;
 LPBITMAPINFOHEADER hb;
 static BOOL bIni=FALSE;
 switch (message){
   case WM_CREATE:
     for(i=0;i<nbitmaps;i++){lpgBitmap[i]=NULL; hPalette[i]=NULL;}
     LoadString(hMainInstance,IDS_DESCRIPTION,wszScreenSaverTitle,
                GEN_STRING_SIZE);
     lpgBitmap[0]=LoadDib("c:\\evi\\gif\\amber.bmp",
                          &hPalette[0],&lpgBitmapBits[0]);
     lpgBitmap[1]=LoadDib("c:\\evi\\gif\\demo1.bmp",
                          &hPalette[1],&lpgBitmapBits[1]);
     hb=(LPBITMAPINFOHEADER)lpgBitmap[0];
     wxSize=hb->biWidth;
     wySize=hb->biHeight;
     initClass(hMainInstance);
     initInstance(hMainInstance,hwnd);
     UpdateWindow(hwndFloat);
     SetWindowPos(hwndFloat,HWND_TOPMOST,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE);
     idTimer=1;
     SetTimer(hwnd,idTimer,16,0);
     ShowWindow(hwndFloat,SW_SHOW);
     break;
   case WM_DESTROY:
     if(idTimer){
       KillTimer(hwnd, idTimer);
       idTimer=0;
     }
     if(IsWindow(hwndFloat))DestroyWindow(hwndFloat);
MessageBeep(MB_OK);
     for(i=0;i<nbitmaps;i++){
       if(lpgBitmap[i] != NULL)EVI_Free(lpgBitmap[i]);   lpgBitmap[i]=NULL;
       if(hPalette[i] != NULL)DeleteObject(hPalette[i]); hPalette[i]=NULL;
     }
     PostQuitMessage(0);
     break;
     case WM_PAINT:
       hdc=BeginPaint(hwnd,&ps);
       if(!bIni){
       }
       EndPaint(hwnd,&ps);
       break;
     case WM_TIMER:
       TimerProc(hwnd);
       break;
     default:
       return DefScreenSaverProc(hwnd, message, wParam, lParam);
    }
    return 0;
}
