/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* saver2.c */

#include <windows.h>
#include <commdlg.h>
#include <scrnsave.h>
#include <GL\gl.h>
#include <GL\glu.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "saver2.h"
#include "..\sfxview\sfxview.h"

#define FI_3DS           0
#define FI_SFX           1

extern int LoadSfxObject(char *);
extern void FreeObject();
extern void ListAdjFaces(void);
extern void StripFaces(void);
extern void FixUpOrientation(void);
extern void Make3dListS(void);
extern void DrawWireFrame(void);

LRESULT WINAPI WndProc(HWND, UINT, WPARAM, LPARAM);
void TimerProc(HWND);
void makeCheckeredImageMap(void);

CHAR    wszClassName[] = "RSF:GlScreenSaverClass";
CHAR    CurrentDirFile[256];
HWND    hwndOpenGL;
HGLRC   hRC = NULL;
HPALETTE ghpalOld, ghPalette = NULL;
UINT    idTimer = 0;
VERTEX  *MainVp=NULL;
FACE    *MainFp=NULL;
MAP     *MapsFp=NULL;
WFRAME  w_frame={0,0,NULL,NULL};
long    Nvert=0,Nface=0,NvertGlue=0,Nmap=0;
int     listID=0;
BOOL    Mapped=FALSE,Cmaterial=TRUE,smooth_shading=TRUE,
        ffastmapping=TRUE,accelerated=TRUE,LightBackground=TRUE,
        ffog=FALSE;
GLfloat ObjScale=1.0;
GLubyte checkImage[checkImageWidth][checkImageHeight][3];

static int       saver_type=1;
static RECT      oldrect;
static float     xTransInc = 2.0f;
static float     yTransInc = 2.5f;
static int       xSize;
static int       ySize;
static int       wxSize;
static int       wySize;
static int       updates = 0;
static GLfloat   up_angle=20.0,round_angle=0.0;
static GLfloat   gbRed=0.0,gbGreen=0.0,gbBlue=0.0;
static GLfloat   near_plane = 1.5,far_plane = 8.5;

static int identify_file_format(char *filename){
 char cf[256];
 int ff=FI_SFX;
 strcpy(cf,filename);
 if     (strstr(filename,".3DS") || strstr(filename,".3ds"))ff=FI_3DS;
 else if(strstr(filename,".SCE") || strstr(filename,".sce"))ff=FI_SFX;
 return ff;
}

static PSTR FileInPath(PSTR pstrPath){
 PSTR pstr;
 pstr = pstrPath + strlen(pstrPath);
 while (pstr > pstrPath) {
     pstr = (AnsiPrev(pstrPath, pstr));
     if (*pstr == '\\' || *pstr == ':' || *pstr == '/') {
         pstr = (AnsiNext(pstr));
         break;
     }
 }
 return pstr;
}

static int SelectFileName(int save, char *szfile, char *szdir,
                          char *szTitle, char *szfilter, HWND parent){
 int i;
 OPENFILENAME ofn;
 char szFilter[256],szdummy[256];
 szdummy[0]='\0';
 if(szfilter == NULL)return 0;
 else{
   strcpy(szFilter,szfilter);
   i=0; while(szFilter[i] != '\0'){
     if(szFilter[i] == '|')szFilter[i]='\0'; i++;
   }
 }
 memset(&ofn,0,sizeof(OPENFILENAME));
 ofn.lStructSize=sizeof(OPENFILENAME);
 ofn.hwndOwner=parent;
 ofn.lpstrFilter=szFilter;
 ofn.nFilterIndex=1;
 ofn.lpstrFile=szfile;              /* will contain file + path */
 ofn.nMaxFile=256;
 ofn.lpstrFileTitle=szdummy;         /* file name only */
 ofn.nMaxFileTitle=256;
 ofn.lpstrInitialDir=szdir;
 ofn.lpstrTitle=szTitle;
 if(save){
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;
   i=(int)(GetSaveFileName(&ofn));
 }
 else{
  ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
   i=(int)(GetOpenFileName(&ofn));
 }
 if(i){
   strcpy(szdummy,szfile);
   *AnsiPrev(szdummy,FileInPath(szdummy)) = '\0';
   strcpy(szdir,szdummy);
 }
 return i;
}

static BOOL WriteRegistryItem(char *keyname, char *keyvaluename,
                              char *keyvalue){
 HKEY hkey;
 DWORD result;
 if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,keyname,0,NULL,REG_OPTION_NON_VOLATILE,
                   KEY_ALL_ACCESS,NULL,&hkey,&result) != ERROR_SUCCESS)
   return FALSE;
 if     (result == REG_CREATED_NEW_KEY    ){;}
 else if(result == REG_OPENED_EXISTING_KEY){;}
 else                                      {;}
 if(keyvalue != NULL){
   if(RegSetValueEx(hkey,keyvaluename,0,REG_SZ,keyvalue,strlen(keyvalue)+1)
     != ERROR_SUCCESS)return FALSE;
 }
 RegCloseKey(hkey);
 return TRUE;
}

static BOOL ReadRegistryItem(char *keyname, char *keyvaluename,
                             LPBYTE value, DWORD size){
 HKEY hkey;
 DWORD result,type;
 if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,keyname,0,NULL,REG_OPTION_NON_VOLATILE,
                   KEY_ALL_ACCESS,NULL,&hkey,&result) != ERROR_SUCCESS)
   return FALSE;
 if     (result == REG_CREATED_NEW_KEY    ){;}
 else if(result == REG_OPENED_EXISTING_KEY){
   if(RegQueryValueEx(hkey,keyvaluename,NULL,&type,value,&size)
                      != ERROR_SUCCESS){
     RegCloseKey(hkey);
     return FALSE;
   }
   RegCloseKey(hkey);
   return TRUE;
 }
 else {;}
 RegCloseKey(hkey);
 return FALSE;
}

static char keyname[]="System\\CurrentControlSet\\Control\\ProductOptions\\RSFSaver2";
static char modelkey[]="model_name";
static char typename[]="type";
static char mapkey[]="mapping";
static char scalekey[]="scaling";
static char sizekey[]="sizing";

BOOL WINAPI RegisterDialogClasses(HANDLE hinst){
 return TRUE;
}

BOOL WINAPI ScreenSaverConfigureDialog(HWND hDlg, UINT message,
                                WPARAM wParam, LPARAM lParam){
 int wTmp;
 static char modelname[256],temp[16];
 static char szFileName[256]=" ",szDirName[256]=" ";
 switch (message) {
   case WM_INITDIALOG:
     if(!ReadRegistryItem(keyname,modelkey,modelname,256))
       strcpy(modelname,"\\*.sce");
     SetDlgItemText(hDlg,DLG_SAVER_NAME,modelname);
     if(ReadRegistryItem(keyname,typename,temp,16)){
       if(temp[0] == '1')saver_type=1;
       if(temp[0] == '2')saver_type=2;
       if(temp[0] == '3')saver_type=3;
     }
     if(ReadRegistryItem(keyname,mapkey,temp,16)){
       if(temp[0] == 'Y')Mapped=TRUE;
     }
     if(saver_type == 1)
       SendDlgItemMessage(hDlg,DLG_SAVER_TYPE1,BM_SETCHECK,TRUE,0);
     else if(saver_type == 2)
       SendDlgItemMessage(hDlg,DLG_SAVER_TYPE2,BM_SETCHECK,TRUE,0);
     else if(saver_type == 3)
       SendDlgItemMessage(hDlg,DLG_SAVER_TYPE3,BM_SETCHECK,TRUE,0);
     if(Mapped)
       SendDlgItemMessage(hDlg,DLG_SAVER_MAP_ON,BM_SETCHECK,TRUE,0);
     SetScrollRange(GetDlgItem(hDlg,DLG_SAVER_SCALE),SB_CTL,0,200,FALSE);
     if(ReadRegistryItem(keyname,scalekey,temp,16))wTmp=atoi(temp);
     else wTmp=150;
     SetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SCALE),SB_CTL,wTmp,TRUE);
     SetScrollRange(GetDlgItem(hDlg,DLG_SAVER_SIZE),SB_CTL,0,200,FALSE);
     if(ReadRegistryItem(keyname,sizekey,temp,16))wTmp=atoi(temp);
     else wTmp=160;
     SetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SIZE),SB_CTL,wTmp,TRUE);
     break;
   case WM_HSCROLL:
     if(lParam == (LPARAM) GetDlgItem(hDlg,DLG_SAVER_SCALE)){
       wTmp = GetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SCALE),SB_CTL);
       switch(LOWORD(wParam)){
         case SB_PAGEDOWN:
           wTmp += 10;
         case SB_LINEDOWN:
           wTmp += 1;
           wTmp = min(200, wTmp);
           break;
         case SB_PAGEUP:
           wTmp -= 10;
         case SB_LINEUP:
           wTmp -= 1;
           wTmp = max(0, (int)wTmp);
           break;
         case SB_THUMBPOSITION:
           wTmp = HIWORD(wParam);
           break;
         default:
           break;
       }
       SetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SCALE),SB_CTL,wTmp,TRUE);
     }
     if(lParam == (LPARAM) GetDlgItem(hDlg,DLG_SAVER_SIZE)){
       wTmp = GetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SIZE),SB_CTL);
       switch(LOWORD(wParam)){
         case SB_PAGEDOWN:
           wTmp += 10;
         case SB_LINEDOWN:
           wTmp += 1;
           wTmp = min(200, wTmp);
           break;
         case SB_PAGEUP:
           wTmp -= 10;
         case SB_LINEUP:
           wTmp -= 1;
           wTmp = max(0, (int)wTmp);
           break;
         case SB_THUMBPOSITION:
           wTmp = HIWORD(wParam);
           break;
         default:
           break;
       }
       SetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SIZE),SB_CTL,wTmp,TRUE);
     }
     break;
   case WM_COMMAND:
     switch (LOWORD(wParam)) {
       case DLG_SAVER_NAME_SET:
         strcpy(szFileName,modelname);
         strcpy(szDirName,modelname); *FileInPath(szDirName)='\0';
         if(SelectFileName(0,szFileName,szDirName,
            "Select Model","All model formats|*.sce;*.3ds|",hDlg) == TRUE){
            strcpy(modelname,szFileName);
            SetDlgItemText(hDlg,DLG_SAVER_NAME,modelname);
          }
         break;
       case IDOK:
         if(SendDlgItemMessage(hDlg,DLG_SAVER_TYPE1,BM_GETCHECK,0,0))
           saver_type=1;
         if(SendDlgItemMessage(hDlg,DLG_SAVER_TYPE2,BM_GETCHECK,0,0))
           saver_type=2;
         if(SendDlgItemMessage(hDlg,DLG_SAVER_TYPE3,BM_GETCHECK,0,0))
           saver_type=3;
         WriteRegistryItem(keyname,modelkey,modelname);
         if(SendDlgItemMessage(hDlg,DLG_SAVER_MAP_ON,BM_GETCHECK,0,0))
              Mapped=TRUE;
         else Mapped=FALSE;
         if(Mapped)WriteRegistryItem(keyname,mapkey,"Y");
         else      WriteRegistryItem(keyname,mapkey,"N");
         if(saver_type == 1)WriteRegistryItem(keyname,typename,"1");
         if(saver_type == 2)WriteRegistryItem(keyname,typename,"2");
         if(saver_type == 3)WriteRegistryItem(keyname,typename,"3");
         wTmp = GetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SCALE),SB_CTL);
         sprintf(temp,"%ld",wTmp); WriteRegistryItem(keyname,scalekey,temp);
         wTmp = GetScrollPos(GetDlgItem(hDlg,DLG_SAVER_SIZE),SB_CTL);
         sprintf(temp,"%ld",wTmp); WriteRegistryItem(keyname,sizekey,temp);
         EndDialog(hDlg, TRUE);
         break;
       case IDCANCEL:
         EndDialog(hDlg, FALSE);
         break;
       default:
       break;
     }
     return TRUE;
   default:
     return 0;
 }
 return 0;
}

BOOL initClass(HANDLE hMainInstance){
 WNDCLASS wclass;
 wclass.style = CS_VREDRAW | CS_HREDRAW;
 wclass.lpfnWndProc = (WNDPROC)WndProc;
 wclass.cbClsExtra = 0;
 wclass.cbWndExtra = 0;
 wclass.hInstance = hMainInstance;
 wclass.hIcon = NULL;
 wclass.hCursor = NULL;
 wclass.hbrBackground = NULL;
 wclass.lpszMenuName = (LPSTR)NULL;
 wclass.lpszClassName = (LPSTR)wszClassName;
 return RegisterClass(&wclass);
}

BOOL initInstance(HANDLE hMainInstance, HWND hwnd, int nCmdShow){
 char temp[16];
 DWORD style;
 ySize=GetSystemMetrics(SM_CYSCREEN);
 xSize=GetSystemMetrics(SM_CXSCREEN);
 if(ReadRegistryItem(keyname,sizekey,temp,16)){
   wxSize=atoi(temp);
   wxSize=(xSize * wxSize)/400;
   wySize=(wxSize*3)/4;
 }
 else{
   wxSize=320; wySize=240;
 }
 style = WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
 hwndOpenGL = CreateWindowEx(WS_EX_TOPMOST,
                                 wszClassName,
                                 " ",
                                 style,
                                 0,                  // horizontal pos
                                 0,                  // vertical pos
                                 wxSize,             // width
                                 wySize,             // height
                                 hwnd, // NULL,
                                 NULL,               // menu
                                 hMainInstance,
                                 NULL
                                );
 if (!hwndOpenGL)return FALSE;
 ShowWindow(hwndOpenGL, SW_SHOW);
 UpdateWindow(hwndOpenGL);
 return TRUE;
}

static unsigned char threeto8[8] = {
  0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

static unsigned char twoto8[4] = {
  0, 0x55, 0xaa, 0xff
};

static unsigned char oneto8[2] = {
  0, 255
};

static int defaultOverride[13] = {
  0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
  { 0,   0,   0,    0 },
  { 0x80,0,   0,    0 },
  { 0,   0x80,0,    0 },
  { 0x80,0x80,0,    0 },
  { 0,   0,   0x80, 0 },
  { 0x80,0,   0x80, 0 },
  { 0,   0x80,0x80, 0 },
  { 0xC0,0xC0,0xC0, 0 },

  { 192, 220, 192,  0 },
  { 166, 202, 240,  0 },
  { 255, 251, 240,  0 },
  { 160, 160, 164,  0 },

  { 0x80,0x80,0x80, 0 },
  { 0xFF,0,   0,    0 },
  { 0,   0xFF,0,    0 },
  { 0xFF,0xFF,0,    0 },
  { 0,   0,   0xFF, 0 },
  { 0xFF,0,   0xFF, 0 },
  { 0,   0xFF,0xFF, 0 },
  { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift){
  unsigned char val;
  val = (unsigned char) (i >> shift);
  switch (nbits) {
  case 1:
    val &= 0x1;
    return oneto8[val];
  case 2:
    val &= 0x3;
    return twoto8[val];
  case 3:
    val &= 0x7;
    return threeto8[val];
  default:
    return 0;
  }
}

static void CreateRGBPalette(HDC hDC){
  PIXELFORMATDESCRIPTOR pfd;
  LOGPALETTE *pPal;
  int n, i;
  n = GetPixelFormat(hDC);
  DescribePixelFormat(hDC, n, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
  if (pfd.dwFlags & PFD_NEED_PALETTE) {
    n = 1 << pfd.cColorBits;
    pPal = (PLOGPALETTE)LocalAlloc(LMEM_FIXED, sizeof(LOGPALETTE) +
                                   n * sizeof(PALETTEENTRY));
    pPal->palVersion = 0x300;
    pPal->palNumEntries = n;
    for (i=0; i<n; i++) {
      pPal->palPalEntry[i].peRed = ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
      pPal->palPalEntry[i].peGreen = ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
      pPal->palPalEntry[i].peBlue = ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
      pPal->palPalEntry[i].peFlags = 0;
    }
    if((pfd.cColorBits == 8)                           &&
       (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
       (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
       (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)) {
      for(i = 1 ; i <= 12 ; i++)
        pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
    }
    ghPalette = CreatePalette(pPal);
    LocalFree(pPal);
    ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
    n = RealizePalette(hDC);
  }
}

static BOOL bSetupPixelFormat(HDC hDC){
  static PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),    // size of this pfd
      1,                              // version number
      PFD_DRAW_TO_WINDOW |            // support window
      PFD_SUPPORT_OPENGL |            // support OpenGL
      PFD_DOUBLEBUFFER,               // double buffered
      PFD_TYPE_RGBA,                  // RGBA type
      24,                             // 24-bit color depth
      0, 0, 0, 0, 0, 0,               // color bits ignored
      0,                              // no alpha buffer
      0,                              // shift bit ignored
      0,                              // no accumulation buffer
      0, 0, 0, 0,                     // accum bits ignored
      32,                             // 32-bit z-buffer
      0,                              // no stencil buffer
      0,                              // no auxiliary buffer
      PFD_MAIN_PLANE,                 // main layer
      0,                              // reserved
      0, 0, 0                         // layer masks ignored
      };
  int pixelformat;
  if ( (pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0 ) {
    MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
    return FALSE;
  }
  if (SetPixelFormat(hDC, pixelformat, &pfd) == FALSE) {
    MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
    return FALSE;
  }
  CreateRGBPalette(hDC);
  return TRUE;
}

static GLvoid initialize(HWND hWnd){
  GLfloat   aspect;
  GLfloat d[]={1.0,1.0,1.0,1.0};
  GLfloat a[]={0.1,0.1,0.1,1.0};
  GLfloat s[]={1.0,1.0,1.0,1.0};
  GLfloat p[]={0.0,0.0,1.0,1.0};  /* (x,y,z,w  w=1 => positional light */
  GetClientRect(hWnd, &oldrect);
  glClearColor(gbRed,gbGreen,gbBlue,1.0);
  glMatrixMode(GL_PROJECTION);
  aspect = (GLfloat) oldrect.right / oldrect.bottom;
  gluPerspective( 45.0, aspect, near_plane, far_plane);
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  glClearDepth(1.0);
  glEnable(GL_DEPTH_TEST);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0,GL_DIFFUSE,d);
  glLightfv(GL_LIGHT0,GL_AMBIENT,a);
  glLightfv(GL_LIGHT0,GL_SPECULAR,s);
  glLightfv(GL_LIGHT0,GL_POSITION,p);
  glLightf(GL_LIGHT0,GL_LINEAR_ATTENUATION,0.15);
  glLightf(GL_LIGHT0,GL_QUADRATIC_ATTENUATION,0.05);
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
}

void vMoveBuffer(int curX, int curY,
                int lastX, int lastY,
                int pxsize, int pysize){
 SetWindowPos(hwndOpenGL, HWND_TOPMOST, curX, curY, 0, 0, SWP_NOSIZE |
                 SWP_NOCOPYBITS);
}

BOOL vShowBuffer(HWND hwnd){
    HDC hdc;
    static float xtrans = 0.0f;
    static float ytrans = 0.0f;
    static int lastXt = 0;
    static int lastYt = 0;
    static int bInited = FALSE;
    static int interBounce = 0;
    int bounce = FALSE;
    int xt, yt;
    if (!bInited) {
        bInited = TRUE;
        srand(clock() & 0xffff);
        rand();
        xtrans = ((float)rand() / (float)RAND_MAX) * (float)(xSize - wxSize);
        ytrans = ((float)rand() / (float)RAND_MAX) * (float)(ySize - wySize);
        lastXt = (int) xtrans;
        lastYt = (int) ytrans;
    }
    xt = (int)xtrans;
    yt = (int)ytrans;
    vMoveBuffer(xt, yt, lastXt, lastYt, wxSize, wySize);
    lastXt = xt;
    lastYt = yt;
    xtrans += xTransInc;
    ytrans += yTransInc;
    if ((xtrans + wxSize) > xSize) {
        xtrans = (float) (xSize - wxSize);
        xTransInc = -xTransInc;
        bounce = TRUE;
    } else if (xtrans < 0.0f) {
        xtrans = 0.0f;
        xTransInc = -xTransInc;
        bounce = TRUE;
    }
    if ((ytrans + wySize) > ySize) {
        ytrans = (float) (ySize - wySize);
        yTransInc = -yTransInc;
        bounce = TRUE;
    } else if (ytrans < 0.0f) {
        ytrans = 0.0f;
        yTransInc = -yTransInc;
        bounce = TRUE;
    }
    glFinish();
    hdc = wglGetCurrentDC();
    SwapBuffers(hdc);
    interBounce++;
    if (bounce) {
        if (interBounce < 10)
            bounce = FALSE;
        else
            interBounce = 0;
    }
    return bounce;
}

void TimerProc(HWND hwnd){
 static int busy = FALSE;
 if (busy) return;
 busy = TRUE;
 updates++;
 if(saver_type == 1){
   round_angle += 10.0;
   if(round_angle > 360)round_angle=0;
 }
 if(saver_type == 2){
   round_angle -= 10.0;
   if(round_angle < -359.0)round_angle=0.0;
   up_angle += 5.0;
   if(up_angle > 359.0)up_angle=0.0;
 }
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 glShadeModel(GL_SMOOTH);
 glPushMatrix();
 glLoadIdentity();
 glTranslatef(0.0, 0.0, -5.0);
 glRotatef(up_angle,1.0,0.0,0.0);
 glRotatef(round_angle+180.0,0.0,1.0,0.0);
 if(listID > 0){
   glEnable(GL_LIGHTING);
   glCallList(listID);
   glDisable(GL_LIGHTING);
 }
 glPopMatrix();
 vShowBuffer(hwnd);
 busy = FALSE;
}

LRESULT WINAPI WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam){
 HDC hDC;
 PAINTSTRUCT ps;
 int retVal,fid;
 char temp[16];
 GLfloat scale;
 switch (message) {
   case WM_CREATE:
     hDC = GetDC(hwnd);
     bSetupPixelFormat(hDC);
     hRC = wglCreateContext( hDC );
     wglMakeCurrent( hDC, hRC );
     initialize(hwnd);
     FreeObject();
     if(ReadRegistryItem(keyname,modelkey,CurrentDirFile,256)){
       char temp[16];
       if(ReadRegistryItem(keyname,mapkey,temp,16)){
         if(temp[0] == 'Y'){ Mapped=TRUE; accelerated=FALSE; }
       }
       if(ReadRegistryItem(keyname,scalekey,temp,16)){
         scale=(GLfloat)atoi(temp);
         if(scale < 150)ObjScale = 0.1 + 0.9*scale/150.0;
         else           ObjScale = 1.0 + (scale - 150.0)/50.0;
       }
       else ObjScale=1.0;
       fid=identify_file_format(CurrentDirFile);
       if(fid == FI_3DS)Load3dsObject(CurrentDirFile);
       else             LoadSfxObject(CurrentDirFile);
       ListAdjFaces();
       FixUpOrientation();
       StripFaces();
       Make3dListS();
       if(ReadRegistryItem(keyname,typename,temp,16)){
         if(temp[0] == '1')saver_type=1;
         if(temp[0] == '2')saver_type=2;
         if(temp[0] == '3')saver_type=3;
       }
     }
     break;
   case WM_DESTROY:
     FreeObject();
     hRC = wglGetCurrentContext();
     hDC = wglGetCurrentDC();
     wglMakeCurrent(NULL, NULL);
     if (hRC != NULL)wglDeleteContext(hRC);
     if (hDC != NULL)ReleaseDC(hwnd,hDC);
     PostQuitMessage(0);
     break;
   case WM_PAINT:
     hDC = BeginPaint(hwnd,&ps);
     EndPaint(hwnd, &ps);
     break;
   case WM_SYSCOMMAND:
   case WM_SETCURSOR:
   case WM_NCACTIVATE:
   case WM_ACTIVATE:
   case WM_ACTIVATEAPP:
   case WM_MOUSEMOVE:
   case WM_LBUTTONDOWN:
   case WM_MBUTTONDOWN:
   case WM_RBUTTONDOWN:
   case WM_KEYDOWN:
   case WM_SYSKEYDOWN:
     retVal = DefScreenSaverProc(hwnd, message, wParam, lParam);
     DefScreenSaverProc(hMainWindow, message, wParam, lParam);
     return retVal;
   default:
     return DefWindowProc(hwnd, message, wParam, lParam);
 }
 return 0;
}

LRESULT WINAPI ScreenSaverProc(HWND hwnd, UINT message, WPARAM wParam,
                               LPARAM lParam){
    HDC hdc;
    PAINTSTRUCT ps;
    static BOOL bInited = FALSE;
    switch (message) {
        case WM_CREATE:
            break;
        case WM_DESTROY:
            if (idTimer) {
                KillTimer(hwnd, idTimer);
                idTimer = 0;
            }
            DestroyWindow(hwndOpenGL);
            PostQuitMessage(0);
            break;
        case WM_PAINT:
            hdc = BeginPaint(hwnd, &ps);
            if (!bInited) {
                bInited = TRUE;
                if ( !initClass(hMainInstance) ||
                     !initInstance(hMainInstance, hwnd, SW_SHOW) ){
                        MessageBox(NULL, "Bad",NULL, MB_OK);
                    return FALSE;
                }
                if (ghPalette) {
                    SelectPalette(hdc, ghPalette, FALSE);
                    RealizePalette(hdc);
                }
                idTimer = 1;
                SetTimer(hwnd, idTimer, 5, 0);
                SetWindowPos(hwndOpenGL, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE);
            }
            EndPaint(hwnd, &ps);
            break;
        case WM_TIMER:
            TimerProc(hwndOpenGL);
            break;
        default:
            return DefScreenSaverProc(hwnd, message, wParam, lParam);
    }
    return 0;
}

void TryCurrentDirect(char *name){
}

void makeCheckeredImageMap(void){
 int i, j, c;
 for (i = 0; i < checkImageWidth; i++) {
   for (j = 0; j < checkImageHeight; j++) {
     c = ((((i&0x8)==0)^((j&0x8))==0))*255;
     checkImage[i][j][0] = (GLubyte) c;
     checkImage[i][j][1] = (GLubyte) c;
     checkImage[i][j][2] = (GLubyte) c;
   }
 }
 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glPixelStorei(GL_PACK_ALIGNMENT,1);
 glTexImage2D(GL_TEXTURE_2D, 0, 3, checkImageWidth,
              checkImageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
              &checkImage[0][0][0]);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
 if(ffastmapping){
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);
 }
 else{
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
 }
 glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
 glTexGenfv(GL_S,GL_SPHERE_MAP,0);
 glTexGenfv(GL_T,GL_SPHERE_MAP,0);
}
