/* --
OpenFX version 1.0 - Modelling, Animation and Rendering Package
Copyright (C) 2000 OpenFX Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the OpenFX development team via elecronic mail
at core@openfx.org, or visit our website at http://openfx.org for
further information and support details.
-- */

/* file SFXLOAD.C  || */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>

#include "sfxview.h"
#include "resource.h"

#define PLANE           0
#define CYLINDER        1
#define PLANE_MOZIAC    2
#define CYLINDER_MOZIAC 3
#define MAP_BY_VERTEX   4
#define STILBRUSH       0
#define ANIMBRUSH       1
#define BUMPBRUSH       2
#define REFLBRUSH       3
#define PSETBRUSH       4

extern VERTEX *MainVp;
extern FACE   *MainFp;
extern MAP    *MapsFp;
extern WFRAME w_frame;
extern long   Nvert,Nface,NvertGlue,Nmap;

extern BOOL Normalize(point, point, point, GLfloat *);

void ListAdjFaces(void);
void StripFaces(void);
void FixUpOrientation(void);

static unsigned char getlon(long *);
static unsigned char getsht(short *);
static unsigned char getusht(unsigned short *);
static void outsht(short);
static void outusht(unsigned short);
static void outlng(long);
static void getchunk(long clen);

#define BRUSHBIT   0x40
#define SMOOTHBIT  0x80
#define TILEBIT    0x20;

 FILE *ff;

#if 0
static void PrintAdjFaces(void){
 int i;
 if(ff != NULL){
//   for(i=0;i<Nvert;i++){
//     fprintf(ff,"V %4ld     %4ld %4ld %4ld\n",i,(MainVp+i)->p[0]/1024,
//     (MainVp+i)->p[1]/1024,(MainVp+i)->p[2]/1024);
//   }
   for(i=0;i<Nface;i++){
     fprintf(ff,"F %3ld =[%3ld %3ld %3ld] Adj [%3ld %3ld %3ld]\n",i,
       (MainFp+i)->V[0],(MainFp+i)->V[1],(MainFp+i)->V[2],
       (MainFp+i)->A[0],(MainFp+i)->A[1],(MainFp+i)->A[2]);
   }
 }
}
static void xListAdjFaces(void){
 int i,j,k,iv0,iv1,iv2;
 FACE *fp1,*fp2;
 for(i=0;i<Nface;i++){
   fp1=(MainFp+i); j=i+1; iv0=fp1->V[0]; iv1=fp1->V[1]; iv2=fp1->V[2];
   for(k=0,j=0;j<Nface;j++){
     if(i == j)continue;
     fp2=(MainFp+j);
     if((fp2->V[0] == iv0 && fp2->V[1] == iv1) ||
        (fp2->V[1] == iv0 && fp2->V[0] == iv1) ||
        (fp2->V[1] == iv0 && fp2->V[2] == iv1) ||
        (fp2->V[2] == iv0 && fp2->V[1] == iv1) ||
        (fp2->V[2] == iv0 && fp2->V[0] == iv1) ||
        (fp2->V[0] == iv0 && fp2->V[2] == iv1)){
       k++; fp1->A[0]=j;// if(k == 3)break;
     }
     else if((fp2->V[0] == iv1 && fp2->V[1] == iv2) ||
             (fp2->V[1] == iv1 && fp2->V[0] == iv2) ||
             (fp2->V[1] == iv1 && fp2->V[2] == iv2) ||
             (fp2->V[2] == iv1 && fp2->V[1] == iv2) ||
             (fp2->V[2] == iv1 && fp2->V[0] == iv2) ||
             (fp2->V[0] == iv1 && fp2->V[2] == iv2)){
       k++; fp1->A[1]=j;// if(k == 3)break;
     }
     else if((fp2->V[0] == iv2 && fp2->V[1] == iv0) ||
             (fp2->V[1] == iv2 && fp2->V[0] == iv0) ||
             (fp2->V[1] == iv2 && fp2->V[2] == iv0) ||
             (fp2->V[2] == iv2 && fp2->V[1] == iv0) ||
             (fp2->V[2] == iv2 && fp2->V[0] == iv0) ||
             (fp2->V[0] == iv2 && fp2->V[2] == iv0)){
       k++; fp1->A[2]=j;// if(k == 3)break;
     }
   }
 }
}
#endif

static FILE *fp;
static long newface,newvert,newglue,minX,minY,minZ,maxX,maxY,maxZ;
static char str[10];
static long FaceBase,EdgeBase,VertexBase;

static unsigned char getlon(long * val){
 *val  = ((long)getc(fp))<<24;
 *val |= ((long)getc(fp))<<16;
 *val |= ((long)getc(fp))<<8;
 *val |= ((long)getc(fp));
 return(1);
}

static unsigned char getsht(short * val){
 *val  = ((short)getc(fp))<<8;
 *val |= ((short)getc(fp));
 return(1);
}

static unsigned char getusht(unsigned short * val){
 *val  = ((short)getc(fp))<<8;
 *val |= ((short)getc(fp));
 return(1);
}

static void getchunk(long clen){
  long i; char c;
  for(i=0;i<clen;i++){ fread(&c,1,1,fp);}
}

static void ReadOldFaces(long clen, short short_form);
static void ReadFaces(long clen, BOOL short_form);
static void ReadVertices(long clen);
static void ReadVertexMapping(long clen);
static void ReadOldTexture(long clen);
static void ReadTexture(long clen);
static void ReadModWireframe(long clen);

void FreeObject(void){
 VERTEX *vp;
 int i;
 if(Nvert > 0){
   for(i=0;i<Nvert;i++){
     vp=(MainVp+i);
     if(vp->nf > 0 && vp->flist != NULL)EVI_Free(vp->flist);
     vp->flist=NULL;
     vp->nf=0;
   }
 }
 if(MainVp != NULL)EVI_Free(MainVp);
 if(MainFp != NULL)EVI_Free(MainFp);
 if(MapsFp != NULL)EVI_Free(MapsFp);
 if(w_frame.p != NULL)EVI_Free(w_frame.p); w_frame.p=NULL; w_frame.Np=0;
 if(w_frame.e != NULL)EVI_Free(w_frame.e); w_frame.e=NULL; w_frame.Ne=0;
 MainVp=NULL; MainFp=NULL; MapsFp=NULL;
 Nvert=0; Nface=0; NvertGlue=0; Nmap=0;
}

int LoadSfxObject(char *FileName){
 int fail=1;
 long CHUNKsize,FORMsize,i,uxx,reset_ruler=1;
 FaceBase=Nface; VertexBase=Nvert;
 if( (fp=fopen(FileName,"rb")) == NULL){
   return 0;
 }
 newface=0; newvert=0; newglue=0;
 if (fread(str,1,4,fp) != 4 || strcmp(str,"FORM") != 0){
   fail=0;
   goto end;
 }
 if (getlon(&FORMsize) != 1) goto end;
 if (fread(str,1,4,fp) != 4 || strcmp(str,"OFXM") != 0){
   fail=0;
   goto end;
}
loop:
 if (fread(str,1,4,fp) != 4 ) goto end;
 if (getlon(&CHUNKsize) != 1) goto end;
 if(strcmp(str,"VERT") == 0){ReadVertices(CHUNKsize);}
 else if(strcmp(str,"FACE") == 0)ReadOldFaces(CHUNKsize,0);
 else if(strcmp(str,"SFAC") == 0)ReadOldFaces(CHUNKsize,1);
 else if(strcmp(str,"SFCE") == 0)ReadFaces(CHUNKsize,TRUE);
 else if(strcmp(str,"LFCE") == 0)ReadFaces(CHUNKsize,FALSE);
 else if(strcmp(str,"VTEX") == 0)ReadVertexMapping(CHUNKsize);
 else if(strcmp(str,"TEXT") == 0)ReadOldTexture(CHUNKsize);
 else if(strcmp(str,"IMAP") == 0)ReadTexture(CHUNKsize);
 else if(strcmp(str,"WFRM") == 0)ReadModWireframe(CHUNKsize);
 else getchunk(CHUNKsize);
 goto loop;
end:
 fclose(fp);
 return fail;
}

static void ReadOldFaces(long clen, short shader){
  long i;
  long id;
  FACE *Fp;
  if(shader)newface=clen/21;
  else      newface=clen/20;
  if(newface == 0)return;
  if((MainFp=(FACE *)EVI_Malloc(sizeof(FACE)*newface)) == NULL)return;
  Fp=MainFp;
  for(i=0;i<newface;i++){
    Fp->A[0] = Fp->A[1] = Fp->A[2] = -1; Fp->next = -1;
    Fp->vid = -1; Fp->start=0;
    getlon(&id);  Fp->V[0]=id+VertexBase;
    getlon(&id);  Fp->V[1]=id+VertexBase;
    getlon(&id);  Fp->V[2]=id+VertexBase;
    fread(Fp->color,1,3,fp);
    fread(&Fp->texture,1,1,fp);
    fread(Fp->matcol,1,3,fp);
    fread(&Fp->brush,1,1,fp);
    if(shader)fread(&Fp->axis,1,1,fp);
    Nface++;
    Fp++;
  }
}

static void ReadFaces(long clen, BOOL short_form){
  long i;
  long id;
  unsigned short ids;
  unsigned char dummy;
  FACE *Fp;
  if(short_form)newface=clen/18;
  else          newface=clen/24;
  if(newface == 0)return;
  if((MainFp=(FACE *)EVI_Malloc(sizeof(FACE)*newface)) == NULL)return;
  Fp=MainFp;
  for(i=0;i<newface;i++){
    Fp->A[0] = Fp->A[1] = Fp->A[2] = -1; Fp->next = -1;
    Fp->vid = -1; Fp->start=0;
    if(short_form){
      getusht(&ids);  Fp->V[0]=ids+VertexBase;
      getusht(&ids);  Fp->V[1]=ids+VertexBase;
      getusht(&ids);  Fp->V[2]=ids+VertexBase;
    }
    else{
      getlon(&id);  Fp->V[0]=id+VertexBase;
      getlon(&id);  Fp->V[1]=id+VertexBase;
      getlon(&id);  Fp->V[2]=id+VertexBase;
    }
    fread(Fp->color,1,3,fp);
    fread(Fp->matcol,1,3,fp);
    fread(&Fp->texture,1,1,fp);
    fread(&dummy,1,1,fp);  // fparam
    fread(&Fp->brush,1,1,fp);
    fread(&dummy,1,1,fp);  // shader
    fread(&Fp->axis,1,1,fp);
    fread(&dummy,1,1,fp);  //pad
    Fp->texture &= 0xe0;
    if((Fp->matcol[1] & 0xf0) > 0)Fp->texture |= 0x04; // make glass
    Nface++;
    Fp++;
  }
}

static void ReadVertices(long clen){
  VERTEX *Vp;
  char c;
  long i,j;
  newvert=clen/12;
  if(newvert == 0)return;
  if((MainVp=(VERTEX *)EVI_Malloc(sizeof(VERTEX)*newvert)) == NULL)return;
  Vp=MainVp;
  for(i=0;i<newvert;i++){
    Vp->gp=0; Vp->nf=0; Vp->flist=NULL;
    getlon(&Vp->p[0]);
    getlon(&Vp->p[1]);
    getlon(&Vp->p[2]);
    Nvert++;
    Vp++;
  }
}

static void ReadVertexMapping(long clen){
  long i,id,ix,iy,szz;
  VERTEX *Vp;
  newglue=clen/12;
  if(newglue == 0)return;
  if(newglue > newvert){ /* dump the mapping */
    for(i=0;i<newglue;i++){
      getlon(&id); getlon(&ix); getlon(&iy);
    }
    return;
  }
  for(i=0;i<newglue;i++){
    getlon(&id); getlon(&ix); getlon(&iy);
    Vp=(MainVp+VertexBase+id);
    Vp->gp=1;
    Vp->x=(float)ix*0.000001;
    Vp->y=(float)iy*0.000001;
    NvertGlue++;
  }
}

static void ReadOldTexture(long clen){
  MAP *temp_map;
  unsigned char t_colour[3];
  char i,j,*cp1,*cp2,brushmaptype,brushmethod,moziac,p,r,b;
  char fsurf[256],frefl[256],fbump[256],tempstr[256];
  long k,l,m,frame=1,firstframe,framedelay,lastframe,thisframe,brushangle,file_len;
  short s,version;
  fread(&j,1,1,fp);
  if(j > 0)for(i=0;i<j;i++){
    version=0;
    if(MapsFp == NULL)temp_map = (MAP *)EVI_Malloc(sizeof(MAP));
    else              temp_map = (MAP *)EVI_Realloc(MapsFp,(Nmap+1)*sizeof(MAP));
    if(temp_map == NULL){
      MessageBeep(MB_OK);
      MessageBox(NULL,"Out of Memory",NULL,MB_OK);
      return;
    }
    MapsFp=temp_map;
    getlon(&file_len);
    fread(tempstr,1,file_len,fp); /* this contains the version2 string */
    getlon(&l); MapsFp[Nmap].p[0]= (double)l;
    getlon(&l); MapsFp[Nmap].p[1]= (double)l;
    getlon(&l); MapsFp[Nmap].p[2]= (double)l;
    getlon(&l); MapsFp[Nmap].x[0]= (double)l;
    getlon(&l); MapsFp[Nmap].x[1]= (double)l;
    getlon(&l); MapsFp[Nmap].x[2]= (double)l;
    getlon(&l); MapsFp[Nmap].y[0]= (double)l;
    getlon(&l); MapsFp[Nmap].y[1]= (double)l;
    getlon(&l); MapsFp[Nmap].y[2]= (double)l;
    getlon(&l); brushmethod=(char)l;   /* still bump etc from old version */
    getlon(&firstframe); getlon(&framedelay); getlon(&lastframe);
    getlon(&l); /* used by skeleton for br. assign. not used by renderer */
    getlon(&l); brushmaptype=(char)l;
    getlon(&l); /* this is the expansion count */
    if(l > 0){
      getlon(&brushangle);
      if(l > 1){
        for(k=1;k<5;k++)getlon(&m); /* dump the brush lock info */
        if(l > 5){             /* this must be a version 2 file */
          version=1;
          fread(tempstr,1,32,fp);
          fread(&p,1,1,fp); fread(&r,1,1,fp); fread(&b,1,1,fp);
          getsht(&s); MapsFp[Nmap].pp=s; /* these will be set to -1 */
          getsht(&s); MapsFp[Nmap].rp=s; /* if a map of this type   */
          getsht(&s); MapsFp[Nmap].bp=s; /* is not used in Material */
          fread(t_colour,1,3,fp);  /* not used yet */
          fread(t_colour,1,3,fp);
          fread(t_colour,1,3,fp);  /* ------------ */
          if(p)fread(fsurf,1,file_len,fp);
          if(r)fread(frefl,1,file_len,fp);
          if(b)fread(fbump,1,file_len,fp);
          getsht(&s); /* should be 0 use as expansion count */
          if(s > 0)for(k=0;k<s;k++)getlon(&m);
        }
      }
    }
    else brushangle=360;
    moziac=NO;
    if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
    if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
    if(version == 0){
      p=r=b=0;
      MapsFp[Nmap].pp=100;
      MapsFp[Nmap].rp= 50;
      MapsFp[Nmap].bp= 50;
      if(brushmethod == REFLBRUSH){
        strcpy(frefl,tempstr); r=1;
      }
      else if(brushmethod == BUMPBRUSH){
        strcpy(fbump,tempstr); b=1;
      }
      else{
        strcpy(fsurf,tempstr); p=1;
      }
      if(brushmethod != ANIMBRUSH)brushmethod=STILBRUSH; /* just switch now */
    }
    if(p && brushmethod == ANIMBRUSH){ /* animated brush calculate frame */
      cp1=strchr(fsurf,'0');
      cp2=strchr(fsurf,'.');
      if(cp1 == NULL || cp2 == NULL){
        sprintf(tempstr,"dummy.evi");
      }
      else{
        int ln;
        char *cp,extn[16],fmat[16];
        strcpy(extn,cp2);
        ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
        *cp1='\0';
        if(framedelay == 0)thisframe=frame+firstframe-1;
        else               thisframe=frame/(framedelay+1)+firstframe;
        while(thisframe > lastframe)thisframe -= lastframe;
        sprintf(fmat,"%%s%%0%ldld%%s",ln);
        sprintf(tempstr,fmat,fsurf,thisframe,extn);
      }
      strcpy(fsurf,tempstr);
    }
    if(p) strcpy(MapsFp[Nmap].filename,fsurf);
    else         MapsFp[Nmap].pp = -1;
    if(r) strcpy(MapsFp[Nmap].frefl,frefl);
    else         MapsFp[Nmap].rp = -1;
    if(b) strcpy(MapsFp[Nmap].fbump,fbump);
    else         MapsFp[Nmap].bp = -1;
    MapsFp[Nmap].map=(short)brushmaptype;
    MapsFp[Nmap].method=(short)brushmethod;
    MapsFp[Nmap].angle=360.0/(double)brushangle;
    MapsFp[Nmap].moziac=moziac;
    Nmap++;
  }
  return;
}

static void ReadTexture(long clen){
  MAP *temp_map;
  unsigned char t_colour[3];
  char *cp1,*cp2,brushmaptype,brushmethod,moziac,p,r,b,t,kt;
  char fsurf[256],frefl[256],fbump[256],tempstr[256];
  long i,j,k,l,m,frame=1,firstframe,framedelay,lastframe,thisframe,
       brushangle;
  short s;
  getlon(&j);
  if(j > 0)for(i=0;i<j;i++){
    if(MapsFp == NULL)temp_map = (MAP *)EVI_Malloc(sizeof(MAP));
    else              temp_map = (MAP *)EVI_Realloc(MapsFp,(Nmap+1)*sizeof(MAP));
    if(temp_map == NULL){
      MessageBeep(MB_OK);
      MessageBox(NULL,"Out of Memory",NULL,MB_OK);
      return;
    }
    MapsFp=temp_map;
    getlon(&l); MapsFp[Nmap].p[0]= (double)l;
    getlon(&l); MapsFp[Nmap].p[1]= (double)l;
    getlon(&l); MapsFp[Nmap].p[2]= (double)l;
    getlon(&l); MapsFp[Nmap].x[0]= (double)l;
    getlon(&l); MapsFp[Nmap].x[1]= (double)l;
    getlon(&l); MapsFp[Nmap].x[2]= (double)l;
    getlon(&l); MapsFp[Nmap].y[0]= (double)l;
    getlon(&l); MapsFp[Nmap].y[1]= (double)l;
    getlon(&l); MapsFp[Nmap].y[2]= (double)l;
    getlon(&l); brushmethod=(char)l;
    getlon(&firstframe); getlon(&framedelay); getlon(&lastframe);
    getlon(&l); getlon(&l); getlon(&l); getlon(&l); //r
    getlon(&l); getlon(&l); getlon(&l); getlon(&l); //b
    getlon(&l); getlon(&l); getlon(&l); getlon(&l); //t
    getlon(&l); /* used by skeleton for br. assign. not used by renderer */
    getlon(&l); brushmaptype=(char)l;
    getlon(&brushangle);
    for(k=0;k<4;k++)getlon(&m); /* dump the brush lock info */
    fread(tempstr,1,32,fp);
    fread(&p,1,1,fp); fread(&r,1,1,fp); fread(&b,1,1,fp); fread(&t,1,1,fp);
    fread(&kt,1,1,fp);
    getsht(&s); MapsFp[Nmap].pp=s; /* these will be set to -1 */
    getsht(&s); MapsFp[Nmap].rp=s; /* if a map of this type   */
    getsht(&s); MapsFp[Nmap].bp=s; /* is not used in Material */
    getsht(&s);
    fread(t_colour,1,3,fp);  /* not used yet */
    fread(t_colour,1,3,fp);
    fread(t_colour,1,3,fp);  /* ------------ */
    fread(t_colour,1,3,fp);  /* key colour   */
    fread(fsurf,1,256,fp);
    fread(frefl,1,256,fp);
    fread(fbump,1,256,fp);
    fread(tempstr,1,256,fp);
    moziac=NO;
    if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
    if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
    if(p && brushmethod == ANIMBRUSH){ /* animated brush calculate frame */
      cp1=strchr(fsurf,'0');
      cp2=strchr(fsurf,'.');
      if(cp1 == NULL || cp2 == NULL){
        sprintf(tempstr,"dummy.evi");
      }
      else{
        int ln;
        char *cp,extn[16],fmat[16];
        strcpy(extn,cp2);
        ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
        *cp1='\0';
        if(framedelay == 0)thisframe=frame+firstframe-1;
        else               thisframe=frame/(framedelay+1)+firstframe;
        while(thisframe > lastframe)thisframe -= lastframe;
        sprintf(fmat,"%%s%%0%ldld%%s",ln);
        sprintf(tempstr,fmat,fsurf,thisframe,extn);
      }
      strcpy(fsurf,tempstr);
    }
    if(p) strcpy(MapsFp[Nmap].filename,fsurf);
    else         MapsFp[Nmap].pp = -1;
    if(r) strcpy(MapsFp[Nmap].frefl,frefl);
    else         MapsFp[Nmap].rp = -1;
    if(b) strcpy(MapsFp[Nmap].fbump,fbump);
    else         MapsFp[Nmap].bp = -1;
    MapsFp[Nmap].map=(short)brushmaptype;
    MapsFp[Nmap].method=(short)brushmethod;
    MapsFp[Nmap].angle=360.0/(double)brushangle;
    MapsFp[Nmap].moziac=moziac;
    Nmap++;
  }
  return;
}

static void ReadModWireframe(long clen){
 long i,ne,np;
 getlon(&np);  getlon(&ne);
 i=((np*3)+(ne*2)+2)*sizeof(long);
 if(clen != i){   /* if chunk is void */
   getchunk(clen-2);
   return;
 }
 w_frame.p = (point *)EVI_Malloc(np*sizeof(point));
 w_frame.e = (twoLongs *)EVI_Malloc(ne*sizeof(twoLongs));
 if(w_frame.p == NULL || w_frame.e == NULL){
   if(w_frame.p != NULL)EVI_Free(w_frame.p); w_frame.p=NULL;
   if(w_frame.e != NULL)EVI_Free(w_frame.e); w_frame.e=NULL;
   getchunk(clen-2);
   return;
 }
 w_frame.Np=np;  w_frame.Ne=ne;
 for(i=0;i<w_frame.Np;i++){
   getlon(&w_frame.p[i][0]);
   getlon(&w_frame.p[i][1]);
   getlon(&w_frame.p[i][2]);
 }
 for(i=0;i<w_frame.Ne;i++){
   getlon(&w_frame.e[i][0]);
   getlon(&w_frame.e[i][1]);
 }
}

void ListAdjFaces(void){
 FACE *f;
 long i,j,*fl,id,n,m,id0,id1,id2,nl[3],ml[3],k,kk,fjd,fkd;
 for(i=0;i<Nface;i++){ /* make list of work faces attached to each vertex */
   f=(MainFp+i);
   for(j=0;j<3;j++){
     id=f->V[j];
     fl=(long *)((MainVp+id)->flist);
     n=(MainVp+id)->nf;
     if(fl == NULL){
       if((fl=(long *)EVI_Malloc(sizeof(long))) == NULL)return;
     }
     else{
       if((fl=(long *)EVI_Realloc(fl,(n+1)*sizeof(long))) == NULL)return;
     }
     fl[n]=i;
     (MainVp+id)->flist=(int *)(fl);
     (MainVp+id)->nf=n+1;
   }
 }
 for(i=0;i<Nface;i++){ /* find faces that are adjacent to face i */
   f=(MainFp+i);
   id0=f->V[0];   id1=f->V[1];   id2=f->V[2];
   nl[0]=id0; ml[0]=id1;
   nl[1]=id1; ml[1]=id2;
   nl[2]=id2; ml[2]=id0;
   for(kk=0;kk<3;kk++){
     f->A[kk] = -1;
     if((n=(MainVp+nl[kk])->nf) > 0 && (m=(MainVp+ml[kk])->nf) > 0){
       for(j=0;j<n;j++){
         fjd=MainVp[nl[kk]].flist[j];
         for(k=0;k<m;k++){
           fkd=MainVp[ml[kk]].flist[k];
           if(fkd == fjd && fkd != i){
             f->A[kk]=fkd;
             goto ENDF;
           }
         }
       }
     }
     ENDF:;
   }
 }
 return;
}

static void MakeSameOrientation(long last, long id){
 FACE *f;
 long i0,i1,i2,l0,l1,l2,a0,a1,a2;
 f=(MainFp+last); l0=f->V[0]; l1=f->V[1]; l2=f->V[2];
 f=(MainFp+id);   i0=f->V[0]; i1=f->V[1]; i2=f->V[2];
                  a0=f->A[0]; a1=f->A[1]; a2=f->A[2];
 if((i2 == l2 && i1 == l1) ||
    (i2 == l1 && i1 == l0) ||
    (i2 == l0 && i1 == l2)){
   f->V[2]=i1; f->V[1]=i2; f->A[2]=a0; f->A[0]=a2;
   return;
 }
 if((i0 == l2 && i2 == l1) ||
    (i0 == l1 && i2 == l0) ||
    (i0 == l0 && i2 == l2)){
   f->V[0]=i2; f->V[2]=i0; f->A[0]=a1; f->A[1]=a0;
   return;
 }
 if((i1 == l1 && i0 == l0) ||
    (i1 == l2 && i0 == l1) ||
    (i1 == l0 && i0 == l2)){
   f->V[1]=i0; f->V[0]=i1; f->A[1]=a2; f->A[2]=a1;
   return;
 }
 return;
}

static void SetConnectingID(long last, long id){
 if((MainFp+id)->start == 1)return;
 (MainFp+id)->start=1;
 if(last >= 0)MakeSameOrientation(last,id);
 last=id;
 if((MainFp+id)->A[0] >= 0)SetConnectingID(last,(MainFp+id)->A[0]);
 if((MainFp+id)->A[1] >= 0)SetConnectingID(last,(MainFp+id)->A[1]);
 if((MainFp+id)->A[2] >= 0)SetConnectingID(last,(MainFp+id)->A[2]);
 return;
}

void FixUpOrientation(void){
 long i;
 while(1){
   for(i=0;i<Nface;i++){
     if((MainFp+i)->start == 0)goto PROCESSIT;
   }
   break;
   PROCESSIT:
   SetConnectingID(-1,i);
 }
 for(i=0;i<Nface;i++)(MainFp+i)->start=0;
}

static long CanExtendStrip(FACE *f, long next){
 if((MainFp+next)->next != -1)next = -1;
 else {
   FACE *fn;
   vector n,nn;
   fn=(MainFp+next);
   if(abs(f->color[0] - fn->color[0]) > 10 ||
      abs(f->color[1] - fn->color[1]) > 10 ||
      abs(f->color[2] - fn->color[2]) > 10 ){
     next = -1;
   }
   else{
     Normalize((MainVp+(f->V[0]))->p,         /* too great an angle     */
               (MainVp+(f->V[1]))->p,         /* assumes that faces     */
               (MainVp+(f->V[2]))->p,n);      /* have been consistently */
     Normalize((MainVp+(fn->V[0]))->p,        /* numbered.              */
               (MainVp+(fn->V[1]))->p,
               (MainVp+(fn->V[2]))->p,nn);
     if(DOT(n,nn) < 0.01)next = -1;
   }
 }
 return next;
}

static long NextFace(int id, int v1, int v2, int last, int *v3){
 FACE *f;
 int i0,i1,i2,next;
 next = -10;
 f=(MainFp+last); i0=f->V[0]; i1=f->V[1]; i2=f->V[2];
 if(     i0 == v1 && i1 == v2){
   *v3 = 2;
   if(id == 0)next=f->A[1];
   else       next=f->A[2];
 }
 else if(i1 == v1 && i0 == v2){
   *v3 = 2;
   if(id == 0)next=f->A[2];
   else       next=f->A[1];
 }
 else if(i1 == v1 && i2 == v2){
   *v3 = 0;
   if(id == 0)next=f->A[2];
   else       next=f->A[0];
 }
 else if(i2 == v1 && i1 == v2){
   *v3 = 0;
   if(id == 0)next=f->A[0];
   else       next=f->A[2];
 }
 else if(i2 == v1 && i0 == v2){
   *v3 = 1;
   if(id == 0)next=f->A[0];
   else       next=f->A[1];
 }
 else if(i0 == v1 && i2 == v2){
   *v3 = 1;
   if(id == 0)next=f->A[1];
   else       next=f->A[0];
 }
 if(next == -10){
   MessageBox(NULL,"None Found",NULL,MB_OK);
   next=-1;
 }
 if(next >= 0 && (MainFp+next)->next != -1)next = -1; /* already used */
 if(next >= 0)next=CanExtendStrip(f,next); /* already used or other attrib */
 return next;
}

void StripFaces(void){
 vector n,nn;
 int i,j,k,last,next,v1,v2,v3,v4,i0,i1,i2;
 FACE *fp1,*fp2;
 for(i=0,fp1=MainFp;i<Nface;i++){
   last=i; k= -1;
   if(fp1->next == -1){ /* not processed */
     if((next=fp1->A[1]) >= 0){
       j=1; next=CanExtendStrip(fp1,next); //already in use/other attrib
       if(next >= 0){
         k=j; /* if this strip is only two long try the others too */
         if(NextFace(0,fp1->V[1],fp1->V[2],next,&v3) < 0)next = -1;
       }
     }
     if(next < 0 && (next=fp1->A[0]) >= 0){
       j=0; next=CanExtendStrip(fp1,next); //already in use/other attrib
       if(next >= 0){
         k=j; /* if this strip is only two long try the others too */
         if(NextFace(0,fp1->V[0],fp1->V[1],next,&v3) < 0)next = -1;
       }
     }
     if(next < 0 && (next=fp1->A[2]) >= 0){
       j=2; next=CanExtendStrip(fp1,next); //already in use/other attrib
       if(next >= 0){
         k=j; /* if this strip is only two long try the others too */
         if(NextFace(0,fp1->V[2],fp1->V[0],next,&v3) < 0)next = -1;
       }
     }
     if(next == -1 && k >= 0){ /* use last best hope */
       j=k; next=fp1->A[j];
     }
     if(next >= 0){
       (MainFp+last)->start=1;
       (MainFp+last)->next=next;  /* face 0 */
       (MainFp+next)->next = -2;  /* end of strip face 1 */
       if(j == 0){
         i0=(MainFp+last)->V[0];
         i1=(MainFp+last)->V[1];
         i2=(MainFp+last)->V[2];
         (MainFp+last)->V[1]=i0;
         (MainFp+last)->V[2]=i1;
         (MainFp+last)->V[0]=i2;
       }
       else if(j == 2){
         i0=(MainFp+last)->V[0];
         i1=(MainFp+last)->V[1];
         i2=(MainFp+last)->V[2];
         (MainFp+last)->V[1]=i2;
         (MainFp+last)->V[2]=i0;
         (MainFp+last)->V[0]=i1;
       }
       v1=(MainFp+last)->V[1];    /* of face 0 */
       v2=(MainFp+last)->V[2];    /* of face 0 */
       while(1){
         last=next;                    /* last now equals face 1      */
         next=NextFace(0,v1,v2,last,&v3);
         (MainFp+last)->vid=v3;        /* index vertex (3) current end*/
         v3=(MainFp+last)->V[v3];      /* next vertex (3)             */
         if(next < 0)break;            /* cannot extend               */
         (MainFp+last)->next=next;     /* set pointer                 */
         (MainFp+next)->next = -2;     /* set this to an end of strip */
         last=next;                    /* last now equal face 2       */
         next=NextFace(1,v3,v2,last,&v4);
         (MainFp+last)->vid=v4;        /* index of next vertex face 2 */
         v4=(MainFp+last)->V[v4];
         if(next < 0)break;            /* cannot extend               */
         (MainFp+last)->next=next;
         (MainFp+next)->next = -2;
         v1=v3;                        /* v1 moves on to v3           */
         v2=v4;                        /* v2 moves on to v4           */
       }
     }
   }
   fp1++;
 }
//#if 0
 for(i=0;i<Nface;i++){ /* any two face strips make quad if possible */
   fp1=(MainFp+i);
   if(fp1->start == 1){
     fp2=(MainFp+(fp1->next));
     if(fp2->next == -2){ /* two face strip */
       Normalize((MainVp+(fp1->V[0]))->p,
                 (MainVp+(fp1->V[1]))->p,
                 (MainVp+(fp1->V[2]))->p,n);
       Normalize((MainVp+(fp2->V[0]))->p,
                 (MainVp+(fp2->V[1]))->p,
                 (MainVp+(fp2->V[2]))->p,nn);
       if(DOT(n,nn) > 0.95){
         fp1->start=2; /* quadilateral */
       }
       else{ /* set remaining two triangles to TRIANGLES ? */
//         fp2->next = -1;
//         fp1->next= -1;
//         fp1->start=FALSE;
       }
     }
   }
 }
//#endif
#if 0
 ff=fopen("d:strip.evi","w");
 if(ff != NULL){
   long kk,kkk,kkc;
   fprintf(ff,"%ld Vertices %ld faces\n",Nvert,Nface);
   for(i=0,k=0,kkk=0;i<Nface;i++){
     if((MainFp+i)->next == -1){
       k++; //fprintf(ff,"%3ld Single\n",i);
     }
   }
   fprintf(ff,"%ld Triangles\n",k); kkc=k;
   k=0;
   for(i=0;i<Nface;i++){
     if((MainFp+i)->start == 2 && (MainFp+i)->next >= 0 &&
         (MainFp+(MainFp+i)->next)->next == -2){
       k++;
     }
   }
   fprintf(ff,"%ld Quads\n",k); kkc += 2*k;
   k=0;
   fprintf(ff,"Triangle Strips\n",k); kkc += 2*k;
   for(i=0;i<Nface;i++){
     if((MainFp+i)->start == 1){
       j=i;
       fprintf(ff,"[%3ld] ",k);
       kk=0; while(j >= 0){
//fprintf(ff," %ld",j);
         j=(MainFp+j)->next; kk++;
       }
       fprintf(ff," %3ld faces\n",kk); kkk += kk;
       k++;
     }
   }
   fprintf(ff,"%ld Strips %ld faces  %ld in total [%ld]\n",k,kkk,kkk+kkc,Nface);
   fclose(ff);
 }
#endif
}
