/* file OFXLOAD.C  || */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>

#include "sim.h"
#include "resource.h"

#define PLANE           0
#define CYLINDER        1
#define PLANE_MOZIAC    2
#define CYLINDER_MOZIAC 3
#define MAP_BY_VERTEX   4
#define STILBRUSH       0
#define ANIMBRUSH       1
#define BUMPBRUSH       2
#define REFLBRUSH       3
#define PSETBRUSH       4

extern VERTEX *MainVp;
extern FACE   *MainFp;
extern MAP    *MapsFp;
extern WFRAME w_frame;
extern long   Nvert,Nface,NvertGlue,Nmap;

extern BOOL Normalize(point, point, point, GLfloat *);

void ListAdjFaces(void);
void FixUpOrientation(void);

static unsigned char getlon(long *);
static unsigned char getsht(short *);
static unsigned char getusht(unsigned short *);
static void outsht(short);
static void outusht(unsigned short);
static void outlng(long);
static void getchunk(long clen);

#define BRUSHBIT   0x40
#define SMOOTHBIT  0x80
#define TILEBIT    0x20;

FILE *ff;

static FILE *fp;
static long newface,newvert,newglue,minX,minY,minZ,maxX,maxY,maxZ;
static char str[10];
static long FaceBase,EdgeBase,VertexBase;

static unsigned char getlon(long * val){
 *val  = ((long)getc(fp))<<24;
 *val |= ((long)getc(fp))<<16;
 *val |= ((long)getc(fp))<<8;
 *val |= ((long)getc(fp));
 return(1);
}

static unsigned char getsht(short * val){
 *val  = ((short)getc(fp))<<8;
 *val |= ((short)getc(fp));
 return(1);
}

static unsigned char getusht(unsigned short * val){
 *val  = ((short)getc(fp))<<8;
 *val |= ((short)getc(fp));
 return(1);
}

static void getchunk(long clen){
  long i; char c;
  for(i=0;i<clen;i++){ fread(&c,1,1,fp);}
}

static void ReadOldFaces(long clen, short short_form);
static void ReadFaces(long clen, BOOL short_form);
static void ReadVertices(long clen);
static void ReadVertexMapping(long clen);
static void ReadOldTexture(long clen);
static void ReadTexture(long clen);
static void ReadModWireframe(long clen);

void FreeObject(void){
 VERTEX *vp;
 int i;
 if(Nvert > 0){
   for(i=0;i<Nvert;i++){
     vp=(MainVp+i);
     if(vp->nf > 0 && vp->flist != NULL)EVI_Free(vp->flist);
     vp->flist=NULL;
     vp->nf=0;
   }
 }
 if(MainVp != NULL)EVI_Free(MainVp);
 if(MainFp != NULL)EVI_Free(MainFp);
 if(MapsFp != NULL)EVI_Free(MapsFp);
 if(w_frame.p != NULL)EVI_Free(w_frame.p); w_frame.p=NULL; w_frame.Np=0;
 if(w_frame.e != NULL)EVI_Free(w_frame.e); w_frame.e=NULL; w_frame.Ne=0;
 MainVp=NULL; MainFp=NULL; MapsFp=NULL;
 Nvert=0; Nface=0; NvertGlue=0; Nmap=0;
}

int LoadSfxObject(char *FileName){
 int fail=1;
 long CHUNKsize,FORMsize,i,uxx,reset_ruler=1;
 FaceBase=Nface; VertexBase=Nvert;
 if( (fp=fopen(FileName,"rb")) == NULL){
   return 0;
 }
 newface=0; newvert=0; newglue=0;
 if (fread(str,1,4,fp) != 4 || strcmp(str,"FORM") != 0){
   fail=0;
   goto end;
 }
 if (getlon(&FORMsize) != 1) goto end;
 if (fread(str,1,4,fp) != 4 || strcmp(str,"OFXM") != 0){
   fail=0;
   goto end;
}
loop:
 if (fread(str,1,4,fp) != 4 ) goto end;
 if (getlon(&CHUNKsize) != 1) goto end;
 if(strcmp(str,"VERT") == 0){ReadVertices(CHUNKsize);}
 else if(strcmp(str,"FACE") == 0)ReadOldFaces(CHUNKsize,0);
 else if(strcmp(str,"SFAC") == 0)ReadOldFaces(CHUNKsize,1);
 else if(strcmp(str,"SFCE") == 0)ReadFaces(CHUNKsize,TRUE);
 else if(strcmp(str,"LFCE") == 0)ReadFaces(CHUNKsize,FALSE);
 else if(strcmp(str,"VTEX") == 0)ReadVertexMapping(CHUNKsize);
 else if(strcmp(str,"TEXT") == 0)ReadOldTexture(CHUNKsize);
 else if(strcmp(str,"IMAP") == 0)ReadTexture(CHUNKsize);
 else if(strcmp(str,"WFRM") == 0)ReadModWireframe(CHUNKsize);
 else getchunk(CHUNKsize);
 goto loop;
end:
 fclose(fp);
 return fail;
}

static void ReadOldFaces(long clen, short shader){
  long i;
  long id;
  FACE *Fp;
  if(shader)newface=clen/21;
  else      newface=clen/20;
  if(newface == 0)return;
  if((MainFp=(FACE *)EVI_Malloc(sizeof(FACE)*newface)) == NULL)return;
  Fp=MainFp;
  for(i=0;i<newface;i++){
    Fp->A[0] = Fp->A[1] = Fp->A[2] = -1; Fp->next = -1;
    Fp->vid = -1; Fp->start=0;
    getlon(&id);  Fp->V[0]=id+VertexBase;
    getlon(&id);  Fp->V[1]=id+VertexBase;
    getlon(&id);  Fp->V[2]=id+VertexBase;
    fread(Fp->color,1,3,fp);
    fread(&Fp->texture,1,1,fp);
    fread(Fp->matcol,1,3,fp);
    fread(&Fp->brush,1,1,fp);
    if(shader)fread(&Fp->axis,1,1,fp);
    Nface++;
    Fp++;
  }
}

static void ReadFaces(long clen, BOOL short_form){
  long i;
  long id;
  unsigned short ids;
  unsigned char dummy;
  FACE *Fp;
  if(short_form)newface=clen/18;
  else          newface=clen/24;
  if(newface == 0)return;
  if((MainFp=(FACE *)EVI_Malloc(sizeof(FACE)*newface)) == NULL)return;
  Fp=MainFp;
  for(i=0;i<newface;i++){
    Fp->A[0] = Fp->A[1] = Fp->A[2] = -1; Fp->next = -1;
    Fp->vid = -1; Fp->start=0;
    if(short_form){
      getusht(&ids);  Fp->V[0]=ids+VertexBase;
      getusht(&ids);  Fp->V[1]=ids+VertexBase;
      getusht(&ids);  Fp->V[2]=ids+VertexBase;
    }
    else{
      getlon(&id);  Fp->V[0]=id+VertexBase;
      getlon(&id);  Fp->V[1]=id+VertexBase;
      getlon(&id);  Fp->V[2]=id+VertexBase;
    }
    fread(Fp->color,1,3,fp);
    fread(Fp->matcol,1,3,fp);
    fread(&Fp->texture,1,1,fp);
    fread(&dummy,1,1,fp);  // fparam
    fread(&Fp->brush,1,1,fp);
    fread(&dummy,1,1,fp);  // shader
    fread(&Fp->axis,1,1,fp);
    fread(&dummy,1,1,fp);  //pad
    Fp->texture &= 0xe0;
    if((Fp->matcol[1] & 0xf0) > 0)Fp->texture |= 0x04; // make glass
    Nface++;
    Fp++;
  }
}

static void ReadVertices(long clen){
  VERTEX *Vp;
  char c;
  long i,j;
  newvert=clen/12;
  if(newvert == 0)return;
  if((MainVp=(VERTEX *)EVI_Malloc(sizeof(VERTEX)*newvert)) == NULL)return;
  Vp=MainVp;
  for(i=0;i<newvert;i++){
    Vp->gp=0; Vp->nf=0; Vp->flist=NULL;
    getlon(&Vp->p[0]);
    getlon(&Vp->p[1]);
    getlon(&Vp->p[2]);
    Nvert++;
    Vp++;
  }
}

static void ReadVertexMapping(long clen){
  long i,id,ix,iy,szz;
  VERTEX *Vp;
  newglue=clen/12;
  if(newglue == 0)return;
  if(newglue > newvert){ /* dump the mapping */
    for(i=0;i<newglue;i++){
      getlon(&id); getlon(&ix); getlon(&iy);
    }
    return;
  }
  for(i=0;i<newglue;i++){
    getlon(&id); getlon(&ix); getlon(&iy);
    Vp=(MainVp+VertexBase+id);
    Vp->gp=1;
    Vp->x=(float)ix*0.000001;
    Vp->y=(float)iy*0.000001;
    NvertGlue++;
  }
}

static void ReadOldTexture(long clen){
  MAP *temp_map;
  unsigned char t_colour[3];
  char i,j,*cp1,*cp2,brushmaptype,brushmethod,moziac,p,r,b;
  char fsurf[256],frefl[256],fbump[256],tempstr[256];
  long k,l,m,frame=1,firstframe,framedelay,lastframe,thisframe,brushangle,file_len;
  short s,version;
  fread(&j,1,1,fp);
  if(j > 0)for(i=0;i<j;i++){
    version=0;
    if(MapsFp == NULL)temp_map = (MAP *)EVI_Malloc(sizeof(MAP));
    else              temp_map = (MAP *)EVI_Realloc(MapsFp,(Nmap+1)*sizeof(MAP));
    if(temp_map == NULL){
      MessageBeep(MB_OK);
      MessageBox(NULL,"Out of Memory",NULL,MB_OK);
      return;
    }
    MapsFp=temp_map;
    getlon(&file_len);
    fread(tempstr,1,file_len,fp); /* this contains the version2 string */
    getlon(&l); MapsFp[Nmap].p[0]= (double)l;
    getlon(&l); MapsFp[Nmap].p[1]= (double)l;
    getlon(&l); MapsFp[Nmap].p[2]= (double)l;
    getlon(&l); MapsFp[Nmap].x[0]= (double)l;
    getlon(&l); MapsFp[Nmap].x[1]= (double)l;
    getlon(&l); MapsFp[Nmap].x[2]= (double)l;
    getlon(&l); MapsFp[Nmap].y[0]= (double)l;
    getlon(&l); MapsFp[Nmap].y[1]= (double)l;
    getlon(&l); MapsFp[Nmap].y[2]= (double)l;
    getlon(&l); brushmethod=(char)l;   /* still bump etc from old version */
    getlon(&firstframe); getlon(&framedelay); getlon(&lastframe);
    getlon(&l); /* used by skeleton for br. assign. not used by renderer */
    getlon(&l); brushmaptype=(char)l;
    getlon(&l); /* this is the expansion count */
    if(l > 0){
      getlon(&brushangle);
      if(l > 1){
        for(k=1;k<5;k++)getlon(&m); /* dump the brush lock info */
        if(l > 5){             /* this must be a version 2 file */
          version=1;
          fread(tempstr,1,32,fp);
          fread(&p,1,1,fp); fread(&r,1,1,fp); fread(&b,1,1,fp);
          getsht(&s); MapsFp[Nmap].pp=s; /* these will be set to -1 */
          getsht(&s); MapsFp[Nmap].rp=s; /* if a map of this type   */
          getsht(&s); MapsFp[Nmap].bp=s; /* is not used in Material */
          fread(t_colour,1,3,fp);  /* not used yet */
          fread(t_colour,1,3,fp);
          fread(t_colour,1,3,fp);  /* ------------ */
          if(p)fread(fsurf,1,file_len,fp);
          if(r)fread(frefl,1,file_len,fp);
          if(b)fread(fbump,1,file_len,fp);
          getsht(&s); /* should be 0 use as expansion count */
          if(s > 0)for(k=0;k<s;k++)getlon(&m);
        }
      }
    }
    else brushangle=360;
    moziac=NO;
    if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
    if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
    if(version == 0){
      p=r=b=0;
      MapsFp[Nmap].pp=100;
      MapsFp[Nmap].rp= 50;
      MapsFp[Nmap].bp= 50;
      if(brushmethod == REFLBRUSH){
        strcpy(frefl,tempstr); r=1;
      }
      else if(brushmethod == BUMPBRUSH){
        strcpy(fbump,tempstr); b=1;
      }
      else{
        strcpy(fsurf,tempstr); p=1;
      }
      if(brushmethod != ANIMBRUSH)brushmethod=STILBRUSH; /* just switch now */
    }
    if(p && brushmethod == ANIMBRUSH){ /* animated brush calculate frame */
      cp1=strchr(fsurf,'0');
      cp2=strchr(fsurf,'.');
      if(cp1 == NULL || cp2 == NULL){
        sprintf(tempstr,"dummy.evi");
      }
      else{
        int ln;
        char *cp,extn[16],fmat[16];
        strcpy(extn,cp2);
        ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
        *cp1='\0';
        if(framedelay == 0)thisframe=frame+firstframe-1;
        else               thisframe=frame/(framedelay+1)+firstframe;
        while(thisframe > lastframe)thisframe -= lastframe;
        sprintf(fmat,"%%s%%0%ldld%%s",ln);
        sprintf(tempstr,fmat,fsurf,thisframe,extn);
      }
      strcpy(fsurf,tempstr);
    }
    if(p) strcpy(MapsFp[Nmap].filename,fsurf);
    else         MapsFp[Nmap].pp = -1;
    if(r) strcpy(MapsFp[Nmap].frefl,frefl);
    else         MapsFp[Nmap].rp = -1;
    if(b) strcpy(MapsFp[Nmap].fbump,fbump);
    else         MapsFp[Nmap].bp = -1;
    MapsFp[Nmap].map=(short)brushmaptype;
    MapsFp[Nmap].method=(short)brushmethod;
    MapsFp[Nmap].angle=360.0/(double)brushangle;
    MapsFp[Nmap].moziac=moziac;
    Nmap++;
  }
  return;
}

static void ReadTexture(long clen){
  MAP *temp_map;
  unsigned char t_colour[3];
  char *cp1,*cp2,brushmaptype,brushmethod,moziac,p,r,b,t,kt;
  char fsurf[256],frefl[256],fbump[256],tempstr[256];
  long i,j,k,l,m,frame=1,firstframe,framedelay,lastframe,thisframe,
       brushangle;
  short s;
  getlon(&j);
  if(j > 0)for(i=0;i<j;i++){
    if(MapsFp == NULL)temp_map = (MAP *)EVI_Malloc(sizeof(MAP));
    else              temp_map = (MAP *)EVI_Realloc(MapsFp,(Nmap+1)*sizeof(MAP));
    if(temp_map == NULL){
      MessageBeep(MB_OK);
      MessageBox(NULL,"Out of Memory",NULL,MB_OK);
      return;
    }
    MapsFp=temp_map;
    getlon(&l); MapsFp[Nmap].p[0]= (double)l;
    getlon(&l); MapsFp[Nmap].p[1]= (double)l;
    getlon(&l); MapsFp[Nmap].p[2]= (double)l;
    getlon(&l); MapsFp[Nmap].x[0]= (double)l;
    getlon(&l); MapsFp[Nmap].x[1]= (double)l;
    getlon(&l); MapsFp[Nmap].x[2]= (double)l;
    getlon(&l); MapsFp[Nmap].y[0]= (double)l;
    getlon(&l); MapsFp[Nmap].y[1]= (double)l;
    getlon(&l); MapsFp[Nmap].y[2]= (double)l;
    getlon(&l); brushmethod=(char)l;
    getlon(&firstframe); getlon(&framedelay); getlon(&lastframe);
    getlon(&l); getlon(&l); getlon(&l); getlon(&l); //r
    getlon(&l); getlon(&l); getlon(&l); getlon(&l); //b
    getlon(&l); getlon(&l); getlon(&l); getlon(&l); //t
    getlon(&l); /* used by skeleton for br. assign. not used by renderer */
    getlon(&l); brushmaptype=(char)l;
    getlon(&brushangle);
    for(k=0;k<4;k++)getlon(&m); /* dump the brush lock info */
    fread(tempstr,1,32,fp);
    fread(&p,1,1,fp); fread(&r,1,1,fp); fread(&b,1,1,fp); fread(&t,1,1,fp);
    fread(&kt,1,1,fp);
    getsht(&s); MapsFp[Nmap].pp=s; /* these will be set to -1 */
    getsht(&s); MapsFp[Nmap].rp=s; /* if a map of this type   */
    getsht(&s); MapsFp[Nmap].bp=s; /* is not used in Material */
    getsht(&s);
    fread(t_colour,1,3,fp);  /* not used yet */
    fread(t_colour,1,3,fp);
    fread(t_colour,1,3,fp);  /* ------------ */
    fread(t_colour,1,3,fp);  /* key colour   */
    fread(fsurf,1,256,fp);
    fread(frefl,1,256,fp);
    fread(fbump,1,256,fp);
    fread(tempstr,1,256,fp);
    moziac=NO;
    if(brushmaptype == PLANE_MOZIAC)   { brushmaptype = PLANE;    moziac=YES;}
    if(brushmaptype == CYLINDER_MOZIAC){ brushmaptype = CYLINDER; moziac=YES;}
    if(p && brushmethod == ANIMBRUSH){ /* animated brush calculate frame */
      cp1=strchr(fsurf,'0');
      cp2=strchr(fsurf,'.');
      if(cp1 == NULL || cp2 == NULL){
        sprintf(tempstr,"dummy.evi");
      }
      else{
        int ln;
        char *cp,extn[16],fmat[16];
        strcpy(extn,cp2);
        ln=0; cp=cp1; while(cp != cp2){ln++; cp++; if(ln > 8)break;}
        *cp1='\0';
        if(framedelay == 0)thisframe=frame+firstframe-1;
        else               thisframe=frame/(framedelay+1)+firstframe;
        while(thisframe > lastframe)thisframe -= lastframe;
        sprintf(fmat,"%%s%%0%ldld%%s",ln);
        sprintf(tempstr,fmat,fsurf,thisframe,extn);
      }
      strcpy(fsurf,tempstr);
    }
    if(p) strcpy(MapsFp[Nmap].filename,fsurf);
    else         MapsFp[Nmap].pp = -1;
    if(r) strcpy(MapsFp[Nmap].frefl,frefl);
    else         MapsFp[Nmap].rp = -1;
    if(b) strcpy(MapsFp[Nmap].fbump,fbump);
    else         MapsFp[Nmap].bp = -1;
    MapsFp[Nmap].map=(short)brushmaptype;
    MapsFp[Nmap].method=(short)brushmethod;
    MapsFp[Nmap].angle=360.0/(double)brushangle;
    MapsFp[Nmap].moziac=moziac;
    Nmap++;
  }
  return;
}

static void ReadModWireframe(long clen){
 long i,ne,np;
 getlon(&np);  getlon(&ne);
 i=((np*3)+(ne*2)+2)*sizeof(long);
 if(clen != i){   /* if chunk is void */
   getchunk(clen-2);
   return;
 }
 w_frame.p = (point *)EVI_Malloc(np*sizeof(point));
 w_frame.e = (twoLongs *)EVI_Malloc(ne*sizeof(twoLongs));
 if(w_frame.p == NULL || w_frame.e == NULL){
   if(w_frame.p != NULL)EVI_Free(w_frame.p); w_frame.p=NULL;
   if(w_frame.e != NULL)EVI_Free(w_frame.e); w_frame.e=NULL;
   getchunk(clen-2);
   return;
 }
 w_frame.Np=np;  w_frame.Ne=ne;
 for(i=0;i<w_frame.Np;i++){
   getlon(&w_frame.p[i][0]);
   getlon(&w_frame.p[i][1]);
   getlon(&w_frame.p[i][2]);
 }
 for(i=0;i<w_frame.Ne;i++){
   getlon(&w_frame.e[i][0]);
   getlon(&w_frame.e[i][1]);
 }
}

void ListAdjFaces(void){
 FACE *f;
 long i,j,*fl,id,n,m,id0,id1,id2,nl[3],ml[3],k,kk,fjd,fkd;
 for(i=0;i<Nface;i++){ /* make list of work faces attached to each vertex */
   f=(MainFp+i);
   for(j=0;j<3;j++){
     id=f->V[j];
     fl=(long *)((MainVp+id)->flist);
     n=(MainVp+id)->nf;
     if(fl == NULL){
       if((fl=(long *)EVI_Malloc(sizeof(long))) == NULL)return;
     }
     else{
       if((fl=(long *)EVI_Realloc(fl,(n+1)*sizeof(long))) == NULL)return;
     }
     fl[n]=i;
     (MainVp+id)->flist=(int *)(fl);
     (MainVp+id)->nf=n+1;
   }
 }
 for(i=0;i<Nface;i++){ /* find faces that are adjacent to face i */
   f=(MainFp+i);
   id0=f->V[0];   id1=f->V[1];   id2=f->V[2];
   nl[0]=id0; ml[0]=id1;
   nl[1]=id1; ml[1]=id2;
   nl[2]=id2; ml[2]=id0;
   for(kk=0;kk<3;kk++){
     f->A[kk] = -1;
     if((n=(MainVp+nl[kk])->nf) > 0 && (m=(MainVp+ml[kk])->nf) > 0){
       for(j=0;j<n;j++){
         fjd=MainVp[nl[kk]].flist[j];
         for(k=0;k<m;k++){
           fkd=MainVp[ml[kk]].flist[k];
           if(fkd == fjd && fkd != i){
             f->A[kk]=fkd;
             goto ENDF;
           }
         }
       }
     }
     ENDF:;
   }
 }
 return;
}

static void MakeSameOrientation(long last, long id){
 FACE *f;
 long i0,i1,i2,l0,l1,l2,a0,a1,a2;
 f=(MainFp+last); l0=f->V[0]; l1=f->V[1]; l2=f->V[2];
 f=(MainFp+id);   i0=f->V[0]; i1=f->V[1]; i2=f->V[2];
                  a0=f->A[0]; a1=f->A[1]; a2=f->A[2];
 if((i2 == l2 && i1 == l1) ||
    (i2 == l1 && i1 == l0) ||
    (i2 == l0 && i1 == l2)){
   f->V[2]=i1; f->V[1]=i2; f->A[2]=a0; f->A[0]=a2;
   return;
 }
 if((i0 == l2 && i2 == l1) ||
    (i0 == l1 && i2 == l0) ||
    (i0 == l0 && i2 == l2)){
   f->V[0]=i2; f->V[2]=i0; f->A[0]=a1; f->A[1]=a0;
   return;
 }
 if((i1 == l1 && i0 == l0) ||
    (i1 == l2 && i0 == l1) ||
    (i1 == l0 && i0 == l2)){
   f->V[1]=i0; f->V[0]=i1; f->A[1]=a2; f->A[2]=a1;
   return;
 }
 return;
}

static void SetConnectingID(long last, long id){
 if((MainFp+id)->start == 1)return;
 (MainFp+id)->start=1;
 if(last >= 0)MakeSameOrientation(last,id);
 last=id;
 if((MainFp+id)->A[0] >= 0)SetConnectingID(last,(MainFp+id)->A[0]);
 if((MainFp+id)->A[1] >= 0)SetConnectingID(last,(MainFp+id)->A[1]);
 if((MainFp+id)->A[2] >= 0)SetConnectingID(last,(MainFp+id)->A[2]);
 return;
}

void FixUpOrientation(void){
 long i;
 while(1){
   for(i=0;i<Nface;i++){
     if((MainFp+i)->start == 0)goto PROCESSIT;
   }
   break;
   PROCESSIT:
   SetConnectingID(-1,i);
 }
 for(i=0;i<Nface;i++)(MainFp+i)->start=0;
}

