/* file SIMLIST.c  */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>

#include "sim.h"

#define PLANE           0
#define CYLINDER        1
#define PLANE_MOZIAC    2
#define CYLINDER_MOZIAC 3
#define MAP_BY_VERTEX   4

extern VERTEX  *MainVp;
extern FACE    *MainFp;
extern WFRAME  w_frame;
extern MAP     *MapsFp;
extern long    Nvert,Nface,NvertGlue,Nmap;
extern int     listID;
extern BOOL    Mapped,Cmaterial,smooth_shading;
extern GLfloat ObjScale;
extern GLubyte checkImage[checkImageWidth][checkImageHeight][3];

extern void TryCurrentDirect(char *name);
extern void makeCheckeredImageMap(void);

static void makeNextMap(unsigned char *pixels24, int x, int y);
static unsigned char *LoadTGA(char *TGAfile, int *xx, int *yy);
static unsigned char *LoadGIF(char *GIFfile, int *xx, int *yy);
static void GetMapNormal(int type, vector P, vector X, vector Y,
                         vector mP, vector mX, vector mY,
                         vector mN);
static void GetMappingCoordP(vector n, vector x, vector y, vector p0,
                             long *pv,/* point of intersection */
                             GLfloat *a, GLfloat *b);
static void GetMappingCoordC(vector n, vector x, vector y, vector p0,
                double angle,
                long *pv,
                GLfloat *a, GLfloat *b);
static unsigned char *LoadMAP(char *CurrentFile, int *xx, int *yy);

static GLfloat scale;
static  point c,pp[8];

static BOOL same_color(unsigned char *color){
 static unsigned char lastcolor[3]={0,0,0};
 if(abs(color[0] - lastcolor[0]) > 10 ||
    abs(color[1] - lastcolor[1]) > 10 ||
    abs(color[2] - lastcolor[2]) > 10 ){
   VECCOPY(color,lastcolor)
   return FALSE;
 }
 return TRUE;
}

static BOOL normalise(GLfloat *v){
 GLfloat n,nn;
 n= (GLfloat)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-10)return FALSE;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE(n,v,v)
 return TRUE;
}

BOOL Normalize(point x, point y, point z, GLfloat *n){
 GLfloat dz[3],dy[3];
 VECSUB((GLfloat)y,(GLfloat)x,dy)
 VECSUB((GLfloat)z,(GLfloat)x,dz)
 CROSS(dy,dz,n)
 if(normalise(n))return TRUE;
 return FALSE;
}

static double GetScaleValue(point c){
 int i;
 int size,xmax,ymax,zmax,xmin,ymin,zmin;
 xmax=ymax=zmax=size= -MAXUNIT;
 xmin=ymin=zmin=MAXUNIT;
 if(Nvert < 1)return 1.0;
 for(i=0;i<Nvert;i++){
   if((MainVp+i)->p[0] > xmax)xmax=(MainVp+i)->p[0];
   if((MainVp+i)->p[1] > ymax)ymax=(MainVp+i)->p[1];
   if((MainVp+i)->p[2] > zmax)zmax=(MainVp+i)->p[2];
   if((MainVp+i)->p[0] < xmin)xmin=(MainVp+i)->p[0];
   if((MainVp+i)->p[1] < ymin)ymin=(MainVp+i)->p[1];
   if((MainVp+i)->p[2] < zmin)zmin=(MainVp+i)->p[2];
 }
 size=max(xmax-xmin,ymax-ymin);
 size=max(size,zmax-zmin);
 c[0]=(xmax+xmin)*0.5;
 c[1]=(ymax+ymin)*0.5;
 c[2]=(zmax+zmin)*0.5;
 pp[0][0]=xmin; pp[0][1]=ymin; pp[0][2]=zmin;
 pp[1][0]=xmax; pp[1][1]=ymin; pp[1][2]=zmin;
 pp[2][0]=xmax; pp[2][1]=ymax; pp[2][2]=zmin;
 pp[3][0]=xmin; pp[3][1]=ymax; pp[3][2]=zmin;
 pp[4][0]=xmin; pp[4][1]=ymin; pp[4][2]=zmax;
 pp[5][0]=xmax; pp[5][1]=ymin; pp[5][2]=zmax;
 pp[6][0]=xmax; pp[6][1]=ymax; pp[6][2]=zmax;
 pp[7][0]=xmin; pp[7][1]=ymax; pp[7][2]=zmax;
 return ((double)4.0/(double)size);
}

void DrawWireFrame(void){
 GLfloat x,y,z;
 int i,i1,i2;
 glPushMatrix();
 glBegin(GL_LINES);
 glColor3f(1.0, 1.0, 1.0);
 if(w_frame.Np == 0 || w_frame.Ne == 0 ||
    w_frame.p == NULL || w_frame.e == NULL){
   static twoLongs e[12]={{0,1},{1,2},{2,3},{3,0},
                          {4,5},{5,6},{6,7},{7,4},
                          {0,4},{1,5},{2,6},{3,7}};
   for(i=0;i<12;i++){
     i1=e[i][0]; i2=e[i][1];
     x=(GLfloat)(pp[i1][0]-c[0])*scale;
     y=(GLfloat)(pp[i1][1]-c[1])*scale;
     z=(GLfloat)(pp[i1][2]-c[2])*scale;
     glVertex3f(x,z,-y);
     x=(GLfloat)(pp[i2][0]-c[0])*scale;
     y=(GLfloat)(pp[i2][1]-c[1])*scale;
     z=(GLfloat)(pp[i2][2]-c[2])*scale;
     glVertex3f(x,z,-y);
   }
 }
 else{
   for(i=0;i<w_frame.Ne;i++){
     i1=w_frame.e[i][0]; i2=w_frame.e[i][1];
     x=((GLfloat)(w_frame.p[i1][0]-c[0]))*scale;
     y=((GLfloat)(w_frame.p[i1][1]-c[1]))*scale;
     z=((GLfloat)(w_frame.p[i1][2]-c[2]))*scale;
     glVertex3f(x,z,-y);
     x=((GLfloat)(w_frame.p[i2][0]-c[0]))*scale;
     y=((GLfloat)(w_frame.p[i2][1]-c[1]))*scale;
     z=((GLfloat)(w_frame.p[i2][2]-c[2]))*scale;
     glVertex3f(x,z,-y);
   }
 }
 glEnd();
 glPopMatrix();
 glFlush();
 return;
}

void Make3dListS(void){
 GLfloat x,y,z,n[3],color[4],*spec_color,
         glos_color[]={1.0,1.0,1.0,1.0},
         matt_color[]={0.0,0.0,0.0,0.0},
         shiny[]={120.0},dot;
 FACE *fp;
 VERTEX *v,*v1,*v2,*v0;
 normal *nv;
 int i,j,k,Vi,V[3],jf,jl,count,strip;
 if(Nface == 0)return;
 scale=GetScaleValue(c)*ObjScale;
 if((nv=(normal *)EVI_Malloc(sizeof(normal)*Nvert)) == NULL){
   MessageBeep(MB_OK); return;
 }
 for(i=0;i<Nvert;i++)nv[i][0]=nv[i][1]=nv[i][2]=0.0;
 fp=MainFp; for(i=0;i<Nface;i++){
   V[0]=fp->V[0]; V[1]=fp->V[1]; V[2]=fp->V[2];
   v0=(MainVp+V[0]); v1=(MainVp+V[1]); v2=(MainVp+V[2]);
   if(fp->texture >> 7 != 0 && /* smoothed */
      Normalize(v0->p,v1->p,v2->p,n)){
     for(j=0;j<3;j++){
       dot=DOT(nv[V[j]],n);
       if(fabs(dot) < 1.e-10){
         VECCOPY(n,nv[V[j]])
       }
       else if(dot >  0.5){
         VECSUM(nv[V[j]],n,nv[V[j]])
         normalise(nv[V[j]]);
       }
       else if(dot < -0.5){
         VECSUB(nv[V[j]],n,nv[V[j]])
         normalise(nv[V[j]]);
       }
     }
   }
   fp++;
 }
// if(listID > 0){glDeleteLists(listID,listID);listID=0;}
 color[3]=1.0;
// listID=1;
// glNewList(listID,GL_COMPILE);
 if(Cmaterial)glEnable(GL_COLOR_MATERIAL);
 count=0;

 glBegin(GL_TRIANGLES);
 glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
 glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,glos_color);
 fp=MainFp; for(i=0;i<Nface;i++,fp++){
   if(Mapped && ((fp->texture & 0x40) == 0x40))continue;
   v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);
   if((fp->texture & 0x1f ) !=  0x4){ /* glass */
     if(!same_color(fp->color)){
       color[0]=(GLfloat)fp->color[0]/255.0;
       color[1]=(GLfloat)fp->color[1]/255.0;
       color[2]=(GLfloat)fp->color[2]/255.0;
       if(Cmaterial)glColor4fv(color);
       else{
         glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,color);
         if((fp->texture & 0x20) == 0x20)spec_color=glos_color;
         else                            spec_color=matt_color;
         glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec_color);
       }
     }
     if(Normalize(v0->p,v1->p,v2->p,n)){
       for(j=0;j<3;j++){
         Vi=fp->V[j]; v=(MainVp+Vi);
         x=((GLfloat)(v->p[0]-c[0]))*scale;
         y=((GLfloat)(v->p[1]-c[1]))*scale;
         z=((GLfloat)(v->p[2]-c[2]))*scale;
         if(smooth_shading && ((fp->texture >> 7) != 0)){ /* smoothed */
           dot=DOT(nv[Vi],n);
           if(fabs(dot) < 0.5)glNormal3f(n[0],n[2],-n[1]);
           else if(dot  < 0.0)glNormal3f(-nv[Vi][0],-nv[Vi][2], nv[Vi][1]);
           else               glNormal3f( nv[Vi][0], nv[Vi][2],-nv[Vi][1]);
         }
         else{/* not smoothed */
           glNormal3f( n[0], n[2],-n[1]);
         }
         glVertex3f(x,z,-y);
         count++;
       }
     }
   }
 }
 glEnd();
 if(Mapped){
   BOOL refmap;
   int k,brushID;
   vector mN,mX,mY,mP;
   GLfloat alpha,beta;
   unsigned char *Image;
   int xm,ym;
   if(Nmap > 0)for(k=0;k<Nmap;k++){
     if(MapsFp[k].pp < 0){
       Image=LoadMAP(MapsFp[k].frefl,&xm,&ym);
       glEnable(GL_TEXTURE_GEN_S);
       glEnable(GL_TEXTURE_GEN_T);
//       glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
       refmap=TRUE;
     }
     else{
       Image=LoadMAP(MapsFp[k].filename,&xm,&ym);
       refmap=FALSE;
     }
     if(Image != NULL){
       makeNextMap(Image,xm,ym);
     }
     else makeCheckeredImageMap();
     GetMapNormal(MapsFp[k].map,MapsFp[k].p,MapsFp[k].x,MapsFp[k].y,
                  mP,mX,mY,mN);
     glEnable(GL_TEXTURE_2D);
     glBegin(GL_TRIANGLES);

     color[0]=1.6; color[1]=1.6; color[2]=1.6;
     if(Cmaterial)glColor4fv(color);
     else{
       glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,color);
        if((fp->texture & 0x20) == 0x20)spec_color=glos_color;
        else                            spec_color=matt_color;
        glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec_color);
     }

     fp=MainFp; for(i=0;i<Nface;i++,fp++){
       if((fp->texture & 0x40) != 0x40)continue; /* if not mapped */
       brushID=(fp->brush & 0x1f);
       if(brushID != k)continue;         /* skip if not this map */
       v0=(MainVp+fp->V[0]); v1=(MainVp+fp->V[1]); v2=(MainVp+fp->V[2]);

       if(Normalize(v0->p,v1->p,v2->p,n)){
         for(j=0;j<3;j++){
           Vi=fp->V[j]; v=(MainVp+Vi);
           x=((GLfloat)(v->p[0]-c[0]))*scale;
           y=((GLfloat)(v->p[1]-c[1]))*scale;
           z=((GLfloat)(v->p[2]-c[2]))*scale;
           if(smooth_shading && ((fp->texture >> 7) != 0)){ /* smoothed */
             dot=DOT(nv[Vi],n);
             if(fabs(dot) < 0.5)glNormal3f(n[0],n[2],-n[1]);
             else if(dot  < 0.0)glNormal3f(-nv[Vi][0],-nv[Vi][2], nv[Vi][1]);
             else               glNormal3f( nv[Vi][0], nv[Vi][2],-nv[Vi][1]);
           }
           else{/* not smoothed */
             glNormal3f( n[0], n[2],-n[1]);
           }
           if(MapsFp[k].map == MAP_BY_VERTEX){
             alpha=(GLfloat)v->x;
             beta=(GLfloat)v->y;
           }
           else if(MapsFp[k].map == CYLINDER ||
                   MapsFp[k].map == CYLINDER_MOZIAC)
                GetMappingCoordC(mN,mX,mY,mP,MapsFp[k].angle,
                                             v->p,&alpha,&beta);
           else GetMappingCoordP(mN,mX,mY,mP,v->p,&alpha,&beta);
           if(!refmap)glTexCoord2f(alpha,beta);
           glVertex3f(x,z,-y);
         }
       }
     }
     glEnd();
     glDisable(GL_TEXTURE_2D);
     if(refmap){
       glDisable(GL_TEXTURE_GEN_S);
       glDisable(GL_TEXTURE_GEN_T);
//       glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
     }
   }
 }

 if(Cmaterial)glDisable(GL_COLOR_MATERIAL);
// glEndList();
 EVI_Free(nv);
}

#define TOL 1000.0

static void GetMapNormal(int type, vector P, vector X, vector Y,
                         vector mP, vector mX, vector mY,
                         vector mN){
 GLfloat n[3];
 if(type == CYLINDER || type == CYLINDER_MOZIAC){
   VECCOPY((double)P,mP)
   VECSUB((double)X,mP,mX)
   VECSUB((double)Y,mP,mY)
   CROSS(mY,mX,mN)
 }
 else{                                      /* arranged this way to */
   VECCOPY((double)Y,mP)                    /* have map with origin */
   VECSUB((double)X,(double)P,mX)           /* at bottom left       */
   VECSUB((double)P,mP,mY)
   CROSS(mX,mY,mN)
 }
 VECCOPY((GLfloat)mN,n)
 if(!normalise(n)){
   n[0]=1.0; n[1]=0.0; n[2]=0.0;
 }
 VECCOPY((double)n,mN)
 return;
}

static void GetMappingCoordP(vector n, vector x, vector y, vector p0,
                             long *pv,/* point of intersection */
                             GLfloat *a, GLfloat *b){
 vector v1,p;
 GLfloat mu,det,ve1,ve2,ve3,vp1,vp2,vp3,p1,p2,p3;
 VECSUB(p0,(GLfloat)pv,v1)
 mu=DOT(n,v1);
 VECSCALE(mu,n,v1)
 VECSUM((GLfloat)pv,v1,p)
 ve1=x[0]; ve2=x[1]; ve3=x[2];
 vp1=y[0]; vp2=y[1]; vp3=y[2];
 p1=p[0]-p0[0];
 p2=p[1]-p0[1];
 p3=p[2]-p0[2];
 det=ve1*vp2-vp1*ve2;
 if(det > TOL || det < -TOL)  /* e-10 appears not to work */
   {  *a=( vp2*p1-vp1*p2)/det;
      *b=(-ve2*p1+ve1*p2)/det;
      goto HIT;
   }
 det=ve1*vp3-vp1*ve3;
 if(det > TOL || det < -TOL)
   {  *a=( vp3*p1-vp1*p3)/det;
      *b=(-ve3*p1+ve1*p3)/det;
      goto HIT;
   }
 det=ve2*vp3-vp2*ve3;
 if(det > TOL || det < -TOL)
   {  *a=( vp3*p2-vp2*p3)/det;
      *b=(-ve3*p2+ve2*p3)/det;
      goto HIT;
   }
 *a = *b = 0.0;
 HIT:
 return;
}

#define RTD 0.159154943

static void GetMappingCoordC(vector n, vector x, vector y, vector p0,
                double angle,
                long *pv,
                GLfloat *a, GLfloat *b){
 GLfloat theta,nn;
 GLfloat p1[3],p2[3],q[3];
 VECSUB((GLfloat)p0,(GLfloat)pv,p1)
 *b=DOT(p1,y)/DOT(y,y);
 VECSCALE(*b,y,p2)
 VECSUB(p1,p2,q)
 normalise(q);
 nn=DOT(q,(GLfloat)n);
 if     (nn >  1.0)nn =  1.0;
 else if(nn < -1.0)nn = -1.0;
 theta=acos(nn)*RTD;
 CROSS(q,n,p1)
 if(DOT(p1,y) <= 0.0){
   if(theta <= 0.25) *a = (0.25 - theta);
   else  *a = (1.25 - theta);
 }
 else    *a = (0.25 + theta);
 *a *= (GLfloat)angle;
 return;
}

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size;

static int rowcount;
static unsigned char *pixels = NULL;

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(y_size - 1 - rowcount)*(long)(x_size);
 memcpy(pixels + l, linepixels, linelen);
 rowcount++;
 return 0;
}

static void makeNextMap(unsigned char *pixels24, int x, int y){
 int xi,yi;
 unsigned char *px;
 if     (x <  96)xi=32;
 else if(x < 320)xi=64;
 else xi=128;
 if     (y <  96)yi=32;
 else if(y < 320)yi=64;
 else yi=128;
 if((px=(unsigned char *)EVI_Malloc(xi*yi*3)) == NULL){
   EVI_Free((char *)pixels24);
   return;
 }
 gluScaleImage(GL_RGB,x,y,GL_UNSIGNED_BYTE,pixels24,xi,yi,GL_UNSIGNED_BYTE,px);
 EVI_Free((char *)pixels24);
 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glTexImage2D(GL_TEXTURE_2D,0,3,xi,yi,0,GL_RGB,GL_UNSIGNED_BYTE,(GLvoid *)px);
 EVI_Free((char *)px);
}

static unsigned char *LoadMAP(char *CurrentFile, int *xx, int *yy){
 FILE *ff;
 if((ff=fopen(CurrentFile,"r")) == NULL){
   TryCurrentDirect(CurrentFile);
 }
 else fclose(ff);
 if     (strstr(CurrentFile,".GIF") || strstr(CurrentFile,".gif"))
   return LoadGIF(CurrentFile,xx,yy);
 else if(strstr(CurrentFile,".TGA") || strstr(CurrentFile,".tga"))
   return LoadTGA(CurrentFile,xx,yy);
 return NULL;
}

static unsigned char *LoadGIF(char *GIFfile, int *xx, int *yy){
 unsigned char *px,*lpB,*pixels24;
 long i,j,k,imagesize,x,y,xi,yi;
 rowcount=0;
 if(read_gif_image(GIFfile,1) < 0)return NULL;
 x=x_size; y=y_size;
 imagesize=x_size*y_size;
 if((pixels=(unsigned char*)EVI_Malloc(imagesize)) == NULL)return NULL;
 if((pixels24=(unsigned char*)EVI_Malloc(imagesize*3)) == NULL){
   EVI_Free(pixels);
   return NULL;
 }
 if(read_gif_image(GIFfile,0) < 0){
   EVI_Free(pixels);
   return NULL;
 }
 px=pixels24;
 lpB=pixels;
 for(j=0;j<y;j++){
   for(i=0;i<x;i++){
     k=(long)(*lpB++);
     *px++ = (unsigned char)(p_red[k]   << 2);
     *px++ = (unsigned char)(p_green[k] << 2);
     *px++ = (unsigned char)(p_blue[k]  << 2);
   }
 }
 *xx=x; *yy=y;
 EVI_Free(pixels);
 return pixels24;
}

static void tga_display_line(unsigned char *prgb,
                             long wide, long line,
                             long upsidedown, long hi){
 register long l,x,k;
 if(upsidedown)l = 3L*(long)(line)*(long)(x_size);
 else          l = 3L*(long)(hi - 1 - line)*(long)(x_size);
 memcpy(pixels + l, prgb, (int)(wide*3));
}

static short R_getvariword(long n, FILE *f, long *vw){
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;          if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

static short read_tga_image(char *filename, short info_only){
 char c0,c1,c2;
 long i,j,k,l,m,n,cmo,cml,xor,yor,wi,hi,iok=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat,wi3;
 FILE *t24;
 unsigned char *prgb;
 prgb=NULL;
 if((t24=fopen(filename,"rb")) != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0 : 1 ;
   if(info_only){
     x_size=(int)wi;
     y_size=(int)hi;
     fclose(t24);
     return 1;
   }
   if((prgb=(unsigned char *)EVI_Malloc((size_t)wi*3)) == NULL)goto CLOSEOUT;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi;
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   wi3=wi*3;
   if(image_type == 2){ /* uncompressed */
     m=0; k=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       prgb[m+2] = (char)( vw        & pat) << s;
       prgb[m+1] = (char)((vw >> s1) & pat) << s;
       prgb[m  ] = (char)((vw >> s2) & pat) << s;
       k++; m+=3;
       if(m == wi3){
         tga_display_line(prgb,wi,i,upsidedown,hi); m=0;
       }
     }
   }
   else{  /* compressed */
     m=0; l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           prgb[m+2] = (char)( vw        & pat) << s;
           prgb[m+1] = (char)((vw >> s1) & pat) << s;
           prgb[m  ] = (char)((vw >> s2) & pat) << s;
           k++; m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           prgb[m+2] = c0;
           prgb[m+1] = c1;
           prgb[m  ] = c2;
           k++;  m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
   if(prgb != NULL)EVI_Free((char *)prgb);
 }
 return (short)iok;
}

static unsigned char *LoadTGA(char *TGAfile, int *xx, int *yy){
 long imagesize;
 rowcount=0;
 if(read_tga_image(TGAfile,1) < 0)return NULL;
 imagesize=x_size*y_size*3;
 if((pixels=(unsigned char*)EVI_Malloc(imagesize)) == NULL)return NULL;
 if(read_tga_image(TGAfile,0) < 0){
   EVI_Free(pixels);
   return NULL;
 }
 *xx=x_size; *yy=y_size;
 return pixels;
}

