/* Win32 Win32s  viewer for GIF / TGA / DIB /DDB  images */

#include <windows.h>
#include <windowsx.h>
#include <io.h>
#include <stdio.h>
#include <string.h>
#include "dib.h"
#include "viewrc.h"
#include "\wing\include\wing.h"

#define FI_DIB 0
#define FI_GIF 1
#define FI_TGA 2
#define PALVERSION  0x300
#define MAXPALETTE  256
#define MAXTEXT     256

extern short read_gif_image(char *, short);
extern unsigned char p_red[],p_green[],p_blue[];
extern int colourmap_size,x_size,y_size;

HMENU     hMenu          = NULL;
HWND      hWndApp        = NULL;
HINSTANCE hInst          = NULL;
HPALETTE  hpalCurrent    = NULL;         /* Handle to current palette            */
HBITMAP   hbmCurrent     = NULL;         /* Handle to current memory BITMAP      */
RGBQUAD   *pRgb          = NULL;
LPSTR     lpBitmapHeader = NULL;
LPSTR     lpBitmapBits   = NULL;

HPALETTE hpalApp=NULL;
static PDIB pdibHalftone=NULL;

#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define SWAP(x,y)   ((x)^=(y)^=(x)^=(y))
#define MIN(x,y) (((x) <= (y)) : x ? y)
#define StartWait() hcurSave = SetCursor(LoadCursor(NULL,IDC_WAIT))
#define EndWait()   SetCursor(hcurSave)
/* macro to determine if resource is a DIB */
#define ISDIB(bft) ((bft) == BFT_BITMAP)
/* Macro to align given value to the closest DWORD (unsigned long ) */
#define ALIGNULONG(i)   ((i+3)/4*4)
/* Macro to determine to round off the given value to the closest byte */
//#define WIDTHBYTES(i)   ((i+31)/32*4)

static SYSTEM_INFO sys_info;
static int DisplayWidth,DisplayHeight,DisplayBits,DisplayPlanes,
           xScreen,yScreen,rowcount,x4_size,file_format=FI_GIF;
static LPSTR pixels;
static char CurrentDirFile[256],CurrentDirectory[256],CurrentFile[256];
static POINT ptSize;
static DWORD fdwStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
                        WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_THICKFRAME;
static HCURSOR hcurSave;

PDIB DibHalftoneDIB( PDIB pSource );
static int identify_file_format(void);
static int ErrMsg(char *sz,...);
static void PutMsg(char *s);
BOOL CALLBACK AboutDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam);
LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,LPARAM lParam);
BOOL MenuCommand(HWND hWnd, WORD id);
static int identify_file_format(void);
static void FreeDib(void);
static void SizeWindow(HWND hWnd);
static void SetScrollRanges(HWND hwnd);
static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi);
static WORD NumColours(LPBITMAPINFOHEADER lpbi);
static void GetRealClientRect(HWND hwnd, PRECT lprc);
static void AppPaint(HWND hWnd, HDC hDC, int x, int y);
static HBITMAP rBitmapFromDib(LPSTR lpHeader,LPSTR lpBits,HPALETTE hpal);

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow){
 MSG msg;
 HWND hDesktopWnd,hWndPrev,hWnd;
 HDC hDCcaps;
 WNDCLASS wndclass;
 hWndPrev = FindWindow("OFX:ViewClass", NULL);
 if (hWndPrev != NULL){
   BringWindowToTop(hWndPrev);
   return FALSE;
 }
 hInst       = hInstance;
 hDesktopWnd = GetDesktopWindow();
 hDCcaps     = GetDC(hDesktopWnd);
 DisplayWidth= GetDeviceCaps(hDCcaps,HORZRES);
 DisplayHeight= GetDeviceCaps(hDCcaps,VERTRES);
 DisplayBits= GetDeviceCaps(hDCcaps,BITSPIXEL);
 DisplayPlanes= GetDeviceCaps(hDCcaps,PLANES);
 ReleaseDC(hDesktopWnd,hDCcaps);
 GetSystemInfo(&sys_info);
 if(!hPrevInstance){
   wndclass.style         = CS_DBLCLKS;
   wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
   wndclass.cbClsExtra    = 0 ;
   wndclass.cbWndExtra    = 0 ;
   wndclass.hInstance     = hInstance ;
   wndclass.hIcon         = LoadIcon(hInst, "VIEWICON");
   wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
   wndclass.hbrBackground = GetStockObject (WHITE_BRUSH) ;
   wndclass.lpszMenuName  = "ShowMenu" ;
   wndclass.lpszClassName = "OFX:ViewClass";
   if (!RegisterClass (&wndclass))  return FALSE ;
 }
 strcpy(CurrentDirFile,(char *)lpCmdLine);
 xScreen = GetSystemMetrics (SM_CXSCREEN) ;
 yScreen = GetSystemMetrics (SM_CYSCREEN) ;
 hWnd = CreateWindow(
        "OFX:ViewClass",
        "OpenFX Viewer",
        fdwStyle,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        xScreen / 2,
        yScreen / 2,
        NULL,
        NULL,
        hInstance,
        NULL) ;
 if(hWnd == NULL)return FALSE;
 ShowWindow (hWndApp = hWnd, nCmdShow);
 { HFONT hFont,hPrevFont;
   HDC hFontDC;
   hFontDC= GetDC(hWndApp);
   hFont= GetStockObject(SYSTEM_FONT);
   hPrevFont= SelectObject(hFontDC,hFont);
   SetTextColor(hFontDC,RGB(191,191,191));
   TextOut(hFontDC,10,10,"Display GIF and TGA images",26);
   TextOut(hFontDC,10,30,"created with Envisage 3D  ",26);
   SetTextColor(hFontDC,RGB(255,255,255));
   SelectObject(hFontDC,hPrevFont);
   ReleaseDC(hWndApp,hFontDC);
 }

 hpalApp = WinGCreateHalftonePalette();

 while(GetMessage (&msg, NULL, 0, 0)) {
   TranslateMessage (&msg);
   DispatchMessage (&msg);
 }

 if (hpalApp != NULL)DeleteObject(hpalApp);

 return msg.wParam ;
}

BOOL CALLBACK AboutDlgProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam){
 lparam = lparam;         /* turn off warning */
 switch( msg ) {
   case WM_INITDIALOG:
     return( TRUE );
   case WM_COMMAND:
      if( wparam == IDOK ) {
        EndDialog( hwnd, TRUE );
        return( TRUE );
      }
      break;
 }
 return( FALSE );
}

LRESULT CALLBACK MainWndProc(HWND hWnd,UINT iMessage,WPARAM wParam,
                             LPARAM lParam){
 PAINTSTRUCT ps;
 HDC hDC;
 int i,iPos,iMax,iMin,dn;
 RECT rc;
 switch (iMessage) {
   case WM_DESTROY:
     FreeDib();
     PostQuitMessage(0);
     break ;
  case WM_CREATE:
  case WM_WININICHANGE:
    hMenu = GetMenu(hWnd);
    break;
  case WM_COMMAND:
    return MenuCommand(hWnd,LOWORD(wParam));
    break;
  case WM_PAINT:
    hDC = BeginPaint(hWnd, &ps);
    AppPaint(hWnd,
       hDC,
       GetScrollPos(hWnd,SB_HORZ),
       GetScrollPos(hWnd,SB_VERT) );
    EndPaint(hWnd, &ps);
    break ;
  case WM_SIZE:
    SetScrollRanges(hWnd);
    break;
  case WM_KEYDOWN:
    switch (wParam) {
      case VK_UP:    PostMessage (hWnd, WM_VSCROLL, SB_LINEUP,   0L);  break;
      case VK_DOWN:  PostMessage (hWnd, WM_VSCROLL, SB_LINEDOWN, 0L);  break;
      case VK_PRIOR: PostMessage (hWnd, WM_VSCROLL, SB_PAGEUP,   0L);  break;
      case VK_NEXT:  PostMessage (hWnd, WM_VSCROLL, SB_PAGEDOWN, 0L);  break;
      case VK_HOME:  PostMessage (hWnd, WM_HSCROLL, SB_PAGEUP,   0L);  break;
      case VK_END:   PostMessage (hWnd, WM_HSCROLL, SB_PAGEDOWN, 0L);  break;
      case VK_LEFT:  PostMessage (hWnd, WM_HSCROLL, SB_LINEUP,   0L);  break;
      case VK_RIGHT: PostMessage (hWnd, WM_HSCROLL, SB_LINEDOWN, 0L);  break;
    }
    break;
  case WM_KEYUP:
    switch (wParam) {
      case VK_UP:
      case VK_DOWN:
      case VK_PRIOR:
      case VK_NEXT:
        PostMessage (hWnd, WM_VSCROLL, SB_ENDSCROLL, 0L);
        break;
      case VK_HOME:
      case VK_END:
      case VK_LEFT:
      case VK_RIGHT:
        PostMessage (hWnd, WM_HSCROLL, SB_ENDSCROLL, 0L);
        break;
    }
    break;
  case WM_VSCROLL:
    /* Calculate new vertical scroll position */
    GetScrollRange (hWnd, SB_VERT, &iMin, &iMax);
    iPos = GetScrollPos (hWnd, SB_VERT);
    GetClientRect (hWnd, &rc);
    switch (LOWORD(wParam)) {
      case SB_LINEDOWN:      dn =  rc.bottom / 16 + 1;    break;
      case SB_LINEUP:        dn = -rc.bottom / 16 + 1;    break;
      case SB_PAGEDOWN:      dn =  rc.bottom / 2  + 1;    break;
      case SB_PAGEUP:        dn = -rc.bottom / 2  + 1;    break;
      case SB_THUMBTRACK:
      case SB_THUMBPOSITION: dn = HIWORD(wParam)-iPos;    break;
      default:               dn = 0;
    }
    /* Limit scrolling to current scroll range */
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      ScrollWindow(hWnd, 0, -dn, NULL, NULL);
      SetScrollPos(hWnd, SB_VERT, iPos + dn, TRUE);
    }
    break;
  case WM_HSCROLL:
    /* Calculate new horizontal scroll position */
    GetScrollRange(hWnd, SB_HORZ, &iMin, &iMax);
    iPos = GetScrollPos(hWnd, SB_HORZ);
    GetClientRect (hWnd, &rc);
    switch (LOWORD(wParam)) {
      case SB_LINEDOWN:      dn =  rc.right / 16 + 1;     break;
      case SB_LINEUP:        dn = -rc.right / 16 + 1;     break;
      case SB_PAGEDOWN:      dn =  rc.right / 2  + 1;     break;
      case SB_PAGEUP:        dn = -rc.right / 2  + 1;     break;
      case SB_THUMBTRACK:
      case SB_THUMBPOSITION: dn = HIWORD(wParam) - iPos; break;
      default:               dn = 0;
    }
    /* Limit scrolling to current scroll range */
    if ((dn = BOUND (iPos + dn, iMin, iMax) - iPos) != 0) {
      ScrollWindow (hWnd, -dn, 0, NULL, NULL);
      SetScrollPos (hWnd, SB_HORZ, iPos + dn, TRUE);
    }
    break;
  default:
      return DefWindowProc (hWnd, iMessage, wParam, lParam) ;
 }
 return 0L ;
}

BOOL MenuCommand (HWND hWnd, WORD id){
 int i;
 char lpMessage[MAXTEXT],lpCaption[64];
 switch (id) {
   case IDM_ABOUT:
     DialogBox(hInst,MAKEINTRESOURCE(DLG_ABOUT),hWnd,(DLGPROC)AboutDlgProc);
     break;
   case IDM_DISPLAY:
       LoadString(hInst,IDS_DISPLAY,lpCaption,64);
     if     (DisplayWidth==640)
       LoadString(hInst,IDS_640,lpMessage,MAXTEXT);
     else if(DisplayWidth==800)
       LoadString(hInst,IDS_800,lpMessage,MAXTEXT);
     else if(DisplayWidth==1024)
       LoadString(hInst,IDS_1024,lpMessage,MAXTEXT);
     else
       LoadString(hInst,IDS_CUSTOM,lpMessage,MAXTEXT);
     MessageBox(GetFocus(),lpMessage,lpCaption,MB_OK);
     break;
   case IDM_COLOURS:
     LoadString(hInst,IDS_COLOUR,lpCaption,64);
     if(DisplayBits < 8)
       LoadString(hInst,IDS_COLOUR16,lpMessage,MAXTEXT);
     else if(DisplayBits==8)
       LoadString(hInst,IDS_COLOUR256,lpMessage,MAXTEXT);
     else if(DisplayBits==16)
       LoadString(hInst,IDS_COLOUR65,lpMessage,MAXTEXT);
     else if(DisplayBits==24)
       LoadString(hInst,IDS_COLOUR16M,lpMessage,MAXTEXT);
     else LoadString(hInst,IDS_CUSTOMC,lpMessage,MAXTEXT);
     MessageBox(GetFocus(),lpMessage,lpCaption,MB_OK);
     break;
   case IDM_MODE:
     sprintf(lpCaption,"Computer System");
     if(sys_info.dwProcessorType == PROCESSOR_INTEL_486)
       sprintf(lpMessage,"Intel 486 type processor");
     else if(sys_info.dwProcessorType == PROCESSOR_INTEL_386)
       sprintf(lpMessage,"Intel 386 type processor");
     else if(sys_info.dwProcessorType == PROCESSOR_ALPHA_21064)
       sprintf(lpMessage,"Alpha 21064 type processor");
     else if(sys_info.dwProcessorType == PROCESSOR_MIPS_R2000 ||
             sys_info.dwProcessorType == PROCESSOR_MIPS_R3000 ||
             sys_info.dwProcessorType == PROCESSOR_MIPS_R4000)
       sprintf(lpMessage,"MIPS type processor");
     else
       sprintf(lpMessage,"Unrecognised processor");
     MessageBox(GetFocus(),lpMessage,lpCaption,MB_OK);
     break;
   case IDM_GENERALHELP:
     LoadString(hInst,IDS_HELP,lpCaption,64);
     LoadString(hInst,IDS_HELPTEXT,lpMessage,MAXTEXT);
     MessageBox(GetFocus(),lpMessage,lpCaption,MB_OK);
     break;
   case IDM_OPEN: {
     OPENFILENAME ofn;
     char szTitle[60]="File to View",szFilter[80];
     CurrentDirFile[0]='\0'; CurrentFile[0]='\0';
     GetCurrentDirectory(sizeof(CurrentDirectory),CurrentDirectory);
     strcpy(szFilter,"8 bit GIF's(*.GIF)|*.gif|24 bit TGA's(*.TGA)|*.tga|");
     i=0; while(szFilter[i] != '\0'){
       if(szFilter[i] == '|')szFilter[i]='\0';
       i++;
     }
     memset(&ofn,0,sizeof(OPENFILENAME));
     ofn.lStructSize=sizeof(OPENFILENAME);
     ofn.hwndOwner=hWnd;
     ofn.lpstrFilter=szFilter;
     if     (file_format == FI_GIF)ofn.nFilterIndex=1;
     else if(file_format == FI_TGA)ofn.nFilterIndex=2;
     ofn.lpstrFile=CurrentDirFile;   /* will contain file + path */
     ofn.nMaxFile=sizeof(CurrentDirFile);
     ofn.lpstrFileTitle=CurrentFile;  /* file name only */
     ofn.nMaxFileTitle=sizeof(CurrentFile);
     ofn.lpstrInitialDir=CurrentDirectory;
     ofn.lpstrTitle=szTitle;
     ofn.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
     if(GetOpenFileName(&ofn)){
       StartWait();
       file_format=identify_file_format();
       switch(file_format){
         case FI_GIF:
           if(gifview(hWnd))InvalidateRect (hWnd, NULL, FALSE);
           break;
         case FI_TGA:
           if(tgaview(hWnd))InvalidateRect (hWnd, NULL, FALSE);
           break;
         default:
//           if (InitDIB(hWnd))InvalidateRect (hWnd, NULL, FALSE);
//           else              bLegitDraw = FALSE;
           break;
       }
       EndWait();
     }
     break;
   }
   case IDM_EXIT:
     PostMessage(hWnd, WM_SYSCOMMAND, SC_CLOSE, 0L);
     break;
   default:
     break;
 }
 return TRUE;
}

static void FreeDib(void){
 if (hpalCurrent) DeleteObject(hpalCurrent);
 if (hbmCurrent)  DeleteObject(hbmCurrent);
 if(lpBitmapHeader != NULL)free((char *)lpBitmapHeader);
 if(pdibHalftone != NULL)DibFree(pdibHalftone);
 pdibHalftone   = NULL;
 lpBitmapHeader = NULL;
 lpBitmapBits   = NULL;
 hpalCurrent    = NULL;
 hbmCurrent     = NULL;
 pRgb           = NULL;
}

static int identify_file_format(void){
 int ff=FI_DIB;
 strupr(CurrentFile);
 if     (strstr(CurrentFile,".GIF"))ff=FI_GIF;
 else if(strstr(CurrentFile,".TGA"))ff=FI_TGA;
 else if(strstr(CurrentFile,".DIB") ||
         strstr(CurrentFile,".BMP") ||
         strstr(CurrentFile,".RLE"))ff=FI_DIB;
 return ff;
}

static int ErrMsg(char *sz,...){
 char ach[128];
 va_list arg_ptr;
 va_start(arg_ptr,sz);
 vsprintf(ach,sz,arg_ptr);   /* Format the string */
 va_end(arg_ptr);
 MessageBox (NULL, ach, NULL, MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
 return FALSE;
}

static void PutMsg(char *s){
 MessageBox (NULL,s, NULL, MB_OK|MB_ICONEXCLAMATION|MB_APPLMODAL);
}

static void SizeWindow(HWND hWnd){
 char Title[120];
 RECT  Rectangle;
 int   dx,dy;
 POINT pt;
 LPBITMAPINFOHEADER bi;
 if((bi=(LPBITMAPINFOHEADER)lpBitmapHeader) == NULL)return;
 sprintf(Title,
          "%s (%s %ldx%ldx%ld)",
          "View",
          CurrentFile,
          (long)x_size,  //(long)bi->biWidth,
          (long)bi->biHeight,
          (long)bi->biBitCount);
 SetWindowText(hWnd,Title);
 ptSize.x = (WORD)bi->biWidth;
 ptSize.y = (WORD)bi->biHeight;
 if (IsZoomed(hWnd))SetScrollRanges(hWnd);
 else {
   Rectangle.left   = 0;
   Rectangle.top    = 0;
   Rectangle.right  = (WORD)bi->biWidth;
   Rectangle.bottom = (WORD)bi->biHeight;
   AdjustWindowRect (&Rectangle,fdwStyle,TRUE);
   SetWindowPos (hWnd, HWND_TOPMOST, 0, 0,
                 Rectangle.right  - Rectangle.left + 1,
                 Rectangle.bottom - Rectangle.top + 1,
                 SWP_NOMOVE);
 }
 InvalidateRect(hWnd,NULL,TRUE);
}

static void SetScrollRanges(HWND hwnd){
 RECT rc;
 int iRangeH,iRangeV,i;
 static int iSem = 0;
 if(!iSem){
   iSem++;
   GetRealClientRect(hwnd,&rc);
   for(i = 0; i < 2; i++){
     iRangeV = ptSize.y - rc.bottom;
     iRangeH = ptSize.x - rc.right;
     if(iRangeH < 0)iRangeH = 0;
     if(iRangeV < 0)iRangeV = 0;
     if (GetScrollPos(hwnd,SB_VERT)  > iRangeV ||
         GetScrollPos(hwnd, SB_HORZ) > iRangeH)
       InvalidateRect (hwnd, NULL, TRUE);
     SetScrollRange(hwnd, SB_VERT, 0, iRangeV, TRUE);
     SetScrollRange(hwnd, SB_HORZ, 0, iRangeH, TRUE);
     GetClientRect(hwnd, &rc);
   }
   iSem--;
 }
}

int out_line(unsigned char *linepixels, int linelen){
 register long l;
 l = (long)(y_size - 1 - rowcount)*(long)(x4_size);
 memcpy(pixels + l, linepixels, linelen);
 rowcount++;
 return 0;
}

int gifview(HWND hWnd){
 int i;
 char temp[128];
 long imagesize;
 LPSTR p;
 LPBITMAPINFOHEADER lpbi;
 FreeDib();
 rowcount=0;
 strcpy(temp,CurrentFile);
 if(read_gif_image(temp,1) < 0){
   ErrMsg("'%ls' is NOT a valid GIF file", (LPSTR)temp);
   return 0;
 }
 x4_size = ALIGNULONG(x_size);
 lpBitmapHeader = (LPSTR)malloc((long)sizeof(BITMAPINFOHEADER)
                                +256L*sizeof(RGBQUAD));
 if(lpBitmapHeader == NULL){
   ErrMsg("Out of memory");
   return 0;
 }
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x4_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=8;
 lpbi->biCompression=BI_RGB;
 imagesize = lpbi->biSizeImage = (DWORD)x4_size*(DWORD)y_size;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=256;
 lpbi->biClrImportant=256;
 pRgb=(RGBQUAD *)((LPSTR)lpbi + lpbi->biSize);
 for(i=0;i<256;i++){
   pRgb[i].rgbRed = p_red[i] << 2;
   pRgb[i].rgbGreen = p_green[i] << 2;
   pRgb[i].rgbBlue = p_blue[i] << 2;
 }
 hpalCurrent=CreateImagePalette(lpbi);
 p=(LPSTR)realloc(lpBitmapHeader,(long)sizeof(BITMAPINFOHEADER)
                                  +256L*sizeof(RGBQUAD)+imagesize);
 if(p == NULL){
   ErrMsg("Out of memory");
   FreeDib();
   return 0;
 }
 lpBitmapHeader=p;
 lpBitmapBits=(LPSTR)(p+(long)sizeof(BITMAPINFOHEADER)+256L*sizeof(RGBQUAD));
 memset(lpBitmapBits,0,imagesize);
 pixels=lpBitmapBits;
 if(read_gif_image(temp,0) < 0){
   PutMsg("Error reading GIF file");
   FreeDib();
   return 0;
 }
// hbmCurrent=rBitmapFromDib(lpBitmapHeader,lpBitmapBits,hpalCurrent);
// if(hbmCurrent == NULL)ErrMsg("Could not create bitmap");
 SizeWindow(hWnd);
 return 1;
}

static HPALETTE CreateImagePalette(LPBITMAPINFOHEADER lpbi){
 LOGPALETTE *pPal;
 HPALETTE   hpal = NULL;
 WORD       nNumColors;
 BYTE       red;
 BYTE       green;
 BYTE       blue;
 int        i;
 RGBQUAD    *pRgb;
 if(lpbi == NULL)return NULL;
 if(lpbi->biSize != sizeof(BITMAPINFOHEADER))return NULL;
 /* Get a pointer to the color table and the number of colors in it */
 pRgb = (RGBQUAD *)((LPSTR)lpbi + (WORD)lpbi->biSize);
 nNumColors = NumColours(lpbi);
 if(nNumColors){
   pPal = (LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   /* Fill in the palette entries from the DIB color table and
    * create a logical color palette. */
   for (i = 0; i < nNumColors; i++){
       pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
       pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
       pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 else if (lpbi->biBitCount == 24){
   /* A 24 bitcount DIB has no color table entries so, set the number of
    * to the maximum value (256). */
   nNumColors = MAXPALETTE;
   pPal=(LOGPALETTE*)LocalAlloc(LPTR,sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
   if(pPal == NULL)return NULL;
   pPal->palNumEntries = nNumColors;
   pPal->palVersion    = PALVERSION;
   red = green = blue = 0;
   /* Generate 256 (= 8*8*4) RGB combinations to fill the palette entries. */
   for (i = 0; i < pPal->palNumEntries; i++){
       pPal->palPalEntry[i].peRed   = red;
       pPal->palPalEntry[i].peGreen = green;
       pPal->palPalEntry[i].peBlue  = blue;
       pPal->palPalEntry[i].peFlags = (BYTE)0;
       if (!(red += 32))if (!(green += 32)) blue += 64;
   }
   hpal = CreatePalette(pPal);
   LocalFree((HANDLE)pPal);
 }
 return hpal;
}

static WORD NumColours(LPBITMAPINFOHEADER lpbi){
 WORD bits;
 if (lpbi->biClrUsed != 0)return (WORD)lpbi->biClrUsed;
 bits = lpbi->biBitCount;
 switch (bits){
  case 1:    return 2;
  case 4:    return 16;
  case 8:    return 256;
  default:   return 0; /* A 24 bitcount DIB has no color table */
 }
}

static void GetRealClientRect(HWND hwnd, PRECT lprc){
 /* take account of scollbars as well */
 DWORD dwStyle;
 dwStyle = GetWindowLong (hwnd, GWL_STYLE);
 GetClientRect (hwnd,lprc);
 if (dwStyle & WS_HSCROLL)lprc->bottom += GetSystemMetrics(SM_CYHSCROLL);
 if (dwStyle & WS_VSCROLL)lprc->right  += GetSystemMetrics(SM_CXVSCROLL);
}

#if 0
static void AppPaint(HWND hWnd,HDC hDC,int x,int y){
 HPALETTE hpalT;
 POINT Origin;
 HDC hDCbits;
 BITMAP bm;
 SetWindowOrgEx(hDC,x,y,&Origin);
 SetBkMode (hDC,TRANSPARENT);
 if(hpalCurrent != NULL){
   hpalT = SelectPalette(hDC,hpalCurrent,FALSE);
   RealizePalette (hDC);
   if(hbmCurrent != NULL){ /* compatible bit map exists; its quicker, use it */
     hDCbits = CreateCompatibleDC(hDC);
     GetObject(hbmCurrent,sizeof(BITMAP),(LPSTR)&bm);
     SelectObject(hDCbits,hbmCurrent);
     BitBlt(hDC,0,0,bm.bmWidth,bm.bmHeight,hDCbits,0,0,SRCCOPY);
     DeleteDC(hDCbits);
   }
   else {
     if(lpBitmapHeader != NULL){
       SetDIBitsToDevice(hDC,
                       0,0,
                       (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biWidth,
                       (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biHeight,
                       0,0,
                       0,
                       (WORD)((LPBITMAPINFOHEADER)lpBitmapHeader)->biHeight,
                       lpBitmapBits,
                       (LPBITMAPINFO)lpBitmapHeader,
                       DIB_RGB_COLORS);
     }
   }
   SelectPalette(hDC,hpalT,FALSE);
 }
}
#endif

static void AppPaint(HWND hWnd,HDC hDC,int x,int y){
 HPALETTE hpalT;
 POINT Origin;
 PDIB pdibOriginal;
 if(lpBitmapHeader == NULL)return;
 SetWindowOrgEx(hDC,x,y,&Origin);
 if(pdibHalftone != NULL){
   if (hpalApp){
     SelectPalette(hDC, hpalApp, FALSE);
     RealizePalette(hDC);
   }
   StretchDIBits(hDC,
      0, 0, DibWidth(pdibHalftone), DibHeight(pdibHalftone),
      0, 0, DibWidth(pdibHalftone), DibHeight(pdibHalftone),
      DibPtr(pdibHalftone), DibInfo(pdibHalftone),
      DIB_RGB_COLORS, SRCCOPY);
 }
 else{
   if(hpalCurrent){
     SelectPalette(hDC,hpalCurrent,FALSE);
     RealizePalette (hDC);
   }
   pdibOriginal=(PDIB)lpBitmapHeader;
   StretchDIBits(hDC,
      0, 0, DibWidth(pdibOriginal), DibHeight(pdibOriginal),
      0, 0, DibWidth(pdibOriginal), DibHeight(pdibOriginal),
      DibPtr(pdibOriginal), DibInfo(pdibOriginal),
                 DIB_RGB_COLORS, SRCCOPY);
  }
}



static HBITMAP rBitmapFromDib(LPSTR lpHeader,LPSTR lpBits,HPALETTE hpal){
 HDC       hdc;         /* make BITMAP from a memory DIB */
 HBITMAP   hbm;
 HPALETTE  hpalT;
 if (lpHeader == NULL || lpBits == NULL)return NULL;
 hdc = GetDC(NULL);
 if(hpal != NULL){
   hpalT = SelectPalette(hdc,hpal,FALSE);
   RealizePalette(hdc);
 }
 hbm = CreateDIBitmap(hdc,
                (LPBITMAPINFOHEADER)lpHeader,
                (LONG)CBM_INIT,
                lpBits,
                (LPBITMAPINFO)lpHeader,
                DIB_RGB_COLORS );
 if(hpalCurrent)SelectPalette(hdc,hpalT,FALSE);
 ReleaseDC(NULL,hdc);
 return hbm;
}

static void tga_display_line(unsigned char *prgb,
                             long wide, long line,
                             long upsidedown, long hi){
 register long l,x,k;
 if(upsidedown)l = 3L*(long)(line)*(long)(x4_size);
 else          l = 3L*(long)(hi - 1 - line)*(long)(x4_size);
 memcpy(pixels + l, prgb, (int)(wide*3));
}

static short R_getvariword(long n, FILE *f, long *vw){
 long c;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw = c;          if(n == 1)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 8);  if(n == 2)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 16); if(n == 3)return 1;
 if((c=(long)getc(f)) == EOF)return 0;
 *vw |= (c << 24);
 return 1;
}

static short LoadTGA(char *filename, short info_only){
 char c0,c1,c2;
 long i,j,k,l,m,n,cmo,cml,xor,yor,wi,hi,iok=0,idsize,cm_type,vw,
      image_type,cmes,pixel_size,image_descriptor,n_attr,upsidedown,b_size,
      s1,s2,s,pat,wi3;
 FILE *t24;
 unsigned char *prgb;
 prgb=NULL;
 if((t24=fopen(filename,"rb")) != NULL){
   if(! R_getvariword(1,t24,&idsize))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cm_type))goto CLOSEOUT; /* 0 for non mapped */
   if(! R_getvariword(1,t24,&image_type))goto CLOSEOUT; /* image type */
   if(image_type != 2 && image_type != 10)goto CLOSEOUT;
   if(! R_getvariword(2,t24,&cmo))goto CLOSEOUT;  /* color map 5 bytes */
   if(! R_getvariword(2,t24,&cml))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&cmes))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&xor))goto CLOSEOUT;  /* Image spec 10 bytes */
   if(! R_getvariword(2,t24,&yor))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&wi))goto CLOSEOUT;
   if(! R_getvariword(2,t24,&hi))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&pixel_size))goto CLOSEOUT;
   if(! R_getvariword(1,t24,&image_descriptor))goto CLOSEOUT;
   if(idsize > 0)for(i=0;i<idsize;i++)if(getc(t24) == EOF)goto CLOSEOUT;
   if(cm_type == 1 && cml > 0){
     for(i=cmo;i<cml;i++)if(! R_getvariword(cmes/8,t24,&j))goto CLOSEOUT;
   }
   n_attr = image_descriptor & 0xf;
   upsidedown = (image_descriptor & 0x20) > 0 ? 0 : 1 ;
   if(info_only){
     x_size=(int)wi;
     y_size=(int)hi;
     fclose(t24);
     return 1;
   }
   if((prgb=(unsigned char *)malloc((size_t)wi*3)) == NULL)goto CLOSEOUT;
   if((image_descriptor & 0xc0) > 0)goto CLOSEOUT; /* interlaced */
   b_size=i=wi*hi;
   if     (pixel_size == 16){n=2;s1=5;s2=10;s=3;pat=0x1f;}
   else if(pixel_size == 24){n=3;s1=8;s2=16;s=0;pat=0xff;}
   else                     {n=4;s1=8;s2=16;s=0;pat=0xff;}
   wi3=wi*3;
   if(image_type == 2){ /* uncompressed */
     m=0; k=0; for(i=0;i<hi;i++)for(j=0;j<wi;j++){
       if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
       prgb[m  ] = (char)( vw        & pat) << s;
       prgb[m+1] = (char)((vw >> s1) & pat) << s;
       prgb[m+2] = (char)((vw >> s2) & pat) << s;
       k++; m+=3;
       if(m == wi3){
         tga_display_line(prgb,wi,i,upsidedown,hi); m=0;
       }
     }
   }
   else{  /* compressed */
     m=0; l=0; k=0; while(k < b_size){
       i=getc(t24); if(feof(t24))goto CLOSEOUT;
        if(i < 128){
         for(j=0;j<i+1;j++){
           if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
           prgb[m  ] = (char)( vw        & pat) << s;
           prgb[m+1] = (char)((vw >> s1) & pat) << s;
           prgb[m+2] = (char)((vw >> s2) & pat) << s;
           k++; m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
       else{
         if(! R_getvariword(n,t24,&vw))goto CLOSEOUT;
         c0 = (char)( vw        & pat) << s;
         c1 = (char)((vw >> s1) & pat) << s;
         c2 = (char)((vw >> s2) & pat) << s;
         i -= 128;
         for(j=0;j<i+1;j++){
           prgb[m  ] = c0;
           prgb[m+1] = c1;
           prgb[m+2] = c2;
           k++;  m+=3;
           if(m == wi3){
             tga_display_line(prgb,wi,l,upsidedown,hi); m=0; l++;
           }
         }
       }
     }
   }
   iok=1;
   CLOSEOUT:
   fclose(t24);
   if(prgb != NULL)free((char *)prgb);
 }
 return (short)iok;
}

int tgaview(HWND hWnd){
 int i;
 long imagesize;
 LPSTR p;
 LPBITMAPINFOHEADER lpbi;
 if(LoadTGA(CurrentFile,1) == 0){
   ErrMsg("'%ls' is NOT a valid TGA file type",CurrentFile);
   return 0;
 }
 x4_size=ALIGNULONG(x_size);
 FreeDib();
 lpBitmapHeader = (LPSTR)malloc((long)sizeof(BITMAPINFOHEADER));
 if(lpBitmapHeader == NULL){
   ErrMsg("Out of memory");
   return 0;
 }
 lpbi=(LPBITMAPINFOHEADER)lpBitmapHeader;
 lpbi->biSize=sizeof(BITMAPINFOHEADER);
 lpbi->biWidth=(DWORD)x4_size;
 lpbi->biHeight=(DWORD)y_size;
 lpbi->biPlanes=1;
 lpbi->biBitCount=24;
 lpbi->biCompression=BI_RGB;
 imagesize = lpbi->biSizeImage = (DWORD)x4_size*(DWORD)y_size*3L;
 lpbi->biXPelsPerMeter=0;
 lpbi->biYPelsPerMeter=0;
 lpbi->biClrUsed=0;
 lpbi->biClrImportant=0;
 hpalCurrent=CreateImagePalette(lpbi);  /* will set up a dummy palette */
 p=(LPSTR)realloc(lpBitmapHeader,(long)sizeof(BITMAPINFOHEADER)+imagesize);
 if(p == NULL){
   ErrMsg("Out of memory");
   FreeDib();
   return 0;
 }
 lpBitmapHeader=p;
 lpBitmapBits=(LPSTR)(p+(long)sizeof(BITMAPINFOHEADER));
 memset(lpBitmapBits,0,imagesize);
 pixels=lpBitmapBits;
 if(LoadTGA(CurrentFile,0) == 0){
   ErrMsg("Error reading TGA file");
   FreeDib();
   return 0;
 }
 if(DisplayBits < 16)pdibHalftone = DibHalftoneDIB((PDIB)lpBitmapHeader);
// hbmCurrent=rBitmapFromDib(lpBitmapHeader,lpBitmapBits,hpalCurrent);
// if(hbmCurrent == NULL)ErrMsg("Could not create bitmap");
 SizeWindow(hWnd);
 return 1;
}

/***************************************************************************
  DibHalftoneDIB

  Halftone a 24-bit DIB to the 8-bit WinG Halftone Palette
*/

//*** These tables are defined in httables.c
extern char unsigned aDividedBy51Rounded[256];
extern char unsigned aDividedBy51[256];
extern char unsigned aModulo51[256];
extern char unsigned aTimes6[6];
extern char unsigned aTimes36[6];

extern char unsigned aHalftone8x8[64];

extern char unsigned aWinGHalftoneTranslation[216];

PDIB DibHalftoneDIB( PDIB pSource ){
  //*** Create an 8-bit DIB to halftone to
  PDIB pDestination = DibCreate(8,DibWidth(pSource),DibHeight(pSource));

  //*** Only work on 24-bit sources
  if(pDestination && hpalApp && (DibBitCount(pSource) == 24))
  {
    long PixelsPerScanline = DibWidth(pSource);
    int Scanline;

    //*** Fill in the DIB color table with the halftone palette
    DibSetUsage(pDestination, hpalApp, DIB_RGB_COLORS);

    //*** Step through, converting each pixel in each scan line
    //*** to the nearest halftone match
    for(Scanline = 0;Scanline < (int)DibHeight(pSource);Scanline++)
    {
      char unsigned *pSourceScanline =
          (char unsigned *)DibXY(pSource,0,Scanline);

      char unsigned  *pDestinationScanline =
          (char unsigned *)DibXY(pDestination,0,Scanline);

      int Pixel;

      for(Pixel = 0;Pixel < (int)DibWidth(pSource);Pixel++)
      {
        //*** This is the meat of the halftoning algorithm:
        //*** Convert an RGB into an index into the halftone palette.

        //*** First, extract the raw RGB information
        int Red = pSourceScanline[Pixel*3 + 2];
        int Green = pSourceScanline[Pixel*3 + 1];
        int Blue = pSourceScanline[Pixel*3];

        //*** Now, look up each value in the halftone matrix
        //*** using an 8x8 ordered dither.
        char unsigned RedTemp = aDividedBy51[Red]
          + (aModulo51[Red] > aHalftone8x8[(Pixel%8)*8
          + Scanline%8]);
        char unsigned GreenTemp = aDividedBy51[(char unsigned)Green]
          + (aModulo51[Green] > aHalftone8x8[
          (Pixel%8)*8 + Scanline%8]);
        char unsigned BlueTemp = aDividedBy51[(char unsigned)Blue]
          + (aModulo51[Blue] > aHalftone8x8[
          (Pixel%8)*8 +Scanline%8]);

        //*** Recombine the halftoned RGB values into a palette index
        char unsigned PaletteIndex =
          RedTemp + aTimes6[GreenTemp] + aTimes36[BlueTemp];

        //*** And translate through the WinG Halftone Palette
        //*** translation vector to give the correct value.
        pDestinationScanline[Pixel] = aWinGHalftoneTranslation[PaletteIndex];
      }
    }
  }
  else
  {
    //*** Punt on anything but 24bpp
    DibFree(pDestination);
    pDestination = 0;
  }

  return pDestination;
}
